diff -pruN 4.3.0-2/.ci/ci.yaml 4.4.0-1/.ci/ci.yaml
--- 4.3.0-2/.ci/ci.yaml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.ci/ci.yaml	2024-05-28 20:08:34.000000000 +0000
@@ -15,7 +15,9 @@ jobs:
   strategy:
     matrix:
       Plain:
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF'
+        # Precompiled headers with debug symbols can cause too much disk space /
+        # RAM usage.
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF -DUSE_PRECOMPILED_HEADERS=OFF'
       Python:
         binding: 'python'
         python.version: '3.7'
@@ -39,22 +41,24 @@ jobs:
     vmImage: macOS-latest
   strategy:
     matrix:
+      # clang on OS X segfaults when using precompiled headers, so we disable
+      # them.
       Plain:
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF'
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF -DUSE_PRECOMPILED_HEADERS=OFF'
         python.version: '3.8'
       Python:
         binding: 'python'
         python.version: '3.8'
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=ON -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF'
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=ON -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF -DUSE_PRECOMPILED_HEADERS=OFF'
       Julia:
         python.version: '3.8'
         julia.version: '1.6.3'
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_JULIA_BINDINGS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF'
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_JULIA_BINDINGS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF -DUSE_PRECOMPILED_HEADERS=OFF'
       Go:
         binding: 'go'
         python.version: '3.8'
         go.version: '1.11.0'
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=ON -DBUILD_R_BINDINGS=OFF'
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=ON -DBUILD_R_BINDINGS=OFF -DUSE_PRECOMPILED_HEADERS=OFF'
 
   steps:
   - template: macos-steps.yaml
@@ -67,7 +71,7 @@ jobs:
   strategy:
     matrix:
       Plain:
-        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF -DCOTIRE_MINIMUM_NUMBER_OF_TARGET_SOURCES=100000'
+        CMakeArgs: '-DDEBUG=ON -DPROFILE=OFF -DBUILD_TESTS=ON -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=OFF'
         python.version: '2.7'
         CMakeGenerator: '-G "Visual Studio 17 2022"'
         MSBuildVersion: '17.0'
diff -pruN 4.3.0-2/.ci/macos-steps.yaml 4.4.0-1/.ci/macos-steps.yaml
--- 4.3.0-2/.ci/macos-steps.yaml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.ci/macos-steps.yaml	2024-05-28 20:08:34.000000000 +0000
@@ -13,7 +13,6 @@ steps:
 - script: |
     set -e
     sudo xcode-select --switch /Applications/Xcode.app/Contents/Developer
-    unset BOOST_ROOT
     brew install libomp openblas armadillo cereal ensmallen
 
     if [ "$(binding)" == "python" ]; then
@@ -29,7 +28,6 @@ steps:
 
 # Configure mlpack (CMake)
 - script: |
-    unset BOOST_ROOT
     mkdir build && cd build
     if [ "$(binding)" == "go" ]; then
       export GOPATH=$PWD/src/mlpack/bindings/go
diff -pruN 4.3.0-2/.ci/windows-steps.yaml 4.4.0-1/.ci/windows-steps.yaml
--- 4.3.0-2/.ci/windows-steps.yaml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.ci/windows-steps.yaml	2024-05-28 20:08:34.000000000 +0000
@@ -9,7 +9,6 @@ steps:
 # Fetch build dependencies
 - powershell: |
     nuget install OpenBLAS -o $(Agent.ToolsDirectory)
-    nuget install unofficial-flayan-cereal -o $(Agent.ToolsDirectory)
     nuget install ensmallen -o $(Agent.ToolsDirectory) -Version 2.17.0
     ## Delete all ensmallen dependencies including armadillo headers, we do not need them here
     Remove-Item $(Agent.ToolsDirectory)\ensmallen.2.17.0\installed\x64-linux\share -Force -Recurse
@@ -18,6 +17,12 @@ steps:
 
   displayName: 'Fetch build dependencies'
 
+# Get cereal
+- bash: |
+    curl -L https://github.com/USCiLab/cereal/archive/refs/tags/v1.3.2.tar.gz -o cereal-1.3.2.tar.gz
+    tar -xvzf cereal-1.3.2.tar.gz
+  displayName: 'Download cereal'
+
 # Configure armadillo
 - bash: |
     git clone --depth 1 https://github.com/mlpack/jenkins-conf.git conf
@@ -57,7 +62,7 @@ steps:
     -DLAPACK_LIBRARIES:FILEPATH=$(Agent.ToolsDirectory)\OpenBLAS.0.2.14.1\lib\native\lib\x64\libopenblas.dll.a `
     -DARMADILLO_INCLUDE_DIR="..\armadillo-9.800.6\tmp\include" `
     -DARMADILLO_LIBRARY="..\armadillo-9.800.6\Release\armadillo.lib" `
-    -DCEREAL_INCLUDE_DIR=$(Agent.ToolsDirectory)\unofficial-flayan-cereal.1.2.2\build\native\include `
+    -DCEREAL_INCLUDE_DIR="..\cereal-1.3.2\include" `
     -DENSMALLEN_INCLUDE_DIR=$(Agent.ToolsDirectory)\ensmallen.2.17.0\installed\x64-linux\include `
     -DBUILD_JULIA_BINDINGS=OFF `
     -DCMAKE_BUILD_TYPE=Release ..
diff -pruN 4.3.0-2/.github/workflows/main.yml 4.4.0-1/.github/workflows/main.yml
--- 4.3.0-2/.github/workflows/main.yml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.github/workflows/main.yml	2024-05-28 20:08:34.000000000 +0000
@@ -24,76 +24,76 @@ jobs:
       r_bindings: ${{ steps.mlpack_version.outputs.mlpack_r_package }}
 
     steps:
-        - uses: actions/checkout@v3
+      - uses: actions/checkout@v3
 
-        - name: Extract mlpack version
-          id: mlpack_version
-          run: |
-            MLPACK_VERSION_MAJOR=$(grep -i ".*#define MLPACK_VERSION_MAJOR.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
-            MLPACK_VERSION_MINOR=$(grep -i ".*#define MLPACK_VERSION_MINOR.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
-            MLPACK_VERSION_PATCH=$(grep -i ".*#define MLPACK_VERSION_PATCH.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
-            MLPACK_VERSION_VALUE=${MLPACK_VERSION_MAJOR}.${MLPACK_VERSION_MINOR}.${MLPACK_VERSION_PATCH}
-            echo ::set-output name=mlpack_r_package::$(echo mlpack_"$MLPACK_VERSION_VALUE".tar.gz)
-
-        # Setup Pandoc 
-        - uses: r-lib/actions/setup-pandoc@v2
-
-        # Setup R actions
-        - uses: r-lib/actions/setup-r@v2
-
-        - name: Query dependencies
-          run: |
-            cp src/mlpack/bindings/R/mlpack/DESCRIPTION.in DESCRIPTION
-            Rscript -e "install.packages('remotes')" -e "saveRDS(remotes::dev_package_deps(dependencies = TRUE), 'depends.Rds')"
-
-        - name: Cache R packages
-          if: runner.os != 'Windows' && runner.os != 'macOS'
-          uses: actions/cache@v3
-          with:
-            path: ${{ env.R_LIBS_USER }}
-            key: ${{ runner.os }}-r-release-${{ hashFiles('depends.Rds') }}
-            restore-keys: ${{ runner.os }}-r-release-
-
-        - name: Install Build Dependencies
-          run: |
-            sudo apt-get update
-            # We don't install cereal via apt, because the Debian packagers
-            # split the rapidjson dependency into a separate package.  We will
-            # bundle the cereal sources with the R package, so we want them to
-            # be exactly the upstream sources (with rapidjson included).
-            sudo apt-get install -y --allow-unauthenticated libopenblas-dev liblapack-dev g++ libensmallen-dev libhdf5-dev libarmadillo-dev libcurl4-openssl-dev
-            wget https://github.com/USCiLab/cereal/archive/refs/tags/v1.3.2.tar.gz
-            tar -xvzpf v1.3.2.tar.gz
-            # These directives cause warnings on CRAN:
-            # https://github.com/USCiLab/cereal/blob/master/include/cereal/external/base64.hpp#L28-L31
-            # The command below comments them out.
-            sed -i 's|#pragma|// #pragma|' cereal-1.3.2/include/cereal/external/base64.hpp
-
-        - name: Install R-bindings dependencies
-          run: |
-            remotes::install_deps(dependencies = TRUE)
-            remotes::install_cran("roxygen2")
-            remotes::install_cran("pkgbuild")
-          shell: Rscript {0}
-
-        - name: CMake
-          run: |
-            mkdir build
-            cd build && cmake -DDEBUG=OFF -DPROFILE=OFF -DBUILD_CLI_EXECUTABLES=OFF -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=ON -DDOWNLOAD_DEPENDENCIES=ON -DBUILD_TESTS=ON -DCEREAL_INCLUDE_DIR=../cereal-1.3.2/include/ ..
-
-        - name: Build
-          run: |
-            cd build && make
-
-        - name: Run tests via ctest
-          run: |
-            cd build && CTEST_OUTPUT_ON_FAILURE=1 ctest -T Test .
-
-        - name: Upload R packages
-          uses: actions/upload-artifact@v3
-          with:
-              name: mlpack_r_tarball
-              path: build/src/mlpack/bindings/R/${{ steps.mlpack_version.outputs.mlpack_r_package }}
+      - name: Extract mlpack version
+        id: mlpack_version
+        run: |
+          MLPACK_VERSION_MAJOR=$(grep -i ".*#define MLPACK_VERSION_MAJOR.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
+          MLPACK_VERSION_MINOR=$(grep -i ".*#define MLPACK_VERSION_MINOR.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
+          MLPACK_VERSION_PATCH=$(grep -i ".*#define MLPACK_VERSION_PATCH.*" src/mlpack/core/util/version.hpp | grep -o "[0-9]*")
+          MLPACK_VERSION_VALUE=${MLPACK_VERSION_MAJOR}.${MLPACK_VERSION_MINOR}.${MLPACK_VERSION_PATCH}
+          echo "mlpack_r_package=$(echo mlpack_"$MLPACK_VERSION_VALUE".tar.gz)" >> $GITHUB_OUTPUT
+
+      # Setup Pandoc
+      - uses: r-lib/actions/setup-pandoc@v2
+
+      # Setup R actions
+      - uses: r-lib/actions/setup-r@v2
+
+      - name: Query dependencies
+        run: |
+          cp src/mlpack/bindings/R/mlpack/DESCRIPTION.in DESCRIPTION
+          Rscript -e "install.packages('remotes')" -e "saveRDS(remotes::dev_package_deps(dependencies = TRUE), 'depends.Rds')"
+
+      - name: Cache R packages
+        if: runner.os != 'Windows' && runner.os != 'macOS'
+        uses: actions/cache@v3
+        with:
+          path: ${{ env.R_LIBS_USER }}
+          key: ${{ runner.os }}-r-release-${{ hashFiles('depends.Rds') }}
+          restore-keys: ${{ runner.os }}-r-release-
+
+      - name: Install Build Dependencies
+        run: |
+          sudo apt-get update
+          # We don't install cereal via apt, because the Debian packagers
+          # split the rapidjson dependency into a separate package.  We will
+          # bundle the cereal sources with the R package, so we want them to
+          # be exactly the upstream sources (with rapidjson included).
+          sudo apt-get install -y --allow-unauthenticated libopenblas-dev liblapack-dev g++ libensmallen-dev libhdf5-dev libarmadillo-dev libcurl4-openssl-dev
+          wget https://github.com/USCiLab/cereal/archive/refs/tags/v1.3.2.tar.gz
+          tar -xvzpf v1.3.2.tar.gz
+          # These directives cause warnings on CRAN:
+          # https://github.com/USCiLab/cereal/blob/master/include/cereal/external/base64.hpp#L28-L31
+          # The command below comments them out.
+          sed -i 's|#pragma|// #pragma|' cereal-1.3.2/include/cereal/external/base64.hpp
+
+      - name: Install R-bindings dependencies
+        run: |
+          remotes::install_deps(dependencies = TRUE)
+          remotes::install_cran("roxygen2")
+          remotes::install_cran("pkgbuild")
+        shell: Rscript {0}
+
+      - name: CMake
+        run: |
+          mkdir build
+          cd build && cmake -DDEBUG=OFF -DPROFILE=OFF -DBUILD_CLI_EXECUTABLES=OFF -DBUILD_PYTHON_BINDINGS=OFF -DBUILD_JULIA_BINDINGS=OFF -DBUILD_GO_BINDINGS=OFF -DBUILD_R_BINDINGS=ON -DDOWNLOAD_DEPENDENCIES=ON -DBUILD_TESTS=ON -DCEREAL_INCLUDE_DIR=../cereal-1.3.2/include/ ..
+
+      - name: Build
+        run: |
+          cd build && make
+
+      - name: Run tests via ctest
+        run: |
+          cd build && CTEST_OUTPUT_ON_FAILURE=1 ctest -T Test .
+
+      - name: Upload R packages
+        uses: actions/upload-artifact@v3
+        with:
+          name: mlpack_r_tarball
+          path: build/src/mlpack/bindings/R/${{ steps.mlpack_version.outputs.mlpack_r_package }}
 
   R-CMD-check:
     needs: jobR
@@ -106,9 +106,14 @@ jobs:
       fail-fast: false
       matrix:
         config:
-        - {os: windows-latest, r: 'release', name: 'Windows R'}
-        - {os: macOS-latest, r: 'release', name: 'macOS R'}
-        - {os: ubuntu-latest, r: 'devel', http-user-agent: 'release', name: 'Linux R'}
+          - { os: windows-latest, r: "release", name: "Windows R" }
+          - { os: macOS-latest, r: "release", name: "macOS R" }
+          - {
+              os: ubuntu-latest,
+              r: "devel",
+              http-user-agent: "release",
+              name: "Linux R",
+            }
 
     env:
       MAKEFLAGS: "-j 2"
@@ -146,14 +151,15 @@ jobs:
       - name: Install check dependencies
         if: runner.os != 'Windows' && runner.os != 'macOS'
         run: |
-            sudo apt-get update
-            sudo apt-get install -y --allow-unauthenticated libcurl4-openssl-dev
+          sudo apt-get update
+          sudo apt-get install -y --allow-unauthenticated libcurl4-openssl-dev
 
       - name: Install dependencies
         run: |
           install.packages('remotes')
           remotes::install_deps('${{ needs.jobR.outputs.r_bindings }}', dependencies = TRUE)
           remotes::install_cran("rcmdcheck")
+          remotes::install_cran("curl")
         shell: Rscript {0}
 
       - name: Check
diff -pruN 4.3.0-2/.github/workflows/update-catch.yaml 4.4.0-1/.github/workflows/update-catch.yaml
--- 4.3.0-2/.github/workflows/update-catch.yaml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.github/workflows/update-catch.yaml	2024-05-28 20:08:34.000000000 +0000
@@ -6,7 +6,7 @@ name: Update Catch
 on:
   workflow_dispatch:
   schedule:
-    - cron:  '0 10 1/16 * *'
+    - cron: "0 10 1/16 * *"
 permissions:
   contents: read
 
@@ -16,7 +16,7 @@ jobs:
       contents: write # for peter-evans/create-pull-request to create branch
       pull-requests: write # for peter-evans/create-pull-request to create a PR
     if: ${{ false }}
-#    if: ${{ github.repository == 'mlpack/mlpack' }}
+    #    if: ${{ github.repository == 'mlpack/mlpack' }}
     runs-on: ubuntu-latest
     steps:
       - uses: actions/checkout@v2
@@ -26,7 +26,7 @@ jobs:
           # Ping version information upstream.
           CATCH_RELEASE_JSON=$(curl -sL https://api.github.com/repos/catchorg/Catch2/releases/latest)
           CATCH_RELEASE_VERSION=$(jq -r ".tag_name" <<< "$CATCH_RELEASE_JSON" | tr -d v)
-          echo ::set-output name=release_tag::$(echo $CATCH_RELEASE_VERSION)
+          echo "release_tag=$(echo $CATCH_RELEASE_VERSION)" >> $GITHUB_OUTPUT
           # Extract out version information from git repository.
           CATCH_VERSION_MAJOR=$(grep -i ".*#define CATCH_VERSION_MAJOR.*" src/mlpack/tests/catch.hpp | grep -o "[0-9]*")
           CATCH_VERSION_MINOR=$(grep -i ".*#define CATCH_VERSION_MINOR.*" src/mlpack/tests/catch.hpp | grep -o "[0-9]*")
@@ -34,7 +34,7 @@ jobs:
           # Combine values to match release tag information.
           CATCH_VERSION_VALUE=${CATCH_VERSION_MAJOR}.${CATCH_VERSION_MINOR}.${CATCH_VERSION_PATCH}
           # Set the current release tag.
-          echo ::set-output name=current_tag::$(echo $CATCH_VERSION_VALUE)
+          echo "current_tag=$(echo $CATCH_VERSION_VALUE)" >> $GITHUB_OUTPUT
 
       - name: Update Catch
         if: steps.catch-header.outputs.current_tag != steps.catch-header.outputs.release_tag
diff -pruN 4.3.0-2/.github/workflows/update-cli11.yaml 4.4.0-1/.github/workflows/update-cli11.yaml
--- 4.3.0-2/.github/workflows/update-cli11.yaml	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/.github/workflows/update-cli11.yaml	2024-05-28 20:08:34.000000000 +0000
@@ -2,15 +2,15 @@ name: Update CLI11
 on:
   workflow_dispatch:
   schedule:
-    - cron:  '0 10 1/16 * *'
+    - cron: "0 10 1/16 * *"
 permissions:
   contents: read
 
 jobs:
   updateCLI11:
     permissions:
-      contents: write  # for peter-evans/create-pull-request to create branch
-      pull-requests: write  # for peter-evans/create-pull-request to create a PR
+      contents: write # for peter-evans/create-pull-request to create branch
+      pull-requests: write # for peter-evans/create-pull-request to create a PR
     if: ${{ github.repository == 'mlpack/mlpack' }}
     runs-on: ubuntu-latest
     steps:
@@ -21,11 +21,11 @@ jobs:
           # Ping version information upstream.
           CLI11_RELEASE_JSON=$(curl -sL https://api.github.com/repos/CLIUtils/CLI11/releases/latest)
           CLI11_RELEASE_VERSION=$(jq -r ".tag_name" <<< "$CLI11_RELEASE_JSON" | tr -d v)
-          echo ::set-output name=release_tag::$(echo $CLI11_RELEASE_VERSION)
+          echo "release_tag=$(echo $CLI11_RELEASE_VERSION)" >> $GITHUB_OUTPUT
           # Extract out version information from git repository.
           CLI11_VERSION_VALUE=$(grep -i ".*#define CLI11_VERSION.*" src/mlpack/bindings/cli/third_party/CLI/CLI11.hpp | grep -Po "(\d+\.)+\d+")
           # Set the current release tag.
-          echo ::set-output name=current_tag::$(echo $CLI11_VERSION_VALUE)
+          echo "current_tag=$(echo $CLI11_VERSION_VALUE)" >> $GITHUB_OUTPUT
 
       - name: Update CLI11
         if: steps.cli11-header.outputs.current_tag != steps.cli11-header.outputs.release_tag
diff -pruN 4.3.0-2/CMake/ConfigureFileOnly.cmake 4.4.0-1/CMake/ConfigureFileOnly.cmake
--- 4.3.0-2/CMake/ConfigureFileOnly.cmake	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/CMake/ConfigureFileOnly.cmake	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,11 @@
+# ConfigureFileOnly.cmake: generate an mlpack binding file given input
+# arguments, using the @ONLY option.
+#
+# This file depends on the following variables being set:
+#
+#  * GENERATE_CPP_IN: the .cpp.in file to configure.
+#  * GENERATE_CPP_OUT: the .cpp file we'll generate.
+#
+# Any other defined variables will be passed on to the file that is being
+# generated.
+configure_file(${GENERATE_CPP_IN} ${GENERATE_CPP_OUT} @ONLY)
diff -pruN 4.3.0-2/CMake/FindStbImage.cmake 4.4.0-1/CMake/FindStbImage.cmake
--- 4.3.0-2/CMake/FindStbImage.cmake	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/CMake/FindStbImage.cmake	2024-05-28 20:08:34.000000000 +0000
@@ -16,7 +16,6 @@ find_path(STB_IMAGE_INCLUDE_DIR_1
     PATHS ${STB_IMAGE_SEARCH_PATHS} ${STB_IMAGE_INCLUDE_DIR})
 
 if(STB_IMAGE_INCLUDE_DIR_1)
-  set(STB_IMAGE_FOUND YES)
   set(STB_IMAGE_INCLUDE_DIR "${STB_IMAGE_INCLUDE_DIR_1}" CACHE PATH
       "stb_image include directory")
 
@@ -32,7 +31,6 @@ else ()
         PATH_SUFFIXES stb/)
 
   if(STB_IMAGE_INCLUDE_DIR_2)
-    set(STB_IMAGE_FOUND YES)
     set(STB_IMAGE_INCLUDE_DIR "${STB_IMAGE_INCLUDE_DIR_2}" CACHE PATH
         "stb_image include directory")
 
@@ -47,7 +45,6 @@ endif ()
 
 # Checks 'REQUIRED'.
 include(FindPackageHandleStandardArgs)
-find_package_handle_standard_args(StbImage
-    REQUIRED_VARS STB_IMAGE_INCLUDE_DIR)
+find_package_handle_standard_args(StbImage REQUIRED_VARS STB_IMAGE_INCLUDE_DIR)
 
 mark_as_advanced(STB_IMAGE_INCLUDE_DIR)
diff -pruN 4.3.0-2/CMake/RunProgram.cmake 4.4.0-1/CMake/RunProgram.cmake
--- 4.3.0-2/CMake/RunProgram.cmake	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/CMake/RunProgram.cmake	2024-05-28 20:08:34.000000000 +0000
@@ -5,7 +5,7 @@
 #
 #   PROGRAM: the program to run to.
 #   OUTPUT_FILE: the file to store the output in.
-execute_process(COMMAND ${PROGRAM} OUTPUT_FILE ${OUTPUT_FILE}
+execute_process(COMMAND ${PROGRAM} ${PROGRAM_OPT1} ${PROGRAM_OPT2} OUTPUT_FILE ${OUTPUT_FILE}
     ERROR_VARIABLE err)
 
 if (err)
diff -pruN 4.3.0-2/CMake/cotire.cmake 4.4.0-1/CMake/cotire.cmake
--- 4.3.0-2/CMake/cotire.cmake	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/CMake/cotire.cmake	1970-01-01 00:00:00.000000000 +0000
@@ -1,4054 +0,0 @@
-# - cotire (compile time reducer)
-#
-# See the cotire manual for usage hints.
-#
-#=============================================================================
-# Copyright 2012-2017 Sascha Kratky
-#
-# Permission is hereby granted, free of charge, to any person
-# obtaining a copy of this software and associated documentation
-# files (the "Software"), to deal in the Software without
-# restriction, including without limitation the rights to use,
-# copy, modify, merge, publish, distribute, sublicense, and/or sell
-# copies of the Software, and to permit persons to whom the
-# Software is furnished to do so, subject to the following
-# conditions:
-#
-# The above copyright notice and this permission notice shall be
-# included in all copies or substantial portions of the Software.
-#
-# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-# OTHER DEALINGS IN THE SOFTWARE.
-#=============================================================================
-
-if(__COTIRE_INCLUDED)
-    return()
-endif()
-set(__COTIRE_INCLUDED TRUE)
-
-# call cmake_minimum_required, but prevent modification of the CMake policy stack in include mode
-# cmake_minimum_required also sets the policy version as a side effect, which we have to avoid
-if (NOT CMAKE_SCRIPT_MODE_FILE)
-    cmake_policy(PUSH)
-endif()
-cmake_minimum_required(VERSION 2.8.12)
-if (NOT CMAKE_SCRIPT_MODE_FILE)
-    cmake_policy(POP)
-endif()
-
-set (COTIRE_CMAKE_MODULE_FILE "${CMAKE_CURRENT_LIST_FILE}")
-set (COTIRE_CMAKE_MODULE_VERSION "1.7.10")
-
-# activate select policies
-if (POLICY CMP0025)
-    # Compiler id for Apple Clang is now AppleClang
-    cmake_policy(SET CMP0025 NEW)
-endif()
-
-if (POLICY CMP0026)
-    # disallow use of the LOCATION target property
-    cmake_policy(SET CMP0026 NEW)
-endif()
-
-if (POLICY CMP0038)
-    # targets may not link directly to themselves
-    cmake_policy(SET CMP0038 NEW)
-endif()
-
-if (POLICY CMP0039)
-    # utility targets may not have link dependencies
-    cmake_policy(SET CMP0039 NEW)
-endif()
-
-if (POLICY CMP0040)
-    # target in the TARGET signature of add_custom_command() must exist
-    cmake_policy(SET CMP0040 NEW)
-endif()
-
-if (POLICY CMP0045)
-    # error on non-existent target in get_target_property
-    cmake_policy(SET CMP0045 NEW)
-endif()
-
-if (POLICY CMP0046)
-    # error on non-existent dependency in add_dependencies
-    cmake_policy(SET CMP0046 NEW)
-endif()
-
-if (POLICY CMP0049)
-    # do not expand variables in target source entries
-    cmake_policy(SET CMP0049 NEW)
-endif()
-
-if (POLICY CMP0050)
-    # disallow add_custom_command SOURCE signatures
-    cmake_policy(SET CMP0050 NEW)
-endif()
-
-if (POLICY CMP0051)
-    # include TARGET_OBJECTS expressions in a target's SOURCES property
-    cmake_policy(SET CMP0051 NEW)
-endif()
-
-if (POLICY CMP0053)
-    # simplify variable reference and escape sequence evaluation
-    cmake_policy(SET CMP0053 NEW)
-endif()
-
-if (POLICY CMP0054)
-    # only interpret if() arguments as variables or keywords when unquoted
-    cmake_policy(SET CMP0054 NEW)
-endif()
-
-if (POLICY CMP0055)
-    # strict checking for break() command
-    cmake_policy(SET CMP0055 NEW)
-endif()
-
-include(CMakeParseArguments)
-include(ProcessorCount)
-
-function (cotire_get_configuration_types _configsVar)
-    set (_configs "")
-    if (CMAKE_CONFIGURATION_TYPES)
-        list (APPEND _configs ${CMAKE_CONFIGURATION_TYPES})
-    endif()
-    if (CMAKE_BUILD_TYPE)
-        list (APPEND _configs "${CMAKE_BUILD_TYPE}")
-    endif()
-    if (_configs)
-        list (REMOVE_DUPLICATES _configs)
-        set (${_configsVar} ${_configs} PARENT_SCOPE)
-    else()
-        set (${_configsVar} "None" PARENT_SCOPE)
-    endif()
-endfunction()
-
-function (cotire_get_source_file_extension _sourceFile _extVar)
-    # get_filename_component returns extension from first occurrence of . in file name
-    # this function computes the extension from last occurrence of . in file name
-    string (FIND "${_sourceFile}" "." _index REVERSE)
-    if (_index GREATER -1)
-        math (EXPR _index "${_index} + 1")
-        string (SUBSTRING "${_sourceFile}" ${_index} -1 _sourceExt)
-    else()
-        set (_sourceExt "")
-    endif()
-    set (${_extVar} "${_sourceExt}" PARENT_SCOPE)
-endfunction()
-
-macro (cotire_check_is_path_relative_to _path _isRelativeVar)
-    set (${_isRelativeVar} FALSE)
-    if (IS_ABSOLUTE "${_path}")
-        foreach (_dir ${ARGN})
-            file (RELATIVE_PATH _relPath "${_dir}" "${_path}")
-            if (NOT _relPath OR (NOT IS_ABSOLUTE "${_relPath}" AND NOT "${_relPath}" MATCHES "^\\.\\."))
-                set (${_isRelativeVar} TRUE)
-                break()
-            endif()
-        endforeach()
-    endif()
-endmacro()
-
-function (cotire_filter_language_source_files _language _target _sourceFilesVar _excludedSourceFilesVar _cotiredSourceFilesVar)
-    if (CMAKE_${_language}_SOURCE_FILE_EXTENSIONS)
-        set (_languageExtensions "${CMAKE_${_language}_SOURCE_FILE_EXTENSIONS}")
-    else()
-        set (_languageExtensions "")
-    endif()
-    if (CMAKE_${_language}_IGNORE_EXTENSIONS)
-        set (_ignoreExtensions "${CMAKE_${_language}_IGNORE_EXTENSIONS}")
-    else()
-        set (_ignoreExtensions "")
-    endif()
-    if (COTIRE_UNITY_SOURCE_EXCLUDE_EXTENSIONS)
-        set (_excludeExtensions "${COTIRE_UNITY_SOURCE_EXCLUDE_EXTENSIONS}")
-    else()
-        set (_excludeExtensions "")
-    endif()
-    if (COTIRE_DEBUG AND _languageExtensions)
-        message (STATUS "${_language} source file extensions: ${_languageExtensions}")
-    endif()
-    if (COTIRE_DEBUG AND _ignoreExtensions)
-        message (STATUS "${_language} ignore extensions: ${_ignoreExtensions}")
-    endif()
-    if (COTIRE_DEBUG AND _excludeExtensions)
-        message (STATUS "${_language} exclude extensions: ${_excludeExtensions}")
-    endif()
-    if (CMAKE_VERSION VERSION_LESS "3.1.0")
-        set (_allSourceFiles ${ARGN})
-    else()
-        # as of CMake 3.1 target sources may contain generator expressions
-        # since we cannot obtain required property information about source files added
-        # through generator expressions at configure time, we filter them out
-        string (GENEX_STRIP "${ARGN}" _allSourceFiles)
-    endif()
-    set (_filteredSourceFiles "")
-    set (_excludedSourceFiles "")
-    foreach (_sourceFile ${_allSourceFiles})
-        get_source_file_property(_sourceIsHeaderOnly "${_sourceFile}" HEADER_FILE_ONLY)
-        get_source_file_property(_sourceIsExternal "${_sourceFile}" EXTERNAL_OBJECT)
-        get_source_file_property(_sourceIsSymbolic "${_sourceFile}" SYMBOLIC)
-        if (NOT _sourceIsHeaderOnly AND NOT _sourceIsExternal AND NOT _sourceIsSymbolic)
-            cotire_get_source_file_extension("${_sourceFile}" _sourceExt)
-            if (_sourceExt)
-                list (FIND _ignoreExtensions "${_sourceExt}" _ignoreIndex)
-                if (_ignoreIndex LESS 0)
-                    list (FIND _excludeExtensions "${_sourceExt}" _excludeIndex)
-                    if (_excludeIndex GREATER -1)
-                        list (APPEND _excludedSourceFiles "${_sourceFile}")
-                    else()
-                        list (FIND _languageExtensions "${_sourceExt}" _sourceIndex)
-                        if (_sourceIndex GREATER -1)
-                            # consider source file unless it is excluded explicitly
-                            get_source_file_property(_sourceIsExcluded "${_sourceFile}" COTIRE_EXCLUDED)
-                            if (_sourceIsExcluded)
-                                list (APPEND _excludedSourceFiles "${_sourceFile}")
-                            else()
-                                list (APPEND _filteredSourceFiles "${_sourceFile}")
-                            endif()
-                        else()
-                            get_source_file_property(_sourceLanguage "${_sourceFile}" LANGUAGE)
-                            if ("${_sourceLanguage}" STREQUAL "${_language}")
-                                # add to excluded sources, if file is not ignored and has correct language without having the correct extension
-                                list (APPEND _excludedSourceFiles "${_sourceFile}")
-                            endif()
-                        endif()
-                    endif()
-                endif()
-            endif()
-        endif()
-    endforeach()
-    # separate filtered source files from already cotired ones
-    # the COTIRE_TARGET property of a source file may be set while a target is being processed by cotire
-    set (_sourceFiles "")
-    set (_cotiredSourceFiles "")
-    foreach (_sourceFile ${_filteredSourceFiles})
-        get_source_file_property(_sourceIsCotired "${_sourceFile}" COTIRE_TARGET)
-        if (_sourceIsCotired)
-            list (APPEND _cotiredSourceFiles "${_sourceFile}")
-        else()
-            get_source_file_property(_sourceCompileFlags "${_sourceFile}" COMPILE_FLAGS)
-            if (_sourceCompileFlags)
-                # add to excluded sources, if file has custom compile flags
-                list (APPEND _excludedSourceFiles "${_sourceFile}")
-            else()
-                list (APPEND _sourceFiles "${_sourceFile}")
-            endif()
-        endif()
-    endforeach()
-    if (COTIRE_DEBUG)
-        if (_sourceFiles)
-            message (STATUS "Filtered ${_target} ${_language} sources: ${_sourceFiles}")
-        endif()
-        if (_excludedSourceFiles)
-            message (STATUS "Excluded ${_target} ${_language} sources: ${_excludedSourceFiles}")
-        endif()
-        if (_cotiredSourceFiles)
-            message (STATUS "Cotired ${_target} ${_language} sources: ${_cotiredSourceFiles}")
-        endif()
-    endif()
-    set (${_sourceFilesVar} ${_sourceFiles} PARENT_SCOPE)
-    set (${_excludedSourceFilesVar} ${_excludedSourceFiles} PARENT_SCOPE)
-    set (${_cotiredSourceFilesVar} ${_cotiredSourceFiles} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_objects_with_property_on _filteredObjectsVar _property _type)
-    set (_filteredObjects "")
-    foreach (_object ${ARGN})
-        get_property(_isSet ${_type} "${_object}" PROPERTY ${_property} SET)
-        if (_isSet)
-            get_property(_propertyValue ${_type} "${_object}" PROPERTY ${_property})
-            if (_propertyValue)
-                list (APPEND _filteredObjects "${_object}")
-            endif()
-        endif()
-    endforeach()
-    set (${_filteredObjectsVar} ${_filteredObjects} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_objects_with_property_off _filteredObjectsVar _property _type)
-    set (_filteredObjects "")
-    foreach (_object ${ARGN})
-        get_property(_isSet ${_type} "${_object}" PROPERTY ${_property} SET)
-        if (_isSet)
-            get_property(_propertyValue ${_type} "${_object}" PROPERTY ${_property})
-            if (NOT _propertyValue)
-                list (APPEND _filteredObjects "${_object}")
-            endif()
-        endif()
-    endforeach()
-    set (${_filteredObjectsVar} ${_filteredObjects} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_file_property_values _valuesVar _property)
-    set (_values "")
-    foreach (_sourceFile ${ARGN})
-        get_source_file_property(_propertyValue "${_sourceFile}" ${_property})
-        if (_propertyValue)
-            list (APPEND _values "${_propertyValue}")
-        endif()
-    endforeach()
-    set (${_valuesVar} ${_values} PARENT_SCOPE)
-endfunction()
-
-function (cotire_resolve_config_properties _configurations _propertiesVar)
-    set (_properties "")
-    foreach (_property ${ARGN})
-        if ("${_property}" MATCHES "<CONFIG>")
-            foreach (_config ${_configurations})
-                string (TOUPPER "${_config}" _upperConfig)
-                string (REPLACE "<CONFIG>" "${_upperConfig}" _configProperty "${_property}")
-                list (APPEND _properties ${_configProperty})
-            endforeach()
-        else()
-            list (APPEND _properties ${_property})
-        endif()
-    endforeach()
-    set (${_propertiesVar} ${_properties} PARENT_SCOPE)
-endfunction()
-
-function (cotire_copy_set_properties _configurations _type _source _target)
-    cotire_resolve_config_properties("${_configurations}" _properties ${ARGN})
-    foreach (_property ${_properties})
-        get_property(_isSet ${_type} ${_source} PROPERTY ${_property} SET)
-        if (_isSet)
-            get_property(_propertyValue ${_type} ${_source} PROPERTY ${_property})
-            set_property(${_type} ${_target} PROPERTY ${_property} "${_propertyValue}")
-        endif()
-    endforeach()
-endfunction()
-
-function (cotire_get_target_usage_requirements _target _config _targetRequirementsVar)
-    set (_targetRequirements "")
-    get_target_property(_librariesToProcess ${_target} LINK_LIBRARIES)
-    while (_librariesToProcess)
-        # remove from head
-        list (GET _librariesToProcess 0 _library)
-        list (REMOVE_AT _librariesToProcess 0)
-        if (_library MATCHES "^\\$<\\$<CONFIG:${_config}>:([A-Za-z0-9_:-]+)>$")
-            set (_library "${CMAKE_MATCH_1}")
-        elseif (_config STREQUAL "None" AND _library MATCHES "^\\$<\\$<CONFIG:>:([A-Za-z0-9_:-]+)>$")
-            set (_library "${CMAKE_MATCH_1}")
-        endif()
-        if (TARGET ${_library})
-            list (FIND _targetRequirements ${_library} _index)
-            if (_index LESS 0)
-                list (APPEND _targetRequirements ${_library})
-                # BFS traversal of transitive libraries
-                get_target_property(_libraries ${_library} INTERFACE_LINK_LIBRARIES)
-                if (_libraries)
-                    list (APPEND _librariesToProcess ${_libraries})
-                    list (REMOVE_DUPLICATES _librariesToProcess)
-                endif()
-            endif()
-        endif()
-    endwhile()
-    set (${_targetRequirementsVar} ${_targetRequirements} PARENT_SCOPE)
-endfunction()
-
-function (cotire_filter_compile_flags _language _flagFilter _matchedOptionsVar _unmatchedOptionsVar)
-    if (WIN32 AND CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-        set (_flagPrefix "[/-]")
-    else()
-        set (_flagPrefix "--?")
-    endif()
-    set (_optionFlag "")
-    set (_matchedOptions "")
-    set (_unmatchedOptions "")
-    foreach (_compileFlag ${ARGN})
-        if (_compileFlag)
-            if (_optionFlag AND NOT "${_compileFlag}" MATCHES "^${_flagPrefix}")
-                # option with separate argument
-                list (APPEND _matchedOptions "${_compileFlag}")
-                set (_optionFlag "")
-            elseif ("${_compileFlag}" MATCHES "^(${_flagPrefix})(${_flagFilter})$")
-                # remember option
-                set (_optionFlag "${CMAKE_MATCH_2}")
-            elseif ("${_compileFlag}" MATCHES "^(${_flagPrefix})(${_flagFilter})(.+)$")
-                # option with joined argument
-                list (APPEND _matchedOptions "${CMAKE_MATCH_3}")
-                set (_optionFlag "")
-            else()
-                # flush remembered option
-                if (_optionFlag)
-                    list (APPEND _matchedOptions "${_optionFlag}")
-                    set (_optionFlag "")
-                endif()
-                # add to unfiltered options
-                list (APPEND _unmatchedOptions "${_compileFlag}")
-            endif()
-        endif()
-    endforeach()
-    if (_optionFlag)
-        list (APPEND _matchedOptions "${_optionFlag}")
-    endif()
-    if (COTIRE_DEBUG AND _matchedOptions)
-        message (STATUS "Filter ${_flagFilter} matched: ${_matchedOptions}")
-    endif()
-    if (COTIRE_DEBUG AND _unmatchedOptions)
-        message (STATUS "Filter ${_flagFilter} unmatched: ${_unmatchedOptions}")
-    endif()
-    set (${_matchedOptionsVar} ${_matchedOptions} PARENT_SCOPE)
-    set (${_unmatchedOptionsVar} ${_unmatchedOptions} PARENT_SCOPE)
-endfunction()
-
-function (cotire_is_target_supported _target _isSupportedVar)
-    if (NOT TARGET "${_target}")
-        set (${_isSupportedVar} FALSE PARENT_SCOPE)
-        return()
-    endif()
-    get_target_property(_imported ${_target} IMPORTED)
-    if (_imported)
-        set (${_isSupportedVar} FALSE PARENT_SCOPE)
-        return()
-    endif()
-    get_target_property(_targetType ${_target} TYPE)
-    if (NOT _targetType MATCHES "EXECUTABLE|(STATIC|SHARED|MODULE|OBJECT)_LIBRARY")
-        set (${_isSupportedVar} FALSE PARENT_SCOPE)
-        return()
-    endif()
-    set (${_isSupportedVar} TRUE PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_target_compile_flags _config _language _target _flagsVar)
-    string (TOUPPER "${_config}" _upperConfig)
-    # collect options from CMake language variables
-    set (_compileFlags "")
-    if (CMAKE_${_language}_FLAGS)
-        set (_compileFlags "${_compileFlags} ${CMAKE_${_language}_FLAGS}")
-    endif()
-    if (CMAKE_${_language}_FLAGS_${_upperConfig})
-        set (_compileFlags "${_compileFlags} ${CMAKE_${_language}_FLAGS_${_upperConfig}}")
-    endif()
-    if (_target)
-        # add target compile flags
-        get_target_property(_targetflags ${_target} COMPILE_FLAGS)
-        if (_targetflags)
-            set (_compileFlags "${_compileFlags} ${_targetflags}")
-        endif()
-    endif()
-    if (UNIX)
-        separate_arguments(_compileFlags UNIX_COMMAND "${_compileFlags}")
-    elseif(WIN32)
-        separate_arguments(_compileFlags WINDOWS_COMMAND "${_compileFlags}")
-    else()
-        separate_arguments(_compileFlags)
-    endif()
-    # target compile options
-    if (_target)
-        get_target_property(_targetOptions ${_target} COMPILE_OPTIONS)
-        if (_targetOptions)
-            list (APPEND _compileFlags ${_targetOptions})
-        endif()
-    endif()
-    # interface compile options from linked library targets
-    if (_target)
-        set (_linkedTargets "")
-        cotire_get_target_usage_requirements(${_target} ${_config} _linkedTargets)
-        foreach (_linkedTarget ${_linkedTargets})
-            get_target_property(_targetOptions ${_linkedTarget} INTERFACE_COMPILE_OPTIONS)
-            if (_targetOptions)
-                list (APPEND _compileFlags ${_targetOptions})
-            endif()
-        endforeach()
-    endif()
-    # handle language standard properties
-    if (CMAKE_${_language}_STANDARD_DEFAULT)
-        # used compiler supports language standard levels
-        if (_target)
-            get_target_property(_targetLanguageStandard ${_target} ${_language}_STANDARD)
-            if (_targetLanguageStandard)
-                set (_type "EXTENSION")
-                get_property(_isSet TARGET ${_target} PROPERTY ${_language}_EXTENSIONS SET)
-                if (_isSet)
-                    get_target_property(_targetUseLanguageExtensions ${_target} ${_language}_EXTENSIONS)
-                    if (NOT _targetUseLanguageExtensions)
-                        set (_type "STANDARD")
-                    endif()
-                endif()
-                if (CMAKE_${_language}${_targetLanguageStandard}_${_type}_COMPILE_OPTION)
-                    list (APPEND _compileFlags "${CMAKE_${_language}${_targetLanguageStandard}_${_type}_COMPILE_OPTION}")
-                endif()
-            endif()
-        endif()
-    endif()
-    # handle the POSITION_INDEPENDENT_CODE target property
-    if (_target)
-        get_target_property(_targetPIC ${_target} POSITION_INDEPENDENT_CODE)
-        if (_targetPIC)
-            get_target_property(_targetType ${_target} TYPE)
-            if (_targetType STREQUAL "EXECUTABLE" AND CMAKE_${_language}_COMPILE_OPTIONS_PIE)
-                list (APPEND _compileFlags "${CMAKE_${_language}_COMPILE_OPTIONS_PIE}")
-            elseif (CMAKE_${_language}_COMPILE_OPTIONS_PIC)
-                list (APPEND _compileFlags "${CMAKE_${_language}_COMPILE_OPTIONS_PIC}")
-            endif()
-        endif()
-    endif()
-    # handle visibility target properties
-    if (_target)
-        get_target_property(_targetVisibility ${_target} ${_language}_VISIBILITY_PRESET)
-        if (_targetVisibility AND CMAKE_${_language}_COMPILE_OPTIONS_VISIBILITY)
-            list (APPEND _compileFlags "${CMAKE_${_language}_COMPILE_OPTIONS_VISIBILITY}${_targetVisibility}")
-        endif()
-        get_target_property(_targetVisibilityInlines ${_target} VISIBILITY_INLINES_HIDDEN)
-        if (_targetVisibilityInlines AND CMAKE_${_language}_COMPILE_OPTIONS_VISIBILITY_INLINES_HIDDEN)
-            list (APPEND _compileFlags "${CMAKE_${_language}_COMPILE_OPTIONS_VISIBILITY_INLINES_HIDDEN}")
-        endif()
-    endif()
-    # platform specific flags
-    if (APPLE)
-        get_target_property(_architectures ${_target} OSX_ARCHITECTURES_${_upperConfig})
-        if (NOT _architectures)
-            get_target_property(_architectures ${_target} OSX_ARCHITECTURES)
-        endif()
-        if (_architectures)
-            foreach (_arch ${_architectures})
-                list (APPEND _compileFlags "-arch" "${_arch}")
-            endforeach()
-        endif()
-        if (CMAKE_OSX_SYSROOT)
-            if (CMAKE_${_language}_SYSROOT_FLAG)
-                list (APPEND _compileFlags "${CMAKE_${_language}_SYSROOT_FLAG}" "${CMAKE_OSX_SYSROOT}")
-            else()
-                list (APPEND _compileFlags "-isysroot" "${CMAKE_OSX_SYSROOT}")
-            endif()
-        endif()
-        if (CMAKE_OSX_DEPLOYMENT_TARGET)
-            if (CMAKE_${_language}_OSX_DEPLOYMENT_TARGET_FLAG)
-                list (APPEND _compileFlags "${CMAKE_${_language}_OSX_DEPLOYMENT_TARGET_FLAG}${CMAKE_OSX_DEPLOYMENT_TARGET}")
-            else()
-                list (APPEND _compileFlags "-mmacosx-version-min=${CMAKE_OSX_DEPLOYMENT_TARGET}")
-            endif()
-        endif()
-    endif()
-    if (COTIRE_DEBUG AND _compileFlags)
-        message (STATUS "Target ${_target} compile flags: ${_compileFlags}")
-    endif()
-    set (${_flagsVar} ${_compileFlags} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_target_include_directories _config _language _target _includeDirsVar _systemIncludeDirsVar)
-    set (_includeDirs "")
-    set (_systemIncludeDirs "")
-    # default include dirs
-    if (CMAKE_INCLUDE_CURRENT_DIR)
-        list (APPEND _includeDirs "${CMAKE_CURRENT_BINARY_DIR}")
-        list (APPEND _includeDirs "${CMAKE_CURRENT_SOURCE_DIR}")
-    endif()
-    set (_targetFlags "")
-    cotire_get_target_compile_flags("${_config}" "${_language}" "${_target}" _targetFlags)
-    # parse additional include directories from target compile flags
-    if (CMAKE_INCLUDE_FLAG_${_language})
-        string (STRIP "${CMAKE_INCLUDE_FLAG_${_language}}" _includeFlag)
-        string (REGEX REPLACE "^[-/]+" "" _includeFlag "${_includeFlag}")
-        if (_includeFlag)
-            set (_dirs "")
-            cotire_filter_compile_flags("${_language}" "${_includeFlag}" _dirs _ignore ${_targetFlags})
-            if (_dirs)
-                list (APPEND _includeDirs ${_dirs})
-            endif()
-        endif()
-    endif()
-    # parse additional system include directories from target compile flags
-    if (CMAKE_INCLUDE_SYSTEM_FLAG_${_language})
-        string (STRIP "${CMAKE_INCLUDE_SYSTEM_FLAG_${_language}}" _includeFlag)
-        string (REGEX REPLACE "^[-/]+" "" _includeFlag "${_includeFlag}")
-        if (_includeFlag)
-            set (_dirs "")
-            cotire_filter_compile_flags("${_language}" "${_includeFlag}" _dirs _ignore ${_targetFlags})
-            if (_dirs)
-                list (APPEND _systemIncludeDirs ${_dirs})
-            endif()
-        endif()
-    endif()
-    # target include directories
-    get_directory_property(_dirs DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}" INCLUDE_DIRECTORIES)
-    if (_target)
-        get_target_property(_targetDirs ${_target} INCLUDE_DIRECTORIES)
-        if (_targetDirs)
-            list (APPEND _dirs ${_targetDirs})
-        endif()
-        get_target_property(_targetDirs ${_target} INTERFACE_SYSTEM_INCLUDE_DIRECTORIES)
-        if (_targetDirs)
-            list (APPEND _systemIncludeDirs ${_targetDirs})
-        endif()
-    endif()
-    # interface include directories from linked library targets
-    if (_target)
-        set (_linkedTargets "")
-        cotire_get_target_usage_requirements(${_target} ${_config} _linkedTargets)
-        foreach (_linkedTarget ${_linkedTargets})
-            get_target_property(_linkedTargetType ${_linkedTarget} TYPE)
-            if (CMAKE_INCLUDE_CURRENT_DIR_IN_INTERFACE AND NOT CMAKE_VERSION VERSION_LESS "3.4.0" AND
-                    _linkedTargetType MATCHES "(STATIC|SHARED|MODULE|OBJECT)_LIBRARY")
-                # CMAKE_INCLUDE_CURRENT_DIR_IN_INTERFACE refers to CMAKE_CURRENT_BINARY_DIR and CMAKE_CURRENT_SOURCE_DIR
-                # at the time, when the target was created. These correspond to the target properties BINARY_DIR and SOURCE_DIR
-                # which are only available with CMake 3.4 or later.
-                get_target_property(_targetDirs ${_linkedTarget} BINARY_DIR)
-                if (_targetDirs)
-                    list (APPEND _dirs ${_targetDirs})
-                endif()
-                get_target_property(_targetDirs ${_linkedTarget} SOURCE_DIR)
-                if (_targetDirs)
-                    list (APPEND _dirs ${_targetDirs})
-                endif()
-            endif()
-            get_target_property(_targetDirs ${_linkedTarget} INTERFACE_INCLUDE_DIRECTORIES)
-            if (_targetDirs)
-                list (APPEND _dirs ${_targetDirs})
-            endif()
-            get_target_property(_targetDirs ${_linkedTarget} INTERFACE_SYSTEM_INCLUDE_DIRECTORIES)
-            if (_targetDirs)
-                list (APPEND _systemIncludeDirs ${_targetDirs})
-            endif()
-        endforeach()
-    endif()
-    if (dirs)
-        list (REMOVE_DUPLICATES _dirs)
-    endif()
-    list (LENGTH _includeDirs _projectInsertIndex)
-    foreach (_dir ${_dirs})
-        if (CMAKE_INCLUDE_DIRECTORIES_PROJECT_BEFORE)
-            cotire_check_is_path_relative_to("${_dir}" _isRelative "${CMAKE_SOURCE_DIR}" "${CMAKE_BINARY_DIR}")
-            if (_isRelative)
-                list (LENGTH _includeDirs _len)
-                if (_len EQUAL _projectInsertIndex)
-                    list (APPEND _includeDirs "${_dir}")
-                else()
-                    list (INSERT _includeDirs _projectInsertIndex "${_dir}")
-                endif()
-                math (EXPR _projectInsertIndex "${_projectInsertIndex} + 1")
-            else()
-                list (APPEND _includeDirs "${_dir}")
-            endif()
-        else()
-            list (APPEND _includeDirs "${_dir}")
-        endif()
-    endforeach()
-    list (REMOVE_DUPLICATES _includeDirs)
-    list (REMOVE_DUPLICATES _systemIncludeDirs)
-    if (CMAKE_${_language}_IMPLICIT_INCLUDE_DIRECTORIES)
-        list (REMOVE_ITEM _includeDirs ${CMAKE_${_language}_IMPLICIT_INCLUDE_DIRECTORIES})
-    endif()
-    if (WIN32 AND NOT MINGW)
-        # convert Windows paths in include directories to CMake paths
-        if (_includeDirs)
-            set (_paths "")
-            foreach (_dir ${_includeDirs})
-                file (TO_CMAKE_PATH "${_dir}" _path)
-                list (APPEND _paths "${_path}")
-            endforeach()
-            set (_includeDirs ${_paths})
-        endif()
-        if (_systemIncludeDirs)
-            set (_paths "")
-            foreach (_dir ${_systemIncludeDirs})
-                file (TO_CMAKE_PATH "${_dir}" _path)
-                list (APPEND _paths "${_path}")
-            endforeach()
-            set (_systemIncludeDirs ${_paths})
-        endif()
-    endif()
-    if (COTIRE_DEBUG AND _includeDirs)
-        message (STATUS "Target ${_target} include dirs: ${_includeDirs}")
-    endif()
-    set (${_includeDirsVar} ${_includeDirs} PARENT_SCOPE)
-    if (COTIRE_DEBUG AND _systemIncludeDirs)
-        message (STATUS "Target ${_target} system include dirs: ${_systemIncludeDirs}")
-    endif()
-    set (${_systemIncludeDirsVar} ${_systemIncludeDirs} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_target_export_symbol _target _exportSymbolVar)
-    set (_exportSymbol "")
-    get_target_property(_targetType ${_target} TYPE)
-    get_target_property(_enableExports ${_target} ENABLE_EXPORTS)
-    if (_targetType MATCHES "(SHARED|MODULE)_LIBRARY" OR
-    (_targetType STREQUAL "EXECUTABLE" AND _enableExports))
-        get_target_property(_exportSymbol ${_target} DEFINE_SYMBOL)
-        if (NOT _exportSymbol)
-            set (_exportSymbol "${_target}_EXPORTS")
-        endif()
-        string (MAKE_C_IDENTIFIER "${_exportSymbol}" _exportSymbol)
-    endif()
-    set (${_exportSymbolVar} ${_exportSymbol} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_target_compile_definitions _config _language _target _definitionsVar)
-    string (TOUPPER "${_config}" _upperConfig)
-    set (_configDefinitions "")
-    # CMAKE_INTDIR for multi-configuration build systems
-    if (NOT "${CMAKE_CFG_INTDIR}" STREQUAL ".")
-        list (APPEND _configDefinitions "CMAKE_INTDIR=\"${_config}\"")
-    endif()
-    # target export define symbol
-    cotire_get_target_export_symbol("${_target}" _defineSymbol)
-    if (_defineSymbol)
-        list (APPEND _configDefinitions "${_defineSymbol}")
-    endif()
-    # directory compile definitions
-    get_directory_property(_definitions DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}" COMPILE_DEFINITIONS)
-    if (_definitions)
-        list (APPEND _configDefinitions ${_definitions})
-    endif()
-    get_directory_property(_definitions DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}" COMPILE_DEFINITIONS_${_upperConfig})
-    if (_definitions)
-        list (APPEND _configDefinitions ${_definitions})
-    endif()
-    # target compile definitions
-    get_target_property(_definitions ${_target} COMPILE_DEFINITIONS)
-    if (_definitions)
-        list (APPEND _configDefinitions ${_definitions})
-    endif()
-    get_target_property(_definitions ${_target} COMPILE_DEFINITIONS_${_upperConfig})
-    if (_definitions)
-        list (APPEND _configDefinitions ${_definitions})
-    endif()
-    # interface compile definitions from linked library targets
-    set (_linkedTargets "")
-    cotire_get_target_usage_requirements(${_target} ${_config} _linkedTargets)
-    foreach (_linkedTarget ${_linkedTargets})
-        get_target_property(_definitions ${_linkedTarget} INTERFACE_COMPILE_DEFINITIONS)
-        if (_definitions)
-            list (APPEND _configDefinitions ${_definitions})
-        endif()
-    endforeach()
-    # parse additional compile definitions from target compile flags
-    # and don't look at directory compile definitions, which we already handled
-    set (_targetFlags "")
-    cotire_get_target_compile_flags("${_config}" "${_language}" "${_target}" _targetFlags)
-    cotire_filter_compile_flags("${_language}" "D" _definitions _ignore ${_targetFlags})
-    if (_definitions)
-        list (APPEND _configDefinitions ${_definitions})
-    endif()
-    list (REMOVE_DUPLICATES _configDefinitions)
-    if (COTIRE_DEBUG AND _configDefinitions)
-        message (STATUS "Target ${_target} compile definitions: ${_configDefinitions}")
-    endif()
-    set (${_definitionsVar} ${_configDefinitions} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_target_compiler_flags _config _language _target _compilerFlagsVar)
-    # parse target compile flags omitting compile definitions and include directives
-    set (_targetFlags "")
-    cotire_get_target_compile_flags("${_config}" "${_language}" "${_target}" _targetFlags)
-    set (_flagFilter "D")
-    if (CMAKE_INCLUDE_FLAG_${_language})
-        string (STRIP "${CMAKE_INCLUDE_FLAG_${_language}}" _includeFlag)
-        string (REGEX REPLACE "^[-/]+" "" _includeFlag "${_includeFlag}")
-        if (_includeFlag)
-            set (_flagFilter "${_flagFilter}|${_includeFlag}")
-        endif()
-    endif()
-    if (CMAKE_INCLUDE_SYSTEM_FLAG_${_language})
-        string (STRIP "${CMAKE_INCLUDE_SYSTEM_FLAG_${_language}}" _includeFlag)
-        string (REGEX REPLACE "^[-/]+" "" _includeFlag "${_includeFlag}")
-        if (_includeFlag)
-            set (_flagFilter "${_flagFilter}|${_includeFlag}")
-        endif()
-    endif()
-    set (_compilerFlags "")
-    cotire_filter_compile_flags("${_language}" "${_flagFilter}" _ignore _compilerFlags ${_targetFlags})
-    if (COTIRE_DEBUG AND _compilerFlags)
-        message (STATUS "Target ${_target} compiler flags: ${_compilerFlags}")
-    endif()
-    set (${_compilerFlagsVar} ${_compilerFlags} PARENT_SCOPE)
-endfunction()
-
-function (cotire_add_sys_root_paths _pathsVar)
-    if (APPLE)
-        if (CMAKE_OSX_SYSROOT AND CMAKE_${_language}_HAS_ISYSROOT)
-            foreach (_path IN LISTS ${_pathsVar})
-                if (IS_ABSOLUTE "${_path}")
-                    get_filename_component(_path "${CMAKE_OSX_SYSROOT}/${_path}" ABSOLUTE)
-                    if (EXISTS "${_path}")
-                        list (APPEND ${_pathsVar} "${_path}")
-                    endif()
-                endif()
-            endforeach()
-        endif()
-    endif()
-    set (${_pathsVar} ${${_pathsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_extra_properties _sourceFile _pattern _resultVar)
-    set (_extraProperties ${ARGN})
-    set (_result "")
-    if (_extraProperties)
-        list (FIND _extraProperties "${_sourceFile}" _index)
-        if (_index GREATER -1)
-            math (EXPR _index "${_index} + 1")
-            list (LENGTH _extraProperties _len)
-            math (EXPR _len "${_len} - 1")
-            foreach (_index RANGE ${_index} ${_len})
-                list (GET _extraProperties ${_index} _value)
-                if (_value MATCHES "${_pattern}")
-                    list (APPEND _result "${_value}")
-                else()
-                    break()
-                endif()
-            endforeach()
-        endif()
-    endif()
-    set (${_resultVar} ${_result} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_compile_definitions _config _language _sourceFile _definitionsVar)
-    set (_compileDefinitions "")
-    if (NOT CMAKE_SCRIPT_MODE_FILE)
-        string (TOUPPER "${_config}" _upperConfig)
-        get_source_file_property(_definitions "${_sourceFile}" COMPILE_DEFINITIONS)
-        if (_definitions)
-            list (APPEND _compileDefinitions ${_definitions})
-        endif()
-        get_source_file_property(_definitions "${_sourceFile}" COMPILE_DEFINITIONS_${_upperConfig})
-        if (_definitions)
-            list (APPEND _compileDefinitions ${_definitions})
-        endif()
-    endif()
-    cotire_get_source_extra_properties("${_sourceFile}" "^[a-zA-Z0-9_]+(=.*)?$" _definitions ${ARGN})
-    if (_definitions)
-        list (APPEND _compileDefinitions ${_definitions})
-    endif()
-    if (COTIRE_DEBUG AND _compileDefinitions)
-        message (STATUS "Source ${_sourceFile} compile definitions: ${_compileDefinitions}")
-    endif()
-    set (${_definitionsVar} ${_compileDefinitions} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_files_compile_definitions _config _language _definitionsVar)
-    set (_configDefinitions "")
-    foreach (_sourceFile ${ARGN})
-        cotire_get_source_compile_definitions("${_config}" "${_language}" "${_sourceFile}" _sourceDefinitions)
-        if (_sourceDefinitions)
-            list (APPEND _configDefinitions "${_sourceFile}" ${_sourceDefinitions} "-")
-        endif()
-    endforeach()
-    set (${_definitionsVar} ${_configDefinitions} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_undefs _sourceFile _property _sourceUndefsVar)
-    set (_sourceUndefs "")
-    if (NOT CMAKE_SCRIPT_MODE_FILE)
-        get_source_file_property(_undefs "${_sourceFile}" ${_property})
-        if (_undefs)
-            list (APPEND _sourceUndefs ${_undefs})
-        endif()
-    endif()
-    cotire_get_source_extra_properties("${_sourceFile}" "^[a-zA-Z0-9_]+$" _undefs ${ARGN})
-    if (_undefs)
-        list (APPEND _sourceUndefs ${_undefs})
-    endif()
-    if (COTIRE_DEBUG AND _sourceUndefs)
-        message (STATUS "Source ${_sourceFile} ${_property} undefs: ${_sourceUndefs}")
-    endif()
-    set (${_sourceUndefsVar} ${_sourceUndefs} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_source_files_undefs _property _sourceUndefsVar)
-    set (_sourceUndefs "")
-    foreach (_sourceFile ${ARGN})
-        cotire_get_source_undefs("${_sourceFile}" ${_property} _undefs)
-        if (_undefs)
-            list (APPEND _sourceUndefs "${_sourceFile}" ${_undefs} "-")
-        endif()
-    endforeach()
-    set (${_sourceUndefsVar} ${_sourceUndefs} PARENT_SCOPE)
-endfunction()
-
-macro (cotire_set_cmd_to_prologue _cmdVar)
-    set (${_cmdVar} "${CMAKE_COMMAND}")
-    if (COTIRE_DEBUG)
-        list (APPEND ${_cmdVar} "--warn-uninitialized")
-    endif()
-    list (APPEND ${_cmdVar} "-DCOTIRE_BUILD_TYPE:STRING=$<CONFIGURATION>")
-    if (XCODE)
-        list (APPEND ${_cmdVar} "-DXCODE:BOOL=TRUE")
-    endif()
-    if (COTIRE_VERBOSE)
-        list (APPEND ${_cmdVar} "-DCOTIRE_VERBOSE:BOOL=ON")
-    elseif("${CMAKE_GENERATOR}" MATCHES "Makefiles")
-        list (APPEND ${_cmdVar} "-DCOTIRE_VERBOSE:BOOL=$(VERBOSE)")
-    endif()
-endmacro()
-
-function (cotire_init_compile_cmd _cmdVar _language _compilerLauncher _compilerExe _compilerArg1)
-    if (NOT _compilerLauncher)
-        set (_compilerLauncher ${CMAKE_${_language}_COMPILER_LAUNCHER})
-    endif()
-    if (NOT _compilerExe)
-        set (_compilerExe "${CMAKE_${_language}_COMPILER}")
-    endif()
-    if (NOT _compilerArg1)
-        set (_compilerArg1 ${CMAKE_${_language}_COMPILER_ARG1})
-    endif()
-    string (STRIP "${_compilerArg1}" _compilerArg1)
-    if ("${CMAKE_GENERATOR}" MATCHES "Make|Ninja")
-        # compiler launcher is only supported for Makefile and Ninja
-        set (${_cmdVar} ${_compilerLauncher} "${_compilerExe}" ${_compilerArg1} PARENT_SCOPE)
-    else()
-        set (${_cmdVar} "${_compilerExe}" ${_compilerArg1} PARENT_SCOPE)
-    endif()
-endfunction()
-
-macro (cotire_add_definitions_to_cmd _cmdVar _language)
-    foreach (_definition ${ARGN})
-        if (WIN32 AND CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-            list (APPEND ${_cmdVar} "/D${_definition}")
-        else()
-            list (APPEND ${_cmdVar} "-D${_definition}")
-        endif()
-    endforeach()
-endmacro()
-
-function (cotire_add_includes_to_cmd _cmdVar _language _includesVar _systemIncludesVar)
-    set (_includeDirs ${${_includesVar}} ${${_systemIncludesVar}})
-    if (_includeDirs)
-        list (REMOVE_DUPLICATES _includeDirs)
-        foreach (_include ${_includeDirs})
-            if (WIN32 AND CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-                file (TO_NATIVE_PATH "${_include}" _include)
-                list (APPEND ${_cmdVar} "${CMAKE_INCLUDE_FLAG_${_language}}${CMAKE_INCLUDE_FLAG_SEP_${_language}}${_include}")
-            else()
-                set (_index -1)
-                if ("${CMAKE_INCLUDE_SYSTEM_FLAG_${_language}}" MATCHES ".+")
-                    list (FIND ${_systemIncludesVar} "${_include}" _index)
-                endif()
-                if (_index GREATER -1)
-                    list (APPEND ${_cmdVar} "${CMAKE_INCLUDE_SYSTEM_FLAG_${_language}}${CMAKE_INCLUDE_FLAG_SEP_${_language}}${_include}")
-                else()
-                    list (APPEND ${_cmdVar} "${CMAKE_INCLUDE_FLAG_${_language}}${CMAKE_INCLUDE_FLAG_SEP_${_language}}${_include}")
-                endif()
-            endif()
-        endforeach()
-    endif()
-    set (${_cmdVar} ${${_cmdVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_add_frameworks_to_cmd _cmdVar _language _includesVar _systemIncludesVar)
-    if (APPLE)
-        set (_frameworkDirs "")
-        foreach (_include ${${_includesVar}})
-            if (IS_ABSOLUTE "${_include}" AND _include MATCHES "\\.framework$")
-                get_filename_component(_frameworkDir "${_include}" DIRECTORY)
-                list (APPEND _frameworkDirs "${_frameworkDir}")
-            endif()
-        endforeach()
-        set (_systemFrameworkDirs "")
-        foreach (_include ${${_systemIncludesVar}})
-            if (IS_ABSOLUTE "${_include}" AND _include MATCHES "\\.framework$")
-                get_filename_component(_frameworkDir "${_include}" DIRECTORY)
-                list (APPEND _systemFrameworkDirs "${_frameworkDir}")
-            endif()
-        endforeach()
-        if (_systemFrameworkDirs)
-            list (APPEND _frameworkDirs ${_systemFrameworkDirs})
-        endif()
-        if (_frameworkDirs)
-            list (REMOVE_DUPLICATES _frameworkDirs)
-            foreach (_frameworkDir ${_frameworkDirs})
-                set (_index -1)
-                if ("${CMAKE_${_language}_SYSTEM_FRAMEWORK_SEARCH_FLAG}" MATCHES ".+")
-                    list (FIND _systemFrameworkDirs "${_frameworkDir}" _index)
-                endif()
-                if (_index GREATER -1)
-                    list (APPEND ${_cmdVar} "${CMAKE_${_language}_SYSTEM_FRAMEWORK_SEARCH_FLAG}${_frameworkDir}")
-                else()
-                    list (APPEND ${_cmdVar} "${CMAKE_${_language}_FRAMEWORK_SEARCH_FLAG}${_frameworkDir}")
-                endif()
-            endforeach()
-        endif()
-    endif()
-    set (${_cmdVar} ${${_cmdVar}} PARENT_SCOPE)
-endfunction()
-
-macro (cotire_add_compile_flags_to_cmd _cmdVar)
-    foreach (_flag ${ARGN})
-        list (APPEND ${_cmdVar} "${_flag}")
-    endforeach()
-endmacro()
-
-function (cotire_check_file_up_to_date _fileIsUpToDateVar _file)
-    if (EXISTS "${_file}")
-        set (_triggerFile "")
-        foreach (_dependencyFile ${ARGN})
-            if (EXISTS "${_dependencyFile}")
-                # IS_NEWER_THAN returns TRUE if both files have the same timestamp
-                # thus we do the comparison in both directions to exclude ties
-                if ("${_dependencyFile}" IS_NEWER_THAN "${_file}" AND
-                        NOT "${_file}" IS_NEWER_THAN "${_dependencyFile}")
-                    set (_triggerFile "${_dependencyFile}")
-                    break()
-                endif()
-            endif()
-        endforeach()
-        if (_triggerFile)
-            if (COTIRE_VERBOSE)
-                get_filename_component(_fileName "${_file}" NAME)
-                message (STATUS "${_fileName} update triggered by ${_triggerFile} change.")
-            endif()
-            set (${_fileIsUpToDateVar} FALSE PARENT_SCOPE)
-        else()
-            if (COTIRE_VERBOSE)
-                get_filename_component(_fileName "${_file}" NAME)
-                message (STATUS "${_fileName} is up-to-date.")
-            endif()
-            set (${_fileIsUpToDateVar} TRUE PARENT_SCOPE)
-        endif()
-    else()
-        if (COTIRE_VERBOSE)
-            get_filename_component(_fileName "${_file}" NAME)
-            message (STATUS "${_fileName} does not exist yet.")
-        endif()
-        set (${_fileIsUpToDateVar} FALSE PARENT_SCOPE)
-    endif()
-endfunction()
-
-macro (cotire_find_closest_relative_path _headerFile _includeDirs _relPathVar)
-    set (${_relPathVar} "")
-    foreach (_includeDir ${_includeDirs})
-        if (IS_DIRECTORY "${_includeDir}")
-            file (RELATIVE_PATH _relPath "${_includeDir}" "${_headerFile}")
-            if (NOT IS_ABSOLUTE "${_relPath}" AND NOT "${_relPath}" MATCHES "^\\.\\.")
-                string (LENGTH "${${_relPathVar}}" _closestLen)
-                string (LENGTH "${_relPath}" _relLen)
-                if (_closestLen EQUAL 0 OR _relLen LESS _closestLen)
-                    set (${_relPathVar} "${_relPath}")
-                endif()
-            endif()
-        elseif ("${_includeDir}" STREQUAL "${_headerFile}")
-            # if path matches exactly, return short non-empty string
-            set (${_relPathVar} "1")
-            break()
-        endif()
-    endforeach()
-endmacro()
-
-macro (cotire_check_header_file_location _headerFile _insideIncludeDirs _outsideIncludeDirs _headerIsInside)
-    # check header path against ignored and honored include directories
-    cotire_find_closest_relative_path("${_headerFile}" "${_insideIncludeDirs}" _insideRelPath)
-    if (_insideRelPath)
-        # header is inside, but could be become outside if there is a shorter outside match
-        cotire_find_closest_relative_path("${_headerFile}" "${_outsideIncludeDirs}" _outsideRelPath)
-        if (_outsideRelPath)
-            string (LENGTH "${_insideRelPath}" _insideRelPathLen)
-            string (LENGTH "${_outsideRelPath}" _outsideRelPathLen)
-            if (_outsideRelPathLen LESS _insideRelPathLen)
-                set (${_headerIsInside} FALSE)
-            else()
-                set (${_headerIsInside} TRUE)
-            endif()
-        else()
-            set (${_headerIsInside} TRUE)
-        endif()
-    else()
-        # header is outside
-        set (${_headerIsInside} FALSE)
-    endif()
-endmacro()
-
-macro (cotire_check_ignore_header_file_path _headerFile _headerIsIgnoredVar)
-    if (NOT EXISTS "${_headerFile}")
-        set (${_headerIsIgnoredVar} TRUE)
-    elseif (IS_DIRECTORY "${_headerFile}")
-        set (${_headerIsIgnoredVar} TRUE)
-    elseif ("${_headerFile}" MATCHES "\\.\\.|[_-]fixed" AND "${_headerFile}" MATCHES "\\.h$")
-        # heuristic: ignore C headers with embedded parent directory references or "-fixed" or "_fixed" in path
-        # these often stem from using GCC #include_next tricks, which may break the precompiled header compilation
-        # with the error message "error: no include path in which to search for header.h"
-        set (${_headerIsIgnoredVar} TRUE)
-    else()
-        set (${_headerIsIgnoredVar} FALSE)
-    endif()
-endmacro()
-
-macro (cotire_check_ignore_header_file_ext _headerFile _ignoreExtensionsVar _headerIsIgnoredVar)
-    # check header file extension
-    cotire_get_source_file_extension("${_headerFile}" _headerFileExt)
-    set (${_headerIsIgnoredVar} FALSE)
-    if (_headerFileExt)
-        list (FIND ${_ignoreExtensionsVar} "${_headerFileExt}" _index)
-        if (_index GREATER -1)
-            set (${_headerIsIgnoredVar} TRUE)
-        endif()
-    endif()
-endmacro()
-
-macro (cotire_parse_line _line _headerFileVar _headerDepthVar)
-    if (MSVC)
-        # cl.exe /showIncludes output looks different depending on the language pack used, e.g.:
-        # English: "Note: including file:   C:\directory\file"
-        # German: "Hinweis: Einlesen der Datei:   C:\directory\file"
-        # We use a very general regular expression, relying on the presence of the : characters
-        if (_line MATCHES "( +)([a-zA-Z]:[^:]+)$")
-            # Visual Studio compiler output
-            string (LENGTH "${CMAKE_MATCH_1}" ${_headerDepthVar})
-            get_filename_component(${_headerFileVar} "${CMAKE_MATCH_2}" ABSOLUTE)
-        else()
-            set (${_headerFileVar} "")
-            set (${_headerDepthVar} 0)
-        endif()
-    else()
-        if (_line MATCHES "^(\\.+) (.*)$")
-            # GCC like output
-            string (LENGTH "${CMAKE_MATCH_1}" ${_headerDepthVar})
-            if (IS_ABSOLUTE "${CMAKE_MATCH_2}")
-                set (${_headerFileVar} "${CMAKE_MATCH_2}")
-            else()
-                get_filename_component(${_headerFileVar} "${CMAKE_MATCH_2}" REALPATH)
-            endif()
-        else()
-            set (${_headerFileVar} "")
-            set (${_headerDepthVar} 0)
-        endif()
-    endif()
-endmacro()
-
-function (cotire_parse_includes _language _scanOutput _ignoredIncludeDirs _honoredIncludeDirs _ignoredExtensions _selectedIncludesVar _unparsedLinesVar)
-    if (WIN32)
-        # prevent CMake macro invocation errors due to backslash characters in Windows paths
-        string (REPLACE "\\" "/" _scanOutput "${_scanOutput}")
-    endif()
-    # canonize slashes
-    string (REPLACE "//" "/" _scanOutput "${_scanOutput}")
-    # prevent semicolon from being interpreted as a line separator
-    string (REPLACE ";" "\\;" _scanOutput "${_scanOutput}")
-    # then separate lines
-    string (REGEX REPLACE "\n" ";" _scanOutput "${_scanOutput}")
-    list (LENGTH _scanOutput _len)
-    # remove duplicate lines to speed up parsing
-    list (REMOVE_DUPLICATES _scanOutput)
-    list (LENGTH _scanOutput _uniqueLen)
-    if (COTIRE_VERBOSE OR COTIRE_DEBUG)
-        message (STATUS "Scanning ${_uniqueLen} unique lines of ${_len} for includes")
-        if (_ignoredExtensions)
-            message (STATUS "Ignored extensions: ${_ignoredExtensions}")
-        endif()
-        if (_ignoredIncludeDirs)
-            message (STATUS "Ignored paths: ${_ignoredIncludeDirs}")
-        endif()
-        if (_honoredIncludeDirs)
-            message (STATUS "Included paths: ${_honoredIncludeDirs}")
-        endif()
-    endif()
-    set (_sourceFiles ${ARGN})
-    set (_selectedIncludes "")
-    set (_unparsedLines "")
-    # stack keeps track of inside/outside project status of processed header files
-    set (_headerIsInsideStack "")
-    foreach (_line IN LISTS _scanOutput)
-        if (_line)
-            cotire_parse_line("${_line}" _headerFile _headerDepth)
-            if (_headerFile)
-                cotire_check_header_file_location("${_headerFile}" "${_ignoredIncludeDirs}" "${_honoredIncludeDirs}" _headerIsInside)
-                if (COTIRE_DEBUG)
-                    message (STATUS "${_headerDepth}: ${_headerFile} ${_headerIsInside}")
-                endif()
-                # update stack
-                list (LENGTH _headerIsInsideStack _stackLen)
-                if (_headerDepth GREATER _stackLen)
-                    math (EXPR _stackLen "${_stackLen} + 1")
-                    foreach (_index RANGE ${_stackLen} ${_headerDepth})
-                        list (APPEND _headerIsInsideStack ${_headerIsInside})
-                    endforeach()
-                else()
-                    foreach (_index RANGE ${_headerDepth} ${_stackLen})
-                        list (REMOVE_AT _headerIsInsideStack -1)
-                    endforeach()
-                    list (APPEND _headerIsInsideStack ${_headerIsInside})
-                endif()
-                if (COTIRE_DEBUG)
-                    message (STATUS "${_headerIsInsideStack}")
-                endif()
-                # header is a candidate if it is outside project
-                if (NOT _headerIsInside)
-                    # get parent header file's inside/outside status
-                    if (_headerDepth GREATER 1)
-                        math (EXPR _index "${_headerDepth} - 2")
-                        list (GET _headerIsInsideStack ${_index} _parentHeaderIsInside)
-                    else()
-                        set (_parentHeaderIsInside TRUE)
-                    endif()
-                    # select header file if parent header file is inside project
-                    # (e.g., a project header file that includes a standard header file)
-                    if (_parentHeaderIsInside)
-                        cotire_check_ignore_header_file_path("${_headerFile}" _headerIsIgnored)
-                        if (NOT _headerIsIgnored)
-                            cotire_check_ignore_header_file_ext("${_headerFile}" _ignoredExtensions _headerIsIgnored)
-                            if (NOT _headerIsIgnored)
-                                list (APPEND _selectedIncludes "${_headerFile}")
-                            else()
-                                # fix header's inside status on stack, it is ignored by extension now
-                                list (REMOVE_AT _headerIsInsideStack -1)
-                                list (APPEND _headerIsInsideStack TRUE)
-                            endif()
-                        endif()
-                        if (COTIRE_DEBUG)
-                            message (STATUS "${_headerFile} ${_ignoredExtensions} ${_headerIsIgnored}")
-                        endif()
-                    endif()
-                endif()
-            else()
-                if (MSVC)
-                    # for cl.exe do not keep unparsed lines which solely consist of a source file name
-                    string (FIND "${_sourceFiles}" "${_line}" _index)
-                    if (_index LESS 0)
-                        list (APPEND _unparsedLines "${_line}")
-                    endif()
-                else()
-                    list (APPEND _unparsedLines "${_line}")
-                endif()
-            endif()
-        endif()
-    endforeach()
-    list (REMOVE_DUPLICATES _selectedIncludes)
-    set (${_selectedIncludesVar} ${_selectedIncludes} PARENT_SCOPE)
-    set (${_unparsedLinesVar} ${_unparsedLines} PARENT_SCOPE)
-endfunction()
-
-function (cotire_scan_includes _includesVar)
-    set(_options "")
-    set(_oneValueArgs COMPILER_ID COMPILER_EXECUTABLE COMPILER_ARG1 COMPILER_VERSION LANGUAGE UNPARSED_LINES SCAN_RESULT)
-    set(_multiValueArgs COMPILE_DEFINITIONS COMPILE_FLAGS INCLUDE_DIRECTORIES SYSTEM_INCLUDE_DIRECTORIES
-            IGNORE_PATH INCLUDE_PATH IGNORE_EXTENSIONS INCLUDE_PRIORITY_PATH COMPILER_LAUNCHER)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    set (_sourceFiles ${_option_UNPARSED_ARGUMENTS})
-    if (NOT _option_LANGUAGE)
-        set (_option_LANGUAGE "CXX")
-    endif()
-    if (NOT _option_COMPILER_ID)
-        set (_option_COMPILER_ID "${CMAKE_${_option_LANGUAGE}_ID}")
-    endif()
-    if (NOT _option_COMPILER_VERSION)
-        set (_option_COMPILER_VERSION "${CMAKE_${_option_LANGUAGE}_COMPILER_VERSION}")
-    endif()
-    cotire_init_compile_cmd(_cmd "${_option_LANGUAGE}" "${_option_COMPILER_LAUNCHER}" "${_option_COMPILER_EXECUTABLE}" "${_option_COMPILER_ARG1}")
-    cotire_add_definitions_to_cmd(_cmd "${_option_LANGUAGE}" ${_option_COMPILE_DEFINITIONS})
-    cotire_add_compile_flags_to_cmd(_cmd ${_option_COMPILE_FLAGS})
-    cotire_add_includes_to_cmd(_cmd "${_option_LANGUAGE}" _option_INCLUDE_DIRECTORIES _option_SYSTEM_INCLUDE_DIRECTORIES)
-    cotire_add_frameworks_to_cmd(_cmd "${_option_LANGUAGE}" _option_INCLUDE_DIRECTORIES _option_SYSTEM_INCLUDE_DIRECTORIES)
-    cotire_add_makedep_flags("${_option_LANGUAGE}" "${_option_COMPILER_ID}" "${_option_COMPILER_VERSION}" _cmd)
-    # only consider existing source files for scanning
-    set (_existingSourceFiles "")
-    foreach (_sourceFile ${_sourceFiles})
-        if (EXISTS "${_sourceFile}")
-            list (APPEND _existingSourceFiles "${_sourceFile}")
-        endif()
-    endforeach()
-    if (NOT _existingSourceFiles)
-        set (${_includesVar} "" PARENT_SCOPE)
-        return()
-    endif()
-    list (APPEND _cmd ${_existingSourceFiles})
-    if (COTIRE_VERBOSE)
-        message (STATUS "execute_process: ${_cmd}")
-    endif()
-    if (_option_COMPILER_ID MATCHES "MSVC")
-        # cl.exe messes with the output streams unless the environment variable VS_UNICODE_OUTPUT is cleared
-        unset (ENV{VS_UNICODE_OUTPUT})
-    endif()
-    execute_process(
-            COMMAND ${_cmd}
-            WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-            RESULT_VARIABLE _result
-            OUTPUT_QUIET
-            ERROR_VARIABLE _output)
-    if (_result)
-        message (STATUS "Result ${_result} scanning includes of ${_existingSourceFiles}.")
-    endif()
-    cotire_parse_includes(
-            "${_option_LANGUAGE}" "${_output}"
-            "${_option_IGNORE_PATH}" "${_option_INCLUDE_PATH}"
-            "${_option_IGNORE_EXTENSIONS}"
-            _includes _unparsedLines
-            ${_sourceFiles})
-    if (_option_INCLUDE_PRIORITY_PATH)
-        set (_sortedIncludes "")
-        foreach (_priorityPath ${_option_INCLUDE_PRIORITY_PATH})
-            foreach (_include ${_includes})
-                string (FIND ${_include} ${_priorityPath} _position)
-                if (_position GREATER -1)
-                    list (APPEND _sortedIncludes ${_include})
-                endif()
-            endforeach()
-        endforeach()
-        if (_sortedIncludes)
-            list (INSERT _includes 0 ${_sortedIncludes})
-            list (REMOVE_DUPLICATES _includes)
-        endif()
-    endif()
-    set (${_includesVar} ${_includes} PARENT_SCOPE)
-    if (_option_UNPARSED_LINES)
-        set (${_option_UNPARSED_LINES} ${_unparsedLines} PARENT_SCOPE)
-    endif()
-    if (_option_SCAN_RESULT)
-        set (${_option_SCAN_RESULT} ${_result} PARENT_SCOPE)
-    endif()
-endfunction()
-
-macro (cotire_append_undefs _contentsVar)
-    set (_undefs ${ARGN})
-    if (_undefs)
-        list (REMOVE_DUPLICATES _undefs)
-        foreach (_definition ${_undefs})
-            list (APPEND ${_contentsVar} "#undef ${_definition}")
-        endforeach()
-    endif()
-endmacro()
-
-macro (cotire_comment_str _language _commentText _commentVar)
-    if ("${_language}" STREQUAL "CMAKE")
-        set (${_commentVar} "# ${_commentText}")
-    else()
-        set (${_commentVar} "/* ${_commentText} */")
-    endif()
-endmacro()
-
-function (cotire_write_file _language _file _contents _force)
-    get_filename_component(_moduleName "${COTIRE_CMAKE_MODULE_FILE}" NAME)
-    cotire_comment_str("${_language}" "${_moduleName} ${COTIRE_CMAKE_MODULE_VERSION} generated file" _header1)
-    cotire_comment_str("${_language}" "${_file}" _header2)
-    set (_contents "${_header1}\n${_header2}\n${_contents}")
-    if (COTIRE_DEBUG)
-        message (STATUS "${_contents}")
-    endif()
-    if (_force OR NOT EXISTS "${_file}")
-        file (WRITE "${_file}" "${_contents}")
-    else()
-        file (READ "${_file}" _oldContents)
-        if (NOT "${_oldContents}" STREQUAL "${_contents}")
-            file (WRITE "${_file}" "${_contents}")
-        else()
-            if (COTIRE_DEBUG)
-                message (STATUS "${_file} unchanged")
-            endif()
-        endif()
-    endif()
-endfunction()
-
-function (cotire_generate_unity_source _unityFile)
-    set(_options "")
-    set(_oneValueArgs LANGUAGE)
-    set(_multiValueArgs
-            DEPENDS SOURCES_COMPILE_DEFINITIONS
-            PRE_UNDEFS SOURCES_PRE_UNDEFS POST_UNDEFS SOURCES_POST_UNDEFS PROLOGUE EPILOGUE)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    if (_option_DEPENDS)
-        cotire_check_file_up_to_date(_unityFileIsUpToDate "${_unityFile}" ${_option_DEPENDS})
-        if (_unityFileIsUpToDate)
-            return()
-        endif()
-    endif()
-    set (_sourceFiles ${_option_UNPARSED_ARGUMENTS})
-    if (NOT _option_PRE_UNDEFS)
-        set (_option_PRE_UNDEFS "")
-    endif()
-    if (NOT _option_SOURCES_PRE_UNDEFS)
-        set (_option_SOURCES_PRE_UNDEFS "")
-    endif()
-    if (NOT _option_POST_UNDEFS)
-        set (_option_POST_UNDEFS "")
-    endif()
-    if (NOT _option_SOURCES_POST_UNDEFS)
-        set (_option_SOURCES_POST_UNDEFS "")
-    endif()
-    set (_contents "")
-    if (_option_PROLOGUE)
-        list (APPEND _contents ${_option_PROLOGUE})
-    endif()
-    if (_option_LANGUAGE AND _sourceFiles)
-        if ("${_option_LANGUAGE}" STREQUAL "CXX")
-            list (APPEND _contents "#ifdef __cplusplus")
-        elseif ("${_option_LANGUAGE}" STREQUAL "C")
-            list (APPEND _contents "#ifndef __cplusplus")
-        endif()
-    endif()
-    set (_compileUndefinitions "")
-    foreach (_sourceFile ${_sourceFiles})
-        cotire_get_source_compile_definitions(
-                "${_option_CONFIGURATION}" "${_option_LANGUAGE}" "${_sourceFile}" _compileDefinitions
-                ${_option_SOURCES_COMPILE_DEFINITIONS})
-        cotire_get_source_undefs("${_sourceFile}" COTIRE_UNITY_SOURCE_PRE_UNDEFS _sourcePreUndefs ${_option_SOURCES_PRE_UNDEFS})
-        cotire_get_source_undefs("${_sourceFile}" COTIRE_UNITY_SOURCE_POST_UNDEFS _sourcePostUndefs ${_option_SOURCES_POST_UNDEFS})
-        if (_option_PRE_UNDEFS)
-            list (APPEND _compileUndefinitions ${_option_PRE_UNDEFS})
-        endif()
-        if (_sourcePreUndefs)
-            list (APPEND _compileUndefinitions ${_sourcePreUndefs})
-        endif()
-        if (_compileUndefinitions)
-            cotire_append_undefs(_contents ${_compileUndefinitions})
-            set (_compileUndefinitions "")
-        endif()
-        if (_sourcePostUndefs)
-            list (APPEND _compileUndefinitions ${_sourcePostUndefs})
-        endif()
-        if (_option_POST_UNDEFS)
-            list (APPEND _compileUndefinitions ${_option_POST_UNDEFS})
-        endif()
-        foreach (_definition ${_compileDefinitions})
-            if (_definition MATCHES "^([a-zA-Z0-9_]+)=(.+)$")
-                list (APPEND _contents "#define ${CMAKE_MATCH_1} ${CMAKE_MATCH_2}")
-                list (INSERT _compileUndefinitions 0 "${CMAKE_MATCH_1}")
-            else()
-                list (APPEND _contents "#define ${_definition}")
-                list (INSERT _compileUndefinitions 0 "${_definition}")
-            endif()
-        endforeach()
-        # use absolute path as source file location
-        get_filename_component(_sourceFileLocation "${_sourceFile}" ABSOLUTE)
-        if (WIN32)
-            file (TO_NATIVE_PATH "${_sourceFileLocation}" _sourceFileLocation)
-        endif()
-        list (APPEND _contents "#include \"${_sourceFileLocation}\"")
-    endforeach()
-    if (_compileUndefinitions)
-        cotire_append_undefs(_contents ${_compileUndefinitions})
-        set (_compileUndefinitions "")
-    endif()
-    if (_option_LANGUAGE AND _sourceFiles)
-        list (APPEND _contents "#endif")
-    endif()
-    if (_option_EPILOGUE)
-        list (APPEND _contents ${_option_EPILOGUE})
-    endif()
-    list (APPEND _contents "")
-    string (REPLACE ";" "\n" _contents "${_contents}")
-    if (COTIRE_VERBOSE)
-        message ("${_contents}")
-    endif()
-    cotire_write_file("${_option_LANGUAGE}" "${_unityFile}" "${_contents}" TRUE)
-endfunction()
-
-function (cotire_generate_prefix_header _prefixFile)
-    set(_options "")
-    set(_oneValueArgs LANGUAGE COMPILER_EXECUTABLE COMPILER_ARG1 COMPILER_ID COMPILER_VERSION)
-    set(_multiValueArgs DEPENDS COMPILE_DEFINITIONS COMPILE_FLAGS
-            INCLUDE_DIRECTORIES SYSTEM_INCLUDE_DIRECTORIES IGNORE_PATH INCLUDE_PATH
-            IGNORE_EXTENSIONS INCLUDE_PRIORITY_PATH COMPILER_LAUNCHER)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    if (NOT _option_COMPILER_ID)
-        set (_option_COMPILER_ID "${CMAKE_${_option_LANGUAGE}_ID}")
-    endif()
-    if (NOT _option_COMPILER_VERSION)
-        set (_option_COMPILER_VERSION "${CMAKE_${_option_LANGUAGE}_COMPILER_VERSION}")
-    endif()
-    if (_option_DEPENDS)
-        cotire_check_file_up_to_date(_prefixFileIsUpToDate "${_prefixFile}" ${_option_DEPENDS})
-        if (_prefixFileIsUpToDate)
-            # create empty log file
-            set (_unparsedLinesFile "${_prefixFile}.log")
-            file (WRITE "${_unparsedLinesFile}" "")
-            return()
-        endif()
-    endif()
-    set (_prologue "")
-    set (_epilogue "")
-    if (_option_COMPILER_ID MATCHES "Clang")
-        set (_prologue "#pragma clang system_header")
-    elseif (_option_COMPILER_ID MATCHES "GNU")
-        set (_prologue "#pragma GCC system_header")
-    elseif (_option_COMPILER_ID MATCHES "MSVC")
-        set (_prologue "#pragma warning(push, 0)")
-        set (_epilogue "#pragma warning(pop)")
-    elseif (_option_COMPILER_ID MATCHES "Intel")
-        # Intel compiler requires hdrstop pragma to stop generating PCH file
-        set (_epilogue "#pragma hdrstop")
-    endif()
-    set (_sourceFiles ${_option_UNPARSED_ARGUMENTS})
-    cotire_scan_includes(_selectedHeaders ${_sourceFiles}
-            LANGUAGE "${_option_LANGUAGE}"
-            COMPILER_LAUNCHER "${_option_COMPILER_LAUNCHER}"
-            COMPILER_EXECUTABLE "${_option_COMPILER_EXECUTABLE}"
-            COMPILER_ARG1 "${_option_COMPILER_ARG1}"
-            COMPILER_ID "${_option_COMPILER_ID}"
-            COMPILER_VERSION "${_option_COMPILER_VERSION}"
-            COMPILE_DEFINITIONS ${_option_COMPILE_DEFINITIONS}
-            COMPILE_FLAGS ${_option_COMPILE_FLAGS}
-            INCLUDE_DIRECTORIES ${_option_INCLUDE_DIRECTORIES}
-            SYSTEM_INCLUDE_DIRECTORIES ${_option_SYSTEM_INCLUDE_DIRECTORIES}
-            IGNORE_PATH ${_option_IGNORE_PATH}
-            INCLUDE_PATH ${_option_INCLUDE_PATH}
-            IGNORE_EXTENSIONS ${_option_IGNORE_EXTENSIONS}
-            INCLUDE_PRIORITY_PATH ${_option_INCLUDE_PRIORITY_PATH}
-            UNPARSED_LINES _unparsedLines
-            SCAN_RESULT _scanResult)
-    cotire_generate_unity_source("${_prefixFile}"
-            PROLOGUE ${_prologue} EPILOGUE ${_epilogue} LANGUAGE "${_option_LANGUAGE}" ${_selectedHeaders})
-    set (_unparsedLinesFile "${_prefixFile}.log")
-    if (_unparsedLines)
-        if (COTIRE_VERBOSE OR _scanResult OR NOT _selectedHeaders)
-            list (LENGTH _unparsedLines _skippedLineCount)
-            message (STATUS "${_skippedLineCount} line(s) skipped, see ${_unparsedLinesFile}")
-        endif()
-        string (REPLACE ";" "\n" _unparsedLines "${_unparsedLines}")
-    endif()
-    file (WRITE "${_unparsedLinesFile}" "${_unparsedLines}")
-endfunction()
-
-function (cotire_add_makedep_flags _language _compilerID _compilerVersion _flagsVar)
-    set (_flags ${${_flagsVar}})
-    if (_compilerID MATCHES "MSVC")
-        # cl.exe options used
-        # /nologo suppresses display of sign-on banner
-        # /TC treat all files named on the command line as C source files
-        # /TP treat all files named on the command line as C++ source files
-        # /EP preprocess to stdout without #line directives
-        # /showIncludes list include files
-        set (_sourceFileTypeC "/TC")
-        set (_sourceFileTypeCXX "/TP")
-        if (_flags)
-            # append to list
-            list (APPEND _flags /nologo "${_sourceFileType${_language}}" /EP /showIncludes)
-        else()
-            # return as a flag string
-            set (_flags "${_sourceFileType${_language}} /EP /showIncludes")
-        endif()
-    elseif (_compilerID MATCHES "GNU")
-        # GCC options used
-        # -H print the name of each header file used
-        # -E invoke preprocessor
-        # -fdirectives-only do not expand macros, requires GCC >= 4.3
-        if (_flags)
-            # append to list
-            list (APPEND _flags -H -E)
-            if (NOT "${_compilerVersion}" VERSION_LESS "4.3.0")
-                list (APPEND _flags "-fdirectives-only")
-            endif()
-        else()
-            # return as a flag string
-            set (_flags "-H -E")
-            if (NOT "${_compilerVersion}" VERSION_LESS "4.3.0")
-                set (_flags "${_flags} -fdirectives-only")
-            endif()
-        endif()
-    elseif (_compilerID MATCHES "Clang")
-        # Clang options used
-        # -H print the name of each header file used
-        # -E invoke preprocessor
-        # -fno-color-diagnostics don't prints diagnostics in color
-        if (_flags)
-            # append to list
-            list (APPEND _flags -H -E -fno-color-diagnostics)
-        else()
-            # return as a flag string
-            set (_flags "-H -E -fno-color-diagnostics")
-        endif()
-    elseif (_compilerID MATCHES "Intel")
-        if (WIN32)
-            # Windows Intel options used
-            # /nologo do not display compiler version information
-            # /QH display the include file order
-            # /EP preprocess to stdout, omitting #line directives
-            # /TC process all source or unrecognized file types as C source files
-            # /TP process all source or unrecognized file types as C++ source files
-            set (_sourceFileTypeC "/TC")
-            set (_sourceFileTypeCXX "/TP")
-            if (_flags)
-                # append to list
-                list (APPEND _flags /nologo "${_sourceFileType${_language}}" /EP /QH)
-            else()
-                # return as a flag string
-                set (_flags "${_sourceFileType${_language}} /EP /QH")
-            endif()
-        else()
-            # Linux / Mac OS X Intel options used
-            # -H print the name of each header file used
-            # -EP preprocess to stdout, omitting #line directives
-            # -Kc++ process all source or unrecognized file types as C++ source files
-            if (_flags)
-                # append to list
-                if ("${_language}" STREQUAL "CXX")
-                    list (APPEND _flags -Kc++)
-                endif()
-                list (APPEND _flags -H -EP)
-            else()
-                # return as a flag string
-                if ("${_language}" STREQUAL "CXX")
-                    set (_flags "-Kc++ ")
-                endif()
-                set (_flags "${_flags}-H -EP")
-            endif()
-        endif()
-    else()
-        message (FATAL_ERROR "cotire: unsupported ${_language} compiler ${_compilerID} version ${_compilerVersion}.")
-    endif()
-    set (${_flagsVar} ${_flags} PARENT_SCOPE)
-endfunction()
-
-function (cotire_add_pch_compilation_flags _language _compilerID _compilerVersion _prefixFile _pchFile _hostFile _flagsVar)
-    set (_flags ${${_flagsVar}})
-    if (_compilerID MATCHES "MSVC")
-        file (TO_NATIVE_PATH "${_prefixFile}" _prefixFileNative)
-        file (TO_NATIVE_PATH "${_pchFile}" _pchFileNative)
-        file (TO_NATIVE_PATH "${_hostFile}" _hostFileNative)
-        # cl.exe options used
-        # /Yc creates a precompiled header file
-        # /Fp specifies precompiled header binary file name
-        # /FI forces inclusion of file
-        # /TC treat all files named on the command line as C source files
-        # /TP treat all files named on the command line as C++ source files
-        # /Zs syntax check only
-        # /Zm precompiled header memory allocation scaling factor
-        set (_sourceFileTypeC "/TC")
-        set (_sourceFileTypeCXX "/TP")
-        if (_flags)
-            # append to list
-            list (APPEND _flags /nologo "${_sourceFileType${_language}}"
-                    "/Yc${_prefixFileNative}" "/Fp${_pchFileNative}" "/FI${_prefixFileNative}" /Zs "${_hostFileNative}")
-            if (COTIRE_PCH_MEMORY_SCALING_FACTOR)
-                list (APPEND _flags "/Zm${COTIRE_PCH_MEMORY_SCALING_FACTOR}")
-            endif()
-        else()
-            # return as a flag string
-            set (_flags "/Yc\"${_prefixFileNative}\" /Fp\"${_pchFileNative}\" /FI\"${_prefixFileNative}\"")
-            if (COTIRE_PCH_MEMORY_SCALING_FACTOR)
-                set (_flags "${_flags} /Zm${COTIRE_PCH_MEMORY_SCALING_FACTOR}")
-            endif()
-        endif()
-    elseif (_compilerID MATCHES "GNU|Clang")
-        # GCC / Clang options used
-        # -x specify the source language
-        # -c compile but do not link
-        # -o place output in file
-        # note that we cannot use -w to suppress all warnings upon pre-compiling, because turning off a warning may
-        # alter compile flags as a side effect (e.g., -Wwrite-string implies -fconst-strings)
-        set (_xLanguage_C "c-header")
-        set (_xLanguage_CXX "c++-header")
-        if (_flags)
-            # append to list
-            list (APPEND _flags "-x" "${_xLanguage_${_language}}" "-c" "${_prefixFile}" -o "${_pchFile}")
-        else()
-            # return as a flag string
-            set (_flags "-x ${_xLanguage_${_language}} -c \"${_prefixFile}\" -o \"${_pchFile}\"")
-        endif()
-    elseif (_compilerID MATCHES "Intel")
-        if (WIN32)
-            file (TO_NATIVE_PATH "${_prefixFile}" _prefixFileNative)
-            file (TO_NATIVE_PATH "${_pchFile}" _pchFileNative)
-            file (TO_NATIVE_PATH "${_hostFile}" _hostFileNative)
-            # Windows Intel options used
-            # /nologo do not display compiler version information
-            # /Yc create a precompiled header (PCH) file
-            # /Fp specify a path or file name for precompiled header files
-            # /FI tells the preprocessor to include a specified file name as the header file
-            # /TC process all source or unrecognized file types as C source files
-            # /TP process all source or unrecognized file types as C++ source files
-            # /Zs syntax check only
-            # /Wpch-messages enable diagnostics related to pre-compiled headers (requires Intel XE 2013 Update 2)
-            set (_sourceFileTypeC "/TC")
-            set (_sourceFileTypeCXX "/TP")
-            if (_flags)
-                # append to list
-                list (APPEND _flags /nologo "${_sourceFileType${_language}}"
-                        "/Yc" "/Fp${_pchFileNative}" "/FI${_prefixFileNative}" /Zs "${_hostFileNative}")
-                if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                    list (APPEND _flags "/Wpch-messages")
-                endif()
-            else()
-                # return as a flag string
-                set (_flags "/Yc /Fp\"${_pchFileNative}\" /FI\"${_prefixFileNative}\"")
-                if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                    set (_flags "${_flags} /Wpch-messages")
-                endif()
-            endif()
-        else()
-            # Linux / Mac OS X Intel options used
-            # -pch-dir location for precompiled header files
-            # -pch-create name of the precompiled header (PCH) to create
-            # -Kc++ process all source or unrecognized file types as C++ source files
-            # -fsyntax-only check only for correct syntax
-            # -Wpch-messages enable diagnostics related to pre-compiled headers (requires Intel XE 2013 Update 2)
-            get_filename_component(_pchDir "${_pchFile}" DIRECTORY)
-            get_filename_component(_pchName "${_pchFile}" NAME)
-            set (_xLanguage_C "c-header")
-            set (_xLanguage_CXX "c++-header")
-            set (_pchSuppressMessages FALSE)
-            if ("${CMAKE_${_language}_FLAGS}" MATCHES ".*-Wno-pch-messages.*")
-                set(_pchSuppressMessages TRUE)
-            endif()
-            if (_flags)
-                # append to list
-                if ("${_language}" STREQUAL "CXX")
-                    list (APPEND _flags -Kc++)
-                endif()
-                list (APPEND _flags "-include" "${_prefixFile}" "-pch-dir" "${_pchDir}" "-pch-create" "${_pchName}" "-fsyntax-only" "${_hostFile}")
-                if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                    if (NOT _pchSuppressMessages)
-                        list (APPEND _flags "-Wpch-messages")
-                    endif()
-                endif()
-            else()
-                # return as a flag string
-                set (_flags "-include \"${_prefixFile}\" -pch-dir \"${_pchDir}\" -pch-create \"${_pchName}\"")
-                if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                    if (NOT _pchSuppressMessages)
-                        set (_flags "${_flags} -Wpch-messages")
-                    endif()
-                endif()
-            endif()
-        endif()
-    else()
-        message (FATAL_ERROR "cotire: unsupported ${_language} compiler ${_compilerID} version ${_compilerVersion}.")
-    endif()
-    set (${_flagsVar} ${_flags} PARENT_SCOPE)
-endfunction()
-
-function (cotire_add_prefix_pch_inclusion_flags _language _compilerID _compilerVersion _prefixFile _pchFile _flagsVar)
-    set (_flags ${${_flagsVar}})
-    if (_compilerID MATCHES "MSVC")
-        file (TO_NATIVE_PATH "${_prefixFile}" _prefixFileNative)
-        # cl.exe options used
-        # /Yu uses a precompiled header file during build
-        # /Fp specifies precompiled header binary file name
-        # /FI forces inclusion of file
-        # /Zm precompiled header memory allocation scaling factor
-        if (_pchFile)
-            file (TO_NATIVE_PATH "${_pchFile}" _pchFileNative)
-            if (_flags)
-                # append to list
-                list (APPEND _flags "/Yu${_prefixFileNative}" "/Fp${_pchFileNative}" "/FI${_prefixFileNative}")
-                if (COTIRE_PCH_MEMORY_SCALING_FACTOR)
-                    list (APPEND _flags "/Zm${COTIRE_PCH_MEMORY_SCALING_FACTOR}")
-                endif()
-            else()
-                # return as a flag string
-                set (_flags "/Yu\"${_prefixFileNative}\" /Fp\"${_pchFileNative}\" /FI\"${_prefixFileNative}\"")
-                if (COTIRE_PCH_MEMORY_SCALING_FACTOR)
-                    set (_flags "${_flags} /Zm${COTIRE_PCH_MEMORY_SCALING_FACTOR}")
-                endif()
-            endif()
-        else()
-            # no precompiled header, force inclusion of prefix header
-            if (_flags)
-                # append to list
-                list (APPEND _flags "/FI${_prefixFileNative}")
-            else()
-                # return as a flag string
-                set (_flags "/FI\"${_prefixFileNative}\"")
-            endif()
-        endif()
-    elseif (_compilerID MATCHES "GNU")
-        # GCC options used
-        # -include process include file as the first line of the primary source file
-        # -Winvalid-pch warns if precompiled header is found but cannot be used
-        # note: ccache requires the -include flag to be used in order to process precompiled header correctly
-        if (_flags)
-            # append to list
-            list (APPEND _flags "-Winvalid-pch" "-include" "${_prefixFile}")
-        else()
-            # return as a flag string
-            set (_flags "-Winvalid-pch -include \"${_prefixFile}\"")
-        endif()
-    elseif (_compilerID MATCHES "Clang")
-        # Clang options used
-        # -include process include file as the first line of the primary source file
-        # -include-pch include precompiled header file
-        # -Qunused-arguments don't emit warning for unused driver arguments
-        # note: ccache requires the -include flag to be used in order to process precompiled header correctly
-        if (_flags)
-            # append to list
-            list (APPEND _flags "-Qunused-arguments" "-include" "${_prefixFile}")
-        else()
-            # return as a flag string
-            set (_flags "-Qunused-arguments -include \"${_prefixFile}\"")
-        endif()
-    elseif (_compilerID MATCHES "Intel")
-        if (WIN32)
-            file (TO_NATIVE_PATH "${_prefixFile}" _prefixFileNative)
-            # Windows Intel options used
-            # /Yu use a precompiled header (PCH) file
-            # /Fp specify a path or file name for precompiled header files
-            # /FI tells the preprocessor to include a specified file name as the header file
-            # /Wpch-messages enable diagnostics related to pre-compiled headers (requires Intel XE 2013 Update 2)
-            if (_pchFile)
-                file (TO_NATIVE_PATH "${_pchFile}" _pchFileNative)
-                if (_flags)
-                    # append to list
-                    list (APPEND _flags "/Yu" "/Fp${_pchFileNative}" "/FI${_prefixFileNative}")
-                    if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                        list (APPEND _flags "/Wpch-messages")
-                    endif()
-                else()
-                    # return as a flag string
-                    set (_flags "/Yu /Fp\"${_pchFileNative}\" /FI\"${_prefixFileNative}\"")
-                    if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                        set (_flags "${_flags} /Wpch-messages")
-                    endif()
-                endif()
-            else()
-                # no precompiled header, force inclusion of prefix header
-                if (_flags)
-                    # append to list
-                    list (APPEND _flags "/FI${_prefixFileNative}")
-                else()
-                    # return as a flag string
-                    set (_flags "/FI\"${_prefixFileNative}\"")
-                endif()
-            endif()
-        else()
-            # Linux / Mac OS X Intel options used
-            # -pch-dir location for precompiled header files
-            # -pch-use name of the precompiled header (PCH) to use
-            # -include process include file as the first line of the primary source file
-            # -Wpch-messages enable diagnostics related to pre-compiled headers (requires Intel XE 2013 Update 2)
-            if (_pchFile)
-                get_filename_component(_pchDir "${_pchFile}" DIRECTORY)
-                get_filename_component(_pchName "${_pchFile}" NAME)
-                set (_pchSuppressMessages FALSE)
-                if ("${CMAKE_${_language}_FLAGS}" MATCHES ".*-Wno-pch-messages.*")
-                    set(_pchSuppressMessages TRUE)
-                endif()
-                if (_flags)
-                    # append to list
-                    list (APPEND _flags "-include" "${_prefixFile}" "-pch-dir" "${_pchDir}" "-pch-use" "${_pchName}")
-                    if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                        if (NOT _pchSuppressMessages)
-                            list (APPEND _flags "-Wpch-messages")
-                        endif()
-                    endif()
-                else()
-                    # return as a flag string
-                    set (_flags "-include \"${_prefixFile}\" -pch-dir \"${_pchDir}\" -pch-use \"${_pchName}\"")
-                    if (NOT "${_compilerVersion}" VERSION_LESS "13.1.0")
-                        if (NOT _pchSuppressMessages)
-                            set (_flags "${_flags} -Wpch-messages")
-                        endif()
-                    endif()
-                endif()
-            else()
-                # no precompiled header, force inclusion of prefix header
-                if (_flags)
-                    # append to list
-                    list (APPEND _flags "-include" "${_prefixFile}")
-                else()
-                    # return as a flag string
-                    set (_flags "-include \"${_prefixFile}\"")
-                endif()
-            endif()
-        endif()
-    else()
-        message (FATAL_ERROR "cotire: unsupported ${_language} compiler ${_compilerID} version ${_compilerVersion}.")
-    endif()
-    set (${_flagsVar} ${_flags} PARENT_SCOPE)
-endfunction()
-
-function (cotire_precompile_prefix_header _prefixFile _pchFile _hostFile)
-    set(_options "")
-    set(_oneValueArgs COMPILER_EXECUTABLE COMPILER_ARG1 COMPILER_ID COMPILER_VERSION LANGUAGE)
-    set(_multiValueArgs COMPILE_DEFINITIONS COMPILE_FLAGS INCLUDE_DIRECTORIES SYSTEM_INCLUDE_DIRECTORIES SYS COMPILER_LAUNCHER)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    if (NOT _option_LANGUAGE)
-        set (_option_LANGUAGE "CXX")
-    endif()
-    if (NOT _option_COMPILER_ID)
-        set (_option_COMPILER_ID "${CMAKE_${_option_LANGUAGE}_ID}")
-    endif()
-    if (NOT _option_COMPILER_VERSION)
-        set (_option_COMPILER_VERSION "${CMAKE_${_option_LANGUAGE}_COMPILER_VERSION}")
-    endif()
-    cotire_init_compile_cmd(_cmd "${_option_LANGUAGE}" "${_option_COMPILER_LAUNCHER}" "${_option_COMPILER_EXECUTABLE}" "${_option_COMPILER_ARG1}")
-    cotire_add_definitions_to_cmd(_cmd "${_option_LANGUAGE}" ${_option_COMPILE_DEFINITIONS})
-    cotire_add_compile_flags_to_cmd(_cmd ${_option_COMPILE_FLAGS})
-    cotire_add_includes_to_cmd(_cmd "${_option_LANGUAGE}" _option_INCLUDE_DIRECTORIES _option_SYSTEM_INCLUDE_DIRECTORIES)
-    cotire_add_frameworks_to_cmd(_cmd "${_option_LANGUAGE}" _option_INCLUDE_DIRECTORIES _option_SYSTEM_INCLUDE_DIRECTORIES)
-    cotire_add_pch_compilation_flags(
-            "${_option_LANGUAGE}" "${_option_COMPILER_ID}" "${_option_COMPILER_VERSION}"
-            "${_prefixFile}" "${_pchFile}" "${_hostFile}" _cmd)
-    if (COTIRE_VERBOSE)
-        message (STATUS "execute_process: ${_cmd}")
-    endif()
-    if (_option_COMPILER_ID MATCHES "MSVC")
-        # cl.exe messes with the output streams unless the environment variable VS_UNICODE_OUTPUT is cleared
-        unset (ENV{VS_UNICODE_OUTPUT})
-    elseif (_option_COMPILER_ID MATCHES "GNU|Clang")
-        if (_option_COMPILER_LAUNCHER MATCHES "ccache" OR
-                _option_COMPILER_EXECUTABLE MATCHES "ccache")
-            # Newer versions of Clang and GCC seem to embed a compilation timestamp into the precompiled header binary,
-            # which results in "file has been modified since the precompiled header was built" errors if ccache is used.
-            # We work around the problem by disabling ccache upon pre-compiling the prefix header.
-            set (ENV{CCACHE_DISABLE} "true")
-        endif()
-    endif()
-    execute_process(
-            COMMAND ${_cmd}
-            WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-            RESULT_VARIABLE _result)
-    if (_result)
-        message (FATAL_ERROR "cotire: error ${_result} precompiling ${_prefixFile}.")
-    endif()
-endfunction()
-
-function (cotire_check_precompiled_header_support _language _target _msgVar)
-    set (_unsupportedCompiler
-            "Precompiled headers not supported for ${_language} compiler ${CMAKE_${_language}_COMPILER_ID}")
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "MSVC")
-        # supported since Visual Studio C++ 6.0
-        # and CMake does not support an earlier version
-        set (${_msgVar} "" PARENT_SCOPE)
-    elseif (CMAKE_${_language}_COMPILER_ID MATCHES "GNU")
-        # GCC PCH support requires version >= 3.4
-        if ("${CMAKE_${_language}_COMPILER_VERSION}" VERSION_LESS "3.4.0")
-            set (${_msgVar} "${_unsupportedCompiler} version ${CMAKE_${_language}_COMPILER_VERSION}." PARENT_SCOPE)
-        else()
-            set (${_msgVar} "" PARENT_SCOPE)
-        endif()
-    elseif (CMAKE_${_language}_COMPILER_ID MATCHES "Clang")
-        # all Clang versions have PCH support
-        set (${_msgVar} "" PARENT_SCOPE)
-    elseif (CMAKE_${_language}_COMPILER_ID MATCHES "Intel")
-        # Intel PCH support requires version >= 8.0.0
-        if ("${CMAKE_${_language}_COMPILER_VERSION}" VERSION_LESS "8.0.0")
-            set (${_msgVar} "${_unsupportedCompiler} version ${CMAKE_${_language}_COMPILER_VERSION}." PARENT_SCOPE)
-        else()
-            set (${_msgVar} "" PARENT_SCOPE)
-        endif()
-    else()
-        set (${_msgVar} "${_unsupportedCompiler}." PARENT_SCOPE)
-    endif()
-    get_target_property(_launcher ${_target} ${_language}_COMPILER_LAUNCHER)
-    if (CMAKE_${_language}_COMPILER MATCHES "ccache" OR _launcher MATCHES "ccache")
-        if (DEFINED ENV{CCACHE_SLOPPINESS})
-            if (NOT "$ENV{CCACHE_SLOPPINESS}" MATCHES "pch_defines" OR NOT "$ENV{CCACHE_SLOPPINESS}" MATCHES "time_macros")
-                set (${_msgVar}
-                        "ccache requires the environment variable CCACHE_SLOPPINESS to be set to \"pch_defines,time_macros\"."
-                        PARENT_SCOPE)
-            endif()
-        else()
-            if (_launcher MATCHES "ccache")
-                get_filename_component(_ccacheExe "${_launcher}" REALPATH)
-            else()
-                get_filename_component(_ccacheExe "${CMAKE_${_language}_COMPILER}" REALPATH)
-            endif()
-            execute_process(
-                    COMMAND "${_ccacheExe}" "--print-config"
-                    WORKING_DIRECTORY "${CMAKE_CURRENT_BINARY_DIR}"
-                    RESULT_VARIABLE _result
-                    OUTPUT_VARIABLE _ccacheConfig OUTPUT_STRIP_TRAILING_WHITESPACE
-                    ERROR_QUIET)
-            if (_result OR NOT
-                    _ccacheConfig MATCHES "sloppiness.*=.*time_macros" OR NOT
-                    _ccacheConfig MATCHES "sloppiness.*=.*pch_defines")
-                set (${_msgVar}
-                        "ccache requires configuration setting \"sloppiness\" to be set to \"pch_defines,time_macros\"."
-                        PARENT_SCOPE)
-            endif()
-        endif()
-    endif()
-    if (APPLE)
-        # PCH compilation not supported by GCC / Clang for multi-architecture builds (e.g., i386, x86_64)
-        cotire_get_configuration_types(_configs)
-        foreach (_config ${_configs})
-            set (_targetFlags "")
-            cotire_get_target_compile_flags("${_config}" "${_language}" "${_target}" _targetFlags)
-            cotire_filter_compile_flags("${_language}" "arch" _architectures _ignore ${_targetFlags})
-            list (LENGTH _architectures _numberOfArchitectures)
-            if (_numberOfArchitectures GREATER 1)
-                string (REPLACE ";" ", " _architectureStr "${_architectures}")
-                set (${_msgVar}
-                        "Precompiled headers not supported on Darwin for multi-architecture builds (${_architectureStr})."
-                        PARENT_SCOPE)
-                break()
-            endif()
-        endforeach()
-    endif()
-endfunction()
-
-macro (cotire_get_intermediate_dir _cotireDir)
-    # ${CMAKE_CFG_INTDIR} may reference a build-time variable when using a generator which supports configuration types
-    get_filename_component(${_cotireDir} "${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_CFG_INTDIR}/${COTIRE_INTDIR}" ABSOLUTE)
-endmacro()
-
-macro (cotire_setup_file_extension_variables)
-    set (_unityFileExt_C ".c")
-    set (_unityFileExt_CXX ".cxx")
-    set (_prefixFileExt_C ".h")
-    set (_prefixFileExt_CXX ".hxx")
-    set (_prefixSourceFileExt_C ".c")
-    set (_prefixSourceFileExt_CXX ".cxx")
-endmacro()
-
-function (cotire_make_single_unity_source_file_path _language _target _unityFileVar)
-    cotire_setup_file_extension_variables()
-    if (NOT DEFINED _unityFileExt_${_language})
-        set (${_unityFileVar} "" PARENT_SCOPE)
-        return()
-    endif()
-    set (_unityFileBaseName "${_target}_${_language}${COTIRE_UNITY_SOURCE_FILENAME_SUFFIX}")
-    set (_unityFileName "${_unityFileBaseName}${_unityFileExt_${_language}}")
-    cotire_get_intermediate_dir(_baseDir)
-    set (_unityFile "${_baseDir}/${_unityFileName}")
-    set (${_unityFileVar} "${_unityFile}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_make_unity_source_file_paths _language _target _maxIncludes _unityFilesVar)
-    cotire_setup_file_extension_variables()
-    if (NOT DEFINED _unityFileExt_${_language})
-        set (${_unityFileVar} "" PARENT_SCOPE)
-        return()
-    endif()
-    set (_unityFileBaseName "${_target}_${_language}${COTIRE_UNITY_SOURCE_FILENAME_SUFFIX}")
-    cotire_get_intermediate_dir(_baseDir)
-    set (_startIndex 0)
-    set (_index 0)
-    set (_unityFiles "")
-    set (_sourceFiles ${ARGN})
-    foreach (_sourceFile ${_sourceFiles})
-        get_source_file_property(_startNew "${_sourceFile}" COTIRE_START_NEW_UNITY_SOURCE)
-        math (EXPR _unityFileCount "${_index} - ${_startIndex}")
-        if (_startNew OR (_maxIncludes GREATER 0 AND NOT _unityFileCount LESS _maxIncludes))
-            if (_index GREATER 0)
-                # start new unity file segment
-                math (EXPR _endIndex "${_index} - 1")
-                set (_unityFileName "${_unityFileBaseName}_${_startIndex}_${_endIndex}${_unityFileExt_${_language}}")
-                list (APPEND _unityFiles "${_baseDir}/${_unityFileName}")
-            endif()
-            set (_startIndex ${_index})
-        endif()
-        math (EXPR _index "${_index} + 1")
-    endforeach()
-    list (LENGTH _sourceFiles _numberOfSources)
-    if (_startIndex EQUAL 0)
-        # there is only a single unity file
-        cotire_make_single_unity_source_file_path(${_language} ${_target} _unityFiles)
-    elseif (_startIndex LESS _numberOfSources)
-        # end with final unity file segment
-        math (EXPR _endIndex "${_index} - 1")
-        set (_unityFileName "${_unityFileBaseName}_${_startIndex}_${_endIndex}${_unityFileExt_${_language}}")
-        list (APPEND _unityFiles "${_baseDir}/${_unityFileName}")
-    endif()
-    set (${_unityFilesVar} ${_unityFiles} PARENT_SCOPE)
-    if (COTIRE_DEBUG AND _unityFiles)
-        message (STATUS "unity files: ${_unityFiles}")
-    endif()
-endfunction()
-
-function (cotire_unity_to_prefix_file_path _language _target _unityFile _prefixFileVar)
-    cotire_setup_file_extension_variables()
-    if (NOT DEFINED _unityFileExt_${_language})
-        set (${_prefixFileVar} "" PARENT_SCOPE)
-        return()
-    endif()
-    set (_unityFileBaseName "${_target}_${_language}${COTIRE_UNITY_SOURCE_FILENAME_SUFFIX}")
-    set (_prefixFileBaseName "${_target}_${_language}${COTIRE_PREFIX_HEADER_FILENAME_SUFFIX}")
-    string (REPLACE "${_unityFileBaseName}" "${_prefixFileBaseName}" _prefixFile "${_unityFile}")
-    string (REGEX REPLACE "${_unityFileExt_${_language}}$" "${_prefixFileExt_${_language}}" _prefixFile "${_prefixFile}")
-    set (${_prefixFileVar} "${_prefixFile}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_prefix_header_to_source_file_path _language _prefixHeaderFile _prefixSourceFileVar)
-    cotire_setup_file_extension_variables()
-    if (NOT DEFINED _prefixSourceFileExt_${_language})
-        set (${_prefixSourceFileVar} "" PARENT_SCOPE)
-        return()
-    endif()
-    string (REGEX REPLACE "${_prefixFileExt_${_language}}$" "${_prefixSourceFileExt_${_language}}" _prefixSourceFile "${_prefixHeaderFile}")
-    set (${_prefixSourceFileVar} "${_prefixSourceFile}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_make_prefix_file_name _language _target _prefixFileBaseNameVar _prefixFileNameVar)
-    cotire_setup_file_extension_variables()
-    if (NOT _language)
-        set (_prefixFileBaseName "${_target}${COTIRE_PREFIX_HEADER_FILENAME_SUFFIX}")
-        set (_prefixFileName "${_prefixFileBaseName}${_prefixFileExt_C}")
-    elseif (DEFINED _prefixFileExt_${_language})
-        set (_prefixFileBaseName "${_target}_${_language}${COTIRE_PREFIX_HEADER_FILENAME_SUFFIX}")
-        set (_prefixFileName "${_prefixFileBaseName}${_prefixFileExt_${_language}}")
-    else()
-        set (_prefixFileBaseName "")
-        set (_prefixFileName "")
-    endif()
-    set (${_prefixFileBaseNameVar} "${_prefixFileBaseName}" PARENT_SCOPE)
-    set (${_prefixFileNameVar} "${_prefixFileName}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_make_prefix_file_path _language _target _prefixFileVar)
-    cotire_make_prefix_file_name("${_language}" "${_target}" _prefixFileBaseName _prefixFileName)
-    set (${_prefixFileVar} "" PARENT_SCOPE)
-    if (_prefixFileName)
-        if (NOT _language)
-            set (_language "C")
-        endif()
-        if (CMAKE_${_language}_COMPILER_ID MATCHES "GNU|Clang|Intel|MSVC")
-            cotire_get_intermediate_dir(_baseDir)
-            set (${_prefixFileVar} "${_baseDir}/${_prefixFileName}" PARENT_SCOPE)
-        endif()
-    endif()
-endfunction()
-
-function (cotire_make_pch_file_path _language _target _pchFileVar)
-    cotire_make_prefix_file_name("${_language}" "${_target}" _prefixFileBaseName _prefixFileName)
-    set (${_pchFileVar} "" PARENT_SCOPE)
-    if (_prefixFileBaseName AND _prefixFileName)
-        cotire_check_precompiled_header_support("${_language}" "${_target}" _msg)
-        if (NOT _msg)
-            if (XCODE)
-                # For Xcode, we completely hand off the compilation of the prefix header to the IDE
-                return()
-            endif()
-            cotire_get_intermediate_dir(_baseDir)
-            if (CMAKE_${_language}_COMPILER_ID MATCHES "MSVC")
-                # MSVC uses the extension .pch added to the prefix header base name
-                set (${_pchFileVar} "${_baseDir}/${_prefixFileBaseName}.pch" PARENT_SCOPE)
-            elseif (CMAKE_${_language}_COMPILER_ID MATCHES "Clang")
-                # Clang looks for a precompiled header corresponding to the prefix header with the extension .pch appended
-                set (${_pchFileVar} "${_baseDir}/${_prefixFileName}.pch" PARENT_SCOPE)
-            elseif (CMAKE_${_language}_COMPILER_ID MATCHES "GNU")
-                # GCC looks for a precompiled header corresponding to the prefix header with the extension .gch appended
-                set (${_pchFileVar} "${_baseDir}/${_prefixFileName}.gch" PARENT_SCOPE)
-            elseif (CMAKE_${_language}_COMPILER_ID MATCHES "Intel")
-                # Intel uses the extension .pchi added to the prefix header base name
-                set (${_pchFileVar} "${_baseDir}/${_prefixFileBaseName}.pchi" PARENT_SCOPE)
-            endif()
-        endif()
-    endif()
-endfunction()
-
-function (cotire_select_unity_source_files _unityFile _sourcesVar)
-    set (_sourceFiles ${ARGN})
-    if (_sourceFiles AND "${_unityFile}" MATCHES "${COTIRE_UNITY_SOURCE_FILENAME_SUFFIX}_([0-9]+)_([0-9]+)")
-        set (_startIndex ${CMAKE_MATCH_1})
-        set (_endIndex ${CMAKE_MATCH_2})
-        list (LENGTH _sourceFiles _numberOfSources)
-        if (NOT _startIndex LESS _numberOfSources)
-            math (EXPR _startIndex "${_numberOfSources} - 1")
-        endif()
-        if (NOT _endIndex LESS _numberOfSources)
-            math (EXPR _endIndex "${_numberOfSources} - 1")
-        endif()
-        set (_files "")
-        foreach (_index RANGE ${_startIndex} ${_endIndex})
-            list (GET _sourceFiles ${_index} _file)
-            list (APPEND _files "${_file}")
-        endforeach()
-    else()
-        set (_files ${_sourceFiles})
-    endif()
-    set (${_sourcesVar} ${_files} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_unity_source_dependencies _language _target _dependencySourcesVar)
-    set (_dependencySources "")
-    # depend on target's generated source files
-    get_target_property(_targetSourceFiles ${_target} SOURCES)
-    cotire_get_objects_with_property_on(_generatedSources GENERATED SOURCE ${_targetSourceFiles})
-    if (_generatedSources)
-        # but omit all generated source files that have the COTIRE_EXCLUDED property set to true
-        cotire_get_objects_with_property_on(_excludedGeneratedSources COTIRE_EXCLUDED SOURCE ${_generatedSources})
-        if (_excludedGeneratedSources)
-            list (REMOVE_ITEM _generatedSources ${_excludedGeneratedSources})
-        endif()
-        # and omit all generated source files that have the COTIRE_DEPENDENCY property set to false explicitly
-        cotire_get_objects_with_property_off(_excludedNonDependencySources COTIRE_DEPENDENCY SOURCE ${_generatedSources})
-        if (_excludedNonDependencySources)
-            list (REMOVE_ITEM _generatedSources ${_excludedNonDependencySources})
-        endif()
-        if (_generatedSources)
-            list (APPEND _dependencySources ${_generatedSources})
-        endif()
-    endif()
-    if (COTIRE_DEBUG AND _dependencySources)
-        message (STATUS "${_language} ${_target} unity source dependencies: ${_dependencySources}")
-    endif()
-    set (${_dependencySourcesVar} ${_dependencySources} PARENT_SCOPE)
-endfunction()
-
-function (cotire_get_prefix_header_dependencies _language _target _dependencySourcesVar)
-    set (_dependencySources "")
-    # depend on target source files marked with custom COTIRE_DEPENDENCY property
-    get_target_property(_targetSourceFiles ${_target} SOURCES)
-    cotire_get_objects_with_property_on(_dependencySources COTIRE_DEPENDENCY SOURCE ${_targetSourceFiles})
-    if (COTIRE_DEBUG AND _dependencySources)
-        message (STATUS "${_language} ${_target} prefix header dependencies: ${_dependencySources}")
-    endif()
-    set (${_dependencySourcesVar} ${_dependencySources} PARENT_SCOPE)
-endfunction()
-
-function (cotire_generate_target_script _language _configurations _target _targetScriptVar _targetConfigScriptVar)
-    set (_targetSources ${ARGN})
-    cotire_get_prefix_header_dependencies(${_language} ${_target} COTIRE_TARGET_PREFIX_DEPENDS ${_targetSources})
-    cotire_get_unity_source_dependencies(${_language} ${_target} COTIRE_TARGET_UNITY_DEPENDS ${_targetSources})
-    # set up variables to be configured
-    set (COTIRE_TARGET_LANGUAGE "${_language}")
-    get_target_property(COTIRE_TARGET_IGNORE_PATH ${_target} COTIRE_PREFIX_HEADER_IGNORE_PATH)
-    cotire_add_sys_root_paths(COTIRE_TARGET_IGNORE_PATH)
-    get_target_property(COTIRE_TARGET_INCLUDE_PATH ${_target} COTIRE_PREFIX_HEADER_INCLUDE_PATH)
-    cotire_add_sys_root_paths(COTIRE_TARGET_INCLUDE_PATH)
-    get_target_property(COTIRE_TARGET_PRE_UNDEFS ${_target} COTIRE_UNITY_SOURCE_PRE_UNDEFS)
-    get_target_property(COTIRE_TARGET_POST_UNDEFS ${_target} COTIRE_UNITY_SOURCE_POST_UNDEFS)
-    get_target_property(COTIRE_TARGET_MAXIMUM_NUMBER_OF_INCLUDES ${_target} COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES)
-    get_target_property(COTIRE_TARGET_INCLUDE_PRIORITY_PATH ${_target} COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH)
-    cotire_get_source_files_undefs(COTIRE_UNITY_SOURCE_PRE_UNDEFS COTIRE_TARGET_SOURCES_PRE_UNDEFS ${_targetSources})
-    cotire_get_source_files_undefs(COTIRE_UNITY_SOURCE_POST_UNDEFS COTIRE_TARGET_SOURCES_POST_UNDEFS ${_targetSources})
-    set (COTIRE_TARGET_CONFIGURATION_TYPES "${_configurations}")
-    foreach (_config ${_configurations})
-        string (TOUPPER "${_config}" _upperConfig)
-        cotire_get_target_include_directories(
-                "${_config}" "${_language}" "${_target}" COTIRE_TARGET_INCLUDE_DIRECTORIES_${_upperConfig} COTIRE_TARGET_SYSTEM_INCLUDE_DIRECTORIES_${_upperConfig})
-        cotire_get_target_compile_definitions(
-                "${_config}" "${_language}" "${_target}" COTIRE_TARGET_COMPILE_DEFINITIONS_${_upperConfig})
-        cotire_get_target_compiler_flags(
-                "${_config}" "${_language}" "${_target}" COTIRE_TARGET_COMPILE_FLAGS_${_upperConfig})
-        cotire_get_source_files_compile_definitions(
-                "${_config}" "${_language}" COTIRE_TARGET_SOURCES_COMPILE_DEFINITIONS_${_upperConfig} ${_targetSources})
-    endforeach()
-    get_target_property(COTIRE_TARGET_${_language}_COMPILER_LAUNCHER ${_target} ${_language}_COMPILER_LAUNCHER)
-    # set up COTIRE_TARGET_SOURCES
-    set (COTIRE_TARGET_SOURCES "")
-    foreach (_sourceFile ${_targetSources})
-        get_source_file_property(_generated "${_sourceFile}" GENERATED)
-        if (_generated)
-            # use absolute paths for generated files only, retrieving the LOCATION property is an expensive operation
-            get_source_file_property(_sourceLocation "${_sourceFile}" LOCATION)
-            list (APPEND COTIRE_TARGET_SOURCES "${_sourceLocation}")
-        else()
-            list (APPEND COTIRE_TARGET_SOURCES "${_sourceFile}")
-        endif()
-    endforeach()
-    # copy variable definitions to cotire target script
-    get_cmake_property(_vars VARIABLES)
-    string (REGEX MATCHALL "COTIRE_[A-Za-z0-9_]+" _matchVars "${_vars}")
-    # omit COTIRE_*_INIT variables
-    string (REGEX MATCHALL "COTIRE_[A-Za-z0-9_]+_INIT" _initVars "${_matchVars}")
-    if (_initVars)
-        list (REMOVE_ITEM _matchVars ${_initVars})
-    endif()
-    # omit COTIRE_VERBOSE which is passed as a CMake define on command line
-    list (REMOVE_ITEM _matchVars COTIRE_VERBOSE)
-    set (_contents "")
-    set (_contentsHasGeneratorExpressions FALSE)
-    foreach (_var IN LISTS _matchVars ITEMS
-            XCODE MSVC CMAKE_GENERATOR CMAKE_BUILD_TYPE CMAKE_CONFIGURATION_TYPES
-            CMAKE_${_language}_COMPILER_ID CMAKE_${_language}_COMPILER_VERSION
-            CMAKE_${_language}_COMPILER_LAUNCHER CMAKE_${_language}_COMPILER CMAKE_${_language}_COMPILER_ARG1
-            CMAKE_INCLUDE_FLAG_${_language} CMAKE_INCLUDE_FLAG_SEP_${_language}
-            CMAKE_INCLUDE_SYSTEM_FLAG_${_language}
-            CMAKE_${_language}_FRAMEWORK_SEARCH_FLAG
-            CMAKE_${_language}_SYSTEM_FRAMEWORK_SEARCH_FLAG
-            CMAKE_${_language}_SOURCE_FILE_EXTENSIONS)
-        if (DEFINED ${_var})
-            string (REPLACE "\"" "\\\"" _value "${${_var}}")
-            set (_contents "${_contents}set (${_var} \"${_value}\")\n")
-            if (NOT _contentsHasGeneratorExpressions)
-                if ("${_value}" MATCHES "\\$<.*>")
-                    set (_contentsHasGeneratorExpressions TRUE)
-                endif()
-            endif()
-        endif()
-    endforeach()
-    # generate target script file
-    get_filename_component(_moduleName "${COTIRE_CMAKE_MODULE_FILE}" NAME)
-    set (_targetCotireScript "${CMAKE_CURRENT_BINARY_DIR}/${_target}_${_language}_${_moduleName}")
-    cotire_write_file("CMAKE" "${_targetCotireScript}" "${_contents}" FALSE)
-    if (_contentsHasGeneratorExpressions)
-        # use file(GENERATE ...) to expand generator expressions in the target script at CMake generate-time
-        set (_configNameOrNoneGeneratorExpression "$<$<CONFIG:>:None>$<$<NOT:$<CONFIG:>>:$<CONFIGURATION>>")
-        set (_targetCotireConfigScript "${CMAKE_CURRENT_BINARY_DIR}/${_target}_${_language}_${_configNameOrNoneGeneratorExpression}_${_moduleName}")
-        file (GENERATE OUTPUT "${_targetCotireConfigScript}" INPUT "${_targetCotireScript}")
-    else()
-        set (_targetCotireConfigScript "${_targetCotireScript}")
-    endif()
-    set (${_targetScriptVar} "${_targetCotireScript}" PARENT_SCOPE)
-    set (${_targetConfigScriptVar} "${_targetCotireConfigScript}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_pch_file_compilation _language _target _targetScript _prefixFile _pchFile _hostFile)
-    set (_sourceFiles ${ARGN})
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-        # for Visual Studio and Intel, we attach the precompiled header compilation to the host file
-        # the remaining files include the precompiled header, see cotire_setup_pch_file_inclusion
-        if (_sourceFiles)
-            set (_flags "")
-            cotire_add_pch_compilation_flags(
-                    "${_language}" "${CMAKE_${_language}_COMPILER_ID}" "${CMAKE_${_language}_COMPILER_VERSION}"
-                    "${_prefixFile}" "${_pchFile}" "${_hostFile}" _flags)
-            set_property (SOURCE ${_hostFile} APPEND_STRING PROPERTY COMPILE_FLAGS " ${_flags} ")
-            set_property (SOURCE ${_hostFile} APPEND PROPERTY OBJECT_OUTPUTS "${_pchFile}")
-            # make object file generated from host file depend on prefix header
-            set_property (SOURCE ${_hostFile} APPEND PROPERTY OBJECT_DEPENDS "${_prefixFile}")
-            # mark host file as cotired to prevent it from being used in another cotired target
-            set_property (SOURCE ${_hostFile} PROPERTY COTIRE_TARGET "${_target}")
-        endif()
-    elseif ("${CMAKE_GENERATOR}" MATCHES "Make|Ninja")
-        # for makefile based generator, we add a custom command to precompile the prefix header
-        if (_targetScript)
-            cotire_set_cmd_to_prologue(_cmds)
-            list (APPEND _cmds -P "${COTIRE_CMAKE_MODULE_FILE}" "precompile" "${_targetScript}" "${_prefixFile}" "${_pchFile}" "${_hostFile}")
-            if (MSVC_IDE)
-                file (TO_NATIVE_PATH "${_pchFile}" _pchFileLogPath)
-            else()
-                file (RELATIVE_PATH _pchFileLogPath "${CMAKE_BINARY_DIR}" "${_pchFile}")
-            endif()
-            # make precompiled header compilation depend on the actual compiler executable used to force
-            # re-compilation when the compiler executable is updated. This prevents "created by a different GCC executable"
-            # warnings when the precompiled header is included.
-            get_filename_component(_realCompilerExe "${CMAKE_${_language}_COMPILER}" ABSOLUTE)
-            if (COTIRE_DEBUG)
-                message (STATUS "add_custom_command: OUTPUT ${_pchFile} ${_cmds} DEPENDS ${_prefixFile} ${_realCompilerExe} IMPLICIT_DEPENDS ${_language} ${_prefixFile}")
-            endif()
-            set_property (SOURCE "${_pchFile}" PROPERTY GENERATED TRUE)
-            add_custom_command(
-                    OUTPUT "${_pchFile}"
-                    COMMAND ${_cmds}
-                    DEPENDS "${_prefixFile}" "${_realCompilerExe}"
-                    IMPLICIT_DEPENDS ${_language} "${_prefixFile}"
-                    WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-                    COMMENT "Building ${_language} precompiled header ${_pchFileLogPath}"
-                    VERBATIM)
-        endif()
-    endif()
-endfunction()
-
-function (cotire_setup_pch_file_inclusion _language _target _wholeTarget _prefixFile _pchFile _hostFile)
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-        # for Visual Studio and Intel, we include the precompiled header in all but the host file
-        # the host file does the precompiled header compilation, see cotire_setup_pch_file_compilation
-        set (_sourceFiles ${ARGN})
-        list (LENGTH _sourceFiles _numberOfSourceFiles)
-        if (_numberOfSourceFiles GREATER 0)
-            # mark sources as cotired to prevent them from being used in another cotired target
-            set_source_files_properties(${_sourceFiles} PROPERTIES COTIRE_TARGET "${_target}")
-            set (_flags "")
-            cotire_add_prefix_pch_inclusion_flags(
-                    "${_language}" "${CMAKE_${_language}_COMPILER_ID}" "${CMAKE_${_language}_COMPILER_VERSION}"
-                    "${_prefixFile}" "${_pchFile}" _flags)
-            set_property (SOURCE ${_sourceFiles} APPEND_STRING PROPERTY COMPILE_FLAGS " ${_flags} ")
-            # make object files generated from source files depend on precompiled header
-            set_property (SOURCE ${_sourceFiles} APPEND PROPERTY OBJECT_DEPENDS "${_pchFile}")
-        endif()
-    elseif ("${CMAKE_GENERATOR}" MATCHES "Make|Ninja")
-        set (_sourceFiles ${_hostFile} ${ARGN})
-        if (NOT _wholeTarget)
-            # for makefile based generator, we force the inclusion of the prefix header for a subset
-            # of the source files, if this is a multi-language target or has excluded files
-            set (_flags "")
-            cotire_add_prefix_pch_inclusion_flags(
-                    "${_language}" "${CMAKE_${_language}_COMPILER_ID}" "${CMAKE_${_language}_COMPILER_VERSION}"
-                    "${_prefixFile}" "${_pchFile}" _flags)
-            set_property (SOURCE ${_sourceFiles} APPEND_STRING PROPERTY COMPILE_FLAGS " ${_flags} ")
-            # mark sources as cotired to prevent them from being used in another cotired target
-            set_source_files_properties(${_sourceFiles} PROPERTIES COTIRE_TARGET "${_target}")
-        endif()
-        # make object files generated from source files depend on precompiled header
-        set_property (SOURCE ${_sourceFiles} APPEND PROPERTY OBJECT_DEPENDS "${_pchFile}")
-    endif()
-endfunction()
-
-function (cotire_setup_prefix_file_inclusion _language _target _prefixFile)
-    set (_sourceFiles ${ARGN})
-    # force the inclusion of the prefix header for the given source files
-    set (_flags "")
-    set (_pchFile "")
-    cotire_add_prefix_pch_inclusion_flags(
-            "${_language}" "${CMAKE_${_language}_COMPILER_ID}" "${CMAKE_${_language}_COMPILER_VERSION}"
-            "${_prefixFile}" "${_pchFile}" _flags)
-    set_property (SOURCE ${_sourceFiles} APPEND_STRING PROPERTY COMPILE_FLAGS " ${_flags} ")
-    # mark sources as cotired to prevent them from being used in another cotired target
-    set_source_files_properties(${_sourceFiles} PROPERTIES COTIRE_TARGET "${_target}")
-    # make object files generated from source files depend on prefix header
-    set_property (SOURCE ${_sourceFiles} APPEND PROPERTY OBJECT_DEPENDS "${_prefixFile}")
-endfunction()
-
-function (cotire_get_first_set_property_value _propertyValueVar _type _object)
-    set (_properties ${ARGN})
-    foreach (_property ${_properties})
-        get_property(_propertyValue ${_type} "${_object}" PROPERTY ${_property})
-        if (_propertyValue)
-            set (${_propertyValueVar} ${_propertyValue} PARENT_SCOPE)
-            return()
-        endif()
-    endforeach()
-    set (${_propertyValueVar} "" PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_combine_command _language _targetScript _joinedFile _cmdsVar)
-    set (_files ${ARGN})
-    set (_filesPaths "")
-    foreach (_file ${_files})
-        get_filename_component(_filePath "${_file}" ABSOLUTE)
-        list (APPEND _filesPaths "${_filePath}")
-    endforeach()
-    cotire_set_cmd_to_prologue(_prefixCmd)
-    list (APPEND _prefixCmd -P "${COTIRE_CMAKE_MODULE_FILE}" "combine")
-    if (_targetScript)
-        list (APPEND _prefixCmd "${_targetScript}")
-    endif()
-    list (APPEND _prefixCmd "${_joinedFile}" ${_filesPaths})
-    if (COTIRE_DEBUG)
-        message (STATUS "add_custom_command: OUTPUT ${_joinedFile} COMMAND ${_prefixCmd} DEPENDS ${_files}")
-    endif()
-    set_property (SOURCE "${_joinedFile}" PROPERTY GENERATED TRUE)
-    if (MSVC_IDE)
-        file (TO_NATIVE_PATH "${_joinedFile}" _joinedFileLogPath)
-    else()
-        file (RELATIVE_PATH _joinedFileLogPath "${CMAKE_BINARY_DIR}" "${_joinedFile}")
-    endif()
-    get_filename_component(_joinedFileBaseName "${_joinedFile}" NAME_WE)
-    get_filename_component(_joinedFileExt "${_joinedFile}" EXT)
-    if (_language AND _joinedFileBaseName MATCHES "${COTIRE_UNITY_SOURCE_FILENAME_SUFFIX}$")
-        set (_comment "Generating ${_language} unity source ${_joinedFileLogPath}")
-    elseif (_language AND _joinedFileBaseName MATCHES "${COTIRE_PREFIX_HEADER_FILENAME_SUFFIX}$")
-        if (_joinedFileExt MATCHES "^\\.c")
-            set (_comment "Generating ${_language} prefix source ${_joinedFileLogPath}")
-        else()
-            set (_comment "Generating ${_language} prefix header ${_joinedFileLogPath}")
-        endif()
-    else()
-        set (_comment "Generating ${_joinedFileLogPath}")
-    endif()
-    add_custom_command(
-            OUTPUT "${_joinedFile}"
-            COMMAND ${_prefixCmd}
-            DEPENDS ${_files}
-            COMMENT "${_comment}"
-            WORKING_DIRECTORY "${CMAKE_BINARY_DIR}"
-            VERBATIM)
-    list (APPEND ${_cmdsVar} COMMAND ${_prefixCmd})
-    set (${_cmdsVar} ${${_cmdsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_target_pch_usage _languages _target _wholeTarget)
-    if (XCODE)
-        # for Xcode, we attach a pre-build action to generate the unity sources and prefix headers
-        set (_prefixFiles "")
-        foreach (_language ${_languages})
-            get_property(_prefixFile TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER)
-            if (_prefixFile)
-                list (APPEND _prefixFiles "${_prefixFile}")
-            endif()
-        endforeach()
-        set (_cmds ${ARGN})
-        list (LENGTH _prefixFiles _numberOfPrefixFiles)
-        if (_numberOfPrefixFiles GREATER 1)
-            # we also generate a generic, single prefix header which includes all language specific prefix headers
-            set (_language "")
-            set (_targetScript "")
-            cotire_make_prefix_file_path("${_language}" ${_target} _prefixHeader)
-            cotire_setup_combine_command("${_language}" "${_targetScript}" "${_prefixHeader}" _cmds ${_prefixFiles})
-        else()
-            set (_prefixHeader "${_prefixFiles}")
-        endif()
-        if (COTIRE_DEBUG)
-            message (STATUS "add_custom_command: TARGET ${_target} PRE_BUILD ${_cmds}")
-        endif()
-        # because CMake PRE_BUILD command does not support dependencies,
-        # we check dependencies explicity in cotire script mode when the pre-build action is run
-        add_custom_command(
-                TARGET "${_target}"
-                PRE_BUILD ${_cmds}
-                WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-                COMMENT "Updating target ${_target} prefix headers"
-                VERBATIM)
-        # make Xcode precompile the generated prefix header with ProcessPCH and ProcessPCH++
-        set_target_properties(${_target} PROPERTIES XCODE_ATTRIBUTE_GCC_PRECOMPILE_PREFIX_HEADER "YES")
-        set_target_properties(${_target} PROPERTIES XCODE_ATTRIBUTE_GCC_PREFIX_HEADER "${_prefixHeader}")
-    elseif ("${CMAKE_GENERATOR}" MATCHES "Make|Ninja")
-        # for makefile based generator, we force inclusion of the prefix header for all target source files
-        # if this is a single-language target without any excluded files
-        if (_wholeTarget)
-            set (_language "${_languages}")
-            # for Visual Studio and Intel, precompiled header inclusion is always done on the source file level
-            # see cotire_setup_pch_file_inclusion
-            if (NOT CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-                get_property(_prefixFile TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER)
-                if (_prefixFile)
-                    get_property(_pchFile TARGET ${_target} PROPERTY COTIRE_${_language}_PRECOMPILED_HEADER)
-                    set (_options COMPILE_OPTIONS)
-                    cotire_add_prefix_pch_inclusion_flags(
-                            "${_language}" "${CMAKE_${_language}_COMPILER_ID}" "${CMAKE_${_language}_COMPILER_VERSION}"
-                            "${_prefixFile}" "${_pchFile}" _options)
-                    set_property(TARGET ${_target} APPEND PROPERTY ${_options})
-                endif()
-            endif()
-        endif()
-    endif()
-endfunction()
-
-function (cotire_setup_unity_generation_commands _language _target _targetScript _targetConfigScript _unityFiles _cmdsVar)
-    set (_dependencySources "")
-    cotire_get_unity_source_dependencies(${_language} ${_target} _dependencySources ${ARGN})
-    foreach (_unityFile ${_unityFiles})
-        set_property (SOURCE "${_unityFile}" PROPERTY GENERATED TRUE)
-        # set up compiled unity source dependencies via OBJECT_DEPENDS
-        # this ensures that missing source files are generated before the unity file is compiled
-        if (COTIRE_DEBUG AND _dependencySources)
-            message (STATUS "${_unityFile} OBJECT_DEPENDS ${_dependencySources}")
-        endif()
-        if (_dependencySources)
-            # the OBJECT_DEPENDS property requires a list of full paths
-            set (_objectDependsPaths "")
-            foreach (_sourceFile ${_dependencySources})
-                get_source_file_property(_sourceLocation "${_sourceFile}" LOCATION)
-                list (APPEND _objectDependsPaths "${_sourceLocation}")
-            endforeach()
-            set_property (SOURCE "${_unityFile}" PROPERTY OBJECT_DEPENDS ${_objectDependsPaths})
-        endif()
-        if (WIN32 AND CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-            # unity file compilation results in potentially huge object file, thus use /bigobj by default unter MSVC and Windows Intel
-            set_property (SOURCE "${_unityFile}" APPEND_STRING PROPERTY COMPILE_FLAGS "/bigobj")
-        endif()
-        cotire_set_cmd_to_prologue(_unityCmd)
-        list (APPEND _unityCmd -P "${COTIRE_CMAKE_MODULE_FILE}" "unity" "${_targetConfigScript}" "${_unityFile}")
-        if (CMAKE_VERSION VERSION_LESS "3.1.0")
-            set (_unityCmdDepends "${_targetScript}")
-        else()
-            # CMake 3.1.0 supports generator expressions in arguments to DEPENDS
-            set (_unityCmdDepends "${_targetConfigScript}")
-        endif()
-        if (MSVC_IDE)
-            file (TO_NATIVE_PATH "${_unityFile}" _unityFileLogPath)
-        else()
-            file (RELATIVE_PATH _unityFileLogPath "${CMAKE_BINARY_DIR}" "${_unityFile}")
-        endif()
-        if (COTIRE_DEBUG)
-            message (STATUS "add_custom_command: OUTPUT ${_unityFile} COMMAND ${_unityCmd} DEPENDS ${_unityCmdDepends}")
-        endif()
-        add_custom_command(
-                OUTPUT "${_unityFile}"
-                COMMAND ${_unityCmd}
-                DEPENDS ${_unityCmdDepends}
-                COMMENT "Generating ${_language} unity source ${_unityFileLogPath}"
-                WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-                VERBATIM)
-        list (APPEND ${_cmdsVar} COMMAND ${_unityCmd})
-    endforeach()
-    set (${_cmdsVar} ${${_cmdsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_prefix_generation_command _language _target _targetScript _prefixFile _unityFiles _cmdsVar)
-    set (_sourceFiles ${ARGN})
-    set (_dependencySources "")
-    cotire_get_prefix_header_dependencies(${_language} ${_target} _dependencySources ${_sourceFiles})
-    cotire_set_cmd_to_prologue(_prefixCmd)
-    list (APPEND _prefixCmd -P "${COTIRE_CMAKE_MODULE_FILE}" "prefix" "${_targetScript}" "${_prefixFile}" ${_unityFiles})
-    set_property (SOURCE "${_prefixFile}" PROPERTY GENERATED TRUE)
-    # make prefix header generation depend on the actual compiler executable used to force
-    # re-generation when the compiler executable is updated. This prevents "file not found"
-    # errors for compiler version specific system header files.
-    get_filename_component(_realCompilerExe "${CMAKE_${_language}_COMPILER}" ABSOLUTE)
-    if (COTIRE_DEBUG)
-        message (STATUS "add_custom_command: OUTPUT ${_prefixFile} COMMAND ${_prefixCmd} DEPENDS ${_unityFile} ${_dependencySources} ${_realCompilerExe}")
-    endif()
-    if (MSVC_IDE)
-        file (TO_NATIVE_PATH "${_prefixFile}" _prefixFileLogPath)
-    else()
-        file (RELATIVE_PATH _prefixFileLogPath "${CMAKE_BINARY_DIR}" "${_prefixFile}")
-    endif()
-    get_filename_component(_prefixFileExt "${_prefixFile}" EXT)
-    if (_prefixFileExt MATCHES "^\\.c")
-        set (_comment "Generating ${_language} prefix source ${_prefixFileLogPath}")
-    else()
-        set (_comment "Generating ${_language} prefix header ${_prefixFileLogPath}")
-    endif()
-    # prevent pre-processing errors upon generating the prefix header when a target's generated include file does not yet exist
-    # we do not add a file-level dependency for the target's generated files though, because we only want to depend on their existence
-    # thus we make the prefix header generation depend on a custom helper target which triggers the generation of the files
-    set (_preTargetName "${_target}${COTIRE_PCH_TARGET_SUFFIX}_pre")
-    if (TARGET ${_preTargetName})
-        # custom helper target has already been generated while processing a different language
-        list (APPEND _dependencySources ${_preTargetName})
-    else()
-        get_target_property(_targetSourceFiles ${_target} SOURCES)
-        cotire_get_objects_with_property_on(_generatedSources GENERATED SOURCE ${_targetSourceFiles})
-        if (_generatedSources)
-            add_custom_target("${_preTargetName}" DEPENDS ${_generatedSources})
-            cotire_init_target("${_preTargetName}")
-            list (APPEND _dependencySources ${_preTargetName})
-        endif()
-    endif()
-    add_custom_command(
-            OUTPUT "${_prefixFile}" "${_prefixFile}.log"
-            COMMAND ${_prefixCmd}
-            DEPENDS ${_unityFiles} ${_dependencySources} "${_realCompilerExe}"
-            COMMENT "${_comment}"
-            WORKING_DIRECTORY "${CMAKE_CURRENT_SOURCE_DIR}"
-            VERBATIM)
-    list (APPEND ${_cmdsVar} COMMAND ${_prefixCmd})
-    set (${_cmdsVar} ${${_cmdsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_prefix_generation_from_unity_command _language _target _targetScript _prefixFile _unityFiles _cmdsVar)
-    set (_sourceFiles ${ARGN})
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "GNU|Clang")
-        # GNU and Clang require indirect compilation of the prefix header to make them honor the system_header pragma
-        cotire_prefix_header_to_source_file_path(${_language} "${_prefixFile}" _prefixSourceFile)
-    else()
-        set (_prefixSourceFile "${_prefixFile}")
-    endif()
-    cotire_setup_prefix_generation_command(
-            ${_language} ${_target} "${_targetScript}"
-            "${_prefixSourceFile}" "${_unityFiles}" ${_cmdsVar} ${_sourceFiles})
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "GNU|Clang")
-        # set up generation of a prefix source file which includes the prefix header
-        cotire_setup_combine_command(${_language} "${_targetScript}" "${_prefixFile}" _cmds ${_prefixSourceFile})
-    endif()
-    set (${_cmdsVar} ${${_cmdsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_prefix_generation_from_provided_command _language _target _targetScript _prefixFile _cmdsVar)
-    set (_prefixHeaderFiles ${ARGN})
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "GNU|Clang")
-        # GNU and Clang require indirect compilation of the prefix header to make them honor the system_header pragma
-        cotire_prefix_header_to_source_file_path(${_language} "${_prefixFile}" _prefixSourceFile)
-    else()
-        set (_prefixSourceFile "${_prefixFile}")
-    endif()
-    cotire_setup_combine_command(${_language} "${_targetScript}" "${_prefixSourceFile}" _cmds ${_prefixHeaderFiles})
-    if (CMAKE_${_language}_COMPILER_ID MATCHES "GNU|Clang")
-        # set up generation of a prefix source file which includes the prefix header
-        cotire_setup_combine_command(${_language} "${_targetScript}" "${_prefixFile}" _cmds ${_prefixSourceFile})
-    endif()
-    set (${_cmdsVar} ${${_cmdsVar}} PARENT_SCOPE)
-endfunction()
-
-function (cotire_init_cotire_target_properties _target)
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_ENABLE_PRECOMPILED_HEADER SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_ENABLE_PRECOMPILED_HEADER TRUE)
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_ADD_UNITY_BUILD SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_ADD_UNITY_BUILD TRUE)
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_ADD_CLEAN SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_ADD_CLEAN FALSE)
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_IGNORE_PATH SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_IGNORE_PATH "${CMAKE_SOURCE_DIR}")
-        cotire_check_is_path_relative_to("${CMAKE_BINARY_DIR}" _isRelative "${CMAKE_SOURCE_DIR}")
-        if (NOT _isRelative)
-            set_property(TARGET ${_target} APPEND PROPERTY COTIRE_PREFIX_HEADER_IGNORE_PATH "${CMAKE_BINARY_DIR}")
-        endif()
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_INCLUDE_PATH SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_INCLUDE_PATH "")
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH "")
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_PRE_UNDEFS SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_PRE_UNDEFS "")
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_POST_UNDEFS SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_POST_UNDEFS "")
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_UNITY_LINK_LIBRARIES_INIT SET)
-    if (NOT _isSet)
-        set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_LINK_LIBRARIES_INIT "COPY_UNITY")
-    endif()
-    get_property(_isSet TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES SET)
-    if (NOT _isSet)
-        if (COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES)
-            set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES "${COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES}")
-        else()
-            set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES "")
-        endif()
-    endif()
-endfunction()
-
-function (cotire_make_target_message _target _languages _disableMsg _targetMsgVar)
-    get_target_property(_targetUsePCH ${_target} COTIRE_ENABLE_PRECOMPILED_HEADER)
-    get_target_property(_targetAddSCU ${_target} COTIRE_ADD_UNITY_BUILD)
-    string (REPLACE ";" " " _languagesStr "${_languages}")
-    math (EXPR _numberOfExcludedFiles "${ARGC} - 4")
-    if (_numberOfExcludedFiles EQUAL 0)
-        set (_excludedStr "")
-    elseif (COTIRE_VERBOSE OR _numberOfExcludedFiles LESS 4)
-        string (REPLACE ";" ", " _excludedStr "excluding ${ARGN}")
-    else()
-        set (_excludedStr "excluding ${_numberOfExcludedFiles} files")
-    endif()
-    set (_targetMsg "")
-    if (NOT _languages)
-        set (_targetMsg "Target ${_target} cannot be cotired.")
-        if (_disableMsg)
-            set (_targetMsg "${_targetMsg} ${_disableMsg}")
-        endif()
-    elseif (NOT _targetUsePCH AND NOT _targetAddSCU)
-        set (_targetMsg "${_languagesStr} target ${_target} cotired without unity build and precompiled header.")
-        if (_disableMsg)
-            set (_targetMsg "${_targetMsg} ${_disableMsg}")
-        endif()
-    elseif (NOT _targetUsePCH)
-        if (_excludedStr)
-            set (_targetMsg "${_languagesStr} target ${_target} cotired without precompiled header ${_excludedStr}.")
-        else()
-            set (_targetMsg "${_languagesStr} target ${_target} cotired without precompiled header.")
-        endif()
-        if (_disableMsg)
-            set (_targetMsg "${_targetMsg} ${_disableMsg}")
-        endif()
-    elseif (NOT _targetAddSCU)
-        if (_excludedStr)
-            set (_targetMsg "${_languagesStr} target ${_target} cotired without unity build ${_excludedStr}.")
-        else()
-            set (_targetMsg "${_languagesStr} target ${_target} cotired without unity build.")
-        endif()
-    else()
-        if (_excludedStr)
-            set (_targetMsg "${_languagesStr} target ${_target} cotired ${_excludedStr}.")
-        else()
-            set (_targetMsg "${_languagesStr} target ${_target} cotired.")
-        endif()
-    endif()
-    set (${_targetMsgVar} "${_targetMsg}" PARENT_SCOPE)
-endfunction()
-
-function (cotire_choose_target_languages _target _targetLanguagesVar _wholeTargetVar)
-    set (_languages ${ARGN})
-    set (_allSourceFiles "")
-    set (_allExcludedSourceFiles "")
-    set (_allCotiredSourceFiles "")
-    set (_targetLanguages "")
-    set (_pchEligibleTargetLanguages "")
-    get_target_property(_targetType ${_target} TYPE)
-    get_target_property(_targetSourceFiles ${_target} SOURCES)
-    get_target_property(_targetUsePCH ${_target} COTIRE_ENABLE_PRECOMPILED_HEADER)
-    get_target_property(_targetAddSCU ${_target} COTIRE_ADD_UNITY_BUILD)
-    set (_disableMsg "")
-    foreach (_language ${_languages})
-        get_target_property(_prefixHeader ${_target} COTIRE_${_language}_PREFIX_HEADER)
-        get_target_property(_unityBuildFile ${_target} COTIRE_${_language}_UNITY_SOURCE)
-        if (_prefixHeader OR _unityBuildFile)
-            message (STATUS "cotire: target ${_target} has already been cotired.")
-            set (${_targetLanguagesVar} "" PARENT_SCOPE)
-            return()
-        endif()
-        if (_targetUsePCH AND "${_language}" MATCHES "^C|CXX$" AND DEFINED CMAKE_${_language}_COMPILER_ID)
-            if (CMAKE_${_language}_COMPILER_ID)
-                cotire_check_precompiled_header_support("${_language}" "${_target}" _disableMsg)
-                if (_disableMsg)
-                    set (_targetUsePCH FALSE)
-                endif()
-            endif()
-        endif()
-        set (_sourceFiles "")
-        set (_excludedSources "")
-        set (_cotiredSources "")
-        cotire_filter_language_source_files(${_language} ${_target} _sourceFiles _excludedSources _cotiredSources ${_targetSourceFiles})
-        if (_sourceFiles OR _excludedSources OR _cotiredSources)
-            list (APPEND _targetLanguages ${_language})
-        endif()
-        if (_sourceFiles)
-            list (APPEND _allSourceFiles ${_sourceFiles})
-        endif()
-        list (LENGTH _sourceFiles _numberOfSources)
-        if (NOT _numberOfSources LESS ${COTIRE_MINIMUM_NUMBER_OF_TARGET_SOURCES})
-            list (APPEND _pchEligibleTargetLanguages ${_language})
-        endif()
-        if (_excludedSources)
-            list (APPEND _allExcludedSourceFiles ${_excludedSources})
-        endif()
-        if (_cotiredSources)
-            list (APPEND _allCotiredSourceFiles ${_cotiredSources})
-        endif()
-    endforeach()
-    set (_targetMsgLevel STATUS)
-    if (NOT _targetLanguages)
-        string (REPLACE ";" " or " _languagesStr "${_languages}")
-        set (_disableMsg "No ${_languagesStr} source files.")
-        set (_targetUsePCH FALSE)
-        set (_targetAddSCU FALSE)
-    endif()
-    if (_targetUsePCH)
-        if (_allCotiredSourceFiles)
-            cotire_get_source_file_property_values(_cotireTargets COTIRE_TARGET ${_allCotiredSourceFiles})
-            list (REMOVE_DUPLICATES _cotireTargets)
-            string (REPLACE ";" ", " _cotireTargetsStr "${_cotireTargets}")
-            set (_disableMsg "Target sources already include a precompiled header for target(s) ${_cotireTargets}.")
-            set (_disableMsg "${_disableMsg} Set target property COTIRE_ENABLE_PRECOMPILED_HEADER to FALSE for targets ${_target},")
-            set (_disableMsg "${_disableMsg} ${_cotireTargetsStr} to get a workable build system.")
-            set (_targetMsgLevel SEND_ERROR)
-            set (_targetUsePCH FALSE)
-        elseif (NOT _pchEligibleTargetLanguages)
-            set (_disableMsg "Too few applicable sources.")
-            set (_targetUsePCH FALSE)
-        elseif (XCODE AND _allExcludedSourceFiles)
-            # for Xcode, we cannot apply the precompiled header to individual sources, only to the whole target
-            set (_disableMsg "Exclusion of source files not supported for generator Xcode.")
-            set (_targetUsePCH FALSE)
-        elseif (XCODE AND "${_targetType}" STREQUAL "OBJECT_LIBRARY")
-            # for Xcode, we cannot apply the required PRE_BUILD action to generate the prefix header to an OBJECT_LIBRARY target
-            set (_disableMsg "Required PRE_BUILD action not supported for OBJECT_LIBRARY targets for generator Xcode.")
-            set (_targetUsePCH FALSE)
-        endif()
-    endif()
-    set_property(TARGET ${_target} PROPERTY COTIRE_ENABLE_PRECOMPILED_HEADER ${_targetUsePCH})
-    set_property(TARGET ${_target} PROPERTY COTIRE_ADD_UNITY_BUILD ${_targetAddSCU})
-    cotire_make_target_message(${_target} "${_targetLanguages}" "${_disableMsg}" _targetMsg ${_allExcludedSourceFiles})
-    if (_targetMsg)
-        if (NOT DEFINED COTIREMSG_${_target})
-            set (COTIREMSG_${_target} "")
-        endif()
-        if (COTIRE_VERBOSE OR NOT "${_targetMsgLevel}" STREQUAL "STATUS" OR
-                NOT "${COTIREMSG_${_target}}" STREQUAL "${_targetMsg}")
-            # cache message to avoid redundant messages on re-configure
-            set (COTIREMSG_${_target} "${_targetMsg}" CACHE INTERNAL "${_target} cotire message.")
-            message (${_targetMsgLevel} "${_targetMsg}")
-        endif()
-    endif()
-    list (LENGTH _targetLanguages _numberOfLanguages)
-    if (_numberOfLanguages GREATER 1 OR _allExcludedSourceFiles)
-        set (${_wholeTargetVar} FALSE PARENT_SCOPE)
-    else()
-        set (${_wholeTargetVar} TRUE PARENT_SCOPE)
-    endif()
-    set (${_targetLanguagesVar} ${_targetLanguages} PARENT_SCOPE)
-endfunction()
-
-function (cotire_compute_unity_max_number_of_includes _target _maxIncludesVar)
-    set (_sourceFiles ${ARGN})
-    get_target_property(_maxIncludes ${_target} COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES)
-    if (_maxIncludes MATCHES "(-j|--parallel|--jobs) ?([0-9]*)")
-        set (_numberOfThreads "${CMAKE_MATCH_2}")
-        if (NOT _numberOfThreads)
-            # use all available cores
-            ProcessorCount(_numberOfThreads)
-        endif()
-        list (LENGTH _sourceFiles _numberOfSources)
-        math (EXPR _maxIncludes "(${_numberOfSources} + ${_numberOfThreads} - 1) / ${_numberOfThreads}")
-    elseif (NOT _maxIncludes MATCHES "[0-9]+")
-        set (_maxIncludes 0)
-    endif()
-    if (COTIRE_DEBUG)
-        message (STATUS "${_target} unity source max includes: ${_maxIncludes}")
-    endif()
-    set (${_maxIncludesVar} ${_maxIncludes} PARENT_SCOPE)
-endfunction()
-
-function (cotire_process_target_language _language _configurations _target _wholeTarget _cmdsVar)
-    set (${_cmdsVar} "" PARENT_SCOPE)
-    get_target_property(_targetSourceFiles ${_target} SOURCES)
-    set (_sourceFiles "")
-    set (_excludedSources "")
-    set (_cotiredSources "")
-    cotire_filter_language_source_files(${_language} ${_target} _sourceFiles _excludedSources _cotiredSources ${_targetSourceFiles})
-    if (NOT _sourceFiles AND NOT _cotiredSources)
-        return()
-    endif()
-    set (_cmds "")
-    # check for user provided unity source file list
-    get_property(_unitySourceFiles TARGET ${_target} PROPERTY COTIRE_${_language}_UNITY_SOURCE_INIT)
-    if (NOT _unitySourceFiles)
-        set (_unitySourceFiles ${_sourceFiles} ${_cotiredSources})
-    endif()
-    cotire_generate_target_script(
-            ${_language} "${_configurations}" ${_target} _targetScript _targetConfigScript ${_unitySourceFiles})
-    # set up unity files for parallel compilation
-    cotire_compute_unity_max_number_of_includes(${_target} _maxIncludes ${_unitySourceFiles})
-    cotire_make_unity_source_file_paths(${_language} ${_target} ${_maxIncludes} _unityFiles ${_unitySourceFiles})
-    list (LENGTH _unityFiles _numberOfUnityFiles)
-    if (_numberOfUnityFiles EQUAL 0)
-        return()
-    elseif (_numberOfUnityFiles GREATER 1)
-        cotire_setup_unity_generation_commands(
-                ${_language} ${_target} "${_targetScript}" "${_targetConfigScript}" "${_unityFiles}" _cmds ${_unitySourceFiles})
-    endif()
-    # set up single unity file for prefix header generation
-    cotire_make_single_unity_source_file_path(${_language} ${_target} _unityFile)
-    cotire_setup_unity_generation_commands(
-            ${_language} ${_target} "${_targetScript}" "${_targetConfigScript}" "${_unityFile}" _cmds ${_unitySourceFiles})
-    cotire_make_prefix_file_path(${_language} ${_target} _prefixFile)
-    # set up prefix header
-    if (_prefixFile)
-        # check for user provided prefix header files
-        get_property(_prefixHeaderFiles TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER_INIT)
-        if (_prefixHeaderFiles)
-            cotire_setup_prefix_generation_from_provided_command(
-                    ${_language} ${_target} "${_targetConfigScript}" "${_prefixFile}" _cmds ${_prefixHeaderFiles})
-        else()
-            cotire_setup_prefix_generation_from_unity_command(
-                    ${_language} ${_target} "${_targetConfigScript}" "${_prefixFile}" "${_unityFile}" _cmds ${_unitySourceFiles})
-        endif()
-        # check if selected language has enough sources at all
-        list (LENGTH _sourceFiles _numberOfSources)
-        if (_numberOfSources LESS ${COTIRE_MINIMUM_NUMBER_OF_TARGET_SOURCES})
-            set (_targetUsePCH FALSE)
-        else()
-            get_target_property(_targetUsePCH ${_target} COTIRE_ENABLE_PRECOMPILED_HEADER)
-        endif()
-        if (_targetUsePCH)
-            cotire_make_pch_file_path(${_language} ${_target} _pchFile)
-            if (_pchFile)
-                # first file in _sourceFiles is passed as the host file
-                cotire_setup_pch_file_compilation(
-                        ${_language} ${_target} "${_targetConfigScript}" "${_prefixFile}" "${_pchFile}" ${_sourceFiles})
-                cotire_setup_pch_file_inclusion(
-                        ${_language} ${_target} ${_wholeTarget} "${_prefixFile}" "${_pchFile}" ${_sourceFiles})
-            endif()
-        elseif (_prefixHeaderFiles)
-            # user provided prefix header must be included unconditionally
-            cotire_setup_prefix_file_inclusion(${_language} ${_target} "${_prefixFile}" ${_sourceFiles})
-        endif()
-    endif()
-    # mark target as cotired for language
-    set_property(TARGET ${_target} PROPERTY COTIRE_${_language}_UNITY_SOURCE "${_unityFiles}")
-    if (_prefixFile)
-        set_property(TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER "${_prefixFile}")
-        if (_targetUsePCH AND _pchFile)
-            set_property(TARGET ${_target} PROPERTY COTIRE_${_language}_PRECOMPILED_HEADER "${_pchFile}")
-        endif()
-    endif()
-    set (${_cmdsVar} ${_cmds} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_clean_target _target)
-    set (_cleanTargetName "${_target}${COTIRE_CLEAN_TARGET_SUFFIX}")
-    if (NOT TARGET "${_cleanTargetName}")
-        cotire_set_cmd_to_prologue(_cmds)
-        get_filename_component(_outputDir "${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_CFG_INTDIR}" ABSOLUTE)
-        list (APPEND _cmds -P "${COTIRE_CMAKE_MODULE_FILE}" "cleanup" "${_outputDir}" "${COTIRE_INTDIR}" "${_target}")
-        add_custom_target(${_cleanTargetName}
-                COMMAND ${_cmds}
-                WORKING_DIRECTORY "${CMAKE_BINARY_DIR}"
-                COMMENT "Cleaning up target ${_target} cotire generated files"
-                VERBATIM)
-        cotire_init_target("${_cleanTargetName}")
-    endif()
-endfunction()
-
-function (cotire_setup_pch_target _languages _configurations _target)
-    if ("${CMAKE_GENERATOR}" MATCHES "Make|Ninja")
-        # for makefile based generators, we add a custom target to trigger the generation of the cotire related files
-        set (_dependsFiles "")
-        foreach (_language ${_languages})
-            set (_props COTIRE_${_language}_PREFIX_HEADER COTIRE_${_language}_UNITY_SOURCE)
-            if (NOT CMAKE_${_language}_COMPILER_ID MATCHES "MSVC|Intel")
-                # Visual Studio and Intel only create precompiled header as a side effect
-                list (INSERT _props 0 COTIRE_${_language}_PRECOMPILED_HEADER)
-            endif()
-            cotire_get_first_set_property_value(_dependsFile TARGET ${_target} ${_props})
-            if (_dependsFile)
-                list (APPEND _dependsFiles "${_dependsFile}")
-            endif()
-        endforeach()
-        if (_dependsFiles)
-            set (_pchTargetName "${_target}${COTIRE_PCH_TARGET_SUFFIX}")
-            add_custom_target("${_pchTargetName}" DEPENDS ${_dependsFiles})
-            cotire_init_target("${_pchTargetName}")
-            cotire_add_to_pch_all_target(${_pchTargetName})
-        endif()
-    else()
-        # for other generators, we add the "clean all" target to clean up the precompiled header
-        cotire_setup_clean_all_target()
-    endif()
-endfunction()
-
-function (cotire_filter_object_libraries _target _objectLibrariesVar)
-    set (_objectLibraries "")
-    foreach (_source ${ARGN})
-        if (_source MATCHES "^\\$<TARGET_OBJECTS:.+>$")
-            list (APPEND _objectLibraries "${_source}")
-        endif()
-    endforeach()
-    set (${_objectLibrariesVar} ${_objectLibraries} PARENT_SCOPE)
-endfunction()
-
-function (cotire_collect_unity_target_sources _target _languages _unityTargetSourcesVar)
-    get_target_property(_targetSourceFiles ${_target} SOURCES)
-    set (_unityTargetSources ${_targetSourceFiles})
-    foreach (_language ${_languages})
-        get_property(_unityFiles TARGET ${_target} PROPERTY COTIRE_${_language}_UNITY_SOURCE)
-        if (_unityFiles)
-            # remove source files that are included in the unity source
-            set (_sourceFiles "")
-            set (_excludedSources "")
-            set (_cotiredSources "")
-            cotire_filter_language_source_files(${_language} ${_target} _sourceFiles _excludedSources _cotiredSources ${_targetSourceFiles})
-            if (_sourceFiles OR _cotiredSources)
-                list (REMOVE_ITEM _unityTargetSources ${_sourceFiles} ${_cotiredSources})
-            endif()
-            # add unity source files instead
-            list (APPEND _unityTargetSources ${_unityFiles})
-        endif()
-    endforeach()
-    get_target_property(_linkLibrariesStrategy ${_target} COTIRE_UNITY_LINK_LIBRARIES_INIT)
-    if ("${_linkLibrariesStrategy}" MATCHES "^COPY_UNITY$")
-        cotire_filter_object_libraries(${_target} _objectLibraries ${_targetSourceFiles})
-        if (_objectLibraries)
-            cotire_map_libraries("${_linkLibrariesStrategy}" _unityObjectLibraries ${_objectLibraries})
-            list (REMOVE_ITEM _unityTargetSources ${_objectLibraries})
-            list (APPEND _unityTargetSources ${_unityObjectLibraries})
-        endif()
-    endif()
-    set (${_unityTargetSourcesVar} ${_unityTargetSources} PARENT_SCOPE)
-endfunction()
-
-function (cotire_setup_unity_target_pch_usage _languages _target)
-    foreach (_language ${_languages})
-        get_property(_unityFiles TARGET ${_target} PROPERTY COTIRE_${_language}_UNITY_SOURCE)
-        if (_unityFiles)
-            get_property(_userPrefixFile TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER_INIT)
-            get_property(_prefixFile TARGET ${_target} PROPERTY COTIRE_${_language}_PREFIX_HEADER)
-            if (_userPrefixFile AND _prefixFile)
-                # user provided prefix header must be included unconditionally by unity sources
-                cotire_setup_prefix_file_inclusion(${_language} ${_target} "${_prefixFile}" ${_unityFiles})
-            endif()
-        endif()
-    endforeach()
-endfunction()
-
-function (cotire_setup_unity_build_target _languages _configurations _target)
-    get_target_property(_unityTargetName ${_target} COTIRE_UNITY_TARGET_NAME)
-    if (NOT _unityTargetName)
-        set (_unityTargetName "${_target}${COTIRE_UNITY_BUILD_TARGET_SUFFIX}")
-    endif()
-    # determine unity target sub type
-    get_target_property(_targetType ${_target} TYPE)
-    if ("${_targetType}" STREQUAL "EXECUTABLE")
-        set (_unityTargetSubType "")
-    elseif (_targetType MATCHES "(STATIC|SHARED|MODULE|OBJECT)_LIBRARY")
-        set (_unityTargetSubType "${CMAKE_MATCH_1}")
-    else()
-        message (WARNING "cotire: target ${_target} has unknown target type ${_targetType}.")
-        return()
-    endif()
-    # determine unity target sources
-    set (_unityTargetSources "")
-    cotire_collect_unity_target_sources(${_target} "${_languages}" _unityTargetSources)
-    # handle automatic Qt processing
-    get_target_property(_targetAutoMoc ${_target} AUTOMOC)
-    get_target_property(_targetAutoUic ${_target} AUTOUIC)
-    get_target_property(_targetAutoRcc ${_target} AUTORCC)
-    if (_targetAutoMoc OR _targetAutoUic OR _targetAutoRcc)
-        # if the original target sources are subject to CMake's automatic Qt processing,
-        # also include implicitly generated <targetname>_automoc.cpp file
-        if (CMAKE_VERSION VERSION_LESS "3.8.0")
-            list (APPEND _unityTargetSources "${_target}_automoc.cpp")
-            set_property (SOURCE "${_target}_automoc.cpp" PROPERTY GENERATED TRUE)
-        else()
-            list (APPEND _unityTargetSources "${_target}_autogen/moc_compilation.cpp")
-            set_property (SOURCE "${_target}_autogen/moc_compilation.cpp" PROPERTY GENERATED TRUE)
-        endif()
-    endif()
-    # prevent AUTOMOC, AUTOUIC and AUTORCC properties from being set when the unity target is created
-    set (CMAKE_AUTOMOC OFF)
-    set (CMAKE_AUTOUIC OFF)
-    set (CMAKE_AUTORCC OFF)
-    if (COTIRE_DEBUG)
-        message (STATUS "add target ${_targetType} ${_unityTargetName} ${_unityTargetSubType} EXCLUDE_FROM_ALL ${_unityTargetSources}")
-    endif()
-    # generate unity target
-    if ("${_targetType}" STREQUAL "EXECUTABLE")
-        add_executable(${_unityTargetName} ${_unityTargetSubType} EXCLUDE_FROM_ALL ${_unityTargetSources})
-    else()
-        add_library(${_unityTargetName} ${_unityTargetSubType} EXCLUDE_FROM_ALL ${_unityTargetSources})
-    endif()
-    if ("${CMAKE_GENERATOR}" MATCHES "Visual Studio")
-        # depend on original target's automoc target, if it exists
-        if (TARGET ${_target}_automoc)
-            add_dependencies(${_unityTargetName} ${_target}_automoc)
-        endif()
-    else()
-        if (_targetAutoMoc OR _targetAutoUic OR _targetAutoRcc)
-            # depend on the original target's implicity generated <targetname>_automoc target
-            if (CMAKE_VERSION VERSION_LESS "3.8.0")
-                add_dependencies(${_unityTargetName} ${_target}_automoc)
-            else()
-                add_dependencies(${_unityTargetName} ${_target}_autogen)
-            endif()
-        endif()
-    endif()
-    # copy output location properties
-    set (_outputDirProperties
-            ARCHIVE_OUTPUT_DIRECTORY ARCHIVE_OUTPUT_DIRECTORY_<CONFIG>
-            LIBRARY_OUTPUT_DIRECTORY LIBRARY_OUTPUT_DIRECTORY_<CONFIG>
-            RUNTIME_OUTPUT_DIRECTORY RUNTIME_OUTPUT_DIRECTORY_<CONFIG>)
-    if (COTIRE_UNITY_OUTPUT_DIRECTORY)
-        set (_setDefaultOutputDir TRUE)
-        if (IS_ABSOLUTE "${COTIRE_UNITY_OUTPUT_DIRECTORY}")
-            set (_outputDir "${COTIRE_UNITY_OUTPUT_DIRECTORY}")
-        else()
-            # append relative COTIRE_UNITY_OUTPUT_DIRECTORY to target's actual output directory
-            cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName} ${_outputDirProperties})
-            cotire_resolve_config_properties("${_configurations}" _properties ${_outputDirProperties})
-            foreach (_property ${_properties})
-                get_property(_outputDir TARGET ${_target} PROPERTY ${_property})
-                if (_outputDir)
-                    get_filename_component(_outputDir "${_outputDir}/${COTIRE_UNITY_OUTPUT_DIRECTORY}" ABSOLUTE)
-                    set_property(TARGET ${_unityTargetName} PROPERTY ${_property} "${_outputDir}")
-                    set (_setDefaultOutputDir FALSE)
-                endif()
-            endforeach()
-            if (_setDefaultOutputDir)
-                get_filename_component(_outputDir "${CMAKE_CURRENT_BINARY_DIR}/${COTIRE_UNITY_OUTPUT_DIRECTORY}" ABSOLUTE)
-            endif()
-        endif()
-        if (_setDefaultOutputDir)
-            set_target_properties(${_unityTargetName} PROPERTIES
-                    ARCHIVE_OUTPUT_DIRECTORY "${_outputDir}"
-                    LIBRARY_OUTPUT_DIRECTORY "${_outputDir}"
-                    RUNTIME_OUTPUT_DIRECTORY "${_outputDir}")
-        endif()
-    else()
-        cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-                ${_outputDirProperties})
-    endif()
-    # copy output name
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            ARCHIVE_OUTPUT_NAME ARCHIVE_OUTPUT_NAME_<CONFIG>
-            LIBRARY_OUTPUT_NAME LIBRARY_OUTPUT_NAME_<CONFIG>
-            OUTPUT_NAME OUTPUT_NAME_<CONFIG>
-            RUNTIME_OUTPUT_NAME RUNTIME_OUTPUT_NAME_<CONFIG>
-            PREFIX <CONFIG>_POSTFIX SUFFIX
-            IMPORT_PREFIX IMPORT_SUFFIX)
-    # copy compile stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            COMPILE_DEFINITIONS COMPILE_DEFINITIONS_<CONFIG>
-            COMPILE_FLAGS COMPILE_OPTIONS
-            Fortran_FORMAT Fortran_MODULE_DIRECTORY
-            INCLUDE_DIRECTORIES
-            INTERPROCEDURAL_OPTIMIZATION INTERPROCEDURAL_OPTIMIZATION_<CONFIG>
-            POSITION_INDEPENDENT_CODE
-            C_COMPILER_LAUNCHER CXX_COMPILER_LAUNCHER
-            C_INCLUDE_WHAT_YOU_USE CXX_INCLUDE_WHAT_YOU_USE
-            C_VISIBILITY_PRESET CXX_VISIBILITY_PRESET VISIBILITY_INLINES_HIDDEN
-            C_CLANG_TIDY CXX_CLANG_TIDY)
-    # copy compile features
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            C_EXTENSIONS C_STANDARD C_STANDARD_REQUIRED
-            CXX_EXTENSIONS CXX_STANDARD CXX_STANDARD_REQUIRED
-            COMPILE_FEATURES)
-    # copy interface stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            COMPATIBLE_INTERFACE_BOOL COMPATIBLE_INTERFACE_NUMBER_MAX COMPATIBLE_INTERFACE_NUMBER_MIN
-            COMPATIBLE_INTERFACE_STRING
-            INTERFACE_COMPILE_DEFINITIONS INTERFACE_COMPILE_FEATURES INTERFACE_COMPILE_OPTIONS
-            INTERFACE_INCLUDE_DIRECTORIES INTERFACE_SOURCES
-            INTERFACE_POSITION_INDEPENDENT_CODE INTERFACE_SYSTEM_INCLUDE_DIRECTORIES
-            INTERFACE_AUTOUIC_OPTIONS NO_SYSTEM_FROM_IMPORTED)
-    # copy link stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            BUILD_WITH_INSTALL_RPATH INSTALL_RPATH INSTALL_RPATH_USE_LINK_PATH SKIP_BUILD_RPATH
-            LINKER_LANGUAGE LINK_DEPENDS LINK_DEPENDS_NO_SHARED
-            LINK_FLAGS LINK_FLAGS_<CONFIG>
-            LINK_INTERFACE_LIBRARIES LINK_INTERFACE_LIBRARIES_<CONFIG>
-            LINK_INTERFACE_MULTIPLICITY LINK_INTERFACE_MULTIPLICITY_<CONFIG>
-            LINK_SEARCH_START_STATIC LINK_SEARCH_END_STATIC
-            STATIC_LIBRARY_FLAGS STATIC_LIBRARY_FLAGS_<CONFIG>
-            NO_SONAME SOVERSION VERSION
-            LINK_WHAT_YOU_USE BUILD_RPATH)
-    # copy cmake stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            IMPLICIT_DEPENDS_INCLUDE_TRANSFORM RULE_LAUNCH_COMPILE RULE_LAUNCH_CUSTOM RULE_LAUNCH_LINK)
-    # copy Apple platform specific stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            BUNDLE BUNDLE_EXTENSION FRAMEWORK FRAMEWORK_VERSION INSTALL_NAME_DIR
-            MACOSX_BUNDLE MACOSX_BUNDLE_INFO_PLIST MACOSX_FRAMEWORK_INFO_PLIST MACOSX_RPATH
-            OSX_ARCHITECTURES OSX_ARCHITECTURES_<CONFIG> PRIVATE_HEADER PUBLIC_HEADER RESOURCE XCTEST
-            IOS_INSTALL_COMBINED XCODE_EXPLICIT_FILE_TYPE XCODE_PRODUCT_TYPE)
-    # copy Windows platform specific stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            GNUtoMS
-            COMPILE_PDB_NAME COMPILE_PDB_NAME_<CONFIG>
-            COMPILE_PDB_OUTPUT_DIRECTORY COMPILE_PDB_OUTPUT_DIRECTORY_<CONFIG>
-            PDB_NAME PDB_NAME_<CONFIG> PDB_OUTPUT_DIRECTORY PDB_OUTPUT_DIRECTORY_<CONFIG>
-            VS_DESKTOP_EXTENSIONS_VERSION VS_DOTNET_REFERENCES VS_DOTNET_TARGET_FRAMEWORK_VERSION
-            VS_GLOBAL_KEYWORD VS_GLOBAL_PROJECT_TYPES VS_GLOBAL_ROOTNAMESPACE
-            VS_IOT_EXTENSIONS_VERSION VS_IOT_STARTUP_TASK
-            VS_KEYWORD VS_MOBILE_EXTENSIONS_VERSION
-            VS_SCC_AUXPATH VS_SCC_LOCALPATH VS_SCC_PROJECTNAME VS_SCC_PROVIDER
-            VS_WINDOWS_TARGET_PLATFORM_MIN_VERSION
-            VS_WINRT_COMPONENT VS_WINRT_EXTENSIONS VS_WINRT_REFERENCES
-            WIN32_EXECUTABLE WINDOWS_EXPORT_ALL_SYMBOLS
-            DEPLOYMENT_REMOTE_DIRECTORY VS_CONFIGURATION_TYPE
-            VS_SDK_REFERENCES VS_USER_PROPS VS_DEBUGGER_WORKING_DIRECTORY)
-    # copy Android platform specific stuff
-    cotire_copy_set_properties("${_configurations}" TARGET ${_target} ${_unityTargetName}
-            ANDROID_API ANDROID_API_MIN ANDROID_GUI
-            ANDROID_ANT_ADDITIONAL_OPTIONS ANDROID_ARCH ANDROID_ASSETS_DIRECTORIES
-            ANDROID_JAR_DEPENDENCIES ANDROID_JAR_DIRECTORIES ANDROID_JAVA_SOURCE_DIR
-            ANDROID_NATIVE_LIB_DEPENDENCIES ANDROID_NATIVE_LIB_DIRECTORIES
-            ANDROID_PROCESS_MAX ANDROID_PROGUARD ANDROID_PROGUARD_CONFIG_PATH
-            ANDROID_SECURE_PROPS_PATH ANDROID_SKIP_ANT_STEP ANDROID_STL_TYPE)
-    # use output name from original target
-    get_target_property(_targetOutputName ${_unityTargetName} OUTPUT_NAME)
-    if (NOT _targetOutputName)
-        set_property(TARGET ${_unityTargetName} PROPERTY OUTPUT_NAME "${_target}")
-    endif()
-    # use export symbol from original target
-    cotire_get_target_export_symbol("${_target}" _defineSymbol)
-    if (_defineSymbol)
-        set_property(TARGET ${_unityTargetName} PROPERTY DEFINE_SYMBOL "${_defineSymbol}")
-        if ("${_targetType}" STREQUAL "EXECUTABLE")
-            set_property(TARGET ${_unityTargetName} PROPERTY ENABLE_EXPORTS TRUE)
-        endif()
-    endif()
-    cotire_init_target(${_unityTargetName})
-    cotire_add_to_unity_all_target(${_unityTargetName})
-    set_property(TARGET ${_target} PROPERTY COTIRE_UNITY_TARGET_NAME "${_unityTargetName}")
-endfunction(cotire_setup_unity_build_target)
-
-function (cotire_target _target)
-    set(_options "")
-    set(_oneValueArgs "")
-    set(_multiValueArgs LANGUAGES CONFIGURATIONS)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    if (NOT _option_LANGUAGES)
-        get_property (_option_LANGUAGES GLOBAL PROPERTY ENABLED_LANGUAGES)
-    endif()
-    if (NOT _option_CONFIGURATIONS)
-        cotire_get_configuration_types(_option_CONFIGURATIONS)
-    endif()
-    # check if cotire can be applied to target at all
-    cotire_is_target_supported(${_target} _isSupported)
-    if (NOT _isSupported)
-        get_target_property(_imported ${_target} IMPORTED)
-        get_target_property(_targetType ${_target} TYPE)
-        if (_imported)
-            message (WARNING "cotire: imported ${_targetType} target ${_target} cannot be cotired.")
-        else()
-            message (STATUS "cotire: ${_targetType} target ${_target} cannot be cotired.")
-        endif()
-        return()
-    endif()
-    # resolve alias
-    get_target_property(_aliasName ${_target} ALIASED_TARGET)
-    if (_aliasName)
-        if (COTIRE_DEBUG)
-            message (STATUS "${_target} is an alias. Applying cotire to aliased target ${_aliasName} instead.")
-        endif()
-        set (_target ${_aliasName})
-    endif()
-    # check if target needs to be cotired for build type
-    # when using configuration types, the test is performed at build time
-    cotire_init_cotire_target_properties(${_target})
-    if (NOT CMAKE_CONFIGURATION_TYPES)
-        if (CMAKE_BUILD_TYPE)
-            list (FIND _option_CONFIGURATIONS "${CMAKE_BUILD_TYPE}" _index)
-        else()
-            list (FIND _option_CONFIGURATIONS "None" _index)
-        endif()
-        if (_index EQUAL -1)
-            if (COTIRE_DEBUG)
-                message (STATUS "CMAKE_BUILD_TYPE=${CMAKE_BUILD_TYPE} not cotired (${_option_CONFIGURATIONS})")
-            endif()
-            return()
-        endif()
-    endif()
-    # when not using configuration types, immediately create cotire intermediate dir
-    if (NOT CMAKE_CONFIGURATION_TYPES)
-        cotire_get_intermediate_dir(_baseDir)
-        file (MAKE_DIRECTORY "${_baseDir}")
-    endif()
-    # choose languages that apply to the target
-    cotire_choose_target_languages("${_target}" _targetLanguages _wholeTarget ${_option_LANGUAGES})
-    if (NOT _targetLanguages)
-        return()
-    endif()
-    set (_cmds "")
-    foreach (_language ${_targetLanguages})
-        cotire_process_target_language("${_language}" "${_option_CONFIGURATIONS}" ${_target} ${_wholeTarget} _cmd)
-        if (_cmd)
-            list (APPEND _cmds ${_cmd})
-        endif()
-    endforeach()
-    get_target_property(_targetAddSCU ${_target} COTIRE_ADD_UNITY_BUILD)
-    if (_targetAddSCU)
-        cotire_setup_unity_build_target("${_targetLanguages}" "${_option_CONFIGURATIONS}" ${_target})
-    endif()
-    get_target_property(_targetUsePCH ${_target} COTIRE_ENABLE_PRECOMPILED_HEADER)
-    if (_targetUsePCH)
-        cotire_setup_target_pch_usage("${_targetLanguages}" ${_target} ${_wholeTarget} ${_cmds})
-        cotire_setup_pch_target("${_targetLanguages}" "${_option_CONFIGURATIONS}" ${_target})
-        if (_targetAddSCU)
-            cotire_setup_unity_target_pch_usage("${_targetLanguages}" ${_target})
-        endif()
-    endif()
-    get_target_property(_targetAddCleanTarget ${_target} COTIRE_ADD_CLEAN)
-    if (_targetAddCleanTarget)
-        cotire_setup_clean_target(${_target})
-    endif()
-endfunction(cotire_target)
-
-function (cotire_map_libraries _strategy _mappedLibrariesVar)
-    set (_mappedLibraries "")
-    foreach (_library ${ARGN})
-        if (_library MATCHES "^\\$<LINK_ONLY:(.+)>$")
-            set (_libraryName "${CMAKE_MATCH_1}")
-            set (_linkOnly TRUE)
-            set (_objectLibrary FALSE)
-        elseif (_library MATCHES "^\\$<TARGET_OBJECTS:(.+)>$")
-            set (_libraryName "${CMAKE_MATCH_1}")
-            set (_linkOnly FALSE)
-            set (_objectLibrary TRUE)
-        else()
-            set (_libraryName "${_library}")
-            set (_linkOnly FALSE)
-            set (_objectLibrary FALSE)
-        endif()
-        if ("${_strategy}" MATCHES "COPY_UNITY")
-            cotire_is_target_supported(${_libraryName} _isSupported)
-            if (_isSupported)
-                # use target's corresponding unity target, if available
-                get_target_property(_libraryUnityTargetName ${_libraryName} COTIRE_UNITY_TARGET_NAME)
-                if (TARGET "${_libraryUnityTargetName}")
-                    if (_linkOnly)
-                        list (APPEND _mappedLibraries "$<LINK_ONLY:${_libraryUnityTargetName}>")
-                    elseif (_objectLibrary)
-                        list (APPEND _mappedLibraries "$<TARGET_OBJECTS:${_libraryUnityTargetName}>")
-                    else()
-                        list (APPEND _mappedLibraries "${_libraryUnityTargetName}")
-                    endif()
-                else()
-                    list (APPEND _mappedLibraries "${_library}")
-                endif()
-            else()
-                list (APPEND _mappedLibraries "${_library}")
-            endif()
-        else()
-            list (APPEND _mappedLibraries "${_library}")
-        endif()
-    endforeach()
-    list (REMOVE_DUPLICATES _mappedLibraries)
-    set (${_mappedLibrariesVar} ${_mappedLibraries} PARENT_SCOPE)
-endfunction()
-
-function (cotire_target_link_libraries _target)
-    cotire_is_target_supported(${_target} _isSupported)
-    if (NOT _isSupported)
-        return()
-    endif()
-    get_target_property(_unityTargetName ${_target} COTIRE_UNITY_TARGET_NAME)
-    if (TARGET "${_unityTargetName}")
-        get_target_property(_linkLibrariesStrategy ${_target} COTIRE_UNITY_LINK_LIBRARIES_INIT)
-        if (COTIRE_DEBUG)
-            message (STATUS "unity target ${_unityTargetName} link strategy: ${_linkLibrariesStrategy}")
-        endif()
-        if ("${_linkLibrariesStrategy}" MATCHES "^(COPY|COPY_UNITY)$")
-            get_target_property(_linkLibraries ${_target} LINK_LIBRARIES)
-            if (_linkLibraries)
-                cotire_map_libraries("${_linkLibrariesStrategy}" _unityLinkLibraries ${_linkLibraries})
-                set_target_properties(${_unityTargetName} PROPERTIES LINK_LIBRARIES "${_unityLinkLibraries}")
-                if (COTIRE_DEBUG)
-                    message (STATUS "unity target ${_unityTargetName} link libraries: ${_unityLinkLibraries}")
-                endif()
-            endif()
-            get_target_property(_interfaceLinkLibraries ${_target} INTERFACE_LINK_LIBRARIES)
-            if (_interfaceLinkLibraries)
-                cotire_map_libraries("${_linkLibrariesStrategy}" _unityLinkInterfaceLibraries ${_interfaceLinkLibraries})
-                set_target_properties(${_unityTargetName} PROPERTIES INTERFACE_LINK_LIBRARIES "${_unityLinkInterfaceLibraries}")
-                if (COTIRE_DEBUG)
-                    message (STATUS "unity target ${_unityTargetName} interface link libraries: ${_unityLinkInterfaceLibraries}")
-                endif()
-            endif()
-        endif()
-    endif()
-endfunction(cotire_target_link_libraries)
-
-function (cotire_cleanup _binaryDir _cotireIntermediateDirName _targetName)
-    if (_targetName)
-        file (GLOB_RECURSE _cotireFiles "${_binaryDir}/${_targetName}*.*")
-    else()
-        file (GLOB_RECURSE _cotireFiles "${_binaryDir}/*.*")
-    endif()
-    # filter files in intermediate directory
-    set (_filesToRemove "")
-    foreach (_file ${_cotireFiles})
-        get_filename_component(_dir "${_file}" DIRECTORY)
-        get_filename_component(_dirName "${_dir}" NAME)
-        if ("${_dirName}" STREQUAL "${_cotireIntermediateDirName}")
-            list (APPEND _filesToRemove "${_file}")
-        endif()
-    endforeach()
-    if (_filesToRemove)
-        if (COTIRE_VERBOSE)
-            message (STATUS "cleaning up ${_filesToRemove}")
-        endif()
-        file (REMOVE ${_filesToRemove})
-    endif()
-endfunction()
-
-function (cotire_init_target _targetName)
-    if (COTIRE_TARGETS_FOLDER)
-        set_target_properties(${_targetName} PROPERTIES FOLDER "${COTIRE_TARGETS_FOLDER}")
-    endif()
-    set_target_properties(${_targetName} PROPERTIES EXCLUDE_FROM_ALL TRUE)
-    if (MSVC_IDE)
-        set_target_properties(${_targetName} PROPERTIES EXCLUDE_FROM_DEFAULT_BUILD TRUE)
-    endif()
-endfunction()
-
-function (cotire_add_to_pch_all_target _pchTargetName)
-    set (_targetName "${COTIRE_PCH_ALL_TARGET_NAME}")
-    if (NOT TARGET "${_targetName}")
-        add_custom_target("${_targetName}"
-                WORKING_DIRECTORY "${CMAKE_BINARY_DIR}"
-                VERBATIM)
-        cotire_init_target("${_targetName}")
-    endif()
-    cotire_setup_clean_all_target()
-    add_dependencies(${_targetName} ${_pchTargetName})
-endfunction()
-
-function (cotire_add_to_unity_all_target _unityTargetName)
-    set (_targetName "${COTIRE_UNITY_BUILD_ALL_TARGET_NAME}")
-    if (NOT TARGET "${_targetName}")
-        add_custom_target("${_targetName}"
-                WORKING_DIRECTORY "${CMAKE_BINARY_DIR}"
-                VERBATIM)
-        cotire_init_target("${_targetName}")
-    endif()
-    cotire_setup_clean_all_target()
-    add_dependencies(${_targetName} ${_unityTargetName})
-endfunction()
-
-function (cotire_setup_clean_all_target)
-    set (_targetName "${COTIRE_CLEAN_ALL_TARGET_NAME}")
-    if (NOT TARGET "${_targetName}")
-        cotire_set_cmd_to_prologue(_cmds)
-        list (APPEND _cmds -P "${COTIRE_CMAKE_MODULE_FILE}" "cleanup" "${CMAKE_BINARY_DIR}" "${COTIRE_INTDIR}")
-        add_custom_target(${_targetName}
-                COMMAND ${_cmds}
-                WORKING_DIRECTORY "${CMAKE_BINARY_DIR}"
-                COMMENT "Cleaning up all cotire generated files"
-                VERBATIM)
-        cotire_init_target("${_targetName}")
-    endif()
-endfunction()
-
-function (cotire)
-    set(_options "")
-    set(_oneValueArgs "")
-    set(_multiValueArgs LANGUAGES CONFIGURATIONS)
-    cmake_parse_arguments(_option "${_options}" "${_oneValueArgs}" "${_multiValueArgs}" ${ARGN})
-    set (_targets ${_option_UNPARSED_ARGUMENTS})
-    foreach (_target ${_targets})
-        if (TARGET ${_target})
-            cotire_target(${_target} LANGUAGES ${_option_LANGUAGES} CONFIGURATIONS ${_option_CONFIGURATIONS})
-        else()
-            message (WARNING "cotire: ${_target} is not a target.")
-        endif()
-    endforeach()
-    foreach (_target ${_targets})
-        if (TARGET ${_target})
-            cotire_target_link_libraries(${_target})
-        endif()
-    endforeach()
-endfunction()
-
-if (CMAKE_SCRIPT_MODE_FILE)
-
-    # cotire is being run in script mode
-    # locate -P on command args
-    set (COTIRE_ARGC -1)
-    foreach (_index RANGE ${CMAKE_ARGC})
-        if (COTIRE_ARGC GREATER -1)
-            set (COTIRE_ARGV${COTIRE_ARGC} "${CMAKE_ARGV${_index}}")
-            math (EXPR COTIRE_ARGC "${COTIRE_ARGC} + 1")
-        elseif ("${CMAKE_ARGV${_index}}" STREQUAL "-P")
-            set (COTIRE_ARGC 0)
-        endif()
-    endforeach()
-
-    # include target script if available
-    if ("${COTIRE_ARGV2}" MATCHES "\\.cmake$")
-        # the included target scripts sets up additional variables relating to the target (e.g., COTIRE_TARGET_SOURCES)
-        include("${COTIRE_ARGV2}")
-    endif()
-
-    if (COTIRE_DEBUG)
-        message (STATUS "${COTIRE_ARGV0} ${COTIRE_ARGV1} ${COTIRE_ARGV2} ${COTIRE_ARGV3} ${COTIRE_ARGV4} ${COTIRE_ARGV5}")
-    endif()
-
-    if (NOT COTIRE_BUILD_TYPE)
-        set (COTIRE_BUILD_TYPE "None")
-    endif()
-    string (TOUPPER "${COTIRE_BUILD_TYPE}" _upperConfig)
-    set (_includeDirs ${COTIRE_TARGET_INCLUDE_DIRECTORIES_${_upperConfig}})
-    set (_systemIncludeDirs ${COTIRE_TARGET_SYSTEM_INCLUDE_DIRECTORIES_${_upperConfig}})
-    set (_compileDefinitions ${COTIRE_TARGET_COMPILE_DEFINITIONS_${_upperConfig}})
-    set (_compileFlags ${COTIRE_TARGET_COMPILE_FLAGS_${_upperConfig}})
-    # check if target has been cotired for actual build type COTIRE_BUILD_TYPE
-    list (FIND COTIRE_TARGET_CONFIGURATION_TYPES "${COTIRE_BUILD_TYPE}" _index)
-    if (_index GREATER -1)
-        set (_sources ${COTIRE_TARGET_SOURCES})
-        set (_sourcesDefinitions ${COTIRE_TARGET_SOURCES_COMPILE_DEFINITIONS_${_upperConfig}})
-    else()
-        if (COTIRE_DEBUG)
-            message (STATUS "COTIRE_BUILD_TYPE=${COTIRE_BUILD_TYPE} not cotired (${COTIRE_TARGET_CONFIGURATION_TYPES})")
-        endif()
-        set (_sources "")
-        set (_sourcesDefinitions "")
-    endif()
-    set (_targetPreUndefs ${COTIRE_TARGET_PRE_UNDEFS})
-    set (_targetPostUndefs ${COTIRE_TARGET_POST_UNDEFS})
-    set (_sourcesPreUndefs ${COTIRE_TARGET_SOURCES_PRE_UNDEFS})
-    set (_sourcesPostUndefs ${COTIRE_TARGET_SOURCES_POST_UNDEFS})
-
-    if ("${COTIRE_ARGV1}" STREQUAL "unity")
-
-        if (XCODE)
-            # executing pre-build action under Xcode, check dependency on target script
-            set (_dependsOption DEPENDS "${COTIRE_ARGV2}")
-        else()
-            # executing custom command, no need to re-check for dependencies
-            set (_dependsOption "")
-        endif()
-
-        cotire_select_unity_source_files("${COTIRE_ARGV3}" _sources ${_sources})
-
-        cotire_generate_unity_source(
-                "${COTIRE_ARGV3}" ${_sources}
-                LANGUAGE "${COTIRE_TARGET_LANGUAGE}"
-                SOURCES_COMPILE_DEFINITIONS ${_sourcesDefinitions}
-                PRE_UNDEFS ${_targetPreUndefs}
-                POST_UNDEFS ${_targetPostUndefs}
-                SOURCES_PRE_UNDEFS ${_sourcesPreUndefs}
-                SOURCES_POST_UNDEFS ${_sourcesPostUndefs}
-                ${_dependsOption})
-
-    elseif ("${COTIRE_ARGV1}" STREQUAL "prefix")
-
-        if (XCODE)
-            # executing pre-build action under Xcode, check dependency on unity file and prefix dependencies
-            set (_dependsOption DEPENDS "${COTIRE_ARGV4}" ${COTIRE_TARGET_PREFIX_DEPENDS})
-        else()
-            # executing custom command, no need to re-check for dependencies
-            set (_dependsOption "")
-        endif()
-
-        set (_files "")
-        foreach (_index RANGE 4 ${COTIRE_ARGC})
-            if (COTIRE_ARGV${_index})
-                list (APPEND _files "${COTIRE_ARGV${_index}}")
-            endif()
-        endforeach()
-
-        cotire_generate_prefix_header(
-                "${COTIRE_ARGV3}" ${_files}
-                COMPILER_LAUNCHER "${COTIRE_TARGET_${COTIRE_TARGET_LANGUAGE}_COMPILER_LAUNCHER}"
-                COMPILER_EXECUTABLE "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER}"
-                COMPILER_ARG1 ${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_ARG1}
-                COMPILER_ID "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_ID}"
-                COMPILER_VERSION "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_VERSION}"
-                LANGUAGE "${COTIRE_TARGET_LANGUAGE}"
-                IGNORE_PATH "${COTIRE_TARGET_IGNORE_PATH};${COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_PATH}"
-                INCLUDE_PATH ${COTIRE_TARGET_INCLUDE_PATH}
-                IGNORE_EXTENSIONS "${CMAKE_${COTIRE_TARGET_LANGUAGE}_SOURCE_FILE_EXTENSIONS};${COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_EXTENSIONS}"
-                INCLUDE_PRIORITY_PATH ${COTIRE_TARGET_INCLUDE_PRIORITY_PATH}
-                INCLUDE_DIRECTORIES ${_includeDirs}
-                SYSTEM_INCLUDE_DIRECTORIES ${_systemIncludeDirs}
-                COMPILE_DEFINITIONS ${_compileDefinitions}
-                COMPILE_FLAGS ${_compileFlags}
-                ${_dependsOption})
-
-    elseif ("${COTIRE_ARGV1}" STREQUAL "precompile")
-
-        set (_files "")
-        foreach (_index RANGE 5 ${COTIRE_ARGC})
-            if (COTIRE_ARGV${_index})
-                list (APPEND _files "${COTIRE_ARGV${_index}}")
-            endif()
-        endforeach()
-
-        cotire_precompile_prefix_header(
-                "${COTIRE_ARGV3}" "${COTIRE_ARGV4}" "${COTIRE_ARGV5}"
-                COMPILER_LAUNCHER "${COTIRE_TARGET_${COTIRE_TARGET_LANGUAGE}_COMPILER_LAUNCHER}"
-                COMPILER_EXECUTABLE "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER}"
-                COMPILER_ARG1 ${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_ARG1}
-                COMPILER_ID "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_ID}"
-                COMPILER_VERSION "${CMAKE_${COTIRE_TARGET_LANGUAGE}_COMPILER_VERSION}"
-                LANGUAGE "${COTIRE_TARGET_LANGUAGE}"
-                INCLUDE_DIRECTORIES ${_includeDirs}
-                SYSTEM_INCLUDE_DIRECTORIES ${_systemIncludeDirs}
-                COMPILE_DEFINITIONS ${_compileDefinitions}
-                COMPILE_FLAGS ${_compileFlags})
-
-    elseif ("${COTIRE_ARGV1}" STREQUAL "combine")
-
-        if (COTIRE_TARGET_LANGUAGE)
-            set (_combinedFile "${COTIRE_ARGV3}")
-            set (_startIndex 4)
-        else()
-            set (_combinedFile "${COTIRE_ARGV2}")
-            set (_startIndex 3)
-        endif()
-        set (_files "")
-        foreach (_index RANGE ${_startIndex} ${COTIRE_ARGC})
-            if (COTIRE_ARGV${_index})
-                list (APPEND _files "${COTIRE_ARGV${_index}}")
-            endif()
-        endforeach()
-
-        if (XCODE)
-            # executing pre-build action under Xcode, check dependency on files to be combined
-            set (_dependsOption DEPENDS ${_files})
-        else()
-            # executing custom command, no need to re-check for dependencies
-            set (_dependsOption "")
-        endif()
-
-        if (COTIRE_TARGET_LANGUAGE)
-            cotire_generate_unity_source(
-                    "${_combinedFile}" ${_files}
-                    LANGUAGE "${COTIRE_TARGET_LANGUAGE}"
-                    ${_dependsOption})
-        else()
-            cotire_generate_unity_source("${_combinedFile}" ${_files} ${_dependsOption})
-        endif()
-
-    elseif ("${COTIRE_ARGV1}" STREQUAL "cleanup")
-
-        cotire_cleanup("${COTIRE_ARGV2}" "${COTIRE_ARGV3}" "${COTIRE_ARGV4}")
-
-    else()
-        message (FATAL_ERROR "cotire: unknown command \"${COTIRE_ARGV1}\".")
-    endif()
-
-else()
-
-    # cotire is being run in include mode
-    # set up all variable and property definitions
-
-    if (NOT DEFINED COTIRE_DEBUG_INIT)
-        if (DEFINED COTIRE_DEBUG)
-            set (COTIRE_DEBUG_INIT ${COTIRE_DEBUG})
-        else()
-            set (COTIRE_DEBUG_INIT FALSE)
-        endif()
-    endif()
-    option (COTIRE_DEBUG "Enable cotire debugging output?" ${COTIRE_DEBUG_INIT})
-
-    if (NOT DEFINED COTIRE_VERBOSE_INIT)
-        if (DEFINED COTIRE_VERBOSE)
-            set (COTIRE_VERBOSE_INIT ${COTIRE_VERBOSE})
-        else()
-            set (COTIRE_VERBOSE_INIT FALSE)
-        endif()
-    endif()
-    option (COTIRE_VERBOSE "Enable cotire verbose output?" ${COTIRE_VERBOSE_INIT})
-
-    set (COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_EXTENSIONS "inc;inl;ipp" CACHE STRING
-            "Ignore headers with the listed file extensions from the generated prefix header.")
-
-    set (COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_PATH "" CACHE STRING
-            "Ignore headers from these directories when generating the prefix header.")
-
-    set (COTIRE_UNITY_SOURCE_EXCLUDE_EXTENSIONS "m;mm" CACHE STRING
-            "Ignore sources with the listed file extensions from the generated unity source.")
-
-    set (COTIRE_MINIMUM_NUMBER_OF_TARGET_SOURCES "3" CACHE STRING
-            "Minimum number of sources in target required to enable use of precompiled header.")
-
-    if (NOT DEFINED COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES_INIT)
-        if (DEFINED COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES)
-            set (COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES_INIT ${COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES})
-        elseif ("${CMAKE_GENERATOR}" MATCHES "JOM|Ninja|Visual Studio")
-            # enable parallelization for generators that run multiple jobs by default
-            set (COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES_INIT "-j")
-        else()
-            set (COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES_INIT "0")
-        endif()
-    endif()
-    set (COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES "${COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES_INIT}" CACHE STRING
-            "Maximum number of source files to include in a single unity source file.")
-
-    if (NOT COTIRE_PREFIX_HEADER_FILENAME_SUFFIX)
-        set (COTIRE_PREFIX_HEADER_FILENAME_SUFFIX "_prefix")
-    endif()
-    if (NOT COTIRE_UNITY_SOURCE_FILENAME_SUFFIX)
-        set (COTIRE_UNITY_SOURCE_FILENAME_SUFFIX "_unity")
-    endif()
-    if (NOT COTIRE_INTDIR)
-        set (COTIRE_INTDIR "cotire")
-    endif()
-    if (NOT COTIRE_PCH_ALL_TARGET_NAME)
-        set (COTIRE_PCH_ALL_TARGET_NAME "all_pch")
-    endif()
-    if (NOT COTIRE_UNITY_BUILD_ALL_TARGET_NAME)
-        set (COTIRE_UNITY_BUILD_ALL_TARGET_NAME "all_unity")
-    endif()
-    if (NOT COTIRE_CLEAN_ALL_TARGET_NAME)
-        set (COTIRE_CLEAN_ALL_TARGET_NAME "clean_cotire")
-    endif()
-    if (NOT COTIRE_CLEAN_TARGET_SUFFIX)
-        set (COTIRE_CLEAN_TARGET_SUFFIX "_clean_cotire")
-    endif()
-    if (NOT COTIRE_PCH_TARGET_SUFFIX)
-        set (COTIRE_PCH_TARGET_SUFFIX "_pch")
-    endif()
-    if (MSVC)
-        # MSVC default PCH memory scaling factor of 100 percent (75 MB) is too small for template heavy C++ code
-        # use a bigger default factor of 170 percent (128 MB)
-        if (NOT DEFINED COTIRE_PCH_MEMORY_SCALING_FACTOR)
-            set (COTIRE_PCH_MEMORY_SCALING_FACTOR "170")
-        endif()
-    endif()
-    if (NOT COTIRE_UNITY_BUILD_TARGET_SUFFIX)
-        set (COTIRE_UNITY_BUILD_TARGET_SUFFIX "_unity")
-    endif()
-    if (NOT DEFINED COTIRE_TARGETS_FOLDER)
-        set (COTIRE_TARGETS_FOLDER "cotire")
-    endif()
-    if (NOT DEFINED COTIRE_UNITY_OUTPUT_DIRECTORY)
-        if ("${CMAKE_GENERATOR}" MATCHES "Ninja")
-            # generated Ninja build files do not work if the unity target produces the same output file as the cotired target
-            set (COTIRE_UNITY_OUTPUT_DIRECTORY "unity")
-        else()
-            set (COTIRE_UNITY_OUTPUT_DIRECTORY "")
-        endif()
-    endif()
-
-    # define cotire cache variables
-
-    define_property(
-            CACHED_VARIABLE PROPERTY "COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_PATH"
-            BRIEF_DOCS "Ignore headers from these directories when generating the prefix header."
-            FULL_DOCS
-            "The variable can be set to a semicolon separated list of include directories."
-            "If a header file is found in one of these directories or sub-directories, it will be excluded from the generated prefix header."
-            "If not defined, defaults to empty list."
-    )
-
-    define_property(
-            CACHED_VARIABLE PROPERTY "COTIRE_ADDITIONAL_PREFIX_HEADER_IGNORE_EXTENSIONS"
-            BRIEF_DOCS "Ignore includes with the listed file extensions from the generated prefix header."
-            FULL_DOCS
-            "The variable can be set to a semicolon separated list of file extensions."
-            "If a header file extension matches one in the list, it will be excluded from the generated prefix header."
-            "Includes with an extension in CMAKE_<LANG>_SOURCE_FILE_EXTENSIONS are always ignored."
-            "If not defined, defaults to inc;inl;ipp."
-    )
-
-    define_property(
-            CACHED_VARIABLE PROPERTY "COTIRE_UNITY_SOURCE_EXCLUDE_EXTENSIONS"
-            BRIEF_DOCS "Exclude sources with the listed file extensions from the generated unity source."
-            FULL_DOCS
-            "The variable can be set to a semicolon separated list of file extensions."
-            "If a source file extension matches one in the list, it will be excluded from the generated unity source file."
-            "Source files with an extension in CMAKE_<LANG>_IGNORE_EXTENSIONS are always excluded."
-            "If not defined, defaults to m;mm."
-    )
-
-    define_property(
-            CACHED_VARIABLE PROPERTY "COTIRE_MINIMUM_NUMBER_OF_TARGET_SOURCES"
-            BRIEF_DOCS "Minimum number of sources in target required to enable use of precompiled header."
-            FULL_DOCS
-            "The variable can be set to an integer > 0."
-            "If a target contains less than that number of source files, cotire will not enable the use of the precompiled header for the target."
-            "If not defined, defaults to 3."
-    )
-
-    define_property(
-            CACHED_VARIABLE PROPERTY "COTIRE_MAXIMUM_NUMBER_OF_UNITY_INCLUDES"
-            BRIEF_DOCS "Maximum number of source files to include in a single unity source file."
-            FULL_DOCS
-            "This may be set to an integer >= 0."
-            "If 0, cotire will only create a single unity source file."
-            "If a target contains more than that number of source files, cotire will create multiple unity source files for it."
-            "Can be set to \"-j\" to optimize the count of unity source files for the number of available processor cores."
-            "Can be set to \"-j jobs\" to optimize the number of unity source files for the given number of simultaneous jobs."
-            "Is used to initialize the target property COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES."
-            "Defaults to \"-j\" for the generators Visual Studio, JOM or Ninja. Defaults to 0 otherwise."
-    )
-
-    # define cotire directory properties
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_ENABLE_PRECOMPILED_HEADER"
-            BRIEF_DOCS "Modify build command of cotired targets added in this directory to make use of the generated precompiled header."
-            FULL_DOCS
-            "See target property COTIRE_ENABLE_PRECOMPILED_HEADER."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_ADD_UNITY_BUILD"
-            BRIEF_DOCS "Add a new target that performs a unity build for cotired targets added in this directory."
-            FULL_DOCS
-            "See target property COTIRE_ADD_UNITY_BUILD."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_ADD_CLEAN"
-            BRIEF_DOCS "Add a new target that cleans all cotire generated files for cotired targets added in this directory."
-            FULL_DOCS
-            "See target property COTIRE_ADD_CLEAN."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_PREFIX_HEADER_IGNORE_PATH"
-            BRIEF_DOCS "Ignore headers from these directories when generating the prefix header."
-            FULL_DOCS
-            "See target property COTIRE_PREFIX_HEADER_IGNORE_PATH."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_PREFIX_HEADER_INCLUDE_PATH"
-            BRIEF_DOCS "Honor headers from these directories when generating the prefix header."
-            FULL_DOCS
-            "See target property COTIRE_PREFIX_HEADER_INCLUDE_PATH."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH"
-            BRIEF_DOCS "Header paths matching one of these directories are put at the top of the prefix header."
-            FULL_DOCS
-            "See target property COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_UNITY_SOURCE_PRE_UNDEFS"
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file before the inclusion of each source file."
-            FULL_DOCS
-            "See target property COTIRE_UNITY_SOURCE_PRE_UNDEFS."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_UNITY_SOURCE_POST_UNDEFS"
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file after the inclusion of each source file."
-            FULL_DOCS
-            "See target property COTIRE_UNITY_SOURCE_POST_UNDEFS."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES"
-            BRIEF_DOCS "Maximum number of source files to include in a single unity source file."
-            FULL_DOCS
-            "See target property COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES."
-    )
-
-    define_property(
-            DIRECTORY PROPERTY "COTIRE_UNITY_LINK_LIBRARIES_INIT"
-            BRIEF_DOCS "Define strategy for setting up the unity target's link libraries."
-            FULL_DOCS
-            "See target property COTIRE_UNITY_LINK_LIBRARIES_INIT."
-    )
-
-    # define cotire target properties
-
-    define_property(
-            TARGET PROPERTY "COTIRE_ENABLE_PRECOMPILED_HEADER" INHERITED
-            BRIEF_DOCS "Modify this target's build command to make use of the generated precompiled header."
-            FULL_DOCS
-            "If this property is set to TRUE, cotire will modify the build command to make use of the generated precompiled header."
-            "Irrespective of the value of this property, cotire will setup custom commands to generate the unity source and prefix header for the target."
-            "For makefile based generators cotire will also set up a custom target to manually invoke the generation of the precompiled header."
-            "The target name will be set to this target's name with the suffix _pch appended."
-            "Inherited from directory."
-            "Defaults to TRUE."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_ADD_UNITY_BUILD" INHERITED
-            BRIEF_DOCS "Add a new target that performs a unity build for this target."
-            FULL_DOCS
-            "If this property is set to TRUE, cotire creates a new target of the same type that uses the generated unity source file instead of the target sources."
-            "Most of the relevant target properties will be copied from this target to the new unity build target."
-            "Target dependencies and linked libraries have to be manually set up for the new unity build target."
-            "The unity target name will be set to this target's name with the suffix _unity appended."
-            "Inherited from directory."
-            "Defaults to TRUE."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_ADD_CLEAN" INHERITED
-            BRIEF_DOCS "Add a new target that cleans all cotire generated files for this target."
-            FULL_DOCS
-            "If this property is set to TRUE, cotire creates a new target that clean all files (unity source, prefix header, precompiled header)."
-            "The clean target name will be set to this target's name with the suffix _clean_cotire appended."
-            "Inherited from directory."
-            "Defaults to FALSE."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_PREFIX_HEADER_IGNORE_PATH" INHERITED
-            BRIEF_DOCS "Ignore headers from these directories when generating the prefix header."
-            FULL_DOCS
-            "The property can be set to a list of directories."
-            "If a header file is found in one of these directories or sub-directories, it will be excluded from the generated prefix header."
-            "Inherited from directory."
-            "If not set, this property is initialized to \${CMAKE_SOURCE_DIR};\${CMAKE_BINARY_DIR}."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_PREFIX_HEADER_INCLUDE_PATH" INHERITED
-            BRIEF_DOCS "Honor headers from these directories when generating the prefix header."
-            FULL_DOCS
-            "The property can be set to a list of directories."
-            "If a header file is found in one of these directories or sub-directories, it will be included in the generated prefix header."
-            "If a header file is both selected by COTIRE_PREFIX_HEADER_IGNORE_PATH and COTIRE_PREFIX_HEADER_INCLUDE_PATH,"
-            "the option which yields the closer relative path match wins."
-            "Inherited from directory."
-            "If not set, this property is initialized to the empty list."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_PREFIX_HEADER_INCLUDE_PRIORITY_PATH" INHERITED
-            BRIEF_DOCS "Header paths matching one of these directories are put at the top of prefix header."
-            FULL_DOCS
-            "The property can be set to a list of directories."
-            "Header file paths matching one of these directories will be inserted at the beginning of the generated prefix header."
-            "Header files are sorted according to the order of the directories in the property."
-            "If not set, this property is initialized to the empty list."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_UNITY_SOURCE_PRE_UNDEFS" INHERITED
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file before the inclusion of each target source file."
-            FULL_DOCS
-            "This may be set to a semicolon-separated list of preprocessor symbols."
-            "cotire will add corresponding #undef directives to the generated unit source file before each target source file."
-            "Inherited from directory."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_UNITY_SOURCE_POST_UNDEFS" INHERITED
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file after the inclusion of each target source file."
-            FULL_DOCS
-            "This may be set to a semicolon-separated list of preprocessor symbols."
-            "cotire will add corresponding #undef directives to the generated unit source file after each target source file."
-            "Inherited from directory."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_UNITY_SOURCE_MAXIMUM_NUMBER_OF_INCLUDES" INHERITED
-            BRIEF_DOCS "Maximum number of source files to include in a single unity source file."
-            FULL_DOCS
-            "This may be set to an integer > 0."
-            "If a target contains more than that number of source files, cotire will create multiple unity build files for it."
-            "If not set, cotire will only create a single unity source file."
-            "Inherited from directory."
-            "Defaults to empty."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_<LANG>_UNITY_SOURCE_INIT"
-            BRIEF_DOCS "User provided unity source file to be used instead of the automatically generated one."
-            FULL_DOCS
-            "If set, cotire will only add the given file(s) to the generated unity source file."
-            "If not set, cotire will add all the target source files to the generated unity source file."
-            "The property can be set to a user provided unity source file."
-            "Defaults to empty."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_<LANG>_PREFIX_HEADER_INIT"
-            BRIEF_DOCS "User provided prefix header file to be used instead of the automatically generated one."
-            FULL_DOCS
-            "If set, cotire will add the given header file(s) to the generated prefix header file."
-            "If not set, cotire will generate a prefix header by tracking the header files included by the unity source file."
-            "The property can be set to a user provided prefix header file (e.g., stdafx.h)."
-            "Defaults to empty."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_UNITY_LINK_LIBRARIES_INIT" INHERITED
-            BRIEF_DOCS "Define strategy for setting up unity target's link libraries."
-            FULL_DOCS
-            "If this property is empty or set to NONE, the generated unity target's link libraries have to be set up manually."
-            "If this property is set to COPY, the unity target's link libraries will be copied from this target."
-            "If this property is set to COPY_UNITY, the unity target's link libraries will be copied from this target with considering existing unity targets."
-            "Inherited from directory."
-            "Defaults to empty."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_<LANG>_UNITY_SOURCE"
-            BRIEF_DOCS "Read-only property. The generated <LANG> unity source file(s)."
-            FULL_DOCS
-            "cotire sets this property to the path of the generated <LANG> single computation unit source file for the target."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_<LANG>_PREFIX_HEADER"
-            BRIEF_DOCS "Read-only property. The generated <LANG> prefix header file."
-            FULL_DOCS
-            "cotire sets this property to the full path of the generated <LANG> language prefix header for the target."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_<LANG>_PRECOMPILED_HEADER"
-            BRIEF_DOCS "Read-only property. The generated <LANG> precompiled header file."
-            FULL_DOCS
-            "cotire sets this property to the full path of the generated <LANG> language precompiled header binary for the target."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            TARGET PROPERTY "COTIRE_UNITY_TARGET_NAME"
-            BRIEF_DOCS "The name of the generated unity build target corresponding to this target."
-            FULL_DOCS
-            "This property can be set to the desired name of the unity target that will be created by cotire."
-            "If not set, the unity target name will be set to this target's name with the suffix _unity appended."
-            "After this target has been processed by cotire, the property is set to the actual name of the generated unity target."
-            "Defaults to empty string."
-    )
-
-    # define cotire source properties
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_EXCLUDED"
-            BRIEF_DOCS "Do not modify source file's build command."
-            FULL_DOCS
-            "If this property is set to TRUE, the source file's build command will not be modified to make use of the precompiled header."
-            "The source file will also be excluded from the generated unity source file."
-            "Source files that have their COMPILE_FLAGS property set will be excluded by default."
-            "Defaults to FALSE."
-    )
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_DEPENDENCY"
-            BRIEF_DOCS "Add this source file to dependencies of the automatically generated prefix header file."
-            FULL_DOCS
-            "If this property is set to TRUE, the source file is added to dependencies of the generated prefix header file."
-            "If the file is modified, cotire will re-generate the prefix header source upon build."
-            "Defaults to FALSE."
-    )
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_UNITY_SOURCE_PRE_UNDEFS"
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file before the inclusion of this source file."
-            FULL_DOCS
-            "This may be set to a semicolon-separated list of preprocessor symbols."
-            "cotire will add corresponding #undef directives to the generated unit source file before this file is included."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_UNITY_SOURCE_POST_UNDEFS"
-            BRIEF_DOCS "Preprocessor undefs to place in the generated unity source file after the inclusion of this source file."
-            FULL_DOCS
-            "This may be set to a semicolon-separated list of preprocessor symbols."
-            "cotire will add corresponding #undef directives to the generated unit source file after this file is included."
-            "Defaults to empty string."
-    )
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_START_NEW_UNITY_SOURCE"
-            BRIEF_DOCS "Start a new unity source file which includes this source file as the first one."
-            FULL_DOCS
-            "If this property is set to TRUE, cotire will complete the current unity file and start a new one."
-            "The new unity source file will include this source file as the first one."
-            "This property essentially works as a separator for unity source files."
-            "Defaults to FALSE."
-    )
-
-    define_property(
-            SOURCE PROPERTY "COTIRE_TARGET"
-            BRIEF_DOCS "Read-only property. Mark this source file as cotired for the given target."
-            FULL_DOCS
-            "cotire sets this property to the name of target, that the source file's build command has been altered for."
-            "Defaults to empty string."
-    )
-
-    message (STATUS "cotire ${COTIRE_CMAKE_MODULE_VERSION} loaded.")
-
-endif()
\ No newline at end of file
diff -pruN 4.3.0-2/CMakeLists.txt 4.4.0-1/CMakeLists.txt
--- 4.3.0-2/CMakeLists.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/CMakeLists.txt	2024-05-28 20:08:34.000000000 +0000
@@ -1,7 +1,6 @@
 cmake_minimum_required(VERSION 3.6)
 project(mlpack C CXX)
 
-include(CMake/cotire.cmake)
 include(CMake/CheckHash.cmake)
 include(CMake/Autodownload.cmake)
 include(CMake/ConfigureCrossCompile.cmake)
@@ -17,6 +16,7 @@ option(BUILD_TESTS "Build tests. (Note:
 option(BUILD_CLI_EXECUTABLES "Build command-line executables." ON)
 option(DOWNLOAD_DEPENDENCIES "Automatically download dependencies if not available." OFF)
 option(BUILD_GO_SHLIB "Build Go shared library." OFF)
+option(USE_PRECOMPILED_HEADERS "Use precompiled headers for mlpack_test build." ON)
 
 # Set minimum library versions required by mlpack.
 #
@@ -106,13 +106,13 @@ option(MATHJAX
 option(USE_OPENMP "If available, use OpenMP for parallelization." ON)
 enable_testing()
 
-# Set required standard to C++14.
-set(CMAKE_CXX_STANDARD 14)
+# Set required standard to C++17.
+set(CMAKE_CXX_STANDARD 17)
 set(CMAKE_CXX_STANDARD_REQUIRED ON)
 
 # Ensure that GCC is new enough, if the compiler is GCC.
-if (CMAKE_COMPILER_IS_GNUCC AND CMAKE_CXX_COMPILER_VERSION VERSION_LESS 5)
-  message(FATAL_ERROR "GCC version (${CMAKE_CXX_COMPILER_VERSION}) is too old! 5.x or newer is required.")
+if (CMAKE_COMPILER_IS_GNUCC AND CMAKE_CXX_COMPILER_VERSION VERSION_LESS 8)
+  message(FATAL_ERROR "GCC version (${CMAKE_CXX_COMPILER_VERSION}) is too old! 8.x or newer is required.")
 endif ()
 
 # Include modules in the CMake directory.
@@ -164,7 +164,7 @@ endif ()
 # If we are using MSVC, we need /bigobj.
 if (MSVC)
   set(CMAKE_CXX_STANDARD 17)
-  set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} /bigobj /Zm200 /Zc:__cplusplus")
+  set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} /bigobj /Zc:__cplusplus")
 endif ()
 
 # If we are using MINGW, we need sections and big-obj, otherwise we create too
@@ -312,14 +312,14 @@ if (NOT DOWNLOAD_DEPENDENCIES)
   find_package(StbImage)
 else()
   find_package(StbImage)
-  if (NOT STB_IMAGE_FOUND)
+  if (NOT StbImage_FOUND)
     get_deps(https://mlpack.org/files/stb.tar.gz stb stb.tar.gz)
     set(STB_IMAGE_INCLUDE_DIR ${GENERIC_INCLUDE_DIR})
     find_package(StbImage REQUIRED)
   endif()
 endif()
 
-if (STB_IMAGE_FOUND)
+if (StbImage_FOUND)
   set(STB_AVAILABLE "1")
   set(MLPACK_INCLUDE_DIRS ${MLPACK_INCLUDE_DIRS} "${STB_IMAGE_INCLUDE_DIR}")
 
@@ -339,7 +339,7 @@ if (NOT DOWNLOAD_DEPENDENCIES)
 else()
   find_package(Ensmallen "${ENSMALLEN_VERSION}")
   if (NOT ENSMALLEN_FOUND)
-    get_deps(https://www.ensmallen.org/files/ensmallen-2.20.0.tar.gz ensmallen ensmallen-latest.tar.gz)
+    get_deps(https://www.ensmallen.org/files/ensmallen-2.21.1.tar.gz ensmallen ensmallen-latest.tar.gz)
     set(ENSMALLEN_INCLUDE_DIR ${GENERIC_INCLUDE_DIR})
     find_package(Ensmallen REQUIRED)
   endif()
diff -pruN 4.3.0-2/HISTORY.md 4.4.0-1/HISTORY.md
--- 4.3.0-2/HISTORY.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/HISTORY.md	2024-05-28 20:08:34.000000000 +0000
@@ -1,5 +1,64 @@
-### mlpack 4.3.0
-###### 2023-11-22
+# mlpack changelog
+
+## mlpack 4.4.0
+
+_2024-05-26_
+
+  * Add `print_training_accuracy` option to LogisticRegression bindings (#3552).
+
+  * Fix `preprocess_split()` call in documentation for `LinearRegression` and
+    `AdaBoost` Python classes (#3563).
+
+  * Added `Repeat` ANN layer type (#3565).
+
+  * Remove `round()` implementation for old MSVC compilers (#3570).
+
+  * (R) Added inline plugin to the R bindings to allow for other R packages to
+    link to headers (#3626, h/t @cgiachalis).
+
+  * (R) Removed extra gcc-specific options from `Makevars.win`  (#3627, h/t
+    @kalibera).
+
+  * (R) Changed roxygen package-level documentation from using
+    `@docType package` to `"_PACKAGE"`. (#3636)
+
+  * Fix floating-point accuracy issue for decision trees that sometimes caused
+    crashes (#3595).
+
+  * Use templates metaprog to distinguish between a matrix and a cube type
+    (#3602), (#3585).
+
+  * Use `MatType` instead of `arma::Mat<eT>`, (#3567), (#3607), (#3608),
+    (#3609), (#3568).
+
+  * Generalize matrix operations for armadillo and bandicoot, (#3619), (#3617),
+    (#3610), (#3643), (#3600), (#3605), (#3629).
+
+  * Change `arma::conv_to` to `ConvTo` using a local shim for bandicoot support
+    (#3614).
+
+  * Fix a bug for the stddev and mean in `RandNormal()` #(3651).
+
+  * Allow PCA to take different matrix types (#3677).
+
+  * Fix usage of precompiled headers; remove cotire (#3635).
+
+  * Fix non-working `verbose` option for R bindings (#3691), and add global
+    `mlpack.verbose` option (#3706).
+
+  * Fix divide-by-zero edge case for LARS (#3701).
+
+  * Templatize `SparseCoding` and `LocalCoordinateCoding` to allow different
+    matrix types (#3709, #3711).
+
+  * Fix handling of unused atoms in `LocalCoordinateCoding` (#3711).
+
+  * Move minimum required C++ version from C++14 to C++17 (#3704).
+
+## mlpack 4.3.0
+
+_2023-11-27_
+
   * Fix include ordering issue for `LinearRegression` (#3541).
 
   * Fix L1 regularization in case where weight is zero (#3545).
@@ -17,12 +76,16 @@
   * Fix inconsistent use of the "input" parameter to the Backward method in ANNs
     (#3551).
 
-### mlpack 4.2.1
-###### 2023-09-05
+  * Allow passing weak learner hyperparameters directly to AdaBoost (#3560).
+
+## mlpack 4.2.1
+
+_2023-09-05_
+
   * Reinforcement Learning: Gaussian noise (#3515).
 
-  * Reinforcement Learning: Twin Delayed Deep Deterministic
-    Policy Gradient (#3512).
+  * Reinforcement Learning: Twin Delayed Deep Deterministic Policy Gradient
+    (#3512).
 
   * Reinforcement Learning: Ornstein-Uhlenbeck noise (#3499).
 
@@ -40,9 +103,12 @@
 
   * Fix CRAN package alias for R bindings (#3543).
 
-### mlpack 4.2.0
-###### 2023-06-14
-  * Adapt C_ReLU, ReLU6, FlexibleReLU layer for new neural network API (#3445).
+## mlpack 4.2.0
+
+_2023-06-14_
+
+  * Adapt `C_ReLU`, `ReLU6`, `FlexibleReLU` layers for the new neural network
+    API (#3445).
 
   * Fix PReLU, add integration test to it (#3473).
 
@@ -65,8 +131,10 @@
 
   * Adapt Hyper-Sinh activation function (#3491).
 
-### mlpack 4.1.0
-###### 2023-04-26
+## mlpack 4.1.0
+
+_2023-04-26_
+
   * Adapt HardTanH layer (#3454).
 
   * Adapt Softmin layer for new neural network API (#3437).
@@ -89,8 +157,10 @@
   * Issue runtime error when serialization of neural networks is attempted but
     `MLPACK_ENABLE_ANN_SERIALIZATION` is not defined (#3451).
 
-### mlpack 4.0.1
-###### 2022-12-23
+## mlpack 4.0.1
+
+_2022-12-23_
+
   * Fix mapping of categorical data for Julia bindings (#3305).
 
   * Bugfix: catch all exceptions when running bindings from Julia, instead of
@@ -117,8 +187,10 @@
     element access. In turn this helps tracking down bugs and incorrect
     usage (#3322).
 
-### mlpack 4.0.0
-###### 2022-10-23
+## mlpack 4.0.0
+
+_2022-10-23_
+
   * Bump C++ standard requirement to C++14 (#3233).
 
   * Fix `Perceptron` to work with cross-validation framework (#3190).
@@ -246,16 +318,20 @@
 
   * Replaced Numpy deprecated code in Python bindings (#3126).
 
-### mlpack 3.4.2
-###### 2020-10-26
+## mlpack 3.4.2
+
+_2020-10-26_
+
   * Added Mean Absolute Percentage Error.
 
   * Added Softmin activation function as layer in ann/layer.
 
   * Fix spurious ARMA_64BIT_WORD compilation warnings on 32-bit systems (#2665).
 
-### mlpack 3.4.1
-###### 2020-09-07
+## mlpack 3.4.1
+
+_2020-09-07_
+
   * Fix incorrect parsing of required matrix/model parameters for command-line
     bindings (#2600).
 
@@ -264,8 +340,10 @@
 
   * Remove use of internal Armadillo functionality (#2596, #2601, #2602).
 
-### mlpack 3.4.0
-###### 2020-09-01
+## mlpack 3.4.0
+
+_2020-09-01_
+
   * Issue warnings when metrics produce NaNs in KFoldCV (#2595).
 
   * Added bindings for _R_ during Google Summer of Code (#2556).
@@ -314,8 +392,10 @@
 
   * Change default configuration of `BUILD_PYTHON_BINDINGS` to `OFF` (#2575).
 
-### mlpack 3.3.2
-###### 2020-06-18
+## mlpack 3.3.2
+
+_2020-06-18_
+
   * Added Noisy DQN to q_networks (#2446).
 
   * Add Go bindings (#1884).
@@ -354,8 +434,10 @@
 
   * Add `RBF` layer in ann module to make `RBFN` architecture (#2261).
 
-### mlpack 3.3.1
-###### 2020-04-29
+## mlpack 3.3.1
+
+_2020-04-29_
+
   * Minor Julia and Python documentation fixes (#2373).
 
   * Updated terminal state and fixed bugs for Pendulum environment (#2354,
@@ -375,8 +457,10 @@
   * Add serialization support from Julia; use `mlpack.serialize()` and
     `mlpack.deserialize()` to save and load from `IOBuffer`s.
 
-### mlpack 3.3.0
-###### 2020-04-07
+## mlpack 3.3.0
+
+_2020-04-07_
+
   * Added `Normal Distribution` to `ann/dists` (#2382).
 
   * Templated return type of `Forward function` of loss functions (#2339).
@@ -465,8 +549,10 @@
   * Bugfix for incorrect parameter vector sizes in logistic regression and
     softmax regression (#2359).
 
-### mlpack 3.2.2
-###### 2019-11-26
+## mlpack 3.2.2
+
+_2019-11-26_
+
   * Add `valid` and `same` padding option in `Convolution` and `Atrous
     Convolution` layer (#1988).
 
@@ -489,8 +575,10 @@
 
   * Correctly handle RNN sequences that are shorter than the value of rho (#2102).
 
-### mlpack 3.2.1
-###### 2019-10-01
+## mlpack 3.2.1
+
+_2019-10-01_
+
   * Enforce CMake version check for ensmallen (#2032).
 
   * Fix CMake check for Armadillo version (#2029).
@@ -499,8 +587,10 @@
 
   * Fix Naive Bayes classifier computations in high dimensions (#2022).
 
-### mlpack 3.2.0
-###### 2019-09-25
+## mlpack 3.2.0
+
+_2019-09-25_
+
   * Fix some potential infinity errors in Naive Bayes Classifier (#2022).
 
   * Fix occasionally-failing RADICAL test (#1924).
@@ -541,8 +631,10 @@
   * Bump minimum Armadillo version to 8.400.0 due to ensmallen dependency
     requirement (#2015).
 
-### mlpack 3.1.1
-###### 2019-05-26
+## mlpack 3.1.1
+
+_2019-05-26_
+
   * Fix random forest bug for numerical-only data (#1887).
 
   * Significant speedups for random forest (#1887).
@@ -567,8 +659,10 @@
 
   * Install ensmallen headers when it is downloaded during build (#1900).
 
-### mlpack 3.1.0
-###### 2019-04-25
+## mlpack 3.1.0
+
+_2019-04-25_
+
   * Add DiagonalGaussianDistribution and DiagonalGMM classes to speed up the
     diagonal covariance computation and deprecate DiagonalConstraint (#1666).
 
@@ -601,14 +695,18 @@
   * Python binding improvements: check parameter type (#1717), avoid copying
     Pandas dataframes (#1711), handle Pandas Series objects (#1700).
 
-### mlpack 3.0.4
-###### 2018-11-13
+## mlpack 3.0.4
+
+_2018-11-13_
+
   * Bump minimum CMake version to 3.3.2.
 
   * CMake fixes for Ninja generator by Marc Espie.
 
-### mlpack 3.0.3
-###### 2018-07-27
+## mlpack 3.0.3
+
+_2018-07-27_
+
   * Fix Visual Studio compilation issue (#1443).
 
   * Allow running local_coordinate_coding binding with no initial_dictionary
@@ -624,8 +722,10 @@
 
   * Generate better Python setup.py with documentation (#1460).
 
-### mlpack 3.0.2
-###### 2018-06-08
+## mlpack 3.0.2
+
+_2018-06-08_
+
   * Documentation generation fixes for Python bindings (#1421).
 
   * Fix build error for man pages if command-line bindings are not being built
@@ -641,8 +741,10 @@
   * Add Pendulum environment for reinforcement learning (#1388) and update
     Mountain Car environment (#1394).
 
-### mlpack 3.0.1
-###### 2018-05-10
+## mlpack 3.0.1
+
+_2018-05-10_
+
   * Fix intermittently failing tests (#1387).
 
   * Add big-batch SGD (BBSGD) optimizer in
@@ -667,8 +769,10 @@
     Neural networks: add He and LeCun normal initializations (#1342), add FReLU
     and SELU activation functions (#1346, #1341), add alpha-dropout (#1349).
 
-### mlpack 3.0.0
-###### 2018-03-30
+## mlpack 3.0.0
+
+_2018-03-30_
+
   * Speed and memory improvements for DBSCAN.  --single_mode can now be used for
     situations where previously RAM usage was too high.
 
@@ -697,25 +801,33 @@
 
   * Add randomized Krylov SVD and Block Krylov SVD.
 
-### mlpack 2.2.5
-###### 2017-08-25
+## mlpack 2.2.5
+
+_2017-08-25_
+
   * Compilation fix for some systems (#1082).
 
   * Fix PARAM_INT_OUT() (#1100).
 
-### mlpack 2.2.4
-###### 2017-07-18
+## mlpack 2.2.4
+
+_2017-07-18_
+
   * Speed and memory improvements for DBSCAN. --single_mode can now be used for
     situations where previously RAM usage was too high.
 
   * Fix bug in CF causing incorrect recommendations.
 
-### mlpack 2.2.3
-###### 2017-05-24
+## mlpack 2.2.3
+
+_2017-05-24_
+
   * Bug fix for --predictions_file in mlpack_decision_tree program.
 
-### mlpack 2.2.2
-###### 2017-05-04
+## mlpack 2.2.2
+
+_2017-05-04_
+
   * Install backwards-compatibility mlpack_allknn and mlpack_allkfn programs;
     note they are deprecated and will be removed in mlpack 3.0.0 (#992).
 
@@ -723,13 +835,17 @@
 
   * Small fixes for MiniBatchSGD and SGD and tests.
 
-### mlpack 2.2.1
-###### 2017-04-13
+## mlpack 2.2.1
+
+_2017-04-13_
+
   * Compilation fix for mlpack_nca and mlpack_test on older Armadillo versions
     (#984).
 
-### mlpack 2.2.0
-###### 2017-03-21
+## mlpack 2.2.0
+
+_2017-03-21_
+
   * Bugfix for mlpack_knn program (#816).
 
   * Add decision tree implementation in methods/decision_tree/.  This is very
@@ -745,8 +861,10 @@
   * Refactor categorical CSV loading with boost::spirit for faster loading
     (#681).
 
-### mlpack 2.1.1
-###### 2016-12-22
+## mlpack 2.1.1
+
+_2016-12-22_
+
   * HMMs now use random initialization; this should fix some convergence issues
     (#828).
 
@@ -757,8 +875,10 @@
 
   * Fix DecisionStump to properly work with any input type.
 
-### mlpack 2.1.0
-###### 2016-10-31
+## mlpack 2.1.0
+
+_2016-10-31_
+
   * Fixed CoverTree to properly handle single-point datasets.
 
   * Fixed a bug in CosineTree (and thus QUIC-SVD) that caused split failures for
@@ -779,8 +899,10 @@
     DrusillaSelect in src/mlpack/methods/approx_kfn/, with command-line program
     mlpack_approx_kfn.
 
-### mlpack 2.0.3
-###### 2016-07-21
+## mlpack 2.0.3
+
+_2016-07-21_
+
   * Added multiprobe LSH (#691).  The parameter 'T' to LSHSearch::Search() can
     now be used to control the number of extra bins that are probed, as can the
     -T (--num_probes) option to mlpack_lsh.
@@ -795,8 +917,10 @@
 
   * Added OpenMP support to LSHSearch and mlpack_lsh (#700).
 
-### mlpack 2.0.2
-###### 2016-06-20
+## mlpack 2.0.2
+
+_2016-06-20_
+
   * Added the function LSHSearch::Projections(), which returns an arma::cube
     with each projection table in a slice (#663).  Instead of Projection(i), you
     should now use Projections().slice(i).
@@ -832,8 +956,10 @@
   * Add --kill_empty_clusters option to mlpack_kmeans and KillEmptyClusters
     policy for the KMeans class (#595, #596).
 
-### mlpack 2.0.1
-###### 2016-02-04
+## mlpack 2.0.1
+
+_2016-02-04_
+
   * Fix CMake to properly detect when MKL is being used with Armadillo.
 
   * Minor parameter handling fixes to mlpack_logistic_regression (#504, #505).
@@ -859,8 +985,10 @@
   * CMake fix for projects using mlpack's CMake configuration from elsewhere
     (#512).
 
-### mlpack 2.0.0
-###### 2015-12-24
+## mlpack 2.0.0
+
+_2015-12-24_
+
   * Removed overclustering support from k-means because it is not well-tested,
     may be buggy, and is (I think) unused.  If this was support you were using,
     open a bug or get in touch with us; it would not be hard for us to
@@ -890,31 +1018,35 @@
     DET training code.
 
   * Add support for predicting new test point values to LARS and the
-    command-line 'lars' program.
+    command-line `lars` program.
 
-  * Add serialization support for Perceptron and LogisticRegression.
+  * Add serialization support for `Perceptron` and `LogisticRegression`.
 
-  * Refactor SoftmaxRegression to predict into an arma::Row<size_t> object, and
-    add a softmax_regression program.
+  * Refactor SoftmaxRegression to predict into an `arma::Row<size_t>` object,
+    and add a `softmax_regression` program.
 
   * Refactor LSH to allow loading and saving of models.
 
   * ToString() is removed entirely (#487).
 
-  * Add --input_model_file and --output_model_file options to appropriate
+  * Add `--input_model_file` and `--output_model_file` options to appropriate
     machine learning algorithms.
 
   * Rename all executables to start with an "mlpack" prefix (#229).
 
-  * Add HoeffdingTree and mlpack_hoeffding_tree, an implementation of the
+  * Add HoeffdingTree and `mlpack_hoeffding_tree`, an implementation of the
     streaming decision tree methodology from Domingos and Hulten in 2000.
 
-### mlpack 1.0.12
-###### 2015-01-07
+## mlpack 1.0.12
+
+_2015-01-07_
+
   * Switch to 3-clause BSD license (from LGPL).
 
-### mlpack 1.0.11
-###### 2014-12-11
+## mlpack 1.0.11
+
+_2014-12-11_
+
   * Proper handling of dimension calculation in PCA.
 
   * Load parameter vectors properly for LinearRegression models.
@@ -923,7 +1055,7 @@
 
   * Add support for observation weights to LinearRegression.
 
-  * MahalanobisDistance<> now takes the root of the distance by default and
+  * `MahalanobisDistance<>` now takes the root of the distance by default and
     therefore satisfies the triangle inequality (TakeRoot now defaults to true).
 
   * Better handling of optional Armadillo HDF5 dependency.
@@ -939,8 +1071,10 @@
   * Known bug: CosineTree construction may fail in some cases on i386 systems
     (#358).
 
-### mlpack 1.0.10
-###### 2014-08-29
+## mlpack 1.0.10
+
+_2014-08-29_
+
   * Bugfix for NeighborSearch regression which caused very slow allknn/allkfn.
     Speeds are now restored to approximately 1.0.8 speeds, with significant
     improvement for the cover tree (#347).
@@ -950,8 +1084,10 @@
 
   * Bugfix for compilation under Visual Studio (#348).
 
-### mlpack 1.0.9
-###### 2014-07-28
+## mlpack 1.0.9
+
+_2014-07-28_
+
   * GMM initialization is now safer and provides a working GMM when constructed
     with only the dimensionality and number of Gaussians (#301).
 
@@ -1004,8 +1140,10 @@
 
   * Sparse autoencoder added by Siddharth Agrawal.
 
-### mlpack 1.0.8
-###### 2014-01-06
+## mlpack 1.0.8
+
+_2014-01-06_
+
   * Memory leak in NeighborSearch index-mapping code fixed (#298).
 
   * GMMs can be trained using the existing model as a starting point by
@@ -1022,8 +1160,10 @@
 
   * Fix typos in allkfn and allkrann output.
 
-### mlpack 1.0.7
-###### 2013-10-04
+## mlpack 1.0.7
+
+_2013-10-04_
+
   * Cover tree support for range search (range_search), rank-approximate nearest
     neighbors (allkrann), minimum spanning tree calculation (emst), and FastMKS
     (fastmks).
@@ -1053,12 +1193,16 @@
 
   * Minor updates and fixes for kernels (in mlpack::kernel).
 
-### mlpack 1.0.6
-###### 2013-06-13
+## mlpack 1.0.6
+
+_2013-06-13_
+
   * Minor bugfix so that FastMKS gets built.
 
-### mlpack 1.0.5
-###### 2013-05-01
+## mlpack 1.0.5
+
+_2013-05-01_
+
   * Speedups of cover tree traversers (#235).
 
   * Addition of rank-approximate nearest neighbors (RANN), found in
@@ -1081,8 +1225,10 @@
 
   * Fix for k-furthest-neighbor-search.
 
-### mlpack 1.0.4
-###### 2013-02-08
+## mlpack 1.0.4
+
+_2013-02-08_
+
   * Force minimum Armadillo version to 2.4.2.
 
   * Better output of class types to streams; a class with a ToString() method
@@ -1099,8 +1245,9 @@
   * Better tests for SGD (stochastic gradient descent) and NCA (neighborhood
     components analysis).
 
-### mlpack 1.0.3
-###### 2012-09-16
+## mlpack 1.0.3
+
+_2012-09-16_
 
   * Remove internal sparse matrix support because Armadillo 3.4.0 now includes
     it.  When using Armadillo versions older than 3.4.0, sparse matrix support
@@ -1109,8 +1256,10 @@
   * NCA (neighborhood components analysis) now support an arbitrary optimizer
     (#245), including stochastic gradient descent (#249).
 
-### mlpack 1.0.2
-###### 2012-08-15
+## mlpack 1.0.2
+
+_2012-08-15_
+
   * Added density estimation trees, found in src/mlpack/methods/det/.
 
   * Added non-negative matrix factorization, found in src/mlpack/methods/nmf/.
@@ -1126,8 +1275,10 @@
 
   * Sparse coding dictionary initialization is now a template parameter (#220).
 
-### mlpack 1.0.1
-###### 2012-03-03
+## mlpack 1.0.1
+
+_2012-03-03_
+
   * Added kernel principal components analysis (kernel PCA), found in
     src/mlpack/methods/kernel_pca/ (#74).
 
@@ -1144,8 +1295,10 @@
 
   * Support seeds for random numbers (#177).
 
-### mlpack 1.0.0
-###### 2011-12-17
+## mlpack 1.0.0
+
+_2011-12-17_
+
   * Initial release.  See any resolved tickets numbered less than #196 or
     execute this query:
     http://www.mlpack.org/trac/query?status=closed&milestone=mlpack+1.0.0
diff -pruN 4.3.0-2/README.md 4.4.0-1/README.md
--- 4.3.0-2/README.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/README.md	2024-05-28 20:08:34.000000000 +0000
@@ -1,6 +1,5 @@
 <h2 align="center">
-  <a href="http://mlpack.org"><img
-src="https://cdn.rawgit.com/mlpack/mlpack.org/e7d36ed8/mlpack-black.svg" style="background-color:rgba(0,0,0,0);" height=230 alt="mlpack: a fast, header-only machine learning library"></a>
+  <a href="https://mlpack.org"><img src="https://cdn.jsdelivr.net/gh/mlpack/mlpack.org@e7d36ed8/mlpack-black.svg" style="background-color:rgba(0,0,0,0);" height=230 alt="mlpack: a fast, header-only machine learning library"></a>
   <br>a fast, header-only machine learning library<br>
 </h2>
 
@@ -14,14 +13,14 @@ src="https://cdn.rawgit.com/mlpack/mlpac
 
 <p align="center">
   <a href="https://dev.azure.com/mlpack/mlpack/_build?definitionId=1"><img alt="Azure DevOps builds (job)" src="https://img.shields.io/azure-devops/build/mlpack/84320e87-76e3-4b6e-8b6e-3adaf6b36eed/1/master?job=Linux&label=Linux%20Build&style=flat-square"></a>
-  <a href="https://opensource.org/licenses/BSD-3-Clause"><img src="https://img.shields.io/badge/License-BSD%203--Clause-blue.svg?style=flat-square" alt="License"></a>
-  <a href="http://numfocus.org/donate-to-mlpack"><img src="https://img.shields.io/badge/sponsored%20by-NumFOCUS-orange.svg?style=flat-square&colorA=E1523D&colorB=007D8A" alt="NumFOCUS"></a>
+  <a href="https://opensource.org/license/BSD-3-Clause"><img src="https://img.shields.io/badge/License-BSD%203--Clause-blue.svg?style=flat-square" alt="License"></a>
+  <a href="https://numfocus.org/donate-to-mlpack"><img src="https://img.shields.io/badge/sponsored%20by-NumFOCUS-orange.svg?style=flat-square&colorA=E1523D&colorB=007D8A" alt="NumFOCUS"></a>
 </p>
 
 <p align="center">
   <em>
     Download:
-    <a href="https://www.mlpack.org/files/mlpack-4.3.0.tar.gz">current stable version (4.3.0)</a>
+    <a href="https://www.mlpack.org/files/mlpack-4.4.0.tar.gz">current stable version (4.4.0)</a>
   </em>
 </p>
 
@@ -41,13 +40,13 @@ programs, Python bindings, Julia binding
 
  - Quickstart guides: [C++](doc/quickstart/cpp.md),
    [CLI](doc/quickstart/cli.md), [Python](doc/quickstart/python.md),
-   [R](doc/quickstart/R.md), [Julia](doc/quickstart/julia.md),
+   [R](doc/quickstart/r.md), [Julia](doc/quickstart/julia.md),
    [Go](doc/quickstart/go.md)
  - [mlpack homepage](https://www.mlpack.org/)
  - [mlpack documentation](https://www.mlpack.org/docs.html)
  - [Examples repository](https://github.com/mlpack/examples/)
  - [Tutorials](doc/tutorials/README.md)
- - [Development Site (Github)](https://www.github.com/mlpack/mlpack/)
+ - [Development Site (Github)](https://github.com/mlpack/mlpack/)
 
 [//]: # (numfocus-fiscal-sponsor-attribution)
 
@@ -59,9 +58,10 @@ variety of other needs.
 
 <div align="center">
   <a href="https://numfocus.org/">
-    <img height="60px"
+    <img height="60"
          src="https://raw.githubusercontent.com/numfocus/templates/master/images/numfocus-logo.png"
-         align="center">
+         align="middle"
+         alt="NumFOCUS logo">
   </a>
 </div>
 <br>
@@ -108,7 +108,7 @@ Citations are beneficial for the growth
 ## 2. Dependencies
 
 **mlpack** requires the following additional dependencies:
- - C++14 compiler
+ - C++17 compiler
  - [Armadillo](https://arma.sourceforge.net)      &nbsp;&emsp;>= 9.800
  - [ensmallen](https://ensmallen.org)      &emsp;>= 2.10.0
  - [cereal](http://uscilab.github.io/cereal/)         &ensp;&nbsp;&emsp;&emsp;>= 1.1.2
@@ -182,7 +182,7 @@ is `my_program.cpp`, your compiler is GC
 OpenMP support (recommended) and optimizations, compile like this:
 
 ```sh
-g++ -O3 -std=c++14 -o my_program my_program.cpp -larmadillo -fopenmp
+g++ -O3 -std=c++17 -o my_program my_program.cpp -larmadillo -fopenmp
 ```
 
 Note that if you want to serialize (save or load) neural networks, you should
@@ -191,8 +191,8 @@ If you don't define `MLPACK_ENABLE_ANN_S
 neural network, a compilation error will occur.
 
 See the [C++ quickstart](doc/quickstart/cpp.md) and the
-[examples](https://github.com/mlpack/examples) repository for some examples of
-mlpack applications in C++, with corresponding `Makefile`s.
+[examples](https://github.com/mlpack/examples) repository for some examples
+of mlpack applications in C++, with corresponding `Makefile`s.
 
 #### 3.1.a. Linking with autodownloaded Armadillo
 
@@ -217,7 +217,7 @@ dependencies of Armadillo.  For example,
 available, compilation can be done like this:
 
 ```sh
-g++ -O3 -std=c++14 -o my_program my_program.cpp -lopenblas -fopenmp
+g++ -O3 -std=c++17 -o my_program my_program.cpp -lopenblas -fopenmp
 ```
 
 See [the Armadillo documentation](https://arma.sourceforge.net/faq.html#linking)
@@ -270,7 +270,7 @@ bindings for many languages at once.
 
 The command-line programs have no extra dependencies.  The set of programs that
 will be compiled is detailed and documented on the [command-line program
-documentation page](https://www.mlpack.org/doc/stable/cli_documentation.html).
+documentation page](doc/user/bindings/cli.md).
 
 From the root of the mlpack sources, run the following commands to build and
 install the command-line bindings:
@@ -290,7 +290,7 @@ build in parallel; e.g., `make -j4` will
 *See also the [Python quickstart](doc/quickstart/python.md).*
 
 mlpack's Python bindings are available on
-[PyPI](https://pypi.org/project/mlpack) and
+[PyPI](https://pypi.org/project/mlpack/) and
 [conda-forge](https://anaconda.org/conda-forge/mlpack), and can be installed
 with either `pip install mlpack` or `conda install -c conda-forge mlpack`.
 These sources are recommended, as building the Python bindings by hand can be
@@ -322,7 +322,7 @@ specify a custom Python interpreter with
 
 ### 4.iii. R bindings
 
-*See also the [R quickstart](doc/quickstart/R.md).*
+*See also the [R quickstart](doc/quickstart/r.md).*
 
 mlpack's R bindings are available as the R package
 [mlpack](https://cran.r-project.org/web/packages/mlpack/index.html) on CRAN.
@@ -439,12 +439,13 @@ More documentation is available for both
 
 ***User documentation***:
 
- - [File formats and loading data in mlpack](doc/user/formats.md)
  - [Matrices in mlpack](doc/user/matrices.md)
+ - [Loading and saving mlpack objects](doc/user/load_save.md)
  - [Cross-Validation](doc/user/cv.md)
  - [Hyper-parameter Tuning](doc/user/hpt.md)
  - [Building mlpack from source on Windows](doc/user/build_windows.md)
  - [Sample C++ ML App for Windows](doc/user/sample_ml_app.md)
+ - [mlpack core library documentation](doc/user/core.md)
  - [Examples repository](https://github.com/mlpack/examples/)
 
 ***Tutorials:***
@@ -466,7 +467,6 @@ More documentation is available for both
 
 ***Developer documentation***:
 
- - [mlpack versions in code](doc/developer/version.md)
  - [Writing an mlpack binding](doc/developer/iodoc.md)
  - [mlpack Timers](doc/developer/timer.md)
  - [mlpack automatic bindings to other languages](doc/developer/bindings.md)
@@ -480,6 +480,6 @@ future, see the [vision document](https:
 
 If you have problems, find a bug, or need help, you can try visiting
 the [mlpack help](https://www.mlpack.org/questions.html) page, or [mlpack on
-Github](https://www.github.com/mlpack/mlpack/).  Alternately, mlpack help can be
+Github](https://github.com/mlpack/mlpack/).  Alternately, mlpack help can be
 found on Matrix at `#mlpack`; see also the
 [community](https://www.mlpack.org/community.html) page.
diff -pruN 4.3.0-2/board/flags-config.cmake 4.4.0-1/board/flags-config.cmake
--- 4.3.0-2/board/flags-config.cmake	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/board/flags-config.cmake	2024-05-28 20:08:34.000000000 +0000
@@ -9,7 +9,7 @@ set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS}
 set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -fomit-frame-pointer -fno-unwind-tables")
 set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -fno-asynchronous-unwind-tables -fvisibility=hidden")
 set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -fshort-enums -finline-small-functions")
-set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -findirect-inlining -fno-common") 
+set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -findirect-inlining -fno-common")
 #-flto -fuse-ld=gold # There is an issue with gold link when compiling on 
 # Ubuntu 16. At that point gcc linker did not integrate the flto support 
 # inside and it was a separate plugin that need to be added. Therefore, 
diff -pruN 4.3.0-2/debian/changelog 4.4.0-1/debian/changelog
--- 4.3.0-2/debian/changelog	2024-01-11 09:40:13.000000000 +0000
+++ 4.4.0-1/debian/changelog	2024-05-29 15:23:48.000000000 +0000
@@ -1,3 +1,10 @@
+mlpack (4.4.0-1) unstable; urgency=medium
+
+  * New upstream version
+  * Forward port patches
+
+ -- Barak A. Pearlmutter <bap@debian.org>  Wed, 29 May 2024 16:23:48 +0100
+
 mlpack (4.3.0-2) unstable; urgency=medium
 
   * Team upload
diff -pruN 4.3.0-2/debian/patches/0001-do-not-disable-optimization.patch 4.4.0-1/debian/patches/0001-do-not-disable-optimization.patch
--- 4.3.0-2/debian/patches/0001-do-not-disable-optimization.patch	2023-10-20 19:57:18.000000000 +0000
+++ 4.4.0-1/debian/patches/0001-do-not-disable-optimization.patch	2024-05-29 15:23:48.000000000 +0000
@@ -8,7 +8,7 @@ Do not disable optimization (-O0) just b
  1 file changed, 2 insertions(+), 2 deletions(-)
 
 diff --git a/CMakeLists.txt b/CMakeLists.txt
-index cc82692..034663f 100644
+index bf61ad9..8072726 100644
 --- a/CMakeLists.txt
 +++ b/CMakeLists.txt
 @@ -226,8 +226,8 @@ set (BFD_DL_AVAILABLE "NO")
diff -pruN 4.3.0-2/debian/patches/no-runtime-cython.patch 4.4.0-1/debian/patches/no-runtime-cython.patch
--- 4.3.0-2/debian/patches/no-runtime-cython.patch	2024-01-10 15:14:10.000000000 +0000
+++ 4.4.0-1/debian/patches/no-runtime-cython.patch	2024-05-29 15:23:48.000000000 +0000
@@ -1,10 +1,17 @@
-Description: Drop the runtime dependency on Cython
-Author: Graham Inggs <ginggs@debian.org>
+From: Graham Inggs <ginggs@debian.org>
+Date: Wed, 29 May 2024 16:24:18 +0100
+Subject: Drop the runtime dependency on Cython
+
 Last-Update: 2024-01-10
+---
+ src/mlpack/bindings/python/setup.py.in | 2 +-
+ 1 file changed, 1 insertion(+), 1 deletion(-)
 
+diff --git a/src/mlpack/bindings/python/setup.py.in b/src/mlpack/bindings/python/setup.py.in
+index 54cf2f0..1b4ad89 100644
 --- a/src/mlpack/bindings/python/setup.py.in
 +++ b/src/mlpack/bindings/python/setup.py.in
-@@ -138,7 +138,7 @@
+@@ -144,7 +144,7 @@ setup(name='mlpack',
            'Documentation': 'http://www.mlpack.org/doc/mlpack-${PACKAGE_VERSION}/python.html',
            'Source': 'https://github.com/mlpack/mlpack/',
            'Tracker': 'https://github.com/mlpack/mlpack/issues'},
@@ -12,4 +19,4 @@ Last-Update: 2024-01-10
 +      install_requires=['numpy', 'pandas'],
        package_dir={ '': '.' }, # Might be superfluous.
        packages=['mlpack'],
-       cmdclass={ 'build_ext': build_ext },
+       package_data={ 'mlpack': include_files + ['../mlpack/share/pkgconfig/mlpack.pc'] },
diff -pruN 4.3.0-2/doc/citation.md 4.4.0-1/doc/citation.md
--- 4.3.0-2/doc/citation.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/citation.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,52 @@
+# Citation details
+
+If you use mlpack in your research or software, please cite mlpack using the
+citation for
+[the paper below](https://joss.theoj.org/papers/10.21105/joss.05026):
+
+ * ***mlpack 4: a fast, header-only C++ machine learning library***.  R.R.
+   Curtin, M. Edel, O. Shrit, S. Agrawal, S. Basak, J.J. Balamuta, R.
+   Birmingham, K. Dutt, D. Eddelbuettel, R. Garg, S. Jaiswal, A. Kaushik, S.
+   Kim, A. Mukherjee, N.G. Sai, N. Sharma, Y.S. Parihar, R. Swain, C. Sanderson.
+   _Journal of Open Source Software_ 8:82, p. 5026, 2023.
+
+In BibTeX format:
+
+```
+@article{mlpack2023,
+    title     = {mlpack 4: a fast, header-only C++ machine learning library},
+    author    = {Ryan R. Curtin and Marcus Edel and Omar Shrit and
+                 Shubham Agrawal and Suryoday Basak and James J. Balamuta and
+                 Ryan Birmingham and Kartik Dutt and Dirk Eddelbuettel and
+                 Rishabh Garg and Shikhar Jaiswal and Aakash Kaushik and
+                 Sangyeon Kim and Anjishnu Mukherjee and Nanubala Gnana Sai and
+                 Nippun Sharma and Yashwant Singh Parihar and Roshan Swain and
+                 Conrad Sanderson},
+    journal   = {Journal of Open Source Software},
+    volume    = {8},
+    number    = {82},
+    pages     = {5026},
+    year      = {2023},
+    doi       = {10.21105/joss.05026},
+    url       = {https://doi.org/10.21105/joss.05026}
+}
+```
+
+Citations are beneficial for the growth and improvement of mlpack.
+
+### Older papers
+
+See also the following older papers concerning mlpack, its community, and its
+internal design.
+
+ * [mlpack 3: a fast, flexible machine learning
+   library](https://joss.theoj.org/papers/10.21105/joss.00726) (2018)
+
+ * [mlpack open-source machine learning library and
+   community](http://kurg.org/pub/pdf/2018mlossmlpack.pdf) (2018)
+
+ * [Designing and building the mlpack open-source machine learning
+   library](https://arxiv.org/abs/1708.05279) (2017)
+
+ * [mlpack: a scalable C++ machine learning
+   library](https://www.jmlr.org/papers/volume14/curtin13a/curtin13a.pdf) (2013)
diff -pruN 4.3.0-2/doc/css/gfm-mod.css 4.4.0-1/doc/css/gfm-mod.css
--- 4.3.0-2/doc/css/gfm-mod.css	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/css/gfm-mod.css	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,1155 @@
+/**
+ * GitHub Flavored Markdown Stylesheet
+ *
+ * MIT License
+ * Dan Untenzu - @pixelbrackets (https://pixelbrackets.de)
+ */
+
+/**
+ * tuzz/github.css body styles
+ * Source https://gist.github.com/tuzz/3331384 (MIT), version 8
+ */
+
+body {
+  font-family: Helvetica, arial, sans-serif;
+  font-size: 14px;
+  line-height: 1.4;
+  background-color: white;
+  color: #333;
+}
+
+div#content {
+  padding-top: 10px;
+  padding-bottom: 10px;
+  padding: 30px;
+  max-width: 920px;
+  margin: auto;
+}
+
+body > *:first-child {
+  margin-top: 0 !important;
+}
+
+body > *:last-child {
+  margin-bottom: 0 !important;
+}
+
+/**
+ * sindresorhus/github-markdown-css without the ».markdown-body« prefix
+ * Source https://github.com/sindresorhus/github-markdown-css (MIT), version 4.0
+ */
+
+.octicon {
+  display: inline-block;
+  fill: currentColor;
+  vertical-align: text-bottom;
+}
+
+.anchor {
+  float: left;
+  line-height: 1;
+  margin-left: -20px;
+  padding-right: 4px;
+}
+
+.anchor:focus {
+  outline: none;
+}
+
+h1 .octicon-link,
+h2 .octicon-link,
+h3 .octicon-link,
+h4 .octicon-link,
+h5 .octicon-link,
+h6 .octicon-link {
+  color: #1b1f23;
+  vertical-align: middle;
+  visibility: hidden;
+}
+
+h1:hover .anchor,
+h2:hover .anchor,
+h3:hover .anchor,
+h4:hover .anchor,
+h5:hover .anchor,
+h6:hover .anchor {
+  text-decoration: none;
+}
+
+h1:hover .anchor .octicon-link,
+h2:hover .anchor .octicon-link,
+h3:hover .anchor .octicon-link,
+h4:hover .anchor .octicon-link,
+h5:hover .anchor .octicon-link,
+h6:hover .anchor .octicon-link {
+  visibility: visible;
+}
+
+h1:hover .anchor .octicon-link:before,
+h2:hover .anchor .octicon-link:before,
+h3:hover .anchor .octicon-link:before,
+h4:hover .anchor .octicon-link:before,
+h5:hover .anchor .octicon-link:before,
+h6:hover .anchor .octicon-link:before {
+  width: 16px;
+  height: 16px;
+  content: ' ';
+  display: inline-block;
+  background-image: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' version='1.1' width='16' height='16' aria-hidden='true'%3E%3Cpath fill-rule='evenodd' d='M4 9h1v1H4c-1.5 0-3-1.69-3-3.5S2.55 3 4 3h4c1.45 0 3 1.69 3 3.5 0 1.41-.91 2.72-2 3.25V8.59c.58-.45 1-1.27 1-2.09C10 5.22 8.98 4 8 4H4c-.98 0-2 1.22-2 2.5S3 9 4 9zm9-3h-1v1h1c1 0 2 1.22 2 2.5S13.98 12 13 12H9c-.98 0-2-1.22-2-2.5 0-.83.42-1.64 1-2.09V6.25c-1.09.53-2 1.84-2 3.25C6 11.31 7.55 13 9 13h4c1.45 0 3-1.69 3-3.5S14.5 6 13 6z'%3E%3C/path%3E%3C/svg%3E");
+}{
+  -ms-text-size-adjust: 100%;
+  -webkit-text-size-adjust: 100%;
+  line-height: 1.5;
+  color: #24292e;
+  font-family: -apple-system,BlinkMacSystemFont,Segoe UI,Helvetica,Arial,sans-serif,Apple Color Emoji,Segoe UI Emoji;
+  font-size: 16px;
+  line-height: 1.5;
+  word-wrap: break-word;
+}
+
+details {
+  display: block;
+}
+
+summary {
+  display: list-item;
+}
+
+a {
+  background-color: initial;
+}
+
+a:active,
+a:hover {
+  outline-width: 0;
+}
+
+strong {
+  font-weight: inherit;
+  font-weight: bolder;
+}
+
+h1 {
+  font-size: 2em;
+  margin: .67em 0;
+}
+
+img {
+  border-style: none;
+}
+
+code,
+kbd,
+pre {
+  font-family: monospace,monospace;
+  font-size: 1em;
+  white-space: pre;
+}
+
+hr {
+  box-sizing: initial;
+  height: 0;
+  overflow: visible;
+}
+
+input {
+  font: inherit;
+  margin: 0;
+}
+
+input {
+  overflow: visible;
+}
+
+[type=checkbox] {
+  box-sizing: border-box;
+  padding: 0;
+}
+
+* {
+  box-sizing: border-box;
+}
+
+input {
+  font-family: inherit;
+  font-size: inherit;
+  line-height: inherit;
+}
+
+a {
+  color: #0366d6;
+  text-decoration: none;
+}
+
+a:hover {
+  text-decoration: underline;
+}
+
+strong {
+  font-weight: 600;
+}
+
+hr {
+  height: 0;
+  margin: 15px 0;
+  overflow: hidden;
+  background: transparent;
+  border: 0;
+  border-bottom: 1px solid #dfe2e5;
+}
+
+hr:after,
+hr:before {
+  display: table;
+  content: "";
+}
+
+hr:after {
+  clear: both;
+}
+
+table {
+  border-spacing: 0;
+  border-collapse: collapse;
+}
+
+td,
+th {
+  padding: 0;
+}
+
+details summary {
+  cursor: pointer;
+}
+
+kbd {
+  display: inline-block;
+  padding: 3px 5px;
+  font: 11px SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  line-height: 10px;
+  color: #444d56;
+  vertical-align: middle;
+  background-color: #fafbfc;
+  border: 1px solid #d1d5da;
+  border-radius: 3px;
+  box-shadow: inset 0 -1px 0 #d1d5da;
+}
+
+h1,
+h2,
+h3,
+h4,
+h5,
+h6 {
+  margin-top: 0;
+  margin-bottom: 0;
+}
+
+h1 {
+  font-size: 32px;
+}
+
+h1,
+h2 {
+  font-weight: 600;
+}
+
+h2 {
+  font-size: 24px;
+}
+
+h3 {
+  font-size: 20px;
+}
+
+h3,
+h4 {
+  font-weight: 600;
+}
+
+h4 {
+  font-size: 16px;
+  text-decoration: underline;
+}
+
+h5 {
+  font-size: 14px;
+}
+
+h5,
+h6 {
+  font-weight: 600;
+}
+
+h6 {
+  font-size: 12px;
+}
+
+p {
+  margin-top: 0;
+  margin-bottom: 10px;
+}
+
+blockquote {
+  margin: 0;
+}
+
+ol,
+ul {
+  padding-left: 0;
+  margin-top: 0;
+  margin-bottom: 0;
+}
+
+ol ol,
+ul ol {
+  list-style-type: lower-roman;
+}
+
+ol ol ol,
+ol ul ol,
+ul ol ol,
+ul ul ol {
+  list-style-type: lower-alpha;
+}
+
+dd {
+  margin-left: 0;
+}
+
+code,
+pre {
+  font-family: SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  font-size: 12px;
+}
+
+pre {
+  margin-top: 0;
+  margin-bottom: 0;
+}
+
+input::-webkit-inner-spin-button,
+input::-webkit-outer-spin-button {
+  margin: 0;
+  -webkit-appearance: none;
+  appearance: none;
+}
+
+:checked+.radio-label {
+  position: relative;
+  z-index: 1;
+  border-color: #0366d6;
+}
+
+.border {
+  border: 1px solid #e1e4e8!important;
+}
+
+.border-0 {
+  border: 0!important;
+}
+
+.border-bottom {
+  border-bottom: 1px solid #e1e4e8!important;
+}
+
+.rounded-1 {
+  border-radius: 3px!important;
+}
+
+.bg-white {
+  background-color: #fff!important;
+}
+
+.bg-gray-light {
+  background-color: #fafbfc!important;
+}
+
+.text-gray-light {
+  color: #6a737d!important;
+}
+
+.mb-0 {
+  margin-bottom: 0!important;
+}
+
+.my-2 {
+  margin-top: 8px!important;
+  margin-bottom: 8px!important;
+}
+
+.pl-0 {
+  padding-left: 0!important;
+}
+
+.py-0 {
+  padding-top: 0!important;
+  padding-bottom: 0!important;
+}
+
+.pl-1 {
+  padding-left: 4px!important;
+}
+
+.pl-2 {
+  padding-left: 8px!important;
+}
+
+.py-2 {
+  padding-top: 8px!important;
+  padding-bottom: 8px!important;
+}
+
+.pl-3,
+.px-3 {
+  padding-left: 16px!important;
+}
+
+.px-3 {
+  padding-right: 16px!important;
+}
+
+.pl-4 {
+  padding-left: 24px!important;
+}
+
+.pl-5 {
+  padding-left: 32px!important;
+}
+
+.pl-6 {
+  padding-left: 40px!important;
+}
+
+.f6 {
+  font-size: 12px!important;
+}
+
+.lh-condensed {
+  line-height: 1.25!important;
+}
+
+.text-bold {
+  font-weight: 600!important;
+}
+
+.pl-c {
+  color: #6a737d;
+}
+
+.pl-c1,
+.pl-s .pl-v {
+  color: #005cc5;
+}
+
+.pl-e,
+.pl-en {
+  color: #6f42c1;
+}
+
+.pl-s .pl-s1,
+.pl-smi {
+  color: #24292e;
+}
+
+.pl-ent {
+  color: #22863a;
+}
+
+.pl-k {
+  color: #d73a49;
+}
+
+.pl-pds,
+.pl-s,
+.pl-s .pl-pse .pl-s1,
+.pl-sr,
+.pl-sr .pl-cce,
+.pl-sr .pl-sra,
+.pl-sr .pl-sre {
+  color: #032f62;
+}
+
+.pl-smw,
+.pl-v {
+  color: #e36209;
+}
+
+.pl-bu {
+  color: #b31d28;
+}
+
+.pl-ii {
+  color: #fafbfc;
+  background-color: #b31d28;
+}
+
+.pl-c2 {
+  color: #fafbfc;
+  background-color: #d73a49;
+}
+
+.pl-c2:before {
+  content: "^M";
+}
+
+.pl-sr .pl-cce {
+  font-weight: 700;
+  color: #22863a;
+}
+
+.pl-ml {
+  color: #735c0f;
+}
+
+.pl-mh,
+.pl-mh .pl-en,
+.pl-ms {
+  font-weight: 700;
+  color: #005cc5;
+}
+
+.pl-mi {
+  font-style: italic;
+  color: #24292e;
+}
+
+.pl-mb {
+  font-weight: 700;
+  color: #24292e;
+}
+
+.pl-md {
+  color: #b31d28;
+  background-color: #ffeef0;
+}
+
+.pl-mi1 {
+  color: #22863a;
+  background-color: #f0fff4;
+}
+
+.pl-mc {
+  color: #e36209;
+  background-color: #ffebda;
+}
+
+.pl-mi2 {
+  color: #f6f8fa;
+  background-color: #005cc5;
+}
+
+.pl-mdr {
+  font-weight: 700;
+  color: #6f42c1;
+}
+
+.pl-ba {
+  color: #586069;
+}
+
+.pl-sg {
+  color: #959da5;
+}
+
+.pl-corl {
+  text-decoration: underline;
+  color: #032f62;
+}
+
+.mb-0 {
+  margin-bottom: 0!important;
+}
+
+.my-2 {
+  margin-bottom: 8px!important;
+}
+
+.my-2 {
+  margin-top: 8px!important;
+}
+
+.pl-0 {
+  padding-left: 0!important;
+}
+
+.py-0 {
+  padding-top: 0!important;
+  padding-bottom: 0!important;
+}
+
+.pl-1 {
+  padding-left: 4px!important;
+}
+
+.pl-2 {
+  padding-left: 8px!important;
+}
+
+.py-2 {
+  padding-top: 8px!important;
+  padding-bottom: 8px!important;
+}
+
+.pl-3 {
+  padding-left: 16px!important;
+}
+
+.pl-4 {
+  padding-left: 24px!important;
+}
+
+.pl-5 {
+  padding-left: 32px!important;
+}
+
+.pl-6 {
+  padding-left: 40px!important;
+}
+
+.pl-7 {
+  padding-left: 48px!important;
+}
+
+.pl-8 {
+  padding-left: 64px!important;
+}
+
+.pl-9 {
+  padding-left: 80px!important;
+}
+
+.pl-10 {
+  padding-left: 96px!important;
+}
+
+.pl-11 {
+  padding-left: 112px!important;
+}
+
+.pl-12 {
+  padding-left: 128px!important;
+}
+
+hr {
+  border-bottom-color: #eee;
+}
+
+kbd {
+  display: inline-block;
+  padding: 3px 5px;
+  font: 11px SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  line-height: 10px;
+  color: #444d56;
+  vertical-align: middle;
+  background-color: #fafbfc;
+  border: 1px solid #d1d5da;
+  border-radius: 3px;
+  box-shadow: inset 0 -1px 0 #d1d5da;
+}
+
+.markdown-body:after,
+.markdown-body:before {
+  display: table;
+  content: "";
+}
+
+.markdown-body:after {
+  clear: both;
+}
+
+.markdown-body>:first-child {
+  margin-top: 0!important;
+}
+
+.markdown-body>:last-child {
+  margin-bottom: 0!important;
+}
+
+a:not([href]) {
+  color: inherit;
+  text-decoration: none;
+}
+
+blockquote,
+details,
+dl,
+ol,
+p,
+pre,
+table,
+ul {
+  margin-top: 0;
+  margin-bottom: 16px;
+}
+
+hr {
+  height: 1px;
+  padding: 0;
+  margin: 24px 0;
+  background-color: #7a7c7f;
+  border: 0;
+  width: 75%;
+  position: relative;
+  left: 50%;
+  transform: translateX(-50%);
+}
+
+blockquote {
+  padding: 0 1em;
+  color: #6a737d;
+  border-left: .25em solid #dfe2e5;
+}
+
+blockquote>:first-child {
+  margin-top: 0;
+}
+
+blockquote>:last-child {
+  margin-bottom: 0;
+}
+
+h1,
+h2,
+h3,
+h4,
+h5,
+h6 {
+  margin-top: 24px;
+  margin-bottom: 16px;
+  font-weight: 600;
+  line-height: 1.25;
+}
+
+h1 {
+  font-size: 2em;
+}
+
+h1,
+h2 {
+  padding-bottom: .3em;
+  border-bottom: 3px solid #7a7c7f;
+}
+
+h2 {
+  font-size: 1.5em;
+}
+
+h3 {
+  font-size: 1.25em;
+  border-bottom: 1px solid #7a7c7f;
+}
+
+h4 {
+  font-size: 1em;
+  margin-top: 1em;
+}
+
+h5 {
+  font-size: .875em;
+}
+
+h6 {
+  font-size: .85em;
+  color: #6a737d;
+}
+
+ol,
+ul {
+  padding-left: 2em;
+}
+
+ol ol,
+ol ul,
+ul ol,
+ul ul {
+  margin-top: 0.5em;
+  margin-bottom: 1em;
+}
+
+li {
+  word-wrap: break-all;
+  line-height: 1.2;
+}
+
+li>p {
+  margin-top: 16px;
+}
+
+li+li {
+  margin-top: .25em;
+}
+
+dl {
+  padding: 0;
+}
+
+dl dt {
+  padding: 0;
+  margin-top: 16px;
+  font-size: 1em;
+  font-style: italic;
+  font-weight: 600;
+}
+
+dl dd {
+  padding: 0 16px;
+  margin-bottom: 16px;
+}
+
+table {
+  font-size: 14px;
+  display: inline-block;
+  width: auto;
+  overflow: auto;
+  position: relative;
+  left: 50%;
+  transform: translateX(-50%);
+}
+
+table th {
+  font-weight: 600;
+  min-width: 80px;
+}
+
+table td,
+table th {
+  padding: 4px 10px;
+}
+
+table tr {
+  background-color: #fff;
+}
+
+table tr:nth-child(2n) {
+  background-color: #f6f8fa;
+}
+
+table thead tr {
+  border-top: none;
+  border-bottom: 2px solid #7a7c7f;
+}
+
+img {
+  max-width: 100%;
+  box-sizing: initial;
+  background-color: #fff;
+}
+
+img[align=right] {
+  padding-left: 20px;
+}
+
+img[align=left] {
+  padding-right: 20px;
+}
+
+code {
+  padding: .2em .4em;
+  margin: 0;
+  font-size: 85%;
+  background-color: rgba(27,31,35,.05);
+  border-radius: 3px;
+}
+
+pre {
+  word-wrap: normal;
+}
+
+pre>code {
+  padding: 0;
+  margin: 0;
+  font-size: 100%;
+  word-break: normal;
+  white-space: pre;
+  background: transparent;
+  border: 0;
+}
+
+.highlight {
+  margin-bottom: 16px;
+}
+
+.highlight pre {
+  margin-bottom: 0;
+  word-break: normal;
+}
+
+.highlight pre,
+pre {
+  padding: 16px;
+  overflow: auto;
+  font-size: 85%;
+  line-height: 1.4;
+  background-color: #f6f8fa;
+  border-radius: 3px;
+  display: inline-block;
+}
+
+pre code {
+  display: inline;
+  max-width: auto;
+  padding: 0;
+  margin: 0;
+  overflow: visible;
+  line-height: inherit;
+  word-wrap: normal;
+  background-color: initial;
+  border: 0;
+}
+
+.commit-tease-sha {
+  display: inline-block;
+  font-family: SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  font-size: 90%;
+  color: #444d56;
+}
+
+.full-commit .btn-outline:not(:disabled):hover {
+  color: #005cc5;
+  border-color: #005cc5;
+}
+
+.blob-wrapper {
+  overflow-x: auto;
+  overflow-y: hidden;
+}
+
+.blob-wrapper-embedded {
+  max-height: 240px;
+  overflow-y: auto;
+}
+
+.blob-num {
+  width: 1%;
+  min-width: 50px;
+  padding-right: 10px;
+  padding-left: 10px;
+  font-family: SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  font-size: 12px;
+  line-height: 20px;
+  color: rgba(27,31,35,.3);
+  text-align: right;
+  white-space: nowrap;
+  vertical-align: top;
+  cursor: pointer;
+  -webkit-user-select: none;
+  -moz-user-select: none;
+  -ms-user-select: none;
+  user-select: none;
+}
+
+.blob-num:hover {
+  color: rgba(27,31,35,.6);
+}
+
+.blob-num:before {
+  content: attr(data-line-number);
+}
+
+.blob-code {
+  position: relative;
+  padding-right: 10px;
+  padding-left: 10px;
+  line-height: 20px;
+  vertical-align: top;
+}
+
+.blob-code-inner {
+  overflow: visible;
+  font-family: SFMono-Regular,Consolas,Liberation Mono,Menlo,monospace;
+  font-size: 12px;
+  color: #24292e;
+  word-wrap: normal;
+  white-space: pre;
+}
+
+.pl-token.active,
+.pl-token:hover {
+  cursor: pointer;
+  background: #ffea7f;
+}
+
+.tab-size[data-tab-size="1"] {
+  -moz-tab-size: 1;
+  tab-size: 1;
+}
+
+.tab-size[data-tab-size="2"] {
+  -moz-tab-size: 2;
+  tab-size: 2;
+}
+
+.tab-size[data-tab-size="3"] {
+  -moz-tab-size: 3;
+  tab-size: 3;
+}
+
+.tab-size[data-tab-size="4"] {
+  -moz-tab-size: 4;
+  tab-size: 4;
+}
+
+.tab-size[data-tab-size="5"] {
+  -moz-tab-size: 5;
+  tab-size: 5;
+}
+
+.tab-size[data-tab-size="6"] {
+  -moz-tab-size: 6;
+  tab-size: 6;
+}
+
+.tab-size[data-tab-size="7"] {
+  -moz-tab-size: 7;
+  tab-size: 7;
+}
+
+.tab-size[data-tab-size="8"] {
+  -moz-tab-size: 8;
+  tab-size: 8;
+}
+
+.tab-size[data-tab-size="9"] {
+  -moz-tab-size: 9;
+  tab-size: 9;
+}
+
+.tab-size[data-tab-size="10"] {
+  -moz-tab-size: 10;
+  tab-size: 10;
+}
+
+.tab-size[data-tab-size="11"] {
+  -moz-tab-size: 11;
+  tab-size: 11;
+}
+
+.tab-size[data-tab-size="12"] {
+  -moz-tab-size: 12;
+  tab-size: 12;
+}
+
+.task-list-item {
+  list-style-type: none;
+}
+
+.task-list-item+.task-list-item {
+  margin-top: 3px;
+}
+
+.task-list-item input {
+  margin: 0 .2em .25em -1.6em;
+  vertical-align: middle;
+}
+
+a.pl {
+  color: #000000;
+  font-size: 75%;
+}
+
+a.pl:hover {
+  text-decoration: none;
+}
+
+table td em {
+  font-size: 85%;
+}
+
+div#sidebar {
+  float: left;
+  position: fixed;
+  top: 55px;
+  min-width: calc(50% - 460px);
+  font-size: 90%;
+  max-width: calc(50% - 460px);
+  overflow-y: scroll;
+  bottom: 0;
+}
+
+div#sidebar ul {
+  border-top: 2px solid #ccc;
+  padding: 0.5em;
+  list-style-type: none;
+  padding-left: 0.5em;
+  margin-bottom: 0px;
+}
+
+div#sidebar ul ul {
+  border-top: none;
+  list-style-type: circle;
+  padding-left: 1.0em;
+  margin-top: 0;
+  margin-bottom: 0;
+}
+
+div#sidebar ul ul ul {
+  border-top: none;
+  padding-left: 1.5em;
+  margin-top: 2px;
+  margin-bottom: 0;
+  list-style-type: disc;
+  padding-top: 0;
+  padding-bottom: 0;
+}
+
+div#sidebar details {
+  margin-bottom: 0px;
+}
+
+div#sidebar details summary {
+  list-style: none;
+}
+div#sidebar details summary::-webkit-details-marker {
+  display: none;
+}
+
+div#sidebar details summary::after {
+  content: " [+]";
+}
+
+div#sidebar details[open] summary::after {
+  content: " [-]";
+}
+
+div#sidebar details[open] details summary::after {
+  content: " [+]";
+}
+
+div#sidebar details[open] details[open] summary::after {
+  content: " [-]";
+}
+
+@media screen and (max-width: 1140px) {
+  div#sidebar {
+    display: none;
+  }
+}
+
+@media screen and (max-width: 1320px) and (min-width: 1140px) {
+  div#content {
+    margin-left: 200px;
+  }
+
+  div#sidebar {
+    max-width: 200px;
+  }
+}
+
+div#data-formats-div li {
+  margin-bottom: 1em;
+}
+
+li.bottom-space {
+  padding-bottom: 0.5em;
+  margin-bottom: 0.5em;
+  border-bottom: 1px solid #aaa;
+}
+
+a.textlink {
+  color: #333;
+}
+
+ul#binding_sidebar {
+  background: #eee;
+}
diff -pruN 4.3.0-2/doc/css/rouge-github-mod.css 4.4.0-1/doc/css/rouge-github-mod.css
--- 4.3.0-2/doc/css/rouge-github-mod.css	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/css/rouge-github-mod.css	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,71 @@
+.highlight table td { padding: 5px; }
+.highlight table pre { margin: 0; }
+.highlight .cm, span.cm { color: #3322ff; font-style: italic; }
+.highlight .cp, span.cp { color: #3322ff; font-weight: bold;  }
+.highlight .c1, span.c1 { color: #3322ff; font-style: italic; }
+.highlight .cs, span.cs { color: #3322ff; font-weight: bold; font-style: italic; }
+.highlight .c, .highlight .cd, span.c, span.cd { color: #3322ff; font-style: italic; }
+.highlight .err, span.err { color: #a61717; background-color: #e3d2d2; }
+.highlight .gd, span.gd { color: #000000; background-color: #ffdddd; }
+.highlight .ge, span.ge { color: #000000; font-style: italic; }
+.highlight .gr, span.gr { color: #aa0000; }
+.highlight .gh, span.gh { color: #999999; }
+.highlight .gi, span.gi { color: #000000; background-color: #ddffdd; }
+.highlight .go, span.go { color: #888888; }
+.highlight .gp, span.gp { color: #555555; }
+.highlight .gs, span.gs { font-weight: bold; }
+.highlight .gu, span.gu { color: #aaaaaa; }
+.highlight .gt, span.gt { color: #aa0000; }
+.highlight .kc, span.kc { color: #000000; font-weight: bold; }
+.highlight .kd, span.kd { color: #000000; font-weight: bold; }
+.highlight .kn, span.kn { color: #000000; font-weight: bold; }
+.highlight .kp, span.kp { color: #000000; font-weight: bold; }
+.highlight .kr, span.kr { color: #000000; font-weight: bold; }
+.highlight .kt, span.kt { color: #449955; font-weight: bold; }
+.highlight .k, .highlight .kv, span.k, span.kv { color: #000000; font-weight: bold; }
+.highlight .mf, span.mf { color: #009999; }
+.highlight .mh, span.mh { color: #009999; }
+.highlight .il, span.il { color: #009999; }
+.highlight .mi, span.mi { color: #009999; }
+.highlight .mo, span.mo { color: #009999; }
+.highlight .m, .highlight .mb, .highlight .mx, span.m, span.mb, span.mx { color: #009999; }
+.highlight .sb, span.sb { color: #d14; }
+.highlight .sc, span.sc { color: #d14; }
+.highlight .sd, span.sd { color: #d14; }
+.highlight .s2, span.s2 { color: #d14; }
+.highlight .se, span.se { color: #d14; }
+.highlight .sh, span.sh { color: #d14; }
+.highlight .si, span.si { color: #d14; }
+.highlight .sx, span.sx { color: #d14; }
+.highlight .sr, span.sr { color: #009926; }
+.highlight .s1, span.s1 { color: #d14; }
+.highlight .ss, span.ss { color: #990073; }
+.highlight .s, span.s { color: #d14; }
+.highlight .na, span.na { color: #008080; }
+.highlight .bp, span.bp { color: #999999; }
+.highlight .nb, span.nb { color: #0086B3; }
+.highlight .nc, span.nc { color: #449955; font-weight: bold; }
+.highlight .no, span.no { color: #008080; }
+.highlight .nd, span.nd { color: #3c5d5d; font-weight: bold; }
+.highlight .ni, span.ni { color: #800080; }
+.highlight .ne, span.ne { color: #990000; font-weight: bold; }
+.highlight .nf, span.nf { }
+.highlight .nl, span.nl { color: #990000; font-weight: bold; }
+.highlight .nn, span.nn { color: #555555; }
+.highlight .nt, span.nt { color: #000080; }
+.highlight .vc, span.vc { color: #008080; }
+.highlight .vg, span.vg { color: #008080; }
+.highlight .vi, span.vi { color: #008080; }
+.highlight .nv, span.nv { color: #008080; }
+.highlight .ow, span.ow { color: #000000; font-weight: bold; }
+.highlight .o, span.o { }
+.highlight .w, span.w { color: #bbbbbb; }
+.highlight {
+  background-color: #f8f8f8;
+  display: inline-block;
+  position: relative;
+  left: 50%;
+  transform: translateX(-50%);
+  max-width: 100%;
+  overflow: scroll;
+}
diff -pruN 4.3.0-2/doc/developer/bindings.md 4.4.0-1/doc/developer/bindings.md
--- 4.3.0-2/doc/developer/bindings.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/developer/bindings.md	2024-05-28 20:08:34.000000000 +0000
@@ -31,7 +31,7 @@ tend to avoid writing native C++ and ins
 most notably Python.
 
 In the case of Python, many projects will use tools like
-[SWIG](http://www.swig.org) to automatically generate bindings, or they might
+[SWIG](https://www.swig.org) to automatically generate bindings, or they might
 hand-write Cython.  The same types of strategies may be used for other
 languages; hand-written MEX files may be used for MATLAB, hand-written Rcpp
 bindings might be used for R bindings, and so forth.
@@ -777,8 +777,8 @@ PARAM_MODEL_OUT(LinearRegression, "outpu
     "linear regression output model.", "M");
 ```
 
-Note that even the parameter documentation strings must be a little be agnostic
-to the binding type, because the command-line interface is so different than the
+Note that even the parameter documentation strings must be a little agnostic
+to the binding type, because the command-line interface is so different from the
 Python interface to the user.
 
 ### Using `Params` in a `BINDING_FUNCTION()` function
@@ -815,7 +815,7 @@ To access a string that a user passed in
 following code could be used:
 
 ```c++
-const std::string& str = params.Has<std::string>("string");
+const std::string& str = params.Get<std::string>("string");
 ```
 
 Matrix types are accessed in the same way:
@@ -943,7 +943,7 @@ populating them with the correct options
 `BINDING_FUNCTION()` with those instantiated objects.
 
 In order to do this, each parameter and the program documentation must make
-themselves known to the IO singleton.  This is accomplished by having the @c
+themselves known to the IO singleton.  This is accomplished by having the
 `BINDING_USER_NAME()`, `BINDING_SHORT_DESC()`, `BINDING_LONG_DESC()`,
 `BINDING_EXAMPLE()`, `BINDING_SEE_ALSO()` and `PARAM_*()` macros declare global
 variables that, in their constructors, register themselves with the `IO`
diff -pruN 4.3.0-2/doc/developer/community.md 4.4.0-1/doc/developer/community.md
--- 4.3.0-2/doc/developer/community.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/developer/community.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,93 @@
+# Community
+
+mlpack is a community-led effort, and so the code is not possible without the
+community.  Since mlpack is an open-source project, anyone is welcome to become
+a part of the community and contribute.  There is no need to be a a machine
+learning expert to participate; often, there are many tasks to be done that
+don't require in-depth knowledge.
+
+Over the past several years, mlpack has participated in
+[Google Summer of Code](https://summerofcode.withgoogle.com/).  For more
+information, see [this page](gsoc.md).
+
+All mlpack development is done on [GitHub](https://github.com/mlpack/mlpack).
+Commits and issue comments can be tracked via the
+[mlpack-git](https://freelists.org/list/mlpack-git) list (graciously hosted by
+[FreeLists](https://freelists.org).  Communication is generally either via
+issues on GitHub, or via chat:
+
+## Real-time chat
+
+ * #mlpack:matrix.org on [Matrix](https://www.matrix.org/)
+ * [mlpack Slack workspace](https://mlpack.slack.com/)
+   - You will need to request an invite from the
+     [auto-inviter](http://slack-inviter.mlpack.org:4000).
+   - The Slack workspace is a bridged version of the Matrix room and it is
+     generally a better experience to use Matrix directly.
+
+## Video meetup
+
+On the first and third Friday of every month, at ***1700 UTC*** on Fridays, we
+have *casual video meetups* with no particular agenda.  Feel free to join up!
+We often talk about code changes that we are working on, issues that people are
+having with mlpack, general design direction, and whatever else might be on our
+mind.
+
+We use [this Zoom room](https://zoom.us/j/3820896170).  For security, we use a
+password for the meeting to keep malicious bots out.  The password is simple:
+it's just the name of the library (in all lowercase).
+
+## Getting involved
+
+Everyone is welcome to contribute to mlpack.  But before becoming a contributor,
+it's often useful to understand mlpack as a user.  So, a good place to start is
+to:
+
+ - [download mlpack](https://www.mlpack.org/download.html)
+ - use it in C++ (see [the documentation](../index.md))
+ - use the bindings to other languages to perform machine learning tasks:
+    [Python](../user/bindings/python.md), [Julia](../user/bindings/julia.md),
+    [R](../user/bindings/r.md), [Go](../user/bindings/go.md), and the
+    [command-line programs](../user/bindings/cli.md)
+
+There is also the [examples repository](https://github.com/mlpack/examples) that
+contains many examples you can build and play around with.
+
+Once you have an idea of what's included in mlpack and how a user might use it,
+then a good next step would be to set up a development environment.  Once you
+have that set up, you can
+[build mlpack from source](../README.md#3-installing-and-using-mlpack-in-c),
+and explore the codebase to see how it's organized.
+
+Try making small changes to the code, or adding new tests to the test suite, and
+then rebuild to see how your changes work.
+
+Now you're set up to contribute!  There are lots of ways you can contribute.
+Here are a couple ideas:
+
+ * Help others figure out their mlpack issues and questions.
+   [Here](https://github.com/mlpack/mlpack/issues?q=is%3Aopen+is%3Aissue+label%3A%22t%3A+question%22)
+   is a list of GitHub issues tagged `question`.  Helping others figure out
+   their problems is really one of the best ways to learn about the library.
+
+ * Read through the [vision document](https://www.mlpack.org/papers/vision.pdf)
+   to learn about the development goals of the mlpack community and see the
+   high-level tasks that need to be done to accomplish that vision.
+
+ * Find an issue that needs implementation help;
+   [here](https://github.com/mlpack/mlpack/issues) is the list of issues.
+
+ * Find an abandoned pull request;
+   [here](https://github.com/mlpack/mlpack/pulls?q=is%3Aclosed+is%3Apr+label%3A%22s%3A+stale%22)
+   is a list of pull requests that were closed for inactivity.  Often these have
+   comments that need to be addressed, but the original author didn't have time
+   to finish the work.  So, you can pick up where they left off!
+
+ * Implement a new machine learning algorithm that mlpack doesn't currently
+   have.
+
+ * Take a look at the ideas on the Google Summer of Code
+   [Ideas List](https://github.com/mlpack/mlpack/wiki/SummerOfCOdeIdeas) and see
+   if you find any of them interesting or exciting.  Even if you're not planning
+   to apply for Summer of Code, it's okay to take these ideas and implement them
+   separately.
diff -pruN 4.3.0-2/doc/developer/gsoc.md 4.4.0-1/doc/developer/gsoc.md
--- 4.3.0-2/doc/developer/gsoc.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/developer/gsoc.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,80 @@
+# mlpack and Google Summer of Code
+
+mlpack is a proud participant in Google Summer of Code. We have been a part of
+the program since 2013, with over 50 students accepted over the years for mlpack
+(all of whom succeeded in the program).  Each year, we receive very many
+applications and it is a competitive process, so this page exists to help you
+determine if you could be a strong candidate.
+
+## How to be a good candidate
+
+The two most important qualities that an mlpack GSoC candidate can possess is
+the ability to be self-sufficient and the willingness to learn.
+
+Self-sufficiency is key because mentors have limited time and can’t put in as
+much time helping the student as the student is putting in. However, this of
+course does not mean that a student (or a prospective student) can never ask any
+questions! mlpack is a complex library and can sometimes take help and
+explanation to understand.
+
+A willingness to learn is also important because virtually every potential
+project will require the student to become familiar with a new algorithm or C++
+technique. mlpack is a complex library with many components and it is likely
+during your project that you might have to use or interact with some other part
+of the codebase.
+
+## Things to be familiar with
+
+Of course, these are not the only important ingredients for a successful GSoC
+project. A student should ideally be familiar with
+
+ - *open source software development*: opening pull requests, using git, opening
+   issues. mlpack uses Github, which has great documentation. You can learn
+   about the workflow using [this link](https://docs.github.com/en), if you are
+   not already familiar.
+
+ - *using the development toolchain on your computer*: you should be able to
+   download and compile mlpack, make changes to the code, and recompile with the
+   new changes. There is a
+   [section in the README](../README.md#3-installing-and-using-mlpack-in-c)
+   for how to build mlpack and would be a great place to get started. If you’re
+   on Windows, then the [Windows build guide](../user/build_windows.md) could be
+   very useful. See also the [Community page](community.md) for more information
+   on getting started.
+
+ - *at least intermediate C++ knowledge*: mlpack uses lots of different C++
+   paradigms including template metaprogramming, C++ features like rvalue
+   references, and other strategies, in order to make the code fast. You should
+   be at least familiar with some of these language features and what templates
+   are, even if you have not used them in-depth, so that you can understand the
+   mlpack codebase. Some examples of patterns that are often used inside of
+   mlpack are SFINAE ([example in mlpack, see std::enable_if usages](https://github.com/mlpack/mlpack/blob/565cfd3aad22deec0656b86e801052593a937723/src/mlpack/methods/mean_shift/mean_shift.hpp)),
+   [policy-based design](https://www.drdobbs.com/policy-based-design-in-the-real-world/184401861),
+   and [compile-time class traits](https://accu.org/index.php/journals/442).
+   Here are some [other useful resources](https://www.codeproject.com/Articles/3743/A-gentle-introduction-to-Template-Metaprogramming)
+   for learning template metaprogramming, and some useful
+   [reference books](https://www.amazon.com/gp/product/0201749629?ie=UTF8tag=aristeia.com-20linkCode=as2camp=1789creative=9325creativeASIN=0201749629).
+   If some of this sounds new to you, don’t feel overwhelmed; it’s not a
+   necessity, but it is helpful. You should at least be prepared to learn about
+   it!
+
+ - *project-specific knowledge*: you’ll need to have an in-depth understanding
+   of the specific project that you choose. If you’re not sure what project you
+   want to work on, see the
+   [SummerOfCodeIdeas](https://github.com/mlpack/mlpack/wiki/SummerOfCodeIdeas)
+   wiki page, which has ideas for GSoC projects that you might find interesting.
+   But you aren’t required to do one of those projects; if you have another
+   interesting idea, propose it and see if a mentor is interested in supervising
+   the project! If you have any questions about a project, be aware that it’s
+   possible that the question has already been answered on Github. Take a look
+   through the closed issues or search to see if there’s already an answer to
+   your question.  Get involved!
+
+For a strong proposal, it’s important to be a part of the mlpack community, via
+contributions, code reviews, helping others solve their issues, and so forth.
+There are some pointers on the [community page](community.md) on how to start
+contributing and get involved.
+
+When it comes time to write your proposal, we do have a
+[proposal guide](https://github.com/mlpack/mlpack/wiki/Google-Summer-of-Code-Application-Guide)
+that you can take a look at to guide your application.
diff -pruN 4.3.0-2/doc/developer/trees.md 4.4.0-1/doc/developer/trees.md
--- 4.3.0-2/doc/developer/trees.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/developer/trees.md	2024-05-28 20:08:34.000000000 +0000
@@ -378,8 +378,8 @@ This section is divided into five parts,
 An earlier section discussed the three different template parameters that are
 required by the `TreeType` policy.
 
-The [MetricType policy](metrictype.md) provides one method that will be useful
-for tree building and other operations:
+The [MetricType policy](metrics.md) provides one method that will be useful for
+tree building and other operations:
 
 ```c++
 // This function is required by the MetricType policy.
diff -pruN 4.3.0-2/doc/developer/version.md 4.4.0-1/doc/developer/version.md
--- 4.3.0-2/doc/developer/version.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/developer/version.md	1970-01-01 00:00:00.000000000 +0000
@@ -1,25 +0,0 @@
-# mlpack versions in code
-
-mlpack provides a couple of convenience macros and functions to get the version
-of mlpack.  More information (and straightforward code) can be found in
-`src/mlpack/core/util/version.hpp`.
-
-The following three macros provide major, minor, and patch versions of mlpack
-(i.e. for `mlpack-x.y.z`, `x` is the major version, `y` is the minor version,
-and `z` is the patch version):
-
-```c++
-MLPACK_VERSION_MAJOR
-MLPACK_VERSION_MINOR
-MLPACK_VERSION_PATCH
-```
-
-In addition, the function `mlpack::util::GetVersion()` returns the mlpack
-version as a string (for instance, `"mlpack 1.0.8"`).
-
-## mlpack command-line program versions
-
-Each mlpack command-line program supports the `--version` (or `-V`) option,
-which will print the version of mlpack used.  If the version is not an official
-release but instead from git, the version will be `mlpack git` (and will have a
-git revision SHA appended to `git`).
diff -pruN 4.3.0-2/doc/examples/sample-ml-app/README.txt 4.4.0-1/doc/examples/sample-ml-app/README.txt
--- 4.3.0-2/doc/examples/sample-ml-app/README.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/examples/sample-ml-app/README.txt	2024-05-28 20:08:34.000000000 +0000
@@ -21,15 +21,15 @@ and library files into C:\mlpack\.
  * ensmallen: https://ensmallen.org/files/ensmallen-2.19.0.tar.gz
    Download the .tar.gz, and extract it into C:\mlpack\ensmallen-2.19.0\
 
-Now, install mlpack into C:\mlpack\mlpack-4.3.0\.  If you downloaded the mlpack
+Now, install mlpack into C:\mlpack\mlpack-4.4.0\.  If you downloaded the mlpack
 source, you can either use the Windows build guide (see
 doc/user/build_windows.md) to build and install, or, since mlpack is
-header-only, copy the src/ directory to C:\mlpack\mlpack-4.3.0\ and rename it
-"include" (so there will now be a directory C:\mlpack\mlpack-4.3.0\include\,
+header-only, copy the src/ directory to C:\mlpack\mlpack-4.4.0\ and rename it
+"include" (so there will now be a directory C:\mlpack\mlpack-4.4.0\include\,
 which contains only base.hpp and the mlpack/ subdirectory).
 
 Alternately, if you downloaded the Windows MSI installer, you can install to
-C:\mlpack\mlpack-4.3.0\.
+C:\mlpack\mlpack-4.4.0\.
 
 Once all of that setup is done, the example should compile as-is.
 
diff -pruN 4.3.0-2/doc/examples/sample-ml-app/sample-ml-app/sample-ml-app.vcxproj 4.4.0-1/doc/examples/sample-ml-app/sample-ml-app/sample-ml-app.vcxproj
--- 4.3.0-2/doc/examples/sample-ml-app/sample-ml-app/sample-ml-app.vcxproj	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/examples/sample-ml-app/sample-ml-app/sample-ml-app.vcxproj	2024-05-28 20:08:34.000000000 +0000
@@ -104,7 +104,7 @@
       <SDLCheck>true</SDLCheck>
       <PreprocessorDefinitions>_DEBUG;_CONSOLE;%(PreprocessorDefinitions)</PreprocessorDefinitions>
       <ConformanceMode>Default</ConformanceMode>
-      <AdditionalIncludeDirectories>C:\mlpack\armadillo-11.4.1\include;C:\mlpack\mlpack-4.3.0\include\;C:\mlpack\cereal-1.3.2\include;C:\mlpack\ensmallen-2.19.0\include\%(AdditionalIncludeDirectories)</AdditionalIncludeDirectories>
+      <AdditionalIncludeDirectories>C:\mlpack\armadillo-11.4.1\include;C:\mlpack\mlpack-4.4.0\include\;C:\mlpack\cereal-1.3.2\include;C:\mlpack\ensmallen-2.19.0\include\%(AdditionalIncludeDirectories)</AdditionalIncludeDirectories>
       <LanguageStandard>stdcpp17</LanguageStandard>
       <OpenMPSupport>false</OpenMPSupport>
       <AdditionalOptions>%(AdditionalOptions)</AdditionalOptions>
Binary files 4.3.0-2/doc/img/favicons-blocks.png and 4.4.0-1/doc/img/favicons-blocks.png differ
Binary files 4.3.0-2/doc/img/favicons-matrix.png and 4.4.0-1/doc/img/favicons-matrix.png differ
diff -pruN 4.3.0-2/doc/index.md 4.4.0-1/doc/index.md
--- 4.3.0-2/doc/index.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/index.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,189 @@
+# Documentation for mlpack
+
+<!--
+This file contains the landing page for mlpack documentation.  Note that if you
+change any section headers, or add any new algorithms, the sidebar in
+sidebar.html will need to be manually modified!
+-->
+
+## A fast, flexible machine learning library
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning
+library with bindings to other languages.  It aims to provide fast, lightweight
+implementations of both common and cutting-edge machine learning algorithms.
+
+mlpack's lightweight C++ implementation makes it ideal for deployment, and it
+can also be used for interactive prototyping via C++ notebooks (these can be
+seen in action on mlpack's [homepage](https://www.mlpack.org/)).
+
+In addition to its [powerful C++ interface](quickstart/cpp.md), mlpack also
+provides [command-line programs](quickstart/cli.md), and bindings to the
+[Python](quickstart/python.md), [R](quickstart/r.md),
+[Julia](quickstart/julia.md), and [Go](quickstart/go.md) languages.
+
+_If you use mlpack, please [cite the software](citation.md)._
+
+## mlpack basics
+
+Installing mlpack can be done using the
+[instructions in the README](README.md#3-installing-and-using-mlpack-in-c);
+or the [Windows build guide](user/build_windows.md).
+The following basic guides are *highly recommended* before using mlpack.
+
+ * ***First steps***:
+   - [mlpack C++ quickstart](quickstart/cpp.md): create a couple simple C++
+     programs that use mlpack
+   - [Sample Windows mlpack C++ application](user/sample_ml_app.md): create a
+     working mlpack Windows program using Visual Studio
+
+ * ***Basics of matrices and data in mlpack***:
+   - [Matrices and data in mlpack](user/matrices.md)
+   - [Loading and saving mlpack objects](user/load_save.md)
+
+ * ***Reference for mlpack core classes***:
+   - [Core mlpack documentation](user/core.md)
+
+ * ***Using mlpack natively with our extensions in Python, R, CLI, Julia, and Go***:
+   - [Links to quickstarts and references](#bindings-to-other-languages)
+
+## mlpack algorithm documentation
+
+Documentation for each machine learning algorithm that mlpack implements is
+detailed in the sections below.
+
+ * [Classification algorithms](#classification-algorithms): classify points as
+   discrete labels (`0`, `1`, `2`, ...).
+ * [Regression algorithms](#regression-algorithms): predict continuous values.
+ * [Clustering algorithms](#clustering-algorithms): group points into clusters.
+ * [Geometric algorithms](#geometric-algorithms): computations based on distance
+   metrics (nearest neighbors, kernel density estimation, etc.).
+ * [Preprocessing utilities](#preprocessing-utilities): prepare data for machine
+   learning algorithms.
+ * [Transformations](#transformations): transform data from one space to
+   another (principal components analysis, etc.).
+ * [Modeling utilities](#modeling-utilities): cross-validation, hyperparameter
+   tuning, etc.
+
+### Classification algorithms
+
+Classify points as discrete labels (`0`, `1`, `2`, ...).
+
+ * [`AdaBoost`](user/methods/adaboost.md): Adaptive Boosting
+ * [`DecisionTree`](user/methods/decision_tree.md): ID3-style decision tree
+   classifier
+ * [`HoeffdingTree`](user/methods/hoeffding_tree.md): streaming/incremental
+   decision tree classifier
+ * [`LinearSVM`](user/methods/linear_svm.md): simple linear support vector
+   machine classifier
+ * [`LogisticRegression`](user/methods/logistic_regression.md): L2-regularized
+   logistic regression (two-class only)
+ * [`NaiveBayesClassifier`](user/methods/naive_bayes_classifier.md): simple
+   multi-class naive Bayes classifier
+ * [`Perceptron`](user/methods/perceptron.md): simple Perceptron classifier
+ * [`RandomForest`](user/methods/random_forest.md): parallelized random forest
+   classifier
+ * [`SoftmaxRegression`](user/methods/softmax_regression.md): L2-regularized
+   softmax regression (i.e. multi-class logistic regression)
+
+### Regression algorithms
+
+Predict continuous values.
+
+ * [`BayesianLinearRegression`](user/methods/bayesian_linear_regression.md):
+   Bayesian L2-penalized linear regression
+ * [`DecisionTreeRegressor`](user/methods/decision_tree_regressor.md): ID3-style
+   decision tree regressor
+ * [`LARS`](user/methods/lars.md): Least Angle Regression (LARS), L1-regularized
+   and L2-regularized
+ * [`LinearRegression`](user/methods/linear_regression.md): L2-regularized
+   linear regression (ridge regression)
+
+### Clustering algorithms
+
+Group points into clusters.
+
+<!-- TODO: add some -->
+
+### Geometric algorithms
+
+Computations based on distance metrics.
+
+<!-- TODO: add some -->
+
+### Preprocessing utilities
+
+Prepare data for machine learning algorithms.
+
+<!-- TODO: add some -->
+
+### Transformations
+
+Transform data from one space to another.
+
+ * [`LocalCoordinateCoding`](user/methods/local_coordinate_coding.md): local
+   coordinate coding with dictionary learning
+ * [`NMF`](user/methods/nmf.md): non-negative matrix factorization
+ * [`PCA`](user/methods/pca.md): principal components analysis
+ * [`SparseCoding`](user/methods/sparse_coding.md): sparse coding with
+   dictionary learning
+
+### Modeling utilities
+
+Tools for assembling a full data science pipeline.
+
+ * [Cross-validation](user/cv.md): k-fold cross-validation tools for any mlpack
+   algorithm
+ * [Hyperparameter tuning](user/hpt.md): generic hyperparameter tuner to find
+   good hyperparameters for any mlpack algorithm
+
+## Bindings to other languages
+
+mlpack's bindings to other languages have less complete functionality than
+mlpack in C++, but almost all the same algorithms are available.
+
+| ***Python*** | -- | [quickstart](quickstart/python.md) | -- | [reference](user/bindings/python.md) |
+| ***Julia*** | -- | [quickstart](quickstart/julia.md) | -- | [reference](user/bindings/julia.md) |
+| ***R*** | -- | [quickstart](quickstart/r.md) | -- | [reference](user/bindings/r.md)
+| ***Command-line programs*** | -- | [quickstart](quickstart/cli.md) | -- | [reference](user/bindings/cli.md) |
+| ***Go*** | -- | [quickstart](quickstart/go.md) | -- | [reference](user/bindings/go.md) |
+
+## Examples and further documentation
+
+ * [mlpack examples repository](https://github.com/mlpack/examples/): numerous
+   fully-working example applications of mlpack, in C++ and other languages.
+ * [mlpack models repository](https://github.com/mlpack/models/): complex models
+   in C++ built with mlpack
+
+For additional documentation beyond what is covered in all the resources above,
+the source code should be consulted.  Each method is fully documented.
+
+## Developer documentation
+
+The following general documentation can be useful if you are interested in
+contributing to mlpack:
+
+ * [The mlpack community](developer/community.md)
+ * [mlpack and Google Summer of Code](developer/gsoc.md)
+
+Throughout the codebase, mlpack uses some common template parameter policies.
+These are documented below.
+
+ * [The `ElemType` policy](developer/elemtype.md): element types for data
+ * [The `MetricType` policy](developer/metrics.md): distance metrics
+ * [The `KernelType` policy](developer/kernels.md): kernel functions
+ * [The `TreeType` policy](developer/trees.md): space trees (ball trees,
+   KD-trees, etc.)
+
+In addition, the following documentation may be useful when developing bindings
+for other languages:
+
+ * [Timers](developer/timer.md): timing parts of bindings
+ * [Writing an mlpack binding](developer/iodoc.md): simple examples of mlpack
+   bindings
+ * [Automatic bindings](developer/bindings.md): details on mlpack's automatic
+   binding generator system.
+
+## Changelog
+
+For a list of changes in each version of mlpack, see the
+[changelog](HISTORY.md).
diff -pruN 4.3.0-2/doc/quickstart/R.md 4.4.0-1/doc/quickstart/R.md
--- 4.3.0-2/doc/quickstart/R.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/R.md	1970-01-01 00:00:00.000000000 +0000
@@ -1,162 +0,0 @@
-# mlpack in R quickstart guide
-
-This page describes how you can quickly get started using mlpack from R and
-gives a few examples of usage, and pointers to deeper documentation.
-
-This quickstart guide is also available for [C++](cpp.md), [Python](python.md),
-[Julia](julia.md), [the command line](cli.md), and [Go](go.md).
-
-## Installing mlpack
-
-Installing the mlpack bindings for R is straightforward; you can just use
-CRAN:
-
-```r
-install.packages('mlpack')
-```
-
-Building the R bindings from scratch is a little more in-depth, though.  For
-information on that, follow the instructions in the
-[main README](../../README.md).
-
-## Simple mlpack quickstart example
-
-As a really simple example of how to use mlpack from R, let's do some
-simple classification on a subset of the standard machine learning `covertype`
-dataset.  We'll first split the dataset into a training set and a testing set,
-then we'll train an mlpack random forest on the training data, and finally we'll
-print the accuracy of the random forest on the test dataset.
-
-You can copy-paste this code directly into R to run it.
-
-```r
-if(!requireNamespace("data.table", quietly = TRUE)) { install.packages("data.table") }
-suppressMessages({
-    library("mlpack")
-    library("data.table")
-})
-
-# Load the dataset from an online URL.  Replace with 'covertype.csv.gz' if you
-# want to use on the full dataset.
-df <- fread("https://www.mlpack.org/datasets/covertype-small.csv.gz")
-
-# Split the labels.
-labels <- df[, .(label)]
-dataset <- df[, label:=NULL]
-
-# Split the dataset using mlpack.
-prepdata <- preprocess_split(input = dataset,
-                             input_labels = labels,
-                             test_ratio = 0.3,
-                             verbose = TRUE)
-
-# Train a random forest.
-output <- random_forest(training = prepdata$training,
-                        labels = prepdata$training_labels,
-                        print_training_accuracy = TRUE,
-                        num_trees = 10,
-                        minimum_leaf_size = 3,
-                        verbose = TRUE)
-rf_model <- output$output_model
-
-# Predict the labels of the test points.
-output <- random_forest(input_model = rf_model,
-                        test = prepdata$test,
-                        verbose = TRUE)
-
-# Now print the accuracy.  The third return value ('probabilities'), which we
-# ignored here, could also be used to generate an ROC curve.
-correct <- sum(output$predictions == prepdata$test_labels)
-cat(correct, "out of", length(prepdata$test_labels), "test points correct",
-    correct / length(prepdata$test_labels) * 100.0, "%\n")
-```
-
-We can see that we achieve reasonably good accuracy on the test dataset (80%+);
-if we use the full `covertype.csv.gz`, the accuracy should increase
-significantly (but training will take longer).
-
-It's easy to modify the code above to do more complex things, or to use
-different mlpack learners, or to interface with other machine learning toolkits.
-
-## Using mlpack for movie recommendations
-
-In this example, we'll train a collaborative filtering model using mlpack's
-[`cf()`](https://www.mlpack.org/doc/stable/r_documentation.html#cf) method.
-We'll train this on the
-[MovieLens dataset](https://grouplens.org/datasets/movielens/), and then we'll
-use the model that we train to give recommendations.
-
-You can copy-paste this code directly into R to run it.
-
-```r
-if(!requireNamespace("data.table", quietly = TRUE)) { install.packages("data.table") }
-suppressMessages({
-    library("mlpack")
-    library("data.table")
-})
-
-# First, load the MovieLens dataset.  This is taken from files.grouplens.org/
-# but reposted on mlpack.org as unpacked and slightly preprocessed data.
-ratings <- fread("http://www.mlpack.org/datasets/ml-20m/ratings-only.csv.gz")
-movies <- fread("http://www.mlpack.org/datasets/ml-20m/movies.csv.gz")
-
-# Hold out 10% of the dataset into a test set so we can evaluate performance.
-predata <- preprocess_split(input = ratings,
-                            test_ratio = 0.1,
-                            verbose = TRUE)
-
-# Train the model.  Change the rank to increase/decrease the complexity of the
-# model.
-output <- cf(training = predata$training,
-             test = predata$test,
-             rank = 10,
-             verbose = TRUE,
-             max_iteration=2,
-             algorithm = "RegSVD")
-cf_model <- output$output_model
-
-# Now query the 5 top movies for user 1.
-output <- cf(input_model = cf_model,
-             query = matrix(1),
-             recommendations = 10,
-             verbose = TRUE)
-
-# Get the names of the movies for user 1.
-cat("Recommendations for user 1:\n")
-for (i in 1:10) {
-  cat("  ", i, ":", as.character(movies[output$output[i], 3]), "\n")
-}
-```
-
-Here is some example output, showing that user 1 seems to have good taste in
-movies:
-
-```
-Recommendations for user 1:
-  0: Casablanca (1942)
-  1: Pan's Labyrinth (Laberinto del fauno, El) (2006)
-  2: Godfather, The (1972)
-  3: Answer This! (2010)
-  4: Life Is Beautiful (La Vita è bella) (1997)
-  5: Adventures of Tintin, The (2011)
-  6: Dark Knight, The (2008)
-  7: Out for Justice (1991)
-  8: Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb (1964)
-  9: Schindler's List (1993)
-```
-
-## Next steps with mlpack
-
-After working through this overview to `mlpack`'s R package, we hope you are
-inspired to use `mlpack`' in your data science workflow.  However, the two
-examples above have only shown a little bit of the functionality of mlpack.
-Lots of other functions are available with different functionality.  A full list
-of each of these functions and full documentation can be found on the following
-page:
-
- - [R documentation](https://www.mlpack.org/doc/stable/r_documentation.html)
-
-Also, mlpack is much more flexible from C++ and allows much greater
-functionality.  So, more complicated tasks are possible if you are willing to
-write C++ (or perhaps Rcpp).  To get started learning about mlpack in C++, a
-good starting point is the [C++ quickstart guide](cpp.md).
diff -pruN 4.3.0-2/doc/quickstart/cli.md 4.4.0-1/doc/quickstart/cli.md
--- 4.3.0-2/doc/quickstart/cli.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/cli.md	2024-05-28 20:08:34.000000000 +0000
@@ -5,7 +5,7 @@ command-line and gives a few examples of
 documentation.
 
 This quickstart guide is also available for [C++](cpp.md), [Python](python.md),
-[R](R.md), [Julia](julia.md), and [Go](go.md).
+[R](r.md), [Julia](julia.md), and [Go](go.md).
 
 ## Installing mlpack
 
@@ -164,6 +164,7 @@ Out for Justice (1991)
 Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb (1964)
 Schindler's List (1993)
 ```
+
 ## Next steps with mlpack
 
 Now that you have done some simple work with mlpack, you have seen how it can
diff -pruN 4.3.0-2/doc/quickstart/cpp.md 4.4.0-1/doc/quickstart/cpp.md
--- 4.3.0-2/doc/quickstart/cpp.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/cpp.md	2024-05-28 20:08:34.000000000 +0000
@@ -6,18 +6,19 @@ gives a few examples of usage, and point
 Keep in mind that mlpack also has interfaces to other languages, and quickstart
 guides for those other languages are available too.  If that is what you are
 looking for, see the quickstarts for [Python](python.md),
-[the command line](cli.md), [Julia](julia.md), [R](R.md), or [Go](go.md).
+[the command line](cli.md), [Julia](julia.md), [R](r.md), or [Go](go.md).
 
 ## Installing mlpack
 
 To use mlpack in C++, you only need the header files associated with the
 libraries, and the dependencies Armadillo and ensmallen (detailed in the
-[main README](../../README.md)).  The headers may already be pre-packaged for
-your distribution; for instance, for Ubuntu and Debian you can simply run the
-command
+[main README](../../README.md)).
+
+The headers may already be pre-packaged for your distribution;
+for instance, for Ubuntu and Debian you can simply run the command
 
 ```sh
-sudo apt-get install mlpack-dev
+sudo apt-get install libmlpack-dev
 ```
 
 and on Fedora or Red Hat:
@@ -26,18 +27,23 @@ and on Fedora or Red Hat:
 sudo dnf install mlpack
 ```
 
-If you run a different distribution, mlpack may be packaged under a different
-name.  And if it is not packaged, you can use a Docker image from Dockerhub:
+You can also use a Docker image from Dockerhub, 
+which has mlpack headers already installed:
 
 ```sh
 docker run -it mlpack/mlpack /bin/bash
 ```
 
-This Docker image has mlpack headers already installed.
-
 If you prefer to build mlpack from scratch, see the
 [main README](../../README.md).
 
+**Note for Ubuntu LTS Users**: The libmlpack-dev version in the Ubuntu LTS 
+repositories may not always be the latest. This can lead to issues, such as
+missing header files (e.g., `mlpack.hpp` missing in versions prior to 4.0).
+To ensure compatibility with the latest mlpack features and examples,
+we recommend building mlpack from source, as explained in the
+[main README](../../README.md).
+
 ## Installing mlpack from vcpkg
 
 The mlpack port in vcpkg is kept up to date by Microsoft team members and community contributors. The url of vcpkg is: https://github.com/Microsoft/vcpkg . You can download and install mlpack using the vcpkg dependency manager:
@@ -52,6 +58,20 @@ cd vcpkg
 
 If the version is out of date, please [create an issue or pull request](https://github.com/Microsoft/vcpkg) on the vcpkg repository.
 
+## Installing mlpack from Conan
+
+The mlpack recipe in [Conan](https://conan.io/) is kept up to date by the Conan team members and the community contributors.
+
+Follow the instructions on [this page on how to set up Conan](https://conan.io/downloads).
+
+Install mlpack:
+
+```shell
+conan install --requires="mlpack/[*]" --build=missing
+```
+
+If the version is outdated or there is a new release version, please [create an issue or pull request](https://github.com/conan-io/conan-center-index) on the conan-center-index repository.
+
 ## Simple quickstart example
 
 As a really simple example of how to use mlpack in C++, let's do some simple
@@ -132,7 +152,7 @@ example command that uses `g++`, and ass
 `cpp_quickstart_1.cpp`.
 
 ```sh
-g++ -O3 -std=c++14 -o cpp_quickstart_1 cpp_quickstart_1.cpp -larmadillo -fopenmp
+g++ -O3 -std=c++17 -o cpp_quickstart_1 cpp_quickstart_1.cpp -larmadillo -fopenmp
 ```
 
 Then, you can run the program easily:
@@ -239,7 +259,7 @@ This can be compiled the same way as bef
 `cpp_quickstart_2.cpp`:
 
 ```sh
-g++ -O3 -std=c++14 -o cpp_quickstart_2 cpp_quickstart_2.cpp -fopenmp -larmadillo
+g++ -O3 -std=c++17 -o cpp_quickstart_2 cpp_quickstart_2.cpp -fopenmp -larmadillo
 ```
 
 And then it can be easily run:
diff -pruN 4.3.0-2/doc/quickstart/go.md 4.4.0-1/doc/quickstart/go.md
--- 4.3.0-2/doc/quickstart/go.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/go.md	2024-05-28 20:08:34.000000000 +0000
@@ -4,7 +4,7 @@ This page describes how you can quickly
 gives a few examples of usage, and pointers to deeper documentation.
 
 This quickstart guide is also available for [C++](cpp.md), [Python](python.md),
-[Julia](julia.md), [the command line](cli.md), and [R](R.md).
+[Julia](julia.md), [the command line](cli.md), and [R](r.md).
 
 ## Installing mlpack
 
@@ -100,7 +100,7 @@ different mlpack learners, or to interfa
 ## Using mlpack for movie recommendations
 
 In this example, we'll train a collaborative filtering model using mlpack's
-[`cf()`](https://www.mlpack.org/doc/stable/go_documentation.html#cf) method.
+[`cf()`](../user/bindings/go.md#cf) method.
 We'll train this on the
 [MovieLens dataset](https://grouplens.org/datasets/movielens/), and then we'll
 use the model that we train to give recommendations.
@@ -190,7 +190,7 @@ only shown a little bit of the functiona
 are available with different functionality.  A full list of each of these
 methods and full documentation can be found on the following page:
 
- - [mlpack Go binding documentation](https://www.mlpack.org/doc/stable/go_documentation.html)
+ - [mlpack Go binding documentation](../user/bindings/go.md)
 
 You can also use GoDoc to explore the `mlpack` module and its functions; every
 function comes with comprehensive documentation.
diff -pruN 4.3.0-2/doc/quickstart/julia.md 4.4.0-1/doc/quickstart/julia.md
--- 4.3.0-2/doc/quickstart/julia.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/julia.md	2024-05-28 20:08:34.000000000 +0000
@@ -4,7 +4,7 @@ This page describes how you can quickly
 gives a few examples of usage, and pointers to deeper documentation.
 
 This quickstart guide is also available for [C++](cpp.md), [Python](python.md),
-[the command line](cli.md), [R](R.md), and [Go](go.md).
+[the command line](cli.md), [R](r.md), and [Go](go.md).
 
 ## Installing mlpack
 
@@ -81,7 +81,7 @@ different mlpack learners, or to interfa
 ## Using mlpack for movie recommendations
 
 In this example, we'll train a collaborative filtering model using mlpack's
-[`cf()`](https://www.mlpack.org/doc/stable/julia_documentation.html#cf) method.
+[`cf()`](../user/bindings/julia.md#cf) method.
 We'll train this on the
 [MovieLens dataset](https://grouplens.org/datasets/movielens/), and then we'll
 use the model that we train to give recommendations.
@@ -151,7 +151,7 @@ have only shown a little bit of the func
 functions are available with different functionality.  A full list of each of
 these commands and full documentation can be found on the following page:
 
- - [Julia documentation](https://www.mlpack.org/doc/stable/julia_documentation.html)
+ - [Julia documentation](../user/bindings/julia.md)
 
 You can also use the Julia REPL to explore the `mlpack` module and its
 functions; every function comes with comprehensive documentation.
diff -pruN 4.3.0-2/doc/quickstart/python.md 4.4.0-1/doc/quickstart/python.md
--- 4.3.0-2/doc/quickstart/python.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/quickstart/python.md	2024-05-28 20:08:34.000000000 +0000
@@ -4,7 +4,7 @@ This page describes how you can quickly
 gives a few examples of usage, and pointers to deeper documentation.
 
 This quickstart guide is also available for [C++](cpp.md),
-[the command line](cli.md), [Julia](julia.md), [R](R.md), and [Go](go.md).
+[the command line](cli.md), [Julia](julia.md), [R](r.md), and [Go](go.md).
 
 ## Installing mlpack
 
@@ -50,7 +50,7 @@ df = pd.read_csv('http://www.mlpack.org/
 
 # Split the labels.
 labels = df['label']
-dataset = df.drop('label', 1)
+dataset = df.drop('label', axis = 1)
 
 # Split the dataset using mlpack.  The output comes back as a dictionary,
 # which we'll unpack for clarity of code.
@@ -92,7 +92,7 @@ different mlpack learners, or to interfa
 ## Using mlpack for movie recommendations
 
 In this example, we'll train a collaborative filtering model using mlpack's
-[`cf()`](https://www.mlpack.org/doc/stable/python_documentation.html#cf) method.
+[`cf()`](../user/bindings/python.md#cf) method.
 We'll train this on the
 [MovieLens dataset](https://grouplens.org/datasets/movielens/), and then we'll
 use the model that we train to give recommendations.
@@ -161,7 +161,7 @@ have only shown a little bit of the func
 commands are available with different functionality.  A full list of each of
 these commands and full documentation can be found on the following page:
 
- - [Python documentation](https://www.mlpack.org/doc/stable/python_documentation.html)
+ - [Python documentation](../user/bindings/python.md)
 
 Also, mlpack is much more flexible from C++ and allows much greater
 functionality.  So, more complicated tasks are possible if you are willing to
diff -pruN 4.3.0-2/doc/quickstart/r.md 4.4.0-1/doc/quickstart/r.md
--- 4.3.0-2/doc/quickstart/r.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/quickstart/r.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,162 @@
+# mlpack in R quickstart guide
+
+This page describes how you can quickly get started using mlpack from R and
+gives a few examples of usage, and pointers to deeper documentation.
+
+This quickstart guide is also available for [C++](cpp.md), [Python](python.md),
+[Julia](julia.md), [the command line](cli.md), and [Go](go.md).
+
+## Installing mlpack
+
+Installing the mlpack bindings for R is straightforward; you can just use
+CRAN:
+
+```r
+install.packages('mlpack')
+```
+
+Building the R bindings from scratch is a little more in-depth, though.  For
+information on that, follow the instructions in the
+[main README](../../README.md).
+
+## Simple mlpack quickstart example
+
+As a really simple example of how to use mlpack from R, let's do some
+simple classification on a subset of the standard machine learning `covertype`
+dataset.  We'll first split the dataset into a training set and a testing set,
+then we'll train an mlpack random forest on the training data, and finally we'll
+print the accuracy of the random forest on the test dataset.
+
+You can copy-paste this code directly into R to run it.
+
+```r
+if(!requireNamespace("data.table", quietly = TRUE)) { install.packages("data.table") }
+suppressMessages({
+    library("mlpack")
+    library("data.table")
+})
+
+# Load the dataset from an online URL.  Replace with 'covertype.csv.gz' if you
+# want to use on the full dataset.
+df <- fread("https://www.mlpack.org/datasets/covertype-small.csv.gz")
+
+# Split the labels.
+labels <- df[, .(label)]
+dataset <- df[, label:=NULL]
+
+# Split the dataset using mlpack.
+prepdata <- preprocess_split(input = dataset,
+                             input_labels = labels,
+                             test_ratio = 0.3,
+                             verbose = TRUE)
+
+# Train a random forest.
+output <- random_forest(training = prepdata$training,
+                        labels = prepdata$training_labels,
+                        print_training_accuracy = TRUE,
+                        num_trees = 10,
+                        minimum_leaf_size = 3,
+                        verbose = TRUE)
+rf_model <- output$output_model
+
+# Predict the labels of the test points.
+output <- random_forest(input_model = rf_model,
+                        test = prepdata$test,
+                        verbose = TRUE)
+
+# Now print the accuracy.  The third return value ('probabilities'), which we
+# ignored here, could also be used to generate an ROC curve.
+correct <- sum(output$predictions == prepdata$test_labels)
+cat(correct, "out of", length(prepdata$test_labels), "test points correct",
+    correct / length(prepdata$test_labels) * 100.0, "%\n")
+```
+
+We can see that we achieve reasonably good accuracy on the test dataset (80%+);
+if we use the full `covertype.csv.gz`, the accuracy should increase
+significantly (but training will take longer).
+
+It's easy to modify the code above to do more complex things, or to use
+different mlpack learners, or to interface with other machine learning toolkits.
+
+## Using mlpack for movie recommendations
+
+In this example, we'll train a collaborative filtering model using mlpack's
+[`cf()`](../user/bindings/r.md#cf) method.
+We'll train this on the
+[MovieLens dataset](https://grouplens.org/datasets/movielens/), and then we'll
+use the model that we train to give recommendations.
+
+You can copy-paste this code directly into R to run it.
+
+```r
+if(!requireNamespace("data.table", quietly = TRUE)) { install.packages("data.table") }
+suppressMessages({
+    library("mlpack")
+    library("data.table")
+})
+
+# First, load the MovieLens dataset.  This is taken from files.grouplens.org/
+# but reposted on mlpack.org as unpacked and slightly preprocessed data.
+ratings <- fread("http://www.mlpack.org/datasets/ml-20m/ratings-only.csv.gz")
+movies <- fread("http://www.mlpack.org/datasets/ml-20m/movies.csv.gz")
+
+# Hold out 10% of the dataset into a test set so we can evaluate performance.
+predata <- preprocess_split(input = ratings,
+                            test_ratio = 0.1,
+                            verbose = TRUE)
+
+# Train the model.  Change the rank to increase/decrease the complexity of the
+# model.
+output <- cf(training = predata$training,
+             test = predata$test,
+             rank = 10,
+             verbose = TRUE,
+             max_iteration=2,
+             algorithm = "RegSVD")
+cf_model <- output$output_model
+
+# Now query the 5 top movies for user 1.
+output <- cf(input_model = cf_model,
+             query = matrix(1),
+             recommendations = 10,
+             verbose = TRUE)
+
+# Get the names of the movies for user 1.
+cat("Recommendations for user 1:\n")
+for (i in 1:10) {
+  cat("  ", i, ":", as.character(movies[output$output[i], 3]), "\n")
+}
+```
+
+Here is some example output, showing that user 1 seems to have good taste in
+movies:
+
+```
+Recommendations for user 1:
+  0: Casablanca (1942)
+  1: Pan's Labyrinth (Laberinto del fauno, El) (2006)
+  2: Godfather, The (1972)
+  3: Answer This! (2010)
+  4: Life Is Beautiful (La Vita è bella) (1997)
+  5: Adventures of Tintin, The (2011)
+  6: Dark Knight, The (2008)
+  7: Out for Justice (1991)
+  8: Dr. Strangelove or: How I Learned to Stop Worrying and Love the Bomb (1964)
+  9: Schindler's List (1993)
+```
+
+## Next steps with mlpack
+
+After working through this overview to `mlpack`'s R package, we hope you are
+inspired to use `mlpack`' in your data science workflow.  However, the two
+examples above have only shown a little bit of the functionality of mlpack.
+Lots of other functions are available with different functionality.  A full list
+of each of these functions and full documentation can be found on the following
+page:
+
+ - [R documentation](../user/bindings/r.md)
+
+Also, mlpack is much more flexible from C++ and allows much greater
+functionality.  So, more complicated tasks are possible if you are willing to
+write C++ (or perhaps Rcpp).  To get started learning about mlpack in C++, a
+good starting point is the [C++ quickstart guide](cpp.md).
diff -pruN 4.3.0-2/doc/sidebar.html 4.4.0-1/doc/sidebar.html
--- 4.3.0-2/doc/sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,294 @@
+<!--
+sidebar.html
+
+This file defines the documentation sidebar.  If you are adding a new
+documentation file, be sure to add the link here in the right section.
+
+This is in pure HTML due to the kramdown parser not being able to read bulleted
+lists with details.
+
+All relative links should be prefixed with "LINKROOT".  These will be replaced
+when the sidebar is built for each page.
+-->
+
+<div id="sidebar">
+<ul>
+  <li>
+  <details open="true"> <!-- default open for non-binding pages -->
+    <summary>
+      <a href="LINKROOTindex.html" class="textlink"><b>C++ API Documentation</b></a>
+    </summary>
+
+    <ul>
+      <!-- General "getting started" documentation -->
+      <li>
+        <a href="LINKROOTquickstart/cpp.html">Quickstart</a>
+      </li>
+      <li>
+        <a href="LINKROOTuser/matrices.html">Matrices and data</a>
+      </li>
+      <li>
+        <a href="LINKROOTuser/load_save.html">Loading and saving</a>
+      </li>
+      <li class="bottom-space">
+        <a href="LINKROOTcitation.html">Citation</a>
+      </li>
+
+      <!-- Documentation for library core functions -->
+      <li>
+        <a href="LINKROOTuser/core.html">Core library functions</a>
+      </li>
+
+      <!-- Classification algorithms -->
+      <li>
+        <details>
+          <summary>
+            <a href="LINKROOTindex.html#classification-algorithms">
+              Classification
+            </a>
+          </summary>
+          <ul>
+            <li>
+              <a href="LINKROOTuser/methods/adaboost.html">
+                <code>AdaBoost</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/decision_tree.html">
+                <code>DecisionTree</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/hoeffding_tree.html">
+                <code>HoeffdingTree</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/linear_svm.html">
+                <code>LinearSVM</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/logistic_regression.html">
+                <code>LogisticRegression</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/naive_bayes_classifier.html">
+                <code>NaiveBayesClassifier</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/perceptron.html">
+                <code>Perceptron</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/random_forest.html">
+                <code>RandomForest</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/softmax_regression.html">
+                <code>SoftmaxRegression</code>
+              </a>
+            </li>
+          </ul>
+        </details>
+      </li>
+
+      <!-- Regression algorithms -->
+      <li>
+        <details>
+          <summary>
+            <a href="LINKROOTindex.html#regression-algorithms">
+              Regression
+            </a>
+          </summary>
+          <ul>
+            <li>
+              <a href="LINKROOTuser/methods/bayesian_linear_regression.html">
+                <code>BayesianLinearRegression</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/decision_tree_regressor.html">
+                <code>DecisionTreeRegressor</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/lars.html">
+                <code>LARS</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/linear_regression.html">
+                <code>LinearRegression</code>
+              </a>
+            </li>
+          </ul>
+        </details>
+      </li>
+
+      <!-- Clustering algorithms -->
+      <li>
+        <a href="LINKROOTindex.html#clustering-algorithms">
+          Clustering
+        </a>
+      </li>
+
+      <!-- Geometric algorithms -->
+      <li>
+        <a href="LINKROOTindex.html#geometric-algorithms">
+          Geometry
+        </a>
+      </li>
+
+      <!-- Preprocessing utilities -->
+      <li>
+        <a href="LINKROOTindex.html#preprocessing-utilities">
+          Preprocessing
+        </a>
+      </li>
+
+      <!-- Transformations -->
+      <li>
+        <details>
+          <summary>
+            <a href="LINKROOTindex.html#transformations">
+              Transformations
+            </a>
+          </summary>
+          <ul>
+            <li>
+              <a href="LINKROOTuser/methods/local_coordinate_coding.html">
+                <code>LocalCoordinateCoding</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/nmf.html">
+                <code>NMF</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/pca.html">
+                <code>PCA</code>
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/methods/sparse_coding.html">
+                <code>SparseCoding</code>
+              </a>
+            </li>
+          </ul>
+        </details>
+      </li>
+      
+      <!-- Modeling utilities -->
+      <li class="bottom-space">
+        <details>
+          <summary>
+            <a href="LINKROOTindex.html#modeling-utilities">
+              Modeling
+            </a>
+          </summary>
+          <ul>
+            <li>
+              <a href="LINKROOTuser/cv.html">
+                Cross-validation
+              </a>
+            </li>
+            <li>
+              <a href="LINKROOTuser/hpt.html">
+                Hyperparameter tuning
+              </a>
+            </li>
+          </ul>
+        </details>
+      </li>
+
+      <li>
+        <a href="LINKROOTindex.html#examples-and-further-documentation">
+          Examples
+        </a>
+      </li>
+
+      <li>
+        <a href="LINKROOTindex.html#developer-documentation">
+          Developers
+        </a>
+      </li>
+    </ul>
+  </details>
+</ul>
+
+<!-- Binding documentation section -->
+<ul id="binding_sidebar">
+  <li>
+  <details> <!-- default closed for non-binding pages -->
+    <summary>
+      <a href="LINKROOTindex.html#bindings-to-other-languages" class="textlink"><b>Binding API</b></a>
+    </summary>
+
+    <ul>
+      <li>
+        <a href="LINKROOTquickstart/python.html">
+          Python quickstart
+        </a>
+      </li>
+      <li class="bottom-space">
+        <a href="LINKROOTuser/bindings/python.html">
+          Python binding documentation
+        </a>
+      </li>
+
+      <li>
+        <a href="LINKROOTquickstart/julia.html">
+          Julia quickstart
+        </a>
+      </li>
+      <li class="bottom-space">
+        <a href="LINKROOTuser/bindings/julia.html">
+          Julia binding documentation
+        </a>
+      </li>
+
+      <li>
+        <a href="LINKROOTquickstart/r.html">
+          R quickstart
+        </a>
+      </li>
+      <li class="bottom-space">
+        <a href="LINKROOTuser/bindings/r.html">
+          R binding documentation
+        </a>
+      </li>
+
+      <li>
+        <a href="LINKROOTquickstart/cli.html">
+          CLI quickstart
+        </a>
+      </li>
+      <li class="bottom-space">
+        <a href="LINKROOTuser/bindings/cli.html">
+          CLI binding documentation
+        </a>
+      </li>
+
+      <li>
+        <a href="LINKROOTquickstart/go.html">
+          Go quickstart
+        </a>
+      </li>
+      <li>
+        <a href="LINKROOTuser/bindings/go.html">
+          Go binding documentation
+        </a>
+      </li>
+    </ul>
+  </details>
+
+  </li>
+</ul>
+
+<!-- Page-specific sidebar section will go next, and will close the div. -->
diff -pruN 4.3.0-2/doc/tutorials/ann.md 4.4.0-1/doc/tutorials/ann.md
--- 4.3.0-2/doc/tutorials/ann.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/tutorials/ann.md	2024-05-28 20:08:34.000000000 +0000
@@ -484,7 +484,7 @@ $ cat dataset.csv
 The type does not necessarily need to be a CSV; it can be any supported storage
 format, assuming that it is a coordinate-format file in the format specified
 above.  For more information on mlpack file formats, see the
-[tutorial](../user/formats.md).
+[load/save documentation](../user/load_save.md).
 
 *Note*: it’s often a good idea to normalize or standardize your data, for
 example using:
diff -pruN 4.3.0-2/doc/user/bindings/cli.md 4.4.0-1/doc/user/bindings/cli.md
--- 4.3.0-2/doc/user/bindings/cli.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/cli.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,3724 @@
+# mlpack CLI binding documentation
+
+## mlpack overview
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning library with bindings to other languages.  It aims to provide fast, lightweight implementations of both common and cutting-edge machine learning algorithms.
+
+This reference page details mlpack's bindings to the command-line.
+
+Further useful mlpack documentation links are given below.
+
+ - [mlpack homepage](https://www.mlpack.org/)
+ - [mlpack on Github](https://github.com/mlpack/mlpack)
+ - [mlpack main documentation page](https://www.mlpack.org/docs.html)
+
+See also the quickstart guide for CLI:
+
+ - [CLI Quickstart](../../quickstart/cli.md)
+
+## Data Formats
+
+<div id="data-formats-div" markdown="1">
+mlpack bindings for CLI take and return a restricted set of types, for simplicity.  These include primitive types, matrix/vector types, categorical matrix types, and model types. Each type is detailed below.
+
+ - `int`{: #doc_int}: An integer (i.e., "1").
+ - `double`{: #doc_double }: A floating-point number (i.e., "0.5").
+ - `flag`{: #doc_flag }: A boolean flag option.  If not specified, it is false; if specified, it is true.
+ - `string`{: #doc_string }: A character string (i.e., "hello").
+ - `int vector`{: #doc_int_vector }: A vector of integers, separated by commas (i.e., "1,2,3").
+ - `string vector`{: #doc_string_vector }: A vector of strings, separated by commas (i.e., "hello","goodbye").
+ - `2-d matrix file`{: #doc_a_2_d_matrix_file }: A data matrix filename.  The file can be CSV (.csv), TSV (.csv), ASCII (space-separated values, .txt), Armadillo ASCII (.txt), PGM (.pgm), PPM (.ppm), Armadillo binary (.bin), or HDF5 (.h5, .hdf, .hdf5, or .he5), if mlpack was compiled with HDF5 support.  The type of the data is detected by the extension of the filename.  The storage should be such that one row corresponds to one point, and one column corresponds to one dimension (this is the typical storage format for on-disk data).  CSV files will be checked for a header; if no header is found, the first row will be loaded as a data point.  All values of the matrix will be loaded as double-precision floating point data.
+ - `2-d index matrix file`{: #doc_a_2_d_index_matrix_file }: A data matrix filename, where the matrix holds only non-negative integer values.  This type is often used for labels or indices.  The file can be CSV (.csv), TSV (.csv), ASCII (space-separated values, .txt), Armadillo ASCII (.txt), PGM (.pgm), PPM (.ppm), Armadillo binary (.bin), or HDF5 (.h5, .hdf, .hdf5, or .he5), if mlpack was compiled with HDF5 support.  The type of the data is detected by the extension of the filename.  The storage should be such that one row corresponds to one point, and one column corresponds to one dimension (this is the typical storage format for on-disk data).  CSV files will be checked for a header; if no header is found, the first row will be loaded as a data point.  All values of the matrix will be loaded as unsigned integers.
+ - `1-d matrix file`{: #doc_a_1_d_matrix_file }: A one-dimensional vector filename.  This file can take the same formats as the data matrix filenames; however, it must either contain one row and many columns, or one column and many rows.
+ - `1-d index matrix file`{: #doc_a_1_d_index_matrix_file }: A one-dimensional vector filename, where the matrix holds only non-negative integer values.  This type is typically used for labels or predictions or other indices.  This file can take the same formats as the data matrix filenames; however, it must either contain one row and many columns, or one column and many rows.
+ - `2-d categorical matrix file`{: #doc_a_2_d_categorical_matrix_file }: A filename for a data matrix that can contain categorical (non-numeric) data.  If the file contains only numeric data, then the same formats for regular data matrices can be used.  If the file contains strings or other values that can't be parsed as numbers, then the type to be loaded must be CSV (.csv) or ARFF (.arff).  Any non-numeric data will be converted to an unsigned integer value, and dimensions where the data is converted will be treated as categorical dimensions.  When using this format, there is no need for one-hot encoding of categorical data.
+ - `mlpackModel file`{: #doc_model }: A filename containing an mlpack model.  These can have one of three formats: binary (.bin), text (.txt), and XML (.xml).  The XML format produces the largest (but most human-readable) files, while the binary format can be significantly more compact and quicker to load and save.
+</div>
+
+
+## mlpack_approx_kfn
+{: #approx_kfn }
+
+#### Approximate furthest neighbor search
+{: #approx_kfn_descr }
+
+```bash
+$ mlpack_approx_kfn [--algorithm 'ds'] [--calculate_error]
+        [--exact_distances_file <string>] [--help] [--info <string>]
+        [--input_model_file <string>] [--k 0] [--num_projections 5]
+        [--num_tables 5] [--query_file <string>] [--reference_file <string>]
+        [--verbose] [--version] [--distances_file <string>] [--neighbors_file
+        <string>] [--output_model_file <string>]
+```
+
+An implementation of two strategies for furthest neighbor search.  This can be used to compute the furthest neighbor of query point(s) from a set of points; furthest neighbor models can be saved and reused with future query point(s). [Detailed documentation](#approx_kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--algorithm (-a)` | [`string`](#doc_string) | Algorithm to use: 'ds' or 'qdafn'. | `'ds'` |
+| `--calculate_error (-e)` | [`flag`](#doc_flag) | If set, calculate the average distance error for the first furthest neighbor only. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--exact_distances_file (-x)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing exact distances to furthest neighbors; this can be used to avoid explicit calculation when --calculate_error is set. | `''` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`ApproxKFNModel file`](#doc_model) | File containing input model. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of furthest neighbors to search for. | `0` |
+| `--num_projections (-p)` | [`int`](#doc_int) | Number of projections to use in each hash table. | `5` |
+| `--num_tables (-t)` | [`int`](#doc_int) | Number of hash tables to use. | `5` |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing query points. | `''` |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save furthest neighbor distances to. | 
+| `--neighbors_file (-n)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to save neighbor indices to. | 
+| `--output_model_file (-M)` | [`ApproxKFNModel file`](#doc_model) | File to save output model to. | 
+
+### Detailed documentation
+{: #approx_kfn_detailed-documentation }
+
+This program implements two strategies for furthest neighbor search. These strategies are:
+
+ - The 'qdafn' algorithm from "Approximate Furthest Neighbor in High Dimensions" by R. Pagh, F. Silvestri, J. Sivertsen, and M. Skala, in Similarity Search and Applications 2015 (SISAP).
+ - The 'DrusillaSelect' algorithm from "Fast approximate furthest neighbors with data-dependent candidate selection", by R.R. Curtin and A.B. Gardner, in Similarity Search and Applications 2016 (SISAP).
+
+These two strategies give approximate results for the furthest neighbor search problem and can be used as fast replacements for other furthest neighbor techniques such as those found in the mlpack_kfn program.  Note that typically, the 'ds' algorithm requires far fewer tables and projections than the 'qdafn' algorithm.
+
+Specify a reference set (set to search in) with `--reference_file (-r)`, specify a query set with `--query_file (-q)`, and specify algorithm parameters with `--num_tables (-t)` and `--num_projections (-p)` (or don't and defaults will be used).  The algorithm to be used (either 'ds'---the default---or 'qdafn')  may be specified with `--algorithm (-a)`.  Also specify the number of neighbors to search for with `--k (-k)`.
+
+Note that for 'qdafn' in lower dimensions, `--num_projections (-p)` may need to be set to a high value in order to return results for each query point.
+
+If no query set is specified, the reference set will be used as the query set.  The `--output_model_file (-M)` output parameter may be used to store the built model, and an input model may be loaded instead of specifying a reference set with the `--input_model_file (-m)` option.
+
+Results for each query point can be stored with the `--neighbors_file (-n)` and `--distances_file (-d)` output parameters.  Each row of these output matrices holds the k distances or neighbor indices for each query point.
+
+### Example
+For example, to find the 5 approximate furthest neighbors with `'reference_set.csv'` as the reference set and `'query_set.csv'` as the query set using DrusillaSelect, storing the furthest neighbor indices to `'neighbors.csv'` and the furthest neighbor distances to `'distances.csv'`, one could call
+
+```bash
+$ mlpack_approx_kfn --query_file query_set.csv --reference_file
+  reference_set.csv --k 5 --algorithm ds --neighbors_file neighbors.csv
+  --distances_file distances.csv
+```
+
+and to perform approximate all-furthest-neighbors search with k=1 on the set `'data.csv'` storing only the furthest neighbor distances to `'distances.csv'`, one could call
+
+```bash
+$ mlpack_approx_kfn --reference_file reference_set.csv --k 1 --distances_file
+  distances.csv
+```
+
+A trained model can be re-used.  If a model has been previously saved to `'model.bin'`, then we may find 3 approximate furthest neighbors on a query set `'new_query_set.csv'` using that model and store the furthest neighbor indices into `'neighbors.csv'` by calling
+
+```bash
+$ mlpack_approx_kfn --input_model_file model.bin --query_file
+  new_query_set.csv --k 3 --neighbors_file neighbors.csv
+```
+
+### See also
+
+ - [k-furthest-neighbor search](#kfn)
+ - [k-nearest-neighbor search](#knn)
+ - [Fast approximate furthest neighbors with data-dependent candidate selection (pdf)](http://ratml.org/pub/pdf/2016fast.pdf)
+ - [Approximate furthest neighbor in high dimensions (pdf)](https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf)
+ - [QDAFN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/qdafn.hpp)
+ - [DrusillaSelect class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/drusilla_select.hpp)
+
+## mlpack_bayesian_linear_regression
+{: #bayesian_linear_regression }
+
+#### BayesianLinearRegression
+{: #bayesian_linear_regression_descr }
+
+```bash
+$ mlpack_bayesian_linear_regression [--center] [--help] [--info
+        <string>] [--input_file <string>] [--input_model_file <string>]
+        [--responses_file <string>] [--scale] [--test_file <string>] [--verbose]
+        [--version] [--output_model_file <string>] [--predictions_file <string>]
+        [--stds_file <string>]
+```
+
+An implementation of the bayesian linear regression. [Detailed documentation](#bayesian_linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--center (-c)` | [`flag`](#doc_flag) | Center the data and fit the intercept if enabled. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of covariates (X). | `''` |
+| `--input_model_file (-m)` | [`BayesianLinearRegression<> file`](#doc_model) | Trained BayesianLinearRegression model to use. | `''` |
+| `--responses_file (-r)` | [`1-d matrix file`](#doc_a_1_d_matrix_file) | Matrix of responses/observations (y). | `''` |
+| `--scale (-s)` | [`flag`](#doc_flag) | Scale each feature by their standard deviations if enabled. |  |
+| `--test_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing points to regress on (test points). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`BayesianLinearRegression<> file`](#doc_model) | Output BayesianLinearRegression model. | 
+| `--predictions_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+| `--stds_file (-u)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If specified, this is where the standard deviations of the predictive distribution will be saved. | 
+
+### Detailed documentation
+{: #bayesian_linear_regression_detailed-documentation }
+
+An implementation of the bayesian linear regression.
+This model is a probabilistic view and implementation of the linear regression. The final solution is obtained by computing a posterior distribution from gaussian likelihood and a zero mean gaussian isotropic  prior distribution on the solution. 
+Optimization is AUTOMATIC and does not require cross validation. The optimization is performed by maximization of the evidence function. Parameters are tuned during the maximization of the marginal likelihood. This procedure includes the Ockham's razor that penalizes over complex solutions. 
+
+This program is able to train a Bayesian linear regression model or load a model from file, output regression predictions for a test set, and save the trained model to a file.
+
+To train a BayesianLinearRegression model, the `--input_file (-i)` and `--responses_file (-r)`parameters must be given. The `--center (-c)`and `--scale (-s)` parameters control the centering and the normalizing options. A trained model can be saved with the `--output_model_file (-M)`. If no training is desired at all, a model can be passed via the `--input_model_file (-m)` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `--test_file (-t)` parameter.  Predicted responses to the test points can be saved with the `--predictions_file (-o)` output parameter. The corresponding standard deviation can be save by precising the `--stds_file (-u)` parameter.
+
+### Example
+For example, the following command trains a model on the data `'data.csv'` and responses `'responses.csv'`with center set to true and scale set to false (so, Bayesian linear regression is being solved, and then the model is saved to `'blr_model.bin'`:
+
+```bash
+$ mlpack_bayesian_linear_regression --input_file data.csv --responses_file
+  responses.csv --center --scale --output_model_file blr_model.bin
+```
+
+The following command uses the `'blr_model.bin'` to provide predicted  responses for the data `'test.csv'` and save those  responses to `'test_predictions.csv'`: 
+
+```bash
+$ mlpack_bayesian_linear_regression --input_model_file blr_model.bin
+  --test_file test.csv --predictions_file test_predictions.csv
+```
+
+Because the estimator computes a predictive distribution instead of a simple point estimate, the `--stds_file (-u)` parameter allows one to save the prediction uncertainties: 
+
+```bash
+$ mlpack_bayesian_linear_regression --input_model_file blr_model.bin
+  --test_file test.csv --predictions_file test_predictions.csv --stds_file
+  stds.csv
+```
+
+### See also
+
+ - [Bayesian Interpolation](https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf)
+ - [Bayesian Linear Regression, Section 3.3](https://www.microsoft.com/en-us/research/uploads/prod/2006/01/Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf)
+ - [BayesianLinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp)
+
+## mlpack_cf
+{: #cf }
+
+#### Collaborative Filtering
+{: #cf_descr }
+
+```bash
+$ mlpack_cf [--algorithm 'NMF'] [--all_user_recommendations] [--help]
+        [--info <string>] [--input_model_file <string>] [--interpolation
+        'average'] [--iteration_only_termination] [--max_iterations 1000]
+        [--min_residue 1e-05] [--neighbor_search 'euclidean'] [--neighborhood 5]
+        [--normalization 'none'] [--query_file <string>] [--rank 0]
+        [--recommendations 5] [--seed 0] [--test_file <string>] [--training_file
+        <string>] [--verbose] [--version] [--output_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of several collaborative filtering (CF) techniques for recommender systems.  This can be used to train a new CF model, or use an existing CF model to compute recommendations. [Detailed documentation](#cf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--algorithm (-a)` | [`string`](#doc_string) | Algorithm used for matrix factorization. | `'NMF'` |
+| `--all_user_recommendations (-A)` | [`flag`](#doc_flag) | Generate recommendations for all users. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`CFModel file`](#doc_model) | Trained CF model to load. | `''` |
+| `--interpolation (-i)` | [`string`](#doc_string) | Algorithm used for weight interpolation. | `'average'` |
+| `--iteration_only_termination (-I)` | [`flag`](#doc_flag) | Terminate only when the maximum number of iterations is reached. |  |
+| `--max_iterations (-N)` | [`int`](#doc_int) | Maximum number of iterations. If set to zero, there is no limit on the number of iterations. | `1000` |
+| `--min_residue (-r)` | [`double`](#doc_double) | Residue required to terminate the factorization (lower values generally mean better fits). | `1e-05` |
+| `--neighbor_search (-S)` | [`string`](#doc_string) | Algorithm used for neighbor search. | `'euclidean'` |
+| `--neighborhood (-n)` | [`int`](#doc_int) | Size of the neighborhood of similar users to consider for each query user. | `5` |
+| `--normalization (-z)` | [`string`](#doc_string) | Normalization performed on the ratings. | `'none'` |
+| `--query_file (-q)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | List of query users for which recommendations should be generated. | `''` |
+| `--rank (-R)` | [`int`](#doc_int) | Rank of decomposed matrices (if 0, a heuristic is used to estimate the rank). | `0` |
+| `--recommendations (-c)` | [`int`](#doc_int) | Number of recommendations to generate for each query user. | `5` |
+| `--seed (-s)` | [`int`](#doc_int) | Set the random seed (0 uses std::time(NULL)). | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Test set to calculate RMSE on. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform CF on. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix that will store output recommendations. | 
+| `--output_model_file (-M)` | [`CFModel file`](#doc_model) | Output for trained CF model. | 
+
+### Detailed documentation
+{: #cf_detailed-documentation }
+
+This program performs collaborative filtering (CF) on the given dataset. Given a list of user, item and preferences (the `--training_file (-t)` parameter), the program will perform a matrix decomposition and then can perform a series of actions related to collaborative filtering.  Alternately, the program can load an existing saved CF model with the `--input_model_file (-m)` parameter and then use that model to provide recommendations or predict values.
+
+The input matrix should be a 3-dimensional matrix of ratings, where the first dimension is the user, the second dimension is the item, and the third dimension is that user's rating of that item.  Both the users and items should be numeric indices, not names. The indices are assumed to start from 0.
+
+A set of query users for which recommendations can be generated may be specified with the `--query_file (-q)` parameter; alternately, recommendations may be generated for every user in the dataset by specifying the `--all_user_recommendations (-A)` parameter.  In addition, the number of recommendations per user to generate can be specified with the `--recommendations (-c)` parameter, and the number of similar users (the size of the neighborhood) to be considered when generating recommendations can be specified with the `--neighborhood (-n)` parameter.
+
+For performing the matrix decomposition, the following optimization algorithms can be specified via the `--algorithm (-a)` parameter:
+
+ - 'RegSVD' -- Regularized SVD using a SGD optimizer
+ - 'NMF' -- Non-negative matrix factorization with alternating least squares update rules
+ - 'BatchSVD' -- SVD batch learning
+ - 'SVDIncompleteIncremental' -- SVD incomplete incremental learning
+ - 'SVDCompleteIncremental' -- SVD complete incremental learning
+ - 'BiasSVD' -- Bias SVD using a SGD optimizer
+ - 'SVDPP' -- SVD++ using a SGD optimizer
+ - 'RandSVD' -- RandomizedSVD learning
+ - 'QSVD' -- QuicSVD learning
+ - 'BKSVD' -- Block Krylov SVD learning
+
+
+The following neighbor search algorithms can be specified via the `--neighbor_search (-S)` parameter:
+
+ - 'cosine'  -- Cosine Search Algorithm
+ - 'euclidean'  -- Euclidean Search Algorithm
+ - 'pearson'  -- Pearson Search Algorithm
+
+
+The following weight interpolation algorithms can be specified via the `--interpolation (-i)` parameter:
+
+ - 'average'  -- Average Interpolation Algorithm
+ - 'regression'  -- Regression Interpolation Algorithm
+ - 'similarity'  -- Similarity Interpolation Algorithm
+
+
+The following ranking normalization algorithms can be specified via the `--normalization (-z)` parameter:
+
+ - 'none'  -- No Normalization
+ - 'item_mean'  -- Item Mean Normalization
+ - 'overall_mean'  -- Overall Mean Normalization
+ - 'user_mean'  -- User Mean Normalization
+ - 'z_score'  -- Z-Score Normalization
+
+A trained model may be saved to with the `--output_model_file (-M)` output parameter.
+
+### Example
+To train a CF model on a dataset `'training_set.csv'` using NMF for decomposition and saving the trained model to `'model.bin'`, one could call: 
+
+```bash
+$ mlpack_cf --training_file training_set.csv --algorithm NMF
+  --output_model_file model.bin
+```
+
+Then, to use this model to generate recommendations for the list of users in the query set `'users.csv'`, storing 5 recommendations in `'recommendations.csv'`, one could call 
+
+```bash
+$ mlpack_cf --input_model_file model.bin --query_file users.csv
+  --recommendations 5 --output_file recommendations.csv
+```
+
+### See also
+
+ - [Collaborative Filtering on Wikipedia](https://en.wikipedia.org/wiki/Collaborative_filtering)
+ - [Matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Matrix_factorization_(recommender_systems))
+ - [Matrix factorization techniques for recommender systems (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5)
+ - [CFType class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/cf/cf.hpp)
+
+## mlpack_dbscan
+{: #dbscan }
+
+#### DBSCAN clustering
+{: #dbscan_descr }
+
+```bash
+$ mlpack_dbscan [--epsilon 1] [--help] [--info <string>] --input_file
+        <string> [--min_size 5] [--naive] [--selection_type 'ordered']
+        [--single_mode] [--tree_type 'kd'] [--verbose] [--version]
+        [--assignments_file <string>] [--centroids_file <string>]
+```
+
+An implementation of DBSCAN clustering.  Given a dataset, this can compute and return a clustering of that dataset. [Detailed documentation](#dbscan_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--epsilon (-e)` | [`double`](#doc_double) | Radius of each range search. | `1` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to cluster. | `**--**` |
+| `--min_size (-m)` | [`int`](#doc_int) | Minimum number of points for a cluster. | `5` |
+| `--naive (-N)` | [`flag`](#doc_flag) | If set, brute-force range search (not tree-based) will be used. |  |
+| `--selection_type (-s)` | [`string`](#doc_string) | If using point selection policy, the type of selection to use ('ordered', 'random'). | `'ordered'` |
+| `--single_mode (-S)` | [`flag`](#doc_flag) | If set, single-tree range search (not dual-tree) will be used. |  |
+| `--tree_type (-t)` | [`string`](#doc_string) | If using single-tree or dual-tree search, the type of tree to use ('kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'). | `'kd'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--assignments_file (-a)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Output matrix for assignments of each point. | 
+| `--centroids_file (-C)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save output centroids to. | 
+
+### Detailed documentation
+{: #dbscan_detailed-documentation }
+
+This program implements the DBSCAN algorithm for clustering using accelerated tree-based range search.  The type of tree that is used may be parameterized, or brute-force range search may also be used.
+
+The input dataset to be clustered may be specified with the `--input_file (-i)` parameter; the radius of each range search may be specified with the `--epsilon (-e)` parameters, and the minimum number of points in a cluster may be specified with the `--min_size (-m)` parameter.
+
+The `--assignments_file (-a)` and `--centroids_file (-C)` output parameters may be used to save the output of the clustering. `--assignments_file (-a)` contains the cluster assignments of each point, and `--centroids_file (-C)` contains the centroids of each cluster.
+
+The range search may be controlled with the `--tree_type (-t)`, `--single_mode (-S)`, and `--naive (-N)` parameters.  `--tree_type (-t)` can control the type of tree used for range search; this can take a variety of values: 'kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'. The `--single_mode (-S)` parameter will force single-tree search (as opposed to the default dual-tree search), and '`--naive (-N)` will force brute-force range search.
+
+### Example
+An example usage to run DBSCAN on the dataset in `'input.csv'` with a radius of 0.5 and a minimum cluster size of 5 is given below:
+
+```bash
+$ mlpack_dbscan --input_file input.csv --epsilon 0.5 --min_size 5
+```
+
+### See also
+
+ - [DBSCAN on Wikipedia](https://en.wikipedia.org/wiki/DBSCAN)
+ - [A density-based algorithm for discovering clusters in large spatial databases with noise (pdf)](https://cdn.aaai.org/KDD/1996/KDD96-037.pdf)
+ - [DBSCAN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/dbscan/dbscan.hpp)
+
+## mlpack_decision_tree
+{: #decision_tree }
+
+#### Decision tree
+{: #decision_tree_descr }
+
+```bash
+$ mlpack_decision_tree [--help] [--info <string>] [--input_model_file
+        <string>] [--labels_file <string>] [--maximum_depth 0]
+        [--minimum_gain_split 1e-07] [--minimum_leaf_size 20]
+        [--print_training_accuracy] [--print_training_error] [--test_file
+        <string>] [--test_labels_file <string>] [--training_file <string>]
+        [--verbose] [--version] [--weights_file <string>] [--output_model_file
+        <string>] [--predictions_file <string>] [--probabilities_file <string>]
+```
+
+An implementation of an ID3-style decision tree for classification, which supports categorical data.  Given labeled data with numeric or categorical features, a decision tree can be trained and saved; or, an existing decision tree can be used for classification on new points. [Detailed documentation](#decision_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`DecisionTreeModel file`](#doc_model) | Pre-trained decision tree, to be used with test points. | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Training labels. | `''` |
+| `--maximum_depth (-D)` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `--minimum_gain_split (-g)` | [`double`](#doc_double) | Minimum gain for node splitting. | `1e-07` |
+| `--minimum_leaf_size (-n)` | [`int`](#doc_int) | Minimum number of points in a leaf. | `20` |
+| `--print_training_accuracy (-a)` | [`flag`](#doc_flag) | Print the training accuracy. |  |
+| `--print_training_error (-e)` | [`flag`](#doc_flag) | Print the training error (deprecated; will be removed in mlpack 4.0.0). |  |
+| `--test_file (-T)` | [`2-d categorical matrix file`](#doc_a_2_d_categorical_matrix_file) | Testing dataset (may be categorical). | `''` |
+| `--test_labels_file (-L)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Test point labels, if accuracy calculation is desired. | `''` |
+| `--training_file (-t)` | [`2-d categorical matrix file`](#doc_a_2_d_categorical_matrix_file) | Training dataset (may be categorical). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--weights_file (-w)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The weight of labels | `''` |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`DecisionTreeModel file`](#doc_model) | Output for trained decision tree. | 
+| `--predictions_file (-p)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Class predictions for each test point. | 
+| `--probabilities_file (-P)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Class probabilities for each test point. | 
+
+### Detailed documentation
+{: #decision_tree_detailed-documentation }
+
+Train and evaluate using a decision tree.  Given a dataset containing numeric or categorical features, and associated labels for each point in the dataset, this program can train a decision tree on that data.
+
+The training set and associated labels are specified with the `--training_file (-t)` and `--labels_file (-l)` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `--labels_file (-l)` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `--output_model_file (-M)` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `--input_model_file (-m)` parameter.  The `--input_model_file (-m)` parameter may not be specified when the `--training_file (-t)` parameter is specified.  The `--minimum_leaf_size (-n)` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `--minimum_gain_split (-g)` parameter specifies the minimum gain that is needed for the node to split.  The `--maximum_depth (-D)` parameter specifies the maximum depth of the tree.  If `--print_training_error (-e)` is specified, the training error will be printed.
+
+Test data may be specified with the `--test_file (-T)` parameter, and if performance numbers are desired for that test set, labels may be specified with the `--test_labels_file (-L)` parameter.  Predictions for each test point may be saved via the `--predictions_file (-p)` output parameter.  Class probabilities for each prediction may be saved with the `--probabilities_file (-P)` output parameter.
+
+### Example
+For example, to train a decision tree with a minimum leaf size of 20 on the dataset contained in `'data.csv'` with labels `'labels.csv'`, saving the output model to `'tree.bin'` and printing the training error, one could call
+
+```bash
+$ mlpack_decision_tree --training_file data.arff --labels_file labels.csv
+  --output_model_file tree.bin --minimum_leaf_size 20 --minimum_gain_split 0.001
+  --print_training_accuracy
+```
+
+Then, to use that model to classify points in `'test_set.csv'` and print the test error given the labels `'test_labels.csv'` using that model, while saving the predictions for each point to `'predictions.csv'`, one could call 
+
+```bash
+$ mlpack_decision_tree --input_model_file tree.bin --test_file test_set.arff
+  --test_labels_file test_labels.csv --predictions_file predictions.csv
+```
+
+### See also
+
+ - [Random forest](#random_forest)
+ - [Decision trees on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+ - [Induction of Decision Trees (pdf)](http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf)
+ - [DecisionTree C++ class documentation](../../user/methods/decision_tree.md)
+
+## mlpack_det
+{: #det }
+
+#### Density Estimation With Density Estimation Trees
+{: #det_descr }
+
+```bash
+$ mlpack_det [--folds 10] [--help] [--info <string>] [--input_model_file
+        <string>] [--max_leaf_size 10] [--min_leaf_size 5] [--path_format 'lr']
+        [--skip_pruning] [--test_file <string>] [--training_file <string>]
+        [--verbose] [--version] [--output_model_file <string>]
+        [--tag_counters_file <string>] [--tag_file <string>]
+        [--test_set_estimates_file <string>] [--training_set_estimates_file
+        <string>] [--vi_file <string>]
+```
+
+An implementation of density estimation trees for the density estimation task.  Density estimation trees can be trained or used to predict the density at locations given by query points. [Detailed documentation](#det_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--folds (-f)` | [`int`](#doc_int) | The number of folds of cross-validation to perform for the estimation (0 is LOOCV) | `10` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`DTree<> file`](#doc_model) | Trained density estimation tree to load. | `''` |
+| `--max_leaf_size (-L)` | [`int`](#doc_int) | The maximum size of a leaf in the unpruned, fully grown DET. | `10` |
+| `--min_leaf_size (-l)` | [`int`](#doc_int) | The minimum size of a leaf in the unpruned, fully grown DET. | `5` |
+| `--path_format (-p)` | [`string`](#doc_string) | The format of path printing: 'lr', 'id-lr', or 'lr-id'. | `'lr'` |
+| `--skip_pruning (-s)` | [`flag`](#doc_flag) | Whether to bypass the pruning process and output the unpruned tree only. |  |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A set of test points to estimate the density of. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The data set on which to build a density estimation tree. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`DTree<> file`](#doc_model) | Output to save trained density estimation tree to. | 
+| `--tag_counters_file (-c)` | [`string`](#doc_string) | The file to output the number of points that went to each leaf. | 
+| `--tag_file (-g)` | [`string`](#doc_string) | The file to output the tags (and possibly paths) for each sample in the test set. | 
+| `--test_set_estimates_file (-E)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The output estimates on the test set from the final optimally pruned tree. | 
+| `--training_set_estimates_file (-e)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The output density estimates on the training set from the final optimally pruned tree. | 
+| `--vi_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The output variable importance values for each feature. | 
+
+### Detailed documentation
+{: #det_detailed-documentation }
+
+This program performs a number of functions related to Density Estimation Trees.  The optimal Density Estimation Tree (DET) can be trained on a set of data (specified by `--training_file (-t)`) using cross-validation (with number of folds specified with the `--folds (-f)` parameter).  This trained density estimation tree may then be saved with the `--output_model_file (-M)` output parameter.
+
+The variable importances (that is, the feature importance values for each dimension) may be saved with the `--vi_file (-i)` output parameter, and the density estimates for each training point may be saved with the `--training_set_estimates_file (-e)` output parameter.
+
+Enabling path printing for each node outputs the path from the root node to a leaf for each entry in the test set, or training set (if a test set is not provided).  Strings like 'LRLRLR' (indicating that traversal went to the left child, then the right child, then the left child, and so forth) will be output. If 'lr-id' or 'id-lr' are given as the `--path_format (-p)` parameter, then the ID (tag) of every node along the path will be printed after or before the L or R character indicating the direction of traversal, respectively.
+
+This program also can provide density estimates for a set of test points, specified in the `--test_file (-T)` parameter.  The density estimation tree used for this task will be the tree that was trained on the given training points, or a tree given as the parameter `--input_model_file (-m)`.  The density estimates for the test points may be saved using the `--test_set_estimates_file (-E)` output parameter.
+
+### See also
+
+ - [Density estimation on Wikipedia](https://en.wikipedia.org/wiki/Density_estimation)
+ - [Density estimation trees (pdf)](https://www.mlpack.org/papers/det.pdf)
+ - [DTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/det/dtree.hpp)
+
+## mlpack_emst
+{: #emst }
+
+#### Fast Euclidean Minimum Spanning Tree
+{: #emst_descr }
+
+```bash
+$ mlpack_emst [--help] [--info <string>] --input_file <string>
+        [--leaf_size 1] [--naive] [--verbose] [--version] [--output_file
+        <string>]
+```
+
+An implementation of the Dual-Tree Boruvka algorithm for computing the Euclidean minimum spanning tree of a set of input points. [Detailed documentation](#emst_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input data matrix. | `**--**` |
+| `--leaf_size (-l)` | [`int`](#doc_int) | Leaf size in the kd-tree.  One-element leaves give the empirically best performance, but at the cost of greater memory requirements. | `1` |
+| `--naive (-n)` | [`flag`](#doc_flag) | Compute the MST using O(n^2) naive algorithm. |  |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output data.  Stored as an edge list. | 
+
+### Detailed documentation
+{: #emst_detailed-documentation }
+
+This program can compute the Euclidean minimum spanning tree of a set of input points using the dual-tree Boruvka algorithm.
+
+The set to calculate the minimum spanning tree of is specified with the `--input_file (-i)` parameter, and the output may be saved with the `--output_file (-o)` output parameter.
+
+The `--leaf_size (-l)` parameter controls the leaf size of the kd-tree that is used to calculate the minimum spanning tree, and if the `--naive (-n)` option is given, then brute-force search is used (this is typically much slower in low dimensions).  The leaf size does not affect the results, but it may have some effect on the runtime of the algorithm.
+
+### Example
+For example, the minimum spanning tree of the input dataset `'data.csv'` can be calculated with a leaf size of 20 and stored as `'spanning_tree.csv'` using the following command:
+
+```bash
+$ mlpack_emst --input_file data.csv --leaf_size 20 --output_file
+  spanning_tree.csv
+```
+
+The output matrix is a three-dimensional matrix, where each row indicates an edge.  The first dimension corresponds to the lesser index of the edge; the second dimension corresponds to the greater index of the edge; and the third column corresponds to the distance between the two points.
+
+### See also
+
+ - [Minimum spanning tree on Wikipedia](https://en.wikipedia.org/wiki/Minimum_spanning_tree)
+ - [Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications (pdf)](https://www.mlpack.org/papers/emst.pdf)
+ - [DualTreeBoruvka class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/emst/dtb.hpp)
+
+## mlpack_fastmks
+{: #fastmks }
+
+#### FastMKS (Fast Max-Kernel Search)
+{: #fastmks_descr }
+
+```bash
+$ mlpack_fastmks [--bandwidth 1] [--base 2] [--degree 2] [--help]
+        [--info <string>] [--input_model_file <string>] [--k 0] [--kernel
+        'linear'] [--naive] [--offset 0] [--query_file <string>]
+        [--reference_file <string>] [--scale 1] [--single] [--verbose]
+        [--version] [--indices_file <string>] [--kernels_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of the single-tree and dual-tree fast max-kernel search (FastMKS) algorithm.  Given a set of reference points and a set of query points, this can find the reference point with maximum kernel value for each query point; trained models can be reused for future queries. [Detailed documentation](#fastmks_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--bandwidth (-w)` | [`double`](#doc_double) | Bandwidth (for Gaussian, Epanechnikov, and triangular kernels). | `1` |
+| `--base (-b)` | [`double`](#doc_double) | Base to use during cover tree construction. | `2` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--degree (-d)` | [`double`](#doc_double) | Degree of polynomial kernel. | `2` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`FastMKSModel file`](#doc_model) | Input FastMKS model to use. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of maximum kernels to find. | `0` |
+| `--kernel (-K)` | [`string`](#doc_string) | Kernel type to use: 'linear', 'polynomial', 'cosine', 'gaussian', 'epanechnikov', 'triangular', 'hyptan'. | `'linear'` |
+| `--naive (-N)` | [`flag`](#doc_flag) | If true, O(n^2) naive mode is used for computation. |  |
+| `--offset (-o)` | [`double`](#doc_double) | Offset of kernel (for polynomial and hyptan kernels). | `0` |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The query dataset. | `''` |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The reference dataset. | `''` |
+| `--scale (-s)` | [`double`](#doc_double) | Scale of kernel (for hyptan kernel). | `1` |
+| `--single (-S)` | [`flag`](#doc_flag) | If true, single-tree search is used (as opposed to dual-tree search. |  |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--indices_file (-i)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Output matrix of indices. | 
+| `--kernels_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output matrix of kernels. | 
+| `--output_model_file (-M)` | [`FastMKSModel file`](#doc_model) | Output for FastMKS model. | 
+
+### Detailed documentation
+{: #fastmks_detailed-documentation }
+
+This program will find the k maximum kernels of a set of points, using a query set and a reference set (which can optionally be the same set). More specifically, for each point in the query set, the k points in the reference set with maximum kernel evaluations are found.  The kernel function used is specified with the `--kernel (-K)` parameter.
+
+### Example
+For example, the following command will calculate, for each point in the query set `'query.csv'`, the five points in the reference set `'reference.csv'` with maximum kernel evaluation using the linear kernel.  The kernel evaluations may be saved with the  `'kernels.csv'` output parameter and the indices may be saved with the `'indices.csv'` output parameter.
+
+```bash
+$ mlpack_fastmks --k 5 --reference_file reference.csv --query_file query.csv
+  --indices_file indices.csv --kernels_file kernels.csv --kernel linear
+```
+
+The output matrices are organized such that row i and column j in the indices matrix corresponds to the index of the point in the reference set that has j'th largest kernel evaluation with the point in the query set with index i.  Row i and column j in the kernels matrix corresponds to the kernel evaluation between those two points.
+
+This program performs FastMKS using a cover tree.  The base used to build the cover tree can be specified with the `--base (-b)` parameter.
+
+### See also
+
+ - [k-nearest-neighbor search](#knn)
+ - [Dual-tree Fast Exact Max-Kernel Search (pdf)](https://mlpack.org/papers/fmks.pdf)
+ - [FastMKS class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/fastmks/fastmks.hpp)
+
+## mlpack_gmm_train
+{: #gmm_train }
+
+#### Gaussian Mixture Model (GMM) Training
+{: #gmm_train_descr }
+
+```bash
+$ mlpack_gmm_train [--diagonal_covariance] --gaussians 0 [--help]
+        [--info <string>] --input_file <string> [--input_model_file <string>]
+        [--kmeans_max_iterations 1000] [--max_iterations 250]
+        [--no_force_positive] [--noise 0] [--percentage 0.02] [--refined_start]
+        [--samplings 100] [--seed 0] [--tolerance 1e-10] [--trials 1]
+        [--verbose] [--version] [--output_model_file <string>]
+```
+
+An implementation of the EM algorithm for training Gaussian mixture models (GMMs).  Given a dataset, this can train a GMM for future use with other tools. [Detailed documentation](#gmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--diagonal_covariance (-d)` | [`flag`](#doc_flag) | Force the covariance of the Gaussians to be diagonal.  This can accelerate training time significantly. |  |
+| `--gaussians (-g)` | [`int`](#doc_int) | Number of Gaussians in the GMM. | `**--**` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The training data on which the model will be fit. | `**--**` |
+| `--input_model_file (-m)` | [`GMM file`](#doc_model) | Initial input GMM model to start training with. | `''` |
+| `--kmeans_max_iterations (-k)` | [`int`](#doc_int) | Maximum number of iterations for the k-means algorithm (used to initialize EM). | `1000` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations of EM algorithm (passing 0 will run until convergence). | `250` |
+| `--no_force_positive (-P)` | [`flag`](#doc_flag) | Do not force the covariance matrices to be positive definite. |  |
+| `--noise (-N)` | [`double`](#doc_double) | Variance of zero-mean Gaussian noise to add to data. | `0` |
+| `--percentage (-p)` | [`double`](#doc_double) | If using --refined_start, specify the percentage of the dataset used for each sampling (should be between 0.0 and 1.0). | `0.02` |
+| `--refined_start (-r)` | [`flag`](#doc_flag) | During the initialization, use refined initial positions for k-means clustering (Bradley and Fayyad, 1998). |  |
+| `--samplings (-S)` | [`int`](#doc_int) | If using --refined_start, specify the number of samplings used for initial points. | `100` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--tolerance (-T)` | [`double`](#doc_double) | Tolerance for convergence of EM. | `1e-10` |
+| `--trials (-t)` | [`int`](#doc_int) | Number of trials to perform in training GMM. | `1` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`GMM file`](#doc_model) | Output for trained GMM model. | 
+
+### Detailed documentation
+{: #gmm_train_detailed-documentation }
+
+This program takes a parametric estimate of a Gaussian mixture model (GMM) using the EM algorithm to find the maximum likelihood estimate.  The model may be saved and reused by other mlpack GMM tools.
+
+The input data to train on must be specified with the `--input_file (-i)` parameter, and the number of Gaussians in the model must be specified with the `--gaussians (-g)` parameter.  Optionally, many trials with different random initializations may be run, and the result with highest log-likelihood on the training data will be taken.  The number of trials to run is specified with the `--trials (-t)` parameter.  By default, only one trial is run.
+
+The tolerance for convergence and maximum number of iterations of the EM algorithm are specified with the `--tolerance (-T)` and `--max_iterations (-n)` parameters, respectively.  The GMM may be initialized for training with another model, specified with the `--input_model_file (-m)` parameter. Otherwise, the model is initialized by running k-means on the data.  The k-means clustering initialization can be controlled with the `--kmeans_max_iterations (-k)`, `--refined_start (-r)`, `--samplings (-S)`, and `--percentage (-p)` parameters.  If `--refined_start (-r)` is specified, then the Bradley-Fayyad refined start initialization will be used.  This can often lead to better clustering results.
+
+The 'diagonal_covariance' flag will cause the learned covariances to be diagonal matrices.  This significantly simplifies the model itself and causes training to be faster, but restricts the ability to fit more complex GMMs.
+
+If GMM training fails with an error indicating that a covariance matrix could not be inverted, make sure that the `--no_force_positive (-P)` parameter is not specified.  Alternately, adding a small amount of Gaussian noise (using the `--noise (-N)` parameter) to the entire dataset may help prevent Gaussians with zero variance in a particular dimension, which is usually the cause of non-invertible covariance matrices.
+
+The `--no_force_positive (-P)` parameter, if set, will avoid the checks after each iteration of the EM algorithm which ensure that the covariance matrices are positive definite.  Specifying the flag can cause faster runtime, but may also cause non-positive definite covariance matrices, which will cause the program to crash.
+
+### Example
+As an example, to train a 6-Gaussian GMM on the data in `'data.csv'` with a maximum of 100 iterations of EM and 3 trials, saving the trained GMM to `'gmm.bin'`, the following command can be used:
+
+```bash
+$ mlpack_gmm_train --input_file data.csv --gaussians 6 --trials 3
+  --output_model_file gmm.bin
+```
+
+To re-train that GMM on another set of data `'data2.csv'`, the following command may be used: 
+
+```bash
+$ mlpack_gmm_train --input_model_file gmm.bin --input_file data2.csv
+  --gaussians 6 --output_model_file new_gmm.bin
+```
+
+### See also
+
+ - [mlpack_gmm_generate](#gmm_generate)
+ - [mlpack_gmm_probability](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## mlpack_gmm_generate
+{: #gmm_generate }
+
+#### GMM Sample Generator
+{: #gmm_generate_descr }
+
+```bash
+$ mlpack_gmm_generate [--help] [--info <string>] --input_model_file
+        <string> --samples 0 [--seed 0] [--verbose] [--version] [--output_file
+        <string>]
+```
+
+A sample generator for pre-trained GMMs.  Given a pre-trained GMM, this can sample new points randomly from that distribution. [Detailed documentation](#gmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`GMM file`](#doc_model) | Input GMM model to generate samples from. | `**--**` |
+| `--samples (-n)` | [`int`](#doc_int) | Number of samples to generate. | `**--**` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save output samples in. | 
+
+### Detailed documentation
+{: #gmm_generate_detailed-documentation }
+
+This program is able to generate samples from a pre-trained GMM (use gmm_train to train a GMM).  The pre-trained GMM must be specified with the `--input_model_file (-m)` parameter.  The number of samples to generate is specified by the `--samples (-n)` parameter.  Output samples may be saved with the `--output_file (-o)` output parameter.
+
+### Example
+The following command can be used to generate 100 samples from the pre-trained GMM `'gmm.bin'` and store those generated samples in `'samples.csv'`:
+
+```bash
+$ mlpack_gmm_generate --input_model_file gmm.bin --samples 100 --output_file
+  samples.csv
+```
+
+### See also
+
+ - [mlpack_gmm_train](#gmm_train)
+ - [mlpack_gmm_probability](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## mlpack_gmm_probability
+{: #gmm_probability }
+
+#### GMM Probability Calculator
+{: #gmm_probability_descr }
+
+```bash
+$ mlpack_gmm_probability [--help] [--info <string>] --input_file
+        <string> --input_model_file <string> [--verbose] [--version]
+        [--output_file <string>]
+```
+
+A probability calculator for GMMs.  Given a pre-trained GMM and a set of points, this can compute the probability that each point is from the given GMM. [Detailed documentation](#gmm_probability_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input matrix to calculate probabilities of. | `**--**` |
+| `--input_model_file (-m)` | [`GMM file`](#doc_model) | Input GMM to use as model. | `**--**` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to store calculated probabilities in. | 
+
+### Detailed documentation
+{: #gmm_probability_detailed-documentation }
+
+This program calculates the probability that given points came from a given GMM (that is, P(X \| gmm)).  The GMM is specified with the `--input_model_file (-m)` parameter, and the points are specified with the `--input_file (-i)` parameter.  The output probabilities may be saved via the `--output_file (-o)` output parameter.
+
+### Example
+So, for example, to calculate the probabilities of each point in `'points.csv'` coming from the pre-trained GMM `'gmm.bin'`, while storing those probabilities in `'probs.csv'`, the following command could be used:
+
+```bash
+$ mlpack_gmm_probability --input_model_file gmm.bin --input_file points.csv
+  --output_file probs.csv
+```
+
+### See also
+
+ - [mlpack_gmm_train](#gmm_train)
+ - [mlpack_gmm_generate](#gmm_generate)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## mlpack_hmm_train
+{: #hmm_train }
+
+#### Hidden Markov Model (HMM) Training
+{: #hmm_train_descr }
+
+```bash
+$ mlpack_hmm_train [--batch] [--gaussians 0] [--help] [--info <string>]
+        --input_file <string> [--input_model_file <string>] [--labels_file
+        <string>] [--seed 0] [--states 0] [--tolerance 1e-05] [--type
+        'gaussian'] [--verbose] [--version] [--output_model_file <string>]
+```
+
+An implementation of training algorithms for Hidden Markov Models (HMMs). Given labeled or unlabeled data, an HMM can be trained for further use with other mlpack HMM tools. [Detailed documentation](#hmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--batch (-b)` | [`flag`](#doc_flag) | If true, input_file (and if passed, labels_file) are expected to contain a list of files to use as input observation sequences (and label sequences). |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--gaussians (-g)` | [`int`](#doc_int) | Number of gaussians in each GMM (necessary when type is 'gmm'). | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`string`](#doc_string) | File containing input observations. | `**--**` |
+| `--input_model_file (-m)` | [`HMMModel file`](#doc_model) | Pre-existing HMM model to initialize training with. | `''` |
+| `--labels_file (-l)` | [`string`](#doc_string) | Optional file of hidden states, used for labeled training. | `''` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--states (-n)` | [`int`](#doc_int) | Number of hidden states in HMM (necessary, unless model_file is specified). | `0` |
+| `--tolerance (-T)` | [`double`](#doc_double) | Tolerance of the Baum-Welch algorithm. | `1e-05` |
+| `--type (-t)` | [`string`](#doc_string) | Type of HMM: discrete \| gaussian \| diag_gmm \| gmm. | `'gaussian'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`HMMModel file`](#doc_model) | Output for trained HMM. | 
+
+### Detailed documentation
+{: #hmm_train_detailed-documentation }
+
+This program allows a Hidden Markov Model to be trained on labeled or unlabeled data.  It supports four types of HMMs: Discrete HMMs, Gaussian HMMs, GMM HMMs, or Diagonal GMM HMMs
+
+Either one input sequence can be specified (with `--input_file (-i)`), or, a file containing files in which input sequences can be found (when `--input_file (-i)`and`--batch (-b)` are used together).  In addition, labels can be provided in the file specified by `--labels_file (-l)`, and if `--batch (-b)` is used, the file given to `--labels_file (-l)` should contain a list of files of labels corresponding to the sequences in the file given to `--input_file (-i)`.
+
+The HMM is trained with the Baum-Welch algorithm if no labels are provided.  The tolerance of the Baum-Welch algorithm can be set with the `--tolerance (-T)`option.  By default, the transition matrix is randomly initialized and the emission distributions are initialized to fit the extent of the data.
+
+Optionally, a pre-created HMM model can be used as a guess for the transition matrix and emission probabilities; this is specifiable with `--output_model_file (-M)`.
+
+### See also
+
+ - [mlpack_hmm_generate](#hmm_generate)
+ - [mlpack_hmm_loglik](#hmm_loglik)
+ - [mlpack_hmm_viterbi](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## mlpack_hmm_generate
+{: #hmm_generate }
+
+#### Hidden Markov Model (HMM) Sequence Generator
+{: #hmm_generate_descr }
+
+```bash
+$ mlpack_hmm_generate [--help] [--info <string>] --length 0 --model_file
+        <string> [--seed 0] [--start_state 0] [--verbose] [--version]
+        [--output_file <string>] [--state_file <string>]
+```
+
+A utility to generate random sequences from a pre-trained Hidden Markov Model (HMM).  The length of the desired sequence can be specified, and a random sequence of observations is returned. [Detailed documentation](#hmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--length (-l)` | [`int`](#doc_int) | Length of sequence to generate. | `**--**` |
+| `--model_file (-m)` | [`HMMModel file`](#doc_model) | Trained HMM to generate sequences with. | `**--**` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--start_state (-t)` | [`int`](#doc_int) | Starting state of sequence. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save observation sequence to. | 
+| `--state_file (-S)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to save hidden state sequence to. | 
+
+### Detailed documentation
+{: #hmm_generate_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as the `--model_file (-m)` parameter, and generates a random observation sequence and hidden state sequence based on its parameters. The observation sequence may be saved with the `--output_file (-o)` output parameter, and the internal state  sequence may be saved with the `--state_file (-S)` output parameter.
+
+The state to start the sequence in may be specified with the `--start_state (-t)` parameter.
+
+### Example
+For example, to generate a sequence of length 150 from the HMM `'hmm.bin'` and save the observation sequence to `'observations.csv'` and the hidden state sequence to `'states.csv'`, the following command may be used: 
+
+```bash
+$ mlpack_hmm_generate --model_file hmm.bin --length 150 --output_file
+  observations.csv --state_file states.csv
+```
+
+### See also
+
+ - [mlpack_hmm_train](#hmm_train)
+ - [mlpack_hmm_loglik](#hmm_loglik)
+ - [mlpack_hmm_viterbi](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## mlpack_hmm_loglik
+{: #hmm_loglik }
+
+#### Hidden Markov Model (HMM) Sequence Log-Likelihood
+{: #hmm_loglik_descr }
+
+```bash
+$ mlpack_hmm_loglik [--help] [--info <string>] --input_file <string>
+        --input_model_file <string> [--verbose] [--version] [--log_likelihood
+        0]
+```
+
+A utility for computing the log-likelihood of a sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observation sequence, this computes and returns the log-likelihood of that sequence being observed from that HMM. [Detailed documentation](#hmm_loglik_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | File containing observations, | `**--**` |
+| `--input_model_file (-m)` | [`HMMModel file`](#doc_model) | File containing HMM. | `**--**` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--log_likelihood` | [`double`](#doc_double) | Log-likelihood of the sequence. | 
+
+### Detailed documentation
+{: #hmm_loglik_detailed-documentation }
+
+This utility takes an already-trained HMM, specified with the `--input_model_file (-m)` parameter, and evaluates the log-likelihood of a sequence of observations, given with the `--input_file (-i)` parameter.  The computed log-likelihood is given as output.
+
+### Example
+For example, to compute the log-likelihood of the sequence `'seq.csv'` with the pre-trained HMM `'hmm.bin'`, the following command may be used: 
+
+```bash
+$ mlpack_hmm_loglik --input_file seq.csv --input_model_file hmm.bin
+```
+
+### See also
+
+ - [mlpack_hmm_train](#hmm_train)
+ - [mlpack_hmm_generate](#hmm_generate)
+ - [mlpack_hmm_viterbi](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## mlpack_hmm_viterbi
+{: #hmm_viterbi }
+
+#### Hidden Markov Model (HMM) Viterbi State Prediction
+{: #hmm_viterbi_descr }
+
+```bash
+$ mlpack_hmm_viterbi [--help] [--info <string>] --input_file <string>
+        --input_model_file <string> [--verbose] [--version] [--output_file
+        <string>]
+```
+
+A utility for computing the most probable hidden state sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observed sequence, this uses the Viterbi algorithm to compute and return the most probable hidden state sequence. [Detailed documentation](#hmm_viterbi_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing observations, | `**--**` |
+| `--input_model_file (-m)` | [`HMMModel file`](#doc_model) | Trained HMM to use. | `**--**` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | File to save predicted state sequence to. | 
+
+### Detailed documentation
+{: #hmm_viterbi_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as `--input_model_file (-m)`, and evaluates the most probable hidden state sequence of a given sequence of observations (specified as '`--input_file (-i)`, using the Viterbi algorithm.  The computed state sequence may be saved using the `--output_file (-o)` output parameter.
+
+### Example
+For example, to predict the state sequence of the observations `'obs.csv'` using the HMM `'hmm.bin'`, storing the predicted state sequence to `'states.csv'`, the following command could be used:
+
+```bash
+$ mlpack_hmm_viterbi --input_file obs.csv --input_model_file hmm.bin
+  --output_file states.csv
+```
+
+### See also
+
+ - [mlpack_hmm_train](#hmm_train)
+ - [mlpack_hmm_generate](#hmm_generate)
+ - [mlpack_hmm_loglik](#hmm_loglik)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## mlpack_hoeffding_tree
+{: #hoeffding_tree }
+
+#### Hoeffding trees
+{: #hoeffding_tree_descr }
+
+```bash
+$ mlpack_hoeffding_tree [--batch_mode] [--bins 10] [--confidence 0.95]
+        [--help] [--info <string>] [--info_gain] [--input_model_file <string>]
+        [--labels_file <string>] [--max_samples 5000] [--min_samples 100]
+        [--numeric_split_strategy 'binary'] [--observations_before_binning 100]
+        [--passes 1] [--test_file <string>] [--test_labels_file <string>]
+        [--training_file <string>] [--verbose] [--version] [--output_model_file
+        <string>] [--predictions_file <string>] [--probabilities_file <string>]
+```
+
+An implementation of Hoeffding trees, a form of streaming decision tree for classification.  Given labeled data, a Hoeffding tree can be trained and saved for later use, or a pre-trained Hoeffding tree can be used for predicting the classifications of new points. [Detailed documentation](#hoeffding_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--batch_mode (-b)` | [`flag`](#doc_flag) | If true, samples will be considered in batch instead of as a stream.  This generally results in better trees but at the cost of memory usage and runtime. |  |
+| `--bins (-B)` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of bins for each numeric split. | `10` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--confidence (-c)` | [`double`](#doc_double) | Confidence before splitting (between 0 and 1). | `0.95` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--info_gain (-i)` | [`flag`](#doc_flag) | If set, information gain is used instead of Gini impurity for calculating Hoeffding bounds. |  |
+| `--input_model_file (-m)` | [`HoeffdingTreeModel file`](#doc_model) | Input trained Hoeffding tree model. | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels for training dataset. | `''` |
+| `--max_samples (-n)` | [`int`](#doc_int) | Maximum number of samples before splitting. | `5000` |
+| `--min_samples (-I)` | [`int`](#doc_int) | Minimum number of samples before splitting. | `100` |
+| `--numeric_split_strategy (-N)` | [`string`](#doc_string) | The splitting strategy to use for numeric features: 'domingos' or 'binary'. | `'binary'` |
+| `--observations_before_binning (-o)` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of samples observed before binning is performed. | `100` |
+| `--passes (-s)` | [`int`](#doc_int) | Number of passes to take over the dataset. | `1` |
+| `--test_file (-T)` | [`2-d categorical matrix file`](#doc_a_2_d_categorical_matrix_file) | Testing dataset (may be categorical). | `''` |
+| `--test_labels_file (-L)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels of test data. | `''` |
+| `--training_file (-t)` | [`2-d categorical matrix file`](#doc_a_2_d_categorical_matrix_file) | Training dataset (may be categorical). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`HoeffdingTreeModel file`](#doc_model) | Output for trained Hoeffding tree model. | 
+| `--predictions_file (-p)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Matrix to output label predictions for test data into. | 
+| `--probabilities_file (-P)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | In addition to predicting labels, provide rediction probabilities in this matrix. | 
+
+### Detailed documentation
+{: #hoeffding_tree_detailed-documentation }
+
+This program implements Hoeffding trees, a form of streaming decision tree suited best for large (or streaming) datasets.  This program supports both categorical and numeric data.  Given an input dataset, this program is able to train the tree with numerous training options, and save the model to a file.  The program is also able to use a trained model or a model from file in order to predict classes for a given test set.
+
+The training file and associated labels are specified with the `--training_file (-t)` and `--labels_file (-l)` parameters, respectively. Optionally, if `--labels_file (-l)` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+The training may be performed in batch mode (like a typical decision tree algorithm) by specifying the `--batch_mode (-b)` option, but this may not be the best option for large datasets.
+
+When a model is trained, it may be saved via the `--output_model_file (-M)` output parameter.  A model may be loaded from file for further training or testing with the `--input_model_file (-m)` parameter.
+
+Test data may be specified with the `--test_file (-T)` parameter, and if performance statistics are desired for that test set, labels may be specified with the `--test_labels_file (-L)` parameter.  Predictions for each test point may be saved with the `--predictions_file (-p)` output parameter, and class probabilities for each prediction may be saved with the `--probabilities_file (-P)` output parameter.
+
+### Example
+For example, to train a Hoeffding tree with confidence 0.99 with data `'dataset.csv'`, saving the trained tree to `'tree.bin'`, the following command may be used:
+
+```bash
+$ mlpack_hoeffding_tree --training_file dataset.arff --confidence 0.99
+  --output_model_file tree.bin
+```
+
+Then, this tree may be used to make predictions on the test set `'test_set.csv'`, saving the predictions into `'predictions.csv'` and the class probabilities into `'class_probs.csv'` with the following command: 
+
+```bash
+$ mlpack_hoeffding_tree --input_model_file tree.bin --test_file test_set.arff
+  --predictions_file predictions.csv --probabilities_file class_probs.csv
+```
+
+### See also
+
+ - [mlpack_decision_tree](#decision_tree)
+ - [mlpack_random_forest](#random_forest)
+ - [Mining High-Speed Data Streams (pdf)](http://dm.cs.washington.edu/papers/vfdt-kdd00.pdf)
+ - [HoeffdingTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp)
+
+## mlpack_kde
+{: #kde }
+
+#### Kernel Density Estimation
+{: #kde_descr }
+
+```bash
+$ mlpack_kde [--abs_error 0] [--algorithm 'dual-tree'] [--bandwidth 1]
+        [--help] [--info <string>] [--initial_sample_size 100]
+        [--input_model_file <string>] [--kernel 'gaussian'] [--mc_break_coef
+        0.4] [--mc_entry_coef 3] [--mc_probability 0.95] [--monte_carlo]
+        [--query_file <string>] [--reference_file <string>] [--rel_error 0.05]
+        [--tree 'kd-tree'] [--verbose] [--version] [--output_model_file
+        <string>] [--predictions_file <string>]
+```
+
+An implementation of kernel density estimation with dual-tree algorithms. Given a set of reference points and query points and a kernel function, this can estimate the density function at the location of each query point using trees; trees that are built can be saved for later use. [Detailed documentation](#kde_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--abs_error (-E)` | [`double`](#doc_double) | Relative error tolerance for the prediction. | `0` |
+| `--algorithm (-a)` | [`string`](#doc_string) | Algorithm to use for the prediction.('dual-tree', 'single-tree'). | `'dual-tree'` |
+| `--bandwidth (-b)` | [`double`](#doc_double) | Bandwidth of the kernel. | `1` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--initial_sample_size (-s)` | [`int`](#doc_int) | Initial sample size for Monte Carlo estimations. | `100` |
+| `--input_model_file (-m)` | [`KDEModel file`](#doc_model) | Contains pre-trained KDE model. | `''` |
+| `--kernel (-k)` | [`string`](#doc_string) | Kernel to use for the prediction.('gaussian', 'epanechnikov', 'laplacian', 'spherical', 'triangular'). | `'gaussian'` |
+| `--mc_break_coef (-c)` | [`double`](#doc_double) | Controls what fraction of the amount of node's descendants is the limit for the sample size before it recurses. | `0.4` |
+| `--mc_entry_coef (-C)` | [`double`](#doc_double) | Controls how much larger does the amount of node descendants has to be compared to the initial sample size in order to be a candidate for Monte Carlo estimations. | `3` |
+| `--mc_probability (-P)` | [`double`](#doc_double) | Probability of the estimation being bounded by relative error when using Monte Carlo estimations. | `0.95` |
+| `--monte_carlo (-S)` | [`flag`](#doc_flag) | Whether to use Monte Carlo estimations when possible. |  |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Query dataset to KDE on. | `''` |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input reference dataset use for KDE. | `''` |
+| `--rel_error (-e)` | [`double`](#doc_double) | Relative error tolerance for the prediction. | `0.05` |
+| `--tree (-t)` | [`string`](#doc_string) | Tree to use for the prediction.('kd-tree', 'ball-tree', 'cover-tree', 'octree', 'r-tree'). | `'kd-tree'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`KDEModel file`](#doc_model) | If specified, the KDE model will be saved here. | 
+| `--predictions_file (-p)` | [`1-d matrix file`](#doc_a_1_d_matrix_file) | Vector to store density predictions. | 
+
+### Detailed documentation
+{: #kde_detailed-documentation }
+
+This program performs a Kernel Density Estimation. KDE is a non-parametric way of estimating probability density function. For each query point the program will estimate its probability density by applying a kernel function to each reference point. The computational complexity of this is O(N^2) where there are N query points and N reference points, but this implementation will typically see better performance as it uses an approximate dual or single tree algorithm for acceleration.
+
+Dual or single tree optimization avoids many barely relevant calculations (as kernel function values decrease with distance), so it is an approximate computation. You can specify the maximum relative error tolerance for each query value with `--rel_error (-e)` as well as the maximum absolute error tolerance with the parameter `--abs_error (-E)`. This program runs using an Euclidean metric. Kernel function can be selected using the `--kernel (-k)` option. You can also choose what which type of tree to use for the dual-tree algorithm with `--tree (-t)`. It is also possible to select whether to use dual-tree algorithm or single-tree algorithm using the `--algorithm (-a)` option.
+
+Monte Carlo estimations can be used to accelerate the KDE estimate when the Gaussian Kernel is used. This provides a probabilistic guarantee on the the error of the resulting KDE instead of an absolute guarantee.To enable Monte Carlo estimations, the `--monte_carlo (-S)` flag can be used, and success probability can be set with the `--mc_probability (-P)` option. It is possible to set the initial sample size for the Monte Carlo estimation using `--initial_sample_size (-s)`. This implementation will only consider a node, as a candidate for the Monte Carlo estimation, if its number of descendant nodes is bigger than the initial sample size. This can be controlled using a coefficient that will multiply the initial sample size and can be set using `--mc_entry_coef (-C)`. To avoid using the same amount of computations an exact approach would take, this program recurses the tree whenever a fraction of the amount of the node's descendant points have already been computed. This fraction is set using `--mc_break_coef (-c)`.
+
+### Example
+For example, the following will run KDE using the data in `'ref_data.csv'` for training and the data in `'qu_data.csv'` as query data. It will apply an Epanechnikov kernel with a 0.2 bandwidth to each reference point and use a KD-Tree for the dual-tree optimization. The returned predictions will be within 5% of the real KDE value for each query point.
+
+```bash
+$ mlpack_kde --reference_file ref_data.csv --query_file qu_data.csv
+  --bandwidth 0.2 --kernel epanechnikov --tree kd-tree --rel_error 0.05
+  --predictions_file out_data.csv
+```
+
+the predicted density estimations will be stored in `'out_data.csv'`.
+If no `--query_file (-q)` is provided, then KDE will be computed on the `--reference_file (-r)` dataset.
+It is possible to select either a reference dataset or an input model but not both at the same time. If an input model is selected and parameter values are not set (e.g. `--bandwidth (-b)`) then default parameter values will be used.
+
+In addition to the last program call, it is also possible to activate Monte Carlo estimations if a Gaussian kernel is used. This can provide faster results, but the KDE will only have a probabilistic guarantee of meeting the desired error bound (instead of an absolute guarantee). The following example will run KDE using a Monte Carlo estimation when possible. The results will be within a 5% of the real KDE value with a 95% probability. Initial sample size for the Monte Carlo estimation will be 200 points and a node will be a candidate for the estimation only when it contains 700 (i.e. 3.5*200) points. If a node contains 700 points and 420 (i.e. 0.6*700) have already been sampled, then the algorithm will recurse instead of keep sampling.
+
+```bash
+$ mlpack_kde --reference_file ref_data.csv --query_file qu_data.csv
+  --bandwidth 0.2 --kernel gaussian --tree kd-tree --rel_error 0.05
+  --predictions_file out_data.csv --monte_carlo --mc_probability 0.95
+  --initial_sample_size 200 --mc_entry_coef 3.5 --mc_break_coef 0.6
+```
+
+### See also
+
+ - [mlpack_knn](#knn)
+ - [Kernel density estimation on Wikipedia](https://en.wikipedia.org/wiki/Kernel_density_estimation)
+ - [Tree-Independent Dual-Tree Algorithms](https://arxiv.org/pdf/1304.4327)
+ - [Fast High-dimensional Kernel Summations Using the Monte Carlo Multipole Method](http://papers.nips.cc/paper/3539-fast-high-dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf)
+ - [KDE C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kde/kde.hpp)
+
+## mlpack_kernel_pca
+{: #kernel_pca }
+
+#### Kernel Principal Components Analysis
+{: #kernel_pca_descr }
+
+```bash
+$ mlpack_kernel_pca [--bandwidth 1] [--center] [--degree 1] [--help]
+        [--info <string>] --input_file <string> --kernel <string>
+        [--kernel_scale 1] [--new_dimensionality 0] [--nystroem_method]
+        [--offset 0] [--sampling 'kmeans'] [--verbose] [--version]
+        [--output_file <string>]
+```
+
+An implementation of Kernel Principal Components Analysis (KPCA).  This can be used to perform nonlinear dimensionality reduction or preprocessing on a given dataset. [Detailed documentation](#kernel_pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--bandwidth (-b)` | [`double`](#doc_double) | Bandwidth, for 'gaussian' and 'laplacian' kernels. | `1` |
+| `--center (-c)` | [`flag`](#doc_flag) | If set, the transformed data will be centered about the origin. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--degree (-D)` | [`double`](#doc_double) | Degree of polynomial, for 'polynomial' kernel. | `1` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform KPCA on. | `**--**` |
+| `--kernel (-k)` | [`string`](#doc_string) | The kernel to use; see the above documentation for the list of usable kernels. | `**--**` |
+| `--kernel_scale (-S)` | [`double`](#doc_double) | Scale, for 'hyptan' kernel. | `1` |
+| `--new_dimensionality (-d)` | [`int`](#doc_int) | If not 0, reduce the dimensionality of the output dataset by ignoring the dimensions with the smallest eigenvalues. | `0` |
+| `--nystroem_method (-n)` | [`flag`](#doc_flag) | If set, the Nystroem method will be used. |  |
+| `--offset (-O)` | [`double`](#doc_double) | Offset, for 'hyptan' and 'polynomial' kernels. | `0` |
+| `--sampling (-s)` | [`string`](#doc_string) | Sampling scheme to use for the Nystroem method: 'kmeans', 'random', 'ordered' | `'kmeans'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #kernel_pca_detailed-documentation }
+
+This program performs Kernel Principal Components Analysis (KPCA) on the specified dataset with the specified kernel.  This will transform the data onto the kernel principal components, and optionally reduce the dimensionality by ignoring the kernel principal components with the smallest eigenvalues.
+
+For the case where a linear kernel is used, this reduces to regular PCA.
+
+The kernels that are supported are listed below:
+
+ * 'linear': the standard linear dot product (same as normal PCA):
+    `K(x, y) = x^T y`
+
+ * 'gaussian': a Gaussian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\| ^ 2) / (2 * (bandwidth ^ 2)))`
+
+ * 'polynomial': polynomial kernel; requires offset and degree:
+    `K(x, y) = (x^T y + offset) ^ degree`
+
+ * 'hyptan': hyperbolic tangent kernel; requires scale and offset:
+    `K(x, y) = tanh(scale * (x^T y) + offset)`
+
+ * 'laplacian': Laplacian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\|) / bandwidth)`
+
+ * 'epanechnikov': Epanechnikov kernel; requires bandwidth:
+    `K(x, y) = max(0, 1 - \|\| x - y \|\|^2 / bandwidth^2)`
+
+ * 'cosine': cosine distance:
+    `K(x, y) = 1 - (x^T y) / (\|\| x \|\| * \|\| y \|\|)`
+
+The parameters for each of the kernels should be specified with the options `--bandwidth (-b)`, `--kernel_scale (-S)`, `--offset (-O)`, or `--degree (-D)` (or a combination of those parameters).
+
+Optionally, the Nystroem method ("Using the Nystroem method to speed up kernel machines", 2001) can be used to calculate the kernel matrix by specifying the `--nystroem_method (-n)` parameter. This approach works by using a subset of the data as basis to reconstruct the kernel matrix; to specify the sampling scheme, the `--sampling (-s)` parameter is used.  The sampling scheme for the Nystroem method can be chosen from the following list: 'kmeans', 'random', 'ordered'.
+
+### Example
+For example, the following command will perform KPCA on the dataset `'input.csv'` using the Gaussian kernel, and saving the transformed data to `'transformed.csv'`: 
+
+```bash
+$ mlpack_kernel_pca --input_file input.csv --kernel gaussian --output_file
+  transformed.csv
+```
+
+### See also
+
+ - [Kernel principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Kernel_principal_component_analysis)
+ - [Kernel Principal Component Analysis (pdf)](https://pca.narod.ru/scholkopf_kernel.pdf)
+ - [KernelPCA class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+
+## mlpack_kmeans
+{: #kmeans }
+
+#### K-Means Clustering
+{: #kmeans_descr }
+
+```bash
+$ mlpack_kmeans [--algorithm 'naive'] [--allow_empty_clusters]
+        --clusters 0 [--help] [--in_place] [--info <string>]
+        [--initial_centroids_file <string>] --input_file <string>
+        [--kill_empty_clusters] [--kmeans_plus_plus] [--labels_only]
+        [--max_iterations 1000] [--percentage 0.02] [--refined_start]
+        [--samplings 100] [--seed 0] [--verbose] [--version] [--centroid_file
+        <string>] [--output_file <string>]
+```
+
+An implementation of several strategies for efficient k-means clustering. Given a dataset and a value of k, this computes and returns a k-means clustering on that data. [Detailed documentation](#kmeans_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--algorithm (-a)` | [`string`](#doc_string) | Algorithm to use for the Lloyd iteration ('naive', 'pelleg-moore', 'elkan', 'hamerly', 'dualtree', or 'dualtree-covertree'). | `'naive'` |
+| `--allow_empty_clusters (-e)` | [`flag`](#doc_flag) | Allow empty clusters to be persist. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--clusters (-c)` | [`int`](#doc_int) | Number of clusters to find (0 autodetects from initial centroids). | `**--**` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--in_place (-P)` | [`flag`](#doc_flag) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden. (Do not use in Python.) |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--initial_centroids_file (-I)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Start with the specified initial centroids. | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform clustering on. | `**--**` |
+| `--kill_empty_clusters (-E)` | [`flag`](#doc_flag) | Remove empty clusters when they occur. |  |
+| `--kmeans_plus_plus (-K)` | [`flag`](#doc_flag) | Use the k-means++ initialization strategy to choose initial points. |  |
+| `--labels_only (-l)` | [`flag`](#doc_flag) | Only output labels into output file. |  |
+| `--max_iterations (-m)` | [`int`](#doc_int) | Maximum number of iterations before k-means terminates. | `1000` |
+| `--percentage (-p)` | [`double`](#doc_double) | Percentage of dataset to use for each refined start sampling (use when --refined_start is specified). | `0.02` |
+| `--refined_start (-r)` | [`flag`](#doc_flag) | Use the refined initial point strategy by Bradley and Fayyad to choose initial points. |  |
+| `--samplings (-S)` | [`int`](#doc_int) | Number of samplings to perform for refined start (use when --refined_start is specified). | `100` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--centroid_file (-C)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If specified, the centroids of each cluster will  be written to the given file. | 
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to store output labels or labeled data to. | 
+
+### Detailed documentation
+{: #kmeans_detailed-documentation }
+
+This program performs K-Means clustering on the given dataset.  It can return the learned cluster assignments, and the centroids of the clusters.  Empty clusters are not allowed by default; when a cluster becomes empty, the point furthest from the centroid of the cluster with maximum variance is taken to fill that cluster.
+
+Optionally, the strategy to choose initial centroids can be specified.  The k-means++ algorithm can be used to choose initial centroids with the `--kmeans_plus_plus (-K)` parameter.  The Bradley and Fayyad approach ("Refining initial points for k-means clustering", 1998) can be used to select initial points by specifying the `--refined_start (-r)` parameter.  This approach works by taking random samplings of the dataset; to specify the number of samplings, the `--samplings (-S)` parameter is used, and to specify the percentage of the dataset to be used in each sample, the `--percentage (-p)` parameter is used (it should be a value between 0.0 and 1.0).
+
+There are several options available for the algorithm used for each Lloyd iteration, specified with the `--algorithm (-a)`  option.  The standard O(kN) approach can be used ('naive').  Other options include the Pelleg-Moore tree-based algorithm ('pelleg-moore'), Elkan's triangle-inequality based algorithm ('elkan'), Hamerly's modification to Elkan's algorithm ('hamerly'), the dual-tree k-means algorithm ('dualtree'), and the dual-tree k-means algorithm using the cover tree ('dualtree-covertree').
+
+The behavior for when an empty cluster is encountered can be modified with the `--allow_empty_clusters (-e)` option.  When this option is specified and there is a cluster owning no points at the end of an iteration, that cluster's centroid will simply remain in its position from the previous iteration. If the `--kill_empty_clusters (-E)` option is specified, then when a cluster owns no points at the end of an iteration, the cluster centroid is simply filled with DBL_MAX, killing it and effectively reducing k for the rest of the computation.  Note that the default option when neither empty cluster option is specified can be time-consuming to calculate; therefore, specifying either of these parameters will often accelerate runtime.
+
+Initial clustering assignments may be specified using the `--initial_centroids_file (-I)` parameter, and the maximum number of iterations may be specified with the `--max_iterations (-m)` parameter.
+
+### Example
+As an example, to use Hamerly's algorithm to perform k-means clustering with k=10 on the dataset `'data.csv'`, saving the centroids to `'centroids.csv'` and the assignments for each point to `'assignments.csv'`, the following command could be used:
+
+```bash
+$ mlpack_kmeans --input_file data.csv --clusters 10 --output_file
+  assignments.csv --centroid_file centroids.csv
+```
+
+To run k-means on that same dataset with initial centroids specified in `'initial.csv'` with a maximum of 500 iterations, storing the output centroids in `'final.csv'` the following command may be used:
+
+```bash
+$ mlpack_kmeans --input_file data.csv --initial_centroids_file initial.csv
+  --clusters 10 --max_iterations 500 --centroid_file final.csv
+```
+
+### See also
+
+ - [mlpack_dbscan](#dbscan)
+ - [k-means++](https://en.wikipedia.org/wiki/K-means%2B%2B)
+ - [Using the triangle inequality to accelerate k-means (pdf)](https://cdn.aaai.org/ICML/2003/ICML03-022.pdf)
+ - [Making k-means even faster (pdf)](https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12)
+ - [Accelerating exact k-means algorithms with geometric reasoning (pdf)](http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/ftp/usr0/ftp/2000/CMU-CS-00-105.pdf)
+ - [A dual-tree algorithm for fast k-means clustering with large k (pdf)](http://www.ratml.org/pub/pdf/2017dual.pdf)
+ - [KMeans class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kmeans/kmeans.hpp)
+
+## mlpack_lars
+{: #lars }
+
+#### LARS
+{: #lars_descr }
+
+```bash
+$ mlpack_lars [--help] [--info <string>] [--input_file <string>]
+        [--input_model_file <string>] [--lambda1 0] [--lambda2 0]
+        [--no_intercept] [--no_normalize] [--responses_file <string>]
+        [--test_file <string>] [--use_cholesky] [--verbose] [--version]
+        [--output_model_file <string>] [--output_predictions_file <string>]
+```
+
+An implementation of Least Angle Regression (Stagewise/laSso), also known as LARS.  This can train a LARS/LASSO/Elastic Net model and use that model or a pre-trained model to output regression predictions for a test set. [Detailed documentation](#lars_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of covariates (X). | `''` |
+| `--input_model_file (-m)` | [`LARS<> file`](#doc_model) | Trained LARS model to use. | `''` |
+| `--lambda1 (-l)` | [`double`](#doc_double) | Regularization parameter for l1-norm penalty. | `0` |
+| `--lambda2 (-L)` | [`double`](#doc_double) | Regularization parameter for l2-norm penalty. | `0` |
+| `--no_intercept (-n)` | [`flag`](#doc_flag) | Do not fit an intercept in the model. |  |
+| `--no_normalize (-N)` | [`flag`](#doc_flag) | Do not normalize data to unit variance before modeling. |  |
+| `--responses_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of responses/observations (y). | `''` |
+| `--test_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing points to regress on (test points). | `''` |
+| `--use_cholesky (-c)` | [`flag`](#doc_flag) | Use Cholesky decomposition during computation rather than explicitly computing the full Gram matrix. |  |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`LARS<> file`](#doc_model) | Output LARS model. | 
+| `--output_predictions_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #lars_detailed-documentation }
+
+An implementation of LARS: Least Angle Regression (Stagewise/laSso).  This is a stage-wise homotopy-based algorithm for L1-regularized linear regression (LASSO) and L1+L2-regularized linear regression (Elastic Net).
+
+This program is able to train a LARS/LASSO/Elastic Net model or load a model from file, output regression predictions for a test set, and save the trained model to a file.  The LARS algorithm is described in more detail below:
+
+Let X be a matrix where each row is a point and each column is a dimension, and let y be a vector of targets.
+
+The Elastic Net problem is to solve
+
+  min_beta 0.5 \|\| X * beta - y \|\|_2^2 + lambda_1 \|\|beta\|\|_1 +
+      0.5 lambda_2 \|\|beta\|\|_2^2
+
+If lambda1 > 0 and lambda2 = 0, the problem is the LASSO.
+If lambda1 > 0 and lambda2 > 0, the problem is the Elastic Net.
+If lambda1 = 0 and lambda2 > 0, the problem is ridge regression.
+If lambda1 = 0 and lambda2 = 0, the problem is unregularized linear regression.
+
+For efficiency reasons, it is not recommended to use this algorithm with `--lambda1 (-l)` = 0.  In that case, use the 'linear_regression' program, which implements both unregularized linear regression and ridge regression.
+
+To train a LARS/LASSO/Elastic Net model, the `--input_file (-i)` and `--responses_file (-r)` parameters must be given.  The `--lambda1 (-l)`, `--lambda2 (-L)`, and `--use_cholesky (-c)` parameters control the training options.  A trained model can be saved with the `--output_model_file (-M)`.  If no training is desired at all, a model can be passed via the `--input_model_file (-m)` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `--test_file (-t)` parameter.  Predicted responses to the test points can be saved with the `--output_predictions_file (-o)` output parameter.
+
+### Example
+For example, the following command trains a model on the data `'data.csv'` and responses `'responses.csv'` with lambda1 set to 0.4 and lambda2 set to 0 (so, LASSO is being solved), and then the model is saved to `'lasso_model.bin'`:
+
+```bash
+$ mlpack_lars --input_file data.csv --responses_file responses.csv --lambda1
+  0.4 --lambda2 0 --output_model_file lasso_model.bin
+```
+
+The following command uses the `'lasso_model.bin'` to provide predicted responses for the data `'test.csv'` and save those responses to `'test_predictions.csv'`: 
+
+```bash
+$ mlpack_lars --input_model_file lasso_model.bin --test_file test.csv
+  --output_predictions_file test_predictions.csv
+```
+
+### See also
+
+ - [mlpack_linear_regression](#linear_regression)
+ - [Least angle regression (pdf)](https://mlpack.org/papers/lars.pdf)
+ - [LARS C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lars/lars.hpp)
+
+## mlpack_linear_svm
+{: #linear_svm }
+
+#### Linear SVM is an L2-regularized support vector machine.
+{: #linear_svm_descr }
+
+```bash
+$ mlpack_linear_svm [--delta 1] [--epochs 50] [--help] [--info <string>]
+        [--input_model_file <string>] [--labels_file <string>] [--lambda 0.0001]
+        [--max_iterations 10000] [--no_intercept] [--num_classes 0] [--optimizer
+        'lbfgs'] [--seed 0] [--shuffle] [--step_size 0.01] [--test_file
+        <string>] [--test_labels_file <string>] [--tolerance 1e-10]
+        [--training_file <string>] [--verbose] [--version] [--output_model_file
+        <string>] [--predictions_file <string>] [--probabilities_file <string>]
+```
+
+An implementation of linear SVM for multiclass classification. Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#linear_svm_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--delta (-d)` | [`double`](#doc_double) | Margin of difference between correct class and other classes. | `1` |
+| `--epochs (-E)` | [`int`](#doc_int) | Maximum number of full epochs over dataset for psgd | `50` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`LinearSVMModel file`](#doc_model) | Existing model (parameters). | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | A matrix containing labels (0 or 1) for the points in the training set (y). | `''` |
+| `--lambda (-r)` | [`double`](#doc_double) | L2-regularization parameter for training. | `0.0001` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `--no_intercept (-N)` | [`flag`](#doc_flag) | Do not add the intercept term to the model. |  |
+| `--num_classes (-c)` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `--optimizer (-O)` | [`string`](#doc_string) | Optimizer to use for training ('lbfgs' or 'psgd'). | `'lbfgs'` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--shuffle (-S)` | [`flag`](#doc_flag) | Don't shuffle the order in which data points are visited for parallel SGD. |  |
+| `--step_size (-a)` | [`double`](#doc_double) | Step size for parallel SGD optimizer. | `0.01` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing test dataset. | `''` |
+| `--test_labels_file (-L)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Matrix containing test labels. | `''` |
+| `--tolerance (-e)` | [`double`](#doc_double) | Convergence tolerance for optimizer. | `1e-10` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the training set (the matrix of predictors, X). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`LinearSVMModel file`](#doc_model) | Output for trained linear svm model. | 
+| `--predictions_file (-P)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `--probabilities_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #linear_svm_detailed-documentation }
+
+An implementation of linear SVMs that uses either L-BFGS or parallel SGD (stochastic gradient descent) to train the model.
+
+This program allows loading a linear SVM model (via the `--input_model_file (-m)` parameter) or training a linear SVM model given training data (specified with the `--training_file (-t)` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `--test_file (-T)` parameter) and the classification results may be saved with the `--predictions_file (-P)` output parameter. The trained linear SVM model may be saved using the `--output_model_file (-M)` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `--labels_file (-l)` parameter may be used to specify a separate vector of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `--lambda (-r)` option, and the number of classes can be manually specified with the `--num_classes (-c)`and if an intercept term is not desired in the model, the `--no_intercept (-N)` parameter can be specified.Margin of difference between correct class and other classes can be specified with the `--delta (-d)` option.The optimizer used to train the model can be specified with the `--optimizer (-O)` parameter.  Available options are 'psgd' (parallel stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `--max_iterations (-n)` parameter specifies the maximum number of allowed iterations, and the `--tolerance (-e)` parameter specifies the tolerance for convergence.  For the parallel SGD optimizer, the `--step_size (-a)` parameter controls the step size taken at each iteration by the optimizer and the maximum number of epochs (specified with `--epochs (-E)`). If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+Optionally, the model can be used to predict the labels for another matrix of data points, if `--test_file (-T)` is specified.  The `--test_file (-T)` parameter can be specified without the `--training_file (-t)` parameter, so long as an existing linear SVM model is given with the `--input_model_file (-m)` parameter.  The output predictions from the linear SVM model may be saved with the `--predictions_file (-P)` parameter.
+
+### Example
+As an example, to train a LinaerSVM on the data '`'data.csv'`' with labels '`'labels.csv'`' with L2 regularization of 0.1, saving the model to '`'lsvm_model.bin'`', the following command may be used:
+
+```bash
+$ mlpack_linear_svm --training_file data.csv --labels_file labels.csv --lambda
+  0.1 --delta 1 --num_classes 0 --output_model_file lsvm_model.bin
+```
+
+Then, to use that model to predict classes for the dataset '`'test.csv'`', storing the output predictions in '`'predictions.csv'`', the following command may be used: 
+
+```bash
+$ mlpack_linear_svm --input_model_file lsvm_model.bin --test_file test.csv
+  --predictions_file predictions.csv
+```
+
+### See also
+
+ - [mlpack_random_forest](#random_forest)
+ - [mlpack_logistic_regression](#logistic_regression)
+ - [LinearSVM on Wikipedia](https://en.wikipedia.org/wiki/Support-vector_machine)
+ - [LinearSVM C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_svm/linear_svm.hpp)
+
+## mlpack_lmnn
+{: #lmnn }
+
+#### Large Margin Nearest Neighbors (LMNN)
+{: #lmnn_descr }
+
+```bash
+$ mlpack_lmnn [--batch_size 50] [--center] [--distance_file <string>]
+        [--help] [--info <string>] --input_file <string> [--k 1] [--labels_file
+        <string>] [--linear_scan] [--max_iterations 100000] [--normalize]
+        [--optimizer 'amsgrad'] [--passes 50] [--print_accuracy] [--range 1]
+        [--rank 0] [--regularization 0.5] [--seed 0] [--step_size 0.01]
+        [--tolerance 1e-07] [--verbose] [--version] [--centered_data_file
+        <string>] [--output_file <string>] [--transformed_data_file <string>]
+```
+
+An implementation of Large Margin Nearest Neighbors (LMNN), a distance learning technique.  Given a labeled dataset, this learns a transformation of the data that improves k-nearest-neighbor performance; this can be useful as a preprocessing step. [Detailed documentation](#lmnn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--batch_size (-b)` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `--center (-C)` | [`flag`](#doc_flag) | Perform mean-centering on the dataset. It is useful when the centroid of the data is far from the origin. |  |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--distance_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Initial distance matrix to be used as starting point | `''` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to run LMNN on. | `**--**` |
+| `--k (-k)` | [`int`](#doc_int) | Number of target neighbors to use for each datapoint. | `1` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels for input dataset. | `''` |
+| `--linear_scan (-L)` | [`flag`](#doc_flag) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. |  |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations for L-BFGS (0 indicates no limit). | `100000` |
+| `--normalize (-N)` | [`flag`](#doc_flag) | Use a normalized starting point for optimization. Itis useful for when points are far apart, or when SGD is returning NaN. |  |
+| `--optimizer (-O)` | [`string`](#doc_string) | Optimizer to use; 'amsgrad', 'bbsgd', 'sgd', or 'lbfgs'. | `'amsgrad'` |
+| `--passes (-p)` | [`int`](#doc_int) | Maximum number of full passes over dataset for AMSGrad, BB_SGD and SGD. | `50` |
+| `--print_accuracy (-P)` | [`flag`](#doc_flag) | Print accuracies on initial and transformed dataset |  |
+| `--range (-R)` | [`int`](#doc_int) | Number of iterations after which impostors needs to be recalculated | `1` |
+| `--rank (-A)` | [`int`](#doc_int) | Rank of distance matrix to be optimized.  | `0` |
+| `--regularization (-r)` | [`double`](#doc_double) | Regularization for LMNN objective function  | `0.5` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--step_size (-a)` | [`double`](#doc_double) | Step size for AMSGrad, BB_SGD and SGD (alpha). | `0.01` |
+| `--tolerance (-t)` | [`double`](#doc_double) | Maximum tolerance for termination of AMSGrad, BB_SGD, SGD or L-BFGS. | `1e-07` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--centered_data_file (-c)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output matrix for mean-centered dataset. | 
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output matrix for learned distance matrix. | 
+| `--transformed_data_file (-D)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output matrix for transformed dataset. | 
+
+### Detailed documentation
+{: #lmnn_detailed-documentation }
+
+This program implements Large Margin Nearest Neighbors, a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset.  The method employes the strategy of reducing distance between similar labeled data points (a.k.a target neighbors) and increasing distance between differently labeled points (a.k.a impostors) using standard optimization techniques over the gradient of the distance between data points.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `--input_file (-i)`), or alternatively as a separate matrix (specified with `--labels_file (-l)`).  Additionally, a starting point for optimization (specified with `--distance_file (-d)`can be given, having (r x d) dimensionality.  Here r should satisfy 1 <= r <= d, Consequently a Low-Rank matrix will be optimized. Alternatively, Low-Rank distance can be learned by specifying the `--rank (-A)`parameter (A Low-Rank matrix with uniformly distributed values will be used as initial learning point). 
+
+The program also requires number of targets neighbors to work with ( specified with `--k (-k)`), A regularization parameter can also be passed, It acts as a trade of between the pulling and pushing terms (specified with `--regularization (-r)`), In addition, this implementation of LMNN includes a parameter to decide the interval after which impostors must be re-calculated (specified with `--range (-R)`).
+
+Output can either be the learned distance matrix (specified with `--output_file (-o)`), or the transformed dataset  (specified with `--transformed_data_file (-D)`), or both. Additionally mean-centered dataset (specified with `--centered_data_file (-c)`) can be accessed given mean-centering (specified with `--center (-C)`) is performed on the dataset. Accuracy on initial dataset and final transformed dataset can be printed by specifying the `--print_accuracy (-P)`parameter. 
+
+This implementation of LMNN uses AdaGrad, BigBatch_SGD, stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer. 
+
+AdaGrad, specified by the value 'adagrad' for the parameter `--optimizer (-O)`, uses maximum of past squared gradients. It primarily on six parameters: the step size (specified with `--step_size (-a)`), the batch size (specified with `--batch_size (-b)`), the maximum number of passes (specified with `--passes (-p)`). Inaddition, a normalized starting point can be used by specifying the `--normalize (-N)` parameter. 
+
+BigBatch_SGD, specified by the value 'bbsgd' for the parameter `--optimizer (-O)`, depends primarily on four parameters: the step size (specified with `--step_size (-a)`), the batch size (specified with `--batch_size (-b)`), the maximum number of passes (specified with `--passes (-p)`).  In addition, a normalized starting point can be used by specifying the `--normalize (-N)` parameter. 
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `--optimizer (-O)`, depends primarily on three parameters: the step size (specified with `--step_size (-a)`), the batch size (specified with `--batch_size (-b)`), and the maximum number of passes (specified with `--passes (-p)`).  In addition, a normalized starting point can be used by specifying the `--normalize (-N)` parameter. Furthermore, mean-centering can be performed on the dataset by specifying the `--center (-C)`parameter. 
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `--optimizer (-O)`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `--max_iterations (-n)`, `--tolerance (-t)`(the optimization is terminated when the gradient norm is below this value).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.  In addition, a normalized starting point can be used by specifying the `--normalize (-N)` parameter.
+
+By default, the AMSGrad optimizer is used.
+
+### Example
+Example - Let's say we want to learn distance on iris dataset with number of targets as 3 using BigBatch_SGD optimizer. A simple call for the same will look like: 
+
+```bash
+$ mlpack_lmnn --input_file iris.csv --labels_file iris_labels.csv --k 3
+  --optimizer bbsgd --output_file output.csv
+```
+
+An another program call making use of range & regularization parameter with dataset having labels as last column can be made as: 
+
+```bash
+$ mlpack_lmnn --input_file letter_recognition.csv --k 5 --range 10
+  --regularization 0.4 --output_file output.csv
+```
+
+### See also
+
+ - [mlpack_nca](#nca)
+ - [Large margin nearest neighbor on Wikipedia](https://en.wikipedia.org/wiki/Large_margin_nearest_neighbor)
+ - [Distance metric learning for large margin nearest neighbor classification (pdf)](http://papers.nips.cc/paper/2795-distance-metric-learning-for-large-margin-nearest-neighbor-classification.pdf)
+ - [LMNN C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lmnn/lmnn.hpp)
+
+## mlpack_local_coordinate_coding
+{: #local_coordinate_coding }
+
+#### Local Coordinate Coding
+{: #local_coordinate_coding_descr }
+
+```bash
+$ mlpack_local_coordinate_coding [--atoms 0] [--help] [--info <string>]
+        [--initial_dictionary_file <string>] [--input_model_file <string>]
+        [--lambda 0] [--max_iterations 0] [--normalize] [--seed 0] [--test_file
+        <string>] [--tolerance 0.01] [--training_file <string>] [--verbose]
+        [--version] [--codes_file <string>] [--dictionary_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of Local Coordinate Coding (LCC), a data transformation technique.  Given input data, this transforms each point to be expressed as a linear combination of a few points in the dataset; once an LCC model is trained, it can be used to transform points later also. [Detailed documentation](#local_coordinate_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--atoms (-k)` | [`int`](#doc_int) | Number of atoms in the dictionary. | `0` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--initial_dictionary_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Optional initial dictionary. | `''` |
+| `--input_model_file (-m)` | [`LocalCoordinateCoding file`](#doc_model) | Input LCC model. | `''` |
+| `--lambda (-l)` | [`double`](#doc_double) | Weighted l1-norm regularization parameter. | `0` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations for LCC (0 indicates no limit). | `0` |
+| `--normalize (-N)` | [`flag`](#doc_flag) | If set, the input data matrix will be normalized before coding. |  |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Test points to encode. | `''` |
+| `--tolerance (-o)` | [`double`](#doc_double) | Tolerance for objective function. | `0.01` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of training data (X). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--codes_file (-c)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output codes matrix. | 
+| `--dictionary_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output dictionary matrix. | 
+| `--output_model_file (-M)` | [`LocalCoordinateCoding file`](#doc_model) | Output for trained LCC model. | 
+
+### Detailed documentation
+{: #local_coordinate_coding_detailed-documentation }
+
+An implementation of Local Coordinate Coding (LCC), which codes data that approximately lives on a manifold using a variation of l1-norm regularized sparse coding.  Given a dense data matrix X with n points and d dimensions, LCC seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a coding matrix Z with n points in k dimensions.  Because of the regularization method used, the atoms in D should lie close to the manifold on which the data points lie.
+
+The original data matrix X can then be reconstructed as D * Z.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a coding step, which updates the coding matrix Z.
+
+To run this program, the input matrix X must be specified (with -i), along with the number of atoms in the dictionary (-k).  An initial dictionary may also be specified with the `--initial_dictionary_file (-i)` parameter.  The l1-norm regularization parameter is specified with the `--lambda (-l)` parameter.
+
+### Example
+For example, to run LCC on the dataset `'data.csv'` using 200 atoms and an l1-regularization parameter of 0.1, saving the dictionary `--dictionary_file (-d)` and the codes into `--codes_file (-c)`, use
+
+```bash
+$ mlpack_local_coordinate_coding --training_file data.csv --atoms 200 --lambda
+  0.1 --dictionary_file dict.csv --codes_file codes.csv
+```
+
+The maximum number of iterations may be specified with the `--max_iterations (-n)` parameter. Optionally, the input data matrix X can be normalized before coding with the `--normalize (-N)` parameter.
+
+An LCC model may be saved using the `--output_model_file (-M)` output parameter.  Then, to encode new points from the dataset `'points.csv'` with the previously saved model `'lcc_model.bin'`, saving the new codes to `'new_codes.csv'`, the following command can be used:
+
+```bash
+$ mlpack_local_coordinate_coding --input_model_file lcc_model.bin --test_file
+  points.csv --codes_file new_codes.csv
+```
+
+### See also
+
+ - [mlpack_sparse_coding](#sparse_coding)
+ - [Nonlinear learning using local coordinate coding (pdf)](https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-coordinate-coding.pdf)
+ - [LocalCoordinateCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding.hpp)
+
+## mlpack_logistic_regression
+{: #logistic_regression }
+
+#### L2-regularized Logistic Regression and Prediction
+{: #logistic_regression_descr }
+
+```bash
+$ mlpack_logistic_regression [--batch_size 64] [--decision_boundary 0.5]
+        [--help] [--info <string>] [--input_model_file <string>] [--labels_file
+        <string>] [--lambda 0] [--max_iterations 10000] [--optimizer 'lbfgs']
+        [--print_training_accuracy] [--step_size 0.01] [--test_file <string>]
+        [--tolerance 1e-10] [--training_file <string>] [--verbose] [--version]
+        [--output_model_file <string>] [--predictions_file <string>]
+        [--probabilities_file <string>]
+```
+
+An implementation of L2-regularized logistic regression for two-class classification.  Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#logistic_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--batch_size (-b)` | [`int`](#doc_int) | Batch size for SGD. | `64` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--decision_boundary (-d)` | [`double`](#doc_double) | Decision boundary for prediction; if the logistic function for a point is less than the boundary, the class is taken to be 0; otherwise, the class is 1. | `0.5` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`LogisticRegression<> file`](#doc_model) | Existing model (parameters). | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | A matrix containing labels (0 or 1) for the points in the training set (y). | `''` |
+| `--lambda (-L)` | [`double`](#doc_double) | L2-regularization parameter for training. | `0` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `--optimizer (-O)` | [`string`](#doc_string) | Optimizer to use for training ('lbfgs' or 'sgd'). | `'lbfgs'` |
+| `--print_training_accuracy (-a)` | [`flag`](#doc_flag) | If set, then the accuracy of the model on the training set will be printed (verbose must also be specified). |  |
+| `--step_size (-s)` | [`double`](#doc_double) | Step size for SGD optimizer. | `0.01` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing test dataset. | `''` |
+| `--tolerance (-e)` | [`double`](#doc_double) | Convergence tolerance for optimizer. | `1e-10` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the training set (the matrix of predictors, X). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`LogisticRegression<> file`](#doc_model) | Output for trained logistic regression model. | 
+| `--predictions_file (-P)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `--probabilities_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #logistic_regression_detailed-documentation }
+
+An implementation of L2-regularized logistic regression using either the L-BFGS optimizer or SGD (stochastic gradient descent).  This solves the regression problem
+
+  y = (1 / 1 + e^-(X * b)).
+
+In this setting, y corresponds to class labels and X corresponds to data.
+
+This program allows loading a logistic regression model (via the `--input_model_file (-m)` parameter) or training a logistic regression model given training data (specified with the `--training_file (-t)` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `--test_file (-T)` parameter) and the classification results may be saved with the `--predictions_file (-P)` output parameter. The trained logistic regression model may be saved using the `--output_model_file (-M)` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `--labels_file (-l)` parameter may be used to specify a separate matrix of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `--lambda (-L)` option, and the optimizer used to train the model can be specified with the `--optimizer (-O)` parameter.  Available options are 'sgd' (stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `--max_iterations (-n)` parameter specifies the maximum number of allowed iterations, and the `--tolerance (-e)` parameter specifies the tolerance for convergence.  For the SGD optimizer, the `--step_size (-s)` parameter controls the step size taken at each iteration by the optimizer.  The batch size for SGD is controlled with the `--batch_size (-b)` parameter. If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+For SGD, an iteration refers to a single point. So to take a single pass over the dataset with SGD, `--max_iterations (-n)` should be set to the number of points in the dataset.
+
+Optionally, the model can be used to predict the responses for another matrix of data points, if `--test_file (-T)` is specified.  The `--test_file (-T)` parameter can be specified without the `--training_file (-t)` parameter, so long as an existing logistic regression model is given with the `--input_model_file (-m)` parameter.  The output predictions from the logistic regression model may be saved with the `--predictions_file (-P)` parameter.
+
+This implementation of logistic regression does not support the general multi-class case but instead only the two-class case.  Any labels must be either 0 or 1.  For more classes, see the softmax regression implementation.
+
+### Example
+As an example, to train a logistic regression model on the data '`'data.csv'`' with labels '`'labels.csv'`' with L2 regularization of 0.1, saving the model to '`'lr_model.bin'`', the following command may be used:
+
+```bash
+$ mlpack_logistic_regression --training_file data.csv --labels_file labels.csv
+  --lambda 0.1 --output_model_file lr_model.bin --print_training_accuracy
+```
+
+Then, to use that model to predict classes for the dataset '`'test.csv'`', storing the output predictions in '`'predictions.csv'`', the following command may be used: 
+
+```bash
+$ mlpack_logistic_regression --input_model_file lr_model.bin --test_file
+  test.csv --predictions_file predictions.csv
+```
+
+### See also
+
+ - [mlpack_softmax_regression](#softmax_regression)
+ - [mlpack_random_forest](#random_forest)
+ - [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+ - [:LogisticRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/logistic_regression/logistic_regression.hpp)
+
+## mlpack_lsh
+{: #lsh }
+
+#### K-Approximate-Nearest-Neighbor Search with LSH
+{: #lsh_descr }
+
+```bash
+$ mlpack_lsh [--bucket_size 500] [--hash_width 0] [--help] [--info
+        <string>] [--input_model_file <string>] [--k 0] [--num_probes 0]
+        [--projections 10] [--query_file <string>] [--reference_file <string>]
+        [--second_hash_size 99901] [--seed 0] [--tables 30]
+        [--true_neighbors_file <string>] [--verbose] [--version]
+        [--distances_file <string>] [--neighbors_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of approximate k-nearest-neighbor search with locality-sensitive hashing (LSH).  Given a set of reference points and a set of query points, this will compute the k approximate nearest neighbors of each query point in the reference set; models can be saved for future use. [Detailed documentation](#lsh_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--bucket_size (-B)` | [`int`](#doc_int) | The size of a bucket in the second level hash. | `500` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--hash_width (-H)` | [`double`](#doc_double) | The hash width for the first-level hashing in the LSH preprocessing. By default, the LSH class automatically estimates a hash width for its use. | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`LSHSearch<> file`](#doc_model) | Input LSH model. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `--num_probes (-T)` | [`int`](#doc_int) | Number of additional probes for multiprobe LSH; if 0, traditional LSH is used. | `0` |
+| `--projections (-K)` | [`int`](#doc_int) | The number of hash functions for each table | `10` |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing query points (optional). | `''` |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--second_hash_size (-S)` | [`int`](#doc_int) | The size of the second level hash table. | `99901` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--tables (-L)` | [`int`](#doc_int) | The number of hash tables to be used. | `30` |
+| `--true_neighbors_file (-t)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix of true neighbors to compute recall with (the recall is printed when -v is specified). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to output distances into. | 
+| `--neighbors_file (-n)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to output neighbors into. | 
+| `--output_model_file (-M)` | [`LSHSearch<> file`](#doc_model) | Output for trained LSH model. | 
+
+### Detailed documentation
+{: #lsh_detailed-documentation }
+
+This program will calculate the k approximate-nearest-neighbors of a set of points using locality-sensitive hashing. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. 
+
+### Example
+For example, the following will return 5 neighbors from the data for each point in `'input.csv'` and store the distances in `'distances.csv'` and the neighbors in `'neighbors.csv'`:
+
+```bash
+$ mlpack_lsh --k 5 --reference_file input.csv --distances_file distances.csv
+  --neighbors_file neighbors.csv
+```
+
+The output is organized such that row i and column j in the neighbors output corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output file corresponds to the distance between those two points.
+
+Because this is approximate-nearest-neighbors search, results may be different from run to run.  Thus, the `--seed (-s)` parameter can be specified to set the random seed.
+
+This program also has many other parameters to control its functionality; see the parameter-specific documentation for more information.
+
+### See also
+
+ - [mlpack_knn](#knn)
+ - [mlpack_krann](#krann)
+ - [Locality-sensitive hashing on Wikipedia](https://en.wikipedia.org/wiki/Locality-sensitive_hashing)
+ - [Locality-sensitive hashing scheme based on p-stable  distributions(pdf)](https://www.mlpack.org/papers/lsh.pdf)
+ - [LSHSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lsh/lsh.hpp)
+
+## mlpack_mean_shift
+{: #mean_shift }
+
+#### Mean Shift Clustering
+{: #mean_shift_descr }
+
+```bash
+$ mlpack_mean_shift [--force_convergence] [--help] [--in_place] [--info
+        <string>] --input_file <string> [--labels_only] [--max_iterations 1000]
+        [--radius 0] [--verbose] [--version] [--centroid_file <string>]
+        [--output_file <string>]
+```
+
+A fast implementation of mean-shift clustering using dual-tree range search.  Given a dataset, this uses the mean shift algorithm to produce and return a clustering of the data. [Detailed documentation](#mean_shift_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--force_convergence (-f)` | [`flag`](#doc_flag) | If specified, the mean shift algorithm will continue running regardless of max_iterations until the clusters converge. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--in_place (-P)` | [`flag`](#doc_flag) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden.  (Do not use with Python.) |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform clustering on. | `**--**` |
+| `--labels_only (-l)` | [`flag`](#doc_flag) | If specified, only the output labels will be written to the file specified by --output_file. |  |
+| `--max_iterations (-m)` | [`int`](#doc_int) | Maximum number of iterations before mean shift terminates. | `1000` |
+| `--radius (-r)` | [`double`](#doc_double) | If the distance between two centroids is less than the given radius, one will be removed.  A radius of 0 or less means an estimate will be calculated and used for the radius. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--centroid_file (-C)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | If specified, the centroids of each cluster will be written to the given matrix. | 
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to write output labels or labeled data to. | 
+
+### Detailed documentation
+{: #mean_shift_detailed-documentation }
+
+This program performs mean shift clustering on the given dataset, storing the learned cluster assignments either as a column of labels in the input dataset or separately.
+
+The input dataset should be specified with the `--input_file (-i)` parameter, and the radius used for search can be specified with the `--radius (-r)` parameter.  The maximum number of iterations before algorithm termination is controlled with the `--max_iterations (-m)` parameter.
+
+The output labels may be saved with the `--output_file (-o)` output parameter and the centroids of each cluster may be saved with the `--centroid_file (-C)` output parameter.
+
+### Example
+For example, to run mean shift clustering on the dataset `'data.csv'` and store the centroids to `'centroids.csv'`, the following command may be used: 
+
+```bash
+$ mlpack_mean_shift --input_file data.csv --centroid_file centroids.csv
+```
+
+### See also
+
+ - [mlpack_kmeans](#kmeans)
+ - [mlpack_dbscan](#dbscan)
+ - [Mean shift on Wikipedia](https://en.wikipedia.org/wiki/Mean_shift)
+ - [Mean Shift, Mode Seeking, and Clustering (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=1c168275c59ba382588350ee1443537f59978183)
+ - [mlpack::mean_shift::MeanShift C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/mean_shift/mean_shift.hpp)
+
+## mlpack_nbc
+{: #nbc }
+
+#### Parametric Naive Bayes Classifier
+{: #nbc_descr }
+
+```bash
+$ mlpack_nbc [--help] [--incremental_variance] [--info <string>]
+        [--input_model_file <string>] [--labels_file <string>] [--test_file
+        <string>] [--training_file <string>] [--verbose] [--version]
+        [--output_file <string>] [--output_model_file <string>]
+        [--output_probs_file <string>] [--predictions_file <string>]
+        [--probabilities_file <string>]
+```
+
+An implementation of the Naive Bayes Classifier, used for classification. Given labeled data, an NBC model can be trained and saved, or, a pre-trained model can be used for classification. [Detailed documentation](#nbc_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--incremental_variance (-I)` | [`flag`](#doc_flag) | The variance of each class will be calculated incrementally. |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`NBCModel file`](#doc_model) | Input Naive Bayes model. | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | A file containing labels for the training set. | `''` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the test set. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the training set. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | The matrix in which the predicted labels for the test set will be written (deprecated). | 
+| `--output_model_file (-M)` | [`NBCModel file`](#doc_model) | File to save trained Naive Bayes model to. | 
+| `--output_probs_file` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The matrix in which the predicted probability of labels for the test set will be written (deprecated). | 
+| `--predictions_file (-a)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | The matrix in which the predicted labels for the test set will be written. | 
+| `--probabilities_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | The matrix in which the predicted probability of labels for the test set will be written. | 
+
+### Detailed documentation
+{: #nbc_detailed-documentation }
+
+This program trains the Naive Bayes classifier on the given labeled training set, or loads a model from the given model file, and then may use that trained model to classify the points in a given test set.
+
+The training set is specified with the `--training_file (-t)` parameter.  Labels may be either the last row of the training set, or alternately the `--labels_file (-l)` parameter may be specified to pass a separate matrix of labels.
+
+If training is not desired, a pre-existing model may be loaded with the `--input_model_file (-m)` parameter.
+
+
+
+The `--incremental_variance (-I)` parameter can be used to force the training to use an incremental algorithm for calculating variance.  This is slower, but can help avoid loss of precision in some cases.
+
+If classifying a test set is desired, the test set may be specified with the `--test_file (-T)` parameter, and the classifications may be saved with the `--predictions_file (-a)`predictions  parameter.  If saving the trained model is desired, this may be done with the `--output_model_file (-M)` output parameter.
+
+Note: the `--output_file (-o)` and `--output_probs_file` parameters are deprecated and will be removed in mlpack 4.0.0.  Use `--predictions_file (-a)` and `--probabilities_file (-p)` instead.
+
+### Example
+For example, to train a Naive Bayes classifier on the dataset `'data.csv'` with labels `'labels.csv'` and save the model to `'nbc_model.bin'`, the following command may be used:
+
+```bash
+$ mlpack_nbc --training_file data.csv --labels_file labels.csv
+  --output_model_file nbc_model.bin
+```
+
+Then, to use `'nbc_model.bin'` to predict the classes of the dataset `'test_set.csv'` and save the predicted classes to `'predictions.csv'`, the following command may be used:
+
+```bash
+$ mlpack_nbc --input_model_file nbc_model.bin --test_file test_set.csv
+  --output_file predictions.csv
+```
+
+### See also
+
+ - [mlpack_softmax_regression](#softmax_regression)
+ - [mlpack_random_forest](#random_forest)
+ - [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+ - [NaiveBayesClassifier C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp)
+
+## mlpack_nca
+{: #nca }
+
+#### Neighborhood Components Analysis (NCA)
+{: #nca_descr }
+
+```bash
+$ mlpack_nca [--armijo_constant 0.0001] [--batch_size 50] [--help]
+        [--info <string>] --input_file <string> [--labels_file <string>]
+        [--linear_scan] [--max_iterations 500000] [--max_line_search_trials 50]
+        [--max_step 1e+20] [--min_step 1e-20] [--normalize] [--num_basis 5]
+        [--optimizer 'sgd'] [--seed 0] [--step_size 0.01] [--tolerance 1e-07]
+        [--verbose] [--version] [--wolfe 0.9] [--output_file <string>]
+```
+
+An implementation of neighborhood components analysis, a distance learning technique that can be used for preprocessing.  Given a labeled dataset, this uses NCA, which seeks to improve the k-nearest-neighbor classification, and returns the learned distance metric. [Detailed documentation](#nca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--armijo_constant (-A)` | [`double`](#doc_double) | Armijo constant for L-BFGS. | `0.0001` |
+| `--batch_size (-b)` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to run NCA on. | `**--**` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels for input dataset. | `''` |
+| `--linear_scan (-L)` | [`flag`](#doc_flag) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. |  |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations for SGD or L-BFGS (0 indicates no limit). | `500000` |
+| `--max_line_search_trials (-T)` | [`int`](#doc_int) | Maximum number of line search trials for L-BFGS. | `50` |
+| `--max_step (-M)` | [`double`](#doc_double) | Maximum step of line search for L-BFGS. | `1e+20` |
+| `--min_step (-m)` | [`double`](#doc_double) | Minimum step of line search for L-BFGS. | `1e-20` |
+| `--normalize (-N)` | [`flag`](#doc_flag) | Use a normalized starting point for optimization. This is useful for when points are far apart, or when SGD is returning NaN. |  |
+| `--num_basis (-B)` | [`int`](#doc_int) | Number of memory points to be stored for L-BFGS. | `5` |
+| `--optimizer (-O)` | [`string`](#doc_string) | Optimizer to use; 'sgd' or 'lbfgs'. | `'sgd'` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--step_size (-a)` | [`double`](#doc_double) | Step size for stochastic gradient descent (alpha). | `0.01` |
+| `--tolerance (-t)` | [`double`](#doc_double) | Maximum tolerance for termination of SGD or L-BFGS. | `1e-07` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--wolfe (-w)` | [`double`](#doc_double) | Wolfe condition parameter for L-BFGS. | `0.9` |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Output matrix for learned distance matrix. | 
+
+### Detailed documentation
+{: #nca_detailed-documentation }
+
+This program implements Neighborhood Components Analysis, both a linear dimensionality reduction technique and a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset by scaling the dimensions.  The method is nonparametric, and does not require a value of k.  It works by using stochastic ("soft") neighbor assignments and using optimization techniques over the gradient of the accuracy of the neighbor assignments.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `--input_file (-i)`), or alternatively as a separate matrix (specified with `--labels_file (-l)`).
+
+This implementation of NCA uses stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer.  These optimizers do not guarantee global convergence for a nonconvex objective function (NCA's objective function is nonconvex), so the final results could depend on the random seed or other optimizer parameters.
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `--optimizer (-O)`, depends primarily on three parameters: the step size (specified with `--step_size (-a)`), the batch size (specified with `--batch_size (-b)`), and the maximum number of iterations (specified with `--max_iterations (-n)`).  In addition, a normalized starting point can be used by specifying the `--normalize (-N)` parameter, which is necessary if many warnings of the form 'Denominator of p_i is 0!' are given.  Tuning the step size can be a tedious affair.  In general, the step size is too large if the objective is not mostly uniformly decreasing, or if zero-valued denominator warnings are being issued.  The step size is too small if the objective is changing very slowly.  Setting the termination condition can be done easily once a good step size parameter is found; either increase the maximum iterations to a large number and allow SGD to find a minimum, or set the maximum iterations to 0 (allowing infinite iterations) and set the tolerance (specified by `--tolerance (-t)`) to define the maximum allowed difference between objectives for SGD to terminate.  Be careful---setting the tolerance instead of the maximum iterations can take a very long time and may actually never converge due to the properties of the SGD optimizer. Note that a single iteration of SGD refers to a single point, so to take a single pass over the dataset, set the value of the `--max_iterations (-n)` parameter equal to the number of points in the dataset.
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `--optimizer (-O)`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `--num_basis (-B)` (specifies the number of memory points used by L-BFGS), `--max_iterations (-n)`, `--armijo_constant (-A)`, `--wolfe (-w)`, `--tolerance (-t)` (the optimization is terminated when the gradient norm is below this value), `--max_line_search_trials (-T)`, `--min_step (-m)`, and `--max_step (-M)` (which both refer to the line search routine).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.
+
+By default, the SGD optimizer is used.
+
+### See also
+
+ - [mlpack_lmnn](#lmnn)
+ - [Neighbourhood components analysis on Wikipedia](https://en.wikipedia.org/wiki/Neighbourhood_components_analysis)
+ - [Neighbourhood components analysis (pdf)](http://papers.nips.cc/paper/2566-neighbourhood-components-analysis.pdf)
+ - [NCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/nca/nca.hpp)
+
+## mlpack_knn
+{: #knn }
+
+#### k-Nearest-Neighbors Search
+{: #knn_descr }
+
+```bash
+$ mlpack_knn [--algorithm 'dual_tree'] [--epsilon 0] [--help] [--info
+        <string>] [--input_model_file <string>] [--k 0] [--leaf_size 20]
+        [--query_file <string>] [--random_basis] [--reference_file <string>]
+        [--rho 0.7] [--seed 0] [--tau 0] [--tree_type 'kd']
+        [--true_distances_file <string>] [--true_neighbors_file <string>]
+        [--verbose] [--version] [--distances_file <string>] [--neighbors_file
+        <string>] [--output_model_file <string>]
+```
+
+An implementation of k-nearest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#knn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--algorithm (-a)` | [`string`](#doc_string) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `'dual_tree'` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--epsilon (-e)` | [`double`](#doc_double) | If specified, will do approximate nearest neighbor search with given relative error. | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`KNNModel file`](#doc_model) | Pre-trained kNN model. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `--leaf_size (-l)` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, spill trees, and octrees). | `20` |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing query points (optional). | `''` |
+| `--random_basis (-R)` | [`flag`](#doc_flag) | Before tree-building, project the data onto a random orthogonal basis. |  |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--rho (-b)` | [`double`](#doc_double) | Balance threshold (only valid for spill trees). | `0.7` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `--tau (-u)` | [`double`](#doc_double) | Overlapping size (only valid for spill trees). | `0` |
+| `--tree_type (-t)` | [`string`](#doc_string) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'spill', 'oct'. | `'kd'` |
+| `--true_distances_file (-D)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `''` |
+| `--true_neighbors_file (-T)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to output distances into. | 
+| `--neighbors_file (-n)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to output neighbors into. | 
+| `--output_model_file (-M)` | [`KNNModel file`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #knn_detailed-documentation }
+
+This program will calculate the k-nearest-neighbors of a set of points using kd-trees or cover trees (cover tree support is experimental and may be slow). You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following command will calculate the 5 nearest neighbors of each point in `'input.csv'` and store the distances in `'distances.csv'` and the neighbors in `'neighbors.csv'`: 
+
+```bash
+$ mlpack_knn --k 5 --reference_file input.csv --neighbors_file neighbors.csv
+  --distances_file distances.csv
+```
+
+The output is organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output matrix corresponds to the distance between those two points.
+
+### See also
+
+ - [mlpack_lsh](#lsh)
+ - [mlpack_krann](#krann)
+ - [mlpack_kfn](#kfn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## mlpack_kfn
+{: #kfn }
+
+#### k-Furthest-Neighbors Search
+{: #kfn_descr }
+
+```bash
+$ mlpack_kfn [--algorithm 'dual_tree'] [--epsilon 0] [--help] [--info
+        <string>] [--input_model_file <string>] [--k 0] [--leaf_size 20]
+        [--percentage 1] [--query_file <string>] [--random_basis]
+        [--reference_file <string>] [--seed 0] [--tree_type 'kd']
+        [--true_distances_file <string>] [--true_neighbors_file <string>]
+        [--verbose] [--version] [--distances_file <string>] [--neighbors_file
+        <string>] [--output_model_file <string>]
+```
+
+An implementation of k-furthest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k furthest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--algorithm (-a)` | [`string`](#doc_string) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `'dual_tree'` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--epsilon (-e)` | [`double`](#doc_double) | If specified, will do approximate furthest neighbor search with given relative error. Must be in the range [0,1). | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`KFNModel file`](#doc_model) | Pre-trained kFN model. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of furthest neighbors to find. | `0` |
+| `--leaf_size (-l)` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `--percentage (-p)` | [`double`](#doc_double) | If specified, will do approximate furthest neighbor search. Must be in the range (0,1] (decimal form). Resultant neighbors will be at least (p*100) % of the distance as the true furthest neighbor. | `1` |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing query points (optional). | `''` |
+| `--random_basis (-R)` | [`flag`](#doc_flag) | Before tree-building, project the data onto a random orthogonal basis. |  |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `--tree_type (-t)` | [`string`](#doc_string) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `'kd'` |
+| `--true_distances_file (-D)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `''` |
+| `--true_neighbors_file (-T)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to output distances into. | 
+| `--neighbors_file (-n)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to output neighbors into. | 
+| `--output_model_file (-M)` | [`KFNModel file`](#doc_model) | If specified, the kFN model will be output here. | 
+
+### Detailed documentation
+{: #kfn_detailed-documentation }
+
+This program will calculate the k-furthest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following will calculate the 5 furthest neighbors of eachpoint in `'input.csv'` and store the distances in `'distances.csv'` and the neighbors in `'neighbors.csv'`: 
+
+```bash
+$ mlpack_kfn --k 5 --reference_file input.csv --distances_file distances.csv
+  --neighbors_file neighbors.csv
+```
+
+The output files are organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th furthest neighbor from the point in the query set with index i.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [mlpack_approx_kfn](#approx_kfn)
+ - [mlpack_knn](#knn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## mlpack_nmf
+{: #nmf }
+
+#### Non-negative Matrix Factorization
+{: #nmf_descr }
+
+```bash
+$ mlpack_nmf [--help] [--info <string>] [--initial_h_file <string>]
+        [--initial_w_file <string>] --input_file <string> [--max_iterations
+        10000] [--min_residue 1e-05] --rank 0 [--seed 0] [--update_rules
+        'multdist'] [--verbose] [--version] [--h_file <string>] [--w_file
+        <string>]
+```
+
+An implementation of non-negative matrix factorization.  This can be used to decompose an input dataset into two low-rank non-negative components. [Detailed documentation](#nmf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--initial_h_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Initial H matrix. | `''` |
+| `--initial_w_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Initial W matrix. | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform NMF on. | `**--**` |
+| `--max_iterations (-m)` | [`int`](#doc_int) | Number of iterations before NMF terminates (0 runs until convergence. | `10000` |
+| `--min_residue (-e)` | [`double`](#doc_double) | The minimum root mean square residue allowed for each iteration, below which the program terminates. | `1e-05` |
+| `--rank (-r)` | [`int`](#doc_int) | Rank of the factorization. | `**--**` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--update_rules (-u)` | [`string`](#doc_string) | Update rules for each iteration; ( multdist \| multdiv \| als ). | `'multdist'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--h_file (-H)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save the calculated H to. | 
+| `--w_file (-W)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save the calculated W to. | 
+
+### Detailed documentation
+{: #nmf_detailed-documentation }
+
+This program performs non-negative matrix factorization on the given dataset, storing the resulting decomposed matrices in the specified files.  For an input dataset V, NMF decomposes V into two matrices W and H such that 
+
+V = W * H
+
+where all elements in W and H are non-negative.  If V is of size (n x m), then W will be of size (n x r) and H will be of size (r x m), where r is the rank of the factorization (specified by the `--rank (-r)` parameter).
+
+Optionally, the desired update rules for each NMF iteration can be chosen from the following list:
+
+ - multdist: multiplicative distance-based update rules (Lee and Seung 1999)
+ - multdiv: multiplicative divergence-based update rules (Lee and Seung 1999)
+ - als: alternating least squares update rules (Paatero and Tapper 1994)
+
+The maximum number of iterations is specified with `--max_iterations (-m)`, and the minimum residue required for algorithm termination is specified with the `--min_residue (-e)` parameter.
+
+### Example
+For example, to run NMF on the input matrix `'V.csv'` using the 'multdist' update rules with a rank-10 decomposition and storing the decomposed matrices into `'W.csv'` and `'H.csv'`, the following command could be used: 
+
+```bash
+$ mlpack_nmf --input_file V.csv --w_file W.csv --h_file H.csv --rank 10
+  --update_rules multdist
+```
+
+### See also
+
+ - [mlpack_cf](#cf)
+ - [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ - [Algorithms for non-negative matrix factorization (pdf)](http://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf)
+ - [AMF C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/amf/amf.hpp)
+
+## mlpack_pca
+{: #pca }
+
+#### Principal Components Analysis
+{: #pca_descr }
+
+```bash
+$ mlpack_pca [--decomposition_method 'exact'] [--help] [--info <string>]
+        --input_file <string> [--new_dimensionality 0] [--scale]
+        [--var_to_retain 0] [--verbose] [--version] [--output_file <string>]
+```
+
+An implementation of several strategies for principal components analysis (PCA), a common preprocessing step.  Given a dataset and a desired new dimensionality, this can reduce the dimensionality of the data using the linear transformation determined by PCA. [Detailed documentation](#pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--decomposition_method (-c)` | [`string`](#doc_string) | Method used for the principal components analysis: 'exact', 'randomized', 'randomized-block-krylov', 'quic'. | `'exact'` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset to perform PCA on. | `**--**` |
+| `--new_dimensionality (-d)` | [`int`](#doc_int) | Desired dimensionality of output dataset. If 0, no dimensionality reduction is performed. | `0` |
+| `--scale (-s)` | [`flag`](#doc_flag) | If set, the data will be scaled before running PCA, such that the variance of each feature is 1. |  |
+| `--var_to_retain (-r)` | [`double`](#doc_double) | Amount of variance to retain; should be between 0 and 1.  If 1, all variance is retained.  Overrides -d. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #pca_detailed-documentation }
+
+This program performs principal components analysis on the given dataset using the exact, randomized, randomized block Krylov, or QUIC SVD method. It will transform the data onto its principal components, optionally performing dimensionality reduction by ignoring the principal components with the smallest eigenvalues.
+
+Use the `--input_file (-i)` parameter to specify the dataset to perform PCA on.  A desired new dimensionality can be specified with the `--new_dimensionality (-d)` parameter, or the desired variance to retain can be specified with the `--var_to_retain (-r)` parameter.  If desired, the dataset can be scaled before running PCA with the `--scale (-s)` parameter.
+
+Multiple different decomposition techniques can be used.  The method to use can be specified with the `--decomposition_method (-c)` parameter, and it may take the values 'exact', 'randomized', or 'quic'.
+
+### Example
+For example, to reduce the dimensionality of the matrix `'data.csv'` to 5 dimensions using randomized SVD for the decomposition, storing the output matrix to `'data_mod.csv'`, the following command can be used:
+
+```bash
+$ mlpack_pca --input_file data.csv --new_dimensionality 5
+  --decomposition_method randomized --output_file data_mod.csv
+```
+
+### See also
+
+ - [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+ - [PCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/pca/pca.hpp)
+
+## mlpack_perceptron
+{: #perceptron }
+
+#### Perceptron
+{: #perceptron_descr }
+
+```bash
+$ mlpack_perceptron [--help] [--info <string>] [--input_model_file
+        <string>] [--labels_file <string>] [--max_iterations 1000] [--test_file
+        <string>] [--training_file <string>] [--verbose] [--version]
+        [--output_file <string>] [--output_model_file <string>]
+        [--predictions_file <string>]
+```
+
+An implementation of a perceptron---a single level neural network--=for classification.  Given labeled data, a perceptron can be trained and saved for future use; or, a pre-trained perceptron can be used for classification on new points. [Detailed documentation](#perceptron_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`PerceptronModel file`](#doc_model) | Input perceptron model. | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | A matrix containing labels for the training set. | `''` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | The maximum number of iterations the perceptron is to be run | `1000` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the test set. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the training set. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | The matrix in which the predicted labels for the test set will be written. | 
+| `--output_model_file (-M)` | [`PerceptronModel file`](#doc_model) | Output for trained perceptron model. | 
+| `--predictions_file (-P)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | The matrix in which the predicted labels for the test set will be written. | 
+
+### Detailed documentation
+{: #perceptron_detailed-documentation }
+
+This program implements a perceptron, which is a single level neural network. The perceptron makes its predictions based on a linear predictor function combining a set of weights with the feature vector.  The perceptron learning rule is able to converge, given enough iterations (specified using the `--max_iterations (-n)` parameter), if the data supplied is linearly separable.  The perceptron is parameterized by a matrix of weight vectors that denote the numerical weights of the neural network.
+
+This program allows loading a perceptron from a model (via the `--input_model_file (-m)` parameter) or training a perceptron given training data (via the `--training_file (-t)` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (via the `--test_file (-T)` parameter) and the classification results on the test set may be saved with the `--predictions_file (-P)` output parameter.  The perceptron model may be saved with the `--output_model_file (-M)` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `--output_file (-o)`.
+Use `--predictions_file (-P)` instead of `--output_file (-o)`.
+
+### Example
+The training data given with the `--training_file (-t)` option may have class labels as its last dimension (so, if the training data is in CSV format, labels should be the last column).  Alternately, the `--labels_file (-l)` parameter may be used to specify a separate matrix of labels.
+
+All these options make it easy to train a perceptron, and then re-use that perceptron for later classification.  The invocation below trains a perceptron on `'training_data.csv'` with labels `'training_labels.csv'`, and saves the model to `'perceptron_model.bin'`.
+
+```bash
+$ mlpack_perceptron --training_file training_data.csv --labels_file
+  training_labels.csv --output_model_file perceptron_model.bin
+```
+
+Then, this model can be re-used for classification on the test data `'test_data.csv'`.  The example below does precisely that, saving the predicted classes to `'predictions.csv'`.
+
+```bash
+$ mlpack_perceptron --input_model_file perceptron_model.bin --test_file
+  test_data.csv --predictions_file predictions.csv
+```
+
+Note that all of the options may be specified at once: predictions may be calculated right after training a model, and model training can occur even if an existing perceptron model is passed with the `--input_model_file (-m)` parameter.  However, note that the number of classes and the dimensionality of all data must match.  So you cannot pass a perceptron model trained on 2 classes and then re-train with a 4-class dataset.  Similarly, attempting classification on a 3-dimensional dataset with a perceptron that has been trained on 8 dimensions will cause an error.
+
+### See also
+
+ - [mlpack_adaboost](#adaboost)
+ - [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+ - [Perceptron C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/perceptron/perceptron.hpp)
+
+## mlpack_preprocess_split
+{: #preprocess_split }
+
+#### Split Data
+{: #preprocess_split_descr }
+
+```bash
+$ mlpack_preprocess_split [--help] [--info <string>] --input_file
+        <string> [--input_labels_file <string>] [--no_shuffle] [--seed 0]
+        [--stratify_data] [--test_ratio 0.2] [--verbose] [--version]
+        [--test_file <string>] [--test_labels_file <string>] [--training_file
+        <string>] [--training_labels_file <string>]
+```
+
+A utility to split data into a training and testing dataset.  This can also split labels according to the same split. [Detailed documentation](#preprocess_split_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing data. | `**--**` |
+| `--input_labels_file (-I)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix containing labels. | `''` |
+| `--no_shuffle (-S)` | [`flag`](#doc_flag) | Avoid shuffling the data before splitting. |  |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `--stratify_data (-z)` | [`flag`](#doc_flag) | Stratify the data according to labels |  |
+| `--test_ratio (-r)` | [`double`](#doc_double) | Ratio of test set; if not set,the ratio defaults to 0.2 | `0.2` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save test data to. | 
+| `--test_labels_file (-L)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to save test labels to. | 
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save training data to. | 
+| `--training_labels_file (-l)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to save train labels to. | 
+
+### Detailed documentation
+{: #preprocess_split_detailed-documentation }
+
+This utility takes a dataset and optionally labels and splits them into a training set and a test set. Before the split, the points in the dataset are randomly reordered. The percentage of the dataset to be used as the test set can be specified with the `--test_ratio (-r)` parameter; the default is 0.2 (20%).
+
+The output training and test matrices may be saved with the `--training_file (-t)` and `--test_file (-T)` output parameters.
+
+Optionally, labels can also be split along with the data by specifying the `--input_labels_file (-I)` parameter.  Splitting labels works the same way as splitting the data. The output training and test labels may be saved with the `--training_labels_file (-l)` and `--test_labels_file (-L)` output parameters, respectively.
+
+### Example
+So, a simple example where we want to split the dataset `'X.csv'` into `'X_train.csv'` and `'X_test.csv'` with 60% of the data in the training set and 40% of the dataset in the test set, we could run 
+
+```bash
+$ mlpack_preprocess_split --input_file X.csv --training_file X_train.csv
+  --test_file X_test.csv --test_ratio 0.4
+```
+
+Also by default the dataset is shuffled and split; you can provide the `--no_shuffle (-S)` option to avoid shuffling the data; an example to avoid shuffling of data is:
+
+```bash
+$ mlpack_preprocess_split --input_file X.csv --training_file X_train.csv
+  --test_file X_test.csv --test_ratio 0.4 --no_shuffle
+```
+
+If we had a dataset `'X.csv'` and associated labels `'y.csv'`, and we wanted to split these into `'X_train.csv'`, `'y_train.csv'`, `'X_test.csv'`, and `'y_test.csv'`, with 30% of the data in the test set, we could run
+
+```bash
+$ mlpack_preprocess_split --input_file X.csv --input_labels_file y.csv
+  --test_ratio 0.3 --training_file X_train.csv --training_labels_file
+  y_train.csv --test_file X_test.csv --test_labels_file y_test.csv
+```
+
+To maintain the ratio of each class in the train and test sets, the`--stratify_data (-z)` option can be used.
+
+```bash
+$ mlpack_preprocess_split --input_file X.csv --training_file X_train.csv
+  --test_file X_test.csv --test_ratio 0.4 --stratify_data
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_describe](#preprocess_describe)
+
+## mlpack_preprocess_binarize
+{: #preprocess_binarize }
+
+#### Binarize Data
+{: #preprocess_binarize_descr }
+
+```bash
+$ mlpack_preprocess_binarize [--dimension 0] [--help] [--info <string>]
+        --input_file <string> [--threshold 0] [--verbose] [--version]
+        [--output_file <string>]
+```
+
+A utility to binarize a dataset.  Given a dataset, this utility converts each value in the desired dimension(s) to 0 or 1; this can be a useful preprocessing step. [Detailed documentation](#preprocess_binarize_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--dimension (-d)` | [`int`](#doc_int) | Dimension to apply the binarization. If not set, the program will binarize every dimension by default. | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input data matrix. | `**--**` |
+| `--threshold (-t)` | [`double`](#doc_double) | Threshold to be applied for binarization. If not set, the threshold defaults to 0.0. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix in which to save the output. | 
+
+### Detailed documentation
+{: #preprocess_binarize_detailed-documentation }
+
+This utility takes a dataset and binarizes the variables into either 0 or 1 given threshold. User can apply binarization on a dimension or the whole dataset.  The dimension to apply binarization to can be specified using the `--dimension (-d)` parameter; if left unspecified, every dimension will be binarized.  The threshold for binarization can also be specified with the `--threshold (-t)` parameter; the default threshold is 0.0.
+
+The binarized matrix may be saved with the `--output_file (-o)` output parameter.
+
+### Example
+For example, if we want to set all variables greater than 5 in the dataset `'X.csv'` to 1 and variables less than or equal to 5.0 to 0, and save the result to `'Y.csv'`, we could run
+
+```bash
+$ mlpack_preprocess_binarize --input_file X.csv --threshold 5 --output_file
+  Y.csv
+```
+
+But if we want to apply this to only the first (0th) dimension of `'X.csv'`,  we could instead run
+
+```bash
+$ mlpack_preprocess_binarize --input_file X.csv --threshold 5 --dimension 0
+  --output_file Y.csv
+```
+
+### See also
+
+ - [mlpack_preprocess_describe](#preprocess_describe)
+ - [mlpack_preprocess_split](#preprocess_split)
+
+## mlpack_preprocess_describe
+{: #preprocess_describe }
+
+#### Descriptive Statistics
+{: #preprocess_describe_descr }
+
+```bash
+$ mlpack_preprocess_describe [--dimension 0] [--help] [--info <string>]
+        --input_file <string> [--population] [--precision 4] [--row_major]
+        [--verbose] [--version] [--width 8]
+```
+
+A utility for printing descriptive statistics about a dataset.  This prints a number of details about a dataset in a tabular format. [Detailed documentation](#preprocess_describe_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--dimension (-d)` | [`int`](#doc_int) | Dimension of the data. Use this to specify a dimension | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing data, | `**--**` |
+| `--population (-P)` | [`flag`](#doc_flag) | If specified, the program will calculate statistics assuming the dataset is the population. By default, the program will assume the dataset as a sample. |  |
+| `--precision (-p)` | [`int`](#doc_int) | Precision of the output statistics. | `4` |
+| `--row_major (-r)` | [`flag`](#doc_flag) | If specified, the program will calculate statistics across rows, not across columns.  (Remember that in mlpack, a column represents a point, so this option is generally not necessary.) |  |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--width (-w)` | [`int`](#doc_int) | Width of the output table. | `8` |
+
+
+### Detailed documentation
+{: #preprocess_describe_detailed-documentation }
+
+This utility takes a dataset and prints out the descriptive statistics of the data. Descriptive statistics is the discipline of quantitatively describing the main features of a collection of information, or the quantitative description itself. The program does not modify the original file, but instead prints out the statistics to the console. The printed result will look like a table.
+
+Optionally, width and precision of the output can be adjusted by a user using the `--width (-w)` and `--precision (-p)` parameters. A user can also select a specific dimension to analyze if there are too many dimensions. The `--population (-P)` parameter can be specified when the dataset should be considered as a population.  Otherwise, the dataset will be considered as a sample.
+
+### Example
+So, a simple example where we want to print out statistical facts about the dataset `'X.csv'` using the default settings, we could run 
+
+```bash
+$ mlpack_preprocess_describe --input_file X.csv --verbose
+```
+
+If we want to customize the width to 10 and precision to 5 and consider the dataset as a population, we could run
+
+```bash
+$ mlpack_preprocess_describe --input_file X.csv --width 10 --precision 5
+  --verbose
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_split](#preprocess_split)
+
+## mlpack_preprocess_scale
+{: #preprocess_scale }
+
+#### Scale Data
+{: #preprocess_scale_descr }
+
+```bash
+$ mlpack_preprocess_scale [--epsilon 1e-06] [--help] [--info <string>]
+        --input_file <string> [--input_model_file <string>] [--inverse_scaling]
+        [--max_value 1] [--min_value 0] [--scaler_method 'standard_scaler']
+        [--seed 0] [--verbose] [--version] [--output_file <string>]
+        [--output_model_file <string>]
+```
+
+A utility to perform feature scaling on datasets using one of sixtechniques.  Both scaling and inverse scaling are supported, andscalers can be saved and then applied to other datasets. [Detailed documentation](#preprocess_scale_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--epsilon (-r)` | [`double`](#doc_double) | regularization Parameter for pcawhitening, or zcawhitening, should be between -1 to 1. | `1e-06` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing data. | `**--**` |
+| `--input_model_file (-m)` | [`ScalingModel file`](#doc_model) | Input Scaling model. | `''` |
+| `--inverse_scaling (-f)` | [`flag`](#doc_flag) | Inverse Scaling to get original dataset |  |
+| `--max_value (-e)` | [`int`](#doc_int) | Ending value of range for min_max_scaler. | `1` |
+| `--min_value (-b)` | [`int`](#doc_int) | Starting value of range for min_max_scaler. | `0` |
+| `--scaler_method (-a)` | [`string`](#doc_string) | method to use for scaling, the default is standard_scaler. | `'standard_scaler'` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save scaled data to. | 
+| `--output_model_file (-M)` | [`ScalingModel file`](#doc_model) | Output scaling model. | 
+
+### Detailed documentation
+{: #preprocess_scale_detailed-documentation }
+
+This utility takes a dataset and performs feature scaling using one of the six scaler methods namely: 'max_abs_scaler', 'mean_normalization', 'min_max_scaler' ,'standard_scaler', 'pca_whitening' and 'zca_whitening'. The function takes a matrix as `--input_file (-i)` and a scaling method type which you can specify using `--scaler_method (-a)` parameter; the default is standard scaler, and outputs a matrix with scaled feature.
+
+The output scaled feature matrix may be saved with the `--output_file (-o)` output parameters.
+
+The model to scale features can be saved using `--output_model_file (-M)` and later can be loaded back using`--input_model_file (-m)`.
+
+### Example
+So, a simple example where we want to scale the dataset `'X.csv'` into `'X_scaled.csv'` with  standard_scaler as scaler_method, we could run 
+
+```bash
+$ mlpack_preprocess_scale --input_file X.csv --output_file X_scaled.csv
+  --scaler_method standard_scaler
+```
+
+A simple example where we want to whiten the dataset `'X.csv'` into `'X_whitened.csv'` with  PCA as whitening_method and use 0.01 as regularization parameter, we could run 
+
+```bash
+$ mlpack_preprocess_scale --input_file X.csv --output_file X_scaled.csv
+  --scaler_method pca_whitening --epsilon 0.01
+```
+
+You can also retransform the scaled dataset back using`--inverse_scaling (-f)`. An example to rescale : `'X_scaled.csv'` into `'X.csv'`using the saved model `--input_model_file (-m)` is:
+
+```bash
+$ mlpack_preprocess_scale --input_file X_scaled.csv --output_file X.csv
+  --inverse_scaling --input_model_file saved.bin
+```
+
+Another simple example where we want to scale the dataset `'X.csv'` into `'X_scaled.csv'` with  min_max_scaler as scaler method, where scaling range is 1 to 3 instead of default 0 to 1. We could run 
+
+```bash
+$ mlpack_preprocess_scale --input_file X.csv --output_file X_scaled.csv
+  --scaler_method min_max_scaler --min_value 1 --max_value 3
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_describe](#preprocess_describe)
+
+## mlpack_preprocess_one_hot_encoding
+{: #preprocess_one_hot_encoding }
+
+#### One Hot Encoding
+{: #preprocess_one_hot_encoding_descr }
+
+```bash
+$ mlpack_preprocess_one_hot_encoding [--dimensions []] [--help] [--info
+        <string>] --input_file <string> [--verbose] [--version] [--output_file
+        <string>]
+```
+
+A utility to do one-hot encoding on features of dataset. [Detailed documentation](#preprocess_one_hot_encoding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--dimensions (-d)` | [`int vector`](#doc_int_vector) | Index of dimensions that need to be one-hot encoded (if unspecified, all categorical dimensions are one-hot encoded). | `[]` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d categorical matrix file`](#doc_a_2_d_categorical_matrix_file) | Matrix containing data. | `**--**` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save one-hot encoded features data to. | 
+
+### Detailed documentation
+{: #preprocess_one_hot_encoding_detailed-documentation }
+
+This utility takes a dataset and a vector of indices and does one-hot encoding of the respective features at those indices. Indices represent the IDs of the dimensions to be one-hot encoded.
+
+If no dimensions are specified with `--dimensions (-d)`, then all categorical-type dimensions will be one-hot encoded. Otherwise, only the dimensions given in `--dimensions (-d)` will be one-hot encoded.
+
+The output matrix with encoded features may be saved with the `--output_file (-o)` parameters.
+
+### Example
+So, a simple example where we want to encode 1st and 3rd feature from dataset `'X.csv'` into `'X_output.csv'` would be
+
+```bash
+$ mlpack_preprocess_one_hot_encoding --input_file X.arff --output_file
+  X_ouput.csv --dimensions 1 --dimensions 3
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_describe](#preprocess_describe)
+ - [One-hot encoding on Wikipedia](https://en.m.wikipedia.org/wiki/One-hot)
+
+## mlpack_radical
+{: #radical }
+
+#### RADICAL
+{: #radical_descr }
+
+```bash
+$ mlpack_radical [--angles 150] [--help] [--info <string>] --input_file
+        <string> [--noise_std_dev 0.175] [--objective] [--replicates 30] [--seed
+        0] [--sweeps 0] [--verbose] [--version] [--output_ic_file <string>]
+        [--output_unmixing_file <string>]
+```
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Given a dataset, this can decompose the dataset into an unmixing matrix and an independent component matrix; this can be useful for preprocessing. [Detailed documentation](#radical_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--angles (-a)` | [`int`](#doc_int) | Number of angles to consider in brute-force search during Radical2D. | `150` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input dataset for ICA. | `**--**` |
+| `--noise_std_dev (-n)` | [`double`](#doc_double) | Standard deviation of Gaussian noise. | `0.175` |
+| `--objective (-O)` | [`flag`](#doc_flag) | If set, an estimate of the final objective function is printed. |  |
+| `--replicates (-r)` | [`int`](#doc_int) | Number of Gaussian-perturbed replicates to use (per point) in Radical2D. | `30` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--sweeps (-S)` | [`int`](#doc_int) | Number of sweeps; each sweep calls Radical2D once for each pair of dimensions. | `0` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_ic_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save independent components to. | 
+| `--output_unmixing_file (-u)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save unmixing matrix to. | 
+
+### Detailed documentation
+{: #radical_detailed-documentation }
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Assuming that we have an input matrix X, the goal is to find a square unmixing matrix W such that Y = W * X and the dimensions of Y are independent components.  If the algorithm is running particularly slowly, try reducing the number of replicates.
+
+The input matrix to perform ICA on should be specified with the `--input_file (-i)` parameter.  The output matrix Y may be saved with the `--output_ic_file (-o)` output parameter, and the output unmixing matrix W may be saved with the `--output_unmixing_file (-u)` output parameter.
+
+### Example
+For example, to perform ICA on the matrix `'X.csv'` with 40 replicates, saving the independent components to `'ic.csv'`, the following command may be used: 
+
+```bash
+$ mlpack_radical --input_file X.csv --replicates 40 --output_ic_file ic.csv
+```
+
+### See also
+
+ - [Independent component analysis on Wikipedia](https://en.wikipedia.org/wiki/Independent_component_analysis)
+ - [ICA using spacings estimates of entropy (pdf)](https://www.jmlr.org/papers/volume4/learned-miller03a/learned-miller03a.pdf)
+ - [Radical C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/radical/radical.hpp)
+
+## mlpack_random_forest
+{: #random_forest }
+
+#### Random forests
+{: #random_forest_descr }
+
+```bash
+$ mlpack_random_forest [--help] [--info <string>] [--input_model_file
+        <string>] [--labels_file <string>] [--maximum_depth 0]
+        [--minimum_gain_split 0] [--minimum_leaf_size 1] [--num_trees 10]
+        [--print_training_accuracy] [--seed 0] [--subspace_dim 0] [--test_file
+        <string>] [--test_labels_file <string>] [--training_file <string>]
+        [--verbose] [--version] [--warm_start] [--output_model_file <string>]
+        [--predictions_file <string>] [--probabilities_file <string>]
+```
+
+An implementation of the standard random forest algorithm by Leo Breiman for classification.  Given labeled data, a random forest can be trained and saved for future use; or, a pre-trained random forest can be used for classification. [Detailed documentation](#random_forest_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`RandomForestModel file`](#doc_model) | Pre-trained random forest to use for classification. | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels for training dataset. | `''` |
+| `--maximum_depth (-D)` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `--minimum_gain_split (-g)` | [`double`](#doc_double) | Minimum gain needed to make a split when building a tree. | `0` |
+| `--minimum_leaf_size (-n)` | [`int`](#doc_int) | Minimum number of points in each leaf node. | `1` |
+| `--num_trees (-N)` | [`int`](#doc_int) | Number of trees in the random forest. | `10` |
+| `--print_training_accuracy (-a)` | [`flag`](#doc_flag) | If set, then the accuracy of the model on the training set will be predicted (verbose must also be specified). |  |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--subspace_dim (-d)` | [`int`](#doc_int) | Dimensionality of random subspace to use for each split.  '0' will autoselect the square root of data dimensionality. | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Test dataset to produce predictions for. | `''` |
+| `--test_labels_file (-L)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Test dataset labels, if accuracy calculation is desired. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Training dataset. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--warm_start (-w)` | [`flag`](#doc_flag) | If true and passed along with `training` and `input_model` then trains more trees on top of existing model. |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`RandomForestModel file`](#doc_model) | Model to save trained random forest to. | 
+| `--predictions_file (-p)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Predicted classes for each point in the test set. | 
+| `--probabilities_file (-P)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #random_forest_detailed-documentation }
+
+This program is an implementation of the standard random forest classification algorithm by Leo Breiman.  A random forest can be trained and saved for later use, or a random forest may be loaded and predictions or class probabilities for points may be generated.
+
+The training set and associated labels are specified with the `--training_file (-t)` and `--labels_file (-l)` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `--labels_file (-l)` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `--output_model_file (-M)` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `--input_model_file (-m)`parameter. The `--input_model_file (-m)` parameter may not be specified when the `--training_file (-t)` parameter is specified.  The `--minimum_leaf_size (-n)` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `--num_trees (-N)` controls the number of trees in the random forest.  The `--minimum_gain_split (-g)` parameter controls the minimum required gain for a decision tree node to split.  Larger values will force higher-confidence splits.  The `--maximum_depth (-D)` parameter specifies the maximum depth of the tree.  The `--subspace_dim (-d)` parameter is used to control the number of random dimensions chosen for an individual node's split.  If `--print_training_accuracy (-a)` is specified, the calculated accuracy on the training set will be printed.
+
+Test data may be specified with the `--test_file (-T)` parameter, and if performance measures are desired for that test set, labels for the test points may be specified with the `--test_labels_file (-L)` parameter.  Predictions for each test point may be saved via the `--predictions_file (-p)`output parameter.  Class probabilities for each prediction may be saved with the `--probabilities_file (-P)` output parameter.
+
+### Example
+For example, to train a random forest with a minimum leaf size of 20 using 10 trees on the dataset contained in `'data.csv'`with labels `'labels.csv'`, saving the output random forest to `'rf_model.bin'` and printing the training error, one could call
+
+```bash
+$ mlpack_random_forest --training_file data.csv --labels_file labels.csv
+  --minimum_leaf_size 20 --num_trees 10 --output_model_file rf_model.bin
+  --print_training_accuracy
+```
+
+Then, to use that model to classify points in `'test_set.csv'` and print the test error given the labels `'test_labels.csv'` using that model, while saving the predictions for each point to `'predictions.csv'`, one could call 
+
+```bash
+$ mlpack_random_forest --input_model_file rf_model.bin --test_file
+  test_set.csv --test_labels_file test_labels.csv --predictions_file
+  predictions.csv
+```
+
+### See also
+
+ - [mlpack_decision_tree](#decision_tree)
+ - [mlpack_hoeffding_tree](#hoeffding_tree)
+ - [mlpack_softmax_regression](#softmax_regression)
+ - [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ - [Random forests (pdf)](https://www.eecis.udel.edu/~shatkay/Course/papers/BreimanRandomForests2001.pdf)
+ - [RandomForest C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/random_forest/random_forest.hpp)
+
+## mlpack_krann
+{: #krann }
+
+#### K-Rank-Approximate-Nearest-Neighbors (kRANN)
+{: #krann_descr }
+
+```bash
+$ mlpack_krann [--alpha 0.95] [--first_leaf_exact] [--help] [--info
+        <string>] [--input_model_file <string>] [--k 0] [--leaf_size 20]
+        [--naive] [--query_file <string>] [--random_basis] [--reference_file
+        <string>] [--sample_at_leaves] [--seed 0] [--single_mode]
+        [--single_sample_limit 20] [--tau 5] [--tree_type 'kd'] [--verbose]
+        [--version] [--distances_file <string>] [--neighbors_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of rank-approximate k-nearest-neighbor search (kRANN)  using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#krann_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--alpha (-a)` | [`double`](#doc_double) | The desired success probability. | `0.95` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--first_leaf_exact (-X)` | [`flag`](#doc_flag) | The flag to trigger sampling only after exactly exploring the first leaf. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`RAModel file`](#doc_model) | Pre-trained kNN model. | `''` |
+| `--k (-k)` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `--leaf_size (-l)` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `--naive (-N)` | [`flag`](#doc_flag) | If true, sampling will be done without using a tree. |  |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing query points (optional). | `''` |
+| `--random_basis (-R)` | [`flag`](#doc_flag) | Before tree-building, project the data onto a random orthogonal basis. |  |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--sample_at_leaves (-L)` | [`flag`](#doc_flag) | The flag to trigger sampling at leaves. |  |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `--single_mode (-S)` | [`flag`](#doc_flag) | If true, single-tree search is used (as opposed to dual-tree search. |  |
+| `--single_sample_limit (-z)` | [`int`](#doc_int) | The limit on the maximum number of samples (and hence the largest node you can approximate). | `20` |
+| `--tau (-T)` | [`double`](#doc_double) | The allowed rank-error in terms of the percentile of the data. | `5` |
+| `--tree_type (-t)` | [`string`](#doc_string) | Type of tree to use: 'kd', 'ub', 'cover', 'r', 'x', 'r-star', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `'kd'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to output distances into. | 
+| `--neighbors_file (-n)` | [`2-d index matrix file`](#doc_a_2_d_index_matrix_file) | Matrix to output neighbors into. | 
+| `--output_model_file (-M)` | [`RAModel file`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #krann_detailed-documentation }
+
+This program will calculate the k rank-approximate-nearest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. You must specify the rank approximation (in %) (and optionally the success probability).
+
+### Example
+For example, the following will return 5 neighbors from the top 0.1% of the data (with probability 0.95) for each point in `'input.csv'` and store the distances in `'distances.csv'` and the neighbors in `'neighbors.csv.csv'`:
+
+```bash
+$ mlpack_krann --reference_file input.csv --k 5 --distances_file distances.csv
+  --neighbors_file neighbors.csv --tau 0.1
+```
+
+Note that tau must be set such that the number of points in the corresponding percentile of the data is greater than k.  Thus, if we choose tau = 0.1 with a dataset of 1000 points and k = 5, then we are attempting to choose 5 nearest neighbors out of the closest 1 point -- this is invalid and the program will terminate with an error message.
+
+The output matrices are organized such that row i and column j in the neighbors output file corresponds to the index of the point in the reference set which is the i'th nearest neighbor from the point in the query set with index j.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [mlpack_knn](#knn)
+ - [mlpack_lsh](#lsh)
+ - [Rank-approximate nearest neighbor search: Retaining meaning and speed in high dimensions (pdf)](https://papers.nips.cc/paper/3864-rank-approximate-nearest-neighbor-search-retaining-meaning-and-speed-in-high-dimensions.pdf)
+ - [RASearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/rann/ra_search.hpp)
+
+## mlpack_softmax_regression
+{: #softmax_regression }
+
+#### Softmax Regression
+{: #softmax_regression_descr }
+
+```bash
+$ mlpack_softmax_regression [--help] [--info <string>]
+        [--input_model_file <string>] [--labels_file <string>] [--lambda 0.0001]
+        [--max_iterations 400] [--no_intercept] [--number_of_classes 0]
+        [--test_file <string>] [--test_labels_file <string>] [--training_file
+        <string>] [--verbose] [--version] [--output_model_file <string>]
+        [--predictions_file <string>] [--probabilities_file <string>]
+```
+
+An implementation of softmax regression for classification, which is a multiclass generalization of logistic regression.  Given labeled data, a softmax regression model can be trained and saved for future use, or, a pre-trained softmax regression model can be used for classification of new points. [Detailed documentation](#softmax_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`SoftmaxRegression<> file`](#doc_model) | File containing existing model (parameters). | `''` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | A matrix containing labels (0 or 1) for the points in the training set (y). The labels must order as a row. | `''` |
+| `--lambda (-r)` | [`double`](#doc_double) | L2-regularization constant | `0.0001` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations before termination. | `400` |
+| `--no_intercept (-N)` | [`flag`](#doc_flag) | Do not add the intercept term to the model. |  |
+| `--number_of_classes (-c)` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing test dataset. | `''` |
+| `--test_labels_file (-L)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Matrix containing test labels. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | A matrix containing the training set (the matrix of predictors, X). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`SoftmaxRegression<> file`](#doc_model) | File to save trained softmax regression model to. | 
+| `--predictions_file (-p)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Matrix to save predictions for test dataset into. | 
+| `--probabilities_file (-P)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save class probabilities for test dataset into. | 
+
+### Detailed documentation
+{: #softmax_regression_detailed-documentation }
+
+This program performs softmax regression, a generalization of logistic regression to the multiclass case, and has support for L2 regularization.  The program is able to train a model, load  an existing model, and give predictions (and optionally their accuracy) for test data.
+
+Training a softmax regression model is done by giving a file of training points with the `--training_file (-t)` parameter and their corresponding labels with the `--labels_file (-l)` parameter. The number of classes can be manually specified with the `--number_of_classes (-c)` parameter, and the maximum number of iterations of the L-BFGS optimizer can be specified with the `--max_iterations (-n)` parameter.  The L2 regularization constant can be specified with the `--lambda (-r)` parameter and if an intercept term is not desired in the model, the `--no_intercept (-N)` parameter can be specified.
+
+The trained model can be saved with the `--output_model_file (-M)` output parameter. If training is not desired, but only testing is, a model can be loaded with the `--input_model_file (-m)` parameter.  At the current time, a loaded model cannot be trained further, so specifying both `--input_model_file (-m)` and `--training_file (-t)` is not allowed.
+
+The program is also able to evaluate a model on test data.  A test dataset can be specified with the `--test_file (-T)` parameter. Class predictions can be saved with the `--predictions_file (-p)` output parameter.  If labels are specified for the test data with the `--test_labels_file (-L)` parameter, then the program will print the accuracy of the predictions on the given test set and its corresponding labels.
+
+### Example
+For example, to train a softmax regression model on the data `'dataset.csv'` with labels `'labels.csv'` with a maximum of 1000 iterations for training, saving the trained model to `'sr_model.bin'`, the following command can be used: 
+
+```bash
+$ mlpack_softmax_regression --training_file dataset.csv --labels_file
+  labels.csv --output_model_file sr_model.bin
+```
+
+Then, to use `'sr_model.bin'` to classify the test points in `'test_points.csv'`, saving the output predictions to `'predictions.csv'`, the following command can be used:
+
+```bash
+$ mlpack_softmax_regression --input_model_file sr_model.bin --test_file
+  test_points.csv --predictions_file predictions.csv
+```
+
+### See also
+
+ - [mlpack_logistic_regression](#logistic_regression)
+ - [mlpack_random_forest](#random_forest)
+ - [Multinomial logistic regression (softmax regression) on Wikipedia](https://en.wikipedia.org/wiki/Multinomial_logistic_regression)
+ - [SoftmaxRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/softmax_regression/softmax_regression.hpp)
+
+## mlpack_sparse_coding
+{: #sparse_coding }
+
+#### Sparse Coding
+{: #sparse_coding_descr }
+
+```bash
+$ mlpack_sparse_coding [--atoms 15] [--help] [--info <string>]
+        [--initial_dictionary_file <string>] [--input_model_file <string>]
+        [--lambda1 0] [--lambda2 0] [--max_iterations 0] [--newton_tolerance
+        1e-06] [--normalize] [--objective_tolerance 0.01] [--seed 0]
+        [--test_file <string>] [--training_file <string>] [--verbose]
+        [--version] [--codes_file <string>] [--dictionary_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of Sparse Coding with Dictionary Learning.  Given a dataset, this will decompose the dataset into a sparse combination of a few dictionary elements, where the dictionary is learned during computation; a dictionary can be reused for future sparse coding of new points. [Detailed documentation](#sparse_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--atoms (-k)` | [`int`](#doc_int) | Number of atoms in the dictionary. | `15` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--initial_dictionary_file (-i)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Optional initial dictionary matrix. | `''` |
+| `--input_model_file (-m)` | [`SparseCoding file`](#doc_model) | File containing input sparse coding model. | `''` |
+| `--lambda1 (-l)` | [`double`](#doc_double) | Sparse coding l1-norm regularization parameter. | `0` |
+| `--lambda2 (-L)` | [`double`](#doc_double) | Sparse coding l2-norm regularization parameter. | `0` |
+| `--max_iterations (-n)` | [`int`](#doc_int) | Maximum number of iterations for sparse coding (0 indicates no limit). | `0` |
+| `--newton_tolerance (-w)` | [`double`](#doc_double) | Tolerance for convergence of Newton method. | `1e-06` |
+| `--normalize (-N)` | [`flag`](#doc_flag) | If set, the input data matrix will be normalized before coding. |  |
+| `--objective_tolerance (-o)` | [`double`](#doc_double) | Tolerance for convergence of the objective function. | `0.01` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Optional matrix to be encoded by trained model. | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix of training data (X). | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--codes_file (-c)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save the output sparse codes of the test matrix (--test_file) to. | 
+| `--dictionary_file (-d)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save the output dictionary to. | 
+| `--output_model_file (-M)` | [`SparseCoding file`](#doc_model) | File to save trained sparse coding model to. | 
+
+### Detailed documentation
+{: #sparse_coding_detailed-documentation }
+
+An implementation of Sparse Coding with Dictionary Learning, which achieves sparsity via an l1-norm regularizer on the codes (LASSO) or an (l1+l2)-norm regularizer on the codes (the Elastic Net).  Given a dense data matrix X with d dimensions and n points, sparse coding seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a sparse coding matrix Z with n points in k dimensions.
+
+The original data matrix X can then be reconstructed as Z * D.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The sparse coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a sparse coding step, which updates the sparse coding matrix.
+
+Once a dictionary D is found, the sparse coding model may be used to encode other matrices, and saved for future usage.
+
+To run this program, either an input matrix or an already-saved sparse coding model must be specified.  An input matrix may be specified with the `--training_file (-t)` option, along with the number of atoms in the dictionary (specified with the `--atoms (-k)` parameter).  It is also possible to specify an initial dictionary for the optimization, with the `--initial_dictionary_file (-i)` parameter.  An input model may be specified with the `--input_model_file (-m)` parameter.
+
+### Example
+As an example, to build a sparse coding model on the dataset `'data.csv'` using 200 atoms and an l1-regularization parameter of 0.1, saving the model into `'model.bin'`, use 
+
+```bash
+$ mlpack_sparse_coding --training_file data.csv --atoms 200 --lambda1 0.1
+  --output_model_file model.bin
+```
+
+Then, this model could be used to encode a new matrix, `'otherdata.csv'`, and save the output codes to `'codes.csv'`: 
+
+```bash
+$ mlpack_sparse_coding --input_model_file model.bin --test_file otherdata.csv
+  --codes_file codes.csv
+```
+
+### See also
+
+ - [mlpack_local_coordinate_coding](#local_coordinate_coding)
+ - [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ - [Efficient sparse coding algorithms (pdf)](http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf)
+ - [Regularization and variable selection via the elastic net](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=46217f372a75dddc2254fdbc6b9418ba3554e453)
+ - [SparseCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/sparse_coding/sparse_coding.hpp)
+
+## mlpack_adaboost
+{: #adaboost }
+
+#### AdaBoost
+{: #adaboost_descr }
+
+```bash
+$ mlpack_adaboost [--help] [--info <string>] [--input_model_file
+        <string>] [--iterations 1000] [--labels_file <string>] [--test_file
+        <string>] [--tolerance 1e-10] [--training_file <string>] [--verbose]
+        [--version] [--weak_learner 'decision_stump'] [--output_file <string>]
+        [--output_model_file <string>] [--predictions_file <string>]
+        [--probabilities_file <string>]
+```
+
+An implementation of the AdaBoost.MH (Adaptive Boosting) algorithm for classification.  This can be used to train an AdaBoost model on labeled data or use an existing AdaBoost model to predict the classes of new points. [Detailed documentation](#adaboost_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`AdaBoostModel file`](#doc_model) | Input AdaBoost model. | `''` |
+| `--iterations (-i)` | [`int`](#doc_int) | The maximum number of boosting iterations to be run (0 will run until convergence.) | `1000` |
+| `--labels_file (-l)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Labels for the training set. | `''` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Test dataset. | `''` |
+| `--tolerance (-e)` | [`double`](#doc_double) | The tolerance for change in values of the weighted error during training. | `1e-10` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Dataset for training AdaBoost. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--weak_learner (-w)` | [`string`](#doc_string) | The type of weak learner to use: 'decision_stump', or 'perceptron'. | `'decision_stump'` |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Predicted labels for the test set. | 
+| `--output_model_file (-M)` | [`AdaBoostModel file`](#doc_model) | Output trained AdaBoost model. | 
+| `--predictions_file (-P)` | [`1-d index matrix file`](#doc_a_1_d_index_matrix_file) | Predicted labels for the test set. | 
+| `--probabilities_file (-p)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #adaboost_detailed-documentation }
+
+This program implements the AdaBoost (or Adaptive Boosting) algorithm. The variant of AdaBoost implemented here is AdaBoost.MH. It uses a weak learner, either decision stumps or perceptrons, and over many iterations, creates a strong learner that is a weighted ensemble of weak learners. It runs these iterations until a tolerance value is crossed for change in the value of the weighted training error.
+
+For more information about the algorithm, see the paper "Improved Boosting Algorithms Using Confidence-Rated Predictions", by R.E. Schapire and Y. Singer.
+
+This program allows training of an AdaBoost model, and then application of that model to a test dataset.  To train a model, a dataset must be passed with the `--training_file (-t)` option.  Labels can be given with the `--labels_file (-l)` option; if no labels are specified, the labels will be assumed to be the last column of the input dataset.  Alternately, an AdaBoost model may be loaded with the `--input_model_file (-m)` option.
+
+Once a model is trained or loaded, it may be used to provide class predictions for a given test dataset.  A test dataset may be specified with the `--test_file (-T)` parameter.  The predicted classes for each point in the test dataset are output to the `--predictions_file (-P)` output parameter.  The AdaBoost model itself is output to the `--output_model_file (-M)` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `--output_file (-o)`.
+Use `--predictions_file (-P)` instead of `--output_file (-o)`.
+
+### Example
+For example, to run AdaBoost on an input dataset `'data.csv'` with labels `'labels.csv'`and perceptrons as the weak learner type, storing the trained model in `'model.bin'`, one could use the following command: 
+
+```bash
+$ mlpack_adaboost --training_file data.csv --labels_file labels.csv
+  --output_model_file model.bin --weak_learner perceptron
+```
+
+Similarly, an already-trained model in `'model.bin'` can be used to provide class predictions from test data `'test_data.csv'` and store the output in `'predictions.csv'` with the following command: 
+
+```bash
+$ mlpack_adaboost --input_model_file model.bin --test_file test_data.csv
+  --predictions_file predictions.csv
+```
+
+### See also
+
+ - [AdaBoost on Wikipedia](https://en.wikipedia.org/wiki/AdaBoost)
+ - [Improved boosting algorithms using confidence-rated predictions (pdf)](http://rob.schapire.net/papers/SchapireSi98.pdf)
+ - [Perceptron](#perceptron)
+ - [Decision Trees](#decision_tree)
+ - [AdaBoost C++ class documentation](../../user/methods/adaboost.md)
+
+## mlpack_linear_regression
+{: #linear_regression }
+
+#### Simple Linear Regression and Prediction
+{: #linear_regression_descr }
+
+```bash
+$ mlpack_linear_regression [--help] [--info <string>]
+        [--input_model_file <string>] [--lambda 0] [--test_file <string>]
+        [--training_file <string>] [--training_responses_file <string>]
+        [--verbose] [--version] [--output_model_file <string>]
+        [--output_predictions_file <string>]
+```
+
+An implementation of simple linear regression and ridge regression using ordinary least squares.  Given a dataset and responses, a model can be trained and saved for later use, or a pre-trained model can be used to output regression predictions for a test set. [Detailed documentation](#linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`LinearRegression<> file`](#doc_model) | Existing LinearRegression model to use. | `''` |
+| `--lambda (-l)` | [`double`](#doc_double) | Tikhonov regularization for ridge regression.  If 0, the method reduces to linear regression. | `0` |
+| `--test_file (-T)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing X' (test regressors). | `''` |
+| `--training_file (-t)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing training set X (regressors). | `''` |
+| `--training_responses_file (-r)` | [`1-d matrix file`](#doc_a_1_d_matrix_file) | Optional vector containing y (responses). If not given, the responses are assumed to be the last row of the input file. | `''` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_model_file (-M)` | [`LinearRegression<> file`](#doc_model) | Output LinearRegression model. | 
+| `--output_predictions_file (-o)` | [`1-d matrix file`](#doc_a_1_d_matrix_file) | If --test_file is specified, this matrix is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #linear_regression_detailed-documentation }
+
+An implementation of simple linear regression and simple ridge regression using ordinary least squares. This solves the problem
+
+  y = X * b + e
+
+where X (specified by `--training_file (-t)`) and y (specified either as the last column of the input matrix `--training_file (-t)` or via the `--training_responses_file (-r)` parameter) are known and b is the desired variable.  If the covariance matrix (X'X) is not invertible, or if the solution is overdetermined, then specify a Tikhonov regularization constant (with `--lambda (-l)`) greater than 0, which will regularize the covariance matrix to make it invertible.  The calculated b may be saved with the `--output_predictions_file (-o)` output parameter.
+
+Optionally, the calculated value of b is used to predict the responses for another matrix X' (specified by the `--test_file (-T)` parameter):
+
+   y' = X' * b
+
+and the predicted responses y' may be saved with the `--output_predictions_file (-o)` output parameter.  This type of regression is related to least-angle regression, which mlpack implements as the 'lars' program.
+
+### Example
+For example, to run a linear regression on the dataset `'X.csv'` with responses `'y.csv'`, saving the trained model to `'lr_model.bin'`, the following command could be used:
+
+```bash
+$ mlpack_linear_regression --training_file X.csv --training_responses_file
+  y.csv --output_model_file lr_model.bin
+```
+
+Then, to use `'lr_model.bin'` to predict responses for a test set `'X_test.csv'`, saving the predictions to `'X_test_responses.csv'`, the following command could be used:
+
+```bash
+$ mlpack_linear_regression --input_model_file lr_model.bin --test_file
+  X_test.csv --output_predictions_file X_test_responses.csv
+```
+
+### See also
+
+ - [mlpack_lars](#lars)
+ - [Linear regression on Wikipedia](https://en.wikipedia.org/wiki/Linear_regression)
+ - [LinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_regression/linear_regression.hpp)
+
+## mlpack_preprocess_imputer
+{: #preprocess_imputer }
+
+#### Impute Data
+{: #preprocess_imputer_descr }
+
+```bash
+$ mlpack_preprocess_imputer [--custom_value 0] [--dimension 0] [--help]
+        [--info <string>] --input_file <string> --missing_value <string>
+        --strategy <string> [--verbose] [--version] [--output_file <string>]
+```
+
+This utility provides several imputation strategies for missing data. Given a dataset with missing values, this can impute according to several strategies, including user-defined values. [Detailed documentation](#preprocess_imputer_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--custom_value (-c)` | [`double`](#doc_double) | User-defined custom imputation value. | `0` |
+| `--dimension (-d)` | [`int`](#doc_int) | The dimension to apply imputation to. | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_file (-i)` | [`string`](#doc_string) | File containing data. | `**--**` |
+| `--missing_value (-m)` | [`string`](#doc_string) | User defined missing value. | `**--**` |
+| `--strategy (-s)` | [`string`](#doc_string) | imputation strategy to be applied. Strategies should be one of 'custom', 'mean', 'median', and 'listwise_deletion'. | `**--**` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`string`](#doc_string) | File to save output into. | 
+
+### Detailed documentation
+{: #preprocess_imputer_detailed-documentation }
+
+This utility takes a dataset and converts a user-defined missing variable to another to provide more meaningful analysis.
+
+The program does not modify the original file, but instead makes a separate file to save the output data; You can save the output by specifying the file name with`--output_file (-o)`.
+
+### Example
+For example, if we consider 'NULL' in dimension 0 to be a missing variable and want to delete whole row containing the NULL in the column-wise`'dataset.csv'`, and save the result to `'result.csv'`, we could run :
+
+```bash
+$ mlpack_preprocess_imputer --input_file dataset --output_file result
+  --missing_value NULL --dimension 0 --strategy listwise_deletion
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_describe](#preprocess_describe)
+ - [mlpack_preprocess_split](#preprocess_split)
+
+## mlpack_image_converter
+{: #image_converter }
+
+#### Image Converter
+{: #image_converter_descr }
+
+```bash
+$ mlpack_image_converter [--channels 0] [--dataset_file <string>]
+        [--height 0] [--help] [--info <string>] --input [] [--quality 90]
+        [--save] [--verbose] [--version] [--width 0] [--output_file <string>]
+```
+
+A utility to load an image or set of images into a single dataset that can then be used by other mlpack methods and utilities. This can also unpack an image dataset into individual files, for instance after mlpack methods have been used. [Detailed documentation](#image_converter_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--channels (-c)` | [`int`](#doc_int) | Number of channels in the image. | `0` |
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--dataset_file (-I)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Input matrix to save as images. | `''` |
+| `--height (-H)` | [`int`](#doc_int) | Height of the images. | `0` |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input (-i)` | [`string vector`](#doc_string_vector) | Image filenames which have to be loaded/saved. | `**--**` |
+| `--quality (-q)` | [`int`](#doc_int) | Compression of the image if saved as jpg (0-100). | `90` |
+| `--save (-s)` | [`flag`](#doc_flag) | Save a dataset as images. |  |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--width (-w)` | [`int`](#doc_int) | Width of the image. | `0` |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--output_file (-o)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix to save images data to, Onlyneeded if you are specifying 'save' option. | 
+
+### Detailed documentation
+{: #image_converter_detailed-documentation }
+
+This utility takes an image or an array of images and loads them to a matrix. You can optionally specify the height `--height (-H)` width `--width (-w)` and channel `--channels (-c)` of the images that needs to be loaded; otherwise, these parameters will be automatically detected from the image.
+There are other options too, that can be specified such as `--quality (-q)`.
+
+You can also provide a dataset and save them as images using `--dataset_file (-I)` and `--save (-s)` as an parameter.
+
+### Example
+ An example to load an image : 
+
+```bash
+$ mlpack_image_converter --input X --height 256 --width 256 --channels 3
+  --output_file Y.csv
+```
+
+ An example to save an image is :
+
+```bash
+$ mlpack_image_converter --input X --height 256 --width 256 --channels 3
+  --dataset_file Y.csv --save
+```
+
+### See also
+
+ - [mlpack_preprocess_binarize](#preprocess_binarize)
+ - [mlpack_preprocess_describe](#preprocess_describe)
+
+## mlpack_range_search
+{: #range_search }
+
+#### Range Search
+{: #range_search_descr }
+
+```bash
+$ mlpack_range_search [--help] [--info <string>] [--input_model_file
+        <string>] [--leaf_size 20] [--max 0] [--min 0] [--naive] [--query_file
+        <string>] [--random_basis] [--reference_file <string>] [--seed 0]
+        [--single_mode] [--tree_type 'kd'] [--verbose] [--version]
+        [--distances_file <string>] [--neighbors_file <string>]
+        [--output_model_file <string>]
+```
+
+An implementation of range search with single-tree and dual-tree algorithms.  Given a set of reference points and a set of query points and a range, this can find the set of reference points within the desired range for each query point, and any trees built during the computation can be saved for reuse with future range searches. [Detailed documentation](#range_search_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `--check_input_matrices` | [`flag`](#doc_flag) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. |  |
+| `--help (-h)` | [`flag`](#doc_flag) | Default help info.  <span class="special">Only exists in CLI binding.</span> |  |
+| `--info` | [`string`](#doc_string) | Print help on a specific option.  <span class="special">Only exists in CLI binding.</span> | `''` |
+| `--input_model_file (-m)` | [`RSModel file`](#doc_model) | File containing pre-trained range search model. | `''` |
+| `--leaf_size (-l)` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `--max (-U)` | [`double`](#doc_double) | Upper bound in range (if not specified, +inf will be used. | `0` |
+| `--min (-L)` | [`double`](#doc_double) | Lower bound in range. | `0` |
+| `--naive (-N)` | [`flag`](#doc_flag) | If true, O(n^2) naive mode is used for computation. |  |
+| `--query_file (-q)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | File containing query points (optional). | `''` |
+| `--random_basis (-R)` | [`flag`](#doc_flag) | Before tree-building, project the data onto a random orthogonal basis. |  |
+| `--reference_file (-r)` | [`2-d matrix file`](#doc_a_2_d_matrix_file) | Matrix containing the reference dataset. | `''` |
+| `--seed (-s)` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `--single_mode (-S)` | [`flag`](#doc_flag) | If true, single-tree search is used (as opposed to dual-tree search). |  |
+| `--tree_type (-t)` | [`string`](#doc_string) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `'kd'` |
+| `--verbose (-v)` | [`flag`](#doc_flag) | Display informational messages and the full list of parameters and timers at the end of execution. |  |
+| `--version (-V)` | [`flag`](#doc_flag) | Display the version of mlpack.  <span class="special">Only exists in CLI binding.</span> |  |
+
+### Output options
+
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `--distances_file (-d)` | [`string`](#doc_string) | File to output distances into. | 
+| `--neighbors_file (-n)` | [`string`](#doc_string) | File to output neighbors into. | 
+| `--output_model_file (-M)` | [`RSModel file`](#doc_model) | If specified, the range search model will be saved to the given file. | 
+
+### Detailed documentation
+{: #range_search_detailed-documentation }
+
+This program implements range search with a Euclidean distance metric. For a given query point, a given range, and a given set of reference points, the program will return all of the reference points with distance to the query point in the given range.  This is performed for an entire set of query points. You may specify a separate set of reference and query points, or only a reference set -- which is then used as both the reference and query set.  The given range is taken to be inclusive (that is, points with a distance exactly equal to the minimum and maximum of the range are included in the results).
+
+### Example
+For example, the following will calculate the points within the range `[2, 5]` of each point in `'input.csv'` and store the distances in`'distances.csv'` and the neighbors in `'neighbors.csv'`
+
+```bash
+$ mlpack_range_search --min 2 --max 5 --distances_file input --distances_file
+  distances --neighbors_file neighbors
+```
+
+The output files are organized such that line i corresponds to the points found for query point i.  Because sometimes 0 points may be found in the given range, lines of the output files may be empty.  The points are not ordered in any specific manner.
+
+Because the number of points returned for each query point may differ, the resultant CSV-like files may not be loadable by many programs.  However, at this time a better way to store this non-square result is not known.  As a result, any output files will be written as CSVs in this manner, regardless of the given extension.
+
+### See also
+
+ - [mlpack_knn](#knn)
+ - [Range search tutorial](../../tutorials/range_search.md)
+ - [Range searching on Wikipedia](https://en.wikipedia.org/wiki/Range_searching)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [RangeSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/range_search/range_search.hpp)
+
diff -pruN 4.3.0-2/doc/user/bindings/cli.sidebar.html 4.4.0-1/doc/user/bindings/cli.sidebar.html
--- 4.3.0-2/doc/user/bindings/cli.sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/cli.sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,102 @@
+<ul>
+<li class="page_title">
+<b>mlpack CLI binding documentation</b>
+<a href="#">[top]</a>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#mlpack-overview">Overview</a></li>
+<li><a href="LINKROOTquickstart/cli.html">CLI Quickstart</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#data-formats">Data Formats</a></li>
+<li><details><summary>Classification</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#decision_tree">mlpack_decision_tree</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#linear_svm">mlpack_linear_svm</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#logistic_regression">mlpack_logistic_regression</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#nbc">mlpack_nbc</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#perceptron">mlpack_perceptron</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#random_forest">mlpack_random_forest</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#softmax_regression">mlpack_softmax_regression</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#adaboost">mlpack_adaboost</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Regression</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#bayesian_linear_regression">mlpack_bayesian_linear_regression</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#lars">mlpack_lars</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#linear_regression">mlpack_linear_regression</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Clustering</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#dbscan">mlpack_dbscan</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#gmm_train">mlpack_gmm_train</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#gmm_generate">mlpack_gmm_generate</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#gmm_probability">mlpack_gmm_probability</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#hoeffding_tree">mlpack_hoeffding_tree</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#kmeans">mlpack_kmeans</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#mean_shift">mlpack_mean_shift</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Geometry</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#approx_kfn">mlpack_approx_kfn</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#emst">mlpack_emst</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#fastmks">mlpack_fastmks</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#lsh">mlpack_lsh</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#knn">mlpack_knn</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#kfn">mlpack_kfn</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#krann">mlpack_krann</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#range_search">mlpack_range_search</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Preprocessing</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_split">mlpack_preprocess_split</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_binarize">mlpack_preprocess_binarize</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_describe">mlpack_preprocess_describe</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_scale">mlpack_preprocess_scale</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_one_hot_encoding">mlpack_preprocess_one_hot_encoding</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#preprocess_imputer">mlpack_preprocess_imputer</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#image_converter">mlpack_image_converter</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Misc. / Other</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#cf">mlpack_cf</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#det">mlpack_det</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#hmm_train">mlpack_hmm_train</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#hmm_generate">mlpack_hmm_generate</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#hmm_loglik">mlpack_hmm_loglik</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#hmm_viterbi">mlpack_hmm_viterbi</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#kde">mlpack_kde</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#nmf">mlpack_nmf</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Transformations</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/cli.html#kernel_pca">mlpack_kernel_pca</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#lmnn">mlpack_lmnn</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#local_coordinate_coding">mlpack_local_coordinate_coding</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#nca">mlpack_nca</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#pca">mlpack_pca</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#radical">mlpack_radical</a></li>
+<li><a href="LINKROOTuser/bindings/cli.html#sparse_coding">mlpack_sparse_coding</a></li>
+</ul>
+</details>
+</li>
+
+</ul>
+</li>
+</ul>
+</div>
diff -pruN 4.3.0-2/doc/user/bindings/go.md 4.4.0-1/doc/user/bindings/go.md
--- 4.3.0-2/doc/user/bindings/go.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/go.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,4455 @@
+# mlpack Go binding documentation
+
+## mlpack overview
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning library with bindings to other languages.  It aims to provide fast, lightweight implementations of both common and cutting-edge machine learning algorithms.
+
+This reference page details mlpack's bindings to Go.
+
+Further useful mlpack documentation links are given below.
+
+ - [mlpack homepage](https://www.mlpack.org/)
+ - [mlpack on Github](https://github.com/mlpack/mlpack)
+ - [mlpack main documentation page](https://www.mlpack.org/docs.html)
+
+See also the quickstart guide for Go:
+
+ - [Go Quickstart](../../quickstart/go.md)
+
+## Data Formats
+
+<div id="data-formats-div" markdown="1">
+mlpack bindings for Go take and return a restricted set of types, for simplicity.  These include primitive types, matrix/vector types, categorical matrix types, and model types. Each type is detailed below.
+
+ - `int`{: #doc_int}: An integer (i.e., `1`).
+ - `float64`{: #doc_float64 }: A floating-point number (i.e., `0.5`).
+ - `bool`{: #doc_bool }: A boolean flag option (`true` or `false`).
+ - `string`{: #doc_string }: A character string (i.e., `"hello"`).
+ - `array of ints`{: #doc_array_of_ints }: An array of integers; i.e., `[]int{0, 1, 2}`.
+ - `array of strings`{: #doc_array_of_strings }: An array of strings; i.e., `[]string{"hello", "goodbye"}`.
+ - `*mat.Dense`{: #doc_a__mat_Dense }: A 2-d gonum Matrix. If the type is not already `float64`, it will be converted.
+ - `*mat.Dense (1d)`{: #doc_a__mat_Dense__1d_ }: A 1-d gonum Matrix (that is, a Matrix where either the number of rows or number of columns is 1).
+ - `matrixWithInfo`{: #doc_matrixWithInfo }: A Tuple(matrixWithInfo) containing `float64` data (Data) along with a boolean array (Categoricals) indicating which dimensions are categorical (represented by `true`) and which are numeric (represented by `false`).  The number of elements in the boolean array should be the same as the dimensionality of the data matrix.  It is expected that each row of the matrix corresponds to a single data point when calling mlpack bindings.
+ - `mlpackModel`{: #doc_model }: An mlpack model pointer.  This type holds a pointer to C++ memory containing the mlpack model.  Note that this means the mlpack model itself cannot be easily inspected in Go.  However, the pointer can be passed to subsequent calls to mlpack functions.
+</div>
+
+
+## ApproxKfn()
+{: #approx_kfn }
+
+#### Approximate furthest neighbor search
+{: #approx_kfn_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for ApproxKfn().
+param := mlpack.ApproxKfnOptions()
+param.Algorithm = "ds"
+param.CalculateError = false
+param.ExactDistances = mat.NewDense(1, 1, nil)
+param.InputModel = nil
+param.K = 0
+param.NumProjections = 5
+param.NumTables = 5
+param.Query = mat.NewDense(1, 1, nil)
+param.Reference = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+distances, neighbors, output_model := mlpack.ApproxKfn(param)
+```
+
+An implementation of two strategies for furthest neighbor search.  This can be used to compute the furthest neighbor of query point(s) from a set of points; furthest neighbor models can be saved and reused with future query point(s). [Detailed documentation](#approx_kfn_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Algorithm` | [`string`](#doc_string) | Algorithm to use: 'ds' or 'qdafn'. | `"ds"` |
+| `CalculateError` | [`bool`](#doc_bool) | If set, calculate the average distance error for the first furthest neighbor only. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `ExactDistances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing exact distances to furthest neighbors; this can be used to avoid explicit calculation when --calculate_error is set. | `mat.NewDense(1, 1, nil)` |
+| `InputModel` | [`approxkfnModel`](#doc_model) | File containing input model. | `nil` |
+| `K` | [`int`](#doc_int) | Number of furthest neighbors to search for. | `0` |
+| `NumProjections` | [`int`](#doc_int) | Number of projections to use in each hash table. | `5` |
+| `NumTables` | [`int`](#doc_int) | Number of hash tables to use. | `5` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing query points. | `mat.NewDense(1, 1, nil)` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing the reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Distances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save furthest neighbor distances to. | 
+| `Neighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save neighbor indices to. | 
+| `OutputModel` | [`approxkfnModel`](#doc_model) | File to save output model to. | 
+
+### Detailed documentation
+{: #approx_kfn_detailed-documentation }
+
+This program implements two strategies for furthest neighbor search. These strategies are:
+
+ - The 'qdafn' algorithm from "Approximate Furthest Neighbor in High Dimensions" by R. Pagh, F. Silvestri, J. Sivertsen, and M. Skala, in Similarity Search and Applications 2015 (SISAP).
+ - The 'DrusillaSelect' algorithm from "Fast approximate furthest neighbors with data-dependent candidate selection", by R.R. Curtin and A.B. Gardner, in Similarity Search and Applications 2016 (SISAP).
+
+These two strategies give approximate results for the furthest neighbor search problem and can be used as fast replacements for other furthest neighbor techniques such as those found in the mlpack_kfn program.  Note that typically, the 'ds' algorithm requires far fewer tables and projections than the 'qdafn' algorithm.
+
+Specify a reference set (set to search in) with `Reference`, specify a query set with `Query`, and specify algorithm parameters with `NumTables` and `NumProjections` (or don't and defaults will be used).  The algorithm to be used (either 'ds'---the default---or 'qdafn')  may be specified with `Algorithm`.  Also specify the number of neighbors to search for with `K`.
+
+Note that for 'qdafn' in lower dimensions, `NumProjections` may need to be set to a high value in order to return results for each query point.
+
+If no query set is specified, the reference set will be used as the query set.  The `OutputModel` output parameter may be used to store the built model, and an input model may be loaded instead of specifying a reference set with the `InputModel` option.
+
+Results for each query point can be stored with the `Neighbors` and `Distances` output parameters.  Each row of these output matrices holds the k distances or neighbor indices for each query point.
+
+### Example
+For example, to find the 5 approximate furthest neighbors with `reference_set` as the reference set and `query_set` as the query set using DrusillaSelect, storing the furthest neighbor indices to `neighbors` and the furthest neighbor distances to `distances`, one could call
+
+```go
+// Initialize optional parameters for ApproxKfn().
+param := mlpack.ApproxKfnOptions()
+param.Query = query_set
+param.Reference = reference_set
+param.K = 5
+param.Algorithm = "ds"
+
+distances, neighbors, _ := mlpack.ApproxKfn(param)
+```
+
+and to perform approximate all-furthest-neighbors search with k=1 on the set `data` storing only the furthest neighbor distances to `distances`, one could call
+
+```go
+// Initialize optional parameters for ApproxKfn().
+param := mlpack.ApproxKfnOptions()
+param.Reference = reference_set
+param.K = 1
+
+distances, _, _ := mlpack.ApproxKfn(param)
+```
+
+A trained model can be re-used.  If a model has been previously saved to `model`, then we may find 3 approximate furthest neighbors on a query set `new_query_set` using that model and store the furthest neighbor indices into `neighbors` by calling
+
+```go
+// Initialize optional parameters for ApproxKfn().
+param := mlpack.ApproxKfnOptions()
+param.InputModel = &model
+param.Query = new_query_set
+param.K = 3
+
+_, neighbors, _ := mlpack.ApproxKfn(param)
+```
+
+### See also
+
+ - [k-furthest-neighbor search](#kfn)
+ - [k-nearest-neighbor search](#knn)
+ - [Fast approximate furthest neighbors with data-dependent candidate selection (pdf)](http://ratml.org/pub/pdf/2016fast.pdf)
+ - [Approximate furthest neighbor in high dimensions (pdf)](https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf)
+ - [QDAFN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/qdafn.hpp)
+ - [DrusillaSelect class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/drusilla_select.hpp)
+
+## BayesianLinearRegression()
+{: #bayesian_linear_regression }
+
+#### BayesianLinearRegression
+{: #bayesian_linear_regression_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for BayesianLinearRegression().
+param := mlpack.BayesianLinearRegressionOptions()
+param.Center = false
+param.Input = mat.NewDense(1, 1, nil)
+param.InputModel = nil
+param.Responses = mat.NewDense(1, 1, nil)
+param.Scale = false
+param.Test = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, predictions, stds := mlpack.BayesianLinearRegression(param)
+```
+
+An implementation of the bayesian linear regression. [Detailed documentation](#bayesian_linear_regression_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Center` | [`bool`](#doc_bool) | Center the data and fit the intercept if enabled. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of covariates (X). | `mat.NewDense(1, 1, nil)` |
+| `InputModel` | [`bayesianLinearRegression`](#doc_model) | Trained BayesianLinearRegression model to use. | `nil` |
+| `Responses` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Matrix of responses/observations (y). | `mat.NewDense(1, 1, nil)` |
+| `Scale` | [`bool`](#doc_bool) | Scale each feature by their standard deviations if enabled. | `false` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing points to regress on (test points). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`bayesianLinearRegression`](#doc_model) | Output BayesianLinearRegression model. | 
+| `Predictions` | [`*mat.Dense`](#doc_a__mat_Dense) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+| `Stds` | [`*mat.Dense`](#doc_a__mat_Dense) | If specified, this is where the standard deviations of the predictive distribution will be saved. | 
+
+### Detailed documentation
+{: #bayesian_linear_regression_detailed-documentation }
+
+An implementation of the bayesian linear regression.
+This model is a probabilistic view and implementation of the linear regression. The final solution is obtained by computing a posterior distribution from gaussian likelihood and a zero mean gaussian isotropic  prior distribution on the solution. 
+Optimization is AUTOMATIC and does not require cross validation. The optimization is performed by maximization of the evidence function. Parameters are tuned during the maximization of the marginal likelihood. This procedure includes the Ockham's razor that penalizes over complex solutions. 
+
+This program is able to train a Bayesian linear regression model or load a model from file, output regression predictions for a test set, and save the trained model to a file.
+
+To train a BayesianLinearRegression model, the `Input` and `Responses`parameters must be given. The `Center`and `Scale` parameters control the centering and the normalizing options. A trained model can be saved with the `OutputModel`. If no training is desired at all, a model can be passed via the `InputModel` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `Test` parameter.  Predicted responses to the test points can be saved with the `Predictions` output parameter. The corresponding standard deviation can be save by precising the `Stds` parameter.
+
+### Example
+For example, the following command trains a model on the data `data` and responses `responses`with center set to true and scale set to false (so, Bayesian linear regression is being solved, and then the model is saved to `blr_model`:
+
+```go
+// Initialize optional parameters for BayesianLinearRegression().
+param := mlpack.BayesianLinearRegressionOptions()
+param.Input = data
+param.Responses = responses
+param.Center = 1
+param.Scale = 0
+
+blr_model, _, _ := mlpack.BayesianLinearRegression(param)
+```
+
+The following command uses the `blr_model` to provide predicted  responses for the data `test` and save those  responses to `test_predictions`: 
+
+```go
+// Initialize optional parameters for BayesianLinearRegression().
+param := mlpack.BayesianLinearRegressionOptions()
+param.InputModel = &blr_model
+param.Test = test
+
+_, test_predictions, _ := mlpack.BayesianLinearRegression(param)
+```
+
+Because the estimator computes a predictive distribution instead of a simple point estimate, the `Stds` parameter allows one to save the prediction uncertainties: 
+
+```go
+// Initialize optional parameters for BayesianLinearRegression().
+param := mlpack.BayesianLinearRegressionOptions()
+param.InputModel = &blr_model
+param.Test = test
+
+_, test_predictions, stds := mlpack.BayesianLinearRegression(param)
+```
+
+### See also
+
+ - [Bayesian Interpolation](https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf)
+ - [Bayesian Linear Regression, Section 3.3](https://www.microsoft.com/en-us/research/uploads/prod/2006/01/Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf)
+ - [BayesianLinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp)
+
+## Cf()
+{: #cf }
+
+#### Collaborative Filtering
+{: #cf_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Cf().
+param := mlpack.CfOptions()
+param.Algorithm = "NMF"
+param.AllUserRecommendations = false
+param.InputModel = nil
+param.Interpolation = "average"
+param.IterationOnlyTermination = false
+param.MaxIterations = 1000
+param.MinResidue = 1e-05
+param.NeighborSearch = "euclidean"
+param.Neighborhood = 5
+param.Normalization = "none"
+param.Query = mat.NewDense(1, 1, nil)
+param.Rank = 0
+param.Recommendations = 5
+param.Seed = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output, output_model := mlpack.Cf(param)
+```
+
+An implementation of several collaborative filtering (CF) techniques for recommender systems.  This can be used to train a new CF model, or use an existing CF model to compute recommendations. [Detailed documentation](#cf_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Algorithm` | [`string`](#doc_string) | Algorithm used for matrix factorization. | `"NMF"` |
+| `AllUserRecommendations` | [`bool`](#doc_bool) | Generate recommendations for all users. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`cfModel`](#doc_model) | Trained CF model to load. | `nil` |
+| `Interpolation` | [`string`](#doc_string) | Algorithm used for weight interpolation. | `"average"` |
+| `IterationOnlyTermination` | [`bool`](#doc_bool) | Terminate only when the maximum number of iterations is reached. | `false` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations. If set to zero, there is no limit on the number of iterations. | `1000` |
+| `MinResidue` | [`float64`](#doc_float64) | Residue required to terminate the factorization (lower values generally mean better fits). | `1e-05` |
+| `NeighborSearch` | [`string`](#doc_string) | Algorithm used for neighbor search. | `"euclidean"` |
+| `Neighborhood` | [`int`](#doc_int) | Size of the neighborhood of similar users to consider for each query user. | `5` |
+| `Normalization` | [`string`](#doc_string) | Normalization performed on the ratings. | `"none"` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | List of query users for which recommendations should be generated. | `mat.NewDense(1, 1, nil)` |
+| `Rank` | [`int`](#doc_int) | Rank of decomposed matrices (if 0, a heuristic is used to estimate the rank). | `0` |
+| `Recommendations` | [`int`](#doc_int) | Number of recommendations to generate for each query user. | `5` |
+| `Seed` | [`int`](#doc_int) | Set the random seed (0 uses std::time(NULL)). | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Test set to calculate RMSE on. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform CF on. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix that will store output recommendations. | 
+| `OutputModel` | [`cfModel`](#doc_model) | Output for trained CF model. | 
+
+### Detailed documentation
+{: #cf_detailed-documentation }
+
+This program performs collaborative filtering (CF) on the given dataset. Given a list of user, item and preferences (the `Training` parameter), the program will perform a matrix decomposition and then can perform a series of actions related to collaborative filtering.  Alternately, the program can load an existing saved CF model with the `InputModel` parameter and then use that model to provide recommendations or predict values.
+
+The input matrix should be a 3-dimensional matrix of ratings, where the first dimension is the user, the second dimension is the item, and the third dimension is that user's rating of that item.  Both the users and items should be numeric indices, not names. The indices are assumed to start from 0.
+
+A set of query users for which recommendations can be generated may be specified with the `Query` parameter; alternately, recommendations may be generated for every user in the dataset by specifying the `AllUserRecommendations` parameter.  In addition, the number of recommendations per user to generate can be specified with the `Recommendations` parameter, and the number of similar users (the size of the neighborhood) to be considered when generating recommendations can be specified with the `Neighborhood` parameter.
+
+For performing the matrix decomposition, the following optimization algorithms can be specified via the `Algorithm` parameter:
+
+ - 'RegSVD' -- Regularized SVD using a SGD optimizer
+ - 'NMF' -- Non-negative matrix factorization with alternating least squares update rules
+ - 'BatchSVD' -- SVD batch learning
+ - 'SVDIncompleteIncremental' -- SVD incomplete incremental learning
+ - 'SVDCompleteIncremental' -- SVD complete incremental learning
+ - 'BiasSVD' -- Bias SVD using a SGD optimizer
+ - 'SVDPP' -- SVD++ using a SGD optimizer
+ - 'RandSVD' -- RandomizedSVD learning
+ - 'QSVD' -- QuicSVD learning
+ - 'BKSVD' -- Block Krylov SVD learning
+
+
+The following neighbor search algorithms can be specified via the `NeighborSearch` parameter:
+
+ - 'cosine'  -- Cosine Search Algorithm
+ - 'euclidean'  -- Euclidean Search Algorithm
+ - 'pearson'  -- Pearson Search Algorithm
+
+
+The following weight interpolation algorithms can be specified via the `Interpolation` parameter:
+
+ - 'average'  -- Average Interpolation Algorithm
+ - 'regression'  -- Regression Interpolation Algorithm
+ - 'similarity'  -- Similarity Interpolation Algorithm
+
+
+The following ranking normalization algorithms can be specified via the `Normalization` parameter:
+
+ - 'none'  -- No Normalization
+ - 'item_mean'  -- Item Mean Normalization
+ - 'overall_mean'  -- Overall Mean Normalization
+ - 'user_mean'  -- User Mean Normalization
+ - 'z_score'  -- Z-Score Normalization
+
+A trained model may be saved to with the `OutputModel` output parameter.
+
+### Example
+To train a CF model on a dataset `training_set` using NMF for decomposition and saving the trained model to `model`, one could call: 
+
+```go
+// Initialize optional parameters for Cf().
+param := mlpack.CfOptions()
+param.Training = training_set
+param.Algorithm = "NMF"
+
+_, model := mlpack.Cf(param)
+```
+
+Then, to use this model to generate recommendations for the list of users in the query set `users`, storing 5 recommendations in `recommendations`, one could call 
+
+```go
+// Initialize optional parameters for Cf().
+param := mlpack.CfOptions()
+param.InputModel = &model
+param.Query = users
+param.Recommendations = 5
+
+recommendations, _ := mlpack.Cf(param)
+```
+
+### See also
+
+ - [Collaborative Filtering on Wikipedia](https://en.wikipedia.org/wiki/Collaborative_filtering)
+ - [Matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Matrix_factorization_(recommender_systems))
+ - [Matrix factorization techniques for recommender systems (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5)
+ - [CFType class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/cf/cf.hpp)
+
+## Dbscan()
+{: #dbscan }
+
+#### DBSCAN clustering
+{: #dbscan_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Dbscan().
+param := mlpack.DbscanOptions()
+param.Epsilon = 1
+param.MinSize = 5
+param.Naive = false
+param.SelectionType = "ordered"
+param.SingleMode = false
+param.TreeType = "kd"
+param.Verbose = false
+
+assignments, centroids := mlpack.Dbscan(input, param)
+```
+
+An implementation of DBSCAN clustering.  Given a dataset, this can compute and return a clustering of that dataset. [Detailed documentation](#dbscan_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Epsilon` | [`float64`](#doc_float64) | Radius of each range search. | `1` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to cluster. | `**--**` |
+| `MinSize` | [`int`](#doc_int) | Minimum number of points for a cluster. | `5` |
+| `Naive` | [`bool`](#doc_bool) | If set, brute-force range search (not tree-based) will be used. | `false` |
+| `SelectionType` | [`string`](#doc_string) | If using point selection policy, the type of selection to use ('ordered', 'random'). | `"ordered"` |
+| `SingleMode` | [`bool`](#doc_bool) | If set, single-tree range search (not dual-tree) will be used. | `false` |
+| `TreeType` | [`string`](#doc_string) | If using single-tree or dual-tree search, the type of tree to use ('kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'). | `"kd"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Assignments` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Output matrix for assignments of each point. | 
+| `Centroids` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save output centroids to. | 
+
+### Detailed documentation
+{: #dbscan_detailed-documentation }
+
+This program implements the DBSCAN algorithm for clustering using accelerated tree-based range search.  The type of tree that is used may be parameterized, or brute-force range search may also be used.
+
+The input dataset to be clustered may be specified with the `Input` parameter; the radius of each range search may be specified with the `Epsilon` parameters, and the minimum number of points in a cluster may be specified with the `MinSize` parameter.
+
+The `Assignments` and `Centroids` output parameters may be used to save the output of the clustering. `Assignments` contains the cluster assignments of each point, and `Centroids` contains the centroids of each cluster.
+
+The range search may be controlled with the `TreeType`, `SingleMode`, and `Naive` parameters.  `TreeType` can control the type of tree used for range search; this can take a variety of values: 'kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'. The `SingleMode` parameter will force single-tree search (as opposed to the default dual-tree search), and '`Naive` will force brute-force range search.
+
+### Example
+An example usage to run DBSCAN on the dataset in `input` with a radius of 0.5 and a minimum cluster size of 5 is given below:
+
+```go
+// Initialize optional parameters for Dbscan().
+param := mlpack.DbscanOptions()
+param.Epsilon = 0.5
+param.MinSize = 5
+
+_, _ := mlpack.Dbscan(input, param)
+```
+
+### See also
+
+ - [DBSCAN on Wikipedia](https://en.wikipedia.org/wiki/DBSCAN)
+ - [A density-based algorithm for discovering clusters in large spatial databases with noise (pdf)](https://cdn.aaai.org/KDD/1996/KDD96-037.pdf)
+ - [DBSCAN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/dbscan/dbscan.hpp)
+
+## DecisionTree()
+{: #decision_tree }
+
+#### Decision tree
+{: #decision_tree_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for DecisionTree().
+param := mlpack.DecisionTreeOptions()
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.MaximumDepth = 0
+param.MinimumGainSplit = 1e-07
+param.MinimumLeafSize = 20
+param.PrintTrainingAccuracy = false
+param.PrintTrainingError = false
+param.Test = mat.NewDense(1, 1, nil)
+param.TestLabels = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+param.Weights = mat.NewDense(1, 1, nil)
+
+output_model, predictions, probabilities := mlpack.DecisionTree(param)
+```
+
+An implementation of an ID3-style decision tree for classification, which supports categorical data.  Given labeled data with numeric or categorical features, a decision tree can be trained and saved; or, an existing decision tree can be used for classification on new points. [Detailed documentation](#decision_tree_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`decisionTreeModel`](#doc_model) | Pre-trained decision tree, to be used with test points. | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Training labels. | `mat.NewDense(1, 1, nil)` |
+| `MaximumDepth` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `MinimumGainSplit` | [`float64`](#doc_float64) | Minimum gain for node splitting. | `1e-07` |
+| `MinimumLeafSize` | [`int`](#doc_int) | Minimum number of points in a leaf. | `20` |
+| `PrintTrainingAccuracy` | [`bool`](#doc_bool) | Print the training accuracy. | `false` |
+| `PrintTrainingError` | [`bool`](#doc_bool) | Print the training error (deprecated; will be removed in mlpack 4.0.0). | `false` |
+| `Test` | [`matrixWithInfo`](#doc_matrixWithInfo) | Testing dataset (may be categorical). | `mat.NewDense(1, 1, nil)` |
+| `TestLabels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Test point labels, if accuracy calculation is desired. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`matrixWithInfo`](#doc_matrixWithInfo) | Training dataset (may be categorical). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `Weights` | [`*mat.Dense`](#doc_a__mat_Dense) | The weight of labels | `mat.NewDense(1, 1, nil)` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`decisionTreeModel`](#doc_model) | Output for trained decision tree. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Class predictions for each test point. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | Class probabilities for each test point. | 
+
+### Detailed documentation
+{: #decision_tree_detailed-documentation }
+
+Train and evaluate using a decision tree.  Given a dataset containing numeric or categorical features, and associated labels for each point in the dataset, this program can train a decision tree on that data.
+
+The training set and associated labels are specified with the `Training` and `Labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `Labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `OutputModel` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `InputModel` parameter.  The `InputModel` parameter may not be specified when the `Training` parameter is specified.  The `MinimumLeafSize` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `MinimumGainSplit` parameter specifies the minimum gain that is needed for the node to split.  The `MaximumDepth` parameter specifies the maximum depth of the tree.  If `PrintTrainingError` is specified, the training error will be printed.
+
+Test data may be specified with the `Test` parameter, and if performance numbers are desired for that test set, labels may be specified with the `TestLabels` parameter.  Predictions for each test point may be saved via the `Predictions` output parameter.  Class probabilities for each prediction may be saved with the `Probabilities` output parameter.
+
+### Example
+For example, to train a decision tree with a minimum leaf size of 20 on the dataset contained in `data` with labels `labels`, saving the output model to `tree` and printing the training error, one could call
+
+```go
+// Initialize optional parameters for DecisionTree().
+param := mlpack.DecisionTreeOptions()
+param.Training = data
+param.Labels = labels
+param.MinimumLeafSize = 20
+param.MinimumGainSplit = 0.001
+param.PrintTrainingAccuracy = true
+
+tree, _, _ := mlpack.DecisionTree(param)
+```
+
+Then, to use that model to classify points in `test_set` and print the test error given the labels `test_labels` using that model, while saving the predictions for each point to `predictions`, one could call 
+
+```go
+// Initialize optional parameters for DecisionTree().
+param := mlpack.DecisionTreeOptions()
+param.InputModel = &tree
+param.Test = test_set
+param.TestLabels = test_labels
+
+_, predictions, _ := mlpack.DecisionTree(param)
+```
+
+### See also
+
+ - [Random forest](#random_forest)
+ - [Decision trees on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+ - [Induction of Decision Trees (pdf)](http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf)
+ - [DecisionTree C++ class documentation](../../user/methods/decision_tree.md)
+
+## Det()
+{: #det }
+
+#### Density Estimation With Density Estimation Trees
+{: #det_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Det().
+param := mlpack.DetOptions()
+param.Folds = 10
+param.InputModel = nil
+param.MaxLeafSize = 10
+param.MinLeafSize = 5
+param.PathFormat = "lr"
+param.SkipPruning = false
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, tag_counters_file, tag_file, test_set_estimates,
+    training_set_estimates, vi := mlpack.Det(param)
+```
+
+An implementation of density estimation trees for the density estimation task.  Density estimation trees can be trained or used to predict the density at locations given by query points. [Detailed documentation](#det_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Folds` | [`int`](#doc_int) | The number of folds of cross-validation to perform for the estimation (0 is LOOCV) | `10` |
+| `InputModel` | [`dTree`](#doc_model) | Trained density estimation tree to load. | `nil` |
+| `MaxLeafSize` | [`int`](#doc_int) | The maximum size of a leaf in the unpruned, fully grown DET. | `10` |
+| `MinLeafSize` | [`int`](#doc_int) | The minimum size of a leaf in the unpruned, fully grown DET. | `5` |
+| `PathFormat` | [`string`](#doc_string) | The format of path printing: 'lr', 'id-lr', or 'lr-id'. | `"lr"` |
+| `SkipPruning` | [`bool`](#doc_bool) | Whether to bypass the pruning process and output the unpruned tree only. | `false` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | A set of test points to estimate the density of. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | The data set on which to build a density estimation tree. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`dTree`](#doc_model) | Output to save trained density estimation tree to. | 
+| `TagCountersFile` | [`string`](#doc_string) | The file to output the number of points that went to each leaf. | 
+| `TagFile` | [`string`](#doc_string) | The file to output the tags (and possibly paths) for each sample in the test set. | 
+| `TestSetEstimates` | [`*mat.Dense`](#doc_a__mat_Dense) | The output estimates on the test set from the final optimally pruned tree. | 
+| `TrainingSetEstimates` | [`*mat.Dense`](#doc_a__mat_Dense) | The output density estimates on the training set from the final optimally pruned tree. | 
+| `Vi` | [`*mat.Dense`](#doc_a__mat_Dense) | The output variable importance values for each feature. | 
+
+### Detailed documentation
+{: #det_detailed-documentation }
+
+This program performs a number of functions related to Density Estimation Trees.  The optimal Density Estimation Tree (DET) can be trained on a set of data (specified by `Training`) using cross-validation (with number of folds specified with the `Folds` parameter).  This trained density estimation tree may then be saved with the `OutputModel` output parameter.
+
+The variable importances (that is, the feature importance values for each dimension) may be saved with the `Vi` output parameter, and the density estimates for each training point may be saved with the `TrainingSetEstimates` output parameter.
+
+Enabling path printing for each node outputs the path from the root node to a leaf for each entry in the test set, or training set (if a test set is not provided).  Strings like 'LRLRLR' (indicating that traversal went to the left child, then the right child, then the left child, and so forth) will be output. If 'lr-id' or 'id-lr' are given as the `PathFormat` parameter, then the ID (tag) of every node along the path will be printed after or before the L or R character indicating the direction of traversal, respectively.
+
+This program also can provide density estimates for a set of test points, specified in the `Test` parameter.  The density estimation tree used for this task will be the tree that was trained on the given training points, or a tree given as the parameter `InputModel`.  The density estimates for the test points may be saved using the `TestSetEstimates` output parameter.
+
+### See also
+
+ - [Density estimation on Wikipedia](https://en.wikipedia.org/wiki/Density_estimation)
+ - [Density estimation trees (pdf)](https://www.mlpack.org/papers/det.pdf)
+ - [DTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/det/dtree.hpp)
+
+## Emst()
+{: #emst }
+
+#### Fast Euclidean Minimum Spanning Tree
+{: #emst_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Emst().
+param := mlpack.EmstOptions()
+param.LeafSize = 1
+param.Naive = false
+param.Verbose = false
+
+output := mlpack.Emst(input, param)
+```
+
+An implementation of the Dual-Tree Boruvka algorithm for computing the Euclidean minimum spanning tree of a set of input points. [Detailed documentation](#emst_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input data matrix. | `**--**` |
+| `LeafSize` | [`int`](#doc_int) | Leaf size in the kd-tree.  One-element leaves give the empirically best performance, but at the cost of greater memory requirements. | `1` |
+| `Naive` | [`bool`](#doc_bool) | Compute the MST using O(n^2) naive algorithm. | `false` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Output data.  Stored as an edge list. | 
+
+### Detailed documentation
+{: #emst_detailed-documentation }
+
+This program can compute the Euclidean minimum spanning tree of a set of input points using the dual-tree Boruvka algorithm.
+
+The set to calculate the minimum spanning tree of is specified with the `Input` parameter, and the output may be saved with the `Output` output parameter.
+
+The `LeafSize` parameter controls the leaf size of the kd-tree that is used to calculate the minimum spanning tree, and if the `Naive` option is given, then brute-force search is used (this is typically much slower in low dimensions).  The leaf size does not affect the results, but it may have some effect on the runtime of the algorithm.
+
+### Example
+For example, the minimum spanning tree of the input dataset `data` can be calculated with a leaf size of 20 and stored as `spanning_tree` using the following command:
+
+```go
+// Initialize optional parameters for Emst().
+param := mlpack.EmstOptions()
+param.LeafSize = 20
+
+spanning_tree := mlpack.Emst(data, param)
+```
+
+The output matrix is a three-dimensional matrix, where each row indicates an edge.  The first dimension corresponds to the lesser index of the edge; the second dimension corresponds to the greater index of the edge; and the third column corresponds to the distance between the two points.
+
+### See also
+
+ - [Minimum spanning tree on Wikipedia](https://en.wikipedia.org/wiki/Minimum_spanning_tree)
+ - [Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications (pdf)](https://www.mlpack.org/papers/emst.pdf)
+ - [DualTreeBoruvka class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/emst/dtb.hpp)
+
+## Fastmks()
+{: #fastmks }
+
+#### FastMKS (Fast Max-Kernel Search)
+{: #fastmks_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Fastmks().
+param := mlpack.FastmksOptions()
+param.Bandwidth = 1
+param.Base = 2
+param.Degree = 2
+param.InputModel = nil
+param.K = 0
+param.Kernel = "linear"
+param.Naive = false
+param.Offset = 0
+param.Query = mat.NewDense(1, 1, nil)
+param.Reference = mat.NewDense(1, 1, nil)
+param.Scale = 1
+param.Single = false
+param.Verbose = false
+
+indices, kernels, output_model := mlpack.Fastmks(param)
+```
+
+An implementation of the single-tree and dual-tree fast max-kernel search (FastMKS) algorithm.  Given a set of reference points and a set of query points, this can find the reference point with maximum kernel value for each query point; trained models can be reused for future queries. [Detailed documentation](#fastmks_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Bandwidth` | [`float64`](#doc_float64) | Bandwidth (for Gaussian, Epanechnikov, and triangular kernels). | `1` |
+| `Base` | [`float64`](#doc_float64) | Base to use during cover tree construction. | `2` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Degree` | [`float64`](#doc_float64) | Degree of polynomial kernel. | `2` |
+| `InputModel` | [`fastmksModel`](#doc_model) | Input FastMKS model to use. | `nil` |
+| `K` | [`int`](#doc_int) | Number of maximum kernels to find. | `0` |
+| `Kernel` | [`string`](#doc_string) | Kernel type to use: 'linear', 'polynomial', 'cosine', 'gaussian', 'epanechnikov', 'triangular', 'hyptan'. | `"linear"` |
+| `Naive` | [`bool`](#doc_bool) | If true, O(n^2) naive mode is used for computation. | `false` |
+| `Offset` | [`float64`](#doc_float64) | Offset of kernel (for polynomial and hyptan kernels). | `0` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | The query dataset. | `mat.NewDense(1, 1, nil)` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | The reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `Scale` | [`float64`](#doc_float64) | Scale of kernel (for hyptan kernel). | `1` |
+| `Single` | [`bool`](#doc_bool) | If true, single-tree search is used (as opposed to dual-tree search. | `false` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Indices` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix of indices. | 
+| `Kernels` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix of kernels. | 
+| `OutputModel` | [`fastmksModel`](#doc_model) | Output for FastMKS model. | 
+
+### Detailed documentation
+{: #fastmks_detailed-documentation }
+
+This program will find the k maximum kernels of a set of points, using a query set and a reference set (which can optionally be the same set). More specifically, for each point in the query set, the k points in the reference set with maximum kernel evaluations are found.  The kernel function used is specified with the `Kernel` parameter.
+
+### Example
+For example, the following command will calculate, for each point in the query set `query`, the five points in the reference set `reference` with maximum kernel evaluation using the linear kernel.  The kernel evaluations may be saved with the  `kernels` output parameter and the indices may be saved with the `indices` output parameter.
+
+```go
+// Initialize optional parameters for Fastmks().
+param := mlpack.FastmksOptions()
+param.K = 5
+param.Reference = reference
+param.Query = query
+param.Kernel = "linear"
+
+indices, kernels, _ := mlpack.Fastmks(param)
+```
+
+The output matrices are organized such that row i and column j in the indices matrix corresponds to the index of the point in the reference set that has j'th largest kernel evaluation with the point in the query set with index i.  Row i and column j in the kernels matrix corresponds to the kernel evaluation between those two points.
+
+This program performs FastMKS using a cover tree.  The base used to build the cover tree can be specified with the `Base` parameter.
+
+### See also
+
+ - [k-nearest-neighbor search](#knn)
+ - [Dual-tree Fast Exact Max-Kernel Search (pdf)](https://mlpack.org/papers/fmks.pdf)
+ - [FastMKS class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/fastmks/fastmks.hpp)
+
+## GmmTrain()
+{: #gmm_train }
+
+#### Gaussian Mixture Model (GMM) Training
+{: #gmm_train_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for GmmTrain().
+param := mlpack.GmmTrainOptions()
+param.DiagonalCovariance = false
+param.InputModel = nil
+param.KmeansMaxIterations = 1000
+param.MaxIterations = 250
+param.NoForcePositive = false
+param.Noise = 0
+param.Percentage = 0.02
+param.RefinedStart = false
+param.Samplings = 100
+param.Seed = 0
+param.Tolerance = 1e-10
+param.Trials = 1
+param.Verbose = false
+
+output_model := mlpack.GmmTrain(gaussians, input, param)
+```
+
+An implementation of the EM algorithm for training Gaussian mixture models (GMMs).  Given a dataset, this can train a GMM for future use with other tools. [Detailed documentation](#gmm_train_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `DiagonalCovariance` | [`bool`](#doc_bool) | Force the covariance of the Gaussians to be diagonal.  This can accelerate training time significantly. | `false` |
+| `gaussians` | [`int`](#doc_int) | Number of Gaussians in the GMM. | `**--**` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | The training data on which the model will be fit. | `**--**` |
+| `InputModel` | [`gmm`](#doc_model) | Initial input GMM model to start training with. | `nil` |
+| `KmeansMaxIterations` | [`int`](#doc_int) | Maximum number of iterations for the k-means algorithm (used to initialize EM). | `1000` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations of EM algorithm (passing 0 will run until convergence). | `250` |
+| `NoForcePositive` | [`bool`](#doc_bool) | Do not force the covariance matrices to be positive definite. | `false` |
+| `Noise` | [`float64`](#doc_float64) | Variance of zero-mean Gaussian noise to add to data. | `0` |
+| `Percentage` | [`float64`](#doc_float64) | If using --refined_start, specify the percentage of the dataset used for each sampling (should be between 0.0 and 1.0). | `0.02` |
+| `RefinedStart` | [`bool`](#doc_bool) | During the initialization, use refined initial positions for k-means clustering (Bradley and Fayyad, 1998). | `false` |
+| `Samplings` | [`int`](#doc_int) | If using --refined_start, specify the number of samplings used for initial points. | `100` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Tolerance` | [`float64`](#doc_float64) | Tolerance for convergence of EM. | `1e-10` |
+| `Trials` | [`int`](#doc_int) | Number of trials to perform in training GMM. | `1` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`gmm`](#doc_model) | Output for trained GMM model. | 
+
+### Detailed documentation
+{: #gmm_train_detailed-documentation }
+
+This program takes a parametric estimate of a Gaussian mixture model (GMM) using the EM algorithm to find the maximum likelihood estimate.  The model may be saved and reused by other mlpack GMM tools.
+
+The input data to train on must be specified with the `Input` parameter, and the number of Gaussians in the model must be specified with the `Gaussians` parameter.  Optionally, many trials with different random initializations may be run, and the result with highest log-likelihood on the training data will be taken.  The number of trials to run is specified with the `Trials` parameter.  By default, only one trial is run.
+
+The tolerance for convergence and maximum number of iterations of the EM algorithm are specified with the `Tolerance` and `MaxIterations` parameters, respectively.  The GMM may be initialized for training with another model, specified with the `InputModel` parameter. Otherwise, the model is initialized by running k-means on the data.  The k-means clustering initialization can be controlled with the `KmeansMaxIterations`, `RefinedStart`, `Samplings`, and `Percentage` parameters.  If `RefinedStart` is specified, then the Bradley-Fayyad refined start initialization will be used.  This can often lead to better clustering results.
+
+The 'diagonal_covariance' flag will cause the learned covariances to be diagonal matrices.  This significantly simplifies the model itself and causes training to be faster, but restricts the ability to fit more complex GMMs.
+
+If GMM training fails with an error indicating that a covariance matrix could not be inverted, make sure that the `NoForcePositive` parameter is not specified.  Alternately, adding a small amount of Gaussian noise (using the `Noise` parameter) to the entire dataset may help prevent Gaussians with zero variance in a particular dimension, which is usually the cause of non-invertible covariance matrices.
+
+The `NoForcePositive` parameter, if set, will avoid the checks after each iteration of the EM algorithm which ensure that the covariance matrices are positive definite.  Specifying the flag can cause faster runtime, but may also cause non-positive definite covariance matrices, which will cause the program to crash.
+
+### Example
+As an example, to train a 6-Gaussian GMM on the data in `data` with a maximum of 100 iterations of EM and 3 trials, saving the trained GMM to `gmm`, the following command can be used:
+
+```go
+// Initialize optional parameters for GmmTrain().
+param := mlpack.GmmTrainOptions()
+param.Trials = 3
+
+gmm := mlpack.GmmTrain(data, 6, param)
+```
+
+To re-train that GMM on another set of data `data2`, the following command may be used: 
+
+```go
+// Initialize optional parameters for GmmTrain().
+param := mlpack.GmmTrainOptions()
+param.InputModel = &gmm
+
+new_gmm := mlpack.GmmTrain(data2, 6, param)
+```
+
+### See also
+
+ - [GmmGenerate()](#gmm_generate)
+ - [GmmProbability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## GmmGenerate()
+{: #gmm_generate }
+
+#### GMM Sample Generator
+{: #gmm_generate_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for GmmGenerate().
+param := mlpack.GmmGenerateOptions()
+param.Seed = 0
+param.Verbose = false
+
+output := mlpack.GmmGenerate(inputModel, samples, param)
+```
+
+A sample generator for pre-trained GMMs.  Given a pre-trained GMM, this can sample new points randomly from that distribution. [Detailed documentation](#gmm_generate_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `inputModel` | [`gmm`](#doc_model) | Input GMM model to generate samples from. | `**--**` |
+| `samples` | [`int`](#doc_int) | Number of samples to generate. | `**--**` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save output samples in. | 
+
+### Detailed documentation
+{: #gmm_generate_detailed-documentation }
+
+This program is able to generate samples from a pre-trained GMM (use gmm_train to train a GMM).  The pre-trained GMM must be specified with the `InputModel` parameter.  The number of samples to generate is specified by the `Samples` parameter.  Output samples may be saved with the `Output` output parameter.
+
+### Example
+The following command can be used to generate 100 samples from the pre-trained GMM `gmm` and store those generated samples in `samples`:
+
+```go
+// Initialize optional parameters for GmmGenerate().
+param := mlpack.GmmGenerateOptions()
+
+samples := mlpack.GmmGenerate(&gmm, 100, param)
+```
+
+### See also
+
+ - [GmmTrain()](#gmm_train)
+ - [GmmProbability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## GmmProbability()
+{: #gmm_probability }
+
+#### GMM Probability Calculator
+{: #gmm_probability_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for GmmProbability().
+param := mlpack.GmmProbabilityOptions()
+param.Verbose = false
+
+output := mlpack.GmmProbability(input, inputModel, param)
+```
+
+A probability calculator for GMMs.  Given a pre-trained GMM and a set of points, this can compute the probability that each point is from the given GMM. [Detailed documentation](#gmm_probability_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input matrix to calculate probabilities of. | `**--**` |
+| `inputModel` | [`gmm`](#doc_model) | Input GMM to use as model. | `**--**` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to store calculated probabilities in. | 
+
+### Detailed documentation
+{: #gmm_probability_detailed-documentation }
+
+This program calculates the probability that given points came from a given GMM (that is, P(X \| gmm)).  The GMM is specified with the `InputModel` parameter, and the points are specified with the `Input` parameter.  The output probabilities may be saved via the `Output` output parameter.
+
+### Example
+So, for example, to calculate the probabilities of each point in `points` coming from the pre-trained GMM `gmm`, while storing those probabilities in `probs`, the following command could be used:
+
+```go
+// Initialize optional parameters for GmmProbability().
+param := mlpack.GmmProbabilityOptions()
+
+probs := mlpack.GmmProbability(&gmm, points, param)
+```
+
+### See also
+
+ - [GmmTrain()](#gmm_train)
+ - [GmmGenerate()](#gmm_generate)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## HmmTrain()
+{: #hmm_train }
+
+#### Hidden Markov Model (HMM) Training
+{: #hmm_train_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for HmmTrain().
+param := mlpack.HmmTrainOptions()
+param.Batch = false
+param.Gaussians = 0
+param.InputModel = nil
+param.LabelsFile = ""
+param.Seed = 0
+param.States = 0
+param.Tolerance = 1e-05
+param.Type = "gaussian"
+param.Verbose = false
+
+output_model := mlpack.HmmTrain(inputFile, param)
+```
+
+An implementation of training algorithms for Hidden Markov Models (HMMs). Given labeled or unlabeled data, an HMM can be trained for further use with other mlpack HMM tools. [Detailed documentation](#hmm_train_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Batch` | [`bool`](#doc_bool) | If true, input_file (and if passed, labels_file) are expected to contain a list of files to use as input observation sequences (and label sequences). | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Gaussians` | [`int`](#doc_int) | Number of gaussians in each GMM (necessary when type is 'gmm'). | `0` |
+| `inputFile` | [`string`](#doc_string) | File containing input observations. | `**--**` |
+| `InputModel` | [`hmmModel`](#doc_model) | Pre-existing HMM model to initialize training with. | `nil` |
+| `LabelsFile` | [`string`](#doc_string) | Optional file of hidden states, used for labeled training. | `""` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `States` | [`int`](#doc_int) | Number of hidden states in HMM (necessary, unless model_file is specified). | `0` |
+| `Tolerance` | [`float64`](#doc_float64) | Tolerance of the Baum-Welch algorithm. | `1e-05` |
+| `Type` | [`string`](#doc_string) | Type of HMM: discrete \| gaussian \| diag_gmm \| gmm. | `"gaussian"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`hmmModel`](#doc_model) | Output for trained HMM. | 
+
+### Detailed documentation
+{: #hmm_train_detailed-documentation }
+
+This program allows a Hidden Markov Model to be trained on labeled or unlabeled data.  It supports four types of HMMs: Discrete HMMs, Gaussian HMMs, GMM HMMs, or Diagonal GMM HMMs
+
+Either one input sequence can be specified (with `InputFile`), or, a file containing files in which input sequences can be found (when `InputFile`and`Batch` are used together).  In addition, labels can be provided in the file specified by `LabelsFile`, and if `Batch` is used, the file given to `LabelsFile` should contain a list of files of labels corresponding to the sequences in the file given to `InputFile`.
+
+The HMM is trained with the Baum-Welch algorithm if no labels are provided.  The tolerance of the Baum-Welch algorithm can be set with the `Tolerance`option.  By default, the transition matrix is randomly initialized and the emission distributions are initialized to fit the extent of the data.
+
+Optionally, a pre-created HMM model can be used as a guess for the transition matrix and emission probabilities; this is specifiable with `OutputModel`.
+
+### See also
+
+ - [HmmGenerate()](#hmm_generate)
+ - [HmmLoglik()](#hmm_loglik)
+ - [HmmViterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## HmmGenerate()
+{: #hmm_generate }
+
+#### Hidden Markov Model (HMM) Sequence Generator
+{: #hmm_generate_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for HmmGenerate().
+param := mlpack.HmmGenerateOptions()
+param.Seed = 0
+param.StartState = 0
+param.Verbose = false
+
+output, state := mlpack.HmmGenerate(length, model, param)
+```
+
+A utility to generate random sequences from a pre-trained Hidden Markov Model (HMM).  The length of the desired sequence can be specified, and a random sequence of observations is returned. [Detailed documentation](#hmm_generate_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `length` | [`int`](#doc_int) | Length of sequence to generate. | `**--**` |
+| `model` | [`hmmModel`](#doc_model) | Trained HMM to generate sequences with. | `**--**` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `StartState` | [`int`](#doc_int) | Starting state of sequence. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save observation sequence to. | 
+| `State` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save hidden state sequence to. | 
+
+### Detailed documentation
+{: #hmm_generate_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as the `Model` parameter, and generates a random observation sequence and hidden state sequence based on its parameters. The observation sequence may be saved with the `Output` output parameter, and the internal state  sequence may be saved with the `State` output parameter.
+
+The state to start the sequence in may be specified with the `StartState` parameter.
+
+### Example
+For example, to generate a sequence of length 150 from the HMM `hmm` and save the observation sequence to `observations` and the hidden state sequence to `states`, the following command may be used: 
+
+```go
+// Initialize optional parameters for HmmGenerate().
+param := mlpack.HmmGenerateOptions()
+
+observations, states := mlpack.HmmGenerate(&hmm, 150, param)
+```
+
+### See also
+
+ - [HmmTrain()](#hmm_train)
+ - [HmmLoglik()](#hmm_loglik)
+ - [HmmViterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## HmmLoglik()
+{: #hmm_loglik }
+
+#### Hidden Markov Model (HMM) Sequence Log-Likelihood
+{: #hmm_loglik_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for HmmLoglik().
+param := mlpack.HmmLoglikOptions()
+param.Verbose = false
+
+log_likelihood := mlpack.HmmLoglik(input, inputModel, param)
+```
+
+A utility for computing the log-likelihood of a sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observation sequence, this computes and returns the log-likelihood of that sequence being observed from that HMM. [Detailed documentation](#hmm_loglik_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | File containing observations, | `**--**` |
+| `inputModel` | [`hmmModel`](#doc_model) | File containing HMM. | `**--**` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `LogLikelihood` | [`float64`](#doc_float64) | Log-likelihood of the sequence. | 
+
+### Detailed documentation
+{: #hmm_loglik_detailed-documentation }
+
+This utility takes an already-trained HMM, specified with the `InputModel` parameter, and evaluates the log-likelihood of a sequence of observations, given with the `Input` parameter.  The computed log-likelihood is given as output.
+
+### Example
+For example, to compute the log-likelihood of the sequence `seq` with the pre-trained HMM `hmm`, the following command may be used: 
+
+```go
+// Initialize optional parameters for HmmLoglik().
+param := mlpack.HmmLoglikOptions()
+
+_ := mlpack.HmmLoglik(seq, &hmm, param)
+```
+
+### See also
+
+ - [HmmTrain()](#hmm_train)
+ - [HmmGenerate()](#hmm_generate)
+ - [HmmViterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## HmmViterbi()
+{: #hmm_viterbi }
+
+#### Hidden Markov Model (HMM) Viterbi State Prediction
+{: #hmm_viterbi_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for HmmViterbi().
+param := mlpack.HmmViterbiOptions()
+param.Verbose = false
+
+output := mlpack.HmmViterbi(input, inputModel, param)
+```
+
+A utility for computing the most probable hidden state sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observed sequence, this uses the Viterbi algorithm to compute and return the most probable hidden state sequence. [Detailed documentation](#hmm_viterbi_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing observations, | `**--**` |
+| `inputModel` | [`hmmModel`](#doc_model) | Trained HMM to use. | `**--**` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | File to save predicted state sequence to. | 
+
+### Detailed documentation
+{: #hmm_viterbi_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as `InputModel`, and evaluates the most probable hidden state sequence of a given sequence of observations (specified as '`Input`, using the Viterbi algorithm.  The computed state sequence may be saved using the `Output` output parameter.
+
+### Example
+For example, to predict the state sequence of the observations `obs` using the HMM `hmm`, storing the predicted state sequence to `states`, the following command could be used:
+
+```go
+// Initialize optional parameters for HmmViterbi().
+param := mlpack.HmmViterbiOptions()
+
+states := mlpack.HmmViterbi(obs, &hmm, param)
+```
+
+### See also
+
+ - [HmmTrain()](#hmm_train)
+ - [HmmGenerate()](#hmm_generate)
+ - [HmmLoglik()](#hmm_loglik)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## HoeffdingTree()
+{: #hoeffding_tree }
+
+#### Hoeffding trees
+{: #hoeffding_tree_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for HoeffdingTree().
+param := mlpack.HoeffdingTreeOptions()
+param.BatchMode = false
+param.Bins = 10
+param.Confidence = 0.95
+param.InfoGain = false
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.MaxSamples = 5000
+param.MinSamples = 100
+param.NumericSplitStrategy = "binary"
+param.ObservationsBeforeBinning = 100
+param.Passes = 1
+param.Test = mat.NewDense(1, 1, nil)
+param.TestLabels = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, predictions, probabilities := mlpack.HoeffdingTree(param)
+```
+
+An implementation of Hoeffding trees, a form of streaming decision tree for classification.  Given labeled data, a Hoeffding tree can be trained and saved for later use, or a pre-trained Hoeffding tree can be used for predicting the classifications of new points. [Detailed documentation](#hoeffding_tree_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `BatchMode` | [`bool`](#doc_bool) | If true, samples will be considered in batch instead of as a stream.  This generally results in better trees but at the cost of memory usage and runtime. | `false` |
+| `Bins` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of bins for each numeric split. | `10` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Confidence` | [`float64`](#doc_float64) | Confidence before splitting (between 0 and 1). | `0.95` |
+| `InfoGain` | [`bool`](#doc_bool) | If set, information gain is used instead of Gini impurity for calculating Hoeffding bounds. | `false` |
+| `InputModel` | [`hoeffdingTreeModel`](#doc_model) | Input trained Hoeffding tree model. | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels for training dataset. | `mat.NewDense(1, 1, nil)` |
+| `MaxSamples` | [`int`](#doc_int) | Maximum number of samples before splitting. | `5000` |
+| `MinSamples` | [`int`](#doc_int) | Minimum number of samples before splitting. | `100` |
+| `NumericSplitStrategy` | [`string`](#doc_string) | The splitting strategy to use for numeric features: 'domingos' or 'binary'. | `"binary"` |
+| `ObservationsBeforeBinning` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of samples observed before binning is performed. | `100` |
+| `Passes` | [`int`](#doc_int) | Number of passes to take over the dataset. | `1` |
+| `Test` | [`matrixWithInfo`](#doc_matrixWithInfo) | Testing dataset (may be categorical). | `mat.NewDense(1, 1, nil)` |
+| `TestLabels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels of test data. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`matrixWithInfo`](#doc_matrixWithInfo) | Training dataset (may be categorical). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`hoeffdingTreeModel`](#doc_model) | Output for trained Hoeffding tree model. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Matrix to output label predictions for test data into. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | In addition to predicting labels, provide rediction probabilities in this matrix. | 
+
+### Detailed documentation
+{: #hoeffding_tree_detailed-documentation }
+
+This program implements Hoeffding trees, a form of streaming decision tree suited best for large (or streaming) datasets.  This program supports both categorical and numeric data.  Given an input dataset, this program is able to train the tree with numerous training options, and save the model to a file.  The program is also able to use a trained model or a model from file in order to predict classes for a given test set.
+
+The training file and associated labels are specified with the `Training` and `Labels` parameters, respectively. Optionally, if `Labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+The training may be performed in batch mode (like a typical decision tree algorithm) by specifying the `BatchMode` option, but this may not be the best option for large datasets.
+
+When a model is trained, it may be saved via the `OutputModel` output parameter.  A model may be loaded from file for further training or testing with the `InputModel` parameter.
+
+Test data may be specified with the `Test` parameter, and if performance statistics are desired for that test set, labels may be specified with the `TestLabels` parameter.  Predictions for each test point may be saved with the `Predictions` output parameter, and class probabilities for each prediction may be saved with the `Probabilities` output parameter.
+
+### Example
+For example, to train a Hoeffding tree with confidence 0.99 with data `dataset`, saving the trained tree to `tree`, the following command may be used:
+
+```go
+// Initialize optional parameters for HoeffdingTree().
+param := mlpack.HoeffdingTreeOptions()
+param.Training = dataset
+param.Confidence = 0.99
+
+tree, _, _ := mlpack.HoeffdingTree(param)
+```
+
+Then, this tree may be used to make predictions on the test set `test_set`, saving the predictions into `predictions` and the class probabilities into `class_probs` with the following command: 
+
+```go
+// Initialize optional parameters for HoeffdingTree().
+param := mlpack.HoeffdingTreeOptions()
+param.InputModel = &tree
+param.Test = test_set
+
+_, predictions, class_probs := mlpack.HoeffdingTree(param)
+```
+
+### See also
+
+ - [DecisionTree()](#decision_tree)
+ - [RandomForest()](#random_forest)
+ - [Mining High-Speed Data Streams (pdf)](http://dm.cs.washington.edu/papers/vfdt-kdd00.pdf)
+ - [HoeffdingTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp)
+
+## Kde()
+{: #kde }
+
+#### Kernel Density Estimation
+{: #kde_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Kde().
+param := mlpack.KdeOptions()
+param.AbsError = 0
+param.Algorithm = "dual-tree"
+param.Bandwidth = 1
+param.InitialSampleSize = 100
+param.InputModel = nil
+param.Kernel = "gaussian"
+param.McBreakCoef = 0.4
+param.McEntryCoef = 3
+param.McProbability = 0.95
+param.MonteCarlo = false
+param.Query = mat.NewDense(1, 1, nil)
+param.Reference = mat.NewDense(1, 1, nil)
+param.RelError = 0.05
+param.Tree = "kd-tree"
+param.Verbose = false
+
+output_model, predictions := mlpack.Kde(param)
+```
+
+An implementation of kernel density estimation with dual-tree algorithms. Given a set of reference points and query points and a kernel function, this can estimate the density function at the location of each query point using trees; trees that are built can be saved for later use. [Detailed documentation](#kde_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `AbsError` | [`float64`](#doc_float64) | Relative error tolerance for the prediction. | `0` |
+| `Algorithm` | [`string`](#doc_string) | Algorithm to use for the prediction.('dual-tree', 'single-tree'). | `"dual-tree"` |
+| `Bandwidth` | [`float64`](#doc_float64) | Bandwidth of the kernel. | `1` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InitialSampleSize` | [`int`](#doc_int) | Initial sample size for Monte Carlo estimations. | `100` |
+| `InputModel` | [`kdeModel`](#doc_model) | Contains pre-trained KDE model. | `nil` |
+| `Kernel` | [`string`](#doc_string) | Kernel to use for the prediction.('gaussian', 'epanechnikov', 'laplacian', 'spherical', 'triangular'). | `"gaussian"` |
+| `McBreakCoef` | [`float64`](#doc_float64) | Controls what fraction of the amount of node's descendants is the limit for the sample size before it recurses. | `0.4` |
+| `McEntryCoef` | [`float64`](#doc_float64) | Controls how much larger does the amount of node descendants has to be compared to the initial sample size in order to be a candidate for Monte Carlo estimations. | `3` |
+| `McProbability` | [`float64`](#doc_float64) | Probability of the estimation being bounded by relative error when using Monte Carlo estimations. | `0.95` |
+| `MonteCarlo` | [`bool`](#doc_bool) | Whether to use Monte Carlo estimations when possible. | `false` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Query dataset to KDE on. | `mat.NewDense(1, 1, nil)` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Input reference dataset use for KDE. | `mat.NewDense(1, 1, nil)` |
+| `RelError` | [`float64`](#doc_float64) | Relative error tolerance for the prediction. | `0.05` |
+| `Tree` | [`string`](#doc_string) | Tree to use for the prediction.('kd-tree', 'ball-tree', 'cover-tree', 'octree', 'r-tree'). | `"kd-tree"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`kdeModel`](#doc_model) | If specified, the KDE model will be saved here. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Vector to store density predictions. | 
+
+### Detailed documentation
+{: #kde_detailed-documentation }
+
+This program performs a Kernel Density Estimation. KDE is a non-parametric way of estimating probability density function. For each query point the program will estimate its probability density by applying a kernel function to each reference point. The computational complexity of this is O(N^2) where there are N query points and N reference points, but this implementation will typically see better performance as it uses an approximate dual or single tree algorithm for acceleration.
+
+Dual or single tree optimization avoids many barely relevant calculations (as kernel function values decrease with distance), so it is an approximate computation. You can specify the maximum relative error tolerance for each query value with `RelError` as well as the maximum absolute error tolerance with the parameter `AbsError`. This program runs using an Euclidean metric. Kernel function can be selected using the `Kernel` option. You can also choose what which type of tree to use for the dual-tree algorithm with `Tree`. It is also possible to select whether to use dual-tree algorithm or single-tree algorithm using the `Algorithm` option.
+
+Monte Carlo estimations can be used to accelerate the KDE estimate when the Gaussian Kernel is used. This provides a probabilistic guarantee on the the error of the resulting KDE instead of an absolute guarantee.To enable Monte Carlo estimations, the `MonteCarlo` flag can be used, and success probability can be set with the `McProbability` option. It is possible to set the initial sample size for the Monte Carlo estimation using `InitialSampleSize`. This implementation will only consider a node, as a candidate for the Monte Carlo estimation, if its number of descendant nodes is bigger than the initial sample size. This can be controlled using a coefficient that will multiply the initial sample size and can be set using `McEntryCoef`. To avoid using the same amount of computations an exact approach would take, this program recurses the tree whenever a fraction of the amount of the node's descendant points have already been computed. This fraction is set using `McBreakCoef`.
+
+### Example
+For example, the following will run KDE using the data in `ref_data` for training and the data in `qu_data` as query data. It will apply an Epanechnikov kernel with a 0.2 bandwidth to each reference point and use a KD-Tree for the dual-tree optimization. The returned predictions will be within 5% of the real KDE value for each query point.
+
+```go
+// Initialize optional parameters for Kde().
+param := mlpack.KdeOptions()
+param.Reference = ref_data
+param.Query = qu_data
+param.Bandwidth = 0.2
+param.Kernel = "epanechnikov"
+param.Tree = "kd-tree"
+param.RelError = 0.05
+
+_, out_data := mlpack.Kde(param)
+```
+
+the predicted density estimations will be stored in `out_data`.
+If no `Query` is provided, then KDE will be computed on the `Reference` dataset.
+It is possible to select either a reference dataset or an input model but not both at the same time. If an input model is selected and parameter values are not set (e.g. `Bandwidth`) then default parameter values will be used.
+
+In addition to the last program call, it is also possible to activate Monte Carlo estimations if a Gaussian kernel is used. This can provide faster results, but the KDE will only have a probabilistic guarantee of meeting the desired error bound (instead of an absolute guarantee). The following example will run KDE using a Monte Carlo estimation when possible. The results will be within a 5% of the real KDE value with a 95% probability. Initial sample size for the Monte Carlo estimation will be 200 points and a node will be a candidate for the estimation only when it contains 700 (i.e. 3.5*200) points. If a node contains 700 points and 420 (i.e. 0.6*700) have already been sampled, then the algorithm will recurse instead of keep sampling.
+
+```go
+// Initialize optional parameters for Kde().
+param := mlpack.KdeOptions()
+param.Reference = ref_data
+param.Query = qu_data
+param.Bandwidth = 0.2
+param.Kernel = "gaussian"
+param.Tree = "kd-tree"
+param.RelError = 0.05
+param.MonteCarlo = 
+param.McProbability = 0.95
+param.InitialSampleSize = 200
+param.McEntryCoef = 3.5
+param.McBreakCoef = 0.6
+
+_, out_data := mlpack.Kde(param)
+```
+
+### See also
+
+ - [Knn()](#knn)
+ - [Kernel density estimation on Wikipedia](https://en.wikipedia.org/wiki/Kernel_density_estimation)
+ - [Tree-Independent Dual-Tree Algorithms](https://arxiv.org/pdf/1304.4327)
+ - [Fast High-dimensional Kernel Summations Using the Monte Carlo Multipole Method](http://papers.nips.cc/paper/3539-fast-high-dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf)
+ - [KDE C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kde/kde.hpp)
+
+## KernelPca()
+{: #kernel_pca }
+
+#### Kernel Principal Components Analysis
+{: #kernel_pca_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for KernelPca().
+param := mlpack.KernelPcaOptions()
+param.Bandwidth = 1
+param.Center = false
+param.Degree = 1
+param.KernelScale = 1
+param.NewDimensionality = 0
+param.NystroemMethod = false
+param.Offset = 0
+param.Sampling = "kmeans"
+param.Verbose = false
+
+output := mlpack.KernelPca(input, kernel, param)
+```
+
+An implementation of Kernel Principal Components Analysis (KPCA).  This can be used to perform nonlinear dimensionality reduction or preprocessing on a given dataset. [Detailed documentation](#kernel_pca_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Bandwidth` | [`float64`](#doc_float64) | Bandwidth, for 'gaussian' and 'laplacian' kernels. | `1` |
+| `Center` | [`bool`](#doc_bool) | If set, the transformed data will be centered about the origin. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Degree` | [`float64`](#doc_float64) | Degree of polynomial, for 'polynomial' kernel. | `1` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform KPCA on. | `**--**` |
+| `kernel` | [`string`](#doc_string) | The kernel to use; see the above documentation for the list of usable kernels. | `**--**` |
+| `KernelScale` | [`float64`](#doc_float64) | Scale, for 'hyptan' kernel. | `1` |
+| `NewDimensionality` | [`int`](#doc_int) | If not 0, reduce the dimensionality of the output dataset by ignoring the dimensions with the smallest eigenvalues. | `0` |
+| `NystroemMethod` | [`bool`](#doc_bool) | If set, the Nystroem method will be used. | `false` |
+| `Offset` | [`float64`](#doc_float64) | Offset, for 'hyptan' and 'polynomial' kernels. | `0` |
+| `Sampling` | [`string`](#doc_string) | Sampling scheme to use for the Nystroem method: 'kmeans', 'random', 'ordered' | `"kmeans"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #kernel_pca_detailed-documentation }
+
+This program performs Kernel Principal Components Analysis (KPCA) on the specified dataset with the specified kernel.  This will transform the data onto the kernel principal components, and optionally reduce the dimensionality by ignoring the kernel principal components with the smallest eigenvalues.
+
+For the case where a linear kernel is used, this reduces to regular PCA.
+
+The kernels that are supported are listed below:
+
+ * 'linear': the standard linear dot product (same as normal PCA):
+    `K(x, y) = x^T y`
+
+ * 'gaussian': a Gaussian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\| ^ 2) / (2 * (bandwidth ^ 2)))`
+
+ * 'polynomial': polynomial kernel; requires offset and degree:
+    `K(x, y) = (x^T y + offset) ^ degree`
+
+ * 'hyptan': hyperbolic tangent kernel; requires scale and offset:
+    `K(x, y) = tanh(scale * (x^T y) + offset)`
+
+ * 'laplacian': Laplacian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\|) / bandwidth)`
+
+ * 'epanechnikov': Epanechnikov kernel; requires bandwidth:
+    `K(x, y) = max(0, 1 - \|\| x - y \|\|^2 / bandwidth^2)`
+
+ * 'cosine': cosine distance:
+    `K(x, y) = 1 - (x^T y) / (\|\| x \|\| * \|\| y \|\|)`
+
+The parameters for each of the kernels should be specified with the options `Bandwidth`, `KernelScale`, `Offset`, or `Degree` (or a combination of those parameters).
+
+Optionally, the Nystroem method ("Using the Nystroem method to speed up kernel machines", 2001) can be used to calculate the kernel matrix by specifying the `NystroemMethod` parameter. This approach works by using a subset of the data as basis to reconstruct the kernel matrix; to specify the sampling scheme, the `Sampling` parameter is used.  The sampling scheme for the Nystroem method can be chosen from the following list: 'kmeans', 'random', 'ordered'.
+
+### Example
+For example, the following command will perform KPCA on the dataset `input` using the Gaussian kernel, and saving the transformed data to `transformed`: 
+
+```go
+// Initialize optional parameters for KernelPca().
+param := mlpack.KernelPcaOptions()
+
+transformed := mlpack.KernelPca(input, "gaussian", param)
+```
+
+### See also
+
+ - [Kernel principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Kernel_principal_component_analysis)
+ - [Kernel Principal Component Analysis (pdf)](https://pca.narod.ru/scholkopf_kernel.pdf)
+ - [KernelPCA class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+
+## Kmeans()
+{: #kmeans }
+
+#### K-Means Clustering
+{: #kmeans_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Kmeans().
+param := mlpack.KmeansOptions()
+param.Algorithm = "naive"
+param.AllowEmptyClusters = false
+param.InPlace = false
+param.InitialCentroids = mat.NewDense(1, 1, nil)
+param.KillEmptyClusters = false
+param.KmeansPlusPlus = false
+param.LabelsOnly = false
+param.MaxIterations = 1000
+param.Percentage = 0.02
+param.RefinedStart = false
+param.Samplings = 100
+param.Seed = 0
+param.Verbose = false
+
+centroid, output := mlpack.Kmeans(clusters, input, param)
+```
+
+An implementation of several strategies for efficient k-means clustering. Given a dataset and a value of k, this computes and returns a k-means clustering on that data. [Detailed documentation](#kmeans_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Algorithm` | [`string`](#doc_string) | Algorithm to use for the Lloyd iteration ('naive', 'pelleg-moore', 'elkan', 'hamerly', 'dualtree', or 'dualtree-covertree'). | `"naive"` |
+| `AllowEmptyClusters` | [`bool`](#doc_bool) | Allow empty clusters to be persist. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `clusters` | [`int`](#doc_int) | Number of clusters to find (0 autodetects from initial centroids). | `**--**` |
+| `InPlace` | [`bool`](#doc_bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden. (Do not use in Python.) | `false` |
+| `InitialCentroids` | [`*mat.Dense`](#doc_a__mat_Dense) | Start with the specified initial centroids. | `mat.NewDense(1, 1, nil)` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform clustering on. | `**--**` |
+| `KillEmptyClusters` | [`bool`](#doc_bool) | Remove empty clusters when they occur. | `false` |
+| `KmeansPlusPlus` | [`bool`](#doc_bool) | Use the k-means++ initialization strategy to choose initial points. | `false` |
+| `LabelsOnly` | [`bool`](#doc_bool) | Only output labels into output file. | `false` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations before k-means terminates. | `1000` |
+| `Percentage` | [`float64`](#doc_float64) | Percentage of dataset to use for each refined start sampling (use when --refined_start is specified). | `0.02` |
+| `RefinedStart` | [`bool`](#doc_bool) | Use the refined initial point strategy by Bradley and Fayyad to choose initial points. | `false` |
+| `Samplings` | [`int`](#doc_int) | Number of samplings to perform for refined start (use when --refined_start is specified). | `100` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Centroid` | [`*mat.Dense`](#doc_a__mat_Dense) | If specified, the centroids of each cluster will  be written to the given file. | 
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to store output labels or labeled data to. | 
+
+### Detailed documentation
+{: #kmeans_detailed-documentation }
+
+This program performs K-Means clustering on the given dataset.  It can return the learned cluster assignments, and the centroids of the clusters.  Empty clusters are not allowed by default; when a cluster becomes empty, the point furthest from the centroid of the cluster with maximum variance is taken to fill that cluster.
+
+Optionally, the strategy to choose initial centroids can be specified.  The k-means++ algorithm can be used to choose initial centroids with the `KmeansPlusPlus` parameter.  The Bradley and Fayyad approach ("Refining initial points for k-means clustering", 1998) can be used to select initial points by specifying the `RefinedStart` parameter.  This approach works by taking random samplings of the dataset; to specify the number of samplings, the `Samplings` parameter is used, and to specify the percentage of the dataset to be used in each sample, the `Percentage` parameter is used (it should be a value between 0.0 and 1.0).
+
+There are several options available for the algorithm used for each Lloyd iteration, specified with the `Algorithm`  option.  The standard O(kN) approach can be used ('naive').  Other options include the Pelleg-Moore tree-based algorithm ('pelleg-moore'), Elkan's triangle-inequality based algorithm ('elkan'), Hamerly's modification to Elkan's algorithm ('hamerly'), the dual-tree k-means algorithm ('dualtree'), and the dual-tree k-means algorithm using the cover tree ('dualtree-covertree').
+
+The behavior for when an empty cluster is encountered can be modified with the `AllowEmptyClusters` option.  When this option is specified and there is a cluster owning no points at the end of an iteration, that cluster's centroid will simply remain in its position from the previous iteration. If the `KillEmptyClusters` option is specified, then when a cluster owns no points at the end of an iteration, the cluster centroid is simply filled with DBL_MAX, killing it and effectively reducing k for the rest of the computation.  Note that the default option when neither empty cluster option is specified can be time-consuming to calculate; therefore, specifying either of these parameters will often accelerate runtime.
+
+Initial clustering assignments may be specified using the `InitialCentroids` parameter, and the maximum number of iterations may be specified with the `MaxIterations` parameter.
+
+### Example
+As an example, to use Hamerly's algorithm to perform k-means clustering with k=10 on the dataset `data`, saving the centroids to `centroids` and the assignments for each point to `assignments`, the following command could be used:
+
+```go
+// Initialize optional parameters for Kmeans().
+param := mlpack.KmeansOptions()
+
+centroids, assignments := mlpack.Kmeans(data, 10, param)
+```
+
+To run k-means on that same dataset with initial centroids specified in `initial` with a maximum of 500 iterations, storing the output centroids in `final` the following command may be used:
+
+```go
+// Initialize optional parameters for Kmeans().
+param := mlpack.KmeansOptions()
+param.InitialCentroids = initial
+param.MaxIterations = 500
+
+final, _ := mlpack.Kmeans(data, 10, param)
+```
+
+### See also
+
+ - [Dbscan()](#dbscan)
+ - [k-means++](https://en.wikipedia.org/wiki/K-means%2B%2B)
+ - [Using the triangle inequality to accelerate k-means (pdf)](https://cdn.aaai.org/ICML/2003/ICML03-022.pdf)
+ - [Making k-means even faster (pdf)](https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12)
+ - [Accelerating exact k-means algorithms with geometric reasoning (pdf)](http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/ftp/usr0/ftp/2000/CMU-CS-00-105.pdf)
+ - [A dual-tree algorithm for fast k-means clustering with large k (pdf)](http://www.ratml.org/pub/pdf/2017dual.pdf)
+ - [KMeans class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kmeans/kmeans.hpp)
+
+## Lars()
+{: #lars }
+
+#### LARS
+{: #lars_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Lars().
+param := mlpack.LarsOptions()
+param.Input = mat.NewDense(1, 1, nil)
+param.InputModel = nil
+param.Lambda1 = 0
+param.Lambda2 = 0
+param.NoIntercept = false
+param.NoNormalize = false
+param.Responses = mat.NewDense(1, 1, nil)
+param.Test = mat.NewDense(1, 1, nil)
+param.UseCholesky = false
+param.Verbose = false
+
+output_model, output_predictions := mlpack.Lars(param)
+```
+
+An implementation of Least Angle Regression (Stagewise/laSso), also known as LARS.  This can train a LARS/LASSO/Elastic Net model and use that model or a pre-trained model to output regression predictions for a test set. [Detailed documentation](#lars_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of covariates (X). | `mat.NewDense(1, 1, nil)` |
+| `InputModel` | [`lars`](#doc_model) | Trained LARS model to use. | `nil` |
+| `Lambda1` | [`float64`](#doc_float64) | Regularization parameter for l1-norm penalty. | `0` |
+| `Lambda2` | [`float64`](#doc_float64) | Regularization parameter for l2-norm penalty. | `0` |
+| `NoIntercept` | [`bool`](#doc_bool) | Do not fit an intercept in the model. | `false` |
+| `NoNormalize` | [`bool`](#doc_bool) | Do not normalize data to unit variance before modeling. | `false` |
+| `Responses` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of responses/observations (y). | `mat.NewDense(1, 1, nil)` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing points to regress on (test points). | `mat.NewDense(1, 1, nil)` |
+| `UseCholesky` | [`bool`](#doc_bool) | Use Cholesky decomposition during computation rather than explicitly computing the full Gram matrix. | `false` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`lars`](#doc_model) | Output LARS model. | 
+| `OutputPredictions` | [`*mat.Dense`](#doc_a__mat_Dense) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #lars_detailed-documentation }
+
+An implementation of LARS: Least Angle Regression (Stagewise/laSso).  This is a stage-wise homotopy-based algorithm for L1-regularized linear regression (LASSO) and L1+L2-regularized linear regression (Elastic Net).
+
+This program is able to train a LARS/LASSO/Elastic Net model or load a model from file, output regression predictions for a test set, and save the trained model to a file.  The LARS algorithm is described in more detail below:
+
+Let X be a matrix where each row is a point and each column is a dimension, and let y be a vector of targets.
+
+The Elastic Net problem is to solve
+
+  min_beta 0.5 \|\| X * beta - y \|\|_2^2 + lambda_1 \|\|beta\|\|_1 +
+      0.5 lambda_2 \|\|beta\|\|_2^2
+
+If lambda1 > 0 and lambda2 = 0, the problem is the LASSO.
+If lambda1 > 0 and lambda2 > 0, the problem is the Elastic Net.
+If lambda1 = 0 and lambda2 > 0, the problem is ridge regression.
+If lambda1 = 0 and lambda2 = 0, the problem is unregularized linear regression.
+
+For efficiency reasons, it is not recommended to use this algorithm with `Lambda1` = 0.  In that case, use the 'linear_regression' program, which implements both unregularized linear regression and ridge regression.
+
+To train a LARS/LASSO/Elastic Net model, the `Input` and `Responses` parameters must be given.  The `Lambda1`, `Lambda2`, and `UseCholesky` parameters control the training options.  A trained model can be saved with the `OutputModel`.  If no training is desired at all, a model can be passed via the `InputModel` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `Test` parameter.  Predicted responses to the test points can be saved with the `OutputPredictions` output parameter.
+
+### Example
+For example, the following command trains a model on the data `data` and responses `responses` with lambda1 set to 0.4 and lambda2 set to 0 (so, LASSO is being solved), and then the model is saved to `lasso_model`:
+
+```go
+// Initialize optional parameters for Lars().
+param := mlpack.LarsOptions()
+param.Input = data
+param.Responses = responses
+param.Lambda1 = 0.4
+param.Lambda2 = 0
+
+lasso_model, _ := mlpack.Lars(param)
+```
+
+The following command uses the `lasso_model` to provide predicted responses for the data `test` and save those responses to `test_predictions`: 
+
+```go
+// Initialize optional parameters for Lars().
+param := mlpack.LarsOptions()
+param.InputModel = &lasso_model
+param.Test = test
+
+_, test_predictions := mlpack.Lars(param)
+```
+
+### See also
+
+ - [LinearRegression()](#linear_regression)
+ - [Least angle regression (pdf)](https://mlpack.org/papers/lars.pdf)
+ - [LARS C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lars/lars.hpp)
+
+## LinearSvm()
+{: #linear_svm }
+
+#### Linear SVM is an L2-regularized support vector machine.
+{: #linear_svm_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for LinearSvm().
+param := mlpack.LinearSvmOptions()
+param.Delta = 1
+param.Epochs = 50
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.Lambda = 0.0001
+param.MaxIterations = 10000
+param.NoIntercept = false
+param.NumClasses = 0
+param.Optimizer = "lbfgs"
+param.Seed = 0
+param.Shuffle = false
+param.StepSize = 0.01
+param.Test = mat.NewDense(1, 1, nil)
+param.TestLabels = mat.NewDense(1, 1, nil)
+param.Tolerance = 1e-10
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, predictions, probabilities := mlpack.LinearSvm(param)
+```
+
+An implementation of linear SVM for multiclass classification. Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#linear_svm_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Delta` | [`float64`](#doc_float64) | Margin of difference between correct class and other classes. | `1` |
+| `Epochs` | [`int`](#doc_int) | Maximum number of full epochs over dataset for psgd | `50` |
+| `InputModel` | [`linearsvmModel`](#doc_model) | Existing model (parameters). | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | A matrix containing labels (0 or 1) for the points in the training set (y). | `mat.NewDense(1, 1, nil)` |
+| `Lambda` | [`float64`](#doc_float64) | L2-regularization parameter for training. | `0.0001` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `NoIntercept` | [`bool`](#doc_bool) | Do not add the intercept term to the model. | `false` |
+| `NumClasses` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `Optimizer` | [`string`](#doc_string) | Optimizer to use for training ('lbfgs' or 'psgd'). | `"lbfgs"` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Shuffle` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for parallel SGD. | `false` |
+| `StepSize` | [`float64`](#doc_float64) | Step size for parallel SGD optimizer. | `0.01` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing test dataset. | `mat.NewDense(1, 1, nil)` |
+| `TestLabels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Matrix containing test labels. | `mat.NewDense(1, 1, nil)` |
+| `Tolerance` | [`float64`](#doc_float64) | Convergence tolerance for optimizer. | `1e-10` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the training set (the matrix of predictors, X). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`linearsvmModel`](#doc_model) | Output for trained linear svm model. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #linear_svm_detailed-documentation }
+
+An implementation of linear SVMs that uses either L-BFGS or parallel SGD (stochastic gradient descent) to train the model.
+
+This program allows loading a linear SVM model (via the `InputModel` parameter) or training a linear SVM model given training data (specified with the `Training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `Test` parameter) and the classification results may be saved with the `Predictions` output parameter. The trained linear SVM model may be saved using the `OutputModel` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `Labels` parameter may be used to specify a separate vector of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `Lambda` option, and the number of classes can be manually specified with the `NumClasses`and if an intercept term is not desired in the model, the `NoIntercept` parameter can be specified.Margin of difference between correct class and other classes can be specified with the `Delta` option.The optimizer used to train the model can be specified with the `Optimizer` parameter.  Available options are 'psgd' (parallel stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `MaxIterations` parameter specifies the maximum number of allowed iterations, and the `Tolerance` parameter specifies the tolerance for convergence.  For the parallel SGD optimizer, the `StepSize` parameter controls the step size taken at each iteration by the optimizer and the maximum number of epochs (specified with `Epochs`). If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+Optionally, the model can be used to predict the labels for another matrix of data points, if `Test` is specified.  The `Test` parameter can be specified without the `Training` parameter, so long as an existing linear SVM model is given with the `InputModel` parameter.  The output predictions from the linear SVM model may be saved with the `Predictions` parameter.
+
+### Example
+As an example, to train a LinaerSVM on the data '`data`' with labels '`labels`' with L2 regularization of 0.1, saving the model to '`lsvm_model`', the following command may be used:
+
+```go
+// Initialize optional parameters for LinearSvm().
+param := mlpack.LinearSvmOptions()
+param.Training = data
+param.Labels = labels
+param.Lambda = 0.1
+param.Delta = 1
+param.NumClasses = 0
+
+lsvm_model, _, _ := mlpack.LinearSvm(param)
+```
+
+Then, to use that model to predict classes for the dataset '`test`', storing the output predictions in '`predictions`', the following command may be used: 
+
+```go
+// Initialize optional parameters for LinearSvm().
+param := mlpack.LinearSvmOptions()
+param.InputModel = &lsvm_model
+param.Test = test
+
+_, predictions, _ := mlpack.LinearSvm(param)
+```
+
+### See also
+
+ - [RandomForest()](#random_forest)
+ - [LogisticRegression()](#logistic_regression)
+ - [LinearSVM on Wikipedia](https://en.wikipedia.org/wiki/Support-vector_machine)
+ - [LinearSVM C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_svm/linear_svm.hpp)
+
+## Lmnn()
+{: #lmnn }
+
+#### Large Margin Nearest Neighbors (LMNN)
+{: #lmnn_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Lmnn().
+param := mlpack.LmnnOptions()
+param.BatchSize = 50
+param.Center = false
+param.Distance = mat.NewDense(1, 1, nil)
+param.K = 1
+param.Labels = mat.NewDense(1, 1, nil)
+param.LinearScan = false
+param.MaxIterations = 100000
+param.Normalize = false
+param.Optimizer = "amsgrad"
+param.Passes = 50
+param.PrintAccuracy = false
+param.Range = 1
+param.Rank = 0
+param.Regularization = 0.5
+param.Seed = 0
+param.StepSize = 0.01
+param.Tolerance = 1e-07
+param.Verbose = false
+
+centered_data, output, transformed_data := mlpack.Lmnn(input, param)
+```
+
+An implementation of Large Margin Nearest Neighbors (LMNN), a distance learning technique.  Given a labeled dataset, this learns a transformation of the data that improves k-nearest-neighbor performance; this can be useful as a preprocessing step. [Detailed documentation](#lmnn_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `BatchSize` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `Center` | [`bool`](#doc_bool) | Perform mean-centering on the dataset. It is useful when the centroid of the data is far from the origin. | `false` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Distance` | [`*mat.Dense`](#doc_a__mat_Dense) | Initial distance matrix to be used as starting point | `mat.NewDense(1, 1, nil)` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to run LMNN on. | `**--**` |
+| `K` | [`int`](#doc_int) | Number of target neighbors to use for each datapoint. | `1` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels for input dataset. | `mat.NewDense(1, 1, nil)` |
+| `LinearScan` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `false` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations for L-BFGS (0 indicates no limit). | `100000` |
+| `Normalize` | [`bool`](#doc_bool) | Use a normalized starting point for optimization. Itis useful for when points are far apart, or when SGD is returning NaN. | `false` |
+| `Optimizer` | [`string`](#doc_string) | Optimizer to use; 'amsgrad', 'bbsgd', 'sgd', or 'lbfgs'. | `"amsgrad"` |
+| `Passes` | [`int`](#doc_int) | Maximum number of full passes over dataset for AMSGrad, BB_SGD and SGD. | `50` |
+| `PrintAccuracy` | [`bool`](#doc_bool) | Print accuracies on initial and transformed dataset | `false` |
+| `Range` | [`int`](#doc_int) | Number of iterations after which impostors needs to be recalculated | `1` |
+| `Rank` | [`int`](#doc_int) | Rank of distance matrix to be optimized.  | `0` |
+| `Regularization` | [`float64`](#doc_float64) | Regularization for LMNN objective function  | `0.5` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `StepSize` | [`float64`](#doc_float64) | Step size for AMSGrad, BB_SGD and SGD (alpha). | `0.01` |
+| `Tolerance` | [`float64`](#doc_float64) | Maximum tolerance for termination of AMSGrad, BB_SGD, SGD or L-BFGS. | `1e-07` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `CenteredData` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix for mean-centered dataset. | 
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix for learned distance matrix. | 
+| `TransformedData` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix for transformed dataset. | 
+
+### Detailed documentation
+{: #lmnn_detailed-documentation }
+
+This program implements Large Margin Nearest Neighbors, a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset.  The method employes the strategy of reducing distance between similar labeled data points (a.k.a target neighbors) and increasing distance between differently labeled points (a.k.a impostors) using standard optimization techniques over the gradient of the distance between data points.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `Input`), or alternatively as a separate matrix (specified with `Labels`).  Additionally, a starting point for optimization (specified with `Distance`can be given, having (r x d) dimensionality.  Here r should satisfy 1 <= r <= d, Consequently a Low-Rank matrix will be optimized. Alternatively, Low-Rank distance can be learned by specifying the `Rank`parameter (A Low-Rank matrix with uniformly distributed values will be used as initial learning point). 
+
+The program also requires number of targets neighbors to work with ( specified with `K`), A regularization parameter can also be passed, It acts as a trade of between the pulling and pushing terms (specified with `Regularization`), In addition, this implementation of LMNN includes a parameter to decide the interval after which impostors must be re-calculated (specified with `Range`).
+
+Output can either be the learned distance matrix (specified with `Output`), or the transformed dataset  (specified with `TransformedData`), or both. Additionally mean-centered dataset (specified with `CenteredData`) can be accessed given mean-centering (specified with `Center`) is performed on the dataset. Accuracy on initial dataset and final transformed dataset can be printed by specifying the `PrintAccuracy`parameter. 
+
+This implementation of LMNN uses AdaGrad, BigBatch_SGD, stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer. 
+
+AdaGrad, specified by the value 'adagrad' for the parameter `Optimizer`, uses maximum of past squared gradients. It primarily on six parameters: the step size (specified with `StepSize`), the batch size (specified with `BatchSize`), the maximum number of passes (specified with `Passes`). Inaddition, a normalized starting point can be used by specifying the `Normalize` parameter. 
+
+BigBatch_SGD, specified by the value 'bbsgd' for the parameter `Optimizer`, depends primarily on four parameters: the step size (specified with `StepSize`), the batch size (specified with `BatchSize`), the maximum number of passes (specified with `Passes`).  In addition, a normalized starting point can be used by specifying the `Normalize` parameter. 
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `Optimizer`, depends primarily on three parameters: the step size (specified with `StepSize`), the batch size (specified with `BatchSize`), and the maximum number of passes (specified with `Passes`).  In addition, a normalized starting point can be used by specifying the `Normalize` parameter. Furthermore, mean-centering can be performed on the dataset by specifying the `Center`parameter. 
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `Optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `MaxIterations`, `Tolerance`(the optimization is terminated when the gradient norm is below this value).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.  In addition, a normalized starting point can be used by specifying the `Normalize` parameter.
+
+By default, the AMSGrad optimizer is used.
+
+### Example
+Example - Let's say we want to learn distance on iris dataset with number of targets as 3 using BigBatch_SGD optimizer. A simple call for the same will look like: 
+
+```go
+// Initialize optional parameters for Lmnn().
+param := mlpack.LmnnOptions()
+param.Labels = iris_labels
+param.K = 3
+param.Optimizer = "bbsgd"
+
+_, output, _ := mlpack.Lmnn(iris, param)
+```
+
+An another program call making use of range & regularization parameter with dataset having labels as last column can be made as: 
+
+```go
+// Initialize optional parameters for Lmnn().
+param := mlpack.LmnnOptions()
+param.K = 5
+param.Range = 10
+param.Regularization = 0.4
+
+_, output, _ := mlpack.Lmnn(letter_recognition, param)
+```
+
+### See also
+
+ - [Nca()](#nca)
+ - [Large margin nearest neighbor on Wikipedia](https://en.wikipedia.org/wiki/Large_margin_nearest_neighbor)
+ - [Distance metric learning for large margin nearest neighbor classification (pdf)](http://papers.nips.cc/paper/2795-distance-metric-learning-for-large-margin-nearest-neighbor-classification.pdf)
+ - [LMNN C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lmnn/lmnn.hpp)
+
+## LocalCoordinateCoding()
+{: #local_coordinate_coding }
+
+#### Local Coordinate Coding
+{: #local_coordinate_coding_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for LocalCoordinateCoding().
+param := mlpack.LocalCoordinateCodingOptions()
+param.Atoms = 0
+param.InitialDictionary = mat.NewDense(1, 1, nil)
+param.InputModel = nil
+param.Lambda = 0
+param.MaxIterations = 0
+param.Normalize = false
+param.Seed = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.Tolerance = 0.01
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+codes, dictionary, output_model := mlpack.LocalCoordinateCoding(param)
+```
+
+An implementation of Local Coordinate Coding (LCC), a data transformation technique.  Given input data, this transforms each point to be expressed as a linear combination of a few points in the dataset; once an LCC model is trained, it can be used to transform points later also. [Detailed documentation](#local_coordinate_coding_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Atoms` | [`int`](#doc_int) | Number of atoms in the dictionary. | `0` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InitialDictionary` | [`*mat.Dense`](#doc_a__mat_Dense) | Optional initial dictionary. | `mat.NewDense(1, 1, nil)` |
+| `InputModel` | [`localCoordinateCoding`](#doc_model) | Input LCC model. | `nil` |
+| `Lambda` | [`float64`](#doc_float64) | Weighted l1-norm regularization parameter. | `0` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations for LCC (0 indicates no limit). | `0` |
+| `Normalize` | [`bool`](#doc_bool) | If set, the input data matrix will be normalized before coding. | `false` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Test points to encode. | `mat.NewDense(1, 1, nil)` |
+| `Tolerance` | [`float64`](#doc_float64) | Tolerance for objective function. | `0.01` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of training data (X). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Codes` | [`*mat.Dense`](#doc_a__mat_Dense) | Output codes matrix. | 
+| `Dictionary` | [`*mat.Dense`](#doc_a__mat_Dense) | Output dictionary matrix. | 
+| `OutputModel` | [`localCoordinateCoding`](#doc_model) | Output for trained LCC model. | 
+
+### Detailed documentation
+{: #local_coordinate_coding_detailed-documentation }
+
+An implementation of Local Coordinate Coding (LCC), which codes data that approximately lives on a manifold using a variation of l1-norm regularized sparse coding.  Given a dense data matrix X with n points and d dimensions, LCC seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a coding matrix Z with n points in k dimensions.  Because of the regularization method used, the atoms in D should lie close to the manifold on which the data points lie.
+
+The original data matrix X can then be reconstructed as D * Z.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a coding step, which updates the coding matrix Z.
+
+To run this program, the input matrix X must be specified (with -i), along with the number of atoms in the dictionary (-k).  An initial dictionary may also be specified with the `InitialDictionary` parameter.  The l1-norm regularization parameter is specified with the `Lambda` parameter.
+
+### Example
+For example, to run LCC on the dataset `data` using 200 atoms and an l1-regularization parameter of 0.1, saving the dictionary `Dictionary` and the codes into `Codes`, use
+
+```go
+// Initialize optional parameters for LocalCoordinateCoding().
+param := mlpack.LocalCoordinateCodingOptions()
+param.Training = data
+param.Atoms = 200
+param.Lambda = 0.1
+
+codes, dict, _ := mlpack.LocalCoordinateCoding(param)
+```
+
+The maximum number of iterations may be specified with the `MaxIterations` parameter. Optionally, the input data matrix X can be normalized before coding with the `Normalize` parameter.
+
+An LCC model may be saved using the `OutputModel` output parameter.  Then, to encode new points from the dataset `points` with the previously saved model `lcc_model`, saving the new codes to `new_codes`, the following command can be used:
+
+```go
+// Initialize optional parameters for LocalCoordinateCoding().
+param := mlpack.LocalCoordinateCodingOptions()
+param.InputModel = &lcc_model
+param.Test = points
+
+new_codes, _, _ := mlpack.LocalCoordinateCoding(param)
+```
+
+### See also
+
+ - [SparseCoding()](#sparse_coding)
+ - [Nonlinear learning using local coordinate coding (pdf)](https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-coordinate-coding.pdf)
+ - [LocalCoordinateCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding.hpp)
+
+## LogisticRegression()
+{: #logistic_regression }
+
+#### L2-regularized Logistic Regression and Prediction
+{: #logistic_regression_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for LogisticRegression().
+param := mlpack.LogisticRegressionOptions()
+param.BatchSize = 64
+param.DecisionBoundary = 0.5
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.Lambda = 0
+param.MaxIterations = 10000
+param.Optimizer = "lbfgs"
+param.PrintTrainingAccuracy = false
+param.StepSize = 0.01
+param.Test = mat.NewDense(1, 1, nil)
+param.Tolerance = 1e-10
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, predictions, probabilities :=
+    mlpack.LogisticRegression(param)
+```
+
+An implementation of L2-regularized logistic regression for two-class classification.  Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#logistic_regression_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `BatchSize` | [`int`](#doc_int) | Batch size for SGD. | `64` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `DecisionBoundary` | [`float64`](#doc_float64) | Decision boundary for prediction; if the logistic function for a point is less than the boundary, the class is taken to be 0; otherwise, the class is 1. | `0.5` |
+| `InputModel` | [`logisticRegression`](#doc_model) | Existing model (parameters). | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | A matrix containing labels (0 or 1) for the points in the training set (y). | `mat.NewDense(1, 1, nil)` |
+| `Lambda` | [`float64`](#doc_float64) | L2-regularization parameter for training. | `0` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `Optimizer` | [`string`](#doc_string) | Optimizer to use for training ('lbfgs' or 'sgd'). | `"lbfgs"` |
+| `PrintTrainingAccuracy` | [`bool`](#doc_bool) | If set, then the accuracy of the model on the training set will be printed (verbose must also be specified). | `false` |
+| `StepSize` | [`float64`](#doc_float64) | Step size for SGD optimizer. | `0.01` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing test dataset. | `mat.NewDense(1, 1, nil)` |
+| `Tolerance` | [`float64`](#doc_float64) | Convergence tolerance for optimizer. | `1e-10` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the training set (the matrix of predictors, X). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`logisticRegression`](#doc_model) | Output for trained logistic regression model. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #logistic_regression_detailed-documentation }
+
+An implementation of L2-regularized logistic regression using either the L-BFGS optimizer or SGD (stochastic gradient descent).  This solves the regression problem
+
+  y = (1 / 1 + e^-(X * b)).
+
+In this setting, y corresponds to class labels and X corresponds to data.
+
+This program allows loading a logistic regression model (via the `InputModel` parameter) or training a logistic regression model given training data (specified with the `Training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `Test` parameter) and the classification results may be saved with the `Predictions` output parameter. The trained logistic regression model may be saved using the `OutputModel` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `Labels` parameter may be used to specify a separate matrix of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `Lambda` option, and the optimizer used to train the model can be specified with the `Optimizer` parameter.  Available options are 'sgd' (stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `MaxIterations` parameter specifies the maximum number of allowed iterations, and the `Tolerance` parameter specifies the tolerance for convergence.  For the SGD optimizer, the `StepSize` parameter controls the step size taken at each iteration by the optimizer.  The batch size for SGD is controlled with the `BatchSize` parameter. If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+For SGD, an iteration refers to a single point. So to take a single pass over the dataset with SGD, `MaxIterations` should be set to the number of points in the dataset.
+
+Optionally, the model can be used to predict the responses for another matrix of data points, if `Test` is specified.  The `Test` parameter can be specified without the `Training` parameter, so long as an existing logistic regression model is given with the `InputModel` parameter.  The output predictions from the logistic regression model may be saved with the `Predictions` parameter.
+
+This implementation of logistic regression does not support the general multi-class case but instead only the two-class case.  Any labels must be either 0 or 1.  For more classes, see the softmax regression implementation.
+
+### Example
+As an example, to train a logistic regression model on the data '`data`' with labels '`labels`' with L2 regularization of 0.1, saving the model to '`lr_model`', the following command may be used:
+
+```go
+// Initialize optional parameters for LogisticRegression().
+param := mlpack.LogisticRegressionOptions()
+param.Training = data
+param.Labels = labels
+param.Lambda = 0.1
+param.PrintTrainingAccuracy = true
+
+lr_model, _, _ := mlpack.LogisticRegression(param)
+```
+
+Then, to use that model to predict classes for the dataset '`test`', storing the output predictions in '`predictions`', the following command may be used: 
+
+```go
+// Initialize optional parameters for LogisticRegression().
+param := mlpack.LogisticRegressionOptions()
+param.InputModel = &lr_model
+param.Test = test
+
+_, predictions, _ := mlpack.LogisticRegression(param)
+```
+
+### See also
+
+ - [SoftmaxRegression()](#softmax_regression)
+ - [RandomForest()](#random_forest)
+ - [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+ - [:LogisticRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/logistic_regression/logistic_regression.hpp)
+
+## Lsh()
+{: #lsh }
+
+#### K-Approximate-Nearest-Neighbor Search with LSH
+{: #lsh_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Lsh().
+param := mlpack.LshOptions()
+param.BucketSize = 500
+param.HashWidth = 0
+param.InputModel = nil
+param.K = 0
+param.NumProbes = 0
+param.Projections = 10
+param.Query = mat.NewDense(1, 1, nil)
+param.Reference = mat.NewDense(1, 1, nil)
+param.SecondHashSize = 99901
+param.Seed = 0
+param.Tables = 30
+param.TrueNeighbors = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+distances, neighbors, output_model := mlpack.Lsh(param)
+```
+
+An implementation of approximate k-nearest-neighbor search with locality-sensitive hashing (LSH).  Given a set of reference points and a set of query points, this will compute the k approximate nearest neighbors of each query point in the reference set; models can be saved for future use. [Detailed documentation](#lsh_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `BucketSize` | [`int`](#doc_int) | The size of a bucket in the second level hash. | `500` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `HashWidth` | [`float64`](#doc_float64) | The hash width for the first-level hashing in the LSH preprocessing. By default, the LSH class automatically estimates a hash width for its use. | `0` |
+| `InputModel` | [`lshSearch`](#doc_model) | Input LSH model. | `nil` |
+| `K` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `NumProbes` | [`int`](#doc_int) | Number of additional probes for multiprobe LSH; if 0, traditional LSH is used. | `0` |
+| `Projections` | [`int`](#doc_int) | The number of hash functions for each table | `10` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing query points (optional). | `mat.NewDense(1, 1, nil)` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing the reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `SecondHashSize` | [`int`](#doc_int) | The size of the second level hash table. | `99901` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Tables` | [`int`](#doc_int) | The number of hash tables to be used. | `30` |
+| `TrueNeighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of true neighbors to compute recall with (the recall is printed when -v is specified). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Distances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output distances into. | 
+| `Neighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output neighbors into. | 
+| `OutputModel` | [`lshSearch`](#doc_model) | Output for trained LSH model. | 
+
+### Detailed documentation
+{: #lsh_detailed-documentation }
+
+This program will calculate the k approximate-nearest-neighbors of a set of points using locality-sensitive hashing. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. 
+
+### Example
+For example, the following will return 5 neighbors from the data for each point in `input` and store the distances in `distances` and the neighbors in `neighbors`:
+
+```go
+// Initialize optional parameters for Lsh().
+param := mlpack.LshOptions()
+param.K = 5
+param.Reference = input
+
+distances, neighbors, _ := mlpack.Lsh(param)
+```
+
+The output is organized such that row i and column j in the neighbors output corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output file corresponds to the distance between those two points.
+
+Because this is approximate-nearest-neighbors search, results may be different from run to run.  Thus, the `Seed` parameter can be specified to set the random seed.
+
+This program also has many other parameters to control its functionality; see the parameter-specific documentation for more information.
+
+### See also
+
+ - [Knn()](#knn)
+ - [Krann()](#krann)
+ - [Locality-sensitive hashing on Wikipedia](https://en.wikipedia.org/wiki/Locality-sensitive_hashing)
+ - [Locality-sensitive hashing scheme based on p-stable  distributions(pdf)](https://www.mlpack.org/papers/lsh.pdf)
+ - [LSHSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lsh/lsh.hpp)
+
+## MeanShift()
+{: #mean_shift }
+
+#### Mean Shift Clustering
+{: #mean_shift_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for MeanShift().
+param := mlpack.MeanShiftOptions()
+param.ForceConvergence = false
+param.InPlace = false
+param.LabelsOnly = false
+param.MaxIterations = 1000
+param.Radius = 0
+param.Verbose = false
+
+centroid, output := mlpack.MeanShift(input, param)
+```
+
+A fast implementation of mean-shift clustering using dual-tree range search.  Given a dataset, this uses the mean shift algorithm to produce and return a clustering of the data. [Detailed documentation](#mean_shift_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `ForceConvergence` | [`bool`](#doc_bool) | If specified, the mean shift algorithm will continue running regardless of max_iterations until the clusters converge. | `false` |
+| `InPlace` | [`bool`](#doc_bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden.  (Do not use with Python.) | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform clustering on. | `**--**` |
+| `LabelsOnly` | [`bool`](#doc_bool) | If specified, only the output labels will be written to the file specified by --output_file. | `false` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations before mean shift terminates. | `1000` |
+| `Radius` | [`float64`](#doc_float64) | If the distance between two centroids is less than the given radius, one will be removed.  A radius of 0 or less means an estimate will be calculated and used for the radius. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Centroid` | [`*mat.Dense`](#doc_a__mat_Dense) | If specified, the centroids of each cluster will be written to the given matrix. | 
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to write output labels or labeled data to. | 
+
+### Detailed documentation
+{: #mean_shift_detailed-documentation }
+
+This program performs mean shift clustering on the given dataset, storing the learned cluster assignments either as a column of labels in the input dataset or separately.
+
+The input dataset should be specified with the `Input` parameter, and the radius used for search can be specified with the `Radius` parameter.  The maximum number of iterations before algorithm termination is controlled with the `MaxIterations` parameter.
+
+The output labels may be saved with the `Output` output parameter and the centroids of each cluster may be saved with the `Centroid` output parameter.
+
+### Example
+For example, to run mean shift clustering on the dataset `data` and store the centroids to `centroids`, the following command may be used: 
+
+```go
+// Initialize optional parameters for MeanShift().
+param := mlpack.MeanShiftOptions()
+
+centroids, _ := mlpack.MeanShift(data, param)
+```
+
+### See also
+
+ - [Kmeans()](#kmeans)
+ - [Dbscan()](#dbscan)
+ - [Mean shift on Wikipedia](https://en.wikipedia.org/wiki/Mean_shift)
+ - [Mean Shift, Mode Seeking, and Clustering (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=1c168275c59ba382588350ee1443537f59978183)
+ - [mlpack::mean_shift::MeanShift C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/mean_shift/mean_shift.hpp)
+
+## Nbc()
+{: #nbc }
+
+#### Parametric Naive Bayes Classifier
+{: #nbc_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Nbc().
+param := mlpack.NbcOptions()
+param.IncrementalVariance = false
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output, output_model, output_probs, predictions, probabilities :=
+    mlpack.Nbc(param)
+```
+
+An implementation of the Naive Bayes Classifier, used for classification. Given labeled data, an NBC model can be trained and saved, or, a pre-trained model can be used for classification. [Detailed documentation](#nbc_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `IncrementalVariance` | [`bool`](#doc_bool) | The variance of each class will be calculated incrementally. | `false` |
+| `InputModel` | [`nbcModel`](#doc_model) | Input Naive Bayes model. | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | A file containing labels for the training set. | `mat.NewDense(1, 1, nil)` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the test set. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the training set. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | The matrix in which the predicted labels for the test set will be written (deprecated). | 
+| `OutputModel` | [`nbcModel`](#doc_model) | File to save trained Naive Bayes model to. | 
+| `OutputProbs` | [`*mat.Dense`](#doc_a__mat_Dense) | The matrix in which the predicted probability of labels for the test set will be written (deprecated). | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | The matrix in which the predicted labels for the test set will be written. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | The matrix in which the predicted probability of labels for the test set will be written. | 
+
+### Detailed documentation
+{: #nbc_detailed-documentation }
+
+This program trains the Naive Bayes classifier on the given labeled training set, or loads a model from the given model file, and then may use that trained model to classify the points in a given test set.
+
+The training set is specified with the `Training` parameter.  Labels may be either the last row of the training set, or alternately the `Labels` parameter may be specified to pass a separate matrix of labels.
+
+If training is not desired, a pre-existing model may be loaded with the `InputModel` parameter.
+
+
+
+The `IncrementalVariance` parameter can be used to force the training to use an incremental algorithm for calculating variance.  This is slower, but can help avoid loss of precision in some cases.
+
+If classifying a test set is desired, the test set may be specified with the `Test` parameter, and the classifications may be saved with the `Predictions`predictions  parameter.  If saving the trained model is desired, this may be done with the `OutputModel` output parameter.
+
+Note: the `Output` and `OutputProbs` parameters are deprecated and will be removed in mlpack 4.0.0.  Use `Predictions` and `Probabilities` instead.
+
+### Example
+For example, to train a Naive Bayes classifier on the dataset `data` with labels `labels` and save the model to `nbc_model`, the following command may be used:
+
+```go
+// Initialize optional parameters for Nbc().
+param := mlpack.NbcOptions()
+param.Training = data
+param.Labels = labels
+
+_, nbc_model, _, _, _ := mlpack.Nbc(param)
+```
+
+Then, to use `nbc_model` to predict the classes of the dataset `test_set` and save the predicted classes to `predictions`, the following command may be used:
+
+```go
+// Initialize optional parameters for Nbc().
+param := mlpack.NbcOptions()
+param.InputModel = &nbc_model
+param.Test = test_set
+
+predictions, _, _, _, _ := mlpack.Nbc(param)
+```
+
+### See also
+
+ - [SoftmaxRegression()](#softmax_regression)
+ - [RandomForest()](#random_forest)
+ - [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+ - [NaiveBayesClassifier C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp)
+
+## Nca()
+{: #nca }
+
+#### Neighborhood Components Analysis (NCA)
+{: #nca_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Nca().
+param := mlpack.NcaOptions()
+param.ArmijoConstant = 0.0001
+param.BatchSize = 50
+param.Labels = mat.NewDense(1, 1, nil)
+param.LinearScan = false
+param.MaxIterations = 500000
+param.MaxLineSearchTrials = 50
+param.MaxStep = 1e+20
+param.MinStep = 1e-20
+param.Normalize = false
+param.NumBasis = 5
+param.Optimizer = "sgd"
+param.Seed = 0
+param.StepSize = 0.01
+param.Tolerance = 1e-07
+param.Verbose = false
+param.Wolfe = 0.9
+
+output := mlpack.Nca(input, param)
+```
+
+An implementation of neighborhood components analysis, a distance learning technique that can be used for preprocessing.  Given a labeled dataset, this uses NCA, which seeks to improve the k-nearest-neighbor classification, and returns the learned distance metric. [Detailed documentation](#nca_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `ArmijoConstant` | [`float64`](#doc_float64) | Armijo constant for L-BFGS. | `0.0001` |
+| `BatchSize` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to run NCA on. | `**--**` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels for input dataset. | `mat.NewDense(1, 1, nil)` |
+| `LinearScan` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `false` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations for SGD or L-BFGS (0 indicates no limit). | `500000` |
+| `MaxLineSearchTrials` | [`int`](#doc_int) | Maximum number of line search trials for L-BFGS. | `50` |
+| `MaxStep` | [`float64`](#doc_float64) | Maximum step of line search for L-BFGS. | `1e+20` |
+| `MinStep` | [`float64`](#doc_float64) | Minimum step of line search for L-BFGS. | `1e-20` |
+| `Normalize` | [`bool`](#doc_bool) | Use a normalized starting point for optimization. This is useful for when points are far apart, or when SGD is returning NaN. | `false` |
+| `NumBasis` | [`int`](#doc_int) | Number of memory points to be stored for L-BFGS. | `5` |
+| `Optimizer` | [`string`](#doc_string) | Optimizer to use; 'sgd' or 'lbfgs'. | `"sgd"` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `StepSize` | [`float64`](#doc_float64) | Step size for stochastic gradient descent (alpha). | `0.01` |
+| `Tolerance` | [`float64`](#doc_float64) | Maximum tolerance for termination of SGD or L-BFGS. | `1e-07` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `Wolfe` | [`float64`](#doc_float64) | Wolfe condition parameter for L-BFGS. | `0.9` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Output matrix for learned distance matrix. | 
+
+### Detailed documentation
+{: #nca_detailed-documentation }
+
+This program implements Neighborhood Components Analysis, both a linear dimensionality reduction technique and a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset by scaling the dimensions.  The method is nonparametric, and does not require a value of k.  It works by using stochastic ("soft") neighbor assignments and using optimization techniques over the gradient of the accuracy of the neighbor assignments.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `Input`), or alternatively as a separate matrix (specified with `Labels`).
+
+This implementation of NCA uses stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer.  These optimizers do not guarantee global convergence for a nonconvex objective function (NCA's objective function is nonconvex), so the final results could depend on the random seed or other optimizer parameters.
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `Optimizer`, depends primarily on three parameters: the step size (specified with `StepSize`), the batch size (specified with `BatchSize`), and the maximum number of iterations (specified with `MaxIterations`).  In addition, a normalized starting point can be used by specifying the `Normalize` parameter, which is necessary if many warnings of the form 'Denominator of p_i is 0!' are given.  Tuning the step size can be a tedious affair.  In general, the step size is too large if the objective is not mostly uniformly decreasing, or if zero-valued denominator warnings are being issued.  The step size is too small if the objective is changing very slowly.  Setting the termination condition can be done easily once a good step size parameter is found; either increase the maximum iterations to a large number and allow SGD to find a minimum, or set the maximum iterations to 0 (allowing infinite iterations) and set the tolerance (specified by `Tolerance`) to define the maximum allowed difference between objectives for SGD to terminate.  Be careful---setting the tolerance instead of the maximum iterations can take a very long time and may actually never converge due to the properties of the SGD optimizer. Note that a single iteration of SGD refers to a single point, so to take a single pass over the dataset, set the value of the `MaxIterations` parameter equal to the number of points in the dataset.
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `Optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `NumBasis` (specifies the number of memory points used by L-BFGS), `MaxIterations`, `ArmijoConstant`, `Wolfe`, `Tolerance` (the optimization is terminated when the gradient norm is below this value), `MaxLineSearchTrials`, `MinStep`, and `MaxStep` (which both refer to the line search routine).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.
+
+By default, the SGD optimizer is used.
+
+### See also
+
+ - [Lmnn()](#lmnn)
+ - [Neighbourhood components analysis on Wikipedia](https://en.wikipedia.org/wiki/Neighbourhood_components_analysis)
+ - [Neighbourhood components analysis (pdf)](http://papers.nips.cc/paper/2566-neighbourhood-components-analysis.pdf)
+ - [NCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/nca/nca.hpp)
+
+## Knn()
+{: #knn }
+
+#### k-Nearest-Neighbors Search
+{: #knn_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Knn().
+param := mlpack.KnnOptions()
+param.Algorithm = "dual_tree"
+param.Epsilon = 0
+param.InputModel = nil
+param.K = 0
+param.LeafSize = 20
+param.Query = mat.NewDense(1, 1, nil)
+param.RandomBasis = false
+param.Reference = mat.NewDense(1, 1, nil)
+param.Rho = 0.7
+param.Seed = 0
+param.Tau = 0
+param.TreeType = "kd"
+param.TrueDistances = mat.NewDense(1, 1, nil)
+param.TrueNeighbors = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+distances, neighbors, output_model := mlpack.Knn(param)
+```
+
+An implementation of k-nearest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#knn_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Algorithm` | [`string`](#doc_string) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Epsilon` | [`float64`](#doc_float64) | If specified, will do approximate nearest neighbor search with given relative error. | `0` |
+| `InputModel` | [`knnModel`](#doc_model) | Pre-trained kNN model. | `nil` |
+| `K` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `LeafSize` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, spill trees, and octrees). | `20` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing query points (optional). | `mat.NewDense(1, 1, nil)` |
+| `RandomBasis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing the reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `Rho` | [`float64`](#doc_float64) | Balance threshold (only valid for spill trees). | `0.7` |
+| `Seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `Tau` | [`float64`](#doc_float64) | Overlapping size (only valid for spill trees). | `0` |
+| `TreeType` | [`string`](#doc_string) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'spill', 'oct'. | `"kd"` |
+| `TrueDistances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `mat.NewDense(1, 1, nil)` |
+| `TrueNeighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Distances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output distances into. | 
+| `Neighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output neighbors into. | 
+| `OutputModel` | [`knnModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #knn_detailed-documentation }
+
+This program will calculate the k-nearest-neighbors of a set of points using kd-trees or cover trees (cover tree support is experimental and may be slow). You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following command will calculate the 5 nearest neighbors of each point in `input` and store the distances in `distances` and the neighbors in `neighbors`: 
+
+```go
+// Initialize optional parameters for Knn().
+param := mlpack.KnnOptions()
+param.K = 5
+param.Reference = input
+
+distances, neighbors, _ := mlpack.Knn(param)
+```
+
+The output is organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output matrix corresponds to the distance between those two points.
+
+### See also
+
+ - [Lsh()](#lsh)
+ - [Krann()](#krann)
+ - [Kfn()](#kfn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## Kfn()
+{: #kfn }
+
+#### k-Furthest-Neighbors Search
+{: #kfn_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Kfn().
+param := mlpack.KfnOptions()
+param.Algorithm = "dual_tree"
+param.Epsilon = 0
+param.InputModel = nil
+param.K = 0
+param.LeafSize = 20
+param.Percentage = 1
+param.Query = mat.NewDense(1, 1, nil)
+param.RandomBasis = false
+param.Reference = mat.NewDense(1, 1, nil)
+param.Seed = 0
+param.TreeType = "kd"
+param.TrueDistances = mat.NewDense(1, 1, nil)
+param.TrueNeighbors = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+distances, neighbors, output_model := mlpack.Kfn(param)
+```
+
+An implementation of k-furthest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k furthest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#kfn_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Algorithm` | [`string`](#doc_string) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Epsilon` | [`float64`](#doc_float64) | If specified, will do approximate furthest neighbor search with given relative error. Must be in the range [0,1). | `0` |
+| `InputModel` | [`kfnModel`](#doc_model) | Pre-trained kFN model. | `nil` |
+| `K` | [`int`](#doc_int) | Number of furthest neighbors to find. | `0` |
+| `LeafSize` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `Percentage` | [`float64`](#doc_float64) | If specified, will do approximate furthest neighbor search. Must be in the range (0,1] (decimal form). Resultant neighbors will be at least (p*100) % of the distance as the true furthest neighbor. | `1` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing query points (optional). | `mat.NewDense(1, 1, nil)` |
+| `RandomBasis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing the reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `Seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `TreeType` | [`string`](#doc_string) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `TrueDistances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `mat.NewDense(1, 1, nil)` |
+| `TrueNeighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Distances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output distances into. | 
+| `Neighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output neighbors into. | 
+| `OutputModel` | [`kfnModel`](#doc_model) | If specified, the kFN model will be output here. | 
+
+### Detailed documentation
+{: #kfn_detailed-documentation }
+
+This program will calculate the k-furthest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following will calculate the 5 furthest neighbors of eachpoint in `input` and store the distances in `distances` and the neighbors in `neighbors`: 
+
+```go
+// Initialize optional parameters for Kfn().
+param := mlpack.KfnOptions()
+param.K = 5
+param.Reference = input
+
+distances, neighbors, _ := mlpack.Kfn(param)
+```
+
+The output files are organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th furthest neighbor from the point in the query set with index i.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [ApproxKfn()](#approx_kfn)
+ - [Knn()](#knn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## Nmf()
+{: #nmf }
+
+#### Non-negative Matrix Factorization
+{: #nmf_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Nmf().
+param := mlpack.NmfOptions()
+param.InitialH = mat.NewDense(1, 1, nil)
+param.InitialW = mat.NewDense(1, 1, nil)
+param.MaxIterations = 10000
+param.MinResidue = 1e-05
+param.Seed = 0
+param.UpdateRules = "multdist"
+param.Verbose = false
+
+h, w := mlpack.Nmf(input, rank, param)
+```
+
+An implementation of non-negative matrix factorization.  This can be used to decompose an input dataset into two low-rank non-negative components. [Detailed documentation](#nmf_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InitialH` | [`*mat.Dense`](#doc_a__mat_Dense) | Initial H matrix. | `mat.NewDense(1, 1, nil)` |
+| `InitialW` | [`*mat.Dense`](#doc_a__mat_Dense) | Initial W matrix. | `mat.NewDense(1, 1, nil)` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform NMF on. | `**--**` |
+| `MaxIterations` | [`int`](#doc_int) | Number of iterations before NMF terminates (0 runs until convergence. | `10000` |
+| `MinResidue` | [`float64`](#doc_float64) | The minimum root mean square residue allowed for each iteration, below which the program terminates. | `1e-05` |
+| `rank` | [`int`](#doc_int) | Rank of the factorization. | `**--**` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `UpdateRules` | [`string`](#doc_string) | Update rules for each iteration; ( multdist \| multdiv \| als ). | `"multdist"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `H` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save the calculated H to. | 
+| `W` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save the calculated W to. | 
+
+### Detailed documentation
+{: #nmf_detailed-documentation }
+
+This program performs non-negative matrix factorization on the given dataset, storing the resulting decomposed matrices in the specified files.  For an input dataset V, NMF decomposes V into two matrices W and H such that 
+
+V = W * H
+
+where all elements in W and H are non-negative.  If V is of size (n x m), then W will be of size (n x r) and H will be of size (r x m), where r is the rank of the factorization (specified by the `Rank` parameter).
+
+Optionally, the desired update rules for each NMF iteration can be chosen from the following list:
+
+ - multdist: multiplicative distance-based update rules (Lee and Seung 1999)
+ - multdiv: multiplicative divergence-based update rules (Lee and Seung 1999)
+ - als: alternating least squares update rules (Paatero and Tapper 1994)
+
+The maximum number of iterations is specified with `MaxIterations`, and the minimum residue required for algorithm termination is specified with the `MinResidue` parameter.
+
+### Example
+For example, to run NMF on the input matrix `V` using the 'multdist' update rules with a rank-10 decomposition and storing the decomposed matrices into `W` and `H`, the following command could be used: 
+
+```go
+// Initialize optional parameters for Nmf().
+param := mlpack.NmfOptions()
+param.UpdateRules = "multdist"
+
+H, W := mlpack.Nmf(V, 10, param)
+```
+
+### See also
+
+ - [Cf()](#cf)
+ - [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ - [Algorithms for non-negative matrix factorization (pdf)](http://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf)
+ - [AMF C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/amf/amf.hpp)
+
+## Pca()
+{: #pca }
+
+#### Principal Components Analysis
+{: #pca_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Pca().
+param := mlpack.PcaOptions()
+param.DecompositionMethod = "exact"
+param.NewDimensionality = 0
+param.Scale = false
+param.VarToRetain = 0
+param.Verbose = false
+
+output := mlpack.Pca(input, param)
+```
+
+An implementation of several strategies for principal components analysis (PCA), a common preprocessing step.  Given a dataset and a desired new dimensionality, this can reduce the dimensionality of the data using the linear transformation determined by PCA. [Detailed documentation](#pca_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `DecompositionMethod` | [`string`](#doc_string) | Method used for the principal components analysis: 'exact', 'randomized', 'randomized-block-krylov', 'quic'. | `"exact"` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset to perform PCA on. | `**--**` |
+| `NewDimensionality` | [`int`](#doc_int) | Desired dimensionality of output dataset. If 0, no dimensionality reduction is performed. | `0` |
+| `Scale` | [`bool`](#doc_bool) | If set, the data will be scaled before running PCA, such that the variance of each feature is 1. | `false` |
+| `VarToRetain` | [`float64`](#doc_float64) | Amount of variance to retain; should be between 0 and 1.  If 1, all variance is retained.  Overrides -d. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #pca_detailed-documentation }
+
+This program performs principal components analysis on the given dataset using the exact, randomized, randomized block Krylov, or QUIC SVD method. It will transform the data onto its principal components, optionally performing dimensionality reduction by ignoring the principal components with the smallest eigenvalues.
+
+Use the `Input` parameter to specify the dataset to perform PCA on.  A desired new dimensionality can be specified with the `NewDimensionality` parameter, or the desired variance to retain can be specified with the `VarToRetain` parameter.  If desired, the dataset can be scaled before running PCA with the `Scale` parameter.
+
+Multiple different decomposition techniques can be used.  The method to use can be specified with the `DecompositionMethod` parameter, and it may take the values 'exact', 'randomized', or 'quic'.
+
+### Example
+For example, to reduce the dimensionality of the matrix `data` to 5 dimensions using randomized SVD for the decomposition, storing the output matrix to `data_mod`, the following command can be used:
+
+```go
+// Initialize optional parameters for Pca().
+param := mlpack.PcaOptions()
+param.NewDimensionality = 5
+param.DecompositionMethod = "randomized"
+
+data_mod := mlpack.Pca(data, param)
+```
+
+### See also
+
+ - [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+ - [PCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/pca/pca.hpp)
+
+## Perceptron()
+{: #perceptron }
+
+#### Perceptron
+{: #perceptron_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Perceptron().
+param := mlpack.PerceptronOptions()
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.MaxIterations = 1000
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output, output_model, predictions := mlpack.Perceptron(param)
+```
+
+An implementation of a perceptron---a single level neural network--=for classification.  Given labeled data, a perceptron can be trained and saved for future use; or, a pre-trained perceptron can be used for classification on new points. [Detailed documentation](#perceptron_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`perceptronModel`](#doc_model) | Input perceptron model. | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | A matrix containing labels for the training set. | `mat.NewDense(1, 1, nil)` |
+| `MaxIterations` | [`int`](#doc_int) | The maximum number of iterations the perceptron is to be run | `1000` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the test set. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the training set. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | The matrix in which the predicted labels for the test set will be written. | 
+| `OutputModel` | [`perceptronModel`](#doc_model) | Output for trained perceptron model. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | The matrix in which the predicted labels for the test set will be written. | 
+
+### Detailed documentation
+{: #perceptron_detailed-documentation }
+
+This program implements a perceptron, which is a single level neural network. The perceptron makes its predictions based on a linear predictor function combining a set of weights with the feature vector.  The perceptron learning rule is able to converge, given enough iterations (specified using the `MaxIterations` parameter), if the data supplied is linearly separable.  The perceptron is parameterized by a matrix of weight vectors that denote the numerical weights of the neural network.
+
+This program allows loading a perceptron from a model (via the `InputModel` parameter) or training a perceptron given training data (via the `Training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (via the `Test` parameter) and the classification results on the test set may be saved with the `Predictions` output parameter.  The perceptron model may be saved with the `OutputModel` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `Output`.
+Use `Predictions` instead of `Output`.
+
+### Example
+The training data given with the `Training` option may have class labels as its last dimension (so, if the training data is in CSV format, labels should be the last column).  Alternately, the `Labels` parameter may be used to specify a separate matrix of labels.
+
+All these options make it easy to train a perceptron, and then re-use that perceptron for later classification.  The invocation below trains a perceptron on `training_data` with labels `training_labels`, and saves the model to `perceptron_model`.
+
+```go
+// Initialize optional parameters for Perceptron().
+param := mlpack.PerceptronOptions()
+param.Training = training_data
+param.Labels = training_labels
+
+_, perceptron_model, _ := mlpack.Perceptron(param)
+```
+
+Then, this model can be re-used for classification on the test data `test_data`.  The example below does precisely that, saving the predicted classes to `predictions`.
+
+```go
+// Initialize optional parameters for Perceptron().
+param := mlpack.PerceptronOptions()
+param.InputModel = &perceptron_model
+param.Test = test_data
+
+_, _, predictions := mlpack.Perceptron(param)
+```
+
+Note that all of the options may be specified at once: predictions may be calculated right after training a model, and model training can occur even if an existing perceptron model is passed with the `InputModel` parameter.  However, note that the number of classes and the dimensionality of all data must match.  So you cannot pass a perceptron model trained on 2 classes and then re-train with a 4-class dataset.  Similarly, attempting classification on a 3-dimensional dataset with a perceptron that has been trained on 8 dimensions will cause an error.
+
+### See also
+
+ - [Adaboost()](#adaboost)
+ - [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+ - [Perceptron C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/perceptron/perceptron.hpp)
+
+## PreprocessSplit()
+{: #preprocess_split }
+
+#### Split Data
+{: #preprocess_split_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for PreprocessSplit().
+param := mlpack.PreprocessSplitOptions()
+param.InputLabels = mat.NewDense(1, 1, nil)
+param.NoShuffle = false
+param.Seed = 0
+param.StratifyData = false
+param.TestRatio = 0.2
+param.Verbose = false
+
+test, test_labels, training, training_labels :=
+    mlpack.PreprocessSplit(input, param)
+```
+
+A utility to split data into a training and testing dataset.  This can also split labels according to the same split. [Detailed documentation](#preprocess_split_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing data. | `**--**` |
+| `InputLabels` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing labels. | `mat.NewDense(1, 1, nil)` |
+| `NoShuffle` | [`bool`](#doc_bool) | Avoid shuffling the data before splitting. | `false` |
+| `Seed` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `StratifyData` | [`bool`](#doc_bool) | Stratify the data according to labels | `false` |
+| `TestRatio` | [`float64`](#doc_float64) | Ratio of test set; if not set,the ratio defaults to 0.2 | `0.2` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save test data to. | 
+| `TestLabels` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save test labels to. | 
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save training data to. | 
+| `TrainingLabels` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save train labels to. | 
+
+### Detailed documentation
+{: #preprocess_split_detailed-documentation }
+
+This utility takes a dataset and optionally labels and splits them into a training set and a test set. Before the split, the points in the dataset are randomly reordered. The percentage of the dataset to be used as the test set can be specified with the `TestRatio` parameter; the default is 0.2 (20%).
+
+The output training and test matrices may be saved with the `Training` and `Test` output parameters.
+
+Optionally, labels can also be split along with the data by specifying the `InputLabels` parameter.  Splitting labels works the same way as splitting the data. The output training and test labels may be saved with the `TrainingLabels` and `TestLabels` output parameters, respectively.
+
+### Example
+So, a simple example where we want to split the dataset `X` into `X_train` and `X_test` with 60% of the data in the training set and 40% of the dataset in the test set, we could run 
+
+```go
+// Initialize optional parameters for PreprocessSplit().
+param := mlpack.PreprocessSplitOptions()
+param.TestRatio = 0.4
+
+X_test, _, X_train, _ := mlpack.PreprocessSplit(X, param)
+```
+
+Also by default the dataset is shuffled and split; you can provide the `NoShuffle` option to avoid shuffling the data; an example to avoid shuffling of data is:
+
+```go
+// Initialize optional parameters for PreprocessSplit().
+param := mlpack.PreprocessSplitOptions()
+param.TestRatio = 0.4
+param.NoShuffle = true
+
+X_test, _, X_train, _ := mlpack.PreprocessSplit(X, param)
+```
+
+If we had a dataset `X` and associated labels `y`, and we wanted to split these into `X_train`, `y_train`, `X_test`, and `y_test`, with 30% of the data in the test set, we could run
+
+```go
+// Initialize optional parameters for PreprocessSplit().
+param := mlpack.PreprocessSplitOptions()
+param.InputLabels = y
+param.TestRatio = 0.3
+
+X_test, y_test, X_train, y_train := mlpack.PreprocessSplit(X, param)
+```
+
+To maintain the ratio of each class in the train and test sets, the`StratifyData` option can be used.
+
+```go
+// Initialize optional parameters for PreprocessSplit().
+param := mlpack.PreprocessSplitOptions()
+param.TestRatio = 0.4
+param.StratifyData = true
+
+X_test, _, X_train, _ := mlpack.PreprocessSplit(X, param)
+```
+
+### See also
+
+ - [PreprocessBinarize()](#preprocess_binarize)
+ - [PreprocessDescribe()](#preprocess_describe)
+
+## PreprocessBinarize()
+{: #preprocess_binarize }
+
+#### Binarize Data
+{: #preprocess_binarize_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for PreprocessBinarize().
+param := mlpack.PreprocessBinarizeOptions()
+param.Dimension = 0
+param.Threshold = 0
+param.Verbose = false
+
+output := mlpack.PreprocessBinarize(input, param)
+```
+
+A utility to binarize a dataset.  Given a dataset, this utility converts each value in the desired dimension(s) to 0 or 1; this can be a useful preprocessing step. [Detailed documentation](#preprocess_binarize_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Dimension` | [`int`](#doc_int) | Dimension to apply the binarization. If not set, the program will binarize every dimension by default. | `0` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input data matrix. | `**--**` |
+| `Threshold` | [`float64`](#doc_float64) | Threshold to be applied for binarization. If not set, the threshold defaults to 0.0. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix in which to save the output. | 
+
+### Detailed documentation
+{: #preprocess_binarize_detailed-documentation }
+
+This utility takes a dataset and binarizes the variables into either 0 or 1 given threshold. User can apply binarization on a dimension or the whole dataset.  The dimension to apply binarization to can be specified using the `Dimension` parameter; if left unspecified, every dimension will be binarized.  The threshold for binarization can also be specified with the `Threshold` parameter; the default threshold is 0.0.
+
+The binarized matrix may be saved with the `Output` output parameter.
+
+### Example
+For example, if we want to set all variables greater than 5 in the dataset `X` to 1 and variables less than or equal to 5.0 to 0, and save the result to `Y`, we could run
+
+```go
+// Initialize optional parameters for PreprocessBinarize().
+param := mlpack.PreprocessBinarizeOptions()
+param.Threshold = 5
+
+Y := mlpack.PreprocessBinarize(X, param)
+```
+
+But if we want to apply this to only the first (0th) dimension of `X`,  we could instead run
+
+```go
+// Initialize optional parameters for PreprocessBinarize().
+param := mlpack.PreprocessBinarizeOptions()
+param.Threshold = 5
+param.Dimension = 0
+
+Y := mlpack.PreprocessBinarize(X, param)
+```
+
+### See also
+
+ - [PreprocessDescribe()](#preprocess_describe)
+ - [PreprocessSplit()](#preprocess_split)
+
+## PreprocessDescribe()
+{: #preprocess_describe }
+
+#### Descriptive Statistics
+{: #preprocess_describe_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for PreprocessDescribe().
+param := mlpack.PreprocessDescribeOptions()
+param.Dimension = 0
+param.Population = false
+param.Precision = 4
+param.RowMajor = false
+param.Verbose = false
+param.Width = 8
+
+ := mlpack.PreprocessDescribe(input, param)
+```
+
+A utility for printing descriptive statistics about a dataset.  This prints a number of details about a dataset in a tabular format. [Detailed documentation](#preprocess_describe_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Dimension` | [`int`](#doc_int) | Dimension of the data. Use this to specify a dimension | `0` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing data, | `**--**` |
+| `Population` | [`bool`](#doc_bool) | If specified, the program will calculate statistics assuming the dataset is the population. By default, the program will assume the dataset as a sample. | `false` |
+| `Precision` | [`int`](#doc_int) | Precision of the output statistics. | `4` |
+| `RowMajor` | [`bool`](#doc_bool) | If specified, the program will calculate statistics across rows, not across columns.  (Remember that in mlpack, a column represents a point, so this option is generally not necessary.) | `false` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `Width` | [`int`](#doc_int) | Width of the output table. | `8` |
+
+
+### Detailed documentation
+{: #preprocess_describe_detailed-documentation }
+
+This utility takes a dataset and prints out the descriptive statistics of the data. Descriptive statistics is the discipline of quantitatively describing the main features of a collection of information, or the quantitative description itself. The program does not modify the original file, but instead prints out the statistics to the console. The printed result will look like a table.
+
+Optionally, width and precision of the output can be adjusted by a user using the `Width` and `Precision` parameters. A user can also select a specific dimension to analyze if there are too many dimensions. The `Population` parameter can be specified when the dataset should be considered as a population.  Otherwise, the dataset will be considered as a sample.
+
+### Example
+So, a simple example where we want to print out statistical facts about the dataset `X` using the default settings, we could run 
+
+```go
+// Initialize optional parameters for PreprocessDescribe().
+param := mlpack.PreprocessDescribeOptions()
+param.Verbose = true
+
+ := mlpack.PreprocessDescribe(X, param)
+```
+
+If we want to customize the width to 10 and precision to 5 and consider the dataset as a population, we could run
+
+```go
+// Initialize optional parameters for PreprocessDescribe().
+param := mlpack.PreprocessDescribeOptions()
+param.Width = 10
+param.Precision = 5
+param.Verbose = true
+
+ := mlpack.PreprocessDescribe(X, param)
+```
+
+### See also
+
+ - [PreprocessBinarize()](#preprocess_binarize)
+ - [PreprocessSplit()](#preprocess_split)
+
+## PreprocessScale()
+{: #preprocess_scale }
+
+#### Scale Data
+{: #preprocess_scale_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for PreprocessScale().
+param := mlpack.PreprocessScaleOptions()
+param.Epsilon = 1e-06
+param.InputModel = nil
+param.InverseScaling = false
+param.MaxValue = 1
+param.MinValue = 0
+param.ScalerMethod = "standard_scaler"
+param.Seed = 0
+param.Verbose = false
+
+output, output_model := mlpack.PreprocessScale(input, param)
+```
+
+A utility to perform feature scaling on datasets using one of sixtechniques.  Both scaling and inverse scaling are supported, andscalers can be saved and then applied to other datasets. [Detailed documentation](#preprocess_scale_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Epsilon` | [`float64`](#doc_float64) | regularization Parameter for pcawhitening, or zcawhitening, should be between -1 to 1. | `1e-06` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing data. | `**--**` |
+| `InputModel` | [`scalingModel`](#doc_model) | Input Scaling model. | `nil` |
+| `InverseScaling` | [`bool`](#doc_bool) | Inverse Scaling to get original dataset | `false` |
+| `MaxValue` | [`int`](#doc_int) | Ending value of range for min_max_scaler. | `1` |
+| `MinValue` | [`int`](#doc_int) | Starting value of range for min_max_scaler. | `0` |
+| `ScalerMethod` | [`string`](#doc_string) | method to use for scaling, the default is standard_scaler. | `"standard_scaler"` |
+| `Seed` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save scaled data to. | 
+| `OutputModel` | [`scalingModel`](#doc_model) | Output scaling model. | 
+
+### Detailed documentation
+{: #preprocess_scale_detailed-documentation }
+
+This utility takes a dataset and performs feature scaling using one of the six scaler methods namely: 'max_abs_scaler', 'mean_normalization', 'min_max_scaler' ,'standard_scaler', 'pca_whitening' and 'zca_whitening'. The function takes a matrix as `Input` and a scaling method type which you can specify using `ScalerMethod` parameter; the default is standard scaler, and outputs a matrix with scaled feature.
+
+The output scaled feature matrix may be saved with the `Output` output parameters.
+
+The model to scale features can be saved using `OutputModel` and later can be loaded back using`InputModel`.
+
+### Example
+So, a simple example where we want to scale the dataset `X` into `X_scaled` with  standard_scaler as scaler_method, we could run 
+
+```go
+// Initialize optional parameters for PreprocessScale().
+param := mlpack.PreprocessScaleOptions()
+param.ScalerMethod = "standard_scaler"
+
+X_scaled, _ := mlpack.PreprocessScale(X, param)
+```
+
+A simple example where we want to whiten the dataset `X` into `X_whitened` with  PCA as whitening_method and use 0.01 as regularization parameter, we could run 
+
+```go
+// Initialize optional parameters for PreprocessScale().
+param := mlpack.PreprocessScaleOptions()
+param.ScalerMethod = "pca_whitening"
+param.Epsilon = 0.01
+
+X_scaled, _ := mlpack.PreprocessScale(X, param)
+```
+
+You can also retransform the scaled dataset back using`InverseScaling`. An example to rescale : `X_scaled` into `X`using the saved model `InputModel` is:
+
+```go
+// Initialize optional parameters for PreprocessScale().
+param := mlpack.PreprocessScaleOptions()
+param.InverseScaling = true
+param.InputModel = &saved
+
+X, _ := mlpack.PreprocessScale(X_scaled, param)
+```
+
+Another simple example where we want to scale the dataset `X` into `X_scaled` with  min_max_scaler as scaler method, where scaling range is 1 to 3 instead of default 0 to 1. We could run 
+
+```go
+// Initialize optional parameters for PreprocessScale().
+param := mlpack.PreprocessScaleOptions()
+param.ScalerMethod = "min_max_scaler"
+param.MinValue = 1
+param.MaxValue = 3
+
+X_scaled, _ := mlpack.PreprocessScale(X, param)
+```
+
+### See also
+
+ - [PreprocessBinarize()](#preprocess_binarize)
+ - [PreprocessDescribe()](#preprocess_describe)
+
+## PreprocessOneHotEncoding()
+{: #preprocess_one_hot_encoding }
+
+#### One Hot Encoding
+{: #preprocess_one_hot_encoding_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for PreprocessOneHotEncoding().
+param := mlpack.PreprocessOneHotEncodingOptions()
+param.Dimensions = []int{}
+param.Verbose = false
+
+output := mlpack.PreprocessOneHotEncoding(input, param)
+```
+
+A utility to do one-hot encoding on features of dataset. [Detailed documentation](#preprocess_one_hot_encoding_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Dimensions` | [`array of ints`](#doc_array_of_ints) | Index of dimensions that need to be one-hot encoded (if unspecified, all categorical dimensions are one-hot encoded). | `[]int{}` |
+| `input` | [`matrixWithInfo`](#doc_matrixWithInfo) | Matrix containing data. | `**--**` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save one-hot encoded features data to. | 
+
+### Detailed documentation
+{: #preprocess_one_hot_encoding_detailed-documentation }
+
+This utility takes a dataset and a vector of indices and does one-hot encoding of the respective features at those indices. Indices represent the IDs of the dimensions to be one-hot encoded.
+
+If no dimensions are specified with `Dimensions`, then all categorical-type dimensions will be one-hot encoded. Otherwise, only the dimensions given in `Dimensions` will be one-hot encoded.
+
+The output matrix with encoded features may be saved with the `Output` parameters.
+
+### Example
+So, a simple example where we want to encode 1st and 3rd feature from dataset `X` into `X_output` would be
+
+```go
+// Initialize optional parameters for PreprocessOneHotEncoding().
+param := mlpack.PreprocessOneHotEncodingOptions()
+param.Dimensions = 1
+param.Dimensions = 3
+
+X_ouput := mlpack.PreprocessOneHotEncoding(X, param)
+```
+
+### See also
+
+ - [PreprocessBinarize()](#preprocess_binarize)
+ - [PreprocessDescribe()](#preprocess_describe)
+ - [One-hot encoding on Wikipedia](https://en.m.wikipedia.org/wiki/One-hot)
+
+## Radical()
+{: #radical }
+
+#### RADICAL
+{: #radical_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Radical().
+param := mlpack.RadicalOptions()
+param.Angles = 150
+param.NoiseStdDev = 0.175
+param.Objective = false
+param.Replicates = 30
+param.Seed = 0
+param.Sweeps = 0
+param.Verbose = false
+
+output_ic, output_unmixing := mlpack.Radical(input, param)
+```
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Given a dataset, this can decompose the dataset into an unmixing matrix and an independent component matrix; this can be useful for preprocessing. [Detailed documentation](#radical_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Angles` | [`int`](#doc_int) | Number of angles to consider in brute-force search during Radical2D. | `150` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`*mat.Dense`](#doc_a__mat_Dense) | Input dataset for ICA. | `**--**` |
+| `NoiseStdDev` | [`float64`](#doc_float64) | Standard deviation of Gaussian noise. | `0.175` |
+| `Objective` | [`bool`](#doc_bool) | If set, an estimate of the final objective function is printed. | `false` |
+| `Replicates` | [`int`](#doc_int) | Number of Gaussian-perturbed replicates to use (per point) in Radical2D. | `30` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Sweeps` | [`int`](#doc_int) | Number of sweeps; each sweep calls Radical2D once for each pair of dimensions. | `0` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputIc` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save independent components to. | 
+| `OutputUnmixing` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save unmixing matrix to. | 
+
+### Detailed documentation
+{: #radical_detailed-documentation }
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Assuming that we have an input matrix X, the goal is to find a square unmixing matrix W such that Y = W * X and the dimensions of Y are independent components.  If the algorithm is running particularly slowly, try reducing the number of replicates.
+
+The input matrix to perform ICA on should be specified with the `Input` parameter.  The output matrix Y may be saved with the `OutputIc` output parameter, and the output unmixing matrix W may be saved with the `OutputUnmixing` output parameter.
+
+### Example
+For example, to perform ICA on the matrix `X` with 40 replicates, saving the independent components to `ic`, the following command may be used: 
+
+```go
+// Initialize optional parameters for Radical().
+param := mlpack.RadicalOptions()
+param.Replicates = 40
+
+ic, _ := mlpack.Radical(X, param)
+```
+
+### See also
+
+ - [Independent component analysis on Wikipedia](https://en.wikipedia.org/wiki/Independent_component_analysis)
+ - [ICA using spacings estimates of entropy (pdf)](https://www.jmlr.org/papers/volume4/learned-miller03a/learned-miller03a.pdf)
+ - [Radical C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/radical/radical.hpp)
+
+## RandomForest()
+{: #random_forest }
+
+#### Random forests
+{: #random_forest_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for RandomForest().
+param := mlpack.RandomForestOptions()
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.MaximumDepth = 0
+param.MinimumGainSplit = 0
+param.MinimumLeafSize = 1
+param.NumTrees = 10
+param.PrintTrainingAccuracy = false
+param.Seed = 0
+param.SubspaceDim = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.TestLabels = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+param.WarmStart = false
+
+output_model, predictions, probabilities := mlpack.RandomForest(param)
+```
+
+An implementation of the standard random forest algorithm by Leo Breiman for classification.  Given labeled data, a random forest can be trained and saved for future use; or, a pre-trained random forest can be used for classification. [Detailed documentation](#random_forest_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`randomForestModel`](#doc_model) | Pre-trained random forest to use for classification. | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels for training dataset. | `mat.NewDense(1, 1, nil)` |
+| `MaximumDepth` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `MinimumGainSplit` | [`float64`](#doc_float64) | Minimum gain needed to make a split when building a tree. | `0` |
+| `MinimumLeafSize` | [`int`](#doc_int) | Minimum number of points in each leaf node. | `1` |
+| `NumTrees` | [`int`](#doc_int) | Number of trees in the random forest. | `10` |
+| `PrintTrainingAccuracy` | [`bool`](#doc_bool) | If set, then the accuracy of the model on the training set will be predicted (verbose must also be specified). | `false` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `SubspaceDim` | [`int`](#doc_int) | Dimensionality of random subspace to use for each split.  '0' will autoselect the square root of data dimensionality. | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Test dataset to produce predictions for. | `mat.NewDense(1, 1, nil)` |
+| `TestLabels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Test dataset labels, if accuracy calculation is desired. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Training dataset. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `WarmStart` | [`bool`](#doc_bool) | If true and passed along with `training` and `input_model` then trains more trees on top of existing model. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`randomForestModel`](#doc_model) | Model to save trained random forest to. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Predicted classes for each point in the test set. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #random_forest_detailed-documentation }
+
+This program is an implementation of the standard random forest classification algorithm by Leo Breiman.  A random forest can be trained and saved for later use, or a random forest may be loaded and predictions or class probabilities for points may be generated.
+
+The training set and associated labels are specified with the `Training` and `Labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `Labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `OutputModel` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `InputModel`parameter. The `InputModel` parameter may not be specified when the `Training` parameter is specified.  The `MinimumLeafSize` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `NumTrees` controls the number of trees in the random forest.  The `MinimumGainSplit` parameter controls the minimum required gain for a decision tree node to split.  Larger values will force higher-confidence splits.  The `MaximumDepth` parameter specifies the maximum depth of the tree.  The `SubspaceDim` parameter is used to control the number of random dimensions chosen for an individual node's split.  If `PrintTrainingAccuracy` is specified, the calculated accuracy on the training set will be printed.
+
+Test data may be specified with the `Test` parameter, and if performance measures are desired for that test set, labels for the test points may be specified with the `TestLabels` parameter.  Predictions for each test point may be saved via the `Predictions`output parameter.  Class probabilities for each prediction may be saved with the `Probabilities` output parameter.
+
+### Example
+For example, to train a random forest with a minimum leaf size of 20 using 10 trees on the dataset contained in `data`with labels `labels`, saving the output random forest to `rf_model` and printing the training error, one could call
+
+```go
+// Initialize optional parameters for RandomForest().
+param := mlpack.RandomForestOptions()
+param.Training = data
+param.Labels = labels
+param.MinimumLeafSize = 20
+param.NumTrees = 10
+param.PrintTrainingAccuracy = true
+
+rf_model, _, _ := mlpack.RandomForest(param)
+```
+
+Then, to use that model to classify points in `test_set` and print the test error given the labels `test_labels` using that model, while saving the predictions for each point to `predictions`, one could call 
+
+```go
+// Initialize optional parameters for RandomForest().
+param := mlpack.RandomForestOptions()
+param.InputModel = &rf_model
+param.Test = test_set
+param.TestLabels = test_labels
+
+_, predictions, _ := mlpack.RandomForest(param)
+```
+
+### See also
+
+ - [DecisionTree()](#decision_tree)
+ - [HoeffdingTree()](#hoeffding_tree)
+ - [SoftmaxRegression()](#softmax_regression)
+ - [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ - [Random forests (pdf)](https://www.eecis.udel.edu/~shatkay/Course/papers/BreimanRandomForests2001.pdf)
+ - [RandomForest C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/random_forest/random_forest.hpp)
+
+## Krann()
+{: #krann }
+
+#### K-Rank-Approximate-Nearest-Neighbors (kRANN)
+{: #krann_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Krann().
+param := mlpack.KrannOptions()
+param.Alpha = 0.95
+param.FirstLeafExact = false
+param.InputModel = nil
+param.K = 0
+param.LeafSize = 20
+param.Naive = false
+param.Query = mat.NewDense(1, 1, nil)
+param.RandomBasis = false
+param.Reference = mat.NewDense(1, 1, nil)
+param.SampleAtLeaves = false
+param.Seed = 0
+param.SingleMode = false
+param.SingleSampleLimit = 20
+param.Tau = 5
+param.TreeType = "kd"
+param.Verbose = false
+
+distances, neighbors, output_model := mlpack.Krann(param)
+```
+
+An implementation of rank-approximate k-nearest-neighbor search (kRANN)  using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#krann_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Alpha` | [`float64`](#doc_float64) | The desired success probability. | `0.95` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `FirstLeafExact` | [`bool`](#doc_bool) | The flag to trigger sampling only after exactly exploring the first leaf. | `false` |
+| `InputModel` | [`raModel`](#doc_model) | Pre-trained kNN model. | `nil` |
+| `K` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `LeafSize` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `Naive` | [`bool`](#doc_bool) | If true, sampling will be done without using a tree. | `false` |
+| `Query` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing query points (optional). | `mat.NewDense(1, 1, nil)` |
+| `RandomBasis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `Reference` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing the reference dataset. | `mat.NewDense(1, 1, nil)` |
+| `SampleAtLeaves` | [`bool`](#doc_bool) | The flag to trigger sampling at leaves. | `false` |
+| `Seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `SingleMode` | [`bool`](#doc_bool) | If true, single-tree search is used (as opposed to dual-tree search. | `false` |
+| `SingleSampleLimit` | [`int`](#doc_int) | The limit on the maximum number of samples (and hence the largest node you can approximate). | `20` |
+| `Tau` | [`float64`](#doc_float64) | The allowed rank-error in terms of the percentile of the data. | `5` |
+| `TreeType` | [`string`](#doc_string) | Type of tree to use: 'kd', 'ub', 'cover', 'r', 'x', 'r-star', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Distances` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output distances into. | 
+| `Neighbors` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to output neighbors into. | 
+| `OutputModel` | [`raModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #krann_detailed-documentation }
+
+This program will calculate the k rank-approximate-nearest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. You must specify the rank approximation (in %) (and optionally the success probability).
+
+### Example
+For example, the following will return 5 neighbors from the top 0.1% of the data (with probability 0.95) for each point in `input` and store the distances in `distances` and the neighbors in `neighbors.csv`:
+
+```go
+// Initialize optional parameters for Krann().
+param := mlpack.KrannOptions()
+param.Reference = input
+param.K = 5
+param.Tau = 0.1
+
+distances, neighbors, _ := mlpack.Krann(param)
+```
+
+Note that tau must be set such that the number of points in the corresponding percentile of the data is greater than k.  Thus, if we choose tau = 0.1 with a dataset of 1000 points and k = 5, then we are attempting to choose 5 nearest neighbors out of the closest 1 point -- this is invalid and the program will terminate with an error message.
+
+The output matrices are organized such that row i and column j in the neighbors output file corresponds to the index of the point in the reference set which is the i'th nearest neighbor from the point in the query set with index j.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [Knn()](#knn)
+ - [Lsh()](#lsh)
+ - [Rank-approximate nearest neighbor search: Retaining meaning and speed in high dimensions (pdf)](https://papers.nips.cc/paper/3864-rank-approximate-nearest-neighbor-search-retaining-meaning-and-speed-in-high-dimensions.pdf)
+ - [RASearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/rann/ra_search.hpp)
+
+## SoftmaxRegression()
+{: #softmax_regression }
+
+#### Softmax Regression
+{: #softmax_regression_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for SoftmaxRegression().
+param := mlpack.SoftmaxRegressionOptions()
+param.InputModel = nil
+param.Labels = mat.NewDense(1, 1, nil)
+param.Lambda = 0.0001
+param.MaxIterations = 400
+param.NoIntercept = false
+param.NumberOfClasses = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.TestLabels = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, predictions, probabilities := mlpack.SoftmaxRegression(param)
+```
+
+An implementation of softmax regression for classification, which is a multiclass generalization of logistic regression.  Given labeled data, a softmax regression model can be trained and saved for future use, or, a pre-trained softmax regression model can be used for classification of new points. [Detailed documentation](#softmax_regression_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`softmaxRegression`](#doc_model) | File containing existing model (parameters). | `nil` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | A matrix containing labels (0 or 1) for the points in the training set (y). The labels must order as a row. | `mat.NewDense(1, 1, nil)` |
+| `Lambda` | [`float64`](#doc_float64) | L2-regularization constant | `0.0001` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations before termination. | `400` |
+| `NoIntercept` | [`bool`](#doc_bool) | Do not add the intercept term to the model. | `false` |
+| `NumberOfClasses` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing test dataset. | `mat.NewDense(1, 1, nil)` |
+| `TestLabels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Matrix containing test labels. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | A matrix containing the training set (the matrix of predictors, X). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`softmaxRegression`](#doc_model) | File to save trained softmax regression model to. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Matrix to save predictions for test dataset into. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save class probabilities for test dataset into. | 
+
+### Detailed documentation
+{: #softmax_regression_detailed-documentation }
+
+This program performs softmax regression, a generalization of logistic regression to the multiclass case, and has support for L2 regularization.  The program is able to train a model, load  an existing model, and give predictions (and optionally their accuracy) for test data.
+
+Training a softmax regression model is done by giving a file of training points with the `Training` parameter and their corresponding labels with the `Labels` parameter. The number of classes can be manually specified with the `NumberOfClasses` parameter, and the maximum number of iterations of the L-BFGS optimizer can be specified with the `MaxIterations` parameter.  The L2 regularization constant can be specified with the `Lambda` parameter and if an intercept term is not desired in the model, the `NoIntercept` parameter can be specified.
+
+The trained model can be saved with the `OutputModel` output parameter. If training is not desired, but only testing is, a model can be loaded with the `InputModel` parameter.  At the current time, a loaded model cannot be trained further, so specifying both `InputModel` and `Training` is not allowed.
+
+The program is also able to evaluate a model on test data.  A test dataset can be specified with the `Test` parameter. Class predictions can be saved with the `Predictions` output parameter.  If labels are specified for the test data with the `TestLabels` parameter, then the program will print the accuracy of the predictions on the given test set and its corresponding labels.
+
+### Example
+For example, to train a softmax regression model on the data `dataset` with labels `labels` with a maximum of 1000 iterations for training, saving the trained model to `sr_model`, the following command can be used: 
+
+```go
+// Initialize optional parameters for SoftmaxRegression().
+param := mlpack.SoftmaxRegressionOptions()
+param.Training = dataset
+param.Labels = labels
+
+sr_model, _, _ := mlpack.SoftmaxRegression(param)
+```
+
+Then, to use `sr_model` to classify the test points in `test_points`, saving the output predictions to `predictions`, the following command can be used:
+
+```go
+// Initialize optional parameters for SoftmaxRegression().
+param := mlpack.SoftmaxRegressionOptions()
+param.InputModel = &sr_model
+param.Test = test_points
+
+_, predictions, _ := mlpack.SoftmaxRegression(param)
+```
+
+### See also
+
+ - [LogisticRegression()](#logistic_regression)
+ - [RandomForest()](#random_forest)
+ - [Multinomial logistic regression (softmax regression) on Wikipedia](https://en.wikipedia.org/wiki/Multinomial_logistic_regression)
+ - [SoftmaxRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/softmax_regression/softmax_regression.hpp)
+
+## SparseCoding()
+{: #sparse_coding }
+
+#### Sparse Coding
+{: #sparse_coding_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for SparseCoding().
+param := mlpack.SparseCodingOptions()
+param.Atoms = 15
+param.InitialDictionary = mat.NewDense(1, 1, nil)
+param.InputModel = nil
+param.Lambda1 = 0
+param.Lambda2 = 0
+param.MaxIterations = 0
+param.NewtonTolerance = 1e-06
+param.Normalize = false
+param.ObjectiveTolerance = 0.01
+param.Seed = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+codes, dictionary, output_model := mlpack.SparseCoding(param)
+```
+
+An implementation of Sparse Coding with Dictionary Learning.  Given a dataset, this will decompose the dataset into a sparse combination of a few dictionary elements, where the dictionary is learned during computation; a dictionary can be reused for future sparse coding of new points. [Detailed documentation](#sparse_coding_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Atoms` | [`int`](#doc_int) | Number of atoms in the dictionary. | `15` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InitialDictionary` | [`*mat.Dense`](#doc_a__mat_Dense) | Optional initial dictionary matrix. | `mat.NewDense(1, 1, nil)` |
+| `InputModel` | [`sparseCoding`](#doc_model) | File containing input sparse coding model. | `nil` |
+| `Lambda1` | [`float64`](#doc_float64) | Sparse coding l1-norm regularization parameter. | `0` |
+| `Lambda2` | [`float64`](#doc_float64) | Sparse coding l2-norm regularization parameter. | `0` |
+| `MaxIterations` | [`int`](#doc_int) | Maximum number of iterations for sparse coding (0 indicates no limit). | `0` |
+| `NewtonTolerance` | [`float64`](#doc_float64) | Tolerance for convergence of Newton method. | `1e-06` |
+| `Normalize` | [`bool`](#doc_bool) | If set, the input data matrix will be normalized before coding. | `false` |
+| `ObjectiveTolerance` | [`float64`](#doc_float64) | Tolerance for convergence of the objective function. | `0.01` |
+| `Seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Optional matrix to be encoded by trained model. | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix of training data (X). | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Codes` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save the output sparse codes of the test matrix (--test_file) to. | 
+| `Dictionary` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save the output dictionary to. | 
+| `OutputModel` | [`sparseCoding`](#doc_model) | File to save trained sparse coding model to. | 
+
+### Detailed documentation
+{: #sparse_coding_detailed-documentation }
+
+An implementation of Sparse Coding with Dictionary Learning, which achieves sparsity via an l1-norm regularizer on the codes (LASSO) or an (l1+l2)-norm regularizer on the codes (the Elastic Net).  Given a dense data matrix X with d dimensions and n points, sparse coding seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a sparse coding matrix Z with n points in k dimensions.
+
+The original data matrix X can then be reconstructed as Z * D.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The sparse coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a sparse coding step, which updates the sparse coding matrix.
+
+Once a dictionary D is found, the sparse coding model may be used to encode other matrices, and saved for future usage.
+
+To run this program, either an input matrix or an already-saved sparse coding model must be specified.  An input matrix may be specified with the `Training` option, along with the number of atoms in the dictionary (specified with the `Atoms` parameter).  It is also possible to specify an initial dictionary for the optimization, with the `InitialDictionary` parameter.  An input model may be specified with the `InputModel` parameter.
+
+### Example
+As an example, to build a sparse coding model on the dataset `data` using 200 atoms and an l1-regularization parameter of 0.1, saving the model into `model`, use 
+
+```go
+// Initialize optional parameters for SparseCoding().
+param := mlpack.SparseCodingOptions()
+param.Training = data
+param.Atoms = 200
+param.Lambda1 = 0.1
+
+_, _, model := mlpack.SparseCoding(param)
+```
+
+Then, this model could be used to encode a new matrix, `otherdata`, and save the output codes to `codes`: 
+
+```go
+// Initialize optional parameters for SparseCoding().
+param := mlpack.SparseCodingOptions()
+param.InputModel = &model
+param.Test = otherdata
+
+codes, _, _ := mlpack.SparseCoding(param)
+```
+
+### See also
+
+ - [LocalCoordinateCoding()](#local_coordinate_coding)
+ - [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ - [Efficient sparse coding algorithms (pdf)](http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf)
+ - [Regularization and variable selection via the elastic net](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=46217f372a75dddc2254fdbc6b9418ba3554e453)
+ - [SparseCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/sparse_coding/sparse_coding.hpp)
+
+## Adaboost()
+{: #adaboost }
+
+#### AdaBoost
+{: #adaboost_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for Adaboost().
+param := mlpack.AdaboostOptions()
+param.InputModel = nil
+param.Iterations = 1000
+param.Labels = mat.NewDense(1, 1, nil)
+param.Test = mat.NewDense(1, 1, nil)
+param.Tolerance = 1e-10
+param.Training = mat.NewDense(1, 1, nil)
+param.Verbose = false
+param.WeakLearner = "decision_stump"
+
+output, output_model, predictions, probabilities := mlpack.Adaboost(param)
+```
+
+An implementation of the AdaBoost.MH (Adaptive Boosting) algorithm for classification.  This can be used to train an AdaBoost model on labeled data or use an existing AdaBoost model to predict the classes of new points. [Detailed documentation](#adaboost_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`adaBoostModel`](#doc_model) | Input AdaBoost model. | `nil` |
+| `Iterations` | [`int`](#doc_int) | The maximum number of boosting iterations to be run (0 will run until convergence.) | `1000` |
+| `Labels` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Labels for the training set. | `mat.NewDense(1, 1, nil)` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Test dataset. | `mat.NewDense(1, 1, nil)` |
+| `Tolerance` | [`float64`](#doc_float64) | The tolerance for change in values of the weighted error during training. | `1e-10` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Dataset for training AdaBoost. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `WeakLearner` | [`string`](#doc_string) | The type of weak learner to use: 'decision_stump', or 'perceptron'. | `"decision_stump"` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Predicted labels for the test set. | 
+| `OutputModel` | [`adaBoostModel`](#doc_model) | Output trained AdaBoost model. | 
+| `Predictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Predicted labels for the test set. | 
+| `Probabilities` | [`*mat.Dense`](#doc_a__mat_Dense) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #adaboost_detailed-documentation }
+
+This program implements the AdaBoost (or Adaptive Boosting) algorithm. The variant of AdaBoost implemented here is AdaBoost.MH. It uses a weak learner, either decision stumps or perceptrons, and over many iterations, creates a strong learner that is a weighted ensemble of weak learners. It runs these iterations until a tolerance value is crossed for change in the value of the weighted training error.
+
+For more information about the algorithm, see the paper "Improved Boosting Algorithms Using Confidence-Rated Predictions", by R.E. Schapire and Y. Singer.
+
+This program allows training of an AdaBoost model, and then application of that model to a test dataset.  To train a model, a dataset must be passed with the `Training` option.  Labels can be given with the `Labels` option; if no labels are specified, the labels will be assumed to be the last column of the input dataset.  Alternately, an AdaBoost model may be loaded with the `InputModel` option.
+
+Once a model is trained or loaded, it may be used to provide class predictions for a given test dataset.  A test dataset may be specified with the `Test` parameter.  The predicted classes for each point in the test dataset are output to the `Predictions` output parameter.  The AdaBoost model itself is output to the `OutputModel` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `Output`.
+Use `Predictions` instead of `Output`.
+
+### Example
+For example, to run AdaBoost on an input dataset `data` with labels `labels`and perceptrons as the weak learner type, storing the trained model in `model`, one could use the following command: 
+
+```go
+// Initialize optional parameters for Adaboost().
+param := mlpack.AdaboostOptions()
+param.Training = data
+param.Labels = labels
+param.WeakLearner = "perceptron"
+
+_, model, _, _ := mlpack.Adaboost(param)
+```
+
+Similarly, an already-trained model in `model` can be used to provide class predictions from test data `test_data` and store the output in `predictions` with the following command: 
+
+```go
+// Initialize optional parameters for Adaboost().
+param := mlpack.AdaboostOptions()
+param.InputModel = &model
+param.Test = test_data
+
+_, _, predictions, _ := mlpack.Adaboost(param)
+```
+
+### See also
+
+ - [AdaBoost on Wikipedia](https://en.wikipedia.org/wiki/AdaBoost)
+ - [Improved boosting algorithms using confidence-rated predictions (pdf)](http://rob.schapire.net/papers/SchapireSi98.pdf)
+ - [Perceptron](#perceptron)
+ - [Decision Trees](#decision_tree)
+ - [AdaBoost C++ class documentation](../../user/methods/adaboost.md)
+
+## LinearRegression()
+{: #linear_regression }
+
+#### Simple Linear Regression and Prediction
+{: #linear_regression_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for LinearRegression().
+param := mlpack.LinearRegressionOptions()
+param.InputModel = nil
+param.Lambda = 0
+param.Test = mat.NewDense(1, 1, nil)
+param.Training = mat.NewDense(1, 1, nil)
+param.TrainingResponses = mat.NewDense(1, 1, nil)
+param.Verbose = false
+
+output_model, output_predictions := mlpack.LinearRegression(param)
+```
+
+An implementation of simple linear regression and ridge regression using ordinary least squares.  Given a dataset and responses, a model can be trained and saved for later use, or a pre-trained model can be used to output regression predictions for a test set. [Detailed documentation](#linear_regression_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `InputModel` | [`linearRegression`](#doc_model) | Existing LinearRegression model to use. | `nil` |
+| `Lambda` | [`float64`](#doc_float64) | Tikhonov regularization for ridge regression.  If 0, the method reduces to linear regression. | `0` |
+| `Test` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing X' (test regressors). | `mat.NewDense(1, 1, nil)` |
+| `Training` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix containing training set X (regressors). | `mat.NewDense(1, 1, nil)` |
+| `TrainingResponses` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | Optional vector containing y (responses). If not given, the responses are assumed to be the last row of the input file. | `mat.NewDense(1, 1, nil)` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `OutputModel` | [`linearRegression`](#doc_model) | Output LinearRegression model. | 
+| `OutputPredictions` | [`*mat.Dense (1d)`](#doc_a__mat_Dense__1d_) | If --test_file is specified, this matrix is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #linear_regression_detailed-documentation }
+
+An implementation of simple linear regression and simple ridge regression using ordinary least squares. This solves the problem
+
+  y = X * b + e
+
+where X (specified by `Training`) and y (specified either as the last column of the input matrix `Training` or via the `TrainingResponses` parameter) are known and b is the desired variable.  If the covariance matrix (X'X) is not invertible, or if the solution is overdetermined, then specify a Tikhonov regularization constant (with `Lambda`) greater than 0, which will regularize the covariance matrix to make it invertible.  The calculated b may be saved with the `OutputPredictions` output parameter.
+
+Optionally, the calculated value of b is used to predict the responses for another matrix X' (specified by the `Test` parameter):
+
+   y' = X' * b
+
+and the predicted responses y' may be saved with the `OutputPredictions` output parameter.  This type of regression is related to least-angle regression, which mlpack implements as the 'lars' program.
+
+### Example
+For example, to run a linear regression on the dataset `X` with responses `y`, saving the trained model to `lr_model`, the following command could be used:
+
+```go
+// Initialize optional parameters for LinearRegression().
+param := mlpack.LinearRegressionOptions()
+param.Training = X
+param.TrainingResponses = y
+
+lr_model, _ := mlpack.LinearRegression(param)
+```
+
+Then, to use `lr_model` to predict responses for a test set `X_test`, saving the predictions to `X_test_responses`, the following command could be used:
+
+```go
+// Initialize optional parameters for LinearRegression().
+param := mlpack.LinearRegressionOptions()
+param.InputModel = &lr_model
+param.Test = X_test
+
+_, X_test_responses := mlpack.LinearRegression(param)
+```
+
+### See also
+
+ - [Lars()](#lars)
+ - [Linear regression on Wikipedia](https://en.wikipedia.org/wiki/Linear_regression)
+ - [LinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_regression/linear_regression.hpp)
+
+## ImageConverter()
+{: #image_converter }
+
+#### Image Converter
+{: #image_converter_descr }
+
+```go
+import (
+  "mlpack.org/v1/mlpack"
+  "gonum.org/v1/gonum/mat"
+)
+
+// Initialize optional parameters for ImageConverter().
+param := mlpack.ImageConverterOptions()
+param.Channels = 0
+param.Dataset = mat.NewDense(1, 1, nil)
+param.Height = 0
+param.Quality = 90
+param.Save = false
+param.Verbose = false
+param.Width = 0
+
+output := mlpack.ImageConverter(input, param)
+```
+
+A utility to load an image or set of images into a single dataset that can then be used by other mlpack methods and utilities. This can also unpack an image dataset into individual files, for instance after mlpack methods have been used. [Detailed documentation](#image_converter_detailed-documentation).
+
+
+
+### Input options
+There are two types of input options: required options, which are passed directly to the function call, and optional options, which are passed via an initialized struct, which allows keyword access to each of the options.
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `Channels` | [`int`](#doc_int) | Number of channels in the image. | `0` |
+| `CheckInputMatrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `Dataset` | [`*mat.Dense`](#doc_a__mat_Dense) | Input matrix to save as images. | `mat.NewDense(1, 1, nil)` |
+| `Height` | [`int`](#doc_int) | Height of the images. | `0` |
+| `input` | [`array of strings`](#doc_array_of_strings) | Image filenames which have to be loaded/saved. | `**--**` |
+| `Quality` | [`int`](#doc_int) | Compression of the image if saved as jpg (0-100). | `90` |
+| `Save` | [`bool`](#doc_bool) | Save a dataset as images. | `false` |
+| `Verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `Width` | [`int`](#doc_int) | Width of the image. | `0` |
+
+### Output options
+
+Output options are returned via Go's support for multiple return values, in the order listed below.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `Output` | [`*mat.Dense`](#doc_a__mat_Dense) | Matrix to save images data to, Onlyneeded if you are specifying 'save' option. | 
+
+### Detailed documentation
+{: #image_converter_detailed-documentation }
+
+This utility takes an image or an array of images and loads them to a matrix. You can optionally specify the height `Height` width `Width` and channel `Channels` of the images that needs to be loaded; otherwise, these parameters will be automatically detected from the image.
+There are other options too, that can be specified such as `Quality`.
+
+You can also provide a dataset and save them as images using `Dataset` and `Save` as an parameter.
+
+### Example
+ An example to load an image : 
+
+```go
+// Initialize optional parameters for ImageConverter().
+param := mlpack.ImageConverterOptions()
+param.Height = 256
+param.Width = 256
+param.Channels = 3
+
+Y := mlpack.ImageConverter(X, param)
+```
+
+ An example to save an image is :
+
+```go
+// Initialize optional parameters for ImageConverter().
+param := mlpack.ImageConverterOptions()
+param.Height = 256
+param.Width = 256
+param.Channels = 3
+param.Dataset = Y
+param.Save = true
+
+_ := mlpack.ImageConverter(X, param)
+```
+
+### See also
+
+ - [PreprocessBinarize()](#preprocess_binarize)
+ - [PreprocessDescribe()](#preprocess_describe)
+
diff -pruN 4.3.0-2/doc/user/bindings/go.sidebar.html 4.4.0-1/doc/user/bindings/go.sidebar.html
--- 4.3.0-2/doc/user/bindings/go.sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/go.sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,100 @@
+<ul>
+<li class="page_title">
+<b>mlpack Go binding documentation</b>
+<a href="#">[top]</a>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#mlpack-overview">Overview</a></li>
+<li><a href="LINKROOTquickstart/go.html">Go Quickstart</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#data-formats">Data Formats</a></li>
+<li><details><summary>Classification</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#decision_tree">DecisionTree()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#linear_svm">LinearSvm()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#logistic_regression">LogisticRegression()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#nbc">Nbc()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#perceptron">Perceptron()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#random_forest">RandomForest()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#softmax_regression">SoftmaxRegression()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#adaboost">Adaboost()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Regression</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#bayesian_linear_regression">BayesianLinearRegression()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#lars">Lars()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#linear_regression">LinearRegression()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Clustering</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#dbscan">Dbscan()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#gmm_train">GmmTrain()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#gmm_generate">GmmGenerate()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#gmm_probability">GmmProbability()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#hoeffding_tree">HoeffdingTree()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#kmeans">Kmeans()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#mean_shift">MeanShift()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Geometry</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#approx_kfn">ApproxKfn()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#emst">Emst()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#fastmks">Fastmks()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#lsh">Lsh()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#knn">Knn()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#kfn">Kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#krann">Krann()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Preprocessing</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#preprocess_split">PreprocessSplit()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#preprocess_binarize">PreprocessBinarize()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#preprocess_describe">PreprocessDescribe()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#preprocess_scale">PreprocessScale()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#preprocess_one_hot_encoding">PreprocessOneHotEncoding()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#image_converter">ImageConverter()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Misc. / Other</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#cf">Cf()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#det">Det()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#hmm_train">HmmTrain()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#hmm_generate">HmmGenerate()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#hmm_loglik">HmmLoglik()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#hmm_viterbi">HmmViterbi()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#kde">Kde()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#nmf">Nmf()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Transformations</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/go.html#kernel_pca">KernelPca()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#lmnn">Lmnn()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#local_coordinate_coding">LocalCoordinateCoding()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#nca">Nca()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#pca">Pca()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#radical">Radical()</a></li>
+<li><a href="LINKROOTuser/bindings/go.html#sparse_coding">SparseCoding()</a></li>
+</ul>
+</details>
+</li>
+
+</ul>
+</li>
+</ul>
+</div>
diff -pruN 4.3.0-2/doc/user/bindings/julia.md 4.4.0-1/doc/user/bindings/julia.md
--- 4.3.0-2/doc/user/bindings/julia.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/julia.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,3631 @@
+# mlpack Julia binding documentation
+
+## mlpack overview
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning library with bindings to other languages.  It aims to provide fast, lightweight implementations of both common and cutting-edge machine learning algorithms.
+
+This reference page details mlpack's bindings to Julia.
+
+Further useful mlpack documentation links are given below.
+
+ - [mlpack homepage](https://www.mlpack.org/)
+ - [mlpack on Github](https://github.com/mlpack/mlpack)
+ - [mlpack main documentation page](https://www.mlpack.org/docs.html)
+
+See also the quickstart guide for Julia:
+
+ - [Julia Quickstart](../../quickstart/julia.md)
+
+## Data Formats
+
+<div id="data-formats-div" markdown="1">
+mlpack bindings for Julia take and return a restricted set of types, for simplicity.  These include primitive types, matrix/vector types, categorical matrix types, and model types. Each type is detailed below.
+
+ - `Int`{: #doc_Int}: An integer (i.e., `1`).
+ - `Float64`{: #doc_Float64 }: A floating-point number (i.e., `0.5`).
+ - `Bool`{: #doc_Bool }: A boolean flag option (`true` or `false`).
+ - `String`{: #doc_String }: A character string (i.e., `"hello"`).
+ - `Array{Int, 1}`{: #doc_Array_Int__1_ }: A vector of integers; i.e., `[0, 1, 2]`.
+ - `Array{String, 1}`{: #doc_Array_String__1_ }: A vector of strings; i.e., `["hello", "goodbye"]`.
+ - `Float64 matrix-like`{: #doc_Float64_matrix_like }: A 2-d matrix-like containing `Float64` data (could be an `Array{Float64, 2}` or a `DataFrame` or anything convertible to an `Array{Float64, 2}`).  It is expected that each row of the matrix corresponds to a data point, unless `points_are_rows` is set to `false` when calling mlpack bindings.
+ - `Int matrix-like`{: #doc_Int_matrix_like }: A 2-d matrix-like containing `Int` data (elements should be greater than or equal to 0).  Could be an `Array{Int, 2}` or a `DataFrame` or anything convertible to an `Array{Int, 2}`.  It is expected that each row of the matrix corresponds to a data point, unless `points_are_rows` is set to `false` when calling mlpack bindings.
+ - `Float64 vector-like`{: #doc_Float64_vector_like }: A 1-d vector-like containing `Float64` data (could be an `Array{Float64, 1}`, an `Array{Float64, 2}` with one dimension of size 1, or anything convertible to `Array{Float64, 1}`.
+ - `Int vector-like`{: #doc_Int_vector_like }: A 1-d vector-like containing `Int` data (elements should be greater than or equal to 0).  Could be an `Array{Int, 1}`, an `Array{Int, 2}` with one dimension of size 1, or anything convertible to `Array{Int, 1}`.
+ - `Tuple{Array{Bool, 1}, Array{Float64, 2}}`{: #doc_Tuple_Array_Bool__1___Array_Float64__2__ }: A 2-d array containing `Float64` data along with a boolean array indicating which dimensions are categorical (represented by `true`) and which are numeric (represented by `false`).  The number of elements in the boolean array should be the same as the dimensionality of the data matrix.  Categorical dimensions should take integer values between 1 and the number of categories.  It is expected that each row of the matrix corresponds to a single data point, unless `points_are_rows` is set to `false` when calling mlpack bindings.
+ - `<Model> (mlpack model)`{: #doc_model }: An mlpack model pointer.  `<Model>` refers to the type of model that is being stored, so, e.g., for `CF()`, the type will be `CFModel`. This type holds a pointer to C++ memory containing the mlpack model.  Note that this means the mlpack model itself cannot be easily inspected in Julia.  However, the pointer can be passed to subsequent calls to mlpack functions, and can be serialized and deserialized via either the `Serialization` package, or the `mlpack.serialize_bin()` and `mlpack.deserialize_bin()` functions.
+</div>
+
+
+## approx_kfn()
+{: #approx_kfn }
+
+#### Approximate furthest neighbor search
+{: #approx_kfn_descr }
+
+```julia
+julia> using mlpack: approx_kfn
+julia> distances, neighbors, output_model = approx_kfn( ;
+          algorithm="ds", calculate_error=false, exact_distances=zeros(0, 0),
+          input_model=nothing, k=0, num_projections=5, num_tables=5,
+          query=zeros(0, 0), reference=zeros(0, 0), verbose=false)
+```
+
+An implementation of two strategies for furthest neighbor search.  This can be used to compute the furthest neighbor of query point(s) from a set of points; furthest neighbor models can be saved and reused with future query point(s). [Detailed documentation](#approx_kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`String`](#doc_String) | Algorithm to use: 'ds' or 'qdafn'. | `"ds"` |
+| `calculate_error` | [`Bool`](#doc_Bool) | If set, calculate the average distance error for the first furthest neighbor only. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `exact_distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing exact distances to furthest neighbors; this can be used to avoid explicit calculation when --calculate_error is set. | `zeros(0, 0)` |
+| `input_model` | [`ApproxKFNModel`](#doc_model) | File containing input model. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of furthest neighbors to search for. | `0` |
+| `num_projections` | [`Int`](#doc_Int) | Number of projections to use in each hash table. | `5` |
+| `num_tables` | [`Int`](#doc_Int) | Number of hash tables to use. | `5` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing query points. | `zeros(0, 0)` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing the reference dataset. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save furthest neighbor distances to. | 
+| `neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to save neighbor indices to. | 
+| `output_model` | [`ApproxKFNModel`](#doc_model) | File to save output model to. | 
+
+### Detailed documentation
+{: #approx_kfn_detailed-documentation }
+
+This program implements two strategies for furthest neighbor search. These strategies are:
+
+ - The 'qdafn' algorithm from "Approximate Furthest Neighbor in High Dimensions" by R. Pagh, F. Silvestri, J. Sivertsen, and M. Skala, in Similarity Search and Applications 2015 (SISAP).
+ - The 'DrusillaSelect' algorithm from "Fast approximate furthest neighbors with data-dependent candidate selection", by R.R. Curtin and A.B. Gardner, in Similarity Search and Applications 2016 (SISAP).
+
+These two strategies give approximate results for the furthest neighbor search problem and can be used as fast replacements for other furthest neighbor techniques such as those found in the mlpack_kfn program.  Note that typically, the 'ds' algorithm requires far fewer tables and projections than the 'qdafn' algorithm.
+
+Specify a reference set (set to search in) with `reference`, specify a query set with `query`, and specify algorithm parameters with `num_tables` and `num_projections` (or don't and defaults will be used).  The algorithm to be used (either 'ds'---the default---or 'qdafn')  may be specified with `algorithm`.  Also specify the number of neighbors to search for with `k`.
+
+Note that for 'qdafn' in lower dimensions, `num_projections` may need to be set to a high value in order to return results for each query point.
+
+If no query set is specified, the reference set will be used as the query set.  The `output_model` output parameter may be used to store the built model, and an input model may be loaded instead of specifying a reference set with the `input_model` option.
+
+Results for each query point can be stored with the `neighbors` and `distances` output parameters.  Each row of these output matrices holds the k distances or neighbor indices for each query point.
+
+### Example
+For example, to find the 5 approximate furthest neighbors with ``reference_set`` as the reference set and ``query_set`` as the query set using DrusillaSelect, storing the furthest neighbor indices to ``neighbors`` and the furthest neighbor distances to ``distances``, one could call
+
+```julia
+julia> using CSV
+julia> query_set = CSV.read("query_set.csv")
+julia> reference_set = CSV.read("reference_set.csv")
+julia> distances, neighbors, _ = approx_kfn(algorithm="ds", k=5,
+            query=query_set, reference=reference_set)
+```
+
+and to perform approximate all-furthest-neighbors search with k=1 on the set ``data`` storing only the furthest neighbor distances to ``distances``, one could call
+
+```julia
+julia> using CSV
+julia> reference_set = CSV.read("reference_set.csv")
+julia> distances, _, _ = approx_kfn(k=1, reference=reference_set)
+```
+
+A trained model can be re-used.  If a model has been previously saved to ``model``, then we may find 3 approximate furthest neighbors on a query set ``new_query_set`` using that model and store the furthest neighbor indices into ``neighbors`` by calling
+
+```julia
+julia> using CSV
+julia> new_query_set = CSV.read("new_query_set.csv")
+julia> _, neighbors, _ = approx_kfn(input_model=model, k=3,
+            query=new_query_set)
+```
+
+### See also
+
+ - [k-furthest-neighbor search](#kfn)
+ - [k-nearest-neighbor search](#knn)
+ - [Fast approximate furthest neighbors with data-dependent candidate selection (pdf)](http://ratml.org/pub/pdf/2016fast.pdf)
+ - [Approximate furthest neighbor in high dimensions (pdf)](https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf)
+ - [QDAFN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/qdafn.hpp)
+ - [DrusillaSelect class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/drusilla_select.hpp)
+
+## bayesian_linear_regression()
+{: #bayesian_linear_regression }
+
+#### BayesianLinearRegression
+{: #bayesian_linear_regression_descr }
+
+```julia
+julia> using mlpack: bayesian_linear_regression
+julia> output_model, predictions, stds = bayesian_linear_regression( ;
+          center=false, input=zeros(0, 0), input_model=nothing,
+          responses=Float64[], scale=false, test=zeros(0, 0), verbose=false)
+```
+
+An implementation of the bayesian linear regression. [Detailed documentation](#bayesian_linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `center` | [`Bool`](#doc_Bool) | Center the data and fit the intercept if enabled. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of covariates (X). | `zeros(0, 0)` |
+| `input_model` | [`BayesianLinearRegression`](#doc_model) | Trained BayesianLinearRegression model to use. | `nothing` |
+| `responses` | [`Float64 vector-like`](#doc_Float64_vector_like) | Matrix of responses/observations (y). | `Float64[]` |
+| `scale` | [`Bool`](#doc_Bool) | Scale each feature by their standard deviations if enabled. | `false` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing points to regress on (test points). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`BayesianLinearRegression`](#doc_model) | Output BayesianLinearRegression model. | 
+| `predictions` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+| `stds` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If specified, this is where the standard deviations of the predictive distribution will be saved. | 
+
+### Detailed documentation
+{: #bayesian_linear_regression_detailed-documentation }
+
+An implementation of the bayesian linear regression.
+This model is a probabilistic view and implementation of the linear regression. The final solution is obtained by computing a posterior distribution from gaussian likelihood and a zero mean gaussian isotropic  prior distribution on the solution. 
+Optimization is AUTOMATIC and does not require cross validation. The optimization is performed by maximization of the evidence function. Parameters are tuned during the maximization of the marginal likelihood. This procedure includes the Ockham's razor that penalizes over complex solutions. 
+
+This program is able to train a Bayesian linear regression model or load a model from file, output regression predictions for a test set, and save the trained model to a file.
+
+To train a BayesianLinearRegression model, the `input` and `responses`parameters must be given. The `center`and `scale` parameters control the centering and the normalizing options. A trained model can be saved with the `output_model`. If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `predictions` output parameter. The corresponding standard deviation can be save by precising the `stds` parameter.
+
+### Example
+For example, the following command trains a model on the data ``data`` and responses ``responses``with center set to true and scale set to false (so, Bayesian linear regression is being solved, and then the model is saved to ``blr_model``:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> responses = CSV.read("responses.csv")
+julia> blr_model, _, _ = bayesian_linear_regression(center=1,
+            input=data, responses=responses, scale=0)
+```
+
+The following command uses the ``blr_model`` to provide predicted  responses for the data ``test`` and save those  responses to ``test_predictions``: 
+
+```julia
+julia> using CSV
+julia> test = CSV.read("test.csv")
+julia> _, test_predictions, _ =
+            bayesian_linear_regression(input_model=blr_model, test=test)
+```
+
+Because the estimator computes a predictive distribution instead of a simple point estimate, the `stds` parameter allows one to save the prediction uncertainties: 
+
+```julia
+julia> using CSV
+julia> test = CSV.read("test.csv")
+julia> _, test_predictions, stds =
+            bayesian_linear_regression(input_model=blr_model, test=test)
+```
+
+### See also
+
+ - [Bayesian Interpolation](https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf)
+ - [Bayesian Linear Regression, Section 3.3](https://www.microsoft.com/en-us/research/uploads/prod/2006/01/Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf)
+ - [BayesianLinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp)
+
+## cf()
+{: #cf }
+
+#### Collaborative Filtering
+{: #cf_descr }
+
+```julia
+julia> using mlpack: cf
+julia> output, output_model = cf( ;
+                                 algorithm="NMF",
+                                 all_user_recommendations=false,
+                                 input_model=nothing, interpolation="average",
+                                 iteration_only_termination=false,
+                                 max_iterations=1000, min_residue=1e-05,
+                                 neighbor_search="euclidean", neighborhood=5,
+                                 normalization="none", query=zeros(Int, 0, 0),
+                                 rank=0, recommendations=5, seed=0,
+                                 test=zeros(0, 0), training=zeros(0, 0),
+                                 verbose=false)
+```
+
+An implementation of several collaborative filtering (CF) techniques for recommender systems.  This can be used to train a new CF model, or use an existing CF model to compute recommendations. [Detailed documentation](#cf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`String`](#doc_String) | Algorithm used for matrix factorization. | `"NMF"` |
+| `all_user_recommendations` | [`Bool`](#doc_Bool) | Generate recommendations for all users. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`CFModel`](#doc_model) | Trained CF model to load. | `nothing` |
+| `interpolation` | [`String`](#doc_String) | Algorithm used for weight interpolation. | `"average"` |
+| `iteration_only_termination` | [`Bool`](#doc_Bool) | Terminate only when the maximum number of iterations is reached. | `false` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations. If set to zero, there is no limit on the number of iterations. | `1000` |
+| `min_residue` | [`Float64`](#doc_Float64) | Residue required to terminate the factorization (lower values generally mean better fits). | `1e-05` |
+| `neighbor_search` | [`String`](#doc_String) | Algorithm used for neighbor search. | `"euclidean"` |
+| `neighborhood` | [`Int`](#doc_Int) | Size of the neighborhood of similar users to consider for each query user. | `5` |
+| `normalization` | [`String`](#doc_String) | Normalization performed on the ratings. | `"none"` |
+| `query` | [`Int matrix-like`](#doc_Int_matrix_like) | List of query users for which recommendations should be generated. | `zeros(Int, 0, 0)` |
+| `rank` | [`Int`](#doc_Int) | Rank of decomposed matrices (if 0, a heuristic is used to estimate the rank). | `0` |
+| `recommendations` | [`Int`](#doc_Int) | Number of recommendations to generate for each query user. | `5` |
+| `seed` | [`Int`](#doc_Int) | Set the random seed (0 uses std::time(NULL)). | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Test set to calculate RMSE on. | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform CF on. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix that will store output recommendations. | 
+| `output_model` | [`CFModel`](#doc_model) | Output for trained CF model. | 
+
+### Detailed documentation
+{: #cf_detailed-documentation }
+
+This program performs collaborative filtering (CF) on the given dataset. Given a list of user, item and preferences (the `training` parameter), the program will perform a matrix decomposition and then can perform a series of actions related to collaborative filtering.  Alternately, the program can load an existing saved CF model with the `input_model` parameter and then use that model to provide recommendations or predict values.
+
+The input matrix should be a 3-dimensional matrix of ratings, where the first dimension is the user, the second dimension is the item, and the third dimension is that user's rating of that item.  Both the users and items should be numeric indices, not names. The indices are assumed to start from 0.
+
+A set of query users for which recommendations can be generated may be specified with the `query` parameter; alternately, recommendations may be generated for every user in the dataset by specifying the `all_user_recommendations` parameter.  In addition, the number of recommendations per user to generate can be specified with the `recommendations` parameter, and the number of similar users (the size of the neighborhood) to be considered when generating recommendations can be specified with the `neighborhood` parameter.
+
+For performing the matrix decomposition, the following optimization algorithms can be specified via the `algorithm` parameter:
+
+ - 'RegSVD' -- Regularized SVD using a SGD optimizer
+ - 'NMF' -- Non-negative matrix factorization with alternating least squares update rules
+ - 'BatchSVD' -- SVD batch learning
+ - 'SVDIncompleteIncremental' -- SVD incomplete incremental learning
+ - 'SVDCompleteIncremental' -- SVD complete incremental learning
+ - 'BiasSVD' -- Bias SVD using a SGD optimizer
+ - 'SVDPP' -- SVD++ using a SGD optimizer
+ - 'RandSVD' -- RandomizedSVD learning
+ - 'QSVD' -- QuicSVD learning
+ - 'BKSVD' -- Block Krylov SVD learning
+
+
+The following neighbor search algorithms can be specified via the `neighbor_search` parameter:
+
+ - 'cosine'  -- Cosine Search Algorithm
+ - 'euclidean'  -- Euclidean Search Algorithm
+ - 'pearson'  -- Pearson Search Algorithm
+
+
+The following weight interpolation algorithms can be specified via the `interpolation` parameter:
+
+ - 'average'  -- Average Interpolation Algorithm
+ - 'regression'  -- Regression Interpolation Algorithm
+ - 'similarity'  -- Similarity Interpolation Algorithm
+
+
+The following ranking normalization algorithms can be specified via the `normalization` parameter:
+
+ - 'none'  -- No Normalization
+ - 'item_mean'  -- Item Mean Normalization
+ - 'overall_mean'  -- Overall Mean Normalization
+ - 'user_mean'  -- User Mean Normalization
+ - 'z_score'  -- Z-Score Normalization
+
+A trained model may be saved to with the `output_model` output parameter.
+
+### Example
+To train a CF model on a dataset ``training_set`` using NMF for decomposition and saving the trained model to ``model``, one could call: 
+
+```julia
+julia> using CSV
+julia> training_set = CSV.read("training_set.csv")
+julia> _, model = cf(algorithm="NMF", training=training_set)
+```
+
+Then, to use this model to generate recommendations for the list of users in the query set ``users``, storing 5 recommendations in ``recommendations``, one could call 
+
+```julia
+julia> using CSV
+julia> users = CSV.read("users.csv"; type=Int)
+julia> recommendations, _ = cf(input_model=model, query=users,
+            recommendations=5)
+```
+
+### See also
+
+ - [Collaborative Filtering on Wikipedia](https://en.wikipedia.org/wiki/Collaborative_filtering)
+ - [Matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Matrix_factorization_(recommender_systems))
+ - [Matrix factorization techniques for recommender systems (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5)
+ - [CFType class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/cf/cf.hpp)
+
+## dbscan()
+{: #dbscan }
+
+#### DBSCAN clustering
+{: #dbscan_descr }
+
+```julia
+julia> using mlpack: dbscan
+julia> assignments, centroids = dbscan(input; epsilon=1, min_size=5,
+          naive=false, selection_type="ordered", single_mode=false,
+          tree_type="kd", verbose=false)
+```
+
+An implementation of DBSCAN clustering.  Given a dataset, this can compute and return a clustering of that dataset. [Detailed documentation](#dbscan_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `epsilon` | [`Float64`](#doc_Float64) | Radius of each range search. | `1` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to cluster. | `**--**` |
+| `min_size` | [`Int`](#doc_Int) | Minimum number of points for a cluster. | `5` |
+| `naive` | [`Bool`](#doc_Bool) | If set, brute-force range search (not tree-based) will be used. | `false` |
+| `selection_type` | [`String`](#doc_String) | If using point selection policy, the type of selection to use ('ordered', 'random'). | `"ordered"` |
+| `single_mode` | [`Bool`](#doc_Bool) | If set, single-tree range search (not dual-tree) will be used. | `false` |
+| `tree_type` | [`String`](#doc_String) | If using single-tree or dual-tree search, the type of tree to use ('kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'). | `"kd"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `assignments` | [`Int vector-like`](#doc_Int_vector_like) | Output matrix for assignments of each point. | 
+| `centroids` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save output centroids to. | 
+
+### Detailed documentation
+{: #dbscan_detailed-documentation }
+
+This program implements the DBSCAN algorithm for clustering using accelerated tree-based range search.  The type of tree that is used may be parameterized, or brute-force range search may also be used.
+
+The input dataset to be clustered may be specified with the `input` parameter; the radius of each range search may be specified with the `epsilon` parameters, and the minimum number of points in a cluster may be specified with the `min_size` parameter.
+
+The `assignments` and `centroids` output parameters may be used to save the output of the clustering. `assignments` contains the cluster assignments of each point, and `centroids` contains the centroids of each cluster.
+
+The range search may be controlled with the `tree_type`, `single_mode`, and `naive` parameters.  `tree_type` can control the type of tree used for range search; this can take a variety of values: 'kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'. The `single_mode` parameter will force single-tree search (as opposed to the default dual-tree search), and '`naive` will force brute-force range search.
+
+### Example
+An example usage to run DBSCAN on the dataset in ``input`` with a radius of 0.5 and a minimum cluster size of 5 is given below:
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> _, _ = dbscan(input; epsilon=0.5, min_size=5)
+```
+
+### See also
+
+ - [DBSCAN on Wikipedia](https://en.wikipedia.org/wiki/DBSCAN)
+ - [A density-based algorithm for discovering clusters in large spatial databases with noise (pdf)](https://cdn.aaai.org/KDD/1996/KDD96-037.pdf)
+ - [DBSCAN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/dbscan/dbscan.hpp)
+
+## decision_tree()
+{: #decision_tree }
+
+#### Decision tree
+{: #decision_tree_descr }
+
+```julia
+julia> using mlpack: decision_tree
+julia> output_model, predictions, probabilities = decision_tree( ;
+          input_model=nothing, labels=Int[], maximum_depth=0,
+          minimum_gain_split=1e-07, minimum_leaf_size=20,
+          print_training_accuracy=false, print_training_error=false,
+          test=zeros(0, 0), test_labels=Int[], training=zeros(0, 0),
+          verbose=false, weights=zeros(0, 0))
+```
+
+An implementation of an ID3-style decision tree for classification, which supports categorical data.  Given labeled data with numeric or categorical features, a decision tree can be trained and saved; or, an existing decision tree can be used for classification on new points. [Detailed documentation](#decision_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`DecisionTreeModel`](#doc_model) | Pre-trained decision tree, to be used with test points. | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Training labels. | `Int[]` |
+| `maximum_depth` | [`Int`](#doc_Int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`Float64`](#doc_Float64) | Minimum gain for node splitting. | `1e-07` |
+| `minimum_leaf_size` | [`Int`](#doc_Int) | Minimum number of points in a leaf. | `20` |
+| `print_training_accuracy` | [`Bool`](#doc_Bool) | Print the training accuracy. | `false` |
+| `print_training_error` | [`Bool`](#doc_Bool) | Print the training error (deprecated; will be removed in mlpack 4.0.0). | `false` |
+| `test` | [`Tuple{Array{Bool, 1}, Array{Float64, 2}}`](#doc_Tuple_Array_Bool__1___Array_Float64__2__) | Testing dataset (may be categorical). | `zeros(0, 0)` |
+| `test_labels` | [`Int vector-like`](#doc_Int_vector_like) | Test point labels, if accuracy calculation is desired. | `Int[]` |
+| `training` | [`Tuple{Array{Bool, 1}, Array{Float64, 2}}`](#doc_Tuple_Array_Bool__1___Array_Float64__2__) | Training dataset (may be categorical). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `weights` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The weight of labels | `zeros(0, 0)` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DecisionTreeModel`](#doc_model) | Output for trained decision tree. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | Class predictions for each test point. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Class probabilities for each test point. | 
+
+### Detailed documentation
+{: #decision_tree_detailed-documentation }
+
+Train and evaluate using a decision tree.  Given a dataset containing numeric or categorical features, and associated labels for each point in the dataset, this program can train a decision tree on that data.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model` parameter.  The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `minimum_gain_split` parameter specifies the minimum gain that is needed for the node to split.  The `maximum_depth` parameter specifies the maximum depth of the tree.  If `print_training_error` is specified, the training error will be printed.
+
+Test data may be specified with the `test` parameter, and if performance numbers are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions` output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a decision tree with a minimum leaf size of 20 on the dataset contained in ``data`` with labels ``labels``, saving the output model to ``tree`` and printing the training error, one could call
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> tree, _, _ = decision_tree(labels=labels,
+            minimum_gain_split=0.001, minimum_leaf_size=20,
+            print_training_accuracy=1, training=data)
+```
+
+Then, to use that model to classify points in ``test_set`` and print the test error given the labels ``test_labels`` using that model, while saving the predictions for each point to ``predictions``, one could call 
+
+```julia
+julia> using CSV
+julia> test_set = CSV.read("test_set.csv")
+julia> test_labels = CSV.read("test_labels.csv"; type=Int)
+julia> _, predictions, _ = decision_tree(input_model=tree,
+            test=test_set, test_labels=test_labels)
+```
+
+### See also
+
+ - [Random forest](#random_forest)
+ - [Decision trees on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+ - [Induction of Decision Trees (pdf)](http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf)
+ - [DecisionTree C++ class documentation](../../user/methods/decision_tree.md)
+
+## det()
+{: #det }
+
+#### Density Estimation With Density Estimation Trees
+{: #det_descr }
+
+```julia
+julia> using mlpack: det
+julia> output_model, tag_counters_file, tag_file, test_set_estimates,
+          training_set_estimates, vi = det( ; folds=10, input_model=nothing,
+          max_leaf_size=10, min_leaf_size=5, path_format="lr",
+          skip_pruning=false, test=zeros(0, 0), training=zeros(0, 0),
+          verbose=false)
+```
+
+An implementation of density estimation trees for the density estimation task.  Density estimation trees can be trained or used to predict the density at locations given by query points. [Detailed documentation](#det_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `folds` | [`Int`](#doc_Int) | The number of folds of cross-validation to perform for the estimation (0 is LOOCV) | `10` |
+| `input_model` | [`DTree`](#doc_model) | Trained density estimation tree to load. | `nothing` |
+| `max_leaf_size` | [`Int`](#doc_Int) | The maximum size of a leaf in the unpruned, fully grown DET. | `10` |
+| `min_leaf_size` | [`Int`](#doc_Int) | The minimum size of a leaf in the unpruned, fully grown DET. | `5` |
+| `path_format` | [`String`](#doc_String) | The format of path printing: 'lr', 'id-lr', or 'lr-id'. | `"lr"` |
+| `skip_pruning` | [`Bool`](#doc_Bool) | Whether to bypass the pruning process and output the unpruned tree only. | `false` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A set of test points to estimate the density of. | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The data set on which to build a density estimation tree. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DTree`](#doc_model) | Output to save trained density estimation tree to. | 
+| `tag_counters_file` | [`String`](#doc_String) | The file to output the number of points that went to each leaf. | 
+| `tag_file` | [`String`](#doc_String) | The file to output the tags (and possibly paths) for each sample in the test set. | 
+| `test_set_estimates` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The output estimates on the test set from the final optimally pruned tree. | 
+| `training_set_estimates` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The output density estimates on the training set from the final optimally pruned tree. | 
+| `vi` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The output variable importance values for each feature. | 
+
+### Detailed documentation
+{: #det_detailed-documentation }
+
+This program performs a number of functions related to Density Estimation Trees.  The optimal Density Estimation Tree (DET) can be trained on a set of data (specified by `training`) using cross-validation (with number of folds specified with the `folds` parameter).  This trained density estimation tree may then be saved with the `output_model` output parameter.
+
+The variable importances (that is, the feature importance values for each dimension) may be saved with the `vi` output parameter, and the density estimates for each training point may be saved with the `training_set_estimates` output parameter.
+
+Enabling path printing for each node outputs the path from the root node to a leaf for each entry in the test set, or training set (if a test set is not provided).  Strings like 'LRLRLR' (indicating that traversal went to the left child, then the right child, then the left child, and so forth) will be output. If 'lr-id' or 'id-lr' are given as the `path_format` parameter, then the ID (tag) of every node along the path will be printed after or before the L or R character indicating the direction of traversal, respectively.
+
+This program also can provide density estimates for a set of test points, specified in the `test` parameter.  The density estimation tree used for this task will be the tree that was trained on the given training points, or a tree given as the parameter `input_model`.  The density estimates for the test points may be saved using the `test_set_estimates` output parameter.
+
+### See also
+
+ - [Density estimation on Wikipedia](https://en.wikipedia.org/wiki/Density_estimation)
+ - [Density estimation trees (pdf)](https://www.mlpack.org/papers/det.pdf)
+ - [DTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/det/dtree.hpp)
+
+## emst()
+{: #emst }
+
+#### Fast Euclidean Minimum Spanning Tree
+{: #emst_descr }
+
+```julia
+julia> using mlpack: emst
+julia> output = emst(input; leaf_size=1, naive=false,
+                     verbose=false)
+```
+
+An implementation of the Dual-Tree Boruvka algorithm for computing the Euclidean minimum spanning tree of a set of input points. [Detailed documentation](#emst_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input data matrix. | `**--**` |
+| `leaf_size` | [`Int`](#doc_Int) | Leaf size in the kd-tree.  One-element leaves give the empirically best performance, but at the cost of greater memory requirements. | `1` |
+| `naive` | [`Bool`](#doc_Bool) | Compute the MST using O(n^2) naive algorithm. | `false` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output data.  Stored as an edge list. | 
+
+### Detailed documentation
+{: #emst_detailed-documentation }
+
+This program can compute the Euclidean minimum spanning tree of a set of input points using the dual-tree Boruvka algorithm.
+
+The set to calculate the minimum spanning tree of is specified with the `input` parameter, and the output may be saved with the `output` output parameter.
+
+The `leaf_size` parameter controls the leaf size of the kd-tree that is used to calculate the minimum spanning tree, and if the `naive` option is given, then brute-force search is used (this is typically much slower in low dimensions).  The leaf size does not affect the results, but it may have some effect on the runtime of the algorithm.
+
+### Example
+For example, the minimum spanning tree of the input dataset ``data`` can be calculated with a leaf size of 20 and stored as ``spanning_tree`` using the following command:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> spanning_tree = emst(data; leaf_size=20)
+```
+
+The output matrix is a three-dimensional matrix, where each row indicates an edge.  The first dimension corresponds to the lesser index of the edge; the second dimension corresponds to the greater index of the edge; and the third column corresponds to the distance between the two points.
+
+### See also
+
+ - [Minimum spanning tree on Wikipedia](https://en.wikipedia.org/wiki/Minimum_spanning_tree)
+ - [Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications (pdf)](https://www.mlpack.org/papers/emst.pdf)
+ - [DualTreeBoruvka class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/emst/dtb.hpp)
+
+## fastmks()
+{: #fastmks }
+
+#### FastMKS (Fast Max-Kernel Search)
+{: #fastmks_descr }
+
+```julia
+julia> using mlpack: fastmks
+julia> indices, kernels, output_model = fastmks( ; bandwidth=1,
+          base=2, degree=2, input_model=nothing, k=0, kernel="linear",
+          naive=false, offset=0, query=zeros(0, 0), reference=zeros(0, 0),
+          scale=1, single=false, verbose=false)
+```
+
+An implementation of the single-tree and dual-tree fast max-kernel search (FastMKS) algorithm.  Given a set of reference points and a set of query points, this can find the reference point with maximum kernel value for each query point; trained models can be reused for future queries. [Detailed documentation](#fastmks_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`Float64`](#doc_Float64) | Bandwidth (for Gaussian, Epanechnikov, and triangular kernels). | `1` |
+| `base` | [`Float64`](#doc_Float64) | Base to use during cover tree construction. | `2` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `degree` | [`Float64`](#doc_Float64) | Degree of polynomial kernel. | `2` |
+| `input_model` | [`FastMKSModel`](#doc_model) | Input FastMKS model to use. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of maximum kernels to find. | `0` |
+| `kernel` | [`String`](#doc_String) | Kernel type to use: 'linear', 'polynomial', 'cosine', 'gaussian', 'epanechnikov', 'triangular', 'hyptan'. | `"linear"` |
+| `naive` | [`Bool`](#doc_Bool) | If true, O(n^2) naive mode is used for computation. | `false` |
+| `offset` | [`Float64`](#doc_Float64) | Offset of kernel (for polynomial and hyptan kernels). | `0` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The query dataset. | `zeros(0, 0)` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The reference dataset. | `zeros(0, 0)` |
+| `scale` | [`Float64`](#doc_Float64) | Scale of kernel (for hyptan kernel). | `1` |
+| `single` | [`Bool`](#doc_Bool) | If true, single-tree search is used (as opposed to dual-tree search. | `false` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `indices` | [`Int matrix-like`](#doc_Int_matrix_like) | Output matrix of indices. | 
+| `kernels` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output matrix of kernels. | 
+| `output_model` | [`FastMKSModel`](#doc_model) | Output for FastMKS model. | 
+
+### Detailed documentation
+{: #fastmks_detailed-documentation }
+
+This program will find the k maximum kernels of a set of points, using a query set and a reference set (which can optionally be the same set). More specifically, for each point in the query set, the k points in the reference set with maximum kernel evaluations are found.  The kernel function used is specified with the `kernel` parameter.
+
+### Example
+For example, the following command will calculate, for each point in the query set ``query``, the five points in the reference set ``reference`` with maximum kernel evaluation using the linear kernel.  The kernel evaluations may be saved with the  ``kernels`` output parameter and the indices may be saved with the ``indices`` output parameter.
+
+```julia
+julia> using CSV
+julia> reference = CSV.read("reference.csv")
+julia> query = CSV.read("query.csv")
+julia> indices, kernels, _ = fastmks(k=5, kernel="linear",
+            query=query, reference=reference)
+```
+
+The output matrices are organized such that row i and column j in the indices matrix corresponds to the index of the point in the reference set that has j'th largest kernel evaluation with the point in the query set with index i.  Row i and column j in the kernels matrix corresponds to the kernel evaluation between those two points.
+
+This program performs FastMKS using a cover tree.  The base used to build the cover tree can be specified with the `base` parameter.
+
+### See also
+
+ - [k-nearest-neighbor search](#knn)
+ - [Dual-tree Fast Exact Max-Kernel Search (pdf)](https://mlpack.org/papers/fmks.pdf)
+ - [FastMKS class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/fastmks/fastmks.hpp)
+
+## gmm_train()
+{: #gmm_train }
+
+#### Gaussian Mixture Model (GMM) Training
+{: #gmm_train_descr }
+
+```julia
+julia> using mlpack: gmm_train
+julia> output_model = gmm_train(gaussians,
+                                input; diagonal_covariance=false,
+                                input_model=nothing, kmeans_max_iterations=1000,
+                                max_iterations=250, no_force_positive=false,
+                                noise=0, percentage=0.02, refined_start=false,
+                                samplings=100, seed=0, tolerance=1e-10,
+                                trials=1, verbose=false)
+```
+
+An implementation of the EM algorithm for training Gaussian mixture models (GMMs).  Given a dataset, this can train a GMM for future use with other tools. [Detailed documentation](#gmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `diagonal_covariance` | [`Bool`](#doc_Bool) | Force the covariance of the Gaussians to be diagonal.  This can accelerate training time significantly. | `false` |
+| `gaussians` | [`Int`](#doc_Int) | Number of Gaussians in the GMM. | `**--**` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The training data on which the model will be fit. | `**--**` |
+| `input_model` | [`GMM`](#doc_model) | Initial input GMM model to start training with. | `nothing` |
+| `kmeans_max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations for the k-means algorithm (used to initialize EM). | `1000` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations of EM algorithm (passing 0 will run until convergence). | `250` |
+| `no_force_positive` | [`Bool`](#doc_Bool) | Do not force the covariance matrices to be positive definite. | `false` |
+| `noise` | [`Float64`](#doc_Float64) | Variance of zero-mean Gaussian noise to add to data. | `0` |
+| `percentage` | [`Float64`](#doc_Float64) | If using --refined_start, specify the percentage of the dataset used for each sampling (should be between 0.0 and 1.0). | `0.02` |
+| `refined_start` | [`Bool`](#doc_Bool) | During the initialization, use refined initial positions for k-means clustering (Bradley and Fayyad, 1998). | `false` |
+| `samplings` | [`Int`](#doc_Int) | If using --refined_start, specify the number of samplings used for initial points. | `100` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tolerance` | [`Float64`](#doc_Float64) | Tolerance for convergence of EM. | `1e-10` |
+| `trials` | [`Int`](#doc_Int) | Number of trials to perform in training GMM. | `1` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`GMM`](#doc_model) | Output for trained GMM model. | 
+
+### Detailed documentation
+{: #gmm_train_detailed-documentation }
+
+This program takes a parametric estimate of a Gaussian mixture model (GMM) using the EM algorithm to find the maximum likelihood estimate.  The model may be saved and reused by other mlpack GMM tools.
+
+The input data to train on must be specified with the `input` parameter, and the number of Gaussians in the model must be specified with the `gaussians` parameter.  Optionally, many trials with different random initializations may be run, and the result with highest log-likelihood on the training data will be taken.  The number of trials to run is specified with the `trials` parameter.  By default, only one trial is run.
+
+The tolerance for convergence and maximum number of iterations of the EM algorithm are specified with the `tolerance` and `max_iterations` parameters, respectively.  The GMM may be initialized for training with another model, specified with the `input_model` parameter. Otherwise, the model is initialized by running k-means on the data.  The k-means clustering initialization can be controlled with the `kmeans_max_iterations`, `refined_start`, `samplings`, and `percentage` parameters.  If `refined_start` is specified, then the Bradley-Fayyad refined start initialization will be used.  This can often lead to better clustering results.
+
+The 'diagonal_covariance' flag will cause the learned covariances to be diagonal matrices.  This significantly simplifies the model itself and causes training to be faster, but restricts the ability to fit more complex GMMs.
+
+If GMM training fails with an error indicating that a covariance matrix could not be inverted, make sure that the `no_force_positive` parameter is not specified.  Alternately, adding a small amount of Gaussian noise (using the `noise` parameter) to the entire dataset may help prevent Gaussians with zero variance in a particular dimension, which is usually the cause of non-invertible covariance matrices.
+
+The `no_force_positive` parameter, if set, will avoid the checks after each iteration of the EM algorithm which ensure that the covariance matrices are positive definite.  Specifying the flag can cause faster runtime, but may also cause non-positive definite covariance matrices, which will cause the program to crash.
+
+### Example
+As an example, to train a 6-Gaussian GMM on the data in ``data`` with a maximum of 100 iterations of EM and 3 trials, saving the trained GMM to ``gmm``, the following command can be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> gmm = gmm_train(6, data; trials=3)
+```
+
+To re-train that GMM on another set of data ``data2``, the following command may be used: 
+
+```julia
+julia> using CSV
+julia> data2 = CSV.read("data2.csv")
+julia> new_gmm = gmm_train(6, data2; input_model=gmm)
+```
+
+### See also
+
+ - [gmm_generate()](#gmm_generate)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_generate()
+{: #gmm_generate }
+
+#### GMM Sample Generator
+{: #gmm_generate_descr }
+
+```julia
+julia> using mlpack: gmm_generate
+julia> output = gmm_generate(input_model, samples;
+                             seed=0, verbose=false)
+```
+
+A sample generator for pre-trained GMMs.  Given a pre-trained GMM, this can sample new points randomly from that distribution. [Detailed documentation](#gmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`GMM`](#doc_model) | Input GMM model to generate samples from. | `**--**` |
+| `samples` | [`Int`](#doc_Int) | Number of samples to generate. | `**--**` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save output samples in. | 
+
+### Detailed documentation
+{: #gmm_generate_detailed-documentation }
+
+This program is able to generate samples from a pre-trained GMM (use gmm_train to train a GMM).  The pre-trained GMM must be specified with the `input_model` parameter.  The number of samples to generate is specified by the `samples` parameter.  Output samples may be saved with the `output` output parameter.
+
+### Example
+The following command can be used to generate 100 samples from the pre-trained GMM ``gmm`` and store those generated samples in ``samples``:
+
+```julia
+julia> samples = gmm_generate(gmm, 100)
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_probability()
+{: #gmm_probability }
+
+#### GMM Probability Calculator
+{: #gmm_probability_descr }
+
+```julia
+julia> using mlpack: gmm_probability
+julia> output = gmm_probability(input,
+                                input_model; verbose=false)
+```
+
+A probability calculator for GMMs.  Given a pre-trained GMM and a set of points, this can compute the probability that each point is from the given GMM. [Detailed documentation](#gmm_probability_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input matrix to calculate probabilities of. | `**--**` |
+| `input_model` | [`GMM`](#doc_model) | Input GMM to use as model. | `**--**` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to store calculated probabilities in. | 
+
+### Detailed documentation
+{: #gmm_probability_detailed-documentation }
+
+This program calculates the probability that given points came from a given GMM (that is, P(X \| gmm)).  The GMM is specified with the `input_model` parameter, and the points are specified with the `input` parameter.  The output probabilities may be saved via the `output` output parameter.
+
+### Example
+So, for example, to calculate the probabilities of each point in ``points`` coming from the pre-trained GMM ``gmm``, while storing those probabilities in ``probs``, the following command could be used:
+
+```julia
+julia> using CSV
+julia> points = CSV.read("points.csv")
+julia> probs = gmm_probability(points, gmm)
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_generate()](#gmm_generate)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## hmm_train()
+{: #hmm_train }
+
+#### Hidden Markov Model (HMM) Training
+{: #hmm_train_descr }
+
+```julia
+julia> using mlpack: hmm_train
+julia> output_model = hmm_train(input_file;
+                                batch=false, gaussians=0, input_model=nothing,
+                                labels_file="", seed=0, states=0,
+                                tolerance=1e-05, type="gaussian",
+                                verbose=false)
+```
+
+An implementation of training algorithms for Hidden Markov Models (HMMs). Given labeled or unlabeled data, an HMM can be trained for further use with other mlpack HMM tools. [Detailed documentation](#hmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch` | [`Bool`](#doc_Bool) | If true, input_file (and if passed, labels_file) are expected to contain a list of files to use as input observation sequences (and label sequences). | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `gaussians` | [`Int`](#doc_Int) | Number of gaussians in each GMM (necessary when type is 'gmm'). | `0` |
+| `input_file` | [`String`](#doc_String) | File containing input observations. | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | Pre-existing HMM model to initialize training with. | `nothing` |
+| `labels_file` | [`String`](#doc_String) | Optional file of hidden states, used for labeled training. | `""` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `states` | [`Int`](#doc_Int) | Number of hidden states in HMM (necessary, unless model_file is specified). | `0` |
+| `tolerance` | [`Float64`](#doc_Float64) | Tolerance of the Baum-Welch algorithm. | `1e-05` |
+| `type` | [`String`](#doc_String) | Type of HMM: discrete \| gaussian \| diag_gmm \| gmm. | `"gaussian"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HMMModel`](#doc_model) | Output for trained HMM. | 
+
+### Detailed documentation
+{: #hmm_train_detailed-documentation }
+
+This program allows a Hidden Markov Model to be trained on labeled or unlabeled data.  It supports four types of HMMs: Discrete HMMs, Gaussian HMMs, GMM HMMs, or Diagonal GMM HMMs
+
+Either one input sequence can be specified (with `input_file`), or, a file containing files in which input sequences can be found (when `input_file`and`batch` are used together).  In addition, labels can be provided in the file specified by `labels_file`, and if `batch` is used, the file given to `labels_file` should contain a list of files of labels corresponding to the sequences in the file given to `input_file`.
+
+The HMM is trained with the Baum-Welch algorithm if no labels are provided.  The tolerance of the Baum-Welch algorithm can be set with the `tolerance`option.  By default, the transition matrix is randomly initialized and the emission distributions are initialized to fit the extent of the data.
+
+Optionally, a pre-created HMM model can be used as a guess for the transition matrix and emission probabilities; this is specifiable with `output_model`.
+
+### See also
+
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_generate()
+{: #hmm_generate }
+
+#### Hidden Markov Model (HMM) Sequence Generator
+{: #hmm_generate_descr }
+
+```julia
+julia> using mlpack: hmm_generate
+julia> output, state = hmm_generate(length, model; seed=0,
+          start_state=0, verbose=false)
+```
+
+A utility to generate random sequences from a pre-trained Hidden Markov Model (HMM).  The length of the desired sequence can be specified, and a random sequence of observations is returned. [Detailed documentation](#hmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `length` | [`Int`](#doc_Int) | Length of sequence to generate. | `**--**` |
+| `model` | [`HMMModel`](#doc_model) | Trained HMM to generate sequences with. | `**--**` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `start_state` | [`Int`](#doc_Int) | Starting state of sequence. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save observation sequence to. | 
+| `state` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to save hidden state sequence to. | 
+
+### Detailed documentation
+{: #hmm_generate_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as the `model` parameter, and generates a random observation sequence and hidden state sequence based on its parameters. The observation sequence may be saved with the `output` output parameter, and the internal state  sequence may be saved with the `state` output parameter.
+
+The state to start the sequence in may be specified with the `start_state` parameter.
+
+### Example
+For example, to generate a sequence of length 150 from the HMM ``hmm`` and save the observation sequence to ``observations`` and the hidden state sequence to ``states``, the following command may be used: 
+
+```julia
+julia> observations, states = hmm_generate(150, hmm)
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_loglik()
+{: #hmm_loglik }
+
+#### Hidden Markov Model (HMM) Sequence Log-Likelihood
+{: #hmm_loglik_descr }
+
+```julia
+julia> using mlpack: hmm_loglik
+julia> log_likelihood = hmm_loglik(input,
+                                   input_model; verbose=false)
+```
+
+A utility for computing the log-likelihood of a sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observation sequence, this computes and returns the log-likelihood of that sequence being observed from that HMM. [Detailed documentation](#hmm_loglik_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | File containing observations, | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | File containing HMM. | `**--**` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `log_likelihood` | [`Float64`](#doc_Float64) | Log-likelihood of the sequence. | 
+
+### Detailed documentation
+{: #hmm_loglik_detailed-documentation }
+
+This utility takes an already-trained HMM, specified with the `input_model` parameter, and evaluates the log-likelihood of a sequence of observations, given with the `input` parameter.  The computed log-likelihood is given as output.
+
+### Example
+For example, to compute the log-likelihood of the sequence ``seq`` with the pre-trained HMM ``hmm``, the following command may be used: 
+
+```julia
+julia> using CSV
+julia> seq = CSV.read("seq.csv")
+julia> _ = hmm_loglik(seq, hmm)
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_viterbi()
+{: #hmm_viterbi }
+
+#### Hidden Markov Model (HMM) Viterbi State Prediction
+{: #hmm_viterbi_descr }
+
+```julia
+julia> using mlpack: hmm_viterbi
+julia> output = hmm_viterbi(input, input_model;
+                            verbose=false)
+```
+
+A utility for computing the most probable hidden state sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observed sequence, this uses the Viterbi algorithm to compute and return the most probable hidden state sequence. [Detailed documentation](#hmm_viterbi_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing observations, | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | Trained HMM to use. | `**--**` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Int matrix-like`](#doc_Int_matrix_like) | File to save predicted state sequence to. | 
+
+### Detailed documentation
+{: #hmm_viterbi_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as `input_model`, and evaluates the most probable hidden state sequence of a given sequence of observations (specified as '`input`, using the Viterbi algorithm.  The computed state sequence may be saved using the `output` output parameter.
+
+### Example
+For example, to predict the state sequence of the observations ``obs`` using the HMM ``hmm``, storing the predicted state sequence to ``states``, the following command could be used:
+
+```julia
+julia> using CSV
+julia> obs = CSV.read("obs.csv")
+julia> states = hmm_viterbi(obs, hmm)
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hoeffding_tree()
+{: #hoeffding_tree }
+
+#### Hoeffding trees
+{: #hoeffding_tree_descr }
+
+```julia
+julia> using mlpack: hoeffding_tree
+julia> output_model, predictions, probabilities = hoeffding_tree( ;
+          batch_mode=false, bins=10, confidence=0.95, info_gain=false,
+          input_model=nothing, labels=Int[], max_samples=5000, min_samples=100,
+          numeric_split_strategy="binary", observations_before_binning=100,
+          passes=1, test=zeros(0, 0), test_labels=Int[], training=zeros(0, 0),
+          verbose=false)
+```
+
+An implementation of Hoeffding trees, a form of streaming decision tree for classification.  Given labeled data, a Hoeffding tree can be trained and saved for later use, or a pre-trained Hoeffding tree can be used for predicting the classifications of new points. [Detailed documentation](#hoeffding_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_mode` | [`Bool`](#doc_Bool) | If true, samples will be considered in batch instead of as a stream.  This generally results in better trees but at the cost of memory usage and runtime. | `false` |
+| `bins` | [`Int`](#doc_Int) | If the 'domingos' split strategy is used, this specifies the number of bins for each numeric split. | `10` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `confidence` | [`Float64`](#doc_Float64) | Confidence before splitting (between 0 and 1). | `0.95` |
+| `info_gain` | [`Bool`](#doc_Bool) | If set, information gain is used instead of Gini impurity for calculating Hoeffding bounds. | `false` |
+| `input_model` | [`HoeffdingTreeModel`](#doc_model) | Input trained Hoeffding tree model. | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels for training dataset. | `Int[]` |
+| `max_samples` | [`Int`](#doc_Int) | Maximum number of samples before splitting. | `5000` |
+| `min_samples` | [`Int`](#doc_Int) | Minimum number of samples before splitting. | `100` |
+| `numeric_split_strategy` | [`String`](#doc_String) | The splitting strategy to use for numeric features: 'domingos' or 'binary'. | `"binary"` |
+| `observations_before_binning` | [`Int`](#doc_Int) | If the 'domingos' split strategy is used, this specifies the number of samples observed before binning is performed. | `100` |
+| `passes` | [`Int`](#doc_Int) | Number of passes to take over the dataset. | `1` |
+| `test` | [`Tuple{Array{Bool, 1}, Array{Float64, 2}}`](#doc_Tuple_Array_Bool__1___Array_Float64__2__) | Testing dataset (may be categorical). | `zeros(0, 0)` |
+| `test_labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels of test data. | `Int[]` |
+| `training` | [`Tuple{Array{Bool, 1}, Array{Float64, 2}}`](#doc_Tuple_Array_Bool__1___Array_Float64__2__) | Training dataset (may be categorical). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HoeffdingTreeModel`](#doc_model) | Output for trained Hoeffding tree model. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | Matrix to output label predictions for test data into. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | In addition to predicting labels, provide rediction probabilities in this matrix. | 
+
+### Detailed documentation
+{: #hoeffding_tree_detailed-documentation }
+
+This program implements Hoeffding trees, a form of streaming decision tree suited best for large (or streaming) datasets.  This program supports both categorical and numeric data.  Given an input dataset, this program is able to train the tree with numerous training options, and save the model to a file.  The program is also able to use a trained model or a model from file in order to predict classes for a given test set.
+
+The training file and associated labels are specified with the `training` and `labels` parameters, respectively. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+The training may be performed in batch mode (like a typical decision tree algorithm) by specifying the `batch_mode` option, but this may not be the best option for large datasets.
+
+When a model is trained, it may be saved via the `output_model` output parameter.  A model may be loaded from file for further training or testing with the `input_model` parameter.
+
+Test data may be specified with the `test` parameter, and if performance statistics are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved with the `predictions` output parameter, and class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a Hoeffding tree with confidence 0.99 with data ``dataset``, saving the trained tree to ``tree``, the following command may be used:
+
+```julia
+julia> using CSV
+julia> dataset = CSV.read("dataset.csv")
+julia> tree, _, _ = hoeffding_tree(confidence=0.99,
+            training=dataset)
+```
+
+Then, this tree may be used to make predictions on the test set ``test_set``, saving the predictions into ``predictions`` and the class probabilities into ``class_probs`` with the following command: 
+
+```julia
+julia> using CSV
+julia> test_set = CSV.read("test_set.csv")
+julia> _, predictions, class_probs =
+            hoeffding_tree(input_model=tree, test=test_set)
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [random_forest()](#random_forest)
+ - [Mining High-Speed Data Streams (pdf)](http://dm.cs.washington.edu/papers/vfdt-kdd00.pdf)
+ - [HoeffdingTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp)
+
+## kde()
+{: #kde }
+
+#### Kernel Density Estimation
+{: #kde_descr }
+
+```julia
+julia> using mlpack: kde
+julia> output_model, predictions = kde( ; abs_error=0,
+          algorithm="dual-tree", bandwidth=1, initial_sample_size=100,
+          input_model=nothing, kernel="gaussian", mc_break_coef=0.4,
+          mc_entry_coef=3, mc_probability=0.95, monte_carlo=false,
+          query=zeros(0, 0), reference=zeros(0, 0), rel_error=0.05,
+          tree="kd-tree", verbose=false)
+```
+
+An implementation of kernel density estimation with dual-tree algorithms. Given a set of reference points and query points and a kernel function, this can estimate the density function at the location of each query point using trees; trees that are built can be saved for later use. [Detailed documentation](#kde_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `abs_error` | [`Float64`](#doc_Float64) | Relative error tolerance for the prediction. | `0` |
+| `algorithm` | [`String`](#doc_String) | Algorithm to use for the prediction.('dual-tree', 'single-tree'). | `"dual-tree"` |
+| `bandwidth` | [`Float64`](#doc_Float64) | Bandwidth of the kernel. | `1` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `initial_sample_size` | [`Int`](#doc_Int) | Initial sample size for Monte Carlo estimations. | `100` |
+| `input_model` | [`KDEModel`](#doc_model) | Contains pre-trained KDE model. | `nothing` |
+| `kernel` | [`String`](#doc_String) | Kernel to use for the prediction.('gaussian', 'epanechnikov', 'laplacian', 'spherical', 'triangular'). | `"gaussian"` |
+| `mc_break_coef` | [`Float64`](#doc_Float64) | Controls what fraction of the amount of node's descendants is the limit for the sample size before it recurses. | `0.4` |
+| `mc_entry_coef` | [`Float64`](#doc_Float64) | Controls how much larger does the amount of node descendants has to be compared to the initial sample size in order to be a candidate for Monte Carlo estimations. | `3` |
+| `mc_probability` | [`Float64`](#doc_Float64) | Probability of the estimation being bounded by relative error when using Monte Carlo estimations. | `0.95` |
+| `monte_carlo` | [`Bool`](#doc_Bool) | Whether to use Monte Carlo estimations when possible. | `false` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Query dataset to KDE on. | `zeros(0, 0)` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input reference dataset use for KDE. | `zeros(0, 0)` |
+| `rel_error` | [`Float64`](#doc_Float64) | Relative error tolerance for the prediction. | `0.05` |
+| `tree` | [`String`](#doc_String) | Tree to use for the prediction.('kd-tree', 'ball-tree', 'cover-tree', 'octree', 'r-tree'). | `"kd-tree"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`KDEModel`](#doc_model) | If specified, the KDE model will be saved here. | 
+| `predictions` | [`Float64 vector-like`](#doc_Float64_vector_like) | Vector to store density predictions. | 
+
+### Detailed documentation
+{: #kde_detailed-documentation }
+
+This program performs a Kernel Density Estimation. KDE is a non-parametric way of estimating probability density function. For each query point the program will estimate its probability density by applying a kernel function to each reference point. The computational complexity of this is O(N^2) where there are N query points and N reference points, but this implementation will typically see better performance as it uses an approximate dual or single tree algorithm for acceleration.
+
+Dual or single tree optimization avoids many barely relevant calculations (as kernel function values decrease with distance), so it is an approximate computation. You can specify the maximum relative error tolerance for each query value with `rel_error` as well as the maximum absolute error tolerance with the parameter `abs_error`. This program runs using an Euclidean metric. Kernel function can be selected using the `kernel` option. You can also choose what which type of tree to use for the dual-tree algorithm with `tree`. It is also possible to select whether to use dual-tree algorithm or single-tree algorithm using the `algorithm` option.
+
+Monte Carlo estimations can be used to accelerate the KDE estimate when the Gaussian Kernel is used. This provides a probabilistic guarantee on the the error of the resulting KDE instead of an absolute guarantee.To enable Monte Carlo estimations, the `monte_carlo` flag can be used, and success probability can be set with the `mc_probability` option. It is possible to set the initial sample size for the Monte Carlo estimation using `initial_sample_size`. This implementation will only consider a node, as a candidate for the Monte Carlo estimation, if its number of descendant nodes is bigger than the initial sample size. This can be controlled using a coefficient that will multiply the initial sample size and can be set using `mc_entry_coef`. To avoid using the same amount of computations an exact approach would take, this program recurses the tree whenever a fraction of the amount of the node's descendant points have already been computed. This fraction is set using `mc_break_coef`.
+
+### Example
+For example, the following will run KDE using the data in ``ref_data`` for training and the data in ``qu_data`` as query data. It will apply an Epanechnikov kernel with a 0.2 bandwidth to each reference point and use a KD-Tree for the dual-tree optimization. The returned predictions will be within 5% of the real KDE value for each query point.
+
+```julia
+julia> using CSV
+julia> ref_data = CSV.read("ref_data.csv")
+julia> qu_data = CSV.read("qu_data.csv")
+julia> _, out_data = kde(bandwidth=0.2, kernel="epanechnikov",
+            query=qu_data, reference=ref_data, rel_error=0.05, tree="kd-tree")
+```
+
+the predicted density estimations will be stored in ``out_data``.
+If no `query` is provided, then KDE will be computed on the `reference` dataset.
+It is possible to select either a reference dataset or an input model but not both at the same time. If an input model is selected and parameter values are not set (e.g. `bandwidth`) then default parameter values will be used.
+
+In addition to the last program call, it is also possible to activate Monte Carlo estimations if a Gaussian kernel is used. This can provide faster results, but the KDE will only have a probabilistic guarantee of meeting the desired error bound (instead of an absolute guarantee). The following example will run KDE using a Monte Carlo estimation when possible. The results will be within a 5% of the real KDE value with a 95% probability. Initial sample size for the Monte Carlo estimation will be 200 points and a node will be a candidate for the estimation only when it contains 700 (i.e. 3.5*200) points. If a node contains 700 points and 420 (i.e. 0.6*700) have already been sampled, then the algorithm will recurse instead of keep sampling.
+
+```julia
+julia> using CSV
+julia> ref_data = CSV.read("ref_data.csv")
+julia> qu_data = CSV.read("qu_data.csv")
+julia> _, out_data = kde(bandwidth=0.2, initial_sample_size=200,
+            kernel="gaussian", mc_break_coef=0.6, mc_entry_coef=3.5,
+            mc_probability=0.95, monte_carlo=, query=qu_data,
+            reference=ref_data, rel_error=0.05, tree="kd-tree")
+```
+
+### See also
+
+ - [knn()](#knn)
+ - [Kernel density estimation on Wikipedia](https://en.wikipedia.org/wiki/Kernel_density_estimation)
+ - [Tree-Independent Dual-Tree Algorithms](https://arxiv.org/pdf/1304.4327)
+ - [Fast High-dimensional Kernel Summations Using the Monte Carlo Multipole Method](http://papers.nips.cc/paper/3539-fast-high-dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf)
+ - [KDE C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kde/kde.hpp)
+
+## kernel_pca()
+{: #kernel_pca }
+
+#### Kernel Principal Components Analysis
+{: #kernel_pca_descr }
+
+```julia
+julia> using mlpack: kernel_pca
+julia> output = kernel_pca(input, kernel;
+                           bandwidth=1, center=false, degree=1, kernel_scale=1,
+                           new_dimensionality=0, nystroem_method=false,
+                           offset=0, sampling="kmeans", verbose=false)
+```
+
+An implementation of Kernel Principal Components Analysis (KPCA).  This can be used to perform nonlinear dimensionality reduction or preprocessing on a given dataset. [Detailed documentation](#kernel_pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`Float64`](#doc_Float64) | Bandwidth, for 'gaussian' and 'laplacian' kernels. | `1` |
+| `center` | [`Bool`](#doc_Bool) | If set, the transformed data will be centered about the origin. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `degree` | [`Float64`](#doc_Float64) | Degree of polynomial, for 'polynomial' kernel. | `1` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform KPCA on. | `**--**` |
+| `kernel` | [`String`](#doc_String) | The kernel to use; see the above documentation for the list of usable kernels. | `**--**` |
+| `kernel_scale` | [`Float64`](#doc_Float64) | Scale, for 'hyptan' kernel. | `1` |
+| `new_dimensionality` | [`Int`](#doc_Int) | If not 0, reduce the dimensionality of the output dataset by ignoring the dimensions with the smallest eigenvalues. | `0` |
+| `nystroem_method` | [`Bool`](#doc_Bool) | If set, the Nystroem method will be used. | `false` |
+| `offset` | [`Float64`](#doc_Float64) | Offset, for 'hyptan' and 'polynomial' kernels. | `0` |
+| `sampling` | [`String`](#doc_String) | Sampling scheme to use for the Nystroem method: 'kmeans', 'random', 'ordered' | `"kmeans"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #kernel_pca_detailed-documentation }
+
+This program performs Kernel Principal Components Analysis (KPCA) on the specified dataset with the specified kernel.  This will transform the data onto the kernel principal components, and optionally reduce the dimensionality by ignoring the kernel principal components with the smallest eigenvalues.
+
+For the case where a linear kernel is used, this reduces to regular PCA.
+
+The kernels that are supported are listed below:
+
+ * 'linear': the standard linear dot product (same as normal PCA):
+    `K(x, y) = x^T y`
+
+ * 'gaussian': a Gaussian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\| ^ 2) / (2 * (bandwidth ^ 2)))`
+
+ * 'polynomial': polynomial kernel; requires offset and degree:
+    `K(x, y) = (x^T y + offset) ^ degree`
+
+ * 'hyptan': hyperbolic tangent kernel; requires scale and offset:
+    `K(x, y) = tanh(scale * (x^T y) + offset)`
+
+ * 'laplacian': Laplacian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\|) / bandwidth)`
+
+ * 'epanechnikov': Epanechnikov kernel; requires bandwidth:
+    `K(x, y) = max(0, 1 - \|\| x - y \|\|^2 / bandwidth^2)`
+
+ * 'cosine': cosine distance:
+    `K(x, y) = 1 - (x^T y) / (\|\| x \|\| * \|\| y \|\|)`
+
+The parameters for each of the kernels should be specified with the options `bandwidth`, `kernel_scale`, `offset`, or `degree` (or a combination of those parameters).
+
+Optionally, the Nystroem method ("Using the Nystroem method to speed up kernel machines", 2001) can be used to calculate the kernel matrix by specifying the `nystroem_method` parameter. This approach works by using a subset of the data as basis to reconstruct the kernel matrix; to specify the sampling scheme, the `sampling` parameter is used.  The sampling scheme for the Nystroem method can be chosen from the following list: 'kmeans', 'random', 'ordered'.
+
+### Example
+For example, the following command will perform KPCA on the dataset ``input`` using the Gaussian kernel, and saving the transformed data to ``transformed``: 
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> transformed = kernel_pca(input, "gaussian")
+```
+
+### See also
+
+ - [Kernel principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Kernel_principal_component_analysis)
+ - [Kernel Principal Component Analysis (pdf)](https://pca.narod.ru/scholkopf_kernel.pdf)
+ - [KernelPCA class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+
+## kmeans()
+{: #kmeans }
+
+#### K-Means Clustering
+{: #kmeans_descr }
+
+```julia
+julia> using mlpack: kmeans
+julia> centroid, output = kmeans(clusters,
+                                 input; algorithm="naive",
+                                 allow_empty_clusters=false, in_place=false,
+                                 initial_centroids=zeros(0, 0),
+                                 kill_empty_clusters=false,
+                                 kmeans_plus_plus=false, labels_only=false,
+                                 max_iterations=1000, percentage=0.02,
+                                 refined_start=false, samplings=100, seed=0,
+                                 verbose=false)
+```
+
+An implementation of several strategies for efficient k-means clustering. Given a dataset and a value of k, this computes and returns a k-means clustering on that data. [Detailed documentation](#kmeans_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`String`](#doc_String) | Algorithm to use for the Lloyd iteration ('naive', 'pelleg-moore', 'elkan', 'hamerly', 'dualtree', or 'dualtree-covertree'). | `"naive"` |
+| `allow_empty_clusters` | [`Bool`](#doc_Bool) | Allow empty clusters to be persist. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `clusters` | [`Int`](#doc_Int) | Number of clusters to find (0 autodetects from initial centroids). | `**--**` |
+| `in_place` | [`Bool`](#doc_Bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden. (Do not use in Python.) | `false` |
+| `initial_centroids` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Start with the specified initial centroids. | `zeros(0, 0)` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform clustering on. | `**--**` |
+| `kill_empty_clusters` | [`Bool`](#doc_Bool) | Remove empty clusters when they occur. | `false` |
+| `kmeans_plus_plus` | [`Bool`](#doc_Bool) | Use the k-means++ initialization strategy to choose initial points. | `false` |
+| `labels_only` | [`Bool`](#doc_Bool) | Only output labels into output file. | `false` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations before k-means terminates. | `1000` |
+| `percentage` | [`Float64`](#doc_Float64) | Percentage of dataset to use for each refined start sampling (use when --refined_start is specified). | `0.02` |
+| `refined_start` | [`Bool`](#doc_Bool) | Use the refined initial point strategy by Bradley and Fayyad to choose initial points. | `false` |
+| `samplings` | [`Int`](#doc_Int) | Number of samplings to perform for refined start (use when --refined_start is specified). | `100` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If specified, the centroids of each cluster will  be written to the given file. | 
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to store output labels or labeled data to. | 
+
+### Detailed documentation
+{: #kmeans_detailed-documentation }
+
+This program performs K-Means clustering on the given dataset.  It can return the learned cluster assignments, and the centroids of the clusters.  Empty clusters are not allowed by default; when a cluster becomes empty, the point furthest from the centroid of the cluster with maximum variance is taken to fill that cluster.
+
+Optionally, the strategy to choose initial centroids can be specified.  The k-means++ algorithm can be used to choose initial centroids with the `kmeans_plus_plus` parameter.  The Bradley and Fayyad approach ("Refining initial points for k-means clustering", 1998) can be used to select initial points by specifying the `refined_start` parameter.  This approach works by taking random samplings of the dataset; to specify the number of samplings, the `samplings` parameter is used, and to specify the percentage of the dataset to be used in each sample, the `percentage` parameter is used (it should be a value between 0.0 and 1.0).
+
+There are several options available for the algorithm used for each Lloyd iteration, specified with the `algorithm`  option.  The standard O(kN) approach can be used ('naive').  Other options include the Pelleg-Moore tree-based algorithm ('pelleg-moore'), Elkan's triangle-inequality based algorithm ('elkan'), Hamerly's modification to Elkan's algorithm ('hamerly'), the dual-tree k-means algorithm ('dualtree'), and the dual-tree k-means algorithm using the cover tree ('dualtree-covertree').
+
+The behavior for when an empty cluster is encountered can be modified with the `allow_empty_clusters` option.  When this option is specified and there is a cluster owning no points at the end of an iteration, that cluster's centroid will simply remain in its position from the previous iteration. If the `kill_empty_clusters` option is specified, then when a cluster owns no points at the end of an iteration, the cluster centroid is simply filled with DBL_MAX, killing it and effectively reducing k for the rest of the computation.  Note that the default option when neither empty cluster option is specified can be time-consuming to calculate; therefore, specifying either of these parameters will often accelerate runtime.
+
+Initial clustering assignments may be specified using the `initial_centroids` parameter, and the maximum number of iterations may be specified with the `max_iterations` parameter.
+
+### Example
+As an example, to use Hamerly's algorithm to perform k-means clustering with k=10 on the dataset ``data``, saving the centroids to ``centroids`` and the assignments for each point to ``assignments``, the following command could be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> centroids, assignments = kmeans(10, data)
+```
+
+To run k-means on that same dataset with initial centroids specified in ``initial`` with a maximum of 500 iterations, storing the output centroids in ``final`` the following command may be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> initial = CSV.read("initial.csv")
+julia> final, _ = kmeans(10, data; initial_centroids=initial,
+            max_iterations=500)
+```
+
+### See also
+
+ - [dbscan()](#dbscan)
+ - [k-means++](https://en.wikipedia.org/wiki/K-means%2B%2B)
+ - [Using the triangle inequality to accelerate k-means (pdf)](https://cdn.aaai.org/ICML/2003/ICML03-022.pdf)
+ - [Making k-means even faster (pdf)](https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12)
+ - [Accelerating exact k-means algorithms with geometric reasoning (pdf)](http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/ftp/usr0/ftp/2000/CMU-CS-00-105.pdf)
+ - [A dual-tree algorithm for fast k-means clustering with large k (pdf)](http://www.ratml.org/pub/pdf/2017dual.pdf)
+ - [KMeans class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kmeans/kmeans.hpp)
+
+## lars()
+{: #lars }
+
+#### LARS
+{: #lars_descr }
+
+```julia
+julia> using mlpack: lars
+julia> output_model, output_predictions = lars( ; input=zeros(0, 0),
+          input_model=nothing, lambda1=0, lambda2=0, no_intercept=false,
+          no_normalize=false, responses=zeros(0, 0), test=zeros(0, 0),
+          use_cholesky=false, verbose=false)
+```
+
+An implementation of Least Angle Regression (Stagewise/laSso), also known as LARS.  This can train a LARS/LASSO/Elastic Net model and use that model or a pre-trained model to output regression predictions for a test set. [Detailed documentation](#lars_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of covariates (X). | `zeros(0, 0)` |
+| `input_model` | [`LARS`](#doc_model) | Trained LARS model to use. | `nothing` |
+| `lambda1` | [`Float64`](#doc_Float64) | Regularization parameter for l1-norm penalty. | `0` |
+| `lambda2` | [`Float64`](#doc_Float64) | Regularization parameter for l2-norm penalty. | `0` |
+| `no_intercept` | [`Bool`](#doc_Bool) | Do not fit an intercept in the model. | `false` |
+| `no_normalize` | [`Bool`](#doc_Bool) | Do not normalize data to unit variance before modeling. | `false` |
+| `responses` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of responses/observations (y). | `zeros(0, 0)` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing points to regress on (test points). | `zeros(0, 0)` |
+| `use_cholesky` | [`Bool`](#doc_Bool) | Use Cholesky decomposition during computation rather than explicitly computing the full Gram matrix. | `false` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LARS`](#doc_model) | Output LARS model. | 
+| `output_predictions` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #lars_detailed-documentation }
+
+An implementation of LARS: Least Angle Regression (Stagewise/laSso).  This is a stage-wise homotopy-based algorithm for L1-regularized linear regression (LASSO) and L1+L2-regularized linear regression (Elastic Net).
+
+This program is able to train a LARS/LASSO/Elastic Net model or load a model from file, output regression predictions for a test set, and save the trained model to a file.  The LARS algorithm is described in more detail below:
+
+Let X be a matrix where each row is a point and each column is a dimension, and let y be a vector of targets.
+
+The Elastic Net problem is to solve
+
+  min_beta 0.5 \|\| X * beta - y \|\|_2^2 + lambda_1 \|\|beta\|\|_1 +
+      0.5 lambda_2 \|\|beta\|\|_2^2
+
+If lambda1 > 0 and lambda2 = 0, the problem is the LASSO.
+If lambda1 > 0 and lambda2 > 0, the problem is the Elastic Net.
+If lambda1 = 0 and lambda2 > 0, the problem is ridge regression.
+If lambda1 = 0 and lambda2 = 0, the problem is unregularized linear regression.
+
+For efficiency reasons, it is not recommended to use this algorithm with `lambda1` = 0.  In that case, use the 'linear_regression' program, which implements both unregularized linear regression and ridge regression.
+
+To train a LARS/LASSO/Elastic Net model, the `input` and `responses` parameters must be given.  The `lambda1`, `lambda2`, and `use_cholesky` parameters control the training options.  A trained model can be saved with the `output_model`.  If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `output_predictions` output parameter.
+
+### Example
+For example, the following command trains a model on the data ``data`` and responses ``responses`` with lambda1 set to 0.4 and lambda2 set to 0 (so, LASSO is being solved), and then the model is saved to ``lasso_model``:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> responses = CSV.read("responses.csv")
+julia> lasso_model, _ = lars(input=data, lambda1=0.4, lambda2=0,
+            responses=responses)
+```
+
+The following command uses the ``lasso_model`` to provide predicted responses for the data ``test`` and save those responses to ``test_predictions``: 
+
+```julia
+julia> using CSV
+julia> test = CSV.read("test.csv")
+julia> _, test_predictions = lars(input_model=lasso_model,
+            test=test)
+```
+
+### See also
+
+ - [linear_regression()](#linear_regression)
+ - [Least angle regression (pdf)](https://mlpack.org/papers/lars.pdf)
+ - [LARS C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lars/lars.hpp)
+
+## linear_svm()
+{: #linear_svm }
+
+#### Linear SVM is an L2-regularized support vector machine.
+{: #linear_svm_descr }
+
+```julia
+julia> using mlpack: linear_svm
+julia> output_model, predictions, probabilities = linear_svm( ;
+          delta=1, epochs=50, input_model=nothing, labels=Int[], lambda=0.0001,
+          max_iterations=10000, no_intercept=false, num_classes=0,
+          optimizer="lbfgs", seed=0, shuffle=false, step_size=0.01,
+          test=zeros(0, 0), test_labels=Int[], tolerance=1e-10,
+          training=zeros(0, 0), verbose=false)
+```
+
+An implementation of linear SVM for multiclass classification. Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#linear_svm_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `delta` | [`Float64`](#doc_Float64) | Margin of difference between correct class and other classes. | `1` |
+| `epochs` | [`Int`](#doc_Int) | Maximum number of full epochs over dataset for psgd | `50` |
+| `input_model` | [`LinearSVMModel`](#doc_model) | Existing model (parameters). | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | A matrix containing labels (0 or 1) for the points in the training set (y). | `Int[]` |
+| `lambda` | [`Float64`](#doc_Float64) | L2-regularization parameter for training. | `0.0001` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `no_intercept` | [`Bool`](#doc_Bool) | Do not add the intercept term to the model. | `false` |
+| `num_classes` | [`Int`](#doc_Int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `optimizer` | [`String`](#doc_String) | Optimizer to use for training ('lbfgs' or 'psgd'). | `"lbfgs"` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `shuffle` | [`Bool`](#doc_Bool) | Don't shuffle the order in which data points are visited for parallel SGD. | `false` |
+| `step_size` | [`Float64`](#doc_Float64) | Step size for parallel SGD optimizer. | `0.01` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing test dataset. | `zeros(0, 0)` |
+| `test_labels` | [`Int vector-like`](#doc_Int_vector_like) | Matrix containing test labels. | `Int[]` |
+| `tolerance` | [`Float64`](#doc_Float64) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the training set (the matrix of predictors, X). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LinearSVMModel`](#doc_model) | Output for trained linear svm model. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #linear_svm_detailed-documentation }
+
+An implementation of linear SVMs that uses either L-BFGS or parallel SGD (stochastic gradient descent) to train the model.
+
+This program allows loading a linear SVM model (via the `input_model` parameter) or training a linear SVM model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained linear SVM model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate vector of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda` option, and the number of classes can be manually specified with the `num_classes`and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.Margin of difference between correct class and other classes can be specified with the `delta` option.The optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'psgd' (parallel stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the parallel SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer and the maximum number of epochs (specified with `epochs`). If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+Optionally, the model can be used to predict the labels for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing linear SVM model is given with the `input_model` parameter.  The output predictions from the linear SVM model may be saved with the `predictions` parameter.
+
+### Example
+As an example, to train a LinaerSVM on the data '``data``' with labels '``labels``' with L2 regularization of 0.1, saving the model to '``lsvm_model``', the following command may be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> lsvm_model, _, _ = linear_svm(delta=1, labels=labels,
+            lambda=0.1, num_classes=0, training=data)
+```
+
+Then, to use that model to predict classes for the dataset '``test``', storing the output predictions in '``predictions``', the following command may be used: 
+
+```julia
+julia> using CSV
+julia> test = CSV.read("test.csv")
+julia> _, predictions, _ = linear_svm(input_model=lsvm_model,
+            test=test)
+```
+
+### See also
+
+ - [random_forest()](#random_forest)
+ - [logistic_regression()](#logistic_regression)
+ - [LinearSVM on Wikipedia](https://en.wikipedia.org/wiki/Support-vector_machine)
+ - [LinearSVM C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_svm/linear_svm.hpp)
+
+## lmnn()
+{: #lmnn }
+
+#### Large Margin Nearest Neighbors (LMNN)
+{: #lmnn_descr }
+
+```julia
+julia> using mlpack: lmnn
+julia> centered_data, output, transformed_data = lmnn(input;
+          batch_size=50, center=false, distance=zeros(0, 0), k=1, labels=Int[],
+          linear_scan=false, max_iterations=100000, normalize=false,
+          optimizer="amsgrad", passes=50, print_accuracy=false, range=1, rank=0,
+          regularization=0.5, seed=0, step_size=0.01, tolerance=1e-07,
+          verbose=false)
+```
+
+An implementation of Large Margin Nearest Neighbors (LMNN), a distance learning technique.  Given a labeled dataset, this learns a transformation of the data that improves k-nearest-neighbor performance; this can be useful as a preprocessing step. [Detailed documentation](#lmnn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`Int`](#doc_Int) | Batch size for mini-batch SGD. | `50` |
+| `center` | [`Bool`](#doc_Bool) | Perform mean-centering on the dataset. It is useful when the centroid of the data is far from the origin. | `false` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `distance` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Initial distance matrix to be used as starting point | `zeros(0, 0)` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to run LMNN on. | `**--**` |
+| `k` | [`Int`](#doc_Int) | Number of target neighbors to use for each datapoint. | `1` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels for input dataset. | `Int[]` |
+| `linear_scan` | [`Bool`](#doc_Bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `false` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations for L-BFGS (0 indicates no limit). | `100000` |
+| `normalize` | [`Bool`](#doc_Bool) | Use a normalized starting point for optimization. Itis useful for when points are far apart, or when SGD is returning NaN. | `false` |
+| `optimizer` | [`String`](#doc_String) | Optimizer to use; 'amsgrad', 'bbsgd', 'sgd', or 'lbfgs'. | `"amsgrad"` |
+| `passes` | [`Int`](#doc_Int) | Maximum number of full passes over dataset for AMSGrad, BB_SGD and SGD. | `50` |
+| `print_accuracy` | [`Bool`](#doc_Bool) | Print accuracies on initial and transformed dataset | `false` |
+| `range` | [`Int`](#doc_Int) | Number of iterations after which impostors needs to be recalculated | `1` |
+| `rank` | [`Int`](#doc_Int) | Rank of distance matrix to be optimized.  | `0` |
+| `regularization` | [`Float64`](#doc_Float64) | Regularization for LMNN objective function  | `0.5` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`Float64`](#doc_Float64) | Step size for AMSGrad, BB_SGD and SGD (alpha). | `0.01` |
+| `tolerance` | [`Float64`](#doc_Float64) | Maximum tolerance for termination of AMSGrad, BB_SGD, SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centered_data` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output matrix for mean-centered dataset. | 
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output matrix for learned distance matrix. | 
+| `transformed_data` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output matrix for transformed dataset. | 
+
+### Detailed documentation
+{: #lmnn_detailed-documentation }
+
+This program implements Large Margin Nearest Neighbors, a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset.  The method employes the strategy of reducing distance between similar labeled data points (a.k.a target neighbors) and increasing distance between differently labeled points (a.k.a impostors) using standard optimization techniques over the gradient of the distance between data points.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input`), or alternatively as a separate matrix (specified with `labels`).  Additionally, a starting point for optimization (specified with `distance`can be given, having (r x d) dimensionality.  Here r should satisfy 1 <= r <= d, Consequently a Low-Rank matrix will be optimized. Alternatively, Low-Rank distance can be learned by specifying the `rank`parameter (A Low-Rank matrix with uniformly distributed values will be used as initial learning point). 
+
+The program also requires number of targets neighbors to work with ( specified with `k`), A regularization parameter can also be passed, It acts as a trade of between the pulling and pushing terms (specified with `regularization`), In addition, this implementation of LMNN includes a parameter to decide the interval after which impostors must be re-calculated (specified with `range`).
+
+Output can either be the learned distance matrix (specified with `output`), or the transformed dataset  (specified with `transformed_data`), or both. Additionally mean-centered dataset (specified with `centered_data`) can be accessed given mean-centering (specified with `center`) is performed on the dataset. Accuracy on initial dataset and final transformed dataset can be printed by specifying the `print_accuracy`parameter. 
+
+This implementation of LMNN uses AdaGrad, BigBatch_SGD, stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer. 
+
+AdaGrad, specified by the value 'adagrad' for the parameter `optimizer`, uses maximum of past squared gradients. It primarily on six parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`). Inaddition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+BigBatch_SGD, specified by the value 'bbsgd' for the parameter `optimizer`, depends primarily on four parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. Furthermore, mean-centering can be performed on the dataset by specifying the `center`parameter. 
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `max_iterations`, `tolerance`(the optimization is terminated when the gradient norm is below this value).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.  In addition, a normalized starting point can be used by specifying the `normalize` parameter.
+
+By default, the AMSGrad optimizer is used.
+
+### Example
+Example - Let's say we want to learn distance on iris dataset with number of targets as 3 using BigBatch_SGD optimizer. A simple call for the same will look like: 
+
+```julia
+julia> using CSV
+julia> iris = CSV.read("iris.csv")
+julia> iris_labels = CSV.read("iris_labels.csv"; type=Int)
+julia> _, output, _ = lmnn(iris; k=3, labels=iris_labels,
+            optimizer="bbsgd")
+```
+
+An another program call making use of range & regularization parameter with dataset having labels as last column can be made as: 
+
+```julia
+julia> using CSV
+julia> letter_recognition = CSV.read("letter_recognition.csv")
+julia> _, output, _ = lmnn(letter_recognition; k=5, range=10,
+            regularization=0.4)
+```
+
+### See also
+
+ - [nca()](#nca)
+ - [Large margin nearest neighbor on Wikipedia](https://en.wikipedia.org/wiki/Large_margin_nearest_neighbor)
+ - [Distance metric learning for large margin nearest neighbor classification (pdf)](http://papers.nips.cc/paper/2795-distance-metric-learning-for-large-margin-nearest-neighbor-classification.pdf)
+ - [LMNN C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lmnn/lmnn.hpp)
+
+## local_coordinate_coding()
+{: #local_coordinate_coding }
+
+#### Local Coordinate Coding
+{: #local_coordinate_coding_descr }
+
+```julia
+julia> using mlpack: local_coordinate_coding
+julia> codes, dictionary, output_model = local_coordinate_coding( ;
+          atoms=0, initial_dictionary=zeros(0, 0), input_model=nothing,
+          lambda=0, max_iterations=0, normalize=false, seed=0, test=zeros(0, 0),
+          tolerance=0.01, training=zeros(0, 0), verbose=false)
+```
+
+An implementation of Local Coordinate Coding (LCC), a data transformation technique.  Given input data, this transforms each point to be expressed as a linear combination of a few points in the dataset; once an LCC model is trained, it can be used to transform points later also. [Detailed documentation](#local_coordinate_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`Int`](#doc_Int) | Number of atoms in the dictionary. | `0` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `initial_dictionary` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Optional initial dictionary. | `zeros(0, 0)` |
+| `input_model` | [`LocalCoordinateCoding`](#doc_model) | Input LCC model. | `nothing` |
+| `lambda` | [`Float64`](#doc_Float64) | Weighted l1-norm regularization parameter. | `0` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations for LCC (0 indicates no limit). | `0` |
+| `normalize` | [`Bool`](#doc_Bool) | If set, the input data matrix will be normalized before coding. | `false` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Test points to encode. | `zeros(0, 0)` |
+| `tolerance` | [`Float64`](#doc_Float64) | Tolerance for objective function. | `0.01` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of training data (X). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output codes matrix. | 
+| `dictionary` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output dictionary matrix. | 
+| `output_model` | [`LocalCoordinateCoding`](#doc_model) | Output for trained LCC model. | 
+
+### Detailed documentation
+{: #local_coordinate_coding_detailed-documentation }
+
+An implementation of Local Coordinate Coding (LCC), which codes data that approximately lives on a manifold using a variation of l1-norm regularized sparse coding.  Given a dense data matrix X with n points and d dimensions, LCC seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a coding matrix Z with n points in k dimensions.  Because of the regularization method used, the atoms in D should lie close to the manifold on which the data points lie.
+
+The original data matrix X can then be reconstructed as D * Z.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a coding step, which updates the coding matrix Z.
+
+To run this program, the input matrix X must be specified (with -i), along with the number of atoms in the dictionary (-k).  An initial dictionary may also be specified with the `initial_dictionary` parameter.  The l1-norm regularization parameter is specified with the `lambda` parameter.
+
+### Example
+For example, to run LCC on the dataset ``data`` using 200 atoms and an l1-regularization parameter of 0.1, saving the dictionary `dictionary` and the codes into `codes`, use
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> codes, dict, _ = local_coordinate_coding(atoms=200,
+            lambda=0.1, training=data)
+```
+
+The maximum number of iterations may be specified with the `max_iterations` parameter. Optionally, the input data matrix X can be normalized before coding with the `normalize` parameter.
+
+An LCC model may be saved using the `output_model` output parameter.  Then, to encode new points from the dataset ``points`` with the previously saved model ``lcc_model``, saving the new codes to ``new_codes``, the following command can be used:
+
+```julia
+julia> using CSV
+julia> points = CSV.read("points.csv")
+julia> new_codes, _, _ =
+            local_coordinate_coding(input_model=lcc_model, test=points)
+```
+
+### See also
+
+ - [sparse_coding()](#sparse_coding)
+ - [Nonlinear learning using local coordinate coding (pdf)](https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-coordinate-coding.pdf)
+ - [LocalCoordinateCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding.hpp)
+
+## logistic_regression()
+{: #logistic_regression }
+
+#### L2-regularized Logistic Regression and Prediction
+{: #logistic_regression_descr }
+
+```julia
+julia> using mlpack: logistic_regression
+julia> output_model, predictions, probabilities = logistic_regression(
+          ; batch_size=64, decision_boundary=0.5, input_model=nothing,
+          labels=Int[], lambda=0, max_iterations=10000, optimizer="lbfgs",
+          print_training_accuracy=false, step_size=0.01, test=zeros(0, 0),
+          tolerance=1e-10, training=zeros(0, 0), verbose=false)
+```
+
+An implementation of L2-regularized logistic regression for two-class classification.  Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#logistic_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`Int`](#doc_Int) | Batch size for SGD. | `64` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `decision_boundary` | [`Float64`](#doc_Float64) | Decision boundary for prediction; if the logistic function for a point is less than the boundary, the class is taken to be 0; otherwise, the class is 1. | `0.5` |
+| `input_model` | [`LogisticRegression`](#doc_model) | Existing model (parameters). | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | A matrix containing labels (0 or 1) for the points in the training set (y). | `Int[]` |
+| `lambda` | [`Float64`](#doc_Float64) | L2-regularization parameter for training. | `0` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `optimizer` | [`String`](#doc_String) | Optimizer to use for training ('lbfgs' or 'sgd'). | `"lbfgs"` |
+| `print_training_accuracy` | [`Bool`](#doc_Bool) | If set, then the accuracy of the model on the training set will be printed (verbose must also be specified). | `false` |
+| `step_size` | [`Float64`](#doc_Float64) | Step size for SGD optimizer. | `0.01` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing test dataset. | `zeros(0, 0)` |
+| `tolerance` | [`Float64`](#doc_Float64) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the training set (the matrix of predictors, X). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LogisticRegression`](#doc_model) | Output for trained logistic regression model. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #logistic_regression_detailed-documentation }
+
+An implementation of L2-regularized logistic regression using either the L-BFGS optimizer or SGD (stochastic gradient descent).  This solves the regression problem
+
+  y = (1 / 1 + e^-(X * b)).
+
+In this setting, y corresponds to class labels and X corresponds to data.
+
+This program allows loading a logistic regression model (via the `input_model` parameter) or training a logistic regression model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained logistic regression model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda` option, and the optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'sgd' (stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer.  The batch size for SGD is controlled with the `batch_size` parameter. If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+For SGD, an iteration refers to a single point. So to take a single pass over the dataset with SGD, `max_iterations` should be set to the number of points in the dataset.
+
+Optionally, the model can be used to predict the responses for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing logistic regression model is given with the `input_model` parameter.  The output predictions from the logistic regression model may be saved with the `predictions` parameter.
+
+This implementation of logistic regression does not support the general multi-class case but instead only the two-class case.  Any labels must be either 0 or 1.  For more classes, see the softmax regression implementation.
+
+### Example
+As an example, to train a logistic regression model on the data '``data``' with labels '``labels``' with L2 regularization of 0.1, saving the model to '``lr_model``', the following command may be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> lr_model, _, _ = logistic_regression(labels=labels,
+            lambda=0.1, print_training_accuracy=1, training=data)
+```
+
+Then, to use that model to predict classes for the dataset '``test``', storing the output predictions in '``predictions``', the following command may be used: 
+
+```julia
+julia> using CSV
+julia> test = CSV.read("test.csv")
+julia> _, predictions, _ = logistic_regression(input_model=lr_model,
+            test=test)
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+ - [:LogisticRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/logistic_regression/logistic_regression.hpp)
+
+## lsh()
+{: #lsh }
+
+#### K-Approximate-Nearest-Neighbor Search with LSH
+{: #lsh_descr }
+
+```julia
+julia> using mlpack: lsh
+julia> distances, neighbors, output_model = lsh( ; bucket_size=500,
+          hash_width=0, input_model=nothing, k=0, num_probes=0, projections=10,
+          query=zeros(0, 0), reference=zeros(0, 0), second_hash_size=99901,
+          seed=0, tables=30, true_neighbors=zeros(Int, 0, 0), verbose=false)
+```
+
+An implementation of approximate k-nearest-neighbor search with locality-sensitive hashing (LSH).  Given a set of reference points and a set of query points, this will compute the k approximate nearest neighbors of each query point in the reference set; models can be saved for future use. [Detailed documentation](#lsh_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bucket_size` | [`Int`](#doc_Int) | The size of a bucket in the second level hash. | `500` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `hash_width` | [`Float64`](#doc_Float64) | The hash width for the first-level hashing in the LSH preprocessing. By default, the LSH class automatically estimates a hash width for its use. | `0` |
+| `input_model` | [`LSHSearch`](#doc_model) | Input LSH model. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of nearest neighbors to find. | `0` |
+| `num_probes` | [`Int`](#doc_Int) | Number of additional probes for multiprobe LSH; if 0, traditional LSH is used. | `0` |
+| `projections` | [`Int`](#doc_Int) | The number of hash functions for each table | `10` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing query points (optional). | `zeros(0, 0)` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing the reference dataset. | `zeros(0, 0)` |
+| `second_hash_size` | [`Int`](#doc_Int) | The size of the second level hash table. | `99901` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tables` | [`Int`](#doc_Int) | The number of hash tables to be used. | `30` |
+| `true_neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix of true neighbors to compute recall with (the recall is printed when -v is specified). | `zeros(Int, 0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to output distances into. | 
+| `neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to output neighbors into. | 
+| `output_model` | [`LSHSearch`](#doc_model) | Output for trained LSH model. | 
+
+### Detailed documentation
+{: #lsh_detailed-documentation }
+
+This program will calculate the k approximate-nearest-neighbors of a set of points using locality-sensitive hashing. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. 
+
+### Example
+For example, the following will return 5 neighbors from the data for each point in ``input`` and store the distances in ``distances`` and the neighbors in ``neighbors``:
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> distances, neighbors, _ = lsh(k=5, reference=input)
+```
+
+The output is organized such that row i and column j in the neighbors output corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output file corresponds to the distance between those two points.
+
+Because this is approximate-nearest-neighbors search, results may be different from run to run.  Thus, the `seed` parameter can be specified to set the random seed.
+
+This program also has many other parameters to control its functionality; see the parameter-specific documentation for more information.
+
+### See also
+
+ - [knn()](#knn)
+ - [krann()](#krann)
+ - [Locality-sensitive hashing on Wikipedia](https://en.wikipedia.org/wiki/Locality-sensitive_hashing)
+ - [Locality-sensitive hashing scheme based on p-stable  distributions(pdf)](https://www.mlpack.org/papers/lsh.pdf)
+ - [LSHSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lsh/lsh.hpp)
+
+## mean_shift()
+{: #mean_shift }
+
+#### Mean Shift Clustering
+{: #mean_shift_descr }
+
+```julia
+julia> using mlpack: mean_shift
+julia> centroid, output = mean_shift(input; force_convergence=false,
+          in_place=false, labels_only=false, max_iterations=1000, radius=0,
+          verbose=false)
+```
+
+A fast implementation of mean-shift clustering using dual-tree range search.  Given a dataset, this uses the mean shift algorithm to produce and return a clustering of the data. [Detailed documentation](#mean_shift_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `force_convergence` | [`Bool`](#doc_Bool) | If specified, the mean shift algorithm will continue running regardless of max_iterations until the clusters converge. | `false` |
+| `in_place` | [`Bool`](#doc_Bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden.  (Do not use with Python.) | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform clustering on. | `**--**` |
+| `labels_only` | [`Bool`](#doc_Bool) | If specified, only the output labels will be written to the file specified by --output_file. | `false` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations before mean shift terminates. | `1000` |
+| `radius` | [`Float64`](#doc_Float64) | If the distance between two centroids is less than the given radius, one will be removed.  A radius of 0 or less means an estimate will be calculated and used for the radius. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | If specified, the centroids of each cluster will be written to the given matrix. | 
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to write output labels or labeled data to. | 
+
+### Detailed documentation
+{: #mean_shift_detailed-documentation }
+
+This program performs mean shift clustering on the given dataset, storing the learned cluster assignments either as a column of labels in the input dataset or separately.
+
+The input dataset should be specified with the `input` parameter, and the radius used for search can be specified with the `radius` parameter.  The maximum number of iterations before algorithm termination is controlled with the `max_iterations` parameter.
+
+The output labels may be saved with the `output` output parameter and the centroids of each cluster may be saved with the `centroid` output parameter.
+
+### Example
+For example, to run mean shift clustering on the dataset ``data`` and store the centroids to ``centroids``, the following command may be used: 
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> centroids, _ = mean_shift(data)
+```
+
+### See also
+
+ - [kmeans()](#kmeans)
+ - [dbscan()](#dbscan)
+ - [Mean shift on Wikipedia](https://en.wikipedia.org/wiki/Mean_shift)
+ - [Mean Shift, Mode Seeking, and Clustering (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=1c168275c59ba382588350ee1443537f59978183)
+ - [mlpack::mean_shift::MeanShift C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/mean_shift/mean_shift.hpp)
+
+## nbc()
+{: #nbc }
+
+#### Parametric Naive Bayes Classifier
+{: #nbc_descr }
+
+```julia
+julia> using mlpack: nbc
+julia> output, output_model, output_probs, predictions, probabilities
+          = nbc( ; incremental_variance=false, input_model=nothing,
+          labels=Int[], test=zeros(0, 0), training=zeros(0, 0), verbose=false)
+```
+
+An implementation of the Naive Bayes Classifier, used for classification. Given labeled data, an NBC model can be trained and saved, or, a pre-trained model can be used for classification. [Detailed documentation](#nbc_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `incremental_variance` | [`Bool`](#doc_Bool) | The variance of each class will be calculated incrementally. | `false` |
+| `input_model` | [`NBCModel`](#doc_model) | Input Naive Bayes model. | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | A file containing labels for the training set. | `Int[]` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the test set. | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the training set. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Int vector-like`](#doc_Int_vector_like) | The matrix in which the predicted labels for the test set will be written (deprecated). | 
+| `output_model` | [`NBCModel`](#doc_model) | File to save trained Naive Bayes model to. | 
+| `output_probs` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The matrix in which the predicted probability of labels for the test set will be written (deprecated). | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | The matrix in which the predicted labels for the test set will be written. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | The matrix in which the predicted probability of labels for the test set will be written. | 
+
+### Detailed documentation
+{: #nbc_detailed-documentation }
+
+This program trains the Naive Bayes classifier on the given labeled training set, or loads a model from the given model file, and then may use that trained model to classify the points in a given test set.
+
+The training set is specified with the `training` parameter.  Labels may be either the last row of the training set, or alternately the `labels` parameter may be specified to pass a separate matrix of labels.
+
+If training is not desired, a pre-existing model may be loaded with the `input_model` parameter.
+
+
+
+The `incremental_variance` parameter can be used to force the training to use an incremental algorithm for calculating variance.  This is slower, but can help avoid loss of precision in some cases.
+
+If classifying a test set is desired, the test set may be specified with the `test` parameter, and the classifications may be saved with the `predictions`predictions  parameter.  If saving the trained model is desired, this may be done with the `output_model` output parameter.
+
+Note: the `output` and `output_probs` parameters are deprecated and will be removed in mlpack 4.0.0.  Use `predictions` and `probabilities` instead.
+
+### Example
+For example, to train a Naive Bayes classifier on the dataset ``data`` with labels ``labels`` and save the model to ``nbc_model``, the following command may be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> _, nbc_model, _, _, _ = nbc(labels=labels, training=data)
+```
+
+Then, to use ``nbc_model`` to predict the classes of the dataset ``test_set`` and save the predicted classes to ``predictions``, the following command may be used:
+
+```julia
+julia> using CSV
+julia> test_set = CSV.read("test_set.csv")
+julia> predictions, _, _, _, _ = nbc(input_model=nbc_model,
+            test=test_set)
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+ - [NaiveBayesClassifier C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp)
+
+## nca()
+{: #nca }
+
+#### Neighborhood Components Analysis (NCA)
+{: #nca_descr }
+
+```julia
+julia> using mlpack: nca
+julia> output = nca(input; armijo_constant=0.0001,
+                    batch_size=50, labels=Int[], linear_scan=false,
+                    max_iterations=500000, max_line_search_trials=50,
+                    max_step=1e+20, min_step=1e-20, normalize=false,
+                    num_basis=5, optimizer="sgd", seed=0, step_size=0.01,
+                    tolerance=1e-07, verbose=false, wolfe=0.9)
+```
+
+An implementation of neighborhood components analysis, a distance learning technique that can be used for preprocessing.  Given a labeled dataset, this uses NCA, which seeks to improve the k-nearest-neighbor classification, and returns the learned distance metric. [Detailed documentation](#nca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `armijo_constant` | [`Float64`](#doc_Float64) | Armijo constant for L-BFGS. | `0.0001` |
+| `batch_size` | [`Int`](#doc_Int) | Batch size for mini-batch SGD. | `50` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to run NCA on. | `**--**` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels for input dataset. | `Int[]` |
+| `linear_scan` | [`Bool`](#doc_Bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `false` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations for SGD or L-BFGS (0 indicates no limit). | `500000` |
+| `max_line_search_trials` | [`Int`](#doc_Int) | Maximum number of line search trials for L-BFGS. | `50` |
+| `max_step` | [`Float64`](#doc_Float64) | Maximum step of line search for L-BFGS. | `1e+20` |
+| `min_step` | [`Float64`](#doc_Float64) | Minimum step of line search for L-BFGS. | `1e-20` |
+| `normalize` | [`Bool`](#doc_Bool) | Use a normalized starting point for optimization. This is useful for when points are far apart, or when SGD is returning NaN. | `false` |
+| `num_basis` | [`Int`](#doc_Int) | Number of memory points to be stored for L-BFGS. | `5` |
+| `optimizer` | [`String`](#doc_String) | Optimizer to use; 'sgd' or 'lbfgs'. | `"sgd"` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`Float64`](#doc_Float64) | Step size for stochastic gradient descent (alpha). | `0.01` |
+| `tolerance` | [`Float64`](#doc_Float64) | Maximum tolerance for termination of SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `wolfe` | [`Float64`](#doc_Float64) | Wolfe condition parameter for L-BFGS. | `0.9` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Output matrix for learned distance matrix. | 
+
+### Detailed documentation
+{: #nca_detailed-documentation }
+
+This program implements Neighborhood Components Analysis, both a linear dimensionality reduction technique and a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset by scaling the dimensions.  The method is nonparametric, and does not require a value of k.  It works by using stochastic ("soft") neighbor assignments and using optimization techniques over the gradient of the accuracy of the neighbor assignments.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input`), or alternatively as a separate matrix (specified with `labels`).
+
+This implementation of NCA uses stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer.  These optimizers do not guarantee global convergence for a nonconvex objective function (NCA's objective function is nonconvex), so the final results could depend on the random seed or other optimizer parameters.
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of iterations (specified with `max_iterations`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter, which is necessary if many warnings of the form 'Denominator of p_i is 0!' are given.  Tuning the step size can be a tedious affair.  In general, the step size is too large if the objective is not mostly uniformly decreasing, or if zero-valued denominator warnings are being issued.  The step size is too small if the objective is changing very slowly.  Setting the termination condition can be done easily once a good step size parameter is found; either increase the maximum iterations to a large number and allow SGD to find a minimum, or set the maximum iterations to 0 (allowing infinite iterations) and set the tolerance (specified by `tolerance`) to define the maximum allowed difference between objectives for SGD to terminate.  Be careful---setting the tolerance instead of the maximum iterations can take a very long time and may actually never converge due to the properties of the SGD optimizer. Note that a single iteration of SGD refers to a single point, so to take a single pass over the dataset, set the value of the `max_iterations` parameter equal to the number of points in the dataset.
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `num_basis` (specifies the number of memory points used by L-BFGS), `max_iterations`, `armijo_constant`, `wolfe`, `tolerance` (the optimization is terminated when the gradient norm is below this value), `max_line_search_trials`, `min_step`, and `max_step` (which both refer to the line search routine).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.
+
+By default, the SGD optimizer is used.
+
+### See also
+
+ - [lmnn()](#lmnn)
+ - [Neighbourhood components analysis on Wikipedia](https://en.wikipedia.org/wiki/Neighbourhood_components_analysis)
+ - [Neighbourhood components analysis (pdf)](http://papers.nips.cc/paper/2566-neighbourhood-components-analysis.pdf)
+ - [NCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/nca/nca.hpp)
+
+## knn()
+{: #knn }
+
+#### k-Nearest-Neighbors Search
+{: #knn_descr }
+
+```julia
+julia> using mlpack: knn
+julia> distances, neighbors, output_model = knn( ;
+          algorithm="dual_tree", epsilon=0, input_model=nothing, k=0,
+          leaf_size=20, query=zeros(0, 0), random_basis=false,
+          reference=zeros(0, 0), rho=0.7, seed=0, tau=0, tree_type="kd",
+          true_distances=zeros(0, 0), true_neighbors=zeros(Int, 0, 0),
+          verbose=false)
+```
+
+An implementation of k-nearest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#knn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`String`](#doc_String) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `epsilon` | [`Float64`](#doc_Float64) | If specified, will do approximate nearest neighbor search with given relative error. | `0` |
+| `input_model` | [`KNNModel`](#doc_model) | Pre-trained kNN model. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`Int`](#doc_Int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, spill trees, and octrees). | `20` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing query points (optional). | `zeros(0, 0)` |
+| `random_basis` | [`Bool`](#doc_Bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing the reference dataset. | `zeros(0, 0)` |
+| `rho` | [`Float64`](#doc_Float64) | Balance threshold (only valid for spill trees). | `0.7` |
+| `seed` | [`Int`](#doc_Int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tau` | [`Float64`](#doc_Float64) | Overlapping size (only valid for spill trees). | `0` |
+| `tree_type` | [`String`](#doc_String) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'spill', 'oct'. | `"kd"` |
+| `true_distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `zeros(0, 0)` |
+| `true_neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `zeros(Int, 0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to output distances into. | 
+| `neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to output neighbors into. | 
+| `output_model` | [`KNNModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #knn_detailed-documentation }
+
+This program will calculate the k-nearest-neighbors of a set of points using kd-trees or cover trees (cover tree support is experimental and may be slow). You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following command will calculate the 5 nearest neighbors of each point in ``input`` and store the distances in ``distances`` and the neighbors in ``neighbors``: 
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> distances, neighbors, _ = knn(k=5, reference=input)
+```
+
+The output is organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output matrix corresponds to the distance between those two points.
+
+### See also
+
+ - [lsh()](#lsh)
+ - [krann()](#krann)
+ - [kfn()](#kfn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## kfn()
+{: #kfn }
+
+#### k-Furthest-Neighbors Search
+{: #kfn_descr }
+
+```julia
+julia> using mlpack: kfn
+julia> distances, neighbors, output_model = kfn( ;
+          algorithm="dual_tree", epsilon=0, input_model=nothing, k=0,
+          leaf_size=20, percentage=1, query=zeros(0, 0), random_basis=false,
+          reference=zeros(0, 0), seed=0, tree_type="kd", true_distances=zeros(0,
+          0), true_neighbors=zeros(Int, 0, 0), verbose=false)
+```
+
+An implementation of k-furthest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k furthest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`String`](#doc_String) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `epsilon` | [`Float64`](#doc_Float64) | If specified, will do approximate furthest neighbor search with given relative error. Must be in the range [0,1). | `0` |
+| `input_model` | [`KFNModel`](#doc_model) | Pre-trained kFN model. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of furthest neighbors to find. | `0` |
+| `leaf_size` | [`Int`](#doc_Int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `percentage` | [`Float64`](#doc_Float64) | If specified, will do approximate furthest neighbor search. Must be in the range (0,1] (decimal form). Resultant neighbors will be at least (p*100) % of the distance as the true furthest neighbor. | `1` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing query points (optional). | `zeros(0, 0)` |
+| `random_basis` | [`Bool`](#doc_Bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing the reference dataset. | `zeros(0, 0)` |
+| `seed` | [`Int`](#doc_Int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tree_type` | [`String`](#doc_String) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `true_distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `zeros(0, 0)` |
+| `true_neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `zeros(Int, 0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to output distances into. | 
+| `neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to output neighbors into. | 
+| `output_model` | [`KFNModel`](#doc_model) | If specified, the kFN model will be output here. | 
+
+### Detailed documentation
+{: #kfn_detailed-documentation }
+
+This program will calculate the k-furthest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following will calculate the 5 furthest neighbors of eachpoint in ``input`` and store the distances in ``distances`` and the neighbors in ``neighbors``: 
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> distances, neighbors, _ = kfn(k=5, reference=input)
+```
+
+The output files are organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th furthest neighbor from the point in the query set with index i.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [approx_kfn()](#approx_kfn)
+ - [knn()](#knn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## nmf()
+{: #nmf }
+
+#### Non-negative Matrix Factorization
+{: #nmf_descr }
+
+```julia
+julia> using mlpack: nmf
+julia> h, w = nmf(input, rank; initial_h=zeros(0, 0),
+                  initial_w=zeros(0, 0), max_iterations=10000,
+                  min_residue=1e-05, seed=0, update_rules="multdist",
+                  verbose=false)
+```
+
+An implementation of non-negative matrix factorization.  This can be used to decompose an input dataset into two low-rank non-negative components. [Detailed documentation](#nmf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `initial_h` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Initial H matrix. | `zeros(0, 0)` |
+| `initial_w` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Initial W matrix. | `zeros(0, 0)` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform NMF on. | `**--**` |
+| `max_iterations` | [`Int`](#doc_Int) | Number of iterations before NMF terminates (0 runs until convergence. | `10000` |
+| `min_residue` | [`Float64`](#doc_Float64) | The minimum root mean square residue allowed for each iteration, below which the program terminates. | `1e-05` |
+| `rank` | [`Int`](#doc_Int) | Rank of the factorization. | `**--**` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `update_rules` | [`String`](#doc_String) | Update rules for each iteration; ( multdist \| multdiv \| als ). | `"multdist"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `h` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save the calculated H to. | 
+| `w` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save the calculated W to. | 
+
+### Detailed documentation
+{: #nmf_detailed-documentation }
+
+This program performs non-negative matrix factorization on the given dataset, storing the resulting decomposed matrices in the specified files.  For an input dataset V, NMF decomposes V into two matrices W and H such that 
+
+V = W * H
+
+where all elements in W and H are non-negative.  If V is of size (n x m), then W will be of size (n x r) and H will be of size (r x m), where r is the rank of the factorization (specified by the `rank` parameter).
+
+Optionally, the desired update rules for each NMF iteration can be chosen from the following list:
+
+ - multdist: multiplicative distance-based update rules (Lee and Seung 1999)
+ - multdiv: multiplicative divergence-based update rules (Lee and Seung 1999)
+ - als: alternating least squares update rules (Paatero and Tapper 1994)
+
+The maximum number of iterations is specified with `max_iterations`, and the minimum residue required for algorithm termination is specified with the `min_residue` parameter.
+
+### Example
+For example, to run NMF on the input matrix ``V`` using the 'multdist' update rules with a rank-10 decomposition and storing the decomposed matrices into ``W`` and ``H``, the following command could be used: 
+
+```julia
+julia> using CSV
+julia> V = CSV.read("V.csv")
+julia> H, W = nmf(V, 10; update_rules="multdist")
+```
+
+### See also
+
+ - [cf()](#cf)
+ - [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ - [Algorithms for non-negative matrix factorization (pdf)](http://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf)
+ - [AMF C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/amf/amf.hpp)
+
+## pca()
+{: #pca }
+
+#### Principal Components Analysis
+{: #pca_descr }
+
+```julia
+julia> using mlpack: pca
+julia> output = pca(input; decomposition_method="exact",
+                    new_dimensionality=0, scale=false, var_to_retain=0,
+                    verbose=false)
+```
+
+An implementation of several strategies for principal components analysis (PCA), a common preprocessing step.  Given a dataset and a desired new dimensionality, this can reduce the dimensionality of the data using the linear transformation determined by PCA. [Detailed documentation](#pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `decomposition_method` | [`String`](#doc_String) | Method used for the principal components analysis: 'exact', 'randomized', 'randomized-block-krylov', 'quic'. | `"exact"` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset to perform PCA on. | `**--**` |
+| `new_dimensionality` | [`Int`](#doc_Int) | Desired dimensionality of output dataset. If 0, no dimensionality reduction is performed. | `0` |
+| `scale` | [`Bool`](#doc_Bool) | If set, the data will be scaled before running PCA, such that the variance of each feature is 1. | `false` |
+| `var_to_retain` | [`Float64`](#doc_Float64) | Amount of variance to retain; should be between 0 and 1.  If 1, all variance is retained.  Overrides -d. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #pca_detailed-documentation }
+
+This program performs principal components analysis on the given dataset using the exact, randomized, randomized block Krylov, or QUIC SVD method. It will transform the data onto its principal components, optionally performing dimensionality reduction by ignoring the principal components with the smallest eigenvalues.
+
+Use the `input` parameter to specify the dataset to perform PCA on.  A desired new dimensionality can be specified with the `new_dimensionality` parameter, or the desired variance to retain can be specified with the `var_to_retain` parameter.  If desired, the dataset can be scaled before running PCA with the `scale` parameter.
+
+Multiple different decomposition techniques can be used.  The method to use can be specified with the `decomposition_method` parameter, and it may take the values 'exact', 'randomized', or 'quic'.
+
+### Example
+For example, to reduce the dimensionality of the matrix ``data`` to 5 dimensions using randomized SVD for the decomposition, storing the output matrix to ``data_mod``, the following command can be used:
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> data_mod = pca(data; decomposition_method="randomized",
+            new_dimensionality=5)
+```
+
+### See also
+
+ - [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+ - [PCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/pca/pca.hpp)
+
+## perceptron()
+{: #perceptron }
+
+#### Perceptron
+{: #perceptron_descr }
+
+```julia
+julia> using mlpack: perceptron
+julia> output, output_model, predictions = perceptron( ;
+          input_model=nothing, labels=Int[], max_iterations=1000, test=zeros(0,
+          0), training=zeros(0, 0), verbose=false)
+```
+
+An implementation of a perceptron---a single level neural network--=for classification.  Given labeled data, a perceptron can be trained and saved for future use; or, a pre-trained perceptron can be used for classification on new points. [Detailed documentation](#perceptron_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`PerceptronModel`](#doc_model) | Input perceptron model. | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | A matrix containing labels for the training set. | `Int[]` |
+| `max_iterations` | [`Int`](#doc_Int) | The maximum number of iterations the perceptron is to be run | `1000` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the test set. | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the training set. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Int vector-like`](#doc_Int_vector_like) | The matrix in which the predicted labels for the test set will be written. | 
+| `output_model` | [`PerceptronModel`](#doc_model) | Output for trained perceptron model. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | The matrix in which the predicted labels for the test set will be written. | 
+
+### Detailed documentation
+{: #perceptron_detailed-documentation }
+
+This program implements a perceptron, which is a single level neural network. The perceptron makes its predictions based on a linear predictor function combining a set of weights with the feature vector.  The perceptron learning rule is able to converge, given enough iterations (specified using the `max_iterations` parameter), if the data supplied is linearly separable.  The perceptron is parameterized by a matrix of weight vectors that denote the numerical weights of the neural network.
+
+This program allows loading a perceptron from a model (via the `input_model` parameter) or training a perceptron given training data (via the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (via the `test` parameter) and the classification results on the test set may be saved with the `predictions` output parameter.  The perceptron model may be saved with the `output_model` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `output`.
+Use `predictions` instead of `output`.
+
+### Example
+The training data given with the `training` option may have class labels as its last dimension (so, if the training data is in CSV format, labels should be the last column).  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+All these options make it easy to train a perceptron, and then re-use that perceptron for later classification.  The invocation below trains a perceptron on ``training_data`` with labels ``training_labels``, and saves the model to ``perceptron_model``.
+
+```julia
+julia> using CSV
+julia> training_data = CSV.read("training_data.csv")
+julia> training_labels = CSV.read("training_labels.csv"; type=Int)
+julia> _, perceptron_model, _ = perceptron(labels=training_labels,
+            training=training_data)
+```
+
+Then, this model can be re-used for classification on the test data ``test_data``.  The example below does precisely that, saving the predicted classes to ``predictions``.
+
+```julia
+julia> using CSV
+julia> test_data = CSV.read("test_data.csv")
+julia> _, _, predictions = perceptron(input_model=perceptron_model,
+            test=test_data)
+```
+
+Note that all of the options may be specified at once: predictions may be calculated right after training a model, and model training can occur even if an existing perceptron model is passed with the `input_model` parameter.  However, note that the number of classes and the dimensionality of all data must match.  So you cannot pass a perceptron model trained on 2 classes and then re-train with a 4-class dataset.  Similarly, attempting classification on a 3-dimensional dataset with a perceptron that has been trained on 8 dimensions will cause an error.
+
+### See also
+
+ - [adaboost()](#adaboost)
+ - [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+ - [Perceptron C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/perceptron/perceptron.hpp)
+
+## preprocess_split()
+{: #preprocess_split }
+
+#### Split Data
+{: #preprocess_split_descr }
+
+```julia
+julia> using mlpack: preprocess_split
+julia> test, test_labels, training, training_labels =
+          preprocess_split(input; input_labels=zeros(Int, 0, 0),
+          no_shuffle=false, seed=0, stratify_data=false, test_ratio=0.2,
+          verbose=false)
+```
+
+A utility to split data into a training and testing dataset.  This can also split labels according to the same split. [Detailed documentation](#preprocess_split_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing data. | `**--**` |
+| `input_labels` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix containing labels. | `zeros(Int, 0, 0)` |
+| `no_shuffle` | [`Bool`](#doc_Bool) | Avoid shuffling the data before splitting. | `false` |
+| `seed` | [`Int`](#doc_Int) | Random seed (0 for std::time(NULL)). | `0` |
+| `stratify_data` | [`Bool`](#doc_Bool) | Stratify the data according to labels | `false` |
+| `test_ratio` | [`Float64`](#doc_Float64) | Ratio of test set; if not set,the ratio defaults to 0.2 | `0.2` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save test data to. | 
+| `test_labels` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to save test labels to. | 
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save training data to. | 
+| `training_labels` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to save train labels to. | 
+
+### Detailed documentation
+{: #preprocess_split_detailed-documentation }
+
+This utility takes a dataset and optionally labels and splits them into a training set and a test set. Before the split, the points in the dataset are randomly reordered. The percentage of the dataset to be used as the test set can be specified with the `test_ratio` parameter; the default is 0.2 (20%).
+
+The output training and test matrices may be saved with the `training` and `test` output parameters.
+
+Optionally, labels can also be split along with the data by specifying the `input_labels` parameter.  Splitting labels works the same way as splitting the data. The output training and test labels may be saved with the `training_labels` and `test_labels` output parameters, respectively.
+
+### Example
+So, a simple example where we want to split the dataset ``X`` into ``X_train`` and ``X_test`` with 60% of the data in the training set and 40% of the dataset in the test set, we could run 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_test, _, X_train, _ = preprocess_split(X; test_ratio=0.4)
+```
+
+Also by default the dataset is shuffled and split; you can provide the `no_shuffle` option to avoid shuffling the data; an example to avoid shuffling of data is:
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_test, _, X_train, _ = preprocess_split(X; no_shuffle=1,
+            test_ratio=0.4)
+```
+
+If we had a dataset ``X`` and associated labels ``y``, and we wanted to split these into ``X_train``, ``y_train``, ``X_test``, and ``y_test``, with 30% of the data in the test set, we could run
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> y = CSV.read("y.csv"; type=Int)
+julia> X_test, y_test, X_train, y_train = preprocess_split(X;
+            input_labels=y, test_ratio=0.3)
+```
+
+To maintain the ratio of each class in the train and test sets, the`stratify_data` option can be used.
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_test, _, X_train, _ = preprocess_split(X; stratify_data=1,
+            test_ratio=0.4)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_binarize()
+{: #preprocess_binarize }
+
+#### Binarize Data
+{: #preprocess_binarize_descr }
+
+```julia
+julia> using mlpack: preprocess_binarize
+julia> output = preprocess_binarize(input; dimension=0, threshold=0,
+          verbose=false)
+```
+
+A utility to binarize a dataset.  Given a dataset, this utility converts each value in the desired dimension(s) to 0 or 1; this can be a useful preprocessing step. [Detailed documentation](#preprocess_binarize_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `dimension` | [`Int`](#doc_Int) | Dimension to apply the binarization. If not set, the program will binarize every dimension by default. | `0` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input data matrix. | `**--**` |
+| `threshold` | [`Float64`](#doc_Float64) | Threshold to be applied for binarization. If not set, the threshold defaults to 0.0. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix in which to save the output. | 
+
+### Detailed documentation
+{: #preprocess_binarize_detailed-documentation }
+
+This utility takes a dataset and binarizes the variables into either 0 or 1 given threshold. User can apply binarization on a dimension or the whole dataset.  The dimension to apply binarization to can be specified using the `dimension` parameter; if left unspecified, every dimension will be binarized.  The threshold for binarization can also be specified with the `threshold` parameter; the default threshold is 0.0.
+
+The binarized matrix may be saved with the `output` output parameter.
+
+### Example
+For example, if we want to set all variables greater than 5 in the dataset ``X`` to 1 and variables less than or equal to 5.0 to 0, and save the result to ``Y``, we could run
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> Y = preprocess_binarize(X; threshold=5)
+```
+
+But if we want to apply this to only the first (0th) dimension of ``X``,  we could instead run
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> Y = preprocess_binarize(X; dimension=0, threshold=5)
+```
+
+### See also
+
+ - [preprocess_describe()](#preprocess_describe)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_describe()
+{: #preprocess_describe }
+
+#### Descriptive Statistics
+{: #preprocess_describe_descr }
+
+```julia
+julia> using mlpack: preprocess_describe
+julia> preprocess_describe(input; dimension=0,
+                           population=false, precision=4, row_major=false,
+                           verbose=false, width=8)
+```
+
+A utility for printing descriptive statistics about a dataset.  This prints a number of details about a dataset in a tabular format. [Detailed documentation](#preprocess_describe_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `dimension` | [`Int`](#doc_Int) | Dimension of the data. Use this to specify a dimension | `0` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing data, | `**--**` |
+| `population` | [`Bool`](#doc_Bool) | If specified, the program will calculate statistics assuming the dataset is the population. By default, the program will assume the dataset as a sample. | `false` |
+| `precision` | [`Int`](#doc_Int) | Precision of the output statistics. | `4` |
+| `row_major` | [`Bool`](#doc_Bool) | If specified, the program will calculate statistics across rows, not across columns.  (Remember that in mlpack, a column represents a point, so this option is generally not necessary.) | `false` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `width` | [`Int`](#doc_Int) | Width of the output table. | `8` |
+
+
+### Detailed documentation
+{: #preprocess_describe_detailed-documentation }
+
+This utility takes a dataset and prints out the descriptive statistics of the data. Descriptive statistics is the discipline of quantitatively describing the main features of a collection of information, or the quantitative description itself. The program does not modify the original file, but instead prints out the statistics to the console. The printed result will look like a table.
+
+Optionally, width and precision of the output can be adjusted by a user using the `width` and `precision` parameters. A user can also select a specific dimension to analyze if there are too many dimensions. The `population` parameter can be specified when the dataset should be considered as a population.  Otherwise, the dataset will be considered as a sample.
+
+### Example
+So, a simple example where we want to print out statistical facts about the dataset ``X`` using the default settings, we could run 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> preprocess_describe(X; verbose=1)
+```
+
+If we want to customize the width to 10 and precision to 5 and consider the dataset as a population, we could run
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> preprocess_describe(X; precision=5, verbose=1, width=10)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_scale()
+{: #preprocess_scale }
+
+#### Scale Data
+{: #preprocess_scale_descr }
+
+```julia
+julia> using mlpack: preprocess_scale
+julia> output, output_model = preprocess_scale(input; epsilon=1e-06,
+          input_model=nothing, inverse_scaling=false, max_value=1, min_value=0,
+          scaler_method="standard_scaler", seed=0, verbose=false)
+```
+
+A utility to perform feature scaling on datasets using one of sixtechniques.  Both scaling and inverse scaling are supported, andscalers can be saved and then applied to other datasets. [Detailed documentation](#preprocess_scale_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `epsilon` | [`Float64`](#doc_Float64) | regularization Parameter for pcawhitening, or zcawhitening, should be between -1 to 1. | `1e-06` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing data. | `**--**` |
+| `input_model` | [`ScalingModel`](#doc_model) | Input Scaling model. | `nothing` |
+| `inverse_scaling` | [`Bool`](#doc_Bool) | Inverse Scaling to get original dataset | `false` |
+| `max_value` | [`Int`](#doc_Int) | Ending value of range for min_max_scaler. | `1` |
+| `min_value` | [`Int`](#doc_Int) | Starting value of range for min_max_scaler. | `0` |
+| `scaler_method` | [`String`](#doc_String) | method to use for scaling, the default is standard_scaler. | `"standard_scaler"` |
+| `seed` | [`Int`](#doc_Int) | Random seed (0 for std::time(NULL)). | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save scaled data to. | 
+| `output_model` | [`ScalingModel`](#doc_model) | Output scaling model. | 
+
+### Detailed documentation
+{: #preprocess_scale_detailed-documentation }
+
+This utility takes a dataset and performs feature scaling using one of the six scaler methods namely: 'max_abs_scaler', 'mean_normalization', 'min_max_scaler' ,'standard_scaler', 'pca_whitening' and 'zca_whitening'. The function takes a matrix as `input` and a scaling method type which you can specify using `scaler_method` parameter; the default is standard scaler, and outputs a matrix with scaled feature.
+
+The output scaled feature matrix may be saved with the `output` output parameters.
+
+The model to scale features can be saved using `output_model` and later can be loaded back using`input_model`.
+
+### Example
+So, a simple example where we want to scale the dataset ``X`` into ``X_scaled`` with  standard_scaler as scaler_method, we could run 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_scaled, _ = preprocess_scale(X;
+            scaler_method="standard_scaler")
+```
+
+A simple example where we want to whiten the dataset ``X`` into ``X_whitened`` with  PCA as whitening_method and use 0.01 as regularization parameter, we could run 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_scaled, _ = preprocess_scale(X; epsilon=0.01,
+            scaler_method="pca_whitening")
+```
+
+You can also retransform the scaled dataset back using`inverse_scaling`. An example to rescale : ``X_scaled`` into ``X``using the saved model `input_model` is:
+
+```julia
+julia> using CSV
+julia> X_scaled = CSV.read("X_scaled.csv")
+julia> X, _ = preprocess_scale(X_scaled; input_model=saved,
+            inverse_scaling=1)
+```
+
+Another simple example where we want to scale the dataset ``X`` into ``X_scaled`` with  min_max_scaler as scaler method, where scaling range is 1 to 3 instead of default 0 to 1. We could run 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_scaled, _ = preprocess_scale(X; max_value=3, min_value=1,
+            scaler_method="min_max_scaler")
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_one_hot_encoding()
+{: #preprocess_one_hot_encoding }
+
+#### One Hot Encoding
+{: #preprocess_one_hot_encoding_descr }
+
+```julia
+julia> using mlpack: preprocess_one_hot_encoding
+julia> output = preprocess_one_hot_encoding(input; dimensions=[],
+          verbose=false)
+```
+
+A utility to do one-hot encoding on features of dataset. [Detailed documentation](#preprocess_one_hot_encoding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `dimensions` | [`Array{Int, 1}`](#doc_Array_Int__1_) | Index of dimensions that need to be one-hot encoded (if unspecified, all categorical dimensions are one-hot encoded). | `[]` |
+| `input` | [`Tuple{Array{Bool, 1}, Array{Float64, 2}}`](#doc_Tuple_Array_Bool__1___Array_Float64__2__) | Matrix containing data. | `**--**` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save one-hot encoded features data to. | 
+
+### Detailed documentation
+{: #preprocess_one_hot_encoding_detailed-documentation }
+
+This utility takes a dataset and a vector of indices and does one-hot encoding of the respective features at those indices. Indices represent the IDs of the dimensions to be one-hot encoded.
+
+If no dimensions are specified with `dimensions`, then all categorical-type dimensions will be one-hot encoded. Otherwise, only the dimensions given in `dimensions` will be one-hot encoded.
+
+The output matrix with encoded features may be saved with the `output` parameters.
+
+### Example
+So, a simple example where we want to encode 1st and 3rd feature from dataset ``X`` into ``X_output`` would be
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> X_ouput = preprocess_one_hot_encoding(X; dimensions=1)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+ - [One-hot encoding on Wikipedia](https://en.m.wikipedia.org/wiki/One-hot)
+
+## radical()
+{: #radical }
+
+#### RADICAL
+{: #radical_descr }
+
+```julia
+julia> using mlpack: radical
+julia> output_ic, output_unmixing = radical(input; angles=150,
+          noise_std_dev=0.175, objective=false, replicates=30, seed=0, sweeps=0,
+          verbose=false)
+```
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Given a dataset, this can decompose the dataset into an unmixing matrix and an independent component matrix; this can be useful for preprocessing. [Detailed documentation](#radical_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `angles` | [`Int`](#doc_Int) | Number of angles to consider in brute-force search during Radical2D. | `150` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input dataset for ICA. | `**--**` |
+| `noise_std_dev` | [`Float64`](#doc_Float64) | Standard deviation of Gaussian noise. | `0.175` |
+| `objective` | [`Bool`](#doc_Bool) | If set, an estimate of the final objective function is printed. | `false` |
+| `replicates` | [`Int`](#doc_Int) | Number of Gaussian-perturbed replicates to use (per point) in Radical2D. | `30` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `sweeps` | [`Int`](#doc_Int) | Number of sweeps; each sweep calls Radical2D once for each pair of dimensions. | `0` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_ic` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save independent components to. | 
+| `output_unmixing` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save unmixing matrix to. | 
+
+### Detailed documentation
+{: #radical_detailed-documentation }
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Assuming that we have an input matrix X, the goal is to find a square unmixing matrix W such that Y = W * X and the dimensions of Y are independent components.  If the algorithm is running particularly slowly, try reducing the number of replicates.
+
+The input matrix to perform ICA on should be specified with the `input` parameter.  The output matrix Y may be saved with the `output_ic` output parameter, and the output unmixing matrix W may be saved with the `output_unmixing` output parameter.
+
+### Example
+For example, to perform ICA on the matrix ``X`` with 40 replicates, saving the independent components to ``ic``, the following command may be used: 
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> ic, _ = radical(X; replicates=40)
+```
+
+### See also
+
+ - [Independent component analysis on Wikipedia](https://en.wikipedia.org/wiki/Independent_component_analysis)
+ - [ICA using spacings estimates of entropy (pdf)](https://www.jmlr.org/papers/volume4/learned-miller03a/learned-miller03a.pdf)
+ - [Radical C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/radical/radical.hpp)
+
+## random_forest()
+{: #random_forest }
+
+#### Random forests
+{: #random_forest_descr }
+
+```julia
+julia> using mlpack: random_forest
+julia> output_model, predictions, probabilities = random_forest( ;
+          input_model=nothing, labels=Int[], maximum_depth=0,
+          minimum_gain_split=0, minimum_leaf_size=1, num_trees=10,
+          print_training_accuracy=false, seed=0, subspace_dim=0, test=zeros(0,
+          0), test_labels=Int[], training=zeros(0, 0), verbose=false,
+          warm_start=false)
+```
+
+An implementation of the standard random forest algorithm by Leo Breiman for classification.  Given labeled data, a random forest can be trained and saved for future use; or, a pre-trained random forest can be used for classification. [Detailed documentation](#random_forest_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`RandomForestModel`](#doc_model) | Pre-trained random forest to use for classification. | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels for training dataset. | `Int[]` |
+| `maximum_depth` | [`Int`](#doc_Int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`Float64`](#doc_Float64) | Minimum gain needed to make a split when building a tree. | `0` |
+| `minimum_leaf_size` | [`Int`](#doc_Int) | Minimum number of points in each leaf node. | `1` |
+| `num_trees` | [`Int`](#doc_Int) | Number of trees in the random forest. | `10` |
+| `print_training_accuracy` | [`Bool`](#doc_Bool) | If set, then the accuracy of the model on the training set will be predicted (verbose must also be specified). | `false` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `subspace_dim` | [`Int`](#doc_Int) | Dimensionality of random subspace to use for each split.  '0' will autoselect the square root of data dimensionality. | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Test dataset to produce predictions for. | `zeros(0, 0)` |
+| `test_labels` | [`Int vector-like`](#doc_Int_vector_like) | Test dataset labels, if accuracy calculation is desired. | `Int[]` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Training dataset. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `warm_start` | [`Bool`](#doc_Bool) | If true and passed along with `training` and `input_model` then trains more trees on top of existing model. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`RandomForestModel`](#doc_model) | Model to save trained random forest to. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | Predicted classes for each point in the test set. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #random_forest_detailed-documentation }
+
+This program is an implementation of the standard random forest classification algorithm by Leo Breiman.  A random forest can be trained and saved for later use, or a random forest may be loaded and predictions or class probabilities for points may be generated.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model`parameter. The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `num_trees` controls the number of trees in the random forest.  The `minimum_gain_split` parameter controls the minimum required gain for a decision tree node to split.  Larger values will force higher-confidence splits.  The `maximum_depth` parameter specifies the maximum depth of the tree.  The `subspace_dim` parameter is used to control the number of random dimensions chosen for an individual node's split.  If `print_training_accuracy` is specified, the calculated accuracy on the training set will be printed.
+
+Test data may be specified with the `test` parameter, and if performance measures are desired for that test set, labels for the test points may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions`output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a random forest with a minimum leaf size of 20 using 10 trees on the dataset contained in ``data``with labels ``labels``, saving the output random forest to ``rf_model`` and printing the training error, one could call
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> rf_model, _, _ = random_forest(labels=labels,
+            minimum_leaf_size=20, num_trees=10, print_training_accuracy=1,
+            training=data)
+```
+
+Then, to use that model to classify points in ``test_set`` and print the test error given the labels ``test_labels`` using that model, while saving the predictions for each point to ``predictions``, one could call 
+
+```julia
+julia> using CSV
+julia> test_set = CSV.read("test_set.csv")
+julia> test_labels = CSV.read("test_labels.csv"; type=Int)
+julia> _, predictions, _ = random_forest(input_model=rf_model,
+            test=test_set, test_labels=test_labels)
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [hoeffding_tree()](#hoeffding_tree)
+ - [softmax_regression()](#softmax_regression)
+ - [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ - [Random forests (pdf)](https://www.eecis.udel.edu/~shatkay/Course/papers/BreimanRandomForests2001.pdf)
+ - [RandomForest C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/random_forest/random_forest.hpp)
+
+## krann()
+{: #krann }
+
+#### K-Rank-Approximate-Nearest-Neighbors (kRANN)
+{: #krann_descr }
+
+```julia
+julia> using mlpack: krann
+julia> distances, neighbors, output_model = krann( ; alpha=0.95,
+          first_leaf_exact=false, input_model=nothing, k=0, leaf_size=20,
+          naive=false, query=zeros(0, 0), random_basis=false, reference=zeros(0,
+          0), sample_at_leaves=false, seed=0, single_mode=false,
+          single_sample_limit=20, tau=5, tree_type="kd", verbose=false)
+```
+
+An implementation of rank-approximate k-nearest-neighbor search (kRANN)  using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#krann_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `alpha` | [`Float64`](#doc_Float64) | The desired success probability. | `0.95` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `first_leaf_exact` | [`Bool`](#doc_Bool) | The flag to trigger sampling only after exactly exploring the first leaf. | `false` |
+| `input_model` | [`RAModel`](#doc_model) | Pre-trained kNN model. | `nothing` |
+| `k` | [`Int`](#doc_Int) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`Int`](#doc_Int) | Leaf size for tree building (used for kd-trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `naive` | [`Bool`](#doc_Bool) | If true, sampling will be done without using a tree. | `false` |
+| `query` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing query points (optional). | `zeros(0, 0)` |
+| `random_basis` | [`Bool`](#doc_Bool) | Before tree-building, project the data onto a random orthogonal basis. | `false` |
+| `reference` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing the reference dataset. | `zeros(0, 0)` |
+| `sample_at_leaves` | [`Bool`](#doc_Bool) | The flag to trigger sampling at leaves. | `false` |
+| `seed` | [`Int`](#doc_Int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `single_mode` | [`Bool`](#doc_Bool) | If true, single-tree search is used (as opposed to dual-tree search. | `false` |
+| `single_sample_limit` | [`Int`](#doc_Int) | The limit on the maximum number of samples (and hence the largest node you can approximate). | `20` |
+| `tau` | [`Float64`](#doc_Float64) | The allowed rank-error in terms of the percentile of the data. | `5` |
+| `tree_type` | [`String`](#doc_String) | Type of tree to use: 'kd', 'ub', 'cover', 'r', 'x', 'r-star', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to output distances into. | 
+| `neighbors` | [`Int matrix-like`](#doc_Int_matrix_like) | Matrix to output neighbors into. | 
+| `output_model` | [`RAModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #krann_detailed-documentation }
+
+This program will calculate the k rank-approximate-nearest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. You must specify the rank approximation (in %) (and optionally the success probability).
+
+### Example
+For example, the following will return 5 neighbors from the top 0.1% of the data (with probability 0.95) for each point in ``input`` and store the distances in ``distances`` and the neighbors in ``neighbors.csv``:
+
+```julia
+julia> using CSV
+julia> input = CSV.read("input.csv")
+julia> distances, neighbors, _ = krann(k=5, reference=input,
+            tau=0.1)
+```
+
+Note that tau must be set such that the number of points in the corresponding percentile of the data is greater than k.  Thus, if we choose tau = 0.1 with a dataset of 1000 points and k = 5, then we are attempting to choose 5 nearest neighbors out of the closest 1 point -- this is invalid and the program will terminate with an error message.
+
+The output matrices are organized such that row i and column j in the neighbors output file corresponds to the index of the point in the reference set which is the i'th nearest neighbor from the point in the query set with index j.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [knn()](#knn)
+ - [lsh()](#lsh)
+ - [Rank-approximate nearest neighbor search: Retaining meaning and speed in high dimensions (pdf)](https://papers.nips.cc/paper/3864-rank-approximate-nearest-neighbor-search-retaining-meaning-and-speed-in-high-dimensions.pdf)
+ - [RASearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/rann/ra_search.hpp)
+
+## softmax_regression()
+{: #softmax_regression }
+
+#### Softmax Regression
+{: #softmax_regression_descr }
+
+```julia
+julia> using mlpack: softmax_regression
+julia> output_model, predictions, probabilities = softmax_regression(
+          ; input_model=nothing, labels=Int[], lambda=0.0001,
+          max_iterations=400, no_intercept=false, number_of_classes=0,
+          test=zeros(0, 0), test_labels=Int[], training=zeros(0, 0),
+          verbose=false)
+```
+
+An implementation of softmax regression for classification, which is a multiclass generalization of logistic regression.  Given labeled data, a softmax regression model can be trained and saved for future use, or, a pre-trained softmax regression model can be used for classification of new points. [Detailed documentation](#softmax_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`SoftmaxRegression`](#doc_model) | File containing existing model (parameters). | `nothing` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | A matrix containing labels (0 or 1) for the points in the training set (y). The labels must order as a row. | `Int[]` |
+| `lambda` | [`Float64`](#doc_Float64) | L2-regularization constant | `0.0001` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations before termination. | `400` |
+| `no_intercept` | [`Bool`](#doc_Bool) | Do not add the intercept term to the model. | `false` |
+| `number_of_classes` | [`Int`](#doc_Int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing test dataset. | `zeros(0, 0)` |
+| `test_labels` | [`Int vector-like`](#doc_Int_vector_like) | Matrix containing test labels. | `Int[]` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | A matrix containing the training set (the matrix of predictors, X). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`SoftmaxRegression`](#doc_model) | File to save trained softmax regression model to. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | Matrix to save predictions for test dataset into. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save class probabilities for test dataset into. | 
+
+### Detailed documentation
+{: #softmax_regression_detailed-documentation }
+
+This program performs softmax regression, a generalization of logistic regression to the multiclass case, and has support for L2 regularization.  The program is able to train a model, load  an existing model, and give predictions (and optionally their accuracy) for test data.
+
+Training a softmax regression model is done by giving a file of training points with the `training` parameter and their corresponding labels with the `labels` parameter. The number of classes can be manually specified with the `number_of_classes` parameter, and the maximum number of iterations of the L-BFGS optimizer can be specified with the `max_iterations` parameter.  The L2 regularization constant can be specified with the `lambda` parameter and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.
+
+The trained model can be saved with the `output_model` output parameter. If training is not desired, but only testing is, a model can be loaded with the `input_model` parameter.  At the current time, a loaded model cannot be trained further, so specifying both `input_model` and `training` is not allowed.
+
+The program is also able to evaluate a model on test data.  A test dataset can be specified with the `test` parameter. Class predictions can be saved with the `predictions` output parameter.  If labels are specified for the test data with the `test_labels` parameter, then the program will print the accuracy of the predictions on the given test set and its corresponding labels.
+
+### Example
+For example, to train a softmax regression model on the data ``dataset`` with labels ``labels`` with a maximum of 1000 iterations for training, saving the trained model to ``sr_model``, the following command can be used: 
+
+```julia
+julia> using CSV
+julia> dataset = CSV.read("dataset.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> sr_model, _, _ = softmax_regression(labels=labels,
+            training=dataset)
+```
+
+Then, to use ``sr_model`` to classify the test points in ``test_points``, saving the output predictions to ``predictions``, the following command can be used:
+
+```julia
+julia> using CSV
+julia> test_points = CSV.read("test_points.csv")
+julia> _, predictions, _ = softmax_regression(input_model=sr_model,
+            test=test_points)
+```
+
+### See also
+
+ - [logistic_regression()](#logistic_regression)
+ - [random_forest()](#random_forest)
+ - [Multinomial logistic regression (softmax regression) on Wikipedia](https://en.wikipedia.org/wiki/Multinomial_logistic_regression)
+ - [SoftmaxRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/softmax_regression/softmax_regression.hpp)
+
+## sparse_coding()
+{: #sparse_coding }
+
+#### Sparse Coding
+{: #sparse_coding_descr }
+
+```julia
+julia> using mlpack: sparse_coding
+julia> codes, dictionary, output_model = sparse_coding( ; atoms=15,
+          initial_dictionary=zeros(0, 0), input_model=nothing, lambda1=0,
+          lambda2=0, max_iterations=0, newton_tolerance=1e-06, normalize=false,
+          objective_tolerance=0.01, seed=0, test=zeros(0, 0), training=zeros(0,
+          0), verbose=false)
+```
+
+An implementation of Sparse Coding with Dictionary Learning.  Given a dataset, this will decompose the dataset into a sparse combination of a few dictionary elements, where the dictionary is learned during computation; a dictionary can be reused for future sparse coding of new points. [Detailed documentation](#sparse_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`Int`](#doc_Int) | Number of atoms in the dictionary. | `15` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `initial_dictionary` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Optional initial dictionary matrix. | `zeros(0, 0)` |
+| `input_model` | [`SparseCoding`](#doc_model) | File containing input sparse coding model. | `nothing` |
+| `lambda1` | [`Float64`](#doc_Float64) | Sparse coding l1-norm regularization parameter. | `0` |
+| `lambda2` | [`Float64`](#doc_Float64) | Sparse coding l2-norm regularization parameter. | `0` |
+| `max_iterations` | [`Int`](#doc_Int) | Maximum number of iterations for sparse coding (0 indicates no limit). | `0` |
+| `newton_tolerance` | [`Float64`](#doc_Float64) | Tolerance for convergence of Newton method. | `1e-06` |
+| `normalize` | [`Bool`](#doc_Bool) | If set, the input data matrix will be normalized before coding. | `false` |
+| `objective_tolerance` | [`Float64`](#doc_Float64) | Tolerance for convergence of the objective function. | `0.01` |
+| `seed` | [`Int`](#doc_Int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Optional matrix to be encoded by trained model. | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix of training data (X). | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save the output sparse codes of the test matrix (--test_file) to. | 
+| `dictionary` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save the output dictionary to. | 
+| `output_model` | [`SparseCoding`](#doc_model) | File to save trained sparse coding model to. | 
+
+### Detailed documentation
+{: #sparse_coding_detailed-documentation }
+
+An implementation of Sparse Coding with Dictionary Learning, which achieves sparsity via an l1-norm regularizer on the codes (LASSO) or an (l1+l2)-norm regularizer on the codes (the Elastic Net).  Given a dense data matrix X with d dimensions and n points, sparse coding seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a sparse coding matrix Z with n points in k dimensions.
+
+The original data matrix X can then be reconstructed as Z * D.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The sparse coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a sparse coding step, which updates the sparse coding matrix.
+
+Once a dictionary D is found, the sparse coding model may be used to encode other matrices, and saved for future usage.
+
+To run this program, either an input matrix or an already-saved sparse coding model must be specified.  An input matrix may be specified with the `training` option, along with the number of atoms in the dictionary (specified with the `atoms` parameter).  It is also possible to specify an initial dictionary for the optimization, with the `initial_dictionary` parameter.  An input model may be specified with the `input_model` parameter.
+
+### Example
+As an example, to build a sparse coding model on the dataset ``data`` using 200 atoms and an l1-regularization parameter of 0.1, saving the model into ``model``, use 
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> _, _, model = sparse_coding(atoms=200, lambda1=0.1,
+            training=data)
+```
+
+Then, this model could be used to encode a new matrix, ``otherdata``, and save the output codes to ``codes``: 
+
+```julia
+julia> using CSV
+julia> otherdata = CSV.read("otherdata.csv")
+julia> codes, _, _ = sparse_coding(input_model=model,
+            test=otherdata)
+```
+
+### See also
+
+ - [local_coordinate_coding()](#local_coordinate_coding)
+ - [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ - [Efficient sparse coding algorithms (pdf)](http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf)
+ - [Regularization and variable selection via the elastic net](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=46217f372a75dddc2254fdbc6b9418ba3554e453)
+ - [SparseCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/sparse_coding/sparse_coding.hpp)
+
+## adaboost()
+{: #adaboost }
+
+#### AdaBoost
+{: #adaboost_descr }
+
+```julia
+julia> using mlpack: adaboost
+julia> output, output_model, predictions, probabilities = adaboost( ;
+          input_model=nothing, iterations=1000, labels=Int[], test=zeros(0, 0),
+          tolerance=1e-10, training=zeros(0, 0), verbose=false,
+          weak_learner="decision_stump")
+```
+
+An implementation of the AdaBoost.MH (Adaptive Boosting) algorithm for classification.  This can be used to train an AdaBoost model on labeled data or use an existing AdaBoost model to predict the classes of new points. [Detailed documentation](#adaboost_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`AdaBoostModel`](#doc_model) | Input AdaBoost model. | `nothing` |
+| `iterations` | [`Int`](#doc_Int) | The maximum number of boosting iterations to be run (0 will run until convergence.) | `1000` |
+| `labels` | [`Int vector-like`](#doc_Int_vector_like) | Labels for the training set. | `Int[]` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Test dataset. | `zeros(0, 0)` |
+| `tolerance` | [`Float64`](#doc_Float64) | The tolerance for change in values of the weighted error during training. | `1e-10` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Dataset for training AdaBoost. | `zeros(0, 0)` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `weak_learner` | [`String`](#doc_String) | The type of weak learner to use: 'decision_stump', or 'perceptron'. | `"decision_stump"` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Int vector-like`](#doc_Int_vector_like) | Predicted labels for the test set. | 
+| `output_model` | [`AdaBoostModel`](#doc_model) | Output trained AdaBoost model. | 
+| `predictions` | [`Int vector-like`](#doc_Int_vector_like) | Predicted labels for the test set. | 
+| `probabilities` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #adaboost_detailed-documentation }
+
+This program implements the AdaBoost (or Adaptive Boosting) algorithm. The variant of AdaBoost implemented here is AdaBoost.MH. It uses a weak learner, either decision stumps or perceptrons, and over many iterations, creates a strong learner that is a weighted ensemble of weak learners. It runs these iterations until a tolerance value is crossed for change in the value of the weighted training error.
+
+For more information about the algorithm, see the paper "Improved Boosting Algorithms Using Confidence-Rated Predictions", by R.E. Schapire and Y. Singer.
+
+This program allows training of an AdaBoost model, and then application of that model to a test dataset.  To train a model, a dataset must be passed with the `training` option.  Labels can be given with the `labels` option; if no labels are specified, the labels will be assumed to be the last column of the input dataset.  Alternately, an AdaBoost model may be loaded with the `input_model` option.
+
+Once a model is trained or loaded, it may be used to provide class predictions for a given test dataset.  A test dataset may be specified with the `test` parameter.  The predicted classes for each point in the test dataset are output to the `predictions` output parameter.  The AdaBoost model itself is output to the `output_model` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `output`.
+Use `predictions` instead of `output`.
+
+### Example
+For example, to run AdaBoost on an input dataset ``data`` with labels ``labels``and perceptrons as the weak learner type, storing the trained model in ``model``, one could use the following command: 
+
+```julia
+julia> using CSV
+julia> data = CSV.read("data.csv")
+julia> labels = CSV.read("labels.csv"; type=Int)
+julia> _, model, _, _ = adaboost(labels=labels, training=data,
+            weak_learner="perceptron")
+```
+
+Similarly, an already-trained model in ``model`` can be used to provide class predictions from test data ``test_data`` and store the output in ``predictions`` with the following command: 
+
+```julia
+julia> using CSV
+julia> test_data = CSV.read("test_data.csv")
+julia> _, _, predictions, _ = adaboost(input_model=model,
+            test=test_data)
+```
+
+### See also
+
+ - [AdaBoost on Wikipedia](https://en.wikipedia.org/wiki/AdaBoost)
+ - [Improved boosting algorithms using confidence-rated predictions (pdf)](http://rob.schapire.net/papers/SchapireSi98.pdf)
+ - [Perceptron](#perceptron)
+ - [Decision Trees](#decision_tree)
+ - [AdaBoost C++ class documentation](../../user/methods/adaboost.md)
+
+## linear_regression()
+{: #linear_regression }
+
+#### Simple Linear Regression and Prediction
+{: #linear_regression_descr }
+
+```julia
+julia> using mlpack: linear_regression
+julia> output_model, output_predictions = linear_regression( ;
+          input_model=nothing, lambda=0, test=zeros(0, 0), training=zeros(0, 0),
+          training_responses=Float64[], verbose=false)
+```
+
+An implementation of simple linear regression and ridge regression using ordinary least squares.  Given a dataset and responses, a model can be trained and saved for later use, or a pre-trained model can be used to output regression predictions for a test set. [Detailed documentation](#linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `input_model` | [`LinearRegression`](#doc_model) | Existing LinearRegression model to use. | `nothing` |
+| `lambda` | [`Float64`](#doc_Float64) | Tikhonov regularization for ridge regression.  If 0, the method reduces to linear regression. | `0` |
+| `test` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing X' (test regressors). | `zeros(0, 0)` |
+| `training` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix containing training set X (regressors). | `zeros(0, 0)` |
+| `training_responses` | [`Float64 vector-like`](#doc_Float64_vector_like) | Optional vector containing y (responses). If not given, the responses are assumed to be the last row of the input file. | `Float64[]` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LinearRegression`](#doc_model) | Output LinearRegression model. | 
+| `output_predictions` | [`Float64 vector-like`](#doc_Float64_vector_like) | If --test_file is specified, this matrix is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #linear_regression_detailed-documentation }
+
+An implementation of simple linear regression and simple ridge regression using ordinary least squares. This solves the problem
+
+  y = X * b + e
+
+where X (specified by `training`) and y (specified either as the last column of the input matrix `training` or via the `training_responses` parameter) are known and b is the desired variable.  If the covariance matrix (X'X) is not invertible, or if the solution is overdetermined, then specify a Tikhonov regularization constant (with `lambda`) greater than 0, which will regularize the covariance matrix to make it invertible.  The calculated b may be saved with the `output_predictions` output parameter.
+
+Optionally, the calculated value of b is used to predict the responses for another matrix X' (specified by the `test` parameter):
+
+   y' = X' * b
+
+and the predicted responses y' may be saved with the `output_predictions` output parameter.  This type of regression is related to least-angle regression, which mlpack implements as the 'lars' program.
+
+### Example
+For example, to run a linear regression on the dataset ``X`` with responses ``y``, saving the trained model to ``lr_model``, the following command could be used:
+
+```julia
+julia> using CSV
+julia> X = CSV.read("X.csv")
+julia> y = CSV.read("y.csv")
+julia> lr_model, _ = linear_regression(training=X,
+            training_responses=y)
+```
+
+Then, to use ``lr_model`` to predict responses for a test set ``X_test``, saving the predictions to ``X_test_responses``, the following command could be used:
+
+```julia
+julia> using CSV
+julia> X_test = CSV.read("X_test.csv")
+julia> _, X_test_responses = linear_regression(input_model=lr_model,
+            test=X_test)
+```
+
+### See also
+
+ - [lars()](#lars)
+ - [Linear regression on Wikipedia](https://en.wikipedia.org/wiki/Linear_regression)
+ - [LinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_regression/linear_regression.hpp)
+
+## image_converter()
+{: #image_converter }
+
+#### Image Converter
+{: #image_converter_descr }
+
+```julia
+julia> using mlpack: image_converter
+julia> output = image_converter(input;
+                                channels=0, dataset=zeros(0, 0), height=0,
+                                quality=90, save=false, verbose=false, width=0)
+```
+
+A utility to load an image or set of images into a single dataset that can then be used by other mlpack methods and utilities. This can also unpack an image dataset into individual files, for instance after mlpack methods have been used. [Detailed documentation](#image_converter_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `channels` | [`Int`](#doc_Int) | Number of channels in the image. | `0` |
+| `check_input_matrices` | [`Bool`](#doc_Bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `false` |
+| `dataset` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Input matrix to save as images. | `zeros(0, 0)` |
+| `height` | [`Int`](#doc_Int) | Height of the images. | `0` |
+| `input` | [`Array{String, 1}`](#doc_Array_String__1_) | Image filenames which have to be loaded/saved. | `**--**` |
+| `quality` | [`Int`](#doc_Int) | Compression of the image if saved as jpg (0-100). | `90` |
+| `save` | [`Bool`](#doc_Bool) | Save a dataset as images. | `false` |
+| `verbose` | [`Bool`](#doc_Bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `false` |
+| `width` | [`Int`](#doc_Int) | Width of the image. | `0` |
+
+### Output options
+
+Results are returned as a tuple, and can be unpacked directly into return values or stored directly as a tuple; undesired results can be ignored with the _ keyword.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`Float64 matrix-like`](#doc_Float64_matrix_like) | Matrix to save images data to, Onlyneeded if you are specifying 'save' option. | 
+
+### Detailed documentation
+{: #image_converter_detailed-documentation }
+
+This utility takes an image or an array of images and loads them to a matrix. You can optionally specify the height `height` width `width` and channel `channels` of the images that needs to be loaded; otherwise, these parameters will be automatically detected from the image.
+There are other options too, that can be specified such as `quality`.
+
+You can also provide a dataset and save them as images using `dataset` and `save` as an parameter.
+
+### Example
+ An example to load an image : 
+
+```julia
+julia> Y = image_converter(X; channels=3, height=256, width=256)
+```
+
+ An example to save an image is :
+
+```julia
+julia> using CSV
+julia> Y = CSV.read("Y.csv")
+julia> _ = image_converter(X; channels=3, dataset=Y, height=256,
+            save=1, width=256)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
diff -pruN 4.3.0-2/doc/user/bindings/julia.sidebar.html 4.4.0-1/doc/user/bindings/julia.sidebar.html
--- 4.3.0-2/doc/user/bindings/julia.sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/julia.sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,100 @@
+<ul>
+<li class="page_title">
+<b>mlpack Julia binding documentation</b>
+<a href="#">[top]</a>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#mlpack-overview">Overview</a></li>
+<li><a href="LINKROOTquickstart/julia.html">Julia Quickstart</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#data-formats">Data Formats</a></li>
+<li><details><summary>Classification</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#decision_tree">decision_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#linear_svm">linear_svm()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#logistic_regression">logistic_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#nbc">nbc()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#perceptron">perceptron()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#random_forest">random_forest()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#softmax_regression">softmax_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#adaboost">adaboost()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Regression</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#bayesian_linear_regression">bayesian_linear_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#lars">lars()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#linear_regression">linear_regression()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Clustering</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#dbscan">dbscan()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#gmm_train">gmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#gmm_generate">gmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#gmm_probability">gmm_probability()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#hoeffding_tree">hoeffding_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#kmeans">kmeans()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#mean_shift">mean_shift()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Geometry</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#approx_kfn">approx_kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#emst">emst()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#fastmks">fastmks()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#lsh">lsh()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#knn">knn()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#kfn">kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#krann">krann()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Preprocessing</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#preprocess_split">preprocess_split()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#preprocess_binarize">preprocess_binarize()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#preprocess_describe">preprocess_describe()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#preprocess_scale">preprocess_scale()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#preprocess_one_hot_encoding">preprocess_one_hot_encoding()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#image_converter">image_converter()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Misc. / Other</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#cf">cf()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#det">det()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#hmm_train">hmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#hmm_generate">hmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#hmm_loglik">hmm_loglik()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#hmm_viterbi">hmm_viterbi()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#kde">kde()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#nmf">nmf()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Transformations</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/julia.html#kernel_pca">kernel_pca()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#lmnn">lmnn()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#local_coordinate_coding">local_coordinate_coding()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#nca">nca()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#pca">pca()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#radical">radical()</a></li>
+<li><a href="LINKROOTuser/bindings/julia.html#sparse_coding">sparse_coding()</a></li>
+</ul>
+</details>
+</li>
+
+</ul>
+</li>
+</ul>
+</div>
diff -pruN 4.3.0-2/doc/user/bindings/python.md 4.4.0-1/doc/user/bindings/python.md
--- 4.3.0-2/doc/user/bindings/python.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/python.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,3698 @@
+# mlpack Python binding documentation
+
+## mlpack overview
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning library with bindings to other languages.  It aims to provide fast, lightweight implementations of both common and cutting-edge machine learning algorithms.
+
+This reference page details mlpack's bindings to Python.
+
+Further useful mlpack documentation links are given below.
+
+ - [mlpack homepage](https://www.mlpack.org/)
+ - [mlpack on Github](https://github.com/mlpack/mlpack)
+ - [mlpack main documentation page](https://www.mlpack.org/docs.html)
+
+See also the quickstart guide for Python:
+
+ - [Python Quickstart](../../quickstart/python.md)
+
+## Data Formats
+
+<div id="data-formats-div" markdown="1">
+mlpack bindings for Python take and return a restricted set of types, for simplicity.  These include primitive types, matrix/vector types, categorical matrix types, and model types. Each type is detailed below.
+
+ - `int`{: #doc_int}: An integer (i.e., "1").
+ - `float`{: #doc_float }: A floating-point number (i.e., "0.5").
+ - `bool`{: #doc_bool }: A boolean flag option (True or False).
+ - `str`{: #doc_str }: A character string (i.e., "hello").
+ - `list of ints`{: #doc_list_of_ints }: A list of integers; i.e., `[0, 1, 2]`.
+ - `list of strs`{: #doc_list_of_strs }: A list of strings; i.e., `["hello", "goodbye"]`.
+ - `matrix`{: #doc_matrix }: A 2-d arraylike containing data.  This can be a list of lists, a numpy ndarray, or a pandas DataFrame.  If the dtype is not already float64, it will be converted.
+ - `int matrix`{: #doc_int_matrix }: A 2-d arraylike containing data with a uint64 dtype.  This can be a list of lists, a numpy ndarray, or a pandas DataFrame.  If the dtype is not already uint64, it will be converted.
+ - `vector`{: #doc_vector }: A 1-d arraylike containing data.  This can be a 2-d matrix where one dimension has size 1, or it can also be a list, a numpy 1-d ndarray, or a 1-d pandas DataFrame.  If the dtype is not already float64, it will be converted.
+ - `int vector`{: #doc_int_vector }: A 1-d arraylike containing data with a uint64 dtype.  This can be a 2-d matrix where one dimension has size 1, or it can also be a list, a numpy 1-d ndarray, or a 1-d pandas DataFrame.  If the dtype is not already uint64, it will be converted.
+ - `categorical matrix`{: #doc_categorical_matrix }: A 2-d arraylike containing data.  Like the regular 2-d matrices, this can be a list of lists, a numpy ndarray, or a pandas DataFrame. However, this type can also accept a pandas DataFrame that has columns of type 'CategoricalDtype'.  These categorical values will be converted to numeric indices before being passed to mlpack, and then inside mlpack they will be properly treated as categorical variables, so there is no need to do one-hot encoding for this matrix type.  If the dtype of the given matrix is not already float64, it will be converted.
+ - `mlpackModelType`{: #doc_model }: An mlpack model pointer.  This type can be pickled to or from disk, and internally holds a pointer to C++ memory containing the mlpack model.  This model pointer has 2 methods with which the parameters of the model can be inspected as well as changed through Python.  The `get_cpp_params()` method returns a python ordered dictionary that contains all the parameters of the model.  These parameters can be inspected and changed.  To set new parameters for a model, pass the modified dictionary (without deleting any keys) to the `set_cpp_params()` method.
+</div>
+
+
+## approx_kfn()
+{: #approx_kfn }
+
+#### Approximate furthest neighbor search
+{: #approx_kfn_descr }
+
+```python
+>>> from mlpack import approx_kfn
+>>> d = approx_kfn(algorithm='ds', calculate_error=False,
+        check_input_matrices=False, copy_all_inputs=False,
+        exact_distances=np.empty([0, 0]), input_model=None, k=0,
+        num_projections=5, num_tables=5, query=np.empty([0, 0]),
+        reference=np.empty([0, 0]), verbose=False)
+>>> distances = d['distances']
+>>> neighbors = d['neighbors']
+>>> output_model = d['output_model']
+```
+
+An implementation of two strategies for furthest neighbor search.  This can be used to compute the furthest neighbor of query point(s) from a set of points; furthest neighbor models can be saved and reused with future query point(s). [Detailed documentation](#approx_kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`str`](#doc_str) | Algorithm to use: 'ds' or 'qdafn'. | `'ds'` |
+| `calculate_error` | [`bool`](#doc_bool) | If set, calculate the average distance error for the first furthest neighbor only. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `exact_distances` | [`matrix`](#doc_matrix) | Matrix containing exact distances to furthest neighbors; this can be used to avoid explicit calculation when --calculate_error is set. | `np.empty([0, 0])` |
+| `input_model` | [`ApproxKFNModelType`](#doc_model) | File containing input model. | `None` |
+| `k` | [`int`](#doc_int) | Number of furthest neighbors to search for. | `0` |
+| `num_projections` | [`int`](#doc_int) | Number of projections to use in each hash table. | `5` |
+| `num_tables` | [`int`](#doc_int) | Number of hash tables to use. | `5` |
+| `query` | [`matrix`](#doc_matrix) | Matrix containing query points. | `np.empty([0, 0])` |
+| `reference` | [`matrix`](#doc_matrix) | Matrix containing the reference dataset. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`matrix`](#doc_matrix) | Matrix to save furthest neighbor distances to. | 
+| `neighbors` | [`int matrix`](#doc_int_matrix) | Matrix to save neighbor indices to. | 
+| `output_model` | [`ApproxKFNModelType`](#doc_model) | File to save output model to. | 
+
+### Detailed documentation
+{: #approx_kfn_detailed-documentation }
+
+This program implements two strategies for furthest neighbor search. These strategies are:
+
+ - The 'qdafn' algorithm from "Approximate Furthest Neighbor in High Dimensions" by R. Pagh, F. Silvestri, J. Sivertsen, and M. Skala, in Similarity Search and Applications 2015 (SISAP).
+ - The 'DrusillaSelect' algorithm from "Fast approximate furthest neighbors with data-dependent candidate selection", by R.R. Curtin and A.B. Gardner, in Similarity Search and Applications 2016 (SISAP).
+
+These two strategies give approximate results for the furthest neighbor search problem and can be used as fast replacements for other furthest neighbor techniques such as those found in the mlpack_kfn program.  Note that typically, the 'ds' algorithm requires far fewer tables and projections than the 'qdafn' algorithm.
+
+Specify a reference set (set to search in) with `reference`, specify a query set with `query`, and specify algorithm parameters with `num_tables` and `num_projections` (or don't and defaults will be used).  The algorithm to be used (either 'ds'---the default---or 'qdafn')  may be specified with `algorithm`.  Also specify the number of neighbors to search for with `k`.
+
+Note that for 'qdafn' in lower dimensions, `num_projections` may need to be set to a high value in order to return results for each query point.
+
+If no query set is specified, the reference set will be used as the query set.  The `output_model` output parameter may be used to store the built model, and an input model may be loaded instead of specifying a reference set with the `input_model` option.
+
+Results for each query point can be stored with the `neighbors` and `distances` output parameters.  Each row of these output matrices holds the k distances or neighbor indices for each query point.
+
+### Example
+For example, to find the 5 approximate furthest neighbors with `'reference_set'` as the reference set and `'query_set'` as the query set using DrusillaSelect, storing the furthest neighbor indices to `'neighbors'` and the furthest neighbor distances to `'distances'`, one could call
+
+```python
+>>> output = approx_kfn(query=query_set, reference=reference_set, k=5,
+  algorithm='ds')
+>>> neighbors = output['neighbors']
+>>> distances = output['distances']
+```
+
+and to perform approximate all-furthest-neighbors search with k=1 on the set `'data'` storing only the furthest neighbor distances to `'distances'`, one could call
+
+```python
+>>> output = approx_kfn(reference=reference_set, k=1)
+>>> distances = output['distances']
+```
+
+A trained model can be re-used.  If a model has been previously saved to `'model'`, then we may find 3 approximate furthest neighbors on a query set `'new_query_set'` using that model and store the furthest neighbor indices into `'neighbors'` by calling
+
+```python
+>>> output = approx_kfn(input_model=model, query=new_query_set, k=3)
+>>> neighbors = output['neighbors']
+```
+
+### See also
+
+ - [k-furthest-neighbor search](#kfn)
+ - [k-nearest-neighbor search](#knn)
+ - [Fast approximate furthest neighbors with data-dependent candidate selection (pdf)](http://ratml.org/pub/pdf/2016fast.pdf)
+ - [Approximate furthest neighbor in high dimensions (pdf)](https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf)
+ - [QDAFN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/qdafn.hpp)
+ - [DrusillaSelect class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/drusilla_select.hpp)
+
+## bayesian_linear_regression()
+{: #bayesian_linear_regression }
+
+#### BayesianLinearRegression
+{: #bayesian_linear_regression_descr }
+
+```python
+>>> from mlpack import bayesian_linear_regression
+>>> d = bayesian_linear_regression(center=False,
+        check_input_matrices=False, copy_all_inputs=False, input_=np.empty([0,
+        0]), input_model=None, responses=np.empty([0]), scale=False,
+        test=np.empty([0, 0]), verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> stds = d['stds']
+```
+
+An implementation of the bayesian linear regression. [Detailed documentation](#bayesian_linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `center` | [`bool`](#doc_bool) | Center the data and fit the intercept if enabled. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix of covariates (X). | `np.empty([0, 0])` |
+| `input_model` | [`BayesianLinearRegression<>Type`](#doc_model) | Trained BayesianLinearRegression model to use. | `None` |
+| `responses` | [`vector`](#doc_vector) | Matrix of responses/observations (y). | `np.empty([0])` |
+| `scale` | [`bool`](#doc_bool) | Scale each feature by their standard deviations if enabled. | `False` |
+| `test` | [`matrix`](#doc_matrix) | Matrix containing points to regress on (test points). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`BayesianLinearRegression<>Type`](#doc_model) | Output BayesianLinearRegression model. | 
+| `predictions` | [`matrix`](#doc_matrix) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+| `stds` | [`matrix`](#doc_matrix) | If specified, this is where the standard deviations of the predictive distribution will be saved. | 
+
+### Detailed documentation
+{: #bayesian_linear_regression_detailed-documentation }
+
+An implementation of the bayesian linear regression.
+This model is a probabilistic view and implementation of the linear regression. The final solution is obtained by computing a posterior distribution from gaussian likelihood and a zero mean gaussian isotropic  prior distribution on the solution. 
+Optimization is AUTOMATIC and does not require cross validation. The optimization is performed by maximization of the evidence function. Parameters are tuned during the maximization of the marginal likelihood. This procedure includes the Ockham's razor that penalizes over complex solutions. 
+
+This program is able to train a Bayesian linear regression model or load a model from file, output regression predictions for a test set, and save the trained model to a file.
+
+To train a BayesianLinearRegression model, the `input_` and `responses`parameters must be given. The `center`and `scale` parameters control the centering and the normalizing options. A trained model can be saved with the `output_model`. If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `predictions` output parameter. The corresponding standard deviation can be save by precising the `stds` parameter.
+
+### Example
+For example, the following command trains a model on the data `'data'` and responses `'responses'`with center set to true and scale set to false (so, Bayesian linear regression is being solved, and then the model is saved to `'blr_model'`:
+
+```python
+>>> output = bayesian_linear_regression(input_=data, responses=responses,
+  center=1, scale=0)
+>>> blr_model = output['output_model']
+```
+
+The following command uses the `'blr_model'` to provide predicted  responses for the data `'test'` and save those  responses to `'test_predictions'`: 
+
+```python
+>>> output = bayesian_linear_regression(input_model=blr_model, test=test)
+>>> test_predictions = output['predictions']
+```
+
+Because the estimator computes a predictive distribution instead of a simple point estimate, the `stds` parameter allows one to save the prediction uncertainties: 
+
+```python
+>>> output = bayesian_linear_regression(input_model=blr_model, test=test)
+>>> test_predictions = output['predictions']
+>>> stds = output['stds']
+```
+
+### See also
+
+ - [Bayesian Interpolation](https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf)
+ - [Bayesian Linear Regression, Section 3.3](https://www.microsoft.com/en-us/research/uploads/prod/2006/01/Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf)
+ - [BayesianLinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp)
+
+## cf()
+{: #cf }
+
+#### Collaborative Filtering
+{: #cf_descr }
+
+```python
+>>> from mlpack import cf
+>>> d = cf(algorithm='NMF', all_user_recommendations=False,
+        check_input_matrices=False, copy_all_inputs=False, input_model=None,
+        interpolation='average', iteration_only_termination=False,
+        max_iterations=1000, min_residue=1e-05, neighbor_search='euclidean',
+        neighborhood=5, normalization='none', query=np.empty([0, 0],
+        dtype=np.uint64), rank=0, recommendations=5, seed=0, test=np.empty([0,
+        0]), training=np.empty([0, 0]), verbose=False)
+>>> output = d['output']
+>>> output_model = d['output_model']
+```
+
+An implementation of several collaborative filtering (CF) techniques for recommender systems.  This can be used to train a new CF model, or use an existing CF model to compute recommendations. [Detailed documentation](#cf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`str`](#doc_str) | Algorithm used for matrix factorization. | `'NMF'` |
+| `all_user_recommendations` | [`bool`](#doc_bool) | Generate recommendations for all users. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`CFModelType`](#doc_model) | Trained CF model to load. | `None` |
+| `interpolation` | [`str`](#doc_str) | Algorithm used for weight interpolation. | `'average'` |
+| `iteration_only_termination` | [`bool`](#doc_bool) | Terminate only when the maximum number of iterations is reached. | `False` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations. If set to zero, there is no limit on the number of iterations. | `1000` |
+| `min_residue` | [`float`](#doc_float) | Residue required to terminate the factorization (lower values generally mean better fits). | `1e-05` |
+| `neighbor_search` | [`str`](#doc_str) | Algorithm used for neighbor search. | `'euclidean'` |
+| `neighborhood` | [`int`](#doc_int) | Size of the neighborhood of similar users to consider for each query user. | `5` |
+| `normalization` | [`str`](#doc_str) | Normalization performed on the ratings. | `'none'` |
+| `query` | [`int matrix`](#doc_int_matrix) | List of query users for which recommendations should be generated. | `np.empty([0, 0], dtype=np.uint64)` |
+| `rank` | [`int`](#doc_int) | Rank of decomposed matrices (if 0, a heuristic is used to estimate the rank). | `0` |
+| `recommendations` | [`int`](#doc_int) | Number of recommendations to generate for each query user. | `5` |
+| `seed` | [`int`](#doc_int) | Set the random seed (0 uses std::time(NULL)). | `0` |
+| `test` | [`matrix`](#doc_matrix) | Test set to calculate RMSE on. | `np.empty([0, 0])` |
+| `training` | [`matrix`](#doc_matrix) | Input dataset to perform CF on. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`int matrix`](#doc_int_matrix) | Matrix that will store output recommendations. | 
+| `output_model` | [`CFModelType`](#doc_model) | Output for trained CF model. | 
+
+### Detailed documentation
+{: #cf_detailed-documentation }
+
+This program performs collaborative filtering (CF) on the given dataset. Given a list of user, item and preferences (the `training` parameter), the program will perform a matrix decomposition and then can perform a series of actions related to collaborative filtering.  Alternately, the program can load an existing saved CF model with the `input_model` parameter and then use that model to provide recommendations or predict values.
+
+The input matrix should be a 3-dimensional matrix of ratings, where the first dimension is the user, the second dimension is the item, and the third dimension is that user's rating of that item.  Both the users and items should be numeric indices, not names. The indices are assumed to start from 0.
+
+A set of query users for which recommendations can be generated may be specified with the `query` parameter; alternately, recommendations may be generated for every user in the dataset by specifying the `all_user_recommendations` parameter.  In addition, the number of recommendations per user to generate can be specified with the `recommendations` parameter, and the number of similar users (the size of the neighborhood) to be considered when generating recommendations can be specified with the `neighborhood` parameter.
+
+For performing the matrix decomposition, the following optimization algorithms can be specified via the `algorithm` parameter:
+
+ - 'RegSVD' -- Regularized SVD using a SGD optimizer
+ - 'NMF' -- Non-negative matrix factorization with alternating least squares update rules
+ - 'BatchSVD' -- SVD batch learning
+ - 'SVDIncompleteIncremental' -- SVD incomplete incremental learning
+ - 'SVDCompleteIncremental' -- SVD complete incremental learning
+ - 'BiasSVD' -- Bias SVD using a SGD optimizer
+ - 'SVDPP' -- SVD++ using a SGD optimizer
+ - 'RandSVD' -- RandomizedSVD learning
+ - 'QSVD' -- QuicSVD learning
+ - 'BKSVD' -- Block Krylov SVD learning
+
+
+The following neighbor search algorithms can be specified via the `neighbor_search` parameter:
+
+ - 'cosine'  -- Cosine Search Algorithm
+ - 'euclidean'  -- Euclidean Search Algorithm
+ - 'pearson'  -- Pearson Search Algorithm
+
+
+The following weight interpolation algorithms can be specified via the `interpolation` parameter:
+
+ - 'average'  -- Average Interpolation Algorithm
+ - 'regression'  -- Regression Interpolation Algorithm
+ - 'similarity'  -- Similarity Interpolation Algorithm
+
+
+The following ranking normalization algorithms can be specified via the `normalization` parameter:
+
+ - 'none'  -- No Normalization
+ - 'item_mean'  -- Item Mean Normalization
+ - 'overall_mean'  -- Overall Mean Normalization
+ - 'user_mean'  -- User Mean Normalization
+ - 'z_score'  -- Z-Score Normalization
+
+A trained model may be saved to with the `output_model` output parameter.
+
+### Example
+To train a CF model on a dataset `'training_set'` using NMF for decomposition and saving the trained model to `'model'`, one could call: 
+
+```python
+>>> output = cf(training=training_set, algorithm='NMF')
+>>> model = output['output_model']
+```
+
+Then, to use this model to generate recommendations for the list of users in the query set `'users'`, storing 5 recommendations in `'recommendations'`, one could call 
+
+```python
+>>> output = cf(input_model=model, query=users, recommendations=5)
+>>> recommendations = output['output']
+```
+
+### See also
+
+ - [Collaborative Filtering on Wikipedia](https://en.wikipedia.org/wiki/Collaborative_filtering)
+ - [Matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Matrix_factorization_(recommender_systems))
+ - [Matrix factorization techniques for recommender systems (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5)
+ - [CFType class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/cf/cf.hpp)
+
+## dbscan()
+{: #dbscan }
+
+#### DBSCAN clustering
+{: #dbscan_descr }
+
+```python
+>>> from mlpack import dbscan
+>>> d = dbscan(check_input_matrices=False, copy_all_inputs=False,
+        epsilon=1, input_=np.empty([0, 0]), min_size=5, naive=False,
+        selection_type='ordered', single_mode=False, tree_type='kd',
+        verbose=False)
+>>> assignments = d['assignments']
+>>> centroids = d['centroids']
+```
+
+An implementation of DBSCAN clustering.  Given a dataset, this can compute and return a clustering of that dataset. [Detailed documentation](#dbscan_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `epsilon` | [`float`](#doc_float) | Radius of each range search. | `1` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to cluster. | `**--**` |
+| `min_size` | [`int`](#doc_int) | Minimum number of points for a cluster. | `5` |
+| `naive` | [`bool`](#doc_bool) | If set, brute-force range search (not tree-based) will be used. | `False` |
+| `selection_type` | [`str`](#doc_str) | If using point selection policy, the type of selection to use ('ordered', 'random'). | `'ordered'` |
+| `single_mode` | [`bool`](#doc_bool) | If set, single-tree range search (not dual-tree) will be used. | `False` |
+| `tree_type` | [`str`](#doc_str) | If using single-tree or dual-tree search, the type of tree to use ('kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'). | `'kd'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `assignments` | [`int vector`](#doc_int_vector) | Output matrix for assignments of each point. | 
+| `centroids` | [`matrix`](#doc_matrix) | Matrix to save output centroids to. | 
+
+### Detailed documentation
+{: #dbscan_detailed-documentation }
+
+This program implements the DBSCAN algorithm for clustering using accelerated tree-based range search.  The type of tree that is used may be parameterized, or brute-force range search may also be used.
+
+The input dataset to be clustered may be specified with the `input_` parameter; the radius of each range search may be specified with the `epsilon` parameters, and the minimum number of points in a cluster may be specified with the `min_size` parameter.
+
+The `assignments` and `centroids` output parameters may be used to save the output of the clustering. `assignments` contains the cluster assignments of each point, and `centroids` contains the centroids of each cluster.
+
+The range search may be controlled with the `tree_type`, `single_mode`, and `naive` parameters.  `tree_type` can control the type of tree used for range search; this can take a variety of values: 'kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'. The `single_mode` parameter will force single-tree search (as opposed to the default dual-tree search), and '`naive` will force brute-force range search.
+
+### Example
+An example usage to run DBSCAN on the dataset in `'input'` with a radius of 0.5 and a minimum cluster size of 5 is given below:
+
+```python
+>>> dbscan(input_=input, epsilon=0.5, min_size=5)
+```
+
+### See also
+
+ - [DBSCAN on Wikipedia](https://en.wikipedia.org/wiki/DBSCAN)
+ - [A density-based algorithm for discovering clusters in large spatial databases with noise (pdf)](https://cdn.aaai.org/KDD/1996/KDD96-037.pdf)
+ - [DBSCAN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/dbscan/dbscan.hpp)
+
+## decision_tree()
+{: #decision_tree }
+
+#### Decision tree
+{: #decision_tree_descr }
+
+```python
+>>> from mlpack import decision_tree
+>>> d = decision_tree(check_input_matrices=False, copy_all_inputs=False,
+        input_model=None, labels=np.empty([0], dtype=np.uint64),
+        maximum_depth=0, minimum_gain_split=1e-07, minimum_leaf_size=20,
+        print_training_accuracy=False, print_training_error=False,
+        test=np.empty([0, 0]), test_labels=np.empty([0], dtype=np.uint64),
+        training=np.empty([0, 0]), verbose=False, weights=np.empty([0, 0]))
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of an ID3-style decision tree for classification, which supports categorical data.  Given labeled data with numeric or categorical features, a decision tree can be trained and saved; or, an existing decision tree can be used for classification on new points. [Detailed documentation](#decision_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`DecisionTreeModelType`](#doc_model) | Pre-trained decision tree, to be used with test points. | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | Training labels. | `np.empty([0], dtype=np.uint64)` |
+| `maximum_depth` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`float`](#doc_float) | Minimum gain for node splitting. | `1e-07` |
+| `minimum_leaf_size` | [`int`](#doc_int) | Minimum number of points in a leaf. | `20` |
+| `print_training_accuracy` | [`bool`](#doc_bool) | Print the training accuracy. | `False` |
+| `print_training_error` | [`bool`](#doc_bool) | Print the training error (deprecated; will be removed in mlpack 4.0.0). | `False` |
+| `test` | [`categorical matrix`](#doc_categorical_matrix) | Testing dataset (may be categorical). | `np.empty([0, 0])` |
+| `test_labels` | [`int vector`](#doc_int_vector) | Test point labels, if accuracy calculation is desired. | `np.empty([0], dtype=np.uint64)` |
+| `training` | [`categorical matrix`](#doc_categorical_matrix) | Training dataset (may be categorical). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `weights` | [`matrix`](#doc_matrix) | The weight of labels | `np.empty([0, 0])` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DecisionTreeModelType`](#doc_model) | Output for trained decision tree. | 
+| `predictions` | [`int vector`](#doc_int_vector) | Class predictions for each test point. | 
+| `probabilities` | [`matrix`](#doc_matrix) | Class probabilities for each test point. | 
+
+### Detailed documentation
+{: #decision_tree_detailed-documentation }
+
+Train and evaluate using a decision tree.  Given a dataset containing numeric or categorical features, and associated labels for each point in the dataset, this program can train a decision tree on that data.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model` parameter.  The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `minimum_gain_split` parameter specifies the minimum gain that is needed for the node to split.  The `maximum_depth` parameter specifies the maximum depth of the tree.  If `print_training_error` is specified, the training error will be printed.
+
+Test data may be specified with the `test` parameter, and if performance numbers are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions` output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a decision tree with a minimum leaf size of 20 on the dataset contained in `'data'` with labels `'labels'`, saving the output model to `'tree'` and printing the training error, one could call
+
+```python
+>>> output = decision_tree(training=data, labels=labels, minimum_leaf_size=20,
+  minimum_gain_split=0.001, print_training_accuracy=True)
+>>> tree = output['output_model']
+```
+
+Then, to use that model to classify points in `'test_set'` and print the test error given the labels `'test_labels'` using that model, while saving the predictions for each point to `'predictions'`, one could call 
+
+```python
+>>> output = decision_tree(input_model=tree, test=test_set,
+  test_labels=test_labels)
+>>> predictions = output['predictions']
+```
+
+### See also
+
+ - [Random forest](#random_forest)
+ - [Decision trees on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+ - [Induction of Decision Trees (pdf)](http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf)
+ - [DecisionTree C++ class documentation](../../user/methods/decision_tree.md)
+
+## det()
+{: #det }
+
+#### Density Estimation With Density Estimation Trees
+{: #det_descr }
+
+```python
+>>> from mlpack import det
+>>> d = det(check_input_matrices=False, copy_all_inputs=False, folds=10,
+        input_model=None, max_leaf_size=10, min_leaf_size=5, path_format='lr',
+        skip_pruning=False, test=np.empty([0, 0]), training=np.empty([0, 0]),
+        verbose=False)
+>>> output_model = d['output_model']
+>>> tag_counters_file = d['tag_counters_file']
+>>> tag_file = d['tag_file']
+>>> test_set_estimates = d['test_set_estimates']
+>>> training_set_estimates = d['training_set_estimates']
+>>> vi = d['vi']
+```
+
+An implementation of density estimation trees for the density estimation task.  Density estimation trees can be trained or used to predict the density at locations given by query points. [Detailed documentation](#det_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `folds` | [`int`](#doc_int) | The number of folds of cross-validation to perform for the estimation (0 is LOOCV) | `10` |
+| `input_model` | [`DTree<>Type`](#doc_model) | Trained density estimation tree to load. | `None` |
+| `max_leaf_size` | [`int`](#doc_int) | The maximum size of a leaf in the unpruned, fully grown DET. | `10` |
+| `min_leaf_size` | [`int`](#doc_int) | The minimum size of a leaf in the unpruned, fully grown DET. | `5` |
+| `path_format` | [`str`](#doc_str) | The format of path printing: 'lr', 'id-lr', or 'lr-id'. | `'lr'` |
+| `skip_pruning` | [`bool`](#doc_bool) | Whether to bypass the pruning process and output the unpruned tree only. | `False` |
+| `test` | [`matrix`](#doc_matrix) | A set of test points to estimate the density of. | `np.empty([0, 0])` |
+| `training` | [`matrix`](#doc_matrix) | The data set on which to build a density estimation tree. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DTree<>Type`](#doc_model) | Output to save trained density estimation tree to. | 
+| `tag_counters_file` | [`str`](#doc_str) | The file to output the number of points that went to each leaf. | 
+| `tag_file` | [`str`](#doc_str) | The file to output the tags (and possibly paths) for each sample in the test set. | 
+| `test_set_estimates` | [`matrix`](#doc_matrix) | The output estimates on the test set from the final optimally pruned tree. | 
+| `training_set_estimates` | [`matrix`](#doc_matrix) | The output density estimates on the training set from the final optimally pruned tree. | 
+| `vi` | [`matrix`](#doc_matrix) | The output variable importance values for each feature. | 
+
+### Detailed documentation
+{: #det_detailed-documentation }
+
+This program performs a number of functions related to Density Estimation Trees.  The optimal Density Estimation Tree (DET) can be trained on a set of data (specified by `training`) using cross-validation (with number of folds specified with the `folds` parameter).  This trained density estimation tree may then be saved with the `output_model` output parameter.
+
+The variable importances (that is, the feature importance values for each dimension) may be saved with the `vi` output parameter, and the density estimates for each training point may be saved with the `training_set_estimates` output parameter.
+
+Enabling path printing for each node outputs the path from the root node to a leaf for each entry in the test set, or training set (if a test set is not provided).  Strings like 'LRLRLR' (indicating that traversal went to the left child, then the right child, then the left child, and so forth) will be output. If 'lr-id' or 'id-lr' are given as the `path_format` parameter, then the ID (tag) of every node along the path will be printed after or before the L or R character indicating the direction of traversal, respectively.
+
+This program also can provide density estimates for a set of test points, specified in the `test` parameter.  The density estimation tree used for this task will be the tree that was trained on the given training points, or a tree given as the parameter `input_model`.  The density estimates for the test points may be saved using the `test_set_estimates` output parameter.
+
+### See also
+
+ - [Density estimation on Wikipedia](https://en.wikipedia.org/wiki/Density_estimation)
+ - [Density estimation trees (pdf)](https://www.mlpack.org/papers/det.pdf)
+ - [DTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/det/dtree.hpp)
+
+## emst()
+{: #emst }
+
+#### Fast Euclidean Minimum Spanning Tree
+{: #emst_descr }
+
+```python
+>>> from mlpack import emst
+>>> d = emst(check_input_matrices=False, copy_all_inputs=False,
+        input_=np.empty([0, 0]), leaf_size=1, naive=False, verbose=False)
+>>> output = d['output']
+```
+
+An implementation of the Dual-Tree Boruvka algorithm for computing the Euclidean minimum spanning tree of a set of input points. [Detailed documentation](#emst_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Input data matrix. | `**--**` |
+| `leaf_size` | [`int`](#doc_int) | Leaf size in the kd-tree.  One-element leaves give the empirically best performance, but at the cost of greater memory requirements. | `1` |
+| `naive` | [`bool`](#doc_bool) | Compute the MST using O(n^2) naive algorithm. | `False` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Output data.  Stored as an edge list. | 
+
+### Detailed documentation
+{: #emst_detailed-documentation }
+
+This program can compute the Euclidean minimum spanning tree of a set of input points using the dual-tree Boruvka algorithm.
+
+The set to calculate the minimum spanning tree of is specified with the `input_` parameter, and the output may be saved with the `output` output parameter.
+
+The `leaf_size` parameter controls the leaf size of the kd-tree that is used to calculate the minimum spanning tree, and if the `naive` option is given, then brute-force search is used (this is typically much slower in low dimensions).  The leaf size does not affect the results, but it may have some effect on the runtime of the algorithm.
+
+### Example
+For example, the minimum spanning tree of the input dataset `'data'` can be calculated with a leaf size of 20 and stored as `'spanning_tree'` using the following command:
+
+```python
+>>> output = emst(input_=data, leaf_size=20)
+>>> spanning_tree = output['output']
+```
+
+The output matrix is a three-dimensional matrix, where each row indicates an edge.  The first dimension corresponds to the lesser index of the edge; the second dimension corresponds to the greater index of the edge; and the third column corresponds to the distance between the two points.
+
+### See also
+
+ - [Minimum spanning tree on Wikipedia](https://en.wikipedia.org/wiki/Minimum_spanning_tree)
+ - [Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications (pdf)](https://www.mlpack.org/papers/emst.pdf)
+ - [DualTreeBoruvka class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/emst/dtb.hpp)
+
+## fastmks()
+{: #fastmks }
+
+#### FastMKS (Fast Max-Kernel Search)
+{: #fastmks_descr }
+
+```python
+>>> from mlpack import fastmks
+>>> d = fastmks(bandwidth=1, base=2, check_input_matrices=False,
+        copy_all_inputs=False, degree=2, input_model=None, k=0, kernel='linear',
+        naive=False, offset=0, query=np.empty([0, 0]), reference=np.empty([0,
+        0]), scale=1, single=False, verbose=False)
+>>> indices = d['indices']
+>>> kernels = d['kernels']
+>>> output_model = d['output_model']
+```
+
+An implementation of the single-tree and dual-tree fast max-kernel search (FastMKS) algorithm.  Given a set of reference points and a set of query points, this can find the reference point with maximum kernel value for each query point; trained models can be reused for future queries. [Detailed documentation](#fastmks_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`float`](#doc_float) | Bandwidth (for Gaussian, Epanechnikov, and triangular kernels). | `1` |
+| `base` | [`float`](#doc_float) | Base to use during cover tree construction. | `2` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `degree` | [`float`](#doc_float) | Degree of polynomial kernel. | `2` |
+| `input_model` | [`FastMKSModelType`](#doc_model) | Input FastMKS model to use. | `None` |
+| `k` | [`int`](#doc_int) | Number of maximum kernels to find. | `0` |
+| `kernel` | [`str`](#doc_str) | Kernel type to use: 'linear', 'polynomial', 'cosine', 'gaussian', 'epanechnikov', 'triangular', 'hyptan'. | `'linear'` |
+| `naive` | [`bool`](#doc_bool) | If true, O(n^2) naive mode is used for computation. | `False` |
+| `offset` | [`float`](#doc_float) | Offset of kernel (for polynomial and hyptan kernels). | `0` |
+| `query` | [`matrix`](#doc_matrix) | The query dataset. | `np.empty([0, 0])` |
+| `reference` | [`matrix`](#doc_matrix) | The reference dataset. | `np.empty([0, 0])` |
+| `scale` | [`float`](#doc_float) | Scale of kernel (for hyptan kernel). | `1` |
+| `single` | [`bool`](#doc_bool) | If true, single-tree search is used (as opposed to dual-tree search. | `False` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `indices` | [`int matrix`](#doc_int_matrix) | Output matrix of indices. | 
+| `kernels` | [`matrix`](#doc_matrix) | Output matrix of kernels. | 
+| `output_model` | [`FastMKSModelType`](#doc_model) | Output for FastMKS model. | 
+
+### Detailed documentation
+{: #fastmks_detailed-documentation }
+
+This program will find the k maximum kernels of a set of points, using a query set and a reference set (which can optionally be the same set). More specifically, for each point in the query set, the k points in the reference set with maximum kernel evaluations are found.  The kernel function used is specified with the `kernel` parameter.
+
+### Example
+For example, the following command will calculate, for each point in the query set `'query'`, the five points in the reference set `'reference'` with maximum kernel evaluation using the linear kernel.  The kernel evaluations may be saved with the  `'kernels'` output parameter and the indices may be saved with the `'indices'` output parameter.
+
+```python
+>>> output = fastmks(k=5, reference=reference, query=query, kernel='linear')
+>>> indices = output['indices']
+>>> kernels = output['kernels']
+```
+
+The output matrices are organized such that row i and column j in the indices matrix corresponds to the index of the point in the reference set that has j'th largest kernel evaluation with the point in the query set with index i.  Row i and column j in the kernels matrix corresponds to the kernel evaluation between those two points.
+
+This program performs FastMKS using a cover tree.  The base used to build the cover tree can be specified with the `base` parameter.
+
+### See also
+
+ - [k-nearest-neighbor search](#knn)
+ - [Dual-tree Fast Exact Max-Kernel Search (pdf)](https://mlpack.org/papers/fmks.pdf)
+ - [FastMKS class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/fastmks/fastmks.hpp)
+
+## gmm_train()
+{: #gmm_train }
+
+#### Gaussian Mixture Model (GMM) Training
+{: #gmm_train_descr }
+
+```python
+>>> from mlpack import gmm_train
+>>> d = gmm_train(check_input_matrices=False, copy_all_inputs=False,
+        diagonal_covariance=False, gaussians=0, input_=np.empty([0, 0]),
+        input_model=None, kmeans_max_iterations=1000, max_iterations=250,
+        no_force_positive=False, noise=0, percentage=0.02, refined_start=False,
+        samplings=100, seed=0, tolerance=1e-10, trials=1, verbose=False)
+>>> output_model = d['output_model']
+```
+
+An implementation of the EM algorithm for training Gaussian mixture models (GMMs).  Given a dataset, this can train a GMM for future use with other tools. [Detailed documentation](#gmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `diagonal_covariance` | [`bool`](#doc_bool) | Force the covariance of the Gaussians to be diagonal.  This can accelerate training time significantly. | `False` |
+| `gaussians` | [`int`](#doc_int) | Number of Gaussians in the GMM. | `**--**` |
+| `input_` | [`matrix`](#doc_matrix) | The training data on which the model will be fit. | `**--**` |
+| `input_model` | [`GMMType`](#doc_model) | Initial input GMM model to start training with. | `None` |
+| `kmeans_max_iterations` | [`int`](#doc_int) | Maximum number of iterations for the k-means algorithm (used to initialize EM). | `1000` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations of EM algorithm (passing 0 will run until convergence). | `250` |
+| `no_force_positive` | [`bool`](#doc_bool) | Do not force the covariance matrices to be positive definite. | `False` |
+| `noise` | [`float`](#doc_float) | Variance of zero-mean Gaussian noise to add to data. | `0` |
+| `percentage` | [`float`](#doc_float) | If using --refined_start, specify the percentage of the dataset used for each sampling (should be between 0.0 and 1.0). | `0.02` |
+| `refined_start` | [`bool`](#doc_bool) | During the initialization, use refined initial positions for k-means clustering (Bradley and Fayyad, 1998). | `False` |
+| `samplings` | [`int`](#doc_int) | If using --refined_start, specify the number of samplings used for initial points. | `100` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tolerance` | [`float`](#doc_float) | Tolerance for convergence of EM. | `1e-10` |
+| `trials` | [`int`](#doc_int) | Number of trials to perform in training GMM. | `1` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`GMMType`](#doc_model) | Output for trained GMM model. | 
+
+### Detailed documentation
+{: #gmm_train_detailed-documentation }
+
+This program takes a parametric estimate of a Gaussian mixture model (GMM) using the EM algorithm to find the maximum likelihood estimate.  The model may be saved and reused by other mlpack GMM tools.
+
+The input data to train on must be specified with the `input_` parameter, and the number of Gaussians in the model must be specified with the `gaussians` parameter.  Optionally, many trials with different random initializations may be run, and the result with highest log-likelihood on the training data will be taken.  The number of trials to run is specified with the `trials` parameter.  By default, only one trial is run.
+
+The tolerance for convergence and maximum number of iterations of the EM algorithm are specified with the `tolerance` and `max_iterations` parameters, respectively.  The GMM may be initialized for training with another model, specified with the `input_model` parameter. Otherwise, the model is initialized by running k-means on the data.  The k-means clustering initialization can be controlled with the `kmeans_max_iterations`, `refined_start`, `samplings`, and `percentage` parameters.  If `refined_start` is specified, then the Bradley-Fayyad refined start initialization will be used.  This can often lead to better clustering results.
+
+The 'diagonal_covariance' flag will cause the learned covariances to be diagonal matrices.  This significantly simplifies the model itself and causes training to be faster, but restricts the ability to fit more complex GMMs.
+
+If GMM training fails with an error indicating that a covariance matrix could not be inverted, make sure that the `no_force_positive` parameter is not specified.  Alternately, adding a small amount of Gaussian noise (using the `noise` parameter) to the entire dataset may help prevent Gaussians with zero variance in a particular dimension, which is usually the cause of non-invertible covariance matrices.
+
+The `no_force_positive` parameter, if set, will avoid the checks after each iteration of the EM algorithm which ensure that the covariance matrices are positive definite.  Specifying the flag can cause faster runtime, but may also cause non-positive definite covariance matrices, which will cause the program to crash.
+
+### Example
+As an example, to train a 6-Gaussian GMM on the data in `'data'` with a maximum of 100 iterations of EM and 3 trials, saving the trained GMM to `'gmm'`, the following command can be used:
+
+```python
+>>> output = gmm_train(input_=data, gaussians=6, trials=3)
+>>> gmm = output['output_model']
+```
+
+To re-train that GMM on another set of data `'data2'`, the following command may be used: 
+
+```python
+>>> output = gmm_train(input_model=gmm, input_=data2, gaussians=6)
+>>> new_gmm = output['output_model']
+```
+
+### See also
+
+ - [gmm_generate()](#gmm_generate)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_generate()
+{: #gmm_generate }
+
+#### GMM Sample Generator
+{: #gmm_generate_descr }
+
+```python
+>>> from mlpack import gmm_generate
+>>> d = gmm_generate(check_input_matrices=False, copy_all_inputs=False,
+        input_model=None, samples=0, seed=0, verbose=False)
+>>> output = d['output']
+```
+
+A sample generator for pre-trained GMMs.  Given a pre-trained GMM, this can sample new points randomly from that distribution. [Detailed documentation](#gmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`GMMType`](#doc_model) | Input GMM model to generate samples from. | `**--**` |
+| `samples` | [`int`](#doc_int) | Number of samples to generate. | `**--**` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save output samples in. | 
+
+### Detailed documentation
+{: #gmm_generate_detailed-documentation }
+
+This program is able to generate samples from a pre-trained GMM (use gmm_train to train a GMM).  The pre-trained GMM must be specified with the `input_model` parameter.  The number of samples to generate is specified by the `samples` parameter.  Output samples may be saved with the `output` output parameter.
+
+### Example
+The following command can be used to generate 100 samples from the pre-trained GMM `'gmm'` and store those generated samples in `'samples'`:
+
+```python
+>>> output = gmm_generate(input_model=gmm, samples=100)
+>>> samples = output['output']
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_probability()
+{: #gmm_probability }
+
+#### GMM Probability Calculator
+{: #gmm_probability_descr }
+
+```python
+>>> from mlpack import gmm_probability
+>>> d = gmm_probability(check_input_matrices=False,
+        copy_all_inputs=False, input_=np.empty([0, 0]), input_model=None,
+        verbose=False)
+>>> output = d['output']
+```
+
+A probability calculator for GMMs.  Given a pre-trained GMM and a set of points, this can compute the probability that each point is from the given GMM. [Detailed documentation](#gmm_probability_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Input matrix to calculate probabilities of. | `**--**` |
+| `input_model` | [`GMMType`](#doc_model) | Input GMM to use as model. | `**--**` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to store calculated probabilities in. | 
+
+### Detailed documentation
+{: #gmm_probability_detailed-documentation }
+
+This program calculates the probability that given points came from a given GMM (that is, P(X \| gmm)).  The GMM is specified with the `input_model` parameter, and the points are specified with the `input_` parameter.  The output probabilities may be saved via the `output` output parameter.
+
+### Example
+So, for example, to calculate the probabilities of each point in `'points'` coming from the pre-trained GMM `'gmm'`, while storing those probabilities in `'probs'`, the following command could be used:
+
+```python
+>>> output = gmm_probability(input_model=gmm, input_=points)
+>>> probs = output['output']
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_generate()](#gmm_generate)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## hmm_train()
+{: #hmm_train }
+
+#### Hidden Markov Model (HMM) Training
+{: #hmm_train_descr }
+
+```python
+>>> from mlpack import hmm_train
+>>> d = hmm_train(batch=False, check_input_matrices=False,
+        copy_all_inputs=False, gaussians=0, input_file='', input_model=None,
+        labels_file='', seed=0, states=0, tolerance=1e-05, type='gaussian',
+        verbose=False)
+>>> output_model = d['output_model']
+```
+
+An implementation of training algorithms for Hidden Markov Models (HMMs). Given labeled or unlabeled data, an HMM can be trained for further use with other mlpack HMM tools. [Detailed documentation](#hmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch` | [`bool`](#doc_bool) | If true, input_file (and if passed, labels_file) are expected to contain a list of files to use as input observation sequences (and label sequences). | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `gaussians` | [`int`](#doc_int) | Number of gaussians in each GMM (necessary when type is 'gmm'). | `0` |
+| `input_file` | [`str`](#doc_str) | File containing input observations. | `**--**` |
+| `input_model` | [`HMMModelType`](#doc_model) | Pre-existing HMM model to initialize training with. | `None` |
+| `labels_file` | [`str`](#doc_str) | Optional file of hidden states, used for labeled training. | `''` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `states` | [`int`](#doc_int) | Number of hidden states in HMM (necessary, unless model_file is specified). | `0` |
+| `tolerance` | [`float`](#doc_float) | Tolerance of the Baum-Welch algorithm. | `1e-05` |
+| `type` | [`str`](#doc_str) | Type of HMM: discrete \| gaussian \| diag_gmm \| gmm. | `'gaussian'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HMMModelType`](#doc_model) | Output for trained HMM. | 
+
+### Detailed documentation
+{: #hmm_train_detailed-documentation }
+
+This program allows a Hidden Markov Model to be trained on labeled or unlabeled data.  It supports four types of HMMs: Discrete HMMs, Gaussian HMMs, GMM HMMs, or Diagonal GMM HMMs
+
+Either one input sequence can be specified (with `input_file`), or, a file containing files in which input sequences can be found (when `input_file`and`batch` are used together).  In addition, labels can be provided in the file specified by `labels_file`, and if `batch` is used, the file given to `labels_file` should contain a list of files of labels corresponding to the sequences in the file given to `input_file`.
+
+The HMM is trained with the Baum-Welch algorithm if no labels are provided.  The tolerance of the Baum-Welch algorithm can be set with the `tolerance`option.  By default, the transition matrix is randomly initialized and the emission distributions are initialized to fit the extent of the data.
+
+Optionally, a pre-created HMM model can be used as a guess for the transition matrix and emission probabilities; this is specifiable with `output_model`.
+
+### See also
+
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_generate()
+{: #hmm_generate }
+
+#### Hidden Markov Model (HMM) Sequence Generator
+{: #hmm_generate_descr }
+
+```python
+>>> from mlpack import hmm_generate
+>>> d = hmm_generate(check_input_matrices=False, copy_all_inputs=False,
+        length=0, model=None, seed=0, start_state=0, verbose=False)
+>>> output = d['output']
+>>> state = d['state']
+```
+
+A utility to generate random sequences from a pre-trained Hidden Markov Model (HMM).  The length of the desired sequence can be specified, and a random sequence of observations is returned. [Detailed documentation](#hmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `length` | [`int`](#doc_int) | Length of sequence to generate. | `**--**` |
+| `model` | [`HMMModelType`](#doc_model) | Trained HMM to generate sequences with. | `**--**` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `start_state` | [`int`](#doc_int) | Starting state of sequence. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save observation sequence to. | 
+| `state` | [`int matrix`](#doc_int_matrix) | Matrix to save hidden state sequence to. | 
+
+### Detailed documentation
+{: #hmm_generate_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as the `model` parameter, and generates a random observation sequence and hidden state sequence based on its parameters. The observation sequence may be saved with the `output` output parameter, and the internal state  sequence may be saved with the `state` output parameter.
+
+The state to start the sequence in may be specified with the `start_state` parameter.
+
+### Example
+For example, to generate a sequence of length 150 from the HMM `'hmm'` and save the observation sequence to `'observations'` and the hidden state sequence to `'states'`, the following command may be used: 
+
+```python
+>>> output = hmm_generate(model=hmm, length=150)
+>>> observations = output['output']
+>>> states = output['state']
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_loglik()
+{: #hmm_loglik }
+
+#### Hidden Markov Model (HMM) Sequence Log-Likelihood
+{: #hmm_loglik_descr }
+
+```python
+>>> from mlpack import hmm_loglik
+>>> d = hmm_loglik(check_input_matrices=False, copy_all_inputs=False,
+        input_=np.empty([0, 0]), input_model=None, verbose=False)
+>>> log_likelihood = d['log_likelihood']
+```
+
+A utility for computing the log-likelihood of a sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observation sequence, this computes and returns the log-likelihood of that sequence being observed from that HMM. [Detailed documentation](#hmm_loglik_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | File containing observations, | `**--**` |
+| `input_model` | [`HMMModelType`](#doc_model) | File containing HMM. | `**--**` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `log_likelihood` | [`float`](#doc_float) | Log-likelihood of the sequence. | 
+
+### Detailed documentation
+{: #hmm_loglik_detailed-documentation }
+
+This utility takes an already-trained HMM, specified with the `input_model` parameter, and evaluates the log-likelihood of a sequence of observations, given with the `input_` parameter.  The computed log-likelihood is given as output.
+
+### Example
+For example, to compute the log-likelihood of the sequence `'seq'` with the pre-trained HMM `'hmm'`, the following command may be used: 
+
+```python
+>>> hmm_loglik(input_=seq, input_model=hmm)
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_viterbi()
+{: #hmm_viterbi }
+
+#### Hidden Markov Model (HMM) Viterbi State Prediction
+{: #hmm_viterbi_descr }
+
+```python
+>>> from mlpack import hmm_viterbi
+>>> d = hmm_viterbi(check_input_matrices=False, copy_all_inputs=False,
+        input_=np.empty([0, 0]), input_model=None, verbose=False)
+>>> output = d['output']
+```
+
+A utility for computing the most probable hidden state sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observed sequence, this uses the Viterbi algorithm to compute and return the most probable hidden state sequence. [Detailed documentation](#hmm_viterbi_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix containing observations, | `**--**` |
+| `input_model` | [`HMMModelType`](#doc_model) | Trained HMM to use. | `**--**` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`int matrix`](#doc_int_matrix) | File to save predicted state sequence to. | 
+
+### Detailed documentation
+{: #hmm_viterbi_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as `input_model`, and evaluates the most probable hidden state sequence of a given sequence of observations (specified as '`input_`, using the Viterbi algorithm.  The computed state sequence may be saved using the `output` output parameter.
+
+### Example
+For example, to predict the state sequence of the observations `'obs'` using the HMM `'hmm'`, storing the predicted state sequence to `'states'`, the following command could be used:
+
+```python
+>>> output = hmm_viterbi(input_=obs, input_model=hmm)
+>>> states = output['output']
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hoeffding_tree()
+{: #hoeffding_tree }
+
+#### Hoeffding trees
+{: #hoeffding_tree_descr }
+
+```python
+>>> from mlpack import hoeffding_tree
+>>> d = hoeffding_tree(batch_mode=False, bins=10,
+        check_input_matrices=False, confidence=0.95, copy_all_inputs=False,
+        info_gain=False, input_model=None, labels=np.empty([0],
+        dtype=np.uint64), max_samples=5000, min_samples=100,
+        numeric_split_strategy='binary', observations_before_binning=100,
+        passes=1, test=np.empty([0, 0]), test_labels=np.empty([0],
+        dtype=np.uint64), training=np.empty([0, 0]), verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of Hoeffding trees, a form of streaming decision tree for classification.  Given labeled data, a Hoeffding tree can be trained and saved for later use, or a pre-trained Hoeffding tree can be used for predicting the classifications of new points. [Detailed documentation](#hoeffding_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_mode` | [`bool`](#doc_bool) | If true, samples will be considered in batch instead of as a stream.  This generally results in better trees but at the cost of memory usage and runtime. | `False` |
+| `bins` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of bins for each numeric split. | `10` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `confidence` | [`float`](#doc_float) | Confidence before splitting (between 0 and 1). | `0.95` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `info_gain` | [`bool`](#doc_bool) | If set, information gain is used instead of Gini impurity for calculating Hoeffding bounds. | `False` |
+| `input_model` | [`HoeffdingTreeModelType`](#doc_model) | Input trained Hoeffding tree model. | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | Labels for training dataset. | `np.empty([0], dtype=np.uint64)` |
+| `max_samples` | [`int`](#doc_int) | Maximum number of samples before splitting. | `5000` |
+| `min_samples` | [`int`](#doc_int) | Minimum number of samples before splitting. | `100` |
+| `numeric_split_strategy` | [`str`](#doc_str) | The splitting strategy to use for numeric features: 'domingos' or 'binary'. | `'binary'` |
+| `observations_before_binning` | [`int`](#doc_int) | If the 'domingos' split strategy is used, this specifies the number of samples observed before binning is performed. | `100` |
+| `passes` | [`int`](#doc_int) | Number of passes to take over the dataset. | `1` |
+| `test` | [`categorical matrix`](#doc_categorical_matrix) | Testing dataset (may be categorical). | `np.empty([0, 0])` |
+| `test_labels` | [`int vector`](#doc_int_vector) | Labels of test data. | `np.empty([0], dtype=np.uint64)` |
+| `training` | [`categorical matrix`](#doc_categorical_matrix) | Training dataset (may be categorical). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HoeffdingTreeModelType`](#doc_model) | Output for trained Hoeffding tree model. | 
+| `predictions` | [`int vector`](#doc_int_vector) | Matrix to output label predictions for test data into. | 
+| `probabilities` | [`matrix`](#doc_matrix) | In addition to predicting labels, provide rediction probabilities in this matrix. | 
+
+### Detailed documentation
+{: #hoeffding_tree_detailed-documentation }
+
+This program implements Hoeffding trees, a form of streaming decision tree suited best for large (or streaming) datasets.  This program supports both categorical and numeric data.  Given an input dataset, this program is able to train the tree with numerous training options, and save the model to a file.  The program is also able to use a trained model or a model from file in order to predict classes for a given test set.
+
+The training file and associated labels are specified with the `training` and `labels` parameters, respectively. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+The training may be performed in batch mode (like a typical decision tree algorithm) by specifying the `batch_mode` option, but this may not be the best option for large datasets.
+
+When a model is trained, it may be saved via the `output_model` output parameter.  A model may be loaded from file for further training or testing with the `input_model` parameter.
+
+Test data may be specified with the `test` parameter, and if performance statistics are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved with the `predictions` output parameter, and class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a Hoeffding tree with confidence 0.99 with data `'dataset'`, saving the trained tree to `'tree'`, the following command may be used:
+
+```python
+>>> output = hoeffding_tree(training=dataset, confidence=0.99)
+>>> tree = output['output_model']
+```
+
+Then, this tree may be used to make predictions on the test set `'test_set'`, saving the predictions into `'predictions'` and the class probabilities into `'class_probs'` with the following command: 
+
+```python
+>>> output = hoeffding_tree(input_model=tree, test=test_set)
+>>> predictions = output['predictions']
+>>> class_probs = output['probabilities']
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [random_forest()](#random_forest)
+ - [Mining High-Speed Data Streams (pdf)](http://dm.cs.washington.edu/papers/vfdt-kdd00.pdf)
+ - [HoeffdingTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp)
+
+## kde()
+{: #kde }
+
+#### Kernel Density Estimation
+{: #kde_descr }
+
+```python
+>>> from mlpack import kde
+>>> d = kde(abs_error=0, algorithm='dual-tree', bandwidth=1,
+        check_input_matrices=False, copy_all_inputs=False,
+        initial_sample_size=100, input_model=None, kernel='gaussian',
+        mc_break_coef=0.4, mc_entry_coef=3, mc_probability=0.95,
+        monte_carlo=False, query=np.empty([0, 0]), reference=np.empty([0, 0]),
+        rel_error=0.05, tree='kd-tree', verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+```
+
+An implementation of kernel density estimation with dual-tree algorithms. Given a set of reference points and query points and a kernel function, this can estimate the density function at the location of each query point using trees; trees that are built can be saved for later use. [Detailed documentation](#kde_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `abs_error` | [`float`](#doc_float) | Relative error tolerance for the prediction. | `0` |
+| `algorithm` | [`str`](#doc_str) | Algorithm to use for the prediction.('dual-tree', 'single-tree'). | `'dual-tree'` |
+| `bandwidth` | [`float`](#doc_float) | Bandwidth of the kernel. | `1` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `initial_sample_size` | [`int`](#doc_int) | Initial sample size for Monte Carlo estimations. | `100` |
+| `input_model` | [`KDEModelType`](#doc_model) | Contains pre-trained KDE model. | `None` |
+| `kernel` | [`str`](#doc_str) | Kernel to use for the prediction.('gaussian', 'epanechnikov', 'laplacian', 'spherical', 'triangular'). | `'gaussian'` |
+| `mc_break_coef` | [`float`](#doc_float) | Controls what fraction of the amount of node's descendants is the limit for the sample size before it recurses. | `0.4` |
+| `mc_entry_coef` | [`float`](#doc_float) | Controls how much larger does the amount of node descendants has to be compared to the initial sample size in order to be a candidate for Monte Carlo estimations. | `3` |
+| `mc_probability` | [`float`](#doc_float) | Probability of the estimation being bounded by relative error when using Monte Carlo estimations. | `0.95` |
+| `monte_carlo` | [`bool`](#doc_bool) | Whether to use Monte Carlo estimations when possible. | `False` |
+| `query` | [`matrix`](#doc_matrix) | Query dataset to KDE on. | `np.empty([0, 0])` |
+| `reference` | [`matrix`](#doc_matrix) | Input reference dataset use for KDE. | `np.empty([0, 0])` |
+| `rel_error` | [`float`](#doc_float) | Relative error tolerance for the prediction. | `0.05` |
+| `tree` | [`str`](#doc_str) | Tree to use for the prediction.('kd-tree', 'ball-tree', 'cover-tree', 'octree', 'r-tree'). | `'kd-tree'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`KDEModelType`](#doc_model) | If specified, the KDE model will be saved here. | 
+| `predictions` | [`vector`](#doc_vector) | Vector to store density predictions. | 
+
+### Detailed documentation
+{: #kde_detailed-documentation }
+
+This program performs a Kernel Density Estimation. KDE is a non-parametric way of estimating probability density function. For each query point the program will estimate its probability density by applying a kernel function to each reference point. The computational complexity of this is O(N^2) where there are N query points and N reference points, but this implementation will typically see better performance as it uses an approximate dual or single tree algorithm for acceleration.
+
+Dual or single tree optimization avoids many barely relevant calculations (as kernel function values decrease with distance), so it is an approximate computation. You can specify the maximum relative error tolerance for each query value with `rel_error` as well as the maximum absolute error tolerance with the parameter `abs_error`. This program runs using an Euclidean metric. Kernel function can be selected using the `kernel` option. You can also choose what which type of tree to use for the dual-tree algorithm with `tree`. It is also possible to select whether to use dual-tree algorithm or single-tree algorithm using the `algorithm` option.
+
+Monte Carlo estimations can be used to accelerate the KDE estimate when the Gaussian Kernel is used. This provides a probabilistic guarantee on the the error of the resulting KDE instead of an absolute guarantee.To enable Monte Carlo estimations, the `monte_carlo` flag can be used, and success probability can be set with the `mc_probability` option. It is possible to set the initial sample size for the Monte Carlo estimation using `initial_sample_size`. This implementation will only consider a node, as a candidate for the Monte Carlo estimation, if its number of descendant nodes is bigger than the initial sample size. This can be controlled using a coefficient that will multiply the initial sample size and can be set using `mc_entry_coef`. To avoid using the same amount of computations an exact approach would take, this program recurses the tree whenever a fraction of the amount of the node's descendant points have already been computed. This fraction is set using `mc_break_coef`.
+
+### Example
+For example, the following will run KDE using the data in `'ref_data'` for training and the data in `'qu_data'` as query data. It will apply an Epanechnikov kernel with a 0.2 bandwidth to each reference point and use a KD-Tree for the dual-tree optimization. The returned predictions will be within 5% of the real KDE value for each query point.
+
+```python
+>>> output = kde(reference=ref_data, query=qu_data, bandwidth=0.2,
+  kernel='epanechnikov', tree='kd-tree', rel_error=0.05)
+>>> out_data = output['predictions']
+```
+
+the predicted density estimations will be stored in `'out_data'`.
+If no `query` is provided, then KDE will be computed on the `reference` dataset.
+It is possible to select either a reference dataset or an input model but not both at the same time. If an input model is selected and parameter values are not set (e.g. `bandwidth`) then default parameter values will be used.
+
+In addition to the last program call, it is also possible to activate Monte Carlo estimations if a Gaussian kernel is used. This can provide faster results, but the KDE will only have a probabilistic guarantee of meeting the desired error bound (instead of an absolute guarantee). The following example will run KDE using a Monte Carlo estimation when possible. The results will be within a 5% of the real KDE value with a 95% probability. Initial sample size for the Monte Carlo estimation will be 200 points and a node will be a candidate for the estimation only when it contains 700 (i.e. 3.5*200) points. If a node contains 700 points and 420 (i.e. 0.6*700) have already been sampled, then the algorithm will recurse instead of keep sampling.
+
+```python
+>>> output = kde(reference=ref_data, query=qu_data, bandwidth=0.2,
+  kernel='gaussian', tree='kd-tree', rel_error=0.05, monte_carlo=,
+  mc_probability=0.95, initial_sample_size=200, mc_entry_coef=3.5,
+  mc_break_coef=0.6)
+>>> out_data = output['predictions']
+```
+
+### See also
+
+ - [knn()](#knn)
+ - [Kernel density estimation on Wikipedia](https://en.wikipedia.org/wiki/Kernel_density_estimation)
+ - [Tree-Independent Dual-Tree Algorithms](https://arxiv.org/pdf/1304.4327)
+ - [Fast High-dimensional Kernel Summations Using the Monte Carlo Multipole Method](http://papers.nips.cc/paper/3539-fast-high-dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf)
+ - [KDE C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kde/kde.hpp)
+
+## kernel_pca()
+{: #kernel_pca }
+
+#### Kernel Principal Components Analysis
+{: #kernel_pca_descr }
+
+```python
+>>> from mlpack import kernel_pca
+>>> d = kernel_pca(bandwidth=1, center=False,
+        check_input_matrices=False, copy_all_inputs=False, degree=1,
+        input_=np.empty([0, 0]), kernel='', kernel_scale=1,
+        new_dimensionality=0, nystroem_method=False, offset=0,
+        sampling='kmeans', verbose=False)
+>>> output = d['output']
+```
+
+An implementation of Kernel Principal Components Analysis (KPCA).  This can be used to perform nonlinear dimensionality reduction or preprocessing on a given dataset. [Detailed documentation](#kernel_pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`float`](#doc_float) | Bandwidth, for 'gaussian' and 'laplacian' kernels. | `1` |
+| `center` | [`bool`](#doc_bool) | If set, the transformed data will be centered about the origin. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `degree` | [`float`](#doc_float) | Degree of polynomial, for 'polynomial' kernel. | `1` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to perform KPCA on. | `**--**` |
+| `kernel` | [`str`](#doc_str) | The kernel to use; see the above documentation for the list of usable kernels. | `**--**` |
+| `kernel_scale` | [`float`](#doc_float) | Scale, for 'hyptan' kernel. | `1` |
+| `new_dimensionality` | [`int`](#doc_int) | If not 0, reduce the dimensionality of the output dataset by ignoring the dimensions with the smallest eigenvalues. | `0` |
+| `nystroem_method` | [`bool`](#doc_bool) | If set, the Nystroem method will be used. | `False` |
+| `offset` | [`float`](#doc_float) | Offset, for 'hyptan' and 'polynomial' kernels. | `0` |
+| `sampling` | [`str`](#doc_str) | Sampling scheme to use for the Nystroem method: 'kmeans', 'random', 'ordered' | `'kmeans'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #kernel_pca_detailed-documentation }
+
+This program performs Kernel Principal Components Analysis (KPCA) on the specified dataset with the specified kernel.  This will transform the data onto the kernel principal components, and optionally reduce the dimensionality by ignoring the kernel principal components with the smallest eigenvalues.
+
+For the case where a linear kernel is used, this reduces to regular PCA.
+
+The kernels that are supported are listed below:
+
+ * 'linear': the standard linear dot product (same as normal PCA):
+    `K(x, y) = x^T y`
+
+ * 'gaussian': a Gaussian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\| ^ 2) / (2 * (bandwidth ^ 2)))`
+
+ * 'polynomial': polynomial kernel; requires offset and degree:
+    `K(x, y) = (x^T y + offset) ^ degree`
+
+ * 'hyptan': hyperbolic tangent kernel; requires scale and offset:
+    `K(x, y) = tanh(scale * (x^T y) + offset)`
+
+ * 'laplacian': Laplacian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\|) / bandwidth)`
+
+ * 'epanechnikov': Epanechnikov kernel; requires bandwidth:
+    `K(x, y) = max(0, 1 - \|\| x - y \|\|^2 / bandwidth^2)`
+
+ * 'cosine': cosine distance:
+    `K(x, y) = 1 - (x^T y) / (\|\| x \|\| * \|\| y \|\|)`
+
+The parameters for each of the kernels should be specified with the options `bandwidth`, `kernel_scale`, `offset`, or `degree` (or a combination of those parameters).
+
+Optionally, the Nystroem method ("Using the Nystroem method to speed up kernel machines", 2001) can be used to calculate the kernel matrix by specifying the `nystroem_method` parameter. This approach works by using a subset of the data as basis to reconstruct the kernel matrix; to specify the sampling scheme, the `sampling` parameter is used.  The sampling scheme for the Nystroem method can be chosen from the following list: 'kmeans', 'random', 'ordered'.
+
+### Example
+For example, the following command will perform KPCA on the dataset `'input'` using the Gaussian kernel, and saving the transformed data to `'transformed'`: 
+
+```python
+>>> output = kernel_pca(input_=input, kernel='gaussian')
+>>> transformed = output['output']
+```
+
+### See also
+
+ - [Kernel principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Kernel_principal_component_analysis)
+ - [Kernel Principal Component Analysis (pdf)](https://pca.narod.ru/scholkopf_kernel.pdf)
+ - [KernelPCA class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+
+## kmeans()
+{: #kmeans }
+
+#### K-Means Clustering
+{: #kmeans_descr }
+
+```python
+>>> from mlpack import kmeans
+>>> d = kmeans(algorithm='naive', allow_empty_clusters=False,
+        check_input_matrices=False, clusters=0, copy_all_inputs=False,
+        in_place=False, initial_centroids=np.empty([0, 0]), input_=np.empty([0,
+        0]), kill_empty_clusters=False, kmeans_plus_plus=False,
+        labels_only=False, max_iterations=1000, percentage=0.02,
+        refined_start=False, samplings=100, seed=0, verbose=False)
+>>> centroid = d['centroid']
+>>> output = d['output']
+```
+
+An implementation of several strategies for efficient k-means clustering. Given a dataset and a value of k, this computes and returns a k-means clustering on that data. [Detailed documentation](#kmeans_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`str`](#doc_str) | Algorithm to use for the Lloyd iteration ('naive', 'pelleg-moore', 'elkan', 'hamerly', 'dualtree', or 'dualtree-covertree'). | `'naive'` |
+| `allow_empty_clusters` | [`bool`](#doc_bool) | Allow empty clusters to be persist. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `clusters` | [`int`](#doc_int) | Number of clusters to find (0 autodetects from initial centroids). | `**--**` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `in_place` | [`bool`](#doc_bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden. (Do not use in Python.) | `False` |
+| `initial_centroids` | [`matrix`](#doc_matrix) | Start with the specified initial centroids. | `np.empty([0, 0])` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to perform clustering on. | `**--**` |
+| `kill_empty_clusters` | [`bool`](#doc_bool) | Remove empty clusters when they occur. | `False` |
+| `kmeans_plus_plus` | [`bool`](#doc_bool) | Use the k-means++ initialization strategy to choose initial points. | `False` |
+| `labels_only` | [`bool`](#doc_bool) | Only output labels into output file. | `False` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations before k-means terminates. | `1000` |
+| `percentage` | [`float`](#doc_float) | Percentage of dataset to use for each refined start sampling (use when --refined_start is specified). | `0.02` |
+| `refined_start` | [`bool`](#doc_bool) | Use the refined initial point strategy by Bradley and Fayyad to choose initial points. | `False` |
+| `samplings` | [`int`](#doc_int) | Number of samplings to perform for refined start (use when --refined_start is specified). | `100` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`matrix`](#doc_matrix) | If specified, the centroids of each cluster will  be written to the given file. | 
+| `output` | [`matrix`](#doc_matrix) | Matrix to store output labels or labeled data to. | 
+
+### Detailed documentation
+{: #kmeans_detailed-documentation }
+
+This program performs K-Means clustering on the given dataset.  It can return the learned cluster assignments, and the centroids of the clusters.  Empty clusters are not allowed by default; when a cluster becomes empty, the point furthest from the centroid of the cluster with maximum variance is taken to fill that cluster.
+
+Optionally, the strategy to choose initial centroids can be specified.  The k-means++ algorithm can be used to choose initial centroids with the `kmeans_plus_plus` parameter.  The Bradley and Fayyad approach ("Refining initial points for k-means clustering", 1998) can be used to select initial points by specifying the `refined_start` parameter.  This approach works by taking random samplings of the dataset; to specify the number of samplings, the `samplings` parameter is used, and to specify the percentage of the dataset to be used in each sample, the `percentage` parameter is used (it should be a value between 0.0 and 1.0).
+
+There are several options available for the algorithm used for each Lloyd iteration, specified with the `algorithm`  option.  The standard O(kN) approach can be used ('naive').  Other options include the Pelleg-Moore tree-based algorithm ('pelleg-moore'), Elkan's triangle-inequality based algorithm ('elkan'), Hamerly's modification to Elkan's algorithm ('hamerly'), the dual-tree k-means algorithm ('dualtree'), and the dual-tree k-means algorithm using the cover tree ('dualtree-covertree').
+
+The behavior for when an empty cluster is encountered can be modified with the `allow_empty_clusters` option.  When this option is specified and there is a cluster owning no points at the end of an iteration, that cluster's centroid will simply remain in its position from the previous iteration. If the `kill_empty_clusters` option is specified, then when a cluster owns no points at the end of an iteration, the cluster centroid is simply filled with DBL_MAX, killing it and effectively reducing k for the rest of the computation.  Note that the default option when neither empty cluster option is specified can be time-consuming to calculate; therefore, specifying either of these parameters will often accelerate runtime.
+
+Initial clustering assignments may be specified using the `initial_centroids` parameter, and the maximum number of iterations may be specified with the `max_iterations` parameter.
+
+### Example
+As an example, to use Hamerly's algorithm to perform k-means clustering with k=10 on the dataset `'data'`, saving the centroids to `'centroids'` and the assignments for each point to `'assignments'`, the following command could be used:
+
+```python
+>>> output = kmeans(input_=data, clusters=10)
+>>> assignments = output['output']
+>>> centroids = output['centroid']
+```
+
+To run k-means on that same dataset with initial centroids specified in `'initial'` with a maximum of 500 iterations, storing the output centroids in `'final'` the following command may be used:
+
+```python
+>>> output = kmeans(input_=data, initial_centroids=initial, clusters=10,
+  max_iterations=500)
+>>> final = output['centroid']
+```
+
+### See also
+
+ - [dbscan()](#dbscan)
+ - [k-means++](https://en.wikipedia.org/wiki/K-means%2B%2B)
+ - [Using the triangle inequality to accelerate k-means (pdf)](https://cdn.aaai.org/ICML/2003/ICML03-022.pdf)
+ - [Making k-means even faster (pdf)](https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12)
+ - [Accelerating exact k-means algorithms with geometric reasoning (pdf)](http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/ftp/usr0/ftp/2000/CMU-CS-00-105.pdf)
+ - [A dual-tree algorithm for fast k-means clustering with large k (pdf)](http://www.ratml.org/pub/pdf/2017dual.pdf)
+ - [KMeans class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kmeans/kmeans.hpp)
+
+## lars()
+{: #lars }
+
+#### LARS
+{: #lars_descr }
+
+```python
+>>> from mlpack import lars
+>>> d = lars(check_input_matrices=False, copy_all_inputs=False,
+        input_=np.empty([0, 0]), input_model=None, lambda1=0, lambda2=0,
+        no_intercept=False, no_normalize=False, responses=np.empty([0, 0]),
+        test=np.empty([0, 0]), use_cholesky=False, verbose=False)
+>>> output_model = d['output_model']
+>>> output_predictions = d['output_predictions']
+```
+
+An implementation of Least Angle Regression (Stagewise/laSso), also known as LARS.  This can train a LARS/LASSO/Elastic Net model and use that model or a pre-trained model to output regression predictions for a test set. [Detailed documentation](#lars_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix of covariates (X). | `np.empty([0, 0])` |
+| `input_model` | [`LARS<>Type`](#doc_model) | Trained LARS model to use. | `None` |
+| `lambda1` | [`float`](#doc_float) | Regularization parameter for l1-norm penalty. | `0` |
+| `lambda2` | [`float`](#doc_float) | Regularization parameter for l2-norm penalty. | `0` |
+| `no_intercept` | [`bool`](#doc_bool) | Do not fit an intercept in the model. | `False` |
+| `no_normalize` | [`bool`](#doc_bool) | Do not normalize data to unit variance before modeling. | `False` |
+| `responses` | [`matrix`](#doc_matrix) | Matrix of responses/observations (y). | `np.empty([0, 0])` |
+| `test` | [`matrix`](#doc_matrix) | Matrix containing points to regress on (test points). | `np.empty([0, 0])` |
+| `use_cholesky` | [`bool`](#doc_bool) | Use Cholesky decomposition during computation rather than explicitly computing the full Gram matrix. | `False` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LARS<>Type`](#doc_model) | Output LARS model. | 
+| `output_predictions` | [`matrix`](#doc_matrix) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #lars_detailed-documentation }
+
+An implementation of LARS: Least Angle Regression (Stagewise/laSso).  This is a stage-wise homotopy-based algorithm for L1-regularized linear regression (LASSO) and L1+L2-regularized linear regression (Elastic Net).
+
+This program is able to train a LARS/LASSO/Elastic Net model or load a model from file, output regression predictions for a test set, and save the trained model to a file.  The LARS algorithm is described in more detail below:
+
+Let X be a matrix where each row is a point and each column is a dimension, and let y be a vector of targets.
+
+The Elastic Net problem is to solve
+
+  min_beta 0.5 \|\| X * beta - y \|\|_2^2 + lambda_1 \|\|beta\|\|_1 +
+      0.5 lambda_2 \|\|beta\|\|_2^2
+
+If lambda1 > 0 and lambda2 = 0, the problem is the LASSO.
+If lambda1 > 0 and lambda2 > 0, the problem is the Elastic Net.
+If lambda1 = 0 and lambda2 > 0, the problem is ridge regression.
+If lambda1 = 0 and lambda2 = 0, the problem is unregularized linear regression.
+
+For efficiency reasons, it is not recommended to use this algorithm with `lambda1` = 0.  In that case, use the 'linear_regression' program, which implements both unregularized linear regression and ridge regression.
+
+To train a LARS/LASSO/Elastic Net model, the `input_` and `responses` parameters must be given.  The `lambda1`, `lambda2`, and `use_cholesky` parameters control the training options.  A trained model can be saved with the `output_model`.  If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `output_predictions` output parameter.
+
+### Example
+For example, the following command trains a model on the data `'data'` and responses `'responses'` with lambda1 set to 0.4 and lambda2 set to 0 (so, LASSO is being solved), and then the model is saved to `'lasso_model'`:
+
+```python
+>>> output = lars(input_=data, responses=responses, lambda1=0.4, lambda2=0)
+>>> lasso_model = output['output_model']
+```
+
+The following command uses the `'lasso_model'` to provide predicted responses for the data `'test'` and save those responses to `'test_predictions'`: 
+
+```python
+>>> output = lars(input_model=lasso_model, test=test)
+>>> test_predictions = output['output_predictions']
+```
+
+### See also
+
+ - [linear_regression()](#linear_regression)
+ - [Least angle regression (pdf)](https://mlpack.org/papers/lars.pdf)
+ - [LARS C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lars/lars.hpp)
+
+## linear_svm()
+{: #linear_svm }
+
+#### Linear SVM is an L2-regularized support vector machine.
+{: #linear_svm_descr }
+
+```python
+>>> from mlpack import linear_svm
+>>> d = linear_svm(check_input_matrices=False, copy_all_inputs=False,
+        delta=1, epochs=50, input_model=None, labels=np.empty([0],
+        dtype=np.uint64), lambda_=0.0001, max_iterations=10000,
+        no_intercept=False, num_classes=0, optimizer='lbfgs', seed=0,
+        shuffle=False, step_size=0.01, test=np.empty([0, 0]),
+        test_labels=np.empty([0], dtype=np.uint64), tolerance=1e-10,
+        training=np.empty([0, 0]), verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of linear SVM for multiclass classification. Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#linear_svm_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `delta` | [`float`](#doc_float) | Margin of difference between correct class and other classes. | `1` |
+| `epochs` | [`int`](#doc_int) | Maximum number of full epochs over dataset for psgd | `50` |
+| `input_model` | [`LinearSVMModelType`](#doc_model) | Existing model (parameters). | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). | `np.empty([0], dtype=np.uint64)` |
+| `lambda_` | [`float`](#doc_float) | L2-regularization parameter for training. | `0.0001` |
+| `max_iterations` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `no_intercept` | [`bool`](#doc_bool) | Do not add the intercept term to the model. | `False` |
+| `num_classes` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `optimizer` | [`str`](#doc_str) | Optimizer to use for training ('lbfgs' or 'psgd'). | `'lbfgs'` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `shuffle` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for parallel SGD. | `False` |
+| `step_size` | [`float`](#doc_float) | Step size for parallel SGD optimizer. | `0.01` |
+| `test` | [`matrix`](#doc_matrix) | Matrix containing test dataset. | `np.empty([0, 0])` |
+| `test_labels` | [`int vector`](#doc_int_vector) | Matrix containing test labels. | `np.empty([0], dtype=np.uint64)` |
+| `tolerance` | [`float`](#doc_float) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`matrix`](#doc_matrix) | A matrix containing the training set (the matrix of predictors, X). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LinearSVMModelType`](#doc_model) | Output for trained linear svm model. | 
+| `predictions` | [`int vector`](#doc_int_vector) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`matrix`](#doc_matrix) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #linear_svm_detailed-documentation }
+
+An implementation of linear SVMs that uses either L-BFGS or parallel SGD (stochastic gradient descent) to train the model.
+
+This program allows loading a linear SVM model (via the `input_model` parameter) or training a linear SVM model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained linear SVM model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate vector of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda_` option, and the number of classes can be manually specified with the `num_classes`and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.Margin of difference between correct class and other classes can be specified with the `delta` option.The optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'psgd' (parallel stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the parallel SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer and the maximum number of epochs (specified with `epochs`). If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+Optionally, the model can be used to predict the labels for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing linear SVM model is given with the `input_model` parameter.  The output predictions from the linear SVM model may be saved with the `predictions` parameter.
+
+### Example
+As an example, to train a LinaerSVM on the data '`'data'`' with labels '`'labels'`' with L2 regularization of 0.1, saving the model to '`'lsvm_model'`', the following command may be used:
+
+```python
+>>> output = linear_svm(training=data, labels=labels, lambda_=0.1, delta=1,
+  num_classes=0)
+>>> lsvm_model = output['output_model']
+```
+
+Then, to use that model to predict classes for the dataset '`'test'`', storing the output predictions in '`'predictions'`', the following command may be used: 
+
+```python
+>>> output = linear_svm(input_model=lsvm_model, test=test)
+>>> predictions = output['predictions']
+```
+
+### See also
+
+ - [random_forest()](#random_forest)
+ - [logistic_regression()](#logistic_regression)
+ - [LinearSVM on Wikipedia](https://en.wikipedia.org/wiki/Support-vector_machine)
+ - [LinearSVM C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_svm/linear_svm.hpp)
+
+## lmnn()
+{: #lmnn }
+
+#### Large Margin Nearest Neighbors (LMNN)
+{: #lmnn_descr }
+
+```python
+>>> from mlpack import lmnn
+>>> d = lmnn(batch_size=50, center=False, check_input_matrices=False,
+        copy_all_inputs=False, distance=np.empty([0, 0]), input_=np.empty([0,
+        0]), k=1, labels=np.empty([0], dtype=np.uint64), linear_scan=False,
+        max_iterations=100000, normalize=False, optimizer='amsgrad', passes=50,
+        print_accuracy=False, range=1, rank=0, regularization=0.5, seed=0,
+        step_size=0.01, tolerance=1e-07, verbose=False)
+>>> centered_data = d['centered_data']
+>>> output = d['output']
+>>> transformed_data = d['transformed_data']
+```
+
+An implementation of Large Margin Nearest Neighbors (LMNN), a distance learning technique.  Given a labeled dataset, this learns a transformation of the data that improves k-nearest-neighbor performance; this can be useful as a preprocessing step. [Detailed documentation](#lmnn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `center` | [`bool`](#doc_bool) | Perform mean-centering on the dataset. It is useful when the centroid of the data is far from the origin. | `False` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `distance` | [`matrix`](#doc_matrix) | Initial distance matrix to be used as starting point | `np.empty([0, 0])` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to run LMNN on. | `**--**` |
+| `k` | [`int`](#doc_int) | Number of target neighbors to use for each datapoint. | `1` |
+| `labels` | [`int vector`](#doc_int_vector) | Labels for input dataset. | `np.empty([0], dtype=np.uint64)` |
+| `linear_scan` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `False` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations for L-BFGS (0 indicates no limit). | `100000` |
+| `normalize` | [`bool`](#doc_bool) | Use a normalized starting point for optimization. Itis useful for when points are far apart, or when SGD is returning NaN. | `False` |
+| `optimizer` | [`str`](#doc_str) | Optimizer to use; 'amsgrad', 'bbsgd', 'sgd', or 'lbfgs'. | `'amsgrad'` |
+| `passes` | [`int`](#doc_int) | Maximum number of full passes over dataset for AMSGrad, BB_SGD and SGD. | `50` |
+| `print_accuracy` | [`bool`](#doc_bool) | Print accuracies on initial and transformed dataset | `False` |
+| `range` | [`int`](#doc_int) | Number of iterations after which impostors needs to be recalculated | `1` |
+| `rank` | [`int`](#doc_int) | Rank of distance matrix to be optimized.  | `0` |
+| `regularization` | [`float`](#doc_float) | Regularization for LMNN objective function  | `0.5` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`float`](#doc_float) | Step size for AMSGrad, BB_SGD and SGD (alpha). | `0.01` |
+| `tolerance` | [`float`](#doc_float) | Maximum tolerance for termination of AMSGrad, BB_SGD, SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centered_data` | [`matrix`](#doc_matrix) | Output matrix for mean-centered dataset. | 
+| `output` | [`matrix`](#doc_matrix) | Output matrix for learned distance matrix. | 
+| `transformed_data` | [`matrix`](#doc_matrix) | Output matrix for transformed dataset. | 
+
+### Detailed documentation
+{: #lmnn_detailed-documentation }
+
+This program implements Large Margin Nearest Neighbors, a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset.  The method employes the strategy of reducing distance between similar labeled data points (a.k.a target neighbors) and increasing distance between differently labeled points (a.k.a impostors) using standard optimization techniques over the gradient of the distance between data points.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input_`), or alternatively as a separate matrix (specified with `labels`).  Additionally, a starting point for optimization (specified with `distance`can be given, having (r x d) dimensionality.  Here r should satisfy 1 <= r <= d, Consequently a Low-Rank matrix will be optimized. Alternatively, Low-Rank distance can be learned by specifying the `rank`parameter (A Low-Rank matrix with uniformly distributed values will be used as initial learning point). 
+
+The program also requires number of targets neighbors to work with ( specified with `k`), A regularization parameter can also be passed, It acts as a trade of between the pulling and pushing terms (specified with `regularization`), In addition, this implementation of LMNN includes a parameter to decide the interval after which impostors must be re-calculated (specified with `range`).
+
+Output can either be the learned distance matrix (specified with `output`), or the transformed dataset  (specified with `transformed_data`), or both. Additionally mean-centered dataset (specified with `centered_data`) can be accessed given mean-centering (specified with `center`) is performed on the dataset. Accuracy on initial dataset and final transformed dataset can be printed by specifying the `print_accuracy`parameter. 
+
+This implementation of LMNN uses AdaGrad, BigBatch_SGD, stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer. 
+
+AdaGrad, specified by the value 'adagrad' for the parameter `optimizer`, uses maximum of past squared gradients. It primarily on six parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`). Inaddition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+BigBatch_SGD, specified by the value 'bbsgd' for the parameter `optimizer`, depends primarily on four parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. Furthermore, mean-centering can be performed on the dataset by specifying the `center`parameter. 
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `max_iterations`, `tolerance`(the optimization is terminated when the gradient norm is below this value).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.  In addition, a normalized starting point can be used by specifying the `normalize` parameter.
+
+By default, the AMSGrad optimizer is used.
+
+### Example
+Example - Let's say we want to learn distance on iris dataset with number of targets as 3 using BigBatch_SGD optimizer. A simple call for the same will look like: 
+
+```python
+>>> output = lmnn(input_=iris, labels=iris_labels, k=3, optimizer='bbsgd')
+>>> output = output['output']
+```
+
+An another program call making use of range & regularization parameter with dataset having labels as last column can be made as: 
+
+```python
+>>> output = lmnn(input_=letter_recognition, k=5, range=10,
+  regularization=0.4)
+>>> output = output['output']
+```
+
+### See also
+
+ - [nca()](#nca)
+ - [Large margin nearest neighbor on Wikipedia](https://en.wikipedia.org/wiki/Large_margin_nearest_neighbor)
+ - [Distance metric learning for large margin nearest neighbor classification (pdf)](http://papers.nips.cc/paper/2795-distance-metric-learning-for-large-margin-nearest-neighbor-classification.pdf)
+ - [LMNN C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lmnn/lmnn.hpp)
+
+## local_coordinate_coding()
+{: #local_coordinate_coding }
+
+#### Local Coordinate Coding
+{: #local_coordinate_coding_descr }
+
+```python
+>>> from mlpack import local_coordinate_coding
+>>> d = local_coordinate_coding(atoms=0, check_input_matrices=False,
+        copy_all_inputs=False, initial_dictionary=np.empty([0, 0]),
+        input_model=None, lambda_=0, max_iterations=0, normalize=False, seed=0,
+        test=np.empty([0, 0]), tolerance=0.01, training=np.empty([0, 0]),
+        verbose=False)
+>>> codes = d['codes']
+>>> dictionary = d['dictionary']
+>>> output_model = d['output_model']
+```
+
+An implementation of Local Coordinate Coding (LCC), a data transformation technique.  Given input data, this transforms each point to be expressed as a linear combination of a few points in the dataset; once an LCC model is trained, it can be used to transform points later also. [Detailed documentation](#local_coordinate_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`int`](#doc_int) | Number of atoms in the dictionary. | `0` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `initial_dictionary` | [`matrix`](#doc_matrix) | Optional initial dictionary. | `np.empty([0, 0])` |
+| `input_model` | [`LocalCoordinateCodingType`](#doc_model) | Input LCC model. | `None` |
+| `lambda_` | [`float`](#doc_float) | Weighted l1-norm regularization parameter. | `0` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations for LCC (0 indicates no limit). | `0` |
+| `normalize` | [`bool`](#doc_bool) | If set, the input data matrix will be normalized before coding. | `False` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`matrix`](#doc_matrix) | Test points to encode. | `np.empty([0, 0])` |
+| `tolerance` | [`float`](#doc_float) | Tolerance for objective function. | `0.01` |
+| `training` | [`matrix`](#doc_matrix) | Matrix of training data (X). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`matrix`](#doc_matrix) | Output codes matrix. | 
+| `dictionary` | [`matrix`](#doc_matrix) | Output dictionary matrix. | 
+| `output_model` | [`LocalCoordinateCodingType`](#doc_model) | Output for trained LCC model. | 
+
+### Detailed documentation
+{: #local_coordinate_coding_detailed-documentation }
+
+An implementation of Local Coordinate Coding (LCC), which codes data that approximately lives on a manifold using a variation of l1-norm regularized sparse coding.  Given a dense data matrix X with n points and d dimensions, LCC seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a coding matrix Z with n points in k dimensions.  Because of the regularization method used, the atoms in D should lie close to the manifold on which the data points lie.
+
+The original data matrix X can then be reconstructed as D * Z.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a coding step, which updates the coding matrix Z.
+
+To run this program, the input matrix X must be specified (with -i), along with the number of atoms in the dictionary (-k).  An initial dictionary may also be specified with the `initial_dictionary` parameter.  The l1-norm regularization parameter is specified with the `lambda_` parameter.
+
+### Example
+For example, to run LCC on the dataset `'data'` using 200 atoms and an l1-regularization parameter of 0.1, saving the dictionary `dictionary` and the codes into `codes`, use
+
+```python
+>>> output = local_coordinate_coding(training=data, atoms=200, lambda_=0.1)
+>>> dict = output['dictionary']
+>>> codes = output['codes']
+```
+
+The maximum number of iterations may be specified with the `max_iterations` parameter. Optionally, the input data matrix X can be normalized before coding with the `normalize` parameter.
+
+An LCC model may be saved using the `output_model` output parameter.  Then, to encode new points from the dataset `'points'` with the previously saved model `'lcc_model'`, saving the new codes to `'new_codes'`, the following command can be used:
+
+```python
+>>> output = local_coordinate_coding(input_model=lcc_model, test=points)
+>>> new_codes = output['codes']
+```
+
+### See also
+
+ - [sparse_coding()](#sparse_coding)
+ - [Nonlinear learning using local coordinate coding (pdf)](https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-coordinate-coding.pdf)
+ - [LocalCoordinateCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding.hpp)
+
+## logistic_regression()
+{: #logistic_regression }
+
+#### L2-regularized Logistic Regression and Prediction
+{: #logistic_regression_descr }
+
+```python
+>>> from mlpack import logistic_regression
+>>> d = logistic_regression(batch_size=64, check_input_matrices=False,
+        copy_all_inputs=False, decision_boundary=0.5, input_model=None,
+        labels=np.empty([0], dtype=np.uint64), lambda_=0, max_iterations=10000,
+        optimizer='lbfgs', print_training_accuracy=False, step_size=0.01,
+        test=np.empty([0, 0]), tolerance=1e-10, training=np.empty([0, 0]),
+        verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of L2-regularized logistic regression for two-class classification.  Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#logistic_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`int`](#doc_int) | Batch size for SGD. | `64` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `decision_boundary` | [`float`](#doc_float) | Decision boundary for prediction; if the logistic function for a point is less than the boundary, the class is taken to be 0; otherwise, the class is 1. | `0.5` |
+| `input_model` | [`LogisticRegression<>Type`](#doc_model) | Existing model (parameters). | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). | `np.empty([0], dtype=np.uint64)` |
+| `lambda_` | [`float`](#doc_float) | L2-regularization parameter for training. | `0` |
+| `max_iterations` | [`int`](#doc_int) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `optimizer` | [`str`](#doc_str) | Optimizer to use for training ('lbfgs' or 'sgd'). | `'lbfgs'` |
+| `print_training_accuracy` | [`bool`](#doc_bool) | If set, then the accuracy of the model on the training set will be printed (verbose must also be specified). | `False` |
+| `step_size` | [`float`](#doc_float) | Step size for SGD optimizer. | `0.01` |
+| `test` | [`matrix`](#doc_matrix) | Matrix containing test dataset. | `np.empty([0, 0])` |
+| `tolerance` | [`float`](#doc_float) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`matrix`](#doc_matrix) | A matrix containing the training set (the matrix of predictors, X). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LogisticRegression<>Type`](#doc_model) | Output for trained logistic regression model. | 
+| `predictions` | [`int vector`](#doc_int_vector) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`matrix`](#doc_matrix) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #logistic_regression_detailed-documentation }
+
+An implementation of L2-regularized logistic regression using either the L-BFGS optimizer or SGD (stochastic gradient descent).  This solves the regression problem
+
+  y = (1 / 1 + e^-(X * b)).
+
+In this setting, y corresponds to class labels and X corresponds to data.
+
+This program allows loading a logistic regression model (via the `input_model` parameter) or training a logistic regression model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained logistic regression model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda_` option, and the optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'sgd' (stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer.  The batch size for SGD is controlled with the `batch_size` parameter. If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+For SGD, an iteration refers to a single point. So to take a single pass over the dataset with SGD, `max_iterations` should be set to the number of points in the dataset.
+
+Optionally, the model can be used to predict the responses for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing logistic regression model is given with the `input_model` parameter.  The output predictions from the logistic regression model may be saved with the `predictions` parameter.
+
+This implementation of logistic regression does not support the general multi-class case but instead only the two-class case.  Any labels must be either 0 or 1.  For more classes, see the softmax regression implementation.
+
+### Example
+As an example, to train a logistic regression model on the data '`'data'`' with labels '`'labels'`' with L2 regularization of 0.1, saving the model to '`'lr_model'`', the following command may be used:
+
+```python
+>>> output = logistic_regression(training=data, labels=labels, lambda_=0.1,
+  print_training_accuracy=True)
+>>> lr_model = output['output_model']
+```
+
+Then, to use that model to predict classes for the dataset '`'test'`', storing the output predictions in '`'predictions'`', the following command may be used: 
+
+```python
+>>> output = logistic_regression(input_model=lr_model, test=test)
+>>> predictions = output['predictions']
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+ - [:LogisticRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/logistic_regression/logistic_regression.hpp)
+
+## lsh()
+{: #lsh }
+
+#### K-Approximate-Nearest-Neighbor Search with LSH
+{: #lsh_descr }
+
+```python
+>>> from mlpack import lsh
+>>> d = lsh(bucket_size=500, check_input_matrices=False,
+        copy_all_inputs=False, hash_width=0, input_model=None, k=0,
+        num_probes=0, projections=10, query=np.empty([0, 0]),
+        reference=np.empty([0, 0]), second_hash_size=99901, seed=0, tables=30,
+        true_neighbors=np.empty([0, 0], dtype=np.uint64), verbose=False)
+>>> distances = d['distances']
+>>> neighbors = d['neighbors']
+>>> output_model = d['output_model']
+```
+
+An implementation of approximate k-nearest-neighbor search with locality-sensitive hashing (LSH).  Given a set of reference points and a set of query points, this will compute the k approximate nearest neighbors of each query point in the reference set; models can be saved for future use. [Detailed documentation](#lsh_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bucket_size` | [`int`](#doc_int) | The size of a bucket in the second level hash. | `500` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `hash_width` | [`float`](#doc_float) | The hash width for the first-level hashing in the LSH preprocessing. By default, the LSH class automatically estimates a hash width for its use. | `0` |
+| `input_model` | [`LSHSearch<>Type`](#doc_model) | Input LSH model. | `None` |
+| `k` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `num_probes` | [`int`](#doc_int) | Number of additional probes for multiprobe LSH; if 0, traditional LSH is used. | `0` |
+| `projections` | [`int`](#doc_int) | The number of hash functions for each table | `10` |
+| `query` | [`matrix`](#doc_matrix) | Matrix containing query points (optional). | `np.empty([0, 0])` |
+| `reference` | [`matrix`](#doc_matrix) | Matrix containing the reference dataset. | `np.empty([0, 0])` |
+| `second_hash_size` | [`int`](#doc_int) | The size of the second level hash table. | `99901` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tables` | [`int`](#doc_int) | The number of hash tables to be used. | `30` |
+| `true_neighbors` | [`int matrix`](#doc_int_matrix) | Matrix of true neighbors to compute recall with (the recall is printed when -v is specified). | `np.empty([0, 0], dtype=np.uint64)` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`matrix`](#doc_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`int matrix`](#doc_int_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`LSHSearch<>Type`](#doc_model) | Output for trained LSH model. | 
+
+### Detailed documentation
+{: #lsh_detailed-documentation }
+
+This program will calculate the k approximate-nearest-neighbors of a set of points using locality-sensitive hashing. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. 
+
+### Example
+For example, the following will return 5 neighbors from the data for each point in `'input'` and store the distances in `'distances'` and the neighbors in `'neighbors'`:
+
+```python
+>>> output = lsh(k=5, reference=input)
+>>> distances = output['distances']
+>>> neighbors = output['neighbors']
+```
+
+The output is organized such that row i and column j in the neighbors output corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output file corresponds to the distance between those two points.
+
+Because this is approximate-nearest-neighbors search, results may be different from run to run.  Thus, the `seed` parameter can be specified to set the random seed.
+
+This program also has many other parameters to control its functionality; see the parameter-specific documentation for more information.
+
+### See also
+
+ - [knn()](#knn)
+ - [krann()](#krann)
+ - [Locality-sensitive hashing on Wikipedia](https://en.wikipedia.org/wiki/Locality-sensitive_hashing)
+ - [Locality-sensitive hashing scheme based on p-stable  distributions(pdf)](https://www.mlpack.org/papers/lsh.pdf)
+ - [LSHSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lsh/lsh.hpp)
+
+## mean_shift()
+{: #mean_shift }
+
+#### Mean Shift Clustering
+{: #mean_shift_descr }
+
+```python
+>>> from mlpack import mean_shift
+>>> d = mean_shift(check_input_matrices=False, copy_all_inputs=False,
+        force_convergence=False, in_place=False, input_=np.empty([0, 0]),
+        labels_only=False, max_iterations=1000, radius=0, verbose=False)
+>>> centroid = d['centroid']
+>>> output = d['output']
+```
+
+A fast implementation of mean-shift clustering using dual-tree range search.  Given a dataset, this uses the mean shift algorithm to produce and return a clustering of the data. [Detailed documentation](#mean_shift_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `force_convergence` | [`bool`](#doc_bool) | If specified, the mean shift algorithm will continue running regardless of max_iterations until the clusters converge. | `False` |
+| `in_place` | [`bool`](#doc_bool) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden.  (Do not use with Python.) | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to perform clustering on. | `**--**` |
+| `labels_only` | [`bool`](#doc_bool) | If specified, only the output labels will be written to the file specified by --output_file. | `False` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations before mean shift terminates. | `1000` |
+| `radius` | [`float`](#doc_float) | If the distance between two centroids is less than the given radius, one will be removed.  A radius of 0 or less means an estimate will be calculated and used for the radius. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`matrix`](#doc_matrix) | If specified, the centroids of each cluster will be written to the given matrix. | 
+| `output` | [`matrix`](#doc_matrix) | Matrix to write output labels or labeled data to. | 
+
+### Detailed documentation
+{: #mean_shift_detailed-documentation }
+
+This program performs mean shift clustering on the given dataset, storing the learned cluster assignments either as a column of labels in the input dataset or separately.
+
+The input dataset should be specified with the `input_` parameter, and the radius used for search can be specified with the `radius` parameter.  The maximum number of iterations before algorithm termination is controlled with the `max_iterations` parameter.
+
+The output labels may be saved with the `output` output parameter and the centroids of each cluster may be saved with the `centroid` output parameter.
+
+### Example
+For example, to run mean shift clustering on the dataset `'data'` and store the centroids to `'centroids'`, the following command may be used: 
+
+```python
+>>> output = mean_shift(input_=data)
+>>> centroids = output['centroid']
+```
+
+### See also
+
+ - [kmeans()](#kmeans)
+ - [dbscan()](#dbscan)
+ - [Mean shift on Wikipedia](https://en.wikipedia.org/wiki/Mean_shift)
+ - [Mean Shift, Mode Seeking, and Clustering (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=1c168275c59ba382588350ee1443537f59978183)
+ - [mlpack::mean_shift::MeanShift C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/mean_shift/mean_shift.hpp)
+
+## nbc()
+{: #nbc }
+
+#### Parametric Naive Bayes Classifier
+{: #nbc_descr }
+
+```python
+>>> from mlpack import nbc
+>>> d = nbc(check_input_matrices=False, copy_all_inputs=False,
+        incremental_variance=False, input_model=None, labels=np.empty([0],
+        dtype=np.uint64), test=np.empty([0, 0]), training=np.empty([0, 0]),
+        verbose=False)
+>>> output = d['output']
+>>> output_model = d['output_model']
+>>> output_probs = d['output_probs']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of the Naive Bayes Classifier, used for classification. Given labeled data, an NBC model can be trained and saved, or, a pre-trained model can be used for classification. [Detailed documentation](#nbc_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `incremental_variance` | [`bool`](#doc_bool) | The variance of each class will be calculated incrementally. | `False` |
+| `input_model` | [`NBCModelType`](#doc_model) | Input Naive Bayes model. | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | A file containing labels for the training set. | `np.empty([0], dtype=np.uint64)` |
+| `test` | [`matrix`](#doc_matrix) | A matrix containing the test set. | `np.empty([0, 0])` |
+| `training` | [`matrix`](#doc_matrix) | A matrix containing the training set. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`int vector`](#doc_int_vector) | The matrix in which the predicted labels for the test set will be written (deprecated). | 
+| `output_model` | [`NBCModelType`](#doc_model) | File to save trained Naive Bayes model to. | 
+| `output_probs` | [`matrix`](#doc_matrix) | The matrix in which the predicted probability of labels for the test set will be written (deprecated). | 
+| `predictions` | [`int vector`](#doc_int_vector) | The matrix in which the predicted labels for the test set will be written. | 
+| `probabilities` | [`matrix`](#doc_matrix) | The matrix in which the predicted probability of labels for the test set will be written. | 
+
+### Detailed documentation
+{: #nbc_detailed-documentation }
+
+This program trains the Naive Bayes classifier on the given labeled training set, or loads a model from the given model file, and then may use that trained model to classify the points in a given test set.
+
+The training set is specified with the `training` parameter.  Labels may be either the last row of the training set, or alternately the `labels` parameter may be specified to pass a separate matrix of labels.
+
+If training is not desired, a pre-existing model may be loaded with the `input_model` parameter.
+
+
+
+The `incremental_variance` parameter can be used to force the training to use an incremental algorithm for calculating variance.  This is slower, but can help avoid loss of precision in some cases.
+
+If classifying a test set is desired, the test set may be specified with the `test` parameter, and the classifications may be saved with the `predictions`predictions  parameter.  If saving the trained model is desired, this may be done with the `output_model` output parameter.
+
+Note: the `output` and `output_probs` parameters are deprecated and will be removed in mlpack 4.0.0.  Use `predictions` and `probabilities` instead.
+
+### Example
+For example, to train a Naive Bayes classifier on the dataset `'data'` with labels `'labels'` and save the model to `'nbc_model'`, the following command may be used:
+
+```python
+>>> output = nbc(training=data, labels=labels)
+>>> nbc_model = output['output_model']
+```
+
+Then, to use `'nbc_model'` to predict the classes of the dataset `'test_set'` and save the predicted classes to `'predictions'`, the following command may be used:
+
+```python
+>>> output = nbc(input_model=nbc_model, test=test_set)
+>>> predictions = output['output']
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+ - [NaiveBayesClassifier C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp)
+
+## nca()
+{: #nca }
+
+#### Neighborhood Components Analysis (NCA)
+{: #nca_descr }
+
+```python
+>>> from mlpack import nca
+>>> d = nca(armijo_constant=0.0001, batch_size=50,
+        check_input_matrices=False, copy_all_inputs=False, input_=np.empty([0,
+        0]), labels=np.empty([0], dtype=np.uint64), linear_scan=False,
+        max_iterations=500000, max_line_search_trials=50, max_step=1e+20,
+        min_step=1e-20, normalize=False, num_basis=5, optimizer='sgd', seed=0,
+        step_size=0.01, tolerance=1e-07, verbose=False, wolfe=0.9)
+>>> output = d['output']
+```
+
+An implementation of neighborhood components analysis, a distance learning technique that can be used for preprocessing.  Given a labeled dataset, this uses NCA, which seeks to improve the k-nearest-neighbor classification, and returns the learned distance metric. [Detailed documentation](#nca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `armijo_constant` | [`float`](#doc_float) | Armijo constant for L-BFGS. | `0.0001` |
+| `batch_size` | [`int`](#doc_int) | Batch size for mini-batch SGD. | `50` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to run NCA on. | `**--**` |
+| `labels` | [`int vector`](#doc_int_vector) | Labels for input dataset. | `np.empty([0], dtype=np.uint64)` |
+| `linear_scan` | [`bool`](#doc_bool) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `False` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations for SGD or L-BFGS (0 indicates no limit). | `500000` |
+| `max_line_search_trials` | [`int`](#doc_int) | Maximum number of line search trials for L-BFGS. | `50` |
+| `max_step` | [`float`](#doc_float) | Maximum step of line search for L-BFGS. | `1e+20` |
+| `min_step` | [`float`](#doc_float) | Minimum step of line search for L-BFGS. | `1e-20` |
+| `normalize` | [`bool`](#doc_bool) | Use a normalized starting point for optimization. This is useful for when points are far apart, or when SGD is returning NaN. | `False` |
+| `num_basis` | [`int`](#doc_int) | Number of memory points to be stored for L-BFGS. | `5` |
+| `optimizer` | [`str`](#doc_str) | Optimizer to use; 'sgd' or 'lbfgs'. | `'sgd'` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`float`](#doc_float) | Step size for stochastic gradient descent (alpha). | `0.01` |
+| `tolerance` | [`float`](#doc_float) | Maximum tolerance for termination of SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `wolfe` | [`float`](#doc_float) | Wolfe condition parameter for L-BFGS. | `0.9` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Output matrix for learned distance matrix. | 
+
+### Detailed documentation
+{: #nca_detailed-documentation }
+
+This program implements Neighborhood Components Analysis, both a linear dimensionality reduction technique and a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset by scaling the dimensions.  The method is nonparametric, and does not require a value of k.  It works by using stochastic ("soft") neighbor assignments and using optimization techniques over the gradient of the accuracy of the neighbor assignments.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input_`), or alternatively as a separate matrix (specified with `labels`).
+
+This implementation of NCA uses stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer.  These optimizers do not guarantee global convergence for a nonconvex objective function (NCA's objective function is nonconvex), so the final results could depend on the random seed or other optimizer parameters.
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of iterations (specified with `max_iterations`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter, which is necessary if many warnings of the form 'Denominator of p_i is 0!' are given.  Tuning the step size can be a tedious affair.  In general, the step size is too large if the objective is not mostly uniformly decreasing, or if zero-valued denominator warnings are being issued.  The step size is too small if the objective is changing very slowly.  Setting the termination condition can be done easily once a good step size parameter is found; either increase the maximum iterations to a large number and allow SGD to find a minimum, or set the maximum iterations to 0 (allowing infinite iterations) and set the tolerance (specified by `tolerance`) to define the maximum allowed difference between objectives for SGD to terminate.  Be careful---setting the tolerance instead of the maximum iterations can take a very long time and may actually never converge due to the properties of the SGD optimizer. Note that a single iteration of SGD refers to a single point, so to take a single pass over the dataset, set the value of the `max_iterations` parameter equal to the number of points in the dataset.
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `num_basis` (specifies the number of memory points used by L-BFGS), `max_iterations`, `armijo_constant`, `wolfe`, `tolerance` (the optimization is terminated when the gradient norm is below this value), `max_line_search_trials`, `min_step`, and `max_step` (which both refer to the line search routine).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.
+
+By default, the SGD optimizer is used.
+
+### See also
+
+ - [lmnn()](#lmnn)
+ - [Neighbourhood components analysis on Wikipedia](https://en.wikipedia.org/wiki/Neighbourhood_components_analysis)
+ - [Neighbourhood components analysis (pdf)](http://papers.nips.cc/paper/2566-neighbourhood-components-analysis.pdf)
+ - [NCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/nca/nca.hpp)
+
+## knn()
+{: #knn }
+
+#### k-Nearest-Neighbors Search
+{: #knn_descr }
+
+```python
+>>> from mlpack import knn
+>>> d = knn(algorithm='dual_tree', check_input_matrices=False,
+        copy_all_inputs=False, epsilon=0, input_model=None, k=0, leaf_size=20,
+        query=np.empty([0, 0]), random_basis=False, reference=np.empty([0, 0]),
+        rho=0.7, seed=0, tau=0, tree_type='kd', true_distances=np.empty([0, 0]),
+        true_neighbors=np.empty([0, 0], dtype=np.uint64), verbose=False)
+>>> distances = d['distances']
+>>> neighbors = d['neighbors']
+>>> output_model = d['output_model']
+```
+
+An implementation of k-nearest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#knn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`str`](#doc_str) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `'dual_tree'` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `epsilon` | [`float`](#doc_float) | If specified, will do approximate nearest neighbor search with given relative error. | `0` |
+| `input_model` | [`KNNModelType`](#doc_model) | Pre-trained kNN model. | `None` |
+| `k` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, spill trees, and octrees). | `20` |
+| `query` | [`matrix`](#doc_matrix) | Matrix containing query points (optional). | `np.empty([0, 0])` |
+| `random_basis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `False` |
+| `reference` | [`matrix`](#doc_matrix) | Matrix containing the reference dataset. | `np.empty([0, 0])` |
+| `rho` | [`float`](#doc_float) | Balance threshold (only valid for spill trees). | `0.7` |
+| `seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tau` | [`float`](#doc_float) | Overlapping size (only valid for spill trees). | `0` |
+| `tree_type` | [`str`](#doc_str) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'spill', 'oct'. | `'kd'` |
+| `true_distances` | [`matrix`](#doc_matrix) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `np.empty([0, 0])` |
+| `true_neighbors` | [`int matrix`](#doc_int_matrix) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `np.empty([0, 0], dtype=np.uint64)` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`matrix`](#doc_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`int matrix`](#doc_int_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`KNNModelType`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #knn_detailed-documentation }
+
+This program will calculate the k-nearest-neighbors of a set of points using kd-trees or cover trees (cover tree support is experimental and may be slow). You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following command will calculate the 5 nearest neighbors of each point in `'input'` and store the distances in `'distances'` and the neighbors in `'neighbors'`: 
+
+```python
+>>> output = knn(k=5, reference=input)
+>>> neighbors = output['neighbors']
+>>> distances = output['distances']
+```
+
+The output is organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output matrix corresponds to the distance between those two points.
+
+### See also
+
+ - [lsh()](#lsh)
+ - [krann()](#krann)
+ - [kfn()](#kfn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## kfn()
+{: #kfn }
+
+#### k-Furthest-Neighbors Search
+{: #kfn_descr }
+
+```python
+>>> from mlpack import kfn
+>>> d = kfn(algorithm='dual_tree', check_input_matrices=False,
+        copy_all_inputs=False, epsilon=0, input_model=None, k=0, leaf_size=20,
+        percentage=1, query=np.empty([0, 0]), random_basis=False,
+        reference=np.empty([0, 0]), seed=0, tree_type='kd',
+        true_distances=np.empty([0, 0]), true_neighbors=np.empty([0, 0],
+        dtype=np.uint64), verbose=False)
+>>> distances = d['distances']
+>>> neighbors = d['neighbors']
+>>> output_model = d['output_model']
+```
+
+An implementation of k-furthest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k furthest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`str`](#doc_str) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `'dual_tree'` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `epsilon` | [`float`](#doc_float) | If specified, will do approximate furthest neighbor search with given relative error. Must be in the range [0,1). | `0` |
+| `input_model` | [`KFNModelType`](#doc_model) | Pre-trained kFN model. | `None` |
+| `k` | [`int`](#doc_int) | Number of furthest neighbors to find. | `0` |
+| `leaf_size` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `percentage` | [`float`](#doc_float) | If specified, will do approximate furthest neighbor search. Must be in the range (0,1] (decimal form). Resultant neighbors will be at least (p*100) % of the distance as the true furthest neighbor. | `1` |
+| `query` | [`matrix`](#doc_matrix) | Matrix containing query points (optional). | `np.empty([0, 0])` |
+| `random_basis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `False` |
+| `reference` | [`matrix`](#doc_matrix) | Matrix containing the reference dataset. | `np.empty([0, 0])` |
+| `seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tree_type` | [`str`](#doc_str) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `'kd'` |
+| `true_distances` | [`matrix`](#doc_matrix) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `np.empty([0, 0])` |
+| `true_neighbors` | [`int matrix`](#doc_int_matrix) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `np.empty([0, 0], dtype=np.uint64)` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`matrix`](#doc_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`int matrix`](#doc_int_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`KFNModelType`](#doc_model) | If specified, the kFN model will be output here. | 
+
+### Detailed documentation
+{: #kfn_detailed-documentation }
+
+This program will calculate the k-furthest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following will calculate the 5 furthest neighbors of eachpoint in `'input'` and store the distances in `'distances'` and the neighbors in `'neighbors'`: 
+
+```python
+>>> output = kfn(k=5, reference=input)
+>>> distances = output['distances']
+>>> neighbors = output['neighbors']
+```
+
+The output files are organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th furthest neighbor from the point in the query set with index i.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [approx_kfn()](#approx_kfn)
+ - [knn()](#knn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## nmf()
+{: #nmf }
+
+#### Non-negative Matrix Factorization
+{: #nmf_descr }
+
+```python
+>>> from mlpack import nmf
+>>> d = nmf(check_input_matrices=False, copy_all_inputs=False,
+        initial_h=np.empty([0, 0]), initial_w=np.empty([0, 0]),
+        input_=np.empty([0, 0]), max_iterations=10000, min_residue=1e-05,
+        rank=0, seed=0, update_rules='multdist', verbose=False)
+>>> h = d['h']
+>>> w = d['w']
+```
+
+An implementation of non-negative matrix factorization.  This can be used to decompose an input dataset into two low-rank non-negative components. [Detailed documentation](#nmf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `initial_h` | [`matrix`](#doc_matrix) | Initial H matrix. | `np.empty([0, 0])` |
+| `initial_w` | [`matrix`](#doc_matrix) | Initial W matrix. | `np.empty([0, 0])` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to perform NMF on. | `**--**` |
+| `max_iterations` | [`int`](#doc_int) | Number of iterations before NMF terminates (0 runs until convergence. | `10000` |
+| `min_residue` | [`float`](#doc_float) | The minimum root mean square residue allowed for each iteration, below which the program terminates. | `1e-05` |
+| `rank` | [`int`](#doc_int) | Rank of the factorization. | `**--**` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `update_rules` | [`str`](#doc_str) | Update rules for each iteration; ( multdist \| multdiv \| als ). | `'multdist'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `h` | [`matrix`](#doc_matrix) | Matrix to save the calculated H to. | 
+| `w` | [`matrix`](#doc_matrix) | Matrix to save the calculated W to. | 
+
+### Detailed documentation
+{: #nmf_detailed-documentation }
+
+This program performs non-negative matrix factorization on the given dataset, storing the resulting decomposed matrices in the specified files.  For an input dataset V, NMF decomposes V into two matrices W and H such that 
+
+V = W * H
+
+where all elements in W and H are non-negative.  If V is of size (n x m), then W will be of size (n x r) and H will be of size (r x m), where r is the rank of the factorization (specified by the `rank` parameter).
+
+Optionally, the desired update rules for each NMF iteration can be chosen from the following list:
+
+ - multdist: multiplicative distance-based update rules (Lee and Seung 1999)
+ - multdiv: multiplicative divergence-based update rules (Lee and Seung 1999)
+ - als: alternating least squares update rules (Paatero and Tapper 1994)
+
+The maximum number of iterations is specified with `max_iterations`, and the minimum residue required for algorithm termination is specified with the `min_residue` parameter.
+
+### Example
+For example, to run NMF on the input matrix `'V'` using the 'multdist' update rules with a rank-10 decomposition and storing the decomposed matrices into `'W'` and `'H'`, the following command could be used: 
+
+```python
+>>> output = nmf(input_=V, rank=10, update_rules='multdist')
+>>> W = output['w']
+>>> H = output['h']
+```
+
+### See also
+
+ - [cf()](#cf)
+ - [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ - [Algorithms for non-negative matrix factorization (pdf)](http://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf)
+ - [AMF C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/amf/amf.hpp)
+
+## pca()
+{: #pca }
+
+#### Principal Components Analysis
+{: #pca_descr }
+
+```python
+>>> from mlpack import pca
+>>> d = pca(check_input_matrices=False, copy_all_inputs=False,
+        decomposition_method='exact', input_=np.empty([0, 0]),
+        new_dimensionality=0, scale=False, var_to_retain=0, verbose=False)
+>>> output = d['output']
+```
+
+An implementation of several strategies for principal components analysis (PCA), a common preprocessing step.  Given a dataset and a desired new dimensionality, this can reduce the dimensionality of the data using the linear transformation determined by PCA. [Detailed documentation](#pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `decomposition_method` | [`str`](#doc_str) | Method used for the principal components analysis: 'exact', 'randomized', 'randomized-block-krylov', 'quic'. | `'exact'` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset to perform PCA on. | `**--**` |
+| `new_dimensionality` | [`int`](#doc_int) | Desired dimensionality of output dataset. If 0, no dimensionality reduction is performed. | `0` |
+| `scale` | [`bool`](#doc_bool) | If set, the data will be scaled before running PCA, such that the variance of each feature is 1. | `False` |
+| `var_to_retain` | [`float`](#doc_float) | Amount of variance to retain; should be between 0 and 1.  If 1, all variance is retained.  Overrides -d. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #pca_detailed-documentation }
+
+This program performs principal components analysis on the given dataset using the exact, randomized, randomized block Krylov, or QUIC SVD method. It will transform the data onto its principal components, optionally performing dimensionality reduction by ignoring the principal components with the smallest eigenvalues.
+
+Use the `input_` parameter to specify the dataset to perform PCA on.  A desired new dimensionality can be specified with the `new_dimensionality` parameter, or the desired variance to retain can be specified with the `var_to_retain` parameter.  If desired, the dataset can be scaled before running PCA with the `scale` parameter.
+
+Multiple different decomposition techniques can be used.  The method to use can be specified with the `decomposition_method` parameter, and it may take the values 'exact', 'randomized', or 'quic'.
+
+### Example
+For example, to reduce the dimensionality of the matrix `'data'` to 5 dimensions using randomized SVD for the decomposition, storing the output matrix to `'data_mod'`, the following command can be used:
+
+```python
+>>> output = pca(input_=data, new_dimensionality=5,
+  decomposition_method='randomized')
+>>> data_mod = output['output']
+```
+
+### See also
+
+ - [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+ - [PCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/pca/pca.hpp)
+
+## perceptron()
+{: #perceptron }
+
+#### Perceptron
+{: #perceptron_descr }
+
+```python
+>>> from mlpack import perceptron
+>>> d = perceptron(check_input_matrices=False, copy_all_inputs=False,
+        input_model=None, labels=np.empty([0], dtype=np.uint64),
+        max_iterations=1000, test=np.empty([0, 0]), training=np.empty([0, 0]),
+        verbose=False)
+>>> output = d['output']
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+```
+
+An implementation of a perceptron---a single level neural network--=for classification.  Given labeled data, a perceptron can be trained and saved for future use; or, a pre-trained perceptron can be used for classification on new points. [Detailed documentation](#perceptron_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`PerceptronModelType`](#doc_model) | Input perceptron model. | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | A matrix containing labels for the training set. | `np.empty([0], dtype=np.uint64)` |
+| `max_iterations` | [`int`](#doc_int) | The maximum number of iterations the perceptron is to be run | `1000` |
+| `test` | [`matrix`](#doc_matrix) | A matrix containing the test set. | `np.empty([0, 0])` |
+| `training` | [`matrix`](#doc_matrix) | A matrix containing the training set. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`int vector`](#doc_int_vector) | The matrix in which the predicted labels for the test set will be written. | 
+| `output_model` | [`PerceptronModelType`](#doc_model) | Output for trained perceptron model. | 
+| `predictions` | [`int vector`](#doc_int_vector) | The matrix in which the predicted labels for the test set will be written. | 
+
+### Detailed documentation
+{: #perceptron_detailed-documentation }
+
+This program implements a perceptron, which is a single level neural network. The perceptron makes its predictions based on a linear predictor function combining a set of weights with the feature vector.  The perceptron learning rule is able to converge, given enough iterations (specified using the `max_iterations` parameter), if the data supplied is linearly separable.  The perceptron is parameterized by a matrix of weight vectors that denote the numerical weights of the neural network.
+
+This program allows loading a perceptron from a model (via the `input_model` parameter) or training a perceptron given training data (via the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (via the `test` parameter) and the classification results on the test set may be saved with the `predictions` output parameter.  The perceptron model may be saved with the `output_model` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `output`.
+Use `predictions` instead of `output`.
+
+### Example
+The training data given with the `training` option may have class labels as its last dimension (so, if the training data is in CSV format, labels should be the last column).  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+All these options make it easy to train a perceptron, and then re-use that perceptron for later classification.  The invocation below trains a perceptron on `'training_data'` with labels `'training_labels'`, and saves the model to `'perceptron_model'`.
+
+```python
+>>> output = perceptron(training=training_data, labels=training_labels)
+>>> perceptron_model = output['output_model']
+```
+
+Then, this model can be re-used for classification on the test data `'test_data'`.  The example below does precisely that, saving the predicted classes to `'predictions'`.
+
+```python
+>>> output = perceptron(input_model=perceptron_model, test=test_data)
+>>> predictions = output['predictions']
+```
+
+Note that all of the options may be specified at once: predictions may be calculated right after training a model, and model training can occur even if an existing perceptron model is passed with the `input_model` parameter.  However, note that the number of classes and the dimensionality of all data must match.  So you cannot pass a perceptron model trained on 2 classes and then re-train with a 4-class dataset.  Similarly, attempting classification on a 3-dimensional dataset with a perceptron that has been trained on 8 dimensions will cause an error.
+
+### See also
+
+ - [adaboost()](#adaboost)
+ - [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+ - [Perceptron C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/perceptron/perceptron.hpp)
+
+## preprocess_split()
+{: #preprocess_split }
+
+#### Split Data
+{: #preprocess_split_descr }
+
+```python
+>>> from mlpack import preprocess_split
+>>> d = preprocess_split(check_input_matrices=False,
+        copy_all_inputs=False, input_=np.empty([0, 0]),
+        input_labels=np.empty([0, 0], dtype=np.uint64), no_shuffle=False,
+        seed=0, stratify_data=False, test_ratio=0.2, verbose=False)
+>>> test = d['test']
+>>> test_labels = d['test_labels']
+>>> training = d['training']
+>>> training_labels = d['training_labels']
+```
+
+A utility to split data into a training and testing dataset.  This can also split labels according to the same split. [Detailed documentation](#preprocess_split_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix containing data. | `**--**` |
+| `input_labels` | [`int matrix`](#doc_int_matrix) | Matrix containing labels. | `np.empty([0, 0], dtype=np.uint64)` |
+| `no_shuffle` | [`bool`](#doc_bool) | Avoid shuffling the data before splitting. | `False` |
+| `seed` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `stratify_data` | [`bool`](#doc_bool) | Stratify the data according to labels | `False` |
+| `test_ratio` | [`float`](#doc_float) | Ratio of test set; if not set,the ratio defaults to 0.2 | `0.2` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `test` | [`matrix`](#doc_matrix) | Matrix to save test data to. | 
+| `test_labels` | [`int matrix`](#doc_int_matrix) | Matrix to save test labels to. | 
+| `training` | [`matrix`](#doc_matrix) | Matrix to save training data to. | 
+| `training_labels` | [`int matrix`](#doc_int_matrix) | Matrix to save train labels to. | 
+
+### Detailed documentation
+{: #preprocess_split_detailed-documentation }
+
+This utility takes a dataset and optionally labels and splits them into a training set and a test set. Before the split, the points in the dataset are randomly reordered. The percentage of the dataset to be used as the test set can be specified with the `test_ratio` parameter; the default is 0.2 (20%).
+
+The output training and test matrices may be saved with the `training` and `test` output parameters.
+
+Optionally, labels can also be split along with the data by specifying the `input_labels` parameter.  Splitting labels works the same way as splitting the data. The output training and test labels may be saved with the `training_labels` and `test_labels` output parameters, respectively.
+
+### Example
+So, a simple example where we want to split the dataset `'X'` into `'X_train'` and `'X_test'` with 60% of the data in the training set and 40% of the dataset in the test set, we could run 
+
+```python
+>>> output = preprocess_split(input_=X, test_ratio=0.4)
+>>> X_train = output['training']
+>>> X_test = output['test']
+```
+
+Also by default the dataset is shuffled and split; you can provide the `no_shuffle` option to avoid shuffling the data; an example to avoid shuffling of data is:
+
+```python
+>>> output = preprocess_split(input_=X, test_ratio=0.4, no_shuffle=True)
+>>> X_train = output['training']
+>>> X_test = output['test']
+```
+
+If we had a dataset `'X'` and associated labels `'y'`, and we wanted to split these into `'X_train'`, `'y_train'`, `'X_test'`, and `'y_test'`, with 30% of the data in the test set, we could run
+
+```python
+>>> output = preprocess_split(input_=X, input_labels=y, test_ratio=0.3)
+>>> X_train = output['training']
+>>> y_train = output['training_labels']
+>>> X_test = output['test']
+>>> y_test = output['test_labels']
+```
+
+To maintain the ratio of each class in the train and test sets, the`stratify_data` option can be used.
+
+```python
+>>> output = preprocess_split(input_=X, test_ratio=0.4, stratify_data=True)
+>>> X_train = output['training']
+>>> X_test = output['test']
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_binarize()
+{: #preprocess_binarize }
+
+#### Binarize Data
+{: #preprocess_binarize_descr }
+
+```python
+>>> from mlpack import preprocess_binarize
+>>> d = preprocess_binarize(check_input_matrices=False,
+        copy_all_inputs=False, dimension=0, input_=np.empty([0, 0]),
+        threshold=0, verbose=False)
+>>> output = d['output']
+```
+
+A utility to binarize a dataset.  Given a dataset, this utility converts each value in the desired dimension(s) to 0 or 1; this can be a useful preprocessing step. [Detailed documentation](#preprocess_binarize_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `dimension` | [`int`](#doc_int) | Dimension to apply the binarization. If not set, the program will binarize every dimension by default. | `0` |
+| `input_` | [`matrix`](#doc_matrix) | Input data matrix. | `**--**` |
+| `threshold` | [`float`](#doc_float) | Threshold to be applied for binarization. If not set, the threshold defaults to 0.0. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix in which to save the output. | 
+
+### Detailed documentation
+{: #preprocess_binarize_detailed-documentation }
+
+This utility takes a dataset and binarizes the variables into either 0 or 1 given threshold. User can apply binarization on a dimension or the whole dataset.  The dimension to apply binarization to can be specified using the `dimension` parameter; if left unspecified, every dimension will be binarized.  The threshold for binarization can also be specified with the `threshold` parameter; the default threshold is 0.0.
+
+The binarized matrix may be saved with the `output` output parameter.
+
+### Example
+For example, if we want to set all variables greater than 5 in the dataset `'X'` to 1 and variables less than or equal to 5.0 to 0, and save the result to `'Y'`, we could run
+
+```python
+>>> output = preprocess_binarize(input_=X, threshold=5)
+>>> Y = output['output']
+```
+
+But if we want to apply this to only the first (0th) dimension of `'X'`,  we could instead run
+
+```python
+>>> output = preprocess_binarize(input_=X, threshold=5, dimension=0)
+>>> Y = output['output']
+```
+
+### See also
+
+ - [preprocess_describe()](#preprocess_describe)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_describe()
+{: #preprocess_describe }
+
+#### Descriptive Statistics
+{: #preprocess_describe_descr }
+
+```python
+>>> from mlpack import preprocess_describe
+>>> preprocess_describe(check_input_matrices=False,
+        copy_all_inputs=False, dimension=0, input_=np.empty([0, 0]),
+        population=False, precision=4, row_major=False, verbose=False, width=8)
+```
+
+A utility for printing descriptive statistics about a dataset.  This prints a number of details about a dataset in a tabular format. [Detailed documentation](#preprocess_describe_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `dimension` | [`int`](#doc_int) | Dimension of the data. Use this to specify a dimension | `0` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix containing data, | `**--**` |
+| `population` | [`bool`](#doc_bool) | If specified, the program will calculate statistics assuming the dataset is the population. By default, the program will assume the dataset as a sample. | `False` |
+| `precision` | [`int`](#doc_int) | Precision of the output statistics. | `4` |
+| `row_major` | [`bool`](#doc_bool) | If specified, the program will calculate statistics across rows, not across columns.  (Remember that in mlpack, a column represents a point, so this option is generally not necessary.) | `False` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `width` | [`int`](#doc_int) | Width of the output table. | `8` |
+
+
+### Detailed documentation
+{: #preprocess_describe_detailed-documentation }
+
+This utility takes a dataset and prints out the descriptive statistics of the data. Descriptive statistics is the discipline of quantitatively describing the main features of a collection of information, or the quantitative description itself. The program does not modify the original file, but instead prints out the statistics to the console. The printed result will look like a table.
+
+Optionally, width and precision of the output can be adjusted by a user using the `width` and `precision` parameters. A user can also select a specific dimension to analyze if there are too many dimensions. The `population` parameter can be specified when the dataset should be considered as a population.  Otherwise, the dataset will be considered as a sample.
+
+### Example
+So, a simple example where we want to print out statistical facts about the dataset `'X'` using the default settings, we could run 
+
+```python
+>>> preprocess_describe(input_=X, verbose=True)
+```
+
+If we want to customize the width to 10 and precision to 5 and consider the dataset as a population, we could run
+
+```python
+>>> preprocess_describe(input_=X, width=10, precision=5, verbose=True)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_scale()
+{: #preprocess_scale }
+
+#### Scale Data
+{: #preprocess_scale_descr }
+
+```python
+>>> from mlpack import preprocess_scale
+>>> d = preprocess_scale(check_input_matrices=False,
+        copy_all_inputs=False, epsilon=1e-06, input_=np.empty([0, 0]),
+        input_model=None, inverse_scaling=False, max_value=1, min_value=0,
+        scaler_method='standard_scaler', seed=0, verbose=False)
+>>> output = d['output']
+>>> output_model = d['output_model']
+```
+
+A utility to perform feature scaling on datasets using one of sixtechniques.  Both scaling and inverse scaling are supported, andscalers can be saved and then applied to other datasets. [Detailed documentation](#preprocess_scale_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `epsilon` | [`float`](#doc_float) | regularization Parameter for pcawhitening, or zcawhitening, should be between -1 to 1. | `1e-06` |
+| `input_` | [`matrix`](#doc_matrix) | Matrix containing data. | `**--**` |
+| `input_model` | [`ScalingModelType`](#doc_model) | Input Scaling model. | `None` |
+| `inverse_scaling` | [`bool`](#doc_bool) | Inverse Scaling to get original dataset | `False` |
+| `max_value` | [`int`](#doc_int) | Ending value of range for min_max_scaler. | `1` |
+| `min_value` | [`int`](#doc_int) | Starting value of range for min_max_scaler. | `0` |
+| `scaler_method` | [`str`](#doc_str) | method to use for scaling, the default is standard_scaler. | `'standard_scaler'` |
+| `seed` | [`int`](#doc_int) | Random seed (0 for std::time(NULL)). | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save scaled data to. | 
+| `output_model` | [`ScalingModelType`](#doc_model) | Output scaling model. | 
+
+### Detailed documentation
+{: #preprocess_scale_detailed-documentation }
+
+This utility takes a dataset and performs feature scaling using one of the six scaler methods namely: 'max_abs_scaler', 'mean_normalization', 'min_max_scaler' ,'standard_scaler', 'pca_whitening' and 'zca_whitening'. The function takes a matrix as `input_` and a scaling method type which you can specify using `scaler_method` parameter; the default is standard scaler, and outputs a matrix with scaled feature.
+
+The output scaled feature matrix may be saved with the `output` output parameters.
+
+The model to scale features can be saved using `output_model` and later can be loaded back using`input_model`.
+
+### Example
+So, a simple example where we want to scale the dataset `'X'` into `'X_scaled'` with  standard_scaler as scaler_method, we could run 
+
+```python
+>>> output = preprocess_scale(input_=X, scaler_method='standard_scaler')
+>>> X_scaled = output['output']
+```
+
+A simple example where we want to whiten the dataset `'X'` into `'X_whitened'` with  PCA as whitening_method and use 0.01 as regularization parameter, we could run 
+
+```python
+>>> output = preprocess_scale(input_=X, scaler_method='pca_whitening',
+  epsilon=0.01)
+>>> X_scaled = output['output']
+```
+
+You can also retransform the scaled dataset back using`inverse_scaling`. An example to rescale : `'X_scaled'` into `'X'`using the saved model `input_model` is:
+
+```python
+>>> output = preprocess_scale(input_=X_scaled, inverse_scaling=True,
+  input_model=saved)
+>>> X = output['output']
+```
+
+Another simple example where we want to scale the dataset `'X'` into `'X_scaled'` with  min_max_scaler as scaler method, where scaling range is 1 to 3 instead of default 0 to 1. We could run 
+
+```python
+>>> output = preprocess_scale(input_=X, scaler_method='min_max_scaler',
+  min_value=1, max_value=3)
+>>> X_scaled = output['output']
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_one_hot_encoding()
+{: #preprocess_one_hot_encoding }
+
+#### One Hot Encoding
+{: #preprocess_one_hot_encoding_descr }
+
+```python
+>>> from mlpack import preprocess_one_hot_encoding
+>>> d = preprocess_one_hot_encoding(check_input_matrices=False,
+        copy_all_inputs=False, dimensions=[], input_=np.empty([0, 0]),
+        verbose=False)
+>>> output = d['output']
+```
+
+A utility to do one-hot encoding on features of dataset. [Detailed documentation](#preprocess_one_hot_encoding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `dimensions` | [`list of ints`](#doc_list_of_ints) | Index of dimensions that need to be one-hot encoded (if unspecified, all categorical dimensions are one-hot encoded). | `[]` |
+| `input_` | [`categorical matrix`](#doc_categorical_matrix) | Matrix containing data. | `**--**` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save one-hot encoded features data to. | 
+
+### Detailed documentation
+{: #preprocess_one_hot_encoding_detailed-documentation }
+
+This utility takes a dataset and a vector of indices and does one-hot encoding of the respective features at those indices. Indices represent the IDs of the dimensions to be one-hot encoded.
+
+If no dimensions are specified with `dimensions`, then all categorical-type dimensions will be one-hot encoded. Otherwise, only the dimensions given in `dimensions` will be one-hot encoded.
+
+The output matrix with encoded features may be saved with the `output` parameters.
+
+### Example
+So, a simple example where we want to encode 1st and 3rd feature from dataset `'X'` into `'X_output'` would be
+
+```python
+>>> output = preprocess_one_hot_encoding(input_=X, dimensions=1,
+  dimensions=3)
+>>> X_ouput = output['output']
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+ - [One-hot encoding on Wikipedia](https://en.m.wikipedia.org/wiki/One-hot)
+
+## radical()
+{: #radical }
+
+#### RADICAL
+{: #radical_descr }
+
+```python
+>>> from mlpack import radical
+>>> d = radical(angles=150, check_input_matrices=False,
+        copy_all_inputs=False, input_=np.empty([0, 0]), noise_std_dev=0.175,
+        objective=False, replicates=30, seed=0, sweeps=0, verbose=False)
+>>> output_ic = d['output_ic']
+>>> output_unmixing = d['output_unmixing']
+```
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Given a dataset, this can decompose the dataset into an unmixing matrix and an independent component matrix; this can be useful for preprocessing. [Detailed documentation](#radical_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `angles` | [`int`](#doc_int) | Number of angles to consider in brute-force search during Radical2D. | `150` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_` | [`matrix`](#doc_matrix) | Input dataset for ICA. | `**--**` |
+| `noise_std_dev` | [`float`](#doc_float) | Standard deviation of Gaussian noise. | `0.175` |
+| `objective` | [`bool`](#doc_bool) | If set, an estimate of the final objective function is printed. | `False` |
+| `replicates` | [`int`](#doc_int) | Number of Gaussian-perturbed replicates to use (per point) in Radical2D. | `30` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `sweeps` | [`int`](#doc_int) | Number of sweeps; each sweep calls Radical2D once for each pair of dimensions. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_ic` | [`matrix`](#doc_matrix) | Matrix to save independent components to. | 
+| `output_unmixing` | [`matrix`](#doc_matrix) | Matrix to save unmixing matrix to. | 
+
+### Detailed documentation
+{: #radical_detailed-documentation }
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Assuming that we have an input matrix X, the goal is to find a square unmixing matrix W such that Y = W * X and the dimensions of Y are independent components.  If the algorithm is running particularly slowly, try reducing the number of replicates.
+
+The input matrix to perform ICA on should be specified with the `input_` parameter.  The output matrix Y may be saved with the `output_ic` output parameter, and the output unmixing matrix W may be saved with the `output_unmixing` output parameter.
+
+### Example
+For example, to perform ICA on the matrix `'X'` with 40 replicates, saving the independent components to `'ic'`, the following command may be used: 
+
+```python
+>>> output = radical(input_=X, replicates=40)
+>>> ic = output['output_ic']
+```
+
+### See also
+
+ - [Independent component analysis on Wikipedia](https://en.wikipedia.org/wiki/Independent_component_analysis)
+ - [ICA using spacings estimates of entropy (pdf)](https://www.jmlr.org/papers/volume4/learned-miller03a/learned-miller03a.pdf)
+ - [Radical C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/radical/radical.hpp)
+
+## random_forest()
+{: #random_forest }
+
+#### Random forests
+{: #random_forest_descr }
+
+```python
+>>> from mlpack import random_forest
+>>> d = random_forest(check_input_matrices=False, copy_all_inputs=False,
+        input_model=None, labels=np.empty([0], dtype=np.uint64),
+        maximum_depth=0, minimum_gain_split=0, minimum_leaf_size=1,
+        num_trees=10, print_training_accuracy=False, seed=0, subspace_dim=0,
+        test=np.empty([0, 0]), test_labels=np.empty([0], dtype=np.uint64),
+        training=np.empty([0, 0]), verbose=False, warm_start=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of the standard random forest algorithm by Leo Breiman for classification.  Given labeled data, a random forest can be trained and saved for future use; or, a pre-trained random forest can be used for classification. [Detailed documentation](#random_forest_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`RandomForestModelType`](#doc_model) | Pre-trained random forest to use for classification. | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | Labels for training dataset. | `np.empty([0], dtype=np.uint64)` |
+| `maximum_depth` | [`int`](#doc_int) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`float`](#doc_float) | Minimum gain needed to make a split when building a tree. | `0` |
+| `minimum_leaf_size` | [`int`](#doc_int) | Minimum number of points in each leaf node. | `1` |
+| `num_trees` | [`int`](#doc_int) | Number of trees in the random forest. | `10` |
+| `print_training_accuracy` | [`bool`](#doc_bool) | If set, then the accuracy of the model on the training set will be predicted (verbose must also be specified). | `False` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `subspace_dim` | [`int`](#doc_int) | Dimensionality of random subspace to use for each split.  '0' will autoselect the square root of data dimensionality. | `0` |
+| `test` | [`matrix`](#doc_matrix) | Test dataset to produce predictions for. | `np.empty([0, 0])` |
+| `test_labels` | [`int vector`](#doc_int_vector) | Test dataset labels, if accuracy calculation is desired. | `np.empty([0], dtype=np.uint64)` |
+| `training` | [`matrix`](#doc_matrix) | Training dataset. | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `warm_start` | [`bool`](#doc_bool) | If true and passed along with `training` and `input_model` then trains more trees on top of existing model. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`RandomForestModelType`](#doc_model) | Model to save trained random forest to. | 
+| `predictions` | [`int vector`](#doc_int_vector) | Predicted classes for each point in the test set. | 
+| `probabilities` | [`matrix`](#doc_matrix) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #random_forest_detailed-documentation }
+
+This program is an implementation of the standard random forest classification algorithm by Leo Breiman.  A random forest can be trained and saved for later use, or a random forest may be loaded and predictions or class probabilities for points may be generated.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model`parameter. The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `num_trees` controls the number of trees in the random forest.  The `minimum_gain_split` parameter controls the minimum required gain for a decision tree node to split.  Larger values will force higher-confidence splits.  The `maximum_depth` parameter specifies the maximum depth of the tree.  The `subspace_dim` parameter is used to control the number of random dimensions chosen for an individual node's split.  If `print_training_accuracy` is specified, the calculated accuracy on the training set will be printed.
+
+Test data may be specified with the `test` parameter, and if performance measures are desired for that test set, labels for the test points may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions`output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a random forest with a minimum leaf size of 20 using 10 trees on the dataset contained in `'data'`with labels `'labels'`, saving the output random forest to `'rf_model'` and printing the training error, one could call
+
+```python
+>>> output = random_forest(training=data, labels=labels, minimum_leaf_size=20,
+  num_trees=10, print_training_accuracy=True)
+>>> rf_model = output['output_model']
+```
+
+Then, to use that model to classify points in `'test_set'` and print the test error given the labels `'test_labels'` using that model, while saving the predictions for each point to `'predictions'`, one could call 
+
+```python
+>>> output = random_forest(input_model=rf_model, test=test_set,
+  test_labels=test_labels)
+>>> predictions = output['predictions']
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [hoeffding_tree()](#hoeffding_tree)
+ - [softmax_regression()](#softmax_regression)
+ - [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ - [Random forests (pdf)](https://www.eecis.udel.edu/~shatkay/Course/papers/BreimanRandomForests2001.pdf)
+ - [RandomForest C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/random_forest/random_forest.hpp)
+
+## krann()
+{: #krann }
+
+#### K-Rank-Approximate-Nearest-Neighbors (kRANN)
+{: #krann_descr }
+
+```python
+>>> from mlpack import krann
+>>> d = krann(alpha=0.95, check_input_matrices=False,
+        copy_all_inputs=False, first_leaf_exact=False, input_model=None, k=0,
+        leaf_size=20, naive=False, query=np.empty([0, 0]), random_basis=False,
+        reference=np.empty([0, 0]), sample_at_leaves=False, seed=0,
+        single_mode=False, single_sample_limit=20, tau=5, tree_type='kd',
+        verbose=False)
+>>> distances = d['distances']
+>>> neighbors = d['neighbors']
+>>> output_model = d['output_model']
+```
+
+An implementation of rank-approximate k-nearest-neighbor search (kRANN)  using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#krann_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `alpha` | [`float`](#doc_float) | The desired success probability. | `0.95` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `first_leaf_exact` | [`bool`](#doc_bool) | The flag to trigger sampling only after exactly exploring the first leaf. | `False` |
+| `input_model` | [`RAModelType`](#doc_model) | Pre-trained kNN model. | `None` |
+| `k` | [`int`](#doc_int) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`int`](#doc_int) | Leaf size for tree building (used for kd-trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `naive` | [`bool`](#doc_bool) | If true, sampling will be done without using a tree. | `False` |
+| `query` | [`matrix`](#doc_matrix) | Matrix containing query points (optional). | `np.empty([0, 0])` |
+| `random_basis` | [`bool`](#doc_bool) | Before tree-building, project the data onto a random orthogonal basis. | `False` |
+| `reference` | [`matrix`](#doc_matrix) | Matrix containing the reference dataset. | `np.empty([0, 0])` |
+| `sample_at_leaves` | [`bool`](#doc_bool) | The flag to trigger sampling at leaves. | `False` |
+| `seed` | [`int`](#doc_int) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `single_mode` | [`bool`](#doc_bool) | If true, single-tree search is used (as opposed to dual-tree search. | `False` |
+| `single_sample_limit` | [`int`](#doc_int) | The limit on the maximum number of samples (and hence the largest node you can approximate). | `20` |
+| `tau` | [`float`](#doc_float) | The allowed rank-error in terms of the percentile of the data. | `5` |
+| `tree_type` | [`str`](#doc_str) | Type of tree to use: 'kd', 'ub', 'cover', 'r', 'x', 'r-star', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `'kd'` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`matrix`](#doc_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`int matrix`](#doc_int_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`RAModelType`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #krann_detailed-documentation }
+
+This program will calculate the k rank-approximate-nearest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. You must specify the rank approximation (in %) (and optionally the success probability).
+
+### Example
+For example, the following will return 5 neighbors from the top 0.1% of the data (with probability 0.95) for each point in `'input'` and store the distances in `'distances'` and the neighbors in `'neighbors.csv'`:
+
+```python
+>>> output = krann(reference=input, k=5, tau=0.1)
+>>> distances = output['distances']
+>>> neighbors = output['neighbors']
+```
+
+Note that tau must be set such that the number of points in the corresponding percentile of the data is greater than k.  Thus, if we choose tau = 0.1 with a dataset of 1000 points and k = 5, then we are attempting to choose 5 nearest neighbors out of the closest 1 point -- this is invalid and the program will terminate with an error message.
+
+The output matrices are organized such that row i and column j in the neighbors output file corresponds to the index of the point in the reference set which is the i'th nearest neighbor from the point in the query set with index j.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [knn()](#knn)
+ - [lsh()](#lsh)
+ - [Rank-approximate nearest neighbor search: Retaining meaning and speed in high dimensions (pdf)](https://papers.nips.cc/paper/3864-rank-approximate-nearest-neighbor-search-retaining-meaning-and-speed-in-high-dimensions.pdf)
+ - [RASearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/rann/ra_search.hpp)
+
+## softmax_regression()
+{: #softmax_regression }
+
+#### Softmax Regression
+{: #softmax_regression_descr }
+
+```python
+>>> from mlpack import softmax_regression
+>>> d = softmax_regression(check_input_matrices=False,
+        copy_all_inputs=False, input_model=None, labels=np.empty([0],
+        dtype=np.uint64), lambda_=0.0001, max_iterations=400,
+        no_intercept=False, number_of_classes=0, test=np.empty([0, 0]),
+        test_labels=np.empty([0], dtype=np.uint64), training=np.empty([0, 0]),
+        verbose=False)
+>>> output_model = d['output_model']
+>>> predictions = d['predictions']
+>>> probabilities = d['probabilities']
+```
+
+An implementation of softmax regression for classification, which is a multiclass generalization of logistic regression.  Given labeled data, a softmax regression model can be trained and saved for future use, or, a pre-trained softmax regression model can be used for classification of new points. [Detailed documentation](#softmax_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `input_model` | [`SoftmaxRegression<>Type`](#doc_model) | File containing existing model (parameters). | `None` |
+| `labels` | [`int vector`](#doc_int_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). The labels must order as a row. | `np.empty([0], dtype=np.uint64)` |
+| `lambda_` | [`float`](#doc_float) | L2-regularization constant | `0.0001` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations before termination. | `400` |
+| `no_intercept` | [`bool`](#doc_bool) | Do not add the intercept term to the model. | `False` |
+| `number_of_classes` | [`int`](#doc_int) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `test` | [`matrix`](#doc_matrix) | Matrix containing test dataset. | `np.empty([0, 0])` |
+| `test_labels` | [`int vector`](#doc_int_vector) | Matrix containing test labels. | `np.empty([0], dtype=np.uint64)` |
+| `training` | [`matrix`](#doc_matrix) | A matrix containing the training set (the matrix of predictors, X). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`SoftmaxRegression<>Type`](#doc_model) | File to save trained softmax regression model to. | 
+| `predictions` | [`int vector`](#doc_int_vector) | Matrix to save predictions for test dataset into. | 
+| `probabilities` | [`matrix`](#doc_matrix) | Matrix to save class probabilities for test dataset into. | 
+
+### Detailed documentation
+{: #softmax_regression_detailed-documentation }
+
+This program performs softmax regression, a generalization of logistic regression to the multiclass case, and has support for L2 regularization.  The program is able to train a model, load  an existing model, and give predictions (and optionally their accuracy) for test data.
+
+Training a softmax regression model is done by giving a file of training points with the `training` parameter and their corresponding labels with the `labels` parameter. The number of classes can be manually specified with the `number_of_classes` parameter, and the maximum number of iterations of the L-BFGS optimizer can be specified with the `max_iterations` parameter.  The L2 regularization constant can be specified with the `lambda_` parameter and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.
+
+The trained model can be saved with the `output_model` output parameter. If training is not desired, but only testing is, a model can be loaded with the `input_model` parameter.  At the current time, a loaded model cannot be trained further, so specifying both `input_model` and `training` is not allowed.
+
+The program is also able to evaluate a model on test data.  A test dataset can be specified with the `test` parameter. Class predictions can be saved with the `predictions` output parameter.  If labels are specified for the test data with the `test_labels` parameter, then the program will print the accuracy of the predictions on the given test set and its corresponding labels.
+
+### Example
+For example, to train a softmax regression model on the data `'dataset'` with labels `'labels'` with a maximum of 1000 iterations for training, saving the trained model to `'sr_model'`, the following command can be used: 
+
+```python
+>>> output = softmax_regression(training=dataset, labels=labels)
+>>> sr_model = output['output_model']
+```
+
+Then, to use `'sr_model'` to classify the test points in `'test_points'`, saving the output predictions to `'predictions'`, the following command can be used:
+
+```python
+>>> output = softmax_regression(input_model=sr_model, test=test_points)
+>>> predictions = output['predictions']
+```
+
+### See also
+
+ - [logistic_regression()](#logistic_regression)
+ - [random_forest()](#random_forest)
+ - [Multinomial logistic regression (softmax regression) on Wikipedia](https://en.wikipedia.org/wiki/Multinomial_logistic_regression)
+ - [SoftmaxRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/softmax_regression/softmax_regression.hpp)
+
+## sparse_coding()
+{: #sparse_coding }
+
+#### Sparse Coding
+{: #sparse_coding_descr }
+
+```python
+>>> from mlpack import sparse_coding
+>>> d = sparse_coding(atoms=15, check_input_matrices=False,
+        copy_all_inputs=False, initial_dictionary=np.empty([0, 0]),
+        input_model=None, lambda1=0, lambda2=0, max_iterations=0,
+        newton_tolerance=1e-06, normalize=False, objective_tolerance=0.01,
+        seed=0, test=np.empty([0, 0]), training=np.empty([0, 0]),
+        verbose=False)
+>>> codes = d['codes']
+>>> dictionary = d['dictionary']
+>>> output_model = d['output_model']
+```
+
+An implementation of Sparse Coding with Dictionary Learning.  Given a dataset, this will decompose the dataset into a sparse combination of a few dictionary elements, where the dictionary is learned during computation; a dictionary can be reused for future sparse coding of new points. [Detailed documentation](#sparse_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`int`](#doc_int) | Number of atoms in the dictionary. | `15` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `initial_dictionary` | [`matrix`](#doc_matrix) | Optional initial dictionary matrix. | `np.empty([0, 0])` |
+| `input_model` | [`SparseCodingType`](#doc_model) | File containing input sparse coding model. | `None` |
+| `lambda1` | [`float`](#doc_float) | Sparse coding l1-norm regularization parameter. | `0` |
+| `lambda2` | [`float`](#doc_float) | Sparse coding l2-norm regularization parameter. | `0` |
+| `max_iterations` | [`int`](#doc_int) | Maximum number of iterations for sparse coding (0 indicates no limit). | `0` |
+| `newton_tolerance` | [`float`](#doc_float) | Tolerance for convergence of Newton method. | `1e-06` |
+| `normalize` | [`bool`](#doc_bool) | If set, the input data matrix will be normalized before coding. | `False` |
+| `objective_tolerance` | [`float`](#doc_float) | Tolerance for convergence of the objective function. | `0.01` |
+| `seed` | [`int`](#doc_int) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`matrix`](#doc_matrix) | Optional matrix to be encoded by trained model. | `np.empty([0, 0])` |
+| `training` | [`matrix`](#doc_matrix) | Matrix of training data (X). | `np.empty([0, 0])` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`matrix`](#doc_matrix) | Matrix to save the output sparse codes of the test matrix (--test_file) to. | 
+| `dictionary` | [`matrix`](#doc_matrix) | Matrix to save the output dictionary to. | 
+| `output_model` | [`SparseCodingType`](#doc_model) | File to save trained sparse coding model to. | 
+
+### Detailed documentation
+{: #sparse_coding_detailed-documentation }
+
+An implementation of Sparse Coding with Dictionary Learning, which achieves sparsity via an l1-norm regularizer on the codes (LASSO) or an (l1+l2)-norm regularizer on the codes (the Elastic Net).  Given a dense data matrix X with d dimensions and n points, sparse coding seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a sparse coding matrix Z with n points in k dimensions.
+
+The original data matrix X can then be reconstructed as Z * D.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The sparse coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a sparse coding step, which updates the sparse coding matrix.
+
+Once a dictionary D is found, the sparse coding model may be used to encode other matrices, and saved for future usage.
+
+To run this program, either an input matrix or an already-saved sparse coding model must be specified.  An input matrix may be specified with the `training` option, along with the number of atoms in the dictionary (specified with the `atoms` parameter).  It is also possible to specify an initial dictionary for the optimization, with the `initial_dictionary` parameter.  An input model may be specified with the `input_model` parameter.
+
+### Example
+As an example, to build a sparse coding model on the dataset `'data'` using 200 atoms and an l1-regularization parameter of 0.1, saving the model into `'model'`, use 
+
+```python
+>>> output = sparse_coding(training=data, atoms=200, lambda1=0.1)
+>>> model = output['output_model']
+```
+
+Then, this model could be used to encode a new matrix, `'otherdata'`, and save the output codes to `'codes'`: 
+
+```python
+>>> output = sparse_coding(input_model=model, test=otherdata)
+>>> codes = output['codes']
+```
+
+### See also
+
+ - [local_coordinate_coding()](#local_coordinate_coding)
+ - [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ - [Efficient sparse coding algorithms (pdf)](http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf)
+ - [Regularization and variable selection via the elastic net](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=46217f372a75dddc2254fdbc6b9418ba3554e453)
+ - [SparseCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/sparse_coding/sparse_coding.hpp)
+
+## class Adaboost
+{: #adaboost }
+
+#### AdaBoost
+{: #adaboost_descr }
+
+
+This program implements the AdaBoost (or Adaptive Boosting) algorithm. The variant of AdaBoost implemented here is AdaBoost.MH. It uses a weak learner, either decision stumps or perceptrons, and over many iterations, creates a strong learner that is a weighted ensemble of weak learners. It runs these iterations until a tolerance value is crossed for change in the value of the weighted training error.
+
+For more information about the algorithm, see the paper "Improved Boosting Algorithms Using Confidence-Rated Predictions", by R.E. Schapire and Y. Singer.
+### Parameters
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `iterations` | [`int`](#doc_int) | The maximum number of boosting iterations to be run (0 will run until convergence.) | `1000` |
+| `tolerance` | [`float`](#doc_float) | The tolerance for change in values of the weighted error during training. | `1e-10` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `weak_learner` | [`str`](#doc_str) | The type of weak learner to use: 'decision_stump', or 'perceptron'. | `'decision_stump'` |
+
+### Example
+
+```python
+>>> import pandas as pd
+>>> from mlpack import preprocess_split
+>>> from mlpack import Adaboost
+>>> X = pd.read_csv('https://example.com')
+>>> y = pd.read_csv('https://example.com')
+>>> d = preprocess_split(input_=X, input_labels=y, test_ratio=0.2)
+>>> X_train = d['training']
+>>> y_train = d['training_labels']
+>>> X_test = d['test']
+>>> y_test = d['test_labels']
+>>> model = Adaboost(check_input_matrices=False, copy_all_inputs=False,
+  iterations=1000, tolerance=1e-10, verbose=False,
+  weak_learner='decision_stump')
+>>> output_model = model.fit(training=X_train, labels=y_train)
+>>> predictions = model.predict(test=X_test)
+>>> probabilities = model.predict_proba(test=X_test)
+```
+
+### Methods
+
+| **name** | **description** |
+|----------|-----------------|
+| fit | Training AdaBoost model. |
+| predict | Class predictions from model. |
+| predict_proba | Class probabilities from model. |
+
+### 1. fit
+
+Training AdaBoost model.
+
+#### Input Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `labels` | [`int vector`](#doc_int_vector) | Labels for the training set. | 
+| `training` | [`matrix`](#doc_matrix) | Dataset for training AdaBoost. | 
+
+#### Returns: 
+
+| **type** | **description** |
+|----------|-----------------|
+| [`AdaBoostModelType`](#doc_model) | Output trained AdaBoost model. | 
+
+### 2. predict
+
+Class predictions from model.
+
+#### Input Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `test` | [`matrix`](#doc_matrix) | Test dataset. | 
+
+#### Returns: 
+
+| **type** | **description** |
+|----------|-----------------|
+| [`int vector`](#doc_int_vector) | Predicted labels for the test set. | 
+
+### 3. predict_proba
+
+Class probabilities from model.
+
+#### Input Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `test` | [`matrix`](#doc_matrix) | Test dataset. | 
+
+#### Returns: 
+
+| **type** | **description** |
+|----------|-----------------|
+| [`matrix`](#doc_matrix) | Predicted class probabilities for each point in the test set. | 
+
+## class LinearRegression
+{: #linear_regression }
+
+#### Simple Linear Regression
+{: #linear_regression_descr }
+
+
+An implementation of simple linear regression and simple ridge regression using ordinary least squares. This solves the problem
+
+  y = X * b + e
+### Parameters
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `lambda_` | [`float`](#doc_float) | Tikhonov regularization for ridge regression.  If 0, the method reduces to linear regression. | `0` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+
+### Example
+
+```python
+>>> import pandas as pd
+>>> from mlpack import preprocess_split
+>>> from mlpack import LinearRegression
+>>> X = pd.read_csv('https://example.com')
+>>> y = pd.read_csv('https://example.com')
+>>> d = preprocess_split(input_=X, input_labels=y, test_ratio=0.2)
+>>> X_train = d['training']
+>>> y_train = d['training_labels']
+>>> X_test = d['test']
+>>> y_test = d['test_labels']
+>>> lr = LinearRegression(check_input_matrices=False, copy_all_inputs=False,
+  lambda_=0, verbose=False)
+>>> output_model = lr.fit(training=X_train, training_responses=y_train)
+>>> output_predictions = model.predict(test=X_test)
+```
+
+### Methods
+
+| **name** | **description** |
+|----------|-----------------|
+| fit | Train a linear regression model. |
+| predict | Predictions from model. |
+
+### 1. fit
+
+Train a linear regression model.
+
+#### Input Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `training` | [`matrix`](#doc_matrix) | Matrix containing training set X (regressors). | 
+| `training_responses` | [`vector`](#doc_vector) | Optional vector containing y (responses). If not given, the responses are assumed to be the last row of the input file. | 
+
+#### Returns: 
+
+| **type** | **description** |
+|----------|-----------------|
+| [`LinearRegression<>Type`](#doc_model) | Output LinearRegression model. | 
+
+### 2. predict
+
+Predictions from model.
+
+#### Input Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `test` | [`matrix`](#doc_matrix) | Matrix containing X' (test regressors). | 
+
+#### Returns: 
+
+| **type** | **description** |
+|----------|-----------------|
+| [`vector`](#doc_vector) | If --test_file is specified, this matrix is where the predicted responses will be saved. | 
+
+## image_converter()
+{: #image_converter }
+
+#### Image Converter
+{: #image_converter_descr }
+
+```python
+>>> from mlpack import image_converter
+>>> d = image_converter(channels=0, check_input_matrices=False,
+        copy_all_inputs=False, dataset=np.empty([0, 0]), height=0, input_=[],
+        quality=90, save=False, verbose=False, width=0)
+>>> output = d['output']
+```
+
+A utility to load an image or set of images into a single dataset that can then be used by other mlpack methods and utilities. This can also unpack an image dataset into individual files, for instance after mlpack methods have been used. [Detailed documentation](#image_converter_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `channels` | [`int`](#doc_int) | Number of channels in the image. | `0` |
+| `check_input_matrices` | [`bool`](#doc_bool) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `False` |
+| `copy_all_inputs` | [`bool`](#doc_bool) | If specified, all input parameters will be deep copied before the method is run.  This is useful for debugging problems where the input parameters are being modified by the algorithm, but can slow down the code.  <span class="special">Only exists in Python binding.</span> | `False` |
+| `dataset` | [`matrix`](#doc_matrix) | Input matrix to save as images. | `np.empty([0, 0])` |
+| `height` | [`int`](#doc_int) | Height of the images. | `0` |
+| `input_` | [`list of strs`](#doc_list_of_strs) | Image filenames which have to be loaded/saved. | `**--**` |
+| `quality` | [`int`](#doc_int) | Compression of the image if saved as jpg (0-100). | `90` |
+| `save` | [`bool`](#doc_bool) | Save a dataset as images. | `False` |
+| `verbose` | [`bool`](#doc_bool) | Display informational messages and the full list of parameters and timers at the end of execution. | `False` |
+| `width` | [`int`](#doc_int) | Width of the image. | `0` |
+
+### Output options
+
+Results are returned in a Python dictionary.  The keys of the dictionary are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`matrix`](#doc_matrix) | Matrix to save images data to, Onlyneeded if you are specifying 'save' option. | 
+
+### Detailed documentation
+{: #image_converter_detailed-documentation }
+
+This utility takes an image or an array of images and loads them to a matrix. You can optionally specify the height `height` width `width` and channel `channels` of the images that needs to be loaded; otherwise, these parameters will be automatically detected from the image.
+There are other options too, that can be specified such as `quality`.
+
+You can also provide a dataset and save them as images using `dataset` and `save` as an parameter.
+
+### Example
+ An example to load an image : 
+
+```python
+>>> output = image_converter(input_=X, height=256, width=256, channels=3)
+>>> Y = output['output']
+```
+
+ An example to save an image is :
+
+```python
+>>> image_converter(input_=X, height=256, width=256, channels=3, dataset=Y,
+  save=True)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
diff -pruN 4.3.0-2/doc/user/bindings/python.sidebar.html 4.4.0-1/doc/user/bindings/python.sidebar.html
--- 4.3.0-2/doc/user/bindings/python.sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/python.sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,100 @@
+<ul>
+<li class="page_title">
+<b>mlpack Python binding documentation</b>
+<a href="#">[top]</a>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#mlpack-overview">Overview</a></li>
+<li><a href="LINKROOTquickstart/python.html">Python Quickstart</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#data-formats">Data Formats</a></li>
+<li><details><summary>Classification</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#decision_tree">decision_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#linear_svm">linear_svm()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#logistic_regression">logistic_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#nbc">nbc()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#perceptron">perceptron()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#random_forest">random_forest()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#softmax_regression">softmax_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#adaboost">class Adaboost</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Regression</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#bayesian_linear_regression">bayesian_linear_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#lars">lars()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#linear_regression">class LinearRegression</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Clustering</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#dbscan">dbscan()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#gmm_train">gmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#gmm_generate">gmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#gmm_probability">gmm_probability()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#hoeffding_tree">hoeffding_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#kmeans">kmeans()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#mean_shift">mean_shift()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Geometry</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#approx_kfn">approx_kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#emst">emst()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#fastmks">fastmks()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#lsh">lsh()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#knn">knn()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#kfn">kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#krann">krann()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Preprocessing</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#preprocess_split">preprocess_split()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#preprocess_binarize">preprocess_binarize()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#preprocess_describe">preprocess_describe()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#preprocess_scale">preprocess_scale()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#preprocess_one_hot_encoding">preprocess_one_hot_encoding()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#image_converter">image_converter()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Misc. / Other</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#cf">cf()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#det">det()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#hmm_train">hmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#hmm_generate">hmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#hmm_loglik">hmm_loglik()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#hmm_viterbi">hmm_viterbi()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#kde">kde()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#nmf">nmf()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Transformations</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/python.html#kernel_pca">kernel_pca()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#lmnn">lmnn()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#local_coordinate_coding">local_coordinate_coding()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#nca">nca()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#pca">pca()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#radical">radical()</a></li>
+<li><a href="LINKROOTuser/bindings/python.html#sparse_coding">sparse_coding()</a></li>
+</ul>
+</details>
+</li>
+
+</ul>
+</li>
+</ul>
+</div>
diff -pruN 4.3.0-2/doc/user/bindings/r.md 4.4.0-1/doc/user/bindings/r.md
--- 4.3.0-2/doc/user/bindings/r.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/r.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,3610 @@
+# mlpack R binding documentation
+
+## mlpack overview
+
+mlpack is an intuitive, fast, and flexible header-only C++ machine learning library with bindings to other languages.  It aims to provide fast, lightweight implementations of both common and cutting-edge machine learning algorithms.
+
+This reference page details mlpack's bindings to R.
+
+Further useful mlpack documentation links are given below.
+
+ - [mlpack homepage](https://www.mlpack.org/)
+ - [mlpack on Github](https://github.com/mlpack/mlpack)
+ - [mlpack main documentation page](https://www.mlpack.org/docs.html)
+
+See also the quickstart guide for R:
+
+ - [R Quickstart](../../quickstart/r.md)
+
+## Data Formats
+
+<div id="data-formats-div" markdown="1">
+mlpack bindings for R take and return a restricted set of types, for simplicity.  These include primitive types, matrix/vector types, categorical matrix types, and model types. Each type is detailed below.
+
+ - `integer`{: #doc_integer}: An integer (i.e., `1`).
+ - `numeric`{: #doc_numeric }: A floating-point number (i.e., `0.5`).
+ - `logical`{: #doc_logical }: A boolean flag option (i.e. `TRUE` or `FALSE`).
+ - `character`{: #doc_character }: A character string (i.e., `"hello"`).
+ - `vector of integers`{: #doc_vector_of_integers }: A vector of integers; i.e., `c(0, 1, 2)`.
+ - `vector of characters`{: #doc_vector_of_characters }: A vector of strings; i.e., `c("hello", "goodbye")`.
+ - `numeric matrix`{: #doc_numeric_matrix }: A 2-d matrix-like containing `numeric` data (could be an `matrix` or a `data.frame` or anything convertible to an 2-d `matrix`).
+ - `integer matrix`{: #doc_integer_matrix }: A 2-d matrix-like containing `integer` data (could be an `matrix` or a `data.frame` or anything convertible to an 2-d `matrix`).
+ - `numeric vector`{: #doc_numeric_vector }: A 1-d matrix-like containing `numeric` data (could be an `matrix` or a `data.frame` with one dimension of size 1).
+ - `integer vector`{: #doc_integer_vector }: A 1-d matrix-like containing `integer` data (could be an `matrix` or a `data.frame` with one dimension of size 1).
+ - `categorical matrix/data.frame`{: #doc_categorical_matrix_data_frame }: A 2-d array containing `numeric` data.  Like the regular 2-d matrices, this can be a `matrix`, or a `data.frame`. However, this type can also accept a `data.frame` that has columns of type `character`, `logical` or `factor`.  These values will be converted to `numeric` indices before being passed to mlpack, and then inside mlpack they will be properly treated as categorical variables, so there is no need to do one-hot encoding for this matrix type.
+ - `<Model> (mlpack model)`{: #doc_model }: An mlpack model pointer.  `<Model>` refers to the type of model that is being stored, so, e.g., for `cf()`, the type will be `CFModel`. This type holds a pointer to C++ memory containing the mlpack model.  Note that this means the mlpack model itself cannot be easily inspected in R.  However, the pointer can be passed to subsequent calls to mlpack functions, and can be serialized and deserialized via either the `Serialize()` and `Unserialize()` functions.
+</div>
+
+
+## approx_kfn()
+{: #approx_kfn }
+
+#### Approximate furthest neighbor search
+{: #approx_kfn_descr }
+
+```R
+R> library(mlpack)
+R> d <- approx_kfn(algorithm="ds", calculate_error=FALSE,
+        exact_distances=matrix(numeric(), 0, 0), input_model=NA, k=0,
+        num_projections=5, num_tables=5, query=matrix(numeric(), 0, 0),
+        reference=matrix(numeric(), 0, 0), verbose=FALSE)
+R> distances <- d$distances
+R> neighbors <- d$neighbors
+R> output_model <- d$output_model
+```
+
+An implementation of two strategies for furthest neighbor search.  This can be used to compute the furthest neighbor of query point(s) from a set of points; furthest neighbor models can be saved and reused with future query point(s). [Detailed documentation](#approx_kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`character`](#doc_character) | Algorithm to use: 'ds' or 'qdafn'. | `"ds"` |
+| `calculate_error` | [`logical`](#doc_logical) | If set, calculate the average distance error for the first furthest neighbor only. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `exact_distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing exact distances to furthest neighbors; this can be used to avoid explicit calculation when --calculate_error is set. | `matrix(numeric(), 0, 0)` |
+| `input_model` | [`ApproxKFNModel`](#doc_model) | File containing input model. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of furthest neighbors to search for. | `0` |
+| `num_projections` | [`integer`](#doc_integer) | Number of projections to use in each hash table. | `5` |
+| `num_tables` | [`integer`](#doc_integer) | Number of hash tables to use. | `5` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing query points. | `matrix(numeric(), 0, 0)` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing the reference dataset. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save furthest neighbor distances to. | 
+| `neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix to save neighbor indices to. | 
+| `output_model` | [`ApproxKFNModel`](#doc_model) | File to save output model to. | 
+
+### Detailed documentation
+{: #approx_kfn_detailed-documentation }
+
+This program implements two strategies for furthest neighbor search. These strategies are:
+
+ - The 'qdafn' algorithm from "Approximate Furthest Neighbor in High Dimensions" by R. Pagh, F. Silvestri, J. Sivertsen, and M. Skala, in Similarity Search and Applications 2015 (SISAP).
+ - The 'DrusillaSelect' algorithm from "Fast approximate furthest neighbors with data-dependent candidate selection", by R.R. Curtin and A.B. Gardner, in Similarity Search and Applications 2016 (SISAP).
+
+These two strategies give approximate results for the furthest neighbor search problem and can be used as fast replacements for other furthest neighbor techniques such as those found in the mlpack_kfn program.  Note that typically, the 'ds' algorithm requires far fewer tables and projections than the 'qdafn' algorithm.
+
+Specify a reference set (set to search in) with `reference`, specify a query set with `query`, and specify algorithm parameters with `num_tables` and `num_projections` (or don't and defaults will be used).  The algorithm to be used (either 'ds'---the default---or 'qdafn')  may be specified with `algorithm`.  Also specify the number of neighbors to search for with `k`.
+
+Note that for 'qdafn' in lower dimensions, `num_projections` may need to be set to a high value in order to return results for each query point.
+
+If no query set is specified, the reference set will be used as the query set.  The `output_model` output parameter may be used to store the built model, and an input model may be loaded instead of specifying a reference set with the `input_model` option.
+
+Results for each query point can be stored with the `neighbors` and `distances` output parameters.  Each row of these output matrices holds the k distances or neighbor indices for each query point.
+
+### Example
+For example, to find the 5 approximate furthest neighbors with `"reference_set"` as the reference set and `"query_set"` as the query set using DrusillaSelect, storing the furthest neighbor indices to `"neighbors"` and the furthest neighbor distances to `"distances"`, one could call
+
+```R
+R> output <- approx_kfn(query=query_set, reference=reference_set, k=5,
+  algorithm="ds")
+R> neighbors <- output$neighbors
+R> distances <- output$distances
+```
+
+and to perform approximate all-furthest-neighbors search with k=1 on the set `"data"` storing only the furthest neighbor distances to `"distances"`, one could call
+
+```R
+R> output <- approx_kfn(reference=reference_set, k=1)
+R> distances <- output$distances
+```
+
+A trained model can be re-used.  If a model has been previously saved to `"model"`, then we may find 3 approximate furthest neighbors on a query set `"new_query_set"` using that model and store the furthest neighbor indices into `"neighbors"` by calling
+
+```R
+R> output <- approx_kfn(input_model=model, query=new_query_set, k=3)
+R> neighbors <- output$neighbors
+```
+
+### See also
+
+ - [k-furthest-neighbor search](#kfn)
+ - [k-nearest-neighbor search](#knn)
+ - [Fast approximate furthest neighbors with data-dependent candidate selection (pdf)](http://ratml.org/pub/pdf/2016fast.pdf)
+ - [Approximate furthest neighbor in high dimensions (pdf)](https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf)
+ - [QDAFN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/qdafn.hpp)
+ - [DrusillaSelect class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/approx_kfn/drusilla_select.hpp)
+
+## bayesian_linear_regression()
+{: #bayesian_linear_regression }
+
+#### BayesianLinearRegression
+{: #bayesian_linear_regression_descr }
+
+```R
+R> library(mlpack)
+R> d <- bayesian_linear_regression(center=FALSE, input=matrix(numeric(),
+        0, 0), input_model=NA, responses=matrix(numeric(), 0, 0), scale=FALSE,
+        test=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> stds <- d$stds
+```
+
+An implementation of the bayesian linear regression. [Detailed documentation](#bayesian_linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `center` | [`logical`](#doc_logical) | Center the data and fit the intercept if enabled. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of covariates (X). | `matrix(numeric(), 0, 0)` |
+| `input_model` | [`BayesianLinearRegression`](#doc_model) | Trained BayesianLinearRegression model to use. | `NA` |
+| `responses` | [`numeric vector`](#doc_numeric_vector) | Matrix of responses/observations (y). | `matrix(numeric(), 0, 0)` |
+| `scale` | [`logical`](#doc_logical) | Scale each feature by their standard deviations if enabled. | `FALSE` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing points to regress on (test points). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`BayesianLinearRegression`](#doc_model) | Output BayesianLinearRegression model. | 
+| `predictions` | [`numeric matrix`](#doc_numeric_matrix) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+| `stds` | [`numeric matrix`](#doc_numeric_matrix) | If specified, this is where the standard deviations of the predictive distribution will be saved. | 
+
+### Detailed documentation
+{: #bayesian_linear_regression_detailed-documentation }
+
+An implementation of the bayesian linear regression.
+This model is a probabilistic view and implementation of the linear regression. The final solution is obtained by computing a posterior distribution from gaussian likelihood and a zero mean gaussian isotropic  prior distribution on the solution. 
+Optimization is AUTOMATIC and does not require cross validation. The optimization is performed by maximization of the evidence function. Parameters are tuned during the maximization of the marginal likelihood. This procedure includes the Ockham's razor that penalizes over complex solutions. 
+
+This program is able to train a Bayesian linear regression model or load a model from file, output regression predictions for a test set, and save the trained model to a file.
+
+To train a BayesianLinearRegression model, the `input` and `responses`parameters must be given. The `center`and `scale` parameters control the centering and the normalizing options. A trained model can be saved with the `output_model`. If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `predictions` output parameter. The corresponding standard deviation can be save by precising the `stds` parameter.
+
+### Example
+For example, the following command trains a model on the data `"data"` and responses `"responses"`with center set to true and scale set to false (so, Bayesian linear regression is being solved, and then the model is saved to `"blr_model"`:
+
+```R
+R> output <- bayesian_linear_regression(input=data, responses=responses,
+  center=1, scale=0)
+R> blr_model <- output$output_model
+```
+
+The following command uses the `"blr_model"` to provide predicted  responses for the data `"test"` and save those  responses to `"test_predictions"`: 
+
+```R
+R> output <- bayesian_linear_regression(input_model=blr_model, test=test)
+R> test_predictions <- output$predictions
+```
+
+Because the estimator computes a predictive distribution instead of a simple point estimate, the `stds` parameter allows one to save the prediction uncertainties: 
+
+```R
+R> output <- bayesian_linear_regression(input_model=blr_model, test=test)
+R> test_predictions <- output$predictions
+R> stds <- output$stds
+```
+
+### See also
+
+ - [Bayesian Interpolation](https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf)
+ - [Bayesian Linear Regression, Section 3.3](https://www.microsoft.com/en-us/research/uploads/prod/2006/01/Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf)
+ - [BayesianLinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp)
+
+## cf()
+{: #cf }
+
+#### Collaborative Filtering
+{: #cf_descr }
+
+```R
+R> library(mlpack)
+R> d <- cf(algorithm="NMF", all_user_recommendations=FALSE,
+        input_model=NA, interpolation="average",
+        iteration_only_termination=FALSE, max_iterations=1000,
+        min_residue=1e-05, neighbor_search="euclidean", neighborhood=5,
+        normalization="none", query=matrix(integer(), 0, 0), rank=0,
+        recommendations=5, seed=0, test=matrix(numeric(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output <- d$output
+R> output_model <- d$output_model
+```
+
+An implementation of several collaborative filtering (CF) techniques for recommender systems.  This can be used to train a new CF model, or use an existing CF model to compute recommendations. [Detailed documentation](#cf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`character`](#doc_character) | Algorithm used for matrix factorization. | `"NMF"` |
+| `all_user_recommendations` | [`logical`](#doc_logical) | Generate recommendations for all users. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`CFModel`](#doc_model) | Trained CF model to load. | `NA` |
+| `interpolation` | [`character`](#doc_character) | Algorithm used for weight interpolation. | `"average"` |
+| `iteration_only_termination` | [`logical`](#doc_logical) | Terminate only when the maximum number of iterations is reached. | `FALSE` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations. If set to zero, there is no limit on the number of iterations. | `1000` |
+| `min_residue` | [`numeric`](#doc_numeric) | Residue required to terminate the factorization (lower values generally mean better fits). | `1e-05` |
+| `neighbor_search` | [`character`](#doc_character) | Algorithm used for neighbor search. | `"euclidean"` |
+| `neighborhood` | [`integer`](#doc_integer) | Size of the neighborhood of similar users to consider for each query user. | `5` |
+| `normalization` | [`character`](#doc_character) | Normalization performed on the ratings. | `"none"` |
+| `query` | [`integer matrix`](#doc_integer_matrix) | List of query users for which recommendations should be generated. | `matrix(integer(), 0, 0)` |
+| `rank` | [`integer`](#doc_integer) | Rank of decomposed matrices (if 0, a heuristic is used to estimate the rank). | `0` |
+| `recommendations` | [`integer`](#doc_integer) | Number of recommendations to generate for each query user. | `5` |
+| `seed` | [`integer`](#doc_integer) | Set the random seed (0 uses std::time(NULL)). | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Test set to calculate RMSE on. | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform CF on. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`integer matrix`](#doc_integer_matrix) | Matrix that will store output recommendations. | 
+| `output_model` | [`CFModel`](#doc_model) | Output for trained CF model. | 
+
+### Detailed documentation
+{: #cf_detailed-documentation }
+
+This program performs collaborative filtering (CF) on the given dataset. Given a list of user, item and preferences (the `training` parameter), the program will perform a matrix decomposition and then can perform a series of actions related to collaborative filtering.  Alternately, the program can load an existing saved CF model with the `input_model` parameter and then use that model to provide recommendations or predict values.
+
+The input matrix should be a 3-dimensional matrix of ratings, where the first dimension is the user, the second dimension is the item, and the third dimension is that user's rating of that item.  Both the users and items should be numeric indices, not names. The indices are assumed to start from 0.
+
+A set of query users for which recommendations can be generated may be specified with the `query` parameter; alternately, recommendations may be generated for every user in the dataset by specifying the `all_user_recommendations` parameter.  In addition, the number of recommendations per user to generate can be specified with the `recommendations` parameter, and the number of similar users (the size of the neighborhood) to be considered when generating recommendations can be specified with the `neighborhood` parameter.
+
+For performing the matrix decomposition, the following optimization algorithms can be specified via the `algorithm` parameter:
+
+ - 'RegSVD' -- Regularized SVD using a SGD optimizer
+ - 'NMF' -- Non-negative matrix factorization with alternating least squares update rules
+ - 'BatchSVD' -- SVD batch learning
+ - 'SVDIncompleteIncremental' -- SVD incomplete incremental learning
+ - 'SVDCompleteIncremental' -- SVD complete incremental learning
+ - 'BiasSVD' -- Bias SVD using a SGD optimizer
+ - 'SVDPP' -- SVD++ using a SGD optimizer
+ - 'RandSVD' -- RandomizedSVD learning
+ - 'QSVD' -- QuicSVD learning
+ - 'BKSVD' -- Block Krylov SVD learning
+
+
+The following neighbor search algorithms can be specified via the `neighbor_search` parameter:
+
+ - 'cosine'  -- Cosine Search Algorithm
+ - 'euclidean'  -- Euclidean Search Algorithm
+ - 'pearson'  -- Pearson Search Algorithm
+
+
+The following weight interpolation algorithms can be specified via the `interpolation` parameter:
+
+ - 'average'  -- Average Interpolation Algorithm
+ - 'regression'  -- Regression Interpolation Algorithm
+ - 'similarity'  -- Similarity Interpolation Algorithm
+
+
+The following ranking normalization algorithms can be specified via the `normalization` parameter:
+
+ - 'none'  -- No Normalization
+ - 'item_mean'  -- Item Mean Normalization
+ - 'overall_mean'  -- Overall Mean Normalization
+ - 'user_mean'  -- User Mean Normalization
+ - 'z_score'  -- Z-Score Normalization
+
+A trained model may be saved to with the `output_model` output parameter.
+
+### Example
+To train a CF model on a dataset `"training_set"` using NMF for decomposition and saving the trained model to `"model"`, one could call: 
+
+```R
+R> output <- cf(training=training_set, algorithm="NMF")
+R> model <- output$output_model
+```
+
+Then, to use this model to generate recommendations for the list of users in the query set `"users"`, storing 5 recommendations in `"recommendations"`, one could call 
+
+```R
+R> output <- cf(input_model=model, query=users, recommendations=5)
+R> recommendations <- output$output
+```
+
+### See also
+
+ - [Collaborative Filtering on Wikipedia](https://en.wikipedia.org/wiki/Collaborative_filtering)
+ - [Matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Matrix_factorization_(recommender_systems))
+ - [Matrix factorization techniques for recommender systems (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5)
+ - [CFType class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/cf/cf.hpp)
+
+## dbscan()
+{: #dbscan }
+
+#### DBSCAN clustering
+{: #dbscan_descr }
+
+```R
+R> library(mlpack)
+R> d <- dbscan(epsilon=1, input=matrix(numeric(), 0, 0), min_size=5,
+        naive=FALSE, selection_type="ordered", single_mode=FALSE,
+        tree_type="kd", verbose=FALSE)
+R> assignments <- d$assignments
+R> centroids <- d$centroids
+```
+
+An implementation of DBSCAN clustering.  Given a dataset, this can compute and return a clustering of that dataset. [Detailed documentation](#dbscan_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `epsilon` | [`numeric`](#doc_numeric) | Radius of each range search. | `1` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to cluster. | `**--**` |
+| `min_size` | [`integer`](#doc_integer) | Minimum number of points for a cluster. | `5` |
+| `naive` | [`logical`](#doc_logical) | If set, brute-force range search (not tree-based) will be used. | `FALSE` |
+| `selection_type` | [`character`](#doc_character) | If using point selection policy, the type of selection to use ('ordered', 'random'). | `"ordered"` |
+| `single_mode` | [`logical`](#doc_logical) | If set, single-tree range search (not dual-tree) will be used. | `FALSE` |
+| `tree_type` | [`character`](#doc_character) | If using single-tree or dual-tree search, the type of tree to use ('kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'). | `"kd"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `assignments` | [`integer vector`](#doc_integer_vector) | Output matrix for assignments of each point. | 
+| `centroids` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save output centroids to. | 
+
+### Detailed documentation
+{: #dbscan_detailed-documentation }
+
+This program implements the DBSCAN algorithm for clustering using accelerated tree-based range search.  The type of tree that is used may be parameterized, or brute-force range search may also be used.
+
+The input dataset to be clustered may be specified with the `input` parameter; the radius of each range search may be specified with the `epsilon` parameters, and the minimum number of points in a cluster may be specified with the `min_size` parameter.
+
+The `assignments` and `centroids` output parameters may be used to save the output of the clustering. `assignments` contains the cluster assignments of each point, and `centroids` contains the centroids of each cluster.
+
+The range search may be controlled with the `tree_type`, `single_mode`, and `naive` parameters.  `tree_type` can control the type of tree used for range search; this can take a variety of values: 'kd', 'r', 'r-star', 'x', 'hilbert-r', 'r-plus', 'r-plus-plus', 'cover', 'ball'. The `single_mode` parameter will force single-tree search (as opposed to the default dual-tree search), and '`naive` will force brute-force range search.
+
+### Example
+An example usage to run DBSCAN on the dataset in `"input"` with a radius of 0.5 and a minimum cluster size of 5 is given below:
+
+```R
+R> dbscan(input=input, epsilon=0.5, min_size=5)
+```
+
+### See also
+
+ - [DBSCAN on Wikipedia](https://en.wikipedia.org/wiki/DBSCAN)
+ - [A density-based algorithm for discovering clusters in large spatial databases with noise (pdf)](https://cdn.aaai.org/KDD/1996/KDD96-037.pdf)
+ - [DBSCAN class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/dbscan/dbscan.hpp)
+
+## decision_tree()
+{: #decision_tree }
+
+#### Decision tree
+{: #decision_tree_descr }
+
+```R
+R> library(mlpack)
+R> d <- decision_tree(input_model=NA, labels=matrix(integer(), 0, 0),
+        maximum_depth=0, minimum_gain_split=1e-07, minimum_leaf_size=20,
+        print_training_accuracy=FALSE, print_training_error=FALSE,
+        test=matrix(numeric(), 0, 0), test_labels=matrix(integer(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE,
+        weights=matrix(numeric(), 0, 0))
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of an ID3-style decision tree for classification, which supports categorical data.  Given labeled data with numeric or categorical features, a decision tree can be trained and saved; or, an existing decision tree can be used for classification on new points. [Detailed documentation](#decision_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`DecisionTreeModel`](#doc_model) | Pre-trained decision tree, to be used with test points. | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Training labels. | `matrix(integer(), 0, 0)` |
+| `maximum_depth` | [`integer`](#doc_integer) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`numeric`](#doc_numeric) | Minimum gain for node splitting. | `1e-07` |
+| `minimum_leaf_size` | [`integer`](#doc_integer) | Minimum number of points in a leaf. | `20` |
+| `print_training_accuracy` | [`logical`](#doc_logical) | Print the training accuracy. | `FALSE` |
+| `print_training_error` | [`logical`](#doc_logical) | Print the training error (deprecated; will be removed in mlpack 4.0.0). | `FALSE` |
+| `test` | [`categorical matrix/data.frame`](#doc_categorical_matrix_data_frame) | Testing dataset (may be categorical). | `matrix(numeric(), 0, 0)` |
+| `test_labels` | [`integer vector`](#doc_integer_vector) | Test point labels, if accuracy calculation is desired. | `matrix(integer(), 0, 0)` |
+| `training` | [`categorical matrix/data.frame`](#doc_categorical_matrix_data_frame) | Training dataset (may be categorical). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `weights` | [`numeric matrix`](#doc_numeric_matrix) | The weight of labels | `matrix(numeric(), 0, 0)` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DecisionTreeModel`](#doc_model) | Output for trained decision tree. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | Class predictions for each test point. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | Class probabilities for each test point. | 
+
+### Detailed documentation
+{: #decision_tree_detailed-documentation }
+
+Train and evaluate using a decision tree.  Given a dataset containing numeric or categorical features, and associated labels for each point in the dataset, this program can train a decision tree on that data.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model` parameter.  The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `minimum_gain_split` parameter specifies the minimum gain that is needed for the node to split.  The `maximum_depth` parameter specifies the maximum depth of the tree.  If `print_training_error` is specified, the training error will be printed.
+
+Test data may be specified with the `test` parameter, and if performance numbers are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions` output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a decision tree with a minimum leaf size of 20 on the dataset contained in `"data"` with labels `"labels"`, saving the output model to `"tree"` and printing the training error, one could call
+
+```R
+R> output <- decision_tree(training=data, labels=labels, minimum_leaf_size=20,
+  minimum_gain_split=0.001, print_training_accuracy=TRUE)
+R> tree <- output$output_model
+```
+
+Then, to use that model to classify points in `"test_set"` and print the test error given the labels `"test_labels"` using that model, while saving the predictions for each point to `"predictions"`, one could call 
+
+```R
+R> output <- decision_tree(input_model=tree, test=test_set,
+  test_labels=test_labels)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [Random forest](#random_forest)
+ - [Decision trees on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+ - [Induction of Decision Trees (pdf)](http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf)
+ - [DecisionTree C++ class documentation](../../user/methods/decision_tree.md)
+
+## det()
+{: #det }
+
+#### Density Estimation With Density Estimation Trees
+{: #det_descr }
+
+```R
+R> library(mlpack)
+R> d <- det(folds=10, input_model=NA, max_leaf_size=10, min_leaf_size=5,
+        path_format="lr", skip_pruning=FALSE, test=matrix(numeric(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output_model <- d$output_model
+R> tag_counters_file <- d$tag_counters_file
+R> tag_file <- d$tag_file
+R> test_set_estimates <- d$test_set_estimates
+R> training_set_estimates <- d$training_set_estimates
+R> vi <- d$vi
+```
+
+An implementation of density estimation trees for the density estimation task.  Density estimation trees can be trained or used to predict the density at locations given by query points. [Detailed documentation](#det_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `folds` | [`integer`](#doc_integer) | The number of folds of cross-validation to perform for the estimation (0 is LOOCV) | `10` |
+| `input_model` | [`DTree`](#doc_model) | Trained density estimation tree to load. | `NA` |
+| `max_leaf_size` | [`integer`](#doc_integer) | The maximum size of a leaf in the unpruned, fully grown DET. | `10` |
+| `min_leaf_size` | [`integer`](#doc_integer) | The minimum size of a leaf in the unpruned, fully grown DET. | `5` |
+| `path_format` | [`character`](#doc_character) | The format of path printing: 'lr', 'id-lr', or 'lr-id'. | `"lr"` |
+| `skip_pruning` | [`logical`](#doc_logical) | Whether to bypass the pruning process and output the unpruned tree only. | `FALSE` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | A set of test points to estimate the density of. | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | The data set on which to build a density estimation tree. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`DTree`](#doc_model) | Output to save trained density estimation tree to. | 
+| `tag_counters_file` | [`character`](#doc_character) | The file to output the number of points that went to each leaf. | 
+| `tag_file` | [`character`](#doc_character) | The file to output the tags (and possibly paths) for each sample in the test set. | 
+| `test_set_estimates` | [`numeric matrix`](#doc_numeric_matrix) | The output estimates on the test set from the final optimally pruned tree. | 
+| `training_set_estimates` | [`numeric matrix`](#doc_numeric_matrix) | The output density estimates on the training set from the final optimally pruned tree. | 
+| `vi` | [`numeric matrix`](#doc_numeric_matrix) | The output variable importance values for each feature. | 
+
+### Detailed documentation
+{: #det_detailed-documentation }
+
+This program performs a number of functions related to Density Estimation Trees.  The optimal Density Estimation Tree (DET) can be trained on a set of data (specified by `training`) using cross-validation (with number of folds specified with the `folds` parameter).  This trained density estimation tree may then be saved with the `output_model` output parameter.
+
+The variable importances (that is, the feature importance values for each dimension) may be saved with the `vi` output parameter, and the density estimates for each training point may be saved with the `training_set_estimates` output parameter.
+
+Enabling path printing for each node outputs the path from the root node to a leaf for each entry in the test set, or training set (if a test set is not provided).  Strings like 'LRLRLR' (indicating that traversal went to the left child, then the right child, then the left child, and so forth) will be output. If 'lr-id' or 'id-lr' are given as the `path_format` parameter, then the ID (tag) of every node along the path will be printed after or before the L or R character indicating the direction of traversal, respectively.
+
+This program also can provide density estimates for a set of test points, specified in the `test` parameter.  The density estimation tree used for this task will be the tree that was trained on the given training points, or a tree given as the parameter `input_model`.  The density estimates for the test points may be saved using the `test_set_estimates` output parameter.
+
+### See also
+
+ - [Density estimation on Wikipedia](https://en.wikipedia.org/wiki/Density_estimation)
+ - [Density estimation trees (pdf)](https://www.mlpack.org/papers/det.pdf)
+ - [DTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/det/dtree.hpp)
+
+## emst()
+{: #emst }
+
+#### Fast Euclidean Minimum Spanning Tree
+{: #emst_descr }
+
+```R
+R> library(mlpack)
+R> d <- emst(input=matrix(numeric(), 0, 0), leaf_size=1, naive=FALSE,
+        verbose=FALSE)
+R> output <- d$output
+```
+
+An implementation of the Dual-Tree Boruvka algorithm for computing the Euclidean minimum spanning tree of a set of input points. [Detailed documentation](#emst_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input data matrix. | `**--**` |
+| `leaf_size` | [`integer`](#doc_integer) | Leaf size in the kd-tree.  One-element leaves give the empirically best performance, but at the cost of greater memory requirements. | `1` |
+| `naive` | [`logical`](#doc_logical) | Compute the MST using O(n^2) naive algorithm. | `FALSE` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Output data.  Stored as an edge list. | 
+
+### Detailed documentation
+{: #emst_detailed-documentation }
+
+This program can compute the Euclidean minimum spanning tree of a set of input points using the dual-tree Boruvka algorithm.
+
+The set to calculate the minimum spanning tree of is specified with the `input` parameter, and the output may be saved with the `output` output parameter.
+
+The `leaf_size` parameter controls the leaf size of the kd-tree that is used to calculate the minimum spanning tree, and if the `naive` option is given, then brute-force search is used (this is typically much slower in low dimensions).  The leaf size does not affect the results, but it may have some effect on the runtime of the algorithm.
+
+### Example
+For example, the minimum spanning tree of the input dataset `"data"` can be calculated with a leaf size of 20 and stored as `"spanning_tree"` using the following command:
+
+```R
+R> output <- emst(input=data, leaf_size=20)
+R> spanning_tree <- output$output
+```
+
+The output matrix is a three-dimensional matrix, where each row indicates an edge.  The first dimension corresponds to the lesser index of the edge; the second dimension corresponds to the greater index of the edge; and the third column corresponds to the distance between the two points.
+
+### See also
+
+ - [Minimum spanning tree on Wikipedia](https://en.wikipedia.org/wiki/Minimum_spanning_tree)
+ - [Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis, and Applications (pdf)](https://www.mlpack.org/papers/emst.pdf)
+ - [DualTreeBoruvka class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/emst/dtb.hpp)
+
+## fastmks()
+{: #fastmks }
+
+#### FastMKS (Fast Max-Kernel Search)
+{: #fastmks_descr }
+
+```R
+R> library(mlpack)
+R> d <- fastmks(bandwidth=1, base=2, degree=2, input_model=NA, k=0,
+        kernel="linear", naive=FALSE, offset=0, query=matrix(numeric(), 0, 0),
+        reference=matrix(numeric(), 0, 0), scale=1, single=FALSE,
+        verbose=FALSE)
+R> indices <- d$indices
+R> kernels <- d$kernels
+R> output_model <- d$output_model
+```
+
+An implementation of the single-tree and dual-tree fast max-kernel search (FastMKS) algorithm.  Given a set of reference points and a set of query points, this can find the reference point with maximum kernel value for each query point; trained models can be reused for future queries. [Detailed documentation](#fastmks_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`numeric`](#doc_numeric) | Bandwidth (for Gaussian, Epanechnikov, and triangular kernels). | `1` |
+| `base` | [`numeric`](#doc_numeric) | Base to use during cover tree construction. | `2` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `degree` | [`numeric`](#doc_numeric) | Degree of polynomial kernel. | `2` |
+| `input_model` | [`FastMKSModel`](#doc_model) | Input FastMKS model to use. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of maximum kernels to find. | `0` |
+| `kernel` | [`character`](#doc_character) | Kernel type to use: 'linear', 'polynomial', 'cosine', 'gaussian', 'epanechnikov', 'triangular', 'hyptan'. | `"linear"` |
+| `naive` | [`logical`](#doc_logical) | If true, O(n^2) naive mode is used for computation. | `FALSE` |
+| `offset` | [`numeric`](#doc_numeric) | Offset of kernel (for polynomial and hyptan kernels). | `0` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | The query dataset. | `matrix(numeric(), 0, 0)` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | The reference dataset. | `matrix(numeric(), 0, 0)` |
+| `scale` | [`numeric`](#doc_numeric) | Scale of kernel (for hyptan kernel). | `1` |
+| `single` | [`logical`](#doc_logical) | If true, single-tree search is used (as opposed to dual-tree search. | `FALSE` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `indices` | [`integer matrix`](#doc_integer_matrix) | Output matrix of indices. | 
+| `kernels` | [`numeric matrix`](#doc_numeric_matrix) | Output matrix of kernels. | 
+| `output_model` | [`FastMKSModel`](#doc_model) | Output for FastMKS model. | 
+
+### Detailed documentation
+{: #fastmks_detailed-documentation }
+
+This program will find the k maximum kernels of a set of points, using a query set and a reference set (which can optionally be the same set). More specifically, for each point in the query set, the k points in the reference set with maximum kernel evaluations are found.  The kernel function used is specified with the `kernel` parameter.
+
+### Example
+For example, the following command will calculate, for each point in the query set `"query"`, the five points in the reference set `"reference"` with maximum kernel evaluation using the linear kernel.  The kernel evaluations may be saved with the  `"kernels"` output parameter and the indices may be saved with the `"indices"` output parameter.
+
+```R
+R> output <- fastmks(k=5, reference=reference, query=query, kernel="linear")
+R> indices <- output$indices
+R> kernels <- output$kernels
+```
+
+The output matrices are organized such that row i and column j in the indices matrix corresponds to the index of the point in the reference set that has j'th largest kernel evaluation with the point in the query set with index i.  Row i and column j in the kernels matrix corresponds to the kernel evaluation between those two points.
+
+This program performs FastMKS using a cover tree.  The base used to build the cover tree can be specified with the `base` parameter.
+
+### See also
+
+ - [k-nearest-neighbor search](#knn)
+ - [Dual-tree Fast Exact Max-Kernel Search (pdf)](https://mlpack.org/papers/fmks.pdf)
+ - [FastMKS class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/fastmks/fastmks.hpp)
+
+## gmm_train()
+{: #gmm_train }
+
+#### Gaussian Mixture Model (GMM) Training
+{: #gmm_train_descr }
+
+```R
+R> library(mlpack)
+R> d <- gmm_train(diagonal_covariance=FALSE, gaussians=0,
+        input=matrix(numeric(), 0, 0), input_model=NA,
+        kmeans_max_iterations=1000, max_iterations=250, no_force_positive=FALSE,
+        noise=0, percentage=0.02, refined_start=FALSE, samplings=100, seed=0,
+        tolerance=1e-10, trials=1, verbose=FALSE)
+R> output_model <- d$output_model
+```
+
+An implementation of the EM algorithm for training Gaussian mixture models (GMMs).  Given a dataset, this can train a GMM for future use with other tools. [Detailed documentation](#gmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `diagonal_covariance` | [`logical`](#doc_logical) | Force the covariance of the Gaussians to be diagonal.  This can accelerate training time significantly. | `FALSE` |
+| `gaussians` | [`integer`](#doc_integer) | Number of Gaussians in the GMM. | `**--**` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | The training data on which the model will be fit. | `**--**` |
+| `input_model` | [`GMM`](#doc_model) | Initial input GMM model to start training with. | `NA` |
+| `kmeans_max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations for the k-means algorithm (used to initialize EM). | `1000` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations of EM algorithm (passing 0 will run until convergence). | `250` |
+| `no_force_positive` | [`logical`](#doc_logical) | Do not force the covariance matrices to be positive definite. | `FALSE` |
+| `noise` | [`numeric`](#doc_numeric) | Variance of zero-mean Gaussian noise to add to data. | `0` |
+| `percentage` | [`numeric`](#doc_numeric) | If using --refined_start, specify the percentage of the dataset used for each sampling (should be between 0.0 and 1.0). | `0.02` |
+| `refined_start` | [`logical`](#doc_logical) | During the initialization, use refined initial positions for k-means clustering (Bradley and Fayyad, 1998). | `FALSE` |
+| `samplings` | [`integer`](#doc_integer) | If using --refined_start, specify the number of samplings used for initial points. | `100` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tolerance` | [`numeric`](#doc_numeric) | Tolerance for convergence of EM. | `1e-10` |
+| `trials` | [`integer`](#doc_integer) | Number of trials to perform in training GMM. | `1` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`GMM`](#doc_model) | Output for trained GMM model. | 
+
+### Detailed documentation
+{: #gmm_train_detailed-documentation }
+
+This program takes a parametric estimate of a Gaussian mixture model (GMM) using the EM algorithm to find the maximum likelihood estimate.  The model may be saved and reused by other mlpack GMM tools.
+
+The input data to train on must be specified with the `input` parameter, and the number of Gaussians in the model must be specified with the `gaussians` parameter.  Optionally, many trials with different random initializations may be run, and the result with highest log-likelihood on the training data will be taken.  The number of trials to run is specified with the `trials` parameter.  By default, only one trial is run.
+
+The tolerance for convergence and maximum number of iterations of the EM algorithm are specified with the `tolerance` and `max_iterations` parameters, respectively.  The GMM may be initialized for training with another model, specified with the `input_model` parameter. Otherwise, the model is initialized by running k-means on the data.  The k-means clustering initialization can be controlled with the `kmeans_max_iterations`, `refined_start`, `samplings`, and `percentage` parameters.  If `refined_start` is specified, then the Bradley-Fayyad refined start initialization will be used.  This can often lead to better clustering results.
+
+The 'diagonal_covariance' flag will cause the learned covariances to be diagonal matrices.  This significantly simplifies the model itself and causes training to be faster, but restricts the ability to fit more complex GMMs.
+
+If GMM training fails with an error indicating that a covariance matrix could not be inverted, make sure that the `no_force_positive` parameter is not specified.  Alternately, adding a small amount of Gaussian noise (using the `noise` parameter) to the entire dataset may help prevent Gaussians with zero variance in a particular dimension, which is usually the cause of non-invertible covariance matrices.
+
+The `no_force_positive` parameter, if set, will avoid the checks after each iteration of the EM algorithm which ensure that the covariance matrices are positive definite.  Specifying the flag can cause faster runtime, but may also cause non-positive definite covariance matrices, which will cause the program to crash.
+
+### Example
+As an example, to train a 6-Gaussian GMM on the data in `"data"` with a maximum of 100 iterations of EM and 3 trials, saving the trained GMM to `"gmm"`, the following command can be used:
+
+```R
+R> output <- gmm_train(input=data, gaussians=6, trials=3)
+R> gmm <- output$output_model
+```
+
+To re-train that GMM on another set of data `"data2"`, the following command may be used: 
+
+```R
+R> output <- gmm_train(input_model=gmm, input=data2, gaussians=6)
+R> new_gmm <- output$output_model
+```
+
+### See also
+
+ - [gmm_generate()](#gmm_generate)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_generate()
+{: #gmm_generate }
+
+#### GMM Sample Generator
+{: #gmm_generate_descr }
+
+```R
+R> library(mlpack)
+R> d <- gmm_generate(input_model=NA, samples=0, seed=0, verbose=FALSE)
+R> output <- d$output
+```
+
+A sample generator for pre-trained GMMs.  Given a pre-trained GMM, this can sample new points randomly from that distribution. [Detailed documentation](#gmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`GMM`](#doc_model) | Input GMM model to generate samples from. | `**--**` |
+| `samples` | [`integer`](#doc_integer) | Number of samples to generate. | `**--**` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save output samples in. | 
+
+### Detailed documentation
+{: #gmm_generate_detailed-documentation }
+
+This program is able to generate samples from a pre-trained GMM (use gmm_train to train a GMM).  The pre-trained GMM must be specified with the `input_model` parameter.  The number of samples to generate is specified by the `samples` parameter.  Output samples may be saved with the `output` output parameter.
+
+### Example
+The following command can be used to generate 100 samples from the pre-trained GMM `"gmm"` and store those generated samples in `"samples"`:
+
+```R
+R> output <- gmm_generate(input_model=gmm, samples=100)
+R> samples <- output$output
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_probability()](#gmm_probability)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## gmm_probability()
+{: #gmm_probability }
+
+#### GMM Probability Calculator
+{: #gmm_probability_descr }
+
+```R
+R> library(mlpack)
+R> d <- gmm_probability(input=matrix(numeric(), 0, 0), input_model=NA,
+        verbose=FALSE)
+R> output <- d$output
+```
+
+A probability calculator for GMMs.  Given a pre-trained GMM and a set of points, this can compute the probability that each point is from the given GMM. [Detailed documentation](#gmm_probability_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input matrix to calculate probabilities of. | `**--**` |
+| `input_model` | [`GMM`](#doc_model) | Input GMM to use as model. | `**--**` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to store calculated probabilities in. | 
+
+### Detailed documentation
+{: #gmm_probability_detailed-documentation }
+
+This program calculates the probability that given points came from a given GMM (that is, P(X \| gmm)).  The GMM is specified with the `input_model` parameter, and the points are specified with the `input` parameter.  The output probabilities may be saved via the `output` output parameter.
+
+### Example
+So, for example, to calculate the probabilities of each point in `"points"` coming from the pre-trained GMM `"gmm"`, while storing those probabilities in `"probs"`, the following command could be used:
+
+```R
+R> output <- gmm_probability(input_model=gmm, input=points)
+R> probs <- output$output
+```
+
+### See also
+
+ - [gmm_train()](#gmm_train)
+ - [gmm_generate()](#gmm_generate)
+ - [Gaussian Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Mixture_model#Gaussian_mixture_model)
+ - [GMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/gmm/gmm.hpp)
+
+## hmm_train()
+{: #hmm_train }
+
+#### Hidden Markov Model (HMM) Training
+{: #hmm_train_descr }
+
+```R
+R> library(mlpack)
+R> d <- hmm_train(batch=FALSE, gaussians=0, input_file="",
+        input_model=NA, labels_file="", seed=0, states=0, tolerance=1e-05,
+        type="gaussian", verbose=FALSE)
+R> output_model <- d$output_model
+```
+
+An implementation of training algorithms for Hidden Markov Models (HMMs). Given labeled or unlabeled data, an HMM can be trained for further use with other mlpack HMM tools. [Detailed documentation](#hmm_train_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch` | [`logical`](#doc_logical) | If true, input_file (and if passed, labels_file) are expected to contain a list of files to use as input observation sequences (and label sequences). | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `gaussians` | [`integer`](#doc_integer) | Number of gaussians in each GMM (necessary when type is 'gmm'). | `0` |
+| `input_file` | [`character`](#doc_character) | File containing input observations. | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | Pre-existing HMM model to initialize training with. | `NA` |
+| `labels_file` | [`character`](#doc_character) | Optional file of hidden states, used for labeled training. | `""` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `states` | [`integer`](#doc_integer) | Number of hidden states in HMM (necessary, unless model_file is specified). | `0` |
+| `tolerance` | [`numeric`](#doc_numeric) | Tolerance of the Baum-Welch algorithm. | `1e-05` |
+| `type` | [`character`](#doc_character) | Type of HMM: discrete \| gaussian \| diag_gmm \| gmm. | `"gaussian"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HMMModel`](#doc_model) | Output for trained HMM. | 
+
+### Detailed documentation
+{: #hmm_train_detailed-documentation }
+
+This program allows a Hidden Markov Model to be trained on labeled or unlabeled data.  It supports four types of HMMs: Discrete HMMs, Gaussian HMMs, GMM HMMs, or Diagonal GMM HMMs
+
+Either one input sequence can be specified (with `input_file`), or, a file containing files in which input sequences can be found (when `input_file`and`batch` are used together).  In addition, labels can be provided in the file specified by `labels_file`, and if `batch` is used, the file given to `labels_file` should contain a list of files of labels corresponding to the sequences in the file given to `input_file`.
+
+The HMM is trained with the Baum-Welch algorithm if no labels are provided.  The tolerance of the Baum-Welch algorithm can be set with the `tolerance`option.  By default, the transition matrix is randomly initialized and the emission distributions are initialized to fit the extent of the data.
+
+Optionally, a pre-created HMM model can be used as a guess for the transition matrix and emission probabilities; this is specifiable with `output_model`.
+
+### See also
+
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_generate()
+{: #hmm_generate }
+
+#### Hidden Markov Model (HMM) Sequence Generator
+{: #hmm_generate_descr }
+
+```R
+R> library(mlpack)
+R> d <- hmm_generate(length=0, model=NA, seed=0, start_state=0,
+        verbose=FALSE)
+R> output <- d$output
+R> state <- d$state
+```
+
+A utility to generate random sequences from a pre-trained Hidden Markov Model (HMM).  The length of the desired sequence can be specified, and a random sequence of observations is returned. [Detailed documentation](#hmm_generate_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `length` | [`integer`](#doc_integer) | Length of sequence to generate. | `**--**` |
+| `model` | [`HMMModel`](#doc_model) | Trained HMM to generate sequences with. | `**--**` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `start_state` | [`integer`](#doc_integer) | Starting state of sequence. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save observation sequence to. | 
+| `state` | [`integer matrix`](#doc_integer_matrix) | Matrix to save hidden state sequence to. | 
+
+### Detailed documentation
+{: #hmm_generate_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as the `model` parameter, and generates a random observation sequence and hidden state sequence based on its parameters. The observation sequence may be saved with the `output` output parameter, and the internal state  sequence may be saved with the `state` output parameter.
+
+The state to start the sequence in may be specified with the `start_state` parameter.
+
+### Example
+For example, to generate a sequence of length 150 from the HMM `"hmm"` and save the observation sequence to `"observations"` and the hidden state sequence to `"states"`, the following command may be used: 
+
+```R
+R> output <- hmm_generate(model=hmm, length=150)
+R> observations <- output$output
+R> states <- output$state
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_loglik()](#hmm_loglik)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_loglik()
+{: #hmm_loglik }
+
+#### Hidden Markov Model (HMM) Sequence Log-Likelihood
+{: #hmm_loglik_descr }
+
+```R
+R> library(mlpack)
+R> d <- hmm_loglik(input=matrix(numeric(), 0, 0), input_model=NA,
+        verbose=FALSE)
+R> log_likelihood <- d$log_likelihood
+```
+
+A utility for computing the log-likelihood of a sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observation sequence, this computes and returns the log-likelihood of that sequence being observed from that HMM. [Detailed documentation](#hmm_loglik_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | File containing observations, | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | File containing HMM. | `**--**` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `log_likelihood` | [`numeric`](#doc_numeric) | Log-likelihood of the sequence. | 
+
+### Detailed documentation
+{: #hmm_loglik_detailed-documentation }
+
+This utility takes an already-trained HMM, specified with the `input_model` parameter, and evaluates the log-likelihood of a sequence of observations, given with the `input` parameter.  The computed log-likelihood is given as output.
+
+### Example
+For example, to compute the log-likelihood of the sequence `"seq"` with the pre-trained HMM `"hmm"`, the following command may be used: 
+
+```R
+R> hmm_loglik(input=seq, input_model=hmm)
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_viterbi()](#hmm_viterbi)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hmm_viterbi()
+{: #hmm_viterbi }
+
+#### Hidden Markov Model (HMM) Viterbi State Prediction
+{: #hmm_viterbi_descr }
+
+```R
+R> library(mlpack)
+R> d <- hmm_viterbi(input=matrix(numeric(), 0, 0), input_model=NA,
+        verbose=FALSE)
+R> output <- d$output
+```
+
+A utility for computing the most probable hidden state sequence for Hidden Markov Models (HMMs).  Given a pre-trained HMM and an observed sequence, this uses the Viterbi algorithm to compute and return the most probable hidden state sequence. [Detailed documentation](#hmm_viterbi_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing observations, | `**--**` |
+| `input_model` | [`HMMModel`](#doc_model) | Trained HMM to use. | `**--**` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`integer matrix`](#doc_integer_matrix) | File to save predicted state sequence to. | 
+
+### Detailed documentation
+{: #hmm_viterbi_detailed-documentation }
+
+This utility takes an already-trained HMM, specified as `input_model`, and evaluates the most probable hidden state sequence of a given sequence of observations (specified as '`input`, using the Viterbi algorithm.  The computed state sequence may be saved using the `output` output parameter.
+
+### Example
+For example, to predict the state sequence of the observations `"obs"` using the HMM `"hmm"`, storing the predicted state sequence to `"states"`, the following command could be used:
+
+```R
+R> output <- hmm_viterbi(input=obs, input_model=hmm)
+R> states <- output$output
+```
+
+### See also
+
+ - [hmm_train()](#hmm_train)
+ - [hmm_generate()](#hmm_generate)
+ - [hmm_loglik()](#hmm_loglik)
+ - [Hidden Mixture Models on Wikipedia](https://en.wikipedia.org/wiki/Hidden_Markov_model)
+ - [HMM class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hmm/hmm.hpp)
+
+## hoeffding_tree()
+{: #hoeffding_tree }
+
+#### Hoeffding trees
+{: #hoeffding_tree_descr }
+
+```R
+R> library(mlpack)
+R> d <- hoeffding_tree(batch_mode=FALSE, bins=10, confidence=0.95,
+        info_gain=FALSE, input_model=NA, labels=matrix(integer(), 0, 0),
+        max_samples=5000, min_samples=100, numeric_split_strategy="binary",
+        observations_before_binning=100, passes=1, test=matrix(numeric(), 0, 0),
+        test_labels=matrix(integer(), 0, 0), training=matrix(numeric(), 0, 0),
+        verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of Hoeffding trees, a form of streaming decision tree for classification.  Given labeled data, a Hoeffding tree can be trained and saved for later use, or a pre-trained Hoeffding tree can be used for predicting the classifications of new points. [Detailed documentation](#hoeffding_tree_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_mode` | [`logical`](#doc_logical) | If true, samples will be considered in batch instead of as a stream.  This generally results in better trees but at the cost of memory usage and runtime. | `FALSE` |
+| `bins` | [`integer`](#doc_integer) | If the 'domingos' split strategy is used, this specifies the number of bins for each numeric split. | `10` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `confidence` | [`numeric`](#doc_numeric) | Confidence before splitting (between 0 and 1). | `0.95` |
+| `info_gain` | [`logical`](#doc_logical) | If set, information gain is used instead of Gini impurity for calculating Hoeffding bounds. | `FALSE` |
+| `input_model` | [`HoeffdingTreeModel`](#doc_model) | Input trained Hoeffding tree model. | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Labels for training dataset. | `matrix(integer(), 0, 0)` |
+| `max_samples` | [`integer`](#doc_integer) | Maximum number of samples before splitting. | `5000` |
+| `min_samples` | [`integer`](#doc_integer) | Minimum number of samples before splitting. | `100` |
+| `numeric_split_strategy` | [`character`](#doc_character) | The splitting strategy to use for numeric features: 'domingos' or 'binary'. | `"binary"` |
+| `observations_before_binning` | [`integer`](#doc_integer) | If the 'domingos' split strategy is used, this specifies the number of samples observed before binning is performed. | `100` |
+| `passes` | [`integer`](#doc_integer) | Number of passes to take over the dataset. | `1` |
+| `test` | [`categorical matrix/data.frame`](#doc_categorical_matrix_data_frame) | Testing dataset (may be categorical). | `matrix(numeric(), 0, 0)` |
+| `test_labels` | [`integer vector`](#doc_integer_vector) | Labels of test data. | `matrix(integer(), 0, 0)` |
+| `training` | [`categorical matrix/data.frame`](#doc_categorical_matrix_data_frame) | Training dataset (may be categorical). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`HoeffdingTreeModel`](#doc_model) | Output for trained Hoeffding tree model. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | Matrix to output label predictions for test data into. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | In addition to predicting labels, provide rediction probabilities in this matrix. | 
+
+### Detailed documentation
+{: #hoeffding_tree_detailed-documentation }
+
+This program implements Hoeffding trees, a form of streaming decision tree suited best for large (or streaming) datasets.  This program supports both categorical and numeric data.  Given an input dataset, this program is able to train the tree with numerous training options, and save the model to a file.  The program is also able to use a trained model or a model from file in order to predict classes for a given test set.
+
+The training file and associated labels are specified with the `training` and `labels` parameters, respectively. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+The training may be performed in batch mode (like a typical decision tree algorithm) by specifying the `batch_mode` option, but this may not be the best option for large datasets.
+
+When a model is trained, it may be saved via the `output_model` output parameter.  A model may be loaded from file for further training or testing with the `input_model` parameter.
+
+Test data may be specified with the `test` parameter, and if performance statistics are desired for that test set, labels may be specified with the `test_labels` parameter.  Predictions for each test point may be saved with the `predictions` output parameter, and class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a Hoeffding tree with confidence 0.99 with data `"dataset"`, saving the trained tree to `"tree"`, the following command may be used:
+
+```R
+R> output <- hoeffding_tree(training=dataset, confidence=0.99)
+R> tree <- output$output_model
+```
+
+Then, this tree may be used to make predictions on the test set `"test_set"`, saving the predictions into `"predictions"` and the class probabilities into `"class_probs"` with the following command: 
+
+```R
+R> output <- hoeffding_tree(input_model=tree, test=test_set)
+R> predictions <- output$predictions
+R> class_probs <- output$probabilities
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [random_forest()](#random_forest)
+ - [Mining High-Speed Data Streams (pdf)](http://dm.cs.washington.edu/papers/vfdt-kdd00.pdf)
+ - [HoeffdingTree class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp)
+
+## kde()
+{: #kde }
+
+#### Kernel Density Estimation
+{: #kde_descr }
+
+```R
+R> library(mlpack)
+R> d <- kde(abs_error=0, algorithm="dual-tree", bandwidth=1,
+        initial_sample_size=100, input_model=NA, kernel="gaussian",
+        mc_break_coef=0.4, mc_entry_coef=3, mc_probability=0.95,
+        monte_carlo=FALSE, query=matrix(numeric(), 0, 0),
+        reference=matrix(numeric(), 0, 0), rel_error=0.05, tree="kd-tree",
+        verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+```
+
+An implementation of kernel density estimation with dual-tree algorithms. Given a set of reference points and query points and a kernel function, this can estimate the density function at the location of each query point using trees; trees that are built can be saved for later use. [Detailed documentation](#kde_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `abs_error` | [`numeric`](#doc_numeric) | Relative error tolerance for the prediction. | `0` |
+| `algorithm` | [`character`](#doc_character) | Algorithm to use for the prediction.('dual-tree', 'single-tree'). | `"dual-tree"` |
+| `bandwidth` | [`numeric`](#doc_numeric) | Bandwidth of the kernel. | `1` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `initial_sample_size` | [`integer`](#doc_integer) | Initial sample size for Monte Carlo estimations. | `100` |
+| `input_model` | [`KDEModel`](#doc_model) | Contains pre-trained KDE model. | `NA` |
+| `kernel` | [`character`](#doc_character) | Kernel to use for the prediction.('gaussian', 'epanechnikov', 'laplacian', 'spherical', 'triangular'). | `"gaussian"` |
+| `mc_break_coef` | [`numeric`](#doc_numeric) | Controls what fraction of the amount of node's descendants is the limit for the sample size before it recurses. | `0.4` |
+| `mc_entry_coef` | [`numeric`](#doc_numeric) | Controls how much larger does the amount of node descendants has to be compared to the initial sample size in order to be a candidate for Monte Carlo estimations. | `3` |
+| `mc_probability` | [`numeric`](#doc_numeric) | Probability of the estimation being bounded by relative error when using Monte Carlo estimations. | `0.95` |
+| `monte_carlo` | [`logical`](#doc_logical) | Whether to use Monte Carlo estimations when possible. | `FALSE` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Query dataset to KDE on. | `matrix(numeric(), 0, 0)` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Input reference dataset use for KDE. | `matrix(numeric(), 0, 0)` |
+| `rel_error` | [`numeric`](#doc_numeric) | Relative error tolerance for the prediction. | `0.05` |
+| `tree` | [`character`](#doc_character) | Tree to use for the prediction.('kd-tree', 'ball-tree', 'cover-tree', 'octree', 'r-tree'). | `"kd-tree"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`KDEModel`](#doc_model) | If specified, the KDE model will be saved here. | 
+| `predictions` | [`numeric vector`](#doc_numeric_vector) | Vector to store density predictions. | 
+
+### Detailed documentation
+{: #kde_detailed-documentation }
+
+This program performs a Kernel Density Estimation. KDE is a non-parametric way of estimating probability density function. For each query point the program will estimate its probability density by applying a kernel function to each reference point. The computational complexity of this is O(N^2) where there are N query points and N reference points, but this implementation will typically see better performance as it uses an approximate dual or single tree algorithm for acceleration.
+
+Dual or single tree optimization avoids many barely relevant calculations (as kernel function values decrease with distance), so it is an approximate computation. You can specify the maximum relative error tolerance for each query value with `rel_error` as well as the maximum absolute error tolerance with the parameter `abs_error`. This program runs using an Euclidean metric. Kernel function can be selected using the `kernel` option. You can also choose what which type of tree to use for the dual-tree algorithm with `tree`. It is also possible to select whether to use dual-tree algorithm or single-tree algorithm using the `algorithm` option.
+
+Monte Carlo estimations can be used to accelerate the KDE estimate when the Gaussian Kernel is used. This provides a probabilistic guarantee on the the error of the resulting KDE instead of an absolute guarantee.To enable Monte Carlo estimations, the `monte_carlo` flag can be used, and success probability can be set with the `mc_probability` option. It is possible to set the initial sample size for the Monte Carlo estimation using `initial_sample_size`. This implementation will only consider a node, as a candidate for the Monte Carlo estimation, if its number of descendant nodes is bigger than the initial sample size. This can be controlled using a coefficient that will multiply the initial sample size and can be set using `mc_entry_coef`. To avoid using the same amount of computations an exact approach would take, this program recurses the tree whenever a fraction of the amount of the node's descendant points have already been computed. This fraction is set using `mc_break_coef`.
+
+### Example
+For example, the following will run KDE using the data in `"ref_data"` for training and the data in `"qu_data"` as query data. It will apply an Epanechnikov kernel with a 0.2 bandwidth to each reference point and use a KD-Tree for the dual-tree optimization. The returned predictions will be within 5% of the real KDE value for each query point.
+
+```R
+R> output <- kde(reference=ref_data, query=qu_data, bandwidth=0.2,
+  kernel="epanechnikov", tree="kd-tree", rel_error=0.05)
+R> out_data <- output$predictions
+```
+
+the predicted density estimations will be stored in `"out_data"`.
+If no `query` is provided, then KDE will be computed on the `reference` dataset.
+It is possible to select either a reference dataset or an input model but not both at the same time. If an input model is selected and parameter values are not set (e.g. `bandwidth`) then default parameter values will be used.
+
+In addition to the last program call, it is also possible to activate Monte Carlo estimations if a Gaussian kernel is used. This can provide faster results, but the KDE will only have a probabilistic guarantee of meeting the desired error bound (instead of an absolute guarantee). The following example will run KDE using a Monte Carlo estimation when possible. The results will be within a 5% of the real KDE value with a 95% probability. Initial sample size for the Monte Carlo estimation will be 200 points and a node will be a candidate for the estimation only when it contains 700 (i.e. 3.5*200) points. If a node contains 700 points and 420 (i.e. 0.6*700) have already been sampled, then the algorithm will recurse instead of keep sampling.
+
+```R
+R> output <- kde(reference=ref_data, query=qu_data, bandwidth=0.2,
+  kernel="gaussian", tree="kd-tree", rel_error=0.05, monte_carlo=,
+  mc_probability=0.95, initial_sample_size=200, mc_entry_coef=3.5,
+  mc_break_coef=0.6)
+R> out_data <- output$predictions
+```
+
+### See also
+
+ - [knn()](#knn)
+ - [Kernel density estimation on Wikipedia](https://en.wikipedia.org/wiki/Kernel_density_estimation)
+ - [Tree-Independent Dual-Tree Algorithms](https://arxiv.org/pdf/1304.4327)
+ - [Fast High-dimensional Kernel Summations Using the Monte Carlo Multipole Method](http://papers.nips.cc/paper/3539-fast-high-dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf)
+ - [KDE C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kde/kde.hpp)
+
+## kernel_pca()
+{: #kernel_pca }
+
+#### Kernel Principal Components Analysis
+{: #kernel_pca_descr }
+
+```R
+R> library(mlpack)
+R> d <- kernel_pca(bandwidth=1, center=FALSE, degree=1,
+        input=matrix(numeric(), 0, 0), kernel="", kernel_scale=1,
+        new_dimensionality=0, nystroem_method=FALSE, offset=0,
+        sampling="kmeans", verbose=FALSE)
+R> output <- d$output
+```
+
+An implementation of Kernel Principal Components Analysis (KPCA).  This can be used to perform nonlinear dimensionality reduction or preprocessing on a given dataset. [Detailed documentation](#kernel_pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bandwidth` | [`numeric`](#doc_numeric) | Bandwidth, for 'gaussian' and 'laplacian' kernels. | `1` |
+| `center` | [`logical`](#doc_logical) | If set, the transformed data will be centered about the origin. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `degree` | [`numeric`](#doc_numeric) | Degree of polynomial, for 'polynomial' kernel. | `1` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform KPCA on. | `**--**` |
+| `kernel` | [`character`](#doc_character) | The kernel to use; see the above documentation for the list of usable kernels. | `**--**` |
+| `kernel_scale` | [`numeric`](#doc_numeric) | Scale, for 'hyptan' kernel. | `1` |
+| `new_dimensionality` | [`integer`](#doc_integer) | If not 0, reduce the dimensionality of the output dataset by ignoring the dimensions with the smallest eigenvalues. | `0` |
+| `nystroem_method` | [`logical`](#doc_logical) | If set, the Nystroem method will be used. | `FALSE` |
+| `offset` | [`numeric`](#doc_numeric) | Offset, for 'hyptan' and 'polynomial' kernels. | `0` |
+| `sampling` | [`character`](#doc_character) | Sampling scheme to use for the Nystroem method: 'kmeans', 'random', 'ordered' | `"kmeans"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #kernel_pca_detailed-documentation }
+
+This program performs Kernel Principal Components Analysis (KPCA) on the specified dataset with the specified kernel.  This will transform the data onto the kernel principal components, and optionally reduce the dimensionality by ignoring the kernel principal components with the smallest eigenvalues.
+
+For the case where a linear kernel is used, this reduces to regular PCA.
+
+The kernels that are supported are listed below:
+
+ * 'linear': the standard linear dot product (same as normal PCA):
+    `K(x, y) = x^T y`
+
+ * 'gaussian': a Gaussian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\| ^ 2) / (2 * (bandwidth ^ 2)))`
+
+ * 'polynomial': polynomial kernel; requires offset and degree:
+    `K(x, y) = (x^T y + offset) ^ degree`
+
+ * 'hyptan': hyperbolic tangent kernel; requires scale and offset:
+    `K(x, y) = tanh(scale * (x^T y) + offset)`
+
+ * 'laplacian': Laplacian kernel; requires bandwidth:
+    `K(x, y) = exp(-(\|\| x - y \|\|) / bandwidth)`
+
+ * 'epanechnikov': Epanechnikov kernel; requires bandwidth:
+    `K(x, y) = max(0, 1 - \|\| x - y \|\|^2 / bandwidth^2)`
+
+ * 'cosine': cosine distance:
+    `K(x, y) = 1 - (x^T y) / (\|\| x \|\| * \|\| y \|\|)`
+
+The parameters for each of the kernels should be specified with the options `bandwidth`, `kernel_scale`, `offset`, or `degree` (or a combination of those parameters).
+
+Optionally, the Nystroem method ("Using the Nystroem method to speed up kernel machines", 2001) can be used to calculate the kernel matrix by specifying the `nystroem_method` parameter. This approach works by using a subset of the data as basis to reconstruct the kernel matrix; to specify the sampling scheme, the `sampling` parameter is used.  The sampling scheme for the Nystroem method can be chosen from the following list: 'kmeans', 'random', 'ordered'.
+
+### Example
+For example, the following command will perform KPCA on the dataset `"input"` using the Gaussian kernel, and saving the transformed data to `"transformed"`: 
+
+```R
+R> output <- kernel_pca(input=input, kernel="gaussian")
+R> transformed <- output$output
+```
+
+### See also
+
+ - [Kernel principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Kernel_principal_component_analysis)
+ - [Kernel Principal Component Analysis (pdf)](https://pca.narod.ru/scholkopf_kernel.pdf)
+ - [KernelPCA class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+
+## kmeans()
+{: #kmeans }
+
+#### K-Means Clustering
+{: #kmeans_descr }
+
+```R
+R> library(mlpack)
+R> d <- kmeans(algorithm="naive", allow_empty_clusters=FALSE,
+        clusters=0, in_place=FALSE, initial_centroids=matrix(numeric(), 0, 0),
+        input=matrix(numeric(), 0, 0), kill_empty_clusters=FALSE,
+        kmeans_plus_plus=FALSE, labels_only=FALSE, max_iterations=1000,
+        percentage=0.02, refined_start=FALSE, samplings=100, seed=0,
+        verbose=FALSE)
+R> centroid <- d$centroid
+R> output <- d$output
+```
+
+An implementation of several strategies for efficient k-means clustering. Given a dataset and a value of k, this computes and returns a k-means clustering on that data. [Detailed documentation](#kmeans_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`character`](#doc_character) | Algorithm to use for the Lloyd iteration ('naive', 'pelleg-moore', 'elkan', 'hamerly', 'dualtree', or 'dualtree-covertree'). | `"naive"` |
+| `allow_empty_clusters` | [`logical`](#doc_logical) | Allow empty clusters to be persist. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `clusters` | [`integer`](#doc_integer) | Number of clusters to find (0 autodetects from initial centroids). | `**--**` |
+| `in_place` | [`logical`](#doc_logical) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden. (Do not use in Python.) | `FALSE` |
+| `initial_centroids` | [`numeric matrix`](#doc_numeric_matrix) | Start with the specified initial centroids. | `matrix(numeric(), 0, 0)` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform clustering on. | `**--**` |
+| `kill_empty_clusters` | [`logical`](#doc_logical) | Remove empty clusters when they occur. | `FALSE` |
+| `kmeans_plus_plus` | [`logical`](#doc_logical) | Use the k-means++ initialization strategy to choose initial points. | `FALSE` |
+| `labels_only` | [`logical`](#doc_logical) | Only output labels into output file. | `FALSE` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations before k-means terminates. | `1000` |
+| `percentage` | [`numeric`](#doc_numeric) | Percentage of dataset to use for each refined start sampling (use when --refined_start is specified). | `0.02` |
+| `refined_start` | [`logical`](#doc_logical) | Use the refined initial point strategy by Bradley and Fayyad to choose initial points. | `FALSE` |
+| `samplings` | [`integer`](#doc_integer) | Number of samplings to perform for refined start (use when --refined_start is specified). | `100` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`numeric matrix`](#doc_numeric_matrix) | If specified, the centroids of each cluster will  be written to the given file. | 
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to store output labels or labeled data to. | 
+
+### Detailed documentation
+{: #kmeans_detailed-documentation }
+
+This program performs K-Means clustering on the given dataset.  It can return the learned cluster assignments, and the centroids of the clusters.  Empty clusters are not allowed by default; when a cluster becomes empty, the point furthest from the centroid of the cluster with maximum variance is taken to fill that cluster.
+
+Optionally, the strategy to choose initial centroids can be specified.  The k-means++ algorithm can be used to choose initial centroids with the `kmeans_plus_plus` parameter.  The Bradley and Fayyad approach ("Refining initial points for k-means clustering", 1998) can be used to select initial points by specifying the `refined_start` parameter.  This approach works by taking random samplings of the dataset; to specify the number of samplings, the `samplings` parameter is used, and to specify the percentage of the dataset to be used in each sample, the `percentage` parameter is used (it should be a value between 0.0 and 1.0).
+
+There are several options available for the algorithm used for each Lloyd iteration, specified with the `algorithm`  option.  The standard O(kN) approach can be used ('naive').  Other options include the Pelleg-Moore tree-based algorithm ('pelleg-moore'), Elkan's triangle-inequality based algorithm ('elkan'), Hamerly's modification to Elkan's algorithm ('hamerly'), the dual-tree k-means algorithm ('dualtree'), and the dual-tree k-means algorithm using the cover tree ('dualtree-covertree').
+
+The behavior for when an empty cluster is encountered can be modified with the `allow_empty_clusters` option.  When this option is specified and there is a cluster owning no points at the end of an iteration, that cluster's centroid will simply remain in its position from the previous iteration. If the `kill_empty_clusters` option is specified, then when a cluster owns no points at the end of an iteration, the cluster centroid is simply filled with DBL_MAX, killing it and effectively reducing k for the rest of the computation.  Note that the default option when neither empty cluster option is specified can be time-consuming to calculate; therefore, specifying either of these parameters will often accelerate runtime.
+
+Initial clustering assignments may be specified using the `initial_centroids` parameter, and the maximum number of iterations may be specified with the `max_iterations` parameter.
+
+### Example
+As an example, to use Hamerly's algorithm to perform k-means clustering with k=10 on the dataset `"data"`, saving the centroids to `"centroids"` and the assignments for each point to `"assignments"`, the following command could be used:
+
+```R
+R> output <- kmeans(input=data, clusters=10)
+R> assignments <- output$output
+R> centroids <- output$centroid
+```
+
+To run k-means on that same dataset with initial centroids specified in `"initial"` with a maximum of 500 iterations, storing the output centroids in `"final"` the following command may be used:
+
+```R
+R> output <- kmeans(input=data, initial_centroids=initial, clusters=10,
+  max_iterations=500)
+R> final <- output$centroid
+```
+
+### See also
+
+ - [dbscan()](#dbscan)
+ - [k-means++](https://en.wikipedia.org/wiki/K-means%2B%2B)
+ - [Using the triangle inequality to accelerate k-means (pdf)](https://cdn.aaai.org/ICML/2003/ICML03-022.pdf)
+ - [Making k-means even faster (pdf)](https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12)
+ - [Accelerating exact k-means algorithms with geometric reasoning (pdf)](http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/ftp/usr0/ftp/2000/CMU-CS-00-105.pdf)
+ - [A dual-tree algorithm for fast k-means clustering with large k (pdf)](http://www.ratml.org/pub/pdf/2017dual.pdf)
+ - [KMeans class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/kmeans/kmeans.hpp)
+
+## lars()
+{: #lars }
+
+#### LARS
+{: #lars_descr }
+
+```R
+R> library(mlpack)
+R> d <- lars(input=matrix(numeric(), 0, 0), input_model=NA, lambda1=0,
+        lambda2=0, no_intercept=FALSE, no_normalize=FALSE,
+        responses=matrix(numeric(), 0, 0), test=matrix(numeric(), 0, 0),
+        use_cholesky=FALSE, verbose=FALSE)
+R> output_model <- d$output_model
+R> output_predictions <- d$output_predictions
+```
+
+An implementation of Least Angle Regression (Stagewise/laSso), also known as LARS.  This can train a LARS/LASSO/Elastic Net model and use that model or a pre-trained model to output regression predictions for a test set. [Detailed documentation](#lars_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of covariates (X). | `matrix(numeric(), 0, 0)` |
+| `input_model` | [`LARS`](#doc_model) | Trained LARS model to use. | `NA` |
+| `lambda1` | [`numeric`](#doc_numeric) | Regularization parameter for l1-norm penalty. | `0` |
+| `lambda2` | [`numeric`](#doc_numeric) | Regularization parameter for l2-norm penalty. | `0` |
+| `no_intercept` | [`logical`](#doc_logical) | Do not fit an intercept in the model. | `FALSE` |
+| `no_normalize` | [`logical`](#doc_logical) | Do not normalize data to unit variance before modeling. | `FALSE` |
+| `responses` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of responses/observations (y). | `matrix(numeric(), 0, 0)` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing points to regress on (test points). | `matrix(numeric(), 0, 0)` |
+| `use_cholesky` | [`logical`](#doc_logical) | Use Cholesky decomposition during computation rather than explicitly computing the full Gram matrix. | `FALSE` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LARS`](#doc_model) | Output LARS model. | 
+| `output_predictions` | [`numeric matrix`](#doc_numeric_matrix) | If --test_file is specified, this file is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #lars_detailed-documentation }
+
+An implementation of LARS: Least Angle Regression (Stagewise/laSso).  This is a stage-wise homotopy-based algorithm for L1-regularized linear regression (LASSO) and L1+L2-regularized linear regression (Elastic Net).
+
+This program is able to train a LARS/LASSO/Elastic Net model or load a model from file, output regression predictions for a test set, and save the trained model to a file.  The LARS algorithm is described in more detail below:
+
+Let X be a matrix where each row is a point and each column is a dimension, and let y be a vector of targets.
+
+The Elastic Net problem is to solve
+
+  min_beta 0.5 \|\| X * beta - y \|\|_2^2 + lambda_1 \|\|beta\|\|_1 +
+      0.5 lambda_2 \|\|beta\|\|_2^2
+
+If lambda1 > 0 and lambda2 = 0, the problem is the LASSO.
+If lambda1 > 0 and lambda2 > 0, the problem is the Elastic Net.
+If lambda1 = 0 and lambda2 > 0, the problem is ridge regression.
+If lambda1 = 0 and lambda2 = 0, the problem is unregularized linear regression.
+
+For efficiency reasons, it is not recommended to use this algorithm with `lambda1` = 0.  In that case, use the 'linear_regression' program, which implements both unregularized linear regression and ridge regression.
+
+To train a LARS/LASSO/Elastic Net model, the `input` and `responses` parameters must be given.  The `lambda1`, `lambda2`, and `use_cholesky` parameters control the training options.  A trained model can be saved with the `output_model`.  If no training is desired at all, a model can be passed via the `input_model` parameter.
+
+The program can also provide predictions for test data using either the trained model or the given input model.  Test points can be specified with the `test` parameter.  Predicted responses to the test points can be saved with the `output_predictions` output parameter.
+
+### Example
+For example, the following command trains a model on the data `"data"` and responses `"responses"` with lambda1 set to 0.4 and lambda2 set to 0 (so, LASSO is being solved), and then the model is saved to `"lasso_model"`:
+
+```R
+R> output <- lars(input=data, responses=responses, lambda1=0.4, lambda2=0)
+R> lasso_model <- output$output_model
+```
+
+The following command uses the `"lasso_model"` to provide predicted responses for the data `"test"` and save those responses to `"test_predictions"`: 
+
+```R
+R> output <- lars(input_model=lasso_model, test=test)
+R> test_predictions <- output$output_predictions
+```
+
+### See also
+
+ - [linear_regression()](#linear_regression)
+ - [Least angle regression (pdf)](https://mlpack.org/papers/lars.pdf)
+ - [LARS C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lars/lars.hpp)
+
+## linear_svm()
+{: #linear_svm }
+
+#### Linear SVM is an L2-regularized support vector machine.
+{: #linear_svm_descr }
+
+```R
+R> library(mlpack)
+R> d <- linear_svm(delta=1, epochs=50, input_model=NA,
+        labels=matrix(integer(), 0, 0), lambda=0.0001, max_iterations=10000,
+        no_intercept=FALSE, num_classes=0, optimizer="lbfgs", seed=0,
+        shuffle=FALSE, step_size=0.01, test=matrix(numeric(), 0, 0),
+        test_labels=matrix(integer(), 0, 0), tolerance=1e-10,
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of linear SVM for multiclass classification. Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#linear_svm_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `delta` | [`numeric`](#doc_numeric) | Margin of difference between correct class and other classes. | `1` |
+| `epochs` | [`integer`](#doc_integer) | Maximum number of full epochs over dataset for psgd | `50` |
+| `input_model` | [`LinearSVMModel`](#doc_model) | Existing model (parameters). | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). | `matrix(integer(), 0, 0)` |
+| `lambda` | [`numeric`](#doc_numeric) | L2-regularization parameter for training. | `0.0001` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `no_intercept` | [`logical`](#doc_logical) | Do not add the intercept term to the model. | `FALSE` |
+| `num_classes` | [`integer`](#doc_integer) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `optimizer` | [`character`](#doc_character) | Optimizer to use for training ('lbfgs' or 'psgd'). | `"lbfgs"` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `shuffle` | [`logical`](#doc_logical) | Don't shuffle the order in which data points are visited for parallel SGD. | `FALSE` |
+| `step_size` | [`numeric`](#doc_numeric) | Step size for parallel SGD optimizer. | `0.01` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing test dataset. | `matrix(numeric(), 0, 0)` |
+| `test_labels` | [`integer vector`](#doc_integer_vector) | Matrix containing test labels. | `matrix(integer(), 0, 0)` |
+| `tolerance` | [`numeric`](#doc_numeric) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the training set (the matrix of predictors, X). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LinearSVMModel`](#doc_model) | Output for trained linear svm model. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #linear_svm_detailed-documentation }
+
+An implementation of linear SVMs that uses either L-BFGS or parallel SGD (stochastic gradient descent) to train the model.
+
+This program allows loading a linear SVM model (via the `input_model` parameter) or training a linear SVM model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained linear SVM model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate vector of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda` option, and the number of classes can be manually specified with the `num_classes`and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.Margin of difference between correct class and other classes can be specified with the `delta` option.The optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'psgd' (parallel stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the parallel SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer and the maximum number of epochs (specified with `epochs`). If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+Optionally, the model can be used to predict the labels for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing linear SVM model is given with the `input_model` parameter.  The output predictions from the linear SVM model may be saved with the `predictions` parameter.
+
+### Example
+As an example, to train a LinaerSVM on the data '`"data"`' with labels '`"labels"`' with L2 regularization of 0.1, saving the model to '`"lsvm_model"`', the following command may be used:
+
+```R
+R> output <- linear_svm(training=data, labels=labels, lambda=0.1, delta=1,
+  num_classes=0)
+R> lsvm_model <- output$output_model
+```
+
+Then, to use that model to predict classes for the dataset '`"test"`', storing the output predictions in '`"predictions"`', the following command may be used: 
+
+```R
+R> output <- linear_svm(input_model=lsvm_model, test=test)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [random_forest()](#random_forest)
+ - [logistic_regression()](#logistic_regression)
+ - [LinearSVM on Wikipedia](https://en.wikipedia.org/wiki/Support-vector_machine)
+ - [LinearSVM C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_svm/linear_svm.hpp)
+
+## lmnn()
+{: #lmnn }
+
+#### Large Margin Nearest Neighbors (LMNN)
+{: #lmnn_descr }
+
+```R
+R> library(mlpack)
+R> d <- lmnn(batch_size=50, center=FALSE, distance=matrix(numeric(), 0,
+        0), input=matrix(numeric(), 0, 0), k=1, labels=matrix(integer(), 0, 0),
+        linear_scan=FALSE, max_iterations=100000, normalize=FALSE,
+        optimizer="amsgrad", passes=50, print_accuracy=FALSE, range=1, rank=0,
+        regularization=0.5, seed=0, step_size=0.01, tolerance=1e-07,
+        verbose=FALSE)
+R> centered_data <- d$centered_data
+R> output <- d$output
+R> transformed_data <- d$transformed_data
+```
+
+An implementation of Large Margin Nearest Neighbors (LMNN), a distance learning technique.  Given a labeled dataset, this learns a transformation of the data that improves k-nearest-neighbor performance; this can be useful as a preprocessing step. [Detailed documentation](#lmnn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`integer`](#doc_integer) | Batch size for mini-batch SGD. | `50` |
+| `center` | [`logical`](#doc_logical) | Perform mean-centering on the dataset. It is useful when the centroid of the data is far from the origin. | `FALSE` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `distance` | [`numeric matrix`](#doc_numeric_matrix) | Initial distance matrix to be used as starting point | `matrix(numeric(), 0, 0)` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to run LMNN on. | `**--**` |
+| `k` | [`integer`](#doc_integer) | Number of target neighbors to use for each datapoint. | `1` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Labels for input dataset. | `matrix(integer(), 0, 0)` |
+| `linear_scan` | [`logical`](#doc_logical) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `FALSE` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations for L-BFGS (0 indicates no limit). | `100000` |
+| `normalize` | [`logical`](#doc_logical) | Use a normalized starting point for optimization. Itis useful for when points are far apart, or when SGD is returning NaN. | `FALSE` |
+| `optimizer` | [`character`](#doc_character) | Optimizer to use; 'amsgrad', 'bbsgd', 'sgd', or 'lbfgs'. | `"amsgrad"` |
+| `passes` | [`integer`](#doc_integer) | Maximum number of full passes over dataset for AMSGrad, BB_SGD and SGD. | `50` |
+| `print_accuracy` | [`logical`](#doc_logical) | Print accuracies on initial and transformed dataset | `FALSE` |
+| `range` | [`integer`](#doc_integer) | Number of iterations after which impostors needs to be recalculated | `1` |
+| `rank` | [`integer`](#doc_integer) | Rank of distance matrix to be optimized.  | `0` |
+| `regularization` | [`numeric`](#doc_numeric) | Regularization for LMNN objective function  | `0.5` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`numeric`](#doc_numeric) | Step size for AMSGrad, BB_SGD and SGD (alpha). | `0.01` |
+| `tolerance` | [`numeric`](#doc_numeric) | Maximum tolerance for termination of AMSGrad, BB_SGD, SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centered_data` | [`numeric matrix`](#doc_numeric_matrix) | Output matrix for mean-centered dataset. | 
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Output matrix for learned distance matrix. | 
+| `transformed_data` | [`numeric matrix`](#doc_numeric_matrix) | Output matrix for transformed dataset. | 
+
+### Detailed documentation
+{: #lmnn_detailed-documentation }
+
+This program implements Large Margin Nearest Neighbors, a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset.  The method employes the strategy of reducing distance between similar labeled data points (a.k.a target neighbors) and increasing distance between differently labeled points (a.k.a impostors) using standard optimization techniques over the gradient of the distance between data points.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input`), or alternatively as a separate matrix (specified with `labels`).  Additionally, a starting point for optimization (specified with `distance`can be given, having (r x d) dimensionality.  Here r should satisfy 1 <= r <= d, Consequently a Low-Rank matrix will be optimized. Alternatively, Low-Rank distance can be learned by specifying the `rank`parameter (A Low-Rank matrix with uniformly distributed values will be used as initial learning point). 
+
+The program also requires number of targets neighbors to work with ( specified with `k`), A regularization parameter can also be passed, It acts as a trade of between the pulling and pushing terms (specified with `regularization`), In addition, this implementation of LMNN includes a parameter to decide the interval after which impostors must be re-calculated (specified with `range`).
+
+Output can either be the learned distance matrix (specified with `output`), or the transformed dataset  (specified with `transformed_data`), or both. Additionally mean-centered dataset (specified with `centered_data`) can be accessed given mean-centering (specified with `center`) is performed on the dataset. Accuracy on initial dataset and final transformed dataset can be printed by specifying the `print_accuracy`parameter. 
+
+This implementation of LMNN uses AdaGrad, BigBatch_SGD, stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer. 
+
+AdaGrad, specified by the value 'adagrad' for the parameter `optimizer`, uses maximum of past squared gradients. It primarily on six parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`). Inaddition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+BigBatch_SGD, specified by the value 'bbsgd' for the parameter `optimizer`, depends primarily on four parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. 
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of passes (specified with `passes`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter. Furthermore, mean-centering can be performed on the dataset by specifying the `center`parameter. 
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `max_iterations`, `tolerance`(the optimization is terminated when the gradient norm is below this value).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.  In addition, a normalized starting point can be used by specifying the `normalize` parameter.
+
+By default, the AMSGrad optimizer is used.
+
+### Example
+Example - Let's say we want to learn distance on iris dataset with number of targets as 3 using BigBatch_SGD optimizer. A simple call for the same will look like: 
+
+```R
+R> output <- lmnn(input=iris, labels=iris_labels, k=3, optimizer="bbsgd")
+R> output <- output$output
+```
+
+An another program call making use of range & regularization parameter with dataset having labels as last column can be made as: 
+
+```R
+R> output <- lmnn(input=letter_recognition, k=5, range=10,
+  regularization=0.4)
+R> output <- output$output
+```
+
+### See also
+
+ - [nca()](#nca)
+ - [Large margin nearest neighbor on Wikipedia](https://en.wikipedia.org/wiki/Large_margin_nearest_neighbor)
+ - [Distance metric learning for large margin nearest neighbor classification (pdf)](http://papers.nips.cc/paper/2795-distance-metric-learning-for-large-margin-nearest-neighbor-classification.pdf)
+ - [LMNN C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lmnn/lmnn.hpp)
+
+## local_coordinate_coding()
+{: #local_coordinate_coding }
+
+#### Local Coordinate Coding
+{: #local_coordinate_coding_descr }
+
+```R
+R> library(mlpack)
+R> d <- local_coordinate_coding(atoms=0,
+        initial_dictionary=matrix(numeric(), 0, 0), input_model=NA, lambda=0,
+        max_iterations=0, normalize=FALSE, seed=0, test=matrix(numeric(), 0, 0),
+        tolerance=0.01, training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> codes <- d$codes
+R> dictionary <- d$dictionary
+R> output_model <- d$output_model
+```
+
+An implementation of Local Coordinate Coding (LCC), a data transformation technique.  Given input data, this transforms each point to be expressed as a linear combination of a few points in the dataset; once an LCC model is trained, it can be used to transform points later also. [Detailed documentation](#local_coordinate_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`integer`](#doc_integer) | Number of atoms in the dictionary. | `0` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `initial_dictionary` | [`numeric matrix`](#doc_numeric_matrix) | Optional initial dictionary. | `matrix(numeric(), 0, 0)` |
+| `input_model` | [`LocalCoordinateCoding`](#doc_model) | Input LCC model. | `NA` |
+| `lambda` | [`numeric`](#doc_numeric) | Weighted l1-norm regularization parameter. | `0` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations for LCC (0 indicates no limit). | `0` |
+| `normalize` | [`logical`](#doc_logical) | If set, the input data matrix will be normalized before coding. | `FALSE` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Test points to encode. | `matrix(numeric(), 0, 0)` |
+| `tolerance` | [`numeric`](#doc_numeric) | Tolerance for objective function. | `0.01` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of training data (X). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`numeric matrix`](#doc_numeric_matrix) | Output codes matrix. | 
+| `dictionary` | [`numeric matrix`](#doc_numeric_matrix) | Output dictionary matrix. | 
+| `output_model` | [`LocalCoordinateCoding`](#doc_model) | Output for trained LCC model. | 
+
+### Detailed documentation
+{: #local_coordinate_coding_detailed-documentation }
+
+An implementation of Local Coordinate Coding (LCC), which codes data that approximately lives on a manifold using a variation of l1-norm regularized sparse coding.  Given a dense data matrix X with n points and d dimensions, LCC seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a coding matrix Z with n points in k dimensions.  Because of the regularization method used, the atoms in D should lie close to the manifold on which the data points lie.
+
+The original data matrix X can then be reconstructed as D * Z.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a coding step, which updates the coding matrix Z.
+
+To run this program, the input matrix X must be specified (with -i), along with the number of atoms in the dictionary (-k).  An initial dictionary may also be specified with the `initial_dictionary` parameter.  The l1-norm regularization parameter is specified with the `lambda` parameter.
+
+### Example
+For example, to run LCC on the dataset `"data"` using 200 atoms and an l1-regularization parameter of 0.1, saving the dictionary `dictionary` and the codes into `codes`, use
+
+```R
+R> output <- local_coordinate_coding(training=data, atoms=200, lambda=0.1)
+R> dict <- output$dictionary
+R> codes <- output$codes
+```
+
+The maximum number of iterations may be specified with the `max_iterations` parameter. Optionally, the input data matrix X can be normalized before coding with the `normalize` parameter.
+
+An LCC model may be saved using the `output_model` output parameter.  Then, to encode new points from the dataset `"points"` with the previously saved model `"lcc_model"`, saving the new codes to `"new_codes"`, the following command can be used:
+
+```R
+R> output <- local_coordinate_coding(input_model=lcc_model, test=points)
+R> new_codes <- output$codes
+```
+
+### See also
+
+ - [sparse_coding()](#sparse_coding)
+ - [Nonlinear learning using local coordinate coding (pdf)](https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-coordinate-coding.pdf)
+ - [LocalCoordinateCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding.hpp)
+
+## logistic_regression()
+{: #logistic_regression }
+
+#### L2-regularized Logistic Regression and Prediction
+{: #logistic_regression_descr }
+
+```R
+R> library(mlpack)
+R> d <- logistic_regression(batch_size=64, decision_boundary=0.5,
+        input_model=NA, labels=matrix(integer(), 0, 0), lambda=0,
+        max_iterations=10000, optimizer="lbfgs", print_training_accuracy=FALSE,
+        step_size=0.01, test=matrix(numeric(), 0, 0), tolerance=1e-10,
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of L2-regularized logistic regression for two-class classification.  Given labeled data, a model can be trained and saved for future use; or, a pre-trained model can be used to classify new points. [Detailed documentation](#logistic_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `batch_size` | [`integer`](#doc_integer) | Batch size for SGD. | `64` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `decision_boundary` | [`numeric`](#doc_numeric) | Decision boundary for prediction; if the logistic function for a point is less than the boundary, the class is taken to be 0; otherwise, the class is 1. | `0.5` |
+| `input_model` | [`LogisticRegression`](#doc_model) | Existing model (parameters). | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). | `matrix(integer(), 0, 0)` |
+| `lambda` | [`numeric`](#doc_numeric) | L2-regularization parameter for training. | `0` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum iterations for optimizer (0 indicates no limit). | `10000` |
+| `optimizer` | [`character`](#doc_character) | Optimizer to use for training ('lbfgs' or 'sgd'). | `"lbfgs"` |
+| `print_training_accuracy` | [`logical`](#doc_logical) | If set, then the accuracy of the model on the training set will be printed (verbose must also be specified). | `FALSE` |
+| `step_size` | [`numeric`](#doc_numeric) | Step size for SGD optimizer. | `0.01` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing test dataset. | `matrix(numeric(), 0, 0)` |
+| `tolerance` | [`numeric`](#doc_numeric) | Convergence tolerance for optimizer. | `1e-10` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the training set (the matrix of predictors, X). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LogisticRegression`](#doc_model) | Output for trained logistic regression model. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | If test data is specified, this matrix is where the predictions for the test set will be saved. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | If test data is specified, this matrix is where the class probabilities for the test set will be saved. | 
+
+### Detailed documentation
+{: #logistic_regression_detailed-documentation }
+
+An implementation of L2-regularized logistic regression using either the L-BFGS optimizer or SGD (stochastic gradient descent).  This solves the regression problem
+
+  y = (1 / 1 + e^-(X * b)).
+
+In this setting, y corresponds to class labels and X corresponds to data.
+
+This program allows loading a logistic regression model (via the `input_model` parameter) or training a logistic regression model given training data (specified with the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (specified with the `test` parameter) and the classification results may be saved with the `predictions` output parameter. The trained logistic regression model may be saved using the `output_model` output parameter.
+
+The training data, if specified, may have class labels as its last dimension.  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+When a model is being trained, there are many options.  L2 regularization (to prevent overfitting) can be specified with the `lambda` option, and the optimizer used to train the model can be specified with the `optimizer` parameter.  Available options are 'sgd' (stochastic gradient descent) and 'lbfgs' (the L-BFGS optimizer).  There are also various parameters for the optimizer; the `max_iterations` parameter specifies the maximum number of allowed iterations, and the `tolerance` parameter specifies the tolerance for convergence.  For the SGD optimizer, the `step_size` parameter controls the step size taken at each iteration by the optimizer.  The batch size for SGD is controlled with the `batch_size` parameter. If the objective function for your data is oscillating between Inf and 0, the step size is probably too large.  There are more parameters for the optimizers, but the C++ interface must be used to access these.
+
+For SGD, an iteration refers to a single point. So to take a single pass over the dataset with SGD, `max_iterations` should be set to the number of points in the dataset.
+
+Optionally, the model can be used to predict the responses for another matrix of data points, if `test` is specified.  The `test` parameter can be specified without the `training` parameter, so long as an existing logistic regression model is given with the `input_model` parameter.  The output predictions from the logistic regression model may be saved with the `predictions` parameter.
+
+This implementation of logistic regression does not support the general multi-class case but instead only the two-class case.  Any labels must be either 0 or 1.  For more classes, see the softmax regression implementation.
+
+### Example
+As an example, to train a logistic regression model on the data '`"data"`' with labels '`"labels"`' with L2 regularization of 0.1, saving the model to '`"lr_model"`', the following command may be used:
+
+```R
+R> output <- logistic_regression(training=data, labels=labels, lambda=0.1,
+  print_training_accuracy=TRUE)
+R> lr_model <- output$output_model
+```
+
+Then, to use that model to predict classes for the dataset '`"test"`', storing the output predictions in '`"predictions"`', the following command may be used: 
+
+```R
+R> output <- logistic_regression(input_model=lr_model, test=test)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+ - [:LogisticRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/logistic_regression/logistic_regression.hpp)
+
+## lsh()
+{: #lsh }
+
+#### K-Approximate-Nearest-Neighbor Search with LSH
+{: #lsh_descr }
+
+```R
+R> library(mlpack)
+R> d <- lsh(bucket_size=500, hash_width=0, input_model=NA, k=0,
+        num_probes=0, projections=10, query=matrix(numeric(), 0, 0),
+        reference=matrix(numeric(), 0, 0), second_hash_size=99901, seed=0,
+        tables=30, true_neighbors=matrix(integer(), 0, 0), verbose=FALSE)
+R> distances <- d$distances
+R> neighbors <- d$neighbors
+R> output_model <- d$output_model
+```
+
+An implementation of approximate k-nearest-neighbor search with locality-sensitive hashing (LSH).  Given a set of reference points and a set of query points, this will compute the k approximate nearest neighbors of each query point in the reference set; models can be saved for future use. [Detailed documentation](#lsh_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `bucket_size` | [`integer`](#doc_integer) | The size of a bucket in the second level hash. | `500` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `hash_width` | [`numeric`](#doc_numeric) | The hash width for the first-level hashing in the LSH preprocessing. By default, the LSH class automatically estimates a hash width for its use. | `0` |
+| `input_model` | [`LSHSearch`](#doc_model) | Input LSH model. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of nearest neighbors to find. | `0` |
+| `num_probes` | [`integer`](#doc_integer) | Number of additional probes for multiprobe LSH; if 0, traditional LSH is used. | `0` |
+| `projections` | [`integer`](#doc_integer) | The number of hash functions for each table | `10` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing query points (optional). | `matrix(numeric(), 0, 0)` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing the reference dataset. | `matrix(numeric(), 0, 0)` |
+| `second_hash_size` | [`integer`](#doc_integer) | The size of the second level hash table. | `99901` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `tables` | [`integer`](#doc_integer) | The number of hash tables to be used. | `30` |
+| `true_neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix of true neighbors to compute recall with (the recall is printed when -v is specified). | `matrix(integer(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`LSHSearch`](#doc_model) | Output for trained LSH model. | 
+
+### Detailed documentation
+{: #lsh_detailed-documentation }
+
+This program will calculate the k approximate-nearest-neighbors of a set of points using locality-sensitive hashing. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. 
+
+### Example
+For example, the following will return 5 neighbors from the data for each point in `"input"` and store the distances in `"distances"` and the neighbors in `"neighbors"`:
+
+```R
+R> output <- lsh(k=5, reference=input)
+R> distances <- output$distances
+R> neighbors <- output$neighbors
+```
+
+The output is organized such that row i and column j in the neighbors output corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output file corresponds to the distance between those two points.
+
+Because this is approximate-nearest-neighbors search, results may be different from run to run.  Thus, the `seed` parameter can be specified to set the random seed.
+
+This program also has many other parameters to control its functionality; see the parameter-specific documentation for more information.
+
+### See also
+
+ - [knn()](#knn)
+ - [krann()](#krann)
+ - [Locality-sensitive hashing on Wikipedia](https://en.wikipedia.org/wiki/Locality-sensitive_hashing)
+ - [Locality-sensitive hashing scheme based on p-stable  distributions(pdf)](https://www.mlpack.org/papers/lsh.pdf)
+ - [LSHSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/lsh/lsh.hpp)
+
+## mean_shift()
+{: #mean_shift }
+
+#### Mean Shift Clustering
+{: #mean_shift_descr }
+
+```R
+R> library(mlpack)
+R> d <- mean_shift(force_convergence=FALSE, in_place=FALSE,
+        input=matrix(numeric(), 0, 0), labels_only=FALSE, max_iterations=1000,
+        radius=0, verbose=FALSE)
+R> centroid <- d$centroid
+R> output <- d$output
+```
+
+A fast implementation of mean-shift clustering using dual-tree range search.  Given a dataset, this uses the mean shift algorithm to produce and return a clustering of the data. [Detailed documentation](#mean_shift_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `force_convergence` | [`logical`](#doc_logical) | If specified, the mean shift algorithm will continue running regardless of max_iterations until the clusters converge. | `FALSE` |
+| `in_place` | [`logical`](#doc_logical) | If specified, a column containing the learned cluster assignments will be added to the input dataset file.  In this case, --output_file is overridden.  (Do not use with Python.) | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform clustering on. | `**--**` |
+| `labels_only` | [`logical`](#doc_logical) | If specified, only the output labels will be written to the file specified by --output_file. | `FALSE` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations before mean shift terminates. | `1000` |
+| `radius` | [`numeric`](#doc_numeric) | If the distance between two centroids is less than the given radius, one will be removed.  A radius of 0 or less means an estimate will be calculated and used for the radius. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `centroid` | [`numeric matrix`](#doc_numeric_matrix) | If specified, the centroids of each cluster will be written to the given matrix. | 
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to write output labels or labeled data to. | 
+
+### Detailed documentation
+{: #mean_shift_detailed-documentation }
+
+This program performs mean shift clustering on the given dataset, storing the learned cluster assignments either as a column of labels in the input dataset or separately.
+
+The input dataset should be specified with the `input` parameter, and the radius used for search can be specified with the `radius` parameter.  The maximum number of iterations before algorithm termination is controlled with the `max_iterations` parameter.
+
+The output labels may be saved with the `output` output parameter and the centroids of each cluster may be saved with the `centroid` output parameter.
+
+### Example
+For example, to run mean shift clustering on the dataset `"data"` and store the centroids to `"centroids"`, the following command may be used: 
+
+```R
+R> output <- mean_shift(input=data)
+R> centroids <- output$centroid
+```
+
+### See also
+
+ - [kmeans()](#kmeans)
+ - [dbscan()](#dbscan)
+ - [Mean shift on Wikipedia](https://en.wikipedia.org/wiki/Mean_shift)
+ - [Mean Shift, Mode Seeking, and Clustering (pdf)](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=1c168275c59ba382588350ee1443537f59978183)
+ - [mlpack::mean_shift::MeanShift C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/mean_shift/mean_shift.hpp)
+
+## nbc()
+{: #nbc }
+
+#### Parametric Naive Bayes Classifier
+{: #nbc_descr }
+
+```R
+R> library(mlpack)
+R> d <- nbc(incremental_variance=FALSE, input_model=NA,
+        labels=matrix(integer(), 0, 0), test=matrix(numeric(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output <- d$output
+R> output_model <- d$output_model
+R> output_probs <- d$output_probs
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of the Naive Bayes Classifier, used for classification. Given labeled data, an NBC model can be trained and saved, or, a pre-trained model can be used for classification. [Detailed documentation](#nbc_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `incremental_variance` | [`logical`](#doc_logical) | The variance of each class will be calculated incrementally. | `FALSE` |
+| `input_model` | [`NBCModel`](#doc_model) | Input Naive Bayes model. | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | A file containing labels for the training set. | `matrix(integer(), 0, 0)` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the test set. | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the training set. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`integer vector`](#doc_integer_vector) | The matrix in which the predicted labels for the test set will be written (deprecated). | 
+| `output_model` | [`NBCModel`](#doc_model) | File to save trained Naive Bayes model to. | 
+| `output_probs` | [`numeric matrix`](#doc_numeric_matrix) | The matrix in which the predicted probability of labels for the test set will be written (deprecated). | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | The matrix in which the predicted labels for the test set will be written. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | The matrix in which the predicted probability of labels for the test set will be written. | 
+
+### Detailed documentation
+{: #nbc_detailed-documentation }
+
+This program trains the Naive Bayes classifier on the given labeled training set, or loads a model from the given model file, and then may use that trained model to classify the points in a given test set.
+
+The training set is specified with the `training` parameter.  Labels may be either the last row of the training set, or alternately the `labels` parameter may be specified to pass a separate matrix of labels.
+
+If training is not desired, a pre-existing model may be loaded with the `input_model` parameter.
+
+
+
+The `incremental_variance` parameter can be used to force the training to use an incremental algorithm for calculating variance.  This is slower, but can help avoid loss of precision in some cases.
+
+If classifying a test set is desired, the test set may be specified with the `test` parameter, and the classifications may be saved with the `predictions`predictions  parameter.  If saving the trained model is desired, this may be done with the `output_model` output parameter.
+
+Note: the `output` and `output_probs` parameters are deprecated and will be removed in mlpack 4.0.0.  Use `predictions` and `probabilities` instead.
+
+### Example
+For example, to train a Naive Bayes classifier on the dataset `"data"` with labels `"labels"` and save the model to `"nbc_model"`, the following command may be used:
+
+```R
+R> output <- nbc(training=data, labels=labels)
+R> nbc_model <- output$output_model
+```
+
+Then, to use `"nbc_model"` to predict the classes of the dataset `"test_set"` and save the predicted classes to `"predictions"`, the following command may be used:
+
+```R
+R> output <- nbc(input_model=nbc_model, test=test_set)
+R> predictions <- output$output
+```
+
+### See also
+
+ - [softmax_regression()](#softmax_regression)
+ - [random_forest()](#random_forest)
+ - [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+ - [NaiveBayesClassifier C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp)
+
+## nca()
+{: #nca }
+
+#### Neighborhood Components Analysis (NCA)
+{: #nca_descr }
+
+```R
+R> library(mlpack)
+R> d <- nca(armijo_constant=0.0001, batch_size=50,
+        input=matrix(numeric(), 0, 0), labels=matrix(integer(), 0, 0),
+        linear_scan=FALSE, max_iterations=500000, max_line_search_trials=50,
+        max_step=1e+20, min_step=1e-20, normalize=FALSE, num_basis=5,
+        optimizer="sgd", seed=0, step_size=0.01, tolerance=1e-07, verbose=FALSE,
+        wolfe=0.9)
+R> output <- d$output
+```
+
+An implementation of neighborhood components analysis, a distance learning technique that can be used for preprocessing.  Given a labeled dataset, this uses NCA, which seeks to improve the k-nearest-neighbor classification, and returns the learned distance metric. [Detailed documentation](#nca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `armijo_constant` | [`numeric`](#doc_numeric) | Armijo constant for L-BFGS. | `0.0001` |
+| `batch_size` | [`integer`](#doc_integer) | Batch size for mini-batch SGD. | `50` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to run NCA on. | `**--**` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Labels for input dataset. | `matrix(integer(), 0, 0)` |
+| `linear_scan` | [`logical`](#doc_logical) | Don't shuffle the order in which data points are visited for SGD or mini-batch SGD. | `FALSE` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations for SGD or L-BFGS (0 indicates no limit). | `500000` |
+| `max_line_search_trials` | [`integer`](#doc_integer) | Maximum number of line search trials for L-BFGS. | `50` |
+| `max_step` | [`numeric`](#doc_numeric) | Maximum step of line search for L-BFGS. | `1e+20` |
+| `min_step` | [`numeric`](#doc_numeric) | Minimum step of line search for L-BFGS. | `1e-20` |
+| `normalize` | [`logical`](#doc_logical) | Use a normalized starting point for optimization. This is useful for when points are far apart, or when SGD is returning NaN. | `FALSE` |
+| `num_basis` | [`integer`](#doc_integer) | Number of memory points to be stored for L-BFGS. | `5` |
+| `optimizer` | [`character`](#doc_character) | Optimizer to use; 'sgd' or 'lbfgs'. | `"sgd"` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `step_size` | [`numeric`](#doc_numeric) | Step size for stochastic gradient descent (alpha). | `0.01` |
+| `tolerance` | [`numeric`](#doc_numeric) | Maximum tolerance for termination of SGD or L-BFGS. | `1e-07` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `wolfe` | [`numeric`](#doc_numeric) | Wolfe condition parameter for L-BFGS. | `0.9` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Output matrix for learned distance matrix. | 
+
+### Detailed documentation
+{: #nca_detailed-documentation }
+
+This program implements Neighborhood Components Analysis, both a linear dimensionality reduction technique and a distance learning technique.  The method seeks to improve k-nearest-neighbor classification on a dataset by scaling the dimensions.  The method is nonparametric, and does not require a value of k.  It works by using stochastic ("soft") neighbor assignments and using optimization techniques over the gradient of the accuracy of the neighbor assignments.
+
+To work, this algorithm needs labeled data.  It can be given as the last row of the input dataset (specified with `input`), or alternatively as a separate matrix (specified with `labels`).
+
+This implementation of NCA uses stochastic gradient descent, mini-batch stochastic gradient descent, or the L_BFGS optimizer.  These optimizers do not guarantee global convergence for a nonconvex objective function (NCA's objective function is nonconvex), so the final results could depend on the random seed or other optimizer parameters.
+
+Stochastic gradient descent, specified by the value 'sgd' for the parameter `optimizer`, depends primarily on three parameters: the step size (specified with `step_size`), the batch size (specified with `batch_size`), and the maximum number of iterations (specified with `max_iterations`).  In addition, a normalized starting point can be used by specifying the `normalize` parameter, which is necessary if many warnings of the form 'Denominator of p_i is 0!' are given.  Tuning the step size can be a tedious affair.  In general, the step size is too large if the objective is not mostly uniformly decreasing, or if zero-valued denominator warnings are being issued.  The step size is too small if the objective is changing very slowly.  Setting the termination condition can be done easily once a good step size parameter is found; either increase the maximum iterations to a large number and allow SGD to find a minimum, or set the maximum iterations to 0 (allowing infinite iterations) and set the tolerance (specified by `tolerance`) to define the maximum allowed difference between objectives for SGD to terminate.  Be careful---setting the tolerance instead of the maximum iterations can take a very long time and may actually never converge due to the properties of the SGD optimizer. Note that a single iteration of SGD refers to a single point, so to take a single pass over the dataset, set the value of the `max_iterations` parameter equal to the number of points in the dataset.
+
+The L-BFGS optimizer, specified by the value 'lbfgs' for the parameter `optimizer`, uses a back-tracking line search algorithm to minimize a function.  The following parameters are used by L-BFGS: `num_basis` (specifies the number of memory points used by L-BFGS), `max_iterations`, `armijo_constant`, `wolfe`, `tolerance` (the optimization is terminated when the gradient norm is below this value), `max_line_search_trials`, `min_step`, and `max_step` (which both refer to the line search routine).  For more details on the L-BFGS optimizer, consult either the mlpack L-BFGS documentation (in lbfgs.hpp) or the vast set of published literature on L-BFGS.
+
+By default, the SGD optimizer is used.
+
+### See also
+
+ - [lmnn()](#lmnn)
+ - [Neighbourhood components analysis on Wikipedia](https://en.wikipedia.org/wiki/Neighbourhood_components_analysis)
+ - [Neighbourhood components analysis (pdf)](http://papers.nips.cc/paper/2566-neighbourhood-components-analysis.pdf)
+ - [NCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/nca/nca.hpp)
+
+## knn()
+{: #knn }
+
+#### k-Nearest-Neighbors Search
+{: #knn_descr }
+
+```R
+R> library(mlpack)
+R> d <- knn(algorithm="dual_tree", epsilon=0, input_model=NA, k=0,
+        leaf_size=20, query=matrix(numeric(), 0, 0), random_basis=FALSE,
+        reference=matrix(numeric(), 0, 0), rho=0.7, seed=0, tau=0,
+        tree_type="kd", true_distances=matrix(numeric(), 0, 0),
+        true_neighbors=matrix(integer(), 0, 0), verbose=FALSE)
+R> distances <- d$distances
+R> neighbors <- d$neighbors
+R> output_model <- d$output_model
+```
+
+An implementation of k-nearest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#knn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`character`](#doc_character) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `epsilon` | [`numeric`](#doc_numeric) | If specified, will do approximate nearest neighbor search with given relative error. | `0` |
+| `input_model` | [`KNNModel`](#doc_model) | Pre-trained kNN model. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`integer`](#doc_integer) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, spill trees, and octrees). | `20` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing query points (optional). | `matrix(numeric(), 0, 0)` |
+| `random_basis` | [`logical`](#doc_logical) | Before tree-building, project the data onto a random orthogonal basis. | `FALSE` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing the reference dataset. | `matrix(numeric(), 0, 0)` |
+| `rho` | [`numeric`](#doc_numeric) | Balance threshold (only valid for spill trees). | `0.7` |
+| `seed` | [`integer`](#doc_integer) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tau` | [`numeric`](#doc_numeric) | Overlapping size (only valid for spill trees). | `0` |
+| `tree_type` | [`character`](#doc_character) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'spill', 'oct'. | `"kd"` |
+| `true_distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `matrix(numeric(), 0, 0)` |
+| `true_neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `matrix(integer(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`KNNModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #knn_detailed-documentation }
+
+This program will calculate the k-nearest-neighbors of a set of points using kd-trees or cover trees (cover tree support is experimental and may be slow). You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following command will calculate the 5 nearest neighbors of each point in `"input"` and store the distances in `"distances"` and the neighbors in `"neighbors"`: 
+
+```R
+R> output <- knn(k=5, reference=input)
+R> neighbors <- output$neighbors
+R> distances <- output$distances
+```
+
+The output is organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th nearest neighbor from the point in the query set with index i.  Row j and column i in the distances output matrix corresponds to the distance between those two points.
+
+### See also
+
+ - [lsh()](#lsh)
+ - [krann()](#krann)
+ - [kfn()](#kfn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## kfn()
+{: #kfn }
+
+#### k-Furthest-Neighbors Search
+{: #kfn_descr }
+
+```R
+R> library(mlpack)
+R> d <- kfn(algorithm="dual_tree", epsilon=0, input_model=NA, k=0,
+        leaf_size=20, percentage=1, query=matrix(numeric(), 0, 0),
+        random_basis=FALSE, reference=matrix(numeric(), 0, 0), seed=0,
+        tree_type="kd", true_distances=matrix(numeric(), 0, 0),
+        true_neighbors=matrix(integer(), 0, 0), verbose=FALSE)
+R> distances <- d$distances
+R> neighbors <- d$neighbors
+R> output_model <- d$output_model
+```
+
+An implementation of k-furthest-neighbor search using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k furthest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#kfn_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `algorithm` | [`character`](#doc_character) | Type of neighbor search: 'naive', 'single_tree', 'dual_tree', 'greedy'. | `"dual_tree"` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `epsilon` | [`numeric`](#doc_numeric) | If specified, will do approximate furthest neighbor search with given relative error. Must be in the range [0,1). | `0` |
+| `input_model` | [`KFNModel`](#doc_model) | Pre-trained kFN model. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of furthest neighbors to find. | `0` |
+| `leaf_size` | [`integer`](#doc_integer) | Leaf size for tree building (used for kd-trees, vp trees, random projection trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `percentage` | [`numeric`](#doc_numeric) | If specified, will do approximate furthest neighbor search. Must be in the range (0,1] (decimal form). Resultant neighbors will be at least (p*100) % of the distance as the true furthest neighbor. | `1` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing query points (optional). | `matrix(numeric(), 0, 0)` |
+| `random_basis` | [`logical`](#doc_logical) | Before tree-building, project the data onto a random orthogonal basis. | `FALSE` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing the reference dataset. | `matrix(numeric(), 0, 0)` |
+| `seed` | [`integer`](#doc_integer) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `tree_type` | [`character`](#doc_character) | Type of tree to use: 'kd', 'vp', 'rp', 'max-rp', 'ub', 'cover', 'r', 'r-star', 'x', 'ball', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `true_distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of true distances to compute the effective error (average relative error) (it is printed when -v is specified). | `matrix(numeric(), 0, 0)` |
+| `true_neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix of true neighbors to compute the recall (it is printed when -v is specified). | `matrix(integer(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`KFNModel`](#doc_model) | If specified, the kFN model will be output here. | 
+
+### Detailed documentation
+{: #kfn_detailed-documentation }
+
+This program will calculate the k-furthest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set.
+
+### Example
+For example, the following will calculate the 5 furthest neighbors of eachpoint in `"input"` and store the distances in `"distances"` and the neighbors in `"neighbors"`: 
+
+```R
+R> output <- kfn(k=5, reference=input)
+R> distances <- output$distances
+R> neighbors <- output$neighbors
+```
+
+The output files are organized such that row i and column j in the neighbors output matrix corresponds to the index of the point in the reference set which is the j'th furthest neighbor from the point in the query set with index i.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [approx_kfn()](#approx_kfn)
+ - [knn()](#knn)
+ - [Tree-independent dual-tree algorithms (pdf)](http://proceedings.mlr.press/v28/curtin13.pdf)
+ - [NeighborSearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+
+## nmf()
+{: #nmf }
+
+#### Non-negative Matrix Factorization
+{: #nmf_descr }
+
+```R
+R> library(mlpack)
+R> d <- nmf(initial_h=matrix(numeric(), 0, 0),
+        initial_w=matrix(numeric(), 0, 0), input=matrix(numeric(), 0, 0),
+        max_iterations=10000, min_residue=1e-05, rank=0, seed=0,
+        update_rules="multdist", verbose=FALSE)
+R> h <- d$h
+R> w <- d$w
+```
+
+An implementation of non-negative matrix factorization.  This can be used to decompose an input dataset into two low-rank non-negative components. [Detailed documentation](#nmf_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `initial_h` | [`numeric matrix`](#doc_numeric_matrix) | Initial H matrix. | `matrix(numeric(), 0, 0)` |
+| `initial_w` | [`numeric matrix`](#doc_numeric_matrix) | Initial W matrix. | `matrix(numeric(), 0, 0)` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform NMF on. | `**--**` |
+| `max_iterations` | [`integer`](#doc_integer) | Number of iterations before NMF terminates (0 runs until convergence. | `10000` |
+| `min_residue` | [`numeric`](#doc_numeric) | The minimum root mean square residue allowed for each iteration, below which the program terminates. | `1e-05` |
+| `rank` | [`integer`](#doc_integer) | Rank of the factorization. | `**--**` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `update_rules` | [`character`](#doc_character) | Update rules for each iteration; ( multdist \| multdiv \| als ). | `"multdist"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `h` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save the calculated H to. | 
+| `w` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save the calculated W to. | 
+
+### Detailed documentation
+{: #nmf_detailed-documentation }
+
+This program performs non-negative matrix factorization on the given dataset, storing the resulting decomposed matrices in the specified files.  For an input dataset V, NMF decomposes V into two matrices W and H such that 
+
+V = W * H
+
+where all elements in W and H are non-negative.  If V is of size (n x m), then W will be of size (n x r) and H will be of size (r x m), where r is the rank of the factorization (specified by the `rank` parameter).
+
+Optionally, the desired update rules for each NMF iteration can be chosen from the following list:
+
+ - multdist: multiplicative distance-based update rules (Lee and Seung 1999)
+ - multdiv: multiplicative divergence-based update rules (Lee and Seung 1999)
+ - als: alternating least squares update rules (Paatero and Tapper 1994)
+
+The maximum number of iterations is specified with `max_iterations`, and the minimum residue required for algorithm termination is specified with the `min_residue` parameter.
+
+### Example
+For example, to run NMF on the input matrix `"V"` using the 'multdist' update rules with a rank-10 decomposition and storing the decomposed matrices into `"W"` and `"H"`, the following command could be used: 
+
+```R
+R> output <- nmf(input=V, rank=10, update_rules="multdist")
+R> W <- output$w
+R> H <- output$h
+```
+
+### See also
+
+ - [cf()](#cf)
+ - [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ - [Algorithms for non-negative matrix factorization (pdf)](http://papers.nips.cc/paper/1861-algorithms-for-non-negative-matrix-factorization.pdf)
+ - [AMF C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/amf/amf.hpp)
+
+## pca()
+{: #pca }
+
+#### Principal Components Analysis
+{: #pca_descr }
+
+```R
+R> library(mlpack)
+R> d <- pca(decomposition_method="exact", input=matrix(numeric(), 0, 0),
+        new_dimensionality=0, scale=FALSE, var_to_retain=0, verbose=FALSE)
+R> output <- d$output
+```
+
+An implementation of several strategies for principal components analysis (PCA), a common preprocessing step.  Given a dataset and a desired new dimensionality, this can reduce the dimensionality of the data using the linear transformation determined by PCA. [Detailed documentation](#pca_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `decomposition_method` | [`character`](#doc_character) | Method used for the principal components analysis: 'exact', 'randomized', 'randomized-block-krylov', 'quic'. | `"exact"` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset to perform PCA on. | `**--**` |
+| `new_dimensionality` | [`integer`](#doc_integer) | Desired dimensionality of output dataset. If 0, no dimensionality reduction is performed. | `0` |
+| `scale` | [`logical`](#doc_logical) | If set, the data will be scaled before running PCA, such that the variance of each feature is 1. | `FALSE` |
+| `var_to_retain` | [`numeric`](#doc_numeric) | Amount of variance to retain; should be between 0 and 1.  If 1, all variance is retained.  Overrides -d. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save modified dataset to. | 
+
+### Detailed documentation
+{: #pca_detailed-documentation }
+
+This program performs principal components analysis on the given dataset using the exact, randomized, randomized block Krylov, or QUIC SVD method. It will transform the data onto its principal components, optionally performing dimensionality reduction by ignoring the principal components with the smallest eigenvalues.
+
+Use the `input` parameter to specify the dataset to perform PCA on.  A desired new dimensionality can be specified with the `new_dimensionality` parameter, or the desired variance to retain can be specified with the `var_to_retain` parameter.  If desired, the dataset can be scaled before running PCA with the `scale` parameter.
+
+Multiple different decomposition techniques can be used.  The method to use can be specified with the `decomposition_method` parameter, and it may take the values 'exact', 'randomized', or 'quic'.
+
+### Example
+For example, to reduce the dimensionality of the matrix `"data"` to 5 dimensions using randomized SVD for the decomposition, storing the output matrix to `"data_mod"`, the following command can be used:
+
+```R
+R> output <- pca(input=data, new_dimensionality=5,
+  decomposition_method="randomized")
+R> data_mod <- output$output
+```
+
+### See also
+
+ - [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+ - [PCA C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/pca/pca.hpp)
+
+## perceptron()
+{: #perceptron }
+
+#### Perceptron
+{: #perceptron_descr }
+
+```R
+R> library(mlpack)
+R> d <- perceptron(input_model=NA, labels=matrix(integer(), 0, 0),
+        max_iterations=1000, test=matrix(numeric(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output <- d$output
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+```
+
+An implementation of a perceptron---a single level neural network--=for classification.  Given labeled data, a perceptron can be trained and saved for future use; or, a pre-trained perceptron can be used for classification on new points. [Detailed documentation](#perceptron_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`PerceptronModel`](#doc_model) | Input perceptron model. | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | A matrix containing labels for the training set. | `matrix(integer(), 0, 0)` |
+| `max_iterations` | [`integer`](#doc_integer) | The maximum number of iterations the perceptron is to be run | `1000` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the test set. | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the training set. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`integer vector`](#doc_integer_vector) | The matrix in which the predicted labels for the test set will be written. | 
+| `output_model` | [`PerceptronModel`](#doc_model) | Output for trained perceptron model. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | The matrix in which the predicted labels for the test set will be written. | 
+
+### Detailed documentation
+{: #perceptron_detailed-documentation }
+
+This program implements a perceptron, which is a single level neural network. The perceptron makes its predictions based on a linear predictor function combining a set of weights with the feature vector.  The perceptron learning rule is able to converge, given enough iterations (specified using the `max_iterations` parameter), if the data supplied is linearly separable.  The perceptron is parameterized by a matrix of weight vectors that denote the numerical weights of the neural network.
+
+This program allows loading a perceptron from a model (via the `input_model` parameter) or training a perceptron given training data (via the `training` parameter), or both those things at once.  In addition, this program allows classification on a test dataset (via the `test` parameter) and the classification results on the test set may be saved with the `predictions` output parameter.  The perceptron model may be saved with the `output_model` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `output`.
+Use `predictions` instead of `output`.
+
+### Example
+The training data given with the `training` option may have class labels as its last dimension (so, if the training data is in CSV format, labels should be the last column).  Alternately, the `labels` parameter may be used to specify a separate matrix of labels.
+
+All these options make it easy to train a perceptron, and then re-use that perceptron for later classification.  The invocation below trains a perceptron on `"training_data"` with labels `"training_labels"`, and saves the model to `"perceptron_model"`.
+
+```R
+R> output <- perceptron(training=training_data, labels=training_labels)
+R> perceptron_model <- output$output_model
+```
+
+Then, this model can be re-used for classification on the test data `"test_data"`.  The example below does precisely that, saving the predicted classes to `"predictions"`.
+
+```R
+R> output <- perceptron(input_model=perceptron_model, test=test_data)
+R> predictions <- output$predictions
+```
+
+Note that all of the options may be specified at once: predictions may be calculated right after training a model, and model training can occur even if an existing perceptron model is passed with the `input_model` parameter.  However, note that the number of classes and the dimensionality of all data must match.  So you cannot pass a perceptron model trained on 2 classes and then re-train with a 4-class dataset.  Similarly, attempting classification on a 3-dimensional dataset with a perceptron that has been trained on 8 dimensions will cause an error.
+
+### See also
+
+ - [adaboost()](#adaboost)
+ - [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+ - [Perceptron C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/perceptron/perceptron.hpp)
+
+## preprocess_split()
+{: #preprocess_split }
+
+#### Split Data
+{: #preprocess_split_descr }
+
+```R
+R> library(mlpack)
+R> d <- preprocess_split(input=matrix(numeric(), 0, 0),
+        input_labels=matrix(integer(), 0, 0), no_shuffle=FALSE, seed=0,
+        stratify_data=FALSE, test_ratio=0.2, verbose=FALSE)
+R> test <- d$test
+R> test_labels <- d$test_labels
+R> training <- d$training
+R> training_labels <- d$training_labels
+```
+
+A utility to split data into a training and testing dataset.  This can also split labels according to the same split. [Detailed documentation](#preprocess_split_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing data. | `**--**` |
+| `input_labels` | [`integer matrix`](#doc_integer_matrix) | Matrix containing labels. | `matrix(integer(), 0, 0)` |
+| `no_shuffle` | [`logical`](#doc_logical) | Avoid shuffling the data before splitting. | `FALSE` |
+| `seed` | [`integer`](#doc_integer) | Random seed (0 for std::time(NULL)). | `0` |
+| `stratify_data` | [`logical`](#doc_logical) | Stratify the data according to labels | `FALSE` |
+| `test_ratio` | [`numeric`](#doc_numeric) | Ratio of test set; if not set,the ratio defaults to 0.2 | `0.2` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save test data to. | 
+| `test_labels` | [`integer matrix`](#doc_integer_matrix) | Matrix to save test labels to. | 
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save training data to. | 
+| `training_labels` | [`integer matrix`](#doc_integer_matrix) | Matrix to save train labels to. | 
+
+### Detailed documentation
+{: #preprocess_split_detailed-documentation }
+
+This utility takes a dataset and optionally labels and splits them into a training set and a test set. Before the split, the points in the dataset are randomly reordered. The percentage of the dataset to be used as the test set can be specified with the `test_ratio` parameter; the default is 0.2 (20%).
+
+The output training and test matrices may be saved with the `training` and `test` output parameters.
+
+Optionally, labels can also be split along with the data by specifying the `input_labels` parameter.  Splitting labels works the same way as splitting the data. The output training and test labels may be saved with the `training_labels` and `test_labels` output parameters, respectively.
+
+### Example
+So, a simple example where we want to split the dataset `"X"` into `"X_train"` and `"X_test"` with 60% of the data in the training set and 40% of the dataset in the test set, we could run 
+
+```R
+R> output <- preprocess_split(input=X, test_ratio=0.4)
+R> X_train <- output$training
+R> X_test <- output$test
+```
+
+Also by default the dataset is shuffled and split; you can provide the `no_shuffle` option to avoid shuffling the data; an example to avoid shuffling of data is:
+
+```R
+R> output <- preprocess_split(input=X, test_ratio=0.4, no_shuffle=TRUE)
+R> X_train <- output$training
+R> X_test <- output$test
+```
+
+If we had a dataset `"X"` and associated labels `"y"`, and we wanted to split these into `"X_train"`, `"y_train"`, `"X_test"`, and `"y_test"`, with 30% of the data in the test set, we could run
+
+```R
+R> output <- preprocess_split(input=X, input_labels=y, test_ratio=0.3)
+R> X_train <- output$training
+R> y_train <- output$training_labels
+R> X_test <- output$test
+R> y_test <- output$test_labels
+```
+
+To maintain the ratio of each class in the train and test sets, the`stratify_data` option can be used.
+
+```R
+R> output <- preprocess_split(input=X, test_ratio=0.4, stratify_data=TRUE)
+R> X_train <- output$training
+R> X_test <- output$test
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_binarize()
+{: #preprocess_binarize }
+
+#### Binarize Data
+{: #preprocess_binarize_descr }
+
+```R
+R> library(mlpack)
+R> d <- preprocess_binarize(dimension=0, input=matrix(numeric(), 0, 0),
+        threshold=0, verbose=FALSE)
+R> output <- d$output
+```
+
+A utility to binarize a dataset.  Given a dataset, this utility converts each value in the desired dimension(s) to 0 or 1; this can be a useful preprocessing step. [Detailed documentation](#preprocess_binarize_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `dimension` | [`integer`](#doc_integer) | Dimension to apply the binarization. If not set, the program will binarize every dimension by default. | `0` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input data matrix. | `**--**` |
+| `threshold` | [`numeric`](#doc_numeric) | Threshold to be applied for binarization. If not set, the threshold defaults to 0.0. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix in which to save the output. | 
+
+### Detailed documentation
+{: #preprocess_binarize_detailed-documentation }
+
+This utility takes a dataset and binarizes the variables into either 0 or 1 given threshold. User can apply binarization on a dimension or the whole dataset.  The dimension to apply binarization to can be specified using the `dimension` parameter; if left unspecified, every dimension will be binarized.  The threshold for binarization can also be specified with the `threshold` parameter; the default threshold is 0.0.
+
+The binarized matrix may be saved with the `output` output parameter.
+
+### Example
+For example, if we want to set all variables greater than 5 in the dataset `"X"` to 1 and variables less than or equal to 5.0 to 0, and save the result to `"Y"`, we could run
+
+```R
+R> output <- preprocess_binarize(input=X, threshold=5)
+R> Y <- output$output
+```
+
+But if we want to apply this to only the first (0th) dimension of `"X"`,  we could instead run
+
+```R
+R> output <- preprocess_binarize(input=X, threshold=5, dimension=0)
+R> Y <- output$output
+```
+
+### See also
+
+ - [preprocess_describe()](#preprocess_describe)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_describe()
+{: #preprocess_describe }
+
+#### Descriptive Statistics
+{: #preprocess_describe_descr }
+
+```R
+R> library(mlpack)
+R> preprocess_describe(dimension=0, input=matrix(numeric(), 0, 0),
+        population=FALSE, precision=4, row_major=FALSE, verbose=FALSE, width=8)
+```
+
+A utility for printing descriptive statistics about a dataset.  This prints a number of details about a dataset in a tabular format. [Detailed documentation](#preprocess_describe_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `dimension` | [`integer`](#doc_integer) | Dimension of the data. Use this to specify a dimension | `0` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing data, | `**--**` |
+| `population` | [`logical`](#doc_logical) | If specified, the program will calculate statistics assuming the dataset is the population. By default, the program will assume the dataset as a sample. | `FALSE` |
+| `precision` | [`integer`](#doc_integer) | Precision of the output statistics. | `4` |
+| `row_major` | [`logical`](#doc_logical) | If specified, the program will calculate statistics across rows, not across columns.  (Remember that in mlpack, a column represents a point, so this option is generally not necessary.) | `FALSE` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `width` | [`integer`](#doc_integer) | Width of the output table. | `8` |
+
+
+### Detailed documentation
+{: #preprocess_describe_detailed-documentation }
+
+This utility takes a dataset and prints out the descriptive statistics of the data. Descriptive statistics is the discipline of quantitatively describing the main features of a collection of information, or the quantitative description itself. The program does not modify the original file, but instead prints out the statistics to the console. The printed result will look like a table.
+
+Optionally, width and precision of the output can be adjusted by a user using the `width` and `precision` parameters. A user can also select a specific dimension to analyze if there are too many dimensions. The `population` parameter can be specified when the dataset should be considered as a population.  Otherwise, the dataset will be considered as a sample.
+
+### Example
+So, a simple example where we want to print out statistical facts about the dataset `"X"` using the default settings, we could run 
+
+```R
+R> preprocess_describe(input=X, verbose=TRUE)
+```
+
+If we want to customize the width to 10 and precision to 5 and consider the dataset as a population, we could run
+
+```R
+R> preprocess_describe(input=X, width=10, precision=5, verbose=TRUE)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_split()](#preprocess_split)
+
+## preprocess_scale()
+{: #preprocess_scale }
+
+#### Scale Data
+{: #preprocess_scale_descr }
+
+```R
+R> library(mlpack)
+R> d <- preprocess_scale(epsilon=1e-06, input=matrix(numeric(), 0, 0),
+        input_model=NA, inverse_scaling=FALSE, max_value=1, min_value=0,
+        scaler_method="standard_scaler", seed=0, verbose=FALSE)
+R> output <- d$output
+R> output_model <- d$output_model
+```
+
+A utility to perform feature scaling on datasets using one of sixtechniques.  Both scaling and inverse scaling are supported, andscalers can be saved and then applied to other datasets. [Detailed documentation](#preprocess_scale_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `epsilon` | [`numeric`](#doc_numeric) | regularization Parameter for pcawhitening, or zcawhitening, should be between -1 to 1. | `1e-06` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing data. | `**--**` |
+| `input_model` | [`ScalingModel`](#doc_model) | Input Scaling model. | `NA` |
+| `inverse_scaling` | [`logical`](#doc_logical) | Inverse Scaling to get original dataset | `FALSE` |
+| `max_value` | [`integer`](#doc_integer) | Ending value of range for min_max_scaler. | `1` |
+| `min_value` | [`integer`](#doc_integer) | Starting value of range for min_max_scaler. | `0` |
+| `scaler_method` | [`character`](#doc_character) | method to use for scaling, the default is standard_scaler. | `"standard_scaler"` |
+| `seed` | [`integer`](#doc_integer) | Random seed (0 for std::time(NULL)). | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save scaled data to. | 
+| `output_model` | [`ScalingModel`](#doc_model) | Output scaling model. | 
+
+### Detailed documentation
+{: #preprocess_scale_detailed-documentation }
+
+This utility takes a dataset and performs feature scaling using one of the six scaler methods namely: 'max_abs_scaler', 'mean_normalization', 'min_max_scaler' ,'standard_scaler', 'pca_whitening' and 'zca_whitening'. The function takes a matrix as `input` and a scaling method type which you can specify using `scaler_method` parameter; the default is standard scaler, and outputs a matrix with scaled feature.
+
+The output scaled feature matrix may be saved with the `output` output parameters.
+
+The model to scale features can be saved using `output_model` and later can be loaded back using`input_model`.
+
+### Example
+So, a simple example where we want to scale the dataset `"X"` into `"X_scaled"` with  standard_scaler as scaler_method, we could run 
+
+```R
+R> output <- preprocess_scale(input=X, scaler_method="standard_scaler")
+R> X_scaled <- output$output
+```
+
+A simple example where we want to whiten the dataset `"X"` into `"X_whitened"` with  PCA as whitening_method and use 0.01 as regularization parameter, we could run 
+
+```R
+R> output <- preprocess_scale(input=X, scaler_method="pca_whitening",
+  epsilon=0.01)
+R> X_scaled <- output$output
+```
+
+You can also retransform the scaled dataset back using`inverse_scaling`. An example to rescale : `"X_scaled"` into `"X"`using the saved model `input_model` is:
+
+```R
+R> output <- preprocess_scale(input=X_scaled, inverse_scaling=TRUE,
+  input_model=saved)
+R> X <- output$output
+```
+
+Another simple example where we want to scale the dataset `"X"` into `"X_scaled"` with  min_max_scaler as scaler method, where scaling range is 1 to 3 instead of default 0 to 1. We could run 
+
+```R
+R> output <- preprocess_scale(input=X, scaler_method="min_max_scaler",
+  min_value=1, max_value=3)
+R> X_scaled <- output$output
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
+## preprocess_one_hot_encoding()
+{: #preprocess_one_hot_encoding }
+
+#### One Hot Encoding
+{: #preprocess_one_hot_encoding_descr }
+
+```R
+R> library(mlpack)
+R> d <- preprocess_one_hot_encoding(dimensions=c(),
+        input=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output <- d$output
+```
+
+A utility to do one-hot encoding on features of dataset. [Detailed documentation](#preprocess_one_hot_encoding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `dimensions` | [`vector of integers`](#doc_vector_of_integers) | Index of dimensions that need to be one-hot encoded (if unspecified, all categorical dimensions are one-hot encoded). | `c()` |
+| `input` | [`categorical matrix/data.frame`](#doc_categorical_matrix_data_frame) | Matrix containing data. | `**--**` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save one-hot encoded features data to. | 
+
+### Detailed documentation
+{: #preprocess_one_hot_encoding_detailed-documentation }
+
+This utility takes a dataset and a vector of indices and does one-hot encoding of the respective features at those indices. Indices represent the IDs of the dimensions to be one-hot encoded.
+
+If no dimensions are specified with `dimensions`, then all categorical-type dimensions will be one-hot encoded. Otherwise, only the dimensions given in `dimensions` will be one-hot encoded.
+
+The output matrix with encoded features may be saved with the `output` parameters.
+
+### Example
+So, a simple example where we want to encode 1st and 3rd feature from dataset `"X"` into `"X_output"` would be
+
+```R
+R> output <- preprocess_one_hot_encoding(input=X, dimensions=1, dimensions=3)
+R> X_ouput <- output$output
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+ - [One-hot encoding on Wikipedia](https://en.m.wikipedia.org/wiki/One-hot)
+
+## radical()
+{: #radical }
+
+#### RADICAL
+{: #radical_descr }
+
+```R
+R> library(mlpack)
+R> d <- radical(angles=150, input=matrix(numeric(), 0, 0),
+        noise_std_dev=0.175, objective=FALSE, replicates=30, seed=0, sweeps=0,
+        verbose=FALSE)
+R> output_ic <- d$output_ic
+R> output_unmixing <- d$output_unmixing
+```
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Given a dataset, this can decompose the dataset into an unmixing matrix and an independent component matrix; this can be useful for preprocessing. [Detailed documentation](#radical_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `angles` | [`integer`](#doc_integer) | Number of angles to consider in brute-force search during Radical2D. | `150` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input` | [`numeric matrix`](#doc_numeric_matrix) | Input dataset for ICA. | `**--**` |
+| `noise_std_dev` | [`numeric`](#doc_numeric) | Standard deviation of Gaussian noise. | `0.175` |
+| `objective` | [`logical`](#doc_logical) | If set, an estimate of the final objective function is printed. | `FALSE` |
+| `replicates` | [`integer`](#doc_integer) | Number of Gaussian-perturbed replicates to use (per point) in Radical2D. | `30` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `sweeps` | [`integer`](#doc_integer) | Number of sweeps; each sweep calls Radical2D once for each pair of dimensions. | `0` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_ic` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save independent components to. | 
+| `output_unmixing` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save unmixing matrix to. | 
+
+### Detailed documentation
+{: #radical_detailed-documentation }
+
+An implementation of RADICAL, a method for independent component analysis (ICA).  Assuming that we have an input matrix X, the goal is to find a square unmixing matrix W such that Y = W * X and the dimensions of Y are independent components.  If the algorithm is running particularly slowly, try reducing the number of replicates.
+
+The input matrix to perform ICA on should be specified with the `input` parameter.  The output matrix Y may be saved with the `output_ic` output parameter, and the output unmixing matrix W may be saved with the `output_unmixing` output parameter.
+
+### Example
+For example, to perform ICA on the matrix `"X"` with 40 replicates, saving the independent components to `"ic"`, the following command may be used: 
+
+```R
+R> output <- radical(input=X, replicates=40)
+R> ic <- output$output_ic
+```
+
+### See also
+
+ - [Independent component analysis on Wikipedia](https://en.wikipedia.org/wiki/Independent_component_analysis)
+ - [ICA using spacings estimates of entropy (pdf)](https://www.jmlr.org/papers/volume4/learned-miller03a/learned-miller03a.pdf)
+ - [Radical C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/radical/radical.hpp)
+
+## random_forest()
+{: #random_forest }
+
+#### Random forests
+{: #random_forest_descr }
+
+```R
+R> library(mlpack)
+R> d <- random_forest(input_model=NA, labels=matrix(integer(), 0, 0),
+        maximum_depth=0, minimum_gain_split=0, minimum_leaf_size=1,
+        num_trees=10, print_training_accuracy=FALSE, seed=0, subspace_dim=0,
+        test=matrix(numeric(), 0, 0), test_labels=matrix(integer(), 0, 0),
+        training=matrix(numeric(), 0, 0), verbose=FALSE, warm_start=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of the standard random forest algorithm by Leo Breiman for classification.  Given labeled data, a random forest can be trained and saved for future use; or, a pre-trained random forest can be used for classification. [Detailed documentation](#random_forest_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`RandomForestModel`](#doc_model) | Pre-trained random forest to use for classification. | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Labels for training dataset. | `matrix(integer(), 0, 0)` |
+| `maximum_depth` | [`integer`](#doc_integer) | Maximum depth of the tree (0 means no limit). | `0` |
+| `minimum_gain_split` | [`numeric`](#doc_numeric) | Minimum gain needed to make a split when building a tree. | `0` |
+| `minimum_leaf_size` | [`integer`](#doc_integer) | Minimum number of points in each leaf node. | `1` |
+| `num_trees` | [`integer`](#doc_integer) | Number of trees in the random forest. | `10` |
+| `print_training_accuracy` | [`logical`](#doc_logical) | If set, then the accuracy of the model on the training set will be predicted (verbose must also be specified). | `FALSE` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `subspace_dim` | [`integer`](#doc_integer) | Dimensionality of random subspace to use for each split.  '0' will autoselect the square root of data dimensionality. | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Test dataset to produce predictions for. | `matrix(numeric(), 0, 0)` |
+| `test_labels` | [`integer vector`](#doc_integer_vector) | Test dataset labels, if accuracy calculation is desired. | `matrix(integer(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Training dataset. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `warm_start` | [`logical`](#doc_logical) | If true and passed along with `training` and `input_model` then trains more trees on top of existing model. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`RandomForestModel`](#doc_model) | Model to save trained random forest to. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | Predicted classes for each point in the test set. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #random_forest_detailed-documentation }
+
+This program is an implementation of the standard random forest classification algorithm by Leo Breiman.  A random forest can be trained and saved for later use, or a random forest may be loaded and predictions or class probabilities for points may be generated.
+
+The training set and associated labels are specified with the `training` and `labels` parameters, respectively.  The labels should be in the range `[0, num_classes - 1]`. Optionally, if `labels` is not specified, the labels are assumed to be the last dimension of the training dataset.
+
+When a model is trained, the `output_model` output parameter may be used to save the trained model.  A model may be loaded for predictions with the `input_model`parameter. The `input_model` parameter may not be specified when the `training` parameter is specified.  The `minimum_leaf_size` parameter specifies the minimum number of training points that must fall into each leaf for it to be split.  The `num_trees` controls the number of trees in the random forest.  The `minimum_gain_split` parameter controls the minimum required gain for a decision tree node to split.  Larger values will force higher-confidence splits.  The `maximum_depth` parameter specifies the maximum depth of the tree.  The `subspace_dim` parameter is used to control the number of random dimensions chosen for an individual node's split.  If `print_training_accuracy` is specified, the calculated accuracy on the training set will be printed.
+
+Test data may be specified with the `test` parameter, and if performance measures are desired for that test set, labels for the test points may be specified with the `test_labels` parameter.  Predictions for each test point may be saved via the `predictions`output parameter.  Class probabilities for each prediction may be saved with the `probabilities` output parameter.
+
+### Example
+For example, to train a random forest with a minimum leaf size of 20 using 10 trees on the dataset contained in `"data"`with labels `"labels"`, saving the output random forest to `"rf_model"` and printing the training error, one could call
+
+```R
+R> output <- random_forest(training=data, labels=labels, minimum_leaf_size=20,
+  num_trees=10, print_training_accuracy=TRUE)
+R> rf_model <- output$output_model
+```
+
+Then, to use that model to classify points in `"test_set"` and print the test error given the labels `"test_labels"` using that model, while saving the predictions for each point to `"predictions"`, one could call 
+
+```R
+R> output <- random_forest(input_model=rf_model, test=test_set,
+  test_labels=test_labels)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [decision_tree()](#decision_tree)
+ - [hoeffding_tree()](#hoeffding_tree)
+ - [softmax_regression()](#softmax_regression)
+ - [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ - [Random forests (pdf)](https://www.eecis.udel.edu/~shatkay/Course/papers/BreimanRandomForests2001.pdf)
+ - [RandomForest C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/random_forest/random_forest.hpp)
+
+## krann()
+{: #krann }
+
+#### K-Rank-Approximate-Nearest-Neighbors (kRANN)
+{: #krann_descr }
+
+```R
+R> library(mlpack)
+R> d <- krann(alpha=0.95, first_leaf_exact=FALSE, input_model=NA, k=0,
+        leaf_size=20, naive=FALSE, query=matrix(numeric(), 0, 0),
+        random_basis=FALSE, reference=matrix(numeric(), 0, 0),
+        sample_at_leaves=FALSE, seed=0, single_mode=FALSE,
+        single_sample_limit=20, tau=5, tree_type="kd", verbose=FALSE)
+R> distances <- d$distances
+R> neighbors <- d$neighbors
+R> output_model <- d$output_model
+```
+
+An implementation of rank-approximate k-nearest-neighbor search (kRANN)  using single-tree and dual-tree algorithms.  Given a set of reference points and query points, this can find the k nearest neighbors in the reference set of each query point using trees; trees that are built can be saved for future use. [Detailed documentation](#krann_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `alpha` | [`numeric`](#doc_numeric) | The desired success probability. | `0.95` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `first_leaf_exact` | [`logical`](#doc_logical) | The flag to trigger sampling only after exactly exploring the first leaf. | `FALSE` |
+| `input_model` | [`RAModel`](#doc_model) | Pre-trained kNN model. | `NA` |
+| `k` | [`integer`](#doc_integer) | Number of nearest neighbors to find. | `0` |
+| `leaf_size` | [`integer`](#doc_integer) | Leaf size for tree building (used for kd-trees, UB trees, R trees, R* trees, X trees, Hilbert R trees, R+ trees, R++ trees, and octrees). | `20` |
+| `naive` | [`logical`](#doc_logical) | If true, sampling will be done without using a tree. | `FALSE` |
+| `query` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing query points (optional). | `matrix(numeric(), 0, 0)` |
+| `random_basis` | [`logical`](#doc_logical) | Before tree-building, project the data onto a random orthogonal basis. | `FALSE` |
+| `reference` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing the reference dataset. | `matrix(numeric(), 0, 0)` |
+| `sample_at_leaves` | [`logical`](#doc_logical) | The flag to trigger sampling at leaves. | `FALSE` |
+| `seed` | [`integer`](#doc_integer) | Random seed (if 0, std::time(NULL) is used). | `0` |
+| `single_mode` | [`logical`](#doc_logical) | If true, single-tree search is used (as opposed to dual-tree search. | `FALSE` |
+| `single_sample_limit` | [`integer`](#doc_integer) | The limit on the maximum number of samples (and hence the largest node you can approximate). | `20` |
+| `tau` | [`numeric`](#doc_numeric) | The allowed rank-error in terms of the percentile of the data. | `5` |
+| `tree_type` | [`character`](#doc_character) | Type of tree to use: 'kd', 'ub', 'cover', 'r', 'x', 'r-star', 'hilbert-r', 'r-plus', 'r-plus-plus', 'oct'. | `"kd"` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `distances` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to output distances into. | 
+| `neighbors` | [`integer matrix`](#doc_integer_matrix) | Matrix to output neighbors into. | 
+| `output_model` | [`RAModel`](#doc_model) | If specified, the kNN model will be output here. | 
+
+### Detailed documentation
+{: #krann_detailed-documentation }
+
+This program will calculate the k rank-approximate-nearest-neighbors of a set of points. You may specify a separate set of reference points and query points, or just a reference set which will be used as both the reference and query set. You must specify the rank approximation (in %) (and optionally the success probability).
+
+### Example
+For example, the following will return 5 neighbors from the top 0.1% of the data (with probability 0.95) for each point in `"input"` and store the distances in `"distances"` and the neighbors in `"neighbors.csv"`:
+
+```R
+R> output <- krann(reference=input, k=5, tau=0.1)
+R> distances <- output$distances
+R> neighbors <- output$neighbors
+```
+
+Note that tau must be set such that the number of points in the corresponding percentile of the data is greater than k.  Thus, if we choose tau = 0.1 with a dataset of 1000 points and k = 5, then we are attempting to choose 5 nearest neighbors out of the closest 1 point -- this is invalid and the program will terminate with an error message.
+
+The output matrices are organized such that row i and column j in the neighbors output file corresponds to the index of the point in the reference set which is the i'th nearest neighbor from the point in the query set with index j.  Row i and column j in the distances output file corresponds to the distance between those two points.
+
+### See also
+
+ - [knn()](#knn)
+ - [lsh()](#lsh)
+ - [Rank-approximate nearest neighbor search: Retaining meaning and speed in high dimensions (pdf)](https://papers.nips.cc/paper/3864-rank-approximate-nearest-neighbor-search-retaining-meaning-and-speed-in-high-dimensions.pdf)
+ - [RASearch C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/rann/ra_search.hpp)
+
+## softmax_regression()
+{: #softmax_regression }
+
+#### Softmax Regression
+{: #softmax_regression_descr }
+
+```R
+R> library(mlpack)
+R> d <- softmax_regression(input_model=NA, labels=matrix(integer(), 0,
+        0), lambda=0.0001, max_iterations=400, no_intercept=FALSE,
+        number_of_classes=0, test=matrix(numeric(), 0, 0),
+        test_labels=matrix(integer(), 0, 0), training=matrix(numeric(), 0, 0),
+        verbose=FALSE)
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of softmax regression for classification, which is a multiclass generalization of logistic regression.  Given labeled data, a softmax regression model can be trained and saved for future use, or, a pre-trained softmax regression model can be used for classification of new points. [Detailed documentation](#softmax_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`SoftmaxRegression`](#doc_model) | File containing existing model (parameters). | `NA` |
+| `labels` | [`integer vector`](#doc_integer_vector) | A matrix containing labels (0 or 1) for the points in the training set (y). The labels must order as a row. | `matrix(integer(), 0, 0)` |
+| `lambda` | [`numeric`](#doc_numeric) | L2-regularization constant | `0.0001` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations before termination. | `400` |
+| `no_intercept` | [`logical`](#doc_logical) | Do not add the intercept term to the model. | `FALSE` |
+| `number_of_classes` | [`integer`](#doc_integer) | Number of classes for classification; if unspecified (or 0), the number of classes found in the labels will be used. | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing test dataset. | `matrix(numeric(), 0, 0)` |
+| `test_labels` | [`integer vector`](#doc_integer_vector) | Matrix containing test labels. | `matrix(integer(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | A matrix containing the training set (the matrix of predictors, X). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`SoftmaxRegression`](#doc_model) | File to save trained softmax regression model to. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | Matrix to save predictions for test dataset into. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save class probabilities for test dataset into. | 
+
+### Detailed documentation
+{: #softmax_regression_detailed-documentation }
+
+This program performs softmax regression, a generalization of logistic regression to the multiclass case, and has support for L2 regularization.  The program is able to train a model, load  an existing model, and give predictions (and optionally their accuracy) for test data.
+
+Training a softmax regression model is done by giving a file of training points with the `training` parameter and their corresponding labels with the `labels` parameter. The number of classes can be manually specified with the `number_of_classes` parameter, and the maximum number of iterations of the L-BFGS optimizer can be specified with the `max_iterations` parameter.  The L2 regularization constant can be specified with the `lambda` parameter and if an intercept term is not desired in the model, the `no_intercept` parameter can be specified.
+
+The trained model can be saved with the `output_model` output parameter. If training is not desired, but only testing is, a model can be loaded with the `input_model` parameter.  At the current time, a loaded model cannot be trained further, so specifying both `input_model` and `training` is not allowed.
+
+The program is also able to evaluate a model on test data.  A test dataset can be specified with the `test` parameter. Class predictions can be saved with the `predictions` output parameter.  If labels are specified for the test data with the `test_labels` parameter, then the program will print the accuracy of the predictions on the given test set and its corresponding labels.
+
+### Example
+For example, to train a softmax regression model on the data `"dataset"` with labels `"labels"` with a maximum of 1000 iterations for training, saving the trained model to `"sr_model"`, the following command can be used: 
+
+```R
+R> output <- softmax_regression(training=dataset, labels=labels)
+R> sr_model <- output$output_model
+```
+
+Then, to use `"sr_model"` to classify the test points in `"test_points"`, saving the output predictions to `"predictions"`, the following command can be used:
+
+```R
+R> output <- softmax_regression(input_model=sr_model, test=test_points)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [logistic_regression()](#logistic_regression)
+ - [random_forest()](#random_forest)
+ - [Multinomial logistic regression (softmax regression) on Wikipedia](https://en.wikipedia.org/wiki/Multinomial_logistic_regression)
+ - [SoftmaxRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/softmax_regression/softmax_regression.hpp)
+
+## sparse_coding()
+{: #sparse_coding }
+
+#### Sparse Coding
+{: #sparse_coding_descr }
+
+```R
+R> library(mlpack)
+R> d <- sparse_coding(atoms=15, initial_dictionary=matrix(numeric(), 0,
+        0), input_model=NA, lambda1=0, lambda2=0, max_iterations=0,
+        newton_tolerance=1e-06, normalize=FALSE, objective_tolerance=0.01,
+        seed=0, test=matrix(numeric(), 0, 0), training=matrix(numeric(), 0, 0),
+        verbose=FALSE)
+R> codes <- d$codes
+R> dictionary <- d$dictionary
+R> output_model <- d$output_model
+```
+
+An implementation of Sparse Coding with Dictionary Learning.  Given a dataset, this will decompose the dataset into a sparse combination of a few dictionary elements, where the dictionary is learned during computation; a dictionary can be reused for future sparse coding of new points. [Detailed documentation](#sparse_coding_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `atoms` | [`integer`](#doc_integer) | Number of atoms in the dictionary. | `15` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `initial_dictionary` | [`numeric matrix`](#doc_numeric_matrix) | Optional initial dictionary matrix. | `matrix(numeric(), 0, 0)` |
+| `input_model` | [`SparseCoding`](#doc_model) | File containing input sparse coding model. | `NA` |
+| `lambda1` | [`numeric`](#doc_numeric) | Sparse coding l1-norm regularization parameter. | `0` |
+| `lambda2` | [`numeric`](#doc_numeric) | Sparse coding l2-norm regularization parameter. | `0` |
+| `max_iterations` | [`integer`](#doc_integer) | Maximum number of iterations for sparse coding (0 indicates no limit). | `0` |
+| `newton_tolerance` | [`numeric`](#doc_numeric) | Tolerance for convergence of Newton method. | `1e-06` |
+| `normalize` | [`logical`](#doc_logical) | If set, the input data matrix will be normalized before coding. | `FALSE` |
+| `objective_tolerance` | [`numeric`](#doc_numeric) | Tolerance for convergence of the objective function. | `0.01` |
+| `seed` | [`integer`](#doc_integer) | Random seed.  If 0, 'std::time(NULL)' is used. | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Optional matrix to be encoded by trained model. | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Matrix of training data (X). | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `codes` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save the output sparse codes of the test matrix (--test_file) to. | 
+| `dictionary` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save the output dictionary to. | 
+| `output_model` | [`SparseCoding`](#doc_model) | File to save trained sparse coding model to. | 
+
+### Detailed documentation
+{: #sparse_coding_detailed-documentation }
+
+An implementation of Sparse Coding with Dictionary Learning, which achieves sparsity via an l1-norm regularizer on the codes (LASSO) or an (l1+l2)-norm regularizer on the codes (the Elastic Net).  Given a dense data matrix X with d dimensions and n points, sparse coding seeks to find a dense dictionary matrix D with k atoms in d dimensions, and a sparse coding matrix Z with n points in k dimensions.
+
+The original data matrix X can then be reconstructed as Z * D.  Therefore, this program finds a representation of each point in X as a sparse linear combination of atoms in the dictionary D.
+
+The sparse coding is found with an algorithm which alternates between a dictionary step, which updates the dictionary D, and a sparse coding step, which updates the sparse coding matrix.
+
+Once a dictionary D is found, the sparse coding model may be used to encode other matrices, and saved for future usage.
+
+To run this program, either an input matrix or an already-saved sparse coding model must be specified.  An input matrix may be specified with the `training` option, along with the number of atoms in the dictionary (specified with the `atoms` parameter).  It is also possible to specify an initial dictionary for the optimization, with the `initial_dictionary` parameter.  An input model may be specified with the `input_model` parameter.
+
+### Example
+As an example, to build a sparse coding model on the dataset `"data"` using 200 atoms and an l1-regularization parameter of 0.1, saving the model into `"model"`, use 
+
+```R
+R> output <- sparse_coding(training=data, atoms=200, lambda1=0.1)
+R> model <- output$output_model
+```
+
+Then, this model could be used to encode a new matrix, `"otherdata"`, and save the output codes to `"codes"`: 
+
+```R
+R> output <- sparse_coding(input_model=model, test=otherdata)
+R> codes <- output$codes
+```
+
+### See also
+
+ - [local_coordinate_coding()](#local_coordinate_coding)
+ - [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ - [Efficient sparse coding algorithms (pdf)](http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf)
+ - [Regularization and variable selection via the elastic net](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=46217f372a75dddc2254fdbc6b9418ba3554e453)
+ - [SparseCoding C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/sparse_coding/sparse_coding.hpp)
+
+## adaboost()
+{: #adaboost }
+
+#### AdaBoost
+{: #adaboost_descr }
+
+```R
+R> library(mlpack)
+R> d <- adaboost(input_model=NA, iterations=1000,
+        labels=matrix(integer(), 0, 0), test=matrix(numeric(), 0, 0),
+        tolerance=1e-10, training=matrix(numeric(), 0, 0), verbose=FALSE,
+        weak_learner="decision_stump")
+R> output <- d$output
+R> output_model <- d$output_model
+R> predictions <- d$predictions
+R> probabilities <- d$probabilities
+```
+
+An implementation of the AdaBoost.MH (Adaptive Boosting) algorithm for classification.  This can be used to train an AdaBoost model on labeled data or use an existing AdaBoost model to predict the classes of new points. [Detailed documentation](#adaboost_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`AdaBoostModel`](#doc_model) | Input AdaBoost model. | `NA` |
+| `iterations` | [`integer`](#doc_integer) | The maximum number of boosting iterations to be run (0 will run until convergence.) | `1000` |
+| `labels` | [`integer vector`](#doc_integer_vector) | Labels for the training set. | `matrix(integer(), 0, 0)` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Test dataset. | `matrix(numeric(), 0, 0)` |
+| `tolerance` | [`numeric`](#doc_numeric) | The tolerance for change in values of the weighted error during training. | `1e-10` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Dataset for training AdaBoost. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `weak_learner` | [`character`](#doc_character) | The type of weak learner to use: 'decision_stump', or 'perceptron'. | `"decision_stump"` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`integer vector`](#doc_integer_vector) | Predicted labels for the test set. | 
+| `output_model` | [`AdaBoostModel`](#doc_model) | Output trained AdaBoost model. | 
+| `predictions` | [`integer vector`](#doc_integer_vector) | Predicted labels for the test set. | 
+| `probabilities` | [`numeric matrix`](#doc_numeric_matrix) | Predicted class probabilities for each point in the test set. | 
+
+### Detailed documentation
+{: #adaboost_detailed-documentation }
+
+This program implements the AdaBoost (or Adaptive Boosting) algorithm. The variant of AdaBoost implemented here is AdaBoost.MH. It uses a weak learner, either decision stumps or perceptrons, and over many iterations, creates a strong learner that is a weighted ensemble of weak learners. It runs these iterations until a tolerance value is crossed for change in the value of the weighted training error.
+
+For more information about the algorithm, see the paper "Improved Boosting Algorithms Using Confidence-Rated Predictions", by R.E. Schapire and Y. Singer.
+
+This program allows training of an AdaBoost model, and then application of that model to a test dataset.  To train a model, a dataset must be passed with the `training` option.  Labels can be given with the `labels` option; if no labels are specified, the labels will be assumed to be the last column of the input dataset.  Alternately, an AdaBoost model may be loaded with the `input_model` option.
+
+Once a model is trained or loaded, it may be used to provide class predictions for a given test dataset.  A test dataset may be specified with the `test` parameter.  The predicted classes for each point in the test dataset are output to the `predictions` output parameter.  The AdaBoost model itself is output to the `output_model` output parameter.
+
+Note: the following parameter is deprecated and will be removed in mlpack 4.0.0: `output`.
+Use `predictions` instead of `output`.
+
+### Example
+For example, to run AdaBoost on an input dataset `"data"` with labels `"labels"`and perceptrons as the weak learner type, storing the trained model in `"model"`, one could use the following command: 
+
+```R
+R> output <- adaboost(training=data, labels=labels,
+  weak_learner="perceptron")
+R> model <- output$output_model
+```
+
+Similarly, an already-trained model in `"model"` can be used to provide class predictions from test data `"test_data"` and store the output in `"predictions"` with the following command: 
+
+```R
+R> output <- adaboost(input_model=model, test=test_data)
+R> predictions <- output$predictions
+```
+
+### See also
+
+ - [AdaBoost on Wikipedia](https://en.wikipedia.org/wiki/AdaBoost)
+ - [Improved boosting algorithms using confidence-rated predictions (pdf)](http://rob.schapire.net/papers/SchapireSi98.pdf)
+ - [Perceptron](#perceptron)
+ - [Decision Trees](#decision_tree)
+ - [AdaBoost C++ class documentation](../../user/methods/adaboost.md)
+
+## linear_regression()
+{: #linear_regression }
+
+#### Simple Linear Regression and Prediction
+{: #linear_regression_descr }
+
+```R
+R> library(mlpack)
+R> d <- linear_regression(input_model=NA, lambda=0,
+        test=matrix(numeric(), 0, 0), training=matrix(numeric(), 0, 0),
+        training_responses=matrix(numeric(), 0, 0), verbose=FALSE)
+R> output_model <- d$output_model
+R> output_predictions <- d$output_predictions
+```
+
+An implementation of simple linear regression and ridge regression using ordinary least squares.  Given a dataset and responses, a model can be trained and saved for later use, or a pre-trained model can be used to output regression predictions for a test set. [Detailed documentation](#linear_regression_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `input_model` | [`LinearRegression`](#doc_model) | Existing LinearRegression model to use. | `NA` |
+| `lambda` | [`numeric`](#doc_numeric) | Tikhonov regularization for ridge regression.  If 0, the method reduces to linear regression. | `0` |
+| `test` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing X' (test regressors). | `matrix(numeric(), 0, 0)` |
+| `training` | [`numeric matrix`](#doc_numeric_matrix) | Matrix containing training set X (regressors). | `matrix(numeric(), 0, 0)` |
+| `training_responses` | [`numeric vector`](#doc_numeric_vector) | Optional vector containing y (responses). If not given, the responses are assumed to be the last row of the input file. | `matrix(numeric(), 0, 0)` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output_model` | [`LinearRegression`](#doc_model) | Output LinearRegression model. | 
+| `output_predictions` | [`numeric vector`](#doc_numeric_vector) | If --test_file is specified, this matrix is where the predicted responses will be saved. | 
+
+### Detailed documentation
+{: #linear_regression_detailed-documentation }
+
+An implementation of simple linear regression and simple ridge regression using ordinary least squares. This solves the problem
+
+  y = X * b + e
+
+where X (specified by `training`) and y (specified either as the last column of the input matrix `training` or via the `training_responses` parameter) are known and b is the desired variable.  If the covariance matrix (X'X) is not invertible, or if the solution is overdetermined, then specify a Tikhonov regularization constant (with `lambda`) greater than 0, which will regularize the covariance matrix to make it invertible.  The calculated b may be saved with the `output_predictions` output parameter.
+
+Optionally, the calculated value of b is used to predict the responses for another matrix X' (specified by the `test` parameter):
+
+   y' = X' * b
+
+and the predicted responses y' may be saved with the `output_predictions` output parameter.  This type of regression is related to least-angle regression, which mlpack implements as the 'lars' program.
+
+### Example
+For example, to run a linear regression on the dataset `"X"` with responses `"y"`, saving the trained model to `"lr_model"`, the following command could be used:
+
+```R
+R> output <- linear_regression(training=X, training_responses=y)
+R> lr_model <- output$output_model
+```
+
+Then, to use `"lr_model"` to predict responses for a test set `"X_test"`, saving the predictions to `"X_test_responses"`, the following command could be used:
+
+```R
+R> output <- linear_regression(input_model=lr_model, test=X_test)
+R> X_test_responses <- output$output_predictions
+```
+
+### See also
+
+ - [lars()](#lars)
+ - [Linear regression on Wikipedia](https://en.wikipedia.org/wiki/Linear_regression)
+ - [LinearRegression C++ class documentation](https://github.com/mlpack/mlpack/blob/master/src/mlpack/methods/linear_regression/linear_regression.hpp)
+
+## image_converter()
+{: #image_converter }
+
+#### Image Converter
+{: #image_converter_descr }
+
+```R
+R> library(mlpack)
+R> d <- image_converter(channels=0, dataset=matrix(numeric(), 0, 0),
+        height=0, input=c(), quality=90, save=FALSE, verbose=FALSE, width=0)
+R> output <- d$output
+```
+
+A utility to load an image or set of images into a single dataset that can then be used by other mlpack methods and utilities. This can also unpack an image dataset into individual files, for instance after mlpack methods have been used. [Detailed documentation](#image_converter_detailed-documentation).
+
+
+
+### Input options
+
+| ***name*** | ***type*** | ***description*** | ***default*** |
+|------------|------------|-------------------|---------------|
+| `channels` | [`integer`](#doc_integer) | Number of channels in the image. | `0` |
+| `check_input_matrices` | [`logical`](#doc_logical) | If specified, the input matrix is checked for NaN and inf values; an exception is thrown if any are found. | `FALSE` |
+| `dataset` | [`numeric matrix`](#doc_numeric_matrix) | Input matrix to save as images. | `matrix(numeric(), 0, 0)` |
+| `height` | [`integer`](#doc_integer) | Height of the images. | `0` |
+| `input` | [`vector of characters`](#doc_vector_of_characters) | Image filenames which have to be loaded/saved. | `**--**` |
+| `quality` | [`integer`](#doc_integer) | Compression of the image if saved as jpg (0-100). | `90` |
+| `save` | [`logical`](#doc_logical) | Save a dataset as images. | `FALSE` |
+| `verbose` | [`logical`](#doc_logical) | Display informational messages and the full list of parameters and timers at the end of execution. | `FALSE` |
+| `width` | [`integer`](#doc_integer) | Width of the image. | `0` |
+
+### Output options
+
+Results are returned in a R list.  The keys of the list are the names of the output parameters.
+
+| ***name*** | ***type*** | ***description*** |
+|------------|------------|-------------------|
+| `output` | [`numeric matrix`](#doc_numeric_matrix) | Matrix to save images data to, Onlyneeded if you are specifying 'save' option. | 
+
+### Detailed documentation
+{: #image_converter_detailed-documentation }
+
+This utility takes an image or an array of images and loads them to a matrix. You can optionally specify the height `height` width `width` and channel `channels` of the images that needs to be loaded; otherwise, these parameters will be automatically detected from the image.
+There are other options too, that can be specified such as `quality`.
+
+You can also provide a dataset and save them as images using `dataset` and `save` as an parameter.
+
+### Example
+ An example to load an image : 
+
+```R
+R> output <- image_converter(input=X, height=256, width=256, channels=3)
+R> Y <- output$output
+```
+
+ An example to save an image is :
+
+```R
+R> image_converter(input=X, height=256, width=256, channels=3, dataset=Y,
+  save=TRUE)
+```
+
+### See also
+
+ - [preprocess_binarize()](#preprocess_binarize)
+ - [preprocess_describe()](#preprocess_describe)
+
diff -pruN 4.3.0-2/doc/user/bindings/r.sidebar.html 4.4.0-1/doc/user/bindings/r.sidebar.html
--- 4.3.0-2/doc/user/bindings/r.sidebar.html	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/bindings/r.sidebar.html	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,100 @@
+<ul>
+<li class="page_title">
+<b>mlpack R binding documentation</b>
+<a href="#">[top]</a>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#mlpack-overview">Overview</a></li>
+<li><a href="LINKROOTquickstart/r.html">R Quickstart</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#data-formats">Data Formats</a></li>
+<li><details><summary>Classification</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#decision_tree">decision_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#linear_svm">linear_svm()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#logistic_regression">logistic_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#nbc">nbc()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#perceptron">perceptron()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#random_forest">random_forest()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#softmax_regression">softmax_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#adaboost">adaboost()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Regression</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#bayesian_linear_regression">bayesian_linear_regression()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#lars">lars()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#linear_regression">linear_regression()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Clustering</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#dbscan">dbscan()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#gmm_train">gmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#gmm_generate">gmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#gmm_probability">gmm_probability()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#hoeffding_tree">hoeffding_tree()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#kmeans">kmeans()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#mean_shift">mean_shift()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Geometry</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#approx_kfn">approx_kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#emst">emst()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#fastmks">fastmks()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#lsh">lsh()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#knn">knn()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#kfn">kfn()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#krann">krann()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Preprocessing</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#preprocess_split">preprocess_split()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#preprocess_binarize">preprocess_binarize()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#preprocess_describe">preprocess_describe()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#preprocess_scale">preprocess_scale()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#preprocess_one_hot_encoding">preprocess_one_hot_encoding()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#image_converter">image_converter()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Misc. / Other</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#cf">cf()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#det">det()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#hmm_train">hmm_train()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#hmm_generate">hmm_generate()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#hmm_loglik">hmm_loglik()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#hmm_viterbi">hmm_viterbi()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#kde">kde()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#nmf">nmf()</a></li>
+</ul>
+</details>
+</li>
+
+<li><details><summary>Transformations</summary>
+<ul>
+<li><a href="LINKROOTuser/bindings/r.html#kernel_pca">kernel_pca()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#lmnn">lmnn()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#local_coordinate_coding">local_coordinate_coding()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#nca">nca()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#pca">pca()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#radical">radical()</a></li>
+<li><a href="LINKROOTuser/bindings/r.html#sparse_coding">sparse_coding()</a></li>
+</ul>
+</details>
+</li>
+
+</ul>
+</li>
+</ul>
+</div>
diff -pruN 4.3.0-2/doc/user/build_windows.md 4.4.0-1/doc/user/build_windows.md
--- 4.3.0-2/doc/user/build_windows.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/user/build_windows.md	2024-05-28 20:08:34.000000000 +0000
@@ -94,7 +94,7 @@ OpenBLAS with LLVM's `clang-cl` and `fla
 library (`.lib`) files 
 compatible with the MSVC compiler. A comprehensive guide on the
 compilation of OpenBLAS for Windows can be found
-[here](https://github.com/xianyi/OpenBLAS/wiki/How-to-use-OpenBLAS-in-Microsoft-Visual-Studio).
+[here](https://github.com/OpenMathLib/OpenBLAS/wiki/How-to-use-OpenBLAS-in-Microsoft-Visual-Studio).
 
 One could always download prebuilt `LAPACK` and `BLAS` libraries for Windows.
 However, there are few official sources, and some of those libraries may require
@@ -109,7 +109,7 @@ project.
 ### Armadillo Dependency
 
 - Download the newest version of Armadillo from
-  [Sourceforge](http://arma.sourceforge.net/download.html)
+  [Sourceforge](https://arma.sourceforge.net/download.html)
 - Unzip to `C:\mlpack\armadillo`
 - Create a `build` directory into `C:\mlpack\armadillo\`
 - Open the Command Prompt and navigate to `C:\mlpack\armadillo\build`
@@ -162,7 +162,7 @@ You are ready to create your first appli
 
 This project can be directly built from the `CMakeLists.txt` with the latest
 version of MS Visual Studio, given you have CMake integration via the
-[C++ CMake tools for Windows](https://docs.microsoft.com/en-us/cpp/build/cmake-projects-in-visual-studio?view=msvc-160).
+[C++ CMake tools for Windows](https://learn.microsoft.com/en-us/cpp/build/cmake-projects-in-visual-studio?view=msvc-160).
 To open the CMake project with Visual Studio, select File -> Open -> CMake in
 the top menu, followed by selecting the root `CMakeLists.txt` located in
 mlpack's root directory.
diff -pruN 4.3.0-2/doc/user/core.md 4.4.0-1/doc/user/core.md
--- 4.3.0-2/doc/user/core.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/core.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,2184 @@
+# mlpack core class documentation
+
+Underlying the implementations of [mlpack's machine learning
+algorithms](../index.md#mlpack-algorithm-documentation) are mlpack core support
+classes, each of which are documented on this page.
+
+ * [Core math utilities](#core-math-utilities): utility classes for mathematical
+   purposes
+ * [Distributions](#distributions): probability distributions
+ * [Metrics](#metrics): distance metrics for geometric algorithms
+ * [Kernels](#kernels): Mercer kernels for kernel-based algorithms
+
+## Core math utilities
+
+mlpack provides a number of additional mathematical utility classes and
+functions on top of Armadillo.
+
+ * [Aliases](#aliases): utilities to create and manage aliases (`MakeAlias()`,
+   `ClearAlias()`, `UnwrapAlias()`).
+
+ * [`Range`](#range): simple mathematical range (i.e. `[0, 3]`)
+
+ * [`ColumnCovariance()`](#columncovariance): compute covariance of
+   [column-major](matrices.md#representing-data-in-mlpack) data
+
+ * [`ColumnsToBlocks`](#columnstoblocks): reshape data points into a block
+   matrix for visualization (useful for images)
+
+ * [Distribution utilities](#distribution-utilities): `Digamma()`, `Trigamma()`
+
+ * [`RandVector()`](#randvector): generate random vector on the unit sphere
+   using the Box-Muller transform
+
+ * [Logarithmic utilities](#logarithmic-utilities): `LogAdd()`, `AccuLog()`,
+   `LogSumExp()`, `LogSumExpT()`.
+
+ * [`MultiplyCube2Cube()`](#multiplycube2cube): multiply each slice in a cube by each slice in another cube
+ * [`MultiplyMat2Cube()`](#multiplymat2cube): multiply a matrix by each slice in a cube
+ * [`MultiplyCube2Mat()`](#multiplycube2mat): multiply each slice in a cube by a matrix
+ * [`Quantile()`](#quantile): compute the quantile function of the Gaussian
+   distribution
+
+ * [RNG and random number utilities](#rng-and-random-number-utilities): extended
+   scalar random number generation functions
+ * [`RandomBasis()`](#randombasis): generate a random orthogonal basis
+ * [`ShuffleData()`](#shuffledata): shuffle a dataset and associated labels
+
+---
+
+### Aliases
+
+Aliases are matrix, vector, or cube objects that share memory with another
+matrix, vector, or cube.  They are often used internally inside of mlpack to
+avoid copies.
+
+***Important caveats about aliases***:
+
+ - An alias represents the same memory block as the input.  As such, changes to
+   the alias object will also be reflected in the original object.
+
+ - The `MakeAlias()` function is not guaranteed to return an alias; it only
+   returns an alias *if possible*, and makes a copy otherwise.
+
+ - If `mat` goes out of scope or is destructed, then `a` ***becomes invalid***.
+   _You are responsible for ensuring an invalid alias is not used!_
+
+---
+
+ * `MakeAlias(a, vector, rows, cols, offset=0, strict=true)`
+   - Make `a` into an alias of `vector` with the given size.
+   - If `offset` is `0`, then the alias is identical: the first element of
+     `a` is the first element of `vector`. Otherwise, the first element of `a`
+     is the `offset`'th element of `vector`.
+   - If `strict` is `true`, the size of `a` cannot be changed.
+   - `vector` and `a` should have the same vector type (e.g. `arma::vec`,
+     `arma::fvec`).
+   - If an alias cannot be created, the vector will be copied.
+
+ * `MakeAlias(a, mat, rows, cols, offset=0, strict=true)`
+   - Make `a` into an alias of `mat` with the given size.
+   - If `offset` is `0`, then the alias is identical: the first element of
+     `a` is the first element of `mat`. Otherwise, the first element of `a`
+     is the `offset`'th element of `mat`; elements in `mat` are ordered in
+     a [column-major way](matrices.md#representing-data-in-mlpack).
+   - If `strict` is `true`, the size of `a` cannot be changed.
+   - `mat` and `a` should have the same matrix type (e.g. `arma::mat`,
+     `arma::fmat`, `arma::sp_mat`).
+   - If an alias cannot be created, the matrix will be copied.  Sparse types
+     cannot have aliases and will be copied.
+
+ * `MakeAlias(a, cube, rows, cols, slices, offset=0, strict=true)`
+   - Make `a` into an alias of `cube` with the given size.
+   - If `offset` is `0`, then the alias is identical: the first element of
+     `a` is the first element of `cube`. Otherwise, the first element of `a`
+     is the `offset`'th element of `cube`; elements in `cube` are ordered in
+     a [column-major way](matrices.md#representing-data-in-mlpack).
+   - If `strict` is `true`, the size of `a` cannot be changed.
+   - `cube` and `a` should have the same cube type (e.g. `arma::cube`,
+     `arma::fcube`).
+   - If an alias cannot be created, the cube will be copied.
+
+---
+
+ * `ClearAlias(a)`
+   - If `a` is an alias, reset `a` to an empty matrix, without modifying the
+     aliased memory.  `a` is no longer an alias after this call.
+
+---
+
+ * `UnwrapAlias(a, in)`
+   - If `in` is a matrix type (e.g. `arma::mat`), make `a` into an alias of
+     `in`.
+   - If `in` is not a matrix type, but instead, e.g., an Armadillo expression,
+     fill `a` with the results of the evaluated expression `in`.
+   - This can be used in place of, e.g., `a = in`, to avoid a copy when
+     possible.
+   - `a` should be a matrix type that matches the type of the expression or
+     matrix `in`.
+
+---
+
+### `Range`
+
+The `Range` class represents a simple mathematical range (i.e. `[0, 3]`),
+with the bounds represented as `double`s.
+
+---
+
+#### Constructors
+
+ * `r = Range()`
+   - Construct an empty range.
+
+ * `r = Range(p)`
+   - Construct the range `[p, p]`.
+
+ * `r = Range(lo, hi)`
+   - Construct the range `[lo, hi]`.
+
+---
+
+#### Accessing and modifying range properties
+
+ * `r.Lo()` and `r.Hi()` return the lower and upper bounds of the range as
+   `double`s.
+   - A range is considered empty if `r.Lo() > r.Hi()`.
+   - These can be used to modify the bounds, e.g., `r.Lo() = 3.0`.
+
+ * `r.Width()` returns the span of the range (i.e. `r.Hi() - r.Lo()`) as a
+   `double`.
+
+ * `r.Mid()` returns the midpoint of the range as a `double`.
+
+---
+
+#### Working with ranges
+
+ * Given two ranges `r1` and `r2`,
+   - `r1 | r2` returns the union of the ranges,
+   - `r1 |= r2` expands `r1` to include the range `r2`,
+   - `r1 & r2` returns the intersection of the ranges (possibly an empty range),
+   - `r1 &= r2` shrinks `r1` to the intersection of `r1` and `r2`,
+   - `r1 == r2` returns `true` if the two ranges are strictly equal (i.e. lower
+     and upper bounds are equal),
+   - `r1 != r2` returns `true` if the two ranges are not strictly equal,
+   - `r1 < r2` returns `true` if `r1.Hi() < r2.Lo()`,
+   - `r1 > r2` returns `true` if `r1.Lo() > r2.Hi()`, and
+   - `r1.Contains(r2)` returns `true` if the ranges overlap at all.
+
+ * Given a range `r` and a `double` scalar `d`,
+   - `r * d` returns a new range `[d * r.Lo(), d * r.Hi()]`,
+   - `r *= d` scales `r.Lo()` and `r.Hi()` by `d`, and
+   - `r.Contains(d)` returns `true` if `d` is contained in the range.
+
+---
+
+ * To use ranges with different element types (e.g. `float`), use the type
+   `RangeType<float>` or similar.
+
+---
+
+Example:
+
+```c++
+mlpack::Range r1(5.0, 6.0); // [5, 6]
+mlpack::Range r2(7.0, 8.0); // [7, 8]
+
+mlpack::Range r3 = r1 | r2; // [5, 8]
+mlpack::Range r4 = r1 & r2; // empty range
+
+bool b1 = r1.Contains(r2); // false
+bool b2 = r1.Contains(5.5); // true
+bool b3 = r1.Contains(r3); // true
+bool b4 = r3.Contains(r4); // false
+
+// Create a range of `float`s and a range of `int`s.
+mlpack::RangeType<float> r5(1.0f, 1.5f); // [1.0, 1.5]
+mlpack::RangeType<int> r6(3, 4); // [3, 4]
+```
+
+---
+
+`Range` is used by:
+
+ * [`RangeSearch`](/src/mlpack/methods/range_search/range_search.hpp)
+ * [mlpack trees](../developer/trees.md) <!-- TODO: link to local trees section -->
+
+---
+
+### `ColumnCovariance()`
+
+ * `ColumnCovariance(X, normType=0)`
+   - `X`: a [column-major](matrices.md#representing-data-in-mlpack) data matrix
+   - `normType`: either `0` or `1` (see below)
+
+ * Computes the covariance of the data matrix `X`.
+
+ * Equivalent to `arma::cov(X.t(), normType)`, but avoids computing the
+   transpose and is thus slightly more efficient.
+
+ * `normType` controls the type of normalization done when computing the
+   covariance:
+   - `0` will normalize with `X.n_cols - 1`, providing the best unbiased
+     estimation of the covariance matrix (if the columns are from a normal
+     distribution);
+   - `1` will normalize with `X.n_cols`, providing the second moment about the
+     mean of the columns.
+
+ * Any dense matrix type can be used so long as it supports the Armadillo API
+   (e.g., `arma::mat`, `arma::fmat`, etc.).
+
+Example:
+
+```c++
+// Generate a random data matrix with 100 points in 5 dimensions.
+arma::mat data(5, 100, arma::fill::randu);
+
+// Compute the covariance matrix of the column-major matrix.
+arma::mat cov = mlpack::ColumnCovariance(data);
+cov.print("Covariance of random matrix:");
+```
+
+---
+
+### `ColumnsToBlocks`
+
+The `ColumnsToBlocks` class provides a way to transform data points (e.g.
+columns in a matrix) into a block matrix format, primarily useful for
+visualization as an image.
+
+As a simple example, given a matrix with four columns `A`, `B`, `C`, and `D`,
+`ColumnsToBlocks` can transform this matrix into the form
+
+```
+[[m m m m m]
+ [m A m B m]
+ [m m m m m]
+ [m C m D m]
+ [m m m m m]]
+```
+
+where `m` is a margin, and where each column may itself be reshaped into a
+block.
+
+---
+
+#### Constructors
+
+ * `ctb = ColumnsToBlocks(rows, cols)`
+   - Create a `ColumnsToBlocks` object that will reshape the input matrix into
+     blocks of shape `rows` by `cols`.
+   - Each input column will be reshaped into a square (e.g. `ctb.BlockHeight()`
+     and `ctb.BlockWidth()` are set to `0`).
+
+ * `ctb = ColumnsToBlocks(rows, cols, blockHeight, blockWidth)`
+   - Create a `ColumnsToBlocks` object that will reshape the input matrix into
+     blocks of shape `rows` by `cols`.
+   - Each individual column will also be reshaped into a block of shape
+     `blockHeight` by `blockWidth`.
+
+---
+
+#### Properties
+
+ * `ctb.Rows(rows)` will set the number of rows in the block output to `rows`.
+   - `ctb.Rows()` will return a `size_t` with the current setting.
+
+ * `ctb.Cols(cols)` will set the number of columns in the block output to
+   `cols`.
+   - `ctb.Cols()` will return a `size_t` with the current setting.
+
+ * `ctb.BlockHeight(blockHeight)` will set the number of rows in each individual
+   block to `blockHeight`.
+   - `ctb.BlockHeight()` will return a `size_t` with the current setting.
+   - If `ctb.BlockHeight()` is `0`, each input column will be reshaped into a
+     square; if this is not possible, an exception will be thrown.
+
+ * `ctb.BlockWidth()` will set the number of columns in each individual block to
+   `blockWidth`.
+   - `ctb.BlockWidth()` will return a `size_t` with the current setting.
+   - If `ctb.BlockWidth()` is `0`, each input column will be reshaped into a
+     square; if this is not possible, an exception will be thrown.
+
+ * `ctb.BufSize(bufSize)` will set the number of margin elements to `bufSize`.
+   - `ctb.BufSize()` will return a `size_t` with the current setting.
+   - The default setting is `1`.
+
+ * `ctb.BufValue(bufValue)` will set the element used for margins to `bufValue`.
+   - `ctb.BufValue()` will return a `size_t` with the current setting.
+   - The default setting is `-1.0`.
+
+---
+
+### Scaling values
+
+`ColumnsToBlocks` also has the capability of linearly scaling values of the
+inputs to a given range.
+
+ * `ctb.Scale(true)` enables scaling values.
+   - By default scaling is disabled.
+   - `ctb.Scale(false)` will disable scaling.
+   - `ctb.Scale()` will return a `bool` indicating whether scaling is enabled.
+
+ * `ctb.MinRange(value)` sets the lower bound of the scaling range to `value`.
+   - `ctb.MinRange()` returns the current value as a `double`.
+
+ * `ctb.MaxRange(value)` sets the upper bound of the scaling range to `value`.
+   - `ctb.MaxRange()` returns the current value as a `double`.
+   - Must be greater than `ctb.MinRange()`, if `ctb.Scale() == true`.
+
+***Note:*** the margin element (`ctb.BufValue()`) is considered during the
+scaling process.
+
+---
+
+#### Transforming into block format
+
+ * `ctb.Transform(input, output)` will perform the columns-to-blocks
+   transformation on the given matrix `input`, storing the result in the matrix
+   `output`.
+   - An exception will be thrown if `input.n_rows` is not equal to
+     `ctb.BlockHeight() * ctb.BlockWidth()` (if neither of those are `0`).
+   - If either `ctb.BlockHeight()` or `ctb.BlockWidth()` is `0`, each column
+     will be reshaped into a square, and an exception will be thrown if
+     `input.n_rows` is not a perfect square (i.e. if `sqrt(input.n_rows)` is not
+     an integer).
+
+---
+
+#### Examples
+
+Reshape two 4-element vectors into one row of two blocks.
+
+```c++
+// This matrix has two columns.
+arma::mat input;
+input = { { -1.0000, 0.1429 },
+          { -0.7143, 0.4286 },
+          { -0.4286, 0.7143 },
+          { -0.1429, 1.0000 } };
+input.print("Input columns:");
+
+arma::mat output;
+mlpack::ColumnsToBlocks ctb(1, 2);
+ctb.Transform(input, output);
+
+// The columns of the input will be reshaped as a square which is
+// surrounded by padding value -1 (this value could be changed with the
+// BufValue() method):
+// -1.0000  -1.0000  -1.0000  -1.0000  -1.0000  -1.0000  -1.0000
+// -1.0000  -1.0000  -0.4286  -1.0000   0.1429   0.7143  -1.0000
+// -1.0000  -0.7143  -0.1429  -1.0000   0.4286   1.0000  -1.0000
+// -1.0000  -1.0000  -1.0000  -1.0000  -1.0000  -1.0000  -1.0000
+output.print("Output using 2x2 block size:");
+
+// Now, let's change some parameters; let's have each input column output not
+// as a square, but as a 4x1 vector.
+ctb.BlockWidth(1);
+ctb.BlockHeight(4);
+ctb.Transform(input, output);
+
+// The output here will be similar, but each maximal input is 4x1:
+// -1.0000 -1.0000 -1.0000 -1.0000 -1.0000
+// -1.0000 -1.0000 -1.0000  0.1429 -1.0000
+// -1.0000 -0.7143 -1.0000  0.4286 -1.0000
+// -1.0000 -0.4286 -1.0000  0.7143 -1.0000
+// -1.0000 -0.1429 -1.0000  1.0000 -1.0000
+// -1.0000 -1.0000 -1.0000 -1.0000 -1.0000
+output.print("Output using 4x1 block size:");
+```
+
+---
+
+Load simple images and reshape into blocks.
+
+```c++
+// Load some favicons from websites associated with mlpack.
+std::vector<std::string> images;
+// See the following files:
+// - https://datasets.mlpack.org/images/mlpack-favicon.png
+// - https://datasets.mlpack.org/images/ensmallen-favicon.png
+// - https://datasets.mlpack.org/images/armadillo-favicon.png 
+// - https://datasets.mlpack.org/images/bandicoot-favicon.png
+images.push_back("mlpack-favicon.png");
+images.push_back("ensmallen-favicon.png");
+images.push_back("armadillo-favicon.png");
+images.push_back("bandicoot-favicon.png");
+
+mlpack::data::ImageInfo info;
+info.Channels() = 1; // Force loading in grayscale.
+
+arma::mat matrix;
+mlpack::data::Load(images, matrix, info, true);
+
+// Now `matrix` has 4 columns, each of which is an individual image.
+// Let's save that as its own image just for visualization.
+mlpack::data::ImageInfo outInfo(matrix.n_cols, matrix.n_rows, 1);
+mlpack::data::Save("favicons-matrix.png", matrix, outInfo, true);
+
+// Use ColumnsToBlocks to create a 2x2 block matrix holding each image.
+mlpack::ColumnsToBlocks ctb(2, 2);
+ctb.BufValue(0.0); // Use 0 for the margin value.
+ctb.BufSize(2); // Use 2-pixel margins.
+
+arma::mat blocks;
+ctb.Transform(matrix, blocks);
+
+mlpack::data::ImageInfo blockOutInfo(blocks.n_cols, blocks.n_rows, 1);
+mlpack::data::Save("favicons-blocks.png", blocks, blockOutInfo, true);
+```
+
+The resulting images (before and after using `ColumnsToBlocks`) are shown below.
+
+*Before*:
+
+<center>
+<img src="../img/favicons-matrix.png" alt="four favicons each as a column in a matrix, unintelligible">
+</center>
+
+*After*:
+
+<center>
+<img src="../img/favicons-blocks.png" alt="four favicons each as a block in a larger image, much better">
+</center>
+
+---
+
+#### See also
+
+ * [Loading and saving image data](load_save.md#image-data)
+ * [`SparseAutoencoder`](/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp)
+
+---
+
+### Distribution utilities
+
+ * `Digamma(x)` returns the logarithmic derivative of the gamma function (see
+   [Wikipedia](https://en.wikipedia.org/wiki/Digamma_function)).
+    - `x` should have type `double`.
+    - The return type is `double`.
+
+ * `Trigamma(x)` returns the [trigamma function](https://en.wikipedia.org/wiki/Trigamma_function) at the value `x`.
+    - `x` should have type `double`.
+    - The return type is `double`.
+
+ * Both of these functions are used internally by the
+   `GammaDistribution` class. <!-- TODO: document the class! -->
+
+*Example*:
+
+```
+const double d1 = mlpack::Digamma(0.25);
+const double d2 = mlpack::Digamma(1.0);
+
+const double t1 = mlpack::Trigamma(0.25);
+const double t2 = mlpack::Trigamma(1.0);
+
+std::cout << "Digamma(0.25):  " << d1 << "." << std::endl;
+std::cout << "Digamma(1.0):   " << d2 << "." << std::endl;
+std::cout << "Trigamma(0.25): " << t1 << "." << std::endl;
+std::cout << "Trigamma(1.0):  " << t2 << "." << std::endl;
+```
+
+---
+
+### `RandVector()`
+
+ * `RandVector(v)` generates a random vector on the unit sphere (i.e. with an
+   L2-norm of 1) and stores it in `v` (an `arma::vec`).
+
+ * The [Box-Muller transform](https://en.wikipedia.org/wiki/Box-Muller_transform)
+   is used to generate the vector.
+
+ * `v` is not resized, and should have size equal to the desired dimensionality
+   when `RandVector()` is called.
+
+*Example*:
+
+```
+// Generate a random 10-dimensional vector.
+arma::vec v;
+v.set_size(10);
+RandVector(v);
+v.print("Random 10-dimensional vector: ");
+
+std::cout << "Random 10-dimensional vector: " << std::endl;
+std::cout << v.t();
+std::cout << "L2-norm of vector (should be 1): " << arma::norm(v, 2) << "."
+    << std::endl;
+```
+
+### Logarithmic utilities
+
+mlpack contains a few functions that are useful for working with logarithms, or
+vectors containing logarithms.
+
+ * `LogAdd(x, y)` for scalars `x` and `y` (e.g. `double`, `float`, `int`, etc.)
+   will return `log(e^x + e^y)`.
+
+ * `AccuLog(v)`, given a vector `v` containing log values, will return the
+   scalar log-sum of those values:
+   `log(e^(v[0]) + e^(v[1]) + ... + e^(v[v.n_elem - 1]))`.
+
+---
+
+ * `LogSumExp(m, out)`, given a matrix `m` (`arma::mat`) containing log values,
+   will compute the scalar log-sum of each *column*, storing the result in the
+   column vector `out` (type `arma::vec`).
+   - `out` will be set to size `m.n_cols`.
+   - `out[i]` will be equal to `AccuLog(m.col(i))`.
+   - Different element types can be used for `m` and `out` (e.g. `arma::fmat`
+     and `arma::fvec`).
+
+ * `LogSumExpT(m, out)`, given a matrix `m` (type `arma::mat`) containing log
+   values, will compute the scalar log-sum of each *row*, storing the result in
+   the column vector `out` (type `arma::vec`)
+   - `out` will be set to size `m.n_rows`.
+   - `out[i]` will be equal to `AccuLog(m.row(i))`.
+   - Different element types can be used for `m` and `out` (e.g. `arma::fmat`
+     and `arma::fvec`).
+
+---
+
+ * `LogSumExp<eT, true>(m, out)` performs an incremental sum, otherwise
+   identical to `LogSumExp()`.
+   - The input values of `out` are not ignored.
+   - `out[i]` will be equal to `log(e^(out[i]) + e^(AccuLog(m.col(i))))`.
+   - `eT` represents the element type of `m` and `out` (e.g., `double` if `m` is
+     `arma::mat` and `out` is `arma::vec`).
+
+ * `LogSumExpT<eT, true>(m, out)` performs an incremental sum, otherwise
+   identical to `LogSumExpT()`.
+   - The input values of `out` are not ignored.
+   - `out[i]` will be equal to `log(e^(out[i]) + e^(AccuLog(m.row(i))))`.
+   - `eT` represents the element type of `m` and `out` (e.g., `double` if `m` is
+     `arma::mat` and `out` is `arma::vec`).
+
+---
+
+### `MultiplyCube2Cube()`
+
+ * `z = MultiplyCube2Cube(x, y, transX=false, transY=false)`
+   - Inputs `x` and `y` are cubes (e.g. `arma::cube`), and must have the same
+     number of slices
+   - `z` is a cube whose slices are the slices of `x` and `y` multiplied
+   - `transX` and `transY` indicate whether each slice of `x` and `y` should be
+     transposed before multiplication.
+
+ * If `transX` and `transY` are `false`, then
+   `z.slice(i) = x.slice(i) * y.slice(i)`.
+
+ * If `transX` is `false` and `transY` is `true`, then
+   `z.slice(i) = x.slice(i) * y.slice(i).t()`.
+
+ * The inner dimensions of `x` and `y` must match for multiplication, or an
+   exception will be thrown.
+
+*Example usage:*
+
+```c++
+// Generate two random cubes.
+arma::cube x(10, 100, 5, arma::fill::randu); // 5 matrices, each 10x100.
+arma::cube y(12, 100, 5, arma::fill::randu); // 5 matrices, each 12x100.
+
+arma::cube z = mlpack::MultiplyCube2Cube(x, y, false, true);
+
+// Output size should be 10x12x5.
+std::cout << "Output size: " << z.n_rows << "x" << z.n_cols << "x" << z.n_slices
+    << "." << std::endl;
+```
+
+---
+
+### `MultiplyMat2Cube()`
+
+ * `z = MultiplyMat2Cube(x, y, transX=false, transY=false)`
+   - Input `x` is a matrix and `y` is a cube (e.g. `arma::cube`).
+   - `z` is a cube whose slices are `x` multiplied by the slices of `y`.
+   - `transX` and `transY` indicate whether `x` and each slice of `y` should be
+     transposed before multiplication.
+
+ * If `transX` and `transY` are `false`, then `z.slice(i) = x * y.slice(i)`.
+
+ * If `transX` is `false` and `transY` is `true`, then
+   `z.slice(i) = x * y.slice(i).t()`.
+
+ * The inner dimensions of `x` and `y` must match for multiplication, or an
+   exception will be thrown.
+
+*Example usage:*
+
+```c++
+// Generate random inputs.
+arma::mat  x(10, 100,    arma::fill::randu); // Random 10x100 matrix.
+arma::cube y(12, 100, 5, arma::fill::randu); // 5 matrices, each 12x100.
+
+arma::cube z = mlpack::MultiplyMat2Cube(x, y, false, true);
+
+// Output size should be 10x12x5.
+std::cout << "Output size: " << z.n_rows << "x" << z.n_cols << "x" << z.n_slices
+    << "." << std::endl;
+```
+
+---
+
+### `MultiplyCube2Mat()`
+
+ * `z = MultiplyCube2Mat(x, y, transX=false, transY=false)`
+   - Input `x` is a cube (e.g. `arma::cube`) and `y` is a matrix.
+   - `z` is a cube whose slices are the slices of `x` multiplied with `y`.
+   - `transX` and `transY` indicate whether each slice of `x` and `y` should be
+     transposed before multiplication.
+
+ * If `transX` and `transY` are `false`, then `z.slice(i) = x.slice(i) * y`.
+
+ * If `transX` is `true` and `transY` is `false`, then
+   `z.slice(i) = x.slice(i).t() * y`.
+
+ * The inner dimensions of `x` and `y` must match for multiplication, or an
+   exception will be thrown.
+
+*Example usage:*
+
+```c++
+// Generate two random cubes.
+arma::cube x(12, 50, 5, arma::fill::randu); // 5 matrices, each 12x50.
+arma::mat  y(12, 60,    arma::fill::randu); // Random 12x60 matrix.
+
+arma::cube z = mlpack::MultiplyCube2Mat(x, y, true, false);
+
+// Output size should be 50x60x5.
+std::cout << "Output size: " << z.n_rows << "x" << z.n_cols << "x" << z.n_slices
+    << "." << std::endl;
+```
+
+---
+
+### `Quantile()`
+
+ * Compute the quantile function of the Gaussian distribution at the given
+   probability.
+
+ * `double q = Quantile(p, mu=0.0, sigma=1.0)`
+   - `q` is the computed quantile.
+   - `p` is the probability to compute the quantile of (between 0 and 1).
+   - `mu` is the (optional) mean of the Gaussian distribution.
+   - `sigma` is the (optional) standard deviation of the Gaussian distribution.
+   - All arguments are `double`s.
+
+ * See also [Quantile function on Wikipedia](https://en.wikipedia.org/wiki/Quantile_function).
+
+*Example usage:*
+
+```c++
+// 70% of points from N(0, 1) are less than q1 = 0.524.
+double q1 = mlpack::Quantile(0.7);
+
+// 90% of points from N(0, 1) are less than q2 = 1.282.
+double q2 = mlpack::Quantile(0.9);
+
+// 50% of points from N(1, 1) are less than q3 = 1.0.
+double q3 = mlpack::Quantile(0.5, 1.0); // Quantile of 1.0 for N(1, 1) is 1.0.
+
+// 10% of points from N(1, 0.1) are less than q4 = 0.871.
+double q4 = mlpack::Quantile(0.1, 1.0, 0.1);
+
+std::cout << "Quantile(0.7): " << q1 << "." << std::endl;
+std::cout << "Quantile(0.9): " << q2 << "." << std::endl;
+std::cout << "Quantile(0.5, 1.0): " << q3 << "." << std::endl;
+std::cout << "Quantile(0.1, 1.0, 0.1): " << q4 << "." << std::endl;
+```
+
+### RNG and random number utilities
+
+On top of the random number generation support that Armadillo provides via
+[randu()](https://arma.sourceforge.net/docs.html#randu),
+[randn()](https://arma.sourceforge.net/docs.html#randn), and
+[randi()](https://arma.sourceforge.net/docs.html#randi), mlpack provides
+a few additional thread-safe random number generation functions for generating
+random scalar values.
+
+ * `RandomSeed(seed)` will set the random seed of mlpack's RNGs ***and***
+   Armadillo's RNG to `seed`.
+   - This internally calls `arma::arma_rng::set_seed()`.
+   - In a multithreaded application, each thread's RNG will be deterministically
+     set to a different value based on `seed`.
+
+ * `Random()` returns a random `double` uniformly distributed between `0` and
+   `1`, *not including 1*.
+
+ * `Random(lo, hi)` returns a random `double` uniformly distributed between `lo`
+   and `hi`, *not including `hi`*.
+
+ * `RandBernoulli(p)` samples from a Bernoulli distribution with parameter `p`:
+   with probability `p`, `1` is returned; with probability `1 - p`, `0` is
+   returned.
+
+ * `RandInt(hiExclusive)` returns a random `int` uniformly distributed in the
+   range `[0, hiExclusive)`.
+
+ * `RandInt(lo, hiExclusive)` returns a random `int` uniformly distributed in
+   the range `[lo, hiExclusive)`.
+
+ * `RandNormal()` returns a random `double` normally distributed with mean `0`
+   and standard deviation `1`.
+
+ * `RandNormal(mean, stddev)` returns a random `double` normally distributed
+   with mean `mean` and standard deviation `stddev`.
+
+*Examples*:
+
+```c++
+mlpack::RandomSeed(123); // Set a specific random seed.
+
+const double r1 = mlpack::Random();             // In the range [0, 1).
+const double r2 = mlpack::Random(3, 4);         // In the range [3, 4).
+const double r3 = mlpack::RandBernoulli(0.25);  // P(1) = 0.25.
+const int    r4 = mlpack::RandInt(10);          // In the range [0, 10).
+const int    r5 = mlpack::RandInt(5, 10);       // In the range [5, 10).
+const double r6 = mlpack::RandNormal();         // r6 ~ N(0, 1).
+const double r7 = mlpack::RandNormal(2.0, 3.0); // r7 ~ N(2, 3).
+
+std::cout << "Random():            " << r1 << "." << std::endl;
+std::cout << "Random(3, 4):        " << r2 << "." << std::endl;
+std::cout << "RandBernoulli(0.25): " << r3 << "." << std::endl;
+std::cout << "RandInt(10):         " << r4 << "." << std::endl;
+std::cout << "RandInt(5, 10):      " << r5 << "." << std::endl;
+std::cout << "RandNormal():        " << r6 << "." << std::endl;
+std::cout << "RandNormal(2, 3):    " << r7 << "." << std::endl;
+```
+
+---
+
+### `RandomBasis()`
+
+The `RandomBasis()` function generates a random d-dimensional orthogonal basis.
+
+ * `RandomBasis(basis, d)` fills the matrix `basis` with `d` orthogonal vectors,
+   each of dimension `d`.
+   - `basis.col(i)` represents the `i`th basis vector.
+   - `basis` will have size `d` rows by `d` cols.
+
+ * The random basis is generated using the QR decomposition.
+
+*Example*:
+
+```c++
+arma::mat basis;
+
+// Generate a 10-dimensional random basis.
+mlpack::RandomBasis(basis, 10);
+
+// Each two vectors are orthogonal.
+std::cout << "Dot product of basis vectors 2 and 4: "
+    << arma::dot(basis.col(2), basis.col(4))
+    << " (should be zero or very close!)." << std::endl;
+```
+
+---
+
+### `ShuffleData()`
+
+Shuffle a [column-major](matrices.md#representing-data-in-mlpack) dataset and
+associated labels/responses, optionally with weights.  This preserves the
+connection of each data point to its label (and optionally its weight).
+
+ * `ShuffleData(inputData, inputLabels, outputData, outputLabels)`
+   - Randomly permute data points and labels from `inputData` and `inputLabels`
+     into `outputData` and `outputLabels`.
+   - `outputData` will be set to the same size as `inputData`.
+   - `outputLabels` will be set to the same size as `inputLabels`.
+   - `inputData` can be a dense matrix, a sparse matrix, or a cube, with any
+     element type.  (That is, `inputData` may have type `arma::mat`,
+     `arma::fmat`, `arma::sp_mat`, `arma::cube`, etc.)
+   - `inputLabels` must be a dense vector type but may hold any element type
+     (e.g.  `arma::Row<size_t>`, `arma::uvec`, `arma::vec`, etc.).
+   - `outputData` must have the same type as `inputData`, and `outputLabels`
+     must have the same type as `inputLabels`.
+
+ * `ShuffleData(inputData, inputLabels, inputWeights, outputData, outputLabels, outputWeights)`
+   - Identical to the previous overload, but also handles weights via
+     `inputWeights` and `outputWeights`.
+   - `inputWeights` must be a dense vector type but may hold any element type
+     (e.g.  `arma::rowvec`, `arma::frowvec`, `arma::vec`, etc.)
+   - `outputWeights` must have the same type as `inputWeights`.
+
+***Note:*** when `inputData` is a cube (e.g. `arma::cube` or similar), the
+columns of the cube will be shuffled.
+
+*Example usage:*
+
+```c++
+// See https://datasets.mlpack.org/iris.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.csv", dataset, true);
+// See https://datasets.mlpack.org/iris.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("iris.labels.csv", labels, true);
+
+// Now shuffle the points in the iris dataset.
+arma::mat shuffledDataset;
+arma::Row<size_t> shuffledLabels;
+mlpack::ShuffleData(dataset, labels, shuffledDataset, shuffledLabels);
+
+std::cout << "Before shuffling, the first point was: " << std::endl;
+std::cout << "  " << dataset.col(0).t();
+std::cout << "with label " << labels[0] << "." << std::endl;
+std::cout << std::endl;
+std::cout << "After shuffling, the first point is: " << std::endl;
+std::cout << "  " << shuffledDataset.col(0).t();
+std::cout << "with label " << shuffledLabels[0] << "." << std::endl;
+
+// Generate random weights, then shuffle those also.
+arma::rowvec weights(dataset.n_cols, arma::fill::randu);
+arma::rowvec shuffledWeights;
+mlpack::ShuffleData(dataset, labels, weights, shuffledDataset, shuffledLabels,
+    shuffledWeights);
+
+std::cout << std::endl << std::endl;
+std::cout << "Before shuffling with weights, the first point was: "
+    << std::endl;
+std::cout << "  " << dataset.col(0).t();
+std::cout << "with label " << labels[0] << " and weight " << weights[0] << "."
+    << std::endl;
+std::cout << std::endl;
+std::cout << "After shuffling with weights, the first point is: " << std::endl;
+std::cout << "  " << shuffledDataset.col(0).t();
+std::cout << "with label " << shuffledLabels[0] << " and weight "
+    << shuffledWeights[0] << "." << std::endl;
+```
+
+---
+
+## Distributions
+
+<!-- TODO: link to the completed HMM documentation -->
+
+mlpack has support for a number of different distributions, each supporting the
+same API.  These can be used with, for instance, the
+[`HMM`](/src/mlpack/methods/hmm/hmm.hpp) class.
+
+ * [`DiscreteDistribution`](#discretedistribution): multidimensional categorical
+   distribution (generalized Bernoulli distribution)
+ * [`GaussianDistribution`](#gaussiandistribution): multidimensional Gaussian
+   distribution
+
+### `DiscreteDistribution`
+
+`DiscreteDistribution` represents a multidimensional categorical distribution
+(or generalized Bernoulli distribution) where integer-valued vectors (e.g.
+`[0, 3, 4]`) are associated with specific probabilities in each dimension.
+
+*Example:* a 3-dimensional `DiscreteDistribution` will have a specific
+probability value associated with each integer value in each dimension.  So, for
+the vector `[0, 3, 4]`, `P(0)` in dimension 0 could be, e.g., `0.3`, `P(3)` in
+dimension 1 could be, e.g., `0.4`, and `P(4)` in dimension 2 could be, e.g.,
+`0.6`.  Then, `P([0, 3, 4])` would be `0.3 * 0.4 * 0.6 = 0.072`.
+
+---
+
+#### Constructors
+
+ * `d = DiscreteDistribution(numObservations)`
+   - Create a one-dimensional discrete distribution with `numObservations`
+     different observations in the one and only dimension.  `numObservations` is
+     of type `size_t`.
+
+ * `d = DiscreteDistribution(numObservationsVec)`
+   - Create a multidimensional discrete distribution with
+     `numObservationsVec.n_elem` dimensions and `numObservationsVec[i]`
+     different observations in dimension `i`.
+   - `numObservationsVec` is of type `arma::Col<size_t>`.
+
+ * `d = DiscreteDistribution(probabilities)`
+   - Create a multidimensional discrete distribution with the given
+     probabilities.
+   - `probabilities` should have type `std::vector<arma::vec>`, and
+     `probabilities.size()` should be equal to the dimensionality of the
+     distribution.
+   - `probabilities[i]` is a vector such that `probabilities[i][j]` contains the
+     probability of `j` in dimension `i`.
+
+---
+
+#### Access and modify properties of distribution
+
+ * `d.Dimensionality()` returns a `size_t` indicating the number of dimensions
+   in the multidimensional discrete distribution.
+
+ * `d.Probabilities(i)` returns an `arma::vec&` containing the probabilities of
+   each observation in dimension `i`.
+   - `d.Probabilities(i)[j]` is the probability of `j` in dimension `i`.
+   - This can be used to modify probabilities: `d.Probabilities(0)[1] = 0.7`
+     sets the probability of observing the value `1` in dimension `0` to `0.7`.
+   - *Note:* when setting probabilities manually, be sure that the sum of
+     probabilities in a dimension is 1!
+
+---
+
+#### Compute probabilities of points
+
+ * `d.Probability(observation)` returns the probability of the given
+   observation as a `double`.
+   - `observation` should be an `arma::vec` of size `d.Dimensionality()`.
+   - `observation[i]` should take integer values between `0` and
+     `d.Probabilities(i).n_elem - 1`.
+
+ * `d.Probability(observations, probabilities)` computes the probabilities of
+   many observations.
+   - `observations` should be an `arma::mat` with number of rows equal to
+     `d.Dimensionality()`; `observations.n_cols` is the number of observations.
+   - `probabilities` will be set to size `observations.n_cols`.
+   - `probabilities[i]` will be set to `d.Probability(observations.col(i))`.
+
+ * `d.LogProbability(observation)` returns the log-probability of the given
+   observation as a `double`.
+
+ * `d.LogProbability(observations, probabilities)` computes the
+   log-probabilities of many observations.
+
+---
+
+#### Sample from the distribution
+
+ * `d.Random()` returns an `arma::vec` with a random sample from the
+   multidimensional discrete distribution.
+
+---
+
+#### Fit the distribution to observations
+
+ * `d.Train(observations)`
+   - Fit the distribution to the given observations.
+   - `observations` should be an `arma::mat` with number of rows equal to
+     `d.Dimensionality()`; `observations.n_cols` is the number of observations.
+   - `observations(j, i)` should be an integer value between `0` and the number
+     of observations for dimension `i`.
+
+ * `d.Train(observations, observationProbabilities)`
+   - Fit the distribution to the given observations, as above, but also provide
+     probabilities that each observation is from this distribution.
+   - `observationProbabilities` should be an `arma::vec` of length
+     `observations.n_cols`.
+   - `observationProbabilities[i]` should be equal to the probability that
+     `observations.col(i)` is from `d`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a single-dimension Bernoulli distribution: P([0]) = 0.3, P([1]) = 0.7.
+mlpack::DiscreteDistribution bernoulli(2);
+bernoulli.Probabilities(0)[0] = 0.3;
+bernoulli.Probabilities(0)[1] = 0.7;
+
+const double p1 = bernoulli.Probability(arma::vec("0")); // p1 = 0.3.
+const double p2 = bernoulli.Probability(arma::vec("1")); // p2 = 0.7.
+
+// Create a 3-dimensional discrete distribution by specifying the probabilities
+// manually.
+arma::vec probDim0 = arma::vec("0.1 0.3 0.5 0.1"); // 4 possible values.
+arma::vec probDim1 = arma::vec("0.7 0.3");         // 2 possible values.
+arma::vec probDim2 = arma::vec("0.4 0.4 0.2");     // 3 possible values.
+std::vector<arma::vec> probs { probDim0, probDim1, probDim2 };
+mlpack::DiscreteDistribution d(probs);
+
+arma::vec obs("2 0 1");
+const double p3 = d.Probability(obs); // p3 = 0.5 * 0.7 * 0.4 = 0.14.
+
+// Estimate a 10-dimensional discrete distribution.
+// Each dimension takes values between 0 and 9.
+arma::mat observations = arma::randi<arma::mat>(10, 1000,
+    arma::distr_param(0, 9));
+
+// Create a distribution with 10 observations in each of the 10 dimensions.
+mlpack::DiscreteDistribution d2(
+    arma::Col<size_t>("10 10 10 10 10 10 10 10 10 10"));
+d2.Train(observations);
+
+// Compute the probabilities of each point.
+arma::vec probabilities;
+d2.Probability(observations, probabilities);
+std::cout << "Average probability: " << arma::mean(probabilities) << "."
+    << std::endl;
+```
+
+---
+
+### `GaussianDistribution`
+
+`GaussianDistribution` is a standard multivariate Gaussian distribution with
+parameterized mean and covariance.
+
+---
+
+#### Constructors
+
+ * `g = GaussianDistribution(dimensionality)`
+   - Create the distribution with the given dimensionality.
+   - The distribution will have a zero mean and unit diagonal covariance matrix.
+
+ * `g = GaussianDistribution(mean, covariance)`
+   - Create the distribution with the given mean and covariance.
+   - `mean` is of type `arma::vec` and should have length equal to the
+     dimensionality of the distribution.
+   - `covariance` is of type `arma::mat`, and should be symmetric and square,
+     with rows and columns equal to the dimensionality of the distribution.
+
+---
+
+#### Access and modify properties of distribution
+
+ * `g.Dimensionality()` returns the dimensionality of the distribution as a
+   `size_t`.
+
+ * `g.Mean()` returns an `arma::vec&` holding the mean of the distribution.
+   This can be modified.
+
+ * `g.Covariance()` returns a `const arma::mat&` holding the covariance of the
+   distribution.  To set a new covariance, use `g.Covariance(newCov)` or
+   `g.Covariance(std::move(newCov))`.
+
+ * `g.InvCov()` returns a `const arma::mat&` holding the precomputed inverse of
+   the covariance.
+
+ * `g.LogDetCov()` returns a `double` holding the log-determinant of the
+   covariance.
+
+---
+
+#### Compute probabilities of points
+
+ * `g.Probability(observation)` returns the probability of the given
+   observation as a `double`.
+   - `observation` should be an `arma::vec` of size `d.Dimensionality()`.
+
+ * `g.Probability(observations, probabilities)` computes the probabilities of
+   many observations.
+   - `observations` should be an `arma::mat` with number of rows equal to
+     `d.Dimensionality()`; `observations.n_cols` is the number of observations.
+   - `probabilities` will be set to size `observations.n_cols`.
+   - `probabilities[i]` will be set to `g.Probability(observations.col(i))`.
+
+ * `g.LogProbability(observation)` returns the log-probability of the given
+   observation as a `double`.
+
+ * `g.LogProbability(observations, probabilities)` computes the
+   log-probabilities of many observations.
+
+---
+
+#### Sample from the distribution
+
+ * `g.Random()` returns an `arma::vec` with a random sample from the
+   multidimensional discrete distribution.
+
+---
+
+#### Fit the distribution to observations
+
+ * `g.Train(observations)`
+   - Fit the distribution to the given observations.
+   - `observations` should be an `arma::mat` with number of rows equal to
+     `d.Dimensionality()`; `observations.n_cols` is the number of observations.
+
+ * `g.Train(observations, observationProbabilities)`
+   - Fit the distribution to the given observations, as above, but also provide
+     probabilities that each observation is from this distribution.
+   - `observationProbabilities` should be an `arma::vec` of length
+     `observations.n_cols`.
+   - `observationProbabilities[i]` should be equal to the probability that
+     `observations.col(i)` is from `d`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a Gaussian distribution in 3 dimensions with zero mean and unit
+// covariance.
+mlpack::GaussianDistribution g(3);
+
+// Compute the probability of the point [0, 0.5, 0.25].
+const double p = g.Probability(arma::vec("0 0.5 0.25"));
+
+// Modify the mean in dimension 0.
+g.Mean()[0] = 0.5;
+
+// Set a random covariance.
+arma::mat newCov(3, 3, arma::fill::randu);
+newCov *= newCov.t(); // Ensure covariance is positive semidefinite.
+g.Covariance(std::move(newCov)); // Set new covariance.
+
+// Compute the probability of the same point [0, 0.5, 0.25].
+const double p2 = g.Probability(arma::vec("0 0.5 0.25"));
+
+// Create a Gaussian distribution that is estimated from random samples in 50
+// dimensions.
+arma::mat samples(50, 10000, arma::fill::randn); // Normally distributed.
+
+mlpack::GaussianDistribution g2(50);
+g2.Train(samples);
+
+// Compute the probability of all of the samples.
+arma::vec probabilities;
+g2.Probability(samples, probabilities);
+
+std::cout << "Average probability is: " << arma::mean(probabilities) << "."
+    << std::endl;
+```
+
+## Metrics
+
+mlpack includes a number of distance metrics for its distance-based techniques.
+These all implement the [same API](../developer/metrics.md), providing one
+`Evaluate()` method, and can be used with a variety of different techniques,
+including:
+
+<!-- TODO: better names for each link -->
+
+ * [`NeighborSearch`](/src/mlpack/methods/neighbor_search/neighbor_search.hpp)
+ * [`RangeSearch`](/src/mlpack/methods/range_search/range_search.hpp)
+ * [`LMNN`](/src/mlpack/methods/lmnn/lmnn.hpp)
+ * [`EMST`](/src/mlpack/methods/emst/emst.hpp)
+ * [`NCA`](/src/mlpack/methods/nca/nca.hpp)
+ * [`RANN`](/src/mlpack/methods/rann/rann.hpp)
+ * [`KMeans`](/src/mlpack/methods/kmeans/kmeans.hpp)
+
+Supported metrics:
+
+ * [`LMetric`](#lmetric): generalized L-metric/Lp-metric, including
+   Manhattan/Euclidean/Chebyshev distances
+ * [Implement a custom metric](../developer/metrics.md)
+
+### `LMetric`
+
+The `LMetric` template class implements a [generalized
+L-metric](https://en.wikipedia.org/wiki/Lp_space#Definition)
+(L1-metric, L2-metric, etc.).  The class has two template parameters:
+
+```
+LMetric<Power, TakeRoot>
+```
+
+ * `Power` is an `int` representing the type of the metric; e.g., `2` would
+   represent the L2-metric (Euclidean distance).
+   - `Power` must be `1` or greater.
+   - If `Power` is `INT_MAX`, the metric is the L-infinity distance (Chebyshev
+     distance).
+
+ * `TakeRoot` is a `bool` (default `true`) indicating whether the root of the
+   distance should be taken.
+   - If set to `false`, the metric will no longer satisfy the triangle
+     inequality.
+
+---
+
+Several convenient typedefs are available:
+
+ * `ManhattanDistance` (defined as `LMetric<1>`)
+ * `EuclideanDistance` (defined as `LMetric<2>`)
+ * `SquaredEuclideanDistance` (defined as `LMetric<2, false>`)
+ * `ChebyshevDistance` (defined as `LMetric<INT_MAX>`)
+
+---
+
+The static `Evaluate()` method can be used to compute the distance between two
+vectors.
+
+*Note:* The vectors given to `Evaluate()` can have any type so long as the type
+implements the Armadillo API (e.g. `arma::fvec`, `arma::sp_fvec`, etc.).
+
+---
+
+*Example usage:*
+
+```c++
+// Create two vectors: [0, 1.0, 5.0] and [1.0, 3.0, 5.0].
+arma::vec a("0.0 1.0 5.0");
+arma::vec b("1.0 3.0 5.0");
+
+const double d1 = mlpack::ManhattanDistance::Evaluate(a, b);        // d1 = 3.0
+const double d2 = mlpack::EuclideanDistance::Evaluate(a, b);        // d2 = 2.24
+const double d3 = mlpack::SquaredEuclideanDistance::Evaluate(a, b); // d3 = 5.0
+const double d4 = mlpack::ChebyshevDistance::Evaluate(a, b);        // d4 = 2.0
+const double d5 = mlpack::LMetric<4>::Evaluate(a, b);               // d5 = 2.03
+const double d6 = mlpack::LMetric<3, false>::Evaluate(a, b);        // d6 = 9.0
+
+std::cout << "Manhattan distance:         " << d1 << "." << std::endl;
+std::cout << "Euclidean distance:         " << d2 << "." << std::endl;
+std::cout << "Squared Euclidean distance: " << d3 << "." << std::endl;
+std::cout << "Chebyshev distance:         " << d4 << "." << std::endl;
+std::cout << "L4-distance:                " << d5 << "." << std::endl;
+std::cout << "Cubed L3-distance:          " << d6 << "." << std::endl;
+
+// Compute the distance between two random 10-dimensional vectors in a matrix.
+arma::mat m(10, 100, arma::fill::randu);
+
+const double d7 = mlpack::EuclideanDistance::Evaluate(m.col(0), m.col(7));
+
+std::cout << std::endl;
+std::cout << "Distance between two random vectors: " << d7 << "." << std::endl;
+std::cout << std::endl;
+
+// Compute the distance between two 32-bit precision `float` vectors.
+arma::fvec fa("0.0 1.0 5.0");
+arma::fvec fb("1.0 3.0 5.0");
+
+const double d8 = mlpack::EuclideanDistance::Evaluate(fa, fb); // d8 = 2.236
+
+std::cout << "Euclidean distance (fvec): " << d8 << "." << std::endl;
+```
+
+## Kernels
+
+mlpack includes a number of Mercer kernels for its kernel-based techniques.
+These all implement the [same API](../developer/kernels.md), providing one
+`Evaluate()` method, and can be used with a variety of different techniques,
+including:
+
+<!-- TODO: document everything below -->
+
+ * [`KDE`](/src/mlpack/methods/kde/kde.hpp)
+ * [`MeanShift`](/src/mlpack/methods/mean_shift/mean_shift.hpp)
+ * [`KernelPCA`](/src/mlpack/methods/kernel_pca/kernel_pca.hpp)
+ * [`FastMKS`](/src/mlpack/methods/fastmks/fastmks.hpp)
+ * [`NystroemMethod`](/src/mlpack/methods/nystroem_method/nystroem_method.hpp)
+
+Supported kernels:
+
+ * [`GaussianKernel`](#gaussiankernel): standard Gaussian/radial basis
+   function/RBF kernel
+ * [`CauchyKernel`](#cauchykernel): Cauchy kernel, with longer tails than the
+   standard Gaussian kernel
+ * [`CosineSimilarity`](#cosinesimilarity): dot-product vector similarity
+ * [`EpanechnikovKernel`](#epanechnikovkernel): Epanechnikov kernel (parabolic),
+   with zero tails
+ * [`HyperbolicTangentKernel`](#hyperbolictangentkernel): hyperbolic tangent
+   kernel (not positive definite)
+ * [`LaplacianKernel`](#laplaciankernel): Laplacian kernel/exponential kernel
+ * [`LinearKernel`](#linearkernel): linear (dot-product) kernel
+ * [`PolynomialKernel`](#polynomialkernel): arbitrary-power polynomial kernel
+   with offset
+ * [`PSpectrumStringKernel`](#pspectrumstringkernel): kernel to compute length-p
+   subsequence match counts
+ * [`SphericalKernel`](#sphericalkernel): spherical/uniform/rectangular window
+   kernel
+ * [`TriangularKernel`](#triangularkernel): triangular kernel, with zero tails
+ * [Implement a custom kernel](#implement-a-custom-kernel)
+
+### `GaussianKernel`
+
+The `GaussianKernel` class implements the standard [Gaussian
+kernel](https://en.wikipedia.org/wiki/Radial_basis_function_kernel) (also called
+the _radial basis function kernel_ or _RBF kernel_).
+
+The Gaussian kernel is defined as:
+`k(x1, x2) = exp(-|| x1 - x2 ||^2 / (2 * bw^2))`
+where `bw` is the bandwidth parameter of the kernel.
+
+---
+
+#### Constructors and properties
+
+ * `g = GaussianKernel(bw=1.0)`
+   - Create a `GaussianKernel` with the given bandwidth `bw`.
+
+ * `g.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `g.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `g.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `g.Evaluate(distance)`
+   - Compute the kernel value between two vectors, given that the distance
+     between those two vectors (`distance`) is already known.
+   - `distance` should have type `double`.
+
+---
+
+#### Other utilities
+
+ * `g.Gradient(distance)`
+   - Compute the (one-dimensional) gradient of the kernel function with respect
+     to the distance between two points, evaluated at `distance`.
+
+ * `g.Normalizer(dimensionality)`
+   - Return the
+     [normalizing constant](https://en.wikipedia.org/wiki/Normalizing_constant)
+     of the Gaussian kernel for points in the given dimensionality as a
+     `double`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a Gaussian kernel with default bandwidth.
+mlpack::GaussianKernel g;
+
+// Create a Gaussian kernel with bandwidth 5.0.
+mlpack::GaussianKernel g2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = g.Evaluate(x1, x2);
+const double k2 = g2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Evaluate the kernel value when the distance between two points is already
+// computed.
+const double distance = 1.5;
+const double k3 = g.Evaluate(distance);
+
+// Change the bandwidth of the kernel to 2.5.
+g.Bandwidth(2.5);
+const double k4 = g.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=2.5: " << k4 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = g.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k5 = g.Evaluate(fx1, fx2);
+const double k6 = g2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k5
+    << " (bw=2.5), " << k6 << " (bw=5.0)." << std::endl;
+```
+
+### `CauchyKernel`
+
+The `CauchyKernel` class implements the Cauchy kernel, a kernel function with a
+longer tail than the Gaussian kernel, defined as:
+`k(x1, x2) = 1 / (1 + (|| x1 - x2 ||^2 / bw^2))`
+where `bw` is the bandwidth parameter of the kernel.
+
+---
+
+#### Constructors and properties
+
+ * `c = CauchyKernel(bw=1.0)`
+   - Create a `CauchyKernel` with the given bandwidth `bw`.
+
+ * `c.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `c.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `c.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+---
+
+*Example usage:*
+
+```c++
+// Create a Cauchy kernel with default bandwidth.
+mlpack::CauchyKernel c;
+
+// Create a Cauchy kernel with bandwidth 5.0.
+mlpack::CauchyKernel c2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = c.Evaluate(x1, x2);
+const double k2 = c2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Change the bandwidth of the kernel to 2.5.
+c.Bandwidth(2.5);
+const double k3 = c.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=2.5: " << k3 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = c.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k4 = c.Evaluate(fx1, fx2);
+const double k5 = c2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k4
+    << " (bw=2.5), " << k5 << " (bw=5.0)." << std::endl;
+```
+
+### `CosineSimilarity`
+
+The `CosineSimilarity` class implements the dot-product cosine similarity,
+defined as:
+`k(x1, x2) = (x1^T x2) / (|| x1 || * || x2 ||)`.
+The value of the kernel is limited to the range `[-1, 1]`.
+The cosine similarity is often used in text mining tasks.
+
+---
+
+#### Constructor
+
+ * `c = CosineSimilarity()`
+   - Create a `CosineSimilarity` object.
+
+***Note:*** because the `CosineSimilarity` kernel has no parameters, it is not
+necessary to create an object and the `Evaluate()` function (below) can be
+called statically.
+
+---
+
+#### Kernel evaluation
+
+ * `c.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2` with an
+     instantiated `CosineSimilarity` object.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `CosineDistance::Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2` without an
+     instantiated `CosineSimilarity` object (e.g. call `Evaluate()` statically).
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+---
+
+*Example usage:*
+
+```c++
+// Create a cosine similarity kernel.
+mlpack::CosineSimilarity c;
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = c.Evaluate(x1, x2);
+const double k2 = c.Evaluate(x1, x1);
+const double k3 = c.Evaluate(x2, x2);
+std::cout << "Cosine similarity values:" << std::endl;
+std::cout << "  - k(x1, x2): " << k1 << "." << std::endl;
+std::cout << "  - k(x1, x1): " << k2 << "." << std::endl;
+std::cout << "  - k(x2, x2): " << k3 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix,
+// using the static Evaluate() function.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = mlpack::CosineSimilarity::Evaluate(x1, r.col(i));
+std::cout << "Average cosine similarity for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the cosine similarity between two sparse 32-bit floating point
+// vectors.
+arma::sp_fvec x3, x4;
+x3.sprandu(100, 1, 0.2);
+x4.sprandu(100, 1, 0.2);
+const double k4 = mlpack::CosineSimilarity::Evaluate(x3, x4);
+std::cout << "Cosine similarity between two random sparse 32-bit floating "
+    << "point vectors: " << k4 << "." << std::endl;
+```
+
+### `EpanechnikovKernel`
+
+The `EpanechnikovKernel` implements the
+[parabolic or Epanechnikov kernel](https://en.wikipedia.org/wiki/Kernel_(statistics)#Kernel_functions_in_common_use),
+defined as the following function:
+`k(x1, x2) = max(0, (3 / 4) * (1 - (|| x1 - x2 ||_2 / bw)^2))`,
+where `bw` is the bandwidth parameter of the kernel.
+
+The kernel takes the value `0` when `|| x1 - x2 ||_2` (the Euclidean
+distance between `x1` and `x2`) is greater than or equal to `bw`.
+
+---
+
+#### Constructors and properties
+
+ * `e = EpanechnikovKernel(bw=1.0)`
+   - Create an `EpanechnikovKernel` with the given bandwidth `bw`.
+
+ * `e.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `e.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `e.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `e.Evaluate(distance)`
+   - Compute the kernel value between two vectors, given that the distance
+     between those two vectors (`distance`) is already known.
+   - `distance` should have type `double`.
+
+---
+
+#### Other utilities
+
+ * `e.Gradient(distance)`
+   - Compute the (one-dimensional) gradient of the kernel function with respect
+     to the distance between two points, evaluated at `distance`.
+
+ * `e.Normalizer(dimensionality)`
+   - Return the
+     [normalizing constant](https://en.wikipedia.org/wiki/Normalizing_constant)
+     of the Epanechnikov kernel for points in the given dimensionality as a
+     `double`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create an Epanechnikov kernel with default bandwidth.
+mlpack::EpanechnikovKernel e;
+
+// Create an Epanechnikov kernel with bandwidth 5.0.
+mlpack::EpanechnikovKernel e2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = e.Evaluate(x1, x2);
+const double k2 = e2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Evaluate the kernel value when the distance between two points is already
+// computed.
+const double distance = 1.5;
+const double k3 = e.Evaluate(distance);
+
+// Change the bandwidth of the kernel to 2.5.
+e.Bandwidth(2.5);
+const double k4 = e.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=2.5: " << k4 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = e.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k5 = e.Evaluate(fx1, fx2);
+const double k6 = e2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k5
+    << " (bw=2.5), " << k6 << " (bw=5.0)." << std::endl;
+```
+
+### `HyperbolicTangentKernel`
+
+The `HyperbolicTangentKernel` implements the
+[hyperbolic tangent kernel](https://en.wikipedia.org/wiki/Support_vector_machine#Nonlinear_Kernels),
+which is defined by the following equation:
+`f(x1, x2) = tanh(s * (x1^T x2) + t)`
+where `s` is the scale parameter and `t` is the offset parameter.
+
+The hyperbolic tangent kernel is *not a positive definite Mercer kernel* and
+thus does not satisfy the theoretical requirements of many kernel methods.  See
+[this discussion](https://stats.stackexchange.com/questions/199620/on-the-properties-of-hyperbolic-tangent-kernel)
+for more details.  In practice, for many kernel methods, it may still provide
+compelling results despite this theoretical limitation.
+
+---
+
+#### Constructors and properties
+
+ * `h = HyperbolicTangentKernel(s=1.0, t=0.0)`
+   - Create a `HyperbolicTangentKernel` with the given scale factor `s` and the
+     given offset `t`.
+
+ * `h.Scale()` returns the scale factor of the kernel as a `double`.
+   - To set the scale parameter, use `h.Scale(scale)`.
+
+ * `h.Offset()` returns the offset parameter of the kernel as a `double`.
+   - To set the offset parameter, use `h.Offset(offset)`.
+
+---
+
+#### Kernel evaluation
+
+ * `h.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+---
+
+*Example usage:*
+
+```c++
+// Create a hyperbolic tangent kernel with default scale and offset.
+mlpack::HyperbolicTangentKernel h;
+
+// Create a hyperbolic tangent kernel with scale 2.0 and offset 1.0.
+mlpack::HyperbolicTangentKernel h2(2.0, 1.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = h.Evaluate(x1, x2);
+const double k2 = h2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (s=1.0, t=0.0), " << k2
+    << " (s=2.0, t=1.0)." << std::endl;
+
+// Change the scale and offset of the kernel.
+h.Scale(2.5);
+h.Offset(-1.0);
+const double k3 = h.Evaluate(x1, x2);
+std::cout << "Kernel value with s=2.5, t=-1.0: " << k3 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = h.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k4 = h.Evaluate(fx1, fx2);
+const double k5 = h2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k4
+    << " (s=2.5, t=-1.0), " << k5 << " (s=2.0, t=1.0)." << std::endl;
+```
+
+### `LaplacianKernel`
+
+The `LaplacianKernel` class implements the Laplacian kernel, also known as the
+exponential kernel, defined by the following equation:
+`k(x1, x2) = exp(-|| x1 - x2 || / bw)`
+where `bw` is the bandwidth parameter.
+
+---
+
+#### Constructors and properties
+
+ * `l = LaplacianKernel(bw=1.0)`
+   - Create a `LaplacianKernel` with the given bandwidth `bw`.
+
+ * `l.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `l.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `l.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `l.Evaluate(distance)`
+   - Compute the kernel value between two vectors, given that the distance
+     between those two vectors (`distance`) is already known.
+   - `distance` should have type `double`.
+
+---
+
+#### Other utilities
+
+ * `l.Gradient(distance)`
+   - Compute the (one-dimensional) gradient of the kernel function with respect
+     to the distance between two points, evaluated at `distance`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a Laplacian kernel with default bandwidth.
+mlpack::LaplacianKernel l;
+
+// Create a Laplacian kernel with bandwidth 5.0.
+mlpack::LaplacianKernel l2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = l.Evaluate(x1, x2);
+const double k2 = l2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Evaluate the kernel value when the distance between two points is already
+// computed.
+const double distance = 1.5;
+const double k3 = l.Evaluate(distance);
+
+// Change the bandwidth of the kernel to 2.5.
+l.Bandwidth(2.5);
+const double k4 = l.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=2.5: " << k4 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = l.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k5 = l.Evaluate(fx1, fx2);
+const double k6 = l2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k5
+    << " (bw=2.5), " << k6 << " (bw=5.0)." << std::endl;
+```
+
+### `LinearKernel`
+
+The `LinearKernel` class implements the simple linear dot product kernel,
+defined by the following equation:
+`k(x1, x2) = x1^T x2`.
+
+The use of the linear kernel for kernel methods generally results in the
+non-kernelized version of the algorithm; for instance, a kernel support
+vector machine using the linear kernel amounts to a [linear
+SVM](methods/linear_svm.md).
+
+---
+
+#### Constructor
+
+ * `l = LinearKernel()`
+   - Create a `LinearKernel` object.
+
+***Note:*** because the `LinearKernel` kernel has no parameters, it is not
+necessary to create an object and the `Evaluate()` function (below) can be
+called statically.
+
+---
+
+#### Kernel evaluation
+
+ * `l.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2` with an
+     instantiated `LinearKernel` object.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `LinearKernel::Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2` without an
+     instantiated `LinearKernel` object (e.g. call `Evaluate()` statically).
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+---
+
+*Example usage:*
+
+```c++
+// Create a linear kernel.
+mlpack::LinearKernel l;
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = l.Evaluate(x1, x2); // Identical to arma::dot(x1, x2).
+const double k2 = l.Evaluate(x1, x1);
+const double k3 = l.Evaluate(x2, x2);
+std::cout << "Linear kernel values:" << std::endl;
+std::cout << "  - k(x1, x2): " << k1 << "." << std::endl;
+std::cout << "  - k(x1, x1): " << k2 << "." << std::endl;
+std::cout << "  - k(x2, x2): " << k3 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix,
+// using the static Evaluate() function.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = mlpack::LinearKernel::Evaluate(x1, r.col(i));
+std::cout << "Average linear kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the cosine similarity between two sparse 32-bit floating point
+// vectors.
+arma::sp_fvec x3, x4;
+x3.sprandu(100, 1, 0.2);
+x4.sprandu(100, 1, 0.2);
+const double k4 = mlpack::LinearKernel::Evaluate(x3, x4);
+std::cout << "Linear kernel value between two random sparse 32-bit floating "
+    << "point vectors: " << k4 << "." << std::endl;
+```
+
+### `PolynomialKernel`
+
+The `PolynomialKernel` class implements the standard
+[polynomial kernel](https://en.wikipedia.org/wiki/Polynomial_kernel), which is
+defined by the following equation:
+`k(x1, x2) = (x1^T x2 + t)^d`
+where `d` is the degree of the polynomial and `t` is the offset.
+
+The use of the polynomial kernel has a similar effect to the use of polynomial
+(interaction) features in standard machine learning methods.
+
+---
+
+#### Constructors and properties
+
+ * `p = PolynomialKernel(d=2.0, t=0.0)`
+   - Create a `PolynomialKernel` with the given degree `d` and given offset `t`.
+
+ * `p.Degree()` returns the degree of the kernel as a `double`.
+   - To set the degree, use `p.Degree(newDegree)`.
+
+ * `p.Offset()` returns the offset of the kernel as a `double`.
+   - To set the offset, use `p.Offset(newOffset)`.
+
+---
+
+#### Kernel evaluation
+
+ * `p.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+---
+
+*Example usage:*
+
+```c++
+// Create a polynomial kernel with default degree (2) and offset (0).
+mlpack::PolynomialKernel p;
+
+// Create a polynomial kernel with degree 3.0 and offset -1.0.
+mlpack::PolynomialKernel p2(3.0, -1.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = p.Evaluate(x1, x2);
+const double k2 = p2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Change the degree of the kernel to 2.5 and the offset to 1.0.
+p.Degree(2.5);
+p.Offset(1.0);
+const double k3 = p.Evaluate(x1, x2);
+std::cout << "Kernel value with d=2.5, t=1.0: " << k3 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = p.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k4 = p.Evaluate(fx1, fx2);
+const double k5 = p2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k4
+    << " (d=2.5, t=1.0), " << k5 << " (d=3.0, t=-1.0)." << std::endl;
+```
+
+### `PSpectrumStringKernel`
+
+The `PSpectrumStringKernel` class implements the length-`p` string spectrum
+kernel, proposed by
+[Leslie, Eskin, and Noble (pdf)](http://psb.stanford.edu/psb-online/proceedings/psb02/leslie.pdf).
+The kernel finds the contiguous subsequence match count between two strings.
+
+Due to mlpack's use of Armadillo, which requires that all matrix data be
+numeric, this class operates by internally storing all strings, and passing in
+numeric vectors such as `[0 1]` that reference string index `1` in dataset index
+`0`.  In turn, this means that the data points given to the
+`PSpectrumStringKernel` are simply IDs and have no geometric meaning.
+
+---
+
+#### Constructors and properties
+
+ * `p = PSpectrumStringKernel(datasets, p)`
+    - Create a `PSpectrumStringKernel` on the given set of string datasets,
+      using the given substring length `p`.
+    - `datasets` should have type `std::vector<std::vector<std::string>>`, and
+      contains a list of datasets, each of which is made up of a list of
+      strings.
+      * Multiple datasets are supported for the case where, e.g., there are
+        multiple files containing different sets of strings.
+    - So, e.g., `datasets[0]` represents the `0`th dataset, and `datasets[0][1]`
+      is the string with index `1` inside the `0`th dataset.
+    - `p` (a `size_t`) is the length of substring to use for the kernel, and
+      must be greater than `0`.
+    - The constructor will build counts of all substrings in the dataset, and
+      for large data may be computationally intensive.
+
+ * `p.P()` returns the substring length `p` of the kernel as a `size_t`.
+    - The value of `p` cannot be changed once the object is constructed.
+
+ * `p.Counts()` returns a `std::vector<std::vector<std::map<std::string, int>>>`
+   that maps a substring to the number of times it appears in the original
+   string.  So, given a substring length of `5`, `p.Counts()[0][1]["hello"]`
+   would be the number of times the substring `hello` appears in the string with
+   index `1` in the dataset with index `0`.
+
+---
+
+#### Kernel evaluation
+
+ * `p.Evaluate(x1, x2)`
+   - Compute the kernel value between two index vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`, `arma::uvec`, etc.).
+   - `x1` and `x2` do not contain string data directly, but instead are each
+     length-2 vectors that represent the index of the datasets and strings to be
+     compared in the `datasets` object that was passed to the constructor.
+   - So, e.g., if `x1 = [0, 0]` and `x2 = [1, 1]`, then the first string from
+     the first dataset will be compared with the second string from the second
+     dataset.
+
+---
+
+*Example usage:*
+
+```c++
+// Create two example datasets:
+//      ["hello", "goodbye", "package"],
+//      ["mlpack", "is", "really", "great"]
+std::vector<std::vector<std::string>> datasets;
+datasets.push_back({ "hello", "goodbye", "package" });
+datasets.push_back({ "mlpack", "is", "really", "great" });
+
+// Create a p-spectrum string kernel with a substring length of 2,
+// and another with a substring length of 3.
+mlpack::PSpectrumStringKernel p(datasets, 2);
+mlpack::PSpectrumStringKernel p2(datasets, 3);
+
+// Evaluate the kernel value between "mlpack" and "package".
+arma::uvec x1("1 0"); // "mlpack": dataset 1, string 0.
+arma::uvec x2("0 2"); // "package": dataset 0, string 2.
+const double k1 = p.Evaluate(x1, x2);
+const double k2 = p2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (p=2), " << k2 << " (p=3)."
+    << std::endl;
+```
+
+### `SphericalKernel`
+
+The `SphericalKernel` class implements the simple spherical kernel, also known
+as the uniform kernel, or rectangular window kernel.  The value of the
+`SphericalKernel` is `1` when the Euclidean distance between two points `x1` and
+`x2` is less than the bandwidth `bw`, and `0` otherwise:
+`k(x1, x2) = 1(|| x1 - x2 || <= bw)`.
+
+---
+
+#### Constructors and properties
+
+ * `s = SphericalKernel(bw=1.0)`
+   - Create a `SphericalKernel` with the given bandwidth `bw`.
+
+ * `s.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `s.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `s.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `s.Evaluate(distance)`
+   - Compute the kernel value between two vectors, given that the distance
+     between those two vectors (`distance`) is already known.
+   - `distance` should have type `double`.
+
+---
+
+#### Other utilities
+
+ * `s.Gradient(distance)`
+   - Compute the (one-dimensional) gradient of the kernel function with respect
+     to the distance between two points, evaluated at `distance`.
+
+ * `s.Normalizer(dimensionality)`
+   - Return the
+     [normalizing constant](https://en.wikipedia.org/wiki/Normalizing_constant)
+     of the spherical kernel for points in the given dimensionality as a
+     `double`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a spherical kernel with default bandwidth.
+mlpack::SphericalKernel s;
+
+// Create a spherical kernel with bandwidth 5.0.
+mlpack::SphericalKernel s2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 2.5");
+arma::vec x2("2.5 1.0 0.5");
+const double k1 = s.Evaluate(x1, x2);
+const double k2 = s2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Evaluate the kernel value when the distance between two points is already
+// computed.
+const double distance = 0.9;
+const double k3 = s.Evaluate(distance);
+
+// Change the bandwidth of the kernel to 3.0.
+s.Bandwidth(3.0);
+const double k4 = s.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=3.0: " << k4 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix, using
+// a kernel bandwidth of 2.5.
+s.Bandwidth(2.5);
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = s.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 2.5");
+arma::fvec fx2("2.5 1.0 0.5");
+const double k5 = s.Evaluate(fx1, fx2);
+const double k6 = s2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k5
+    << " (bw=2.5), " << k6 << " (bw=5.0)." << std::endl;
+```
+
+### `TriangularKernel`
+
+The `TriangularKernel` class implements the
+[simple triangular kernel](https://en.wikipedia.org/wiki/Kernel_(statistics)#Kernel_functions_in_common_use),
+defined by the following equation:
+`k(x1, x2) = max(0, 1 - || x1 - x2 || / bw)`
+where `bw` is the bandwidth of the kernel.
+
+---
+
+#### Constructors and properties
+
+ * `t = TriangularKernel(bw=1.0)`
+   - Create a `TriangularKernel` with the given bandwidth `bw`.
+
+ * `t.Bandwidth()` returns the bandwidth of the kernel as a `double`.
+   - To set the bandwidth, use `t.Bandwidth(newBandwidth)`.
+
+---
+
+#### Kernel evaluation
+
+ * `t.Evaluate(x1, x2)`
+   - Compute the kernel value between two vectors `x1` and `x2`.
+   - `x1` and `x2` should be vector types that implement the Armadillo API
+     (e.g., `arma::vec`).
+
+ * `t.Evaluate(distance)`
+   - Compute the kernel value between two vectors, given that the distance
+     between those two vectors (`distance`) is already known.
+   - `distance` should have type `double`.
+
+---
+
+#### Other utilities
+
+ * `t.Gradient(distance)`
+   - Compute the (one-dimensional) gradient of the kernel function with respect
+     to the distance between two points, evaluated at `distance`.
+
+---
+
+*Example usage:*
+
+```c++
+// Create a triangular kernel with default bandwidth.
+mlpack::TriangularKernel t;
+
+// Create a triangular kernel with bandwidth 5.0.
+mlpack::TriangularKernel t2(5.0);
+
+// Evaluate the kernel value between two 3-dimensional points.
+arma::vec x1("0.5 1.0 1.5");
+arma::vec x2("1.5 1.0 0.5");
+const double k1 = t.Evaluate(x1, x2);
+const double k2 = t2.Evaluate(x1, x2);
+std::cout << "Kernel values: " << k1 << " (bw=1.0), " << k2 << " (bw=5.0)."
+    << std::endl;
+
+// Evaluate the kernel value when the distance between two points is already
+// computed.
+const double distance = 0.75;
+const double k3 = t.Evaluate(distance);
+
+// Change the bandwidth of the kernel to 2.5.
+t.Bandwidth(2.5);
+const double k4 = t.Evaluate(x1, x2);
+std::cout << "Kernel value with bw=2.5: " << k4 << "." << std::endl;
+
+// Evaluate the kernel value between x1 and all points in a random matrix.
+arma::mat r(3, 100, arma::fill::randu);
+arma::vec kernelValues(100);
+for (size_t i = 0; i < r.n_cols; ++i)
+  kernelValues[i] = t.Evaluate(x1, r.col(i));
+std::cout << "Average kernel value for random points: "
+    << arma::mean(kernelValues) << "." << std::endl;
+
+// Compute the kernel value between two 32-bit floating-point vectors.
+arma::fvec fx1("0.5 1.0 1.5");
+arma::fvec fx2("1.5 1.0 0.5");
+const double k5 = t.Evaluate(fx1, fx2);
+const double k6 = t2.Evaluate(fx1, fx2);
+std::cout << "Kernel values between two floating-point vectors: " << k5
+    << " (bw=2.5), " << k6 << " (bw=5.0)." << std::endl;
+```
+
+### Implement a custom kernel
+
+mlpack supports custom kernels, so long as they implement an appropriate
+`Evaluate()` function.
+
+See [The KernelType Policy in mlpack](../developer/kernels.md) for more
+information.
diff -pruN 4.3.0-2/doc/user/formats.md 4.4.0-1/doc/user/formats.md
--- 4.3.0-2/doc/user/formats.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/user/formats.md	1970-01-01 00:00:00.000000000 +0000
@@ -1,525 +0,0 @@
-# File formats and loading data in mlpack
-
-mlpack supports a wide variety of data (including images) and model formats for
-use in both its command-line programs, and in C++ programs via the
-`mlpack::data::Load()` function.  This tutorial discusses the formats that are
-supported and how to use them.
-
-## Simple examples to load data in C++
-
-The example code snippets below load data from different formats into an
-Armadillo matrix object (`arma::mat`) or model when using C++.
-
-```c++
-using namespace mlpack;
-
-arma::mat matrix1;
-data::Load("dataset.csv", matrix1);
-```
-
-```c++
-using namespace mlpack;
-
-arma::mat matrix2;
-data::Load("dataset.bin", matrix2);
-```
-
-```c++
-using namespace mlpack;
-
-arma::mat matrix3;
-data::Load("dataset.h5", matrix3);
-```
-
-```c++
-using namespace mlpack;
-
-// ARFF loading is a little different, since sometimes mapping has to be done
-// for string types.
-arma::mat matrix4;
-data::DatasetInfo datasetInfo;
-data::Load("dataset.arff", matrix4, datasetInfo);
-
-// The datasetInfo object now holds information about each dimension.
-```
-
-```c++
-using namespace mlpack;
-
-LogisticRegression lr;
-data::Load("model.bin", "logistic_regression_model", lr);
-```
-
-## Supported dataset types
-
-Datasets in mlpack are represented internally as sparse or dense numeric
-matrices (specifically, as `arma::mat` or `arma::sp_mat` or similar).  This
-means that when datasets are loaded from file, they must be converted to a
-suitable numeric representation.  Therefore, in general, datasets on disk should
-contain only numeric features in order to be loaded successfully by mlpack.
-
-The types of datasets that mlpack can load are roughly the same as the types of
-matrices that Armadillo can load.  However, the load functionality that mlpack
-provides ***only supports loading dense datasets***.  When datasets are loaded
-by mlpack, ***the file's type is detected using the file's extension***.
-mlpack supports the following file types:
-
- - csv (comma-separated values), denoted by `.csv` or `.txt`
- - tsv (tab-separated values), denoted by `.tsv`, `.csv`, or `.txt`
- - ASCII (raw ASCII, with space-separated values), denoted by `.txt`
- - Armadillo ASCII (Armadillo's text format with a header), denoted by `.txt`
- - PGM, denoted by `.pgm`
- - PPM, denoted by `.ppm`
- - Armadillo binary, denoted by `.bin`
- - Raw binary, denoted by `.bin` ***(note: this will be loaded as
-   one-dimensional data, which is likely not what is desired.)***
- - HDF5, denoted by `.hdf`, `.hdf5`, `.h5`, or `.he5` ***(note: HDF5 must be
-   enabled in the Armadillo configuration)***
- - ARFF, denoted by .arff ***(note: this is not supported by all mlpack
-   command-line programs***; see below)
-
-Datasets that are loaded by mlpack should be stored with ***one row for one
-point*** and ***one column for one dimension***.  Therefore, a dataset with
-three two-dimensional points `(0, 1)`, `(3, 1)`, and `(5, -5)` would be stored
-in a csv file as:
-
-```
-0, 1
-3, 1
-5, -5
-```
-
-As noted earlier, for command-line programs, the format is automatically
-detected at load time.  Therefore, a dataset can be loaded in many ways:
-
-```
-$ mlpack_logistic_regression -t dataset.csv -v
-[INFO ] Loading 'dataset.csv' as CSV data.  Size is 32 x 37749.
-...
-
-$ mlpack_logistic_regression -t dataset.txt -v
-[INFO ] Loading 'dataset.txt' as raw ASCII formatted data.  Size is 32 x 37749.
-...
-
-$ mlpack_logistic_regression -t dataset.h5 -v
-[INFO ] Loading 'dataset.h5' as HDF5 data.  Size is 32 x 37749.
-...
-```
-
-Similarly, the format to save to is detected by the extension of the given
-filename.
-
-## Loading simple matrices in C++
-
-When C++ is being written, the `mlpack::data::Load()` and `mlpack::data::Save()`
-functions are used to load and save datasets, respectively.  These functions
-should be preferred over the built-in Armadillo `.load()` and `.save()`
-functions.
-
-Matrices in mlpack are column-major, meaning that each column should correspond
-to a point in the dataset and each row should correspond to a dimension; for
-more information, see [matrices in mlpack](matrices.md).  This is at odds with
-how the data is stored in files; therefore, a transposition is required during
-load and save.  The `mlpack::data::Load()` and `mlpack::data::Save()` functions
-do this automatically (unless otherwise specified), which is why they are
-preferred over the Armadillo functions.
-
-To load a matrix from file, the call is straightforward.  After creating a
-matrix object, the data can be loaded:
-
-```c++
-arma::mat dataset; // The data will be loaded into this matrix.
-mlpack::data::Load("dataset.csv", dataset);
-```
-
-Saving matrices is equally straightforward.  The code below generates a random
-matrix with 10 points in 3 dimensions and saves it to a file as HDF5.
-
-```c++
-// 3 dimensions (rows), with 10 points (columns).
-arma::mat dataset = arma::randu<arma::mat>(3, 10);
-mlpack::data::Save("dataset.h5", dataset);
-```
-
-As with the command-line programs, the type of data to be loaded is
-automatically detected from the filename extension.  For more details, see the
-`mlpack::data::Load()` and `mlpack::data::Save()` documentation.
-
-## Dealing with sparse matrices
-
-As mentioned earlier, support for loading sparse matrices in mlpack is not
-available at this time.  To use a sparse matrix with mlpack code, you will have
-to write a C++ program instead of using any of the command-line tools, because
-the command-line tools all use dense datasets internally.  (There is one
-exception: the `mlpack_cf` command-line program, for collaborative filtering,
-loads sparse coordinate lists.)
-
-In addition, the `mlpack::data::Load()` function does not support loading any
-sparse format; so the best idea is to use Armadillo functionality to load
-coordinate lists.  Suppose you have a coordinate list file like the one below:
-
-```sh
-$ cat cl.csv
-0 0 0.332
-1 3 3.126
-4 4 1.333
-```
-
-This represents a 5x5 matrix with three nonzero elements.  We can load this
-using Armadillo:
-
-```c++
-arma::sp_mat matrix;
-matrix.load("cl.csv", arma::coord_ascii);
-matrix = matrix.t(); // We must transpose after load!
-```
-
-The transposition after loading is necessary if the coordinate list is in
-row-major format (that is, if each row in the matrix represents a point and each
-column represents a feature).  Be sure that the matrix you use with mlpack
-methods has points as columns and features as rows!  See [matrices](matrices.md)
-for more information.
-
-## Categorical features and command line programs
-
-In some situations it is useful to represent data not just as a numeric matrix
-but also as categorical data (i.e. with numeric but unordered categories).  This
-support is useful for, e.g., decision trees and other models that support
-categorical features.
-
-In some machine learning situations, such as, e.g., decision trees, categorical
-data can be used.  Categorical data might look like this (in CSV format):
-
-```
-0, 1, "true", 3
-5, -2, "false", 5
-2, 2, "true", 4
-3, -1, "true", 3
-4, 4, "not sure", 0
-0, 7, "false", 6
-```
-
-In the example above, the third dimension (which takes values `"true"`,
-`"false"`, and `"not sure"`) is categorical.  mlpack can load and work with this
-data, but the strings must be mapped to numbers, because all dataset in mlpack
-are represented by Armadillo matrix objects.
-
-From the perspective of an mlpack command-line program, this support is
-transparent; mlpack will attempt to load the data file, and if it detects
-entries in the file that are not numeric, it will map them to numbers and then
-print, for each dimension, the number of mappings.  For instance, if we run the
-`mlpack_hoeffding_tree` program (which supports categorical data) on the dataset
-above (stored as `dataset.csv`), we receive this output during loading:
-
-```sh
-$ mlpack_hoeffding_tree -t dataset.csv -l dataset.labels.csv -v
-[INFO ] Loading 'dataset.csv' as CSV data.  Size is 6 x 4.
-[INFO ] 0 mappings in dimension 0.
-[INFO ] 0 mappings in dimension 1.
-[INFO ] 3 mappings in dimension 2.
-[INFO ] 0 mappings in dimension 3.
-...
-```
-
-Currently, only the `mlpack_hoeffding_tree` and `mlpack_decision_tree` programs
-supports loading categorical data, and this is also the only program that
-supports loading an ARFF dataset.
-
-## Categorical features and C++
-
-When writing C++, loading categorical data is slightly more tricky: the mappings
-from strings to integers must be preserved.  This is the purpose of the
-`mlpack::data::DatasetInfo` class, which stores these mappings and can be used
-and load and save time to apply and de-apply the mappings.
-
-When loading a dataset with categorical data, the overload of
-`mlpack::data::Load()` that takes an `mlpack::data::DatasetInfo` object should
-be used.  An example is below:
-
-```c++
-arma::mat dataset; // Load into this matrix.
-mlpack::data::DatasetInfo info; // Store information about dataset in this.
-
-// Load the ARFF dataset.
-mlpack::data::Load("dataset.arff", dataset, info);
-```
-
-After this load completes, the `info` object will hold the information about the
-mappings necessary to load the dataset.  It is possible to re-use the
-`DatasetInfo` object to load another dataset with the same mappings.  This is
-useful when, for instance, both a training and test set are being loaded, and it
-is necessary that the mappings from strings to integers for categorical features
-are identical.  An example is given below.
-
-```c++
-arma::mat trainingData; // Load training data into this matrix.
-mlpack::data::DatasetInfo info; // This will store the mappings.
-
-// Load the training data, and create the mappings in the 'info' object.
-mlpack::data::Load("training_data.arff", trainingData, info);
-
-// Load the test data, but re-use the 'info' object with the already initialized
-// mappings.  This means that the same mappings will be applied to the test set.
-mlpack::data::Load("test_data.arff", trainingData, info);
-```
-
-When saving data, pass the same `DatasetInfo` object it was loaded with in order
-to unmap the categorical features correctly.  The example below demonstrates
-this functionality: it loads the dataset, increments all non-categorical
-features by 1, and then saves the dataset with the same DatasetInfo it was
-loaded with.
-
-```c++
-arma::mat dataset; // Load data into this matrix.
-mlpack::data::DatasetInfo info; // This will store the mappings.
-
-// Load the dataset.
-mlpack::data::Load("dataset.tsv", dataset, info);
-
-// Loop over all features, and add 1 to all non-categorical features.
-for (size_t i = 0; i < info.Dimensionality(); ++i)
-{
-  // The Type() function returns whether or not the data is numeric or
-  // categorical.
-  if (info.Type(i) != mlpack::data::Datatype::categorical)
-    dataset.row(i) += 1.0;
-}
-
-// Save the modified dataset using the same DatasetInfo.
-mlpack::data::Save("dataset-new.tsv", dataset, info);
-```
-
-There is more functionality to the `DatasetInfo` class; for more information,
-see the `mlpack::data::DatasetInfo` documentation.
-
-## Loading and Saving Images
-
-Image datasets are becoming increasingly popular in deep learning.
-
-mlpack's image saving/loading functionality is based on
-[STB](https://github.com/nothings/stb).
-
-### Image Utilities API
-
-mlpack's image utilities support loading and saving of images, if STB is
-available on the system.
-
-There is support for the following filetypes: `jpg`, `png`, `tga`, `bmp`, `psd`,
-`gif`, `hdr`, `pic`, `pnm` for loading, and `jpg`, `png`, `tga`, `bmp`, `hdr`
-for saving.
-
-The datatype associated is unsigned char to support RGB values in the range
-1-255. To feed data into the network typecast of `arma::Mat` may be required.
-Images are stored in the matrix as
-`(width * height * channels, numberOfImages)`.  Therefore `imageMatrix.col(0)`
-would be the first image if images are loaded in `imageMatrix`.
-
-### Accessing Metadata of Images: ImageInfo
-
-`ImageInfo` class contains the metadata of the images.
-
-```c++
-ImageInfo(const size_t width,
-          const size_t height,
-          const size_t channels,
-          const size_t quality = 90);
-```
-
-The `quality` member denotes the compression of the image if it is saved as
-`jpg`; it takes values from 0 to 100.
-
-### Loading Images in C++
-
-Standalone loading of images.  Below is the signature of the
-`mlpack::data::Load()` method for images:
-
-```c++
-template<typename eT>
-bool Load(const std::string& filename,
-          arma::Mat<eT>& matrix,
-          ImageInfo& info,
-          const bool fatal);
-```
-
-The example below loads a test image. It also fills up the ImageInfo class
-object.
-
-```c++
-data::ImageInfo info;
-data::Load("test_image.png", matrix, info, false);
-```
-
-`ImageInfo` requires height, width, number of channels of the image.
-
-```c++
-size_t height = 64, width = 64, channels = 1;
-data::ImageInfo info(width, height, channels);
-```
-
-More than one image can be loaded into the same matrix.
-
-Loading multiple images: below is the signature of the `mlpack::data::Load()`
-method to load multiple images.
-
-```c++
-template<typename eT>
-bool Load(const std::vector<std::string>& files,
-          arma::Mat<eT>& matrix,
-          ImageInfo& info,
-          const bool fatal);
-```
-
-Here is example usage:
-
-```c++
-data::ImageInfo info;
-std::vector<std::string>> files{"test_image1.bmp","test_image2.bmp"};
-data::Load(files, matrix, info, false);
-```
-
-### Saving Images in C++
-
-`mlpack::data::Save()` images expects a matrix of type `unsigned char` in the
-form `(width * height * channels, numberOfImages)`.  Just like `Load()`, it can
-be used to save one image or multiple images. Besides image data it also expects
-the shape of the image as input `(width, height, channels)`.
-
-Saving one image: below is the signature of the `mlpack::data::Save()` method
-for saving one image.
-
-```c++
-template<typename eT>
-bool Save(const std::string& filename,
-          arma::Mat<eT>& matrix,
-          ImageInfo& info,
-          const bool fatal,
-          const bool transpose);
-```
-
-Below is example usage:
-
-```c++
-data::ImageInfo info;
-info.width = info.height = 25;
-info.channels = 3;
-info.quality = 90;
-data::Save("test_image.bmp", matrix, info, false, true);
-```
-
-If the matrix contains more than one image, only the first one is saved.
-
-Saving multiple images: below is the signature of the `mlpack::data::Save()`
-method for saving multiple images.
-
-```c++
-template<typename eT>
-bool Save(const std::vector<std::string>& files,
-          arma::Mat<eT>& matrix,
-          ImageInfo& info,
-          const bool fatal,
-          const bool transpose);
-```
-
-Below is example usage:
-
-```c++
-data::ImageInfo info;
-info.width = info.height = 25;
-info.channels = 3;
-info.quality = 90;
-std::vector<std::string>> files{"test_image1.bmp", "test_image2.bmp"};
-data::Save(files, matrix, info, false, true);
-```
-
-Multiple images are saved according to the vector of filenames specified.
-
-## Loading and Saving Models
-
-Using `cereal`, mlpack is able to load and save machine learning
-models with ease.  These models can currently be saved in three formats:
-
- - binary (`.bin`); this is not human-readable, but it is small
- - json (`.json`); this is sort of human-readable and relatively small
- - xml (`.xml`); this is human-readable but very verbose and large
-
-The type of file to save is determined by the given file extension, as with the
-other loading and saving functionality in mlpack.  Below is an example where a
-dataset stored as TSV and labels stored as ASCII text are used with mlpack's
-command line programs to train a logistic regression model, which is then saved
-to `model.xml`.
-
-```sh
-$ mlpack_logistic_regression -t training_dataset.tsv -l training_labels.txt \
-> -M model.xml
-```
-
-Many mlpack command-line programs have support for loading and saving models
-through the `--input_model_file` (`-m`) and `--output_model_file` (`-M`)
-options; for more information, see the documentation for each program
-(accessible by passing `--help` as a parameter).
-
-mlpack's bindings to other languages, similarly, have `input_model` parameters,
-and depending on the language, may have `output_model` parameters (or may simply
-return models as part of the output).
-
-### Loading and Saving Models in C++
-
-mlpack uses the `cereal` library internally to perform loading and saving of
-models, and provides convenience overloads of `mlpack::data::Load()` and
-`mlpack::data::Save()` to load and save these models.
-
-To be serializable, a class must implement the method
-
-```c++
-template<typename Archive>
-void serialize(Archive& ar);
-```
-
-For more information on this method and how it works, see the
-[cereal documentation](https://uscilab.github.io/cereal/index.html).
-
-Examples of `serialize()` methods can be found in most classes; one simple one
-is in the `Range` class.  A more complex example is found in the
-`BinarySpaceTree` class.
-
-Using the `mlpack::data::Load()` and `mlpack::data::Save()` classes is easy if
-the type being saved has a `serialize()` method implemented: simply call either
-function with a filename, a name for the object to save, and the object itself.
-The example below, for instance, creates an `Range` object and saves it
-with the name `"range"` as `range.txt`.  Then, that range is loaded from file
-into another `Range` object.
-
-```c++
-// Create range and save it.
-Range r(0.0, 5.0);
-mlpack::data::Save("range.json", "range", r);
-
-// Load into new range.
-Range newRange;
-mlpack::data::Load("range.json", "range", newRange);
-```
-
-It is important to be sure that you load the appropriate type; if you save, for
-instance, a `LogisticRegression` object and attempt to load it as an `Range`
-object, the load will fail and an exception will be thrown.  (When the object is
-saved as binary (.bin), it is possible that the load will not fail, but instead
-load with mangled data, which is perhaps even worse!)
-
-Note that all models saved by command-line bindings are given the name
-`"model"` by default.  Use, e.g., `mlpack::data::Load("model.bin", "model", m)`
-to load a model saved by a command-line binding.
-
-## Final Notes
-
-If the examples here are unclear, it would be worth looking into the ways that
-`mlpack::data::Load()` and `mlpack::data::Save()` are used in the code.  Some
-example files that may be useful to this end:
-
- - `src/mlpack/methods/logistic_regression/logistic_regression_main.cpp`
- - `src/mlpack/methods/hoeffding_trees/hoeffding_tree_main.cpp`
- - `src/mlpack/methods/neighbor_search/knn_main.cpp`
-
-If you are interested in adding support for more data types to mlpack, it would
-be preferable to add the support upstream to Armadillo instead, so that may be a
-better direction to go first.  Then very little code modification for mlpack
-will be necessary.
diff -pruN 4.3.0-2/doc/user/load_save.md 4.4.0-1/doc/user/load_save.md
--- 4.3.0-2/doc/user/load_save.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/load_save.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,760 @@
+# Loading and saving in mlpack
+
+mlpack provides the `data::Load()` and `data::Save()` functions to load and save
+[Armadillo matrices](matrices.md) (e.g. numeric and categorical datasets) and
+any mlpack object via the [cereal](https://uscilab.github.io/cereal/)
+serialization toolkit.  A number of other utilities related to loading and
+saving data and objects are also available.
+
+ * [Numeric data](#numeric-data)
+ * [Mixed categorical data](#mixed-categorical-data)
+   - [`data::DatasetInfo`](#datadatasetinfo)
+   - [Loading categorical data](#loading-categorical-data)
+ * [Image data](#image-data)
+   - [`data::ImageInfo`](#dataimageinfo)
+   - [Loading images](#loading-images)
+ * [mlpack objects](#mlpack-objects): load or save any mlpack object
+ * [Normalizing labels](#normalizing-labels): convert labels to ranges required
+   by mlpack classifiers
+ * [Formats](#formats): supported formats for each load/save variant
+
+## Numeric data
+
+Numeric data or general numeric matrices can be loaded or saved with the
+following functions.
+
+ - `data::Load(filename, matrix, fatal=false, transpose=true, format=FileType::AutoDetect)`
+ - `data::Save(filename, matrix, fatal=false, transpose=true, format=FileType::AutoDetect)`
+   * `filename` is a `std::string` with a path to the file to be loaded.
+
+   * By default the format is auto-detected based on the file extension, but can
+     be explicitly specified with `format`; see [Formats](#formats).
+
+   * `matrix` is an `arma::mat&`, `arma::Mat<size_t>&`, `arma::sp_mat&`, or
+     similar (e.g., a reference to an Armadillo object that data will be loaded
+     into or saved from).
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown on failure.
+
+   * If `transpose` is `true`, then for plaintext formats (CSV/TSV/ASCII), the
+     matrix will be transposed on save.  (Keep this `true` if you want a
+     column-major matrix to be saved with points as rows and dimensions as
+     columns; that is generally what is desired.)
+
+   * A `bool` is returned indicating whether the operation was successful.
+
+---
+
+Example usage:
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat dataset;
+mlpack::data::Load("satellite.train.csv", dataset, true);
+
+// See https://datasets.mlpack.org/satellite.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("satellite.train.labels.csv", labels, true);
+
+// Print information about the data.
+std::cout << "The data in 'satellite.train.csv' has: " << std::endl;
+std::cout << " - " << dataset.n_cols << " points." << std::endl;
+std::cout << " - " << dataset.n_rows << " dimensions." << std::endl;
+
+std::cout << "The labels in 'satellite.train.labels.csv' have: " << std::endl;
+std::cout << " - " << labels.n_elem << " labels." << std::endl;
+std::cout << " - A maximum label of " << labels.max() << "." << std::endl;
+std::cout << " - A minimum label of " << labels.min() << "." << std::endl;
+
+// Modify and save the data.  Add 2 to the data and drop the last column.
+dataset += 2;
+dataset.shed_col(dataset.n_cols - 1);
+labels.shed_col(labels.n_cols - 1);
+
+mlpack::data::Save("satellite.train.mod.csv", dataset);
+mlpack::data::Save("satellite.train.labels.mod.csv", labels);
+```
+
+---
+
+## Mixed categorical data
+
+Some mlpack techniques support mixed categorical data, e.g., data where some
+dimensions take only categorical values (e.g. `0`, `1`, `2`, etc.).  When using
+mlpack, string data and other non-numerical data must be mapped to categorical
+values and represented as part of an `arma::mat`.  Category information is
+stored in an auxiliary `data::DatasetInfo` object.
+
+### `data::DatasetInfo`
+
+<!-- TODO: also document in core.md? -->
+
+mlpack represents categorical data via the use of the auxiliary
+`data::DatasetInfo` object, which stores information about which dimensions are
+numeric or categorical and allows conversion from the original category values
+to the numeric values used to represent those categories.
+
+---
+
+#### Constructors
+
+ - `info = data::DatasetInfo()`
+   * Create an empty `data::DatasetInfo` object.
+   * Use this constructor if you intend to populate the `data::DatasetInfo` via
+     a `data::Load()` call.
+
+ - `info = data::DatasetInfo(dimensionality)`
+   * Create a `data::DatasetInfo` object with the given dimensionality
+   * All dimensions are assumed to be numeric (not categorical).
+
+---
+
+#### Accessing and setting properties
+
+ - `info.Type(d)`
+   * Get the type (categorical or numeric) of dimension `d`.
+   * Returns a `data::Datatype`, either `data::Datatype::numeric` or
+     `data::Datatype::categorical`.
+   * Calling `info.Type(d) = t` will set a dimension to type `t`, but this
+     should only be done before `info` is used with `data::Load()` or
+     `data::Save()`.
+
+ - `info.NumMappings(d)`
+   * Get the number of categories in dimension `d` as a `size_t`.
+   * Returns `0` if dimension `d` is numeric.
+
+ - `info.Dimensionality()`
+   * Return the dimensionality of the object as a `size_t`.
+
+---
+
+#### Map to and from numeric values
+
+ - `info.MapString<double>(value, d)`
+   * Given `value` (a `std::string`), return the `double` representing the
+     categorical mapping (an integer value) of `value` in dimension `d`.
+   * If a mapping for `value` does not exist in dimension `d`, a new mapping is
+     created, and `info.NumMappings(d)` is increased by one.
+   * If dimension `d` is numeric and `value` cannot be parsed as a numeric
+     value, then dimension `d` is changed to categorical and a new mapping is
+     returned.
+
+ - `info.UnmapString(mappedValue, d)`
+   * Given `mappedValue` (a `size_t`), return the `std::string` containing the
+     original category that mapped to the value `mappedValue` in dimension `d`.
+   * If dimension `d` is not categorical, a `std::invalid_argument` is thrown.
+
+---
+
+### Loading categorical data
+
+With a `data::DatasetInfo` object, categorical data can be loaded:
+
+ - `data::Load(filename, matrix, info, fatal=false, transpose=true)`
+   * `filename` is a `std::string` with a path to the file to be loaded.
+
+   * The format is auto-detected based on the extension of the filename and the
+     contents of the file:
+     - `.csv`, `.tsv`, or `.txt` for CSV/TSV (tab-separated)/ASCII
+       (space-separated)
+     - `.arff` for [ARFF](https://ml.cms.waikato.ac.nz/weka/arff.html)
+
+   * `matrix` is an `arma::mat&`, `arma::Mat<size_t>&`, or similar (e.g., a
+     reference to an Armadillo object that data will be loaded into or saved
+     from).
+
+   * `info` is a `data::DatasetInfo&` object.  This will be populated with the
+     category information of the file when loading, and used to unmap values
+     when saving.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown on failure.
+
+   * If `transpose` is `true`, then for plaintext formats (CSV/TSV/ASCII), the
+     matrix will be transposed on save.  (Keep this `true` if you want a
+     column-major matrix to be saved with points as rows and dimensions as
+     columns; that is generally what is desired.)
+
+   * A `bool` is returned indicating whether the operation was successful.
+
+Saving should be performed with the [numeric](#numeric-data) `data::Load()`
+variant.
+
+---
+
+Example usage to load and manipulate an ARFF file.
+
+```c++
+// Load a categorical dataset.
+arma::mat dataset;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/covertype.train.arff.
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+
+arma::Row<size_t> labels;
+// See https://datasets.mlpack.org/covertype.train.labels.csv.
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+
+// Print information about the data.
+std::cout << "The data in 'covertype.train.arff' has: " << std::endl;
+std::cout << " - " << dataset.n_cols << " points." << std::endl;
+std::cout << " - " << info.Dimensionality() << " dimensions." << std::endl;
+
+// Print information about each dimension.
+for (size_t d = 0; d < info.Dimensionality(); ++d)
+{
+  if (info.Type(d) == mlpack::data::Datatype::categorical)
+  {
+    std::cout << " - Dimension " << d << " is categorical with "
+        << info.NumMappings(d) << " categories." << std::endl;
+  }
+  else
+  {
+    std::cout << " - Dimension " << d << " is numeric." << std::endl;
+  }
+}
+
+// Modify the 5th point.  Increment any numeric values, and set any categorical
+// values to the string "hooray!".
+for (size_t d = 0; d < info.Dimensionality(); ++d)
+{
+  if (info.Type(d) == mlpack::data::Datatype::categorical)
+  {
+    // This will create a new mapping if the string "hooray!" does not already
+    // exist as a category for dimension d..
+    dataset(d, 4) = info.MapString<double>("hooray!", d);
+  }
+  else
+  {
+    dataset(d, 4) += 1.0;
+  }
+}
+```
+
+---
+
+Example usage to manually create a `data::DatasetInfo` object.
+
+```c++
+// This will manually create the following data matrix (shown as it would appear
+// in a CSV):
+//
+// 1, TRUE, "good", 7.0, 4
+// 2, FALSE, "good", 5.6, 3
+// 3, FALSE, "bad", 6.1, 4
+// 4, TRUE, "bad", 6.1, 1
+// 5, TRUE, "unknown", 6.3, 0
+// 6, FALSE, "unknown", 5.1, 2
+//
+// Although the last dimension is numeric, we will take it as a categorical
+// dimension.
+
+arma::mat dataset(5, 6); // 6 data points in 5 dimensions.
+mlpack::data::DatasetInfo info(5);
+
+// Set types of dimensions.  By default they are numeric so we only set
+// categorical dimensions.
+info.Type(1) = mlpack::data::Datatype::categorical;
+info.Type(2) = mlpack::data::Datatype::categorical;
+info.Type(4) = mlpack::data::Datatype::categorical;
+
+// The first dimension is numeric.
+dataset(0, 0) = 1;
+dataset(0, 1) = 2;
+dataset(0, 2) = 3;
+dataset(0, 3) = 4;
+dataset(0, 4) = 5;
+dataset(0, 5) = 6;
+
+// The second dimension is categorical.
+dataset(1, 0) = info.MapString<double>("TRUE", 1);
+dataset(1, 1) = info.MapString<double>("FALSE", 1);
+dataset(1, 2) = info.MapString<double>("FALSE", 1);
+dataset(1, 3) = info.MapString<double>("TRUE", 1);
+dataset(1, 4) = info.MapString<double>("TRUE", 1);
+dataset(1, 5) = info.MapString<double>("FALSE", 1);
+
+// The third dimension is categorical.
+dataset(2, 0) = info.MapString<double>("good", 2);
+dataset(2, 1) = info.MapString<double>("good", 2);
+dataset(2, 2) = info.MapString<double>("bad", 2);
+dataset(2, 3) = info.MapString<double>("bad", 2);
+dataset(2, 4) = info.MapString<double>("unknown", 2);
+dataset(2, 5) = info.MapString<double>("unknown", 2);
+
+// The fourth dimension is numeric.
+dataset(3, 0) = 7.0;
+dataset(3, 1) = 5.6;
+dataset(3, 2) = 6.1;
+dataset(3, 3) = 6.1;
+dataset(3, 4) = 6.3;
+dataset(3, 5) = 5.1;
+
+// The fifth dimension is categorical.  Note that `info` will choose to assign
+// category values in the order they are seen, even if the category can be
+// parsed as a number.  So, here, the value '4' will be assigned category '0',
+// since it is seen first.
+dataset(4, 0) = info.MapString<double>("4", 4);
+dataset(4, 1) = info.MapString<double>("3", 4);
+dataset(4, 2) = info.MapString<double>("4", 4);
+dataset(4, 3) = info.MapString<double>("1", 4);
+dataset(4, 4) = info.MapString<double>("0", 4);
+dataset(4, 5) = info.MapString<double>("2", 4);
+
+// Print the dataset with mapped categories.
+dataset.print("Dataset with mapped categories");
+
+// Print the mappings for the third dimension.
+std::cout << "Mappings for dimension 3: " << std::endl;
+for (size_t i = 0; i < info.NumMappings(2); ++i)
+{
+  std::cout << " - \"" << info.UnmapString(i, 2) << "\" maps to " << i << "."
+      << std::endl;
+}
+
+// Now `dataset` is ready for use with an mlpack algorithm that supports
+// categorical data.
+```
+
+---
+
+## Image data
+
+If the STB image library is available on the system (`stb_image.h` and
+`stb_image_write.h` must be available on the compiler's include search path),
+then mlpack will define the `MLPACK_HAS_STB` macro, and support for loading
+individual images or sets of images will be available.
+
+Supported formats for loading are `jpg`, `png`, `tga`, `bmp`, `psd`, `gif`, `hdr`, `pic`, and `pnm`.
+
+Supported formats for saving are `jpg`, `png`, `tga`, `bmp`, and `hdr`.
+
+When loading images, each image is represented as a flattened single column
+vector in a data matrix; each row of the resulting vector will correspond to a
+single pixel value in a single channel.  An auxiliary `data::ImageInfo` class is
+used to store information about the images.
+
+### `data::ImageInfo`
+
+The `data::ImageInfo` class contains the metadata of the images.
+
+---
+
+#### Constructors
+
+ - `info = data::ImageInfo()`
+   * Create a `data::ImageInfo` object with no data.
+   * Use this constructor if you intend to populate the `data::ImageInfo` via a
+     `data::Load()` call.
+
+ - `info = data::ImageInfo(width, height, channels)`
+   * Create a `data::ImageInfo` object with the given image specifications.
+   * `width` and `height` are specified as pixels.
+
+---
+
+#### Accessing and modifying image metadata
+
+ - `info.Quality() = q` will set the compression quality (e.g. for saving JPEGs)
+   to `q`.
+   * `q` should take values between `0` and `100`.
+   * The quality value is ignored unless calling `data::Save()` with `info`.
+
+ - Calling `info.Channels() = 1` before loading will cause images to be loaded
+   in grayscale.
+
+ - Metadata stored in the `data::ImageInfo` can be accessed with the following
+   members:
+   * `info.Width()` returns the image width in pixels.
+   * `info.Height()` returns the image height in pixels.
+   * `info.Channels()` returns the number of color channels in the image.
+   * `info.Quality()` returns the compression quality that will be used to save
+     images (between 0 and 100).
+
+---
+
+### Loading images
+
+With a `data::ImageInfo` object, image data can be loaded or saved, handling
+either one or multiple images at a time:
+
+<!-- TODO: add parameter to force use of what's in `info` -->
+
+ - `data::Load(filename, matrix, info, fatal=false)`
+   * Load a ***single image*** from `filename` into `matrix`.
+     - Format is chosen by extension (e.g. `image.png` will load as PNG).
+
+   * `matrix` will have one column representing the image as a flattened vector.
+
+   * `info` will be populated with information from the image in `filename`.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown upon load
+     failure.
+
+   * Returns a `bool` indicating the success of the operation.
+
+---
+
+ - `data::Load(files, matrix, info, fatal=false)`
+   * Load ***multiple images*** from `files` into `matrix`.
+     - `files` is of type `std::vector<std::string>` and should contain the list
+       of images to be loaded.
+     - `matrix` will have `files.size()` columns, each representing the
+       corresponding image as a flattened vector.
+
+   * `info` will be populated with information from the images in `files`.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown if any files
+     fail to load.
+
+   * Returns a `bool` indicating the success of the operation.
+
+---
+
+ - `data::Save(filename, matrix, info, fatal=false)`
+   * Save a ***single image*** from `matrix` into the file `filename`.
+     - Format is chosen by extension (e.g. `image.png` will save as PNG).
+
+   * `matrix` is expected to have only one column representing the image as a
+     flattened vector.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown in the event of
+     save failure.
+
+   * Returns a `bool` indicating the success of the operation.
+
+---
+
+ - `data::Save(files, matrix, info, fatal=false)`
+   * Save ***multiple images*** from `matrix` into `files`.
+     - `files` is of type `std::vector<std::string>` and should contain the list
+       of files to save to.
+     - The format of each file is chosen by extension (e.g. `image.png` will
+       save as PNG); it is allowed for filenames in `files` to have different
+       extensions.
+
+   * `matrix` is expected to have `files.size()` columns representing images as
+     flattened vectors.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown if any images
+     fail to save.
+
+   * Returns a `bool` indicating the success of the operation.
+
+---
+
+Images are flattened along rows, with channel values interleaved, starting from
+the top left.  Thus, the value of the pixel at position `(x, y)` in channel `c`
+will be contained in element/row `y * (width * channels) + x * (channels) + c`
+of the flattened vector.
+
+Pixels take values between 0 and 255.
+
+---
+
+Example of loading and saving a single image:
+
+```c++
+// See https://www.mlpack.org/static/img/numfocus-logo.png.
+mlpack::data::ImageInfo info;
+arma::mat matrix;
+mlpack::data::Load("numfocus-logo.png", matrix, info, true);
+
+// `matrix` should now contain one column.
+
+// Print information about the image.
+std::cout << "Information about the image in 'numfocus-logo.png': "
+    << std::endl;
+std::cout << " - " << info.Width() << " pixels in width." << std::endl;
+std::cout << " - " << info.Height() << " pixels in height." << std::endl;
+std::cout << " - " << info.Channels() << " color channels." << std::endl;
+
+std::cout << "Value at pixel (x=3, y=4) in the first channel: ";
+const size_t index = (4 * info.Width() * info.Channels()) +
+    (3 * info.Channels());
+std::cout << matrix[index] << "." << std::endl;
+
+// Increment each pixel value, but make sure they are still within the bounds.
+matrix += 1;
+matrix = arma::clamp(matrix, 0, 255);
+
+mlpack::data::Save("numfocus-logo-mod.png", matrix, info);
+```
+
+---
+
+Example of loading and saving multiple images:
+
+```c++
+// Load some favicons from websites associated with mlpack.
+std::vector<std::string> images;
+// See the following files:
+// - https://datasets.mlpack.org/images/mlpack-favicon.png
+// - https://datasets.mlpack.org/images/ensmallen-favicon.png
+// - https://datasets.mlpack.org/images/armadillo-favicon.png
+// - https://datasets.mlpack.org/images/bandicoot-favicon.png
+images.push_back("mlpack-favicon.png");
+images.push_back("ensmallen-favicon.png");
+images.push_back("armadillo-favicon.png");
+images.push_back("bandicoot-favicon.png");
+
+mlpack::data::ImageInfo info;
+info.Channels() = 1; // Force loading in grayscale.
+
+arma::mat matrix;
+mlpack::data::Load(images, matrix, info, true);
+
+// Print information about what we loaded.
+std::cout << "Loaded " << matrix.n_cols << " images.  Images are of size "
+    << info.Width() << " x " << info.Height() << " with " << info.Channels()
+    << " color channel." << std::endl;
+
+// Invert images.
+matrix = (255.0 - matrix);
+
+// Save as compressed JPEGs with low quality.
+info.Quality() = 75;
+std::vector<std::string> outImages;
+outImages.push_back("mlpack-favicon-inv.jpeg");
+outImages.push_back("ensmallen-favicon-inv.jpeg");
+outImages.push_back("armadillo-favicon-inv.jpeg");
+outImages.push_back("bandicoot-favicon-inv.jpeg");
+
+mlpack::data::Save(outImages, matrix, info);
+```
+
+## mlpack objects
+
+All mlpack objects can be saved with `data::Save()` and loaded with
+`data::Load()`.  Serialization is performed using the
+[cereal](https://uscilab.github.io/cereal/) serialization toolkit.
+Each object must be given a logical name.
+
+ - `data::Load(filename, name, object, fatal=false, format=data::format::autodetect)`
+ - `data::Save(filename, name, object, fatal=false, format=data::format::autodetect)`
+   * Load/save `object` to/from `filename` with the logical name `name`.
+
+   * If `fatal` is `true`, a `std::runtime_error` will be thrown in the event of
+     load or save failure.
+
+   * The format is autodetected based on extension (`.bin`, `.json`, or `.xml`),
+     but can be manually specified:
+     - `data::format::binary`: binary blob (smallest and fastest).  No checks;
+       assumes all data is correct.
+     - `data::format::json`: JSON.
+     - `data::format::xml`: XML (largest and slowest).
+
+   * For JSON and XML types, when loading, `name` must match the name used to
+     save the object.
+
+   * Returns a `bool` indicating the success of the operation.
+
+***Note:*** when loading an object that was saved as a binary blob, the C++ type
+of the object must be ***exactly the same*** (including template parameters) as
+the type used to save the object.  If not, undefined behavior will occur---most
+likely a crash.
+
+---
+
+Simple example: create a `math::Range` object, then save and load it.
+
+```c++
+mlpack::math::Range r(3.0, 6.0);
+
+// Save the Range to 'range.bin', using the name "range".
+mlpack::data::Save("range.bin", "range", r, true);
+
+// Load the range into a new object.
+mlpack::math::Range r2;
+mlpack::data::Load("range.bin", "range", r2, true);
+
+std::cout << "Loaded range: [" << r2.Lo() << ", " << r2.Hi() << "]."
+    << std::endl;
+
+// Modify and save the range as JSON.
+r2.Lo() = 4.0;
+mlpack::data::Save("range.json", "range", r2, true);
+
+// Now 'range.json' will contain the following:
+//
+// {
+//     "range": {
+//         "cereal_class_version": 0,
+//         "hi": 6.0,
+//         "lo": 4.0
+//     }
+// }
+```
+
+---
+
+## Normalizing labels
+
+mlpack classifiers and other algorithms require labels to be in the range `0` to
+`numClasses - 1`.  A vector of labels with arbitrary (`size_t`) values can be
+normalized to the required range with the `NormalizeLabels()` function.
+
+---
+
+ * `data::NormalizeLabels(labelsIn, labelsOut, mappings)`
+    - Map vector `labelsIn` into the range `0` to `numClasses - 1`, storing as
+      `labelsOut` (of type `arma::Row<size_t>`).
+      * `numClasses` is automatically detected using the number of unique values
+        in `labelsIn`.
+
+    - The column vector `mappings` will be filled with the reverse mappings to
+      convert back to the old labels; this can be used by `RevertLabels()`.
+
+    - `mappings[i]` contains the original class label for the mapped label `i`.
+
+---
+
+ * `data::RevertLabels(labelsIn, mappings, labelsOut)`
+    - Unmap normalized labels `labelsIn` using `mappings` into `labelsOut`.
+
+    - Performs the reverse operation of `NormalizeLabels()`; `mappings` should
+      be the same vector output by `NormalizeLabels()`.
+
+---
+
+Simple example: convert labels into `0`, `1`, `2`, learn a model, then convert
+predictions back to the original label values.
+
+```c++
+// Create a random dataset with 5 points in 10 dimensions.
+arma::mat dataset(10, 5, arma::fill::randu);
+
+// Manually assemble labels vector: [3, 7, 3, 3, 5]
+arma::Row<size_t> labels = { 3, 7, 3, 3, 5 };
+
+// Note that these labels are not in the range `0` to `2`, and thus cannot be
+// used directly by mlpack classifiers!
+// We will map them to that range using NormalizeLabels().
+arma::Row<size_t> mappedLabels;
+arma::Col<size_t> mappings;
+mlpack::data::NormalizeLabels(labels, mappedLabels, mappings);
+const size_t numClasses = mappedLabels.max() + 1;
+
+// Print the mapped values:
+// [3, 7, 3, 3, 5] maps to [0, 1, 0, 0, 2].
+// The `mappings` vector will be [3, 7, 5].
+std::cout << "Original labels: " << labels;
+std::cout << "Mapped labels:   " << mappedLabels;
+std::cout << "Mappings: " << mappings;
+
+// Learn a model with the mapped labels.
+mlpack::DecisionTree d(dataset, mappedLabels, numClasses, 1 /* leaf size */);
+
+// Make predictions on the training dataset.
+arma::Row<size_t> mappedPredictions;
+d.Classify(dataset, mappedPredictions);
+
+// The predictions use mapped labels (0, 1, 2), which we will need to map back
+// to the original labels using RevertLabels().
+arma::Row<size_t> predictions;
+mlpack::data::RevertLabels(mappedPredictions, mappings, predictions);
+
+// Print the predictions before and after unmapping.
+// The mapped predictions will take values 0, 1, or 2; the predictions will take
+// values 3, 7, or 5 (like the original data).
+std::cout << "Mapped predictions: " << mappedPredictions;
+std::cout << "Predictions:        " << predictions;
+```
+
+## Formats
+
+mlpack's `data::Load()` and `data::Save()` functions support a variety of
+different formats in different contexts.
+
+---
+
+#### [Numeric data](#numeric-data)
+
+By default, load/save format is ***autodetected***, but can be manually
+specified with the `format` parameter using one of the options below:
+
+ - `FileType::AutoDetect` (default): auto-detects the format as one of the
+   formats below using the extension of the filename and inspecting the file
+   contents.
+
+ - `FileType::CSVASCII` (autodetect extensions `.csv`, `.tsv`): CSV format
+   with no header.  If loading a sparse matrix and the CSV has three columns,
+   the data is interpreted as a
+   [coordinate list](https://arma.sourceforge.net/docs.html#save_load_mat).
+
+ - `FileType::RawASCII` (autodetect extensions `.csv`, `.txt`):
+   space-separated values or tab-separated values (TSV) with no header.
+
+ - `FileType::ArmaASCII` (autodetect extension `.txt`): space-separated
+   values as saved by Armadillo with the
+   [`arma_ascii`](https://arma.sourceforge.net/docs.html#save_load_mat)
+   format.
+
+ - `FileType::CoordASCII` (autodetect extensions `.txt`, `.tsv`; must be
+   loading a sparse matrix type): coordinate list format for sparse data (see
+   [`coord_ascii`](https://arma.sourceforge.net/docs.html#save_load_mat)).
+
+ - `FileType::ArmaBinary` (autodetect extension `.bin`): Armadillo's
+   efficient binary matrix format
+   ([`arma_binary`](https://arma.sourceforge.net/docs.html#save_load_mat)).
+
+ - `FileType::HDF5Binary` (autodetect extensions `.h5`, `.hdf5`, `.hdf`,
+  `.he5`): [HDF5](https://en.wikipedia.org/wiki/Hierarchical_Data_Format)
+   binary format; only available if Armadillo is configured with
+   [HDF5 support](https://arma.sourceforge.net/docs.html#config_hpp).
+
+ - `FileType::RawBinary` (autodetect extension `.bin`): packed binary data
+   with no header and no size information; data will be loaded as a single
+   column vector _(not recommended)_.
+
+ - `FileType::PGMBinary` (autodetect extension `.pgm`): PGM image format
+
+***Notes:***
+
+   - ASCII formats (`CSVASCII`, `RawASCII`, `ArmaASCII`) are human-readable but
+     large; to reduce dataset size, consider a binary format such as
+      `ArmaBinary` or `HDF5Binary`.
+   - Sparse data (`arma::sp_mat`, `arma::sp_fmat`, etc.) should be saved in a
+     binary format (`ArmaBinary` or `HDF5Binary`) or as a coordinate list
+     (`CoordASCII`).
+
+---
+
+#### [Mixed categorical data](#mixed-categorical-data)
+
+The format of mixed categorical data is detected automatically based on the
+file extension and inspecting the file contents:
+
+ - `.csv`, `.txt`, or `.tsv` indicates CSV/TSV/ASCII format
+ - `.arff` indicates [ARFF](https://ml.cms.waikato.ac.nz/weka/arff.html)
+
+---
+
+#### [Image data](#image-data)
+
+The format of images are detected automatically based on the file extension.
+
+ - The following formats are supported for loading: `.jpg`, `.jpeg`, `.png`,
+   `.tga`, `.bmp`, `.psd`, `.gif`, `.hdr`, `.pic`, `.pnm`
+
+ - The following formats are supported for saving: `.jpg`, `.png`, `.tga`,
+   `.bmp`, `.hdr`
+
+---
+
+#### [mlpack objects](#mlpack-objects)
+
+By default, load/save format for mlpack objects is autodetected, but can be
+manually specified with the `format` parameter using one of the options below:
+
+ - `format::autodetect` (default): auto-detects the format as one of the
+    formats below using the extension of the filename
+ - `format::json` (autodetect extension `.json`)
+ - `format::xml` (autodetect extension `.xml`)
+ - `format::binary` (autodetect extension `.bin`)
+
+***Notes:***
+
+ - `format::json` (`.json`) and `format::xml` (`.xml`) produce human-readable
+   files, but they may be quite large.
+ - `format::binary` (`.bin`) is recommended for the sake of size; objects in
+   binary format may be an order of magnitude or more smaller than JSON!
diff -pruN 4.3.0-2/doc/user/matrices.md 4.4.0-1/doc/user/matrices.md
--- 4.3.0-2/doc/user/matrices.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/user/matrices.md	2024-05-28 20:08:34.000000000 +0000
@@ -1,50 +1,95 @@
 # Matrices in mlpack
 
-mlpack uses Armadillo matrices for matrix support.  Armadillo is a fast C++
-matrix library which makes use of advanced template techniques to provide the
-fastest possible matrix operations.
+mlpack uses Armadillo matrices for linear algebra support.  Armadillo is a fast
+C++ matrix library which uses advanced template metaprogramming techniques to
+provide the fastest possible linear algebra operations.
 
-Documentation on Armadillo can be found on [the Armadillo
-website](http://arma.sourceforge.net/docs.html).
+<center><p><img src="https://arma.sourceforge.net/img/armadillo_logo2.png" alt="Armadillo logo"></p></center>
 
-Nonetheless, there are a few further caveats for mlpack Armadillo usage.
+Detailed documentation on Armadillo can be found on [the Armadillo
+website](https://arma.sourceforge.net/docs.html).
 
-## Column-major matrices
+Nonetheless, there are a few further caveats for mlpack Armadillo usage.
 
-Armadillo matrices are stored in a column-major format; this means that on disk,
-each column is located in contiguous memory.
+ * [An Armadillo primer](#an-armadillo-primer)
+ * [Representing data in mlpack](#representing-data-in-mlpack)
+ * [Loading data](#loading-data)
+ * [Loading and using categorical data](#loading-and-using-categorical-data)
+ * [Alternate matrix types](#alternate-matrix-types)
+ * [Adapting from other toolkits (Eigen, etc.)](#adapting-from-other-toolkits-eigen-etc)
+
+## An Armadillo primer
+
+The Armadillo syntax is straightforward and is aimed at ease-of-use and
+readability.  To give a flavor of what a linear algebra program using Armadillo
+looks like, see the trivial (contrived) program below that performs some basic
+matrix operations.
+
+```c++
+// Create a 10x15 matrix with random elements.
+arma::mat m(10, 15, arma::fill::randu);
+
+std::cout << "Size of m: " << m.n_rows << " x " << m.n_cols << "." << std::endl;
+
+// Sum all elements in the matrix.
+const double sumVal = arma::accu(m);
+std::cout << "Sum of all elements: " << sumVal << "." << std::endl;
+
+// Sum the elements in each column.
+arma::rowvec sums = arma::sum(m, 0);
+std::cout << "Sums in each column: " << sums;
 
-This means that, for the vast majority of machine learning methods, it is faster
-to store observations as columns and dimensions as rows.  This is counter to
-most standard machine learning texts!
+// Add 1 to all elements.
+m += 1;
 
-Major implications of this are for linear algebra.  For instance, the covariance
-of a matrix is typically
+// Subtract sums from each row.
+m.each_row() -= sums;
 
-```
-C = X^T X
+// Print an individual element.
+std::cout << "m(3, 4) is: " << m(3, 4) << "." << std::endl;
 ```
 
-but for a column-wise matrix, it is
+For more information on Armadillo, see the following resources:
 
-```
-C = X X^T
-```
+ * [Armadillo documentation](https://arma.sourceforge.net/docs.html)
+ * [Armadillo example
+    program](https://arma.sourceforge.net/docs.html#example_prog)
+ * [Armadillo/MATLAB syntax conversion
+   table](https://arma.sourceforge.net/docs.html#syntax)
+
+## Representing data in mlpack
+
+Armadillo matrices, unlike numpy and some other toolkits, store data in a
+***column-major*** format.  This means that each column is located in contiguous
+memory; i.e., `x(0, 0)` is adjacent to `x(1, 0)` in memory.
 
-and this is very important to keep in mind!  If your mlpack code is not working,
-this may be a factor in why.
-
-## Loading matrices
+This means that, for the vast majority of machine learning methods, it is faster
+to store ***observations as columns*** and ***dimensions as rows***.  This is
+counter to most standard machine learning texts!  It also has some implications
+for linear algebra operations; for instance, computing the Gram matrix of a
+matrix `X` is typically expressed as `X^T X`, but when using column-major
+matrices, the expression must be `X X^T`.
+
+In general, the following Armadillo types are commonly used inside mlpack:
+
+ * `arma::mat`: datasets and general-purpose matrices
+ * `arma::Row<size_t>`: integer response data, e.g., labels for classification
+   datasets
+ * `arma::rowvec`: floating-point response data, e.g., responses for regression
+   datasets
+ * `arma::vec`: general-purpose column vectors
+ * `arma::sp_mat`, `arma::fmat`: alternate types for representing data; see
+   [Alternate matrix types](#alternate-matrix-types)
+
+## Loading data
 
-mlpack provides a `data::Load()` and `data::Save()` function, which should be
-used instead of Armadillo's loading and saving functions.
+mlpack provides two simple functions for loading and saving data matrices in a
+column-major form:
 
-Most machine learning data is stored in row-major format; a CSV, for example,
-will generally have one observation per line and each column will correspond to
-a dimension.
+ * `data::Load(filename, matrix, fatal=false, transpose=true, type=FileType::AutoDetect)` ([full documentation](load_save.md#numeric-data))
+ * `data::Save(filename, matrix, fatal=false, transpose=true, type=FileType::AutoDetect)` ([full documentation](load_save.md#numeric-data))
 
-The `data::Load()` and `data::Save()` functions transpose the matrix upon
-loading, meaning that the following CSV:
+As an example, consider the following CSV file:
 
 ```sh
 $ cat data.csv
@@ -63,6 +108,284 @@ $ cat data.csv
 2,4,4,2,0
 ```
 
-is actually loaded with 5 rows and 13 columns, not 13 rows and 5 columns like
-the CSV is written.  More information on mlpack's loading functionality can be
-found in [the formats tutorial](formats.md).
+The following program will load the data, print information about it, and save a
+modified dataset to disk.
+
+```c++
+// Load data from `data.csv` into `m`.  Throw an exception on failure (i.e. set
+// `fatal` to `true`).
+arma::mat m;
+mlpack::data::Load("data.csv", m, true);
+
+// Since mlpack uses column-major data,
+//
+//  - each column corresponds to a data point!
+//  - each row corresponds to a dimension!
+//
+std::cout << "The matrix in 'data.csv' has: " << std::endl;
+std::cout << " - " << m.n_cols << " points." << std::endl;
+std::cout << " - " << m.n_rows << " dimensions." << std::endl;
+
+std::cout << "The second point in the dataset: " << std::endl;
+std::cout << m.col(1).t();
+
+// Now modify the matrix and save to a different format (space-separated
+// values).
+m += 3;
+mlpack::data::Save("data-mod.txt", m);
+```
+
+Although Armadillo does provide a `.load()` and `.save()` member function for
+matrices, the `data::Load()` and `data::Save()` functions offer additional
+flexibility, and ensure that data is saved and loaded in a column-major format.
+
+## Loading and using categorical data
+
+Some mlpack techniques support mixed categorical data, e.g., data where some
+dimensions take only categorical values (e.g. `0`, `1`, `2`, etc.).  String data
+and other non-numerical data can be represented as categorical values, and
+mlpack has support to load mixed categorical data:
+
+ * The `data::DatasetInfo` auxiliary class stores information about whether each
+   dimension is numeric or categorical.  ([full
+   documentation](load_save.md#datadatasetinfo))
+ * `data::Load(filename, matrix, info, fatal=false, transpose=true)` ([full
+   documentation](load_save.md#loading-categorical-data))
+
+For example, consider the following CSV file that contains strings:
+
+```sh
+$ cat mixed_string_data.csv
+3,"hello",3,"f",0
+3,"goodbye",4,"f",0
+3,"goodbye",4,"e",0
+3,"hello",4,"d",0
+3,"hello",4,"d",0
+2,"hello",4,"d",0
+2,"hello",4,"d",0
+3,"goodbye",3,"f",0
+3,"goodbye",4,"f",0
+3,"hello",4,"f",0
+3,"hello",4,"c",0
+3,"hello",4,"f",0
+2,"hello",4,"c",0
+```
+
+The following program will load the data file, print information about
+categorical dimensions, and prepare the data for use with an mlpack algorithm
+that supports mixed categorical data.
+
+```c++
+// Load data from `mixed_string_data.csv` into `m`.  Throw an exception on
+// failure (i.e. set `fatal` to `true`).  This populates the `info` object.
+arma::mat m;
+mlpack::data::DatasetInfo info;
+mlpack::data::Load("mixed_string_data.csv", m, info, true);
+
+// Print information about the data.
+std::cout << "The matrix in 'mixed_string_data.csv' has: " << std::endl;
+std::cout << " - " << m.n_cols << " points." << std::endl;
+std::cout << " - " << info.Dimensionality() << " dimensions." << std::endl;
+
+// Print which dimensions are categorical.
+for (size_t d = 0; d < info.Dimensionality(); ++d)
+{
+  if (info.Type(d) == mlpack::data::Datatype::categorical)
+  {
+    std::cout << " - Dimension " << d << " is categorical with "
+        << info.NumMappings(d) << " distinct categories." << std::endl;
+  }
+}
+
+// Modify the third point to be 4,"wonderful",1,"c",0.
+// Note that we manually map the string values; MapString() returns the category
+// for a given value.
+m(0, 2) = 4;
+m(1, 2) = info.MapString<double>("wonderful", 1); // Create new third category.
+m(2, 2) = 1;
+m(3, 2) = info.MapString<double>("c", 1);
+m(4, 2) = 0;
+
+// `m` can now be used with any mlpack algorithm that supports categorical data.
+```
+
+Not every mlpack method supports categorical data.  Below are the list of
+methods that do have categorical data support:
+
+ * [`DecisionTree`](methods/decision_tree.md)
+ * [`DecisionTreeRegressor`](methods/decision_tree_regressor.md)
+ * [`RandomForest`](methods/random_forest.md)
+ * [`HoeffdingTree`](methods/hoeffding_tree.md)
+
+## Alternate matrix types
+
+mlpack's documentation focuses on the use of the `arma::mat`, `arma::vec`, and
+`arma::rowvec` types, with an underlying `double` numeric type, e.g., 64-bit
+floating-point.  But, many mlpack algorithms and support utilities have support
+for alternate matrix types and element types:
+
+ * Many methods, such as
+   [`LogisticRegression`](methods/logistic_regression.md#advanced-functionality-different-element-types),
+   allow specifying the matrix types as a template parameter.
+
+ * Some methods, such as
+   [`DecisionTree`](methods/decision_tree.md#using-different-element-types),
+   accept different matrix types to `Train()`, `Classify()`, or `Predict()`,
+   without needing to specify an explicit template parameter.
+
+ * In general, any matrix type that supports the Armadillo API can be used; this
+   includes:
+    - Single-precision floating point matrices (`arma::fmat`, `arma::frowvec`,
+      `arma::fvec`)
+    - Sparse matrices (`arma::sp_mat`, `arma::sp_fmat`)
+    - GPU matrices via [Bandicoot](https://coot.sourceforge.io) (`coot::mat`,
+      `coot::fmat`) --- ***(note: support is under development and still
+       experimental)***
+
+---
+
+A simple example of using single-precision floating point data to train an
+[`AdaBoost`](methods/adaboost.md) model is below.
+
+```c++
+// 1000 random points in 10 dimensions, using 32-bit precision (float).
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor, using floating-point data.
+// The weak learner type is now a floating-point Perceptron.
+typedef mlpack::Perceptron<
+    mlpack::SimpleWeightUpdate,
+    mlpack::ZeroInitialization,
+    arma::fmat> PerceptronType;
+mlpack::AdaBoost<PerceptronType, arma::fmat> ab(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+ab.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 3) << " test points classified as class "
+    << "3." << std::endl;
+```
+
+---
+
+A simple example of using sparse 32-bit floating point data to train a
+[`LogisticRegression`](methods/logistic_regression.md) model is below.
+
+```c++
+// Create random, sparse 100-dimensional data.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 1));
+
+// Train with L2 regularization penalty parameter of 0.1.
+mlpack::LogisticRegression<arma::sp_fmat> lr(dataset, labels, 0.1);
+
+// Now classify a test point.
+arma::sp_fvec point;
+point.sprandu(100, 1, 0.3);
+
+size_t prediction;
+arma::fvec probabilitiesVec;
+lr.Classify(point, prediction, probabilitiesVec);
+
+std::cout << "Prediction for random test point: " << prediction << "."
+    << std::endl;
+std::cout << "Class probabilities for random test point: "
+    << probabilitiesVec.t();
+```
+
+<!-- TODO: a simple Bandicoot example! -->
+
+## Adapting from other toolkits (Eigen, etc.)
+
+In general, C++ linear algebra toolkits store data in a column-major
+representation, and transitioning between toolkits is a matter of getting access
+to the underlying memory.
+
+---
+
+Copy an [Eigen](https://eigen.tuxfamily.org/index.php?title=Main_Page) matrix
+into an Armadillo matrix.
+
+```c++
+// Note: this will only work if the Eigen matrix is stored in column-major
+// order.  See https://eigen.tuxfamily.org/dox/group__TopicStorageOrders.html
+// for more details.
+Eigen::MatrixXd m;
+const size_t rows = 10;
+const size_t cols = 20;
+m.setRandom(rows, cols); // 10x20 random matrix.
+
+// Copy into an Armadillo matrix.
+arma::mat mCopy(&m(0, 0), rows, cols);
+```
+
+---
+
+Copy an [XTensor](https://xtensor.readthedocs.io/en/latest/) matrix into an
+Armadillo matrix.
+
+```c++
+// Note: this will only work correctly if the layout_type of the XTensor matrix
+// is column-major (i.e. xt::layout_type::column_major).  See
+// https://xtensor.readthedocs.io/en/latest/container.html for more details.
+const size_t rows = 10;
+const size_t cols = 20;
+
+// Create a random 10 x 20 matrix with normally distributed values.
+// Note that we must ensure that the matrix is laid out in column-major form.
+xt::xarray<double, xt::layout_type::column_major> m =
+    xt::random::randn<double>({ rows, cols });
+
+// Copy into an Armadillo matrix.
+arma::mat mCopy(m.data(), rows, cols);
+```
+
+---
+
+Make an Armadillo matrix that is an alias of an
+[Eigen](https://eigen.tuxfamily.org/index.php?title=Main_Page) matrix.  Note
+that changes to the Eigen matrix will be reflected in the Armadillo matrix, and
+vice versa.
+
+*If the Eigen matrix is deallocated, the Armadillo matrix will become invalid.
+Be careful!  [More details
+here](https://arma.sourceforge.net/docs.html#adv_constructors_mat).*
+
+```c++
+// Note: this will only work if the Eigen matrix is stored in column-major
+// order.  See https://eigen.tuxfamily.org/dox/group__TopicStorageOrders.html
+// for more details.
+Eigen::MatrixXd m;
+const size_t rows = 10;
+const size_t cols = 20;
+m.setRandom(rows, cols); // 10x20 random matrix.
+
+// Make an Armadillo matrix that is an alias of the Eigen matrix.  This avoids
+// the copy, but is potentially dangerous: be careful that the Eigen matrix is
+// not deleted while the Armadillo matrix is in use!
+//
+// See https://arma.sourceforge.net/docs.html#adv_constructors_mat
+arma::mat mAlias(&m(0, 0), rows, cols, false, true);
+```
+
+---
+
+Copy an Armadillo matrix to an Eigen matrix.
+
+```c++
+const size_t rows = 10;
+const size_t cols = 20;
+arma::mat m(10, 20, arma::fill::randu);
+
+// Construct the Eigen matrix by using a map of the Armadillo memory.
+Eigen::MatrixXd eigenM(Eigen::Map<Eigen::MatrixXd>(m.memptr(), rows, cols));
+```
diff -pruN 4.3.0-2/doc/user/methods/adaboost.md 4.4.0-1/doc/user/methods/adaboost.md
--- 4.3.0-2/doc/user/methods/adaboost.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/adaboost.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,435 @@
+## `AdaBoost`
+
+The `AdaBoost` class implements the 'adaptive boosting' classifier AdaBoost.MH.
+This classifier is an ensemble of weak learners.  The `AdaBoost` class offers
+control over the weak learners and other behavior via template parameters.  By
+default, the `Perceptron` class is used as a weak learner.
+
+`AdaBoost` is useful for classifying points with _discrete labels_ (i.e. `0`,
+`1`, `2`).
+
+#### Simple usage example:
+
+Train an AdaBoost model on random data and predict labels on a random test set.
+
+```c++
+// Train an AdaBoost model on random data and predict labels on test data:
+
+// All data and labels are uniform random; 10 dimensional data, 5 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::AdaBoost ab;                   // Step 1: create model.
+ab.Train(dataset, labels, 5);          // Step 2: train model.
+arma::Row<size_t> predictions;
+ab.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 3) << " test points classified as class "
+    << "3." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `AdaBoost` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+ * [Advanced template examples](#advanced-functionality-examples) of use with
+   custom template parameters.
+
+#### See also:
+
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [`Perceptron`](perceptron.md)
+ * [`DecisionTree`](decision_tree.md)
+ * [AdaBoost on Wikipedia](https://en.wikipedia.org/wiki/AdaBoost)
+ * [AdaBoost.MH paper (pdf)](https://dl.acm.org/doi/pdf/10.1145/279943.279960)
+
+### Constructors
+
+ * `ab = AdaBoost(tolerance=1e-6)`
+   - Initialize model without training.
+   - You will need to call [`Train()`](#training) later to train the tree before
+     calling [`Classify()`](#classification).
+
+---
+
+ * `ab = AdaBoost(data, labels, numClasses, maxIterations=100, tolerance=1e-6)`
+   - Train model using default weak learner hyperparameters.
+
+---
+
+ * `ab = AdaBoost(data, labels, numClasses, maxIterations=100, tolerance=1e-6, _[weak learner hyperparameters...]_)`
+   - Train model with custom weak learner hyperparameters.
+   - Hyperparameters for the weak learner are any arguments to the weak
+     learner's `Train()` function that come after `numClasses` or `weights`.
+   - The only hyperparameter for the default weak learner (`Perceptron`) is
+     `maxIterations`.
+   - See [examples of this constructor in use](#simple-examples).
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `weakLearner` | `Perceptron` | An initialized weak learner whose hyperparameters will be used as settings for weak learners during training. | _(N/A)_ |
+| `maxIterations` | `size_t` | Maximum number of iterations of AdaBoost.MH to use.  This is the maximum number of weak learners to train.  (0 means no limit, and weak learners will be trained until the tolerance is met.) | `100` |
+| `tolerance` | `double` | When the weighted residual (`r_t`) of the model goes below `tolerance`, training will terminate and no more weak learners will be added. | `1e-6` |
+
+As an alternative to passing hyperparameters, each hyperparameter can be set
+with a standalone method.  The following functions can be used before calling
+`Train()` to set hyperparameters:
+
+ * `ab.MaxIterations() = maxIter;` will set the maximum number of weak learners
+   during training to `maxIter`.
+ * `ab.Tolerance() = tol;` will set the tolerance to `tol`.
+
+***Note:*** different types of weak learners can be used than
+[`Perceptron`](perceptron.md), by changing the [`WeakLearnerType` template
+parameter](#advanced-functionality-template-parameters).
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the versions of the `Train()` member function.  For an instance of `AdaBoost`
+named `ab`, the following functions for training are available:
+
+ * `ab.Train(data, labels, numClasses, maxIterations=100, tolerance=1e-6)`
+   - Train model using default weak learner parameters.
+
+---
+
+ * `ab.Train(data, labels, numClasses, maxIterations=100, tolerance=1e-6, [weak learner hyperparameters...])`
+   - Train model with custom weak learner parameters.
+   - Hyperparameters for the weak learner are any arguments to the weak
+     learner's `Train()` function that come after `numClasses` or `weights`.
+   - The only hyperparameter for the default weak learner (`Perceptron`) is
+     `maxIterations`.
+   - See [examples of this form in use](#simple-examples).
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   AdaBoost model from scratch.
+
+ * `Train()` returns a `double` indicating an upper bound on the training error
+   (specifically, the product of _Zt_ values, as described in the paper).
+
+### Classification
+
+Once an `AdaBoost` model is trained, the `Classify()` member function can be
+used to make class predictions for new data.
+
+ * `size_t predictedClass = ab.Classify(point)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class.
+
+---
+
+ * `ab.Classify(point, prediction, probabilitiesVec)`
+   - ***(Single-point)***
+   - Classify a single point and compute class probabilities.
+   - The predicted class is stored in `prediction`.
+   - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `ab.Classify(data, predictions)`
+   - ***(Multi-point)***
+   - Classify a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `ab.Classify(data, predictions, probabilities)`
+   - ***(Multi-point)***
+   - Classify a set of points and compute class probabilities for each point.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The probability of class `j` for data point `i` can be accessed with
+     `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into; will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to number of classes; number of columns will be equal to `data.n_cols`). |
+
+### Other Functionality
+
+ * An `AdaBoost` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `ab.NumClasses()` will return a `size_t` indicating the number of classes the
+   model was trained on.
+
+ * `ab.WeakLearners()` will return a `size_t` indicating the number of weak
+   learners that the model currently contains.
+
+ * `ab.Alpha(i)` will return the weight of weak learner `i`.
+
+ * `ab.WeakLearner(i)` will return the `i`th weak learner.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/adaboost/adaboost.hpp) can be consulted.  Each method
+is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `AdaBoost` class.
+
+---
+
+Train an AdaBoost model using the hyperparameters from an existing weak learner.
+
+```c++
+// See https://datasets.mlpack.org/iris.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.csv", dataset, true);
+// See https://datasets.mlpack.org/iris.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("iris.labels.csv", labels, true);
+
+mlpack::AdaBoost ab;
+// Train with a custom number of perceptron iterations, and custom AdaBoost
+// parameters.
+ab.Train(dataset, labels, 3, 75 /* maximum number of weak learners */,
+                             1e-6 /* tolerance for AdaBoost convergence */,
+                             100 /* maximum number of perceptron iterations */);
+
+// Now predict the label of a point and the probabilities of each class.
+size_t prediction;
+arma::rowvec probabilities;
+ab.Classify(dataset.col(10), prediction, probabilities);
+
+std::cout << "Point 11 is predicted to have class " << prediction << "."
+    << std::endl;
+std::cout << "Probabilities of each class: " << probabilities;
+```
+
+---
+
+Before training an AdaBoost model, set hyperparameters individually.  Save the
+trained model to disk.
+
+```c++
+// See https://datasets.mlpack.org/iris.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.csv", dataset, true);
+// See https://datasets.mlpack.org/iris.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("iris.labels.csv", labels, true);
+
+mlpack::AdaBoost ab;
+ab.MaxIterations() = 50; // Use at most 50 weak learners.
+ab.Tolerance() = 1e-4; // Set a custom tolerance for convergence.
+
+// Now train, using the hyperparameters specified above.
+ab.Train(dataset, labels, 3);
+
+// Save the model to `adaboost_model.bin`.
+mlpack::data::Save("adaboost_model.bin", "adaboost_model", ab, true);
+```
+
+---
+
+Load an AdaBoost model and print some information about it.
+
+```c++
+// Load a saved model named "adaboost_model" from `adaboost_model.bin`.
+mlpack::AdaBoost ab;
+mlpack::data::Load("adaboost_model.bin", "adaboost_model", ab, true);
+
+std::cout << "Details about the model in `adaboost_model.bin`:" << std::endl;
+std::cout << "  - Trained on " << ab.NumClasses() << " classes." << std::endl;
+std::cout << "  - Tolerance used for training: " << ab.Tolerance() << "."
+    << std::endl;
+std::cout << "  - Number of perceptron weak learners in model"
+    << ab.WeakLearners() << "." << std::endl;
+
+// Print some details about the first weak learner, if available.  The weak
+// learner type is `Perceptron<>`.
+if (ab.WeakLearners() > 0)
+{
+  std::cout << "  - Weight of first perceptron weak learner: " << ab.Alpha(0)
+      << "." << std::endl;
+  std::cout << "  - Biases of first perceptron learner: "
+      << ab.WeakLearner(0).Biases().t();
+}
+```
+
+---
+
+See also the following fully-working examples:
+
+ - [Graduate admission classification with `AdaBoost`](https://github.com/mlpack/examples/blob/master/graduate_admission_classification_with_Adaboost/graduate-admission-classification-with-adaboost-cpp.ipynb)
+
+### Advanced Functionality: Template Parameters
+
+The `AdaBoost` class has two template parameters that can be used for custom
+behavior.  The full signature of the class is:
+
+```
+AdaBoost<WeakLearnerType, MatType>
+```
+
+ * `WeakLearnerType`: the weak classifier to ensemble in the AdaBoost model.
+ * `MatType`: specifies the type of matrix used for learning and internal
+   representation of model parameters.
+
+---
+
+#### `WeakLearnerType`
+
+ * Specifies the weak learner to use when constructing an AdaBoost model.
+ * The default `WeakLearnerType` is [`Perceptron<>`](perceptron.md).
+ * The `ID3DecisionStump` class (a custom variant of
+   [`DecisionTree`](decision_tree.md)) is available for drop-in usage as a weak
+   learner.
+ * Any custom variant of `Perceptron<>` or `DecisionTree<>` can be used; e.g.,
+   `Perceptron<SimpleWeightUpdate, RandomPerceptronInitialization>`.
+ * A custom class must implement the following functions for training and
+   classification.  Note that this is the same API as mlpack's classifiers that
+   support instance weights for learning, and so any mlpack classifier
+   supporting instance weights can also be used.
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomWeakLearner
+{
+ public:
+  // Train the model with the given hyperparameters.
+  //
+  //  * `MatType` will be an Armadillo-like matrix type (typically `arma::mat`);
+  //    this is the same type as the `MatType` template parameter for the
+  //    `AdaBoost` class.
+  //
+  //  * `data` and `labels` are the same dataset passed to the `Train()` method
+  //    of `AdaBoost`.
+  //
+  //  * `weights` contains instance weights for each point (column) of `data`.
+  //
+  // Note: there is no restriction on the number or types of hyperparameters
+  // that can be used, but, they do need default arguments.  The example here
+  // includes two.
+  template<typename MatType>
+  void Train(const MatType& data,
+             const arma::Row<size_t>& labels,
+             const size_t numClasses,
+             const arma::rowvec& weights,
+             const size_t hyperparameterA = 10,
+             const double hyperparameterB = 0.1);
+
+  // Classify the given point.  `VecType` will be an Armadillo-like type that is
+  // a vector that represents a single point.
+  template<typename VecType>
+  size_t Classify(const VecType& point);
+
+  // Classify the given points in `data`, storing the predicted classifications
+  // in `predictions`.
+  template<typename MatType>
+  void Classify(const MatType& data, arma::Row<size_t>& predictions);
+};
+```
+
+---
+
+#### `MatType`
+
+ * Specifies the matrix type to use for data when learning a model (or
+   predicting with one).
+ * By default, `MatType` is `arma::mat` (dense 64-bit precision matrix).
+ * Any matrix type implementing the Armadillo API will work; so, for instance,
+   `arma::fmat` or `arma::sp_mat` can also be used.
+
+### Advanced Functionality Examples
+
+Train an AdaBoost model using decision stumps as the weak learner (use a
+different `WeakLearnerType`).
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::mat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.
+// Note that we specify decision stumps as the weak learner type, and pass
+// hyperparameters for the decision stump (these could be omitted).  See the
+// DecisionTree documentation for more details on the ID3DecisionStump-specific
+// hyperparameters.
+mlpack::AdaBoost<mlpack::ID3DecisionStump> ab(dataset, labels, 5,
+    25 /* maximum number of decision stumps */,
+    1e-6 /* tolerance for convergence of AdaBoost */,
+    /** Hyperparameters specific to ID3DecisionStump: **/
+    10 /* minimum number of points in each leaf of the decision stump */,
+    1e-5 /* minimum gain for splitting the root node of the decision stump */);
+
+// Create test data (500 points).
+arma::mat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+ab.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 3) << " test points classified as class "
+    << "3." << std::endl;
+```
+
+---
+
+Train an AdaBoost model on 32-bit floating-point precision data (use a different
+`MatType`).
+
+```c++
+// 1000 random points in 10 dimensions, using 32-bit precision (float).
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor, using floating-point data.
+// The weak learner type is now a floating-point Perceptron.
+typedef mlpack::Perceptron<mlpack::SimpleWeightUpdate,
+                           mlpack::ZeroInitialization,
+                           arma::fmat> PerceptronType;
+mlpack::AdaBoost<PerceptronType, arma::fmat> ab(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+ab.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 3) << " test points classified as class "
+    << "3." << std::endl;
+```
diff -pruN 4.3.0-2/doc/user/methods/bayesian_linear_regression.md 4.4.0-1/doc/user/methods/bayesian_linear_regression.md
--- 4.3.0-2/doc/user/methods/bayesian_linear_regression.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/bayesian_linear_regression.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,311 @@
+## `BayesianLinearRegression`
+
+The `BayesianLinearRegression` class implements a Bayesian ridge regression
+model for numerical data that optimally tunes the regularization strength to the
+given data.  The class offers configurable functionality and template parameters
+to control the data type used for storing the model.
+
+#### Simple usage example:
+
+```c++
+// Train a Bayesian linear regression model on random data and make predictions.
+
+// All data and responses are uniform random; this uses 10 dimensional data.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::rowvec responses = arma::randn<arma::rowvec>(1000);
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::BayesianLinearRegression blr;  // Step 1: create model.
+blr.Train(dataset, responses);         // Step 2: train model.
+arma::rowvec predictions;
+blr.Predict(testDataset, predictions); // Step 3: use model to predict.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions > 0.6) << " test points predicted to have"
+    << " responses greater than 0.6." << std::endl;
+std::cout << arma::accu(predictions < 0) << " test points predicted to have "
+    << "negative responses." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `BayesianLinearRegression` objects.
+ * [`Train()`](#training): train model.
+ * [`Predict()`](#prediction): predict with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [mlpack regression techniques](../../index.md#regression-algorithms)
+ * [`LinearRegression`](linear_regression.md)
+ * [`LARS`](lars.md)
+ * [Bayesian linear regression on Wikipedia](https://en.wikipedia.org/wiki/Bayesian_linear_regression)
+
+### Constructors
+
+ * `blr = BayesianLinearRegression(centerData=true, scaleData=false, maxIterations=50, tolerance=1e-4)`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Predict()`](#prediction).
+
+---
+
+ * `blr = BayesianLinearRegression(data, responses)`
+ * `blr = BayesianLinearRegression(data, responses, centerData=true, scaleData=false, maxIterations=50, tolerance=1e-4)`
+   - Train model on the given data.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `responses` | [`arma::rowvec`](../matrices.md) | Training responses (e.g. values to predict).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `centerData` | `bool` | Whether to center the data before learning. | `true` |
+| `scaleData` | `bool` | Whether to scale the data to unit variance before learning. | `false` |
+| `maxIterations` | `size_t` | Maximum number of iterations for convergence. | `50` |
+| `tolerance` | `double` | Tolerance for convergence of the model. | `1e-4` |
+
+As an alternative to passing `centerData`, `scaleData`, `maxIterations`, or
+`tolerance`, they can each be set or accessed with standalone methods:
+
+ * `blr.CenterData() = centerData;` will set whether to center the data before
+   learning to `centerData`.
+ * `blr.ScaleData() = scaleData;` will set whether to scale the data to unit
+   variance before learning to `scaleData`.
+ * `blr.MaxIterations() = maxIterations;` will set the maximum number of
+   iterations to `maxIterations`.
+ * `blr.Tolerance() = tolerance;` will set the tolerance for convergence to
+   `tolerance`.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `blr.Train(data, responses, centerData=true, scaleData=false, maxIterations=50, tolerance=1e-4)`
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes:***
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   model from scratch.
+
+ * `Train()` returns the root mean squared error (RMSE) of the model on the
+   training set as a `double`.
+
+### Prediction
+
+Once a `LinearRegression` model is trained, the `Predict()` member function
+can be used to make predictions for new data.
+
+ * `double predictedValue = blr.Predict(point)`
+   - ***(Single-point)***
+   - Make a prediction for a single point, returning the predicted value.
+
+---
+ * `blr.Predict(point, prediction, stddev)`
+   - ***(Single-point)***
+   - Make a prediction for a single point, storing the predicted value in
+     `prediction` and the standard deviation of the prediction in `stddev`.
+
+---
+
+ * `blr.Predict(data, predictions)`
+   - ***(Multi-point)***
+   - Make predictions for a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `blr.Predict(data, predictions, stddevs)`
+   - ***(Multi-point)***
+   - Make predictions for a set of points and compute standard deviations of
+     predictions.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The standard deviation of the prediction for data point `i` can be accessed
+     with `stddevs[i]`.
+
+---
+
+#### Prediction Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for prediction. |
+| _single-point_ | `prediction` | `double&` | `double` to store predicted value into. |
+| _single-point_ | `stddev` | `double&` | `double` to store standard deviation of predicted value into. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::rowvec&`](../matrices.md) | Vector of `double`s to store predictions into.  Will be set to length `data.n_cols`. |
+| _multi-point_ | `stddevs` | [`arma::rowvec&`](../matrices.md) | Vector of `double`s to store standard deviations of predictions into.  Will be set to length `data.n_cols`. |
+
+### Other Functionality
+
+ * A `BayesianLinearRegression` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * After training is complete, the following methods can be used to inspect the
+   model:
+   - `blr.Omega()` returns the weights of the trained model as an
+     `const arma::vec&` of length `data.n_rows`.  The weight for the `i`th
+     dimension can be accessed with `blr.Omega()[i]`.
+
+   - `blr.Alpha()` returns the precision (or inverse variance) of the Gaussian
+     prior of the model as a `double`.
+
+   - `blr.Beta()` returns the precision (or inverse variance) of the model as a
+     `double`.
+
+   - `blr.Variance()` returns the estimated variance as a `double`.
+
+   - `blr.DataOffset()` returns a `const arma::vec&` containing the mean values
+     of the training data in each dimension.  The vector has length
+     `data.n_rows`.  The result is only meaningful if `centerData` is `true`.
+
+   - `blr.DataScale()` returns a `const arma::vec&` containing the standard
+     deviations of the training data in each dimension.  The vector has length
+     `data.n_rows`.  The result is only meaningful if `scaleData` is `true`.
+
+   - `blr.ResponsesOffset()` returns the mean value of the training responses as
+     a `double`.  This is the intercept of the model.
+
+ * `blr.RMSE(data, responses)` returns a `double` containing the RMSE (root mean
+   squared error) of the model on the given `data` and `responses`.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `BayesianLinearRegression` class.
+
+---
+
+Train a Bayesian linear regression model in the constructor on weighted data,
+compute the RMSE with `RMSE()`, and save the model.
+
+```c++
+// See https://datasets.mlpack.org/admission_predict.csv.
+arma::mat data;
+mlpack::data::Load("admission_predict.csv", data, true);
+
+// See https://datasets.mlpack.org/admission_predict.responses.csv.
+arma::rowvec responses;
+mlpack::data::Load("admission_predict.responses.csv", responses, true);
+
+// Generate random instance weights for each point, in the range 0.5 to 1.5.
+arma::rowvec weights(data.n_cols, arma::fill::randu);
+weights += 0.5;
+
+// Train Bayesian linear regression model.  The data will be both centered and
+// scaled to have unit variance.
+mlpack::BayesianLinearRegression blr(data, responses, true, true);
+
+// Now compute the RMSE on the training set.
+std::cout << "RMSE on the training set: " << blr.RMSE(data, responses)
+    << "." << std::endl;
+
+// Finally, save the model with the name "blr".
+mlpack::data::Save("blr_model.bin", "blr", blr, true);
+```
+
+---
+
+Load a saved Bayesian linear regression model and print some information about
+it, then make some predictions individually for random points.
+
+```
+mlpack::BayesianLinearRegression blr;
+
+// Load the model named "blr" from "lr_model.bin".
+mlpack::data::Load("blr_model.bin", "blr", blr, true);
+
+// Print some information about the model.
+const size_t dimensionality = blr.Omega().n_elem;
+if (dimensionality == 0)
+{
+  std::cout << "The model in `blr_model.bin` has not been trained."
+      << std::endl;
+  return 0;
+}
+
+std::cout << "Information on the BayesianLinearRegression model in "
+    << "'blr_model.bin':" << std::endl;
+std::cout << " - Data was centered when training: "
+    << (blr.CenterData() ? std::string("yes") : std::string("no")) << "."
+    << std::endl;
+std::cout << " - Data was scaled to unit variance when training: "
+    << (blr.ScaleData() ? std::string("yes") : std::string("no")) << "."
+    << std::endl;
+std::cout << " - Model intercept: " << blr.ResponsesOffset() << "."
+    << std::endl;
+std::cout << " - Precision of Gaussian prior: " << blr.Alpha() << "."
+    << std::endl;
+std::cout << " - Precision of model: " << blr.Beta() << "." << std::endl;
+
+// Now make a prediction for three random points.
+for (size_t t = 0; t < 3; ++t)
+{
+  arma::vec randomPoint(dimensionality, arma::fill::randu);
+  double prediction, stddev;
+  blr.Predict(randomPoint, prediction, stddev);
+
+  std::cout << "Prediction for random point " << t << ": " << prediction
+      << " +/- " << stddev << "." << std::endl;
+}
+```
+
+---
+
+### Advanced Functionality: Different Element Types
+
+The `BayesianLinearRegression` class has one template parameter that can be used
+to control the element type of the model.  The full signature of the class is:
+
+```
+BayesianLinearRegression<ModelMatType>
+```
+
+`ModelMatType` specifies the type of matrix used for the internal representation
+of model parameters.  Any matrix type that implements the Armadillo API can be
+used; however, the matrix should be dense, as in general
+`BayesianLinearRegression` will produce models that are not sparse.
+
+The example below trains a Bayesian linear regression model on 32-bit floating
+point data.
+
+```c++
+// Create random, sparse 100-dimensional data.
+arma::fmat dataset(100, 5000, arma::fill::randu);
+
+// Generate noisy responses from random data.
+arma::fvec trueWeights(100, arma::fill::randu);
+arma::frowvec responses = trueWeights.t() * dataset +
+    0.01 * arma::randu<arma::frowvec>(5000) /* noise term */;
+
+mlpack::BayesianLinearRegression<arma::fmat> blr;
+blr.ScaleData() = true;
+blr.MaxIterations() = 75;
+
+blr.Train(dataset, responses);
+
+// Compute the RMSE on the training set and a random test set.
+arma::fmat testDataset(100, 1000, arma::fill::randu);
+
+arma::frowvec testResponses = trueWeights.t() * testDataset +
+    0.01 * arma::randu<arma::frowvec>(1000) /* noise term */;
+
+std::cout << "RMSE on training set: "
+    << blr.RMSE(dataset, responses) << "." << std::endl;
+std::cout << "RMSE on test set:     "
+    << blr.RMSE(testDataset, testResponses) << "." << std::endl;
+```
diff -pruN 4.3.0-2/doc/user/methods/decision_tree.md 4.4.0-1/doc/user/methods/decision_tree.md
--- 4.3.0-2/doc/user/methods/decision_tree.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/decision_tree.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,586 @@
+## `DecisionTree`
+
+The `DecisionTree` class implements a decision tree classifier that supports
+numerical and categorical features, by default using Gini gain to choose which
+feature to split on.  The class offers several template parameters and several
+runtime options that can be used to control the behavior of the tree.
+
+Decision trees are useful for classifying points with _discrete labels_ (i.e.
+`0`, `1`, `2`).  For predicting _continuous values_ (regression), see
+[`DecisionTreeRegressor`](decision_tree_regressor.md).
+
+#### Simple usage example:
+
+```c++
+// Train a decision tree on random numeric data and predict labels on test data:
+
+// All data and labels are uniform random; 10 dimensional data, 5 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::DecisionTree tree;               // Step 1: create model.
+tree.Train(dataset, labels, 5);          // Step 2: train model.
+arma::Row<size_t> predictions;
+tree.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `DecisionTree` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+
+#### See also:
+
+ * [`DecisionTreeRegressor`](decision_tree_regressor.md)
+ * [Random forests](random_forest.md)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [Decision tree on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree)
+ * [Decision tree learning on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+
+### Constructors
+
+ * `tree = DecisionTree()`
+   - Initialize tree without training.
+   - You will need to call [`Train()`](#training) later to train the tree before
+     calling [`Classify()`](#classification).
+
+---
+
+ * `tree = DecisionTree(data, labels, numClasses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree = DecisionTree(data, labels, numClasses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on numerical-only data (optionally with instance weights).
+
+---
+
+ * `tree = DecisionTree(data, datasetInfo, labels, numClasses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree = DecisionTree(data, datasetInfo, labels, numClasses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `datasetInfo` | [`data::DatasetInfo`](../load_save.md#loading-categorical-data) | Dataset information, specifying type information for each dimension. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `weights` | [`arma::rowvec`](../matrices.md) | Weights for each training point.  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `minLeafSize` | `size_t` | Minimum number of points in each leaf node. | `10` |
+| `minGainSplit` | `double` | Minimum gain for a node to split. | `1e-7` |
+| `maxDepth` | `size_t` | Maximum depth for the tree. (0 means no limit.) | `0` |
+
+ * Setting `minLeafSize` too small (e.g. `1`) may cause the tree to overfit to
+   its training data, and may create a very large tree.  However, setting it too
+   large may cause the tree to be very small and underfit.
+ * `minGainSplit` has similar behavior: if it is too small, the tree may
+   overfit; if too large, it may underfit.
+
+***Note:*** different types can be used for `data` and `weights` (e.g.,
+`arma::fmat`, `arma::sp_mat`).  However, the element type of `data` and
+`weights` must match; for example, if `data` has type `arma::fmat`, then
+`weights` must have type `arma::frowvec`.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the following versions of the `Train()` member function:
+
+ * `tree.Train(data, labels, numClasses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree.Train(data, labels, numClasses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on numerical-only data (optionally with instance weights).
+
+---
+
+ * `tree.Train(data, datasetInfo, labels, numClasses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree.Train(data, datasetInfo, labels, numClasses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   decision tree from scratch.
+
+ * `Train()` returns a `double` with the final gain of the tree (the Gini gain,
+   unless a different
+   [`FitnessFunction` template parameter](#fully-custom-behavior) is specified.
+
+### Classification
+
+Once a `DecisionTree` is trained, the `Classify()` member function can be used
+to make class predictions for new data.
+
+ * `size_t predictedClass = tree.Classify(point)`
+    - ***(Single-point)***
+    - Classify a single point, returning the predicted class.
+
+---
+
+ * `tree.Classify(point, prediction, probabilitiesVec)`
+    - ***(Single-point)***
+    - Classify a single point and compute class probabilities.
+    - The predicted class is stored in `prediction`.
+    - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `tree.Classify(data, predictions)`
+    - ***(Multi-point)***
+    - Classify a set of points.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `tree.Classify(data, predictions, probabilities)`
+    - ***(Multi-point)***
+    - Classify a set of points and compute class probabilities for each point.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+    - The probability of class `j` for data point `i` can be accessed with
+      `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into.  Will be set to length `numClasses`. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into.  Will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to number of classes, number of columns will be equal to `data.n_cols`). |
+
+***Note:*** different types can be used for `data` and `point` (e.g.
+`arma::fmat`, `arma::sp_mat`, `arma::sp_vec`, etc.).  However, the element type
+that is used should be the same type that was used for training.
+
+### Other Functionality
+
+ * A `DecisionTree` can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `tree.NumChildren()` will return a `size_t` indicating the number of children
+   in the node `tree`.
+
+ * `tree.Child(i)` will return a `DecisionTree` object representing the `i`th
+   child of the node `tree`.
+
+ * `tree.SplitDimension()` returns a `size_t` indicating which dimension the
+   node `tree` splits on.
+
+ * `tree.NumClasses()` returns a `size_t` indicating the number of classes the
+   tree was trained on.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/decision_tree/decision_tree.hpp) can be consulted.
+Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`DecisionTree`.
+
+---
+
+Train a decision tree on mixed categorical data and save it:
+
+```c++
+// Load a categorical dataset.
+arma::mat dataset;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/covertype.train.arff.
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+
+arma::Row<size_t> labels;
+// See https://datasets.mlpack.org/covertype.train.labels.csv.
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+
+// Create the tree.
+mlpack::DecisionTree tree;
+// Train on the given dataset, specifying a minimum leaf size of 5.
+tree.Train(dataset, info, labels, 7 /* classes */, 5 /* minimum leaf size */);
+
+// Load categorical test data.
+arma::mat testDataset;
+// See https://datasets.mlpack.org/covertype.test.arff.
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+
+// Predict class of first test point.
+const size_t firstPrediction = tree.Classify(testDataset.col(0));
+std::cout << "Predicted class of first test point is " << firstPrediction << "."
+    << std::endl;
+
+// Predict class and probabilities of second test point.
+size_t secondPrediction;
+arma::vec secondProbabilities;
+tree.Classify(testDataset.col(1), secondPrediction, secondProbabilities);
+std::cout << "Class probabilities of second test point: " <<
+    secondProbabilities.t();
+
+// Save the tree to `tree.bin`.
+mlpack::data::Save("tree.bin", "tree", tree);
+```
+
+---
+
+Load a tree and print some information about it.
+
+```c++
+mlpack::DecisionTree tree;
+// This call assumes a tree called "tree" has already been saved to `tree.bin`
+// with `data::Save()`.
+mlpack::data::Load("tree.bin", "tree", tree, true);
+
+if (tree.NumChildren() > 0)
+{
+  std::cout << "The split dimension of the root node of the tree in `tree.bin` "
+      << "is dimension " << tree.SplitDimension() << "." << std::endl;
+}
+else
+{
+  std::cout << "The tree in `tree.bin` is a leaf (it has no children)."
+      << std::endl;
+}
+```
+
+---
+
+See also the following fully-working examples:
+
+ - [Loan default prediction with `DecisionTree`](https://github.com/mlpack/examples/blob/master/loan_default_prediction_with_decision_tree/loan-default-prediction-with-decision-tree-cpp.ipynb)
+
+### Advanced Functionality: Template Parameters
+
+#### Using different element types.
+
+`DecisionTree`'s constructors, `Train()`, and `Classify()` functions support
+any data type, so long as it supports the Armadillo matrix API.  So, for
+instance, learning can be done on single-precision floating-point data:
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.
+mlpack::DecisionTree tree(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+tree.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+
+---
+
+#### Fully custom behavior.
+
+The `DecisionTree` class also supports several template parameters, which can
+be used for custom behavior during learning.  The full signature of the class is
+as follows:
+
+```
+DecisionTree<FitnessFunction,
+             NumericSplitType,
+             CategoricalSplitType,
+             DimensionSelectionType,
+             NoRecursion>
+```
+
+ * `FitnessFunction`: the measure of goodness to use when deciding on tree
+   splits
+ * `NumericSplitType`: the strategy used for finding splits on numeric data
+   dimensions
+ * `CategoricalSplitType`: the strategy used for finding splits on categorical
+   data dimensions
+ * `DimensionSelectionType`: the strategy used for proposing dimensions to
+   attempt to split on
+ * `NoRecursion`: a boolean indicating whether to build a tree or a stump
+   (one level tree)
+
+Below, details are given for the requirements of each of these template types.
+
+---
+
+#### `FitnessFunction`
+
+ * Specifies the fitness function to use when learning a decision tree.
+ * The `GiniGain` _(default)_ and `InformationGain` classes are available for
+   drop-in usage.
+ * A custom class must implement three functions:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomFitnessFunction
+{
+  // Return the range (difference between maximum and minimum gain values).
+  double Range(const size_t numClasses);
+
+  // Compute the gain for the given vector of labels, where `labels[i]` has an
+  // associated instance weight `weights[i]`.
+  //
+  // `RowType` and `WeightVecType` will be vector types following the Armadillo
+  // API.  If `UseWeights` is `false`, then the `weights` vector should be
+  // ignored (e.g. the labels are not weighted).
+  template<bool UseWeights, typename RowType, typename WeightVecType>
+  double Evaluate(const RowType& labels,
+                  const size_t numClasses,
+                  const WeightVecType& weights);
+
+  // Compute the gain for the given counted set of labels, where `counts[i]`
+  // contains the number of points with label `i`.  There are `totalCount`
+  // labels total, and `counts` has length `numClasses`.
+  //
+  // `UseWeights` is ignored, and `CountType` will be an integral type (e.g.
+  // `size_t`).
+  template<bool UseWeights, typename CountType>
+  double EvaluatePtr(const CountType* counts,
+                     const size_t numClasses,
+                     const CountType totalCount);
+};
+```
+
+---
+
+#### `NumericSplitType`
+
+ * Specifies the strategy to be used during training when splitting a numeric
+   feature.
+ * The `BestBinaryNumericSplit` _(default)_ class is available for drop-in
+   usage and finds the best binary (two-way) split among all possible binary
+   splits.
+ * The `RandomBinaryNumericSplit` class is available for drop-in usage and
+   will select a split randomly between the minimum and maximum values of a
+   dimension.  It is very efficient but does not yield splits that maximize
+   the gain.  (Used by the `ExtraTrees` variant of
+   [`RandomForest`](random_forest.md).)
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal
+   structure `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+template<typename FitnessFunction>
+class CustomNumericSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one dimension
+  // of the input data).
+  template<bool UseWeights, typename VecType, typename WeightVecType>
+  double SplitIfBetter(const double bestGain,
+                       const VecType& data,
+                       const arma::Row<size_t>& labels,
+                       const size_t numClasses,
+                       const WeightVecType& weights,
+                       const size_t minLeafSize,
+                       const double minGainSplit,
+                       arma::vec& splitInfo,
+                       AuxiliarySplitInfo& aux);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with value `point`, and split information `splitInfo` and
+  // `aux`, return the index of the child that corresponds to the point.  So,
+  // e.g., if the split type was a binary split on the value `splitInfo`, you
+  // might return `0` if `point < splitInfo`, and `1` otherwise.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+---
+
+#### `CategoricalSplitType`
+
+ * Specifies the strategy to be used during training when splitting a
+    categorical feature.
+ * The `AllCategoricalSplit` _(default)_ is available for drop-in usage and
+   splits all categories into their own node.
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal structure
+   `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+template<typename FitnessFunction>
+class CustomCategoricalSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one
+  // categorical dimension of the input data, which takes values between `0` and
+  // `numCategories - 1`).
+  template<bool UseWeights, typename VecType, typename LabelsType,
+           typename WeightVecType>
+  static double SplitIfBetter(
+      const double bestGain,
+      const VecType& data,
+      const size_t numCategories,
+      const LabelsType& labels,
+      const size_t numClasses,
+      const WeightVecType& weights,
+      const size_t minLeafSize,
+      const double minGainSplit,
+      arma::vec& splitInfo,
+      AuxiliarySplitInfo& aux);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with (categorical) value `point`, and split information
+  // `splitInfo` and `aux`, return the index of the child that corresponds to
+  // the point.  So, e.g., for `AllCategoricalSplit`, which splits a categorical
+  // dimension into one child for each category, this simply returns `point`.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+---
+
+#### `DimensionSelectionType`
+
+ * When splitting a decision tree, `DimensionSelectionType` proposes possible
+   dimensions to try splitting on.
+ * `AllDimensionSplit` _(default)_ is available for drop-in usage and proposes
+   all dimensions for splits.
+ * `MultipleRandomDimensionSelect` proposes a different random subset of
+   dimensions at each decision tree node.
+    - By default each random subset is of size `sqrt(d)` where `d` is the number
+      of dimensions in the data.
+    - If constructed as `MultipleRandomDimensionSelect(n)` and passed to the
+      constructor of `DecisionTree` or the `Train()` function, each random
+      subset will be of size `n`.
+ * Each `DecisionTree` [constructor](#constructors) and each version of the
+   [`Train()`](#training) function optionally accept an instantiated
+   `DimensionSelectionType` object as the very last parameter (after
+   `maxDepth`), in case some internal state in the dimension selection mechanism
+   is required.
+ * A custom class must implement three simple functions:
+
+```c++
+class CustomDimensionSelect
+{
+ public:
+  // Get the first dimension to try.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Begin();
+
+  // Get the next dimension to try.  Note that internal state can be used to
+  // track which candidate dimension is currently being looked at.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Next();
+
+  // Get a value indicating that all dimensions have been tried.
+  size_t End() const;
+
+  // The usage pattern of `DimensionSelectionType` by `DecisionTree` is as
+  // follows, assuming that `dim` is an instantiated `DimensionSelectionType`
+  // object:
+  //
+  // for (size_t dim = dim.Begin(); dim != dim.End(); dim = dim.Next())
+  // {
+  //   // ... try to split on dimension `dim` ...
+  // }
+};
+```
+
+---
+
+#### `NoRecursion`
+
+ * A `bool` value that indicates whether a decision tree should be
+   constructed recursively.
+ * If `true`, only the root node will be split (producing a decision
+   stump).
+ * If `false` _(default)_, a full decision tree will be built.
diff -pruN 4.3.0-2/doc/user/methods/decision_tree_regressor.md 4.4.0-1/doc/user/methods/decision_tree_regressor.md
--- 4.3.0-2/doc/user/methods/decision_tree_regressor.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/decision_tree_regressor.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,588 @@
+## `DecisionTreeRegressor`
+
+The `DecisionTreeRegressor` class implements a decision tree regressor that
+supports numerical and categorical features, by default using MSE (minimum
+squared error) to choose which feature to split on.  The class offers several
+template parameters and runtime options that can be used to control the behavior
+of the tree.
+
+The `DecisionTreeRegressor` class is useful for regressions; i.e., predicting
+_continuous values_ (`0.3`, `1.2`, etc.).  For predicting _discrete labels_
+(classification), see [`DecisionTree`](decision_tree.md).
+
+#### Simple usage example:
+
+Train a decision tree regressor on random numeric data and make predictions on a
+test set:
+
+```c++
+// Train a decision tree regressor on random numeric data and make predictions.
+
+// All data and responses are uniform random; this uses 10 dimensional data.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::rowvec responses = arma::randn<arma::rowvec>(1000);
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::DecisionTreeRegressor tree;     // Step 1: create tree.
+tree.Train(dataset, responses);         // Step 2: train model.
+arma::rowvec predictions;
+tree.Predict(testDataset, predictions); // Step 3: use model to predict.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions > 0.7) << " test points predicted to have"
+    << " responses greater than 0.7." << std::endl;
+std::cout << arma::accu(predictions < 0) << " test points predicted to have "
+    << "negative responses." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `DecisionTreeRegressor` objects.
+ * [`Train()`](#training): train model.
+ * [`Predict()`](#prediction): predict values with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+
+#### See also:
+
+ * [`DecisionTree`](decision_tree.md)
+ * [Random forests](random_forest.md)
+ * [mlpack regression techniques](../../index.md#regression-algorithms)
+ * [Decision tree on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree)
+ * [Decision tree learning on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree_learning)
+
+### Constructors
+
+ * `tree = DecisionTreeRegressor()`
+   - Initialize tree without training.
+   - You will need to call [`Train()`](#training) later to train the tree before
+     calling [`Predict()`](#prediction).
+
+---
+
+ * `tree = DecisionTreeRegressor(data, responses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree = DecisionTreeRegressor(data, responses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on numerical-only data (optionally with instance weights).
+
+---
+
+ * `tree = DecisionTreeRegressor(data, datasetInfo, responses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree = DecisionTreeRegressor(data, datasetInfo, responses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+#### Constructor parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `datasetInfo` | [`data::DatasetInfo`](../load_save.md#loading-categorical-data) | Dataset information, specifying type information for each dimension. | _(N/A)_ |
+| `responses` | [`arma::rowvec`](../matrices.md) | Training responses (e.g. values to predict).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `weights` | [`arma::rowvec`](../matrices.md) | Weights for each training point.  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `minLeafSize` | `size_t` | Minimum number of points in each leaf node. | `10` |
+| `minGainSplit` | `double` | Minimum gain for a node to split. | `1e-7` |
+| `maxDepth` | `size_t` | Maximum depth for the tree. (0 means no limit.) | `0` |
+
+ * Setting `minLeafSize` too small (e.g. `1`) may cause the tree to overfit to
+   its training data, and may create a very large tree.  However, setting it too
+   large may cause the tree to be very small and underfit.
+ * `minGainSplit` has similar behavior: if it is too small, the tree may
+   overfit; if too large, it may underfit.
+
+***Note:*** different types can be used for `data`, `responses`, and `weights`
+(e.g., `arma::fmat`, `arma::sp_mat`).  However, the element type of `data`,
+`responses`, and `weights` all must match; for example, if `data` has type
+`arma::fmat`, then `responses` and `weights` must have type `arma::frowvec`.
+
+### Training
+
+If training is not done as a part of the constructor call, it can be done with
+one of the following versions of the `Train()` member function:
+
+ * `tree.Train(data, responses,          minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+ * `tree.Train(data, responses, weights, minLeafSize=10, minGainSplit=1e-7, maxDepth=0)`
+   - Train on numerical-only data (optionally with instance weights).
+
+---
+
+ * `tree.Train(data, datasetInfo, responses)`
+ * `tree.Train(data, datasetInfo, responses, weights)`
+ * `tree.Train(data, datasetInfo, responses,          minLeafSize, minGainSplit, maxDepth)`
+ * `tree.Train(data, datasetInfo, responses, weights, minLeafSize, minGainSplit, maxDepth)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   decision tree from scratch.
+
+ * `Train()` returns a `double` with the final gain of the tree (the Gini gain,
+   unless a different
+   [`FitnessFunction` template parameter](#fully-custom-behavior) is specified.
+
+### Prediction
+
+Once a `DecisionTreeRegressor` is trained, the `Predict()` member function can
+be used to make class predictions for new data.
+
+ * `double predictedValue = tree.Predict(point)`
+   - ***(Single-point)***
+   - Predict and return the value for a single point.
+
+---
+
+ * `tree.Predict(data, predictions)`
+   - ***(Multi-point)***
+   - Predict and return values for every point in the given matrix `data`.
+   - The predictions for each point are stored in `predictions`, which is set to
+     length `data.n_cols`.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+#### Prediction Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for prediction. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for prediction. |
+| _multi-point_ | `predictions` | [`arma::rowvec&`](../matrices.md) | Vector to store predictions into. |
+
+***Note:*** different types can be used for `data` and `point` (e.g.
+`arma::fmat`, `arma::sp_mat`, `arma::sp_vec`, etc.).  However, the element type
+that is used should be the same type that was used for training.
+
+### Other Functionality
+
+ * A `DecisionTreeRegressor` can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `tree.NumChildren()` will return a `size_t` indicating the number of children
+   in the node `tree`.
+
+ * `tree.NumLeaves()` will return the total number of leaf nodes that are
+   descendants of the node `tree`.
+
+ * `tree.Child(i)` will return a `DecisionTreeRegressor` object representing the
+   `i`th child of the node `tree`.
+
+ * `tree.SplitDimension()` returns a `size_t` indicating which dimension the
+   node `tree` splits on.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/decision_tree/decision_tree_regressor.hpp) can be
+consulted.  Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`DecisionTreeRegressor`.
+
+---
+
+Train a decision tree regressor on mixed categorical data.
+
+```c++
+// Load a categorical dataset.
+arma::mat data;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/telecom_churn.arff.
+mlpack::data::Load("telecom_churn.arff", data, info, true);
+
+arma::rowvec responses;
+// See https://datasets.mlpack.org/telecom_churn.responses.csv.
+mlpack::data::Load("telecom_churn.responses.csv", responses, true);
+
+// Split data into training set (80%) and test set (20%).
+arma::mat trainData, testData;
+arma::rowvec trainResponses, testResponses;
+mlpack::data::Split(data, responses, trainData, testData, trainResponses,
+    testResponses, 0.2);
+
+// Create the tree.
+mlpack::DecisionTreeRegressor tree;
+// Train on the given dataset, specifying a minimum gain of 1e-6 and keeping the
+// default minimum leaf size.
+const double mse = tree.Train(trainData, info, trainResponses,
+    10 /* minimum leaf size */, 1e-6 /* minimum gain */);
+// Print the MSE of the trained tree.
+std::cout << "MSE of trained tree is " << mse << "." << std::endl;
+
+// Compute prediction on the first test point.
+const double firstPrediction = tree.Predict(testData.col(0));
+std::cout << "Predicted value for first test point is " << firstPrediction
+    << "." << std::endl;
+
+// Compute predictions on test data.
+arma::rowvec testPredictions;
+tree.Predict(testData, testPredictions);
+
+// Compute the average error on the test set.
+const double testAverageError = arma::mean(testResponses - testPredictions);
+std::cout << "Average error on test set: " << testAverageError << "."
+    << std::endl;
+```
+
+---
+
+Load a tree and print some information about it.
+
+```c++
+mlpack::DecisionTreeRegressor tree;
+// This call assumes a tree called "tree" has already been saved to `tree.bin`
+// with `data::Save()`.
+mlpack::data::Load("tree.bin", "tree", tree, true);
+
+std::cout << "Information about the DecisionTreeRegressor in `tree.bin`:"
+    << std::endl;
+std::cout << " * The root node has " << tree.NumChildren() << " children."
+    << std::endl;
+std::cout << " * The tree has " << tree.NumLeaves() << " leaves." << std::endl;
+if (tree.NumChildren() > 0)
+{
+  for (size_t i = 0; i < tree.NumChildren(); ++i)
+  {
+    std::cout << " * Child " << i << " of the root has "
+        << tree.Child(i).NumLeaves() << " leaves in its subtree." << std::endl;
+  }
+}
+```
+
+### Advanced Functionality: Template Parameters
+
+#### Using different element types.
+
+`DecisionTreeRegressor`'s constructors, `Train()`, and `Predict()` functions
+support any data type, so long as it supports the Armadillo matrix API.  So, for
+instance, learning can be done on single-precision floating-point data:
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random responses for each point, with a normal distribution.
+arma::frowvec responses = arma::randn<arma::frowvec>(1000);
+
+// Train in the constructor.
+mlpack::DecisionTreeRegressor tree(dataset, responses, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::frowvec predictions;
+tree.Predict(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions > 1) << " test points predicted to have "
+    << "value greater than 1." << std::endl;
+```
+
+---
+
+#### Fully custom behavior.
+
+The `DecisionTreeRegressor` class also supports several template parameters,
+which can be used for custom behavior during learning.  The full signature of
+the class is as follows:
+
+```
+DecisionTreeRegressor<FitnessFunction,
+                      NumericSplitType,
+                      CategoricalSplitType,
+                      DimensionSelectionType,
+                      NoRecursion>
+```
+
+ * `FitnessFunction`: the measure of goodness to use when deciding on tree
+   splits
+ * `NumericSplitType`: the strategy used for finding splits on numeric data
+   dimensions
+ * `CategoricalSplitType`: the strategy used for finding splits on categorical
+   data dimensions
+ * `DimensionSelectionType`: the strategy used for proposing dimensions to
+   attempt to split on
+ * `NoRecursion`: a boolean indicating whether to build a tree or a stump
+   (one level tree)
+
+Below, details are given for the requirements of each of these template types.
+
+---
+
+#### `FitnessFunction`
+
+ * Specifies the fitness function to use when learning a decision tree.
+ * The `MSEGain` _(default)_ and `MADGain` classes are available for drop-in
+   usage.
+ * A custom class must implement three functions:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomFitnessFunction
+{
+  // Compute the gain for the given vector of values, where `values[i]` has an
+  // associated instance weight `weights[i]`.
+  //
+  // `RowType` and `WeightVecType` will be vector types following the Armadillo
+  // API.  If `UseWeights` is `false`, then the `weights` vector should be
+  // ignored (e.g. the responses are not weighted).
+  //
+  // In the version with `begin` and `end` parameters, only the subset between
+  // `labels[begin]` and `labels[end]` (inclusive) should be considered.
+  template<bool UseWeights, typename RowType, typename WeightVecType>
+  double Evaluate(const RowType& labels,
+                  const WeightVecType& weights);
+
+  template<bool UseWeights, typename RowType, typename WeightVecType>
+  double Evaluate(const RowType& labels,
+                  const WeightVecType& weights,
+                  const size_t begin,
+                  const size_t end);
+
+  // Return the output value for prediction for a leaf node whose training
+  // values are made up of the values in the vector `responses` (optionally with
+  // associated instance weights `weights`).
+  //
+  // `ResponsesType` and `WeightsType` will be vector types following the
+  // Armadillo API.  If `UseWeights` is `false`, then the `weights` vector
+  // should be ignored (e.g. the responses are not weighted).
+  template<bool UseWeights, typename ResponsesType, typename WeightsType>
+  double OutputLeafValue(const ResponsesType& responses,
+                         const WeightsType& weights);
+};
+```
+
+***Note:*** this API differs from the `FitnessFunction` API required for
+[`DecisionTree`](decision_tree.md)!
+
+---
+
+#### `NumericSplitType`
+
+ * Specifies the strategy to be used during training when splitting a numeric
+   feature.
+ * The `BestBinaryNumericSplit` _(default)_ class is available for drop-in
+   usage and finds the best binary (two-way) split among all possible binary
+   splits.
+ * The `RandomBinaryNumericSplit` class is available for drop-in usage and
+   will select a split randomly between the minimum and maximum values of a
+   dimension.  It is very efficient but does not yield splits that maximize
+   the gain.  (Used by the `ExtraTrees` variant of
+   [`RandomForest`](random_forest.md).)
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal
+   structure `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+class CustomNumericSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one dimension
+  // of the input data).
+  template<bool UseWeights, typename VecType, typename ResponsesType,
+      typename WeightVecType>
+  double SplitIfBetter(const double bestGain,
+                       const VecType& data,
+                       const ResponsesType& responses,
+                       const WeightVecType& weights,
+                       const size_t minLeafSize,
+                       const double minGainSplit,
+                       arma::vec& splitInfo,
+                       AuxiliarySplitInfo& aux,
+                       FitnessFunction& function);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with value `point`, and split information `splitInfo` and
+  // `aux`, return the index of the child that corresponds to the point.  So,
+  // e.g., if the split type was a binary split on the value `splitInfo`, you
+  // might return `0` if `point < splitInfo`, and `1` otherwise.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+***Note:*** this API differs from the `NumericSplitType` API required for
+[`DecisionTree`](decision_tree.md)!
+
+---
+
+#### `CategoricalSplitType`
+
+ * Specifies the strategy to be used during training when splitting a
+    categorical feature.
+ * The `AllCategoricalSplit` _(default)_ is available for drop-in usage and
+   splits all categories into their own node.
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal
+   structure `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+class CustomCategoricalSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one
+  // categorical dimension of the input data, which takes values between `0` and
+  // `numCategories - 1`).
+  template<bool UseWeights, typename VecType, typename ResponsesType,
+           typename WeightVecType>
+  static double SplitIfBetter(
+      const double bestGain,
+      const VecType& data,
+      const size_t numCategories,
+      const ResponsesType& labels,
+      const WeightVecType& weights,
+      const size_t minLeafSize,
+      const double minGainSplit,
+      arma::vec& splitInfo,
+      AuxiliarySplitInfo& aux,
+      FitnessFunction& fitnessFunction);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with (categorical) value `point`, and split information
+  // `splitInfo` and `aux`, return the index of the child that corresponds to
+  // the point.  So, e.g., for `AllCategoricalSplit`, which splits a categorical
+  // dimension into one child for each category, this simply returns `point`.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+***Note:*** this API differs from the `CategoricalSplitType` API required for
+[`DecisionTree`](decision_tree.md)!
+
+---
+
+#### `DimensionSelectionType`
+
+ * When splitting a decision tree, `DimensionSelectionType` proposes possible
+   dimensions to try splitting on.
+ * `AllDimensionSplit` _(default)_ is available for drop-in usage and proposes
+   all dimensions for splits.
+ * `MultipleRandomDimensionSelect` proposes a different random subset of
+   dimensions at each decision tree node.
+    - By default each random subset is of size `sqrt(d)` where `d` is the number
+      of dimensions in the data.
+    - If constructed as `MultipleRandomDimensionSelect(n)` and passed to the
+      constructor of `DecisionTree<>` or the `Train()` function, each random
+      subset will be of size `n`.
+ * Each `DecisionTreeRegressor` [constructor](#constructors) and each version of
+   the [`Train()`](#training) function optionally accept an instantiated
+   `DimensionSelectionType` object as the very last parameter (after
+   `maxDepth`), in case some internal state in the dimension selection mechanism
+   is required.
+ * A custom class must implement three simple functions:
+
+```c++
+class CustomDimensionSelect
+{
+ public:
+  // Get the first dimension to try.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Begin();
+
+  // Get the next dimension to try.  Note that internal state can be used to
+  // track which candidate dimension is currently being looked at.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Next();
+
+  // Get a value indicating that all dimensions have been tried.
+  size_t End() const;
+
+  // The usage pattern of `DimensionSelectionType` by `DecisionTree` is as
+  // follows, assuming that `dim` is an instantiated `DimensionSelectionType`
+  // object:
+  //
+  // for (size_t dim = dim.Begin(); dim != dim.End(); dim = dim.Next())
+  // {
+  //   // ... try to split on dimension `dim` ...
+  // }
+};
+```
+
+---
+
+#### `NoRecursion`
+
+ * A `bool` value that indicates whether a decision tree should be
+   constructed recursively.
+ * If `true`, only the root node will be split (producing a decision
+   stump).
+ * If `false` _(default)_, a full decision tree will be built.
diff -pruN 4.3.0-2/doc/user/methods/hoeffding_tree.md 4.4.0-1/doc/user/methods/hoeffding_tree.md
--- 4.3.0-2/doc/user/methods/hoeffding_tree.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/hoeffding_tree.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,730 @@
+## `HoeffdingTree`
+
+The `HoeffdingTree` class implements a streaming (or incremental) decision tree
+classifier that supports numerical and categorical features, by default using
+Gini impurity to choose which feature to split on.  The class offers several
+template parameters and several runtime options that can be used to control the
+behavior of the tree.
+
+Hoeffding trees (also known as "Very Fast Decision Trees" or VFDTs) are useful
+for classifying points with _discrete labels_ (i.e.  `0`, `1`, `2`).
+
+#### Simple usage example:
+
+```c++
+// Train a Hoeffding tree on random numeric data; predict labels on test data:
+
+// All data and labels are uniform random; 10 dimensional data, 5 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::HoeffdingTree tree;              // Step 1: create model.
+tree.Train(dataset, labels, 5);          // Step 2a: train model (batch).
+tree.Train(dataset.col(0), labels[0]);   // Step 2b: train model (incremental).
+arma::Row<size_t> predictions;
+tree.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `HoeffdingTree` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+
+#### See also:
+
+ * [`DecisionTree`](decision_tree.md)
+ * [Random forests](random_forest.md)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [Incremental decision tree on Wikipedia](https://en.wikipedia.org/wiki/Incremental_decision_tree)
+ * [Mining High-Speed Data Streams (pdf)](https://dl.acm.org/doi/pdf/10.1145/347090.347107)
+
+### Constructors
+
+ * `tree = HoeffdingTree()`
+   - Initialize tree without training.
+   - You will need to call the batch version of [`Train()`](#training) later to
+     train the tree before calling [`Classify()`](#classification).
+
+---
+
+ * `tree = HoeffdingTree(dimensionality, numClasses)`
+ * `tree = HoeffdingTree(dimensionality, numClasses, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Initialize tree for incremental training on numerical-only data.
+   - The single-point [`Train()`](#training) function can be used to train
+     incrementally.
+
+---
+
+ * `tree = HoeffdingTree(datasetInfo, numClasses)`
+ * `tree = HoeffdingTree(datasetInfo, numClasses, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Initialize tree for incremental training on mixed categorical data.
+   - The single-point [`Train()`](#training) function can be used to train
+     incrementally.
+
+---
+
+ * `tree = HoeffdingTree(data, labels, numClasses)`
+ * `tree = HoeffdingTree(data, labels, numClasses, batchTraining=true, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Train non-incrementally on the given data.
+   - The tree will be reset if `numClasses` or the data's dimensionality does
+     not match the current settings of the tree.
+
+---
+
+ * `tree = HoeffdingTree(data, datasetInfo, labels, numClasses)
+ * `tree = HoeffdingTree(data, datasetInfo, labels, numClasses, batchTraining=true, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Train non-incrementally on mixed categorical data.
+   - The tree will be reset if `numClasses` or `datasetInfo` does not match the
+     current settings of the tree.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `datasetInfo` | [`data::DatasetInfo`](../load_save.md#loading-categorical-data) | Dataset information, specifying type information for each dimension. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `dimensionality` | `size_t` | When using on numeric-only data, this specifies the number of dimensions in the data. | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `batchTraining` | `bool` | If `true`, a batch training algorithm is used, instead of the usual incremental algorithm.  This is generally more efficient for larger datasets. | `true` |
+| `successProbability` | `double` | Probability of success required for Hoeffding bound before a node split can happen. | `0.95` |
+| `maxSamples` | `size_t` | Maximum number of samples before a node split is forced.  `0` means no limit. | `0` |
+| `checkInterval` | `size_t` | Number of samples required before each split check.  Higher values check less often, which is more efficient, but may not split a node as early as possible. | `100` |
+| `minSamples` | `size_t` | Minimum number of samples for a node to see before a split is allowed. | `100` |
+
+As an alternative to passing hyperparameters, these can be set with a
+standalone method.  The following functions can be used before calling
+`Train()`:
+
+ * `tree.SuccessProbability(successProbability);` will set the required success
+   probability for splitting to `successProbability`.
+ * `tree.MaxSamples(maxSamples);` will set the maximum number of samples before
+   a split to `maxSamples`.
+ * `tree.CheckInterval(checkInterval);` will set the number of samples between
+   split checks to `checkInterval`.
+ * `tree.MinSamples(minSamples);` will set the minimum number of samples before
+   a split to `minSamples`.
+
+***Notes:***
+
+ * Setting `successProbability` higher than the default means that the Hoeffding
+   tree is less likely (and will take more samples) to split a node.  This can
+   result in a smaller tree.
+
+ * Different types can be used for `data` (e.g., `arma::fmat`, `arma::sp_mat`).
+   See [template parameters](#advanced-functionality-template-parameters) for
+   using different `NumericSplitType`s that accept different element types.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the following versions of the `Train()` member function:
+
+ * `tree.Train(point, label)`
+   - Streaming (incremental) training: train on a single data point.
+   - The number of classes and dataset information must have already been
+     specified by a previous constructor, `Train()`, or `Reset()` call.
+
+---
+
+ * `tree.Train(data, labels)`
+ * `tree.Train(data, labels, numClasses)`
+ * `tree.Train(data, labels, numClasses, batchTraining=true, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Train on the given data.
+   - If the data is mixed categorical, then `datasetInfo` should have already
+     been passed via a previous constructor, `Train()`, or `Reset()` call.
+   - `numClasses` does not need to be specified if it has been specified in an
+     earlier constructor or `Train()` call.
+
+---
+
+ * `tree.Train(data, datasetInfo, labels)`
+ * `tree.Train(data, datasetInfo, labels, numClasses)`
+ * `tree.Train(data, datasetInfo, labels, numClasses, batchTraining=true, successProbability=0.95, maxSamples=0, checkInterval=100, minSamples=100)`
+   - Train on mixed categorical data.
+   - The previous overload (without `datasetInfo`) can be used instead if
+     `datasetInfo` has already been passed in a previous constructor, `Train()`,
+     or `Reset()` call, and has not changed.
+   - `numClasses` does not need to be specified if it has been specified in an
+     earlier constructor or `Train()` call.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is incremental.  Successive calls to `Train()` will train the
+   Hoeffding tree further.  To reset the tree, call
+   [`Reset()`](#other-functionality).
+
+### Classification
+
+Once a `DecisionTree` is trained, the `Classify()` member function can be used
+to make class predictions for new data.
+
+ * `size_t predictedClass = tree.Classify(point)`
+    - ***(Single-point)***
+    - Classify a single point, returning the predicted class.
+
+---
+
+ * `tree.Classify(point, prediction, probability)`
+    - ***(Single-point)***
+    - Classify a single point and compute class probabilities.
+    - The predicted class is stored in `prediction`.
+    - The probability of class `i` is stored in `probability`.
+
+---
+
+ * `tree.Classify(data, predictions)`
+    - ***(Multi-point)***
+    - Classify a set of points.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `tree.Classify(data, predictions, probabilities)`
+    - ***(Multi-point)***
+    - Classify a set of points and compute class probabilities for each point.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+    - The probability of class `predictions[i]` for data point `i` can be
+      accessed with `probabilities[i]`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probability` | `double&` | `double` to store predicted class probability into. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into.  Will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::rowvec&`](../matrices.md) | Vector to store probability of predicted class in for each point.  Will be set to length `data.n_cols`. |
+
+***Note:*** different types can be used for `data` and `point` (e.g.
+`arma::fmat`, `arma::sp_mat`, `arma::sp_vec`, etc.).  However, the element type
+that is used should be the same type that was used for training.
+
+### Other Functionality
+
+ * A `HoeffdingTree` can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `tree.NumChildren()` will return a `size_t` indicating the number of children
+   in the node `tree`.
+
+ * `tree.NumDescendants()` will return a `size_t` indicating the total number of
+   descendant nodes of the tree.
+
+ * `tree.Child(i)` will return a `HoeffdingTree` object representing the `i`th
+   child of the node `tree`.
+
+ * `tree.SplitDimension()` returns a `size_t` indicating which dimension the
+   node `tree` splits on.
+
+ * `tree.NumSamples()` returns a `size_t` indicating the number of points seen
+   so far by `tree`, if `tree` has not yet split.  If `tree` has split (i.e. if
+   `tree.NumChildren() > 0`), then what is returned is the number of points seen
+   up until the split occurred.
+
+ * `tree.NumClasses()` returns a `size_t` indicating the number of classes the
+   tree was trained on.
+
+ * `tree.Reset()` will reset the tree to an empty tree, and:
+   - `tree.Reset()` will leave the number of classes and dataset information
+     (e.g. `datasetInfo`) intact.
+   - `tree.Reset(dimensionality, numClasses)` will set the number of classes to
+     `numClasses` and set the dimensionality of the data to `dimensionality`,
+     assuming all dimensions are numeric.
+   - `tree.Reset(datasetInfo, numClasses)` will set the number of classes to
+     `numClasses` and set the dataset information to `datasetInfo`.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp) can be consulted.
+Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`HoeffdingTree`.
+
+---
+
+Train a Hoeffding tree incrementally on mixed categorical data:
+
+```c++
+// Load a categorical dataset.
+arma::mat dataset;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/covertype.train.arff.
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+
+arma::Row<size_t> labels;
+// See https://datasets.mlpack.org/covertype.train.labels.csv.
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+
+// Create the tree.
+mlpack::HoeffdingTree tree(info, 7 /* classes */);
+
+// Train on each point in the given dataset.
+for (size_t i = 0; i < dataset.n_cols; ++i)
+  tree.Train(dataset.col(i), labels[i]);
+
+// Load categorical test data.
+arma::mat testDataset;
+// See https://datasets.mlpack.org/covertype.test.arff.
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+
+// Predict class of first test point.
+const size_t firstPrediction = tree.Classify(testDataset.col(0));
+std::cout << "First test point has predicted class " << firstPrediction << "."
+    << std::endl;
+
+// Predict class and probabilities of second test point.
+size_t secondPrediction;
+double secondProbability;
+tree.Classify(testDataset.col(1), secondPrediction, secondProbability);
+std::cout << "Second test point has predicted class " << secondPrediction
+    << " with probability " << secondProbability << "." << std::endl;
+```
+
+---
+
+Train a Hoeffding tree on blocks of a dataset, print accuracy measures on a test
+set during training, and save the model to disk.
+
+```c++
+// Load a categorical dataset.
+arma::mat dataset;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/covertype.train.arff.
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+
+arma::Row<size_t> labels;
+// See https://datasets.mlpack.org/covertype.train.labels.csv.
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+
+// Also load test data.
+
+// See https://datasets.mlpack.org/covertype.test.arff.
+arma::mat testDataset;
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+
+// See https://datasets.mlpack.org/covertype.test.labels.csv.
+arma::Row<size_t> testLabels;
+mlpack::data::Load("covertype.test.labels.csv", testLabels, true);
+
+// Create the tree with custom parameters.
+mlpack::HoeffdingTree tree(info, 7 /* number of classes */);
+tree.SuccessProbability(0.99);
+tree.CheckInterval(500);
+
+// Now iterate over 10k-point chunks in the dataset.
+for (size_t start = 0; start < dataset.n_cols; start += 10000)
+{
+  size_t end = std::min(start + 9999, (size_t) dataset.n_cols - 1);
+
+  tree.Train(dataset.cols(start, end), info, labels.subvec(start, end));
+
+  // Compute accuracy on the test set.
+  arma::Row<size_t> predictions;
+  tree.Classify(testDataset, predictions);
+  const double accuracy = 100.0 * arma::accu(predictions == testLabels) /
+      testLabels.n_elem;
+
+  std::cout << "Accuracy after " << (end + 1) << " points: " << accuracy
+      << "\%." << std::endl;
+}
+
+// Save the fully trained tree in `tree.bin` with name `tree`.
+mlpack::data::Save("tree.bin", "tree", tree, true);
+```
+
+---
+
+Load a tree and print some information about it.
+
+```c++
+mlpack::HoeffdingTree tree;
+// This call assumes a tree called "tree" has already been saved to `tree.bin`
+// with `data::Save()`.
+mlpack::data::Load("tree.bin", "tree", tree, true);
+
+if (tree.NumChildren() > 0)
+{
+  std::cout << "The split dimension of the root node of the tree in `tree.bin` "
+      << "is dimension " << tree.SplitDimension() << "." << std::endl;
+}
+else
+{
+  std::cout << "The tree in `tree.bin` is a leaf (it has no children)."
+      << std::endl;
+}
+```
+
+---
+
+Train a tree, reset a tree, and train again.
+
+```c++
+// See the following files:
+//  - https://datasets.mlpack.org/covertype.train.arff.
+//  - https://datasets.mlpack.org/covertype.train.labels.csv.
+//  - https://datasets.mlpack.org/covertype.test.arff.
+//  - https://datasets.mlpack.org/covertype.test.labels.arff.
+
+arma::mat dataset, testDataset;
+arma::Row<size_t> labels, testLabels;
+mlpack::data::DatasetInfo info;
+
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+mlpack::data::Load("covertype.test.labels.csv", testLabels, true);
+
+// Create a tree, and train on the training data.
+mlpack::HoeffdingTree tree(info, 7 /* number of classes */, 0.98);
+tree.MinSamples(500);
+tree.CheckInterval(500);
+
+tree.Train(dataset, labels);
+
+// Print accuracy on the training and test set.
+arma::Row<size_t> predictions, testPredictions;
+tree.Classify(dataset, predictions);
+tree.Classify(testDataset, testPredictions);
+
+double trainAcc = (100.0 * arma::accu(predictions == labels)) / labels.n_elem;
+double testAcc = (100.0 * arma::accu(testPredictions == testLabels)) /
+    testLabels.n_elem;
+
+std::cout << "When trained on the training data:" << std::endl;
+std::cout << "  - Training set accuracy: " << trainAcc << "\%." << std::endl;
+std::cout << "  - Test set accuracy:     " << testAcc << "\%." << std::endl;
+
+// Now reset the tree, and train on the test set instead.
+// The dataset info and number of classes has not changed, so we can just call
+// Reset() with no arguments.
+tree.Reset();
+tree.Train(testDataset, testLabels);
+
+// Print accuracy on the training and test set, now that we have trained on the
+// test set.
+tree.Classify(dataset, predictions);
+tree.Classify(testDataset, testPredictions);
+
+trainAcc = (100.0 * arma::accu(predictions == labels)) / labels.n_elem;
+testAcc = (100.0 * arma::accu(testPredictions == testLabels)) /
+    testLabels.n_elem;
+
+std::cout << "When trained on the test data:" << std::endl;
+std::cout << "  - Training set accuracy: " << trainAcc << "\%." << std::endl;
+std::cout << "  - Test set accuracy:     " << testAcc << "\%." << std::endl;
+```
+
+---
+
+### Advanced Functionality: Template Parameters
+
+#### Using different element types.
+
+`HoeffdingTree`'s constructors, `Train()`, and `Classify()` functions support
+any data type, so long as it supports the Armadillo matrix API.  So, for
+instance, learning can be done on single-precision floating-point data:
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.
+mlpack::HoeffdingTree tree(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+tree.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+
+---
+
+#### Fully custom behavior.
+
+The `HoeffdingTree` class also supports several template parameters, which can
+be used for custom behavior during learning.  The full signature of the class is
+as follows:
+
+```
+HoeffdingTree<FitnessFunction,
+              NumericSplitType,
+              CategoricalSplitType>
+```
+
+ * `FitnessFunction`: the measure of goodness to use when deciding on tree
+   splits
+ * `NumericSplitType`: the strategy used for finding splits on numeric data
+   dimensions
+ * `CategoricalSplitType`: the strategy used for finding splits on categorical
+   data dimensions
+
+Below, details are given for the requirements of each of these template types.
+
+---
+
+#### `FitnessFunction`
+
+ * Specifies the fitness function to use when learning a decision tree.
+ * The `GiniImpurity` _(default)_ and `HoeffdingInformationGain` classes are
+   available for drop-in usage.
+ * `GiniImpurity` uses the [Gini impurity](https://en.wikipedia.org/wiki/Decision_tree_learning#Gini_impurity),
+   which measures the probability of a randomly labeled random element in the
+   split nodes being correctly labeled.
+ * `HoeffdingInformationGain` uses the [information gain](https://en.wikipedia.org/wiki/Decision_tree_learning#Information_gain),
+   which is based on the information-theoretic entropy of the possible splits.
+ * A custom class must implement two functions:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomFitnessFunction
+{
+  // Return the range (difference between maximum and minimum gain values).
+  double Range(const size_t numClasses);
+
+  // Compute the gain for the given split candidates represented in the matrix
+  // `counts`.  `counts` is a matrix with `numChildren` columns and `numClasses`
+  // rows, containing the number of points for each class held by each child.
+  //
+  // Note that the gain returned should be the gain for *all* child nodes (e.g.
+  // all columns of `counts`).
+  double Evaluate(const arma::Mat<size_t>& counts);
+};
+```
+
+---
+
+#### `NumericSplitType`
+
+ * Specifies the strategy to be used during training when splitting a numeric
+   feature.
+
+ * Several options are already implemented and available for drop-in usage.
+   - The `HoeffdingDoubleNumericSplit` _(default)_ class discretizes the given
+     numeric data into a default of 10 bins.  This expects `double` to be the
+     type of the input data.
+   - The `HoeffdingFloatNumericSplit` class operates similarly to
+     `HoeffdingDoubleNumericSplit`, but expects `float` to be the type of the
+     input data.
+   - The `BinaryNumericSplit` class splits numeric features in two in the way
+     that maximizes gain.  This split type is more computationally expensive
+     during training.
+
+ * If a non-default `NumericSplitType` is specified, the following constructor
+   forms can be used to pass constructed `NumericSplitType`s to the
+   `HoeffdingTree` to use as copy-constructed templates during splitting:
+    - `HoeffdingTree(dimensionality, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(datasetInfo, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(data, labels, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(data, datasetInfo, labels, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement several functions, and have an internal
+   structure `SplitInfo` that is used at classification time:
+
+```c++
+// The job of this class is to track sufficient statistics of training data,
+// returning gain information if a split were to happen according to this
+// class's split strategy.
+//
+// For details, consult the HoeffdingNumericSplit and BinaryNumericSplit class
+// implementations.
+template<typename FitnessFunction>
+class CustomNumericSplit
+{
+ public:
+  // Create the split object with the given number of classes.
+  CustomNumericSplit(const size_t numClasses);
+
+  // Create the split from another split object with the given number of
+  // classes.
+  CustomNumericSplit(const size_t numClasses, const CustomNumericSplit& other);
+
+  // Train on the given value with the given label.
+  // Note that the type used here must match the element type of the training
+  // data (so, e.g., if you plan to use `arma::fmat`, use `float` instead of
+  // `double`).
+  void Train(double value, const size_t label);
+
+  // Given the points seen so far, evaluate the fitness function, returning the
+  // gain if a split were to occur.  If this `NumericSplitType` class could
+  // provide multiple possible splits, also return the second best fitness
+  // value.  (If not, set secondBestFitness to 0.)
+  void EvaluateFitnessFunction(double& bestFitness, double& secondBestFitness);
+
+  // Return the number of children that would be created if a split were to
+  // occur.  (For example, if this class implements a binary split, this should
+  // return 2.)
+  size_t NumChildren() const;
+
+  // Given that a split should happen, return the majority classes of the
+  // children and an initialized SplitInfo object.
+  //
+  // childMajorities should be set to have length equal to the number of
+  // children that this strategy splits into, and the i'th element should be the
+  // majority class label of the i'th child after splitting.
+  void Split(arma::Col<size_t>& childMajorities, SplitInfo& splitInfo);
+
+  // Return the current majority class of points seen so far.
+  size_t MajorityClass() const;
+  // Return the probability of the majority class given the points seen so far.
+  double MajorityProbability() const;
+
+  // Serialize (load/save) the split object using cereal.
+  template<typename Archive>
+  void serialize(Archive& ar, const uint32_t version);
+
+  // The SplitInfo class should implement two functions.  It is used at
+  // prediction time, after a split has occurred, and should contain the
+  // information necessary to classify a point.
+  //
+  // The SplitInfo class must implement two methods; one for classification and
+  // one for serialization.
+  class SplitInfo
+  {
+   public:
+    // Given that the point in the split dimension has the value `value`, return
+    // the index of the child that the traversal should go to.
+    template<typename eT>
+    size_t CalculateDirection(const eT& value) const;
+
+    // Serialize the split (load/save) using cereal.
+    template<typename Archive>
+    void serialize(Archive& ar, const uint32_t version);
+  };
+};
+```
+
+---
+
+#### `CategoricalSplitType`
+
+ * Specifies the strategy to be used during training when splitting a
+    categorical feature.
+
+ * The `HoeffdingCategoricalSplit` _(default)_ is available for drop-in usage
+   and splits all categories into their own node.
+
+ * If a non-default `CategoricalSplitType` is specified, the following
+   constructor forms can be used to pass constructed `CategoricalSplitType`s to
+   the `HoeffdingTree` to use as copy-constructed templates during splitting:
+    - `HoeffdingTree(dimensionality, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(datasetInfo, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(data, labels, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+    - `HoeffdingTree(data, datasetInfo, labels, numClasses, successProbability, maxSamples, checkInterval, minSamples, categoricalSplit, numericSplit)`
+
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement several functions, and have an internal
+   structure `SplitInfo` that is used at classification time:
+
+```c++
+// The job of this class is to track sufficient statistics of training data,
+// returning gain information if a split were to happen according to this
+// class's split strategy.
+//
+// For details, consult the HoeffdingCategoricalSplit class implementation.
+template<typename FitnessFunction>
+class CustomCategoricalSplit
+{
+ public:
+  // Create the split object with the given number of classes.  The dimension
+  // that this object tracks has `numCategories` possible category values.
+  CustomCategoricalSplit(const size_t numCategories, const size_t numClasses);
+
+  // Create the split object from another split object with the given number of
+  // classes.  The dimension that this object tracks has `numCategories`
+  // possible category values.
+  CustomCategoricalSplit(const size_t numCategories, const size_t numClasses,
+                         const CustomCategoricalSplit& other);
+
+  // Train on the given value with the given label.
+  // Note that the type used here must match the element type of the training
+  // data (so, e.g., if you plan to use `arma::fmat`, use `float` instead of
+  // `double`).
+  void Train(double value, const size_t label);
+
+  // Given the points seen so far, evaluate the fitness function, returning the
+  // gain if a split were to occur.  If this `NumericSplitType` class could
+  // provide multiple possible splits, also return the second best fitness
+  // value.  (If not, set secondBestFitness to 0.)
+  void EvaluateFitnessFunction(double& bestFitness, double& secondBestFitness);
+
+  // Return the number of children that would be created if a split were to
+  // occur.  (For example, if this class implements a binary split, this should
+  // return 2.)
+  size_t NumChildren() const;
+
+  // Given that a split should happen, return the majority classes of the
+  // children and an initialized SplitInfo object.
+  //
+  // childMajorities should be set to have length equal to the number of
+  // children that this strategy splits into, and the i'th element should be the
+  // majority class label of the i'th child after splitting.
+  void Split(arma::Col<size_t>& childMajorities, SplitInfo& splitInfo);
+
+  // Return the current majority class of points seen so far.
+  size_t MajorityClass() const;
+  // Return the probability of the majority class given the points seen so far.
+  double MajorityProbability() const;
+
+  // Serialize (load/save) the split object using cereal.
+  template<typename Archive>
+  void serialize(Archive& ar, const uint32_t version);
+
+  // The SplitInfo class should implement two functions.  It is used at
+  // prediction time, after a split has occurred, and should contain the
+  // information necessary to classify a point.
+  //
+  // The SplitInfo class must implement two methods; one for classification and
+  // one for serialization.
+  class SplitInfo
+  {
+   public:
+    // Given that the point in the split dimension has the value `value`, return
+    // the index of the child that the traversal should go to.
+    template<typename eT>
+    size_t CalculateDirection(const eT& value) const;
+
+    // Serialize the split (load/save) using cereal.
+    template<typename Archive>
+    void serialize(Archive& ar, const uint32_t version);
+  };
+};
+```
diff -pruN 4.3.0-2/doc/user/methods/lars.md 4.4.0-1/doc/user/methods/lars.md
--- 4.3.0-2/doc/user/methods/lars.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/lars.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,461 @@
+## `LARS`
+
+The `LARS` class implements the least-angle regression (LARS) algorithm for
+L1-penalized and L2-penalized linear regression.  `LARS` can also solve the
+LASSO (least absolute shrinkage and selection operator) problem.  The LARS
+algorithm is a *path* algorithm, and thus will recover solutions for *all* L1
+penalty parameters greater than or equal to the given L1 penalty parameter.
+
+#### Simple usage example:
+
+```c++
+// Train a LARS model on random numeric data and make predictions.
+
+// All data and responses are uniform random; this uses 10 dimensional data.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::rowvec responses = arma::randn<arma::rowvec>(1000);
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::LARS lars(true, 0.1 /* L1 penalty */); // Step 1: create model.
+lars.Train(dataset, responses);                // Step 2: train model.
+arma::rowvec predictions;
+lars.Predict(testDataset, predictions);        // Step 3: use model to predict.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions > 0.7) << " test points predicted to have"
+    << " responses greater than 0.7." << std::endl;
+std::cout << arma::accu(predictions < 0) << " test points predicted to have "
+    << "negative responses." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `LARS` objects.
+ * [`Train()`](#training): train model.
+ * [`Predict()`](#prediction): predict with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [The LARS path](#the-lars-path): use models from the LARS path with different
+   L1 penalty values.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [`LinearRegression`](linear_regression.md)
+ * [mlpack regression techniques](../../index.md#regression-algorithms)
+ * [Least-angle Regression on Wikipedia](https://en.wikipedia.org/wiki/Least-angle_regression)
+
+### Constructors
+
+ * `lars = LARS(useCholesky=false, lambda1=0.0, lambda2=0.0, tolerance=1e-16, fitIntercept=true, normalizeData=true)`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Predict()`](#prediction).
+
+---
+
+ * `lars = LARS(data, responses, colMajor=true, useCholesky=true, lambda1=0.0, lambda2=0.0, tolerance=1e-16, fitIntercept=true, normalizeData=true)`
+   - Train model on the given data and responses, using the given settings for
+     hyperparameters.
+
+---
+
+ * `lars = LARS(data, responses, colMajor, useCholesky, gramMatrix, lambda1=0.0, lambda2=0.0, tolerance=1e-16, fitIntercept=true, normalizeData=true)`
+   - *(Advanced constructor)*.
+   - Train model on the given data and responses, using a precomputed Gram
+     matrix (`gramMatrix`, equivalent to `data * data.t()`).
+   - Using a precomputed Gram matrix can save time, if it has already been
+     computed.
+   - ***Note:*** any precomputed Gram matrix must also match the settings of
+     `fitIntercept` and `normalizeData`; so, if both are `true`, then
+     `gramMatrix` must be computed on mean-centered data whose features are
+     normalized to have unit variance.  In addition, if `lambda2 > 0`, then
+     it is expected that `lambda2` is added to each element on the diagonal of
+     `gramMatrix`.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | Training matrix. | _(N/A)_ |
+| `responses` | [`arma::rowvec`](../matrices.md) | Training responses (e.g. values to predict).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `colMajor` | `bool` | Should be set to `true` if `data` is [column-major](../matrices.md#representing-data-in-mlpack).  Passing row-major data can avoid a transpose operation. | `false` |
+| `useCholesky` | `bool` | If `true`, use the Cholesky decomposition of the Gram matrix to solve linear systems (as opposed to the full Gram matrix). | `false` |
+| `gramMatrix` | [`arma::mat`](../matrices.md) | Precomputed Gram matrix of `data` (i.e.  `data * data.t()` for column-major data). | _(N/A)_ |
+| `lambda1` | `double` | L1 regularization penalty parameter. | `0.0` |
+| `lambda2` | `double` | L2 regularization penalty parameter. | `0.0` |
+| `tolerance` | `double` | Tolerance on feature correlations for convergence. | `1e-16` |
+| `fitIntercept` | `bool` | If `true`, an intercept term will be included in the model. | `true` |
+| `normalizeData` | `bool` | If `true`, data will be normalized before fitting the model. | `true` |
+
+As an alternative to passing hyperparameters, each hyperparameter can be set
+with a standalone method.  The following functions can be used before calling
+`Train()` to set hyperparameters:
+
+ * `lars.UseCholesky() = useChol;` will set whether or not the Cholesky
+   decomposition will be used during training to `useChol`.
+ * `lars.Lambda1() = lambda1;` will set the L1 regularization penalty parameter
+   to `lambda1`.
+ * `lars.Lambda2() = lambda2;` will set the L2 regularization penalty parameter
+   to `lambda2`.
+ * `lars.Tolerance() = tol;` will set the convergence tolerance to `tol`.
+ * `lars.FitIntercept(fitIntercept);` will set whether an intercept will be fit
+   to `fitIntercept`.  If an external Gram matrix has been specified, this will
+   throw an exception.
+ * `lars.NormalizeData(normalizeData);` will set whether data should be
+   normalized to `normalizeData`.  If an external Gram matrix has been
+   specified, this will throw an exception.
+
+***Notes:***
+
+ - The `lambda1` parameter implicitly controls the sparsity of the model; for
+   more sparse models (i.e. fewer nonzero weights), specify a larger `lambda1`.
+
+ - Specifying a too-small `lambda1` or `lambda2` value may cause the model to
+   overfit; however, setting it too large may cause the model to underfit.
+   Because LARS is a path algorithm, the
+   [`SelectBeta()`](#other-functionality) functions can be used to select models
+   with different values of `lambda1`.  For tuning `lambda2`,
+   [Automatic hyperparameter tuning](../hpt.md) can be used.
+
+<!-- TODO: fix the link to the hyperparameter tuner -->
+
+ - `fitIntercept` and `normalizeData` are recommended to be set as `true`, in
+   accordance with the original LARS algorithm.  `false` can be used for
+   `fitIntercept` if the features and responses are already mean-centered, and
+   `false` can also be used for `normalizeData` if the features are already
+   unit-variance.  Using `false` for either option can provide a small amount of
+   speedup.
+
+ - `useCholesky` should generally be set to `true` and in most situations will
+   result in faster training.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `lars.Train(data, responses, colMajor=true, useCholesky=true, lambda1=0.0, lambda2=0.0, tolerance=1e-16, fitIntercept=true, normalizeData=true)`
+   - Train the model on the given data.
+
+---
+
+ * `lars.Train(data, responses, colMajor, useCholesky, gramMatrix, lambda1=0.0, lambda2=0.0, tolerance=1e-16, fitIntercept=true, normalizeData=true)`
+   - *(Advanced training.)*
+   - Train model on the given data and responses, using a precomputed Gram
+     matrix (`gramMatrix`, equivalent to `data * data.t()`).
+   - Using a precomputed Gram matrix can save time, if it has already been
+     computed.
+   - ***Note:*** any precomputed Gram matrix must also match the settings of
+     `fitIntercept` and `normalizeData`; so, if both are `true`, then
+     `gramMatrix` must be computed on mean-centered data whose features are
+     normalized to have unit variance.  In addition, if `lambda2 > 0`, then
+     it is expected that `lambda2` is added to each element on the diagonal of
+     `gramMatrix`.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes:***
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   model from scratch.
+
+ * `Train()` returns the squared error (loss) of the model on the training set
+   as a `double`.  To obtain the MSE, divide by the number of training points.
+
+### Prediction
+
+Once a `LARS` model is trained, the `Predict()` member function
+can be used to make predictions for new data.
+
+ * `double predictedValue = lars.Predict(point)`
+   - ***(Single-point)***
+   - Make a prediction for a single point, returning the predicted value.
+
+---
+
+ * `lars.Predict(data, predictions, colMajor=true)`
+   - ***(Multi-point)***
+   - Make predictions for a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+#### Prediction Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for prediction. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md) points for classification. |
+| _multi-point_ | `predictions` | [`arma::rowvec&`](../matrices.md) | Vector of `double`s to store predictions into.  Will be set to length `data.n_cols`. |
+| _multi-point_ | `colMajor` | `bool` | Should be set to `true` if `data` is [column-major](../matrices.md).  Passing row-major data can avoid a transpose operation.  (Default `true`.) |
+
+### Other Functionality
+
+ * A `LARS` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `lars.Beta()` will return an `arma::vec` with the model parameters.  This
+   will have length equal to the dimensionality of the model.  Note that
+   `lars.Beta()` can be changed to a different model on the LARS path using the
+   [`lars.SelectBeta()` method](#the-lars-path).
+
+ * `lars.Intercept()` will return a `double` representing the fitted intercept
+   term, or 0 if `lars.FitIntercept()` is `false`.
+
+ * `lars.ActiveSet()` will return a `std::vector<size_t>&` containing the
+   indices of nonzero dimensions in the model parameters (`lars.Beta()`).
+
+ * `lars.ComputeError(data, responses, colMajor=true)` will return a `double`
+   containing the squared error of the model on `data`, given that
+   the true responses are `responses`.  To obtain the MSE, divide by the number
+   of points in `data`.
+
+### The LARS Path
+
+LARS is a *path* (or stepwise) algorithm, meaning it adds one feature at a time
+to the model.  This in turn means that when we train a LARS model with
+`lambda1` set to `l`, we also recover every possible LARS model on the same data
+with a `lambda1` greater than `l`.
+
+The `LARS` class provides a way to access all of the models on the path, and
+switch between them for prediction purposes:
+
+ * `lars.BetaPath()` returns a `std::vector<arma::vec>&` containing each set of
+   model weights on the LARS path.
+
+ * `lars.InterceptPath()` returns a `std::vector<double>&` containing each
+   intercept value on the LARS path.  These values are only meaningful if
+   `lars.FitIntercept()` is `true`.
+
+ * `lars.LambdaPath()` returns a `std::vector<double>&` containing each
+   `lambda1` value that is associated with each element in `lars.BetaPath()` and
+   `lars.InterceptPath()`.  That is, `lars.LambdaPath()[i]` is the `lambda1`
+    value corresponding to the model defined by `lars.BetaPath()[i]` and
+    `lars.InterceptPath()[i]`.
+
+ * `lars.SelectBeta(lambda1)` will set the model weights (`lars.ActiveSet()`,
+   `lars.Beta()` and `lars.Intercept()`) to the path location with L1 penalty
+   `lambda1`.  This is equivalent to calling `lars.Train(data, responses,
+   colMajor, useCholesky, lambda1)`---but much more efficient!  `lambda1`
+   cannot be less than `lars.Lambda1()`, or an exception will be thrown.
+
+ * `lars.SelectedLambda1()` returns the currently selected L1 regularization
+   penalty parameter.
+
+ * For any value `lambda1` between `lars.LambdaPath()[i]` and
+   `lars.LambdaPath()[i + 1]`, the corresponding model is a linear interpolation
+   between `lars.BetaPath()[i]` and `lars.BetaPath()[i + 1]` (and
+   `lars.InterceptPath()[i]` and `lars.InterceptPath()[i + 1]`).  This exact
+   linear interpolation is what is computed by `lars.SelectBeta(lambda1)`.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `LARS` class.
+
+---
+
+Train a LARS model in the constructor, and print the MSE on training and
+test data for each set of weights in the path.
+
+```c++
+// See https://datasets.mlpack.org/wave_energy_farm_100.csv.
+arma::mat data;
+mlpack::data::Load("wave_energy_farm_100.csv", data, true);
+
+// Split the last row off: it is the responses.  Also, normalize the responses
+// to [0, 1].
+arma::rowvec responses = data.row(data.n_rows - 1);
+responses /= responses.max();
+data.shed_row(data.n_rows - 1);
+
+// Split into a training and test dataset.  20% of the data is held out as a
+// test set.
+arma::mat trainingData, testData;
+arma::rowvec trainingResponses, testResponses;
+mlpack::data::Split(data, responses, trainingData, testData, trainingResponses,
+    testResponses, 0.2);
+
+// Train a LARS model with lambda1 = 1e-5 and lambda2 = 1e-6.
+mlpack::LARS lars(trainingData, trainingResponses, true, true, 1e-5, 1e-6);
+
+// Iterate over all the models in the path.
+const size_t pathLength = lars.BetaPath().size();
+for (size_t i = 0; i < pathLength; ++i)
+{
+  // Use the i'th model in the path.
+  lars.SelectBeta(lars.LambdaPath()[i]);
+
+  // ComputeError() returns the total loss, which we need to divide by the
+  // number of points to get the MSE.
+  const double trainMSE = lars.ComputeError(trainingData, trainingResponses) /
+      trainingData.n_cols;
+  const double testMSE = lars.ComputeError(testData, testResponses) /
+      testData.n_cols;
+  std::cout << "L1 penalty parameter: " << lars.SelectedLambda1() << std::endl;
+  std::cout << "  MSE on training set: " << trainMSE << "." << std::endl;
+  std::cout << "  MSE on test set:     " << testMSE << "." << std::endl;
+}
+```
+
+---
+
+Train a LARS model, print predictions for a random point, and save to a file.
+
+```c++
+// See https://datasets.mlpack.org/admission_predict.csv.
+arma::mat data;
+mlpack::data::Load("admission_predict.csv", data, true); 
+
+// See https://datasets.mlpack.org/admission_predict.responses.csv.
+arma::rowvec responses;
+mlpack::data::Load("admission_predict.responses.csv", responses, true);
+
+// Train a LARS model with only L2 regularization.
+mlpack::LARS lars(data, responses, true, true, 0.0, 0.1 /* lambda2 */);
+
+// Predict on a random point.
+arma::vec point = arma::randu<arma::vec>(data.n_rows);
+const double prediction = lars.Predict(point);
+
+std::cout << "Prediction on random point: " << prediction << "." << std::endl;
+
+// Save the model to "lars_model.bin" with the name "lars".
+mlpack::data::Save("lars_model.bin", "lars", lars, true);
+```
+
+---
+
+Load a LARS model from disk and print some information about it.
+
+```c++
+// This assumes a model named "lars" has previously been saved to
+// "lars_model.bin".
+mlpack::LARS lars;
+mlpack::data::Load("lars_model.bin", "lars", lars, true);
+
+if (lars.BetaPath().size() == 0)
+{
+  std::cout << "lars_model.bin contains an untrained LARS model." << std::endl;
+}
+else
+{
+  std::cout << "Information on the LARS model in lars_model.bin:" << std::endl;
+
+  std::cout << " - Model dimensionality: " << lars.Beta().n_elem << "."
+      << std::endl;
+  std::cout << " - Has intercept: "
+      << (lars.FitIntercept() ? std::string("yes") : std::string("no")) << "."
+      << std::endl;
+  std::cout << " - Current L1 regularization penalty parameter value: "
+      << lars.SelectedLambda1() << "." << std::endl;
+  std::cout << " - L2 regularization penalty parameter: " << lars.Lambda2()
+      << "." << std::endl;
+  std::cout << " - Number of nonzero elements in model: "
+      << lars.ActiveSet().size() << "." << std::endl;
+  std::cout << " - Number of models in LARS path: " << lars.BetaPath().size()
+      << "." << std::endl;
+  std::cout << " - Model weight for dimension 0: " << lars.Beta()[0] << "."
+      << std::endl;
+
+  if (lars.FitIntercept())
+  {
+    std::cout << " - Intercept value: " << lars.Intercept() << "." << std::endl;
+  }
+}
+```
+
+---
+
+Train several models with different L2 regularization penalty parameters, using
+a precomputed Gram matrix.
+
+```c++
+// See https://datasets.mlpack.org/admission_predict.csv.
+arma::mat data;
+mlpack::data::Load("admission_predict.csv", data, true);
+
+// See https://datasets.mlpack.org/admission_predict.responses.csv.
+arma::rowvec responses;
+mlpack::data::Load("admission_predict.responses.csv", responses, true);
+
+// Precompute Gram matrix.
+arma::mat gramMatrix = data * data.t();
+
+std::vector<double> lambda2Values = { 0.01, 0.1, 1.0, 10.0, 100.0 };
+for (double lambda2 : lambda2Values)
+{
+  // Build a LARS model using the precomputed Gram matrix.  We did not normalize
+  // or center the data before computing the Gram matrix, so we have to set
+  // fitIntercept and normalizeData accordingly.
+  mlpack::LARS lars(data, responses, true, true, gramMatrix, 0.01, lambda2,
+      1e-16, false, false);
+
+  std::cout << "MSE with L2 penalty " << lambda2 << ": "
+      << (lars.ComputeError(data, responses) / data.n_cols) << "." << std::endl;
+}
+```
+
+### Advanced Functionality: Different Element Types
+
+The `LARS` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+LARS<ModelMatType>
+```
+
+`ModelMatType` specifies the type of matrix used for the internal representation
+of model parameters.  Any matrix type that implements the Armadillo API can be
+used.
+
+Note that the `Train()` and `Predict()` functions themselves are templatized and
+can allow any matrix type that has the same element type.  So, for instance, a
+`LARS<arma::sp_mat>` can accept an `arma::mat` for training.
+
+The example below trains a LARS model on 32-bit precision data, using
+`arma::sp_fmat` to store the model parameters.
+
+```c++
+// Create random, sparse 1000-dimensional data.
+arma::fmat dataset(1000, 5000, arma::fill::randu);
+
+// Generate noisy responses from random data.
+arma::fvec trueWeights(1000, arma::fill::randu);
+arma::frowvec responses = trueWeights.t() * dataset +
+    0.01 * arma::randu<arma::frowvec>(5000) /* noise term */;
+
+mlpack::LARS<arma::sp_fmat> lars;
+lars.Lambda1() = 0.1;
+lars.Lambda2() = 0.01;
+
+lars.Train(dataset, responses);
+
+// Compute the MSE on the training set and a random test set.
+arma::fmat testDataset(1000, 2500, arma::fill::randu);
+arma::frowvec testResponses = trueWeights.t() * testDataset +
+    0.01 * arma::randu<arma::frowvec>(2500) /* noise term */;
+
+const float trainMSE = lars.ComputeError(dataset, responses) / dataset.n_cols;
+const float testMSE = lars.ComputeError(testDataset, testResponses) /
+    testDataset.n_cols;
+
+std::cout << "MSE on training set: " << trainMSE << "." << std::endl;
+std::cout << "MSE on test set:     " << testMSE << "." << std::endl;
+```
+
+***Note:*** it is generally only more efficient to use a sparse type (e.g.
+`arma::sp_mat`) for `ModelMatType` when the L1 regularization parameter is set
+such that a highly sparse model is produced.
diff -pruN 4.3.0-2/doc/user/methods/linear_regression.md 4.4.0-1/doc/user/methods/linear_regression.md
--- 4.3.0-2/doc/user/methods/linear_regression.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/linear_regression.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,286 @@
+## `LinearRegression`
+
+The `LinearRegression` class implements a standard L2-regularized linear
+regression model for numerical data, trained by direct decomposition of the
+training data.  The class offers configurable functionality and template
+parameters to control the data type used for storing the model.
+
+#### Simple usage example:
+
+```c++
+// Train a linear regression model on random numeric data and make predictions.
+
+// All data and responses are uniform random; this uses 10 dimensional data.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::rowvec responses = arma::randn<arma::rowvec>(1000);
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::LinearRegression lr;          // Step 1: create model.
+lr.Train(dataset, responses);         // Step 2: train model.
+arma::rowvec predictions;
+lr.Predict(testDataset, predictions); // Step 3: use model to predict.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions > 0.7) << " test points predicted to have"
+    << " responses greater than 0.7." << std::endl;
+std::cout << arma::accu(predictions < 0) << " test points predicted to have "
+    << "negative responses." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `LinearRegression` objects.
+ * [`Train()`](#training): train model.
+ * [`Predict()`](#prediction): predict with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [mlpack regression techniques](../../index.md#regression-algorithms)
+ * [`LARS`](lars.md)
+ * [Linear Regression on Wikipedia](https://en.wikipedia.org/wiki/Linear_regression)
+
+### Constructors
+
+ * `lr = LinearRegression()`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Predict()`](#prediction).
+
+---
+
+ * `lr = LinearRegression(data, responses,          lambda=0.0, intercept=true)`
+ * `lr = LinearRegression(data, responses, weights, lambda=0.0, intercept=true)`
+   - Train model, optionally with instance weights.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `responses` | [`arma::rowvec`](../matrices.md) | Training responses (e.g. values to predict).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `weights` | [`arma::rowvec`](../matrices.md) | Weights for each training point.  Should have length `data.n_cols`. | _(N/A)_ |
+| `lambda` | `double` | L2 regularization penalty parameter. | `0.0` |
+| `intercept` | `bool` | Whether to fit an intercept term in the model. | `bool` |
+
+As an alternative to passing `lambda`, it can be set with the standalone
+`Lambda()` method: `lr.Lambda() = l;` will set the value of `lambda` to `l` for
+the next time `Train()` is called.
+
+***Note***: setting `lambda` too small may cause the model to overfit; however,
+setting it too large may cause the model to underfit.  [Automatic hyperparameter
+tuning](../hpt.md) can be used to find a good value of `lambda` instead of a
+manual setting.
+
+<!-- TODO: update link to hyperparameter tuner -->
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `lr.Train(data, responses,          lambda=0.0, intercept=true)`
+ * `lr.Train(data, responses, weights, lambda=0.0, intercept=true)`
+   - Train model on the given data, optionally with instance weights.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes:***
+
+ * Training is not incremental.  A second call to `Train()` will retrain the
+   model from scratch.
+
+ * `Train()` returns the mean squared error (MSE) of the model on the training
+   set as a `double`.
+
+### Prediction
+
+Once a `LinearRegression` model is trained, the `Predict()` member function
+can be used to make predictions for new data.
+
+ * `double predictedValue = lr.Predict(point)`
+   - ***(Single-point)***
+   - Make a prediction for a single point, returning the predicted value.
+
+---
+
+ * `lr.Predict(data, predictions)`
+   - ***(Multi-point)***
+   - Make predictions for a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+#### Prediction Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for prediction. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::rowvec&`](../matrices.md) | Vector of `double`s to store predictions into.  Will be set to length `data.n_cols`. |
+
+### Other Functionality
+
+ * A `LinearRegression` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `lr.Intercept()` will return a `bool` indicating whether the model was
+   trained with an intercept term.
+
+ * `lr.Parameters()` will return an `arma::vec&` with the model parameters.
+   This will have length equal to the dimensionality of the model if
+   `lr.Intercept()` is `false`, and length equal to the dimensionality of the
+   model plus one if `lr.Intercept()` is `true`.  If an intercept was fitted,
+   the intercept term is the first element of `lr.Parameters()`.
+
+ * `lr.ComputeError(data, responses)` will return a `double` containing the mean
+   squared error (MSE) of the model on `data`, given that the true responses are
+   `responses`.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `LinearRegression` class.
+
+---
+
+Train a linear regression model in the constructor on weighted data, compute the
+objective function with `ComputeError()`, and save the model.
+
+```c++
+// See https://datasets.mlpack.org/admission_predict.csv.
+arma::mat data;
+mlpack::data::Load("admission_predict.csv", data, true);
+
+// See https://datasets.mlpack.org/admission_predict.responses.csv.
+arma::rowvec responses;
+mlpack::data::Load("admission_predict.responses.csv", responses, true);
+
+// Generate random instance weights for each point, in the range 0.5 to 1.5.
+arma::rowvec weights(data.n_cols, arma::fill::randu);
+weights += 0.5;
+
+// Train a linear regression model, fitting an intercept term and using an L2
+// regularization parameter of 0.3.
+mlpack::LinearRegression lr(data, responses, weights, 0.3, true);
+
+// Now compute the MSE on the training set.
+std::cout << "MSE on the training set: " << lr.ComputeError(data, responses)
+    << "." << std::endl;
+
+// Finally, save the model with the name "lr".
+mlpack::data::Save("lr_model.bin", "lr", lr, true);
+```
+
+---
+
+Load a saved linear regression model and print some information about it, then
+make some predictions individually for random points.
+
+```
+mlpack::LinearRegression lr;
+
+// Load the model named "lr" from "lr_model.bin".
+mlpack::data::Load("lr_model.bin", "lr", lr, true);
+
+// Print some information about the model.
+const size_t dimensionality =
+    (lr.Intercept() ? (lr.Parameters().n_elem - 1) : lr.Parameters().n_elem);
+
+std::cout << "Information on the LinearRegression model in 'lr_model.bin':"
+    << std::endl;
+std::cout << " - Model has intercept: "
+    << (lr.Intercept() ? std::string("yes") : std::string("no")) << "."
+    << std::endl;
+if (lr.Intercept())
+{
+  std::cout << " - Intercept weight: " << lr.Parameters()[0] << "."
+      << std::endl;
+}
+std::cout << " - Model dimensionality: " << dimensionality << "." << std::endl;
+std::cout << " - Lambda value: " << lr.Lambda() << "." << std::endl;
+std::cout << std::endl;
+
+// Now make a prediction for three random points.
+for (size_t t = 0; t < 3; ++t)
+{
+  arma::vec randomPoint(dimensionality, arma::fill::randu);
+  const double prediction = lr.Predict(randomPoint);
+
+  std::cout << "Prediction for random point " << t << ": " << prediction << "."
+      << std::endl;
+}
+```
+
+---
+
+See also the following fully-working examples:
+
+ - [Salary prediction with `LinearRegression`](https://github.com/mlpack/examples/blob/master/salary_prediction_with_linear_regression/salary-prediction-linear-regression-cpp.ipynb)
+ - [Avocado price prediction with `LinearRegression`](https://github.com/mlpack/examples/blob/master/avocado_price_prediction_with_linear_regression/avocado_price_prediction_with_lr_cpp.ipynb)
+ - [California housing price prediction with `LinearRegression`](https://github.com/mlpack/examples/blob/master/california_housing_price_prediction_with_linear_regression/california_housing_price_prediction_with_lr_cpp.ipynb)
+
+### Advanced Functionality: Different Element Types
+
+The `LinearRegression` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+LinearRegression<ModelMatType>
+```
+
+`ModelMatType` specifies the type of matrix used for the internal representation
+of model parameters.  Any matrix type that implements the Armadillo API can be
+used.
+
+Note that the `Train()` and `Predict()` functions themselves are templatized and
+can allow any matrix type that has the same element type.  So, for instance, a
+`LinearRegression<arma::mat>` can accept an `arma::sp_mat` for training.
+
+The example below trains a linear regression model on sparse 32-bit floating
+point data, but uses a dense 32-bit floating point vector to store the model
+itself.
+
+```c++
+// Create random, sparse 100-dimensional data.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+
+// Generate noisy responses from random data.
+arma::fvec trueWeights(100, arma::fill::randu);
+arma::frowvec responses = trueWeights.t() * dataset +
+    0.01 * arma::randu<arma::frowvec>(5000) /* noise term */;
+
+mlpack::LinearRegression<arma::fmat> lr;
+lr.Lambda() = 0.01;
+
+lr.Train(dataset, responses);
+
+// Compute the MSE on the training set and a random test set.
+arma::sp_fmat testDataset;
+testDataset.sprandu(100, 1000, 0.3);
+
+arma::frowvec testResponses = trueWeights.t() * testDataset +
+    0.01 * arma::randu<arma::frowvec>(1000) /* noise term */;
+
+std::cout << "MSE on training set: "
+    << lr.ComputeError(dataset, responses) << "." << std::endl;
+std::cout << "MSE on test set:     "
+    << lr.ComputeError(testDataset, testResponses) << "." << std::endl;
+```
+
+***Note:*** dense objects should be used for `ModelMatType`, since in general an
+L2-regularized linear regression model will not be sparse.
diff -pruN 4.3.0-2/doc/user/methods/linear_svm.md 4.4.0-1/doc/user/methods/linear_svm.md
--- 4.3.0-2/doc/user/methods/linear_svm.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/linear_svm.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,374 @@
+## `LinearSVM`
+
+The `LinearSVM` class implements an L2-regularized support vector machine for
+numerical data that can train using any ensmallen optimizer.  The class offers
+standard classification functionality.  Linear SVM is useful for multi-class
+classification (i.e. classes are `0`, `1`, `2`, etc.).
+
+#### Simple usage example:
+
+```c++
+// Train a linear SVM classifier on random data and predict labels:
+
+// All data and labels are uniform random; 5 dimensional data, 4 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+arma::mat testDataset(5, 500, arma::fill::randu); // 500 test points.
+
+mlpack::LinearSVM svm;                  // Step 1: create model.
+svm.Train(dataset, labels, 4);          // Step 2: train model.
+arma::Row<size_t> predictions;
+svm.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 1) << " test points classified as class "
+    << "1." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `LinearSVM` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [`GaussianDistribution`](../core.md#gaussiandistribution)
+ * [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+
+### Constructors
+
+ * `svm = LinearSVM()`
+   - Initialize the parameters of the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `svm = LinearSVM(dimensionality, numClasses, lambda=0.0001, delta=1.0, fitIntercept=false)`
+   - Initialize the model without training, to default weights.
+   - [`Classify()`](#classification) can immediately be called and
+     `Parameters()` returns valid weights, but the model is otherwise untrained.
+   - The model should be trained with [`Train()`](#training) before calling
+     [`Classify()`](#classification).
+
+---
+
+ * `svm = LinearSVM(data, labels, numClasses, lambda=0.0001, delta=1.0, fitIntercept=false, [callbacks...])`
+   - Train model, optionally specifying ensmallen callbacks for use during
+     optimization.
+
+---
+
+ * `svm = LinearSVM(data, labels, numClasses, optimizer, lambda=0.0001, delta=1.0, fitIntercept=false, [callbacks...])`
+   - Train model with a custom ensmallen optimizer, optionally specifying
+     callbacks for use during optimization.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `dimensionality` | `size_t` | Dimension of input data (if data is not specified).  Should be equal to `data.n_rows`. | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `optimizer` | [any ensmallen optimizer](https://www.ensmallen.org) | Instantiated ensmallen optimizer for [differentiable functions](https://www.ensmallen.org/docs.html#differentiable-functions) or [differentiable separable functions](https://www.ensmallen.org/docs.html#differentiable-separable-functions). | `ens::L_BFGS()` |
+| `lambda` | `double` | L2 regularization penalty parameter.  Must be nonnegative. | `0.0` |
+| `delta` | `double` | Margin of difference between correct class and other classes. | `1.0` |
+| `fitIntercept` | `bool` | If `true`, then an intercept term is fitted to the model. | `false` |
+| `callbacks...` | [any set of ensmallen callbacks](https://www.ensmallen.org/docs.html#callback-documentation) | Optional callbacks for the ensmallen optimizer, such as e.g. `ens::ProgressBar()`, `ens::Report()`, or others. | _(N/A)_ |
+
+As an alternative to passing `lambda`, `delta`, or `fitIntercept`, these can be
+set with a standalone method.  The following functions can be used before
+calling `Train()`:
+
+ * `svm.Lambda() = lambda;` will set the L2 regularization penalty parameter to
+   `lambda`.
+ * `svm.Delta() = delta;` will set the margin of difference to `delta`.
+ * `svm.FitIntercept() = fitIntercept;` will set whether the model fits an
+   intercept to `fitIntercept`.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `svm.Train(data, labels, numClasses,            [callbacks...])`
+ * `svm.Train(data, labels, numClasses, optimizer, [callbacks...])`
+   - Train model without changing any hyperparameters, optionally using a custom
+     ensmallen optimizer and specifying callbacks for use during optimization.
+
+---
+
+ * `svm.Train(data, labels, numClasses,            lambda=0.0001, delta=1.0, fitIntercept=false, [callbacks...])`
+ * `svm.Train(data, labels, numClasses, optimizer, lambda=0.0001, delta=1.0, fitIntercept=false, [callbacks...])`
+   - Train model on the given data, specifying hyperparameters and optionally
+     also a custom ensmallen optimizer and callbacks for use during
+     optimization.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Note:*** Training is not incremental.  Successive calls to `Train()` will
+train entirely new models.
+
+### Classification
+
+Once a `LinearSVM` model is trained, the `Classify()` member function
+can be used to make class predictions for new data.
+
+ * `size_t predictedClass = svm.Classify(point)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class (`0` through
+     `numClasses - 1`, inclusive).
+
+---
+
+ * `svm.Classify(point, prediction, probabilitiesVec)`
+   - ***(Single-point)***
+   - Classify a single point and compute class probabilities.
+   - The predicted class is stored in `prediction`.
+   - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `svm.Classify(data, predictions)`
+   - ***(Multi-point)***
+   - Classify a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `svm.Classify(data, predictions, probabilities)`
+   - ***(Multi-point)***
+   - Classify a set of points and compute class probabilities.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The probability of class `j` for data point `i` can be accessed with
+     `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into; will have length 2. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into; will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to 2; number of columns will be equal to `data.n_cols`). |
+
+### Other Functionality
+
+ * A `LinearSVM` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `svm.Parameters()` will return the parameters of the model as an `arma::mat`
+   with either `data.n_rows` rows (if `FitIntercept()` is `false`) or
+   `data.n_rows + 1` rows (if `FitIntercept()` is `true`), and `numClasses`
+   columns.  The weight for dimension `i` for class `j` can be accessed with
+   `svm.Parameters()(i, j)`.  If `FitIntercept()` is `true`, the last row of
+   `svm.Parameters()` represents the bias parameters for each class.
+
+ * `svm.FeatureSize()` will return the number of features in the model.  This is
+   equivalent to `data.n_rows` when the model was trained.  The output is only
+   valid if the model has been trained.
+
+ * `svm.ComputeAccuracy(data, labels)` will return the accuracy of the model on
+   the given `data` with the given `labels`.  The returned accuracy is between 0
+   and 100.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `LinearSVM` class.
+
+---
+
+Train a linear SVM using a custom SGD-like optimizer with callbacks.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat dataset;
+mlpack::data::Load("satellite.train.csv", dataset, true);
+// See https://datasets.mlpack.org/satellite.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("satellite.train.labels.csv", labels, true);
+
+mlpack::LinearSVM svm;
+svm.Lambda() = 0.1;
+
+// Create AMSGrad optimizer with custom step size and batch size.
+ens::AMSGrad optimizer(0.01 /* step size */, 16 /* batch size */);
+optimizer.MaxIterations() = 100 * dataset.n_cols; // Allow 100 epochs.
+
+// Print a progress bar and an optimization report when training is finished.
+svm.Train(dataset, labels, 2, optimizer, ens::ProgressBar(), ens::Report());
+
+// Now predict on test labels and compute accuracy.
+
+// See https://datasets.mlpack.org/satellite.test.csv.
+arma::mat testDataset;
+mlpack::data::Load("satellite.test.csv", testDataset, true);
+// See https://datasets.mlpack.org/satellite.test.labels.csv.
+arma::Row<size_t> testLabels;
+mlpack::data::Load("satellite.test.labels.csv", testLabels, true);
+
+std::cout << std::endl;
+std::cout << "Accuracy on training set: "
+    << svm.ComputeAccuracy(dataset, labels) << "\%." << std::endl;
+std::cout << "Accuracy on test set:     "
+    << svm.ComputeAccuracy(testDataset, testLabels) << "\%." << std::endl;
+```
+
+---
+
+Train a linear SVM with SGD and save the model every epoch using a [custom
+ensmallen callback](https://www.ensmallen.org/docs.html#custom-callbacks):
+
+```c++
+// This callback saves the model into "model-<epoch>.bin" after every epoch.
+class ModelCheckpoint
+{
+ public:
+  ModelCheckpoint(mlpack::LinearSVM<>& model) : model(model) { }
+
+  template<typename OptimizerType, typename FunctionType, typename MatType>
+  bool EndEpoch(OptimizerType& /* optimizer */,
+                FunctionType& /* function */,
+                const MatType& /* coordinates */,
+                const size_t epoch,
+                const double /* objective */)
+  {
+    const std::string filename = "model-" + std::to_string(epoch) + ".bin";
+    mlpack::data::Save(filename, "svm", model, true);
+    return false; // Do not terminate the optimization.
+  }
+
+ private:
+  mlpack::LinearSVM<>& model;
+};
+```
+
+With that callback available, the code to train the model is below:
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat dataset;
+mlpack::data::Load("satellite.train.csv", dataset, true);
+// See https://datasets.mlpack.org/satellite.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("satellite.train.labels.csv", labels, true);
+
+mlpack::LinearSVM svm;
+
+// Create AdaDelta optimizer with a small step size and batch size of 1.
+ens::AdaDelta adaDelta(0.001, 1);
+adaDelta.MaxIterations() = 100 * dataset.n_cols; // 100 epochs maximum.
+
+// Use the custom callback and an L2 penalty parameter of 0.01, with default
+// delta and fitting an intercept.
+svm.Train(dataset, labels, 2, adaDelta, 0.01, 1.0, true, ModelCheckpoint(svm),
+    ens::ProgressBar());
+
+// Now files like model-1.bin, model-2.bin, etc. should be saved on disk.
+```
+
+---
+
+Load a linear SVM from disk and print some information about it.
+
+```c++
+mlpack::LinearSVM svm;
+// This assumes that a model called "svm" has been saved to the file
+// "model-1.bin" (as in the previous example).
+mlpack::data::Load("model-1.bin", "svm", svm, true);
+
+// Print the dimensionality of the model and some other statistics.
+std::cout << "The dimensionality of the model in model-1.bin is "
+    << svm.FeatureSize() << "." << std::endl;
+if (svm.FitIntercept())
+{
+  std::cout << "Intercept values for each class: " << std::endl;
+  for (size_t i = 0; i < svm.Parameters().n_cols; ++i)
+  {
+    std::cout << "  - Class " << i << ": "
+        << svm.Parameters()(svm.Parameters().n_rows - 1, i) << "." << std::endl;
+  }
+}
+else
+{
+  std::cout << "The model does not have an intercept fitted." << std::endl;
+}
+
+std::cout << "The L2 regularization penalty parameter is: " << svm.Lambda()
+    << "." << std::endl;
+
+std::cout << "Weights for the first dimension are: "
+    << svm.Parameters().row(0);
+```
+
+---
+
+### Advanced Functionality: Different Element Types
+
+The `LinearSVM` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+LinearSVM<ModelMatType>
+```
+
+`ModelMatType` specifies the type of matrix used for training data and internal
+representation of model parameters.
+
+ * Any matrix type that implements the Armadillo API can be used.
+
+ * `Train()` and `Classify()` functions themselves are templatized and can allow
+   any matrix type that has the same element type.  So, for instance, a
+   `LinearSVM<arma::mat>` can accept an `arma::sp_mat` for training.
+
+The example below trains a linear SVM on sparse 32-bit floating point
+data, but uses dense 32-bit floating point matrices to store the model itself.
+
+```c++
+// Create random, sparse 100-dimensional data, with 3 classes.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 2));
+
+mlpack::LinearSVM<arma::fmat> svm(dataset, labels, 3);
+
+// Now classify a test point.
+arma::sp_fvec point;
+point.sprandu(100, 1, 0.3);
+
+size_t prediction;
+arma::fvec probabilitiesVec;
+svm.Classify(point, prediction, probabilitiesVec);
+
+std::cout << "Prediction for random test point: " << prediction << "."
+    << std::endl;
+std::cout << "Class probabilities for random test point: "
+    << probabilitiesVec.t();
+```
+
+***Note:*** dense objects should be used for `ModelMatType`, since in general
+L2-regularized models are fully dense.
diff -pruN 4.3.0-2/doc/user/methods/local_coordinate_coding.md 4.4.0-1/doc/user/methods/local_coordinate_coding.md
--- 4.3.0-2/doc/user/methods/local_coordinate_coding.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/local_coordinate_coding.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,375 @@
+## `LocalCoordinateCoding`
+
+The `LocalCoordinateCoding` class implements local coordinate coding, a
+variation of [sparse coding](sparse_coding.md) with dictionary learning.  Local
+coordinate coding is a form of representation learning, and can be used to
+represent each point in a dataset as a linear combination of a few nearby
+*atoms* in the learned dictionary.
+
+#### Simple usage example:
+
+```c++
+// Create a random dataset with 100 points in 40 dimensions, and then a random
+// test dataset with 50 points.
+arma::mat data(40, 100, arma::fill::randn);
+arma::mat testData(40, 50, arma::fill::randn);
+
+// Perform local coordinate coding with 20 atoms and an L1 penalty of 0.1.
+mlpack::LocalCoordinateCoding lcc(20, 0.1); // Step 1: create object.
+double objective = lcc.Train(data);         // Step 2: learn dictionary.
+arma::mat codes;
+lcc.Encode(testData, codes);                // Step 3: encode new data.
+
+// Print some information about the test encoding.
+std::cout << "Average density of encoded test data: "
+    << 100.0 * arma::mean(arma::sum(codes != 0)) / codes.n_rows << "\%."
+    << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `LocalCoordinateCoding` objects.
+ * [`Train()`](#training): train model (learn dictionary).
+ * [`Encode()`](#encoding): encode points with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for
+   advanced functionality: different element types and dictionary initialization
+   strategies.
+
+#### See also:
+
+ * [`SparseCoding`](sparse_coding.md)
+ * [`LARS`](lars.md) (used internally by `LocalCoordinateCoding`)
+ * [mlpack transformations](../../index.md#transformations)
+ * [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ * [Nonlinear learning using local coordinate coding (pdf)](https://proceedings.neurips.cc/paper_files/paper/2009/file/2afe4567e1bf64d32a5527244d104cea-Paper.pdf)
+
+### Constructors
+
+ * `lcc = LocalCoordinateCoding()`
+ * `lcc = LocalCoordinateCoding(atoms=0, lambda=0.0, maxIter=0, tol=0.01)`
+   - Create a `LocalCoordinateCoding` object without learning a dictionary on
+     data.
+   - If `atoms` is set to `0` (the default), it will need to be set to a value
+     greater than `0` before `Train()` is called (`lcc.Atoms() = atoms` can be
+     used for this).
+
+ * `lcc = LocalCoordinateCoding(data, atoms, lambda=0.0, maxIter=0, tol=0.01)`
+   - Create a `LocalCoordinateCoding` object and train the dictionary on the
+     given `data`.
+   - The dictionary will contain `atoms` elements.
+
+ * `lcc = LocalCoordinateCoding(data, atoms, lambda, maxIter, tol, initializer)`
+   - *Advanced constructor*: create a `LocalCoordinateCoding` object that will
+     use a custom dictionary initializer and train on the given `data`.
+   - The dictionary will contain `atoms` elements.
+   - `initializer` will be used to initialize the dictionary; see [Advanced
+     Functionality: Different Dictionary Initialization
+     Strategies](#dictionaryinitializer-different-dictionary-initialization-strategies)
+     for details.
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `atoms` | `size_t` | Number of atoms in dictionary. | _(N/A)_ |
+| `lambda` | `double` | L1 regularization penalty.  Used in both `Train()` and `Encode()` steps. | `0.0` |
+| `maxIter` | `size_t` | Maximum number of iterations for dictionary learning.  `0` means no limit. | `0` |
+| `tol` | `double` | Objective function tolerance for terminating dictionary learning. | `0.01` |
+
+As an alternative to passing `atoms`, `lambda`, `maxIter`, or `tol`, these can
+be set with a standalone method.  The following functions can be used before
+calling `Train()`:
+
+ * `lcc.Atoms() = a;` will set the number of atoms to use in the dictionary to
+   `a`.  Changing this after calling `Train()` will not make a difference to the
+   dictionary size.
+
+ * `lcc.Lambda() = l;` will set the L1 regularization penalty to `l1`.  This can
+   be set after `Train()` to force sparser encodings when `Encode()` is called.
+
+ * `lcc.MaxIterations() = m;` will set the maximum number of iterations for
+   dictionary learning to `m`.  `0` means that the algorithm will run until
+   convergence.
+
+ * `lcc.Tolerance() = t;` will set the objective tolerance for convergence of
+   the dictionary learning algorithm to `t`.
+
+***Caveats***:
+
+ * Larger settings of `atoms` (i.e. larger dictionary sizes) will be able to
+   more accurately represent the data, but may take longer to learn.
+
+ * Larger values of `lambda` will cause the model to use sparser encodings for
+   data (e.g. fewer nearby anchor points) when `Train()` and `Encode()` are
+   called, but when `lambda` is too large, the codings may be inaccurate
+   representations of the original points.
+
+<!-- TODO: indicate that you can get this info with MLPACK_PRINT_INFO and
+MLPACK_PRINT_WARN, once those are documented -->
+
+ * If `lambda` is set too large, encodings may be empty (e.g. all zeros).
+
+ * Training is not incremental; a second call to `Train()` will reinitialize the
+   dictionary and restart the learning process.
+
+### Training
+
+If training the dictionary is not done as part of the constructor call, it can
+be done with one of the following versions of the `Train()` member function:
+
+ * `lcc.Train(data)`
+ * `lcc.Train(data, initializer)`
+   - Train the local coordinate coding dictionary on the given `data`.
+   - Optionally, use the given `initializer` to initialize the dictionary (see
+     [`DictionaryInitializer`](#dictionaryinitializer-different-dictionary-initialization-strategies)
+     for more details).
+
+### Encoding
+
+Once a `LocalCoordinateCoding` model has a trained dictionary, the `Encode()`
+member function can be used to encode new data points.
+
+ * `lcc.Encode(data, codes)`
+   - Encode `data` (a [column-major data
+     matrix](../matrices.md#representing-data-in-mlpack)) as a sparse set of
+     local atoms of the dictionary, storing the result in `codes`.
+   - Both `data` and `codes` should be the same matrix type (e.g. `arma::mat`);
+     see [Different Element Types](#mattype-different-element-types) for more
+     details.
+   - `codes` will be set to have `atoms` rows and `data.n_cols` columns.
+   - Column `i` of `codes` corresponds to the coding of the `i`'th column of
+     `data`.  Each row represents the weight associated with each atom in the
+     dictionary.
+
+After encoding, the original data can be recovered (approximately) as
+`lcc.Dictionary() * data`.
+
+### Other Functionality
+
+ * A `LocalCoordinateCoding` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `lcc.Dictionary()` will return an `arma::mat&` containing the dictionary
+   matrix.  The matrix has `data.n_rows` rows and `atoms` columns; each column
+   corresponds to an atom in the dictionary.  Dictionary atoms are regularized
+   to be close to the manifold that data lie on.
+
+ * `double obj = lcc.Objective(data, codes)` computes the local coordinate
+   coding objective function on the given `data` and encodings `codes`.  This
+   can be used after `Encode()` to test the quality of the encodings (a smaller
+   objective is better).
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `LocalCoordinateCoding` class.
+
+---
+
+Train a local coordinate coding model on the cloud dataset and print the
+reconstruction error.
+
+```c++
+// See https://datasets.mlpack.org/cloud.csv.
+arma::mat dataset;
+mlpack::data::Load("cloud.csv", dataset, true);
+
+mlpack::LocalCoordinateCoding lcc;
+lcc.Atoms() = 50;
+lcc.Lambda() = 1e-5;
+lcc.MaxIterations() = 25;
+lcc.Train(dataset);
+
+// Encode the training dataset.
+arma::mat codes;
+lcc.Encode(dataset, codes);
+
+std::cout << "Input matrix size: " << dataset.n_rows << " x " << dataset.n_cols
+    << "." << std::endl;
+std::cout << "Codes matrix size: " << codes.n_rows << " x " << codes.n_cols
+    << "." << std::endl;
+
+// Reconstruct the original matrix.
+arma::mat recon = lcc.Dictionary() * codes;
+double error = std::sqrt(arma::norm(dataset - recon, "fro") / dataset.n_elem);
+std::cout << "RMSE of reconstructed matrix: " << error << "." << std::endl;
+```
+
+---
+
+Train a local coordinate coding model on the iris dataset and save the model to
+disk.
+
+```c++
+// See https://datasets.mlpack.org/iris.train.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.train.csv", dataset, true);
+
+// Train the model in the constructor.
+mlpack::LocalCoordinateCoding lcc(dataset,
+                                  10 /* atoms */,
+                                  0.1 /* L1 penalty */);
+
+// Save the model to disk.
+mlpack::data::Save("lcc.bin", "lcc", lcc);
+```
+
+---
+
+Train a local coordinate coding model on the satellite dataset, trying several
+different regularization parameters and checking the objective value on a
+held-out test dataset.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat trainData;
+mlpack::data::Load("satellite.train.csv", trainData, true);
+// See https://datasets.mlpack.org/satellite.test.csv.
+arma::mat testData;
+mlpack::data::Load("satellite.test.csv", testData, true);
+
+for (double lambdaPow = -6; lambdaPow <= -2; lambdaPow += 1)
+{
+  const double lambda = std::pow(10.0, lambdaPow);
+  mlpack::LocalCoordinateCoding lcc(50 /* atoms */);
+  lcc.Lambda() = lambda;
+  lcc.MaxIterations() = 25; // Keep iterations low so this runs relatively fast.
+
+  const double trainObj = lcc.Train(trainData);
+
+  // Compute the objective on the test set.
+  arma::mat codes;
+  lcc.Encode(testData, codes);
+  const double testObj = lcc.Objective(testData, codes);
+
+  std::cout << "Lambda: " << std::setfill(' ') << std::setw(3) << lambda
+      << "; ";
+  std::cout << "training set objective: " << std::setw(6) << trainObj << "; ";
+  std::cout << "test set objective: " << std::setw(6) << testObj << "."
+      << std::endl;
+}
+```
+
+### Advanced Functionality: Template Parameters
+
+The `LocalCoordinateCoding` class has one class template parameter that can be
+used for custom behavior.  The full signature of the class is:
+
+```
+LocalCoordinateCoding<MatType>
+```
+
+In addition, the [constructors](#constructors) and [`Train()`
+functions](#training) have a template parameter `DictionaryInitializer` that can
+be used for custom behavior.
+
+ * `MatType`: the type of the matrix to use (e.g. `arma::mat`, `arma::fmat`,
+   etc.).  The given `MatType` must support the Armadillo API and hold a
+   floating-point element type (e.g. `float`, `double`, etc.).
+
+ * `DictionaryInitializer`: the strategy used to initialize the dictionary.  By
+   default, `DataDependentRandomInitializer` is used.
+
+#### `MatType`: Different Element Types
+
+`MatType` specifies the type of matrix used for training data and internal
+representation of the dictionary.  Any matrix type that implements the Armadillo
+API can be used.  The example below trains a local coordinate coding model on
+32-bit floating point data.
+
+```c++
+// See https://datasets.mlpack.org/cloud.csv.
+arma::fmat dataset;
+mlpack::data::Load("cloud.csv", dataset, true);
+
+mlpack::LocalCoordinateCoding<arma::fmat> lcc;
+lcc.Atoms() = 30;
+lcc.Lambda() = 1e-5;
+lcc.MaxIterations() = 100;
+lcc.Train(dataset);
+
+// Encode the training dataset.
+arma::fmat codes;
+lcc.Encode(dataset, codes);
+
+std::cout << "Input matrix size: " << dataset.n_rows << " x " << dataset.n_cols
+    << "." << std::endl;
+std::cout << "Codes matrix size: " << codes.n_rows << " x " << codes.n_cols
+    << "." << std::endl;
+
+// Reconstruct the original matrix.
+arma::fmat recon = lcc.Dictionary() * codes;
+double error = std::sqrt(arma::norm(dataset - recon, "fro") / dataset.n_elem);
+std::cout << "RMSE of reconstructed matrix: " << error << "." << std::endl;
+```
+
+#### `DictionaryInitializer`: Different Dictionary Initialization Strategies
+
+The `DictionaryInitializer` template class specifies the strategy to be used to
+initialize the dictionary when `Train()` is called.
+
+ * The `DataDependentRandomInitalizer` class (the default) uses the average of
+   three random points in the dataset to initialize each atom in the dictionary.
+
+ * The `NothingInitializer` class does not modify the dictionary matrix in any
+   way, and could be used either to set a specific dictionary before training
+   with `sc.Dictionary()`, or to allow incremental training that does not modify
+   the existing dictionary when `Train()` is called a second time.
+
+ * The `RandomInitializer` class initializes the dictionary by sampling norm-1
+   atoms from a normal distribution.
+
+***Note:*** none of the classes above have any members, and as such it is not
+necessary to use the constructor or `Train()` variants that take an initialized
+`initializer` object.  That would only be necessary for a custom
+`DictionaryInitializer` class that stored internal members.
+
+---
+
+The example below uses `NothingInitializer` to set a specific initial
+dictionary.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat trainData;
+mlpack::data::Load("satellite.train.csv", trainData, true);
+
+const size_t atoms = 25;
+const double lambda = 1e-5;
+const size_t maxIterations = 50;
+
+// Use a uniform random matrix as the initial dictionary.
+arma::mat initialDictionary(trainData.n_rows, atoms, arma::fill::randu);
+
+mlpack::LocalCoordinateCoding lcc(atoms, lambda, maxIterations);
+lcc.Dictionary() = initialDictionary;
+
+const double obj = lcc.Train<mlpack::NothingInitializer>(trainData);
+std::cout << "Training set objective: " << obj << "." << std::endl;
+```
+
+---
+
+ * An entirely custom class can also be implemented.  The class must implement
+   one method, `Initialize()`:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomDictionaryInitializer
+{
+ public:
+  // Initialize the dictionary to have the given number of atoms, given the
+  // dataset.  MatType will be the matrix type used by the local coordinate
+  // coding model (e.g. `arma::mat`, `arma::fmat`, etc.).
+  template<typename MatType>
+  void Initialize(const MatType& data,
+                  const size_t atoms,
+                  MatType& dictionary);
+};
+```
diff -pruN 4.3.0-2/doc/user/methods/logistic_regression.md 4.4.0-1/doc/user/methods/logistic_regression.md
--- 4.3.0-2/doc/user/methods/logistic_regression.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/logistic_regression.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,409 @@
+## `LogisticRegression`
+
+The `LogisticRegression` class implements a simple L2-regularized two-class
+logistic regression classifier for numerical data, by default using L-BFGS to
+learn the model.  The class offers easy configurability, and arbitrary
+optimizers can be used to learn the model.
+
+Logistic regression is useful for two-class classification (i.e. classes are `0`
+or `1`).  For multi-class logistic regression, see
+[`SoftmaxRegression`](softmax_regression.md).
+
+#### Simple usage example:
+
+```c++
+// Train a logistic regression model on random data and predict labels:
+
+// All data and labels are uniform random; 5 dimensional data, 2 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 1));
+arma::mat testDataset(5, 500, arma::fill::randu); // 500 test points.
+
+mlpack::LogisticRegression lr;         // Step 1: create model.
+lr.Train(dataset, labels);             // Step 2: train model.
+arma::Row<size_t> predictions;
+lr.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 0) << " test points classified as class "
+    << "0." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `LogisticRegression` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [`SoftmaxRegression`](softmax_regression.md)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [Logistic regression on Wikipedia](https://en.wikipedia.org/wiki/Logistic_regression)
+
+### Constructors
+
+ * `lr = LogisticRegression()`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `lr = LogisticRegression(data, labels,               lambda=0.0, [callbacks...])`
+ * `lr = LogisticRegression(data, labels, initialPoint, lambda=0.0, [callbacks...])`
+   - Train model, optionally specifying an initial set of weights for the
+     optimization and callbacks.
+
+---
+
+ * `lr = LogisticRegression(data, labels, optimizer,               lambda=0.0, [callbacks...])`
+ * `lr = LogisticRegression(data, labels, optimizer, initialPoint, lambda=0.0, [callbacks...])`
+   - Train model with a custom ensmallen optimizer, optionally specifying an
+     initial set of weights to start the optimization from and callbacks for the
+     optimizer.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, either [`0` or `1`](../load_save.md#normalizing-labels).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `initialPoint` | `arma::rowvec` | Initial model weights to start optimization from.  Should have length `data.n_rows + 1`.  The first element is the bias.  If not specified, a zero vector will be used. | zero vector |
+| `optimizer` | [any ensmallen optimizer](https://www.ensmallen.org) | Instantiated ensmallen optimizer for [differentiable functions](https://www.ensmallen.org/docs.html#differentiable-functions) or [differentiable separable functions](https://www.ensmallen.org/docs.html#differentiable-separable-functions). | `ens::L_BFGS()` |
+| `lambda` | `double` | L2 regularization penalty parameter.  Must be nonnegative. | `0.0` |
+| `callbacks...` | [any set of ensmallen callbacks](https://www.ensmallen.org/docs.html#callback-documentation) | Optional callbacks for the ensmallen optimizer, such as e.g. `ens::ProgressBar()`, `ens::Report()`, or others. | _(N/A)_ |
+
+As an alternative to passing `lambda` or `initialPoint`, these can be set with a
+standalone method.  The following functions can be used before calling
+`Train()`:
+
+ * `lr.Lambda() = l;` will set the value of the L2 regularization penalty
+   parameter to `l`.
+ * `lr.Parameters() = initialPoint;` will set the initial point for the training
+   optimization to `initialPoint`.
+
+***Note***: Setting `lambda` too small may cause the model to overfit; however,
+setting it too large may cause the model to underfit.  [Automatic hyperparameter
+tuning](../hpt.md) can be used to find a good value of `lambda` instead of a
+manual setting.
+
+<!-- TODO: fix link for hyperparameter tuner -->
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the following versions of the `Train()` member function:
+
+ * `lr.Train(data, labels)`
+ * `lr.Train(data, labels, lambda=0.0, [callbacks...])`
+   - Train model, optionally specifying callbacks for the default L-BFGS
+     optimizer.
+
+---
+
+ * `lr.Train(data, labels, optimizer)`
+ * `lr.Train(data, labels, optimizer, lambda=0.0, [callbacks...])`
+   - Train model with a custom ensmallen optimizer, optionally specifying
+     callbacks.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is incremental.  Successive calls to `Train()` will not reinitialize
+   the model, unless the given data has different dimensionality.  To
+   reinitialize the model, call `Reset()` (see
+   [Other Functionality](#other-functionality)).
+
+ * To set the initial point of the optimization, call `Parameters()`; see
+   [Other Functionality](#other-functionality).
+
+ * `Train()` returns a `double` with the final logistic regression loss value
+   (including L2 penalty term) of the trained model.
+
+### Classification
+
+Once a `LogisticRegression` model is trained, the `Classify()` member function
+can be used to make class predictions for new data.
+
+ * `size_t predictedClass = lr.Classify(point, decisionBoundary=0.5)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class (`0` or `1`).
+
+---
+
+ * `lr.Classify(point, prediction, probabilitiesVec, decisionBoundary=0.5)`
+   - ***(Single-point)***
+   - Classify a single point and compute class probabilities.
+   - The predicted class is stored in `prediction`.
+   - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `lr.Classify(data, predictions, decisionBoundary=0.5)`
+   - ***(Multi-point)***
+   - Classify a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `lr.Classify(data, predictions, probabilities, decisionBoundary=0.5)`
+   - ***(Multi-point)***
+   - Classify a set of points and compute class probabilities.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The probability of class `j` for data point `i` can be accessed with
+     `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into; will have length 2. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into; will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to 2; number of columns will be equal to `data.n_cols`). |
+||||
+| _all_ | `decisionBoundary` | `double` | If the logistic function value for a point is greater than `decisionBoundary`, it is classified as class `1`.  Defaults to `0.5`. |
+
+### Other Functionality
+
+ * A `LogisticRegression` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `lr.Parameters()` will return an `arma::rowvec` filled with the weights of
+   the model.  This vector has length equal to the dimensionality plus one, and
+   the first element is the bias.
+
+ * `lr.Lambda()` will return the L2 regularization penalty parameter.
+
+ * `lr.ComputeAccuracy(data, labels, decisionBoundary=0.5)` will return the
+   accuracy of the model on the given `data` with the given `labels`.  The
+   returned accuracy is between 0 and 100.
+
+ * `lr.ComputeError(data, labels)` will return the loss of the logistic
+   regression objective function on the given `data` with the given `labels`.
+
+ * `lr.Reset()` will reset the weights of the model to zeros.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/logistic_regression/logistic_regression.hpp) can be
+consulted.  Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `LogisticRegression` class.
+
+---
+
+Train a logistic regression model using a custom SGD-like optimizer with
+callbacks.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat dataset;
+mlpack::data::Load("satellite.train.csv", dataset, true);
+// See https://datasets.mlpack.org/satellite.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("satellite.train.labels.csv", labels, true);
+
+mlpack::LogisticRegression lr;
+lr.Lambda() = 0.1;
+
+// Create AMSGrad optimizer with custom step size and batch size.
+ens::AMSGrad optimizer(0.01 /* step size */, 16 /* batch size */);
+optimizer.MaxIterations() = 100 * dataset.n_cols; // Allow 100 epochs.
+
+// Print a progress bar and an optimization report when training is finished.
+lr.Train(dataset, labels, optimizer, ens::ProgressBar(), ens::Report());
+
+// Now predict on test labels and compute accuracy.
+
+// See https://datasets.mlpack.org/satellite.test.csv.
+arma::mat testDataset;
+mlpack::data::Load("satellite.test.csv", testDataset, true);
+// See https://datasets.mlpack.org/satellite.test.labels.csv.
+arma::Row<size_t> testLabels;
+mlpack::data::Load("satellite.test.labels.csv", testLabels, true);
+
+std::cout << std::endl;
+std::cout << "Accuracy on training set: "
+    << lr.ComputeAccuracy(dataset, labels) << "\%." << std::endl;
+std::cout << "Accuracy on test set:     "
+    << lr.ComputeAccuracy(testDataset, testLabels) << "\%." << std::endl;
+std::cout << "Objective on training set: "
+    << lr.ComputeError(dataset, labels) << "." << std::endl;
+std::cout << "Objective on test set:     "
+    << lr.ComputeError(testDataset, testLabels) << "." << std::endl;
+```
+
+---
+
+Train a logistic regression model with SGD and save the model every epoch using
+a [custom ensmallen
+callback](https://www.ensmallen.org/docs.html#custom-callbacks):
+
+```c++
+// This callback saves the model into "model-<epoch>.bin" after every epoch.
+class ModelCheckpoint
+{
+ public:
+  ModelCheckpoint(mlpack::LogisticRegression<>& model) : model(model) { }
+
+  template<typename OptimizerType, typename FunctionType, typename MatType>
+  bool EndEpoch(OptimizerType& /* optimizer */,
+                FunctionType& /* function */,
+                const MatType& /* coordinates */,
+                const size_t epoch,
+                const double /* objective */)
+  {
+    const std::string filename = "model-" + std::to_string(epoch) + ".bin";
+    mlpack::data::Save(filename, "lr_model", model, true);
+    return false; // Do not terminate the optimization.
+  }
+
+ private:
+  mlpack::LogisticRegression<>& model;
+};
+```
+
+With that callback available, the code to train the model is below:
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat dataset;
+mlpack::data::Load("satellite.train.csv", dataset, true);
+// See https://datasets.mlpack.org/satellite.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("satellite.train.labels.csv", labels, true);
+
+mlpack::LogisticRegression lr;
+
+// Create AdaDelta optimizer with a small step size and batch size of 1.
+ens::AdaDelta adaDelta(0.001, 1);
+adaDelta.MaxIterations() = 100 * dataset.n_cols; // 100 epochs maximum.
+
+// Use the custom callback and an L2 penalty parameter of 0.01.
+lr.Train(dataset, labels, adaDelta, 0.01, ModelCheckpoint(lr),
+    ens::ProgressBar());
+
+// Now files like model-1.bin, model-2.bin, etc. should be saved on disk.
+```
+
+---
+
+Load an existing logistic regression model and print some information about it.
+
+```c++
+mlpack::LogisticRegression lr;
+// This assumes that a model called "lr_model" has been saved to the file
+// "model-1.bin" (as in the previous example).
+mlpack::data::Load("model-1.bin", "lr_model", lr, true);
+
+// Print the dimensionality of the model and some other statistics.
+std::cout << "The dimensionality of the model in model-1.bin is "
+    << (lr.Parameters().n_elem - 1) << "." << std::endl;
+std::cout << "The bias parameter for the model is " << lr.Parameters()[0]
+    << "." << std::endl;
+
+arma::vec point(lr.Parameters().n_elem - 1, arma::fill::randu);
+std::cout << "The predicted class for a random point, using a decision boundary"
+    << " of 0.2, is " << lr.Classify(point, 0.2) << "." << std::endl;
+```
+
+---
+
+Perform incremental training on multiple datasets with multiple calls to
+`Train()`.
+
+```c++
+// Generate two random datasets.
+arma::mat firstDataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> firstLabels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 1));
+
+arma::mat secondDataset(5, 1500, arma::fill::randu); // 1500 points.
+arma::Row<size_t> secondLabels =
+    arma::randi<arma::Row<size_t>>(1500, arma::distr_param(0, 1));
+
+// Train a model on the first dataset with an L2 regularization penalty
+// parameter of 0.01.
+mlpack::LogisticRegression lr(firstDataset, firstLabels, 0.01);
+
+// Now compute the objective on the second dataset and print it.
+std::cout << "Objective on second dataset: "
+    << lr.ComputeError(secondDataset, secondLabels) << "." << std::endl;
+
+// Train for a second round on the second dataset.
+lr.Train(secondDataset, secondLabels);
+
+// Now compute the objective on the second dataset again and print it.
+// (Note that it may not be all that much better because this is random data!)
+std::cout << "Objective on second dataset after second training: "
+    << lr.ComputeError(secondDataset, secondLabels) << "." << std::endl;
+```
+
+---
+
+### Advanced Functionality: Different Element Types
+
+The `LogisticRegression` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+LogisticRegression<MatType>
+```
+
+`MatType` specifies the type of matrix used for training data and internal
+representation of model parameters.  Any matrix type that implements the
+Armadillo API can be used.  The example below trains a logistic regression model
+on sparse 32-bit floating point data.
+
+```c++
+// Create random, sparse 100-dimensional data.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 1));
+
+// Train with L2 regularization penalty parameter of 0.1.
+mlpack::LogisticRegression<arma::sp_fmat> lr(dataset, labels, 0.1);
+
+// Now classify a test point.
+arma::sp_fvec point;
+point.sprandu(100, 1, 0.3);
+
+size_t prediction;
+arma::fvec probabilitiesVec;
+lr.Classify(point, prediction, probabilitiesVec);
+
+std::cout << "Prediction for random test point: " << prediction << "."
+    << std::endl;
+std::cout << "Class probabilities for random test point: "
+    << probabilitiesVec.t();
+```
+
+***Note***: if `MatType` is a sparse object (e.g. `sp_fmat`), the internal
+parameter representation will be a *dense* vector containing elements of the
+same type (e.g. `frowvec`).  This is because L2-regularized logistic regression,
+even when training on sparse data, does not necessarily produce sparse models.
diff -pruN 4.3.0-2/doc/user/methods/naive_bayes_classifier.md 4.4.0-1/doc/user/methods/naive_bayes_classifier.md
--- 4.3.0-2/doc/user/methods/naive_bayes_classifier.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/naive_bayes_classifier.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,325 @@
+## `NaiveBayesClassifier`
+
+The `NaiveBayesClassifier` implements a trivial Naive Bayes classifier for
+numerical data.  The class offers standard classification functionality.  Naive
+Bayes is useful for multi-class classification (i.e. classes are `0`, `1`, `2`,
+etc.), and due to its simplicity scales well to large-data scenarios.
+
+#### Simple usage example:
+
+```c++
+// Train a Naive Bayes classifier on random data and predict labels:
+
+// All data and labels are uniform random; 5 dimensional data, 4 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+arma::mat testDataset(5, 500, arma::fill::randu); // 500 test points.
+
+mlpack::NaiveBayesClassifier nbc;       // Step 1: create model.
+nbc.Train(dataset, labels, 4);          // Step 2: train model.
+arma::Row<size_t> predictions;
+nbc.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `NaiveBayesClassifier` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [`GaussianDistribution`](../core.md#gaussiandistribution)
+ * [Naive Bayes classifier on Wikipedia](https://en.wikipedia.org/wiki/Naive_Bayes_classifier)
+
+### Constructors
+
+ * `nbc = NaiveBayesClassifier()`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `nbc = NaiveBayesClassifier(dimensionality, numClasses, epsilon=1e-10)`
+   - Initialize model to the given dimensionality and number of classes without
+     training.
+   - This is meant to be used with the incremental version of `Train()` that
+     takes only a single point.
+
+---
+
+ * `nbc = NaiveBayesClassifier(data, labels, numClasses, incremental=true, epsilon=1e-10)`
+   - Train model, optionally specifying whether to do incremental training.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `incremental` | `bool` | If `true`, then the model will not be reset before training, and will use a robust incremental algorithm for variance computation. | `true` |
+| `epsilon` | `double` | Initial small value for sample variances, to prevent underflow (via `log(0)`). | 1e-10 |
+
+As an alternative to passing the `epsilon` parameter, it can be set with the
+standalone `Epsilon()` method: `nbc.Epsilon() = eps;` will set the value of
+`epsilon` to `eps` for the next time non-incremental `Train()` or `Reset()` is
+called.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `nbc.Train(data, labels, numClasses, incremental=true, epsilon=1e-10)`
+   - Train model on the given data, optionally specifying whether to do
+     incremental training.
+   - Arguments described in [Constructor Parameters](#constructor-parameters)
+     table above.
+
+---
+
+ * `nbc.Train(point, label)`
+   - Incrementally train on a single data point with the given label.
+   - Ensure that the model has the right size and number of classes by using the
+     appropriate constructor form to set `dimensionality`, or by calling
+     `Reset()` (see [other functionality](#other-functionality)).
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `point` | [`arma::vec`](../matrices.md) | [Column-major](../matrices.md) training point (i.e. one column). | _(N/A)_ |
+| `label` | `size_t` | Training label, in range `0` to `numClasses`. | _(N/A)_ |
+
+***Note***: when performing incremental training, if `data` has a different
+dimensionality than the model, or if `numClasses` is different, the model will
+be reset.  For single-point `Train()`, if `point` has different dimensionality,
+an exception will be thrown.
+
+### Classification
+
+Once a `NaiveBayesClassifier` model is trained, the `Classify()` member function
+can be used to make class predictions for new data.
+
+ * `size_t predictedClass = nbc.Classify(point)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class (`0` through
+     `numClasses - 1`, inclusive).
+
+---
+
+ * `nbc.Classify(point, prediction, probabilitiesVec)`
+   - ***(Single-point)***
+   - Classify a single point and compute class probabilities.
+   - The predicted class is stored in `prediction`.
+   - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `nbc.Classify(data, predictions)`
+   - ***(Multi-point)***
+   - Classify a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `nbc.Classify(data, predictions, probabilities)`
+   - ***(Multi-point)***
+   - Classify a set of points and compute class probabilities.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The probability of class `j` for data point `i` can be accessed with
+     `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into; will have length 2. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into; will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to 2; number of columns will be equal to `data.n_cols`). |
+
+### Other Functionality
+
+ * A `NaiveBayesClassifier` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `nbc.Probabilities()` will return a column vector of length `numClasses`
+   representing the prior probability of each class.
+
+ * `nbc.Means()` will return a matrix with rows equal to the dimensionality of
+   the model and `numClasses` columns.  Column `i` represents the sample mean of
+   class `i`.
+
+ * `nbc.Variances()` will return a matrix with rows equal to the dimensionality
+   of the model and `numClasses` columns.  The element at row `i` and column `j`
+   represents the sample variance in dimension `i` of class `j`.
+
+ * `nbc.Reset()` will reset the model to zeros; this is useful before
+   incremental training.  The form
+   `nbc.Reset(dimensionality, numClasses, epsilon=1e-10)` can
+   also be used to set the dimensionality and number of classes in the reset
+   model.
+
+ * `nbc.TrainingPoints()` will return the number of points that the model has
+   been trained on.  When `nbc.Reset()` is called, this is reset to 0.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `NaiveBayesClassifier` class.
+
+---
+
+Train a Naive Bayes classifier incrementally, one point at a time, then compute
+accuracy on a test set and save the model to disk.
+
+```c++
+// See https://datasets.mlpack.org/mnist.train.csv.
+arma::mat dataset;
+mlpack::data::Load("mnist.train.csv", dataset, true);
+// See https://datasets.mlpack.org/mnist.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("mnist.train.labels.csv", labels, true);
+
+mlpack::NaiveBayesClassifier nbc(dataset.n_rows /* dimensionality */,
+                                 10 /* numClasses */);
+
+// Iterate over all points in the dataset and call Train() on each point.
+for (size_t i = 0; i < dataset.n_cols; ++i)
+  nbc.Train(dataset.col(i), labels[i]);
+
+// Now compute the accuracy of the fully trained model on a test set.
+
+// See https://datasets.mlpack.org/mnist.test.csv.
+arma::mat testDataset;
+mlpack::data::Load("mnist.test.csv", testDataset, true);
+// See https://datasets.mlpack.org/mnist.test.labels.csv.
+arma::Row<size_t> testLabels;
+mlpack::data::Load("mnist.test.labels.csv", testLabels, true);
+
+arma::Row<size_t> predictions;
+nbc.Classify(dataset, predictions);
+const double trainAccuracy = 100.0 *
+    ((double) arma::accu(predictions == labels)) / labels.n_elem;
+std::cout << "Accuracy of model on training data: " << trainAccuracy << "\%."
+    << std::endl;
+
+nbc.Classify(testDataset, predictions);
+
+const double testAccuracy = 100.0 *
+    ((double) arma::accu(predictions == testLabels)) / testLabels.n_elem;
+std::cout << "Accuracy of model on test data:     " << testAccuracy << "\%."
+    << std::endl;
+
+// Save the model to disk with the name "nbc".
+mlpack::data::Save("nbc_model.bin", "nbc", nbc, true);
+```
+
+---
+
+Load a saved Naive Bayes classifier and print some information about it.
+
+```c++
+mlpack::NaiveBayesClassifier nbc;
+
+// Load the model named "nbc" from "nbc_model.bin".
+mlpack::data::Load("nbc_model.bin", "nbc", nbc, true);
+
+// Print information about the model.
+std::cout << "The dimensionality of the model in nbc_model.bin is "
+    << nbc.Means().n_rows << "." << std::endl;
+std::cout << "The number of classes in the model is "
+    << nbc.Probabilities().n_elem << "." << std::endl;
+std::cout << "The model was trained on " << nbc.TrainingPoints() << " points."
+    << std::endl;
+std::cout << "The prior probabilities of each class are: "
+    << nbc.Probabilities().t();
+
+// Compute the class probabilities of a random point.
+// For our random point, we'll use one of the means plus some noise.
+arma::vec randomPoint = nbc.Means().col(2) +
+    10.0 * arma::randu<arma::vec>(nbc.Means().n_rows);
+
+size_t prediction;
+arma::vec probabilities;
+nbc.Classify(randomPoint, prediction, probabilities);
+
+std::cout << "Random point class prediction: " << prediction << "."
+    << std::endl;
+std::cout << "Random point class probabilities: " << probabilities.t();
+```
+
+---
+
+See also the following fully-working examples:
+
+ - [Microchip QA Classification using `NaiveBayesClassifier`](https://github.com/mlpack/examples/blob/master/microchip_quality_control_naive_bayes/microchip-quality-control-naive-bayes-cpp.ipynb)
+
+### Advanced Functionality: Different Element Types
+
+The `NaiveBayesClassifier` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+NaiveBayesClassifier<ModelMatType>
+```
+
+`ModelMatType` specifies the type of matrix used for training data and internal
+representation of model parameters.
+
+ * Any matrix type that implements the Armadillo API can be used.
+
+ * `Train()` and `Classify()` functions themselves are templatized and can allow
+   any matrix type that has the same element type.  So, for instance, a
+   `NaiveBayesClassifier<arma::mat>` can accept an `arma::sp_mat` for training.
+
+The example below trains a Naive Bayes model on sparse 32-bit floating point
+data, but uses dense 32-bit floating point matrices to store the model itself.
+
+```c++
+// Create random, sparse 100-dimensional data, with 3 classes.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 2));
+
+mlpack::NaiveBayesClassifier<arma::fmat> nbc(dataset, labels, 3);
+
+// Now classify a test point.
+arma::sp_fvec point;
+point.sprandu(100, 1, 0.3);
+
+size_t prediction;
+arma::fvec probabilitiesVec;
+nbc.Classify(point, prediction, probabilitiesVec);
+
+std::cout << "Prediction for random test point: " << prediction << "."
+    << std::endl;
+std::cout << "Class probabilities for random test point: "
+    << probabilitiesVec.t();
+```
+
+***Note:*** dense objects should be used for `ModelMatType`, since in general
+the mean and sample variance of sparse data is dense.
diff -pruN 4.3.0-2/doc/user/methods/nmf.md 4.4.0-1/doc/user/methods/nmf.md
--- 4.3.0-2/doc/user/methods/nmf.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/nmf.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,625 @@
+## `NMF`
+
+The `NMF` class implements non-negative matrix factorization, a technique to
+decompose a large (potentially sparse) matrix `V` into two smaller matrices `W`
+and `H`, such that `V ~= W * H`, and `W` and `H` only contain nonnegative
+elements.  This technique may be used for dimensionality reduction, or as part
+of a recommender system.
+
+The `NMF` class allows fully configurable behavior via [template
+parameters](#advanced-functionality-template-parameters).  For more general
+matrix factorization strategies, see the
+[`AMF`](/src/mlpack/methods/amf/amf.hpp) (alternating matrix factorization)
+class documentation.
+
+<!-- TODO: clean up AMF link -->
+
+#### Simple usage example:
+
+```c++
+// Create a random sparse matrix (V) of size 10x100, with 15% nonzeros.
+arma::sp_mat V;
+V.sprandu(100, 100, 0.15);
+
+// W and H will be low-rank matrices of size 100x10 and 10x100.
+arma::mat W, H;
+
+mlpack::NMF nmf;                         // Step 1: create object.
+double residue = nmf.Apply(V, 10, W, H); // Step 2: apply NMF to decompose V.
+
+// Now print some information about the factorized matrices.
+std::cout << "W has size: " << W.n_rows << " x " << W.n_cols << "."
+    << std::endl;
+std::cout << "H has size: " << H.n_rows << " x " << H.n_cols << "."
+    << std::endl;
+std::cout << "RMSE of reconstructed matrix: "
+    << arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem) << "." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `NMF` objects.
+ * [`Apply()`](#applying-decompositions): apply `NMF` decomposition to data.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for
+   using different update rules, initialization strategies, and termination
+   criteria.
+ * [Advanced template examples](#advanced-functionality-examples) of use with
+   custom template parameters.
+
+#### See also:
+
+<!-- TODO: add these links
+ * [`AMF`](amf.md): alternating matrix factorization
+ * [`CF`](cf.md): collaborative filtering (recommender system)
+ * [`SparseCoding`](sparse_coding.md)
+-->
+
+ * [mlpack transformations](../../index.md#transformations)
+ * [Non-negative matrix factorization on Wikipedia](https://en.wikipedia.org/wiki/Non-negative_matrix_factorization)
+ * [Learning the parts of objects by non-negative matrix factorization](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=29bae9472203546847ec1352a604566d0f602728) (original NMF paper, pdf)
+
+### Constructors
+
+ * `nmf = NMF()`
+   - Create an `NMF` object.
+   - The rank of the decomposition is specified in the call to
+     [`Apply()`](#applying-decompositions).
+
+---
+
+ * `nmf = NMF(SimpleResidueTermination(minResidue=1e-5, maxIterations=10000))`
+   - Create an NMF object with custom termination parameters.
+   - `minResidue` (a `double`) specifies the minimum difference of the norm of
+     `W * H` between iterations for termination.
+   - `maxIterations` specifies the maximum number of iterations before
+     decomposition terminates.
+
+---
+
+### Applying Decompositions
+
+ * `double residue = nmf.Apply(V, rank, W, H)`
+   - Decompose the matrix `V` into two non-negative matrices `W` and `H` with
+     rank `rank`.
+   - `W` will be set to size `V.n_rows` x `rank`.
+   - `H` will be set to size `rank` x `V.n_cols`.
+   - `W` and `H` are initialized randomly using the
+     [Acol](#advanced-functionality-initializationruletype) initialization
+     strategy; i.e., each column of `W` is an average of 5 random columns of
+     `V`, and `H` is initialized uniformly randomly.
+   - The residue (change in the norm of `W * H` between iterations) is returned.
+
+---
+
+***Notes***:
+
+ - Low values of `rank` will give smaller matrices `W` and `H`, but the
+   decomposition will be less accurate.  Larger values of `rank` will give more
+   accurate decompositions, but will take longer to compute.  Every problem is
+   different, so `rank` must be specified manually.
+
+ - The expression `W * H` can be used to reconstruct the matrix `V`.
+
+ - Custom behavior, such as custom initialization of `W` and `H`, different or
+   custom termination rules, and different update rules are discussed in the
+   [advanced functionality](#advanced-functionality-template-parameters)
+   section.
+
+---
+
+#### `Apply()` Parameters:
+
+| **name** | **type** | **description** |
+|----------|----------|-----------------|
+| `V` | [`arma::sp_mat` or `arma::mat`](../matrices.md) | Input matrix to be factorized. |
+| `rank` | `size_t` | Rank of decomposition; lower is smaller, higher is more accurate. |
+| `W` | [`arma::mat`](../matrices.md) | Output matrix in which `W` will be stored. |
+| `H` | [`arma::mat`](../matrices.md) | Output matrix in which `H` will be stored. |
+
+***Note:*** Matrices with different element types can be used for `V`, `W`, and
+`H`; e.g., `arma::fmat`.  While `V` can be sparse or dense, `W` and `H` must be
+dense matrices.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`NMF`.
+
+---
+
+Decompose a dense matrix with custom termination parameters.
+
+```c++
+// Create a low-rank V matrix by multiplying together two random matrices.
+arma::mat V = arma::randu<arma::mat>(500, 25) *
+              arma::randn<arma::mat>(25, 5000);
+
+// Create the NMF object with a looser tolerance of 1e-3 and a maximum of 100
+// iterations only.
+mlpack::NMF nmf(mlpack::SimpleResidueTermination(1e-3, 500));
+
+arma::mat W, H;
+
+// Decompose with a rank of 25.
+// W will have size 500 x 25, and H will have size 25 x 5000.
+const double residue = nmf.Apply(V, 25, W, H);
+
+std::cout << "Residue of decomposition: " << residue << "." << std::endl;
+
+// Compute RMSE of decomposition.
+const double rmse = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+std::cout << "RMSE of decomposition: " << rmse << "." << std::endl;
+```
+
+---
+
+Decompose the sparse MovieLens dataset using a rank-12 decomposition and `float`
+element type.
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv.
+arma::sp_fmat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+// Create the NMF object.
+mlpack::NMF nmf;
+
+arma::fmat W, H;
+
+// Decompose the Movielens dataset with rank 12.
+const double residue = nmf.Apply(V, 12, W, H);
+
+std::cout << "Residue of MovieLens decomposition: " << residue << "."
+    << std::endl;
+
+// Compute RMSE of decomposition.
+const double rmse = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+std::cout << "RMSE of decomposition: " << rmse << "." << std::endl;
+```
+
+---
+
+Compare quality of decompositions of MovieLens with different ranks.
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv
+arma::sp_mat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+// Create the NMF object.
+mlpack::NMF nmf;
+arma::mat W, H;
+
+for (size_t rank = 10; rank <= 100; rank += 15)
+{
+  // Decompose with the given rank.
+  const double residue = nmf.Apply(V, rank, W, H);
+
+  const double rmse = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+  std::cout << "RMSE for rank-" << rank << " decomposition: " << rmse << "."
+      << std::endl;
+}
+```
+
+---
+
+### Advanced Functionality: Template Parameters
+
+The `NMF` class has three template parameters that can be used for custom
+behavior.  The full signature of the class is:
+
+```
+NMF<TerminationPolicyType, InitializationRuleType, UpdateRuleType>
+```
+
+ * `TerminationPolicyType`: the strategy used to choose when to terminate NMF.
+ * `InitializationRuleType`: the strategy used to choose the initial `W` and `H`
+   matrices.
+ * `UpdateRuleType`: the update rules used for NMF:
+   - `NMFMultiplicativeDistanceUpdate`: update rule that ensure the Frobenius
+     norm of the reconstruction error is decreasing at each iteration.
+   - `NMFMultiplicativeDivergenceUpdate`: update rules that ensure
+     Kullback-Leibler divergence is decreasing at each iteration.
+   - `NMFALSUpdate`: alternating least-squares projections for `W` and `H`.
+   - For custom update rules, use the more general
+     [`AMF`](/src/mlpack/methods/amf/amf.hpp) class.
+
+<!-- TODO: update link above to Markdown documentation -->
+
+---
+
+### Advanced Functionality: `TerminationPolicyType`
+
+ * Specifies the strategy to use to choose when to stop the NMF algorithm.
+ * An instantiated `TerminationPolicyType` can be passed to the NMF constructor.
+ * The following choices are available for drop-in usage:
+
+---
+
+#### ***`SimpleResidueTermination`*** (default):
+
+ - Terminates when a maximum number of iterations is reached, or when the
+   residue (change in norm of `W * H` between iterations) is sufficiently small.
+ - Constructor: `SimpleResidueTermination(minResidue=1e-5, maxIterations=10000)`
+   * `minResidue` (a `double`) specifies the sufficiently small residue for
+     termination.
+   * `maxIterations` (a `size_t`) specifies the maximum number of iterations.
+ - `nmf.Apply()` will return the residue of the last iteration.
+
+---
+
+#### ***`MaxIterationTermination`***:
+
+ - Terminates when the maximum number of iterations is reached.
+ - No other condition is checked.
+ - Constructor: `MaxIterationTermination(maxIterations=1000)`
+ - `nmf.Apply()` will return the number of iterations performed.
+
+---
+
+#### ***`SimpleToleranceTermination<MatType, WHMatType>`***:
+
+ - Terminates when the nonzero residual decreases a sufficiently small relative
+   amount between iterations (e.g.
+   `(lastNonzeroResidual - nonzeroResidual) / lastNonzeroResidual` is below a
+   threshold), or when the maximum number of iterations is reached.
+ - The residual must remain below the threshold for a specified number of
+   iterations.
+ - The nonzero residual is defined as the root of the sum of squared elements in
+   the reconstruction error matrix `(V - WH)`, limited to locations where `V` is
+   nonzero.
+ - Constructor: `SimpleToleranceTermination<MatType, WHMatType>(tol=1e-5, maxIter=10000, extraSteps=3)`
+   * `MatType` should be set to the type of `V` (see
+     [`Apply()` Parameters](#apply-parameters)).
+   * `WHMatType` (default `arma::mat`) should be set to the type of `W` and `H`
+     (see [`Apply()` Parameters](#apply-parameters)).
+   * `tol` (a `double`) specifies the relative nonzero residual tolerance for
+     convergence.
+   * `maxIter` (a `size_t`) specifies the maximum number of iterations
+     before termination.
+   * `extraSteps` (a `size_t`) specifies the number of iterations
+     where the relative nonzero residual must be below the tolerance for
+     convergence.
+ - The best `W` and `H` matrices (according to the nonzero residual) from the
+   final `extraSteps` iterations are returned by `nmf.Apply()`.
+ - `nmf.Apply()` will return the nonzero residue of the iteration corresponding
+   to the best `W` and `H` matrices.
+
+---
+
+#### ***`ValidationRMSETermination<MatType>`***:
+
+ - Holds out a validation set of nonzero elements from `V`, and terminates when
+   the RMSE (root mean squared error) on this validation set is sufficiently
+   small between iterations.
+ - The validation RMSE must remain below the threshold for a specified number of
+   iterations.
+ - `MatType` should be set to the type of `V` (see
+   [`Apply()` Parameters](#apply-parameters)).
+ - Constructor: `ValidationRMSETermination<MatType>(V, numValPoints, tol=1e-5, maxIter=10000, extraSteps=3)`
+   * `V` is the matrix to be decomposed by `Apply()`.  This will be modified
+     (validation elements will be removed).
+   * `numValPoints` (a `size_t`) specifies number of test points from `V` to be
+     held out.
+   * `tol` (a `double`) specifies the relative tolerance for the validation RMSE
+     for termination.
+   * `maxIter` (a `size_t`) specifies the maximum number of iterations before
+     termination.
+   * `extraSteps` (a `size_t`) specifies the number of iterations where the
+     validation RMSE must be below the tolerance for convergence.
+ - The best `W` and `H` matrices (according to the validation RMSE) from the
+   final `extraSteps` iterations are returned by `nmf.Apply()`.
+ - `nmf.Apply()` will return the best validation RMSE.
+
+---
+
+#### ***Custom policies***:
+
+ - A custom class for termination behavior must implement the following
+   functions.
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomTerminationPolicy
+{
+ public:
+  // Initialize the termination policy for the given matrix V.  (It is okay to
+  // do nothing.)  This function is called at the beginning of Apply().
+  //
+  // If the termination policy requires V to compute convergence, store a
+  // reference or pointer to it in this function.
+  template<typename MatType>
+  void Initialize(const MatType& V);
+
+  // Check if convergence has occurred for the given W and H matrices.  Return
+  // `true` if so.
+  //
+  // Note that W and H may have different types than V (i.e. V may be sparse,
+  // and W and H must be dense.)
+  template<typename WHMatType>
+  bool IsConverged(const MatType& H, const MatType& W);
+
+  // Return the value that should be returned for the `nmf.Apply()` function
+  // when convergence has been reached.  This is called at the end of
+  // `nmf.Apply()`.
+  const double Index();
+
+  // Return the number of iterations that have been completed.  This is called
+  // at the end of `nmf.Apply()`.
+  const size_t Iteration();
+};
+```
+
+---
+
+### Advanced Functionality: `InitializationRuleType`
+
+ * Specifies the strategy to use to initialize `W` and `H` at the beginning of
+   the NMF algorithm.
+ * An initialization `InitializationRuleType` can be passed to the following
+   constructor:
+   - `nmf = NMF(terminationPolicy, initializationRule)`
+ * The following choices are available for drop-in usage:
+
+---
+
+#### ***`RandomAcolInitialization<N>`*** (default):
+
+ - Initialize `W` by averaging `N` randomly chosen columns of `V`.
+ - Initialize `H` as uniform random in the range `[0, 1]`.
+ - The default value for `N` is 5.
+ - See also [the paper](https://arxiv.org/abs/1407.7299) describing the
+   strategy.
+
+---
+
+#### ***`NoInitialization`***:
+
+ - When `nmf.Apply(V, rank, W, H)`, the existing values of `W` and `H` will be
+   used.
+ - If `W` is not of size `V.n_rows` x `rank`, or if `H` is not of size `rank` x
+   `V.n_cols`, a `std::invalid_argument` exception will be thrown.
+
+---
+
+#### ***`GivenInitialization<MatType>`***:
+
+ - Set `W` and/or `H` to the given matrices when `Apply()` is called.
+ - `MatType` should be set to the type of `W` or `H` (default `arma::mat`); see
+   [`Apply()` Parameters](#apply-parameters).
+ - Constructors:
+   * `GivenInitialization<MatType>(W, H)`
+     - Specify both initial `W` and `H` matrices.
+   * `GivenInitialization<MatType>(M, isW=true)`
+     - If `isW` is `true`, then set initial `W` to `M`.
+     - If `isW` is `false`, then set initial `H` to `M`.
+     - This constructor is meant to only be used with `MergeInitialization`
+       (below).
+
+---
+
+#### ***`RandomInit`***:
+
+ - Initialize `W` and `H` as uniform random in the range `[0, 1]`.
+
+---
+
+#### ***`AverageInitialization`***:
+
+ - Initialize each element of `W` and `H` to the square root of the average
+   value of `V`, adding uniform random noise in the range `[0, 1]`.
+
+---
+
+#### ***`MergeInitialization<WRule, HRule>`***:
+
+ - Use two different initialization rules, one for `W` (`WRule`) and one for `H`
+   (`HRule`).
+ - Constructors:
+   * `MergeInitialization<WRule, HRule>()`
+     - Create the merge initialization with default-constructed rules for `W`
+       and `H`.
+   * `MergeInitialization<WRule, HRule>(wRule, hRule)`
+     - Create the merge initialization with instantiated rules for `W` and `H`.
+     - `wRule` and `hRule` will be copied.
+ - Any `WRule` and `HRule` classes must implement the `InitializeOne()`
+   function.
+
+---
+
+#### ***Custom rules***:
+
+ - A custom class for initializing `W` and `H` must implement the following
+   functions.
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomInitialization
+{
+ public:
+  // Initialize the W and H matrices, given V and the rank of the decomposition.
+  // This is called at the start of `Apply()`.
+  //
+  // Note that `MatType` may be different from `WHMatType`; e.g., `V` could be
+  // sparse, but `W` and `H` must be dense.
+  template<typename MatType, typename WHMatType>
+  void Initialize(const MatType& V,
+                  const size_t rank,
+                  WHMatType& W,
+                  WHMatType& H);
+
+  // Initialize one of the W or H matrices, given V and the rank of the
+  // decomposition.
+  //
+  // If `isW` is `true`, then `M` should be treated as though it is `W`;
+  // if `isW` is `false`, then `M` should be treated as thought it is `H`.
+  //
+  // This function only needs to be implemented if it is intended to use the
+  // custom initialization strategy with `MergeInitialization`.
+  template<typename MatType, typename WHMatType>
+  void InitializeOne(const MatType& V,
+                     const size_t rank,
+                     WHMatType& M,
+                     const bool isW);
+};
+```
+
+---
+
+### Advanced Functionality Examples
+
+Use a pre-specified initialization for `W` and `H`.
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv.
+arma::sp_mat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+arma::mat W, H;
+
+// Pre-initialize W and H.
+// W will be filled with random values from a normal distribution.
+// H will be filled with 1s.
+W.randn(V.n_rows, 15);
+H.set_size(15, V.n_cols);
+H.fill(0.2);
+
+mlpack::NMF<mlpack::SimpleResidueTermination, mlpack::NoInitialization> nmf;
+const double residue = nmf.Apply(V, 15, W, H);
+const double rmse = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+
+std::cout << "RMSE of NMF decomposition with pre-specified W and H: " << rmse
+    << "." << std::endl;
+```
+
+---
+
+Use `ValidationRMSETermination` to decompose the MovieLens dataset until the
+RMSE of the held-out validation set is sufficiently low.
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv.
+arma::sp_mat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+arma::mat W, H;
+
+// Create a ValidationRMSETermination class that will hold out 3k points from V.
+// This will remove 3000 nonzero entries from V.
+mlpack::ValidationRMSETermination<arma::sp_mat> t(V, 3000);
+
+// Create the NMF object with the instantiated termination policy.
+mlpack::NMF<mlpack::ValidationRMSETermination<arma::sp_mat>> nmf(t);
+
+// Perform NMF with a rank of 20.
+// Note the RMSE returned here is the RMSE on the validation set.
+const double rmse = nmf.Apply(V, 20, W, H);
+const double rmseTrain = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+
+std::cout << "Training RMSE:   " << rmseTrain << "." << std::endl;
+std::cout << "Validation RMSE: " << rmse << "." << std::endl;
+```
+
+---
+
+Use all three sets of NMF update rules and compare the RMSE on a held-out
+validation set.
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv.
+arma::sp_mat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+arma::mat W1, W2, W3;
+arma::mat H1, H2, H3;
+
+// Create a ValidationRMSETermination class that will hold out 3k points from V.
+// This will remove 3000 nonzero entries from V.
+mlpack::ValidationRMSETermination<arma::sp_mat> t(V, 3000);
+
+// Multiplicative distance update rule.
+mlpack::NMF<mlpack::ValidationRMSETermination<arma::sp_mat>,
+            mlpack::RandomAcolInitialization<5>,
+            mlpack::NMFMultiplicativeDistanceUpdate> nmf1(t);
+
+// Multiplicative divergence update rule.
+mlpack::NMF<mlpack::ValidationRMSETermination<arma::sp_mat>,
+            mlpack::RandomAcolInitialization<5>,
+            mlpack::NMFMultiplicativeDivergenceUpdate> nmf2(t);
+
+// Alternating least squares update rule.
+mlpack::NMF<mlpack::ValidationRMSETermination<arma::sp_mat>,
+            mlpack::RandomAcolInitialization<5>,
+            mlpack::NMFALSUpdate> nmf3(t);
+
+const double rmse1 = nmf1.Apply(V, 15, W1, H1);
+const double rmse2 = nmf2.Apply(V, 15, W2, H2);
+const double rmse3 = nmf3.Apply(V, 15, W3, H3);
+
+// Print the RMSEs.
+std::cout << "Mult. dist. update RMSE: " << rmse1 << "." << std::endl;
+std::cout << "Mult. div. update RMSE:  " << rmse2 << "." << std::endl;
+std::cout << "ALS update RMSE:         " << rmse3 << "." << std::endl;
+```
+
+---
+
+Use a custom termination policy that sets a limit on how long NMF is allowed to
+take.  First, we define the termination policy:
+
+```c++
+class CustomTimeTermination
+{
+ public:
+  CustomTimeTermination(const double totalAllowedTime) :
+      totalAllowedTime(totalAllowedTime) { }
+
+  template<typename MatType>
+  void Initialize(const MatType& /* V */)
+  {
+    totalTime = 0.0;
+    iteration = 0;
+    c.tic();
+  }
+
+  template<typename WHMatType>
+  bool IsConverged(const WHMatType& /* W */, const WHMatType& /* H */)
+  {
+    totalTime += c.toc();
+    c.tic();
+    ++iteration;
+    return (totalTime > totalAllowedTime);
+  }
+
+  const double Index() const { return totalTime; }
+  const size_t Iteration() const { return iteration; }
+
+ private:
+  double totalAllowedTime;
+  double totalTime;
+  size_t iteration;
+  arma::wall_clock c; // used for convenient timing
+};
+```
+
+Then we can use it in the test program:
+
+```c++
+// See https://datasets.mlpack.org/movielens-100k.csv.
+arma::sp_fmat V;
+mlpack::data::Load("movielens-100k.csv", V, true);
+
+CustomTimeTermination t(5 /* seconds */);
+mlpack::NMF<CustomTimeTermination> nmf(t);
+
+arma::fmat W, H;
+const double actualTime = nmf.Apply(V, 10, W, H);
+const double rmse = arma::norm(V - W * H, "fro") / std::sqrt(V.n_elem);
+
+std::cout << "Actual time used for decomposition: " << actualTime << "."
+    << std::endl;
+std::cout << "RMSE after ~5 seconds: " << rmse << "." << std::endl;
+```
diff -pruN 4.3.0-2/doc/user/methods/pca.md 4.4.0-1/doc/user/methods/pca.md
--- 4.3.0-2/doc/user/methods/pca.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/pca.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,326 @@
+## `PCA`
+
+The `PCA` class implements principal components analysis (PCA), a standard
+machine learning data preparation technique.  PCA can be used to reduce the
+number of dimensions in a dataset, or to preserve a certain percentage of the
+variance of a dataset.
+
+By default, `PCA` uses the full exact singular value decomposition (SVD), but
+supports the use of other more efficient decompositions, including approximate
+singular value decompositions.
+
+#### Simple usage example:
+
+```c++
+// Use PCA to reduce the number of dimensions to 5 on uniform random data.
+
+// This dataset is uniform random in 10 dimensions.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+
+mlpack::PCA pca;       // Step 1: create PCA object.
+pca.Apply(dataset, 5); // Step 2: reduce data dimension to 5.
+
+// Print some information about the modified dataset.
+std::cout << "The transformed data matrix has size " << dataset.n_rows /* 5 */
+    << " x " << dataset.n_cols << "." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `PCA` objects.
+ * [`Apply()`](#applying-transformations): apply PCA transformation to data.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-decomposition-strategies)
+   for using different decomposition strategies.
+
+#### See also:
+
+<!-- TODO: add link -->
+<!-- * [`RADICAL`](radical.md): independent components analysis -->
+ * [mlpack preprocessing utilities](../../index.md#preprocessing-utilities)
+ * [mlpack transformations](../../index.md#transformations)
+ * [Principal component analysis on Wikipedia](https://en.wikipedia.org/wiki/Principal_component_analysis)
+
+### Constructors
+
+ * `pca = PCA(scaleData=false)`
+   - Construct a `PCA` object.
+   - If `scaleData` is `true`, then all dimensions will have variance scaled to
+     1 before applying PCA.
+   - The `scaleData` parameter can be inspected with `pca.ScaleData()`, and also
+     set; `pca.ScaleData() = true` will enable data variance scaling.
+
+---
+
+ * `pca = PCA(scaleData, decompositionPolicy)`
+   - Construct a `PCA` object with a custom decomposition policy.
+   - See the documentation for using
+     [different decomposition strategies](#advanced-functionality-different-decomposition-strategies).
+
+### Applying Transformations
+
+ * `pca.Apply(data, transformedData)`
+ * `pca.Apply(data, transformedData, eigVal)`
+ * `pca.Apply(data, transformedData, eigVal, eigVec)`
+   - Transform the
+     [column-major matrix](../matrices.md#representing-data-in-mlpack) `data`
+     using PCA, storing the result in `transformedData`.
+   - `data` should be a floating-point matrix (e.g. `arma::mat`, `arma::fmat`,
+     `arma::sp_mat`, etc.) or an expression that evaluates to one.
+   - `transformedData` should be a dense floating-point matrix (e.g.,
+     `arma::mat`, `arma::sp_mat`).
+   - The size of `transformedData` will be the same as the size of `data`.
+   - Dimensions in `transformedData` will be ordered decreasing in variance;
+     that is, the first row of `transformedData` will correspond to the
+     dimension with maximum variance.
+   - Optionally, eigenvalues and eigenvectors of the covariance matrix can be
+     returned:
+     * If specified, `eigVal` should be a dense floating-point vector (e.g.
+       `arma::vec`, `arma::fvec`, etc.) and will be filled with the eigenvalues
+       of `transformedData`.
+     * If specified, `eigvec` should be a dense floating-point matrix (e.g.
+       `arma::mat`, `arma::fmat`, etc.) and will be filled with the eigenvectors
+       of `transformedData`.
+
+---
+
+ * `double varRetained = pca.Apply(data, transformedData, newDimension)`
+   - Use PCA to reduce the number of dimensions in the
+     [column-major matrix](../matrices.md#representing-data-in-mlpack) `data`
+     to `newDimension`, storing the result in `transformedData`.
+   - `data` should be a floating-point matrix (e.g. `arma::mat`,
+     `arma::fmat`, `arma::sp_mat`, etc.) or an expression that evaluates to
+     one.
+   - `transformedData` should be a dense floating-point matrix with the same
+     element type as `data` (e.g. `arma::mat`, `arma::fmat`).
+   - `transformedData` will have `newDimension` rows after the transformation.
+   - Returns a `double` indicating the percentage of variance retained (between
+     `0.0` and `1.0`).
+
+---
+
+ * `double varRetained = pca.Apply(data, transformedData, varianceToKeep)`
+   - Use PCA to retain the dimensions of the
+     [column-major matrix](../matrices.md#representing-data-in-mlpack) `data`
+     that capture a factor of `varianceToKeep` of the data variance.
+   - `data` should be a floating-point matrix (e.g. `arma::mat`, `arma::fmat`,
+     `arma::sp_mat`, etc.) or an expression that evaluates to one.
+   - `transformedData` should be a dense floating-point matrix with the same
+     element type as `data` (e.g. `arma::mat`, `arma::fmat`).
+   - `transformedData` will have `newDimension` rows after the transformation.
+   - `varianceToKeep` should be a floating-point value between `0.0` and `1.0`.
+     If `1.0`, all of the data variance is retained, and this is equivalent to
+     the first version of `Apply()` (above).
+   - Returns a `double` indicating the percentage of variance actually retained
+     (between `0.0` and `1.0`).
+
+---
+
+ * `double varRetained = pca.Apply(data, newDimension)`
+ * `double varRetained = pca.Apply(data, varianceToKeep)`
+   - In-place versions of the two `Apply()` functions above.
+   - Equivalent to `pca.Apply(data, data, newDimension)` or
+     `pca.Apply(data, data, varianceToKeep)`.
+   - `data` should be a dense floating-point matrix (e.g. `arma::mat`,
+     `arma::fmat`, etc.).
+
+---
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `PCA` class.
+
+---
+
+Apply PCA to a dataset, keeping dimensions that capture 90% of the data
+variance.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat data;
+mlpack::data::Load("satellite.train.csv", data, true);
+const size_t origDim = data.n_rows;
+
+mlpack::PCA pca;
+
+// Keep 90% of the data variance.
+pca.Apply(data, 0.9);
+
+std::cout << "PCA kept " << data.n_rows << " of " << origDim << " dimensions "
+    << "to capture 90\% of the data variance." << std::endl;
+```
+
+---
+
+Apply PCA to a 32-bit floating point dataset with dimension scaling, keeping all
+dimensions, and printing the 5 largest eigenvalues of the covariance matrix of
+the transformed data.
+
+```c++
+// See https://datasets.mlpack.org/iris.csv.
+arma::fmat data;
+mlpack::data::Load("iris.csv", data, true);
+
+mlpack::PCA pca(true /* scale data when transforming */);
+
+arma::fvec eigval;
+arma::fmat transformedData;
+
+pca.Apply(data, transformedData, eigval);
+
+std::cout << "First point, before PCA: " << data.col(0).t();
+std::cout << "First point, after PCA:  " << transformedData.col(0).t();
+std::cout << std::endl;
+
+// Now print the top 5 eigenvalues.
+for (size_t i = 0; i < 5; ++i)
+  std::cout << "Eigenvalue " << i << ": " << eigval[i] << "." << std::endl;
+```
+
+---
+
+Apply PCA to a random sparse dataset, to reduce the dimensionality to a
+20-dimensional dense dataset.
+
+```c++
+arma::sp_mat data;
+// This dataset has 10k points in 1k dimensions, with 1% density.
+data.sprandn(1000, 10000, 0.01);
+
+mlpack::PCA pca(true /* scale data when transforming */);
+
+arma::mat transformedData;
+const double varianceRetained = pca.Apply(data, transformedData, 20);
+
+std::cout << "First point, before PCA: " << data.col(0).t();
+std::cout << "First point, after PCA: " << transformedData.col(0).t();
+
+// Note that for random uniform data, this won't capture very much of the
+// variance!  It would be much more for a real, structured dataset.
+std::cout << "50 dimensions captured " << (100.0 * varianceRetained) << "\% of "
+    << "the data variance." << std::endl;
+```
+
+---
+
+### Advanced Functionality: Different Decomposition Strategies
+
+By default, `PCA` uses the full exact singular value decomposition (SVD) to
+transform data.  However, for very large datasets, it may be faster to use
+alternative strategies, some of which may be approximate.  The `PCA` class has
+one template parameter that allows different decomposition strategies to be
+used.  The full signature of the class is:
+
+```
+PCA<DecompositionPolicy>
+```
+
+`DecompositionPolicy` specifies the strategy to be used to compute the singular
+values and vectors of a data matrix.
+
+Several decomposition policies are already implemented and ready for drop-in
+usage:
+
+ * `ExactSVDPolicy` _(default)_: use Armadillo's `svd()` and `svd_econ()`
+   functions to compute the SVD
+ * `RandomizedSVDPCAPolicy`: use the randomized SVD algorithm to compute the SVD
+   <!-- TODO: add link to documentation! -->
+ * `RandomizedBlockKrylovSVDPolicy`: use the randomized Block Krylov SVD
+   algorithm to compute the SVD <!-- TODO: add link to documentation! -->
+ * `QUICSVDPolicy`: use the tree-based `QUIC-SVD` algorithm to compute the SVD
+   <!-- TODO: add link to documentation -->
+
+The simple example program below uses all four decomposition types on the same
+MNIST data, timing how long each decomposition takes.
+
+```c++
+arma::mat data;
+// See https://datasets.mlpack.org/mnist.train.csv.
+mlpack::data::Load("mnist.train.csv", data, true);
+
+arma::mat output1, output2, output3, output4;
+
+mlpack::PCA<mlpack::ExactSVDPolicy> pca1;
+mlpack::PCA<mlpack::RandomizedSVDPCAPolicy> pca2;
+mlpack::PCA<mlpack::RandomizedBlockKrylovSVDPolicy> pca3;
+mlpack::PCA<mlpack::QUICSVDPolicy> pca4;
+
+// Compute decompositions on all four, timing each one.
+arma::wall_clock c;
+
+c.tic();
+pca1.Apply(data, output1);
+const double pca1Time = c.toc();
+
+c.tic();
+pca2.Apply(data, output2);
+const double pca2Time = c.toc();
+
+c.tic();
+pca3.Apply(data, output3);
+const double pca3Time = c.toc();
+
+c.tic();
+pca4.Apply(data, output4);
+const double pca4Time = c.toc();
+
+std::cout << "PCA computation times for " << data.n_rows << " x " << data.n_cols
+    << " data:" << std::endl;
+std::cout << " - ExactSVDPolicy:                 " << pca1Time << "s."
+    << std::endl;
+std::cout << " - RandomizedSVDPCAPolicy:         " << pca2Time << "s."
+    << std::endl;
+std::cout << " - RandomizedBlockKrylovSVDPolicy: " << pca3Time << "s."
+    << std::endl;
+std::cout << " - QUICSVDPolicy:                  " << pca4Time << "s."
+    << std::endl;
+```
+
+---
+
+#### Custom decomposition policies
+
+Instead of using the predefined classes above, it is also possible to implement
+fully custom functionality via a new decomposition policy.  Any new
+decomposition policy must implement one method:
+
+```c++
+class CustomDecompositionPolicy
+{
+ public:
+  // Given input data `data` and `centeredData`, compute the singular value
+  // decomposition of the data, and then project the data onto the first `rank`
+  // singular vectors.
+  //
+  //  * `data` is the input matrix.  It is not guaranteed to be centered or
+  //      scaled.
+  //  * `centeredData` is the centered (and possibly scaled) version of the
+  //      input matrix (e.g. the mean of each dimension is 0).
+  //  * `transformedData` should be overwritten with the centered data's
+  //      projection onto the singular vectors.
+  //  * `svals` and `svecs` should be filled with the singular values and
+  //      vectors of the centered data.
+  //  * `rank` specifies the number of singular values/vectors to keep, and the
+  //      dimension of `transformedData` should be equivalent to `rank`.  `rank`
+  //      will be at most equal to `data.n_rows`.
+  //
+  //  * `InMatType` is a dense floating-point matrix type, but may be a subview
+  //      or expression.
+  //  * `MatType` is the type of matrix used to represent data, and will be a
+  //      dense floating-point matrix type (e.g. `arma::mat`, `arma::fmat`,
+  //      etc.).
+  //  * `VecType` is the corresponding vector type to `MatType` (e.g., a
+  //      `MatType` of `arma::mat` would mean a `VecType` of `arma::vec`, etc.).
+  template<typename MatType, typename MatType, typename VecType>
+  static void Apply(const InMatType& data,
+                    const MatType& centeredData,
+                    MatType& transformedData,
+                    VecType& svals,
+                    MatType& svecs,
+                    const size_t rank);
+};
+```
diff -pruN 4.3.0-2/doc/user/methods/perceptron.md 4.4.0-1/doc/user/methods/perceptron.md
--- 4.3.0-2/doc/user/methods/perceptron.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/perceptron.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,412 @@
+## `Perceptron`
+
+The `Perceptron` class implements the simple perceptron classifier originally
+implemented by Frank Rosenblatt in 1958.  The perceptron is a linear classifier,
+and can be understood as a trivial neural network with one neuron that uses the
+step function as an activation function.  mlpack's implementation of the
+`Perceptron` class also offers several template parameters that can be used to
+control the behavior of the perceptron.
+
+Perceptrons are useful for classifying points with _discrete labels_ (i.e., `0`,
+`1`, `2`).  Because they are simple classifiers, they are also useful as _weak
+learners_ for the [`AdaBoost`](adaboost.md) boosting classifier.
+
+#### Simple usage example:
+
+```c++
+// Train a perceptron on random numeric data and predict labels on test data:
+
+// All data and labels are uniform random; 10 dimensional data, 5 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+arma::mat testDataset(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::Perceptron p;                 // Step 1: create model.
+p.Train(dataset, labels, 5);          // Step 2: train model.
+arma::Row<size_t> predictions;
+p.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 1) << " test points classified as class "
+    << "1." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `DecisionTree` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+ * [Advanced template examples](#advanced-functionality-examples) of use with
+   custom template parameters.
+
+#### See also:
+
+ * [`NaiveBayesClassifier`](naive_bayes_classifier.md), another simple classifier
+ * [`AdaBoost`](adaboost.md)
+ * [`FFN`](/src/mlpack/methods/ann/ffn.hpp)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [Perceptron on Wikipedia](https://en.wikipedia.org/wiki/Perceptron)
+
+### Constructors
+
+Construct a `Perceptron` object using one of the constructors below.  Defaults
+and types are detailed in the [Constructor Parameters](#constructor-parameters)
+section below.
+
+#### Forms:
+
+ * `p = Perceptron()`
+   - Initialize perceptron without training.
+   - You will need to call [`Train()`](#training) later to train the perceptron
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `p = Perceptron(numClasses, dimensionality, maxIterations=1000)`
+   - Initialize perceptron with all-zero weights and biases.
+   - `Classify()` can immediately be used; training is not required with this
+     form.
+
+---
+
+ * `p = Perceptron(data, labels, numClasses,          maxIterations=1000)`
+ * `p = Perceptron(data, labels, numClasses, weights, maxIterations=1000)`
+   - Train the perceptron (optionally with instance weights).
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `datasetInfo` | [`data::DatasetInfo`](../load_save.md#mixed-categorical-data) | Dataset information, specifying type information for each dimension. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, between [`0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `weights` | [`arma::rowvec`](../matrices.md) | Weights for each training point.  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `dimensionality` | `size_t` | Dimensionality of data (only used if an initialized but untrained model is desired). | _(N/A)_ |
+| `maxIterations` | `size_t` | Maximum number of iterations during training.  Can also be set with `MaxIterations()`. | `1000` |
+
+As an alternative to passing `maxIterations`, it can be set with a standalone
+method.  The following function can be used before calling `Train()` to set
+the maximum number of iterations:
+
+ * `p.MaxIterations() = maxIter;` will set the maximum number of iterations
+   during training to `maxIter`.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the following versions of the `Train()` member function:
+
+ * `p.Train(data, labels, numClasses, maxIterations=1000)`
+   - Train the perceptron on unweighted data.
+
+---
+
+ * `p.Train(data, labels, numClasses, weights, maxIterations=1000)`
+   - Train the perceptron on data with instance weights.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is incremental.  Successive calls to `Train()` will not reinitialize
+   the model, unless the given data has different dimensionality or `numClasses`
+   is different.  To reinitialize the model, call `Reset()` (see
+   [Other Functionality](#other-functionality)).
+
+ * If `maxIterations` is not passed, but has been set in the constructor or with
+   `MaxIterations()`, the previous setting will be used.
+
+### Classification
+
+Once a `Perceptron` is trained, the `Classify()` member function can be used to
+make class predictions for new data.
+
+ * `size_t predictedClass = p.Classify(point)`
+    - ***(Single-point)***
+    - Classify a single point, returning the predicted class.
+
+---
+
+ * `p.Classify(data, predictions)`
+    - ***(Multi-point)***
+    - Classify a set of points.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+***Note***: perceptrons do not provide any measure resembling probabilities
+during classification, and thus a version of `Classify()` that computes class
+probabilities is not available.
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into.  Will be set to length `data.n_cols`. |
+
+### Other Functionality
+
+ * A `Perceptron` can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `p.NumClasses()` will return a `size_t` indicating the number of classes the
+   perceptron was trained on.
+
+ * `p.Biases()` will return an `arma::vec` with the biases of the model (each
+   element corresponds to the bias for a class).
+
+ * `p.Weights()` will return an `arma::mat` with the weights of the model (each
+   column corresponds to the weights for one class label).
+
+ * `p.Reset()` will re-initialize the weights and biases of the model.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/perceptron/perceptron.hpp) can be consulted.  Each
+method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`Perceptron`.
+
+---
+
+Train a perceptron multiple times, incrementally, with custom hyperparameters,
+and save the resulting model to disk.
+
+```c++
+// See https://datasets.mlpack.org/iris.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.csv", dataset, true);
+// See https://datasets.mlpack.org/iris.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("iris.labels.csv", labels, true);
+
+// Create a Perceptron object.
+mlpack::Perceptron p;
+// Set the maximum number of iterations to 100.  (This can also be done in the
+// constructor.)
+p.MaxIterations() = 100;
+
+// Train the model for up to 100 iterations.
+p.Train(dataset, labels, 3);
+
+// Now, compute and print accuracy on the training set.
+arma::Row<size_t> predictions;
+p.Classify(dataset, predictions);
+std::cout << "Training set accuracy after 100 iterations: "
+    << (100.0 * double(arma::accu(labels == predictions)) / labels.n_elem)
+    << "\%." << std::endl;
+
+// Train for another 250 iterations and compute training set accuracy again.
+p.Train(dataset, labels, 3, 250);
+p.Classify(dataset, predictions);
+std::cout << "Training set accuracy after 350 iterations: "
+    << (100.0 * double(arma::accu(labels == predictions)) / labels.n_elem)
+    << "\%." << std::endl;
+
+// Save the perceptron to disk for later use.
+mlpack::data::Save("perceptron.bin", "perceptron", p);
+```
+
+---
+
+Load a saved perceptron from disk and print information about it.
+
+```c++
+mlpack::Perceptron p;
+// This call assumes a perceptron called "p" has already been saved to
+// `perceptron.bin` with `data::Save()`.
+mlpack::data::Load("perceptron.bin", "p", p, true);
+
+if (p.NumClasses() > 0)
+{
+  std::cout << "The perceptron in `perceptron.bin` was trained on "
+      << p.NumClasses() << " classes." << std::endl;
+  std::cout << "The dimensionality of the perceptron model is "
+      << p.Weights().n_rows << "." << std::endl;
+  std::cout << "The bias weights for each class are:" << std::endl;
+  for (size_t i = 0; i < p.NumClasses(); ++i)
+    std::cout << "  - Class " << i << ": " << p.Biases()[i] << std::endl;
+}
+else
+{
+  std::cout << "The perceptron in `perceptron.bin` has not been trained."
+      << std::endl;
+}
+```
+
+---
+
+### Advanced Functionality: Template Parameters
+
+The `Perceptron` class also supports several template parameters, which can be
+used for custom behavior.  The full signature of the class is as follows:
+
+```
+Perceptron<LearnPolicy,
+           WeightInitializationPolicy,
+           MatType>
+```
+
+ * `LearnPolicy`: the strategy used to learn the weights during training.
+ * `WeightInitializationPolicy`: the way that weights are initialized before
+   training.
+ * `MatType`: specifies the type of matrix used for learning and internal
+   representation of weights and biases.
+
+---
+
+#### `LearnPolicy`
+
+ * Specifies the step to be taken when a point is misclassified.
+ * The `SimpleWeightUpdate` class is available, and is the default.
+ * A custom class must implement only one function:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomLearnPolicy
+{
+  // Update the weights and biases in the `weights` matrix and the `biases`
+  // vector given that the model currently classified `trainingPoint` as having
+  // the label `incorrectClass`, when in reality it has the label
+  // `correctClass`.  If `instanceWeight` is given, it specifies the instance
+  // weight for the given `trainingPoint`.
+  //
+  // `VecType` will be an Armadillo-like vector type.  It will be a column from
+  // the training data matrix (`data`) given to `Train()` or to the constructor.
+  //
+  // `eT` is the element type of the Perceptron (e.g. `float`, `double`).
+  template<typename VecType, typename eT>
+  void UpdateWeights(const VecType& trainingPoint,
+                     arma::Mat<eT>& weights,
+                     arma::Col<eT>& biases,
+                     const size_t incorrectClass,
+                     const size_t correctClass,
+                     const double instanceWeight = 1.0);
+};
+```
+
+---
+
+#### `WeightInitializationPolicy`
+
+ * Specifies how the weights matrix and biases vector should be initialized when
+   the `Perceptron` object is created, or when `Reset()` is called.
+ * The `ZeroInitialization` _(default)_ and `RandomPerceptronInitialization`
+   classes are available for drop-in usage.
+ * `RandomPerceptronInitialization` will initialize weights and biases using a
+   uniform random distribution between 0 and 1.
+ * A custom class must implement only one function:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomWeightInitializationPolicy
+{
+  // Initialize the `weights` matrix and `biases` vector, given that the model
+  // will have dimensionality of `numFeatures` (that is, the training data
+  // matrix will have `numFeatures` rows), and the training data has
+  // `numClasses` classes.
+  //
+  // The initialized `weights` matrix should have `numFeatures` rows and
+  // `numClasses` columns, and the initialized `biases` vector should have
+  // `numClasses` elements.
+  //
+  // `eT` specifies the element type of the weights and biases; it may be
+  // `double`, `float`, or another floating-point type.
+  template<typename eT>
+  inline static void Initialize(arma::Mat<eT>& weights,
+                                arma::Col<eT>& biases,
+                                const size_t numFeatures,
+                                const size_t numClasses)
+  {
+    weights.randu(numFeatures, numClasses);
+    biases.randu(numClasses);
+  }
+};
+```
+
+---
+
+#### `MatType`
+
+ * Specifies the matrix type to use for data when learning a perceptron.
+ * By default, `MatType` is `arma::mat` (dense 64-bit precision matrix).
+ * Any matrix type implementing the Armadillo API will work; so, for instance,
+   `arma::fmat` or `arma::sp_mat` can be used.
+
+### Advanced Functionality Examples
+
+Train a `Perceptron` with random initialization, instead of zero initialization
+of weights.
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::mat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.  Weights will be initialized randomly.
+mlpack::Perceptron<mlpack::SimpleWeightUpdate,
+                   mlpack::RandomPerceptronInitialization> p(
+    dataset, labels, 5);
+
+// Create test data (500 points).
+arma::mat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+p.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 1) << " test points classified as class "
+    << "1." << std::endl;
+```
+
+---
+
+Train a `Perceptron` on sparse 32-bit floating point data.
+
+```c++
+
+// 1000 sparse random points in 100 dimensions, with 1% nonzero elements.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 1000, 0.01);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.
+mlpack::Perceptron p(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::sp_fmat testDataset;
+testDataset.sprandu(100, 500, 0.01);
+arma::Row<size_t> predictions;
+p.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 1) << " test points classified as class "
+    << "1." << std::endl;
+```
+
+---
diff -pruN 4.3.0-2/doc/user/methods/random_forest.md 4.4.0-1/doc/user/methods/random_forest.md
--- 4.3.0-2/doc/user/methods/random_forest.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/random_forest.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,714 @@
+## `RandomForest`
+
+The `RandomForest` class implements a parallelized random forest classifier that
+supports numerical and categorical features, by default using Gini gain to
+choose which feature to split on in each tree.
+
+Random forests are a collection of decision trees that give better performance
+than a single decision tree.  They are useful for classifying points with
+_discrete labels_ (i.e.  `0`, `1`, `2`).  This implementation of the
+`RandomForest` class is not for regression (i.e. predicting _continuous
+values_).
+
+mlpack's `RandomForest` class offers configurability via template parameters and
+runtime parameters.  This is used to provide the additional API-compatible
+`ExtraTrees` class.  To use `ExtraTrees`, simply replace `RandomForest` with
+`ExtraTrees` in any of the documentation below.  ([More
+information...](#fully-custom-behavior))
+
+#### Simple usage example:
+
+```c++
+// Train a random forest on random numeric data and predict labels on test data:
+
+// All data and labels are uniform random; 10 dimensional data, 5 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(10, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+arma::mat testData(10, 500, arma::fill::randu); // 500 test points.
+
+mlpack::RandomForest rf;            // Step 1: create model.
+rf.Train(dataset, labels, 5, 10);   // Step 2: train model.
+arma::Row<size_t> predictions;
+rf.Classify(testData, predictions); // Step 3: classify points.
+// You can also use `ExtraTrees` instead of `RandomForest`!
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 3) << " test points classified as class "
+    << "3." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `RandomForest` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for custom
+   behavior.
+
+#### See also:
+
+ * [`DecisionTree`](decision_tree.md)
+ * [`DecisionTreeRegressor`](decision_tree_regressor.md)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [Random forest on Wikipedia](https://en.wikipedia.org/wiki/Random_forest)
+ * [Decision tree on Wikipedia](https://en.wikipedia.org/wiki/Decision_tree)
+ * [Leo Breiman's Random Forests page](https://www.stat.berkeley.edu/~breiman/RandomForests/cc_home.htm)
+
+### Constructors
+
+ * `rf = RandomForest()`
+   - Initialize the random forest without training.
+   - You will need to call [`Train()`](#training) later to train the tree
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `rf = RandomForest(data, labels, numClasses,          numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0)`
+ * `rf = RandomForest(data, labels, numClasses, weights, numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0)`
+   - Train on numerical-only data (optionally with instance weights).
+
+---
+
+ * `rf = RandomForest(data, info, labels, numClasses,          numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0)`
+ * `rf = RandomForest(data, info, labels, numClasses, weights, numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `info` | [`data::DatasetInfo`](../load_save.md#loading-categorical-data) | Dataset information, specifying type information for each dimension. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `weights` | [`arma::rowvec`](../matrices.md) | Instance weights for each training point.  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numTrees` | `size_t` | Number of trees to train in the random forest. | `20`
+|
+| `minLeafSize` | `size_t` | Minimum number of points in each leaf node of each decision tree. | `1` |
+| `minGainSplit` | `double` | Minimum gain for a node to split in each decision tree. | `1e-7` |
+| `maxDepth` | `size_t` | Maximum depth for each decision tree. (0 means no limit.) | `0` |
+| `warmStart` | `bool` | (Only available in `Train()`.)  If true, training adds `numTrees` trees to the random forest.  If `false`, an entirely new random forest will be created. | `false` |
+
+ * If OpenMP is enabled<!-- TODO: link! -->, one thread will be used to train
+   each of the `numTrees` trees in the random forest.  The computational effort
+   involved with training a random forest increases linearly with the number of
+   trees.
+ * The default `minLeafSize` is `1`, unlike `DecisionTree`.  This is because
+   random forests are less susceptible to overfitting due to their ensembled
+   nature.
+ * Note that the default `minLeafSize` of `1` will make large decision trees,
+   and so if a smaller-sized model is desired, this value should be increased
+   (at the potential cost of accuracy).
+ * `minGainSplit` can also be increased if a smaller-sized model is desired.
+
+***Note:*** different types can be used for `data` and `weights` (e.g.,
+`arma::fmat`, `arma::sp_mat`).  However, the element type of `data` and
+`weights` must match; for example, if `data` has type `arma::fmat`, then
+`weights` must have type `arma::frowvec`.
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with one
+of the following versions of the `Train()` member function:
+
+ * `rf.Train(data, labels, numClasses,          numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0, warmStart=false)`
+ * `rf.Train(data, labels, numClasses, weights, numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0, warmStart=false)`
+   - Train on numerical-only data (optionally with instance weights).
+   - Returns a `double` with the average gain of each tree in the random forest.
+     By default, this is the Gini gain, unless a different
+     [`FitnessFunction` template parameter](#fully-custom-behavior) is
+     specified.
+
+---
+
+ * `rf.Train(data, info, labels, numClasses,          numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0, warmStart=false)`
+ * `rf.Train(data, info, labels, numClasses, weights, numTrees=20, minLeafSize=1, minGainSplit=1e-7, maxDepth=0, warmStart=false)`
+   - Train on mixed categorical data (optionally with instance weights).
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+**Notes**:
+
+ * The `warmStart` option, which allows incremental training (i.e. additional
+   training on top of an existing model) is of type `bool` and defaults to
+   `false`.  This option is not available in the [constructors](#constructors).
+
+ * `Train()` returns a `double` with the average gain of each tree in the random
+   forest.  By default, this is the Gini gain, unless a different
+   [`FitnessFunction` template parameter](#fully-custom-behavior) is specified.
+
+### Classification
+
+Once a `RandomForest` is trained, the `Classify()` member function can be used
+to make class predictions for new data.
+
+ * `size_t predictedClass = rf.Classify(point)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class.
+
+---
+
+ * `rf.Classify(point, prediction, probabilitiesVec)`
+    - ***(Single-point)***
+    - Classify a single point and compute class probabilities.
+    - The predicted class is stored in `prediction`.
+    - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `rf.Classify(data, predictions)`
+    - ***(Multi-point)***
+    - Classify a set of points.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `rf.Classify(data, predictions, probabilities)`
+    - ***(Multi-point)***
+    - Classify a set of points and compute class probabilities for each point.
+    - The prediction for data point `i` can be accessed with `predictions[i]`.
+    - The probability of class `j` for data point `i` can be accessed with
+      `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into.  Will be set to length `numClasses`. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into.  Will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to number of classes, number of columns will be equal to `data.n_cols`). |
+
+***Note:*** different types can be used for `data` and `point` (e.g.
+`arma::fmat`, `arma::sp_mat`, `arma::sp_vec`, etc.).  However, the element type
+that is used should be the same type that was used for training.
+
+### Other Functionality
+
+ * A `RandomForest` can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `rf.NumTrees()` will return a `size_t` indicating the number of trees in the
+   random forest.
+
+ * `rf.Tree(i)` will return a [`DecisionTree` object](decision_tree.md)
+   representing the `i`th decision tree in the random forest.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/random_forest/random_forest.hpp) can be consulted.
+Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial use of
+`RandomForest`.
+
+---
+
+Train a random forest incrementally on random mixed categorical data and save it
+to disk:
+
+```c++
+// Load a categorical dataset.
+arma::mat dataset;
+mlpack::data::DatasetInfo info;
+// See https://datasets.mlpack.org/covertype.train.arff.
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+
+arma::Row<size_t> labels;
+// See https://datasets.mlpack.org/covertype.train.labels.csv.
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+
+// Create the random forest.
+mlpack::RandomForest rf;
+// Train 10 trees on the given dataset, with a minimum leaf size of 3.
+rf.Train(dataset, info, labels, 7 /* classes */, 10 /* trees */,
+         3 /* minimum leaf size */);
+
+// Now load categorical test data.
+arma::mat testDataset;
+// See https://datasets.mlpack.org/covertype.test.arff.
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+
+arma::Row<size_t> testLabels;
+// See https://datasets.mlpack.org/covertype.test.labels.csv.
+mlpack::data::Load("covertype.test.labels.csv", testLabels, true);
+
+// Compute test set accuracy.
+arma::Row<size_t> testPredictions;
+rf.Classify(testDataset, testPredictions);
+double accuracy = 100.0 * ((double) arma::accu(testPredictions == testLabels)) /
+    testLabels.n_elem;
+std::cout << "After training 10 trees, test set accuracy is " << accuracy
+    << "%." << std::endl;
+
+// Now train another 10 trees and compute the test accuracy.
+rf.Train(dataset, info, labels, 7 /* classes */, 10 /* trees */,
+         3 /* minimum leaf size */, 0.0 /* minimum split gain */,
+         0 /* maximum depth (unlimited) */, true /* incremental training */);
+
+rf.Classify(testDataset, testPredictions);
+accuracy = 100.0 * ((double) arma::accu(testPredictions == testLabels)) /
+    testLabels.n_elem;
+std::cout << "After training 20 trees, test set accuracy is " << accuracy
+    << "%." << std::endl;
+
+// Save the random forest to disk.
+mlpack::data::Save("rf.bin", "rf", rf);
+```
+
+---
+
+Load a random forest and print some information about it.
+
+```c++
+mlpack::RandomForest rf;
+// This call assumes a random forest called "rf" has already been saved to
+// `rf.bin` with `data::Save()`.
+mlpack::data::Load("rf.bin", "rf", rf, true);
+
+std::cout << "The random forest in 'rf.bin' contains " << rf.NumTrees()
+    << " trees." << std::endl;
+if (rf.NumTrees() > 0)
+{
+  std::cout << "The first tree's root node has " << rf.Tree(0).NumChildren()
+      << " children." << std::endl;
+}
+```
+
+---
+
+Train a random forest on categorical data, and compare its performance with the
+performance of each individual tree:
+
+```c++
+// Load a categorical dataset (training and test sets).
+arma::mat dataset, testDataset;
+mlpack::data::DatasetInfo info;
+arma::Row<size_t> labels, testLabels;
+
+// See the following files:
+//  * https://datasets.mlpack.org/covertype.train.arff
+//  * https://datasets.mlpack.org/covertype.train.labels.csv
+//  * https://datasets.mlpack.org/covertype.test.arff
+//  * https://datasets.mlpack.org/covertype.test.labels.csv
+mlpack::data::Load("covertype.train.arff", dataset, info, true);
+mlpack::data::Load("covertype.train.labels.csv", labels, true);
+mlpack::data::Load("covertype.test.arff", testDataset, info, true);
+mlpack::data::Load("covertype.test.labels.csv", testLabels, true);
+
+// Create the random forest.
+mlpack::RandomForest rf;
+// Train 20 trees on the given dataset, with a minimum leaf size of 5.
+rf.Train(dataset, info, labels, 7 /* classes */, 20 /* trees */,
+         5 /* minimum leaf size */);
+
+// Compute test set accuracy for each tree.
+arma::Row<size_t> testPredictions;
+for (size_t i = 0; i < rf.NumTrees(); ++i)
+{
+  rf.Tree(i).Classify(testDataset, testPredictions);
+  const double accuracy = 100.0 *
+      ((double) arma::accu(testPredictions == testLabels)) / testLabels.n_elem;
+  std::cout << "Tree " << i << " has test accuracy " << accuracy << "%."
+      << std::endl;
+}
+
+// Now compute accuracy using the whole forest.
+rf.Classify(testDataset, testPredictions);
+const double accuracy = 100.0 *
+    ((double) arma::accu(testPredictions == testLabels)) / testLabels.n_elem;
+std::cout << "The whole forest has test accuracy " << accuracy << "%."
+    << std::endl;
+```
+
+---
+
+Train an `ExtraTrees` model on random numeric data.
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::mat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor, using 10 trees in the forest.
+// Note that `ExtraTrees` has exactly the same API as `RandomForest`.
+mlpack::ExtraTrees<> rf(dataset, labels, 5, 10);
+
+// Create a single test point.
+arma::vec testPoint(10, arma::fill::randu);
+
+size_t prediction;
+arma::vec probabilities;
+rf.Classify(testPoint, prediction, probabilities);
+std::cout << "Test point predicted to be class " << prediction << "."
+    << std::endl;
+std::cout << "Probabilities of each class: " << probabilities.t();
+```
+
+---
+
+See also the following fully-working examples:
+
+ - [Rainfall prediction with `RandomForest`](https://github.com/mlpack/examples/blob/master/rainfall_prediction_with_random_forest/rainfall-prediction-with-random-forest-cpp.ipynb)
+ - [Forest cover type prediction with `RandomForest`](https://github.com/mlpack/examples/blob/master/rainfall_prediction_with_random_forest/rainfall-prediction-with-random-forest-cpp.ipynb)
+
+### Advanced Functionality: Template Parameters
+
+#### Using different element types.
+
+`RandomForest`'s constructors, `Train()`, and `Predict()` functions support any
+data type, so long as it supports the Armadillo matrix API.  So, for instance,
+learning can be done on single-precision floating-point data:
+
+```c++
+// 1000 random points in 10 dimensions.
+arma::fmat dataset(10, 1000, arma::fill::randu);
+// Random labels for each point, totaling 5 classes.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 4));
+
+// Train in the constructor.
+mlpack::RandomForest rf(dataset, labels, 5);
+
+// Create test data (500 points).
+arma::fmat testDataset(10, 500, arma::fill::randu);
+arma::Row<size_t> predictions;
+rf.Classify(testDataset, predictions);
+// Now `predictions` holds predictions for the test dataset.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 0) << " test points classified as class "
+    << "0." << std::endl;
+```
+
+---
+
+#### Fully custom behavior.
+
+mlpack provides a few variants of the random forest classifier, using the
+template parameters of the `RandomForest` class.  The following types can be
+used as drop-in replacements throughout this documentation page:
+
+ * `RandomForest`
+    - This is an implementation of Breiman's seminal random forest algorithm
+      ([website](https://www.stat.berkeley.edu/~breiman/RandomForests/cc_home.htm),
+      [paper pdf](https://www.stat.berkeley.edu/~breiman/randomforest2001.pdf)).
+    - The [`DecisionTree`](decision_tree.md) class is used for each individual
+      decision tree.
+    - When training each individual decision tree, bootstrapping is used to
+      compute the samples given to each tree for training.
+
+ * `ExtraTrees`
+    - This is an implementation of the Extremely Randomized Trees algorithm
+      ([paper pdf](https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=336a165c17c9c56160d332b9f4a2b403fccbdbfb)).
+    - When training an `ExtraTrees` model, each individual decision tree chooses
+      splits for numeric data randomly.
+    - Training an `ExtraTrees` model is generally much faster than
+      `RandomForest`, but the accuracy of the `ExtraTrees` model will be lower.
+    - To use `ExtraTrees`, simply replace `RandomForest` with `ExtraTrees` in
+      the documentation below.
+
+---
+
+Fully custom classes can also be used to control the behavior of the
+`RandomForest` class.  The full signature of the class is as follows:
+
+```
+RandomForest<FitnessFunction,
+             DimensionSelectionType,
+             NumericSplitType,
+             CategoricalSplitType,
+             UseBootstrap>
+```
+
+ * `FitnessFunction`: the measure of goodness to use when deciding on tree
+   splits
+ * `DimensionSelectionType`: the strategy used for proposing dimensions to
+   attempt to split on
+ * `NumericSplitType`: the strategy used for finding splits on numeric data
+   dimensions
+ * `CategoricalSplitType`: the strategy used for finding splits on categorical
+   data dimensions
+ * `UseBootstrap`: a boolean indicating whether or not to use a bootstrap sample
+   when training each tree in the forest
+
+Note that the first four of these template parameters are exactly the same as
+the template parameters for the
+[`DecisionTree`](decision_tree.md#fully-custom-behavior) class.
+
+Below, details are given for the requirements of each of these template types.
+
+---
+
+#### `FitnessFunction`
+
+ * Specifies the fitness function to use when learning a decision tree.
+ * The `GiniGain` _(default)_ and `InformationGain` classes are available for
+   drop-in usage.
+ * A custom class must implement three functions:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomFitnessFunction
+{
+  // Return the range (difference between maximum and minimum gain values).
+  double Range(const size_t numClasses);
+
+  // Compute the gain for the given vector of labels, where `labels[i]` has an
+  // associated instance weight `weights[i]`.
+  //
+  // `RowType` and `WeightVecType` will be vector types following the Armadillo
+  // API.  If `UseWeights` is `false`, then the `weights` vector should be
+  // ignored (e.g. the labels are not weighted).
+  template<bool UseWeights, typename RowType, typename WeightVecType>
+  double Evaluate(const RowType& labels,
+                  const size_t numClasses,
+                  const WeightVecType& weights);
+
+  // Compute the gain for the given counted set of labels, where `counts[i]`
+  // contains the number of points with label `i`.  There are `totalCount`
+  // labels total, and `counts` has length `numClasses`.
+  //
+  // `UseWeights` is ignored, and `CountType` will be an integral type (e.g.
+  // `size_t`).
+  template<bool UseWeights, typename CountType>
+  double EvaluatePtr(const CountType* counts,
+                     const size_t numClasses,
+                     const CountType totalCount);
+};
+```
+
+---
+
+#### `DimensionSelectionType`
+
+ * When splitting a tree in the forest, `DimensionSelectionType` proposes
+   possible dimensions to try splitting on.
+ * `MultipleRandomDimensionSelect` _(default)_ is available for drop-in usage
+   and proposes a different random subset of dimensions at each decision tree
+   node.
+    - By default each random subset is of size `sqrt(d)` where `d` is the number
+      of dimensions in the data.
+    - If constructed as `MultipleRandomDimensionSelect(n)` and passed to the
+      constructor of `RandomForest` or the `Train()` function, each random
+      subset will be of size `n`.
+ * Each `RandomForest` [constructor](#constructors) and each version of
+   the [`Train()`](#training) function optionally accept an instantiated
+   `DimensionSelectionType` object as the very last parameter (after
+   `maxDepth` in the constructor, or `warmStart` in `Train()`), in case some
+   internal state in the dimension selection mechanism is required.
+ * A custom class must implement three simple functions:
+
+```c++
+class CustomDimensionSelect
+{
+ public:
+  // Get the first dimension to try.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Begin();
+
+  // Get the next dimension to try.  Note that internal state can be used to
+  // track which candidate dimension is currently being looked at.
+  // This should return a value between `0` and `data.n_rows`.
+  size_t Next();
+
+  // Get a value indicating that all dimensions have been tried.
+  size_t End() const;
+
+  // The usage pattern of `DimensionSelectionType` by `DecisionTree` is as
+  // follows, assuming that `dim` is an instantiated `DimensionSelectionType`
+  // object:
+  //
+  // for (size_t dim = dim.Begin(); dim != dim.End(); dim = dim.Next())
+  // {
+  //   // ... try to split on dimension `dim` ...
+  // }
+};
+```
+
+---
+
+#### `NumericSplitType`
+
+ * Specifies the strategy to be used during training when splitting a numeric
+   feature.
+ * The `BestBinaryNumericSplit` _(default)_ class is available for drop-in
+   usage and finds the best binary (two-way) split among all possible binary
+   splits.
+ * The `RandomBinaryNumericSplit` class is available for drop-in usage and
+   will select a split randomly between the minimum and maximum values of a
+   dimension.  It is very efficient but does not yield splits that maximize
+   the gain.  (Used by the `ExtraTrees` [variant](#fully-custom-behavior).)
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal
+   structure `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+template<typename FitnessFunction>
+class CustomNumericSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one dimension
+  // of the input data).
+  template<bool UseWeights, typename VecType, typename WeightVecType>
+  double SplitIfBetter(const double bestGain,
+                       const VecType& data,
+                       const arma::Row<size_t>& labels,
+                       const size_t numClasses,
+                       const WeightVecType& weights,
+                       const size_t minLeafSize,
+                       const double minGainSplit,
+                       arma::vec& splitInfo,
+                       AuxiliarySplitInfo& aux);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with value `point`, and split information `splitInfo` and
+  // `aux`, return the index of the child that corresponds to the point.  So,
+  // e.g., if the split type was a binary split on the value `splitInfo`, you
+  // might return `0` if `point < splitInfo`, and `1` otherwise.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+---
+
+#### `CategoricalSplitType`
+
+ * Specifies the strategy to be used during training when splitting a
+    categorical feature.
+ * The `AllCategoricalSplit` _(default)_ is available for drop-in usage and
+   splits all categories into their own node.
+ * A custom class must take a [`FitnessFunction`](#fitnessfunction) as a
+   template parameter, implement three functions, and have an internal structure
+   `AuxiliarySplitInfo` that is used at classification time:
+
+```c++
+template<typename FitnessFunction>
+class CustomCategoricalSplit
+{
+ public:
+  // If a split with better resulting gain than `bestGain` is found, then
+  // information about the new, better split should be stored in `splitInfo` and
+  // `aux`.  Specifically, a split is better than `bestGain` if the sum of the
+  // gains that the children will have (call this `sumChildrenGains`) is
+  // sufficiently better than the gain of the unsplit node (call this
+  // `unsplitGain`):
+  //
+  //    split if `sumChildrenGains - unsplitGain > bestGain`, and
+  //             `sumChildrenGains - unsplitGain > minGainSplit`, and
+  //             each child will have at least `minLeafSize` points
+  //
+  // The new best split value should be returned (or anything greater than or
+  // equal to `bestGain` if no better split is found).
+  //
+  // If a new best split is found, then `splitInfo` and `aux` should be
+  // populated with the information that will be needed for
+  // `CalculateDirection()` to successfully choose the child for a given point.
+  // `splitInfo` should be set to a vector of length 1.  The format of `aux` is
+  // arbitrary and is detailed more below.
+  //
+  // If `UseWeights` is false, the vector `weights` should be ignored.
+  // Otherwise, they are instance weighs for each value in `data` (one
+  // categorical dimension of the input data, which takes values between `0` and
+  // `numCategories - 1`).
+  template<bool UseWeights, typename VecType, typename LabelsType,
+           typename WeightVecType>
+  static double SplitIfBetter(
+      const double bestGain,
+      const VecType& data,
+      const size_t numCategories,
+      const LabelsType& labels,
+      const size_t numClasses,
+      const WeightVecType& weights,
+      const size_t minLeafSize,
+      const double minGainSplit,
+      arma::vec& splitInfo,
+      AuxiliarySplitInfo& aux);
+
+  // Return the number of children for a given split (stored as the single
+  // element from `splitInfo` and auxiliary data `aux` in `SplitIfBetter()`).
+  size_t NumChildren(const double& splitInfo,
+                     const AuxiliarySplitInfo& aux);
+
+  // Given a point with (categorical) value `point`, and split information
+  // `splitInfo` and `aux`, return the index of the child that corresponds to
+  // the point.  So, e.g., for `AllCategoricalSplit`, which splits a categorical
+  // dimension into one child for each category, this simply returns `point`.
+  template<typename ElemType>
+  static size_t CalculateDirection(
+      const ElemType& point,
+      const double& splitInfo,
+      const AuxiliarySplitInfo& /* aux */);
+
+  // This class can hold any extra data that is necessary to encode a split.  It
+  // should only be non-empty if a single `double` value cannot be used to hold
+  // the information corresponding to a split.
+  class AuxiliarySplitInfo { };
+};
+```
+
+---
+
+#### `UseBootstrap`
+
+ * A `bool` value that indicates whether or not a bootstrap sample of the
+   dataset should be used for the training of each individual decision tree in
+   the random forest.
+ * If `true` _(default)_, a different bootstrap sample of the same size as the
+   dataset will be used to train each decision tree.
+ * If `false` _(default for the `ExtraTrees` [variant](#fully-custom-behavior))_, the full
+   dataset will be used to train each decision tree.
diff -pruN 4.3.0-2/doc/user/methods/softmax_regression.md 4.4.0-1/doc/user/methods/softmax_regression.md
--- 4.3.0-2/doc/user/methods/softmax_regression.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/softmax_regression.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,408 @@
+## `SoftmaxRegression`
+
+The `SoftmaxRegression` class implements an L2-regularized multi-class softmax
+regression classifier for numerical data.  This is a multi-class extension of
+the logistic regression classifier.  By default, L-BFGS is used to learn the
+model.  The `SoftmaxRegression` class offers easy configurability, and arbitrary
+optimizers can be used to learn the model.
+
+Softmax regression is useful for multi-class classification (i.e. classes are
+`0`, `1`, `2`).  For two-class situations, see also the
+[`LogisticRegression`](logistic_regression.md) class.
+
+#### Simple usage example:
+
+```c++
+// Train a softmax regression model on random data and predict labels:
+
+// All data and labels are uniform random; 5 dimensional data, 4 classes.
+// Replace with a data::Load() call or similar for a real application.
+arma::mat dataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+arma::mat testDataset(5, 500, arma::fill::randu); // 500 test points.
+
+mlpack::SoftmaxRegression sr;          // Step 1: create model.
+sr.Train(dataset, labels, 4);          // Step 2: train model.
+arma::Row<size_t> predictions;
+sr.Classify(testDataset, predictions); // Step 3: classify points.
+
+// Print some information about the test predictions.
+std::cout << arma::accu(predictions == 2) << " test points classified as class "
+    << "2." << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `SoftmaxRegression` objects.
+ * [`Train()`](#training): train model.
+ * [`Classify()`](#classification): classify with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-different-element-types) for
+   using different element types for a model.
+
+#### See also:
+
+ * [`LogisticRegression`](logistic_regression.md)
+ * [mlpack classifiers](../../index.md#classification-algorithms)
+ * [UFLDL Softmax Regression Tutorial](http://deeplearning.stanford.edu/tutorial/supervised/SoftmaxRegression/)
+
+### Constructors
+
+ * `sr = SoftmaxRegression()`
+   - Initialize the model without training.
+   - You will need to call [`Train()`](#training) later to train the model
+     before calling [`Classify()`](#classification).
+
+---
+
+ * `sr = SoftmaxRegression(data, labels, numClasses, lambda=0.0001, fitIntercept=true)`
+ * `sr = SoftmaxRegression(data, labels, numClasses, lambda=0.0001, fitIntercept=true, [callbacks...])`
+   - Train model, optionally specifying hyperparameters and callbacks for the
+     default ensmallen optimizer.
+
+---
+
+ * `sr = SoftmaxRegression(data, labels, numClasses, optimizer, lambda=0.0001, fitIntercept=true)`
+ * `sr = SoftmaxRegression(data, labels, numClasses, optimizer, lambda=0.0001, fitIntercept=true, [callbacks...])`
+    - Train model with a custom ensmallen optimizer, optionally specifying
+      hyperparameters and callbacks for the custom optimizer.
+
+---
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `labels` | [`arma::Row<size_t>`](../matrices.md) | Training labels, [between `0` and `numClasses - 1`](../load_save.md#normalizing-labels) (inclusive).  Should have length `data.n_cols`.  | _(N/A)_ |
+| `numClasses` | `size_t` | Number of classes in the dataset. | _(N/A)_ |
+| `lambda` | `double` | L2 regularization penalty parameter.  Must be nonnegative. | `0.0001` |
+| `fitIntercept` | `bool` | If true, an intercept term is fit to the model. | `true` |
+| `optimizer` | [any ensmallen optimizer](https://www.ensmallen.org) | Instantiated ensmallen optimizer for [differentiable functions](https://www.ensmallen.org/docs.html#differentiable-functions) or [differentiable separable functions](https://www.ensmallen.org/docs.html#differentiable-separable-functions). | _(N/A)_ |
+| `callbacks...` | [any set of ensmallen callbacks](https://www.ensmallen.org/docs.html#callback-documentation) | Optional callbacks for the ensmallen optimizer, such as e.g. `ens::ProgressBar()`, `ens::Report()`, or others. | _(N/A)_ |
+
+As an alternative to passing `lambda`, it can be set with a
+standalone method:
+
+ * `sr.Lambda() = l;` will set the value of the L2 regularization penalty
+   parameter to `l`.
+
+It is not possible to set `fitIntercept` except in the constructor or the call
+to `Train()`.
+
+***Note***: Setting `lambda` too small may cause the model to overfit; however,
+setting it too large may cause the model to underfit.  [Automatic hyperparameter
+tuning](../hpt.md) can be used to find a good value of `lambda` instead of a
+manual setting.
+
+<!-- TODO: fix hyperparameter tuner link with real documentation -->
+
+### Training
+
+If training is not done as part of the constructor call, it can be done with the
+`Train()` function:
+
+ * `sr.Train(data, labels, numClasses, lambda=0.0001, fitIntercept=true)`
+ * `sr.Train(data, labels, numClasses, lambda=0.0001, fitIntercept=true, [callbacks...])`
+   - Train model, optionally specifying hyperparameters and callbacks for the
+     default ensmallen optimizer.
+
+---
+
+ * `sr.Train(data, labels, numClasses, optimizer, lambda=0.0001, fitIntercept=true)`
+ * `sr.Train(data, labels, numClasses, optimizer, lambda=0.0001, fitIntercept=true, optimizer=ens::L_BFGS(), [callbacks...])`
+   - Train model with a custom ensmallen optimizer, optionally specifying
+    hyperparameters and callbacks for the optimizer.
+
+---
+
+Types of each argument are the same as in the table for constructors
+[above](#constructor-parameters).
+
+***Notes***:
+
+ * Training is incremental.  Successive calls to `Train()` will not reinitialize
+   the model, unless the given data has different dimensionality.  To
+   reinitialize the model, call `Reset()` (see
+   [Other Functionality](#other-functionality)).
+
+ * To set the initial point of the optimization, call `Parameters()`; see
+   [Other Functionality](#other-functionality).
+
+ * `Train()` returns a `double` with the final softmax regression loss value
+   (including L2 penalty term) of the trained model.
+
+### Classification
+
+Once a `SoftmaxRegression` model is trained, the `Classify()` member function
+can be used to make class predictions for new data.
+
+ * `size_t predictedClass = sr.Classify(point)`
+   - ***(Single-point)***
+   - Classify a single point, returning the predicted class (`0` through
+     `numClasses - 1`, inclusive).
+
+---
+
+ * `sr.Classify(point, prediction, probabilitiesVec)`
+   - ***(Single-point)***
+   - Classify a single point and compute class probabilities.
+   - The predicted class is stored in `prediction`.
+   - The probability of class `i` can be accessed with `probabilitiesVec[i]`.
+
+---
+
+ * `sr.Classify(data, predictions)`
+   - ***(Multi-point)***
+   - Classify a set of points.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+
+---
+
+ * `sr.Classify(data, predictions, probabilities)`
+   - ***(Multi-point)***
+   - Classify a set of points and compute class probabilities.
+   - The prediction for data point `i` can be accessed with `predictions[i]`.
+   - The probability of class `j` for data point `i` can be accessed with
+     `probabilities(j, i)`.
+
+---
+
+#### Classification Parameters:
+
+| **usage** | **name** | **type** | **description** |
+|-----------|----------|----------|-----------------|
+| _single-point_ | `point` | [`arma::vec`](../matrices.md) | Single point for classification. |
+| _single-point_ | `prediction` | `size_t&` | `size_t` to store class prediction into. |
+| _single-point_ | `probabilitiesVec` | [`arma::vec&`](../matrices.md) | `arma::vec&` to store class probabilities into; will have length 2. |
+||||
+| _multi-point_ | `data` | [`arma::mat`](../matrices.md) | Set of [column-major](../matrices.md#representing-data-in-mlpack) points for classification. |
+| _multi-point_ | `predictions` | [`arma::Row<size_t>&`](../matrices.md) | Vector of `size_t`s to store class prediction into; will be set to length `data.n_cols`. |
+| _multi-point_ | `probabilities` | [`arma::mat&`](../matrices.md) | Matrix to store class probabilities into (number of rows will be equal to 2; number of columns will be equal to `data.n_cols`). |
+
+### Other Functionality
+
+ * A `SoftmaxRegression` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `sr.Parameters()` will return an `arma::mat` filled with the weights of the
+   model.  The matrix will have rows equal to `numClasses` and columns equal to
+   the dimensionality of the model (plus one if `fitIntercept` is `true`).  Each
+   row of the matrix can be considered its own two-class one-vs.-all logistic
+   regression model.
+
+ * `sr.NumClasses()` will return the number of classes the model was trained on.
+
+ * `sr.ComputeAccuracy(data, labels)` will return the accuracy of the model on
+   the given `data` with the given `labels`.  The returned accuracy is between 0
+   and 100.
+
+ * `sr.Reset()` will reset the weights of the model to small random values.
+
+For complete functionality, the [source
+code](/src/mlpack/methods/softmax_regression/softmax_regression.hpp) can be
+consulted.  Each method is fully documented.
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `SoftmaxRegression` class.
+
+---
+
+Train a softmax regression model using a custom SGD-like optimizer with
+callbacks.
+
+```c++
+// See https://datasets.mlpack.org/mnist.train.csv.
+arma::mat dataset;
+mlpack::data::Load("mnist.train.csv", dataset, true);
+// See https://datasets.mlpack.org/mnist.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("mnist.train.labels.csv", labels, true);
+
+mlpack::SoftmaxRegression sr;
+
+// Create AdaGrad optimizer with custom step size and batch size.
+ens::AdaGrad optimizer(0.0005 /* step size */, 16 /* batch size */);
+optimizer.MaxIterations() = 10 * dataset.n_cols; // Allow 10 epochs.
+
+// Print a progress bar and an optimization report when training is finished.
+sr.Train(dataset, labels, 10 /* numClasses */, optimizer,
+    0.01 /* lambda */, true /* fit intercept */, ens::ProgressBar(),
+    ens::Report());
+
+// Now predict on test labels and compute accuracy.
+
+// See https://datasets.mlpack.org/mnist.test.csv.
+arma::mat testDataset;
+mlpack::data::Load("mnist.test.csv", testDataset, true);
+// See https://datasets.mlpack.org/mnist.test.labels.csv.
+arma::Row<size_t> testLabels;
+mlpack::data::Load("mnist.test.labels.csv", testLabels, true);
+
+std::cout << std::endl;
+std::cout << "Accuracy on training set: "
+    << sr.ComputeAccuracy(dataset, labels) << "\%." << std::endl;
+std::cout << "Accuracy on test set:     "
+    << sr.ComputeAccuracy(testDataset, testLabels) << "\%." << std::endl;
+```
+
+---
+
+Train a softmax regression model with AdaGrad and save the model every epoch
+using a [custom ensmallen
+callback](https://www.ensmallen.org/docs.html#custom-callbacks):
+
+```c++
+// This callback saves the model into "model-<epoch>.bin" after every epoch.
+class ModelCheckpoint
+{
+ public:
+  ModelCheckpoint(mlpack::SoftmaxRegression<>& model) : model(model) { }
+
+  template<typename OptimizerType, typename FunctionType, typename MatType>
+  bool EndEpoch(OptimizerType& /* optimizer */,
+                FunctionType& /* function */,
+                const MatType& /* coordinates */,
+                const size_t epoch,
+                const double /* objective */)
+  {
+    const std::string filename = "model-" + std::to_string(epoch) + ".bin";
+    mlpack::data::Save(filename, "sr_model", model, true);
+    return false; // Do not terminate the optimization.
+  }
+
+ private:
+  mlpack::SoftmaxRegression<>& model;
+};
+```
+
+With that callback available, the code to train the model is below:
+
+```c++
+// See https://datasets.mlpack.org/mnist.train.csv.
+arma::mat dataset;
+mlpack::data::Load("mnist.train.csv", dataset, true);
+// See https://datasets.mlpack.org/mnist.train.labels.csv.
+arma::Row<size_t> labels;
+mlpack::data::Load("mnist.train.labels.csv", labels, true);
+
+mlpack::SoftmaxRegression sr;
+
+// Create AdaGrad optimizer with small step size and batch size of 32.
+ens::AdaGrad adaGrad(0.0005, 32);
+adaGrad.MaxIterations() = 10 * dataset.n_cols; // 10 epochs maximum.
+
+// Use the custom callback and an L2 penalty parameter of 0.01.
+sr.Train(dataset, labels, 10 /* numClasses */, adaGrad, 0.01, true,
+    ModelCheckpoint(sr), ens::ProgressBar());
+
+// Now files like model-1.bin, model-2.bin, etc. should be saved on disk.
+```
+
+---
+
+Load an existing softmax regression model and print some information about it.
+
+```c++
+mlpack::SoftmaxRegression sr;
+// This assumes that a model called "sr_model" has been saved to the file
+// "model-1.bin" (as in the previous example).
+mlpack::data::Load("model-1.bin", "sr_model", sr, true);
+
+// Print the dimensionality of the model and some other statistics.
+const size_t dimensionality = (sr.FitIntercept()) ?
+    (sr.Parameters().n_cols - 1) : (sr.Parameters().n_cols);
+std::cout << "The dimensionality of the model in model-1.bin is "
+    << dimensionality << "." << std::endl;
+std::cout << "The bias parameters for the model, for each class, are: "
+    << std::endl;
+std::cout << sr.Parameters().col(0).t();
+
+arma::vec point(dimensionality, arma::fill::randu);
+std::cout << "The predicted class for a random point is " << sr.Classify(point)
+    << "." << std::endl;
+```
+
+---
+
+Perform incremental training on multiple datasets with multiple calls to
+`Train()`.
+
+```c++
+// Generate two random datasets with four classes.
+arma::mat firstDataset(5, 1000, arma::fill::randu); // 1000 points.
+arma::Row<size_t> firstLabels =
+    arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+
+arma::mat secondDataset(5, 1500, arma::fill::randu); // 1500 points.
+arma::Row<size_t> secondLabels =
+    arma::randi<arma::Row<size_t>>(1500, arma::distr_param(0, 3));
+
+// Train a model on the first dataset with an L2 regularization penalty
+// parameter of 0.01, not fitting an intercept.
+mlpack::SoftmaxRegression sr(firstDataset, firstLabels, 4, 0.01, false);
+
+// Now compute the accuracy on the second dataset and print it.
+std::cout << "Accuracy on second dataset: "
+    << sr.ComputeAccuracy(secondDataset, secondLabels) << "\%." << std::endl;
+
+// Train for a second round on the second dataset.
+sr.Train(secondDataset, secondLabels, 4);
+
+// Now compute the accuracy on the second dataset again and print it.
+// (Note that it may not be all that much better because this is random data!)
+std::cout << "Accuracy on second dataset after second training: "
+    << sr.ComputeAccuracy(secondDataset, secondLabels) << "\%." << std::endl;
+```
+
+---
+
+### Advanced Functionality: Different Element Types
+
+The `SoftmaxRegression` class has one template parameter that can be used to
+control the element type of the model.  The full signature of the class is:
+
+```
+SoftmaxRegression<MatType>
+```
+
+`MatType` specifies the type of matrix used for training data and internal
+representation of model parameters.  Any matrix type that implements the
+Armadillo API can be used.  The example below trains a softmax regression model
+on sparse 32-bit floating point data.
+
+```c++
+// Create random, sparse 100-dimensional data, with 3 classes.
+arma::sp_fmat dataset;
+dataset.sprandu(100, 5000, 0.3);
+arma::Row<size_t> labels =
+    arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 2));
+
+// Train with L2 regularization penalty parameter of 0.1.
+mlpack::SoftmaxRegression<arma::sp_fmat> sr(dataset, labels, 3, 0.1);
+
+// Now classify a test point.
+arma::sp_fvec point;
+point.sprandu(100, 1, 0.3);
+
+size_t prediction;
+arma::fvec probabilitiesVec;
+sr.Classify(point, prediction, probabilitiesVec);
+
+std::cout << "Prediction for random test point: " << prediction << "."
+    << std::endl;
+std::cout << "Class probabilities for random test point: "
+    << probabilitiesVec.t();
+```
+
+***Note***: if `MatType` is a sparse object (e.g. `sp_fmat`), the internal
+parameter representation will be a *dense* matrix containing elements of the
+same type (e.g. `fmat`).  This is because L2-regularized softmax regression,
+even when training on sparse data, does not necessarily produce sparse models.
diff -pruN 4.3.0-2/doc/user/methods/sparse_coding.md 4.4.0-1/doc/user/methods/sparse_coding.md
--- 4.3.0-2/doc/user/methods/sparse_coding.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/doc/user/methods/sparse_coding.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,399 @@
+## `SparseCoding`
+
+The `SparseCoding` class implements sparse coding with dictionary learning using
+L1 regularization or elastic net (L1+L2) regularization.  Sparse coding is a
+form of representation learning, and can be used to represent each point in a
+dataset as a sparse combination of *atoms* in the dictionary.
+
+#### Simple usage example:
+
+```c++
+// Create a random dataset with 100 points in 40 dimensions, and then a random
+// test dataset with 50 points.
+arma::mat data(40, 100, arma::fill::randu);
+arma::mat testData(40, 50, arma::fill::randu);
+
+// Perform sparse coding with 20 atoms and an L1 penalty of 0.5.
+mlpack::SparseCoding sc(20, 0.5);  // Step 1: create object.
+double objective = sc.Train(data); // Step 2: learn dictionary.
+arma::mat codes;
+sc.Encode(testData, codes);        // Step 3: encode new data.
+
+// Print some information about the test encoding.
+std::cout << "Average density of encoded test data: "
+    << 100.0 * arma::mean(arma::sum(codes != 0)) / codes.n_rows << "\%."
+    << std::endl;
+```
+<p style="text-align: center; font-size: 85%"><a href="#simple-examples">More examples...</a></p>
+
+#### Quick links:
+
+ * [Constructors](#constructors): create `SparseCoding` objects.
+ * [`Train()`](#training): train model (learn dictionary).
+ * [`Encode()`](#encoding): encode points with a trained model.
+ * [Other functionality](#other-functionality) for loading, saving, and
+   inspecting.
+ * [Examples](#simple-examples) of simple usage and links to detailed example
+   projects.
+ * [Template parameters](#advanced-functionality-template-parameters) for
+   advanced functionality: different element types and dictionary initialization
+   strategies.
+
+#### See also:
+
+ * [`LocalCoordinateCoding`](local_coordinate_coding.md)
+ * [`LARS`](lars.md) (used internally by `SparseCoding`)
+ * [mlpack transformations](../../index.md#transformations)
+ * [Sparse dictionary learning on Wikipedia](https://en.wikipedia.org/wiki/Sparse_dictionary_learning)
+ * [Efficient sparse coding algorithms (pdf)](https://proceedings.neurips.cc/paper/2006/file/2d71b2ae158c7c5912cc0bbde2bb9d95-Paper.pdf)
+
+### Constructors
+
+ * `sc = SparseCoding()`
+ * `sc = SparseCoding(atoms=0, lambda1=0.0, lambda2=0.0, maxIter=0, objTol=0.01, newtonTol=1e-6)`
+   - Create a `SparseCoding` object without learning a dictionary on data.
+   - If `atoms` is set to `0` (the default), it will need to be set to a value
+     greater than `0` before `Train()` is called (`sc.Atoms() = atoms` can be
+     used for this).
+
+ * `sc = SparseCoding(data, atoms, lambda1=0.0, lambda2=0.0, maxIter=0, objTol=0.01, newtonTol=1e-6)`
+   - Create a `SparseCoding` object and train the dictionary on the given
+     `data`.
+   - The dictionary will contain `atoms` elements.
+
+ * `sc = SparseCoding(data, atoms, lambda1, lambda2, maxIter, objTol, newtonTol, initializer)`
+   - *Advanced constructor*: create a `SparseCoding` object that will use a
+     custom dictionary initializer and train on the given `data`.
+   - The dictionary will contain `atoms` elements.
+   - `initializer` will be used to initialize the dictionary; see [Advanced
+     Functionality: Different Dictionary Initialization
+     Strategies](#dictionaryinitializer-different-dictionary-initialization-strategies)
+     for details.
+
+#### Constructor Parameters:
+
+| **name** | **type** | **description** | **default** |
+|----------|----------|-----------------|-------------|
+| `data` | [`arma::mat`](../matrices.md) | [Column-major](../matrices.md#representing-data-in-mlpack) training matrix. | _(N/A)_ |
+| `atoms` | `size_t` | Number of atoms in dictionary. | _(N/A)_ |
+| `lambda1` | `double` | L1 regularization penalty.  Used in both `Train()` and `Encode()` steps. | `0.0` |
+| `lambda2` | `double` | L2 regularization penalty (for elastic net regularization).  Used in both `Train()` and `Encode()` steps. | `0.0` |
+| `maxIter` | `size_t` | Maximum number of iterations for dictionary learning.  `0` means no limit. | `0` |
+| `objTol` | `double` | Objective function tolerance for terminating dictionary learning. | `0.01` |
+| `newtonTol` | `double` | Tolerance for the Newton's method dictionary optimization step. | `1e-6` |
+
+As an alternative to passing `atoms`, `lambda1`, `lambda2`, `maxIter`, `objTol`,
+or `newtonTol`, these can be set with a standalone method.  The following
+functions can be used before calling `Train()`:
+
+ * `sc.Atoms() = a;` will set the number of atoms to use in the dictionary to
+   `a`.  Changing this after calling `Train()` will not make a difference to the
+   dictionary size.
+
+ * `sc.Lambda1() = l1;` will set the L1 regularization penalty to `l1`.  This
+   can be set after `Train()` to force sparser encodings when `Encode()` is
+   called.
+
+ * `sc.Lambda2() = l2;` will set the L2 regularization penalty to `l2`.  This
+   can be set after `Train()` to increase the regularization when `Encode()` is
+   called.
+
+ * `sc.MaxIterations() = m;` will set the maximum number of iterations for
+   dictionary learning to `m`.  `0` means that the algorithm will run until
+   convergence.
+
+ * `sc.ObjTolerance() = ot;` will set the objective tolerance for convergence of
+   the dictionary learning algorithm to `ot`.
+
+ * `sc.NewtonTolerance() = nt;` will set the tolerance for the Newton's method
+   dictionary optimization step to `nt`.
+
+***Caveats***:
+
+ * Larger settings of `atoms` (i.e. larger dictionary sizes) will be able to
+   more accurately represent the data, but may take longer to learn.
+
+ * Larger values of `lambda1` will cause the model to use sparser codings for
+   data when `Train()` and `Encode()` are called, but codings that are too
+   sparse may be inaccurate representations of the original points.
+
+ * Training is not incremental; a second call to `Train()` will reinitialize the
+   dictionary and restart the learning process.
+
+### Training
+
+If training the dictionary is not done as part of the constructor call, it can
+be done with one of the following versions of the `Train()` member function:
+
+ * `sc.Train(data)`
+ * `sc.Train(data, initializer)`
+   - Train the sparse coding dictionary on the given `data`.
+   - Optionally, use the given `initializer` to initialize the dictionary (see
+     [`DictionaryInitializer`](#dictionaryinitializer-different-dictionary-initialization-strategies)
+     for more details).
+
+### Encoding
+
+Once a `SparseCoding` model has a trained dictionary, the `Encode()` member
+function can be used to encode new data points.
+
+ * `sc.Encode(data, codes)`
+   - Encode `data` (a
+     [column-major data matrix](../matrices.md#representing-data-in-mlpack))
+     as a set of sparse codes of the dictionary, storing the result in `codes`.
+   - Both `data` and `codes` should be the same matrix type (e.g. `arma::mat`);
+     see [Different Element Types](#mattype-different-element-types) for more
+     details.
+   - `codes` will be set to have `atoms` rows and `data.n_cols` columns.
+   - Column `i` of `codes` corresponds to the sparse coding of the `i`'th column
+     of `data`.  Each row represents the weight associated with each atom in
+     the dictionary.
+
+After encoding, the original data can be recovered (approximately) as
+`sc.Dictionary() * data`.
+
+### Other Functionality
+
+ * A `SparseCoding` model can be serialized with
+   [`data::Save()` and `data::Load()`](../load_save.md#mlpack-objects).
+
+ * `sc.Dictionary()` will return an `arma::mat&` containing the dictionary
+   matrix.  The matrix has `data.n_rows` rows and `atoms` columns; each column
+   corresponds to an atom in the dictionary.
+
+ * `double obj = sc.Objective(data, codes)` computes the sparse coding objective
+   function on the given `data` and encodings `codes`.  This can be used after
+   `Encode()` to test the quality of the encodings (a smaller objective is
+   better).
+
+### Simple Examples
+
+See also the [simple usage example](#simple-usage-example) for a trivial usage
+of the `SparseCoding` class.
+
+---
+
+Train a sparse coding model on the cloud dataset and print the reconstruction
+error.
+
+```c++
+// See https://datasets.mlpack.org/cloud.csv.
+arma::mat dataset;
+mlpack::data::Load("cloud.csv", dataset, true);
+
+mlpack::SparseCoding sc;
+sc.Atoms() = 50;
+sc.Lambda1() = 0.1;
+sc.Lambda2() = 0.001;
+sc.MaxIterations() = 25;
+sc.Train(dataset);
+
+// Encode the training dataset.
+arma::mat codes;
+sc.Encode(dataset, codes);
+
+std::cout << "Input matrix size: " << dataset.n_rows << " x " << dataset.n_cols
+    << "." << std::endl;
+std::cout << "Codes matrix size: " << codes.n_rows << " x " << codes.n_cols
+    << "." << std::endl;
+
+// Reconstruct the original matrix.
+arma::mat recon = sc.Dictionary() * codes;
+double error = std::sqrt(arma::norm(dataset - recon, "fro") / dataset.n_elem);
+std::cout << "RMSE of reconstructed matrix: " << error << "." << std::endl;
+```
+
+---
+
+Train a sparse coding model on the iris dataset and save the model to disk.
+
+```c++
+// See https://datasets.mlpack.org/iris.train.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.train.csv", dataset, true);
+
+// Train the model in the constructor.
+mlpack::SparseCoding sc(dataset, 10 /* atoms */, 0.1 /* L1 penalty */);
+
+// Save the model to disk.
+mlpack::data::Save("sc.bin", "sc", sc);
+```
+
+---
+
+Load a sparse coding model from disk and encode some new points from the
+iris dataset.
+
+```c++
+// Load model from disk.
+mlpack::SparseCoding sc;
+mlpack::data::Load("sc.bin", "sc", sc);
+
+// See https://datasets.mlpack.org/iris.test.csv.
+arma::mat dataset;
+mlpack::data::Load("iris.test.csv", dataset, true);
+
+// Encode the test points.
+arma::mat codes;
+sc.Encode(dataset, codes);
+
+// Compute the sparse coding objective on the test points.
+const double obj = sc.Objective(dataset, codes);
+std::cout << "Sparse coding objective on test set: " << obj << "." << std::endl;
+```
+
+---
+
+Train a sparse coding model on the satellite dataset, trying several different
+dictionary sizes and checking the objective value on a held-out test dataset.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat trainData;
+mlpack::data::Load("satellite.train.csv", trainData, true);
+// See https://datasets.mlpack.org/satellite.test.csv.
+arma::mat testData;
+mlpack::data::Load("satellite.test.csv", testData, true);
+
+for (size_t atoms = 20; atoms < 100; atoms += 10)
+{
+  mlpack::SparseCoding sc(atoms);
+  sc.Lambda1() = 0.1;
+  sc.MaxIterations() = 20; // Keep iterations low so this runs relatively fast.
+
+  const double trainObj = sc.Train(trainData);
+
+  // Compute the objective on the test set.
+  arma::mat codes;
+  sc.Encode(testData, codes);
+  const double testObj = sc.Objective(testData, codes);
+
+  std::cout << "Atoms: " << std::setfill(' ') << std::setw(3) << atoms << "; ";
+  std::cout << "training set objective: " << std::setw(6) << trainObj << "; ";
+  std::cout << "test set objective: " << std::setw(6) << testObj << "."
+      << std::endl;
+}
+```
+
+### Advanced Functionality: Template Parameters
+
+The `SparseCoding` class has one class template parameter that can be used for
+custom behavior.  The full signature of the class is:
+
+```
+SparseCoding<MatType>
+```
+
+In addition, the [constructors](#constructors) and
+[`Train()` functions](#training) have a template parameter
+`DictionaryInitializer` that can be used for custom behavior.
+
+ * `MatType`: the type of the matrix to use (e.g. `arma::mat`, `arma::fmat`,
+   etc.).  The given `MatType` must support the Armadillo API and hold a
+   floating-point element type (e.g. `float`, `double`, etc.).
+
+ * `DictionaryInitializer`: the strategy used to initialize the dictionary.  By
+   default, `DataDependentRandomInitializer` is used.
+
+#### `MatType`: Different Element Types
+
+`MatType` specifies the type of matrix used for training data and internal
+representation of the dictionary.  Any matrix type that implements the Armadillo
+API can be used.  The example below trains a sparse coding model on 32-bit
+floating point data.
+
+```c++
+// See https://datasets.mlpack.org/cloud.csv.
+arma::fmat dataset;
+mlpack::data::Load("cloud.csv", dataset, true);
+
+mlpack::SparseCoding<arma::fmat> sc;
+sc.Atoms() = 30;
+sc.Lambda1() = 0.15;
+sc.Lambda2() = 0.001;
+sc.MaxIterations() = 100;
+// Note: a looser tolerance is often required when using floats instead of
+// doubles.
+sc.ObjTolerance() = 0.01;
+sc.Train(dataset);
+
+// Encode the training dataset.
+arma::fmat codes;
+sc.Encode(dataset, codes);
+
+std::cout << "Input matrix size: " << dataset.n_rows << " x " << dataset.n_cols
+    << "." << std::endl;
+std::cout << "Codes matrix size: " << codes.n_rows << " x " << codes.n_cols
+    << "." << std::endl;
+
+// Reconstruct the original matrix.
+arma::fmat recon = sc.Dictionary() * codes;
+double error = std::sqrt(arma::norm(dataset - recon, "fro") / dataset.n_elem);
+std::cout << "RMSE of reconstructed matrix: " << error << "." << std::endl;
+```
+
+#### `DictionaryInitializer`: Different Dictionary Initialization Strategies
+
+The `DictionaryInitializer` template class specifies the strategy to be used to
+initialize the dictionary when `Train()` is called.
+
+ * The `DataDependentRandomInitalizer` class (the default) uses the average of
+   three random points in the dataset to initialize each atom in the dictionary.
+
+ * The `NothingInitializer` class does not modify the dictionary matrix in any
+   way, and could be used either to set a specific dictionary before training
+   with `sc.Dictionary()`, or to allow incremental training that does not modify
+   the existing dictionary when `Train()` is called a second time.
+
+ * The `RandomInitializer` class initializes the dictionary by sampling norm-1
+   atoms from a normal distribution.
+
+***Note:*** none of the classes above have any members, and as such it is not
+necessary to use the constructor or `Train()` variants that take an initialized
+`initializer` object.  That would only be necessary for a custom
+`DictionaryInitializer` class that stored internal members.
+
+---
+
+The example below uses `NothingInitializer` to set a specific initial
+dictionary.
+
+```c++
+// See https://datasets.mlpack.org/satellite.train.csv.
+arma::mat trainData;
+mlpack::data::Load("satellite.train.csv", trainData, true);
+
+const size_t atoms = 25;
+const double lambda1 = 0.1;
+const double lambda2 = 0.0;
+const size_t maxIterations = 50;
+
+// Use a uniform random matrix as the initial dictionary.
+arma::mat initialDictionary(trainData.n_rows, atoms, arma::fill::randu);
+
+mlpack::SparseCoding sc(atoms, lambda1, lambda2, maxIterations);
+sc.Dictionary() = initialDictionary;
+
+const double obj = sc.Train<mlpack::NothingInitializer>(trainData);
+std::cout << "Training set objective: " << obj << "." << std::endl;
+```
+
+---
+
+ * An entirely custom class can also be implemented.  The class must implement
+   one method, `Initialize()`:
+
+```c++
+// You can use this as a starting point for implementation.
+class CustomDictionaryInitializer
+{
+ public:
+  // Initialize the dictionary to have the given number of atoms, given the
+  // dataset.  MatType will be the matrix type used by the sparse coding model
+  // (e.g. `arma::mat`, `arma::fmat`, etc.).
+  template<typename MatType>
+  void Initialize(const MatType& data,
+                  const size_t atoms,
+                  MatType& dictionary);
+};
+```
diff -pruN 4.3.0-2/doc/user/sample_ml_app.md 4.4.0-1/doc/user/sample_ml_app.md
--- 4.3.0-2/doc/user/sample_ml_app.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/doc/user/sample_ml_app.md	2024-05-28 20:08:34.000000000 +0000
@@ -27,13 +27,13 @@ dependencies in Release Mode).
 - Under C/C++ > General > Additional Include Directories add:
 ```
  - C:\mlpack\armadillo-9.800.3\include
- - C:\mlpack\mlpack-4.3.0\src
+ - C:\mlpack\mlpack-4.4.0\src
  - C:\mlpack\ensmallen-2.19.0\include
  - C:\mlpack\cereal-3.1.2\include
 ```
 - Under Build Events > Post-Build Event > Command Line add:
 ```
- - xcopy /y "C:\mlpack\mlpack-4.3.0\packages\OpenBLAS.0.2.14.1\lib\native\bin\x64\*.dll" $(OutDir)
+ - xcopy /y "C:\mlpack\mlpack-4.4.0\packages\OpenBLAS.0.2.14.1\lib\native\bin\x64\*.dll" $(OutDir)
 ```
 
 *Note*: recent versions of Visual Studio set "Conformance Mode" enabled by
diff -pruN 4.3.0-2/scripts/README.md 4.4.0-1/scripts/README.md
--- 4.3.0-2/scripts/README.md	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/scripts/README.md	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,71 @@
+This directory contains some utility scripts for mlpack, meant to be used during
+the development and release process.  All scripts should be run from the root
+directory.
+
+#### `build-docs.sh`
+
+This will convert all the Markdown documentation in `doc/` into HTML in
+`doc/html/`.  This script is used as part of the documentation deployment
+process but also can be run manually.
+
+The `kramdown` parser with the `parser-gfm` and `rouge` extensions installed is
+necessary, as are the `tidy` and `checklink` HTML checking packages.
+
+```sh
+scripts/build-docs.sh
+```
+
+This will build all the documentation and check for any broken links.  You can
+then go to `doc/html/index.html` to browse it.
+
+#### `test-docs.sh`
+
+This can be used to compile and run C++ code blocks from Markdown documentation,
+either for all Markdown files in a directory or for a specific Markdown file.
+
+```sh
+export CXX=g++
+# Adapt the flags below to your setup.
+export CXXFLAGS="-O3 -I/path/to/ensmallen/include/ -I/path/to/armadillo/include/ -I/path/to/cereal/include/"
+export LDFLAGS="-fopenmp"
+scripts/test-docs.sh doc/
+```
+
+That command first sets the compiler to use and the compilation flags, and then
+all C++ snippets from all Markdown files in `doc/` will be compiled with those
+flags.  Each snippet is enclosed in an `int main() { }` function, and
+`mlpack.hpp` is included.
+
+All snippets are extracted into `doc/build/`.
+
+It is also possible to run for just one file:
+
+```
+export CXX=g++
+# Adapt the flags below to your setup.
+export CXXFLAGS="-O3 -I/path/to/ensmallen/include/ -I/path/to/armadillo/include/ -I/path/to/cereal/include/"
+export LDFLAGS="-fopenmp"
+scripts/test-docs.sh doc/user/core.md
+```
+
+When running for just one file, the full code for each snippet will be printed,
+along with the output when the program is run.  When running for multiple files,
+this will be omitted (it would be too much output!).
+
+#### `release-mlpack.sh`
+
+Run this to open a pull request to release a new version of mlpack.
+You need the Hub command-line tool installed (see https://hub.github.com/).
+
+```
+scripts/release-mlpack.sh rcurtin 4 3 0
+```
+
+This will use Hub as the user `rcurtin` to open a PR to release mlpack 4.3.0.
+
+Be sure that `HISTORY.md` is up to date before running this!
+
+#### `update-website-after-release.sh`
+
+This is run automatically on mlpack.org after a new version of mlpack is
+released.
diff -pruN 4.3.0-2/scripts/build-docs.sh 4.4.0-1/scripts/build-docs.sh
--- 4.3.0-2/scripts/build-docs.sh	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/scripts/build-docs.sh	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,432 @@
+#!/usr/bin/env bash
+#
+# Convert all the Markdown files in doc/ to HTML.
+# This requires `kramdown` to be available on the path.
+# `tidy` and `checklink` (from Debian's w3c-linkchecker package) are used to
+# test the output and must also be available and on the path.
+# Run this from the root directory of the repository.
+# The output directory can be specified as the first option.
+
+if [ "$#" -gt 1 ]; then
+  echo "Usage: $0 [output_dir/]";
+  exit 1;
+elif [ "$#" -eq 1 ]; then
+  output_dir=$1;
+else
+  output_dir=doc/html;
+fi
+
+# If the header and footer already exist, they will not be overwritten.
+template_html_header="${output_dir}/template.html.header";
+template_html_footer="${output_dir}/template.html.footer";
+template_html_sidebar="${output_dir}/template.html.sidebar";
+
+if ! command -v kramdown &>/dev/null
+then
+  echo "kramdown not installed!  Cannot build documentation.";
+  exit 1;
+fi
+
+if ! command -v tidy &>/dev/null
+then
+  echo "tidy not installed!  Cannot build documentation.";
+  exit 1;
+fi
+
+if ! command -v checklink &>/dev/null
+then
+  echo "checklink not installed!  Cannot build documentation.";
+  exit 1;
+fi
+
+if [ ! -d doc/ ];
+then
+  echo "Run this script from the root of the mlpack repository.";
+  exit 1;
+fi
+
+# Define utility function to run kramdown and turn an .md file to an .html file.
+run_kramdown()
+{
+  input_file=$1;
+  # This converts, e.g., ./doc/user/index.md -> doc/html/user/index.html.
+  tmp=${input_file#./doc/}; # Strip leading ./doc/.
+  output_file="$output_dir/${tmp%.md}.html";
+
+  # Determine what the link root is.  If we're in the root directory, it's
+  # nothing, otherwise it's one of more '../'s.
+  dir_name=$(dirname $tmp);
+  link_root="";
+  if [[ "$dir_name" != "." ]];
+  then
+    levels_below_root=`echo $dir_name | awk -F'/' '{ print NF }'`;
+    link_root=$(printf '../%.0s' `seq 1 $levels_below_root`);
+  fi
+
+  # Make the enclosing directory if needed.
+  out_dir=`dirname "$output_file"`;
+  mkdir -p "$out_dir";
+
+  # Kramdown doesn't detect languages correctly with the "```" fence; instead it
+  # needs the "~~~" fence.
+  sed 's/^```/~~~/' $input_file > $input_file.tmp;
+
+  # Our documentation is full of relative links, like
+  # [name](other_file.md#anchor).  We need these to turn into links to the
+  # rendered HTML file, like [name](other_file.html#anchor).  We'll do this with
+  # regular expressions...
+  #
+  # - Note that this assumes there are no spaces in any filenames.
+  # - We also only catch the second part of the link '](' because the name of
+  #   the link could be spread on multiple lines.
+  #
+  # We start by trying to catch the special cases README.md and HISTORY.md,
+  # which our documentation puts in a slightly different place.  In addition,
+  # because those files are being moved to the root of the documentation, we
+  # must adjust links differently.
+  if [[ $input_file != "README.md" ]] && [[ $input_file != "HISTORY.md" ]];
+  then
+    sed -i "s|\]([./]*README.md)|](${link_root}README.html)|g" $input_file.tmp;
+    sed -i "s|\]([./]*README.md#[0-9]-\([^ ]*\))|](${link_root}README.html#\1)|g" $input_file.tmp;
+    sed -i 's/\](\([^ ]*\).md)/](\1.html)/g' $input_file.tmp;
+    sed -i 's/\](\([^ ]*\).md#\([^ ]*\))/](\1.html#\2)/g' $input_file.tmp;
+  else
+    sed -i 's/\](doc\/\([^ ]*\).md)/](\1.html)/g' $input_file.tmp;
+    sed -i 's/\](doc\/\([^ ]*\).md#\([^ ]*\))/](\1.html#\2)/g' $input_file.tmp;
+
+    # The README specifically has a link to GOVERNANCE.md, but we want to
+    # preserve that.  We're not building that file into Markdown.
+    sed -i 's|(./GOVERNANCE.md)|(https://github.com/mlpack/mlpack/blob/master/GOVERNANCE.md)|' $input_file.tmp;
+
+    # Ugh!  Github naming of anchors is different than kramdown, and so we have
+    # to adjust all the table-of-contents anchor links in the README (and in
+    # that file only).
+    sed -i 's/\](#[0-9][0-9]-\([^ ]*\))/](#\1)/g' $input_file.tmp;
+    sed -i 's/\](#[0-9]-\([^ ]*\))/](#\1)/g' $input_file.tmp;
+    sed -i 's/\](#[0-9][0-9]\([^ ]*\))/](#\1)/g' $input_file.tmp;
+    sed -i 's/\](#[0-9]\([^ ]*\))/](#\1)/g' $input_file.tmp;
+
+    # For HISTORY.md, we want to turn all references to Github issues into
+    # actual links, and all references to Github usernames into links to their
+    # profile.
+    if [[ $input_file == "HISTORY.md" ]];
+    then
+      sed -i 's/#\([0-9][0-9]*\)/[#\1](https:\/\/github.com\/mlpack\/mlpack\/issues\/\1)/g' $input_file.tmp;
+      sed -i 's/\([^`]\)@\([a-zA-Z0-9_-][a-zA-Z0-9_-]*\)/\1[@\2](https:\/\/github.com\/\2)/g' $input_file.tmp;
+    fi
+  fi
+
+  # Replace any links to source files with a link to the current version of the
+  # source file on Github.
+  sed -i 's/\](\/src\/\([^ ]*\)\.hpp)/](https:\/\/github.com\/mlpack\/mlpack\/blob\/master\/src\/\1.hpp)/' $input_file.tmp;
+
+  # If this is binding documentation or quickstart documentation, don't set the
+  # default language to C++.
+  set_lang=1;
+  if [[ `dirname $input_file` == "./doc/user/bindings" ]];
+  then
+    set_lang=0;
+  elif [[ `dirname $input_file` == "./doc/quickstart" ]];
+  then
+    if [[ `basename $input_file .md` != "cpp" ]];
+    then
+      set_lang=0;
+    fi
+  elif [[ $input_file == "HISTORY.md" ]];
+  then
+    set_lang=0;
+  fi
+
+  if [[ "$set_lang" == "0" ]];
+  then
+    kramdown \
+        -x parser-gfm \
+        --syntax-highlighter rouge \
+        --auto_ids \
+        $input_file.tmp > "$output_file.tmp" || exit 1;
+  else
+    kramdown \
+        -x parser-gfm \
+        --syntax-highlighter rouge \
+        --syntax-highlighter-opts '{ default_lang: c++ }' \
+        --auto_ids \
+        $input_file.tmp > "$output_file.tmp" || exit 1;
+  fi
+  cat "$template_html_header" | sed "s|LINKROOT|$link_root|" > "$output_file";
+
+  # Create the sidebar.  Extract anchors from the page, unless we are looking at
+  # index.md, since the permanent part of the sidebar links all over index.md
+  # anyway.  If we are looking at binding documentation, use a slightly
+  # different sidebar.
+  if { [[ $dir_name != "user/bindings" ]] && \
+       [[ $dir_name != "quickstart" ]] } ||
+     [[ $input_file == "./doc/quickstart/cpp.md" ]];
+  then
+    cat "$template_html_sidebar" | sed "s|LINKROOT|$link_root|" \
+        >> "$output_file";
+    create_page_sidebar_section "$output_file.tmp" "$output_file" "$dir_name";
+  else
+    echo "Using custom sidebar...";
+    cat "$template_html_sidebar" | sed "s|LINKROOT|$link_root|" |\
+        sed 's|<details> <!-- default closed for non-binding pages -->|<details open="true">|' |\
+        sed 's|<details open="true"> <!-- default open for non-binding pages -->|<details>|' \
+        >> "$output_file";
+    # Some pages may have a custom sidebar HTML file.  (Specifically,
+    # generated language bindings.)
+    if [[ $dir_name == "user/bindings" ]];
+    then
+      cat "${input_file/%.md/.sidebar.html}" | sed "s|LINKROOT|$link_root|" \
+          >> "$output_file";
+    else
+      sidebar_file=`basename $input_file .md`.sidebar.html;
+      cat "./doc/user/bindings/$sidebar_file" | sed "s|LINKROOT|$link_root|" \
+          >> "$output_file";
+    fi
+  fi
+
+  # Add clickable anchors to h2 and h3 headers.
+  echo "<div id=\"content\">" >> "$output_file";
+  sed -E 's/<h([23]) id="([^"]*)">/<h\1 id="\2"><a href="#\2" class="pl">🔗<\/a> /' "$output_file.tmp" >> "$output_file";
+
+  # Simple postprocessing to make tidy a little happier.
+  # (Muting the warning won't change the error code!)
+  sed -i 's/<table>/<table summary="">/' "$output_file";
+
+  cat "$template_html_footer" >> "$output_file";
+  rm -f $input_file.tmp "$output_file.tmp";
+}
+
+# Create the template header file.
+create_template_header()
+{
+  output_file="$1";
+
+  # Note that LINKROOT will be substituted into place by run_kramdown.
+  cat > "$output_file" << EOF
+<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
+<html>
+<head>
+  <meta content="text/html; charset=utf-8" http-equiv="Content-Type">
+  <meta name="viewport" content="width=device-width, initial-scale=1.0">
+  <link type="text/css" rel="stylesheet" href="LINKROOTgfm-mod.css">
+  <link type="text/css" rel="stylesheet" href="LINKROOTrouge-github-mod.css">
+  <title>mlpack documentation</title>
+</head>
+<body>
+EOF
+}
+
+# Create the template footer.
+create_template_footer()
+{
+  output_file="$1";
+
+  cat > "$output_file" << EOF
+</div>
+</body>
+</html>
+EOF
+}
+
+# Extract anchors to build a sidebar.
+# This should take the input HTML (before anchor elements are added), and it
+# appends a sidebar list to the output HTML.
+create_page_sidebar_section()
+{
+  sb_input_file="$1";
+  sb_output_file="$2";
+  sb_dir_name="$3"; # The directory containing the documentation.
+  sb_input_file_base=`basename "$sb_input_file" .html.tmp`;
+
+  # Extract h2/h3 anchors into a list.  For individual method documentation, we
+  # only extract h3 anchors because those use h2s as their headings.  And, for
+  # core.md, we want to extract both h2 and h3 anchors.
+  if [[ "$sb_dir_name" == "user/methods" ]];
+  then
+    # The page title on individual methods is encoded as an h2.
+    page_title=`grep '<h2 id=' "$sb_input_file" |\
+        head -1 |\
+        sed 's/^<h2 id="[^"]*">\(.*\)<\/h2>/\1/'`;
+
+    grep '<h3 id=' "$sb_input_file" | sed 's/<h3 id="\([^"]*\)">\(.*\)<\/h3>/<li><a href="#\1">\2<\/a><\/li>/' > "$sb_output_file.side.tmp";
+  elif [[ "$sb_input_file_base" == "core" ]];
+  then
+    # The page title on the core class documentation page is encoded as an h1.
+    page_title=`grep '<h1 id=' "$sb_input_file" |\
+        head -1 |\
+        sed 's/^<h1 id="[^"]*">\(.*\)<\/h1>/\1/'`;
+
+    # We want to collect h2s and h3s as individual documentation; each h2 should
+    # have a summary/details block.  This is a little tedious to create... we'll
+    # do this by creating a temporary tab-separated file with lines like
+    #
+    # h2  anchor_name   Anchor Title
+    # h3  anchor_name   Anchor Title
+    # ...
+    #
+    # and then we'll construct the actual sidebar using that list.
+    grep '<h[23] id=' "$sb_input_file" |\
+        sed 's/^<\(h[23]\) id="\([^"]*\)">\(.*\)<\/h[23]>/\1\t\2\t\3/' \
+        > "$sb_output_file.side.list.tmp";
+    in_block=0;
+    while read line; do
+      # First, extract the pieces of each line.
+      line_type=`echo "$line"    | awk -F'\t' '{ print $1 }'`;
+      anchor_name=`echo "$line"  | awk -F'\t' '{ print $2 }'`;
+      anchor_title=`echo "$line" | awk -F'\t' '{ print $3 }'`;
+
+      # For an h2, we have to print a summary block.
+      # Note that this assumes that *all* h2s have h3 children.  If that's not
+      # true, some extra processing will be needed.
+      if [ "$line_type" = "h2" ];
+      then
+        if [ "$in_block" = "1" ];
+        then
+          # We have to close the previous block.
+          echo "</ul></details></li>" >> "$sb_output_file.side.tmp";
+        fi
+
+        # Create the new details block.
+        echo "<li><details><summary>" >> "$sb_output_file.side.tmp";
+        echo "<a href=\"#$anchor_name\">" >> "$sb_output_file.side.tmp";
+        echo "$anchor_title" >> "$sb_output_file.side.tmp";
+        echo "</a>" >> "$sb_output_file.side.tmp";
+        echo "</summary>" >> "$sb_output_file.side.tmp";
+        echo "<ul>" >> "$sb_output_file.side.tmp";
+        in_block=1;
+      else
+        echo "  <li><a href=\"#$anchor_name\">" >> "$sb_output_file.side.tmp";
+        echo "  $anchor_title" >> "$sb_output_file.side.tmp";
+        echo "  </a></li>" >> "$sb_output_file.side.tmp";
+      fi
+    done < "$sb_output_file.side.list.tmp";
+
+    # Close the last h2 block, if we need to.
+    if [ "$in_block" = "1" ];
+    then
+      echo "</ul></details></li>" >> "$sb_output_file.side.tmp";
+    fi
+
+    rm -f "$sb_output_file.side.list.tmp";
+  else
+    # On other pages, the page title is encoded as an h1.
+    page_title=`grep '<h1 id=' "$sb_input_file" |\
+        head -1 |\
+        sed 's/^<h1 id="[^"]*">\(.*\)<\/h1>/\1/'`;
+
+    grep '<h2 id=' "$sb_input_file" |\
+        sed 's/<h2 id="\([^"]*\)">\(.*\)<\/h2>/<li><a href="#\1">\2<\/a><\/li>/' \
+        > "$sb_output_file.side.tmp";
+  fi
+  lines=`cat "$sb_output_file.side.tmp" | wc -l`;
+
+  echo "<ul>" >> "$sb_output_file";
+
+  # Make the top of the sidebar.
+  if [ -n "$page_title" ];
+  then
+    echo "<li class=\"page_title\"><b>$page_title</b> <a href=\"#\">[top]</a>" >> "$sb_output_file";
+  else
+    echo "<li><a href=\"#\">[top of page]</a>" >> "$sb_output_file";
+  fi
+
+  if [[ "$lines" -gt 0 ]];
+  then
+    echo "<ul>" >> "$sb_output_file";
+    cat "$sb_output_file.side.tmp" >> "$sb_output_file";
+    echo "</ul>" >> "$sb_output_file";
+  fi
+  echo "</li>" >> "$sb_output_file";
+  echo "</ul>" >> "$sb_output_file";
+  echo "</div>" >> "$sb_output_file";
+
+  rm -f "$sb_output_file.side.tmp";
+}
+
+# Save any existing template.
+if [ -f "$template_html_header" ];
+then
+  mv "$template_html_header" template.html.header.tmp;
+fi
+
+if [ -f "$template_html_footer" ];
+then
+  mv "$template_html_footer" template.html.footer.tmp;
+fi
+
+rm -rf "$output_dir";
+mkdir -p "$output_dir";
+cp -v doc/css/* "$output_dir";
+mkdir -p "$output_dir/img/";
+cp -v doc/img/* "$output_dir/img/";
+mkdir -p "$output_dir/tutorials/res/";
+cp -v doc/tutorials/res/* "$output_dir/tutorials/res/";
+
+# Create the template files we will use, if they don't already exist.
+if [ -f template.html.header.tmp ];
+then
+  mv template.html.header.tmp "$template_html_header";
+else
+  create_template_header "$template_html_header";
+  del_header=1;
+fi
+
+if [ -f template.html.footer.tmp ];
+then
+  mv template.html.footer.tmp "$template_html_footer";
+else
+  create_template_footer "$template_html_footer";
+  del_footer=1;
+fi
+
+cp doc/sidebar.html "$template_html_sidebar";
+
+# Process all the .md files.
+for f in README.md HISTORY.md `find ./doc/ -iname '*.md'`;
+do
+  # Skip the JOSS paper...
+  if [[ $f == *"joss_paper"* ]]; then
+    continue;
+  fi
+
+  echo "Processing $f...";
+  run_kramdown $f;
+
+  # This converts, e.g., ./doc/user/index.md -> doc/html/user/index.html.
+  tmp=${f#./doc/}; # Strip leading ./doc/.
+  of="$output_dir/${tmp%.md}.html";
+
+  tidy -qe "$of" || exit 1;
+done
+
+# Now take a second pass to check all the links.
+find "$output_dir" -iname '*.html' -print0 | while read -d $'\0' f
+do
+  echo "Checking links in $f...";
+
+  # To run checklink we have to strip out some perl stderr warnings...
+  checklink -qs \
+      --follow-file-links \
+      --suppress-broken 405 \
+      --suppress-broken 301 \
+      -X "https://eigen.tuxfamily.org/index.php\?title=Main_Page" \
+      -X "https://mlpack.slack.com/" "$f" 2>&1 |
+      grep -v 'Use of uninitialized value' > checklink_out;
+  if [ -s checklink_out ];
+  then
+    cat checklink_out;
+    exit 1;
+  fi
+  rm -f checklink_out;
+done
+
+# Remove temporary files.
+if [ "a$del_header" == "a1" ];
+then
+  rm -f "$template_html_header";
+fi
+
+if [ "a$del_footer" == "a1" ];
+then
+  rm -f "$template_html_footer";
+fi
diff -pruN 4.3.0-2/scripts/check-markdown-docs.sh 4.4.0-1/scripts/check-markdown-docs.sh
--- 4.3.0-2/scripts/check-markdown-docs.sh	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/scripts/check-markdown-docs.sh	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,53 @@
+#!/usr/bin/env bash
+#
+# This script checks to ensure that generated Markdown documentation is the same
+# as what's committed to the repository.  After configuring mlpack with
+# -DBUILD_MARKDOWN_BINDINGS=ON, use this simple script to detect changes in the
+# documentation that should be committed.
+#
+# The only argument of the script is to the build directory.
+if [ "$#" -ne 1 ];
+then
+  echo "Usage: $0 build/";
+  echo "  (replace build/ with your build directory, where you already"
+  echo "   ran 'make markdown')";
+  exit 1;
+fi
+
+build_dir="$1";
+
+# Check that Markdown documentation has been built.
+if [[ ! -d "$build_dir/doc/" ]];
+then
+  echo "$build_dir/doc/ does not exist!";
+  echo "Did you run 'make markdown' in your build directory ($build_dir)?";
+  exit 1;
+fi
+
+# Now check every file in the main repository bindings.
+for f in doc/user/bindings/*;
+do
+  echo "Checking $f...";
+  f_base=`basename $f`;
+
+  if [[ ! -f "$build_dir/doc/$f_base" ]];
+  then
+    echo "$build_dir/doc/$f_base does not exist!";
+    echo "Did you run 'make markdown' in your build directory ($build_dir)?";
+    echo "Or does the file need to be removed from the repository?";
+    exit 1;
+  fi
+
+  diff -Nau $f "$build_dir/doc/$f_base";
+
+  if [ "$?" -ne 0 ];
+  then
+    echo "";
+    echo "";
+    echo "Files $f and $build_dir/doc/$f_base differ!  (See above.)";
+    echo "";
+    echo "If the sidebar differs, be sure to check if updates are needed in ";
+    echo "quickstart/*.sidebar.html!";
+    exit 1;
+  fi
+done
diff -pruN 4.3.0-2/scripts/release-mlpack.sh 4.4.0-1/scripts/release-mlpack.sh
--- 4.3.0-2/scripts/release-mlpack.sh	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/scripts/release-mlpack.sh	2024-05-28 20:08:34.000000000 +0000
@@ -2,7 +2,7 @@
 #
 # Release a new version of mlpack.
 #
-# Usage: release-mlpack.sh X Y Z
+# Usage: release-mlpack.sh github_username X Y Z
 #
 # where X is the major version, Y is the minor version, and Z is the patch
 # version.  Run this from the root of the repository.
diff -pruN 4.3.0-2/scripts/test-docs.sh 4.4.0-1/scripts/test-docs.sh
--- 4.3.0-2/scripts/test-docs.sh	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/scripts/test-docs.sh	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,407 @@
+#!/usr/bin/env bash
+#
+# Extract C++ code blocks from either an individual Markdown file or a directory
+# full of Markdown files.  This does roughly what you would expect it to, but
+# there is a little bit of magic:
+#
+#  * All ```c++ code blocks are extracted into their own .cpp files.
+#
+#  * mlpack.hpp is included in each file, and each code block is placed in an
+#      `int main() { }` block.
+#
+#  * If Eigen or xtensor is detected, correct includes are added.
+#
+#  * All data referenced in http URLs is downloaded.
+#
+#  * If a code block is just a class declaration, it will be added to the *next*
+#      code block file.
+#
+# Once all code blocks are compiled, they are run just to make sure they run
+# correctly.  If a single file was passed, all inputs and outputs are printed;
+# if an entire directory is given, program output is only printed on error.
+
+if [[ $# -ne 1 ]];
+then
+  echo "Usages: " >&2;
+  echo " - $0 input_file.md" >&2;
+  echo " - $0 markdown_directory/" >&2;
+  exit 2;
+fi
+
+if [ -z "$CXX" ];
+then
+  echo "You must set \$CXX to the compiler you want to use!" >&2;
+  exit 1;
+fi
+
+if [ -z "$CXXFLAGS" ];
+then
+  echo "Warning: \$CXXFLAGS is unset.  If ensmallen, Armadillo, STB, or cereal ";
+  echo "  are not in standard locations, builds will fail!  Be sure to use ";
+  echo "  absolute paths, not relative paths.";
+fi
+
+if [ -z "$LDFLAGS" ];
+then
+  echo "Warning: \$LDFLAGS is unset.  If libarmadillo.so is not in a standard ";
+  echo "  location, builds will fail!";
+fi
+
+# First determine what files we are looking at.
+if [ -d $1 ];
+then
+  files=`find $1 -iname '*.md'`;
+  mode="directory";
+else
+  files=$1;
+  mode="file";
+fi
+
+# Extract the C++ code blocks from a particular file, creating
+# $output_prefix1.cpp, $output_prefix2.cpp, and so on and so forth.
+#
+# The code in those snippets will be placed into an int main() { } block, and
+# mlpack.hpp will be included.
+extract_code_blocks()
+{
+  input_file=$1;
+  output_prefix=$2;
+
+  # Extract into temporary files.
+  sed -n '/^```c++/,/^```/ p' < $input_file > $input_file.tmp;
+  output_file_id=0;
+  output_file_display="00"; # Hopefully no file has more than 100 examples...
+
+  # Track whether or not the last line was a fence, since we get them two at a
+  # time.  We initially set this to 1, because the first fence does not have a
+  # preceding fence close above it.
+  last_line_fence=1;
+
+  # Track whether or not the entire last file corresponded to a class
+  # declaration.
+  class_decl=0;
+
+  while IFS= read -r line;
+  do
+    if [[ $last_line_fence == 1 ]];
+    then
+      # Skip this line---it will be a fence opening.
+      last_line_fence=0;
+
+      # Create main() function to wrap the code in.
+      echo "#include <mlpack.hpp>" > $output_prefix$output_file_display.cpp;
+      echo "" >> $output_prefix$output_file_display.cpp;
+
+      # If we have a class declaration from the previous file, insert it.
+      if [[ $class_decl == 1 ]];
+      then
+        class_decl=0;
+        last_output_file_id=$(($output_file_id - 1));
+        last_output_file_display=$(printf "%02d" $last_output_file_id);
+
+        cat $output_prefix$last_output_file_display.cpp | awk '
+            BEGIN { p=0 }
+            /int main()/ { p=1 }
+            /^{/ { if(p == 1) { p=2; o=1 } }
+            /^}/ { p=0; }
+            // { if (p == 2 && o == 0) { print substr($0, 3) } o=0 }' >> $output_prefix$output_file_display.cpp;
+        echo "" >> $output_prefix$output_file_display.cpp;
+        rm -f $output_prefix$last_output_file_display.cpp;
+      fi
+
+      echo "int main()" >> $output_prefix$output_file_display.cpp;
+      echo "{" >> $output_prefix$output_file_display.cpp;
+      continue;
+    fi
+
+    if [[ $line == '```'* ]];
+    then
+      last_line_fence=1;
+
+      # Close main() function.
+      echo "}" >> $output_prefix$output_file_display.cpp;
+
+      # Check after the fact: was this file only a class declaration?  If so, we
+      # want to put it instead into the next file.
+      has_class1=`grep '^  class' $output_prefix$output_file_display.cpp | wc -l`;
+      has_class2=`grep '^  };' $output_prefix$output_file_display.cpp | wc -l`;
+      if [[ "$has_class1" != "0" && "$has_class2" != "0" ]];
+      then
+        class_decl=1;
+      fi;
+
+      # Detect if we need any to add any special headers.  We have to do this
+      # when we finish with the file...
+      if [[ `grep 'Eigen::' $output_prefix$output_file_display.cpp | wc -l` -gt 0 ]];
+      then
+        sed -i '1s/^/#include <Eigen\/Dense>\n/' $output_prefix$output_file_display.cpp;
+      fi
+
+      if [[ `grep 'xt::' $output_prefix$output_file_display.cpp | wc -l` -gt 0 ]];
+      then
+        sed -i '1s/^/#include <xtensor\/xrandom.hpp>\n/' $output_prefix$output_file_display.cpp;
+        sed -i '1s/^/#include <xtensor\/xarray.hpp>\n/' $output_prefix$output_file_display.cpp;
+      fi
+
+      output_file_id=$(($output_file_id + 1));
+      output_file_display=$(printf "%02d" $output_file_id);
+
+      continue;
+    fi
+
+    # Include indentation (two spaces).
+    echo "  $line" >> $output_prefix$output_file_display.cpp;
+  done < $input_file.tmp;
+
+  # The last file is always invalid---we opened it without knowing whether
+  # anything would be in it.
+  rm -f $output_prefix$output_file_display.cpp;
+
+  # Check the "true" last file: if it's only class declarations, no need to
+  # compile it.
+  output_file_id=$(($output_file_id - 1));
+  output_file_display=$(printf "%02d" $output_file_id);
+  if [ -f $output_prefix$output_file_display.cpp ];
+  then
+    cat $output_prefix$output_file_display.cpp | awk '
+        BEGIN { p=0 }
+        /int main()/ { p=1 }
+        /^{/ { if(p == 1) { p=2; o=1 } }
+        /^}/ { p=0 }
+        // { if (p == 2 && o == 0) { print substr($0, 3) } o=0 }' >> $output_prefix$output_file_display.cpp.tmp;
+    has_class1=`grep '^class' $output_prefix$output_file_display.cpp.tmp | wc -l`;
+    has_class2=`grep '^};' $output_prefix$output_file_display.cpp.tmp | wc -l`;
+    if [[ "$has_class1" != "0" && "$has_class2" != "0" ]];
+    then
+      # The file's main() function is just a class declaration.  Nuke it.
+      rm -f $output_prefix$output_file_display.cpp;
+    fi
+    rm -f $output_prefix$output_file_display.cpp.tmp;
+  fi
+
+  rm -f $input_file.tmp;
+}
+
+compile_code_blocks()
+{
+  input_dir=$1;
+
+  # If there are no files to compile, leave early.
+  if ! compgen -G $input_dir/*.cpp >/dev/null;
+  then
+    return;
+  fi
+
+  for f in $input_dir/*.cpp;
+  do
+    echo "  Compiling $f...";
+    of=${f%.cpp};
+
+    if ! $CXX -std=c++17 -Isrc/ $CXXFLAGS -o $of $f $LDFLAGS -larmadillo 2>$of.tmp;
+    then
+      echo "Compilation of the following program failed:";
+      echo "";
+      cat $f;
+      echo "";
+      echo "First ten lines of error output:";
+      head $of.tmp;
+      echo "";
+      echo "For full error output run either:";
+      echo " - less $of.tmp";
+      echo " - $CXX -std=c++17 -Isrc/ $CXXFLAGS -o $of $f $LDFLAGS -larmadillo";
+      echo "";
+      echo "Did you set \$CXX, \$CXXFLAGS, and \$LDFLAGS correctly?"
+      exit 1;
+    fi
+  done
+}
+
+download_http_artifacts()
+{
+  input_dir=$1;
+  output_dir=$2;
+
+  # Get a list of all HTTP resources.
+  artifacts=`grep 'http[s]*://' $input_dir/*.cpp |\
+      sed 's/^.*\(http[^ ]*\).*$/\1/' |\
+      sort |\
+      uniq |\
+      grep 'csv\|arff\|bin\|png' |\
+      sed 's/\.$//'`;
+  cd $output_dir;
+  for a in $artifacts;
+  do
+    out_a=`basename $a`;
+    if [ ! -f $out_a ];
+    then
+      echo "  Downloading $a...";
+      if ! curl -s -O $a;
+      then
+        echo "Error downloading $a!";
+        exit 1;
+      fi
+    fi
+  done
+  cd - >/dev/null;
+
+  # Special case: if we are looking at core.md, this has two special files we
+  # need to create that is used in the example.
+  f=`basename $input_dir`;
+  if [[ "$f" == "core" || "$f" == "matrices" ]];
+  then
+    cd $output_dir;
+    echo "  Creating data.csv...";
+    cat > data.csv << EOF
+3,3,3,3,0
+3,4,4,3,0
+3,4,4,3,0
+3,3,4,3,0
+3,6,4,3,0
+2,4,4,3,0
+2,4,4,1,0
+3,3,3,2,0
+3,4,4,2,0
+3,4,4,2,0
+3,3,4,2,0
+3,6,4,2,0
+2,4,4,2,0
+EOF
+
+    echo "  Creating mixed_string_data.csv...";
+    cat > mixed_string_data.csv << EOF
+3,"hello",3,"f",0
+3,"goodbye",4,"f",0
+3,"goodbye",4,"e",0
+3,"hello",4,"d",0
+3,"hello",4,"d",0
+2,"hello",4,"d",0
+2,"hello",4,"d",0
+3,"goodbye",3,"f",0
+3,"goodbye",4,"f",0
+3,"hello",4,"f",0
+3,"hello",4,"c",0
+3,"hello",4,"f",0
+2,"hello",4,"c",0
+EOF
+    cd - >/dev/null;
+  fi
+}
+
+run_code_blocks()
+{
+  input_dir=$1;
+
+  for f in $input_dir/*.cpp;
+  do
+    f_exec=${f%.cpp};
+    if [[ "$mode" == "directory" ]];
+    then
+      echo "  Running $f_exec...";
+      if ! ./$f_exec 2>&1 >/dev/null;
+      then
+        echo "  Error running $f_exec!";
+        exit 1;
+      fi
+    else
+      echo "  ---------------------------------------------------------------------  ";
+      echo "  Contents of $f:";
+      echo "";
+      cat $f;
+      echo "";
+      echo "  ---------------------------------------------------------------------  ";
+      echo "  Output of $f_exec:";
+      echo "";
+
+      if ! ./$f_exec;
+      then
+        echo "";
+        echo "Error running $f_exec!  See output above.";
+        exit 1;
+      fi
+      echo "";
+      echo "  ---------------------------------------------------------------------  ";
+    fi
+  done
+}
+
+# Main loop: process the files we were asked to process.
+mkdir -p doc/build/;
+for f in $files;
+do
+  if [[ "$mode" == "directory" ]];
+  then
+    declare -a files_to_skip=(
+        # These files have small incomplete snippets that can't compile into
+        # standalone programs.
+        "sample_ml_app.md"
+        "hpt.md"
+        "cv.md"
+        "timer.md"
+        "bindings.md"
+        "elemtype.md"
+        "iodoc.md"
+        "kernels.md"
+        "metrics.md"
+        "trees.md"
+        # The tutorials are old and are likely to be replaced, so let's not test
+        # them.
+        "amf.md"
+        "ann.md"
+        "approx_kfn.md"
+        "cf.md"
+        "datasetmapper.md"
+        "det.md"
+        "emst.md"
+        "fastmks.md"
+        "image.md"
+        "kmeans.md"
+        "linear_regression.md"
+        "neighbor_search.md"
+        "range_search.md"
+        "reinforcement_learning.md"
+        "asynchronous_learning.md"
+        "ddpg.md"
+        "q_learning.md"
+        "sac.md"
+        "td3.md"
+        # Skip quickstarts, although we should eventually test them.
+        "cpp.md"
+    );
+
+    skip=0;
+    for skip_f in "${files_to_skip[@]}";
+    do
+        base_f=`basename $f`;
+        if [ "$base_f" = "$skip_f" ];
+        then
+            skip=1;
+            break;
+        fi
+    done
+
+    if [[ $skip -eq 1 ]];
+    then
+      continue;
+    fi
+  fi
+
+  echo "Building documentation for $f...";
+
+  build_dir_tmp=${f#doc/};
+  build_dir=${build_dir_tmp%.md};
+  base_file=`basename $f .md`;
+  mkdir -p doc/build/$build_dir/;
+
+  extract_code_blocks $f doc/build/$build_dir/$base_file;
+  # If there are no C++ files, don't do anything else..
+  if ! compgen -G doc/build/$build_dir/*.cpp >/dev/null;
+  then
+    continue;
+  fi
+
+  compile_code_blocks doc/build/$build_dir;
+  download_http_artifacts doc/build/$build_dir doc/build/;
+  cd doc/build/;
+  run_code_blocks $build_dir;
+  cd ../../;
+done
diff -pruN 4.3.0-2/src/mlpack/base.hpp 4.4.0-1/src/mlpack/base.hpp
--- 4.3.0-2/src/mlpack/base.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/base.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,6 +24,7 @@
 #include <cmath>
 
 // Next, standard includes.
+#include <any>
 #include <cctype>
 #include <cfloat>
 #include <climits>
@@ -31,7 +32,9 @@
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
+#include <optional>
 #include <stdexcept>
+#include <string_view>
 #include <tuple>
 #include <utility>
 #include <numeric>
@@ -81,23 +84,8 @@
   #endif
 #endif
 
-#if defined(MLPACK_HAVE_CXX17)
-  #include <any>
-  #include <string_view>
-  #define MLPACK_ANY std::any
-  #define MLPACK_ANY_CAST std::any_cast
-  #define MLPACK_STRING_VIEW std::string_view
-#elif defined(_MSC_VER)
-  #error "When using Visual Studio, mlpack should be compiled with /std:c++17 or newer."
-#else
-  // Backport std::any from C+17 to C++11 to replace boost::any.
-  // Use mnmlstc backport implementation only if compiler does not
-  // support C++17.
-  #include <mlpack/core/std_backport/any.hpp>
-  #include <mlpack/core/std_backport/string_view.hpp>
-  #define MLPACK_ANY core::v2::any
-  #define MLPACK_ANY_CAST core::v2::any_cast
-  #define MLPACK_STRING_VIEW core::v2::string_view
+#if !defined(MLPACK_HAVE_CXX17)
+  #error "Need to enable C++17 mode in your compiler"
 #endif
 
 // Now include Armadillo through the special mlpack extensions.
diff -pruN 4.3.0-2/src/mlpack/bindings/Categories.cmake 4.4.0-1/src/mlpack/bindings/Categories.cmake
--- 4.3.0-2/src/mlpack/bindings/Categories.cmake	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/Categories.cmake	2024-05-28 20:08:34.000000000 +0000
@@ -2,10 +2,10 @@
 # category you choose is not in this list, an error will
 # be thrown.
 set (CATEGORIES
-    "classification"
-    "regression"
-    "clustering"
-    "geometry"
-    "preprocessing"
-    "misc. / other"
-    "transformations")
+    "Classification"
+    "Regression"
+    "Clustering"
+    "Geometry"
+    "Preprocessing"
+    "Misc. / Other"
+    "Transformations")
diff -pruN 4.3.0-2/src/mlpack/bindings/R/default_param_impl.hpp 4.4.0-1/src/mlpack/bindings/R/default_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/R/default_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/default_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,10 +34,22 @@ std::string DefaultParamImpl(
 {
   std::ostringstream oss;
   if (std::is_same<T, bool>::value)
-    oss << "FALSE";
+  {
+    // If this is the verbose option, print the default that uses the global
+    // package option.
+    if (data.name == "verbose")
+    {
+      oss << "getOption(\"mlpack.verbose\", FALSE)";
+    }
+    else
+    {
+      oss << "FALSE";
+    }
+  }
   else
-    oss << MLPACK_ANY_CAST<T>(data.value);
-
+  {
+    oss << std::any_cast<T>(data.value);
+  }
   return oss.str();
 }
 
@@ -51,7 +63,7 @@ std::string DefaultParamImpl(
 {
   // Print each element in an array delimited by square brackets.
   std::ostringstream oss;
-  const T& vector = MLPACK_ANY_CAST<T>(data.value);
+  const T& vector = std::any_cast<T>(data.value);
   oss << "c(";
   if (std::is_same<T, std::vector<std::string>>::value)
   {
@@ -92,7 +104,7 @@ std::string DefaultParamImpl(
     util::ParamData& data,
     const typename std::enable_if<std::is_same<T, std::string>::value>::type*)
 {
-  const std::string& s = *MLPACK_ANY_CAST<std::string>(&data.value);
+  const std::string& s = *std::any_cast<std::string>(&data.value);
   return "\"" + s + "\"";
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/R/get_param.hpp 4.4.0-1/src/mlpack/bindings/R/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/R/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ void GetParam(util::ParamData& d,
               const void* /* input */,
               void* output)
 {
-  *((T**) output) = const_cast<T*>(MLPACK_ANY_CAST<T>(&d.value));
+  *((T**) output) = const_cast<T*>(std::any_cast<T>(&d.value));
 }
 
 } // namespace r
diff -pruN 4.3.0-2/src/mlpack/bindings/R/get_printable_param.hpp 4.4.0-1/src/mlpack/bindings/R/get_printable_param.hpp
--- 4.3.0-2/src/mlpack/bindings/R/get_printable_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/get_printable_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -44,7 +44,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* = 0)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -61,7 +61,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& matrix = MLPACK_ANY_CAST<T>(data.value);
+  const T& matrix = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   oss << matrix.n_rows << "x" << matrix.n_cols << " matrix";
@@ -78,7 +78,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << data.cppType << " model at " << MLPACK_ANY_CAST<T*>(data.value);
+  oss << data.cppType << " model at " << std::any_cast<T*>(data.value);
   return oss.str();
 }
 
@@ -92,7 +92,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& tuple = MLPACK_ANY_CAST<T>(data.value);
+  const T& tuple = std::any_cast<T>(data.value);
   const arma::mat& matrix = std::get<1>(tuple);
 
   std::ostringstream oss;
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/DESCRIPTION.in 4.4.0-1/src/mlpack/bindings/R/mlpack/DESCRIPTION.in
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/DESCRIPTION.in	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/DESCRIPTION.in	2024-05-28 20:08:34.000000000 +0000
@@ -7,9 +7,9 @@ Description: A fast, flexible machine le
              aims to provide fast, extensible implementations of cutting-edge
              machine learning algorithms.  See also Curtin et al. (2023)
              <doi:10.21105/joss.05026>.
-SystemRequirements: A C++14 compiler. Version 5 or later of GCC will be fine.
+SystemRequirements: A C++17 compiler. Version 8 or later of GCC will be fine.
 License: BSD_3_clause + file LICENSE
-Depends: R (>= 4.0.0)
+Depends: R (>= 4.3.0)
 Imports: Rcpp (>= 0.12.12)
 LinkingTo: Rcpp,
            RcppArmadillo (>= @RcppArmadillo_Version@),
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/R/inline.R 4.4.0-1/src/mlpack/bindings/R/mlpack/R/inline.R
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/R/inline.R	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/R/inline.R	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,17 @@
+inlineCxxPlugin <- function(...) {
+    is_macos <- Sys.info()[["sysname"]] == "Darwin"
+    openmp_flag <- if (is_macos) "" else "$(SHLIB_OPENMP_CFLAGS)"
+    plugin <-
+        Rcpp::Rcpp.plugin.maker(
+            include.before = "#include <mlpack.h>",
+            libs           = paste(openmp_flag, "$(LAPACK_LIBS) $(BLAS_LIBS) $(FLIBS)"),
+            LinkingTo      = c("RcppArmadillo", "Rcpp", "RcppEnsmallen", "mlpack"),
+            package        = "mlpack"
+        )
+    settings <- plugin()
+    settings$env$PKG_CPPFLAGS <- paste("-I../inst/include", openmp_flag)
+    # C++17 is required
+    if (getRversion() < "4.2.0") settings$env$USE_CXX17 <- "yes"
+
+    settings
+}
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/R/package.R 4.4.0-1/src/mlpack/bindings/R/mlpack/R/package.R
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/R/package.R	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/R/package.R	2024-05-28 20:08:34.000000000 +0000
@@ -5,11 +5,9 @@
 #' learning algorithms. mlpack provides these algorithms as simple command-line
 #' programs, C++ classes and bindings for : Python, Julia, Go and R which can
 #' then be integrated into larger-scale machine learning solutions.
-#'
-#' @docType package
+#' 
 #' @name mlpack
 #' @aliases mlpack-package
-#' @author mlpack Team
 #' @importFrom Rcpp evalCpp
 #' @useDynLib mlpack
-NULL
+"_PACKAGE"
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/src/Makevars.win 4.4.0-1/src/mlpack/bindings/R/mlpack/src/Makevars.win
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/src/Makevars.win	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/src/Makevars.win	2024-05-28 20:08:34.000000000 +0000
@@ -1,3 +1,3 @@
-PKG_CXXFLAGS = -I. -I../inst/include $(SHLIB_OPENMP_CXXFLAGS) -ftrack-macro-expansion=0 -pipe --param ggc-min-expand=10 --param ggc-min-heapsize=8192
+PKG_CXXFLAGS = -I. -I../inst/include $(SHLIB_OPENMP_CXXFLAGS)
 PKG_LIBS = $(SHLIB_OPENMP_CXXFLAGS) $(LAPACK_LIBS) $(BLAS_LIBS) $(FLIBS)
 CXX_STD = CXX17
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/src/r_util.cpp 4.4.0-1/src/mlpack/bindings/R/mlpack/src/r_util.cpp
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/src/r_util.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/src/r_util.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -363,20 +363,6 @@ List IO_GetParamMatWithInfo(SEXP params,
                       Rcpp::Named("Data") = std::move(m));
 }
 
-// Enable verbose output.
-// [[Rcpp::export]]
-void EnableVerbose()
-{
-  Log::Info.ignoreInput = false;
-}
-
-// Disable verbose output.
-// [[Rcpp::export]]
-void DisableVerbose()
-{
-  Log::Info.ignoreInput = true;
-}
-
 // Reset the state of all timers.
 // [[Rcpp::export]]
 void ResetTimers()
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/src/rcpp_mlpack.h 4.4.0-1/src/mlpack/bindings/R/mlpack/src/rcpp_mlpack.h
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/src/rcpp_mlpack.h	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/src/rcpp_mlpack.h	2024-05-28 20:08:34.000000000 +0000
@@ -15,9 +15,6 @@
 
 #include <Rcpp.h>
 
-// To suppress Found '__assert_fail', possibly from 'assert' (C).
-#define BOOST_DISABLE_ASSERTS
-
 // Rcpp has its own stream object which cooperates more nicely with R's i/o
 // And as of armadillo and mlpack, we can use this stream object as well.
 #if !defined(ARMA_COUT_STREAM)
diff -pruN 4.3.0-2/src/mlpack/bindings/R/mlpack/tests/testthat/test-R_binding.R 4.4.0-1/src/mlpack/bindings/R/mlpack/tests/testthat/test-R_binding.R
--- 4.3.0-2/src/mlpack/bindings/R/mlpack/tests/testthat/test-R_binding.R	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/mlpack/tests/testthat/test-R_binding.R	2024-05-28 20:08:34.000000000 +0000
@@ -315,3 +315,45 @@ test_that("TestSerialization", {
 
   expect_true(output2$model_bw_out == 20)
 })
+
+# Make sure that the verbose argument does anything at all.
+test_that("TestVerbose", {
+  expect_output(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE,
+                               verbose=TRUE))
+})
+
+# Test that we get no output when verbose output is disabled.
+test_that("TestNotVerbose", {
+  expect_silent(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE))
+})
+
+# Test that we get no output when verbose output is explicitly disabled.
+test_that("TestReallyNotVerbose", {
+  expect_silent(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE,
+                               verbose=FALSE))
+})
+
+# Make sure that the mlpack verbose global option does anything at all.
+test_that("TestGlobalVerbose", {
+  options(mlpack.verbose = TRUE)
+  expect_output(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE))
+})
+
+# Test that we get no output when the global verbose option is set to false.
+test_that("TestGlobalNotVerbose", {
+  options(mlpack.verbose = FALSE)
+  expect_silent(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE))
+})
+
+# Test that we can override the global verbose option.
+test_that("TestGlobalVerboseOverride", {
+  options(mlpack.verbose = TRUE)
+  expect_silent(test_r_binding(4.0, 12, "hello",
+                               build_model=TRUE,
+                               verbose=FALSE))
+})
diff -pruN 4.3.0-2/src/mlpack/bindings/R/print_R.cpp 4.4.0-1/src/mlpack/bindings/R/print_R.cpp
--- 4.3.0-2/src/mlpack/bindings/R/print_R.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/print_R.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -184,21 +184,10 @@ void PrintR(util::Params& params,
        << endl;
   for (const string& opt : inputOptions)
   {
-    if (opt != "verbose")
-    {
-      util::ParamData& d = parameters.at(opt);
-      params.functionMap[d.tname]["PrintInputProcessing"](d, NULL, NULL);
-    }
+    util::ParamData& d = parameters.at(opt);
+    params.functionMap[d.tname]["PrintInputProcessing"](d, NULL, NULL);
   }
 
-  // Special handling for verbose output.
-  cout << "  if (verbose) {" << endl;
-  cout << "    EnableVerbose()" << endl;
-  cout << "  } else {" << endl;
-  cout << "    DisableVerbose()" << endl;
-  cout << "  }" << endl;
-  cout << endl;
-
   // Mark output parameters as passed.
   cout << "  # Mark all output options as passed." << endl;
   for (const string& opt : outputOptions)
diff -pruN 4.3.0-2/src/mlpack/bindings/R/print_doc.hpp 4.4.0-1/src/mlpack/bindings/R/print_doc.hpp
--- 4.3.0-2/src/mlpack/bindings/R/print_doc.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/print_doc.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -55,19 +55,28 @@ void PrintDoc(util::ParamData& d,
       oss << ".  Default value \"";
       if (d.cppType == "std::string")
       {
-        oss << MLPACK_ANY_CAST<std::string>(d.value);
+        oss << std::any_cast<std::string>(d.value);
       }
       else if (d.cppType == "double")
       {
-        oss << MLPACK_ANY_CAST<double>(d.value);
+        oss << std::any_cast<double>(d.value);
       }
       else if (d.cppType == "int")
       {
-        oss << MLPACK_ANY_CAST<int>(d.value);
+        oss << std::any_cast<int>(d.value);
       }
       else if (d.cppType == "bool")
       {
-        oss << (MLPACK_ANY_CAST<bool>(d.value) ? "TRUE" : "FALSE");
+        // If the option is `verbose`, be sure to print the use of the global
+        // mlpack package option as a default.
+        if (d.name == "verbose")
+        {
+          oss << "getOption(\"mlpack.verbose\", FALSE)";
+        }
+        else
+        {
+          oss << (std::any_cast<bool>(d.value) ? "TRUE" : "FALSE");
+        }
       }
       oss << "\"";
     }
diff -pruN 4.3.0-2/src/mlpack/bindings/R/print_input_param.hpp 4.4.0-1/src/mlpack/bindings/R/print_input_param.hpp
--- 4.3.0-2/src/mlpack/bindings/R/print_input_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/print_input_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,9 +30,22 @@ void PrintInputParam(util::ParamData& d,
 {
   MLPACK_COUT_STREAM << d.name;
   if (std::is_same<T, bool>::value)
-    MLPACK_COUT_STREAM << "=FALSE";
+  {
+    if (d.name == "verbose")
+    {
+      // Make sure that we use the global verbose option for the mlpack package
+      // as the default.
+      MLPACK_COUT_STREAM << "=getOption(\"mlpack.verbose\", FALSE)";
+    }
+    else
+    {
+      MLPACK_COUT_STREAM << "=FALSE";
+    }
+  }
   else if (!d.required)
+  {
     MLPACK_COUT_STREAM << "=NA";
+  }
 }
 
 } // namespace r
diff -pruN 4.3.0-2/src/mlpack/bindings/R/r_method.cpp.in 4.4.0-1/src/mlpack/bindings/R/r_method.cpp.in
--- 4.3.0-2/src/mlpack/bindings/R/r_method.cpp.in	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/r_method.cpp.in	2024-05-28 20:08:34.000000000 +0000
@@ -17,6 +17,11 @@ void ${PROGRAM_NAME}_call(SEXP params, S
   util::Params& p = *Rcpp::as<Rcpp::XPtr<util::Params>>(params);
   util::Timers& t = *Rcpp::as<Rcpp::XPtr<util::Timers>>(timers);
 
+  if (p.Has("verbose"))
+    Log::Info.ignoreInput = false;
+  else
+    Log::Info.ignoreInput = true;
+
   BINDING_FUNCTION(p, t);
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/R/tests/test_r_binding_main.cpp 4.4.0-1/src/mlpack/bindings/R/tests/test_r_binding_main.cpp
--- 4.3.0-2/src/mlpack/bindings/R/tests/test_r_binding_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/R/tests/test_r_binding_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -237,4 +237,7 @@ void BINDING_FUNCTION(util::Params& para
     params.Get<double>("model_bw_out") =
         params.Get<GaussianKernel*>("model_in")->Bandwidth() * 2.0;
   }
+
+  // Provide some output if the user asked for it.
+  Log::Info << "Here is some verbose output!" << std::endl;
 }
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/add_to_cli11.hpp 4.4.0-1/src/mlpack/bindings/cli/add_to_cli11.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/add_to_cli11.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/add_to_cli11.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,8 +47,8 @@ void AddToCLI11(const std::string& cliNa
       [&param](const std::string& value)
       {
         using TupleType = std::tuple<T, typename ParameterType<T>::type>;
-        TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&param.value);
-        std::get<0>(std::get<1>(tuple)) = MLPACK_ANY_CAST<std::string>(value);
+        TupleType& tuple = *std::any_cast<TupleType>(&param.value);
+        std::get<0>(std::get<1>(tuple)) = std::any_cast<std::string>(value);
         param.wasPassed = true;
       },
       param.desc.c_str());
@@ -79,8 +79,8 @@ void AddToCLI11(const std::string& cliNa
       [&param](const std::string& value)
       {
         using TupleType = std::tuple<T*, typename ParameterType<T>::type>;
-        TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&param.value);
-        std::get<1>(tuple) = MLPACK_ANY_CAST<std::string>(value);
+        TupleType& tuple = *std::any_cast<TupleType>(&param.value);
+        std::get<1>(tuple) = std::any_cast<std::string>(value);
         param.wasPassed = true;
       },
       param.desc.c_str());
@@ -109,8 +109,8 @@ void AddToCLI11(const std::string& cliNa
       [&param](const std::string& value)
       {
         using TupleType = std::tuple<T, typename ParameterType<T>::type>;
-        TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&param.value);
-        std::get<0>(std::get<1>(tuple)) = MLPACK_ANY_CAST<std::string>(value);
+        TupleType& tuple = *std::any_cast<TupleType>(&param.value);
+        std::get<0>(std::get<1>(tuple)) = std::any_cast<std::string>(value);
         param.wasPassed = true;
       },
       param.desc.c_str());
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/default_param_impl.hpp 4.4.0-1/src/mlpack/bindings/cli/default_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/default_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/default_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,7 +34,7 @@ std::string DefaultParamImpl(
 {
   std::ostringstream oss;
   if (!std::is_same<T, bool>::value)
-    oss << MLPACK_ANY_CAST<T>(data.value);
+    oss << std::any_cast<T>(data.value);
 
   return oss.str();
 }
@@ -49,7 +49,7 @@ std::string DefaultParamImpl(
 {
   // Print each element in an array delimited by square brackets.
   std::ostringstream oss;
-  const T& vector = MLPACK_ANY_CAST<T>(data.value);
+  const T& vector = std::any_cast<T>(data.value);
   oss << "[";
   if (std::is_same<T, std::vector<std::string>>::value)
   {
@@ -91,7 +91,7 @@ std::string DefaultParamImpl(
     util::ParamData& data,
     const typename std::enable_if<std::is_same<T, std::string>::value>::type*)
 {
-  const std::string& s = *MLPACK_ANY_CAST<std::string>(&data.value);
+  const std::string& s = *std::any_cast<std::string>(&data.value);
   return "'" + s + "'";
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/delete_allocated_memory.hpp 4.4.0-1/src/mlpack/bindings/cli/delete_allocated_memory.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/delete_allocated_memory.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/delete_allocated_memory.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -43,7 +43,7 @@ void DeleteAllocatedMemoryImpl(
 {
   // Delete the allocated memory (hopefully we actually own it).
   typedef std::tuple<T*, std::string> TupleType;
-  delete std::get<0>(*MLPACK_ANY_CAST<TupleType>(&d.value));
+  delete std::get<0>(*std::any_cast<TupleType>(&d.value));
 }
 
 template<typename T>
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/get_allocated_memory.hpp 4.4.0-1/src/mlpack/bindings/cli/get_allocated_memory.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/get_allocated_memory.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/get_allocated_memory.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,7 +45,7 @@ void* GetAllocatedMemory(
   // Here we have a model, which is a tuple, and we need the address of the
   // memory.
   typedef std::tuple<T*, std::string> TupleType;
-  return std::get<0>(*MLPACK_ANY_CAST<TupleType>(&d.value));
+  return std::get<0>(*std::any_cast<TupleType>(&d.value));
 }
 
 template<typename T>
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/get_param.hpp 4.4.0-1/src/mlpack/bindings/cli/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,7 +34,7 @@ T& GetParam(
         std::tuple<mlpack::data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // No mapping is needed, so just cast it directly.
-  return *MLPACK_ANY_CAST<T>(&d.value);
+  return *std::any_cast<T>(&d.value);
 }
 
 /**
@@ -52,7 +52,7 @@ T& GetParam(
   // times, but I am not bothered by that---it shouldn't be something that
   // happens.
   typedef std::tuple<T, typename ParameterType<T>::type> TupleType;
-  TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&d.value);
+  TupleType& tuple = *std::any_cast<TupleType>(&d.value);
   const std::string& value = std::get<0>(std::get<1>(tuple));
   T& matrix = std::get<0>(tuple);
   size_t& n_rows = std::get<1>(std::get<1>(tuple));
@@ -86,7 +86,7 @@ T& GetParam(
   // If this is an input parameter, we need to load both the matrix and the
   // dataset info.
   typedef std::tuple<T, std::tuple<std::string, size_t, size_t>> TupleType;
-  TupleType* tuple = MLPACK_ANY_CAST<TupleType>(&d.value);
+  TupleType* tuple = std::any_cast<TupleType>(&d.value);
   const std::string& value = std::get<0>(std::get<1>(*tuple));
   T& t = std::get<0>(*tuple);
   size_t& n_rows = std::get<1>(std::get<1>(*tuple));
@@ -116,7 +116,7 @@ T*& GetParam(
   // If the model is an input model, we have to load it from file.  'value'
   // contains the filename.
   typedef std::tuple<T*, std::string> TupleType;
-  TupleType* tuple = MLPACK_ANY_CAST<TupleType>(&d.value);
+  TupleType* tuple = std::any_cast<TupleType>(&d.value);
   const std::string& value = std::get<1>(*tuple);
   if (d.input && !d.loaded)
   {
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/get_printable_param_impl.hpp 4.4.0-1/src/mlpack/bindings/cli/get_printable_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/get_printable_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/get_printable_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,7 +30,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* /* junk */)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -41,7 +41,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<util::IsStdVector<T>::value>::type*
         /* junk */)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -80,7 +80,7 @@ std::string GetPrintableParam(
 {
   // Extract the string from the tuple that's being held.
   typedef std::tuple<T, typename ParameterType<T>::type> TupleType;
-  const TupleType* tuple = MLPACK_ANY_CAST<TupleType>(&data.value);
+  const TupleType* tuple = std::any_cast<TupleType>(&data.value);
 
   std::ostringstream oss;
   oss << "'" << std::get<0>(std::get<1>(*tuple)) << "'";
@@ -108,7 +108,7 @@ std::string GetPrintableParam(
 {
   // Extract the string from the tuple that's being held.
   typedef std::tuple<T*, typename ParameterType<T>::type> TupleType;
-  const TupleType* tuple = MLPACK_ANY_CAST<TupleType>(&data.value);
+  const TupleType* tuple = std::any_cast<TupleType>(&data.value);
 
   std::ostringstream oss;
   oss << std::get<1>(*tuple);
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/get_raw_param.hpp 4.4.0-1/src/mlpack/bindings/cli/get_raw_param.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/get_raw_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/get_raw_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -33,7 +33,7 @@ T& GetRawParam(
         std::tuple<mlpack::data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // No mapping is needed, so just cast it directly.
-  return *MLPACK_ANY_CAST<T>(&d.value);
+  return *std::any_cast<T>(&d.value);
 }
 
 /**
@@ -49,7 +49,7 @@ T& GetRawParam(
 {
   // Don't load the matrix.
   typedef std::tuple<T, std::tuple<std::string, size_t, size_t>> TupleType;
-  T& value = std::get<0>(*MLPACK_ANY_CAST<TupleType>(&d.value));
+  T& value = std::get<0>(*std::any_cast<TupleType>(&d.value));
   return value;
 }
 
@@ -64,7 +64,7 @@ T*& GetRawParam(
 {
   // Don't load the model.
   typedef std::tuple<T*, std::string> TupleType;
-  T*& value = std::get<0>(*MLPACK_ANY_CAST<TupleType>(&d.value));
+  T*& value = std::get<0>(*std::any_cast<TupleType>(&d.value));
   return value;
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/in_place_copy.hpp 4.4.0-1/src/mlpack/bindings/cli/in_place_copy.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/in_place_copy.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/in_place_copy.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -58,10 +58,10 @@ void InPlaceCopyInternal(
 {
   // Make the output filename the same as the input filename.
   typedef std::tuple<T, typename ParameterType<T>::type> TupleType;
-  TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&d.value);
+  TupleType& tuple = *std::any_cast<TupleType>(&d.value);
   std::string& value = std::get<0>(std::get<1>(tuple));
 
-  const TupleType& inputTuple = *MLPACK_ANY_CAST<TupleType>(&input.value);
+  const TupleType& inputTuple = *std::any_cast<TupleType>(&input.value);
   value = std::get<0>(std::get<1>(inputTuple));
 }
 
@@ -81,10 +81,10 @@ void InPlaceCopyInternal(
 {
   // Make the output filename the same as the input filename.
   typedef std::tuple<T*, typename ParameterType<T>::type> TupleType;
-  TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&d.value);
+  TupleType& tuple = *std::any_cast<TupleType>(&d.value);
   std::string& value = std::get<1>(tuple);
 
-  const TupleType& inputTuple = *MLPACK_ANY_CAST<TupleType>(&input.value);
+  const TupleType& inputTuple = *std::any_cast<TupleType>(&input.value);
   value = std::get<1>(inputTuple);
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/output_param_impl.hpp 4.4.0-1/src/mlpack/bindings/cli/output_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/output_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/output_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,7 +30,7 @@ void OutputParamImpl(
     const typename std::enable_if<!std::is_same<T,
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* /* junk */)
 {
-  std::cout << data.name << ": " << *MLPACK_ANY_CAST<T>(&data.value)
+  std::cout << data.name << ": " << *std::any_cast<T>(&data.value)
       << std::endl;
 }
 
@@ -41,7 +41,7 @@ void OutputParamImpl(
     const typename std::enable_if<util::IsStdVector<T>::value>::type* /* junk */)
 {
   std::cout << data.name << ": ";
-  const T& t = *MLPACK_ANY_CAST<T>(&data.value);
+  const T& t = *std::any_cast<T>(&data.value);
   for (size_t i = 0; i < t.size(); ++i)
     std::cout << t[i] << " ";
   std::cout << std::endl;
@@ -54,9 +54,9 @@ void OutputParamImpl(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* /* junk */)
 {
   typedef std::tuple<T, std::tuple<std::string, size_t, size_t>> TupleType;
-  const T& output = std::get<0>(*MLPACK_ANY_CAST<TupleType>(&data.value));
+  const T& output = std::get<0>(*std::any_cast<TupleType>(&data.value));
   const std::string& filename =
-      std::get<0>(std::get<1>(*MLPACK_ANY_CAST<TupleType>(&data.value)));
+      std::get<0>(std::get<1>(*std::any_cast<TupleType>(&data.value)));
 
   if (output.n_elem > 0 && filename != "")
   {
@@ -78,10 +78,10 @@ void OutputParamImpl(
   // const.  In this case we can assume it though, since we will be saving and
   // not loading.
   typedef std::tuple<T*, std::string> TupleType;
-  T*& output = const_cast<T*&>(std::get<0>(*MLPACK_ANY_CAST<TupleType>(
+  T*& output = const_cast<T*&>(std::get<0>(*std::any_cast<TupleType>(
       &data.value)));
   const std::string& filename =
-      std::get<1>(*MLPACK_ANY_CAST<TupleType>(&data.value));
+      std::get<1>(*std::any_cast<TupleType>(&data.value));
 
   if (filename != "")
     data::Save(filename, "model", *output);
@@ -96,9 +96,9 @@ void OutputParamImpl(
 {
   // Output the matrix with the mappings.
   typedef std::tuple<T, std::tuple<std::string, size_t, size_t>> TupleType;
-  const T& tuple = std::get<0>(*MLPACK_ANY_CAST<TupleType>(&data.value));
+  const T& tuple = std::get<0>(*std::any_cast<TupleType>(&data.value));
   const std::string& filename =
-      std::get<0>(std::get<1>(*MLPACK_ANY_CAST<TupleType>(&data.value)));
+      std::get<0>(std::get<1>(*std::any_cast<TupleType>(&data.value)));
   const arma::mat& matrix = std::get<1>(tuple);
 
   // The mapping isn't taken into account.  We should write a data::Save()
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/print_help_impl.hpp 4.4.0-1/src/mlpack/bindings/cli/print_help_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/print_help_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/print_help_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -104,11 +104,6 @@ inline void PrintHelp(util::Params& para
       if ((pass == 2) && data.input) // Output options only (always optional).
         continue;
 
-      // For reverse compatibility: this can be removed when these options are
-      // gone in mlpack 3.0.0.  We don't want to print the deprecated options.
-      if (data.name == "inputFile")
-        continue;
-
       if (!printedHeader)
       {
         printedHeader = true;
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/set_param.hpp 4.4.0-1/src/mlpack/bindings/cli/set_param.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/set_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/set_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ namespace cli {
 template<typename T>
 void SetParam(
     util::ParamData& d,
-    const MLPACK_ANY& value,
+    const std::any& value,
     const typename std::enable_if<!arma::is_arma_type<T>::value>::type* = 0,
     const typename std::enable_if<!data::HasSerialize<T>::value>::type* = 0,
     const typename std::enable_if<!std::is_same<T,
@@ -34,7 +34,7 @@ void SetParam(
     const typename std::enable_if<!std::is_same<T, bool>::value>::type* = 0)
 {
   // No mapping is needed.
-  d.value = *MLPACK_ANY_CAST<T>(&value);
+  d.value = *std::any_cast<T>(&value);
 }
 
 /**
@@ -43,7 +43,7 @@ void SetParam(
 template<typename T>
 void SetParam(
     util::ParamData& d,
-    const MLPACK_ANY& /* value */,
+    const std::any& /* value */,
     const typename std::enable_if<std::is_same<T, bool>::value>::type* = 0)
 {
   // Force set to the value of whether or not this was passed.
@@ -57,15 +57,15 @@ void SetParam(
 template<typename T>
 void SetParam(
     util::ParamData& d,
-    const MLPACK_ANY& value,
+    const std::any& value,
     const typename std::enable_if<arma::is_arma_type<T>::value ||
                                   std::is_same<T,
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // We're setting the string filename.
   typedef std::tuple<T, typename ParameterType<T>::type> TupleType;
-  TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&d.value);
-  std::get<0>(std::get<1>(tuple)) = MLPACK_ANY_CAST<std::string>(value);
+  TupleType& tuple = *std::any_cast<TupleType>(&d.value);
+  std::get<0>(std::get<1>(tuple)) = std::any_cast<std::string>(value);
 }
 
 /**
@@ -75,14 +75,14 @@ void SetParam(
 template<typename T>
 void SetParam(
     util::ParamData& d,
-    const MLPACK_ANY& value,
+    const std::any& value,
     const typename std::enable_if<!arma::is_arma_type<T>::value>::type* = 0,
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   // We're setting the string filename.
   typedef std::tuple<T*, typename ParameterType<T>::type> TupleType;
-  TupleType& tuple = *MLPACK_ANY_CAST<TupleType>(&d.value);
-  std::get<1>(tuple) = MLPACK_ANY_CAST<std::string>(value);
+  TupleType& tuple = *std::any_cast<TupleType>(&d.value);
+  std::get<1>(tuple) = std::any_cast<std::string>(value);
 }
 
 /**
@@ -97,7 +97,7 @@ template<typename T>
 void SetParam(util::ParamData& d, const void* input, void* /* output */)
 {
   SetParam<typename std::remove_pointer<T>::type>(
-      const_cast<util::ParamData&>(d), *((MLPACK_ANY*) input));
+      const_cast<util::ParamData&>(d), *((std::any*) input));
 }
 
 } // namespace cli
diff -pruN 4.3.0-2/src/mlpack/bindings/cli/third_party/CLI/CLI11.hpp 4.4.0-1/src/mlpack/bindings/cli/third_party/CLI/CLI11.hpp
--- 4.3.0-2/src/mlpack/bindings/cli/third_party/CLI/CLI11.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/cli/third_party/CLI/CLI11.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -1,11 +1,11 @@
-// CLI11: Version 2.3.2
+// CLI11: Version 2.4.2
 // Originally designed by Henry Schreiner
 // https://github.com/CLIUtils/CLI11
 //
 // This is a standalone header file generated by MakeSingleHeader.py in CLI11/scripts
-// from: v2.3.2
+// from: v2.4.2
 //
-// CLI11 2.3.2 Copyright (c) 2017-2022 University of Cincinnati, developed by Henry
+// CLI11 2.4.2 Copyright (c) 2017-2024 University of Cincinnati, developed by Henry
 // Schreiner under NSF AWARD 1414736. All rights reserved.
 //
 // Redistribution and use in source and binary forms of CLI11, with or without
@@ -34,34 +34,40 @@
 #pragma once
 
 // Standard combined includes:
-#include <iomanip>
-#include <set>
-#include <memory>
-#include <vector>
-#include <utility>
-#include <stdexcept>
-#include <locale>
+#include <algorithm>
+#include <array>
+#include <cctype>
+#include <clocale>
+#include <cmath>
+#include <cstdint>
+#include <cstdlib>
+#include <cstring>
+#include <cwchar>
+#include <exception>
+#include <fstream>
 #include <functional>
+#include <iomanip>
 #include <iostream>
 #include <iterator>
-#include <exception>
+#include <limits>
+#include <locale>
+#include <map>
+#include <memory>
 #include <numeric>
-#include <fstream>
+#include <set>
+#include <sstream>
+#include <stdexcept>
 #include <string>
-#include <type_traits>
 #include <tuple>
-#include <map>
-#include <algorithm>
-#include <cstdint>
-#include <sstream>
-#include <cmath>
-#include <limits>
+#include <type_traits>
+#include <utility>
+#include <vector>
 
 
 #define CLI11_VERSION_MAJOR 2
-#define CLI11_VERSION_MINOR 3
+#define CLI11_VERSION_MINOR 4
 #define CLI11_VERSION_PATCH 2
-#define CLI11_VERSION "2.3.2"
+#define CLI11_VERSION "2.4.2"
 
 
 
@@ -124,17 +130,7 @@
 #endif
 #endif
 
-/** Inline macro **/
-#ifdef CLI11_COMPILE
-#define CLI11_INLINE
-#else
-#define CLI11_INLINE inline
-#endif
-
-
-
-// C standard library
-// Only needed for existence checking
+/** <filesystem> availability */
 #if defined CLI11_CPP17 && defined __has_include && !defined CLI11_HAS_FILESYSTEM
 #if __has_include(<filesystem>)
 // Filesystem cannot be used if targeting macOS < 10.15
@@ -161,6 +157,44 @@
 #endif
 #endif
 
+/** <codecvt> availability */
+#if defined(__GNUC__) && !defined(__llvm__) && !defined(__INTEL_COMPILER) && __GNUC__ < 5
+#define CLI11_HAS_CODECVT 0
+#else
+#define CLI11_HAS_CODECVT 1
+#include <codecvt>
+#endif
+
+/** disable deprecations */
+#if defined(__GNUC__)  // GCC or clang
+#define CLI11_DIAGNOSTIC_PUSH _Pragma("GCC diagnostic push")
+#define CLI11_DIAGNOSTIC_POP _Pragma("GCC diagnostic pop")
+
+#define CLI11_DIAGNOSTIC_IGNORE_DEPRECATED _Pragma("GCC diagnostic ignored \"-Wdeprecated-declarations\"")
+
+#elif defined(_MSC_VER)
+#define CLI11_DIAGNOSTIC_PUSH __pragma(warning(push))
+#define CLI11_DIAGNOSTIC_POP __pragma(warning(pop))
+
+#define CLI11_DIAGNOSTIC_IGNORE_DEPRECATED __pragma(warning(disable : 4996))
+
+#else
+#define CLI11_DIAGNOSTIC_PUSH
+#define CLI11_DIAGNOSTIC_POP
+
+#define CLI11_DIAGNOSTIC_IGNORE_DEPRECATED
+
+#endif
+
+/** Inline macro **/
+#ifdef CLI11_COMPILE
+#define CLI11_INLINE
+#else
+#define CLI11_INLINE inline
+#endif
+
+
+
 #if defined CLI11_HAS_FILESYSTEM && CLI11_HAS_FILESYSTEM > 0
 #include <filesystem>  // NOLINT(build/include)
 #else
@@ -170,9 +204,244 @@
 
 
 
+
+#ifdef CLI11_CPP17
+#include <string_view>
+#endif  // CLI11_CPP17
+
+#if defined CLI11_HAS_FILESYSTEM && CLI11_HAS_FILESYSTEM > 0
+#include <filesystem>
+#include <string_view>  // NOLINT(build/include)
+#endif                  // CLI11_HAS_FILESYSTEM
+
+
+
+#if defined(_WIN32)
+#if !(defined(_AMD64_) || defined(_X86_) || defined(_ARM_))
+#if defined(__amd64__) || defined(__amd64) || defined(__x86_64__) || defined(__x86_64) || defined(_M_X64) ||           \
+    defined(_M_AMD64)
+#define _AMD64_
+#elif defined(i386) || defined(__i386) || defined(__i386__) || defined(__i386__) || defined(_M_IX86)
+#define _X86_
+#elif defined(__arm__) || defined(_M_ARM) || defined(_M_ARMT)
+#define _ARM_
+#elif defined(__aarch64__) || defined(_M_ARM64)
+#define _ARM64_
+#elif defined(_M_ARM64EC)
+#define _ARM64EC_
+#endif
+#endif
+
+// first
+#ifndef NOMINMAX
+// if NOMINMAX is already defined we don't want to mess with that either way
+#define NOMINMAX
+#include <windef.h>
+#undef NOMINMAX
+#else
+#include <windef.h>
+#endif
+
+// second
+#include <winbase.h>
+// third
+#include <processthreadsapi.h>
+#include <shellapi.h>
+#endif
+
+
 namespace CLI {
 
 
+/// Convert a wide string to a narrow string.
+CLI11_INLINE std::string narrow(const std::wstring &str);
+CLI11_INLINE std::string narrow(const wchar_t *str);
+CLI11_INLINE std::string narrow(const wchar_t *str, std::size_t size);
+
+/// Convert a narrow string to a wide string.
+CLI11_INLINE std::wstring widen(const std::string &str);
+CLI11_INLINE std::wstring widen(const char *str);
+CLI11_INLINE std::wstring widen(const char *str, std::size_t size);
+
+#ifdef CLI11_CPP17
+CLI11_INLINE std::string narrow(std::wstring_view str);
+CLI11_INLINE std::wstring widen(std::string_view str);
+#endif  // CLI11_CPP17
+
+#if defined CLI11_HAS_FILESYSTEM && CLI11_HAS_FILESYSTEM > 0
+/// Convert a char-string to a native path correctly.
+CLI11_INLINE std::filesystem::path to_path(std::string_view str);
+#endif  // CLI11_HAS_FILESYSTEM
+
+
+
+
+namespace detail {
+
+#if !CLI11_HAS_CODECVT
+/// Attempt to set one of the acceptable unicode locales for conversion
+CLI11_INLINE void set_unicode_locale() {
+    static const std::array<const char *, 3> unicode_locales{{"C.UTF-8", "en_US.UTF-8", ".UTF-8"}};
+
+    for(const auto &locale_name : unicode_locales) {
+        if(std::setlocale(LC_ALL, locale_name) != nullptr) {
+            return;
+        }
+    }
+    throw std::runtime_error("CLI::narrow: could not set locale to C.UTF-8");
+}
+
+template <typename F> struct scope_guard_t {
+    F closure;
+
+    explicit scope_guard_t(F closure_) : closure(closure_) {}
+    ~scope_guard_t() { closure(); }
+};
+
+template <typename F> CLI11_NODISCARD CLI11_INLINE scope_guard_t<F> scope_guard(F &&closure) {
+    return scope_guard_t<F>{std::forward<F>(closure)};
+}
+
+#endif  // !CLI11_HAS_CODECVT
+
+CLI11_DIAGNOSTIC_PUSH
+CLI11_DIAGNOSTIC_IGNORE_DEPRECATED
+
+CLI11_INLINE std::string narrow_impl(const wchar_t *str, std::size_t str_size) {
+#if CLI11_HAS_CODECVT
+#ifdef _WIN32
+    return std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>>().to_bytes(str, str + str_size);
+
+#else
+    return std::wstring_convert<std::codecvt_utf8<wchar_t>>().to_bytes(str, str + str_size);
+
+#endif  // _WIN32
+#else   // CLI11_HAS_CODECVT
+    (void)str_size;
+    std::mbstate_t state = std::mbstate_t();
+    const wchar_t *it = str;
+
+    std::string old_locale = std::setlocale(LC_ALL, nullptr);
+    auto sg = scope_guard([&] { std::setlocale(LC_ALL, old_locale.c_str()); });
+    set_unicode_locale();
+
+    std::size_t new_size = std::wcsrtombs(nullptr, &it, 0, &state);
+    if(new_size == static_cast<std::size_t>(-1)) {
+        throw std::runtime_error("CLI::narrow: conversion error in std::wcsrtombs at offset " +
+                                 std::to_string(it - str));
+    }
+    std::string result(new_size, '\0');
+    std::wcsrtombs(const_cast<char *>(result.data()), &str, new_size, &state);
+
+    return result;
+
+#endif  // CLI11_HAS_CODECVT
+}
+
+CLI11_INLINE std::wstring widen_impl(const char *str, std::size_t str_size) {
+#if CLI11_HAS_CODECVT
+#ifdef _WIN32
+    return std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>>().from_bytes(str, str + str_size);
+
+#else
+    return std::wstring_convert<std::codecvt_utf8<wchar_t>>().from_bytes(str, str + str_size);
+
+#endif  // _WIN32
+#else   // CLI11_HAS_CODECVT
+    (void)str_size;
+    std::mbstate_t state = std::mbstate_t();
+    const char *it = str;
+
+    std::string old_locale = std::setlocale(LC_ALL, nullptr);
+    auto sg = scope_guard([&] { std::setlocale(LC_ALL, old_locale.c_str()); });
+    set_unicode_locale();
+
+    std::size_t new_size = std::mbsrtowcs(nullptr, &it, 0, &state);
+    if(new_size == static_cast<std::size_t>(-1)) {
+        throw std::runtime_error("CLI::widen: conversion error in std::mbsrtowcs at offset " +
+                                 std::to_string(it - str));
+    }
+    std::wstring result(new_size, L'\0');
+    std::mbsrtowcs(const_cast<wchar_t *>(result.data()), &str, new_size, &state);
+
+    return result;
+
+#endif  // CLI11_HAS_CODECVT
+}
+
+CLI11_DIAGNOSTIC_POP
+
+}  // namespace detail
+
+CLI11_INLINE std::string narrow(const wchar_t *str, std::size_t str_size) { return detail::narrow_impl(str, str_size); }
+CLI11_INLINE std::string narrow(const std::wstring &str) { return detail::narrow_impl(str.data(), str.size()); }
+// Flawfinder: ignore
+CLI11_INLINE std::string narrow(const wchar_t *str) { return detail::narrow_impl(str, std::wcslen(str)); }
+
+CLI11_INLINE std::wstring widen(const char *str, std::size_t str_size) { return detail::widen_impl(str, str_size); }
+CLI11_INLINE std::wstring widen(const std::string &str) { return detail::widen_impl(str.data(), str.size()); }
+// Flawfinder: ignore
+CLI11_INLINE std::wstring widen(const char *str) { return detail::widen_impl(str, std::strlen(str)); }
+
+#ifdef CLI11_CPP17
+CLI11_INLINE std::string narrow(std::wstring_view str) { return detail::narrow_impl(str.data(), str.size()); }
+CLI11_INLINE std::wstring widen(std::string_view str) { return detail::widen_impl(str.data(), str.size()); }
+#endif  // CLI11_CPP17
+
+#if defined CLI11_HAS_FILESYSTEM && CLI11_HAS_FILESYSTEM > 0
+CLI11_INLINE std::filesystem::path to_path(std::string_view str) {
+    return std::filesystem::path{
+#ifdef _WIN32
+        widen(str)
+#else
+        str
+#endif  // _WIN32
+    };
+}
+#endif  // CLI11_HAS_FILESYSTEM
+
+
+
+
+namespace detail {
+#ifdef _WIN32
+/// Decode and return UTF-8 argv from GetCommandLineW.
+CLI11_INLINE std::vector<std::string> compute_win32_argv();
+#endif
+}  // namespace detail
+
+
+
+namespace detail {
+
+#ifdef _WIN32
+CLI11_INLINE std::vector<std::string> compute_win32_argv() {
+    std::vector<std::string> result;
+    int argc = 0;
+
+    auto deleter = [](wchar_t **ptr) { LocalFree(ptr); };
+    // NOLINTBEGIN(*-avoid-c-arrays)
+    auto wargv = std::unique_ptr<wchar_t *[], decltype(deleter)>(CommandLineToArgvW(GetCommandLineW(), &argc), deleter);
+    // NOLINTEND(*-avoid-c-arrays)
+
+    if(wargv == nullptr) {
+        throw std::runtime_error("CommandLineToArgvW failed with code " + std::to_string(GetLastError()));
+    }
+
+    result.reserve(static_cast<size_t>(argc));
+    for(size_t i = 0; i < static_cast<size_t>(argc); ++i) {
+        result.push_back(narrow(wargv[i]));
+    }
+
+    return result;
+}
+#endif
+
+}  // namespace detail
+
+
+
+
 /// Include the items in this namespace to get free conversion of enums to/from streams.
 /// (This is available inside CLI as well, so CLI11 will use this without a using statement).
 namespace enums {
@@ -270,6 +539,9 @@ inline std::string trim_copy(const std::
 /// remove quotes at the front and back of a string either '"' or '\''
 CLI11_INLINE std::string &remove_quotes(std::string &str);
 
+/// remove quotes from all elements of a string vector and process escaped components
+CLI11_INLINE void remove_quotes(std::vector<std::string> &args);
+
 /// Add a leader to the beginning of all new lines (nothing is added
 /// at the start of the first line). `"; "` would be for ini files
 ///
@@ -290,14 +562,16 @@ CLI11_INLINE std::ostream &format_aliase
 
 /// Verify the first character of an option
 /// - is a trigger character, ! has special meaning and new lines would just be annoying to deal with
-template <typename T> bool valid_first_char(T c) { return ((c != '-') && (c != '!') && (c != ' ') && c != '\n'); }
+template <typename T> bool valid_first_char(T c) {
+    return ((c != '-') && (static_cast<unsigned char>(c) > 33));  // space and '!' not allowed
+}
 
 /// Verify following characters of an option
 template <typename T> bool valid_later_char(T c) {
     // = and : are value separators, { has special meaning for option defaults,
-    // and \n would just be annoying to deal with in many places allowing space here has too much potential for
-    // inadvertent entry errors and bugs
-    return ((c != '=') && (c != ':') && (c != '{') && (c != ' ') && c != '\n');
+    // and control codes other than tab would just be annoying to deal with in many places allowing space here has too
+    // much potential for inadvertent entry errors and bugs
+    return ((c != '=') && (c != ':') && (c != '{') && ((static_cast<unsigned char>(c) > 32) || c == '\t'));
 }
 
 /// Verify an option/subcommand name
@@ -360,18 +634,46 @@ template <typename Callable> inline std:
     return str;
 }
 
+/// close a sequence of characters indicated by a closure character.  Brackets allows sub sequences
+/// recognized bracket sequences include "'`[(<{  other closure characters are assumed to be literal strings
+CLI11_INLINE std::size_t close_sequence(const std::string &str, std::size_t start, char closure_char);
+
 /// Split a string '"one two" "three"' into 'one two', 'three'
-/// Quote characters can be ` ' or "
+/// Quote characters can be ` ' or " or bracket characters [{(< with matching to the matching bracket
 CLI11_INLINE std::vector<std::string> split_up(std::string str, char delimiter = '\0');
 
+/// get the value of an environmental variable or empty string if empty
+CLI11_INLINE std::string get_environment_value(const std::string &env_name);
+
 /// This function detects an equal or colon followed by an escaped quote after an argument
 /// then modifies the string to replace the equality with a space.  This is needed
 /// to allow the split up function to work properly and is intended to be used with the find_and_modify function
 /// the return value is the offset+1 which is required by the find_and_modify function.
 CLI11_INLINE std::size_t escape_detect(std::string &str, std::size_t offset);
 
-/// Add quotes if the string contains spaces
-CLI11_INLINE std::string &add_quotes_if_needed(std::string &str);
+/// @brief  detect if a string has escapable characters
+/// @param str the string to do the detection on
+/// @return true if the string has escapable characters
+CLI11_INLINE bool has_escapable_character(const std::string &str);
+
+/// @brief escape all escapable characters
+/// @param str the string to escape
+/// @return a string with the escapble characters escaped with '\'
+CLI11_INLINE std::string add_escaped_characters(const std::string &str);
+
+/// @brief replace the escaped characters with their equivalent
+CLI11_INLINE std::string remove_escaped_characters(const std::string &str);
+
+/// generate a string with all non printable characters escaped to hex codes
+CLI11_INLINE std::string binary_escape_string(const std::string &string_to_escape);
+
+CLI11_INLINE bool is_binary_escaped_string(const std::string &escaped_string);
+
+/// extract an escaped binary_string
+CLI11_INLINE std::string extract_binary_string(const std::string &escaped_string);
+
+/// process a quoted string, remove the quotes and if appropriate handle escaped characters
+CLI11_INLINE bool process_quoted_string(std::string &str, char string_char = '\"', char literal_char = '\'');
 
 }  // namespace detail
 
@@ -421,7 +723,17 @@ CLI11_INLINE std::string &rtrim(std::str
 }
 
 CLI11_INLINE std::string &remove_quotes(std::string &str) {
-    if(str.length() > 1 && (str.front() == '"' || str.front() == '\'')) {
+    if(str.length() > 1 && (str.front() == '"' || str.front() == '\'' || str.front() == '`')) {
+        if(str.front() == str.back()) {
+            str.pop_back();
+            str.erase(str.begin(), str.begin() + 1);
+        }
+    }
+    return str;
+}
+
+CLI11_INLINE std::string &remove_outer(std::string &str, char key) {
+    if(str.length() > 1 && (str.front() == key)) {
         if(str.front() == str.back()) {
             str.pop_back();
             str.erase(str.begin(), str.begin() + 1);
@@ -541,37 +853,220 @@ find_member(std::string name, const std:
     return (it != std::end(names)) ? (it - std::begin(names)) : (-1);
 }
 
+static const std::string escapedChars("\b\t\n\f\r\"\\");
+static const std::string escapedCharsCode("btnfr\"\\");
+static const std::string bracketChars{"\"'`[(<{"};
+static const std::string matchBracketChars("\"'`])>}");
+
+CLI11_INLINE bool has_escapable_character(const std::string &str) {
+    return (str.find_first_of(escapedChars) != std::string::npos);
+}
+
+CLI11_INLINE std::string add_escaped_characters(const std::string &str) {
+    std::string out;
+    out.reserve(str.size() + 4);
+    for(char s : str) {
+        auto sloc = escapedChars.find_first_of(s);
+        if(sloc != std::string::npos) {
+            out.push_back('\\');
+            out.push_back(escapedCharsCode[sloc]);
+        } else {
+            out.push_back(s);
+        }
+    }
+    return out;
+}
+
+CLI11_INLINE std::uint32_t hexConvert(char hc) {
+    int hcode{0};
+    if(hc >= '0' && hc <= '9') {
+        hcode = (hc - '0');
+    } else if(hc >= 'A' && hc <= 'F') {
+        hcode = (hc - 'A' + 10);
+    } else if(hc >= 'a' && hc <= 'f') {
+        hcode = (hc - 'a' + 10);
+    } else {
+        hcode = -1;
+    }
+    return static_cast<uint32_t>(hcode);
+}
+
+CLI11_INLINE char make_char(std::uint32_t code) { return static_cast<char>(static_cast<unsigned char>(code)); }
+
+CLI11_INLINE void append_codepoint(std::string &str, std::uint32_t code) {
+    if(code < 0x80) {  // ascii code equivalent
+        str.push_back(static_cast<char>(code));
+    } else if(code < 0x800) {  // \u0080 to \u07FF
+        // 110yyyyx 10xxxxxx; 0x3f == 0b0011'1111
+        str.push_back(make_char(0xC0 | code >> 6));
+        str.push_back(make_char(0x80 | (code & 0x3F)));
+    } else if(code < 0x10000) {  // U+0800...U+FFFF
+        if(0xD800 <= code && code <= 0xDFFF) {
+            throw std::invalid_argument("[0xD800, 0xDFFF] are not valid UTF-8.");
+        }
+        // 1110yyyy 10yxxxxx 10xxxxxx
+        str.push_back(make_char(0xE0 | code >> 12));
+        str.push_back(make_char(0x80 | (code >> 6 & 0x3F)));
+        str.push_back(make_char(0x80 | (code & 0x3F)));
+    } else if(code < 0x110000) {  // U+010000 ... U+10FFFF
+        // 11110yyy 10yyxxxx 10xxxxxx 10xxxxxx
+        str.push_back(make_char(0xF0 | code >> 18));
+        str.push_back(make_char(0x80 | (code >> 12 & 0x3F)));
+        str.push_back(make_char(0x80 | (code >> 6 & 0x3F)));
+        str.push_back(make_char(0x80 | (code & 0x3F)));
+    }
+}
+
+CLI11_INLINE std::string remove_escaped_characters(const std::string &str) {
+
+    std::string out;
+    out.reserve(str.size());
+    for(auto loc = str.begin(); loc < str.end(); ++loc) {
+        if(*loc == '\\') {
+            if(str.end() - loc < 2) {
+                throw std::invalid_argument("invalid escape sequence " + str);
+            }
+            auto ecloc = escapedCharsCode.find_first_of(*(loc + 1));
+            if(ecloc != std::string::npos) {
+                out.push_back(escapedChars[ecloc]);
+                ++loc;
+            } else if(*(loc + 1) == 'u') {
+                // must have 4 hex characters
+                if(str.end() - loc < 6) {
+                    throw std::invalid_argument("unicode sequence must have 4 hex codes " + str);
+                }
+                std::uint32_t code{0};
+                std::uint32_t mplier{16 * 16 * 16};
+                for(int ii = 2; ii < 6; ++ii) {
+                    std::uint32_t res = hexConvert(*(loc + ii));
+                    if(res > 0x0F) {
+                        throw std::invalid_argument("unicode sequence must have 4 hex codes " + str);
+                    }
+                    code += res * mplier;
+                    mplier = mplier / 16;
+                }
+                append_codepoint(out, code);
+                loc += 5;
+            } else if(*(loc + 1) == 'U') {
+                // must have 8 hex characters
+                if(str.end() - loc < 10) {
+                    throw std::invalid_argument("unicode sequence must have 8 hex codes " + str);
+                }
+                std::uint32_t code{0};
+                std::uint32_t mplier{16 * 16 * 16 * 16 * 16 * 16 * 16};
+                for(int ii = 2; ii < 10; ++ii) {
+                    std::uint32_t res = hexConvert(*(loc + ii));
+                    if(res > 0x0F) {
+                        throw std::invalid_argument("unicode sequence must have 8 hex codes " + str);
+                    }
+                    code += res * mplier;
+                    mplier = mplier / 16;
+                }
+                append_codepoint(out, code);
+                loc += 9;
+            } else if(*(loc + 1) == '0') {
+                out.push_back('\0');
+                ++loc;
+            } else {
+                throw std::invalid_argument(std::string("unrecognized escape sequence \\") + *(loc + 1) + " in " + str);
+            }
+        } else {
+            out.push_back(*loc);
+        }
+    }
+    return out;
+}
+
+CLI11_INLINE std::size_t close_string_quote(const std::string &str, std::size_t start, char closure_char) {
+    std::size_t loc{0};
+    for(loc = start + 1; loc < str.size(); ++loc) {
+        if(str[loc] == closure_char) {
+            break;
+        }
+        if(str[loc] == '\\') {
+            // skip the next character for escaped sequences
+            ++loc;
+        }
+    }
+    return loc;
+}
+
+CLI11_INLINE std::size_t close_literal_quote(const std::string &str, std::size_t start, char closure_char) {
+    auto loc = str.find_first_of(closure_char, start + 1);
+    return (loc != std::string::npos ? loc : str.size());
+}
+
+CLI11_INLINE std::size_t close_sequence(const std::string &str, std::size_t start, char closure_char) {
+
+    auto bracket_loc = matchBracketChars.find(closure_char);
+    switch(bracket_loc) {
+    case 0:
+        return close_string_quote(str, start, closure_char);
+    case 1:
+    case 2:
+    case std::string::npos:
+        return close_literal_quote(str, start, closure_char);
+    default:
+        break;
+    }
+
+    std::string closures(1, closure_char);
+    auto loc = start + 1;
+
+    while(loc < str.size()) {
+        if(str[loc] == closures.back()) {
+            closures.pop_back();
+            if(closures.empty()) {
+                return loc;
+            }
+        }
+        bracket_loc = bracketChars.find(str[loc]);
+        if(bracket_loc != std::string::npos) {
+            switch(bracket_loc) {
+            case 0:
+                loc = close_string_quote(str, loc, str[loc]);
+                break;
+            case 1:
+            case 2:
+                loc = close_literal_quote(str, loc, str[loc]);
+                break;
+            default:
+                closures.push_back(matchBracketChars[bracket_loc]);
+                break;
+            }
+        }
+        ++loc;
+    }
+    if(loc > str.size()) {
+        loc = str.size();
+    }
+    return loc;
+}
+
 CLI11_INLINE std::vector<std::string> split_up(std::string str, char delimiter) {
 
-    const std::string delims("\'\"`");
     auto find_ws = [delimiter](char ch) {
         return (delimiter == '\0') ? std::isspace<char>(ch, std::locale()) : (ch == delimiter);
     };
     trim(str);
 
     std::vector<std::string> output;
-    bool embeddedQuote = false;
-    char keyChar = ' ';
     while(!str.empty()) {
-        if(delims.find_first_of(str[0]) != std::string::npos) {
-            keyChar = str[0];
-            auto end = str.find_first_of(keyChar, 1);
-            while((end != std::string::npos) && (str[end - 1] == '\\')) {  // deal with escaped quotes
-                end = str.find_first_of(keyChar, end + 1);
-                embeddedQuote = true;
-            }
-            if(end != std::string::npos) {
-                output.push_back(str.substr(1, end - 1));
+        if(bracketChars.find_first_of(str[0]) != std::string::npos) {
+            auto bracketLoc = bracketChars.find_first_of(str[0]);
+            auto end = close_sequence(str, 0, matchBracketChars[bracketLoc]);
+            if(end >= str.size()) {
+                output.push_back(std::move(str));
+                str.clear();
+            } else {
+                output.push_back(str.substr(0, end + 1));
                 if(end + 2 < str.size()) {
                     str = str.substr(end + 2);
                 } else {
                     str.clear();
                 }
-
-            } else {
-                output.push_back(str.substr(1));
-                str = "";
             }
+
         } else {
             auto it = std::find_if(std::begin(str), std::end(str), find_ws);
             if(it != std::end(str)) {
@@ -580,14 +1075,9 @@ CLI11_INLINE std::vector<std::string> sp
                 str = std::string(it + 1, str.end());
             } else {
                 output.push_back(str);
-                str = "";
+                str.clear();
             }
         }
-        // transform any embedded quotes into the regular character
-        if(embeddedQuote) {
-            output.back() = find_and_replace(output.back(), std::string("\\") + keyChar, std::string(1, keyChar));
-            embeddedQuote = false;
-        }
         trim(str);
     }
     return output;
@@ -605,15 +1095,140 @@ CLI11_INLINE std::size_t escape_detect(s
     return offset + 1;
 }
 
-CLI11_INLINE std::string &add_quotes_if_needed(std::string &str) {
-    if((str.front() != '"' && str.front() != '\'') || str.front() != str.back()) {
-        char quote = str.find('"') < str.find('\'') ? '\'' : '"';
-        if(str.find(' ') != std::string::npos) {
-            str.insert(0, 1, quote);
-            str.append(1, quote);
+CLI11_INLINE std::string binary_escape_string(const std::string &string_to_escape) {
+    // s is our escaped output string
+    std::string escaped_string{};
+    // loop through all characters
+    for(char c : string_to_escape) {
+        // check if a given character is printable
+        // the cast is necessary to avoid undefined behaviour
+        if(isprint(static_cast<unsigned char>(c)) == 0) {
+            std::stringstream stream;
+            // if the character is not printable
+            // we'll convert it to a hex string using a stringstream
+            // note that since char is signed we have to cast it to unsigned first
+            stream << std::hex << static_cast<unsigned int>(static_cast<unsigned char>(c));
+            std::string code = stream.str();
+            escaped_string += std::string("\\x") + (code.size() < 2 ? "0" : "") + code;
+
+        } else {
+            escaped_string.push_back(c);
+        }
+    }
+    if(escaped_string != string_to_escape) {
+        auto sqLoc = escaped_string.find('\'');
+        while(sqLoc != std::string::npos) {
+            escaped_string.replace(sqLoc, sqLoc + 1, "\\x27");
+            sqLoc = escaped_string.find('\'');
+        }
+        escaped_string.insert(0, "'B\"(");
+        escaped_string.push_back(')');
+        escaped_string.push_back('"');
+        escaped_string.push_back('\'');
+    }
+    return escaped_string;
+}
+
+CLI11_INLINE bool is_binary_escaped_string(const std::string &escaped_string) {
+    size_t ssize = escaped_string.size();
+    if(escaped_string.compare(0, 3, "B\"(") == 0 && escaped_string.compare(ssize - 2, 2, ")\"") == 0) {
+        return true;
+    }
+    return (escaped_string.compare(0, 4, "'B\"(") == 0 && escaped_string.compare(ssize - 3, 3, ")\"'") == 0);
+}
+
+CLI11_INLINE std::string extract_binary_string(const std::string &escaped_string) {
+    std::size_t start{0};
+    std::size_t tail{0};
+    size_t ssize = escaped_string.size();
+    if(escaped_string.compare(0, 3, "B\"(") == 0 && escaped_string.compare(ssize - 2, 2, ")\"") == 0) {
+        start = 3;
+        tail = 2;
+    } else if(escaped_string.compare(0, 4, "'B\"(") == 0 && escaped_string.compare(ssize - 3, 3, ")\"'") == 0) {
+        start = 4;
+        tail = 3;
+    }
+
+    if(start == 0) {
+        return escaped_string;
+    }
+    std::string outstring;
+
+    outstring.reserve(ssize - start - tail);
+    std::size_t loc = start;
+    while(loc < ssize - tail) {
+        // ssize-2 to skip )" at the end
+        if(escaped_string[loc] == '\\' && (escaped_string[loc + 1] == 'x' || escaped_string[loc + 1] == 'X')) {
+            auto c1 = escaped_string[loc + 2];
+            auto c2 = escaped_string[loc + 3];
+
+            std::uint32_t res1 = hexConvert(c1);
+            std::uint32_t res2 = hexConvert(c2);
+            if(res1 <= 0x0F && res2 <= 0x0F) {
+                loc += 4;
+                outstring.push_back(static_cast<char>(res1 * 16 + res2));
+                continue;
+            }
         }
+        outstring.push_back(escaped_string[loc]);
+        ++loc;
     }
-    return str;
+    return outstring;
+}
+
+CLI11_INLINE void remove_quotes(std::vector<std::string> &args) {
+    for(auto &arg : args) {
+        if(arg.front() == '\"' && arg.back() == '\"') {
+            remove_quotes(arg);
+            // only remove escaped for string arguments not literal strings
+            arg = remove_escaped_characters(arg);
+        } else {
+            remove_quotes(arg);
+        }
+    }
+}
+
+CLI11_INLINE bool process_quoted_string(std::string &str, char string_char, char literal_char) {
+    if(str.size() <= 1) {
+        return false;
+    }
+    if(detail::is_binary_escaped_string(str)) {
+        str = detail::extract_binary_string(str);
+        return true;
+    }
+    if(str.front() == string_char && str.back() == string_char) {
+        detail::remove_outer(str, string_char);
+        if(str.find_first_of('\\') != std::string::npos) {
+            str = detail::remove_escaped_characters(str);
+        }
+        return true;
+    }
+    if((str.front() == literal_char || str.front() == '`') && str.back() == str.front()) {
+        detail::remove_outer(str, str.front());
+        return true;
+    }
+    return false;
+}
+
+std::string get_environment_value(const std::string &env_name) {
+    char *buffer = nullptr;
+    std::string ename_string;
+
+#ifdef _MSC_VER
+    // Windows version
+    std::size_t sz = 0;
+    if(_dupenv_s(&buffer, &sz, env_name.c_str()) == 0 && buffer != nullptr) {
+        ename_string = std::string(buffer);
+        free(buffer);
+    }
+#else
+    // This also works on Windows, but gives a warning
+    buffer = std::getenv(env_name.c_str());
+    if(buffer != nullptr) {
+        ename_string = std::string(buffer);
+    }
+#endif
+    return ename_string;
 }
 
 }  // namespace detail
@@ -722,7 +1337,13 @@ class BadNameString : public Constructio
     CLI11_ERROR_DEF(ConstructionError, BadNameString)
     CLI11_ERROR_SIMPLE(BadNameString)
     static BadNameString OneCharName(std::string name) { return BadNameString("Invalid one char name: " + name); }
+    static BadNameString MissingDash(std::string name) {
+        return BadNameString("Long names strings require 2 dashes " + name);
+    }
     static BadNameString BadLongName(std::string name) { return BadNameString("Bad long name: " + name); }
+    static BadNameString BadPositionalName(std::string name) {
+        return BadNameString("Invalid positional Name: " + name);
+    }
     static BadNameString DashesOnly(std::string name) {
         return BadNameString("Must have a name, not just dashes: " + name);
     }
@@ -830,22 +1451,22 @@ class RequiredError : public ParseError
         if((min_option == 1) && (max_option == 1) && (used == 0))
             return RequiredError("Exactly 1 option from [" + option_list + "]");
         if((min_option == 1) && (max_option == 1) && (used > 1)) {
-            return {"Exactly 1 option from [" + option_list + "] is required and " + std::to_string(used) +
+            return {"Exactly 1 option from [" + option_list + "] is required but " + std::to_string(used) +
                         " were given",
                     ExitCodes::RequiredError};
         }
         if((min_option == 1) && (used == 0))
             return RequiredError("At least 1 option from [" + option_list + "]");
         if(used < min_option) {
-            return {"Requires at least " + std::to_string(min_option) + " options used and only " +
-                        std::to_string(used) + "were given from [" + option_list + "]",
+            return {"Requires at least " + std::to_string(min_option) + " options used but only " +
+                        std::to_string(used) + " were given from [" + option_list + "]",
                     ExitCodes::RequiredError};
         }
         if(max_option == 1)
             return {"Requires at most 1 options be given from [" + option_list + "]", ExitCodes::RequiredError};
 
-        return {"Requires at most " + std::to_string(max_option) + " options be used and " + std::to_string(used) +
-                    "were given from [" + option_list + "]",
+        return {"Requires at most " + std::to_string(max_option) + " options be used but " + std::to_string(used) +
+                    " were given from [" + option_list + "]",
                 ExitCodes::RequiredError};
     }
 };
@@ -1012,6 +1633,23 @@ template <> struct IsMemberType<const ch
     using type = std::string;
 };
 
+namespace adl_detail {
+/// Check for existence of user-supplied lexical_cast.
+///
+/// This struct has to be in a separate namespace so that it doesn't see our lexical_cast overloads in CLI::detail.
+/// Standard says it shouldn't see them if it's defined before the corresponding lexical_cast declarations, but this
+/// requires a working implementation of two-phase lookup, and not all compilers can boast that (msvc, ahem).
+template <typename T, typename S = std::string> class is_lexical_castable {
+    template <typename TT, typename SS>
+    static auto test(int) -> decltype(lexical_cast(std::declval<const SS &>(), std::declval<TT &>()), std::true_type());
+
+    template <typename, typename> static auto test(...) -> std::false_type;
+
+  public:
+    static constexpr bool value = decltype(test<T, S>(0))::value;
+};
+}  // namespace adl_detail
+
 namespace detail {
 
 // These are utilities for IsMember and other transforming objects
@@ -1087,14 +1725,22 @@ template <typename T, typename C> class
     static auto test(int, std::true_type) -> decltype(
 // NVCC warns about narrowing conversions here
 #ifdef __CUDACC__
+#ifdef __NVCC_DIAG_PRAGMA_SUPPORT__
+#pragma nv_diag_suppress 2361
+#else
 #pragma diag_suppress 2361
 #endif
-        TT{std::declval<CC>()}
+#endif
+                                              TT{std::declval<CC>()}
 #ifdef __CUDACC__
+#ifdef __NVCC_DIAG_PRAGMA_SUPPORT__
+#pragma nv_diag_default 2361
+#else
 #pragma diag_default 2361
 #endif
-        ,
-        std::is_move_assignable<TT>());
+#endif
+                                              ,
+                                              std::is_move_assignable<TT>());
 
     template <typename TT, typename CC> static auto test(int, std::false_type) -> std::false_type;
 
@@ -1171,8 +1817,10 @@ struct is_mutable_container<
                          decltype(std::declval<T>().clear()),
                          decltype(std::declval<T>().insert(std::declval<decltype(std::declval<T>().end())>(),
                                                            std::declval<const typename T::value_type &>()))>,
-                  void>>
-    : public conditional_t<std::is_constructible<T, std::string>::value, std::false_type, std::true_type> {};
+                  void>> : public conditional_t<std::is_constructible<T, std::string>::value ||
+                                                    std::is_constructible<T, std::wstring>::value,
+                                                std::false_type,
+                                                std::true_type> {};
 
 // check to see if an object is a mutable container (fail by default)
 template <typename T, typename _ = void> struct is_readable_container : std::false_type {};
@@ -1475,6 +2123,8 @@ enum class object_category : int {
     // string like types
     string_assignable = 23,
     string_constructible = 24,
+    wstring_assignable = 25,
+    wstring_constructible = 26,
     other = 45,
     // special wrapper or container types
     wrapper_value = 50,
@@ -1523,12 +2173,23 @@ template <typename T> struct classify_ob
 template <typename T> struct classify_object<T, typename std::enable_if<std::is_floating_point<T>::value>::type> {
     static constexpr object_category value{object_category::floating_point};
 };
+#if defined _MSC_VER
+// in MSVC wstring should take precedence if available this isn't as useful on other compilers due to the broader use of
+// utf-8 encoding
+#define WIDE_STRING_CHECK                                                                                              \
+    !std::is_assignable<T &, std::wstring>::value && !std::is_constructible<T, std::wstring>::value
+#define STRING_CHECK true
+#else
+#define WIDE_STRING_CHECK true
+#define STRING_CHECK !std::is_assignable<T &, std::string>::value && !std::is_constructible<T, std::string>::value
+#endif
 
 /// String and similar direct assignment
 template <typename T>
-struct classify_object<T,
-                       typename std::enable_if<!std::is_floating_point<T>::value && !std::is_integral<T>::value &&
-                                               std::is_assignable<T &, std::string>::value>::type> {
+struct classify_object<
+    T,
+    typename std::enable_if<!std::is_floating_point<T>::value && !std::is_integral<T>::value && WIDE_STRING_CHECK &&
+                            std::is_assignable<T &, std::string>::value>::type> {
     static constexpr object_category value{object_category::string_assignable};
 };
 
@@ -1538,10 +2199,27 @@ struct classify_object<
     T,
     typename std::enable_if<!std::is_floating_point<T>::value && !std::is_integral<T>::value &&
                             !std::is_assignable<T &, std::string>::value && (type_count<T>::value == 1) &&
-                            std::is_constructible<T, std::string>::value>::type> {
+                            WIDE_STRING_CHECK && std::is_constructible<T, std::string>::value>::type> {
     static constexpr object_category value{object_category::string_constructible};
 };
 
+/// Wide strings
+template <typename T>
+struct classify_object<T,
+                       typename std::enable_if<!std::is_floating_point<T>::value && !std::is_integral<T>::value &&
+                                               STRING_CHECK && std::is_assignable<T &, std::wstring>::value>::type> {
+    static constexpr object_category value{object_category::wstring_assignable};
+};
+
+template <typename T>
+struct classify_object<
+    T,
+    typename std::enable_if<!std::is_floating_point<T>::value && !std::is_integral<T>::value &&
+                            !std::is_assignable<T &, std::wstring>::value && (type_count<T>::value == 1) &&
+                            STRING_CHECK && std::is_constructible<T, std::wstring>::value>::type> {
+    static constexpr object_category value{object_category::wstring_constructible};
+};
+
 /// Enumerations
 template <typename T> struct classify_object<T, typename std::enable_if<std::is_enum<T>::value>::type> {
     static constexpr object_category value{object_category::enumeration};
@@ -1554,12 +2232,13 @@ template <typename T> struct classify_ob
 /// Handy helper to contain a bunch of checks that rule out many common types (integers, string like, floating point,
 /// vectors, and enumerations
 template <typename T> struct uncommon_type {
-    using type = typename std::conditional<!std::is_floating_point<T>::value && !std::is_integral<T>::value &&
-                                               !std::is_assignable<T &, std::string>::value &&
-                                               !std::is_constructible<T, std::string>::value && !is_complex<T>::value &&
-                                               !is_mutable_container<T>::value && !std::is_enum<T>::value,
-                                           std::true_type,
-                                           std::false_type>::type;
+    using type = typename std::conditional<
+        !std::is_floating_point<T>::value && !std::is_integral<T>::value &&
+            !std::is_assignable<T &, std::string>::value && !std::is_constructible<T, std::string>::value &&
+            !std::is_assignable<T &, std::wstring>::value && !std::is_constructible<T, std::wstring>::value &&
+            !is_complex<T>::value && !is_mutable_container<T>::value && !std::is_enum<T>::value,
+        std::true_type,
+        std::false_type>::type;
     static constexpr bool value = type::value;
 };
 
@@ -1748,7 +2427,7 @@ bool integral_conversion(const std::stri
     if(input.empty() || input.front() == '-') {
         return false;
     }
-    char *val = nullptr;
+    char *val{nullptr};
     errno = 0;
     std::uint64_t output_ll = std::strtoull(input.c_str(), &val, 0);
     if(errno == ERANGE) {
@@ -1764,6 +2443,33 @@ bool integral_conversion(const std::stri
         output = (output_sll < 0) ? static_cast<T>(0) : static_cast<T>(output_sll);
         return (static_cast<std::int64_t>(output) == output_sll);
     }
+    // remove separators
+    if(input.find_first_of("_'") != std::string::npos) {
+        std::string nstring = input;
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '_'), nstring.end());
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '\''), nstring.end());
+        return integral_conversion(nstring, output);
+    }
+    if(input.compare(0, 2, "0o") == 0) {
+        val = nullptr;
+        errno = 0;
+        output_ll = std::strtoull(input.c_str() + 2, &val, 8);
+        if(errno == ERANGE) {
+            return false;
+        }
+        output = static_cast<T>(output_ll);
+        return (val == (input.c_str() + input.size()) && static_cast<std::uint64_t>(output) == output_ll);
+    }
+    if(input.compare(0, 2, "0b") == 0) {
+        val = nullptr;
+        errno = 0;
+        output_ll = std::strtoull(input.c_str() + 2, &val, 2);
+        if(errno == ERANGE) {
+            return false;
+        }
+        output = static_cast<T>(output_ll);
+        return (val == (input.c_str() + input.size()) && static_cast<std::uint64_t>(output) == output_ll);
+    }
     return false;
 }
 
@@ -1788,11 +2494,38 @@ bool integral_conversion(const std::stri
         output = static_cast<T>(1);
         return true;
     }
+    // remove separators
+    if(input.find_first_of("_'") != std::string::npos) {
+        std::string nstring = input;
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '_'), nstring.end());
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '\''), nstring.end());
+        return integral_conversion(nstring, output);
+    }
+    if(input.compare(0, 2, "0o") == 0) {
+        val = nullptr;
+        errno = 0;
+        output_ll = std::strtoll(input.c_str() + 2, &val, 8);
+        if(errno == ERANGE) {
+            return false;
+        }
+        output = static_cast<T>(output_ll);
+        return (val == (input.c_str() + input.size()) && static_cast<std::int64_t>(output) == output_ll);
+    }
+    if(input.compare(0, 2, "0b") == 0) {
+        val = nullptr;
+        errno = 0;
+        output_ll = std::strtoll(input.c_str() + 2, &val, 2);
+        if(errno == ERANGE) {
+            return false;
+        }
+        output = static_cast<T>(output_ll);
+        return (val == (input.c_str() + input.size()) && static_cast<std::int64_t>(output) == output_ll);
+    }
     return false;
 }
 
-/// Convert a flag into an integer value  typically binary flags
-inline std::int64_t to_flag_value(std::string val) {
+/// Convert a flag into an integer value  typically binary flags sets errno to nonzero if conversion failed
+inline std::int64_t to_flag_value(std::string val) noexcept {
     static const std::string trueString("true");
     static const std::string falseString("false");
     if(val == trueString) {
@@ -1820,7 +2553,8 @@ inline std::int64_t to_flag_value(std::s
             ret = 1;
             break;
         default:
-            throw std::invalid_argument("unrecognized character");
+            errno = EINVAL;
+            return -1;
         }
         return ret;
     }
@@ -1829,7 +2563,11 @@ inline std::int64_t to_flag_value(std::s
     } else if(val == falseString || val == "off" || val == "no" || val == "disable") {
         ret = -1;
     } else {
-        ret = std::stoll(val);
+        char *loc_ptr{nullptr};
+        ret = std::strtoll(val.c_str(), &loc_ptr, 0);
+        if(loc_ptr != (val.c_str() + val.size()) && errno == 0) {
+            errno = EINVAL;
+        }
     }
     return ret;
 }
@@ -1858,18 +2596,16 @@ bool lexical_cast(const std::string &inp
 template <typename T,
           enable_if_t<classify_object<T>::value == object_category::boolean_value, detail::enabler> = detail::dummy>
 bool lexical_cast(const std::string &input, T &output) {
-    try {
-        auto out = to_flag_value(input);
+    errno = 0;
+    auto out = to_flag_value(input);
+    if(errno == 0) {
         output = (out > 0);
-        return true;
-    } catch(const std::invalid_argument &) {
-        return false;
-    } catch(const std::out_of_range &) {
-        // if the number is out of the range of a 64 bit value then it is still a number and for this purpose is still
-        // valid all we care about the sign
+    } else if(errno == ERANGE) {
         output = (input[0] != '-');
-        return true;
+    } else {
+        return false;
     }
+    return true;
 }
 
 /// Floats
@@ -1882,7 +2618,17 @@ bool lexical_cast(const std::string &inp
     char *val = nullptr;
     auto output_ld = std::strtold(input.c_str(), &val);
     output = static_cast<T>(output_ld);
-    return val == (input.c_str() + input.size());
+    if(val == (input.c_str() + input.size())) {
+        return true;
+    }
+    // remove separators
+    if(input.find_first_of("_'") != std::string::npos) {
+        std::string nstring = input;
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '_'), nstring.end());
+        nstring.erase(std::remove(nstring.begin(), nstring.end(), '\''), nstring.end());
+        return lexical_cast(nstring, output);
+    }
+    return false;
 }
 
 /// complex
@@ -1934,6 +2680,23 @@ bool lexical_cast(const std::string &inp
     return true;
 }
 
+/// Wide strings
+template <
+    typename T,
+    enable_if_t<classify_object<T>::value == object_category::wstring_assignable, detail::enabler> = detail::dummy>
+bool lexical_cast(const std::string &input, T &output) {
+    output = widen(input);
+    return true;
+}
+
+template <
+    typename T,
+    enable_if_t<classify_object<T>::value == object_category::wstring_constructible, detail::enabler> = detail::dummy>
+bool lexical_cast(const std::string &input, T &output) {
+    output = T{widen(input)};
+    return true;
+}
+
 /// Enumerations
 template <typename T,
           enable_if_t<classify_object<T>::value == object_category::enumeration, detail::enabler> = detail::dummy>
@@ -2047,13 +2810,24 @@ bool lexical_cast(const std::string &inp
 
 /// Non-string parsable by a stream
 template <typename T,
-          enable_if_t<classify_object<T>::value == object_category::other && !std::is_assignable<T &, int>::value,
+          enable_if_t<classify_object<T>::value == object_category::other && !std::is_assignable<T &, int>::value &&
+                          is_istreamable<T>::value,
                       detail::enabler> = detail::dummy>
 bool lexical_cast(const std::string &input, T &output) {
-    static_assert(is_istreamable<T>::value,
+    return from_stream(input, output);
+}
+
+/// Fallback overload that prints a human-readable error for types that we don't recognize and that don't have a
+/// user-supplied lexical_cast overload.
+template <typename T,
+          enable_if_t<classify_object<T>::value == object_category::other && !std::is_assignable<T &, int>::value &&
+                          !is_istreamable<T>::value && !adl_detail::is_lexical_castable<T>::value,
+                      detail::enabler> = detail::dummy>
+bool lexical_cast(const std::string & /*input*/, T & /*output*/) {
+    static_assert(!std::is_same<T, T>::value,  // Can't just write false here.
                   "option object type must have a lexical cast overload or streaming input operator(>>) defined, if it "
                   "is convertible from another type use the add_option<T, XC>(...) with XC being the known type");
-    return from_stream(input, output);
+    return false;
 }
 
 /// Assign a value through lexical cast operations
@@ -2062,7 +2836,9 @@ template <typename AssignTo,
           typename ConvertTo,
           enable_if_t<std::is_same<AssignTo, ConvertTo>::value &&
                           (classify_object<AssignTo>::value == object_category::string_assignable ||
-                           classify_object<AssignTo>::value == object_category::string_constructible),
+                           classify_object<AssignTo>::value == object_category::string_constructible ||
+                           classify_object<AssignTo>::value == object_category::wstring_assignable ||
+                           classify_object<AssignTo>::value == object_category::wstring_constructible),
                       detail::enabler> = detail::dummy>
 bool lexical_assign(const std::string &input, AssignTo &output) {
     return lexical_cast(input, output);
@@ -2073,7 +2849,9 @@ template <typename AssignTo,
           typename ConvertTo,
           enable_if_t<std::is_same<AssignTo, ConvertTo>::value && std::is_assignable<AssignTo &, AssignTo>::value &&
                           classify_object<AssignTo>::value != object_category::string_assignable &&
-                          classify_object<AssignTo>::value != object_category::string_constructible,
+                          classify_object<AssignTo>::value != object_category::string_constructible &&
+                          classify_object<AssignTo>::value != object_category::wstring_assignable &&
+                          classify_object<AssignTo>::value != object_category::wstring_constructible,
                       detail::enabler> = detail::dummy>
 bool lexical_assign(const std::string &input, AssignTo &output) {
     if(input.empty()) {
@@ -2112,9 +2890,17 @@ bool lexical_assign(const std::string &i
         output = 0;
         return true;
     }
-    int val = 0;
+    int val{0};
     if(lexical_cast(input, val)) {
+#if defined(__clang__)
+/* on some older clang compilers */
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wsign-conversion"
+#endif
         output = val;
+#if defined(__clang__)
+#pragma clang diagnostic pop
+#endif
         return true;
     }
     return false;
@@ -2169,12 +2955,12 @@ template <typename AssignTo,
                       detail::enabler> = detail::dummy>
 bool lexical_conversion(const std::vector<std ::string> &strings, AssignTo &output) {
     // the remove const is to handle pair types coming from a container
-    typename std::remove_const<typename std::tuple_element<0, ConvertTo>::type>::type v1;
-    typename std::tuple_element<1, ConvertTo>::type v2;
-    bool retval = lexical_assign<decltype(v1), decltype(v1)>(strings[0], v1);
-    if(strings.size() > 1) {
-        retval = retval && lexical_assign<decltype(v2), decltype(v2)>(strings[1], v2);
-    }
+    using FirstType = typename std::remove_const<typename std::tuple_element<0, ConvertTo>::type>::type;
+    using SecondType = typename std::tuple_element<1, ConvertTo>::type;
+    FirstType v1;
+    SecondType v2;
+    bool retval = lexical_assign<FirstType, FirstType>(strings[0], v1);
+    retval = retval && lexical_assign<SecondType, SecondType>((strings.size() > 1) ? strings[1] : std::string{}, v2);
     if(retval) {
         output = AssignTo{v1, v2};
     }
@@ -2189,6 +2975,9 @@ template <class AssignTo,
                       detail::enabler> = detail::dummy>
 bool lexical_conversion(const std::vector<std ::string> &strings, AssignTo &output) {
     output.erase(output.begin(), output.end());
+    if(strings.empty()) {
+        return true;
+    }
     if(strings.size() == 1 && strings[0] == "{}") {
         return true;
     }
@@ -2333,7 +3122,7 @@ tuple_type_conversion(std::vector<std::s
 
     std::size_t index{subtype_count_min<ConvertTo>::value};
     const std::size_t mx_count{subtype_count<ConvertTo>::value};
-    const std::size_t mx{(std::max)(mx_count, strings.size())};
+    const std::size_t mx{(std::min)(mx_count, strings.size() - 1)};
 
     while(index < mx) {
         if(is_separator(strings[index])) {
@@ -2343,7 +3132,11 @@ tuple_type_conversion(std::vector<std::s
     }
     bool retval = lexical_conversion<AssignTo, ConvertTo>(
         std::vector<std::string>(strings.begin(), strings.begin() + static_cast<std::ptrdiff_t>(index)), output);
-    strings.erase(strings.begin(), strings.begin() + static_cast<std::ptrdiff_t>(index) + 1);
+    if(strings.size() > index) {
+        strings.erase(strings.begin(), strings.begin() + static_cast<std::ptrdiff_t>(index) + 1);
+    } else {
+        strings.clear();
+    }
     return retval;
 }
 
@@ -2487,12 +3280,13 @@ inline std::string sum_string_vector(con
         double tv{0.0};
         auto comp = lexical_cast(arg, tv);
         if(!comp) {
-            try {
-                tv = static_cast<double>(detail::to_flag_value(arg));
-            } catch(const std::exception &) {
-                fail = true;
+            errno = 0;
+            auto fv = detail::to_flag_value(arg);
+            fail = (errno != 0);
+            if(fail) {
                 break;
             }
+            tv = static_cast<double>(fv);
         }
         val += tv;
     }
@@ -2501,13 +3295,10 @@ inline std::string sum_string_vector(con
             output.append(arg);
         }
     } else {
-        if(val <= static_cast<double>((std::numeric_limits<std::int64_t>::min)()) ||
-           val >= static_cast<double>((std::numeric_limits<std::int64_t>::max)()) ||
-           std::ceil(val) == std::floor(val)) {
-            output = detail::value_string(static_cast<int64_t>(val));
-        } else {
-            output = detail::value_string(val);
-        }
+        std::ostringstream out;
+        out.precision(16);
+        out << val;
+        output = out.str();
     }
     return output;
 }
@@ -2553,7 +3344,7 @@ CLI11_INLINE bool split_short(const std:
 }
 
 CLI11_INLINE bool split_long(const std::string &current, std::string &name, std::string &value) {
-    if(current.size() > 2 && current.substr(0, 2) == "--" && valid_first_char(current[2])) {
+    if(current.size() > 2 && current.compare(0, 2, "--") == 0 && valid_first_char(current[2])) {
         auto loc = current.find_first_of('=');
         if(loc != std::string::npos) {
             name = current.substr(2, loc - 2);
@@ -2625,7 +3416,6 @@ get_names(const std::vector<std::string>
     std::vector<std::string> short_names;
     std::vector<std::string> long_names;
     std::string pos_name;
-
     for(std::string name : input) {
         if(name.length() == 0) {
             continue;
@@ -2633,6 +3423,8 @@ get_names(const std::vector<std::string>
         if(name.length() > 1 && name[0] == '-' && name[1] != '-') {
             if(name.length() == 2 && valid_first_char(name[1]))
                 short_names.emplace_back(1, name[1]);
+            else if(name.length() > 2)
+                throw BadNameString::MissingDash(name);
             else
                 throw BadNameString::OneCharName(name);
         } else if(name.length() > 2 && name.substr(0, 2) == "--") {
@@ -2644,12 +3436,15 @@ get_names(const std::vector<std::string>
         } else if(name == "-" || name == "--") {
             throw BadNameString::DashesOnly(name);
         } else {
-            if(pos_name.length() > 0)
+            if(!pos_name.empty())
                 throw BadNameString::MultiPositionalNames(name);
-            pos_name = name;
+            if(valid_name_string(name)) {
+                pos_name = name;
+            } else {
+                throw BadNameString::BadPositionalName(name);
+            }
         }
     }
-
     return std::make_tuple(short_names, long_names, pos_name);
 }
 
@@ -2666,7 +3461,6 @@ struct ConfigItem {
 
     /// This is the name
     std::string name{};
-
     /// Listing of inputs
     std::vector<std::string> inputs{};
 
@@ -2729,8 +3523,8 @@ class ConfigBase : public Config {
     char valueDelimiter = '=';
     /// the character to use around strings
     char stringQuote = '"';
-    /// the character to use around single characters
-    char characterQuote = '\'';
+    /// the character to use around single characters and literal strings
+    char literalQuote = '\'';
     /// the maximum number of layers to allow
     uint8_t maximumLayers{255};
     /// the separator used to separator parent layers
@@ -2766,10 +3560,10 @@ class ConfigBase : public Config {
         valueDelimiter = vSep;
         return this;
     }
-    /// Specify the quote characters used around strings and characters
-    ConfigBase *quoteCharacter(char qString, char qChar) {
+    /// Specify the quote characters used around strings and literal strings
+    ConfigBase *quoteCharacter(char qString, char literalChar) {
         stringQuote = qString;
-        characterQuote = qChar;
+        literalQuote = literalChar;
         return this;
     }
     /// Specify the maximum number of parents
@@ -3001,6 +3795,11 @@ class IPV4Validator : public Validator {
     IPV4Validator();
 };
 
+class EscapedStringTransformer : public Validator {
+  public:
+    EscapedStringTransformer();
+};
+
 }  // namespace detail
 
 // Static is not needed here, because global const implies static.
@@ -3020,6 +3819,9 @@ const detail::NonexistentPathValidator N
 /// Check for an IP4 address
 const detail::IPV4Validator ValidIPV4;
 
+/// convert escaped characters into their associated values
+const detail::EscapedStringTransformer EscapedString;
+
 /// Validate the input as a particular type
 template <typename DesiredType> class TypeValidator : public Validator {
   public:
@@ -3772,14 +4574,14 @@ namespace detail {
 #if defined CLI11_HAS_FILESYSTEM && CLI11_HAS_FILESYSTEM > 0
 CLI11_INLINE path_type check_path(const char *file) noexcept {
     std::error_code ec;
-    auto stat = std::filesystem::status(file, ec);
+    auto stat = std::filesystem::status(to_path(file), ec);
     if(ec) {
         return path_type::nonexistent;
     }
     switch(stat.type()) {
     case std::filesystem::file_type::none:  // LCOV_EXCL_LINE
     case std::filesystem::file_type::not_found:
-        return path_type::nonexistent;
+        return path_type::nonexistent;  // LCOV_EXCL_LINE
     case std::filesystem::file_type::directory:
         return path_type::directory;
     case std::filesystem::file_type::symlink:
@@ -3873,10 +4675,29 @@ CLI11_INLINE IPV4Validator::IPV4Validato
                 return std::string("Each IP number must be between 0 and 255 ") + var;
             }
         }
-        return std::string();
+        return std::string{};
     };
 }
 
+CLI11_INLINE EscapedStringTransformer::EscapedStringTransformer() {
+    func_ = [](std::string &str) {
+        try {
+            if(str.size() > 1 && (str.front() == '\"' || str.front() == '\'' || str.front() == '`') &&
+               str.front() == str.back()) {
+                process_quoted_string(str);
+            } else if(str.find_first_of('\\') != std::string::npos) {
+                if(detail::is_binary_escaped_string(str)) {
+                    str = detail::extract_binary_string(str);
+                } else {
+                    str = remove_escaped_characters(str);
+                }
+            }
+            return std::string{};
+        } catch(const std::invalid_argument &ia) {
+            return std::string(ia.what());
+        }
+    };
+}
 }  // namespace detail
 
 CLI11_INLINE FileOnDefaultPath::FileOnDefaultPath(std::string default_path, bool enableErrorReturn)
@@ -4175,7 +4996,8 @@ enum class MultiOptionPolicy : char {
     TakeFirst,  //!< take only the first Expected number of arguments
     Join,       //!< merge all the arguments together into a single string via the delimiter character default('\n')
     TakeAll,    //!< just get all the passed argument regardless
-    Sum         //!< sum all the arguments together if numerical or concatenate directly without delimiter
+    Sum,        //!< sum all the arguments together if numerical or concatenate directly without delimiter
+    Reverse,    //!< take only the last Expected number of arguments in reverse order
 };
 
 /// This is the CRTP base class for Option and OptionDefaults. It was designed this way
@@ -4683,12 +5505,12 @@ class Option : public OptionBase<Option>
         if(!lnames_.empty()) {
             return lnames_[0];
         }
-        if(!pname_.empty()) {
-            return pname_;
-        }
         if(!snames_.empty()) {
             return snames_[0];
         }
+        if(!pname_.empty()) {
+            return pname_;
+        }
         return envname_;
     }
     /// The number of times the option expects to be included
@@ -4711,13 +5533,13 @@ class Option : public OptionBase<Option>
     CLI11_NODISCARD int get_items_expected() const { return get_items_expected_min(); }
 
     /// True if the argument can be given directly
-    CLI11_NODISCARD bool get_positional() const { return pname_.length() > 0; }
+    CLI11_NODISCARD bool get_positional() const { return !pname_.empty(); }
 
     /// True if option has at least one non-positional name
-    CLI11_NODISCARD bool nonpositional() const { return (snames_.size() + lnames_.size()) > 0; }
+    CLI11_NODISCARD bool nonpositional() const { return (!lnames_.empty() || !snames_.empty()); }
 
     /// True if option has description
-    CLI11_NODISCARD bool has_description() const { return description_.length() > 0; }
+    CLI11_NODISCARD bool has_description() const { return !description_.empty(); }
 
     /// Get the description
     CLI11_NODISCARD const std::string &get_description() const { return description_; }
@@ -5226,13 +6048,29 @@ CLI11_INLINE void Option::run_callback()
 
 CLI11_NODISCARD CLI11_INLINE const std::string &Option::matching_name(const Option &other) const {
     static const std::string estring;
-    for(const std::string &sname : snames_)
+    for(const std::string &sname : snames_) {
         if(other.check_sname(sname))
             return sname;
-    for(const std::string &lname : lnames_)
+        if(other.check_lname(sname))
+            return sname;
+    }
+    for(const std::string &lname : lnames_) {
         if(other.check_lname(lname))
             return lname;
-
+        if(lname.size() == 1) {
+            if(other.check_sname(lname)) {
+                return lname;
+            }
+        }
+    }
+    if(snames_.empty() && lnames_.empty() && !pname_.empty()) {
+        if(other.check_sname(pname_) || other.check_lname(pname_) || pname_ == other.pname_)
+            return pname_;
+    }
+    if(other.snames_.empty() && other.fnames_.empty() && !other.pname_.empty()) {
+        if(check_sname(other.pname_) || check_lname(other.pname_) || (pname_ == other.pname_))
+            return other.pname_;
+    }
     if(ignore_case_ ||
        ignore_underscore_) {  // We need to do the inverse, in case we are ignore_case or ignore underscore
         for(const std::string &sname : other.snames_)
@@ -5286,6 +6124,9 @@ CLI11_NODISCARD CLI11_INLINE std::string
             if(default_ind >= 0) {
                 // We can static cast this to std::size_t because it is more than 0 in this block
                 if(default_flag_values_[static_cast<std::size_t>(default_ind)].second != input_value) {
+                    if(input_value == default_str_ && force_callback_) {
+                        return input_value;
+                    }
                     throw(ArgumentMismatch::FlagOverride(name));
                 }
             } else {
@@ -5306,15 +6147,15 @@ CLI11_NODISCARD CLI11_INLINE std::string
         return input_value;
     }
     if(default_flag_values_[static_cast<std::size_t>(ind)].second == falseString) {
-        try {
-            auto val = detail::to_flag_value(input_value);
-            return (val == 1) ? falseString : (val == (-1) ? trueString : std::to_string(-val));
-        } catch(const std::invalid_argument &) {
+        errno = 0;
+        auto val = detail::to_flag_value(input_value);
+        if(errno != 0) {
+            errno = 0;
             return input_value;
         }
-    } else {
-        return input_value;
+        return (val == 1) ? falseString : (val == (-1) ? trueString : std::to_string(-val));
     }
+    return input_value;
 }
 
 CLI11_INLINE Option *Option::add_result(std::string s) {
@@ -5417,7 +6258,8 @@ CLI11_INLINE void Option::_validate_resu
         if(type_size_max_ > 1) {  // in this context index refers to the index in the type
             int index = 0;
             if(get_items_expected_max() < static_cast<int>(res.size()) &&
-               multi_option_policy_ == CLI::MultiOptionPolicy::TakeLast) {
+               (multi_option_policy_ == CLI::MultiOptionPolicy::TakeLast ||
+                multi_option_policy_ == CLI::MultiOptionPolicy::Reverse)) {
                 // create a negative index for the earliest ones
                 index = get_items_expected_max() - static_cast<int>(res.size());
             }
@@ -5435,7 +6277,8 @@ CLI11_INLINE void Option::_validate_resu
         } else {
             int index = 0;
             if(expected_max_ < static_cast<int>(res.size()) &&
-               multi_option_policy_ == CLI::MultiOptionPolicy::TakeLast) {
+               (multi_option_policy_ == CLI::MultiOptionPolicy::TakeLast ||
+                multi_option_policy_ == CLI::MultiOptionPolicy::Reverse)) {
                 // create a negative index for the earliest ones
                 index = expected_max_ - static_cast<int>(res.size());
             }
@@ -5467,6 +6310,15 @@ CLI11_INLINE void Option::_reduce_result
             out.assign(original.end() - static_cast<results_t::difference_type>(trim_size), original.end());
         }
     } break;
+    case MultiOptionPolicy::Reverse: {
+        // Allow multi-option sizes (including 0)
+        std::size_t trim_size = std::min<std::size_t>(
+            static_cast<std::size_t>(std::max<int>(get_items_expected_max(), 1)), original.size());
+        if(original.size() != trim_size || trim_size > 1) {
+            out.assign(original.end() - static_cast<results_t::difference_type>(trim_size), original.end());
+        }
+        std::reverse(out.begin(), out.end());
+    } break;
     case MultiOptionPolicy::TakeFirst: {
         std::size_t trim_size = std::min<std::size_t>(
             static_cast<std::size_t>(std::max<int>(get_items_expected_max(), 1)), original.size());
@@ -5496,7 +6348,12 @@ CLI11_INLINE void Option::_reduce_result
             throw ArgumentMismatch::AtLeast(get_name(), static_cast<int>(num_min), original.size());
         }
         if(original.size() > num_max) {
-            throw ArgumentMismatch::AtMost(get_name(), static_cast<int>(num_max), original.size());
+            if(original.size() == 2 && num_max == 1 && original[1] == "%%" && original[0] == "{}") {
+                // this condition is a trap for the following empty indicator check on config files
+                out = original;
+            } else {
+                throw ArgumentMismatch::AtMost(get_name(), static_cast<int>(num_max), original.size());
+            }
         }
         break;
     }
@@ -5505,11 +6362,11 @@ CLI11_INLINE void Option::_reduce_result
     // {} is the indicator for an empty container
     if(out.empty()) {
         if(original.size() == 1 && original[0] == "{}" && get_items_expected_min() > 0) {
-            out.push_back("{}");
-            out.push_back("%%");
+            out.emplace_back("{}");
+            out.emplace_back("%%");
         }
     } else if(out.size() == 1 && out[0] == "{}" && get_items_expected_min() > 0) {
-        out.push_back("%%");
+        out.emplace_back("%%");
     }
 }
 
@@ -5570,9 +6427,9 @@ CLI11_INLINE int Option::_add_result(std
 
 
 #ifndef CLI11_PARSE
-#define CLI11_PARSE(app, argc, argv)                                                                                   \
+#define CLI11_PARSE(app, ...)                                                                                          \
     try {                                                                                                              \
-        (app).parse((argc), (argv));                                                                                   \
+        (app).parse(__VA_ARGS__);                                                                                      \
     } catch(const CLI::ParseError &e) {                                                                                \
         return (app).exit(e);                                                                                          \
     }
@@ -5685,6 +6542,12 @@ class App {
     /// @name Help
     ///@{
 
+    /// Usage to put after program/subcommand description in the help output INHERITABLE
+    std::string usage_{};
+
+    /// This is a function that generates a usage to put after program/subcommand description in help output
+    std::function<std::string()> usage_callback_{};
+
     /// Footer to put after all options in the help output INHERITABLE
     std::string footer_{};
 
@@ -5819,6 +6682,14 @@ class App {
 
     ///@}
 
+#ifdef _WIN32
+    /// When normalizing argv to UTF-8 on Windows, this is the storage for normalized args.
+    std::vector<std::string> normalized_argv_{};
+
+    /// When normalizing argv to UTF-8 on Windows, this is the `char**` value returned to the user.
+    std::vector<char *> normalized_argv_view_{};
+#endif
+
     /// Special private constructor for subcommand
     App(std::string app_description, std::string app_name, App *parent);
 
@@ -5838,6 +6709,9 @@ class App {
     /// virtual destructor
     virtual ~App() = default;
 
+    /// Convert the contents of argv to UTF-8. Only does something on Windows, does nothing elsewhere.
+    CLI11_NODISCARD char **ensure_utf8(char **argv);
+
     /// Set a callback for execution when all parsing and processing has completed
     ///
     /// Due to a bug in c++11,
@@ -6251,6 +7125,10 @@ class App {
     /// Check to see if a subcommand is part of this command (text version)
     CLI11_NODISCARD App *get_subcommand(std::string subcom) const;
 
+    /// Get a subcommand by name (noexcept non-const version)
+    /// returns null if subcommand doesn't exist
+    CLI11_NODISCARD App *get_subcommand_no_throw(std::string subcom) const noexcept;
+
     /// Get a pointer to subcommand by index
     CLI11_NODISCARD App *get_subcommand(int index = 0) const;
 
@@ -6369,12 +7247,18 @@ class App {
     /// Parses the command line - throws errors.
     /// This must be called after the options are in but before the rest of the program.
     void parse(int argc, const char *const *argv);
+    void parse(int argc, const wchar_t *const *argv);
 
+  private:
+    template <class CharT> void parse_char_t(int argc, const CharT *const *argv);
+
+  public:
     /// Parse a single string as if it contained command line arguments.
     /// This function splits the string into arguments then calls parse(std::vector<std::string> &)
     /// the function takes an optional boolean argument specifying if the programName is included in the string to
     /// process
     void parse(std::string commandline, bool program_name_included = false);
+    void parse(std::wstring commandline, bool program_name_included = false);
 
     /// The real work is done here. Expects a reversed vector.
     /// Changes the vector to the remaining options.
@@ -6419,8 +7303,9 @@ class App {
     }
 
     /// Check with name instead of pointer to see if subcommand was selected
-    CLI11_NODISCARD bool got_subcommand(std::string subcommand_name) const {
-        return get_subcommand(subcommand_name)->parsed_ > 0;
+    CLI11_NODISCARD bool got_subcommand(std::string subcommand_name) const noexcept {
+        App *sub = get_subcommand_no_throw(subcommand_name);
+        return (sub != nullptr) ? (sub->parsed_ > 0) : false;
     }
 
     /// Sets excluded options for the subcommand
@@ -6482,6 +7367,16 @@ class App {
     /// @name Help
     ///@{
 
+    /// Set usage.
+    App *usage(std::string usage_string) {
+        usage_ = std::move(usage_string);
+        return this;
+    }
+    /// Set usage.
+    App *usage(std::function<std::string()> usage_function) {
+        usage_callback_ = std::move(usage_function);
+        return this;
+    }
     /// Set footer.
     App *footer(std::string footer_string) {
         footer_ = std::move(footer_string);
@@ -6540,7 +7435,7 @@ class App {
     std::vector<Option *> get_options(const std::function<bool(Option *)> filter = {});
 
     /// Get an option by name (noexcept non-const version)
-    Option *get_option_no_throw(std::string option_name) noexcept;
+    CLI11_NODISCARD Option *get_option_no_throw(std::string option_name) noexcept;
 
     /// Get an option by name (noexcept const version)
     CLI11_NODISCARD const Option *get_option_no_throw(std::string option_name) const noexcept;
@@ -6590,6 +7485,11 @@ class App {
     /// Get the group of this subcommand
     CLI11_NODISCARD const std::string &get_group() const { return group_; }
 
+    /// Generate and return the usage.
+    CLI11_NODISCARD std::string get_usage() const {
+        return (usage_callback_) ? usage_callback_() + '\n' + usage_ : usage_;
+    }
+
     /// Generate and return the footer.
     CLI11_NODISCARD std::string get_footer() const {
         return (footer_callback_) ? footer_callback_() + '\n' + footer_ : footer_;
@@ -6727,6 +7627,9 @@ class App {
     /// Read and process a configuration file (main app only)
     void _process_config_file();
 
+    /// Read and process a particular configuration file
+    bool _process_config_file(const std::string &config_file, bool throw_error);
+
     /// Get envname options if not yet passed. Runs on *all* subcommands.
     void _process_env();
 
@@ -6799,8 +7702,9 @@ class App {
     bool _parse_subcommand(std::vector<std::string> &args);
 
     /// Parse a short (false) or long (true) argument, must be at the top of the list
+    /// if local_processing_only is set to true then fallthrough is disabled will return false if not found
     /// return true if the argument was processed or false if nothing was done
-    bool _parse_arg(std::vector<std::string> &args, detail::Classifier current_type);
+    bool _parse_arg(std::vector<std::string> &args, detail::Classifier current_type, bool local_processing_only);
 
     /// Trigger the pre_parse callback if needed
     void _trigger_pre_parse(std::size_t remaining_args);
@@ -6955,6 +7859,7 @@ CLI11_INLINE App::App(std::string app_de
         configurable_ = parent_->configurable_;
         allow_windows_style_options_ = parent_->allow_windows_style_options_;
         group_ = parent_->group_;
+        usage_ = parent_->usage_;
         footer_ = parent_->footer_;
         formatter_ = parent_->formatter_;
         config_formatter_ = parent_->config_formatter_;
@@ -6962,10 +7867,32 @@ CLI11_INLINE App::App(std::string app_de
     }
 }
 
+CLI11_NODISCARD CLI11_INLINE char **App::ensure_utf8(char **argv) {
+#ifdef _WIN32
+    (void)argv;
+
+    normalized_argv_ = detail::compute_win32_argv();
+
+    if(!normalized_argv_view_.empty()) {
+        normalized_argv_view_.clear();
+    }
+
+    normalized_argv_view_.reserve(normalized_argv_.size());
+    for(auto &arg : normalized_argv_) {
+        // using const_cast is well-defined, string is known to not be const.
+        normalized_argv_view_.push_back(const_cast<char *>(arg.data()));
+    }
+
+    return normalized_argv_view_.data();
+#else
+    return argv;
+#endif
+}
+
 CLI11_INLINE App *App::name(std::string app_name) {
 
     if(parent_ != nullptr) {
-        auto oname = name_;
+        std::string oname = name_;
         name_ = app_name;
         const auto &res = _compare_subcommand_names(*this, *_get_fallthrough_parent());
         if(!res.empty()) {
@@ -7046,6 +7973,32 @@ CLI11_INLINE Option *App::add_option(std
 
     if(std::find_if(std::begin(options_), std::end(options_), [&myopt](const Option_p &v) { return *v == myopt; }) ==
        std::end(options_)) {
+        if(myopt.lnames_.empty() && myopt.snames_.empty()) {
+            // if the option is positional only there is additional potential for ambiguities in config files and needs
+            // to be checked
+            std::string test_name = "--" + myopt.get_single_name();
+            if(test_name.size() == 3) {
+                test_name.erase(0, 1);
+            }
+
+            auto *op = get_option_no_throw(test_name);
+            if(op != nullptr) {
+                throw(OptionAlreadyAdded("added option positional name matches existing option: " + test_name));
+            }
+        } else if(parent_ != nullptr) {
+            for(auto &ln : myopt.lnames_) {
+                auto *op = parent_->get_option_no_throw(ln);
+                if(op != nullptr) {
+                    throw(OptionAlreadyAdded("added option matches existing positional option: " + ln));
+                }
+            }
+            for(auto &sn : myopt.snames_) {
+                auto *op = parent_->get_option_no_throw(sn);
+                if(op != nullptr) {
+                    throw(OptionAlreadyAdded("added option matches existing positional option: " + sn));
+                }
+            }
+        }
         options_.emplace_back();
         Option_p &option = options_.back();
         option.reset(new Option(option_name, option_description, option_callback, this));
@@ -7136,8 +8089,8 @@ App::set_version_flag(std::string flag_n
 
     // Empty name will simply remove the version flag
     if(!flag_name.empty()) {
-        version_ptr_ = add_flag_callback(
-            flag_name, [vfunc]() { throw(CLI::CallForVersion(vfunc(), 0)); }, version_help);
+        version_ptr_ =
+            add_flag_callback(flag_name, [vfunc]() { throw(CLI::CallForVersion(vfunc(), 0)); }, version_help);
         version_ptr_->configurable(false);
     }
 
@@ -7220,8 +8173,11 @@ CLI11_INLINE Option *App::set_config(std
         }
         if(!default_filename.empty()) {
             config_ptr_->default_str(std::move(default_filename));
+            config_ptr_->force_callback_ = true;
         }
         config_ptr_->configurable(false);
+        // set the option to take the last value and reverse given by default
+        config_ptr_->multi_option_policy(MultiOptionPolicy::Reverse);
     }
 
     return config_ptr_;
@@ -7251,13 +8207,14 @@ CLI11_INLINE bool App::remove_option(Opt
 CLI11_INLINE App *App::add_subcommand(std::string subcommand_name, std::string subcommand_description) {
     if(!subcommand_name.empty() && !detail::valid_name_string(subcommand_name)) {
         if(!detail::valid_first_char(subcommand_name[0])) {
-            throw IncorrectConstruction("Subcommand name starts with invalid character, '!' and '-' are not allowed");
+            throw IncorrectConstruction(
+                "Subcommand name starts with invalid character, '!' and '-' and control characters");
         }
         for(auto c : subcommand_name) {
             if(!detail::valid_later_char(c)) {
                 throw IncorrectConstruction(std::string("Subcommand name contains invalid character ('") + c +
                                             "'), all characters are allowed except"
-                                            "'=',':','{','}', and ' '");
+                                            "'=',':','{','}', ' ', and control characters");
             }
         }
     }
@@ -7310,6 +8267,10 @@ CLI11_NODISCARD CLI11_INLINE App *App::g
     return subc;
 }
 
+CLI11_NODISCARD CLI11_INLINE App *App::get_subcommand_no_throw(std::string subcom) const noexcept {
+    return _find_subcommand(subcom, false, false);
+}
+
 CLI11_NODISCARD CLI11_INLINE App *App::get_subcommand(int index) const {
     if(index >= 0) {
         auto uindex = static_cast<unsigned>(index);
@@ -7382,17 +8343,29 @@ CLI11_INLINE void App::clear() {
     }
 }
 
-CLI11_INLINE void App::parse(int argc, const char *const *argv) {
+CLI11_INLINE void App::parse(int argc, const char *const *argv) { parse_char_t(argc, argv); }
+CLI11_INLINE void App::parse(int argc, const wchar_t *const *argv) { parse_char_t(argc, argv); }
+
+namespace detail {
+
+// Do nothing or perform narrowing
+CLI11_INLINE const char *maybe_narrow(const char *str) { return str; }
+CLI11_INLINE std::string maybe_narrow(const wchar_t *str) { return narrow(str); }
+
+}  // namespace detail
+
+template <class CharT> CLI11_INLINE void App::parse_char_t(int argc, const CharT *const *argv) {
     // If the name is not set, read from command line
     if(name_.empty() || has_automatic_name_) {
         has_automatic_name_ = true;
-        name_ = argv[0];
+        name_ = detail::maybe_narrow(argv[0]);
     }
 
     std::vector<std::string> args;
     args.reserve(static_cast<std::size_t>(argc) - 1U);
     for(auto i = static_cast<std::size_t>(argc) - 1U; i > 0U; --i)
-        args.emplace_back(argv[i]);
+        args.emplace_back(detail::maybe_narrow(argv[i]));
+
     parse(std::move(args));
 }
 
@@ -7418,11 +8391,19 @@ CLI11_INLINE void App::parse(std::string
     auto args = detail::split_up(std::move(commandline));
     // remove all empty strings
     args.erase(std::remove(args.begin(), args.end(), std::string{}), args.end());
+    try {
+        detail::remove_quotes(args);
+    } catch(const std::invalid_argument &arg) {
+        throw CLI::ParseError(arg.what(), CLI::ExitCodes::InvalidError);
+    }
     std::reverse(args.begin(), args.end());
-
     parse(std::move(args));
 }
 
+CLI11_INLINE void App::parse(std::wstring commandline, bool program_name_included) {
+    parse(narrow(commandline), program_name_included);
+}
+
 CLI11_INLINE void App::parse(std::vector<std::string> &args) {
     // Clear if parsed
     if(parsed_ > 0)
@@ -7489,7 +8470,7 @@ CLI11_INLINE int App::exit(const Error &
     }
 
     if(e.get_name() == "CallForVersion") {
-        out << e.what() << std::endl;
+        out << e.what() << '\n';
         return e.get_exit_code();
     }
 
@@ -7585,7 +8566,8 @@ CLI11_NODISCARD CLI11_INLINE std::string
 CLI11_NODISCARD CLI11_INLINE std::string App::version() const {
     std::string val;
     if(version_ptr_ != nullptr) {
-        auto rv = version_ptr_->results();
+        // copy the results for reuse later
+        results_t rv = version_ptr_->results();
         version_ptr_->clear();
         version_ptr_->add_result("true");
         try {
@@ -7628,7 +8610,7 @@ CLI11_INLINE std::vector<Option *> App::
     return options;
 }
 
-CLI11_INLINE Option *App::get_option_no_throw(std::string option_name) noexcept {
+CLI11_NODISCARD CLI11_INLINE Option *App::get_option_no_throw(std::string option_name) noexcept {
     for(Option_p &opt : options_) {
         if(opt->check_name(option_name)) {
             return opt.get();
@@ -7694,7 +8676,7 @@ CLI11_NODISCARD CLI11_INLINE bool App::c
     if(local_name == name_to_check) {
         return true;
     }
-    for(auto les : aliases_) {  // NOLINT(performance-for-range-copy)
+    for(std::string les : aliases_) {  // NOLINT(performance-for-range-copy)
         if(ignore_underscore_) {
             les = detail::remove_underscore(les);
         }
@@ -7884,70 +8866,94 @@ CLI11_NODISCARD CLI11_INLINE detail::Cla
         return detail::Classifier::WINDOWS_STYLE;
     if((current == "++") && !name_.empty() && parent_ != nullptr)
         return detail::Classifier::SUBCOMMAND_TERMINATOR;
+    auto dotloc = current.find_first_of('.');
+    if(dotloc != std::string::npos) {
+        auto *cm = _find_subcommand(current.substr(0, dotloc), true, ignore_used_subcommands);
+        if(cm != nullptr) {
+            auto res = cm->_recognize(current.substr(dotloc + 1), ignore_used_subcommands);
+            if(res == detail::Classifier::SUBCOMMAND) {
+                return res;
+            }
+        }
+    }
     return detail::Classifier::NONE;
 }
 
+CLI11_INLINE bool App::_process_config_file(const std::string &config_file, bool throw_error) {
+    auto path_result = detail::check_path(config_file.c_str());
+    if(path_result == detail::path_type::file) {
+        try {
+            std::vector<ConfigItem> values = config_formatter_->from_file(config_file);
+            _parse_config(values);
+            return true;
+        } catch(const FileError &) {
+            if(throw_error) {
+                throw;
+            }
+            return false;
+        }
+    } else if(throw_error) {
+        throw FileError::Missing(config_file);
+    } else {
+        return false;
+    }
+}
+
 CLI11_INLINE void App::_process_config_file() {
     if(config_ptr_ != nullptr) {
         bool config_required = config_ptr_->get_required();
         auto file_given = config_ptr_->count() > 0;
+        if(!(file_given || config_ptr_->envname_.empty())) {
+            std::string ename_string = detail::get_environment_value(config_ptr_->envname_);
+            if(!ename_string.empty()) {
+                config_ptr_->add_result(ename_string);
+            }
+        }
+        config_ptr_->run_callback();
+
         auto config_files = config_ptr_->as<std::vector<std::string>>();
+        bool files_used{file_given};
         if(config_files.empty() || config_files.front().empty()) {
             if(config_required) {
-                throw FileError::Missing("no specified config file");
+                throw FileError("config file is required but none was given");
             }
             return;
         }
-        for(auto rit = config_files.rbegin(); rit != config_files.rend(); ++rit) {
-            const auto &config_file = *rit;
-            auto path_result = detail::check_path(config_file.c_str());
-            if(path_result == detail::path_type::file) {
-                try {
-                    std::vector<ConfigItem> values = config_formatter_->from_file(config_file);
-                    _parse_config(values);
-                    if(!file_given) {
-                        config_ptr_->add_result(config_file);
-                    }
-                } catch(const FileError &) {
-                    if(config_required || file_given)
-                        throw;
-                }
-            } else if(config_required || file_given) {
-                throw FileError::Missing(config_file);
+        for(const auto &config_file : config_files) {
+            if(_process_config_file(config_file, config_required || file_given)) {
+                files_used = true;
             }
         }
+        if(!files_used) {
+            // this is done so the count shows as 0 if no callbacks were processed
+            config_ptr_->clear();
+            bool force = config_ptr_->force_callback_;
+            config_ptr_->force_callback_ = false;
+            config_ptr_->run_callback();
+            config_ptr_->force_callback_ = force;
+        }
     }
 }
 
 CLI11_INLINE void App::_process_env() {
     for(const Option_p &opt : options_) {
         if(opt->count() == 0 && !opt->envname_.empty()) {
-            char *buffer = nullptr;
-            std::string ename_string;
-
-#ifdef _MSC_VER
-            // Windows version
-            std::size_t sz = 0;
-            if(_dupenv_s(&buffer, &sz, opt->envname_.c_str()) == 0 && buffer != nullptr) {
-                ename_string = std::string(buffer);
-                free(buffer);
-            }
-#else
-            // This also works on Windows, but gives a warning
-            buffer = std::getenv(opt->envname_.c_str());
-            if(buffer != nullptr)
-                ename_string = std::string(buffer);
-#endif
-
+            std::string ename_string = detail::get_environment_value(opt->envname_);
             if(!ename_string.empty()) {
-                opt->add_result(ename_string);
+                std::string result = ename_string;
+                result = opt->_validate(result, 0);
+                if(result.empty()) {
+                    opt->add_result(ename_string);
+                }
             }
         }
     }
 
     for(App_p &sub : subcommands_) {
-        if(sub->get_name().empty() || !sub->parse_complete_callback_)
+        if(sub->get_name().empty() || (sub->count_all() > 0 && !sub->parse_complete_callback_)) {
+            // only process environment variables if the callback has actually been triggered already
             sub->_process_env();
+        }
     }
 }
 
@@ -8247,15 +9253,10 @@ CLI11_INLINE void App::_parse_config(con
 }
 
 CLI11_INLINE bool App::_parse_single_config(const ConfigItem &item, std::size_t level) {
-    if(level < item.parents.size()) {
-        try {
-            auto *subcom = get_subcommand(item.parents.at(level));
-            auto result = subcom->_parse_single_config(item, level + 1);
 
-            return result;
-        } catch(const OptionNotFound &) {
-            return false;
-        }
+    if(level < item.parents.size()) {
+        auto *subcom = get_subcommand_no_throw(item.parents.at(level));
+        return (subcom != nullptr) ? subcom->_parse_single_config(item, level + 1) : false;
     }
     // check for section open
     if(item.name == "++") {
@@ -8282,15 +9283,20 @@ CLI11_INLINE bool App::_parse_single_con
         if(item.name.size() == 1) {
             op = get_option_no_throw("-" + item.name);
         }
+        if(op == nullptr) {
+            op = get_option_no_throw(item.name);
+        }
     }
-    if(op == nullptr) {
-        op = get_option_no_throw(item.name);
-    }
+
     if(op == nullptr) {
         // If the option was not present
-        if(get_allow_config_extras() == config_extras_mode::capture)
+        if(get_allow_config_extras() == config_extras_mode::capture) {
             // Should we worry about classifying the extras properly?
             missing_.emplace_back(detail::Classifier::NONE, item.fullname());
+            for(const auto &input : item.inputs) {
+                missing_.emplace_back(detail::Classifier::NONE, input);
+            }
+        }
         return false;
     }
 
@@ -8309,29 +9315,54 @@ CLI11_INLINE bool App::_parse_single_con
                 auto res = config_formatter_->to_flag(item);
                 bool converted{false};
                 if(op->get_disable_flag_override()) {
-
-                    try {
-                        auto val = detail::to_flag_value(res);
-                        if(val == 1) {
-                            res = op->get_flag_value(item.name, "{}");
-                            converted = true;
-                        }
-                    } catch(...) {
+                    auto val = detail::to_flag_value(res);
+                    if(val == 1) {
+                        res = op->get_flag_value(item.name, "{}");
+                        converted = true;
                     }
                 }
 
                 if(!converted) {
+                    errno = 0;
                     res = op->get_flag_value(item.name, res);
                 }
 
                 op->add_result(res);
                 return true;
             }
-            if(static_cast<int>(item.inputs.size()) > op->get_items_expected_max()) {
+            if(static_cast<int>(item.inputs.size()) > op->get_items_expected_max() &&
+               op->get_multi_option_policy() != MultiOptionPolicy::TakeAll) {
                 if(op->get_items_expected_max() > 1) {
                     throw ArgumentMismatch::AtMost(item.fullname(), op->get_items_expected_max(), item.inputs.size());
                 }
-                throw ConversionError::TooManyInputsFlag(item.fullname());
+
+                if(!op->get_disable_flag_override()) {
+                    throw ConversionError::TooManyInputsFlag(item.fullname());
+                }
+                // if the disable flag override is set then we must have the flag values match a known flag value
+                // this is true regardless of the output value, so an array input is possible and must be accounted for
+                for(const auto &res : item.inputs) {
+                    bool valid_value{false};
+                    if(op->default_flag_values_.empty()) {
+                        if(res == "true" || res == "false" || res == "1" || res == "0") {
+                            valid_value = true;
+                        }
+                    } else {
+                        for(const auto &valid_res : op->default_flag_values_) {
+                            if(valid_res.second == res) {
+                                valid_value = true;
+                                break;
+                            }
+                        }
+                    }
+
+                    if(valid_value) {
+                        op->add_result(res);
+                    } else {
+                        throw InvalidError("invalid flag argument given");
+                    }
+                }
+                return true;
             }
         }
         op->add_result(item.inputs);
@@ -8366,7 +9397,7 @@ CLI11_INLINE bool App::_parse_single(std
     case detail::Classifier::SHORT:
     case detail::Classifier::WINDOWS_STYLE:
         // If already parsed a subcommand, don't accept options_
-        _parse_arg(args, classifier);
+        retval = _parse_arg(args, classifier, false);
         break;
     case detail::Classifier::NONE:
         // Probably a positional or something for a parent (sub)command
@@ -8408,6 +9439,7 @@ CLI11_NODISCARD CLI11_INLINE bool App::_
 CLI11_INLINE bool App::_parse_positional(std::vector<std::string> &args, bool haltOnSubcommand) {
 
     const std::string &positional = args.back();
+    Option *posOpt{nullptr};
 
     if(positionals_at_end_) {
         // deal with the case of required arguments at the end which should take precedence over other arguments
@@ -8424,56 +9456,47 @@ CLI11_INLINE bool App::_parse_positional
                                 continue;
                             }
                         }
-
-                        parse_order_.push_back(opt.get());
-                        /// if we require a separator add it here
-                        if(opt->get_inject_separator()) {
-                            if(!opt->results().empty() && !opt->results().back().empty()) {
-                                opt->add_result(std::string{});
-                            }
-                        }
-                        if(opt->get_trigger_on_parse() &&
-                           opt->current_option_state_ == Option::option_state::callback_run) {
-                            opt->clear();
-                        }
-                        opt->add_result(positional);
-                        if(opt->get_trigger_on_parse()) {
-                            opt->run_callback();
-                        }
-                        args.pop_back();
-                        return true;
+                        posOpt = opt.get();
+                        break;
                     }
                 }
             }
         }
     }
-    for(const Option_p &opt : options_) {
-        // Eat options, one by one, until done
-        if(opt->get_positional() &&
-           (static_cast<int>(opt->count()) < opt->get_items_expected_min() || opt->get_allow_extra_args())) {
-            if(validate_positionals_) {
-                std::string pos = positional;
-                pos = opt->_validate(pos, 0);
-                if(!pos.empty()) {
-                    continue;
-                }
-            }
-            if(opt->get_inject_separator()) {
-                if(!opt->results().empty() && !opt->results().back().empty()) {
-                    opt->add_result(std::string{});
+    if(posOpt == nullptr) {
+        for(const Option_p &opt : options_) {
+            // Eat options, one by one, until done
+            if(opt->get_positional() &&
+               (static_cast<int>(opt->count()) < opt->get_items_expected_min() || opt->get_allow_extra_args())) {
+                if(validate_positionals_) {
+                    std::string pos = positional;
+                    pos = opt->_validate(pos, 0);
+                    if(!pos.empty()) {
+                        continue;
+                    }
                 }
+                posOpt = opt.get();
+                break;
             }
-            if(opt->get_trigger_on_parse() && opt->current_option_state_ == Option::option_state::callback_run) {
-                opt->clear();
-            }
-            opt->add_result(positional);
-            if(opt->get_trigger_on_parse()) {
-                opt->run_callback();
+        }
+    }
+    if(posOpt != nullptr) {
+        parse_order_.push_back(posOpt);
+        if(posOpt->get_inject_separator()) {
+            if(!posOpt->results().empty() && !posOpt->results().back().empty()) {
+                posOpt->add_result(std::string{});
             }
-            parse_order_.push_back(opt.get());
-            args.pop_back();
-            return true;
         }
+        if(posOpt->get_trigger_on_parse() && posOpt->current_option_state_ == Option::option_state::callback_run) {
+            posOpt->clear();
+        }
+        posOpt->add_result(positional);
+        if(posOpt->get_trigger_on_parse()) {
+            posOpt->run_callback();
+        }
+
+        args.pop_back();
+        return true;
     }
 
     for(auto &subc : subcommands_) {
@@ -8554,6 +9577,17 @@ CLI11_INLINE bool App::_parse_subcommand
         return true;
     }
     auto *com = _find_subcommand(args.back(), true, true);
+    if(com == nullptr) {
+        // the main way to get here is using .notation
+        auto dotloc = args.back().find_first_of('.');
+        if(dotloc != std::string::npos) {
+            com = _find_subcommand(args.back().substr(0, dotloc), true, true);
+            if(com != nullptr) {
+                args.back() = args.back().substr(dotloc + 1);
+                args.push_back(com->get_display_name());
+            }
+        }
+    }
     if(com != nullptr) {
         args.pop_back();
         if(!com->silent_) {
@@ -8576,7 +9610,8 @@ CLI11_INLINE bool App::_parse_subcommand
     return false;
 }
 
-CLI11_INLINE bool App::_parse_arg(std::vector<std::string> &args, detail::Classifier current_type) {
+CLI11_INLINE bool
+App::_parse_arg(std::vector<std::string> &args, detail::Classifier current_type, bool local_processing_only) {
 
     std::string current = args.back();
 
@@ -8618,7 +9653,7 @@ CLI11_INLINE bool App::_parse_arg(std::v
     if(op_ptr == std::end(options_)) {
         for(auto &subc : subcommands_) {
             if(subc->name_.empty() && !subc->disabled_) {
-                if(subc->_parse_arg(args, current_type)) {
+                if(subc->_parse_arg(args, current_type, local_processing_only)) {
                     if(!subc->pre_parse_called_) {
                         subc->_trigger_pre_parse(args.size());
                     }
@@ -8632,9 +9667,57 @@ CLI11_INLINE bool App::_parse_arg(std::v
             return false;
         }
 
+        // now check for '.' notation of subcommands
+        auto dotloc = arg_name.find_first_of('.', 1);
+        if(dotloc != std::string::npos) {
+            // using dot notation is equivalent to single argument subcommand
+            auto *sub = _find_subcommand(arg_name.substr(0, dotloc), true, false);
+            if(sub != nullptr) {
+                auto v = args.back();
+                args.pop_back();
+                arg_name = arg_name.substr(dotloc + 1);
+                if(arg_name.size() > 1) {
+                    args.push_back(std::string("--") + v.substr(dotloc + 3));
+                    current_type = detail::Classifier::LONG;
+                } else {
+                    auto nval = v.substr(dotloc + 2);
+                    nval.front() = '-';
+                    if(nval.size() > 2) {
+                        // '=' not allowed in short form arguments
+                        args.push_back(nval.substr(3));
+                        nval.resize(2);
+                    }
+                    args.push_back(nval);
+                    current_type = detail::Classifier::SHORT;
+                }
+                auto val = sub->_parse_arg(args, current_type, true);
+                if(val) {
+                    if(!sub->silent_) {
+                        parsed_subcommands_.push_back(sub);
+                    }
+                    // deal with preparsing
+                    increment_parsed();
+                    _trigger_pre_parse(args.size());
+                    // run the parse complete callback since the subcommand processing is now complete
+                    if(sub->parse_complete_callback_) {
+                        sub->_process_env();
+                        sub->_process_callbacks();
+                        sub->_process_help_flags();
+                        sub->_process_requirements();
+                        sub->run_callback(false, true);
+                    }
+                    return true;
+                }
+                args.pop_back();
+                args.push_back(v);
+            }
+        }
+        if(local_processing_only) {
+            return false;
+        }
         // If a subcommand, try the main command
         if(parent_ != nullptr && fallthrough_)
-            return _get_fallthrough_parent()->_parse_arg(args, current_type);
+            return _get_fallthrough_parent()->_parse_arg(args, current_type, false);
 
         // Otherwise, add to missing
         args.pop_back();
@@ -8756,7 +9839,7 @@ CLI11_INLINE void App::_trigger_pre_pars
     } else if(immediate_callback_) {
         if(!name_.empty()) {
             auto pcnt = parsed_;
-            auto extras = std::move(missing_);
+            missing_t extras = std::move(missing_);
             clear();
             parsed_ = pcnt;
             pre_parse_called_ = true;
@@ -8942,12 +10025,12 @@ CLI11_INLINE void retire_option(App *app
                             ->allow_extra_args(opt->get_allow_extra_args());
 
     app->remove_option(opt);
-    auto *opt2 = app->add_option(option_copy->get_name(false, true), "option has been retired and has no effect")
-                     ->type_name("RETIRED")
-                     ->default_str("RETIRED")
-                     ->type_size(option_copy->get_type_size_min(), option_copy->get_type_size_max())
-                     ->expected(option_copy->get_expected_min(), option_copy->get_expected_max())
-                     ->allow_extra_args(option_copy->get_allow_extra_args());
+    auto *opt2 = app->add_option(option_copy->get_name(false, true), "option has been retired and has no effect");
+    opt2->type_name("RETIRED")
+        ->default_str("RETIRED")
+        ->type_size(option_copy->get_type_size_min(), option_copy->get_type_size_max())
+        ->expected(option_copy->get_expected_min(), option_copy->get_expected_max())
+        ->allow_extra_args(option_copy->get_allow_extra_args());
 
     Validator retired_warning{[opt2](std::string &) {
                                   std::cout << "WARNING " << opt2->get_name() << " is retired and has no effect\n";
@@ -9015,7 +10098,10 @@ CLI11_INLINE std::string help(const App
 
 namespace detail {
 
-std::string convert_arg_for_ini(const std::string &arg, char stringQuote = '"', char characterQuote = '\'');
+std::string convert_arg_for_ini(const std::string &arg,
+                                char stringQuote = '"',
+                                char literalQuote = '\'',
+                                bool disable_multi_line = false);
 
 /// Comma separated join, adds quotes if needed
 std::string ini_join(const std::vector<std::string> &args,
@@ -9023,7 +10109,9 @@ std::string ini_join(const std::vector<s
                      char arrayStart = '[',
                      char arrayEnd = ']',
                      char stringQuote = '"',
-                     char characterQuote = '\'');
+                     char literalQuote = '\'');
+
+void clean_name_string(std::string &name, const std::string &keyChars);
 
 std::vector<std::string> generate_parents(const std::string &section, std::string &name, char parentSeparator);
 
@@ -9034,9 +10122,19 @@ void checkParentSegments(std::vector<Con
 
 
 
+static constexpr auto multiline_literal_quote = R"(''')";
+static constexpr auto multiline_string_quote = R"(""")";
+
 namespace detail {
 
-CLI11_INLINE std::string convert_arg_for_ini(const std::string &arg, char stringQuote, char characterQuote) {
+CLI11_INLINE bool is_printable(const std::string &test_string) {
+    return std::all_of(test_string.begin(), test_string.end(), [](char x) {
+        return (isprint(static_cast<unsigned char>(x)) != 0 || x == '\n' || x == '\t');
+    });
+}
+
+CLI11_INLINE std::string
+convert_arg_for_ini(const std::string &arg, char stringQuote, char literalQuote, bool disable_multi_line) {
     if(arg.empty()) {
         return std::string(2, stringQuote);
     }
@@ -9049,12 +10147,20 @@ CLI11_INLINE std::string convert_arg_for
         using CLI::detail::lexical_cast;
         double val = 0.0;
         if(lexical_cast(arg, val)) {
-            return arg;
+            if(arg.find_first_not_of("0123456789.-+eE") == std::string::npos) {
+                return arg;
+            }
         }
     }
     // just quote a single non numeric character
     if(arg.size() == 1) {
-        return std::string(1, characterQuote) + arg + characterQuote;
+        if(isprint(static_cast<unsigned char>(arg.front())) == 0) {
+            return binary_escape_string(arg);
+        }
+        if(arg == "'") {
+            return std::string(1, stringQuote) + "'" + stringQuote;
+        }
+        return std::string(1, literalQuote) + arg + literalQuote;
     }
     // handle hex, binary or octal arguments
     if(arg.front() == '0') {
@@ -9074,10 +10180,16 @@ CLI11_INLINE std::string convert_arg_for
             }
         }
     }
-    if(arg.find_first_of(stringQuote) == std::string::npos) {
-        return std::string(1, stringQuote) + arg + stringQuote;
+    if(!is_printable(arg)) {
+        return binary_escape_string(arg);
+    }
+    if(detail::has_escapable_character(arg)) {
+        if(arg.size() > 100 && !disable_multi_line) {
+            return std::string(multiline_literal_quote) + arg + multiline_literal_quote;
+        }
+        return std::string(1, stringQuote) + detail::add_escaped_characters(arg) + stringQuote;
     }
-    return characterQuote + arg + characterQuote;
+    return std::string(1, stringQuote) + arg + stringQuote;
 }
 
 CLI11_INLINE std::string ini_join(const std::vector<std::string> &args,
@@ -9085,10 +10197,12 @@ CLI11_INLINE std::string ini_join(const
                                   char arrayStart,
                                   char arrayEnd,
                                   char stringQuote,
-                                  char characterQuote) {
+                                  char literalQuote) {
+    bool disable_multi_line{false};
     std::string joined;
     if(args.size() > 1 && arrayStart != '\0') {
         joined.push_back(arrayStart);
+        disable_multi_line = true;
     }
     std::size_t start = 0;
     for(const auto &arg : args) {
@@ -9098,7 +10212,7 @@ CLI11_INLINE std::string ini_join(const
                 joined.push_back(' ');
             }
         }
-        joined.append(convert_arg_for_ini(arg, stringQuote, characterQuote));
+        joined.append(convert_arg_for_ini(arg, stringQuote, literalQuote, disable_multi_line));
     }
     if(args.size() > 1 && arrayEnd != '\0') {
         joined.push_back(arrayEnd);
@@ -9111,22 +10225,22 @@ generate_parents(const std::string &sect
     std::vector<std::string> parents;
     if(detail::to_lower(section) != "default") {
         if(section.find(parentSeparator) != std::string::npos) {
-            parents = detail::split(section, parentSeparator);
+            parents = detail::split_up(section, parentSeparator);
         } else {
             parents = {section};
         }
     }
     if(name.find(parentSeparator) != std::string::npos) {
-        std::vector<std::string> plist = detail::split(name, parentSeparator);
+        std::vector<std::string> plist = detail::split_up(name, parentSeparator);
         name = plist.back();
-        detail::remove_quotes(name);
         plist.pop_back();
         parents.insert(parents.end(), plist.begin(), plist.end());
     }
-
     // clean up quotes on the parents
-    for(auto &parent : parents) {
-        detail::remove_quotes(parent);
+    try {
+        detail::remove_quotes(parents);
+    } catch(const std::invalid_argument &iarg) {
+        throw CLI::ParseError(iarg.what(), CLI::ExitCodes::InvalidError);
     }
     return parents;
 }
@@ -9179,30 +10293,59 @@ checkParentSegments(std::vector<ConfigIt
     output.back().parents = std::move(parents);
     output.back().name = "++";
 }
+
+/// @brief  checks if a string represents a multiline comment
+CLI11_INLINE bool hasMLString(std::string const &fullString, char check) {
+    if(fullString.length() < 3) {
+        return false;
+    }
+    auto it = fullString.rbegin();
+    return (*it == check) && (*(it + 1) == check) && (*(it + 2) == check);
+}
 }  // namespace detail
 
 inline std::vector<ConfigItem> ConfigBase::from_config(std::istream &input) const {
     std::string line;
+    std::string buffer;
     std::string currentSection = "default";
     std::string previousSection = "default";
     std::vector<ConfigItem> output;
     bool isDefaultArray = (arrayStart == '[' && arrayEnd == ']' && arraySeparator == ',');
     bool isINIArray = (arrayStart == '\0' || arrayStart == ' ') && arrayStart == arrayEnd;
     bool inSection{false};
+    bool inMLineComment{false};
+    bool inMLineValue{false};
+
     char aStart = (isINIArray) ? '[' : arrayStart;
     char aEnd = (isINIArray) ? ']' : arrayEnd;
     char aSep = (isINIArray && arraySeparator == ' ') ? ',' : arraySeparator;
     int currentSectionIndex{0};
-    while(getline(input, line)) {
+
+    std::string line_sep_chars{parentSeparatorChar, commentChar, valueDelimiter};
+    while(getline(input, buffer)) {
         std::vector<std::string> items_buffer;
         std::string name;
-
-        detail::trim(line);
+        line = detail::trim_copy(buffer);
         std::size_t len = line.length();
         // lines have to be at least 3 characters to have any meaning to CLI just skip the rest
         if(len < 3) {
             continue;
         }
+        if(line.compare(0, 3, multiline_string_quote) == 0 || line.compare(0, 3, multiline_literal_quote) == 0) {
+            inMLineComment = true;
+            auto cchar = line.front();
+            while(inMLineComment) {
+                if(getline(input, line)) {
+                    detail::trim(line);
+                } else {
+                    break;
+                }
+                if(detail::hasMLString(line, cchar)) {
+                    inMLineComment = false;
+                }
+            }
+            continue;
+        }
         if(line.front() == '[' && line.back() == ']') {
             if(currentSection != "default") {
                 // insert a section end which is just an empty items_buffer
@@ -9234,49 +10377,130 @@ inline std::vector<ConfigItem> ConfigBas
         if(line.front() == ';' || line.front() == '#' || line.front() == commentChar) {
             continue;
         }
-
-        // Find = in string, split and recombine
-        auto pos = line.find(valueDelimiter);
-        if(pos != std::string::npos) {
-            name = detail::trim_copy(line.substr(0, pos));
-            std::string item = detail::trim_copy(line.substr(pos + 1));
-            auto cloc = item.find(commentChar);
-            if(cloc != std::string::npos) {
-                item.erase(cloc, std::string::npos);  // NOLINT(readability-suspicious-call-argument)
-                detail::trim(item);
+        std::size_t search_start = 0;
+        if(line.find_first_of("\"'`") != std::string::npos) {
+            while(search_start < line.size()) {
+                auto test_char = line[search_start];
+                if(test_char == '\"' || test_char == '\'' || test_char == '`') {
+                    search_start = detail::close_sequence(line, search_start, line[search_start]);
+                    ++search_start;
+                } else if(test_char == valueDelimiter || test_char == commentChar) {
+                    --search_start;
+                    break;
+                } else if(test_char == ' ' || test_char == '\t' || test_char == parentSeparatorChar) {
+                    ++search_start;
+                } else {
+                    search_start = line.find_first_of(line_sep_chars, search_start);
+                }
             }
-            if(item.size() > 1 && item.front() == aStart) {
+        }
+        // Find = in string, split and recombine
+        auto delimiter_pos = line.find_first_of(valueDelimiter, search_start + 1);
+        auto comment_pos = line.find_first_of(commentChar, search_start);
+        if(comment_pos < delimiter_pos) {
+            delimiter_pos = std::string::npos;
+        }
+        if(delimiter_pos != std::string::npos) {
+
+            name = detail::trim_copy(line.substr(0, delimiter_pos));
+            std::string item = detail::trim_copy(line.substr(delimiter_pos + 1, std::string::npos));
+            bool mlquote =
+                (item.compare(0, 3, multiline_literal_quote) == 0 || item.compare(0, 3, multiline_string_quote) == 0);
+            if(!mlquote && comment_pos != std::string::npos) {
+                auto citems = detail::split_up(item, commentChar);
+                item = detail::trim_copy(citems.front());
+            }
+            if(mlquote) {
+                // mutliline string
+                auto keyChar = item.front();
+                item = buffer.substr(delimiter_pos + 1, std::string::npos);
+                detail::ltrim(item);
+                item.erase(0, 3);
+                inMLineValue = true;
+                bool lineExtension{false};
+                bool firstLine = true;
+                if(!item.empty() && item.back() == '\\') {
+                    item.pop_back();
+                    lineExtension = true;
+                }
+                while(inMLineValue) {
+                    std::string l2;
+                    if(!std::getline(input, l2)) {
+                        break;
+                    }
+                    line = l2;
+                    detail::rtrim(line);
+                    if(detail::hasMLString(line, keyChar)) {
+                        line.pop_back();
+                        line.pop_back();
+                        line.pop_back();
+                        if(lineExtension) {
+                            detail::ltrim(line);
+                        } else if(!(firstLine && item.empty())) {
+                            item.push_back('\n');
+                        }
+                        firstLine = false;
+                        item += line;
+                        inMLineValue = false;
+                        if(!item.empty() && item.back() == '\n') {
+                            item.pop_back();
+                        }
+                        if(keyChar == '\"') {
+                            try {
+                                item = detail::remove_escaped_characters(item);
+                            } catch(const std::invalid_argument &iarg) {
+                                throw CLI::ParseError(iarg.what(), CLI::ExitCodes::InvalidError);
+                            }
+                        }
+                    } else {
+                        if(lineExtension) {
+                            detail::trim(l2);
+                        } else if(!(firstLine && item.empty())) {
+                            item.push_back('\n');
+                        }
+                        lineExtension = false;
+                        firstLine = false;
+                        if(!l2.empty() && l2.back() == '\\') {
+                            lineExtension = true;
+                            l2.pop_back();
+                        }
+                        item += l2;
+                    }
+                }
+                items_buffer = {item};
+            } else if(item.size() > 1 && item.front() == aStart) {
                 for(std::string multiline; item.back() != aEnd && std::getline(input, multiline);) {
                     detail::trim(multiline);
                     item += multiline;
                 }
-                items_buffer = detail::split_up(item.substr(1, item.length() - 2), aSep);
+                if(item.back() == aEnd) {
+                    items_buffer = detail::split_up(item.substr(1, item.length() - 2), aSep);
+                } else {
+                    items_buffer = detail::split_up(item.substr(1, std::string::npos), aSep);
+                }
             } else if((isDefaultArray || isINIArray) && item.find_first_of(aSep) != std::string::npos) {
                 items_buffer = detail::split_up(item, aSep);
             } else if((isDefaultArray || isINIArray) && item.find_first_of(' ') != std::string::npos) {
-                items_buffer = detail::split_up(item);
+                items_buffer = detail::split_up(item, '\0');
             } else {
                 items_buffer = {item};
             }
         } else {
-            name = detail::trim_copy(line);
-            auto cloc = name.find(commentChar);
-            if(cloc != std::string::npos) {
-                name.erase(cloc, std::string::npos);  // NOLINT(readability-suspicious-call-argument)
-                detail::trim(name);
-            }
-
+            name = detail::trim_copy(line.substr(0, comment_pos));
             items_buffer = {"true"};
         }
-        if(name.find(parentSeparatorChar) == std::string::npos) {
-            detail::remove_quotes(name);
-        }
-        // clean up quotes on the items
-        for(auto &it : items_buffer) {
-            detail::remove_quotes(it);
+        std::vector<std::string> parents;
+        try {
+            parents = detail::generate_parents(currentSection, name, parentSeparatorChar);
+            detail::process_quoted_string(name);
+            // clean up quotes on the items and check for escaped strings
+            for(auto &it : items_buffer) {
+                detail::process_quoted_string(it, stringQuote, literalQuote);
+            }
+        } catch(const std::invalid_argument &ia) {
+            throw CLI::ParseError(ia.what(), CLI::ExitCodes::InvalidError);
         }
 
-        std::vector<std::string> parents = detail::generate_parents(currentSection, name, parentSeparatorChar);
         if(parents.size() > maximumLayers) {
             continue;
         }
@@ -9313,6 +10537,23 @@ inline std::vector<ConfigItem> ConfigBas
     return output;
 }
 
+CLI11_INLINE std::string &clean_name_string(std::string &name, const std::string &keyChars) {
+    if(name.find_first_of(keyChars) != std::string::npos || (name.front() == '[' && name.back() == ']') ||
+       (name.find_first_of("'`\"\\") != std::string::npos)) {
+        if(name.find_first_of('\'') == std::string::npos) {
+            name.insert(0, 1, '\'');
+            name.push_back('\'');
+        } else {
+            if(detail::has_escapable_character(name)) {
+                name = detail::add_escaped_characters(name);
+            }
+            name.insert(0, 1, '\"');
+            name.push_back('\"');
+        }
+    }
+    return name;
+}
+
 CLI11_INLINE std::string
 ConfigBase::to_config(const App *app, bool default_also, bool write_description, std::string prefix) const {
     std::stringstream out;
@@ -9320,6 +10561,18 @@ ConfigBase::to_config(const App *app, bo
     commentLead.push_back(commentChar);
     commentLead.push_back(' ');
 
+    std::string commentTest = "#;";
+    commentTest.push_back(commentChar);
+    commentTest.push_back(parentSeparatorChar);
+
+    std::string keyChars = commentTest;
+    keyChars.push_back(literalQuote);
+    keyChars.push_back(stringQuote);
+    keyChars.push_back(arrayStart);
+    keyChars.push_back(arrayEnd);
+    keyChars.push_back(valueDelimiter);
+    keyChars.push_back(arraySeparator);
+
     std::vector<std::string> groups = app->get_groups();
     bool defaultUsed = false;
     groups.insert(groups.begin(), std::string("Options"));
@@ -9345,13 +10598,17 @@ ConfigBase::to_config(const App *app, bo
                         continue;
                     }
                 }
-                std::string name = prefix + opt->get_single_name();
+                std::string single_name = opt->get_single_name();
+                if(single_name.empty()) {
+                    continue;
+                }
+
                 std::string value = detail::ini_join(
-                    opt->reduced_results(), arraySeparator, arrayStart, arrayEnd, stringQuote, characterQuote);
+                    opt->reduced_results(), arraySeparator, arrayStart, arrayEnd, stringQuote, literalQuote);
 
                 if(value.empty() && default_also) {
                     if(!opt->get_default_str().empty()) {
-                        value = detail::convert_arg_for_ini(opt->get_default_str(), stringQuote, characterQuote);
+                        value = detail::convert_arg_for_ini(opt->get_default_str(), stringQuote, literalQuote, false);
                     } else if(opt->get_expected_min() == 0) {
                         value = "false";
                     } else if(opt->get_run_callback_for_default()) {
@@ -9360,13 +10617,35 @@ ConfigBase::to_config(const App *app, bo
                 }
 
                 if(!value.empty()) {
+
                     if(!opt->get_fnames().empty()) {
-                        value = opt->get_flag_value(name, value);
+                        try {
+                            value = opt->get_flag_value(single_name, value);
+                        } catch(const CLI::ArgumentMismatch &) {
+                            bool valid{false};
+                            for(const auto &test_name : opt->get_fnames()) {
+                                try {
+                                    value = opt->get_flag_value(test_name, value);
+                                    single_name = test_name;
+                                    valid = true;
+                                } catch(const CLI::ArgumentMismatch &) {
+                                    continue;
+                                }
+                            }
+                            if(!valid) {
+                                value = detail::ini_join(
+                                    opt->results(), arraySeparator, arrayStart, arrayEnd, stringQuote, literalQuote);
+                            }
+                        }
                     }
                     if(write_description && opt->has_description()) {
                         out << '\n';
                         out << commentLead << detail::fix_newlines(commentLead, opt->get_description()) << '\n';
                     }
+                    clean_name_string(single_name, keyChars);
+
+                    std::string name = prefix + single_name;
+
                     out << name << valueDelimiter << value << '\n';
                 }
             }
@@ -9375,31 +10654,56 @@ ConfigBase::to_config(const App *app, bo
     auto subcommands = app->get_subcommands({});
     for(const App *subcom : subcommands) {
         if(subcom->get_name().empty()) {
+            if(!default_also && (subcom->count_all() == 0)) {
+                continue;
+            }
             if(write_description && !subcom->get_group().empty()) {
                 out << '\n' << commentLead << subcom->get_group() << " Options\n";
             }
+            /*if (!prefix.empty() || app->get_parent() == nullptr) {
+                out << '[' << prefix << "___"<< subcom->get_group() << "]\n";
+            } else {
+                std::string subname = app->get_name() + parentSeparatorChar + "___"+subcom->get_group();
+                const auto *p = app->get_parent();
+                while(p->get_parent() != nullptr) {
+                    subname = p->get_name() + parentSeparatorChar +subname;
+                    p = p->get_parent();
+                }
+                out << '[' << subname << "]\n";
+            }
+            */
             out << to_config(subcom, default_also, write_description, prefix);
         }
     }
 
     for(const App *subcom : subcommands) {
         if(!subcom->get_name().empty()) {
+            if(!default_also && (subcom->count_all() == 0)) {
+                continue;
+            }
+            std::string subname = subcom->get_name();
+            clean_name_string(subname, keyChars);
+
             if(subcom->get_configurable() && app->got_subcommand(subcom)) {
                 if(!prefix.empty() || app->get_parent() == nullptr) {
-                    out << '[' << prefix << subcom->get_name() << "]\n";
+
+                    out << '[' << prefix << subname << "]\n";
                 } else {
-                    std::string subname = app->get_name() + parentSeparatorChar + subcom->get_name();
+                    std::string appname = app->get_name();
+                    clean_name_string(appname, keyChars);
+                    subname = appname + parentSeparatorChar + subname;
                     const auto *p = app->get_parent();
                     while(p->get_parent() != nullptr) {
-                        subname = p->get_name() + parentSeparatorChar + subname;
+                        std::string pname = p->get_name();
+                        clean_name_string(pname, keyChars);
+                        subname = pname + parentSeparatorChar + subname;
                         p = p->get_parent();
                     }
                     out << '[' << subname << "]\n";
                 }
                 out << to_config(subcom, default_also, write_description, "");
             } else {
-                out << to_config(
-                    subcom, default_also, write_description, prefix + subcom->get_name() + parentSeparatorChar);
+                out << to_config(subcom, default_also, write_description, prefix + subname + parentSeparatorChar);
             }
         }
     }
@@ -9463,7 +10767,7 @@ CLI11_INLINE std::string Formatter::make
     auto min_options = app->get_require_option_min();
     auto max_options = app->get_require_option_max();
     if(app->get_required()) {
-        desc += " REQUIRED ";
+        desc += " " + get_label("REQUIRED") + " ";
     }
     if((max_options == min_options) && (min_options > 0)) {
         if(min_options == 1) {
@@ -9485,6 +10789,11 @@ CLI11_INLINE std::string Formatter::make
 }
 
 CLI11_INLINE std::string Formatter::make_usage(const App *app, std::string name) const {
+    std::string usage = app->get_usage();
+    if(!usage.empty()) {
+        return usage + "\n";
+    }
+
     std::stringstream out;
 
     out << get_label("Usage") << ":" << (name.empty() ? "" : " ") << name;
@@ -9521,7 +10830,7 @@ CLI11_INLINE std::string Formatter::make
             << (app->get_require_subcommand_min() == 0 ? "]" : "");
     }
 
-    out << std::endl;
+    out << '\n';
 
     return out.str();
 }
@@ -9602,7 +10911,10 @@ CLI11_INLINE std::string Formatter::make
 
 CLI11_INLINE std::string Formatter::make_subcommand(const App *sub) const {
     std::stringstream out;
-    detail::format_help(out, sub->get_display_name(true), sub->get_description(), column_width_);
+    detail::format_help(out,
+                        sub->get_display_name(true) + (sub->get_required() ? " " + get_label("REQUIRED") : ""),
+                        sub->get_description(),
+                        column_width_);
     return out.str();
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/go/default_param_impl.hpp 4.4.0-1/src/mlpack/bindings/go/default_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/go/default_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/default_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ std::string DefaultParamImpl(
   if (std::is_same<T, bool>::value)
     oss << "false";
   else
-    oss << MLPACK_ANY_CAST<T>(data.value);
+    oss << std::any_cast<T>(data.value);
 
   return oss.str();
 }
@@ -51,7 +51,7 @@ std::string DefaultParamImpl(
 {
   // Print each element in an array delimited by square brackets.
   std::ostringstream oss;
-  const T& vector = MLPACK_ANY_CAST<T>(data.value);
+  const T& vector = std::any_cast<T>(data.value);
   if (std::is_same<T, std::vector<std::string>>::value)
   {
     oss << "[]string{";
@@ -93,7 +93,7 @@ std::string DefaultParamImpl(
     util::ParamData& data,
     const typename std::enable_if<std::is_same<T, std::string>::value>::type*)
 {
-  const std::string& s = *MLPACK_ANY_CAST<std::string>(&data.value);
+  const std::string& s = *std::any_cast<std::string>(&data.value);
   return "\"" + s + "\"";
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/go/get_param.hpp 4.4.0-1/src/mlpack/bindings/go/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/go/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ void GetParam(util::ParamData& d,
               const void* /* input */,
               void* output)
 {
-  *((T**) output) = const_cast<T*>(MLPACK_ANY_CAST<T>(&d.value));
+  *((T**) output) = const_cast<T*>(std::any_cast<T>(&d.value));
 }
 
 } // namespace go
diff -pruN 4.3.0-2/src/mlpack/bindings/go/get_printable_param.hpp 4.4.0-1/src/mlpack/bindings/go/get_printable_param.hpp
--- 4.3.0-2/src/mlpack/bindings/go/get_printable_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/get_printable_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -44,7 +44,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* = 0)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -61,7 +61,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& matrix = MLPACK_ANY_CAST<T>(data.value);
+  const T& matrix = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   oss << matrix.n_rows << "x" << matrix.n_cols << " matrix";
@@ -78,7 +78,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << data.cppType << " model at " << MLPACK_ANY_CAST<T*>(data.value);
+  oss << data.cppType << " model at " << std::any_cast<T*>(data.value);
   return oss.str();
 }
 
@@ -92,7 +92,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& tuple = MLPACK_ANY_CAST<T>(data.value);
+  const T& tuple = std::any_cast<T>(data.value);
   const arma::mat& matrix = std::get<1>(tuple);
 
   std::ostringstream oss;
diff -pruN 4.3.0-2/src/mlpack/bindings/go/print_doc.hpp 4.4.0-1/src/mlpack/bindings/go/print_doc.hpp
--- 4.3.0-2/src/mlpack/bindings/go/print_doc.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/print_doc.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,16 +53,16 @@ void PrintDoc(util::ParamData& d,
   {
     if (d.cppType == "std::string")
     {
-      oss << "  Default value '" << MLPACK_ANY_CAST<std::string>(d.value)
+      oss << "  Default value '" << std::any_cast<std::string>(d.value)
           << "'.";
     }
     else if (d.cppType == "double")
     {
-      oss << "  Default value " << MLPACK_ANY_CAST<double>(d.value) << ".";
+      oss << "  Default value " << std::any_cast<double>(d.value) << ".";
     }
     else if (d.cppType == "int")
     {
-      oss << "  Default value " << MLPACK_ANY_CAST<int>(d.value) << ".";
+      oss << "  Default value " << std::any_cast<int>(d.value) << ".";
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/go/print_input_processing.hpp 4.4.0-1/src/mlpack/bindings/go/print_input_processing.hpp
--- 4.3.0-2/src/mlpack/bindings/go/print_input_processing.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/print_input_processing.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,22 +67,22 @@ void PrintInputProcessing(
     // Print out default value.
     if (d.cppType == "std::string")
     {
-      std::string value = MLPACK_ANY_CAST<std::string>(d.value);
+      std::string value = std::any_cast<std::string>(d.value);
       std::cout << "\"" << value << "\"";
     }
     else if (d.cppType == "double")
     {
-      double value = MLPACK_ANY_CAST<double>(d.value);
+      double value = std::any_cast<double>(d.value);
       std::cout << value;
     }
     else if (d.cppType == "int")
     {
-      int value = MLPACK_ANY_CAST<int>(d.value);
+      int value = std::any_cast<int>(d.value);
       std::cout << value;
     }
     else if (d.cppType == "bool")
     {
-      bool value = MLPACK_ANY_CAST<bool>(d.value);
+      bool value = std::any_cast<bool>(d.value);
       if (value == 0)
       std::cout << "false";
       else
diff -pruN 4.3.0-2/src/mlpack/bindings/go/print_method_init.hpp 4.4.0-1/src/mlpack/bindings/go/print_method_init.hpp
--- 4.3.0-2/src/mlpack/bindings/go/print_method_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/go/print_method_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -54,23 +54,23 @@ void PrintMethodInit(
   {
     if (d.cppType == "std::string")
     {
-      std::string value = MLPACK_ANY_CAST<std::string>(d.value);
+      std::string value = std::any_cast<std::string>(d.value);
       std::cout << prefix << goParamName << ": \""
                 << value << "\"," << std::endl;
     }
     else if (d.cppType == "double")
     {
-      double value = MLPACK_ANY_CAST<double>(d.value);
+      double value = std::any_cast<double>(d.value);
       std::cout << prefix << goParamName << ": " << value << "," << std::endl;
     }
     else if (d.cppType == "int")
     {
-      int value = MLPACK_ANY_CAST<int>(d.value);
+      int value = std::any_cast<int>(d.value);
       std::cout << prefix << goParamName << ": " << value << "," << std::endl;
     }
     else if (d.cppType == "bool")
     {
-      bool value = MLPACK_ANY_CAST<bool>(d.value);
+      bool value = std::any_cast<bool>(d.value);
       if (value == 0)
         std::cout << prefix << goParamName << ": false," << std::endl;
       else
diff -pruN 4.3.0-2/src/mlpack/bindings/julia/default_param_impl.hpp 4.4.0-1/src/mlpack/bindings/julia/default_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/julia/default_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/julia/default_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ std::string DefaultParamImpl(
   if (std::is_same<T, bool>::value)
     oss << "false";
   else
-    oss << MLPACK_ANY_CAST<T>(data.value);
+    oss << std::any_cast<T>(data.value);
 
   return oss.str();
 }
@@ -51,7 +51,7 @@ std::string DefaultParamImpl(
 {
   // Print each element in an array delimited by square brackets.
   std::ostringstream oss;
-  const T& vector = MLPACK_ANY_CAST<T>(data.value);
+  const T& vector = std::any_cast<T>(data.value);
   oss << "[";
   if (std::is_same<T, std::vector<std::string>>::value)
   {
@@ -92,7 +92,7 @@ std::string DefaultParamImpl(
     util::ParamData& data,
     const typename std::enable_if<std::is_same<T, std::string>::value>::type*)
 {
-  const std::string& s = *MLPACK_ANY_CAST<std::string>(&data.value);
+  const std::string& s = *std::any_cast<std::string>(&data.value);
   return "\"" + s + "\"";
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/julia/get_param.hpp 4.4.0-1/src/mlpack/bindings/julia/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/julia/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/julia/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ void GetParam(util::ParamData& d,
               const void* /* input */,
               void* output)
 {
-  *((T**) output) = const_cast<T*>(MLPACK_ANY_CAST<T>(&d.value));
+  *((T**) output) = const_cast<T*>(std::any_cast<T>(&d.value));
 }
 
 } // namespace julia
diff -pruN 4.3.0-2/src/mlpack/bindings/julia/get_printable_param.hpp 4.4.0-1/src/mlpack/bindings/julia/get_printable_param.hpp
--- 4.3.0-2/src/mlpack/bindings/julia/get_printable_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/julia/get_printable_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -44,7 +44,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* = 0)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -61,7 +61,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& matrix = MLPACK_ANY_CAST<T>(data.value);
+  const T& matrix = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   oss << matrix.n_rows << "x" << matrix.n_cols << " matrix";
@@ -78,7 +78,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << data.cppType << " model at " << MLPACK_ANY_CAST<T*>(data.value);
+  oss << data.cppType << " model at " << std::any_cast<T*>(data.value);
   return oss.str();
 }
 
@@ -92,7 +92,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& tuple = MLPACK_ANY_CAST<T>(data.value);
+  const T& tuple = std::any_cast<T>(data.value);
   const arma::mat& matrix = std::get<1>(tuple);
 
   std::ostringstream oss;
diff -pruN 4.3.0-2/src/mlpack/bindings/julia/print_doc.hpp 4.4.0-1/src/mlpack/bindings/julia/print_doc.hpp
--- 4.3.0-2/src/mlpack/bindings/julia/print_doc.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/julia/print_doc.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -39,19 +39,19 @@ void PrintDoc(util::ParamData& d, const
       oss << "  Default value `";
       if (d.cppType == "std::string")
       {
-        oss << MLPACK_ANY_CAST<std::string>(d.value);
+        oss << std::any_cast<std::string>(d.value);
       }
       else if (d.cppType == "double")
       {
-        oss << MLPACK_ANY_CAST<double>(d.value);
+        oss << std::any_cast<double>(d.value);
       }
       else if (d.cppType == "int")
       {
-        oss << MLPACK_ANY_CAST<int>(d.value);
+        oss << std::any_cast<int>(d.value);
       }
       else if (d.cppType == "bool")
       {
-        oss << (MLPACK_ANY_CAST<bool>(d.value) ? "true" : "false");
+        oss << (std::any_cast<bool>(d.value) ? "true" : "false");
       }
       oss << "`." << std::endl;
     }
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/CMakeLists.txt 4.4.0-1/src/mlpack/bindings/markdown/CMakeLists.txt
--- 4.3.0-2/src/mlpack/bindings/markdown/CMakeLists.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/CMakeLists.txt	2024-05-28 20:08:34.000000000 +0000
@@ -106,7 +106,7 @@ macro (add_markdown_docs directory name
     set(WRAPPER_PUSH_BACK_CODE
         "${WRAPPER_PUSH_BACK_CODE}\n  addWrapperDocs.push_back(${val});")
   endforeach()
-  
+
   if(NOT ${WRAPPERS} EQUAL "")
     foreach(method ${${name}_VALID_METHODS})
       set(VALID_METHODS_PUSH_BACK_CODE
@@ -163,10 +163,10 @@ function (post_markdown_setup)
   math(EXPR cat_limit "${NUM_MARKDOWN_CATEGORIES} - 1")
   foreach (i RANGE ${cat_limit})
     list (GET CATEGORIES ${i} cat)
-    # Put the things in this category in a div.
-    string(CONCAT header_code "${MARKDOWN_HEADER_CODE}  cout << "
-        "\"<div class=\\\"category\\\" markdown=\\\"1\\\">\" << endl;\n  "
-        "cout << \"<h5>${cat}:</h5>\" << endl;\n")
+    # Put the things in this category in a details block.
+    string(CONCAT header_code "${MARKDOWN_HEADER_CODE}"
+        "  cout << \"<li><details><summary>${cat}</summary>\" << endl;\n"
+        "  cout << \"<ul>\" << endl;\n")
     set(MARKDOWN_HEADER_CODE ${header_code})
 
     # Add an option for this binding.
@@ -183,7 +183,11 @@ function (post_markdown_setup)
     endforeach ()
 
     set (MARKDOWN_HEADER_CODE
-        "${MARKDOWN_HEADER_CODE}\n  cout << \"</div>\" << endl << endl;")
+        "${MARKDOWN_HEADER_CODE}\n"
+        "  cout << \"</ul>\" << endl;\n"
+        "  cout << \"</details>\" << endl;\n"
+        "  cout << \"</li>\" << endl;\n"
+        "  cout << endl;\n\n")
   endforeach ()
 
   foreach (name ${MARKDOWN_NAMES})
@@ -218,30 +222,35 @@ function (post_markdown_setup)
       "${BINDING_SOURCE_DIR}/print_param_table.hpp"
       "${BINDING_SOURCE_DIR}/print_param_table.cpp"
       "${BINDING_SOURCE_DIR}/program_doc_wrapper.hpp"
-      "${BINDING_SOURCE_DIR}/replace_all_copy.hpp"
-      "${BINDING_SOURCE_DIR}/generate_markdown.cpp")
+      "${BINDING_SOURCE_DIR}/replace_all_copy.hpp")
   target_link_libraries(generate_markdown ${MLPACK_LIBRARIES})
   add_dependencies(generate_markdown markdown_copy)
   set_target_properties(generate_markdown PROPERTIES
       COMPILE_FLAGS -DBINDING_TYPE=BINDING_TYPE_MARKDOWN
       RUNTIME_OUTPUT_DIRECTORY ${BINDING_BINARY_DIR})
 
-  add_custom_target(markdown ALL
+  add_custom_target(markdown_setup ALL
       ${CMAKE_COMMAND} -E make_directory
           ${CMAKE_BINARY_DIR}/doc/
-      COMMAND ${CMAKE_COMMAND}
-          -DPROGRAM=${BINDING_BINARY_DIR}/generate_markdown
-          -DOUTPUT_FILE=${CMAKE_BINARY_DIR}/doc/mlpack.md
-          -P ${CMAKE_SOURCE_DIR}/CMake/RunProgram.cmake
-      COMMAND ${CMAKE_COMMAND} -E copy
-          ${BINDING_SOURCE_DIR}/res/change_language.js
-          ${CMAKE_BINARY_DIR}/doc/res/change_languages.js
-      COMMAND ${CMAKE_COMMAND} -E copy
-          ${BINDING_SOURCE_DIR}/res/menu_bg.png
-          ${CMAKE_BINARY_DIR}/doc/res/menu_bg.png
-      COMMAND ${CMAKE_COMMAND} -E copy
-          ${BINDING_SOURCE_DIR}/res/formatting.css
-          ${CMAKE_BINARY_DIR}/doc/res/formatting.css
       DEPENDS generate_markdown
-      COMMENT "Generating Markdown documentation for mlpack bindings...")
+      COMMENT "Setting up environment for Markdown documentation build...")
+  add_custom_target(markdown DEPENDS markdown_setup)
+  foreach (lang ${MARKDOWN_ALL_LANGUAGES_LIST})
+    add_custom_target(markdown_${lang} ALL
+        COMMAND ${CMAKE_COMMAND}
+            -DPROGRAM=${BINDING_BINARY_DIR}/generate_markdown
+            -DPROGRAM_OPT1="${lang}"
+            -DPROGRAM_OPT2="docs"
+            -DOUTPUT_FILE=${CMAKE_BINARY_DIR}/doc/${lang}.md
+            -P ${CMAKE_SOURCE_DIR}/CMake/RunProgram.cmake
+        COMMAND ${CMAKE_COMMAND}
+            -DPROGRAM=${BINDING_BINARY_DIR}/generate_markdown
+            -DPROGRAM_OPT1="${lang}"
+            -DPROGRAM_OPT2="sidebar"
+            -DOUTPUT_FILE=${CMAKE_BINARY_DIR}/doc/${lang}.sidebar.html
+            -P ${CMAKE_SOURCE_DIR}/CMake/RunProgram.cmake
+        COMMENT "Building Markdown documentation for ${lang}..."
+        DEPENDS markdown_setup)
+    add_dependencies(markdown markdown_${lang})
+  endforeach ()
 endfunction ()
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/generate_markdown.binding.cpp.in 4.4.0-1/src/mlpack/bindings/markdown/generate_markdown.binding.cpp.in
--- 4.3.0-2/src/mlpack/bindings/markdown/generate_markdown.binding.cpp.in	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/generate_markdown.binding.cpp.in	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ namespace markdown {
 
 void Print${BINDING}Headers()
 {
-  // Fill the vector of languages for which we want to print.
+  // Stores the names of languages this binding works with.
   vector<string> languages;
 
   // Stores if a wrapper is added for a particular language.
@@ -35,12 +35,27 @@ void Print${BINDING}Headers()
   ${LANGUAGES_PUSH_BACK_CODE}
   ${WRAPPER_PUSH_BACK_CODE}
 
-  PrintHeaders("${BINDING}", languages, addWrapperDocs);
+  // Find whether or not we print anything for the current language.
+  size_t langIndex = languages.size();
+  for (size_t i = 0; i < languages.size(); ++i)
+  {
+    if (languages[i] == BindingInfo::Language())
+    {
+      langIndex = i;
+      break;
+    }
+  }
+
+  // Don't do anything if we didn't find the language.
+  if (langIndex == languages.size())
+    return;
+
+  PrintHeaders("${BINDING}", addWrapperDocs[langIndex]);
 }
 
 void Print${BINDING}Docs()
 {
-  // Fill the vector of languages for which we want to print.
+  // Stores the names of languages this binding works with.
   vector<string> languages;
 
   // Stores the names of all methods in a vector.
@@ -53,7 +68,22 @@ void Print${BINDING}Docs()
   ${WRAPPER_PUSH_BACK_CODE}
   ${VALID_METHODS_PUSH_BACK_CODE}
 
-  PrintDocs("${BINDING}", languages, validMethods, addWrapperDocs);
+  // Find whether or not we print anything for the current language.
+  size_t langIndex = languages.size();
+  for (size_t i = 0; i < languages.size(); ++i)
+  {
+    if (languages[i] == BindingInfo::Language())
+    {
+      langIndex = i;
+      break;
+    }
+  }
+
+  // Don't do anything if we didn't find the language.
+  if (langIndex == languages.size())
+    return;
+
+  PrintDocs("${BINDING}", validMethods, addWrapperDocs[langIndex]);
 }
 
 } // namespace markdown
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/generate_markdown.cpp.in 4.4.0-1/src/mlpack/bindings/markdown/generate_markdown.cpp.in
--- 4.3.0-2/src/mlpack/bindings/markdown/generate_markdown.cpp.in	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/generate_markdown.cpp.in	2024-05-28 20:08:34.000000000 +0000
@@ -20,100 +20,100 @@ using namespace mlpack::bindings;
 using namespace mlpack::bindings::markdown;
 using namespace std;
 
-int main()
+void usage(const char* name)
 {
-  // These come to use from CMake separated by semicolons.
-  string languageList = "${MARKDOWN_ALL_LANGUAGES_LIST}";
-  vector<string> languages;
-  size_t index;
-  while ((index = languageList.find(';')) != string::npos)
-  {
-    languages.push_back(languageList.substr(0, index));
-    languageList = languageList.substr(index + 1);
-  }
-  languages.push_back(languageList);
+  cerr << "Usage: " << name << " <language> <mode>" << endl;
+  cerr << "  <language> must be one of ${MARKDOWN_ALL_LANGUAGES_LIST}" << endl;
+  cerr << "  <mode> must be 'docs' or 'sidebar'" << endl;
+  exit(1);
+}
 
-  cout << "<div id=\"header\" markdown=\"1\">" << endl;
+int main(int argc, char** argv)
+{
+  // Sanity check: make sure the user told us what language they want to print
+  // docs for.
+  if (argc != 3)
+    usage(argv[0]);
+
+  // This will throw an exception if the language is not known later on.
+  BindingInfo::Language() = std::string(argv[1]);
+  const std::string& mode = std::string(argv[2]);
 
-  // We need to create the input form selector for the language.
-  cout << "<form>" << endl;
-  cout << "  <div class=\"language-select-div\">" << endl;
-  cout << "  <select id=\"language-select\" onchange=\"changeLanguage()\">"
-      << endl;
-
-  // Print the first manually, because it is selected.
-  cout << "    <option value=\"" << languages[0] << "\" selected>"
-      << languages[0] << "</option>" << endl;
-  for (size_t i = 1; i < languages.size(); ++i)
+  if (mode == "sidebar")
   {
-    cout << "    <option value=\"" << languages[i] << "\">" << languages[i]
-        << "</option>" << endl;
+    cout << "<ul>" << endl;
+    cout << "<li class=\"page_title\">" << endl;
+    cout << "<b>mlpack " << PrintLanguage(BindingInfo::Language())
+        << " binding documentation</b>" << endl;
+    cout << "<a href=\"#\">[top]</a>" << endl;
+    cout << "<ul>" << endl;
+
+    // The links to the data type sections get put here.
+    cout << "<li><a href=\"LINKROOTuser/bindings/" << BindingInfo::Language()
+        << ".html#mlpack-overview\">Overview</a></li>" << endl;
+    cout << "<li><a href=\"LINKROOTquickstart/" << BindingInfo::Language()
+        << ".html\">" << PrintLanguage(BindingInfo::Language())
+        << " Quickstart</a></li>" << endl;
+    cout << "<li><a href=\"LINKROOTuser/bindings/" << BindingInfo::Language()
+        << ".html#data-formats\">Data Formats</a></li>" << endl;
+
+    ${MARKDOWN_HEADER_CODE}
+
+    cout << "</ul>" << endl;
+    cout << "</li>" << endl;
+    cout << "</ul>" << endl;
+    cout << "</div>" << endl;
   }
-
-  cout << "  </select>" << endl;
-  cout << "  </div>" << endl;
-  cout << "</form>" << endl;
-  cout << endl;
-
-  // The links to the data type sections get put here.
-  cout << " - [mlpack overview](#mlpack-overview){: .language-link #always }"
-      << endl;
-  for (size_t i = 0; i < languages.size(); ++i)
+  else if (mode == "docs")
   {
-    cout << " - [data formats](#" << languages[i] << "_data-formats){: "
-        << ".language-link #" << languages[i] << " }" << endl;
-  }
-
-  ${MARKDOWN_HEADER_CODE}
-
-  cout << endl << "</div>" << endl << endl;
+    // Create the header for the given language.
+    cout << "# mlpack " << PrintLanguage(BindingInfo::Language())
+        << " binding documentation" << endl;
+    cout << endl;
 
-  cout << "<div id=\"docs\" markdown=\"1\">" << endl;
-  cout << endl;
+    /**
+     * "mlpack overview" section.  This will go at the top of the page.
+     */
+    cout << "## mlpack overview" << endl;
+    cout << endl;
+    cout << "mlpack is an intuitive, fast, and flexible header-only C++ machine"
+        << " learning library with bindings to other languages.  It aims to "
+        << "provide fast, lightweight implementations of both common and "
+        << "cutting-edge machine learning algorithms." << endl;
+    cout << endl;
+    cout << "This reference page details mlpack's bindings to ";
+    if (BindingInfo::Language() == "cli")
+      cout << "the command-line";
+    else
+      cout << PrintLanguage(BindingInfo::Language());
+    cout << "." << endl;
+    cout << endl;
+    cout << "Further useful mlpack documentation links are given below.";
+    cout << endl << endl;
+    cout << " - [mlpack homepage](https://www.mlpack.org/)" << endl;
+    cout << " - [mlpack on Github](https://github.com/mlpack/mlpack)" << endl;
+    cout << " - [mlpack main documentation page]"
+        << "(https://www.mlpack.org/docs.html)" << endl;
+    cout << endl;
+    cout << "See also the quickstart guide for "
+        << PrintLanguage(BindingInfo::Language()) << ":" << endl;
+    cout << endl;
+    cout << " - [" << PrintLanguage(BindingInfo::Language())
+        << " Quickstart](../../quickstart/" << BindingInfo::Language() << ".md)"
+        << endl;
+    cout << endl;
+
+    /**
+     * Discussion of different data types section.  This goes just below the
+     * overview section at the top of the page.
+     */
+    cout << PrintTypeDocs() << endl;
 
-  // Create all the headers for each language.
-  for (size_t i = 0; i < languages.size(); ++i)
-  {
-    cout << "<div id=\"" << languages[i] << "\" class=\"language-header\" "
-        << "markdown=\"1\">" << endl;
-    cout << "# " << util::GetVersion() << " " << PrintLanguage(languages[i])
-        << " binding documentation" << endl;
-    cout << "</div>" << endl;
+    ${MARKDOWN_CALL_CODE}
   }
-
-  /**
-   * "mlpack overview" section.  This will go at the top of the page.
-   */
-  cout << "## mlpack overview" << endl;
-  cout << endl;
-  cout << "mlpack is an intuitive, fast, and flexible C++ machine learning "
-      "library with bindings to other languages.  It is meant to be a machine "
-      "learning analog to LAPACK, and aims to implement a wide array of machine"
-      " learning methods and functions as a \"swiss army knife\" for machine "
-      "learning researchers.";
-  cout << endl << endl;
-  cout << "This reference page details mlpack's bindings to other languages. "
-      "Further useful mlpack documentation links are given below.";
-  cout << endl << endl;
-  cout << " - [mlpack homepage](https://www.mlpack.org/)" << endl;
-  cout << " - [mlpack on Github](https://github.com/mlpack/mlpack)" << endl;
-  cout << " - [mlpack main documentation page]"
-      << "(https://www.mlpack.org/docs.html)" << endl;
-  cout << endl;
-
-  /**
-   * Discussion of different data types section.  This goes just below the
-   * overview section at the top of the page.
-   */
-  for (size_t i = 0; i < languages.size(); ++i)
+  else
   {
-    BindingInfo::Language() = languages[i];
-    cout << PrintTypeDocs() << endl;
+    // Unknown mode.
+    usage(argv[0]);
   }
-
-  ${MARKDOWN_CALL_CODE}
-  cout << "</div>" << endl << endl;
-
-  // Make sure script gets included for changeLanguage().
-  cout << "<script src=\"res/change_language.js\"></script>" << endl;
 }
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/get_param.hpp 4.4.0-1/src/mlpack/bindings/markdown/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/markdown/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -28,7 +28,7 @@ void GetParam(util::ParamData& d,
               void* output)
 {
   util::ParamData& dmod = const_cast<util::ParamData&>(d);
-  *((T**) output) = MLPACK_ANY_CAST<T>(&dmod.value);
+  *((T**) output) = std::any_cast<T>(&dmod.value);
 }
 
 } // namespace markdown
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/get_printable_param.hpp 4.4.0-1/src/mlpack/bindings/markdown/get_printable_param.hpp
--- 4.3.0-2/src/mlpack/bindings/markdown/get_printable_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/get_printable_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -44,7 +44,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* = 0)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -61,7 +61,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& matrix = MLPACK_ANY_CAST<T>(data.value);
+  const T& matrix = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   oss << matrix.n_rows << "x" << matrix.n_cols << " matrix";
@@ -78,7 +78,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << data.cppType << " model at " << MLPACK_ANY_CAST<T*>(data.value);
+  oss << data.cppType << " model at " << std::any_cast<T*>(data.value);
   return oss.str();
 }
 
@@ -92,7 +92,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& tuple = MLPACK_ANY_CAST<T>(data.value);
+  const T& tuple = std::any_cast<T>(data.value);
   const arma::mat& matrix = std::get<1>(tuple);
 
   std::ostringstream oss;
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/print_doc_functions_impl.hpp 4.4.0-1/src/mlpack/bindings/markdown/print_doc_functions_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/markdown/print_doc_functions_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/print_doc_functions_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -107,7 +107,7 @@ inline std::string PrintLanguage(const s
   else
   {
     throw std::invalid_argument("PrintLanguage(): unknown "
-        "BindingInfo::Language(): " + BindingInfo::Language() + "!");
+        "BindingInfo::Language(): " + language + "!");
   }
 }
 
@@ -220,14 +220,25 @@ class mlpackModel
 
 } // namespace priv
 
-// Utility function that returns the first word (as delimited by spaces) of a
-// string.
-inline std::string ToUnderscores(const std::string& str)
+// Utility function that replaces all characters that may not be valid as anchor
+// names.
+inline std::string ToValidHTMLAnchor(const std::string& str)
 {
   std::string ret(str);
   std::replace(ret.begin(), ret.end(), ' ', '_');
   std::replace(ret.begin(), ret.end(), '{', '_');
   std::replace(ret.begin(), ret.end(), '}', '_');
+  std::replace(ret.begin(), ret.end(), '*', '_');
+  std::replace(ret.begin(), ret.end(), '-', '_');
+  std::replace(ret.begin(), ret.end(), ',', '_');
+  std::replace(ret.begin(), ret.end(), '(', '_');
+  std::replace(ret.begin(), ret.end(), ')', '_');
+  std::replace(ret.begin(), ret.end(), '.', '_');
+  std::replace(ret.begin(), ret.end(), '/', '_');
+  if (std::isdigit(ret[0]) || ret[0] == '_')
+  {
+    ret = "a_" + ret;
+  }
   return ret;
 }
 
@@ -237,12 +248,10 @@ inline std::string ToUnderscores(const s
 inline std::string PrintTypeDocs()
 {
   std::ostringstream oss;
-  oss << "<div id=\"" << BindingInfo::Language()
-      << "\" class=\"language-types\" markdown=\"1\">" << std::endl;
-  oss << "## data formats" << std::endl;
-  oss << "{: .language-types-h2 #" << BindingInfo::Language()
-      << "_data-formats }" << std::endl;
+  oss << "## Data Formats" << std::endl;
   oss << std::endl;
+  // This is used for formatting, to provide wider spacing between types.
+  oss << "<div id=\"data-formats-div\" markdown=\"1\">" << std::endl;
 
   // Iterate through each of the types that we care about.
   oss << "mlpack bindings for " << PrintLanguage(BindingInfo::Language())
@@ -263,144 +272,142 @@ inline std::string PrintTypeDocs()
   data.required = false;
   data.input = true;
   data.loaded = false;
-  data.value = MLPACK_ANY(int(0));
+  data.value = std::any(int(0));
 
   std::string type = GetPrintableType<int>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<int>(data) << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << "}: "
+      << PrintTypeDoc<int>(data) << std::endl;
 
   data.tname = std::string(typeid(double).name());
   data.cppType = "double";
-  data.value = MLPACK_ANY(double(0.0));
+  data.value = std::any(double(0.0));
 
   type = GetPrintableType<double>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<double>(data)
-      << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+      << PrintTypeDoc<double>(data) << std::endl;
 
   data.tname = std::string(typeid(bool).name());
   data.cppType = "double";
-  data.value = MLPACK_ANY(bool(0.0));
+  data.value = std::any(bool(0.0));
 
   type = GetPrintableType<bool>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<bool>(data) << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+      << PrintTypeDoc<bool>(data) << std::endl;
 
   data.tname = std::string(typeid(std::string).name());
   data.cppType = "std::string";
-  data.value = MLPACK_ANY(std::string(""));
+  data.value = std::any(std::string(""));
 
   type = GetPrintableType<std::string>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<std::string>(data)
-      << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+      << PrintTypeDoc<std::string>(data) << std::endl;
 
   data.tname = std::string(typeid(std::vector<int>).name());
   data.cppType = "std::vector<int>";
-  data.value = MLPACK_ANY(std::vector<int>());
+  data.value = std::any(std::vector<int>());
 
   type = GetPrintableType<std::vector<int>>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<std::vector<int>>(data)
-      << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+      << PrintTypeDoc<std::vector<int>>(data) << std::endl;
 
   data.tname = std::string(typeid(std::vector<std::string>).name());
   data.cppType = "std::vector<std::string>";
-  data.value = MLPACK_ANY(std::vector<std::string>());
+  data.value = std::any(std::vector<std::string>());
 
   type = GetPrintableType<std::vector<std::string>>(data);
-  oss << " - `" << type << "`{: " << "#doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: "
-      << PrintTypeDoc<std::vector<std::string>>(data) << std::endl;
+  oss << " - `" << type << "`{: " << "#doc_" << ToValidHTMLAnchor(type)
+      << " }: " << PrintTypeDoc<std::vector<std::string>>(data) << std::endl;
 
   data.tname = std::string(typeid(arma::mat).name());
   data.cppType = "arma::mat";
-  data.value = MLPACK_ANY(arma::mat());
+  data.value = std::any(arma::mat());
 
   type = GetPrintableType<arma::mat>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<arma::mat>(data)
-      << std::endl;
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+      << PrintTypeDoc<arma::mat>(data) << std::endl;
 
   data.tname = std::string(typeid(arma::Mat<size_t>).name());
   data.cppType = "arma::Mat<size_t>";
-  data.value = MLPACK_ANY(arma::Mat<size_t>());
+  data.value = std::any(arma::Mat<size_t>());
 
+  // Make sure that we don't print the same type twice, if the integer matrix
+  // type is the same type as the regular matrix type.
+  std::string oldType = type;
   type = GetPrintableType<arma::Mat<size_t>>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: " << PrintTypeDoc<arma::Mat<size_t>>(data)
-      << std::endl;
+  if (type != oldType)
+  {
+    oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
+        << PrintTypeDoc<arma::Mat<size_t>>(data) << std::endl;
+  }
 
   data.tname = std::string(typeid(arma::rowvec).name());
   data.cppType = "arma::rowvec";
-  data.value = MLPACK_ANY(arma::rowvec());
+  data.value = std::any(arma::rowvec());
   const std::string& rowType = GetPrintableType<arma::rowvec>(data);
 
-  oss << " - `" << rowType << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(rowType) << " }: " << PrintTypeDoc<arma::rowvec>(data)
-      << std::endl;
+  oss << " - `" << rowType << "`{: #doc_" << ToValidHTMLAnchor(rowType)
+      << " }: " << PrintTypeDoc<arma::rowvec>(data) << std::endl;
 
   data.tname = std::string(typeid(arma::Row<size_t>).name());
   data.cppType = "arma::Row<size_t>";
-  data.value = MLPACK_ANY(arma::Row<size_t>());
+  data.value = std::any(arma::Row<size_t>());
   const std::string& urowType = GetPrintableType<arma::Row<size_t>>(data);
 
-  oss << " - `" << urowType << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(urowType) << " }: "
-      << PrintTypeDoc<arma::Row<size_t>>(data)
-      << std::endl;
+  if (urowType != rowType)
+  {
+    oss << " - `" << urowType << "`{: #doc_" << ToValidHTMLAnchor(urowType)
+        << " }: " << PrintTypeDoc<arma::Row<size_t>>(data) << std::endl;
+  }
 
   data.tname = std::string(typeid(arma::vec).name());
   data.cppType = "arma::vec";
-  data.value = MLPACK_ANY(arma::vec());
+  data.value = std::any(arma::vec());
   const std::string& colType = GetPrintableType<arma::vec>(data);
 
   // For some languages there is no distinction between column and row vectors.
   // If that is the case, then don't print both.
   if (colType != rowType)
   {
-    oss << " - `" << colType << "`{: #doc_" << BindingInfo::Language() << "_"
-        << ToUnderscores(colType) << " }: " << PrintTypeDoc<arma::vec>(data)
-        << std::endl;
+    oss << " - `" << colType << "`{: #doc_" << ToValidHTMLAnchor(colType)
+        << " }: " << PrintTypeDoc<arma::vec>(data) << std::endl;
   }
 
   data.tname = std::string(typeid(arma::Col<size_t>).name());
   data.cppType = "arma::Col<size_t>";
-  data.value = MLPACK_ANY(arma::Col<size_t>());
+  data.value = std::any(arma::Col<size_t>());
   const std::string& ucolType = GetPrintableType<arma::Col<size_t>>(data);
 
   // For some languages there is no distinction between column and row vectors.
   // If that is the case, then don't print both.
   if (ucolType != urowType)
   {
-    oss << " - `" << ucolType << "`{ #doc_" << BindingInfo::Language() << "_"
-        << ToUnderscores(ucolType) << " }: "
-        << PrintTypeDoc<arma::Col<size_t>>(data) << std::endl;
+    oss << " - `" << ucolType << "`{ #doc_" << ToValidHTMLAnchor(ucolType)
+        << " }: " << PrintTypeDoc<arma::Col<size_t>>(data) << std::endl;
   }
 
   data.tname =
       std::string(typeid(std::tuple<data::DatasetInfo, arma::mat>).name());
   data.cppType = "std::tuple<data::DatasetInfo, arma::mat>";
-  data.value = MLPACK_ANY(std::tuple<data::DatasetInfo, arma::mat>());
+  data.value = std::any(std::tuple<data::DatasetInfo, arma::mat>());
 
   type = GetPrintableType<std::tuple<data::DatasetInfo, arma::mat>>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language() << "_"
-      << ToUnderscores(type) << " }: "
+  oss << " - `" << type << "`{: #doc_" << ToValidHTMLAnchor(type) << " }: "
       << PrintTypeDoc<std::tuple<data::DatasetInfo, arma::mat>>(data)
       << std::endl;
 
   data.tname = std::string(typeid(priv::mlpackModel).name());
   data.cppType = "mlpackModel";
-  data.value = MLPACK_ANY(new priv::mlpackModel());
+  data.value = std::any(new priv::mlpackModel());
 
   type = GetPrintableType<priv::mlpackModel*>(data);
-  oss << " - `" << type << "`{: #doc_" << BindingInfo::Language()
-      << "_model }: " << PrintTypeDoc<priv::mlpackModel*>(data) << std::endl;
+  oss << " - `" << type << "`{: #doc_model }: "
+      << PrintTypeDoc<priv::mlpackModel*>(data) << std::endl;
 
   // Clean up memory.
-  delete MLPACK_ANY_CAST<priv::mlpackModel*>(data.value);
+  delete std::any_cast<priv::mlpackModel*>(data.value);
 
-  oss << std::endl << "</div>" << std::endl;
+  oss << "</div>" << std::endl;
+  oss << std::endl;
 
   return oss.str();
 }
@@ -750,8 +757,7 @@ inline std::string ParamType(util::Param
   if (result)
     anchorType = "model";
 
-  return "[`" + output + "`](#doc_" + BindingInfo::Language() + "_" +
-      ToUnderscores(anchorType) + ")";
+  return "[`" + output + "`](#doc_" + ToValidHTMLAnchor(anchorType) + ")";
 }
 
 inline std::string ImportExtLib()
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/print_docs.cpp 4.4.0-1/src/mlpack/bindings/markdown/print_docs.cpp
--- 4.3.0-2/src/mlpack/bindings/markdown/print_docs.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/print_docs.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -23,8 +23,10 @@
 #define SRC_PREFIX "https://github.com/mlpack/mlpack/blob/master/src/"
 #endif
 
+// Relative path to the root of the documentation directory.
+// (The assumption is that the documentation is built in doc/user/bindings/.)
 #ifndef DOC_PREFIX
-#define DOC_PREFIX "https://github.com/mlpack/mlpack/blob/master/doc/"
+#define DOC_PREFIX "../../"
 #endif
 
 using namespace std;
@@ -34,38 +36,21 @@ using namespace mlpack::bindings;
 using namespace mlpack::bindings::markdown;
 
 void PrintHeaders(const string& bindingName,
-                  const vector<string>& languages,
-                  const vector<bool>& addWrapperDocs)
+                  const bool addWrapperDocs)
 {
-  // We just want to print the name of the function and a link, as Markdown.
-  // We have to mark it as having the right language with a div.
-  for (size_t i = 0; i < languages.size(); ++i)
-  {
-    BindingInfo::Language() = languages[i];
-
-    // Get the name of the binding in the target language, and convert it to
-    // lowercase (since the anchor link will be in lowercase).
-    const string langBindingName =
-        addWrapperDocs[i] ? GetWrapperName(bindingName) :
-            GetBindingName(bindingName);
-
-    cout << " - [" << langBindingName << "](#" << languages[i]
-        << "_";
-
-    if (addWrapperDocs[i])
-      cout << GetWrapperLink(bindingName);
-    else
-      cout << bindingName;
-
-    cout << "){: .language-link #" << languages[i] << " }"
-        << endl;
-  }
+  // Get the name of the binding in the target language, and convert it to
+  // lowercase (since the anchor link will be in lowercase).
+  const string langBindingName = addWrapperDocs ? GetWrapperName(bindingName) :
+          GetBindingName(bindingName);
+
+  cout << "<li><a href=\"LINKROOTuser/bindings/" << BindingInfo::Language()
+      << ".html#" << bindingName << "\">" << langBindingName << "</a></li>"
+      << endl;
 }
 
 void PrintDocs(const string& bindingName,
-               const vector<string>& languages,
                const vector<string>& validMethods,
-               const vector<bool>& addWrapperDocs)
+               const bool addWrapperDocs)
 {
   // Here we have to handle both wrapper and non wrapper bindings.
 
@@ -85,297 +70,239 @@ void PrintDocs(const string& bindingName
   }
 
   // First, for this section, print each of the names.
-  for (size_t i = 0; i < languages.size(); ++i)
-  {
-    BindingInfo::Language() = languages[i];
-
-    cout << "<div class=\"language-title\" id=\"" << languages[i]
-        << "\" markdown=\"1\">" << endl;
+  // We have to check if docs are printed for wrappers or not.
+  const string langBindingName = addWrapperDocs ? GetWrapperName(bindingName) :
+      GetBindingName(bindingName);
 
-    // here we have to check if docs are printed for wrappers or not.
-    const string langBindingName = 
-        addWrapperDocs[i] ? GetWrapperName(bindingName) :
-            GetBindingName(bindingName);
-
-    cout << "## " << langBindingName << endl;
-    cout << "{: #" << languages[i] << "_";
-
-    if (addWrapperDocs[i])
-      cout << GetWrapperLink(bindingName);
-    else
-      cout << bindingName;
-
-    cout << " }" << endl;
-    cout << "</div>" << endl;
-  }
+  cout << "## " << langBindingName << endl;
+  cout << "{: #" << bindingName << " }" << endl;
   cout << endl;
 
   // Next, print the logical name of the binding (that's known by
   // ProgramInfo).
-  for (size_t i = 0; i < languages.size(); ++i)
-  {
-    BindingInfo::Language() = languages[i];
-    cout << "<div class=\"language-logical-name\" id=\"" << languages[i]
-        << "\" markdown=\"1\">" << endl;
-    cout << "#### ";
-    if (addWrapperDocs[i])
-      cout << paramMethods[0].Doc().name << endl;
-    else
-      cout << doc.name << endl; 
-    cout << "</div>" << endl;
-  }
+  cout << "#### ";
+  if (addWrapperDocs)
+    cout << paramMethods[0].Doc().name << endl;
+  else
+    cout << doc.name << endl;
+  cout << "{: #" << bindingName << "_descr }" << endl;
   cout << endl;
 
-  for (size_t i = 0; i < languages.size(); ++i)
+  // No need to print ProgramCall for wrappers (already included in example).
+  if (!addWrapperDocs)
   {
-    // no need to print ProgramCall for wrappers (already included in example).
-    if (addWrapperDocs[i])
-      continue;
-
-    BindingInfo::Language() = languages[i];
-
-    cout << "<div class=\"language-decl\" id=\"" << languages[i]
-        << "\" markdown=\"1\">" << endl;
     cout << ProgramCall(bindingName);
-    cout << "</div>" << endl;
   }
   cout << endl;
 
-  // Print short description for each language.
-  for (size_t i = 0; i < languages.size(); ++i)
+  // Print short description.
+  if (addWrapperDocs)
   {
-    BindingInfo::Language() = languages[i];
-    cout << "<div class=\"language-short-desc\" id=\"" << languages[i]
-        << "\" markdown=\"1\">" << endl;
-    if (addWrapperDocs[i])
-    {
-      string desc = ReplaceAllCopy(
-          paramMethods[0].Doc().longDescription(), "|", "\\|");
-      cout << desc;
-    }
-    else
-    {
-      cout << doc.shortDescription << " ";
-      cout << "[Detailed documentation](#" << languages[i] << "_";
-      cout << bindingName << "_detailed-documentation){: .language-detail-link #"
-          << languages[i] << " }.";
-    }
+    string desc = ReplaceAllCopy(paramMethods[0].Doc().longDescription(), "|",
+        "\\|");
+    cout << desc;
+  }
+  else
+  {
+    cout << doc.shortDescription << " ";
+    cout << "[Detailed documentation](#" << bindingName
+        << "_detailed-documentation)." << endl;
     cout << endl;
-    cout << "</div>" << endl;
   }
   cout << endl;
 
-  // Next, print the PROGRAM_INFO() documentation for each language.
-  for (size_t i = 0; i < languages.size(); ++i)
+  // Next, print the PROGRAM_INFO() documentation.
+  if (!addWrapperDocs) // different structure for wrappers.
   {
-    if (!addWrapperDocs[i]) // different structure for wrappers.
-    {
-      BindingInfo::Language() = languages[i];
+    // We need to print the signature.
+
+    // Now, iterate through each of the input options.
+    cout << endl;
+    cout << "### Input options" << endl;
+    string inputInfo = PrintInputOptionInfo();
+    if (inputInfo.size() > 0)
+      cout << inputInfo << endl;
+    cout << endl;
 
-      // This works with the Kramdown processor.
-      cout << "<div class=\"language-section\" id=\"" << languages[i]
-          << "\" markdown=\"1\">" << endl;
+    cout << "| ***name*** | ***type*** | ***description*** | ***default*** |"
+          << endl;
+    cout << "|------------|------------|-------------------|---------------|"
+        << endl;
 
-      // We need to print the signature.
+    unordered_set<string> paramsSet;
+    PrintParamTable(bindingName, params,
+        {"name", "type", "description", "default"}, paramsSet,
+        false, false, true, false);
+    paramsSet.clear();
 
-      // Now, iterate through each of the input options.
+    cout << endl;
+
+    // Determine if there are any output options, to see if we need
+    // to print the header of the output options table.
+    bool hasOutputOptions = false;
+    map<string, ParamData>& parameters = params.Parameters();
+    for (map<string, ParamData>::iterator it = parameters.begin();
+        it != parameters.end(); ++it)
+    {
+      if (!it->second.input)
+      {
+        hasOutputOptions = true;
+        break;
+      }
+    }
+
+    if (hasOutputOptions)
+    {
+      // Next, iterate through the list of output options.
+      cout << "### Output options" << endl;
       cout << endl;
-      cout << "### Input options" << endl;
-      string inputInfo = PrintInputOptionInfo();
-      if (inputInfo.size() > 0)
-        cout << inputInfo << endl;
+      string outputInfo = PrintOutputOptionInfo();
+      if (outputInfo.size() > 0)
+        cout << outputInfo << endl;
       cout << endl;
+      cout << "| ***name*** | ***type*** | ***description*** |" << endl;
+      cout << "|------------|------------|-------------------|" << endl;
+    }
 
-      cout << "| ***name*** | ***type*** | ***description*** | ***default*** |"
-          << endl;
-      cout << "|------------|------------|-------------------|---------------|"
-          << endl;
+    PrintParamTable(bindingName, params,
+        {"name", "type", "description"}, paramsSet,
+        false, false, false, true);
+    paramsSet.clear();
 
-      unordered_set<string> paramsSet;
-      PrintParamTable(bindingName, languages[i], params,
-          {"name", "type", "description", "default"}, paramsSet,
-          false, false, true, false);
-      paramsSet.clear();
+    cout << endl;
 
-      cout << endl;
+    cout << "### Detailed documentation" << endl;
+    cout << "{: #" << bindingName << "_detailed-documentation }" << endl;
+    cout << endl;
+    string desc = ReplaceAllCopy(doc.longDescription(), "|", "\\|");
+    cout << desc << endl << endl;
+
+    if (doc.example.size() > 0)
+      cout << "### Example" << endl;
+    for (size_t j = 0; j < doc.example.size(); ++j)
+    {
+      string eg = ReplaceAllCopy(doc.example[j](), "|", "\\|");
+      cout << eg << endl << endl;
+    }
+    cout << "### See also" << endl;
+    cout << endl;
+
+    for (size_t j = 0; j < doc.seeAlso.size(); ++j)
+    {
+      cout << " - " << "[";
+      // We need special processing if the user has specified a binding name
+      // starting with @ (i.e., '@kfn' or similar).
+      if (doc.seeAlso[j].first[0] == '@')
+        cout << GetBindingName(doc.seeAlso[j].first.substr(1));
+      else
+        cout << doc.seeAlso[j].first;
+      cout << "](";
 
-      // Determine if there are any output options, to see if we need
-      // to print the header of the output options table.
-      bool hasOutputOptions = false;
-      map<string, ParamData>& parameters = params.Parameters();
-      for (map<string, ParamData>::iterator it = parameters.begin();
-          it != parameters.end(); ++it)
+      // We need special handling of source links.
+      if (doc.seeAlso[j].second.substr(0, 4) == "@src")
       {
-        if (!it->second.input)
-        {
-          hasOutputOptions = true;
-          break;
-        }
+        cout << SRC_PREFIX << doc.seeAlso[j].second.substr(5);
       }
-
-      if (hasOutputOptions)
+      else if (doc.seeAlso[j].second.substr(0, 4) == "@doc")
       {
-        // Next, iterate through the list of output options.
-        cout << "### Output options" << endl;
-        cout << endl;
-        string outputInfo = PrintOutputOptionInfo();
-        if (outputInfo.size() > 0)
-          cout << outputInfo << endl;
-        cout << endl;
-        cout << "| ***name*** | ***type*** | ***description*** |" << endl;
-        cout << "|------------|------------|-------------------|" << endl;
+        cout << DOC_PREFIX << doc.seeAlso[j].second.substr(5);
       }
-
-      PrintParamTable(bindingName, languages[i], params,
-          {"name", "type", "description"}, paramsSet,
-          false, false, false, true);
-      paramsSet.clear();
-
-      cout << endl;
-
-      cout << "### Detailed documentation" << endl;
-      cout << "{: #" << languages[i] << "_" << bindingName
-          << "_detailed-documentation }" << endl;
-      cout << endl;
-      string desc = ReplaceAllCopy(doc.longDescription(),
-                                     "|", "\\|");
-      cout << desc << endl << endl;
-
-      if (doc.example.size() > 0)
-        cout << "### Example" << endl;
-      for (size_t j = 0; j < doc.example.size(); ++j)
+      else if (doc.seeAlso[j].second[0] == '#')
       {
-        string eg = ReplaceAllCopy(doc.example[j](), "|", "\\|");
-        cout << eg << endl << endl;
+        cout << "#" << doc.seeAlso[j].second.substr(1);
       }
-      cout << "### See also" << endl;
-      cout << endl;
-      for (size_t j = 0; j < doc.seeAlso.size(); ++j)
+      else
       {
-        cout << " - " << "[";
-        // We need special processing if the user has specified a binding name
-        // starting with @ (i.e., '@kfn' or similar).
-        if (doc.seeAlso[j].first[0] == '@')
-          cout << GetBindingName(doc.seeAlso[j].first.substr(1));
-        else
-          cout << doc.seeAlso[j].first;
-        cout << "](";
-
-        // We need special handling of source links.
-        if (doc.seeAlso[j].second.substr(0, 4) == "@src")
-        {
-          cout << SRC_PREFIX << doc.seeAlso[j].second.substr(5);
-        }
-        else if (doc.seeAlso[j].second.substr(0, 4) == "@doc")
-        {
-          cout << DOC_PREFIX << doc.seeAlso[j].second.substr(5);
-        }
-        else if (doc.seeAlso[j].second[0] == '#')
-        {
-          cout << "#" << languages[i] << "_"
-              << doc.seeAlso[j].second.substr(1);
-        }
-        else
-        {
-          cout << doc.seeAlso[j].second;
-        }
-
-        cout << ")" << endl;
+        cout << doc.seeAlso[j].second;
       }
-      cout << endl;
 
-      cout << "</div>" << endl;
-      cout << endl;
+      cout << ")" << endl;
     }
-    else // for wrappers.
-    {
-      BindingInfo::Language() = languages[i];
-      cout << "<div class=\"language-section\" id=\"" << languages[i]
-          << "\" markdown=\"1\">" << endl;
-      cout << endl;
+    cout << endl;
+  }
+  else // for wrappers.
+  {
+    cout << "### Parameters" << endl;
+    cout << endl;
 
-      cout << "### Parameters" << endl;
-      cout << endl;
+    cout << "| ***name*** | ***type*** | ***description*** | ***default*** |"
+        << endl;
+    cout << "|------------|------------|-------------------|---------------|"
+        << endl;
 
-      cout << "| ***name*** | ***type*** | ***description*** | ***default*** |"
-          << endl;
-      cout << "|------------|------------|-------------------|---------------|"
-          << endl;
+    unordered_set<string> paramsSet; // to prevent duplicates.
+
+    for(size_t j = 0; j < paramMethods.size(); ++j)
+    {
+      PrintParamTable(bindingName + "_" + validMethods[j],
+          paramMethods[j], {"name", "type", "description",
+          "default"}, paramsSet, true, false, true, false);
+    }
+    paramsSet.clear(); // for reusing.
+    cout << endl;
 
-      unordered_set<string> paramsSet; // to prevent duplicates.
+    cout << "### Example" << endl;
+    cout << endl;
+    string example;
 
-      for(size_t j = 0; j < paramMethods.size(); ++j)
+    for (size_t j = 0; j < paramMethods.size(); ++j)
+    {
+      for (size_t k = 0; k < paramMethods[j].Doc().example.size(); ++k)
       {
-        PrintParamTable(bindingName + "_" + validMethods[j],
-            languages[i], paramMethods[j], {"name", "type", "description",
-            "default"}, paramsSet, true, false, true, false);
+        string eg = ReplaceAllCopy(paramMethods[j].Doc().example[k](), "|",
+            "\\|");
+        example += eg + "\n";
       }
-      paramsSet.clear(); // for reusing.
+    }
+
+    cout << "```" << BindingInfo::Language() << endl;
+    cout << example.substr(0, example.size() - 1) << endl; // do not want the last "\n".
+    cout << "```" << endl;
+    cout << endl;
+
+    cout << "### Methods" << endl;
+    cout << endl;
+    cout << "| **name** | **description** |" << endl;
+    cout << "|----------|-----------------|" << endl;
+    for (size_t j = 0; j < validMethods.size(); j++)
+    {
+      cout << "| " << GetMappedName(validMethods[j]) << " | ";
+      cout << paramMethods[j].Doc().shortDescription;
+      cout << " |" << endl;
+    }
+    cout << endl;
+
+    // Print information for each method.
+    for (size_t j = 0; j < validMethods.size(); j++)
+    {
+      cout << "### " << j + 1 << ". " << GetMappedName(validMethods[j])
+          << endl;
+      cout << endl;
+      cout << paramMethods[j].Doc().shortDescription << endl;
       cout << endl;
 
-      cout << "### Example" << endl;
+      cout << "#### Input Parameters:" << endl;
       cout << endl;
-      string example;
 
-      for(size_t j = 0; j < paramMethods.size(); ++j)
-      {
-        for(size_t k = 0; k < paramMethods[j].Doc().example.size();
-            ++k)
-        {
-          string eg = ReplaceAllCopy(paramMethods[j].Doc().example[k](),
-              "|", "\\|");
-          example += eg + "\n";
-        }
-      }
+      cout << "| **name** | **type** | **description** |" << endl;
+      cout << "|----------|----------|-----------------|" << endl;
 
-      cout << "```" << languages[i] << endl;
-      cout << example.substr(0, example.size() - 1) << endl; // do not want the last "\n".
-      cout << "```" << endl;
+      PrintParamTable(bindingName + "_" + validMethods[j],
+          paramMethods[j], {"name", "type", "description"},
+          paramsSet, false, true, true, false);
+      paramsSet.clear();
       cout << endl;
 
-      cout << "### Methods" << endl;
+      cout << "#### Returns: " << endl;
       cout << endl;
-      cout << "| **name** | **description** |" << endl;
+
+      cout << "| **type** | **description** |" << endl;
       cout << "|----------|-----------------|" << endl;
-      for(size_t j = 0; j < validMethods.size(); j++)
-      {
-        cout << "| " << GetMappedName(validMethods[j]) << " | ";
-        cout << paramMethods[j].Doc().shortDescription;
-        cout << " |" << endl;
-      }
-      cout << endl;
 
-      // Print information for each method.
-      for(size_t j = 0; j < validMethods.size(); j++)
-      {
-        cout << "### " << j+1 << ". " << GetMappedName(validMethods[j]) << endl;
-        cout << endl;
-        cout << paramMethods[j].Doc().shortDescription << endl;
-        cout << endl;
-
-        cout << "#### Input Parameters:" << endl;
-        cout << endl;
-
-        cout << "| **name** | **type** | **description** |" << endl;
-        cout << "|----------|----------|-----------------|" << endl;
-
-        PrintParamTable(bindingName + "_" + validMethods[j], languages[i],
-            paramMethods[j], {"name", "type", "description"},
-            paramsSet, false, true, true, false);
-        paramsSet.clear();
-        cout << endl;
-
-        cout << "#### Returns: " << endl;
-        cout << endl;
-
-        cout << "| **type** | **description** |" << endl;
-        cout << "|----------|-----------------|" << endl;
-
-        PrintParamTable(bindingName + "_" + validMethods[j], languages[i],
-            paramMethods[j], {"type", "description"},
-            paramsSet, false, false, false, true);
-        cout << endl;
-      }
+      PrintParamTable(bindingName + "_" + validMethods[j],
+          paramMethods[j], {"type", "description"},
+          paramsSet, false, false, false, true);
+      cout << endl;
     }
   }
 }
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/print_docs.hpp 4.4.0-1/src/mlpack/bindings/markdown/print_docs.hpp
--- 4.3.0-2/src/mlpack/bindings/markdown/print_docs.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/print_docs.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,8 +21,7 @@
  * Output is printed to stdout.
  */
 void PrintHeaders(const std::string& bindingName,
-                  const std::vector<std::string>& languages,
-                  const std::vector<bool>& addWrapperDocs);
+                  const bool addWrapperDocs);
 
 /**
  * Given the current settings of IO, print Markdown documentation for the
@@ -34,8 +33,7 @@ void PrintHeaders(const std::string& bin
  * @param languages The set of languages to print documentation for.
  */
 void PrintDocs(const std::string& bindingName,
-               const std::vector<std::string>& languages,
                const std::vector<std::string>& validMethods,
-               const std::vector<bool>& addWrapperDocs);
+               const bool addWrapperDocs);
 
 #endif
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/print_param_table.cpp 4.4.0-1/src/mlpack/bindings/markdown/print_param_table.cpp
--- 4.3.0-2/src/mlpack/bindings/markdown/print_param_table.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/print_param_table.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,7 +24,6 @@ using namespace mlpack::bindings;
 using namespace mlpack::bindings::markdown;
 
 void PrintParamTable(const string& bindingName,
-                     const string& language,
                      Params& params,
                      const set<string>& headers,
                      unordered_set<string>& paramsSet,
@@ -33,6 +32,7 @@ void PrintParamTable(const string& bindi
                      const bool onlyInputParams,
                      const bool onlyOutputParams)
 {
+  const std::string& language = BindingInfo::Language();
   map<string, ParamData>& parameters = params.Parameters();
   for (map<string, ParamData>::iterator it = parameters.begin();
       it != parameters.end(); ++it)
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/print_param_table.hpp 4.4.0-1/src/mlpack/bindings/markdown/print_param_table.hpp
--- 4.3.0-2/src/mlpack/bindings/markdown/print_param_table.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/print_param_table.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,7 +30,6 @@
  * @param onlyOutputParams If true, print only output-parameters.
  */
 void PrintParamTable(const std::string& bindingName,
-                     const std::string& language,
                      mlpack::util::Params& params,
                      const std::set<std::string>& headers,
                      std::unordered_set<std::string>& paramsSet,
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/res/change_language.js 4.4.0-1/src/mlpack/bindings/markdown/res/change_language.js
--- 4.3.0-2/src/mlpack/bindings/markdown/res/change_language.js	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/res/change_language.js	1970-01-01 00:00:00.000000000 +0000
@@ -1,110 +0,0 @@
-/**
- * @file change_language.js
- * @author Ryan Curtin
- *
- * A utility function to change the language displayed on the page.  This
- * function should be called whenever the language is changed from the
- * drop-down.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-function changeLanguage()
-{
-  lang = document.getElementById("language-select").value;
-  var links = document.getElementsByClassName("language-link");
-  for (i = 0; i < links.length; ++i)
-  {
-    // With each of the links, we get the inner <a>, but we need the parent
-    // <li>.
-    if (links[i].id == lang || links[i].id == "always")
-      links[i].parentElement.style.display = "list-item";
-    else
-      links[i].parentElement.style.display = "none";
-  }
-
-  var titles = document.getElementsByClassName("language-title");
-  for (i = 0; i < titles.length; ++i)
-  {
-    if (titles[i].id == lang)
-      titles[i].style.display = "inline";
-    else
-      titles[i].style.display = "none";
-  }
-
-  var headers = document.getElementsByClassName("language-header");
-  for (i = 0; i < headers.length; ++i)
-  {
-    if (headers[i].id == lang)
-      headers[i].style.display = "inline";
-    else
-      headers[i].style.display = "none";
-  }
-
-  var decls = document.getElementsByClassName("language-decl");
-  for (i = 0; i < decls.length; ++i)
-  {
-    if (decls[i].id == lang)
-      decls[i].style.display = "inline";
-    else
-      decls[i].style.display = "none";
-  }
-
-  var types = document.getElementsByClassName("language-types");
-  for (i = 0; i < types.length; ++i)
-  {
-    if (types[i].id == lang)
-      types[i].style.display = "inline";
-    else
-      types[i].style.display = "none";
-  }
-
-  var details = document.getElementsByClassName("language-detail-link");
-  for (i = 0; i < details.length; ++i)
-  {
-    if (details[i].id == lang)
-      details[i].style.display = "inline";
-    else
-      details[i].style.display = "none";
-  }
-
-  var sections = document.getElementsByClassName("language-section");
-  for (i = 0; i < sections.length; ++i)
-  {
-    if (sections[i].id == lang)
-      sections[i].style.display = "inline";
-    else
-      sections[i].style.display = "none";
-  }
-}
-
-document.body.onload = function()
-{
-  // Do we need to manually set the language because the user came with an
-  // anchor?
-  if (window.location.hash)
-  {
-    // Try to extract the language.
-    firstUnderscore = window.location.hash.indexOf("_");
-    if (firstUnderscore !== -1)
-    {
-      var lang = window.location.hash.substring(1, firstUnderscore);
-      // Now see if it's in the list of languages.
-      var select = document.getElementById("language-select");
-      for (i = 0; i < select.length; ++i)
-      {
-        var select_lang = select[i].value;
-        // Is the language a match?
-        if (lang === select_lang)
-        {
-          select.value = select_lang;
-          break;
-        }
-      }
-    }
-  }
-
-  changeLanguage();
-}
diff -pruN 4.3.0-2/src/mlpack/bindings/markdown/res/formatting.css 4.4.0-1/src/mlpack/bindings/markdown/res/formatting.css
--- 4.3.0-2/src/mlpack/bindings/markdown/res/formatting.css	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/markdown/res/formatting.css	1970-01-01 00:00:00.000000000 +0000
@@ -1,151 +0,0 @@
-/*
- * @file formatting.css
- * @author Ryan Curtin
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-body
-{
-  background: #000000;
-}
-
-/* Don't display other languages by default. */
-div.language-title, div.language-header, div.language-decl,
-div.language-detail-link, div.language-types, div.language-section
-{
-  display: none;
-}
-
-/* Just display cli documentation. */
-div.language-title#cli, div.language-header#cli, div.language-decl#cli,
-div.language-detail-link#cli, div.language-types#cli, div.language-section#cli
-{
-  display: none;
-}
-
-div#header
-{
-  padding-top: 10px;
-  width: 200px;
-  background: #000000;
-  border-right: 2px solid #333333;
-  height: 100%;
-  position: fixed;
-  z-index: 1;
-  top: 0;
-  left: 0;
-  overflow-y: scroll;
-}
-
-div#header .language-select-div select
-{
-  color: #ffffff;
-  background: transparent;
-  border: none;
-  height: 29px;
-  padding: 5px;
-  width: 190px;
-}
-
-div#header .language-select-div
-{
-  color: #ffffff;
-  border: 2px solid #333333;
-  background: url('../res/menu_bg.png') no-repeat 96% 0;
-  height: 34px;
-  width: 180px;
-  overflow: hidden;
-  margin-bottom: 20px;
-
-  -webkit-border-radius: 5px;
-  -moz-border-radius: 5px;
-  border-radius: 5px;
-}
-
-div#header ul
-{
-  padding-top: 5px;
-  margin-bottom: 5px;
-  color: #bb2000;
-  display: table;
-  text-align: left;
-  padding-left: 0px;
-  margin-left: 15px;
-  font-size: 75%;
-}
-
-div#header ul li
-{
-  line-height: 1.3em;
-}
-
-div#header ul li a
-{
-  color: #eab72c;
-  font-weight: none;
-}
-
-div#header ul li a:hover
-{
-  color: #ffffff;
-  font-weight: none;
-}
-
-div#docs
-{
-  margin-left: 200px;
-}
-
-span.special
-{
-  font-size: 85%;
-  color: #eab72c;
-}
-
-div.category
-{
-  text-align: left;
-}
-
-div.category h5
-{
-  text-align: left;
-  margin-top: 0;
-  margin-bottom: 0;
-  color: #ffffff;
-  font-size: 100%;
-  font-weight: normal;
-  padding: 0;
-}
-
-h2#mlpack-overview, h2.language-types-h2
-{
-  padding-bottom: 0.75em;
-}
-
-div.language-header h1
-{
-  color: #ffffff;
-  text-align: center;
-  border-top: none;
-  border-bottom: 1px solid #eab72c;
-}
-
-div.language-types li code
-{
-  font-weight: bold;
-  color: #eab72c;
-}
-
-div.language-types li
-{
-  padding-bottom: 1em;
-}
-
-div.language-section table td a code:hover
-{
-  color: #bb2000;
-}
Binary files 4.3.0-2/src/mlpack/bindings/markdown/res/menu_bg.png and 4.4.0-1/src/mlpack/bindings/markdown/res/menu_bg.png differ
diff -pruN 4.3.0-2/src/mlpack/bindings/python/CMakeLists.txt 4.4.0-1/src/mlpack/bindings/python/CMakeLists.txt
--- 4.3.0-2/src/mlpack/bindings/python/CMakeLists.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/CMakeLists.txt	2024-05-28 20:08:34.000000000 +0000
@@ -158,6 +158,28 @@ add_custom_command(TARGET python_copy PR
         ${CMAKE_CURRENT_SOURCE_DIR}/setup_readme.md
         ${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/)
 
+# Copy all mlpack headers for inclusion in the package, but remove the bindings/
+# and tests/ directories as they should not be included.
+add_custom_command(TARGET python_copy PRE_BUILD
+    COMMAND ${CMAKE_COMMAND} ARGS -E copy_directory
+        ${CMAKE_SOURCE_DIR}/src/
+        ${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/mlpack/include/)
+add_custom_command(TARGET python_copy PRE_BUILD
+    COMMAND ${CMAKE_COMMAND} ARGS -E rm -r
+        ${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/mlpack/include/mlpack/bindings/
+        ${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/mlpack/include/mlpack/tests/)
+
+# Generate pkgconfig file for easy use of included headers.
+add_custom_target(python_pkgconfig
+    COMMAND ${CMAKE_COMMAND}
+        -D GENERATE_CPP_IN=${CMAKE_SOURCE_DIR}/src/mlpack/bindings/python/mlpack.pc.in
+        -D GENERATE_CPP_OUT=${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/mlpack/share/pkgconfig/mlpack.pc
+        -D PACKAGE_VERSION="${PACKAGE_VERSION}"
+        -P "${CMAKE_SOURCE_DIR}/CMake/ConfigureFileOnly.cmake"
+        BYPRODUCTS "${CMAKE_BINARY_DIR}/src/mlpack/bindings/python/mlpack/share/pkgconfig/mlpack.pc"
+        COMMENT "Configuring Python mlpack.pc...")
+add_dependencies(python_copy python_pkgconfig)
+
 if (BUILD_TESTS)
   foreach(test_file ${TEST_SOURCES})
     add_custom_command(TARGET python_copy PRE_BUILD
diff -pruN 4.3.0-2/src/mlpack/bindings/python/default_param_impl.hpp 4.4.0-1/src/mlpack/bindings/python/default_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/python/default_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/default_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ std::string DefaultParamImpl(
   if (std::is_same<T, bool>::value)
     oss << "False";
   else
-    oss << MLPACK_ANY_CAST<T>(data.value);
+    oss << std::any_cast<T>(data.value);
 
   return oss.str();
 }
@@ -51,7 +51,7 @@ std::string DefaultParamImpl(
 {
   // Print each element in an array delimited by square brackets.
   std::ostringstream oss;
-  const T& vector = MLPACK_ANY_CAST<T>(data.value);
+  const T& vector = std::any_cast<T>(data.value);
   oss << "[";
   if (std::is_same<T, std::vector<std::string>>::value)
   {
@@ -92,7 +92,7 @@ std::string DefaultParamImpl(
     util::ParamData& data,
     const typename std::enable_if<std::is_same<T, std::string>::value>::type*)
 {
-  const std::string& s = *MLPACK_ANY_CAST<std::string>(&data.value);
+  const std::string& s = *std::any_cast<std::string>(&data.value);
   return "'" + s + "'";
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/python/get_param.hpp 4.4.0-1/src/mlpack/bindings/python/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/python/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ void GetParam(util::ParamData& d,
               const void* /* input */,
               void* output)
 {
-  *((T**) output) = const_cast<T*>(MLPACK_ANY_CAST<T>(&d.value));
+  *((T**) output) = const_cast<T*>(std::any_cast<T>(&d.value));
 }
 
 } // namespace python
diff -pruN 4.3.0-2/src/mlpack/bindings/python/get_printable_param.hpp 4.4.0-1/src/mlpack/bindings/python/get_printable_param.hpp
--- 4.3.0-2/src/mlpack/bindings/python/get_printable_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/get_printable_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -44,7 +44,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* = 0)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
@@ -61,7 +61,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& matrix = MLPACK_ANY_CAST<T>(data.value);
+  const T& matrix = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   oss << matrix.n_rows << "x" << matrix.n_cols << " matrix";
@@ -78,7 +78,7 @@ std::string GetPrintableParam(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   std::ostringstream oss;
-  oss << data.cppType << " model at " << MLPACK_ANY_CAST<T*>(data.value);
+  oss << data.cppType << " model at " << std::any_cast<T*>(data.value);
   return oss.str();
 }
 
@@ -92,7 +92,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* = 0)
 {
   // Get the matrix.
-  const T& tuple = MLPACK_ANY_CAST<T>(data.value);
+  const T& tuple = std::any_cast<T>(data.value);
   const arma::mat& matrix = std::get<1>(tuple);
 
   std::ostringstream oss;
diff -pruN 4.3.0-2/src/mlpack/bindings/python/mlpack/io_util.hpp 4.4.0-1/src/mlpack/bindings/python/mlpack/io_util.hpp
--- 4.3.0-2/src/mlpack/bindings/python/mlpack/io_util.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/mlpack/io_util.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -22,16 +22,14 @@ namespace util {
 // Utility functions to correctly handle transposed Armadillo matrices.
 template<typename T>
 inline void TransposeIfNeeded(
-    const std::string& identifier,
-    T& value,
-    bool transpose)
+    T& /* value */,
+    bool /* transpose */)
 {
   // No transpose needed for non-matrices.
   return;
 }
 
 inline void TransposeIfNeeded(
-    const std::string& identifier,
     arma::mat& value,
     bool transpose)
 {
@@ -59,7 +57,7 @@ inline void SetParam(util::Params& param
                      T& value,
                      bool transpose = false)
 {
-  TransposeIfNeeded(identifier, value, transpose);
+  TransposeIfNeeded(value, transpose);
   params.Get<T>(identifier) = std::move(value);
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/python/mlpack.pc.in 4.4.0-1/src/mlpack/bindings/python/mlpack.pc.in
--- 4.3.0-2/src/mlpack/bindings/python/mlpack.pc.in	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/mlpack.pc.in	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,7 @@
+prefix=${pcfiledir}/../../
+includedir=${prefix}/include
+
+Name: mlpack
+Description: a flexible, fast machine learning library
+Version: @PACKAGE_VERSION@
+Cflags: -I${includedir}
diff -pruN 4.3.0-2/src/mlpack/bindings/python/print_doc_functions_impl.hpp 4.4.0-1/src/mlpack/bindings/python/print_doc_functions_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/python/print_doc_functions_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/print_doc_functions_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -413,12 +413,13 @@ inline std::string SplitTrainTest(const
                                   const std::string& testLabels,
                                   const std::string& splitRatio)
 {
-  std::string splitString = ">>> ";
-  splitString += testDataset + ", " + testLabels + ", ";
-  splitString += trainDataset + ", " + trainLabels;
-  splitString += " = ";
-  splitString += "preprocess_split(input_=" + datasetName + ", input_labels=";
-  splitString += labelName + ", test_ratio=" + splitRatio + ")";
+  std::string splitString;
+  splitString += ">>> d = preprocess_split(input_=" + datasetName + ", input_labels=";
+  splitString += labelName + ", test_ratio=" + splitRatio + ")\n";
+  splitString += ">>> " + trainDataset + " = d['training']\n";
+  splitString += ">>> " + trainLabels + " = d['training_labels']\n";
+  splitString += ">>> " + testDataset + " = d['test']\n";
+  splitString += ">>> " + testLabels + " = d['test_labels']";
   return splitString;
 }
 
diff -pruN 4.3.0-2/src/mlpack/bindings/python/print_type_doc_impl.hpp 4.4.0-1/src/mlpack/bindings/python/print_type_doc_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/python/print_type_doc_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/print_type_doc_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,11 +67,11 @@ std::string PrintTypeDoc(
 {
   if (std::is_same<T, std::vector<int>>::value)
   {
-    return "A list of integers; i.e., [0, 1, 2].";
+    return "A list of integers; i.e., `[0, 1, 2]`.";
   }
   else if (std::is_same<T, std::vector<std::string>>::value)
   {
-    return "A list of strings; i.e., [\"hello\", \"goodbye\"].";
+    return "A list of strings; i.e., `[\"hello\", \"goodbye\"]`.";
   }
   else
   {
diff -pruN 4.3.0-2/src/mlpack/bindings/python/print_wrapper_py.cpp 4.4.0-1/src/mlpack/bindings/python/print_wrapper_py.cpp
--- 4.3.0-2/src/mlpack/bindings/python/print_wrapper_py.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/print_wrapper_py.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -13,6 +13,7 @@
 #include "print_wrapper_py.hpp"
 #include "get_arma_type.hpp"
 #include "wrapper_functions.hpp"
+#include "strip_type.hpp"
 #include <mlpack/core/util/io.hpp>
 
 using namespace mlpack::util;
@@ -42,18 +43,18 @@ void PrintWrapperPY(const std::string& c
   // keeps track of indentation at each point.
   int indent = 0;
 
-  for(string i: methods)
+  for (string i: methods)
   {
     params[i] = IO::Parameters(groupName + "_" + i);
   }
 
-  if(category == "regression")
+  if (category == "regression")
   {
     cout << "class BaseEstimator:" << endl;
     cout << "  pass" << endl;
     cout << endl;
   }
-  else if(category == "classification")
+  else if (category == "classification")
   {
     cout << "class BaseEstimator:" << endl;
     cout << "  pass" << endl;
@@ -64,7 +65,7 @@ void PrintWrapperPY(const std::string& c
   }
 
   // Import different mlpack programs that are to be wrapped.
-  for(size_t i = 0; i < methods.size(); i++)
+  for (size_t i = 0; i < methods.size(); i++)
   {
     cout << "from mlpack." << groupName << "_" << methods[i] << " ";
     cout << "import " << groupName << "_" << methods[i] << endl;
@@ -72,19 +73,19 @@ void PrintWrapperPY(const std::string& c
 
   // Try importing scikit-learn, only for classification and
   // regression.
-  if(category == "regression")
+  if (category == "regression")
   {
     cout << "try:" << endl;
     cout << "  from sklearn.base import BaseEstimator" << endl;
   }
-  else if(category == "classification")
+  else if (category == "classification")
   {
     cout << "try: " << endl;
     cout << "  from sklearn.base import BaseEstimator, ClassifierMixin";
     cout << endl;
   }
 
-  if(category == "regression" || category == "classification")
+  if (category == "regression" || category == "classification")
   {
     cout << "except:" << endl;
     cout << "  pass" << endl;
@@ -93,30 +94,34 @@ void PrintWrapperPY(const std::string& c
 
   // Check and store if every parameter is serializable,
   // a hyperparameter and boolean.
-  for(map<string, Params>::iterator i=params.begin();
-      i!=params.end(); i++)
+  for (map<string, Params>::iterator i = params.begin();
+       i != params.end(); i++)
   {
     map<string, ParamData> methodParams = i->second.Parameters();
 
-    for(map<string, ParamData>::iterator itr=methodParams.begin();
-        itr!=methodParams.end(); itr++)
+    for (map<string, ParamData>::iterator itr = methodParams.begin();
+         itr != methodParams.end(); itr++)
     {
+      // Get the sanitized name of the class.
+      string strippedName, unused1, unused2;
+      StripType(itr->second.cppType, strippedName, unused1, unused2);
+
       // Checking for serializability.
       bool isSerial;
       i->second.functionMap[itr->second.tname]["IsSerializable"](
           itr->second, NULL, (void*)& isSerial);
-      if(isSerial)
-        serializable.insert(itr->second.cppType);
+      if (isSerial)
+        serializable.insert(strippedName);
 
       // Checking for hyperparameter.
       bool isHyperParam = false;
       size_t foundArma = itr->second.cppType.find("arma");
-      if(itr->second.input && foundArma == string::npos && !isSerial)
+      if (itr->second.input && foundArma == string::npos && !isSerial)
         isHyperParam = true;
       hyperParams[itr->first] = isHyperParam;
 
       // Checking for boolean.
-      if(itr->second.cppType == "bool")
+      if (itr->second.cppType == "bool")
         isBool[itr->first] = true;
       else
         isBool[itr->first] = false;
@@ -126,9 +131,9 @@ void PrintWrapperPY(const std::string& c
   string className = GetClassName(groupName);
 
   // print class.
-  if(category == "regression")
+  if (category == "regression")
     cout << "class " << className << "(BaseEstimator)" << ":" << endl;
-  else if(category == "classification")
+  else if (category == "classification")
   {
     cout << "class " << className << "(BaseEstimator, ClassifierMixin)" <<
         ":" << endl;
@@ -140,15 +145,15 @@ void PrintWrapperPY(const std::string& c
   indent += 2;
   cout << string(indent, ' ') << "def __init__(self," << endl;
 
-  for(auto itr=hyperParams.begin(); itr!=hyperParams.end(); itr++)
+  for (auto itr = hyperParams.begin(); itr != hyperParams.end(); itr++)
   {
     // indent + 13, here 13 -> def __init__(
-    if(itr->second && isBool[itr->first])
+    if (itr->second && isBool[itr->first])
     {
       cout << string(indent+13, ' ') << GetValidName(itr->first)
           << " = False," << endl;
     }
-    else if(itr->second && !isBool[itr->first])
+    else if (itr->second && !isBool[itr->first])
     {
       cout << string(indent+13, ' ') << GetValidName(itr->first)
           << " = None," << endl;
@@ -161,10 +166,10 @@ void PrintWrapperPY(const std::string& c
   // storing given arguments in attributes.
   indent += 2;
   cout << string(indent, ' ') << "# serializable attributes." << endl;
-  if(serializable.size() == 0)
+  if (serializable.size() == 0)
     cout << string(indent, ' ') << "# None" << endl;
 
-  for(auto itr=serializable.begin(); itr!=serializable.end(); itr++)
+  for (auto itr = serializable.begin(); itr != serializable.end(); itr++)
   {
     cout << string(indent, ' ');
     cout << "self._" << *itr << " = None" << endl;
@@ -173,9 +178,9 @@ void PrintWrapperPY(const std::string& c
   cout << endl;
   cout << string(indent, ' ') << "# hyper-parameters." << endl;
 
-  for(auto itr=hyperParams.begin(); itr!=hyperParams.end(); itr++)
+  for (auto itr = hyperParams.begin(); itr != hyperParams.end(); itr++)
   {
-    if(itr->second)
+    if (itr->second)
     {
       string validName = GetValidName(itr->first);
       cout << string(indent, ' ');
@@ -187,7 +192,7 @@ void PrintWrapperPY(const std::string& c
   indent -= 2;
 
   // print all method definitions.
-  for(string methodName: methods)
+  for (string methodName: methods)
   {
     // print method name.
     cout << string(indent, ' ') << "def " << GetMappedName(methodName);
@@ -205,39 +210,44 @@ void PrintWrapperPY(const std::string& c
     int numMatrixInputs = 0;
     int numVectorInputs = 0;
 
-    if(category == "regression" || category == "classification")
+    if (category == "regression" || category == "classification")
     {
-      for(map<string, ParamData>::iterator itr=methodParams.begin();
-          itr!=methodParams.end(); itr++)
+      for (map<string, ParamData>::iterator itr = methodParams.begin();
+           itr != methodParams.end(); itr++)
       {
         // Throw error if there are more than one matrix input params,
         // or more than one vector input params.
-        if(numMatrixInputs > 1)
-          Log::Fatal << "More than one matrix input parameters for " <<
+        if (numMatrixInputs > 1)
+        {
+          Log::Fatal << "More than one matrix input parameter for " <<
               methodName << "(" << GetMappedName(methodName) << ") method!" <<
               endl;
-        if(numVectorInputs > 1)
-          Log::Fatal << "More than one vector input parameters for " <<
+        }
+
+        if (numVectorInputs > 1)
+        {
+          Log::Fatal << "More than one vector input parameter for " <<
               methodName << "(" << GetMappedName(methodName) << ") method!" <<
               endl;
+        }
 
-        if(itr->second.input)
+        if (itr->second.input)
         {
           // If this is a matrix parameter.
-          if(itr->second.cppType == "arma::mat" ||
-             itr->second.cppType ==
-             "std::tuple<mlpack::data::DatasetInfo, arma::mat>" ||
-             itr->second.cppType == "arma::Mat<size_t>")
+          if (itr->second.cppType == "arma::mat" ||
+              itr->second.cppType ==
+                  "std::tuple<mlpack::data::DatasetInfo, arma::mat>" ||
+              itr->second.cppType == "arma::Mat<size_t>")
           {
             numMatrixInputs++;
             mapToScikitNames[itr->first] = "X";
             invMapToScikitNames["X"] = itr->first;
           }
           // If this is a vector parameter.
-          else if(itr->second.cppType == "arma::vec" ||
-                  itr->second.cppType == "arma::rowvec" ||
-                  itr->second.cppType == "arma::Row<size_t>" ||
-                  itr->second.cppType == "arma::Col<size_t>")
+          else if (itr->second.cppType == "arma::vec" ||
+                   itr->second.cppType == "arma::rowvec" ||
+                   itr->second.cppType == "arma::Row<size_t>" ||
+                   itr->second.cppType == "arma::Col<size_t>")
           {
             numVectorInputs++;
             mapToScikitNames[itr->first] = "y";
@@ -248,13 +258,13 @@ void PrintWrapperPY(const std::string& c
     }
 
     // Now we print the matrix and vector params.
-    if(category == "classification" || category == "regression")
+    if (category == "classification" || category == "regression")
     {
       bool hasX = false;
       bool hasy = false;
 
       // First print X, if it is present.
-      if(invMapToScikitNames.find("X") != invMapToScikitNames.end())
+      if (invMapToScikitNames.find("X") != invMapToScikitNames.end())
       {
         cout << string(indent + addIndent, ' ');
         cout << "X = None," << endl;
@@ -262,25 +272,25 @@ void PrintWrapperPY(const std::string& c
       }
 
       // Now print y, if it is present.
-      if(invMapToScikitNames.find("y") != invMapToScikitNames.end())
+      if (invMapToScikitNames.find("y") != invMapToScikitNames.end())
       {
         cout << string(indent + addIndent, ' ');
         cout << "y = None," << endl;
         hasy = true;
       }
 
-      // Now print actual names, to make it work for
-      // both mapped and actual names.
+      // Now print actual names, to make it work for both mapped and actual
+      // names.
 
       // Now print actual name for X.
-      if(hasX)
+      if (hasX)
       {
         cout << string(indent + addIndent, ' ');
         cout << GetValidName(invMapToScikitNames["X"]) << " = None," << endl;
       }
 
       // Now print actual name for y.
-      if(hasy)
+      if (hasy)
       {
         cout << string(indent + addIndent, ' ');
         cout << GetValidName(invMapToScikitNames["y"]) << " = None," << endl;
@@ -295,10 +305,10 @@ void PrintWrapperPY(const std::string& c
       string logicString = "";
       indent += 2;
 
-      if(hasX)
+      if (hasX)
       {
         string realName = GetValidName(invMapToScikitNames["X"]);
-        
+
         cout << string(indent, ' ') << "if X is not None and ";
         cout << realName << " is None:" << endl;
 
@@ -312,7 +322,7 @@ void PrintWrapperPY(const std::string& c
         cout << endl;
       }
 
-      if(hasy)
+      if (hasy)
       {
         string realName = GetValidName(invMapToScikitNames["y"]);
 
@@ -332,13 +342,17 @@ void PrintWrapperPY(const std::string& c
     else
     {
       // print input parameters.
-      for(map<string, ParamData>::iterator itr=methodParams.begin();
-          itr!=methodParams.end(); itr++)
+      for (map<string, ParamData>::iterator itr = methodParams.begin();
+           itr != methodParams.end(); itr++)
       {
         string validName = GetValidName(itr->first);
 
-        if(itr->second.input && serializable.find(itr->second.cppType) ==
-            serializable.end() && !hyperParams[itr->first])
+        // Get valid stripped name.
+        string strippedName, unused1, unused2;
+        StripType(itr->second.cppType, strippedName, unused1, unused2);
+
+        if (itr->second.input && serializable.find(strippedName) ==
+             serializable.end() && !hyperParams[itr->first])
         {
           cout << string(indent + addIndent, ' ');
           cout << validName << " = None," << endl;
@@ -357,8 +371,8 @@ void PrintWrapperPY(const std::string& c
     int count = 0; // just for reference.
 
     // first pass through the parameters and print all required parameters.
-    for(map<string, ParamData>::iterator itr=methodParams.begin();
-        itr!=methodParams.end(); itr++)
+    for (map<string, ParamData>::iterator itr = methodParams.begin();
+         itr != methodParams.end(); itr++)
     {
       if(itr->second.input && itr->second.required)
       {
@@ -367,9 +381,11 @@ void PrintWrapperPY(const std::string& c
           cout << string(indent + addIndent, ' ');
         cout << validName << " = ";
 
-        if(serializable.find(itr->second.cppType) != serializable.end())
-          cout << "self._" << itr->second.cppType << "," << endl;
-        else if(hyperParams[itr->first])
+        string strippedName, unused1, unused2;
+        StripType(itr->second.cppType, strippedName, unused1, unused2);
+        if (serializable.find(strippedName) != serializable.end())
+          cout << "self._" << strippedName << "," << endl;
+        else if (hyperParams[itr->first])
           cout << "self." << validName << "," << endl;
         else
           cout << validName << "," << endl;
@@ -379,20 +395,22 @@ void PrintWrapperPY(const std::string& c
     }
 
     // Now print all non-required parameters.
-    for(map<string, ParamData>::iterator itr=methodParams.begin();
-        itr!=methodParams.end(); itr++)
+    for (map<string, ParamData>::iterator itr = methodParams.begin();
+         itr != methodParams.end(); itr++)
     {
-      if(itr->second.input && !itr->second.required)
+      if (itr->second.input && !itr->second.required)
       {
         string validName = GetValidName(itr->first);
 
-        if(count != 0)
+        if (count != 0)
           cout << string(indent + addIndent, ' ');
         cout << validName << " = ";
 
-        if(serializable.find(itr->second.cppType) != serializable.end())
-          cout << "self._" << itr->second.cppType << "," << endl;
-        else if(hyperParams[itr->first])
+        string strippedName, unused1, unused2;
+        StripType(itr->second.cppType, strippedName, unused1, unused2);
+        if (serializable.find(strippedName) != serializable.end())
+          cout << "self._" << strippedName << "," << endl;
+        else if (hyperParams[itr->first])
           cout << "self." << validName << "," << endl;
         else
           cout << validName << "," << endl;
@@ -409,15 +427,17 @@ void PrintWrapperPY(const std::string& c
     string returnString = string(indent, ' ') + "return ";
     bool outputsOnlySerial = true;
 
-    for(map<string, ParamData>::iterator itr=methodParams.begin();
-        itr!=methodParams.end(); itr++)
+    for (map<string, ParamData>::iterator itr = methodParams.begin();
+         itr != methodParams.end(); itr++)
     {
-      if(!itr->second.input)
+      if (!itr->second.input)
       {
-        if(serializable.find(itr->second.cppType) != serializable.end())
+        string strippedName, unused1, unused2;
+        StripType(itr->second.cppType, strippedName, unused1, unused2);
+        if (serializable.find(strippedName) != serializable.end())
         {
           cout << string(indent, ' ');
-          cout << "self._" << itr->second.cppType << " = out[\"" <<
+          cout << "self._" << strippedName << " = out[\"" <<
               itr->first << "\"]" << endl;
         }
         else
@@ -430,7 +450,7 @@ void PrintWrapperPY(const std::string& c
     cout << endl;
 
     // return somethings.
-    if(outputsOnlySerial)
+    if (outputsOnlySerial)
       cout << returnString + "self" << endl;
     else
       cout << returnString.substr(0, returnString.size() - 2) << endl;
diff -pruN 4.3.0-2/src/mlpack/bindings/python/setup.py.in 4.4.0-1/src/mlpack/bindings/python/setup.py.in
--- 4.3.0-2/src/mlpack/bindings/python/setup.py.in	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/python/setup.py.in	2024-05-28 20:08:34.000000000 +0000
@@ -116,6 +116,12 @@ else:
                       else '${DISABLE_CFLAGS}'.split(' ')) \
         for name in pyxs]
 
+# Find all include files.
+include_files = []
+for (path, directories, filenames) in os.walk('mlpack/include/'):
+  for filename in filenames:
+    include_files.append(os.path.join('..', path, filename))
+
 setup(name='mlpack',
       version='${PACKAGE_VERSION}',
       description='a flexible, fast machine learning library',
@@ -141,6 +147,8 @@ setup(name='mlpack',
       install_requires=['cython>=0.24', 'numpy', 'pandas'],
       package_dir={ '': '.' }, # Might be superfluous.
       packages=['mlpack'],
+      package_data={ 'mlpack': include_files + ['../mlpack/share/pkgconfig/mlpack.pc'] },
+      include_package_data=True,
       cmdclass={ 'build_ext': build_ext },
       ext_modules = modules,
       setup_requires=['cython', 'pytest-runner'],
diff -pruN 4.3.0-2/src/mlpack/bindings/tests/delete_allocated_memory.hpp 4.4.0-1/src/mlpack/bindings/tests/delete_allocated_memory.hpp
--- 4.3.0-2/src/mlpack/bindings/tests/delete_allocated_memory.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/tests/delete_allocated_memory.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ void DeleteAllocatedMemoryImpl(
     util::ParamData& d,
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
-  (*MLPACK_ANY_CAST<T>(&d.value)).clear();
+  (*std::any_cast<T>(&d.value)).clear();
 }
 
 template<typename T>
@@ -42,7 +42,7 @@ void DeleteAllocatedMemoryImpl(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   // Delete the allocated memory (hopefully we actually own it).
-  delete *MLPACK_ANY_CAST<T*>(&d.value);
+  delete *std::any_cast<T*>(&d.value);
 }
 
 template<typename T>
diff -pruN 4.3.0-2/src/mlpack/bindings/tests/get_allocated_memory.hpp 4.4.0-1/src/mlpack/bindings/tests/get_allocated_memory.hpp
--- 4.3.0-2/src/mlpack/bindings/tests/get_allocated_memory.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/tests/get_allocated_memory.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -33,7 +33,7 @@ void* GetAllocatedMemory(
     util::ParamData& d,
     const typename std::enable_if<arma::is_arma_type<T>::value>::type* = 0)
 {
-  return (*MLPACK_ANY_CAST<T>(&d.value)).memptr();
+  return (*std::any_cast<T>(&d.value)).memptr();
 }
 
 template<typename T>
@@ -43,7 +43,7 @@ void* GetAllocatedMemory(
     const typename std::enable_if<data::HasSerialize<T>::value>::type* = 0)
 {
   // Here we have a model; return its memory location.
-  return *MLPACK_ANY_CAST<T*>(&d.value);
+  return *std::any_cast<T*>(&d.value);
 }
 
 template<typename T>
diff -pruN 4.3.0-2/src/mlpack/bindings/tests/get_param.hpp 4.4.0-1/src/mlpack/bindings/tests/get_param.hpp
--- 4.3.0-2/src/mlpack/bindings/tests/get_param.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/tests/get_param.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ template<typename T>
 T& GetParam(util::ParamData& d)
 {
   // No mapping is needed, so just cast it directly.
-  return *MLPACK_ANY_CAST<T>(&d.value);
+  return *std::any_cast<T>(&d.value);
 }
 
 /**
diff -pruN 4.3.0-2/src/mlpack/bindings/tests/get_printable_param_impl.hpp 4.4.0-1/src/mlpack/bindings/tests/get_printable_param_impl.hpp
--- 4.3.0-2/src/mlpack/bindings/tests/get_printable_param_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/bindings/tests/get_printable_param_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ std::string GetPrintableParam(
         std::tuple<data::DatasetInfo, arma::mat>>::value>::type* /* junk */)
 {
   std::ostringstream oss;
-  oss << MLPACK_ANY_CAST<T>(data.value);
+  oss << std::any_cast<T>(data.value);
   return oss.str();
 }
 
@@ -39,7 +39,7 @@ std::string GetPrintableParam(
     util::ParamData& data,
     const typename std::enable_if<util::IsStdVector<T>::value>::type* /* junk */)
 {
-  const T& t = MLPACK_ANY_CAST<T>(data.value);
+  const T& t = std::any_cast<T>(data.value);
 
   std::ostringstream oss;
   for (size_t i = 0; i < t.size(); ++i)
diff -pruN 4.3.0-2/src/mlpack/core/cereal/template_class_version.hpp 4.4.0-1/src/mlpack/core/cereal/template_class_version.hpp
--- 4.3.0-2/src/mlpack/core/cereal/template_class_version.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cereal/template_class_version.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,78 @@
+/**
+ * @file core/cereal/template_class_version.hpp
+ * @author Ryan Curtin
+ *
+ * Implementation of CEREAL_TEMPLATE_CLASS_VERSION() macro, useful for
+ * templatized types where CEREAL_CLASS_VERSION() will not work.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_CEREAL_TEMPLATE_CLASS_VERSION_HPP
+#define MLPACK_CORE_CEREAL_TEMPLATE_CLASS_VERSION_HPP
+
+#include <cereal/cereal.hpp>
+
+// This useful implementation is adapted from @lubensky on Github:
+// https://github.com/uscilab/cereal/issues/319#issuecomment-1512927210
+
+#define CEREAL_UNPACK(...) __VA_ARGS__
+
+#ifdef MLPACK_HAVE_CXX17
+
+// The C++17 version sets `version` as `inline`.
+#define CEREAL_TEMPLATE_CLASS_VERSION(ARGS, TYPE, VERSION_NUMBER)              \
+namespace cereal {                                                             \
+namespace detail {                                                             \
+template<CEREAL_UNPACK ARGS>                                                   \
+struct Version<CEREAL_UNPACK TYPE>                                             \
+{                                                                              \
+  static std::uint32_t registerVersion()                                       \
+  {                                                                            \
+    ::cereal::detail::StaticObject<Versions>::getInstance().mapping.emplace(   \
+        std::type_index(typeid(CEREAL_UNPACK TYPE)).hash_code(),               \
+                        CEREAL_UNPACK VERSION_NUMBER);                         \
+    return CEREAL_UNPACK VERSION_NUMBER;                                       \
+  }                                                                            \
+                                                                               \
+  static inline const std::uint32_t version = registerVersion();               \
+                                                                               \
+  static void unused() { (void) version; }                                     \
+}; /* end Version */                                                           \
+                                                                               \
+}                                                                              \
+}
+
+#else
+
+// Here we cannot use inline variables.
+#define CEREAL_TEMPLATE_CLASS_VERSION(ARGS, TYPE, VERSION_NUMBER)              \
+namespace cereal {                                                             \
+namespace detail {                                                             \
+template<CEREAL_UNPACK ARGS>                                                   \
+struct Version<CEREAL_UNPACK TYPE>                                             \
+{                                                                              \
+  static const std::uint32_t version;                                          \
+  static std::uint32_t registerVersion()                                       \
+  {                                                                            \
+    ::cereal::detail::StaticObject<Versions>::getInstance().mapping.emplace(   \
+        std::type_index(typeid(CEREAL_UNPACK TYPE)).hash_code(),               \
+                        CEREAL_UNPACK VERSION_NUMBER);                         \
+    return CEREAL_UNPACK VERSION_NUMBER;                                       \
+  }                                                                            \
+                                                                               \
+  static void unused() { (void) version; }                                     \
+}; /* end Version */                                                           \
+                                                                               \
+template<CEREAL_UNPACK ARGS>                                                   \
+const std::uint32_t Version<CEREAL_UNPACK TYPE>::version =                     \
+    Version<CEREAL_UNPACK TYPE>::registerVersion();                            \
+                                                                               \
+}                                                                              \
+}
+
+#endif // MLPACK_HAVE_CXX17
+
+#endif // TEMPLATE_CLASS_VERSION_HPP
diff -pruN 4.3.0-2/src/mlpack/core/cv/k_fold_cv_impl.hpp 4.4.0-1/src/mlpack/core/cv/k_fold_cv_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/k_fold_cv_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/k_fold_cv_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -227,7 +227,7 @@ void KFoldCV<MLAlgorithm,
   binSize = source.n_cols / k;
   lastBinSize = source.n_cols - ((k - 1) * binSize);
 
-  destination = (k == 2) ? source : arma::join_rows(source,
+  destination = (k == 2) ? source : join_rows(source,
       source.cols(0, source.n_cols - lastBinSize - 1));
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/accuracy_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/accuracy_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/accuracy_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/accuracy_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -23,7 +23,7 @@ double Accuracy::Evaluate(MLAlgorithm& m
 
   arma::Row<size_t> predictedLabels;
   model.Classify(data, predictedLabels);
-  size_t amountOfCorrectPredictions = arma::sum(predictedLabels == labels);
+  size_t amountOfCorrectPredictions = sum(predictedLabels == labels);
 
   return (double) amountOfCorrectPredictions / labels.n_elem;
 }
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/f1_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/f1_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/f1_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/f1_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,9 +36,9 @@ double F1<AS, PC>::Evaluate(MLAlgorithm&
   arma::Row<size_t> predictedLabels;
   model.Classify(data, predictedLabels);
 
-  size_t tp = arma::sum((labels == PC) % (predictedLabels == PC));
-  size_t numberOfPositivePredictions = arma::sum(predictedLabels == PC);
-  size_t numberOfPositiveClassInstances = arma::sum(labels == PC);
+  size_t tp = sum((labels == PC) % (predictedLabels == PC));
+  size_t numberOfPositivePredictions = sum(predictedLabels == PC);
+  size_t numberOfPositiveClassInstances = sum(labels == PC);
 
   double precision = double(tp) / numberOfPositivePredictions;
   double recall = double(tp) / numberOfPositiveClassInstances;
@@ -78,9 +78,9 @@ double F1<AS, PC>::Evaluate(MLAlgorithm&
   arma::vec f1s = arma::vec(numClasses);
   for (size_t c = 0; c < numClasses; ++c)
   {
-    size_t tp = arma::sum((labels == c) % (predictedLabels == c));
-    size_t positivePredictions = arma::sum(predictedLabels == c);
-    size_t positiveLabels = arma::sum(labels == c);
+    size_t tp = sum((labels == c) % (predictedLabels == c));
+    size_t positivePredictions = sum(predictedLabels == c);
+    size_t positiveLabels = sum(labels == c);
 
     double precision = double(tp) / positivePredictions;
     double recall = double(tp) / positiveLabels;
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/mse_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/mse_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/mse_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/mse_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,7 +24,7 @@ double MSE::Evaluate(MLAlgorithm& model,
 
   ResponsesType predictedResponses;
   model.Predict(data, predictedResponses);
-  double sum = arma::accu(arma::square(responses - predictedResponses));
+  double sum = accu(arma::square(responses - predictedResponses));
 
   return sum / responses.n_elem;
 }
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/precision_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/precision_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/precision_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/precision_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,8 +36,8 @@ double Precision<AS, PC>::Evaluate(MLAlg
   arma::Row<size_t> predictedLabels;
   model.Classify(data, predictedLabels);
 
-  size_t tp = arma::sum((labels == PC) % (predictedLabels == PC));
-  size_t numberOfPositivePredictions = arma::sum(predictedLabels == PC);
+  size_t tp = sum((labels == PC) % (predictedLabels == PC));
+  size_t numberOfPositivePredictions = sum(predictedLabels == PC);
 
   return double(tp) / numberOfPositivePredictions;
 }
@@ -72,8 +72,8 @@ double Precision<AS, PC>::Evaluate(MLAlg
   arma::vec precisions = arma::vec(numClasses);
   for (size_t c = 0; c < numClasses; ++c)
   {
-    size_t tp = arma::sum((labels == c) % (predictedLabels == c));
-    size_t numberOfPositivePredictions = arma::sum(predictedLabels == c);
+    size_t tp = sum((labels == c) % (predictedLabels == c));
+    size_t numberOfPositivePredictions = sum(predictedLabels == c);
     precisions(c) = double(tp) / numberOfPositivePredictions;
   }
 
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/r2_score_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/r2_score_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/r2_score_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/r2_score_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,11 +30,11 @@ double R2Score<AdjustedR2>::Evaluate(MLA
   double meanResponses = arma::mean(responses);
 
   // Calculate the numerator i.e. residual sum of squares.
-  double residualSumSquared = arma::accu(arma::square(responses -
+  double residualSumSquared = accu(arma::square(responses -
       predictedResponses));
 
   // Calculate the denominator i.e.total sum of squares.
-  double totalSumSquared = arma::accu(arma::square(responses - meanResponses));
+  double totalSumSquared = accu(arma::square(responses - meanResponses));
 
   // Handling undefined R2 Score when both denominator and numerator is 0.0.
   if (residualSumSquared == 0.0)
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/recall_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/recall_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/recall_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/recall_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,8 +36,8 @@ double Recall<AS, PC>::Evaluate(MLAlgori
   arma::Row<size_t> predictedLabels;
   model.Classify(data, predictedLabels);
 
-  size_t tp = arma::sum((labels == PC) % (predictedLabels == PC));
-  size_t numberOfPositiveClassInstances = arma::sum(labels == PC);
+  size_t tp = sum((labels == PC) % (predictedLabels == PC));
+  size_t numberOfPositiveClassInstances = sum(labels == PC);
 
   return double(tp) / numberOfPositiveClassInstances;
 }
@@ -72,8 +72,8 @@ double Recall<AS, PC>::Evaluate(MLAlgori
   arma::vec recalls = arma::vec(numClasses);
   for (size_t c = 0; c < numClasses; ++c)
   {
-    size_t tp = arma::sum((labels == c) % (predictedLabels == c));
-    size_t positiveLabels = arma::sum(labels == c);
+    size_t tp = sum((labels == c) % (predictedLabels == c));
+    size_t positiveLabels = sum(labels == c);
     recalls(c) = double(tp) / positiveLabels;
   }
 
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/roc_auc_score_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/roc_auc_score_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/roc_auc_score_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/roc_auc_score_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,7 +35,7 @@ double ROCAUCScore<PositiveClass>::Evalu
   // Converting probability scores of PositiveClass, from row to column vector.
   arma::vec colScores = arma::conv_to<arma::vec>::from(scores);
 
-  size_t numberOfTrueLabels  = arma::sum(binaryLabels);
+  size_t numberOfTrueLabels  = sum(binaryLabels);
   size_t numberOfFalseLabels = binaryLabels.n_rows - numberOfTrueLabels;
 
   // Check if only one class is given in labels.
diff -pruN 4.3.0-2/src/mlpack/core/cv/metrics/silhouette_score_impl.hpp 4.4.0-1/src/mlpack/core/cv/metrics/silhouette_score_impl.hpp
--- 4.3.0-2/src/mlpack/core/cv/metrics/silhouette_score_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/cv/metrics/silhouette_score_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -98,7 +98,7 @@ inline double SilhouetteScore::MeanDista
   }
   else
   {
-    double distance = arma::accu(distances.elem(sameClusterIndices));
+    double distance = accu(distances.elem(sameClusterIndices));
     distance /= (numSameCluster - sameCluster);
     return distance;
   }
diff -pruN 4.3.0-2/src/mlpack/core/data/detect_file_type.hpp 4.4.0-1/src/mlpack/core/data/detect_file_type.hpp
--- 4.3.0-2/src/mlpack/core/data/detect_file_type.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/detect_file_type.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,6 +64,12 @@ inline FileType AutoDetect(std::fstream&
  */
 inline FileType DetectFromExtension(const std::string& filename);
 
+/**
+ * Count the number of columns in the file.  The file must be a CSV/TSV/TXT file
+ * with no header.
+ */
+inline size_t CountCols(std::fstream& stream);
+
 } // namespace data
 } // namespace mlpack
 
diff -pruN 4.3.0-2/src/mlpack/core/data/detect_file_type_impl.hpp 4.4.0-1/src/mlpack/core/data/detect_file_type_impl.hpp
--- 4.3.0-2/src/mlpack/core/data/detect_file_type_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/detect_file_type_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,6 +35,7 @@ inline std::string GetStringType(const F
     case FileType::ArmaBinary:  return "Armadillo binary formatted data";
     case FileType::PGMBinary:   return "PGM data";
     case FileType::HDF5Binary:  return "HDF5 data";
+    case FileType::CoordASCII:  return "ASCII formatted sparse coordinate data";
     default:                    return "";
   }
 }
@@ -264,6 +265,7 @@ inline FileType AutoDetect(std::fstream&
     // This could be raw binary or Armadillo binary (binary with header).  We
     // will check to see if it is Armadillo binary.
     const std::string ARMA_MAT_BIN = "ARMA_MAT_BIN";
+    const std::string ARMA_SPM_BIN = "ARMA_SPM_BIN";
     std::string rawHeader(ARMA_MAT_BIN.length(), '\0');
 
     std::streampos pos = stream.tellg();
@@ -272,7 +274,7 @@ inline FileType AutoDetect(std::fstream&
     stream.clear();
     stream.seekg(pos); // Reset stream position after peeking.
 
-    if (rawHeader == ARMA_MAT_BIN)
+    if (rawHeader == ARMA_MAT_BIN || rawHeader == ARMA_SPM_BIN)
     {
       detectedLoadType = FileType::ArmaBinary;
     }
@@ -335,5 +337,32 @@ inline FileType DetectFromExtension(cons
   }
 }
 
+/**
+ * Count the number of columns in the file.  The file must be a CSV/TSV/TXT file
+ * with no header.
+ */
+inline size_t CountCols(std::fstream& f)
+{
+  f.clear();
+  const std::fstream::pos_type pos1 = f.tellg();
+
+  std::string firstLine;
+  std::getline(f, firstLine);
+
+  // Extract tokens from the first line using whitespace.
+  std::stringstream str(firstLine);
+  size_t cols = 0;
+  std::string token;
+
+  while (str >> token)
+    ++cols;
+
+  // Reset to wherever we were.
+  f.clear();
+  f.seekg(pos1);
+
+  return cols;
+}
+
 } // namespace data
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/core/data/load.hpp 4.4.0-1/src/mlpack/core/data/load.hpp
--- 4.3.0-2/src/mlpack/core/data/load.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/load.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -83,6 +83,7 @@ bool Load(const std::string& filename,
  *
  * The supported types of files are the same as found in Armadillo:
  *
+ *  - CSV (coord_ascii), denoted by .csv or .txt
  *  - TSV (coord_ascii), denoted by .tsv or .txt
  *  - TXT (coord_ascii), denoted by .txt
  *  - Raw binary (raw_binary), denoted by .bin
@@ -109,7 +110,8 @@ template<typename eT>
 bool Load(const std::string& filename,
           arma::SpMat<eT>& matrix,
           const bool fatal = false,
-          const bool transpose = true);
+          const bool transpose = true,
+          const FileType inputLoadType = FileType::AutoDetect);
 
 /**
  * Load a column vector from a file, guessing the filetype from the extension.
diff -pruN 4.3.0-2/src/mlpack/core/data/load_image_impl.hpp 4.4.0-1/src/mlpack/core/data/load_image_impl.hpp
--- 4.3.0-2/src/mlpack/core/data/load_image_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/load_image_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -154,14 +154,16 @@ inline bool LoadImage(const std::string&
 
   info.Width() = tempWidth;
   info.Height() = tempHeight;
-  info.Channels() = tempChannels;
+  // Only set the new number of channels if we didn't force grayscale loading.
+  if (info.Channels() != 1)
+    info.Channels() = tempChannels;
 
   // Copy image into armadillo Mat.
   matrix = arma::Mat<unsigned char>(image, info.Width() * info.Height() *
       info.Channels(), 1, true, true);
 
   // Free the image pointer.
-  free(image);
+  stbi_image_free(image);
   return true;
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/data/load_impl.hpp 4.4.0-1/src/mlpack/core/data/load_impl.hpp
--- 4.3.0-2/src/mlpack/core/data/load_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/load_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -63,7 +63,7 @@ bool inline inplace_transpose(MatType& X
 {
   try
   {
-    X = arma::trans(X);
+    X = trans(X);
     return true;
   }
   catch (const std::exception& e)
@@ -123,7 +123,7 @@ bool Load(const std::string& filename,
             << "incorrect extension?" << std::endl;
       else
         Log::Warn << "Unable to detect type of '" << filename << "'; load "
-            << " failed. Incorrect extension?" << std::endl;
+            << "failed. Incorrect extension?" << std::endl;
 
       return false;
     }
@@ -160,7 +160,7 @@ bool Load(const std::string& filename,
   // We can't use the stream if the type is HDF5.
   bool success;
   LoadCSV loader;
-  
+
   if (loadType != FileType::HDF5Binary)
   {
     if (loadType == FileType::CSVASCII)
@@ -299,7 +299,8 @@ template <typename eT>
 bool Load(const std::string& filename,
           arma::SpMat<eT>& matrix,
           const bool fatal,
-          const bool transpose)
+          const bool transpose,
+          const FileType inputLoadType)
 {
   Timer::Start("loading_data");
 
@@ -325,71 +326,85 @@ bool Load(const std::string& filename,
     return false;
   }
 
-  bool unknownType = false;
-  arma::file_type loadType;
+  FileType loadType = inputLoadType;
   std::string stringType;
-
-  if (extension == "tsv" || extension == "txt")
-  {
-    loadType = arma::coord_ascii;
-    stringType = "Coordinate Formatted Data for Sparse Matrix";
-  }
-  else if (extension == "bin")
+  if (inputLoadType == FileType::AutoDetect)
   {
-    // This could be raw binary or Armadillo binary (binary with header).  We
-    // will check to see if it is Armadillo binary.
-    const std::string ARMA_SPM_BIN = "ARMA_SPM_BIN";
-    std::string rawHeader(ARMA_SPM_BIN.length(), '\0');
-
-    std::streampos pos = stream.tellg();
-
-    stream.read(&rawHeader[0], std::streamsize(ARMA_SPM_BIN.length()));
-    stream.clear();
-    stream.seekg(pos); // Reset stream position after peeking.
-
-    if (rawHeader == ARMA_SPM_BIN)
+    // Attempt to auto-detect the type from the given file.
+    loadType = AutoDetect(stream, filename);
+    // Provide an error if we don't know the type, or if the type can't be used
+    // to load sparse matrices.
+    if (loadType == FileType::FileTypeUnknown)
     {
-      stringType = "Armadillo binary formatted data for sparse matrix";
-      loadType = arma::arma_binary;
+      if (fatal)
+        Log::Fatal << "Unable to detect type of '" << filename << "'; "
+            << "incorrect extension?" << std::endl;
+      else
+        Log::Warn << "Unable to detect type of '" << filename << "'; load "
+            << "failed. Incorrect extension?" << std::endl;
+
+      return false;
     }
-    else // We can only assume it's raw binary.
+
+    // There is still a small amount of differentiation that needs to be done:
+    // if we got a text type, it could be a coordinate list.  We will make an
+    // educated guess based on the shape of the input.
+    if (loadType == FileType::RawASCII)
     {
-      stringType = "raw binary formatted data";
-      loadType = arma::raw_binary;
+      // Get the number of columns in the file.  If it is the right shape, we
+      // will assume it is sparse.
+      const size_t cols = CountCols(stream);
+      if (cols == 3)
+      {
+        // We have the right number of columns, so assume the type is a
+        // coordinate list.
+        loadType = FileType::CoordASCII;
+      }
     }
   }
-  else // Unknown extension...
-  {
-    unknownType = true;
-    loadType = arma::raw_binary; // Won't be used; prevent a warning.
-    stringType = "";
-  }
 
-  // Provide error if we don't know the type.
-  if (unknownType)
+  // Filter out invalid types.
+  if ((loadType == FileType::PGMBinary) ||
+      (loadType == FileType::PPMBinary) ||
+      (loadType == FileType::ArmaASCII) ||
+      (loadType == FileType::RawBinary))
   {
-    Timer::Stop("loading_data");
     if (fatal)
-      Log::Fatal << "Unable to detect type of '" << filename << "'; "
-          << "incorrect extension?" << std::endl;
+      Log::Fatal << "Cannot load '" << filename << "' with type "
+          << GetStringType(loadType) << " into a sparse matrix; format is "
+          << "only supported for dense matrices." << std::endl;
     else
-      Log::Warn << "Unable to detect type of '" << filename << "'; load failed."
-          << " Incorrect extension?" << std::endl;
+      Log::Warn << "Cannot load '" << filename << "' with type "
+          << GetStringType(loadType) << " into a sparse matrix; format is "
+          << "only supported for dense matrices; load failed." << std::endl;
 
     return false;
   }
 
-  // Try to load the file; but if it's raw_binary, it could be a problem.
-  if (loadType == arma::raw_binary)
-    Log::Warn << "Loading '" << filename << "' as " << stringType << "; "
-        << "but this may not be the actual filetype!" << std::endl;
-  else
-    Log::Info << "Loading '" << filename << "' as " << stringType << ".  "
-        << std::flush;
-
   bool success;
 
-  success = matrix.load(stream, loadType);
+  if (loadType == FileType::CSVASCII)
+  {
+    // Armadillo sparse matrices can't load CSVs, so we have to load a separate
+    // matrix to do that.  If the CSV has three columns, we assume it's a
+    // coordinate list.
+    arma::Mat<eT> dense;
+    success = dense.load(stream, ToArmaFileType(loadType));
+    if (dense.n_cols == 3)
+    {
+      arma::umat locations = arma::conv_to<arma::umat>::from(
+          dense.cols(0, 1).t());
+      matrix = arma::SpMat<eT>(locations, dense.col(2));
+    }
+    else
+    {
+      matrix = arma::conv_to<arma::SpMat<eT>>::from(dense);
+    }
+  }
+  else
+  {
+    success = matrix.load(stream, ToArmaFileType(loadType));
+  }
 
   if (!success)
   {
@@ -403,8 +418,10 @@ bool Load(const std::string& filename,
     return false;
   }
   else
+  {
     Log::Info << "Size is " << (transpose ? matrix.n_cols : matrix.n_rows)
         << " x " << (transpose ? matrix.n_rows : matrix.n_cols) << ".\n";
+  }
 
   // Now transpose the matrix, if necessary.
   if (transpose)
diff -pruN 4.3.0-2/src/mlpack/core/data/scaler_methods/max_abs_scaler.hpp 4.4.0-1/src/mlpack/core/data/scaler_methods/max_abs_scaler.hpp
--- 4.3.0-2/src/mlpack/core/data/scaler_methods/max_abs_scaler.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/scaler_methods/max_abs_scaler.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -54,7 +54,7 @@ class MaxAbsScaler
   template<typename MatType>
   void Fit(const MatType& input)
   {
-    itemMin = arma::min(input, 1);
+    itemMin = min(input, 1);
     itemMax = arma::max(input, 1);
     scale = arma::max(arma::abs(itemMin), arma::abs(itemMax));
     // Handling zeros in scale vector.
diff -pruN 4.3.0-2/src/mlpack/core/data/scaler_methods/mean_normalization.hpp 4.4.0-1/src/mlpack/core/data/scaler_methods/mean_normalization.hpp
--- 4.3.0-2/src/mlpack/core/data/scaler_methods/mean_normalization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/scaler_methods/mean_normalization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -55,7 +55,7 @@ class MeanNormalization
   void Fit(const MatType& input)
   {
     itemMean = arma::mean(input, 1);
-    itemMin = arma::min(input, 1);
+    itemMin = min(input, 1);
     itemMax = arma::max(input, 1);
     scale = itemMax - itemMin;
     // Handling zeros in scale vector.
diff -pruN 4.3.0-2/src/mlpack/core/data/scaler_methods/min_max_scaler.hpp 4.4.0-1/src/mlpack/core/data/scaler_methods/min_max_scaler.hpp
--- 4.3.0-2/src/mlpack/core/data/scaler_methods/min_max_scaler.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/scaler_methods/min_max_scaler.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,7 +73,7 @@ class MinMaxScaler
   template<typename MatType>
   void Fit(const MatType& input)
   {
-    itemMin = arma::min(input, 1);
+    itemMin = min(input, 1);
     itemMax = arma::max(input, 1);
     scale = itemMax - itemMin;
     // Handle zeros in scale vector.
diff -pruN 4.3.0-2/src/mlpack/core/data/scaler_methods/pca_whitening.hpp 4.4.0-1/src/mlpack/core/data/scaler_methods/pca_whitening.hpp
--- 4.3.0-2/src/mlpack/core/data/scaler_methods/pca_whitening.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/scaler_methods/pca_whitening.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -13,7 +13,6 @@
 #define MLPACK_CORE_DATA_PCA_WHITENING_SCALE_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/lin_alg.hpp>
 #include <mlpack/core/math/ccov.hpp>
 
 namespace mlpack {
@@ -93,7 +92,7 @@ class PCAWhitening
     }
     output.copy_size(input);
     output = (input.each_col() - itemMean);
-    output = arma::diagmat(1.0 / (arma::sqrt(eigenValues))) * eigenVectors.t()
+    output = arma::diagmat(1.0 / (sqrt(eigenValues))) * eigenVectors.t()
         * output;
   }
 
@@ -106,7 +105,7 @@ class PCAWhitening
   template<typename MatType>
   void InverseTransform(const MatType& input, MatType& output)
   {
-    output = arma::diagmat(arma::sqrt(eigenValues)) * inv(eigenVectors.t())
+    output = arma::diagmat(sqrt(eigenValues)) * inv(eigenVectors.t())
         * input;
     output = (output.each_col() + itemMean);
   }
diff -pruN 4.3.0-2/src/mlpack/core/data/scaler_methods/zca_whitening.hpp 4.4.0-1/src/mlpack/core/data/scaler_methods/zca_whitening.hpp
--- 4.3.0-2/src/mlpack/core/data/scaler_methods/zca_whitening.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/scaler_methods/zca_whitening.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -13,7 +13,6 @@
 #define MLPACK_CORE_DATA_ZCA_WHITENING_SCALE_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/lin_alg.hpp>
 #include <mlpack/core/data/scaler_methods/pca_whitening.hpp>
 
 namespace mlpack {
@@ -87,7 +86,7 @@ class ZCAWhitening
   template<typename MatType>
   void InverseTransform(const MatType& input, MatType& output)
   {
-    output = inv(pca.EigenVectors()) * arma::diagmat(arma::sqrt(
+    output = inv(pca.EigenVectors()) * arma::diagmat(sqrt(
         pca.EigenValues())) * inv(pca.EigenVectors().t()) * input;
     output = (output.each_col() + pca.ItemMean());
   }
diff -pruN 4.3.0-2/src/mlpack/core/data/string_encoding.hpp 4.4.0-1/src/mlpack/core/data/string_encoding.hpp
--- 4.3.0-2/src/mlpack/core/data/string_encoding.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/string_encoding.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -74,7 +74,7 @@ class StringEncoding
    * @param tokenizer The tokenizer object.
    *
    * The tokenization algorithm has to be an object with two public methods:
-   * 1. operator() which accepts a reference to MLPACK_STRING_VIEW, extracts
+   * 1. operator() which accepts a reference to std::string_view, extracts
    * the next token from the given view, removes the prefix containing
    * the extracted token and returns the token;
    * 2. IsTokenEmpty() that accepts a token and returns true if the given
@@ -108,7 +108,7 @@ class StringEncoding
    * @param tokenizer The tokenizer object.
    *
    * The tokenization algorithm has to be an object with two public methods:
-   * 1. operator() which accepts a reference to MLPACK_STRING_VIEW, extracts
+   * 1. operator() which accepts a reference to std::string_view, extracts
    * the next token from the given view, removes the prefix containing
    * the extracted token and returns the token;
    * 2. IsTokenEmpty() that accepts a token and returns true if the given
@@ -158,7 +158,7 @@ class StringEncoding
    * @param policy The policy object.
    *
    * The tokenization algorithm has to be an object with two public methods:
-   * 1. operator() which accepts a reference to MLPACK_STRING_VIEW, extracts
+   * 1. operator() which accepts a reference to std::string_view, extracts
    * the next token from the given view, removes the prefix containing
    * the extracted token and returns the token;
    * 2. IsTokenEmpty() that accepts a token and returns true if the given
@@ -187,7 +187,7 @@ class StringEncoding
    * @param policy The policy object.
    *
    * The tokenization algorithm has to be an object with two public methods:
-   * 1. operator() which accepts a reference to MLPACK_STRING_VIEW, extracts
+   * 1. operator() which accepts a reference to std::string_view, extracts
    * the next token from the given view, removes the prefix containing
    * the extracted token and returns the token;
    * 2. IsTokenEmpty() that accepts a token and returns true if the given
diff -pruN 4.3.0-2/src/mlpack/core/data/string_encoding_dictionary.hpp 4.4.0-1/src/mlpack/core/data/string_encoding_dictionary.hpp
--- 4.3.0-2/src/mlpack/core/data/string_encoding_dictionary.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/string_encoding_dictionary.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -106,20 +106,20 @@ class StringEncodingDictionary
 };
 
 /*
- * Specialization of the StringEncodingDictionary class for MLPACK_STRING_VIEW.
+ * Specialization of the StringEncodingDictionary class for std::string_view.
  */
 template<>
-class StringEncodingDictionary<MLPACK_STRING_VIEW>
+class StringEncodingDictionary<std::string_view>
 {
  public:
   //! A convenient alias for the internal type of the map.
   using MapType = std::unordered_map<
-      MLPACK_STRING_VIEW,
+      std::string_view,
       size_t,
-      std::hash<MLPACK_STRING_VIEW>>;
+      std::hash<std::string_view>>;
 
   //! The type of the token that the dictionary stores.
-  using TokenType = MLPACK_STRING_VIEW;
+  using TokenType = std::string_view;
 
   //! Construct the default class.
   StringEncodingDictionary() = default;
@@ -156,7 +156,7 @@ class StringEncodingDictionary<MLPACK_ST
    *
    * @param token The given token.
    */
-  bool HasToken(const MLPACK_STRING_VIEW token) const
+  bool HasToken(const std::string_view token) const
   {
     return mapping.find(token) != mapping.end();
   }
@@ -168,7 +168,7 @@ class StringEncodingDictionary<MLPACK_ST
    *
    * @param token The given token.
    */
-  size_t AddToken(const MLPACK_STRING_VIEW token)
+  size_t AddToken(const std::string_view token)
   {
     tokens.emplace_back(token);
 
@@ -185,7 +185,7 @@ class StringEncodingDictionary<MLPACK_ST
    *
    * @param token The given token.
    */
-  size_t Value(const MLPACK_STRING_VIEW token) const
+  size_t Value(const std::string_view token) const
   {
     return mapping.at(token);
   }
diff -pruN 4.3.0-2/src/mlpack/core/data/string_encoding_impl.hpp 4.4.0-1/src/mlpack/core/data/string_encoding_impl.hpp
--- 4.3.0-2/src/mlpack/core/data/string_encoding_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/string_encoding_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -66,7 +66,7 @@ void StringEncoding<EncodingPolicyType,
     const std::string& input,
     const TokenizerType& tokenizer)
 {
-  MLPACK_STRING_VIEW strView(input);
+  std::string_view strView(input);
   auto token = tokenizer(strView);
 
   static_assert(
@@ -112,7 +112,7 @@ EncodeHelper(const std::vector<std::stri
   // The first pass adds the extracted tokens to the dictionary.
   for (size_t i = 0; i < input.size(); ++i)
   {
-    MLPACK_STRING_VIEW strView(input[i]);
+    std::string_view strView(input[i]);
     auto token = tokenizer(strView);
 
     static_assert(
@@ -143,7 +143,7 @@ EncodeHelper(const std::vector<std::stri
   // The second pass writes the encoded values to the output.
   for (size_t i = 0; i < input.size(); ++i)
   {
-    MLPACK_STRING_VIEW strView(input[i]);
+    std::string_view strView(input[i]);
     auto token = tokenizer(strView);
     size_t numTokens = 0;
 
@@ -172,7 +172,7 @@ EncodeHelper(const std::vector<std::stri
   // at once.
   for (size_t i = 0; i < input.size(); ++i)
   {
-    MLPACK_STRING_VIEW strView(input[i]);
+    std::string_view strView(input[i]);
     auto token = tokenizer(strView);
 
     static_assert(
diff -pruN 4.3.0-2/src/mlpack/core/data/tokenizers/char_extract.hpp 4.4.0-1/src/mlpack/core/data/tokenizers/char_extract.hpp
--- 4.3.0-2/src/mlpack/core/data/tokenizers/char_extract.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/tokenizers/char_extract.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ class CharExtract
    *
    * @param str String view to retrieve the next token from.
    */
-  int operator()(MLPACK_STRING_VIEW& str) const
+  int operator()(std::string_view& str) const
   {
     if (str.empty())
       return EOF;
diff -pruN 4.3.0-2/src/mlpack/core/data/tokenizers/split_by_any_of.hpp 4.4.0-1/src/mlpack/core/data/tokenizers/split_by_any_of.hpp
--- 4.3.0-2/src/mlpack/core/data/tokenizers/split_by_any_of.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/tokenizers/split_by_any_of.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ class SplitByAnyOf
 {
  public:
   //! The type of the token which the tokenizer extracts.
-  using TokenType = MLPACK_STRING_VIEW;
+  using TokenType = std::string_view;
 
   //! A convenient alias for the mask type.
   using MaskType = std::array<bool, 1 << CHAR_BIT>;
@@ -37,7 +37,7 @@ class SplitByAnyOf
    *
    * @param delimiters The given delimiters.
    */
-  SplitByAnyOf(const MLPACK_STRING_VIEW delimiters)
+  SplitByAnyOf(const std::string_view delimiters)
   {
     mask.fill(false);
 
@@ -51,13 +51,13 @@ class SplitByAnyOf
    *
    * @param str String view to retrieve the token from.
    */
-  MLPACK_STRING_VIEW operator()(MLPACK_STRING_VIEW& str) const
+  std::string_view operator()(std::string_view& str) const
   {
-    MLPACK_STRING_VIEW retval;
+    std::string_view retval;
     // std::basic_string_view does not have empty function.
     // Therefore, we are assiging an empty string when reaching the last
     // delimiter.
-    MLPACK_STRING_VIEW empty_string{""};
+    std::string_view empty_string{""};
 
     while (retval.empty())
     {
@@ -79,7 +79,7 @@ class SplitByAnyOf
    *
    * @param token The given token.
    */
-  static bool IsTokenEmpty(const MLPACK_STRING_VIEW token)
+  static bool IsTokenEmpty(const std::string_view token)
   {
     return token.empty();
   }
@@ -93,11 +93,11 @@ class SplitByAnyOf
   /**
    * The function finds the first character in the given string view equal to 
    * any of the delimiters and returns the position of the character or 
-   * MLPACK_STRING_VIEW::npos if no such character is found.
+   * std::string_view::npos if no such character is found.
    *
    * @param str String where to find the character.
    */
-  size_t FindFirstDelimiter(const MLPACK_STRING_VIEW str) const
+  size_t FindFirstDelimiter(const std::string_view str) const
   {
     for (size_t pos = 0; pos < str.size(); pos++)
     {
diff -pruN 4.3.0-2/src/mlpack/core/data/types_impl.hpp 4.4.0-1/src/mlpack/core/data/types_impl.hpp
--- 4.3.0-2/src/mlpack/core/data/types_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/data/types_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,43 +35,43 @@ inline arma::file_type ToArmaFileType(co
     case FileType::AutoDetect:
       return arma::auto_detect;
       break;
-    
+
     case FileType::RawASCII:
       return arma::raw_ascii;
       break;
-    
+
     case FileType::ArmaASCII:
       return arma::arma_ascii;
       break;
-    
+
     case FileType::CSVASCII:
       return arma::csv_ascii;
       break;
-    
+
     case FileType::RawBinary:
       return arma::raw_binary;
       break;
-    
+
     case FileType::ArmaBinary:
       return arma::arma_binary;
       break;
-    
+
     case FileType::PGMBinary:
       return arma::pgm_binary;
       break;
-    
+
     case FileType::PPMBinary:
       return arma::ppm_binary;
       break;
-    
+
     case FileType::HDF5Binary:
       return arma::hdf5_binary;
       break;
-    
+
     case FileType::CoordASCII:
       return arma::coord_ascii;
       break;
-    
+
     default:
       return arma::file_type_unknown;
       break;
diff -pruN 4.3.0-2/src/mlpack/core/dists/diagonal_gaussian_distribution.hpp 4.4.0-1/src/mlpack/core/dists/diagonal_gaussian_distribution.hpp
--- 4.3.0-2/src/mlpack/core/dists/diagonal_gaussian_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/diagonal_gaussian_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,7 +65,7 @@ class DiagonalGaussianDistribution
   //! Return the probability of the given observation.
   double Probability(const arma::vec& observation) const
   {
-    return exp(LogProbability(observation));
+    return std::exp(LogProbability(observation));
   }
 
   //! Return the log probability of the given observation.
@@ -82,7 +82,7 @@ class DiagonalGaussianDistribution
   {
     arma::vec logProbabilities;
     LogProbability(x, logProbabilities);
-    probabilities = arma::exp(logProbabilities);
+    probabilities = exp(logProbabilities);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/core/dists/diagonal_gaussian_distribution_impl.hpp 4.4.0-1/src/mlpack/core/dists/diagonal_gaussian_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/core/dists/diagonal_gaussian_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/diagonal_gaussian_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -28,14 +28,14 @@ inline DiagonalGaussianDistribution::Dia
 inline void DiagonalGaussianDistribution::Covariance(const arma::vec& covariance)
 {
   invCov = 1 / covariance;
-  logDetCov = arma::accu(log(covariance));
+  logDetCov = accu(log(covariance));
   this->covariance = covariance;
 }
 
 inline void DiagonalGaussianDistribution::Covariance(arma::vec&& covariance)
 {
   invCov = 1 / covariance;
-  logDetCov = arma::accu(log(covariance));
+  logDetCov = accu(log(covariance));
   this->covariance = std::move(covariance);
 }
 
@@ -60,14 +60,14 @@ inline void DiagonalGaussianDistribution
 
   // Calculates log of exponent equation in multivariate Gaussian
   // distribution. We use only diagonal part for faster computation.
-  arma::vec logExponents = -0.5 * arma::trans(diffs % diffs) * invCov;
+  arma::vec logExponents = -0.5 * trans(diffs % diffs) * invCov;
 
   logProbabilities = -0.5 * k * log2pi - 0.5 * logDetCov + logExponents;
 }
 
 inline arma::vec DiagonalGaussianDistribution::Random() const
 {
-  return (arma::sqrt(covariance) % arma::randn<arma::vec>(mean.n_elem)) + mean;
+  return (sqrt(covariance) % arma::randn<arma::vec>(mean.n_elem)) + mean;
 }
 
 inline void DiagonalGaussianDistribution::Train(const arma::mat& observations)
@@ -84,17 +84,17 @@ inline void DiagonalGaussianDistribution
   }
 
   // Calculate and normalize the mean.
-  mean = arma::sum(observations, 1) / observations.n_cols;
+  mean = sum(observations, 1) / observations.n_cols;
 
   // Now calculate the covariance.
   const arma::mat diffs = observations.each_col() - mean;
-  covariance += arma::sum(diffs % diffs, 1);
+  covariance += sum(diffs % diffs, 1);
 
   // Finish estimating the covariance by normalizing, with the (1 / (n - 1))
   // to make the estimator unbiased.
   covariance /= (observations.n_cols - 1);
   invCov = 1 / covariance;
-  logDetCov = arma::accu(log(covariance));
+  logDetCov = accu(log(covariance));
 }
 
 inline void DiagonalGaussianDistribution::Train(const arma::mat& observations,
@@ -116,14 +116,14 @@ inline void DiagonalGaussianDistribution
   // of the weights, and the v2 is the sum of the each weight squared.
   // If you want to know more detailed description,
   // please refer to https://en.wikipedia.org/wiki/Weighted_arithmetic_mean.
-  double v1 = arma::accu(probabilities);
+  double v1 = accu(probabilities);
 
   // If their sum is 0, there is nothing in this Gaussian.
   // At least, set the covariance so that it's invertible.
   if (v1 == 0)
   {
     invCov = 1 / (covariance += 1e-50);
-    logDetCov = arma::accu(log(covariance));
+    logDetCov = accu(log(covariance));
     return;
   }
 
@@ -138,7 +138,7 @@ inline void DiagonalGaussianDistribution
   covariance += (diffs % diffs) * normalizedProbs;
 
   // Calculate the sum of each weight squared.
-  const double v2 = arma::accu(normalizedProbs % normalizedProbs);
+  const double v2 = accu(normalizedProbs % normalizedProbs);
 
   // Finish estimating the covariance by normalizing, with
   // the (1 / (v1 - (v2 / v1))) to make the estimator unbiased.
@@ -146,7 +146,7 @@ inline void DiagonalGaussianDistribution
     covariance /= (1 - v2);
 
   invCov = 1 / covariance;
-  logDetCov = arma::accu(log(covariance));
+  logDetCov = accu(log(covariance));
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/core/dists/discrete_distribution.hpp 4.4.0-1/src/mlpack/core/dists/discrete_distribution.hpp
--- 4.3.0-2/src/mlpack/core/dists/discrete_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/discrete_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -161,7 +161,7 @@ class DiscreteDistribution
   double LogProbability(const arma::vec& observation) const
   {
     // TODO: consider storing log probabilities instead?
-    return log(Probability(observation));
+    return std::log(Probability(observation));
   }
 
   /**
@@ -190,7 +190,7 @@ class DiscreteDistribution
   {
     logProbabilities.set_size(x.n_cols);
     for (size_t i = 0; i < x.n_cols; ++i)
-      logProbabilities(i) = log(Probability(x.unsafe_col(i)));
+      logProbabilities(i) = std::log(Probability(x.unsafe_col(i)));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/core/dists/gamma_distribution_impl.hpp 4.4.0-1/src/mlpack/core/dists/gamma_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/core/dists/gamma_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/gamma_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -56,9 +56,9 @@ inline void GammaDistribution::Train(con
     return;
 
   // Calculate log(mean(x)) and mean(log(x)) of each dataset row.
-  const arma::vec meanLogxVec = arma::mean(arma::log(rdata), 1);
+  const arma::vec meanLogxVec = arma::mean(log(rdata), 1);
   const arma::vec meanxVec = arma::mean(rdata, 1);
-  const arma::vec logMeanxVec = arma::log(meanxVec);
+  const arma::vec logMeanxVec = log(meanxVec);
 
   // Call the statistics-only GammaDistribution::Train() function to fit the
   // parameters. That function does all the work so we're done.
@@ -80,15 +80,15 @@ inline void GammaDistribution::Train(con
 
   for (size_t i = 0; i < rdata.n_cols; ++i)
   {
-    meanLogxVec += probabilities(i) * arma::log(rdata.col(i));
+    meanLogxVec += probabilities(i) * log(rdata.col(i));
     meanxVec += probabilities(i) * rdata.col(i);
   }
 
-  double totProbability = arma::accu(probabilities);
+  double totProbability = accu(probabilities);
 
   meanLogxVec /= totProbability;
   meanxVec /= totProbability;
-  logMeanxVec = arma::log(meanxVec);
+  logMeanxVec = log(meanxVec);
 
   // Call the statistics-only GammaDistribution::Train() function to fit the
   // parameters. That function does all the work so we're done.
@@ -135,8 +135,8 @@ inline void GammaDistribution::Train(con
       aOld = aEst;
 
       // Calculate new value for alpha.
-      double nominator = meanLogx - logMeanx + log(aEst) - Digamma(aEst);
-      double denominator = pow(aEst, 2) * (1 / aEst - Trigamma(aEst));
+      double nominator = meanLogx - logMeanx + std::log(aEst) - Digamma(aEst);
+      double denominator = std::pow(aEst, 2) * (1 / aEst - Trigamma(aEst));
 
       // Protect against division by 0.
       if (denominator == 0)
diff -pruN 4.3.0-2/src/mlpack/core/dists/gaussian_distribution.hpp 4.4.0-1/src/mlpack/core/dists/gaussian_distribution.hpp
--- 4.3.0-2/src/mlpack/core/dists/gaussian_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/gaussian_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -72,7 +72,7 @@ class GaussianDistribution
    */
   double Probability(const arma::vec& observation) const
   {
-    return exp(LogProbability(observation));
+    return std::exp(LogProbability(observation));
   }
 
   /**
@@ -93,7 +93,7 @@ class GaussianDistribution
     // logspace.
     arma::vec logProbs;
     LogProbability(x, logProbs);
-    probabilities = arma::exp(logProbs);
+    probabilities = exp(logProbs);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/core/dists/gaussian_distribution_impl.hpp 4.4.0-1/src/mlpack/core/dists/gaussian_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/core/dists/gaussian_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/gaussian_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,7 +67,7 @@ inline void GaussianDistribution::Factor
 
   invCov = invCovLower.t() * invCovLower;
   double sign = 0.;
-  arma::log_det(logDetCov, sign, covLower);
+  log_det(logDetCov, sign, covLower);
   logDetCov *= 2;
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/dists/laplace_distribution.hpp 4.4.0-1/src/mlpack/core/dists/laplace_distribution.hpp
--- 4.3.0-2/src/mlpack/core/dists/laplace_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/laplace_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -87,7 +87,7 @@ class LaplaceDistribution
    */
   double Probability(const arma::vec& observation) const
   {
-    return exp(LogProbability(observation));
+    return std::exp(LogProbability(observation));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/core/dists/laplace_distribution_impl.hpp 4.4.0-1/src/mlpack/core/dists/laplace_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/core/dists/laplace_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/laplace_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -25,7 +25,7 @@ inline double LaplaceDistribution::LogPr
 {
   // Evaluate the PDF of the Laplace distribution to determine
   // the log probability.
-  return -log(2. * scale) - arma::norm(observation - mean, 2) / scale;
+  return -std::log(2. * scale) - arma::norm(observation - mean, 2) / scale;
 }
 
 /**
@@ -90,7 +90,7 @@ inline void LaplaceDistribution::Estimat
   mean.zeros(observations.n_rows);
   for (size_t i = 0; i < observations.n_cols; ++i)
     mean += observations.col(i) * probabilities(i);
-  mean /= arma::accu(probabilities);
+  mean /= accu(probabilities);
 
   // This is the same formula as the previous function, but here we are
   // multiplying by the probability that the point is actually from
@@ -98,7 +98,7 @@ inline void LaplaceDistribution::Estimat
   scale = 0.0;
   for (size_t i = 0; i < observations.n_cols; ++i)
     scale += probabilities(i) * arma::norm(observations.col(i) - mean, 2);
-  scale /= arma::accu(probabilities);
+  scale /= accu(probabilities);
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/core/dists/regression_distribution.hpp 4.4.0-1/src/mlpack/core/dists/regression_distribution.hpp
--- 4.3.0-2/src/mlpack/core/dists/regression_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/regression_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,7 +31,7 @@ class RegressionDistribution
 {
  private:
   //! Regression function for representing conditional mean.
-  LinearRegression rf;
+  LinearRegression<> rf;
   //! Error distribution.
   GaussianDistribution err;
 
@@ -48,18 +48,6 @@ class RegressionDistribution
    * @param predictors Matrix of predictors (X).
    * @param responses Vector of responses (y).
    */
-  mlpack_deprecated RegressionDistribution(const arma::mat& predictors,
-                                           const arma::vec& responses) :
-    RegressionDistribution(predictors, arma::rowvec(responses.t()))
-  {}
-
-  /**
-   * Create a Conditional Gaussian distribution with conditional mean function
-   * obtained by running RegressionFunction on predictors, responses.
-   *
-   * @param predictors Matrix of predictors (X).
-   * @param responses Vector of responses (y).
-   */
   RegressionDistribution(const arma::mat& predictors,
                          const arma::rowvec& responses)
   {
@@ -81,9 +69,9 @@ class RegressionDistribution
   }
 
   //! Return regression function.
-  const LinearRegression& Rf() const { return rf; }
+  const LinearRegression<>& Rf() const { return rf; }
   //! Modify regression function.
-  LinearRegression& Rf() { return rf; }
+  LinearRegression<>& Rf() { return rf; }
 
   //! Return error distribution.
   const GaussianDistribution& Err() const { return err; }
@@ -103,15 +91,6 @@ class RegressionDistribution
    * @param observations List of observations.
    * @param weights Probability that given observation is from distribution.
    */
-  mlpack_deprecated void Train(const arma::mat& observations,
-                               const arma::vec& weights);
-
-  /**
-   * Estimate parameters using provided observation weights.
-   *
-   * @param observations List of observations.
-   * @param weights Probability that given observation is from distribution.
-   */
   void Train(const arma::mat& observations, const arma::rowvec& weights);
 
   /**
@@ -128,20 +107,11 @@ class RegressionDistribution
    */
   double LogProbability(const arma::vec& observation) const
   {
-    return log(Probability(observation));
+    return std::log(Probability(observation));
   }
 
   /**
    * Calculate y_i for each data point in points.
-   *
-   * @param points The data points to calculate with.
-   * @param predictions Y, will contain calculated values on completion.
-   */
-  mlpack_deprecated void Predict(const arma::mat& points,
-                                 arma::vec& predictions) const;
-
-  /**
-   * Calculate y_i for each data point in points.
    *
    * @param points The data points to calculate with.
    * @param predictions Y, will contain calculated values on completion.
diff -pruN 4.3.0-2/src/mlpack/core/dists/regression_distribution_impl.hpp 4.4.0-1/src/mlpack/core/dists/regression_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/core/dists/regression_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/dists/regression_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,7 +24,7 @@ namespace mlpack {
  */
 inline void RegressionDistribution::Train(const arma::mat& observations)
 {
-  LinearRegression lr(observations.rows(1, observations.n_rows - 1),
+  LinearRegression<> lr(observations.rows(1, observations.n_rows - 1),
       arma::rowvec(observations.row(0)), 0, true);
   rf = lr;
   arma::rowvec fitted;
@@ -32,21 +32,10 @@ inline void RegressionDistribution::Trai
   err.Train(observations.row(0) - fitted);
 }
 
-/**
- * Estimate parameters using provided observation weights.
- *
- * @param weights Probability that given observation is from distribution.
- */
-inline void RegressionDistribution::Train(const arma::mat& observations,
-                                          const arma::vec& weights)
-{
-  Train(observations, arma::rowvec(weights.t()));
-}
-
 inline void RegressionDistribution::Train(const arma::mat& observations,
                                           const arma::rowvec& weights)
 {
-  LinearRegression lr(observations.rows(1, observations.n_rows - 1),
+  LinearRegression<> lr(observations.rows(1, observations.n_rows - 1),
       arma::rowvec(observations.row(0)), weights, 0, true);
   rf = lr;
   arma::rowvec fitted;
@@ -63,16 +52,8 @@ inline double RegressionDistribution::Pr
     const arma::vec& observation) const
 {
   arma::rowvec fitted;
-  rf.Predict(observation.rows(1, observation.n_rows-1), fitted);
-  return err.Probability(observation(0)-fitted.t());
-}
-
-inline void RegressionDistribution::Predict(const arma::mat& points,
-                                            arma::vec& predictions) const
-{
-  arma::rowvec rowPredictions;
-  Predict(points, rowPredictions);
-  predictions = rowPredictions.t();
+  rf.Predict(observation.rows(1, observation.n_rows - 1), fitted);
+  return err.Probability(observation(0) - fitted.t());
 }
 
 inline void RegressionDistribution::Predict(const arma::mat& points,
diff -pruN 4.3.0-2/src/mlpack/core/kernels/cauchy_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/cauchy_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/cauchy_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/cauchy_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -76,6 +76,11 @@ class CauchyKernel
     ar(CEREAL_NVP(bandwidth));
   }
 
+  // Get the kernel bandwidth.
+  const double Bandwidth() const { return bandwidth; }
+  // Modify the kernel bandwidth.
+  void Bandwidth(const double bw) { this->bandwidth = bw; }
+
  private:
   //! Kernel bandwidth.
   double bandwidth;
diff -pruN 4.3.0-2/src/mlpack/core/kernels/cosine_distance.hpp 4.4.0-1/src/mlpack/core/kernels/cosine_distance.hpp
--- 4.3.0-2/src/mlpack/core/kernels/cosine_distance.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/cosine_distance.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,65 +0,0 @@
-/**
- * @file core/kernels/cosine_distance.hpp
- * @author Ryan Curtin
- *
- * This implements the cosine distance (or cosine similarity) between two
- * vectors, which is a measure of the angle between the two vectors.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_KERNELS_COSINE_DISTANCE_HPP
-#define MLPACK_CORE_KERNELS_COSINE_DISTANCE_HPP
-
-#include <mlpack/prereqs.hpp>
-#include <mlpack/core/kernels/kernel_traits.hpp>
-
-namespace mlpack {
-
-/**
- * The cosine distance (or cosine similarity).  It is defined by
- *
- * @f[
- * d(a, b) = \frac{a^T b}{|| a || || b ||}
- * @f]
- *
- * and this class assumes the standard L2 inner product.
- */
-class CosineDistance
-{
- public:
-  /**
-   * Computes the cosine distance between two points.
-   *
-   * @param a First vector.
-   * @param b Second vector.
-   * @return d(a, b).
-   */
-  template<typename VecTypeA, typename VecTypeB>
-  static double Evaluate(const VecTypeA& a, const VecTypeB& b);
-
-  //! Serialize the class (there's nothing to save).
-  template<typename Archive>
-  void serialize(Archive& /* ar */, const uint32_t /* version */) { }
-};
-
-//! Kernel traits for the cosine distance.
-template<>
-class KernelTraits<CosineDistance>
-{
- public:
-  //! The cosine kernel is normalized: K(x, x) = 1 for all x.
-  static const bool IsNormalized = true;
-
-  //! The cosine kernel doesn't include a squared distance.
-  static const bool UsesSquaredDistance = false;
-};
-
-} // namespace mlpack
-
-// Include implementation.
-#include "cosine_distance_impl.hpp"
-
-#endif
diff -pruN 4.3.0-2/src/mlpack/core/kernels/cosine_distance_impl.hpp 4.4.0-1/src/mlpack/core/kernels/cosine_distance_impl.hpp
--- 4.3.0-2/src/mlpack/core/kernels/cosine_distance_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/cosine_distance_impl.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,36 +0,0 @@
-/**
- * @file core/kernels/cosine_distance_impl.hpp
- * @author Ryan Curtin
- *
- * This implements the cosine distance.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_KERNELS_COSINE_DISTANCE_IMPL_HPP
-#define MLPACK_CORE_KERNELS_COSINE_DISTANCE_IMPL_HPP
-
-#include "cosine_distance.hpp"
-
-namespace mlpack {
-
-template<typename VecTypeA, typename VecTypeB>
-double CosineDistance::Evaluate(const VecTypeA& a, const VecTypeB& b)
-{
-  // Since we are using the L2 inner product, this is easy.  But we have to make
-  // sure we aren't dividing by zero (if we are, then the cosine similarity is
-  // 0: we reason this value because the cosine distance is just a normalized
-  // dot product; take away the normalization, and if ||a|| or ||b|| is equal to
-  // 0, then a^T b is zero too).
-  const double denominator = norm(a, 2) * norm(b, 2);
-  if (denominator == 0.0)
-    return 0;
-  else
-    return dot(a, b) / denominator;
-}
-
-} // namespace mlpack
-
-#endif
diff -pruN 4.3.0-2/src/mlpack/core/kernels/cosine_similarity.hpp 4.4.0-1/src/mlpack/core/kernels/cosine_similarity.hpp
--- 4.3.0-2/src/mlpack/core/kernels/cosine_similarity.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/cosine_similarity.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,68 @@
+/**
+ * @file core/kernels/cosine_similarity.hpp
+ * @author Ryan Curtin
+ *
+ * This implements the cosine similarity between two vectors, which is a measure
+ * of the angle between the two vectors.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_KERNELS_COSINE_SIMILARITY_HPP
+#define MLPACK_CORE_KERNELS_COSINE_SIMILARITY_HPP
+
+#include <mlpack/prereqs.hpp>
+#include <mlpack/core/kernels/kernel_traits.hpp>
+
+namespace mlpack {
+
+/**
+ * The cosine distance (or cosine similarity).  It is defined by
+ *
+ * @f[
+ * d(a, b) = \frac{a^T b}{|| a || || b ||}
+ * @f]
+ *
+ * and this class assumes the standard L2 inner product.
+ */
+class CosineSimilarity
+{
+ public:
+  /**
+   * Computes the cosine distance between two points.
+   *
+   * @param a First vector.
+   * @param b Second vector.
+   * @return d(a, b).
+   */
+  template<typename VecTypeA, typename VecTypeB>
+  static double Evaluate(const VecTypeA& a, const VecTypeB& b);
+
+  //! Serialize the class (there's nothing to save).
+  template<typename Archive>
+  void serialize(Archive& /* ar */, const uint32_t /* version */) { }
+};
+
+//! Kernel traits for the cosine distance.
+template<>
+class KernelTraits<CosineSimilarity>
+{
+ public:
+  //! The cosine kernel is normalized: K(x, x) = 1 for all x.
+  static const bool IsNormalized = true;
+
+  //! The cosine kernel doesn't include a squared distance.
+  static const bool UsesSquaredDistance = false;
+};
+
+// This name is deprecated and can be removed in mlpack 5.0.0.
+typedef CosineSimilarity CosineDistance;
+
+} // namespace mlpack
+
+// Include implementation.
+#include "cosine_similarity_impl.hpp"
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/kernels/cosine_similarity_impl.hpp 4.4.0-1/src/mlpack/core/kernels/cosine_similarity_impl.hpp
--- 4.3.0-2/src/mlpack/core/kernels/cosine_similarity_impl.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/cosine_similarity_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,36 @@
+/**
+ * @file core/kernels/cosine_similarity_impl.hpp
+ * @author Ryan Curtin
+ *
+ * This implements the cosine similarity.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_KERNELS_COSINE_SIMILARITY_IMPL_HPP
+#define MLPACK_CORE_KERNELS_COSINE_SIMILARITY_IMPL_HPP
+
+#include "cosine_similarity.hpp"
+
+namespace mlpack {
+
+template<typename VecTypeA, typename VecTypeB>
+double CosineSimilarity::Evaluate(const VecTypeA& a, const VecTypeB& b)
+{
+  // Since we are using the L2 inner product, this is easy.  But we have to make
+  // sure we aren't dividing by zero (if we are, then the cosine similarity is
+  // 0: we reason this value because the cosine distance is just a normalized
+  // dot product; take away the normalization, and if ||a|| or ||b|| is equal to
+  // 0, then a^T b is zero too).
+  const double denominator = norm(a, 2) * norm(b, 2);
+  if (denominator == 0.0)
+    return 0;
+  else
+    return dot(a, b) / denominator;
+}
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/kernels/epanechnikov_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/epanechnikov_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/epanechnikov_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/epanechnikov_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,30 +64,21 @@ class EpanechnikovKernel
   inline double Gradient(const double distance) const;
 
   /**
-   * Evaluate the Gradient of Epanechnikov kernel
-   * given that the squared distance between the two
-   * input points is known.
-   */
-  inline double GradientForSquaredDistance(const double distanceSquared) const;
-  /**
-   * Obtains the convolution integral [integral of K(||x-a||) K(||b-x||) dx]
-   * for the two vectors.
-   *
-   * @tparam VecType Type of vector (arma::vec, arma::spvec should be expected).
-   * @param a First vector.
-   * @param b Second vector.
-   * @return the convolution integral value.
-   */
-  template<typename VecTypeA, typename VecTypeB>
-  double ConvolutionIntegral(const VecTypeA& a, const VecTypeB& b);
-
-  /**
    * Compute the normalizer of this Epanechnikov kernel for the given dimension.
    *
    * @param dimension Dimension to calculate the normalizer for.
    */
   inline double Normalizer(const size_t dimension);
 
+  // Get the bandwidth of the kernel.
+  const double Bandwidth() const { return bandwidth; }
+  // Modify the bandwidth of the kernel.
+  void Bandwidth(const double bandwidth)
+  {
+    this->bandwidth = bandwidth;
+    this->inverseBandwidthSquared = 1.0 / (bandwidth * bandwidth);
+  }
+
   /**
    * Serialize the kernel.
    */
diff -pruN 4.3.0-2/src/mlpack/core/kernels/epanechnikov_kernel_impl.hpp 4.4.0-1/src/mlpack/core/kernels/epanechnikov_kernel_impl.hpp
--- 4.3.0-2/src/mlpack/core/kernels/epanechnikov_kernel_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/epanechnikov_kernel_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,57 +29,13 @@ inline double EpanechnikovKernel::Evalua
 }
 
 /**
- * Obtains the convolution integral [integral of K(||x-a||) K(||b-x||) dx]
- * for the two vectors.
- *
- * @tparam VecTypeA Type of first vector (arma::vec, arma::sp_vec should be
- *      expected).
- * @tparam VecTypeB Type of second vector (arma::vec, arma::sp_vec).
- * @param a First vector.
- * @param b Second vector.
- * @return the convolution integral value.
- */
-template<typename VecTypeA, typename VecTypeB>
-inline double EpanechnikovKernel::ConvolutionIntegral(const VecTypeA& a,
-                                                      const VecTypeB& b)
-{
-  double distance = sqrt(SquaredEuclideanDistance::Evaluate(a, b));
-  if (distance >= 2.0 * bandwidth)
-    return 0.0;
-
-  double volumeSquared = std::pow(Normalizer(a.n_rows), 2.0);
-
-  switch (a.n_rows)
-  {
-    case 1:
-      return 1.0 / volumeSquared *
-          (16.0 / 15.0 * bandwidth - 4.0 * distance * distance /
-          (3.0 * bandwidth) + 2.0 * distance * distance * distance /
-          (3.0 * bandwidth * bandwidth) -
-          std::pow(distance, 5.0) / (30.0 * std::pow(bandwidth, 4.0)));
-    case 2:
-      return 1.0 / volumeSquared *
-          ((2.0 / 3.0 * bandwidth * bandwidth - distance * distance) *
-          asin(sqrt(1.0 - std::pow(distance / (2.0 * bandwidth), 2.0))) +
-          sqrt(4.0 * bandwidth * bandwidth - distance * distance) *
-          (distance / 6.0 + 2.0 / 9.0 * distance *
-          std::pow(distance / bandwidth, 2.0) - distance / 72.0 *
-          std::pow(distance / bandwidth, 4.0)));
-    default:
-      Log::Fatal << "EpanechnikovKernel::ConvolutionIntegral(): dimension "
-          << a.n_rows << " not supported.";
-      return -1.0; // This line will not execute.
-  }
-}
-
-/**
  * Compute the normalizer of this Epanechnikov kernel for the given dimension.
  *
  * @param dimension Dimension to calculate the normalizer for.
  */
 inline double EpanechnikovKernel::Normalizer(const size_t dimension)
 {
-  return 2.0 * pow(bandwidth, (double) dimension) *
+  return 2.0 * std::pow(bandwidth, (double) dimension) *
       std::pow(M_PI, dimension / 2.0) /
       (std::tgamma(dimension / 2.0 + 1.0) * (dimension + 2.0));
 }
@@ -113,29 +69,6 @@ inline double EpanechnikovKernel::Gradie
   }
 }
 
-/**
- * Evaluate gradient of the kernel not for two points
- * but for a numerical value.
- */
-inline double EpanechnikovKernel::GradientForSquaredDistance(
-    const double distanceSquared) const
-{
-  double bandwidthSquared = bandwidth * bandwidth;
-  if (distanceSquared < bandwidthSquared)
-  {
-    return -1 * inverseBandwidthSquared;
-  }
-  else if (distanceSquared > bandwidthSquared &&
-           distanceSquared >= 0)
-  {
-    return  0;
-  }
-  else
-  {
-    // The gradient doesn't exist.
-    return arma::datum::nan;
-  }
-}
 //! Serialize the kernel.
 template<typename Archive>
 void EpanechnikovKernel::serialize(Archive& ar,
diff -pruN 4.3.0-2/src/mlpack/core/kernels/example_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/example_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/example_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/example_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -90,24 +90,6 @@ class ExampleKernel
   void serialize(Archive& /* ar */, const uint32_t /* version */) { }
 
   /**
-   * Obtains the convolution integral [integral K(||x-a||)K(||b-x||)dx]
-   * for the two vectors.  In this case, because
-   * our simple example kernel has no internal parameters, we can declare the
-   * function static.  For a more complex example which cannot be declared
-   * static, see the GaussianKernel, which stores an internal parameter.
-   *
-   * @tparam VecTypeA Type of first vector (arma::vec, arma::sp_vec should be
-   *      expected).
-   * @tparam VecTypeB Type of second vector (arma::vec, arma::sp_vec).
-   * @param * (a) First vector.
-   * @param * (b) Second vector.
-   * @return the convolution integral value.
-   */
-  template<typename VecTypeA, typename VecTypeB>
-  static double ConvolutionIntegral(const VecTypeA& /* a */,
-                                    const VecTypeB& /* b */) { return 0; }
-
-  /**
    * Obtains the normalizing volume for the kernel with dimension $dimension$.
    * In this case, because our simple example kernel has no internal parameters,
    * we can declare the function static.  For a more complex example which
diff -pruN 4.3.0-2/src/mlpack/core/kernels/gaussian_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/gaussian_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/gaussian_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/gaussian_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -46,7 +46,7 @@ class GaussianKernel
    */
   GaussianKernel(const double bandwidth) :
       bandwidth(bandwidth),
-      gamma(-0.5 * pow(bandwidth, -2.0))
+      gamma(-0.5 * std::pow(bandwidth, -2.0))
   { }
 
   /**
@@ -64,7 +64,7 @@ class GaussianKernel
   double Evaluate(const VecTypeA& a, const VecTypeB& b) const
   {
     // The precalculation of gamma saves us a little computation time.
-    return exp(gamma * SquaredEuclideanDistance::Evaluate(a, b));
+    return std::exp(gamma * SquaredEuclideanDistance::Evaluate(a, b));
   }
 
   /**
@@ -77,7 +77,7 @@ class GaussianKernel
   double Evaluate(const double t) const
   {
     // The precalculation of gamma saves us a little computation time.
-    return exp(gamma * std::pow(t, 2.0));
+    return std::exp(gamma * std::pow(t, 2.0));
   }
 
   /**
@@ -89,19 +89,7 @@ class GaussianKernel
    *     constructor.
    */
   double Gradient(const double t) const {
-    return 2 * t * gamma * exp(gamma * std::pow(t, 2.0));
-  }
-
-  /**
-   * Evaluation of the gradient of Gaussian kernel
-   * given the squared distance between two points.
-   *
-   * @param t The squared distance between the two points
-   * @return K(t) using the bandwidth (@f$\mu@f$) specified in the
-   *     constructor.
-   */
-  double GradientForSquaredDistance(const double t) const {
-    return gamma * exp(gamma * t);
+    return 2 * t * gamma * std::exp(gamma * std::pow(t, 2.0));
   }
 
   /**
@@ -112,24 +100,9 @@ class GaussianKernel
    */
   double Normalizer(const size_t dimension)
   {
-    return pow(sqrt(2.0 * M_PI) * bandwidth, (double) dimension);
-  }
-
-  /**
-   * Obtain a convolution integral of the Gaussian kernel.
-   *
-   * @param a First vector.
-   * @param b Second vector.
-   * @return The convolution integral.
-   */
-  template<typename VecTypeA, typename VecTypeB>
-  double ConvolutionIntegral(const VecTypeA& a, const VecTypeB& b)
-  {
-    return Evaluate(sqrt(SquaredEuclideanDistance::Evaluate(a, b) /
-        2.0)) / (Normalizer(a.n_rows) * pow(2.0, (double) a.n_rows / 2.0));
+    return std::pow(std::sqrt(2.0 * M_PI) * bandwidth, (double) dimension);
   }
 
-
   //! Get the bandwidth.
   double Bandwidth() const { return bandwidth; }
 
@@ -138,7 +111,7 @@ class GaussianKernel
   void Bandwidth(const double bandwidth)
   {
     this->bandwidth = bandwidth;
-    this->gamma = -0.5 * pow(bandwidth, -2.0);
+    this->gamma = -0.5 * std::pow(bandwidth, -2.0);
   }
 
   //! Get the precalculated constant.
diff -pruN 4.3.0-2/src/mlpack/core/kernels/hyperbolic_tangent_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/hyperbolic_tangent_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/hyperbolic_tangent_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/hyperbolic_tangent_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -58,18 +58,18 @@ class HyperbolicTangentKernel
   template<typename VecTypeA, typename VecTypeB>
   double Evaluate(const VecTypeA& a, const VecTypeB& b)
   {
-    return tanh(scale * arma::dot(a, b) + offset);
+    return tanh(scale * dot(a, b) + offset);
   }
 
   //! Get scale factor.
   double Scale() const { return scale; }
   //! Modify scale factor.
-  double& Scale() { return scale; }
+  void Scale(const double scale) { this->scale = scale; }
 
   //! Get offset for the kernel.
   double Offset() const { return offset; }
   //! Modify offset for the kernel.
-  double& Offset() { return offset; }
+  void Offset(const double offset) { this->offset = offset; }
 
   //! Serialize the kernel.
   template<typename Archive>
diff -pruN 4.3.0-2/src/mlpack/core/kernels/kernels.hpp 4.4.0-1/src/mlpack/core/kernels/kernels.hpp
--- 4.3.0-2/src/mlpack/core/kernels/kernels.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/kernels.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -15,7 +15,7 @@
 #include "kernel_traits.hpp"
 
 #include "cauchy_kernel.hpp"
-#include "cosine_distance.hpp"
+#include "cosine_similarity.hpp"
 #include "epanechnikov_kernel.hpp"
 #include "example_kernel.hpp"
 #include "gaussian_kernel.hpp"
diff -pruN 4.3.0-2/src/mlpack/core/kernels/laplacian_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/laplacian_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/laplacian_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/laplacian_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,19 +30,11 @@ class LaplacianKernel
 {
  public:
   /**
-   * Default constructor; sets bandwidth to 1.0.
-   */
-  LaplacianKernel() : bandwidth(1.0)
-  { }
-
-  /**
-   * Construct the Laplacian kernel with a custom bandwidth.
+   * Construct the Laplacian kernel with the given bandwidth.
    *
    * @param bandwidth The bandwidth of the kernel (@f$\mu@f$).
    */
-  LaplacianKernel(double bandwidth) :
-      bandwidth(bandwidth)
-  { }
+  LaplacianKernel(const double bandwidth = 1.0) : bandwidth(bandwidth) { }
 
   /**
    * Evaluation of the Laplacian kernel.  This could be generalized to use any
@@ -60,7 +52,7 @@ class LaplacianKernel
   double Evaluate(const VecTypeA& a, const VecTypeB& b) const
   {
     // The precalculation of gamma saves us a little computation time.
-    return exp(-EuclideanDistance::Evaluate(a, b) / bandwidth);
+    return std::exp(-EuclideanDistance::Evaluate(a, b) / bandwidth);
   }
 
   /**
@@ -74,7 +66,7 @@ class LaplacianKernel
   double Evaluate(const double t) const
   {
     // The precalculation of gamma saves us a little computation time.
-    return exp(-t / bandwidth);
+    return std::exp(-t / bandwidth);
   }
 
   /**
@@ -87,13 +79,13 @@ class LaplacianKernel
    *     constructor.
    */
   double Gradient(const double t) const  {
-    return exp(-t / bandwidth) / -bandwidth;
+    return std::exp(-t / bandwidth) / -bandwidth;
   }
 
   //! Get the bandwidth.
   double Bandwidth() const { return bandwidth; }
   //! Modify the bandwidth.
-  double& Bandwidth() { return bandwidth; }
+  void Bandwidth(const double bandwidth) { this->bandwidth = bandwidth; }
 
   //! Serialize the kernel.
   template<typename Archive>
diff -pruN 4.3.0-2/src/mlpack/core/kernels/linear_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/linear_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/linear_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/linear_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,7 +51,7 @@ class LinearKernel
   template<typename VecTypeA, typename VecTypeB>
   static double Evaluate(const VecTypeA& a, const VecTypeB& b)
   {
-    return arma::dot(a, b);
+    return dot(a, b);
   }
 
   //! Serialize the kernel (it has no members... do nothing).
diff -pruN 4.3.0-2/src/mlpack/core/kernels/polynomial_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/polynomial_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/polynomial_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/polynomial_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,18 +53,18 @@ class PolynomialKernel
   template<typename VecTypeA, typename VecTypeB>
   double Evaluate(const VecTypeA& a, const VecTypeB& b) const
   {
-    return pow((arma::dot(a, b) + offset), degree);
+    return std::pow((dot(a, b) + offset), degree);
   }
 
   //! Get the degree of the polynomial.
   const double& Degree() const { return degree; }
   //! Modify the degree of the polynomial.
-  double& Degree() { return degree; }
+  void Degree(const double degree) { this->degree = degree; }
 
   //! Get the offset of the dot product of the arguments.
   const double& Offset() const { return offset; }
   //! Modify the offset of the dot product of the arguments.
-  double& Offset() { return offset; }
+  void Offset(const double offset) { this->offset = offset; }
 
   //! Serialize the kernel.
   template<typename Archive>
diff -pruN 4.3.0-2/src/mlpack/core/kernels/pspectrum_string_kernel_impl.hpp 4.4.0-1/src/mlpack/core/kernels/pspectrum_string_kernel_impl.hpp
--- 4.3.0-2/src/mlpack/core/kernels/pspectrum_string_kernel_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/pspectrum_string_kernel_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -25,6 +25,12 @@ inline PSpectrumStringKernel::PSpectrumS
     const std::vector<std::vector<std::string> >& datasets,
     const size_t p) : p(p)
 {
+  if (p == 0)
+  {
+    throw std::invalid_argument(
+        "PSpectrumStringKernel::PSpectrumStringKernel(): p must be positive");
+  }
+
   // We have to assemble the counts of substrings.  This is not a particularly
   // fast operation, unfortunately, but it only needs to be done once.
   Log::Info << "Assembling counts of substrings of length " << p << "."
diff -pruN 4.3.0-2/src/mlpack/core/kernels/spherical_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/spherical_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/spherical_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/spherical_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,6 +29,15 @@ class SphericalKernel
     bandwidthSquared(std::pow(bandwidth, 2.0))
   { /* Nothing to do. */ }
 
+  // Get the bandwidth.
+  const double Bandwidth() const { return bandwidth; }
+  // Modify the bandwidth.
+  void Bandwidth(const double bandwidth)
+  {
+    this->bandwidth = bandwidth;
+    this->bandwidthSquared = bandwidth * bandwidth;
+  }
+
   /**
    * Evaluate the spherical kernel with the given two vectors.
    *
@@ -44,45 +53,11 @@ class SphericalKernel
     return (SquaredEuclideanDistance::Evaluate(a, b) <= bandwidthSquared) ?
         1.0 : 0.0;
   }
-  /**
-   * Obtains the convolution integral [integral K(||x-a||)K(||b-x||)dx]
-   * for the two vectors.
-   *
-   * @tparam VecTypeA Type of first vector (arma::vec, arma::sp_vec should be
-   *       expected).
-   * @tparam VecTypeB Type of second vector.
-   * @param a First vector.
-   * @param b Second vector.
-   * @return The convolution integral value.
-   */
-  template<typename VecTypeA, typename VecTypeB>
-  double ConvolutionIntegral(const VecTypeA& a, const VecTypeB& b) const
-  {
-    double distance = sqrt(SquaredEuclideanDistance::Evaluate(a, b));
-    if (distance >= 2.0 * bandwidth)
-    {
-      return 0.0;
-    }
-    double volumeSquared = pow(Normalizer(a.n_rows), 2.0);
 
-    switch (a.n_rows)
-    {
-      case 1:
-        return 1.0 / volumeSquared * (2.0 * bandwidth - distance);
-      case 2:
-        return 1.0 / volumeSquared *
-          (2.0 * bandwidth * bandwidth * acos(distance/(2.0 * bandwidth)) -
-          distance / 4.0 * sqrt(4.0*bandwidth*bandwidth-distance*distance));
-      default:
-        Log::Fatal << "The spherical kernel does not support convolution\
-          integrals above dimension two, yet..." << std::endl;
-        return -1.0;
-    }
-  }
   double Normalizer(size_t dimension) const
   {
-    return pow(bandwidth, (double) dimension) * pow(M_PI, dimension / 2.0) /
-        std::tgamma(dimension / 2.0 + 1.0);
+    return std::pow(bandwidth, (double) dimension) *
+        std::pow(M_PI, dimension / 2.0) / std::tgamma(dimension / 2.0 + 1.0);
   }
 
   /**
@@ -94,6 +69,7 @@ class SphericalKernel
   {
     return (t <= bandwidth) ? 1.0 : 0.0;
   }
+
   double Gradient(double t)
   {
     return t == bandwidth ? arma::datum::nan : 0.0;
diff -pruN 4.3.0-2/src/mlpack/core/kernels/triangular_kernel.hpp 4.4.0-1/src/mlpack/core/kernels/triangular_kernel.hpp
--- 4.3.0-2/src/mlpack/core/kernels/triangular_kernel.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/kernels/triangular_kernel.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -87,7 +87,7 @@ class TriangularKernel
   //! Get the bandwidth of the kernel.
   double Bandwidth() const { return bandwidth; }
   //! Modify the bandwidth of the kernel.
-  double& Bandwidth() { return bandwidth; }
+  void Bandwidth(const double bandwidth) { this->bandwidth = bandwidth; }
 
   //! Serialize the kernel.
   template<typename Archive>
diff -pruN 4.3.0-2/src/mlpack/core/math/ccov.hpp 4.4.0-1/src/mlpack/core/math/ccov.hpp
--- 4.3.0-2/src/mlpack/core/math/ccov.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/ccov.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -3,8 +3,8 @@
  * @author Ryan Curtin
  * @author Conrad Sanderson
  *
- * ColumnCovariance(X) is same as cov(trans(X)) but without the cost
- * of computing trans(X)
+ * ColumnCovariance(X) is the same as cov(trans(X)) but without the cost of
+ * computing trans(X).
  *
  * mlpack is free software; you may redistribute it and/or modify it under the
  * terms of the 3-clause BSD license.  You should have received a copy of the
@@ -21,13 +21,13 @@ namespace mlpack {
 template<typename eT>
 inline
 arma::Mat<eT>
-ColumnCovariance(const arma::Mat<eT>& A, const size_t norm_type = 0);
+ColumnCovariance(const arma::Mat<eT>& A, const size_t normType = 0);
 
 template<typename T>
 inline
-arma::Mat< std::complex<T> >
-ColumnCovariance(const arma::Mat< std::complex<T> >& A,
-     const size_t norm_type = 0);
+arma::Mat<std::complex<T>>
+ColumnCovariance(const arma::Mat<std::complex<T>>& A,
+                 const size_t normType = 0);
 
 } // namespace mlpack
 
diff -pruN 4.3.0-2/src/mlpack/core/math/ccov_impl.hpp 4.4.0-1/src/mlpack/core/math/ccov_impl.hpp
--- 4.3.0-2/src/mlpack/core/math/ccov_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/ccov_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,7 +24,7 @@ inline arma::Mat<eT> ColumnCovariance(co
 {
   if (normType > 1)
   {
-    Log::Fatal << "ColumnCovariance(): norm_type must be 0 or 1!" << std::endl;
+    Log::Fatal << "ColumnCovariance(): normType must be 0 or 1!" << std::endl;
   }
 
   arma::Mat<eT> out;
@@ -56,7 +56,7 @@ inline arma::Mat<std::complex<T>> Column
 {
   if (normType > 1)
   {
-    Log::Fatal << "ColumnCovariance(): norm_type must be 0 or 1" << std::endl;
+    Log::Fatal << "ColumnCovariance(): normType must be 0 or 1" << std::endl;
   }
 
   typedef typename std::complex<T> eT;
@@ -67,7 +67,7 @@ inline arma::Mat<std::complex<T>> Column
   {
     if (x.n_rows == 1)
     {
-      const arma::Mat<T> tmpMat = arma::var(arma::trans(x), normType);
+      const arma::Mat<T> tmpMat = arma::var(trans(x), normType);
       out.set_size(1, 1);
       out[0] = tmpMat[0];
     }
@@ -84,10 +84,10 @@ inline arma::Mat<std::complex<T>> Column
     const eT normVal = (normType == 0) ?
         ((n > 1) ? eT(n - 1) : eT(1)) : eT(n);
 
-    const arma::Col<eT> acc = arma::sum(x, 1);
+    const arma::Col<eT> acc = sum(x, 1);
 
-    out = x * arma::trans(arma::conj(x));
-    out -= (acc * arma::trans(arma::conj(acc))) / eT(n);
+    out = x * trans(arma::conj(x));
+    out -= (acc * trans(arma::conj(acc))) / eT(n);
     out /= normVal;
   }
 
diff -pruN 4.3.0-2/src/mlpack/core/math/clamp.hpp 4.4.0-1/src/mlpack/core/math/clamp.hpp
--- 4.3.0-2/src/mlpack/core/math/clamp.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/clamp.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,65 +0,0 @@
-/**
- * @file core/math/clamp.hpp
- *
- * Miscellaneous math clamping routines.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_MATH_CLAMP_HPP
-#define MLPACK_CORE_MATH_CLAMP_HPP
-
-#include <stdlib.h>
-#include <math.h>
-#include <float.h>
-
-namespace mlpack {
-
-/**
- * Forces a number to be non-negative, turning negative numbers into zero.
- * Avoids branching costs (this is a measurable improvement).
- *
- * @param d Double to clamp.
- * @return 0 if d < 0, d otherwise.
- */
-inline double ClampNonNegative(const double d)
-{
-  return (d + fabs(d)) / 2;
-}
-
-/**
- * Forces a number to be non-positive, turning positive numbers into zero.
- * Avoids branching costs (this is a measurable improvement).
- *
- * @param d Double to clamp.
- * @return 0 if d > 0, d otherwise.
- */
-inline double ClampNonPositive(const double d)
-{
-  return (d - fabs(d)) / 2;
-}
-
-/**
- * Clamp a number between a particular range.
- *
- * @param value The number to clamp.
- * @param rangeMin The first of the range.
- * @param rangeMax The last of the range.
- * @return max(rangeMin, min(rangeMax, d)).
- */
-inline double ClampRange(double value,
-                         const double rangeMin,
-                         const double rangeMax)
-{
-  value -= rangeMax;
-  value = ClampNonPositive(value) + rangeMax;
-  value -= rangeMin;
-  value = ClampNonNegative(value) + rangeMin;
-  return value;
-}
-
-} // namespace mlpack
-
-#endif // MLPACK_CORE_MATH_CLAMP_HPP
diff -pruN 4.3.0-2/src/mlpack/core/math/columns_to_blocks.hpp 4.4.0-1/src/mlpack/core/math/columns_to_blocks.hpp
--- 4.3.0-2/src/mlpack/core/math/columns_to_blocks.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/columns_to_blocks.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -10,8 +10,8 @@
  * 3-clause BSD license along with mlpack.  If not, see
  * http://www.opensource.org/licenses/BSD-3-Clause for more information.
  */
-#ifndef MLPACK_METHODS_NN_COLUMNS_TO_BLOCKS_HPP
-#define MLPACK_METHODS_NN_COLUMNS_TO_BLOCKS_HPP
+#ifndef MLPACK_CORE_MATH_COLUMNS_TO_BLOCKS_HPP
+#define MLPACK_CORE_MATH_COLUMNS_TO_BLOCKS_HPP
 
 #include <mlpack/prereqs.hpp>
 
diff -pruN 4.3.0-2/src/mlpack/core/math/columns_to_blocks_impl.hpp 4.4.0-1/src/mlpack/core/math/columns_to_blocks_impl.hpp
--- 4.3.0-2/src/mlpack/core/math/columns_to_blocks_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/columns_to_blocks_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -9,8 +9,8 @@
  * 3-clause BSD license along with mlpack.  If not, see
  * http://www.opensource.org/licenses/BSD-3-Clause for more information.
  */
-#ifndef MLPACK_METHODS_NN_COLUMNS_TO_BLOCKS_IMPL_HPP
-#define MLPACK_METHODS_NN_COLUMNS_TO_BLOCKS_IMPL_HPP
+#ifndef MLPACK_CORE_MATH_COLUMNS_TO_BLOCKS_IMPL_HPP
+#define MLPACK_CORE_MATH_COLUMNS_TO_BLOCKS_IMPL_HPP
 
 #include "columns_to_blocks.hpp"
 
@@ -41,14 +41,17 @@ inline bool ColumnsToBlocks::IsPerfectSq
 inline void ColumnsToBlocks::Transform(const arma::mat& maximalInputs,
                                        arma::mat& output)
 {
-  //! TODO: Maybe replace std::runtime_error with Log::Fatal.
-  if (!IsPerfectSquare(maximalInputs.n_rows))
-  {
-    throw std::runtime_error("maximalInputs.n_rows should be perfect square");
-  }
-
   if (blockHeight == 0 || blockWidth == 0)
   {
+    //! TODO: Maybe replace std::runtime_error with Log::Fatal.
+    if (!IsPerfectSquare(maximalInputs.n_rows))
+    {
+      std::ostringstream oss;
+      oss << "ColumnsToBlocks::Transform(): input.n_rows ("
+          << maximalInputs.n_rows << ") should be a perfect square!";
+      throw std::runtime_error(oss.str());
+    }
+
     size_t const squareRows =
         static_cast<size_t>(std::sqrt(maximalInputs.n_rows));
     blockHeight = squareRows;
@@ -57,15 +60,17 @@ inline void ColumnsToBlocks::Transform(c
 
   if (blockHeight * blockWidth != maximalInputs.n_rows)
   {
-    throw std::runtime_error("blockHeight * blockWidth should "
-                             "equal to maximalInputs.n_rows");
+    std::ostringstream oss;
+    oss << "ColumnsToBlocks::Transform(): blockHeight * blockWidth ("
+        << blockHeight << " * " << blockWidth << ") should be equal to "
+        << "input.n_rows (" << maximalInputs.n_rows << ")!";
+    throw std::runtime_error(oss.str());
   }
 
-  const size_t rowOffset = blockHeight+bufSize;
-  const size_t colOffset = blockWidth+bufSize;
-  output.ones(bufSize + rows * rowOffset,
-              bufSize + cols * colOffset);
-  output *= bufValue;
+  const size_t rowOffset = blockHeight + bufSize;
+  const size_t colOffset = blockWidth + bufSize;
+  output.set_size(bufSize + rows * rowOffset, bufSize + cols * colOffset);
+  output.fill(bufValue);
 
   size_t k = 0;
   const size_t maxSize = std::min(rows * cols, (size_t) maximalInputs.n_cols);
@@ -76,8 +81,8 @@ inline void ColumnsToBlocks::Transform(c
       // Now, copy the elements of the row to the output submatrix.
       const size_t minRow = bufSize + i * rowOffset;
       const size_t minCol = bufSize + j * colOffset;
-      const size_t maxRow = i * rowOffset + blockHeight;
-      const size_t maxCol = j * colOffset + blockWidth;
+      const size_t maxRow = bufSize + i * rowOffset + blockHeight - 1;
+      const size_t maxCol = bufSize + j * colOffset + blockWidth - 1;
 
       output.submat(minRow, minCol, maxRow, maxCol) =
           arma::reshape(maximalInputs.col(k++), blockHeight, blockWidth);
diff -pruN 4.3.0-2/src/mlpack/core/math/digamma.hpp 4.4.0-1/src/mlpack/core/math/digamma.hpp
--- 4.3.0-2/src/mlpack/core/math/digamma.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/digamma.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -1,11 +1,11 @@
 /**
  * @file core/math/digamma.hpp
- * @author Gopi Tatiraju 
+ * @author Gopi Tatiraju
  *
  * Some parts of the implementation are inspired from boost.
  * Boost Software License, Version 1.0. (See accompanying file
  * LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
- * 
+ *
  * mlpack is free software; you may redistribute it and/or modify it under the
  * terms of the 3-clause BSD license.  You should have received a copy of the
  * 3-clause BSD license along with mlpack.  If not, see
@@ -32,20 +32,20 @@ EvaluatePolyLarge(const T(&a)[N], const
 {
   T x2 = x * x;
   T t[2];
-  
+
   t[0] = a[7] * x2 + a[5];
   t[0] *= x2;
   t[0] += a[3];
   t[0] *= x2;
   t[0] += a[1];
   t[0] *= x;
-  
+
   t[1] = a[6] * x2 + a[4];
   t[1] *= x2;
   t[1] += a[2];
   t[1] *= x2;
   t[1] += a[0];
-  
+
   return t[0] + t[1];
 }
 
@@ -178,7 +178,7 @@ T DigammaLarge(T x)
   T result = std::log(x);
   result += 1 / (2 * x);
   const T z = 1 / (x * x);
-  result -= z * EvaluatePolyLarge(P, z); 
+  result -= z * EvaluatePolyLarge(P, z);
 
   return result;
 }
@@ -192,7 +192,7 @@ T DigammaLarge(T x)
  * For x > 2 we reduce it to the interval [1, 2].
  * For x < 1 we bring it to the interval [1, 2].
  *
- * @param x Input for which digamma will be calculated. 
+ * @param x Input for which digamma will be calculated.
  */
 template<typename T>
 T Digamma(T x)
diff -pruN 4.3.0-2/src/mlpack/core/math/lin_alg.hpp 4.4.0-1/src/mlpack/core/math/lin_alg.hpp
--- 4.3.0-2/src/mlpack/core/math/lin_alg.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/lin_alg.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,139 +0,0 @@
-/**
- * @file core/math/lin_alg.hpp
- * @author Nishant Mehta
- *
- * Linear algebra utilities.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_MATH_LIN_ALG_HPP
-#define MLPACK_CORE_MATH_LIN_ALG_HPP
-
-#include <mlpack/prereqs.hpp>
-#include "ccov.hpp"
-#include "random.hpp"
-
-/**
- * Linear algebra utility functions, generally performed on matrices or vectors.
- */
-namespace mlpack {
-
-/**
- * Auxiliary function to raise vector elements to a specific power.  The sign
- * is ignored in the power operation and then re-added.  Useful for
- * eigenvalues.
- */
-inline void VectorPower(arma::vec& vec, const double power);
-
-/**
- * Creates a centered matrix, where centering is done by subtracting
- * the sum over the columns (a column vector) from each column of the matrix.
- *
- * @param x Input matrix
- * @param xCentered Matrix to write centered output into
- */
-inline void Center(const arma::mat& x, arma::mat& xCentered);
-
-/**
- * Whitens a matrix using the singular value decomposition of the covariance
- * matrix. Whitening means the covariance matrix of the result is the identity
- * matrix.
- */
-inline void WhitenUsingSVD(const arma::mat& x,
-                           arma::mat& xWhitened,
-                           arma::mat& whiteningMatrix);
-
-
-/**
- * Overwrites a dimension-N vector to a random vector on the unit sphere in R^N.
- */
-inline void RandVector(arma::vec& v);
-
-/**
- * Orthogonalize x and return the result in W, using eigendecomposition.
- * We will be using the formula \f$ W = x (x^T x)^{-0.5} \f$.
- */
-inline void Orthogonalize(const arma::mat& x, arma::mat& W);
-
-/**
- * Orthogonalize x in-place.  This could be sped up by a custom
- * implementation.
- */
-inline void Orthogonalize(arma::mat& x);
-
-/**
- * Remove a certain set of rows in a matrix while copying to a second matrix.
- *
- * @param input Input matrix to copy.
- * @param rowsToRemove Vector containing indices of rows to be removed.
- * @param output Matrix to copy non-removed rows into.
- */
-inline void RemoveRows(const arma::mat& input,
-                       const std::vector<size_t>& rowsToRemove,
-                       arma::mat& output);
-
-/**
- * Upper triangular representation of a symmetric matrix, scaled such that,
- * dot(Svec(A), Svec(B)) == dot(A, B) for symmetric A, B. Specifically,
- *
- * Svec(K) = [ K_11, sqrt(2) K_12, ..., sqrt(2) K_1n, K_22, ..., sqrt(2) K_2n, ..., K_nn ]^T
- *
- * @param input A symmetric matrix
- * @param output
- */
-inline void Svec(const arma::mat& input, arma::vec& output);
-
-inline void Svec(const arma::sp_mat& input, arma::sp_vec& output);
-
-/**
- * The inverse of Svec. That is, Smat(Svec(A)) == A.
- *
- * @param input
- * @param output A symmetric matrix
- */
-inline void Smat(const arma::vec& input, arma::mat& output);
-
-/**
- * Return the index such that A[i,j] == factr(i, j) * svec(A)[pos(i, j)],
- * where factr(i, j) = sqrt(2) if i != j and 1 otherwise.
- *
- * @param i
- * @param j
- * @param n
- */
-inline size_t SvecIndex(size_t i, size_t j, size_t n);
-
-/**
- * If A is a symmetric matrix, then SymKronId returns an operator Op such that
- *
- *    Op * svec(X) == svec(0.5 * (AX + XA))
- *
- * for every symmetric matrix X
- *
- * @param A
- * @param op
- */
-inline void SymKronId(const arma::mat& A, arma::mat& op);
-
-/**
- * Signum function.
- * Return 1 if x>0; return 0 if x=0; return -1 if x<0.
- * Return type are the same as input type.
- *
- * @param x Number of any type.
- */
-template <typename T>
-T Sign(const T x)
-{
-  return (T(0) < x) - (x < T(0));
-}
-
-} // namespace mlpack
-
-// Partially include implementation
-#include "lin_alg_impl.hpp"
-
-#endif // MLPACK_CORE_MATH_LIN_ALG_HPP
diff -pruN 4.3.0-2/src/mlpack/core/math/lin_alg_impl.hpp 4.4.0-1/src/mlpack/core/math/lin_alg_impl.hpp
--- 4.3.0-2/src/mlpack/core/math/lin_alg_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/lin_alg_impl.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,289 +0,0 @@
-/**
- * @file core/math/lin_alg_impl.hpp
- * @author Stephen Tu
- * @author Nishant Mehta
- *
- * Linear algebra utilities.
- * 
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_MATH_LIN_ALG_IMPL_HPP
-#define MLPACK_CORE_MATH_LIN_ALG_IMPL_HPP
-
-#include "lin_alg.hpp"
-
-namespace mlpack {
-
-inline size_t SvecIndex(size_t i, size_t j, size_t n)
-{
-  if (i > j)
-    std::swap(i, j);
-  return (j-i) + (n*(n+1) - (n-i)*(n-i+1))/2;
-}
-
-/**
- * Auxiliary function to raise vector elements to a specific power.  The sign
- * is ignored in the power operation and then re-added.  Useful for
- * eigenvalues.
- */
-inline void VectorPower(arma::vec& vec, const double power)
-{
-  for (size_t i = 0; i < vec.n_elem; ++i)
-  {
-    if (std::abs(vec(i)) > 1e-12)
-    {
-      vec(i) = (vec(i) > 0) ? std::pow(vec(i), (double) power) :
-          -std::pow(-vec(i), (double) power);
-     }
-    else
-      vec(i) = 0;
-  }
-}
-
-/**
- * Creates a centered matrix, where centering is done by subtracting
- * the sum over the columns (a column vector) from each column of the matrix.
- *
- * @param x Input matrix
- * @param xCentered Matrix to write centered output into
- */
-inline void Center(const arma::mat& x, arma::mat& xCentered)
-{
-  // Get the mean of the elements in each row.
-  arma::vec rowMean = arma::sum(x, 1) / x.n_cols;
-
-  xCentered = x - arma::repmat(rowMean, 1, x.n_cols);
-}
-
-/**
- * Whitens a matrix using the singular value decomposition of the covariance
- * matrix. Whitening means the covariance matrix of the result is the identity
- * matrix.
- */
-inline void WhitenUsingSVD(const arma::mat& x,
-                           arma::mat& xWhitened,
-                           arma::mat& whiteningMatrix)
-{
-  arma::mat covX, u, v, invSMatrix, temp1;
-  arma::vec sVector;
-
-  covX = ColumnCovariance(x);
-
-  svd(u, sVector, v, covX);
-
-  size_t d = sVector.n_elem;
-  invSMatrix.zeros(d, d);
-  invSMatrix.diag() = 1 / sqrt(sVector);
-
-  whiteningMatrix = v * invSMatrix * trans(u);
-
-  xWhitened = whiteningMatrix * x;
-}
-
-/**
- * Overwrites a dimension-N vector to a random vector on the unit sphere in R^N.
- */
-inline void RandVector(arma::vec& v)
-{
-  for (size_t i = 0; i + 1 < v.n_elem; i += 2)
-  {
-    double a = Random();
-    double b = Random();
-    double first_term = sqrt(-2 * log(a));
-    double second_term = 2 * M_PI * b;
-    v[i]     = first_term * cos(second_term);
-    v[i + 1] = first_term * sin(second_term);
-  }
-
-  if ((v.n_elem % 2) == 1)
-  {
-    v[v.n_elem - 1] = sqrt(-2 * log(Random())) * cos(2 * M_PI * Random());
-  }
-
-  v /= sqrt(dot(v, v));
-}
-
-/**
- * Orthogonalize x and return the result in W, using eigendecomposition.
- * We will be using the formula \f$ W = x (x^T x)^{-0.5} \f$.
- */
-inline void Orthogonalize(const arma::mat& x, arma::mat& W)
-{
-  // For a matrix A, A^N = V * D^N * V', where VDV' is the
-  // eigendecomposition of the matrix A.
-  arma::mat eigenvalues, eigenvectors;
-  arma::vec egval;
-  eig_sym(egval, eigenvectors, ColumnCovariance(x));
-  VectorPower(egval, -0.5);
-
-  eigenvalues.zeros(egval.n_elem, egval.n_elem);
-  eigenvalues.diag() = egval;
-
-  arma::mat at = (eigenvectors * eigenvalues * trans(eigenvectors));
-
-  W = at * x;
-}
-
-/**
- * Orthogonalize x in-place.  This could be sped up by a custom
- * implementation.
- */
-inline void Orthogonalize(arma::mat& x)
-{
-  Orthogonalize(x, x);
-}
-
-/**
- * Remove a certain set of rows in a matrix while copying to a second matrix.
- *
- * @param input Input matrix to copy.
- * @param rowsToRemove Vector containing indices of rows to be removed.
- * @param output Matrix to copy non-removed rows into.
- */
-inline void RemoveRows(const arma::mat& input,
-                       const std::vector<size_t>& rowsToRemove,
-                       arma::mat& output)
-{
-  const size_t nRemove = rowsToRemove.size();
-  const size_t nKeep = input.n_rows - nRemove;
-
-  if (nRemove == 0)
-  {
-    output = input; // Copy everything.
-  }
-  else
-  {
-    output.set_size(nKeep, input.n_cols);
-
-    size_t curRow = 0;
-    size_t removeInd = 0;
-    // First, check 0 to first row to remove.
-    if (rowsToRemove[0] > 0)
-    {
-      // Note that this implies that n_rows > 1.
-      output.rows(0, rowsToRemove[0] - 1) = input.rows(0, rowsToRemove[0] - 1);
-      curRow += rowsToRemove[0];
-    }
-
-    // Now, check i'th row to remove to (i + 1)'th row to remove, until i is the
-    // penultimate row.
-    while (removeInd < nRemove - 1)
-    {
-      const size_t height = rowsToRemove[removeInd + 1] -
-          rowsToRemove[removeInd] - 1;
-
-      if (height > 0)
-      {
-        output.rows(curRow, curRow + height - 1) =
-            input.rows(rowsToRemove[removeInd] + 1,
-                       rowsToRemove[removeInd + 1] - 1);
-        curRow += height;
-      }
-
-      removeInd++;
-    }
-
-    // Now that i is the last row to remove, check last row to remove to last
-    // row.
-    if (rowsToRemove[removeInd] < input.n_rows - 1)
-    {
-      output.rows(curRow, nKeep - 1) = input.rows(rowsToRemove[removeInd] + 1,
-          input.n_rows - 1);
-    }
-  }
-}
-
-inline void Svec(const arma::mat& input, arma::vec& output)
-{
-  const size_t n = input.n_rows;
-  const size_t n2bar = n * (n + 1) / 2;
-
-  output.zeros(n2bar);
-
-  size_t idx = 0;
-  for (size_t i = 0; i < n; ++i)
-  {
-    for (size_t j = i; j < n; ++j)
-    {
-      if (i == j)
-        output(idx++) = input(i, j);
-      else
-        output(idx++) = M_SQRT2 * input(i, j);
-    }
-  }
-}
-
-inline void Svec(const arma::sp_mat& input, arma::sp_vec& output)
-{
-  const size_t n = input.n_rows;
-  const size_t n2bar = n * (n + 1) / 2;
-
-  output.zeros(n2bar, 1);
-
-  for (auto it = input.begin(); it != input.end(); ++it)
-  {
-    const size_t i = it.row();
-    const size_t j = it.col();
-    if (i > j)
-      continue;
-    if (i == j)
-      output(SvecIndex(i, j, n)) = *it;
-    else
-      output(SvecIndex(i, j, n)) = M_SQRT2 * (*it);
-  }
-}
-
-inline void Smat(const arma::vec& input, arma::mat& output)
-{
-  const size_t n = static_cast<size_t>
-      (ceil((-1. + sqrt(1. + 8. * input.n_elem)) / 2.));
-
-  output.zeros(n, n);
-
-  size_t idx = 0;
-  for (size_t i = 0; i < n; ++i)
-  {
-    for (size_t j = i; j < n; ++j)
-    {
-      if (i == j)
-        output(i, j) = input(idx++);
-      else
-        output(i, j) = output(j, i) = M_SQRT1_2 * input(idx++);
-    }
-  }
-}
-
-inline void SymKronId(const arma::mat& A, arma::mat& op)
-{
-  // TODO(stephentu): there's probably an easier way to build this operator
-
-  const size_t n = A.n_rows;
-  const size_t n2bar = n * (n + 1) / 2;
-  op.zeros(n2bar, n2bar);
-
-  size_t idx = 0;
-  for (size_t i = 0; i < n; ++i)
-  {
-    for (size_t j = i; j < n; ++j)
-    {
-      for (size_t k = 0; k < n; ++k)
-      {
-        op(idx, SvecIndex(k, j, n)) +=
-          ((k == j) ? 1. : M_SQRT1_2) * A(i, k);
-        op(idx, SvecIndex(i, k, n)) +=
-          ((k == i) ? 1. : M_SQRT1_2) * A(k, j);
-      }
-      op.row(idx) *= 0.5;
-      if (i != j)
-        op.row(idx) *= M_SQRT2;
-      idx++;
-    }
-  }
-}
-
-} // namespace mlpack
-
-#endif
diff -pruN 4.3.0-2/src/mlpack/core/math/log_add_impl.hpp 4.4.0-1/src/mlpack/core/math/log_add_impl.hpp
--- 4.3.0-2/src/mlpack/core/math/log_add_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/log_add_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -20,7 +20,7 @@ namespace mlpack {
  *
  * @f[
  * e^z = e^x + e^y
- * e^z = e^x(1 + e^{y-x})     = e^y(1 + e^{x-y})
+ * e^z = e^x(1 + e^{y-x})      = e^y(1 + e^{x-y})
  * z   = x + \log(1 + e^{y-x}) = y + \log(1 + e^{x-y})
  * @f]
  *
@@ -49,7 +49,7 @@ T LogAdd(T x, T y)
   if (std::isinf(d) || std::isinf(r))
     return r;
 
-  return r + log(1 + exp(d));
+  return r + std::log(1 + std::exp(d));
 }
 
 /**
@@ -65,7 +65,7 @@ typename T::elem_type AccuLog(const T& x
   if (maxVal == -std::numeric_limits<typename T::elem_type>::infinity())
     return maxVal;
 
-  return maxVal + log(sum(exp(x - maxVal)));;
+  return maxVal + std::log(sum(exp(x - maxVal)));;
 }
 
 /**
@@ -83,8 +83,7 @@ void LogSumExp(const T& x, arma::Col<typ
     // Compute the maximum in each column (treating y as a column too).
     maxs = max(max(x, 1), y);
 
-    y = maxs + log(sum(exp(x - repmat(maxs, 1, x.n_cols)), 1) +
-        exp(y - maxs));
+    y = maxs + log(sum(exp(x - repmat(maxs, 1, x.n_cols)), 1) + exp(y - maxs));
   }
   else
   {
diff -pruN 4.3.0-2/src/mlpack/core/math/make_alias.hpp 4.4.0-1/src/mlpack/core/math/make_alias.hpp
--- 4.3.0-2/src/mlpack/core/math/make_alias.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/make_alias.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,89 +16,84 @@
 namespace mlpack {
 
 /**
- * Make an alias of a dense cube.  If strict is true, then the alias cannot be
- * resized or pointed at new memory.
+ * Reconstruct `v` as an alias around the memory `newMem`, with size `numRows` x
+ * `numCols`.
  */
-template<typename ElemType>
-arma::Cube<ElemType> MakeAlias(arma::Cube<ElemType>& input,
-                               const bool strict = true)
-{
-  // Use the advanced constructor.
-  return arma::Cube<ElemType>(input.memptr(), input.n_rows, input.n_cols,
-      input.n_slices, false, strict);
-}
-
-/**
- * Make an alias of a dense matrix.  If strict is true, then the alias cannot be
- * resized or pointed at new memory.
- */
-template<typename ElemType>
-arma::Mat<ElemType> MakeAlias(arma::Mat<ElemType>& input,
-                              const bool strict = true)
-{
-  // Use the advanced constructor.
-  return arma::Mat<ElemType>(input.memptr(), input.n_rows, input.n_cols, false,
-      strict);
-}
-
-/**
- * Make an alias of a dense row.  If strict is true, then the alias cannot be
- * resized or pointed at new memory.
- */
-template<typename ElemType>
-arma::Row<ElemType> MakeAlias(arma::Row<ElemType>& input,
-                              const bool strict = true)
-{
-  // Use the advanced constructor.
-  return arma::Row<ElemType>(input.memptr(), input.n_elem, false, strict);
-}
-
-/**
- * Make an alias of a dense column.  If strict is true, then the alias cannot be
- * resized or pointed at new memory.
- */
-template<typename ElemType>
-arma::Col<ElemType> MakeAlias(arma::Col<ElemType>& input,
-                              const bool strict = true)
-{
-  // Use the advanced constructor.
-  return arma::Col<ElemType>(input.memptr(), input.n_elem, false, strict);
-}
-
-/**
- * Make a copy of a sparse matrix (an alias is not possible).  The strict
- * parameter is ignored.
- */
-template<typename ElemType>
-arma::SpMat<ElemType> MakeAlias(const arma::SpMat<ElemType>& input,
-                                const bool /* strict */ = true)
-{
-  // Make a copy...
-  return arma::SpMat<ElemType>(input);
-}
-
-/**
- * Make a copy of a sparse row (an alias is not possible).  The strict
- * parameter is ignored.
- */
-template<typename ElemType>
-arma::SpRow<ElemType> MakeAlias(const arma::SpRow<ElemType>& input,
-                                const bool /* strict */ = true)
-{
-  // Make a copy...
-  return arma::SpRow<ElemType>(input);
-}
-
-/**
- * Make a copy of a sparse column (an alias is not possible).  The strict
- * parameter is ignored.
- */
-template<typename ElemType>
-arma::SpCol<ElemType> MakeAlias(const arma::SpCol<ElemType>& input,
-                                const bool /* strict */ = true)
+template<typename InVecType, typename OutVecType>
+void MakeAlias(OutVecType& v,
+               const InVecType& oldVec,
+               const size_t numElems,
+               const size_t offset = 0,
+               const bool strict = true,
+               const typename std::enable_if_t<IsVector<OutVecType>::value>* = 0)
+{
+  // We use placement new to reinitialize the object, since the copy and move
+  // assignment operators in Armadillo will end up copying memory instead of
+  // making an alias.
+  typename InVecType::elem_type* newMem =
+      const_cast<typename InVecType::elem_type*>(oldVec.memptr()) + offset;
+  v.~OutVecType();
+  new (&v) OutVecType(newMem, numElems, false, strict);
+}
+
+/**
+ * Reconstruct `m` as an alias around the memory `newMem`, with size `numRows` x
+ * `numCols`.
+ */
+template<typename InMatType, typename OutMatType>
+void MakeAlias(OutMatType& m,
+               const InMatType& oldMat,
+               const size_t numRows,
+               const size_t numCols,
+               const size_t offset = 0,
+               const bool strict = true,
+               const typename std::enable_if_t<IsMatrix<OutMatType>::value>* = 0)
+{
+  // We use placement new to reinitialize the object, since the copy and move
+  // assignment operators in Armadillo will end up copying memory instead of
+  // making an alias.
+  typename InMatType::elem_type* newMem =
+      const_cast<typename InMatType::elem_type*>(oldMat.memptr()) + offset;
+  m.~OutMatType();
+  new (&m) OutMatType(newMem, numRows, numCols, false, strict);
+}
+
+/**
+ * Reconstruct `c` as an alias around the memory` newMem`, with size `numRows` x
+ * `numCols` x `numSlices`.
+ */
+template<typename InCubeType, typename OutCubeType>
+void MakeAlias(OutCubeType& c,
+               const InCubeType& oldCube,
+               const size_t numRows,
+               const size_t numCols,
+               const size_t numSlices,
+               const size_t offset = 0,
+               const bool strict = true,
+               const typename std::enable_if_t<IsCube<OutCubeType>::value>* = 0)
+{
+  // We use placement new to reinitialize the object, since the copy and move
+  // assignment operators in Armadillo will end up copying memory instead of
+  // making an alias.
+  typename InCubeType::elem_type* newMem =
+      const_cast<typename InCubeType::elem_type*>(oldCube.memptr()) + offset;
+  c.~OutCubeType();
+  new (&c) OutCubeType(newMem, numRows, numCols, numSlices, false, strict);
+}
+
+/**
+ * Make `m` an alias of `in`, using the given size.
+ */
+template<typename eT>
+void MakeAlias(arma::SpMat<eT>& m,
+               const arma::SpMat<eT>& in,
+               const size_t /* numRows */,
+               const size_t /* numCols */,
+               const size_t /* offset */,
+               const bool /* strict */)
 {
-  // Make a copy...
-  return arma::SpCol<ElemType>(input);
+  // We can't make aliases of sparse objects, so just copy it.
+  m = in;
 }
 
 /**
@@ -113,16 +108,15 @@ void ClearAlias(arma::Mat<ElemType>& mat
 }
 
 /**
- * Clear an alias for a sparse matrix.  This does nothing because no sparse
- * matrices can have aliases.
+ * Clear an alias so that no data is overwritten.  This resets the matrix if it
+ * is an alias (and does nothing otherwise).
  */
 template<typename ElemType>
 void ClearAlias(arma::SpMat<ElemType>& /* mat */)
 {
-  // Nothing to do.
+  // We cannot make aliases of sparse matrices, so, nothing to do.
 }
 
-
 } // namespace mlpack
 
 #endif
diff -pruN 4.3.0-2/src/mlpack/core/math/math.hpp 4.4.0-1/src/mlpack/core/math/math.hpp
--- 4.3.0-2/src/mlpack/core/math/math.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/math.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -13,19 +13,18 @@
 #define MLPACK_CORE_MATH_MATH_HPP
 
 #include "ccov.hpp"
-#include "clamp.hpp"
 #include "columns_to_blocks.hpp"
 #include "digamma.hpp"
-#include "lin_alg.hpp"
 #include "log_add.hpp"
 #include "make_alias.hpp"
 #include "multiply_slices.hpp"
 #include "quantile.hpp"
 #include "random_basis.hpp"
 #include "random.hpp"
+#include "rand_vector.hpp"
 #include "range.hpp"
-#include "round.hpp"
 #include "shuffle_data.hpp"
 #include "trigamma.hpp"
+#include "unwrap_alias.hpp"
 
 #endif
diff -pruN 4.3.0-2/src/mlpack/core/math/multiply_slices_impl.hpp 4.4.0-1/src/mlpack/core/math/multiply_slices_impl.hpp
--- 4.3.0-2/src/mlpack/core/math/multiply_slices_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/multiply_slices_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -58,7 +58,7 @@ CubeType MultiplyCube2Cube(const CubeTyp
   if (aTranspose && bTranspose)
   {
     for (size_t i = 0; i < slices; ++i)
-      z.slice(i) = arma::trans(cubeB.slice(i) * cubeA.slice(i));
+      z.slice(i) = trans(cubeB.slice(i) * cubeA.slice(i));
   }
   else if (bTranspose && !aTranspose)
   {
@@ -116,7 +116,7 @@ CubeType MultiplyMat2Cube(const MatType&
   if (aTranspose && bTranspose)
   {
     for (size_t i = 0; i < slices; ++i)
-      z.slice(i) = arma::trans(cubeB.slice(i) * matA);
+      z.slice(i) = trans(cubeB.slice(i) * matA);
   }
   else if (bTranspose)
   {
@@ -172,7 +172,7 @@ CubeType MultiplyCube2Mat(const CubeType
   if (aTranspose && bTranspose)
   {
     for (size_t i = 0; i < slices; ++i)
-      z.slice(i) = arma::trans(matB * cubeA.slice(i));
+      z.slice(i) = trans(matB * cubeA.slice(i));
   }
   else if (bTranspose && !aTranspose)
   {
diff -pruN 4.3.0-2/src/mlpack/core/math/quantile.hpp 4.4.0-1/src/mlpack/core/math/quantile.hpp
--- 4.3.0-2/src/mlpack/core/math/quantile.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/quantile.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -19,7 +19,7 @@ namespace mlpack {
 /**
  * Computes the inverse erf function using the rational approximation from
  * Numerical Recipes.
- * Code accompanying the article "Approximating the erfinv function" in 
+ * Code accompanying the article "Approximating the erfinv function" in
  * GPU Computing Gems, Volume 2.
  *
  * @param x Input value.
@@ -28,9 +28,9 @@ inline double ErfInverse(double x)
 {
   double w, p;
 
-  w = -log((1.0 - x) * (1.0 + x));
+  w = -std::log((1.0 - x) * (1.0 + x));
 
-  if (w < 6.250000) 
+  if (w < 6.250000)
   {
     w = w - 3.125000;
     p = -3.6444120640178196996e-21;
@@ -56,10 +56,10 @@ inline double ErfInverse(double x)
     p = -0.0060336708714301490533 + p * w;
     p = 0.24015818242558961693 + p * w;
     p = 1.6536545626831027356 + p * w;
-  } 
-  else if (w < 16.000000) 
+  }
+  else if (w < 16.000000)
   {
-    w = sqrt(w) - 3.250000;
+    w = std::sqrt(w) - 3.250000;
     p = 2.2137376921775787049e-09;
     p = 9.0756561938885390979e-08 + p * w;
     p = -2.7517406297064545428e-07 + p * w;
@@ -79,10 +79,10 @@ inline double ErfInverse(double x)
     p = 0.005370914553590063617 + p * w;
     p = 1.0052589676941592334 + p * w;
     p = 3.0838856104922207635 + p * w;
-  } 
-  else 
+  }
+  else
   {
-    w = sqrt(w) - 5.000000;
+    w = std::sqrt(w) - 5.000000;
     p = -2.7109920616438573243e-11;
     p = -2.5556418169965252055e-10 + p * w;
     p = 1.5076572693500548083e-09 + p * w;
@@ -105,7 +105,7 @@ inline double ErfInverse(double x)
 }
 
 /**
- * Computes the quantile function of Guassian distribution at given probability.
+ * Computes the quantile function of Gaussian distribution at given probability.
  *
  * @param p Probability value.
  * @param mu Mean of the distribution. (Default 0)
diff -pruN 4.3.0-2/src/mlpack/core/math/rand_vector.hpp 4.4.0-1/src/mlpack/core/math/rand_vector.hpp
--- 4.3.0-2/src/mlpack/core/math/rand_vector.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/rand_vector.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,42 @@
+/**
+ * @file core/math/rand_vector.hpp
+ * @author Nishant Mehta
+ *
+ * Utility to generate a random vector on the unit sphere.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_MATH_RAND_VECTOR_HPP
+#define MLPACK_CORE_MATH_RAND_VECTOR_HPP
+
+namespace mlpack {
+
+/**
+ * Overwrites a dimension-N vector to a random vector on the unit sphere in R^N.
+ */
+inline void RandVector(arma::vec& v)
+{
+  for (size_t i = 0; i + 1 < v.n_elem; i += 2)
+  {
+    double a = Random();
+    double b = Random();
+    double first_term = std::sqrt(-2 * std::log(a));
+    double second_term = 2 * M_PI * b;
+    v[i]     = first_term * cos(second_term);
+    v[i + 1] = first_term * sin(second_term);
+  }
+
+  if ((v.n_elem % 2) == 1)
+  {
+    v[v.n_elem - 1] = std::sqrt(-2 * std::log(Random())) * cos(2 * M_PI * Random());
+  }
+
+  v /= std::sqrt(dot(v, v));
+}
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/math/random.hpp 4.4.0-1/src/mlpack/core/math/random.hpp
--- 4.3.0-2/src/mlpack/core/math/random.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/random.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -146,7 +146,8 @@ inline int RandInt(const int lo, const i
 }
 
 /**
- * Generates a normally distributed random number with mean 0 and variance 1.
+ * Generates a normally distributed random number with mean 0 and standard
+ * deviation of 1.
  */
 inline double RandNormal()
 {
@@ -155,52 +156,14 @@ inline double RandNormal()
 
 /**
  * Generates a normally distributed random number with specified mean and
- * variance.
+ * standard deviation.
  *
  * @param mean Mean of distribution.
- * @param variance Variance of distribution.
+ * @param stddev Standard deviation of distribution.
  */
-inline double RandNormal(const double mean, const double variance)
+inline double RandNormal(const double mean, const double stddev)
 {
-  return variance * RandNormalDist()(RandGen()) + mean;
-}
-
-/**
- * Obtains no more than maxNumSamples distinct samples. Each sample belongs to
- * [loInclusive, hiExclusive).
- *
- * @param loInclusive The lower bound (inclusive).
- * @param hiExclusive The high bound (exclusive).
- * @param maxNumSamples The maximum number of samples to obtain.
- * @param distinctSamples The samples that will be obtained.
- */
-inline void ObtainDistinctSamples(const size_t loInclusive,
-                                  const size_t hiExclusive,
-                                  const size_t maxNumSamples,
-                                  arma::uvec& distinctSamples)
-{
-  const size_t samplesRangeSize = hiExclusive - loInclusive;
-
-  if (samplesRangeSize > maxNumSamples)
-  {
-    arma::Col<size_t> samples;
-
-    samples.zeros(samplesRangeSize);
-
-    for (size_t i = 0; i < maxNumSamples; ++i)
-      samples [ (size_t) RandInt(samplesRangeSize) ]++;
-
-    distinctSamples = arma::find(samples > 0);
-
-    if (loInclusive > 0)
-      distinctSamples += loInclusive;
-  }
-  else
-  {
-    distinctSamples.set_size(samplesRangeSize);
-    for (size_t i = 0; i < samplesRangeSize; ++i)
-      distinctSamples[i] = loInclusive + i;
-  }
+  return stddev * RandNormalDist()(RandGen()) + mean;
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/core/math/round.hpp 4.4.0-1/src/mlpack/core/math/round.hpp
--- 4.3.0-2/src/mlpack/core/math/round.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/round.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,31 +0,0 @@
-/**
- * @file core/math/round.hpp
- * @author Ryan Curtin
- *
- * Implementation of round() for use on Visual Studio, where C99 isn't
- * implemented.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_CORE_MATH_ROUND_HPP
-#define MLPACK_CORE_MATH_ROUND_HPP
-
-// _MSC_VER should only be defined for Visual Studio, which doesn't implement
-// C99.
-#ifdef _MSC_VER
-
-// This function ends up going into the global namespace, so it can be used in
-// place of C99's round().
-
-//! Round a number to the nearest integer.
-inline double round(double a)
-{
-  return floor(a + 0.5);
-}
-
-#endif
-
-#endif
diff -pruN 4.3.0-2/src/mlpack/core/math/trigamma.hpp 4.4.0-1/src/mlpack/core/math/trigamma.hpp
--- 4.3.0-2/src/mlpack/core/math/trigamma.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/trigamma.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -189,7 +189,7 @@ T Trigamma(T x)
 
     T s = fabs(x) < fabs(z) ? sin(M_PI * x) : sin(M_PI * z);
 
-    return -Trigamma(z) + pow(M_PI, 2) / (s * s);
+    return -Trigamma(z) + std::pow(M_PI, 2) / (s * s);
   }
 
   if (x < 1)
diff -pruN 4.3.0-2/src/mlpack/core/math/unwrap_alias.hpp 4.4.0-1/src/mlpack/core/math/unwrap_alias.hpp
--- 4.3.0-2/src/mlpack/core/math/unwrap_alias.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/math/unwrap_alias.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,38 @@
+/**
+ * @file core/math/unwrap_alias.hpp
+ * @author Ryan Curtin
+ *
+ * Make an alias of a matrix if possible, or unwrap it if an expression is
+ * given.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_MATH_UNWRAP_ALIAS_HPP
+#define MLPACK_CORE_MATH_UNWRAP_ALIAS_HPP
+
+namespace mlpack {
+
+/**
+ * If `in` is an expression, unwrap it into `m`.  If `in` is a matrix, then
+ * create `m` as an alias of `in`.
+ * `numCols`.
+ */
+template<typename MatType>
+void UnwrapAlias(MatType& m, const MatType& in)
+{
+  MakeAlias(m, in, in.n_rows, in.n_cols);
+}
+
+template<typename MatType, typename InMatType>
+void UnwrapAlias(MatType& m,
+                 const InMatType& in)
+{
+  m = in;
+}
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/metrics/ip_metric_impl.hpp 4.4.0-1/src/mlpack/core/metrics/ip_metric_impl.hpp
--- 4.3.0-2/src/mlpack/core/metrics/ip_metric_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/metrics/ip_metric_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -76,7 +76,7 @@ inline typename Vec1Type::elem_type IPMe
 {
   // This is the metric induced by the kernel function.
   // Maybe we can do better by caching some of this?
-  return sqrt(kernel->Evaluate(a, a) + kernel->Evaluate(b, b) -
+  return std::sqrt(kernel->Evaluate(a, a) + kernel->Evaluate(b, b) -
       2 * kernel->Evaluate(a, b));
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/metrics/lmetric_impl.hpp 4.4.0-1/src/mlpack/core/metrics/lmetric_impl.hpp
--- 4.3.0-2/src/mlpack/core/metrics/lmetric_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/metrics/lmetric_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,7 +41,7 @@ typename VecTypeA::elem_type LMetric<1,
     const VecTypeA& a,
     const VecTypeB& b)
 {
-  return arma::accu(abs(a - b));
+  return accu(abs(a - b));
 }
 
 template<>
@@ -50,7 +50,7 @@ typename VecTypeA::elem_type LMetric<1,
     const VecTypeA& a,
     const VecTypeB& b)
 {
-  return arma::accu(abs(a - b));
+  return accu(abs(a - b));
 }
 
 // L2-metric specializations.
@@ -83,7 +83,7 @@ typename VecTypeA::elem_type LMetric<3,
   for (size_t i = 0; i < a.n_elem; ++i)
     sum += std::pow(fabs(a[i] - b[i]), 3.0);
 
-  return std::pow(arma::accu(arma::pow(arma::abs(a - b), 3.0)), 1.0 / 3.0);
+  return std::pow(accu(pow(arma::abs(a - b), 3.0)), 1.0 / 3.0);
 }
 
 template<>
@@ -92,7 +92,7 @@ typename VecTypeA::elem_type LMetric<3,
     const VecTypeA& a,
     const VecTypeB& b)
 {
-  return arma::accu(arma::pow(arma::abs(a - b), 3.0));
+  return accu(pow(arma::abs(a - b), 3.0));
 }
 
 // L-infinity (Chebyshev distance) specialization
diff -pruN 4.3.0-2/src/mlpack/core/metrics/mahalanobis_distance_impl.hpp 4.4.0-1/src/mlpack/core/metrics/mahalanobis_distance_impl.hpp
--- 4.3.0-2/src/mlpack/core/metrics/mahalanobis_distance_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/metrics/mahalanobis_distance_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -43,7 +43,7 @@ double MahalanobisDistance<true>::Evalua
 
   arma::vec m = (a - b);
   arma::mat out = trans(m) * covariance * m; // 1x1;
-  return sqrt(out[0]);
+  return std::sqrt(out[0]);
 }
 
 // Serialize the Mahalanobis distance.
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/LICENSE.txt 4.4.0-1/src/mlpack/core/std_backport/LICENSE.txt
--- 4.3.0-2/src/mlpack/core/std_backport/LICENSE.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/LICENSE.txt	1970-01-01 00:00:00.000000000 +0000
@@ -1,12 +0,0 @@
-Copyright © 2013 - 2015 MNMLSTC
-
-Licensed under the Apache License, Version 2.0 (the "License"); you may
-not use this software except in compliance with the License. You may 
-obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
-
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS, 
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
-implied. See the License for the specific language governing 
-permissions and limitations under the License.
-
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/README.md 4.4.0-1/src/mlpack/core/std_backport/README.md
--- 4.3.0-2/src/mlpack/core/std_backport/README.md	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/README.md	1970-01-01 00:00:00.000000000 +0000
@@ -1,15 +0,0 @@
-The files in this directory are taken from MNMLSTC Core 1.1.0 in order to 
-backport features from C++ 17 standard library:
-
- * C++17 STL algorithms such as std::any and std::basic_string_view. 
- * Dependencies files that are used to implement these features.
-
-These files are licensed under the Apache 2.0 License, available in LICENSE.txt
-in this directory.
-
-If you want a copy of mlpack without a dependence on the Apache License or 
-without the backported version then you will need to
-
- * Remove this entire directory.
- * Remove the line "std_backport" from src/mlpack/core/CMakeLists.txt.
- * Use the C++17 standard by modifying the mlpack/CMakeLists.txt.
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/algorithm.hpp 4.4.0-1/src/mlpack/core/std_backport/algorithm.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/algorithm.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/algorithm.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,2280 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_ALGORITHM_HPP
-#define CORE_ALGORITHM_HPP
-
-#include <algorithm>
-
-#include "functional.hpp"
-#include "utility.hpp"
-#include "range.hpp"
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-template <class InputIt1, class InputIt2, class Predicate>
-bool equal (
-  range<InputIt1> r1,
-  range<InputIt2> r2,
-  Predicate&& p,
-  ::std::random_access_iterator_tag,
-  ::std::random_access_iterator_tag
-) {
-  if (r1.size() != r2.size()) { return false; }
-  return ::std::equal(
-    begin(r1),
-    end(r1),
-    begin(r2),
-    ::core::forward<Predicate>(p)
-  );
-}
-
-template <class InputIt1, class InputIt2, class Predicate>
-bool equal (
-  range<InputIt1> r1,
-  range<InputIt2> r2,
-  Predicate&& p,
-  ::std::input_iterator_tag,
-  ::std::input_iterator_tag
-) {
-  while (not r1.empty() and not r2.empty()) {
-    if (
-      not ::core::invoke(
-        ::core::forward<Predicate>(p),
-        r1.front(),
-        r2.front())
-    ) { return false; }
-    r1.pop_front();
-    r2.pop_front();
-  }
-  return r1.empty() and r2.empty();
-}
-
-} /* namespace impl */
-
-/* non-range based algorithms */
-template <class T>
-constexpr T const& min (T const& lhs, T const& rhs) {
-  return (rhs < lhs) ? rhs : lhs;
-}
-
-template <class T, class Compare>
-constexpr T const& min (T const& lhs, T const& rhs, Compare compare) {
-  return compare(rhs, lhs) ? rhs : lhs;
-}
-
-template <class T>
-constexpr T const& max (T const& lhs, T const& rhs) {
-  return (lhs < rhs) ? rhs : lhs;
-}
-
-template <class T, class Compare>
-constexpr T const& max (T const& lhs, T const& rhs, Compare compare) {
-  return compare(lhs, rhs) ? rhs : lhs;
-}
-
-/* extensions */
-template <class T, class Compare=less<>>
-constexpr T const& clamp (
-  T const& value,
-  T const& low,
-  T const& high,
-  Compare compare = Compare { }
-) {
-  return compare(value, low)
-    ? low
-    : compare(high, value)
-      ? high
-      : value;
-}
-
-/* N4318 (modified) */
-template <
-  class T,
-  class Compare = ::core::less<>,
-  class Difference = ::core::minus<>
-> constexpr auto abs_diff (
-  T const& a,
-  T const& b,
-  Compare compare = Compare { },
-  Difference diff = Difference { }
-) -> decltype(compare(a, b) ? diff(b, a) : diff(a, b)) {
-  return compare(a, b) ? diff(b, a) : diff(a, b);
-}
-
-/* non-modifying sequence algorithms */
-template <class Range, class UnaryPredicate>
-auto all_of (Range&& rng, UnaryPredicate&& p) -> meta::when<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "all_of requires InputIterators");
-  return ::std::all_of(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto any_of (Range&& rng, UnaryPredicate&& p) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "any_of requires InputIterators");
-  return ::std::any_of(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto none_of (Range&& rng, UnaryPredicate&& p) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "none_of requires InputIterators");
-  return ::std::none_of(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class Range, class UnaryFunction>
-auto for_each (Range&& rng, UnaryFunction&& f) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<UnaryFunction>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "for_each requires InputIterators");
-  return ::std::for_each(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryFunction>(f)
-  );
-}
-
-template <class Range, class UnaryFunction, class UnaryPredicate>
-UnaryFunction for_each_if (Range&& r, UnaryFunction uf, UnaryPredicate up) {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "for_each_if requires InputIterators");
-  while (not range.empty()) {
-    if (up(range.front())) { uf(range.front()); }
-    range.pop_front();
-  }
-  return uf;
-}
-
-template <class Range, class UnaryFunction, class UnaryPredicate>
-auto for_each_while (
-  Range&& r,
-  UnaryFunction f,
-  UnaryPredicate p
-) -> decltype(begin(make_range(::core::forward<Range>(r)))) {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "for_each_while requires InputIterators");
-  while (not range.empty()) {
-    if (not p(range.front())) { break; }
-    f(range.front());
-    range.pop_front();
-  }
-  return range.begin();
-}
-
-template <class Range, class UnaryFunction, class T>
-auto for_each_until (
-  Range&& r,
-  UnaryFunction f,
-  T const& value
-) -> decltype(begin(make_range(::core::forward<Range>(r)))) {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "for_each_until requires InputIterators");
-  while (not range.empty()) {
-    if (range.front() == value) { break; }
-    f(range.front());
-    range.pop_front();
-  }
-  return range.begin();
-}
-
-template <class Range, class T>
-auto count (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(
-    ::std::count(
-      ::std::begin(::core::forward<Range>(rng)),
-      ::std::end(::core::forward<Range>(rng)),
-      value
-    )
-  )
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "count requires InputIterators");
-  return ::std::count(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class UnaryPredicate>
-auto count_if (Range&& rng, UnaryPredicate&& p) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(
-    ::std::count_if(
-      ::std::begin(::core::forward<Range>(rng)),
-      ::std::end(::core::forward<Range>(rng)),
-      ::core::forward<UnaryPredicate>(p)
-    )
-  )
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "count_if requires InputIterators");
-  return ::std::count_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class InputIt1, class InputIt2, class BinaryPredicate>
-::std::pair<InputIt1, InputIt2> mismatch (
-  InputIt1 first1,
-  InputIt1 last1,
-  InputIt2 first2,
-  InputIt2 last2,
-  BinaryPredicate predicate
-) {
-  auto r1 = make_range(first1, last1);
-  auto r2 = make_range(first2, last2);
-  while (not r1.empty() and not r2.empty()) {
-    if (not predicate(r1.front(), r2.front())) { break; }
-    r1.pop_front();
-    r2.pop_front();
-  }
-  return ::std::make_pair(r1.begin(), r2.begin());
-}
-
-template <class InputIt1, class InputIt2>
-::std::pair<InputIt1, InputIt2> mismatch (
-  InputIt1 first1,
-  InputIt1 last1,
-  InputIt2 first2,
-  InputIt2 last2
-) { return (mismatch)(first1, last1, first2, last2, equal_to<> { }); }
-
-template <
-  class Range1,
-  class Range2,
-  class BinaryPred,
-  meta::require<
-    meta::all_of<meta::list<Range1, Range2>, is_range>()
-  > = __LINE__
-> auto mismatch (Range1&& r1, Range2&& r2, BinaryPred&& bp) -> ::std::pair<
-  decltype(::std::begin(::core::forward<Range1>(r1))),
-  decltype(::std::begin(::core::forward<Range2>(r2)))
-> {
-  auto range1 = make_range(::core::forward<Range1>(r1));
-  auto range2 = make_range(::core::forward<Range2>(r2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1 and is_input2, "mismatch requires InputIterators");
-  return (mismatch)(
-    range1.begin(),
-    range1.end(),
-    range2.begin(),
-    range2.end(),
-    ::core::forward<BinaryPred>(bp));
-}
-
-template <
-  class Range,
-  class InputIt,
-  meta::require<
-    meta::all<is_range<Range>::value, meta::none<is_range<InputIt>::value>()>()
-  > = __LINE__
-> auto mismatch(Range&& rng, InputIt&& it) -> ::std::pair<
-  decltype(make_range(::core::forward<Range>(rng)).begin()),
-  decay_t<InputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "mismatch requires InputIterators");
-  return ::std::mismatch(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<InputIt>(it)
-  );
-}
-
-template <
-  class Range,
-  class InputIt,
-  class BinaryPredicate,
-  meta::require<
-    meta::all<is_range<Range>::value, meta::none<is_range<InputIt>>()>()
-  > = __LINE__
-> auto mismatch(Range&& r, InputIt&& it, BinaryPredicate&& bp) -> ::std::pair<
-  decltype(core::make_range(::core::forward<Range>(r).begin())),
-  decay_t<InputIt>
-> {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "mismatch requires InputIterators");
-  return ::std::mismatch(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<InputIt>(it),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <class InputIt1, class InputIt2, class BinaryPredicate>
-bool equal (
-  InputIt1 first1,
-  InputIt1 last1,
-  InputIt2 first2,
-  InputIt2 last2,
-  BinaryPredicate bp
-) {
-  auto r1 = make_range(first1, last1);
-  auto r2 = make_range(first2, last2);
-  using tag1 = typename decltype(r1)::iterator_category;
-  using tag2 = typename decltype(r2)::iterator_category;
-  return impl::equal(r1, r2, bp, tag1 { }, tag2 { });
-}
-
-template <class InputIt1, class InputIt2>
-bool equal (InputIt1 first1, InputIt1 last1, InputIt2 first2, InputIt2 last2) {
-  return equal(first1, last1, first2, last2, equal_to<> { });
-}
-
-template <
-  class Range1,
-  class Range2,
-  meta::require<
-    meta::all_of<meta::list<Range1, Range2>, is_range>()
-  > = __LINE__
-> bool equal (Range1&& range1, Range2&& range2) {
-  auto r1 = make_range(::core::forward<Range1>(range1));
-  auto r2 = make_range(::core::forward<Range2>(range2));
-  static constexpr auto is_input1 = decltype(r1)::is_input;
-  static constexpr auto is_input2 = decltype(r2)::is_input;
-  static_assert(is_input1, "equal requires InputIterators");
-  static_assert(is_input2, "equal requires InputIterators");
-  return (equal)(r1.begin(), r1.end(), r2.begin(), r2.end());
-}
-
-template <
-  class Range1,
-  class Range2,
-  class BinaryPredicate,
-  meta::require<
-    meta::all_of<meta::list<Range1, Range2>, is_range>()
-  > = __LINE__
-> bool equal (Range1&& range1, Range2&& range2, BinaryPredicate&& bp) {
-  auto r1 = make_range(::core::forward<Range1>(range1));
-  auto r2 = make_range(::core::forward<Range2>(range2));
-  static constexpr auto is_input1 = decltype(r1)::is_input;
-  static constexpr auto is_input2 = decltype(r2)::is_input;
-  static_assert(is_input1, "equal requires InputIterators");
-  static_assert(is_input2, "equal requires InputIterators");
-  return equal(
-    r1.begin(),
-    r1.end(),
-    r2.begin(),
-    r2.end(),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <
-  class Range,
-  class InputIt,
-  meta::require<
-    meta::all<is_range<Range>::value, meta::none<is_range<InputIt>::value>()>()
-  > = __LINE__
-> bool equal (Range&& rng, InputIt&& it) {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "equal requires InputIterators");
-  return ::std::equal(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<InputIt>(it)
-  );
-}
-
-template <
-  class Range,
-  class InputIt,
-  class BinaryPredicate,
-  meta::require<
-    meta::all<is_range<Range>::value, meta::none<is_range<InputIt>>()>()
-  > = __LINE__
-> bool equal (Range&& rng, InputIt&& it, BinaryPredicate&& bp) {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "equal requires InputIterators");
-  return ::std::equal(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<InputIt>(it),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <class Range, class T>
-auto find (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "find requires InputIterators");
-  return ::std::find(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class UnaryPredicate>
-auto find_if (Range&& rng, UnaryPredicate&& p) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "find_if requires InputIterators");
-  return ::std::find_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto find_if_not (Range&& rng, UnaryPredicate&& p) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "find_if_not requires InputIterators");
-  return ::std::find_if_not(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(p)
-  );
-}
-
-template <class Range1, class Range2>
-auto find_end (Range1&& rng1, Range2&& rng2) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decltype(::std::begin(::core::forward<Range1>(rng1)))
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "find_end requires ForwardIterators");
-  static_assert(is_forward2, "find_end requires ForwardIterators");
-  return ::std::find_end(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2)
-  );
-}
-
-template <class Range1, class Range2, class BinaryPred>
-auto find_end (Range1&& rng1, Range2&& rng2, BinaryPred& bp) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decltype(::std::begin(::core::forward<Range1>(rng1)))
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "find_end requires ForwardIterators");
-  static_assert(is_forward2, "find_end requires ForwardIterators");
-  return ::std::find_end(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<BinaryPred>(bp)
-  );
-}
-
-template <class IRange, class FRange>
-auto find_first_of (IRange&& irng, FRange&& frng) -> meta::when<
-  meta::all_of<meta::list<IRange, FRange>, is_range>(),
-  decltype(::std::begin(::core::forward<IRange>(irng)))
-> {
-  auto irange = make_range(::core::forward<IRange>(irng));
-  auto frange = make_range(::core::forward<FRange>(frng));
-  static constexpr auto is_input = decltype(irange)::is_input;
-  static constexpr auto is_forward = decltype(frange)::is_forward;
-  static_assert(is_input, "find_first_of requires InputIterators");
-  static_assert(is_forward, "find_first_of requires ForwardIterators");
-  return ::std::find_first_of(
-    ::std::begin(irange),
-    ::std::end(irange),
-    ::std::begin(frange),
-    ::std::end(frange)
-  );
-}
-
-template <class IRange, class FRange, class BinaryPred>
-auto find_first_of (
-  IRange&& irng,
-  FRange&& frng,
-  BinaryPred&& bp
-) -> meta::when<
-  meta::all_of<meta::list<IRange, FRange>, is_range>(),
-  decltype(::std::begin(::core::forward<IRange>(irng)))
-> {
-  auto irange = make_range(::core::forward<IRange>(irng));
-  auto frange = make_range(::core::forward<FRange>(frng));
-  static constexpr auto is_input = decltype(irange)::is_input;
-  static constexpr auto is_forward = decltype(frange)::is_forward;
-  static_assert(is_input, "find_first_of requires InputIterators");
-  static_assert(is_forward, "find_first_of requires ForwardIterators");
-  return ::std::find_first_of(
-    ::std::begin(irange),
-    ::std::end(irange),
-    ::std::begin(frange),
-    ::std::end(frange),
-    ::core::forward<BinaryPred>(bp)
-  );
-}
-
-template <class Range>
-auto adjacent_find (Range&& rng) -> meta::when<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "adjacent_find requires ForwardIterators");
-  return ::std::adjacent_find(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class BinaryPredicate>
-auto adjacent_find (Range&& rng, BinaryPredicate&& bp) -> meta::when<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "adjacent_find requires ForwardIterators");
-  return ::std::adjacent_find(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <class Range1, class Range2>
-auto search (Range1&& rng1, Range2&& rng2) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decltype(::std::begin(::core::forward<Range1>(rng1)))
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "search requires ForwardIterators");
-  static_assert(is_forward2, "search requires ForwardIterators");
-  return ::std::search(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2)
-  );
-}
-
-template <class Range1, class Range2, class BinaryPred>
-auto search (Range1&& rng1, Range2&& rng2, BinaryPred&& bp) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decltype(::std::begin(::core::forward<Range1>(rng1)))
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "search requires ForwardIterators");
-  static_assert(is_forward2, "search requires ForwardIterators");
-  return ::std::search(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<BinaryPred>(bp)
-  );
-}
-
-template <class Range, class Size, class T>
-auto search_n (Range&& rng, Size&& count, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "search_n requires ForwardIterators");
-  return ::std::search_n(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Size>(count),
-    value
-  );
-}
-
-template <class Range, class Size, class T, class BinaryPred>
-auto search_n (
-  Range&& rng,
-  Size&& count,
-  T const& value,
-  BinaryPred&& bp
-) -> meta::when<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "search_n requires ForwardIterators");
-  return ::std::search_n(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Size>(count),
-    value,
-    ::core::forward<BinaryPred>(bp)
-  );
-}
-
-/* modifying sequence algorithms */
-template <class Range, class OutputIt>
-auto copy (Range&& rng, OutputIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "copy requires InputIterators");
-  return ::std::copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range, class OutputIt, class UnaryPredicate>
-auto copy_if (Range&& rng, OutputIt&& it, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "copy_if requires InputIterators");
-  return ::std::copy_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-template <class Range, class OutputIt, class T>
-OutputIt copy_until (Range&& r, OutputIt it, T const& value) {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "copy_until requires InputIterators");
-  while (not range.empty()) {
-    if (range.front() == value) { break; }
-    *it++ = range.front();
-    range.pop_front();
-  }
-  return it;
-}
-
-template <class Range, class BidirIt>
-auto copy_backward (Range&& rng, BidirIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<BidirIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "copy_backward requires BidirectionalIterators");
-  return ::std::copy_backward(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<BidirIt>(it)
-  );
-}
-
-template <class Range, class OutputIt>
-auto move (Range&& rng, OutputIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "move requires InputIterators");
-  return ::std::move(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range, class BidirIt>
-auto move_backward (Range&& rng, BidirIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<BidirIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "move_backward requires BidirectionalIterators");
-  return ::std::move_backward(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<BidirIt>(it)
-  );
-}
-
-template <class Range, class T>
-auto fill (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "fill requires ForwardIterators");
-  return ::std::fill(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class OutputIt, class UnaryOperation>
-auto transform (
-  Range&& rng,
-  OutputIt&& it,
-  UnaryOperation&& op
-) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "transform requires InputIterators");
-  return ::std::transform(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<UnaryOperation>(op)
-  );
-}
-
-template <class Range, class OutputIt, class UnaryOperation, class UnaryPred>
-auto transform_if (
-  Range&& rng,
-  OutputIt it,
-  UnaryOperation op,
-  UnaryPred up
-) -> enable_if_t<
-  is_range<Range>::value,
-  OutputIt
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "transform_if requires ForwardIterators");
-  while (not range.empty()) {
-    if (invoke(up, range.front())) {
-      *it = invoke(op, range.front());
-      ++it;
-    }
-    range.pop_front();
-  }
-  return it;
-}
-
-template <class InputIt, class Size, class OutputIt, class UnaryOp>
-OutputIt transform_n (InputIt in, Size count, OutputIt out, UnaryOp op) {
-  while (count > 0) {
-    *out = invoke(op, *in);
-    ++out;
-    ++in;
-    --count;
-  }
-  return out;
-}
-
-template <
-  class InputIt1,
-  class InputIt2,
-  class Size,
-  class OutputIt,
-  class UnaryOp
-> OutputIt transform_n (
-  InputIt1 in1,
-  InputIt2 in2,
-  Size count,
-  OutputIt out,
-  UnaryOp op
-) {
-  while (count > 0) {
-    *out = invoke(op, *in1, *in2);
-    ++out;
-    ++in1;
-    ++in2;
-    --count;
-  }
-  return out;
-}
-
-
-template <
-  class Range,
-  class InputIt,
-  class OutputIt,
-  class BinaryOperation,
-  meta::require<
-    meta::all<is_range<Range>::value, meta::none<is_range<InputIt>::value>()>()
-  > = __LINE__
-> decay_t<OutputIt> transform (
-  Range&& r,
-  InputIt&& in,
-  OutputIt&& out,
-  BinaryOperation&& op
-) {
-  auto range = make_range(::core::forward<Range>(r));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "transform requires InputIterators");
-  return ::std::transform(
-    range.begin(),
-    range.end(),
-    ::core::forward<InputIt>(in),
-    ::core::forward<OutputIt>(out),
-    ::core::forward<BinaryOperation>(op));
-}
-
-template <
-  class Range1,
-  class Range2,
-  class OutputIt,
-  class BinaryOperation,
-  meta::require<
-    meta::all_of<meta::list<Range1, Range2>, is_range>()
-  > = __LINE__
-> decay_t<OutputIt> transform (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  BinaryOperation&& op
-) {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "transform requires InputIterators");
-  static_assert(is_input2, "transform requires InputIterators");
-  return ::std::transform(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<BinaryOperation>(op)
-  );
-}
-
-template <
-  class Range1,
-  class Range2,
-  class OutputIt,
-  class BinaryOperation,
-  class BinaryPredicate
-> auto transform_if (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt it,
-  BinaryOperation op,
-  BinaryPredicate bp
-) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  OutputIt
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "transform_if requires ForwardIterators");
-  static_assert(is_forward2, "transform_if requires ForwardIterators");
-  while (not range1.empty()) {
-    if (bp(range1.front(), range2.front())) {
-      *it = op(range1.front(), range2.front());
-      ++it;
-    }
-    range1.pop_front();
-    range2.pop_front();
-  }
-  return it;
-}
-
-template <class Range, class T>
-auto remove (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "remove requires ForwardIterators");
-  return ::std::remove(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class UnaryPredicate>
-auto remove_if (Range&& rng, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "remove_if requires ForwardIterators");
-  return ::std::remove_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-template <class Range, class OutputIt, class T>
-auto remove_copy (Range&& rng, OutputIt&& it, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "remove_copy requires InputIterators");
-  return ::std::remove_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    value
-  );
-}
-
-template <class Range, class OutputIt, class UnaryPred>
-auto remove_copy_if (Range&& rng, OutputIt&& it, UnaryPred&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "remove_copy_if requires InputIterators");
-  return ::std::remove_copy_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<UnaryPred>(up)
-  );
-}
-
-template <class Range, class T>
-auto remove_erase (Range&& rng, T const& val) -> enable_if_t<
-  is_range<Range>::value
-> {
-  ::core::forward<Range>(rng).erase(
-    remove(::core::forward<Range>(rng), val),
-    ::std::end(::core::forward<Range>(rng))
-  );
-}
-
-template <class Range, class UnaryPred>
-auto remove_erase_if (Range&& rng, UnaryPred&& up) -> enable_if_t<
-  is_range<Range>::value
-> {
-  ::core::forward<Range>(rng).erase(
-    remove_if(
-      ::core::forward<Range>(rng),
-      ::core::forward<UnaryPred>(up)
-    ),
-    ::std::end(::core::forward<Range>(rng))
-  );
-}
-
-template <class Range, class T>
-auto replace (Range&& rng, T const& old, T const& value) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_input;
-  static_assert(is_forward, "replace requires ForwardIterators");
-  return ::std::replace(
-    ::std::begin(range),
-    ::std::end(range),
-    old,
-    value
-  );
-}
-
-template <class Range, class UnaryPred, class T>
-auto replace_if (Range&& rng, UnaryPred&& up, T const& value) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "replace_if requires ForwardIterators");
-  return ::std::replace_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPred>(up),
-    value
-  );
-}
-
-template <class Range, class OutputIt, class T>
-auto replace_copy (
-  Range&& rng,
-  OutputIt&& it,
-  T const& old,
-  T const& value
-) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "replace_copy requires InputIterators");
-  return ::std::replace_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    old,
-    value
-  );
-}
-
-template <class Range, class OutputIt, class UnaryPred, class T>
-auto replace_copy_if (
-  Range&& rng,
-  OutputIt&& it,
-  UnaryPred&& up,
-  T const& value
-) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "replace_copy_if requires InputIterators");
-  return ::std::replace_copy_if(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<UnaryPred>(up),
-    value
-  );
-}
-
-template <class Range, class ForwardIt>
-auto swap_ranges (Range&& rng, ForwardIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<ForwardIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "swap_ranges requires ForwardIterators");
-  return ::std::swap_ranges(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<ForwardIt>(it)
-  );
-}
-
-template <class Range>
-auto reverse (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "reverse requires BidirectionalIterators");
-  return ::std::reverse(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class OutputIt>
-auto reverse_copy (Range&& rng, OutputIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "reverse_copy requires BidirectionalIterators");
-  return ::std::reverse_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range, class ForwardIt>
-auto rotate (Range&& rng, ForwardIt&& it) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "rotate requires ForwardIterators");
-  ::std::rotate(
-    ::std::begin(range),
-    ::core::forward<ForwardIt>(it),
-    ::std::end(range)
-  );
-}
-
-template <class Range, class ForwardIt, class OutputIt>
-auto rotate_copy (Range&& rng, ForwardIt&& it, OutputIt&& ot) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "rotate_copy requires ForwardIterators");
-  return ::std::rotate_copy(
-    ::std::begin(range),
-    ::core::forward<ForwardIt>(it),
-    ::std::end(range),
-    ::core::forward<OutputIt>(ot)
-  );
-}
-
-template <class Range, class URNG>
-auto shuffle (Range&& rng, URNG&& g) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "shuffle requires RandomAccessIterators");
-  return ::std::shuffle(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<URNG>(g)
-  );
-}
-
-template <class Range>
-auto unique (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "unique requires ForwardIterators");
-  return ::std::unique(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class BinaryPredicate>
-auto unique (Range&& rng, BinaryPredicate&& bp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "unique requires ForwardIterators");
-  return ::std::unique(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <class Range, class OutputIt>
-auto unique_copy (Range&& rng, OutputIt&& it) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "unique_copy requires InputIterators");
-  return ::std::unique_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range, class OutputIt, class BinaryPred>
-auto unique_copy (Range&& rng, OutputIt&& it, BinaryPred&& bp) -> enable_if_t<
-  is_range<Range>::value,
-  decay_t<OutputIt>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "unique_copy requires InputIterators");
-  return ::std::unique_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<BinaryPred>(bp)
-  );
-}
-
-/* partitioning operations */
-template <class Range, class UnaryPredicate>
-auto is_partitioned (Range&& rng, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "is_partitioned requires InputIterators");
-  return ::std::is_partitioned(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto partition (Range&& rng, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "partition requires ForwardIterators");
-  return ::std::partition(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-template <class Range, class OutputTrue, class OutputFalse, class UnaryPred>
-auto partition_copy (
-  Range&& rng,
-  OutputTrue&& ot,
-  OutputFalse&& of,
-  UnaryPred&& up
-) -> enable_if_t<
-  is_range<Range>::value,
-  ::std::pair<decay_t<OutputTrue>, decay_t<OutputFalse>>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_input = decltype(range)::is_input;
-  static_assert(is_input, "partition_copy requires InputIterators");
-  return ::std::partition_copy(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<OutputTrue>(ot),
-    ::core::forward<OutputFalse>(of),
-    ::core::forward<UnaryPred>(up)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto stable_partition (Range&& rng, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "stable_partition requires BidirectionalIterators");
-  return ::std::stable_partition(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-template <class Range, class UnaryPredicate>
-auto partition_point (Range&& rng, UnaryPredicate&& up) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "partition_point requires ForwardIterators");
-  return ::std::partition_point(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<UnaryPredicate>(up)
-  );
-}
-
-/* sorting operations */
-
-template <class Range>
-auto is_sorted (Range&& rng) -> enable_if_t<is_range<Range>::value, bool> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "is_sorted requires ForwardIterators");
-  return ::std::is_sorted(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto is_sorted (Range&& rng, Compare&& compare) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "is_sorted requires ForwardIterators");
-  return ::std::is_sorted(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(compare)
-  );
-}
-
-template <class Range>
-auto is_sorted_until (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "is_sorted_until requires ForwardIterators");
-  return ::std::is_sorted_until(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto is_sorted_until (Range&& rng, Compare&& compare) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "is_sorted_until requires ForwardIterators");
-  return ::std::is_sorted_until(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(compare)
-  );
-}
-
-template <class Range>
-auto sort (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "sort requires RandomAccessIterators");
-  return ::std::sort(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto sort (Range&& rng, Compare&& cmp) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "sort requires RandomAccessIterators");
-  return ::std::sort(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class RandomIt>
-auto partial_sort (Range&& rng, RandomIt&& it) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "partial_sort requires RandomAccessIterators");
-  return ::std::partial_sort(
-    ::std::begin(range),
-    ::core::forward<RandomIt>(it),
-    ::std::end(range)
-  );
-}
-
-template <class Range, class RandomIt, class Compare>
-auto partial_sort (Range&& rng, RandomIt&& it, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "partial_sort requires RandomAccessIterators");
-  return ::std::partial_sort(
-    ::std::begin(range),
-    ::core::forward<RandomIt>(it),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class IRange, class RRange>
-auto partial_sort_copy (IRange&& irng, RRange&& rrng) -> meta::when<
-  meta::all_of<meta::list<IRange, RRange>, is_range>(),
-  decltype(::std::begin(::core::forward<RRange>(rrng)))
-> {
-  auto irange = make_range(::core::forward<IRange>(irng));
-  auto rrange = make_range(::core::forward<RRange>(rrng));
-  static constexpr auto is_input = decltype(irange)::is_input;
-  static constexpr auto is_random = decltype(rrange)::is_random_access;
-  static_assert(is_input, "partial_sort_copy requires InputIterators");
-  static_assert(is_random, "partial_sort_copy requires RandomAccessIterators");
-  return ::std::partial_sort_copy(
-    ::std::begin(irange),
-    ::std::end(irange),
-    ::std::begin(rrange),
-    ::std::end(rrange)
-  );
-}
-
-template <class IRange, class RRange, class Compare>
-auto partial_sort_copy (
-  IRange&& irng,
-  RRange&& rrng,
-  Compare&& cmp
-) -> meta::when<
-  meta::all_of<meta::list<IRange, RRange>, is_range>(),
-  decltype(::std::begin(::core::forward<RRange>(rrng)))
-> {
-  auto irange = make_range(::core::forward<IRange>(irng));
-  auto rrange = make_range(::core::forward<RRange>(rrng));
-  static constexpr auto is_input = decltype(irange)::is_input;
-  static constexpr auto is_random = decltype(rrange)::is_random_access;
-  static_assert(is_input, "partial_sort_copy requires InputIterators");
-  static_assert(is_random, "partial_sort_copy requires RandomAccessIterators");
-  return ::std::partial_sort_copy(
-    ::std::begin(irange),
-    ::std::end(irange),
-    ::std::begin(rrange),
-    ::std::end(rrange),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto stable_sort (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "stable_sort requires RandomAccessIterators");
-  return ::std::stable_sort(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto stable_sort (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "stable_sort requires RandomAccessIterators");
-  return ::std::stable_sort(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class RandomIt>
-auto nth_element (Range&& rng, RandomIt&& it) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "nth_element requires RandomAccessIterators");
-  return ::std::nth_element(
-    ::std::begin(range),
-    ::core::forward<RandomIt>(it),
-    ::std::end(range)
-  );
-}
-
-template <class Range, class RandomIt, class Compare>
-auto nth_element (Range&& rng, RandomIt&& it, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "nth_element requires RandomAccessIterators");
-  return ::std::nth_element(
-    ::std::begin(range),
-    ::core::forward<RandomIt>(it),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-/* binary search operations (on sorted ranges) */
-template <class Range, class T>
-auto lower_bound (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "lower_bound requires ForwardIterators");
-  return ::std::lower_bound(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class T, class Compare>
-auto lower_bound (Range&& rng, T const& value, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "lower_bound requires ForwardIterators");
-  return ::std::lower_bound(
-    ::std::begin(range),
-    ::std::end(range),
-    value,
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class T>
-auto upper_bound (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "upper_bound requires ForwardIterators");
-  return ::std::upper_bound(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class T, class Compare>
-auto upper_bound (Range&& rng, T const& value, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "upper_bound requires ForwardIterators");
-  return ::std::upper_bound(
-    ::std::begin(range),
-    ::std::end(range),
-    value,
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class T>
-auto binary_search (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "binary_search requires ForwardIterators");
-  return ::std::binary_search(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class T, class Compare>
-auto binary_search (Range&& rng, T const& value, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "binary_search requires ForwardIterators");
-  return ::std::binary_search(
-    ::std::begin(range),
-    ::std::end(range),
-    value,
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class T>
-auto equal_range (Range&& rng, T const& value) -> enable_if_t<
-  is_range<Range>::value,
-  range<decltype(::std::begin(::core::forward<Range>(rng)))>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "equal_range requires ForwardIterators");
-  return ::std::equal_range(::std::begin(range), ::std::end(range), value);
-}
-
-template <class Range, class T, class Compare>
-auto equal_range (Range&& rng, T const& value, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  range<decltype(::std::begin(::core::forward<Range>(rng)))>
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "equal_range requires ForwardIterators");
-  return ::std::equal_range(
-    ::std::begin(range),
-    ::std::end(range),
-    value,
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-/* set operations (on sorted ranges) */
-template <class Range1, class Range2, class OutputIt>
-auto merge (Range1&& rng1, Range2&& rng2, OutputIt&& it) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "merge requires InputIterators");
-  static_assert(is_input2, "merge requires InputIterators");
-  return ::std::merge(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt, class Compare>
-auto merge (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  Compare&& cmp
-) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "merge requires InputIterators");
-  static_assert(is_input2, "merge requires InputIterators");
-  return ::std::merge(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range, class BidirIt>
-auto inplace_merge (Range&& rng, BidirIt&& it) -> meta::when<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "inplace_merge requires BidirectionalIterators");
-  return ::std::inplace_merge(
-    ::std::begin(range),
-    ::core::forward<BidirIt>(it),
-    ::std::end(range)
-  );
-}
-
-template <class Range, class BidirIt, class Compare>
-auto inplace_merge (Range&& rng, BidirIt&& it, Compare&& cmp) -> meta::when<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "inplace_merge requires BidirectionalIterators");
-  return ::std::inplace_merge(
-    ::std::begin(range),
-    ::core::forward<BidirIt>(it),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2>
-auto includes (Range1&& rng1, Range2&& rng2) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "includes requires InputIterators");
-  static_assert(is_input2, "includes requires InputIterators");
-  return ::std::includes(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2)
-  );
-}
-
-template <class Range1, class Range2, class Compare>
-auto includes (Range1&& rng1, Range2&& rng2, Compare&& cmp) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "includes requires InputIterators");
-  static_assert(is_input2, "includes requires InputIterators");
-  return ::std::includes(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt>
-auto set_difference (Range1&& rng1, Range2&& rng2, OutputIt&& it) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_difference requires InputIterators");
-  static_assert(is_input2, "set_difference requires InputIterators");
-  return ::std::set_difference(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt, class Compare>
-auto set_difference (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  Compare&& cmp
-) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_difference requires InputIterators");
-  static_assert(is_input2, "set_difference requires InputIterators");
-  return ::std::set_difference(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt>
-auto set_intersection (Range1&& rng1, Range2&& rng2, OutputIt&& it) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_intersection requires InputIterators");
-  static_assert(is_input2, "set_intersection requires InputIterators");
-  return ::std::set_intersection(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt, class Compare>
-auto set_intersection (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  Compare&& cmp
-) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_intersection requires InputIterators");
-  static_assert(is_input2, "set_intersection requires InputIterators");
-  return ::std::set_intersection(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt>
-auto set_symmetric_difference (Range1&& rng1, Range2&& rng2, OutputIt&& it) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_symmetric_difference requires InputIterators");
-  static_assert(is_input2, "set_symmetric_difference requires InputIterators");
-  return ::std::set_symmetric_difference(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt, class Compare>
-auto set_symmetric_difference (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  Compare&& cmp
-) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_symmetric_difference requires InputIterators");
-  static_assert(is_input2, "set_symmetric_difference requires InputIterators");
-  return ::std::set_symmetric_difference(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt>
-auto set_union (Range1&& rng1, Range2&& rng2, OutputIt&& it) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_union requires InputIterators");
-  static_assert(is_input2, "set_union requires InputIterators");
-  return ::std::set_union(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it)
-  );
-}
-
-template <class Range1, class Range2, class OutputIt, class Compare>
-auto set_union (
-  Range1&& rng1,
-  Range2&& rng2,
-  OutputIt&& it,
-  Compare&& cmp
-) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  decay_t<OutputIt>
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "set_union requires InputIterators");
-  static_assert(is_input2, "set_union requires InputIterators");
-  return ::std::set_union(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<OutputIt>(it),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-/* heap operations */
-template <class Range>
-auto is_heap (Range&& rng) -> enable_if_t<is_range<Range>::value, bool> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "is_heap requires RandomAccessIterators");
-  return ::std::is_heap(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto is_heap (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "is_heap requires RandomAccessIterators");
-  return ::std::is_heap(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto is_heap_until (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "is_heap_until requires RandomAccessIterators");
-  return ::std::is_heap_until(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto is_heap_until (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "is_heap_until requires RandomAccessIterators");
-  return ::std::is_heap_until(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto make_heap (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "make_heap requires RandomAccessIterators");
-  return ::std::make_heap(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto make_heap (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "make_heap requires RandomAccessIterators");
-  return ::std::make_heap(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto push_heap (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "push_heap requires RandomAccessIterators");
-  return ::std::push_heap(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto push_heap (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "push_heap requires RandomAccessIterators");
-  return ::std::push_heap(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto pop_heap (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "pop_heap requires RandomAccessIterators");
-  return ::std::pop_heap(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto pop_heap (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "pop_heap requires RandomAccessIterators");
-  return ::std::pop_heap(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto sort_heap (Range&& rng) -> enable_if_t<is_range<Range>::value> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "sort_heap requires RandomAccessIterators");
-  return ::std::sort_heap(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto sort_heap (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_random = decltype(range)::is_random_access;
-  static_assert(is_random, "sort_heap requires RandomAccessIterators");
-  return ::std::sort_heap(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-/* min/max operations */
-template <class Range>
-auto max_element (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "max_element requires ForwardIterators");
-  return ::std::max_element(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto max_element (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "max_element requires ForwardIterators");
-  return ::std::max_element(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto min_element (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "min_element requires ForwardIterators");
-  return ::std::min_element(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto min_element (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  decltype(::std::begin(::core::forward<Range>(rng)))
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "min_element requires ForwardIterators");
-  return ::std::min_element(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto minmax_element (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  ::std::pair<
-    decltype(::std::begin(::core::forward<Range>(rng))),
-    decltype(::std::end(::core::forward<Range>(rng)))
-  >
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "minmax_element requires ForwardIterators");
-  return ::std::minmax_element(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto minmax_element (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  ::std::pair<
-    range<decltype(::std::begin(::core::forward<Range>(rng)))>,
-    range<decltype(::std::end(::core::forward<Range>(rng)))>
-  >
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_forward = decltype(range)::is_forward;
-  static_assert(is_forward, "minmax_element requires ForwardIterators");
-  return ::std::minmax_element(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2>
-auto lexicographical_compare (Range1&& rng1, Range2&& rng2) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "lexicographical_compare requires InputIterators");
-  static_assert(is_input2, "lexicographical_compare requires InputIterators");
-  return ::std::lexicographical_compare(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2)
-  );
-}
-
-template <class Range1, class Range2, class Compare>
-auto lexicographical_compare (
-  Range1&& rng1,
-  Range2&& rng2,
-  Compare&& cmp
-) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_input1 = decltype(range1)::is_input;
-  static constexpr auto is_input2 = decltype(range2)::is_input;
-  static_assert(is_input1, "lexicographical_compare requires InputIterators");
-  static_assert(is_input2, "lexicographical_compare requires InputIterators");
-  return ::std::lexicographical_compare(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::std::end(range2),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range1, class Range2>
-auto is_permutation (Range1&& rng1, Range2&& rng2) -> enable_if_t<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "is_permutation requires ForwardIterators");
-  static_assert(is_forward2, "is_permutation requires ForwardIterators");
-  return ::std::is_permutation(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2)
-  );
-}
-
-template <class Range1, class Range2, class BinaryPredicate>
-auto is_permutation (
-  Range1&& rng1,
-  Range2&& rng2,
-  BinaryPredicate&& bp
-) -> meta::when<
-  meta::all_of<meta::list<Range1, Range2>, is_range>(),
-  bool
-> {
-  auto range1 = make_range(::core::forward<Range1>(rng1));
-  auto range2 = make_range(::core::forward<Range2>(rng2));
-  static constexpr auto is_forward1 = decltype(range1)::is_forward;
-  static constexpr auto is_forward2 = decltype(range2)::is_forward;
-  static_assert(is_forward1, "is_permutation requires ForwardIterators");
-  static_assert(is_forward2, "is_permutation requires ForwardIterators");
-  return ::std::is_permutation(
-    ::std::begin(range1),
-    ::std::end(range1),
-    ::std::begin(range2),
-    ::core::forward<BinaryPredicate>(bp)
-  );
-}
-
-template <class Range>
-auto next_permutation (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "next_permutation requires BidirectionalIterators");
-  return ::std::next_permutation(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto next_permutation (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "next_permutation requires BidirectionalIterators");
-  return ::std::next_permutation(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-template <class Range>
-auto prev_permutation (Range&& rng) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "prev_permutation requires BidirectionalIterators");
-  return ::std::prev_permutation(::std::begin(range), ::std::end(range));
-}
-
-template <class Range, class Compare>
-auto prev_permutation (Range&& rng, Compare&& cmp) -> enable_if_t<
-  is_range<Range>::value,
-  bool
-> {
-  auto range = make_range(::core::forward<Range>(rng));
-  static constexpr auto is_bidir = decltype(range)::is_bidirectional;
-  static_assert(is_bidir, "prev_permutation requires BidirectionalIterators");
-  return ::std::prev_permutation(
-    ::std::begin(range),
-    ::std::end(range),
-    ::core::forward<Compare>(cmp)
-  );
-}
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_ALGORITHM_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/any.hpp 4.4.0-1/src/mlpack/core/std_backport/any.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/any.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/any.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,337 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_ANY_HPP
-#define CORE_ANY_HPP
-
-#include <memory>
-
-#include <cstdlib>
-#include <cstring>
-
-#include "type_traits.hpp"
-#include "algorithm.hpp"
-#include "typeinfo.hpp"
-#include "utility.hpp"
-
-#ifndef CORE_NO_EXCEPTIONS
-#include <stdexcept>
-#endif /* CORE_NO_EXCEPTIONS */
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-using data_type = add_pointer_t<void>;
-
-template <class T>
-struct is_small final : meta::all_t<
-  sizeof(decay_t<T>) <= sizeof(data_type),
-  alignof(decay_t<T>) <= alignof(data_type),
-  ::std::is_nothrow_copy_constructible<decay_t<T>>::value
-> { };
-template <> struct is_small<void> final : ::std::true_type { };
-
-template <class T=void, bool=is_small<T>::value> struct dispatch;
-template <> struct dispatch<void, true> {
-  dispatch () noexcept = default;
-  virtual ~dispatch () noexcept = default;
-
-  virtual void clone (data_type const&, data_type&) const { }
-  virtual void move (data_type&, data_type&) const noexcept { }
-  virtual void destroy (data_type&) const noexcept { }
-  virtual type_info const& type () const noexcept { return type_of<void>(); }
-};
-
-template <class T>
-struct dispatch<T, true> final : dispatch<> {
-  using value_type = T;
-  using const_pointer = add_pointer_t<add_const_t<value_type>>;
-  using pointer = add_pointer_t<value_type>;
-  using allocator_type = ::std::allocator<value_type>;
-  using allocator_traits = ::std::allocator_traits<allocator_type>;
-
-  virtual void clone (data_type const& src, data_type& dst) const final {
-    allocator_type alloc { };
-    auto val = reinterpret_cast<add_const_t<const_pointer>>(&src);
-    auto ptr = reinterpret_cast<pointer>(&dst);
-    allocator_traits::construct(alloc, ptr, *val);
-  }
-
-  virtual void move (data_type& src, data_type& dst) const noexcept final {
-    allocator_type alloc { };
-    auto val = reinterpret_cast<pointer>(&src);
-    auto ptr = reinterpret_cast<pointer>(&dst);
-    allocator_traits::construct(alloc, ptr, ::core::move(*val));
-  }
-
-  virtual void destroy (data_type& src) const noexcept final {
-    allocator_type alloc { };
-    auto ptr = reinterpret_cast<pointer>(&src);
-    allocator_traits::destroy(alloc, ptr);
-  }
-
-  virtual type_info const& type () const noexcept final {
-    return type_of<value_type>();
-  }
-};
-
-template <class T>
-struct dispatch<T, false> final : dispatch<> {
-  using value_type = T;
-  using pointer = add_pointer_t<value_type>;
-  using allocator_type = ::std::allocator<value_type>;
-  using allocator_traits = ::std::allocator_traits<allocator_type>;
-
-  virtual void clone (data_type const& src, data_type& dst) const final {
-    allocator_type alloc { };
-    auto const& value = *static_cast<add_const_t<pointer>>(src);
-    auto ptr = allocator_traits::allocate(alloc, 1);
-    auto scope = make_scope_guard([&alloc, ptr] {
-      allocator_traits::deallocate(alloc, ptr, 1);
-    });
-    allocator_traits::construct(alloc, ptr, value);
-    scope.dismiss();
-    dst = ptr;
-  }
-
-  virtual void move (data_type& src, data_type& dst) const noexcept final {
-    allocator_type alloc { };
-    auto& value = *static_cast<pointer>(src);
-    auto ptr = allocator_traits::allocate(alloc, 1);
-    auto scope = make_scope_guard([&alloc, ptr] {
-      allocator_traits::deallocate(alloc, ptr, 1);
-    });
-    allocator_traits::construct(alloc, ptr, ::core::move(value));
-    scope.dismiss();
-    dst = ptr;
-  }
-
-  virtual void destroy (data_type& src) const noexcept final {
-    allocator_type alloc { };
-    auto ptr = static_cast<pointer>(src);
-    allocator_traits::destroy(alloc, ptr);
-    allocator_traits::deallocate(alloc, ptr, 1);
-  }
-
-  virtual type_info const& type () const noexcept final {
-    return type_of<value_type>();
-  }
-};
-
-template <class T> dispatch<> const* lookup () noexcept {
-  static dispatch<T> instance;
-  return ::std::addressof(instance);
-}
-
-template <> inline dispatch<> const* lookup<void> () noexcept {
-  static dispatch<> instance;
-  return ::std::addressof(instance);
-}
-
-} /* namespace impl */
-
-#ifndef CORE_NO_EXCEPTIONS
-class bad_any_cast final : public ::std::bad_cast {
-public:
-  virtual char const* what () const noexcept override {
-    return "bad any cast";
-  }
-};
-
-[[noreturn]] inline void throw_bad_any_cast () { throw bad_any_cast { }; }
-#else /* CORE_NO_EXCEPTIONS */
-[[noreturn]] inline void throw_bad_any_cast () { ::std::abort(); }
-#endif /* CORE_NO_EXCEPTIONS */
-
-struct any final {
-
-  template <class T> friend T const* any_cast (any const*) noexcept;
-  template <class T> friend T* any_cast (any*) noexcept;
-
-  any (any const& that) :
-    table { that.table },
-    data { nullptr }
-  { this->table->clone(that.data, this->data); }
-
-  any (any&& that) noexcept :
-    table { that.table },
-    data { nullptr }
-  { this->table->move(that.data, this->data); }
-
-  any () noexcept :
-    table { impl::lookup<void>() },
-    data { nullptr }
-  { }
-
-  template <
-    class T,
-    class=enable_if_t<not ::std::is_same<any, decay_t<T>>::value>
-  > any (T&& value) :
-    any { ::std::forward<T>(value), impl::is_small<T> { } }
-  { }
-
-  ~any () noexcept { this->clear(); }
-
-  any& operator = (any const& that) {
-    any { that }.swap(*this);
-    return *this;
-  }
-
-  any& operator = (any&& that) noexcept {
-    any { ::std::move(that) }.swap(*this);
-    return *this;
-  }
-
-  template <
-    class T,
-    class=enable_if_t<not ::std::is_same<any, decay_t<T>>::value>
-  > any& operator = (T&& value) {
-    any {
-      ::std::forward<T>(value),
-      impl::is_small<T> { }
-    }.swap(*this);
-    return *this;
-  }
-
-  void swap (any& that) noexcept {
-    using ::std::swap;
-    swap(this->table, that.table);
-    swap(this->data, that.data);
-  }
-
-  void clear () noexcept {
-    this->table->destroy(this->data);
-    this->table = impl::lookup<void>();
-  }
-
-  type_info const& type () const noexcept { return this->table->type(); }
-
-  bool empty () const noexcept { return this->table == impl::lookup<void>(); }
-
-private:
-  impl::dispatch<> const* table;
-  impl::data_type data;
-
-  template <class T>
-  any (T&& value, ::std::true_type&&) :
-    table { impl::lookup<decay_t<T>>() },
-    data { nullptr }
-  {
-    using value_type = decay_t<T>;
-    using allocator_type = ::std::allocator<value_type>;
-    using allocator_traits = ::std::allocator_traits<allocator_type>;
-    allocator_type alloc { };
-    auto pointer = reinterpret_cast<value_type*>(::std::addressof(this->data));
-    allocator_traits::construct(alloc, pointer, ::core::forward<T>(value));
-  }
-
-  template <class T>
-  any (T&& value, ::std::false_type&&) :
-    table { impl::lookup<decay_t<T>>() },
-    data { nullptr }
-  {
-    using value_type = decay_t<T>;
-    using allocator_type = ::std::allocator<value_type>;
-    using allocator_traits = ::std::allocator_traits<allocator_type>;
-    allocator_type alloc { };
-    auto pointer = allocator_traits::allocate(alloc, 1);
-    allocator_traits::construct(alloc, pointer, ::core::forward<T>(value));
-    this->data = pointer;
-  }
-
-  template <class T>
-  T const* cast (::std::true_type&&) const {
-    return reinterpret_cast<T const*>(::std::addressof(this->data));
-  }
-
-  template <class T>
-  T* cast (::std::true_type&&) {
-    return reinterpret_cast<T*>(::std::addressof(this->data));
-  }
-
-  template <class T>
-  T const* cast (::std::false_type&&) const {
-    return static_cast<T const*>(this->data);
-  }
-
-  template <class T>
-  T* cast (::std::false_type&&) {
-    return static_cast<T*>(this->data);
-  }
-};
-
-template <class T>
-T const* any_cast (any const* operand) noexcept {
-  return operand and operand->type() == type_of<T>()
-    ? operand->cast<T>(impl::is_small<T> { })
-    : nullptr;
-}
-
-template <class T>
-T* any_cast (any* operand) noexcept {
-  return operand and operand->type() == type_of<T>()
-    ? operand->cast<T>(impl::is_small<T> { })
-    : nullptr;
-}
-
-template <
-  class T,
-  class=meta::when<
-    meta::any<
-      ::std::is_reference<T>::value,
-      ::std::is_copy_constructible<T>::value
-    >()
-  >
-> T any_cast (any const& operand) {
-  using type = remove_reference_t<T>;
-  auto pointer = any_cast<add_const_t<type>>(::std::addressof(operand));
-  if (not pointer) { throw_bad_any_cast(); }
-  return *pointer;
-}
-
-template <
-  class T,
-  class=meta::when<
-    meta::any<
-      ::std::is_reference<T>::value,
-      ::std::is_copy_constructible<T>::value
-    >()
-  >
-> T any_cast (any&& operand) {
-  using type = remove_reference_t<T>;
-  auto pointer = any_cast<type>(::std::addressof(operand));
-  if (not pointer) { throw_bad_any_cast(); }
-  return *pointer;
-}
-
-template <
-  class T,
-  class=meta::when<
-    meta::any<
-      ::std::is_reference<T>::value,
-      ::std::is_copy_constructible<T>::value
-    >()
-  >
-> T any_cast (any& operand) {
-  using type = remove_reference_t<T>;
-  auto pointer = any_cast<type>(::std::addressof(operand));
-  if (not pointer) { throw_bad_any_cast(); }
-  return *pointer;
-}
-
-inline void swap (any& lhs, any& rhs) noexcept { lhs.swap(rhs); }
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_ANY_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/functional.hpp 4.4.0-1/src/mlpack/core/std_backport/functional.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/functional.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/functional.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,435 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_FUNCTIONAL_HPP
-#define CORE_FUNCTIONAL_HPP
-
-#include <functional>
-#include <tuple>
-#include <array>
-
-#include "type_traits.hpp"
-#include "utility.hpp"
-
-namespace core {
-inline namespace v2 {
-
-template <class T> using is_reference_wrapper = meta::is_specialization_of<
-  remove_cv_t<T>,
-  ::std::reference_wrapper
->;
-
-template <class F> struct function_traits;
-
-template <class R, class... Args>
-struct function_traits<R(*)(Args...)> : function_traits<R(Args...)> { };
-
-template <class C, class R>
-struct function_traits<R(C::*)> : function_traits<R(C&)> { };
-
-template <class C, class R, class... Args>
-struct function_traits<R(C::*)(Args...)> : function_traits<R(C&, Args...)> { };
-
-template <class C, class R, class... Args>
-struct function_traits<R(C::*)(Args...) const volatile> :
-  function_traits<R(C volatile const&, Args...)>
-{ };
-
-template <class C, class R, class... Args>
-struct function_traits<R(C::*)(Args...) volatile> :
-  function_traits<R(C volatile&, Args...)>
-{ };
-
-template <class C, class R, class... Args>
-struct function_traits<R(C::*)(Args...) const> :
-  function_traits<R(C const&, Args...)>
-{ };
-
-template <class R, class... Args>
-struct function_traits<R(Args...)> {
-  using typelist = meta::list<Args...>;
-  using return_type = R;
-
-  using pointer = add_pointer_t<return_type(Args...)>;
-  static constexpr auto arity = typelist::size();
-
-  template <::std::size_t N> using argument = meta::get<typelist, N>;
-};
-
-template <class F> struct function_traits {
-  using functor_type = function_traits<decltype(&decay_t<F>::operator())>;
-  using return_type = typename functor_type::return_type;
-  using pointer = typename functor_type::pointer;
-  static constexpr auto arity = functor_type::arity - 1;
-  template <::std::size_t N>
-  using argument = typename functor_type::template argument<N>;
-};
-
-/* N3727 */
-template <class Functor, class... Args>
-auto invoke (Functor&& f, Args&&... args) -> enable_if_t<
-  ::std::is_member_pointer<decay_t<Functor>>::value,
-  result_of_t<Functor&&(Args&&...)>
-> { return ::std::mem_fn(f)(core::forward<Args>(args)...); }
-
-template <class Functor, class... Args>
-auto invoke (Functor&& f, Args&&... args) -> enable_if_t<
-  not ::std::is_member_pointer<decay_t<Functor>>::value,
-  result_of_t<Functor&&(Args&&...)>
-> { return core::forward<Functor>(f)(core::forward<Args>(args)...); }
-
-template <class F, class T, ::std::size_t... I>
-auto apply (F&& f, T&& t, index_sequence<I...>) -> decltype(
-  invoke(core::forward<F>(f), ::std::get<I>(core::forward<T>(t))...)
-) { return invoke(core::forward<F>(f), ::std::get<I>(core::forward<T>(t))...); }
-
-template <
-  class Functor,
-  class T,
-  class I = make_index_sequence<::std::tuple_size<decay_t<T>>::value>
-> auto apply (Functor&& f, T&& t) -> decltype(
-  apply(core::forward<Functor>(f), core::forward<T>(t), I { })
-) { return apply(core::forward<Functor>(f), core::forward<T>(t), I { }); }
-
-template <class F>
-struct apply_functor {
-  template <class G>
-  explicit apply_functor (G&& g) : f(core::forward<G>(g)) { }
-
-  template <class Applicable>
-  auto operator () (Applicable&& args) -> decltype(
-    core::apply(core::forward<F>(this->f), core::forward<Applicable>(args))
-  ) { return apply(core::forward<F>(f), core::forward<Applicable>(args)); }
-private:
-  F f;
-};
-
-template <class F>
-auto make_apply (F&& f) -> apply_functor<F> {
-  return apply_functor<F> { core::forward<F>(f) };
-}
-
-template <class F>
-struct not_fn_functor {
-  template <class G>
-  explicit not_fn_functor (G&& g) : f(core::forward<G>(g)) { }
-
-  template <class... Args>
-  auto operator () (Args&&... args) const -> decltype(
-    not (invoke)(::std::declval<F>(), core::forward<Args>(args)...)
-  ) { return not (invoke)(f, core::forward<Args>(args)...); }
-
-  template <class... Args>
-  auto operator () (Args&&... args) -> decltype(
-    not (invoke)(::std::declval<F>(), core::forward<Args>(args)...)
-  ) { return not (invoke)(f, core::forward<Args>(args)...); }
-
-private:
-  F f;
-};
-
-/* Were this C++14, we could just use a lambda with a capture. Oh Well! */
-template <class F>
-not_fn_functor<decay_t<F>> not_fn (F&& f) {
-  return not_fn_functor<decay_t<F>> { core::forward<F>(f) };
-}
-
-/* converter function object */
-template <class T>
-struct converter {
-  template <class... Args>
-  constexpr T operator () (Args&&... args) const {
-    return T(core::forward<Args>(args)...);
-  }
-};
-
-/* function objects -- arithmetic */
-template <class T=void>
-struct plus {
-  constexpr T operator () (T const& l, T const& r) const { return l + r; }
-};
-
-template <class T=void>
-struct minus {
-  constexpr T operator () (T const& l, T const& r) const { return l - r; }
-};
-
-template <class T=void>
-struct multiplies {
-  constexpr T operator () (T const& l, T const& r) const { return l * r; }
-};
-
-template <class T=void>
-struct divides {
-  constexpr T operator () (T const& l, T const& r) const { return l / r; }
-};
-
-template <class T=void>
-struct modulus {
-  constexpr T operator () (T const& l, T const& r) const { return l % r; }
-};
-
-template <class T=void>
-struct negate {
-  constexpr T operator () (T const& arg) const { return -arg; }
-};
-
-/* function objects -- comparisons */
-template <class T=void>
-struct equal_to {
-  constexpr bool operator () (T const& l, T const& r) const { return l == r; }
-};
-
-template <class T=void>
-struct not_equal_to {
-  constexpr bool operator () (T const& l, T const& r) const { return l != r; }
-};
-
-template <class T=void>
-struct greater_equal {
-  constexpr bool operator () (T const& l, T const& r) const { return l >= r; }
-};
-
-template <class T=void>
-struct less_equal {
-  constexpr bool operator () (T const& l, T const& r) const { return l <= r; }
-};
-
-template <class T=void>
-struct greater {
-  constexpr bool operator () (T const& l, T const& r) const { return l > r; }
-};
-
-template <class T=void>
-struct less {
-  constexpr bool operator () (T const& l, T const& r) const { return l < r; }
-};
-
-/* function objects -- logical */
-template <class T=void>
-struct logical_and {
-  constexpr bool operator () (T const& l, T const& r) const { return l and r; }
-};
-
-template <class T=void>
-struct logical_or {
-  constexpr bool operator () (T const& l, T const& r) const { return l or r; }
-};
-
-template <class T=void>
-struct logical_not {
-  constexpr bool operator () (T const& arg) const { return not arg; }
-};
-
-/* function objects -- bitwise */
-
-template <class T=void>
-struct bit_and {
-  constexpr bool operator () (T const& l, T const& r) const { return l & r; }
-};
-
-template <class T=void>
-struct bit_or {
-  constexpr bool operator () (T const& l, T const& r) const { return l | r; }
-};
-
-template <class T=void>
-struct bit_xor {
-  constexpr bool operator () (T const& l, T const& r) const { return l ^ r; }
-};
-
-template <class T=void>
-struct bit_not {
-  constexpr bool operator () (T const& arg) const { return ~arg; }
-};
-
-/* function objects -- arithmetic specializations */
-template <> struct plus<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) + core::forward<U>(u)
-  ) { return core::forward<T>(t) + core::forward<U>(u); }
-};
-
-template <> struct minus<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) - core::forward<U>(u)
-  ) { return core::forward<T>(t) - core::forward<U>(u); }
-};
-
-template <> struct multiplies<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) * core::forward<U>(u)
-  ) { return core::forward<T>(t) * core::forward<U>(u); }
-};
-
-template <> struct divides<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) / core::forward<U>(u)
-  ) { return core::forward<T>(t) / core::forward<U>(u); }
-};
-
-template <> struct modulus<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) % core::forward<U>(u)
-  ) { return core::forward<T>(t) % core::forward<U>(u); }
-};
-
-template <> struct negate<void> {
-  using is_transparent = void;
-
-  template <class T>
-  constexpr auto operator () (T&& t) const -> decltype(core::forward<T>(t)) {
-    return core::forward<T>(t);
-  }
-};
-
-/* function objects -- comparison specialization */
-template <> struct equal_to<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) == core::forward<U>(u)
-  ) { return core::forward<T>(t) == core::forward<U>(u); }
-};
-
-template <> struct not_equal_to<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) != core::forward<U>(u)
-  ) { return core::forward<T>(t) != core::forward<U>(u); }
-};
-
-template <> struct greater_equal<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) >= core::forward<U>(u)
-  ) { return core::forward<T>(t) >= core::forward<U>(u); }
-};
-
-template <> struct less_equal<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) <= core::forward<U>(u)
-  ) { return core::forward<T>(t) <= core::forward<U>(u); }
-};
-
-template <> struct greater<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) > core::forward<U>(u)
-  ) { return core::forward<T>(t) > core::forward<U>(u); }
-};
-
-template <> struct less<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) < core::forward<U>(u)
-  ) { return core::forward<T>(t) < core::forward<U>(u); }
-};
-
-/* function objects -- logical specializations */
-template <> struct logical_and<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) and core::forward<U>(u)
-  ) { return core::forward<T>(t) and core::forward<U>(u); }
-};
-
-template <> struct logical_or<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) or core::forward<U>(u)
-  ) { return core::forward<T>(t) or core::forward<U>(u); }
-};
-
-template <> struct logical_not<void> {
-  using is_transparent = void;
-
-  template <class T>
-  constexpr auto operator () (T&& t) const -> decltype(
-    not core::forward<T>(t)
-  ) { return not core::forward<T>(t); }
-};
-
-/* function objects -- bitwise specializations */
-template <> struct bit_and<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) & core::forward<U>(u)
-  ) { return core::forward<T>(t) & core::forward<U>(u); }
-};
-
-template <> struct bit_or<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) | core::forward<U>(u)
-  ) { return core::forward<T>(t) | core::forward<U>(u); }
-};
-
-template <> struct bit_xor<void> {
-  using is_transparent = void;
-
-  template <class T, class U>
-  constexpr auto operator () (T&& t, U&& u) const -> decltype(
-    core::forward<T>(t) ^ core::forward<U>(u)
-  ) { return core::forward<T>(t) ^ core::forward<U>(u); }
-};
-
-template <> struct bit_not<void> {
-  using is_transparent = void;
-
-  template <class T>
-  constexpr auto operator () (T&& t) const -> decltype(~core::forward<T>(t)) {
-    return ~core::forward<T>(t);
-  }
-};
-
-/* N3980 Implementation */
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_FUNCTIONAL_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/internal.hpp 4.4.0-1/src/mlpack/core/std_backport/internal.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/internal.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/internal.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,156 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_INTERNAL_HPP
-#define CORE_INTERNAL_HPP
-
-/* This is a header containing common implementation specific code, to
- * reduce the complexity of the other headers, especially those that are
- * closely intertwined, such as <core/functional.hpp> and <core/type_traits.hpp>
- *
- * Additionally, some of this code is duplicated elsewhere (such as class_of,
- * and meta::identity), but aliases are placed to lessen any impact that this
- * might have.
- */
-
-#include <type_traits>
-#include <functional>
-#include <utility>
-#include <ciso646>
-
-#include "meta.hpp"
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-template <class T, class Void, template <class...> class, class...>
-struct make_detect : meta::identity<T> { using value_t = ::std::false_type; };
-template <class T, template <class...> class U, class... Args>
-struct make_detect<T, meta::deduce<U<Args...>>, U, Args...> :
-  meta::identity<U<Args...>>
-{ using value_t = ::std::true_type; };
-
-/* extremely useful custom type traits */
-template <class T> struct class_of : meta::identity<T> { };
-template <class Signature, class T>
-struct class_of<Signature T::*> : meta::identity<T> { };
-
-/* aliases */
-template <class T> using class_of_t = typename class_of<T>::type;
-template <class T> using decay_t = typename ::std::decay<T>::type;
-template <class T>
-using remove_reference_t = typename ::std::remove_reference<T>::type;
-template <bool B, class T = void>
-using enable_if_t = typename ::std::enable_if<B, T>::type;
-
-/* is_nothrow_swappable plumbing */
-using ::std::declval;
-using ::std::swap;
-
-// MSVC 2015 workaround
-template <class T, class U>
-struct is_swappable_with {
-  template <class X, class Y>
-  static auto test (void*) noexcept(true) -> decltype(
-    swap(declval<X&>(), declval<Y&>())
-  );
-
-  template <class, class>
-  static void test (...) noexcept(false);
-
-  static constexpr bool value = noexcept(test<T, U>(nullptr));
-};
-
-// MSVC 2015 workaround
-template <class T, class U>
-struct is_noexcept_swappable_with {
-  template <
-    class X,
-    class Y,
-    bool B=noexcept(swap(declval<X&>(), declval<Y&>()))
-  > static void test (enable_if_t<B>*) noexcept(true);
-
-  template <class, class>
-  static void test (...) noexcept(false);
-
-  static constexpr bool value = noexcept(test<T, U>(nullptr));
-};
-
-template <class, class, class=void>
-struct is_swappable : ::std::false_type { };
-
-template <class T, class U>
-struct is_swappable<
-  T,
-  U,
-  meta::deduce<
-    is_swappable_with<T, U>,
-    is_swappable_with<U, T>
-  >
-> : ::std::true_type { };
-
-template <class T, class U=T>
-struct is_nothrow_swappable : meta::all_t<
-  is_swappable<T, U>::value,
-  is_noexcept_swappable_with<T, U>::value,
-  is_noexcept_swappable_with<U, T>::value
-> { };
-
-/*
- * If I can't amuse myself when working with C++ templates, then life isn't
- * worth living. Bury me with my chevrons.
- */
-template <class T>
-constexpr T&& pass (remove_reference_t<T>& t) noexcept {
-  return static_cast<T&&>(t);
-}
-
-template <class T>
-constexpr T&& pass (remove_reference_t<T>&& t) noexcept {
-  return static_cast<T&&>(t);
-}
-
-/* INVOKE pseudo-expression plumbing, *much* more simplified than previous
- * versions of Core
- */
-struct undefined { constexpr undefined (...) noexcept { } };
-
-/* We get some weird warnings under clang, so we actually give these functions
- * a body to get rid of it.
- */
-template <class... Args>
-constexpr undefined INVOKE (undefined, Args&&...) noexcept {
-  return undefined { };
-}
-
-template <class Functor, class... Args>
-constexpr auto INVOKE (Functor&& f, Args&&... args) -> enable_if_t<
-  not ::std::is_member_pointer<decay_t<Functor>>::value,
-  decltype(pass<Functor>(f)(pass<Args>(args)...))
-> { return pass<Functor>(f)(pass<Args>(args)...); }
-
-template <class Functor, class... Args>
-auto INVOKE (Functor&& f, Args&&... args) -> enable_if_t<
-  ::std::is_member_pointer<decay_t<Functor>>::value,
-  decltype(::std::mem_fn(pass<Functor>(f))(pass<Args>(args)...))
-> { return ::std::mem_fn(pass<Functor>(f))(pass<Args>(args)...); }
-
-template <bool, class...> struct invoke_of { };
-template <class... Args> struct invoke_of<true, Args...> :
-  meta::identity<decltype(INVOKE(declval<Args>()...))>
-{ };
-
-}}} /* namespace core::v2::impl */
-
-#endif /* CORE_INTERNAL_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/iterator.hpp 4.4.0-1/src/mlpack/core/std_backport/iterator.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/iterator.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/iterator.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,272 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_ITERATOR_HPP
-#define CORE_ITERATOR_HPP
-
-#include <functional>
-#include <iterator>
-#include <ostream>
-
-#include "type_traits.hpp"
-#include "utility.hpp"
-
-namespace core {
-inline namespace v2 {
-
-/* capacity */
-template <class Container>
-constexpr auto size (Container const& container) noexcept -> decltype(
-  container.size()
-) { return container.size(); }
-
-template <class T, ::std::size_t N>
-constexpr ::std::size_t size (T const (&)[N]) noexcept { return N; }
-
-template <class Container>
-constexpr bool empty (Container const& container) noexcept {
-  return container.empty();
-}
-
-template <class T, std::size_t N>
-constexpr bool empty (T const (&)[N]) noexcept { return false; }
-
-/* element access */
-template <class Container>
-constexpr auto front (Container const& container) -> decltype(
-  container.front()
-) { return container.front(); }
-
-template <class Container>
-constexpr auto front (Container& container) -> decltype(container.front()) {
-  return container.front();
-}
-
-template <class T, ::std::size_t N>
-constexpr T const& front (T const (&array)[N]) noexcept { return array[0]; }
-
-template <class T, ::std::size_t N>
-constexpr T& front (T (&array)[N]) noexcept { return array[0]; }
-
-template <class Container>
-constexpr auto back (Container const& container) -> decltype(
-  container.back()
-) { return container.back(); }
-
-template <class Container>
-constexpr auto back (Container& container) -> decltype(container.back()) {
-  return container.back();
-}
-
-template <class T, ::std::size_t N>
-constexpr T const& back (T const (&array)[N]) noexcept { return array[N - 1]; }
-
-template <class T, ::std::size_t N>
-constexpr T& back (T (&array)[N]) noexcept { return array[N - 1]; }
-
-/* data access */
-template <class Container>
-constexpr auto data (Container const& container) noexcept -> decltype(
-  container.data()
-) { return container.data(); }
-
-template <class Container>
-constexpr auto data (Container& container) noexcept -> decltype(
-  container.data()
-) { return container.data(); }
-
-template <class T, ::std::size_t N>
-constexpr T const* data (T const (&array)[N]) noexcept { return array; }
-
-template <class T, ::std::size_t N>
-constexpr T* data (T (&array)[N]) noexcept { return array; }
-
-/* iteration */
-template <class Container>
-auto cbegin (Container const& container) -> decltype(::std::begin(container)) {
-  return ::std::begin(container);
-}
-
-template <class Container>
-auto cend (Container const& container) -> decltype(::std::end(container)) {
-  return ::std::end(container);
-}
-
-template <class Container>
-auto rbegin (Container const& container) -> decltype(container.rbegin()) {
-  return container.rbegin();
-}
-
-template <class Container>
-auto rbegin (Container& container) -> decltype(container.rbegin()) {
-  return container.rbegin();
-}
-
-template <class Container>
-auto crbegin (Container const& container) -> decltype(rbegin(container)) {
-  return rbegin(container);
-}
-
-template <class Container>
-auto rend (Container const& container) -> decltype(container.rend()) {
-  return container.rend();
-}
-
-template <class Container>
-auto rend (Container& container) -> decltype(container.rend()) {
-  return container.rend();
-}
-
-template <class Container>
-auto crend (Container const& container) -> decltype(rend(container)) {
-  return rend(container);
-}
-
-template <class Iterator>
-::std::reverse_iterator<Iterator> make_reverse_iterator (Iterator iter) {
-  return ::std::reverse_iterator<Iterator>(iter);
-}
-
-template <
-  class DelimT,
-  class CharT=char,
-  class Traits=::std::char_traits<CharT>
-> struct ostream_joiner final : ::std::iterator<
-  ::std::output_iterator_tag,
-  void,
-  void,
-  void,
-  void
-> {
-  using delimiter_type = DelimT;
-  using ostream_type = ::std::basic_ostream<CharT, Traits>;
-  using traits_type = Traits;
-  using char_type = CharT;
-
-  ostream_joiner (ostream_type& stream, delimiter_type const& delimiter) :
-    stream(stream),
-    delimiter { delimiter }
-  { }
-
-  ostream_joiner (ostream_type& stream, delimiter_type&& delimiter) :
-    stream(stream),
-    delimiter { ::core::move(delimiter) },
-    first { true }
-  { }
-
-  template <class T>
-  ostream_joiner& operator = (T const& item) {
-    if (not first and delimiter) { this->stream << delimiter; }
-    this->stream << item;
-    this->first = false;
-    return *this;
-  }
-
-  ostream_joiner& operator ++ (int) noexcept { return *this; }
-  ostream_joiner& operator ++ () noexcept { return *this; }
-  ostream_joiner& operator * () noexcept { return *this; }
-
-private:
-  ostream_type& stream;
-  delimiter_type delimiter;
-  bool first;
-};
-
-template <class T>
-struct number_iterator {
-  using iterator_category = ::std::bidirectional_iterator_tag;
-  using difference_type = T;
-  using value_type = T;
-  using reference = add_lvalue_reference_t<T>;
-  using pointer = add_pointer_t<T>;
-
-  static_assert(::std::is_integral<value_type>::value, "");
-
-  explicit number_iterator (value_type value, value_type step=1) noexcept :
-    value { value },
-    step { step }
-  { }
-
-  number_iterator (number_iterator const&) noexcept = default;
-  number_iterator () noexcept = default;
-  ~number_iterator () noexcept = default;
-
-  number_iterator& operator = (number_iterator const&) noexcept = default;
-
-  void swap (number_iterator& that) noexcept {
-    ::std::swap(this->value, that.value);
-    ::std::swap(this->step, that.step);
-  }
-
-  reference operator * () noexcept { return this->value; }
-
-  number_iterator& operator ++ () noexcept {
-    this->value += this->step;
-    return *this;
-  }
-
-  number_iterator& operator -- () noexcept {
-    this->value -= this->step;
-    return *this;
-  }
-
-  number_iterator operator ++ (int) const noexcept {
-    return number_iterator { this->value + this->step };
-  }
-
-  number_iterator operator -- (int) const noexcept {
-    return number_iterator { this->value - this->step };
-  }
-
-  bool operator == (number_iterator const& that) const noexcept {
-    return this->value == that.value and this->step == that.step;
-  }
-
-  bool operator != (number_iterator const& that) const noexcept {
-    return this->value != that.value and this->step == that.step;
-  }
-
-private:
-  value_type value { };
-  value_type step { static_cast<value_type>(1) };
-};
-
-template <class T>
-void swap (number_iterator<T>& lhs, number_iterator<T>& rhs) noexcept {
-  lhs.swap(rhs);
-}
-
-template <class CharT, class Traits, class DelimT>
-ostream_joiner<decay_t<DelimT>, CharT, Traits> make_ostream_joiner (
-  ::std::basic_ostream<CharT, Traits>& stream,
-  DelimT&& delimiter
-) {
-  return ostream_joiner<decay_t<DelimT>, CharT, Traits> {
-    stream,
-    ::core::forward<DelimT>(delimiter)
-  };
-}
-
-template <class T>
-number_iterator<T> make_number_iterator (T value, T step) noexcept {
-  return number_iterator<T> { value, step };
-}
-
-template <class T>
-number_iterator<T> make_number_iterator (T value) noexcept {
-  return number_iterator<T> { value };
-}
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_ITERATOR_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/meta.hpp 4.4.0-1/src/mlpack/core/std_backport/meta.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/meta.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/meta.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,325 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_META_HPP
-#define CORE_META_HPP
-
-#include <type_traits>
-#include <limits>
-#include <tuple>
-
-#include <cstdint>
-#include <cstddef>
-
-namespace core {
-inline namespace v2 {
-namespace meta {
-namespace impl {
-
-using false_t = ::std::false_type;
-using true_t = ::std::true_type;
-
-template <class T, T V> using integral = ::std::integral_constant<T, V>;
-template <bool B> using boolean = integral<bool, B>;
-
-template <class T> struct identity { using type = T; };
-
-template <class...> struct deducer : identity<void> { };
-template <class... Ts> using deduce = typename deducer<Ts...>::type;
-
-template <class T, class V, template <class...> class Detector, class... Args>
-struct detect : identity<T> { using value = false_t; };
-
-template <class T, template <class...> class Detector, class... Args>
-struct detect<T, deduce<Detector<Args...>>, Detector, Args...> :
-  identity<Detector<Args...>>
-{ using value = true_t; };
-
-template <class... Ts>
-struct list {
-  static constexpr ::std::size_t size () noexcept { return sizeof...(Ts); }
-  static constexpr bool empty () noexcept { return size() == 0u; }
-};
-
-template <class, template <class...> class> struct convert;
-template <class, class> struct rebind;
-template <class> struct into;
-
-template <class, template <class...> class, class...> struct filter;
-template <class, template <class...> class, class...> struct map;
-template <class...> struct join;
-
-template <template <class...> class, class...> struct apply;
-
-template <class, ::std::size_t> struct get;
-template <class, class> struct index_of;
-template <class> struct head;
-template <class> struct tail;
-
-template <class, template <class...> class, class...> struct count_if;
-template <class, template <class...> class, class...> struct find_if;
-template <class, class> struct count;
-template <class, class> struct find;
-
-template <class, ::std::size_t> struct rotate;
-template <class> struct reverse;
-
-template <class, class> struct equal;
-
-template <class, template <class...> class, class...> struct none_of;
-template <class, template <class...> class, class...> struct all_of;
-template <class, template <class...> class, class...> struct any_of;
-
-template <bool...> struct none;
-template <bool...> struct all;
-template <bool...> struct any;
-
-template <class... Ts, template <class...> class To>
-struct convert<list<Ts...>, To> : identity<To<Ts...>> { };
-
-template <template <class...> class To, class... Ts, class... Us>
-struct rebind<list<Ts...>, To<Us...>> : identity<To<Ts...>> { };
-
-template <template <class...> class T, class... Ts>
-struct into<T<Ts...>> : identity<list<Ts...>> { };
-
-template <template <class...> class F, class... Ts, class... Args>
-struct filter<list<Ts...>, F, Args...> :
-  join<
-    typename ::std::conditional<
-      F<Ts, Args...>::value,
-      list<Ts>,
-      list<>
-    >::type...
-  >
-{ };
-
-template <template <class...> class F, class... Ts, class... Args>
-struct map<list<Ts...>, F, Args...> : identity<list<F<Ts, Args...>...>> { };
-
-template <class... Ts> struct join<list<Ts...>> : identity<list<Ts...>> { };
-template <class... Ts, class... Us, class... Vs>
-struct join<list<Ts...>, list<Us...>, Vs...> :
-  join<list<Ts..., Us...>, Vs...>
-{ };
-
-template <class T, class... Ts>
-struct index_of<list<Ts...>, T> : integral<
-  ::std::size_t,
-  list<Ts...>::size() - find<list<Ts...>, T>::type::size()
-> { };
-
-template <class T, class... Ts>
-struct get<list<T, Ts...>, 0> : identity<T> { };
-template <class T, class... Ts, size_t N>
-struct get<list<T, Ts...>, N> :
-  get<list<Ts...>, N - 1>
-{ static_assert(N < (sizeof...(Ts) + 1),""); };
-
-
-template <class T, class... Ts>
-struct head<list<T, Ts...>> : identity<T> { };
-template <class T, class... Ts>
-struct tail<list<T, Ts...>> : identity<list<Ts...>> { };
-
-template <class... Ts, template <class...> class F, class... Args>
-struct count_if<list<Ts...>, F, Args...> : integral<
-  ::std::size_t,
-  filter<list<Ts...>, F, Args...>::type::size()
-> { };
-
-template <template <class...> class F, class... Args>
-struct find_if<list<>, F, Args...> : identity<list<>> { };
-
-template <template <class...> class F, class T, class... Ts, class... Args>
-struct find_if<list<T, Ts...>, F, Args...> : ::std::conditional<
-  F<Args..., T>::value,
-  list<T, Ts...>,
-  typename find_if<list<Ts...>, F, Args...>::type
-> { };
-
-template <class T, class U> struct count : count_if<T, ::std::is_same, U> { };
-template <class T, class U> struct find : find_if<T, ::std::is_same, U> { };
-
-template <template <class...> class F, class... Ts, class... Args>
-struct none_of<list<Ts...>, F, Args...> : none<F<Ts, Args...>::value...> { };
-
-template <template <class...> class F, class... Ts, class... Args>
-struct all_of<list<Ts...>, F, Args...> : all<F<Ts, Args...>::value...> { };
-
-template <template <class...> class F, class... Ts, class... Args>
-struct any_of<list<Ts...>, F, Args...> : any<F<Ts, Args...>::value...> { };
-
-template <bool B, bool... Bs> struct none<B, Bs...> :
-  boolean<not B and none<Bs...>::value>
-{ };
-template <> struct none<> : true_t { };
-
-template <bool B, bool... Bs> struct all<B, Bs...> :
-  boolean<B and all<Bs...>::value>
-{ };
-template <> struct all<> : true_t { };
-
-template <bool B, bool... Bs> struct any<B, Bs...> :
-  boolean<B or any<Bs...>::value>
-{ };
-template <> struct any<> : false_t { };
-
-template <class T, template <class...> class U>
-struct is_specialization_of : false_t { };
-template <template <class...> class T, class... Ts>
-struct is_specialization_of<T<Ts...>, T> : true_t { };
-
-}}}} /* namespace core::v2::meta::impl */
-
-namespace core {
-inline namespace v2 {
-namespace meta {
-
-template <bool B, class T = void>
-using unless = typename ::std::enable_if<not B, T>::type;
-
-template <bool B, class T = void>
-using when = typename ::std::enable_if<B, T>::type;
-
-template <bool B> using inhibit = unless<B, ::std::size_t>;
-template <bool B> using require = when<B, ::std::size_t>;
-
-template <bool B, class T, class F>
-using either = typename ::std::conditional<B, T, F>::type;
-
-using impl::integral;
-using impl::boolean;
-
-using impl::false_t;
-using impl::true_t;
-
-using impl::identity;
-using impl::list;
-
-using impl::is_specialization_of;
-
-template <class... Ts> using deduce = typename impl::deduce<Ts...>;
-
-template <class T, template <class...> class U>
-using convert = typename impl::convert<T, U>::type;
-
-template <class T, class U> using rebind = typename impl::rebind<T, U>::type;
-template <class T> using into = typename impl::into<T>::type;
-
-template <class... Ts> using join = typename impl::join<Ts...>::type;
-
-template <class T, size_t N> using get = typename impl::get<T, N>::type;
-template <class T> using head = typename impl::head<T>::type;
-template <class T> using tail = typename impl::tail<T>::type;
-
-template <class T, class U>
-using index_of_t = impl::index_of<T, U>;
-
-template <class T, template <class...> class F, class... Args>
-using count_if_t = impl::count_if<T, F, Args...>;
-
-template <class T, template <class...> class F, class... Args>
-using find_if = typename impl::find_if<T, F>::type;
-
-template <class T, class U> using count_t = impl::count<T, U>;
-
-template <class T, class U>
-using find = typename impl::find<T, U>::type;
-
-template <class T, template <class...> class F, class... Args>
-using none_of_t = impl::none_of<T, F, Args...>;
-
-template <class T, template <class...> class F, class... Args>
-using all_of_t = impl::all_of<T, F, Args...>;
-
-template <class T, template <class...> class F, class... Args>
-using any_of_t = impl::any_of<T, F, Args...>;
-
-template <bool... Bs> using none_t = impl::none<Bs...>;
-template <bool... Bs> using all_t = impl::all<Bs...>;
-template <bool... Bs> using any_t = impl::any<Bs...>;
-
-template <class T, class U>
-constexpr ::std::size_t index_of () noexcept {
-  return index_of_t<T, U>::value;
-}
-
-template <class T, template <class...> class F, class... Args>
-constexpr ::std::size_t count_if () noexcept {
-  return count_if_t<T, F, Args...>::value;
-}
-
-template <class T, class U>
-constexpr ::std::size_t count () noexcept { return count_t<T, U>::value; }
-
-template <class T, template <class...> class F, class... Args>
-constexpr bool none_of () noexcept { return none_of_t<T, F, Args...>::value; }
-
-template <class T, template <class...> class F, class... Args>
-constexpr bool all_of () noexcept { return all_of_t<T, F, Args...>::value; }
-
-template <class T, template <class...> class F, class... Args>
-constexpr bool any_of () noexcept { return any_of_t<T, F, Args...>::value; }
-
-template <bool... Bs>
-constexpr bool none () noexcept { return none_t<Bs...>::value; }
-
-template <bool... Bs>
-constexpr bool all () noexcept { return all_t<Bs...>::value; }
-
-template <bool... Bs>
-constexpr bool any () noexcept { return any_t<Bs...>::value; }
-
-}}} /* namespace core::v2::meta */
-
-namespace core {
-inline namespace v2 {
-namespace meta {
-
-template <class T, T... I> struct integer_sequence : identity<T> {
-  static_assert(
-    ::std::is_integral<T>::value,
-    "integer_sequence must use an integral type"
-  );
-
-  template <T N> using append = integer_sequence<T, I..., N>;
-  static constexpr ::std::size_t size() noexcept { return sizeof...(I); }
-  using next = append<sizeof...(I)>;
-};
-
-template <class T, T Index, ::std::size_t N>
-struct iota : identity<
-  typename iota<T, Index - 1, N - 1u>::type::next
-> { static_assert(Index >= 0, "Index cannot be negative"); };
-
-template <class T, T Index>
-struct iota<T, Index, 0u> : identity<integer_sequence<T>> { };
-
-template <::std::size_t... I>
-using index_sequence = integer_sequence<::std::size_t, I...>;
-
-template <class T, T N>
-using make_integer_sequence = typename iota<T, N, N>::type;
-
-template <::std::size_t N>
-using make_index_sequence = make_integer_sequence<::std::size_t, N>;
-
-template <class... Ts>
-using index_sequence_for = make_index_sequence<sizeof...(Ts)>;
-
-template <class> struct index_sequence_from;
-
-}}} /* namespace core::meta::v2 */
-
-#endif /* CORE_META_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/range.hpp 4.4.0-1/src/mlpack/core/std_backport/range.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/range.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/range.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,355 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_RANGE_HPP
-#define CORE_RANGE_HPP
-
-#include <istream>
-#include <utility>
-#include <memory>
-
-#include <cstdlib>
-
-#include "type_traits.hpp"
-#include "iterator.hpp"
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-using ::std::begin;
-using ::std::end;
-
-template <class T> using adl_begin_t = decltype(begin(::std::declval<T>()));
-template <class T> using adl_end_t = decltype(end(::std::declval<T>()));
-
-template <class T>
-adl_begin_t<T> adl_begin (T&& t) {
-  using ::std::begin;
-  return begin(::core::forward<T>(t));
-}
-
-template <class T>
-adl_end_t<T> adl_end (T&& t) {
-  using ::std::end;
-  return end(::core::forward<T>(t));
-}
-
-} /* namespace impl */
-
-template <class R>
-struct is_range : meta::all_t<
-  is_detected<impl::adl_begin_t, R>::value,
-  is_detected<impl::adl_end_t, R>::value
-> { };
-
-template <class Iterator>
-struct range {
-  using traits = ::std::iterator_traits<Iterator>;
-
-  using iterator_category = typename traits::iterator_category;
-
-  using difference_type = typename traits::difference_type;
-  using value_type = typename traits::value_type;
-
-  using reference = typename traits::reference;
-  using pointer = typename traits::pointer;
-
-  using iterator = Iterator;
-
-  static constexpr bool is_input = ::std::is_convertible<
-    iterator_category,
-    ::std::input_iterator_tag
-  >::value;
-
-  static constexpr bool is_output = ::std::is_convertible<
-    iterator_category,
-    ::std::output_iterator_tag
-  >::value;
-
-  static constexpr bool is_forward = ::std::is_convertible<
-    iterator_category,
-    ::std::forward_iterator_tag
-  >::value;
-
-  static constexpr bool is_bidirectional = ::std::is_convertible<
-    iterator_category,
-    ::std::bidirectional_iterator_tag
-  >::value;
-
-  static constexpr bool is_random_access = ::std::is_convertible<
-    iterator_category,
-    ::std::random_access_iterator_tag
-  >::value;
-
-  template <
-    class Range,
-    class=meta::when<
-      meta::all<
-        meta::none<
-          ::std::is_pointer<iterator>::value,
-          ::std::is_same<decay_t<Range>, range>::value
-        >(),
-        is_range<Range>::value,
-        is_detected_convertible<iterator, impl::adl_begin_t, Range>::value
-      >()
-    >
-  > explicit range (Range&& r) noexcept :
-    begin_ { impl::adl_begin(::core::forward<Range>(r)) },
-    end_ { impl::adl_end(::core::forward<Range>(r)) }
-  { }
-
-  range (::std::pair<iterator, iterator> pair) noexcept :
-    range { ::std::get<0>(pair), ::std::get<1>(pair) }
-  { }
-
-  range (iterator begin_, iterator end_) noexcept :
-    begin_ { begin_ },
-    end_ { end_ }
-  { }
-
-  range (range const& that) :
-    range { that.begin_, that.end_ }
-  { }
-
-  range (range&& that) noexcept :
-    range { ::core::move(that.begin_), ::core::move(that.end_) }
-  { that.begin_ = that.end_; }
-
-  range () = default;
-  ~range () = default;
-
-  range& operator = (range const& that) {
-    return *this = range { that };
-  }
-
-  range& operator = (range&& that) {
-    range { ::std::move(that) }.swap(*this);
-    return *this;
-  }
-
-  reference operator [](difference_type idx) const {
-    static_assert(is_random_access, "can only subscript into random-access");
-    return idx < 0 ? this->end()[idx] : this->begin()[idx];
-  }
-
-  iterator begin () const { return this->begin_; }
-  iterator end () const { return this->end_; }
-
-  reference front () const { return *this->begin(); }
-  reference back () const {
-    static_assert(is_bidirectional, "can only get back of bidirectional");
-    return *::std::prev(this->end());
-  }
-
-  bool empty () const { return this->begin() == this->end(); }
-
-  difference_type size () const {
-    static_assert(is_forward, "can only get size of forward-range");
-    return ::std::distance(this->begin(), this->end());
-  }
-
-  /* Creates an open-ended range of [start, stop) */
-  range slice (difference_type start, difference_type stop) const {
-    static_assert(is_forward, "can only slice forward-range");
-    /* Behavior is:
-     * if start is negative, the begin marker is this->end() - start
-     * if stop is negative, the end marker is this->end() - stop
-     * if start is positive, the begin marker is this->begin() + start
-     * if stop is positive, the end marker is this->begin() + stop
-     *
-     * if start and stop are positive, and stop is less than or equal to start,
-     * an empty range is returned.
-     *
-     * if start and stop are negative and stop is less than or equal to start,
-     * an empty range is returned.
-     *
-     * if start is positive and stop is negative and abs(stop) + start is
-     * greater than or equal to this->size(), an empty range is returned.
-     *
-     * if start is negative and stop is positive and this->size() + start is
-     * greater or equal to stop, an empty range is returned.
-     *
-     * The first two conditions can be computed cheaply, while the third and
-     * fourth are a bit more expensive, but WILL be required no matter what
-     * iterator type we are. However we don't compute the size until after
-     * we've checked the first two conditions
-     *
-     * An example with python style slicing for each would be:
-     * [4:3] -> empty range
-     * [-4:-4] -> empty range
-     * [7:-4] -> empty range for string of size 11 or more
-     * [-4:15] -> empty range for a string of size 19 or less.
-     */
-    bool const start_positive = start > 0;
-    bool const stop_positive = stop > 0;
-    bool const stop_less = stop < start;
-    bool const first_return_empty = start_positive == stop_positive and stop_less;
-    if (first_return_empty) { return range { }; }
-
-    /* now safe to compute size */
-    auto const size = this->size();
-    auto const third_empty = ::std::abs(stop) + start;
-
-    bool const second_return_empty =
-      (start_positive and not stop_positive and third_empty >= size) or
-      (not start_positive and stop_positive and size + start >= stop);
-    if (second_return_empty) { return range { }; }
-
-    /* While the code below technically works for all iterators it is
-     * ineffecient in some cases for bidirectional ranges, where either of
-     * start or stop are negative.
-     * TODO: Specialize for bidirectional operators
-     */
-    if (not start_positive) { start += size; }
-    if (not stop_positive) { stop += size; }
-
-    auto begin = this->begin();
-    ::std::advance(begin, start);
-
-    auto end = begin;
-    ::std::advance(end, stop - start);
-
-    return range { begin, end };
-  }
-
-  /* Creates an open-ended range of [start, end()) */
-  range slice (difference_type start) const {
-    static_assert(is_forward, "can only slice forward-range");
-    return range { split(start).second };
-  }
-
-  ::std::pair<range, range> split (difference_type idx) const {
-    static_assert(is_forward,"can only split a forward-range");
-    if (idx >= 0) {
-      range second { *this };
-      second.pop_front_upto(idx);
-      return ::std::make_pair(range { this->begin(), second.begin() }, second);
-    }
-
-    range first { *this };
-    first.pop_back_upto(-idx);
-    return ::std::make_pair(first, range { first.end(), this->end() });
-  }
-
-  /* mutates range */
-  void pop_front (difference_type n) { ::std::advance(this->begin_, n); }
-  void pop_front () { ++this->begin_; }
-
-  void pop_back (difference_type n) {
-    static_assert(is_bidirectional, "can only pop-back bidirectional-range");
-    ::std::advance(this->end_, -n);
-  }
-
-  void pop_back () {
-    static_assert(is_bidirectional, "can only pop-back bidirectional-range");
-    --this->end_;
-  }
-
-  /* Negative argument causes no change */
-  void pop_front_upto (difference_type n) {
-    ::std::advance(
-      this->begin_,
-      ::std::min(::std::max<difference_type>(0, n), this->size())
-    );
-  }
-
-  /* Negative argument causes no change */
-  void pop_back_upto (difference_type n) {
-    static_assert(is_bidirectional, "can only pop-back-upto bidirectional");
-    ::std::advance(
-      this->end_,
-      -::std::min(::std::max<difference_type>(0, n), this->size())
-    );
-  }
-
-  void swap (range& that) noexcept(is_nothrow_swappable<iterator>::value) {
-    using ::std::swap;
-    swap(this->begin_, that.begin_);
-    swap(this->end_, that.end_);
-  }
-
-private:
-  iterator begin_;
-  iterator end_;
-};
-
-template <class T>
-auto make_range (T* ptr, ::std::size_t n) -> range<T*> {
-  return range<T*> { ptr, ptr + n };
-}
-
-template <class Iterator>
-auto make_range (Iterator begin, Iterator end) -> range<Iterator> {
-  return range<Iterator> { begin, end };
-}
-
-template <class Range>
-auto make_range (Range&& value) -> range<decltype(begin(value))> {
-  using ::std::begin;
-  using ::std::end;
-  return make_range(begin(value), end(value));
-}
-
-/* Used like: core::make_range<char>(::std::cin) */
-template <
-  class T,
-  class CharT,
-  class Traits=::std::char_traits<CharT>
-> auto make_range (::std::basic_istream<CharT, Traits>& stream) -> range<
-  ::std::istream_iterator<T, CharT, Traits>
-> {
-  using iterator = ::std::istream_iterator<T, CharT, Traits>;
-  return make_range(iterator { stream }, iterator { });
-}
-
-template <class CharT, class Traits=::std::char_traits<CharT>>
-auto make_range (::std::basic_streambuf<CharT, Traits>* buffer) -> range<
-  ::std::istreambuf_iterator<CharT, Traits>
-> {
-  using iterator = ::std::istreambuf_iterator<CharT, Traits>;
-  return make_range(iterator { buffer }, iterator { });
-}
-
-template <class Iter>
-range<::std::move_iterator<Iter>> make_move_range (Iter start, Iter stop) {
-  return make_range(
-    ::std::make_move_iterator(start),
-    ::std::make_move_iterator(stop));
-}
-
-template <class T>
-range<::std::move_iterator<T*>> make_move_range (T* ptr, ::std::size_t n) {
-  return make_move_range(ptr, ptr + n);
-}
-
-template <class T>
-range<number_iterator<T>> make_number_range(T start, T stop, T step) noexcept {
-  auto begin = make_number_iterator(start, step);
-  auto end = make_number_iterator(stop, step);
-  return make_range(begin, end);
-}
-
-template <class T>
-range<number_iterator<T>> make_number_range (T start, T stop) noexcept {
-  return make_range(make_number_iterator(start), make_number_iterator(stop));
-}
-
-template <class Iterator>
-void swap (range<Iterator>& lhs, range<Iterator>& rhs) noexcept(
-  noexcept(lhs.swap(rhs))
-) { lhs.swap(rhs); }
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_RANGE_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/string_view.hpp 4.4.0-1/src/mlpack/core/std_backport/string_view.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/string_view.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/string_view.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,760 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_STRING_VIEW_HPP
-#define CORE_STRING_VIEW_HPP
-
-#include <initializer_list>
-#include <functional>
-#include <stdexcept>
-#include <algorithm>
-#include <iterator>
-#include <string>
-#include <limits>
-
-#include <cstdlib>
-#include <ciso646>
-
-#if defined(_MSC_VER)
-  #pragma warning(push)
-  #pragma warning(disable:5030)
-  #pragma warning(disable:4702)
-#endif /* defined(_MSC_VER) */
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-/* implementations of MurmurHash2 *Endian Neutral* (but not alignment!) */
-template <::std::size_t=sizeof(::std::size_t)> struct murmur;
-template <> struct murmur<4> {
-  constexpr murmur () = default;
-
-  ::std::uint32_t operator () (void const* p, ::std::size_t len) const noexcept {
-    static constexpr ::std::uint32_t magic = UINT32_C(0x5BD1E995);
-    static constexpr auto shift = 24;
-
-    auto hash = static_cast<::std::uint32_t>(len);
-    auto data = static_cast<::std::uint8_t const*>(p);
-
-    while (len >= sizeof(::std::uint32_t)) {
-      ::std::uint32_t mix = data[0];
-      mix |= ::std::uint32_t(data[1]) <<  8;
-      mix |= ::std::uint32_t(data[2]) << 16;
-      mix |= ::std::uint32_t(data[3]) << 24;
-
-      mix *= magic;
-      mix ^= mix >> shift;
-      mix *= magic;
-
-      hash *= magic;
-      hash ^= mix;
-
-      data += sizeof(::std::uint32_t);
-      len -= sizeof(::std::uint32_t);
-    }
-
-    switch (len) {
-      case 3: hash ^= ::std::uint32_t(data[2]) << 16; [[clang::fallthrough]];
-      case 2: hash ^= ::std::uint32_t(data[1]) <<  8; [[clang::fallthrough]];
-      case 1: hash ^= ::std::uint32_t(data[0]);
-              hash *= magic;
-    }
-
-    hash ^= hash >> 13;
-    hash *= magic;
-    hash ^= hash >> 15;
-
-    return hash;
-  }
-};
-
-template <> struct murmur<8> {
-  constexpr murmur () = default;
-
-  ::std::uint64_t operator () (void const* p, ::std::size_t len) const noexcept {
-    static constexpr ::std::uint64_t magic = UINT64_C(0xC6A4A7935BD1E995);
-    static constexpr auto shift = 47;
-
-    ::std::uint64_t hash = len * magic;
-
-    auto data = static_cast<::std::uint8_t const*>(p);
-
-    while (len >= sizeof(::std::uint64_t)) {
-      ::std::uint64_t mix = data[0];
-      mix |= ::std::uint64_t(data[1]) <<  8;
-      mix |= ::std::uint64_t(data[2]) << 16;
-      mix |= ::std::uint64_t(data[3]) << 24;
-      mix |= ::std::uint64_t(data[4]) << 32;
-      mix |= ::std::uint64_t(data[5]) << 40;
-      mix |= ::std::uint64_t(data[6]) << 48;
-      mix |= ::std::uint64_t(data[7]) << 54;
-
-      mix *= magic;
-      mix ^= mix >> shift;
-      mix *= magic;
-
-      hash ^= mix;
-      hash *= magic;
-
-      data += sizeof(::std::uint64_t);
-      len -= sizeof(::std::uint64_t);
-    }
-
-    switch (len & 7) {
-      case 7: hash ^= ::std::uint64_t(data[6]) << 48; [[clang::fallthrough]];
-      case 6: hash ^= ::std::uint64_t(data[5]) << 40; [[clang::fallthrough]];
-      case 5: hash ^= ::std::uint64_t(data[4]) << 32; [[clang::fallthrough]];
-      case 4: hash ^= ::std::uint64_t(data[3]) << 24; [[clang::fallthrough]];
-      case 3: hash ^= ::std::uint64_t(data[2]) << 16; [[clang::fallthrough]];
-      case 2: hash ^= ::std::uint64_t(data[1]) << 8;  [[clang::fallthrough]];
-      case 1: hash ^= ::std::uint64_t(data[0]);
-              hash *= magic;
-    }
-
-    hash ^= hash >> shift;
-    hash *= magic;
-    hash ^= hash >> shift;
-
-    return hash;
-  }
-};
-
-}}} /* namespace core::v2::impl */
-
-namespace core {
-inline namespace v2 {
-
-#ifndef CORE_NO_EXCEPTIONS
-[[noreturn]] inline void throw_out_of_range (char const* msg) {
-  throw ::std::out_of_range { msg };
-}
-#else /* CORE_NO_EXCEPTIONS */
-[[noreturn]] inline void throw_out_of_range (char const*) { ::std::abort(); }
-#endif /* CORE_NO_EXCEPTIONS */
-
-template <class CharT, class Traits=::std::char_traits<CharT>>
-struct basic_string_view {
-  using difference_type = ::std::ptrdiff_t;
-  using value_type = CharT;
-  using size_type = ::std::size_t;
-
-  using reference = value_type const&;
-  using pointer = value_type const*;
-
-  using const_reference = reference;
-  using const_pointer = pointer;
-
-  using const_iterator = pointer;
-  using iterator = const_iterator;
-
-  using const_reverse_iterator = ::std::reverse_iterator<const_iterator>;
-  using reverse_iterator = const_reverse_iterator;
-
-  using traits = Traits;
-
-  static constexpr size_type npos = ::std::numeric_limits<size_type>::max();
-
-  template <class Allocator>
-  basic_string_view (
-    ::std::basic_string<CharT, Traits, Allocator> const& that
-  ) : str { that.data() }, len { that.size() } { }
-
-  constexpr basic_string_view (pointer str, size_type len) noexcept :
-    str { str },
-    len { len }
-  { }
-
-  basic_string_view (pointer str) noexcept :
-    basic_string_view { str, traits::length(str) }
-  { }
-
-  constexpr basic_string_view (basic_string_view const&) noexcept = default;
-  constexpr basic_string_view () noexcept = default;
-  basic_string_view& operator = (basic_string_view const&) noexcept = default;
-
-  template <class Allocator>
-  explicit operator ::std::basic_string<CharT, Traits, Allocator> () const {
-    return ::std::basic_string<CharT, Traits, Allocator> {
-      this->data(),
-      this->size()
-    };
-  }
-
-  template <class Allocator=std::allocator<CharT>>
-  ::std::basic_string<CharT, Traits, Allocator> to_string (
-    Allocator const& allocator = Allocator()
-  ) const {
-    return ::std::basic_string<CharT, Traits, Allocator> {
-      this->data(),
-      this->size(),
-      allocator
-    };
-  }
-
-  constexpr const_iterator begin () const noexcept { return this->data(); }
-  constexpr const_iterator end () const noexcept {
-    return this->data() + this->size();
-  }
-
-  constexpr const_iterator cbegin () const noexcept { return this->begin(); }
-  constexpr const_iterator cend () const noexcept { return this->end(); }
-
-  const_reverse_iterator rbegin () const noexcept {
-    return const_reverse_iterator { this->end()};
-  }
-
-  const_reverse_iterator rend () const noexcept {
-    return const_reverse_iterator { this->begin() };
-  }
-
-  const_reverse_iterator crbegin () const noexcept { return this->rbegin(); }
-  const_reverse_iterator crend () const noexcept { return this->rend(); }
-
-  constexpr size_type max_size () const noexcept {
-    return ::std::numeric_limits<size_type>::max();
-  }
-  constexpr size_type length () const noexcept { return this->size(); }
-  constexpr size_type size () const noexcept { return this->len; }
-
-  constexpr bool empty () const noexcept { return this->size() == 0; }
-
-  constexpr reference operator [] (size_type idx) const {
-    return this->str[idx];
-  }
-
-  constexpr reference front () const { return this->str[0]; }
-  constexpr reference back () const { return this->str[this->size() - 1]; }
-  constexpr pointer data () const { return this->str; }
-
-  void remove_prefix (size_type n) {
-    if (n > this->size()) { n = this->size(); }
-    this->str += n;
-    this->len -= n;
-  }
-
-  void remove_suffix (size_type n) {
-    if (n > this->size()) { n = this->size(); }
-    this->len -= n;
-  }
-
-  void clear () noexcept {
-    this->str = nullptr;
-    this->len = 0;
-  }
-
-  size_type copy (CharT* s, size_type n, size_type pos = 0) const {
-    if (pos > this->size()) {
-      throw_out_of_range("position greater than size");
-    }
-    auto const rlen = std::min(n, this->size() - pos);
-    ::std::copy_n(this->begin() + pos, rlen, s);
-    return rlen;
-  }
-
-  constexpr basic_string_view substr (
-    size_type pos=0,
-    size_type n=npos
-  ) const noexcept {
-    return pos > this->size()
-      ? (throw_out_of_range("start position out of range"), *this)
-      : basic_string_view {
-        this->data() + pos,
-        n == npos or pos + n > this->size()
-          ? (this->size() - pos)
-          : n
-      };
-  }
-
-  bool starts_with (value_type value) const noexcept {
-    return not this->empty() and traits::eq(value, this->front());
-  }
-
-  bool ends_with (value_type value) const noexcept {
-    return not this->empty() and traits::eq(value, this->back());
-  }
-
-  bool starts_with (basic_string_view that) const noexcept {
-    return this->size() >= that.size() and
-      traits::compare(this->data(), that.data(), that.size()) == 0;
-  }
-
-  bool ends_with (basic_string_view that) const noexcept {
-    return this->size() >= that.size() and
-      traits::compare(
-        this->data() + this->size() - that.size(),
-        that.data(),
-        that.size()
-      ) == 0;
-  }
-
-  /* compare */
-  difference_type compare (basic_string_view s) const noexcept {
-    auto cmp = traits::compare(
-      this->data(),
-      s.data(),
-      ::std::min(this->size(), s.size())
-    );
-
-    if (cmp != 0) { return cmp; }
-    if (this->size() == s.size()) { return 0; }
-    if (this->size() < s.size()) { return -1; }
-    return 1;
-  }
-
-  difference_type compare (
-    size_type pos,
-    size_type n,
-    basic_string_view s
-  ) const noexcept { return this->substr(pos, n).compare(s); }
-
-  difference_type compare (
-    size_type pos1,
-    size_type n1,
-    basic_string_view s,
-    size_type pos2,
-    size_type n2
-  ) const noexcept {
-    return this->substr(pos1, n1).compare(s.substr(pos2, n2));
-  }
-
-  difference_type compare (pointer s) const noexcept {
-    return this->compare(basic_string_view { s });
-  }
-
-  difference_type compare (
-    size_type pos,
-    size_type n,
-    pointer s
-  ) const noexcept {
-    return this->substr(pos, n).compare(basic_string_view { s });
-  }
-
-  difference_type compare (
-    size_type pos,
-    size_type n1,
-    pointer s,
-    size_type n2
-  ) const noexcept {
-    return this->substr(pos, n1).compare(basic_string_view { s, n2 });
-  }
-
-  reference at (size_type idx) const {
-    static constexpr auto error = "requested index out of range";
-    if (idx >= this->size()) { throw_out_of_range(error); }
-    return this->str[idx];
-  }
-
-  /* find-first-not-of */
-  size_type find_first_not_of (
-    basic_string_view str,
-    size_type pos = 0) const noexcept {
-    if (pos > this->size()) { return npos; }
-    auto begin = this->begin() + pos;
-    auto end = this->end();
-    auto const predicate = [str] (value_type v) { return str.find(v) == npos; };
-    auto iter = std::find_if(begin, end, predicate);
-    if (iter == end) { return npos; }
-    return static_cast<size_type>(::std::distance(this->begin(), iter));
-  }
-
-  size_type find_first_not_of (
-    pointer s,
-    size_type pos,
-    size_type n) const noexcept {
-      return this->find_first_not_of(basic_string_view { s, n }, pos);
-  }
-
-  size_type find_first_not_of (pointer s, size_type pos = 0) const noexcept {
-    return this->find_first_not_of(basic_string_view { s }, pos);
-  }
-
-  size_type find_first_not_of (value_type c, size_type pos = 0) const noexcept {
-    return this->find_first_not_of(
-      basic_string_view { ::std::addressof(c), 1 },
-      pos);
-  }
-
-  /* find-first-of */
-  size_type find_first_of (
-    basic_string_view str,
-    size_type pos = 0) const noexcept {
-    if (pos > this->size()) { return npos; }
-    auto iter = ::std::find_first_of(
-      this->begin() + pos, this->end(),
-      str.begin(), str.end(),
-      traits::eq);
-    if (iter == this->end()) { return npos; }
-    return static_cast<size_type>(::std::distance(this->begin(), iter));
-  }
-
-  size_type find_first_of (pointer s, size_type p, size_type n) const noexcept {
-    return this->find_first_of(basic_string_view { s, n }, p);
-  }
-
-  size_type find_first_of (pointer s, size_type pos = 0) const noexcept {
-    return this->find_first_of(basic_string_view { s }, pos);
-  }
-
-  size_type find_first_of (value_type c, size_type pos = 0) const noexcept {
-    return this->find_first_of(
-      basic_string_view { ::std::addressof(c), 1 },
-      pos);
-  }
-
-  /* find */
-  size_type find (basic_string_view str, size_type pos = 0) const noexcept {
-    if (pos >= this->size()) { return npos; }
-    auto iter = ::std::search(
-      this->begin() + pos, this->end(),
-      str.begin(), str.end(),
-      traits::eq);
-    if (iter == this->end()) { return npos; }
-    return static_cast<size_type>(::std::distance(this->begin(), iter));
-  }
-
-  size_type find (pointer s, size_type p, size_type n) const noexcept {
-    return this->find(basic_string_view { s, n }, p);
-  }
-
-  size_type find (pointer s, size_type pos = 0) const noexcept {
-    return this->find(basic_string_view { s }, pos);
-  }
-
-  size_type find (value_type c, size_type pos = 0) const noexcept {
-    return this->find(basic_string_view { ::std::addressof(c), 1 }, pos);
-  }
-
-  size_type find_last_not_of (
-    basic_string_view str,
-    size_type pos = npos) const noexcept {
-    auto const offset = this->size() - ::std::min(this->size(), pos);
-    auto begin = this->rbegin() + static_cast<difference_type>(offset);
-    auto end = this->rend();
-    auto const predicate = [str] (value_type v) { return str.find(v) == npos; };
-    auto iter = ::std::find_if(begin, end, predicate);
-    if (iter == end) { return npos; }
-    auto const distance = static_cast<size_type>(
-      ::std::distance(this->rbegin(), iter));
-    return this->size() - distance - 1;
-  }
-
-  size_type find_last_not_of (
-    pointer s,
-    size_type p,
-    size_type n) const noexcept {
-    return this->find_last_not_of(basic_string_view { s, n }, p);
-  }
-
-  size_type find_last_not_of (pointer s, size_type p = npos) const noexcept {
-    return this->find_last_not_of(basic_string_view { s }, p);
-  }
-
-  size_type find_last_not_of (
-    value_type c,
-    size_type pos = npos) const noexcept {
-    return this->find_last_not_of(
-      basic_string_view { ::std::addressof(c), 1 },
-      pos);
-  }
-
-  size_type find_last_of (
-    basic_string_view str,
-    size_type pos = npos) const noexcept {
-    auto const offset = this->size() - ::std::min(this->size(), pos);
-    auto begin = this->rbegin() + static_cast<difference_type>(offset);
-    auto end = this->rend();
-
-    auto iter = ::std::find_first_of(
-      begin, end,
-      str.rbegin(), str.rend(),
-      traits::eq);
-    if (iter == end) { return npos; }
-    auto const distance = static_cast<size_type>(
-      ::std::distance(this->rbegin(), iter));
-    return this->size() - distance - 1;
-  }
-
-  size_type find_last_of (pointer s, size_type p, size_type n) const noexcept {
-    return this->find_last_of(basic_string_view { s, n }, p);
-  }
-
-  size_type find_last_of (pointer s, size_type p=npos) const noexcept {
-    return this->find_last_of(basic_string_view { s }, p);
-  }
-
-  size_type find_last_of (value_type c, size_type p=npos) const noexcept {
-    return this->find_last_of(basic_string_view { ::std::addressof(c), 1 }, p);
-  }
-
-  size_type rfind (basic_string_view str, size_type pos=npos) const noexcept {
-    auto const offset = this->size() - ::std::min(this->size(), pos);
-    auto begin = this->rbegin() + offset;
-    auto end = this->rend();
-    auto iter = ::std::search(
-      begin, end,
-      str.rbegin(), str.rend(),
-      traits::eq);
-    if (iter == end) { return npos; }
-    auto const distance = static_cast<size_type>(
-      ::std::distance(this->rbegin(), iter));
-    return this->size() - distance - 1;
-  }
-
-  size_type rfind (pointer s, size_type p, size_type n) const noexcept {
-    return this->rfind(basic_string_view { s, n }, p);
-  }
-
-  size_type rfind (pointer s, size_type p=npos) const noexcept {
-    return this->rfind(basic_string_view { s }, p);
-  }
-
-  size_type rfind (value_type c, size_type p=npos) const noexcept {
-    return this->rfind(basic_string_view { ::std::addressof(c), 1 }, p);
-  }
-
-  void swap (basic_string_view& that) noexcept {
-    using ::std::swap;
-    swap(this->str, that.str);
-    swap(this->len, that.len);
-  }
-
-private:
-  pointer str { nullptr };
-  size_type len { 0 };
-};
-
-using u32string_view = basic_string_view<char32_t>;
-using u16string_view = basic_string_view<char16_t>;
-using wstring_view = basic_string_view<wchar_t>;
-using string_view = basic_string_view<char>;
-
-/* string_view comparison string_view */
-template <class CharT, typename Traits>
-bool operator == (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.size() == rhs.size() and lhs.compare(rhs) == 0; }
-
-template <class CharT, typename Traits>
-bool operator != (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.size() != rhs.size() or lhs.compare(rhs) != 0; }
-
-template <class CharT, typename Traits>
-bool operator >= (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.compare(rhs) >= 0; }
-
-template <class CharT, typename Traits>
-bool operator <= (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.compare(rhs) <= 0; }
-
-template <class CharT, typename Traits>
-bool operator > (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.compare(rhs) > 0; }
-
-template <class CharT, typename Traits>
-bool operator < (
-  basic_string_view<CharT, Traits> lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return lhs.compare(rhs) < 0; }
-
-/* string_view comparison string */
-template <class CharT, class Traits, class Allocator>
-bool operator == (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs == basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator != (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs != basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator >= (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs >= basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator <= (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs <= basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator > (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs > basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator < (
-  basic_string_view<CharT, Traits> lhs,
-  ::std::basic_string<CharT, Traits, Allocator> const& rhs
-) noexcept { return lhs < basic_string_view<CharT, Traits> { rhs }; }
-
-/* string comparison string_view */
-template <class CharT, class Traits, class Allocator>
-bool operator == (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } == rhs; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator != (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } != rhs; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator >= (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } >= rhs; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator <= (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } <= rhs; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator > (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } > rhs; }
-
-template <class CharT, class Traits, class Allocator>
-bool operator < (
-  ::std::basic_string<CharT, Traits, Allocator> const& lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } < rhs; }
-
-/* string_view comparison CharT* */
-template <class CharT, class Traits>
-bool operator == (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs == basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits>
-bool operator != (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs != basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits>
-bool operator >= (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs >= basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits>
-bool operator <= (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs <= basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits>
-bool operator > (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs > basic_string_view<CharT, Traits> { rhs }; }
-
-template <class CharT, class Traits>
-bool operator < (
-  basic_string_view<CharT, Traits> lhs,
-  CharT const* rhs
-) noexcept { return lhs < basic_string_view<CharT, Traits> { rhs }; }
-
-/* CharT* comparison string_view */
-template <class CharT, class Traits>
-bool operator == (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } == rhs; }
-
-template <class CharT, class Traits>
-bool operator != (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } != rhs; }
-
-template <class CharT, class Traits>
-bool operator >= (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } >= rhs; }
-
-template <class CharT, class Traits>
-bool operator <= (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } <= rhs; }
-
-template <class CharT, class Traits>
-bool operator > (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } > rhs; }
-
-template <class CharT, class Traits>
-bool operator < (
-  CharT const* lhs,
-  basic_string_view<CharT, Traits> rhs
-) noexcept { return basic_string_view<CharT, Traits> { lhs } < rhs; }
-
-template <class CharT, class Traits>
-::std::basic_ostream<CharT, Traits>& operator << (
-  ::std::basic_ostream<CharT, Traits>& os,
-  basic_string_view<CharT, Traits> const& str
-) { return os << str.to_string(); }
-
-template <class CharT, class Traits>
-void swap (
-  basic_string_view<CharT, Traits>& lhs,
-  basic_string_view<CharT, Traits>& rhs
-) noexcept { return lhs.swap(rhs); }
-
-}} /* namespace core::v2 */
-
-namespace std {
-
-template <typename CharT, typename Traits>
-struct hash<core::v2::basic_string_view<CharT, Traits>> {
-  using argument_type = core::v2::basic_string_view<CharT, Traits>;
-  using result_type = size_t;
-
-  result_type operator ()(argument_type const& ref) const noexcept {
-    static constexpr core::impl::murmur<sizeof(size_t)> hasher { };
-    return hasher(ref.data(), ref.size());
-  }
-};
-
-} /* namespace std */
-
-
-#if defined(_MSC_VER)
-  #pragma warning(pop)
-#endif /* defined(_MSC_VER) */
-#endif /* CORE_STRING_VIEW_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/type_traits.hpp 4.4.0-1/src/mlpack/core/std_backport/type_traits.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/type_traits.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/type_traits.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,266 +0,0 @@
-#ifndef CORE_TYPE_TRAITS_HPP
-#define CORE_TYPE_TRAITS_HPP
-
-#include <type_traits>
-#include <utility>
-#include <tuple>
-
-#include "internal.hpp"
-
-namespace core {
-inline namespace v2 {
-namespace impl {
-
-/* union used for variant<Ts...> and implementing aligned_union, which is
- * not provided by gcc 4.8.x, but is provided by clang. (aligned_union_t is
- * the only alias missing from <type_traits>)
- */
-template <class... Ts> union discriminate;
-template <> union discriminate<> { };
-template <class T, class... Ts>
-union discriminate<T, Ts...> {
-  T value;
-  discriminate<Ts...> rest;
-};
-
-} /* namespace impl */
-
-/* custom type traits and types */
-template <class T> using identity_t = typename meta::identity<T>::type;
-template <class T> using identity = meta::identity<T>;
-
-/* extracts the class of a member function ponter */
-template <class T> using class_of_t = impl::class_of_t<T>;
-template <class T> using class_of = impl::class_of<T>;
-
-template <::std::size_t I, class T>
-using tuple_element_t = typename ::std::tuple_element<I, T>::type;
-template <class T> using tuple_size_t = typename ::std::tuple_size<T>::type;
-
-/* Implementation of N4389 */
-template <bool B> using bool_constant = ::std::integral_constant<bool, B>;
-
-template <class...> struct conjunction;
-template <class...> struct disjunction;
-template <class...> struct negation;
-
-template <class T, class... Ts>
-struct conjunction<T, Ts...> :
-  bool_constant<T::value and conjunction<Ts...>::value>
-{ };
-template <> struct conjunction<> : ::std::true_type { };
-
-template <class T, class... Ts>
-struct disjunction<T, Ts...> :
-  bool_constant<T::value or disjunction<Ts...>::value>
-{ };
-template <> struct disjunction<> : ::std::false_type { };
-
-template <class T, class... Ts>
-struct negation<T, Ts...> :
-  bool_constant<not T::value and negation<Ts...>::value>
-{ };
-template <> struct negation<> : ::std::false_type { };
-
-/* C++ Library Fundamentals V2 TS detection idiom */
-template <class... Ts> using void_t = meta::deduce<Ts...>;
-
-struct nonesuch {
-  nonesuch (nonesuch const&) = delete;
-  nonesuch () = delete;
-  ~nonesuch () = delete;
-  void operator = (nonesuch const&) = delete;
-};
-
-template <class T, template <class...> class U, class... Args>
-using detected_or = impl::make_detect<T, void, U, Args...>;
-
-template <template <class...> class T, class... Args>
-using detected_t = typename detected_or<nonesuch, T, Args...>::type;
-
-template <class T, template <class...> class U, class... Args>
-using detected_or_t = typename detected_or<T, U, Args...>::type;
-
-template <class To, template <class...> class T, class... Args>
-using is_detected_convertible = ::std::is_convertible<
-  detected_t<T, Args...>,
-  To
->;
-
-template <class T, template <class...> class U, class... Args>
-using is_detected_same = ::std::is_same<T, detected_t<U, Args...>>;
-
-template <class T, template<class...> class U, class... Args>
-using is_detected_convertible = ::std::is_convertible<
-  detected_t<U, Args...>,
-  T
->;
-
-template <template <class...> class T, class... Args>
-using is_detected = typename detected_or<nonesuch, T, Args...>::value_t;
-
-/* forward declaration */
-template <::std::size_t, class...> struct aligned_union;
-template <class...> struct invokable;
-template <class...> struct invoke_of;
-template <class T> struct result_of; /* SFINAE result_of */
-
-/* C++14 style aliases for standard traits */
-template <class T>
-using remove_volatile_t = typename ::std::remove_volatile<T>::type;
-
-template <class T>
-using remove_const_t = typename ::std::remove_const<T>::type;
-template <class T> using remove_cv_t = typename ::std::remove_cv<T>::type;
-
-template <class T>
-using add_volatile_t = typename ::std::add_volatile<T>::type;
-template <class T> using add_const_t = typename ::std::add_const<T>::type;
-template <class T> using add_cv_t = typename ::std::add_cv<T>::type;
-
-template <class T>
-using add_lvalue_reference_t = typename ::std::add_lvalue_reference<T>::type;
-
-template <class T>
-using add_rvalue_reference_t = typename ::std::add_rvalue_reference<T>::type;
-
-template <class T>
-using remove_reference_t = typename ::std::remove_reference<T>::type;
-
-template <class T>
-using remove_pointer_t = typename ::std::remove_pointer<T>::type;
-
-template <class T> using add_pointer_t = typename ::std::add_pointer<T>::type;
-
-template <class T>
-using make_unsigned_t = typename ::std::make_unsigned<T>::type;
-template <class T> using make_signed_t = typename ::std::make_signed<T>::type;
-
-template <class T>
-using remove_extent_t = typename ::std::remove_extent<T>::type;
-
-template <class T>
-using remove_all_extents_t = typename ::std::remove_all_extents<T>::type;
-
-template <
-  ::std::size_t Len,
-  ::std::size_t Align = alignof(typename ::std::aligned_storage<Len>::type)
-> using aligned_storage_t = typename ::std::aligned_storage<Len, Align>::type;
-
-template <::std::size_t Len, class... Types>
-using aligned_union_t = typename aligned_union<Len, Types...>::type;
-
-template <class T> using decay_t = impl::decay_t<T>;
-
-template <bool B, class T = void>
-using enable_if_t = typename ::std::enable_if<B, T>::type;
-
-template <bool B, class T, class F>
-using conditional_t = typename ::std::conditional<B, T, F>::type;
-
-template <class T>
-using underlying_type_t = typename ::std::underlying_type<T>::type;
-
-template <::std::size_t Len, class... Types>
-struct aligned_union {
-  using union_type = impl::discriminate<Types...>;
-  static constexpr ::std::size_t size () noexcept {
-    return Len > sizeof(union_type) ? Len : sizeof(union_type);
-  }
-
-  static constexpr ::std::size_t alignment_value = alignof(
-    impl::discriminate<Types...>
-  );
-
-  using type = aligned_storage_t<
-    (Len > sizeof(union_type) ? Len : sizeof(union_type)),
-    alignment_value
-  >;
-};
-
-/* custom type trait specializations */
-template <class... Args> using invoke_of_t = typename invoke_of<Args...>::type;
-
-template <class... Args>
-struct invokable : meta::none_t<
-  std::is_same<
-    decltype(impl::INVOKE(::std::declval<Args>()...)),
-    impl::undefined
-  >::value
-> { };
-
-template <class... Args> struct invoke_of :
-  impl::invoke_of<invokable<Args...>::value, Args...>
-{ };
-
-template <class F, class... Args>
-struct result_of<F(Args...)> : invoke_of<F, Args...> { };
-
-template <class T> using result_of_t = typename result_of<T>::type;
-
-template <class... Ts> struct common_type;
-
-template <class T> struct common_type<T> : identity<decay_t<T>> { };
-template <class T, class U>
-struct common_type<T, U> : identity<
-  decay_t<decltype(true ? ::std::declval<T>() : ::std::declval<U>())>
-> { };
-
-template <class T, class U, class... Ts>
-struct common_type<T, U, Ts...> : identity<
-  typename common_type<
-    typename common_type<T, U>::type,
-    Ts...
-  >::type
-> { };
-
-template <class... T> using common_type_t = typename common_type<T...>::type;
-
-/* is_null_pointer */
-template <class T> struct is_null_pointer : ::std::false_type { };
-
-template <>
-struct is_null_pointer<add_cv_t<::std::nullptr_t>> : ::std::true_type { };
-template <>
-struct is_null_pointer<::std::nullptr_t volatile> : ::std::true_type { };
-template <>
-struct is_null_pointer<::std::nullptr_t const> : ::std::true_type { };
-template <>
-struct is_null_pointer<::std::nullptr_t> : ::std::true_type { };
-
-/* is_nothrow_swappable - N4426 (implemented before paper was proposed) */
-template <class T, class U=T>
-using is_nothrow_swappable = impl::is_nothrow_swappable<T, U>;
-
-/* propagates const or volatile without using the name propagate :) */
-template <class T, class U>
-struct transmit_volatile : ::std::conditional<
-  ::std::is_volatile<T>::value,
-  add_volatile_t<U>,
-  U
-> { };
-
-template <class T, class U>
-struct transmit_const : ::std::conditional<
-  ::std::is_const<T>::value,
-  add_const_t<U>,
-  U
-> { };
-
-template <class T, class U>
-struct transmit_cv : transmit_volatile<
-  T, typename transmit_const<T, U>::type
-> { };
-
-template <class T, class U>
-using transmit_volatile_t = typename transmit_volatile<T, U>::type;
-
-template <class T, class U>
-using transmit_const_t = typename transmit_const<T, U>::type;
-
-template <class T, class U>
-using transmit_cv_t = typename transmit_cv<T, U>::type;
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_TYPE_TRAITS_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/typeinfo.hpp 4.4.0-1/src/mlpack/core/std_backport/typeinfo.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/typeinfo.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/typeinfo.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,79 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_TYPEINFO_HPP
-#define CORE_TYPEINFO_HPP
-
-#include "type_traits.hpp"
-#include "utility.hpp"
-
-#ifndef CORE_NO_RTTI
-  #include <typeindex>
-  #include <typeinfo>
-#endif /* CORE_NO_RTTI */
-
-namespace core {
-inline namespace v2 {
-
-#ifndef CORE_NO_RTTI
-using type_info = ::std::type_info;
-
-template <class T>
-type_info const& type_of () noexcept { return typeid(T); }
-#else /* CORE_NO_RTTI */
-struct type_info final {
-
-  type_info (type_info const&) = delete;
-  type_info () = delete;
-  virtual ~type_info () = default;
-
-  type_info& operator = (type_info const&) = delete;
-
-  /* If we had C++14 template variables, this would actually be easier */
-  template <class T>
-  friend type_info const& type_of () noexcept {
-    return type_info::cref<remove_reference_t<remove_cv_t<T>>>();
-  }
-
-  bool operator == (type_info const& that) const noexcept {
-    return this->id == that.id;
-  }
-
-  bool operator != (type_info const& that) const noexcept {
-    return this->id != that.id;
-  }
-
-  bool before (type_info const& that) const noexcept {
-    return this->id < that.id;
-  }
-
-  ::std::size_t hash_code () const noexcept { return this->id; }
-
-private:
-  type_info (::std::uintptr_t id) noexcept : id { id } { }
-
-  template <class T>
-  static type_info const& cref () noexcept {
-    static ::std::uintptr_t const value { };
-    static type_info const instance { as_int(::std::addressof(value)) };
-    return instance;
-  }
-
-  ::std::uintptr_t const id;
-};
-
-#endif /* CORE_NO_RTTI */
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_TYPEINFO_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/std_backport/utility.hpp 4.4.0-1/src/mlpack/core/std_backport/utility.hpp
--- 4.3.0-2/src/mlpack/core/std_backport/utility.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/std_backport/utility.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,180 +0,0 @@
-/**
- * Copyright © 2013 - 2015 MNMLSTC
- * 
- * Licensed under the Apache License, Version 2.0 (the "License"); you may
- * not use this software except in compliance with the License. You may 
- * obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
- * 
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS, 
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
- * implied. See the License for the specific language governing 
- * permissions and limitations under the License.
- */
-#ifndef CORE_UTILITY_HPP
-#define CORE_UTILITY_HPP
-
-#include <functional>
-
-#include <cstddef>
-
-#include "type_traits.hpp"
-
-namespace core {
-inline namespace v2 {
-
-template <class T>
-constexpr T&& forward (remove_reference_t<T>& t) noexcept {
-  return static_cast<T&&>(t);
-}
-
-template <class T>
-constexpr T&& forward (remove_reference_t<T>&& t) noexcept {
-  return static_cast<T&&>(t);
-}
-
-template <class T>
-constexpr auto move (T&& t) noexcept -> decltype(
-  static_cast<remove_reference_t<T>&&>(t)
-) { return static_cast<remove_reference_t<T>&&>(t); }
-
-
-template <class T, T... I>
-using integer_sequence = meta::integer_sequence<T, I...>;
-
-template <::std::size_t... I>
-using index_sequence = integer_sequence<::std::size_t, I...>;
-
-template <class T, T N>
-using make_integer_sequence = typename meta::iota<T, N, N>::type;
-
-template <::std::size_t N>
-using make_index_sequence = make_integer_sequence<::std::size_t, N>;
-
-template <class... Ts>
-using index_sequence_for = make_index_sequence<sizeof...(Ts)>;
-
-template <::std::size_t N, class T, class... Ts>
-constexpr auto value_at (T&& value, Ts&&...) -> enable_if_t<
-  N == 0 and N < (sizeof...(Ts) + 1),
-  decltype(::core::forward<T>(value))
-> { return ::core::forward<T>(value); }
-
-template <::std::size_t N, class T, class... Ts>
-constexpr auto value_at (T&&, Ts&&... values) -> enable_if_t<
-  N != 0 and N < (sizeof...(Ts) + 1),
-  meta::get<meta::list<T, Ts...>, N>
-> { return value_at<N - 1, Ts...>(::core::forward<Ts>(values)...); }
-
-template <class Callable>
-struct scope_guard final {
-
-  static_assert(
-    ::std::is_nothrow_move_constructible<Callable>::value,
-    "Given type must be nothrow move constructible"
-  );
-
-  explicit scope_guard (Callable callable) noexcept :
-    callable { ::core::move(callable) },
-    dismissed { false }
-  { }
-
-  scope_guard (scope_guard const&) = delete;
-  scope_guard (scope_guard&&) = default;
-  scope_guard () = delete;
-  ~scope_guard () noexcept { if (not this->dismissed) { callable(); } }
-
-  scope_guard& operator = (scope_guard const&) = delete;
-  scope_guard& operator = (scope_guard&&) = default;
-
-  void dismiss () noexcept { this->dismissed = true; }
-
-private:
-  Callable callable;
-  bool dismissed;
-};
-
-template <class Callable>
-auto make_scope_guard(Callable&& callable) -> scope_guard<decay_t<Callable>> {
-  return scope_guard<decay_t<Callable>> {
-    ::core::forward<Callable>(callable)
-  };
-}
-
-template <class T, class U=T>
-T exchange (T& obj, U&& value) noexcept(
-  meta::all<
-    ::std::is_nothrow_move_constructible<T>,
-    ::std::is_nothrow_assignable<add_lvalue_reference_t<T>, U>
-  >()
-) {
-  T old = ::core::move(obj);
-  obj = ::core::forward<U>(value);
-  return old;
-}
-
-inline ::std::uintptr_t as_int (void const* ptr) noexcept {
-  return reinterpret_cast<::std::uintptr_t>(ptr);
-}
-
-template <class T>
-void const* as_void (T const* ptr) { return static_cast<void const*>(ptr); }
-
-template <class T>
-void* as_void (T* ptr) { return static_cast<void*>(ptr); }
-
-template <class T>
-void const* as_void (T const& ref) { return as_void(::std::addressof(ref)); }
-
-template <class T>
-void* as_void (T& ref) { return as_void(::std::addressof(ref)); }
-
-template <class E>
-constexpr auto as_under(E e) noexcept -> meta::when<
-  std::is_enum<E>::value,
-  underlying_type_t<E>
-> { return static_cast<underlying_type_t<E>>(e); }
-
-template <class T>
-struct capture final {
-  static_assert(::std::is_move_constructible<T>::value, "T must be movable");
-  using value_type = T;
-  using reference = add_lvalue_reference_t<value_type>;
-  using pointer = add_pointer_t<value_type>;
-
-  capture (T&& data) : data { core::move(data) } { }
-
-  capture (capture&&) = default;
-  capture (capture& that) : data { core::move(that.data) } { }
-  capture () = delete;
-
-  capture& operator = (capture const&) = delete;
-  capture& operator = (capture&&) = delete;
-
-  operator reference () const noexcept { return this->get(); }
-  reference operator * () const noexcept { return this->get(); }
-  pointer operator -> () const noexcept {
-    return ::std::addressof(this->get());
-  }
-
-  reference get () const noexcept { return this->data; }
-
-private:
-  value_type data;
-};
-
-template <class T>
-auto make_capture (remove_reference_t<T>& ref) -> capture<T> {
-  return capture<T> { core::move(ref) };
-}
-
-template <class T>
-auto make_capture (remove_reference_t<T>&& ref) -> capture<T> {
-  return capture<T> { core::move(ref) };
-}
-
-struct erased_type { };
-
-}} /* namespace core::v2 */
-
-#endif /* CORE_UTILITY_HPP */
diff -pruN 4.3.0-2/src/mlpack/core/tree/ballbound_impl.hpp 4.4.0-1/src/mlpack/core/tree/ballbound_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/ballbound_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/ballbound_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -14,7 +14,6 @@
 
 // In case it hasn't been included already.
 #include "ballbound.hpp"
-#include <mlpack/core/math/clamp.hpp>
 
 #include <string>
 
@@ -159,7 +158,7 @@ BallBound<MetricType, VecType>::MinDista
   if (radius < 0)
     return std::numeric_limits<ElemType>::max();
   else
-    return ClampNonNegative(metric->Evaluate(point, center) - radius);
+    return std::max(metric->Evaluate(point, center) - radius, (ElemType) 0.0);
 }
 
 /**
@@ -176,7 +175,7 @@ BallBound<MetricType, VecType>::MinDista
   {
     const ElemType delta = metric->Evaluate(center, other.center) - radius -
         other.radius;
-    return ClampNonNegative(delta);
+    return std::max(delta, (ElemType) 0.0);
   }
 }
 
@@ -228,8 +227,7 @@ BallBound<MetricType, VecType>::RangeDis
   else
   {
     const ElemType dist = metric->Evaluate(center, point);
-    return Range(ClampNonNegative(dist - radius),
-                                              dist + radius);
+    return Range(std::max(dist - radius, (ElemType) 0.0), dist + radius);
   }
 }
 
@@ -245,8 +243,7 @@ BallBound<MetricType, VecType>::RangeDis
   {
     const ElemType dist = metric->Evaluate(center, other.center);
     const ElemType sumradius = radius + other.radius;
-    return Range(ClampNonNegative(dist - sumradius),
-                                              dist + sumradius);
+    return Range(std::max(dist - sumradius, (ElemType) 0.0), dist + sumradius);
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/binary_space_tree.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/binary_space_tree.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/binary_space_tree.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/binary_space_tree.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -504,7 +504,7 @@ class BinarySpaceTree
   size_t& Count() { return count; }
 
   //! Store the center of the bounding region in the given vector.
-  void Center(arma::vec& center) const { bound.Center(center); }
+  void Center(arma::Col<ElemType>& center) const { bound.Center(center); }
 
  private:
   /**
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/binary_space_tree_impl.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/binary_space_tree_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/binary_space_tree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/binary_space_tree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -909,7 +909,7 @@ void BinarySpaceTree<MetricType, Statist
       splitter, maxLeafSize);
 
   // Calculate parent distances for those two nodes.
-  arma::vec center, leftCenter, rightCenter;
+  arma::Col<ElemType> center, leftCenter, rightCenter;
   Center(center);
   left->Center(leftCenter);
   right->Center(rightCenter);
@@ -977,7 +977,7 @@ SplitNode(std::vector<size_t>& oldFromNe
       oldFromNew, splitter, maxLeafSize);
 
   // Calculate parent distances for those two nodes.
-  arma::vec center, leftCenter, rightCenter;
+  arma::Col<ElemType> center, leftCenter, rightCenter;
   Center(center);
   left->Center(leftCenter);
   right->Center(rightCenter);
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -114,7 +114,7 @@ class RPTreeMaxSplit
   template<typename VecType>
   static bool AssignToLeftNode(const VecType& point, const SplitInfo& splitInfo)
   {
-    return (arma::dot(point, splitInfo.direction) <= splitInfo.splitVal);
+    return (dot(point, splitInfo.direction) <= splitInfo.splitVal);
   }
 
  private:
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split_impl.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_max_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,19 +45,21 @@ bool RPTreeMaxSplit<BoundType, MatType>:
 {
   const size_t maxNumSamples = 100;
   const size_t numSamples = std::min(maxNumSamples, count);
-  arma::uvec samples;
-
   // Get no more than numSamples distinct samples.
-  ObtainDistinctSamples(begin, begin + count, numSamples, samples);
+  arma::uvec samples;
+  if (numSamples < count)
+    samples = begin + arma::randperm(count, numSamples);
+  else
+    samples = begin + arma::linspace<arma::uvec>(0, count - 1, count);
 
   arma::Col<ElemType> values(samples.n_elem);
 
   // Find the median of scalar products of the samples and the normal vector.
   for (size_t k = 0; k < samples.n_elem; ++k)
-    values[k] = arma::dot(data.col(samples[k]), direction);
+    values[k] = dot(data.col(samples[k]), direction);
 
   const ElemType maximum = arma::max(values);
-  const ElemType minimum = arma::min(values);
+  const ElemType minimum = min(values);
   if (minimum == maximum)
     return false;
 
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,7 +16,6 @@
 #include <mlpack/prereqs.hpp>
 #include "rp_tree_max_split.hpp"
 #include <mlpack/core/tree/perform_split.hpp>
-#include <mlpack/core/math/lin_alg.hpp>
 
 namespace mlpack {
 
@@ -122,10 +121,10 @@ class RPTreeMeanSplit
   static bool AssignToLeftNode(const VecType& point, const SplitInfo& splitInfo)
   {
     if (splitInfo.meanSplit)
-      return arma::dot(point - splitInfo.mean, point - splitInfo.mean) <=
+      return dot(point - splitInfo.mean, point - splitInfo.mean) <=
           splitInfo.splitVal;
 
-    return (arma::dot(point, splitInfo.direction) <= splitInfo.splitVal);
+    return (dot(point, splitInfo.direction) <= splitInfo.splitVal);
   }
 
  private:
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split_impl.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/rp_tree_mean_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,10 +26,13 @@ bool RPTreeMeanSplit<BoundType, MatType>
 {
   const size_t maxNumSamples = 100;
   const size_t numSamples = std::min(maxNumSamples, count);
-  arma::uvec samples;
 
   // Get no more than numSamples distinct samples.
-  ObtainDistinctSamples(begin, begin + count, numSamples, samples);
+  arma::uvec samples;
+  if (numSamples < count)
+    samples = begin + arma::randperm(count, numSamples);
+  else
+    samples = begin + arma::linspace<arma::uvec>(0, count - 1, count);
 
   // Find the average distance between points.
   ElemType averageDistanceSq = GetAveragePointDistance(data, samples);
@@ -89,10 +92,10 @@ bool RPTreeMeanSplit<BoundType, MatType>
   arma::Col<ElemType> values(samples.n_elem);
 
   for (size_t k = 0; k < samples.n_elem; ++k)
-    values[k] = arma::dot(data.col(samples[k]), direction);
+    values[k] = dot(data.col(samples[k]), direction);
 
   const ElemType maximum = arma::max(values);
-  const ElemType minimum = arma::min(values);
+  const ElemType minimum = min(values);
   if (minimum == maximum)
     return false;
 
@@ -122,11 +125,11 @@ bool RPTreeMeanSplit<BoundType, MatType>
     tmp = data.col(samples[k]);
     tmp -= mean;
 
-    values[k] = arma::dot(tmp, tmp);
+    values[k] = dot(tmp, tmp);
   }
 
   const ElemType maximum = arma::max(values);
-  const ElemType minimum = arma::min(values);
+  const ElemType minimum = min(values);
   if (minimum == maximum)
     return false;
 
diff -pruN 4.3.0-2/src/mlpack/core/tree/binary_space_tree/vantage_point_split_impl.hpp 4.4.0-1/src/mlpack/core/tree/binary_space_tree/vantage_point_split_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/binary_space_tree/vantage_point_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/binary_space_tree/vantage_point_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -43,12 +43,19 @@ void VantagePointSplit<BoundType, MatTyp
 SelectVantagePoint(const MetricType& metric, const MatType& data,
     const size_t begin, const size_t count, size_t& vantagePoint, ElemType& mu)
 {
-  arma::uvec vantagePointCandidates;
   arma::Col<ElemType> distances(MaxNumSamples);
 
   // Get no more than max(MaxNumSamples, count) vantage point candidates
-  ObtainDistinctSamples(begin, begin + count, MaxNumSamples,
-      vantagePointCandidates);
+  arma::uvec vantagePointCandidates;
+  if (MaxNumSamples > count)
+  {
+    vantagePointCandidates = begin + arma::linspace<arma::uvec>(0, count - 1,
+        count);
+  }
+  else
+  {
+    vantagePointCandidates = begin + arma::randperm(count, MaxNumSamples);
+  }
 
   ElemType bestSpread = 0;
 
@@ -57,7 +64,10 @@ SelectVantagePoint(const MetricType& met
   for (size_t i = 0; i < vantagePointCandidates.n_elem; ++i)
   {
     // Get no more than min(MaxNumSamples, count) random samples
-    ObtainDistinctSamples(begin, begin + count, MaxNumSamples, samples);
+    if (MaxNumSamples > count)
+      samples = begin + arma::linspace<arma::uvec>(0, count - 1, count);
+    else
+      samples = begin + arma::randperm(count, MaxNumSamples);
 
     // Calculate the second moment of the distance to the vantage point
     // candidate using these random samples.
@@ -67,7 +77,7 @@ SelectVantagePoint(const MetricType& met
       distances[j] = metric.Evaluate(data.col(vantagePointCandidates[i]),
           data.col(samples[j]));
 
-    const ElemType spread = arma::sum(distances % distances) / samples.n_elem;
+    const ElemType spread = sum(distances % distances) / samples.n_elem;
 
     if (spread > bestSpread)
     {
diff -pruN 4.3.0-2/src/mlpack/core/tree/cellbound_impl.hpp 4.4.0-1/src/mlpack/core/tree/cellbound_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/cellbound_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/cellbound_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -504,8 +504,8 @@ inline ElemType CellBound<MetricType, El
       }
       else
       {
-        sum += pow((lower + fabs(lower)) + (higher + fabs(higher)),
-            (ElemType) MetricType::Power);
+        sum += std::pow((lower + std::fabs(lower)) +
+            (higher + std::fabs(higher)), (ElemType) MetricType::Power);
       }
 
       if (sum >= minSum)
@@ -532,10 +532,10 @@ inline ElemType CellBound<MetricType, El
   else
   {
     if (MetricType::TakeRoot)
-      return (ElemType) pow((double) minSum,
+      return (ElemType) std::pow((double) minSum,
           1.0 / (double) MetricType::Power) / 2.0;
     else
-      return minSum / pow(2.0, MetricType::Power);
+      return minSum / std::pow(2.0, MetricType::Power);
   }
 }
 
@@ -574,8 +574,8 @@ ElemType CellBound<MetricType, ElemType>
         }
         else
         {
-          sum += pow((lower + fabs(lower)) + (higher + fabs(higher)),
-              (ElemType) MetricType::Power);
+          sum += std::pow((lower + std::fabs(lower)) +
+              (higher + std::fabs(higher)), (ElemType) MetricType::Power);
         }
 
         if (sum >= minSum)
@@ -599,10 +599,10 @@ ElemType CellBound<MetricType, ElemType>
   else
   {
     if (MetricType::TakeRoot)
-      return (ElemType) pow((double) minSum,
+      return (ElemType) std::pow((double) minSum,
           1.0 / (double) MetricType::Power) / 2.0;
     else
-      return minSum / pow(2.0, MetricType::Power);
+      return minSum / std::pow(2.0, MetricType::Power);
   }
 }
 
@@ -647,7 +647,8 @@ inline ElemType CellBound<MetricType, El
     else if (MetricType::Power == 2)
       return (ElemType) std::sqrt(maxSum);
     else
-      return (ElemType) pow((double) maxSum, 1.0 / (double) MetricType::Power);
+      return (ElemType) std::pow((double) maxSum, 1.0 /
+          (double) MetricType::Power);
   }
 
   return maxSum;
@@ -696,7 +697,8 @@ inline ElemType CellBound<MetricType, El
     else if (MetricType::Power == 2)
       return (ElemType) std::sqrt(maxSum);
     else
-      return (ElemType) pow((double) maxSum, 1.0 / (double) MetricType::Power);
+      return (ElemType) std::pow((double) maxSum, 1.0 /
+          (double) MetricType::Power);
   }
 
   return maxSum;
@@ -772,8 +774,10 @@ CellBound<MetricType, ElemType>::RangeDi
     else
     {
       return RangeType<ElemType>(
-          (ElemType) pow((double) minLoSum, 1.0 / (double) MetricType::Power),
-          (ElemType) pow((double) maxHiSum, 1.0 / (double) MetricType::Power));
+          (ElemType) std::pow((double) minLoSum, 1.0 /
+            (double) MetricType::Power),
+          (ElemType) std::pow((double) maxHiSum, 1.0 /
+            (double) MetricType::Power));
     }
   }
 
@@ -858,8 +862,10 @@ CellBound<MetricType, ElemType>::RangeDi
     else
     {
       return RangeType<ElemType>(
-          (ElemType) pow((double) minLoSum, 1.0 / (double) MetricType::Power),
-          (ElemType) pow((double) maxHiSum, 1.0 / (double) MetricType::Power));
+          (ElemType) std::pow((double) minLoSum, 1.0 /
+              (double) MetricType::Power),
+          (ElemType) std::pow((double) maxHiSum, 1.0 /
+              (double) MetricType::Power));
     }
   }
 
@@ -876,7 +882,7 @@ CellBound<MetricType, ElemType>::operato
 {
   Log::Assert(data.n_rows == dim);
 
-  arma::Col<ElemType> mins(arma::min(data, 1));
+  arma::Col<ElemType> mins(min(data, 1));
   arma::Col<ElemType> maxs(arma::max(data, 1));
 
   minWidth = std::numeric_limits<ElemType>::max();
diff -pruN 4.3.0-2/src/mlpack/core/tree/cosine_tree/cosine_tree.hpp 4.4.0-1/src/mlpack/core/tree/cosine_tree/cosine_tree.hpp
--- 4.3.0-2/src/mlpack/core/tree/cosine_tree/cosine_tree.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/cosine_tree/cosine_tree.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -20,14 +20,20 @@ namespace mlpack {
 
 // Predeclare classes for CosineNodeQueue typedef.
 class CompareCosineNode;
+
+template<typename MatType>
 class CosineTree;
 
 // CosineNodeQueue typedef.
-typedef std::vector<CosineTree*> CosineNodeQueue;
+template<typename MatType = arma::mat>
+using CosineNodeQueue = std::vector<CosineTree<MatType>*>;
 
+template<typename MatType = arma::mat>
 class CosineTree
 {
  public:
+  typedef typename GetDenseColType<MatType>::type VecType;
+
   /**
    * CosineTree constructor for the root node of the tree. It initializes the
    * necessary variables required for splitting of the node, and building the
@@ -36,7 +42,7 @@ class CosineTree
    *
    * @param dataset Matrix for which cosine tree is constructed.
    */
-  CosineTree(const arma::mat& dataset);
+  CosineTree(const MatType& dataset);
 
   /**
    * CosineTree constructor for nodes other than the root node of the tree. It
@@ -63,7 +69,7 @@ class CosineTree
    * @param epsilon Error tolerance fraction for calculated subspace.
    * @param delta Cumulative probability for Monte Carlo error lower bound.
    */
-  CosineTree(const arma::mat& dataset,
+  CosineTree(const MatType& dataset,
              const double epsilon,
              const double delta);
 
@@ -110,10 +116,10 @@ class CosineTree
    * @param newBasisVector Orthonormalized centroid of the node.
    * @param addBasisVector Address to additional basis vector.
    */
-  void ModifiedGramSchmidt(CosineNodeQueue& treeQueue,
-                           arma::vec& centroid,
-                           arma::vec& newBasisVector,
-                           arma::vec* addBasisVector = NULL);
+  void ModifiedGramSchmidt(CosineNodeQueue<MatType>& treeQueue,
+                           VecType& centroid,
+                           VecType& newBasisVector,
+                           VecType* addBasisVector = NULL);
 
   /**
    * Estimates the squared error of the projection of the input node's matrix
@@ -128,16 +134,16 @@ class CosineTree
    * @param addBasisVector2 Address to second additional basis vector.
    */
   double MonteCarloError(CosineTree* node,
-                         CosineNodeQueue& treeQueue,
-                         arma::vec* addBasisVector1 = NULL,
-                         arma::vec* addBasisVector2 = NULL);
+                         CosineNodeQueue<MatType>& treeQueue,
+                         VecType* addBasisVector1 = NULL,
+                         VecType* addBasisVector2 = NULL);
 
   /**
    * Constructs the final basis matrix, after the cosine tree construction.
    *
    * @param treeQueue Priority queue of cosine nodes.
    */
-  void ConstructBasis(CosineNodeQueue& treeQueue);
+  void ConstructBasis(CosineNodeQueue<MatType>& treeQueue);
 
   /**
    * This function splits the cosine node into two children based on the cosines
@@ -153,7 +159,8 @@ class CosineTree
    * randomly generated values in the range [0, 1].
    */
   void ColumnSamplesLS(std::vector<size_t>& sampledIndices,
-                       arma::vec& probabilities, size_t numSamples);
+                       VecType& probabilities,
+                       size_t numSamples);
 
   /**
    * Sample a point from the Length-Squared distribution of the cosine node. The
@@ -175,7 +182,9 @@ class CosineTree
    * @param start Starting index of the distribution interval to search in.
    * @param end Ending index of the distribution interval to search in.
    */
-  size_t BinarySearch(arma::vec& cDistribution, double value, size_t start,
+  size_t BinarySearch(VecType& cDistribution,
+                      double value,
+                      size_t start,
                       size_t end);
 
   /**
@@ -185,7 +194,7 @@ class CosineTree
    *
    * @param cosines Vector to store the cosine values in.
    */
-  void CalculateCosines(arma::vec& cosines);
+  void CalculateCosines(VecType& cosines);
 
   /**
    * Calculate centroid of the columns present in the node. The calculated
@@ -194,10 +203,10 @@ class CosineTree
   void CalculateCentroid();
 
   //! Returns the basis of the constructed subspace.
-  void GetFinalBasis(arma::mat& finalBasis) { finalBasis = basis; }
+  void GetFinalBasis(MatType& finalBasis) { finalBasis = basis; }
 
   //! Get pointer to the dataset matrix.
-  const arma::mat& GetDataset() const { return *dataset; }
+  const MatType& GetDataset() const { return *dataset; }
 
   //! Get the indices of columns in the node.
   std::vector<size_t>& VectorIndices() { return indices; }
@@ -208,13 +217,13 @@ class CosineTree
   double L2Error() const { return l2Error; }
 
   //! Get pointer to the centroid vector.
-  arma::vec& Centroid() { return centroid; }
+  VecType& Centroid() { return centroid; }
 
   //! Set the basis vector of the node.
-  void BasisVector(arma::vec& bVector) { this->basisVector = bVector; }
+  void BasisVector(VecType& bVector) { this->basisVector = bVector; }
 
   //! Get the basis vector of the node.
-  arma::vec& BasisVector() { return basisVector; }
+  VecType& BasisVector() { return basisVector; }
 
   //! Get pointer to the parent node.
   CosineTree* Parent() const { return parent; }
@@ -242,11 +251,11 @@ class CosineTree
 
  private:
   //! Matrix for which cosine tree is constructed.
-  const arma::mat* dataset;
+  const MatType* dataset;
   //! Cumulative probability for Monte Carlo error lower bound.
   double delta;
   //! Subspace basis of the input dataset.
-  arma::mat basis;
+  MatType basis;
   //! Parent of the node.
   CosineTree* parent;
   //! Left child of the node.
@@ -256,11 +265,11 @@ class CosineTree
   //! Indices of columns of input matrix in the node.
   std::vector<size_t> indices;
   //! L2-norm squared of columns in the node.
-  arma::vec l2NormsSquared;
+  VecType l2NormsSquared;
   //! Centroid of columns of input matrix in the node.
-  arma::vec centroid;
+  VecType centroid;
   //! Orthonormalized basis vector of the node.
-  arma::vec basisVector;
+  VecType basisVector;
   //! Index of split point of cosine node.
   size_t splitPointIndex;
   //! Number of columns of input matrix in the node.
@@ -277,7 +286,9 @@ class CompareCosineNode
 {
  public:
   // Comparison function for construction of priority queue.
-  bool operator() (const CosineTree* a, const CosineTree* b) const
+  template<typename MatType>
+  bool operator() (const CosineTree<MatType>* a,
+                   const CosineTree<MatType>* b) const
   {
     return a->L2Error() < b->L2Error();
   }
diff -pruN 4.3.0-2/src/mlpack/core/tree/cosine_tree/cosine_tree_impl.hpp 4.4.0-1/src/mlpack/core/tree/cosine_tree/cosine_tree_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/cosine_tree/cosine_tree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/cosine_tree/cosine_tree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,7 +16,8 @@
 
 namespace mlpack {
 
-inline CosineTree::CosineTree(const arma::mat& dataset) :
+template<typename MatType>
+inline CosineTree<MatType>::CosineTree(const MatType& dataset) :
     dataset(&dataset),
     parent(NULL),
     left(NULL),
@@ -32,12 +33,12 @@ inline CosineTree::CosineTree(const arma
   for (size_t i = 0; i < numColumns; ++i)
   {
     indices[i] = i;
-    double l2Norm = arma::norm(dataset.col(i), 2);
+    double l2Norm = (double) arma::norm(dataset.col(i), 2);
     l2NormsSquared(i) = l2Norm * l2Norm;
   }
 
   // Frobenius norm of columns in the node.
-  frobNormSquared = arma::accu(l2NormsSquared);
+  frobNormSquared = accu(l2NormsSquared);
 
   // Calculate centroid of columns in the node.
   CalculateCentroid();
@@ -45,8 +46,9 @@ inline CosineTree::CosineTree(const arma
   splitPointIndex = ColumnSampleLS();
 }
 
-inline CosineTree::CosineTree(CosineTree& parentNode,
-                              const std::vector<size_t>& subIndices) :
+template<typename MatType>
+inline CosineTree<MatType>::CosineTree(CosineTree& parentNode,
+                                       const std::vector<size_t>& subIndices) :
     dataset(&parentNode.GetDataset()),
     parent(&parentNode),
     left(NULL),
@@ -66,7 +68,7 @@ inline CosineTree::CosineTree(CosineTree
   }
 
   // Frobenius norm of columns in the node.
-  frobNormSquared = arma::accu(l2NormsSquared);
+  frobNormSquared = accu(l2NormsSquared);
 
   // Calculate centroid of columns in the node.
   CalculateCentroid();
@@ -74,9 +76,10 @@ inline CosineTree::CosineTree(CosineTree
   splitPointIndex = ColumnSampleLS();
 }
 
-inline CosineTree::CosineTree(const arma::mat& dataset,
-                              const double epsilon,
-                              const double delta) :
+template<typename MatType>
+inline CosineTree<MatType>::CosineTree(const MatType& dataset,
+                                       const double epsilon,
+                                       const double delta) :
     dataset(&dataset),
     delta(delta),
     left(NULL),
@@ -84,15 +87,16 @@ inline CosineTree::CosineTree(const arma
     localDataset(false)
 {
   // Declare the cosine tree priority queue.
-  CosineNodeQueue treeQueue;
+  CosineNodeQueue<MatType> treeQueue;
   CompareCosineNode comp;
 
   // Define root node of the tree and add it to the queue.
   CosineTree root(dataset);
-  arma::vec tempVector = arma::zeros(dataset.n_rows);
+  VecType tempVector = arma::zeros<VecType>(dataset.n_rows);
   root.L2Error(-1.0); // We don't know what the error is.
   root.BasisVector(tempVector);
-  treeQueue.push_back(&root); // treeQueue is empty now, so we don't need to call std::push_heap here.
+  treeQueue.push_back(&root);
+  // treeQueue is empty now, so we don't need to call std::push_heap here.
 
   // Initialize Monte Carlo error estimate for comparison.
   double monteCarloError = root.FrobNormSquared();
@@ -128,7 +132,7 @@ inline CosineTree::CosineTree(const arma
     currentRight = currentNode->Right();
 
     // Calculate basis vectors of left and right children.
-    arma::vec lBasisVector, rBasisVector;
+    VecType lBasisVector, rBasisVector;
 
     ModifiedGramSchmidt(treeQueue, currentLeft->Centroid(), lBasisVector);
     ModifiedGramSchmidt(treeQueue, currentRight->Centroid(), rBasisVector,
@@ -157,9 +161,10 @@ inline CosineTree::CosineTree(const arma
 }
 
 //! Copy the given tree.
-inline CosineTree::CosineTree(const CosineTree& other) :
+template<typename MatType>
+inline CosineTree<MatType>::CosineTree(const CosineTree& other) :
     // Copy matrix, but only if we are the root.
-    dataset((other.parent == NULL) ? new arma::mat(*other.dataset) : NULL),
+    dataset((other.parent == NULL) ? new MatType(*other.dataset) : NULL),
     delta(other.delta),
     parent(NULL),
     left(NULL),
@@ -210,7 +215,9 @@ inline CosineTree::CosineTree(const Cosi
 }
 
 //! Copy assignment operator: copy the given other tree.
-inline CosineTree& CosineTree::operator=(const CosineTree& other)
+template<typename MatType>
+inline CosineTree<MatType>& CosineTree<MatType>::operator=(
+    const CosineTree& other)
 {
   // Return if it's the same tree.
   if (this == &other)
@@ -224,7 +231,7 @@ inline CosineTree& CosineTree::operator=
   delete right;
 
   // Performing a deep copy of the dataset.
-  dataset = (other.parent == NULL) ? new arma::mat(*other.dataset) : NULL;
+  dataset = (other.parent == NULL) ? new MatType(*other.dataset) : NULL;
 
   delta = other.delta;
   parent = other.Parent();
@@ -278,7 +285,8 @@ inline CosineTree& CosineTree::operator=
 }
 
 //! Move the given tree.
-inline CosineTree::CosineTree(CosineTree&& other) :
+template<typename MatType>
+inline CosineTree<MatType>::CosineTree(CosineTree&& other) :
     dataset(other.dataset),
     delta(std::move(other.delta)),
     parent(other.parent),
@@ -313,7 +321,8 @@ inline CosineTree::CosineTree(CosineTree
 }
 
 //! Move assignment operator: take ownership of the given tree.
-inline CosineTree& CosineTree::operator=(CosineTree&& other)
+template<typename MatType>
+inline CosineTree<MatType>& CosineTree<MatType>::operator=(CosineTree&& other)
 {
   // Return if it's the same tree.
   if (this == &other)
@@ -360,7 +369,8 @@ inline CosineTree& CosineTree::operator=
   return *this;
 }
 
-inline CosineTree::~CosineTree()
+template<typename MatType>
+inline CosineTree<MatType>::~CosineTree()
 {
   if (localDataset)
     delete dataset;
@@ -370,17 +380,19 @@ inline CosineTree::~CosineTree()
     delete right;
 }
 
-inline void CosineTree::ModifiedGramSchmidt(CosineNodeQueue& treeQueue,
-                                            arma::vec& centroid,
-                                            arma::vec& newBasisVector,
-                                            arma::vec* addBasisVector)
+template<typename MatType>
+inline void CosineTree<MatType>::ModifiedGramSchmidt(
+    CosineNodeQueue<MatType>& treeQueue,
+    typename CosineTree<MatType>::VecType& centroid,
+    typename CosineTree<MatType>::VecType& newBasisVector,
+    typename CosineTree<MatType>::VecType* addBasisVector)
 {
   // Set new basis vector to centroid.
   newBasisVector = centroid;
 
   // Variables for iterating throught the priority queue.
   CosineTree *currentNode;
-  CosineNodeQueue::const_iterator i = treeQueue.cbegin();
+  typename CosineNodeQueue<MatType>::const_iterator i = treeQueue.cbegin();
 
   // For every vector in the current basis, remove its projection from the
   // centroid.
@@ -388,14 +400,14 @@ inline void CosineTree::ModifiedGramSchm
   {
     currentNode = *i;
 
-    double projection = arma::dot(currentNode->BasisVector(), centroid);
+    double projection = (double) dot(currentNode->BasisVector(), centroid);
     newBasisVector -= projection * currentNode->BasisVector();
   }
 
   // If additional basis vector is passed, take it into account.
   if (addBasisVector)
   {
-    double projection = arma::dot(*addBasisVector, centroid);
+    double projection = (double) dot(*addBasisVector, centroid);
     newBasisVector -= *addBasisVector * projection;
   }
 
@@ -404,24 +416,26 @@ inline void CosineTree::ModifiedGramSchm
     newBasisVector /= arma::norm(newBasisVector, 2);
 }
 
-inline double CosineTree::MonteCarloError(CosineTree* node,
-                                          CosineNodeQueue& treeQueue,
-                                          arma::vec* addBasisVector1,
-                                          arma::vec* addBasisVector2)
+template<typename MatType>
+inline double CosineTree<MatType>::MonteCarloError(
+    CosineTree* node,
+    CosineNodeQueue<MatType>& treeQueue,
+    typename CosineTree<MatType>::VecType* addBasisVector1,
+    typename CosineTree<MatType>::VecType* addBasisVector2)
 {
   std::vector<size_t> sampledIndices;
-  arma::vec probabilities;
+  VecType probabilities;
 
   // Sample O(log m) points from the input node's distribution.
   // 'm' is the number of columns present in the node.
-  size_t numSamples = log(node->NumColumns()) + 1;
+  size_t numSamples = std::log(node->NumColumns()) + 1;
   node->ColumnSamplesLS(sampledIndices, probabilities, numSamples);
 
   // Get pointer to the original dataset.
-  const arma::mat& dataset = node->GetDataset();
+  const MatType& dataset = node->GetDataset();
 
   // Initialize weighted projection magnitudes as zeros.
-  arma::vec weightedMagnitudes;
+  VecType weightedMagnitudes;
   weightedMagnitudes.zeros(numSamples);
 
   // Set size of projection vector, depending on whether additional basis
@@ -436,11 +450,11 @@ inline double CosineTree::MonteCarloErro
   for (size_t i = 0; i < numSamples; ++i)
   {
     // Initialize projection as a vector of zeros.
-    arma::vec projection;
+    VecType projection;
     projection.zeros(projectionSize);
 
     CosineTree *currentNode;
-    CosineNodeQueue::const_iterator j = treeQueue.cbegin();
+    typename CosineNodeQueue<MatType>::const_iterator j = treeQueue.cbegin();
 
     size_t k = 0;
     // Compute the projection of the sampled vector onto the existing subspace.
@@ -448,15 +462,15 @@ inline double CosineTree::MonteCarloErro
     {
       currentNode = *j;
 
-      projection(k) = arma::dot(dataset.col(sampledIndices[i]),
+      projection(k) = dot(dataset.col(sampledIndices[i]),
                                 currentNode->BasisVector());
     }
     // If two additional vectors are passed, take their projections.
     if (addBasisVector1 && addBasisVector2)
     {
-      projection(k++) = arma::dot(dataset.col(sampledIndices[i]),
+      projection(k++) = dot(dataset.col(sampledIndices[i]),
                                   *addBasisVector1);
-      projection(k) = arma::dot(dataset.col(sampledIndices[i]),
+      projection(k) = dot(dataset.col(sampledIndices[i]),
                                 *addBasisVector2);
     }
 
@@ -488,14 +502,16 @@ inline double CosineTree::MonteCarloErro
   return (node->FrobNormSquared() - lowerBound);
 }
 
-inline void CosineTree::ConstructBasis(CosineNodeQueue& treeQueue)
+template<typename MatType>
+inline void CosineTree<MatType>::ConstructBasis(
+    CosineNodeQueue<MatType>& treeQueue)
 {
   // Initialize basis as matrix of zeros.
   basis.zeros(dataset->n_rows, treeQueue.size());
 
   // Variables for iterating through the priority queue.
   CosineTree *currentNode;
-  CosineNodeQueue::const_iterator i = treeQueue.cbegin();
+  typename CosineNodeQueue<MatType>::const_iterator i = treeQueue.cbegin();
 
   // Transfer basis vectors from the queue to the basis matrix.
   size_t j = 0;
@@ -506,7 +522,8 @@ inline void CosineTree::ConstructBasis(C
   }
 }
 
-inline void CosineTree::CosineNodeSplit()
+template<typename MatType>
+inline void CosineTree<MatType>::CosineNodeSplit()
 {
   // If less than two points, splitting does not make sense---there is nothing
   // to split.
@@ -514,13 +531,13 @@ inline void CosineTree::CosineNodeSplit(
     return;
 
   // Calculate cosines with respect to the splitting point.
-  arma::vec cosines;
+  VecType cosines;
   CalculateCosines(cosines);
 
   // Compute maximum and minimum cosine values.
   double cosineMax, cosineMin;
   cosineMax = arma::max(cosines % (cosines < 1));
-  cosineMin = arma::min(cosines);
+  cosineMin = min(cosines);
 
   std::vector<size_t> leftIndices, rightIndices;
 
@@ -543,12 +560,14 @@ inline void CosineTree::CosineNodeSplit(
   right = new CosineTree(*this, rightIndices);
 }
 
-inline void CosineTree::ColumnSamplesLS(std::vector<size_t>& sampledIndices,
-                                        arma::vec& probabilities,
-                                        size_t numSamples)
+template<typename MatType>
+inline void CosineTree<MatType>::ColumnSamplesLS(
+    std::vector<size_t>& sampledIndices,
+    typename CosineTree<MatType>::VecType& probabilities,
+    size_t numSamples)
 {
   // Initialize the cumulative distribution vector size.
-  arma::vec cDistribution;
+  VecType cDistribution;
   cDistribution.zeros(numColumns + 1);
 
   // Calculate cumulative length-squared distribution for the node.
@@ -575,7 +594,8 @@ inline void CosineTree::ColumnSamplesLS(
   }
 }
 
-inline size_t CosineTree::ColumnSampleLS()
+template<typename MatType>
+inline size_t CosineTree<MatType>::ColumnSampleLS()
 {
   // If only one element is present, there can only be one sample.
   if (numColumns < 2)
@@ -584,7 +604,7 @@ inline size_t CosineTree::ColumnSampleLS
   }
 
   // Initialize the cumulative distribution vector size.
-  arma::vec cDistribution;
+  VecType cDistribution;
   cDistribution.zeros(numColumns + 1);
 
   // Calculate cumulative length-squared distribution for the node.
@@ -602,10 +622,12 @@ inline size_t CosineTree::ColumnSampleLS
   return BinarySearch(cDistribution, randValue, start, end);
 }
 
-inline size_t CosineTree::BinarySearch(arma::vec& cDistribution,
-                                       double value,
-                                       size_t start,
-                                       size_t end)
+template<typename MatType>
+inline size_t CosineTree<MatType>::BinarySearch(
+    typename CosineTree<MatType>::VecType& cDistribution,
+    double value,
+    size_t start,
+    size_t end)
 {
   size_t pivot = (start + end) / 2;
 
@@ -630,7 +652,9 @@ inline size_t CosineTree::BinarySearch(a
   }
 }
 
-inline void CosineTree::CalculateCosines(arma::vec& cosines)
+template<typename MatType>
+inline void CosineTree<MatType>::CalculateCosines(
+    typename CosineTree<MatType>::VecType& cosines)
 {
   // Initialize cosine vector as a vector of zeros.
   cosines.zeros(numColumns);
@@ -652,7 +676,8 @@ inline void CosineTree::CalculateCosines
   }
 }
 
-inline void CosineTree::CalculateCentroid()
+template<typename MatType>
+inline void CosineTree<MatType>::CalculateCentroid()
 {
   // Initialize centroid as vector of zeros.
   centroid.zeros(dataset->n_rows);
diff -pruN 4.3.0-2/src/mlpack/core/tree/cover_tree/cover_tree_impl.hpp 4.4.0-1/src/mlpack/core/tree/cover_tree/cover_tree_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/cover_tree/cover_tree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/cover_tree/cover_tree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -122,7 +122,8 @@ CoverTree<MetricType, StatisticType, Mat
   else if (furthestDescendantDistance == 0.0)
     scale = INT_MIN + 1;
   else
-    scale = (int) ceil(log(furthestDescendantDistance) / log(base));
+    scale = (int) std::ceil(std::log(furthestDescendantDistance) /
+        std::log(base));
 
   // Initialize statistics recursively after the entire tree construction is
   // complete.
@@ -217,7 +218,8 @@ CoverTree<MetricType, StatisticType, Mat
   else if (furthestDescendantDistance == 0.0)
     scale = INT_MIN + 1;
   else
-    scale = (int) ceil(log(furthestDescendantDistance) / log(base));
+    scale = (int) std::ceil(std::log(furthestDescendantDistance) /
+        std::log(base));
 
   // Initialize statistics recursively after the entire tree construction is
   // complete.
@@ -313,7 +315,8 @@ CoverTree<MetricType, StatisticType, Mat
   else if (furthestDescendantDistance == 0.0)
     scale = INT_MIN + 1;
   else
-    scale = (int) ceil(log(furthestDescendantDistance) / log(base));
+    scale = (int) std::ceil(std::log(furthestDescendantDistance) /
+        std::log(base));
 
   // Initialize statistics recursively after the entire tree construction is
   // complete.
@@ -408,7 +411,8 @@ CoverTree<MetricType, StatisticType, Mat
   else if (furthestDescendantDistance == 0.0)
     scale = INT_MIN + 1;
   else
-    scale = (int) ceil(log(furthestDescendantDistance) / log(base));
+    scale = (int) std::ceil(std::log(furthestDescendantDistance) /
+        std::log(base));
 
   // Initialize statistics recursively after the entire tree construction is
   // complete.
@@ -1183,8 +1187,8 @@ CoverTree<MetricType, StatisticType, Mat
   }
 
   const int nextScale = std::min(scale,
-      (int) ceil(log(maxDistance) / log(base))) - 1;
-  const ElemType bound = pow(base, nextScale);
+      (int) std::ceil(std::log(maxDistance) / std::log(base))) - 1;
+  const ElemType bound = std::pow(base, nextScale);
 
   // First, make the self child.  We must split the given near set into the near
   // set and far set for the self child.
diff -pruN 4.3.0-2/src/mlpack/core/tree/hollow_ball_bound_impl.hpp 4.4.0-1/src/mlpack/core/tree/hollow_ball_bound_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/hollow_ball_bound_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/hollow_ball_bound_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -231,8 +231,8 @@ ElemType HollowBallBound<TMetricType, El
       return outerDistance; // The outer ball does not contain the point.
 
     // Check if the point is situated in the hole.
-    const ElemType innerDistance = ClampNonNegative(radii.Lo() -
-        metric->Evaluate(point, hollowCenter));
+    const ElemType innerDistance = std::max(radii.Lo() -
+        metric->Evaluate(point, hollowCenter), (ElemType) 0.0);
 
     return innerDistance;
   }
@@ -264,8 +264,9 @@ ElemType HollowBallBound<TMetricType, El
 
     // Check if the hole of the first bound contains the outer ball of the
     // second bound.
-    const ElemType innerDistance2 = ClampNonNegative(radii.Lo() -
-        metric->Evaluate(hollowCenter, other.center) - other.radii.Hi());
+    const ElemType innerDistance2 = std::max(radii.Lo() -
+        metric->Evaluate(hollowCenter, other.center) - other.radii.Hi(),
+        (ElemType) 0.0);
 
     return innerDistance2;
   }
@@ -325,8 +326,8 @@ RangeType<ElemType> HollowBallBound<TMet
     else
     {
       // Check if the point is situated in the hole.
-      range.Lo() = ClampNonNegative(radii.Lo() -
-          metric->Evaluate(point, hollowCenter));
+      range.Lo() = std::max(radii.Lo() -
+          metric->Evaluate(point, hollowCenter), (ElemType) 0.0);
     }
     range.Hi() = dist + radii.Hi();
 
@@ -362,8 +363,9 @@ RangeType<ElemType> HollowBallBound<TMet
       {
         // Check if the outer ball of the second bound is contained in the
         // hole of the first bound.
-        range.Lo() = ClampNonNegative(radii.Lo() -
-            metric->Evaluate(hollowCenter, other.center) - other.radii.Hi());
+        range.Lo() = std::max(radii.Lo() -
+            metric->Evaluate(hollowCenter, other.center) - other.radii.Hi(),
+            (ElemType) 0.0);
       }
     }
     range.Hi() = dist + radii.Hi() + other.radii.Hi();
@@ -435,8 +437,8 @@ HollowBallBound<TMetricType, ElemType>::
   if (radii.Hi() < dist + other.radii.Hi())
     radii.Hi() = dist + other.radii.Hi();
 
-  const ElemType innerDist = ClampNonNegative(other.radii.Lo() -
-      metric->Evaluate(hollowCenter, other.hollowCenter));
+  const ElemType innerDist = std::max(other.radii.Lo() -
+      metric->Evaluate(hollowCenter, other.hollowCenter), (ElemType) 0.0);
   // Check if the hole of the first bound is not contained in the hole of the
   // second bound.
   if (radii.Lo() > innerDist)
diff -pruN 4.3.0-2/src/mlpack/core/tree/hrectbound_impl.hpp 4.4.0-1/src/mlpack/core/tree/hrectbound_impl.hpp
--- 4.3.0-2/src/mlpack/core/tree/hrectbound_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/hrectbound_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -211,7 +211,7 @@ inline ElemType HRectBound<MetricType, E
     }
     else
     {
-      sum += pow((lower + fabs(lower)) + (higher + fabs(higher)),
+      sum += std::pow((lower + std::fabs(lower)) + (higher + std::fabs(higher)),
           (ElemType) MetricType::Power);
     }
   }
@@ -232,10 +232,10 @@ inline ElemType HRectBound<MetricType, E
   else
   {
     if (MetricType::TakeRoot)
-      return (ElemType) pow((double) sum,
+      return (ElemType) std::pow((double) sum,
           1.0 / (double) MetricType::Power) / 2.0;
     else
-      return sum / pow(2.0, MetricType::Power);
+      return sum / std::pow(2.0, MetricType::Power);
   }
 }
 
@@ -271,7 +271,7 @@ ElemType HRectBound<MetricType, ElemType
     }
     else
     {
-      sum += pow((lower + fabs(lower)) + (higher + fabs(higher)),
+      sum += std::pow((lower + std::fabs(lower)) + (higher + std::fabs(higher)),
           (ElemType) MetricType::Power);
     }
 
@@ -293,10 +293,10 @@ ElemType HRectBound<MetricType, ElemType
   else
   {
     if (MetricType::TakeRoot)
-      return (ElemType) pow((double) sum,
+      return (ElemType) std::pow((double) sum,
           1.0 / (double) MetricType::Power) / 2.0;
     else
-      return sum / pow(2.0, MetricType::Power);
+      return sum / std::pow(2.0, MetricType::Power);
   }
 }
 
@@ -335,7 +335,8 @@ inline ElemType HRectBound<MetricType, E
     else if (MetricType::Power == 2)
       return (ElemType) std::sqrt(sum);
     else
-      return (ElemType) pow((double) sum, 1.0 / (double) MetricType::Power);
+      return (ElemType) std::pow((double) sum, 1.0 /
+          (double) MetricType::Power);
   }
   else
     return sum;
@@ -376,7 +377,8 @@ inline ElemType HRectBound<MetricType, E
     else if (MetricType::Power == 2)
       return (ElemType) std::sqrt(sum);
     else
-      return (ElemType) pow((double) sum, 1.0 / (double) MetricType::Power);
+      return (ElemType) std::pow((double) sum, 1.0 /
+          (double) MetricType::Power);
   }
   else
     return sum;
@@ -440,8 +442,9 @@ HRectBound<MetricType, ElemType>::RangeD
     else
     {
       return RangeType<ElemType>(
-          (ElemType) pow((double) loSum, 1.0 / (double) MetricType::Power),
-          (ElemType) pow((double) hiSum, 1.0 / (double) MetricType::Power));
+          (ElemType) std::pow((double) loSum, 1.0 / (double) MetricType::Power),
+          (ElemType) std::pow((double) hiSum,
+              1.0 / (double) MetricType::Power));
     }
   }
   else
@@ -516,8 +519,9 @@ HRectBound<MetricType, ElemType>::RangeD
     else
     {
       return RangeType<ElemType>(
-          (ElemType) pow((double) loSum, 1.0 / (double) MetricType::Power),
-          (ElemType) pow((double) hiSum, 1.0 / (double) MetricType::Power));
+          (ElemType) std::pow((double) loSum, 1.0 / (double) MetricType::Power),
+          (ElemType) std::pow((double) hiSum,
+              1.0 / (double) MetricType::Power));
     }
   }
   else
diff -pruN 4.3.0-2/src/mlpack/core/tree/space_split/projection_vector.hpp 4.4.0-1/src/mlpack/core/tree/space_split/projection_vector.hpp
--- 4.3.0-2/src/mlpack/core/tree/space_split/projection_vector.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/tree/space_split/projection_vector.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -107,7 +107,7 @@ class ProjVector
    * @param vect Vector to be considered.
    */
   ProjVector(const arma::vec& vect) :
-      projVect(arma::normalise(vect))
+      projVect(normalise(vect))
   {};
 
   /**
@@ -119,7 +119,7 @@ class ProjVector
   double Project(const VecType& point,
                  typename std::enable_if_t<IsVector<VecType>::value>* = 0) const
   {
-    return arma::dot(point, projVect);
+    return dot(point, projVect);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/core/util/arma_traits.hpp 4.4.0-1/src/mlpack/core/util/arma_traits.hpp
--- 4.3.0-2/src/mlpack/core/util/arma_traits.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/arma_traits.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,6 +37,18 @@ struct IsVector
   const static bool value = false;
 };
 
+template<typename MatType>
+struct IsMatrix
+{
+  const static bool value = false;
+};
+
+template<typename CubeType>
+struct IsCube
+{
+  const static bool value = false;
+};
+
 // Commenting out the first template per case, because
 // Visual Studio doesn't like this instantiaion pattern (error C2910).
 // template<>
@@ -81,6 +93,18 @@ struct IsVector<arma::subview_row<eT> >
   const static bool value = true;
 };
 
+template<typename eT>
+struct IsMatrix<arma::Mat<eT> >
+{
+  const static bool value = true;
+};
+
+template<typename eT>
+struct IsCube<arma::Cube<eT> >
+{
+  const static bool value = true;
+};
+
 
 #if ((ARMA_VERSION_MAJOR >= 10) || \
     ((ARMA_VERSION_MAJOR == 9) && (ARMA_VERSION_MINOR >= 869)))
@@ -111,4 +135,168 @@ struct IsVector<arma::subview_row<eT> >
 
 #endif
 
+// Get the row vector type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetRowType
+{
+  typedef arma::Row<typename MatType::elem_type> type;
+};
+
+template<typename eT>
+struct GetRowType<arma::Mat<eT>>
+{
+  typedef arma::Row<eT> type;
+};
+
+template<typename eT>
+struct GetRowType<arma::SpMat<eT>>
+{
+  typedef arma::SpRow<eT> type;
+};
+
+// Get the column vector type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetColType
+{
+  typedef arma::Col<typename MatType::elem_type> type;
+};
+
+template<typename MatType>
+struct GetUColType
+{
+  typedef arma::Col<arma::uword> type;
+};
+
+template<typename eT>
+struct GetColType<arma::Mat<eT>>
+{
+  typedef arma::Col<eT> type;
+};
+
+template<typename eT>
+struct GetColType<arma::SpMat<eT>>
+{
+  typedef arma::SpCol<eT> type;
+};
+
+// Get the dense row vector type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetDenseRowType
+{
+  typedef typename GetRowType<MatType>::type type;
+};
+
+template<typename eT>
+struct GetDenseRowType<arma::SpMat<eT>>
+{
+  typedef arma::Row<eT> type;
+};
+
+// Get the dense column vector type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetDenseColType
+{
+  typedef typename GetColType<MatType>::type type;
+};
+
+template<typename eT>
+struct GetDenseColType<arma::SpMat<eT>>
+{
+  typedef arma::Col<eT> type;
+};
+
+// Get the dense matrix type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetDenseMatType
+{
+  typedef arma::Mat<typename MatType::elem_type> type;
+};
+
+template<typename MatType>
+struct GetUDenseMatType
+{
+  typedef arma::Mat<arma::uword> type;
+};
+
+template<typename eT>
+struct GetDenseMatType<arma::SpMat<eT>>
+{
+  typedef arma::Mat<eT> type;
+};
+
+// Get the cube type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetCubeType;
+
+template<typename eT>
+struct GetCubeType<arma::Mat<eT>>
+{
+  typedef arma::Cube<eT> type;
+};
+
+// Get the sparse matrix type corresponding to a given MatType.
+
+template<typename MatType>
+struct GetSparseMatType
+{
+  typedef arma::SpMat<typename MatType::elem_type> type;
+};
+
+template<typename eT>
+struct GetSparseMatType<arma::SpMat<eT>>
+{
+  typedef arma::SpMat<eT> type;
+};
+
+// Get whether or not the given type is a base matrix type (e.g. not an
+// expression).
+
+template<typename MatType>
+struct IsBaseMatType
+{
+  constexpr static bool value = false;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::Mat<eT>>
+{
+  constexpr static bool value = true;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::Col<eT>>
+{
+  constexpr static bool value = true;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::Row<eT>>
+{
+  constexpr static bool value = true;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::SpMat<eT>>
+{
+  constexpr static bool value = true;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::SpCol<eT>>
+{
+  constexpr static bool value = true;
+};
+
+template<typename eT>
+struct IsBaseMatType<arma::SpRow<eT>>
+{
+  constexpr static bool value = true;
+};
+
 #endif
diff -pruN 4.3.0-2/src/mlpack/core/util/conv_to.hpp 4.4.0-1/src/mlpack/core/util/conv_to.hpp
--- 4.3.0-2/src/mlpack/core/util/conv_to.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/conv_to.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,65 @@
+/**
+ * @file core/util/conv_to.hpp
+ * @author Marcus Edel
+ * @author Omar Shrit
+ *
+ * A simple `conv_to` wrapper that based on the data type forwards to
+ * `coot::conv_to` or `arma::conv_to`. This file is borrowed from ensmallen.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_UTIL_CONV_TO_HPP
+#define MLPACK_CORE_UTIL_CONV_TO_HPP
+
+namespace mlpack {
+
+/**
+ * A utility class that based on the data type forwards to `coot::conv_to` or
+ * `arma::conv_to`.
+ *
+ * @tparam OutputType The data type to convert to.
+ */
+template<typename OutputType>
+class ConvTo
+{
+  public:
+
+#ifdef MLPACK_HAS_COOT
+
+    /**
+    * Convert from one matrix type to another by forwarding to `coot::conv_to`.
+    *
+    * @param input The input that is converted.
+    */
+   template<typename InputType>
+   inline static OutputType From(const InputType& input,
+                                 const typename std::enable_if_t<
+                                 coot::is_coot_type<InputType>::value ||
+                                 coot::is_coot_type<OutputType>::value>* = 0)
+   {
+     return coot::conv_to<OutputType>::from(input);
+   }
+
+#endif
+
+   /**
+    * Convert from one matrix type to another by forwarding to `arma::conv_to`.
+    *
+    * @param input The input that is converted.
+    */
+   template<typename InputType>
+   inline static OutputType From(const InputType& input,
+                                 const typename std::enable_if_t<
+                                 arma::is_arma_type<InputType>::value ||
+                                 arma::is_arma_type<OutputType>::value>* = 0)
+   {
+     return arma::conv_to<OutputType>::from(input);
+   }
+};
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/util/ens_traits.hpp 4.4.0-1/src/mlpack/core/util/ens_traits.hpp
--- 4.3.0-2/src/mlpack/core/util/ens_traits.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/ens_traits.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,101 @@
+/**
+ * @file core/util/ens_traits.hpp
+ * @author Ryan Curtin
+ *
+ * This file contains utilities for SFINAE on ensmallen types (optimizers and
+ * callbacks).
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_UTIL_ENS_TRAITS
+#define MLPACK_CORE_UTIL_ENS_TRAITS
+
+#include "sfinae_utility.hpp"
+
+namespace mlpack {
+
+HAS_MEM_FUNC(Optimize, HasOptimize);
+
+// Utility struct for IsEnsOptimizer (below).  By default returns false.  (See
+// specialization below for the real logic.)
+template<typename OptimizerType,
+         typename FunctionType,
+         typename MatType,
+         bool OptimizerTypeIsClass>
+struct IsEnsOptimizerInternal
+{
+  constexpr static bool value = false;
+};
+
+/**
+ * If the given argument is an ensmallen-compatible optimizer for the given
+ * MatType and FunctionType (e.g. if it has an `Optimize()` function that can
+ * handle the given `FunctionType` and `MatType`), then the `value` member will
+ * be `true`.
+ */
+template<typename OptimizerType, typename FunctionType, typename MatType>
+struct IsEnsOptimizer
+{
+  // Dispatch to IsEnsOptimizerInternal, which will filter out when
+  // OptimizerType is a non-class.
+  constexpr static bool value = IsEnsOptimizerInternal<
+      OptimizerType,
+      FunctionType,
+      MatType,
+      std::is_class<OptimizerType>::value
+  >::value;
+};
+
+// Logic for detecting ensmallen optimizers when OptimizerType is a class.
+template<typename OptimizerType, typename FunctionType, typename MatType>
+struct IsEnsOptimizerInternal<OptimizerType, FunctionType, MatType, true>
+{
+  // If OptimizerType is a reference type, then forming the types below will
+  // fail.  So we need to strip the reference (and the const for good measure).
+  typedef typename std::remove_cv<
+      typename std::remove_reference<OptimizerType>::type>::type
+      SafeOptimizerType;
+
+  using OptimizeElemReturnForm =
+      typename MatType::elem_type(SafeOptimizerType::*)(FunctionType&,
+                                                        MatType&);
+
+  using OptimizeVoidReturnForm =
+      void(SafeOptimizerType::*)(FunctionType&, MatType&);
+
+  constexpr static bool value =
+      HasOptimize<SafeOptimizerType, OptimizeElemReturnForm>::value ||
+      HasOptimize<SafeOptimizerType, OptimizeVoidReturnForm>::value;
+};
+
+/**
+ * If the given template parameter pack could all be valid ensmallen callbacks
+ * (i.e. they are all classes), then the `value` member will be `true`.
+ *
+ * This is not a perfect check, but it is sufficient to differentiate from
+ * hyperparameters.
+ */
+template<typename... CallbackTypes>
+struct IsEnsCallbackTypes;
+
+template<typename CallbackType, typename... CallbackTypes>
+struct IsEnsCallbackTypes<CallbackType, CallbackTypes...>
+{
+  constexpr static bool value =
+      std::is_class<typename std::remove_cv<
+          typename std::remove_reference<CallbackType>::type
+      >::type>::value && IsEnsCallbackTypes<CallbackTypes...>::value;
+};
+
+template<>
+struct IsEnsCallbackTypes<>
+{
+  constexpr static bool value = true;
+};
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/util/param_data.hpp 4.4.0-1/src/mlpack/core/util/param_data.hpp
--- 4.3.0-2/src/mlpack/core/util/param_data.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/param_data.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,7 +62,7 @@ struct ParamData
   bool loaded;
   //! The actual value that is held.  If the user has passed a different type,
   //! this may be a tuple containing multiple values.
-  MLPACK_ANY value;
+  std::any value;
   //! The true name of the type, as it would be written in C++.
   std::string cppType;
 };
diff -pruN 4.3.0-2/src/mlpack/core/util/params_impl.hpp 4.4.0-1/src/mlpack/core/util/params_impl.hpp
--- 4.3.0-2/src/mlpack/core/util/params_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/params_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -106,7 +106,7 @@ T& Params::Get(const std::string& identi
   }
   else
   {
-    return *MLPACK_ANY_CAST<T>(&d.value);
+    return *std::any_cast<T>(&d.value);
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/core/util/sfinae_utility.hpp 4.4.0-1/src/mlpack/core/util/sfinae_utility.hpp
--- 4.3.0-2/src/mlpack/core/util/sfinae_utility.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/sfinae_utility.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -132,7 +132,7 @@ struct NAME
 <                                                                              \
   T,                                                                           \
   sig,                                                                         \
-  std::integral_constant<bool, mlpack::SigCheck<sig, &T::FUNC>::value> \
+  std::integral_constant<bool, mlpack::SigCheck<sig, &T::FUNC>::value>         \
 > : std::true_type {};
 
 /**
diff -pruN 4.3.0-2/src/mlpack/core/util/size_checks.hpp 4.4.0-1/src/mlpack/core/util/size_checks.hpp
--- 4.3.0-2/src/mlpack/core/util/size_checks.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/size_checks.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,13 +32,16 @@ namespace util {
  *     before size-check. Default is false.
  */
 template<typename DataType, typename LabelsType>
-inline void CheckSameSizes(const DataType& data,
-                           const LabelsType& label,
-                           const std::string& callerDescription,
-                           const std::string& addInfo = "labels",
-                           const bool& isDataTranspose = false,
-                           const bool& isLabelTranspose = false)
-{ 
+inline void CheckSameSizes(
+    const DataType& data,
+    const LabelsType& label,
+    const std::string& callerDescription,
+    const std::string& addInfo = "labels",
+    const bool& isDataTranspose = false,
+    const bool& isLabelTranspose = false,
+    const typename std::enable_if<
+        !std::is_integral<LabelsType>::value>::type* = 0)
+{
   const size_t dataPoints = (isDataTranspose == true) ? data.n_rows : data.n_cols;
   const size_t labelPoints = (isLabelTranspose == true) ? label.n_rows : label.n_cols;
   
@@ -56,11 +59,13 @@ inline void CheckSameSizes(const DataTyp
  * An overload of CheckSameSizes() where the size to be checked is known
  * previously. The second parameter is of type unsigned int.
  */
-template<typename DataType>
-inline void CheckSameSizes(const DataType& data,
-                           const size_t& size,
-                           const std::string& callerDescription,
-                           const std::string& addInfo = "labels")
+template<typename DataType, typename SizeType>
+inline void CheckSameSizes(
+    const DataType& data,
+    const SizeType& size,
+    const std::string& callerDescription,
+    const std::string& addInfo = "labels",
+    const typename std::enable_if<std::is_integral<SizeType>::value>::type* = 0)
 {
   if (data.n_cols != size)
   {
@@ -84,10 +89,12 @@ inline void CheckSameSizes(const DataTyp
  *     is "dataset"; for example, "weights" could also be used.
  */
 template<typename DataType, typename DimType>
-inline void CheckSameDimensionality(const DataType& data,
-                                    const DimType& dimension,
-                                    const std::string& callerDescription,
-                                    const std::string& addInfo = "dataset")
+inline void CheckSameDimensionality(
+    const DataType& data,
+    const DimType& dimension,
+    const std::string& callerDescription,
+    const std::string& addInfo = "dataset",
+    const typename std::enable_if<!std::is_integral<DimType>::value>::type* = 0)
 {
   if (data.n_rows != dimension.n_rows)
   {
@@ -104,11 +111,13 @@ inline void CheckSameDimensionality(cons
  * An overload of CheckSameDimensionality() where the dimension to be checked
  * is known second param is unsigned long int.
  */
-template<typename DataType>
-inline void CheckSameDimensionality(const DataType& data,
-                                    const size_t& dimension,
-                                    const std::string& callerDescription,
-                                    const std::string& addInfo = "dataset")
+template<typename DataType, typename DimType>
+inline void CheckSameDimensionality(
+    const DataType& data,
+    const DimType& dimension,
+    const std::string& callerDescription,
+    const std::string& addInfo = "dataset",
+    const typename std::enable_if<std::is_integral<DimType>::value>::type* = 0)
 {
   if (data.n_rows != dimension)
   {
diff -pruN 4.3.0-2/src/mlpack/core/util/using.hpp 4.4.0-1/src/mlpack/core/util/using.hpp
--- 4.3.0-2/src/mlpack/core/util/using.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/using.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,80 @@
+/**
+ * @file core/util/using.hpp
+ * @author Omar Shrit
+ *
+ * This is a set of `using` statements to mitigate any possible risks or
+ * conflicts with local functions. The compiler is supposed to proritise the
+ * following functions to be looked up first. This is to be considered as a
+ * replacement to the ADL solution that we had deployed earlier.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_CORE_UTIL_USING_HPP
+#define MLPACK_CORE_UTIL_USING_HPP
+
+namespace mlpack {
+
+  /* using for armadillo namespace*/
+  using arma::conv_to;
+  using arma::exp;
+  using arma::distr_param;
+  using arma::dot;
+  using arma::join_cols;
+  using arma::join_rows;
+  using arma::log;
+  using arma::min;
+  using arma::max;
+  using arma::mean;
+  using arma::norm;
+  using arma::normalise;
+  using arma::ones;
+  using arma::pow;
+  using arma::randi;
+  using arma::randn;
+  using arma::randu;
+  using arma::repmat;
+  using arma::sign;
+  using arma::sqrt;
+  using arma::square;
+  using arma::sum;
+  using arma::trans;
+  using arma::vectorise;
+  using arma::zeros;
+
+#ifdef MLPACK_HAS_COOT
+
+  /* using for bandicoot namespace*/
+  using coot::conv_to;
+  using coot::exp;
+  using coot::distr_param;
+  using coot::dot;
+  using coot::join_cols;
+  using coot::join_rows;
+  using coot::log;
+  using coot::min;
+  using coot::max;
+  using coot::mean;
+  using coot::norm;
+  using coot::normalise;
+  using coot::ones;
+  using coot::pow;
+  using coot::randi;
+  using coot::randn;
+  using coot::randu;
+  using coot::repmat;
+  using coot::sign;
+  using coot::sqrt;
+  using coot::square;
+  using coot::sum;
+  using coot::trans;
+  using coot::vectorise;
+  using coot::zeros;
+
+#endif
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/core/util/version.hpp 4.4.0-1/src/mlpack/core/util/version.hpp
--- 4.3.0-2/src/mlpack/core/util/version.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core/util/version.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,7 +17,7 @@
 // The version of mlpack.  If this is a git repository, this will be a version
 // with higher number than the most recent release.
 #define MLPACK_VERSION_MAJOR 4
-#define MLPACK_VERSION_MINOR 3
+#define MLPACK_VERSION_MINOR 4
 #define MLPACK_VERSION_PATCH 0
 
 // The name of the version (for use by --version).
diff -pruN 4.3.0-2/src/mlpack/core.hpp 4.4.0-1/src/mlpack/core.hpp
--- 4.3.0-2/src/mlpack/core.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/core.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,6 +36,9 @@
 
 // Now the core mlpack classes.
 #include <mlpack/core/util/arma_traits.hpp>
+#include <mlpack/core/util/ens_traits.hpp>
+#include <mlpack/core/util/using.hpp>
+#include <mlpack/core/util/conv_to.hpp>
 #include <mlpack/core/util/log.hpp>
 #include <mlpack/core/util/io.hpp>
 #include <mlpack/core/util/deprecated.hpp>
diff -pruN 4.3.0-2/src/mlpack/methods/CMakeLists.txt 4.4.0-1/src/mlpack/methods/CMakeLists.txt
--- 4.3.0-2/src/mlpack/methods/CMakeLists.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/CMakeLists.txt	2024-05-28 20:08:34.000000000 +0000
@@ -10,57 +10,57 @@ macro(add_all_bindings directory name ty
 endmacro()
 
 # First, define all the "regular" bindings that exist for all languages.
-add_all_bindings(approx_kfn approx_kfn "geometry")
+add_all_bindings(approx_kfn approx_kfn "Geometry")
 add_all_bindings(bayesian_linear_regression bayesian_linear_regression
-    "regression")
-add_all_bindings(cf cf "misc. / other")
-add_all_bindings(dbscan dbscan "clustering")
-add_all_bindings(decision_tree decision_tree "classification")
-add_all_bindings(det det "misc. / other")
-add_all_bindings(emst emst "geometry")
-add_all_bindings(fastmks fastmks "geometry")
-add_all_bindings(gmm gmm_train "clustering")
-add_all_bindings(gmm gmm_generate "clustering")
-add_all_bindings(gmm gmm_probability "clustering")
-add_all_bindings(hmm hmm_train "misc. / other")
-add_all_bindings(hmm hmm_generate "misc. / other")
-add_all_bindings(hmm hmm_loglik "misc. / other")
-add_all_bindings(hmm hmm_viterbi "misc. / other")
-add_all_bindings(hoeffding_trees hoeffding_tree "clustering")
-add_all_bindings(kde kde "misc. / other")
-add_all_bindings(kernel_pca kernel_pca "transformations")
-add_all_bindings(kmeans kmeans "clustering")
-add_all_bindings(lars lars "regression")
-add_all_bindings(linear_svm linear_svm "classification")
-add_all_bindings(lmnn lmnn "transformations")
+    "Regression")
+add_all_bindings(cf cf "Misc. / Other")
+add_all_bindings(dbscan dbscan "Clustering")
+add_all_bindings(decision_tree decision_tree "Classification")
+add_all_bindings(det det "Misc. / Other")
+add_all_bindings(emst emst "Geometry")
+add_all_bindings(fastmks fastmks "Geometry")
+add_all_bindings(gmm gmm_train "Clustering")
+add_all_bindings(gmm gmm_generate "Clustering")
+add_all_bindings(gmm gmm_probability "Clustering")
+add_all_bindings(hmm hmm_train "Misc. / Other")
+add_all_bindings(hmm hmm_generate "Misc. / Other")
+add_all_bindings(hmm hmm_loglik "Misc. / Other")
+add_all_bindings(hmm hmm_viterbi "Misc. / Other")
+add_all_bindings(hoeffding_trees hoeffding_tree "Clustering")
+add_all_bindings(kde kde "Misc. / Other")
+add_all_bindings(kernel_pca kernel_pca "Transformations")
+add_all_bindings(kmeans kmeans "Clustering")
+add_all_bindings(lars lars "Regression")
+add_all_bindings(linear_svm linear_svm "Classification")
+add_all_bindings(lmnn lmnn "Transformations")
 add_all_bindings(local_coordinate_coding local_coordinate_coding
-    "transformations")
-add_all_bindings(logistic_regression logistic_regression "classification")
-add_all_bindings(lsh lsh "geometry")
-add_all_bindings(mean_shift mean_shift "clustering")
-add_all_bindings(naive_bayes nbc "classification")
-add_all_bindings(nca nca "transformations")
-add_all_bindings(neighbor_search knn "geometry")
-add_all_bindings(neighbor_search kfn "geometry")
-add_all_bindings(nmf nmf "misc. / other")
-add_all_bindings(pca pca "transformations")
-add_all_bindings(perceptron perceptron "classification")
-add_all_bindings(preprocess preprocess_split "preprocessing")
-add_all_bindings(preprocess preprocess_binarize "preprocessing")
-add_all_bindings(preprocess preprocess_describe "preprocessing")
-add_all_bindings(preprocess preprocess_scale "preprocessing")
-add_all_bindings(preprocess preprocess_one_hot_encoding "preprocessing")
-add_all_bindings(radical radical "transformations")
-add_all_bindings(random_forest random_forest "classification")
-add_all_bindings(rann krann "geometry")
-add_all_bindings(softmax_regression softmax_regression "classification")
-add_all_bindings(sparse_coding sparse_coding "transformations")
+    "Transformations")
+add_all_bindings(logistic_regression logistic_regression "Classification")
+add_all_bindings(lsh lsh "Geometry")
+add_all_bindings(mean_shift mean_shift "Clustering")
+add_all_bindings(naive_bayes nbc "Classification")
+add_all_bindings(nca nca "Transformations")
+add_all_bindings(neighbor_search knn "Geometry")
+add_all_bindings(neighbor_search kfn "Geometry")
+add_all_bindings(nmf nmf "Misc. / Other")
+add_all_bindings(pca pca "Transformations")
+add_all_bindings(perceptron perceptron "Classification")
+add_all_bindings(preprocess preprocess_split "Preprocessing")
+add_all_bindings(preprocess preprocess_binarize "Preprocessing")
+add_all_bindings(preprocess preprocess_describe "Preprocessing")
+add_all_bindings(preprocess preprocess_scale "Preprocessing")
+add_all_bindings(preprocess preprocess_one_hot_encoding "Preprocessing")
+add_all_bindings(radical radical "Transformations")
+add_all_bindings(random_forest random_forest "Classification")
+add_all_bindings(rann krann "Geometry")
+add_all_bindings(softmax_regression softmax_regression "Classification")
+add_all_bindings(sparse_coding sparse_coding "Transformations")
 
 # Now, define the "special" bindings that are different somehow.
 
 # Adaboost has a Python wrapper class that encapsulates its functionality in a
 # few methods.
-add_category(adaboost "classification")
+add_category(adaboost "Classification")
 group_bindings(adaboost adaboost "train;classify;probabilities")
 add_python_wrapper(adaboost adaboost)
 
@@ -70,9 +70,9 @@ add_go_binding(adaboost adaboost)
 add_r_binding(adaboost adaboost)
 add_markdown_docs(adaboost adaboost "cli;python;julia;go;r" "python")
 
-# Linear regression has a Python wrapper class that encapsulates its
+# Linear Regression has a Python wrapper class that encapsulates its
 # functionality in a few methods.
-add_category(linear_regression "regression")
+add_category(linear_regression "Regression")
 group_bindings(linear_regression linear_regression "train;predict")
 add_python_wrapper(linear_regression linear_regression)
 
@@ -84,17 +84,17 @@ add_markdown_docs(linear_regression line
     "python")
 
 # The imputer is only defined for CLI bindings right now.
-add_category(preprocess_imputer "preprocessing")
+add_category(preprocess_imputer "Preprocessing")
 add_cli_executable(preprocess preprocess_imputer)
 add_markdown_docs(preprocess preprocess_imputer "cli" "")
 
 # The image converter is only enabled if STB is available.
 if (STB_AVAILABLE)
-  add_all_bindings(preprocess image_converter "preprocessing")
+  add_all_bindings(preprocess image_converter "Preprocessing")
 endif()
 
 # Range search provides a vector of vector of results, and this is only
 # supported for the CLI bindings.
-add_category(range_search "geometry")
+add_category(range_search "Geometry")
 add_cli_executable(range_search range_search)
 add_markdown_docs(range_search range_search "cli" "")
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost.hpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost.hpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -80,6 +80,35 @@ template<typename WeakLearnerType = Perc
 class AdaBoost
 {
  public:
+  typedef typename MatType::elem_type ElemType;
+
+  /**
+   * Create the AdaBoost object without training.  Be sure to call Train()
+   * before calling Classify()!
+   */
+  AdaBoost(const ElemType tolerance = 1e-6);
+
+  /**
+   * Construct an AdaBoost model.  Any extra parameters are used as
+   * hyperparameters for the weak learner.  These should be the last arguments
+   * to the weak learner's constructor or `Train()` function (i.e. anything
+   * after `numClasses` or `weights`).
+   *
+   * @param data Input data.
+   * @param labels Corresponding labels.
+   * @param numClasses The number of classes.
+   * @param maxIterations Number of boosting rounds.
+   * @param tolerance The tolerance for change in values of rt.
+   * @param weakLearnerParams... Any hyperparameters for the weak learner.
+   */
+  template<typename... WeakLearnerArgs>
+  AdaBoost(const MatType& data,
+           const arma::Row<size_t>& labels,
+           const size_t numClasses,
+           const size_t maxIterations = 100,
+           const ElemType tolerance = 1e-6,
+           WeakLearnerArgs&&... weakLearnerArgs);
+
   /**
    * Constructor.  This runs the AdaBoost.MH algorithm to provide a trained
    * boosting model.  This constructor takes an already-initialized weak
@@ -89,27 +118,31 @@ class AdaBoost
    * @param data Input data.
    * @param labels Corresponding labels.
    * @param numClasses The number of classes.
-   * @param iterations Number of boosting rounds.
+   * @param maxIterations Number of boosting rounds.
    * @param tolerance The tolerance for change in values of rt.
    * @param other Weak learner that has already been initialized.
    */
+  template<typename WeakLearnerInType>
+  mlpack_deprecated /* to be removed in mlpack 5.0.0 */
   AdaBoost(const MatType& data,
            const arma::Row<size_t>& labels,
            const size_t numClasses,
-           const WeakLearnerType& other,
-           const size_t iterations = 100,
-           const double tolerance = 1e-6);
-
-  /**
-   * Create the AdaBoost object without training.  Be sure to call Train()
-   * before calling Classify()!
-   */
-  AdaBoost(const double tolerance = 1e-6);
+           const WeakLearnerInType& other,
+           const size_t maxIterations = 100,
+           const ElemType tolerance = 1e-6,
+           const typename std::enable_if<
+              std::is_same<WeakLearnerType, WeakLearnerInType>::value
+           >::type* = 0);
+
+  //! Get the maximum number of weak learners allowed in the model.
+  size_t MaxIterations() const { return maxIterations; }
+  //! Modify the maximum number of weak learners allowed in the model.
+  size_t& MaxIterations() { return maxIterations; }
 
   //! Get the tolerance for stopping the optimization during training.
-  double Tolerance() const { return tolerance; }
+  ElemType Tolerance() const { return tolerance; }
   //! Modify the tolerance for stopping the optimization during training.
-  double& Tolerance() { return tolerance; }
+  ElemType& Tolerance() { return tolerance; }
 
   //! Get the number of classes this model is trained on.
   size_t NumClasses() const { return numClasses; }
@@ -118,9 +151,9 @@ class AdaBoost
   size_t WeakLearners() const { return alpha.size(); }
 
   //! Get the weights for the given weak learner.
-  double Alpha(const size_t i) const { return alpha[i]; }
+  ElemType Alpha(const size_t i) const { return alpha[i]; }
   //! Modify the weight for the given weak learner (be careful!).
-  double& Alpha(const size_t i) { return alpha[i]; }
+  ElemType& Alpha(const size_t i) { return alpha[i]; }
 
   //! Get the given weak learner.
   const WeakLearnerType& WeakLearner(const size_t i) const { return wl[i]; }
@@ -134,20 +167,76 @@ class AdaBoost
    * completely overwrite any model that has already been trained with this
    * object.
    *
+   * Default values are not used for `maxIterations` and `tolerance`; instead,
+   * multiple overloads are allowed; this is because we want to use the existing
+   * setting internal to the class, if one is not specified.
+   *
    * @param data Dataset to train on.
    * @param labels Labels for each point in the dataset.
    * @param numClasses The number of classes.
    * @param learner Learner to use for training.
-   * @param iterations Number of boosting rounds.
+   * @param maxIterations Number of boosting rounds.
    * @param tolerance The tolerance for change in values of rt.
    * @return The upper bound for training error.
+   */ 
+  template<typename WeakLearnerInType>
+  mlpack_deprecated /* to be removed in mlpack 5.0.0 */
+  ElemType Train(
+      const MatType& data,
+      const arma::Row<size_t>& labels,
+      const size_t numClasses,
+      const WeakLearnerInType& learner,
+      const std::optional<size_t> maxIterations = std::nullopt,
+      const std::optional<double> tolerance = std::nullopt,
+      // Necessary to distinguish from other overloads.
+      const typename std::enable_if<
+          std::is_same<WeakLearnerType, WeakLearnerInType>::value>::type* = 0);
+
+  /**
+   * Train AdaBoost on the given dataset, using the given parameters.  The last
+   * parameters are the hyperparameters to use for the weak learners; these are
+   * all the arguments to `WeakLearnerType::Train()` after `numClasses` and
+   * `weights`.
+   *
+   * Default values are not used for `maxIterations` and `tolerance`; instead,
+   * multiple overloads are allowed; this is because we want to use the existing
+   * setting internal to the class, if one is not specified.
+   *
+   * @param data Dataset to train on.
+   * @param labels Labels for each point in the dataset.
+   * @param numClasses The number of classes in the dataset.
+   * @param maxIterations Number of boosting rounds.
+   * @param tolerance The tolerance for change in values of rt.
+   * @param weakLearnerArgs Hyperparameters to use for each weak learner.
+   * @return The upper bound for training error.
    */
-  double Train(const MatType& data,
-               const arma::Row<size_t>& labels,
-               const size_t numClasses,
-               const WeakLearnerType& learner,
-               const size_t iterations = 100,
-               const double tolerance = 1e-6);
+  template<typename... WeakLearnerArgs>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const std::optional<size_t> maxIterations = std::nullopt,
+                 const std::optional<double> tolerance = std::nullopt,
+                 WeakLearnerArgs&&... weakLearnerArgs);
+
+  /**
+   * Classify the given test point.
+   *
+   * @param point Test point.
+   */
+  template<typename VecType>
+  size_t Classify(const VecType& point) const;
+
+  /**
+   * Classify the given test point and compute class probabilities.
+   *
+   * @param point Test point.
+   * @param prediction Will be filled with the predicted class of `point`.
+   * @param probabilities Will be filled with the class probabilities.
+   */
+  template<typename VecType>
+  void Classify(const VecType& point,
+                size_t& prediction,
+                arma::Row<ElemType>& probabilities) const;
 
   /**
    * Classify the given test points.
@@ -155,12 +244,9 @@ class AdaBoost
    * @param test Testing data.
    * @param predictedLabels Vector in which the predicted labels of the test
    *      set will be stored.
-   * @param probabilities matrix to store the predicted class probabilities for
-   *      each point in the test set.
    */
   void Classify(const MatType& test,
-                arma::Row<size_t>& predictedLabels,
-                arma::mat& probabilities);
+                arma::Row<size_t>& predictedLabels) const;
 
   /**
    * Classify the given test points.
@@ -168,9 +254,12 @@ class AdaBoost
    * @param test Testing data.
    * @param predictedLabels Vector in which the predicted labels of the test
    *      set will be stored.
+   * @param probabilities matrix to store the predicted class probabilities for
+   *      each point in the test set.
    */
   void Classify(const MatType& test,
-                arma::Row<size_t>& predictedLabels);
+                arma::Row<size_t>& predictedLabels,
+                arma::Mat<ElemType>& probabilities) const;
 
   /**
    * Serialize the AdaBoost model.
@@ -179,19 +268,36 @@ class AdaBoost
   void serialize(Archive& ar, const uint32_t /* version */);
 
  private:
+  /**
+   * Internal utility training function.  `wl` is not used if
+   * `UseExistingWeakLearner` is false.  `weakLearnerArgs` are not used if
+   * `UseExistingWeakLearner` is true.
+   */
+  template<bool UseExistingWeakLearner, typename... WeakLearnerArgs>
+  ElemType TrainInternal(const MatType& data,
+                         const arma::Row<size_t>& labels,
+                         const size_t numClasses,
+                         const WeakLearnerType& wl,
+                         WeakLearnerArgs&&... weakLearnerArgs);
+
   //! The number of classes in the model.
   size_t numClasses;
-  // The tolerance for change in rt and when to stop.
-  double tolerance;
+  //! The maximum number of weak learners allowed in the model.
+  size_t maxIterations;
+  //! The tolerance for change in rt and when to stop.
+  ElemType tolerance;
 
   //! The vector of weak learners.
   std::vector<WeakLearnerType> wl;
   //! The weights corresponding to each weak learner.
-  std::vector<double> alpha;
+  std::vector<ElemType> alpha;
 }; // class AdaBoost
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename WeakLearnerType, typename MatType),
+    (mlpack::AdaBoost<WeakLearnerType, MatType>), (1));
+
 // Include implementation.
 #include "adaboost_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost_classify_main.cpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost_classify_main.cpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost_classify_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost_classify_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -61,7 +61,6 @@ BINDING_EXAMPLE(
 
 // Classification options.
 PARAM_MATRIX_IN_REQ("test", "Test dataset.", "T");
-// PARAM_UROW_OUT("output") is deprecated and will be removed in mlpack 4.0.0.
 PARAM_UROW_OUT("predictions", "Predicted labels for the test set.", "P");
 
 // Loading/saving of a model.
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost_impl.hpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost_impl.hpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,45 +30,306 @@
 
 namespace mlpack {
 
+// Empty constructor.
+template<typename WeakLearnerType, typename MatType>
+AdaBoost<WeakLearnerType, MatType>::AdaBoost(const ElemType tolerance) :
+    numClasses(0),
+    tolerance(tolerance)
+{
+  // Nothing to do.
+}
+
 /**
- * Constructor. Currently runs the AdaBoost.MH algorithm.
+ * Constructor. Runs the AdaBoost.MH algorithm.
  *
  * @param data Input data
  * @param labels Corresponding labels
- * @param iterations Number of boosting rounds
+ * @param maxIterations Number of boosting rounds
  * @param tol Tolerance for termination of Adaboost.MH.
  * @param other Weak Learner, which has been initialized already.
  */
 template<typename WeakLearnerType, typename MatType>
+template<typename WeakLearnerInType>
 AdaBoost<WeakLearnerType, MatType>::AdaBoost(
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    const WeakLearnerType& other,
-    const size_t iterations,
-    const double tol)
+    const WeakLearnerInType& other,
+    const size_t maxIterations,
+    const typename MatType::elem_type tol,
+    const typename std::enable_if<
+        std::is_same<WeakLearnerType, WeakLearnerInType>::value>::type*) :
+    maxIterations(maxIterations),
+    tolerance(tol)
 {
-  Train(data, labels, numClasses, other, iterations, tol);
+  (void) TrainInternal<true>(data, labels, numClasses, other);
 }
 
-// Empty constructor.
+/**
+ * Constructor. Runs the AdaBoost.MH algorithm.
+ *
+ * @param data Input data
+ * @param labels Corresponding labels
+ * @param maxIterations Number of boosting rounds
+ * @param tol Tolerance for termination of Adaboost.MH.
+ * @param other Weak Learner, which has been initialized already.
+ */
 template<typename WeakLearnerType, typename MatType>
-AdaBoost<WeakLearnerType, MatType>::AdaBoost(const double tolerance) :
-    numClasses(0),
-    tolerance(tolerance)
+template<typename... WeakLearnerArgs>
+AdaBoost<WeakLearnerType, MatType>::AdaBoost(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const size_t maxIterations,
+    const typename MatType::elem_type tol,
+    WeakLearnerArgs&&... weakLearnerArgs) :
+    maxIterations(maxIterations),
+    tolerance(tol)
 {
-  // Nothing to do.
+  WeakLearnerType other; // Will not be used.
+  (void) TrainInternal<false>(data, labels, numClasses, other,
+      weakLearnerArgs...);
+}
+
+// Train AdaBoost with a given weak learner, and set the maximum number of
+// iterations and tolerance.
+template<typename WeakLearnerType, typename MatType>
+template<typename WeakLearnerInType>
+typename MatType::elem_type AdaBoost<WeakLearnerType, MatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const WeakLearnerInType& other,
+    const std::optional<size_t> maxIterations,
+    const std::optional<double> tolerance,
+    const typename std::enable_if<
+        std::is_same<WeakLearnerType, WeakLearnerInType>::value>::type*)
+{
+  if (maxIterations.has_value())
+    this->maxIterations = maxIterations.value();
+
+  if (tolerance.has_value())
+    this->tolerance = tolerance.value();
+
+  return TrainInternal<true>(data, labels, numClasses, other);
 }
 
 // Train AdaBoost.
 template<typename WeakLearnerType, typename MatType>
-double AdaBoost<WeakLearnerType, MatType>::Train(
+template<typename... WeakLearnerArgs>
+typename MatType::elem_type AdaBoost<WeakLearnerType, MatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const std::optional<size_t> maxIterations,
+    const std::optional<double> tolerance,
+    WeakLearnerArgs&&... weakLearnerArgs)
+{
+  if (maxIterations.has_value())
+    this->maxIterations = maxIterations.value();
+
+  if (tolerance.has_value())
+    this->tolerance = tolerance.value();
+
+  WeakLearnerType other; // Will not be used.
+  return TrainInternal<false>(data, labels, numClasses, other,
+      weakLearnerArgs...);
+}
+
+// Classify the given test point.
+template<typename WeakLearnerType, typename MatType>
+template<typename VecType>
+size_t AdaBoost<WeakLearnerType, MatType>::Classify(const VecType& point) const
+{
+  arma::Row<ElemType> probabilities;
+  size_t prediction;
+  Classify(point, prediction, probabilities);
+
+  return prediction;
+}
+
+// Classify the given test point and return class probabilities.
+template<typename WeakLearnerType, typename MatType>
+template<typename VecType>
+void AdaBoost<WeakLearnerType, MatType>::Classify(
+    const VecType& point,
+    size_t& prediction,
+    arma::Row<typename MatType::elem_type>& probabilities) const
+{
+  probabilities.zeros(numClasses);
+  for (size_t i = 0; i < wl.size(); ++i)
+  {
+    prediction = wl[i].Classify(point);
+    probabilities(prediction) += alpha[i];
+  }
+
+  arma::uword maxIndex = 0;
+  probabilities /= accu(probabilities);
+  probabilities.max(maxIndex);
+  prediction = (size_t) maxIndex;
+}
+
+// Classify the given test points.
+template<typename WeakLearnerType, typename MatType>
+void AdaBoost<WeakLearnerType, MatType>::Classify(
+    const MatType& test,
+    arma::Row<size_t>& predictedLabels) const
+{
+  arma::Row<size_t> tempPredictedLabels(test.n_cols);
+  arma::Mat<ElemType> probabilities;
+
+  Classify(test, predictedLabels, probabilities);
+}
+
+// Classify the given test points.
+template<typename WeakLearnerType, typename MatType>
+void AdaBoost<WeakLearnerType, MatType>::Classify(
+    const MatType& test,
+    arma::Row<size_t>& predictedLabels,
+    arma::Mat<typename MatType::elem_type>& probabilities) const
+{
+  probabilities.zeros(numClasses, test.n_cols);
+  predictedLabels.set_size(test.n_cols);
+
+  for (size_t i = 0; i < wl.size(); ++i)
+  {
+    wl[i].Classify(test, predictedLabels);
+
+    for (size_t j = 0; j < predictedLabels.n_cols; ++j)
+      probabilities(predictedLabels(j), j) += alpha[i];
+  }
+
+  arma::uword maxIndex = 0;
+
+  for (size_t i = 0; i < predictedLabels.n_cols; ++i)
+  {
+    probabilities.col(i) /= accu(probabilities.col(i));
+    probabilities.col(i).max(maxIndex);
+    predictedLabels(i) = maxIndex;
+  }
+}
+
+/**
+ * Serialize the AdaBoost model.
+ */
+template<typename WeakLearnerType, typename MatType>
+template<typename Archive>
+void AdaBoost<WeakLearnerType, MatType>::serialize(Archive& ar,
+                                                   const uint32_t version)
+{
+  // Between version 0 and 1, the maxIterations member was added, and `alpha`
+  // was switched to type arma::Row<ElemType> instead of arma::rowvec.  These
+  // require a little bit of special handling when loading older versions.
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // This is the legacy version.
+    ar(CEREAL_NVP(numClasses));
+    ar(CEREAL_NVP(tolerance));
+    ar(CEREAL_NVP(alpha));
+
+    // In earlier versions, `alpha` was a vector of doubles---but it might not
+    // be now.
+    if (std::is_same<ElemType, double>::value)
+    {
+      ar(CEREAL_NVP(alpha)); // The easy case.
+    }
+    else
+    {
+      arma::rowvec alphaTmp;
+      // Avoid CEREAL_NVP so we can specify a custom name.
+      ar(cereal::make_nvp("alpha", alphaTmp));
+      alpha.clear();
+      alpha.resize(alphaTmp.size());
+      for (size_t i = 0; i < alphaTmp.size(); ++i)
+        alpha[i] = (ElemType) alphaTmp[i];
+    }
+
+    // Now serialize each weak learner.
+    ar(CEREAL_NVP(wl));
+
+    // Attempt to set maxIterations to something reasonable.
+    maxIterations = std::max((size_t) 100, alpha.size());
+  }
+  else
+  {
+    // This is the current version.
+    // (Once there is a major version bump, we should make this version 0.)
+    ar(CEREAL_NVP(numClasses));
+    ar(CEREAL_NVP(tolerance));
+    ar(CEREAL_NVP(maxIterations));
+    ar(CEREAL_NVP(alpha));
+
+    // Now serialize each weak learner.
+    ar(CEREAL_NVP(wl));
+  }
+}
+
+template<
+    bool UseExistingWeakLearner,
+    typename MatType,
+    typename WeightsType,
+    typename WeakLearnerType,
+    typename... WeakLearnerArgs
+>
+struct WeakLearnerTrainer
+{
+  static WeakLearnerType Train(
+      const MatType& data,
+      const arma::Row<size_t>& labels,
+      const size_t numClasses,
+      const WeightsType& weights,
+      const WeakLearnerType& wl,
+      WeakLearnerArgs&&... /* weakLearnerArgs */)
+  {
+    // Use the existing weak learner to train a new one with new weights.
+    // API requirement: there is a constructor with this signature:
+    //
+    //    WeakLearnerType(const WeakLearnerType&,
+    //                    MatType& data,
+    //                    LabelsType& labels,
+    //                    const size_t numClasses,
+    //                    WeightsType& weights)
+    //
+    // This trains the new WeakLearnerType using the hyperparameters from the
+    // given WeakLearnerType.
+    return WeakLearnerType(wl, data, labels, numClasses, weights);
+  }
+};
+
+template<
+    typename MatType,
+    typename WeightsType,
+    typename WeakLearnerType,
+    typename... WeakLearnerArgs
+>
+struct WeakLearnerTrainer<
+    false, MatType, WeightsType, WeakLearnerType, WeakLearnerArgs...
+>
+{
+  static WeakLearnerType Train(
+      const MatType& data,
+      const arma::Row<size_t>& labels,
+      const size_t numClasses,
+      const WeightsType& weights,
+      const WeakLearnerType& /* wl */,
+      WeakLearnerArgs&&... weakLearnerArgs)
+  {
+    // When UseExistingWeakLearner is false, we use the given hyperparameters.
+    // (This is the preferred approach that supports more types of weak
+    // learners.)
+    return WeakLearnerType(data, labels, numClasses, weights,
+        weakLearnerArgs...);
+  }
+};
+
+template<typename WeakLearnerType, typename MatType>
+template<bool UseExistingWeakLearner, typename... WeakLearnerArgs>
+typename MatType::elem_type AdaBoost<WeakLearnerType, MatType>::TrainInternal(
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const WeakLearnerType& other,
-    const size_t iterations,
-    const double tolerance)
+    WeakLearnerArgs&&... weakLearnerArgs)
 {
   // Clear information from previous runs.
   wl.clear();
@@ -79,9 +340,9 @@ double AdaBoost<WeakLearnerType, MatType
 
   // crt is the cumulative rt value for terminating the optimization when rt is
   // changing by less than the tolerance.
-  double rt, crt = 0.0, alphat = 0.0, zt;
+  ElemType rt, crt = 0.0, alphat = 0.0, zt;
 
-  double ztProduct = 1.0;
+  ElemType ztProduct = 1.0;
 
   // To be used for prediction by the weak learner.
   arma::Row<size_t> predictedLabels(labels.n_cols);
@@ -90,22 +351,22 @@ double AdaBoost<WeakLearnerType, MatType
   MatType tempData(data);
 
   // This matrix is a helper matrix used to calculate the final hypothesis.
-  arma::mat sumFinalH = arma::zeros<arma::mat>(numClasses,
-      predictedLabels.n_cols);
+  MatType sumFinalH(numClasses, predictedLabels.n_cols);
+  sumFinalH.zeros();
 
   // Load the initial weights into a 2-D matrix.
-  const double initWeight = 1.0 / double(data.n_cols * numClasses);
-  arma::mat D(numClasses, data.n_cols);
+  const ElemType initWeight = 1.0 / ElemType(data.n_cols * numClasses);
+  MatType D(numClasses, data.n_cols);
   D.fill(initWeight);
 
   // Weights are stored in this row vector.
-  arma::rowvec weights(predictedLabels.n_cols);
+  arma::Row<ElemType> weights(predictedLabels.n_cols);
 
   // This is the final hypothesis.
   arma::Row<size_t> finalH(predictedLabels.n_cols);
 
   // Now, start the boosting rounds.
-  for (size_t i = 0; i < iterations; ++i)
+  for (size_t i = 0; i < maxIterations; ++i)
   {
     // Initialized to zero in every round.  rt is used for calculation of
     // alphat; it is the weighted error.
@@ -116,40 +377,26 @@ double AdaBoost<WeakLearnerType, MatType
     zt = 0.0;
 
     // Build the weight vectors.
-    weights = arma::sum(D);
-
-    // Use the existing weak learner to train a new one with new weights.
-    // API requirement: there is a constructor with this signature:
-    //
-    //    WeakLearnerType(const WeakLearnerType&,
-    //                    MatType& data,
-    //                    LabelsType& labels,
-    //                    const size_t numClasses,
-    //                    WeightsType& weights)
-    //
-    // This trains the new WeakLearnerType using the hyperparameters from the
-    // given WeakLearnerType.
+    weights = sum(D);
 
-    WeakLearnerType w(other, tempData, labels, numClasses, weights);
-    // There is a bug with Adaboost!  It will not use the specified
-    // hyperparameters for the decision tree because they are not properly
-    // passed to the new weak learners!  (And: it's a hard bug, because the
-    // decision tree itself doesn't even store the hyperparameters it was
-    // trained with!)
+    // This is split into a separate function, so that we can still call
+    // AdaBoost::Train() with extra hyperparameters, even when the weak learner
+    // type does not support the special constructor that takes another weak
+    // learner.
+    WeakLearnerType w = WeakLearnerTrainer<
+        UseExistingWeakLearner, MatType, arma::Row<ElemType>, WeakLearnerType,
+        WeakLearnerArgs...
+    >::Train(tempData, labels, numClasses, weights, other, weakLearnerArgs...);
 
-    // DecisionTree(DecisionTree&, MatType&, LabelsType&, size_t, WeightsType&, double = 0.0, double = 0.0, ...);
     w.Classify(tempData, predictedLabels);
 
-    // Now from predictedLabels, build ht, the weak hypothesis
-    // buildClassificationMatrix(ht, predictedLabels);
-
     // Now, calculate alpha(t) using ht.
     for (size_t j = 0; j < D.n_cols; ++j) // instead of D, ht
     {
       if (predictedLabels(j) == labels(j))
-        rt += arma::accu(D.col(j));
+        rt += accu(D.col(j));
       else
-        rt -= arma::accu(D.col(j));
+        rt -= accu(D.col(j));
     }
 
     if ((i > 0) && (std::abs(rt - crt) < tolerance))
@@ -166,9 +413,9 @@ double AdaBoost<WeakLearnerType, MatType
 
     crt = rt;
 
-    // Our goal is to find alphat which mizimizes or approximately minimizes the
+    // Our goal is to find alphat which minimizes or approximately minimizes the
     // value of Z as a function of alpha.
-    alphat = 0.5 * log((1 + rt) / (1 - rt));
+    alphat = 0.5 * std::log((1 + rt) / (1 - rt));
 
     alpha.push_back(alphat);
     wl.push_back(w);
@@ -176,7 +423,7 @@ double AdaBoost<WeakLearnerType, MatType
     // Now start modifying the weights.
     for (size_t j = 0; j < D.n_cols; ++j)
     {
-      const double expo = exp(alphat);
+      const ElemType expo = std::exp(alphat);
       if (predictedLabels(j) == labels(j))
       {
         for (size_t k = 0; k < D.n_rows; ++k)
@@ -216,74 +463,8 @@ double AdaBoost<WeakLearnerType, MatType
     // Accumulate the value of zt for the Hamming loss bound.
     ztProduct *= zt;
   }
-  return ztProduct;
-}
 
-/**
- * Classify the given test points.
- */
-template<typename WeakLearnerType, typename MatType>
-void AdaBoost<WeakLearnerType, MatType>::Classify(
-    const MatType& test,
-    arma::Row<size_t>& predictedLabels)
-{
-  arma::Row<size_t> tempPredictedLabels(test.n_cols);
-  arma::mat probabilities;
-
-  Classify(test, predictedLabels, probabilities);
-}
-
-/**
- * Classify the given test points.
- */
-template<typename WeakLearnerType, typename MatType>
-void AdaBoost<WeakLearnerType, MatType>::Classify(
-    const MatType& test,
-    arma::Row<size_t>& predictedLabels,
-    arma::mat& probabilities)
-{
-  arma::Row<size_t> tempPredictedLabels(test.n_cols);
-
-  probabilities.zeros(numClasses, test.n_cols);
-  predictedLabels.set_size(test.n_cols);
-
-  for (size_t i = 0; i < wl.size(); ++i)
-  {
-    wl[i].Classify(test, tempPredictedLabels);
-
-    for (size_t j = 0; j < tempPredictedLabels.n_cols; ++j)
-      probabilities(tempPredictedLabels(j), j) += alpha[i];
-  }
-
-  arma::uword maxIndex = 0;
-
-  for (size_t i = 0; i < predictedLabels.n_cols; ++i)
-  {
-    probabilities.col(i) /= arma::accu(probabilities.col(i));
-    probabilities.col(i).max(maxIndex);
-    predictedLabels(i) = maxIndex;
-  }
-}
-
-/**
- * Serialize the AdaBoost model.
- */
-template<typename WeakLearnerType, typename MatType>
-template<typename Archive>
-void AdaBoost<WeakLearnerType, MatType>::serialize(Archive& ar,
-                                                   const uint32_t /* version */)
-{
-  ar(CEREAL_NVP(numClasses));
-  ar(CEREAL_NVP(tolerance));
-  ar(CEREAL_NVP(alpha));
-
-  // Now serialize each weak learner.
-  if (cereal::is_loading<Archive>())
-  {
-    wl.clear();
-    wl.resize(alpha.size());
-  }
-  ar(CEREAL_NVP(wl));
+  return ztProduct;
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost_main.cpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost_main.cpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -84,14 +84,7 @@ BINDING_LONG_DESC(
     "classes for each point in the test dataset are output to the " +
     PRINT_PARAM_STRING("predictions") + " output parameter.  The AdaBoost "
     "model itself is output to the " + PRINT_PARAM_STRING("output_model") +
-    " output parameter."
-    "\n\n"
-    "Note: the following parameter is deprecated and "
-    "will be removed in mlpack 4.0.0: " + PRINT_PARAM_STRING("output") +
-    "."
-    "\n"
-    "Use " + PRINT_PARAM_STRING("predictions") + " instead of " +
-    PRINT_PARAM_STRING("output") + '.');
+    " output parameter.");
 
 // Example.
 BINDING_EXAMPLE(
@@ -113,13 +106,13 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("AdaBoost on Wikipedia", "https://en.wikipedia.org/wiki/"
-        "AdaBoost");
+    "AdaBoost");
 BINDING_SEE_ALSO("Improved boosting algorithms using confidence-rated "
-        "predictions (pdf)", "http://rob.schapire.net/papers/SchapireSi98.pdf");
+    "predictions (pdf)", "http://rob.schapire.net/papers/SchapireSi98.pdf");
 BINDING_SEE_ALSO("Perceptron", "#perceptron");
-BINDING_SEE_ALSO("Decision Stump", "#decision_stump");
-BINDING_SEE_ALSO("mlpack::adaboost::AdaBoost C++ class documentation",
-        "@src/mlpack/methods/adaboost/adaboost.hpp");
+BINDING_SEE_ALSO("Decision Trees", "#decision_tree");
+BINDING_SEE_ALSO("AdaBoost C++ class documentation",
+    "@doc/user/methods/adaboost.md");
 
 // Input for training.
 PARAM_MATRIX_IN("training", "Dataset for training AdaBoost.", "t");
@@ -127,8 +120,6 @@ PARAM_UROW_IN("labels", "Labels for the
 
 // Classification options.
 PARAM_MATRIX_IN("test", "Test dataset.", "T");
-// PARAM_UROW_OUT("output") is deprecated and will be removed in mlpack 4.0.0.
-PARAM_UROW_OUT("output", "Predicted labels for the test set.", "o");
 PARAM_UROW_OUT("predictions", "Predicted labels for the test set.", "P");
 PARAM_MATRIX_OUT("probabilities", "Predicted class probabilities for each "
     "point in the test set.", "p");
@@ -178,10 +169,9 @@ void BINDING_FUNCTION(util::Params& para
         "no task will be performed");
   }
 
-  RequireAtLeastOnePassed(params, { "output_model", "output", "predictions" },
-      false, "no results will be saved");
+  RequireAtLeastOnePassed(params, { "output_model", "predictions" }, false,
+      "no results will be saved");
 
-  // "output" will be removed in mlpack 4.0.0.
   ReportIgnoredParam(params, {{ "test", false }}, "predictions");
 
   AdaBoostModel* m;
@@ -203,7 +193,7 @@ void BINDING_FUNCTION(util::Params& para
       // Extract the labels as the last dimension of the training data.
       Log::Info << "Using the last dimension of training set as labels."
           << endl;
-      labelsIn = conv_to<Row<size_t>>::from(
+      labelsIn = ConvTo<arma::Row<size_t>>::From(
           trainingData.row(trainingData.n_rows - 1));
       trainingData.shed_row(trainingData.n_rows - 1);
     }
@@ -266,8 +256,6 @@ void BINDING_FUNCTION(util::Params& para
     data::RevertLabels(predictedLabels, m->Mappings(), results);
 
     // Save the predicted labels.
-    if (params.Has("output"))
-      params.Get<arma::Row<size_t>>("output") = results;
     if (params.Has("predictions"))
       params.Get<arma::Row<size_t>>("predictions") = std::move(results);
     if (params.Has("probabilities"))
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost_model_impl.hpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost_model_impl.hpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost_model_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost_model_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -124,15 +124,13 @@ inline void AdaBoostModel::Train(const a
   if (weakLearnerType == WeakLearnerTypes::DECISION_STUMP)
   {
     delete dsBoost;
-    ID3DecisionStump ds(data, labels, max(labels) + 1);
-    dsBoost = new AdaBoost<ID3DecisionStump>(data, labels, numClasses, ds,
+    dsBoost = new AdaBoost<ID3DecisionStump>(data, labels, numClasses,
         iterations, tolerance);
   }
   else if (weakLearnerType == WeakLearnerTypes::PERCEPTRON)
   {
     delete pBoost;
-    Perceptron<> p(data, labels, max(labels) + 1);
-    pBoost = new AdaBoost<Perceptron<>>(data, labels, numClasses, p, iterations,
+    pBoost = new AdaBoost<Perceptron<>>(data, labels, numClasses, iterations,
         tolerance);
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/adaboost/adaboost_train_main.cpp 4.4.0-1/src/mlpack/methods/adaboost/adaboost_train_main.cpp
--- 4.3.0-2/src/mlpack/methods/adaboost/adaboost_train_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/adaboost/adaboost_train_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -82,13 +82,13 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("AdaBoost on Wikipedia", "https://en.wikipedia.org/wiki/"
-        "AdaBoost");
+    "AdaBoost");
 BINDING_SEE_ALSO("Improved boosting algorithms using confidence-rated "
-        "predictions (pdf)", "http://rob.schapire.net/papers/SchapireSi98.pdf");
+    "predictions (pdf)", "http://rob.schapire.net/papers/SchapireSi98.pdf");
 BINDING_SEE_ALSO("Perceptron", "#perceptron");
-BINDING_SEE_ALSO("Decision Stump", "#decision_stump");
-BINDING_SEE_ALSO("mlpack::adaboost::AdaBoost C++ class documentation",
-        "@src/mlpack/methods/adaboost/adaboost.hpp");
+BINDING_SEE_ALSO("Decision Trees", "#decision_tree");
+BINDING_SEE_ALSO("AdaBoost C++ class documentation",
+    "@doc/user/methods/adaboost.md");
 
 // Input for training.
 PARAM_MATRIX_IN_REQ("training", "Dataset for training AdaBoost.", "t");
@@ -142,7 +142,7 @@ void BINDING_FUNCTION(util::Params& para
           << "dimensions are less than 2." << endl;
     }
 
-    labelsIn = conv_to<Row<size_t>>::from(
+    labelsIn = ConvTo<arma::Row<size_t>>::From(
         trainingData.row(trainingData.n_rows - 1));
     trainingData.shed_row(trainingData.n_rows - 1);
   }
diff -pruN 4.3.0-2/src/mlpack/methods/amf/amf.hpp 4.4.0-1/src/mlpack/methods/amf/amf.hpp
--- 4.3.0-2/src/mlpack/methods/amf/amf.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/amf.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -97,11 +97,11 @@ class AMF
    * @param H Encoding matrix to output.
    * @param r Rank r of the factorization.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   double Apply(const MatType& V,
                const size_t r,
-               arma::mat& W,
-               arma::mat& H);
+               WHMatType& W,
+               WHMatType& H);
 
   //! Access the termination policy.
   const TerminationPolicyType& TerminationPolicy() const
@@ -161,6 +161,7 @@ using SVDIncompleteIncrementalFactorizer
     SimpleResidueTermination,
     RandomAcolInitialization<>,
     SVDIncompleteIncrementalLearning>;
+
 /**
  * SVDCompleteIncrementalFactorizer factorizes given matrix V into two matrices
  * W and H by complete incremental gradient descent. SVD complete incremental
@@ -175,6 +176,27 @@ using SVDCompleteIncrementalFactorizer =
     RandomAcolInitialization<>,
     SVDCompleteIncrementalLearning<MatType>>;
 
+/**
+ * Convenience typedef: NMF can be represented by the AMF class, so for
+ * usability and discoverability we make the AMF class available under both
+ * names.
+ */
+template<typename TerminationPolicyType = SimpleResidueTermination,
+         typename InitializationRuleType = RandomAcolInitialization<>,
+         typename UpdateRuleType = NMFMultiplicativeDistanceUpdate>
+class NMF : public AMF<TerminationPolicyType,
+                       InitializationRuleType,
+                       UpdateRuleType>
+{
+ public:
+  // Mirror constructor from AMF.
+  NMF(const TerminationPolicyType& terminationPolicy = TerminationPolicyType(),
+      const InitializationRuleType& initializeRule = InitializationRuleType(),
+      const UpdateRuleType& update = UpdateRuleType()) :
+      AMF<TerminationPolicyType, InitializationRuleType, UpdateRuleType>(
+          terminationPolicy, initializeRule, update) { }
+};
+
 } // namespace mlpack
 
 // Include implementation.
diff -pruN 4.3.0-2/src/mlpack/methods/amf/amf_impl.hpp 4.4.0-1/src/mlpack/methods/amf/amf_impl.hpp
--- 4.3.0-2/src/mlpack/methods/amf/amf_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/amf_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -39,12 +39,12 @@ AMF<TerminationPolicyType, Initializatio
 template<typename TerminationPolicyType,
          typename InitializationRuleType,
          typename UpdateRuleType>
-template<typename MatType>
+template<typename MatType, typename WHMatType>
 double AMF<TerminationPolicyType, InitializationRuleType, UpdateRuleType>::
 Apply(const MatType& V,
       const size_t r,
-      arma::mat& W,
-      arma::mat& H)
+      WHMatType& W,
+      WHMatType& H)
 {
   // Initialize W and H.
   initializationRule.Initialize(V, r, W, H);
@@ -57,11 +57,15 @@ Apply(const MatType& V,
   terminationPolicy.Initialize(V);
 
   // check if termination conditions are met
+  size_t iter = 0;
   while (!terminationPolicy.IsConverged(W, H))
   {
     // Update the values of W and H based on the update rules provided.
     update.WUpdate(V, W, H);
     update.HUpdate(V, W, H);
+    ++iter;
+    Log::Debug << "AMF::Apply(): iteration " << iter
+        << "; termination condition " << terminationPolicy.Index() << ".\n";
   }
 
   // get final residue and iteration count from termination policy
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/average_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/average_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/average_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/average_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,11 +38,11 @@ class AverageInitialization
    * @param W W matrix, to be initialized.
    * @param H H matrix, to be initialized.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void Initialize(const MatType& V,
                                 const size_t r,
-                                arma::mat& W,
-                                arma::mat& H)
+                                WHMatType& W,
+                                WHMatType& H)
   {
     const size_t n = V.n_rows;
     const size_t m = V.n_cols;
@@ -61,7 +61,7 @@ class AverageInitialization
         min = *it;
     }
 
-    avgV = sqrt(((avgV / (n * m)) - min) / r);
+    avgV = std::sqrt(((avgV / (n * m)) - min) / r);
 
     // Initialize to random values.
     W.randu(n, r);
@@ -81,10 +81,10 @@ class AverageInitialization
    * with uniform random noise added.
    * @param whichMatrix If true, initialize W. Otherwise, initialize H.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void InitializeOne(const MatType& V,
                                    const size_t r,
-                                   arma::mat& M,
+                                   WHMatType& M,
                                    const bool whichMatrix = true)
   {
     const size_t n = V.n_rows;
@@ -113,7 +113,7 @@ class AverageInitialization
       // Initialize H to random values
       M.randu(r, m);
     }
-    M += sqrt(((avgV / (n * m)) - min) / r);
+    M += std::sqrt(((avgV / (n * m)) - min) / r);
   }
 
   //! Serialize the object (in this case, there is nothing to do).
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/given_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/given_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/given_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/given_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -23,6 +23,7 @@ namespace mlpack {
  * not use std::move() during the Initialize() method, so it can be reused for
  * multiple AMF objects, but will incur copies of the W and H matrices.
  */
+template<typename MatType = arma::mat>
 class GivenInitialization
 {
  public:
@@ -30,12 +31,12 @@ class GivenInitialization
   GivenInitialization() : wIsGiven(false), hIsGiven(false) { }
 
   // Initialize the GivenInitialization object with the given matrices.
-  GivenInitialization(const arma::mat& w, const arma::mat& h) :
+  GivenInitialization(const MatType& w, const MatType& h) :
     w(w), h(h), wIsGiven(true), hIsGiven(true) { }
 
   // Initialize the GivenInitialization object, taking control of the given
   // matrices.
-  GivenInitialization(const arma::mat&& w, const arma::mat&& h) :
+  GivenInitialization(const MatType&& w, const MatType&& h) :
     w(std::move(w)),
     h(std::move(h)),
     wIsGiven(true),
@@ -43,7 +44,7 @@ class GivenInitialization
   { }
 
   // Initialize either H or W with the given matrix.
-  GivenInitialization(const arma::mat& m, const bool whichMatrix = true)
+  GivenInitialization(const MatType& m, const bool whichMatrix = true)
   {
     if (whichMatrix)
     {
@@ -60,7 +61,7 @@ class GivenInitialization
   }
 
   // Initialize either H or W, taking control of the given matrix.
-  GivenInitialization(const arma::mat&& m, const bool whichMatrix = true)
+  GivenInitialization(const MatType&& m, const bool whichMatrix = true)
   {
     if (whichMatrix)
     {
@@ -84,11 +85,11 @@ class GivenInitialization
    * @param W W matrix, to be initialized to given matrix.
    * @param H H matrix, to be initialized to given matrix.
    */
-  template<typename MatType>
-  inline void Initialize(const MatType& V,
+  template<typename VMatType>
+  inline void Initialize(const VMatType& V,
                          const size_t r,
-                         arma::mat& W,
-                         arma::mat& H)
+                         MatType& W,
+                         MatType& H)
   {
     // Make sure the initial W, H matrices are given
     if (!wIsGiven)
@@ -139,10 +140,10 @@ class GivenInitialization
    * @param M W or H matrix, to be initialized to given matrix.
    * @param whichMatrix If true, initialize W. Otherwise, initialize H.
    */
-  template<typename MatType>
-  inline void InitializeOne(const MatType& V,
+  template<typename VMatType>
+  inline void InitializeOne(const VMatType& V,
                             const size_t r,
-                            arma::mat& M,
+                            MatType& M,
                             const bool whichMatrix = true)
   {
     if (whichMatrix)
@@ -207,9 +208,9 @@ class GivenInitialization
 
  private:
   //! The W matrix for initialization.
-  arma::mat w;
+  MatType w;
   //! The H matrix for initialization.
-  arma::mat h;
+  MatType h;
   //! Whether initial W is given.
   bool wIsGiven;
   //! Whether initial H is given.
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/init_rules.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/init_rules.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/init_rules.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/init_rules.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -15,6 +15,7 @@
 #include "average_init.hpp"
 #include "given_init.hpp"
 #include "merge_init.hpp"
+#include "no_init.hpp"
 #include "random_acol_init.hpp"
 #include "random_init.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/merge_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/merge_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/merge_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/merge_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,11 +45,11 @@ class MergeInitialization
    * @param W W matrix, to be initialized to given matrix.
    * @param H H matrix, to be initialized to given matrix.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline void Initialize(const MatType& V,
                          const size_t r,
-                         arma::mat& W,
-                         arma::mat& H)
+                         WHMatType& W,
+                         WHMatType& H)
   {
     wInitializationRule.InitializeOne(V, r, W);
     hInitializationRule.InitializeOne(V, r, H, false);
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/no_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/no_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/no_init.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/no_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,103 @@
+/**
+ * @file methods/amf/init_rules/no_init.hpp
+ * @author Ryan Curtin
+ *
+ * Initialization rule for alternating matrix factorization (AMF). This simple
+ * initialization leaves matrices the way they are, only checking the size.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_METHODS_AMF_INIT_RULES_NO_INIT_HPP
+#define MLPACK_METHODS_AMF_INIT_RULES_NO_INIT_HPP
+
+#include <mlpack/prereqs.hpp>
+
+namespace mlpack {
+
+/**
+ * This initialization rule for AMF does not initialize W and H, but instead
+ * simply sets them to the right size.  An exception is thrown if W or H is not
+ * the correct size.
+ */
+class NoInitialization
+{
+ public:
+  // Nothing to do in the constructor.
+  NoInitialization() { }
+
+  /**
+   * Check that W and H are the right sizes.
+   *
+   * @param V Input matrix.
+   * @param r Rank of decomposition.
+   * @param W W matrix, to be initialized to given matrix.
+   * @param H H matrix, to be initialized to given matrix.
+   */
+  template<typename MatType, typename WHMatType>
+  static inline void Initialize(const MatType& V,
+                                const size_t r,
+                                WHMatType& W,
+                                WHMatType& H)
+  {
+    // Make sure the initial W, H matrices have correct size.
+    if (W.n_rows != V.n_rows || W.n_cols != r)
+    {
+      std::ostringstream oss;
+      oss << "NoInitialization::Initialize(): W has incorrect size; expected "
+          << V.n_rows << " x " << r << ", got " << W.n_rows << " x " << W.n_cols
+          << "!";
+      throw std::invalid_argument(oss.str());
+    }
+
+    if (H.n_rows != r || H.n_cols != V.n_cols)
+    {
+      std::ostringstream oss;
+      oss << "NoInitialization::Initialize(): H has incorrect size; expected "
+          << r << " x " << V.n_cols << ", got " << H.n_rows << " x " << H.n_cols
+          << "!";
+      throw std::invalid_argument(oss.str());
+    }
+  }
+
+  // Initialize either W or H.
+  template<typename MatType, typename WHMatType>
+  static inline void InitializeOne(const MatType& V,
+                                   const size_t r,
+                                   WHMatType& M,
+                                   const bool whichMatrix = true)
+  {
+    if (whichMatrix)
+    {
+      if (M.n_rows != V.n_rows || M.n_cols != r)
+      {
+        std::ostringstream oss;
+        oss << "NoInitialization::Initialize(): W has incorrect size; expected "
+            << V.n_rows << " x " << r << ", got " << M.n_rows << " x "
+            << M.n_cols << "!";
+        throw std::invalid_argument(oss.str());
+      }
+    }
+    else
+    {
+      if (M.n_rows != r || M.n_cols != V.n_cols)
+      {
+        std::ostringstream oss;
+        oss << "NoInitialization::Initialize(): H has incorrect size; expected "
+            << r << " x " << V.n_cols << ", got " << M.n_rows << " x "
+            << M.n_cols << "!";
+        throw std::invalid_argument(oss.str());
+      }
+    }
+  }
+
+  //! Serialize the object (in this case, there is nothing to serialize).
+  template<typename Archive>
+  void serialize(Archive& ar, const uint32_t /* version */) { }
+};
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/random_acol_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/random_acol_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/random_acol_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/random_acol_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,11 +47,11 @@ class RandomAcolInitialization
   RandomAcolInitialization()
   { }
 
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void Initialize(const MatType& V,
                                 const size_t r,
-                                arma::mat& W,
-                                arma::mat& H)
+                                WHMatType& W,
+                                WHMatType& H)
   {
     const size_t n = V.n_rows;
     const size_t m = V.n_cols;
@@ -70,8 +70,7 @@ class RandomAcolInitialization
     {
       for (size_t randCol = 0; randCol < columnsToAverage; randCol++)
       {
-        // .col() does not work in this case, as of Armadillo 3.920.
-        W.unsafe_col(col) += V.col(RandInt(0, m));
+        W.col(col) += V.col(RandInt(0, m));
       }
     }
 
diff -pruN 4.3.0-2/src/mlpack/methods/amf/init_rules/random_init.hpp 4.4.0-1/src/mlpack/methods/amf/init_rules/random_init.hpp
--- 4.3.0-2/src/mlpack/methods/amf/init_rules/random_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/init_rules/random_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,11 +35,11 @@ class RandomAMFInitialization
    * @param W W matrix, to be filled with random noise.
    * @param H H matrix, to be filled with random noise.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void Initialize(const MatType& V,
                                 const size_t r,
-                                arma::mat& W,
-                                arma::mat& H)
+                                WHMatType& W,
+                                WHMatType& H)
   {
     // Simple implementation (left in the header file due to its simplicity).
     const size_t n = V.n_rows;
@@ -58,10 +58,10 @@ class RandomAMFInitialization
    * @param M W or H matrix, to be filled with random noise.
    * @param whichMatrix If true, initialize W. Otherwise, initialize H.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline void InitializeOne(const MatType& V,
                             const size_t r,
-                            arma::mat& M,
+                            WHMatType& M,
                             const bool whichMatrix = true)
   {
     // Simple implementation (left in the header file due to its simplicity).
diff -pruN 4.3.0-2/src/mlpack/methods/amf/termination_policies/complete_incremental_termination.hpp 4.4.0-1/src/mlpack/methods/amf/termination_policies/complete_incremental_termination.hpp
--- 4.3.0-2/src/mlpack/methods/amf/termination_policies/complete_incremental_termination.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/termination_policies/complete_incremental_termination.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,7 +49,7 @@ class CompleteIncrementalTermination
     tPolicy.Initialize(V);
 
     // Get the number of non-zero entries.
-    incrementalIndex = arma::accu(V != 0);
+    incrementalIndex = accu(V != 0);
     iteration = 0;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/amf/termination_policies/max_iteration_termination.hpp 4.4.0-1/src/mlpack/methods/amf/termination_policies/max_iteration_termination.hpp
--- 4.3.0-2/src/mlpack/methods/amf/termination_policies/max_iteration_termination.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/termination_policies/max_iteration_termination.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ class MaxIterationTermination
    *
    * @param maxIterations Maximum number of allowed iterations.
    */
-  MaxIterationTermination(const size_t maxIterations) :
+  MaxIterationTermination(const size_t maxIterations = 1000) :
       maxIterations(maxIterations),
       iteration(0)
   {
@@ -48,7 +48,8 @@ class MaxIterationTermination
   /**
    * Check if convergence has occurred.
    */
-  bool IsConverged(const arma::mat& /* H */, const arma::mat& /* W */)
+  template<typename MatType>
+  bool IsConverged(const MatType& /* H */, const MatType& /* W */)
   {
     // Return true if we have performed the correct number of iterations.
     return (++iteration >= maxIterations);
diff -pruN 4.3.0-2/src/mlpack/methods/amf/termination_policies/simple_residue_termination.hpp 4.4.0-1/src/mlpack/methods/amf/termination_policies/simple_residue_termination.hpp
--- 4.3.0-2/src/mlpack/methods/amf/termination_policies/simple_residue_termination.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/termination_policies/simple_residue_termination.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -72,7 +72,8 @@ class SimpleResidueTermination
    * @param W Basis matrix of output.
    * @param H Encoding matrix of output.
    */
-  bool IsConverged(arma::mat& W, arma::mat& H)
+  template<typename MatType>
+  bool IsConverged(MatType& W, MatType& H)
   {
     // Calculate the norm and compute the residue, but do it by hand, so as to
     // avoid calculating (W*H), which may be very large.
diff -pruN 4.3.0-2/src/mlpack/methods/amf/termination_policies/simple_tolerance_termination.hpp 4.4.0-1/src/mlpack/methods/amf/termination_policies/simple_tolerance_termination.hpp
--- 4.3.0-2/src/mlpack/methods/amf/termination_policies/simple_tolerance_termination.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/termination_policies/simple_tolerance_termination.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,15 +18,15 @@ namespace mlpack {
 
 /**
  * This class implements residue tolerance termination policy. Termination
- * criterion is met when increase in residue value drops below the given tolerance.
- * To accommodate spikes certain number of successive residue drops are accepted.
- * This upper imit on successive drops can be adjusted with reverseStepCount.
- * Secondary termination criterion terminates algorithm when iteration count
- * goes above the threshold.
+ * criterion is met when increase in residue value drops below the given
+ * tolerance.  To accommodate spikes certain number of successive residue drops
+ * are accepted.  This upper limit on successive drops can be adjusted with
+ * reverseStepCount.  Secondary termination criterion terminates algorithm when
+ * iteration count goes above the threshold.
  *
  * @see AMF
  */
-template <class MatType>
+template<typename MatType, typename WHMatType = arma::mat>
 class SimpleToleranceTermination
 {
  public:
@@ -43,8 +43,8 @@ class SimpleToleranceTermination
       reverseStepTolerance(reverseStepTolerance),
       reverseStepCount(0),
       isCopy(false),
-      c_indexOld(0),
-      c_index(0)
+      cIndexOld(0),
+      cIndex(0)
   { }
 
   /**
@@ -62,21 +62,19 @@ class SimpleToleranceTermination
 
     this->V = &V;
 
-    c_index = 0;
-    c_indexOld = 0;
+    cIndex = 0;
+    cIndexOld = 0;
   }
 
   /**
-   * Check if termination criterio is met.
+   * Check if termination criterion is met.
    *
    * @param W Basis matrix of output.
    * @param H Encoding matrix of output.
    */
-  bool IsConverged(arma::mat& W, arma::mat& H)
+  bool IsConverged(WHMatType& W, WHMatType& H)
   {
-    arma::mat WH;
-
-    WH = W * H;
+    WHMatType WH = W * H;
 
     // Compute residue.
     residueOld = residue;
@@ -86,23 +84,23 @@ class SimpleToleranceTermination
     size_t count = 0;
     for (size_t i = 0; i < n; ++i)
     {
-        for (size_t j = 0; j < m; ++j)
+      for (size_t j = 0; j < m; ++j)
+      {
+        double temp = 0;
+        if ((temp = (*V)(i, j)) != 0)
         {
-            double temp = 0;
-            if ((temp = (*V)(i, j)) != 0)
-            {
-                temp = (temp - WH(i, j));
-                temp = temp * temp;
-                sum += temp;
-                count++;
-            }
+          temp = (temp - WH(i, j));
+          temp = temp * temp;
+          sum += temp;
+          count++;
         }
+      }
     }
 
     residue = sum;
     if (count > 0)
       residue /= count;
-    residue = sqrt(residue);
+    residue = std::sqrt(residue);
 
     // Increment iteration count.
     iteration++;
@@ -120,8 +118,8 @@ class SimpleToleranceTermination
         this->W = W;
         this->H = H;
         // Store residue values.
-        c_index = residue;
-        c_indexOld = residueOld;
+        cIndex = residue;
+        cIndexOld = residueOld;
       }
       // Increase successive drop count.
       reverseStepCount++;
@@ -132,7 +130,7 @@ class SimpleToleranceTermination
       // Initialize successive drop count.
       reverseStepCount = 0;
       // If residue is droped below minimum scrap stored values.
-      if (residue <= c_indexOld && isCopy == true)
+      if (residue <= cIndexOld && isCopy == true)
       {
         isCopy = false;
       }
@@ -147,7 +145,7 @@ class SimpleToleranceTermination
       {
         W = this->W;
         H = this->H;
-        residue = c_index;
+        residue = cIndex;
       }
       return true;
     }
@@ -195,10 +193,10 @@ class SimpleToleranceTermination
   bool isCopy;
 
   //! Variables to store information of minimum residue poi.
-  arma::mat W;
-  arma::mat H;
-  double c_indexOld;
-  double c_index;
+  WHMatType W;
+  WHMatType H;
+  double cIndexOld;
+  double cIndex;
 }; // class SimpleToleranceTermination
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/amf/termination_policies/validation_rmse_termination.hpp 4.4.0-1/src/mlpack/methods/amf/termination_policies/validation_rmse_termination.hpp
--- 4.3.0-2/src/mlpack/methods/amf/termination_policies/validation_rmse_termination.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/termination_policies/validation_rmse_termination.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,19 +18,21 @@ namespace mlpack {
 
 /**
  * This class implements validation termination policy based on RMSE index.
- * The input data matrix is divided into 2 sets, training set and validation set.
+ * The input data matrix is divided into 2 sets, training set and validation
+ * set.
+ *
  * Entries of validation set are nullifed in the input matrix. Termination
  * criterion is met when increase in validation set RMSe value drops below the
- * given tolerance. To accommodate spikes certain number of successive validation
- * RMSE drops are accepted. This upper imit on successive drops can be adjusted
- * with reverseStepCount. Secondary termination criterion terminates algorithm
- * when iteration count goes above the threshold.
+ * given tolerance. To accommodate spikes certain number of successive
+ * validation RMSE drops are accepted. This upper imit on successive drops can
+ * be adjusted with reverseStepCount. Secondary termination criterion terminates
+ * algorithm when iteration count goes above the threshold.
  *
  * @note The input matrix is modified by this termination policy.
  *
  * @see AMF
  */
-template <class MatType>
+template<typename MatType, typename WHMatType = arma::mat>
 class ValidationRMSETermination
 {
  public:
@@ -39,29 +41,29 @@ class ValidationRMSETermination
    * set in data matrix(training set).
    *
    * @param V Input matrix to be factorized.
-   * @param num_test_points number of validation test points
+   * @param numTestPoints number of validation test points
    * @param tolerance the tolerance value to compare RMSe against
    * @param maxIterations max iteration count before termination
    * @param reverseStepTolerance max successive RMSE drops allowed
    */
   ValidationRMSETermination(MatType& V,
-                            size_t num_test_points,
+                            size_t numTestPoints,
                             double tolerance = 1e-5,
                             size_t maxIterations = 10000,
                             size_t reverseStepTolerance = 3)
         : tolerance(tolerance),
           maxIterations(maxIterations),
-          num_test_points(num_test_points),
+          numTestPoints(numTestPoints),
           reverseStepTolerance(reverseStepTolerance)
   {
     size_t n = V.n_rows;
     size_t m = V.n_cols;
 
     // initialize validation set matrix
-    test_points.zeros(num_test_points, 3);
+    testPoints.zeros(numTestPoints, 3);
 
     // fill validation set matrix with random chosen entries
-    for (size_t i = 0; i < num_test_points; ++i)
+    for (size_t i = 0; i < numTestPoints; ++i)
     {
       double t_val;
       size_t t_row;
@@ -75,9 +77,9 @@ class ValidationRMSETermination
       } while ((t_val = V(t_row, t_col)) == 0);
 
       // add the entry to the validation set
-      test_points(i, 0) = t_row;
-      test_points(i, 1) = t_col;
-      test_points(i, 2) = t_val;
+      testPoints(i, 0) = t_row;
+      testPoints(i, 1) = t_col;
+      testPoints(i, 2) = t_val;
 
       // nullify the added entry from data matrix (training set)
       V(t_row, t_col) = 0;
@@ -96,8 +98,8 @@ class ValidationRMSETermination
     rmse = DBL_MAX;
     rmseOld = DBL_MAX;
 
-    c_index = 0;
-    c_indexOld = 0;
+    cIndex = 0;
+    cIndexOld = 0;
 
     reverseStepCount = 0;
     isCopy = false;
@@ -109,28 +111,26 @@ class ValidationRMSETermination
    * @param W Basis matrix of output.
    * @param H Encoding matrix of output.
    */
-  bool IsConverged(arma::mat& W, arma::mat& H)
+  bool IsConverged(WHMatType& W, WHMatType& H)
   {
-    arma::mat WH;
-
-    WH = W * H;
+    WHMatType WH = W * H;
 
     // compute validation RMSE
     if (iteration != 0)
     {
       rmseOld = rmse;
       rmse = 0;
-      for (size_t i = 0; i < num_test_points; ++i)
+      for (size_t i = 0; i < numTestPoints; ++i)
       {
-        size_t t_row = test_points(i, 0);
-        size_t t_col = test_points(i, 1);
-        double t_val = test_points(i, 2);
+        size_t t_row = testPoints(i, 0);
+        size_t t_col = testPoints(i, 1);
+        double t_val = testPoints(i, 2);
         double temp = (t_val - WH(t_row, t_col));
         temp *= temp;
         rmse += temp;
       }
-      rmse /= num_test_points;
-      rmse = sqrt(rmse);
+      rmse /= numTestPoints;
+      rmse = std::sqrt(rmse);
     }
 
     // increment iteration count
@@ -147,8 +147,8 @@ class ValidationRMSETermination
         this->W = W;
         this->H = H;
         // store residue values
-        c_indexOld = rmseOld;
-        c_index = rmse;
+        cIndexOld = rmseOld;
+        cIndex = rmse;
       }
       // increase successive drop count
       reverseStepCount++;
@@ -159,7 +159,7 @@ class ValidationRMSETermination
       // initialize successive drop count
       reverseStepCount = 0;
       // if residue is droped below minimum scrap stored values
-      if (rmse <= c_indexOld && isCopy == true)
+      if (rmse <= cIndexOld && isCopy == true)
       {
         isCopy = false;
       }
@@ -174,7 +174,7 @@ class ValidationRMSETermination
       {
         W = this->W;
         H = this->H;
-        rmse = c_index;
+        rmse = cIndex;
       }
       return true;
     }
@@ -183,12 +183,13 @@ class ValidationRMSETermination
 
   //! Get current value of residue
   const double& Index() const { return rmse; }
+  const double& RMSE() const { return rmse; }
 
   //! Get current iteration count
   const size_t& Iteration() const { return iteration; }
 
   //! Get number of validation points
-  const size_t& NumTestPoints() const { return num_test_points; }
+  const size_t& NumTestPoints() const { return numTestPoints; }
 
   //! Access upper limit of iteration count
   const size_t& MaxIterations() const { return maxIterations; }
@@ -204,13 +205,13 @@ class ValidationRMSETermination
   //! max iteration limit
   size_t maxIterations;
   //! number of validation test points
-  size_t num_test_points;
+  size_t numTestPoints;
 
   //! current iteration count
   size_t iteration;
 
   //! validation point matrix
-  arma::mat test_points;
+  WHMatType testPoints;
 
   //! rmse values
   double rmseOld;
@@ -226,10 +227,10 @@ class ValidationRMSETermination
   bool isCopy;
 
   //! variables to store information of minimum residue point
-  arma::mat W;
-  arma::mat H;
-  double c_indexOld;
-  double c_index;
+  WHMatType W;
+  WHMatType H;
+  double cIndexOld;
+  double cIndex;
 }; // class ValidationRMSETermination
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_als.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_als.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_als.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_als.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,10 +67,10 @@ class NMFALSUpdate
    * @param W Basis matrix to be updated.
    * @param H Encoding matrix.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void WUpdate(const MatType& V,
-                             arma::mat& W,
-                             const arma::mat& H)
+                             WHMatType& W,
+                             const WHMatType& H)
   {
     // The call to inv() sometimes fails; so we are using the psuedoinverse.
     // W = (inv(H * H.t()) * H * V.t()).t();
@@ -100,10 +100,10 @@ class NMFALSUpdate
    * @param W Basis matrix.
    * @param H Encoding matrix to be updated.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void HUpdate(const MatType& V,
-                             const arma::mat& W,
-                             arma::mat& H)
+                             const WHMatType& W,
+                             WHMatType& H)
   {
     H = pinv(W.t() * W) * W.t() * V;
 
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_mult_dist.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_mult_dist.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_mult_dist.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_mult_dist.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,12 +65,13 @@ class NMFMultiplicativeDistanceUpdate
    * @param W Basis matrix to be updated.
    * @param H Encoding matrix.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void WUpdate(const MatType& V,
-                             arma::mat& W,
-                             const arma::mat& H)
+                             WHMatType& W,
+                             const WHMatType& H)
   {
-    W = (W % (V * H.t())) / (W * H * H.t());
+    WHMatType zz = W * H * H.t();
+    W = (W % (V * H.t())) / (W * H * H.t() + 1e-15);
   }
 
   /**
@@ -87,12 +88,13 @@ class NMFMultiplicativeDistanceUpdate
    * @param W Basis matrix.
    * @param H Encoding matrix to be updated.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void HUpdate(const MatType& V,
-                             const arma::mat& W,
-                             arma::mat& H)
+                             const WHMatType& W,
+                             WHMatType& H)
   {
-    H = (H % (W.t() * V)) / (W.t() * W * H);
+    WHMatType zz = W.t() * W * H;
+    H = (H % (W.t() * V)) / (W.t() * W * H + 1e-15);
   }
 
   //! Serialize the object (in this case, there is nothing to serialize).
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_mult_div.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_mult_div.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/nmf_mult_div.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/nmf_mult_div.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,33 +75,14 @@ class NMFMultiplicativeDivergenceUpdate
    * @param W Basis matrix to be updated.
    * @param H Encoding matrix.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void WUpdate(const MatType& V,
-                             arma::mat& W,
-                             const arma::mat& H)
+                             WHMatType& W,
+                             const WHMatType& H)
   {
     // Simple implementation left in the header file.
-    arma::mat t1;
-    arma::rowvec t2;
-
-    t1 = W * H;
-    for (size_t i = 0; i < W.n_rows; ++i)
-    {
-      for (size_t j = 0; j < W.n_cols; ++j)
-      {
-        // Writing this as a single expression does not work as of Armadillo
-        // 3.920.  This should be fixed in a future release, and then the code
-        // below can be fixed.
-        // t2 = H.row(j) % V.row(i) / t1.row(i);
-        t2.set_size(H.n_cols);
-        for (size_t k = 0; k < t2.n_elem; ++k)
-        {
-          t2(k) = H(j, k) * V(i, k) / t1(i, k);
-        }
-
-        W(i, j) = W(i, j) * sum(t2) / sum(H.row(j));
-      }
-    }
+    W %= ((V / (W * H + 1e-15)) * H.t()) /
+        (repmat(sum(H, 1).t(), W.n_rows, 1) + 1e-15);
   }
 
   /**
@@ -119,33 +100,14 @@ class NMFMultiplicativeDivergenceUpdate
    * @param W Basis matrix.
    * @param H Encoding matrix to updated.
    */
-  template<typename MatType>
+  template<typename MatType, typename WHMatType>
   inline static void HUpdate(const MatType& V,
-                            const arma::mat& W,
-                            arma::mat& H)
+                             const WHMatType& W,
+                             WHMatType& H)
   {
     // Simple implementation left in the header file.
-    arma::mat t1;
-    arma::colvec t2;
-
-    t1 = W * H;
-    for (size_t i = 0; i < H.n_rows; ++i)
-    {
-      for (size_t j = 0; j < H.n_cols; ++j)
-      {
-        // Writing this as a single expression does not work as of Armadillo
-        // 3.920.  This should be fixed in a future release, and then the code
-        // below can be fixed.
-        // t2 = W.col(i) % V.col(j) / t1.col(j);
-        t2.set_size(W.n_rows);
-        for (size_t k = 0; k < t2.n_elem; ++k)
-        {
-          t2(k) = W(k, i) * V(k, j) / t1(k, j);
-        }
-
-        H(i, j) = H(i, j) * sum(t2) / sum(W.col(i));
-      }
-    }
+    H %= (W.t() * (V / (W * H + 1e-15))) /
+        (repmat(sum(W, 0).t(), 1, H.n_cols) + 1e-15);
   }
 
   //! Serialize the object (in this case, there is nothing to serialize).
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_batch_learning.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_batch_learning.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_batch_learning.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_batch_learning.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -105,8 +105,8 @@ class SVDBatchLearning
       {
         const double val = V(i, j);
         if (val != 0)
-          deltaW.row(i) += (val - arma::dot(W.row(i), H.col(j))) *
-                                            arma::trans(H.col(j));
+          deltaW.row(i) += (val - dot(W.row(i), H.col(j))) *
+                                            trans(H.col(j));
       }
       // Add regularization.
       if (kw != 0)
@@ -150,7 +150,7 @@ class SVDBatchLearning
       {
         const double val = V(i, j);
         if (val != 0)
-          deltaH.col(j) += (val - arma::dot(W.row(i), H.col(j))) * W.row(i).t();
+          deltaH.col(j) += (val - dot(W.row(i), H.col(j))) * W.row(i).t();
       }
       // Add regularization.
       if (kh != 0)
@@ -214,8 +214,8 @@ inline void SVDBatchLearning::WUpdate<ar
   {
     const size_t row = it.row();
     const size_t col = it.col();
-    deltaW.row(it.row()) += (*it - arma::dot(W.row(row), H.col(col))) *
-                                             arma::trans(H.col(col));
+    deltaW.row(it.row()) += (*it - dot(W.row(row), H.col(col))) *
+                                             trans(H.col(col));
   }
 
   if (kw != 0)
@@ -242,7 +242,7 @@ inline void SVDBatchLearning::HUpdate<ar
   {
     const size_t row = it.row();
     const size_t col = it.col();
-    deltaH.col(col) += (*it - arma::dot(W.row(row), H.col(col))) *
+    deltaH.col(col) += (*it - dot(W.row(row), H.col(col))) *
         W.row(row).t();
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_complete_incremental_learning.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_complete_incremental_learning.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_complete_incremental_learning.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_complete_incremental_learning.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -95,7 +95,7 @@ class SVDCompleteIncrementalLearning
       // Update feature vector if current entry is non-zero and break the loop.
       if (val != 0)
       {
-        deltaW += (val - arma::dot(W.row(currentItemIndex),
+        deltaW += (val - dot(W.row(currentItemIndex),
             H.col(currentUserIndex))) * H.col(currentUserIndex).t();
 
         // Add regularization.
@@ -127,7 +127,7 @@ class SVDCompleteIncrementalLearning
     const double val = V(currentItemIndex, currentUserIndex);
 
     // Update H matrix based on the non-zero entry found in WUpdate function.
-    deltaH += (val - arma::dot(W.row(currentItemIndex),
+    deltaH += (val - dot(W.row(currentItemIndex),
         H.col(currentUserIndex))) * W.row(currentItemIndex).t();
     // Add regularization.
     if (kh != 0)
@@ -216,8 +216,8 @@ class SVDCompleteIncrementalLearning<arm
     arma::mat deltaW(1, W.n_cols);
     deltaW.zeros();
 
-    deltaW += (**it - arma::dot(W.row(currentItemIndex),
-        H.col(currentUserIndex))) * arma::trans(H.col(currentUserIndex));
+    deltaW += (**it - dot(W.row(currentItemIndex),
+        H.col(currentUserIndex))) * trans(H.col(currentUserIndex));
     if (kw != 0) deltaW -= kw * W.row(currentItemIndex);
 
     W.row(currentItemIndex) += u*deltaW;
@@ -242,8 +242,8 @@ class SVDCompleteIncrementalLearning<arm
     size_t currentUserIndex = it->col();
     size_t currentItemIndex = it->row();
 
-    deltaH += (**it - arma::dot(W.row(currentItemIndex),
-        H.col(currentUserIndex))) * arma::trans(W.row(currentItemIndex));
+    deltaH += (**it - dot(W.row(currentItemIndex),
+        H.col(currentUserIndex))) * trans(W.row(currentItemIndex));
     if (kh != 0) deltaH -= kh * H.col(currentUserIndex);
 
     H.col(currentUserIndex) += u * deltaH;
diff -pruN 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_incomplete_incremental_learning.hpp 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_incomplete_incremental_learning.hpp
--- 4.3.0-2/src/mlpack/methods/amf/update_rules/svd_incomplete_incremental_learning.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/amf/update_rules/svd_incomplete_incremental_learning.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -95,7 +95,7 @@ class SVDIncompleteIncrementalLearning
       // Update only if the rating is non-zero.
       if (val != 0)
       {
-        deltaW.row(i) += (val - arma::dot(W.row(i), H.col(currentUserIndex))) *
+        deltaW.row(i) += (val - dot(W.row(i), H.col(currentUserIndex))) *
             H.col(currentUserIndex).t();
       }
       // Add regularization.
@@ -130,7 +130,7 @@ class SVDIncompleteIncrementalLearning
       // Update only if the rating is non-zero.
       if (val != 0)
       {
-        deltaH += (val - arma::dot(W.row(i), H.col(currentUserIndex))) *
+        deltaH += (val - dot(W.row(i), H.col(currentUserIndex))) *
             W.row(i).t();
       }
     }
@@ -170,8 +170,8 @@ inline void SVDIncompleteIncrementalLear
   {
     double val = *it;
     size_t i = it.row();
-    deltaW.row(i) += (val - arma::dot(W.row(i), H.col(currentUserIndex))) *
-        arma::trans(H.col(currentUserIndex));
+    deltaW.row(i) += (val - dot(W.row(i), H.col(currentUserIndex))) *
+        trans(H.col(currentUserIndex));
     if (kw != 0) deltaW.row(i) -= kw * W.row(i);
   }
 
@@ -192,8 +192,8 @@ inline void SVDIncompleteIncrementalLear
     size_t i = it.row();
     if ((val = V(i, currentUserIndex)) != 0)
     {
-      deltaH += (val - arma::dot(W.row(i), H.col(currentUserIndex))) *
-          arma::trans(W.row(i));
+      deltaH += (val - dot(W.row(i), H.col(currentUserIndex))) *
+          trans(W.row(i));
     }
   }
   if (kh != 0) deltaH -= kh * H.col(currentUserIndex);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/bipolar_sigmoid_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/bipolar_sigmoid_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/bipolar_sigmoid_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/bipolar_sigmoid_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,7 @@ class BipolarSigmoidFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = (1 - arma::exp(-x)) / (1 + arma::exp(-x));
+    y = (1 - exp(-x)) / (1 + exp(-x));
   }
 
   /**
@@ -73,7 +73,7 @@ class BipolarSigmoidFunction
   template<typename InputVecType, typename OutputVecType, typename DerivVecType>
   static void Deriv(const InputVecType& /* x */, const OutputVecType& y, DerivVecType& dy)
   {
-    dy =  (1.0 - arma::pow(y, 2)) / 2.0;
+    dy =  (1.0 - pow(y, 2)) / 2.0;
   }
 }; // class BipolarSigmoidFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/elish_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/elish_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/elish_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/elish_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,8 +71,8 @@ class ElishFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = ((x < 0.0) % ((arma::exp(x) - 1) / (1 + arma::exp(-x))))
-        + ((x >= 0.0) % (x / (1 + arma::exp(-x))));
+    y = ((x < 0.0) % ((exp(x) - 1) / (1 + exp(-x))))
+        + ((x >= 0.0) % (x / (1 + exp(-x))));
   }
 
   /**
@@ -115,8 +115,8 @@ class ElishFunction
     // the expression solely in terms of x is defined (= 0.5)
     // only calculate exp(x) once for each element where x < 0
     // this gives approx 3x speedup, despite allocating the temp vector
-    DerivVecType ex = (x < 0) % arma::exp(x);
-    dy = ((x < 0) % ((ex - 2 / (1 + ex) + 2 / arma::pow(1 + ex, 2)))) +
+    DerivVecType ex = (x < 0) % exp(x);
+    dy = ((x < 0) % ((ex - 2 / (1 + ex) + 2 / pow(1 + ex, 2)))) +
          ((x > 0) % ((y / x) % (1.0 + x - y)));
     // need to do this here, because the /x above gives nans even when the
     // condition is not met (e.g. when x > 0 is false)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/elliot_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/elliot_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/elliot_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/elliot_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -86,7 +86,7 @@ class ElliotFunction
                     const OutputVecType& /* y */,
                     DerivVecType &dy)
   {
-    dy = 1.0 / arma::pow(1.0 + arma::abs(x), 2);
+    dy = 1.0 / pow(1.0 + arma::abs(x), 2);
   }
 }; // class ElliotFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/gaussian_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/gaussian_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/gaussian_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/gaussian_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,7 +49,7 @@ class GaussianFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = arma::exp(-1 * arma::pow(x, 2));
+    y = exp(-1 * pow(x, 2));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/gelu_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/gelu_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/gelu_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/gelu_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,7 +53,7 @@ class GELUFunction
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
     y = 0.5 * x % (1 + arma::tanh(std::sqrt(2 / M_PI) *
-        (x + 0.044715 * arma::pow(x, 3))));
+        (x + 0.044715 * pow(x, 3))));
   }
 
   /**
@@ -84,9 +84,9 @@ class GELUFunction
                     const OutputVecType& /* y */,
                     DerivVecType& dy)
   {
-    dy = 0.5 * arma::tanh(0.0356774 * arma::pow(x, 3) + 0.797885 * x) +
-        (0.0535161 * arma::pow(x, 3) + 0.398942 * x) %
-        arma::pow(1 / arma::cosh(0.0356774 * arma::pow(x, 3) +
+    dy = 0.5 * arma::tanh(0.0356774 * pow(x, 3) + 0.797885 * x) +
+        (0.0535161 * pow(x, 3) + 0.398942 * x) %
+        pow(1 / arma::cosh(0.0356774 * pow(x, 3) +
         0.797885 * x), 2) + 0.5;
     dy(arma::find(x < -10)).fill(0); // catch overflows
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/hyper_sinh_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/hyper_sinh_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/hyper_sinh_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/hyper_sinh_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -2,7 +2,24 @@
  * @file methods/ann/activation_functions/hyper_sinh_function.hpp
  * @author Mayank Raj
  *
- * Definition and implementation of the Hyper-sinh function.
+ * Definition of the hyper-sinh function as described by
+ * L. Parisi et al. (2021) (full reference below).
+ *
+ * Original implementation in Python by Luca Parisi as per https://github.com/luca-parisi/hyper_sinh
+ *
+ * For more information, see the following paper.
+ *
+ * @code
+ * @article{parisi2021hyper,
+ *   title={hyper-sinh: An accurate and reliable function from shallow to deep learning in TensorFlow and Keras},
+ *   author={Parisi, Luca and Ma, Renfei and RaviChandran, Narrendar and Lanzillotta, Matteo},
+ *   journal={Machine Learning with Applications},
+ *   volume={6},
+ *   pages={100112},
+ *   year={2021},
+ *   publisher={Elsevier}
+ * }
+ * @endcode
  *
  * mlpack is free software; you may redistribute it and/or modify it under the
  * terms of the 3-clause BSD license.  You should have received a copy of the
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/identity_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/identity_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/identity_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/identity_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,7 +45,9 @@ class IdentityFunction
    * @param y The resulting output activation.
    */
   template<typename InputVecType, typename OutputVecType>
-  static void Fn(const InputVecType& x, OutputVecType& y)
+  static void Fn(const InputVecType& x, OutputVecType& y,
+      const typename std::enable_if_t<IsVector<InputVecType>::value>* = 0,
+      const typename std::enable_if_t<IsVector<OutputVecType>::value>* = 0)
   {
     y = x;
   }
@@ -82,8 +84,9 @@ class IdentityFunction
    * @param y Input activations.
    * @param dy The resulting derivatives.
    */
-  template<typename eT>
-  static void Deriv(const arma::Cube<eT>& y, arma::Cube<eT>& x)
+  template<typename CubeType>
+  static void Deriv(const CubeType& y, CubeType& x,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     x.ones(y.n_rows, y.n_cols, y.n_slices);
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/inverse_quadratic_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/inverse_quadratic_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/inverse_quadratic_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/inverse_quadratic_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ class InvQuadFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = 1 / (1 + arma::pow(x, 2));
+    y = 1 / (1 + pow(x, 2));
   }
 
   /**
@@ -72,7 +72,7 @@ class InvQuadFunction
   template<typename InputVecType, typename OutputVecType, typename DerivVecType>
   static void Deriv(const InputVecType& x, const OutputVecType& /* y */, DerivVecType &dy)
   {
-    dy = - 2 * x / arma::pow(1 + arma::pow(x, 2), 2);
+    dy = - 2 * x / pow(1 + pow(x, 2), 2);
   }
 }; // class InvQuadFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/lisht_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/lisht_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/lisht_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/lisht_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -88,7 +88,7 @@ class LiSHTFunction
                     const OutputVecType& /* y */,
                     DerivVecType& dy)
   {
-    dy = arma::tanh(x) + x % (1 - arma::pow(arma::tanh(x), 2));
+    dy = arma::tanh(x) + x % (1 - pow(arma::tanh(x), 2));
   }
 }; // class LishtFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/logistic_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/logistic_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/logistic_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/logistic_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -57,7 +57,7 @@ class LogisticFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = (1.0 / (1 + arma::exp(-x)));
+    y = (1.0 / (1 + exp(-x)));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/mish_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/mish_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/mish_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/mish_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,8 +60,7 @@ class MishFunction
   template <typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType &x, OutputVecType &y)
   {
-    y = x % (arma::exp(2 * x) + 2 * arma::exp(x)) /
-        (2 + 2 * arma::exp(x) + arma::exp(2 * x));
+    y = x % (exp(2 * x) + 2 * exp(x)) / (2 + 2 * exp(x) + exp(2 * x));
   }
 
   /**
@@ -90,9 +89,9 @@ class MishFunction
                     const OutputVecType& /* y */,
                     DerivVecType& dy)
   {
-    dy = arma::exp(x) % (4 * (x + 1) + arma::exp(x) % (4 * x + 6) +
-        4 * arma::exp(2 * x) + arma::exp(3 * x)) /
-        arma::pow(arma::exp(2 * x) + 2 * arma::exp(x) + 2, 2);
+    dy = exp(x) % (4 * (x + 1) + exp(x) % (4 * x + 6) +
+        4 * exp(2 * x) + exp(3 * x)) /
+        pow(exp(2 * x) + 2 * exp(x) + 2, 2);
   }
 }; // class MishFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/multi_quadratic_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/multi_quadratic_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/multi_quadratic_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/multi_quadratic_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,7 @@ class MultiQuadFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = arma::pow((1 + arma::pow(x, 2)), 0.5);
+    y = pow((1 + pow(x, 2)), 0.5);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/poisson1_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/poisson1_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/poisson1_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/poisson1_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ class Poisson1Function
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = (x - 1) % arma::exp(-x);
+    y = (x - 1) % exp(-x);
   }
 
   /**
@@ -74,7 +74,7 @@ class Poisson1Function
                     const OutputVecType& /* y */,
                     DerivVecType &dy)
   {
-    dy = -arma::exp(-x) % (x - 2);
+    dy = -exp(-x) % (x - 2);
   }
 }; // class Poisson1Function
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/quadratic_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/quadratic_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/quadratic_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/quadratic_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ class QuadraticFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = arma::pow(x, 2);
+    y = pow(x, 2);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/rectifier_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/rectifier_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/rectifier_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/rectifier_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -56,29 +56,17 @@ class RectifierFunction
   }
 
   /**
-   * Computes the rectifier function using a dense matrix as input.
+   * Computes the rectifier function using a 2nd /3rd-order tensor as input.
    *
    * @param x Input data.
    * @param y The resulting output activation.
    */
-  template<typename eT>
-  static void Fn(const arma::Mat<eT>& x, arma::Mat<eT>& y)
+  template<typename MatType>
+  static void Fn(const MatType& x, MatType& y)
   {
-    y.zeros(x.n_rows, x.n_cols);
-    y = arma::max(y, x);
-  }
-
-  /**
-   * Computes the rectifier function using a 3rd-order tensor as input.
-   *
-   * @param x Input data.
-   * @param y The resulting output activation.
-   */
-  template<typename eT>
-  static void Fn(const arma::Cube<eT>& x, arma::Cube<eT>& y)
-  {
-    y.zeros(x.n_rows, x.n_cols, x.n_slices);
-    y = arma::max(y, x);
+    y.set_size(size(x));
+    y.zeros();
+    y = max(y, x);
   }
 
   /**
@@ -105,7 +93,7 @@ class RectifierFunction
                     const OutputType& /* y */,
                     DerivType& dy)
   {
-    dy = arma::conv_to<DerivType>::from(x > 0);
+    dy = ConvTo<DerivType>::From(x > 0);
   }
 }; // class RectifierFunction
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/silu_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/silu_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/silu_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/silu_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -63,7 +63,7 @@ class SILUFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType &x, OutputVecType &y)
   {
-    y = x / (1.0 + arma::exp(-x));
+    y = x / (1.0 + exp(-x));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/softplus_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/softplus_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/softplus_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/softplus_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -95,7 +95,7 @@ class SoftplusFunction
                     const OutputType& /* y */,
                     DerivType& dy)
   {
-    dy = 1.0 / (1 + arma::exp(-x));
+    dy = 1.0 / (1 + exp(-x));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/softsign_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/softsign_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/softsign_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/softsign_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -98,7 +98,7 @@ class SoftsignFunction
                     const OutputVecType& /* y */,
                     DerivVecType& dy)
   {
-    dy = 1.0 / arma::pow(1.0 + arma::abs(x), 2);
+    dy = 1.0 / pow(1.0 + arma::abs(x), 2);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/spline_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/spline_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/spline_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/spline_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,7 @@ class SplineFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = arma::pow(x, 2) % arma::log(1 + x);
+    y = pow(x, 2) % log(1 + x);
   }
 
   /**
@@ -75,7 +75,7 @@ class SplineFunction
                     const OutputVecType& y,
                     DerivVecType& dy)
   {
-    dy = 2 * y / x + arma::pow(x, 2) / (1 + x);
+    dy = 2 * y / x + pow(x, 2) / (1 + x);
     // the expression above is indeterminate at 0, even though
     // the expression solely in terms of x is defined (= 0)
     dy(arma::find(x == 0)).zeros();
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/swish_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/swish_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/swish_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/swish_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,10 +47,11 @@ class SwishFunction
    * @param x Input data.
    * @param y The resulting output activation.
    */
-  template<typename eT>
-  static void Fn(const arma::Mat<eT>& x, arma::Mat<eT>& y)
+  template<typename MatType>
+  static void Fn(const MatType& x, MatType& y,
+     const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0) 
   {
-    y = x / (1.0 + arma::exp(-x));
+    y = x / (1.0 + exp(-x));
   }
 
   /**
@@ -59,8 +60,9 @@ class SwishFunction
    * @param x Input data.
    * @param y The resulting output activation.
    */
-  template<typename InputVecType, typename OutputVecType>
-  static void Fn(const InputVecType& x, OutputVecType& y)
+  template<typename VecType>
+  static void Fn(const VecType& x, VecType& y,
+      const typename std::enable_if_t<IsVector<VecType>::value>* = 0)
   {
     y.set_size(arma::size(x));
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/tanh_exponential_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/tanh_exponential_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/tanh_exponential_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/tanh_exponential_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,7 +62,7 @@ class TanhExpFunction
   template<typename InputVecType, typename OutputVecType>
   static void Fn(const InputVecType& x, OutputVecType& y)
   {
-    y = x % arma::tanh(arma::exp(x));
+    y = x % arma::tanh(exp(x));
   }
 
   /**
@@ -92,7 +92,7 @@ class TanhExpFunction
                     DerivVecType& dy)
   {
     // leverage both y and x
-    dy = y / x + x % arma::exp(x) % (1 - arma::pow(y / x, 2));
+    dy = y / x + x % exp(x) % (1 - pow(y / x, 2));
     // the expression above is indeterminate at 0, even though
     // the expression solely in terms of x is defined (= tanh(1))
     dy(arma::find(x == 0)).fill(std::tanh(1));
diff -pruN 4.3.0-2/src/mlpack/methods/ann/activation_functions/tanh_function.hpp 4.4.0-1/src/mlpack/methods/ann/activation_functions/tanh_function.hpp
--- 4.3.0-2/src/mlpack/methods/ann/activation_functions/tanh_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/activation_functions/tanh_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,7 +75,7 @@ class TanhFunction
                     const OutputVecType& y,
                     DerivVecType& dy)
   {
-    dy = 1 - arma::pow(y, 2);
+    dy = 1 - pow(y, 2);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/ann.hpp 4.4.0-1/src/mlpack/methods/ann/ann.hpp
--- 4.3.0-2/src/mlpack/methods/ann/ann.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/ann.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,7 +17,6 @@
 #define MLPACK_METHODS_ANN_ANN_HPP
 
 #include "forward_decls.hpp"
-#include "make_alias.hpp"
 
 #include "activation_functions/activation_functions.hpp"
 #include "augmented/augmented.hpp"
diff -pruN 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/add_impl.hpp 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/add_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/add_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/add_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -42,20 +42,19 @@ inline void AddTask::Generate(arma::fiel
     if (!fixedLength)
     {
       arma::vec weights(bitLen - 1);
-      weights = arma::exp2(arma::linspace(1, bitLen - 1, bitLen - 1));
+      weights = exp2(arma::linspace(1, bitLen - 1, bitLen - 1));
 
       DiscreteDistribution d(1);
       // We have two binary numbers with exactly two digits (10 and 11).
       // Increasing length by 1 double the number of valid numbers.
-      d.Probabilities(0) = arma::exp2(
-          arma::linspace(1, bitLen - 1, bitLen - 1));
+      d.Probabilities(0) = exp2(arma::linspace(1, bitLen - 1, bitLen - 1));
 
       sizeA = 2 + d.Random()(0);
       sizeB = 2 + d.Random()(0);
     }
     // Construct sequence of the form
     // (binary number with sizeA bits) + '+' + (binary number with sizeB bits).
-    vecInput(i) = arma::randi<arma::colvec>(
+    vecInput(i) = randi<arma::colvec>(
         sizeA + sizeB + 1, arma::distr_param(0, 1));
     // Insert special value for '+' delimiter.
     vecInput(i).at(sizeA) = 2;
@@ -135,7 +134,7 @@ inline void AddTask::Binarize(const arma
   output = arma::field<arma::mat>(input.n_elem);
   for (size_t i = 0; i < input.n_elem; ++i)
   {
-    output.at(i) = arma::zeros(3, input.at(i).n_elem);
+    output.at(i) = zeros(3, input.at(i).n_elem);
     for (size_t j = 0; j < input.at(i).n_elem; ++j)
     {
       size_t val = input.at(i).at(j);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/copy_impl.hpp 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/copy_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/copy_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/copy_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,25 +62,25 @@ inline void CopyTask::Generate(arma::fie
       // We have two binary numbers with exactly two digits (10 and 11).
       // Increasing length by 1 double the number of valid numbers.
       d.Probabilities(0) =
-          arma::exp2(arma::linspace(1, maxLength - 1, maxLength - 1));
+          exp2(arma::linspace(1, maxLength - 1, maxLength - 1));
 
       size = 2 + d.Random()(0);
     }
-    arma::colvec vecInput = arma::randi<arma::colvec>(
+    arma::colvec vecInput = randi<arma::colvec>(
       size, arma::distr_param(0, 1));
-    arma::colvec vecLabel = arma::conv_to<arma::colvec>::from(
-      arma::repmat(vecInput, nRepeats, 1));
+    arma::colvec vecLabel = ConvTo<arma::colvec>::From(
+        repmat(vecInput, nRepeats, 1));
     size_t totSize = vecInput.n_elem + addSeparator + vecLabel.n_elem;
-    input(i) = arma::zeros(totSize, 2);
+    input(i) = zeros(totSize, 2);
     input(i).col(0).rows(0, vecInput.n_elem - 1) =
         vecInput;
     if (addSeparator)
       input(i).at(vecInput.n_elem, 0) = 0.5;
     input(i).col(1).rows(addSeparator + vecInput.n_elem, totSize - 1) =
-        arma::ones(totSize-vecInput.n_elem - addSeparator);
+        ones(totSize - vecInput.n_elem - addSeparator);
     input(i) = input(i).t();
     input(i).reshape(input(i).n_elem, 1);
-    labels(i) = arma::zeros(totSize, 1);
+    labels(i) = zeros(totSize, 1);
     labels(i).col(0).rows(addSeparator + vecInput.n_elem, totSize - 1) =
         vecLabel;
   }
@@ -93,8 +93,8 @@ inline void CopyTask::Generate(arma::mat
   arma::field<arma::mat> fieldInput, fieldLabels;
   Generate(fieldInput, fieldLabels, batchSize, true);
   size_t cols = batchSize;
-  input = arma::zeros(fieldInput(0).n_rows, cols);
-  labels = arma::zeros(fieldLabels(0).n_rows, cols);
+  input = zeros(fieldInput(0).n_rows, cols);
+  labels = zeros(fieldLabels(0).n_rows, cols);
   for (size_t i = 0; i < cols; ++i)
   {
     input.col(i) = fieldInput.at(i);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/score_impl.hpp 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/score_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/score_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/score_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,9 +37,8 @@ double SequencePrecision(arma::field<Mat
 
   for (size_t i = 0; i < testSize; ++i)
   {
-    arma::vec delta = arma::vectorise(arma::abs(
-      trueOutputs.at(i) - predOutputs.at(i)));
-    double maxDelta = arma::max(delta);
+    arma::vec delta = vectorise(abs(trueOutputs.at(i) - predOutputs.at(i)));
+    double maxDelta = max(delta);
     if (maxDelta < tol)
     {
       score++;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/sort_impl.hpp 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/sort_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/augmented/tasks/sort_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/augmented/tasks/sort_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -57,7 +57,7 @@ inline void SortTask::Generate(arma::fie
       // Generate random uniform length from [2..maxLength].
       size = RandInt(2, maxLength+1);
     }
-    input(i) = arma::randi<arma::mat>(bitLen, size, arma::distr_param(0, 1));
+    input(i) = randi<arma::mat>(bitLen, size, arma::distr_param(0, 1));
     arma::mat itemAns(bitLen, size);
     arma::colvec vals(size);
     for (size_t j = 0; j < size; ++j)
@@ -79,7 +79,7 @@ inline void SortTask::Generate(arma::fie
     input(i).reshape(input(i).n_elem, 1);
     if (addSeparator)
     {
-      arma::mat sepInput = arma::zeros(input(i).n_elem + size, 1);
+      arma::mat sepInput = zeros(input(i).n_elem + size, 1);
       size_t ptr = 0, origPtr = 0;
       for (size_t j = 0; j < size; ++j)
       {
@@ -105,8 +105,8 @@ inline void SortTask::Generate(arma::mat
   size_t inputRows = fieldInput(0).n_rows;
   size_t labelRows = fieldLabels(0).n_rows;
   size_t cols = batchSize;
-  input = arma::zeros(inputRows, cols);
-  labels = arma::zeros(labelRows, cols);
+  input = zeros(inputRows, cols);
+  labels = zeros(labelRows, cols);
   for (size_t i = 0; i < cols; ++i)
   {
     input.col(i) = fieldInput.at(i);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/convolution_rules/fft_convolution.hpp 4.4.0-1/src/mlpack/methods/ann/convolution_rules/fft_convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/convolution_rules/fft_convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/convolution_rules/fft_convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,15 +47,16 @@ class FFTConvolution
    * @param filter Filter used to perform the convolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT, typename Border = BorderMode>
+  template<typename MatType, typename Border = BorderMode>
   static typename std::enable_if<
       std::is_same<Border, ValidConvolution>::value, void>::type
-  Convolution(const arma::Mat<eT>& input,
-              const arma::Mat<eT>& filter,
-              arma::Mat<eT>& output)
+  Convolution(const MatType& input,
+              const MatType& filter,
+              MatType& output,
+              const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
-    arma::Mat<eT> inputPadded = input;
-    arma::Mat<eT> filterPadded = filter;
+    MatType inputPadded = input;
+    MatType filterPadded = filter;
 
     if (padLastDim)
       inputPadded.resize(inputPadded.n_rows, inputPadded.n_cols + 1);
@@ -63,8 +64,7 @@ class FFTConvolution
     // Pad filter and input to the output shape.
     filterPadded.resize(inputPadded.n_rows, inputPadded.n_cols);
 
-    arma::Mat<eT> temp = arma::real(ifft2(arma::fft2(inputPadded) % arma::fft2(
-        filterPadded)));
+    MatType temp = real(ifft2(fft2(inputPadded) % fft2(filterPadded)));
 
     // Extract the region of interest. We don't need to handle the padLastDim in
     // a special way we just cut it out from the output matrix.
@@ -82,12 +82,13 @@ class FFTConvolution
    * @param filter Filter used to perform the convolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT, typename Border = BorderMode>
+  template<typename MatType, typename Border = BorderMode>
   static typename std::enable_if<
       std::is_same<Border, FullConvolution>::value, void>::type
-  Convolution(const arma::Mat<eT>& input,
-              const arma::Mat<eT>& filter,
-              arma::Mat<eT>& output)
+  Convolution(const MatType& input,
+              const MatType& filter,
+              MatType& output,
+              const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     // In case of the full convolution outputRows and outputCols doesn't
     // represent the true output size when the padLastDim parameter is set,
@@ -99,18 +100,16 @@ class FFTConvolution
         outputCols++;
 
     // Pad filter and input to the working output shape.
-    arma::Mat<eT> inputPadded = arma::zeros<arma::Mat<eT> >(outputRows,
-        outputCols);
+    MatType inputPadded = zeros<MatType>(outputRows, outputCols);
     inputPadded.submat(filter.n_rows - 1, filter.n_cols - 1,
           filter.n_rows - 1 + input.n_rows - 1,
           filter.n_cols - 1 + input.n_cols - 1) = input;
 
-    arma::Mat<eT> filterPadded = filter;
+    MatType filterPadded = filter;
     filterPadded.resize(outputRows, outputCols);
 
     // Perform FFT and IFFT
-    arma::Mat<eT> temp = arma::real(ifft2(arma::fft2(inputPadded) % arma::fft2(
-        filterPadded)));
+    MatType temp = real(ifft2(fft2(inputPadded) % fft2(filterPadded)));
 
     // Extract the region of interest. We don't need to handle the padLastDim
     // parameter in a special way we just cut it out from the output matrix.
@@ -130,17 +129,18 @@ class FFTConvolution
    * @param filter Filter used to perform the convolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const CubeType& filter,
+                          CubeType& output,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    typedef typename GetDenseMatType<CubeType>::type MatType;
+    MatType convOutput;
     FFTConvolution<BorderMode>::Convolution(input.slice(0), filter.slice(0),
         convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-        input.n_slices);
+    output = CubeType(convOutput.n_rows, convOutput.n_cols, input.n_slices);
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < input.n_slices; ++i)
@@ -161,17 +161,16 @@ class FFTConvolution
    * @param filter Filter used to perform the convolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Mat<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename MatType, typename CubeType>
+  static void Convolution(const MatType& input,
+                          const CubeType& filter,
+                          CubeType& output)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     FFTConvolution<BorderMode>::Convolution(input, filter.slice(0),
         convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-        filter.n_slices);
+    output = CubeType(convOutput.n_rows, convOutput.n_cols, filter.n_slices);
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < filter.n_slices; ++i)
@@ -189,16 +188,16 @@ class FFTConvolution
    * @param filter Filter used to perform the convolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Mat<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename MatType, typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const MatType& filter,
+                          CubeType& output)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     FFTConvolution<BorderMode>::Convolution(input.slice(0), filter,
         convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
+    output = CubeType(convOutput.n_rows, convOutput.n_cols,
         input.n_slices);
     output.slice(0) = convOutput;
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/convolution_rules/naive_convolution.hpp 4.4.0-1/src/mlpack/methods/ann/convolution_rules/naive_convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/convolution_rules/naive_convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/convolution_rules/naive_convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,18 +47,21 @@ class NaiveConvolution
    * @param appending If true, it will not initialize the output. Instead,
    *                  it will append the results to the output.
    */
-  template<typename eT, typename Border = BorderMode>
+  template<typename InMatType, typename FilMatType, typename OutMatType,
+      typename Border = BorderMode>
   static typename std::enable_if<
       std::is_same<Border, ValidConvolution>::value, void>::type
-  Convolution(const arma::Mat<eT>& input,
-              const arma::Mat<eT>& filter,
-              arma::Mat<eT>& output,
+  Convolution(const InMatType& input,
+              const FilMatType& filter,
+              OutMatType& output,
               const size_t dW = 1,
               const size_t dH = 1,
               const size_t dilationW = 1,
               const size_t dilationH = 1,
-              const bool appending = false)
+              const bool appending = false,
+              const typename std::enable_if_t<IsMatrix<InMatType>::value>* = 0)
   {
+    typedef typename InMatType::elem_type eT;
     // Compute the output size.  The filterRows and filterCols computation must
     // take into account the fact that dilation only adds rows or columns
     // *between* filter elements.  So, e.g., a dilation of 2 on a kernel size of
@@ -105,17 +108,19 @@ class NaiveConvolution
    * @param appending If true, it will not initialize the output. Instead,
    *                  it will append the results to the output.
    */
-  template<typename eT, typename Border = BorderMode>
+  template<typename InMatType, typename FilMatType, typename OutMatType,
+      typename Border = BorderMode>
   static typename std::enable_if<
       std::is_same<Border, FullConvolution>::value, void>::type
-  Convolution(const arma::Mat<eT>& input,
-              const arma::Mat<eT>& filter,
-              arma::Mat<eT>& output,
+  Convolution(const InMatType& input,
+              const FilMatType& filter,
+              OutMatType& output,
               const size_t dW = 1,
               const size_t dH = 1,
               const size_t dilationW = 1,
               const size_t dilationH = 1,
-              const bool appending = false)
+              const bool appending = false,
+              const typename std::enable_if_t<IsMatrix<InMatType>::value>* = 0)
   {
     // First, compute the necessary padding for the full convolution.  It is
     // possible that this might be an overestimate.  Note that these variables
@@ -126,7 +131,7 @@ class NaiveConvolution
     const size_t paddingCols = filterCols - 1;
 
     // Pad filter and input to the working output shape.
-    arma::Mat<eT> inputPadded(input.n_rows + 2 * paddingRows,
+    InMatType inputPadded(input.n_rows + 2 * paddingRows,
         input.n_cols + 2 * paddingCols, arma::fill::zeros);
     inputPadded.submat(paddingRows, paddingCols, paddingRows + input.n_rows - 1,
         paddingCols + input.n_cols - 1) = input;
@@ -148,23 +153,25 @@ class NaiveConvolution
    * @param appending If true, it will not initialize the output. Instead,
    *                  it will append the results to the output.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output,
+  template<typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const CubeType& filter,
+                          CubeType& output,
                           const size_t dW = 1,
                           const size_t dH = 1,
                           const size_t dilationW = 1,
                           const size_t dilationH = 1,
-                          const bool appending = false)
+                          const bool appending = false,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    typedef typename GetDenseMatType<CubeType>::type MatType;
+    MatType convOutput;
     NaiveConvolution<BorderMode>::Convolution(input.slice(0), filter.slice(0),
         convOutput, dW, dH, dilationW, dilationH, appending);
 
     if (!appending)
-      output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-          input.n_slices);
+      output = CubeType(convOutput.n_rows, convOutput.n_cols, input.n_slices);
+
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < input.n_slices; ++i)
@@ -188,23 +195,25 @@ class NaiveConvolution
    * @param appending If true, it will not initialize the output. Instead,
    *                  it will append the results to the output.
    */
-  template<typename eT>
-  static void Convolution(const arma::Mat<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output,
+  template<typename MatType, typename CubeType>
+  static void Convolution(const MatType& input,
+                          const CubeType& filter,
+                          CubeType& output,
                           const size_t dW = 1,
                           const size_t dH = 1,
                           const size_t dilationW = 1,
                           const size_t dilationH = 1,
-                          const bool appending = false)
+                          const bool appending = false,
+                          const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     NaiveConvolution<BorderMode>::Convolution(input, filter.slice(0),
         convOutput, dW, dH, dilationW, dilationH, appending);
 
     if (!appending)
-      output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-          filter.n_slices);
+      output = CubeType(convOutput.n_rows, convOutput.n_cols, filter.n_slices);
+
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < filter.n_slices; ++i)
@@ -228,23 +237,25 @@ class NaiveConvolution
    * @param appending If true, it will not initialize the output. Instead,
    *                  it will append the results to the output.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Mat<eT>& filter,
-                          arma::Cube<eT>& output,
+  template<typename MatType, typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const MatType& filter,
+                          CubeType& output,
                           const size_t dW = 1,
                           const size_t dH = 1,
                           const size_t dilationW = 1,
                           const size_t dilationH = 1,
-                          const bool appending = false)
+                          const bool appending = false,
+                          const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     NaiveConvolution<BorderMode>::Convolution(input.slice(0), filter,
         convOutput, dW, dH, dilationW, dilationH, appending);
 
     if (!appending)
-      output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-          input.n_slices);
+      output = CubeType(convOutput.n_rows, convOutput.n_cols, input.n_slices);
+
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < input.n_slices; ++i)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/convolution_rules/svd_convolution.hpp 4.4.0-1/src/mlpack/methods/ann/convolution_rules/svd_convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/convolution_rules/svd_convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/convolution_rules/svd_convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,11 +49,13 @@ class SVDConvolution
    * @param filter Filter used to perform the conolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Mat<eT>& input,
-                          const arma::Mat<eT>& filter,
-                          arma::Mat<eT>& output)
+  template<typename MatType>
+  static void Convolution(const MatType& input,
+                          const MatType& filter,
+                          MatType& output,
+                          const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
+    typedef typename GetColType<MatType>::type ColType;
     // Use the naive convolution in case the filter isn't two dimensional or the
     // filter is bigger than the input.
     if (filter.n_rows > input.n_rows || filter.n_cols > input.n_cols ||
@@ -63,28 +65,27 @@ class SVDConvolution
     }
     else
     {
-      arma::Mat<eT> U, V, subOutput;
-      arma::Col<eT> s;
+      MatType U, V, subOutput;
+      ColType s;
 
-      arma::svd_econ(U, s, V, filter);
+      svd_econ(U, s, V, filter);
 
       // Rank approximation using the singular values calculated with singular
       // value decomposition of dense filter matrix.
-      const size_t rank = arma::sum(s > (s.n_elem * arma::max(s) *
-          arma::datum::eps));
+      const size_t rank = sum(s > (s.n_elem * max(s) * arma::datum::eps));
 
       // Test for separability based on the rank of the kernel and take
       // advantage of the low rank.
       if (rank * (filter.n_rows + filter.n_cols) < filter.n_elem)
       {
-        arma::Mat<eT> subFilter = V.unsafe_col(0) * s(0);
+        MatType subFilter = V.unsafe_col(0) * s(0);
         NaiveConvolution<BorderMode>::Convolution(input, subFilter, subOutput);
 
         subOutput = subOutput.t();
         NaiveConvolution<BorderMode>::Convolution(subOutput, U.unsafe_col(0),
             output);
 
-        arma::Mat<eT> temp;
+        MatType temp;
         for (size_t r = 1; r < rank; r++)
         {
           subFilter = V.unsafe_col(r) * s(r);
@@ -113,17 +114,18 @@ class SVDConvolution
    * @param filter Filter used to perform the conolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const CubeType& filter,
+                          CubeType& output,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    typedef typename GetDenseMatType<CubeType>::type MatType;
+    MatType convOutput;
     SVDConvolution<BorderMode>::Convolution(input.slice(0), filter.slice(0),
         convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-        input.n_slices);
+    output = CubeType(convOutput.n_rows, convOutput.n_cols, input.n_slices);
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < input.n_slices; ++i)
@@ -141,16 +143,17 @@ class SVDConvolution
    * @param filter Filter used to perform the conolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Mat<eT>& input,
-                          const arma::Cube<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename MatType, typename CubeType>
+  static void Convolution(const MatType& input,
+                          const CubeType& filter,
+                          CubeType& output,
+                          const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     SVDConvolution<BorderMode>::Convolution(input, filter.slice(0), convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-        filter.n_slices);
+    output = CubeType(convOutput.n_rows, convOutput.n_cols, filter.n_slices);
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < filter.n_slices; ++i)
@@ -168,16 +171,17 @@ class SVDConvolution
    * @param filter Filter used to perform the conolution.
    * @param output Output data that contains the results of the convolution.
    */
-  template<typename eT>
-  static void Convolution(const arma::Cube<eT>& input,
-                          const arma::Mat<eT>& filter,
-                          arma::Cube<eT>& output)
+  template<typename MatType, typename CubeType>
+  static void Convolution(const CubeType& input,
+                          const MatType& filter,
+                          CubeType& output,
+                          const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0,
+                          const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
-    arma::Mat<eT> convOutput;
+    MatType convOutput;
     SVDConvolution<BorderMode>::Convolution(input.slice(0), filter, convOutput);
 
-    output = arma::Cube<eT>(convOutput.n_rows, convOutput.n_cols,
-        input.n_slices);
+    output = CubeType(convOutput.n_rows, convOutput.n_cols, input.n_slices);
     output.slice(0) = convOutput;
 
     for (size_t i = 1; i < input.n_slices; ++i)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/dists/bernoulli_distribution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/dists/bernoulli_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/dists/bernoulli_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/dists/bernoulli_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,8 +48,8 @@ BernoulliDistribution<DataType>::Bernoul
 template<typename DataType>
 DataType BernoulliDistribution<DataType>::Sample() const
 {
-  DataType sample = arma::randu<DataType>
-      (probability.n_rows, probability.n_cols);
+  DataType sample;
+  sample.randu(probability.n_rows, probability.n_cols);
 
   for (size_t i = 0; i < sample.n_elem; ++i)
       sample(i) = sample(i) < probability(i);
@@ -61,8 +61,8 @@ template<typename DataType>
 double BernoulliDistribution<DataType>::LogProbability(
     const DataType& observation) const
 {
-  return arma::accu(arma::log(probability + eps) % observation +
-      arma::log(1 - probability + eps) % (1 - observation)) /
+  return accu(log(probability + eps) % observation +
+      log(1 - probability + eps) % (1 - observation)) /
       observation.n_cols;
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/dists/normal_distribution.hpp 4.4.0-1/src/mlpack/methods/ann/dists/normal_distribution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/dists/normal_distribution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/dists/normal_distribution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,7 +52,7 @@ class NormalDistribution
    */
   DataType Probability(const DataType& observation) const
   {
-    return arma::exp(LogProbability(observation));
+    return exp(LogProbability(observation));
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/ann/dists/normal_distribution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/dists/normal_distribution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/dists/normal_distribution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/dists/normal_distribution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,16 +37,15 @@ NormalDistribution<DataType>::NormalDist
 template<typename DataType>
 DataType NormalDistribution<DataType>::Sample() const
 {
-  return sigma * arma::randn<DataType>(mean.n_elem) + mean;
+  return sigma * randn<DataType>(mean.n_elem) + mean;
 }
 
 template<typename DataType>
 DataType NormalDistribution<DataType>::LogProbability(
     const DataType& observation) const
 {
-  const DataType v1 = arma::log(sigma) + std::log(std::sqrt(2 * M_PI));
-  const DataType v2 = arma::square(observation - mean) /
-                      (2 * arma::square(sigma));
+  const DataType v1 = log(sigma) + std::log(std::sqrt(2 * M_PI));
+  const DataType v2 = square(observation - mean) / (2 * square(sigma));
   return  (-v1 - v2);
 }
 
@@ -56,9 +55,8 @@ void NormalDistribution<DataType>::ProbB
     DataType& dmu,
     DataType& dsigma) const
 {
-  dmu = (observation - mean) / (arma::square(sigma)) % Probability(observation);
-  dsigma = (- 1.0 / sigma +
-            (arma::square(observation - mean) / arma::pow(sigma, 3)))
+  dmu = (observation - mean) / (square(sigma)) % Probability(observation);
+  dsigma = (-1.0 / sigma + (square(observation - mean) / pow(sigma, 3)))
             % Probability(observation);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/ffn_impl.hpp 4.4.0-1/src/mlpack/methods/ann/ffn_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/ffn_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/ffn_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -15,8 +15,6 @@
 // In case it hasn't been included yet.
 #include "ffn.hpp"
 
-#include "make_alias.hpp"
-
 namespace mlpack {
 
 template<typename OutputLayerType,
@@ -206,11 +204,12 @@ void FFN<
     const size_t effectiveBatchSize = std::min(batchSize,
         size_t(predictors.n_cols) - i);
 
-    const MatType predictorAlias(
-        const_cast<typename MatType::elem_type*>(predictors.colptr(i)),
-        predictors.n_rows, effectiveBatchSize, false, true);
-    MatType resultAlias(results.colptr(i), results.n_rows,
-        effectiveBatchSize, false, true);
+    MatType predictorAlias, resultAlias;
+
+    MakeAlias(predictorAlias, predictors, predictors.n_rows,
+        effectiveBatchSize, i * predictors.n_rows);
+    MakeAlias(resultAlias, results, results.n_rows, effectiveBatchSize,
+        i * results.n_rows);
 
     network.Forward(predictorAlias, resultAlias);
   }
@@ -451,8 +450,10 @@ typename MatType::elem_type FFN<
   // pass.
   networkOutput.set_size(network.OutputSize(), batchSize);
   MatType predictorsBatch, responsesBatch;
-  MakeAlias(predictorsBatch, predictors.colptr(begin), predictors.n_rows, batchSize);
-  MakeAlias(responsesBatch, responses.colptr(begin), responses.n_rows, batchSize);
+  MakeAlias(predictorsBatch, predictors, predictors.n_rows, batchSize,
+      begin * predictors.n_rows);
+  MakeAlias(responsesBatch, responses, responses.n_rows, batchSize,
+      begin * responses.n_rows);
   network.Forward(predictorsBatch, networkOutput);
 
   return outputLayer.Forward(networkOutput, responsesBatch) + network.Loss();
@@ -499,10 +500,10 @@ typename MatType::elem_type FFN<
 
   // Alias the batches so we don't copy memory.
   MatType predictorsBatch, responsesBatch;
-  MakeAlias(predictorsBatch, predictors.colptr(begin), predictors.n_rows,
-      batchSize);
-  MakeAlias(responsesBatch, responses.colptr(begin), responses.n_rows,
-      batchSize);
+  MakeAlias(predictorsBatch, predictors, predictors.n_rows,
+      batchSize, begin * predictors.n_rows);
+  MakeAlias(responsesBatch, responses, responses.n_rows,
+      batchSize, begin * responses.n_rows);
 
   network.Forward(predictorsBatch, networkOutput);
 
@@ -598,7 +599,7 @@ void FFN<
       "FFN::SetLayerMemory(): total layer weight size does not match parameter "
       "size!");
 
-  network.SetWeights(parameters.memptr());
+  network.SetWeights(parameters);
   layerMemoryIsSet = true;
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/const_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/const_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/const_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/const_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,8 +37,8 @@ class ConstInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
     if (W.is_empty())
       W.set_size(rows, cols);
@@ -51,8 +51,9 @@ class ConstInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
@@ -68,8 +69,8 @@ class ConstInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -85,8 +86,9 @@ class ConstInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/gaussian_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/gaussian_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/gaussian_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/gaussian_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -1,10 +1,11 @@
 /**
  * @file methods/ann/init_rules/gaussian_init.hpp
  * @author Kris Singh
+ * @author Omar Shrit
  *
  * Intialization rule for the neural networks. This simple initialization is
- * performed by assigning a gaussian matrix with a given mean and variance
- * to the weight matrix.
+ * performed by assigning a gaussian matrix with a given mean and standard
+ * deviation to the weight matrix.
  *
  * mlpack is free software; you may redistribute it and/or modify it under the
  * terms of the 3-clause BSD license.  You should have received a copy of the
@@ -29,10 +30,10 @@ class GaussianInitialization
    * Initialize the gaussian with the given mean and variance.
    *
    * @param mean Mean of the gaussian.
-   * @param variance Variance of the gaussian.
+   * @param stddev Standard deviation of the gaussian.
    */
-  GaussianInitialization(const double mean = 0, const double variance = 1) :
-      mean(mean), variance(variance)
+  GaussianInitialization(const double mean = 0, const double stddev = 1) :
+      mean(mean), stddev(stddev)
   {
     // Nothing to do here.
   }
@@ -44,15 +45,15 @@ class GaussianInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W,
+  template<typename MatType>
+  void Initialize(MatType& W,
                   const size_t rows,
                   const size_t cols)
   {
     if (W.is_empty())
       W.set_size(rows, cols);
 
-    W.imbue( [&]() { return arma::as_scalar(RandNormal(mean, variance)); } );
+    W = randn<MatType>(rows, cols) * stddev + mean;
   }
 
   /**
@@ -60,13 +61,14 @@ class GaussianInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
 
-    W.imbue( [&]() { return arma::as_scalar(RandNormal(mean, variance)); } );
+    W = randn<MatType>(W.n_rows, W.n_cols) * stddev + mean;
   }
 
   /**
@@ -77,8 +79,8 @@ class GaussianInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT> & W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -95,8 +97,9 @@ class GaussianInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT> & W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
@@ -112,15 +115,15 @@ class GaussianInitialization
   void serialize(Archive& ar, const uint32_t /* version */)
   {
     ar(CEREAL_NVP(mean));
-    ar(CEREAL_NVP(variance));
+    ar(CEREAL_NVP(stddev));
   }
 
  private:
   //! Mean of the gaussian.
   double mean;
 
-  //! Variance of the gaussian.
-  double variance;
+  //! Standard deviation of the gaussian.
+  double stddev;
 }; // class GaussianInitialization
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/glorot_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/glorot_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/glorot_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/glorot_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,8 +67,8 @@ class GlorotInitializationType
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W,
+  template<typename MatType>
+  void Initialize(MatType& W,
                   const size_t rows,
                   const size_t cols);
 
@@ -77,8 +77,9 @@ class GlorotInitializationType
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W);
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0);
 
   /**
    * Initialize the elements of the specified weight 3rd order tensor with
@@ -89,8 +90,8 @@ class GlorotInitializationType
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices);
@@ -101,8 +102,9 @@ class GlorotInitializationType
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W);
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0);
 
   /**
    * Serialize the initialization.  (Nothing to serialize for this one.)
@@ -112,34 +114,22 @@ class GlorotInitializationType
 }; // class GlorotInitializationType
 
 template<>
-template<typename eT>
-inline void GlorotInitializationType<false>::Initialize(arma::Mat<eT>& W,
+template<typename MatType>
+inline void GlorotInitializationType<false>::Initialize(MatType& W,
                                                         const size_t rows,
                                                         const size_t cols)
 {
   if (W.is_empty())
     W.set_size(rows, cols);
 
-  double var = 2.0 / double(rows + cols);
-  GaussianInitialization normalInit(0.0, var);
+  double stddev = std::sqrt(2.0 / double(rows + cols));
+  GaussianInitialization normalInit(0.0, stddev);
   normalInit.Initialize(W, rows, cols);
 }
 
 template<>
-template<typename eT>
-inline void GlorotInitializationType<false>::Initialize(arma::Mat<eT>& W)
-{
-  if (W.is_empty())
-    Log::Fatal << "Cannot initialize and empty matrix." << std::endl;
-
-  double var = 2.0 / double(W.n_rows + W.n_cols);
-  GaussianInitialization normalInit(0.0, var);
-  normalInit.Initialize(W);
-}
-
-template<>
-template<typename eT>
-inline void GlorotInitializationType<true>::Initialize(arma::Mat<eT>& W,
+template<typename MatType> 
+inline void GlorotInitializationType<true>::Initialize(MatType& W,
                                                        const size_t rows,
                                                        const size_t cols)
 {
@@ -147,27 +137,41 @@ inline void GlorotInitializationType<tru
     W.set_size(rows, cols);
 
   // Limit of distribution.
-  double a = sqrt(6) / sqrt(rows + cols);
+  double a = std::sqrt(6) / std::sqrt(rows + cols);
   RandomInitialization randomInit(-a, a);
   randomInit.Initialize(W, rows, cols);
 }
 
 template<>
-template<typename eT>
-inline void GlorotInitializationType<true>::Initialize(arma::Mat<eT>& W)
+template<typename MatType>
+inline void GlorotInitializationType<false>::Initialize(MatType& W,
+    const typename std::enable_if_t<IsMatrix<MatType>::value>*)
+{
+  if (W.is_empty())
+    Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
+
+  double stddev = std::sqrt(2.0 / double(W.n_rows + W.n_cols));
+  GaussianInitialization normalInit(0.0, stddev);
+  normalInit.Initialize(W);
+}
+
+template<>
+template<typename MatType>
+inline void GlorotInitializationType<true>::Initialize(MatType& W,
+    const typename std::enable_if_t<IsMatrix<MatType>::value>*)
 {
   if (W.is_empty())
     Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
 
   // Limit of distribution.
-  double a = sqrt(6) / sqrt(W.n_rows + W.n_cols);
+  double a = std::sqrt(6) / std::sqrt(W.n_rows + W.n_cols);
   RandomInitialization randomInit(-a, a);
   randomInit.Initialize(W);
 }
 
 template <bool Uniform>
-template<typename eT>
-inline void GlorotInitializationType<Uniform>::Initialize(arma::Cube<eT>& W,
+template<typename CubeType> 
+inline void GlorotInitializationType<Uniform>::Initialize(CubeType& W,
                                                           const size_t rows,
                                                           const size_t cols,
                                                           const size_t slices)
@@ -180,8 +184,9 @@ inline void GlorotInitializationType<Uni
 }
 
 template <bool Uniform>
-template<typename eT>
-inline void GlorotInitializationType<Uniform>::Initialize(arma::Cube<eT>& W)
+template<typename CubeType>
+inline void GlorotInitializationType<Uniform>::Initialize(CubeType& W,
+    const typename std::enable_if_t<IsCube<CubeType>::value>*)
 {
   if (W.is_empty())
     Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/he_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/he_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/he_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/he_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,8 +60,8 @@ class HeInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template <typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template <typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
     // He initialization rule says to initialize weights with random
     // values taken from a gaussian distribution with mean = 0 and
@@ -73,7 +73,7 @@ class HeInitialization
 
     // Multipling a random variable X with variance V(X) by some factor c,
     // then the variance V(cX) = (c^2) * V(X).
-    W.imbue( [&]() { return sqrt(variance) * arma::randn(); } );
+    W = randn<MatType>(rows, cols) * std::sqrt(variance);
   }
 
   /**
@@ -82,8 +82,9 @@ class HeInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template <typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template <typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     // He initialization rule says to initialize weights with random
     // values taken from a gaussian distribution with mean = 0 and
@@ -95,7 +96,7 @@ class HeInitialization
 
     // Multipling a random variable X with variance V(X) by some factor c,
     // then the variance V(cX) = (c^2) * V(X).
-    W.imbue( [&]() { return sqrt(variance) * arma::randn(); } );
+    W = randn<MatType>(W.n_rows, W.n_cols) * std::sqrt(variance);
   }
 
   /**
@@ -107,8 +108,8 @@ class HeInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template <typename eT>
-  void Initialize(arma::Cube<eT> & W,
+  template <typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -126,8 +127,9 @@ class HeInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template <typename eT>
-  void Initialize(arma::Cube<eT> & W)
+  template <typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix" << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/kathirvalavakumar_subavathi_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/kathirvalavakumar_subavathi_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/kathirvalavakumar_subavathi_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/kathirvalavakumar_subavathi_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,11 +65,11 @@ class KathirvalavakumarSubavathiInitiali
    * @param data The input patterns.
    * @param s Parameter that defines the active region.
    */
-  template<typename eT>
-  KathirvalavakumarSubavathiInitialization(const arma::Mat<eT>& data,
+  template<typename MatType>
+  KathirvalavakumarSubavathiInitialization(const MatType& data,
                                            const double s) : s(s)
   {
-    dataSum = arma::sum(data % data);
+    dataSum = sum(data % data);
   }
 
   /**
@@ -80,10 +80,11 @@ class KathirvalavakumarSubavathiInitiali
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
-    arma::Row<eT> b = s * arma::sqrt(3 / (rows * dataSum));
+    typedef typename GetRowType<MatType>::type RowType; 
+    RowType b = s * sqrt(3 / (rows * dataSum));
     const double theta = b.min();
     RandomInitialization randomInit(-theta, theta);
     randomInit.Initialize(W, rows, cols);
@@ -95,10 +96,12 @@ class KathirvalavakumarSubavathiInitiali
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
-    arma::Row<eT> b = s * arma::sqrt(3 / (W.n_rows * dataSum));
+    typedef typename GetRowType<MatType>::type RowType; 
+    RowType b = s * sqrt(3 / (W.n_rows * dataSum));
     const double theta = b.min();
     RandomInitialization randomInit(-theta, theta);
     randomInit.Initialize(W);
@@ -113,8 +116,8 @@ class KathirvalavakumarSubavathiInitiali
    * @param cols Number of columns.
    * @param slices Number of slices
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -132,8 +135,9 @@ class KathirvalavakumarSubavathiInitiali
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/lecun_normal_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/lecun_normal_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/lecun_normal_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/lecun_normal_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,12 +64,12 @@ class LecunNormalInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template <typename eT>
-  void Initialize(arma::Mat<eT>& W,
+  template <typename MatType>
+  void Initialize(MatType& W,
                   const size_t rows,
                   const size_t cols)
   {
-    // He initialization rule says to initialize weights with random
+    // Lecun initialization rule says to initialize weights with random
     // values taken from a gaussian distribution with mean = 0 and
     // standard deviation = sqrt(1 / rows), i.e. variance = (1 / rows).
     const double variance = 1.0 / ((double) rows);
@@ -79,7 +79,7 @@ class LecunNormalInitialization
 
     // Multipling a random variable X with variance V(X) by some factor c,
     // then the variance V(cX) = (c ^ 2) * V(X).
-    W.imbue( [&]() { return sqrt(variance) * arma::randn(); } );
+    W = randn<MatType>(rows, cols) * std::sqrt(variance);
   }
 
   /**
@@ -88,10 +88,11 @@ class LecunNormalInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template <typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template <typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
-    // He initialization rule says to initialize weights with random
+    // Lecun initialization rule says to initialize weights with random
     // values taken from a gaussian distribution with mean = 0 and
     // standard deviation = sqrt(1 / rows), i.e. variance = (1 / rows).
     const double variance = 1.0 / (double) W.n_rows;
@@ -101,7 +102,7 @@ class LecunNormalInitialization
 
     // Multipling a random variable X with variance V(X) by some factor c,
     // then the variance V(cX) = (c ^ 2) * V(X).
-    W.imbue( [&]() { return sqrt(variance) * arma::randn(); } );
+    W = randn<MatType>(W.n_rows, W.n_cols) * std::sqrt(variance);
   }
 
   /**
@@ -113,8 +114,8 @@ class LecunNormalInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template <typename eT>
-  void Initialize(arma::Cube<eT> & W,
+  template <typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -132,8 +133,9 @@ class LecunNormalInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template <typename eT>
-  void Initialize(arma::Cube<eT> & W)
+  template <typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/nguyen_widrow_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/nguyen_widrow_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/nguyen_widrow_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/nguyen_widrow_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,14 +71,14 @@ class NguyenWidrowInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
     RandomInitialization randomInit(lowerBound, upperBound);
     randomInit.Initialize(W, rows, cols);
 
     double beta = 0.7 * std::pow(cols, 1.0 / rows);
-    W *= (beta / arma::norm(W));
+    W *= (beta / norm(W));
   }
 
   /**
@@ -87,14 +87,15 @@ class NguyenWidrowInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     RandomInitialization randomInit(lowerBound, upperBound);
     randomInit.Initialize(W);
 
     double beta = 0.7 * std::pow(W.n_cols, 1.0 / W.n_rows);
-    W *= (beta / arma::norm(W));
+    W *= (beta / norm(W));
   }
 
   /**
@@ -106,8 +107,8 @@ class NguyenWidrowInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -125,8 +126,9 @@ class NguyenWidrowInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/oivs_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/oivs_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/oivs_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/oivs_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -79,13 +79,13 @@ class OivsInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
     RandomInitialization randomInit(-gamma, gamma);
     randomInit.Initialize(W, rows, cols);
 
-    W = (b / (k  * rows)) * arma::sqrt(W + 1);
+    W = (b / (k  * rows)) * sqrt(W + 1);
   }
 
   /**
@@ -93,13 +93,14 @@ class OivsInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     RandomInitialization randomInit(-gamma, gamma);
     randomInit.Initialize(W);
 
-    W = (b / (k  * W.n_rows)) * arma::sqrt(W + 1);
+    W = (b / (k  * W.n_rows)) * sqrt(W + 1);
   }
 
   /**
@@ -111,8 +112,8 @@ class OivsInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -130,8 +131,9 @@ class OivsInitialization
    *
    * @param W 3rd order tensor to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/orthogonal_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/orthogonal_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/orthogonal_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/orthogonal_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,13 +38,14 @@ class OrthogonalInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
-    arma::Mat<eT> V;
-    arma::Col<eT> s;
+    MatType V;
+    typedef typename GetColType<MatType>::type ColType;
+    ColType s;
 
-    arma::svd_econ(W, s, V, arma::randu<arma::Mat<eT> >(rows, cols));
+    svd_econ(W, s, V, randu<MatType>(rows, cols));
     W *= gain;
   }
 
@@ -54,13 +55,15 @@ class OrthogonalInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
-    arma::Mat<eT> V;
-    arma::Col<eT> s;
+    MatType V;
+    typedef typename GetColType<MatType>::type ColType;
+    ColType s;
 
-    arma::svd_econ(W, s, V, arma::randu<arma::Mat<eT> >(W.n_rows, W.n_cols));
+    svd_econ(W, s, V, randu<MatType>(W.n_rows, W.n_cols));
     W *= gain;
   }
 
@@ -73,8 +76,8 @@ class OrthogonalInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -92,8 +95,9 @@ class OrthogonalInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/init_rules/random_init.hpp 4.4.0-1/src/mlpack/methods/ann/init_rules/random_init.hpp
--- 4.3.0-2/src/mlpack/methods/ann/init_rules/random_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/init_rules/random_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,8 +51,8 @@ class RandomInitialization
    * @param rows Number of rows.
    * @param cols Number of columns.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W, const size_t rows, const size_t cols)
+  template<typename MatType>
+  void Initialize(MatType& W, const size_t rows, const size_t cols)
   {
     if (W.is_empty())
       W.set_size(rows, cols);
@@ -67,8 +67,9 @@ class RandomInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Mat<eT>& W)
+  template<typename MatType>
+  void Initialize(MatType& W,
+      const typename std::enable_if_t<IsMatrix<MatType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty matrix." << std::endl;
@@ -86,8 +87,8 @@ class RandomInitialization
    * @param cols Number of columns.
    * @param slices Number of slices.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W,
+  template<typename CubeType>
+  void Initialize(CubeType& W,
                   const size_t rows,
                   const size_t cols,
                   const size_t slices)
@@ -104,8 +105,9 @@ class RandomInitialization
    *
    * @param W Weight matrix to initialize.
    */
-  template<typename eT>
-  void Initialize(arma::Cube<eT>& W)
+  template<typename CubeType>
+  void Initialize(CubeType& W,
+      const typename std::enable_if_t<IsCube<CubeType>::value>* = 0)
   {
     if (W.is_empty())
       Log::Fatal << "Cannot initialize an empty cube." << std::endl;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/add.hpp 4.4.0-1/src/mlpack/methods/ann/layer/add.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/add.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/add.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -94,7 +94,7 @@ class AddType : public Layer<MatType>
   void ComputeOutputDimensions();
 
   //! Set the weights of the layer to use the given memory.
-  void SetWeights(typename MatType::elem_type* weightPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Serialize the layer.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/add_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/add_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/add_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/add_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,7 +71,7 @@ AddType<MatType>::operator=(AddType&& ot
 template<typename MatType>
 void AddType<MatType>::Forward(const MatType& input, MatType& output)
 {
-  output = input + arma::repmat(arma::vectorise(weights), 1, input.n_cols);
+  output = input + repmat(vectorise(weights), 1, input.n_cols);
 }
 
 template<typename MatType>
@@ -90,14 +90,15 @@ void AddType<MatType>::Gradient(
     const MatType& error,
     MatType& gradient)
 {
-  gradient = error;
+  // The gradient is the sum of the error across all input points.
+  gradient = sum(error, 1);
 }
 
 template<typename MatType>
-void AddType<MatType>::SetWeights(typename MatType::elem_type* weightPtr)
+void AddType<MatType>::SetWeights(const MatType& weightsIn)
 {
   // Set the weights to wrap the given memory.
-  MakeAlias(weights, weightPtr, 1, outSize);
+  MakeAlias(weights, weightsIn, 1, outSize);
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/add_merge.hpp 4.4.0-1/src/mlpack/methods/ann/layer/add_merge.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/add_merge.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/add_merge.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -12,7 +12,6 @@
 #ifndef MLPACK_METHODS_ANN_LAYER_ADD_MERGE_HPP
 #define MLPACK_METHODS_ANN_LAYER_ADD_MERGE_HPP
 
-#include "../make_alias.hpp"
 #include "multi_layer.hpp"
 
 namespace mlpack {
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/alpha_dropout.hpp 4.4.0-1/src/mlpack/methods/ann/layer/alpha_dropout.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/alpha_dropout.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/alpha_dropout.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -114,7 +114,7 @@ class AlphaDropoutType : public Layer<Ma
   void Ratio(const double r)
   {
     ratio = r;
-    a = pow((1 - ratio) * (1 + ratio * pow(alphaDash, 2)), -0.5);
+    a = std::pow((1 - ratio) * (1 + ratio * std::pow(alphaDash, 2)), -0.5);
     b = -a * alphaDash * ratio;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/alpha_dropout_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/alpha_dropout_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/alpha_dropout_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/alpha_dropout_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -103,7 +103,7 @@ void AlphaDropoutType<MatType>::Forward(
     // Set values to alphaDash with probability ratio.  Then apply affine
     // transformation so as to keep mean and variance of outputs to their
     // original values.
-    mask = arma::randu<MatType>(input.n_rows, input.n_cols);
+    mask.randu(input.n_rows, input.n_cols);
     mask.transform( [&](double val) { return (val > ratio); } );
     output = (input % mask + alphaDash * (1 - mask)) * a + b;
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/batch_norm.hpp 4.4.0-1/src/mlpack/methods/ann/layer/batch_norm.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/batch_norm.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/batch_norm.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -118,7 +118,7 @@ class BatchNormType : public Layer<MatTy
   /**
    * Reset the layer parameters.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Initialize the weight matrix of the layer.
@@ -126,9 +126,7 @@ class BatchNormType : public Layer<MatTy
    * @param W Weight matrix to initialize.
    * @param elements Number of elements.
    */
-  void CustomInitialize(
-      MatType& W,
-      const size_t elements);
+  void CustomInitialize(MatType& W, const size_t elements);
 
   /**
    * Forward pass of the Batch Normalization layer. Transforms the input data
@@ -160,9 +158,7 @@ class BatchNormType : public Layer<MatTy
    * @param error The calculated error
    * @param gradient The calculated gradient.
    */
-  void Gradient(const MatType& input,
-                const MatType& error,
-                MatType& gradient);
+  void Gradient(const MatType& input, const MatType& error, MatType& gradient);
 
   //! Get the parameters.
   const MatType& Parameters() const { return weights; }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/batch_norm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/batch_norm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/batch_norm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/batch_norm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -148,14 +148,13 @@ BatchNormType<MatType>::operator=(
 }
 
 template<typename MatType>
-void BatchNormType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void BatchNormType<MatType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, WeightSize(), 1);
+  MakeAlias(weights, weightsIn, WeightSize(), 1);
   // Gamma acts as the scaling parameters for the normalized output.
-  MakeAlias(gamma, weightsPtr, size, 1);
+  MakeAlias(gamma, weightsIn, size, 1);
   // Beta acts as the shifting parameters for the normalized output.
-  MakeAlias(beta, weightsPtr + gamma.n_elem, size, 1);
+  MakeAlias(beta, weightsIn, size, 1, gamma.n_elem);
 }
 
 template<typename MatType>
@@ -170,9 +169,9 @@ void BatchNormType<MatType>::CustomIniti
   MatType gammaTemp;
   MatType betaTemp;
   // Gamma acts as the scaling parameters for the normalized output.
-  MakeAlias(gammaTemp, W.memptr(), size, 1);
+  MakeAlias(gammaTemp, W, size, 1);
   // Beta acts as the shifting parameters for the normalized output.
-  MakeAlias(betaTemp, W.memptr() + gammaTemp.n_elem, size, 1);
+  MakeAlias(betaTemp, W, size, 1, gammaTemp.n_elem);
 
   gammaTemp.fill(1.0);
   betaTemp.fill(0.0);
@@ -214,29 +213,25 @@ void BatchNormType<MatType>::Forward(
     outputTemp = inputTemp;
 
     // Calculate mean and variance over all channels.
-    MatType mean = arma::sum(arma::sum(inputTemp, 2), 0) / m;
-    variance = arma::sum(arma::sum(arma::pow(
-        inputTemp.each_slice() - arma::repmat(mean,
-        inputSize, 1), 2), 2), 0) / m;
+    MatType mean = sum(sum(inputTemp, 2), 0) / m;
+    variance = sum(sum(pow(
+        inputTemp.each_slice() - repmat(mean, inputSize, 1), 2), 2), 0) / m;
 
-    outputTemp.each_slice() -= arma::repmat(mean, inputSize, 1);
+    outputTemp.each_slice() -= repmat(mean, inputSize, 1);
 
     // Used in backward propagation.
     inputMean.set_size(arma::size(inputTemp));
     inputMean = outputTemp;
 
     // Normalize output.
-    outputTemp.each_slice() /= arma::sqrt(arma::repmat(variance,
-        inputSize, 1) + eps);
+    outputTemp.each_slice() /= sqrt(repmat(variance, inputSize, 1) + eps);
 
     // Re-used in backward propagation.
     normalized.set_size(arma::size(inputTemp));
     normalized = outputTemp;
 
-    outputTemp.each_slice() %= arma::repmat(gamma.t(),
-        inputSize, 1);
-    outputTemp.each_slice() += arma::repmat(beta.t(),
-        inputSize, 1);
+    outputTemp.each_slice() %= repmat(gamma.t(), inputSize, 1);
+    outputTemp.each_slice() += repmat(beta.t(), inputSize, 1);
 
     count += 1;
     // Value for average factor which used to update running parameters.
@@ -260,14 +255,11 @@ void BatchNormType<MatType>::Forward(
         const_cast<MatType&>(output).memptr(), inputSize, size,
         batchSize * higherDimension, false, false);
 
-    outputTemp.each_slice() -= arma::repmat(runningMean.t(),
-        inputSize, 1);
-    outputTemp.each_slice() /= arma::sqrt(arma::repmat(runningVariance.t(),
+    outputTemp.each_slice() -= repmat(runningMean.t(), inputSize, 1);
+    outputTemp.each_slice() /= sqrt(repmat(runningVariance.t(),
         inputSize, 1) + eps);
-    outputTemp.each_slice() %= arma::repmat(gamma.t(),
-        inputSize, 1);
-    outputTemp.each_slice() += arma::repmat(beta.t(),
-        inputSize, 1);
+    outputTemp.each_slice() %= repmat(gamma.t(), inputSize, 1);
+    outputTemp.each_slice() += repmat(beta.t(), inputSize, 1);
   }
 }
 
@@ -278,7 +270,7 @@ void BatchNormType<MatType>::Backward(
     const MatType& gy,
     MatType& g)
 {
-  const MatType stdInv = 1.0 / arma::sqrt(variance + eps);
+  const MatType stdInv = 1.0 / sqrt(variance + eps);
 
   const size_t batchSize = gy.n_cols;
   const size_t inputSize = inputDimension;
@@ -293,25 +285,24 @@ void BatchNormType<MatType>::Backward(
 
   // Step 1: dl / dxhat.
   arma::Cube<typename MatType::elem_type> norm =
-      gyTemp.each_slice() % arma::repmat(gamma.t(), inputSize, 1);
+      gyTemp.each_slice() % repmat(gamma.t(), inputSize, 1);
 
   // Step 2: sum dl / dxhat * (x - mu) * -0.5 * stdInv^3.
-  MatType temp = arma::sum(arma::sum(norm % inputMean, 2), 0);
-  MatType vars = temp % arma::pow(stdInv, 3) * (-0.5);
+  MatType temp = sum(sum(norm % inputMean, 2), 0);
+  MatType vars = temp % pow(stdInv, 3) * (-0.5);
 
   // Step 3: dl / dxhat * 1 / stdInv + variance * 2 * (x - mu) / m +
   // dl / dmu * 1 / m.
-  gTemp = (norm.each_slice() % arma::repmat(stdInv,
-      inputSize, 1)) +
-      ((inputMean.each_slice() % arma::repmat(vars, inputSize, 1) * 2.0) / m);
+  gTemp = (norm.each_slice() % repmat(stdInv, inputSize, 1)) +
+      ((inputMean.each_slice() % repmat(vars, inputSize, 1) * 2.0) / m);
 
   // Step 4: sum (dl / dxhat * -1 / stdInv) + variance *
   // sum (-2 * (x - mu)) / m.
-  MatType normTemp = arma::sum(arma::sum((norm.each_slice() %
-      arma::repmat(-stdInv, inputSize, 1)) + 
-      (inputMean.each_slice() % arma::repmat(vars, inputSize, 1) * (-2.0) / m),
+  MatType normTemp = sum(sum((norm.each_slice() %
+      repmat(-stdInv, inputSize, 1)) + 
+      (inputMean.each_slice() % repmat(vars, inputSize, 1) * (-2.0) / m),
       2), 0) / m;
-  gTemp.each_slice() += arma::repmat(normTemp, inputSize, 1);
+  gTemp.each_slice() += repmat(normTemp, inputSize, 1);
 }
 
 template<typename MatType>
@@ -327,11 +318,11 @@ void BatchNormType<MatType>::Gradient(
       error.n_cols * higherDimension, false, false);
 
   // Step 5: dl / dy * xhat.
-  MatType temp = arma::sum(arma::sum(normalized % errorTemp, 0), 2);
+  MatType temp = sum(sum(normalized % errorTemp, 0), 2);
   gradient.submat(0, 0, gamma.n_elem - 1, 0) = temp.t();
 
   // Step 6: dl / dy.
-  temp = arma::sum(arma::sum(errorTemp, 0), 2);
+  temp = sum(sum(errorTemp, 0), 2);
   gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) = temp.t();
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/concat_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/concat_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/concat_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/concat_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -123,19 +123,12 @@ void ConcatType<MatType>::Forward(const
       this->layerOutputs.size());
   for (size_t i = 0; i < this->layerOutputs.size(); ++i)
   {
-    MakeAlias(layerOutputAliases[i],
-              (typename MatType::elem_type*) this->layerOutputs[i].memptr(),
-              rows,
-              this->network[i]->OutputDimensions()[axis],
-              slices);
+    MakeAlias(layerOutputAliases[i], this->layerOutputs[i], rows, 
+        this->network[i]->OutputDimensions()[axis], slices);
   }
 
   arma::Cube<typename MatType::elem_type> outputAlias;
-  MakeAlias(outputAlias,
-            (typename MatType::elem_type*) output.memptr(),
-            rows,
-            this->outputDimensions[axis],
-            slices);
+  MakeAlias(outputAlias, output, rows, this->outputDimensions[axis], slices);
 
   // Now get the columns from each output.
   size_t startCol = 0;
@@ -171,11 +164,7 @@ void ConcatType<MatType>::Backward(
     slices *= this->outputDimensions[i];
 
   arma::Cube<typename MatType::elem_type> gyTmp;
-  MakeAlias(gyTmp,
-            (typename MatType::elem_type*) gy.memptr(),
-            rows,
-            this->outputDimensions[axis],
-            slices);
+  MakeAlias(gyTmp, gy, rows, this->outputDimensions[axis], slices);
 
   size_t startCol = 0;
   for (size_t i = 0; i < this->network.size(); ++i)
@@ -221,11 +210,7 @@ void ConcatType<MatType>::Backward(
     slices *= this->outputDimensions[i];
 
   arma::Cube<typename MatType::elem_type> gyTmp;
-  MakeAlias(gyTmp,
-            (typename MatType::elem_type*) gy.memptr(),
-            rows,
-            this->outputDimensions[axis],
-            slices);
+  MakeAlias(gyTmp, gy, rows, this->outputDimensions[axis], slices);
 
   size_t startCol = 0;
   for (size_t i = 0; i < index; ++i)
@@ -238,11 +223,7 @@ void ConcatType<MatType>::Backward(
   // Reshape so that the batch size is the number of columns.
   delta.reshape(delta.n_elem / gy.n_cols, gy.n_cols);
 
-  this->network[index]->Backward(
-          input,
-          this->layerOutputs[index],
-          delta,
-          g);
+  this->network[index]->Backward(input, this->layerOutputs[index], delta, g);
 }
 
 template<typename MatType>
@@ -263,11 +244,7 @@ void ConcatType<MatType>::Gradient(
     slices *= this->outputDimensions[i];
 
   arma::Cube<typename MatType::elem_type> errorTmp;
-  MakeAlias(errorTmp,
-            (typename MatType::elem_type*) error.memptr(),
-            rows,
-            this->outputDimensions[axis],
-            slices);
+  MakeAlias(errorTmp, error, rows, this->outputDimensions[axis], slices);
 
   size_t startCol = 0;
   size_t startParam = 0;
@@ -279,10 +256,7 @@ void ConcatType<MatType>::Gradient(
     MatType err = errorTmp.cols(startCol, startCol + cols - 1);
     err.reshape(err.n_elem / input.n_cols, input.n_cols);
     MatType gradientAlias;
-    MakeAlias(gradientAlias,
-              (typename MatType::elem_type*) gradient.memptr() + startParam,
-              params,
-              1);
+    MakeAlias(gradientAlias, gradient, params, 1, startParam);
     this->network[i]->Gradient(input, err, gradientAlias);
 
     startCol += cols;
@@ -309,11 +283,7 @@ void ConcatType<MatType>::Gradient(
     slices *= this->outputDimensions[i];
 
   arma::Cube<typename MatType::elem_type> errorTmp;
-  MakeAlias(errorTmp,
-            (typename MatType::elem_type*) error.memptr(),
-            rows,
-            this->outputDimensions[axis],
-            slices);
+  MakeAlias(errorTmp, error, rows, this->outputDimensions[axis], slices);
 
   size_t startCol = 0;
   size_t startParam = 0;
@@ -329,10 +299,7 @@ void ConcatType<MatType>::Gradient(
   MatType err = errorTmp.cols(startCol, startCol + cols - 1);
   err.reshape(err.n_elem / input.n_cols, input.n_cols);
   MatType gradientAlias;
-  MakeAlias(gradientAlias,
-            (typename MatType::elem_type*) gradient.memptr() + startParam,
-            params,
-            1);
+  MakeAlias(gradientAlias, gradient, params, 1, startParam);
   this->network[index]->Gradient(input, err, gradientAlias);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/concatenate_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/concatenate_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/concatenate_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/concatenate_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -82,7 +82,7 @@ void ConcatenateType<MatType>::Forward(c
 
   output.submat(0, 0, input.n_rows - 1, input.n_cols - 1) = input;
   output.submat(input.n_rows, 0, output.n_rows - 1, input.n_cols - 1) =
-      arma::repmat(arma::vectorise(concat), 1, input.n_cols);
+      repmat(vectorise(concat), 1, input.n_cols);
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/convolution.hpp 4.4.0-1/src/mlpack/methods/ann/layer/convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,6 +73,8 @@ template <
 class ConvolutionType : public Layer<MatType>
 {
  public:
+  typedef typename GetCubeType<MatType>::type CubeType;
+
   //! Create the ConvolutionType object.
   ConvolutionType();
 
@@ -153,7 +155,7 @@ class ConvolutionType : public Layer<Mat
   /*
    * Set the weight and bias term.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Ordinary feed forward pass of a neural network, evaluating the function
@@ -196,12 +198,12 @@ class ConvolutionType : public Layer<Mat
   MatType& Parameters() { return weights; }
 
   //! Get the weight of the layer as a cube.
-  arma::Cube<typename MatType::elem_type> const& Weight() const
+  CubeType const& Weight() const
   {
     return weight;
   }
   //! Modify the weight of the layer as a cube.
-  arma::Cube<typename MatType::elem_type>& Weight() { return weight; }
+  CubeType& Weight() { return weight; }
 
   //! Get the bias of the layer.
   MatType const& Bias() const { return bias; }
@@ -298,14 +300,13 @@ class ConvolutionType : public Layer<Mat
    * @param input The input data to be rotated.
    * @param output The rotated output.
    */
-  template<typename eT>
-  void Rotate180(const arma::Cube<eT>& input, arma::Cube<eT>& output)
+  void Rotate180(const CubeType& input, CubeType& output)
   {
-    output = arma::Cube<eT>(input.n_rows, input.n_cols, input.n_slices);
+    output = CubeType(input.n_rows, input.n_cols, input.n_slices);
 
     // * left-right flip, up-down flip */
     for (size_t s = 0; s < output.n_slices; s++)
-      output.slice(s) = arma::fliplr(arma::flipud(input.slice(s)));
+      output.slice(s) = fliplr(flipud(input.slice(s)));
   }
 
   /**
@@ -314,11 +315,10 @@ class ConvolutionType : public Layer<Mat
    * @param input The input data to be rotated.
    * @param output The rotated output.
    */
-  template<typename eT>
-  void Rotate180(const arma::Mat<eT>& input, arma::Mat<eT>& output)
+  void Rotate180(const MatType& input, MatType& output)
   {
     // * left-right flip, up-down flip */
-    output = arma::fliplr(arma::flipud(input));
+    output = fliplr(flipud(input));
   }
 
   //! Locally-stored number of output channels.
@@ -358,22 +358,22 @@ class ConvolutionType : public Layer<Mat
   MatType weights;
 
   //! Locally-stored weight object.
-  arma::Cube<typename MatType::elem_type> weight;
+  CubeType weight;
 
   //! Locally-stored bias term object.
   MatType bias;
 
   //! Locally-stored transformed output parameter.
-  arma::Cube<typename MatType::elem_type> outputTemp;
+  CubeType outputTemp;
 
   //! Locally-stored transformed padded input parameter.
   MatType inputPadded;
 
   //! Locally-stored transformed error parameter.
-  arma::Cube<typename MatType::elem_type> gTemp;
+  CubeType gTemp;
 
   //! Locally-stored transformed gradient parameter.
-  arma::Cube<typename MatType::elem_type> gradientTemp;
+  CubeType gradientTemp;
 
   //! Locally-stored padding layer.
   PaddingType<MatType> padding;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/convolution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/convolution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/convolution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/convolution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -272,17 +272,17 @@ void ConvolutionType<
     BackwardConvolutionRule,
     GradientConvolutionRule,
     MatType
->::SetWeights(typename MatType::elem_type* weightPtr)
+>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weight, weightPtr, kernelWidth, kernelHeight, maps * inMaps);
+  MakeAlias(weight, weightsIn, kernelWidth, kernelHeight, maps * inMaps);
   if (useBias)
   {
-    MakeAlias(bias, weightPtr + weight.n_elem, maps, 1);
-    MakeAlias(weights, weightPtr, weight.n_elem + bias.n_elem, 1);
+    MakeAlias(bias, weightsIn, maps, 1, weight.n_elem);
+    MakeAlias(weights, weightsIn, weight.n_elem + bias.n_elem, 1);
   }
   else
   {
-    MakeAlias(weights, weightPtr, weight.n_elem, 1);
+    MakeAlias(weights, weightsIn, weight.n_elem, 1);
   }
 }
 
@@ -313,12 +313,11 @@ void ConvolutionType<
     padding.Forward(input, inputPadded);
   }
 
-  arma::Cube<typename MatType::elem_type> inputTemp;
-  MakeAlias(inputTemp,
-      const_cast<MatType&>(usingPadding ? inputPadded : input).memptr(),
-      paddedRows, paddedCols, inMaps * higherInDimensions * batchSize);
+  CubeType inputTemp;
+  MakeAlias(inputTemp, (usingPadding ? inputPadded : input), paddedRows,
+      paddedCols, inMaps * higherInDimensions * batchSize);
 
-  MakeAlias(outputTemp, output.memptr(), this->outputDimensions[0],
+  MakeAlias(outputTemp, output, this->outputDimensions[0],
       this->outputDimensions[1], maps * higherInDimensions * batchSize);
   outputTemp.zeros();
 
@@ -375,27 +374,27 @@ void ConvolutionType<
     const MatType& gy,
     MatType& g)
 {
-  arma::Cube<typename MatType::elem_type> mappedError;
-  MakeAlias(mappedError, ((MatType&) gy).memptr(), this->outputDimensions[0],
+  CubeType mappedError;
+  MakeAlias(mappedError, gy, this->outputDimensions[0],
       this->outputDimensions[1], higherInDimensions * maps * batchSize);
 
-  MakeAlias(gTemp, g.memptr(), this->inputDimensions[0],
-      this->inputDimensions[1], inMaps * higherInDimensions * batchSize);
+  MakeAlias(gTemp, g, this->inputDimensions[0], this->inputDimensions[1],
+      inMaps * higherInDimensions * batchSize);
   gTemp.zeros();
 
   const bool usingPadding =
       (padWLeft != 0 || padWRight != 0 || padHTop != 0 || padHBottom != 0);
 
   // To perform the backward pass, we need to rotate all the filters.
-  arma::Cube<typename MatType::elem_type> rotatedFilters(weight.n_rows,
+  CubeType rotatedFilters(weight.n_rows,
       weight.n_cols, weight.n_slices);
 
   // To perform the backward pass, we need to dilate all the mappedError.
-  arma::Cube<typename MatType::elem_type> dilatedMappedError;
+  CubeType dilatedMappedError;
   if (strideHeight == 1 && strideWidth == 1)
   {
-    MakeAlias(dilatedMappedError, mappedError.memptr(),
-        mappedError.n_rows, mappedError.n_cols, mappedError.n_slices);
+    MakeAlias(dilatedMappedError, mappedError, mappedError.n_rows,
+        mappedError.n_cols, mappedError.n_slices);
   }
   else
   {
@@ -424,8 +423,8 @@ void ConvolutionType<
 
   MatType output(apparentWidth * apparentHeight * inMaps * higherInDimensions,
       batchSize, arma::fill::zeros);
-  arma::Cube<typename MatType::elem_type> outputCube;
-  MakeAlias(outputCube, output.memptr(), apparentWidth, apparentHeight,
+  CubeType outputCube;
+  MakeAlias(outputCube, output, apparentWidth, apparentHeight,
       inMaps * higherInDimensions * batchSize);
 
   // See Forward() for the overall iteration strategy.
@@ -456,8 +455,8 @@ void ConvolutionType<
   }
   MatType temp(padding.OutputDimensions()[0] * padding.OutputDimensions()[1] *
       inMaps * higherInDimensions, batchSize);
-  arma::Cube<typename MatType::elem_type> tempCube;
-  MakeAlias(tempCube, temp.memptr(), padding.OutputDimensions()[0],
+  CubeType tempCube;
+  MakeAlias(tempCube, temp, padding.OutputDimensions()[0],
       padding.OutputDimensions()[1], inMaps * higherInDimensions * batchSize);
   paddingBackward.Forward(output, temp);
   if (usingPadding)
@@ -490,10 +489,9 @@ void ConvolutionType<
     const MatType& error,
     MatType& gradient)
 {
-  arma::Cube<typename MatType::elem_type> mappedError;
-  MakeAlias(mappedError, ((MatType&) error).memptr(),
-      this->outputDimensions[0], this->outputDimensions[1],
-      higherInDimensions * maps * batchSize);
+  CubeType mappedError;
+  MakeAlias(mappedError, error, this->outputDimensions[0],
+      this->outputDimensions[1], higherInDimensions * maps * batchSize);
 
   // We are depending here on `inputPadded` being properly set from a call to
   // Forward().
@@ -502,14 +500,14 @@ void ConvolutionType<
   const size_t paddedRows = this->inputDimensions[0] + padWLeft + padWRight;
   const size_t paddedCols = this->inputDimensions[1] + padHTop + padHBottom;
 
-  arma::Cube<typename MatType::elem_type> inputTemp(
+  CubeType inputTemp(
       const_cast<MatType&>(usingPadding ? inputPadded : input).memptr(),
       paddedRows, paddedCols, inMaps * batchSize, false, false);
 
   MatType temp(apparentWidth * apparentHeight * inMaps * higherInDimensions,
       batchSize);
-  arma::Cube<typename MatType::elem_type> tempCube;
-  MakeAlias(tempCube, temp.memptr(), apparentWidth, apparentHeight,
+  CubeType tempCube;
+  MakeAlias(tempCube, temp, apparentWidth, apparentHeight,
       inMaps * higherInDimensions * batchSize);
   paddingBackward.Backward(input, {} /* unused */, usingPadding ? inputPadded : input, temp);
 
@@ -517,8 +515,7 @@ void ConvolutionType<
   // convolution map weights!  The bias will be handled by direct accesses into
   // `gradient`.
   gradient.zeros();
-  MakeAlias(gradientTemp, gradient.memptr(), weight.n_rows, weight.n_cols,
-      weight.n_slices);
+  MakeAlias(gradientTemp, gradient, weight.n_rows, weight.n_cols, weight.n_slices);
 
   // See Forward() for our iteration strategy.
   for (size_t offset = 0; offset < higherInDimensions * batchSize; ++offset)
@@ -544,7 +541,7 @@ void ConvolutionType<
       }
 
       if (useBias)
-        gradient[weight.n_elem + outMap] += arma::accu(curError);
+        gradient[weight.n_elem + outMap] += accu(curError);
     }
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/dropconnect.hpp 4.4.0-1/src/mlpack/methods/ann/layer/dropconnect.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/dropconnect.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/dropconnect.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -124,7 +124,7 @@ class DropConnectType : public Layer<Mat
   size_t WeightSize() const { return baseLayer->WeightSize(); }
 
   // Set the weights to use the given memory `weightsPtr`.
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Serialize the layer.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/dropconnect_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/dropconnect_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/dropconnect_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/dropconnect_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -114,7 +114,7 @@ void DropConnectType<MatType>::Forward(c
 
     // Scale with input / (1 - ratio) and set values to zero with
     // probability ratio.
-    mask = arma::randu<MatType>(denoise.n_rows, denoise.n_cols);
+    mask.randu(denoise.n_rows, denoise.n_cols);
     mask.transform([&](double val) { return (val > ratio); });
 
     baseLayer->Parameters() = denoise % mask;
@@ -155,10 +155,9 @@ void DropConnectType<MatType>::ComputeOu
 }
 
 template<typename MatType>
-void DropConnectType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void DropConnectType<MatType>::SetWeights(const MatType& weightsIn)
 {
-  baseLayer->SetWeights(weightsPtr);
+  baseLayer->SetWeights(weightsIn);
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/dropout_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/dropout_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/dropout_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/dropout_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -86,7 +86,7 @@ void DropoutType<MatType>::Forward(const
   {
     // Scale with input / (1 - ratio) and set values to zero with probability
     // 'ratio'.
-    mask = arma::randu<MatType>(input.n_rows, input.n_cols);
+    mask.randu(input.n_rows, input.n_cols);
     mask.transform([&](double val) { return (val > ratio); });
     output = input % mask * scale;
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/flexible_relu.hpp 4.4.0-1/src/mlpack/methods/ann/layer/flexible_relu.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/flexible_relu.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/flexible_relu.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -83,7 +83,7 @@ class FlexibleReLUType : public Layer<Ma
    * Reset the layer parameter (alpha). The method is called to
    * assign the allocated memory to the learnable layer parameter.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Initialize the weight matrix of the layer.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/flexible_relu_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/flexible_relu_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/flexible_relu_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/flexible_relu_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -74,10 +74,9 @@ FlexibleReLUType<MatType>::operator=(Fle
 }
 
 template<typename MatType>
-void FlexibleReLUType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void FlexibleReLUType<MatType>::SetWeights(const MatType& weights)
 {
-  MakeAlias(alpha, weightsPtr, 1, 1);
+  MakeAlias(alpha, weights, 1, 1);
 }
 
 template<typename MatType>
@@ -119,7 +118,7 @@ void FlexibleReLUType<MatType>::Gradient
     const MatType& error,
     MatType& gradient)
 {
-  gradient(0) = arma::accu(error) / input.n_cols;
+  gradient(0) = accu(error) / input.n_cols;
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/grouped_convolution.hpp 4.4.0-1/src/mlpack/methods/ann/layer/grouped_convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/grouped_convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/grouped_convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -77,6 +77,8 @@ template <
 class GroupedConvolutionType : public Layer<MatType>
 {
  public:
+  typedef typename GetCubeType<MatType>::type CubeType;
+
   //! Create the GroupedConvolutionType object.
   GroupedConvolutionType();
 
@@ -161,7 +163,7 @@ class GroupedConvolutionType : public La
   /*
    * Set the weight and bias term.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Ordinary feed forward pass of a neural network, evaluating the function
@@ -204,12 +206,12 @@ class GroupedConvolutionType : public La
   MatType& Parameters() { return weights; }
 
   //! Get the weight of the layer as a cube.
-  arma::Cube<typename MatType::elem_type> const& Weight() const
+  CubeType const& Weight() const
   {
     return weight;
   }
   //! Modify the weight of the layer as a cube.
-  arma::Cube<typename MatType::elem_type>& Weight() { return weight; }
+  CubeType& Weight() { return weight; }
 
   //! Get the bias of the layer.
   MatType const& Bias() const { return bias; }
@@ -311,14 +313,13 @@ class GroupedConvolutionType : public La
    * @param input The input data to be rotated.
    * @param output The rotated output.
    */
-  template<typename eT>
-  void Rotate180(const arma::Cube<eT>& input, arma::Cube<eT>& output)
+  void Rotate180(const CubeType& input, CubeType& output)
   {
-    output = arma::Cube<eT>(input.n_rows, input.n_cols, input.n_slices);
+    output = CubeType(input.n_rows, input.n_cols, input.n_slices);
 
     // * left-right flip, up-down flip */
     for (size_t s = 0; s < output.n_slices; s++)
-      output.slice(s) = arma::fliplr(arma::flipud(input.slice(s)));
+      output.slice(s) = fliplr(flipud(input.slice(s)));
   }
 
   /**
@@ -327,11 +328,10 @@ class GroupedConvolutionType : public La
    * @param input The input data to be rotated.
    * @param output The rotated output.
    */
-  template<typename eT>
-  void Rotate180(const arma::Mat<eT>& input, arma::Mat<eT>& output)
+  void Rotate180(const MatType& input, MatType& output)
   {
     // * left-right flip, up-down flip */
-    output = arma::fliplr(arma::flipud(input));
+    output = fliplr(flipud(input));
   }
 
   //! Locally-stored number of output channels.
@@ -374,22 +374,22 @@ class GroupedConvolutionType : public La
   MatType weights;
 
   //! Locally-stored weight object.
-  arma::Cube<typename MatType::elem_type> weight;
+  CubeType weight;
 
   //! Locally-stored bias term object.
   MatType bias;
 
   //! Locally-stored transformed output parameter.
-  arma::Cube<typename MatType::elem_type> outputTemp;
+  CubeType outputTemp;
 
   //! Locally-stored transformed padded input parameter.
   MatType inputPadded;
 
   //! Locally-stored transformed error parameter.
-  arma::Cube<typename MatType::elem_type> gTemp;
+  CubeType gTemp;
 
   //! Locally-stored transformed gradient parameter.
-  arma::Cube<typename MatType::elem_type> gradientTemp;
+  CubeType gradientTemp;
 
   //! Locally-stored padding layer.
   PaddingType<MatType> padding;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/grouped_convolution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/grouped_convolution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/grouped_convolution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/grouped_convolution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -281,18 +281,18 @@ void GroupedConvolutionType<
     BackwardConvolutionRule,
     GradientConvolutionRule,
     MatType
->::SetWeights(typename MatType::elem_type* weightPtr)
+>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weight, weightPtr, kernelWidth, kernelHeight, 
+  MakeAlias(weight, weightsIn, kernelWidth, kernelHeight, 
       (maps * inMaps) / groups);
   if (useBias)
   {
-    MakeAlias(bias, weightPtr + weight.n_elem, maps, 1);
-    MakeAlias(weights, weightPtr, weight.n_elem + bias.n_elem, 1);
+    MakeAlias(bias, weightsIn, maps, 1, weight.n_elem);
+    MakeAlias(weights, weightsIn, weight.n_elem + bias.n_elem, 1);
   }
   else
   {
-    MakeAlias(weights, weightPtr, weight.n_elem, 1);
+    MakeAlias(weights, weightsIn, weight.n_elem, 1);
   }
 }
 
@@ -323,12 +323,11 @@ void GroupedConvolutionType<
     padding.Forward(input, inputPadded);
   }
 
-  arma::Cube<typename MatType::elem_type> inputTemp;
-  MakeAlias(inputTemp,
-      const_cast<MatType&>(usingPadding ? inputPadded : input).memptr(),
+  CubeType inputTemp;
+  MakeAlias(inputTemp, (usingPadding ? inputPadded : input),
       paddedRows, paddedCols, inMaps * higherInDimensions * batchSize);
 
-  MakeAlias(outputTemp, output.memptr(), this->outputDimensions[0],
+  MakeAlias(outputTemp, output, this->outputDimensions[0],
       this->outputDimensions[1], maps * higherInDimensions * batchSize);
   outputTemp.zeros();
 
@@ -393,19 +392,19 @@ void GroupedConvolutionType<
     const MatType& gy,
     MatType& g)
 {
-  arma::Cube<typename MatType::elem_type> mappedError;
-  MakeAlias(mappedError, ((MatType&) gy).memptr(), this->outputDimensions[0],
+  CubeType mappedError;
+  MakeAlias(mappedError, gy, this->outputDimensions[0],
       this->outputDimensions[1], higherInDimensions * maps * batchSize);
 
-  MakeAlias(gTemp, g.memptr(), this->inputDimensions[0],
-      this->inputDimensions[1], inMaps * higherInDimensions * batchSize);
+  MakeAlias(gTemp, g, this->inputDimensions[0], this->inputDimensions[1],
+      inMaps * higherInDimensions * batchSize);
   gTemp.zeros();
 
   const bool usingPadding =
       (padWLeft != 0 || padWRight != 0 || padHTop != 0 || padHBottom != 0);
 
   // To perform the backward pass, we need to rotate all the filters.
-  arma::Cube<typename MatType::elem_type> rotatedFilters(weight.n_rows,
+  CubeType rotatedFilters(weight.n_rows,
       weight.n_cols, weight.n_slices);
 
   #pragma omp parallel for
@@ -415,11 +414,11 @@ void GroupedConvolutionType<
   }
 
   // To perform the backward pass, we need to dilate all the mappedError.
-  arma::Cube<typename MatType::elem_type> dilatedMappedError;
+  CubeType dilatedMappedError;
   if (strideHeight == 1 && strideWidth == 1)
   {
-    MakeAlias(dilatedMappedError, mappedError.memptr(),
-        mappedError.n_rows, mappedError.n_cols, mappedError.n_slices);
+    MakeAlias(dilatedMappedError, mappedError, mappedError.n_rows,
+        mappedError.n_cols, mappedError.n_slices);
   }
   else
   {
@@ -442,8 +441,8 @@ void GroupedConvolutionType<
 
   MatType output(apparentWidth * apparentHeight * inMaps * higherInDimensions,
       batchSize, arma::fill::zeros);
-  arma::Cube<typename MatType::elem_type> outputCube;
-  MakeAlias(outputCube, output.memptr(), apparentWidth, apparentHeight,
+  CubeType outputCube;
+  MakeAlias(outputCube, output, apparentWidth, apparentHeight,
       inMaps * higherInDimensions * batchSize);
 
   size_t inGroupSize = inMaps / groups;
@@ -483,8 +482,8 @@ void GroupedConvolutionType<
 
   MatType temp(padding.OutputDimensions()[0] * padding.OutputDimensions()[1] *
       inMaps * higherInDimensions, batchSize);
-  arma::Cube<typename MatType::elem_type> tempCube;
-  MakeAlias(tempCube, temp.memptr(), padding.OutputDimensions()[0],
+  CubeType tempCube;
+  MakeAlias(tempCube, temp, padding.OutputDimensions()[0],
       padding.OutputDimensions()[1], inMaps * higherInDimensions * batchSize);
   paddingBackward.Forward(output, temp);
   if (usingPadding)
@@ -517,10 +516,9 @@ void GroupedConvolutionType<
     const MatType& error,
     MatType& gradient)
 {
-  arma::Cube<typename MatType::elem_type> mappedError;
-  MakeAlias(mappedError, ((MatType&) error).memptr(),
-      this->outputDimensions[0], this->outputDimensions[1],
-      higherInDimensions * maps * batchSize);
+  CubeType mappedError;
+  MakeAlias(mappedError, error, this->outputDimensions[0],
+      this->outputDimensions[1], higherInDimensions * maps * batchSize);
 
   // We are depending here on `inputPadded` being properly set from a call to
   // Forward().
@@ -529,14 +527,14 @@ void GroupedConvolutionType<
   const size_t paddedRows = this->inputDimensions[0] + padWLeft + padWRight;
   const size_t paddedCols = this->inputDimensions[1] + padHTop + padHBottom;
 
-  arma::Cube<typename MatType::elem_type> inputTemp(
+  CubeType inputTemp(
       const_cast<MatType&>(usingPadding ? inputPadded : input).memptr(),
       paddedRows, paddedCols, inMaps * batchSize, false, false);
 
   MatType temp(apparentWidth * apparentHeight * inMaps * higherInDimensions,
       batchSize);
-  arma::Cube<typename MatType::elem_type> tempCube;
-  MakeAlias(tempCube, temp.memptr(), apparentWidth, apparentHeight,
+  CubeType tempCube;
+  MakeAlias(tempCube, temp, apparentWidth, apparentHeight,
       inMaps * higherInDimensions * batchSize);
   paddingBackward.Backward(input, {} /* unused */, usingPadding ? inputPadded : input, temp);
 
@@ -544,7 +542,7 @@ void GroupedConvolutionType<
   // convolution map weights!  The bias will be handled by direct accesses into
   // `gradient`.
   gradient.zeros();
-  MakeAlias(gradientTemp, gradient.memptr(), weight.n_rows, weight.n_cols,
+  MakeAlias(gradientTemp, gradient, weight.n_rows, weight.n_cols,
       weight.n_slices);
 
   size_t inGroupSize = inMaps / groups;
@@ -582,7 +580,7 @@ void GroupedConvolutionType<
 
         if (useBias)
           gradient[weight.n_elem + group * outGroupSize + outMap] += 
-              arma::accu(curError);
+              accu(curError);
       }
     }
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/layer.hpp 4.4.0-1/src/mlpack/methods/ann/layer/layer.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/layer.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/layer.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -81,7 +81,7 @@ class Layer
   { /* Nothing to do here */ }
 
   //! Copy assignment operator.  This is not responsible for copying weights!
-  virtual Layer& operator=(const Layer& layer)
+  Layer& operator=(const Layer& layer)
   {
     if (&layer != this)
     {
@@ -95,7 +95,7 @@ class Layer
   }
 
   //! Move assignment operator.  This is not responsible for moving weights!
-  virtual Layer& operator=(Layer&& layer)
+  Layer& operator=(Layer&& layer)
   {
     if (&layer != this)
     {
@@ -191,7 +191,7 @@ class Layer
    *    implementations should use MakeAlias() with weightsPtr to wrap the
    *    weights of a layer.
    */
-  virtual void SetWeights(typename MatType::elem_type* /* weightsPtr */) { }
+  virtual void SetWeights(const MatType& /* weightsIn */) { }
 
   /**
    * Get the total number of trainable weights in the layer.
@@ -220,7 +220,7 @@ class Layer
 
   //! Get the layer loss.  Overload this if the layer should add any extra loss
   //! to the loss function when computing the objective.  (TODO: better comment)
-  virtual double Loss() { return 0; }
+  virtual double Loss() const { return 0; }
 
   //! Get the input dimensions.
   const std::vector<size_t>& InputDimensions() const { return inputDimensions; }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/layer_norm.hpp 4.4.0-1/src/mlpack/methods/ann/layer/layer_norm.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/layer_norm.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/layer_norm.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -134,7 +134,7 @@ class LayerNormType : public Layer<MatTy
       size *= this->inputDimensions[i];
   }
 
-  void SetWeights(typename MatType::elem_type* /* weightsPtr */) override;
+  void SetWeights(const MatType& weightsIn) override;
 
   void CustomInitialize(
       MatType& /* W */,
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/layer_norm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/layer_norm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/layer_norm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/layer_norm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,12 +27,11 @@ LayerNormType<MatType>::LayerNormType(co
 }
 
 template<typename MatType>
-void LayerNormType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void LayerNormType<MatType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, 2 * size, 1);
-  MakeAlias(gamma, weightsPtr, size, 1);
-  MakeAlias(beta, weightsPtr + gamma.n_elem, size, 1);
+  MakeAlias(weights, weightsIn, 2 * size, 1);
+  MakeAlias(gamma, weightsIn, size, 1);
+  MakeAlias(beta, weightsIn, size, 1, gamma.n_elem);
 }
 
 template<typename MatType>
@@ -48,9 +47,9 @@ void LayerNormType<MatType>::CustomIniti
   MatType gammaTemp;
   MatType betaTemp;
   // Gamma acts as the scaling parameters for the normalized output.
-  MakeAlias(gammaTemp, W.memptr(), size, 1);
+  MakeAlias(gammaTemp, W, size, 1);
   // Beta acts as the shifting parameters for the normalized output.
-  MakeAlias(betaTemp, W.memptr() + gammaTemp.n_elem, size, 1);
+  MakeAlias(betaTemp, W, size, 1, gammaTemp.n_elem);
 
   gammaTemp.fill(1.0);
   betaTemp.fill(0.0);
@@ -66,7 +65,7 @@ void LayerNormType<MatType>::Forward(
   // Normalize the input.
   output = input.each_row() - mean;
   inputMean = output;
-  output.each_row() /= arma::sqrt(variance + eps);
+  output.each_row() /= sqrt(variance + eps);
 
   // Reused in the backward and gradient step.
   normalized = output;
@@ -83,14 +82,13 @@ void LayerNormType<MatType>::Backward(
     const MatType& gy,
     MatType& g)
 {
-  const MatType stdInv = 1.0 / arma::sqrt(variance + eps);
+  const MatType stdInv = 1.0 / sqrt(variance + eps);
 
   // dl / dxhat.
   const MatType norm = gy.each_col() % gamma;
 
   // sum dl / dxhat * (x - mu) * -0.5 * stdInv^3.
-  const MatType var = arma::sum(norm % inputMean, 0) %
-      arma::pow(stdInv, 3.0) * -0.5;
+  const MatType var = sum(norm % inputMean, 0) % pow(stdInv, 3.0) * -0.5;
 
   // dl / dxhat * 1 / stdInv + variance * 2 * (x - mu) / m +
   // dl / dmu * 1 / m.
@@ -99,7 +97,7 @@ void LayerNormType<MatType>::Backward(
 
   // sum (dl / dxhat * -1 / stdInv) + variance *
   // (sum -2 * (x - mu)) / m.
-  g.each_row() += arma::sum(norm.each_row() % -stdInv, 0) / gy.n_rows;
+  g.each_row() += sum(norm.each_row() % -stdInv, 0) / gy.n_rows;
 }
 
 template<typename MatType>
@@ -111,11 +109,10 @@ void LayerNormType<MatType>::Gradient(
   gradient.set_size(size + size, 1);
 
   // Step 5: dl / dy * xhat.
-  gradient.submat(0, 0, gamma.n_elem - 1, 0) = arma::sum(normalized % error, 1);
+  gradient.submat(0, 0, gamma.n_elem - 1, 0) = sum(normalized % error, 1);
 
   // Step 6: dl / dy.
-  gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) =
-      arma::sum(error, 1);
+  gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) = sum(error, 1);
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/layer_types.hpp 4.4.0-1/src/mlpack/methods/ann/layer/layer_types.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/layer_types.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/layer_types.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,6 +51,7 @@
 #include <mlpack/methods/ann/layer/parametric_relu.hpp>
 #include <mlpack/methods/ann/layer/radial_basis_function.hpp>
 #include <mlpack/methods/ann/layer/relu6.hpp>
+#include <mlpack/methods/ann/layer/repeat.hpp>
 #include <mlpack/methods/ann/layer/softmax.hpp>
 #include <mlpack/methods/ann/layer/softmin.hpp>
 #include <mlpack/methods/ann/layer/ftswish.hpp>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -76,7 +76,7 @@ class LinearType : public Layer<MatType>
    * Reset the layer parameter (weights and bias). The method is called to
    * assign the allocated memory to the internal learnable parameters.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Ordinary feed forward pass of a neural network, evaluating the function
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear3d.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear3d.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear3d.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear3d.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -68,7 +68,7 @@ class Linear3DType : public Layer<MatTyp
   /*
    * Reset the layer parameter.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Ordinary feed forward pass of a neural network, evaluating the function
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear3d_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear3d_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear3d_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear3d_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -85,13 +85,12 @@ Linear3DType<MatType, RegularizerType>::
 }
 
 template<typename MatType, typename RegularizerType>
-void Linear3DType<MatType, RegularizerType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void Linear3DType<MatType, RegularizerType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, outSize * this->inputDimensions[0] + outSize,
+  MakeAlias(weights, weightsIn, outSize * this->inputDimensions[0] + outSize,
       1);
-  MakeAlias(weight, weightsPtr, outSize, this->inputDimensions[0]);
-  MakeAlias(bias, weightsPtr + weight.n_elem, outSize, 1);
+  MakeAlias(weight, weightsIn, outSize, this->inputDimensions[0]);
+  MakeAlias(bias, weightsIn, outSize, 1, weight.n_elem);
 }
 
 template<typename MatType, typename RegularizerType>
@@ -112,7 +111,7 @@ void Linear3DType<MatType, RegularizerTy
     // Shape of inputTemp : (inSize, nPoints, batchSize).
     MatType z = weight * inputTemp.slice(i);
     z.each_col() += bias;
-    output.col(i) = arma::vectorise(z);
+    output.col(i) = vectorise(z);
   }
 }
 
@@ -141,7 +140,7 @@ void Linear3DType<MatType, RegularizerTy
   {
     // Shape of weight : (outSize, inSize).
     // Shape of gyTemp : (outSize, nPoints, batchSize).
-    g.col(i) = arma::vectorise(weight.t() * gyTemp.slice(i));
+    g.col(i) = vectorise(weight.t() * gyTemp.slice(i));
   }
 }
 
@@ -172,11 +171,10 @@ void Linear3DType<MatType, RegularizerTy
     dW.slice(i) = errorTemp.slice(i) * inputTemp.slice(i).t();
   }
 
-  gradient.submat(0, 0, weight.n_elem - 1, 0)
-      = arma::vectorise(arma::sum(dW, 2));
+  gradient.submat(0, 0, weight.n_elem - 1, 0) = vectorise(sum(dW, 2));
 
   gradient.submat(weight.n_elem, 0, weights.n_elem - 1, 0)
-      = arma::vectorise(arma::sum(arma::sum(errorTemp, 2), 1));
+      = vectorise(sum(sum(errorTemp, 2), 1));
 
   regularizer.Evaluate(weights, gradient);
 }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -93,12 +93,11 @@ LinearType<MatType, RegularizerType>::op
 }
 
 template<typename MatType, typename RegularizerType>
-void LinearType<MatType, RegularizerType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void LinearType<MatType, RegularizerType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, outSize * inSize + outSize, 1);
-  MakeAlias(weight, weightsPtr, outSize, inSize);
-  MakeAlias(bias, weightsPtr + weight.n_elem, outSize, 1);
+  MakeAlias(weights, weightsIn, outSize * inSize + outSize, 1);
+  MakeAlias(weight, weightsIn, outSize, inSize);
+  MakeAlias(bias, weightsIn, outSize, 1, weight.n_elem);
 }
 
 template<typename MatType, typename RegularizerType>
@@ -128,11 +127,8 @@ void LinearType<MatType, RegularizerType
     const MatType& error,
     MatType& gradient)
 {
-  gradient.submat(0, 0, weight.n_elem - 1, 0) = arma::vectorise(
-      error * input.t());
-  gradient.submat(weight.n_elem, 0, gradient.n_elem - 1, 0) =
-      arma::sum(error, 1);
-
+  gradient.submat(0, 0, weight.n_elem - 1, 0) = vectorise(error * input.t());
+  gradient.submat(weight.n_elem, 0, gradient.n_elem - 1, 0) = sum(error, 1);
   regularizer.Evaluate(weights, gradient);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear_no_bias.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear_no_bias.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear_no_bias.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear_no_bias.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,7 +52,7 @@ class LinearNoBiasType : public Layer<Ma
   LinearNoBiasType* Clone() const { return new LinearNoBiasType(*this); }
 
   //! Reset the layer parameter.
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   //! Copy constructor.
   LinearNoBiasType(const LinearNoBiasType& layer);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/linear_no_bias_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/linear_no_bias_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/linear_no_bias_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/linear_no_bias_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -95,9 +95,9 @@ LinearNoBiasType<MatType, RegularizerTyp
 
 template<typename MatType, typename RegularizerType>
 void LinearNoBiasType<MatType, RegularizerType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+    const MatType& weights)
 {
-  MakeAlias(weight, weightsPtr, outSize, inSize);
+  MakeAlias(weight, weights, outSize, inSize);
 }
 
 template<typename MatType, typename RegularizerType>
@@ -123,8 +123,7 @@ void LinearNoBiasType<MatType, Regulariz
     const MatType& error,
     MatType& gradient)
 {
-  gradient.submat(0, 0, weight.n_elem - 1, 0) = arma::vectorise(
-      error * input.t());
+  gradient.submat(0, 0, weight.n_elem - 1, 0) = vectorise(error * input.t());
   regularizer.Evaluate(weight, gradient);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/log_softmax.hpp 4.4.0-1/src/mlpack/methods/ann/layer/log_softmax.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/log_softmax.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/log_softmax.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,13 +53,30 @@ class LogSoftMaxType : public Layer<MatT
   LogSoftMaxType& operator=(LogSoftMaxType&& other);
 
   /**
+   * A wrapper function to call the correct implementation according to the
+   * specific matrix type (e.g., arma, coot).
+   *
+   * @param input Input data used for evaluating the specified function.
+   * @param output Resulting output activation.
+   */
+  void Forward(const MatType& input, MatType& output);
+ 
+  /**
    * Ordinary feed forward pass of a neural network, evaluating the function
    * f(x) by propagating the activity forward through f.
    *
    * @param input Input data used for evaluating the specified function.
    * @param output Resulting output activation.
    */
-  void Forward(const MatType& input, MatType& output);
+  void ForwardImpl(const MatType& input, MatType& output,
+                   const typename std::enable_if_t<
+                       arma::is_arma_type<MatType>::value>* = 0);
+
+#ifdef MLPACK_HAS_COOT
+  void ForwardImpl(const MatType& input, MatType& output,
+                   const typename std::enable_if_t<
+                       coot::is_coot_type<MatType>::value>* = 0);
+#endif
 
   /**
    * Ordinary feed backward pass of a neural network, calculating the function
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/log_softmax_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/log_softmax_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/log_softmax_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/log_softmax_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,7 +65,16 @@ LogSoftMaxType<MatType>::operator=(LogSo
 template<typename MatType>
 void LogSoftMaxType<MatType>::Forward(const MatType& input, MatType& output)
 {
-  MatType maxInput = arma::repmat(arma::max(input), input.n_rows, 1);
+  ForwardImpl(input, output);
+}
+
+template<typename MatType>
+void LogSoftMaxType<MatType>::ForwardImpl(const MatType& input,
+                                          MatType& output,
+                                          const typename std::enable_if_t<
+                                              arma::is_arma_type<MatType>::value>*)
+{
+  MatType maxInput = repmat(max(input), input.n_rows, 1);
   output = (maxInput - input);
 
   // Approximation of the base-e exponential function. The acuracy however is
@@ -92,11 +101,27 @@ void LogSoftMaxType<MatType>::Forward(co
 
     return 0.0;
   });
+  maxInput.each_row() += log(sum(output));
+  output = input - maxInput;
+}
 
-  maxInput.each_row() += arma::log(arma::sum(output));
+#ifdef MLPACK_HAS_COOT
+
+template<typename MatType>
+void LogSoftMaxType<MatType>::ForwardImpl(const MatType& input,
+                                          MatType& output,
+                                          const typename std::enable_if_t<
+                                              coot::is_coot_type<MatType>::value>*)
+{
+  MatType maxInput = repmat(max(input), input.n_rows, 1);
+  output = (maxInput - input);
+  output = exp(output * -1);
+  maxInput.each_row() += log(sum(output));
   output = input - maxInput;
 }
 
+#endif
+
 template<typename MatType>
 void LogSoftMaxType<MatType>::Backward(
     const MatType& /* input */,
@@ -104,7 +129,7 @@ void LogSoftMaxType<MatType>::Backward(
     const MatType& gy,
     MatType& g)
 {
-  g = gy - arma::exp(output) % arma::repmat(arma::sum(gy), output.n_rows, 1);
+  g = gy - exp(output) % repmat(sum(gy), output.n_rows, 1);
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/lstm.hpp 4.4.0-1/src/mlpack/methods/ann/layer/lstm.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/lstm.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/lstm.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -87,7 +87,7 @@ class LSTMType : public RecurrentLayer<M
    * Reset the layer parameter. The method is called to
    * assign the allocated memory to the internal learnable parameters.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Ordinary feed-forward pass of a neural network, evaluating the function
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/lstm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/lstm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/lstm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/lstm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -93,59 +93,58 @@ void LSTMType<MatType>::ClearRecurrentSt
 }
 
 template<typename MatType>
-void LSTMType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void LSTMType<MatType>::SetWeights(const MatType& weights)
 {
   // Set the weight parameter for the output gate.
-  MakeAlias(input2GateOutputWeight, weightsPtr, outSize, inSize);
+  MakeAlias(input2GateOutputWeight, weights, outSize, inSize);
   size_t offset = input2GateOutputWeight.n_elem;
-  MakeAlias(input2GateOutputBias, weightsPtr + offset, outSize, 1);
+  MakeAlias(input2GateOutputBias, weights, outSize, 1, offset);
   offset += input2GateOutputBias.n_elem;
 
   // Set the weight parameter for the forget gate.
-  MakeAlias(input2GateForgetWeight, weightsPtr + offset, outSize, inSize);
+  MakeAlias(input2GateForgetWeight, weights, outSize, inSize, offset);
   offset += input2GateForgetWeight.n_elem;
-  MakeAlias(input2GateForgetBias, weightsPtr + offset, outSize, 1);
+  MakeAlias(input2GateForgetBias, weights, outSize, 1, offset);
   offset += input2GateForgetBias.n_elem;
 
   // Set the weight parameter for the input gate.
-  MakeAlias(input2GateInputWeight, weightsPtr + offset, outSize, inSize);
+  MakeAlias(input2GateInputWeight, weights, outSize, inSize, offset);
   offset += input2GateInputWeight.n_elem;
-  MakeAlias(input2GateInputBias, weightsPtr + offset, outSize, 1);
+  MakeAlias(input2GateInputBias, weights, outSize, 1, offset);
   offset += input2GateInputBias.n_elem;
 
   // Set the weight parameter for the hidden gate.
-  MakeAlias(input2HiddenWeight, weightsPtr + offset, outSize, inSize);
+  MakeAlias(input2HiddenWeight, weights, outSize, inSize, offset);
   offset += input2HiddenWeight.n_elem;
-  MakeAlias(input2HiddenBias, weightsPtr + offset, outSize, 1);
+  MakeAlias(input2HiddenBias, weights, outSize, 1, offset);
   offset += input2HiddenBias.n_elem;
 
   // Set the weight parameter for the output multiplication.
-  MakeAlias(output2GateOutputWeight, weightsPtr + offset, outSize, outSize);
+  MakeAlias(output2GateOutputWeight, weights, outSize, outSize, offset);
   offset += output2GateOutputWeight.n_elem;
 
   // Set the weight parameter for the output multiplication.
-  MakeAlias(output2GateForgetWeight, weightsPtr + offset, outSize, outSize);
+  MakeAlias(output2GateForgetWeight, weights, outSize, outSize, offset);
   offset += output2GateForgetWeight.n_elem;
 
   // Set the weight parameter for the input multiplication.
-  MakeAlias(output2GateInputWeight, weightsPtr + offset, outSize, outSize);
+  MakeAlias(output2GateInputWeight, weights, outSize, outSize, offset);
   offset += output2GateInputWeight.n_elem;
 
   // Set the weight parameter for the hidden multiplication.
-  MakeAlias(output2HiddenWeight, weightsPtr + offset, outSize, outSize);
+  MakeAlias(output2HiddenWeight, weights, outSize, outSize, offset);
   offset += output2HiddenWeight.n_elem;
 
   // Set the weight parameter for the cell multiplication.
-  MakeAlias(cell2GateOutputWeight, weightsPtr + offset, outSize, 1);
+  MakeAlias(cell2GateOutputWeight, weights, outSize, 1, offset);
   offset += cell2GateOutputWeight.n_elem;
 
   // Set the weight parameter for the cell - forget gate multiplication.
-  MakeAlias(cell2GateForgetWeight, weightsPtr + offset, outSize, 1);
+  MakeAlias(cell2GateForgetWeight, weights, outSize, 1, offset);
   offset += cell2GateOutputWeight.n_elem;
 
   // Set the weight parameter for the cell - input gate multiplication.
-  MakeAlias(cell2GateInputWeight, weightsPtr + offset, outSize, 1);
+  MakeAlias(cell2GateInputWeight, weights, outSize, 1, offset);
 }
 
 // Forward when cellState is not needed.
@@ -173,17 +172,17 @@ void LSTMType<MatType>::Forward(const Ma
 
   if (this->HasPreviousStep())
   {
-    inputGate += arma::repmat(cell2GateInputWeight, 1, batchSize) %
+    inputGate += repmat(cell2GateInputWeight, 1, batchSize) %
         cell.slice(this->PreviousStep());
 
-    forgetGate += arma::repmat(cell2GateForgetWeight, 1, batchSize) %
+    forgetGate += repmat(cell2GateForgetWeight, 1, batchSize) %
         cell.slice(this->PreviousStep());
   }
 
   inputGateActivation.slice(this->CurrentStep()) =
-      1.0 / (1.0 + arma::exp(-inputGate));
+      1.0 / (1.0 + exp(-inputGate));
   forgetGateActivation.slice(this->CurrentStep()) =
-      1.0 / (1.0 + arma::exp(-forgetGate));
+      1.0 / (1.0 + exp(-forgetGate));
 
   hiddenLayer = input2HiddenWeight * input;
   if (this->HasPreviousStep())
@@ -220,7 +219,7 @@ void LSTMType<MatType>::Forward(const Ma
   outputGate.each_col() += input2GateOutputBias;
 
   outputGateActivation.slice(this->CurrentStep()) =
-      1.0 / (1.0 + arma::exp(-outputGate));
+      1.0 / (1.0 + exp(-outputGate));
 
   cellActivation.slice(this->CurrentStep()) =
       arma::tanh(cell.slice(this->CurrentStep()));
@@ -264,7 +263,7 @@ void LSTMType<MatType>::Backward(
 
   MatType cellError = gyLocal %
       outputGateActivation.slice(this->CurrentStep()) %
-      (1 - arma::pow(cellActivation.slice(this->CurrentStep()), 2)) +
+      (1 - pow(cellActivation.slice(this->CurrentStep()), 2)) +
       outputGateError.each_col() % cell2GateOutputWeight;
 
   if (this->HasPreviousStep())
@@ -289,7 +288,7 @@ void LSTMType<MatType>::Backward(
       (1.0 - inputGateActivation.slice(this->CurrentStep())));
 
   hiddenError = inputGateActivation.slice(this->CurrentStep()) % cellError %
-      (1 - arma::pow(hiddenLayerActivation.slice(this->CurrentStep()), 2));
+      (1 - pow(hiddenLayerActivation.slice(this->CurrentStep()), 2));
 
   inputCellError = forgetGateActivation.slice(this->CurrentStep()) % cellError +
       forgetGateError.each_col() % cell2GateForgetWeight +
@@ -312,73 +311,69 @@ void LSTMType<MatType>::Gradient(
 
   // Input2GateOutputWeight and input2GateOutputBias gradients.
   gradient.submat(0, 0, input2GateOutputWeight.n_elem - 1, 0) =
-      arma::vectorise(outputGateError * input.t());
+      vectorise(outputGateError * input.t());
   gradient.submat(input2GateOutputWeight.n_elem, 0,
       input2GateOutputWeight.n_elem + input2GateOutputBias.n_elem - 1, 0) =
-      arma::sum(outputGateError, 1);
+      sum(outputGateError, 1);
   size_t offset = input2GateOutputWeight.n_elem + input2GateOutputBias.n_elem;
 
   // input2GateForgetWeight and input2GateForgetBias gradients.
   gradient.submat(offset, 0, offset + input2GateForgetWeight.n_elem - 1, 0) =
-      arma::vectorise(forgetGateError * input.t());
+      vectorise(forgetGateError * input.t());
   gradient.submat(offset + input2GateForgetWeight.n_elem, 0,
       offset + input2GateForgetWeight.n_elem +
-      input2GateForgetBias.n_elem - 1, 0) = arma::sum(forgetGateError, 1);
+      input2GateForgetBias.n_elem - 1, 0) = sum(forgetGateError, 1);
   offset += input2GateForgetWeight.n_elem + input2GateForgetBias.n_elem;
 
   // input2GateInputWeight and input2GateInputBias gradients.
   gradient.submat(offset, 0, offset + input2GateInputWeight.n_elem - 1, 0) =
-      arma::vectorise(inputGateError * input.t());
+      vectorise(inputGateError * input.t());
   gradient.submat(offset + input2GateInputWeight.n_elem, 0,
       offset + input2GateInputWeight.n_elem +
-      input2GateInputBias.n_elem - 1, 0) = arma::sum(inputGateError, 1);
+      input2GateInputBias.n_elem - 1, 0) = sum(inputGateError, 1);
   offset += input2GateInputWeight.n_elem + input2GateInputBias.n_elem;
 
   // input2HiddenWeight and input2HiddenBias gradients.
   gradient.submat(offset, 0, offset + input2HiddenWeight.n_elem - 1, 0) =
-      arma::vectorise(hiddenError * input.t());
+      vectorise(hiddenError * input.t());
   gradient.submat(offset + input2HiddenWeight.n_elem, 0,
       offset + input2HiddenWeight.n_elem + input2HiddenBias.n_elem - 1, 0) =
-      arma::sum(hiddenError, 1);
+      sum(hiddenError, 1);
   offset += input2HiddenWeight.n_elem + input2HiddenBias.n_elem;
 
   // output2GateOutputWeight gradients.
   gradient.submat(offset, 0, offset + output2GateOutputWeight.n_elem - 1, 0) =
-      arma::vectorise(outputGateError *
-      outParameter.slice(this->CurrentStep()).t());
+      vectorise(outputGateError * outParameter.slice(this->CurrentStep()).t());
   offset += output2GateOutputWeight.n_elem;
 
   // output2GateForgetWeight gradients.
   gradient.submat(offset, 0, offset + output2GateForgetWeight.n_elem - 1, 0) =
-      arma::vectorise(forgetGateError *
-      outParameter.slice(this->CurrentStep()).t());
+      vectorise(forgetGateError * outParameter.slice(this->CurrentStep()).t());
   offset += output2GateForgetWeight.n_elem;
 
   // output2GateInputWeight gradients.
   gradient.submat(offset, 0, offset + output2GateInputWeight.n_elem - 1, 0) =
-      arma::vectorise(inputGateError *
-      outParameter.slice(this->CurrentStep()).t());
+      vectorise(inputGateError * outParameter.slice(this->CurrentStep()).t());
   offset += output2GateInputWeight.n_elem;
 
   // output2HiddenWeight gradients.
   gradient.submat(offset, 0, offset + output2HiddenWeight.n_elem - 1, 0) =
-      arma::vectorise(hiddenError *
-      outParameter.slice(this->CurrentStep()).t());
+      vectorise(hiddenError * outParameter.slice(this->CurrentStep()).t());
   offset += output2HiddenWeight.n_elem;
 
   // cell2GateOutputWeight gradients.
   gradient.submat(offset, 0, offset + cell2GateOutputWeight.n_elem - 1, 0) =
-      arma::sum(outputGateError % cell.slice(this->CurrentStep()), 1);
+      sum(outputGateError % cell.slice(this->CurrentStep()), 1);
   offset += cell2GateOutputWeight.n_elem;
 
   // cell2GateForgetWeight and cell2GateInputWeight gradients.
   if (this->HasPreviousStep())
   {
     gradient.submat(offset, 0, offset + cell2GateForgetWeight.n_elem - 1, 0) =
-        arma::sum(forgetGateError % cell.slice(this->PreviousStep()), 1);
+        sum(forgetGateError % cell.slice(this->PreviousStep()), 1);
     gradient.submat(offset + cell2GateForgetWeight.n_elem, 0, offset +
         cell2GateForgetWeight.n_elem + cell2GateInputWeight.n_elem - 1, 0) =
-        arma::sum(inputGateError % cell.slice(this->PreviousStep()), 1);
+        sum(inputGateError % cell.slice(this->PreviousStep()), 1);
   }
   else
   {
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/max_pooling.hpp 4.4.0-1/src/mlpack/methods/ann/layer/max_pooling.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/max_pooling.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/max_pooling.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,7 +35,7 @@ class MaxPoolingRule
   template<typename MatType>
   typename MatType::elem_type Pooling(const MatType& input)
   {
-    return arma::max(arma::vectorise(input));
+    return max(vectorise(input));
   }
 
   template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/mean_pooling.hpp 4.4.0-1/src/mlpack/methods/ann/layer/mean_pooling.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/mean_pooling.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/mean_pooling.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -148,7 +148,7 @@ class MeanPoolingType : public Layer<Mat
    */
   typename MatType::elem_type Pooling(const MatType& input)
   {
-    return arma::mean(arma::vectorise(input));
+    return arma::mean(vectorise(input));
   }
 
   //! Locally-stored width of the pooling window.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/multi_layer.hpp 4.4.0-1/src/mlpack/methods/ann/layer/multi_layer.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/multi_layer.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/multi_layer.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -12,7 +12,6 @@
 #ifndef MLPACK_METHODS_ANN_LAYER_MULTI_LAYER_HPP
 #define MLPACK_METHODS_ANN_LAYER_MULTI_LAYER_HPP
 
-#include "../make_alias.hpp"
 #include "layer.hpp"
 
 namespace mlpack {
@@ -129,7 +128,7 @@ class MultiLayer : public Layer<MatType>
   /**
    * Set the weights of the layer to use the memory given as `weightsPtr`.
    */
-  virtual void SetWeights(typename MatType::elem_type* weightsPtr);
+  virtual void SetWeights(const MatType& weightsIn);
 
   /**
    * Initialize the weight matrix of the layer.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/multi_layer_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/multi_layer_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/multi_layer_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/multi_layer_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -235,7 +235,7 @@ void MultiLayer<MatType>::Gradient(
 }
 
 template<typename MatType>
-void MultiLayer<MatType>::SetWeights(typename MatType::elem_type* weightsPtr)
+void MultiLayer<MatType>::SetWeights(const MatType& weightsIn)
 {
   size_t start = 0;
   const size_t totalWeightSize = WeightSize();
@@ -248,8 +248,9 @@ void MultiLayer<MatType>::SetWeights(typ
     Log::Assert(start + weightSize <= totalWeightSize,
         "FNN::SetLayerMemory(): parameter size does not match total layer "
         "weight size!");
-
-    network[i]->SetWeights(weightsPtr + start);
+    MatType tmpWeights;
+    MakeAlias(tmpWeights, weightsIn, weightSize, 1, start);
+    network[i]->SetWeights(tmpWeights);
     start += weightSize;
   }
 
@@ -278,7 +279,7 @@ void MultiLayer<MatType>::CustomInitiali
         "weight size!");
     
     MatType WTemp;
-    MakeAlias(WTemp, W.memptr() + start, weightSize, 1);
+    MakeAlias(WTemp, W, weightSize, 1, start);
     network[i]->CustomInitialize(WTemp, weightSize);
     
     start += weightSize;
@@ -388,8 +389,8 @@ void MultiLayer<MatType>::InitializeForw
   for (size_t i = 0; i < layerOutputs.size(); ++i)
   {
     const size_t layerOutputSize = network[i]->OutputSize();
-    MakeAlias(layerOutputs[i], layerOutputMatrix.colptr(start),
-        layerOutputSize, batchSize);
+    MakeAlias(layerOutputs[i], layerOutputMatrix, layerOutputSize, batchSize,
+        start * layerOutputMatrix.n_rows);
     start += batchSize * layerOutputSize;
   }
 }
@@ -417,8 +418,8 @@ void MultiLayer<MatType>::InitializeBack
     for (size_t j = 0; j < this->network[i]->InputDimensions().size(); ++j)
       layerInputSize *= this->network[i]->InputDimensions()[j];
 
-    MakeAlias(layerDeltas[i], layerDeltaMatrix.colptr(start), layerInputSize,
-        batchSize);
+    MakeAlias(layerDeltas[i], layerDeltaMatrix, layerInputSize,
+        batchSize, start * layerDeltaMatrix.n_rows);
     start += batchSize * layerInputSize;
   }
 }
@@ -432,8 +433,7 @@ void MultiLayer<MatType>::InitializeGrad
   for (size_t i = 0; i < network.size(); ++i)
   {
     const size_t weightSize = network[i]->WeightSize();
-    MakeAlias(layerGradients[i], gradient.memptr() + gradientStart,
-        weightSize, 1);
+    MakeAlias(layerGradients[i], gradient, weightSize, 1, gradientStart);
     gradientStart += weightSize;
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/multihead_attention.hpp 4.4.0-1/src/mlpack/methods/ann/layer/multihead_attention.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/multihead_attention.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/multihead_attention.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -77,8 +77,7 @@ class MultiheadAttentionType : public La
   /**
    * Create the MultiheadAttention object using the specified modules.
    * The attnMask and keyPaddingMask arguments are defined as a mask
-   * where 1 passes values and 0 masks them, unlike the original
-   * paper which uses 0 to pass values and -inf to mask them.
+   * 0 passes values and -inf masks them, as in the original paper.
    *
    * @param tgtSeqLen Target sequence length.
    * @param numHeads Number of parallel attention heads.
@@ -103,7 +102,7 @@ class MultiheadAttentionType : public La
   /**
    * Reset the layer parameters.
    */
-  void SetWeights(typename MatType::elem_type* weightsPtr) override;
+  void SetWeights(const MatType& weightsIn) override;
 
   /**
    * Ordinary feed forward pass of a neural network, evaluating the function
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/multihead_attention_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/multihead_attention_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/multihead_attention_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/multihead_attention_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -54,19 +54,19 @@ MultiheadAttentionType(
 
 template <typename MatType, typename RegularizerType>
 void MultiheadAttentionType<MatType, RegularizerType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+    const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, (4 * embedDim + 4) * embedDim, 1);
+  MakeAlias(weights, weightsIn, (4 * embedDim + 4) * embedDim, 1);
 
-  MakeAlias(queryWt, weightsPtr, embedDim, embedDim);
-  MakeAlias(keyWt, weightsPtr + embedDim * embedDim, embedDim, embedDim);
-  MakeAlias(valueWt, weightsPtr + 2 * embedDim * embedDim, embedDim, embedDim);
-  MakeAlias(outWt, weightsPtr + 3 * embedDim * embedDim, embedDim, embedDim);
-
-  MakeAlias(qBias, weightsPtr + 4 * embedDim * embedDim, embedDim, 1);
-  MakeAlias(kBias, weightsPtr + (4 * embedDim + 1) * embedDim, embedDim, 1);
-  MakeAlias(vBias, weightsPtr + (4 * embedDim + 2) * embedDim, embedDim, 1);
-  MakeAlias(outBias, weightsPtr + (4 * embedDim + 3) * embedDim, 1, embedDim);
+  MakeAlias(queryWt, weightsIn, embedDim, embedDim);
+  MakeAlias(keyWt, weightsIn, embedDim, embedDim, embedDim * embedDim);
+  MakeAlias(valueWt, weightsIn, embedDim, embedDim, 2 * embedDim * embedDim);
+  MakeAlias(outWt, weightsIn, embedDim, embedDim, 3 * embedDim * embedDim);
+
+  MakeAlias(qBias, weightsIn, embedDim, 1, 4 * embedDim * embedDim);
+  MakeAlias(kBias, weightsIn, embedDim, 1, (4 * embedDim + 1) * embedDim);
+  MakeAlias(vBias, weightsIn, embedDim, 1, (4 * embedDim + 2) * embedDim);
+  MakeAlias(outBias, weightsIn, 1, embedDim, (4 * embedDim + 3) * embedDim);
 }
 
 template <typename MatType, typename RegularizerType>
@@ -114,12 +114,12 @@ Forward(const MatType& input, MatType& o
 
   for (size_t i = 0; i < batchSize; ++i)
   {
-    qProj.slice(i) = arma::trans(
-        queryWt * q.slice(i) + arma::repmat(qBias, 1, tgtSeqLen));
-    kProj.slice(i) = arma::trans(
-        keyWt * k.slice(i) + arma::repmat(kBias, 1, srcSeqLen));
-    vProj.slice(i) = arma::trans(
-        valueWt * v.slice(i) + arma::repmat(vBias, 1, srcSeqLen));
+    qProj.slice(i) = trans(
+        queryWt * q.slice(i) + repmat(qBias, 1, tgtSeqLen));
+    kProj.slice(i) = trans(
+        keyWt * k.slice(i) + repmat(kBias, 1, srcSeqLen));
+    vProj.slice(i) = trans(
+        valueWt * v.slice(i) + repmat(vBias, 1, srcSeqLen));
   }
 
   // The scaling factor sqrt(headDim) is used to prevent exploding values
@@ -155,7 +155,7 @@ Forward(const MatType& input, MatType& o
   {
     if (keyPaddingMask.n_rows != 1 || keyPaddingMask.n_cols != srcSeqLen)
         Log::Fatal << "The size of the 'keyPaddingMask' is not correct.\n";
-    scores.each_slice() += arma::repmat(keyPaddingMask, tgtSeqLen, 1);
+    scores.each_slice() += repmat(keyPaddingMask, tgtSeqLen, 1);
   }
 
   for (size_t i = 0; i < numHeads * batchSize; ++i)
@@ -175,8 +175,8 @@ Forward(const MatType& input, MatType& o
   // The final output is the linear projection of attention output.
   for (size_t i = 0; i < batchSize; ++i)
   {
-    output.col(i) = arma::vectorise(arma::trans(attnOut.slice(i) * outWt
-        + arma::repmat(outBias, tgtSeqLen, 1)));
+    output.col(i) = vectorise(trans(attnOut.slice(i) * outWt
+        + repmat(outBias, tgtSeqLen, 1)));
   }
 }
 
@@ -229,12 +229,12 @@ Backward(const MatType& /* input */,
     if (selfAttention)
     {
       g.submat(0, i, g.n_rows - 1, i) =
-          arma::vectorise(arma::trans(tmp.slice(i) * valueWt));
+          vectorise(trans(tmp.slice(i) * valueWt));
     }
     else
     {
       g.submat((tgtSeqLen + srcSeqLen) * embedDim, i, g.n_rows - 1, i) =
-          arma::vectorise(arma::trans(tmp.slice(i) * valueWt));
+          vectorise(trans(tmp.slice(i) * valueWt));
     }
   }
 
@@ -264,13 +264,13 @@ Backward(const MatType& /* input */,
     {
       // Sum the query, key, and value deltas.
       g.submat(0, i, g.n_rows - 1, i) +=
-          arma::vectorise(arma::trans(tmp.slice(i) * keyWt));
+          vectorise(trans(tmp.slice(i) * keyWt));
     }
     else
     {
       g.submat(tgtSeqLen * embedDim, i,
                (tgtSeqLen + srcSeqLen) * embedDim - 1, i) =
-          arma::vectorise(arma::trans(tmp.slice(i) * keyWt));
+          vectorise(trans(tmp.slice(i) * keyWt));
     }
   }
 
@@ -289,12 +289,12 @@ Backward(const MatType& /* input */,
     {
       // Sum the query, key, and value deltas.
       g.submat(0, i, g.n_rows - 1, i) +=
-          arma::vectorise(arma::trans(tmp.slice(i) * queryWt));
+          vectorise(trans(tmp.slice(i) * queryWt));
     }
     else
     {
       g.submat(0, i, tgtSeqLen * embedDim - 1, i) =
-          arma::vectorise(arma::trans(tmp.slice(i) * queryWt));
+          vectorise(trans(tmp.slice(i) * queryWt));
     }
   }
 }
@@ -347,7 +347,7 @@ Gradient(const MatType& input,
 
   // Gradient wrt. outBias, i.e. dL/d(outBias).
   gradient.rows(4 * wtSize + 3 * embedDim, 4 * wtSize + 4 * embedDim - 1)
-      = arma::vectorise(arma::sum(arma::sum(errorTemp, 2), 1));
+      = vectorise(sum(sum(errorTemp, 2), 1));
 
   // The shape of attnOut : (tgtSeqLen, embedDim, batchSize).
   // The shape of errorTemp : (embedDim, tgtSeqLen, batchSize).
@@ -356,8 +356,7 @@ Gradient(const MatType& input,
 
   // Gradient wrt. outWt, i.e. dL/d(outWt). We will take sum of gyTemp along
   // the slices and vectorise the output.
-  gradient.rows(3 * wtSize, 4 * wtSize - 1)
-      = arma::vectorise(arma::sum(gyTemp, 2));
+  gradient.rows(3 * wtSize, 4 * wtSize - 1) = vectorise(sum(gyTemp, 2));
 
   // Partial derivative wrt. attnOut.
   // The shape of outWt : (embedDim, embedDim).
@@ -381,7 +380,7 @@ Gradient(const MatType& input,
   // Gradient wrt. vBias, i.e. dL/d(vBias). We will take summation of errorTemp
   // over all the batches and over all the sequences.
   gradient.rows(4 * wtSize + 2 * embedDim, 4 * wtSize + 3 * embedDim - 1)
-      = arma::vectorise(arma::sum(arma::sum(errorTemp, 2), 0));
+      = vectorise(sum(sum(errorTemp, 2), 0));
 
   // Shape of v : (srcSeqLen, embedDim, batchSize).
   // Shape of errorTemp : (srcSeqLen, embedDim, bathSize).
@@ -390,8 +389,7 @@ Gradient(const MatType& input,
 
   // Gradient wrt. valueWt, i.e. dL/d(valueWt). We will take summation over all
   // batches of errorTemp.
-  gradient.rows(2 * wtSize, 3 * wtSize - 1)
-      = arma::vectorise(arma::sum(errorTemp, 2));
+  gradient.rows(2 * wtSize, 3 * wtSize - 1) = vectorise(sum(errorTemp, 2));
 
   // Now, the shape of gyTemp : (tgtSeqLen, headDim, numHeads * batchSize).
   // The shape of vProj : (srcSeqLen, headDim, numHeads * batchSize).
@@ -419,7 +417,7 @@ Gradient(const MatType& input,
   // Gradient wrt. kBias, i.e. dL/d(kBias). We will take summation over all the
   // batches of gyTemp and then over all the sequences.
   gradient.rows(4 * wtSize + embedDim, 4 * wtSize + 2 * embedDim - 1)
-      = arma::vectorise(arma::sum(arma::sum(gyTemp, 2), 0));
+      = vectorise(sum(sum(gyTemp, 2), 0));
 
   // The shape of k : (embedDim, srcSeqLen, batchSize).
   // The shape of gyTemp : (srcSeqLen, embedDim, batchSize).
@@ -428,7 +426,7 @@ Gradient(const MatType& input,
 
   // Gradient wrt. keyWt, i.e. dL/d(keyWt). We will take summation over all the
   // batches of dkeyWt.
-  gradient.rows(wtSize, 2 * wtSize - 1) = arma::vectorise(arma::sum(gyTemp, 2));
+  gradient.rows(wtSize, 2 * wtSize - 1) = vectorise(sum(gyTemp, 2));
 
   // The shape of kProj : (srcSeqLen, headDim, numHeads * batchSize).
   // The shape of errorTemp : (tgtSeqLen, srcSeqLen, numHeads * batchSize).
@@ -442,7 +440,7 @@ Gradient(const MatType& input,
   // Gradient wrt. qBias, i.e. dL/d(qBias). We will take summation over all the
   // batches of gyTemp and over all the sequences.
   gradient.rows(4 * wtSize, 4 * wtSize + embedDim - 1)
-      = arma::vectorise(arma::sum(arma::sum(gyTemp, 2), 0));
+      = vectorise(sum(sum(gyTemp, 2), 0));
 
   // The shape of gyTemp : (tgtSeqLen, embedDim, batchSize).
   // The shape of q : (embedDim, tgtSeqLen, batchSize).
@@ -451,7 +449,7 @@ Gradient(const MatType& input,
 
   // Gradient wrt. queryWt, i.e. dL/d(queryBias). We will take summation over
   // all the batches of gyTemp.
-  gradient.rows(0, wtSize - 1) = arma::vectorise(arma::sum(gyTemp, 2));
+  gradient.rows(0, wtSize - 1) = vectorise(sum(gyTemp, 2));
 
   // Regularize according to the given regularization rule.
   regularizer.Evaluate(weights, gradient);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/noisylinear.hpp 4.4.0-1/src/mlpack/methods/ann/layer/noisylinear.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/noisylinear.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/noisylinear.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,7 +52,7 @@ class NoisyLinearType : public Layer<Mat
   NoisyLinearType& operator=(NoisyLinearType&& other);
 
   //! Reset the layer parameter.
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   //! Reset the noise parameters (epsilons).
   void ResetNoise();
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/noisylinear_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/noisylinear_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/noisylinear_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/noisylinear_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,17 +73,15 @@ NoisyLinearType<MatType>::operator=(Nois
 }
 
 template<typename MatType>
-void NoisyLinearType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void NoisyLinearType<MatType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(weights, weightsPtr, 1, (outSize * inSize + outSize) * 2);
-
-  MakeAlias(weightMu, weightsPtr, outSize, inSize);
-  MakeAlias(biasMu, weightsPtr + weightMu.n_elem, outSize, 1);
-  MakeAlias(weightSigma, weightsPtr + weightMu.n_elem + biasMu.n_elem, outSize,
-      inSize);
-  MakeAlias(biasSigma, weightsPtr + weightMu.n_elem * 2 + biasMu.n_elem,
-      outSize, 1);
+  MakeAlias(weights, weightsIn, 1, (outSize * inSize + outSize) * 2);
+  MakeAlias(weightMu, weightsIn, outSize, inSize);
+  MakeAlias(biasMu, weightsIn, outSize, 1, weightMu.n_elem);
+  MakeAlias(weightSigma, weightsIn, outSize, inSize,
+      weightMu.n_elem + biasMu.n_elem);
+  MakeAlias(biasSigma, weightsIn, outSize, 1,
+      weightMu.n_elem * 2 + biasMu.n_elem);
 
   this->ResetNoise();
 }
@@ -91,11 +89,13 @@ void NoisyLinearType<MatType>::SetWeight
 template<typename MatType>
 void NoisyLinearType<MatType>::ResetNoise()
 {
-  MatType epsilonIn = arma::randn<MatType>(inSize, 1);
-  epsilonIn = arma::sign(epsilonIn) % arma::sqrt(arma::abs(epsilonIn));
-
-  MatType epsilonOut = arma::randn<MatType>(outSize, 1);
-  epsilonOut = arma::sign(epsilonOut) % arma::sqrt(arma::abs(epsilonOut));
+  MatType epsilonIn;
+  epsilonIn.randn(inSize, 1);
+  epsilonIn = sign(epsilonIn) % sqrt(arma::abs(epsilonIn));
+
+  MatType epsilonOut;
+  epsilonOut.randn(outSize, 1);
+  epsilonOut = sign(epsilonOut) % sqrt(arma::abs(epsilonOut));
 
   weightEpsilon = epsilonOut * epsilonIn.t();
   biasEpsilon = epsilonOut;
@@ -140,15 +140,14 @@ void NoisyLinearType<MatType>::Gradient(
   MatType weightGrad = error * input.t();
 
   // Gradients for mu values.
-  gradient.rows(0, weight.n_elem - 1) = arma::vectorise(weightGrad);
-  gradient.rows(weight.n_elem, weight.n_elem + bias.n_elem - 1)
-      = arma::sum(error, 1);
+  gradient.rows(0, weight.n_elem - 1) = vectorise(weightGrad);
+  gradient.rows(weight.n_elem, weight.n_elem + bias.n_elem - 1) = sum(error, 1);
 
   // Gradients for sigma values.
   gradient.rows(weight.n_elem + bias.n_elem, gradient.n_elem - bias.n_elem - 1)
-      = arma::vectorise(weightGrad % weightEpsilon);
+      = vectorise(weightGrad % weightEpsilon);
   gradient.rows(gradient.n_elem - bias.n_elem, gradient.n_elem - 1)
-      = arma::sum(error, 1) % biasEpsilon;
+      = sum(error, 1) % biasEpsilon;
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/atrous_convolution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/atrous_convolution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/atrous_convolution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/atrous_convolution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -395,7 +395,7 @@ void AtrousConvolution<
     }
 
     gradient.submat(weight.n_elem + (outMap % outSize), 0, weight.n_elem +
-        (outMap % outSize), 0) = arma::accu(mappedError.slice(outMap));
+        (outMap % outSize), 0) = accu(mappedError.slice(outMap));
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/bicubic_interpolation_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/bicubic_interpolation_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/bicubic_interpolation_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/bicubic_interpolation_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -89,7 +89,7 @@ void BicubicInterpolation<InputDataType,
   for (size_t k = 0; k < depth * batchSize; ++k)
   {
     // The input is padded on all sides using replication of the input boundary.
-    arma::mat grid = arma::zeros(inRowSize + 4, inColSize + 4);
+    arma::mat grid = zeros(inRowSize + 4, inColSize + 4);
     grid(arma::span(2, inRowSize + 1), arma::span(2, inColSize + 1)) = inputAsCube.slice(k);
     grid(arma::span(2, inRowSize + 1), 0) = grid(arma::span(2, inRowSize + 1), 2);
     grid(arma::span(2, inRowSize + 1), 1) = grid(arma::span(2, inRowSize + 1), 2);
@@ -128,7 +128,7 @@ void BicubicInterpolation<InputDataType,
         GetKernalWeight(fc, weightC);
 
         arma::mat weightMatrix = weightR * weightC;
-        outputAsCube(i, j, k) = arma::accu(weightMatrix % kernal);
+        outputAsCube(i, j, k) = accu(weightMatrix % kernal);
       }
     }
   }
@@ -168,7 +168,7 @@ void BicubicInterpolation<InputDataType,
   {
     for (size_t k = 0; k < depth * batchSize; ++k)
     {
-      arma::mat temp = arma::zeros(inRowSize + 4, inColSize + 4);
+      arma::mat temp = zeros(inRowSize + 4, inColSize + 4);
       for (size_t i = 0; i < outRowSize; ++i)
       {
         double rOrigin = (i + 0.5) * scaleRow;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/bilinear_interpolation_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/bilinear_interpolation_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/bilinear_interpolation_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/bilinear_interpolation_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -88,7 +88,7 @@ void BilinearInterpolationType<InputType
 
       for (size_t k = 0; k < depth * batchSize; ++k)
       {
-        outputAsCube(i, j, k) = arma::accu(inputAsCube.slice(k).submat(
+        outputAsCube(i, j, k) = accu(inputAsCube.slice(k).submat(
             rOrigin, cOrigin, rOrigin + 1, cOrigin + 1) % coeffs);
       }
     }
@@ -148,7 +148,7 @@ void BilinearInterpolationType<InputType
 
         for (size_t k = 0; k < depth * batchSize; ++k)
         {
-          outputAsCube(i, j, k) = arma::accu(gradientAsCube.slice(k).submat(
+          outputAsCube(i, j, k) = accu(gradientAsCube.slice(k).submat(
               rOrigin, cOrigin, rOrigin + 1, cOrigin + 1) % coeffs);
         }
       }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/concat_performance_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/concat_performance_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/concat_performance_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/concat_performance_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -78,7 +78,7 @@ void ConcatPerformance<
 
   outputLayer.Backward(subInput, target, subOutput);
 
-  output = arma::zeros(subOutput.n_elem, inputDimensions[0]);
+  output = zeros(subOutput.n_elem, inputDimensions[0]);
   output.col(0) = subOutput;
 
   for (size_t i = elements, j = 0; i < input.n_elem; i+= elements, ++j)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/fast_lstm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/fast_lstm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/fast_lstm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/fast_lstm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -258,7 +258,7 @@ void FastLSTMType<InputType, OutputType>
 
   cellActivationError = gyLocal % gateActivation.submat(outSize,
       backwardStep - batchStep, 2 * outSize - 1, backwardStep) %
-      (1 - arma::pow(cellActivation.cols(backwardStep - batchStep,
+      (1 - pow(cellActivation.cols(backwardStep - batchStep,
       backwardStep), 2));
 
   if (gradientStepIdx > 0)
@@ -291,7 +291,7 @@ void FastLSTMType<InputType, OutputType>
 
   prevError.submat(3 * outSize, 0, 4 * outSize - 1, batchStep) =
       gateActivation.submat(0, backwardStep - batchStep,
-      outSize - 1, backwardStep) % cellActivationError % (1 - arma::pow(
+      outSize - 1, backwardStep) % cellActivationError % (1 - pow(
       stateActivation.cols(backwardStep - batchStep, backwardStep), 2));
 
   prevError.submat(outSize, 0, 2 * outSize - 1, batchStep) =
@@ -320,14 +320,14 @@ void FastLSTMType<InputType, OutputType>
 {
   // Gradient of the input to gate layer.
   gradient.submat(0, 0, input2GateWeight.n_elem - 1, 0) =
-      arma::vectorise(prevError * input.t());
+      vectorise(prevError * input.t());
 
   gradient.submat(input2GateWeight.n_elem, 0, input2GateWeight.n_elem +
-      input2GateBias.n_elem - 1, 0) = arma::sum(prevError, 1);
+      input2GateBias.n_elem - 1, 0) = sum(prevError, 1);
 
   // Gradient of the output to gate layer.
   gradient.submat(input2GateWeight.n_elem + input2GateBias.n_elem, 0,
-      gradient.n_elem - 1, 0) = arma::vectorise(prevError *
+      gradient.n_elem - 1, 0) = vectorise(prevError *
       outParameter.cols(gradientStep - batchStep, gradientStep).t());
 
   if (gradientStep > batchStep)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/glimpse.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/glimpse.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/glimpse.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/glimpse.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,7 +64,7 @@ class MeanPoolingRule
   template<typename MatType>
   void Unpooling(const MatType& input, const double value, MatType& output)
   {
-    output = arma::zeros<MatType>(input.n_rows, input.n_cols);
+    output = zeros<MatType>(input.n_rows, input.n_cols);
     const double mean = arma::mean(arma::mean(input));
 
     output.elem(arma::find(mean == input, 1)).fill(value);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/glimpse_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/glimpse_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/glimpse_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/glimpse_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,7 +67,7 @@ void GlimpseType<InputType, OutputType>:
       size_t padSize = std::floor((glimpseSize - 1) / 2);
 
       arma::Cube<typename InputType::elem_type> inputPadded =
-          arma::zeros<arma::Cube<typename InputType::elem_type>>(
+          zeros<arma::Cube<typename InputType::elem_type>>(
           inputTemp.n_rows + padSize * 2, inputTemp.n_cols + padSize * 2,
           inputTemp.n_slices / inSize);
 
@@ -117,7 +117,7 @@ void GlimpseType<InputType, OutputType>:
 
   for (size_t i = 0; i < outputTemp.n_slices; ++i)
   {
-    outputTemp.slice(i) = arma::trans(outputTemp.slice(i));
+    outputTemp.slice(i) = trans(outputTemp.slice(i));
   }
 
   output = OutputType(outputTemp.memptr(), outputTemp.n_elem, 1);
@@ -132,7 +132,7 @@ void GlimpseType<InputType, OutputType>:
 {
   // Generate a cube using the backpropagated error matrix.
   arma::Cube<typename OutputType::elem_type> mappedError =
-      arma::zeros<arma::Cube<typename OutputType::elem_type>>(outputWidth,
+      zeros<arma::Cube<typename OutputType::elem_type>>(outputWidth,
       outputHeight, 1);
 
   location = locationParameter.back();
@@ -147,7 +147,7 @@ void GlimpseType<InputType, OutputType>:
     }
   }
 
-  gTemp = arma::zeros<arma::Cube<typename InputType::elem_type>>(
+  gTemp = zeros<arma::Cube<typename InputType::elem_type>>(
       inputTemp.n_rows, inputTemp.n_cols, inputTemp.n_slices);
 
   for (size_t inputIdx = 0; inputIdx < inSize; inputIdx++)
@@ -158,7 +158,7 @@ void GlimpseType<InputType, OutputType>:
       size_t padSize = std::floor((glimpseSize - 1) / 2);
 
       arma::Cube<typename InputType::elem_type> inputPadded =
-          arma::zeros<arma::Cube<typename InputType::elem_type>>(
+          zeros<arma::Cube<typename InputType::elem_type>>(
           inputTemp.n_rows + padSize * 2, inputTemp.n_cols +
           padSize * 2, inputTemp.n_slices / inSize);
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/group_norm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/group_norm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/group_norm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/group_norm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -80,7 +80,7 @@ void GroupNorm<InputDataType, OutputData
   // Normalize the input.
   output = reshapedInput.each_row() - mean;
   inputMean = output;
-  output.each_row() /= arma::sqrt(variance + eps);
+  output.each_row() /= sqrt(variance + eps);
 
   output.reshape(input.n_rows, input.n_cols);
   // Reused in the backward and gradient step.
@@ -125,11 +125,11 @@ void GroupNorm<InputDataType, OutputData
   arma::mat normReshaped(const_cast<arma::Mat<eT>&>(norm).memptr(),
       gy.n_rows / groupCount, gy.n_cols * groupCount, false, false);
 
-  const arma::mat stdInv = 1.0 / arma::sqrt(variance + eps);
+  const arma::mat stdInv = 1.0 / sqrt(variance + eps);
 
   // sum dl / dxhat * (x - mu) * -0.5 * stdInv^3.
-  const arma::mat var = arma::sum(normReshaped % inputMean, 0) %
-      arma::pow(stdInv, 3.0) * -0.5;
+  const arma::mat var = sum(normReshaped % inputMean, 0) %
+      pow(stdInv, 3.0) * -0.5;
 
   // dl / dxhat * 1 / stdInv + variance * 2 * (x - mu) / m +
   // dl / dmu * 1 / m.
@@ -138,7 +138,7 @@ void GroupNorm<InputDataType, OutputData
 
   // sum (dl / dxhat * -1 / stdInv) + variance *
   // (sum -2 * (x - mu)) / m.
-  g.each_row() += arma::sum(normReshaped.each_row() % -stdInv, 0) /
+  g.each_row() += sum(normReshaped.each_row() % -stdInv, 0) /
       inputReshaped.n_rows;
   g.reshape(input.n_rows, input.n_cols);
 }
@@ -152,22 +152,22 @@ void GroupNorm<InputDataType, OutputData
 {
   assert(error.n_rows % size == 0);
   const size_t channelSize = error.n_rows / size;
-  temp = arma::sum(normalized % error, 1);
+  temp = sum(normalized % error, 1);
   arma::mat tempReshaped((temp).memptr(),
       channelSize, temp.n_elem / channelSize, false, false);
 
   gradient.set_size(size + size, 1);
 
   // Step 5: dl / dy * xhat.
-  gradient.submat(0, 0, gamma.n_elem - 1, 0) = arma::sum(tempReshaped, 0).t();
+  gradient.submat(0, 0, gamma.n_elem - 1, 0) = sum(tempReshaped, 0).t();
 
-  temp = arma::sum(error, 1);
+  temp = sum(error, 1);
   arma::mat tempErrorReshaped((temp).memptr(),
       channelSize, temp.n_elem / channelSize, false, false);
 
   // Step 6: dl / dy.
   gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) =
-      arma::sum(tempErrorReshaped, 0).t();
+      sum(tempErrorReshaped, 0).t();
 }
 
 template<typename InputDataType, typename OutputDataType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/gru_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/gru_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/gru_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/gru_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,9 +60,9 @@ GRU<InputType, OutputType>::GRU(
   network.push_back(hiddenStateModule);
   network.push_back(forgetGateModule);
 
-  prevError = arma::zeros<OutputType>(3 * outSize, batchSize);
+  prevError = zeros<OutputType>(3 * outSize, batchSize);
 
-  allZeros = arma::zeros<OutputType>(outSize, batchSize);
+  allZeros = zeros<OutputType>(outSize, batchSize);
 
   outParameter.emplace_back(allZeros.memptr(),
       allZeros.n_rows, allZeros.n_cols, false, true);
@@ -220,7 +220,7 @@ void GRU<InputType, OutputType>::Backwar
       hiddenStateModule->OutputParameter());
 
   // Delta ot.
-  OutputType dOt = gyLocal % (arma::ones<OutputType>(outSize, batchSize) -
+  OutputType dOt = gyLocal % (ones<OutputType>(outSize, batchSize) -
       inputGateModule->OutputParameter());
 
   // Delta of input gate.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/hardshrink_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/hardshrink_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/hardshrink_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/hardshrink_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,7 +37,7 @@ template<typename InputType, typename Ou
 void HardShrinkType<InputType, OutputType>::Backward(
     const InputType& input, const OutputType& gy, OutputType& g)
 {
-  g = gy % (arma::ones<OutputType>(arma::size(input)) - (input == 0));
+  g = gy % (ones<OutputType>(arma::size(input)) - (input == 0));
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/highway_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/highway_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/highway_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/highway_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -110,7 +110,7 @@ void HighwayType<InputType, OutputType>:
 
   transformGate = transformWeight * input;
   transformGate.each_col() += transformBias;
-  transformGateActivation = 1.0 /(1 + arma::exp(-transformGate));
+  transformGateActivation = 1.0 /(1 + exp(-transformGate));
   output = (this->layerOutputs.back() % transformGateActivation) +
       (input % (1 - transformGateActivation));
 }
@@ -173,10 +173,10 @@ void HighwayType<InputType, OutputType>:
       this->layerDeltas[1],
       this->layerGradients.front());
 
-  gradient.submat(0, 0, transformWeight.n_elem - 1, 0) = arma::vectorise(
+  gradient.submat(0, 0, transformWeight.n_elem - 1, 0) = vectorise(
       transformGateError * input.t());
   gradient.submat(transformWeight.n_elem, 0, transformWeight.n_elem +
-      transformBias.n_elem - 1, 0) = arma::sum(transformGateError, 1);
+      transformBias.n_elem - 1, 0) = sum(transformGateError, 1);
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/isrlu_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/isrlu_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/isrlu_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/isrlu_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ template<typename InputType, typename Ou
 void ISRLU<InputDataType, OutputDataType>::Forward(
     const InputType& input, OutputType& output)
 {
-  output = arma::ones<OutputDataType>(arma::size(input));
+  output = ones<OutputDataType>(arma::size(input));
   for (size_t i = 0; i < input.n_elem; ++i)
   {
     output(i) = (input(i) >= 0) ? input(i) : input(i) *
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/join_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/join_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/join_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/join_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,7 +31,7 @@ void JoinType<InputType, OutputType>::Fo
 {
   inSizeRows = input.n_rows;
   inSizeCols = input.n_cols;
-  output = arma::vectorise(input);
+  output = vectorise(input);
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lookup_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lookup_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lookup_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lookup_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,8 +47,8 @@ void LookupType<InputType, OutputType>::
     // ith column of output is a vectorized form of a matrix of shape
     // (embeddingSize, seqLength) selected as a combination of columns from the
     // weights.
-    output.col(i) = arma::vectorise(weights.cols(
-        arma::conv_to<arma::uvec>::from(input.col(i)) - 1));
+    output.col(i) = vectorise(weights.cols(
+        ConvTo<arma::uvec>::From(input.col(i)) - 1));
   }
 }
 
@@ -77,7 +77,7 @@ void LookupType<InputType, OutputType>::
   gradient.zeros();
   for (size_t i = 0; i < batchSize; ++i)
   {
-    gradient.cols(arma::conv_to<arma::uvec>::from(input.col(i)) - 1)
+    gradient.cols(ConvTo<arma::uvec>::From(input.col(i)) - 1)
         += errorTemp.slice(i);
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lp_pooling.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lp_pooling.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lp_pooling.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lp_pooling.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -161,7 +161,7 @@ class LpPooling
   void Pooling(const arma::Mat<eT>& input, arma::Mat<eT>& output)
   {
     arma::Mat<eT> inputPre = input;
-    inputPre = arma::pow(inputPre, normType);
+    inputPre = pow(inputPre, normType);
 
     for (size_t i = 1; i < input.n_cols; ++i)
       inputPre.col(i) += inputPre.col(i - 1);
@@ -199,7 +199,7 @@ class LpPooling
       }
     }
 
-    output = arma::pow(output, 1.0 / normType);
+    output = pow(output, 1.0 / normType);
   }
 
   /**
@@ -240,11 +240,11 @@ class LpPooling
         arma::mat InputArea = input(arma::span(i, rowEnd),
             arma::span(j, colEnd));
 
-        size_t sum = pow(arma::accu(arma::pow(InputArea, normType)),
+        size_t sum = pow(accu(pow(InputArea, normType)),
             (normType - 1) / normType);
         unpooledError = arma::Mat<eT>(InputArea.n_rows, InputArea.n_cols);
         unpooledError.fill(error(rowidx, colidx) / InputArea.n_elem);
-        unpooledError %= arma::pow(InputArea, normType - 1);
+        unpooledError %= pow(InputArea, normType - 1);
         unpooledError /= sum;
         output(arma::span(i, i + InputArea.n_rows - 1),
             arma::span(j, j + InputArea.n_cols - 1)) += unpooledError;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lp_pooling_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lp_pooling_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/lp_pooling_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/lp_pooling_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,7 +75,7 @@ void LpPooling<InputDataType, OutputData
         (double) kernelHeight) / (double) strideHeight + 1);
   }
 
-  outputTemp = arma::zeros<arma::Cube<eT> >(outputWidth, outputHeight,
+  outputTemp = zeros<arma::Cube<eT> >(outputWidth, outputHeight,
       batchSize * inSize);
 
   for (size_t s = 0; s < inputTemp.n_slices; s++)
@@ -99,7 +99,7 @@ void LpPooling<InputDataType, OutputData
   arma::cube mappedError = arma::cube(((arma::Mat<eT>&) gy).memptr(),
       outputWidth, outputHeight, outSize, false, false);
 
-  gTemp = arma::zeros<arma::cube>(inputTemp.n_rows,
+  gTemp = zeros<arma::cube>(inputTemp.n_rows,
       inputTemp.n_cols, inputTemp.n_slices);
 
   for (size_t s = 0; s < mappedError.n_slices; s++)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/minibatch_discrimination_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/minibatch_discrimination_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/minibatch_discrimination_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/minibatch_discrimination_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,7 +75,7 @@ void MiniBatchDiscrimination<InputType,
       else
       {
         distances.slice(i).col(j) =
-            arma::exp(-arma::sum(abs(cubeM.slice(i) - cubeM.slice(j)), 1));
+            exp(-sum(abs(cubeM.slice(i) - cubeM.slice(j)), 1));
         output.col(i) += distances.slice(i).col(j);
       }
     }
@@ -98,7 +98,7 @@ void MiniBatchDiscrimination<InputType,
       {
         continue;
       }
-      InputType t = arma::sign(M.slice(i) - M.slice(j));
+      InputType t = sign(M.slice(i) - M.slice(j));
       t.each_col() %=
           distances.slice(std::min(i, j)).col(std::max(i, j)) % gM.col(i);
       deltaM.slice(i) -= t;
@@ -116,7 +116,7 @@ void MiniBatchDiscrimination<InputType,
     const OutputType& /* error */,
     OutputType& gradient)
 {
-  gradient = arma::vectorise(deltaTemp * input.t());
+  gradient = vectorise(deltaTemp * input.t());
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/positional_encoding_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/positional_encoding_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/positional_encoding_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/positional_encoding_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,7 +41,7 @@ void PositionalEncodingType<InputType, O
 {
   positionalEncoding.set_size(maxSequenceLength, embedDim);
   const InputType position = arma::regspace(0, 1, maxSequenceLength - 1);
-  const InputType divTerm = arma::exp(arma::regspace(0, 2, embedDim - 1)
+  const InputType divTerm = exp(arma::regspace(0, 2, embedDim - 1)
       * (- std::log(10000.0) / embedDim));
   const InputType theta = position * divTerm.t();
   for (size_t i = 0; i < theta.n_cols; ++i)
@@ -49,7 +49,7 @@ void PositionalEncodingType<InputType, O
     positionalEncoding.col(2 * i) = arma::sin(theta.col(i));
     positionalEncoding.col(2 * i + 1) = arma::cos(theta.col(i));
   }
-  positionalEncoding = arma::vectorise(positionalEncoding.t());
+  positionalEncoding = vectorise(positionalEncoding.t());
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/recurrent_attention_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/recurrent_attention_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/recurrent_attention_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/recurrent_attention_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,7 +60,7 @@ void RecurrentAttention<InputType, Outpu
   // Initialize the action input.
   if (initialInput.is_empty())
   {
-    initialInput = arma::zeros(outSize, input.n_cols);
+    initialInput = zeros(outSize, input.n_cols);
   }
 
   // Propagate through the action and recurrent module.
@@ -76,7 +76,7 @@ void RecurrentAttention<InputType, Outpu
     }
 
     // Initialize the glimpse input.
-    InputType glimpseInput = arma::zeros(input.n_elem, 2);
+    InputType glimpseInput = zeros(input.n_elem, 2);
     glimpseInput.col(0) = input;
     glimpseInput.submat(0, 1, actionOutput.n_elem - 1, 1) =
         actionOutput;
@@ -123,11 +123,11 @@ void RecurrentAttention<InputType, Outpu
     size_t weights = rnnModule->Parameters().n_elem +
         actionModule->Parameters().n_elem;
 
-    intermediateGradient = arma::zeros(weights, 1);
-    attentionGradient = arma::zeros(weights, 1);
+    intermediateGradient = zeros(weights, 1);
+    attentionGradient = zeros(weights, 1);
 
     // Initialize the action error.
-    actionError = arma::zeros(actionOutput.n_rows, actionOutput.n_cols);
+    actionError = zeros(actionOutput.n_rows, actionOutput.n_cols);
   }
 
   // Propagate the attention gradients.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/reinforce_normal_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/reinforce_normal_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/reinforce_normal_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/reinforce_normal_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,9 +32,7 @@ void ReinforceNormalType<InputType, Outp
   if (!deterministic)
   {
     // Multiply by standard deviations and re-center the means to the mean.
-    output = arma::randn<OutputType>(input.n_rows, input.n_cols) *
-        stdev + input;
-
+    output.randn(input.n_rows, input.n_cols) * stdev + input;
     moduleInputParameter.push_back(input);
   }
   else
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/reparametrization_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/reparametrization_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/reparametrization_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/reparametrization_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -97,9 +97,9 @@ void ReparametrizationType<InputType, Ou
   preStdDev = input.submat(0, 0, latentSize - 1, input.n_cols - 1);
 
   if (stochastic)
-    gaussianSample = arma::randn<OutputType>(latentSize, input.n_cols);
+    gaussianSample.randn(latentSize, input.n_cols);
   else
-    gaussianSample = arma::ones<OutputType>(latentSize, input.n_cols) * 0.7;
+    gaussianSample.ones(latentSize, input.n_cols) * 0.7;
 
   SoftplusFunction::Fn(preStdDev, stdDev);
   output = mean + stdDev % gaussianSample;
@@ -129,8 +129,8 @@ double ReparametrizationType<InputType,
   if (!includeKl)
     return 0;
 
-  return -0.5 * beta * arma::accu(2 * arma::log(stdDev) - arma::pow(stdDev, 2)
-      - arma::pow(mean, 2) + 1) / mean.n_cols;
+  return -0.5 * beta * accu(2 * log(stdDev) - pow(stdDev, 2)
+      - pow(mean, 2) + 1) / mean.n_cols;
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/softshrink_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/softshrink_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/softshrink_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/softshrink_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,7 +38,7 @@ template<typename InputType, typename Ou
 void SoftShrinkType<InputType, OutputType>::Backward(
     const InputType& input, const OutputType& gy, OutputType& g)
 {
-  g = gy % (arma::ones(arma::size(input)) - (input == 0));
+  g = gy % (ones(arma::size(input)) - (input == 0));
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/spatial_dropout_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/spatial_dropout_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/spatial_dropout_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/spatial_dropout_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,7 +75,7 @@ void SpatialDropoutType<InputType, Outpu
     probabilities.fill(ratio);
     BernoulliDistribution<> bernoulli_dist(probabilities, false);
     maskRow = bernoulli_dist.Sample();
-    mask = arma::repmat(maskRow, inputSize, 1);
+    mask = repmat(maskRow, inputSize, 1);
 
     for (size_t n = 0; n < batchSize; n++)
       outputTemp.slice(n) = inputTemp.slice(n) % mask * scale;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/subview.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/subview.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/subview.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/subview.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -89,7 +89,7 @@ class SubviewType : public Layer<InputTy
     {
       for (size_t i = 0; i < batchSize; ++i)
       {
-        output.col(i) = arma::vectorise(
+        output.col(i) = vectorise(
             input.submat(beginRow, batchBegin, endRow, batchEnd));
 
         // Move to next batch.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -332,7 +332,7 @@ class TransposedConvolutionType : public
     if (output.n_rows != input.n_rows * strideWidth - strideWidth + 1 ||
         output.n_cols != input.n_cols * strideHeight - strideHeight + 1)
     {
-      output = arma::zeros(input.n_rows * strideWidth - strideWidth + 1,
+      output = zeros(input.n_rows * strideWidth - strideWidth + 1,
           input.n_cols * strideHeight - strideHeight + 1);
     }
 
@@ -362,7 +362,7 @@ class TransposedConvolutionType : public
                    const size_t strideHeight,
                    arma::Cube<eT>& output)
   {
-    output = arma::zeros(input.n_rows * strideWidth - strideWidth + 1,
+    output = zeros(input.n_rows * strideWidth - strideWidth + 1,
         input.n_cols * strideHeight - strideHeight + 1, input.n_slices);
 
     for (size_t i = 0; i < input.n_slices; ++i)
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/transposed_convolution_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -429,7 +429,7 @@ void TransposedConvolutionType<
     }
 
     gradient.submat(weight.n_elem + (outMap % outSize), 0, weight.n_elem +
-        (outMap % outSize), 0) = arma::accu(mappedError.slices(outMap, outMap));
+        (outMap % outSize), 0) = accu(mappedError.slices(outMap, outMap));
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/virtual_batch_norm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/virtual_batch_norm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/virtual_batch_norm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/virtual_batch_norm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,7 +38,7 @@ VirtualBatchNormType<InputType, OutputTy
     loading(false)
 {
   referenceBatchMean = arma::mean(referenceBatch, 1);
-  referenceBatchMeanSquared = arma::mean(arma::square(referenceBatch), 1);
+  referenceBatchMeanSquared = arma::mean(square(referenceBatch), 1);
   newCoefficient = 1.0 / (referenceBatch.n_cols + 1);
   oldCoefficient = 1 - newCoefficient;
 }
@@ -68,16 +68,16 @@ void VirtualBatchNormType<InputType, Out
 
   inputParameter = input;
   InputType inputMean = arma::mean(input, 1);
-  InputType inputMeanSquared = arma::mean(arma::square(input), 1);
+  InputType inputMeanSquared = arma::mean(square(input), 1);
 
   mean = oldCoefficient * referenceBatchMean + newCoefficient * inputMean;
   OutputType meanSquared = oldCoefficient * referenceBatchMeanSquared +
       newCoefficient * inputMeanSquared;
-  variance = meanSquared - arma::square(mean);
+  variance = meanSquared - square(mean);
   // Normalize the input.
   output = input.each_col() - mean;
   inputSubMean = output;
-  output.each_col() /= arma::sqrt(variance + eps);
+  output.each_col() /= sqrt(variance + eps);
 
   // Reused in the backward and gradient step.
   normalized = output;
@@ -92,14 +92,13 @@ void VirtualBatchNormType<InputType, Out
     const OutputType& gy,
     OutputType& g)
 {
-  const OutputType stdInv = 1.0 / arma::sqrt(variance + eps);
+  const OutputType stdInv = 1.0 / sqrt(variance + eps);
 
   // dl / dxhat.
   const OutputType norm = gy.each_col() % gamma;
 
   // sum dl / dxhat * (x - mu) * -0.5 * stdInv^3.
-  const OutputType var = arma::sum(norm % inputSubMean, 1) %
-      arma::pow(stdInv, 3.0) * -0.5;
+  const OutputType var = sum(norm % inputSubMean, 1) % pow(stdInv, 3.0) * -0.5;
 
   // dl / dxhat * 1 / stdInv + variance * 2 * (x - mu) / m +
   // dl / dmu * newCoefficient / m.
@@ -108,7 +107,7 @@ void VirtualBatchNormType<InputType, Out
 
   // (sum (dl / dxhat * -1 / stdInv) + (variance * mean * -2)) *
   // newCoefficient / m.
-  g.each_col() += (arma::sum(norm.each_col() % -stdInv, 1) + (var %
+  g.each_col() += (sum(norm.each_col() % -stdInv, 1) + (var %
       mean * -2)) * newCoefficient / inputParameter.n_cols;
 }
 
@@ -121,11 +120,10 @@ void VirtualBatchNormType<InputType, Out
   gradient.set_size(size + size, 1);
 
   // Step 5: dl / dy * xhat.
-  gradient.submat(0, 0, gamma.n_elem - 1, 0) = arma::sum(normalized % error, 1);
+  gradient.submat(0, 0, gamma.n_elem - 1, 0) = sum(normalized % error, 1);
 
   // Step 6: dl / dy.
-  gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) =
-      arma::sum(error, 1);
+  gradient.submat(gamma.n_elem, 0, gradient.n_elem - 1, 0) = sum(error, 1);
 }
 
 template<typename InputType, typename OutputType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/weight_norm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/weight_norm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/not_adapted/weight_norm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/not_adapted/weight_norm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -136,7 +136,7 @@ void WeightNormType<InputType, OutputTyp
     const InputType& input, OutputType& output)
 {
   // Initialize the non-bias weights of wrapped layer.
-  const double normVectorParameter = arma::norm(vectorParameter, 2);
+  const double normVectorParameter = norm(vectorParameter, 2);
   layerWeights.rows(0, layerWeightSize - biasWeightSize - 1) =
       scalarParameter(0) * vectorParameter / normVectorParameter;
 
@@ -163,7 +163,7 @@ void WeightNormType<InputType, OutputTyp
   wrappedLayer->Gradient(input, error, gradient);
 
   // Store the norm of vector parameter temporarily.
-  const double normVectorParameter = arma::norm(vectorParameter, 2);
+  const double normVectorParameter = norm(vectorParameter, 2);
 
   // Set the gradients of the bias terms.
   if (biasWeightSize != 0)
@@ -173,7 +173,7 @@ void WeightNormType<InputType, OutputTyp
   }
 
   // Calculate the gradients of the scalar parameter.
-  gradient[gradient.n_rows - 1] = arma::accu(layerGradients.rows(0,
+  gradient[gradient.n_rows - 1] = accu(layerGradients.rows(0,
       layerWeightSize - biasWeightSize - 1) % vectorParameter) /
       normVectorParameter;
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/parametric_relu.hpp 4.4.0-1/src/mlpack/methods/ann/layer/parametric_relu.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/parametric_relu.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/parametric_relu.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -68,7 +68,7 @@ class PReLUType : public Layer<MatType>
   PReLUType& operator=(PReLUType&& other);
 
   //! Reset the layer parameter.
-  void SetWeights(typename MatType::elem_type* weightsPtr);
+  void SetWeights(const MatType& weightsIn);
 
   /**
    * Initialize the weight matrix of the layer.
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/parametric_relu_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/parametric_relu_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/parametric_relu_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/parametric_relu_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,10 +73,9 @@ PReLUType<MatType>::operator=(PReLUType&
 }
 
 template<typename MatType>
-void PReLUType<MatType>::SetWeights(
-    typename MatType::elem_type* weightsPtr)
+void PReLUType<MatType>::SetWeights(const MatType& weightsIn)
 {
-  MakeAlias(alpha, weightsPtr, 1, 1);
+  MakeAlias(alpha, weightsIn, 1, 1);
 }
 
 template<typename MatType>
@@ -125,9 +124,9 @@ void PReLUType<MatType>::Gradient(
     const MatType& error,
     MatType& gradient)
 {
-  MatType zeros = arma::zeros<MatType>(input.n_rows, input.n_cols);
+  MatType zerosMat = zeros<MatType>(input.n_rows, input.n_cols);
   gradient.set_size(1, 1);
-  gradient(0) = arma::accu(error % arma::min(zeros, input)) / input.n_cols;
+  gradient(0) = accu(error % min(zerosMat, input)) / input.n_cols;
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/radial_basis_function_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/radial_basis_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/radial_basis_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/radial_basis_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,8 +41,7 @@ RBFType<MatType, Activation>::RBFType(
     {
       double maxDis = 0;
       MatType temp = centres.each_col() - centres.col(i);
-      maxDis = arma::accu(arma::max(arma::pow(arma::sum(
-          arma::pow((temp), 2), 0), 0.5).t()));
+      maxDis = max(pow(sum(pow((temp), 2), 0), 0.5).t());
       if (maxDis > sigmas)
         sigmas = maxDis;
     }
@@ -120,8 +119,7 @@ void RBFType<MatType, Activation>::Forwa
   for (size_t i = 0; i < input.n_cols; i++)
   {
     MatType temp = centres.each_col() - input.col(i);
-    distances.col(i) = arma::pow(arma::sum(
-        arma::pow((temp), 2), 0), 0.5).t();
+    distances.col(i) = pow(sum(pow((temp), 2), 0), 0.5).t();
   }
   Activation::Fn(distances * std::pow(betas, 0.5), output);
 }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/repeat.hpp 4.4.0-1/src/mlpack/methods/ann/layer/repeat.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/repeat.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/repeat.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,157 @@
+/**
+ * @file methods/ann/layer/repeat.hpp
+ * @author Adam Kropp
+ *
+ * Definition of the Repeat class, which repeats the input n times
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_METHODS_ANN_LAYER_REPEAT_HPP
+#define MLPACK_METHODS_ANN_LAYER_REPEAT_HPP
+
+#include <mlpack/prereqs.hpp>
+
+#include "layer.hpp"
+
+namespace mlpack {
+
+/**
+ * Implementation of the Repeat class. The Repeat class repeats the
+ * input a specified number of times along a each dimension.  The output
+ * will have the same number of dimensions as the input, with each dimension
+ * multiplied by a specified multiple.  The input can be repeated in an
+ * interleaved or block fashion, depending on the parameters given to the
+ * constructor.
+ *
+ * @tparam MatType Matrix representation to accept as input and use for
+ *    computation.
+ */
+template <typename MatType = arma::mat>
+class RepeatType : public Layer<MatType>
+{
+ public:
+  //! Get Specific Col type, not only arma
+  typedef typename GetUColType<MatType>::type UintCol;
+  typedef typename GetUDenseMatType<MatType>::type UintMat;
+  /**
+   * Create the Repeat object.  Multiples will be empty (e.g. 1s for all
+   * dimensions), so this is the equivalent of an Identity Layer.
+   * Interleave will be false (e.g. repeat in blocks).
+   */
+  RepeatType();
+
+  /**
+   * Create the Repeat object, specifying the number of times to repeat
+   * along each dimension, as well as whether to interleave the output or
+   * repeat in blocks.
+   *
+   * @param multiples The number of times to repeat along each axis. Must be
+   *        the same size or smaller than InputDimensions.
+   * @apram interleave If true, the output will be interleaved (similar to
+   *        arma::repelem).  If false, the output will be repeated in blocks.
+   */
+  RepeatType(std::vector<size_t> multiples, bool interleave = false);
+
+  /**
+   * Destroy the layers held by the model.
+   */
+  virtual ~RepeatType() { }
+
+  //! Clone the RepeatType object. This handles polymorphism correctly.
+  RepeatType* Clone() const override { return new RepeatType(*this); }
+
+  //! Copy the given RepeatType layer.
+  RepeatType(const RepeatType& other);
+  //! Take ownership of the given RepeatType layer.
+  RepeatType(RepeatType&& other) noexcept;
+  //! Copy the given RepeatType layer.
+  RepeatType& operator=(const RepeatType& other);
+  //! Take ownership of the given RepeatType layer.
+  RepeatType& operator=(RepeatType&& other) noexcept;
+
+  /**
+   * Ordinary feed forward pass of a neural network, evaluating the function
+   * f(x) by propagating the activity forward through f.
+   *
+   * @param input Input data used for evaluating the specified function.
+   * @param output Resulting output activation.
+   */
+  void Forward(const MatType& input, MatType& output) override;
+
+  /**
+   * Ordinary feed backward pass of a neural network, using 3rd-order tensors as
+   * input, calculating the function f(x) by propagating x backwards through f.
+   * Using the results from the feed forward pass.
+   *
+   * @param * (input) The input data (x) given to the forward pass.
+   * @param * (output) The propagated data (f(x)) resulting from Forward()
+   * @param gy The backpropagated error.
+   * @param g The calculated gradient.
+   */
+  void Backward(const MatType& /* input */,
+                const MatType& /* output */,
+                const MatType& gy,
+                MatType& g) override;
+
+  //! Get the repeat multiples
+  const std::vector<size_t>& Multiples() const { return multiples; }
+
+  //! Get the repeat multiples for modification
+  std::vector<size_t>& Multiples()
+  {
+    this->validOutputDimensions = false;
+    return multiples;
+  }
+
+  //! Get the interleave parameter
+  bool Interleave() const { return interleave; }
+
+  //! Get the interleave parameter for modification
+  bool& Interleave() { return interleave; }
+
+  /**
+   * @brief Computes the output dimensions of the Repeat layer.
+   *
+   * The ComputeOutputDimensions function computes the output dimensions of the
+   * Repeat layer based on the input dimensions and the repeat multiples. The
+   * output dimensions will have the same number of dimensions as the input,
+   * with all dimensions other than the one specified in the axis being the same
+   * size as the input.
+   */
+  void ComputeOutputDimensions() override;
+
+  /**
+   * Serialize the layer.
+   */
+  template<typename Archive>
+  void serialize(Archive& ar, const uint32_t /* version */);
+
+ private:
+  //! Parameter to indicate number of times to repeat along each dimension
+  std::vector<size_t> multiples;
+
+  //! Parameter to indicate whether to interleave the output
+  bool interleave;
+
+  // Cache the target indices for a single tensor for use
+  // in the forward pass.
+  UintCol outIdxs;
+
+  // Cache the contributions of each output element to the
+  // input elements for use in the backward pass.
+  size_t sizeMult;
+  UintMat backIdxs;
+}; // class RepeatType.
+
+// Standard Repeat layer.
+typedef RepeatType<arma::mat> Repeat;
+
+} // namespace mlpack
+
+// Include implementation.
+#include "repeat_impl.hpp"
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/repeat_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/repeat_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/repeat_impl.hpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/repeat_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,204 @@
+/**
+ * @file methods/ann/layer/repeat_impl.hpp
+ * @author Adam Kropp
+ *
+ * Implementation of the Repeat class, which repeats the input n times
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#ifndef MLPACK_METHODS_ANN_LAYER_REPEAT_IMPL_HPP
+#define MLPACK_METHODS_ANN_LAYER_REPEAT_IMPL_HPP
+
+#include "repeat.hpp"
+
+namespace mlpack {
+
+template<typename MatType>
+RepeatType<MatType>::RepeatType(
+    std::vector<size_t> multiples, bool interleave) :
+    Layer<MatType>(),
+    multiples(std::move(multiples)),
+    interleave(interleave)
+{
+  // Nothing to do.
+}
+
+template<typename MatType>
+RepeatType<MatType>::RepeatType() :
+    Layer<MatType>(),
+    interleave(false)
+{
+  // Nothing to do.
+}
+
+template<typename MatType>
+RepeatType<MatType>::RepeatType(const RepeatType& other) :
+    Layer<MatType>(other),
+    multiples(other.multiples),
+    interleave(other.interleave),
+    outIdxs(other.outIdxs),
+    sizeMult(other.sizeMult),
+    backIdxs(other.backIdxs)
+{
+  // Nothing else to do.
+}
+
+template<typename MatType>
+RepeatType<MatType>::RepeatType(RepeatType&& other)  noexcept :
+    Layer<MatType>(std::move(other)),
+    multiples(other.multiples),
+    interleave(other.interleave),
+    outIdxs(other.outIdxs),
+    sizeMult(other.sizeMult),
+    backIdxs(other.backIdxs)
+{
+  // Nothing else to do.
+}
+
+template<typename MatType>
+RepeatType<MatType>& RepeatType<MatType>::operator=(const RepeatType& other)
+{
+  if (this != &other)
+  {
+    Layer<MatType>::operator=(other);
+    multiples = other.multiples;
+    interleave = other.interleave;
+    outIdxs = other.outIdxs;
+    sizeMult = other.sizeMult;
+    backIdxs = other.backIdxs;
+  }
+
+  return *this;
+}
+
+template<typename MatType>
+RepeatType<MatType>& RepeatType<MatType>::operator=(RepeatType&& other) noexcept
+{
+  if (this != &other)
+  {
+    Layer<MatType>::operator=(std::move(other));
+    multiples = std::move(other.multiples);
+    interleave = std::move(other.interleave);
+    outIdxs = std::move(other.outIdxs);
+    sizeMult = other.sizeMult;
+    backIdxs = other.backIdxs;
+  }
+
+  return *this;
+}
+
+template<typename MatType>
+void RepeatType<MatType>::ComputeOutputDimensions()
+{
+  if (multiples.size() > this->inputDimensions.size())
+  {
+    std::ostringstream oss;
+    oss << "Repeat::ComputeOutputDimensions(): multiples vector must "
+        << "have the same or fewer dimensions than InputDimensions";
+    throw std::invalid_argument(oss.str());
+  }
+
+  size_t inputSize = this->inputDimensions[0];
+  for (size_t i = 1; i < this->inputDimensions.size(); i++)
+  {
+    inputSize *= this->inputDimensions[i];
+  }
+  UintCol idxs_col = arma::regspace<UintCol>(0, inputSize - 1);
+  UintMat idxs = ConvTo<UintMat>::From(idxs_col);
+  // Here, we are going to pre-compute the source index for each output
+  // for a single tensor.  Since the tensors are flattened into 1-d
+  // vectors, we can fill the output row-wise based on these
+  // indices.
+  this->outputDimensions = this->inputDimensions;
+  sizeMult = 1;
+  size_t outSize = 1;
+
+  // iteratively reshape the index matrix such that the dimension
+  // to be repeated (and all prior) are flattened to a column, and
+  // then repelem rowwise or repmat columwise.
+  for (size_t i = 0; i < multiples.size(); i++)
+  {
+    if (multiples[i] != 1)
+    {
+      if (interleave)
+      {
+        // For the first dimension, we need to do the repelem columnwise.
+        if (i == 0)
+        {
+          idxs.reshape(outSize * this->inputDimensions[i],
+                       idxs.n_elem / (outSize * this->inputDimensions[i]));
+          idxs = repelem(idxs, multiples[i], 1);
+        }
+        else
+        {
+          idxs.reshape(outSize,
+                       idxs.n_elem / outSize);
+          idxs = repelem(idxs, 1, multiples[i]);
+        }
+      }
+      else
+      {
+        idxs.reshape(outSize * this->inputDimensions[i],
+                     idxs.n_elem / (outSize * this->inputDimensions[i]));
+        idxs = repmat(idxs, multiples[i], 1);
+      }
+      this->outputDimensions[i] *= multiples[i];
+      sizeMult *= multiples[i];
+    }
+    outSize *= this->outputDimensions[i];
+  }
+  outIdxs = idxs.as_col();
+
+  // Now, we are going to pre-compute the contribution of each output
+  // element to the input elements.  This will be used in the backward
+  // pass with a simple matrix multiplication.
+  backIdxs.set_size(inputSize, sizeMult);
+  UintCol counts(inputSize, arma::fill::zeros);
+  for (size_t i = 0; i < outIdxs.n_elem; i++)
+  {
+    arma::uword r = outIdxs.at(i);
+    backIdxs.at(r, counts.at(r)++) = i;
+  }
+}
+
+template<typename MatType>
+void RepeatType<MatType>::Forward(const MatType& input, MatType& output)
+{
+  output = input.rows(outIdxs);
+}
+
+template<typename MatType>
+void RepeatType<MatType>::Backward(
+    const MatType& /* input */,
+    const MatType& /* output */,
+    const MatType& gy,
+    MatType& g)
+{
+  g = gy.rows(backIdxs.col(0));
+  for (size_t c = 1; c < sizeMult; c++)
+  {
+    g += gy.rows(backIdxs.col(c));
+  }
+  g /= sizeMult;
+}
+
+template<typename MatType>
+template<typename Archive>
+void RepeatType<MatType>::serialize(
+    Archive& ar, const uint32_t /* version */)
+{
+  ar(cereal::base_class<Layer<MatType>>(this));
+
+  ar(CEREAL_NVP(multiples));
+  ar(CEREAL_NVP(interleave));
+  ar(CEREAL_NVP(outIdxs));
+  ar(CEREAL_NVP(sizeMult));
+  ar(CEREAL_NVP(backIdxs));
+}
+
+} // namespace mlpack
+
+#endif
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/serialization.hpp 4.4.0-1/src/mlpack/methods/ann/layer/serialization.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/serialization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/serialization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -72,6 +72,7 @@
     CEREAL_REGISTER_TYPE(mlpack::PReLUType<__VA_ARGS__>); \
     CEREAL_REGISTER_TYPE(mlpack::RBFType<__VA_ARGS__>); \
     CEREAL_REGISTER_TYPE(mlpack::ReLU6Type<__VA_ARGS__>); \
+    CEREAL_REGISTER_TYPE(mlpack::RepeatType<__VA_ARGS__>); \
     CEREAL_REGISTER_TYPE(mlpack::SoftmaxType<__VA_ARGS__>); \
     CEREAL_REGISTER_TYPE(mlpack::SoftminType<__VA_ARGS__>); \
     CEREAL_REGISTER_TYPE(mlpack::HardTanHType<__VA_ARGS__>); \
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/softmax_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/softmax_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/softmax_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/softmax_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,8 +62,7 @@ SoftmaxType<MatType>::operator=(SoftmaxT
 template<typename MatType>
 void SoftmaxType<MatType>::Forward(const MatType& input, MatType& output)
 {
-  MatType softmaxInput = arma::exp(input.each_row() -
-      arma::max(input, 0));
+  MatType softmaxInput = exp(input.each_row() - max(input, 0));
   output = softmaxInput.each_row() / sum(softmaxInput, 0);
 }
 
@@ -74,7 +73,7 @@ void SoftmaxType<MatType>::Backward(
     const MatType& gy,
     MatType& g)
 {
-  g = output % (gy - arma::repmat(arma::sum(gy % output), output.n_rows, 1));
+  g = output % (gy - repmat(sum(gy % output), output.n_rows, 1));
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/layer/softmin_impl.hpp 4.4.0-1/src/mlpack/methods/ann/layer/softmin_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/layer/softmin_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/layer/softmin_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,8 +62,8 @@ void SoftminType<MatType>::Forward(
     const MatType& input,
     MatType& output)
 {
-  MatType softminInput = arma::exp(-(input.each_row() -
-      arma::min(input, 0)));
+  MatType softminInput = exp(-(input.each_row() -
+      min(input, 0)));
   output = softminInput.each_row() / sum(softminInput, 0);
 }
 
@@ -74,7 +74,7 @@ void SoftminType<MatType>::Backward(
     const MatType& gy,
     MatType& g)
 {
-  g = output % (gy - arma::repmat(arma::sum(gy % output), output.n_rows, 1));
+  g = output % (gy - repmat(sum(gy % output), output.n_rows, 1));
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/binary_cross_entropy_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/binary_cross_entropy_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/binary_cross_entropy_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/binary_cross_entropy_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,8 +31,8 @@ typename MatType::elem_type BCELossType<
 {
   typedef typename MatType::elem_type ElemType;
 
-  ElemType lossSum = -arma::accu(target % arma::log(prediction + eps) +
-      (1. - target) % arma::log(1. - prediction + eps));
+  ElemType lossSum = -accu(target % log(prediction + eps) +
+      (1. - target) % log(1. - prediction + eps));
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/cosine_embedding_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/cosine_embedding_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/cosine_embedding_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/cosine_embedding_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,8 +37,8 @@ typename MatType::elem_type CosineEmbedd
   if (arma::size(prediction) != arma::size(target))
     Log::Fatal << "Input Tensors must have same dimensions." << std::endl;
 
-  arma::Col<ElemType> inputTemp1 = arma::vectorise(prediction);
-  arma::Col<ElemType> inputTemp2 = arma::vectorise(target);
+  arma::Col<ElemType> inputTemp1 = vectorise(prediction);
+  arma::Col<ElemType> inputTemp2 = vectorise(target);
   ElemType lossSum = 0.0;
 
   for (size_t i = 0; i < inputTemp1.n_elem; i += cols)
@@ -74,8 +74,8 @@ void CosineEmbeddingLossType<MatType>::B
   if (arma::size(prediction) != arma::size(target))
     Log::Fatal << "Input Tensors must have same dimensions." << std::endl;
 
-  arma::Col<ElemType> inputTemp1 = arma::vectorise(prediction);
-  arma::Col<ElemType> inputTemp2 = arma::vectorise(target);
+  arma::Col<ElemType> inputTemp1 = vectorise(prediction);
+  arma::Col<ElemType> inputTemp2 = vectorise(target);
   loss.set_size(arma::size(inputTemp1));
 
   arma::Col<ElemType> outputTemp(loss.memptr(), inputTemp1.n_elem,
@@ -91,9 +91,9 @@ void CosineEmbeddingLossType<MatType>::B
     {
       const int multiplier = similarity ? 1 : -1;
       outputTemp(arma::span(i, i + cols -1)) = -1 * multiplier *
-          (arma::normalise(inputTemp2(arma::span(i, i + cols - 1))) -
-          cosDist * arma::normalise(inputTemp1(arma::span(i, i + cols -
-          1)))) / std::sqrt(arma::accu(arma::pow(inputTemp1(arma::span(i, i +
+          (normalise(inputTemp2(arma::span(i, i + cols - 1))) -
+          cosDist * normalise(inputTemp1(arma::span(i, i + cols -
+          1)))) / std::sqrt(accu(pow(inputTemp1(arma::span(i, i +
           cols - 1)), 2)));
     }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/dice_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/dice_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/dice_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/dice_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -28,8 +28,8 @@ typename MatType::elem_type DiceLossType
     const MatType& prediction,
     const MatType& target)
 {
-  return 1 - ((2 * arma::accu(target % prediction) + smooth) /
-      (arma::accu(target % target) + arma::accu(
+  return 1 - ((2 * accu(target % prediction) + smooth) /
+      (accu(target % target) + accu(
       prediction % prediction) + smooth));
 }
 
@@ -39,10 +39,10 @@ void DiceLossType<MatType>::Backward(
     const MatType& target,
     MatType& loss)
 {
-  loss = -2 * (target * (arma::accu(prediction % prediction) +
-      arma::accu(target % target) + smooth) - prediction *
-      (2 * arma::accu(target % prediction) + smooth)) / std::pow(
-      arma::accu(target % target) + arma::accu(prediction % prediction)
+  loss = -2 * (target * (accu(prediction % prediction) +
+      accu(target % target) + smooth) - prediction *
+      (2 * accu(target % prediction) + smooth)) / std::pow(
+      accu(target % target) + accu(prediction % prediction)
       + smooth, 2.0);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/earth_mover_distance_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/earth_mover_distance_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/earth_mover_distance_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/earth_mover_distance_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ typename MatType::elem_type EarthMoverDi
     const MatType& prediction,
     const MatType& target)
 {
-  typename MatType::elem_type lossSum = -arma::accu(target % prediction);
+  typename MatType::elem_type lossSum = -accu(target % prediction);
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/hinge_embedding_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/hinge_embedding_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/hinge_embedding_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/hinge_embedding_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,7 +31,7 @@ typename MatType::elem_type HingeEmbeddi
     const MatType& target)
 {
   MatType loss = (1 - target) / 2 + prediction % (target);
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/hinge_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/hinge_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/hinge_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/hinge_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -33,9 +33,9 @@ typename MatType::elem_type HingeLossTyp
   MatType temp = target - (target == 0);
   MatType temp_zeros(size(target), arma::fill::zeros);
 
-  MatType loss = arma::max(temp_zeros, 1 - prediction % temp);
+  MatType loss = max(temp_zeros, 1 - prediction % temp);
 
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/kl_divergence_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/kl_divergence_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/kl_divergence_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/kl_divergence_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,8 +30,8 @@ typename MatType::elem_type KLDivergence
     const MatType& prediction,
     const MatType& target)
 {
-  MatType loss = target % (arma::log(target) - prediction);
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  MatType loss = target % (log(target) - prediction);
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/l1_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/l1_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/l1_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/l1_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,7 +30,7 @@ typename MatType::elem_type L1LossType<M
     const MatType& target)
 {
   MatType loss = arma::abs(prediction - target);
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
@@ -44,7 +44,7 @@ void L1LossType<MatType>::Backward(
     const MatType& target,
     MatType& loss)
 {
-  loss = arma::sign(prediction - target);
+  loss = sign(prediction - target);
 
   if (!reduction)
     loss = loss / target.n_elem;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/log_cosh_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/log_cosh_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/log_cosh_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/log_cosh_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,7 +34,7 @@ typename MatType::elem_type LogCoshLossT
     const MatType& target)
 {
   typename MatType::elem_type lossSum =
-      arma::accu(arma::log(arma::cosh(a * (target - prediction)))) / a;
+      accu(log(arma::cosh(a * (target - prediction)))) / a;
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/margin_ranking_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/margin_ranking_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/margin_ranking_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/margin_ranking_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,7 +37,7 @@ typename MatType::elem_type MarginRankin
   const MatType& prediction2 = prediction.rows(predictionRows / 2,
       predictionRows - 1);
 
-  double lossSum = arma::accu(arma::max(arma::zeros(size(target)),
+  double lossSum = accu(max(zeros(size(target)),
       -target % (prediction1 - prediction2) + margin));
 
   if (reduction)
@@ -63,7 +63,7 @@ void MarginRankingLossType<MatType>::Bac
   MatType lossPrediction2 = lossPrediction1;
   lossPrediction1 = -target % lossPrediction1;
   lossPrediction2 = target % lossPrediction2;
-  loss = arma::join_cols(lossPrediction1, lossPrediction2);
+  loss = join_cols(lossPrediction1, lossPrediction2);
 
   if (!reduction)
     loss = loss / target.n_elem;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_absolute_percentage_error_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_absolute_percentage_error_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_absolute_percentage_error_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_absolute_percentage_error_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ typename MatType::elem_type MeanAbsolute
     const MatType& target)
 {
   MatType loss = arma::abs((prediction - target) / target);
-  return arma::accu(loss) * (100 / target.n_cols);
+  return accu(loss) * (100 / target.n_cols);
 }
 
 template<typename MatType>
@@ -39,7 +39,7 @@ void MeanAbsolutePercentageErrorType<Mat
     MatType& loss)
 
 {
-  loss = (((arma::conv_to<arma::mat>::from(prediction < target) * -2) + 1) /
+  loss = (((ConvTo<arma::mat>::From(prediction < target) * -2) + 1) /
       target) * (100 / target.n_cols);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_bias_error_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_bias_error_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_bias_error_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_bias_error_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,7 +31,7 @@ typename MatType::elem_type MeanBiasErro
     const MatType& target)
 {
   MatType loss = target - prediction;
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_squared_error_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_squared_error_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_squared_error_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_squared_error_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,8 +29,7 @@ typename MatType::elem_type MeanSquaredE
     const MatType& prediction,
     const MatType& target)
 {
-  typename MatType::elem_type lossSum =
-      arma::accu(arma::square(prediction - target));
+  typename MatType::elem_type lossSum = accu(square(prediction - target));
 
   if (reduction)
     return lossSum;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_squared_logarithmic_error_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_squared_logarithmic_error_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/mean_squared_logarithmic_error_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/mean_squared_logarithmic_error_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,8 +31,7 @@ typename MatType::elem_type MeanSquaredL
     const MatType& target)
 {
   typename MatType::elem_type lossSum =
-      arma::accu(arma::square(arma::log(1.0 + target) -
-      arma::log(1.0 + prediction)));
+      accu(square(log(1.0 + target) - log(1.0 + prediction)));
 
   if (reduction)
     return lossSum;
@@ -46,8 +45,7 @@ void MeanSquaredLogarithmicErrorType<Mat
     const MatType& target,
     MatType& loss)
 {
-  loss = 2 * (arma::log(1. + prediction) - arma::log(1. + target)) /
-      (1. + prediction);
+  loss = 2 * (log(1. + prediction) - log(1. + target)) / (1. + prediction);
 
   if (!reduction)
     loss = loss / target.n_elem;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/multilabel_softmargin_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/multilabel_softmargin_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/multilabel_softmargin_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/multilabel_softmargin_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,9 +41,9 @@ typename MatType::elem_type MultiLabelSo
     weighted = true;
   }
 
-  MatType logSigmoid = arma::log((1 / (1 + arma::exp(-input))));
-  MatType logSigmoidNeg = arma::log(1 / (1 + arma::exp(input)));
-  MatType loss = arma::mean(arma::sum(-(target % logSigmoid +
+  MatType logSigmoid = log((1 / (1 + exp(-input))));
+  MatType logSigmoidNeg = log(1 / (1 + exp(input)));
+  MatType loss = arma::mean(sum(-(target % logSigmoid +
       (1 - target) % logSigmoidNeg)) % classWeights, 1);
 
   if (reduction)
@@ -59,9 +59,9 @@ void MultiLabelSoftMarginLossType<MatTyp
     MatType& output)
 {
   output.set_size(size(input));
-  MatType sigmoid = (1 / (1 + arma::exp(-input)));
+  MatType sigmoid = (1 / (1 + exp(-input)));
   output = -(target % (1 - sigmoid) - (1 - target) % sigmoid) %
-        arma::repmat(classWeights, target.n_rows, 1) / output.n_elem;
+        repmat(classWeights, target.n_rows, 1) / output.n_elem;
 
   if (reduction)
     output = output * input.n_rows;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/negative_log_likelihood_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/negative_log_likelihood_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/negative_log_likelihood_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/negative_log_likelihood_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,7 +51,7 @@ void NegativeLogLikelihoodType<MatType>:
       const MatType& target,
       MatType& loss)
 {
-  loss = arma::zeros<MatType>(prediction.n_rows, prediction.n_cols);
+  loss = zeros<MatType>(prediction.n_rows, prediction.n_cols);
   for (size_t i = 0; i < prediction.n_cols; ++i)
   {
     Log::Assert(target(i) >= 0 && target(i) < prediction.n_rows,
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/poisson_nll_loss.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/poisson_nll_loss.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/poisson_nll_loss.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/poisson_nll_loss.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -111,8 +111,7 @@ class PoissonNLLLossType
 
  private:
   //! Check if the probabilities lie in the range [0, 1].
-  template<typename eT>
-  void CheckProbs(const arma::Mat<eT>& probs)
+  void CheckProbs(const MatType& probs)
   {
     for (size_t i = 0; i < probs.size(); ++i)
     {
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/poisson_nll_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/poisson_nll_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/poisson_nll_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/poisson_nll_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -40,21 +40,21 @@ typename MatType::elem_type PoissonNLLLo
   MatType loss(arma::size(prediction));
 
   if (logInput)
-    loss = arma::exp(prediction) - target % prediction;
+    loss = exp(prediction) - target % prediction;
   else
   {
     CheckProbs(prediction);
-    loss = prediction - target % arma::log(prediction + eps);
+    loss = prediction - target % log(prediction + eps);
   }
 
   if (full)
   {
     const auto mask = target > 1.0;
-    const MatType approx = target % arma::log(target) - target
-        + 0.5 * arma::log(2 * M_PI * target);
+    const MatType approx = target % log(target) - target
+        + 0.5 * log(2 * M_PI * target);
     loss.elem(arma::find(mask)) += approx.elem(arma::find(mask));
   }
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  typename MatType::elem_type lossSum = accu(loss);
   
   if (reduction)
     return lossSum;
@@ -71,7 +71,7 @@ void PoissonNLLLossType<MatType>::Backwa
   loss.set_size(size(prediction));
 
   if (logInput)
-    loss = (arma::exp(prediction) - target);
+    loss = (exp(prediction) - target);
   else
     loss = (1 - target / (prediction + eps));
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/sigmoid_cross_entropy_error_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/sigmoid_cross_entropy_error_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/sigmoid_cross_entropy_error_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/sigmoid_cross_entropy_error_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,7 +41,7 @@ SigmoidCrossEntropyErrorType<MatType>::F
         std::log(1 + std::exp(-std::abs(prediction[i])));
   }
 
-  ElemType lossSum = maximum - arma::accu(prediction % target);
+  ElemType lossSum = maximum - accu(prediction % target);
 
   if (reduction)
     return lossSum;
@@ -55,7 +55,7 @@ inline void SigmoidCrossEntropyErrorType
     const MatType& target,
     MatType& loss)
 {
-  loss = 1.0 / (1.0 + arma::exp(-prediction)) - target;
+  loss = 1.0 / (1.0 + exp(-prediction)) - target;
 
   if (!reduction)
     loss = loss / target.n_elem;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/soft_margin_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/soft_margin_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/soft_margin_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/soft_margin_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -28,8 +28,8 @@ template<typename MatType>
 typename MatType::elem_type SoftMarginLossType<MatType>::Forward(
     const MatType& prediction, const MatType& target)
 {
-  MatType loss = arma::log(1 + arma::exp(-target % prediction));
-  typename MatType::elem_type lossSum = arma::accu(loss);
+  MatType loss = log(1 + exp(-target % prediction));
+  typename MatType::elem_type lossSum = accu(loss);
 
   if (reduction)
     return lossSum;
@@ -44,7 +44,7 @@ void SoftMarginLossType<MatType>::Backwa
     MatType& loss)
 {
   loss.set_size(size(prediction));
-  MatType temp = arma::exp(-target % prediction);
+  MatType temp = exp(-target % prediction);
   MatType numerator = -target % temp;
   MatType denominator = 1 + temp;
   loss = numerator / denominator;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/triplet_margin_loss_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/triplet_margin_loss_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/triplet_margin_loss_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/triplet_margin_loss_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,8 +35,8 @@ typename MatType::elem_type TripletMargi
   MatType positive =
       prediction.submat(prediction.n_rows / 2, 0, prediction.n_rows - 1,
       prediction.n_cols - 1);
-  return std::max(0.0, arma::accu(arma::pow(anchor - positive, 2)) -
-      arma::accu(arma::pow(anchor - target, 2)) + margin) / anchor.n_cols;
+  return std::max(0.0, accu(pow(anchor - positive, 2)) -
+      accu(pow(anchor - target, 2)) + margin) / anchor.n_cols;
 }
 
 template<typename MatType>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/loss_functions/vr_class_reward_impl.hpp 4.4.0-1/src/mlpack/methods/ann/loss_functions/vr_class_reward_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/loss_functions/vr_class_reward_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/loss_functions/vr_class_reward_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,7 +65,7 @@ void VRClassRewardType<MatType>::Backwar
     const MatType& target,
     MatType& output)
 {
-  output = arma::zeros<MatType>(input.n_rows, input.n_cols);
+  output = zeros<MatType>(input.n_rows, input.n_cols);
   for (size_t i = 0; i < (input.n_cols - 1); ++i)
   {
     const size_t currentTarget = target(i);
diff -pruN 4.3.0-2/src/mlpack/methods/ann/make_alias.hpp 4.4.0-1/src/mlpack/methods/ann/make_alias.hpp
--- 4.3.0-2/src/mlpack/methods/ann/make_alias.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/make_alias.hpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,58 +0,0 @@
-/**
- * @file make_alias.hpp
- * @author Ryan Curtin
- *
- * Implementation of `MakeAlias()`, a utility function.  This is meant to be
- * used in `SetWeights()` calls in various layers, to wrap internal weight
- * objects as aliases around the given memory pointers.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#ifndef MLPACK_METHODS_ANN_MAKE_ALIAS_HPP
-#define MLPACK_METHODS_ANN_MAKE_ALIAS_HPP
-
-#include <mlpack/prereqs.hpp>
-
-namespace mlpack {
-
-/**
- * Reconstruct `m` as an alias around the memory `newMem`, with size `numRows` x
- * `numCols`.
- */
-template<typename MatType>
-void MakeAlias(MatType& m,
-               typename MatType::elem_type* newMem,
-               const size_t numRows,
-               const size_t numCols)
-{
-  // We use placement new to reinitialize the object, since the copy and move
-  // assignment operators in Armadillo will end up copying memory instead of
-  // making an alias.
-  m.~MatType();
-  new (&m) MatType(newMem, numRows, numCols, false, true);
-}
-
-/**
- * Reconstruct `c` as an alias around the memory` newMem`, with size `numRows` x
- * `numCols` x `numSlices`.
- */
-template<typename CubeType>
-void MakeAlias(CubeType& c,
-               typename CubeType::elem_type* newMem,
-               const size_t numRows,
-               const size_t numCols,
-               const size_t numSlices)
-{
-  // We use placement new to reinitialize the object, since the copy and move
-  // assignment operators in Armadillo will end up copying memory instead of
-  // making an alias.
-  c.~CubeType();
-  new (&c) CubeType(newMem, numRows, numCols, numSlices, false, true);
-}
-
-} // namespace mlpack
-
-#endif
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/brnn_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/brnn_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/brnn_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/brnn_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -203,11 +203,11 @@ void BRNN<OutputLayerType, MergeLayerTyp
 
   if (std::is_same<MergeLayerType, Concat<>>::value)
   {
-    results = arma::zeros<arma::cube>(outputSize * 2, predictors.n_cols, rho);
+    results = zeros<arma::cube>(outputSize * 2, predictors.n_cols, rho);
   }
   else
   {
-    results = arma::zeros<arma::cube>(outputSize, predictors.n_cols, rho);
+    results = zeros<arma::cube>(outputSize, predictors.n_cols, rho);
   }
 
   std::vector<arma::mat> results1, results2;
@@ -375,7 +375,7 @@ EvaluateWithGradient(const arma::mat& /*
     {
       ResetParameters();
     }
-    gradient = arma::zeros<arma::mat>(parameter.n_rows, parameter.n_cols);
+    gradient = zeros<arma::mat>(parameter.n_rows, parameter.n_cols);
   }
   else
   {
@@ -384,12 +384,8 @@ EvaluateWithGradient(const arma::mat& /*
 
   if (backwardGradient.is_empty())
   {
-    backwardGradient = arma::zeros<arma::mat>(
-        parameter.n_rows/ 2,
-        parameter.n_cols);
-    forwardGradient = arma::zeros<arma::mat>(
-        parameter.n_rows/ 2,
-        parameter.n_cols);
+    backwardGradient = zeros<arma::mat>(parameter.n_rows / 2, parameter.n_cols);
+    forwardGradient = zeros<arma::mat>(parameter.n_rows / 2, parameter.n_cols);
   }
   if (this->deterministic)
   {
@@ -444,11 +440,11 @@ EvaluateWithGradient(const arma::mat& /*
   arma::cube results;
   if (std::is_same<MergeLayerType, Concat<>>::value)
   {
-    results = arma::zeros<arma::cube>(outputSize * 2, batchSize, rho);
+    results = zeros<arma::cube>(outputSize * 2, batchSize, rho);
   }
   else
   {
-    results = arma::zeros<arma::cube>(outputSize, batchSize, rho);
+    results = zeros<arma::cube>(outputSize, batchSize, rho);
   }
 
   double performance = 0;
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/gan_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/gan_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/gan_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/gan_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -158,7 +158,7 @@ void GAN<Model, InitializationRuleType,
 
   responses.ones(1, numFunctions + batchSize);
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      arma::zeros(1, batchSize);
+      zeros(1, batchSize);
   this->discriminator.responses = arma::mat(this->responses.memptr(),
       this->responses.n_rows, this->responses.n_cols, false, false);
 
@@ -270,7 +270,7 @@ GAN<Model, InitializationRuleType, Noise
   discriminator.Forward(predictors.cols(numFunctions,
       numFunctions + batchSize - 1));
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      arma::zeros(1, batchSize);
+      zeros(1, batchSize);
 
   currentTarget = arma::mat(responses.memptr() + numFunctions,
       1, batchSize, false, false);
@@ -305,7 +305,7 @@ EvaluateWithGradient(const arma::mat& /*
   {
     if (parameter.is_empty())
       Reset();
-    gradient = arma::zeros<arma::mat>(parameter.n_elem, 1);
+    gradient = zeros<arma::mat>(parameter.n_elem, 1);
   }
   else
     gradient.zeros();
@@ -318,7 +318,7 @@ EvaluateWithGradient(const arma::mat& /*
 
   if (noiseGradientDiscriminator.is_empty())
   {
-    noiseGradientDiscriminator = arma::zeros<arma::mat>(
+    noiseGradientDiscriminator = zeros<arma::mat>(
         gradientDiscriminator.n_elem, 1);
   }
   else
@@ -342,7 +342,7 @@ EvaluateWithGradient(const arma::mat& /*
   predictors.cols(numFunctions, numFunctions + batchSize - 1) =
       boost::apply_visitor(outputParameterVisitor, generator.network.back());
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      arma::zeros(1, batchSize);
+      zeros(1, batchSize);
 
   // Get the gradients of the Generator.
   res += discriminator.EvaluateWithGradient(discriminator.parameter,
@@ -354,7 +354,7 @@ EvaluateWithGradient(const arma::mat& /*
     // Minimize -log(D(G(noise))).
     // Pass the error from Discriminator to Generator.
     responses.cols(numFunctions, numFunctions + batchSize - 1) =
-        arma::ones(1, batchSize);
+        ones(1, batchSize);
 
     discriminator.outputLayer.Backward(
         boost::apply_visitor(outputParameterVisitor,
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/metrics/inception_score_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/metrics/inception_score_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/metrics/inception_score_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/metrics/inception_score_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,11 +41,11 @@ double InceptionScore(ModelType model,
     }
     arma::mat curPreds =
         arma::mat(preds.colptr(index), preds.n_rows, curSize, false, true);
-    arma::colvec c = arma::log(arma::mean(curPreds, 1));
-    arma::mat temp = arma::log(curPreds);
+    arma::colvec c = log(arma::mean(curPreds, 1));
+    arma::mat temp = log(curPreds);
     temp.each_col() -= c;
     curPreds %= temp;
-    scores(i) = exp(arma::as_scalar(arma::mean(arma::sum(curPreds, 0))));
+    scores(i) = exp(arma::as_scalar(arma::mean(sum(curPreds, 0))));
     index += curSize;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/wgan_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/wgan_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/wgan_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/wgan_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -63,7 +63,7 @@ GAN<Model, InitializationRuleType, Noise
   discriminator.Forward(predictors.cols(numFunctions,
       numFunctions + batchSize - 1));
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      -arma::ones(1, batchSize);
+      -ones(1, batchSize);
 
   currentTarget = arma::mat(responses.memptr() + numFunctions,
       1, batchSize, false, false);
@@ -98,7 +98,7 @@ EvaluateWithGradient(const arma::mat& /*
   {
     if (parameter.is_empty())
       Reset();
-    gradient = arma::zeros<arma::mat>(parameter.n_elem, 1);
+    gradient = zeros<arma::mat>(parameter.n_elem, 1);
   }
   else
     gradient.zeros();
@@ -111,7 +111,7 @@ EvaluateWithGradient(const arma::mat& /*
 
   if (noiseGradientDiscriminator.is_empty())
   {
-    noiseGradientDiscriminator = arma::zeros<arma::mat>(
+    noiseGradientDiscriminator = zeros<arma::mat>(
         gradientDiscriminator.n_elem, 1);
   }
   else
@@ -135,7 +135,7 @@ EvaluateWithGradient(const arma::mat& /*
   predictors.cols(numFunctions, numFunctions + batchSize - 1) =
       boost::apply_visitor(outputParameterVisitor, generator.network.back());
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      -arma::ones(1, batchSize);
+      -ones(1, batchSize);
 
   // Get the gradients of the Generator.
   res += discriminator.EvaluateWithGradient(discriminator.parameter,
@@ -149,7 +149,7 @@ EvaluateWithGradient(const arma::mat& /*
     // Minimize -D(G(noise)).
     // Pass the error from Discriminator to Generator.
     responses.cols(numFunctions, numFunctions + batchSize - 1) =
-        arma::ones(1, batchSize);
+        ones(1, batchSize);
 
     discriminator.outputLayer.Backward(
         boost::apply_visitor(outputParameterVisitor,
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/wgangp_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/wgangp_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/gan/wgangp_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/gan/wgangp_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -66,7 +66,7 @@ GAN<Model, InitializationRuleType, Noise
   discriminator.Forward(std::move(predictors.cols(numFunctions,
       numFunctions + batchSize - 1)));
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      -arma::ones(1, batchSize);
+      -ones(1, batchSize);
 
   currentTarget = arma::mat(responses.memptr() + numFunctions,
       1, batchSize, false, false);
@@ -80,10 +80,10 @@ GAN<Model, InitializationRuleType, Noise
   predictors.cols(numFunctions, numFunctions + batchSize - 1) =
       (epsilon * currentInput) + ((1.0 - epsilon) * generatedData);
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      -arma::ones(1, batchSize);
+      -ones(1, batchSize);
   discriminator.Gradient(discriminator.parameter, numFunctions,
       normGradientDiscriminator, batchSize);
-  res += lambda * std::pow(arma::norm(normGradientDiscriminator, 2) - 1, 2);
+  res += lambda * std::pow(norm(normGradientDiscriminator, 2) - 1, 2);
 
   return res;
 }
@@ -112,7 +112,7 @@ EvaluateWithGradient(const arma::mat& /*
   {
     if (parameter.is_empty())
       Reset();
-    gradient = arma::zeros<arma::mat>(parameter.n_elem, 1);
+    gradient = zeros<arma::mat>(parameter.n_elem, 1);
   }
   else
     gradient.zeros();
@@ -125,7 +125,7 @@ EvaluateWithGradient(const arma::mat& /*
 
   if (noiseGradientDiscriminator.is_empty())
   {
-    noiseGradientDiscriminator = arma::zeros<arma::mat>(
+    noiseGradientDiscriminator = zeros<arma::mat>(
         gradientDiscriminator.n_elem, 1);
   }
   else
@@ -157,10 +157,10 @@ EvaluateWithGradient(const arma::mat& /*
   predictors.cols(numFunctions, numFunctions + batchSize - 1) =
       (epsilon * currentInput) + ((1.0 - epsilon) * generatedData);
   responses.cols(numFunctions, numFunctions + batchSize - 1) =
-      -arma::ones(1, batchSize);
+      -ones(1, batchSize);
   discriminator.Gradient(discriminator.parameter, numFunctions,
       normGradientDiscriminator, batchSize);
-  res += lambda * std::pow(arma::norm(normGradientDiscriminator, 2) - 1, 2);
+  res += lambda * std::pow(norm(normGradientDiscriminator, 2) - 1, 2);
 
   predictors.cols(numFunctions, numFunctions + batchSize - 1) =
       generatedData;
@@ -173,7 +173,7 @@ EvaluateWithGradient(const arma::mat& /*
     // Minimize -D(G(noise)).
     // Pass the error from Discriminator to Generator.
     responses.cols(numFunctions, numFunctions + batchSize - 1) =
-        arma::ones(1, batchSize);
+        ones(1, batchSize);
 
     discriminator.outputLayer.Backward(
         boost::apply_visitor(outputParameterVisitor,
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/rbm/rbm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/rbm/rbm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/rbm/rbm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/rbm/rbm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -114,8 +114,8 @@ RBM<InitializationRuleType, DataType, Po
 {
   preActivation = (weight.slice(0) * input);
   preActivation.each_col() += hiddenBias;
-  return -(arma::accu(arma::log(1 + arma::trunc_exp(preActivation))) +
-      arma::dot(input, arma::repmat(visibleBias, 1, input.n_cols)));
+  return -(accu(log(1 + arma::trunc_exp(preActivation))) +
+      dot(input, repmat(visibleBias, 1, input.n_cols)));
 }
 
 template<
diff -pruN 4.3.0-2/src/mlpack/methods/ann/not_adapted/rbm/spike_slab_rbm_impl.hpp 4.4.0-1/src/mlpack/methods/ann/not_adapted/rbm/spike_slab_rbm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/not_adapted/rbm/spike_slab_rbm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/not_adapted/rbm/spike_slab_rbm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -69,14 +69,14 @@ typename std::enable_if<std::is_same<Pol
 RBM<InitializationRuleType, DataType, PolicyType>::FreeEnergy(
     const arma::Mat<ElemType>& input)
 {
-  ElemType freeEnergy = 0.5 * visiblePenalty(0) * arma::dot(input, input);
+  ElemType freeEnergy = 0.5 * visiblePenalty(0) * dot(input, input);
 
   freeEnergy -= 0.5 * hiddenSize * poolSize *
       std::log((2.0 * M_PI) / slabPenalty);
 
   for (size_t i = 0; i < hiddenSize; ++i)
   {
-    ElemType sum = arma::accu(arma::square(input.t() * weight.slice(i))) /
+    ElemType sum = accu(square(input.t() * weight.slice(i))) /
         (2.0 * slabPenalty);
     freeEnergy -= SoftplusFunction::Fn(spikeBias(i) - sum);
   }
@@ -107,13 +107,13 @@ RBM<InitializationRuleType, DataType, Po
 
   for (size_t i = 0 ; i < hiddenSize; ++i)
   {
-    weightGrad.slice(i) = input * arma::repmat(slabMean.col(i).t(),
+    weightGrad.slice(i) = input * repmat(slabMean.col(i).t(),
         input.n_cols, 1) * spikeMean(i);
   }
 
   spikeBiasGrad = spikeMean;
   // Setting visiblePenaltyGrad.
-  gradient.row(weightGrad.n_elem + spikeBiasGrad.n_elem) = -0.5 * arma::dot(
+  gradient.row(weightGrad.n_elem + spikeBiasGrad.n_elem) = -0.5 * dot(
        input, input) / std::pow(input.n_cols, 2);
 }
 
@@ -163,7 +163,7 @@ RBM<InitializationRuleType, DataType, Po
     {
       output(i) = RandNormal(visibleMean(i), 1.0 / visiblePenalty(0));
     }
-    if (arma::norm(output, 2) < radius)
+    if (norm(output, 2) < radius)
     {
       break;
     }
@@ -172,7 +172,7 @@ RBM<InitializationRuleType, DataType, Po
   if (k == numMaxTrials)
   {
     Log::Warn << "Outputs are still not in visible unit "
-        << arma::norm(output, 2)
+        << norm(output, 2)
         << " terminating optimization."
         << std::endl;
   }
@@ -238,7 +238,7 @@ RBM<InitializationRuleType, DataType, Po
 {
   for (size_t i = 0; i < hiddenSize; ++i)
   {
-    spikeMean(i) = LogisticFunction::Fn(0.5 * (1.0 / slabPenalty) * arma::accu(
+    spikeMean(i) = LogisticFunction::Fn(0.5 * (1.0 / slabPenalty) * accu(
         visible.t() * (weight.slice(i) * weight.slice(i).t()) * visible)
         / std::pow(visible.n_cols, 2) + spikeBias(i));
   }
diff -pruN 4.3.0-2/src/mlpack/methods/ann/regularizer/lregularizer_impl.hpp 4.4.0-1/src/mlpack/methods/ann/regularizer/lregularizer_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/regularizer/lregularizer_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/regularizer/lregularizer_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ template<int Power>
 template<typename MatType>
 void LRegularizer<Power>::Evaluate(const MatType& weight, MatType& gradient)
 {
-  gradient += arma::vectorise(arma::pow(weight, Power - 1) * Power * factor);
+  gradient += vectorise(pow(weight, Power - 1) * Power * factor);
 }
 
 // L1-Regularizer specializations.
@@ -37,7 +37,7 @@ template<>
 template<typename MatType>
 void LRegularizer<1>::Evaluate(const MatType& weight, MatType& gradient)
 {
-  gradient += arma::vectorise(factor * arma::sign(weight));
+  gradient += vectorise(factor * sign(weight));
 }
 
 // L2-Regularizer specializations.
@@ -45,7 +45,7 @@ template<>
 template<typename MatType>
 void LRegularizer<2>::Evaluate(const MatType& weight, MatType& gradient)
 {
-  gradient += arma::vectorise(2 * factor * weight);
+  gradient += vectorise(2 * factor * weight);
 }
 
 template<int Power>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/regularizer/orthogonal_regularizer_impl.hpp 4.4.0-1/src/mlpack/methods/ann/regularizer/orthogonal_regularizer_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/regularizer/orthogonal_regularizer_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/regularizer/orthogonal_regularizer_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ inline OrthogonalRegularizer::Orthogonal
 template<typename MatType>
 void OrthogonalRegularizer::Evaluate(const MatType& weight, MatType& gradient)
 {
-  arma::mat grad = arma::zeros(arma::size(weight));
+  arma::mat grad = zeros(arma::size(weight));
 
   for (size_t i = 0; i < weight.n_rows; ++i)
   {
@@ -34,22 +34,19 @@ void OrthogonalRegularizer::Evaluate(con
     {
       if (i == j)
       {
-        double s =
-            arma::as_scalar(
-            arma::sign((weight.row(i) * weight.row(i).t()) - 1));
+        double s = arma::as_scalar(sign((weight.row(i) * weight.row(i).t()) - 1));
         grad.row(i) += 2 * s * weight.row(i);
       }
       else
       {
-        double s = arma::as_scalar(
-            arma::sign(weight.row(i) * weight.row(j).t()));
+        double s = arma::as_scalar(sign(weight.row(i) * weight.row(j).t()));
         grad.row(i) += s * weight.row(j);
         grad.row(j) += s * weight.row(i);
       }
     }
   }
 
-  gradient += arma::vectorise(grad) * factor;
+  gradient += vectorise(grad) * factor;
 }
 
 template<typename Archive>
diff -pruN 4.3.0-2/src/mlpack/methods/ann/rnn_impl.hpp 4.4.0-1/src/mlpack/methods/ann/rnn_impl.hpp
--- 4.3.0-2/src/mlpack/methods/ann/rnn_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/ann/rnn_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -239,11 +239,10 @@ void RNN<
         SetPreviousStep(size_t(0));
 
       // Create aliases for the input and output.
-      MakeAlias(inputAlias,
-          (typename MatType::elem_type*) predictors.slice(t).colptr(i),
-          predictors.n_rows, effectiveBatchSize);
-      MakeAlias(outputAlias, results.slice(t).colptr(i), results.n_rows,
-          effectiveBatchSize);
+      MakeAlias(inputAlias, predictors.slice(t), predictors.n_rows,
+          effectiveBatchSize, i * predictors.slice(t).n_rows);
+      MakeAlias(outputAlias, results.slice(t), results.n_rows,
+          effectiveBatchSize, i * results.slice(t).n_rows);
 
       network.Forward(inputAlias, outputAlias);
     }
@@ -349,11 +348,12 @@ typename MatType::elem_type RNN<
 
     // Manually reset the data of the network to be an alias of the current time
     // step.
-    MakeAlias(network.predictors, predictors.slice(t).colptr(begin),
-        predictors.n_rows, batchSize);
+    MakeAlias(network.predictors, predictors.slice(t), predictors.n_rows,
+        batchSize, begin * predictors.slice(t).n_rows);
     const size_t responseStep = (single) ? 0 : t;
-    MakeAlias(network.responses, responses.slice(responseStep).colptr(begin),
-        responses.n_rows, batchSize);
+    MakeAlias(network.responses, responses.slice(responseStep),
+        responses.n_rows, batchSize,
+        begin * responses.slice(responseStep).n_rows);
 
     loss += network.Evaluate(output, begin, batchSize);
   }
@@ -418,15 +418,15 @@ typename MatType::elem_type RNN<
     SetCurrentStep(0);
 
     // Make an alias of the step's data.
-    MakeAlias(stepData, predictors.slice(t).colptr(begin), predictors.n_rows,
-        batchSize);
-    MakeAlias(outputData, outputs.slice(t).memptr(), outputs.n_rows,
-        outputs.n_cols);
+    MakeAlias(stepData, predictors.slice(t), predictors.n_rows, batchSize,
+        begin * predictors.slice(t).n_rows);
+    MakeAlias(outputData, outputs.slice(t), outputs.n_rows, outputs.n_cols);
     network.network.Forward(stepData, outputData);
 
     const size_t responseStep = (single) ? 0 : t;
-    MakeAlias(responseData, responses.slice(responseStep).colptr(begin),
-        responses.n_rows, batchSize);
+    MakeAlias(responseData, responses.slice(responseStep),
+        responses.n_rows, batchSize,
+        begin * responses.slice(responseStep).n_rows);
 
     loss += network.outputLayer.Forward(outputData, responseData);
 
@@ -440,15 +440,15 @@ typename MatType::elem_type RNN<
     SetCurrentStep(t - extraSteps + 1);
 
     // Wrap a matrix around our data to avoid a copy.
-    MakeAlias(stepData, predictors.slice(t).colptr(begin), predictors.n_rows,
-        batchSize);
-    MakeAlias(outputData, outputs.slice(t).memptr(), outputs.n_rows,
-        outputs.n_cols);
+    MakeAlias(stepData, predictors.slice(t), predictors.n_rows, batchSize,
+        begin * predictors.slice(t).n_rows);
+    MakeAlias(outputData, outputs.slice(t), outputs.n_rows, outputs.n_cols);
     network.network.Forward(stepData, outputData);
 
     const size_t responseStep = (single) ? 0 : t;
-    MakeAlias(responseData, responses.slice(responseStep).colptr(begin),
-        responses.n_rows, batchSize);
+    MakeAlias(responseData, responses.slice(responseStep),
+        responses.n_rows, batchSize,
+        begin * responses.slice(responseStep).n_rows);
 
     loss += network.outputLayer.Forward(outputData, responseData);
 
@@ -484,18 +484,18 @@ typename MatType::elem_type RNN<
     }
     else
     {
-      MakeAlias(outputData, outputs.slice(t - 1).colptr(0), outputs.n_rows,
+      MakeAlias(outputData, outputs.slice(t - 1), outputs.n_rows,
           outputs.n_cols);
       const size_t respStep = (single) ? 0 : t - 1;
-      MakeAlias(responseData, responses.slice(respStep).colptr(begin),
-          responses.n_rows, batchSize);
+      MakeAlias(responseData, responses.slice(respStep), responses.n_rows,
+          batchSize, begin * responses.slice(respStep).n_rows);
       network.outputLayer.Backward(outputData, responseData, error);
     }
 
     // Now pass that error backwards through the network.
-    MakeAlias(stepData, predictors.slice(t - 1).colptr(begin),
-        predictors.n_rows, batchSize);
-    MakeAlias(outputData, outputs.slice(t - 1).colptr(0), outputs.n_rows,
+    MakeAlias(stepData, predictors.slice(t - 1), predictors.n_rows, batchSize,
+        begin * predictors.slice(t - 1).n_rows);
+    MakeAlias(outputData, outputs.slice(t - 1), outputs.n_rows,
         outputs.n_cols);
 
     MatType networkDelta;
diff -pruN 4.3.0-2/src/mlpack/methods/approx_kfn/approx_kfn_main.cpp 4.4.0-1/src/mlpack/methods/approx_kfn/approx_kfn_main.cpp
--- 4.3.0-2/src/mlpack/methods/approx_kfn/approx_kfn_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/approx_kfn/approx_kfn_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -114,8 +114,7 @@ BINDING_SEE_ALSO("k-nearest-neighbor sea
 BINDING_SEE_ALSO("Fast approximate furthest neighbors with data-dependent"
     " candidate selection (pdf)", "http://ratml.org/pub/pdf/2016fast.pdf");
 BINDING_SEE_ALSO("Approximate furthest neighbor in high dimensions (pdf)",
-    "https://pdfs.semanticscholar.org/a4b5/7b9cbf37201fb1d9a56c0f4eefad0466"
-    "9c20.pdf");
+    "https://www.rasmuspagh.net/papers/approx-furthest-neighbor-SISAP15.pdf");
 BINDING_SEE_ALSO("QDAFN class documentation",
     "@src/mlpack/methods/approx_kfn/qdafn.hpp");
 BINDING_SEE_ALSO("DrusillaSelect class documentation",
@@ -336,11 +335,11 @@ void BINDING_FUNCTION(util::Params& para
         Log::Info << "Calculation complete." << endl;
       }
 
-      const double averageError = arma::sum(exactDistances.row(0) /
+      const double averageError = sum(exactDistances.row(0) /
           distances.row(0)) / distances.n_cols;
-      const double minError = arma::min(exactDistances.row(0) /
+      const double minError = min(exactDistances.row(0) /
           distances.row(0));
-      const double maxError = arma::max(exactDistances.row(0) /
+      const double maxError = max(exactDistances.row(0) /
           distances.row(0));
 
       Log::Info << "Average error: " << averageError << "." << endl;
diff -pruN 4.3.0-2/src/mlpack/methods/approx_kfn/drusilla_select_impl.hpp 4.4.0-1/src/mlpack/methods/approx_kfn/drusilla_select_impl.hpp
--- 4.3.0-2/src/mlpack/methods/approx_kfn/drusilla_select_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/approx_kfn/drusilla_select_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -87,7 +87,7 @@ void DrusillaSelect<MatType>::Train(
   for (size_t i = 0; i < refCopy.n_cols; ++i)
   {
     refCopy.col(i) = referenceSet.col(i) - dataMean;
-    norms[i] = arma::norm(refCopy.col(i));
+    norms[i] = norm(refCopy.col(i));
   }
 
   // Find the top m points for each of the l projections...
@@ -97,7 +97,7 @@ void DrusillaSelect<MatType>::Train(
     arma::uword maxIndex = 0;
     norms.max(maxIndex);
 
-    arma::vec line(refCopy.col(maxIndex) / arma::norm(refCopy.col(maxIndex)));
+    arma::vec line(refCopy.col(maxIndex) / norm(refCopy.col(maxIndex)));
 
     // Calculate distortion and offset and make scores.
     std::vector<bool> closeAngle(referenceSet.n_cols, false);
@@ -106,8 +106,8 @@ void DrusillaSelect<MatType>::Train(
     {
       if (norms[j] > 0.0)
       {
-        const double offset = arma::dot(refCopy.col(j), line);
-        const double distortion = arma::norm(refCopy.col(j) - offset * line);
+        const double offset = dot(refCopy.col(j), line);
+        const double distortion = norm(refCopy.col(j) - offset * line);
         sums[j] = std::abs(offset) - std::abs(distortion);
         closeAngle[j] =
             (std::atan(distortion / std::abs(offset)) < (M_PI / 8.0));
diff -pruN 4.3.0-2/src/mlpack/methods/approx_kfn/qdafn_impl.hpp 4.4.0-1/src/mlpack/methods/approx_kfn/qdafn_impl.hpp
--- 4.3.0-2/src/mlpack/methods/approx_kfn/qdafn_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/approx_kfn/qdafn_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -112,14 +112,14 @@ void QDAFN<MatType>::Search(const MatTyp
     std::priority_queue<std::pair<double, size_t>> queue;
     for (size_t i = 0; i < l; ++i)
     {
-      const double val = sValues(0, i) - arma::dot(querySet.col(q),
+      const double val = sValues(0, i) - dot(querySet.col(q),
           lines.col(i));
       queue.push(std::make_pair(val, i));
     }
 
     // To track where we are in each S table, we keep the next index to look at
     // in each table (they start at 0).
-    arma::Col<size_t> tableLocations = arma::zeros<arma::Col<size_t>>(l);
+    arma::Col<size_t> tableLocations = zeros<arma::Col<size_t>>(l);
 
     // Now that the queue is initialized, iterate over m elements.
     std::vector<std::pair<double, size_t>> v(k, std::make_pair(-1.0,
diff -pruN 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp
--- 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -95,13 +95,18 @@ namespace mlpack {
  * estimator.Predict(xTest, responses, stds)
  * @endcode
  */
+template<typename ModelMatType = arma::mat>
 class BayesianLinearRegression
 {
  public:
+  typedef typename ModelMatType::elem_type ElemType;
+  typedef typename GetDenseColType<ModelMatType>::type DenseVecType;
+  typedef typename GetDenseRowType<ModelMatType>::type DenseRowType;
+
   /**
-   * Set the parameters of Bayesian Ridge regression object. The
-   * regularization parameter is automatically set to its optimal value by
-   * maximization of the marginal likelihood.
+   * Set the parameters of Bayesian Ridge regression object. The regularization
+   * parameter will be automatically set to its optimal value by maximization of
+   * the marginal likelihood when training is done.
    *
    * @param centerData Whether or not center the data according to the
    *    examples.
@@ -117,16 +122,105 @@ class BayesianLinearRegression
                            const double tolerance = 1e-4);
 
   /**
+   * Create the BayesianLinearRegression object and train the model.  The
+   * regularization parameter is automatically set to its optimal value by
+   * maximization of the maginal likelihood.
+   *
+   * @param data Column-major input data, dim(P, N).
+   * @param responses A vector of targets, dim(N).
+   * @param centerData Whether or not center the data according to the
+   *    examples.
+   * @param scaleData Whether or not scale the data according to the
+   *    standard deviation of each feature.
+   * @param maxIterations Maximum number of iterations for convergency.
+   * @param tolerance Level from which the solution is considered sufficientlly
+   *    stable.
+   * @return Root mean squared error.
+   */
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  BayesianLinearRegression(const MatType& data,
+                           const ResponsesType& responses,
+                           const bool centerData = true,
+                           const bool scaleData = false,
+                           const size_t maxIterations = 50,
+                           const double tolerance = 1e-4);
+
+  /**
    * Run BayesianLinearRegression. The input matrix (like all mlpack matrices)
    * should be column-major -- each column is an observation and each row is a
    * dimension.
    *
    * @param data Column-major input data, dim(P, N).
    * @param responses A vector of targets, dim(N).
+   * @param centerData Whether or not center the data according to the
+   *    examples.
+   * @param scaleData Whether or not scale the data according to the
+   *    standard deviation of each feature.
+   * @param maxIterations Maximum number of iterations for convergency.
+   * @param tolerance Level from which the solution is considered sufficientlly
+   *    stable.
    * @return Root mean squared error.
    */
-  double Train(const arma::mat& data,
-               const arma::rowvec& responses);
+  // Many overloads necessary here until std::optional is available with C++17.
+  // The first overload is also necessary to avoid confusing the hyperparameter
+  // tuner, so that this can be correctly detected as a regression algorithm.
+  template<typename MatType>
+  ElemType Train(const MatType& data,
+                 const arma::rowvec& responses);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const std::optional<bool> centerData = std::nullopt,
+                 const std::optional<bool> scaleData = std::nullopt,
+                 const std::optional<size_t> maxIterations = std::nullopt);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool centerData,
+                 const bool scaleData,
+                 const size_t maxIterations,
+                 const double tolerance);
+
+  /**
+   * Predict \f$y\f$ for a single data point \f$x\f$ using the currently-trained
+   * Bayesian ridge regression model.
+   *
+   * @param point The data point to apply the model to.
+   * @return Prediction for the `point`.
+   */
+  template<typename VecType>
+  ElemType Predict(const VecType& point) const;
+
+  /**
+   * Predict \f$y\f$ for a single data point \f$x\f$ using the currently-trained
+   * Bayesian ridge regression model, storing the prediction in `prediction` and
+   * the standard deviation of the prediction in `stddev`.
+   *
+   * @param point The data point to apply the model to.
+   * @param prediction `double` to store the prediction into.
+   * @param stddev `double` to store the standard deviation of the prediction
+   * into.
+   */
+  template<typename VecType>
+  void Predict(const VecType& point,
+               ElemType& prediction,
+               ElemType& stddev) const;
 
   /**
    * Predict \f$y_{i}\f$ for each data point in the given data matrix using the
@@ -134,10 +228,14 @@ class BayesianLinearRegression
    *
    * @param points The data points to apply the model.
    * @param predictions y, Contains the  predicted values on completion.
-   * @return Root mean squared error computed on the train set.
    */
-  void Predict(const arma::mat& points,
-               arma::rowvec& predictions) const;
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  void Predict(const MatType& points,
+               ResponsesType& predictions) const;
 
   /**
    * Predict \f$y_{i}\f$ and the standard deviation of the predictive posterior
@@ -149,9 +247,14 @@ class BayesianLinearRegression
    *     completion.
    * @param std Standard deviations of the predictions.
    */
-  void Predict(const arma::mat& points,
-               arma::rowvec& predictions,
-               arma::rowvec& std) const;
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  void Predict(const MatType& points,
+               ResponsesType& predictions,
+               ResponsesType& std) const;
 
   /**
    * Compute the Root Mean Square Error between the predictions returned by the
@@ -161,8 +264,13 @@ class BayesianLinearRegression
    * @param responses A vector of targets.
    * @return Root mean squared error.
    **/
-  double RMSE(const arma::mat& data,
-              const arma::rowvec& responses) const;
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType RMSE(const MatType& data,
+                const ResponsesType& responses) const;
 
   /**
    * Get the solution vector.
@@ -199,7 +307,7 @@ class BayesianLinearRegression
    *
    * @return responsesOffset
    */
-  const arma::colvec& DataOffset() const { return dataOffset; }
+  const DenseVecType& DataOffset() const { return dataOffset; }
 
   /**
    * Get the vector of standard deviations computed on the features over the
@@ -207,14 +315,14 @@ class BayesianLinearRegression
    *
    * @return dataOffset
    */
-  const arma::colvec& DataScale() const { return dataScale; }
+  const DenseVecType& DataScale() const { return dataScale; }
 
   /**
    * Get the mean value of the train responses.
    *
    * @return responsesOffset
    */
-  double ResponsesOffset() const { return responsesOffset; }
+  ElemType ResponsesOffset() const { return responsesOffset; }
 
   //! Get whether the data will be centered during training.
   bool CenterData() const { return centerData; }
@@ -258,28 +366,28 @@ class BayesianLinearRegression
   double tolerance;
 
   //! Mean vector computed over the points.
-  arma::colvec dataOffset;
+  DenseVecType dataOffset;
 
   //! Std vector computed over the points.
-  arma::colvec dataScale;
+  DenseVecType dataScale;
 
   //! Mean of the response vector computed over the points.
-  double responsesOffset;
+  ElemType responsesOffset;
 
   //! Precision of the prior pdf (gaussian).
-  double alpha;
+  ElemType alpha;
 
   //! Noise inverse variance.
-  double beta;
+  ElemType beta;
 
   //! Effective number of parameters.
-  double gamma;
+  ElemType gamma;
 
   //! Solution vector.
-  arma::colvec omega;
+  DenseVecType omega;
 
   //! Covariance matrix of the solution vector omega.
-  arma::mat matCovariance;
+  ModelMatType matCovariance;
 
   /**
    * Center and scale the data accordind to centerData and scaleData.
@@ -291,23 +399,30 @@ class BayesianLinearRegression
    * @param responsesProc Responses processed, dim(N).
    * @return reponsesOffset Mean of responses.
    */
-  double CenterScaleData(const arma::mat& data,
-                         const arma::rowvec& responses,
-                         arma::mat& dataProc,
-                         arma::rowvec& responsesProc);
+  template<typename MatType, typename ResponsesType>
+  double CenterScaleData(const MatType& data,
+                         const ResponsesType& responses,
+                         MatType& dataProc,
+                         ResponsesType& responsesProc);
 
   /**
-   * Center and scale the points before prediction.
+   * Center and scale the points before prediction.  This should only be called
+   * if centerData or scaleData is true; if neither is true, then `dataProc`
+   * will be unmodified.
    *
    * @param data Design matrix in column-major format, dim(P, N).
    * @param dataProc Data processed, dim(P, N).
    */
-  void CenterScaleDataPred(const arma::mat& data,
-                           arma::mat& dataProc) const;
+  template<typename MatType, typename OutMatType>
+  void CenterScaleDataPred(const MatType& data,
+                           OutMatType& dataProc) const;
 };
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename ModelMatType),
+    (mlpack::BayesianLinearRegression<ModelMatType>), (1));
+
 // Include implementation of serialize.
 #include "bayesian_linear_regression_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_impl.hpp 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_impl.hpp
--- 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,7 +16,8 @@
 
 namespace mlpack {
 
-inline BayesianLinearRegression::BayesianLinearRegression(
+template<typename ModelMatType>
+inline BayesianLinearRegression<ModelMatType>::BayesianLinearRegression(
     const bool centerData,
     const bool scaleData,
     const size_t maxIterations,
@@ -29,15 +30,81 @@ inline BayesianLinearRegression::Bayesia
     alpha(0.0),
     beta(0.0),
     gamma(0.0)
-{/* Nothing to do */}
+{ /* Nothing to do */ }
 
-inline double BayesianLinearRegression::Train(const arma::mat& data,
-                                              const arma::rowvec& responses)
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline BayesianLinearRegression<ModelMatType>::BayesianLinearRegression(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool centerData,
+    const bool scaleData,
+    const size_t maxIterations,
+    const double tolerance) :
+    centerData(centerData),
+    scaleData(scaleData),
+    maxIterations(maxIterations),
+    tolerance(tolerance),
+    responsesOffset(0.0),
+    alpha(0.0),
+    beta(0.0),
+    gamma(0.0)
+{
+  // Train the model.
+  Train(data, responses);
+}
+
+template<typename ModelMatType>
+template<typename MatType>
+inline
+typename BayesianLinearRegression<ModelMatType>::ElemType
+BayesianLinearRegression<ModelMatType>::Train(
+    const MatType& data,
+    const arma::rowvec& responses)
+{
+  return Train(data, responses, this->centerData, this->scaleData,
+      this->maxIterations, this->tolerance);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline
+typename BayesianLinearRegression<ModelMatType>::ElemType
+BayesianLinearRegression<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const std::optional<bool> centerData,
+    const std::optional<bool> scaleData,
+    const std::optional<size_t> maxIterations)
+{
+  return Train(data, responses,
+       (centerData.has_value()) ? centerData.value() :  this->centerData,
+       (scaleData.has_value()) ?   scaleData.value() :  this->scaleData,
+       (maxIterations.has_value()) ? maxIterations.value() : this->maxIterations,
+       this->tolerance);
+ }
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline
+typename BayesianLinearRegression<ModelMatType>::ElemType
+BayesianLinearRegression<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool centerData,
+    const bool scaleData,
+    const size_t maxIterations,
+    const double tolerance)
 {
-  arma::mat phi;
-  arma::rowvec t;
-  arma::colvec eigVal;
-  arma::mat eigVec;
+  this->centerData = centerData;
+  this->scaleData = scaleData;
+  this->maxIterations = maxIterations;
+  this->tolerance = tolerance;
+
+  ModelMatType phi;
+  DenseRowType t;
+  DenseVecType eigVal;
+  ModelMatType eigVec;
 
   // Preprocess the data. Center and scale.
   responsesOffset = CenterScaleData(data, responses, phi, t);
@@ -49,20 +116,20 @@ inline double BayesianLinearRegression::
   }
 
   // Compute this quantities once and for all.
-  const arma::mat eigVecInv = inv(eigVec);
-  const arma::colvec eigVecInvPhitT = eigVecInv * phi * t.t();
+  const ModelMatType eigVecInv = inv(eigVec);
+  const DenseVecType eigVecInvPhitT = eigVecInv * phi * t.t();
 
   // Initialize the hyperparameters and begin with an infinitely broad prior.
-  alpha = 1e-6;
-  beta =  1 / (var(t, 1) * 0.1);
+  alpha = ((ElemType) 1e-6);
+  beta =  ((ElemType) 1 / (var(t, 1) * 0.1));
 
   unsigned short i = 0;
-  double crit = 1.0;
+  ElemType crit = ((ElemType) 1.0);
 
-  while ((crit > tolerance) && (i < maxIterations))
+  while (((double) crit > tolerance) && (i < maxIterations))
   {
-    double deltaAlpha = -alpha;
-    double deltaBeta = -beta;
+    ElemType deltaAlpha = -alpha;
+    ElemType deltaBeta = -beta;
 
     // Update the solution.
     omega = eigVec * diagmat(1 / (eigVal + (alpha / beta))) * eigVecInvPhitT;
@@ -72,7 +139,7 @@ inline double BayesianLinearRegression::
     alpha = gamma / dot(omega, omega);
 
     // Update beta.
-    const arma::rowvec temp = t - omega.t() * phi;
+    const DenseRowType temp = t - omega.t() * phi;
     beta = (data.n_cols - gamma) / dot(temp, temp);
 
     // Compute the stopping criterion.
@@ -82,56 +149,136 @@ inline double BayesianLinearRegression::
     i++;
   }
   // Compute the covariance matrix for the uncertainties later.
-  matCovariance = eigVec * diagmat(1 / (beta * eigVal + alpha)) * eigVecInv;
+  matCovariance = eigVec * diagmat(((ElemType) 1) / (beta * eigVal + alpha)) *
+      eigVecInv;
 
   return RMSE(data, responses);
 }
 
-inline void BayesianLinearRegression::Predict(const arma::mat& points,
-                                              arma::rowvec& predictions) const
+template<typename ModelMatType>
+template<typename VecType>
+inline
+typename BayesianLinearRegression<ModelMatType>::ElemType
+BayesianLinearRegression<ModelMatType>::Predict(const VecType& point) const
+{
+  // Center and scale the point before applying the model, if needed.
+  if (!centerData && !scaleData)
+    return dot(omega, point) + responsesOffset;
+  else if (centerData && !scaleData)
+    return dot(omega, point - dataOffset) + responsesOffset;
+  else if (!centerData && scaleData)
+    return dot(omega, point / dataScale) + responsesOffset;
+  else
+    return dot(omega, (point - dataOffset) / dataScale) + responsesOffset;
+}
+
+template<typename ModelMatType>
+template<typename VecType>
+inline void BayesianLinearRegression<ModelMatType>::Predict(
+    const VecType& point,
+    typename BayesianLinearRegression<ModelMatType>::ElemType& prediction,
+    typename BayesianLinearRegression<ModelMatType>::ElemType& stddev) const
+{
+  prediction = Predict(point);
+
+  ElemType inner;
+  if (!centerData && !scaleData)
+  {
+    stddev = std::sqrt(Variance() +
+        accu(point % (matCovariance * point)));
+    inner = accu(point % (matCovariance * point));
+  }
+  else if (centerData && !scaleData)
+  {
+    inner = accu((point - dataOffset) %
+        (matCovariance * (point - dataOffset)));
+  }
+  else if (!centerData && scaleData)
+  {
+    inner = accu((point / dataScale) %
+        (matCovariance * (point / dataScale)));
+  }
+  else
+  {
+    inner = accu(((point - dataOffset) / dataScale) %
+        (matCovariance * ((point - dataOffset) / dataScale)));
+  }
+
+  stddev = std::sqrt(Variance() + inner);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline void BayesianLinearRegression<ModelMatType>::Predict(
+    const MatType& points,
+    ResponsesType& predictions) const
 {
-  // Center and scale the points before applying the model.
-  arma::mat matX;
-  CenterScaleDataPred(points, matX);
-  predictions = omega.t() * matX + responsesOffset;
+  if (!centerData && !scaleData)
+  {
+    predictions = omega.t() * points + responsesOffset;
+  }
+  else
+  {
+    // Center and scale the points before applying the model.
+    arma::Mat<ElemType> pointsProc;
+    CenterScaleDataPred(points, pointsProc);
+
+    predictions = omega.t() * pointsProc + responsesOffset;
+  }
 }
 
-inline void BayesianLinearRegression::Predict(const arma::mat& points,
-                                              arma::rowvec& predictions,
-                                              arma::rowvec& std) const
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline void BayesianLinearRegression<ModelMatType>::Predict(
+    const MatType& points,
+    ResponsesType& predictions,
+    ResponsesType& std) const
 {
-  // Center and scale the points before applying the model.
-  arma::mat matX;
-  CenterScaleDataPred(points, matX);
-  predictions = omega.t() * matX + responsesOffset;
-  // Compute the standard deviation for each point.
-  std = sqrt(Variance() + sum(matX % (matCovariance * matX), 0));
+  if (!centerData && !scaleData)
+  {
+    Predict(points, predictions);
+    std = sqrt(Variance() + sum(points % (matCovariance * points), 0));
+  }
+  else
+  {
+    // Center or scale data.
+    arma::Mat<ElemType> pointsProc;
+    CenterScaleDataPred(points, pointsProc);
+
+    predictions = omega.t() * pointsProc + responsesOffset;
+    std = sqrt(Variance() + sum(pointsProc % (matCovariance * pointsProc), 0));
+  }
 }
 
-inline double BayesianLinearRegression::RMSE(
-    const arma::mat& data,
-    const arma::rowvec& responses) const
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline
+typename BayesianLinearRegression<ModelMatType>::ElemType
+BayesianLinearRegression<ModelMatType>::RMSE(
+    const MatType& data,
+    const ResponsesType& responses) const
 {
-  arma::rowvec predictions;
+  typename GetDenseRowType<ResponsesType>::type predictions;
   Predict(data, predictions);
-  return sqrt(mean(square(responses - predictions)));
+  return std::sqrt(mean(square(responses - predictions)));
 }
 
-inline double BayesianLinearRegression::CenterScaleData(
-    const arma::mat& data,
-    const arma::rowvec& responses,
-    arma::mat& dataProc,
-    arma::rowvec& responsesProc)
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType>
+inline double BayesianLinearRegression<ModelMatType>::CenterScaleData(
+    const MatType& data,
+    const ResponsesType& responses,
+    MatType& dataProc,
+    ResponsesType& responsesProc)
 {
   if (!centerData && !scaleData)
   {
-    dataProc = arma::mat(const_cast<double*>(data.memptr()), data.n_rows,
+    dataProc = MatType(const_cast<ElemType*>(data.memptr()), data.n_rows,
                                              data.n_cols, false, true);
-    responsesProc = arma::rowvec(const_cast<double*>(responses.memptr()),
-                                                     responses.n_elem, false,
-                                                     true);
+    responsesProc = ResponsesType(const_cast<ElemType*>(responses.memptr()),
+                                                        responses.n_elem, false,
+                                                        true);
   }
-
   else if (centerData && !scaleData)
   {
     dataOffset = mean(data, 1);
@@ -139,16 +286,14 @@ inline double BayesianLinearRegression::
     dataProc = data.each_col() - dataOffset;
     responsesProc = responses - responsesOffset;
   }
-
   else if (!centerData && scaleData)
   {
     dataScale = stddev(data, 0, 1);
     dataProc = data.each_col() / dataScale;
-    responsesProc = arma::rowvec(const_cast<double*>(responses.memptr()),
-                                                     responses.n_elem, false,
-                                                     true);
+    responsesProc = ResponsesType(const_cast<ElemType*>(responses.memptr()),
+                                                        responses.n_elem, false,
+                                                        true);
   }
-
   else
   {
     dataOffset = mean(data, 1);
@@ -157,29 +302,28 @@ inline double BayesianLinearRegression::
     dataProc = (data.each_col() - dataOffset).each_col() / dataScale;
     responsesProc = responses - responsesOffset;
   }
+
   return responsesOffset;
 }
 
-inline void BayesianLinearRegression::CenterScaleDataPred(
-    const arma::mat& data,
-    arma::mat& dataProc) const
+template<typename ModelMatType>
+template<typename MatType, typename OutMatType>
+inline void BayesianLinearRegression<ModelMatType>::CenterScaleDataPred(
+    const MatType& data,
+    OutMatType& dataProc) const
 {
   if (!centerData && !scaleData)
   {
-    dataProc = arma::mat(const_cast<double*>(data.memptr()), data.n_rows,
-                         data.n_cols, false, true);
+    return; // Don't modify dataProc.
   }
-
   else if (centerData && !scaleData)
   {
     dataProc = data.each_col() - dataOffset;
   }
-
   else if (!centerData && scaleData)
   {
     dataProc = data.each_col() / dataScale;
   }
-
   else
   {
     dataProc = (data.each_col() - dataOffset).each_col() / dataScale;
@@ -189,22 +333,59 @@ inline void BayesianLinearRegression::Ce
 /**
  * Serialize the Bayesian linear regression model.
  */
+template<typename ModelMatType>
 template<typename Archive>
-void BayesianLinearRegression::serialize(Archive& ar,
-                                         const uint32_t /* version */)
+void BayesianLinearRegression<ModelMatType>::serialize(Archive& ar,
+                                                       const uint32_t version)
 {
   ar(CEREAL_NVP(centerData));
   ar(CEREAL_NVP(scaleData));
   ar(CEREAL_NVP(maxIterations));
   ar(CEREAL_NVP(tolerance));
-  ar(CEREAL_NVP(dataOffset));
-  ar(CEREAL_NVP(dataScale));
-  ar(CEREAL_NVP(responsesOffset));
-  ar(CEREAL_NVP(alpha));
-  ar(CEREAL_NVP(beta));
-  ar(CEREAL_NVP(gamma));
-  ar(CEREAL_NVP(omega));
-  ar(CEREAL_NVP(matCovariance));
+
+  // In older versions, dataOffset and dataScale were of type arma::colvec,
+  // responsesOffset, alpha, beta, and gamma were of type double, omega was of
+  // type arma::colvec, and matCovariance was of type arma::mat.
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    arma::colvec colvecTmp;
+    ar(cereal::make_nvp("dataOffset", colvecTmp));
+    dataOffset = ConvTo<DenseVecType>::From(colvecTmp);
+
+    ar(cereal::make_nvp("dataScale", colvecTmp));
+    dataScale = ConvTo<DenseVecType>::From(colvecTmp);
+
+    double dblTmp;
+    ar(cereal::make_nvp("responsesOffset", dblTmp));
+    responsesOffset = (ElemType) dblTmp;
+
+    ar(cereal::make_nvp("alpha", dblTmp));
+    alpha = (ElemType) dblTmp;
+
+    ar(cereal::make_nvp("beta", dblTmp));
+    beta = (ElemType) dblTmp;
+
+    ar(cereal::make_nvp("gamma", dblTmp));
+    gamma = (ElemType) dblTmp;
+
+    ar(cereal::make_nvp("omega", colvecTmp));
+    omega = ConvTo<DenseVecType>::From(colvecTmp);
+
+    arma::mat matTmp;
+    ar(cereal::make_nvp("matCovariance", matTmp));
+    matCovariance = ConvTo<ModelMatType>::From(matCovariance);
+  }
+  else
+  {
+    ar(CEREAL_NVP(dataOffset));
+    ar(CEREAL_NVP(dataScale));
+    ar(CEREAL_NVP(responsesOffset));
+    ar(CEREAL_NVP(alpha));
+    ar(CEREAL_NVP(beta));
+    ar(CEREAL_NVP(gamma));
+    ar(CEREAL_NVP(omega));
+    ar(CEREAL_NVP(matCovariance));
+  }
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_main.cpp 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_main.cpp
--- 4.3.0-2/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bayesian_linear_regression/bayesian_linear_regression_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -97,10 +97,11 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("Bayesian Interpolation",
-    "https://authors.library.caltech.edu/13792/1/MACnc92a.pdf");
-BINDING_SEE_ALSO("Bayesian Linear Regression, Section 3.3", "MLA Bishop, "
-    "Christopher M. Pattern Recognition and Machine Learning. New York: "
-    "Springer, 2006, section 3.3.");
+    "https://cs.uwaterloo.ca/~mannr/cs886-w10/mackay-bayesian.pdf");
+BINDING_SEE_ALSO("Bayesian Linear Regression, Section 3.3",
+    // I wonder how long this full text PDF will remain available...
+    "https://www.microsoft.com/en-us/research/uploads/prod/2006/01/"
+    "Bishop-Pattern-Recognition-and-Machine-Learning-2006.pdf");
 BINDING_SEE_ALSO("BayesianLinearRegression C++ class documentation",
     "@src/mlpack/methods/bayesian_linear_regression/"
     "bayesian_linear_regression.hpp");
@@ -109,10 +110,10 @@ PARAM_MATRIX_IN("input", "Matrix of cova
 
 PARAM_ROW_IN("responses", "Matrix of responses/observations (y).", "r");
 
-PARAM_MODEL_IN(BayesianLinearRegression, "input_model", "Trained "
+PARAM_MODEL_IN(BayesianLinearRegression<>, "input_model", "Trained "
                "BayesianLinearRegression model to use.", "m");
 
-PARAM_MODEL_OUT(BayesianLinearRegression, "output_model", "Output "
+PARAM_MODEL_OUT(BayesianLinearRegression<>, "output_model", "Output "
                 "BayesianLinearRegression model.", "M");
 
 PARAM_MATRIX_IN("test", "Matrix containing points to regress on (test "
@@ -149,12 +150,12 @@ void BINDING_FUNCTION(util::Params& para
   // Ignore out_predictions unless test is specified.
   ReportIgnoredParam(params, {{"test", false}}, "predictions");
 
-  BayesianLinearRegression* bayesLinReg;
+  BayesianLinearRegression<>* bayesLinReg;
   if (params.Has("input"))
   {
     Log::Info << "Input given; model will be trained." << std::endl;
     // Initialize the object.
-    bayesLinReg = new BayesianLinearRegression(center, scale);
+    bayesLinReg = new BayesianLinearRegression<>(center, scale);
 
     // Load covariates.
     mat matX = std::move(params.Get<arma::mat>("input"));
@@ -180,7 +181,7 @@ void BINDING_FUNCTION(util::Params& para
   }
   else // We must have --input_model_file.
   {
-    bayesLinReg = params.Get<BayesianLinearRegression*>("input_model");
+    bayesLinReg = params.Get<BayesianLinearRegression<>*>("input_model");
   }
 
   if (params.Has("test"))
@@ -209,5 +210,5 @@ void BINDING_FUNCTION(util::Params& para
     params.Get<arma::mat>("predictions") = std::move(predictions);
   }
 
-  params.Get<BayesianLinearRegression*>("output_model") = bayesLinReg;
+  params.Get<BayesianLinearRegression<>*>("output_model") = bayesLinReg;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd.hpp 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd.hpp
--- 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -50,7 +50,9 @@ namespace mlpack {
  * @endcode
  *
  */
-template<typename OptimizerType = ens::StandardSGD>
+template<typename OptimizerType = ens::StandardSGD,
+         typename MatType = arma::mat,
+         typename VecType = arma::vec>
 class BiasSVD
 {
  public:
@@ -76,12 +78,12 @@ class BiasSVD
    * @param p Item bias.
    * @param q User bias.
    */
-  void Apply(const arma::mat& data,
+  void Apply(const MatType& data,
              const size_t rank,
-             arma::mat& u,
-             arma::mat& v,
-             arma::vec& p,
-             arma::vec& q);
+             MatType& u,
+             MatType& v,
+             VecType& p,
+             VecType& q);
 
  private:
   //! Number of optimization iterations.
diff -pruN 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_function.hpp 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_function.hpp
--- 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -15,6 +15,7 @@
 #define MLPACK_METHODS_BIAS_SVD_BIAS_SVD_FUNCTION_HPP
 
 #include <mlpack/prereqs.hpp>
+#include <mlpack/core/math/make_alias.hpp>
 #include <ensmallen.hpp>
 
 namespace mlpack {
@@ -54,7 +55,7 @@ class BiasSVDFunction
    * @param parameters Parameters(user/item matrices/bias) of the
    *     decomposition.
    */
-  double Evaluate(const arma::mat& parameters) const;
+  double Evaluate(const MatType& parameters) const;
 
   /**
    * Evaluates the cost function for one training example. Useful for the SGD
@@ -65,7 +66,7 @@ class BiasSVDFunction
    * @param start First index of the training examples to be used.
    * @param batchSize Size of batch to evaluate.
    */
-  double Evaluate(const arma::mat& parameters,
+  double Evaluate(const MatType& parameters,
                   const size_t start,
                   const size_t batchSize = 1) const;
 
@@ -77,8 +78,8 @@ class BiasSVDFunction
    *     decomposition.
    * @param gradient Calculated gradient for the parameters.
    */
-  void Gradient(const arma::mat& parameters,
-                arma::mat& gradient) const;
+  void Gradient(const MatType& parameters,
+                MatType& gradient) const;
 
   /**
    * Evaluates the gradient of the cost function over one training example.
@@ -94,16 +95,16 @@ class BiasSVDFunction
    * @param batchSize Size of batch to calculate gradient for.
    */
   template <typename GradType>
-  void Gradient(const arma::mat& parameters,
+  void Gradient(const MatType& parameters,
                 const size_t start,
                 GradType& gradient,
                 const size_t batchSize = 1) const;
 
   //! Return the initial point for the optimization.
-  const arma::mat& GetInitialPoint() const { return initialPoint; }
+  const MatType& GetInitialPoint() const { return initialPoint; }
 
   //! Return the dataset passed into the constructor.
-  const arma::mat& Dataset() const { return data; }
+  const MatType& Dataset() const { return data; }
 
   //! Return the number of training examples. Useful for SGD optimizer.
   size_t NumFunctions() const { return data.n_cols; }
@@ -124,7 +125,7 @@ class BiasSVDFunction
   //! Rating data.  This will be an alias until Shuffle() is called.
   MatType data;
   //! Initial parameter point.
-  arma::mat initialPoint;
+  MatType initialPoint;
   //! Rank used for matrix factorization.
   size_t rank;
   //! Regularization parameter for the optimization.
diff -pruN 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_function_impl.hpp 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -14,18 +14,18 @@
 #define MLPACK_METHODS_BIAS_SVD_BIAS_SVD_FUNCTION_IMPL_HPP
 
 #include "bias_svd_function.hpp"
-#include <mlpack/core/math/make_alias.hpp>
 
 namespace mlpack {
 
 template <typename MatType>
-BiasSVDFunction<MatType>::BiasSVDFunction(const MatType& data,
+BiasSVDFunction<MatType>::BiasSVDFunction(const MatType& dataIn,
                                           const size_t rank,
                                           const double lambda) :
-    data(MakeAlias(const_cast<MatType&>(data), false)),
     rank(rank),
     lambda(lambda)
 {
+  MakeAlias(data, dataIn, dataIn.n_rows, dataIn.n_cols, false);
+
   // Number of users and items in the data.
   numUsers = max(data.row(0)) + 1;
   numItems = max(data.row(1)) + 1;
@@ -43,13 +43,13 @@ void BiasSVDFunction<MatType>::Shuffle()
 }
 
 template <typename MatType>
-double BiasSVDFunction<MatType>::Evaluate(const arma::mat& parameters) const
+double BiasSVDFunction<MatType>::Evaluate(const MatType& parameters) const
 {
   return Evaluate(parameters, 0, data.n_cols);
 }
 
 template <typename MatType>
-double BiasSVDFunction<MatType>::Evaluate(const arma::mat& parameters,
+double BiasSVDFunction<MatType>::Evaluate(const MatType& parameters,
                                           const size_t start,
                                           const size_t batchSize) const
 {
@@ -74,13 +74,13 @@ double BiasSVDFunction<MatType>::Evaluat
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     double ratingError = rating - userBias - itemBias -
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
     double ratingErrorSquared = ratingError * ratingError;
 
     // Calculate the regularization penalty corresponding to the parameters.
-    double userVecNorm = arma::norm(parameters.col(user), 2);
-    double itemVecNorm = arma::norm(parameters.col(item), 2);
+    double userVecNorm = norm(parameters.col(user), 2);
+    double itemVecNorm = norm(parameters.col(item), 2);
     double regularizationError = lambda * (userVecNorm * userVecNorm +
                                            itemVecNorm * itemVecNorm);
 
@@ -91,8 +91,8 @@ double BiasSVDFunction<MatType>::Evaluat
 }
 
 template <typename MatType>
-void BiasSVDFunction<MatType>::Gradient(const arma::mat& parameters,
-                                        arma::mat& gradient) const
+void BiasSVDFunction<MatType>::Gradient(const MatType& parameters,
+                                        MatType& gradient) const
 {
   // For an example with rating corresponding to user 'i' and item 'j', the
   // gradients for the parameters is as follows:
@@ -118,7 +118,7 @@ void BiasSVDFunction<MatType>::Gradient(
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     double ratingError = rating - userBias - itemBias -
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
@@ -138,7 +138,7 @@ void BiasSVDFunction<MatType>::Gradient(
 
 template <typename MatType>
 template <typename GradType>
-void BiasSVDFunction<MatType>::Gradient(const arma::mat& parameters,
+void BiasSVDFunction<MatType>::Gradient(const MatType& parameters,
                                         const size_t start,
                                         GradType& gradient,
                                         const size_t batchSize) const
@@ -156,7 +156,7 @@ void BiasSVDFunction<MatType>::Gradient(
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     double ratingError = rating - userBias - itemBias -
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
@@ -230,7 +230,7 @@ double StandardSGD::Optimize(
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     double ratingError = rating - userBias - itemBias -
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
 
     double lambda = function.Lambda();
@@ -339,7 +339,7 @@ inline double ParallelSGD<ExponentialBac
         const double userBias = iterate(rank, user);
         const double itemBias = iterate(rank, item);
         double ratingError = rating - userBias - itemBias -
-            arma::dot(iterate.col(user).subvec(0, rank - 1),
+            dot(iterate.col(user).subvec(0, rank - 1),
                       iterate.col(item).subvec(0, rank - 1));
 
         arma::mat userVecUpdate = stepSize * 2 * (
diff -pruN 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_impl.hpp 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_impl.hpp
--- 4.3.0-2/src/mlpack/methods/bias_svd/bias_svd_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/bias_svd/bias_svd_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,10 +16,10 @@
 
 namespace mlpack {
 
-template<typename OptimizerType>
-BiasSVD<OptimizerType>::BiasSVD(const size_t iterations,
-                                const double alpha,
-                                const double lambda) :
+template<typename OptimizerType, typename MatType, typename VecType>
+BiasSVD<OptimizerType, MatType, VecType>::BiasSVD(const size_t iterations,
+                                                  const double alpha,
+                                                  const double lambda) :
     iterations(iterations),
     alpha(alpha),
     lambda(lambda)
@@ -27,13 +27,13 @@ BiasSVD<OptimizerType>::BiasSVD(const si
   // Nothing to do.
 }
 
-template<typename OptimizerType>
-void BiasSVD<OptimizerType>::Apply(const arma::mat& data,
-                                   const size_t rank,
-                                   arma::mat& u,
-                                   arma::mat& v,
-                                   arma::vec& p,
-                                   arma::vec& q)
+template<typename OptimizerType, typename MatType, typename VecType>
+void BiasSVD<OptimizerType, MatType, VecType>::Apply(const MatType& data,
+                                                     const size_t rank,
+                                                     MatType& u,
+                                                     MatType& v,
+                                                     VecType& p,
+                                                     VecType& q)
 {
   // batchSize is 1 in our implementation of Bias SVD.
   // batchSize other than 1 has not been supported yet.
@@ -47,7 +47,7 @@ void BiasSVD<OptimizerType>::Apply(const
       iterations * data.n_cols);
 
   // Get optimized parameters.
-  arma::mat parameters = biasSVDFunc.GetInitialPoint();
+  MatType parameters = biasSVDFunc.GetInitialPoint();
   optimizer.Optimize(biasSVDFunc, parameters);
 
   // Constants for extracting user and item matrices.
diff -pruN 4.3.0-2/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd.hpp 4.4.0-1/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd.hpp
--- 4.3.0-2/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -69,10 +69,11 @@ class RandomizedBlockKrylovSVD
    * @param rank Rank of the approximation (Default: number of rows.)
    * @param blockSize The block size, must be >= rank (Default: rank + 10).
    */
-  RandomizedBlockKrylovSVD(const arma::mat& data,
-                           arma::mat& u,
-                           arma::vec& s,
-                           arma::mat& v,
+  template<typename InMatType, typename MatType, typename VecType>
+  RandomizedBlockKrylovSVD(const InMatType& data,
+                           MatType& u,
+                           VecType& s,
+                           MatType& v,
                            const size_t maxIterations = 2,
                            const size_t rank = 0,
                            const size_t blockSize = 0);
@@ -97,10 +98,11 @@ class RandomizedBlockKrylovSVD
    * @param s Diagonal matrix of singular values.
    * @param rank Rank of the approximation.
    */
-  void Apply(const arma::mat& data,
-             arma::mat& u,
-             arma::vec& s,
-             arma::mat& v,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& data,
+             MatType& u,
+             VecType& s,
+             MatType& v,
              const size_t rank);
 
   //! Get the number of iterations for the power method.
diff -pruN 4.3.0-2/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd_impl.hpp 4.4.0-1/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd_impl.hpp
--- 4.3.0-2/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/block_krylov_svd/randomized_block_krylov_svd_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,11 +16,12 @@
 
 namespace mlpack {
 
+template<typename InMatType, typename MatType, typename VecType>
 inline RandomizedBlockKrylovSVD::RandomizedBlockKrylovSVD(
-    const arma::mat& data,
-    arma::mat& u,
-    arma::vec& s,
-    arma::mat& v,
+    const InMatType& data,
+    MatType& u,
+    VecType& s,
+    MatType& v,
     const size_t maxIterations,
     const size_t rank,
     const size_t blockSize) :
@@ -46,42 +47,46 @@ inline RandomizedBlockKrylovSVD::Randomi
   /* Nothing to do here */
 }
 
-inline void RandomizedBlockKrylovSVD::Apply(const arma::mat& data,
-                                            arma::mat& u,
-                                            arma::vec& s,
-                                            arma::mat& v,
+template<typename InMatType, typename MatType, typename VecType>
+inline void RandomizedBlockKrylovSVD::Apply(const InMatType& data,
+                                            MatType& u,
+                                            VecType& s,
+                                            MatType& v,
                                             const size_t rank)
 {
-  arma::mat Q, R, block, blockIteration;
+  MatType Q, R, block, blockIteration;
 
   if (blockSize == 0)
   {
-    blockSize = rank + 10;
+    // The block size cannot be greater than the number of points in the
+    // dataset or the dimensionality of the dataset.
+    blockSize = std::min((size_t) data.n_rows, std::min((size_t) data.n_cols,
+        rank + 10));
   }
 
   // Random block initialization.
-  arma::mat G = arma::randn(data.n_cols, blockSize);
+  MatType G = arma::randn<MatType>(data.n_cols, blockSize);
 
   // Construct and orthonormalize Krylov subspace.
-  arma::mat K(data.n_rows, blockSize * (maxIterations + 1));
+  MatType K(data.n_rows, blockSize * (maxIterations + 1));
 
   // Create a working matrix using data from writable auxiliary memory
-  // (K matrix). Doing so avoids an uncessary copy in upcoming step.
-  block = arma::mat(K.memptr(), data.n_rows, blockSize, false, false);
+  // (K matrix). Doing so avoids an unnecessary copy in upcoming step.
+  MakeAlias(block, K, data.n_rows, blockSize, false);
   arma::qr_econ(block, R, data * G);
 
   for (size_t blockOffset = block.n_elem; blockOffset < K.n_elem;
       blockOffset += block.n_elem)
   {
     // Temporary working matrix to store the result in the correct place.
-    blockIteration = arma::mat(K.memptr() + blockOffset, block.n_rows,
-        block.n_cols, false, false);
+    MakeAlias(blockIteration, K, block.n_rows, block.n_cols, blockOffset,
+        false);
 
     arma::qr_econ(blockIteration, R, data * (data.t() * block));
 
     // Update working matrix for the next iteration.
-    block = arma::mat(K.memptr() + blockOffset, block.n_rows, block.n_cols,
-        false, false);
+    MakeAlias(block, K, block.n_rows, block.n_cols, blockOffset,
+        false);
   }
 
   arma::qr_econ(Q, R, K);
diff -pruN 4.3.0-2/src/mlpack/methods/cf/cf_main.cpp 4.4.0-1/src/mlpack/methods/cf/cf_main.cpp
--- 4.3.0-2/src/mlpack/methods/cf/cf_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/cf_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,7 +62,7 @@ BINDING_LONG_DESC(
     "\n\n"
     "For performing the matrix decomposition, the following optimization "
     "algorithms can be specified via the " + PRINT_PARAM_STRING("algorithm") +
-    " parameter: "
+    " parameter:\n"
     "\n"
     " - 'RegSVD' -- Regularized SVD using a SGD optimizer\n"
     " - 'NMF' -- Non-negative matrix factorization with alternating least "
@@ -77,21 +77,21 @@ BINDING_LONG_DESC(
     " - 'BKSVD' -- Block Krylov SVD learning\n"
     "\n\n"
     "The following neighbor search algorithms can be specified via" +
-    " the " + PRINT_PARAM_STRING("neighbor_search") + " parameter:"
+    " the " + PRINT_PARAM_STRING("neighbor_search") + " parameter:\n"
     "\n"
     " - 'cosine'  -- Cosine Search Algorithm\n"
     " - 'euclidean'  -- Euclidean Search Algorithm\n"
     " - 'pearson'  -- Pearson Search Algorithm\n"
     "\n\n"
     "The following weight interpolation algorithms can be specified via" +
-    " the " + PRINT_PARAM_STRING("interpolation") + " parameter:"
+    " the " + PRINT_PARAM_STRING("interpolation") + " parameter:\n"
     "\n"
     " - 'average'  -- Average Interpolation Algorithm\n"
     " - 'regression'  -- Regression Interpolation Algorithm\n"
     " - 'similarity'  -- Similarity Interpolation Algorithm\n"
     "\n\n"
     "The following ranking normalization algorithms can be specified via" +
-    " the " + PRINT_PARAM_STRING("normalization") + " parameter:"
+    " the " + PRINT_PARAM_STRING("normalization") + " parameter:\n"
     "\n"
     " - 'none'  -- No Normalization\n"
     " - 'item_mean'  -- Item Mean Normalization\n"
@@ -120,18 +120,14 @@ BINDING_EXAMPLE(
         "recommendations", 5, "output", "recommendations"));
 
 // See also...
-BINDING_SEE_ALSO("Collaborative filtering tutorial",
-        "@doc/tutorials/cf.md");
-BINDING_SEE_ALSO("Alternating Matrix Factorization tutorial",
-        "@doc/tutorials/amf.md");
 BINDING_SEE_ALSO("Collaborative Filtering on Wikipedia",
         "https://en.wikipedia.org/wiki/Collaborative_filtering");
 BINDING_SEE_ALSO("Matrix factorization on Wikipedia",
         "https://en.wikipedia.org/wiki/Matrix_factorization_"
         "(recommender_systems)");
 BINDING_SEE_ALSO("Matrix factorization techniques for recommender systems"
-        " (pdf)", "http://citeseerx.ist.psu.edu/viewdoc/download?doi="
-        "10.1.1.441.3234&rep=rep1&type=pdf");
+        " (pdf)", "https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf"
+        "&doi=cf17f85a0a7991fa01dbfb3e5878fbf71ea4bdc5");
 BINDING_SEE_ALSO("CFType class documentation", "@src/mlpack/methods/cf/cf.hpp");
 
 // Parameters for training a model.
@@ -422,7 +418,7 @@ void BINDING_FUNCTION(util::Params& para
     // of points to get the RMSE.  It turns out this is just the L2-norm divided
     // by the square root of the number of points, if we interpret the
     // predictions and the true values as vectors.
-    const double rmse = arma::norm(predictions - testData.row(2).t(), 2) /
+    const double rmse = norm(predictions - testData.row(2).t(), 2) /
         std::sqrt((double) testData.n_cols);
 
     Log::Info << "RMSE is " << rmse << "." << endl;
diff -pruN 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/block_krylov_svd_method.hpp 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/block_krylov_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/block_krylov_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/block_krylov_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -81,7 +81,7 @@ class BlockKrylovSVDPolicy
     w = w * arma::diagmat(sigma);
 
     // Take transpose of the matrix h as required by CF class.
-    h = arma::trans(h);
+    h = trans(h);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/quic_svd_method.hpp 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/quic_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/quic_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/quic_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -74,14 +74,14 @@ class QUIC_SVDPolicy
     arma::mat data(cleanedData);
 
     // Do singular value decomposition using the quic SVD algorithm.
-    QUIC_SVD quicsvd;
+    QUIC_SVD<> quicsvd;
     quicsvd.Apply(data, w, h, sigma);
 
     // Sigma matrix is multiplied to w.
     w = w * sigma;
 
     // Take transpose of the matrix h as required by CF class.
-    h = arma::trans(h);
+    h = trans(h);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/randomized_svd_method.hpp 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/randomized_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/cf/decomposition_policies/randomized_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/decomposition_policies/randomized_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -86,7 +86,7 @@ class RandomizedSVDPolicy
     w = w * arma::diagmat(sigma);
 
     // Take transpose of the matrix h as required by CF class.
-    h = arma::trans(h);
+    h = trans(h);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/cf/interpolation_policies/regression_interpolation.hpp 4.4.0-1/src/mlpack/methods/cf/interpolation_policies/regression_interpolation.hpp
--- 4.3.0-2/src/mlpack/methods/cf/interpolation_policies/regression_interpolation.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/interpolation_policies/regression_interpolation.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -116,7 +116,7 @@ class RegressionInterpolation
     arma::vec constant(neighborNum);
 
     arma::vec userRating(cleanedData.col(queryUser));
-    const size_t support = arma::accu(userRating != 0);
+    const size_t support = accu(userRating != 0);
 
     // If user has no rating at all, average interpolation is used.
     if (support == 0)
@@ -144,7 +144,7 @@ class RegressionInterpolation
             // Avoid recalculation of iPrediction.
             iPrediction = w * h.col(neighbors(i));
           arma::vec jPrediction = w * h.col(neighbors(j));
-          coeff(i, j) = arma::dot(iPrediction, jPrediction) / itemNum;
+          coeff(i, j) = dot(iPrediction, jPrediction) / itemNum;
           if (coeff(i, j) == 0)
             coeff(i, j) = std::numeric_limits<double>::min();
           coeff(j, i) = coeff(i, j);
@@ -164,7 +164,7 @@ class RegressionInterpolation
         if (iPrediction.size() == 0)
             // Avoid recalculation of iPrediction.
             iPrediction = w * h.col(neighbors(i));
-        constant(i) = arma::dot(iPrediction, userRating) / support;
+        constant(i) = dot(iPrediction, userRating) / support;
         if (constant(i) == 0)
           constant(i) = std::numeric_limits<double>::min();
         // Cache calculated constant term.
diff -pruN 4.3.0-2/src/mlpack/methods/cf/interpolation_policies/similarity_interpolation.hpp 4.4.0-1/src/mlpack/methods/cf/interpolation_policies/similarity_interpolation.hpp
--- 4.3.0-2/src/mlpack/methods/cf/interpolation_policies/similarity_interpolation.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/interpolation_policies/similarity_interpolation.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -83,7 +83,7 @@ class SimilarityInterpolation
           << std::endl;
     }
 
-    double similaritiesSum = arma::sum(similarities);
+    double similaritiesSum = sum(similarities);
     if (std::fabs(similaritiesSum) < 1e-14)
     {
       weights.fill(1.0 / similarities.n_elem);
diff -pruN 4.3.0-2/src/mlpack/methods/cf/neighbor_search_policies/cosine_search.hpp 4.4.0-1/src/mlpack/methods/cf/neighbor_search_policies/cosine_search.hpp
--- 4.3.0-2/src/mlpack/methods/cf/neighbor_search_policies/cosine_search.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/neighbor_search_policies/cosine_search.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,7 +52,7 @@ class CosineSearch
   CosineSearch(const arma::mat& referenceSet)
   {
     // Normalize all vectors to unit length.
-    arma::mat normalizedSet = arma::normalise(referenceSet, 2, 0);
+    arma::mat normalizedSet = normalise(referenceSet, 2, 0);
 
     neighborSearch.Train(std::move(normalizedSet));
   }
@@ -70,7 +70,7 @@ class CosineSearch
               arma::Mat<size_t>& neighbors, arma::mat& similarities)
   {
     // Normalize query vectors to unit length.
-    arma::mat normalizedQuery = arma::normalise(query, 2, 0);
+    arma::mat normalizedQuery = normalise(query, 2, 0);
 
     neighborSearch.Search(normalizedQuery, k, neighbors, similarities);
 
@@ -80,7 +80,7 @@ class CosineSearch
     // Furthermore, we restrict the range of similarity to be [0, 1]:
     // similarities = (cos(a,b) + 1) / 2.0. As a result we have the following
     // formula.
-    similarities = 1 - arma::pow(similarities, 2) / 4.0;
+    similarities = 1 - pow(similarities, 2) / 4.0;
   }
 
  private:
diff -pruN 4.3.0-2/src/mlpack/methods/cf/neighbor_search_policies/pearson_search.hpp 4.4.0-1/src/mlpack/methods/cf/neighbor_search_policies/pearson_search.hpp
--- 4.3.0-2/src/mlpack/methods/cf/neighbor_search_policies/pearson_search.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/neighbor_search_policies/pearson_search.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -57,8 +57,7 @@ class PearsonSearch
     // For each vector x, first subtract mean(x) from each element in x.
     // Then normalize the vector to unit length.
     arma::mat normalizedSet(arma::size(referenceSet));
-    normalizedSet = arma::normalise(
-        referenceSet.each_row() - arma::mean(referenceSet));
+    normalizedSet = normalise(referenceSet.each_row() - arma::mean(referenceSet));
 
     neighborSearch.Train(std::move(normalizedSet));
   }
@@ -79,7 +78,7 @@ class PearsonSearch
     // For each vector x, first subtract mean(x) from each element in x.
     // Then normalize the vector to unit length.
     arma::mat normalizedQuery;
-    normalizedQuery = arma::normalise(query.each_row() - arma::mean(query));
+    normalizedQuery = normalise(query.each_row() - arma::mean(query));
 
     neighborSearch.Search(normalizedQuery, k, neighbors, similarities);
 
@@ -89,7 +88,7 @@ class PearsonSearch
     // Furthermore, we restrict the range of similarity to be [0, 1]:
     // similarities = (pearson(a,b) + 1) / 2.0. As a result we have the
     // following formula.
-    similarities = 1 - arma::pow(similarities, 2) / 4.0;
+    similarities = 1 - pow(similarities, 2) / 4.0;
   }
 
  private:
diff -pruN 4.3.0-2/src/mlpack/methods/cf/normalization/item_mean_normalization.hpp 4.4.0-1/src/mlpack/methods/cf/normalization/item_mean_normalization.hpp
--- 4.3.0-2/src/mlpack/methods/cf/normalization/item_mean_normalization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/normalization/item_mean_normalization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,7 @@ class ItemMeanNormalization
    */
   void Normalize(arma::mat& data)
   {
-    const size_t itemNum = arma::max(data.row(1)) + 1;
+    const size_t itemNum = max(data.row(1)) + 1;
     itemMean = arma::vec(itemNum, arma::fill::zeros);
     // Number of ratings for each item.
     arma::Row<size_t> ratingNum(itemNum, arma::fill::zeros);
diff -pruN 4.3.0-2/src/mlpack/methods/cf/normalization/user_mean_normalization.hpp 4.4.0-1/src/mlpack/methods/cf/normalization/user_mean_normalization.hpp
--- 4.3.0-2/src/mlpack/methods/cf/normalization/user_mean_normalization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/normalization/user_mean_normalization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,7 @@ class UserMeanNormalization
    */
   void Normalize(arma::mat& data)
   {
-    const size_t userNum = arma::max(data.row(0)) + 1;
+    const size_t userNum = max(data.row(0)) + 1;
     userMean = arma::vec(userNum, arma::fill::zeros);
     // Number of ratings for each user.
     arma::Row<size_t> ratingNum(userNum, arma::fill::zeros);
diff -pruN 4.3.0-2/src/mlpack/methods/cf/svd_wrapper.hpp 4.4.0-1/src/mlpack/methods/cf/svd_wrapper.hpp
--- 4.3.0-2/src/mlpack/methods/cf/svd_wrapper.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/svd_wrapper.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,7 +53,7 @@ class SVDWrapper
    * @param sigma eigenvalue matrix
    * @param H second unitary matrix
    *
-   * @note V = W * sigma * arma::trans(H)
+   * @note V = W * sigma * trans(H)
    */
   double Apply(const arma::mat& V,
                arma::mat& W,
diff -pruN 4.3.0-2/src/mlpack/methods/cf/svd_wrapper_impl.hpp 4.4.0-1/src/mlpack/methods/cf/svd_wrapper_impl.hpp
--- 4.3.0-2/src/mlpack/methods/cf/svd_wrapper_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/cf/svd_wrapper_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,10 +34,10 @@ double SVDWrapper<Factorizer>::Apply(con
   for (size_t i = 0; i < sigma.n_rows && i < sigma.n_cols; ++i)
     sigma(i, i) = E(i, 0);
 
-  arma::mat V_rec = W * sigma * arma::trans(H);
+  arma::mat V_rec = W * sigma * trans(H);
 
   // return normalized frobenius error
-  return arma::norm(V - V_rec, "fro") / arma::norm(V, "fro");
+  return norm(V - V_rec, "fro") / norm(V, "fro");
 }
 
 template<>
@@ -56,10 +56,10 @@ double SVDWrapper<DummyClass>::Apply(con
   for (size_t i = 0; i < sigma.n_rows && i < sigma.n_cols; ++i)
     sigma(i, i) = E(i, 0);
 
-  arma::mat V_rec = W * sigma * arma::trans(H);
+  arma::mat V_rec = W * sigma * trans(H);
 
   // return normalized frobenius error
-  return arma::norm(V - V_rec, "fro") / arma::norm(V, "fro");
+  return norm(V - V_rec, "fro") / norm(V, "fro");
 }
 
 template<class Factorizer>
@@ -94,13 +94,13 @@ double SVDWrapper<Factorizer>::Apply(con
   W = W * arma::diagmat(sigma);
 
   // take transpose of the matrix H as required by CF module
-  H = arma::trans(H);
+  H = trans(H);
 
   // reconstruct the matrix
   arma::mat V_rec = W * H;
 
   // return the normalized frobenius norm
-  return arma::norm(V - V_rec, "fro") / arma::norm(V, "fro");
+  return norm(V - V_rec, "fro") / norm(V, "fro");
 }
 
 template<>
@@ -135,13 +135,13 @@ double SVDWrapper<DummyClass>::Apply(con
   W = W * arma::diagmat(sigma);
 
   // take transpose of the matrix H as required by CF module
-  H = arma::trans(H);
+  H = trans(H);
 
   // reconstruct the matrix
   arma::mat V_rec = W * H;
 
   // return the normalized frobenius norm
-  return arma::norm(V - V_rec, "fro") / arma::norm(V, "fro");
+  return norm(V - V_rec, "fro") / norm(V, "fro");
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/dbscan/dbscan.hpp 4.4.0-1/src/mlpack/methods/dbscan/dbscan.hpp
--- 4.3.0-2/src/mlpack/methods/dbscan/dbscan.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/dbscan/dbscan.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,6 +51,10 @@ template<typename RangeSearchType = Rang
 class DBSCAN
 {
  public:
+   //! Easy access to the MatType.
+  typedef typename RangeSearchType::Mat MatType;
+  //! Easy access to Element Type of the matrix.
+  typedef typename RangeSearchType::Mat::elem_type ElemType;
   /**
    * Construct the DBSCAN object with the given parameters.  The batchMode
    * parameter should be set to false in the case where RAM issues will be
@@ -64,7 +68,7 @@ class DBSCAN
    * @param rangeSearch Optional instantiated RangeSearch object.
    * @param pointSelector OptionL instantiated PointSelectionPolicy object.
    */
-  DBSCAN(const double epsilon,
+  DBSCAN(const ElemType epsilon,
          const size_t minPoints,
          const bool batchMode = true,
          RangeSearchType rangeSearch = RangeSearchType(),
@@ -74,24 +78,22 @@ class DBSCAN
    * Performs DBSCAN clustering on the data, returning number of clusters
    * and also the centroid of each cluster.
    *
-   * @tparam MatType Type of matrix (arma::mat or arma::sp_mat).
+   * @tparam MatType Only dense matrix type (e.g., arma::mat or coot::mat).
    * @param data Dataset to cluster.
    * @param centroids Matrix in which centroids are stored.
    */
-  template<typename MatType>
   size_t Cluster(const MatType& data,
-                 arma::mat& centroids);
+                 MatType& centroids);
 
   /**
    * Performs DBSCAN clustering on the data, returning number of clusters
    * and also the list of cluster assignments.  If assignments[i] == SIZE_MAX,
    * then the point is considered "noise".
    *
-   * @tparam MatType Type of matrix (arma::mat or arma::sp_mat).
+   * @tparam MatType Only dense matrix type (e.g., arma::mat or coot::mat).
    * @param data Dataset to cluster.
    * @param assignments Vector to store cluster assignments.
    */
-  template<typename MatType>
   size_t Cluster(const MatType& data,
                  arma::Row<size_t>& assignments);
 
@@ -100,19 +102,18 @@ class DBSCAN
    * the centroid of each cluster and also the list of cluster assignments.
    * If assignments[i] == SIZE_MAX, then the point is considered "noise".
    *
-   * @tparam MatType Type of matrix (arma::mat or arma::sp_mat).
+   * @tparam MatType Only dense matrix type (e.g., arma::mat or coot::mat).
    * @param data Dataset to cluster.
    * @param assignments Vector to store cluster assignments.
    * @param centroids Matrix in which centroids are stored.
    */
-  template<typename MatType>
   size_t Cluster(const MatType& data,
                  arma::Row<size_t>& assignments,
-                 arma::mat& centroids);
+                 MatType& centroids);
 
  private:
   //! Maximum distance between two points to be part of same cluster.
-  double epsilon;
+  ElemType epsilon;
 
   //! Minimum number of points to be in the epsilon-neighborhood (including
   //! itself) for the point to be a core-point.
@@ -137,7 +138,6 @@ class DBSCAN
    * @param assignments Assignments for each point.
    * @param uf UnionFind structure that will be modified.
    */
-  template<typename MatType>
   void PointwiseCluster(const MatType& data, UnionFind& uf);
 
   /**
@@ -149,7 +149,6 @@ class DBSCAN
    * @param assignments Assignments for each point.
    * @param uf UnionFind structure that will be modified.
    */
-  template<typename MatType>
   void BatchCluster(const MatType& data, UnionFind& uf);
 };
 
diff -pruN 4.3.0-2/src/mlpack/methods/dbscan/dbscan_impl.hpp 4.4.0-1/src/mlpack/methods/dbscan/dbscan_impl.hpp
--- 4.3.0-2/src/mlpack/methods/dbscan/dbscan_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/dbscan/dbscan_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,7 +21,7 @@ namespace mlpack {
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
 DBSCAN<RangeSearchType, PointSelectionPolicy>::DBSCAN(
-    const double epsilon,
+    const ElemType epsilon,
     const size_t minPoints,
     const bool batchMode,
     RangeSearchType rangeSearch,
@@ -40,10 +40,9 @@ DBSCAN<RangeSearchType, PointSelectionPo
  * and also the centroid of each cluster.
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
-template<typename MatType>
 size_t DBSCAN<RangeSearchType, PointSelectionPolicy>::Cluster(
     const MatType& data,
-    arma::mat& centroids)
+    MatType& centroids)
 {
   // These assignments will be thrown away, but there is no way to avoid
   // calculating them.
@@ -58,11 +57,10 @@ size_t DBSCAN<RangeSearchType, PointSele
  * the centroid of each cluster and also the list of cluster assignments.
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
-template<typename MatType>
 size_t DBSCAN<RangeSearchType, PointSelectionPolicy>::Cluster(
     const MatType& data,
     arma::Row<size_t>& assignments,
-    arma::mat& centroids)
+    MatType& centroids)
 {
   const size_t numClusters = Cluster(data, assignments);
 
@@ -94,7 +92,6 @@ size_t DBSCAN<RangeSearchType, PointSele
  * also the list of cluster assignments.
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
-template<typename MatType>
 size_t DBSCAN<RangeSearchType, PointSelectionPolicy>::Cluster(
     const MatType& data,
     arma::Row<size_t>& assignments)
@@ -114,7 +111,7 @@ size_t DBSCAN<RangeSearchType, PointSele
     assignments[i] = uf.Find(i);
 
   // Get a count of all clusters.
-  const size_t numClusters = arma::max(assignments) + 1;
+  const size_t numClusters = max(assignments) + 1;
   arma::Col<size_t> counts(numClusters, arma::fill::zeros);
   for (size_t i = 0; i < assignments.n_elem; ++i)
     counts[assignments[i]]++;
@@ -146,13 +143,12 @@ size_t DBSCAN<RangeSearchType, PointSele
  * dual-tree algorithm.
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
-template<typename MatType>
 void DBSCAN<RangeSearchType, PointSelectionPolicy>::PointwiseCluster(
     const MatType& data,
     UnionFind& uf)
 {
   std::vector<std::vector<size_t>> neighbors;
-  std::vector<std::vector<double>> distances;
+  std::vector<std::vector<ElemType>> distances;
 
   // Note that the strategy here is somewhat different from the original DBSCAN
   // paper.  The original DBSCAN paper grows each cluster individually to its
@@ -181,8 +177,8 @@ void DBSCAN<RangeSearchType, PointSelect
     visited[index] = true;
 
     // Do the range search for only this point.
-    rangeSearch.Search(data.col(index), Range(0.0, epsilon), neighbors,
-        distances);
+    rangeSearch.Search(data.col(index), RangeType<ElemType>(ElemType(0.0), epsilon),
+        neighbors, distances);
 
     // Union to all neighbors if the point is not noise.
     //
@@ -225,17 +221,17 @@ void DBSCAN<RangeSearchType, PointSelect
  * naive search).
  */
 template<typename RangeSearchType, typename PointSelectionPolicy>
-template<typename MatType>
 void DBSCAN<RangeSearchType, PointSelectionPolicy>::BatchCluster(
     const MatType& data,
     UnionFind& uf)
 {
   // For each point, find the points in epsilon-neighborhood and their distances.
   std::vector<std::vector<size_t>> neighbors;
-  std::vector<std::vector<double>> distances;
+  std::vector<std::vector<ElemType>> distances;
   Log::Info << "Performing range search." << std::endl;
   rangeSearch.Train(data);
-  rangeSearch.Search(Range(0.0, epsilon), neighbors, distances);
+  rangeSearch.Search(RangeType<ElemType>(ElemType(0.0), epsilon), neighbors,
+      distances);
   Log::Info << "Range search complete." << std::endl;
 
   // See the description of the algorithm in `PointwiseCluster()`.  The strategy
diff -pruN 4.3.0-2/src/mlpack/methods/dbscan/dbscan_main.cpp 4.4.0-1/src/mlpack/methods/dbscan/dbscan_main.cpp
--- 4.3.0-2/src/mlpack/methods/dbscan/dbscan_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/dbscan/dbscan_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,7 +71,7 @@ BINDING_EXAMPLE(
 BINDING_SEE_ALSO("DBSCAN on Wikipedia", "https://en.wikipedia.org/wiki/DBSCAN");
 BINDING_SEE_ALSO("A density-based algorithm for discovering clusters in large "
     "spatial databases with noise (pdf)",
-    "http://www.aaai.org/Papers/KDD/1996/KDD96-037.pdf");
+    "https://cdn.aaai.org/KDD/1996/KDD96-037.pdf");
 BINDING_SEE_ALSO("DBSCAN class documentation",
     "@src/mlpack/methods/dbscan/dbscan.hpp");
 
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/all_categorical_split_impl.hpp 4.4.0-1/src/mlpack/methods/decision_tree/all_categorical_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/all_categorical_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/all_categorical_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,7 +53,7 @@ double AllCategoricalSplit<FitnessFuncti
 
   // If each child will have the minimum number of points in it, we can split.
   // Otherwise we can't.
-  if (arma::min(counts) < minimumLeafSize)
+  if (min(counts) < minimumLeafSize)
     return DBL_MAX;
 
   // Calculate the gain of the split.  First we have to calculate the labels
@@ -150,7 +150,7 @@ double AllCategoricalSplit<FitnessFuncti
 
   // If each child will have the minimum number of points in it, we can split.
   // Otherwise we can't.
-  if (arma::min(counts) < minimumLeafSize)
+  if (min(counts) < minimumLeafSize)
     return DBL_MAX;
 
   // Calculate the gain of the split.  First we have to calculate the labels
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/best_binary_numeric_split.hpp 4.4.0-1/src/mlpack/methods/decision_tree/best_binary_numeric_split.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/best_binary_numeric_split.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/best_binary_numeric_split.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -103,7 +103,7 @@ class BestBinaryNumericSplit
    * @param splitInfo Stores split information on a successful split.
    * @param aux Auxiliary split information, which may be modified on a
    *      successful split.
-   * @param fitnessFunction The FitnessFunction object instance. It it used to
+   * @param fitnessFunction The FitnessFunction object instance. It is used to
    *      evaluate the gain for the split.
    */
   template<bool UseWeights, typename VecType, typename ResponsesType,
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/best_binary_numeric_split_impl.hpp 4.4.0-1/src/mlpack/methods/decision_tree/best_binary_numeric_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/best_binary_numeric_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/best_binary_numeric_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -70,7 +70,7 @@ double BestBinaryNumericSplit<FitnessFun
   if (UseWeights)
   {
     classWeightSums.zeros(numClasses, 2);
-    totalWeight = arma::accu(sortedWeights);
+    totalWeight = accu(sortedWeights);
     bestFoundGain *= totalWeight;
 
     // Initialize the counts.
@@ -158,6 +158,16 @@ double BestBinaryNumericSplit<FitnessFun
       splitInfo[0] = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
 
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo[0] == data[sortedIndices[index]])
+      {
+        splitInfo[0] = std::nexttoward(splitInfo[0],
+            data[sortedIndices[index - 1]]);
+      }
+
       return gain;
     }
     else if (gain > bestFoundGain)
@@ -168,6 +178,16 @@ double BestBinaryNumericSplit<FitnessFun
       splitInfo[0] = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
       improved = true;
+
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo[0] == data[sortedIndices[index]])
+      {
+        splitInfo[0] = std::nexttoward(splitInfo[0],
+            data[sortedIndices[index - 1]]);
+      }
     }
   }
 
@@ -243,7 +263,7 @@ BestBinaryNumericSplit<FitnessFunction>:
 
   if (UseWeights)
   {
-    totalWeight = arma::accu(sortedWeights);
+    totalWeight = accu(sortedWeights);
     bestFoundGain *= totalWeight;
 
     for (size_t i = 0; i < minimum - 1; ++i)
@@ -297,6 +317,13 @@ BestBinaryNumericSplit<FitnessFunction>:
       splitInfo = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
 
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo == data[sortedIndices[index]])
+        splitInfo = std::nexttoward(splitInfo, data[sortedIndices[index - 1]]);
+
       return gain;
     }
      if (gain > bestFoundGain)
@@ -306,6 +333,13 @@ BestBinaryNumericSplit<FitnessFunction>:
       splitInfo = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
       improved = true;
+
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo == data[sortedIndices[index]])
+        splitInfo = std::nexttoward(splitInfo, data[sortedIndices[index - 1]]);
     }
   }
 
@@ -382,7 +416,7 @@ BestBinaryNumericSplit<FitnessFunction>:
 
   if (UseWeights)
   {
-    totalWeight = arma::accu(sortedWeights);
+    totalWeight = accu(sortedWeights);
     bestFoundGain *= totalWeight;
 
     for (size_t i = 0; i < minimum - 1; ++i)
@@ -444,6 +478,13 @@ BestBinaryNumericSplit<FitnessFunction>:
       splitInfo = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
 
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo == data[sortedIndices[index]])
+        splitInfo = std::nexttoward(splitInfo, data[sortedIndices[index - 1]]);
+
       return gain;
     }
     if (gain > bestFoundGain)
@@ -453,6 +494,13 @@ BestBinaryNumericSplit<FitnessFunction>:
       splitInfo = (data[sortedIndices[index - 1]] +
           data[sortedIndices[index]]) / 2.0;
       improved = true;
+
+      // In some very extreme cases, floating-point inaccuracies can lead to the
+      // split result being the upper bound, which is problematic for later as
+      // all the child points will be sent to the left child.  If this happens,
+      // bump it down incrementally.
+      if (splitInfo == data[sortedIndices[index]])
+        splitInfo = std::nexttoward(splitInfo, data[sortedIndices[index - 1]]);
     }
   }
   // If we didn't improve, return the original gain exactly as we got it
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree.hpp 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -143,64 +143,65 @@ class DecisionTree :
           typename std::remove_reference<WeightsType>::type>::value>* = 0);
 
   /**
-   * Using the hyperparameters of another decision tree, train on the given data
-   * and labels with weights, where the data can be both numeric and
-   * categorical.  Setting minimumLeafSize and minimumGainSplit too small may
-   * cause the tree to overfit, but setting them too large may cause it to
-   * underfit.
+   * Construct the decision tree on the given data and labels with weights,
+   * assuming that the data is all of the numeric type. Setting minimumLeafSize
+   * and minimumGainSplit too small may cause the tree to overfit, but setting
+   * them too large may cause it to underfit.
    *
    * Use std::move if data, labels or weights are no longer needed to avoid
    * copies.
    *
-   * @param other Tree to take ownership of.
    * @param data Dataset to train on.
-   * @param datasetInfo Type information for each dimension of the dataset.
    * @param labels Labels for each training point.
    * @param numClasses Number of classes in the dataset.
    * @param weights The weight list of given label.
    * @param minimumLeafSize Minimum number of points in each leaf node.
    * @param minimumGainSplit Minimum gain for the node to split.
+   * @param maximumDepth Maximum depth for the tree.
+   * @param dimensionSelector Instantiated dimension selection policy.
    */
   template<typename MatType, typename LabelsType, typename WeightsType>
   DecisionTree(
-      const DecisionTree& other,
       MatType data,
-      const data::DatasetInfo& datasetInfo,
       LabelsType labels,
       const size_t numClasses,
       WeightsType weights,
       const size_t minimumLeafSize = 10,
       const double minimumGainSplit = 1e-7,
+      const size_t maximumDepth = 0,
+      DimensionSelectionType dimensionSelector = DimensionSelectionType(),
       const std::enable_if_t<arma::is_arma_type<
           typename std::remove_reference<WeightsType>::type>::value>* = 0);
+
   /**
-   * Construct the decision tree on the given data and labels with weights,
-   * assuming that the data is all of the numeric type. Setting minimumLeafSize
-   * and minimumGainSplit too small may cause the tree to overfit, but setting
-   * them too large may cause it to underfit.
+   * Using the hyperparameters of another decision tree, train on the given data
+   * and labels with weights, where the data can be both numeric and
+   * categorical.  Setting minimumLeafSize and minimumGainSplit too small may
+   * cause the tree to overfit, but setting them too large may cause it to
+   * underfit.
    *
    * Use std::move if data, labels or weights are no longer needed to avoid
    * copies.
    *
+   * @param other Tree to take ownership of.
    * @param data Dataset to train on.
+   * @param datasetInfo Type information for each dimension of the dataset.
    * @param labels Labels for each training point.
    * @param numClasses Number of classes in the dataset.
-   * @param weights The Weight list of given labels.
+   * @param weights The weight list of given label.
    * @param minimumLeafSize Minimum number of points in each leaf node.
    * @param minimumGainSplit Minimum gain for the node to split.
-   * @param maximumDepth Maximum depth for the tree.
-   * @param dimensionSelector Instantiated dimension selection policy.
    */
   template<typename MatType, typename LabelsType, typename WeightsType>
   DecisionTree(
+      const DecisionTree& other,
       MatType data,
+      const data::DatasetInfo& datasetInfo,
       LabelsType labels,
       const size_t numClasses,
       WeightsType weights,
       const size_t minimumLeafSize = 10,
       const double minimumGainSplit = 1e-7,
-      const size_t maximumDepth = 0,
-      DimensionSelectionType dimensionSelector = DimensionSelectionType(),
       const std::enable_if_t<arma::is_arma_type<
           typename std::remove_reference<WeightsType>::type>::value>* = 0);
 
@@ -546,14 +547,14 @@ class DecisionTree :
    * @param maximumDepth Maximum depth for the tree.
    * @return The final entropy of decision tree.
    */
-  template<bool UseWeights, typename MatType>
+  template<bool UseWeights, typename MatType, typename WeightsType>
   double Train(MatType& data,
                const size_t begin,
                const size_t count,
                const data::DatasetInfo& datasetInfo,
                arma::Row<size_t>& labels,
                const size_t numClasses,
-               arma::rowvec& weights,
+               WeightsType& weights,
                const size_t minimumLeafSize,
                const double minimumGainSplit,
                const size_t maximumDepth,
@@ -575,13 +576,13 @@ class DecisionTree :
    * @param maximumDepth Maximum depth for the tree.
    * @return The final entropy of decision tree.
    */
-  template<bool UseWeights, typename MatType>
+  template<bool UseWeights, typename MatType, typename WeightsType>
   double Train(MatType& data,
                const size_t begin,
                const size_t count,
                arma::Row<size_t>& labels,
                const size_t numClasses,
-               arma::rowvec& weights,
+               WeightsType& weights,
                const size_t minimumLeafSize,
                const double minimumGainSplit,
                const size_t maximumDepth,
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_impl.hpp 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_impl.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -132,30 +132,28 @@ DecisionTree<FitnessFunction,
       dimensionSelector);
 }
 
-//! Construct and train with weights.
+//! Construct and train with weights on numerical data.
 template<typename FitnessFunction,
-        template<typename> class NumericSplitType,
-        template<typename> class CategoricalSplitType,
-        typename DimensionSelectionType,
-        bool NoRecursion>
+         template<typename> class NumericSplitType,
+         template<typename> class CategoricalSplitType,
+         typename DimensionSelectionType,
+         bool NoRecursion>
 template<typename MatType, typename LabelsType, typename WeightsType>
 DecisionTree<FitnessFunction,
-        NumericSplitType,
-        CategoricalSplitType,
-        DimensionSelectionType,
-        NoRecursion>::DecisionTree(
-    const DecisionTree& other,
+             NumericSplitType,
+             CategoricalSplitType,
+             DimensionSelectionType,
+             NoRecursion>::DecisionTree(
     MatType data,
-    const data::DatasetInfo& datasetInfo,
     LabelsType labels,
     const size_t numClasses,
     WeightsType weights,
     const size_t minimumLeafSize,
     const double minimumGainSplit,
+    const size_t maximumDepth,
+    DimensionSelectionType dimensionSelector,
     const std::enable_if_t<arma::is_arma_type<
-        typename std::remove_reference<WeightsType>::type>::value>*):
-        NumericAuxiliarySplitInfo(other),
-        CategoricalAuxiliarySplitInfo(other)
+        typename std::remove_reference<WeightsType>::type>::value>*)
 {
   using TrueMatType = typename std::decay<MatType>::type;
   using TrueLabelsType = typename std::decay<LabelsType>::type;
@@ -166,35 +164,38 @@ DecisionTree<FitnessFunction,
   TrueLabelsType tmpLabels(std::move(labels));
   TrueWeightsType tmpWeights(std::move(weights));
 
+  // Set the correct dimensionality for the dimension selector.
+  dimensionSelector.Dimensions() = tmpData.n_rows;
+
   // Pass off work to the weighted Train() method.
-  Train<true>(tmpData, 0, tmpData.n_cols, datasetInfo, tmpLabels, numClasses,
-              tmpWeights, minimumLeafSize, minimumGainSplit);
+  Train<true>(tmpData, 0, tmpData.n_cols, tmpLabels, numClasses, tmpWeights,
+      minimumLeafSize, minimumGainSplit, maximumDepth, dimensionSelector);
 }
 
 //! Construct and train with weights.
 template<typename FitnessFunction,
-         template<typename> class NumericSplitType,
-         template<typename> class CategoricalSplitType,
-         typename DimensionSelectionType,
-         bool NoRecursion>
+        template<typename> class NumericSplitType,
+        template<typename> class CategoricalSplitType,
+        typename DimensionSelectionType,
+        bool NoRecursion>
 template<typename MatType, typename LabelsType, typename WeightsType>
 DecisionTree<FitnessFunction,
-             NumericSplitType,
-             CategoricalSplitType,
-             DimensionSelectionType,
-             NoRecursion>::DecisionTree(
+        NumericSplitType,
+        CategoricalSplitType,
+        DimensionSelectionType,
+        NoRecursion>::DecisionTree(
+    const DecisionTree& other,
     MatType data,
+    const data::DatasetInfo& datasetInfo,
     LabelsType labels,
     const size_t numClasses,
     WeightsType weights,
     const size_t minimumLeafSize,
     const double minimumGainSplit,
-    const size_t maximumDepth,
-    DimensionSelectionType dimensionSelector,
-    const std::enable_if_t<
-        arma::is_arma_type<
-        typename std::remove_reference<
-        WeightsType>::type>::value>*)
+    const std::enable_if_t<arma::is_arma_type<
+        typename std::remove_reference<WeightsType>::type>::value>*):
+        NumericAuxiliarySplitInfo(other),
+        CategoricalAuxiliarySplitInfo(other)
 {
   using TrueMatType = typename std::decay<MatType>::type;
   using TrueLabelsType = typename std::decay<LabelsType>::type;
@@ -205,12 +206,9 @@ DecisionTree<FitnessFunction,
   TrueLabelsType tmpLabels(std::move(labels));
   TrueWeightsType tmpWeights(std::move(weights));
 
-  // Set the correct dimensionality for the dimension selector.
-  dimensionSelector.Dimensions() = tmpData.n_rows;
-
   // Pass off work to the weighted Train() method.
-  Train<true>(tmpData, 0, tmpData.n_cols, tmpLabels, numClasses, tmpWeights,
-      minimumLeafSize, minimumGainSplit, maximumDepth, dimensionSelector);
+  Train<true>(tmpData, 0, tmpData.n_cols, datasetInfo, tmpLabels, numClasses,
+              tmpWeights, minimumLeafSize, minimumGainSplit);
 }
 
 //! Construct and train with weights.
@@ -616,7 +614,7 @@ template<typename FitnessFunction,
          template<typename> class CategoricalSplitType,
          typename DimensionSelectionType,
          bool NoRecursion>
-template<bool UseWeights, typename MatType>
+template<bool UseWeights, typename MatType, typename WeightsType>
 double DecisionTree<FitnessFunction,
                     NumericSplitType,
                     CategoricalSplitType,
@@ -628,7 +626,7 @@ double DecisionTree<FitnessFunction,
     const data::DatasetInfo& datasetInfo,
     arma::Row<size_t>& labels,
     const size_t numClasses,
-    arma::rowvec& weights,
+    WeightsType& weights,
     const size_t minimumLeafSize,
     const double minimumGainSplit,
     const size_t maximumDepth,
@@ -800,7 +798,7 @@ template<typename FitnessFunction,
          template<typename> class CategoricalSplitType,
          typename DimensionSelectionType,
          bool NoRecursion>
-template<bool UseWeights, typename MatType>
+template<bool UseWeights, typename MatType, typename WeightsType>
 double DecisionTree<FitnessFunction,
                     NumericSplitType,
                     CategoricalSplitType,
@@ -811,7 +809,7 @@ double DecisionTree<FitnessFunction,
     const size_t count,
     arma::Row<size_t>& labels,
     const size_t numClasses,
-    arma::rowvec& weights,
+    WeightsType& weights,
     const size_t minimumLeafSize,
     const double minimumGainSplit,
     const size_t maximumDepth,
@@ -1044,7 +1042,7 @@ void DecisionTree<FitnessFunction,
   if (children.size() == 0)
   {
     predictions.fill(majorityClass);
-    probabilities = arma::repmat(classProbabilities, 1, data.n_cols);
+    probabilities = repmat(classProbabilities, 1, data.n_cols);
     return;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_main.cpp 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_main.cpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -40,8 +40,8 @@ BINDING_LONG_DESC(
     "\n\n"
     "The training set and associated labels are specified with the " +
     PRINT_PARAM_STRING("training") + " and " + PRINT_PARAM_STRING("labels") +
-    " parameters, respectively.  The labels should be in the range [0, "
-    "num_classes - 1]. Optionally, if " +
+    " parameters, respectively.  The labels should be in the range `[0, "
+    "num_classes - 1]`. Optionally, if " +
     PRINT_PARAM_STRING("labels") + " is not specified, the labels are assumed "
     "to be the last dimension of the training dataset."
     "\n\n"
@@ -57,8 +57,8 @@ BINDING_LONG_DESC(
     "the minimum gain that is needed for the node to split.  The " +
     PRINT_PARAM_STRING("maximum_depth") + " parameter specifies "
     "the maximum depth of the tree.  If " +
-    PRINT_PARAM_STRING("print_training_error") + " is specified, the training "
-    "error will be printed."
+    PRINT_PARAM_STRING("print_training_accuracy") + " is specified, the "
+    "training accuracy will be printed."
     "\n\n"
     "Test data may be specified with the " + PRINT_PARAM_STRING("test") + " "
     "parameter, and if performance numbers are desired for that test set, "
@@ -90,14 +90,13 @@ BINDING_EXAMPLE(
         "test_labels", "test_labels", "predictions", "predictions"));
 
 // See also...
-BINDING_SEE_ALSO("Decision stump", "#decision_stump");
 BINDING_SEE_ALSO("Random forest", "#random_forest");
 BINDING_SEE_ALSO("Decision trees on Wikipedia",
     "https://en.wikipedia.org/wiki/Decision_tree_learning");
 BINDING_SEE_ALSO("Induction of Decision Trees (pdf)",
-    "https://link.springer.com/content/pdf/10.1007/BF00116251.pdf");
-BINDING_SEE_ALSO("DecisionTree class documentation",
-    "@src/mlpack/methods/decision_tree/decision_tree.hpp");
+    "http://www.cs.bc.edu/~alvarez/ML/QuinlanID3Paper.pdf");
+BINDING_SEE_ALSO("DecisionTree C++ class documentation",
+    "@doc/user/methods/decision_tree.md");
 
 // Datasets.
 PARAM_MATRIX_AND_INFO_IN("training", "Training dataset (may be categorical).",
@@ -115,9 +114,6 @@ PARAM_DOUBLE_IN("minimum_gain_split", "M
     1e-7);
 PARAM_INT_IN("maximum_depth", "Maximum depth of the tree (0 means no limit).",
     "D", 0);
-// This is deprecated and should be removed in mlpack 4.0.0.
-PARAM_FLAG("print_training_error", "Print the training error (deprecated; will "
-      "be removed in mlpack 4.0.0).", "e");
 PARAM_FLAG("print_training_accuracy", "Print the training accuracy.", "a");
 
 // Output parameters.
@@ -180,12 +176,6 @@ void BINDING_FUNCTION(util::Params& para
       [](double x) { return (x > 0.0 && x < 1.0); }, true,
       "gain split must be a fraction in range [0,1]");
 
-  if (params.Has("print_training_error"))
-  {
-    Log::Warn << "The option " << PRINT_PARAM_STRING("print_training_error")
-        << " is deprecated and will be removed in mlpack 4.0.0." << std::endl;
-  }
-
   // Load the model or build the tree.
   DecisionTreeModel* model;
   arma::mat trainingSet;
@@ -205,12 +195,12 @@ void BINDING_FUNCTION(util::Params& para
       // Extract the labels as the last
       Log::Info << "Using the last dimension of training set as labels."
           << endl;
-      labels = arma::conv_to<arma::Row<size_t>>::from(
+      labels = ConvTo<arma::Row<size_t>>::From(
           trainingSet.row(trainingSet.n_rows - 1));
       trainingSet.shed_row(trainingSet.n_rows - 1);
     }
 
-    const size_t numClasses = arma::max(arma::max(labels)) + 1;
+    const size_t numClasses = max(labels) + 1;
 
     // Now build the tree.
     const size_t minLeafSize = (size_t) params.Get<int>("minimum_leaf_size");
@@ -223,8 +213,7 @@ void BINDING_FUNCTION(util::Params& para
     {
       arma::Row<double> weights =
           std::move(params.Get<arma::Mat<double>>("weights"));
-      if (params.Has("print_training_error") ||
-          params.Has("print_training_accuracy"))
+      if (params.Has("print_training_accuracy"))
       {
         model->tree = DecisionTree<>(trainingSet, model->info, labels,
             numClasses, std::move(weights), minLeafSize, minimumGainSplit,
@@ -239,7 +228,7 @@ void BINDING_FUNCTION(util::Params& para
     }
     else
     {
-      if (params.Has("print_training_error"))
+      if (params.Has("print_training_accuracy"))
       {
         model->tree = DecisionTree<>(trainingSet, model->info, labels,
             numClasses, minLeafSize, minimumGainSplit, maxDepth);
@@ -253,8 +242,7 @@ void BINDING_FUNCTION(util::Params& para
     }
 
     // Do we need to print training error?
-    if (params.Has("print_training_error") ||
-        params.Has("print_training_accuracy"))
+    if (params.Has("print_training_accuracy"))
     {
       arma::Row<size_t> predictions;
       arma::mat probabilities;
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_regressor.hpp 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_regressor.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_regressor.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_regressor.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -393,7 +393,7 @@ class DecisionTreeRegressor :
    * @param point Point to predict.
    */
   template<typename VecType>
-  double Predict(const VecType& point) const;
+  typename VecType::elem_type Predict(const VecType& point) const;
 
   /**
    * Make prediction for the given points, using the entire tree. The predicted
@@ -402,9 +402,9 @@ class DecisionTreeRegressor :
    * @param data Set of points to predict.
    * @param predictions This will be filled with predictions for each point.
    */
-  template<typename MatType>
+  template<typename MatType, typename PredVecType>
   void Predict(const MatType& data,
-                arma::Row<double>& predictions) const;
+               PredVecType& predictions) const;
 
   /**
    * Serialize the tree.
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_regressor_impl.hpp 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_regressor_impl.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/decision_tree_regressor_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/decision_tree_regressor_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -943,11 +943,12 @@ template<typename FitnessFunction,
          typename DimensionSelectionType,
          bool NoRecursion>
 template<typename VecType>
-double DecisionTreeRegressor<FitnessFunction,
-                             NumericSplitType,
-                             CategoricalSplitType,
-                             DimensionSelectionType,
-                             NoRecursion>::Predict(const VecType& point) const
+typename VecType::elem_type
+DecisionTreeRegressor<FitnessFunction,
+                      NumericSplitType,
+                      CategoricalSplitType,
+                      DimensionSelectionType,
+                      NoRecursion>::Predict(const VecType& point) const
 {
   if (children.size() == 0)
   {
@@ -955,7 +956,8 @@ double DecisionTreeRegressor<FitnessFunc
     return prediction;
   }
 
-  return children[CalculateDirection(point)]->Predict(point);
+  typedef typename VecType::elem_type ElemType;
+  return (ElemType) children[CalculateDirection(point)]->Predict(point);
 }
 
 //! Return the predictions for a set of points.
@@ -964,25 +966,28 @@ template<typename FitnessFunction,
          template<typename> class CategoricalSplitType,
          typename DimensionSelectionType,
          bool NoRecursion>
-template<typename MatType>
+template<typename MatType, typename PredVecType>
 void DecisionTreeRegressor<FitnessFunction,
                            NumericSplitType,
                            CategoricalSplitType,
                            DimensionSelectionType,
                            NoRecursion
->::Predict(const MatType& data, arma::Row<double>& predictions) const
+>::Predict(const MatType& data,
+           PredVecType& predictions) const
 {
+  typedef typename PredVecType::elem_type ElemType;
+
   predictions.set_size(data.n_cols);
   // If the tree's root is leaf.
   if (children.size() == 0)
   {
-    predictions.fill(prediction);
+    predictions.fill((ElemType) prediction);
     return;
   }
 
   // Loop over each point.
   for (size_t i = 0; i < data.n_cols; ++i)
-    predictions[i] = Predict(data.col(i));
+    predictions[i] = (ElemType) Predict(data.col(i));
 }
 
 template<typename FitnessFunction,
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/information_gain.hpp 4.4.0-1/src/mlpack/methods/decision_tree/information_gain.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/information_gain.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/information_gain.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -55,10 +55,10 @@ class InformationGain
    * @param numClasses Number of classes in the dataset.
    * @param weights Weights associated with labels.
    */
-  template<bool UseWeights>
+  template<bool UseWeights, typename WeightsType>
   static double Evaluate(const arma::Row<size_t>& labels,
                          const size_t numClasses,
-                         const arma::Row<double>& weights)
+                         const WeightsType& weights)
   {
      // Edge case: if there are no elements, the gain is zero.
      if (labels.n_elem == 0)
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/mad_gain.hpp 4.4.0-1/src/mlpack/methods/decision_tree/mad_gain.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/mad_gain.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/mad_gain.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -74,7 +74,7 @@ class MADGain
       Sum(values, begin, end, mean);
       mean /= (double) (end - begin);
 
-      mad = arma::accu(arma::abs(values.subvec(begin, end - 1) - mean));
+      mad = accu(arma::abs(values.subvec(begin, end - 1) - mean));
       mad /= (double) (end - begin);
     }
 
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/mse_gain.hpp 4.4.0-1/src/mlpack/methods/decision_tree/mse_gain.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/mse_gain.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/mse_gain.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -70,7 +70,7 @@ class MSEGain
       Sum(values, begin, end, mean);
       mean /= (double) (end - begin);
 
-      mse = arma::accu(arma::square(values.subvec(begin, end - 1) - mean));
+      mse = accu(square(values.subvec(begin, end - 1) - mean));
       mse /= (double) (end - begin);
     }
 
@@ -164,7 +164,7 @@ class MSEGain
 
     if (UseWeights)
     {
-      totalSumSquares = arma::accu(weights % arma::square(responses));
+      totalSumSquares = accu(weights % square(responses));
       for (size_t i = 0; i < minimum - 1; ++i)
       {
         const WType w = weights[i];
@@ -192,7 +192,7 @@ class MSEGain
     }
     else
     {
-      totalSumSquares = arma::accu(arma::square(responses));
+      totalSumSquares = accu(square(responses));
       for (size_t i = 0; i < minimum - 1; ++i)
       {
         const RType x = responses[i];
diff -pruN 4.3.0-2/src/mlpack/methods/decision_tree/random_binary_numeric_split_impl.hpp 4.4.0-1/src/mlpack/methods/decision_tree/random_binary_numeric_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/decision_tree/random_binary_numeric_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/decision_tree/random_binary_numeric_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,8 +41,8 @@ double RandomBinaryNumericSplit<FitnessF
   if (bestGain == 0.0)
     return DBL_MAX; // It can't be outperformed.
 
-  typename VecType::elem_type maxValue = arma::max(data);
-  typename VecType::elem_type minValue = arma::min(data);
+  typename VecType::elem_type maxValue = max(data);
+  typename VecType::elem_type minValue = min(data);
 
   // Sanity check: if the maximum element is the same as the minimum, we
   // can't split in this dimension.
@@ -63,7 +63,7 @@ double RandomBinaryNumericSplit<FitnessF
   if (UseWeights)
   {
     classWeightSums.zeros(numClasses, 2);
-    totalWeight = arma::accu(weights);
+    totalWeight = accu(weights);
     bestFoundGain *= totalWeight;
 
     for (size_t i = 0; i < data.n_elem; ++i)
@@ -161,8 +161,8 @@ double RandomBinaryNumericSplit<FitnessF
   if (bestGain == 0.0)
     return DBL_MAX; // It can't be outperformed.
 
-  typename VecType::elem_type maxValue = arma::max(data);
-  typename VecType::elem_type minValue = arma::min(data);
+  typename VecType::elem_type maxValue = max(data);
+  typename VecType::elem_type minValue = min(data);
 
   // Sanity check: if the maximum element is the same as the minimum, we
   // can't split in this dimension.
@@ -174,7 +174,7 @@ double RandomBinaryNumericSplit<FitnessF
   double totalRightWeight = 0.0;
   if (UseWeights)
   {
-    totalWeight = arma::accu(weights);
+    totalWeight = accu(weights);
     bestFoundGain *= totalWeight;
   }
   else
diff -pruN 4.3.0-2/src/mlpack/methods/det/det_main.cpp 4.4.0-1/src/mlpack/methods/det/det_main.cpp
--- 4.3.0-2/src/mlpack/methods/det/det_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/det/det_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,12 +64,10 @@ BINDING_LONG_DESC(
     PRINT_PARAM_STRING("test_set_estimates") + " output parameter.");
 
 // See also...
-BINDING_SEE_ALSO("Density estimation tree (DET) tutorial",
-    "@doc/tutorials/det.md");
 BINDING_SEE_ALSO("Density estimation on Wikipedia",
     "https://en.wikipedia.org/wiki/Density_estimation");
 BINDING_SEE_ALSO("Density estimation trees (pdf)",
-    "http://www.mlpack.org/papers/det.pdf");
+    "https://www.mlpack.org/papers/det.pdf");
 BINDING_SEE_ALSO("DTree class documentation",
     "@src/mlpack/methods/det/dtree.hpp");
 
diff -pruN 4.3.0-2/src/mlpack/methods/det/dtree.hpp 4.4.0-1/src/mlpack/methods/det/dtree.hpp
--- 4.3.0-2/src/mlpack/methods/det/dtree.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/det/dtree.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -222,7 +222,7 @@ class DTree
    */
   bool WithinRange(const VecType& query) const;
 
- private:
+ protected: /* these are checked in the tests */
   // The indices in the complete set of points
   // (after all forms of swapping in the original data
   // matrix to align all the points in a node
@@ -331,7 +331,7 @@ class DTree
   template<typename Archive>
   void serialize(Archive& ar, const uint32_t /* version */);
 
- private:
+ protected: /* these are checked in the tests */
   // Utility methods.
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/det/dtree_impl.hpp 4.4.0-1/src/mlpack/methods/det/dtree_impl.hpp
--- 4.3.0-2/src/mlpack/methods/det/dtree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/det/dtree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -332,7 +332,7 @@ DTree<MatType, TagType>::DTree(MatType &
     start(0),
     end(data.n_cols),
     maxVals(arma::max(data, 1)),
-    minVals(arma::min(data, 1)),
+    minVals(min(data, 1)),
     splitDim(size_t(-1)),
     splitValue(std::numeric_limits<ElemType>::max()),
     subtreeLeavesLogNegError(-DBL_MAX),
diff -pruN 4.3.0-2/src/mlpack/methods/emst/emst_main.cpp 4.4.0-1/src/mlpack/methods/emst/emst_main.cpp
--- 4.3.0-2/src/mlpack/methods/emst/emst_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/emst/emst_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,11 +73,10 @@ BINDING_EXAMPLE(
     "the third column corresponds to the distance between the two points.");
 
 // See also...
-BINDING_SEE_ALSO("EMST Tutorial", "@doc/tutorials/emst.md");
 BINDING_SEE_ALSO("Minimum spanning tree on Wikipedia",
         "https://en.wikipedia.org/wiki/Minimum_spanning_tree");
 BINDING_SEE_ALSO("Fast Euclidean Minimum Spanning Tree: Algorithm, Analysis,"
-        " and Applications (pdf)", "http://www.mlpack.org/papers/emst.pdf");
+        " and Applications (pdf)", "https://www.mlpack.org/papers/emst.pdf");
 BINDING_SEE_ALSO("DualTreeBoruvka class documentation",
         "@src/mlpack/methods/emst/dtb.hpp");
 
diff -pruN 4.3.0-2/src/mlpack/methods/fastmks/fastmks_main.cpp 4.4.0-1/src/mlpack/methods/fastmks/fastmks_main.cpp
--- 4.3.0-2/src/mlpack/methods/fastmks/fastmks_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/fastmks/fastmks_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -66,11 +66,9 @@ BINDING_EXAMPLE(
     " parameter.");
 
 // See also...
-BINDING_SEE_ALSO("Fast max-kernel search tutorial (fastmks)",
-    "@doc/tutorials/fastmks.md");
 BINDING_SEE_ALSO("k-nearest-neighbor search", "#knn");
 BINDING_SEE_ALSO("Dual-tree Fast Exact Max-Kernel Search (pdf)",
-    "http://mlpack.org/papers/fmks.pdf");
+    "https://mlpack.org/papers/fmks.pdf");
 BINDING_SEE_ALSO("FastMKS class documentation",
     "@src/mlpack/methods/fastmks/fastmks.hpp");
 
@@ -186,7 +184,7 @@ void BINDING_FUNCTION(util::Params& para
     else if (kernelType == "cosine")
     {
       CosineDistance cd;
-      model->KernelType() = FastMKSModel::COSINE_DISTANCE;
+      model->KernelType() = FastMKSModel::COSINE_SIMILARITY;
       model->BuildModel(timers, std::move(referenceData), cd, single, naive,
           base);
     }
diff -pruN 4.3.0-2/src/mlpack/methods/fastmks/fastmks_model.hpp 4.4.0-1/src/mlpack/methods/fastmks/fastmks_model.hpp
--- 4.3.0-2/src/mlpack/methods/fastmks/fastmks_model.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/fastmks/fastmks_model.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,7 +17,7 @@
 #include <mlpack/core/kernels/kernel_traits.hpp>
 #include <mlpack/core/kernels/linear_kernel.hpp>
 #include <mlpack/core/kernels/polynomial_kernel.hpp>
-#include <mlpack/core/kernels/cosine_distance.hpp>
+#include <mlpack/core/kernels/cosine_similarity.hpp>
 #include <mlpack/core/kernels/gaussian_kernel.hpp>
 #include <mlpack/core/kernels/epanechnikov_kernel.hpp>
 #include <mlpack/core/kernels/hyperbolic_tangent_kernel.hpp>
@@ -38,7 +38,7 @@ class FastMKSModel
   {
     LINEAR_KERNEL,
     POLYNOMIAL_KERNEL,
-    COSINE_DISTANCE,
+    COSINE_SIMILARITY,
     GAUSSIAN_KERNEL,
     EPANECHNIKOV_KERNEL,
     TRIANGULAR_KERNEL,
@@ -142,7 +142,7 @@ class FastMKSModel
   //! This will only be non-NULL if this is the type of kernel we are using.
   FastMKS<PolynomialKernel>* polynomial;
   //! This will only be non-NULL if this is the type of kernel we are using.
-  FastMKS<CosineDistance>* cosine;
+  FastMKS<CosineSimilarity>* cosine;
   //! This will only be non-NULL if this is the type of kernel we are using.
   FastMKS<GaussianKernel>* gaussian;
   //! This will only be non-NULL if this is the type of kernel we are using.
diff -pruN 4.3.0-2/src/mlpack/methods/fastmks/fastmks_model_impl.hpp 4.4.0-1/src/mlpack/methods/fastmks/fastmks_model_impl.hpp
--- 4.3.0-2/src/mlpack/methods/fastmks/fastmks_model_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/fastmks/fastmks_model_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ inline FastMKSModel::FastMKSModel(const
     polynomial(other.polynomial == NULL ? NULL :
         new FastMKS<PolynomialKernel>(*other.polynomial)),
     cosine(other.cosine == NULL ? NULL :
-        new FastMKS<CosineDistance>(*other.cosine)),
+        new FastMKS<CosineSimilarity>(*other.cosine)),
     gaussian(other.gaussian == NULL ? NULL :
         new FastMKS<GaussianKernel>(*other.gaussian)),
     epan(other.epan == NULL ? NULL :
@@ -98,7 +98,7 @@ inline FastMKSModel& FastMKSModel::opera
     if (other.polynomial)
       polynomial = new FastMKS<PolynomialKernel>(*other.polynomial);
     if (other.cosine)
-      cosine = new FastMKS<CosineDistance>(*other.cosine);
+      cosine = new FastMKS<CosineSimilarity>(*other.cosine);
     if (other.gaussian)
       gaussian = new FastMKS<GaussianKernel>(*other.gaussian);
     if (other.epan)
@@ -164,7 +164,7 @@ inline bool FastMKSModel::Naive() const
       return linear->Naive();
     case POLYNOMIAL_KERNEL:
       return polynomial->Naive();
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       return cosine->Naive();
     case GAUSSIAN_KERNEL:
       return gaussian->Naive();
@@ -187,7 +187,7 @@ inline bool& FastMKSModel::Naive()
       return linear->Naive();
     case POLYNOMIAL_KERNEL:
       return polynomial->Naive();
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       return cosine->Naive();
     case GAUSSIAN_KERNEL:
       return gaussian->Naive();
@@ -210,7 +210,7 @@ inline bool FastMKSModel::SingleMode() c
       return linear->SingleMode();
     case POLYNOMIAL_KERNEL:
       return polynomial->SingleMode();
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       return cosine->SingleMode();
     case GAUSSIAN_KERNEL:
       return gaussian->SingleMode();
@@ -233,7 +233,7 @@ inline bool& FastMKSModel::SingleMode()
       return linear->SingleMode();
     case POLYNOMIAL_KERNEL:
       return polynomial->SingleMode();
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       return cosine->SingleMode();
     case GAUSSIAN_KERNEL:
       return gaussian->SingleMode();
@@ -339,8 +339,8 @@ void FastMKSModel::BuildModel(util::Time
           base);
       break;
 
-    case COSINE_DISTANCE:
-      cosine = new FastMKS<CosineDistance>(singleMode, naive);
+    case COSINE_SIMILARITY:
+      cosine = new FastMKS<CosineSimilarity>(singleMode, naive);
       BuildFastMKSModel(timers, *cosine, kernel, std::move(referenceData), base);
       break;
 
@@ -411,7 +411,7 @@ void FastMKSModel::serialize(Archive& ar
       ar(CEREAL_POINTER(polynomial));
       break;
 
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       ar(CEREAL_POINTER(cosine));
       break;
 
@@ -476,7 +476,7 @@ inline void FastMKSModel::Search(
     case POLYNOMIAL_KERNEL:
       Search(timers, *polynomial, querySet, k, indices, kernels, base);
       break;
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       Search(timers, *cosine, querySet, k, indices, kernels, base);
       break;
     case GAUSSIAN_KERNEL:
@@ -511,7 +511,7 @@ inline void FastMKSModel::Search(
     case POLYNOMIAL_KERNEL:
       polynomial->Search(k, indices, kernels);
       break;
-    case COSINE_DISTANCE:
+    case COSINE_SIMILARITY:
       cosine->Search(k, indices, kernels);
       break;
     case GAUSSIAN_KERNEL:
diff -pruN 4.3.0-2/src/mlpack/methods/fastmks/fastmks_rules_impl.hpp 4.4.0-1/src/mlpack/methods/fastmks/fastmks_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/fastmks/fastmks_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/fastmks/fastmks_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,13 +36,13 @@ FastMKSRules<KernelType, TreeType>::Fast
   // Precompute each self-kernel.
   queryKernels.set_size(querySet.n_cols);
   for (size_t i = 0; i < querySet.n_cols; ++i)
-    queryKernels[i] = sqrt(kernel.Evaluate(querySet.col(i),
-                                           querySet.col(i)));
+    queryKernels[i] = std::sqrt(kernel.Evaluate(querySet.col(i),
+                                                querySet.col(i)));
 
   referenceKernels.set_size(referenceSet.n_cols);
   for (size_t i = 0; i < referenceSet.n_cols; ++i)
-    referenceKernels[i] = sqrt(kernel.Evaluate(referenceSet.col(i),
-                                               referenceSet.col(i)));
+    referenceKernels[i] = std::sqrt(kernel.Evaluate(referenceSet.col(i),
+                                                    referenceSet.col(i)));
 
   // Set to invalid memory, so that the first node combination does not try to
   // dereference null pointers.
@@ -141,9 +141,10 @@ double FastMKSRules<KernelType, TreeType
       const double delta = (1 - 0.5 * squaredDist);
       if (lastKernel <= delta)
       {
-        const double gamma = combinedDistBound * sqrt(1 - 0.25 * squaredDist);
+        const double gamma = combinedDistBound * 
+            std::sqrt(1 - 0.25 * squaredDist);
         maxKernelBound = lastKernel * delta +
-             gamma * sqrt(1 - std::pow(lastKernel, 2.0));
+             gamma * std::sqrt(1 - std::pow(lastKernel, 2.0));
       }
       else
       {
@@ -195,9 +196,9 @@ double FastMKSRules<KernelType, TreeType
     const double delta = (1 - 0.5 * squaredDist);
     if (kernelEval <= delta)
     {
-      const double gamma = furthestDist * sqrt(1 - 0.25 * squaredDist);
+      const double gamma = furthestDist * std::sqrt(1 - 0.25 * squaredDist);
       maxKernel = kernelEval * delta +
-          gamma * sqrt(1 - std::pow(kernelEval, 2.0));
+          gamma * std::sqrt(1 - std::pow(kernelEval, 2.0));
     }
     else
     {
@@ -370,12 +371,13 @@ double FastMKSRules<KernelType, TreeType
     if (kernelEval <= (1 - 0.5 * bothSqDist))
     {
       const double queryDelta = (1 - 0.5 * querySqDist);
-      const double queryGamma = queryDescDist * sqrt(1 - 0.25 * querySqDist);
+      const double queryGamma = queryDescDist * 
+          std::sqrt(1 - 0.25 * querySqDist);
       const double refDelta = (1 - 0.5 * refSqDist);
-      const double refGamma = refDescDist * sqrt(1 - 0.25 * refSqDist);
+      const double refGamma = refDescDist * std::sqrt(1 - 0.25 * refSqDist);
 
       maxKernel = kernelEval * (queryDelta * refDelta - queryGamma * refGamma) +
-          sqrt(1 - std::pow(kernelEval, 2.0)) *
+          std::sqrt(1 - std::pow(kernelEval, 2.0)) *
           (queryGamma * refDelta + queryDelta * refGamma);
     }
     else
diff -pruN 4.3.0-2/src/mlpack/methods/fastmks/fastmks_stat.hpp 4.4.0-1/src/mlpack/methods/fastmks/fastmks_stat.hpp
--- 4.3.0-2/src/mlpack/methods/fastmks/fastmks_stat.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/fastmks/fastmks_stat.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -61,7 +61,7 @@ class FastMKSStat
       }
       else
       {
-        selfKernel = sqrt(node.Metric().Kernel().Evaluate(
+        selfKernel = std::sqrt(node.Metric().Kernel().Evaluate(
             node.Dataset().col(node.Point(0)),
             node.Dataset().col(node.Point(0))));
       }
@@ -72,7 +72,7 @@ class FastMKSStat
       arma::vec center;
       node.Center(center);
 
-      selfKernel = sqrt(node.Metric().Kernel().Evaluate(center, center));
+      selfKernel = std::sqrt(node.Metric().Kernel().Evaluate(center, center));
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/gmm/diagonal_gmm_impl.hpp 4.4.0-1/src/mlpack/methods/gmm/diagonal_gmm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/gmm/diagonal_gmm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/gmm/diagonal_gmm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -147,7 +147,7 @@ inline double DiagonalGMM::LogProbabilit
   double sum = -std::numeric_limits<double>::infinity();
   for (size_t i = 0; i < gaussians; ++i)
   {
-    sum = LogAdd(sum, log(weights[i]) +
+    sum = LogAdd(sum, std::log(weights[i]) +
         dists[i].LogProbability(observation));
   }
   return sum;
@@ -177,7 +177,7 @@ inline void DiagonalGMM::LogProbability(
   }
 
   // Save log(weights) as a vector.
-  arma::vec logWeights = arma::log(weights);
+  arma::vec logWeights = log(weights);
 
   // Compute log-probability.
   logProb += repmat(logWeights.t(), logProb.n_rows, 1);
@@ -189,7 +189,7 @@ inline void DiagonalGMM::LogProbability(
  */
 inline double DiagonalGMM::Probability(const arma::vec& observation) const
 {
-  return exp(LogProbability(observation));
+  return std::exp(LogProbability(observation));
 }
 
 /**
@@ -215,7 +215,7 @@ inline double DiagonalGMM::LogProbabilit
 {
   // We are only considering one Gaussian component -- so we only need to call
   // Probability() once.  We do consider the prior probability!
-  return log(weights[component]) +
+  return std::log(weights[component]) +
          dists[component].LogProbability(observation);
 }
 
@@ -226,7 +226,7 @@ inline double DiagonalGMM::LogProbabilit
 inline double DiagonalGMM::Probability(const arma::vec& observation,
                                        const size_t component) const
 {
-  return exp(LogProbability(observation, component));
+  return std::exp(LogProbability(observation, component));
 }
 
 /**
@@ -250,8 +250,8 @@ inline arma::vec DiagonalGMM::Random() c
     }
   }
 
-  return arma::sqrt(dists[gaussian].Covariance()) %
-      arma::randn<arma::vec>(dimensionality) + dists[gaussian].Mean();
+  return sqrt(dists[gaussian].Covariance()) %
+      randn<arma::vec>(dimensionality) + dists[gaussian].Mean();
 }
 
 /**
@@ -306,7 +306,7 @@ inline double DiagonalGMM::LogLikelihood
     if (accu(likelihoods.col(j)) == 0)
       Log::Info << "Likelihood of point " << j << " is 0!  It is probably an "
           << "outlier." << std::endl;
-    logLikelihood += log(accu(likelihoods.col(j)));
+    logLikelihood += std::log(accu(likelihoods.col(j)));
   }
 
   return logLikelihood;
diff -pruN 4.3.0-2/src/mlpack/methods/gmm/em_fit_impl.hpp 4.4.0-1/src/mlpack/methods/gmm/em_fit_impl.hpp
--- 4.3.0-2/src/mlpack/methods/gmm/em_fit_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/gmm/em_fit_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -92,7 +92,7 @@ Estimate(const arma::mat& observations,
       // Gaussian.  First we make an alias of the condLogProb vector.
       arma::vec condLogProbAlias = condLogProb.unsafe_col(i);
       dists[i].LogProbability(observations, condLogProbAlias);
-      condLogProbAlias += log(weights[i]);
+      condLogProbAlias += std::log(weights[i]);
     }
 
     // Normalize row-wise.
@@ -118,7 +118,7 @@ Estimate(const arma::mat& observations,
     {
       // Don't update if there's no probability of the Gaussian having points.
       if (probRowSums[i] != -std::numeric_limits<double>::infinity())
-        dists[i].Mean() = observations * arma::exp(condLogProb.col(i) -
+        dists[i].Mean() = observations * exp(condLogProb.col(i) -
                                                    probRowSums[i]);
       else
         continue;
@@ -131,9 +131,9 @@ Estimate(const arma::mat& observations,
       // covariance only with diagonal components.
       if (std::is_same<Distribution, DiagonalGaussianDistribution>::value)
       {
-        arma::vec covariance = arma::sum((tmp % tmp) %
-            (arma::ones<arma::vec>(observations.n_rows) *
-            trans(arma::exp(condLogProb.col(i) - probRowSums[i]))), 1);
+        arma::vec covariance = sum((tmp % tmp) %
+            (ones<arma::vec>(observations.n_rows) *
+            trans(exp(condLogProb.col(i) - probRowSums[i]))), 1);
 
         // Apply covariance constraint.
         constraint.ApplyConstraint(covariance);
@@ -141,7 +141,7 @@ Estimate(const arma::mat& observations,
       }
       else
       {
-        arma::mat tmpB = tmp.each_row() % trans(arma::exp(condLogProb.col(i) -
+        arma::mat tmpB = tmp.each_row() % trans(exp(condLogProb.col(i) -
                                                           probRowSums[i]));
         arma::mat covariance = tmp * trans(tmpB);
 
@@ -153,7 +153,7 @@ Estimate(const arma::mat& observations,
 
     // Calculate the new values for omega using the updated conditional
     // probabilities.
-    weights = arma::exp(probRowSums - std::log(observations.n_cols));
+    weights = exp(probRowSums - std::log(observations.n_cols));
 
     // Update values of l; calculate new log-likelihood.
     lOld = l;
@@ -196,7 +196,7 @@ Estimate(const arma::mat& observations,
       // Gaussian.  First we make an alias of the condLogProb vector.
       arma::vec condLogProbAlias = condLogProb.unsafe_col(i);
       dists[i].LogProbability(observations, condLogProbAlias);
-      condLogProbAlias += log(weights[i]);
+      condLogProbAlias += std::log(weights[i]);
     }
 
     // Normalize row-wise.
@@ -215,7 +215,7 @@ Estimate(const arma::mat& observations,
 
     // Calculate the new value of the means using the updated conditional
     // probabilities.
-    arma::vec logProbabilities = arma::log(probabilities);
+    arma::vec logProbabilities = log(probabilities);
     for (size_t i = 0; i < dists.size(); ++i)
     {
       // Calculate the sum of probabilities of points, which is the
@@ -229,7 +229,7 @@ Estimate(const arma::mat& observations,
       if (probRowSums[i] != -std::numeric_limits<double>::infinity())
       {
         dists[i].Mean() = observations *
-              arma::exp(condLogProb.col(i) + logProbabilities - probRowSums[i]);
+              exp(condLogProb.col(i) + logProbabilities - probRowSums[i]);
       }
       else
         continue;
@@ -242,9 +242,9 @@ Estimate(const arma::mat& observations,
       // covariance only with diagonal components.
       if (std::is_same<Distribution, DiagonalGaussianDistribution>::value)
       {
-        arma::vec cov = arma::sum((tmp % tmp) %
-            (arma::ones<arma::vec>(observations.n_rows) *
-            trans(arma::exp(condLogProb.col(i) +
+        arma::vec cov = sum((tmp % tmp) %
+            (ones<arma::vec>(observations.n_rows) *
+            trans(exp(condLogProb.col(i) +
                             logProbabilities - probRowSums[i]))), 1);
 
         // Apply covariance constraint.
@@ -253,7 +253,7 @@ Estimate(const arma::mat& observations,
       }
       else
       {
-        arma::mat tmpB = tmp.each_row() % trans(arma::exp(condLogProb.col(i) +
+        arma::mat tmpB = tmp.each_row() % trans(exp(condLogProb.col(i) +
             logProbabilities - probRowSums[i]));
         arma::mat cov = (tmp * trans(tmpB));
 
@@ -265,7 +265,7 @@ Estimate(const arma::mat& observations,
 
     // Calculate the new values for omega using the updated conditional
     // probabilities.
-    weights = arma::exp(probRowSums - AccuLog(logProbabilities));
+    weights = exp(probRowSums - AccuLog(logProbabilities));
 
     // Update values of l; calculate new log-likelihood.
     lOld = l;
@@ -386,7 +386,7 @@ LogLikelihood(const arma::mat& observati
   for (size_t i = 0; i < dists.size(); ++i)
   {
     dists[i].LogProbability(observations, logPhis);
-    logLikelihoods.row(i) = log(weights(i)) + trans(logPhis);
+    logLikelihoods.row(i) = std::log(weights(i)) + trans(logPhis);
   }
   // Now sum over every point.
   for (size_t j = 0; j < observations.n_cols; ++j)
diff -pruN 4.3.0-2/src/mlpack/methods/gmm/gmm_impl.hpp 4.4.0-1/src/mlpack/methods/gmm/gmm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/gmm/gmm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/gmm/gmm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,7 +64,7 @@ inline double GMM::LogProbability(const
   // multiply by the prior for each Gaussian too).
   double sum = -std::numeric_limits<double>::infinity();
   for (size_t i = 0; i < gaussians; ++i)
-    sum = LogAdd(sum, log(weights[i]) +
+    sum = LogAdd(sum, std::log(weights[i]) +
         dists[i].LogProbability(observation));
 
   return sum;
@@ -94,7 +94,7 @@ inline void GMM::LogProbability(const ar
   }
 
   // Save log(weights) as a vector.
-  arma::vec logWeights = arma::log(weights);
+  arma::vec logWeights = log(weights);
 
   // Compute log-probability.
   logProb += repmat(logWeights.t(), logProb.n_rows, 1);
@@ -108,7 +108,7 @@ inline void GMM::LogProbability(const ar
  */
 inline double GMM::Probability(const arma::vec& observation) const
 {
-  return exp(LogProbability(observation));
+  return std::exp(LogProbability(observation));
 }
 
 /**
@@ -137,7 +137,8 @@ inline double GMM::LogProbability(const
 {
   // We are only considering one Gaussian component -- so we only need to call
   // Probability() once.  We do consider the prior probability!
-  return log(weights[component]) + dists[component].LogProbability(observation);
+  return std::log(weights[component]) +
+      dists[component].LogProbability(observation);
 }
 
 /**
@@ -150,7 +151,7 @@ inline double GMM::LogProbability(const
 inline double GMM::Probability(const arma::vec& observation,
                                const size_t component) const
 {
-  return exp(LogProbability(observation, component));
+  return std::exp(LogProbability(observation, component));
 }
 
 /**
@@ -180,7 +181,7 @@ inline arma::vec GMM::Random() const
     Log::Fatal << "Cholesky decomposition failed." << std::endl;
   }
   return trans(cholDecomp) *
-      arma::randn<arma::vec>(dimensionality) + dists[gaussian].Mean();
+      randn<arma::vec>(dimensionality) + dists[gaussian].Mean();
 }
 
 /**
@@ -236,7 +237,7 @@ inline double GMM::LogLikelihood(
   for (size_t i = 0; i < gaussians; ++i)
   {
     distsL[i].LogProbability(data, logPhis);
-    logLikelihoods.row(i) = log(weightsL(i)) + trans(logPhis);
+    logLikelihoods.row(i) = std::log(weightsL(i)) + trans(logPhis);
   }
 
   // Now sum over every point.
diff -pruN 4.3.0-2/src/mlpack/methods/hmm/hmm_impl.hpp 4.4.0-1/src/mlpack/methods/hmm/hmm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hmm/hmm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hmm/hmm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,17 +29,17 @@ HMM<Distribution>::HMM(const size_t stat
                        const Distribution emissions,
                        const double tolerance) :
     emission(states, /* default distribution */ emissions),
-    transitionProxy(arma::randu<arma::mat>(states, states)),
-    initialProxy(arma::randu<arma::vec>(states) / (double) states),
+    transitionProxy(randu<arma::mat>(states, states)),
+    initialProxy(randu<arma::vec>(states) / (double) states),
     dimensionality(emissions.Dimensionality()),
     tolerance(tolerance),
     recalculateInitial(false),
     recalculateTransition(false)
 {
   // Normalize the transition probabilities and initial state probabilities.
-  initialProxy /= arma::accu(initialProxy);
+  initialProxy /= accu(initialProxy);
   for (size_t i = 0; i < transitionProxy.n_cols; ++i)
-    transitionProxy.col(i) /= arma::accu(transitionProxy.col(i));
+    transitionProxy.col(i) /= accu(transitionProxy.col(i));
 
   logTransition = log(transitionProxy);
   logInitial = log(initialProxy);
@@ -193,7 +193,7 @@ double HMM<Distribution>::Train(const st
 
         // Add to list of emission observations, for Distribution::Train().
         for (size_t j = 0; j < logTransition.n_cols; ++j)
-          emissionProb[j][sumTime] = exp(stateLogProb(j, t));
+          emissionProb[j][sumTime] = std::exp(stateLogProb(j, t));
         emissionList.col(sumTime) = dataSeq[seq].col(t);
         sumTime++;
       }
@@ -226,7 +226,7 @@ double HMM<Distribution>::Train(const st
       if (std::isfinite(sum))
         logTransition.col(i) -= sum;
       else
-        logTransition.col(i).fill(-log((double) logTransition.n_rows));
+        logTransition.col(i).fill(-std::log((double) logTransition.n_rows));
     }
 
     initialProxy = exp(logInitial);
@@ -257,9 +257,8 @@ void HMM<Distribution>::Train(const std:
         << ")." << std::endl;
   }
 
-  arma::mat initial = arma::zeros(logInitial.n_elem);
-  arma::mat transition = arma::zeros(logTransition.n_rows,
-                                     logTransition.n_cols);
+  arma::mat initial = zeros(logInitial.n_elem);
+  arma::mat transition = zeros(logTransition.n_rows, logTransition.n_cols);
 
   // Estimate the transition and emission matrices directly from the
   // observations.  The emission list holds the time indices for observations
@@ -463,7 +462,7 @@ void HMM<Distribution>::Generate(const s
     double probSum = 0;
     for (size_t st = 0; st < logTransition.n_rows; st++)
     {
-      probSum += exp(logTransition(st, stateSequence[t - 1]));
+      probSum += std::exp(logTransition(st, stateSequence[t - 1]));
       if (randValue <= probSum)
       {
         stateSequence[t] = st;
diff -pruN 4.3.0-2/src/mlpack/methods/hmm/hmm_train_main.cpp 4.4.0-1/src/mlpack/methods/hmm/hmm_train_main.cpp
--- 4.3.0-2/src/mlpack/methods/hmm/hmm_train_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hmm/hmm_train_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -123,7 +123,7 @@ struct Init
          ++it)
     {
       arma::Col<size_t> maxSeqs =
-          arma::conv_to<arma::Col<size_t>>::from(arma::max(*it, 1)) + 1;
+          ConvTo<arma::Col<size_t>>::From(arma::max(*it, 1)) + 1;
       maxEmissions = arma::max(maxEmissions, maxSeqs);
     }
 
@@ -234,7 +234,7 @@ struct Init
     for (size_t i = 0; i < e.size(); ++i)
     {
       e[i].Probabilities().randu();
-      e[i].Probabilities() /= arma::accu(e[i].Probabilities());
+      e[i].Probabilities() /= accu(e[i].Probabilities());
     }
   }
 
@@ -247,7 +247,8 @@ struct Init
       const size_t dimensionality = e[i].Mean().n_rows;
       e[i].Mean().randu();
       // Generate random covariance.
-      arma::mat r = arma::randu<arma::mat>(dimensionality, dimensionality);
+      arma::mat r;
+      r.randu(dimensionality, dimensionality);
       e[i].Covariance(r * r.t());
     }
   }
@@ -260,7 +261,7 @@ struct Init
     {
       // Random weights.
       e[i].Weights().randu();
-      e[i].Weights() /= arma::accu(e[i].Weights());
+      e[i].Weights() /= accu(e[i].Weights());
 
       // Random means and covariances.
       for (int g = 0; g < params.Get<int>("gaussians"); ++g)
@@ -269,8 +270,8 @@ struct Init
         e[i].Component(g).Mean().randu();
 
         // Generate random covariance.
-        arma::mat r = arma::randu<arma::mat>(dimensionality,
-            dimensionality);
+        arma::mat r;
+        r.randu(dimensionality, dimensionality);
         e[i].Component(g).Covariance(r * r.t());
       }
     }
@@ -284,7 +285,7 @@ struct Init
     {
       // Random weights.
       e[i].Weights().randu();
-      e[i].Weights() /= arma::accu(e[i].Weights());
+      e[i].Weights() /= accu(e[i].Weights());
 
       // Random means and covariances.
       for (int g = 0; g < params.Get<int>("gaussians"); ++g)
@@ -293,7 +294,8 @@ struct Init
         e[i].Component(g).Mean().randu();
 
         // Generate random diagonal covariance.
-        arma::vec r = arma::randu<arma::vec>(dimensionality);
+        arma::vec r;
+        r.randu(dimensionality);
         e[i].Component(g).Covariance(r);
       }
     }
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/binary_numeric_split_impl.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/binary_numeric_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/binary_numeric_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/binary_numeric_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -164,7 +164,7 @@ template<typename FitnessFunction, typen
 double BinaryNumericSplit<FitnessFunction, ObservationType>::
     MajorityProbability() const
 {
-  return double(arma::max(classCounts)) / double(arma::accu(classCounts));
+  return double(max(classCounts)) / double(accu(classCounts));
 }
 
 template<typename FitnessFunction, typename ObservationType>
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/gini_impurity.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/gini_impurity.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/gini_impurity.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/gini_impurity.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ class GiniImpurity
     arma::vec splitCounts(counts.n_cols);
     for (size_t i = 0; i < counts.n_cols; ++i)
     {
-      splitCounts[i] = arma::accu(counts.col(i));
+      splitCounts[i] = accu(counts.col(i));
       numElem += splitCounts[i];
     }
 
@@ -37,7 +37,7 @@ class GiniImpurity
     if (numElem == 0)
       return 0.0;
 
-    arma::Col<size_t> classCounts = arma::sum(counts, 1);
+    arma::Col<size_t> classCounts = sum(counts, 1);
 
     // Calculate the Gini impurity of the un-split node.
     double impurity = 0.0;
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_categorical_split_impl.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_categorical_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_categorical_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_categorical_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -77,7 +77,7 @@ template<typename FitnessFunction>
 size_t HoeffdingCategoricalSplit<FitnessFunction>::MajorityClass() const
 {
   // Calculate the class that we have seen the most of.
-  arma::Col<size_t> classCounts = arma::sum(sufficientStatistics, 1);
+  arma::Col<size_t> classCounts = sum(sufficientStatistics, 1);
 
   arma::uword maxIndex = 0;
   classCounts.max(maxIndex);
@@ -88,9 +88,9 @@ size_t HoeffdingCategoricalSplit<Fitness
 template<typename FitnessFunction>
 double HoeffdingCategoricalSplit<FitnessFunction>::MajorityProbability() const
 {
-  arma::Col<size_t> classCounts = arma::sum(sufficientStatistics, 1);
+  arma::Col<size_t> classCounts = sum(sufficientStatistics, 1);
 
-  return double(classCounts.max()) / double(arma::accu(classCounts));
+  return double(classCounts.max()) / double(accu(classCounts));
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -146,6 +146,10 @@ template<typename FitnessFunction>
 using HoeffdingDoubleNumericSplit = HoeffdingNumericSplit<FitnessFunction,
     double>;
 
+template<typename FitnessFunction>
+using HoeffdingFloatNumericSplit = HoeffdingNumericSplit<FitnessFunction,
+    float>;
+
 } // namespace mlpack
 
 // Include implementation.
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split_impl.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_numeric_split_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ HoeffdingNumericSplit<FitnessFunction, O
     bins(bins),
     observationsBeforeBinning(observationsBeforeBinning),
     samplesSeen(0),
-    sufficientStatistics(arma::zeros<arma::Mat<size_t>>(numClasses, bins))
+    sufficientStatistics(zeros<arma::Mat<size_t>>(numClasses, bins))
 {
   observations.zeros();
   labels.zeros();
@@ -41,7 +41,7 @@ HoeffdingNumericSplit<FitnessFunction, O
     bins(other.bins),
     observationsBeforeBinning(other.observationsBeforeBinning),
     samplesSeen(0),
-    sufficientStatistics(arma::zeros<arma::Mat<size_t>>(numClasses, bins))
+    sufficientStatistics(zeros<arma::Mat<size_t>>(numClasses, bins))
 {
   observations.zeros();
   labels.zeros();
@@ -152,7 +152,7 @@ size_t HoeffdingNumericSplit<FitnessFunc
   {
     // We've calculated the bins, so we can just sum over the sufficient
     // statistics.
-    arma::Col<size_t> classCounts = arma::sum(sufficientStatistics, 1);
+    arma::Col<size_t> classCounts = sum(sufficientStatistics, 1);
 
     arma::uword maxIndex = 0;
     classCounts.max(maxIndex);
@@ -173,15 +173,15 @@ double HoeffdingNumericSplit<FitnessFunc
     for (size_t i = 0; i < samplesSeen; ++i)
       classes[labels[i]]++;
 
-    return double(classes.max()) / double(arma::accu(classes));
+    return double(classes.max()) / double(accu(classes));
   }
   else
   {
     // We've calculated the bins, so we can just sum over the sufficient
     // statistics.
-    arma::Col<size_t> classCounts = arma::sum(sufficientStatistics, 1);
+    arma::Col<size_t> classCounts = sum(sufficientStatistics, 1);
 
-    return double(classCounts.max()) / double(arma::sum(classCounts));
+    return double(classCounts.max()) / double(sum(classCounts));
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,35 +73,33 @@ class HoeffdingTree
   typedef CategoricalSplitType<FitnessFunction> CategoricalSplit;
 
   /**
-   * Construct the Hoeffding tree with the given parameters and given training
-   * data.  The tree may be trained either in batch mode (which looks at all
-   * points before splitting, and propagates these points to the created
-   * children for further training), or in streaming mode, where each point is
-   * only considered once.  (In general, batch mode will give better-performing
-   * trees, but will have higher memory and runtime costs for the same dataset.)
+   * Construct a Hoeffding tree with no data and no information.  Be sure to
+   * call Train() before trying to use the tree.
+   */
+  HoeffdingTree();
+
+  /**
+   * Construct the Hoeffding tree with the given parameters for training on
+   * numerical data, but training on no data.  The dimensionMappings parameter
+   * is only used if it is desired that this node does not create its own
+   * dimensionMappings object (for instance, if this is a child of another node
+   * in the tree).
    *
-   * @param data Dataset to train on.
-   * @param datasetInfo Information on the dataset (types of each feature).
-   * @param labels Labels of each point in the dataset.
    * @param numClasses Number of classes in the dataset.
-   * @param batchTraining Whether or not to train in batch.
    * @param successProbability Probability of success required in Hoeffding
-   *      bounds before a split can happen.
-   * @param maxSamples Maximum number of samples before a split is forced (0
-   *      never forces a split); ignored in batch training mode.
-   * @param checkInterval Number of samples required before each split; ignored
-   *      in batch training mode.
+   *      bound before a split can happen.
+   * @param maxSamples Maximum number of samples before a split is forced.
+   * @param checkInterval Number of samples required before each split check.
    * @param minSamples If the node has seen this many points or fewer, no split
    *      will be allowed.
    * @param categoricalSplitIn Optional instantiated categorical split object.
    * @param numericSplitIn Optional instantiated numeric split object.
+   * @param dimensionMappings Mappings from dimension indices to positions in
+   *      numeric and categorical split vectors.  If left NULL, a new one will
+   *      be created.
    */
-  template<typename MatType>
-  HoeffdingTree(const MatType& data,
-                const data::DatasetInfo& datasetInfo,
-                const arma::Row<size_t>& labels,
+  HoeffdingTree(const size_t dimensionality,
                 const size_t numClasses,
-                const bool batchTraining = true,
                 const double successProbability = 0.95,
                 const size_t maxSamples = 0,
                 const size_t checkInterval = 100,
@@ -109,7 +107,9 @@ class HoeffdingTree
                 const CategoricalSplitType<FitnessFunction>& categoricalSplitIn
                     = CategoricalSplitType<FitnessFunction>(0, 0),
                 const NumericSplitType<FitnessFunction>& numericSplitIn =
-                    NumericSplitType<FitnessFunction>(0));
+                    NumericSplitType<FitnessFunction>(0),
+                std::unordered_map<size_t, std::pair<size_t, size_t>>*
+                    dimensionMappings = NULL);
 
   /**
    * Construct the Hoeffding tree with the given parameters, but training on no
@@ -117,21 +117,21 @@ class HoeffdingTree
    * this node does not create its own dimensionMappings object (for instance,
    * if this is a child of another node in the tree).
    *
-   * @param numClasses Number of classes in the dataset.
    * @param datasetInfo Information on the dataset (types of each feature).
+   * @param numClasses Number of classes in the dataset.
    * @param successProbability Probability of success required in Hoeffding
    *      bound before a split can happen.
    * @param maxSamples Maximum number of samples before a split is forced.
    * @param checkInterval Number of samples required before each split check.
    * @param minSamples If the node has seen this many points or fewer, no split
    *      will be allowed.
+   * @param categoricalSplitIn Optional instantiated categorical split object.
+   * @param numericSplitIn Optional instantiated numeric split object.
    * @param dimensionMappings Mappings from dimension indices to positions in
    *      numeric and categorical split vectors.  If left NULL, a new one will
    *      be created.
    * @param copyDatasetInfo If true, then a copy of the datasetInfo will be
    *      made.
-   * @param categoricalSplitIn Optional instantiated categorical split object.
-   * @param numericSplitIn Optional instantiated numeric split object.
    */
   HoeffdingTree(const data::DatasetInfo& datasetInfo,
                 const size_t numClasses,
@@ -148,10 +148,81 @@ class HoeffdingTree
                 const bool copyDatasetInfo = true);
 
   /**
-   * Construct a Hoeffding tree with no data and no information.  Be sure to
-   * call Train() before trying to use the tree.
+   * Construct the Hoeffding tree with the given parameters and given training
+   * data, where the training data contains only numerical features.  The tree
+   * may be trained either in batch mode (which looks at all points before
+   * splitting, and propagates these points to the created children for further
+   * training), or in streaming mode, where each point is only considered once.
+   * (In general, batch mode will give better-performing trees, but will have
+   * higher memory and runtime costs for the same dataset.)
+   *
+   * @param data Dataset to train on.
+   * @param labels Labels of each point in the dataset.
+   * @param numClasses Number of classes in the dataset.
+   * @param batchTraining Whether or not to train in batch.
+   * @param successProbability Probability of success required in Hoeffding
+   *      bounds before a split can happen.
+   * @param maxSamples Maximum number of samples before a split is forced (0
+   *      never forces a split); ignored in batch training mode.
+   * @param checkInterval Number of samples required before each split; ignored
+   *      in batch training mode.
+   * @param minSamples If the node has seen this many points or fewer, no split
+   *      will be allowed.
+   * @param categoricalSplitIn Optional instantiated categorical split object.
+   * @param numericSplitIn Optional instantiated numeric split object.
    */
-  HoeffdingTree();
+  template<typename MatType>
+  HoeffdingTree(const MatType& data,
+                const arma::Row<size_t>& labels,
+                const size_t numClasses,
+                const bool batchTraining = true,
+                const double successProbability = 0.95,
+                const size_t maxSamples = 0,
+                const size_t checkInterval = 100,
+                const size_t minSamples = 100,
+                const CategoricalSplitType<FitnessFunction>& categoricalSplitIn
+                    = CategoricalSplitType<FitnessFunction>(0, 0),
+                const NumericSplitType<FitnessFunction>& numericSplitIn =
+                    NumericSplitType<FitnessFunction>(0));
+
+  /**
+   * Construct the Hoeffding tree with the given parameters and given training
+   * data.  The tree may be trained either in batch mode (which looks at all
+   * points before splitting, and propagates these points to the created
+   * children for further training), or in streaming mode, where each point is
+   * only considered once.  (In general, batch mode will give better-performing
+   * trees, but will have higher memory and runtime costs for the same dataset.)
+   *
+   * @param data Dataset to train on.
+   * @param datasetInfo Information on the dataset (types of each feature).
+   * @param labels Labels of each point in the dataset.
+   * @param numClasses Number of classes in the dataset.
+   * @param batchTraining Whether or not to train in batch.
+   * @param successProbability Probability of success required in Hoeffding
+   *      bounds before a split can happen.
+   * @param maxSamples Maximum number of samples before a split is forced (0
+   *      never forces a split); ignored in batch training mode.
+   * @param checkInterval Number of samples required before each split; ignored
+   *      in batch training mode.
+   * @param minSamples If the node has seen this many points or fewer, no split
+   *      will be allowed.
+   * @param categoricalSplitIn Optional instantiated categorical split object.
+   * @param numericSplitIn Optional instantiated numeric split object.
+   */
+  template<typename MatType>
+  HoeffdingTree(const MatType& data,
+                const data::DatasetInfo& datasetInfo,
+                const arma::Row<size_t>& labels,
+                const size_t numClasses,
+                const bool batchTraining = true,
+                const double successProbability = 0.95,
+                const size_t maxSamples = 0,
+                const size_t checkInterval = 100,
+                const size_t minSamples = 100,
+                const CategoricalSplitType<FitnessFunction>& categoricalSplitIn
+                    = CategoricalSplitType<FitnessFunction>(0, 0),
+                const NumericSplitType<FitnessFunction>& numericSplitIn =
+                    NumericSplitType<FitnessFunction>(0));
 
   /**
    * Copy another tree (warning: this will duplicate the tree entirely, and may
@@ -194,44 +265,89 @@ class HoeffdingTree
    *
    * Note that the tree will be automatically reset if the dimensionality of
    * `data` does not match the dimensionality that the tree was currently
-   * trained with.  The tree will also be reset if `numClasses` is passed.
+   * trained with.  The tree will also be reset if `numClasses` is passed and
+   * differs from the existing setting.
    *
    * @param data Data points to train on.
    * @param labels Labels of data points.
-   * @param batchTraining If true, perform training in batch.
-   * @param resetTree If true, reset the tree to an empty tree before training.
    * @param numClasses The number of classes in `labels`.  Passing this will
    *      reset the tree.  If not given and `resetTree` is `true`, then the
    *      number of classes will be computed from `labels`.
+   * @param batchTraining If true, perform training in batch.
+   * @param successProbability Probability of success required in Hoeffding
+   *      bounds before a split can happen.
+   * @param maxSamples Maximum number of samples before a split is forced (0
+   *      never forces a split); ignored in batch training mode.
+   * @param checkInterval Number of samples required before each split; ignored
+   *      in batch training mode.
+   * @param minSamples If the node has seen this many points or fewer, no split
+   *      will be allowed.
    */
   template<typename MatType>
   void Train(const MatType& data,
              const arma::Row<size_t>& labels,
+             const size_t numClasses = 0,
              const bool batchTraining = true,
-             const bool resetTree = false,
-             const size_t numClasses = 0);
+             const std::optional<double> successProbability = std::nullopt,
+             const std::optional<size_t> maxSamples = std::nullopt,
+             const std::optional<size_t> checkInterval = std::nullopt);
+
+  template<typename MatType>
+  void Train(const MatType& data,
+             const arma::Row<size_t>& labels,
+             const size_t numClasses,
+             const bool batchTraining,
+             const double successProbability,
+             const size_t maxSamples,
+             const size_t checkInterval,
+             const size_t minSamples);
 
   /**
    * Train on a set of points, either in streaming mode or in batch mode, with
-   * the given labels and the given `DatasetInfo`.  This will reset the tree.
-   * This only needs to be called when the `DatasetInfo` has changed---if you
-   * are training incrementally but have already passed the DatasetInfo once,
-   * use the overload of `Train()` that does not take a `DatasetInfo` and make
-   * sure `resetTree` is set to `false`.
+   * the given labels.  If `resetTree` is set to `true`, then reset the state of
+   * the tree to an empty tree before training.
+   *
+   * Note that the tree will be automatically reset if the dimensionality of
+   * `data` does not match the dimensionality that the tree was currently
+   * trained with.  The tree will also be reset if `numClasses` is passed and
+   * differs from the existing setting.
    *
    * @param data Data points to train on.
-   * @param info DatasetInfo object with information about each dimension.
+   * @param datasetInfo Information on the dataset (types of each feature).
    * @param labels Labels of data points.
+   * @param numClasses The number of classes in `labels`.  Passing this will
+   *      reset the tree.  If not given and `resetTree` is `true`, then the
+   *      number of classes will be computed from `labels`.
    * @param batchTraining If true, perform training in batch.
-   * @param numClasses Number of classes in `labels`.  If not specified, it is
-   *      computed from `labels`.
+   * @param successProbability Probability of success required in Hoeffding
+   *      bounds before a split can happen.
+   * @param maxSamples Maximum number of samples before a split is forced (0
+   *      never forces a split); ignored in batch training mode.
+   * @param checkInterval Number of samples required before each split; ignored
+   *      in batch training mode.
+   * @param minSamples If the node has seen this many points or fewer, no split
+   *      will be allowed.
    */
   template<typename MatType>
   void Train(const MatType& data,
              const data::DatasetInfo& info,
              const arma::Row<size_t>& labels,
+             const size_t numClasses = 0,
              const bool batchTraining = true,
-             const size_t numClasses = 0);
+             const std::optional<double> successProbability = std::nullopt,
+             const std::optional<size_t> maxSamples = std::nullopt,
+             const std::optional<size_t> checkInterval = std::nullopt);
+
+  template<typename MatType>
+  void Train(const MatType& data,
+             const data::DatasetInfo& info,
+             const arma::Row<size_t>& labels,
+             const size_t numClasses,
+             const bool batchTraining,
+             const double successProbability,
+             const size_t maxSamples,
+             const size_t checkInterval,
+             const size_t minSamples);
 
   /**
    * Train on a single point in streaming mode, with the given label.  The tree
@@ -291,6 +407,12 @@ class HoeffdingTree
   //! Modify the number of samples before a split check is performed.
   void CheckInterval(const size_t checkInterval);
 
+  //! Get the number of points seen so far.
+  size_t NumSamples() const { return numSamples; }
+
+  //! Get the number of classes the tree is trained on.
+  size_t NumClasses() const { return numClasses; }
+
   /**
    * Given a point and that this node is not a leaf, calculate the index of the
    * child node this point would go towards.  This method is primarily used by
@@ -301,6 +423,9 @@ class HoeffdingTree
   template<typename VecType>
   size_t CalculateDirection(const VecType& point) const;
 
+  //! Get the size of the Hoeffding Tree.
+  size_t NumDescendants() const;
+
   /**
    * Classify the given point, using this node and the entire (sub)tree beneath
    * it.  The predicted label is returned.
@@ -311,9 +436,6 @@ class HoeffdingTree
   template<typename VecType>
   size_t Classify(const VecType& point) const;
 
-  //! Get the size of the Hoeffding Tree.
-  size_t NumDescendants() const;
-
   /**
    * Classify the given point and also return an estimate of the probability
    * that the prediction is correct.  (This estimate is simply the probability
@@ -360,6 +482,23 @@ class HoeffdingTree
    */
   void CreateChildren();
 
+  /**
+   * Reset the tree, keeping the number of classes and dimension information
+   * intact.
+   */
+  void Reset();
+
+  /**
+   * Reset the tree, setting a new number of classes and dimensionality.  This
+   * assumes all dimensions are numeric.
+   */
+  void Reset(const size_t dimensionality, const size_t numClasses);
+
+  /**
+   * Reset the tree, setting a new number of classes and a new datasetInfo.
+   */
+  void Reset(const data::DatasetInfo& datasetInfo, const size_t numClasses);
+
   //! Serialize the split.
   template<typename Archive>
   void serialize(Archive& ar, const uint32_t /* version */);
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_impl.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -14,38 +14,64 @@
 
 // In case it hasn't been included yet.
 #include "hoeffding_tree.hpp"
-#include <stack>
 
 namespace mlpack {
 
 template<typename FitnessFunction,
          template<typename> class NumericSplitType,
          template<typename> class CategoricalSplitType>
-template<typename MatType>
 HoeffdingTree<
     FitnessFunction,
     NumericSplitType,
     CategoricalSplitType
->::HoeffdingTree(const MatType& data,
-                 const data::DatasetInfo& datasetInfoIn,
-                 const arma::Row<size_t>& labels,
+>::HoeffdingTree() :
+    dimensionMappings(
+        new std::unordered_map<size_t, std::pair<size_t, size_t>>()),
+    ownsMappings(true),
+    numSamples(0),
+    numClasses(0),
+    maxSamples(size_t(-1)),
+    checkInterval(100),
+    minSamples(100),
+    datasetInfo(new data::DatasetInfo()),
+    ownsInfo(true),
+    successProbability(0.95),
+    splitDimension(size_t(-1)),
+    majorityClass(0),
+    majorityProbability(0.0),
+    categoricalSplit(0),
+    numericSplit()
+{
+  // Nothing to do.
+}
+
+template<typename FitnessFunction,
+         template<typename> class NumericSplitType,
+         template<typename> class CategoricalSplitType>
+HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::HoeffdingTree(const size_t dimensionality,
                  const size_t numClasses,
-                 const bool batchTraining,
                  const double successProbability,
                  const size_t maxSamples,
                  const size_t checkInterval,
                  const size_t minSamples,
                  const CategoricalSplitType<FitnessFunction>&
                      categoricalSplitIn,
-                 const NumericSplitType<FitnessFunction>& numericSplitIn) :
-    dimensionMappings(NULL),
-    ownsMappings(false),
+                 const NumericSplitType<FitnessFunction>& numericSplitIn,
+                 std::unordered_map<size_t, std::pair<size_t, size_t>>*
+                     dimensionMappingsIn) :
+    dimensionMappings((dimensionMappingsIn != NULL) ? dimensionMappingsIn :
+        new std::unordered_map<size_t, std::pair<size_t, size_t>>()),
+    ownsMappings(dimensionMappingsIn == NULL),
     numSamples(0),
     numClasses(numClasses),
     maxSamples((maxSamples == 0) ? size_t(-1) : maxSamples),
     checkInterval(checkInterval),
     minSamples(minSamples),
-    datasetInfo(new data::DatasetInfo(datasetInfoIn)),
+    datasetInfo(new data::DatasetInfo(dimensionality)),
     ownsInfo(true),
     successProbability(successProbability),
     splitDimension(size_t(-1)),
@@ -54,11 +80,20 @@ HoeffdingTree<
     categoricalSplit(0),
     numericSplit()
 {
-  // Reset the tree.
-  ResetTree(categoricalSplitIn, numericSplitIn);
-
-  // Now train.
-  Train(data, labels, batchTraining);
+  // Do we need to generate the mappings too?
+  if (ownsMappings)
+  {
+    ResetTree(categoricalSplitIn, numericSplitIn);
+  }
+  else
+  {
+    // All dimensions are numeric.
+    for (size_t i = 0; i < datasetInfo->Dimensionality(); ++i)
+    {
+      numericSplits.push_back(NumericSplitType<FitnessFunction>(numClasses,
+          numericSplitIn));
+    }
+  }
 }
 
 template<typename FitnessFunction,
@@ -124,29 +159,86 @@ HoeffdingTree<
 template<typename FitnessFunction,
          template<typename> class NumericSplitType,
          template<typename> class CategoricalSplitType>
+template<typename MatType>
 HoeffdingTree<
     FitnessFunction,
     NumericSplitType,
     CategoricalSplitType
->::HoeffdingTree() :
-    dimensionMappings(
-        new std::unordered_map<size_t, std::pair<size_t, size_t>>()),
-    ownsMappings(true),
+>::HoeffdingTree(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const bool batchTraining,
+                 const double successProbability,
+                 const size_t maxSamples,
+                 const size_t checkInterval,
+                 const size_t minSamples,
+                 const CategoricalSplitType<FitnessFunction>&
+                     categoricalSplitIn,
+                 const NumericSplitType<FitnessFunction>& numericSplitIn) :
+    dimensionMappings(NULL),
+    ownsMappings(false),
     numSamples(0),
-    numClasses(0),
-    maxSamples(size_t(-1)),
-    checkInterval(100),
-    minSamples(100),
-    datasetInfo(new data::DatasetInfo()),
+    numClasses(numClasses),
+    maxSamples((maxSamples == 0) ? size_t(-1) : maxSamples),
+    checkInterval(checkInterval),
+    minSamples(minSamples),
+    datasetInfo(new data::DatasetInfo(data.n_rows)),
     ownsInfo(true),
-    successProbability(0.95),
+    successProbability(successProbability),
     splitDimension(size_t(-1)),
     majorityClass(0),
     majorityProbability(0.0),
     categoricalSplit(0),
     numericSplit()
 {
-  // Nothing to do.
+  // Reset the tree.
+  ResetTree(categoricalSplitIn, numericSplitIn);
+
+  // Now train.
+  Train(data, labels, numClasses, batchTraining);
+}
+
+template<typename FitnessFunction,
+         template<typename> class NumericSplitType,
+         template<typename> class CategoricalSplitType>
+template<typename MatType>
+HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::HoeffdingTree(const MatType& data,
+                 const data::DatasetInfo& datasetInfoIn,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const bool batchTraining,
+                 const double successProbability,
+                 const size_t maxSamples,
+                 const size_t checkInterval,
+                 const size_t minSamples,
+                 const CategoricalSplitType<FitnessFunction>&
+                     categoricalSplitIn,
+                 const NumericSplitType<FitnessFunction>& numericSplitIn) :
+    dimensionMappings(NULL),
+    ownsMappings(false),
+    numSamples(0),
+    numClasses(numClasses),
+    maxSamples((maxSamples == 0) ? size_t(-1) : maxSamples),
+    checkInterval(checkInterval),
+    minSamples(minSamples),
+    datasetInfo(new data::DatasetInfo(datasetInfoIn)),
+    ownsInfo(true),
+    successProbability(successProbability),
+    splitDimension(size_t(-1)),
+    majorityClass(0),
+    majorityProbability(0.0),
+    categoricalSplit(0),
+    numericSplit()
+{
+  // Reset the tree.
+  ResetTree(categoricalSplitIn, numericSplitIn);
+
+  // Now train.
+  Train(data, labels, numClasses, batchTraining);
 }
 
 // Copy constructor.
@@ -336,7 +428,6 @@ HoeffdingTree<FitnessFunction, NumericSp
     delete children[i];
 }
 
-//! Train on a set of points.
 template<typename FitnessFunction,
          template<typename> class NumericSplitType,
          template<typename> class CategoricalSplitType>
@@ -347,14 +438,46 @@ void HoeffdingTree<
     CategoricalSplitType
 >::Train(const MatType& data,
          const arma::Row<size_t>& labels,
+         const size_t numClasses,
+         const bool batchTraining,
+         const std::optional<double> successProbability,
+         const std::optional<size_t> maxSamples,
+         const std::optional<size_t> checkInterval)
+{
+  Train(data, labels, numClasses, batchTraining,
+      (successProbability.has_value()) ? successProbability.value() :
+          this->successProbability,
+      (maxSamples.has_value()) ? maxSamples.value() : this->maxSamples,
+      (checkInterval.has_value()) ? checkInterval.value() : this->checkInterval,
+          this->minSamples);
+}
+
+template<typename FitnessFunction,
+         template<typename> class NumericSplitType,
+         template<typename> class CategoricalSplitType>
+template<typename MatType>
+void HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::Train(const MatType& data,
+         const arma::Row<size_t>& labels,
+         const size_t numClasses,
          const bool batchTraining,
-         const bool resetTree,
-         const size_t numClassesIn)
+         const double successProbability,
+         const size_t maxSamples,
+         const size_t checkInterval,
+         const size_t minSamples)
 {
+  this->successProbability = successProbability;
+  this->maxSamples = maxSamples;
+  this->checkInterval = checkInterval;
+  this->minSamples = minSamples;
+
   // We need to reset the tree either if the user asked for it, or if they
   // passed data whose dimensionality is different than our datasetInfo object.
-  if (resetTree || data.n_rows != datasetInfo->Dimensionality() ||
-      numClassesIn != 0)
+  if (data.n_rows != datasetInfo->Dimensionality() ||
+      (numClasses != 0 && numClasses != this->numClasses))
   {
     // Create a new datasetInfo, which assumes that all features are numeric.
     if (ownsInfo)
@@ -363,7 +486,14 @@ void HoeffdingTree<
     ownsInfo = true;
 
     // Set the number of classes correctly.
-    numClasses = (numClassesIn != 0) ? numClassesIn : arma::max(labels) + 1;
+    if (numClasses != 0)
+      this->numClasses = numClasses;
+
+    if (this->numClasses == 0)
+    {
+      throw std::invalid_argument("HoeffdingTree::Train(): must specify number "
+          "of classes!");
+    }
 
     ResetTree();
   }
@@ -371,7 +501,6 @@ void HoeffdingTree<
   TrainInternal(data, labels, batchTraining);
 }
 
-//! Train on a set of points.
 template<typename FitnessFunction,
          template<typename> class NumericSplitType,
          template<typename> class CategoricalSplitType>
@@ -383,21 +512,69 @@ void HoeffdingTree<
 >::Train(const MatType& data,
          const data::DatasetInfo& info,
          const arma::Row<size_t>& labels,
+         const size_t numClasses,
+         const bool batchTraining,
+         const std::optional<double> successProbability,
+         const std::optional<size_t> maxSamples,
+         const std::optional<size_t> checkInterval)
+{
+  Train(data, info, labels, numClasses, batchTraining,
+      (successProbability.has_value()) ? successProbability.value() :
+          this->successProbability,
+      (maxSamples.has_value()) ? maxSamples.value() : this->maxSamples,
+      (checkInterval.has_value()) ? checkInterval.value() : this->checkInterval,
+          this->minSamples);
+}
+
+template<typename FitnessFunction,
+         template<typename> class NumericSplitType,
+         template<typename> class CategoricalSplitType>
+template<typename MatType>
+void HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::Train(const MatType& data,
+         const data::DatasetInfo& info,
+         const arma::Row<size_t>& labels,
+         const size_t numClasses,
          const bool batchTraining,
-         const size_t numClassesIn)
+         const double successProbability,
+         const size_t maxSamples,
+         const size_t checkInterval,
+         const size_t minSamples)
 {
-  // Take over new DatasetInfo.
-  if (ownsInfo)
-    delete datasetInfo;
-  datasetInfo = &info;
-  ownsInfo = false;
+  this->successProbability = successProbability;
+  this->maxSamples = maxSamples;
+  this->checkInterval = checkInterval;
+  this->minSamples = minSamples;
 
-  // Set the number of classes correctly.
-  numClasses = (numClassesIn != 0) ? numClassesIn : arma::max(labels) + 1;
+  // We need to reset the tree either if the user asked for it, or if they
+  // passed data whose dimensionality is different than our datasetInfo object.
+  if (data.n_rows != datasetInfo->Dimensionality() ||
+      (numClasses != 0 && numClasses != this->numClasses))
+  {
+    // Set the number of classes correctly.
+    if (numClasses != 0)
+      this->numClasses = numClasses;
 
-  ResetTree();
+    if (this->numClasses == 0)
+    {
+      throw std::invalid_argument("HoeffdingTree::Train(): must specify number "
+          "of classes!");
+    }
+
+    Reset(info, this->numClasses);
+  }
+  else if (datasetInfo != &info)
+  {
+    // Take over new DatasetInfo.
+    if (ownsInfo)
+      delete datasetInfo;
+    datasetInfo = &info;
+    ownsInfo = false;
+  }
 
-  // Now train.
   TrainInternal(data, labels, batchTraining);
 }
 
@@ -817,6 +994,62 @@ template<
     template<typename> class NumericSplitType,
     template<typename> class CategoricalSplitType
 >
+void HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::Reset()
+{
+  ResetTree();
+}
+
+template<
+    typename FitnessFunction,
+    template<typename> class NumericSplitType,
+    template<typename> class CategoricalSplitType
+>
+void HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::Reset(const size_t dimensionality, const size_t numClasses)
+{
+  if (ownsInfo)
+    delete datasetInfo;
+  datasetInfo = new data::DatasetInfo(dimensionality); // All features numeric.
+  ownsInfo = true;
+
+  this->numClasses = numClasses;
+
+  ResetTree();
+}
+
+template<
+    typename FitnessFunction,
+    template<typename> class NumericSplitType,
+    template<typename> class CategoricalSplitType
+>
+void HoeffdingTree<
+    FitnessFunction,
+    NumericSplitType,
+    CategoricalSplitType
+>::Reset(const data::DatasetInfo& info, const size_t numClasses)
+{
+  if (ownsInfo)
+    delete datasetInfo;
+  datasetInfo = &info;
+  ownsInfo = false;
+
+  this->numClasses = numClasses;
+
+  ResetTree();
+}
+
+template<
+    typename FitnessFunction,
+    template<typename> class NumericSplitType,
+    template<typename> class CategoricalSplitType
+>
 template<typename Archive>
 void HoeffdingTree<
     FitnessFunction,
@@ -969,7 +1202,7 @@ void HoeffdingTree<
       // allocated, and then pass subvectors to the submat() function.
       std::vector<arma::uvec> indices(children.size(), arma::uvec(data.n_cols));
       arma::Col<size_t> counts =
-          arma::zeros<arma::Col<size_t>>(children.size());
+          zeros<arma::Col<size_t>>(children.size());
 
       for (size_t i = 0; i < data.n_cols; ++i)
       {
@@ -999,8 +1232,19 @@ void HoeffdingTree<
         // Train(), since the col() function is not provided.  So,
         // unfortunately, instead, we'll just extract the non-contiguous
         // submatrix.
-        MatType childData = data.cols(indices[i].subvec(0, counts[i] - 1));
-        children[i]->Train(childData, childLabels, true);
+        //
+        // I'd rather be able to use:
+        //
+        // arma::Mat<typename MatType::elem_type> childData =
+        //     data.cols(indices[i].subvec(0, counts[i] - 1));
+        //
+        // but this isn't currently supported by Armadillo.
+        arma::Mat<typename MatType::elem_type> childData(data.n_rows,
+            counts[i]);
+        for (size_t j = 0; j < counts[i]; ++j)
+          childData.col(j) = data.col(indices[i][j]);
+
+        children[i]->Train(childData, childLabels, numClasses, true);
       }
     }
   }
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_main.cpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_main.cpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -214,7 +214,7 @@ void BINDING_FUNCTION(util::Params& para
       // Extract the labels from the last dimension of training set.
       Log::Info << "Using the last dimension of training set as labels."
           << endl;
-      labels = arma::conv_to<arma::Row<size_t>>::from(
+      labels = ConvTo<arma::Row<size_t>>::From(
           trainingSet.row(trainingSet.n_rows - 1));
       trainingSet.shed_row(trainingSet.n_rows - 1);
     }
@@ -230,7 +230,7 @@ void BINDING_FUNCTION(util::Params& para
     {
       // Build the model.
       model->BuildModel(trainingSet, datasetInfo, labels,
-          arma::max(labels) + 1, batchTraining, confidence, maxSamples,
+          max(labels) + 1, batchTraining, confidence, maxSamples,
           100, minSamples, bins, observationsBeforeBinning);
       --passes; // This model-building takes one pass.
     }
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_model_impl.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_model_impl.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/hoeffding_tree_model_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/hoeffding_tree_model_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -198,19 +198,23 @@ inline void HoeffdingTreeModel::Train(co
   switch (type)
   {
     case GINI_HOEFFDING:
-      giniHoeffdingTree->Train(dataset, labels, batchTraining);
+      giniHoeffdingTree->Train(dataset, labels, giniHoeffdingTree->NumClasses(),
+          batchTraining);
       break;
 
     case GINI_BINARY:
-      giniBinaryTree->Train(dataset, labels, batchTraining);
+      giniBinaryTree->Train(dataset, labels, giniBinaryTree->NumClasses(),
+          batchTraining);
       break;
 
     case INFO_HOEFFDING:
-      infoHoeffdingTree->Train(dataset, labels, batchTraining);
+      infoHoeffdingTree->Train(dataset, labels, infoHoeffdingTree->NumClasses(),
+          batchTraining);
       break;
 
     case INFO_BINARY:
-      infoBinaryTree->Train(dataset, labels, batchTraining);
+      infoBinaryTree->Train(dataset, labels, infoBinaryTree->NumClasses(),
+          batchTraining);
       break;
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/hoeffding_trees/information_gain.hpp 4.4.0-1/src/mlpack/methods/hoeffding_trees/information_gain.hpp
--- 4.3.0-2/src/mlpack/methods/hoeffding_trees/information_gain.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/hoeffding_trees/information_gain.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,7 +35,7 @@ class HoeffdingInformationGain
     arma::vec splitCounts(counts.n_elem);
     for (size_t i = 0; i < counts.n_cols; ++i)
     {
-      splitCounts[i] = arma::accu(counts.col(i));
+      splitCounts[i] = accu(counts.col(i));
       numElem += splitCounts[i];
     }
 
@@ -43,7 +43,7 @@ class HoeffdingInformationGain
     if (numElem == 0)
       return 0.0;
 
-    arma::Col<size_t> classCounts = arma::sum(counts, 1);
+    arma::Col<size_t> classCounts = sum(counts, 1);
 
     // Calculate the gain of the unsplit node.
     double gain = 0.0;
diff -pruN 4.3.0-2/src/mlpack/methods/kde/kde_main.cpp 4.4.0-1/src/mlpack/methods/kde/kde_main.cpp
--- 4.3.0-2/src/mlpack/methods/kde/kde_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kde/kde_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -122,7 +122,7 @@ BINDING_SEE_ALSO("@knn", "#knn");
 BINDING_SEE_ALSO("Kernel density estimation on Wikipedia",
     "https://en.wikipedia.org/wiki/Kernel_density_estimation");
 BINDING_SEE_ALSO("Tree-Independent Dual-Tree Algorithms",
-    "https://arxiv.org/pdf/1304.4327.pdf");
+    "https://arxiv.org/pdf/1304.4327");
 BINDING_SEE_ALSO("Fast High-dimensional Kernel Summations Using the Monte Carlo"
     " Multipole Method", "http://papers.nips.cc/paper/3539-fast-high-"
     "dimensional-kernel-summations-using-the-monte-carlo-multipole-method.pdf");
diff -pruN 4.3.0-2/src/mlpack/methods/kde/kde_rules_impl.hpp 4.4.0-1/src/mlpack/methods/kde/kde_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/kde/kde_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kde/kde_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -401,7 +401,7 @@ Score(TreeType& queryNode, TreeType& ref
 
     // Auxiliary variables.
     arma::vec sample;
-    arma::vec means = arma::zeros(queryNode.NumDescendants());
+    arma::vec means = zeros(queryNode.NumDescendants());
     size_t m;
     double meanSample = 0;
     bool useMonteCarloPredictions = true;
diff -pruN 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_pca_impl.hpp 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_pca_impl.hpp
--- 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_pca_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_pca_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -42,7 +42,7 @@ void KernelPCA<KernelType, KernelRule>::
   {
     arma::colvec transformedDataMean = arma::mean(transformedData, 1);
     transformedData = transformedData - (transformedDataMean *
-        arma::ones<arma::rowvec>(transformedData.n_cols));
+        ones<arma::rowvec>(transformedData.n_cols));
   }
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_pca_main.cpp 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_pca_main.cpp
--- 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_pca_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_pca_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,25 +49,25 @@ BINDING_LONG_DESC(
     "The kernels that are supported are listed below:"
     "\n\n"
     " * 'linear': the standard linear dot product (same as normal PCA):\n"
-    "    K(x, y) = x^T y\n"
+    "    `K(x, y) = x^T y`\n"
     "\n"
     " * 'gaussian': a Gaussian kernel; requires bandwidth:\n"
-    "    K(x, y) = exp(-(|| x - y || ^ 2) / (2 * (bandwidth ^ 2)))\n"
+    "    `K(x, y) = exp(-(|| x - y || ^ 2) / (2 * (bandwidth ^ 2)))`\n"
     "\n"
     " * 'polynomial': polynomial kernel; requires offset and degree:\n"
-    "    K(x, y) = (x^T y + offset) ^ degree\n"
+    "    `K(x, y) = (x^T y + offset) ^ degree`\n"
     "\n"
     " * 'hyptan': hyperbolic tangent kernel; requires scale and offset:\n"
-    "    K(x, y) = tanh(scale * (x^T y) + offset)\n"
+    "    `K(x, y) = tanh(scale * (x^T y) + offset)`\n"
     "\n"
     " * 'laplacian': Laplacian kernel; requires bandwidth:\n"
-    "    K(x, y) = exp(-(|| x - y ||) / bandwidth)\n"
+    "    `K(x, y) = exp(-(|| x - y ||) / bandwidth)`\n"
     "\n"
     " * 'epanechnikov': Epanechnikov kernel; requires bandwidth:\n"
-    "    K(x, y) = max(0, 1 - || x - y ||^2 / bandwidth^2)\n"
+    "    `K(x, y) = max(0, 1 - || x - y ||^2 / bandwidth^2)`\n"
     "\n"
     " * 'cosine': cosine distance:\n"
-    "    K(x, y) = 1 - (x^T y) / (|| x || * || y ||)\n"
+    "    `K(x, y) = 1 - (x^T y) / (|| x || * || y ||)`\n"
     "\n"
     "The parameters for each of the kernels should be specified with the "
     "options " + PRINT_PARAM_STRING("bandwidth") + ", " +
@@ -97,7 +97,7 @@ BINDING_EXAMPLE(
 BINDING_SEE_ALSO("Kernel principal component analysis on Wikipedia",
     "https://en.wikipedia.org/wiki/Kernel_principal_component_analysis");
 BINDING_SEE_ALSO("Kernel Principal Component Analysis (pdf)",
-    "http://pca.narod.ru/scholkopf_kernel.pdf");
+    "https://pca.narod.ru/scholkopf_kernel.pdf");
 BINDING_SEE_ALSO("KernelPCA class documentation",
     "@src/mlpack/methods/kernel_pca/kernel_pca.hpp");
 
diff -pruN 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_rules/naive_method.hpp 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_rules/naive_method.hpp
--- 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_rules/naive_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_rules/naive_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -66,10 +66,10 @@ class NaiveKernelRule
   // centered. Since we actually never work in the feature space we cannot
   // center the data. So, we perform a "psuedo-centering" using the kernel
   // matrix.
-  arma::rowvec rowMean = arma::sum(kernelMatrix, 0) / kernelMatrix.n_cols;
-  kernelMatrix.each_col() -= arma::sum(kernelMatrix, 1) / kernelMatrix.n_cols;
+  arma::rowvec rowMean = sum(kernelMatrix, 0) / kernelMatrix.n_cols;
+  kernelMatrix.each_col() -= sum(kernelMatrix, 1) / kernelMatrix.n_cols;
   kernelMatrix.each_row() -= rowMean;
-  kernelMatrix += arma::sum(rowMean) / kernelMatrix.n_cols;
+  kernelMatrix += sum(rowMean) / kernelMatrix.n_cols;
 
   // Eigendecompose the centered kernel matrix.
   kernelMatrix = arma::symmatu(kernelMatrix);
@@ -87,7 +87,7 @@ class NaiveKernelRule
   eigvec = arma::fliplr(eigvec);
 
   transformedData = eigvec.t() * kernelMatrix;
-  transformedData.each_col() /= arma::sqrt(eigval);
+  transformedData.each_col() /= sqrt(eigval);
 }
 };
 
diff -pruN 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_rules/nystroem_method.hpp 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_rules/nystroem_method.hpp
--- 4.3.0-2/src/mlpack/methods/kernel_pca/kernel_rules/nystroem_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kernel_pca/kernel_rules/nystroem_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,17 +49,17 @@ class NystroemKernelRule
     transformedData = G.t() * G;
 
     // Center the reconstructed approximation.
-    Center(transformedData, transformedData);
+    transformedData.each_col() -= arma::mean(transformedData, 1);
 
     // For PCA the data has to be centered, even if the data is centered. But
     // it is not guaranteed that the data, when mapped to the kernel space, is
     // also centered. Since we actually never work in the feature space we
     // cannot center the data. So, we perform a "psuedo-centering" using the
     // kernel matrix.
-    arma::colvec colMean = arma::sum(G, 1) / G.n_rows;
-    G.each_row() -= arma::sum(G, 0) / G.n_rows;
+    arma::colvec colMean = sum(G, 1) / G.n_rows;
+    G.each_row() -= sum(G, 0) / G.n_rows;
     G.each_col() -= colMean;
-    G += arma::sum(colMean) / G.n_rows;
+    G += sum(colMean) / G.n_rows;
 
     // Eigendecompose the centered kernel matrix.
     transformedData = arma::symmatu(transformedData);
diff -pruN 4.3.0-2/src/mlpack/methods/kmeans/elkan_kmeans_impl.hpp 4.4.0-1/src/mlpack/methods/kmeans/elkan_kmeans_impl.hpp
--- 4.3.0-2/src/mlpack/methods/kmeans/elkan_kmeans_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kmeans/elkan_kmeans_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -76,7 +76,7 @@ double ElkanKMeans<MetricType, MatType>:
 
   // Now find the closest cluster to each other cluster.  We multiply by 0.5 so
   // that this is equivalent to s(c) for each cluster c.
-  minClusterDistances = 0.5 * arma::min(clusterDistances).t();
+  minClusterDistances = 0.5 * min(clusterDistances).t();
 
   // Now loop over all points, and see which ones need to be updated.
   for (size_t i = 0; i < dataset.n_cols; ++i)
diff -pruN 4.3.0-2/src/mlpack/methods/kmeans/kmeans_main.cpp 4.4.0-1/src/mlpack/methods/kmeans/kmeans_main.cpp
--- 4.3.0-2/src/mlpack/methods/kmeans/kmeans_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kmeans/kmeans_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -106,14 +106,12 @@ BINDING_EXAMPLE(
         "clusters", 10, "max_iterations", 500, "centroid", "final"));
 
 // See also...
-BINDING_SEE_ALSO("K-Means tutorial", "@doc/tutorials/kmeans.md");
 BINDING_SEE_ALSO("@dbscan", "#dbscan");
 BINDING_SEE_ALSO("k-means++", "https://en.wikipedia.org/wiki/K-means%2B%2B");
 BINDING_SEE_ALSO("Using the triangle inequality to accelerate k-means (pdf)",
-    "http://www.aaai.org/Papers/ICML/2003/ICML03-022.pdf");
+    "https://cdn.aaai.org/ICML/2003/ICML03-022.pdf");
 BINDING_SEE_ALSO("Making k-means even faster (pdf)",
-    "http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.586.2554&rep=rep1"
-    "&type=pdf");
+    "https://epubs.siam.org/doi/pdf/10.1137/1.9781611972801.12");
 BINDING_SEE_ALSO("Accelerating exact k-means algorithms with geometric"
     " reasoning (pdf)", "http://reports-archive.adm.cs.cmu.edu/anon/anon/usr/"
     "ftp/usr0/ftp/2000/CMU-CS-00-105.pdf");
@@ -385,7 +383,7 @@ void RunKMeans(util::Params& params,
         // Save only the labels.  TODO: figure out how to get this to output an
         // arma::Mat<size_t> instead of an arma::mat.
         params.Get<arma::mat>("output") =
-            arma::conv_to<arma::mat>::from(assignments);
+            ConvTo<arma::mat>::From(assignments);
       }
       else
       {
diff -pruN 4.3.0-2/src/mlpack/methods/kmeans/kmeans_plus_plus_initialization.hpp 4.4.0-1/src/mlpack/methods/kmeans/kmeans_plus_plus_initialization.hpp
--- 4.3.0-2/src/mlpack/methods/kmeans/kmeans_plus_plus_initialization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kmeans/kmeans_plus_plus_initialization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -89,7 +89,7 @@ class KMeansPlusPlusInitialization
 
       // Next normalize the distribution (actually technically we could avoid
       // this).
-      distribution /= arma::accu(distribution);
+      distribution /= accu(distribution);
 
       // Turn it into a CDF for convenience...
       for (size_t j = 1; j < distribution.n_elem; ++j)
diff -pruN 4.3.0-2/src/mlpack/methods/kmeans/pelleg_moore_kmeans_rules_impl.hpp 4.4.0-1/src/mlpack/methods/kmeans/pelleg_moore_kmeans_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/kmeans/pelleg_moore_kmeans_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/kmeans/pelleg_moore_kmeans_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,7 +64,7 @@ double PellegMooreKMeansRules<MetricType
   // holds all of the points in the dataset.  Our goal is to determine whether
   // or not this node is dominated by a single cluster.
   const size_t whitelisted = centroids.n_cols -
-      arma::accu(referenceNode.Stat().Blacklist());
+      accu(referenceNode.Stat().Blacklist());
 
   distanceCalculations += whitelisted;
 
diff -pruN 4.3.0-2/src/mlpack/methods/lars/lars.hpp 4.4.0-1/src/mlpack/methods/lars/lars.hpp
--- 4.3.0-2/src/mlpack/methods/lars/lars.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lars/lars.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -28,9 +28,6 @@
 
 namespace mlpack {
 
-// beta is the estimator
-// yHat is the prediction from the current estimator
-
 /**
  * An implementation of LARS, a stage-wise homotopy-based algorithm for
  * l1-regularized linear regression (LASSO) and l1+l2 regularized linear
@@ -85,9 +82,14 @@ namespace mlpack {
  * }
  * @endcode
  */
+template<typename ModelMatType = arma::mat>
 class LARS
 {
  public:
+  typedef typename GetColType<ModelMatType>::type ModelColType;
+  typedef typename GetDenseMatType<ModelMatType>::type DenseMatType;
+  typedef typename ModelMatType::elem_type ElemType;
+
   /**
    * Set the parameters to LARS.  Both lambda1 and lambda2 default to 0.
    *
@@ -105,9 +107,9 @@ class LARS
    * for training.
    */
   LARS(const bool useCholesky = false,
-       const double lambda1 = 0.0,
-       const double lambda2 = 0.0,
-       const double tolerance = 1e-16,
+       const ElemType lambda1 = 0.0,
+       const ElemType lambda2 = 0.0,
+       const ElemType tolerance = 1e-16,
        const bool fitIntercept = true,
        const bool normalizeData = true);
 
@@ -134,6 +136,7 @@ class LARS
    * @param normalizeData If true, normalize all features to have unit variance
    * for training.
    */
+  mlpack_deprecated
   LARS(const bool useCholesky,
        const arma::mat& gramMatrix,
        const double lambda1 = 0.0,
@@ -148,8 +151,8 @@ class LARS
    *
    * @param data Input data.
    * @param responses A vector of targets.
-   * @param transposeData Should be true if the input data is column-major and
-   *     false otherwise.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
    * @param useCholesky Whether or not to use Cholesky decomposition when
    *     solving linear system (as opposed to using the full Gram matrix).
    * @param lambda1 Regularization parameter for l1-norm penalty.
@@ -160,13 +163,18 @@ class LARS
    * @param normalizeData If true, normalize all features to have unit variance
    * for training.
    */
-  LARS(const arma::mat& data,
-       const arma::rowvec& responses,
-       const bool transposeData = true,
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  LARS(const MatType& data,
+       const ResponsesType& responses,
+       bool colMajor = true,
        const bool useCholesky = false,
-       const double lambda1 = 0.0,
-       const double lambda2 = 0.0,
-       const double tolerance = 1e-16,
+       const ElemType lambda1 = 0.0,
+       const ElemType lambda2 = 0.0,
+       const ElemType tolerance = 1e-16,
        const bool fitIntercept = true,
        const bool normalizeData = true);
 
@@ -181,8 +189,8 @@ class LARS
    *
    * @param data Input data.
    * @param responses A vector of targets.
-   * @param transposeData Should be true if the input data is column-major and
-   *     false otherwise.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
    * @param useCholesky Whether or not to use Cholesky decomposition when
    *     solving linear system (as opposed to using the full Gram matrix).
    * @param gramMatrix Gram matrix.
@@ -194,14 +202,19 @@ class LARS
    * @param normalizeData If true, normalize all features to have unit variance
    * for training.
    */
-  LARS(const arma::mat& data,
-       const arma::rowvec& responses,
-       const bool transposeData,
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  LARS(const MatType& data,
+       const ResponsesType& responses,
+       const bool colMajor,
        const bool useCholesky,
-       const arma::mat& gramMatrix,
-       const double lambda1 = 0.0,
-       const double lambda2 = 0.0,
-       const double tolerance = 1e-16,
+       const DenseMatType& gramMatrix,
+       const ElemType lambda1 = 0.0,
+       const ElemType lambda2 = 0.0,
+       const ElemType tolerance = 1e-16,
        const bool fitIntercept = true,
        const bool normalizeData = true);
 
@@ -238,38 +251,254 @@ class LARS
    * column-major -- each column is an observation and each row is a dimension.
    * However, because LARS is more efficient on a row-major matrix, this method
    * will (internally) transpose the matrix.  If this transposition is not
-   * necessary (i.e., you want to pass in a row-major matrix), pass 'false' for
-   * the transposeData parameter.
+   * necessary (i.e., you want to pass in a row-major matrix), pass `false` for
+   * the `colMajor` parameter.
    *
-   * @param data Column-major input data (or row-major input data if rowMajor =
-   *     true).
+   * @param data Column-major input data (or row-major input data if colMajor =
+   *     false).
    * @param responses A vector of targets.
    * @param beta Vector to store the solution (the coefficients) in.
-   * @param transposeData Set to false if the data is row-major.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
    * @return minimum cost error(||y-beta*X||2 is used to calculate error).
    */
+  mlpack_deprecated
   double Train(const arma::mat& data,
                const arma::rowvec& responses,
                arma::vec& beta,
-               const bool transposeData = true);
+               const bool colMajor = true);
 
   /**
    * Run LARS.  The input matrix (like all mlpack matrices) should be
    * column-major -- each column is an observation and each row is a dimension.
    * However, because LARS is more efficient on a row-major matrix, this method
    * will (internally) transpose the matrix.  If this transposition is not
-   * necessary (i.e., you want to pass in a row-major matrix), pass 'false' for
-   * the transposeData parameter.
+   * necessary (i.e., you want to pass in a row-major matrix), pass `false` for
+   * the `colMajor` parameter.
+   *
+   * All of the different overloads below are needed until C++17 is the minimum
+   * required standard (then std::optional could be used).
    *
    * @param data Input data.
    * @param responses A vector of targets.
-   * @param transposeData Should be true if the input data is column-major and
-   *     false otherwise.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
    * @return minimum cost error(||y-beta*X||2 is used to calculate error).
    */
-  double Train(const arma::mat& data,
-               const arma::rowvec& responses,
-               const bool transposeData = true);
+
+  // Dummy overload so MetaInfoExtractor can properly detect that LARS is a
+  // regression method.
+  template<typename MatType>
+  ElemType Train(const MatType& data,
+                 const arma::rowvec& responses,
+                 const bool colMajor = true);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type,
+           typename = typename std::enable_if<
+               !std::is_same<ResponsesType, arma::rowvec>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor = true);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const ElemType lambda1);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const ElemType lambda1,
+                 const ElemType lambda2);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance,
+                 const bool fitIntercept);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance,
+                 const bool fitIntercept,
+                 const bool normalizeData);
+
+  /**
+   * Run LARS with a precomputed Gram matrix.  The input matrix (like all mlpack
+   * matrices) should be column-major -- each column is an observation and each
+   * row is a dimension.  However, because LARS is more efficient on a row-major
+   * matrix, this method will (internally) transpose the matrix.  If this
+   * transposition is not necessary (i.e., you want to pass in a row-major
+   * matrix), pass `false` for the `colMajor` parameter.
+   *
+   * All of the different overloads below are needed until C++17 is the minimum
+   * required standard (then std::optional could be used).
+   *
+   * @param data Input data.
+   * @param responses A vector of targets.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
+   * @return minimum cost error(||y-beta*X||2 is used to calculate error).
+   */
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix,
+                 const ElemType lambda1);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix,
+                 const ElemType lambda1,
+                 const ElemType lambda2);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance,
+                 const bool fitIntercept);
+
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& data,
+                 const ResponsesType& responses,
+                 const bool colMajor,
+                 const bool useCholesky,
+                 const DenseMatType& gramMatrix,
+                 const ElemType lambda1,
+                 const ElemType lambda2,
+                 const ElemType tolerance,
+                 const bool fitIntercept,
+                 const bool normalizeData);
+
+  /**
+   * Predict y_i for the given data point.
+   *
+   * @param point The data point to regress on.
+   * @return Predicted value for y_i for `point`.
+   */
+  template<typename VecType>
+  ElemType Predict(const VecType& point) const;
 
   /**
    * Predict y_i for each data point in the given data matrix using the
@@ -277,22 +506,23 @@ class LARS
    *
    * @param points The data points to regress on.
    * @param predictions y, which will contained calculated values on completion.
-   * @param rowMajor Should be true if the data points matrix is row-major and
-   *     false otherwise.
+   * @param colMajor Should be true if the input data is column-major.  Passing
+   *     row-major data can avoid a transpose operation.
    */
-  void Predict(const arma::mat& points,
-               arma::rowvec& predictions,
-               const bool rowMajor = false) const;
+  template<typename MatType, typename ResponsesType>
+  void Predict(const MatType& points,
+               ResponsesType& predictions,
+               const bool colMajor = true) const;
 
   //! Get the L1 regularization coefficient.
-  double Lambda1() const { return lambda1; }
+  ElemType Lambda1() const { return lambda1; }
   //! Modify the L1 regularization coefficient.
-  double& Lambda1() { return lambda1; }
+  ElemType& Lambda1() { return lambda1; }
 
   //! Get the L2 regularization coefficient.
-  double Lambda2() const { return lambda2; }
+  ElemType Lambda2() const { return lambda2; }
   //! Modify the L2 regularization coefficient.
-  double& Lambda2() { return lambda2; }
+  ElemType& Lambda2() { return lambda2; }
 
   //! Get whether to use the Cholesky decomposition.
   bool UseCholesky() const { return useCholesky; }
@@ -300,74 +530,49 @@ class LARS
   bool& UseCholesky() { return useCholesky; }
 
   //! Get the tolerance for maximum correlation during training.
-  double Tolerance() const { return tolerance; }
+  ElemType Tolerance() const { return tolerance; }
   //! Modify the tolerance for maximum correlation during training.
-  double& Tolerance() { return tolerance; }
+  ElemType& Tolerance() { return tolerance; }
 
   //! Get whether or not to fit an intercept.
   bool FitIntercept() const { return fitIntercept; }
   //! Modify whether or not to fit an intercept.
-  void FitIntercept(const bool newFitIntercept)
-  {
-    // If we are storing a Gram matrix internally, but now will be normalizing
-    // data, then the Gram matrix we have computed is incorrect and needs to be
-    // recomputed.
-    if (fitIntercept != newFitIntercept)
-    {
-      if (matGram != &matGramInternal)
-      {
-        throw std::invalid_argument("LARS::FitIntercept(): cannot change value "
-            "when an external Gram matrix was specified!");
-      }
-
-      fitIntercept = newFitIntercept;
-      matGramInternal.clear();
-    }
-  }
+  void FitIntercept(const bool newFitIntercept);
 
   //! Get whether or not to normalize data during training.
   bool NormalizeData() const { return normalizeData; }
   //! Modify whether or not to normalize data during training.
-  void NormalizeData(const bool newNormalizeData)
-  {
-    // If we are storing a Gram matrix internally, but now will be normalizing
-    // data, then the Gram matrix we have computed is incorrect and needs to be
-    // recomputed.
-    if (normalizeData != newNormalizeData)
-    {
-      if (matGram != &matGramInternal)
-      {
-        throw std::invalid_argument("LARS::NormalizeData(): cannot change value "
-            "when an external Gram matrix was specified!");
-      }
-
-      normalizeData = newNormalizeData;
-      matGramInternal.clear();
-    }
-  }
+  void NormalizeData(const bool newNormalizeData);
 
-  //! Access the set of active dimensions.
-  const std::vector<size_t>& ActiveSet() const { return activeSet; }
+  //! Access the set of active dimensions in the currently selected model.
+  const std::vector<size_t>& ActiveSet() const;
 
   //! Access the set of coefficients after each iteration; the solution is the
   //! last element.
-  const std::vector<arma::vec>& BetaPath() const { return betaPath; }
+  const std::vector<ModelColType>& BetaPath() const { return betaPath; }
 
   //! Access the solution coefficients
-  const arma::vec& Beta() const { return betaPath.back(); }
+  const ModelColType& Beta() const;
 
   //! Access the set of values for lambda1 after each iteration; the solution is
   //! the last element.
-  const std::vector<double>& LambdaPath() const { return lambdaPath; }
+  const std::vector<ElemType>& LambdaPath() const { return lambdaPath; }
 
   //! Return the intercept (if fitted, otherwise 0).
-  double Intercept() const { return interceptPath.back(); }
+  ElemType Intercept() const;
 
   //! Return the intercept path (the intercept for every model).
-  const std::vector<double>& InterceptPath() const { return interceptPath; }
+  const std::vector<ElemType>& InterceptPath() const { return interceptPath; }
+
+  //! Set the model to use the given lambda1 value in the path.
+  void SelectBeta(const ElemType lambda1);
+
+  //! Get the L1 penalty parameter corresponding to the currently selected
+  //! model.
+  ElemType SelectedLambda1() const { return selectedLambda1; }
 
   //! Access the upper triangular cholesky factor.
-  const arma::mat& MatUtriCholFactor() const { return matUtriCholFactor; }
+  const DenseMatType& MatUtriCholFactor() const { return matUtriCholFactor; }
 
   /**
    * Serialize the LARS model.
@@ -380,26 +585,26 @@ class LARS
    * currently-trained LARS model. Only ||y-beta*X||2 is used to calculate
    * cost error.
    *
-   * @param matX Column-major input data (or row-major input data if rowMajor =
-   *     true).
+   * @param matX Column-major input data (or row-major input data if colMajor =
+   *     false).
    * @param y responses A vector of targets.
-   * @param rowMajor Should be true if the data points matrix is row-major and
-   *   false otherwise.
+   * @param colMajor Should be true if the data points matrix is column-major.
    * @return The minimum cost error.
    */
-  double ComputeError(const arma::mat& matX,
-                      const arma::rowvec& y,
-                      const bool rowMajor = false);
+  template<typename MatType, typename ResponsesType>
+  ElemType ComputeError(const MatType& matX,
+                        const ResponsesType& y,
+                        const bool colMajor = true);
 
  private:
   //! Gram matrix.
-  arma::mat matGramInternal;
+  DenseMatType matGramInternal;
 
   //! Pointer to the Gram matrix we will use.
-  const arma::mat* matGram;
+  const DenseMatType* matGram;
 
   //! Upper triangular cholesky factor; initially 0x0 matrix.
-  arma::mat matUtriCholFactor;
+  DenseMatType matUtriCholFactor;
 
   //! Whether or not to use Cholesky decomposition when solving linear system.
   bool useCholesky;
@@ -407,15 +612,15 @@ class LARS
   //! True if this is the LASSO problem.
   bool lasso;
   //! Regularization parameter for l1 penalty.
-  double lambda1;
+  ElemType lambda1;
 
   //! True if this is the elastic net problem.
   bool elasticNet;
   //! Regularization parameter for l2 penalty.
-  double lambda2;
+  ElemType lambda2;
 
   //! Tolerance for main loop.
-  double tolerance;
+  ElemType tolerance;
 
   //! Whether or not to fit an intercept.
   bool fitIntercept;
@@ -425,17 +630,36 @@ class LARS
   bool normalizeData;
 
   //! Solution path.
-  std::vector<arma::vec> betaPath;
+  std::vector<ModelColType> betaPath;
 
   //! Value of lambda_1 for each solution in solution path.
-  std::vector<double> lambdaPath;
+  std::vector<ElemType> lambdaPath;
 
   //! Intercept (only if fitIntercept is true).
-  std::vector<double> interceptPath;
+  std::vector<ElemType> interceptPath;
 
   //! Active set of dimensions.
   std::vector<size_t> activeSet;
 
+  //! Selected lambda1 value for Predict().
+  ElemType selectedLambda1;
+
+  //! Index of selected beta (if selectedLambda1 is in lambdaPath).
+  size_t selectedIndex;
+
+  //! Selected beta, if selectedLambda1 is not in lambdaPath.
+  ModelColType selectedBeta;
+
+  //! Selected intercept, if selectedLambda1 is not in lambdaPath.
+  ElemType selectedIntercept;
+
+  //! Selected active set of dimensions, if selectedLambda1 is not the last
+  //! element in the path.
+  std::vector<size_t> selectedActiveSet;
+
+  //! Might be needed to compute the intercept for other lambda values.
+  ElemType offsetY;
+
   //! Active set membership indicator (for each dimension).
   std::vector<bool> isActive;
 
@@ -468,27 +692,34 @@ class LARS
    */
   void Ignore(const size_t varInd);
 
-  // compute "equiangular" direction in output space
-  void ComputeYHatDirection(const arma::mat& matX,
-                            const arma::vec& betaDirection,
-                            arma::vec& yHatDirection);
+  // Compute "equiangular" direction in output space.
+  template<typename MatType, typename VecType>
+  void ComputeYHatDirection(const MatType& matX,
+                            const VecType& betaDirection,
+                            VecType& yHatDirection);
 
-  // interpolate to compute last solution vector
+  // Interpolate to compute last solution vector.
   void InterpolateBeta();
 
-  void CholeskyInsert(const arma::vec& newX, const arma::mat& X);
+  template<typename VecType, typename MatType>
+  void CholeskyInsert(const VecType& newX, const MatType& X);
 
-  void CholeskyInsert(double sqNormNewX, const arma::vec& newGramCol);
+  template<typename VecType>
+  void CholeskyInsert(ElemType sqNormNewX, const VecType& newGramCol);
 
-  void GivensRotate(const arma::vec::fixed<2>& x,
-                    arma::vec::fixed<2>& rotatedX,
-                    arma::mat& G);
+  template<typename MatType>
+  void GivensRotate(const typename arma::Col<ElemType>::template fixed<2>& x,
+                    typename arma::Col<ElemType>::template fixed<2>& rotatedX,
+                    MatType& G);
 
   void CholeskyDelete(const size_t colToKill);
 };
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename ModelMatType),
+    (mlpack::LARS<ModelMatType>), (1));
+
 // Include implementation of serialize().
 #include "lars_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/lars/lars_impl.hpp 4.4.0-1/src/mlpack/methods/lars/lars_impl.hpp
--- 4.3.0-2/src/mlpack/methods/lars/lars_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lars/lars_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,11 +17,12 @@
 
 namespace mlpack {
 
-inline LARS::LARS(
+template<typename ModelMatType>
+inline LARS<ModelMatType>::LARS(
     const bool useCholesky,
-    const double lambda1,
-    const double lambda2,
-    const double tolerance,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance,
     const bool fitIntercept,
     const bool normalizeData) :
     matGram(&matGramInternal),
@@ -32,10 +33,16 @@ inline LARS::LARS(
     lambda2(lambda2),
     tolerance(tolerance),
     fitIntercept(fitIntercept),
-    normalizeData(normalizeData)
+    normalizeData(normalizeData),
+    selectedLambda1(lambda1),
+    selectedIndex(0),
+    selectedIntercept(0.0),
+    offsetY(0.0)
 { /* Nothing left to do. */ }
 
-inline LARS::LARS(
+template<typename ModelMatType>
+mlpack_deprecated
+inline LARS<ModelMatType>::LARS(
     const bool useCholesky,
     const arma::mat& gramMatrix,
     const double lambda1,
@@ -51,43 +58,63 @@ inline LARS::LARS(
     lambda2(lambda2),
     tolerance(tolerance),
     fitIntercept(fitIntercept),
-    normalizeData(normalizeData)
+    normalizeData(normalizeData),
+    selectedLambda1(lambda1),
+    selectedIndex(0),
+    selectedIntercept(0.0),
+    offsetY(0.0)
 { /* Nothing left to do */ }
 
-inline LARS::LARS(
-    const arma::mat& data,
-    const arma::rowvec& responses,
-    const bool transposeData,
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline LARS<ModelMatType>::LARS(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
     const bool useCholesky,
-    const double lambda1,
-    const double lambda2,
-    const double tolerance,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance,
     const bool fitIntercept,
     const bool normalizeData) :
     LARS(useCholesky, lambda1, lambda2, tolerance, fitIntercept, normalizeData)
 {
-  Train(data, responses, transposeData);
+  Train(data, responses, colMajor);
 }
 
-inline LARS::LARS(
-    const arma::mat& data,
-    const arma::rowvec& responses,
-    const bool transposeData,
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline LARS<ModelMatType>::LARS(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
     const bool useCholesky,
-    const arma::mat& gramMatrix,
-    const double lambda1,
-    const double lambda2,
-    const double tolerance,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance,
     const bool fitIntercept,
     const bool normalizeData) :
-    LARS(useCholesky, gramMatrix, lambda1, lambda2, tolerance, fitIntercept,
-        normalizeData)
+    matGram(&gramMatrix),
+    useCholesky(useCholesky),
+    lasso((lambda1 != 0)),
+    lambda1(lambda1),
+    elasticNet((lambda1 != 0) && (lambda2 != 0)),
+    lambda2(lambda2),
+    tolerance(tolerance),
+    fitIntercept(fitIntercept),
+    normalizeData(normalizeData),
+    selectedLambda1(lambda1),
+    selectedIndex(0),
+    selectedIntercept(0.0),
+    offsetY(0.0)
 {
-  Train(data, responses, transposeData);
+  Train(data, responses, colMajor);
 }
 
 // Copy Constructor.
-inline LARS::LARS(const LARS& other) :
+template<typename ModelMatType>
+inline LARS<ModelMatType>::LARS(const LARS<ModelMatType>& other) :
     matGramInternal(other.matGramInternal),
     matGram(other.matGram != &other.matGramInternal ?
         other.matGram : &matGramInternal),
@@ -104,6 +131,12 @@ inline LARS::LARS(const LARS& other) :
     lambdaPath(other.lambdaPath),
     interceptPath(other.interceptPath),
     activeSet(other.activeSet),
+    selectedLambda1(other.selectedLambda1),
+    selectedIndex(other.selectedIndex),
+    selectedBeta(other.selectedBeta),
+    selectedIntercept(other.selectedIntercept),
+    selectedActiveSet(other.selectedActiveSet),
+    offsetY(other.offsetY),
     isActive(other.isActive),
     ignoreSet(other.ignoreSet),
     isIgnored(other.isIgnored)
@@ -112,7 +145,8 @@ inline LARS::LARS(const LARS& other) :
 }
 
 // Move constructor.
-inline LARS::LARS(LARS&& other) :
+template<typename ModelMatType>
+inline LARS<ModelMatType>::LARS(LARS<ModelMatType>&& other) :
     matGramInternal(std::move(other.matGramInternal)),
     matGram(other.matGram != &other.matGramInternal ?
         other.matGram : &matGramInternal),
@@ -129,6 +163,12 @@ inline LARS::LARS(LARS&& other) :
     lambdaPath(std::move(other.lambdaPath)),
     interceptPath(std::move(other.interceptPath)),
     activeSet(std::move(other.activeSet)),
+    selectedLambda1(std::move(other.selectedLambda1)),
+    selectedIndex(std::move(other.selectedIndex)),
+    selectedBeta(std::move(other.selectedBeta)),
+    selectedIntercept(std::move(other.selectedIntercept)),
+    selectedActiveSet(std::move(other.selectedActiveSet)),
+    offsetY(std::move(other.offsetY)),
     isActive(std::move(other.isActive)),
     ignoreSet(std::move(other.ignoreSet)),
     isIgnored(std::move(other.isIgnored))
@@ -137,7 +177,9 @@ inline LARS::LARS(LARS&& other) :
 }
 
 // Copy operator.
-inline LARS& LARS::operator=(const LARS& other)
+template<typename ModelMatType>
+inline LARS<ModelMatType>& LARS<ModelMatType>::operator=(
+    const LARS<ModelMatType>& other)
 {
   if (&other == this)
     return *this;
@@ -158,6 +200,12 @@ inline LARS& LARS::operator=(const LARS&
   lambdaPath = other.lambdaPath;
   interceptPath = other.interceptPath;
   activeSet = other.activeSet;
+  selectedLambda1 = other.selectedLambda1;
+  selectedIndex = other.selectedIndex;
+  selectedBeta = other.selectedBeta;
+  selectedIntercept = other.selectedIntercept;
+  selectedActiveSet = other.selectedActiveSet;
+  offsetY = other.offsetY;
   isActive = other.isActive;
   ignoreSet = other.ignoreSet;
   isIgnored = other.isIgnored;
@@ -165,7 +213,9 @@ inline LARS& LARS::operator=(const LARS&
 }
 
 // Move Operator.
-inline LARS& LARS::operator=(LARS&& other)
+template<typename ModelMatType>
+inline LARS<ModelMatType>& LARS<ModelMatType>::operator=(
+    LARS<ModelMatType>&& other)
 {
   if (&other == this)
     return *this;
@@ -185,6 +235,12 @@ inline LARS& LARS::operator=(LARS&& othe
   betaPath = std::move(other.betaPath);
   lambdaPath = std::move(other.lambdaPath);
   interceptPath = std::move(other.interceptPath);
+  selectedLambda1 = std::move(other.selectedLambda1);
+  selectedIndex = std::move(other.selectedIndex);
+  selectedBeta = std::move(other.selectedBeta);
+  selectedIntercept = std::move(other.selectedIntercept);
+  selectedActiveSet = std::move(other.selectedActiveSet);
+  offsetY = std::move(other.offsetY);
   activeSet = std::move(other.activeSet);
   isActive = std::move(other.isActive);
   ignoreSet = std::move(other.ignoreSet);
@@ -192,11 +248,236 @@ inline LARS& LARS::operator=(LARS&& othe
   return *this;
 }
 
-inline double LARS::Train(const arma::mat& matX,
-                          const arma::rowvec& y,
-                          arma::vec& beta,
-                          const bool transposeData)
+template<typename ModelMatType>
+mlpack_deprecated
+inline double LARS<ModelMatType>::Train(const arma::mat& matX,
+                                        const arma::rowvec& y,
+                                        arma::vec& beta,
+                                        const bool colMajor)
 {
+  const double result = Train(matX, y, colMajor);
+  beta = betaPath.back();
+  return result;
+}
+
+// Dummy overload for MetaInfoExtractor.
+template<typename ModelMatType>
+template<typename MatType>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const arma::rowvec& responses,
+                          const bool colMajor)
+{
+  return Train(data, responses, colMajor, this->useCholesky, this->lambda1,
+      this->lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor)
+{
+  return Train(data, responses, colMajor, this->useCholesky, this->lambda1,
+      this->lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor,
+                          const bool useCholesky)
+{
+  return Train(data, responses, colMajor, useCholesky, this->lambda1,
+      this->lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor,
+                          const bool useCholesky,
+                          const typename LARS<ModelMatType>::ElemType lambda1)
+{
+  return Train(data, responses, colMajor, useCholesky, lambda1,
+      this->lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor,
+                          const bool useCholesky,
+                          const typename LARS<ModelMatType>::ElemType lambda1,
+                          const typename LARS<ModelMatType>::ElemType lambda2)
+{
+  return Train(data, responses, colMajor, useCholesky, lambda1, lambda2,
+      this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor,
+                          const bool useCholesky,
+                          const typename LARS<ModelMatType>::ElemType lambda1,
+                          const typename LARS<ModelMatType>::ElemType lambda2,
+                          const typename LARS<ModelMatType>::ElemType tolerance)
+{
+  return Train(data, responses, colMajor, useCholesky, lambda1, lambda2,
+      tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& data,
+                          const ResponsesType& responses,
+                          const bool colMajor,
+                          const bool useCholesky,
+                          const typename LARS<ModelMatType>::ElemType lambda1,
+                          const typename LARS<ModelMatType>::ElemType lambda2,
+                          const typename LARS<ModelMatType>::ElemType tolerance,
+                          const bool fitIntercept)
+{
+  return Train(data, responses, colMajor, useCholesky, lambda1, lambda2,
+      tolerance, fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix)
+{
+  return Train(data, responses, colMajor, useCholesky, gramMatrix,
+      this->lambda1, this->lambda2, this->tolerance, this->fitIntercept,
+      this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1)
+{
+  return Train(data, responses, colMajor, useCholesky, gramMatrix, lambda1,
+      this->lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2)
+{
+  return Train(data, responses, colMajor, useCholesky, gramMatrix, lambda1,
+      lambda2, this->tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance)
+{
+  return Train(data, responses, colMajor, useCholesky, gramMatrix, lambda1,
+      lambda2, tolerance, this->fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance,
+    const bool fitIntercept)
+{
+  return Train(data, responses, colMajor, useCholesky, gramMatrix, lambda1,
+      lambda2, tolerance, fitIntercept, this->normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(
+    const MatType& data,
+    const ResponsesType& responses,
+    const bool colMajor,
+    const bool useCholesky,
+    const typename LARS<ModelMatType>::DenseMatType& gramMatrix,
+    const typename LARS<ModelMatType>::ElemType lambda1,
+    const typename LARS<ModelMatType>::ElemType lambda2,
+    const typename LARS<ModelMatType>::ElemType tolerance,
+    const bool fitIntercept,
+    const bool normalizeData)
+{
+  // Set Gram matrix.
+  matGramInternal.clear();
+  matGram = &gramMatrix;
+
+  return Train(data, responses, colMajor, useCholesky, lambda1, lambda2,
+      tolerance, fitIntercept, normalizeData);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Train(const MatType& matX,
+                          const ResponsesType& y,
+                          const bool colMajor,
+                          const bool useCholesky,
+                          const typename LARS<ModelMatType>::ElemType lambda1,
+                          const typename LARS<ModelMatType>::ElemType lambda2,
+                          const typename LARS<ModelMatType>::ElemType tolerance,
+                          const bool fitIntercept,
+                          const bool normalizeData)
+{
+  // Update hyperparameter settings.
+  this->useCholesky = useCholesky;
+  this->lambda1 = lambda1;
+  this->lambda2 = lambda2;
+  this->tolerance = tolerance;
+  this->fitIntercept = fitIntercept;
+  this->normalizeData = normalizeData;
+
   // Clear any previous solution information.
   betaPath.clear();
   lambdaPath.clear();
@@ -205,28 +486,29 @@ inline double LARS::Train(const arma::ma
   ignoreSet.clear();
   isIgnored.clear();
   matUtriCholFactor.reset();
+  selectedBeta.clear();
 
   // Update values in case lambda1 or lambda2 changed.
   lasso = (lambda1 != 0);
   elasticNet = (lambda1 != 0 && lambda2 != 0);
 
   // This matrix may end up holding the transpose -- if necessary.
-  arma::mat dataTrans;
+  MatType dataTrans;
   // This vector may hold zero-centered responses, if necessary.
-  arma::rowvec yCentered;
+  ResponsesType yCentered;
 
   // dataRef is row-major.  We can reuse the given matX, but only if we don't
   // need to do any transformations to it.
-  const arma::mat& dataRef =
-      (transposeData || fitIntercept || normalizeData) ? dataTrans : matX;
-  const arma::rowvec& yRef =
+  const MatType& dataRef =
+      (colMajor || fitIntercept || normalizeData) ? dataTrans : matX;
+  const ResponsesType& yRef =
       (fitIntercept) ? yCentered : y;
 
-  arma::vec offsetX; // used only if fitting an intercept
-  double offsetY = 0.0; // used only if fitting an intercept
-  arma::vec stdX; // used only if normalizing
+  arma::Col<ElemType> offsetX; // used only if fitting an intercept
+  this->offsetY = 0.0; // used only if fitting an intercept
+  arma::Col<ElemType> stdX; // used only if normalizing
 
-  if (transposeData)
+  if (colMajor)
   {
     if (fitIntercept)
     {
@@ -278,12 +560,12 @@ inline double LARS::Train(const arma::ma
 
   if (fitIntercept)
   {
-    offsetY = arma::mean(y);
-    yCentered = y - offsetY;
+    this->offsetY = arma::mean(y);
+    yCentered = y - this->offsetY;
   }
 
   // Compute X' * y.
-  arma::vec vecXTy = trans(yRef * dataRef);
+  arma::Col<ElemType> vecXTy = trans(yRef * dataRef);
 
   // Set up active set variables.  In the beginning, the active set has size 0
   // (all dimensions are inactive).
@@ -293,15 +575,15 @@ inline double LARS::Train(const arma::ma
   isIgnored.resize(dataRef.n_cols, false);
 
   // Initialize yHat and beta.
-  beta = arma::zeros(dataRef.n_cols);
-  arma::vec yHat = arma::zeros(dataRef.n_rows);
-  arma::vec yHatDirection(dataRef.n_rows);
+  arma::Col<ElemType> beta(dataRef.n_cols, arma::fill::zeros);
+  arma::Col<ElemType> yHat(dataRef.n_rows, arma::fill::zeros);
+  arma::Col<ElemType> yHatDirection(dataRef.n_rows, arma::fill::none);
 
   bool lassocond = false;
 
   // Compute the initial maximum correlation among all dimensions.
-  arma::vec corr = vecXTy;
-  double maxCorr = 0;
+  arma::Col<ElemType> corr = vecXTy;
+  ElemType maxCorr = 0;
   size_t changeInd = 0;
   size_t lassocondInd = dataRef.n_cols;
   for (size_t i = 0; i < vecXTy.n_elem; ++i)
@@ -322,7 +604,7 @@ inline double LARS::Train(const arma::ma
     lambdaPath[0] = lambda1;
 
     if (fitIntercept)
-      interceptPath.push_back(offsetY - arma::dot(offsetX, betaPath[0]));
+      interceptPath.push_back(this->offsetY - dot(offsetX, betaPath[0]));
     else
       interceptPath.push_back(0.0);
 
@@ -337,7 +619,10 @@ inline double LARS::Train(const arma::ma
     matGramInternal = trans(dataRef) * dataRef;
 
     if (elasticNet && !useCholesky)
-      matGramInternal += lambda2 * arma::eye(dataRef.n_cols, dataRef.n_cols);
+    {
+      matGramInternal += lambda2 *
+          arma::eye<ModelMatType>(dataRef.n_cols, dataRef.n_cols);
+    }
   }
 
   // Main loop.
@@ -346,8 +631,8 @@ inline double LARS::Train(const arma::ma
   {
     // Compute the maximum correlation among inactive dimensions.
     maxCorr = 0;
-    double maxActiveCorr = 0;
-    double minActiveCorr = DBL_MAX;
+    ElemType maxActiveCorr = 0;
+    ElemType minActiveCorr = DBL_MAX;
     for (size_t i = 0; i < dataRef.n_cols; ++i)
     {
       if ((!isActive[i]) && (!isIgnored[i]) && (fabs(corr(i)) > maxCorr))
@@ -373,26 +658,6 @@ inline double LARS::Train(const arma::ma
     if (maxCorr < tolerance)
       break;
 
-    if ((matGram != &matGramInternal) && ((maxActiveCorr - minActiveCorr) >
-        100 * std::numeric_limits<double>::epsilon()))
-    {
-      // Construct the error message to match the user's settings.
-      std::ostringstream oss;
-      oss << "LARS::Train(): correlation conditions violated; check that your "
-          << "given Gram matrix is properly computed on ";
-      if (fitIntercept)
-        oss << "mean-centered ";
-      else
-        oss << "non-mean-centered ";
-      if (normalizeData)
-        oss << "unit-variance (normalized) ";
-      else
-        oss << "non-normalized ";
-      oss << "data!";
-      oss << std::endl;
-      throw std::runtime_error(oss.str());
-    }
-
     // Add the variable to the active set and update the Gram matrix as
     // necessary.
     if (!lassocond)
@@ -405,8 +670,9 @@ inline double LARS::Train(const arma::ma
         //   newGramCol[i] = dot(matX.col(activeSet[i]), matX.col(changeInd));
         // }
         // This is equivalent to the above 5 lines.
-        arma::vec newGramCol = matGram->elem(changeInd * dataRef.n_cols +
-            arma::conv_to<arma::uvec>::from(activeSet));
+        arma::Col<ElemType> newGramCol = matGram->elem(
+            changeInd * dataRef.n_cols +
+            ConvTo<arma::uvec>::From(activeSet));
 
         CholeskyInsert((*matGram)(changeInd, changeInd), newGramCol);
       }
@@ -414,20 +680,23 @@ inline double LARS::Train(const arma::ma
     }
 
     // Compute signs of correlations.
-    arma::vec s = arma::vec(activeSet.size());
+    arma::Col<ElemType> s(activeSet.size());
     for (size_t i = 0; i < activeSet.size(); ++i)
-      s(i) = corr(activeSet[i]) / fabs(corr(activeSet[i]));
+    {
+      const size_t j = activeSet[i];
+      s[i] = (ElemType) (corr(j) == 0.0 ? 0.0 : (corr(j) > 0) ? 1.0 : -1.0);
+    }
 
     // Compute the "equiangular" direction in parameter space (betaDirection).
     // We use quotes because in the case of non-unit norm variables, this need
     // not be equiangular.
-    arma::vec unnormalizedBetaDirection;
-    double normalization;
-    arma::vec betaDirection;
+    arma::Col<ElemType> unnormalizedBetaDirection;
+    ElemType normalization;
+    arma::Col<ElemType> betaDirection;
     if (useCholesky)
     {
       // Check for singularity.
-      const double lastUtriElement = matUtriCholFactor(
+      const ElemType lastUtriElement = matUtriCholFactor(
           matUtriCholFactor.n_cols - 1, matUtriCholFactor.n_rows - 1);
       if (std::abs(lastUtriElement) > tolerance)
       {
@@ -446,7 +715,7 @@ inline double LARS::Train(const arma::ma
         unnormalizedBetaDirection = solve(trimatu(matUtriCholFactor),
             solve(trimatl(trans(matUtriCholFactor)), s));
 
-        normalization = 1.0 / sqrt(dot(s, unnormalizedBetaDirection));
+        normalization = 1.0 / std::sqrt(dot(s, unnormalizedBetaDirection));
         betaDirection = normalization * unnormalizedBetaDirection;
       }
       else
@@ -467,26 +736,26 @@ inline double LARS::Train(const arma::ma
         unnormalizedBetaDirection = solve(trimatu(matUtriCholFactor),
             solve(trimatl(trans(matUtriCholFactor)), s));
 
-        normalization = 1.0 / sqrt(dot(s, unnormalizedBetaDirection));
+        normalization = 1.0 / std::sqrt(dot(s, unnormalizedBetaDirection));
         betaDirection = normalization * unnormalizedBetaDirection;
       }
     }
     else
     {
-      arma::mat matGramActive = arma::mat(activeSet.size(), activeSet.size());
+      MatType matGramActive(activeSet.size(), activeSet.size());
       for (size_t i = 0; i < activeSet.size(); ++i)
         for (size_t j = 0; j < activeSet.size(); ++j)
           matGramActive(i, j) = (*matGram)(activeSet[i], activeSet[j]);
 
       // Check for singularity.
-      arma::mat matS = s * arma::ones<arma::mat>(1, activeSet.size());
+      MatType matS = s * ones<MatType>(1, activeSet.size());
       const bool solvedOk = solve(unnormalizedBetaDirection,
           matGramActive % trans(matS) % matS,
-          arma::ones<arma::mat>(activeSet.size(), 1));
+          ones<MatType>(activeSet.size(), 1));
       if (solvedOk)
       {
         // Ok, no singularity.
-        normalization = 1.0 / sqrt(sum(unnormalizedBetaDirection));
+        normalization = 1.0 / std::sqrt(sum(unnormalizedBetaDirection));
         betaDirection = normalization * unnormalizedBetaDirection % s;
       }
       else
@@ -503,12 +772,14 @@ inline double LARS::Train(const arma::ma
         // need to take a step with the previous beta direction towards the next
         // variable we will add.
         s = s.subvec(0, activeSet.size() - 1); // Drop last element.
-        matS = s * arma::ones<arma::mat>(1, activeSet.size());
+        matGramActive = matGramActive.submat(0, 0, activeSet.size() - 1,
+            activeSet.size() - 1);
+        matS = s * ones<MatType>(1, activeSet.size());
         // This worked last iteration, so there can't be a singularity.
         solve(unnormalizedBetaDirection,
             matGramActive % trans(matS) % matS,
-            arma::ones<arma::mat>(activeSet.size(), 1));
-        normalization = 1.0 / sqrt(sum(unnormalizedBetaDirection));
+            ones<MatType>(activeSet.size(), 1));
+        normalization = 1.0 / std::sqrt(sum(unnormalizedBetaDirection));
         betaDirection = normalization * unnormalizedBetaDirection % s;
       }
     }
@@ -516,7 +787,7 @@ inline double LARS::Train(const arma::ma
     // compute "equiangular" direction in output space
     ComputeYHatDirection(dataRef, betaDirection, yHatDirection);
 
-    double gamma = maxCorr / normalization;
+    ElemType gamma = maxCorr / normalization;
 
     // If not all variables are active.
     if ((activeSet.size() + ignoreSet.size()) < dataRef.n_cols)
@@ -527,9 +798,9 @@ inline double LARS::Train(const arma::ma
         if (isActive[ind] || isIgnored[ind])
           continue;
 
-        const double dirCorr = dot(dataRef.col(ind), yHatDirection);
-        const double val1 = (maxCorr - corr(ind)) / (normalization - dirCorr);
-        const double val2 = (maxCorr + corr(ind)) / (normalization + dirCorr);
+        const ElemType dirCorr = dot(dataRef.col(ind), yHatDirection);
+        const ElemType val1 = (maxCorr - corr(ind)) / (normalization - dirCorr);
+        const ElemType val2 = (maxCorr + corr(ind)) / (normalization + dirCorr);
 
         // If we kicked out a feature due to the LASSO modification last
         // iteration, then we do not allow relaxation of the step size to 0 for
@@ -556,12 +827,12 @@ inline double LARS::Train(const arma::ma
     {
       lassocond = false;
       lassocondInd = dataRef.n_cols;
-      double lassoboundOnGamma = DBL_MAX;
+      ElemType lassoboundOnGamma = DBL_MAX;
       size_t activeIndToKickOut = -1;
 
       for (size_t i = 0; i < activeSet.size(); ++i)
       {
-        double val = -beta(activeSet[i]) / betaDirection(i);
+        ElemType val = -beta(activeSet[i]) / betaDirection(i);
         if ((val > 0) && (val < lassoboundOnGamma))
         {
           lassoboundOnGamma = val;
@@ -611,7 +882,7 @@ inline double LARS::Train(const arma::ma
     if (elasticNet)
       corr -= lambda2 * beta;
 
-    double curLambda = 0;
+    ElemType curLambda = 0;
     for (size_t i = 0; i < activeSet.size(); ++i)
       curLambda += fabs(corr(activeSet[i]));
 
@@ -647,7 +918,7 @@ inline double LARS::Train(const arma::ma
   {
     interceptPath.clear();
     for (size_t i = 0; i < betaPath.size(); ++i)
-      interceptPath.push_back(offsetY - arma::dot(offsetX, betaPath[i]));
+      interceptPath.push_back(this->offsetY - dot(offsetX, betaPath[i]));
   }
   else
   {
@@ -655,71 +926,243 @@ inline double LARS::Train(const arma::ma
     interceptPath.resize(betaPath.size(), 0.0);
   }
 
-  // Unfortunate copy...
-  beta = betaPath.back();
+  // Make the model we use point to the last element in the path after
+  // interpolation.
+  selectedLambda1 = lambda1;
+  selectedIndex = betaPath.size() - 1;
 
-  return ComputeError(matX, y, !transposeData);
+  return ComputeError(matX, y, colMajor);
 }
 
-inline double LARS::Train(const arma::mat& data,
-                          const arma::rowvec& responses,
-                          const bool transposeData)
+template<typename ModelMatType>
+template<typename VecType>
+inline typename LARS<ModelMatType>::ElemType LARS<ModelMatType>::Predict(
+    const VecType& point) const
 {
-  arma::vec beta;
-  return Train(data, responses, beta, transposeData);
+  if (!fitIntercept)
+    return dot(Beta(), point);
+  else
+    return dot(Beta(), point) + Intercept();
 }
 
-inline void LARS::Predict(const arma::mat& points,
-                          arma::rowvec& predictions,
-                          const bool rowMajor) const
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType>
+inline void LARS<ModelMatType>::Predict(const MatType& points,
+                                        ResponsesType& predictions,
+                                        const bool colMajor) const
 {
-  // We really only need to store beta internally...
-  if (rowMajor && !fitIntercept)
-    predictions = trans(points * betaPath.back());
-  else if (rowMajor)
-    predictions = trans(points * betaPath.back()) + interceptPath.back();
+  if (!colMajor && !fitIntercept)
+    predictions = trans(points * Beta());
+  else if (!colMajor)
+    predictions = trans(points * Beta()) + Intercept();
   else if (fitIntercept)
-    predictions = betaPath.back().t() * points + interceptPath.back();
+    predictions = Beta().t() * points + Intercept();
+  else
+    predictions = Beta().t() * points;
+}
+
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::FitIntercept(const bool newFitIntercept)
+{
+  // If we are storing a Gram matrix internally, but now will be normalizing
+  // data, then the Gram matrix we have computed is incorrect and needs to be
+  // recomputed.
+  if (fitIntercept != newFitIntercept)
+  {
+    if (matGram != &matGramInternal)
+    {
+      throw std::invalid_argument("LARS::FitIntercept(): cannot change value "
+          "when an external Gram matrix was specified!");
+    }
+
+    fitIntercept = newFitIntercept;
+    matGramInternal.clear();
+  }
+}
+
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::NormalizeData(const bool newNormalizeData)
+{
+  // If we are storing a Gram matrix internally, but now will be normalizing
+  // data, then the Gram matrix we have computed is incorrect and needs to be
+  // recomputed.
+  if (normalizeData != newNormalizeData)
+  {
+    if (matGram != &matGramInternal)
+    {
+      throw std::invalid_argument("LARS::NormalizeData(): cannot change value"
+          " when an external Gram matrix was specified!");
+    }
+
+    normalizeData = newNormalizeData;
+    matGramInternal.clear();
+  }
+}
+
+template<typename ModelMatType>
+inline const std::vector<size_t>& LARS<ModelMatType>::ActiveSet() const
+{
+  if (selectedIndex != (betaPath.size() - 1))
+    return selectedActiveSet;
+  else
+    return activeSet;
+}
+
+template<typename ModelMatType>
+inline const typename LARS<ModelMatType>::ModelColType&
+LARS<ModelMatType>::Beta() const
+{
+  if (selectedIndex < betaPath.size())
+    return betaPath[selectedIndex];
+  else
+    return selectedBeta;
+}
+
+template<typename ModelMatType>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::Intercept() const
+{
+  if (selectedIndex < betaPath.size())
+    return interceptPath[selectedIndex];
+  else
+    return selectedIntercept;
+}
+
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::SelectBeta(
+    const typename LARS<ModelMatType>::ElemType selLambda1)
+{
+  if (selLambda1 < lambda1)
+  {
+    std::ostringstream oss;
+    oss << "LARS::SelectBeta(): given lambda1 value (" << selLambda1 << ") "
+        << "cannot be less than model's Lambda1() value (" << lambda1
+        << ")!";
+    throw std::invalid_argument(oss.str());
+  }
+  else if (betaPath.size() == 0)
+  {
+    throw std::runtime_error("LARS::SelectBeta(): model must be trained "
+        "before calling SelectBeta()!");
+  }
+
+  this->selectedLambda1 = selLambda1;
+  selectedBeta.clear();
+
+  // Find which lambda values we are interpolating between.  lambdaPath is in
+  // reverse order (due to the fact that LARS is a stepwise algorithm), so the
+  // largest lambdas come first.
+  size_t i = 0;
+  while (i < lambdaPath.size())
+  {
+    if (selLambda1 == lambdaPath[i])
+    {
+      // If it's an exact match, no interpolation is necessary, and we can
+      // directly use the element from the path.
+      selectedIndex = i;
+
+      // However, we may need to compute the active set.
+      if (i != lambdaPath.size() - 1)
+      {
+        selectedActiveSet = ConvTo<std::vector<size_t>>::From(
+            arma::find(betaPath[i] != 0));
+      }
+
+      return;
+    }
+    else if (selLambda1 > lambdaPath[i])
+    {
+      // It's not an exact match, but lambdaPath[i] is the first lambda element
+      // that is smaller than the desired lambda.
+      break;
+    }
+
+    ++i;
+  }
+
+  // In the case where selLambda1 is larger than the largest lambda we have a
+  // model for, we can interpolate between the zero vector and the first model.
+  if (i == 0)
+  {
+    const ElemType interp = selLambda1 / lambdaPath[0];
+    selectedIndex = betaPath.size();
+
+    selectedLambda1 = interp * lambdaPath[0];
+    selectedBeta = interp * betaPath[0];
+    // Computing the intercept differs just a little bit from what's expected,
+    // because we have to account for the offsetY term, which is not zero even
+    // for a zero model.
+    selectedIntercept = (1 - interp) * this->offsetY +
+        interp * interceptPath[0];
+  }
+  else if (i == betaPath.size())
+  {
+    // It's possible that we fit the model perfectly with some lambda1 value
+    // less than this->lambda1.  In that case, the interpolated solution is just
+    // the last solution.
+    selectedIndex = betaPath.size() - 1;
+    return;
+  }
   else
-    predictions = betaPath.back().t() * points;
+  {
+    const ElemType interp = (lambdaPath[i - 1] - selLambda1) /
+        (lambdaPath[i - 1] - lambdaPath[i]);
+    selectedIndex = betaPath.size();
+
+    selectedLambda1 = (1 - interp) * lambdaPath[i - 1] + interp * lambdaPath[i];
+    selectedBeta = (1 - interp) * betaPath[i - 1] + interp * betaPath[i];
+    selectedIntercept = (1 - interp) * interceptPath[i - 1] +
+        interp * interceptPath[i];
+  }
+
+  // Compute the active set of variables.
+  selectedActiveSet = ConvTo<std::vector<size_t>>::From(
+      arma::find(selectedBeta != 0));
 }
 
 // Private functions.
-inline void LARS::Deactivate(const size_t activeVarInd)
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::Deactivate(const size_t activeVarInd)
 {
   isActive[activeSet[activeVarInd]] = false;
   activeSet.erase(activeSet.begin() + activeVarInd);
 }
 
-inline void LARS::Activate(const size_t varInd)
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::Activate(const size_t varInd)
 {
   isActive[varInd] = true;
   activeSet.push_back(varInd);
 }
 
-inline void LARS::Ignore(const size_t varInd)
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::Ignore(const size_t varInd)
 {
   isIgnored[varInd] = true;
   ignoreSet.push_back(varInd);
 }
 
-inline void LARS::ComputeYHatDirection(const arma::mat& matX,
-                                       const arma::vec& betaDirection,
-                                       arma::vec& yHatDirection)
+template<typename ModelMatType>
+template<typename MatType, typename VecType>
+inline void LARS<ModelMatType>::ComputeYHatDirection(
+    const MatType& matX,
+    const VecType& betaDirection,
+    VecType& yHatDirection)
 {
   yHatDirection.fill(0);
   for (size_t i = 0; i < activeSet.size(); ++i)
     yHatDirection += betaDirection(i) * matX.col(activeSet[i]);
 }
 
-inline void LARS::InterpolateBeta()
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::InterpolateBeta()
 {
   const size_t pathLength = betaPath.size();
 
   // interpolate beta and stop
-  double ultimateLambda = lambdaPath[pathLength - 1];
-  double penultimateLambda = lambdaPath[pathLength - 2];
-  double interp = (penultimateLambda - lambda1)
+  ElemType ultimateLambda = lambdaPath[pathLength - 1];
+  ElemType penultimateLambda = lambdaPath[pathLength - 2];
+  ElemType interp = (penultimateLambda - lambda1)
       / (penultimateLambda - ultimateLambda);
 
   betaPath[pathLength - 1] = (1 - interp) * (betaPath[pathLength - 2])
@@ -728,88 +1171,100 @@ inline void LARS::InterpolateBeta()
   lambdaPath[pathLength - 1] = lambda1;
 }
 
-inline void LARS::CholeskyInsert(const arma::vec& newX,
-                                 const arma::mat& X)
+template<typename ModelMatType>
+template<typename VecType, typename MatType>
+inline void LARS<ModelMatType>::CholeskyInsert(const VecType& newX,
+                                               const MatType& X)
 {
   if (matUtriCholFactor.n_rows == 0)
   {
-    matUtriCholFactor = arma::mat(1, 1);
+    matUtriCholFactor.set_size(1, 1);
 
     if (elasticNet)
-      matUtriCholFactor(0, 0) = sqrt(dot(newX, newX) + lambda2);
+      matUtriCholFactor(0, 0) = std::sqrt(dot(newX, newX) + lambda2);
     else
       matUtriCholFactor(0, 0) = norm(newX, 2);
   }
   else
   {
-    arma::vec newGramCol = trans(X) * newX;
+    VecType newGramCol = trans(X) * newX;
     CholeskyInsert(dot(newX, newX), newGramCol);
   }
 }
 
-inline void LARS::CholeskyInsert(double sqNormNewX,
-                                 const arma::vec& newGramCol)
+template<typename ModelMatType>
+template<typename VecType>
+inline void LARS<ModelMatType>::CholeskyInsert(
+    typename LARS<ModelMatType>::ElemType sqNormNewX,
+    const VecType& newGramCol)
 {
   int n = matUtriCholFactor.n_rows;
 
   if (n == 0)
   {
-    matUtriCholFactor = arma::mat(1, 1);
+    matUtriCholFactor.set_size(1, 1);
 
     if (elasticNet)
-      matUtriCholFactor(0, 0) = sqrt(sqNormNewX + lambda2);
+      matUtriCholFactor(0, 0) = std::sqrt(sqNormNewX + lambda2);
     else
-      matUtriCholFactor(0, 0) = sqrt(sqNormNewX);
+      matUtriCholFactor(0, 0) = std::sqrt(sqNormNewX);
   }
   else
   {
-    arma::mat matNewR = arma::mat(n + 1, n + 1);
+    DenseMatType matNewR(n + 1, n + 1);
 
     if (elasticNet)
       sqNormNewX += lambda2;
 
-    arma::vec matUtriCholFactork = solve(trimatl(trans(matUtriCholFactor)),
-        newGramCol);
+    arma::Col<ElemType> matUtriCholFactork =
+        solve(trimatl(trans(matUtriCholFactor)), newGramCol);
 
     matNewR(arma::span(0, n - 1), arma::span(0, n - 1)) = matUtriCholFactor;
     matNewR(arma::span(0, n - 1), n) = matUtriCholFactork;
     matNewR(n, arma::span(0, n - 1)).fill(0.0);
-    matNewR(n, n) = sqrt(sqNormNewX - dot(matUtriCholFactork,
-                                          matUtriCholFactork));
+    matNewR(n, n) = std::sqrt(sqNormNewX - dot(matUtriCholFactork,
+                                               matUtriCholFactork));
 
-    matUtriCholFactor = matNewR;
+    matUtriCholFactor = std::move(matNewR);
   }
 }
 
-inline void LARS::GivensRotate(const arma::vec::fixed<2>& x,
-                               arma::vec::fixed<2>& rotatedX,
-                               arma::mat& matG)
+template<typename ModelMatType>
+template<typename MatType>
+inline void LARS<ModelMatType>::GivensRotate(
+    const typename arma::Col<
+        typename LARS<ModelMatType>::ElemType
+    >::template fixed<2>& x,
+    typename arma::Col<
+        typename LARS<ModelMatType>::ElemType
+    >::template fixed<2>& rotatedX,
+    MatType& matG)
 {
   if (x(1) == 0)
   {
-    matG = arma::eye(2, 2);
+    matG = arma::eye<MatType>(2, 2);
     rotatedX = x;
   }
   else
   {
-    double r = norm(x, 2);
-    matG = arma::mat(2, 2);
+    ElemType r = norm(x, 2);
+    matG.set_size(2, 2);
 
-    double scaledX1 = x(0) / r;
-    double scaledX2 = x(1) / r;
+    ElemType scaledX1 = x(0) / r;
+    ElemType scaledX2 = x(1) / r;
 
     matG(0, 0) = scaledX1;
     matG(1, 0) = -scaledX2;
     matG(0, 1) = scaledX2;
     matG(1, 1) = scaledX1;
 
-    rotatedX = arma::vec(2);
     rotatedX(0) = r;
     rotatedX(1) = 0;
   }
 }
 
-inline void LARS::CholeskyDelete(const size_t colToKill)
+template<typename ModelMatType>
+inline void LARS<ModelMatType>::CholeskyDelete(const size_t colToKill)
 {
   size_t n = matUtriCholFactor.n_rows;
 
@@ -825,8 +1280,8 @@ inline void LARS::CholeskyDelete(const s
 
     for (size_t k = colToKill; k < n; ++k)
     {
-      arma::mat matG;
-      arma::vec::fixed<2> rotatedVec;
+      DenseMatType matG;
+      typename arma::Col<ElemType>::template fixed<2> rotatedVec;
       GivensRotate(matUtriCholFactor(arma::span(k, k + 1), k), rotatedVec,
           matG);
       matUtriCholFactor(arma::span(k, k + 1), k) = rotatedVec;
@@ -842,32 +1297,41 @@ inline void LARS::CholeskyDelete(const s
   }
 }
 
-inline double LARS::ComputeError(const arma::mat& matX,
-                                 const arma::rowvec& y,
-                                 const bool rowMajor)
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType>
+inline typename LARS<ModelMatType>::ElemType
+LARS<ModelMatType>::ComputeError(const MatType& matX,
+                                 const ResponsesType& y,
+                                 const bool colMajor)
 {
-  if (rowMajor)
-  {
-    return arma::accu(arma::pow(y - trans(matX * betaPath.back()), 2.0));
-  }
-
+  if (!colMajor)
+    return accu(pow(y - trans(matX * Beta()) - Intercept(), 2.0));
   else
-  {
-    return arma::accu(arma::pow(y - betaPath.back().t() * matX, 2.0));
-  }
+    return accu(pow(y - Beta().t() * matX - Intercept(), 2.0));
 }
 
 /**
  * Serialize the LARS model.
  */
+template<typename ModelMatType>
 template<typename Archive>
-void LARS::serialize(Archive& ar, const uint32_t /* version */)
+void LARS<ModelMatType>::serialize(Archive& ar, const uint32_t version)
 {
   // If we're loading, we have to use the internal storage.
   if (cereal::is_loading<Archive>())
   {
     matGram = &matGramInternal;
-    ar(CEREAL_NVP(matGramInternal));
+    if (version == 0)
+    {
+      // Older versions stored matGramInternal as type arma::mat.
+      arma::mat matGramInternalTmp;
+      ar(cereal::make_nvp("matGramInternal", matGramInternalTmp));
+      matGramInternal = ConvTo<ModelMatType>::From(matGramInternalTmp);
+    }
+    else
+    {
+      ar(CEREAL_NVP(matGramInternal));
+    }
   }
   else
   {
@@ -875,22 +1339,94 @@ void LARS::serialize(Archive& ar, const
         (const_cast<arma::mat&>(*matGram))));
   }
 
-  ar(CEREAL_NVP(matUtriCholFactor));
-  ar(CEREAL_NVP(useCholesky));
-  ar(CEREAL_NVP(lasso));
-  ar(CEREAL_NVP(lambda1));
-  ar(CEREAL_NVP(elasticNet));
-  ar(CEREAL_NVP(lambda2));
-  ar(CEREAL_NVP(tolerance));
-  ar(CEREAL_NVP(fitIntercept));
-  ar(CEREAL_NVP(normalizeData));
-  ar(CEREAL_NVP(betaPath));
-  ar(CEREAL_NVP(lambdaPath));
-  ar(CEREAL_NVP(interceptPath));
-  ar(CEREAL_NVP(activeSet));
-  ar(CEREAL_NVP(isActive));
-  ar(CEREAL_NVP(ignoreSet));
-  ar(CEREAL_NVP(isIgnored));
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // Older versions stored matUtriCholFactor as type arma::mat, and other
+    // elements as type double.  This version loads everything as
+    // double/arma::mat and converts as needed.
+    arma::mat matUtriCholFactorTmp;
+    ar(cereal::make_nvp("matUtriCholFactor", matUtriCholFactorTmp));
+    matUtriCholFactor = ConvTo<ModelMatType>::From(matUtriCholFactorTmp);
+
+    ar(CEREAL_NVP(useCholesky));
+    ar(CEREAL_NVP(lasso));
+
+    double tmp;
+    ar(cereal::make_nvp("lambda1", tmp));
+    lambda1 = tmp;
+
+    ar(CEREAL_NVP(elasticNet));
+
+    ar(cereal::make_nvp("lambda2", tmp));
+    lambda2 = tmp;
+
+    ar(cereal::make_nvp("tolerance", tmp));
+    tolerance = tmp;
+
+    ar(CEREAL_NVP(fitIntercept));
+    ar(CEREAL_NVP(normalizeData));
+
+    std::vector<arma::vec> betaPathTmp;
+    ar(cereal::make_nvp("betaPath", betaPathTmp));
+    betaPath.resize(betaPathTmp.size());
+    for (size_t i = 0; i < betaPathTmp.size(); ++i)
+      betaPath[i] = ConvTo<ModelColType>::From(betaPathTmp[i]);
+
+    std::vector<double> lambdaPathTmp;
+    ar(cereal::make_nvp("lambdaPath", lambdaPathTmp));
+    lambdaPath.resize(lambdaPathTmp.size());
+    for (size_t i = 0; i < lambdaPathTmp.size(); ++i)
+      lambdaPath[i] = (ElemType) lambdaPathTmp[i];
+
+    std::vector<double> interceptPathTmp;
+    ar(cereal::make_nvp("interceptPath", interceptPathTmp));
+    interceptPath.resize(interceptPathTmp.size());
+    for (size_t i = 0; i < interceptPathTmp.size(); ++i)
+      interceptPath[i] = (ElemType) interceptPathTmp[i];
+
+    ar(CEREAL_NVP(activeSet));
+    ar(CEREAL_NVP(isActive));
+    ar(CEREAL_NVP(ignoreSet));
+    ar(CEREAL_NVP(isIgnored));
+  }
+  else
+  {
+    ar(CEREAL_NVP(matUtriCholFactor));
+    ar(CEREAL_NVP(useCholesky));
+    ar(CEREAL_NVP(lasso));
+    ar(CEREAL_NVP(lambda1));
+    ar(CEREAL_NVP(elasticNet));
+    ar(CEREAL_NVP(lambda2));
+    ar(CEREAL_NVP(tolerance));
+    ar(CEREAL_NVP(fitIntercept));
+    ar(CEREAL_NVP(normalizeData));
+    ar(CEREAL_NVP(betaPath));
+    ar(CEREAL_NVP(lambdaPath));
+    ar(CEREAL_NVP(interceptPath));
+    ar(CEREAL_NVP(activeSet));
+    ar(CEREAL_NVP(isActive));
+    ar(CEREAL_NVP(ignoreSet));
+    ar(CEREAL_NVP(isIgnored));
+  }
+
+  if (version > 0)
+  {
+    ar(CEREAL_NVP(selectedLambda1));
+    ar(CEREAL_NVP(selectedIndex));
+    ar(CEREAL_NVP(selectedBeta));
+    ar(CEREAL_NVP(selectedIntercept));
+    ar(CEREAL_NVP(selectedActiveSet));
+    ar(CEREAL_NVP(offsetY));
+  }
+  else if (cereal::is_loading<Archive>())
+  {
+    selectedLambda1 = lambdaPath.back();
+    selectedIndex = betaPath.size() - 1;
+    selectedBeta.clear();
+    selectedIntercept = 0.0;
+    selectedActiveSet.clear();
+    offsetY = 0.0;
+  }
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/lars/lars_main.cpp 4.4.0-1/src/mlpack/methods/lars/lars_main.cpp
--- 4.3.0-2/src/mlpack/methods/lars/lars_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lars/lars_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -100,15 +100,15 @@ BINDING_EXAMPLE(
 // See also...
 BINDING_SEE_ALSO("@linear_regression", "#linear_regression");
 BINDING_SEE_ALSO("Least angle regression (pdf)",
-    "http://mlpack.org/papers/lars.pdf");
+    "https://mlpack.org/papers/lars.pdf");
 BINDING_SEE_ALSO("LARS C++ class documentation",
     "@src/mlpack/methods/lars/lars.hpp");
 
 PARAM_TMATRIX_IN("input", "Matrix of covariates (X).", "i");
 PARAM_MATRIX_IN("responses", "Matrix of responses/observations (y).", "r");
 
-PARAM_MODEL_IN(LARS, "input_model", "Trained LARS model to use.", "m");
-PARAM_MODEL_OUT(LARS, "output_model", "Output LARS model.", "M");
+PARAM_MODEL_IN(LARS<>, "input_model", "Trained LARS model to use.", "m");
+PARAM_MODEL_OUT(LARS<>, "output_model", "Output LARS model.", "M");
 
 PARAM_TMATRIX_IN("test", "Matrix containing points to regress on (test "
     "points).", "t");
@@ -149,11 +149,11 @@ void BINDING_FUNCTION(util::Params& para
       false, "no results will be saved");
   ReportIgnoredParam(params, {{ "test", true }}, "output_predictions");
 
-  LARS* lars;
+  LARS<>* lars;
   if (params.Has("input"))
   {
     // Initialize the object.
-    lars = new LARS(useCholesky, lambda1, lambda2);
+    lars = new LARS<>(useCholesky, lambda1, lambda2);
     lars->FitIntercept(!noIntercept);
     lars->NormalizeData(!noNormalize);
 
@@ -176,15 +176,14 @@ void BINDING_FUNCTION(util::Params& para
       Log::Fatal << "Number of responses must be equal to number of rows of X!"
           << endl;
 
-    vec beta;
     arma::rowvec y = std::move(matY);
     timers.Start("lars_regression");
-    lars->Train(matX, y, beta, false /* do not transpose */);
+    lars->Train(matX, y, false /* do not transpose */);
     timers.Stop("lars_regression");
   }
   else // We must have --input_model_file.
   {
-    lars = params.Get<LARS*>("input_model");
+    lars = params.Get<LARS<>*>("input_model");
   }
 
   if (params.Has("test"))
@@ -208,5 +207,5 @@ void BINDING_FUNCTION(util::Params& para
     params.Get<arma::mat>("output_predictions") = predictions.t();
   }
 
-  params.Get<LARS*>("output_model") = lars;
+  params.Get<LARS<>*>("output_model") = lars;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression.hpp 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression.hpp
--- 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,9 +26,13 @@ namespace mlpack {
  * Optionally, this class can perform ridge regression, if the lambda parameter
  * is set to a number greater than zero.
  */
+template<typename ModelMatType = arma::mat>
 class LinearRegression
 {
  public:
+  typedef typename GetColType<ModelMatType>::type ModelColType;
+  typedef typename ModelMatType::elem_type ElemType;
+
   /**
    * Creates the model.
    *
@@ -37,23 +41,37 @@ class LinearRegression
    * @param lambda Regularization constant for ridge regression.
    * @param intercept Whether or not to include an intercept term.
    */
-  LinearRegression(const arma::mat& predictors,
-                   const arma::rowvec& responses,
+  template<typename MatType,
+           typename ResponsesType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  LinearRegression(const MatType& predictors,
+                   const ResponsesType& responses,
                    const double lambda = 0,
                    const bool intercept = true);
 
   /**
-   * Creates the model with weighted learning.
+   * Creates the model with instance-weighted learning.
    *
    * @param predictors X, matrix of data points.
    * @param responses y, the measured data for each point in X.
-   * @param weights Observation weights (for boosting).
+   * @param weights Instance weights (for boosting).
    * @param lambda Regularization constant for ridge regression.
    * @param intercept Whether or not to include an intercept term.
    */
-  LinearRegression(const arma::mat& predictors,
-                   const arma::rowvec& responses,
-                   const arma::rowvec& weights,
+  template<typename MatType,
+           typename ResponsesType,
+           typename WeightsType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type,
+           typename = typename std::enable_if<
+               std::is_same<typename WeightsType::elem_type, ElemType>::value
+           >::type>
+  LinearRegression(const MatType& predictors,
+                   const ResponsesType& responses,
+                   const WeightsType& weights,
                    const double lambda = 0,
                    const bool intercept = true);
 
@@ -71,14 +89,55 @@ class LinearRegression
    * regularization parameter lambda, call Lambda() or set a different value in
    * the constructor.
    *
+   * This version of `Train()` is deprecated and will be removed in mlpack
+   * 5.0.0.  Use the version of `Train()` that specifies `lambda` before
+   * `intercept` instead.
+   *
+   * @param predictors X, the matrix of data points to train the model on.
+   * @param responses y, the responses to the data points.
+   * @param intercept Whether or not to fit an intercept term.
+   * @return The least squares error after training.
+   */
+  template<typename T>
+  mlpack_deprecated /** Will be removed in mlpack 5.0.0. */
+  double Train(const arma::mat& predictors,
+               const arma::rowvec& responses,
+               const T intercept,
+               const typename std::enable_if<std::is_same<T, bool>::value>::type* = 0);
+
+  /**
+   * Train the LinearRegression model on the given data and instance weights.
+   * Careful!  This will completely ignore and overwrite the existing model.
+   * This particular implementation does not have an incremental training
+   * algorithm.  To set the regularization parameter lambda, call Lambda() or
+   * set a different value in the constructor.
+   *
+   * This version of `Train()` is deprecated and will be removed in mlpack
+   * 5.0.0.  Use the version of `Train()` that specifies `lambda` before
+   * `intercept` instead.
+   *
    * @param predictors X, the matrix of data points to train the model on.
    * @param responses y, the responses to the data points.
+   * @param weights Instance weights (for boosting).
    * @param intercept Whether or not to fit an intercept term.
    * @return The least squares error after training.
    */
+  template<typename T>
+  mlpack_deprecated /** Will be removed in mlpack 5.0.0. */
   double Train(const arma::mat& predictors,
                const arma::rowvec& responses,
-               const bool intercept = true);
+               const arma::rowvec& weights,
+               const T intercept,
+               const typename std::enable_if<std::is_same<T, bool>::value>::type* = 0);
+
+  /**
+   * Train the LinearRegression model.  This is a dummy overload so that
+   * MetaInfoExtractor can properly detect that LinearRegression is a regression
+   * method.
+   */
+  template<typename MatType>
+  ElemType Train(const MatType& predictors,
+                 const arma::rowvec& responses);
 
   /**
    * Train the LinearRegression model on the given data and weights. Careful!
@@ -89,14 +148,67 @@ class LinearRegression
    *
    * @param predictors X, the matrix of data points to train the model on.
    * @param responses y, the responses to the data points.
-   * @param weights Observation weights (for boosting).
+   * @param lambda L2 regularization penalty parameter to use.
    * @param intercept Whether or not to fit an intercept term.
    * @return The least squares error after training.
    */
-  double Train(const arma::mat& predictors,
-               const arma::rowvec& responses,
-               const arma::rowvec& weights,
-               const bool intercept = true);
+  template<typename MatType,
+           typename ResponsesType,
+           typename = void, /* so MetaInfoExtractor does not get confused */
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& predictors,
+                 const ResponsesType& responses,
+                 const std::optional<double> lambda = std::nullopt,
+                 const std::optional<bool> intercept = std::nullopt);
+
+  /**
+   * Train the LinearRegression model.  This is a dummy overload so that
+   * MetaInfoExtractor can properly detect that LinearRegression is a regression
+   * method.
+   */
+  template<typename MatType>
+  ElemType Train(const MatType& predictors,
+                 const arma::rowvec& responses,
+                 const arma::rowvec& weights);
+
+  /**
+   * Train the LinearRegression model on the given data and instance weights.
+   * Careful!  This will completely ignore and overwrite the existing model.
+   * This particular implementation does not have an incremental training
+   * algorithm.  To set the regularization parameter lambda, call Lambda() or
+   * set a different value in the constructor.
+   *
+   * @param predictors X, the matrix of data points to train the model on.
+   * @param responses y, the responses to the data points.
+   * @param weights Instance weights (for boosting).
+   * @param lambda L2 regularization penalty parameter to use.
+   * @param intercept Whether or not to fit an intercept term.
+   * @return The least squares error after training.
+   */
+  template<typename MatType,
+           typename ResponsesType,
+           typename WeightsType,
+           typename = typename std::enable_if<
+               std::is_same<typename ResponsesType::elem_type, ElemType>::value
+           >::type,
+           typename = typename std::enable_if<
+               std::is_same<typename WeightsType::elem_type, ElemType>::value
+           >::type>
+  ElemType Train(const MatType& predictors,
+                 const ResponsesType& responses,
+                 const WeightsType& weights,
+                 const std::optional<double> lambda = std::nullopt,
+                 const std::optional<bool> intercept = std::nullopt);
+
+  /**
+   * Calculate y_i for a single data point.
+   *
+   * @param point the data point to calculate with.
+   */
+  template<typename VecType>
+  ElemType Predict(const VecType& point) const;
 
   /**
    * Calculate y_i for each data point in points.
@@ -104,7 +216,8 @@ class LinearRegression
    * @param points the data points to calculate with.
    * @param predictions y, will contain calculated values on completion.
    */
-  void Predict(const arma::mat& points, arma::rowvec& predictions) const;
+  template<typename MatType, typename ResponsesType>
+  void Predict(const MatType& points, ResponsesType& predictions) const;
 
   /**
    * Calculate the L2 squared error on the given predictors and responses using
@@ -123,13 +236,14 @@ class LinearRegression
    * @param points Matrix of predictors (X).
    * @param responses Transposed vector of responses (y^T).
    */
-  double ComputeError(const arma::mat& points,
-                      const arma::rowvec& responses) const;
+  template<typename MatType, typename ResponsesType>
+  ElemType ComputeError(const MatType& points,
+                        const ResponsesType& responses) const;
 
   //! Return the parameters (the b vector).
-  const arma::vec& Parameters() const { return parameters; }
+  const ModelColType& Parameters() const { return parameters; }
   //! Modify the parameters (the b vector).
-  arma::vec& Parameters() { return parameters; }
+  ModelColType& Parameters() { return parameters; }
 
   //! Return the Tikhonov regularization parameter for ridge regression.
   double Lambda() const { return lambda; }
@@ -143,19 +257,14 @@ class LinearRegression
    * Serialize the model.
    */
   template<typename Archive>
-  void serialize(Archive& ar, const uint32_t /* version */)
-  {
-    ar(CEREAL_NVP(parameters));
-    ar(CEREAL_NVP(lambda));
-    ar(CEREAL_NVP(intercept));
-  }
+  void serialize(Archive& ar, const uint32_t version);
 
  private:
   /**
    * The calculated B.
    * Initialized and filled by constructor to hold the least squares solution.
    */
-  arma::vec parameters;
+  ModelColType parameters;
 
   /**
    * The Tikhonov regularization parameter for ridge regression (0 for linear
@@ -169,6 +278,9 @@ class LinearRegression
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename ModelMatType),
+    (mlpack::LinearRegression<ModelMatType>), (1));
+
 // Include implementation.
 #include "linear_regression_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_impl.hpp 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_impl.hpp
--- 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,39 +17,114 @@
 
 namespace mlpack {
 
-inline LinearRegression::LinearRegression(
-    const arma::mat& predictors,
-    const arma::rowvec& responses,
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename>
+inline LinearRegression<ModelMatType>::LinearRegression(
+    const MatType& predictors,
+    const ResponsesType& responses,
     const double lambda,
     const bool intercept) :
-    LinearRegression(predictors, responses, arma::rowvec(), lambda, intercept)
+    LinearRegression(predictors, responses,
+        arma::Row<typename ResponsesType::elem_type>(), lambda, intercept)
 { /* Nothing to do. */ }
 
-inline LinearRegression::LinearRegression(
-    const arma::mat& predictors,
-    const arma::rowvec& responses,
-    const arma::rowvec& weights,
+template<typename ModelMatType>
+template<typename MatType,
+         typename ResponsesType,
+         typename WeightsType,
+         typename, typename>
+inline LinearRegression<ModelMatType>::LinearRegression(
+    const MatType& predictors,
+    const ResponsesType& responses,
+    const WeightsType& weights,
     const double lambda,
     const bool intercept) :
     lambda(lambda),
     intercept(intercept)
 {
-  Train(predictors, responses, weights, intercept);
+  Train(predictors, responses, weights, lambda, intercept);
 }
 
-inline double LinearRegression::Train(const arma::mat& predictors,
-                                      const arma::rowvec& responses,
-                                      const bool intercept)
+template<typename ModelMatType>
+template<typename T>
+mlpack_deprecated /** Will be removed in mlpack 5.0.0. */
+inline double LinearRegression<ModelMatType>::Train(
+    const arma::mat& predictors,
+    const arma::rowvec& responses,
+    const T intercept,
+    const typename std::enable_if<std::is_same<T, bool>::value>::type*)
+{
+  return Train(predictors, responses, arma::rowvec(), this->lambda, intercept);
+}
+
+template<typename ModelMatType>
+template<typename T>
+mlpack_deprecated /** Will be removed in mlpack 5.0.0. */
+inline double LinearRegression<ModelMatType>::Train(
+    const arma::mat& predictors,
+    const arma::rowvec& responses,
+    const arma::rowvec& weights,
+    const T intercept,
+    const typename std::enable_if<std::is_same<T, bool>::value>::type*)
 {
-  return Train(predictors, responses, arma::rowvec(), intercept);
+  return Train(predictors, responses, weights, this->lambda, intercept);
 }
 
-inline double LinearRegression::Train(const arma::mat& predictors,
+template<typename ModelMatType>
+template<typename MatType>
+inline
+typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::Train(const MatType& predictors,
+                                      const arma::rowvec& responses)
+{
+  return Train(predictors, responses, arma::rowvec(), this->lambda,
+      this->intercept);
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType, typename, typename>
+inline
+typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::Train(const MatType& predictors,
+                                      const ResponsesType& responses,
+                                      const std::optional<double> lambda,
+                                      const std::optional<bool> intercept)
+{
+  return Train(predictors, responses,
+      arma::Row<typename ResponsesType::elem_type>(),
+      (lambda.has_value()) ? lambda.value() : this->lambda,
+      (intercept.has_value()) ? intercept.value() : this->intercept);
+}
+
+template<typename ModelMatType>
+template<typename MatType>
+inline
+typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::Train(const MatType& predictors,
                                       const arma::rowvec& responses,
-                                      const arma::rowvec& weights,
-                                      const bool intercept)
+                                      const arma::rowvec& weights)
+{
+  return Train(predictors, responses, weights, this->lambda, this->intercept);
+}
+
+template<typename ModelMatType>
+template<typename MatType,
+         typename ResponsesType,
+         typename WeightsType,
+         typename, typename>
+inline
+typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::Train(const MatType& predictors,
+                                      const ResponsesType& responses,
+                                      const WeightsType& weights,
+                                      const std::optional<double> lambda,
+                                      const std::optional<bool> intercept)
 {
-  this->intercept = intercept;
+  if (lambda.has_value())
+    this->lambda = lambda.value();
+  
+  if (intercept.has_value())
+    this->intercept = intercept.value();
 
   /*
    * We want to calculate the a_i coefficients of:
@@ -66,15 +141,16 @@ inline double LinearRegression::Train(co
 
   const size_t nCols = predictors.n_cols;
 
-  arma::mat p = predictors;
-  arma::rowvec r = responses;
+  // TODO: avoid copy if possible.
+  arma::Mat<ElemType> p = ConvTo<arma::Mat<ElemType>>::From(predictors);
+  arma::Row<ElemType> r = responses;
 
   // Here we add the row of ones to the predictors.
   // The intercept is not penalized. Add an "all ones" row to design and set
   // intercept = false to get a penalized intercept.
-  if (intercept)
+  if (this->intercept)
   {
-    p.insert_rows(0, arma::ones<arma::mat>(1, nCols));
+    p.insert_rows(0, ones<arma::Mat<ElemType>>(1, nCols));
   }
 
   if (weights.n_elem > 0)
@@ -88,30 +164,63 @@ inline double LinearRegression::Train(co
   // Then we'll use Armadillo to solve it.
   // The total runtime of this should be O(d^2 N) + O(d^3) + O(dN).
   // (assuming the SVD is used to solve it)
-  arma::mat cov = p * p.t() +
-      lambda * arma::eye<arma::mat>(p.n_rows, p.n_rows);
+  arma::Mat<ElemType> cov = p * p.t() +
+      ((ElemType) this->lambda) *
+      arma::eye<arma::Mat<ElemType>>(p.n_rows, p.n_rows);
 
   parameters = arma::solve(cov, p * r.t());
   return ComputeError(predictors, responses);
 }
 
-inline void LinearRegression::Predict(
-    const arma::mat& points,
-    arma::rowvec& predictions) const
+template<typename ModelMatType>
+template<typename VecType>
+inline
+typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::Predict(const VecType& point) const
+{
+  if (intercept)
+  {
+    // We want to be sure we have the correct number of dimensions in the
+    // dataset.
+    // Prevent underflow.
+    const size_t dimensionality = (parameters.n_rows == 0) ? size_t(0) :
+        size_t(parameters.n_rows - 1);
+    util::CheckSameDimensionality(point, dimensionality,
+        "LinearRegression::Predict()", "point");
+
+    return dot(parameters.subvec(1, parameters.n_elem - 1).t(), point) +
+        parameters(0);
+  }
+  else
+  {
+    // We want to be sure we have the correct number of dimensions in
+    // the dataset.
+    util::CheckSameDimensionality(point, parameters,
+        "LinearRegression::Predict()", "point");
+
+    return dot(parameters.t(), point);
+  }
+}
+
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType>
+inline void LinearRegression<ModelMatType>::Predict(
+    const MatType& points,
+    ResponsesType& predictions) const
 {
   if (intercept)
   {
     // We want to be sure we have the correct number of dimensions in the
     // dataset.
     // Prevent underflow.
-    const size_t labels = (parameters.n_rows == 0) ? size_t(0) :
+    const size_t dimensionality = (parameters.n_rows == 0) ? size_t(0) :
         size_t(parameters.n_rows - 1);
-    util::CheckSameDimensionality(points, labels, "LinearRegression::Predict()", 
-        "points");
+    util::CheckSameDimensionality(points, dimensionality,
+        "LinearRegression::Predict()", "points");
+
     // Get the predictions, but this ignores the intercept value
     // (parameters[0]).
-    predictions = arma::trans(parameters.subvec(1, parameters.n_elem - 1))
-        * points;
+    predictions = parameters.subvec(1, parameters.n_elem - 1).t() * points;
     // Now add the intercept.
     predictions += parameters(0);
   }
@@ -119,26 +228,29 @@ inline void LinearRegression::Predict(
   {
     // We want to be sure we have the correct number of dimensions in
     // the dataset.
-    util::CheckSameDimensionality(points, parameters, 
+    util::CheckSameDimensionality(points, parameters,
         "LinearRegression::Predict()", "points");
-    predictions = arma::trans(parameters) * points;
+    predictions = trans(parameters) * points;
   }
 }
 
-inline double LinearRegression::ComputeError(
-    const arma::mat& predictors,
-    const arma::rowvec& responses) const
+template<typename ModelMatType>
+template<typename MatType, typename ResponsesType>
+inline typename LinearRegression<ModelMatType>::ElemType
+LinearRegression<ModelMatType>::ComputeError(
+    const MatType& predictors,
+    const ResponsesType& responses) const
 {
   // Sanity check on data.
   util::CheckSameSizes(predictors, responses, "LinearRegression::Train()");
-  
+
   // Get the number of columns and rows of the dataset.
   const size_t nCols = predictors.n_cols;
   const size_t nRows = predictors.n_rows;
 
   // Calculate the differences between actual responses and predicted responses.
   // We must also add the intercept (parameters(0)) to the predictions.
-  arma::rowvec temp;
+  arma::Row<typename ResponsesType::elem_type> temp;
   if (intercept)
   {
     // Ensure that we have the correct number of dimensions in the dataset.
@@ -148,7 +260,7 @@ inline double LinearRegression::ComputeE
           "training file." << std::endl;
     }
     temp = responses - (parameters(0) +
-        arma::trans(parameters.subvec(1, parameters.n_elem - 1)) * predictors);
+        parameters.subvec(1, parameters.n_elem - 1).t() * predictors);
   }
   else
   {
@@ -158,13 +270,34 @@ inline double LinearRegression::ComputeE
       Log::Fatal << "The test data must have the same number of columns as the "
           "training file." << std::endl;
     }
-    temp = responses - arma::trans(parameters) * predictors;
+    temp = responses - parameters.t() * predictors;
   }
-  const double cost = arma::dot(temp, temp) / nCols;
+  const ElemType cost = dot(temp, temp) / nCols;
 
   return cost;
 }
 
+template<typename ModelMatType>
+template<typename Archive>
+void LinearRegression<ModelMatType>::serialize(Archive& ar,
+                                               const uint32_t version)
+{
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // Old versions represented `parameters` as an arma::vec.
+    arma::vec parametersTmp;
+    ar(cereal::make_nvp("parameters", parametersTmp));
+    parameters = ConvTo<ModelColType>::From(parametersTmp);
+  }
+  else
+  {
+    ar(CEREAL_NVP(parameters));
+  }
+
+  ar(CEREAL_NVP(lambda));
+  ar(CEREAL_NVP(intercept));
+}
+
 } // namespace mlpack
 
 #endif
diff -pruN 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_main.cpp 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_main.cpp
--- 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -81,8 +81,6 @@ BINDING_EXAMPLE(
         "output_predictions", "X_test_responses"));
 
 // See also...
-BINDING_SEE_ALSO("Linear/ridge regression tutorial",
-    "@doc/tutorials/linear_regression.md");
 BINDING_SEE_ALSO("@lars", "#lars");
 BINDING_SEE_ALSO("Linear regression on Wikipedia",
     "https://en.wikipedia.org/wiki/Linear_regression");
@@ -95,9 +93,9 @@ PARAM_ROW_IN("training_responses", "Opti
     "(responses). If not given, the responses are assumed to be the last row "
     "of the input file.", "r");
 
-PARAM_MODEL_IN(LinearRegression, "input_model", "Existing LinearRegression "
+PARAM_MODEL_IN(LinearRegression<>, "input_model", "Existing LinearRegression "
     "model to use.", "m");
-PARAM_MODEL_OUT(LinearRegression, "output_model", "Output LinearRegression "
+PARAM_MODEL_OUT(LinearRegression<>, "output_model", "Output LinearRegression "
     "model.", "M");
 
 PARAM_MATRIX_IN("test", "Matrix containing X' (test regressors).", "T");
@@ -120,7 +118,7 @@ void BINDING_FUNCTION(util::Params& para
   mat regressors;
   rowvec responses;
 
-  LinearRegression* lr;
+  LinearRegression<>* lr;
 
   const bool computeModel = !params.Has("input_model");
   const bool computePrediction = params.Has("test");
@@ -172,14 +170,14 @@ void BINDING_FUNCTION(util::Params& para
     }
 
     timer.Start("regression");
-    lr = new LinearRegression(regressors, responses, lambda);
+    lr = new LinearRegression<>(regressors, responses, lambda);
     timer.Stop("regression");
   }
   else
   {
     // A model file was passed in, so load it.
     timer.Start("load_model");
-    lr = params.Get<LinearRegression*>("input_model");
+    lr = params.Get<LinearRegression<>*>("input_model");
     timer.Stop("load_model");
   }
 
@@ -221,5 +219,5 @@ void BINDING_FUNCTION(util::Params& para
   }
 
   // Save the model if needed.
-  params.Get<LinearRegression*>("output_model") = lr;
+  params.Get<LinearRegression<>*>("output_model") = lr;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_predict_main.cpp 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_predict_main.cpp
--- 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_predict_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_predict_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,8 +37,8 @@ BINDING_LONG_DESC("");
 BINDING_EXAMPLE(
   CALL_METHOD("model", "predict", "test", "X_test"));
 
-PARAM_MODEL_IN_REQ(LinearRegression, "input_model", "Existing LinearRegression "
-    "model to use.", "m");
+PARAM_MODEL_IN_REQ(LinearRegression<>, "input_model", "Existing "
+    "LinearRegression model to use.", "m");
 
 PARAM_MATRIX_IN_REQ("test", "Matrix containing X' (test regressors).", "T");
 
@@ -50,7 +50,7 @@ void BINDING_FUNCTION(util::Params& para
 {
   // A model file was passed in, so load it.
   timer.Start("load_model");
-  LinearRegression* lr = params.Get<LinearRegression*>("input_model");
+  LinearRegression<>* lr = params.Get<LinearRegression<>*>("input_model");
   timer.Stop("load_model");
 
   // Cache the output of GetPrintable before we std::move() the test
diff -pruN 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_train_main.cpp 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_train_main.cpp
--- 4.3.0-2/src/mlpack/methods/linear_regression/linear_regression_train_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_regression/linear_regression_train_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -46,22 +46,19 @@ BINDING_EXAMPLE(
     GET_DATASET("X", "https://example.com") + "\n" +
     GET_DATASET("y", "https://example.com") + "\n" +
     SPLIT_TRAIN_TEST("X", "y", "X_train", "y_train", "X_test", "y_test", 
-    "0.2") + "\n" +
-    CREATE_OBJECT("model", "linear_regression") + "\n" +
-    CALL_METHOD("model", "train", "training", "X_train",
-        "training_responses", "y_train"));
+        "0.2") + "\n" +
+    CREATE_OBJECT("lr", "linear_regression") + "\n" +
+    CALL_METHOD("lr", "train", "training", "X_train", "training_responses",
+        "y_train"));
 
 // See also...
-BINDING_SEE_ALSO("Linear/ridge regression tutorial",
-       "@doc/tutorials/linear_regression.md");
-
 PARAM_MATRIX_IN_REQ("training", "Matrix containing training set X (regressors).",
     "t");
 PARAM_ROW_IN("training_responses", "Optional vector containing y "
     "(responses). If not given, the responses are assumed to be the last row "
     "of the input file.", "r");
 
-PARAM_MODEL_OUT(LinearRegression, "output_model", "Output LinearRegression "
+PARAM_MODEL_OUT(LinearRegression<>, "output_model", "Output LinearRegression "
     "model.", "M");
 
 PARAM_DOUBLE_IN("lambda", "Tikhonov regularization for ridge regression.  If 0,"
@@ -108,9 +105,10 @@ void BINDING_FUNCTION(util::Params& para
     Log::Fatal << "Regressors and Responses must have the same number of data points!" << endl;
 
   timer.Start("regression");
-  LinearRegression* lr = new LinearRegression(regressors, responses, lambda);
+  LinearRegression<>* lr = new LinearRegression<>(regressors, responses,
+      lambda);
   timer.Stop("regression");
 
   // Save the model if needed.
-  params.Get<LinearRegression*>("output_model") = lr;
+  params.Get<LinearRegression<>*>("output_model") = lr;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm.hpp 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm.hpp
--- 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -73,18 +73,54 @@ namespace mlpack {
  * lsvm.Classify(test_data, predictions);
  * @endcode
  *
- * @tparam MatType Type of data matrix.
+ * @tparam ModelMatType Type of data matrix to use to store model parameters.
  */
-template <typename MatType = arma::mat>
+template<typename ModelMatType = arma::mat>
 class LinearSVM
 {
  public:
+  typedef typename ModelMatType::elem_type ElemType;
+  typedef typename GetDenseMatType<ModelMatType>::type DenseMatType;
+  typedef typename GetDenseColType<ModelMatType>::type DenseColType;
+
+  /**
+   * Initialize the Linear SVM without performing training.  Default
+   * value of lambda is 0.0001.  Be sure to use Train() before calling
+   * Classify() or ComputeAccuracy(), otherwise the results may be meaningless.
+   *
+   * @param lambda L2-regularization constant.
+   * @param delta Margin of difference between correct class and other classes.
+   * @param fitIntercept add intercept term or not.
+   */
+  LinearSVM();
+
+  /**
+   * Initialize the Linear SVM without performing training.  Default
+   * value of lambda is 0.0001.  Be sure to use Train() before calling
+   * Classify() or ComputeAccuracy(), otherwise the results may be meaningless.
+   *
+   * @param dimensionality Size of the input feature vector.
+   * @param numClasses Number of classes for classification.
+   * @param lambda L2-regularization constant.
+   * @param delta Margin of difference between correct class and other classes.
+   * @param fitIntercept add intercept term or not.
+   */
+  LinearSVM(const size_t dimensionality,
+            const size_t numClasses,
+            const double lambda = 0.0001,
+            const double delta = 1.0,
+            const bool fitIntercept = false);
+
   /**
    * Construct the LinearSVM class with the provided data and labels.
    * This will train the model. Optionally, the parameter 'lambda' can be
    * passed, which controls the amount of L2-regularization in the objective
    * function. By default, the model takes a small value.
    *
+   * This constructor is deprecated; if you want to specify a custom optimizer,
+   * use the constructor with the optimizer after `numClasses`.  The constructor
+   * will be removed in mlpack 5.0.0.
+   *
    * @tparam OptimizerType Desired differentiable separable optimizer
    * @tparam CallbackTypes Types of callback functions.
    * @param data Input training features. Each column associate with one sample
@@ -97,8 +133,18 @@ class LinearSVM
    * @param callbacks Callback functions.
    *      See https://www.ensmallen.org/docs.html#callback-documentation.
    */
-  template <typename OptimizerType, typename... CallbackTypes>
-  LinearSVM(const MatType& data,
+  template<typename OptimizerType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType,
+               LinearSVMFunction<arma::mat, ModelMatType>,
+               ModelMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  mlpack_deprecated /** To be removed in mlpack 5.0.0. **/
+  LinearSVM(const arma::mat& data,
             const arma::Row<size_t>& labels,
             const size_t numClasses,
             const double lambda,
@@ -113,6 +159,10 @@ class LinearSVM
    * passed, which controls the amount of L2-regularization in the objective
    * function. By default, the model takes a small value.
    *
+   * This constructor is deprecated; if you want to specify a custom optimizer,
+   * use the constructor with the optimizer after `numClasses`.  The constructor
+   * will be removed in mlpack 5.0.0.
+   *
    * @tparam OptimizerType Desired differentiable separable optimizer
    * @param data Input training features. Each column associate with one sample
    * @param labels Labels associated with the feature data.
@@ -122,45 +172,166 @@ class LinearSVM
    * @param fitIntercept add intercept term or not.
    * @param optimizer Desired optimizer.
    */
-  template <typename OptimizerType = ens::L_BFGS>
+  template<typename OptimizerType = ens::L_BFGS,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType,
+               LinearSVMFunction<arma::mat, ModelMatType>,
+               ModelMatType
+           >::value>::type>
+  mlpack_deprecated /** To be removed in mlpack 5.0.0. **/
+  LinearSVM(const arma::mat& data,
+            const arma::Row<size_t>& labels,
+            const size_t numClasses,
+            const double lambda,
+            const double delta,
+            const bool fitIntercept,
+            OptimizerType optimizer);
+
+  /**
+   * Construct the LinearSVM class with the provided data and labels.
+   * This will train the model. Optionally, hyperparameters can be passed.
+   *
+   * @tparam OptimizerType Desired differentiable separable optimizer
+   * @param data Input training features. Each column associate with one sample
+   * @param labels Labels associated with the feature data.
+   * @param numClasses Number of classes for classification.
+   * @param lambda L2-regularization constant.
+   * @param delta Margin of difference between correct class and other classes.
+   * @param fitIntercept add intercept term or not.
+   * @param callbacks Callback Functions.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   */
+  template<typename MatType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
   LinearSVM(const MatType& data,
             const arma::Row<size_t>& labels,
-            const size_t numClasses = 2,
+            const size_t numClasses,
             const double lambda = 0.0001,
             const double delta = 1.0,
             const bool fitIntercept = false,
-            OptimizerType optimizer = OptimizerType());
+            CallbackTypes&&... callbacks);
 
   /**
-   * Initialize the Linear SVM without performing training.  Default
-   * value of lambda is 0.0001.  Be sure to use Train() before calling
-   * Classify() or ComputeAccuracy(), otherwise the results may be meaningless.
+   * Construct the LinearSVM class with the provided data and labels.  This will
+   * train the model with the given custom optimizer. Optionally,
+   * hyperparameters can be passed.
    *
-   * @param inputSize Size of the input feature vector.
+   * @tparam OptimizerType Desired differentiable separable optimizer
+   * @param data Input training features. Each column associate with one sample
+   * @param labels Labels associated with the feature data.
    * @param numClasses Number of classes for classification.
    * @param lambda L2-regularization constant.
    * @param delta Margin of difference between correct class and other classes.
    * @param fitIntercept add intercept term or not.
+   * @param callbacks Callback Functions.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
    */
-  LinearSVM(const size_t inputSize,
-            const size_t numClasses = 0,
+  template<typename MatType,
+           typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  LinearSVM(const MatType& data,
+            const arma::Row<size_t>& labels,
+            const size_t numClasses,
+            OptimizerType& optimizer,
             const double lambda = 0.0001,
             const double delta = 1.0,
-            const bool fitIntercept = false);
+            const bool fitIntercept = false,
+            CallbackTypes&&... callbacks);
+
   /**
-   * Initialize the Linear SVM without performing training.  Default
-   * value of lambda is 0.0001.  Be sure to use Train() before calling
-   * Classify() or ComputeAccuracy(), otherwise the results may be meaningless.
+   * Train the Linear SVM with the given training data.
    *
+   * @param data Input training features. Each column associate with one sample.
+   * @param labels Labels associated with the feature data.
    * @param numClasses Number of classes for classification.
    * @param lambda L2-regularization constant.
    * @param delta Margin of difference between correct class and other classes.
    * @param fitIntercept add intercept term or not.
+   * @param callbacks Callback Functions.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return Objective value of the final point.
    */
-  LinearSVM(const size_t numClasses = 0,
-            const double lambda = 0.0001,
-            const double delta = 1.0,
-            const bool fitIntercept = false);
+  template<typename MatType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 CallbackTypes&&... callbackTypes);
+
+  template<typename MatType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const double lambda,
+                 const std::optional<double> delta = std::nullopt,
+                 const std::optional<bool> fitIntercept = std::nullopt,
+                 CallbackTypes&&... callbacks);
+
+  /**
+   * Train the Linear SVM with the given training data using a custom ensmallen
+   * optimizer.
+   *
+   * @tparam OptimizerType Desired optimizer.
+   * @param data Input training features. Each column associate with one sample.
+   * @param labels Labels associated with the feature data.
+   * @param numClasses Number of classes for classification.
+   * @param optimizer Desired optimizer.
+   * @param lambda L2-regularization constant.
+   * @param delta Margin of difference between correct class and other classes.
+   * @param fitIntercept add intercept term or not.
+   * @param callbacks Callback Functions.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return Objective value of the final point.
+   */
+  template<typename MatType,
+           typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType,
+               LinearSVMFunction<MatType, ModelMatType>,
+               ModelMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 OptimizerType optimizer,
+                 CallbackTypes&&... callbacks);
+
+  template<typename MatType,
+           typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType,
+               LinearSVMFunction<MatType, ModelMatType>,
+               ModelMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 OptimizerType optimizer,
+                 const double lambda,
+                 const std::optional<double> delta = std::nullopt,
+                 const std::optional<bool> fitIntercept = std::nullopt,
+                 CallbackTypes&&... callbacks);
 
   /**
    * Classify the given points, returning the predicted labels for each point.
@@ -171,6 +342,7 @@ class LinearSVM
    * @param data Set of points to classify.
    * @param labels Predicted labels for each point.
    */
+  template<typename MatType>
   void Classify(const MatType& data,
                 arma::Row<size_t>& labels) const;
 
@@ -185,9 +357,10 @@ class LinearSVM
    * @param labels Predicted labels for each point.
    * @param scores Class probabilities for each point.
    */
+  template<typename MatType>
   void Classify(const MatType& data,
                 arma::Row<size_t>& labels,
-                arma::mat& scores) const;
+                DenseMatType& scores) const;
 
   /**
    * Classify the given points, returning class scores for each point.
@@ -195,7 +368,8 @@ class LinearSVM
    * @param data Matrix of data points to be classified.
    * @param scores Class scores for each point.
    */
-  void Classify(const MatType& data,
+  mlpack_deprecated
+  void Classify(const arma::mat& data,
                 arma::mat& scores) const;
 
   /**
@@ -210,6 +384,20 @@ class LinearSVM
   size_t Classify(const VecType& point) const;
 
   /**
+   * Classify the given point. The predicted class label is stored in `label`,
+   * and the probability of each class is stored in `probabilities`..
+   *
+   * @param point Point to be classified.
+   * @param label size_t to store predicted label into.
+   * @param probabilities Vector to store class probabilities into.
+   * @return Predicted class label of the point.
+   */
+  template<typename VecType>
+  void Classify(const VecType& point,
+                size_t& label,
+                DenseColType& probabilities) const;
+
+  /**
    * Computes accuracy of the learned model given the feature data and the
    * labels associated with each data point. Predictions are made using the
    * provided data and are compared with the actual labels.
@@ -218,46 +406,10 @@ class LinearSVM
    * @param testLabels Vector of labels associated with the data.
    * @return Accuracy of the model.
    */
+  template<typename MatType>
   double ComputeAccuracy(const MatType& testData,
                          const arma::Row<size_t>& testLabels) const;
 
-  /**
-   * Train the Linear SVM with the given training data.
-   *
-   * @tparam OptimizerType Desired optimizer.
-   * @tparam CallbackTypes Types of Callback Functions.
-   * @param data Input training features. Each column associate with one sample.
-   * @param labels Labels associated with the feature data.
-   * @param numClasses Number of classes for classification.
-   * @param optimizer Desired optimizer.
-   * @param callbacks Callback Functions.
-   *      See https://www.ensmallen.org/docs.html#callback-documentation.
-   * @return Objective value of the final point.
-   */
-  template <typename OptimizerType, typename... CallbackTypes>
-  double Train(const MatType& data,
-               const arma::Row<size_t>& labels,
-               const size_t numClasses,
-               OptimizerType optimizer,
-               CallbackTypes&&... callbacks);
-
-  /**
-   * Train the Linear SVM with the given training data.
-   *
-   * @tparam OptimizerType Desired optimizer.
-   * @param data Input training features. Each column associate with one sample.
-   * @param labels Labels associated with the feature data.
-   * @param numClasses Number of classes for classification.
-   * @param optimizer Desired optimizer.
-   * @return Objective value of the final point.
-   */
-  template <typename OptimizerType = ens::L_BFGS>
-  double Train(const MatType& data,
-               const arma::Row<size_t>& labels,
-               const size_t numClasses = 2,
-               OptimizerType optimizer = OptimizerType());
-
-
   //! Sets the number of classes.
   size_t& NumClasses() { return numClasses; }
   //! Gets the number of classes.
@@ -277,9 +429,9 @@ class LinearSVM
   bool& FitIntercept() { return fitIntercept; }
 
   //! Set the model parameters.
-  arma::mat& Parameters() { return parameters; }
+  ModelMatType& Parameters() { return parameters; }
   //! Get the model parameters.
-  const arma::mat& Parameters() const { return parameters; }
+  const ModelMatType& Parameters() const { return parameters; }
 
   //! Gets the features size of the training data
   size_t FeatureSize() const
@@ -290,17 +442,11 @@ class LinearSVM
    * Serialize the LinearSVM model.
    */
   template<typename Archive>
-  void serialize(Archive& ar, const uint32_t /* version */)
-  {
-    ar(CEREAL_NVP(parameters));
-    ar(CEREAL_NVP(numClasses));
-    ar(CEREAL_NVP(lambda));
-    ar(CEREAL_NVP(fitIntercept));
-  }
+  void serialize(Archive& ar, const uint32_t /* version */);
 
  private:
   //! Parameters after optimization.
-  arma::mat parameters;
+  ModelMatType parameters;
   //! Number of classes.
   size_t numClasses;
   //! L2-Regularization constant.
@@ -313,6 +459,9 @@ class LinearSVM
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename ModelMatType),
+    (mlpack::LinearSVM<ModelMatType>), (1));
+
 // Include implementation.
 #include "linear_svm_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_function.hpp 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_function.hpp
--- 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -23,10 +23,16 @@ namespace mlpack {
  * This is used by various ensmallen optimizers to train the linear
  * SVM model.
  */
-template <typename MatType = arma::mat>
+template<typename MatType = arma::mat, typename ParametersType = arma::mat>
 class LinearSVMFunction
 {
  public:
+  typedef typename ParametersType::elem_type ElemType;
+  typedef typename GetDenseMatType<ParametersType>::type DenseMatType;
+  typedef typename GetSparseMatType<ParametersType>::type SparseMatType;
+  typedef typename GetDenseColType<SparseMatType>::type DenseColType;
+  typedef typename GetDenseRowType<ParametersType>::type DenseRowType;
+
   /**
    * Construct the Linear SVM objective function with given parameters.
    *
@@ -58,7 +64,7 @@ class LinearSVMFunction
    * @param numClasses Number of classes for classification.
    * @param fitIntercept If true, an intercept is fitted.
    */
-  static void InitializeWeights(arma::mat& weights,
+  static void InitializeWeights(ParametersType& weights,
                                 const size_t featureSize,
                                 const size_t numClasses,
                                 const bool fitIntercept = false);
@@ -67,10 +73,10 @@ class LinearSVMFunction
    * Constructs the ground truth label matrix with the passed labels.
    *
    * @param labels Labels associated with the training data.
-   * @param groundTruth Pointer to arma::mat which stores the computed matrix.
+   * @param groundTruth Reference to sparse matrix to stores the result.
    */
   void GetGroundTruthMatrix(const arma::Row<size_t>& labels,
-                            arma::sp_mat& groundTruth);
+                            SparseMatType& groundTruth) const;
 
   /**
    * Evaluate the hinge loss function for all the datapoints
@@ -78,7 +84,7 @@ class LinearSVMFunction
    * @param parameters The parameters of the SVM.
    * @return The value of the loss function for the entire dataset.
    */
-  double Evaluate(const arma::mat& parameters);
+  ElemType Evaluate(const ParametersType& parameters) const;
 
   /**
    * Evaluate the hinge loss function on the specified datapoints.
@@ -89,9 +95,9 @@ class LinearSVMFunction
    * @param batchSize Size of batch to process.
    * @return The value of the loss function for the given parameters.
    */
-  double Evaluate(const arma::mat& parameters,
-                  const size_t firstId,
-                  const size_t batchSize = 1);
+  ElemType Evaluate(const ParametersType& parameters,
+                    const size_t firstId,
+                    const size_t batchSize = 1) const;
 
   /**
    * Evaluate the gradient of the hinge loss function following the
@@ -101,9 +107,9 @@ class LinearSVMFunction
    * @param parameters The parameters of the SVM.
    * @param gradient Linear matrix to output the gradient into.
    */
-  template <typename GradType>
-  void Gradient(const arma::mat& parameters,
-                GradType& gradient);
+  template<typename GradType>
+  void Gradient(const ParametersType& parameters,
+                GradType& gradient) const;
 
   /**
    * Evaluate the gradient of the hinge loss function, following
@@ -115,11 +121,11 @@ class LinearSVMFunction
    * @param gradient Linear matrix to output the gradient into.
    * @param batchSize Size of the batch to process.
    */
-  template <typename GradType>
-  void Gradient(const arma::mat& parameters,
+  template<typename GradType>
+  void Gradient(const ParametersType& parameters,
                 const size_t firstId,
                 GradType& gradient,
-                const size_t batchSize = 1);
+                const size_t batchSize = 1) const;
 
   /**
    * Evaluate the gradient of the hinge loss function, following
@@ -132,9 +138,9 @@ class LinearSVMFunction
    * @param gradient Linear matrix to output the gradient into.
    * @return The value of the loss function at the given parameters.
    */
-  template <typename GradType>
-  double EvaluateWithGradient(const arma::mat& parameters,
-                              GradType& gradient) const;
+  template<typename GradType>
+  ElemType EvaluateWithGradient(const ParametersType& parameters,
+                                GradType& gradient) const;
 
   /**
    * Evaluate the gradient of the hinge loss function, following
@@ -150,21 +156,21 @@ class LinearSVMFunction
    * @param batchSize Size of the batch to process.
    * @return The value of the loss function at the given parameters.
    */
-  template <typename GradType>
-  double EvaluateWithGradient(const arma::mat& parameters,
-                              const size_t firstId,
-                              GradType& gradient,
-                              const size_t batchSize = 1) const;
+  template<typename GradType>
+  ElemType EvaluateWithGradient(const ParametersType& parameters,
+                                const size_t firstId,
+                                GradType& gradient,
+                                const size_t batchSize = 1) const;
 
   //! Return the initial point for the optimization.
-  const arma::mat& InitialPoint() const { return initialPoint; }
+  const ParametersType& InitialPoint() const { return initialPoint; }
   //! Modify the initial point for the optimization.
-  arma::mat& InitialPoint() { return initialPoint; }
+  ParametersType& InitialPoint() { return initialPoint; }
 
   //! Get the dataset.
-  const arma::sp_mat& Dataset() const { return dataset; }
+  const SparseMatType& Dataset() const { return dataset; }
   //! Modify the dataset.
-  arma::sp_mat& Dataset() { return dataset; }
+  SparseMatType& Dataset() { return dataset; }
 
   //! Sets the regularization parameter.
   double& Lambda() { return lambda; }
@@ -179,12 +185,13 @@ class LinearSVMFunction
 
  private:
   //! The initial point, from which to start the optimization.
-  arma::mat initialPoint;
+  ParametersType initialPoint;
 
   //! Label matrix for provided data
-  arma::sp_mat groundTruth;
+  SparseMatType groundTruth;
 
-  //! The datapoints for training.
+  //! The datapoints for training.  This will be an alias until Shuffle() is
+  //! called.
   MatType dataset;
 
   //! Number of Classes.
diff -pruN 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_function_impl.hpp 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -22,20 +22,21 @@
 
 namespace mlpack {
 
-template <typename MatType>
-LinearSVMFunction<MatType>::LinearSVMFunction(
-    const MatType& dataset,
+template<typename MatType, typename ParametersType>
+LinearSVMFunction<MatType, ParametersType>::LinearSVMFunction(
+    const MatType& datasetIn,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
     const double delta,
     const bool fitIntercept) :
-    dataset(MakeAlias(const_cast<MatType&>(dataset), false)),
     numClasses(numClasses),
     lambda(lambda),
     delta(delta),
     fitIntercept(fitIntercept)
 {
+  MakeAlias(dataset, datasetIn, datasetIn.n_rows, datasetIn.n_cols, 0, false);
+
   InitializeWeights(initialPoint, dataset.n_rows, numClasses, fitIntercept);
   initialPoint *= 0.005;
 
@@ -48,9 +49,9 @@ LinearSVMFunction<MatType>::LinearSVMFun
  * normal distribution. The weights cannot be initialized to zero, as that will
  * lead to each class output being the same.
  */
-template <typename MatType>
-void LinearSVMFunction<MatType>::InitializeWeights(
-    arma::mat &weights,
+template<typename MatType, typename ParametersType>
+void LinearSVMFunction<MatType, ParametersType>::InitializeWeights(
+    ParametersType& weights,
     const size_t featureSize,
     const size_t numClasses,
     const bool fitIntercept)
@@ -69,10 +70,11 @@ void LinearSVMFunction<MatType>::Initial
  * labels. The output is in the form of a matrix, which leads to simpler
  * calculations in the Evaluate() and Gradient() methods.
  */
-template <typename MatType>
-void LinearSVMFunction<MatType>::GetGroundTruthMatrix(
+template<typename MatType, typename ParametersType>
+void LinearSVMFunction<MatType, ParametersType>::GetGroundTruthMatrix(
     const arma::Row<size_t>& labels,
-    arma::sp_mat& groundTruth)
+    typename LinearSVMFunction<MatType, ParametersType>::SparseMatType&
+        groundTruth) const
 {
   // Calculate the ground truth matrix according to the labels passed. The
   // ground truth matrix is a matrix of dimensions 'numClasses * numExamples',
@@ -95,19 +97,19 @@ void LinearSVMFunction<MatType>::GetGrou
   }
 
   // All entries are '1'.
-  arma::vec values;
+  DenseColType values;
   values.ones(labels.n_elem);
 
   // Calculate the matrix.
-  groundTruth = arma::sp_mat(rowPointers, colPointers, values, numClasses,
-                             labels.n_elem);
+  groundTruth = SparseMatType(rowPointers, colPointers, values, numClasses,
+                              labels.n_elem);
 }
 
 /**
  * Shuffle the data.
  */
-template <typename MatType>
-void LinearSVMFunction<MatType>::Shuffle()
+template<typename MatType, typename ParametersType>
+void LinearSVMFunction<MatType, ParametersType>::Shuffle()
 {
   // Determine new ordering.
   arma::uvec ordering = arma::shuffle(arma::linspace<arma::uvec>(0,
@@ -124,26 +126,27 @@ void LinearSVMFunction<MatType>::Shuffle
     reverseOrdering[ordering[i]] = i;
 
   arma::umat newLocations(2, groundTruth.n_nonzero);
-  arma::vec values(groundTruth.n_nonzero);
-  arma::sp_mat::const_iterator it = groundTruth.begin();
+  DenseColType values(groundTruth.n_nonzero);
+  typename SparseMatType::const_iterator it = groundTruth.begin();
   size_t loc = 0;
   while (it != groundTruth.end())
   {
-    newLocations(0, loc) = reverseOrdering(it.col());
-    newLocations(1, loc) = it.row();
+    newLocations(0, loc) = it.row();
+    newLocations(1, loc) = reverseOrdering(it.col());
     values(loc) = (*it);
 
     ++it;
     ++loc;
   }
 
-  groundTruth = arma::sp_mat(newLocations, values, groundTruth.n_rows,
-                             groundTruth.n_cols);
+  groundTruth = SparseMatType(newLocations, values, groundTruth.n_rows,
+                              groundTruth.n_cols);
 }
 
-template <typename MatType>
-double LinearSVMFunction<MatType>::Evaluate(
-    const arma::mat& parameters)
+template<typename MatType, typename ParametersType>
+typename LinearSVMFunction<MatType, ParametersType>::ElemType
+LinearSVMFunction<MatType, ParametersType>::Evaluate(
+    const ParametersType& parameters) const
 {
   // The objective function is the hinge loss function and it is
   // calculated over all the training examples.
@@ -151,10 +154,10 @@ double LinearSVMFunction<MatType>::Evalu
   // Calculate the loss and regularization terms.
   // L_i = Σ_i Σ_m max(0, Δ + (w_m x_i + b_m) - (w_{y_i} x_i + b_{y_i}))
   // where (m != y_i)
-  double loss, regularization;
+  ElemType loss, regularization;
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   // Check intercept condition.
   if (!fitIntercept)
@@ -169,8 +172,7 @@ double LinearSVMFunction<MatType>::Evalu
     // On calculating the score, we add `b_i` term to each element of
     // `i_th` row of `scores`.
     scores = parameters.rows(0, dataset.n_rows - 1).t() * dataset
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1,
-        dataset.n_cols);
+        + repmat(parameters.row(dataset.n_rows).t(), 1, dataset.n_cols);
   }
 
   // Evaluate the margin by the following steps:
@@ -178,32 +180,34 @@ double LinearSVMFunction<MatType>::Evalu
   //  - Adding the margin parameter `delta`.
   //  - Removing the `delta` parameter from correct class label in each
   //    column.
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
+  DenseMatType margin = scores - (repmat(ones(numClasses).t()
       * (scores % groundTruth), numClasses, 1)) + delta
       - (delta * groundTruth);
 
   // The Hinge Loss Function
-  loss = arma::accu(arma::clamp(margin, 0.0, DBL_MAX)) / dataset.n_cols;
+  loss = accu(arma::clamp(margin, 0.0, DBL_MAX)) / dataset.n_cols;
 
   // Adding the regularization term.
-  regularization = 0.5 * lambda * arma::dot(parameters, parameters);
+  constexpr ElemType half = ((ElemType) 0.5);
+  regularization = half * lambda * dot(parameters, parameters);
 
   return loss + regularization;
 }
 
-template <typename MatType>
-double LinearSVMFunction<MatType>::Evaluate(
-    const arma::mat& parameters,
+template<typename MatType, typename ParametersType>
+typename LinearSVMFunction<MatType, ParametersType>::ElemType
+LinearSVMFunction<MatType, ParametersType>::Evaluate(
+    const ParametersType& parameters,
     const size_t firstId,
-    const size_t batchSize)
+    const size_t batchSize) const
 {
   const size_t lastId = firstId + batchSize - 1;
 
   // Calculate the loss and regularization terms.
-  double loss, regularization, cost;
+  ElemType loss, regularization, cost;
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   // Check intercept condition.
   if (!fitIntercept)
@@ -214,30 +218,31 @@ double LinearSVMFunction<MatType>::Evalu
   {
     scores = parameters.rows(0, dataset.n_rows - 1).t()
         * dataset.cols(firstId, lastId)
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1,
+        + repmat(parameters.row(dataset.n_rows).t(), 1,
         dataset.n_cols);
   }
 
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
+  DenseMatType margin = scores - (repmat(ones(numClasses).t()
       * (scores % groundTruth.cols(firstId, lastId)), numClasses, 1))
       + delta - (delta * groundTruth.cols(firstId, lastId));
 
   // The Hinge Loss Function
-  loss = arma::accu(arma::clamp(margin, 0.0, DBL_MAX));
+  loss = accu(arma::clamp(margin, 0.0, DBL_MAX));
   loss /= batchSize;
 
   // Adding the regularization term.
-  regularization = 0.5 * lambda * arma::dot(parameters, parameters);
+  constexpr ElemType half = ((ElemType) 0.5);
+  regularization = half * lambda * dot(parameters, parameters);
 
   cost = loss + regularization;
   return cost;
 }
 
-template <typename MatType>
-template <typename GradType>
-void LinearSVMFunction<MatType>::Gradient(
-    const arma::mat& parameters,
-    GradType& gradient)
+template<typename MatType, typename ParametersType>
+template<typename GradType>
+void LinearSVMFunction<MatType, ParametersType>::Gradient(
+    const ParametersType& parameters,
+    GradType& gradient) const
 {
   // The objective is to minimize the loss, which is evaluated as the sum
   // of all the positive elements of `margin` matrix.
@@ -245,7 +250,7 @@ void LinearSVMFunction<MatType>::Gradien
   // Also, we need to increase the score of the correct class.
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   if (!fitIntercept)
   {
@@ -254,21 +259,20 @@ void LinearSVMFunction<MatType>::Gradien
   else
   {
     scores = parameters.rows(0, dataset.n_rows - 1).t() * dataset
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1,
-        dataset.n_cols);
+        + repmat(parameters.row(dataset.n_rows).t(), 1, dataset.n_cols);
   }
 
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
+  DenseMatType margin = scores - (repmat(ones(numClasses).t()
       * (scores % groundTruth), numClasses, 1)) + delta
       - (delta * groundTruth);
 
   // An element of `mask` matrix holds `1` corresponding to
   // each positive element of `margin` matrix.
-  arma::mat mask = margin.for_each([](arma::mat::elem_type& val)
+  DenseMatType mask = margin.for_each([](arma::mat::elem_type& val)
       { val = (val > 0) ? 1: 0; });
 
-  arma::mat difference = groundTruth
-      % (-arma::repmat(arma::sum(mask), numClasses, 1)) + mask;
+  DenseMatType difference = groundTruth
+      % (-repmat(sum(mask), numClasses, 1)) + mask;
 
   // The gradient is evaluated as follows:
   //  - Add `x_i` to `w_j` if `margin_i_m`is positive.
@@ -288,7 +292,7 @@ void LinearSVMFunction<MatType>::Gradien
     gradient.submat(0, 0, parameters.n_rows - 2, parameters.n_cols - 1) =
         dataset * difference.t();
     gradient.row(parameters.n_rows - 1) =
-        arma::ones<arma::rowvec>(dataset.n_cols) * difference.t();
+        ones<DenseRowType>(dataset.n_cols) * difference.t();
   }
 
   gradient /= dataset.n_cols;
@@ -297,18 +301,18 @@ void LinearSVMFunction<MatType>::Gradien
   gradient += lambda * parameters;
 }
 
-template <typename MatType>
-template <typename GradType>
-void LinearSVMFunction<MatType>::Gradient(
-    const arma::mat& parameters,
+template<typename MatType, typename ParametersType>
+template<typename GradType>
+void LinearSVMFunction<MatType, ParametersType>::Gradient(
+    const ParametersType& parameters,
     const size_t firstId,
     GradType& gradient,
-    const size_t batchSize)
+    const size_t batchSize) const
 {
   const size_t lastId = firstId + batchSize - 1;
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   // Check intercept condition.
   if (!fitIntercept)
@@ -319,20 +323,20 @@ void LinearSVMFunction<MatType>::Gradien
   {
     scores = parameters.rows(0, dataset.n_rows - 1).t()
         * dataset.cols(firstId, lastId)
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1, batchSize);
+        + repmat(parameters.row(dataset.n_rows).t(), 1, batchSize);
   }
 
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
+  DenseMatType margin = scores - (repmat(ones(numClasses).t()
       * (scores % groundTruth.cols(firstId, lastId)), numClasses, 1))
       + delta - (delta * groundTruth.cols(firstId, lastId));
 
   // For each sample, find the total number of classes where
   // ( margin > 0 ).
-  arma::mat mask = margin.for_each([](arma::mat::elem_type& val)
+  DenseMatType mask = margin.for_each([](arma::mat::elem_type& val)
       { val = (val > 0) ? 1: 0; });
 
-  arma::mat difference = groundTruth.cols(firstId, lastId)
-      % (-arma::repmat(arma::sum(mask), numClasses, 1)) + mask;
+  DenseMatType difference = groundTruth.cols(firstId, lastId)
+      % (-repmat(sum(mask), numClasses, 1)) + mask;
 
   // Check intercept condition
   if (!fitIntercept)
@@ -345,7 +349,7 @@ void LinearSVMFunction<MatType>::Gradien
     gradient.submat(0, 0, parameters.n_rows - 2, parameters.n_cols - 1) =
         dataset.cols(firstId, lastId) * difference.t();
     gradient.row(parameters.n_rows - 1) =
-        arma::ones<arma::rowvec>(batchSize) * difference.t();
+        ones<DenseRowType>(batchSize) * difference.t();
   }
 
   gradient /= batchSize;
@@ -354,16 +358,17 @@ void LinearSVMFunction<MatType>::Gradien
   gradient += lambda * parameters;
 }
 
-template <typename MatType>
-template <typename GradType>
-double LinearSVMFunction<MatType>::EvaluateWithGradient(
-    const arma::mat& parameters,
+template<typename MatType, typename ParametersType>
+template<typename GradType>
+typename LinearSVMFunction<MatType, ParametersType>::ElemType
+LinearSVMFunction<MatType, ParametersType>::EvaluateWithGradient(
+    const ParametersType& parameters,
     GradType& gradient) const
 {
-  double loss, regularization, cost;
+  ElemType loss, regularization, cost;
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   if (!fitIntercept)
   {
@@ -372,21 +377,21 @@ double LinearSVMFunction<MatType>::Evalu
   else
   {
     scores = parameters.rows(0, dataset.n_rows - 1).t() * dataset
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1,
+        + repmat(parameters.row(dataset.n_rows).t(), 1,
         dataset.n_cols);
   }
 
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
-      * (scores % groundTruth), numClasses, 1)) + delta
-      - (delta * groundTruth);
+  DenseMatType margin = scores - (repmat(
+      ones<DenseColType>(numClasses).t() * (scores % groundTruth),
+      numClasses, 1)) + delta - (delta * groundTruth);
 
   // For each sample, find the total number of classes where
   // ( margin > 0 ).
-  arma::mat mask = margin.for_each([](arma::mat::elem_type& val)
+  DenseMatType mask = margin.for_each([](ElemType& val)
       { val = (val > 0) ? 1: 0; });
 
-  arma::mat difference = groundTruth
-      % (-arma::repmat(arma::sum(mask), numClasses, 1)) + mask;
+  DenseMatType difference = groundTruth
+      % (-repmat(sum(mask), numClasses, 1)) + mask;
 
   // Check intercept condition
   if (!fitIntercept)
@@ -399,7 +404,7 @@ double LinearSVMFunction<MatType>::Evalu
     gradient.submat(0, 0, parameters.n_rows - 2, parameters.n_cols - 1) =
             dataset * difference.t();
     gradient.row(parameters.n_rows - 1) =
-            arma::ones<arma::rowvec>(dataset.n_cols) * difference.t();
+            ones<DenseRowType>(dataset.n_cols) * difference.t();
   }
 
   gradient /= dataset.n_cols;
@@ -408,20 +413,22 @@ double LinearSVMFunction<MatType>::Evalu
   gradient += lambda * parameters;
 
   // The Hinge Loss Function
-  loss = arma::accu(arma::clamp(margin, 0.0, DBL_MAX));
+  loss = accu(arma::clamp(margin, 0.0, DBL_MAX));
   loss /= dataset.n_cols;
 
   // Adding the regularization term.
-  regularization = 0.5 * lambda * arma::dot(parameters, parameters);
+  constexpr ElemType half = ((ElemType) 0.5);
+  regularization = half * lambda * dot(parameters, parameters);
 
   cost = loss + regularization;
   return cost;
 }
 
-template <typename MatType>
-template <typename GradType>
-double LinearSVMFunction<MatType>::EvaluateWithGradient(
-    const arma::mat& parameters,
+template<typename MatType, typename ParametersType>
+template<typename GradType>
+typename LinearSVMFunction<MatType, ParametersType>::ElemType
+LinearSVMFunction<MatType, ParametersType>::EvaluateWithGradient(
+    const ParametersType& parameters,
     const size_t firstId,
     GradType& gradient,
     const size_t batchSize) const
@@ -429,10 +436,10 @@ double LinearSVMFunction<MatType>::Evalu
   const size_t lastId = firstId + batchSize - 1;
 
   // Calculate the loss and regularization terms.
-  double loss, regularization, cost;
+  ElemType loss, regularization, cost;
 
   // Scores for each class are evaluated.
-  arma::mat scores;
+  DenseMatType scores;
 
   // Check intercept condition.
   if (!fitIntercept)
@@ -443,20 +450,20 @@ double LinearSVMFunction<MatType>::Evalu
   {
     scores = parameters.rows(0, dataset.n_rows - 1).t()
         * dataset.cols(firstId, lastId)
-        + arma::repmat(parameters.row(dataset.n_rows).t(), 1, dataset.n_cols);
+        + repmat(parameters.row(dataset.n_rows).t(), 1, (lastId - firstId + 1));
   }
 
-  arma::mat margin = scores - (arma::repmat(arma::ones(numClasses).t()
+  DenseMatType margin = scores - (repmat(ones(numClasses).t()
       * (scores % groundTruth.cols(firstId, lastId)), numClasses, 1))
       + delta - (delta * groundTruth.cols(firstId, lastId));
 
   // For each sample, find the total number of classes where
   // ( margin > 0 ).
-  arma::mat mask = margin.for_each([](arma::mat::elem_type& val)
+  DenseMatType mask = margin.for_each([](arma::mat::elem_type& val)
       { val = (val > 0) ? 1: 0; });
 
-  arma::mat difference = groundTruth.cols(firstId, lastId)
-      % (-arma::repmat(arma::sum(mask), numClasses, 1)) + mask;
+  DenseMatType difference = groundTruth.cols(firstId, lastId)
+      % (-repmat(sum(mask), numClasses, 1)) + mask;
 
   // Check intercept condition
   if (!fitIntercept)
@@ -469,7 +476,7 @@ double LinearSVMFunction<MatType>::Evalu
     gradient.submat(0, 0, parameters.n_rows - 2, parameters.n_cols - 1) =
         dataset.cols(firstId, lastId) * difference.t();
     gradient.row(parameters.n_rows - 1) =
-        arma::ones<arma::rowvec>(batchSize) * difference.t();
+        ones<DenseRowType>(batchSize) * difference.t();
   }
 
   gradient /= batchSize;
@@ -479,18 +486,19 @@ double LinearSVMFunction<MatType>::Evalu
   gradient += lambda * parameters;
 
   // The Hinge Loss Function
-  loss = arma::accu(arma::clamp(margin.cols(firstId, lastId), 0.0, DBL_MAX));
+  loss = accu(arma::clamp(margin, 0.0, DBL_MAX));
   loss /= batchSize;
 
   // Adding the regularization term.
-  regularization = 0.5 * lambda * arma::dot(parameters, parameters);
+  constexpr ElemType half = ((ElemType) 0.5);
+  regularization = half * lambda * dot(parameters, parameters);
 
   cost = loss + regularization;
   return cost;
 }
 
-template <typename MatType>
-size_t LinearSVMFunction<MatType>::NumFunctions() const
+template<typename MatType, typename ParametersType>
+size_t LinearSVMFunction<MatType, ParametersType>::NumFunctions() const
 {
   // The number of points in the dataset is the number of functions, as this
   // is a data dependent function.
diff -pruN 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_impl.hpp 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_impl.hpp
--- 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,10 +17,37 @@
 
 namespace mlpack {
 
-template <typename MatType>
-template <typename OptimizerType, typename... CallbackTypes>
-LinearSVM<MatType>::LinearSVM(
-    const MatType& data,
+template<typename ModelMatType>
+LinearSVM<ModelMatType>::LinearSVM() :
+    lambda(0.0001),
+    delta(1.0),
+    fitIntercept(false)
+{
+  // No training to do here.
+}
+
+template<typename ModelMatType>
+LinearSVM<ModelMatType>::LinearSVM(
+    const size_t dimensionality,
+    const size_t numClasses,
+    const double lambda,
+    const double delta,
+    const bool fitIntercept) :
+    numClasses(numClasses),
+    lambda(lambda),
+    delta(delta),
+    fitIntercept(fitIntercept)
+{
+  LinearSVMFunction<ModelMatType /* fake, does not matter for this call */,
+                    ModelMatType>::InitializeWeights(
+      parameters, dimensionality, numClasses, fitIntercept);
+}
+
+template<typename ModelMatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+mlpack_deprecated /** Will be removed in mlpack 5.0.0. **/
+LinearSVM<ModelMatType>::LinearSVM(
+    const arma::mat& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
@@ -33,13 +60,15 @@ LinearSVM<MatType>::LinearSVM(
     delta(delta),
     fitIntercept(fitIntercept)
 {
-  Train(data, labels, numClasses, optimizer, callbacks...);
+  Train(data, labels, numClasses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
-template <typename MatType>
-template <typename OptimizerType>
-LinearSVM<MatType>::LinearSVM(
-    const MatType& data,
+template<typename ModelMatType>
+template<typename OptimizerType, typename>
+mlpack_deprecated /** Will be removed in mlpack 5.0.0. **/
+LinearSVM<ModelMatType>::LinearSVM(
+    const arma::mat& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
@@ -54,84 +83,137 @@ LinearSVM<MatType>::LinearSVM(
   Train(data, labels, numClasses, optimizer);
 }
 
-template <typename MatType>
-LinearSVM<MatType>::LinearSVM(
-    const size_t inputSize,
+template<typename ModelMatType>
+template<typename MatType, typename... CallbackTypes, typename>
+LinearSVM<ModelMatType>::LinearSVM(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
     const double delta,
-    const bool fitIntercept) :
+    const bool fitIntercept,
+    CallbackTypes&&... callbacks) :
     numClasses(numClasses),
     lambda(lambda),
     delta(delta),
     fitIntercept(fitIntercept)
 {
-  LinearSVMFunction<MatType>::InitializeWeights(parameters, inputSize,
-      numClasses, fitIntercept);
+  // By default we use L-BFGS.
+  ens::L_BFGS optimizer;
+  Train(data, labels, numClasses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
-template <typename MatType>
-LinearSVM<MatType>::LinearSVM(
+template<typename ModelMatType>
+template<typename MatType,
+         typename OptimizerType,
+         typename... CallbackTypes,
+         typename>
+LinearSVM<ModelMatType>::LinearSVM(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
     const size_t numClasses,
+    OptimizerType& optimizer,
     const double lambda,
     const double delta,
-    const bool fitIntercept) :
+    const bool fitIntercept,
+    CallbackTypes&&... callbacks) :
     numClasses(numClasses),
     lambda(lambda),
     delta(delta),
     fitIntercept(fitIntercept)
 {
-  // No training to do here.
+  Train(data, labels, numClasses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
-template <typename MatType>
-template <typename OptimizerType, typename... CallbackTypes>
-double LinearSVM<MatType>::Train(
+template<typename ModelMatType>
+template<typename MatType, typename... CallbackTypes, typename>
+typename LinearSVM<ModelMatType>::ElemType LinearSVM<ModelMatType>::Train(
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    OptimizerType optimizer,
     CallbackTypes&&... callbacks)
 {
-  if (numClasses <= 1)
-  {
-    throw std::invalid_argument("LinearSVM dataset has 0 number of classes!");
-  }
-
-  LinearSVMFunction<MatType> svm(data, labels, numClasses, lambda, delta,
-      fitIntercept);
-  if (parameters.is_empty())
-    parameters = svm.InitialPoint();
-
-  // Train the model.
-  const double out = optimizer.Optimize(svm, parameters, callbacks...);
+  return Train(data, labels, numClasses, this->lambda, this->delta,
+      this->fitIntercept, std::forward<CallbackTypes>(callbacks)...);
+}
 
-  Log::Info << "LinearSVM::LinearSVM(): final objective of "
-            << "trained model is " << out << "." << std::endl;
+template<typename ModelMatType>
+template<typename MatType, typename... CallbackTypes, typename>
+typename LinearSVM<ModelMatType>::ElemType LinearSVM<ModelMatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const double lambda,
+    const std::optional<double> delta,
+    const std::optional<bool> fitIntercept,
+    CallbackTypes&&... callbacks)
+{
+  // By default, train with L-BFGS.
+  ens::L_BFGS lbfgs;
+  return Train(data, labels, numClasses, lbfgs, lambda,
+      (delta.has_value()) ? delta.value() : this->delta,
+      (fitIntercept.has_value()) ? fitIntercept.value() : this->fitIntercept,
+      std::forward<CallbackTypes>(callbacks)...);
+}
 
-  return out;
+template<typename ModelMatType>
+template<typename MatType,
+         typename OptimizerType,
+         typename... CallbackTypes,
+         typename, typename>
+typename LinearSVM<ModelMatType>::ElemType LinearSVM<ModelMatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    OptimizerType optimizer,
+    CallbackTypes&&... callbacks)
+{
+  return Train(data, labels, numClasses, optimizer, this->lambda, this->delta,
+      this->fitIntercept, std::forward<CallbackTypes>(callbacks)...);
 }
 
-template <typename MatType>
-template <typename OptimizerType>
-double LinearSVM<MatType>::Train(
+template<typename ModelMatType>
+template<typename MatType,
+         typename OptimizerType,
+         typename... CallbackTypes,
+         typename, typename>
+typename LinearSVM<ModelMatType>::ElemType LinearSVM<ModelMatType>::Train(
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    OptimizerType optimizer)
+    OptimizerType optimizer,
+    const double lambda,
+    const std::optional<double> delta,
+    const std::optional<bool> fitIntercept,
+    CallbackTypes&&... callbacks)
 {
+  this->numClasses = numClasses;
+  this->lambda = lambda;
+  
+  if (delta.has_value())
+    this->delta = delta.value();
+
+  if (fitIntercept.has_value())
+    this->fitIntercept = fitIntercept.value();
+
   if (numClasses <= 1)
   {
     throw std::invalid_argument("LinearSVM dataset has 0 number of classes!");
   }
 
-  LinearSVMFunction<MatType> svm(data, labels, numClasses, lambda, delta,
-      fitIntercept);
-  if (parameters.is_empty())
+  LinearSVMFunction<MatType, ModelMatType> svm(data, labels, numClasses,
+      this->lambda, this->delta, this->fitIntercept);
+  const bool needNewParameters = (parameters.is_empty() ||
+      (this->fitIntercept && (parameters.n_rows != data.n_rows + 1)) ||
+      (!this->fitIntercept && (parameters.n_rows != data.n_rows)) ||
+      (parameters.n_cols != numClasses));
+  if (needNewParameters)
     parameters = svm.InitialPoint();
 
   // Train the model.
-  const double out = optimizer.Optimize(svm, parameters);
+  const ElemType out = optimizer.Optimize(svm, parameters, callbacks...);
 
   Log::Info << "LinearSVM::LinearSVM(): final objective of "
             << "trained model is " << out << "." << std::endl;
@@ -139,33 +221,46 @@ double LinearSVM<MatType>::Train(
   return out;
 }
 
-template <typename MatType>
-void LinearSVM<MatType>::Classify(
+template<typename ModelMatType>
+template<typename MatType>
+void LinearSVM<ModelMatType>::Classify(
     const MatType& data,
     arma::Row<size_t>& labels) const
 {
-  arma::mat scores;
+  DenseMatType scores;
   Classify(data, labels, scores);
 }
 
-template <typename MatType>
-void LinearSVM<MatType>::Classify(
+template<typename ModelMatType>
+template<typename MatType>
+void LinearSVM<ModelMatType>::Classify(
     const MatType& data,
     arma::Row<size_t>& labels,
-    arma::mat& scores) const
+    typename LinearSVM<ModelMatType>::DenseMatType& scores) const
 {
-  Classify(data, scores);
+  util::CheckSameDimensionality(data, FeatureSize(), "LinearSVM::Classify()");
+
+  if (fitIntercept)
+  {
+    scores = parameters.rows(0, parameters.n_rows - 2).t() * data
+        + repmat(parameters.row(parameters.n_rows - 1).t(), 1, data.n_cols);
+  }
+  else
+  {
+    scores = parameters.t() * data;
+  }
 
   // Prepare necessary data.
   labels.zeros(data.n_cols);
 
-  labels = arma::conv_to<arma::Row<size_t>>::from(
+  labels = ConvTo<arma::Row<size_t>>::From(
       arma::index_max(scores));
 }
 
-template <typename MatType>
-void LinearSVM<MatType>::Classify(
-    const MatType& data,
+template<typename ModelMatType>
+mlpack_deprecated
+void LinearSVM<ModelMatType>::Classify(
+    const arma::mat& data,
     arma::mat& scores) const
 {
   util::CheckSameDimensionality(data, FeatureSize(), "LinearSVM::Classify()");
@@ -173,8 +268,7 @@ void LinearSVM<MatType>::Classify(
   if (fitIntercept)
   {
     scores = parameters.rows(0, parameters.n_rows - 2).t() * data
-        + arma::repmat(parameters.row(parameters.n_rows - 1).t(), 1,
-        data.n_cols);
+        + repmat(parameters.row(parameters.n_rows - 1).t(), 1, data.n_cols);
   }
   else
   {
@@ -182,17 +276,30 @@ void LinearSVM<MatType>::Classify(
   }
 }
 
-template <typename MatType>
-template <typename VecType>
-size_t LinearSVM<MatType>::Classify(const VecType& point) const
+template<typename ModelMatType>
+template<typename VecType>
+size_t LinearSVM<ModelMatType>::Classify(const VecType& point) const
 {
   arma::Row<size_t> label(1);
   Classify(point, label);
   return size_t(label(0));
 }
 
-template <typename MatType>
-double LinearSVM<MatType>::ComputeAccuracy(
+template<typename ModelMatType>
+template<typename VecType>
+void LinearSVM<ModelMatType>::Classify(
+    const VecType& point,
+    size_t& label,
+    typename LinearSVM<ModelMatType>::DenseColType& probabilities) const
+{
+  arma::Row<size_t> labelRow(1);
+  Classify(point, labelRow, probabilities);
+  label = labelRow[0];
+}
+
+template<typename ModelMatType>
+template<typename MatType>
+double LinearSVM<ModelMatType>::ComputeAccuracy(
     const MatType& testData,
     const arma::Row<size_t>& testLabels) const
 {
@@ -208,7 +315,28 @@ double LinearSVM<MatType>::ComputeAccura
       count++;
 
   // Return the accuracy.
-  return (double) count / labels.n_elem;
+  return (double) 100.0 * count / labels.n_elem;
+}
+
+template<typename ModelMatType>
+template<typename Archive>
+void LinearSVM<ModelMatType>::serialize(Archive& ar, const uint32_t version)
+{
+  // Old versions used `arma::mat` for the type of `parameters`.
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    arma::mat parametersTmp;
+    ar(cereal::make_nvp("parameters", parametersTmp));
+    parameters = ConvTo<arma::mat>::From(parametersTmp);
+  }
+  else
+  {
+    ar(CEREAL_NVP(parameters));
+  }
+
+  ar(CEREAL_NVP(numClasses));
+  ar(CEREAL_NVP(lambda));
+  ar(CEREAL_NVP(fitIntercept));
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_main.cpp 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_main.cpp
--- 4.3.0-2/src/mlpack/methods/linear_svm/linear_svm_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/linear_svm/linear_svm_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -294,7 +294,7 @@ void BINDING_FUNCTION(util::Params& para
     }
 
     // The initial predictors for y, Nx1.
-    rawLabels = arma::conv_to<arma::Row<size_t>>::from(
+    rawLabels = ConvTo<arma::Row<size_t>>::From(
         trainingSet.row(trainingSet.n_rows - 1));
     trainingSet.shed_row(trainingSet.n_rows - 1);
   }
@@ -405,7 +405,7 @@ void BINDING_FUNCTION(util::Params& para
       Log::Info << "Calculating class probabilities of points in " << testOutput
           << "." << endl;
       arma::mat probabilities;
-      model->svm.Classify(testSet, probabilities);
+      model->svm.Classify(testSet, predictions, probabilities);
       params.Get<arma::mat>("probabilities") = std::move(probabilities);
     }
 
diff -pruN 4.3.0-2/src/mlpack/methods/lmnn/constraints_impl.hpp 4.4.0-1/src/mlpack/methods/lmnn/constraints_impl.hpp
--- 4.3.0-2/src/mlpack/methods/lmnn/constraints_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lmnn/constraints_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -26,7 +26,7 @@ Constraints<MetricType>::Constraints(
     precalculated(false)
 {
   // Ensure a valid k is passed.
-  size_t minCount = arma::min(arma::histc(labels, arma::unique(labels)));
+  size_t minCount = min(arma::histc(labels, arma::unique(labels)));
 
   if (minCount < k + 1)
   {
@@ -66,7 +66,7 @@ inline void Constraints<MetricType>::Reo
         // We must sort these elements by norm.
         arma::Col<size_t> newNeighbors =
             neighbors.col(i).subvec(start, end - 1);
-        arma::uvec indices = arma::conv_to<arma::uvec>::from(newNeighbors);
+        arma::uvec indices = ConvTo<arma::uvec>::From(newNeighbors);
 
         arma::uvec order = arma::sort_index(norms.elem(indices));
         neighbors.col(i).subvec(start, end - 1) =
diff -pruN 4.3.0-2/src/mlpack/methods/lmnn/lmnn_function_impl.hpp 4.4.0-1/src/mlpack/methods/lmnn/lmnn_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/lmnn/lmnn_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lmnn/lmnn_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -19,23 +19,24 @@
 namespace mlpack {
 
 template<typename MetricType>
-LMNNFunction<MetricType>::LMNNFunction(const arma::mat& dataset,
-                                       const arma::Row<size_t>& labels,
+LMNNFunction<MetricType>::LMNNFunction(const arma::mat& datasetIn,
+                                       const arma::Row<size_t>& labelsIn,
                                        size_t k,
                                        double regularization,
                                        size_t range,
                                        MetricType metric) :
-    dataset(MakeAlias(const_cast<arma::mat&>(dataset), false)),
-    labels(MakeAlias(const_cast<arma::Row<size_t>&>(labels), false)),
     k(k),
     metric(metric),
     regularization(regularization),
     iteration(0),
     range(range),
-    constraint(dataset, labels, k),
-    points(dataset.n_cols),
+    constraint(datasetIn, labelsIn, k),
+    points(datasetIn.n_cols),
     impBounds(false)
 {
+  MakeAlias(dataset, datasetIn, datasetIn.n_rows, datasetIn.n_cols, 0, false);
+  MakeAlias(labels, labelsIn, labelsIn.n_elem, 0, false);
+
   // Initialize the initial learning point.
   initialPoint.eye(dataset.n_rows, dataset.n_rows);
   // Initialize transformed dataset to base dataset.
@@ -64,7 +65,7 @@ LMNNFunction<MetricType>::LMNNFunction(c
   oldTransformationCounts.push_back(dataset.n_cols);
 
   // Check if we can impose bounds over impostors.
-  size_t minCount = arma::min(arma::histc(labels, arma::unique(labels)));
+  size_t minCount = min(arma::histc(labels, arma::unique(labels)));
   if (minCount <= k + 1)
   {
     // Initialize target neighbors & impostors.
@@ -517,7 +518,7 @@ void LMNNFunction<MetricType>::Gradient(
   arma::mat cij = pCij;
 
   // Calculate gradient due to impostors.
-  arma::mat cil = arma::zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cil = zeros(dataset.n_rows, dataset.n_rows);
 
   for (size_t i = 0; i < dataset.n_cols; ++i)
   {
@@ -580,10 +581,10 @@ void LMNNFunction<MetricType>::Gradient(
 
         // Caculate gradient due to impostors.
         arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-        cil += diff * arma::trans(diff);
+        cil += diff * trans(diff);
 
         diff = dataset.col(i) - dataset.col(impostors(l, i));
-        cil -= diff * arma::trans(diff);
+        cil -= diff * trans(diff);
       }
     }
   }
@@ -645,8 +646,8 @@ void LMNNFunction<MetricType>::Gradient(
 
   gradient.zeros(transformation.n_rows, transformation.n_cols);
 
-  arma::mat cij = arma::zeros(dataset.n_rows, dataset.n_rows);
-  arma::mat cil = arma::zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cij = zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cil = zeros(dataset.n_rows, dataset.n_rows);
 
   for (size_t i = begin; i < begin + batchSize; ++i)
   {
@@ -654,7 +655,7 @@ void LMNNFunction<MetricType>::Gradient(
     {
       // Calculate gradient due to target neighbors.
       arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-      cij += diff * arma::trans(diff);
+      cij += diff * trans(diff);
     }
 
     for (int j = k - 1; j >= 0; j--)
@@ -718,10 +719,10 @@ void LMNNFunction<MetricType>::Gradient(
 
         // Caculate gradient due to impostors.
         arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-        cil += diff * arma::trans(diff);
+        cil += diff * trans(diff);
 
         diff = dataset.col(i) - dataset.col(impostors(l, i));
-        cil -= diff * arma::trans(diff);
+        cil -= diff * trans(diff);
       }
     }
   }
@@ -792,7 +793,7 @@ double LMNNFunction<MetricType>::Evaluat
   arma::mat cij = pCij;
 
   // Calculate gradient due to impostors.
-  arma::mat cil = arma::zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cil = zeros(dataset.n_rows, dataset.n_rows);
 
   for (size_t i = 0; i < dataset.n_cols; ++i)
   {
@@ -857,10 +858,10 @@ double LMNNFunction<MetricType>::Evaluat
 
         // Caculate gradient due to impostors.
         arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-        cil += diff * arma::trans(diff);
+        cil += diff * trans(diff);
 
         diff = dataset.col(i) - dataset.col(impostors(l, i));
-        cil -= diff * arma::trans(diff);
+        cil -= diff * trans(diff);
       }
     }
   }
@@ -927,8 +928,8 @@ double LMNNFunction<MetricType>::Evaluat
 
   gradient.zeros(transformation.n_rows, transformation.n_cols);
 
-  arma::mat cij = arma::zeros(dataset.n_rows, dataset.n_rows);
-  arma::mat cil = arma::zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cij = zeros(dataset.n_rows, dataset.n_rows);
+  arma::mat cil = zeros(dataset.n_rows, dataset.n_rows);
 
   for (size_t i = begin; i < begin + batchSize; ++i)
   {
@@ -941,7 +942,7 @@ double LMNNFunction<MetricType>::Evaluat
 
       // Calculate gradient due to target neighbors.
       arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-      cij += diff * arma::trans(diff);
+      cij += diff * trans(diff);
     }
 
     for (int j = k - 1; j >= 0; j--)
@@ -997,10 +998,10 @@ double LMNNFunction<MetricType>::Evaluat
 
         // Caculate gradient due to impostors.
         arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-        cil += diff * arma::trans(diff);
+        cil += diff * trans(diff);
 
         diff = dataset.col(i) - dataset.col(impostors(l, i));
-        cil -= diff * arma::trans(diff);
+        cil -= diff * trans(diff);
       }
     }
   }
@@ -1025,7 +1026,7 @@ inline void LMNNFunction<MetricType>::Pr
     {
       // Calculate gradient due to target neighbors.
       arma::vec diff = dataset.col(i) - dataset.col(targetNeighbors(j, i));
-      pCij += diff * arma::trans(diff);
+      pCij += diff * trans(diff);
     }
   }
 }
diff -pruN 4.3.0-2/src/mlpack/methods/lmnn/lmnn_main.cpp 4.4.0-1/src/mlpack/methods/lmnn/lmnn_main.cpp
--- 4.3.0-2/src/mlpack/methods/lmnn/lmnn_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lmnn/lmnn_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -213,8 +213,7 @@ double KNNAccuracy(const arma::mat& data
           1 / std::pow(distances(j, i) + 1, 2);
     }
 
-    arma::vec index = arma::conv_to<arma::vec>::from(arma::find(Map
-        == arma::max(Map)));
+    arma::vec index = ConvTo<arma::vec>::From(arma::find(Map == max(Map)));
 
     // Increase count if labels match.
     if (index(0) == labels(i))
@@ -344,7 +343,7 @@ void BINDING_FUNCTION(util::Params& para
   else if (normalize)
   {
     // Find the minimum and maximum values for each dimension.
-    arma::vec ranges = arma::max(data, 1) - arma::min(data, 1);
+    arma::vec ranges = max(data, 1) - min(data, 1);
     for (size_t d = 0; d < ranges.n_elem; ++d)
       if (ranges[d] == 0.0)
         ranges[d] = 1; // A range of 0 produces NaN later on.
diff -pruN 4.3.0-2/src/mlpack/methods/local_coordinate_coding/lcc.hpp 4.4.0-1/src/mlpack/methods/local_coordinate_coding/lcc.hpp
--- 4.3.0-2/src/mlpack/methods/local_coordinate_coding/lcc.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/local_coordinate_coding/lcc.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -75,9 +75,13 @@ namespace mlpack {
  * }
  * @endcode
  */
+template<typename MatType = arma::mat>
 class LocalCoordinateCoding
 {
  public:
+  typedef typename GetColType<MatType>::type ColType;
+  typedef typename GetRowType<MatType>::type RowType;
+
   /**
    * Set the parameters to LocalCoordinateCoding, and train the dictionary.
    * This constructor will also initialize the dictionary using the given
@@ -85,7 +89,7 @@ class LocalCoordinateCoding
    *
    * If you want to initialize the dictionary to a custom matrix, consider
    * either writing your own DictionaryInitializer class (with void
-   * Initialize(const arma::mat& data, arma::mat& dictionary) function), or call
+   * Initialize(const MatType& data, MatType& dictionary) function), or call
    * the constructor that does not take a data matrix, then call Dictionary() to
    * set the dictionary matrix to a matrix of your choosing, and then call
    * Train() with NothingInitializer (i.e.  Train<NothingInitializer>(data)).
@@ -99,7 +103,7 @@ class LocalCoordinateCoding
    * @param initializer Intializer to use.
    */
   template<typename DictionaryInitializer = DataDependentRandomInitializer>
-  LocalCoordinateCoding(const arma::mat& data,
+  LocalCoordinateCoding(const MatType& data,
                         const size_t atoms,
                         const double lambda,
                         const size_t maxIterations = 0,
@@ -132,7 +136,7 @@ class LocalCoordinateCoding
    * @return The final objective value.
    */
   template<typename DictionaryInitializer = DataDependentRandomInitializer>
-  double Train(const arma::mat& data,
+  double Train(const MatType& data,
                const DictionaryInitializer& initializer =
                    DictionaryInitializer());
 
@@ -142,7 +146,7 @@ class LocalCoordinateCoding
    * @param data Matrix containing points to encode.
    * @param codes Output matrix to store codes in.
    */
-  void Encode(const arma::mat& data, arma::mat& codes);
+  void Encode(const MatType& data, MatType& codes);
 
   /**
    * Learn dictionary by solving linear system.
@@ -153,8 +157,8 @@ class LocalCoordinateCoding
    *    the coding matrix Z that are non-zero (the adjacency matrix for the
    *    bipartite graph of points and atoms)
    */
-  void OptimizeDictionary(const arma::mat& data,
-                          const arma::mat& codes,
+  void OptimizeDictionary(const MatType& data,
+                          const MatType& codes,
                           const arma::uvec& adjacencies);
 
   /**
@@ -162,12 +166,21 @@ class LocalCoordinateCoding
    *
    * @param data Matrix containing points to encode.
    * @param codes Output matrix to store codes in.
+   */
+  double Objective(const MatType& data,
+                   const MatType& codes) const;
+
+  /**
+   * Compute objective function given the list of adjacencies.
+   *
+   * @param data Matrix containing points to encode.
+   * @param codes Output matrix to store codes in.
    * @param adjacencies Indices of entries (unrolled column by column) of
    *    the coding matrix Z that are non-zero (the adjacency matrix for the
    *    bipartite graph of points and atoms)
    */
-  double Objective(const arma::mat& data,
-                   const arma::mat& codes,
+  double Objective(const MatType& data,
+                   const MatType& codes,
                    const arma::uvec& adjacencies) const;
 
   //! Get the number of atoms.
@@ -176,9 +189,9 @@ class LocalCoordinateCoding
   size_t& Atoms() { return atoms; }
 
   //! Accessor for dictionary.
-  const arma::mat& Dictionary() const { return dictionary; }
+  const MatType& Dictionary() const { return dictionary; }
   //! Mutator for dictionary.
-  arma::mat& Dictionary() { return dictionary; }
+  MatType& Dictionary() { return dictionary; }
 
   //! Get the L1 regularization parameter.
   double Lambda() const { return lambda; }
@@ -204,7 +217,7 @@ class LocalCoordinateCoding
   size_t atoms;
 
   //! Dictionary (columns are atoms).
-  arma::mat dictionary;
+  MatType dictionary;
 
   //! l1 regularization term.
   double lambda;
@@ -217,6 +230,9 @@ class LocalCoordinateCoding
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename MatType),
+    (mlpack::LocalCoordinateCoding<MatType>), (1));
+
 // Include implementation.
 #include "lcc_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/local_coordinate_coding/lcc_impl.hpp 4.4.0-1/src/mlpack/methods/local_coordinate_coding/lcc_impl.hpp
--- 4.3.0-2/src/mlpack/methods/local_coordinate_coding/lcc_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/local_coordinate_coding/lcc_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,9 +17,10 @@
 
 namespace mlpack {
 
+template<typename MatType>
 template<typename DictionaryInitializer>
-LocalCoordinateCoding::LocalCoordinateCoding(
-    const arma::mat& data,
+LocalCoordinateCoding<MatType>::LocalCoordinateCoding(
+    const MatType& data,
     const size_t atoms,
     const double lambda,
     const size_t maxIterations,
@@ -34,7 +35,8 @@ LocalCoordinateCoding::LocalCoordinateCo
   Train(data, initializer);
 }
 
-inline LocalCoordinateCoding::LocalCoordinateCoding(
+template<typename MatType>
+inline LocalCoordinateCoding<MatType>::LocalCoordinateCoding(
     const size_t atoms,
     const double lambda,
     const size_t maxIterations,
@@ -47,9 +49,10 @@ inline LocalCoordinateCoding::LocalCoord
   // Nothing to do.
 }
 
+template<typename MatType>
 template<typename DictionaryInitializer>
-double LocalCoordinateCoding::Train(
-    const arma::mat& data,
+double LocalCoordinateCoding<MatType>::Train(
+    const MatType& data,
     const DictionaryInitializer& initializer)
 {
   // Initialize the dictionary.
@@ -61,7 +64,7 @@ double LocalCoordinateCoding::Train(
   // loop.
   Log::Info << "Initial Coding Step." << std::endl;
 
-  arma::mat codes;
+  MatType codes;
   Encode(data, codes);
   arma::uvec adjacencies = find(codes);
 
@@ -115,15 +118,16 @@ double LocalCoordinateCoding::Train(
   return lastObjVal;
 }
 
-inline void LocalCoordinateCoding::Encode(const arma::mat& data,
-                                          arma::mat& codes)
+template<typename MatType>
+inline void LocalCoordinateCoding<MatType>::Encode(const MatType& data,
+                                                   MatType& codes)
 {
-  arma::mat invSqDists = 1.0 / (repmat(trans(sum(square(dictionary))), 1,
+  MatType invSqDists = 1.0 / (repmat(trans(sum(square(dictionary))), 1,
       data.n_cols) + repmat(sum(square(data)), atoms, 1) - 2 * trans(dictionary)
       * data);
 
-  arma::mat dictGram = trans(dictionary) * dictionary;
-  arma::mat dictGramTD(dictGram.n_rows, dictGram.n_cols);
+  MatType dictGram = trans(dictionary) * dictionary;
+  MatType dictGramTD(dictGram.n_rows, dictGram.n_cols);
 
   codes.set_size(atoms, data.n_cols);
   for (size_t i = 0; i < data.n_cols; ++i)
@@ -134,32 +138,35 @@ inline void LocalCoordinateCoding::Encod
       Log::Debug << "Optimization at point " << i << "." << std::endl;
     }
 
-    arma::vec invW = invSqDists.unsafe_col(i);
-    arma::mat dictPrime = dictionary * diagmat(invW);
+    ColType invW = invSqDists.unsafe_col(i);
+    MatType dictPrime = dictionary * diagmat(invW);
 
-    arma::mat dictGramTD = diagmat(invW) * dictGram * diagmat(invW);
+    MatType dictGramTD = diagmat(invW) * dictGram * diagmat(invW);
 
     bool useCholesky = false;
     // Normalization and fitting and intercept are disabled.
-    LARS lars(useCholesky, dictGramTD, 0.5 * lambda, 0,
-        1e-16 /* default tolerance */, false, false);
+    const double tol = std::is_same<typename MatType::elem_type, float>::value ?
+        1e-8 : 1e-16;
+    LARS<MatType> lars(useCholesky, 0.5 * lambda, 0, tol, false, false);
 
     // Run LARS for this point, by making an alias of the point and passing
     // that.
-    arma::vec beta = codes.unsafe_col(i);
-    arma::rowvec responses = data.unsafe_col(i).t();
-    lars.Train(dictPrime, responses, beta, false);
+    ColType beta = codes.unsafe_col(i);
+    RowType responses = data.unsafe_col(i).t();
+    lars.Train(dictPrime, responses, false, useCholesky, dictGramTD);
+    beta = lars.Beta();
     beta %= invW; // Remember, beta is an alias of codes.col(i).
   }
 }
 
-inline void LocalCoordinateCoding::OptimizeDictionary(
-    const arma::mat& data,
-    const arma::mat& codes,
+template<typename MatType>
+inline void LocalCoordinateCoding<MatType>::OptimizeDictionary(
+    const MatType& data,
+    const MatType& codes,
     const arma::uvec& adjacencies)
 {
   // Count number of atomic neighbors for each point x^i.
-  arma::uvec neighborCounts = arma::zeros<arma::uvec>(data.n_cols, 1);
+  arma::uvec neighborCounts = zeros<arma::uvec>(data.n_cols, 1);
   if (adjacencies.n_elem > 0)
   {
     // This gets the column index.  Intentional integer division.
@@ -183,7 +190,7 @@ inline void LocalCoordinateCoding::Optim
 
   // Build dataPrime := [X x^1 ... x^1 ... x^n ... x^n]
   // where each x^i is repeated for the number of neighbors x^i has.
-  arma::mat dataPrime = arma::zeros(data.n_rows,
+  MatType dataPrime = zeros<MatType>(data.n_rows,
       data.n_cols + adjacencies.n_elem);
 
   dataPrime(arma::span::all, arma::span(0, data.n_cols - 1)) = data;
@@ -200,18 +207,18 @@ inline void LocalCoordinateCoding::Optim
   }
 
   // Handle the case of inactive atoms (atoms not used in the given coding).
-  std::vector<size_t> inactiveAtoms;
+  std::vector<arma::uword> activeAtoms;
   for (size_t j = 0; j < atoms; ++j)
-    if (accu(codes.row(j) != 0) == 0)
-      inactiveAtoms.push_back(j);
+    if (accu(codes.row(j) != 0) != 0)
+      activeAtoms.push_back((arma::uword) j);
 
-  const size_t nInactiveAtoms = inactiveAtoms.size();
-  const size_t nActiveAtoms = atoms - nInactiveAtoms;
+  const size_t nActiveAtoms = activeAtoms.size();
+  const size_t nInactiveAtoms = atoms - nActiveAtoms;
 
   // Efficient construction of codes restricted to active atoms.
-  arma::mat codesPrime = arma::zeros(nActiveAtoms, data.n_cols +
+  MatType codesPrime = zeros<MatType>(nActiveAtoms, data.n_cols +
       adjacencies.n_elem);
-  arma::vec wSquared = arma::ones(data.n_cols + adjacencies.n_elem, 1);
+  ColType wSquared = ones<ColType>(data.n_cols + adjacencies.n_elem, 1);
 
   if (nInactiveAtoms > 0)
   {
@@ -219,18 +226,13 @@ inline void LocalCoordinateCoding::Optim
         << " inactive atoms.  They will be re-initialized randomly.\n";
 
     // Create matrix holding only active codes.
-    arma::mat activeCodes;
-    RemoveRows(codes, inactiveAtoms, activeCodes);
+    MatType activeCodes = codes.rows(arma::uvec(activeAtoms));
 
     // Create reverse atom lookup for active atoms.
     arma::uvec atomReverseLookup(atoms);
-    size_t inactiveOffset = 0;
-    for (size_t i = 0; i < atoms; ++i)
+    for (size_t i = 0; i < activeAtoms.size(); ++i)
     {
-      if (inactiveAtoms[inactiveOffset] == i)
-        ++inactiveOffset;
-      else
-        atomReverseLookup(i - inactiveOffset) = i;
+      atomReverseLookup[activeAtoms[i]] = i;
     }
 
     codesPrime(arma::span::all, arma::span(0, data.n_cols - 1)) = activeCodes;
@@ -264,15 +266,18 @@ inline void LocalCoordinateCoding::Optim
     }
   }
 
-  wSquared.subvec(data.n_cols, wSquared.n_elem - 1) = lambda *
-      abs(wSquared.subvec(data.n_cols, wSquared.n_elem - 1));
+  if (adjacencies.n_elem > 0)
+  {
+    wSquared.subvec(data.n_cols, wSquared.n_elem - 1) = lambda *
+        abs(wSquared.subvec(data.n_cols, wSquared.n_elem - 1));
+  }
 
   // Solve system.
   if (nInactiveAtoms == 0)
   {
     // No inactive atoms.  We can solve directly.
-    arma::mat A = codesPrime * diagmat(wSquared) * trans(codesPrime);
-    arma::mat B = codesPrime * diagmat(wSquared) * trans(dataPrime);
+    MatType A = codesPrime * diagmat(wSquared) * trans(codesPrime);
+    MatType B = codesPrime * diagmat(wSquared) * trans(dataPrime);
 
     dictionary = trans(solve(A, B));
     /*
@@ -284,15 +289,16 @@ inline void LocalCoordinateCoding::Optim
   {
     // Inactive atoms must be reinitialized randomly, so we cannot solve
     // directly for the entire dictionary estimate.
-    arma::mat dictionaryActive =
-      trans(solve(codesPrime * diagmat(wSquared) * trans(codesPrime),
-                  codesPrime * diagmat(wSquared) * trans(dataPrime)));
+    MatType dictionaryActive =
+        trans(solve(codesPrime * diagmat(wSquared) * trans(codesPrime),
+                    codesPrime * diagmat(wSquared) * trans(dataPrime)));
 
     // Update all atoms.
-    size_t currentInactiveIndex = 0;
+    size_t currentActiveIndex = 0;
     for (size_t i = 0; i < atoms; ++i)
     {
-      if (inactiveAtoms[currentInactiveIndex] == i)
+      if (currentActiveIndex >= activeAtoms.size() ||
+          activeAtoms[currentActiveIndex] != i)
       {
         // This atom is inactive.  Reinitialize it randomly.
         dictionary.col(i) = (data.col(RandInt(data.n_cols)) +
@@ -301,22 +307,32 @@ inline void LocalCoordinateCoding::Optim
 
         // Now normalize the atom.
         dictionary.col(i) /= norm(dictionary.col(i), 2);
-
-        // Increment inactive atom counter.
-        ++currentInactiveIndex;
       }
       else
       {
         // Update estimate.
-        dictionary.col(i) = dictionaryActive.col(i - currentInactiveIndex);
+        dictionary.col(i) = dictionaryActive.col(currentActiveIndex);
+
+        // Increment active atom counter.
+        ++currentActiveIndex;
       }
     }
   }
 }
 
-inline double LocalCoordinateCoding::Objective(
-    const arma::mat& data,
-    const arma::mat& codes,
+template<typename MatType>
+inline double LocalCoordinateCoding<MatType>::Objective(
+    const MatType& data,
+    const MatType& codes) const
+{
+  // Compute adjacencies and pass off to other overload.
+  return Objective(data, codes, find(codes));
+}
+
+template<typename MatType>
+inline double LocalCoordinateCoding<MatType>::Objective(
+    const MatType& data,
+    const MatType& codes,
     const arma::uvec& adjacencies) const
 {
   double weightedL1NormZ = 0;
@@ -328,18 +344,33 @@ inline double LocalCoordinateCoding::Obj
     const size_t pointInd = (size_t) (adjacencies(l) / atoms);
 
     weightedL1NormZ += fabs(codes(atomInd, pointInd)) * arma::as_scalar(
-        arma::sum(arma::square(dictionary.col(atomInd) - data.col(pointInd))));
+        sum(square(dictionary.col(atomInd) - data.col(pointInd))));
   }
 
   double froNormResidual = norm(data - dictionary * codes, "fro");
   return std::pow(froNormResidual, 2.0) + lambda * weightedL1NormZ;
 }
 
+template<typename MatType>
 template<typename Archive>
-void LocalCoordinateCoding::serialize(Archive& ar,
-                                      const uint32_t /* version */)
+void LocalCoordinateCoding<MatType>::serialize(Archive& ar,
+                                               const uint32_t version)
 {
   ar(CEREAL_NVP(atoms));
+
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // Older versions of LocalCoordinateCoding always stored dictionary as an
+    // arma::mat.
+    arma::mat dictionaryTmp;
+    ar(cereal::make_nvp("dictionary", dictionaryTmp));
+    dictionary = ConvTo<MatType>::From(dictionaryTmp);
+  }
+  else
+  {
+    ar(CEREAL_NVP(dictionary));
+  }
+
   ar(CEREAL_NVP(dictionary));
   ar(CEREAL_NVP(lambda));
   ar(CEREAL_NVP(maxIterations));
diff -pruN 4.3.0-2/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding_main.cpp 4.4.0-1/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding_main.cpp
--- 4.3.0-2/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/local_coordinate_coding/local_coordinate_coding_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -90,8 +90,7 @@ BINDING_SEE_ALSO("Nonlinear learning usi
         "https://papers.nips.cc/paper/3875-nonlinear-learning-using-local-"
         "coordinate-coding.pdf");
 BINDING_SEE_ALSO("LocalCoordinateCoding C++ class documentation",
-        "@src/mlpack/methods/local_coordinate_coding/local_coordinate_coding."
-        "hpp");
+        "@doc/user/methods/local_coordinate_coding.md");
 
 // Training parameters.
 PARAM_MATRIX_IN("training", "Matrix of training data (X).", "t");
@@ -106,9 +105,9 @@ PARAM_FLAG("normalize", "If set, the inp
 PARAM_DOUBLE_IN("tolerance", "Tolerance for objective function.", "o", 0.01);
 
 // Load/save a model.
-PARAM_MODEL_IN(LocalCoordinateCoding, "input_model", "Input LCC model.", "m");
-PARAM_MODEL_OUT(LocalCoordinateCoding, "output_model", "Output for trained LCC "
-    "model.", "M");
+PARAM_MODEL_IN(LocalCoordinateCoding<>, "input_model", "Input LCC model.", "m");
+PARAM_MODEL_OUT(LocalCoordinateCoding<>, "output_model",
+    "Output for trained LCC model.", "M");
 
 // Test on another dataset.
 PARAM_MATRIX_IN("test", "Test points to encode.", "T");
@@ -143,9 +142,9 @@ void BINDING_FUNCTION(util::Params& para
   ReportIgnoredParam(params, {{ "training", false }}, "tolerance");
 
   // Do we have an existing model?
-  LocalCoordinateCoding* lcc = NULL;
+  LocalCoordinateCoding<>* lcc = NULL;
   if (params.Has("input_model"))
-    lcc = params.Get<LocalCoordinateCoding*>("input_model");
+    lcc = params.Get<LocalCoordinateCoding<>*>("input_model");
 
   if (params.Has("training"))
   {
@@ -171,7 +170,7 @@ void BINDING_FUNCTION(util::Params& para
         [](double x) { return x > 0; }, 1,
         "Tolerance should be a positive real number");
 
-    lcc = new LocalCoordinateCoding(0, 0.0);
+    lcc = new LocalCoordinateCoding<>(0, 0.0);
 
     lcc->Lambda() = params.Get<double>("lambda");
     lcc->Atoms() = (size_t) params.Get<int>("atoms");
@@ -257,5 +256,5 @@ void BINDING_FUNCTION(util::Params& para
 
   // Save the dictionary and the model.
   params.Get<mat>("dictionary") = lcc->Dictionary();
-  params.Get<LocalCoordinateCoding*>("output_model") = lcc;
+  params.Get<LocalCoordinateCoding<>*>("output_model") = lcc;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression.hpp 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression.hpp
--- 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,24 +37,43 @@ template<typename MatType = arma::mat>
 class LogisticRegression
 {
  public:
+  typedef typename MatType::elem_type ElemType;
+  typedef typename GetDenseRowType<MatType>::type RowType;
+  typedef typename GetDenseColType<MatType>::type ColType;
+
+  /**
+   * Construct the LogisticRegression class without performing any training.
+   * The dimensionality of the data (which will be used to set the size of the
+   * parameters vector) must be specified, and all of the parameters in the
+   * model will be set to 0.  Note that the dimensionality may be changed later
+   * by directly modifying the parameters vector (using Parameters()).
+   *
+   * @param dimensionality Dimensionality of the data.
+   * @param lambda L2-regularization parameter.
+   */
+  LogisticRegression(const size_t dimensionality = 0,
+                     const double lambda = 0);
+
   /**
    * Construct the LogisticRegression class with the given labeled training
    * data.  This will train the model.  Optionally, specify lambda, which is the
    * penalty parameter for L2-regularization.  If not specified, it is set to 0,
    * which results in standard (unregularized) logistic regression.
    *
-   * It is not possible to set a custom optimizer with this constructor.  Either
-   * use a constructor that does not train and call Train() with a custom
-   * optimizer type, or use the constructor that takes an instantiated
-   * optimizer.  (This unfortunate situation is a language restriction of C++.)
+   * This constructor uses the default L-BFGS optimizer to train the model.
    *
    * @param predictors Input training variables.
    * @param responses Outputs resulting from input training variables.
    * @param lambda L2-regularization parameter.
    */
+  template<typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
   LogisticRegression(const MatType& predictors,
                      const arma::Row<size_t>& responses,
-                     const double lambda = 0);
+                     const double lambda = 0.0,
+                     CallbackTypes&&... callbacks);
 
   /**
    * Construct the LogisticRegression class with the given labeled training
@@ -62,54 +81,84 @@ class LogisticRegression
    * penalty parameter for L2-regularization.  If not specified, it is set to 0,
    * which results in standard (unregularized) logistic regression.
    *
-   * It is not possible to set a custom optimizer with this constructor.  Either
-   * use a constructor that does not train and call Train() with a custom
-   * optimizer type, or use the constructor that takes an instantiated
-   * optimizer.  (This unfortunate situation is a language restriction of C++.)
+   * This constructor uses the default L-BFGS optimizer to train the model.
    *
    * @param predictors Input training variables.
    * @param responses Outputs results from input training variables.
    * @param initialPoint Initial model to train with.
    * @param lambda L2-regularization parameter.
+   * @param callbacks Instantiated ensmallen callbacks fro the default optimizer
+   *     (L-BFGS).
    */
+  template<typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
   LogisticRegression(const MatType& predictors,
                      const arma::Row<size_t>& responses,
-                     const arma::rowvec& initialPoint,
-                     const double lambda = 0);
+                     const RowType& initialPoint,
+                     const double lambda = 0.0,
+                     CallbackTypes&&... callbacks);
 
   /**
-   * Construct the LogisticRegression class without performing any training.
-   * The dimensionality of the data (which will be used to set the size of the
-   * parameters vector) must be specified, and all of the parameters in the
-   * model will be set to 0.  Note that the dimensionality may be changed later
-   * by directly modifying the parameters vector (using Parameters()).
+   * Construct the LogisticRegression class with the given labeled training
+   * data.  This will train the model.  This overload takes an already
+   * instantiated optimizer, which can be configured before training is run by
+   * this constructor.
    *
-   * @param dimensionality Dimensionality of the data.
+   * Before optimization, the model will be initialized to zeros.
+   *
+   * @param predictors Input training variables.
+   * @param responses Outputs results from input training variables.
+   * @param optimizer Instantiated optimizer with instantiated error function.
    * @param lambda L2-regularization parameter.
+   * @param callbacks Instantiated ensmallen callbacks fro the default optimizer
+   *     (L-BFGS).
    */
-  LogisticRegression(const size_t dimensionality = 0,
-                     const double lambda = 0);
+  template<typename OptimizerType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, LogisticRegressionFunction<MatType>, RowType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  LogisticRegression(const MatType& predictors,
+                     const arma::Row<size_t>& responses,
+                     OptimizerType& optimizer,
+                     const double lambda = 0.0,
+                     CallbackTypes&&... callbacks);
 
   /**
    * Construct the LogisticRegression class with the given labeled training
    * data.  This will train the model.  This overload takes an already
-   * instantiated optimizer (which holds the LogisticRegressionFunction error
-   * function, which must also be instantiated), so that the optimizer can be
-   * configured before the training is run by this constructor.  The update
-   * policy of the optimizer can be set through the policy argument.  The
-   * predictors and responses and initial point are all taken from the error
-   * function contained in the optimizer.
+   * instantiated optimizer, which can be configured before training is run by
+   * this constructor.
+   *
+   * Before optimization, the model will be initialized to zeros.
    *
    * @param predictors Input training variables.
    * @param responses Outputs results from input training variables.
    * @param optimizer Instantiated optimizer with instantiated error function.
+   * @param initialPoint Initial model to train with.
    * @param lambda L2-regularization parameter.
+   * @param callbacks Instantiated ensmallen callbacks fro the default optimizer
+   *     (L-BFGS).
    */
-  template<typename OptimizerType>
+  template<typename OptimizerType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, LogisticRegressionFunction<MatType>, RowType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
   LogisticRegression(const MatType& predictors,
                      const arma::Row<size_t>& responses,
                      OptimizerType& optimizer,
-                     const double lambda);
+                     const RowType& initialPoint,
+                     const double lambda = 0.0,
+                     CallbackTypes&&... callbacks);
 
   /**
    * Train the LogisticRegression model on the given input data.  By default,
@@ -128,10 +177,73 @@ class LogisticRegression
    *      See https://www.ensmallen.org/docs.html#callback-documentation.
    * @return The final objective of the trained model (NaN or Inf on error)
    */
-  template<typename OptimizerType = ens::L_BFGS, typename... CallbackTypes>
-  double Train(const MatType& predictors,
-               const arma::Row<size_t>& responses,
-               CallbackTypes&&... callbacks);
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& predictors,
+                 const arma::Row<size_t>& responses,
+                 CallbackTypes&&... callbacks);
+
+  /**
+   * Train the LogisticRegression model on the given input data.  By default,
+   * the L-BFGS optimization algorithm is used, but others can be specified
+   * (such as ens::SGD).
+   *
+   * This will use the existing model parameters as a starting point for the
+   * optimization.  If this is not what you want, then you should access the
+   * parameters vector directly with Parameters() and modify it as desired.
+   *
+   * @tparam OptimizerType Type of optimizer to use to train the model.
+   * @tparam CallbackTypes Types of Callback Functions.
+   * @param predictors Input training variables.
+   * @param responses Outputs results from input training variables.
+   * @param lambda L2 regularization penalty parameter.
+   * @param callbacks Callback function for ensmallen optimizer `OptimizerType`.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return The final objective of the trained model (NaN or Inf on error)
+   */
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& predictors,
+                 const arma::Row<size_t>& responses,
+                 const double lambda,
+                 CallbackTypes&&... callbacks);
+
+  /**
+   * Train the LogisticRegression model with the given instantiated optimizer.
+   * Using this overload allows configuring the instantiated optimizer before
+   * training is performed.
+   *
+   * This will use the existing model parameters as a starting point for the
+   * optimization.  If this is not what you want, then you should access the
+   * parameters vector directly with Parameters() and modify it as desired.
+   *
+   * @tparam OptimizerType Type of optimizer to use to train the model.
+   * @tparam CallbackTypes Types of Callback Functions.
+   * @param predictors Input training variables.
+   * @param responses Outputs results from input training variables.
+   * @param optimizer Instantiated optimizer with instantiated error function.
+   * @param callbacks Callback function for ensmallen optimizer `OptimizerType`.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return The final objective of the trained model (NaN or Inf on error)
+   */
+  template<typename OptimizerType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, LogisticRegressionFunction<MatType>, RowType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& predictors,
+                 const arma::Row<size_t>& responses,
+                 OptimizerType& optimizer,
+                 CallbackTypes&&... callbacks);
 
   /**
    * Train the LogisticRegression model with the given instantiated optimizer.
@@ -147,20 +259,29 @@ class LogisticRegression
    * @param predictors Input training variables.
    * @param responses Outputs results from input training variables.
    * @param optimizer Instantiated optimizer with instantiated error function.
+   * @param lambda L2 regularization penalty parameter.
    * @param callbacks Callback function for ensmallen optimizer `OptimizerType`.
    *      See https://www.ensmallen.org/docs.html#callback-documentation.
    * @return The final objective of the trained model (NaN or Inf on error)
    */
-  template<typename OptimizerType, typename... CallbackTypes>
-  double Train(const MatType& predictors,
-               const arma::Row<size_t>& responses,
-               OptimizerType& optimizer,
-               CallbackTypes&&... callbacks);
+  template<typename OptimizerType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, LogisticRegressionFunction<MatType>, RowType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& predictors,
+                 const arma::Row<size_t>& responses,
+                 OptimizerType& optimizer,
+                 const double lambda,
+                 CallbackTypes&&... callbacks);
 
   //! Return the parameters (the b vector).
-  const arma::rowvec& Parameters() const { return parameters; }
+  const RowType& Parameters() const { return parameters; }
   //! Modify the parameters (the b vector).
-  arma::rowvec& Parameters() { return parameters; }
+  RowType& Parameters() { return parameters; }
 
   //! Return the lambda value for L2-regularization.
   const double& Lambda() const { return lambda; }
@@ -183,18 +304,54 @@ class LogisticRegression
                   const double decisionBoundary = 0.5) const;
 
   /**
-   * Classify the given points, returning the predicted labels for each point.
-   * Optionally, specify the decision boundary; logistic regression returns a
-   * value between 0 and 1.  If the value is greater than the decision boundary,
-   * the response is taken to be 1; otherwise, it is 0.  By default the decision
-   * boundary is 0.5.
+   * Classify the given point, storing the predicted label in `prediction` and
+   * the class probabilities in `probabilities`.
+   *
+   * @param point Point to classify.
+   * @param prediction size_t to store computed label into.
+   * @param probabilities Vector to store class probabilities into.  Will have
+   *     length 2.
+   * @param decisionBoundary Decision boundary (default 0.5).
+   */
+  template<typename VecType>
+  void Classify(const VecType& point,
+                size_t& prediction,
+                ColType& probabilities,
+                const double decisionBoundary = 0.5) const;
+
+  /**
+   * Classify the given points, storing the predicted labels for each point in
+   * `labels`.  Optionally, specify the decision boundary; logistic regression
+   * returns a value between 0 and 1.  If the value is greater than the decision
+   * boundary, the response is taken to be 1; otherwise, it is 0.  By default
+   * the decision boundary is 0.5.
    *
    * @param dataset Set of points to classify.
-   * @param labels Predicted labels for each point.
+   * @param predictions Predicted labels for each point.
    * @param decisionBoundary Decision boundary (default 0.5).
    */
   void Classify(const MatType& dataset,
-                arma::Row<size_t>& labels,
+                arma::Row<size_t>& predictions,
+                const double decisionBoundary = 0.5) const;
+
+  /**
+   * Classify the given points, storing the predicted labels for each point in
+   * `labels` and the class probabilities for each point in `probabilities`.
+   *
+   * Optionally, specify the decision boundary; logistic regression
+   * returns a value between 0 and 1.  If the value is greater than the decision
+   * boundary, the response is taken to be 1; otherwise, it is 0.  By default
+   * the decision boundary is 0.5.
+   *
+   * @param dataset Set of points to classify.
+   * @param predictions Predicted labels for each point.
+   * @param probabilities Matrix to store class probabilities for each point
+   *     into.
+   * @param decisionBoundary Decision boundary (default 0.5).
+   */
+  void Classify(const MatType& dataset,
+                arma::Row<size_t>& predictions,
+                MatType& probabilities,
                 const double decisionBoundary = 0.5) const;
 
   /**
@@ -203,8 +360,16 @@ class LogisticRegression
    * @param dataset Set of points to classify.
    * @param probabilities Class probabilities for each point (output).
    */
+  mlpack_deprecated /* to be removed in mlpack 5.0.0 */
   void Classify(const MatType& dataset,
-                arma::mat& probabilities) const;
+                MatType& probabilities) const;
+
+  /**
+   * Reset the weights in the model to zeros.  This function can be used between
+   * calls to Train(), to force learning of a new model instead of incremental
+   * training.
+   */
+  void Reset();
 
   /**
    * Compute the accuracy of the model on the given predictors and responses,
@@ -241,13 +406,16 @@ class LogisticRegression
 
  private:
   //! Vector of trained parameters (size: dimensionality plus one).
-  arma::rowvec parameters;
+  RowType parameters;
   //! L2-regularization penalty parameter.
   double lambda;
 };
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename MatType),
+    (mlpack::LogisticRegression<MatType>), (1));
+
 // Include implementation.
 #include "logistic_regression_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_function.hpp 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_function.hpp
--- 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,7 +51,8 @@ class LogisticRegressionFunction
   const arma::Row<size_t>& Responses() const { return responses; }
 
   /**
-  * Shuffle the order of function visitation.  This may be called by the optimizer.
+  * Shuffle the order of function visitation.  This may be called by the
+  * optimizer.
   */
   void Shuffle();
 
@@ -66,7 +67,9 @@ class LogisticRegressionFunction
    *
    * @param parameters Vector of logistic regression parameters.
    */
-  double Evaluate(const arma::mat& parameters) const;
+  template<typename CoordinatesType>
+  typename CoordinatesType::elem_type Evaluate(
+      const CoordinatesType& parameters) const;
 
   /**
    * Evaluate the logistic regression log-likelihood function with the given
@@ -85,9 +88,11 @@ class LogisticRegressionFunction
    * @param batchSize Number of points to be passed at a time to use for
    *     objective function evaluation.
    */
-  double Evaluate(const arma::mat& parameters,
-                  const size_t begin,
-                  const size_t batchSize = 1) const;
+  template<typename CoordinatesType>
+  typename CoordinatesType::elem_type Evaluate(
+      const CoordinatesType& parameters,
+      const size_t begin,
+      const size_t batchSize = 1) const;
 
   /**
    * Evaluate the gradient of the logistic regression log-likelihood function
@@ -96,7 +101,8 @@ class LogisticRegressionFunction
    * @param parameters Vector of logistic regression parameters.
    * @param gradient Vector to output gradient into.
    */
-  void Gradient(const arma::mat& parameters, arma::mat& gradient) const;
+  template<typename CoordinatesType, typename GradType>
+  void Gradient(const CoordinatesType& parameters, GradType& gradient) const;
 
   /**
    * Evaluate the gradient of the logistic regression log-likelihood function
@@ -111,8 +117,8 @@ class LogisticRegressionFunction
    * @param batchSize Number of points to be processed as a batch for objective
    *     function gradient evaluation.
    */
-  template<typename GradType>
-  void Gradient(const arma::mat& parameters,
+  template<typename CoordinatesType, typename GradType>
+  void Gradient(const CoordinatesType& parameters,
                 const size_t begin,
                 GradType& gradient,
                 const size_t batchSize = 1) const;
@@ -126,30 +132,33 @@ class LogisticRegressionFunction
    * @param parameters Vector of logistic regression parameters.
    * @param j Index of the feature with respect to which the gradient is to
    *    be computed.
-   * @param gradient Sparse matrix to output gradient into.
+   * @param gradient Matrix to output gradient into.
    */
-  void PartialGradient(const arma::mat& parameters,
+  template<typename CoordinatesType, typename GradType>
+  void PartialGradient(const CoordinatesType& parameters,
                        const size_t j,
-                       arma::sp_mat& gradient) const;
+                       GradType& gradient) const;
 
   /**
    * Evaluate the objective function and gradient of the logistic regression
    * log-likelihood function simultaneously with the given parameters.
    */
-  template<typename GradType>
-  double EvaluateWithGradient(const arma::mat& parameters,
-                              GradType& gradient) const;
+  template<typename CoordinatesType, typename GradType>
+  typename CoordinatesType::elem_type EvaluateWithGradient(
+      const CoordinatesType& parameters,
+      GradType& gradient) const;
 
   /**
    * Evaluate the objective function and gradient of the logistic regression
    * log-likelihood function simultaneously with the given parameters, for
    * the given batch size from a given point in the dataset.
    */
-  template<typename GradType>
-  double EvaluateWithGradient(const arma::mat& parameters,
-                              const size_t begin,
-                              GradType& gradient,
-                              const size_t batchSize = 1) const;
+  template<typename CoordinatesType, typename GradType>
+  typename CoordinatesType::elem_type EvaluateWithGradient(
+      const CoordinatesType& parameters,
+      const size_t begin,
+      GradType& gradient,
+      const size_t batchSize = 1) const;
 
   //! Return the number of separable functions (the number of predictor points).
   size_t NumFunctions() const { return predictors.n_cols; }
diff -pruN 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_function_impl.hpp 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,15 +21,16 @@ namespace mlpack {
 
 template<typename MatType>
 LogisticRegressionFunction<MatType>::LogisticRegressionFunction(
-    const MatType& predictors,
-    const arma::Row<size_t>& responses,
+    const MatType& predictorsIn,
+    const arma::Row<size_t>& responsesIn,
     const double lambda) :
-    // We promise to be well-behaved... the elements won't be modified.
-    predictors(MakeAlias(const_cast<MatType&>(predictors), false)),
-    responses(MakeAlias(const_cast<arma::Row<size_t>&>(responses),
-        false)),
     lambda(lambda)
 {
+  // We promise to be well-behaved... the elements won't be modified.
+  MakeAlias(this->predictors, predictorsIn, predictorsIn.n_rows,
+      predictorsIn.n_cols, 0, false);
+  MakeAlias(this->responses, responsesIn, responsesIn.n_elem, 0, false);
+
   // Sanity check.
   if (responses.n_elem != predictors.n_cols)
   {
@@ -65,8 +66,10 @@ void LogisticRegressionFunction<MatType>
  * parameters.
  */
 template<typename MatType>
-double LogisticRegressionFunction<MatType>::Evaluate(
-    const arma::mat& parameters) const
+template<typename CoordinatesType>
+typename CoordinatesType::elem_type
+LogisticRegressionFunction<MatType>::Evaluate(
+    const CoordinatesType& parameters) const
 {
   // The objective function is the log-likelihood function (w is the parameters
   // vector for the model; y is the responses; x is the predictors; sig() is the
@@ -74,16 +77,24 @@ double LogisticRegressionFunction<MatTyp
   //   f(w) = sum(y log(sig(w'x)) + (1 - y) log(sig(1 - w'x))).
   // We want to minimize this function.  L2-regularization is just lambda
   // multiplied by the squared l2-norm of the parameters then divided by two.
+  typedef typename CoordinatesType::elem_type ElemType;
+
+  // Specifying these here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType half = ((ElemType) 0.5);
+  constexpr ElemType one = ((ElemType) 1);
+  constexpr ElemType two = ((ElemType) 2);
 
   // For the regularization, we ignore the first term, which is the intercept
   // term and take every term except the last one in the decision variable.
-  const double regularization = 0.5 * lambda *
-      arma::dot(parameters.tail_cols(parameters.n_elem - 1),
+  const ElemType regularization = half * lambda *
+      dot(parameters.tail_cols(parameters.n_elem - 1),
       parameters.tail_cols(parameters.n_elem - 1));
 
   // Calculate vectors of sigmoids.  The intercept term is parameters(0, 0) and
   // does not need to be multiplied by any of the predictors.
-  const arma::rowvec sigmoid = 1.0 / (1.0 + arma::exp(-(parameters(0, 0) +
+  const CoordinatesType sigmoid = one / (one + exp(-(parameters(0, 0) +
       parameters.tail_cols(parameters.n_elem - 1) * predictors)));
 
   // Assemble full objective function.  Often the objective function and the
@@ -92,9 +103,9 @@ double LogisticRegressionFunction<MatTyp
   // terms for computational efficiency.  Note that the conversion causes some
   // copy and slowdown, but this is so negligible compared to the rest of the
   // calculation it is not worth optimizing for.
-  const double result = arma::accu(arma::log(1.0 -
-      arma::conv_to<arma::rowvec>::from(responses) + sigmoid %
-      (2 * arma::conv_to<arma::rowvec>::from(responses) - 1.0)));
+  const ElemType result = accu(log(one -
+      ConvTo<CoordinatesType>::From(responses) + sigmoid %
+      (two * ConvTo<CoordinatesType>::From(responses) - one)));
 
   // Invert the result, because it's a minimization.
   return regularization - result;
@@ -105,27 +116,37 @@ double LogisticRegressionFunction<MatTyp
  * parameters for a given batch from a given point.
  */
 template<typename MatType>
-double LogisticRegressionFunction<MatType>::Evaluate(
-                  const arma::mat& parameters,
-                  const size_t begin,
-                  const size_t batchSize) const
+template<typename CoordinatesType>
+typename CoordinatesType::elem_type
+LogisticRegressionFunction<MatType>::Evaluate(
+    const CoordinatesType& parameters,
+    const size_t begin,
+    const size_t batchSize) const
 {
+  typedef typename CoordinatesType::elem_type ElemType;
+
+  // Specifying these here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+  constexpr ElemType two = ((ElemType) 2);
+
   // Calculate the regularization term.
-  const double regularization = lambda *
-      (batchSize / (2.0 * predictors.n_cols)) *
-      arma::dot(parameters.tail_cols(parameters.n_elem - 1),
-                parameters.tail_cols(parameters.n_elem - 1));
+  const ElemType regularization = lambda *
+      (batchSize / (two * predictors.n_cols)) *
+      dot(parameters.tail_cols(parameters.n_elem - 1),
+          parameters.tail_cols(parameters.n_elem - 1));
 
   // Calculate the sigmoid function values.
-  const arma::rowvec sigmoid = 1.0 / (1.0 + arma::exp(-(parameters(0, 0) +
+  const CoordinatesType sigmoid = one / (one + exp(-(parameters(0, 0) +
       parameters.tail_cols(parameters.n_elem - 1) *
       predictors.cols(begin, begin + batchSize - 1))));
 
   // Compute the objective for the given batch size from a given point.
-  arma::rowvec respD = arma::conv_to<arma::rowvec>::from(responses.subvec(begin,
-      begin + batchSize - 1));
-  const double result = arma::accu(arma::log(1.0 - respD + sigmoid %
-      (2 * respD - 1.0)));
+  CoordinatesType respD = ConvTo<CoordinatesType>::From(
+      responses.subvec(begin, begin + batchSize - 1));
+  const ElemType result = accu(log(one - respD + sigmoid %
+      (two * respD - one)));
 
   // Invert the result, because it's a minimization.
   return regularization - result;
@@ -133,19 +154,27 @@ double LogisticRegressionFunction<MatTyp
 
 //! Evaluate the gradient of the logistic regression objective function.
 template<typename MatType>
+template<typename CoordinatesType, typename GradType>
 void LogisticRegressionFunction<MatType>::Gradient(
-    const arma::mat& parameters,
-    arma::mat& gradient) const
+    const CoordinatesType& parameters,
+    GradType& gradient) const
 {
+  typedef typename CoordinatesType::elem_type ElemType;
+
   // Regularization term.
-  arma::mat regularization;
+  GradType regularization;
   regularization = lambda * parameters.tail_cols(parameters.n_elem - 1);
 
-  const arma::rowvec sigmoids = (1 / (1 + arma::exp(-parameters(0, 0)
+  // Specifying this here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+
+  const CoordinatesType sigmoids = (one / (one + exp(-parameters(0, 0)
       - parameters.tail_cols(parameters.n_elem - 1) * predictors)));
 
-  gradient.set_size(arma::size(parameters));
-  gradient[0] = -arma::accu(responses - sigmoids);
+  gradient.set_size(size(parameters));
+  gradient[0] = -accu(responses - sigmoids);
   gradient.tail_cols(parameters.n_elem - 1) = (sigmoids - responses) *
       predictors.t() + regularization;
 }
@@ -153,26 +182,33 @@ void LogisticRegressionFunction<MatType>
 //! Evaluate the gradient of the logistic regression objective function for a
 //! given batch size.
 template<typename MatType>
-template<typename GradType>
+template<typename CoordinatesType, typename GradType>
 void LogisticRegressionFunction<MatType>::Gradient(
-                const arma::mat& parameters,
+                const CoordinatesType& parameters,
                 const size_t begin,
                 GradType& gradient,
                 const size_t batchSize) const
 {
+  typedef typename CoordinatesType::elem_type ElemType;
+
   // Regularization term.
-  arma::mat regularization;
+  GradType regularization;
   regularization = lambda * parameters.tail_cols(parameters.n_elem - 1)
       / predictors.n_cols * batchSize;
 
-  const arma::rowvec exponents = parameters(0, 0) +
+  // Specifying this here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+
+  const CoordinatesType exponents = parameters(0, 0) +
       parameters.tail_cols(parameters.n_elem - 1) *
       predictors.cols(begin, begin + batchSize - 1);
   // Calculating the sigmoid function values.
-  const arma::rowvec sigmoids = 1.0 / (1.0 + arma::exp(-exponents));
+  const CoordinatesType sigmoids = one / (one + exp(-exponents));
 
   gradient.set_size(parameters.n_rows, parameters.n_cols);
-  gradient[0] = -arma::accu(responses.subvec(begin, begin + batchSize - 1) -
+  gradient[0] = -accu(responses.subvec(begin, begin + batchSize - 1) -
       sigmoids);
   gradient.tail_cols(parameters.n_elem - 1) = (sigmoids -
       responses.subvec(begin, begin + batchSize - 1)) *
@@ -183,95 +219,121 @@ void LogisticRegressionFunction<MatType>
  * Evaluate the partial gradient of the logistic regression objective
  * function with respect to the individual features in the parameter.
  */
-template <typename MatType>
+template<typename MatType>
+template<typename CoordinatesType, typename GradType>
 void LogisticRegressionFunction<MatType>::PartialGradient(
-    const arma::mat& parameters,
+    const CoordinatesType& parameters,
     const size_t j,
-    arma::sp_mat& gradient) const
+    GradType& gradient) const
 {
-  const arma::rowvec diffs = responses - (1 / (1 + arma::exp(-parameters(0, 0)
+  typedef typename CoordinatesType::elem_type ElemType;
+
+  // Specifying this here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+
+  const CoordinatesType diffs = responses - (one / (one + exp(-parameters(0, 0)
       - parameters.tail_cols(parameters.n_elem - 1) * predictors)));
 
-  gradient.set_size(arma::size(parameters));
+  gradient.set_size(size(parameters));
 
   if (j == 0)
   {
-    gradient[j] = -arma::accu(diffs);
+    gradient[j] = -accu(diffs);
   }
   else
   {
-    gradient[j] = arma::dot(-predictors.row(j - 1), diffs) + lambda *
-      parameters(0, j);
+    gradient[j] = dot(-predictors.row(j - 1), diffs) + lambda *
+        parameters(0, j);
   }
 }
 
 template<typename MatType>
-template<typename GradType>
-double LogisticRegressionFunction<MatType>::EvaluateWithGradient(
-    const arma::mat& parameters,
+template<typename CoordinatesType, typename GradType>
+typename CoordinatesType::elem_type
+LogisticRegressionFunction<MatType>::EvaluateWithGradient(
+    const CoordinatesType& parameters,
     GradType& gradient) const
 {
+  typedef typename CoordinatesType::elem_type ElemType;
+
+  // Specifying these here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+  constexpr ElemType two = ((ElemType) 2);
+
   // Regularization term.
-  arma::mat regularization = lambda *
+  GradType regularization = lambda *
       parameters.tail_cols(parameters.n_elem - 1);
 
-  const double objectiveRegularization = lambda / 2.0 *
-      arma::dot(parameters.tail_cols(parameters.n_elem - 1),
-                parameters.tail_cols(parameters.n_elem - 1));
+  const ElemType objectiveRegularization = lambda / two *
+      dot(parameters.tail_cols(parameters.n_elem - 1),
+          parameters.tail_cols(parameters.n_elem - 1));
 
   // Calculate the sigmoid function values.
-  const arma::rowvec sigmoids = 1.0 / (1.0 + arma::exp(-(parameters(0, 0) +
+  const CoordinatesType sigmoids = one / (one + exp(-(parameters(0, 0) +
       parameters.tail_cols(parameters.n_elem - 1) * predictors)));
 
-  gradient.set_size(arma::size(parameters));
-  gradient[0] = -arma::accu(responses - sigmoids);
+  gradient.set_size(size(parameters));
+  gradient[0] = -accu(responses - sigmoids);
   gradient.tail_cols(parameters.n_elem - 1) = (sigmoids - responses) *
       predictors.t() + regularization;
 
   // Now compute the objective function using the sigmoids.
-  double result = arma::accu(arma::log(1.0 -
-      arma::conv_to<arma::rowvec>::from(responses) + sigmoids %
-      (2 * arma::conv_to<arma::rowvec>::from(responses) - 1.0)));
+  ElemType result = accu(log(one -
+      ConvTo<CoordinatesType>::From(responses) + sigmoids %
+      (two * ConvTo<CoordinatesType>::From(responses) - one)));
 
   // Invert the result, because it's a minimization.
   return objectiveRegularization - result;
 }
 
 template<typename MatType>
-template<typename GradType>
-double LogisticRegressionFunction<MatType>::EvaluateWithGradient(
-    const arma::mat& parameters,
+template<typename CoordinatesType, typename GradType>
+typename CoordinatesType::elem_type
+LogisticRegressionFunction<MatType>::EvaluateWithGradient(
+    const CoordinatesType& parameters,
     const size_t begin,
     GradType& gradient,
     const size_t batchSize) const
 {
+  typedef typename CoordinatesType::elem_type ElemType;
+
+  // Specifying these here makes the code below a little bit cleaner, and avoids
+  // accidentally casting an entire expression to `double`, e.g., by the use of
+  // `1.0` or similar.
+  constexpr ElemType one = ((ElemType) 1);
+  constexpr ElemType two = ((ElemType) 2);
+
   // Regularization term.
-  arma::mat regularization =
+  GradType regularization =
       lambda * parameters.tail_cols(parameters.n_elem - 1) / predictors.n_cols *
       batchSize;
 
-  const double objectiveRegularization = lambda *
-      (batchSize / (2.0 * predictors.n_cols)) *
-      arma::dot(parameters.tail_cols(parameters.n_elem - 1),
-                parameters.tail_cols(parameters.n_elem - 1));
+  const ElemType objectiveRegularization = lambda *
+      (batchSize / (two * predictors.n_cols)) *
+      dot(parameters.tail_cols(parameters.n_elem - 1),
+          parameters.tail_cols(parameters.n_elem - 1));
 
   // Calculate the sigmoid function values.
-  const arma::rowvec sigmoids = 1.0 / (1.0 + arma::exp(-(parameters(0, 0) +
+  const CoordinatesType sigmoids = one / (one + exp(-(parameters(0, 0) +
       parameters.tail_cols(parameters.n_elem - 1) *
       predictors.cols(begin, begin + batchSize - 1))));
 
   gradient.set_size(parameters.n_rows, parameters.n_cols);
-  gradient[0] = -arma::accu(responses.subvec(begin, begin + batchSize - 1) -
+  gradient[0] = -accu(responses.subvec(begin, begin + batchSize - 1) -
       sigmoids);
   gradient.tail_cols(parameters.n_elem - 1) = (sigmoids -
       responses.subvec(begin, begin + batchSize - 1)) *
       predictors.cols(begin, begin + batchSize - 1).t() + regularization;
 
   // Now compute the objective function using the sigmoids.
-  arma::rowvec respD = arma::conv_to<arma::rowvec>::from(responses.subvec(begin,
-      begin + batchSize - 1));
-  const double result = arma::accu(arma::log(1.0 - respD + sigmoids %
-      (2 * respD - 1.0)));
+  CoordinatesType respD = ConvTo<CoordinatesType>::From(
+      responses.subvec(begin, begin + batchSize - 1));
+  const ElemType result = accu(log(one - respD + sigmoids %
+      (two * respD - one)));
 
   // Invert the result, because it's a minimization.
   return objectiveRegularization - result;
diff -pruN 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_impl.hpp 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_impl.hpp
--- 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,62 +21,103 @@ namespace mlpack {
 
 template<typename MatType>
 LogisticRegression<MatType>::LogisticRegression(
+    const size_t dimensionality,
+    const double lambda) :
+    parameters(dimensionality + 1),
+    lambda(lambda)
+{
+  // No training to do here.
+  parameters.zeros();
+}
+
+template<typename MatType>
+template<typename... CallbackTypes, typename>
+LogisticRegression<MatType>::LogisticRegression(
     const MatType& predictors,
     const arma::Row<size_t>& responses,
-    const double lambda) :
+    const double lambda,
+    CallbackTypes&&... callbacks) :
     lambda(lambda)
 {
-  Train(predictors, responses);
+  Train(predictors, responses, std::forward<CallbackTypes>(callbacks)...);
 }
 
 template<typename MatType>
+template<typename... CallbackTypes, typename>
 LogisticRegression<MatType>::LogisticRegression(
     const MatType& predictors,
     const arma::Row<size_t>& responses,
-    const arma::rowvec& initialPoint,
-    const double lambda) :
+    const LogisticRegression<MatType>::RowType& initialPoint,
+    const double lambda,
+    CallbackTypes&&... callbacks) :
     parameters(initialPoint),
     lambda(lambda)
 {
-  Train(predictors, responses);
+  Train(predictors, responses, std::forward<CallbackTypes>(callbacks)...);
 }
 
 template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
 LogisticRegression<MatType>::LogisticRegression(
-    const size_t dimensionality,
-    const double lambda) :
-    parameters(arma::rowvec(dimensionality + 1, arma::fill::zeros)),
+    const MatType& predictors,
+    const arma::Row<size_t>& responses,
+    OptimizerType& optimizer,
+    const double lambda,
+    CallbackTypes&&... callbacks) :
     lambda(lambda)
 {
-  // No training to do here.
+  Train(predictors, responses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
 template<typename MatType>
-template<typename OptimizerType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
 LogisticRegression<MatType>::LogisticRegression(
     const MatType& predictors,
     const arma::Row<size_t>& responses,
     OptimizerType& optimizer,
-    const double lambda) :
+    const LogisticRegression<MatType>::RowType& initialPoint,
+    const double lambda,
+    CallbackTypes&&... callbacks) :
+    parameters(initialPoint),
     lambda(lambda)
 {
-  Train(predictors, responses, optimizer);
+  Train(predictors, responses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
 template<typename MatType>
-template<typename OptimizerType, typename... CallbackTypes>
-double LogisticRegression<MatType>::Train(
-        const MatType& predictors,
-        const arma::Row<size_t>& responses,
-        CallbackTypes&&... callbacks)
+template<typename OptimizerType, typename... CallbackTypes, typename>
+typename LogisticRegression<MatType>::ElemType
+LogisticRegression<MatType>::Train(
+    const MatType& predictors,
+    const arma::Row<size_t>& responses,
+    CallbackTypes&&... callbacks)
 {
   OptimizerType optimizer;
-  return Train(predictors, responses, optimizer, callbacks...);
+  return Train(predictors, responses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
 template<typename MatType>
-template<typename OptimizerType, typename... CallbackTypes>
-double LogisticRegression<MatType>::Train(
+template<typename OptimizerType, typename... CallbackTypes, typename>
+typename LogisticRegression<MatType>::ElemType
+LogisticRegression<MatType>::Train(
+    const MatType& predictors,
+    const arma::Row<size_t>& responses,
+    const double lambda,
+    CallbackTypes&&... callbacks)
+{
+  this->lambda = lambda;
+  OptimizerType optimizer;
+  return Train(predictors, responses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
+}
+
+template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+typename LogisticRegression<MatType>::ElemType
+LogisticRegression<MatType>::Train(
     const MatType& predictors,
     const arma::Row<size_t>& responses,
     OptimizerType& optimizer,
@@ -87,7 +128,7 @@ double LogisticRegression<MatType>::Trai
 
   // Set size of parameters vector according to the input data received.
   if (parameters.n_elem != predictors.n_rows + 1)
-    parameters = arma::rowvec(predictors.n_rows + 1, arma::fill::zeros);
+    parameters.zeros(predictors.n_rows + 1);
 
   const double out = optimizer.Optimize(errorFunction, parameters,
       callbacks...);
@@ -99,14 +140,53 @@ double LogisticRegression<MatType>::Trai
 }
 
 template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+typename LogisticRegression<MatType>::ElemType
+LogisticRegression<MatType>::Train(
+    const MatType& predictors,
+    const arma::Row<size_t>& responses,
+    OptimizerType& optimizer,
+    const double lambda,
+    CallbackTypes&&... callbacks)
+{
+  this->lambda = lambda;
+  return Train(predictors, responses, optimizer,
+      std::forward<CallbackTypes>(callbacks)...);
+}
+
+template<typename MatType>
 template<typename VecType>
 size_t LogisticRegression<MatType>::Classify(const VecType& point,
                                              const double decisionBoundary)
     const
 {
-  return size_t(1.0 / (1.0 + std::exp(-parameters(0) - arma::dot(point,
-      parameters.tail_cols(parameters.n_elem - 1)))) +
-      (1.0 - decisionBoundary));
+  // Used to prevent automatic casting to double.
+  constexpr ElemType one = ((ElemType) 1);
+
+  return size_t(one / (one + std::exp(-parameters(0) - dot(point,
+      parameters.tail_cols(parameters.n_elem - 1).t()))) +
+      (one - ((ElemType) decisionBoundary)));
+}
+
+template<typename MatType>
+template<typename VecType>
+void LogisticRegression<MatType>::Classify(
+    const VecType& point,
+    size_t& prediction,
+    LogisticRegression<MatType>::ColType& probabilities,
+    const double decisionBoundary) const
+{
+  // Used to prevent automatic casting to double.
+  constexpr ElemType one = ((ElemType) 1);
+
+  const ElemType logit = one / (one + std::exp(-parameters(0) - dot(point,
+      parameters.tail_cols(parameters.n_elem - 1).t())));
+
+  probabilities.set_size(2);
+  probabilities[0] = (1 - logit);
+  probabilities[1] = logit;
+
+  prediction = (size_t) (logit + (one - ((ElemType) decisionBoundary)));
 }
 
 template<typename MatType>
@@ -114,27 +194,59 @@ void LogisticRegression<MatType>::Classi
                                            arma::Row<size_t>& labels,
                                            const double decisionBoundary) const
 {
+  // Used to prevent automatic casting to double.
+  constexpr ElemType one = ((ElemType) 1);
+
   // Calculate sigmoid function for each point.  The (1.0 - decisionBoundary)
   // term correctly sets an offset so that floor() returns 0 or 1 correctly.
-  labels = arma::conv_to<arma::Row<size_t>>::from((1.0 /
-      (1.0 + arma::exp(-parameters(0) -
+  labels = ConvTo<arma::Row<size_t>>::From((one /
+      (one + exp(-parameters(0) -
       parameters.tail_cols(parameters.n_elem - 1) * dataset))) +
-      (1.0 - decisionBoundary));
+      (one - ((ElemType) decisionBoundary)));
 }
 
 template<typename MatType>
+mlpack_deprecated
 void LogisticRegression<MatType>::Classify(const MatType& dataset,
-                                           arma::mat& probabilities) const
+                                           MatType& probabilities) const
 {
   // Set correct size of output matrix.
   probabilities.set_size(2, dataset.n_cols);
 
-  probabilities.row(1) = 1.0 / (1.0 + arma::exp(-parameters(0) -
+  probabilities.row(1) = 1.0 / (1.0 + exp(-parameters(0) -
       parameters.tail_cols(parameters.n_elem - 1) * dataset));
   probabilities.row(0) = 1.0 - probabilities.row(1);
 }
 
 template<typename MatType>
+void LogisticRegression<MatType>::Classify(const MatType& dataset,
+                                           arma::Row<size_t>& predictions,
+                                           MatType& probabilities,
+                                           const double decisionBoundary) const
+{
+  // Used to prevent automatic casting to double.
+  constexpr ElemType one = ((ElemType) 1);
+
+  // Set correct sizes for outputs.
+  predictions.set_size(dataset.n_cols);
+  probabilities.set_size(2, dataset.n_cols);
+
+  probabilities.row(1) = one / (one + exp(-parameters(0) -
+      parameters.tail_cols(parameters.n_elem - 1) * dataset));
+  probabilities.row(0) = one - probabilities.row(1);
+
+  predictions = ConvTo<arma::Row<size_t>>::From(probabilities.row(1) +
+      (one - ((ElemType) decisionBoundary)));
+}
+
+template<typename MatType>
+void LogisticRegression<MatType>::Reset()
+{
+  // Zero out parameters vector.
+  parameters.zeros();
+}
+
+template<typename MatType>
 double LogisticRegression<MatType>::ComputeError(
     const MatType& predictors,
     const arma::Row<size_t>& responses) const
@@ -170,10 +282,22 @@ double LogisticRegression<MatType>::Comp
 template<typename MatType>
 template<typename Archive>
 void LogisticRegression<MatType>::serialize(Archive& ar,
-    const uint32_t /* version */)
+    const uint32_t version)
 {
-  ar(CEREAL_NVP(parameters));
-  ar(CEREAL_NVP(lambda));
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // This is the legacy version: `parameters` is of type arma::rowvec.
+    arma::rowvec parametersTmp;
+    ar(cereal::make_nvp("parameters", parametersTmp));
+    parameters = ConvTo<RowType>::From(parametersTmp);
+
+    ar(CEREAL_NVP(lambda));
+  }
+  else
+  {
+    ar(CEREAL_NVP(parameters));
+    ar(CEREAL_NVP(lambda));
+  }
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_main.cpp 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_main.cpp
--- 4.3.0-2/src/mlpack/methods/logistic_regression/logistic_regression_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/logistic_regression/logistic_regression_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -101,7 +101,8 @@ BINDING_EXAMPLE(
     PRINT_MODEL("lr_model") + "', the following command may be used:"
     "\n\n" +
     PRINT_CALL("logistic_regression", "training", "data", "labels", "labels",
-        "lambda", 0.1, "output_model", "lr_model") +
+        "lambda", 0.1, "output_model", "lr_model", "print_training_accuracy",
+        true) +
     "\n\n"
     "Then, to use that model to predict classes for the dataset '" +
     PRINT_DATASET("test") + "', storing the output predictions in '" +
@@ -153,6 +154,9 @@ PARAM_MATRIX_OUT("probabilities", "If te
 PARAM_DOUBLE_IN("decision_boundary", "Decision boundary for prediction; if the "
     "logistic function for a point is less than the boundary, the class is "
     "taken to be 0; otherwise, the class is 1.", "d", 0.5);
+PARAM_FLAG("print_training_accuracy", "If set, then the accuracy of the model "
+    "on the training set will be printed (verbose must also be specified).",
+    "a");
 
 void BINDING_FUNCTION(util::Params& params, util::Timers& timers)
 {
@@ -182,6 +186,12 @@ void BINDING_FUNCTION(util::Params& para
   ReportIgnoredParam(params, {{ "test", false }}, "predictions");
   ReportIgnoredParam(params, {{ "test", false }}, "probabilities");
 
+  ReportIgnoredParam(params, {{ "training", false }}, "print_training_accuracy");
+  
+  RequireAtLeastOnePassed(params,
+      { "test", "output_model", "print_training_accuracy" }, false,
+      "the trained logistic regression model will not be used or saved");
+
   // Max Iterations needs to be positive.
   RequireParamValue<int>(params, "max_iterations", [](int x) { return x >= 0; },
       true, "max_iterations must be positive or zero");
@@ -245,9 +255,9 @@ void BINDING_FUNCTION(util::Params& para
 
     // Set the size of the parameters vector, if necessary.
     if (!params.Has("labels"))
-      model->Parameters() = arma::zeros<arma::rowvec>(regressors.n_rows);
+      model->Parameters() = zeros<arma::rowvec>(regressors.n_rows);
     else
-      model->Parameters() = arma::zeros<arma::rowvec>(regressors.n_rows + 1);
+      model->Parameters() = zeros<arma::rowvec>(regressors.n_rows + 1);
   }
 
   // Check if the responses are in a separate file.
@@ -278,7 +288,7 @@ void BINDING_FUNCTION(util::Params& para
     }
 
     // The initial predictors for y, Nx1.
-    responses = arma::conv_to<arma::Row<size_t>>::from(
+    responses = ConvTo<arma::Row<size_t>>::From(
         regressors.row(regressors.n_rows - 1));
     regressors.shed_row(regressors.n_rows - 1);
   }
@@ -325,8 +335,23 @@ void BINDING_FUNCTION(util::Params& para
       model->Train(regressors, responses, lbfgsOpt);
       timers.Stop("logistic_regression_optimization");
     }
-  }
 
+    // Did we want training accuracy?
+    if (params.Has("print_training_accuracy"))
+    {
+      timers.Start("lr_prediction");
+      arma::Row<size_t> predictions;
+      model->Classify(regressors, predictions);
+
+      const size_t correct = accu(predictions == responses);
+
+      Log::Info << correct << " of " << responses.n_elem << " correct on training"
+          << " set (" << (double(correct) / double(responses.n_elem) * 100) << ")."
+          << endl;
+      timers.Stop("lr_prediction");
+    }
+  }
+  
   if (params.Has("test"))
   {
     const arma::mat& testSet = params.Get<arma::mat>("test");
@@ -361,7 +386,7 @@ void BINDING_FUNCTION(util::Params& para
       Log::Info << "Calculating class probabilities of points in '"
           << params.GetPrintable<arma::mat>("test") << "'." << endl;
       arma::mat probabilities;
-      model->Classify(testSet, probabilities);
+      model->Classify(testSet, predictions, probabilities);
 
       if (params.Has("probabilities"))
         params.Get<arma::mat>("probabilities") = std::move(probabilities);
diff -pruN 4.3.0-2/src/mlpack/methods/lsh/lsh_main.cpp 4.4.0-1/src/mlpack/methods/lsh/lsh_main.cpp
--- 4.3.0-2/src/mlpack/methods/lsh/lsh_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lsh/lsh_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -70,7 +70,7 @@ BINDING_SEE_ALSO("@krann", "#krann");
 BINDING_SEE_ALSO("Locality-sensitive hashing on Wikipedia",
     "https://en.wikipedia.org/wiki/Locality-sensitive_hashing");
 BINDING_SEE_ALSO("Locality-sensitive hashing scheme based on p-stable"
-    "  distributions(pdf)", "http://mlpack.org/papers/lsh.pdf");
+    "  distributions(pdf)", "https://www.mlpack.org/papers/lsh.pdf");
 BINDING_SEE_ALSO("LSHSearch C++ class documentation",
     "@src/mlpack/methods/lsh/lsh.hpp");
 
diff -pruN 4.3.0-2/src/mlpack/methods/lsh/lsh_search_impl.hpp 4.4.0-1/src/mlpack/methods/lsh/lsh_search_impl.hpp
--- 4.3.0-2/src/mlpack/methods/lsh/lsh_search_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/lsh/lsh_search_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -270,8 +270,8 @@ void LSHSearch<SortPolicy, MatType>::Tra
     // and the corresponding offset be 'offset_i'.  Then the key of a single
     // point is obtained as:
     // key = { floor((<proj_i, point> + offset_i) / 'hashWidth') forall i }
-    arma::mat offsetMat = arma::repmat(offsets.unsafe_col(i), 1,
-                                       this->referenceSet.n_cols);
+    arma::mat offsetMat = repmat(offsets.unsafe_col(i), 1,
+                                 this->referenceSet.n_cols);
     arma::mat hashMat = projections.slice(i).t() * (this->referenceSet);
     hashMat += offsetMat;
     hashMat /= hashWidth;
@@ -308,7 +308,7 @@ void LSHSearch<SortPolicy, MatType>::Tra
   secondHashBinCounts.transform([effectiveBucketSize](size_t val)
       { return std::min(val, effectiveBucketSize); });
 
-  const size_t numRowsInTable = arma::accu(secondHashBinCounts > 0);
+  const size_t numRowsInTable = accu(secondHashBinCounts > 0);
   bucketContentSize.zeros(numRowsInTable);
   secondHashTable.resize(numRowsInTable);
 
@@ -343,8 +343,8 @@ void LSHSearch<SortPolicy, MatType>::Tra
   } // Loop over tables.
 
   Log::Info << "Final hash table size: " << numRowsInTable << " rows, with a "
-            << "maximum length of " << arma::max(secondHashBinCounts) << ", "
-            << "totaling " << arma::accu(secondHashBinCounts) << " elements."
+            << "maximum length of " << max(secondHashBinCounts) << ", "
+            << "totaling " << accu(secondHashBinCounts) << " elements."
             << std::endl;
 }
 
@@ -548,15 +548,15 @@ void LSHSearch<SortPolicy, MatType>::Get
 
   // Calculate scores. score = distance^2.
   arma::vec scores(2 * numProj);
-  scores.rows(0, numProj - 1) = arma::pow(limLow, 2);
-  scores.rows(numProj, (2 * numProj) - 1) = arma::pow(limHigh, 2);
+  scores.rows(0, numProj - 1) = pow(limLow, 2);
+  scores.rows(numProj, (2 * numProj) - 1) = pow(limHigh, 2);
 
   // Actions vector describes what perturbation (-1/+1) corresponds to a score.
   arma::Col<short int> actions(2 * numProj); // will be [-1 ... 1 ...]
   actions.rows(0, numProj - 1) = // First numProj rows.
-    -1 * arma::ones< arma::Col<short int> > (numProj); // -1s
+    -1 * ones<arma::Col<short int>> (numProj); // -1s
   actions.rows(numProj, (2 * numProj) - 1) = // Last numProj rows.
-    arma::ones< arma::Col<short int> > (numProj); // 1s
+    ones<arma::Col<short int>> (numProj); // 1s
 
 
   // Acting dimension vector shows which coordinate to transform according to
@@ -739,8 +739,8 @@ void LSHSearch<SortPolicy, MatType>::Ret
 
   // Compute the primary hash value of each key of the query into a bucket of
   // the secondHashTable using the secondHashWeights.
-  hashMat.row(0) = arma::conv_to<arma::Row<size_t>> // Floor by typecasting
-      ::from(secondHashWeights.t() * allProjInTables);
+  hashMat.row(0) = ConvTo<arma::Row<size_t>> // Floor by typecasting
+      ::From(secondHashWeights.t() * allProjInTables);
   // Mod to compute 2nd-level codes.
   for (size_t i = 0; i < numTablesToSearch; ++i)
     hashMat(0, i) = (hashMat(0, i) % secondHashSize);
@@ -760,8 +760,8 @@ void LSHSearch<SortPolicy, MatType>::Ret
       // Map each probing bin to a bin in secondHashTable (just like we did for
       // the primary hash table).
       hashMat(arma::span(1, T), i) = // Compute code of rows 1:end of column i
-        arma::conv_to< arma::Col<size_t> >:: // floor by typecasting to size_t
-        from(secondHashWeights.t() * additionalProbingBins);
+        ConvTo<arma::Col<size_t>>:: // floor by typecasting to size_t
+        From(secondHashWeights.t() * additionalProbingBins);
       for (size_t p = 1; p < T + 1; ++p)
         hashMat(p, i) = (hashMat(p, i) % secondHashSize);
     }
diff -pruN 4.3.0-2/src/mlpack/methods/matrix_completion/matrix_completion_impl.hpp 4.4.0-1/src/mlpack/methods/matrix_completion/matrix_completion_impl.hpp
--- 4.3.0-2/src/mlpack/methods/matrix_completion/matrix_completion_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/matrix_completion/matrix_completion_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,7 +27,7 @@ inline MatrixCompletion::MatrixCompletio
     n(n),
     indices(indices),
     values(values),
-    sdp(indices.n_cols, 0, arma::randu<arma::mat>(m + n, r))
+    sdp(indices.n_cols, 0, randu<arma::mat>(m + n, r))
 {
   CheckValues();
   InitSDP();
@@ -59,7 +59,7 @@ inline MatrixCompletion::MatrixCompletio
     indices(indices),
     values(values),
     sdp(indices.n_cols, 0,
-        arma::randu<arma::mat>(m + n, DefaultRank(m, n, indices.n_cols)))
+        randu<arma::mat>(m + n, DefaultRank(m, n, indices.n_cols)))
 {
   CheckValues();
   InitSDP();
@@ -119,7 +119,7 @@ inline size_t MatrixCompletion::DefaultR
   //   Pablo Moscato, Michael Norman, and Gabor Pataki.
   //   Math Oper. Res., 23(2). 1998.
   const size_t mpn = m + n;
-  float r = 0.5 + sqrt(0.25 + 2 * p);
+  float r = 0.5 + std::sqrt(0.25 + 2 * p);
   if (ceil(r) > mpn)
     r = mpn; // An upper bound on the dimension.
   return ceil(r);
diff -pruN 4.3.0-2/src/mlpack/methods/mean_shift/mean_shift_impl.hpp 4.4.0-1/src/mlpack/methods/mean_shift/mean_shift_impl.hpp
--- 4.3.0-2/src/mlpack/methods/mean_shift/mean_shift_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/mean_shift/mean_shift_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -67,7 +67,7 @@ EstimateRadius(const MatType& data, doub
   arma::rowvec maxDistances = max(distances);
 
   // Calculate and return the radius.
-  return arma::sum(maxDistances) / (double) data.n_cols;
+  return sum(maxDistances) / (double) data.n_cols;
 }
 
 // Class to compare two vectors.
@@ -220,7 +220,7 @@ inline void MeanShift<UseKernel, KernelT
         || forceConvergence; completedIterations++)
     {
       // Store new centroid in this.
-      arma::colvec newCentroid = arma::zeros<arma::colvec>(pSeeds->n_rows);
+      arma::colvec newCentroid = zeros<arma::colvec>(pSeeds->n_rows);
 
       rangeSearcher.Search(allCentroids.unsafe_col(i), validRadius,
           neighbors, distances);
diff -pruN 4.3.0-2/src/mlpack/methods/mean_shift/mean_shift_main.cpp 4.4.0-1/src/mlpack/methods/mean_shift/mean_shift_main.cpp
--- 4.3.0-2/src/mlpack/methods/mean_shift/mean_shift_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/mean_shift/mean_shift_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,8 +62,8 @@ BINDING_SEE_ALSO("@dbscan", "#dbscan");
 BINDING_SEE_ALSO("Mean shift on Wikipedia",
         "https://en.wikipedia.org/wiki/Mean_shift");
 BINDING_SEE_ALSO("Mean Shift, Mode Seeking, and Clustering (pdf)",
-        "http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.510.1222"
-        "&rep=rep1&type=pdf");
+        "https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf"
+        "&doi=1c168275c59ba382588350ee1443537f59978183");
 BINDING_SEE_ALSO("mlpack::mean_shift::MeanShift C++ class documentation",
         "@src/mlpack/methods/mean_shift/mean_shift.hpp");
 
@@ -156,7 +156,7 @@ void BINDING_FUNCTION(util::Params& para
       // TODO: figure out how to output as an arma::Mat<size_t> so that files
       // aren't way larger than needed.
       params.Get<arma::mat>("output") =
-          arma::conv_to<arma::mat>::from(assignments);
+          ConvTo<arma::mat>::From(assignments);
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp 4.4.0-1/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp
--- 4.3.0-2/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -77,7 +77,8 @@ class NaiveBayesClassifier
    * @param incrementalVariance If true, an incremental algorithm is used to
    *     calculate the variance; this can prevent loss of precision in some
    *     cases, but will be somewhat slower to calculate.
-   * @param epsilon Small value to prevent log of zero.
+   * @param epsilon Small initialization value for variances to prevent log of
+   *     zero.
    */
   template<typename MatType>
   NaiveBayesClassifier(const MatType& data,
@@ -109,7 +110,7 @@ class NaiveBayesClassifier
    *
    * @param data The dataset to train on.
    * @param labels The labels for the dataset.
-   * @param numClasses The numbe of classes in the dataset.
+   * @param numClasses The number of classes in the dataset.
    * @param incremental Whether or not to use the incremental algorithm for
    *      training.
    */
@@ -120,6 +121,32 @@ class NaiveBayesClassifier
              const bool incremental = true);
 
   /**
+   * Train the Naive Bayes classifier on the given dataset.  If the incremental
+   * algorithm is used, the current model is used as a starting point (this is
+   * the default).  If the incremental algorithm is not used, then the current
+   * model is ignored and the new model will be trained only on the given data.
+   * Note that even if the incremental algorithm is not used, the data must have
+   * the same dimensionality and number of classes that the model was
+   * initialized with.  If you want to change the dimensionality or number of
+   * classes, either re-initialize or call Means(), Variances(), and
+   * Probabilities() individually to set them to the right size.
+   *
+   * @param data The dataset to train on.
+   * @param labels The labels for the dataset.
+   * @param numClasses The number of classes in the dataset.
+   * @param incremental Whether or not to use the incremental algorithm for
+   *      training.
+   * @param epsilon Small reinitialization value for variances to prevent log of
+   *      zero (ignored if incremental is true).
+   */
+  template<typename MatType>
+  void Train(const MatType& data,
+             const arma::Row<size_t>& labels,
+             const size_t numClasses,
+             const bool incremental,
+             const double epsilon);
+
+  /**
    * Train the Naive Bayes classifier on the given point.  This will use the
    * incremental algorithm for updating the model parameters.  The data must be
    * the same dimensionality as the existing model parameters.
@@ -198,6 +225,21 @@ class NaiveBayesClassifier
                 arma::Row<size_t>& predictions,
                 ProbabilitiesMatType& probabilities) const;
 
+  /**
+   * Reset the model to zeros, keeping the model's current dimensionality and
+   * number of classes.
+   */
+  void Reset();
+
+  /**
+   * Reset the model to zeros, with a new dimensionality and number of classes.
+   * The value epsilon specifies an initial very small value for the variances,
+   * to prevent log(0) issues.
+   */
+  void Reset(const size_t dimensionality,
+             const size_t numClasses,
+             const double epsilon = 1e-10);
+
   //! Get the sample means for each class.
   const ModelMatType& Means() const { return means; }
   //! Modify the sample means for each class.
@@ -213,9 +255,12 @@ class NaiveBayesClassifier
   //! Modify the prior probabilities for each class.
   ModelMatType& Probabilities() { return probabilities; }
 
+  //! Get the number of points the model has been trained on so far.
+  size_t TrainingPoints() const { return trainingPoints; }
+
   //! Serialize the classifier.
   template<typename Archive>
-  void serialize(Archive& ar, const uint32_t /* version */);
+  void serialize(Archive& ar, const uint32_t version);
 
  private:
   //! Sample mean for each class.
@@ -244,6 +289,9 @@ class NaiveBayesClassifier
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename MatType),
+    (mlpack::NaiveBayesClassifier<MatType>), (1));
+
 // Include implementation.
 #include "naive_bayes_classifier_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/naive_bayes/naive_bayes_classifier_impl.hpp 4.4.0-1/src/mlpack/methods/naive_bayes/naive_bayes_classifier_impl.hpp
--- 4.3.0-2/src/mlpack/methods/naive_bayes/naive_bayes_classifier_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/naive_bayes/naive_bayes_classifier_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -46,7 +46,8 @@ NaiveBayesClassifier<ModelMatType>::Naiv
   {
     probabilities.zeros(numClasses);
     means.zeros(data.n_rows, numClasses);
-    variances.zeros(data.n_rows, numClasses);
+    variances.set_size(data.n_rows, numClasses);
+    variances.fill(epsilon);
   }
   else
   {
@@ -69,7 +70,8 @@ NaiveBayesClassifier<ModelMatType>::Naiv
   // Initialize model to 0.
   probabilities.zeros(numClasses);
   means.zeros(dimensionality, numClasses);
-  variances.zeros(dimensionality, numClasses);
+  variances.set_size(dimensionality, numClasses);
+  variances.fill(epsilon);
 }
 
 template<typename ModelMatType>
@@ -84,30 +86,14 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
-  // Do we need to resize the model?
-  if (probabilities.n_elem != numClasses)
-  {
-    // Perform training, after initializing the model to 0 (that is, if Train()
-    // won't do that for us, which it won't if we're using the incremental
-    // algorithm).
-    if (incremental)
-    {
-      probabilities.zeros(numClasses);
-      means.zeros(data.n_rows, numClasses);
-      variances.zeros(data.n_rows, numClasses);
-    }
-    else
-    {
-      probabilities.set_size(numClasses);
-      means.set_size(data.n_rows, numClasses);
-      variances.set_size(data.n_rows, numClasses);
-    }
-  }
-
   // Calculate the class probabilities as well as the sample mean and variance
   // for each of the features with respect to each of the labels.
   if (incremental)
   {
+    // Do we need to resize the model?
+    if (probabilities.n_elem != numClasses || data.n_rows != means.n_rows)
+      Reset(data.n_rows, numClasses);
+
     // Use incremental algorithm.
     // Fist, de-normalize probabilities.
     probabilities *= trainingPoints;
@@ -117,7 +103,7 @@ void NaiveBayesClassifier<ModelMatType>:
       const size_t label = labels[j];
       ++probabilities[label];
 
-      arma::vec delta = data.col(j) - means.col(label);
+      arma::Col<ElemType> delta = data.col(j) - means.col(label);
       means.col(label) += delta / probabilities[label];
       variances.col(label) += delta % (data.col(j) - means.col(label));
     }
@@ -131,9 +117,9 @@ void NaiveBayesClassifier<ModelMatType>:
   else
   {
     // Set all parameters to zero.
-    probabilities.zeros();
-    means.zeros();
-    variances.zeros();
+    probabilities.zeros(numClasses);
+    means.zeros(data.n_rows, numClasses);
+    variances.zeros(data.n_rows, numClasses);
 
     // Don't use incremental algorithm.  This is a two-pass algorithm.  It is
     // possible to calculate the means and variances using a faster one-pass
@@ -175,6 +161,19 @@ void NaiveBayesClassifier<ModelMatType>:
 }
 
 template<typename ModelMatType>
+template<typename MatType>
+void NaiveBayesClassifier<ModelMatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const bool incremental,
+    const double epsilon)
+{
+  this->epsilon = epsilon;
+  Train(data, labels, numClasses, incremental);
+}
+
+template<typename ModelMatType>
 template<typename VecType>
 void NaiveBayesClassifier<ModelMatType>::Train(const VecType& point,
                                                const size_t label)
@@ -183,11 +182,20 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
+  if (point.n_elem != means.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "NaiveBayesClassifier::Train(): given point has dimensionality "
+        << point.n_elem << ", but model has dimensionality " << means.n_rows
+        << "!";
+    throw std::invalid_argument(oss.str());
+  }
+
   // We must use the incremental algorithm here.
   probabilities *= trainingPoints;
   probabilities[label]++;
 
-  arma::vec delta = point - means.col(label);
+  arma::Col<ElemType> delta = point - means.col(label);
   means.col(label) += delta / probabilities[label];
   if (probabilities[label] > 2)
     variances.col(label) *= (probabilities[label] - 2);
@@ -209,7 +217,7 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
-  logLikelihoods = arma::log(arma::repmat(probabilities, 1, data.n_cols));
+  logLikelihoods = log(repmat(probabilities, 1, data.n_cols));
   ModelMatType invVar = 1.0 / variances;
 
   // Calculate the joint log likelihood of point for each of the
@@ -220,12 +228,12 @@ void NaiveBayesClassifier<ModelMatType>:
   {
     // This is an adaptation of phi() for the case where the covariance is a
     // diagonal matrix.
-    ModelMatType diffs = data - arma::repmat(means.col(i), 1, data.n_cols);
+    ModelMatType diffs = data - repmat(means.col(i), 1, data.n_cols);
     ModelMatType rhs = -0.5 * arma::diagmat(invVar.col(i)) * diffs;
-    arma::Mat<ElemType> exponents = arma::sum(diffs % rhs, 0);
+    arma::Mat<ElemType> exponents = sum(diffs % rhs, 0);
 
-    logLikelihoods.row(i) += (data.n_rows / -2.0 * log(2 * M_PI) - 0.5 *
-        arma::accu(arma::log(variances.col(i))) + exponents);
+    logLikelihoods.row(i) += (data.n_rows / -2.0 * std::log(2 * M_PI) - 0.5 *
+        accu(log(variances.col(i))) + exponents);
   }
 }
 
@@ -237,6 +245,15 @@ size_t NaiveBayesClassifier<ModelMatType
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
+  if (point.n_elem != means.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "NaiveBayesClassifier::Classify(): given point has dimensionality "
+        << point.n_elem << ", but model has dimensionality " << means.n_rows
+        << "!";
+    throw std::invalid_argument(oss.str());
+  }
+
   // Find the label(class) with max log likelihood.
   ModelMatType logLikelihoods;
   LogLikelihood(point, logLikelihoods);
@@ -261,6 +278,15 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
+  if (point.n_elem != means.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "NaiveBayesClassifier::Classify(): given point has dimensionality "
+        << point.n_elem << ", but model has dimensionality " << means.n_rows
+        << "!";
+    throw std::invalid_argument(oss.str());
+  }
+
   // log(Prob(Y|X)) = Log(Prob(X|Y)) + Log(Prob(Y)) - Log(Prob(X));
   // But LogLikelihood() gives us the unnormalized log likelihood which is
   // Log(Prob(X|Y)) + Log(Prob(Y)) so we need to subtract the normalization
@@ -270,8 +296,8 @@ void NaiveBayesClassifier<ModelMatType>:
 
   // To prevent underflow in log of sum of exp of x operation (where x is a
   // small negative value), we use logsumexp(x - max(x)) + max(x).
-  const double maxValue = arma::max(logLikelihoods);
-  const double logProbX = log(arma::accu(exp(logLikelihoods - maxValue))) +
+  const ElemType maxValue = logLikelihoods.max();
+  const ElemType logProbX = std::log(accu(exp(logLikelihoods - maxValue))) +
       maxValue;
   probabilities = exp(logLikelihoods - logProbX); // log(exp(value)) == value.
 
@@ -290,6 +316,15 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
+  if (data.n_rows != means.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "NaiveBayesClassifier::Classify(): given data has dimensionality "
+        << data.n_rows << ", but model has dimensionality " << means.n_rows
+        << "!";
+    throw std::invalid_argument(oss.str());
+  }
+
   predictions.set_size(data.n_cols);
 
   ModelMatType logLikelihoods;
@@ -318,6 +353,15 @@ void NaiveBayesClassifier<ModelMatType>:
       "NaiveBayesClassifier: element type of given data must match the element "
       "type of the model!");
 
+  if (data.n_rows != means.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "NaiveBayesClassifier::Classify(): given data has dimensionality "
+        << data.n_rows << ", but model has dimensionality " << means.n_rows
+        << "!";
+    throw std::invalid_argument(oss.str());
+  }
+
   predictions.set_size(data.n_cols);
 
   ModelMatType logLikelihoods;
@@ -331,10 +375,10 @@ void NaiveBayesClassifier<ModelMatType>:
     // Log(Prob(X|Y)) + Log(Prob(Y)), so we subtract the normalization term.
     // Besides, to prevent underflow in log of sum of exp of x operation (where
     // x is a small negative value), we use logsumexp(x - max(x)) + max(x).
-    maxValue = arma::max(logLikelihoods.col(j));
-    logProbX = log(arma::accu(exp(logLikelihoods.col(j) -
+    maxValue = max(logLikelihoods.col(j));
+    logProbX = std::log(accu(exp(logLikelihoods.col(j) -
         maxValue))) + maxValue;
-    predictionProbs.col(j) = arma::exp(logLikelihoods.col(j) - logProbX);
+    predictionProbs.col(j) = exp(logLikelihoods.col(j) - logProbX);
   }
 
   // Now calculate maximum probabilities for each point.
@@ -347,14 +391,49 @@ void NaiveBayesClassifier<ModelMatType>:
 }
 
 template<typename ModelMatType>
+void NaiveBayesClassifier<ModelMatType>::Reset()
+{
+  means.zeros();
+  probabilities.zeros();
+  variances.fill(epsilon);
+  trainingPoints = 0;
+}
+
+template<typename ModelMatType>
+void NaiveBayesClassifier<ModelMatType>::Reset(const size_t dimensionality,
+                                               const size_t numClasses,
+                                               const double epsilon)
+{
+  this->epsilon = epsilon;
+
+  probabilities.zeros(numClasses);
+  means.zeros(dimensionality, numClasses);
+  variances.set_size(dimensionality, numClasses);
+  variances.fill(epsilon);
+  trainingPoints = 0;
+}
+
+template<typename ModelMatType>
 template<typename Archive>
 void NaiveBayesClassifier<ModelMatType>::serialize(
     Archive& ar,
-    const uint32_t /* version */)
+    const uint32_t version)
 {
   ar(CEREAL_NVP(means));
   ar(CEREAL_NVP(variances));
   ar(CEREAL_NVP(probabilities));
+
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // Old versions did not serialize the trainingPoints or epsilon members.
+    trainingPoints = 0;
+    epsilon = 1e-10;
+  }
+  else
+  {
+    ar(CEREAL_NVP(trainingPoints));
+    ar(CEREAL_NVP(epsilon));
+  }
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/naive_bayes/nbc_main.cpp 4.4.0-1/src/mlpack/methods/naive_bayes/nbc_main.cpp
--- 4.3.0-2/src/mlpack/methods/naive_bayes/nbc_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/naive_bayes/nbc_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -61,12 +61,7 @@ BINDING_LONG_DESC(
     " may be saved with the " + PRINT_PARAM_STRING("predictions") +"predictions"
     "  parameter.  If saving the trained model is desired, this may be "
     "done with the " + PRINT_PARAM_STRING("output_model") + " output "
-    "parameter."
-    "\n\n"
-    "Note: the " + PRINT_PARAM_STRING("output") + " and " +
-    PRINT_PARAM_STRING("output_probs") + " parameters are deprecated and will "
-    "be removed in mlpack 4.0.0.  Use " + PRINT_PARAM_STRING("predictions") +
-    " and " + PRINT_PARAM_STRING("probabilities") + " instead.");
+    "parameter.");
 
 // Example.
 BINDING_EXAMPLE(
@@ -83,8 +78,8 @@ BINDING_EXAMPLE(
     "classes to " + PRINT_DATASET("predictions") + ", the following command "
     "may be used:"
     "\n\n" +
-    PRINT_CALL("nbc", "input_model", "nbc_model", "test", "test_set", "output",
-        "predictions"));
+    PRINT_CALL("nbc", "input_model", "nbc_model", "test", "test_set",
+        "predictions", "predictions"));
 
 // See also...
 BINDING_SEE_ALSO("@softmax_regression", "#softmax_regression");
@@ -92,7 +87,7 @@ BINDING_SEE_ALSO("@random_forest", "#ran
 BINDING_SEE_ALSO("Naive Bayes classifier on Wikipedia",
     "https://en.wikipedia.org/wiki/Naive_Bayes_classifier");
 BINDING_SEE_ALSO("NaiveBayesClassifier C++ class documentation",
-    "@src/mlpack/methods/naive_bayes/naive_bayes_classifier.cpp");
+    "@src/mlpack/methods/naive_bayes/naive_bayes_classifier.hpp");
 
 // A struct for saving the model with mappings.
 struct NBCModel
@@ -126,14 +121,8 @@ PARAM_FLAG("incremental_variance", "The
 
 // Test parameters.
 PARAM_MATRIX_IN("test", "A matrix containing the test set.", "T");
-// The parameter 'output' is deprecated and will be removed in mlpack 4.
-PARAM_UROW_OUT("output", "The matrix in which the predicted labels for the"
-    " test set will be written (deprecated).", "o");
 PARAM_UROW_OUT("predictions", "The matrix in which the predicted labels for the"
     " test set will be written.", "a");
-// The parameter 'output_probs' is deprecated and can be removed in mlpack 4.
-PARAM_MATRIX_OUT("output_probs", "The matrix in which the predicted probability"
-    " of labels for the test set will be written (deprecated).", "");
 PARAM_MATRIX_OUT("probabilities", "The matrix in which the predicted"
     " probability of labels for the test set will be written.", "p");
 
@@ -143,9 +132,8 @@ void BINDING_FUNCTION(util::Params& para
   RequireOnlyOnePassed(params, { "training", "input_model" }, true);
   ReportIgnoredParam(params, {{ "training", false }}, "labels");
   ReportIgnoredParam(params, {{ "training", false }}, "incremental_variance");
-  RequireAtLeastOnePassed(params, { "output", "predictions", "output_model",
-      "output_probs", "probabilities" }, false, "no output will be saved");
-  ReportIgnoredParam(params, {{ "test", false }}, "output");
+  RequireAtLeastOnePassed(params, { "predictions", "output_model",
+      "probabilities" }, false, "no output will be saved");
   ReportIgnoredParam(params, {{ "test", false }}, "predictions");
   if (params.Has("input_model") && !params.Has("test"))
     Log::Warn << "No test set given; no task will be performed!" << std::endl;
@@ -208,23 +196,20 @@ void BINDING_FUNCTION(util::Params& para
     model->nbc.Classify(testingData, predictions, probabilities);
     timers.Stop("nbc_testing");
 
-    if (params.Has("output") || params.Has("predictions"))
+    if (params.Has("predictions"))
     {
       // Un-normalize labels to prepare output.
       Row<size_t> rawResults;
       data::RevertLabels(predictions, model->mappings, rawResults);
 
       if (params.Has("predictions"))
-        params.Get<Row<size_t>>("predictions") = rawResults;
-      if (params.Has("output"))
-        params.Get<Row<size_t>>("output") = std::move(rawResults);
+        params.Get<Row<size_t>>("predictions") = std::move(rawResults);
     }
-    if (params.Has("output_probs") || params.Has("probabilities"))
+
+    if (params.Has("probabilities"))
     {
       if (params.Has("probabilities"))
-        params.Get<mat>("probabilities") = probabilities;
-      if (params.Has("output_probs"))
-        params.Get<mat>("output_probs") = std::move(probabilities);
+        params.Get<mat>("probabilities") = std::move(probabilities);
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/nca/nca_main.cpp 4.4.0-1/src/mlpack/methods/nca/nca_main.cpp
--- 4.3.0-2/src/mlpack/methods/nca/nca_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/nca/nca_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -224,7 +224,7 @@ void BINDING_FUNCTION(util::Params& para
   if (normalize)
   {
     // Find the minimum and maximum values for each dimension.
-    arma::vec ranges = arma::max(data, 1) - arma::min(data, 1);
+    arma::vec ranges = arma::range(data, 1);
     for (size_t d = 0; d < ranges.n_elem; ++d)
       if (ranges[d] == 0.0)
         ranges[d] = 1; // A range of 0 produces NaN later on.
diff -pruN 4.3.0-2/src/mlpack/methods/nca/nca_softmax_error_function_impl.hpp 4.4.0-1/src/mlpack/methods/nca/nca_softmax_error_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/nca/nca_softmax_error_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/nca/nca_softmax_error_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -22,14 +22,15 @@ namespace mlpack {
 // Initialize with the given kernel.
 template<typename MetricType>
 SoftmaxErrorFunction<MetricType>::SoftmaxErrorFunction(
-    const arma::mat& dataset,
-    const arma::Row<size_t>& labels,
+    const arma::mat& datasetIn,
+    const arma::Row<size_t>& labelsIn,
     MetricType metric) :
-    dataset(MakeAlias(const_cast<arma::mat&>(dataset), false)),
-    labels(MakeAlias(const_cast<arma::Row<size_t>&>(labels), false)),
     metric(metric),
     precalculated(false)
-{ /* nothing to do */ }
+{
+  MakeAlias(dataset, datasetIn, datasetIn.n_rows, datasetIn.n_cols, 0, false);
+  MakeAlias(labels, labelsIn, labelsIn.n_elem, 0, false);
+}
 
 //! Shuffle the dataset.
 template<typename MetricType>
@@ -133,7 +134,7 @@ void SoftmaxErrorFunction<MetricType>::G
     for (size_t k = (i + 1); k < stretchedDataset.n_cols; ++k)
     {
       // Calculate p_ik and p_ki first.
-      double eval = exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
+      double eval = std::exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
                                          stretchedDataset.unsafe_col(k)));
       double p_ik = 0, p_ki = 0;
       p_ik = eval / denominators(i);
@@ -188,7 +189,7 @@ void SoftmaxErrorFunction<MetricType>::G
         continue;
 
       // Calculate the numerator of p_ik.
-      double eval = exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
+      double eval = std::exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
                                          stretchedDataset.unsafe_col(k)));
 
       // If the points are in the same class, we must add to the second term of
@@ -269,7 +270,7 @@ void SoftmaxErrorFunction<MetricType>::P
     for (size_t j = (i + 1); j < stretchedDataset.n_cols; ++j)
     {
       // Evaluate exp(-d(x_i, x_j)).
-      double eval = exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
+      double eval = std::exp(-metric.Evaluate(stretchedDataset.unsafe_col(i),
                                          stretchedDataset.unsafe_col(j)));
 
       // Add this to the denominators of both p_i and p_j: K(i, j) = K(j, i).
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/knn_main.cpp 4.4.0-1/src/mlpack/methods/neighbor_search/knn_main.cpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/knn_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/knn_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -66,8 +66,6 @@ BINDING_EXAMPLE(
 BINDING_SEE_ALSO("@lsh", "#lsh");
 BINDING_SEE_ALSO("@krann", "#krann");
 BINDING_SEE_ALSO("@kfn", "#kfn");
-BINDING_SEE_ALSO("NeighborSearch tutorial (k-nearest-neighbors)",
-    "@doc/tutorials/neighbor_search.md");
 BINDING_SEE_ALSO("Tree-independent dual-tree algorithms (pdf)",
     "http://proceedings.mlr.press/v28/curtin13.pdf");
 BINDING_SEE_ALSO("NeighborSearch C++ class documentation",
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search.hpp 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search.hpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -82,6 +82,8 @@ class NeighborSearch
  public:
   //! Convenience typedef.
   typedef TreeType<MetricType, NeighborSearchStat<SortPolicy>, MatType> Tree;
+  //! The type of element held in MatType.
+  typedef typename MatType::elem_type ElemType;
 
   /**
    * Initialize the NeighborSearch object, passing a reference dataset (this is
@@ -223,7 +225,7 @@ class NeighborSearch
   void Search(const MatType& querySet,
               const size_t k,
               arma::Mat<size_t>& neighbors,
-              arma::mat& distances);
+              arma::Mat<ElemType>& distances);
 
   /**
    * Given a pre-built query tree, search for the nearest neighbors of each
@@ -248,7 +250,7 @@ class NeighborSearch
   void Search(Tree& queryTree,
               const size_t k,
               arma::Mat<size_t>& neighbors,
-              arma::mat& distances,
+              arma::Mat<ElemType>& distances,
               bool sameSet = false);
 
   /**
@@ -267,7 +269,7 @@ class NeighborSearch
    */
   void Search(const size_t k,
               arma::Mat<size_t>& neighbors,
-              arma::mat& distances);
+              arma::Mat<ElemType>& distances);
 
   /**
    * Calculate the average relative error (effective error) between the
@@ -284,8 +286,8 @@ class NeighborSearch
    *     query point.
    * @return Average relative error.
    */
-  static double EffectiveError(arma::mat& foundDistances,
-                               arma::mat& realDistances);
+  static double EffectiveError(arma::Mat<ElemType>& foundDistances,
+                               arma::Mat<ElemType>& realDistances);
 
   /**
    * Calculate the recall (% of neighbors found) given the list of found
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_impl.hpp 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_impl.hpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -104,7 +104,7 @@ SingleTreeTraversalType>::NeighborSearch
   // Build the tree on the empty dataset, if necessary.
   if (mode != NAIVE_MODE)
   {
-    referenceTree = BuildTree<Tree>(std::move(arma::mat()),
+    referenceTree = BuildTree<Tree>(std::move(MatType()),
         oldFromNewReferences);
     referenceSet = &referenceTree->Dataset();
   }
@@ -255,7 +255,7 @@ NeighborSearch<SortPolicy,
   if (!other.referenceTree)
     delete other.referenceSet;
 
-  other.referenceTree = BuildTree<Tree>(std::move(arma::mat()),
+  other.referenceTree = BuildTree<Tree>(std::move(MatType()),
       other.oldFromNewReferences);
   other.referenceSet = &other.referenceTree->Dataset();
   other.searchMode = DUAL_TREE_MODE,
@@ -365,7 +365,7 @@ DualTreeTraversalType, SingleTreeTravers
     const MatType& querySet,
     const size_t k,
     arma::Mat<size_t>& neighbors,
-    arma::mat& distances)
+    arma::Mat<ElemType>& distances)
 {
   if (k > referenceSet->n_cols)
   {
@@ -386,14 +386,14 @@ DualTreeTraversalType, SingleTreeTravers
   // To avoid an extra copy, we will store the neighbors and distances in a
   // separate matrix.
   arma::Mat<size_t>* neighborPtr = &neighbors;
-  arma::mat* distancePtr = &distances;
+  arma::Mat<ElemType>* distancePtr = &distances;
 
   // Mapping is only necessary if the tree rearranges points.
   if (TreeTraits<Tree>::RearrangesDataset)
   {
     if (searchMode == DUAL_TREE_MODE)
     {
-      distancePtr = new arma::mat; // Query indices need to be mapped.
+      distancePtr = new arma::Mat<ElemType>; // Query indices need to be mapped.
       neighborPtr = new arma::Mat<size_t>;
     }
     else if (!oldFromNewReferences.empty())
@@ -570,7 +570,7 @@ DualTreeTraversalType, SingleTreeTravers
     Tree& queryTree,
     const size_t k,
     arma::Mat<size_t>& neighbors,
-    arma::mat& distances,
+    arma::Mat<ElemType>& distances,
     bool sameSet)
 {
   if (k > referenceSet->n_cols)
@@ -648,7 +648,7 @@ void NeighborSearch<SortPolicy, MetricTy
 DualTreeTraversalType, SingleTreeTraversalType>::Search(
     const size_t k,
     arma::Mat<size_t>& neighbors,
-    arma::mat& distances)
+    arma::Mat<ElemType>& distances)
 {
   if (k > referenceSet->n_cols)
   {
@@ -670,12 +670,12 @@ DualTreeTraversalType, SingleTreeTravers
   scores = 0;
 
   arma::Mat<size_t>* neighborPtr = &neighbors;
-  arma::mat* distancePtr = &distances;
+  arma::Mat<ElemType>* distancePtr = &distances;
 
   if (!oldFromNewReferences.empty() && TreeTraits<Tree>::RearrangesDataset)
   {
     // We will always need to rearrange in this case.
-    distancePtr = new arma::mat;
+    distancePtr = new MatType;
     neighborPtr = new arma::Mat<size_t>;
   }
 
@@ -825,8 +825,8 @@ template<typename SortPolicy,
          template<typename> class SingleTreeTraversalType>
 double NeighborSearch<SortPolicy, MetricType, MatType, TreeType,
 DualTreeTraversalType, SingleTreeTraversalType>::EffectiveError(
-    arma::mat& foundDistances,
-    arma::mat& realDistances)
+    arma::Mat<ElemType>& foundDistances,
+    arma::Mat<ElemType>& realDistances)
 {
   if (foundDistances.n_rows != realDistances.n_rows ||
       foundDistances.n_cols != realDistances.n_cols)
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_rules.hpp 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_rules.hpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_rules.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_rules.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,6 +34,9 @@ template<typename SortPolicy, typename M
 class NeighborSearchRules
 {
  public:
+  //! The type of element held in MatType.
+  typedef typename TreeType::Mat::elem_type ElemType;
+
   /**
    * Construct the NeighborSearchRules object.  This is usually done from within
    * the NeighborSearch class at search time.
@@ -60,7 +63,7 @@ class NeighborSearchRules
    * @param distances Matrix storing distances of neighbors for each query
    *     point.
    */
-  void GetResults(arma::Mat<size_t>& neighbors, arma::mat& distances);
+  void GetResults(arma::Mat<size_t>& neighbors, arma::Mat<ElemType>& distances);
 
   /**
    * Get the distance from the query point to the reference point.
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_rules_impl.hpp 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/neighbor_search_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/neighbor_search_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -61,7 +61,7 @@ NeighborSearchRules<SortPolicy, MetricTy
 template<typename SortPolicy, typename MetricType, typename TreeType>
 void NeighborSearchRules<SortPolicy, MetricType, TreeType>::GetResults(
     arma::Mat<size_t>& neighbors,
-    arma::mat& distances)
+    arma::Mat<ElemType>& distances)
 {
   neighbors.set_size(k, querySet.n_cols);
   distances.set_size(k, querySet.n_cols);
diff -pruN 4.3.0-2/src/mlpack/methods/neighbor_search/ns_model_impl.hpp 4.4.0-1/src/mlpack/methods/neighbor_search/ns_model_impl.hpp
--- 4.3.0-2/src/mlpack/methods/neighbor_search/ns_model_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/neighbor_search/ns_model_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -576,7 +576,7 @@ void NSModel<SortPolicy>::BuildModel(uti
       // [Q, R] = qr(randn(d, d));
       // Q = Q * diag(sign(diag(R)));
       arma::mat r;
-      if (arma::qr(q, r, arma::randn<arma::mat>(referenceSet.n_rows,
+      if (arma::qr(q, r, randn<arma::mat>(referenceSet.n_rows,
               referenceSet.n_rows)))
       {
         arma::vec rDiag(r.n_rows);
diff -pruN 4.3.0-2/src/mlpack/methods/nmf/nmf_main.cpp 4.4.0-1/src/mlpack/methods/nmf/nmf_main.cpp
--- 4.3.0-2/src/mlpack/methods/nmf/nmf_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/nmf/nmf_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -70,8 +70,6 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("@cf", "#cf");
-BINDING_SEE_ALSO("Alternating matrix factorization tutorial",
-    "@doc/tutorials/amf.md");
 BINDING_SEE_ALSO("Non-negative matrix factorization on Wikipedia",
     "https://en.wikipedia.org/wiki/Non-negative_matrix_factorization");
 BINDING_SEE_ALSO("Algorithms for non-negative matrix factorization (pdf)",
@@ -158,9 +156,9 @@ void ApplyFactorization(util::Params& pa
   if (params.Has("initial_w") && params.Has("initial_h"))
   {
     // Initialize W and H with given matrices
-    GivenInitialization ginit = GivenInitialization(initialW, initialH);
+    GivenInitialization<> ginit = GivenInitialization<>(initialW, initialH);
     AMF<SimpleResidueTermination,
-        GivenInitialization,
+        GivenInitialization<>,
         UpdateRuleType> amf(srt, ginit);
     amf.Apply(V, r, W, H);
   }
@@ -168,13 +166,13 @@ void ApplyFactorization(util::Params& pa
   {
     // Merge GivenInitialization and RandomAMFInitialization rules
     // to initialize W with the given matrix, and H with random noise
-    GivenInitialization ginit = GivenInitialization(initialW);
+    GivenInitialization<> ginit = GivenInitialization<>(initialW);
     RandomAMFInitialization rinit = RandomAMFInitialization();
-    MergeInitialization<GivenInitialization, RandomAMFInitialization> minit =
-        MergeInitialization<GivenInitialization, RandomAMFInitialization>
+    MergeInitialization<GivenInitialization<>, RandomAMFInitialization> minit =
+        MergeInitialization<GivenInitialization<>, RandomAMFInitialization>
         (ginit, rinit);
     AMF<SimpleResidueTermination,
-        MergeInitialization<GivenInitialization, RandomAMFInitialization>,
+        MergeInitialization<GivenInitialization<>, RandomAMFInitialization>,
         UpdateRuleType> amf(srt, minit);
     amf.Apply(V, r, W, H);
   }
@@ -182,13 +180,13 @@ void ApplyFactorization(util::Params& pa
   {
     // Merge GivenInitialization and RandomAMFInitialization rules
     // to initialize H with the given matrix, and W with random noise
-    GivenInitialization ginit = GivenInitialization(initialH, false);
+    GivenInitialization<> ginit = GivenInitialization<>(initialH, false);
     RandomAMFInitialization rinit = RandomAMFInitialization();
-    MergeInitialization<RandomAMFInitialization, GivenInitialization> minit =
-        MergeInitialization<RandomAMFInitialization, GivenInitialization>
+    MergeInitialization<RandomAMFInitialization, GivenInitialization<>> minit =
+        MergeInitialization<RandomAMFInitialization, GivenInitialization<>>
         (rinit, ginit);
     AMF<SimpleResidueTermination,
-        MergeInitialization<RandomAMFInitialization, GivenInitialization>,
+        MergeInitialization<RandomAMFInitialization, GivenInitialization<>>,
         UpdateRuleType> amf(srt, minit);
     amf.Apply(V, r, W, H);
   }
diff -pruN 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/exact_svd_method.hpp 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/exact_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/exact_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/exact_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,19 +37,20 @@ class ExactSVDPolicy
    * @param eigvec Matrix to put eigenvectors (loadings) into.
    * @param * (rank) Rank of the decomposition.
    */
-  void Apply(const arma::mat& data,
-             const arma::mat& centeredData,
-             arma::mat& transformedData,
-             arma::vec& eigVal,
-             arma::mat& eigvec,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& /* data */,
+             const MatType& centeredData,
+             MatType& transformedData,
+             VecType& eigVal,
+             MatType& eigvec,
              const size_t /* rank */)
   {
     // This matrix will store the right singular vectors; we do not need them.
-    arma::mat v;
+    MatType v;
 
     // Do singular value decomposition.  Use the economical singular value
     // decomposition if the columns are much larger than the rows.
-    if (data.n_rows < data.n_cols)
+    if (centeredData.n_rows < centeredData.n_cols)
     {
       // Do economical singular value decomposition and compute only the left
       // singular vectors.
@@ -63,10 +64,10 @@ class ExactSVDPolicy
     // Now we must square the singular values to get the eigenvalues.
     // In addition we must divide by the number of points, because the
     // covariance matrix is X * X' / (N - 1).
-    eigVal %= eigVal / (data.n_cols - 1);
+    eigVal %= eigVal / (centeredData.n_cols - 1);
 
     // Project the samples to the principals.
-    transformedData = arma::trans(eigvec) * centeredData;
+    transformedData = trans(eigvec) * centeredData;
   }
 };
 
diff -pruN 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/quic_svd_method.hpp 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/quic_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/quic_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/quic_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,26 +49,27 @@ class QUICSVDPolicy
    * @param eigvec Matrix to put eigenvectors (loadings) into.
    * @param * (rank) Rank of the decomposition.
    */
-  void Apply(const arma::mat& data,
-             const arma::mat& centeredData,
-             arma::mat& transformedData,
-             arma::vec& eigVal,
-             arma::mat& eigvec,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& /* data */,
+             const MatType& centeredData,
+             MatType& transformedData,
+             VecType& eigVal,
+             MatType& eigvec,
              const size_t /* rank */)
   {
     // This matrix will store the right singular vectors; we do not need them.
-    arma::mat v, sigma;
+    MatType v, sigma;
 
     // Do singular value decomposition using the QUIC-SVD algorithm.
-    QUIC_SVD quicsvd(centeredData, eigvec, v, sigma, epsilon, delta);
+    QUIC_SVD<MatType> quicsvd(centeredData, eigvec, v, sigma, epsilon, delta);
 
     // Now we must square the singular values to get the eigenvalues.
     // In addition we must divide by the number of points, because the
     // covariance matrix is X * X' / (N - 1).
-    eigVal = arma::pow(arma::diagvec(sigma), 2) / (data.n_cols - 1);
+    eigVal = pow(arma::diagvec(sigma), 2) / (centeredData.n_cols - 1);
 
     // Project the samples to the principals.
-    transformedData = arma::trans(eigvec) * centeredData;
+    transformedData = trans(eigvec) * centeredData;
   }
 
   //! Get the error tolerance fraction for calculated subspace.
diff -pruN 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/randomized_block_krylov_method.hpp 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/randomized_block_krylov_method.hpp
--- 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/randomized_block_krylov_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/randomized_block_krylov_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,15 +52,16 @@ class RandomizedBlockKrylovSVDPolicy
    * @param eigvec Matrix to put eigenvectors (loadings) into.
    * @param rank Rank of the decomposition.
    */
-  void Apply(const arma::mat& data,
-             const arma::mat& centeredData,
-             arma::mat& transformedData,
-             arma::vec& eigVal,
-             arma::mat& eigvec,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& /* data */,
+             const MatType& centeredData,
+             MatType& transformedData,
+             VecType& eigVal,
+             MatType& eigvec,
              const size_t rank)
   {
     // This matrix will store the right singular vectors; we do not need them.
-    arma::mat v;
+    MatType v;
 
     // Do singular value decomposition using the randomized block krylov SVD
     // algorithm.
@@ -70,10 +71,10 @@ class RandomizedBlockKrylovSVDPolicy
     // Now we must square the singular values to get the eigenvalues.
     // In addition we must divide by the number of points, because the
     // covariance matrix is X * X' / (N - 1).
-    eigVal %= eigVal / (data.n_cols - 1);
+    eigVal %= eigVal / (centeredData.n_cols - 1);
 
     // Project the samples to the principals.
-    transformedData = arma::trans(eigvec) * centeredData;
+    transformedData = trans(eigvec) * centeredData;
   }
 
   //! Get the number of iterations for the power method.
diff -pruN 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/randomized_svd_method.hpp 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/randomized_svd_method.hpp
--- 4.3.0-2/src/mlpack/methods/pca/decomposition_policies/randomized_svd_method.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/decomposition_policies/randomized_svd_method.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,15 +53,16 @@ class RandomizedSVDPCAPolicy
    * @param eigvec Matrix to put eigenvectors (loadings) into.
    * @param rank Rank of the decomposition.
    */
-  void Apply(const arma::mat& data,
-             const arma::mat& centeredData,
-             arma::mat& transformedData,
-             arma::vec& eigVal,
-             arma::mat& eigvec,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& data,
+             const MatType& centeredData,
+             MatType& transformedData,
+             VecType& eigVal,
+             MatType& eigvec,
              const size_t rank)
   {
     // This matrix will store the right singular vectors; we do not need them.
-    arma::mat v;
+    MatType v;
 
     // Do singular value decomposition using the randomized SVD algorithm.
     RandomizedSVD rsvd(iteratedPower, maxIterations);
@@ -70,10 +71,10 @@ class RandomizedSVDPCAPolicy
     // Now we must square the singular values to get the eigenvalues.
     // In addition we must divide by the number of points, because the
     // covariance matrix is X * X' / (N - 1).
-    eigVal %= eigVal / (data.n_cols - 1);
+    eigVal %= eigVal / (centeredData.n_cols - 1);
 
     // Project the samples to the principals.
-    transformedData = arma::trans(eigvec) * centeredData;
+    transformedData = trans(eigvec) * centeredData;
   }
 
   //! Get the size of the normalized power iterations.
diff -pruN 4.3.0-2/src/mlpack/methods/pca/pca.hpp 4.4.0-1/src/mlpack/methods/pca/pca.hpp
--- 4.3.0-2/src/mlpack/methods/pca/pca.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/pca.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,10 +52,13 @@ class PCA
    * @param eigVal Vector to put eigenvalues into.
    * @param eigvec Matrix to put eigenvectors (loadings) into.
    */
-  void Apply(const arma::mat& data,
-             arma::mat& transformedData,
-             arma::vec& eigVal,
-             arma::mat& eigvec);
+  template<typename MatType = arma::mat,
+           typename OutMatType = arma::mat,
+           typename VecType = arma::vec>
+  void Apply(const MatType& data,
+             OutMatType& transformedData,
+             VecType& eigVal,
+             OutMatType& eigvec);
 
   /**
    * Apply Principal Component Analysis to the provided data set. It is safe
@@ -65,17 +68,21 @@ class PCA
    * @param transformedData Matrix to store results of PCA in.
    * @param eigVal Vector to put eigenvalues into.
    */
-  void Apply(const arma::mat& data,
-             arma::mat& transformedData,
-             arma::vec& eigVal);
+  template<typename MatType = arma::mat,
+           typename OutMatType = arma::mat,
+           typename VecType = arma::vec>
+  void Apply(const MatType& data,
+             OutMatType& transformedData,
+             VecType& eigVal);
   /**
    * Apply Principal Component Analysis to the provided data set. It is safe
    * to pass the same matrix reference for both data and transformedData.
    * @param data Data matrix.
    * @param transformedData Matrix to store results of PCA in.
    */
-  void Apply(const arma::mat& data,
-             arma::mat& transformedData);
+  template<typename MatType = arma::mat, typename OutMatType = arma::mat>
+  void Apply(const MatType& data,
+             OutMatType& transformedData);
 
   /**
    * Use PCA for dimensionality reduction on the given dataset. This will save
@@ -88,16 +95,63 @@ class PCA
    * @param newDimension New dimension of the data.
    * @return Amount of the variance of the data retained (between 0 and 1).
    */
-  double Apply(arma::mat& data, const size_t newDimension);
+  template<typename MatType = arma::mat>
+  double Apply(MatType& data, const size_t newDimension);
 
   //! This overload is here to make sure int gets casted right to size_t.
-  inline double Apply(arma::mat& data, const int newDimension)
+  template<typename MatType = arma::mat>
+  inline double Apply(MatType& data, const int newDimension)
   {
     return Apply(data, size_t(newDimension));
   }
 
   /**
    * Use PCA for dimensionality reduction on the given dataset. This will save
+   * the newDimension largest principal components of the data and remove the
+   * rest, storing the result in `transformedData`. The return value is the
+   * amount of variance of the data that is retained; this is a value between 0
+   * and 1.  For instance, a value of 0.9 indicates that 90% of the variance
+   * present in the data was retained.
+   *
+   * @param data Data matrix.
+   * @param transformedData Output matrix to store transformed data in.
+   * @param newDimension New dimension of the data.
+   * @return Amount of the variance of the data retained (between 0 and 1).
+   */
+  template<typename MatType = arma::mat, typename OutMatType = arma::mat>
+  double Apply(const MatType& data,
+               OutMatType& transformedData,
+               const size_t newDimension);
+
+  //! This overload is here to make sure int gets casted right to size_t.
+  template<typename MatType = arma::mat, typename OutMatType = arma::mat>
+  inline double Apply(const MatType& data,
+                      OutMatType& transformedData,
+                      const int newDimension)
+  {
+    return Apply(data, transformedData, size_t(newDimension));
+  }
+
+  /**
+   * Use PCA for dimensionality reduction on the given dataset. This will save
+   * as many dimensions as necessary to retain at least the given amount of
+   * variance (specified by parameter varRetained).  The amount should be
+   * between 0 and 1; if the amount is 0, then only 1 dimension will be
+   * retained.  If the amount is 1, then all dimensions will be retained.
+   *
+   * The method returns the actual amount of variance retained, which will
+   * always be greater than or equal to the varRetained parameter.
+   *
+   * @param data Data matrix.
+   * @param varRetained Lower bound on amount of variance to retain; should be
+   *     between 0 and 1.
+   * @return Actual amount of variance retained (between 0 and 1).
+   */
+  template<typename MatType>
+  double Apply(MatType& data, const double varRetained);
+
+  /**
+   * Use PCA for dimensionality reduction on the given dataset. This will save
    * as many dimensions as necessary to retain at least the given amount of
    * variance (specified by parameter varRetained).  The amount should be
    * between 0 and 1; if the amount is 0, then only 1 dimension will be
@@ -111,7 +165,10 @@ class PCA
    *     between 0 and 1.
    * @return Actual amount of variance retained (between 0 and 1).
    */
-  double Apply(arma::mat& data, const double varRetained);
+  template<typename MatType = arma::mat, typename OutMatType = arma::mat>
+  double Apply(const MatType& data,
+               OutMatType& transformedData,
+               const double varRetained);
 
   //! Get whether or not this PCA object will scale (by standard deviation)
   //! the data when PCA is performed.
@@ -122,14 +179,15 @@ class PCA
 
  private:
   //! Scaling the data is when we reduce the variance of each dimension to 1.
-  void ScaleData(arma::mat& centeredData)
+  template<typename MatType>
+  void ScaleData(MatType& centeredData)
   {
     if (scaleData)
     {
       // Scaling the data is when we reduce the variance of each dimension
       // to 1. We do this by dividing each dimension by its standard
       // deviation.
-      arma::vec stdDev = arma::stddev(
+      arma::Col<typename MatType::elem_type> stdDev = arma::stddev(
           centeredData, 0, 1 /* for each dimension */);
 
       // If there are any zeroes, make them very small.
@@ -137,7 +195,7 @@ class PCA
         if (stdDev[i] == 0)
           stdDev[i] = 1e-50;
 
-      centeredData /= arma::repmat(stdDev, 1, centeredData.n_cols);
+      centeredData /= repmat(stdDev, 1, centeredData.n_cols);
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/pca/pca_impl.hpp 4.4.0-1/src/mlpack/methods/pca/pca_impl.hpp
--- 4.3.0-2/src/mlpack/methods/pca/pca_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/pca/pca_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -12,12 +12,10 @@
  * 3-clause BSD license along with mlpack.  If not, see
  * http://www.opensource.org/licenses/BSD-3-Clause for more information.
  */
-
 #ifndef MLPACK_METHODS_PCA_PCA_IMPL_HPP
 #define MLPACK_METHODS_PCA_PCA_IMPL_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/lin_alg.hpp>
 #include "pca.hpp"
 
 namespace mlpack {
@@ -38,20 +36,31 @@ PCA<DecompositionPolicy>::PCA(
  * @param eigvec - PCA Loadings/Coeffs/EigenVectors
  */
 template<typename DecompositionPolicy>
-void PCA<DecompositionPolicy>::Apply(const arma::mat& data,
-                                     arma::mat& transformedData,
-                                     arma::vec& eigVal,
-                                     arma::mat& eigvec)
+template<typename MatType, typename OutMatType, typename VecType>
+void PCA<DecompositionPolicy>::Apply(const MatType& data,
+                                     OutMatType& transformedData,
+                                     VecType& eigVal,
+                                     OutMatType& eigvec)
 {
+  // Sanity checks on input types.
+  static_assert(IsBaseMatType<OutMatType>::value,
+      "PCA::Apply(): transformedData must be a matrix type!");
+  static_assert(IsBaseMatType<VecType>::value,
+      "PCA::Apply(): eigVal must be a vector type!");
+  static_assert(std::is_same<typename MatType::elem_type,
+                             typename OutMatType::elem_type>::value,
+      "PCA::Apply(): data and transformedData must have the same element "
+      "types!");
+
   // Center the data into a temporary matrix.
-  arma::mat centeredData;
-  Center(data, centeredData);
+  OutMatType centeredData = arma::conv_to<OutMatType>::from(data);
+  centeredData.each_col() -= arma::mean(centeredData, 1);
 
-  // Scale the data if the user ask for.
+  // Scale the data if the user asked for it.
   ScaleData(centeredData);
 
   decomposition.Apply(data, centeredData, transformedData, eigVal, eigvec,
-      data.n_rows);
+      centeredData.n_rows);
 }
 
 /**
@@ -62,11 +71,22 @@ void PCA<DecompositionPolicy>::Apply(con
  * @param eigVal - contains eigen values in a column vector
  */
 template<typename DecompositionPolicy>
-void PCA<DecompositionPolicy>::Apply(const arma::mat& data,
-                                     arma::mat& transformedData,
-                                     arma::vec& eigVal)
+template<typename MatType, typename OutMatType, typename VecType>
+void PCA<DecompositionPolicy>::Apply(const MatType& data,
+                                     OutMatType& transformedData,
+                                     VecType& eigVal)
 {
-  arma::mat eigvec;
+  // Sanity checks on input types.
+  static_assert(IsBaseMatType<OutMatType>::value,
+      "PCA::Apply(): transformedData must be a matrix type!");
+  static_assert(IsBaseMatType<VecType>::value,
+      "PCA::Apply(): eigVal must be a vector type!");
+  static_assert(std::is_same<typename MatType::elem_type,
+                             typename OutMatType::elem_type>::value,
+      "PCA::Apply(): data and transformedData must have the same element "
+      "types!");
+
+  OutMatType eigvec;
   Apply(data, transformedData, eigVal, eigvec);
 }
 
@@ -77,11 +97,24 @@ void PCA<DecompositionPolicy>::Apply(con
  * @param transformedData Data with PCA applied.
  */
 template<typename DecompositionPolicy>
-void PCA<DecompositionPolicy>::Apply(const arma::mat& data,
-                                     arma::mat& transformedData)
+template<typename MatType, typename OutMatType>
+void PCA<DecompositionPolicy>::Apply(const MatType& data,
+                                     OutMatType& transformedData)
 {
-  arma::mat eigvec;
-  arma::vec eigVal;
+  // Sanity checks on input types.
+  static_assert(IsBaseMatType<OutMatType>::value,
+      "PCA::Apply(): transformedData must be a matrix type!");
+  static_assert(std::is_same<typename MatType::elem_type,
+                             typename OutMatType::elem_type>::value,
+      "PCA::Apply(): data and transformedData must have the same element "
+      "types!");
+
+  // It's possible a user didn't pass in a matrix but instead an expression, but
+  // we need a type that we can store.
+  typedef typename GetDenseColType<MatType>::type BaseColType;
+
+  OutMatType eigvec;
+  BaseColType eigVal;
   Apply(data, transformedData, eigVal, eigvec);
 }
 
@@ -97,33 +130,58 @@ void PCA<DecompositionPolicy>::Apply(con
  * @return Amount of the variance of the data retained (between 0 and 1).
  */
 template<typename DecompositionPolicy>
-double PCA<DecompositionPolicy>::Apply(arma::mat& data,
+template<typename MatType>
+double PCA<DecompositionPolicy>::Apply(MatType& data,
+                                       const size_t newDimension)
+{
+  return Apply(data, data, newDimension);
+}
+
+template<typename DecompositionPolicy>
+template<typename MatType, typename OutMatType>
+double PCA<DecompositionPolicy>::Apply(const MatType& data,
+                                       OutMatType& transformedData,
                                        const size_t newDimension)
 {
   // Parameter validation.
   if (newDimension == 0)
-    Log::Fatal << "PCA::Apply(): newDimension (" << newDimension << ") cannot "
-        << "be zero!" << std::endl;
-  if (newDimension > data.n_rows)
-    Log::Fatal << "PCA::Apply(): newDimension (" << newDimension << ") cannot "
-        << "be greater than the existing dimensionality of the data ("
-        << data.n_rows << ")!" << std::endl;
+  {
+    std::ostringstream oss;
+    oss << "PCA::Apply(): newDimension (" << newDimension << ") cannot be "
+        << "zero!";
+    throw std::invalid_argument(oss.str());
+  }
+
+  typedef typename GetDenseMatType<MatType>::type BaseMatType;
+  typedef typename GetDenseColType<MatType>::type BaseColType;
 
-  arma::mat eigvec;
-  arma::vec eigVal;
+  BaseMatType eigvec;
+  BaseColType eigVal;
 
   // Center the data into a temporary matrix.
-  arma::mat centeredData;
-  Center(data, centeredData);
+  BaseMatType centeredData = arma::conv_to<OutMatType>::from(data);
+  centeredData.each_col() -= arma::mean(centeredData, 1);
+
+  // This check cannot happen until here, as `data` may not have a .n_rows
+  // member if it is an expression.
+  if (newDimension > centeredData.n_rows)
+  {
+    std::ostringstream oss;
+    oss << "PCA::Apply(): newDimension (" << newDimension << ") cannot "
+        << "be greater than the existing dimensionality of the data ("
+        << data.n_rows << ")!";
+    throw std::invalid_argument(oss.str());
+  }
 
   // Scale the data if the user ask for.
   ScaleData(centeredData);
 
-  decomposition.Apply(data, centeredData, data, eigVal, eigvec, newDimension);
+  decomposition.Apply(data, centeredData, transformedData, eigVal, eigvec,
+      newDimension);
 
   if (newDimension < eigvec.n_rows)
     // Drop unnecessary rows.
-    data.shed_rows(newDimension, data.n_rows - 1);
+    transformedData.shed_rows(newDimension, data.n_rows - 1);
 
   // The svd method returns only non-zero eigenvalues so we have to calculate
   // the right dimension before calculating the amount of variance retained.
@@ -144,26 +202,48 @@ double PCA<DecompositionPolicy>::Apply(a
  * always be greater than or equal to the varRetained parameter.
  */
 template<typename DecompositionPolicy>
-double PCA<DecompositionPolicy>::Apply(arma::mat& data,
+template<typename MatType>
+double PCA<DecompositionPolicy>::Apply(MatType& data,
+                                       const double varRetained)
+{
+  return Apply(data, data, varRetained);
+}
+
+template<typename DecompositionPolicy>
+template<typename MatType, typename OutMatType>
+double PCA<DecompositionPolicy>::Apply(const MatType& data,
+                                       OutMatType& transformedData,
                                        const double varRetained)
 {
   // Parameter validation.
   if (varRetained < 0)
-    Log::Fatal << "PCA::Apply(): varRetained (" << varRetained << ") must be "
-        << "greater than or equal to 0." << std::endl;
-  if (varRetained > 1)
-    Log::Fatal << "PCA::Apply(): varRetained (" << varRetained << ") should be "
-        << "less than or equal to 1." << std::endl;
+  {
+    std::ostringstream oss;
+    oss << "PCA::Apply(): varRetained (" << varRetained << ") must be greater "
+        << "than or equal to 0.";
+    throw std::invalid_argument(oss.str());
+  }
+  else if (varRetained > 1)
+  {
+    std::ostringstream oss;
+    oss << "PCA::Apply(): varRetained (" << varRetained << ") should be less "
+        << "than or equal to 1.";
+    throw std::invalid_argument(oss.str());
+  }
+
+  typedef typename GetDenseMatType<MatType>::type BaseMatType;
+  typedef typename GetDenseColType<MatType>::type BaseColType;
 
-  arma::mat eigvec;
-  arma::vec eigVal;
+  BaseMatType eigvec;
+  BaseColType eigVal;
+  BaseMatType out;
 
-  Apply(data, data, eigVal, eigvec);
+  Apply(data, transformedData, eigVal, eigvec);
 
   // Calculate the dimension we should keep.
   size_t newDimension = 0;
   double varSum = 0.0;
-  eigVal /= arma::sum(eigVal); // Normalize eigenvalues.
+  eigVal /= sum(eigVal); // Normalize eigenvalues.
   while ((varSum < varRetained) && (newDimension < eigVal.n_elem))
   {
     varSum += eigVal[newDimension];
@@ -172,7 +252,7 @@ double PCA<DecompositionPolicy>::Apply(a
 
   // varSum is the actual variance we will retain.
   if (newDimension < eigVal.n_elem)
-    data.shed_rows(newDimension, data.n_rows - 1);
+    transformedData.shed_rows(newDimension, transformedData.n_rows - 1);
 
   return varSum;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/initialization_methods/random_init.hpp 4.4.0-1/src/mlpack/methods/perceptron/initialization_methods/random_init.hpp
--- 4.3.0-2/src/mlpack/methods/perceptron/initialization_methods/random_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/initialization_methods/random_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -25,8 +25,9 @@ class RandomPerceptronInitialization
  public:
   RandomPerceptronInitialization() { }
 
-  inline static void Initialize(arma::mat& weights,
-                                arma::vec& biases,
+  template<typename eT>
+  inline static void Initialize(arma::Mat<eT>& weights,
+                                arma::Col<eT>& biases,
                                 const size_t numFeatures,
                                 const size_t numClasses)
   {
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/initialization_methods/zero_init.hpp 4.4.0-1/src/mlpack/methods/perceptron/initialization_methods/zero_init.hpp
--- 4.3.0-2/src/mlpack/methods/perceptron/initialization_methods/zero_init.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/initialization_methods/zero_init.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,8 +24,9 @@ class ZeroInitialization
  public:
   ZeroInitialization() { }
 
-  inline static void Initialize(arma::mat& weights,
-                                arma::vec& biases,
+  template<typename eT>
+  inline static void Initialize(arma::Mat<eT>& weights,
+                                arma::Col<eT>& biases,
                                 const size_t numFeatures,
                                 const size_t numClasses)
   {
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/learning_policies/simple_weight_update.hpp 4.4.0-1/src/mlpack/methods/perceptron/learning_policies/simple_weight_update.hpp
--- 4.3.0-2/src/mlpack/methods/perceptron/learning_policies/simple_weight_update.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/learning_policies/simple_weight_update.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,13 +45,13 @@ class SimpleWeightUpdate
    * @param instanceWeight Weight to be given to this particular point during
    *      training (this is useful for boosting).
    */
-  template<typename VecType>
+  template<typename VecType, typename eT>
   void UpdateWeights(const VecType& trainingPoint,
-                     arma::mat& weights,
-                     arma::vec& biases,
+                     arma::Mat<eT>& weights,
+                     arma::Col<eT>& biases,
                      const size_t incorrectClass,
                      const size_t correctClass,
-                     const double instanceWeight = 1.0)
+                     const eT instanceWeight = 1.0)
   {
     weights.col(incorrectClass) -= instanceWeight * trainingPoint;
     biases(incorrectClass) -= instanceWeight;
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/perceptron.hpp 4.4.0-1/src/mlpack/methods/perceptron/perceptron.hpp
--- 4.3.0-2/src/mlpack/methods/perceptron/perceptron.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/perceptron.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,6 +34,9 @@ template<typename LearnPolicy = SimpleWe
 class Perceptron
 {
  public:
+  //! The element type used in the Perceptron.
+  typedef typename MatType::elem_type ElemType;
+
   /**
    * Constructor: create the perceptron with the given number of classes and
    * initialize the weight matrix, but do not perform any training.  (Call the
@@ -83,11 +86,14 @@ class Perceptron
    * @param maxIterations Maximum number of iterations for the perceptron
    *     learning algorithm.
    */
+  template<typename WeightsType>
   Perceptron(const MatType& data,
              const arma::Row<size_t>& labels,
              const size_t numClasses,
-             const arma::rowvec& instanceWeights,
-             const size_t maxIterations = 1000);
+             const WeightsType& instanceWeights,
+             const size_t maxIterations = 1000,
+             const typename std::enable_if<
+                 arma::is_arma_type<WeightsType>::value>::type* = 0);
 
   /**
    * Alternate constructor which copies parameters from an already initiated
@@ -101,17 +107,46 @@ class Perceptron
    * @param instanceWeights Weight vector to use while training. For boosting
    *      purposes.
    */
+  template<typename WeightsType>
+  mlpack_deprecated /* was previously only used by AdaBoost */
   Perceptron(const Perceptron& other,
              const MatType& data,
              const arma::Row<size_t>& labels,
              const size_t numClasses,
-             const arma::rowvec& instanceWeights);
+             const WeightsType& instanceWeights,
+             const typename std::enable_if<
+                 arma::is_arma_type<WeightsType>::value>::type* = 0);
+
+  /**
+   * Train the perceptron on the given data for up to the given maximum number
+   * of iterations.  A single iteration corresponds to a single pass through the
+   * data, so if you want to pass through the dataset only once, set
+   * `maxIterations` to 1.
+   *
+   * After calling this overload, `MaxIterations()` will return whatever
+   * `maxIterations` was given to this function.
+   *
+   * This training does not reset the model weights, so you can call Train() on
+   * multiple datasets sequentially.
+   *
+   * @param data Dataset on which training should be performed.
+   * @param labels Labels of the dataset.
+   * @param numClasses Number of classes in the data.
+   * @param maxIterations Maximum number of iterations for training.
+   */
+  void Train(const MatType& data,
+             const arma::Row<size_t>& labels,
+             const size_t numClasses,
+             const std::optional<size_t> maxIterations = std::nullopt);
 
   /**
-   * Train the perceptron on the given data for up to the maximum number of
-   * iterations (specified in the constructor or through MaxIterations()).  A
-   * single iteration corresponds to a single pass through the data, so if you
-   * want to pass through the dataset only once, set MaxIterations() to 1.
+   * Train the perceptron on the given data for up to the given maximum number
+   * of iterations.  A single iteration corresponds to a single pass through the
+   * data, so if you want to pass through the dataset only once, set
+   * `maxIterations` to 1.
+   *
+   * After calling this overload, `MaxIterations()` will return whatever
+   * `maxIterations` was given to this function.
    *
    * This training does not reset the model weights, so you can call Train() on
    * multiple datasets sequentially.
@@ -121,11 +156,22 @@ class Perceptron
    * @param numClasses Number of classes in the data.
    * @param instanceWeights Cost matrix. Stores the cost of mispredicting
    *      instances.  This is useful for boosting.
+   * @param maxIterations Maximum number of iterations for training.
    */
   void Train(const MatType& data,
              const arma::Row<size_t>& labels,
              const size_t numClasses,
-             const arma::rowvec& instanceWeights = arma::rowvec());
+             const arma::rowvec& instanceWeights,
+             const std::optional<size_t> maxIterations = std::nullopt);
+
+  /**
+   * After training, use the weights matrix to classify `point`, and return the
+   * predicted class.
+   *
+   * @param point Test point to classify.
+   */
+  template<typename VecType>
+  size_t Classify(const VecType& point) const;
 
   /**
    * Classification function. After training, use the weights matrix to
@@ -135,7 +181,13 @@ class Perceptron
    * @param predictedLabels Vector to store the predicted classes after
    *     classifying test.
    */
-  void Classify(const MatType& test, arma::Row<size_t>& predictedLabels);
+  void Classify(const MatType& test, arma::Row<size_t>& predictedLabels) const;
+
+  /**
+   * Reset the model, so that the next call to `Train()` will not be
+   * incremental.
+   */
+  void Reset();
 
   /**
    * Serialize the perceptron.
@@ -152,16 +204,28 @@ class Perceptron
   size_t NumClasses() const { return weights.n_cols; }
 
   //! Get the weight matrix.
-  const arma::mat& Weights() const { return weights; }
+  const arma::Mat<ElemType>& Weights() const { return weights; }
   //! Modify the weight matrix.  You had better know what you are doing!
-  arma::mat& Weights() { return weights; }
+  arma::Mat<ElemType>& Weights() { return weights; }
 
   //! Get the biases.
-  const arma::vec& Biases() const { return biases; }
+  const arma::Col<ElemType>& Biases() const { return biases; }
   //! Modify the biases.  You had better know what you are doing!
-  arma::vec& Biases() { return biases; }
+  arma::Col<ElemType>& Biases() { return biases; }
 
  private:
+  /**
+   * Internal training function; this assumes that maxIterations has been set.
+   *
+   * If `HasWeights` is `false`, then `instanceWeights` is ignored (and may be
+   * left empty).
+   */
+  template<bool HasWeights, typename WeightsType>
+  void TrainInternal(const MatType& data,
+                     const arma::Row<size_t>& labels,
+                     const size_t numClasses,
+                     const WeightsType& instanceWeights = WeightsType());
+
   //! The maximum number of iterations during training.
   size_t maxIterations;
 
@@ -171,10 +235,10 @@ class Perceptron
    * the weights for one dimension of the input data.  The biases are held in a
    * separate vector.
    */
-  arma::mat weights;
+  arma::Mat<ElemType> weights;
 
   //! The biases for each class.
-  arma::vec biases;
+  arma::Col<ElemType> biases;
 };
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/perceptron_impl.hpp 4.4.0-1/src/mlpack/methods/perceptron/perceptron_impl.hpp
--- 4.3.0-2/src/mlpack/methods/perceptron/perceptron_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/perceptron_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -58,7 +58,8 @@ Perceptron<LearnPolicy, WeightInitializa
     maxIterations(maxIterations)
 {
   // Start training.
-  Train(data, labels, numClasses);
+  TrainInternal<false, arma::Row<typename MatType::elem_type>>(data, labels,
+      numClasses);
 }
 
 /**
@@ -75,16 +76,19 @@ template<
     typename WeightInitializationPolicy,
     typename MatType
 >
+template<typename WeightsType>
 Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Perceptron(
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    const arma::rowvec& instanceWeights,
-    const size_t maxIterations) :
+    const WeightsType& instanceWeights,
+    const size_t maxIterations,
+    const typename std::enable_if<
+        arma::is_arma_type<WeightsType>::value>::type*) :
     maxIterations(maxIterations)
 {
   // Start training.
-  Train(data, labels, numClasses, instanceWeights);
+  TrainInternal<true>(data, labels, numClasses, instanceWeights);
 }
 
 /**
@@ -103,55 +107,75 @@ template<
     typename WeightInitializationPolicy,
     typename MatType
 >
+template<typename WeightsType>
+mlpack_deprecated
 Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Perceptron(
     const Perceptron& other,
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    const arma::rowvec& instanceWeights) :
+    const WeightsType& instanceWeights,
+    const typename std::enable_if<
+        arma::is_arma_type<WeightsType>::value>::type*) :
     maxIterations(other.maxIterations)
 {
-  Train(data, labels, numClasses, instanceWeights);
+  TrainInternal<true>(data, labels, numClasses, instanceWeights);
 }
 
 /**
- * Classification function. After training, use the weights matrix to classify
- * test, and put the predicted classes in predictedLabels.
+ * Train the perceptron on the given data for up to the given maximum number
+ * of iterations.  A single iteration corresponds to a single pass through the
+ * data, so if you want to pass through the dataset only once, set
+ * `maxIterations` to 1.
  *
- * @param test Testing data or data to classify.
- * @param predictedLabels Vector to store the predicted classes after
- *      classifying test.
+ * After calling this overload, `MaxIterations()` will return whatever
+ * `maxIterations` was given to this function.
+ *
+ * This training does not reset the model weights, so you can call Train() on
+ * multiple datasets sequentially.
+ *
+ * @param data Dataset on which training should be performed.
+ * @param labels Labels of the dataset.
+ * @param numClasses Number of classes in the data.
+ * @param maxIterations Maximum number of iterations for training.
  */
 template<
     typename LearnPolicy,
     typename WeightInitializationPolicy,
     typename MatType
 >
-void Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Classify(
-    const MatType& test,
-    arma::Row<size_t>& predictedLabels)
+void Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Train(
+    const MatType& data,
+    const arma::Row<size_t>& labels,
+    const size_t numClasses,
+    const std::optional<size_t> maxIterations)
 {
-  arma::vec tempLabelMat;
-  arma::uword maxIndex = 0;
-  predictedLabels.set_size(test.n_cols);
+  // Set the maximum number of iterations and call unweighted Train().
+  if (maxIterations.has_value())
+    this->maxIterations = maxIterations.value();
 
-  // Could probably be faster if done in batch.
-  for (size_t i = 0; i < test.n_cols; ++i)
-  {
-    tempLabelMat = weights.t() * test.col(i) + biases;
-    tempLabelMat.max(maxIndex);
-    predictedLabels(i) = maxIndex;
-  }
+  TrainInternal<false, arma::Row<typename MatType::elem_type>>(data, labels,
+      numClasses);
 }
 
 /**
- * Training function.  It trains on trainData using the cost matrix
- * instanceWeights.
+ * Train the perceptron on the given data for up to the given maximum number
+ * of iterations.  A single iteration corresponds to a single pass through the
+ * data, so if you want to pass through the dataset only once, set
+ * `maxIterations` to 1.
  *
- * @param data Data to train on.
- * @param labels Labels of data.
+ * After calling this overload, `MaxIterations()` will return whatever
+ * `maxIterations` was given to this function.
+ *
+ * This training does not reset the model weights, so you can call Train() on
+ * multiple datasets sequentially.
+ *
+ * @param data Dataset on which training should be performed.
+ * @param labels Labels of the dataset.
+ * @param numClasses Number of classes in the data.
  * @param instanceWeights Cost matrix. Stores the cost of mispredicting
  *      instances.  This is useful for boosting.
+ * @param maxIterations Maximum number of iterations for training.
  */
 template<
     typename LearnPolicy,
@@ -162,10 +186,40 @@ void Perceptron<LearnPolicy, WeightIniti
     const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
-    const arma::rowvec& instanceWeights)
+    const arma::rowvec& instanceWeights,
+    const std::optional<size_t> maxIterations)
+{
+  // Set the maximum number of iterations and call weighted training.
+  if (maxIterations.has_value())
+    this->maxIterations = maxIterations.value();
+
+  TrainInternal<true>(data, labels, numClasses, instanceWeights);
+}
+
+/**
+ * Training function.  It trains on trainData using the cost matrix
+ * instanceWeights.
+ *
+ * @param data Data to train on.
+ * @param labels Labels of data.
+ * @param instanceWeights Cost matrix. Stores the cost of mispredicting
+ *      instances.  This is useful for boosting.
+ */
+template<
+    typename LearnPolicy,
+    typename WeightInitializationPolicy,
+    typename MatType
+>
+template<bool HasWeights, typename WeightsType>
+void Perceptron<
+    LearnPolicy, WeightInitializationPolicy, MatType
+>::TrainInternal(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const WeightsType& instanceWeights)
 {
   // Do we need to resize the weights?
-  if (weights.n_elem != numClasses)
+  if (weights.n_cols != numClasses || weights.n_rows != data.n_rows)
   {
     WeightInitializationPolicy wip;
     wip.Initialize(weights, biases, data.n_rows, numClasses);
@@ -175,12 +229,10 @@ void Perceptron<LearnPolicy, WeightIniti
   bool converged = false;
   size_t tempLabel;
   arma::uword maxIndexRow = 0, maxIndexCol = 0;
-  arma::mat tempLabelMat;
+  arma::Mat<ElemType> tempLabelMat;
 
   LearnPolicy LP;
 
-  const bool hasWeights = (instanceWeights.n_elem > 0);
-
   while ((i < maxIterations) && (!converged))
   {
     // This outer loop is for each iteration, and we use the 'converged'
@@ -207,9 +259,9 @@ void Perceptron<LearnPolicy, WeightIniti
         // Send maxIndexRow for knowing which weight to update, send j to know
         // the value of the vector to update it with.  Send tempLabel to know
         // the correct class.
-        if (hasWeights)
+        if (HasWeights)
           LP.UpdateWeights(data.col(j), weights, biases, maxIndexRow, tempLabel,
-              instanceWeights(j));
+              (typename MatType::elem_type) instanceWeights(j));
         else
           LP.UpdateWeights(data.col(j), weights, biases, maxIndexRow,
               tempLabel);
@@ -218,6 +270,79 @@ void Perceptron<LearnPolicy, WeightIniti
   }
 }
 
+/**
+ * After training, use the weights matrix to classify `point`, and return the
+ * predicted class.
+ *
+ * @param point Test point to classify.
+ */
+template<
+    typename LearnPolicy,
+    typename WeightInitializationPolicy,
+    typename MatType
+>
+template<typename VecType>
+size_t Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Classify(
+    const VecType& point) const
+{
+  util::CheckSameDimensionality(point, weights.n_rows, "Perceptron::Classify()",
+      "point");
+
+  arma::Col<ElemType> tempLabelVec;
+  arma::uword maxIndex = 0;
+
+  tempLabelVec = weights.t() * point + biases;
+  tempLabelVec.max(maxIndex);
+
+  return size_t(maxIndex);
+}
+
+/**
+ * Classification function. After training, use the weights matrix to classify
+ * test, and put the predicted classes in predictedLabels.
+ *
+ * @param test Testing data or data to classify.
+ * @param predictedLabels Vector to store the predicted classes after
+ *      classifying test.
+ */
+template<
+    typename LearnPolicy,
+    typename WeightInitializationPolicy,
+    typename MatType
+>
+void Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Classify(
+    const MatType& test,
+    arma::Row<size_t>& predictedLabels) const
+{
+  util::CheckSameDimensionality(test, weights.n_rows, "Perceptron::Classify()",
+      "points");
+
+  arma::Col<ElemType> tempLabelMat;
+  arma::uword maxIndex = 0;
+  predictedLabels.set_size(test.n_cols);
+
+  // Could probably be faster if done in batch.
+  for (size_t i = 0; i < test.n_cols; ++i)
+  {
+    tempLabelMat = weights.t() * test.col(i) + biases;
+    tempLabelMat.max(maxIndex);
+    predictedLabels(i) = maxIndex;
+  }
+}
+
+/**
+ * Reset the model, so that the next call to `Train()` will not be
+ * incremental.
+ */
+template<typename LearnPolicy,
+         typename WeightInitializationPolicy,
+         typename MatType>
+void Perceptron<LearnPolicy, WeightInitializationPolicy, MatType>::Reset()
+{
+  weights.clear();
+  biases.clear();
+}
+
 //! Serialize the perceptron.
 template<typename LearnPolicy,
          typename WeightInitializationPolicy,
diff -pruN 4.3.0-2/src/mlpack/methods/perceptron/perceptron_main.cpp 4.4.0-1/src/mlpack/methods/perceptron/perceptron_main.cpp
--- 4.3.0-2/src/mlpack/methods/perceptron/perceptron_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/perceptron/perceptron_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -56,14 +56,7 @@ BINDING_LONG_DESC(
     "on the test set may be saved with the " +
     PRINT_PARAM_STRING("predictions") +
     " output parameter.  The perceptron model may be saved with the " +
-    PRINT_PARAM_STRING("output_model") + " output parameter."
-    "\n\n"
-    "Note: the following parameter is deprecated and "
-    "will be removed in mlpack 4.0.0: " + PRINT_PARAM_STRING("output") +
-    "."
-    "\n"
-    "Use " + PRINT_PARAM_STRING("predictions") + " instead of " +
-    PRINT_PARAM_STRING("output") + '.');
+    PRINT_PARAM_STRING("output_model") + " output parameter.");
 
 // Example.
 BINDING_EXAMPLE(
@@ -144,9 +137,6 @@ PARAM_MODEL_OUT(PerceptronModel, "output
 
 // Testing/classification parameters.
 PARAM_MATRIX_IN("test", "A matrix containing the test set.", "T");
-// PARAM_UROW_OUT("output") is deprecated and will be removed in
-PARAM_UROW_OUT("output", "The matrix in which the predicted labels for the"
-    " test set will be written.", "o");
 PARAM_UROW_OUT("predictions", "The matrix in which the predicted labels for the"
     " test set will be written.", "P");
 
@@ -160,9 +150,8 @@ void BINDING_FUNCTION(util::Params& para
 
   // If the user isn't going to save the output model or any predictions, we
   // should issue a warning.
-  RequireAtLeastOnePassed(params, { "output_model", "output", "predictions" },
-      false, "no output will be saved");
-  // "output" will be removed in mlpack 4.0.0.
+  RequireAtLeastOnePassed(params, { "output_model", "predictions" }, false,
+      "no output will be saved");
   ReportIgnoredParam(params, {{ "test", false }}, "predictions");
 
   // Check parameter validity.
@@ -242,7 +231,7 @@ void BINDING_FUNCTION(util::Params& para
       // Use the last row of the training data as the labels.
       Log::Info << "Using the last dimension of training set as labels."
           << endl;
-      labelsIn = arma::conv_to<Row<size_t>>::from(
+      labelsIn = ConvTo<Row<size_t>>::From(
           trainingData.row(trainingData.n_rows - 1));
       trainingData.shed_row(trainingData.n_rows - 1);
     }
@@ -320,8 +309,6 @@ void BINDING_FUNCTION(util::Params& para
     data::RevertLabels(predictedLabels, p->Map(), results);
 
     // Save the predicted labels.
-    if (params.Has("output"))
-      params.Get<arma::Row<size_t>>("output") = results;
     if (params.Has("predictions"))
       params.Get<arma::Row<size_t>>("predictions") = std::move(results);
   }
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/image_converter_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/image_converter_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/image_converter_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/image_converter_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -10,13 +10,12 @@
  * http://www.opensource.org/licenses/BSD-3-Clause for more information.
  */
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/util/io.hpp>
+#include <mlpack/core.hpp>
 
 #undef BINDING_NAME
 #define BINDING_NAME image_converter
 
 #include <mlpack/core/util/mlpack_main.hpp>
-#include <mlpack/core.hpp>
 
 using namespace mlpack;
 using namespace mlpack::util;
@@ -65,7 +64,9 @@ BINDING_EXAMPLE(
 // See also...
 BINDING_SEE_ALSO("@preprocess_binarize", "#preprocess_binarize");
 BINDING_SEE_ALSO("@preprocess_describe", "#preprocess_describe");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 
 // DEFINE PARAM
 PARAM_VECTOR_IN_REQ(string, "input", "Image filenames which have to "
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/preprocess_binarize_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/preprocess_binarize_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/preprocess_binarize_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/preprocess_binarize_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -57,7 +57,9 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("@preprocess_describe", "#preprocess_describe");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 BINDING_SEE_ALSO("@preprocess_split", "#preprocess_split");
 
 // Define parameters for data.
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/preprocess_describe_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/preprocess_describe_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/preprocess_describe_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/preprocess_describe_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,7 +64,9 @@ BINDING_EXAMPLE(
 
 // See also...
 BINDING_SEE_ALSO("@preprocess_binarize", "#preprocess_binarize");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 BINDING_SEE_ALSO("@preprocess_split", "#preprocess_split");
 
 // Define parameters for data.
@@ -92,7 +94,7 @@ double SumNthPowerDeviations(const arma:
                              const double& fMean,
                              size_t n)
 {
-  return arma::sum(arma::pow(input - fMean, static_cast<double>(n)));
+  return sum(pow(input - fMean, static_cast<double>(n)));
 }
 
 /**
@@ -195,13 +197,13 @@ void BINDING_FUNCTION(util::Params& para
   {
     arma::rowvec feature;
     if (rowMajor)
-      feature = arma::conv_to<arma::rowvec>::from(data.col(dim));
+      feature = ConvTo<arma::rowvec>::From(data.col(dim));
     else
       feature = data.row(dim);
 
     // f at the front of the variable names means "feature".
-    const double fMax = arma::max(feature);
-    const double fMin = arma::min(feature);
+    const double fMax = max(feature);
+    const double fMin = min(feature);
     const double fMean = arma::mean(feature);
     const double fStd = arma::stddev(feature, population);
 
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/preprocess_one_hot_encoding_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/preprocess_one_hot_encoding_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/preprocess_one_hot_encoding_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/preprocess_one_hot_encoding_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,7 +48,9 @@ BINDING_EXAMPLE(
 // See also...
 BINDING_SEE_ALSO("@preprocess_binarize", "#preprocess_binarize");
 BINDING_SEE_ALSO("@preprocess_describe", "#preprocess_describe");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 BINDING_SEE_ALSO("One-hot encoding on Wikipedia",
         "https://en.m.wikipedia.org/wiki/One-hot");
 
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/preprocess_scale_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/preprocess_scale_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/preprocess_scale_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/preprocess_scale_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -85,7 +85,9 @@ BINDING_EXAMPLE(
 // See also...
 BINDING_SEE_ALSO("@preprocess_binarize", "#preprocess_binarize");
 BINDING_SEE_ALSO("@preprocess_describe", "#preprocess_describe");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 
 // Define parameters for data.
 PARAM_MATRIX_IN_REQ("input", "Matrix containing data.", "i");
diff -pruN 4.3.0-2/src/mlpack/methods/preprocess/preprocess_split_main.cpp 4.4.0-1/src/mlpack/methods/preprocess/preprocess_split_main.cpp
--- 4.3.0-2/src/mlpack/methods/preprocess/preprocess_split_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/preprocess/preprocess_split_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -80,7 +80,9 @@ BINDING_EXAMPLE(
 // See also...
 BINDING_SEE_ALSO("@preprocess_binarize", "#preprocess_binarize");
 BINDING_SEE_ALSO("@preprocess_describe", "#preprocess_describe");
+#if BINDING_TYPE == BINDING_TYPE_CLI
 BINDING_SEE_ALSO("@preprocess_imputer", "#preprocess_imputer");
+#endif
 
 // Define parameters for data.
 PARAM_MATRIX_IN_REQ("input", "Matrix containing data.", "i");
diff -pruN 4.3.0-2/src/mlpack/methods/quic_svd/quic_svd.hpp 4.4.0-1/src/mlpack/methods/quic_svd/quic_svd.hpp
--- 4.3.0-2/src/mlpack/methods/quic_svd/quic_svd.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/quic_svd/quic_svd.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -51,6 +51,7 @@ namespace mlpack {
  * qSVD.Apply(data, u, v, sigma, epsilon, delta);
  * @endcode
  */
+template<typename MatType = arma::mat>
 class QUIC_SVD
 {
  public:
@@ -64,10 +65,10 @@ class QUIC_SVD
    * @param epsilon Error tolerance fraction for calculated subspace.
    * @param delta Cumulative probability for Monte Carlo error lower bound.
    */
-  QUIC_SVD(const arma::mat& dataset,
-           arma::mat& u,
-           arma::mat& v,
-           arma::mat& sigma,
+  QUIC_SVD(const MatType& dataset,
+           MatType& u,
+           MatType& v,
+           MatType& sigma,
            const double epsilon = 0.03,
            const double delta = 0.1);
 
@@ -93,10 +94,10 @@ class QUIC_SVD
    * @param epsilon Error tolerance fraction for calculated subspace.
    * @param delta Cumulative probability for Monte Carlo error lower bound.
    */
-  void Apply(const arma::mat& dataset,
-             arma::mat& u,
-             arma::mat& v,
-             arma::mat& sigma,
+  void Apply(const MatType& dataset,
+             MatType& u,
+             MatType& v,
+             MatType& sigma,
              const double epsilon = 0.03,
              const double delta = 0.1);
 
@@ -108,14 +109,14 @@ class QUIC_SVD
    * @param v Second unitary matrix.
    * @param sigma Diagonal matrix of singular values.
    */
-  void ExtractSVD(const arma::mat& dataset,
-                  arma::mat& u, 
-                  arma::mat& v, 
-                  arma::mat& sigma);
+  void ExtractSVD(const MatType& dataset,
+                  MatType& u,
+                  MatType& v,
+                  MatType& sigma);
 
  private:
   //! Subspace basis of the input dataset.
-  arma::mat basis;
+  MatType basis;
 };
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/quic_svd/quic_svd_impl.hpp 4.4.0-1/src/mlpack/methods/quic_svd/quic_svd_impl.hpp
--- 4.3.0-2/src/mlpack/methods/quic_svd/quic_svd_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/quic_svd/quic_svd_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,39 +17,42 @@
 
 namespace mlpack {
 
-inline QUIC_SVD::QUIC_SVD(
-    const arma::mat& dataset,
-    arma::mat& u,
-    arma::mat& v,
-    arma::mat& sigma,
+template<typename MatType>
+inline QUIC_SVD<MatType>::QUIC_SVD(
+    const MatType& dataset,
+    MatType& u,
+    MatType& v,
+    MatType& sigma,
     const double epsilon,
     const double delta)
 {
   Apply(dataset, u, v, sigma, epsilon, delta);
 }
 
-inline QUIC_SVD::QUIC_SVD(
-    const double epsilon,
-    const double delta)
+template<typename MatType>
+inline QUIC_SVD<MatType>::QUIC_SVD(
+    const double /* epsilon */,
+    const double /* delta */)
 {
   /* Nothing to do here */
 }
 
-inline void QUIC_SVD::Apply(
-    const arma::mat& dataset,
-    arma::mat& u,
-    arma::mat& v,
-    arma::mat& sigma,
+template<typename MatType>
+inline void QUIC_SVD<MatType>::Apply(
+    const MatType& dataset,
+    MatType& u,
+    MatType& v,
+    MatType& sigma,
     const double epsilon,
     const double delta)
 {
   // Since columns are sample in the implementation, the matrix is transposed if
   // necessary for maximum speedup.
-  CosineTree* ctree;
+  CosineTree<MatType>* ctree;
   if (dataset.n_cols > dataset.n_rows)
-    ctree = new CosineTree(dataset, epsilon, delta);
+    ctree = new CosineTree<MatType>(dataset, epsilon, delta);
   else
-    ctree = new CosineTree(dataset.t(), epsilon, delta);
+    ctree = new CosineTree<MatType>(dataset.t(), epsilon, delta);
 
   // Get subspace basis by creating the cosine tree.
   ctree->GetFinalBasis(basis);
@@ -62,37 +65,38 @@ inline void QUIC_SVD::Apply(
   ExtractSVD(dataset, u, v, sigma);
 }
 
-inline void QUIC_SVD::ExtractSVD(const arma::mat& dataset,
-                                 arma::mat& u,
-                                 arma::mat& v,
-                                 arma::mat& sigma)
+template<typename MatType>
+inline void QUIC_SVD<MatType>::ExtractSVD(const MatType& dataset,
+                                          MatType& u,
+                                          MatType& v,
+                                          MatType& sigma)
 {
   // Calculate A * V_hat, necessary for further calculations.
-  arma::mat projectedMat;
+  MatType projectedMat;
   if (dataset.n_cols > dataset.n_rows)
     projectedMat = dataset.t() * basis;
   else
     projectedMat = dataset * basis;
 
   // Calculate the squared projected matrix.
-  arma::mat projectedMatSquared = projectedMat.t() * projectedMat;
+  MatType projectedMatSquared = projectedMat.t() * projectedMat;
 
   // Calculate the SVD of the above matrix.
-  arma::mat uBar, vBar;
-  arma::vec sigmaBar;
+  MatType uBar, vBar;
+  arma::Col<typename MatType::elem_type> sigmaBar;
   arma::svd(uBar, sigmaBar, vBar, projectedMatSquared);
 
   // Calculate the approximate SVD of the original matrix, using the SVD of the
   // squared projected matrix.
   v = basis * vBar;
-  sigma = arma::sqrt(diagmat(sigmaBar));
+  sigma = sqrt(diagmat(sigmaBar));
   u = projectedMat * vBar * sigma.i();
 
   // Since columns are sampled, the unitary matrices have to be exchanged, if
   // the transposed matrix is not passed.
   if (dataset.n_cols > dataset.n_rows)
   {
-    arma::mat tempMat = u;
+    MatType tempMat = u;
     u = v;
     v = tempMat;
   }
diff -pruN 4.3.0-2/src/mlpack/methods/radical/radical_impl.hpp 4.4.0-1/src/mlpack/methods/radical/radical_impl.hpp
--- 4.3.0-2/src/mlpack/methods/radical/radical_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/radical/radical_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,8 +35,8 @@ inline Radical::Radical(
 inline void Radical::CopyAndPerturb(arma::mat& xNew,
                                     const arma::mat& x) const
 {
-  xNew = arma::repmat(x, replicates, 1) + noiseStdDev * arma::randn(replicates * x.n_rows,
-      x.n_cols);
+  xNew = repmat(x, replicates, 1) + noiseStdDev * arma::randn(
+      replicates * x.n_rows, x.n_cols);
 }
 
 
@@ -56,7 +56,7 @@ inline double Radical::Vasicek(arma::vec
   arma::uword range = z.n_elem - m;
   for (arma::uword i = 0; i < range; ++i)
   {
-    sum += log(std::max(z(i + m) - z(i), DBL_MIN));
+    sum += std::log(std::max(z(i + m) - z(i), DBL_MIN));
   }
 
   return sum;
@@ -114,7 +114,7 @@ inline void Radical::DoRadical(const arm
   // If m was not specified, initialize m as recommended in
   // (Learned-Miller and Fisher, 2003).
   if (m < 1)
-    m = floor(sqrt((double) matX.n_rows));
+    m = floor(std::sqrt((double) matX.n_rows));
 
   const size_t nDims = matX.n_cols;
   const size_t nPoints = matX.n_rows;
diff -pruN 4.3.0-2/src/mlpack/methods/radical/radical_main.cpp 4.4.0-1/src/mlpack/methods/radical/radical_main.cpp
--- 4.3.0-2/src/mlpack/methods/radical/radical_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/radical/radical_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -54,7 +54,7 @@ BINDING_EXAMPLE(
 BINDING_SEE_ALSO("Independent component analysis on Wikipedia",
     "https://en.wikipedia.org/wiki/Independent_component_analysis");
 BINDING_SEE_ALSO("ICA using spacings estimates of entropy (pdf)",
-    "http://www.jmlr.org/papers/volume4/learned-miller03a/"
+    "https://www.jmlr.org/papers/volume4/learned-miller03a/"
     "learned-miller03a.pdf");
 BINDING_SEE_ALSO("Radical C++ class documentation",
     "@src/mlpack/methods/radical/radical.hpp");
diff -pruN 4.3.0-2/src/mlpack/methods/random_forest/bootstrap.hpp 4.4.0-1/src/mlpack/methods/random_forest/bootstrap.hpp
--- 4.3.0-2/src/mlpack/methods/random_forest/bootstrap.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/random_forest/bootstrap.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,7 +35,7 @@ void Bootstrap(const MatType& dataset,
     bootstrapWeights.set_size(weights.n_elem);
 
   // Random sampling with replacement.
-  arma::uvec indices = arma::randi<arma::uvec>(dataset.n_cols,
+  arma::uvec indices = randi<arma::uvec>(dataset.n_cols,
       arma::distr_param(0, dataset.n_cols - 1));
   bootstrapDataset = dataset.cols(indices);
   bootstrapLabels = labels.cols(indices);
diff -pruN 4.3.0-2/src/mlpack/methods/random_forest/random_forest_main.cpp 4.4.0-1/src/mlpack/methods/random_forest/random_forest_main.cpp
--- 4.3.0-2/src/mlpack/methods/random_forest/random_forest_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/random_forest/random_forest_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -40,8 +40,8 @@ BINDING_LONG_DESC(
     "\n\n"
     "The training set and associated labels are specified with the " +
     PRINT_PARAM_STRING("training") + " and " + PRINT_PARAM_STRING("labels") +
-    " parameters, respectively.  The labels should be in the range [0, "
-    "num_classes - 1]. Optionally, if " +
+    " parameters, respectively.  The labels should be in the range `[0, "
+    "num_classes - 1]`. Optionally, if " +
     PRINT_PARAM_STRING("labels") + " is not specified, the labels are assumed "
     "to be the last dimension of the training dataset."
     "\n\n"
@@ -99,11 +99,11 @@ BINDING_SEE_ALSO("@decision_tree", "#dec
 BINDING_SEE_ALSO("@hoeffding_tree", "#hoeffding_tree");
 BINDING_SEE_ALSO("@softmax_regression", "#softmax_regression");
 BINDING_SEE_ALSO("Random forest on Wikipedia",
-        "https://en.wikipedia.org/wiki/Random_forest");
-BINDING_SEE_ALSO("Random forests (pdf)",
-        "https://link.springer.com/content/pdf/10.1023/A:1010933404324.pdf");
+    "https://en.wikipedia.org/wiki/Random_forest");
+BINDING_SEE_ALSO("Random forests (pdf)", "https://www.eecis.udel.edu/~shatkay"
+    "/Course/papers/BreimanRandomForests2001.pdf");
 BINDING_SEE_ALSO("RandomForest C++ class documentation",
-        "@src/mlpack/methods/random_forest/random_forest.cpp");
+    "@src/mlpack/methods/random_forest/random_forest.hpp");
 
 PARAM_MATRIX_IN("training", "Training dataset.", "t");
 PARAM_UROW_IN("labels", "Labels for training dataset.", "l");
@@ -252,7 +252,7 @@ void BINDING_FUNCTION(util::Params& para
     Log::Info << "Training random forest with " << numTrees << " trees..."
         << endl;
 
-    const size_t numClasses = arma::max(labels) + 1;
+    const size_t numClasses = max(labels) + 1;
 
     // Train the model.
     rfModel->rf.Train(data, labels, numClasses, numTrees, minimumLeafSize,
@@ -267,7 +267,7 @@ void BINDING_FUNCTION(util::Params& para
       arma::Row<size_t> predictions;
       rfModel->rf.Classify(data, predictions);
 
-      const size_t correct = arma::accu(predictions == labels);
+      const size_t correct = accu(predictions == labels);
 
       Log::Info << correct << " of " << labels.n_elem << " correct on training"
           << " set (" << (double(correct) / double(labels.n_elem) * 100) << ")."
@@ -292,7 +292,7 @@ void BINDING_FUNCTION(util::Params& para
       arma::Row<size_t> testLabels =
           std::move(params.Get<arma::Row<size_t>>("test_labels"));
 
-      const size_t correct = arma::accu(predictions == testLabels);
+      const size_t correct = accu(predictions == testLabels);
 
       Log::Info << correct << " of " << testLabels.n_elem << " correct on test"
           << " set (" << (double(correct) / double(testLabels.n_elem) * 100)
diff -pruN 4.3.0-2/src/mlpack/methods/randomized_svd/randomized_svd.hpp 4.4.0-1/src/mlpack/methods/randomized_svd/randomized_svd.hpp
--- 4.3.0-2/src/mlpack/methods/randomized_svd/randomized_svd.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/randomized_svd/randomized_svd.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -80,10 +80,11 @@ class RandomizedSVD
    * @param eps The eps coefficient to avoid division by zero (numerical
    *        stability).
    */
-  RandomizedSVD(const arma::mat& data,
-                arma::mat& u,
-                arma::vec& s,
-                arma::mat& v,
+  template<typename InMatType, typename MatType, typename VecType>
+  RandomizedSVD(const InMatType& data,
+                MatType& u,
+                VecType& s,
+                MatType& v,
                 const size_t iteratedPower = 0,
                 const size_t maxIterations = 2,
                 const size_t rank = 0,
@@ -113,10 +114,11 @@ class RandomizedSVD
    * @param s Diagonal "Sigma" matrix of singular values.
    * @param rank Rank of the approximation.
    */
-  void Apply(const arma::sp_mat& data,
-             arma::mat& u,
-             arma::vec& s,
-             arma::mat& v,
+  template<typename eT, typename MatType, typename VecType>
+  void Apply(const arma::SpMat<eT>& data,
+             MatType& u,
+             VecType& s,
+             MatType& v,
              const size_t rank);
 
 /**
@@ -129,10 +131,11 @@ class RandomizedSVD
    * @param s Diagonal "Sigma" matrix of singular values.
    * @param rank Rank of the approximation.
    */
-  void Apply(const arma::mat& data,
-             arma::mat& u,
-             arma::vec& s,
-             arma::mat& v,
+  template<typename InMatType, typename MatType, typename VecType>
+  void Apply(const InMatType& data,
+             MatType& u,
+             VecType& s,
+             MatType& v,
              const size_t rank);
 
   /**
@@ -146,13 +149,16 @@ class RandomizedSVD
    * @param rank Rank of the approximation.
    * @param rowMean Centered mean value matrix.
    */
-  template<typename MatType>
-  void Apply(const MatType& data,
-             arma::mat& u,
-             arma::vec& s,
-             arma::mat& v,
+  template<typename InMatType,
+           typename MatType,
+           typename VecType,
+           typename MeanType>
+  void Apply(const InMatType& data,
+             MatType& u,
+             VecType& s,
+             MatType& v,
              const size_t rank,
-             MatType rowMean);
+             const MeanType& rowMean);
 
   //! Get the size of the normalized power iterations.
   size_t IteratedPower() const { return iteratedPower; }
diff -pruN 4.3.0-2/src/mlpack/methods/randomized_svd/randomized_svd_impl.hpp 4.4.0-1/src/mlpack/methods/randomized_svd/randomized_svd_impl.hpp
--- 4.3.0-2/src/mlpack/methods/randomized_svd/randomized_svd_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/randomized_svd/randomized_svd_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,11 +17,12 @@
 
 namespace mlpack {
 
+template<typename InMatType, typename MatType, typename VecType>
 inline RandomizedSVD::RandomizedSVD(
-    const arma::mat& data,
-    arma::mat& u,
-    arma::vec& s,
-    arma::mat& v,
+    const InMatType& data,
+    MatType& u,
+    VecType& s,
+    MatType& v,
     const size_t iteratedPower,
     const size_t maxIterations,
     const size_t rank,
@@ -51,61 +52,68 @@ inline RandomizedSVD::RandomizedSVD(
   /* Nothing to do here */
 }
 
-inline void RandomizedSVD::Apply(const arma::sp_mat& data,
-                                 arma::mat& u,
-                                 arma::vec& s,
-                                 arma::mat& v,
+template<typename eT, typename MatType, typename VecType>
+inline void RandomizedSVD::Apply(const arma::SpMat<eT>& data,
+                                 MatType& u,
+                                 VecType& s,
+                                 MatType& v,
                                  const size_t rank)
 {
   // Center the data into a temporary matrix for sparse matrix.
-  arma::sp_mat rowMean = arma::sum(data, 1) / data.n_cols;
+  arma::SpMat<eT> rowMean = sum(data, 1) / data.n_cols;
 
   Apply(data, u, s, v, rank, rowMean);
 }
 
-inline void RandomizedSVD::Apply(const arma::mat& data,
-                                 arma::mat& u,
-                                 arma::vec& s,
-                                 arma::mat& v,
+template<typename InMatType, typename MatType, typename VecType>
+inline void RandomizedSVD::Apply(const InMatType& dataIn,
+                                 MatType& u,
+                                 VecType& s,
+                                 MatType& v,
                                  const size_t rank)
 {
   // Center the data into a temporary matrix.
-  arma::mat rowMean = arma::sum(data, 1) / data.n_cols + eps;
+  MatType data;
+  UnwrapAlias(data, dataIn);
+
+  MatType rowMean = sum(data, 1) / data.n_cols + eps;
 
   Apply(data, u, s, v, rank, rowMean);
 }
 
-template<typename MatType>
-inline void RandomizedSVD::Apply(const MatType& data,
-                                 arma::mat& u,
-                                 arma::vec& s,
-                                 arma::mat& v,
+template<typename InMatType,
+         typename MatType,
+         typename VecType,
+         typename MeanType>
+inline void RandomizedSVD::Apply(const InMatType& data,
+                                 MatType& u,
+                                 VecType& s,
+                                 MatType& v,
                                  const size_t rank,
-                                 MatType rowMean)
+                                 const MeanType& rowMean)
 {
   if (iteratedPower == 0)
       iteratedPower = rank + 2;
 
-  arma::mat R, Q, Qdata;
+  MatType R, Q, Qdata;
 
   // Apply the centered data matrix to a random matrix, obtaining Q.
   if (data.n_cols >= data.n_rows)
   {
-    R = arma::randn<arma::mat>(data.n_rows, iteratedPower);
-    Q = (data.t() * R) - arma::repmat(arma::trans(R.t() * rowMean),
-        data.n_cols, 1);
+    R.randn(data.n_rows, iteratedPower);
+    Q = (data.t() * R) - repmat(trans(R.t() * rowMean), data.n_cols, 1);
   }
   else
   {
-    R = arma::randn<arma::mat>(data.n_cols, iteratedPower);
-    Q = (data * R) - (rowMean * (arma::ones(1, data.n_cols) * R));
+    R.randn(data.n_cols, iteratedPower);
+    Q = (data * R) - (rowMean * (ones<MatType>(1, data.n_cols) * R));
   }
 
   // Form a matrix Q whose columns constitute a
   // well-conditioned basis for the columns of the earlier Q.
   if (maxIterations == 0)
   {
-  arma::qr_econ(Q, v, Q);
+    arma::qr_econ(Q, v, Q);
   }
   else
   {
@@ -117,15 +125,15 @@ inline void RandomizedSVD::Apply(const M
   {
     if (data.n_cols >= data.n_rows)
     {
-      Q = (data * Q) - rowMean * (arma::ones(1, data.n_cols) * Q);
+      Q = (data * Q) - rowMean * (ones<MatType>(1, data.n_cols) * Q);
       arma::lu(Q, v, Q);
-      Q = (data.t() * Q) - arma::repmat(rowMean.t() * Q, data.n_cols, 1);
+      Q = (data.t() * Q) - repmat(rowMean.t() * Q, data.n_cols, 1);
     }
     else
     {
-      Q = (data.t() * Q) - arma::repmat(rowMean.t() * Q, data.n_cols, 1);
+      Q = (data.t() * Q) - repmat(rowMean.t() * Q, data.n_cols, 1);
       arma::lu(Q, v, Q);
-      Q = (data * Q) - (rowMean * (arma::ones(1, data.n_cols) * Q));
+      Q = (data * Q) - (rowMean * (ones<MatType>(1, data.n_cols) * Q));
     }
 
     // Computing the LU decomposition is more efficient than computing the QR
@@ -147,13 +155,13 @@ inline void RandomizedSVD::Apply(const M
   // applied to Q.
   if (data.n_cols >= data.n_rows)
   {
-    Qdata = (data * Q) - rowMean * (arma::ones(1, data.n_cols) * Q);
+    Qdata = (data * Q) - rowMean * (ones<MatType>(1, data.n_cols) * Q);
     arma::svd_econ(u, s, v, Qdata);
     v = Q * v;
   }
   else
   {
-    Qdata = (Q.t() * data) - arma::repmat(Q.t() * rowMean, 1,  data.n_cols);
+    Qdata = (Q.t() * data) - repmat(Q.t() * rowMean, 1,  data.n_cols);
     arma::svd_econ(u, s, v, Qdata);
     u = Q * u;
   }
diff -pruN 4.3.0-2/src/mlpack/methods/range_search/range_search.hpp 4.4.0-1/src/mlpack/methods/range_search/range_search.hpp
--- 4.3.0-2/src/mlpack/methods/range_search/range_search.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/range_search/range_search.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -46,6 +46,10 @@ class RangeSearch
  public:
   //! Convenience typedef.
   typedef TreeType<MetricType, RangeSearchStat, MatType> Tree;
+   //! The type of Matrix.
+  typedef MatType Mat;
+  //! The type of element held in MatType.
+  typedef typename MatType::elem_type ElemType;
 
   /**
    * Initialize the RangeSearch object with a given reference dataset (this is
@@ -189,9 +193,9 @@ class RangeSearch
    *      point which fell into the given range, for each query point.
    */
   void Search(const MatType& querySet,
-              const Range& range,
+              const RangeType<ElemType>& range,
               std::vector<std::vector<size_t>>& neighbors,
-              std::vector<std::vector<double>>& distances);
+              std::vector<std::vector<ElemType>>& distances);
 
   /**
    * Given a pre-built query tree, search for all reference points in the given
@@ -230,9 +234,9 @@ class RangeSearch
    *      point which fell into the given range, for each query point.
    */
   void Search(Tree* queryTree,
-              const Range& range,
+              const RangeType<ElemType>& range,
               std::vector<std::vector<size_t>>& neighbors,
-              std::vector<std::vector<double>>& distances);
+              std::vector<std::vector<ElemType>>& distances);
 
   /**
    * Search for all points in the given range for each point in the reference
@@ -263,9 +267,9 @@ class RangeSearch
    * @param distances Object which will hold the list of distances for each
    *      point which fell into the given range, for each query point.
    */
-  void Search(const Range& range,
+  void Search(const RangeType<ElemType>& range,
               std::vector<std::vector<size_t>>& neighbors,
-              std::vector<std::vector<double>>& distances);
+              std::vector<std::vector<ElemType>>& distances);
 
   //! Get whether single-tree search is being used.
   bool SingleMode() const { return singleMode; }
diff -pruN 4.3.0-2/src/mlpack/methods/range_search/range_search_impl.hpp 4.4.0-1/src/mlpack/methods/range_search/range_search_impl.hpp
--- 4.3.0-2/src/mlpack/methods/range_search/range_search_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/range_search/range_search_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -86,7 +86,7 @@ RangeSearch<MetricType, MatType, TreeTyp
   // Build the tree on the empty dataset, if necessary.
   if (!naive)
   {
-    referenceTree = BuildTree<Tree>(std::move(arma::mat()),
+    referenceTree = BuildTree<Tree>(std::move(MatType()),
         oldFromNewReferences);
     referenceSet = &referenceTree->Dataset();
     treeOwner = true;
@@ -132,7 +132,7 @@ RangeSearch<MetricType, MatType, TreeTyp
 {
   // Clear other object.
   other.referenceTree =
-      BuildTree<Tree>(std::move(arma::mat()), other.oldFromNewReferences);
+      BuildTree<Tree>(std::move(MatType()), other.oldFromNewReferences);
   other.referenceSet = &other.referenceTree->Dataset();
   other.treeOwner = true;
   other.naive = false;
@@ -286,9 +286,9 @@ template<typename MetricType,
                   typename TreeMatType> class TreeType>
 void RangeSearch<MetricType, MatType, TreeType>::Search(
     const MatType& querySet,
-    const Range& range,
+    const RangeType<ElemType>& range,
     std::vector<std::vector<size_t>>& neighbors,
-    std::vector<std::vector<double>>& distances)
+    std::vector<std::vector<ElemType>>& distances)
 {
   util::CheckSameDimensionality(querySet, *referenceSet,
       "RangeSearch::Search()", "query set");
@@ -305,7 +305,7 @@ void RangeSearch<MetricType, MatType, Tr
   // To avoid extra copies, we will store the unmapped neighbors and distances
   // in a separate object.
   std::vector<std::vector<size_t>>* neighborPtr = &neighbors;
-  std::vector<std::vector<double>>* distancePtr = &distances;
+  std::vector<std::vector<ElemType>>* distancePtr = &distances;
 
   // Mapping is only necessary if the tree rearranges points.
   if (TreeTraits<Tree>::RearrangesDataset)
@@ -314,7 +314,7 @@ void RangeSearch<MetricType, MatType, Tr
     // ourselves.
     if (!singleMode && !naive)
     {
-      distancePtr = new std::vector<std::vector<double>>;
+      distancePtr = new std::vector<std::vector<ElemType>>;
       neighborPtr = new std::vector<std::vector<size_t>>;
     }
 
@@ -456,9 +456,9 @@ template<typename MetricType,
                   typename TreeMatType> class TreeType>
 void RangeSearch<MetricType, MatType, TreeType>::Search(
     Tree* queryTree,
-    const Range& range,
+    const RangeType<ElemType>& range,
     std::vector<std::vector<size_t>>& neighbors,
-    std::vector<std::vector<double>>& distances)
+    std::vector<std::vector<ElemType>>& distances)
 {
   // If there are no points, there is no search to be done.
   if (referenceSet->n_cols == 0)
@@ -522,9 +522,9 @@ template<typename MetricType,
                   typename TreeStatType,
                   typename TreeMatType> class TreeType>
 void RangeSearch<MetricType, MatType, TreeType>::Search(
-    const Range& range,
+    const RangeType<ElemType>& range,
     std::vector<std::vector<size_t>>& neighbors,
-    std::vector<std::vector<double>>& distances)
+    std::vector<std::vector<ElemType>>& distances)
 {
   // If there are no points, there is no search to be done.
   if (referenceSet->n_cols == 0)
@@ -532,12 +532,12 @@ void RangeSearch<MetricType, MatType, Tr
 
   // Here, we will use the query set as the reference set.
   std::vector<std::vector<size_t>>* neighborPtr = &neighbors;
-  std::vector<std::vector<double>>* distancePtr = &distances;
+  std::vector<std::vector<ElemType>>* distancePtr = &distances;
 
   if (TreeTraits<Tree>::RearrangesDataset && treeOwner)
   {
     // We will always need to rearrange in this case.
-    distancePtr = new std::vector<std::vector<double>>;
+    distancePtr = new std::vector<std::vector<ElemType>>;
     neighborPtr = new std::vector<std::vector<size_t>>;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/range_search/range_search_main.cpp 4.4.0-1/src/mlpack/methods/range_search/range_search_main.cpp
--- 4.3.0-2/src/mlpack/methods/range_search/range_search_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/range_search/range_search_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -50,8 +50,8 @@ BINDING_LONG_DESC(
 
 // Example.
 BINDING_EXAMPLE(
-    "For example, the following will calculate the points within the range [2, "
-    "5] of each point in "+ PRINT_DATASET("input") + " and store the"
+    "For example, the following will calculate the points within the range `[2,"
+    " 5]` of each point in "+ PRINT_DATASET("input") + " and store the"
     " distances in" + PRINT_DATASET("distances") + " and the neighbors in "
     + PRINT_DATASET("neighbors") +
     "\n\n" +
diff -pruN 4.3.0-2/src/mlpack/methods/range_search/range_search_rules.hpp 4.4.0-1/src/mlpack/methods/range_search/range_search_rules.hpp
--- 4.3.0-2/src/mlpack/methods/range_search/range_search_rules.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/range_search/range_search_rules.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -27,6 +27,11 @@ template<typename MetricType, typename T
 class RangeSearchRules
 {
  public:
+  //! Easy access to MatType.
+  typedef typename TreeType::Mat MatType;
+  //! The type of element held in MatType.
+  typedef typename MatType::elem_type ElemType;
+
   /**
    * Construct the RangeSearchRules object.  This is usually done from within
    * the RangeSearch class at search time.
@@ -40,11 +45,11 @@ class RangeSearchRules
    * @param sameSet If true, the query and reference set are taken to be the
    *      same, and a query point will not return itself in the results.
    */
-  RangeSearchRules(const arma::mat& referenceSet,
-                   const arma::mat& querySet,
-                   const Range& range,
+  RangeSearchRules(const MatType& referenceSet,
+                   const MatType& querySet,
+                   const RangeType<ElemType>& range,
                    std::vector<std::vector<size_t> >& neighbors,
-                   std::vector<std::vector<double> >& distances,
+                   std::vector<std::vector<ElemType> >& distances,
                    MetricType& metric,
                    const bool sameSet = false);
 
@@ -54,7 +59,7 @@ class RangeSearchRules
    * @param queryIndex Index of query point.
    * @param referenceIndex Index of reference point.
    */
-  double BaseCase(const size_t queryIndex, const size_t referenceIndex);
+  ElemType BaseCase(const size_t queryIndex, const size_t referenceIndex);
 
   /**
    * Get the score for recursion order.  A low score indicates priority for
@@ -64,7 +69,7 @@ class RangeSearchRules
    * @param queryIndex Index of query point.
    * @param referenceNode Candidate node to be recursed into.
    */
-  double Score(const size_t queryIndex, TreeType& referenceNode);
+  ElemType Score(const size_t queryIndex, TreeType& referenceNode);
 
   /**
    * Re-evaluate the score for recursion order.  A low score indicates priority
@@ -77,9 +82,9 @@ class RangeSearchRules
    * @param referenceNode Candidate node to be recursed into.
    * @param oldScore Old score produced by Score() (or Rescore()).
    */
-  double Rescore(const size_t queryIndex,
-                 TreeType& referenceNode,
-                 const double oldScore) const;
+  ElemType Rescore(const size_t queryIndex,
+                   TreeType& referenceNode,
+                   const ElemType oldScore) const;
 
   /**
    * Get the score for recursion order.  A low score indicates priority for
@@ -89,7 +94,7 @@ class RangeSearchRules
    * @param queryNode Candidate query node to recurse into.
    * @param referenceNode Candidate reference node to recurse into.
    */
-  double Score(TreeType& queryNode, TreeType& referenceNode);
+  ElemType Score(TreeType& queryNode, TreeType& referenceNode);
 
   /**
    * Re-evaluate the score for recursion order.  A low score indicates priority
@@ -102,9 +107,9 @@ class RangeSearchRules
    * @param referenceNode Candidate reference node to recurse into.
    * @param oldScore Old score produced by Score() (or Rescore()).
    */
-  double Rescore(TreeType& queryNode,
-                 TreeType& referenceNode,
-                 const double oldScore) const;
+  ElemType Rescore(TreeType& queryNode,
+                   TreeType& referenceNode,
+                   const ElemType oldScore) const;
 
   typedef typename mlpack::TraversalInfo<TreeType> TraversalInfoType;
 
@@ -122,19 +127,19 @@ class RangeSearchRules
 
  private:
   //! The reference set.
-  const arma::mat& referenceSet;
+  const MatType& referenceSet;
 
   //! The query set.
-  const arma::mat& querySet;
+  const MatType& querySet;
 
   //! The range of distances for which we are searching.
-  const Range& range;
+  const RangeType<ElemType>& range;
 
   //! The vector the resultant neighbor indices should be stored in.
   std::vector<std::vector<size_t> >& neighbors;
 
   //! The vector the resultant neighbor distances should be stored in.
-  std::vector<std::vector<double> >& distances;
+  std::vector<std::vector<ElemType> >& distances;
 
   //! The instantiated metric.
   MetricType& metric;
diff -pruN 4.3.0-2/src/mlpack/methods/range_search/range_search_rules_impl.hpp 4.4.0-1/src/mlpack/methods/range_search/range_search_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/range_search/range_search_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/range_search/range_search_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -19,11 +19,11 @@ namespace mlpack {
 
 template<typename MetricType, typename TreeType>
 RangeSearchRules<MetricType, TreeType>::RangeSearchRules(
-    const arma::mat& referenceSet,
-    const arma::mat& querySet,
-    const Range& range,
+    const MatType& referenceSet,
+    const MatType& querySet,
+    const RangeType<ElemType>& range,
     std::vector<std::vector<size_t> >& neighbors,
-    std::vector<std::vector<double> >& distances,
+    std::vector<std::vector<ElemType> >& distances,
     MetricType& metric,
     const bool sameSet) :
     referenceSet(referenceSet),
@@ -45,7 +45,8 @@ RangeSearchRules<MetricType, TreeType>::
 //! results if necessary.
 template<typename MetricType, typename TreeType>
 inline mlpack_force_inline
-double RangeSearchRules<MetricType, TreeType>::BaseCase(
+typename RangeSearchRules<MetricType, TreeType>::ElemType
+RangeSearchRules<MetricType, TreeType>::BaseCase(
     const size_t queryIndex,
     const size_t referenceIndex)
 {
@@ -57,7 +58,7 @@ double RangeSearchRules<MetricType, Tree
   if ((lastQueryIndex == queryIndex) && (lastReferenceIndex == referenceIndex))
     return 0.0; // No value to return... this shouldn't do anything bad.
 
-  const double distance = metric.Evaluate(querySet.unsafe_col(queryIndex),
+  const ElemType distance = metric.Evaluate(querySet.unsafe_col(queryIndex),
       referenceSet.unsafe_col(referenceIndex));
   ++baseCases;
 
@@ -76,18 +77,19 @@ double RangeSearchRules<MetricType, Tree
 
 //! Single-tree scoring function.
 template<typename MetricType, typename TreeType>
-double RangeSearchRules<MetricType, TreeType>::Score(const size_t queryIndex,
-                                                     TreeType& referenceNode)
+typename RangeSearchRules<MetricType, TreeType>::ElemType
+RangeSearchRules<MetricType, TreeType>::Score(const size_t queryIndex,
+                                              TreeType& referenceNode)
 {
   // We must get the minimum and maximum distances and store them in this
   // object.
-  Range distances;
+  RangeType<ElemType> distances;
 
   if (TreeTraits<TreeType>::FirstPointIsCentroid)
   {
     // In this situation, we calculate the base case.  So we should check to be
     // sure we haven't already done that.
-    double baseCase;
+    ElemType baseCase;
     if (TreeTraits<TreeType>::HasSelfChildren &&
         (referenceNode.Parent() != NULL) &&
         (referenceNode.Point(0) == referenceNode.Parent()->Point(0)))
@@ -136,10 +138,11 @@ double RangeSearchRules<MetricType, Tree
 
 //! Single-tree rescoring function.
 template<typename MetricType, typename TreeType>
-double RangeSearchRules<MetricType, TreeType>::Rescore(
+typename RangeSearchRules<MetricType, TreeType>::ElemType
+RangeSearchRules<MetricType, TreeType>::Rescore(
     const size_t /* queryIndex */,
     TreeType& /* referenceNode */,
-    const double oldScore) const
+    const ElemType oldScore) const
 {
   // If it wasn't pruned before, it isn't pruned now.
   return oldScore;
@@ -147,14 +150,15 @@ double RangeSearchRules<MetricType, Tree
 
 //! Dual-tree scoring function.
 template<typename MetricType, typename TreeType>
-double RangeSearchRules<MetricType, TreeType>::Score(TreeType& queryNode,
-                                                     TreeType& referenceNode)
+typename RangeSearchRules<MetricType, TreeType>::ElemType
+RangeSearchRules<MetricType, TreeType>::Score(TreeType& queryNode,
+                                              TreeType& referenceNode)
 {
-  Range distances;
+  RangeType<ElemType> distances;
   if (TreeTraits<TreeType>::FirstPointIsCentroid)
   {
     // It is possible that the base case has already been calculated.
-    double baseCase = 0.0;
+    ElemType baseCase = 0.0;
     if ((traversalInfo.LastQueryNode() != NULL) &&
         (traversalInfo.LastReferenceNode() != NULL) &&
         (traversalInfo.LastQueryNode()->Point(0) == queryNode.Point(0)) &&
@@ -209,10 +213,11 @@ double RangeSearchRules<MetricType, Tree
 
 //! Dual-tree rescoring function.
 template<typename MetricType, typename TreeType>
-double RangeSearchRules<MetricType, TreeType>::Rescore(
+typename RangeSearchRules<MetricType, TreeType>::ElemType
+RangeSearchRules<MetricType, TreeType>::Rescore(
     TreeType& /* queryNode */,
     TreeType& /* referenceNode */,
-    const double oldScore) const
+    const ElemType oldScore) const
 {
   // If it wasn't pruned before, it isn't pruned now.
   return oldScore;
@@ -250,7 +255,7 @@ void RangeSearchRules<MetricType, TreeTy
         (queryIndex == referenceNode.Descendant(i)))
       continue;
 
-    const double distance = metric.Evaluate(querySet.unsafe_col(queryIndex),
+    const ElemType distance = metric.Evaluate(querySet.unsafe_col(queryIndex),
         referenceNode.Dataset().unsafe_col(referenceNode.Descendant(i)));
 
     neighbors[queryIndex].push_back(referenceNode.Descendant(i));
diff -pruN 4.3.0-2/src/mlpack/methods/rann/ra_search_impl.hpp 4.4.0-1/src/mlpack/methods/rann/ra_search_impl.hpp
--- 4.3.0-2/src/mlpack/methods/rann/ra_search_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/rann/ra_search_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -271,9 +271,8 @@ Search(const MatType& querySet,
     // from the reference set without replacement.
     const size_t numSamples = RAUtil::MinimumSamplesReqd(referenceSet->n_cols,
         k, tau, alpha);
-    arma::uvec distinctSamples;
-    ObtainDistinctSamples(0, referenceSet->n_cols, numSamples,
-        distinctSamples);
+    arma::uvec distinctSamples = arma::randperm(referenceSet->n_cols,
+        numSamples);
 
     // Run the base case on each combination of query point and sampled
     // reference point.
@@ -487,9 +486,8 @@ void RASearch<SortPolicy, MetricType, Ma
     // from the reference set without replacement.
     const size_t numSamples = RAUtil::MinimumSamplesReqd(referenceSet->n_cols,
         k, tau, alpha);
-    arma::uvec distinctSamples;
-    ObtainDistinctSamples(0, referenceSet->n_cols, numSamples,
-        distinctSamples);
+    arma::uvec distinctSamples = arma::randperm(referenceSet->n_cols,
+        numSamples);
 
     // The naive brute-force solution.
     for (size_t i = 0; i < referenceSet->n_cols; ++i)
diff -pruN 4.3.0-2/src/mlpack/methods/rann/ra_search_rules.hpp 4.4.0-1/src/mlpack/methods/rann/ra_search_rules.hpp
--- 4.3.0-2/src/mlpack/methods/rann/ra_search_rules.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/rann/ra_search_rules.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -232,7 +232,7 @@ class RASearchRules
     if (numSamplesMade.n_elem == 0)
       return 0;
     else
-      return arma::sum(numSamplesMade);
+      return sum(numSamplesMade);
   }
 
   typedef typename mlpack::TraversalInfo<TreeType> TraversalInfoType;
diff -pruN 4.3.0-2/src/mlpack/methods/rann/ra_search_rules_impl.hpp 4.4.0-1/src/mlpack/methods/rann/ra_search_rules_impl.hpp
--- 4.3.0-2/src/mlpack/methods/rann/ra_search_rules_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/rann/ra_search_rules_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,7 +60,7 @@ RASearchRules(const arma::mat& reference
   numSamplesReqd = RAUtil::MinimumSamplesReqd(n, k, tau, alpha);
 
   // Initialize some statistics to be collected during the search.
-  numSamplesMade = arma::zeros<arma::Col<size_t> >(querySet.n_cols);
+  numSamplesMade = zeros<arma::Col<size_t>>(querySet.n_cols);
   numDistComputations = 0;
   samplingRatio = (double) numSamplesReqd / (double) n;
 
@@ -87,7 +87,7 @@ RASearchRules(const arma::mat& reference
     arma::uvec distinctSamples;
     for (size_t i = 0; i < querySet.n_cols; ++i)
     {
-      ObtainDistinctSamples(0, n, numSamplesReqd, distinctSamples);
+      distinctSamples = arma::randperm(n, numSamplesReqd);
       for (size_t j = 0; j < distinctSamples.n_elem; ++j)
         BaseCase(i, (size_t) distinctSamples[j]);
     }
@@ -200,9 +200,8 @@ inline double RASearchRules<SortPolicy,
         {
           // Then samplesReqd <= singleSampleLimit.
           // Hence, approximate the node by sampling enough number of points.
-          arma::uvec distinctSamples;
-          ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-              samplesReqd, distinctSamples);
+          arma::uvec distinctSamples =
+              arma::randperm(referenceNode.NumDescendants(), samplesReqd);
           for (size_t i = 0; i < distinctSamples.n_elem; ++i)
             // The counting of the samples are done in the 'BaseCase' function
             // so no book-keeping is required here.
@@ -216,9 +215,8 @@ inline double RASearchRules<SortPolicy,
           if (sampleAtLeaves) // If allowed to sample at leaves.
           {
             // Approximate node by sampling enough number of points.
-            arma::uvec distinctSamples;
-            ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-                samplesReqd, distinctSamples);
+            arma::uvec distinctSamples =
+                arma::randperm(referenceNode.NumDescendants(), samplesReqd);
             for (size_t i = 0; i < distinctSamples.n_elem; ++i)
               // The counting of the samples are done in the 'BaseCase' function
               // so no book-keeping is required here.
@@ -304,9 +302,8 @@ Rescore(const size_t queryIndex,
       {
         // Then, samplesReqd <= singleSampleLimit.  Hence, approximate the node
         // by sampling enough number of points.
-        arma::uvec distinctSamples;
-        ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-            samplesReqd, distinctSamples);
+        arma::uvec distinctSamples =
+            arma::randperm(referenceNode.NumDescendants(), samplesReqd);
         for (size_t i = 0; i < distinctSamples.n_elem; ++i)
           // The counting of the samples are done in the 'BaseCase' function so
           // no book-keeping is required here.
@@ -320,9 +317,8 @@ Rescore(const size_t queryIndex,
         if (sampleAtLeaves)
         {
           // Approximate node by sampling enough points.
-          arma::uvec distinctSamples;
-          ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-              samplesReqd, distinctSamples);
+          arma::uvec distinctSamples =
+              arma::randperm(referenceNode.NumDescendants(), samplesReqd);
           for (size_t i = 0; i < distinctSamples.n_elem; ++i)
             // The counting of the samples are done in the 'BaseCase' function
             // so no book-keeping is required here.
@@ -508,8 +504,8 @@ inline double RASearchRules<SortPolicy,
           for (size_t i = 0; i < queryNode.NumDescendants(); ++i)
           {
             const size_t queryIndex = queryNode.Descendant(i);
-            ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-                samplesReqd, distinctSamples);
+            distinctSamples = arma::randperm(referenceNode.NumDescendants(),
+                samplesReqd);
             for (size_t j = 0; j < distinctSamples.n_elem; ++j)
               // The counting of the samples are done in the 'BaseCase' function
               // so no book-keeping is required here.
@@ -538,8 +534,8 @@ inline double RASearchRules<SortPolicy,
             for (size_t i = 0; i < queryNode.NumDescendants(); ++i)
             {
               const size_t queryIndex = queryNode.Descendant(i);
-              ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-                  samplesReqd, distinctSamples);
+              distinctSamples = arma::randperm(referenceNode.NumDescendants(),
+                  samplesReqd);
               for (size_t j = 0; j < distinctSamples.n_elem; ++j)
                 // The counting of the samples are done in the 'BaseCase'
                 // function so no book-keeping is required here.
@@ -713,8 +709,8 @@ Rescore(TreeType& queryNode,
         for (size_t i = 0; i < queryNode.NumDescendants(); ++i)
         {
           const size_t queryIndex = queryNode.Descendant(i);
-          ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-              samplesReqd, distinctSamples);
+          distinctSamples = arma::randperm(referenceNode.NumDescendants(),
+              samplesReqd);
           for (size_t j = 0; j < distinctSamples.n_elem; ++j)
             // The counting of the samples are done in the 'BaseCase'
             // function so no book-keeping is required here.
@@ -742,8 +738,8 @@ Rescore(TreeType& queryNode,
           for (size_t i = 0; i < queryNode.NumDescendants(); ++i)
           {
             const size_t queryIndex = queryNode.Descendant(i);
-            ObtainDistinctSamples(0, referenceNode.NumDescendants(),
-                samplesReqd, distinctSamples);
+            distinctSamples = arma::randperm(referenceNode.NumDescendants(),
+                samplesReqd);
             for (size_t j = 0; j < distinctSamples.n_elem; ++j)
               // The counting of the samples are done in BaseCase() so no
               // book-keeping is required here.
diff -pruN 4.3.0-2/src/mlpack/methods/regularized_svd/regularized_svd_function_impl.hpp 4.4.0-1/src/mlpack/methods/regularized_svd/regularized_svd_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/regularized_svd/regularized_svd_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/regularized_svd/regularized_svd_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,13 +18,14 @@
 namespace mlpack {
 
 template <typename MatType>
-RegularizedSVDFunction<MatType>::RegularizedSVDFunction(const MatType& data,
+RegularizedSVDFunction<MatType>::RegularizedSVDFunction(const MatType& dataIn,
                                                         const size_t rank,
                                                         const double lambda) :
-    data(MakeAlias(const_cast<MatType&>(data), false)),
     rank(rank),
     lambda(lambda)
 {
+  MakeAlias(data, dataIn, dataIn.n_rows, dataIn.n_cols, false);
+
   // Number of users and items in the data.
   numUsers = max(data.row(0)) + 1;
   numItems = max(data.row(1)) + 1;
@@ -69,13 +70,13 @@ double RegularizedSVDFunction<MatType>::
 
     // Calculate the squared error in the prediction.
     const double rating = data(2, i);
-    double ratingError = rating - arma::dot(parameters.col(user),
+    double ratingError = rating - dot(parameters.col(user),
                                             parameters.col(item));
     double ratingErrorSquared = ratingError * ratingError;
 
     // Calculate the regularization penalty corresponding to the parameters.
-    double userVecNorm = arma::norm(parameters.col(user), 2);
-    double itemVecNorm = arma::norm(parameters.col(item), 2);
+    double userVecNorm = norm(parameters.col(user), 2);
+    double itemVecNorm = norm(parameters.col(item), 2);
     double regularizationError = lambda * (userVecNorm * userVecNorm +
                                            itemVecNorm * itemVecNorm);
 
@@ -108,7 +109,7 @@ void RegularizedSVDFunction<MatType>::Gr
 
     // Prediction error for the example.
     const double rating = data(2, i);
-    double ratingError = rating - arma::dot(parameters.col(user),
+    double ratingError = rating - dot(parameters.col(user),
                                             parameters.col(item));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
@@ -137,7 +138,7 @@ void RegularizedSVDFunction<MatType>::Gr
 
     // Prediction error for the example.
     const double rating = data(2, i);
-    double ratingError = rating - arma::dot(parameters.col(user),
+    double ratingError = rating - dot(parameters.col(user),
                                             parameters.col(item));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
@@ -196,7 +197,7 @@ double StandardSGD::Optimize(
 
     // Prediction error for the example.
     const double rating = data(2, currentFunction);
-    double ratingError = rating - arma::dot(parameters.col(user),
+    double ratingError = rating - dot(parameters.col(user),
                                             parameters.col(item));
 
     double lambda = function.Lambda();
@@ -293,7 +294,7 @@ inline double ParallelSGD<ExponentialBac
 
         // Prediction error for the example.
         const double rating = data(2, visitationOrder[j]);
-        double ratingError = rating - arma::dot(iterate.col(user),
+        double ratingError = rating - dot(iterate.col(user),
             iterate.col(item));
 
         double lambda = function.Lambda();
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/ddpg_impl.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/ddpg_impl.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/ddpg_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/ddpg_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -197,7 +197,7 @@ void DDPG<
 
   arma::mat sampledActionValues(action.size, sampledActions.size());
   for (size_t i = 0; i < sampledActions.size(); i++)
-    sampledActionValues.col(i) = arma::conv_to<arma::colvec>::from
+    sampledActionValues.col(i) = ConvTo<arma::colvec>::From
                                  (sampledActions[i].action);
   arma::mat learningQInput = arma::join_vert(sampledActionValues,
       sampledStates);
@@ -291,7 +291,7 @@ void DDPG<
     sample = arma::clamp(sample, -0.25, 0.25);
     outputAction = outputAction + sample;
   }
-  action.action = arma::conv_to<std::vector<double>>::from(outputAction);
+  action.action = ConvTo<std::vector<double>>::From(outputAction);
 }
 
 template <
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/acrobot.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/acrobot.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/acrobot.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/acrobot.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -174,10 +174,10 @@ class Acrobot
     nextState.Theta2() = Wrap(currentNextState[1], -M_PI, M_PI);
 
     //! The value of angular velocity is bounded in min and max value.
-    nextState.AngularVelocity1() = ClampRange(currentNextState[2],
-        -maxVel1, maxVel1);
-    nextState.AngularVelocity2() = ClampRange(currentNextState[3],
-        -maxVel2, maxVel2);
+    nextState.AngularVelocity1() = std::min(std::max(currentNextState[2],
+        -maxVel1), maxVel1);
+    nextState.AngularVelocity2() = std::min(std::max(currentNextState[3],
+        -maxVel2), maxVel2);
 
     // Check if the episode has terminated.
     bool done = IsTerminal(nextState);
@@ -212,7 +212,7 @@ class Acrobot
   State InitialSample()
   {
     stepsPerformed = 0;
-    return State((arma::randu<arma::colvec>(4) - 0.5) / 5.0);
+    return State((randu<arma::colvec>(4) - 0.5) / 5.0);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/cart_pole.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/cart_pole.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/cart_pole.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/cart_pole.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -206,7 +206,7 @@ class CartPole
   State InitialSample()
   {
     stepsPerformed = 0;
-    return State((arma::randu<arma::colvec>(4) - 0.5) / 10.0);
+    return State((randu<arma::colvec>(4) - 0.5) / 10.0);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/cont_double_pole_cart.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/cont_double_pole_cart.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/cont_double_pole_cart.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/cont_double_pole_cart.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -282,7 +282,7 @@ class ContinuousDoublePoleCart
   State InitialSample()
   {
     stepsPerformed = 0;
-    return State((arma::randu<arma::vec>(6) - 0.5) / 10.0);
+    return State((randu<arma::vec>(6) - 0.5) / 10.0);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/continuous_mountain_car.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/continuous_mountain_car.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/continuous_mountain_car.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/continuous_mountain_car.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,7 +18,6 @@
 #define MLPACK_METHODS_RL_ENVIRONMENT_CONTINUOUS_MOUNTAIN_CAR_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/clamp.hpp>
 
 namespace mlpack {
 
@@ -136,16 +135,16 @@ class ContinuousMountainCar
     stepsPerformed++;
 
     // Calculate acceleration.
-    double force = ClampRange(action.action[0], -1.0, 1.0);
+    double force = std::min(std::max(action.action[0], -1.0), 1.0);
 
     // Update states.
     nextState.Velocity() = state.Velocity() + force * duration - 0.0025 *
         std::cos(3 * state.Position());
-    nextState.Velocity() = ClampRange(nextState.Velocity(),
-      velocityMin, velocityMax);
+    nextState.Velocity() = std::min(std::max(nextState.Velocity(),
+      velocityMin), velocityMax);
     nextState.Position() = state.Position() + nextState.Velocity();
-    nextState.Position() = ClampRange(nextState.Position(),
-      positionMin, positionMax);
+    nextState.Position() = std::min(std::max(nextState.Position(),
+      positionMin), positionMax);
     if (nextState.Position() == positionMin && nextState.Velocity() < 0)
       nextState.Velocity() = 0.0;
 
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/double_pole_cart.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/double_pole_cart.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/double_pole_cart.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/double_pole_cart.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -288,7 +288,7 @@ class DoublePoleCart
   State InitialSample()
   {
     stepsPerformed = 0;
-    return State((arma::randu<arma::vec>(6) - 0.5) / 10.0);
+    return State((randu<arma::vec>(6) - 0.5) / 10.0);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/ftn.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/ftn.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/ftn.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/ftn.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -141,7 +141,7 @@ class FruitTreeNavigation
 
     // Do not reward the agent if time ran out.
     if (done && maxSteps != 0 && stepsPerformed >= maxSteps)
-      return arma::zeros(rewardSize);
+      return zeros(rewardSize);
 
     return fruitTree.GetReward(state);
   };
@@ -168,7 +168,7 @@ class FruitTreeNavigation
   State InitialSample()
   {
     stepsPerformed = 0;
-    return State(arma::zeros<arma::vec>(2));
+    return State(zeros<arma::vec>(2));
   }
 
   /**
@@ -223,8 +223,8 @@ class FruitTreeNavigation
                                  "Only depth values of: 5, 6, 7 are allowed.");
         }
 
-        arma::mat branches = arma::zeros(rewardSize, (size_t) std::pow(2, depth - 1));
-        tree = arma::join_rows(branches, Fruits());
+        arma::mat branches = zeros(rewardSize, (size_t) std::pow(2, depth - 1));
+        tree = join_rows(branches, Fruits());
       }
 
       // Extract array index from {(row, column)} representation.
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/mountain_car.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/mountain_car.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/mountain_car.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/mountain_car.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,7 +17,6 @@
 #define MLPACK_METHODS_RL_ENVIRONMENT_MOUNTAIN_CAR_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/clamp.hpp>
 
 namespace mlpack {
 
@@ -140,13 +139,13 @@ class MountainCar
     int direction = action.action - 1;
     nextState.Velocity() = state.Velocity() + 0.001 * direction - 0.0025 *
         std::cos(3 * state.Position());
-    nextState.Velocity() = ClampRange(nextState.Velocity(),
-        velocityMin, velocityMax);
+    nextState.Velocity() = std::min(std::max(nextState.Velocity(),
+        velocityMin), velocityMax);
 
     // Update states.
     nextState.Position() = state.Position() + nextState.Velocity();
-    nextState.Position() = ClampRange(nextState.Position(),
-        positionMin, positionMax);
+    nextState.Position() = std::min(std::max(nextState.Position(),
+        positionMin), positionMax);
 
     if (nextState.Position() == positionMin && nextState.Velocity() < 0)
       nextState.Velocity() = 0.0;
@@ -188,7 +187,7 @@ class MountainCar
     State state;
     stepsPerformed = 0;
     state.Velocity() = 0.0;
-    state.Position() = arma::as_scalar(arma::randu(1)) * 0.2 - 0.6;
+    state.Position() = randu() * 0.2 - 0.6;
     return state;
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/pendulum.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/pendulum.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/pendulum.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/pendulum.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,7 +17,6 @@
 #define MLPACK_METHODS_RL_ENVIRONMENT_PENDULUM_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/clamp.hpp>
 
 namespace mlpack {
 
@@ -153,7 +152,7 @@ class Pendulum
     const double length = 1.0;
 
     // Get action and clip the values between max and min limits.
-    double torque = ClampRange(action.action[0], -maxTorque, maxTorque);
+    double torque = std::min(std::max(action.action[0], -maxTorque), maxTorque);
 
     // Calculate costs of taking this action in the current state.
     double costs = std::pow(AngleNormalize(theta), 2) + 0.1 *
@@ -164,8 +163,8 @@ class Pendulum
         length) * std::sin(theta + M_PI) + 3.0 / (mass * std::pow(length, 2)) *
         torque) * dt;
     nextState.Theta() = theta + newAngularVelocity * dt;
-    nextState.AngularVelocity() = ClampRange(newAngularVelocity,
-        -maxAngularVelocity, maxAngularVelocity);
+    nextState.AngularVelocity() = std::min(std::max(newAngularVelocity,
+        -maxAngularVelocity), maxAngularVelocity);
 
     nextState.SetState();
 
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/reward_clipping.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/reward_clipping.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/environment/reward_clipping.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/environment/reward_clipping.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -13,7 +13,6 @@
 #define MLPACK_METHODS_RL_ENVIRONMENT_REWARD_CLIPPING_HPP
 
 #include <mlpack/prereqs.hpp>
-#include <mlpack/core/math/clamp.hpp>
 
 namespace mlpack {
 
@@ -89,9 +88,9 @@ class RewardClipping
                 State& nextState)
   {
     // Get original unclipped reward from base environment.
-    double unclippedReward =  environment.Sample(state, action, nextState);
+    double unclippedReward = environment.Sample(state, action, nextState);
     // Clip rewards according to the min and max limit and return.
-    return ClampRange(unclippedReward, minReward, maxReward);
+    return std::min(std::max(unclippedReward, minReward), maxReward);
   }
 
   /**
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/noise/gaussian.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/noise/gaussian.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/noise/gaussian.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/noise/gaussian.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -45,7 +45,7 @@ class GaussianNoise
    */
   arma::colvec sample()
   {
-    return arma::randi<arma::colvec>(size, arma::distr_param(mu, sigma));
+    return sigma * randn<arma::colvec>(size) + mu;
   }
 
  private:
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/noise/ornstein_uhlenbeck.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/noise/ornstein_uhlenbeck.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/noise/ornstein_uhlenbeck.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/noise/ornstein_uhlenbeck.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,7 +30,7 @@ class OUNoise
           double mu = 0.0,
           double theta = 0.15,
           double sigma = 0.2) :
-      mu(mu * arma::ones<arma::colvec>(size)),
+      mu(mu * ones<arma::colvec>(size)),
       theta(theta),
       sigma(sigma)
   {
@@ -54,7 +54,7 @@ class OUNoise
   {
     arma::colvec x = state;
     arma::colvec dx = theta * (mu - x) + 
-        sigma * arma::randn<arma::colvec>(x.n_elem);
+        sigma * randn<arma::colvec>(x.n_elem);
     state = x + dx;
     return state;
   }
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/q_learning_impl.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/q_learning_impl.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/q_learning_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/q_learning_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -106,7 +106,7 @@ arma::Col<size_t> QLearning<
 {
   // Take best possible action at a particular instance.
   arma::Col<size_t> bestActions(actionValues.n_cols);
-  arma::rowvec maxActionValues = arma::max(actionValues, 0);
+  arma::rowvec maxActionValues = max(actionValues, 0);
   for (size_t i = 0; i < actionValues.n_cols; ++i)
   {
     bestActions(i) = arma::as_scalar(
@@ -260,7 +260,7 @@ void QLearning<
         arma::size(atomSize, 1));
   }
 
-  arma::mat tZ = (arma::conv_to<arma::mat>::from(config.Discount() *
+  arma::mat tZ = (ConvTo<arma::mat>::From(config.Discount() *
       (support * (1 - isTerminal))).each_row() + sampledRewards);
   tZ = arma::clamp(tZ, config.VMin(), config.VMax());
   arma::mat b = (tZ - config.VMin()) / (config.VMax() - config.VMin()) *
@@ -271,7 +271,7 @@ void QLearning<
   arma::mat projDistUpper = nextDist % (u - b);
   arma::mat projDistLower = nextDist % (b - l);
 
-  arma::mat projDist = arma::zeros<arma::mat>(arma::size(nextDist));
+  arma::mat projDist = zeros<arma::mat>(arma::size(nextDist));
   for (size_t batchNo = 0; batchNo < batchSize; batchNo++)
   {
     for (size_t j = 0; j < atomSize; j++)
@@ -282,7 +282,7 @@ void QLearning<
   }
   arma::mat dists;
   learningNetwork.Forward(sampledStates, dists);
-  arma::mat lossGradients = arma::zeros<arma::mat>(arma::size(dists));
+  arma::mat lossGradients = zeros<arma::mat>(arma::size(dists));
   for (size_t i = 0; i < batchSize; ++i)
   {
     lossGradients(sampledActions[i].action * atomSize, i,
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/q_networks/dueling_dqn.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/q_networks/dueling_dqn.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/q_networks/dueling_dqn.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/q_networks/dueling_dqn.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -214,10 +214,10 @@ class DuelingDQN
     arma::mat gradLoss;
     lossFunction.Backward(this->actionValues, target, gradLoss);
 
-    arma::mat gradValue = arma::sum(gradLoss);
+    arma::mat gradValue = sum(gradLoss);
     arma::mat gradAdvantage = gradLoss.each_row() - arma::mean(gradLoss);
 
-    arma::mat grad = arma::join_cols(gradValue, gradAdvantage);
+    arma::mat grad = join_cols(gradValue, gradAdvantage);
     completeNetwork.Backward(state, grad, gradient);
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/replay/prioritized_replay.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/replay/prioritized_replay.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/replay/prioritized_replay.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/replay/prioritized_replay.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -241,7 +241,7 @@ class PrioritizedReplay
     for (size_t i = 0; i < sampledIndices.n_rows; ++i)
     {
       double p_sample = idxSum.Get(sampledIndices(i)) / idxSum.Sum();
-      weights(i) = pow(numSample * p_sample, -beta);
+      weights(i) = std::pow(numSample * p_sample, -beta);
     }
     weights /= weights.max();
   }
@@ -255,7 +255,7 @@ class PrioritizedReplay
   void UpdatePriorities(arma::ucolvec& indices, arma::colvec& priorities)
   {
       arma::colvec alphaPri = alpha * priorities;
-      maxPriority = std::max(maxPriority, arma::max(priorities));
+      maxPriority = std::max(maxPriority, max(priorities));
       idxSum.BatchUpdate(indices, alphaPri);
   }
 
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/replay/random_replay.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/replay/random_replay.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/replay/random_replay.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/replay/random_replay.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -186,7 +186,7 @@ class RandomReplay
               arma::irowvec& isTerminal)
   {
     size_t upperBound = full ? capacity : position;
-    arma::uvec sampledIndices = arma::randi<arma::uvec>(
+    arma::uvec sampledIndices = randi<arma::uvec>(
         batchSize, arma::distr_param(0, upperBound - 1));
 
     sampledStates = states.cols(sampledIndices);
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/sac_impl.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/sac_impl.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/sac_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/sac_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -183,11 +183,11 @@ void SAC<
   targetQ1Network.Predict(targetQInput, Q1);
   targetQ2Network.Predict(targetQInput, Q2);
   arma::rowvec nextQ = sampledRewards + config.Discount() * ((1 - isTerminal)
-      % arma::min(Q1, Q2));
+      % min(Q1, Q2));
 
   arma::mat sampledActionValues(action.size, sampledActions.size());
   for (size_t i = 0; i < sampledActions.size(); i++)
-    sampledActionValues.col(i) = arma::conv_to<arma::colvec>::from
+    sampledActionValues.col(i) = ConvTo<arma::colvec>::From
                                  (sampledActions[i].action);
   arma::mat learningQInput = arma::join_vert(sampledActionValues,
       sampledStates);
@@ -302,11 +302,12 @@ void SAC<
 
   if (!deterministic)
   {
-    arma::colvec noise = arma::randn<arma::colvec>(outputAction.n_rows) * 0.1;
+    arma::colvec noise;
+    noise.randn(outputAction.n_rows) * 0.1;
     noise = arma::clamp(noise, -0.25, 0.25);
     outputAction = outputAction + noise;
   }
-  action.action = arma::conv_to<std::vector<double>>::from(outputAction);
+  action.action = ConvTo<std::vector<double>>::From(outputAction);
 }
 
 template <
diff -pruN 4.3.0-2/src/mlpack/methods/reinforcement_learning/td3_impl.hpp 4.4.0-1/src/mlpack/methods/reinforcement_learning/td3_impl.hpp
--- 4.3.0-2/src/mlpack/methods/reinforcement_learning/td3_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/reinforcement_learning/td3_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -190,11 +190,11 @@ void TD3<
   targetQ1Network.Predict(targetQInput, Q1);
   targetQ2Network.Predict(targetQInput, Q2);
   arma::rowvec nextQ = sampledRewards + config.Discount() * ((1 - isTerminal)
-      % arma::min(Q1, Q2));
+      % min(Q1, Q2));
 
   arma::mat sampledActionValues(action.size, sampledActions.size());
   for (size_t i = 0; i < sampledActions.size(); i++)
-    sampledActionValues.col(i) = arma::conv_to<arma::colvec>::from
+    sampledActionValues.col(i) = ConvTo<arma::colvec>::From
                                  (sampledActions[i].action);
   arma::mat learningQInput = arma::join_vert(sampledActionValues,
       sampledStates);
@@ -297,11 +297,12 @@ void TD3<
 
   if (!deterministic)
   {
-    arma::colvec noise = arma::randn<arma::colvec>(outputAction.n_rows) * 0.1;
+    arma::colvec noise;
+    noise.randn(outputAction.n_rows) * 0.1;
     noise = arma::clamp(noise, -0.25, 0.25);
     outputAction = outputAction + noise;
   }
-  action.action = arma::conv_to<std::vector<double>>::from(outputAction);
+  action.action = ConvTo<std::vector<double>>::From(outputAction);
 }
 
 template <
diff -pruN 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression.hpp 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression.hpp
--- 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -54,9 +54,15 @@ namespace mlpack {
  * regressor.Classify(testData, predictions);
  * @endcode
  */
+template<typename MatType = arma::mat>
 class SoftmaxRegression
 {
  public:
+  typedef typename MatType::elem_type ElemType;
+  typedef typename GetDenseMatType<MatType>::type DenseMatType;
+  typedef typename GetDenseRowType<MatType>::type DenseRowType;
+  typedef typename GetDenseColType<MatType>::type DenseColType;
+
   /**
    * Initialize the SoftmaxRegression without performing training.  Default
    * value of lambda is 0.0001.  Be sure to use Train() before calling
@@ -68,53 +74,167 @@ class SoftmaxRegression
    */
   SoftmaxRegression(const size_t inputSize = 0,
                     const size_t numClasses = 0,
-                    const bool fitIntercept = false);
+                    const bool fitIntercept = true);
+
   /**
    * Construct the SoftmaxRegression class with the provided data and labels.
    * This will train the model. Optionally, the parameter 'lambda' can be
    * passed, which controls the amount of L2-regularization in the objective
-   * function. By default, the model takes a small value.
+   * function. By default, the model takes a small value for lambda.
    *
    * @tparam OptimizerType Desired optimizer type.
    * @param data Input training features. Each column associate with one sample
    * @param labels Labels associated with the feature data.
    * @param numClasses Number of classes for classification.
-   * @param optimizer Desired optimizer.
    * @param lambda L2-regularization constant.
    * @param fitIntercept add intercept term or not.
+   * @param callbacks Callback(s) for ensmallen optimizer `OptimizerType`.
+   *        See https://www.ensmallen.org/docs.html#callback-documentation.
    */
-  template<typename OptimizerType = ens::L_BFGS>
-  SoftmaxRegression(const arma::mat& data,
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, SoftmaxRegressionFunction<MatType>, DenseMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  SoftmaxRegression(const MatType& data,
                     const arma::Row<size_t>& labels,
                     const size_t numClasses,
                     const double lambda = 0.0001,
-                    const bool fitIntercept = false,
-                    OptimizerType optimizer = OptimizerType());
+                    const bool fitIntercept = true,
+                    CallbackTypes&&... callbacks);
+
   /**
-   * Construct the SoftmaxRegression class with the provided data and labels.
-   * This will train the model. Optionally, the parameter 'lambda' can be
-   * passed, which controls the amount of L2-regularization in the objective
-   * function. By default, the model takes a small value.
+   * Construct the SoftmaxRegression class with the provided data and labels,
+   * using the given ensmallen optimizer.  This will train the model.
+   * Optionally, the parameter 'lambda' can be passed, which controls the amount
+   * of L2-regularization in the objective function. By default, the model takes
+   * a small value for lambda.
    *
    * @tparam OptimizerType Desired optimizer type.
-   * @tparam CallbackTypes Types of Callback Functions.
    * @param data Input training features. Each column associate with one sample
    * @param labels Labels associated with the feature data.
    * @param numClasses Number of classes for classification.
+   * @param optimizer Desired optimizer.
    * @param lambda L2-regularization constant.
    * @param fitIntercept add intercept term or not.
-   * @param optimizer Desired optimizer.
-   * @param callbacks Callback function for ensmallen optimizer `OptimizerType`.
+   * @param callbacks Callback(s) for ensmallen optimizer `OptimizerType`.
    *        See https://www.ensmallen.org/docs.html#callback-documentation.
    */
-  template<typename OptimizerType, typename... CallbackTypes>
-  SoftmaxRegression(const arma::mat& data,
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, SoftmaxRegressionFunction<MatType>, DenseMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  SoftmaxRegression(const MatType& data,
                     const arma::Row<size_t>& labels,
                     const size_t numClasses,
-                    const double lambda,
-                    const bool fitIntercept,
-                    OptimizerType optimizer,
+                    OptimizerType& optimizer,
+                    const double lambda = 0.0001,
+                    const bool fitIntercept = true,
                     CallbackTypes&&... callbacks);
+
+  /**
+   * Train the softmax regression with the given training data.
+   *
+   * This function is deprecated and will be removed in mlpack 5.0.0 (for
+   * consistency reasons); use the other overload of Train() that allows
+   * specifying hyperparameters in addition to the optimizer and callbacks.
+   *
+   * @tparam OptimizerType Desired optimizer type.
+   * @param data Input data with each column as one example.
+   * @param labels Labels associated with the feature data.
+   * @param numClasses Number of classes for classification.
+   * @param optimizer Desired optimizer.
+   * @param callbacks Callback(s) for ensmallen optimizer `OptimizerType`.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return Objective value of the final point.
+   */
+  template<typename OptimizerType = ens::L_BFGS,
+           typename FirstCallbackType,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, SoftmaxRegressionFunction<MatType>, DenseMatType
+           >::value>::type,
+           typename = typename std::enable_if<
+              std::is_class<FirstCallbackType>::value
+           >::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  mlpack_deprecated /* To be removed in mlpack 5.0.0; use the overload below. */
+  double Train(const MatType& data,
+               const arma::Row<size_t>& labels,
+               const size_t numClasses,
+               OptimizerType optimizer,
+               FirstCallbackType&& firstCallback,
+               CallbackTypes&&... callbacks);
+
+  /**
+   * Train the softmax regression with the given training data and optionally
+   * the given hyperparameters.
+   *
+   * @tparam OptimizerType Desired optimizer type.
+   * @param data Input data with each column as one example.
+   * @param labels Labels associated with the feature data.
+   * @param numClasses Number of classes for classification.
+   * @param lambda L2-regularization constant.
+   * @param fitIntercept Whether or not to fit an intercept term to the model.
+   * @param callbacks Callback(s) for ensmallen optimizer `OptimizerType`.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return Objective value of the final point.
+   */
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, SoftmaxRegressionFunction<MatType>, DenseMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 const double lambda = 0.0001,
+                 const bool fitIntercept = true,
+                 CallbackTypes&&... callbacks);
+
+  /**
+   * Train the softmax regression with the given training data and optionally
+   * the given hyperparameters, using the given optimizer.
+   *
+   * @tparam OptimizerType Desired optimizer type.
+   * @param data Input data with each column as one example.
+   * @param labels Labels associated with the feature data.
+   * @param numClasses Number of classes for classification.
+   * @param optimizer Desired optimizer.
+   * @param lambda L2-regularization constant.
+   * @param fitIntercept Whether or not to fit an intercept term to the model.
+   * @param callbacks Callback(s) for ensmallen optimizer `OptimizerType`.
+   *      See https://www.ensmallen.org/docs.html#callback-documentation.
+   * @return Objective value of the final point.
+   */
+  template<typename OptimizerType = ens::L_BFGS,
+           typename... CallbackTypes,
+           typename = typename std::enable_if<IsEnsOptimizer<
+               OptimizerType, SoftmaxRegressionFunction<MatType>, DenseMatType
+           >::value>::type,
+           typename = typename std::enable_if<IsEnsCallbackTypes<
+               CallbackTypes...
+           >::value>::type>
+  ElemType Train(const MatType& data,
+                 const arma::Row<size_t>& labels,
+                 const size_t numClasses,
+                 OptimizerType& optimizer,
+                 const double lambda = 0.0001,
+                 const bool fitIntercept = true,
+                 CallbackTypes&&... callbacks);
+
   /**
    * Classify the given points, returning the predicted labels for each point.
    * The function calculates the probabilities for every class, given a data
@@ -123,11 +243,13 @@ class SoftmaxRegression
    * @param dataset Set of points to classify.
    * @param labels Predicted labels for each point.
    */
-  void Classify(const arma::mat& dataset, arma::Row<size_t>& labels) const;
+  void Classify(const MatType& dataset, arma::Row<size_t>& labels) const;
+
   /**
    * Classify the given point. The predicted class label is returned.
    * The function calculates the probabilites for every class, given the point.
    * It then chooses the class which has the highest probability among all.
+   *
    * @param point Point to be classified.
    * @return Predicted class label of the point.
    */
@@ -135,6 +257,19 @@ class SoftmaxRegression
   size_t Classify(const VecType& point) const;
 
   /**
+   * Classify the given point, storing the predicted label in `prediction` and
+   * storing class probabilities in `probabilitiesVec`.
+   *
+   * @param point Point to be classified.
+   * @param prediction size_t to store the label in.
+   * @param probabilitiesVec Vector to store class probabilities in.
+   */
+  template<typename VecType>
+  void Classify(const VecType& point,
+                size_t& prediction,
+                DenseColType& probabilitiesVec) const;
+
+  /**
    * Classify the given points, returning class probabilities and predicted
    * class label for each point.
    * The function calculates the probabilities for every class, given a data
@@ -145,9 +280,9 @@ class SoftmaxRegression
    * @param labels Predicted labels for each point.
    * @param probabilities Class probabilities for each point.
    */
-  void Classify(const arma::mat& dataset,
+  void Classify(const MatType& dataset,
                 arma::Row<size_t>& labels,
-                arma::mat& probabilities) const;
+                DenseMatType& probabilities) const;
 
   /**
    * Classify the given points, returning class probabilities for each point.
@@ -155,8 +290,9 @@ class SoftmaxRegression
    * @param dataset Matrix of data points to be classified.
    * @param probabilities Class probabilities for each point.
    */
-  void Classify(const arma::mat& dataset,
-                arma::mat& probabilities) const;
+  mlpack_deprecated /** To be removed in mlpack 5.0.0. */
+  void Classify(const MatType& dataset,
+                DenseMatType& probabilities) const;
 
   /**
    * Computes accuracy of the learned model given the feature data and the
@@ -166,42 +302,8 @@ class SoftmaxRegression
    * @param testData Matrix of data points using which predictions are made.
    * @param labels Vector of labels associated with the data.
    */
-  double ComputeAccuracy(const arma::mat& testData,
+  double ComputeAccuracy(const MatType& testData,
                          const arma::Row<size_t>& labels) const;
-  /**
-   * Train the softmax regression with the given training data.
-   *
-   * @tparam OptimizerType Desired optimizer type.
-   * @param data Input data with each column as one example.
-   * @param labels Labels associated with the feature data.
-   * @param numClasses Number of classes for classification.
-   * @param optimizer Desired optimizer.
-   * @return Objective value of the final point.
-   */
-  template<typename OptimizerType = ens::L_BFGS>
-  double Train(const arma::mat& data,
-               const arma::Row<size_t>& labels,
-               const size_t numClasses,
-               OptimizerType optimizer = OptimizerType());
-  /**
-   * Train the softmax regression with the given training data.
-   *
-   * @tparam OptimizerType Desired optimizer type.
-   * @tparam CallbackTypes Types of Callback Functions.
-   * @param data Input data with each column as one example.
-   * @param labels Labels associated with the feature data.
-   * @param numClasses Number of classes for classification.
-   * @param optimizer Desired optimizer.
-   * @param callbacks Callback function for ensmallen optimizer `OptimizerType`.
-   *      See https://www.ensmallen.org/docs.html#callback-documentation.
-   * @return Objective value of the final point.
-   */
-  template<typename OptimizerType = ens::L_BFGS, typename... CallbackTypes>
-  double Train(const arma::mat& data,
-               const arma::Row<size_t>& labels,
-               const size_t numClasses,
-               OptimizerType optimizer,
-               CallbackTypes&&... callbacks);
 
   //! Sets the number of classes.
   size_t& NumClasses() { return numClasses; }
@@ -217,9 +319,16 @@ class SoftmaxRegression
   bool FitIntercept() const { return fitIntercept; }
 
   //! Get the model parameters.
-  arma::mat& Parameters() { return parameters; }
+  MatType& Parameters() { return parameters; }
   //! Get the model parameters.
-  const arma::mat& Parameters() const { return parameters; }
+  const MatType& Parameters() const { return parameters; }
+
+  /**
+   * Reset the weights in the model to small random values.  This function can
+   * be used between calls to Train(), to force learning of a new model instead
+   * of incremental training.
+   */
+  void Reset();
 
   //! Gets the features size of the training data
   size_t FeatureSize() const
@@ -230,17 +339,11 @@ class SoftmaxRegression
    * Serialize the SoftmaxRegression model.
    */
   template<typename Archive>
-  void serialize(Archive& ar, const uint32_t /* version */)
-  {
-    ar(CEREAL_NVP(parameters));
-    ar(CEREAL_NVP(numClasses));
-    ar(CEREAL_NVP(lambda));
-    ar(CEREAL_NVP(fitIntercept));
-  }
+  void serialize(Archive& ar, const uint32_t version);
 
  private:
   //! Parameters after optimization.
-  arma::mat parameters;
+  DenseMatType parameters;
   //! Number of classes.
   size_t numClasses;
   //! L2-regularization constant.
@@ -251,6 +354,9 @@ class SoftmaxRegression
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename MatType),
+    (mlpack::SoftmaxRegression<MatType>), (1));
+
 // Include implementation.
 #include "softmax_regression_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_function.hpp 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_function.hpp
--- 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,9 +18,14 @@
 
 namespace mlpack {
 
+template<typename MatType = arma::mat>
 class SoftmaxRegressionFunction
 {
  public:
+  typedef typename MatType::elem_type ElemType;
+  typedef typename GetDenseMatType<MatType>::type DenseMatType;
+  typedef typename GetSparseMatType<MatType>::type SpMatType;
+
   /**
    * Construct the Softmax Regression objective function with the given
    * parameters.
@@ -31,14 +36,14 @@ class SoftmaxRegressionFunction
    * @param lambda L2-regularization constant.
    * @param fitIntercept Intercept term flag.
    */
-  SoftmaxRegressionFunction(const arma::mat& data,
+  SoftmaxRegressionFunction(const MatType& data,
                             const arma::Row<size_t>& labels,
                             const size_t numClasses,
                             const double lambda = 0.0001,
                             const bool fitIntercept = false);
 
   //! Initializes the parameters of the model to suitable values.
-  const arma::mat InitializeWeights();
+  const DenseMatType InitializeWeights();
 
   /**
    * Shuffle the dataset.
@@ -54,9 +59,9 @@ class SoftmaxRegressionFunction
    * @param fitIntercept If true, an intercept is fitted.
    * @return Initialized model weights.
    */
-  static const arma::mat InitializeWeights(const size_t featureSize,
-                                           const size_t numClasses,
-                                           const bool fitIntercept = false);
+  static const DenseMatType InitializeWeights(const size_t featureSize,
+                                              const size_t numClasses,
+                                              const bool fitIntercept = false);
 
   /**
    * Initialize Softmax Regression weights (trainable parameters) with the given
@@ -67,7 +72,7 @@ class SoftmaxRegressionFunction
    * @param numClasses Number of classes for classification.
    * @param fitIntercept Intercept term flag.
    */
-  static void InitializeWeights(arma::mat &weights,
+  static void InitializeWeights(DenseMatType& weights,
                                 const size_t featureSize,
                                 const size_t numClasses,
                                 const bool fitIntercept = false);
@@ -79,7 +84,7 @@ class SoftmaxRegressionFunction
    * @param groundTruth Pointer to arma::mat which stores the computed matrix.
    */
   void GetGroundTruthMatrix(const arma::Row<size_t>& labels,
-                            arma::sp_mat& groundTruth);
+                            SpMatType& groundTruth);
 
   /**
    * Evaluate the probabilities matrix with the passed parameters.
@@ -92,8 +97,8 @@ class SoftmaxRegressionFunction
    * @param start Index of point to start at.
    * @param batchSize Number of points to calculate probabilities for.
    */
-  void GetProbabilitiesMatrix(const arma::mat& parameters,
-                              arma::mat& probabilities,
+  void GetProbabilitiesMatrix(const DenseMatType& parameters,
+                              DenseMatType& probabilities,
                               const size_t start,
                               const size_t batchSize) const;
 
@@ -106,7 +111,7 @@ class SoftmaxRegressionFunction
    *
    * @param parameters Current values of the model parameters.
    */
-  double Evaluate(const arma::mat& parameters) const;
+  ElemType Evaluate(const DenseMatType& parameters) const;
 
   /**
    * Evaluate the objective function of the softmax regression model for a
@@ -119,9 +124,9 @@ class SoftmaxRegressionFunction
    * @param start First index of the data points to use.
    * @param batchSize Number of data points to evaluate objective for.
    */
-  double Evaluate(const arma::mat& parameters,
-                  const size_t start,
-                  const size_t batchSize = 1) const;
+  ElemType Evaluate(const DenseMatType& parameters,
+                    const size_t start,
+                    const size_t batchSize = 1) const;
 
   /**
    * Evaluates the gradient values of the objective function given the current
@@ -132,7 +137,8 @@ class SoftmaxRegressionFunction
    * @param parameters Current values of the model parameters.
    * @param gradient Matrix where gradient values will be stored.
    */
-  void Gradient(const arma::mat& parameters, arma::mat& gradient) const;
+  template<typename GradType>
+  void Gradient(const DenseMatType& parameters, GradType& gradient) const;
 
   /**
    * Evaluate the gradient of the objective function given the current set of
@@ -145,9 +151,10 @@ class SoftmaxRegressionFunction
    * @param gradient Matrix to store gradient into.
    * @param batchSize Number of data points to evaluate gradient for.
    */
-  void Gradient(const arma::mat& parameters,
+  template<typename GradType>
+  void Gradient(const DenseMatType& parameters,
                 const size_t start,
-                arma::mat& gradient,
+                GradType& gradient,
                 const size_t batchSize = 1) const;
 
   /**
@@ -159,12 +166,13 @@ class SoftmaxRegressionFunction
    *    gradient is to be computed.
    * @param gradient Out param for the gradient value.
    */
-  void PartialGradient(const arma::mat& parameters,
+  template<typename GradType>
+  void PartialGradient(const DenseMatType& parameters,
                        size_t j,
-                       arma::sp_mat& gradient) const;
+                       GradType& gradient) const;
 
   //! Return the initial point for the optimization.
-  const arma::mat& GetInitialPoint() const { return initialPoint; }
+  const DenseMatType& GetInitialPoint() const { return initialPoint; }
 
   //! Gets the number of classes.
   size_t NumClasses() const { return numClasses; }
@@ -189,11 +197,11 @@ class SoftmaxRegressionFunction
 
  private:
   //! Training data matrix.  This is an alias until the data is shuffled.
-  arma::mat data;
+  MatType data;
   //! Label matrix for the provided data.
-  arma::sp_mat groundTruth;
+  SpMatType groundTruth;
   //! Initial parameter point.
-  arma::mat initialPoint;
+  DenseMatType initialPoint;
   //! Number of classes.
   size_t numClasses;
   //! L2-regularization constant.
diff -pruN 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_function_impl.hpp 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -16,17 +16,19 @@
 
 namespace mlpack {
 
-inline SoftmaxRegressionFunction::SoftmaxRegressionFunction(
-    const arma::mat& data,
+template<typename MatType>
+inline SoftmaxRegressionFunction<MatType>::SoftmaxRegressionFunction(
+    const MatType& dataIn,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
     const bool fitIntercept) :
-    data(MakeAlias(const_cast<arma::mat&>(data), false)),
     numClasses(numClasses),
     lambda(lambda),
     fitIntercept(fitIntercept)
 {
+  MakeAlias(data, dataIn, dataIn.n_rows, dataIn.n_cols, 0, false);
+
   // Initialize the parameters to suitable values.
   initialPoint = InitializeWeights();
 
@@ -37,14 +39,15 @@ inline SoftmaxRegressionFunction::Softma
 /**
  * Shuffle the data.
  */
-inline void SoftmaxRegressionFunction::Shuffle()
+template<typename MatType>
+inline void SoftmaxRegressionFunction<MatType>::Shuffle()
 {
   // Determine new ordering.
   arma::uvec ordering = arma::shuffle(arma::linspace<arma::uvec>(0,
       data.n_cols - 1, data.n_cols));
 
   // Re-sort data.
-  arma::mat newData = data.cols(ordering);
+  MatType newData = data.cols(ordering);
   ClearAlias(data);
   data = std::move(newData);
 
@@ -54,20 +57,20 @@ inline void SoftmaxRegressionFunction::S
     reverseOrdering[ordering[i]] = i;
 
   arma::umat newLocations(2, groundTruth.n_nonzero);
-  arma::vec values(groundTruth.n_nonzero);
-  arma::sp_mat::const_iterator it = groundTruth.begin();
+  arma::Col<ElemType> values(groundTruth.n_nonzero);
+  typename SpMatType::const_iterator it = groundTruth.begin();
   size_t loc = 0;
   while (it != groundTruth.end())
   {
-    newLocations(0, loc) = reverseOrdering(it.col());
-    newLocations(1, loc) = it.row();
+    newLocations(0, loc) = it.row();
+    newLocations(1, loc) = reverseOrdering(it.col());
     values(loc) = (*it);
 
     ++it;
     ++loc;
   }
 
-  groundTruth = arma::sp_mat(newLocations, values, groundTruth.n_rows,
+  groundTruth = SpMatType(newLocations, values, groundTruth.n_rows,
       groundTruth.n_cols);
 }
 
@@ -76,23 +79,28 @@ inline void SoftmaxRegressionFunction::S
  * normal distribution. The weights cannot be initialized to zero, as that will
  * lead to each class output being the same.
  */
-inline const arma::mat SoftmaxRegressionFunction::InitializeWeights()
+template<typename MatType>
+inline const typename SoftmaxRegressionFunction<MatType>::DenseMatType
+SoftmaxRegressionFunction<MatType>::InitializeWeights()
 {
   return InitializeWeights(data.n_rows, numClasses, fitIntercept);
 }
 
-inline const arma::mat SoftmaxRegressionFunction::InitializeWeights(
+template<typename MatType>
+inline const typename SoftmaxRegressionFunction<MatType>::DenseMatType
+SoftmaxRegressionFunction<MatType>::InitializeWeights(
     const size_t featureSize,
     const size_t numClasses,
     const bool fitIntercept)
 {
-    arma::mat parameters;
-    InitializeWeights(parameters, featureSize, numClasses, fitIntercept);
-    return parameters;
+  DenseMatType parameters;
+  InitializeWeights(parameters, featureSize, numClasses, fitIntercept);
+  return parameters;
 }
 
-inline void SoftmaxRegressionFunction::InitializeWeights(
-    arma::mat &weights,
+template<typename MatType>
+inline void SoftmaxRegressionFunction<MatType>::InitializeWeights(
+    typename SoftmaxRegressionFunction<MatType>::DenseMatType& weights,
     const size_t featureSize,
     const size_t numClasses,
     const bool fitIntercept)
@@ -112,8 +120,10 @@ inline void SoftmaxRegressionFunction::I
  * labels. The output is in the form of a matrix, which leads to simpler
  * calculations in the Evaluate() and Gradient() methods.
  */
-inline void SoftmaxRegressionFunction::GetGroundTruthMatrix(
-    const arma::Row<size_t>& labels, arma::sp_mat& groundTruth)
+template<typename MatType>
+inline void SoftmaxRegressionFunction<MatType>::GetGroundTruthMatrix(
+    const arma::Row<size_t>& labels,
+    typename SoftmaxRegressionFunction<MatType>::SpMatType& groundTruth)
 {
   // Calculate the ground truth matrix according to the labels passed. The
   // ground truth matrix is a matrix of dimensions 'numClasses * numExamples',
@@ -136,53 +146,56 @@ inline void SoftmaxRegressionFunction::G
   }
 
   // All entries are '1'.
-  arma::vec values;
+  arma::Col<ElemType> values;
   values.ones(labels.n_elem);
 
   // Calculate the matrix.
-  groundTruth = arma::sp_mat(rowPointers, colPointers, values, numClasses,
-                             labels.n_elem);
+  groundTruth = SpMatType(rowPointers, colPointers, values, numClasses,
+      labels.n_elem);
 }
 
 /**
  * Evaluate the probabilities matrix. If fitIntercept flag is true,
  * it should consider the parameters.cols(0) intercept term.
  */
-inline void SoftmaxRegressionFunction::GetProbabilitiesMatrix(
-    const arma::mat& parameters,
-    arma::mat& probabilities,
+template<typename MatType>
+inline void SoftmaxRegressionFunction<MatType>::GetProbabilitiesMatrix(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters,
+    typename SoftmaxRegressionFunction<MatType>::DenseMatType& probabilities,
     const size_t start,
     const size_t batchSize) const
 {
-  arma::mat hypothesis;
+  DenseMatType hypothesis;
 
   if (fitIntercept)
   {
     // In order to add the intercept term, we should compute following matrix:
-    //     [1; data] = arma::join_cols(ones(1, data.n_cols), data)
-    //     hypothesis = arma::exp(parameters * [1; data]).
+    //     [1; data] = join_cols(ones(1, data.n_cols), data)
+    //     hypothesis = exp(parameters * [1; data]).
     //
     // Since the cost of join may be high due to the copy of original data,
     // split the hypothesis computation to two components.
-    hypothesis = arma::exp(
-        arma::repmat(parameters.col(0), 1, batchSize) +
+    hypothesis = exp(
+        repmat(parameters.col(0), 1, batchSize) +
         parameters.cols(1, parameters.n_cols - 1) *
         data.cols(start, start + batchSize - 1));
   }
   else
   {
-    hypothesis = arma::exp(parameters *
-        data.cols(start, start + batchSize - 1));
+    hypothesis = exp(parameters * data.cols(start, start + batchSize - 1));
   }
 
-  probabilities = hypothesis / arma::repmat(arma::sum(hypothesis, 0),
-                                            numClasses, 1);
+  probabilities = hypothesis / repmat(sum(hypothesis, 0), numClasses, 1);
 }
 
 /**
  * Evaluates the objective function given the parameters.
  */
-inline double SoftmaxRegressionFunction::Evaluate(const arma::mat& parameters)
+template<typename MatType>
+inline
+typename SoftmaxRegressionFunction<MatType>::ElemType
+SoftmaxRegressionFunction<MatType>::Evaluate(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters)
     const
 {
   // The objective function is the negative log likelihood of the model
@@ -201,15 +214,16 @@ inline double SoftmaxRegressionFunction:
   // The sum is calculated over all the classes.
   // x_i is the input vector for a particular training example.
   // theta_j is the parameter vector associated with a particular class.
-  arma::mat probabilities;
+  DenseMatType probabilities;
   GetProbabilitiesMatrix(parameters, probabilities, 0, data.n_cols);
 
   // Calculate the log likelihood and regularization terms.
-  double logLikelihood, weightDecay, cost;
+  ElemType logLikelihood, weightDecay, cost;
 
-  logLikelihood = arma::accu(groundTruth % arma::log(probabilities)) /
+  logLikelihood = arma::accu(groundTruth % log(probabilities)) /
                   data.n_cols;
-  weightDecay = 0.5 * lambda * arma::accu(parameters % parameters);
+  weightDecay = ((typename MatType::elem_type) 0.5) * lambda *
+      arma::accu(parameters % parameters);
 
   // The cost is the sum of the negative log likelihood and the regularization
   // terms.
@@ -221,19 +235,24 @@ inline double SoftmaxRegressionFunction:
 /**
  * Evaluate the objective function for the given points given the parameters.
  */
-inline double SoftmaxRegressionFunction::Evaluate(const arma::mat& parameters,
-                                                  const size_t start,
-                                                  const size_t batchSize) const
+template<typename MatType>
+inline
+typename SoftmaxRegressionFunction<MatType>::ElemType
+SoftmaxRegressionFunction<MatType>::Evaluate(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters,
+    const size_t start,
+    const size_t batchSize) const
 {
-  arma::mat probabilities;
+  MatType probabilities;
   GetProbabilitiesMatrix(parameters, probabilities, start, batchSize);
 
   // Calculate the log likelihood and regularization terms.
-  double logLikelihood, weightDecay;
+  ElemType logLikelihood, weightDecay;
 
   logLikelihood = arma::accu(groundTruth.cols(start, start + batchSize - 1) %
-      arma::log(probabilities)) / batchSize;
-  weightDecay = 0.5 * lambda * arma::accu(parameters * parameters);
+      log(probabilities)) / batchSize;
+  weightDecay = ((typename MatType::elem_type) 0.5) * lambda *
+      norm(vectorise(parameters), 2);
 
   return -logLikelihood + weightDecay;
 }
@@ -241,8 +260,11 @@ inline double SoftmaxRegressionFunction:
 /**
  * Calculates and stores the gradient values given a set of parameters.
  */
-inline void SoftmaxRegressionFunction::Gradient(const arma::mat& parameters,
-                                                arma::mat& gradient) const
+template<typename MatType>
+template<typename GradType>
+inline void SoftmaxRegressionFunction<MatType>::Gradient(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters,
+    GradType& gradient) const
 {
   // Calculate the class probabilities for each training example. The
   // probabilities for each of the classes are given by:
@@ -250,7 +272,7 @@ inline void SoftmaxRegressionFunction::G
   // The sum is calculated over all the classes.
   // x_i is the input vector for a particular training example.
   // theta_j is the parameter vector associated with a particular class.
-  arma::mat probabilities;
+  DenseMatType probabilities;
   GetProbabilitiesMatrix(parameters, probabilities, 0, data.n_cols);
 
   // Calculate the parameter gradients.
@@ -259,13 +281,13 @@ inline void SoftmaxRegressionFunction::G
   {
     // Treating the intercept term parameters.col(0) seperately to avoid
     // the cost of building matrix [1; data].
-    arma::mat inner = probabilities - groundTruth;
+    DenseMatType inner = probabilities - groundTruth;
     gradient.col(0) =
-      inner * arma::ones<arma::mat>(data.n_cols, 1) / data.n_cols +
-      lambda * parameters.col(0);
+        inner * ones<DenseMatType>(data.n_cols, 1) / data.n_cols +
+        lambda * parameters.col(0);
     gradient.cols(1, parameters.n_cols - 1) =
-      inner * data.t() / data.n_cols +
-      lambda * parameters.cols(1, parameters.n_cols - 1);
+        inner * data.t() / data.n_cols +
+        lambda * parameters.cols(1, parameters.n_cols - 1);
   }
   else
   {
@@ -274,23 +296,25 @@ inline void SoftmaxRegressionFunction::G
   }
 }
 
-inline void SoftmaxRegressionFunction::Gradient(const arma::mat& parameters,
-                                                const size_t start,
-                                                arma::mat& gradient,
-                                                const size_t batchSize)
-    const
+template<typename MatType>
+template<typename GradType>
+inline void SoftmaxRegressionFunction<MatType>::Gradient(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters,
+    const size_t start,
+    GradType& gradient,
+    const size_t batchSize) const
 {
-  arma::mat probabilities;
+  DenseMatType probabilities;
   GetProbabilitiesMatrix(parameters, probabilities, start, batchSize);
 
   // Calculate the parameter gradients.
   gradient.set_size(parameters.n_rows, parameters.n_cols);
   if (fitIntercept)
   {
-    arma::mat inner = probabilities - groundTruth.cols(start, start +
+    DenseMatType inner = probabilities - groundTruth.cols(start, start +
         batchSize - 1);
     gradient.col(0) =
-        inner * arma::ones<arma::mat>(batchSize, 1) / batchSize +
+        inner * ones<DenseMatType>(batchSize, 1) / batchSize +
         lambda * parameters.col(0);
     gradient.cols(1, parameters.n_cols - 1) =
         inner * data.cols(start, start + batchSize - 1).t() / batchSize +
@@ -304,24 +328,26 @@ inline void SoftmaxRegressionFunction::G
   }
 }
 
-inline void SoftmaxRegressionFunction::PartialGradient(const arma::mat& parameters,
-                                                       const size_t j,
-                                                       arma::sp_mat& gradient) 
-    const
+template<typename MatType>
+template<typename GradType>
+inline void SoftmaxRegressionFunction<MatType>::PartialGradient(
+    const typename SoftmaxRegressionFunction<MatType>::DenseMatType& parameters,
+    const size_t j,
+    GradType& gradient) const
 {
   gradient.zeros(arma::size(parameters));
 
-  arma::mat probabilities;
+  DenseMatType probabilities;
   GetProbabilitiesMatrix(parameters, probabilities, 0, data.n_cols);
 
   // Calculate the required part of the gradient.
-  arma::mat inner = probabilities - groundTruth;
+  DenseMatType inner = probabilities - groundTruth;
   if (fitIntercept)
   {
     if (j == 0)
     {
       gradient.col(j) =
-          inner * arma::ones<arma::mat>(data.n_cols, 1) / data.n_cols +
+          inner * ones<DenseMatType>(data.n_cols, 1) / data.n_cols +
           lambda * parameters.col(0);
     }
     else
diff -pruN 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_impl.hpp 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_impl.hpp
--- 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,85 +17,188 @@
 
 namespace mlpack {
 
-template<typename OptimizerType>
-SoftmaxRegression::SoftmaxRegression(
-    const arma::mat& data,
+template<typename MatType>
+inline SoftmaxRegression<MatType>::SoftmaxRegression(
+    const size_t inputSize,
+    const size_t numClasses,
+    const bool fitIntercept) :
+    numClasses(numClasses),
+    lambda(0.0001),
+    fitIntercept(fitIntercept)
+{
+  SoftmaxRegressionFunction<MatType>::InitializeWeights(
+      parameters, inputSize, numClasses, fitIntercept);
+}
+
+template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+SoftmaxRegression<MatType>::SoftmaxRegression(
+    const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
     const double lambda,
     const bool fitIntercept,
-    OptimizerType optimizer) :
+    CallbackTypes&&... callbacks) :
     numClasses(numClasses),
     lambda(lambda),
     fitIntercept(fitIntercept)
 {
-  Train(data, labels, numClasses, optimizer);
+  OptimizerType optimizer;
+  Train(data, labels, numClasses, lambda, fitIntercept, std::move(optimizer),
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
-template<typename OptimizerType, typename... CallbackTypes>
-SoftmaxRegression::SoftmaxRegression(
-    const arma::mat& data,
+template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+SoftmaxRegression<MatType>::SoftmaxRegression(
+    const MatType& data,
     const arma::Row<size_t>& labels,
     const size_t numClasses,
+    OptimizerType& optimizer,
     const double lambda,
     const bool fitIntercept,
-    OptimizerType optimizer,
     CallbackTypes&&... callbacks) :
     numClasses(numClasses),
     lambda(lambda),
     fitIntercept(fitIntercept)
 {
-  Train(data, labels, numClasses, optimizer, callbacks...);
+  Train(data, labels, numClasses, lambda, fitIntercept, std::move(optimizer),
+      std::forward<CallbackTypes>(callbacks)...);
 }
 
-inline SoftmaxRegression::SoftmaxRegression(
-    const size_t inputSize,
-    const size_t numClasses,
-    const bool fitIntercept) :
-    numClasses(numClasses),
-    lambda(0.0001),
-    fitIntercept(fitIntercept)
+template<typename MatType>
+template<typename OptimizerType,
+         typename FirstCallbackType,
+         typename... CallbackTypes,
+         typename, typename, typename>
+mlpack_deprecated /** To be removed in mlpack 5.0.0; use the overload below. */
+double SoftmaxRegression<MatType>::Train(const MatType& data,
+                                         const arma::Row<size_t>& labels,
+                                         const size_t numClasses,
+                                         OptimizerType optimizer,
+                                         FirstCallbackType&& firstCallback,
+                                         CallbackTypes&&... callbacks)
 {
-  SoftmaxRegressionFunction::InitializeWeights(
-      parameters, inputSize, numClasses, fitIntercept);
+  SoftmaxRegressionFunction<MatType> regressor(data, labels, numClasses, lambda,
+                                               fitIntercept);
+  if (parameters.n_elem != regressor.GetInitialPoint().n_elem)
+    parameters = regressor.GetInitialPoint();
+
+  // Train the model.
+  const double out = optimizer.Optimize(regressor, parameters, firstCallback,
+      callbacks...);
+  this->numClasses = numClasses;
+
+  Log::Info << "SoftmaxRegression::SoftmaxRegression(): final objective of "
+            << "trained model is " << out << "." << std::endl;
+
+  return out;
 }
 
-inline void SoftmaxRegression::Classify(const arma::mat& dataset,
-                                        arma::Row<size_t>& labels)
-    const
+template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+typename SoftmaxRegression<MatType>::ElemType
+SoftmaxRegression<MatType>::Train(const MatType& data,
+                                  const arma::Row<size_t>& labels,
+                                  const size_t numClasses,
+                                  const double lambda,
+                                  const bool fitIntercept,
+                                  CallbackTypes&&... callbacks)
 {
-  arma::mat probabilities;
-  Classify(dataset, probabilities);
+  OptimizerType optimizer;
+  return Train(data, labels, numClasses, optimizer, lambda, fitIntercept,
+      std::forward<CallbackTypes>(callbacks)...);
+}
 
-  // Prepare necessary data.
-  labels.zeros(dataset.n_cols);
-  double maxProbability = 0;
+template<typename MatType>
+template<typename OptimizerType, typename... CallbackTypes, typename, typename>
+typename SoftmaxRegression<MatType>::ElemType
+SoftmaxRegression<MatType>::Train(const MatType& data,
+                                  const arma::Row<size_t>& labels,
+                                  const size_t numClasses,
+                                  OptimizerType& optimizer,
+                                  const double lambda,
+                                  const bool fitIntercept,
+                                  CallbackTypes&&... callbacks)
+{
+  this->lambda = lambda;
+  this->fitIntercept = fitIntercept;
+  this->numClasses = numClasses;
 
-  // For each test input.
-  for (size_t i = 0; i < dataset.n_cols; ++i)
-  {
-    // For each class.
-    for (size_t j = 0; j < numClasses; ++j)
-    {
-      // If a higher class probability is encountered, change prediction.
-      if (probabilities(j, i) > maxProbability)
-      {
-        maxProbability = probabilities(j, i);
-        labels(i) = j;
-      }
-    }
+  SoftmaxRegressionFunction<MatType> regressor(data, labels, numClasses, lambda,
+                                               fitIntercept);
+  if (parameters.n_elem != regressor.GetInitialPoint().n_elem)
+    parameters = regressor.GetInitialPoint();
 
-    // Set maximum probability to zero for the next input.
-    maxProbability = 0;
-  }
+  // Train the model.
+  const double out = optimizer.Optimize(regressor, parameters, callbacks...);
+  this->numClasses = numClasses;
+
+  Log::Info << "SoftmaxRegression::SoftmaxRegression(): final objective of "
+            << "trained model is " << out << "." << std::endl;
+
+  return out;
 }
 
-inline void SoftmaxRegression::Classify(const arma::mat& dataset,
-                                        arma::Row<size_t>& labels,
-                                        arma::mat& probabilities)
+template<typename MatType>
+inline void SoftmaxRegression<MatType>::Classify(const MatType& dataset,
+                                                 arma::Row<size_t>& labels)
     const
 {
-  Classify(dataset, probabilities);
+  MatType probabilities;
+  Classify(dataset, labels, probabilities);
+}
+
+template<typename MatType>
+template<typename VecType>
+size_t SoftmaxRegression<MatType>::Classify(const VecType& point) const
+{
+  arma::Row<size_t> label(1);
+  Classify(point, label);
+  return size_t(label(0));
+}
+
+template<typename MatType>
+template<typename VecType>
+void SoftmaxRegression<MatType>::Classify(
+    const VecType& point,
+    size_t& prediction,
+    typename SoftmaxRegression<MatType>::DenseColType& probabilitiesVec) const
+{
+  arma::Row<size_t> label(1);
+  Classify(point, label, probabilitiesVec);
+  prediction = label[0];
+}
+
+template<typename MatType>
+inline void SoftmaxRegression<MatType>::Classify(
+    const MatType& dataset,
+    arma::Row<size_t>& labels,
+    typename SoftmaxRegression<MatType>::DenseMatType& probabilities) const
+{
+  util::CheckSameDimensionality(dataset, FeatureSize(),
+      "SoftmaxRegression::Classify()");
+
+  // Calculate the probabilities for each test input.
+  DenseMatType hypothesis;
+  if (fitIntercept)
+  {
+    // In order to add the intercept term, we should compute following matrix:
+    //     [1; data] = join_cols(ones(1, data.n_cols), data)
+    //     hypothesis = exp(parameters * [1; data]).
+    //
+    // Since the cost of join maybe high due to the copy of original data,
+    // split the hypothesis computation to two components.
+    hypothesis = exp(
+      repmat(parameters.col(0), 1, dataset.n_cols) +
+      parameters.cols(1, parameters.n_cols - 1) * dataset);
+  }
+  else
+  {
+    hypothesis = exp(parameters * dataset);
+  }
+
+  probabilities = hypothesis / repmat(sum(hypothesis, 0), numClasses, 1);
 
   // Prepare necessary data.
   labels.zeros(dataset.n_cols);
@@ -120,46 +223,18 @@ inline void SoftmaxRegression::Classify(
   }
 }
 
-inline void SoftmaxRegression::Classify(const arma::mat& dataset,
-                                        arma::mat& probabilities)
-    const
+template<typename MatType>
+inline mlpack_deprecated void SoftmaxRegression<MatType>::Classify(
+    const MatType& dataset,
+    typename SoftmaxRegression<MatType>::DenseMatType& probabilities) const
 {
-  util::CheckSameDimensionality(dataset, FeatureSize(),
-      "SoftmaxRegression::Classify()");
-
-  // Calculate the probabilities for each test input.
-  arma::mat hypothesis;
-  if (fitIntercept)
-  {
-    // In order to add the intercept term, we should compute following matrix:
-    //     [1; data] = arma::join_cols(ones(1, data.n_cols), data)
-    //     hypothesis = arma::exp(parameters * [1; data]).
-    //
-    // Since the cost of join maybe high due to the copy of original data,
-    // split the hypothesis computation to two components.
-    hypothesis = arma::exp(
-      arma::repmat(parameters.col(0), 1, dataset.n_cols) +
-      parameters.cols(1, parameters.n_cols - 1) * dataset);
-  }
-  else
-  {
-    hypothesis = arma::exp(parameters * dataset);
-  }
-
-  probabilities = hypothesis / arma::repmat(arma::sum(hypothesis, 0),
-                                            numClasses, 1);
-}
-
-template<typename VecType>
-size_t SoftmaxRegression::Classify(const VecType& point) const
-{
-  arma::Row<size_t> label(1);
-  Classify(point, label);
-  return size_t(label(0));
+  arma::Row<size_t> labels;
+  Classify(dataset, labels, probabilities);
 }
 
-inline double SoftmaxRegression::ComputeAccuracy(
-    const arma::mat& testData,
+template<typename MatType>
+inline double SoftmaxRegression<MatType>::ComputeAccuracy(
+    const MatType& testData,
     const arma::Row<size_t>& labels) const
 {
   arma::Row<size_t> predictions;
@@ -177,47 +252,36 @@ inline double SoftmaxRegression::Compute
   return (count * 100.0) / predictions.n_elem;
 }
 
-template<typename OptimizerType>
-double SoftmaxRegression::Train(const arma::mat& data,
-                                const arma::Row<size_t>& labels,
-                                const size_t numClasses,
-                                OptimizerType optimizer)
+template<typename MatType>
+void SoftmaxRegression<MatType>::Reset()
 {
-  SoftmaxRegressionFunction regressor(data, labels, numClasses, lambda,
-                                      fitIntercept);
-  if (parameters.n_elem != regressor.GetInitialPoint().n_elem)
-    parameters = regressor.GetInitialPoint();
-
-  // Train the model.
-  const double out = optimizer.Optimize(regressor, parameters);
-  this->numClasses = numClasses;
-
-  Log::Info << "SoftmaxRegression::SoftmaxRegression(): final objective of "
-            << "trained model is " << out << "." << std::endl;
-
-  return out;
+  SoftmaxRegressionFunction<MatType>::InitializeWeights(
+      parameters, parameters.n_cols, parameters.n_rows, fitIntercept);
 }
 
-template<typename OptimizerType, typename... CallbackTypes>
-double SoftmaxRegression::Train(const arma::mat& data,
-                                const arma::Row<size_t>& labels,
-                                const size_t numClasses,
-                                OptimizerType optimizer,
-                                CallbackTypes&&... callbacks)
+template<typename MatType>
+template<typename Archive>
+void SoftmaxRegression<MatType>::serialize(Archive& ar,
+                                           const unsigned int version)
 {
-  SoftmaxRegressionFunction regressor(data, labels, numClasses, lambda,
-                                      fitIntercept);
-  if (parameters.n_elem != regressor.GetInitialPoint().n_elem)
-    parameters = regressor.GetInitialPoint();
-
-  // Train the model.
-  const double out = optimizer.Optimize(regressor, parameters, callbacks...);
-  this->numClasses = numClasses;
-
-  Log::Info << "SoftmaxRegression::SoftmaxRegression(): final objective of "
-            << "trained model is " << out << "." << std::endl;
-
-  return out;
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // This is the legacy version: `parameters` is of type arma::mat.
+    arma::mat parametersTmp;
+    ar(cereal::make_nvp("parameters", parametersTmp));
+    parameters = ConvTo<DenseMatType>::From(parametersTmp);
+
+    ar(CEREAL_NVP(numClasses));
+    ar(CEREAL_NVP(lambda));
+    ar(CEREAL_NVP(fitIntercept));
+  }
+  else
+  {
+    ar(CEREAL_NVP(parameters));
+    ar(CEREAL_NVP(numClasses));
+    ar(CEREAL_NVP(lambda));
+    ar(CEREAL_NVP(fitIntercept));
+  }
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_main.cpp 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_main.cpp
--- 4.3.0-2/src/mlpack/methods/softmax_regression/softmax_regression_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/softmax_regression/softmax_regression_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -100,9 +100,9 @@ PARAM_UROW_IN("labels", "A matrix contai
     "in the training set (y). The labels must order as a row.", "l");
 
 // Model loading/saving.
-PARAM_MODEL_IN(SoftmaxRegression, "input_model", "File containing existing "
+PARAM_MODEL_IN(SoftmaxRegression<>, "input_model", "File containing existing "
     "model (parameters).", "m");
-PARAM_MODEL_OUT(SoftmaxRegression, "output_model", "File to save trained "
+PARAM_MODEL_OUT(SoftmaxRegression<>, "output_model", "File to save trained "
     "softmax regression model to.", "M");
 
 // Testing.
@@ -171,12 +171,12 @@ void BINDING_FUNCTION(util::Params& para
   RequireAtLeastOnePassed(params, { "output_model", "predictions" }, false,
       "no results will be saved");
 
-  SoftmaxRegression* sm = TrainSoftmax<SoftmaxRegression>(params, timers,
+  SoftmaxRegression<>* sm = TrainSoftmax<SoftmaxRegression<>>(params, timers,
       maxIterations);
 
   TestClassifyAcc(params, timers, sm->NumClasses(), *sm);
 
-  params.Get<SoftmaxRegression*>("output_model") = sm;
+  params.Get<SoftmaxRegression<>*>("output_model") = sm;
 }
 
 size_t CalculateNumberOfClasses(const size_t numClasses,
@@ -215,8 +215,9 @@ void TestClassifyAcc(util::Params& param
   arma::mat testData = std::move(params.Get<arma::mat>("test"));
 
   arma::Row<size_t> predictLabels;
+  arma::mat probabilities;
   timers.Start("softmax_regression_classification");
-  model.Classify(testData, predictLabels);
+  model.Classify(testData, predictLabels, probabilities);
   timers.Stop("softmax_regression_classification");
 
   // Calculate accuracy, if desired.
@@ -261,15 +262,9 @@ void TestClassifyAcc(util::Params& param
   if (params.Has("predictions"))
     params.Get<arma::Row<size_t>>("predictions") = std::move(predictLabels);
 
-  // Compute probabiltiies, if desired.
+  // Save probabilities, if desired.
   if (params.Has("probabilities"))
-  {
-    Log::Info << "Calculating class probabilities of points in '"
-        << params.GetPrintable<arma::mat>("test") << "'." << endl;
-    arma::mat probabilities;
-    model.Classify(testData, probabilities);
     params.Get<arma::mat>("probabilities") = std::move(probabilities);
-  }
 }
 
 template<typename Model>
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_autoencoder/maximal_inputs_impl.hpp 4.4.0-1/src/mlpack/methods/sparse_autoencoder/maximal_inputs_impl.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_autoencoder/maximal_inputs_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_autoencoder/maximal_inputs_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -20,8 +20,7 @@ inline void MaximalInputs(const arma::ma
 {
   arma::mat paramTemp(parameters.submat(0, 0, (parameters.n_rows - 1) / 2 - 1,
                                         parameters.n_cols - 2).t());
-  double const mean = arma::mean(arma::mean(paramTemp));
-  paramTemp -= mean;
+  paramTemp -= arma::mean(arma::mean(paramTemp));
 
   NormalizeColByMax(paramTemp, output);
 }
@@ -32,10 +31,10 @@ inline void NormalizeColByMax(const arma
   output.set_size(input.n_rows, input.n_cols);
   for (arma::uword i = 0; i != input.n_cols; ++i)
   {
-    const double max = arma::max(arma::abs(input.col(i)));
-    if (max != 0.0)
+    const double colMax = max(arma::abs(input.col(i)));
+    if (colMax != 0.0)
     {
-      output.col(i) = input.col(i) / max;
+      output.col(i) = input.col(i) / colMax;
     }
     else
     {
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -133,7 +133,7 @@ class SparseAutoencoder
    */
   void Sigmoid(const arma::mat& x, arma::mat& output) const
   {
-    output = (1.0 / (1 + arma::exp(-x)));
+    output = (1.0 / (1 + exp(-x)));
   }
 
   //! Sets size of the visible layer.
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function.hpp 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -78,7 +78,7 @@ class SparseAutoencoderFunction
    */
   void Sigmoid(const arma::mat& x, arma::mat& output) const
   {
-    output = (1.0 / (1 + arma::exp(-x)));
+    output = (1.0 / (1 + exp(-x)));
   }
 
   //! Return the initial point for the optimization.
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function_impl.hpp 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -64,7 +64,7 @@ inline const arma::mat SparseAutoencoder
 
   // Decide the parameter 'r' depending on the size of the visible and hidden
   // layers. The formula used is r = sqrt(6) / sqrt(vSize + hSize + 1).
-  const double range = sqrt(6) / sqrt(visibleSize + hiddenSize + 1);
+  const double range = std::sqrt(6) / std::sqrt(visibleSize + hiddenSize + 1);
 
   // Shift range of w1 and w2 values from [0, 1] to [-r, r].
   parameters.submat(0, 0, 2 * hiddenSize - 1, visibleSize - 1) = 2 * range *
@@ -103,24 +103,24 @@ inline double SparseAutoencoderFunction:
 
   // Compute activations of the hidden and output layers.
   Sigmoid(parameters.submat(0, 0, l1 - 1, l2 - 1) * data +
-      arma::repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
+      repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
       hiddenLayer);
 
   Sigmoid(parameters.submat(l1, 0, l3 - 1, l2 - 1).t() * hiddenLayer +
-      arma::repmat(parameters.submat(l3, 0, l3, l2 - 1).t(), 1, data.n_cols),
+      repmat(parameters.submat(l3, 0, l3, l2 - 1).t(), 1, data.n_cols),
       outputLayer);
 
   arma::mat rhoCap, diff;
 
   // Average activations of the hidden layer.
-  rhoCap = arma::sum(hiddenLayer, 1) / data.n_cols;
+  rhoCap = sum(hiddenLayer, 1) / data.n_cols;
   // Difference between the reconstructed data and the original data.
   diff = outputLayer - data;
 
   double wL2SquaredNorm;
 
   // Calculate squared L2-norms of w1 and w2.
-  wL2SquaredNorm = arma::accu(parameters.submat(0, 0, l3 - 1, l2 - 1) %
+  wL2SquaredNorm = accu(parameters.submat(0, 0, l3 - 1, l2 - 1) %
       parameters.submat(0, 0, l3 - 1, l2 - 1));
 
   double sumOfSquaresError, weightDecay, klDivergence, cost;
@@ -131,10 +131,10 @@ inline double SparseAutoencoderFunction:
   // of the weights w1 and w2. 'klDivergence' is the cost of the hidden layer
   // activations not being low. It is given by the following formula:
   // KL = sum_over_hSize(rho*log(rho/rhoCaq) + (1-rho)*log((1-rho)/(1-rhoCap)))
-  sumOfSquaresError = 0.5 * arma::accu(diff % diff) / data.n_cols;
+  sumOfSquaresError = 0.5 * accu(diff % diff) / data.n_cols;
   weightDecay = 0.5 * lambda * wL2SquaredNorm;
-  klDivergence = beta * arma::accu(rho * arma::log(rho / rhoCap) + (1 - rho) *
-      arma::log((1 - rho) / (1 - rhoCap)));
+  klDivergence = beta * accu(rho * log(rho / rhoCap) + (1 - rho) *
+      log((1 - rho) / (1 - rhoCap)));
 
   // The cost is the sum of the terms calculated above.
   cost = sumOfSquaresError + weightDecay + klDivergence;
@@ -170,17 +170,17 @@ inline void SparseAutoencoderFunction::G
 
   // Compute activations of the hidden and output layers.
   Sigmoid(parameters.submat(0, 0, l1 - 1, l2 - 1) * data +
-      arma::repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
+      repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
       hiddenLayer);
 
   Sigmoid(parameters.submat(l1, 0, l3 - 1, l2 - 1).t() * hiddenLayer +
-      arma::repmat(parameters.submat(l3, 0, l3, l2 - 1).t(), 1, data.n_cols),
+      repmat(parameters.submat(l3, 0, l3, l2 - 1).t(), 1, data.n_cols),
       outputLayer);
 
   arma::mat rhoCap, diff;
 
   // Average activations of the hidden layer.
-  rhoCap = arma::sum(hiddenLayer, 1) / data.n_cols;
+  rhoCap = sum(hiddenLayer, 1) / data.n_cols;
   // Difference between the reconstructed data and the original data.
   diff = outputLayer - data;
 
@@ -195,7 +195,7 @@ inline void SparseAutoencoderFunction::G
   klDivGrad = beta * (-(rho / rhoCap) + (1 - rho) / (1 - rhoCap));
   delOut = diff % outputLayer % (1 - outputLayer);
   delHid = (parameters.submat(l1, 0, l3 - 1, l2 - 1) * delOut +
-      arma::repmat(klDivGrad, 1, data.n_cols)) % hiddenLayer %
+      repmat(klDivGrad, 1, data.n_cols)) % hiddenLayer %
       (1 - hiddenLayer);
 
   gradient.zeros(2 * hiddenSize + 1, visibleSize + 1);
@@ -208,8 +208,8 @@ inline void SparseAutoencoderFunction::G
   gradient.submat(l1, 0, l3 - 1, l2 - 1) =
       (delOut * hiddenLayer.t() / data.n_cols +
       lambda * parameters.submat(l1, 0, l3 - 1, l2 - 1).t()).t();
-  gradient.submat(0, l2, l1 - 1, l2) = arma::sum(delHid, 1) / data.n_cols;
-  gradient.submat(l3, 0, l3, l2 - 1) = (arma::sum(delOut, 1) / data.n_cols).t();
+  gradient.submat(0, l2, l1 - 1, l2) = sum(delHid, 1) / data.n_cols;
+  gradient.submat(l3, 0, l3, l2 - 1) = (sum(delOut, 1) / data.n_cols).t();
 }
 
 } // namespace mlpack
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_impl.hpp 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_impl.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_autoencoder/sparse_autoencoder_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -78,7 +78,7 @@ inline void SparseAutoencoder::GetNewFea
   const size_t l2 = visibleSize;
 
   Sigmoid(parameters.submat(0, 0, l1 - 1, l2 - 1) * data +
-      arma::repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
+      repmat(parameters.submat(0, l2, l1 - 1, l2), 1, data.n_cols),
       features);
 }
 
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/data_dependent_random_initializer.hpp 4.4.0-1/src/mlpack/methods/sparse_coding/data_dependent_random_initializer.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/data_dependent_random_initializer.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/data_dependent_random_initializer.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,9 +34,10 @@ class DataDependentRandomInitializer
    * @param atoms Number of atoms in dictionary.
    * @param dictionary Dictionary to initialize.
    */
-  static void Initialize(const arma::mat& data,
+  template<typename MatType>
+  static void Initialize(const MatType& data,
                          const size_t atoms,
-                         arma::mat& dictionary)
+                         MatType& dictionary)
   {
     // Set the size of the dictionary.
     dictionary.set_size(data.n_rows, atoms);
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/nothing_initializer.hpp 4.4.0-1/src/mlpack/methods/sparse_coding/nothing_initializer.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/nothing_initializer.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/nothing_initializer.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,9 +31,10 @@ class NothingInitializer
    * for SparseCoding if the dictionary is not set manually before running the
    * method.
    */
-  static void Initialize(const arma::mat& /* data */,
+  template<typename MatType>
+  static void Initialize(const MatType& /* data */,
                          const size_t /* atoms */,
-                         arma::mat& /* dictionary */)
+                         MatType& /* dictionary */)
   {
     // Do nothing!
   }
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/random_initializer.hpp 4.4.0-1/src/mlpack/methods/sparse_coding/random_initializer.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/random_initializer.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/random_initializer.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -33,9 +33,10 @@ class RandomInitializer
    * @param atoms Number of atoms (columns) in the dictionary.
    * @param dictionary Dictionary to initialize.
    */
-  static void Initialize(const arma::mat& data,
+  template<typename MatType>
+  static void Initialize(const MatType& data,
                          const size_t atoms,
-                         arma::mat& dictionary)
+                         MatType& dictionary)
   {
     // Create random dictionary.
     dictionary.randn(data.n_rows, atoms);
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding.hpp 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -108,12 +108,16 @@ namespace mlpack {
  * the Encode() function.
  *
  * @tparam DictionaryInitializationPolicy The class to use to initialize the
- *     dictionary; must have 'void Initialize(const arma::mat& data, arma::mat&
+ *     dictionary; must have 'void Initialize(const MatType& data, MatType&
  *     dictionary)' function.
  */
+template<typename MatType = arma::mat>
 class SparseCoding
 {
  public:
+  typedef typename GetColType<MatType>::type ColType;
+  typedef typename GetRowType<MatType>::type RowType;
+
   /**
    * Set the parameters to SparseCoding.  lambda2 defaults to 0.  This
    * constructor will train the model.  If that is not desired, call the other
@@ -123,7 +127,7 @@ class SparseCoding
    *
    * If you want to initialize the dictionary to a custom matrix, consider
    * either writing your own DictionaryInitializer class (with void
-   * Initialize(const arma::mat& data, arma::mat& dictionary) function), or call
+   * Initialize(const MatType& data, MatType& dictionary) function), or call
    * the constructor that does not take a data matrix, then call Dictionary() to
    * set the dictionary matrix to a matrix of your choosing, and then call
    * Train() with NothingInitializer (i.e. Train<NothingInitializer>(data)).
@@ -142,7 +146,7 @@ class SparseCoding
    * @param initializer The initializer to use.
    */
   template<typename DictionaryInitializer = DataDependentRandomInitializer>
-  SparseCoding(const arma::mat& data,
+  SparseCoding(const MatType& data,
                const size_t atoms,
                const double lambda1,
                const double lambda2 = 0,
@@ -180,7 +184,7 @@ class SparseCoding
    * @return The final objective value.
    */
   template<typename DictionaryInitializer = DataDependentRandomInitializer>
-  double Train(const arma::mat& data,
+  double Train(const MatType& data,
                const DictionaryInitializer& initializer =
                    DictionaryInitializer());
 
@@ -191,7 +195,7 @@ class SparseCoding
    * @param data Input data matrix to be encoded.
    * @param codes Output codes matrix.
    */
-  void Encode(const arma::mat& data, arma::mat& codes);
+  void Encode(const MatType& data, MatType& codes);
 
   /**
    * Learn dictionary via Newton method based on Lagrange dual.
@@ -204,8 +208,8 @@ class SparseCoding
    * @return the norm of the gradient of the Lagrange dual with respect to
    *    the dual variables
    */
-  double OptimizeDictionary(const arma::mat& data,
-                            const arma::mat& codes,
+  double OptimizeDictionary(const MatType& data,
+                            const MatType& codes,
                             const arma::uvec& adjacencies);
 
   /**
@@ -216,12 +220,12 @@ class SparseCoding
   /**
    * Compute the objective function.
    */
-  double Objective(const arma::mat& data, const arma::mat& codes) const;
+  double Objective(const MatType& data, const MatType& codes) const;
 
   //! Access the dictionary.
-  const arma::mat& Dictionary() const { return dictionary; }
+  const MatType& Dictionary() const { return dictionary; }
   //! Modify the dictionary.
-  arma::mat& Dictionary() { return dictionary; }
+  MatType& Dictionary() { return dictionary; }
 
   //! Access the number of atoms.
   size_t Atoms() const { return atoms; }
@@ -262,7 +266,7 @@ class SparseCoding
   size_t atoms;
 
   //! Dictionary (columns are atoms).
-  arma::mat dictionary;
+  MatType dictionary;
 
   //! l1 regularization term.
   double lambda1;
@@ -279,6 +283,9 @@ class SparseCoding
 
 } // namespace mlpack
 
+CEREAL_TEMPLATE_CLASS_VERSION((typename MatType),
+    (mlpack::SparseCoding<MatType>), (1));
+
 // Include implementation.
 #include "sparse_coding_impl.hpp"
 
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding_impl.hpp 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding_impl.hpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,9 +18,10 @@
 
 namespace mlpack {
 
+template<typename MatType>
 template<typename DictionaryInitializer>
-SparseCoding::SparseCoding(
-    const arma::mat& data,
+SparseCoding<MatType>::SparseCoding(
+    const MatType& data,
     const size_t atoms,
     const double lambda1,
     const double lambda2,
@@ -38,7 +39,8 @@ SparseCoding::SparseCoding(
   Train(data, initializer);
 }
 
-inline SparseCoding::SparseCoding(
+template<typename MatType>
+inline SparseCoding<MatType>::SparseCoding(
     const size_t atoms,
     const double lambda1,
     const double lambda2,
@@ -55,12 +57,13 @@ inline SparseCoding::SparseCoding(
   // Nothing to do.
 }
 
-inline void SparseCoding::Encode(const arma::mat& data,
-                                 arma::mat& codes)
+template<typename MatType>
+inline void SparseCoding<MatType>::Encode(const MatType& data,
+                                          MatType& codes)
 {
   // When using the Cholesky version of LARS, this is correct even if
   // lambda2 > 0.
-  arma::mat matGram = trans(dictionary) * dictionary;
+  MatType matGram = trans(dictionary) * dictionary;
 
   codes.set_size(atoms, data.n_cols);
   for (size_t i = 0; i < data.n_cols; ++i)
@@ -71,25 +74,28 @@ inline void SparseCoding::Encode(const a
 
     bool useCholesky = true;
     // Intercept fitting and data normalization is disabled.
-    LARS lars(useCholesky, matGram, lambda1, lambda2,
+    LARS<MatType> lars(useCholesky, lambda1, lambda2,
         1e-16 /* default tolerance */, false, false);
 
     // Create an alias of the code (using the same memory), and then LARS will
     // place the result directly into that; then we will not need to have an
     // extra copy.
-    arma::vec code = codes.unsafe_col(i);
-    arma::rowvec responses = data.unsafe_col(i).t();
-    lars.Train(dictionary, responses, code, false);
+    ColType code = codes.unsafe_col(i);
+    RowType responses = data.unsafe_col(i).t();
+    lars.Train(dictionary, responses, false, useCholesky, matGram);
+    code = lars.Beta();
   }
 }
 
 // Dictionary step for optimization.
-inline double SparseCoding::OptimizeDictionary(const arma::mat& data,
-                                               const arma::mat& codes,
-                                               const arma::uvec& adjacencies)
+template<typename MatType>
+inline double SparseCoding<MatType>::OptimizeDictionary(
+    const MatType& data,
+    const MatType& codes,
+    const arma::uvec& adjacencies)
 {
   // Count the number of atomic neighbors for each point x^i.
-  arma::uvec neighborCounts = arma::zeros<arma::uvec>(data.n_cols, 1);
+  arma::uvec neighborCounts = zeros<arma::uvec>(data.n_cols, 1);
 
   if (adjacencies.n_elem > 0)
   {
@@ -112,22 +118,21 @@ inline double SparseCoding::OptimizeDict
   }
 
   // Handle the case of inactive atoms (atoms not used in the given coding).
-  std::vector<size_t> inactiveAtoms;
-
-  for (size_t j = 0; j < atoms; ++j)
+  std::vector<arma::uword> activeAtoms;
+  for (arma::uword j = 0; j < atoms; ++j)
   {
-    if (arma::accu(codes.row(j) != 0) == 0)
-      inactiveAtoms.push_back(j);
+    if (arma::any(codes.row(j) != 0))
+      activeAtoms.push_back(j);
   }
 
-  const size_t nInactiveAtoms = inactiveAtoms.size();
-  const size_t nActiveAtoms = atoms - nInactiveAtoms;
+  const size_t nActiveAtoms = activeAtoms.size();
+  const size_t nInactiveAtoms = atoms - nActiveAtoms;
 
   // Efficient construction of Z restricted to active atoms.
-  arma::mat matActiveZ;
+  MatType matActiveZ;
   if (nInactiveAtoms > 0)
   {
-    RemoveRows(codes, inactiveAtoms, matActiveZ);
+    matActiveZ = codes.rows(arma::uvec(activeAtoms));
   }
 
   if (nInactiveAtoms > 0)
@@ -142,7 +147,7 @@ inline double SparseCoding::OptimizeDict
   // multiplication with inv(A) seems to be unavoidable. Although more
   // expensive, the code written this way (we use solve()) should be more
   // numerically stable than just using inv(A) for everything.
-  arma::vec dualVars = arma::zeros<arma::vec>(nActiveAtoms);
+  ColType dualVars = zeros<ColType>(nActiveAtoms);
 
   // vec dualVars = 1e-14 * ones<vec>(nActiveAtoms);
 
@@ -159,10 +164,10 @@ inline double SparseCoding::OptimizeDict
   bool converged = false;
 
   // If we have any inactive atoms, we must construct these differently.
-  arma::mat codesXT;
-  arma::mat codesZT;
+  MatType codesXT;
+  MatType codesZT;
 
-  if (inactiveAtoms.empty())
+  if (nInactiveAtoms == 0)
   {
     codesXT = codes * trans(data);
     codesZT = codes * trans(codes);
@@ -177,16 +182,15 @@ inline double SparseCoding::OptimizeDict
   double improvement = 0;
   for (size_t t = 1; (t != maxIterations) && !converged; ++t)
   {
-    arma::mat A = codesZT + diagmat(dualVars);
+    MatType A = codesZT + diagmat(dualVars);
 
-    arma::mat matAInvZXT = solve(A, codesXT);
+    MatType matAInvZXT = solve(A, codesXT);
 
-    arma::vec gradient = -arma::sum(arma::square(matAInvZXT), 1);
-    gradient += 1;
+    ColType gradient = -sum(square(matAInvZXT), 1) + 1;
 
-    arma::mat hessian = -(-2 * (matAInvZXT * trans(matAInvZXT)) % inv(A));
+    MatType hessian = -(-2 * (matAInvZXT * trans(matAInvZXT)) % inv(A));
 
-    arma::vec searchDirection = -solve(hessian, gradient);
+    ColType searchDirection = -solve(hessian, gradient);
 
     // Armijo line search.
     const double c = 1e-4;
@@ -199,11 +203,11 @@ inline double SparseCoding::OptimizeDict
     while (true)
     {
       // Calculate objective.
-      double sumDualVars = arma::sum(dualVars);
+      double sumDualVars = sum(dualVars);
       double fOld = -(-trace(trans(codesXT) * matAInvZXT) - sumDualVars);
       double fNew = -(-trace(trans(codesXT) * solve(codesZT +
           diagmat(dualVars + alpha * searchDirection), codesXT)) -
-          (sumDualVars + alpha * arma::sum(searchDirection)));
+          (sumDualVars + alpha * sum(searchDirection)));
 
       if (fNew <= fOld + alpha * sufficientDecrease)
       {
@@ -217,7 +221,7 @@ inline double SparseCoding::OptimizeDict
 
     // Take step and print useful information.
     dualVars += searchDirection;
-    normGradient = arma::norm(gradient, 2);
+    normGradient = norm(gradient, 2);
     Log::Debug << "Newton Method iteration " << t << ":" << std::endl;
     Log::Debug << "  Gradient norm: " << std::scientific << normGradient
         << "." << std::endl;
@@ -227,36 +231,37 @@ inline double SparseCoding::OptimizeDict
       converged = true;
   }
 
-  if (inactiveAtoms.empty())
+  if (nInactiveAtoms == 0)
   {
     // Directly update dictionary.
     dictionary = trans(solve(codesZT + diagmat(dualVars), codesXT));
   }
   else
   {
-    arma::mat activeDictionary = trans(solve(codesZT +
+    MatType activeDictionary = trans(solve(codesZT +
         diagmat(dualVars), codesXT));
 
     // Update all atoms.
-    size_t currentInactiveIndex = 0;
+    size_t currentActiveIndex = 0;
     for (size_t i = 0; i < atoms; ++i)
     {
-      if (inactiveAtoms[currentInactiveIndex] == i)
+      if (currentActiveIndex >= activeAtoms.size() ||
+          activeAtoms[currentActiveIndex] != i)
       {
         // This atom is inactive.  Reinitialize it randomly.
         dictionary.col(i) = (data.col(RandInt(data.n_cols)) +
                              data.col(RandInt(data.n_cols)) +
                              data.col(RandInt(data.n_cols)));
 
-        dictionary.col(i) /= arma::norm(dictionary.col(i), 2);
-
-        // Increment inactive index counter.
-        ++currentInactiveIndex;
+        dictionary.col(i) /= norm(dictionary.col(i), 2);
       }
       else
       {
         // Update estimate.
-        dictionary.col(i) = activeDictionary.col(i - currentInactiveIndex);
+        dictionary.col(i) = activeDictionary.col(currentActiveIndex);
+
+        // Increment active index counter.
+        ++currentActiveIndex;
       }
     }
   }
@@ -265,11 +270,12 @@ inline double SparseCoding::OptimizeDict
 }
 
 // Project each atom of the dictionary back into the unit ball (if necessary).
-inline void SparseCoding::ProjectDictionary()
+template<typename MatType>
+inline void SparseCoding<MatType>::ProjectDictionary()
 {
   for (size_t j = 0; j < atoms; ++j)
   {
-    double atomNorm = arma::norm(dictionary.col(j), 2);
+    double atomNorm = norm(dictionary.col(j), 2);
     if (atomNorm > 1)
     {
       Log::Info << "Norm of atom " << j << " exceeds 1 (" << std::scientific
@@ -280,16 +286,17 @@ inline void SparseCoding::ProjectDiction
 }
 
 // Compute the objective function.
-inline double SparseCoding::Objective(const arma::mat& data, 
-                                      const arma::mat& codes)
+template<typename MatType>
+inline double SparseCoding<MatType>::Objective(const MatType& data,
+                                               const MatType& codes)
     const
 {
-  double l11NormZ = arma::sum(arma::sum(arma::abs(codes)));
-  double froNormResidual = arma::norm(data - (dictionary * codes), "fro");
+  double l11NormZ = sum(sum(arma::abs(codes)));
+  double froNormResidual = norm(data - (dictionary * codes), "fro");
 
   if (lambda2 > 0)
   {
-    double froNormZ = arma::norm(codes, "fro");
+    double froNormZ = norm(codes, "fro");
     return 0.5 * (std::pow(froNormResidual, 2.0) + (lambda2 *
         std::pow(froNormZ, 2.0))) + (lambda1 * l11NormZ);
   }
@@ -299,9 +306,10 @@ inline double SparseCoding::Objective(co
   }
 }
 
+template<typename MatType>
 template<typename DictionaryInitializer>
-double SparseCoding::Train(
-    const arma::mat& data,
+double SparseCoding<MatType>::Train(
+    const MatType& data,
     const DictionaryInitializer& initializer)
 {
   // Now, train.
@@ -315,7 +323,7 @@ double SparseCoding::Train(
   // optimization loop.
   Log::Info << "Initial coding step." << std::endl;
 
-  arma::mat codes(atoms, data.n_cols);
+  MatType codes(atoms, data.n_cols);
   Encode(data, codes);
   arma::uvec adjacencies = find(codes);
 
@@ -367,11 +375,24 @@ double SparseCoding::Train(
   return lastObjVal;
 }
 
+template<typename MatType>
 template<typename Archive>
-void SparseCoding::serialize(Archive& ar, const uint32_t /* version */)
+void SparseCoding<MatType>::serialize(Archive& ar, const uint32_t version)
 {
   ar(CEREAL_NVP(atoms));
-  ar(CEREAL_NVP(dictionary));
+
+  if (cereal::is_loading<Archive>() && version == 0)
+  {
+    // Older versions of SparseCoding always stored dictionary as an arma::mat.
+    arma::mat dictionaryTmp;
+    ar(cereal::make_nvp("dictionary", dictionaryTmp));
+    dictionary = ConvTo<MatType>::From(dictionaryTmp);
+  }
+  else
+  {
+    ar(CEREAL_NVP(dictionary));
+  }
+
   ar(CEREAL_NVP(lambda1));
   ar(CEREAL_NVP(lambda2));
   ar(CEREAL_NVP(maxIterations));
diff -pruN 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding_main.cpp 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding_main.cpp
--- 4.3.0-2/src/mlpack/methods/sparse_coding/sparse_coding_main.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/sparse_coding/sparse_coding_main.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -87,8 +87,8 @@ BINDING_SEE_ALSO("Sparse dictionary lear
 BINDING_SEE_ALSO("Efficient sparse coding algorithms (pdf)",
     "http://papers.nips.cc/paper/2979-efficient-sparse-coding-algorithms.pdf");
 BINDING_SEE_ALSO("Regularization and variable selection via the elastic net",
-    "http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.124.4696&rep=rep1"
-    "&type=pdf");
+    "https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf"
+    "&doi=46217f372a75dddc2254fdbc6b9418ba3554e453");
 BINDING_SEE_ALSO("SparseCoding C++ class documentation",
     "@src/mlpack/methods/sparse_coding/sparse_coding.hpp");
 
@@ -113,9 +113,9 @@ PARAM_DOUBLE_IN("newton_tolerance", "Tol
     "method.", "w", 1e-6);
 
 // Load/save a model.
-PARAM_MODEL_IN(SparseCoding, "input_model", "File containing input sparse "
+PARAM_MODEL_IN(SparseCoding<>, "input_model", "File containing input sparse "
     "coding model.", "m");
-PARAM_MODEL_OUT(SparseCoding, "output_model", "File to save trained sparse "
+PARAM_MODEL_OUT(SparseCoding<>, "output_model", "File to save trained sparse "
     "coding model to.", "M");
 
 PARAM_MATRIX_OUT("dictionary", "Matrix to save the output dictionary to.", "d");
@@ -177,11 +177,11 @@ void BINDING_FUNCTION(util::Params& para
       "Newton method tolerance must be nonnegative");
 
   // Do we have an existing model?
-  SparseCoding* sc;
+  SparseCoding<>* sc;
   if (params.Has("input_model"))
-    sc = params.Get<SparseCoding*>("input_model");
+    sc = params.Get<SparseCoding<>*>("input_model");
   else
-    sc = new SparseCoding(0, 0.0);
+    sc = new SparseCoding<>(0, 0.0);
 
   if (params.Has("training"))
   {
@@ -207,7 +207,7 @@ void BINDING_FUNCTION(util::Params& para
     if (params.Has("input_model"))
     {
       Log::Info << "Using dictionary from existing model in '"
-          << params.GetPrintable<SparseCoding>("input_model")
+          << params.GetPrintable<SparseCoding<>>("input_model")
           << "' as initial dictionary for training." << endl;
       sc->Train<NothingInitializer>(matX);
     }
@@ -286,5 +286,5 @@ void BINDING_FUNCTION(util::Params& para
       sc->Dictionary().n_rows, sc->Dictionary().n_cols, false, false);
 
   // Save the model.
-  params.Get<SparseCoding*>("output_model") = sc;
+  params.Get<SparseCoding<>*>("output_model") = sc;
 }
diff -pruN 4.3.0-2/src/mlpack/methods/svdplusplus/svdplusplus_function_impl.hpp 4.4.0-1/src/mlpack/methods/svdplusplus/svdplusplus_function_impl.hpp
--- 4.3.0-2/src/mlpack/methods/svdplusplus/svdplusplus_function_impl.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/svdplusplus/svdplusplus_function_impl.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -20,15 +20,16 @@ namespace mlpack {
 
 template <typename MatType>
 SVDPlusPlusFunction<MatType>::SVDPlusPlusFunction(
-    const MatType& data,
+    const MatType& dataIn,
     const arma::sp_mat& implicitData,
     const size_t rank,
     const double lambda) :
-    data(MakeAlias(const_cast<MatType&>(data), false)),
     implicitData(implicitData),
     rank(rank),
     lambda(lambda)
 {
+  MakeAlias(data, dataIn, dataIn.n_rows, dataIn.n_cols, false);
+
   // Number of users and items in the data.
   numUsers = max(data.row(0)) + 1;
   numItems = max(data.row(1)) + 1;
@@ -105,7 +106,7 @@ double SVDPlusPlusFunction<MatType>::Eva
       userVec += parameters.col(implicitStart + it.row()).subvec(0, rank - 1);
       if (implicitVecsNormSquare(it.row()) < 0)
       {
-        implicitVecsNormSquare(it.row()) = arma::dot(
+        implicitVecsNormSquare(it.row()) = dot(
             parameters.col(implicitStart + it.row()).subvec(0, rank - 1),
             parameters.col(implicitStart + it.row()).subvec(0, rank - 1));
       }
@@ -120,12 +121,12 @@ double SVDPlusPlusFunction<MatType>::Eva
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     double ratingError = rating - userBias - itemBias -
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
     double ratingErrorSquared = ratingError * ratingError;
 
     // Calculate the regularization penalty corresponding to the parameters.
-    double userVecNorm = arma::norm(parameters.col(user), 2);
-    double itemVecNorm = arma::norm(parameters.col(item), 2);
+    double userVecNorm = norm(parameters.col(user), 2);
+    double itemVecNorm = norm(parameters.col(item), 2);
     regularizationError += lambda * (userVecNorm * userVecNorm +
                                      itemVecNorm * itemVecNorm);
 
@@ -182,7 +183,7 @@ void SVDPlusPlusFunction<MatType>::Gradi
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     double ratingError = rating - userBias - itemBias -
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
     // example.
@@ -249,7 +250,7 @@ void SVDPlusPlusFunction<MatType>::Gradi
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     double ratingError = rating - userBias - itemBias -
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
     // example.
@@ -356,7 +357,7 @@ double StandardSGD::Optimize(
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     double ratingError = rating - userBias - itemBias -
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
 
     // Gradient is non-zero only for the parameter columns corresponding to the
     // example.
@@ -492,7 +493,7 @@ inline double ParallelSGD<ExponentialBac
         userVec += iterate.col(user).subvec(0, rank - 1);
 
         double ratingError = rating - userBias - itemBias -
-        arma::dot(userVec, iterate.col(item).subvec(0, rank - 1));
+        dot(userVec, iterate.col(item).subvec(0, rank - 1));
 
         arma::mat userVecUpdate = stepSize * 2 * (
             lambda * iterate.col(user).subvec(0, rank - 1) -
diff -pruN 4.3.0-2/src/mlpack/methods/xgboost/loss_functions/sse_loss.hpp 4.4.0-1/src/mlpack/methods/xgboost/loss_functions/sse_loss.hpp
--- 4.3.0-2/src/mlpack/methods/xgboost/loss_functions/sse_loss.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/methods/xgboost/loss_functions/sse_loss.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ class SSELoss
     if (values.n_elem == 0)
       return 0;
 
-    return arma::accu(values) / (typename VecType::elem_type) values.n_elem;
+    return accu(values) / (typename VecType::elem_type) values.n_elem;
   }
 
   /**
@@ -57,7 +57,7 @@ class SSELoss
   double OutputLeafValue(const MatType& /* input */,
                          const WeightVecType& /* weights */)
   {
-    return -ApplyL1(arma::accu(gradients)) / (arma::accu(hessians) + lambda);
+    return -ApplyL1(accu(gradients)) / (accu(hessians) + lambda);
   }
 
   /**
@@ -68,8 +68,8 @@ class SSELoss
    */
   double Evaluate(const size_t begin, const size_t end)
   {
-    return std::pow(ApplyL1(arma::accu(gradients.subvec(begin, end))), 2) /
-        (arma::accu(hessians.subvec(begin, end)) + lambda);
+    return std::pow(ApplyL1(accu(gradients.subvec(begin, end))), 2) /
+        (accu(hessians.subvec(begin, end)) + lambda);
   }
 
   /**
@@ -89,8 +89,7 @@ class SSELoss
     gradients = (input.row(1) - input.row(0)).t();
     hessians = arma::vec(input.n_cols, arma::fill::ones);
 
-    return std::pow(ApplyL1(arma::accu(gradients)), 2) /
-        (arma::accu(hessians) + lambda);
+    return std::pow(ApplyL1(accu(gradients)), 2) / (accu(hessians) + lambda);
   }
  private:
   //! The L1 regularization parameter.
diff -pruN 4.3.0-2/src/mlpack/prereqs.hpp 4.4.0-1/src/mlpack/prereqs.hpp
--- 4.3.0-2/src/mlpack/prereqs.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/prereqs.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,12 +31,9 @@
 #include <mlpack/core/cereal/array_wrapper.hpp>
 #include <mlpack/core/cereal/pointer_vector_wrapper.hpp>
 #include <mlpack/core/cereal/pointer_wrapper.hpp>
+#include <mlpack/core/cereal/template_class_version.hpp>
 #include <mlpack/core/data/has_serialize.hpp>
 
-// All code should have access to logging.
-#include <mlpack/core/util/log.hpp>
-#include <mlpack/core/util/io.hpp>
-
 // Include ready to use utility function to check sizes of datasets.
 #include <mlpack/core/util/size_checks.hpp>
 
diff -pruN 4.3.0-2/src/mlpack/tests/CMakeLists.txt 4.4.0-1/src/mlpack/tests/CMakeLists.txt
--- 4.3.0-2/src/mlpack/tests/CMakeLists.txt	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/CMakeLists.txt	2024-05-28 20:08:34.000000000 +0000
@@ -79,7 +79,6 @@ add_executable(mlpack_test
   knn_test.cpp
   krann_search_test.cpp
   lars_test.cpp
-  lin_alg_test.cpp
   linear_regression_test.cpp
   lmnn_test.cpp
   linear_svm_test.cpp
@@ -226,12 +225,13 @@ endif()
 # This has to be added here so that cotire picks it up (even though it is in
 # individual tests).
 target_compile_definitions(mlpack_test PUBLIC -DMLPACK_ENABLE_ANN_SERIALIZATION)
-set_target_properties(mlpack_test PROPERTIES COTIRE_CXX_PREFIX_HEADER_INIT
-    "../core.hpp")
-# TODO: use the source below, but this requires the DET test to be refactored
-# and cleaned up.
-#    "../../mlpack.hpp")
-cotire(mlpack_test)
+if (CMAKE_VERSION VERSION_GREATER "3.16.0" AND USE_PRECOMPILED_HEADERS)
+  target_precompile_headers(mlpack_test PUBLIC
+      "../core.hpp"
+      "../../mlpack.hpp"
+      <ensmallen.hpp>
+      <armadillo>)
+endif ()
 
 # Copy test data into right place.
 add_custom_command(TARGET mlpack_test
diff -pruN 4.3.0-2/src/mlpack/tests/adaboost_test.cpp 4.4.0-1/src/mlpack/tests/adaboost_test.cpp
--- 4.3.0-2/src/mlpack/tests/adaboost_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/adaboost_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -24,14 +24,17 @@ using namespace mlpack;
  * checks whether the hamming loss breaches the upperbound, which is provided by
  * ztAccumulator.
  */
-TEST_CASE("HammingLossBoundIris", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossBoundIris", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
 
   if (!data::Load("iris.csv", inputData))
     FAIL("Cannot load test dataset iris.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("iris_labels.txt", labels))
     FAIL("Cannot load labels for iris iris_labels.txt");
 
@@ -41,20 +44,20 @@ TEST_CASE("HammingLossBoundIris", "[AdaB
   // Run the perceptron for perceptronIter iterations.
   int perceptronIter = 400;
 
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
-
   // Define parameters for AdaBoost.
   size_t iterations = 100;
-  double tolerance = 1e-10;
-  AdaBoost<> a(tolerance);
-  double ztProduct = a.Train(inputData, labels.row(0), numClasses, p,
-      iterations, tolerance);
+  eT tolerance = 2e-10;
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  AdaBoost<PerceptronType, MatType> a;
+  eT ztProduct = a.Train(inputData, labels.row(0), numClasses, iterations,
+      tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) == true);
@@ -66,15 +69,16 @@ TEST_CASE("HammingLossBoundIris", "[AdaB
  * checks if the error returned by running a single instance of the weak learner
  * close to that of the boosted weak learner using adaboost.
  */
-TEST_CASE("WeakLearnerErrorIris", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorIris", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
 
+  MatType inputData;
   if (!data::Load("iris.csv", inputData))
     FAIL("Cannot load test dataset iris.csv!");
 
-  arma::Mat<size_t> labels;
-
+  Mat<size_t> labels;
   if (!data::Load("iris_labels.txt", labels))
     FAIL("Cannot load labels for iris iris_labels.txt");
 
@@ -84,23 +88,26 @@ TEST_CASE("WeakLearnerErrorIris", "[AdaB
   // Run the perceptron for perceptronIter iterations.
   int perceptronIter = 400;
 
-  arma::Row<size_t> perceptronPrediction(labels.n_cols);
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  Row<size_t> perceptronPrediction(labels.n_cols);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
   p.Classify(inputData, perceptronPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != perceptronPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != perceptronPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 100;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);;
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -110,13 +117,17 @@ TEST_CASE("WeakLearnerErrorIris", "[AdaB
  * dataset.  It checks whether the hamming loss breaches the upperbound, which
  * is provided by ztAccumulator.
  */
-TEST_CASE("HammingLossBoundVertebralColumn", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossBoundVertebralColumn", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("vc2.csv", inputData))
     FAIL("Cannot load test dataset vc2.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("vc2_labels.txt", labels))
     FAIL("Cannot load labels for vc2_labels.txt");
 
@@ -125,20 +136,22 @@ TEST_CASE("HammingLossBoundVertebralColu
   // Define your own weak learner, perceptron in this case.
   // Run the perceptron for perceptronIter iterations.
   size_t perceptronIter = 800;
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(tolerance);
-  double ztProduct = a.Train(inputData, labels.row(0), numClasses, p,
-      iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(tolerance);
+  eT ztProduct = a.Train(inputData, labels.row(0), numClasses, iterations,
+      tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) == true);
@@ -150,13 +163,17 @@ TEST_CASE("HammingLossBoundVertebralColu
  * dataset.  It checks if the error returned by running a single instance of the
  * weak learner is close to that of a boosted weak learner using adaboost.
  */
-TEST_CASE("WeakLearnerErrorVertebralColumn", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorVertebralColumn", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("vc2.csv", inputData))
     FAIL("Cannot load test dataset vc2.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("vc2_labels.txt", labels))
     FAIL("Cannot load labels for vc2_labels.txt");
 
@@ -167,22 +184,25 @@ TEST_CASE("WeakLearnerErrorVertebralColu
   size_t perceptronIter = 800;
 
   Row<size_t> perceptronPrediction(labels.n_cols);
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
   p.Classify(inputData, perceptronPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != perceptronPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != perceptronPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -192,13 +212,17 @@ TEST_CASE("WeakLearnerErrorVertebralColu
  * dataset.  It checks whether the hamming loss breaches the upperbound, which
  * is provided by ztAccumulator.
  */
-TEST_CASE("HammingLossBoundNonLinearSepData", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossBoundNonLinearSepData", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("train_nonlinsep.txt", inputData))
     FAIL("Cannot load test dataset train_nonlinsep.txt!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("train_labels_nonlinsep.txt", labels))
     FAIL("Cannot load labels for train_labels_nonlinsep.txt");
 
@@ -207,20 +231,22 @@ TEST_CASE("HammingLossBoundNonLinearSepD
   // Define your own weak learner, perceptron in this case.
   // Run the perceptron for perceptronIter iterations.
   size_t perceptronIter = 800;
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(tolerance);
-  double ztProduct = a.Train(inputData, labels.row(0), numClasses, p,
-      iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(tolerance);
+  eT ztProduct = a.Train(inputData, labels.row(0), numClasses, iterations,
+      tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels == predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels == predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) <= true);
@@ -232,13 +258,17 @@ TEST_CASE("HammingLossBoundNonLinearSepD
  * dataset.  It checks if the error returned by running a single instance of the
  * weak learner is close to that of a boosted weak learner using AdaBoost.
  */
-TEST_CASE("WeakLearnerErrorNonLinearSepData", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorNonLinearSepData", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("train_nonlinsep.txt", inputData))
     FAIL("Cannot load test dataset train_nonlinsep.txt!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("train_labels_nonlinsep.txt", labels))
     FAIL("Cannot load labels for train_labels_nonlinsep.txt");
 
@@ -249,22 +279,25 @@ TEST_CASE("WeakLearnerErrorNonLinearSepD
   size_t perceptronIter = 800;
 
   Row<size_t> perceptronPrediction(labels.n_cols);
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
   p.Classify(inputData, perceptronPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != perceptronPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != perceptronPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -274,34 +307,36 @@ TEST_CASE("WeakLearnerErrorNonLinearSepD
  * checks whether the Hamming loss breaches the upper bound, which is provided
  * by ztAccumulator.  This uses decision stumps as the weak learner.
  */
-TEST_CASE("HammingLossIris_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossIris_DS", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("iris.csv", inputData))
     FAIL("Cannot load test dataset iris.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("iris_labels.txt", labels))
     FAIL("Cannot load labels for iris_labels.txt");
 
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 3;
   const size_t inpBucketSize = 6;
-  arma::Row<size_t> labelsvec = labels.row(0);
-  ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
+  Row<size_t> labelsvec = labels.row(0);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<ID3DecisionStump> a(tolerance);
-  double ztProduct = a.Train(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<ID3DecisionStump, MatType> a(tolerance);
+  eT ztProduct = a.Train(inputData, labelsvec, numClasses, iterations,
+      tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) == true);
@@ -314,13 +349,16 @@ TEST_CASE("HammingLossIris_DS", "[AdaBoo
  * weak learner is close to that of a boosted weak learner using adaboost.
  * This is for the weak learner: decision stumps.
  */
-TEST_CASE("WeakLearnerErrorIris_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorIris_DS", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("iris.csv", inputData))
     FAIL("Cannot load test dataset iris.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("iris_labels.txt", labels))
     FAIL("Cannot load labels for iris_labels.txt");
 
@@ -329,28 +367,28 @@ TEST_CASE("WeakLearnerErrorIris_DS", "[A
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 3;
   const size_t inpBucketSize = 6;
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> labelsvec = labels.row(0);
 
-  arma::Row<size_t> dsPrediction(labels.n_cols);
+  Row<size_t> dsPrediction(labels.n_cols);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
   ds.Classify(inputData, dsPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != dsPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != dsPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
+  eT tolerance = 1e-10;
 
-  AdaBoost<ID3DecisionStump> a(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  AdaBoost<ID3DecisionStump, MatType> a(inputData, labelsvec, numClasses,
+      iterations, tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -361,36 +399,40 @@ TEST_CASE("WeakLearnerErrorIris_DS", "[A
  * weak learner is close to that of a boosted weak learner using adaboost.
  * This is for the weak learner: decision stumps.
  */
-TEST_CASE("HammingLossBoundVertebralColumn_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossBoundVertebralColumn_DS", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("vc2.csv", inputData))
     FAIL("Cannot load test dataset vc2.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("vc2_labels.txt", labels))
     FAIL("Cannot load labels for vc2_labels.txt");
 
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 3;
   const size_t inpBucketSize = 6;
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> labelsvec = labels.row(0);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
+  eT tolerance = 1e-10;
 
-  AdaBoost<ID3DecisionStump> a(tolerance);
-  double ztProduct = a.Train(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  AdaBoost<ID3DecisionStump, MatType> a(tolerance);
+  eT ztProduct = a.Train(inputData, labelsvec, numClasses, iterations,
+      tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) == true);
@@ -403,39 +445,43 @@ TEST_CASE("HammingLossBoundVertebralColu
  * weak learner is close to that of a boosted weak learner using adaboost.
  * This is for the weak learner: decision stumps.
  */
-TEST_CASE("WeakLearnerErrorVertebralColumn_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorVertebralColumn_DS", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("vc2.csv", inputData))
     FAIL("Cannot load test dataset vc2.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("vc2_labels.txt", labels))
     FAIL("Cannot load labels for vc2_labels.txt");
 
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 3;
   const size_t inpBucketSize = 6;
-  arma::Row<size_t> dsPrediction(labels.n_cols);
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> dsPrediction(labels.n_cols);
+  Row<size_t> labelsvec = labels.row(0);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
   ds.Classify(inputData, dsPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != dsPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != dsPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<ID3DecisionStump> a(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<ID3DecisionStump, MatType> a(inputData, labelsvec, numClasses,
+      iterations, tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -445,36 +491,40 @@ TEST_CASE("WeakLearnerErrorVertebralColu
  * dataset.  It checks whether the hamming loss breaches the upperbound, which
  * is provided by ztAccumulator.  This is for the weak learner: decision stumps.
  */
-TEST_CASE("HammingLossBoundNonLinearSepData_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("HammingLossBoundNonLinearSepData_DS", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("train_nonlinsep.txt", inputData))
     FAIL("Cannot load test dataset train_nonlinsep.txt!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("train_labels_nonlinsep.txt", labels))
     FAIL("Cannot load labels for train_labels_nonlinsep.txt");
 
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 2;
   const size_t inpBucketSize = 6;
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> labelsvec = labels.row(0);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
 
   // Define parameters for Adaboost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
+  eT tolerance = 1e-10;
 
-  AdaBoost<ID3DecisionStump> a(tolerance);
-  double ztProduct = a.Train(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  AdaBoost<ID3DecisionStump, MatType> a(tolerance);
+  eT ztProduct = a.Train(inputData, labelsvec, numClasses, iterations,
+      tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double hammingLoss = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT hammingLoss = (eT) countError / labels.n_cols;
 
   // Check that ztProduct is finite.
   REQUIRE(std::isfinite(ztProduct) == true);
@@ -487,41 +537,45 @@ TEST_CASE("HammingLossBoundNonLinearSepD
  * weak learner is close to that of a boosted weak learner using adaboost.
  * This for the weak learner: decision stumps.
  */
-TEST_CASE("WeakLearnerErrorNonLinearSepData_DS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("WeakLearnerErrorNonLinearSepData_DS", "[AdaBoostTest]", mat,
+    fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("train_nonlinsep.txt", inputData))
     FAIL("Cannot load test dataset train_nonlinsep.txt!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("train_labels_nonlinsep.txt", labels))
     FAIL("Cannot load labels for train_labels_nonlinsep.txt");
 
   // Define your own weak learner, decision stumps in this case.
   const size_t numClasses = 2;
   const size_t inpBucketSize = 3;
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> labelsvec = labels.row(0);
 
-  arma::Row<size_t> dsPrediction(labels.n_cols);
+  Row<size_t> dsPrediction(labels.n_cols);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
   ds.Classify(inputData, dsPrediction);
 
-  size_t countWeakLearnerError = arma::accu(labels != dsPrediction);
-  double weakLearnerErrorRate = (double) countWeakLearnerError / labels.n_cols;
+  size_t countWeakLearnerError = accu(labels != dsPrediction);
+  eT weakLearnerErrorRate = (eT) countWeakLearnerError / labels.n_cols;
 
   // Define parameters for AdaBoost.
   size_t iterations = 500;
-  double tolerance = 1e-23;
+  eT tolerance = 1e-23;
 
-  AdaBoost<ID3DecisionStump > a(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
+  AdaBoost<ID3DecisionStump, MatType> a(inputData, labelsvec, numClasses,
+      iterations, tolerance, inpBucketSize);
 
-  arma::Row<size_t> predictedLabels;
+  Row<size_t> predictedLabels;
   a.Classify(inputData, predictedLabels);
 
-  size_t countError = arma::accu(labels != predictedLabels);
-  double error = (double) countError / labels.n_cols;
+  size_t countError = accu(labels != predictedLabels);
+  eT error = (eT) countError / labels.n_cols;
 
   REQUIRE(error <= weakLearnerErrorRate + 0.03);
 }
@@ -531,13 +585,16 @@ TEST_CASE("WeakLearnerErrorNonLinearSepD
  * dataset.  It tests the Classify function and checks for a satisfactory error
  * rate.
  */
-TEST_CASE("ClassifyTest_VERTEBRALCOL", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("ClassifyTest_VERTEBRALCOL", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("vc2.csv", inputData))
     FAIL("Cannot load test dataset vc2.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("vc2_labels.txt", labels))
     FAIL("Cannot load labels for vc2_labels.txt");
 
@@ -545,12 +602,11 @@ TEST_CASE("ClassifyTest_VERTEBRALCOL", "
   // Run the perceptron for perceptronIter iterations.
   size_t perceptronIter = 1000;
 
-  arma::mat testData;
-
+  MatType testData;
   if (!data::Load("vc2_test.csv", testData))
     FAIL("Cannot load test dataset vc2_test.csv!");
 
-  arma::Mat<size_t> trueTestLabels;
+  Mat<size_t> trueTestLabels;
 
   if (!data::Load("vc2_test_labels.txt", trueTestLabels))
     FAIL("Cannot load labels for vc2_test_labels.txt");
@@ -558,17 +614,20 @@ TEST_CASE("ClassifyTest_VERTEBRALCOL", "
   const size_t numClasses = max(labels.row(0)) + 1;
 
   Row<size_t> perceptronPrediction(labels.n_cols);
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
   p.Classify(inputData, perceptronPrediction);
 
   // Define parameters for AdaBoost.
   size_t iterations = 100;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
-
-  arma::Row<size_t> predictedLabels1(testData.n_cols),
-                    predictedLabels2(testData.n_cols);
-  arma::mat probabilities;
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
+
+  Row<size_t> predictedLabels1(testData.n_cols),
+              predictedLabels2(testData.n_cols);
+  MatType probabilities;
 
   a.Classify(testData, predictedLabels1);
   a.Classify(testData, predictedLabels2, probabilities);
@@ -579,19 +638,19 @@ TEST_CASE("ClassifyTest_VERTEBRALCOL", "
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
     REQUIRE(predictedLabels1[i] == predictedLabels2[i]);
 
-  arma::colvec pRow;
-  arma::uword maxIndex = 0;
+  Col<eT> pRow;
+  uword maxIndex = 0;
 
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
   {
     pRow = probabilities.unsafe_col(i);
     pRow.max(maxIndex);
     REQUIRE(predictedLabels1(i) == maxIndex);
-    REQUIRE(arma::accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
+    REQUIRE(accu(probabilities.col(i)) == Approx(1));
   }
 
-  size_t localError = arma::accu(trueTestLabels != predictedLabels1);
-  double lError = (double) localError / trueTestLabels.n_cols;
+  size_t localError = accu(trueTestLabels != predictedLabels1);
+  eT lError = (eT) localError / trueTestLabels.n_cols;
   REQUIRE(lError <= 0.30);
 }
 
@@ -600,44 +659,46 @@ TEST_CASE("ClassifyTest_VERTEBRALCOL", "
  * dataset.  It tests the Classify function and checks for a satisfactory error
  * rate.
  */
-TEST_CASE("ClassifyTest_NONLINSEP", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("ClassifyTest_NONLINSEP", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("train_nonlinsep.txt", inputData))
     FAIL("Cannot load test dataset train_nonlinsep.txt!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("train_labels_nonlinsep.txt", labels))
     FAIL("Cannot load labels for train_labels_nonlinsep.txt");
 
   // Define your own weak learner; in this test decision stumps are used.
   const size_t numClasses = 2;
   const size_t inpBucketSize = 3;
-  arma::Row<size_t> labelsvec = labels.row(0);
+  Row<size_t> labelsvec = labels.row(0);
 
-  arma::mat testData;
+  MatType testData;
 
   if (!data::Load("test_nonlinsep.txt", testData))
     FAIL("Cannot load test dataset test_nonlinsep.txt!");
 
-  arma::Mat<size_t> trueTestLabels;
-
+  Mat<size_t> trueTestLabels;
   if (!data::Load("test_labels_nonlinsep.txt", trueTestLabels))
     FAIL("Cannot load labels for test_labels_nonlinsep.txt");
 
-  arma::Row<size_t> dsPrediction(labels.n_cols);
+  Row<size_t> dsPrediction(labels.n_cols);
 
   ID3DecisionStump ds(inputData, labelsvec, numClasses, inpBucketSize);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<ID3DecisionStump > a(inputData, labelsvec, numClasses, ds,
-      iterations, tolerance);
-
-  arma::Row<size_t> predictedLabels1(testData.n_cols),
-                    predictedLabels2(testData.n_cols);
-  arma::mat probabilities;
+  eT tolerance = 1e-10;
+  AdaBoost<ID3DecisionStump, MatType> a(inputData, labelsvec, numClasses,
+      iterations, tolerance, inpBucketSize);
+
+  Row<size_t> predictedLabels1(testData.n_cols),
+              predictedLabels2(testData.n_cols);
+  MatType probabilities;
 
   a.Classify(testData, predictedLabels1);
   a.Classify(testData, predictedLabels2, probabilities);
@@ -647,19 +708,19 @@ TEST_CASE("ClassifyTest_NONLINSEP", "[Ad
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
     REQUIRE(predictedLabels1[i] == predictedLabels2[i]);
 
-  arma::colvec pRow;
-  arma::uword maxIndex = 0;
+  Col<eT> pRow;
+  uword maxIndex = 0;
 
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
   {
     pRow = probabilities.unsafe_col(i);
     pRow.max(maxIndex);
     REQUIRE(predictedLabels1(i) == maxIndex);
-    REQUIRE(arma::accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
+    REQUIRE(accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
   }
 
-  size_t localError = arma::accu(trueTestLabels != predictedLabels1);
-  double lError = (double) localError / trueTestLabels.n_cols;
+  size_t localError = accu(trueTestLabels != predictedLabels1);
+  eT lError = (eT) localError / trueTestLabels.n_cols;
   REQUIRE(lError <= 0.30);
 }
 
@@ -669,13 +730,16 @@ TEST_CASE("ClassifyTest_NONLINSEP", "[Ad
  * the remaining third of the dataset (iris_test.csv).  It tests the Classify()
  * function and checks for a satisfactory error rate.
  */
-TEST_CASE("ClassifyTest_IRIS", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("ClassifyTest_IRIS", "[AdaBoostTest]", mat, fmat)
 {
-  arma::mat inputData;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  MatType inputData;
   if (!data::Load("iris_train.csv", inputData))
     FAIL("Cannot load test dataset iris_train.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("iris_train_labels.csv", labels))
     FAIL("Cannot load labels for iris_train_labels.csv");
   const size_t numClasses = max(labels.row(0)) + 1;
@@ -684,27 +748,30 @@ TEST_CASE("ClassifyTest_IRIS", "[AdaBoos
   // Run the perceptron for perceptronIter iterations.
   size_t perceptronIter = 800;
 
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
 
   // Define parameters for AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
 
-  arma::mat testData;
+  MatType testData;
   if (!data::Load("iris_test.csv", testData))
     FAIL("Cannot load test dataset iris_test.csv!");
 
-  arma::Row<size_t> predictedLabels(testData.n_cols);
+  Row<size_t> predictedLabels(testData.n_cols);
   a.Classify(testData, predictedLabels);
 
-  arma::Mat<size_t> trueTestLabels;
+  Mat<size_t> trueTestLabels;
   if (!data::Load("iris_test_labels.csv", trueTestLabels))
     FAIL("Cannot load test dataset iris_test_labels.csv!");
 
-  arma::Row<size_t> predictedLabels1(testData.n_cols),
-                    predictedLabels2(testData.n_cols);
-  arma::mat probabilities;
+  Row<size_t> predictedLabels1(testData.n_cols),
+              predictedLabels2(testData.n_cols);
+  MatType probabilities;
 
   a.Classify(testData, predictedLabels1);
   a.Classify(testData, predictedLabels2, probabilities);
@@ -714,19 +781,19 @@ TEST_CASE("ClassifyTest_IRIS", "[AdaBoos
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
     REQUIRE(predictedLabels1[i] == predictedLabels2[i]);
 
-  arma::colvec pRow;
-  arma::uword maxIndex = 0;
+  Col<eT> pRow;
+  uword maxIndex = 0;
 
   for (size_t i = 0; i < predictedLabels1.n_cols; ++i)
   {
     pRow = probabilities.unsafe_col(i);
     pRow.max(maxIndex);
     REQUIRE(predictedLabels1(i) == maxIndex);
-    REQUIRE(arma::accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
+    REQUIRE(accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
   }
 
-  size_t localError = arma::accu(trueTestLabels != predictedLabels1);
-  double lError = (double) localError / labels.n_cols;
+  size_t localError = accu(trueTestLabels != predictedLabels1);
+  eT lError = (eT) localError / labels.n_cols;
   REQUIRE(lError <= 0.30);
 }
 
@@ -734,26 +801,32 @@ TEST_CASE("ClassifyTest_IRIS", "[AdaBoos
  * Ensure that the Train() function works like it is supposed to, by building
  * AdaBoost on one dataset and then re-training on another dataset.
  */
-TEST_CASE("TrainTest", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("TrainTest", "[AdaBoostTest]", mat, fmat)
 {
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
   // First train on the iris dataset.
-  arma::mat inputData;
+  MatType inputData;
   if (!data::Load("iris_train.csv", inputData))
     FAIL("Cannot load test dataset iris_train.csv!");
 
-  arma::Mat<size_t> labels;
+  Mat<size_t> labels;
   if (!data::Load("iris_train_labels.csv", labels))
     FAIL("Cannot load labels for iris_train_labels.csv");
 
   const size_t numClasses = max(labels.row(0)) + 1;
 
   size_t perceptronIter = 800;
-  Perceptron<> p(inputData, labels.row(0), numClasses, perceptronIter);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  PerceptronType p(inputData, labels.row(0), numClasses, perceptronIter);
 
   // Now train AdaBoost.
   size_t iterations = 50;
-  double tolerance = 1e-10;
-  AdaBoost<> a(inputData, labels.row(0), numClasses, p, iterations, tolerance);
+  eT tolerance = 1e-10;
+  AdaBoost<PerceptronType, MatType> a(inputData, labels.row(0), numClasses,
+      iterations, tolerance, perceptronIter);
 
   // Now load another dataset...
   if (!data::Load("vc2.csv", inputData))
@@ -763,44 +836,48 @@ TEST_CASE("TrainTest", "[AdaBoostTest]")
 
   const size_t newNumClasses = max(labels.row(0)) + 1;
 
-  Perceptron<> p2(inputData, labels.row(0), newNumClasses, perceptronIter);
+  PerceptronType p2(inputData, labels.row(0), newNumClasses, perceptronIter);
 
-  a.Train(inputData, labels.row(0), newNumClasses, p2, iterations, tolerance);
+  a.Train(inputData, labels.row(0), newNumClasses, iterations, tolerance,
+      perceptronIter);
 
   // Load test set to see if it trained on vc2 correctly.
-  arma::mat testData;
+  MatType testData;
   if (!data::Load("vc2_test.csv", testData))
     FAIL("Cannot load test dataset vc2_test.csv!");
 
-  arma::Mat<size_t> trueTestLabels;
+  Mat<size_t> trueTestLabels;
   if (!data::Load("vc2_test_labels.txt", trueTestLabels))
     FAIL("Cannot load labels for vc2_test_labels.txt");
 
   // Define parameters for AdaBoost.
-  arma::Row<size_t> predictedLabels(testData.n_cols);
+  Row<size_t> predictedLabels(testData.n_cols);
   a.Classify(testData, predictedLabels);
 
-  int localError = arma::accu(trueTestLabels != predictedLabels);
-  double lError = (double) localError / trueTestLabels.n_cols;
+  size_t localError = accu(trueTestLabels != predictedLabels);
+  eT lError = (eT) localError / trueTestLabels.n_cols;
 
   REQUIRE(lError <= 0.30);
 }
 
-TEST_CASE("PerceptronSerializationTest", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("PerceptronSerializationTest", "[AdaBoostTest]", fmat, mat)
 {
+  typedef TestType MatType;
+
   // Build an AdaBoost object.
-  mat data = randu<mat>(10, 500);
+  MatType data = randu<MatType>(10, 500);
   Row<size_t> labels(500);
   for (size_t i = 0; i < 250; ++i)
     labels[i] = 0;
   for (size_t i = 250; i < 500; ++i)
     labels[i] = 1;
 
-  Perceptron<> p(data, labels, 2, 800);
-  AdaBoost<> ab(data, labels, 2, p, 50, 1e-10);
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  AdaBoost<PerceptronType, MatType> ab(data, labels, 2, 50, 1e-10, 800);
 
   // Now create another dataset to train with.
-  mat otherData = randu<mat>(5, 200);
+  MatType otherData = randu<MatType>(5, 200);
   Row<size_t> otherLabels(200);
   for (size_t i = 0; i < 100; ++i)
     otherLabels[i] = 1;
@@ -809,10 +886,10 @@ TEST_CASE("PerceptronSerializationTest",
   for (size_t i = 150; i < 200; ++i)
     otherLabels[i] = 2;
 
-  Perceptron<> p2(otherData, otherLabels, 3, 500);
-  AdaBoost<> abText(otherData, otherLabels, 3, p2, 50, 1e-10);
+  AdaBoost<PerceptronType, MatType> abText(otherData, otherLabels, 3, 50, 1e-10,
+      500);
 
-  AdaBoost<> abXml, abBinary;
+  AdaBoost<PerceptronType, MatType> abXml, abBinary;
 
   SerializeObjectAll(ab, abXml, abText, abBinary);
 
@@ -839,21 +916,23 @@ TEST_CASE("PerceptronSerializationTest",
   }
 }
 
-TEST_CASE("ID3DecisionStumpSerializationTest", "[AdaBoostTest]")
+TEMPLATE_TEST_CASE("ID3DecisionStumpSerializationTest", "[AdaBoostTest]", mat,
+    fmat)
 {
+  typedef TestType MatType;
+
   // Build an AdaBoost object.
-  mat data = randu<mat>(10, 500);
+  MatType data = randu<MatType>(10, 500);
   Row<size_t> labels(500);
   for (size_t i = 0; i < 250; ++i)
     labels[i] = 0;
   for (size_t i = 250; i < 500; ++i)
     labels[i] = 1;
 
-  ID3DecisionStump p(data, labels, 2, 800);
-  AdaBoost<ID3DecisionStump> ab(data, labels, 2, p, 50, 1e-10);
+  AdaBoost<ID3DecisionStump, MatType> ab(data, labels, 2, 50, 1e-10, 40);
 
   // Now create another dataset to train with.
-  mat otherData = randu<mat>(5, 200);
+  MatType otherData = randu<MatType>(5, 200);
   Row<size_t> otherLabels(200);
   for (size_t i = 0; i < 100; ++i)
     otherLabels[i] = 1;
@@ -862,10 +941,10 @@ TEST_CASE("ID3DecisionStumpSerialization
   for (size_t i = 150; i < 200; ++i)
     otherLabels[i] = 2;
 
-  ID3DecisionStump p2(otherData, otherLabels, 3, 500);
-  AdaBoost<ID3DecisionStump> abText(otherData, otherLabels, 3, p2, 50, 1e-10);
+  AdaBoost<ID3DecisionStump, MatType> abText(otherData, otherLabels, 3, 50,
+      1e-10, 25);
 
-  AdaBoost<ID3DecisionStump> abXml, abBinary;
+  AdaBoost<ID3DecisionStump, MatType> abXml, abBinary;
 
   SerializeObjectAll(ab, abXml, abText, abBinary);
 
@@ -881,10 +960,147 @@ TEST_CASE("ID3DecisionStumpSerialization
   for (size_t i = 0; i < ab.WeakLearners(); ++i)
   {
     REQUIRE(ab.WeakLearner(i).SplitDimension() ==
-           abXml.WeakLearner(i).SplitDimension());
+            abXml.WeakLearner(i).SplitDimension());
     REQUIRE(ab.WeakLearner(i).SplitDimension() ==
             abText.WeakLearner(i).SplitDimension());
     REQUIRE(ab.WeakLearner(i).SplitDimension() ==
             abBinary.WeakLearner(i).SplitDimension());
   }
 }
+
+TEMPLATE_TEST_CASE("AdaBoostSinglePointClassify", "[AdaBoostTest]", mat, fmat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data = randu<MatType>(10, 100);
+  // Create random labels.
+  Row<size_t> labels = randi<Row<size_t>>(100, distr_param(0, 3));
+
+  // Train a model.
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  AdaBoost<PerceptronType, MatType> ab(data, labels, 4);
+
+  // Ensure that we can get single-point classifications.
+  for (size_t i = 0; i < 100; ++i)
+  {
+    const size_t prediction = ab.Classify(data.col(i));
+
+    REQUIRE(prediction <= 3);
+  }
+}
+
+TEMPLATE_TEST_CASE("AdaBoostSinglePointClassifyWithProbs", "[AdaBoostTest]",
+    mat, fmat)
+{
+  typedef TestType MatType;
+  typedef typename MatType::elem_type eT;
+
+  // Create random data.
+  MatType data = randu<MatType>(10, 100);
+  // Create random labels.
+  Row<size_t> labels = randi<Row<size_t>>(100, distr_param(0, 3));
+
+  // Train a model.
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  AdaBoost<PerceptronType, MatType> ab(data, labels, 4);
+
+  // Ensure that we can get single-point classifications.
+  for (size_t i = 0; i < 100; ++i)
+  {
+    size_t prediction;
+    Row<eT> probabilities;
+    ab.Classify(data.col(i), prediction, probabilities);
+
+    REQUIRE(prediction <= 3);
+    REQUIRE(accu(probabilities) == Approx((eT) 1.0));
+  }
+}
+
+// Make sure that everything works when we use the constructor that takes extra
+// hyperparameters.
+TEMPLATE_TEST_CASE("AdaBoostParamsConstructor", "[AdaBoostTest]", fmat, mat)
+{
+  typedef TestType MatType;
+
+  MatType inputData;
+  if (!data::Load("iris.csv", inputData))
+    FAIL("Cannot load test dataset iris.csv!");
+
+  Mat<size_t> labels;
+  if (!data::Load("iris_labels.txt", labels))
+    FAIL("Cannot load labels for iris iris_labels.txt");
+
+  const size_t numClasses = max(labels.row(0)) + 1;
+
+  // Create two AdaBoost models.  One does not allow the perceptron to train for
+  // more than one iteration, and therefore should get less accuracy than the
+  // one we let train in full.
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+
+  AdaBoost<PerceptronType, MatType> a1(inputData, labels, numClasses, 2, 1e-6,
+      1 /* perceptron max iterations */);
+  AdaBoost<PerceptronType, MatType> a2(inputData, labels, numClasses, 2, 1e-6,
+      100 /* perceptron max iterations */);
+
+  // Make sure test data performance is better for a2.
+  Row<size_t> predictions1, predictions2;
+  a1.Classify(inputData, predictions1);
+  a2.Classify(inputData, predictions2);
+
+  const size_t correct1 = accu(predictions1 == labels);
+  const size_t correct2 = accu(predictions2 == labels);
+
+  REQUIRE(correct2 > 0);
+  REQUIRE(correct2 >= correct1);
+}
+
+// Ensure that all Train() overloads work correctly.
+TEMPLATE_TEST_CASE("AdaBoostTrainOverloads", "[AdaBoostTest]", fmat, mat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data = randu<MatType>(10, 100);
+  // Create random labels.
+  Row<size_t> labels = randi<Row<size_t>>(100, distr_param(0, 3));
+
+  typedef Perceptron<SimpleWeightUpdate, ZeroInitialization, MatType>
+      PerceptronType;
+  AdaBoost<PerceptronType, MatType> a1, a2, a3, a4;
+  a1.MaxIterations() = 65;
+  a1.Tolerance() = 2e-4;
+  a2.Tolerance() = 2e-5;
+
+  a1.Train(data, labels, 4);
+  a2.Train(data, labels, 4, 15);
+  a3.Train(data, labels, 4, 55, 1e-3);
+  a4.Train(data, labels, 4, 60, 2e-3, 100);
+
+  // Make sure hyperparameters were set correctly, where appropriate.
+  REQUIRE(a1.MaxIterations() == 65);
+  REQUIRE(a2.MaxIterations() == 15);
+  REQUIRE(a3.MaxIterations() == 55);
+  REQUIRE(a4.MaxIterations() == 60);
+
+  REQUIRE(a1.Tolerance() == Approx(2e-4));
+  REQUIRE(a2.Tolerance() == Approx(2e-5));
+  REQUIRE(a3.Tolerance() == Approx(1e-3));
+  REQUIRE(a4.Tolerance() == Approx(2e-3));
+
+  // Make sure anything at all was trained.
+  REQUIRE(a1.WeakLearners() > 0);
+  REQUIRE(a2.WeakLearners() > 0);
+  REQUIRE(a3.WeakLearners() > 0);
+  REQUIRE(a4.WeakLearners() > 0);
+
+  // Make sure the maximum number of iterations in the perceptron was set
+  // properly.
+  REQUIRE(a1.WeakLearner(0).MaxIterations() == 1000);
+  REQUIRE(a2.WeakLearner(0).MaxIterations() == 1000);
+  REQUIRE(a3.WeakLearner(0).MaxIterations() == 1000);
+  REQUIRE(a4.WeakLearner(0).MaxIterations() == 100);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/ann/ann_regularizer_test.cpp 4.4.0-1/src/mlpack/tests/ann/ann_regularizer_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/ann_regularizer_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/ann_regularizer_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -92,7 +92,7 @@ TEST_CASE("GradientOrthogonalRegularizer
     {
       arma::mat x = arma::abs(weight * weight.t() -
           arma::eye<arma::mat>(weight.n_rows, weight.n_cols)) * factor;
-      return arma::accu(x);
+      return accu(x);
     }
 
     void Gradient(arma::mat& weight, arma::mat& gradient)
diff -pruN 4.3.0-2/src/mlpack/tests/ann/ann_test_tools.hpp 4.4.0-1/src/mlpack/tests/ann/ann_test_tools.hpp
--- 4.3.0-2/src/mlpack/tests/ann/ann_test_tools.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/ann_test_tools.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -217,14 +217,14 @@ double ActivationJacobianTest(arma::mat&
 }
 
 // Simple numerical gradient checker.
-template<class FunctionType>
+template<class FunctionType, typename MatType = arma::mat>
 double CheckGradient(FunctionType& function, const double eps = 1e-7)
 {
   // Get gradients for the current parameters.
-  arma::mat orgGradient, gradient, estGradient;
+  MatType orgGradient, gradient, estGradient;
   function.Gradient(orgGradient);
 
-  estGradient = arma::zeros(orgGradient.n_rows, orgGradient.n_cols);
+  estGradient = arma::zeros<MatType>(orgGradient.n_rows, orgGradient.n_cols);
 
   // Compute numeric approximations to gradient.
   for (size_t i = 0; i < orgGradient.n_elem; ++i)
@@ -280,7 +280,7 @@ double CheckRegularizerGradient(Function
     }
   }
 
-  estGradient = arma::vectorise(estGradient);
+  estGradient = vectorise(estGradient);
   // Estimate error of gradient.
   return arma::norm(orgGradient - estGradient) /
       arma::norm(orgGradient + estGradient);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/convolutional_network_test.cpp 4.4.0-1/src/mlpack/tests/ann/convolutional_network_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/convolutional_network_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/convolutional_network_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,7 +71,7 @@ void CheckMoveFunction(ModelType* networ
 TEST_CASE("PaddingTest", "[ConvolutionalNetworktest]")
 {
   arma::mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  X.load("mnist_first250_training_4s_and_9s.csv");
 
   // Create the network.
   FFN<NegativeLogLikelihood, RandomInitialization> model;
@@ -148,7 +148,7 @@ TEST_CASE("MaxPoolingTest", "[Convolutio
 TEST_CASE("VanillaNetworkTest", "[ConvolutionalNetworkTest]")
 {
   arma::mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  X.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   arma::uword nPoints = X.n_cols;
@@ -232,7 +232,7 @@ TEST_CASE("VanillaNetworkTest", "[Convol
             arma::max(predictionTemp.col(i)) == predictionTemp.col(i), 1));
     }
 
-    size_t correct = arma::accu(prediction == Y);
+    size_t correct = accu(prediction == Y);
     double classificationError = 1 - double(correct) / X.n_cols;
     if (classificationError <= 0.25)
     {
@@ -264,7 +264,7 @@ TEST_CASE("VanillaNetworkBatchSizeTest",
   model.InputDimensions() = std::vector<size_t>({ 28, 28 });
 
   arma::mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  X.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   arma::uword nPoints = X.n_cols;
@@ -347,7 +347,7 @@ TEST_CASE("VanillaNetworkBatchSizeTest",
 TEST_CASE("CheckCopyVanillaNetworkTest", "[ConvolutionalNetworkTest]")
 {
   arma::mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  X.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   arma::uword nPoints = X.n_cols;
@@ -447,7 +447,7 @@ TEST_CASE("Issue2986", "[ConvolutionalNe
   c.InputDimensions() = std::vector<size_t>({ 6, 6 });
   c.ComputeOutputDimensions();
   arma::mat weights(c.WeightSize(), 1, arma::fill::randu);
-  c.SetWeights(weights.memptr());
+  c.SetWeights(weights);
 
   output.set_size(c.OutputSize(), 1);
   delta.set_size(input.size());
@@ -463,7 +463,7 @@ TEST_CASE("Issue2986", "[ConvolutionalNe
   c.ComputeOutputDimensions();
   weights.set_size(c.WeightSize(), 1);
   weights.randu();
-  c.SetWeights(weights.memptr());
+  c.SetWeights(weights);
 
   output.set_size(c.OutputSize(), 1);
   delta.set_size(input.size());
@@ -479,7 +479,7 @@ TEST_CASE("Issue2986", "[ConvolutionalNe
   c.ComputeOutputDimensions();
   weights.set_size(c.WeightSize(), 1);
   weights.randu();
-  c.SetWeights(weights.memptr());
+  c.SetWeights(weights);
 
   output.set_size(c.OutputSize(), 1);
   delta.set_size(input.size());
@@ -508,7 +508,7 @@ TEST_CASE("CustomPaddingTest", "[Convolu
 
   weights.set_size(c.WeightSize(), 1);
   weights.ones();
-  c.SetWeights(weights.memptr());
+  c.SetWeights(weights);
 
   // Now make sure that the forward pass returns the correct output.
   output.set_size(c.OutputSize(), 1);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/feedforward_network_test.cpp 4.4.0-1/src/mlpack/tests/ann/feedforward_network_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/feedforward_network_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/feedforward_network_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ void TestNetwork(ModelType& model,
         arma::max(predictionTemp.col(i)) == predictionTemp.col(i), 1));
   }
 
-  size_t correct = arma::accu(prediction == testLabels);
+  size_t correct = accu(prediction == testLabels);
 
   double classificationError = 1 - double(correct) / testData.n_cols;
   REQUIRE(classificationError <= classificationErrorThreshold);
@@ -400,7 +400,7 @@ TEST_CASE("FFVanillaNetworkTest", "[Feed
   TestNetwork<>(model, trainData, trainLabels, testData, testLabels, 10, 0.1);
 
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
@@ -421,7 +421,7 @@ TEST_CASE("FFVanillaNetworkTest", "[Feed
 TEST_CASE("ForwardBackwardTest", "[FeedForwardNetworkTest]")
 {
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
@@ -477,7 +477,7 @@ TEST_CASE("ForwardBackwardTest", "[FeedF
             arma::max(currentResuls.col(i)) == currentResuls.col(i), 1));
       }
 
-      size_t correct = arma::accu(prediction == currentLabels);
+      size_t correct = accu(prediction == currentLabels);
       error(1 - (double) correct / batchSize);
     }
     Log::Debug << "Current training error: " << error.mean() << std::endl;
@@ -548,7 +548,7 @@ TEST_CASE("DropoutNetworkTest", "[FeedFo
   // network must be significant better than 92%.
   TestNetwork<>(model, trainData, trainLabels, testData, testLabels, 10, 0.1);
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
@@ -627,7 +627,7 @@ TEST_CASE("DropConnectNetworkTest", "[Fe
   TestNetwork(model, trainData, trainLabels, testData, testLabels, 10, 0.1);
 
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
@@ -952,7 +952,7 @@ TEST_CASE("RBFNetworkTest", "[FeedForwar
   TestNetwork<>(model, trainData, trainLabels1, testData, testLabels, 10, 0.1);
 
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
diff -pruN 4.3.0-2/src/mlpack/tests/ann/init_rules_test.cpp 4.4.0-1/src/mlpack/tests/ann/init_rules_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/init_rules_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/init_rules_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -25,7 +25,7 @@ TEST_CASE("ConstantInitTest", "[InitRule
   RandomInitialization constantInit(1, 1);
   constantInit.Initialize(weights, 100, 100);
 
-  bool b = arma::all(arma::vectorise(weights) == 1);
+  bool b = arma::all(vectorise(weights) == 1);
   REQUIRE(b == 1);
 }
 
@@ -94,7 +94,7 @@ TEST_CASE("ConstInitTest", "[InitRulesTe
   ConstInitialization zeroInit(0);
   zeroInit.Initialize(weights, 100, 100);
 
-  bool b = arma::all(arma::vectorise(weights) == 0);
+  bool b = arma::all(vectorise(weights) == 0);
   REQUIRE(b == 1);
 }
 
@@ -211,7 +211,7 @@ TEST_CASE("NetworkInitTest", "[InitRules
   randomModel.Add<LogSoftMax>();
   randomModel.Predict(input, response);
 
-  bool b = arma::all(arma::vectorise(randomModel.Parameters()) == 0.5);
+  bool b = arma::all(vectorise(randomModel.Parameters()) == 0.5);
   REQUIRE(b == 1);
   REQUIRE(randomModel.Parameters().n_elem == 42);
 
@@ -234,7 +234,7 @@ TEST_CASE("NetworkInitTest", "[InitRules
   zeroModel.Add<LogSoftMax>();
   zeroModel.Predict(input, response);
 
-  REQUIRE(arma::accu(zeroModel.Parameters()) == 0);
+  REQUIRE(accu(zeroModel.Parameters()) == 0);
   REQUIRE(zeroModel.Parameters().n_elem == 42);
 
   // Create a simple network and use the
diff -pruN 4.3.0-2/src/mlpack/tests/ann/ksinit_test.cpp 4.4.0-1/src/mlpack/tests/ann/ksinit_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/ksinit_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/ksinit_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -85,11 +85,11 @@ void BuildVanillaNetwork(MatType& trainD
 
   // Calculating the mean squared error on the training data.
   model.Predict(trainData, prediction);
-  trainError = arma::mean(arma::mean(arma::square(prediction - trainLabels)));
+  trainError = arma::mean(arma::mean(square(prediction - trainLabels)));
 
   // Calculating the mean squared error on the test data
   model.Predict(testData, prediction);
-  testError = arma::mean(arma::mean(arma::square(prediction - testLabels)));
+  testError = arma::mean(arma::mean(square(prediction - testLabels)));
 }
 
 /**
@@ -294,7 +294,7 @@ TEST_CASE("NonLinearFunctionApproximatio
   dataset.row(9) /= 8;
 
   // Eqn 13.3.
-  dataset.row(10) = arma::sqrt(1 - dataset.row(0));
+  dataset.row(10) = sqrt(1 - dataset.row(0));
 
   // Counter for the number of failures.
   size_t numFails = 0;
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/adaptive_max_pooling.cpp 4.4.0-1/src/mlpack/tests/ann/layer/adaptive_max_pooling.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/adaptive_max_pooling.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/adaptive_max_pooling.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -48,13 +48,13 @@ TEST_CASE("AdaptiveMaxPoolingTestCase",
   module1.Training() = true;
   module1.Forward(input, output);
   // Calculated using torch.nn.AdaptiveMaxPool2d().
-  REQUIRE(arma::accu(output) == 28);
+  REQUIRE(accu(output) == 28);
   REQUIRE(output.n_elem == 4);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(12, 1);
   module1.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 28.0);
+  REQUIRE(accu(delta) == 28.0);
 
   // For Square input.
   input = arma::mat(9, 1);
@@ -73,13 +73,13 @@ TEST_CASE("AdaptiveMaxPoolingTestCase",
   module2.Training() = true;
   module2.Forward(input, output);
   // Calculated using torch.nn.AdaptiveMaxPool2d().
-  REQUIRE(arma::accu(output) == 15.0);
+  REQUIRE(accu(output) == 15.0);
   REQUIRE(output.n_elem == 2);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(9, 1);
   module2.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 15.0);
+  REQUIRE(accu(delta) == 15.0);
 
   // For Square input.
   input = arma::mat(16, 1);
@@ -98,13 +98,13 @@ TEST_CASE("AdaptiveMaxPoolingTestCase",
   module3.Training() = true;
   module3.Forward(input, output);
   // Calculated using torch.nn.AdaptiveMaxPool2d().
-  REQUIRE(arma::accu(output) == 30.0);
+  REQUIRE(accu(output) == 30.0);
   REQUIRE(output.n_elem == 9);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(16, 1);
   module3.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 30.0);
+  REQUIRE(accu(delta) == 30.0);
 
   // For Rectangular input.
   input = arma::mat(20, 1);
@@ -121,11 +121,11 @@ TEST_CASE("AdaptiveMaxPoolingTestCase",
   module4.Training() = true;
   module4.Forward(input, output);
   // Calculated using torch.nn.AdaptiveMaxPool2d().
-  REQUIRE(arma::accu(output) == 2);
+  REQUIRE(accu(output) == 2);
   REQUIRE(output.n_elem == 4);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(20, 1);
   module4.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 2.0);
+  REQUIRE(accu(delta) == 2.0);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/adaptive_mean_pooling.cpp 4.4.0-1/src/mlpack/tests/ann/layer/adaptive_mean_pooling.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/adaptive_mean_pooling.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/adaptive_mean_pooling.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,13 +47,13 @@ TEST_CASE("AdaptiveMeanPoolingTestCase",
   module1.ComputeOutputDimensions();
   module1.Forward(input, output);
   // Calculated using torch.nn.AdaptiveAvgPool2d().
-  REQUIRE(arma::accu(output) == 19.75);
+  REQUIRE(accu(output) == 19.75);
   REQUIRE(output.n_elem == 4);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(12, 1);
   module1.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 19.75);
+  REQUIRE(accu(delta) == 19.75);
 
   // For Square input.
   input = arma::mat(9, 1);
@@ -71,13 +71,13 @@ TEST_CASE("AdaptiveMeanPoolingTestCase",
   module2.ComputeOutputDimensions();
   module2.Forward(input, output);
   // Calculated using torch.nn.AdaptiveAvgPool2d().
-  REQUIRE(arma::accu(output) == 4.5);
+  REQUIRE(accu(output) == 4.5);
   REQUIRE(output.n_elem == 2);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(9, 1);
   module2.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 4.50);
+  REQUIRE(accu(delta) == 4.50);
 
   // For Square input.
   input = arma::mat(16, 1);
@@ -95,13 +95,13 @@ TEST_CASE("AdaptiveMeanPoolingTestCase",
   module3.ComputeOutputDimensions();
   module3.Forward(input, output);
   // Calculated using torch.nn.AdaptiveAvgPool2d().
-  REQUIRE(arma::accu(output) == 10.5);
+  REQUIRE(accu(output) == 10.5);
   REQUIRE(output.n_elem == 9);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(16, 1);
   module3.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 10.5);
+  REQUIRE(accu(delta) == 10.5);
 
   // For Rectangular input.
   input = arma::mat(24, 1);
@@ -117,11 +117,11 @@ TEST_CASE("AdaptiveMeanPoolingTestCase",
   module4.ComputeOutputDimensions();
   module4.Forward(input, output);
   // Calculated using torch.nn.AdaptiveAvgPool2d().
-  REQUIRE(arma::accu(output) == 2.25);
+  REQUIRE(accu(output) == 2.25);
   REQUIRE(output.n_elem == 9);
   REQUIRE(output.n_cols == 1);
   // Test the Backward Function.
   delta.set_size(24, 1);
   module4.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 2.25);
+  REQUIRE(accu(delta) == 2.25);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/add.cpp 4.4.0-1/src/mlpack/tests/ann/layer/add.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/add.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/add.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,98 @@
+/**
+ * @file tests/ann/layer/add.cpp
+ * @author Ryan Curtin
+ *
+ * Tests the Add layer (which is just a bias).
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#include <mlpack/core.hpp>
+#include <mlpack/methods/ann.hpp>
+
+#include "../../test_catch_tools.hpp"
+#include "../../catch.hpp"
+#include "../../serialization.hpp"
+#include "../ann_test_tools.hpp"
+
+using namespace mlpack;
+
+/**
+ * Simple test for Add layer.
+ */
+TEST_CASE("AddManualWeightTestCase", "[ANNLayerTest]")
+{
+  arma::mat input = arma::mat(1, 1);
+  input.zeros();
+
+  arma::mat weights(1, 1);
+
+  Add module;
+  module.InputDimensions() = std::vector<size_t>({ 1 });
+  module.ComputeOutputDimensions();
+  module.SetWeights(weights);
+  module.Parameters()[0] = 3.0;
+
+  arma::mat output(1, 1);
+  module.Forward(input, output);
+
+  REQUIRE(output[0] == Approx(input[0] + 3.0));
+
+  arma::mat delta(1, 1);
+
+  // The backwards pass does not modify anything.
+  module.Backward(input, output, output, delta);
+  REQUIRE(delta[0] == Approx(output[0]));
+
+  arma::mat error(1, 1);
+  error(0, 0) = 2.0;
+  arma::mat gradient(1, 1);
+  module.Gradient(input, error, gradient);
+  REQUIRE(gradient[0] == Approx(error[0]));
+}
+
+/**
+ * Test the Add layer with a batch size greater than 1.
+ */
+TEST_CASE("AddManualWeightBatchTestCase", "[ANNLayerTest]")
+{
+  arma::mat input = arma::mat(1, 5);
+  input.zeros();
+  input[1] = 2.0;
+
+  arma::mat weights(1, 1);
+
+  Add module;
+  module.InputDimensions() = std::vector<size_t>({ 1 });
+  module.ComputeOutputDimensions();
+  module.SetWeights(weights);
+  module.Parameters()[0] = 3.0;
+
+  arma::mat output(1, 5);
+  module.Forward(input, output);
+
+  REQUIRE(output[0] == Approx(input[0] + 3.0));
+  REQUIRE(output[1] == Approx(input[1] + 3.0));
+  REQUIRE(output[2] == Approx(input[2] + 3.0));
+  REQUIRE(output[3] == Approx(input[3] + 3.0));
+  REQUIRE(output[4] == Approx(input[4] + 3.0));
+
+  arma::mat delta(1, 5);
+
+  // The backwards pass does not modify anything.
+  module.Backward(input, output, output, delta);
+  REQUIRE(delta[0] == Approx(output[0]));
+  REQUIRE(delta[1] == Approx(output[1]));
+  REQUIRE(delta[2] == Approx(output[2]));
+  REQUIRE(delta[3] == Approx(output[3]));
+  REQUIRE(delta[4] == Approx(output[4]));
+
+  arma::mat error(1, 5);
+  error.fill(2.0);
+  error(0, 1) = 3.0;
+  arma::mat gradient(1, 1);
+  module.Gradient(input, error, gradient);
+  REQUIRE(gradient[0] == Approx(11.0));
+}
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/add_merge.cpp 4.4.0-1/src/mlpack/tests/ann/layer/add_merge.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/add_merge.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/add_merge.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -65,9 +65,9 @@ TEST_CASE("AddMergeTestCase", "[ANNLayer
   output1.set_size(8, 1);
   output2.set_size(6, 1);
   module1.Forward(input, output1);
-  REQUIRE(arma::accu(output1) == 51.0);
+  REQUIRE(accu(output1) == 51.0);
   module2.Forward(input, output2);
-  REQUIRE(arma::accu(output2) == 39.0);
+  REQUIRE(accu(output2) == 39.0);
   output1.reshape(4, 2);
   output2.reshape(3, 2);
   CheckMatrices(output1, result1, 1e-1);
@@ -87,9 +87,9 @@ TEST_CASE("AddMergeTestCase", "[ANNLayer
   prevDelta1.reshape(8, 1);
   prevDelta2.reshape(6, 1);
   module1.Backward(input, output1, prevDelta1, delta1);
-  REQUIRE(arma::accu(delta1) == Approx(21.6).epsilon(1e-3));
+  REQUIRE(accu(delta1) == Approx(21.6).epsilon(1e-3));
   module2.Backward(input, output2, prevDelta2, delta2);
-  REQUIRE(arma::accu(delta2) == Approx(16.2).epsilon(1e-3));
+  REQUIRE(accu(delta2) == Approx(16.2).epsilon(1e-3));
 }
 
 /**
@@ -109,7 +109,7 @@ TEST_CASE("AddMergeAdvanceTestCase", "[A
   r.InputDimensions() = std::vector<size_t>({ 5 });
   r.ComputeOutputDimensions();
   arma::mat rParams(r.WeightSize(), 1);
-  r.SetWeights((double*) rParams.memptr());
+  r.SetWeights(rParams);
   r.Network()[0]->Parameters().fill(2.0);
   ((AddMerge*) r.Network()[1])->Network()[0]->Parameters().fill(-1.0);
 
@@ -117,7 +117,7 @@ TEST_CASE("AddMergeAdvanceTestCase", "[A
   l.InputDimensions() = std::vector<size_t>({ 5 });
   l.ComputeOutputDimensions();
   arma::mat lParams(l.WeightSize(), 1);
-  l.SetWeights((double*) lParams.memptr());
+  l.SetWeights(lParams);
   l.Parameters().fill(1.0);
 
   arma::mat input(arma::randn(5, 10));
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/alpha_dropout.cpp 4.4.0-1/src/mlpack/tests/ann/layer/alpha_dropout.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/alpha_dropout.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/alpha_dropout.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -55,7 +55,7 @@ TEST_CASE("SimpleAlphaDropoutLayerTest",
   // Test the Forward function when testing phase.
   module.Training() = false;
   module.Forward(input, output);
-  REQUIRE(arma::accu(input) == arma::accu(output));
+  REQUIRE(accu(input) == accu(output));
 }
 
 /**
@@ -108,6 +108,6 @@ TEST_CASE("NoAlphaDropoutTest", "[ANNLay
   arma::mat output;
   module.Forward(input, output);
 
-  REQUIRE(arma::accu(output) == arma::accu(input));
+  REQUIRE(accu(output) == accu(input));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/batch_norm.cpp 4.4.0-1/src/mlpack/tests/ann/layer/batch_norm.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/batch_norm.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/batch_norm.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -44,7 +44,7 @@ TEST_CASE("BatchNormTest", "[ANNLayerTes
   module1.ComputeOutputDimensions();
   arma::mat moduleParams(module1.WeightSize(), 1);
   module1.CustomInitialize(moduleParams, module1.WeightSize());
-  module1.SetWeights((double*) moduleParams.memptr());
+  module1.SetWeights(moduleParams);
 
   // BatchNorm layer with average parameter set to false (using momentum).
   BatchNorm module2(2, 2, 1e-5, false);
@@ -53,7 +53,7 @@ TEST_CASE("BatchNormTest", "[ANNLayerTes
   module2.ComputeOutputDimensions();
   arma::mat moduleParams2(module2.WeightSize(), 1);
   module2.CustomInitialize(moduleParams2, module2.WeightSize());
-  module2.SetWeights((double*) moduleParams2.memptr());
+  module2.SetWeights(moduleParams2);
 
   // Training Forward Pass Test.
   output.set_size(module1.OutputSize(), 1);
@@ -250,7 +250,7 @@ TEST_CASE("BatchNormWithMinBatchesTest",
   module1.ComputeOutputDimensions();
   arma::mat moduleParams(module1.WeightSize(), 1);
   module1.CustomInitialize(moduleParams, module1.WeightSize());
-  module1.SetWeights((double*) moduleParams.memptr());
+  module1.SetWeights(moduleParams);
   output.set_size(8, 3);
   module1.Forward(input, output);
   CheckMatrices(output, result, 1e-1);
@@ -275,7 +275,7 @@ TEST_CASE("BatchNormWithMinBatchesTest",
   module2.ComputeOutputDimensions();
   arma::mat moduleParams2(module2.WeightSize(), 1);
   module2.CustomInitialize(moduleParams2, module2.WeightSize());
-  module2.SetWeights((double*) moduleParams2.memptr());
+  module2.SetWeights(moduleParams2);
   output.set_size(8, 3);
   module2.Forward(input, output);
   CheckMatrices(output, result, 1e-1);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/concat.cpp 4.4.0-1/src/mlpack/tests/ann/layer/concat.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/concat.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/concat.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,14 +31,14 @@ TEST_CASE("SimpleConcatLayerTest", "[ANN
   moduleA->InputDimensions() = std::vector<size_t>({ 10 });
   moduleA->ComputeOutputDimensions();
   arma::mat weightsA(moduleA->WeightSize(), 1);
-  moduleA->SetWeights((double*) weightsA.memptr());
+  moduleA->SetWeights(weightsA);
   moduleA->Parameters().randu();
 
   Linear* moduleB = new Linear(10);
   moduleB->InputDimensions() = std::vector<size_t>({ 10 });
   moduleB->ComputeOutputDimensions();
   arma::mat weightsB(moduleB->WeightSize(), 1);
-  moduleB->SetWeights((double*) weightsB.memptr());
+  moduleB->SetWeights(weightsB);
   moduleB->Parameters().randu();
 
   Concat module;
@@ -52,20 +52,20 @@ TEST_CASE("SimpleConcatLayerTest", "[ANN
   output.set_size(module.OutputSize(), 1);
   module.Forward(input, output);
 
-  const double sumModuleA = arma::accu(
+  const double sumModuleA = accu(
       moduleA->Parameters().submat(
       100, 0, moduleA->Parameters().n_elem - 1, 0));
-  const double sumModuleB = arma::accu(
+  const double sumModuleB = accu(
       moduleB->Parameters().submat(
       100, 0, moduleB->Parameters().n_elem - 1, 0));
   REQUIRE(sumModuleA + sumModuleB ==
-      Approx(arma::accu(output.col(0))).epsilon(1e-5));
+      Approx(accu(output.col(0))).epsilon(1e-5));
 
   // Test the Backward function.
   error = arma::zeros(20, 1);
   delta.set_size(input.n_rows, input.n_cols);
   module.Backward(input, output, error, delta);
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 }
 
 /**
@@ -92,13 +92,13 @@ TEST_CASE("ConcatAlongAxisTest", "[ANNLa
   moduleA->InputDimensions() = std::vector<size_t>({ inputWidth, inputHeight });
   moduleA->ComputeOutputDimensions();
   arma::mat weightsA(moduleA->WeightSize(), 1);
-  moduleA->SetWeights((double*) weightsA.memptr());
+  moduleA->SetWeights(weightsA);
   moduleA->Parameters().randu();
 
   moduleB->InputDimensions() = std::vector<size_t>({ inputWidth, inputHeight });
   moduleB->ComputeOutputDimensions();
   arma::mat weightsB(moduleB->WeightSize(), 1);
-  moduleB->SetWeights((double*) weightsB.memptr());
+  moduleB->SetWeights(weightsB);
   moduleB->Parameters().randu();
 
   // Compute output of each layer.
@@ -123,7 +123,7 @@ TEST_CASE("ConcatAlongAxisTest", "[ANNLa
       {
         arma::mat aMat = A.slice(i);
         arma::mat bMat = B.slice(i);
-        calculatedOut.slice(i) = arma::join_cols(aMat, bMat);
+        calculatedOut.slice(i) = join_cols(aMat, bMat);
       }
       x = 2;
     }
@@ -134,7 +134,7 @@ TEST_CASE("ConcatAlongAxisTest", "[ANNLa
       {
         arma::mat aMat = A.slice(i);
         arma::mat bMat = B.slice(i);
-        calculatedOut.slice(i) = arma::join_rows(aMat, bMat);
+        calculatedOut.slice(i) = join_rows(aMat, bMat);
       }
       y = 2;
     }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/concatenate.cpp 4.4.0-1/src/mlpack/tests/ann/layer/concatenate.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/concatenate.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/concatenate.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,12 +38,12 @@ TEST_CASE("SimpleConcatenateLayerTest",
   output.set_size(module.OutputSize(), 1);
   module.Forward(input, output);
 
-  REQUIRE(arma::accu(output) == 7.5);
+  REQUIRE(accu(output) == 7.5);
 
   // Test the Backward function.
   delta.set_size(5, 1);
   module.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 5);
+  REQUIRE(accu(delta) == 5);
 }
 
 /**
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/convolution.cpp 4.4.0-1/src/mlpack/tests/ann/layer/convolution.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/convolution.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/convolution.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -77,7 +77,7 @@ TEST_CASE("ConvolutionLayerPaddingTest",
   module1.ComputeOutputDimensions();
   arma::mat weights1(module1.WeightSize(), 1);
   REQUIRE(weights1.n_elem == 10);
-  module1.SetWeights(weights1.memptr());
+  module1.SetWeights(weights1);
 
   // Test the Forward function.
   input = arma::linspace<arma::colvec>(0, 48, 49);
@@ -85,7 +85,7 @@ TEST_CASE("ConvolutionLayerPaddingTest",
   module1.Parameters().zeros();
   module1.Forward(input, output);
 
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == 25);
   REQUIRE(output.n_cols == 1);
 
@@ -100,7 +100,7 @@ TEST_CASE("ConvolutionLayerPaddingTest",
   module2.ComputeOutputDimensions();
   arma::mat weights2(module2.WeightSize(), 1);
   REQUIRE(weights2.n_elem == 10);
-  module2.SetWeights(weights2.memptr());
+  module2.SetWeights(weights2);
 
   // Test the forward function.
   input = arma::linspace<arma::colvec>(0, 48, 49);
@@ -108,7 +108,7 @@ TEST_CASE("ConvolutionLayerPaddingTest",
   module2.Parameters().zeros();
   module2.Forward(input, output);
 
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == 49);
   REQUIRE(output.n_cols == 1);
 
@@ -218,7 +218,7 @@ TEST_CASE("ConvolutionLayerTestCase", "[
   layer.InputDimensions() = std::vector<size_t>({ 4, 1, 2 });
   layer.ComputeOutputDimensions();
   arma::mat layerWeights(layer.WeightSize(), 1);
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   output.set_size(layer.OutputSize(), 3);
 
   // Set weights to 1.0 and bias to 0.0.
@@ -227,14 +227,14 @@ TEST_CASE("ConvolutionLayerTestCase", "[
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == 4108);
+  REQUIRE(accu(output) == 4108);
 
   // Set bias to one.
   layer.Bias().fill(1.0);
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == 4156);
+  REQUIRE(accu(output) == 4156);
 }
 
 /**
@@ -250,7 +250,7 @@ TEST_CASE("ConvolutionLayerWeightInitial
   module.InputDimensions() = std::vector<size_t>({ 12, 13, 2 });
   module.ComputeOutputDimensions();
   arma::mat weights(module.WeightSize(), 1);
-  module.SetWeights(weights.memptr());
+  module.SetWeights(weights);
 
   RandomInitialization().Initialize(module.Weight());
   module.Bias().ones();
@@ -290,7 +290,7 @@ TEST_CASE("NoBiasConvolutionLayerTestCas
   layer.ComputeOutputDimensions();
   REQUIRE(layer.WeightSize() == 8);
   arma::mat layerWeights(layer.WeightSize(), 1);
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   REQUIRE(layer.Bias().n_elem == 0);
   output.set_size(layer.OutputSize(), 3);
 
@@ -300,14 +300,14 @@ TEST_CASE("NoBiasConvolutionLayerTestCas
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == 4108);
+  REQUIRE(accu(output) == 4108);
 
   // Set bias to one.
   layer.Bias().fill(1.0);
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == 4108);
+  REQUIRE(accu(output) == 4108);
 }
 
 /**
@@ -350,18 +350,18 @@ TEST_CASE("AdvancedConvolutionLayerTest"
   layerWeights(15) = -0.2283206;
   layerWeights(16) = 0.3204123974;
   layerWeights(17) = 0.2334779799;
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   output.set_size(layer.OutputSize(), 3);
 
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == Approx(12.6755657196).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(12.6755657196).epsilon(1e-5));
 
   arma::mat delta;
   delta.set_size(8, 3);
   layer.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == Approx(-1.9237523079).epsilon(1e-5));
+  REQUIRE(accu(delta) == Approx(-1.9237523079).epsilon(1e-5));
 }
 
 /**
@@ -428,18 +428,18 @@ TEST_CASE("AdvancedConvolutionLayerWithS
   layerWeights(15) = -0.12563613;
   layerWeights(16) = -0.1114468053;
   layerWeights(17) = -0.3029643595;
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   output.set_size(layer.OutputSize(), 3);
 
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == Approx(364.7379150391).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(364.7379150391).epsilon(1e-5));
 
   arma::mat delta;
   delta.set_size(32, 3);
   layer.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == Approx(115.3515701294).epsilon(1e-5));
+  REQUIRE(accu(delta) == Approx(115.3515701294).epsilon(1e-5));
 }
 
 // Make a simple convolutional layer with non-square filters, and make sure the
@@ -451,7 +451,7 @@ TEST_CASE("NonSquareConvolutionTest", "[
   module1.InputDimensions() = std::vector<size_t>({ 7, 7 });
   module1.ComputeOutputDimensions();
   arma::mat weights1(module1.WeightSize(), 1);
-  module1.SetWeights(weights1.memptr());
+  module1.SetWeights(weights1);
 
   arma::mat data(49, 10, arma::fill::randu);
   arma::mat forwardResult(module1.OutputSize(), 10, arma::fill::zeros);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/dropout.cpp 4.4.0-1/src/mlpack/tests/ann/layer/dropout.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/dropout.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/dropout.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ TEST_CASE("SimpleDropoutLayerTest", "[AN
   // Test the Forward function.
   module.Training() = false;
   module.Forward(input, output);
-  REQUIRE(arma::accu(input) == arma::accu(output));
+  REQUIRE(accu(input) == accu(output));
 }
 
 /**
@@ -97,6 +97,6 @@ TEST_CASE("NoDropoutTest", "[ANNLayerTes
   arma::mat output;
   module.Forward(input, output);
 
-  REQUIRE(arma::accu(output) == arma::accu(input));
+  REQUIRE(accu(output) == accu(input));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/flexible_relu.cpp 4.4.0-1/src/mlpack/tests/ann/layer/flexible_relu.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/flexible_relu.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/flexible_relu.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -35,7 +35,7 @@ TEST_CASE("JacobianFlexibleReLULayerTest
     FlexibleReLU module;
     arma::mat moduleParams(module.WeightSize(), 1);
     module.CustomInitialize(moduleParams, module.WeightSize());
-    module.SetWeights((double*) moduleParams.memptr());
+    module.SetWeights(moduleParams);
 
     double error = JacobianTest(module, input);
     REQUIRE(error <= 1e-5);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/ftswish.cpp 4.4.0-1/src/mlpack/tests/ann/layer/ftswish.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/ftswish.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/ftswish.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -44,7 +44,7 @@ TEST_CASE("FTSwishTest", "[ANNLayerTest]
   layer.Forward(input, output);
 
   // Test the Forward function
-  REQUIRE(abs(arma::accu(output - actualOutput)) <= 0.0001);
+  REQUIRE(abs(accu(output - actualOutput)) <= 0.0001);
 
   arma::mat delta = {{0.615941, 0.989097, 0.0},
     {1.03315, 1.09083 ,0.0 },
@@ -58,5 +58,5 @@ TEST_CASE("FTSwishTest", "[ANNLayerTest]
   layer.Backward(input, output, gy, g);
 
   //Test the Backward function
-  REQUIRE(abs(arma::accu(g - delta)) <= 0.0001);
+  REQUIRE(abs(accu(g - delta)) <= 0.0001);
 }
\ No newline at end of file
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/grouped_convolution.cpp 4.4.0-1/src/mlpack/tests/ann/layer/grouped_convolution.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/grouped_convolution.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/grouped_convolution.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,18 +49,18 @@ TEST_CASE("GroupedConvolutionLayerTest",
   layerWeights(5) = -0.7586858273;
   layerWeights(6) = -0.1721059084;
   layerWeights(7) = -0.1972532272;
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   output.set_size(layer.OutputSize(), 3);
 
   layer.Forward(input, output);
 
   // Value calculated using torch.nn.Conv2d().
-  REQUIRE(arma::accu(output) == Approx(-647.6369628906).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(-647.6369628906).epsilon(1e-5));
 
   arma::mat delta;
   delta.set_size(8, 3);
   layer.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == Approx(686.7855224609).epsilon(1e-5));
+  REQUIRE(accu(delta) == Approx(686.7855224609).epsilon(1e-5));
 }
 
 /**
@@ -104,7 +104,7 @@ TEST_CASE("GroupedConvolutionEquivalence
   layerWeights(15) = -0.2283206;
   layerWeights(16) = 0.3204123974;
   layerWeights(17) = 0.2334779799;
-  layer.SetWeights(layerWeights.memptr());
+  layer.SetWeights(layerWeights);
   output.set_size(layer.OutputSize(), 3);
 
   GroupedConvolution layerG(2, 2, 2, 1, 1, 1, 0, 0);
@@ -129,7 +129,7 @@ TEST_CASE("GroupedConvolutionEquivalence
   layerWeightsG(15) = -0.2283206;
   layerWeightsG(16) = 0.3204123974;
   layerWeightsG(17) = 0.2334779799;
-  layerG.SetWeights(layerWeightsG.memptr());
+  layerG.SetWeights(layerWeightsG);
   outputG.set_size(layerG.OutputSize(), 3);
 
   layer.Forward(input, output);
@@ -137,8 +137,8 @@ TEST_CASE("GroupedConvolutionEquivalence
 
   // Value calculated using torch.nn.Conv2d().
   CheckMatrices(output, outputG);
-  REQUIRE(arma::accu(output) == Approx(12.6755657196).epsilon(1e-5));
-  REQUIRE(arma::accu(outputG) == Approx(12.6755657196).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(12.6755657196).epsilon(1e-5));
+  REQUIRE(accu(outputG) == Approx(12.6755657196).epsilon(1e-5));
 
   arma::mat delta, deltaG;
   delta.set_size(8, 3);
@@ -147,8 +147,8 @@ TEST_CASE("GroupedConvolutionEquivalence
   layerG.Backward(input, outputG, outputG, deltaG);
 
   CheckMatrices(delta, deltaG);
-  REQUIRE(arma::accu(delta) == Approx(-1.9237523079).epsilon(1e-5));
-  REQUIRE(arma::accu(deltaG) == Approx(-1.9237523079).epsilon(1e-5));
+  REQUIRE(accu(delta) == Approx(-1.9237523079).epsilon(1e-5));
+  REQUIRE(accu(deltaG) == Approx(-1.9237523079).epsilon(1e-5));
 }
 
 TEST_CASE("EdgeCaseFailGroupedConvolutionTest", "[ANNLayerTest]")
@@ -212,7 +212,7 @@ TEST_CASE("NonSquareGroupedConvolutionTe
   module1.InputDimensions() = std::vector<size_t>({ 7, 7 });
   module1.ComputeOutputDimensions();
   arma::mat weights1(module1.WeightSize(), 1);
-  module1.SetWeights(weights1.memptr());
+  module1.SetWeights(weights1);
 
   arma::mat data(49, 10, arma::fill::randu);
   arma::mat forwardResult(module1.OutputSize(), 10, arma::fill::zeros);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/hard_tanh.cpp 4.4.0-1/src/mlpack/tests/ann/layer/hard_tanh.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/hard_tanh.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/hard_tanh.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -43,7 +43,7 @@ TEST_CASE("SimpleHardTanHTest", "[ANNLay
   output.set_size(5,5);
   // Test the Forward function
   module.Forward(input, output);
-  REQUIRE(arma::accu(output - actualOutput) == Approx(0).epsilon(1e-4));
+  REQUIRE(accu(output - actualOutput) == Approx(0).epsilon(1e-4));
   
   arma::mat delta = {{0  , 1.0, 1.0, 1.0, 0.0},
     {0  , 0  , 0  , 0.0, 0.0},
@@ -57,6 +57,6 @@ TEST_CASE("SimpleHardTanHTest", "[ANNLay
   
   //Test the Backward function
   module.Backward(input, output, gy, g);
-  REQUIRE(arma::accu(g - delta) == Approx(0).epsilon(1e-4));
+  REQUIRE(accu(g - delta) == Approx(0).epsilon(1e-4));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/layer_norm.cpp 4.4.0-1/src/mlpack/tests/ann/layer/layer_norm.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/layer_norm.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/layer_norm.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ TEST_CASE("LayerNormTest", "[ANNLayerTes
   model.InputDimensions() = std::vector<size_t>({ 3 });
   model.ComputeOutputDimensions();
   arma::mat weights(model.WeightSize(), 1);
-  model.SetWeights(weights.memptr());
+  model.SetWeights(weights);
   model.CustomInitialize(weights, model.WeightSize());
 
   model.Forward(input, output);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/linear3d.cpp 4.4.0-1/src/mlpack/tests/ann/layer/linear3d.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/linear3d.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/linear3d.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,7 +36,7 @@ TEST_CASE("SimpleLinear3DLayerTest", "[A
   module.InputDimensions() = std::vector<size_t>({ 4, 2 });
   module.ComputeOutputDimensions();
   arma::mat weights(module.WeightSize(), 1);
-  module.SetWeights(weights.memptr());
+  module.SetWeights(weights);
 
   module.Parameters().randu();
 
@@ -44,14 +44,14 @@ TEST_CASE("SimpleLinear3DLayerTest", "[A
   input = arma::zeros(inSize * nPoints, batchSize);
   output.set_size(outSize * nPoints, batchSize);
   module.Forward(input, output);
-  REQUIRE(arma::accu(module.Bias())
-      == Approx(arma::accu(output) / (nPoints * batchSize)).epsilon(1e-3));
+  REQUIRE(accu(module.Bias())
+      == Approx(accu(output) / (nPoints * batchSize)).epsilon(1e-3));
 
   // Test the Backward function.
   delta.set_size(input.n_rows, input.n_cols);
   output.zeros();
   module.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 }
 
 /**
@@ -74,7 +74,7 @@ TEST_CASE("JacobianLinear3DLayerTest", "
     module.InputDimensions() = std::vector<size_t>({ inSize, nPoints });
     module.ComputeOutputDimensions();
     arma::mat weights(module.WeightSize(), 1);
-    module.SetWeights(weights.memptr());
+    module.SetWeights(weights);
 
     module.Parameters().randu();
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/linear_no_bias.cpp 4.4.0-1/src/mlpack/tests/ann/layer/linear_no_bias.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/linear_no_bias.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/linear_no_bias.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,18 +30,18 @@ TEST_CASE("SimpleLinearNoBiasLayerTest",
   arma::mat weights(10 * 10, 1);
   module.InputDimensions() = std::vector<size_t>({ 10 });
   module.ComputeOutputDimensions();
-  module.SetWeights(weights.memptr());
+  module.SetWeights(weights);
 
   module.Parameters().randu();
 
   // Test the Forward function.
   input = arma::zeros(10, 1);
   module.Forward(input, output);
-  REQUIRE(0 == arma::accu(output));
+  REQUIRE(0 == accu(output));
 
   // Test the Backward function.
   module.Backward(input, output, output, delta);
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 }
 
 /**
@@ -61,7 +61,7 @@ TEST_CASE("JacobianLinearNoBiasLayerTest
     arma::mat weights(inputElements * outputElements, 1);
     module.InputDimensions() = std::vector<size_t>({ inputElements });
     module.ComputeOutputDimensions();
-    module.SetWeights(weights.memptr());
+    module.SetWeights(weights);
 
     module.Parameters().randu();
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/log_softmax.cpp 4.4.0-1/src/mlpack/tests/ann/layer/log_softmax.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/log_softmax.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/log_softmax.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,14 +31,14 @@ TEST_CASE("SimpleLogSoftmaxLayerTest", "
   // Test the Forward function.
   input = arma::mat("-0.6871; 0.7898; 0.2011; 0.0949; -0.0550");
   module.Forward(input, output);
-  REQUIRE(arma::accu(arma::abs(arma::mat("-2.4746; -0.9977; -1.5864; -1.6926; -1.8425") - output)) ==
+  REQUIRE(accu(arma::abs(arma::mat("-2.4746; -0.9977; -1.5864; -1.6926; -1.8425") - output)) ==
       Approx(0.0).margin(1e-3));
 
   // Test the Backward function.
   error = arma::ones(input.n_rows, input.n_cols);
   // Assume LogSoftmax layer is always associated with NLL output layer.
   module.Backward(input, output, error, delta);
-  REQUIRE(arma::accu(arma::abs(arma::mat("0.5790; -0.8435; -0.0233; 0.0798; 0.2079") - delta)) ==
+  REQUIRE(accu(arma::abs(arma::mat("0.5790; -0.8435; -0.0233; 0.0798; 0.2079") - delta)) ==
       Approx(0.0).margin(1e-3));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/max_pooling.cpp 4.4.0-1/src/mlpack/tests/ann/layer/max_pooling.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/max_pooling.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/max_pooling.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -47,7 +47,7 @@ TEST_CASE("MaxPoolingTestCase", "[ANNLay
   module1.ComputeOutputDimensions();
   module1.Forward(input, output);
   // Calculated using torch.nn.MaxPool2d().
-  REQUIRE(arma::accu(output) == 28);
+  REQUIRE(accu(output) == 28);
   REQUIRE(output.n_elem == 4);
   REQUIRE(output.n_cols == 1);
 
@@ -68,7 +68,7 @@ TEST_CASE("MaxPoolingTestCase", "[ANNLay
   module2.ComputeOutputDimensions();
   module2.Forward(input, output);
   // Calculated using torch.nn.MaxPool2d().
-  REQUIRE(arma::accu(output) == 12.0);
+  REQUIRE(accu(output) == 12.0);
   REQUIRE(output.n_elem == 2);
   REQUIRE(output.n_cols == 1);
 
@@ -89,7 +89,7 @@ TEST_CASE("MaxPoolingTestCase", "[ANNLay
   module3.ComputeOutputDimensions();
   module3.Forward(input, output);
   // Calculated using torch.nn.MaxPool2d().
-  REQUIRE(arma::accu(output) == 30.0);
+  REQUIRE(accu(output) == 30.0);
   REQUIRE(output.n_elem == 9);
   REQUIRE(output.n_cols == 1);
 
@@ -108,7 +108,7 @@ TEST_CASE("MaxPoolingTestCase", "[ANNLay
   module4.ComputeOutputDimensions();
   module4.Forward(input, output);
   // Calculated using torch.nn.MaxPool2d().
-  REQUIRE(arma::accu(output) == 3);
+  REQUIRE(accu(output) == 3);
   REQUIRE(output.n_elem == 4);
   REQUIRE(output.n_cols == 1);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/mean_pooling.cpp 4.4.0-1/src/mlpack/tests/ann/layer/mean_pooling.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/mean_pooling.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/mean_pooling.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -59,9 +59,9 @@ TEST_CASE("MeanPoolingTestCase", "[ANNLa
   output1.set_size(8, 1);
   output2.set_size(6, 1);
   module1.Forward(input, output1);
-  REQUIRE(arma::accu(output1) == 25.5);
+  REQUIRE(accu(output1) == 25.5);
   module2.Forward(input, output2);
-  REQUIRE(arma::accu(output2) == 19.5);
+  REQUIRE(accu(output2) == 19.5);
   output1.reshape(4, 2);
   output2.reshape(3, 2);
   CheckMatrices(output1, result1, 1e-1);
@@ -82,7 +82,7 @@ TEST_CASE("MeanPoolingTestCase", "[ANNLa
   prevDelta1.reshape(8, 1);
   prevDelta2.reshape(6, 1);
   module1.Backward(input, output1, prevDelta1, delta1);
-  REQUIRE(arma::accu(delta1) == Approx(10.8).epsilon(1e-3));
+  REQUIRE(accu(delta1) == Approx(10.8).epsilon(1e-3));
   module2.Backward(input, output2, prevDelta2, delta2);
-  REQUIRE(arma::accu(delta2) == Approx(8.1).epsilon(1e-3));
+  REQUIRE(accu(delta2) == Approx(8.1).epsilon(1e-3));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/multihead_attention.cpp 4.4.0-1/src/mlpack/tests/ann/layer/multihead_attention.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/multihead_attention.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/multihead_attention.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,13 +53,13 @@ TEST_CASE("SimpleMultiheadAttentionTest"
   module.ComputeOutputDimensions();
   arma::mat weights(module.WeightSize(), 1);
   weights.randu();
-  module.SetWeights(weights.memptr());
+  module.SetWeights(weights);
 
   module.AttentionMask() = attnMask;
   module.KeyPaddingMask() = keyPaddingMask;
 
   // Forward test.
-  arma::mat input = arma::join_cols(arma::join_cols(query, query), query);
+  arma::mat input = join_cols(join_cols(query, query), query);
 
   module.Forward(input, output);
   REQUIRE(output.n_rows == embedDim * tLen);
@@ -94,14 +94,14 @@ TEST_CASE("JacobianMultiheadAttentionTes
     const size_t batchSize = 1;
 
     arma::mat query = arma::randu(embedDim * tgtSeqLen, batchSize);
-    arma::mat input = arma::join_cols(arma::join_cols(query, query), query);
+    arma::mat input = join_cols(join_cols(query, query), query);
 
     MultiheadAttention module(tgtSeqLen, nHeads);
     module.InputDimensions() = std::vector<size_t>({ embedDim, 3 * tgtSeqLen });
     module.ComputeOutputDimensions();
     arma::mat weights(module.WeightSize(), 1);
     weights.randu();
-    module.SetWeights(weights.memptr());
+    module.SetWeights(weights);
 
     double error = CustomJacobianTest(module, input);
     REQUIRE(error <= 1e-5);
@@ -118,14 +118,14 @@ TEST_CASE("JacobianMultiheadAttentionTes
 
     arma::mat query = arma::randu(embedDim * tgtSeqLen, batchSize);
     arma::mat key = 0.091 * arma::randu(embedDim * srcSeqLen, batchSize);
-    arma::mat input = arma::join_cols(arma::join_cols(query, key), key);
+    arma::mat input = join_cols(join_cols(query, key), key);
 
     MultiheadAttention module(tgtSeqLen, nHeads);
     module.InputDimensions() = std::vector<size_t>({ embedDim, 2 * srcSeqLen + tgtSeqLen });
     module.ComputeOutputDimensions();
     arma::mat weights(module.WeightSize(), 1);
     weights.randu();
-    module.SetWeights(weights.memptr());
+    module.SetWeights(weights);
 
     double error = CustomJacobianTest(module, input);
     REQUIRE(error <= 1e-5);
@@ -143,14 +143,14 @@ TEST_CASE("JacobianMultiheadAttentionTes
     arma::mat query = arma::randu(embedDim * tgtSeqLen, batchSize);
     arma::mat key = 0.091 * arma::randu(embedDim * srcSeqLen, batchSize);
     arma::mat value = 0.045 * arma::randu(embedDim * srcSeqLen, batchSize);
-    arma::mat input = arma::join_cols(arma::join_cols(query, key), value);
+    arma::mat input = join_cols(join_cols(query, key), value);
 
     MultiheadAttention module(tgtSeqLen, nHeads);
     module.InputDimensions() = std::vector<size_t>({ embedDim, 2 * srcSeqLen + tgtSeqLen });
     module.ComputeOutputDimensions();
     arma::mat weights(module.WeightSize(), 1);
     weights.randu();
-    module.SetWeights(weights.memptr());
+    module.SetWeights(weights);
 
     double error = JacobianTest(module, input);
     REQUIRE(error <= 1e-5);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/padding.cpp 4.4.0-1/src/mlpack/tests/ann/layer/padding.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/padding.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/padding.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -38,7 +38,7 @@ TEST_CASE("SimplePaddingLayerTest", "[AN
   output.set_size(totalOutputDimensions, input.n_cols);
   output.randu();
   module.Forward(input, output);
-  REQUIRE(arma::accu(input) == Approx(arma::accu(output)));
+  REQUIRE(accu(input) == Approx(accu(output)));
   REQUIRE(output.n_rows == (5 * 12)); // 2x5 --> 5x12
 
   // Test the Backward function.
@@ -60,7 +60,7 @@ TEST_CASE("SimplePaddingLayerTest", "[AN
   output.set_size(totalOutputDimensions, input.n_cols);
   output.randu();
   module.Forward(input, output);
-  REQUIRE(arma::accu(input) == Approx(arma::accu(output)));
+  REQUIRE(accu(input) == Approx(accu(output)));
   REQUIRE(output.n_rows == (226 * 226 * 3));
   REQUIRE(output.n_cols == 1);
 
@@ -79,5 +79,5 @@ TEST_CASE("SimplePaddingLayerTest", "[AN
   module.Forward(input, output);
   REQUIRE(output.n_rows == (246 * 246 * 3));
   REQUIRE(output.n_cols == 3);
-  REQUIRE(arma::accu(input) == Approx(arma::accu(output)));
+  REQUIRE(accu(input) == Approx(accu(output)));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/parametric_relu.cpp 4.4.0-1/src/mlpack/tests/ann/layer/parametric_relu.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/parametric_relu.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/parametric_relu.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,14 +31,14 @@ TEST_CASE("PReLUFORWARDTest", "[ANNLayer
   PReLU module(0.01);
   arma::mat moduleParams(module.WeightSize(), 1);
   module.CustomInitialize(moduleParams, module.WeightSize());
-  module.SetWeights((double*) moduleParams.memptr());
+  module.SetWeights(moduleParams);
   arma::mat predOutput;
   module.Forward(input, predOutput);
   arma::mat actualOutput = {{0.5, 1.2, 3.1},
                            {-0.022, -0.015, 0.8},
                            {5.5, -0.047, 2.1},
                            {0.2, 0.1, -0.005}};
-  REQUIRE(arma::accu(arma::abs(actualOutput - predOutput)) ==
+  REQUIRE(accu(arma::abs(actualOutput - predOutput)) ==
       Approx(0.0).margin(1e-4));
 }
 
@@ -54,7 +54,7 @@ TEST_CASE("PReLUBACKWARDTest", "[ANNLaye
   PReLU module(0.01);
   arma::mat moduleParams(module.WeightSize(), 1);
   module.CustomInitialize(moduleParams, module.WeightSize());
-  module.SetWeights((double*) moduleParams.memptr());
+  module.SetWeights(moduleParams);
   arma::mat gy = {{0.2, -0.5, 0.8},
                  {1.5, -0.6, 0.1},
                  {-0.3, 0.2, -0.5},
@@ -68,7 +68,7 @@ TEST_CASE("PReLUBACKWARDTest", "[ANNLaye
                       {-0.3, 0.002, -0.5},
                       {0.1, -0.1, 0.0030}};
 
-  REQUIRE(arma::accu(arma::abs(actualG - predG)) ==
+  REQUIRE(accu(arma::abs(actualG - predG)) ==
       Approx(0.0).margin(1e-4));
 }
 
@@ -84,7 +84,7 @@ TEST_CASE("PReLUGRADIENTTest", "[ANNLaye
   PReLU module(0.01);
   arma::mat moduleParams(module.WeightSize(), 1);
   module.CustomInitialize(moduleParams, module.WeightSize());
-  module.SetWeights((double*) moduleParams.memptr());
+  module.SetWeights(moduleParams);
   arma::mat error = {{0.2, -0.5,  0.8},
                     {-0.015, -0.006, 0.001},
                     {-0.3,  0.002, -0.005},
@@ -92,13 +92,13 @@ TEST_CASE("PReLUGRADIENTTest", "[ANNLaye
   arma::mat predGradient;
   module.Gradient(input, error, predGradient);
 
-  REQUIRE(0.0103 - arma::accu(predGradient) ==
+  REQUIRE(0.0103 - accu(predGradient) ==
       Approx(0.0).margin(1e-4));
 }
 
 double ComputeMSRE(arma::mat input, arma::mat target)
 {
-  return std::pow(arma::accu(arma::pow(input - target, 2)) / target.n_cols, 0.5);
+  return std::pow(accu(pow(input - target, 2)) / target.n_cols, 0.5);
 }
 
 TEST_CASE("PReLUIntegrationTest", "[ANNLayerTest]")
@@ -132,8 +132,6 @@ TEST_CASE("PReLUIntegrationTest", "[ANNL
     double msreTrain = ComputeMSRE(predictions, trainLabels);
     model.Predict(testData, predictions);
     double msreTest = ComputeMSRE(predictions, testLabels);
-    std::cout << "train: " << msreTrain << "\n";
-    std::cout << "test: " << msreTest << "\n";
 
     double relativeMSRE = std::abs((msreTest - msreTrain) / msreTrain);
     if (relativeMSRE <= 0.35)
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/repeat.cpp 4.4.0-1/src/mlpack/tests/ann/layer/repeat.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/repeat.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/repeat.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -0,0 +1,362 @@
+/**
+ * @file tests/ann/layer/repeat.cpp
+ * @author Adam Kropp
+ *
+ * Tests the repeat layer.
+ *
+ * mlpack is free software; you may redistribute it and/or modify it under the
+ * terms of the 3-clause BSD license.  You should have received a copy of the
+ * 3-clause BSD license along with mlpack.  If not, see
+ * http://www.opensource.org/licenses/BSD-3-Clause for more information.
+ */
+#include <mlpack/core.hpp>
+#include <mlpack/methods/ann.hpp>
+
+#include "../../test_catch_tools.hpp"
+#include "../../catch.hpp"
+#include "../../serialization.hpp"
+#include "../ann_test_tools.hpp"
+
+using namespace mlpack;
+
+/**
+ * Simple test for Repeat layer using interleaved repeats along axis 0.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseI0", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repelem(input, 2, 1);
+  input.reshape(12, 1);
+  target.reshape(24, 1);
+
+  MatType output;
+  // Output-Size should be 8 x 3.
+  output.set_size(24, 1);
+
+  RepeatType<MatType> module1({2, 1}, true);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 8);
+  REQUIRE(module1.OutputDimensions()[1] == 3);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 24);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(24, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(8, 3);
+  MatType targetDelta(4, 3);
+  for (size_t i = 0; i < targetDelta.n_rows; i++)
+  {
+    targetDelta.row(i) = arma::sum(prevDelta.rows(i * 2, i * 2 + 1), 0) / 2;
+  }
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+/**
+ * Simple test for Repeat layer using interleaved repeats along axis 1.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseI1", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repelem(input, 1, 2);
+  input.reshape(12, 1);
+  target.reshape(24, 1);
+
+  MatType output;
+  // Output-Size should be 4 x 6.
+  output.set_size(24, 1);
+
+  RepeatType<MatType> module1({1, 2}, true);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 4);
+  REQUIRE(module1.OutputDimensions()[1] == 6);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 24);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(24, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(4, 6);
+  MatType targetDelta(4, 3);
+  for (size_t i = 0; i < targetDelta.n_cols; i++)
+  {
+    targetDelta.col(i) = arma::sum(prevDelta.cols(i * 2, i * 2 + 1), 1) / 2;
+  }
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+/**
+ * Simple test for Repeat layer using interleaved repeats along axis 0 and 1.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseI2", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repelem(input, 2, 2);
+  input.reshape(12, 1);
+  target.reshape(48, 1);
+
+  MatType output;
+  // Output-Size should be 8 x 6.
+  output.set_size(48, 1);
+
+  RepeatType<MatType> module1({2, 2}, true);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 8);
+  REQUIRE(module1.OutputDimensions()[1] == 6);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 48);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(48, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(8, 6);
+  MatType targetDelta(4, 3);
+  for (size_t i = 0; i < targetDelta.n_rows; i++)
+  {
+    for (size_t j = 0; j < targetDelta.n_cols; j++)
+    {
+      targetDelta.at(i, j) = (prevDelta.at(i * 2, j * 2)
+          + prevDelta.at(i * 2 + 1, j * 2)
+          + prevDelta.at(i * 2, j * 2 + 1)
+          + prevDelta.at(i * 2 + 1, j * 2 + 1)) / 4;
+    }
+  }
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+/**
+ * Simple test for Repeat layer using block repeats.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseB1", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repmat(input, 2, 1);
+  input.reshape(12, 1);
+  target.reshape(24, 1);
+
+  MatType output;
+  // Output-Size should be 8 x 3.
+  output.set_size(24, 1);
+
+  RepeatType<MatType> module1({2, 1}, false);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 8);
+  REQUIRE(module1.OutputDimensions()[1] == 3);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 24);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(24, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(8, 3);
+  MatType targetDelta = (prevDelta.submat(0, 0, 3, 2) +
+      prevDelta.submat(4, 0, 7, 2)) / 2;
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+/**
+ * Simple test for Replicate layer using block repeats.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseB2", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repmat(input, 1, 2);
+  input.reshape(12, 1);
+  target.reshape(24, 1);
+
+  MatType output;
+  // Output-Size should be 4 x 6.
+  output.set_size(24, 1);
+
+  RepeatType<MatType> module1({1, 2}, false);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 4);
+  REQUIRE(module1.OutputDimensions()[1] == 6);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 24);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(24, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(4, 6);
+  MatType targetDelta = (prevDelta.submat(0, 0, 3, 2)
+                         + prevDelta.submat(0, 3, 3, 5)) / 2;
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+/**
+ * Simple test for Replicate layer using block repeats.
+ */
+TEMPLATE_TEST_CASE("RepeatTestCaseB3", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Input will be 4 x 3.
+  MatType input(4, 3, arma::fill::randn);
+  MatType target = arma::repmat(input, 2, 2);
+  input.reshape(12, 1);
+  target.reshape(48, 1);
+
+  MatType output;
+  // Output-Size should be 8 x 6.
+  output.set_size(48, 1);
+
+  RepeatType<MatType> module1({2, 2}, false);
+  module1.InputDimensions() = std::vector<size_t>({ 4, 3 });
+  module1.ComputeOutputDimensions();
+  REQUIRE(module1.OutputDimensions().size() == 2);
+  REQUIRE(module1.OutputDimensions()[0] == 8);
+  REQUIRE(module1.OutputDimensions()[1] == 6);
+  module1.Forward(input, output);
+  CheckMatrices(output, target, 1e-1);
+  REQUIRE(output.n_elem == 48);
+  REQUIRE(output.n_cols == 1);
+
+  MatType prevDelta(48, 1, arma::fill::randn);
+  MatType delta;
+  delta.set_size(12, 1);
+  module1.Backward(input, output, prevDelta, delta);
+  prevDelta.reshape(8, 6);
+  MatType targetDelta = (prevDelta.submat(0, 0, 3, 2)
+                         + prevDelta.submat(4, 0, 7, 2)
+                         + prevDelta.submat(0, 3, 3, 5)
+                         + prevDelta.submat(4, 3, 7, 5)) / 4;
+  targetDelta.reshape(12, 1);
+  CheckMatrices(delta, targetDelta, 1e-1);
+}
+
+template <typename F> struct GradientBound {};
+template <> struct GradientBound<arma::mat>
+{
+  static constexpr double eps = 1e-5;
+  static constexpr double bound = 1e-4;
+};
+template <> struct GradientBound<arma::fmat>
+{
+  static constexpr double eps = 1e-3;
+  static constexpr double bound = 3e-3;
+};
+
+/**
+ * Numerical gradient test for MultiheadAttention layer.
+ */
+TEMPLATE_TEST_CASE("GradientRepeatTest", "[ANNLayerTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+  struct GradientFunction
+  {
+    GradientFunction(std::vector<size_t> multiples, bool interleave) :
+        inputDimensions({3, 4}),
+        multiples(std::move(multiples)),
+        interleave(interleave),
+        outputSize(8),
+        vocabSize(5),
+        batchSize(20)
+    {
+      size_t inputSize = std::accumulate(inputDimensions.begin(),
+                                         inputDimensions.end(), 1,
+                                         std::multiplies<>());
+      input = arma::randu<MatType>(inputSize, batchSize);
+      target = arma::zeros<MatType>(vocabSize, batchSize);
+      for (size_t i = 0; i < target.n_elem; ++i)
+      {
+        const size_t label = RandInt(1, vocabSize);
+        target(i) = label;
+      }
+
+      model = new FFN<NegativeLogLikelihoodType<MatType>,
+                      RandomInitialization, MatType>();
+      model->InputDimensions() = inputDimensions;
+      model->ResetData(input, target);
+      model->template Add<RepeatType<MatType>>(multiples, interleave);
+      model->template Add<LinearType<MatType>>(vocabSize);
+      model->template Add<LogSoftMaxType<MatType>>();
+    }
+
+    ~GradientFunction()
+    {
+      delete model;
+    }
+
+    double Gradient(MatType& gradient)
+    {
+      double error = model->Evaluate(model->Parameters(), 0, batchSize);
+      model->Gradient(model->Parameters(), 0, gradient, batchSize);
+      return error;
+    }
+
+    MatType& Parameters() { return model->Parameters(); }
+
+    FFN<NegativeLogLikelihoodType<MatType>,RandomInitialization,MatType>* model;
+
+    MatType input, target;
+    const std::vector<size_t> inputDimensions;
+    const std::vector<size_t> multiples;
+    const bool interleave;
+    const size_t outputSize;
+    const size_t vocabSize;
+    const size_t batchSize;
+    size_t count;
+  };
+  GradientFunction fn1({2, 1}, true);
+  GradientFunction fn2({2, 1}, false);
+  GradientFunction fn3({1, 2}, true);
+  GradientFunction fn4({1, 2}, false);
+  GradientFunction fn5( {2, 2}, true);
+  GradientFunction fn6( {2, 2}, false);
+  
+  double eps = GradientBound<MatType>::eps;
+  double bound = GradientBound<MatType>::bound;
+
+  CHECK(CheckGradient<GradientFunction,MatType>(fn1, eps) <= bound);
+  CHECK(CheckGradient<GradientFunction,MatType>(fn2, eps) <= bound);
+  CHECK(CheckGradient<GradientFunction,MatType>(fn3, eps) <= bound);
+  CHECK(CheckGradient<GradientFunction,MatType>(fn4, eps) <= bound);
+  CHECK(CheckGradient<GradientFunction,MatType>(fn5, eps) <= bound);
+  CHECK(CheckGradient<GradientFunction,MatType>(fn6, eps) <= bound);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/softmax.cpp 4.4.0-1/src/mlpack/tests/ann/layer/softmax.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/softmax.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/softmax.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -31,14 +31,14 @@ TEST_CASE("SimpleSoftmaxLayerTest", "[AN
   // Test the forward function.
   input = arma::mat("1.7; 3.6");
   module.Forward(input, output);
-  REQUIRE(arma::accu(arma::abs(arma::mat("0.130108; 0.869892") - output)) ==
+  REQUIRE(accu(arma::abs(arma::mat("0.130108; 0.869892") - output)) ==
       Approx(0.0).margin(1e-4));
 
   // Test the backward function.
   gy = arma::zeros(input.n_rows, input.n_cols);
   gy(0) = 1;
   module.Backward(input, output, gy, g);
-  REQUIRE(arma::accu(arma::abs(arma::mat("0.11318; -0.11318") - g)) ==
+  REQUIRE(accu(arma::abs(arma::mat("0.11318; -0.11318") - g)) ==
       Approx(0.0).margin(1e-04));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer/softmin.cpp 4.4.0-1/src/mlpack/tests/ann/layer/softmin.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer/softmin.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer/softmin.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -39,7 +39,7 @@ TEST_CASE("SimpleSoftminLayerTest", "[AN
   			    {0.035311, 0.042794, 0.062526}};
 
   module.Forward(input, output);
-  REQUIRE(arma::accu(arma::abs(actualOutput - output)) ==
+  REQUIRE(accu(arma::abs(actualOutput - output)) ==
              Approx(0.0).margin(1e-4));
 
   // Test the backward function.
@@ -51,7 +51,7 @@ TEST_CASE("SimpleSoftminLayerTest", "[AN
                                   {-0.0205, 0, 0},
                                   {-0.0083, 0, 0}};
 
-  REQUIRE(arma::accu(arma::abs(calculatedGradient - g)) ==
+  REQUIRE(accu(arma::abs(calculatedGradient - g)) ==
       Approx(0.0).margin(1e-04));
 
 }
diff -pruN 4.3.0-2/src/mlpack/tests/ann/layer_test.cpp 4.4.0-1/src/mlpack/tests/ann/layer_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/layer_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/layer_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,6 +18,7 @@
 
 #include "layer/adaptive_max_pooling.cpp"
 #include "layer/adaptive_mean_pooling.cpp"
+#include "layer/add.cpp"
 #include "layer/add_merge.cpp"
 #include "layer/alpha_dropout.cpp"
 #include "layer/batch_norm.cpp"
@@ -38,6 +39,7 @@
 #include "layer/padding.cpp"
 #include "layer/parametric_relu.cpp"
 #include "layer/relu6.cpp"
+#include "layer/repeat.cpp"
 #include "layer/softmax.cpp"
 #include "layer/softmin.cpp"
 #include "layer/ftswish.cpp"
diff -pruN 4.3.0-2/src/mlpack/tests/ann/loss_functions_test.cpp 4.4.0-1/src/mlpack/tests/ann/loss_functions_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/loss_functions_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/loss_functions_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -50,8 +50,7 @@ TEST_CASE("HuberLossTest", "[LossFunctio
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-                Approx(-0.8032).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.8032).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -69,8 +68,7 @@ TEST_CASE("HuberLossTest", "[LossFunctio
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-            Approx(-0.0669333).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.0669333).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -171,7 +169,7 @@ TEST_CASE("SimpleKLDivergenceTest", "[Lo
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(-5.8127).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-5.8127).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -189,7 +187,7 @@ TEST_CASE("SimpleKLDivergenceTest", "[Lo
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(-0.484392).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.484392).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -221,8 +219,7 @@ TEST_CASE("SimpleMeanSquaredLogarithmicE
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(-10.5619).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-10.5619).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -235,7 +232,7 @@ TEST_CASE("SimpleMeanSquaredLogarithmicE
 
   // Test the Backward function on a single input.
   module.Backward(input, target, output);
-  REQUIRE(arma::accu(output) == Approx(-0.1917880483011872).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.1917880483011872).epsilon(1e-3));
   REQUIRE(output.n_elem == 1);
 }
 
@@ -272,7 +269,7 @@ TEST_CASE("SimpleMeanSquaredErrorTest",
   // Test the Backward function on a single input.
   module.Backward(input, target, output);
   // Test whether the output is negative.
-  REQUIRE(arma::accu(output) == -2);
+  REQUIRE(accu(output) == -2);
   REQUIRE(output.n_elem == 1);
 
   // Test for sum reduction
@@ -285,7 +282,7 @@ TEST_CASE("SimpleMeanSquaredErrorTest",
   // Test the Backward function on a single input.
   module.Backward(input, target, output);
   // Test whether the output is negative.
-  REQUIRE(arma::accu(output) == -2);
+  REQUIRE(accu(output) == -2);
   REQUIRE(output.n_elem == 1);
 }
 
@@ -459,8 +456,7 @@ TEST_CASE("SimpleEarthMoverDistanceLayer
 
   // Test the Backward function.
   module.Backward(input3, target3, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == 
-      Approx(-0.168867).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.168867).epsilon(1e-3));
   REQUIRE(output.n_rows == input3.n_rows);
   REQUIRE(output.n_cols == input3.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -634,7 +630,7 @@ TEST_CASE("SimpleMeanBiasErrorTest", "[L
 
   for(double el : output)
     REQUIRE(el == Approx(-0.0833).epsilon(1e-3));
-  REQUIRE(arma::accu(output) == Approx(-1).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(-1).epsilon(1e-5));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 }
@@ -674,7 +670,7 @@ TEST_CASE("LogCoshLossTest", "[LossFunct
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::accu(output) == Approx(2.46962).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(2.46962).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
@@ -687,7 +683,7 @@ TEST_CASE("LogCoshLossTest", "[LossFunct
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::accu(output) == Approx(0.49392).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(0.49392).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 }
@@ -718,7 +714,7 @@ TEST_CASE("HingeEmbeddingLossTest", "[Lo
 
   // Test the Backward function
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(6).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(6).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 1e-3);
@@ -738,7 +734,7 @@ TEST_CASE("HingeEmbeddingLossTest", "[Lo
 
   // Test the backward function
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(0.5).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(0.5).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -791,7 +787,7 @@ TEST_CASE("CosineEmbeddingLossTest", "[L
 
   // Test the Backward function.
   module.Backward(input1, input1, output);
-  REQUIRE(arma::accu(output) == Approx(0.0).margin(1e-6));
+  REQUIRE(accu(output) == Approx(0.0).margin(1e-6));
 
   // Check for dissimilarity.
   module.Similarity() = false;
@@ -800,7 +796,7 @@ TEST_CASE("CosineEmbeddingLossTest", "[L
 
   // Test the Backward function.
   module.Backward(input1, input1, output);
-  REQUIRE(arma::accu(output) == Approx(0.0).margin(1e-6));
+  REQUIRE(accu(output) == Approx(0.0).margin(1e-6));
 
   input1 = arma::mat(3, 2);
   input2 = arma::mat(3, 2);
@@ -816,7 +812,7 @@ TEST_CASE("CosineEmbeddingLossTest", "[L
 
   // Test the Backward function.
   module.Backward(input1, input2, output);
-  REQUIRE(arma::accu(output) == Approx(0.06324556).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(0.06324556).epsilon(1e-3));
 }
 
 /*
@@ -834,7 +830,7 @@ TEST_CASE("MarginRankingLossTest", "[Los
       "-9.4886 2.2755 8.4951");
   expectedOutput = { { 0.0, 0.0,  1.0, 0.0,  1.0, -1.0, 0.0, 0.0,  1.0, -1.0 }, 
                      { 0.0, 0.0, -1.0, 0.0, -1.0,  1.0, 0.0, 0.0, -1.0,  1.0 } };
-  input = arma::join_cols(input1, input2);
+  input = join_cols(input1, input2);
   target = arma::mat("1 1 -1 1 -1 1 1 1 -1 1");
 
   // Test the forward function
@@ -897,8 +893,7 @@ TEST_CASE("SoftMarginLossTest", "[LossFu
 
   // Test the Backward function.
   module1.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(-1.48227).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-1.48227).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -917,8 +912,7 @@ TEST_CASE("SoftMarginLossTest", "[LossFu
 
   // Test the Backward function.
   module2.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(-0.164697).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.164697).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -942,8 +936,7 @@ TEST_CASE("MeanAbsolutePercentageErrorTe
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(-105.625).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-105.625).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1007,7 +1000,7 @@ TEST_CASE("TripletMarginLossTest")
   // Test the Backward function on a single input.
   module.Backward(input, negative, output);
   // Test whether the output is negative.
-  REQUIRE(arma::accu(output) == -12);
+  REQUIRE(accu(output) == -12);
   REQUIRE(output.n_elem == 1);
 }
 
@@ -1071,7 +1064,7 @@ TEST_CASE("HingeLossTest", "[LossFunctio
 
   // Test the Backward function
   module1.Backward(input, target, output);
-  REQUIRE(arma::accu(output) == Approx(-5).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-5).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
@@ -1083,7 +1076,7 @@ TEST_CASE("HingeLossTest", "[LossFunctio
 
   // Test the Backward function.
   module2.Backward(input, target, output);
-  REQUIRE(arma::accu(output) == Approx(-0.41667).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-0.41667).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 }
@@ -1118,8 +1111,7 @@ TEST_CASE("MultiLabelSoftMarginLossTest"
 
   // Test the Backward function.
   module1.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(0.505909).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(0.505909).epsilon(1e-5));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1138,8 +1130,7 @@ TEST_CASE("MultiLabelSoftMarginLossTest"
 
   // Test the Backward function.
   module2.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(0.168636).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(0.168636).epsilon(1e-5));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1177,8 +1168,7 @@ TEST_CASE("MultiLabelSoftMarginLossWeigh
 
   // Test the Backward function.
   module1.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(1.43577).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(1.43577).epsilon(1e-5));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1197,8 +1187,7 @@ TEST_CASE("MultiLabelSoftMarginLossWeigh
 
   // Test the Backward function.
   module2.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) ==
-      Approx(0.358943).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(0.358943).epsilon(1e-5));
   REQUIRE(output.n_rows ==input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1229,7 +1218,7 @@ TEST_CASE("NegativeLogLikelihoodLossTest
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(-4).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-4).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
@@ -1246,7 +1235,7 @@ TEST_CASE("NegativeLogLikelihoodLossTest
 
   // Test the Backward function.
   module.Backward(input, target, output);
-  REQUIRE(arma::as_scalar(arma::accu(output)) == Approx(-1).epsilon(1e-3));
+  REQUIRE(accu(output) == Approx(-1).epsilon(1e-3));
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
   CheckMatrices(output, expectedOutput, 0.1);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/ann_layer_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/ann_layer_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/ann_layer_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/ann_layer_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -78,21 +78,21 @@ TEST_CASE("SimpleAddLayerTest", "[ANNLay
   // Test the Forward function.
   input = arma::zeros(10, 1);
   module.Forward(input, output);
-  REQUIRE(arma::accu(module.Parameters()) == arma::accu(output));
+  REQUIRE(accu(module.Parameters()) == accu(output));
 
   // Test the Backward function.
   module.Backward(input, output, delta);
-  REQUIRE(arma::accu(output) == arma::accu(delta));
+  REQUIRE(accu(output) == accu(delta));
 
   // Test the forward function.
   input = arma::ones(10, 1);
   module.Forward(input, output);
-  REQUIRE(10 + arma::accu(module.Parameters()) ==
-      Approx(arma::accu(output)).epsilon(1e-5));
+  REQUIRE(10 + accu(module.Parameters()) ==
+      Approx(accu(output)).epsilon(1e-5));
 
   // Test the backward function.
   module.Backward(input, output, delta);
-  REQUIRE(arma::accu(output) == Approx(arma::accu(delta)).epsilon(1e-5));
+  REQUIRE(accu(output) == Approx(accu(delta)).epsilon(1e-5));
 }
 */
 
@@ -181,20 +181,20 @@ TEST_CASE("SimpleConstantLayerTest", "[A
   // Test the Forward function.
   input = arma::zeros(10, 1);
   module.Forward(input, output);
-  REQUIRE(arma::accu(output) == 30.0);
+  REQUIRE(accu(output) == 30.0);
 
   // Test the Backward function.
   module.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 
   // Test the forward function.
   input = arma::ones(10, 1);
   module.Forward(input, output);
-  REQUIRE(arma::accu(output) == 30.0);
+  REQUIRE(accu(output) == 30.0);
 
   // Test the backward function.
   module.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 }*/
 
 /**
@@ -241,13 +241,13 @@ TEST_CASE("ConstantLayerParametersTest",
 //   // Test the Forward function.
 //   input = arma::zeros(10, 1);
 //   module.Forward(input, output);
-//   REQUIRE(arma::accu(module.Parameters().submat(100,
+//   REQUIRE(accu(module.Parameters().submat(100,
 //           0, module.Parameters().n_elem - 1, 0)) ==
-//           Approx(arma::accu(output)).epsilon(1e-5));
+//           Approx(accu(output)).epsilon(1e-5));
 
 //   // Test the Backward function.
 //   module.Backward(input, input, delta);
-//   REQUIRE(arma::accu(delta) == 0);
+//   REQUIRE(accu(delta) == 0);
 // }
 
 // /**
@@ -324,7 +324,7 @@ TEST_CASE("ConstantLayerParametersTest",
 
 //   // Test the Backward function.
 //   module.Backward(input, input, delta);
-//   REQUIRE(arma::accu(delta) == 0);
+//   REQUIRE(accu(delta) == 0);
 // }
 
 // /**
@@ -496,20 +496,20 @@ TEST_CASE("SimpleSelectLayerTest", "[ANN
   // Test the Forward function.
   Select moduleA(3);
   moduleA.Forward(input, outputA);
-  REQUIRE(30 == arma::accu(outputA));
+  REQUIRE(30 == accu(outputA));
 
   // Test the Forward function.
   Select moduleB(3, 5);
   moduleB.Forward(input, outputB);
-  REQUIRE(15 == arma::accu(outputB));
+  REQUIRE(15 == accu(outputB));
 
   // Test the Backward function.
   moduleA.Backward(input, outputA, delta);
-  REQUIRE(30 == arma::accu(delta));
+  REQUIRE(30 == accu(delta));
 
   // Test the Backward function.
   moduleB.Backward(input, outputA, delta);
-  REQUIRE(15 == arma::accu(delta));
+  REQUIRE(15 == accu(delta));
 }
 */
 
@@ -539,14 +539,14 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
   // Test the Forward function.
   Join module;
   module.Forward(input, output);
-  REQUIRE(50 == arma::accu(output));
+  REQUIRE(50 == accu(output));
 
   bool b = output.n_rows == 1 || output.n_cols == 1;
   REQUIRE(b == true);
 
   // Test the Backward function.
   module.Backward(input, output, delta);
-  REQUIRE(50 == arma::accu(delta));
+  REQUIRE(50 == accu(delta));
 
   b = delta.n_rows == input.n_rows && input.n_cols;
   REQUIRE(b == true);
@@ -575,11 +575,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 
 //     // Test the Forward function.
 //     module.Forward(input, output);
-//     REQUIRE(10 * numMergeModules == arma::accu(output));
+//     REQUIRE(10 * numMergeModules == accu(output));
 
 //     // Test the Backward function.
 //     module.Backward(input, output, delta);
-//     REQUIRE(arma::accu(output) == arma::accu(delta));
+//     REQUIRE(accu(output) == accu(delta));
 //   }
 // }
 
@@ -901,11 +901,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 
 //       // Compute the value of cell state and output.
 //       // i = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       inputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       inputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // f = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       forgetGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       forgetGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // z = tanh(W.dot(x) + W.dot(h) + b).
@@ -916,7 +916,7 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //       cellCalc = forgetGate % cellCalc + inputGate % hidden;
 
 //       // o = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       outputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       outputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // h = o * tanh(c).
@@ -983,11 +983,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 
 //       // Compute the value of cell state and output.
 //       // i = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       inputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       inputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // f = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       forgetGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       forgetGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // z = tanh(W.dot(x) + W.dot(h) + b).
@@ -998,7 +998,7 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //       cellCalc = forgetGate % cellCalc + inputGate % hidden;
 
 //       // o = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       outputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       outputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 
 //       // h = o * tanh(c).
@@ -1123,36 +1123,36 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //   // Compute the z_t gate output.
 //   arma::mat expectedOutput = arma::ones(3, 1);
 //   expectedOutput *= -4;
-//   expectedOutput = arma::exp(expectedOutput);
+//   expectedOutput = exp(expectedOutput);
 //   expectedOutput = arma::ones(3, 1) / (arma::ones(3, 1) + expectedOutput);
 //   expectedOutput = (arma::ones(3, 1)  - expectedOutput) % expectedOutput;
 
 //   // For the first input the output should be equal to the output of
 //   // gate z_t as the previous output fed to the cell is all zeros.
-//   REQUIRE(arma::as_scalar(arma::trans(output) * expectedOutput) <= 1e-2);
+//   REQUIRE(arma::as_scalar(trans(output) * expectedOutput) <= 1e-2);
 
 //   expectedOutput = output;
 
 //   gru.Forward(input, output);
 
-//   double s = arma::as_scalar(arma::sum(expectedOutput));
+//   double s = arma::as_scalar(sum(expectedOutput));
 
 //   // Compute the value of z_t gate for the second input.
 //   arma::mat z_t = arma::ones(3, 1);
 //   z_t *= -(s + 4);
-//   z_t = arma::exp(z_t);
+//   z_t = exp(z_t);
 //   z_t = arma::ones(3, 1) / (arma::ones(3, 1) + z_t);
 
 //   // Compute the value of o_t gate for the second input.
 //   arma::mat o_t = arma::ones(3, 1);
-//   o_t *= -(arma::as_scalar(arma::sum(expectedOutput % z_t)) + 4);
-//   o_t = arma::exp(o_t);
+//   o_t *= -(arma::as_scalar(sum(expectedOutput % z_t)) + 4);
+//   o_t = exp(o_t);
 //   o_t = arma::ones(3, 1) / (arma::ones(3, 1) + o_t);
 
 //   // Expected output for the second input.
 //   expectedOutput = z_t % expectedOutput + (arma::ones(3, 1) - z_t) % o_t;
 
-//   REQUIRE(arma::as_scalar(arma::trans(output) * expectedOutput) <= 1e-2);
+//   REQUIRE(arma::as_scalar(trans(output) * expectedOutput) <= 1e-2);
 
 //   LayerTypes<> layer(gruAlloc);
 //   boost::apply_visitor(DeleteVisitor(), layer);
@@ -1180,11 +1180,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //
 //     // Test the Forward function.
 //     module.Forward(input, output);
-//     REQUIRE(10 * numMergeModules == arma::accu(output));
+//     REQUIRE(10 * numMergeModules == accu(output));
 //
 //     // Test the Backward function.
 //     module.Backward(input, output, delta);
-//     REQUIRE(arma::accu(output) == arma::accu(delta));
+//     REQUIRE(accu(output) == accu(delta));
 //   }
 // }
 
@@ -1506,11 +1506,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //
 //       // Compute the value of cell state and output.
 //       // i = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       inputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       inputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // f = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       forgetGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       forgetGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // z = tanh(W.dot(x) + W.dot(h) + b).
@@ -1521,7 +1521,7 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //       cellCalc = forgetGate % cellCalc + inputGate % hidden;
 //
 //       // o = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       outputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       outputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // h = o * tanh(c).
@@ -1588,11 +1588,11 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //
 //       // Compute the value of cell state and output.
 //       // i = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       inputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       inputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // f = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       forgetGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       forgetGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // z = tanh(W.dot(x) + W.dot(h) + b).
@@ -1603,7 +1603,7 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //       cellCalc = forgetGate % cellCalc + inputGate % hidden;
 //
 //       // o = sigmoid(W.dot(x) + W.dot(h) + W.dot(c) + b).
-//       outputGate = 1.0 /(1 + arma::exp(-(inputWeight * stepData +
+//       outputGate = 1.0 /(1 + exp(-(inputWeight * stepData +
 //           outputWeight * outCalc + outputWeight % cellCalc + bias)));
 //
 //       // h = o * tanh(c).
@@ -1728,36 +1728,36 @@ TEST_CASE("SimpleJoinLayerTest", "[ANNLa
 //   // Compute the z_t gate output.
 //   arma::mat expectedOutput = arma::ones(3, 1);
 //   expectedOutput *= -4;
-//   expectedOutput = arma::exp(expectedOutput);
+//   expectedOutput = exp(expectedOutput);
 //   expectedOutput = arma::ones(3, 1) / (arma::ones(3, 1) + expectedOutput);
 //   expectedOutput = (arma::ones(3, 1)  - expectedOutput) % expectedOutput;
 //
 //   // For the first input the output should be equal to the output of
 //   // gate z_t as the previous output fed to the cell is all zeros.
-//   REQUIRE(arma::as_scalar(arma::trans(output) * expectedOutput) <= 1e-2);
+//   REQUIRE(arma::as_scalar(trans(output) * expectedOutput) <= 1e-2);
 //
 //   expectedOutput = output;
 //
 //   gru.Forward(input, output);
 //
-//   double s = arma::as_scalar(arma::sum(expectedOutput));
+//   double s = arma::as_scalar(sum(expectedOutput));
 //
 //   // Compute the value of z_t gate for the second input.
 //   arma::mat z_t = arma::ones(3, 1);
 //   z_t *= -(s + 4);
-//   z_t = arma::exp(z_t);
+//   z_t = exp(z_t);
 //   z_t = arma::ones(3, 1) / (arma::ones(3, 1) + z_t);
 //
 //   // Compute the value of o_t gate for the second input.
 //   arma::mat o_t = arma::ones(3, 1);
-//   o_t *= -(arma::as_scalar(arma::sum(expectedOutput % z_t)) + 4);
-//   o_t = arma::exp(o_t);
+//   o_t *= -(arma::as_scalar(sum(expectedOutput % z_t)) + 4);
+//   o_t = exp(o_t);
 //   o_t = arma::ones(3, 1) / (arma::ones(3, 1) + o_t);
 //
 //   // Expected output for the second input.
 //   expectedOutput = z_t % expectedOutput + (arma::ones(3, 1) - z_t) % o_t;
 //
-//   REQUIRE(arma::as_scalar(arma::trans(output) * expectedOutput) <= 1e-2);
+//   REQUIRE(arma::as_scalar(trans(output) * expectedOutput) <= 1e-2);
 //
 //   LayerTypes<> layer(gruAlloc);
 //   boost::apply_visitor(DeleteVisitor(), layer);
@@ -1790,17 +1790,17 @@ TEST_CASE("SimpleLookupLayerTest", "[ANN
   for (size_t i = 0; i < batchSize; ++i)
   {
     // The Lookup module uses index - 1 for the cols.
-    const double outputSum = arma::accu(module.Parameters().cols(
-        arma::conv_to<arma::uvec>::from(input.col(i)) - 1));
+    const double outputSum = accu(module.Parameters().cols(
+        ConvTo<arma::uvec>::From(input.col(i)) - 1));
 
-    REQUIRE(std::fabs(outputSum - arma::accu(output.col(i))) <= 1e-5);
+    REQUIRE(std::fabs(outputSum - accu(output.col(i))) <= 1e-5);
   }
 
   // Test the Gradient function.
   arma::mat error = 0.01 * arma::randu(embeddingSize * seqLength, batchSize);
   module.Gradient(input, error, gradient);
 
-  REQUIRE(std::fabs(arma::accu(error) - arma::accu(gradient)) <= 1e-07);
+  REQUIRE(std::fabs(accu(error) - accu(gradient)) <= 1e-07);
 }
 */
 
@@ -1933,8 +1933,8 @@ TEST_CASE("SimpleNearestInterpolationLay
   layer1.Forward(input1, output1);
   layer1.Backward(output1, output1, unzoomedOutput1);
 
-  REQUIRE(arma::accu(output1) - 1317.00 == Approx(0.0).margin(1e-05));
-  REQUIRE(arma::accu(unzoomedOutput1) - 1317.00 ==
+  REQUIRE(accu(output1) - 1317.00 == Approx(0.0).margin(1e-05));
+  REQUIRE(accu(unzoomedOutput1) - 1317.00 ==
           Approx(0.0).margin(1e-05));
 }
 */
@@ -2225,12 +2225,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module1.Reset();
   module1.Forward(input, output);
   // Value calculated using tensorflow.nn.conv2d_transpose()
-  REQUIRE(arma::accu(output) == 360.0);
+  REQUIRE(accu(output) == 360.0);
 
   // Test the backward function.
   module1.Backward(input, output, delta);
   // Value calculated using tensorflow.nn.conv2d()
-  REQUIRE(arma::accu(delta) == 720.0);
+  REQUIRE(accu(delta) == 720.0);
 
   TransposedConvolution module2(1, 1, 4, 4, 1, 1, 1, 1, 5, 5, 6, 6);
   // Test the forward function.
@@ -2245,12 +2245,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module2.Reset();
   module2.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 1512.0);
+  REQUIRE(accu(output) == 1512.0);
 
   // Test the backward function.
   module2.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 6504.0);
+  REQUIRE(accu(delta) == 6504.0);
 
   TransposedConvolution module3(1, 1, 3, 3, 1, 1, 1, 1, 5, 5, 5, 5);
   // Test the forward function.
@@ -2263,12 +2263,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module3.Reset();
   module3.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 2370.0);
+  REQUIRE(accu(output) == 2370.0);
 
   // Test the backward function.
   module3.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 19154.0);
+  REQUIRE(accu(delta) == 19154.0);
 
   TransposedConvolution module4(1, 1, 3, 3, 1, 1, 0, 0, 5, 5, 7, 7);
   // Test the forward function.
@@ -2281,12 +2281,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module4.Reset();
   module4.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 6000.0);
+  REQUIRE(accu(output) == 6000.0);
 
   // Test the backward function.
   module4.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 86208.0);
+  REQUIRE(accu(delta) == 86208.0);
 
   TransposedConvolution module5(1, 1, 3, 3, 2, 2, 0, 0, 2, 2, 5, 5);
   // Test the forward function.
@@ -2299,12 +2299,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module5.Reset();
   module5.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 120.0);
+  REQUIRE(accu(output) == 120.0);
 
   // Test the backward function.
   module5.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 960.0);
+  REQUIRE(accu(delta) == 960.0);
 
   TransposedConvolution module6(1, 1, 3, 3, 2, 2, 1, 1, 3, 3, 5, 5);
   // Test the forward function.
@@ -2317,12 +2317,12 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module6.Reset();
   module6.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 410.0);
+  REQUIRE(accu(output) == 410.0);
 
   // Test the backward function.
   module6.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 4444.0);
+  REQUIRE(accu(delta) == 4444.0);
 
   TransposedConvolution module7(1, 1, 3, 3, 2, 2, 1, 1, 3, 3, 6, 6);
   // Test the forward function.
@@ -2335,11 +2335,11 @@ TEST_CASE("SimpleTransposedConvolutionLa
   module7.Reset();
   module7.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d()
-  REQUIRE(arma::accu(output) == 606.0);
+  REQUIRE(accu(output) == 606.0);
 
   module7.Backward(input, output, delta);
   // Value calculated using torch.nn.functional.conv2d()
-  REQUIRE(arma::accu(delta) == 7732.0);
+  REQUIRE(accu(delta) == 7732.0);
 }
 */
 
@@ -2415,11 +2415,11 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 
     // Test the Forward function.
     module.Forward(input, output);
-    REQUIRE(10 == arma::accu(output));
+    REQUIRE(10 == accu(output));
 
     // Test the Backward function.
     module.Backward(input, output, delta);
-    REQUIRE(arma::accu(output) == arma::accu(delta));
+    REQUIRE(accu(output) == accu(delta));
   }
 }
 */
@@ -2490,11 +2490,11 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   module1.Reset();
 //   module1.Forward(input, output);
 //   // Value calculated using tensorflow.nn.atrous_conv2d()
-//   REQUIRE(arma::accu(output) == 792.0);
+//   REQUIRE(accu(output) == 792.0);
 
 //   // Test the Backward function.
 //   module1.Backward(input, output, delta);
-//   REQUIRE(arma::accu(delta) == 2376);
+//   REQUIRE(accu(delta) == 2376);
 
 //   AtrousConvolution<> module2(1, 1, 3, 3, 2, 2, 0, 0, 7, 7, 2, 2);
 //   // Test the forward function.
@@ -2506,11 +2506,11 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   module2.Reset();
 //   module2.Forward(input, output);
 //   // Value calculated using tensorflow.nn.conv2d()
-//   REQUIRE(arma::accu(output) == 264.0);
+//   REQUIRE(accu(output) == 264.0);
 
 //   // Test the backward function.
 //   module2.Backward(input, output, delta);
-//   REQUIRE(arma::accu(delta) == 792.0);
+//   REQUIRE(accu(delta) == 792.0);
 // }
 
 // /**
@@ -2633,7 +2633,7 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   module1.Reset();
 //   module1.Forward(input, output);
 
-//   REQUIRE(arma::accu(output) == 0);
+//   REQUIRE(accu(output) == 0);
 //   REQUIRE(output.n_rows == 9);
 //   REQUIRE(output.n_cols == 1);
 
@@ -2651,7 +2651,7 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   module2.Reset();
 //   module2.Forward(input, output);
 
-//   REQUIRE(arma::accu(output) == 0);
+//   REQUIRE(accu(output) == 0);
 //   REQUIRE(output.n_rows == 49);
 //   REQUIRE(output.n_cols == 1);
 
@@ -2756,7 +2756,7 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   input = arma::zeros(10, 1);
 //   module.Forward(input, output);
 
-//   double parameterSum = arma::accu(linear->Parameters().submat(
+//   double parameterSum = accu(linear->Parameters().submat(
 //       100, 0, linear->Parameters().n_elem - 1, 0));
 
 //   // Test the Backward function.
@@ -2765,8 +2765,8 @@ TEST_CASE("SimpleMultiplyMergeLayerTest"
 //   // Clean up before we break,
 //   delete linear;
 
-//   REQUIRE(parameterSum == Approx(arma::accu(output)).epsilon(1e-5));
-//   REQUIRE(arma::accu(delta) == 0);
+//   REQUIRE(parameterSum == Approx(accu(output)).epsilon(1e-5));
+//   REQUIRE(accu(delta) == 0);
 // }
 
 /**
@@ -2788,7 +2788,7 @@ TEST_CASE("MultiplyMergeRunTest", "[ANNL
   input = arma::zeros(10, 1);
   module.Forward(input, output);
 
-  double parameterSum = arma::accu(linear->Parameters().submat(
+  double parameterSum = accu(linear->Parameters().submat(
       100, 0, linear->Parameters().n_elem - 1, 0));
 
   // Test the Backward function.
@@ -2797,8 +2797,8 @@ TEST_CASE("MultiplyMergeRunTest", "[ANNL
   // Clean up before we break,
   delete linear;
 
-  REQUIRE(parameterSum == Approx(arma::accu(output)).epsilon(1e-5));
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(parameterSum == Approx(accu(output)).epsilon(1e-5));
+  REQUIRE(accu(delta) == 0);
 }
 */
 
@@ -2940,12 +2940,12 @@ TEST_CASE("SimpleReparametrizationLayerT
   input = join_cols(arma::ones<arma::mat>(5, 1) * -15,
       arma::zeros<arma::mat>(5, 1));
   module.Forward(input, output);
-  REQUIRE(arma::accu(output) <= 1e-5);
+  REQUIRE(accu(output) <= 1e-5);
 
   // Test the Backward function.
   arma::mat gy = arma::zeros<arma::mat>(5, 1);
   module.Backward(input, gy, delta);
-  REQUIRE(arma::accu(delta) != 0); // klBackward will be added.
+  REQUIRE(accu(delta) != 0); // klBackward will be added.
 }
 */
 
@@ -2985,7 +2985,7 @@ TEST_CASE("ReparametrizationLayerInclude
   gy = arma::zeros(output.n_rows, output.n_cols);
   module.Backward(output, gy, delta);
 
-  REQUIRE(arma::accu(delta) == 0);
+  REQUIRE(accu(delta) == 0);
 }
 */
 
@@ -3377,8 +3377,8 @@ TEST_CASE("HighwayLayerParametersTest",
 //   // Test the Backward function.
 //   module.Backward(input, input, delta);
 
-//   REQUIRE(0 == arma::accu(output));
-//   REQUIRE(arma::accu(delta) == 0);
+//   REQUIRE(0 == accu(output));
+//   REQUIRE(accu(delta) == 0);
 // }
 
 // /**
@@ -3405,11 +3405,11 @@ TEST_CASE("TransposedConvolutionLayerPad
   module1.Reset();
   module1.Forward(input, output);
   // Value calculated using tensorflow.nn.conv2d_transpose().
-  REQUIRE(arma::accu(output) == 0.0);
+  REQUIRE(accu(output) == 0.0);
 
   // Test the Backward Function.
   module1.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0.0);
+  REQUIRE(accu(delta) == 0.0);
 
   // Test Valid for non zero padding.
   TransposedConvolution module2(1, 1, 3, 3, 2, 2,
@@ -3425,11 +3425,11 @@ TEST_CASE("TransposedConvolutionLayerPad
   module2.Reset();
   module2.Forward(input, output);
   // Value calculated using torch.nn.functional.conv_transpose2d().
-  REQUIRE(arma::accu(output) == 120.0);
+  REQUIRE(accu(output) == 120.0);
 
   // Test the Backward Function.
   module2.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 960.0);
+  REQUIRE(accu(delta) == 960.0);
 
   // Test for same padding type.
   TransposedConvolution module3(1, 1, 3, 3, 2, 2, 0, 0, 3, 3, 3, 3, "SAME");
@@ -3438,13 +3438,13 @@ TEST_CASE("TransposedConvolutionLayerPad
   module3.Parameters() = arma::mat(9 + 1, 1, arma::fill::zeros);
   module3.Reset();
   module3.Forward(input, output);
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
   // Test the Backward Function.
   module3.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0.0);
+  REQUIRE(accu(delta) == 0.0);
 
   // Output shape should equal input.
   TransposedConvolution module4(1, 1, 3, 3, 1, 1,
@@ -3455,13 +3455,13 @@ TEST_CASE("TransposedConvolutionLayerPad
   module4.Parameters() = arma::mat(9 + 1, 1, arma::fill::zeros);
   module4.Reset();
   module4.Forward(input, output);
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
   // Test the Backward Function.
   module4.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0.0);
+  REQUIRE(accu(delta) == 0.0);
 
   TransposedConvolution module5(1, 1, 3, 3, 2, 2, 0, 0, 2, 2, 2, 2, "SAME");
   // Test the forward function.
@@ -3469,13 +3469,13 @@ TEST_CASE("TransposedConvolutionLayerPad
   module5.Parameters() = arma::mat(25 + 1, 1, arma::fill::zeros);
   module5.Reset();
   module5.Forward(input, output);
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
   // Test the Backward Function.
   module5.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0.0);
+  REQUIRE(accu(delta) == 0.0);
 
   TransposedConvolution module6(1, 1, 4, 4, 1, 1, 1, 1, 5, 5, 5, 5, "SAME");
   // Test the forward function.
@@ -3483,13 +3483,13 @@ TEST_CASE("TransposedConvolutionLayerPad
   module6.Parameters() = arma::mat(16 + 1, 1, arma::fill::zeros);
   module6.Reset();
   module6.Forward(input, output);
-  REQUIRE(arma::accu(output) == 0);
+  REQUIRE(accu(output) == 0);
   REQUIRE(output.n_rows == input.n_rows);
   REQUIRE(output.n_cols == input.n_cols);
 
   // Test the Backward Function.
   module6.Backward(input, output, delta);
-  REQUIRE(arma::accu(delta) == 0.0);
+  REQUIRE(accu(delta) == 0.0);
 }
 */
 
@@ -3513,7 +3513,7 @@ TEST_CASE("TransposedConvolutionLayerPad
 //   module1.InputWidth() = 4;
 //   module1.Forward(input, output);
 //   // Calculated using torch.nn.LPPool2d().
-//   REQUIRE(arma::accu(output) - 706.0 == Approx(0.0).margin(2e-5));
+//   REQUIRE(accu(output) - 706.0 == Approx(0.0).margin(2e-5));
 //   REQUIRE(output.n_elem == 2);
 //
 //   // For Square input.
@@ -3534,7 +3534,7 @@ TEST_CASE("TransposedConvolutionLayerPad
 //   module3.InputWidth() = 4;
 //   module3.Forward(input, output);
 //   // Calculated using torch.nn.LPPool2d().
-//   REQUIRE(arma::accu(output) - 77.0 == Approx(0.0).margin(2e-5));
+//   REQUIRE(accu(output) - 77.0 == Approx(0.0).margin(2e-5));
 //   REQUIRE(output.n_elem == 4);
 // }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/dcgan_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/dcgan_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/dcgan_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/dcgan_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -53,7 +53,7 @@ TEST_CASE("DCGANMNISTTest", "[DCGANNetwo
       << " shuffle = " << shuffle << std::endl;
 
   arma::mat trainData;
-  trainData.load("mnist_first250_training_4s_and_9s.arm");
+  trainData.load("mnist_first250_training_4s_and_9s.csv");
   Log::Info << arma::size(trainData) << std::endl;
 
   trainData = trainData.cols(0, datasetMaxCols - 1);
@@ -212,7 +212,7 @@ TEST_CASE("DCGANMNISTTest", "[DCGANNetwo
       << " shuffle = " << shuffle << std::endl;
 
   arma::mat trainData;
-  trainData.load("mnist_first250_training_4s_and_9s.arm");
+  trainData.load("mnist_first250_training_4s_and_9s.csv");
   Log::Info << arma::size(trainData) << std::endl;
 
   // trainData = trainData.cols(0, datasetMaxCols - 1);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/feedforward_network_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/feedforward_network_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/feedforward_network_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/feedforward_network_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -134,7 +134,7 @@ TEST_CASE("CheckCopyMovingReparametrizat
 TEST_CASE("HighwayNetworkTest", "[FeedForwardNetworkTest]")
 {
   arma::mat dataset;
-  dataset.load("mnist_first250_training_4s_and_9s.arm");
+  dataset.load("mnist_first250_training_4s_and_9s.csv");
 
   // Normalize each point since these are images.
   for (size_t i = 0; i < dataset.n_cols; ++i)
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/gan_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/gan_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/gan_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/gan_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -46,7 +46,7 @@ TEST_CASE("GANTest", "[GANNetworkTest]")
   double multiplier = 1;
 
   arma::mat trainData(1, 10000);
-  trainData.imbue( [&]() { return arma::as_scalar(RandNormal(4, 0.5));});
+  trainData.imbue( [&]() { return arma::as_scalar(RandNormal(4, std::sqrt(0.5)));});
   trainData = arma::sort(trainData);
 
   // Create the Discriminator network.
@@ -156,7 +156,7 @@ TEST_CASE("GANMNISTTest", "[GANNetworkTe
       << " shuffle = " << shuffle << std::endl;
 
   arma::mat trainData;
-  trainData.load("mnist_first250_training_4s_and_9s.arm");
+  trainData.load("mnist_first250_training_4s_and_9s.csv");
   Log::Info << arma::size(trainData) << std::endl;
 
   trainData = trainData.cols(0, datasetMaxCols - 1);
@@ -297,7 +297,7 @@ TEST_CASE("GANMemorySharingTest", "[GANN
   bool shuffle = true;
 
   arma::mat trainData(1, 10000);
-  trainData.imbue( [&]() { return arma::as_scalar(RandNormal(4, 0.5));});
+  trainData.imbue( [&]() { return arma::as_scalar(RandNormal(4, std::sqrt(0.5)));});
   trainData = arma::sort(trainData);
 
   // Create the Discriminator network.
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/rnn_reber_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/rnn_reber_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/rnn_reber_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/rnn_reber_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -274,10 +274,10 @@ arma::Mat<char> GenerateReberGrammarData
     for (size_t j = 0; j < trainReber.length() - 1; ++j)
     {
       ReberTranslation(trainReber[j], translation);
-      trainInput(0, i) = arma::join_cols(trainInput(0, i), translation);
+      trainInput(0, i) = join_cols(trainInput(0, i), translation);
 
       ReberTranslation(trainReber[j + 1], translation);
-      trainLabels(0, i) = arma::join_cols(trainLabels(0, i), translation);
+      trainLabels(0, i) = join_cols(trainLabels(0, i), translation);
     }
   }
 
@@ -293,7 +293,7 @@ arma::Mat<char> GenerateReberGrammarData
     for (size_t j = 0; j < testReber.length() - 1; ++j)
     {
       ReberTranslation(testReber[j], translation);
-      testInput(0, i) = arma::join_cols(testInput(0, i), translation);
+      testInput(0, i) = join_cols(testInput(0, i), translation);
     }
   }
 
diff -pruN 4.3.0-2/src/mlpack/tests/ann/not_adapted/wgan_test.cpp 4.4.0-1/src/mlpack/tests/ann/not_adapted/wgan_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/not_adapted/wgan_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/not_adapted/wgan_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -60,7 +60,7 @@ TEST_CASE("WGANMNISTTest", "[WGANNetwork
       << " shuffle = " << shuffle << std::endl;
 
   arma::mat trainData;
-  trainData.load("mnist_first250_training_4s_and_9s.arm");
+  trainData.load("mnist_first250_training_4s_and_9s.csv");
   Log::Info << arma::size(trainData) << std::endl;
 
   trainData = trainData.cols(0, datasetMaxCols - 1);
@@ -222,7 +222,7 @@ TEST_CASE("WGANGPMNISTTest", "[WGANNetwo
       << " shuffle = " << shuffle << std::endl;
 
   arma::mat trainData;
-  trainData.load("mnist_first250_training_4s_and_9s.arm");
+  trainData.load("mnist_first250_training_4s_and_9s.csv");
   Log::Info << arma::size(trainData) << std::endl;
 
   trainData = trainData.cols(0, datasetMaxCols - 1);
diff -pruN 4.3.0-2/src/mlpack/tests/ann/recurrent_network_test.cpp 4.4.0-1/src/mlpack/tests/ann/recurrent_network_test.cpp
--- 4.3.0-2/src/mlpack/tests/ann/recurrent_network_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/ann/recurrent_network_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -36,23 +36,21 @@ void GenerateNoisySines(arma::cube& data
 {
   arma::colvec x =  arma::linspace<arma::colvec>(0, points - 1, points) /
       points * 20.0;
-  arma::colvec y1 = arma::sin(x + arma::as_scalar(arma::randu(1)) * 3.0);
-  arma::colvec y2 = arma::sin(x / 2.0 + arma::as_scalar(arma::randu(1)) * 3.0);
+  arma::colvec y1 = arma::sin(x + randu() * 3.0);
+  arma::colvec y2 = arma::sin(x / 2.0 + randu() * 3.0);
 
   data = arma::zeros(1 /* single dimension */, sequences * 2, points);
   labels = arma::zeros(2 /* 2 classes */, sequences * 2);
 
   for (size_t seq = 0; seq < sequences; seq++)
   {
-    arma::vec sequence = arma::randu(points) * noise + y1 +
-        arma::as_scalar(arma::randu(1) - 0.5) * noise;
+    arma::vec sequence = randu(points) * noise + y1 + (randu() - 0.5) * noise;
     for (size_t i = 0; i < points; ++i)
       data(0, seq, i) = sequence[i];
 
     labels(0, seq) = 1;
 
-    sequence = arma::randu(points) * noise + y2 +
-        arma::as_scalar(arma::randu(1) - 0.5) * noise;
+    sequence = randu(points) * noise + y2 + (randu() - 0.5) * noise;
     for (size_t i = 0; i < points; ++i)
       data(0, sequences + seq, i) = sequence[i];
 
@@ -262,7 +260,7 @@ void GenerateNoisySinRNN(arma::cube& dat
 
   arma::colvec y = x;
   if (normalize)
-    y = arma::normalise(x);
+    y = normalise(x);
 
   // Now break this into columns of rho size slices.
   size_t numColumns = y.n_elem / rho;
@@ -317,13 +315,13 @@ double RNNSineTest(size_t hiddenUnits, s
 
   // The prediction must really follow the test data. So convert both the test
   // data and the pediction to vectors and compare the two.
-  arma::colvec testVector = arma::vectorise(testData);
-  arma::colvec predVector = arma::vectorise(prediction);
+  arma::colvec testVector = vectorise(testData);
+  arma::colvec predVector = vectorise(prediction);
 
   // Adjust the vectors for comparison, as the prediction is one step ahead.
   testVector = testVector.rows(1, testVector.n_rows - 1);
   predVector = predVector.rows(0, predVector.n_rows - 2);
-  double error = std::sqrt(arma::sum(arma::square(testVector - predVector))) /
+  double error = std::sqrt(sum(square(testVector - predVector))) /
       testVector.n_rows;
 
   return error;
diff -pruN 4.3.0-2/src/mlpack/tests/bayesian_linear_regression_test.cpp 4.4.0-1/src/mlpack/tests/bayesian_linear_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/bayesian_linear_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/bayesian_linear_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -42,7 +42,7 @@ TEST_CASE("BayesianLinearRegressionRegre
   GenerateProblem(matX, y, 200, 10);
 
   // Instanciate and train the estimator.
-  BayesianLinearRegression estimator(true);
+  BayesianLinearRegression<> estimator(true);
   estimator.Train(matX, y);
   estimator.Predict(matX, predictions);
 
@@ -63,7 +63,7 @@ TEST_CASE("TestCenter0ScaleData0", "[Bay
 
   GenerateProblem(matX, y, nPoints, nDims, 0.5);
 
-  BayesianLinearRegression estimator(false, false);
+  BayesianLinearRegression<> estimator(false, false);
 
   estimator.Train(matX, y);
 
@@ -85,7 +85,7 @@ TEST_CASE("TestCenterDataTrueScaleDataTr
   size_t nDims = 5, nPoints = 100;
   GenerateProblem(matX, y, nPoints, nDims, 0.5);
 
-  BayesianLinearRegression estimator(true, true);
+  BayesianLinearRegression<> estimator(true, true);
   estimator.Train(matX, y);
 
   arma::colvec xMean = arma::mean(matX, 1);
@@ -108,7 +108,7 @@ TEST_CASE("OptionsMakeModelDifferent", "
   size_t nDims = 10, nPoints = 100;
   GenerateProblem(matX, y, nPoints, nDims, 0.5);
 
-  BayesianLinearRegression blr(false, false), blrC(true, false),
+  BayesianLinearRegression<> blr(false, false), blrC(true, false),
       blrCS(true, true);
 
   blr.Train(matX, y);
@@ -133,7 +133,7 @@ TEST_CASE("SingularMatix", "[BayesianLin
   // Now the first and the second rows are indentical.
   matX.row(1) = matX.row(0);
 
-  BayesianLinearRegression estimator;
+  BayesianLinearRegression<> estimator;
   estimator.Train(matX, y);
 }
 
@@ -146,7 +146,7 @@ TEST_CASE("PredictiveUncertainties", "[B
 
   GenerateProblem(matX, y, 100, 10, 1);
 
-  BayesianLinearRegression estimator(true, true);
+  BayesianLinearRegression<> estimator(true, true);
   estimator.Train(matX, y);
 
   arma::rowvec responses, std;
@@ -156,6 +156,15 @@ TEST_CASE("PredictiveUncertainties", "[B
   for (size_t i = 0; i < matX.n_cols; i++)
     REQUIRE(std[i] > estStd);
 
+  // Also make single-point predictions.
+  for (size_t i = 0; i < matX.n_cols; ++i)
+  {
+    double prediction, stddev;
+    estimator.Predict(matX.col(i), prediction, stddev);
+    REQUIRE(prediction == Approx(responses[i]));
+    REQUIRE(stddev > estStd);
+  }
+
   // Check that the estimated variance is close to 1.
   REQUIRE(estStd == Approx(1).epsilon(0.3));
 }
@@ -171,10 +180,10 @@ TEST_CASE("EqualtoRidge", "[BayesianLine
   {
     GenerateProblem(matX, y, 100, 10, 1);
 
-    BayesianLinearRegression blr(false, false);
+    BayesianLinearRegression<> blr(false, false);
     blr.Train(matX, y);
 
-    LinearRegression ridge(matX, y, blr.Alpha() / blr.Beta(), false);
+    LinearRegression<> ridge(matX, y, blr.Alpha() / blr.Beta(), false);
 
     blr.Predict(matX, blrPred);
     ridge.Predict(matX, ridgePred);
@@ -187,9 +196,120 @@ TEST_CASE("EqualtoRidge", "[BayesianLine
     for (size_t i = 0; i < y.size(); ++i)
       REQUIRE(blrPred[i] == Approx(ridgePred[i]).epsilon(1));
 
+    // Also make single-point predictions.
+    for (size_t i = 0; i < y.n_elem; ++i)
+      REQUIRE(blr.Predict(matX.col(i)) == Approx(ridgePred[i]).epsilon(1));
+
     // Exit once a test case has completed.
     break;
   }
 
   REQUIRE(trial <= 3);
 }
+
+// Check that all constructor variants work.
+TEMPLATE_TEST_CASE("BayesianLinearRegressionConstructorVariantTest",
+    "[BayesianLinearRegressionTest]", arma::mat)
+{
+  typedef TestType MatType;
+
+  MatType matX;
+  arma::Row<typename MatType::elem_type> y;
+  size_t nDims = 5, nPoints = 100;
+  GenerateProblem(matX, y, nPoints, nDims, 0.5);
+
+  // The important thing here is that all of the inputs to the constructor are
+  // properly parsed.  The actual details of the models that are learned are
+  // less important and are checked by other tests.
+  BayesianLinearRegression<> blr1;
+  BayesianLinearRegression<> blr2(false);
+  BayesianLinearRegression<> blr3(false, true);
+  BayesianLinearRegression<> blr4(false, true, 100);
+  BayesianLinearRegression<> blr5(false, true, 110, 1e-3);
+  BayesianLinearRegression<> blr6(matX, y);
+  BayesianLinearRegression<> blr7(matX, y, false);
+  BayesianLinearRegression<> blr8(matX, y, false, true);
+  BayesianLinearRegression<> blr9(matX, y, false, true, 120);
+  BayesianLinearRegression<> blr10(matX, y, false, true, 130, 1e-2);
+
+  // Check that the hyperparameters as reported by the model are correct.
+  REQUIRE(blr1.Omega().n_elem == 0);
+
+  REQUIRE(blr2.Omega().n_elem == 0);
+  REQUIRE(blr2.CenterData() == false);
+
+  REQUIRE(blr3.Omega().n_elem == 0);
+  REQUIRE(blr3.CenterData() == false);
+  REQUIRE(blr3.ScaleData() == true);
+
+  REQUIRE(blr4.Omega().n_elem == 0);
+  REQUIRE(blr4.CenterData() == false);
+  REQUIRE(blr4.ScaleData() == true);
+  REQUIRE(blr4.MaxIterations() == 100);
+
+  REQUIRE(blr5.Omega().n_elem == 0);
+  REQUIRE(blr5.CenterData() == false);
+  REQUIRE(blr5.ScaleData() == true);
+  REQUIRE(blr5.MaxIterations() == 110);
+  REQUIRE(blr5.Tolerance() == 1e-3);
+
+  REQUIRE(blr6.Omega().n_elem == matX.n_rows);
+
+  REQUIRE(blr7.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr7.CenterData() == false);
+
+  REQUIRE(blr8.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr8.CenterData() == false);
+  REQUIRE(blr8.ScaleData() == true);
+
+  REQUIRE(blr9.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr9.CenterData() == false);
+  REQUIRE(blr9.ScaleData() == true);
+  REQUIRE(blr9.MaxIterations() == 120);
+
+  REQUIRE(blr10.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr10.CenterData() == false);
+  REQUIRE(blr10.ScaleData() == true);
+  REQUIRE(blr10.MaxIterations() == 130);
+  REQUIRE(blr10.Tolerance() == 1e-2);
+}
+
+// Test that all Train() variants work.
+TEMPLATE_TEST_CASE("BayesianLinearRegressionTrainVariantTest",
+    "[BayesianLinearRegressionTest]", arma::mat)
+{
+  typedef TestType MatType;
+
+  MatType matX;
+  arma::Row<typename MatType::elem_type> y;
+  size_t nDims = 5, nPoints = 100;
+  GenerateProblem(matX, y, nPoints, nDims, 0.5);
+
+  BayesianLinearRegression<> blr1, blr2, blr3, blr4, blr5;
+
+  blr1.Train(matX, y);
+  blr2.Train(matX, y, false);
+  blr3.Train(matX, y, false, true);
+  blr4.Train(matX, y, false, true, 100);
+  blr5.Train(matX, y, false, true, 110, 1e-3);
+
+  REQUIRE(blr1.Omega().n_elem == matX.n_rows);
+
+  REQUIRE(blr2.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr2.CenterData() == false);
+
+  REQUIRE(blr3.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr3.CenterData() == false);
+  REQUIRE(blr3.ScaleData() == true);
+
+  REQUIRE(blr4.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr4.CenterData() == false);
+  REQUIRE(blr4.ScaleData() == true);
+  REQUIRE(blr4.MaxIterations() == 100);
+
+  REQUIRE(blr5.Omega().n_elem == matX.n_rows);
+  REQUIRE(blr5.CenterData() == false);
+  REQUIRE(blr5.ScaleData() == true);
+  REQUIRE(blr5.MaxIterations() == 110);
+  REQUIRE(blr5.Tolerance() == 1e-3);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/bias_svd_test.cpp 4.4.0-1/src/mlpack/tests/bias_svd_test.cpp
--- 4.3.0-2/src/mlpack/tests/bias_svd_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/bias_svd_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -55,7 +55,7 @@ TEST_CASE("BiasSVDFunctionRandomEvaluate
       const double userBias = parameters(rank, user);
       const double itemBias = parameters(rank, item);
       const double ratingError = rating - userBias - itemBias -
-          arma::dot(parameters.col(user).subvec(0, rank - 1),
+          dot(parameters.col(user).subvec(0, rank - 1),
                     parameters.col(item).subvec(0, rank - 1));
       const double ratingErrorSquared = ratingError * ratingError;
 
@@ -263,7 +263,7 @@ TEST_CASE("BiasSVDFunctionOptimize", "[B
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     data(2, i) = userBias + itemBias +
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
   }
 
@@ -284,7 +284,7 @@ TEST_CASE("BiasSVDFunctionOptimize", "[B
     const double userBias = optParameters(rank, user);
     const double itemBias = optParameters(rank, item);
     predictedData(0, i) = userBias + itemBias +
-        arma::dot(optParameters.col(user).subvec(0, rank - 1),
+        dot(optParameters.col(user).subvec(0, rank - 1),
                   optParameters.col(item).subvec(0, rank - 1));
   }
 
@@ -331,7 +331,7 @@ TEST_CASE("BiasSVDFunctionParallelOptimi
     const double userBias = parameters(rank, user);
     const double itemBias = parameters(rank, item);
     data(2, i) = userBias + itemBias +
-        arma::dot(parameters.col(user).subvec(0, rank - 1),
+        dot(parameters.col(user).subvec(0, rank - 1),
                   parameters.col(item).subvec(0, rank - 1));
   }
 
@@ -359,7 +359,7 @@ TEST_CASE("BiasSVDFunctionParallelOptimi
     const double userBias = optParameters(rank, user);
     const double itemBias = optParameters(rank, item);
     predictedData(0, i) = userBias + itemBias +
-        arma::dot(optParameters.col(user).subvec(0, rank - 1),
+        dot(optParameters.col(user).subvec(0, rank - 1),
                   optParameters.col(item).subvec(0, rank - 1));
   }
 
diff -pruN 4.3.0-2/src/mlpack/tests/block_krylov_svd_test.cpp 4.4.0-1/src/mlpack/tests/block_krylov_svd_test.cpp
--- 4.3.0-2/src/mlpack/tests/block_krylov_svd_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/block_krylov_svd_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,9 +32,8 @@ void CreateNoisyLowRankMatrix(arma::mat&
 
   arma::vec ids = arma::linspace<arma::vec>(0, n - 1, n);
 
-  arma::vec lowRank = ((1 - strength) *
-      arma::exp(-1.0 * arma::pow((ids / rank), 2)));
-  arma::vec tail = strength * arma::exp(-0.1 * ids / rank);
+  arma::vec lowRank = ((1 - strength) * exp(-1.0 * pow((ids / rank), 2)));
+  arma::vec tail = strength * exp(-0.1 * ids / rank);
 
   arma::mat s = arma::zeros<arma::mat>(n, n);
   s.diag() = lowRank + tail;
@@ -57,11 +56,10 @@ TEST_CASE("RandomizedBlockKrylovSVDRecon
 
   arma::mat s = arma::diagmat(arma::vec("1 0.1 0.01"));
 
-  arma::mat data = arma::trans(U * arma::diagmat(s) * V.t());
+  arma::mat data = trans(U * arma::diagmat(s) * V.t());
 
   // Center the data into a temporary matrix.
-  arma::mat centeredData;
-  Center(data, centeredData);
+  arma::mat centeredData = data.each_col() - arma::mean(data, 1);
 
   arma::mat U1, U2, V1, V2;
   arma::vec s1, s2, s3;
diff -pruN 4.3.0-2/src/mlpack/tests/callback_test.cpp 4.4.0-1/src/mlpack/tests/callback_test.cpp
--- 4.3.0-2/src/mlpack/tests/callback_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/callback_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -136,8 +136,7 @@ TEST_CASE("LRWithOptimizerCallback", "[C
   ens::StandardSGD sgd(0.1, 1, 5);
   LogisticRegression<> logisticRegression(data, responses, sgd, 0.001);
   std::stringstream stream;
-  logisticRegression.Train<ens::StandardSGD>(data, responses, sgd,
-                                             ens::PrintLoss(stream));
+  logisticRegression.Train(data, responses, sgd, ens::PrintLoss(stream));
 
   REQUIRE(stream.str().length() > 0);
 }
@@ -210,8 +209,9 @@ TEST_CASE("SRWithOptimizerCallback", "[C
   ens::StandardSGD sgd(0.1, 1, 5);
   std::stringstream stream;
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
-  sr.Train(data, labels, numClasses, sgd, ens::ProgressBar(70, stream));
+  SoftmaxRegression<> sr;
+  sr.Train(data, labels, numClasses, sgd, lambda, true,
+      ens::ProgressBar(70, stream));
 
   REQUIRE(stream.str().length() > 0);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/catch.hpp 4.4.0-1/src/mlpack/tests/catch.hpp
--- 4.3.0-2/src/mlpack/tests/catch.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/catch.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -69,6 +69,9 @@
 // See e.g.:
 // https://opensource.apple.com/source/CarbonHeaders/CarbonHeaders-18.1/TargetConditionals.h.auto.html
 #ifdef __APPLE__
+#  ifndef __has_extension
+#    define __has_extension(x) 0
+#  endif
 #  include <TargetConditionals.h>
 #  if (defined(TARGET_OS_OSX) && TARGET_OS_OSX == 1) || \
       (defined(TARGET_OS_MAC) && TARGET_OS_MAC == 1)
diff -pruN 4.3.0-2/src/mlpack/tests/cli_binding_test.cpp 4.4.0-1/src/mlpack/tests/cli_binding_test.cpp
--- 4.3.0-2/src/mlpack/tests/cli_binding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/cli_binding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -476,7 +476,7 @@ TEST_CASE("SetParamDoubleTest", "[CLIOpt
 
   // Now create second value.
   double dd2 = 1.0;
-  MLPACK_ANY a = dd2;
+  std::any a = dd2;
   SetParam<double>((util::ParamData&) d, (const void*) &a, (void*) NULL);
 
   // Make sure it's the right thing.
@@ -499,10 +499,10 @@ TEST_CASE("SetParamBoolTest", "[CLIOptio
 
   // Now create second value.
   bool b2 = true;
-  MLPACK_ANY a = b2;
+  std::any a = b2;
   SetParam<bool>((util::ParamData&) d, (const void*) &a, (void*) NULL);
 
-  REQUIRE(MLPACK_ANY_CAST<bool>(d.value) == true);
+  REQUIRE(std::any_cast<bool>(d.value) == true);
 }
 
 // Test that calling SetParam on a matrix sets the string correctly.
@@ -519,14 +519,14 @@ TEST_CASE("SetParamMatrixTest", "[CLIOpt
 
   // Get a new string.
   string newFilename = "new.csv";
-  MLPACK_ANY a2 = newFilename;
+  std::any a2 = newFilename;
 
   SetParam<arma::mat>((util::ParamData&) d, (const void*) &a2,
       (void*) NULL);
 
   // Make sure the change went through.
   tuple<arma::mat, TupleType>& t =
-      *MLPACK_ANY_CAST<tuple<arma::mat, TupleType>>(&d.value);
+      *std::any_cast<tuple<arma::mat, TupleType>>(&d.value);
   REQUIRE(get<0>(get<1>(t)) == "new.csv");
 }
 
@@ -542,14 +542,14 @@ TEST_CASE("SetParamModelTest", "[CLIOpti
 
   // Get a new string.
   string newFilename = "new_kernel.bin";
-  MLPACK_ANY a2 = newFilename;
+  std::any a2 = newFilename;
 
   SetParam<GaussianKernel>((util::ParamData&) d, (const void*) &a2,
       (void*) NULL);
 
   // Make sure the change went through.
   tuple<GaussianKernel*, string>& t =
-      *MLPACK_ANY_CAST<tuple<GaussianKernel*, string>>(&d.value);
+      *std::any_cast<tuple<GaussianKernel*, string>>(&d.value);
 
   REQUIRE(get<1>(t) == "new_kernel.bin");
 }
@@ -575,14 +575,14 @@ TEST_CASE("SetParamDatasetInfoMatTest",
 
   // Now get new filename.
   string newFilename = "new_filename.csv";
-  MLPACK_ANY a2 = newFilename;
+  std::any a2 = newFilename;
 
   SetParam<tuple<DatasetInfo, arma::mat>>((util::ParamData&) d,
       (const void*) &a2, (void*) NULL);
 
   // Check that the name is right.
   tuple<tuple<DatasetInfo, arma::mat>, TupleType>& t3 =
-      *MLPACK_ANY_CAST<tuple<tuple<DatasetInfo, arma::mat>, TupleType>>(
+      *std::any_cast<tuple<tuple<DatasetInfo, arma::mat>, TupleType>>(
       &d.value);
 
   REQUIRE(get<0>(get<1>(t3)) == "new_filename.csv");
diff -pruN 4.3.0-2/src/mlpack/tests/cosine_tree_test.cpp 4.4.0-1/src/mlpack/tests/cosine_tree_test.cpp
--- 4.3.0-2/src/mlpack/tests/cosine_tree_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/cosine_tree_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,7 +34,7 @@ TEST_CASE("CosineTreeNoSplit", "[CosineT
 
   // Make a cosine tree, with the generated dataset and the defined constants.
   // Note that the value of epsilon is one.
-  CosineTree ctree(data, epsilon, delta);
+  CosineTree<> ctree(data, epsilon, delta);
   arma::mat basis;
   ctree.GetFinalBasis(basis);
 
@@ -57,17 +57,17 @@ TEST_CASE("CosineNodeCosineSplit", "[Cos
 
   // Make a random dataset and the root object.
   arma::mat data = arma::randu(numRows, numCols);
-  CosineTree root(data);
+  CosineTree<> root(data);
 
   // Stack for depth first search of the tree.
-  std::vector<CosineTree*> nodeStack;
+  std::vector<CosineTree<>*> nodeStack;
   nodeStack.push_back(&root);
 
   // While stack is not empty.
   while (nodeStack.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode, *currentLeft, *currentRight;
+    CosineTree<> *currentNode, *currentLeft, *currentRight;
     currentNode = nodeStack.back();
     currentNode->CosineNodeSplit();
     nodeStack.pop_back();
@@ -111,7 +111,7 @@ TEST_CASE("CosineNodeCosineSplit", "[Cos
       // CosineNodeSplit may differ from cosineMax below, so we have to handle
       // minor differences.
       double cosineMax = arma::max(cosines % (cosines < 1.0 + precision));
-      double cosineMin = arma::min(cosines);
+      double cosineMin = min(cosines);
       // If max(cosines) is close to 1.0 cosineMax and cosineMax2 may
       // differ significantly.
       double cosineMax2 = arma::max(cosines % (cosines < 1.0 - precision));
@@ -173,14 +173,14 @@ TEST_CASE("CosineTreeModifiedGramSchmidt
 
   // Declare a queue and a dummy CosineTree object.
   CompareCosineNode comp;
-  CosineNodeQueue basisQueue;
-  CosineTree dummyTree(data, epsilon, delta);
+  CosineNodeQueue<> basisQueue;
+  CosineTree<> dummyTree(data, epsilon, delta);
 
   for (size_t i = 0; i < numCols; ++i)
   {
     // Make a new CosineNode object.
-    CosineTree* basisNode;
-    basisNode = new CosineTree(data);
+    CosineTree<>* basisNode;
+    basisNode = new CosineTree<>(data);
 
     // Use the columns of the dataset as random centroids.
     arma::vec centroid = data.col(i);
@@ -190,13 +190,13 @@ TEST_CASE("CosineTreeModifiedGramSchmidt
     dummyTree.ModifiedGramSchmidt(basisQueue, centroid, newBasisVector);
 
     // Check if the obtained vector is orthonormal to the basis vectors.
-    CosineNodeQueue::const_iterator j = basisQueue.cbegin();
-    CosineTree* currentNode;
+    CosineNodeQueue<>::const_iterator j = basisQueue.cbegin();
+    CosineTree<>* currentNode;
 
     for (; j != basisQueue.cend(); ++j)
     {
       currentNode = *j;
-      REQUIRE(arma::dot(currentNode->BasisVector(), newBasisVector) ==
+      REQUIRE(dot(currentNode->BasisVector(), newBasisVector) ==
           Approx(0.0).margin(1e-5));
     }
 
@@ -210,7 +210,7 @@ TEST_CASE("CosineTreeModifiedGramSchmidt
   // Deallocate memory given to the objects.
   for (size_t i = 0; i < numCols; ++i)
   {
-    CosineTree* currentNode;
+    CosineTree<>* currentNode;
     currentNode = basisQueue.front();
     std::pop_heap(basisQueue.begin(), basisQueue.end(), comp);
     basisQueue.pop_back();
@@ -236,17 +236,17 @@ TEST_CASE("CopyConstructorAndOperatorCos
   arma::mat* data = new arma::mat(numRows, numCols, arma::fill::randu);
 
   // Make a cosine tree, with the generated dataset.
-  CosineTree* ctree1 = new CosineTree(*data);
+  CosineTree<>* ctree1 = new CosineTree<>(*data);
 
   // Stacks for depth first search of the tree.
-  std::vector<CosineTree*> nodeStack1, nodeStack2, nodeStack3;
+  std::vector<CosineTree<>*> nodeStack1, nodeStack2, nodeStack3;
   nodeStack1.push_back(ctree1);
 
   // While stack is not empty.
   while (nodeStack1.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode1, *currentLeft1, *currentRight1;
+    CosineTree<> *currentNode1, *currentLeft1, *currentRight1;
 
     currentNode1 = nodeStack1.back();
     currentNode1->CosineNodeSplit();
@@ -268,8 +268,8 @@ TEST_CASE("CopyConstructorAndOperatorCos
   }
 
   // Copy constructor and operator.
-  CosineTree ctree2(*ctree1);
-  CosineTree ctree3 = *ctree1;
+  CosineTree<> ctree2(*ctree1);
+  CosineTree<> ctree3 = *ctree1;
 
   delete ctree1;
   delete data;
@@ -281,8 +281,8 @@ TEST_CASE("CopyConstructorAndOperatorCos
   while (nodeStack2.size() && nodeStack3.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode2, *currentLeft2, *currentRight2;
-    CosineTree *currentNode3, *currentLeft3, *currentRight3;
+    CosineTree<> *currentNode2, *currentLeft2, *currentRight2;
+    CosineTree<> *currentNode3, *currentLeft3, *currentRight3;
 
     currentNode2 = nodeStack2.back();
     nodeStack2.pop_back();
@@ -337,17 +337,17 @@ TEST_CASE("MoveConstructorAndOperatorCos
   arma::mat data = arma::randu(numRows, numCols);
 
   // Make a cosine tree, with the generated dataset.
-  CosineTree ctree1(data);
+  CosineTree<> ctree1(data);
 
   // Stacks for depth first search of the tree.
-  std::vector<CosineTree*> nodeStack1, nodeStack2, nodeStack3;
+  std::vector<CosineTree<>*> nodeStack1, nodeStack2, nodeStack3;
   nodeStack1.push_back(&ctree1);
 
   // While stack is not empty.
   while (nodeStack1.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode1, *currentLeft1, *currentRight1;
+    CosineTree<> *currentNode1, *currentLeft1, *currentRight1;
 
     currentNode1 = nodeStack1.back();
     currentNode1->CosineNodeSplit();
@@ -369,7 +369,7 @@ TEST_CASE("MoveConstructorAndOperatorCos
   }
 
   // Move constructor.
-  CosineTree ctree2(std::move(ctree1));
+  CosineTree<> ctree2(std::move(ctree1));
 
   nodeStack2.push_back(&ctree2);
 
@@ -377,7 +377,7 @@ TEST_CASE("MoveConstructorAndOperatorCos
   while (nodeStack2.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode2, *currentLeft2, *currentRight2;
+    CosineTree<> *currentNode2, *currentLeft2, *currentRight2;
 
     currentNode2 = nodeStack2.back();
     nodeStack2.pop_back();
@@ -398,7 +398,7 @@ TEST_CASE("MoveConstructorAndOperatorCos
   }
 
   // Move operator.
-  CosineTree ctree3 = std::move(ctree2);
+  CosineTree<> ctree3 = std::move(ctree2);
 
   nodeStack3.push_back(&ctree3);
 
@@ -406,7 +406,7 @@ TEST_CASE("MoveConstructorAndOperatorCos
   while (nodeStack3.size())
   {
     // Pop a node from the stack and split it.
-    CosineTree *currentNode3, *currentLeft3, *currentRight3;
+    CosineTree<> *currentNode3, *currentLeft3, *currentRight3;
 
     currentNode3 = nodeStack3.back();
     nodeStack3.pop_back();
diff -pruN 4.3.0-2/src/mlpack/tests/cv_test.cpp 4.4.0-1/src/mlpack/tests/cv_test.cpp
--- 4.3.0-2/src/mlpack/tests/cv_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/cv_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -207,7 +207,7 @@ TEST_CASE("MSETest", "[CVTest]")
   arma::mat trainingData("0 1");
   arma::rowvec trainingResponses("-1 0");
 
-  LinearRegression lr(trainingData, trainingResponses);
+  LinearRegression<> lr(trainingData, trainingResponses);
 
   // Making three responses that differ from the correct ones by 0, 1, and 2
   // respectively
@@ -229,7 +229,7 @@ TEST_CASE("R2ScoreTest", "[CVTest]")
   arma::mat trainingData("0 1");
   arma::rowvec trainingResponses("-1 0");
 
-  LinearRegression lr(trainingData, trainingResponses);
+  LinearRegression<> lr(trainingData, trainingResponses);
 
   // Making five responses that are the output of regression function f(x)
   // with some responses having a slight deviation of 0.005.
@@ -256,7 +256,7 @@ TEST_CASE("AdjR2ScoreTest", "[CVTest]")
   arma::rowvec Y;
   Y = { 3, 5, 7, 9, 11, 13 };
 
-  LinearRegression lr(X, Y);
+  LinearRegression<> lr(X, Y);
 
   // Theoretically Adjusted R squared should be equal 1
   double expAdjR2 = 1;
@@ -309,11 +309,11 @@ void CheckPredictionsType()
  */
 TEST_CASE("PredictionsTypeTest", "[CVTest]")
 {
-  CheckPredictionsType<LinearRegression, arma::rowvec>();
+  CheckPredictionsType<LinearRegression<>, arma::rowvec>();
   // CheckPredictionsType<FFN<>, arma::mat>();
 
   CheckPredictionsType<LogisticRegression<>, arma::Row<size_t>>();
-  CheckPredictionsType<SoftmaxRegression, arma::Row<size_t>>();
+  CheckPredictionsType<SoftmaxRegression<>, arma::Row<size_t>>();
   CheckPredictionsType<HoeffdingTree<>, arma::Row<size_t>, arma::mat>();
   CheckPredictionsType<HoeffdingTree<>, arma::Row<size_t>, arma::imat>();
   CheckPredictionsType<DecisionTree<>, arma::Row<size_t>, arma::mat,
@@ -328,14 +328,14 @@ TEST_CASE("PredictionsTypeTest", "[CVTes
  */
 TEST_CASE("SupportsWeightsTest", "[CVTest]")
 {
-  static_assert(MetaInfoExtractor<LinearRegression>::SupportsWeights,
+  static_assert(MetaInfoExtractor<LinearRegression<>>::SupportsWeights,
       "Value should be true");
   static_assert(MetaInfoExtractor<DecisionTree<>>::SupportsWeights,
       "Value should be true");
   static_assert(MetaInfoExtractor<DecisionTree<>, arma::mat, arma::urowvec,
       arma::Row<float>>::SupportsWeights, "Value should be true");
 
-  static_assert(!MetaInfoExtractor<LARS>::SupportsWeights,
+  static_assert(!MetaInfoExtractor<LARS<>>::SupportsWeights,
       "Value should be false");
   static_assert(!MetaInfoExtractor<LogisticRegression<>>::SupportsWeights,
       "Value should be false");
@@ -360,7 +360,7 @@ void CheckWeightsType()
  */
 TEST_CASE("WeightsTypeTest", "[CVTest]")
 {
-  CheckWeightsType<LinearRegression, arma::rowvec>();
+  CheckWeightsType<LinearRegression<>, arma::rowvec>();
   CheckWeightsType<DecisionTree<>, arma::rowvec>();
   CheckWeightsType<DecisionTree<>, arma::Row<float>, arma::mat,
       arma::Row<size_t>, arma::Row<float>>();
@@ -374,9 +374,9 @@ TEST_CASE("TakesDatasetInfoTest", "[CVTe
 {
   static_assert(MetaInfoExtractor<DecisionTree<>>::TakesDatasetInfo,
       "Value should be true");
-  static_assert(!MetaInfoExtractor<LinearRegression>::TakesDatasetInfo,
+  static_assert(!MetaInfoExtractor<LinearRegression<>>::TakesDatasetInfo,
       "Value should be false");
-  static_assert(!MetaInfoExtractor<SoftmaxRegression>::TakesDatasetInfo,
+  static_assert(!MetaInfoExtractor<SoftmaxRegression<>>::TakesDatasetInfo,
       "Value should be false");
 }
 
@@ -388,11 +388,11 @@ TEST_CASE("TakesNumClassesTest", "[CVTes
 {
   static_assert(MetaInfoExtractor<DecisionTree<>>::TakesNumClasses,
       "Value should be true");
-  static_assert(MetaInfoExtractor<SoftmaxRegression>::TakesNumClasses,
+  static_assert(MetaInfoExtractor<SoftmaxRegression<>>::TakesNumClasses,
       "Value should be true");
-  static_assert(!MetaInfoExtractor<LinearRegression>::TakesNumClasses,
+  static_assert(!MetaInfoExtractor<LinearRegression<>>::TakesNumClasses,
       "Value should be false");
-  static_assert(!MetaInfoExtractor<LARS>::TakesNumClasses,
+  static_assert(!MetaInfoExtractor<LARS<>>::TakesNumClasses,
       "Value should be false");
 }
 
@@ -425,28 +425,28 @@ TEST_CASE("SimpleCVMSETest", "[CVTest]")
 
   double expectedMSE = (0 * 0 + 1 * 1 + 2 * 2) / 3.0;
 
-  SimpleCV<LinearRegression, MSE> cv(0.6, data, responses);
+  SimpleCV<LinearRegression<>, MSE> cv(0.6, data, responses);
 
   REQUIRE(cv.Evaluate() == Approx(expectedMSE).epsilon(1e-7));
 
   arma::mat noiseData("-1 -2 -3 -4 -5");
   arma::rowvec noiseResponses("10 20 30 40 50");
 
-  arma::mat allData = arma::join_rows(noiseData, data);
-  arma::rowvec allResponces = arma::join_rows(noiseResponses, responses);
+  arma::mat allData = join_rows(noiseData, data);
+  arma::rowvec allResponces = join_rows(noiseResponses, responses);
 
-  arma::rowvec weights = arma::join_rows(arma::zeros(noiseData.n_cols).t(),
+  arma::rowvec weights = join_rows(arma::zeros(noiseData.n_cols).t(),
       arma::ones(data.n_cols).t());
 
-  SimpleCV<LinearRegression, MSE> weightedCV(0.3, allData, allResponces,
+  SimpleCV<LinearRegression<>, MSE> weightedCV(0.3, allData, allResponces,
       weights);
 
   REQUIRE(weightedCV.Evaluate() == Approx(expectedMSE).epsilon(1e-7));
 
-  arma::rowvec weights2 = arma::join_rows(arma::zeros(noiseData.n_cols - 1).t(),
+  arma::rowvec weights2 = join_rows(arma::zeros(noiseData.n_cols - 1).t(),
       arma::ones(data.n_cols + 1).t());
 
-  SimpleCV<LinearRegression, MSE> weightedCV2(0.3, allData, allResponces,
+  SimpleCV<LinearRegression<>, MSE> weightedCV2(0.3, allData, allResponces,
       weights2);
 
   REQUIRE(std::abs(weightedCV2.Evaluate() - expectedMSE) > 1e-5);
@@ -506,21 +506,21 @@ TEST_CASE("SimpleCVWithDTTest", "[CVTest
     arma::Row<size_t> predictedLabels = PredictLabelsWithDT(testData,
         trainingData, trainingLabels, numClasses, minimumLeafSize);
     SimpleCV<DecisionTree<InformationGain>, Accuracy> cv(0.5, data,
-        arma::join_rows(trainingLabels, predictedLabels), numClasses);
+        join_rows(trainingLabels, predictedLabels), numClasses);
     REQUIRE(cv.Evaluate(minimumLeafSize) == Approx(1.0).epsilon(1e-7));
   }
   {
     arma::Row<size_t> predictedLabels = PredictLabelsWithDT(testData,
         trainingData, datasetInfo, trainingLabels, numClasses, minimumLeafSize);
     SimpleCV<DecisionTree<InformationGain>, Accuracy> cv(0.5, data, datasetInfo,
-        arma::join_rows(trainingLabels, predictedLabels), numClasses);
+        join_rows(trainingLabels, predictedLabels), numClasses);
     REQUIRE(cv.Evaluate(minimumLeafSize) == Approx(1.0).epsilon(1e-7));
   }
   {
     arma::Row<size_t> predictedLabels = PredictLabelsWithDT(testData,
         trainingData, trainingLabels, numClasses, weights, minimumLeafSize);
     SimpleCV<DecisionTree<InformationGain>, Accuracy> cv(0.5, data,
-        arma::join_rows(trainingLabels, predictedLabels), numClasses, weights);
+        join_rows(trainingLabels, predictedLabels), numClasses, weights);
     REQUIRE(cv.Evaluate(minimumLeafSize) == Approx(1.0).epsilon(1e-7));
   }
   {
@@ -528,7 +528,7 @@ TEST_CASE("SimpleCVWithDTTest", "[CVTest
         trainingData, datasetInfo, trainingLabels, numClasses, weights,
         minimumLeafSize);
     SimpleCV<DecisionTree<InformationGain>, Accuracy> cv(0.5, data, datasetInfo,
-        arma::join_rows(trainingLabels, predictedLabels), numClasses, weights);
+        join_rows(trainingLabels, predictedLabels), numClasses, weights);
     REQUIRE(cv.Evaluate(minimumLeafSize) == Approx(1.0).epsilon(1e-7));
   }
 }
@@ -543,7 +543,7 @@ TEST_CASE("KFoldCVMSETest", "[CVTest]")
   arma::rowvec responses("0 1  1 3");
 
   // 2-fold cross-validation, no shuffling.
-  KFoldCV<LinearRegression, MSE> cv(2, data, responses, false);
+  KFoldCV<LinearRegression<>, MSE> cv(2, data, responses, false);
 
   // In each of two validation tests the MSE value should be the same.
   double expectedMSE =
@@ -620,8 +620,8 @@ TEST_CASE("KFoldCVWithWeightedLRTest", "
   arma::rowvec responses("1 2 30 40");
   arma::rowvec weights("1 1 0 0");
 
-  KFoldCV<LinearRegression, MSE> cv(2, arma::join_rows(data, data),
-      arma::join_rows(responses, responses), arma::join_rows(weights, weights),
+  KFoldCV<LinearRegression<>, MSE> cv(2, join_rows(data, data),
+      join_rows(responses, responses), join_rows(weights, weights),
       false);
   cv.Evaluate();
 
@@ -649,9 +649,9 @@ TEST_CASE("KFoldCVWithDTTest", "[CVTest]
   arma::Row<size_t> labels = originalLabels.cols(0, 1199);
   arma::rowvec weights(data.n_cols, arma::fill::randu);
 
-  arma::mat doubledData = arma::join_rows(data, data);
-  arma::Row<size_t> doubledLabels = arma::join_rows(labels, labels);
-  arma::rowvec doubledWeights = arma::join_rows(weights, weights);
+  arma::mat doubledData = join_rows(data, data);
+  arma::Row<size_t> doubledLabels = join_rows(labels, labels);
+  arma::rowvec doubledWeights = join_rows(weights, weights);
 
   size_t numClasses = 5;
   size_t minimumLeafSize = 8;
Binary files 4.3.0-2/src/mlpack/tests/data/mnist_first250_training_4s_and_9s.tar.bz2 and 4.4.0-1/src/mlpack/tests/data/mnist_first250_training_4s_and_9s.tar.bz2 differ
diff -pruN 4.3.0-2/src/mlpack/tests/dbscan_test.cpp 4.4.0-1/src/mlpack/tests/dbscan_test.cpp
--- 4.3.0-2/src/mlpack/tests/dbscan_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/dbscan_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,6 +17,18 @@
 
 using namespace mlpack;
 
+/**
+ * A couple of handful declarations for float32 testing.
+ * These will be removed when we refactor the Bounds to accept MatType.
+ * For now, we will keep the following declarations.
+ */
+template<typename MetricType>
+using FloatHRectBound = HRectBound<MetricType, float>;
+
+template<typename MetricType, typename StatisticType, typename MatType>
+using FloatKDTree = BinarySpaceTree<MetricType, StatisticType, MatType,
+                                    FloatHRectBound, MidpointSplit>;
+
 TEST_CASE("OneClusterTest", "[DBSCANTest]")
 {
   // Make sure that if we have points in the unit box, and if we set epsilon
@@ -193,6 +205,35 @@ TEST_CASE("OutlierSingleModeTest", "[DBS
 
   arma::Row<size_t> assignments;
   const size_t clusters = d.Cluster(points, assignments);
+
+  REQUIRE(clusters > 0);
+  REQUIRE(assignments.n_elem == points.n_cols);
+  REQUIRE(assignments[15] == SIZE_MAX);
+  REQUIRE(assignments[45] == SIZE_MAX);
+  REQUIRE(assignments[101] == SIZE_MAX);
+}
+
+/**
+ * Check that outliers are properly labeled as noise
+ * in the case of float32.
+ */
+TEST_CASE("Float32OutlierSingleModeTest", "[DBSCANTest]")
+{
+  arma::Mat<float> points(2, 200, arma::fill::randu);
+
+  // Add 3 outliers.
+  points.col(15) = arma::Col<float>("10.3 1.6");
+  points.col(45) = arma::Col<float>("-100 0.0");
+  points.col(101) = arma::Col<float>("1.5 1.5");
+
+  DBSCAN<RangeSearch<
+         EuclideanDistance,
+         arma::Mat<float>,
+         FloatKDTree>,
+         OrderedPointSelection> d(0.1, 3, false);
+
+  arma::Row<size_t> assignments;
+  const size_t clusters = d.Cluster(points, assignments);
 
   REQUIRE(clusters > 0);
   REQUIRE(assignments.n_elem == points.n_cols);
diff -pruN 4.3.0-2/src/mlpack/tests/decision_tree_regressor_test.cpp 4.4.0-1/src/mlpack/tests/decision_tree_regressor_test.cpp
--- 4.3.0-2/src/mlpack/tests/decision_tree_regressor_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/decision_tree_regressor_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -853,9 +853,8 @@ TEST_CASE("CategoricalWeightedBuildTest_
   for (size_t i = 2000; i < 4000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(trainingData, randomNoise);
-  arma::rowvec fullResponses = arma::join_rows(trainingResponses,
-      randomResponses);
+  arma::mat fullData = join_rows(trainingData, randomNoise);
+  arma::rowvec fullResponses = join_rows(trainingResponses, randomResponses);
 
   // Build the tree.
   DecisionTreeRegressor<> tree(fullData, di, fullResponses, weights, 10);
@@ -908,9 +907,8 @@ TEST_CASE("CategoricalMADGainWeightedBui
   for (size_t i = 2000; i < 4000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(trainingData, randomNoise);
-  arma::rowvec fullResponses = arma::join_rows(trainingResponses,
-      randomResponses);
+  arma::mat fullData = join_rows(trainingData, randomNoise);
+  arma::rowvec fullResponses = join_rows(trainingResponses, randomResponses);
 
   // Build the tree using MAD gain.
   DecisionTreeRegressor<MADGain> tree(fullData, di, fullResponses, weights,
@@ -924,84 +922,61 @@ TEST_CASE("CategoricalMADGainWeightedBui
 
   // Make sure we get reasonable rmse.
   const double rmse = RMSE(predictions, testResponses);
-  REQUIRE(rmse < 1.05);
-}
-
-/**
- * Test that the decision tree generalizes reasonably.
- */
-TEST_CASE("SimpleGeneralizationTest_", "[DecisionTreeRegressorTest]")
-{
-  // Loading data.
-  data::DatasetInfo info;
-  arma::mat trainData, testData;
-  arma::rowvec trainResponses, testResponses;
-  LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
-      info);
-  arma::rowvec weights = arma::ones<arma::rowvec>(trainResponses.n_elem);
-
-  // Build decision tree.
-  DecisionTreeRegressor<> d(trainData, info, trainResponses);
-  DecisionTreeRegressor<> wd(trainData, info, trainResponses, weights);
-
-  // Get the predicted test responses.
-  arma::rowvec predictions;
-  d.Predict(testData, predictions);
-
-  REQUIRE(predictions.n_elem == testData.n_cols);
-
-  // Figure out rmse.
-  double rmse = RMSE(predictions, testResponses);
-  REQUIRE(rmse < 6.1);
-
-  // Reset the predictions.
-  predictions.zeros();
-  wd.Predict(testData, predictions);
-
-  REQUIRE(predictions.n_elem == testData.n_cols);
-
-  // Figure out rmse.
-  rmse = RMSE(predictions, testResponses);
-  REQUIRE(rmse < 6.1);
+  REQUIRE(rmse < 1.25);
 }
 
 /**
  * Test that the decision tree generalizes reasonably when built on float data.
  */
-TEST_CASE("SimpleGeneralizationFMatTest_", "[DecisionTreeRegressorTest]")
+TEMPLATE_TEST_CASE("SimpleGeneralizationTest_",
+    "[DecisionTreeRegressorTest]", float, double)
 {
-  // Loading data.
-  data::DatasetInfo info;
-  arma::fmat trainData, testData;
-  arma::rowvec trainLabels, testLabels;
-  LoadBostonHousingDataset(trainData, testData, trainLabels, testLabels, info);
-
-  // Initialize an all-ones weight matrix.
-  arma::rowvec weights(trainLabels.n_cols, arma::fill::ones);
-
-  // Build decision tree.
-  DecisionTreeRegressor<> d(trainData, trainLabels);
-  DecisionTreeRegressor<> wd(trainData, trainLabels, weights);
-
-  // Get the predicted test labels.
-  arma::rowvec predictions;
-  d.Predict(testData, predictions);
+  typedef TestType ElemType;
 
-  REQUIRE(predictions.n_elem == testData.n_cols);
-
-  // Figure out the rmse.
-  double rmse = RMSE(predictions, testLabels);
-  REQUIRE(rmse < 6.0);
-
-  // Reset the prediction.
-  predictions.zeros();
-  wd.Predict(testData, predictions);
-
-  REQUIRE(predictions.n_elem == testData.n_cols);
+  // Allow three trials.
+  bool success = false;
+  for (size_t trial = 0; trial < 3; ++trial)
+  {
+    // Loading data.
+    data::DatasetInfo info;
+    arma::Mat<ElemType> trainData, testData;
+    arma::Row<ElemType> trainResponses, testResponses;
+    LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
+        info);
+
+    // Initialize an all-ones weight matrix.
+    arma::rowvec weights(trainResponses.n_cols, arma::fill::ones);
+
+    // Build decision tree.
+    DecisionTreeRegressor<> d(trainData, trainResponses);
+    DecisionTreeRegressor<> wd(trainData, trainResponses, weights);
+
+    // Get the predicted test labels.
+    arma::Row<ElemType> predictions;
+    d.Predict(testData, predictions);
+
+    REQUIRE(predictions.n_elem == testData.n_cols);
+
+    // Figure out the rmse.
+    ElemType rmse = RMSE(predictions, testResponses);
+
+    // Reset the prediction.
+    predictions.zeros();
+    wd.Predict(testData, predictions);
+
+    REQUIRE(predictions.n_elem == testData.n_cols);
+
+    // Figure out the rmse.
+    ElemType wdrmse = RMSE(predictions, testResponses);
+
+    if (rmse <= 6.2 && wdrmse <= 6.2)
+    {
+      success = true;
+      break;
+    }
+  }
 
-  // Figure out the rmse.
-  double wdrmse = RMSE(predictions, testLabels);
-  REQUIRE(wdrmse < 6.0);
+  REQUIRE(success == true);
 }
 
 /**
@@ -1011,42 +986,52 @@ TEST_CASE("SimpleGeneralizationFMatTest_
  */
 TEST_CASE("WeightedDecisionTreeTest_", "[DecisionTreeRegressorTest]")
 {
-  // Loading data.
-  data::DatasetInfo info;
-  arma::mat trainData, testData;
-  arma::rowvec trainResponses, testResponses;
-  LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
-      info);
-
-  // Add some noise.
-  arma::mat noise(trainData.n_rows, 100, arma::fill::randu);
-  arma::rowvec noiseResponses(100);
-  for (size_t i = 0; i < noiseResponses.n_elem; ++i)
-    noiseResponses[i] = 15 + Random(0, 10); // Random response.
-
-  // Concatenate data matrices.
-  arma::mat data = arma::join_rows(trainData, noise);
-  arma::rowvec fullResponses = arma::join_rows(trainResponses, noiseResponses);
-
-  // Now set weights.
-  arma::rowvec weights(trainData.n_cols + 100);
-  for (size_t i = 0; i < trainData.n_cols; ++i)
-    weights[i] = Random(0.9, 1.0);
-  for (size_t i = trainData.n_cols; i < trainData.n_cols + 100; ++i)
-    weights[i] = Random(0.0, 0.01); // Low weights for false points.
-
-  // Now build the decision tree.
-  DecisionTreeRegressor<> d(data, fullResponses, weights);
-
-  // Now we can check that we get good performance on the test set.
-  arma::rowvec predictions;
-  d.Predict(testData, predictions);
-
-  REQUIRE(predictions.n_elem == testData.n_cols);
-
-  // Figure out the rmse.
-  double rmse = RMSE(predictions, testResponses);
-  REQUIRE(rmse < 6.0);
+  // Allow three trials for success.
+  bool success = false;
+  for (size_t trial = 0; trial < 3; ++trial)
+  {
+    // Loading data.
+    data::DatasetInfo info;
+    arma::mat trainData, testData;
+    arma::rowvec trainResponses, testResponses;
+    LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
+        info);
+
+    // Add some noise.
+    arma::mat noise(trainData.n_rows, 100, arma::fill::randu);
+    arma::rowvec noiseResponses(100);
+    for (size_t i = 0; i < noiseResponses.n_elem; ++i)
+      noiseResponses[i] = 15 + Random(0, 10); // Random response.
+
+    // Concatenate data matrices.
+    arma::mat data = join_rows(trainData, noise);
+    arma::rowvec fullResponses = join_rows(trainResponses, noiseResponses);
+
+    // Set weights.
+    arma::rowvec weights(trainData.n_cols + 100);
+    for (size_t i = 0; i < trainData.n_cols; ++i)
+      weights[i] = Random(0.9, 1.0);
+    for (size_t i = trainData.n_cols; i < trainData.n_cols + 100; ++i)
+      weights[i] = Random(0.0, 0.01); // Low weights for false points.
+
+    // Now build the decision tree.
+    DecisionTreeRegressor<> d(data, fullResponses, weights);
+
+    // Now we can check that we get good performance on the test set.
+    arma::rowvec predictions;
+    d.Predict(testData, predictions);
+
+    REQUIRE(predictions.n_elem == testData.n_cols);
+
+    // Figure out the rmse.
+    double rmse = RMSE(predictions, testResponses);
+    if (rmse < 6.2)
+    {
+      success = true;
+      break;
+    }
+  }
+  REQUIRE(success == true);
 }
 
 /**
@@ -1056,40 +1041,51 @@ TEST_CASE("WeightedDecisionTreeTest_", "
  */
 TEST_CASE("WeightedDecisionTreeMADGainTest", "[DecisionTreeRegressorTest]")
 {
-  // Loading data.
-  data::DatasetInfo info;
-  arma::mat trainData, testData;
-  arma::rowvec trainResponses, testResponses;
-  LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
-      info);
-
-  // Add some noise.
-  arma::mat noise(trainData.n_rows, 100, arma::fill::randu);
-  arma::rowvec noiseResponses(100);
-  for (size_t i = 0; i < noiseResponses.n_elem; ++i)
-    noiseResponses[i] = 15 + Random(0, 10); // Random response.
-
-  // Concatenate data matrices.
-  arma::mat data = arma::join_rows(trainData, noise);
-  arma::rowvec fullResponses = arma::join_rows(trainResponses, noiseResponses);
-
-  // Now set weights.
-  arma::rowvec weights(trainData.n_cols + 100);
-  for (size_t i = 0; i < trainData.n_cols; ++i)
-    weights[i] = Random(0.9, 1.0);
-  for (size_t i = trainData.n_cols; i < trainData.n_cols + 100; ++i)
-    weights[i] = Random(0.0, 0.01); // Low weights for false points.
-
-  // Now build the decision tree using MADGain.
-  DecisionTreeRegressor<MADGain> d(data, fullResponses, weights);
-
-  // Now we can check that we get good performance on the test set.
-  arma::rowvec predictions;
-  d.Predict(testData, predictions);
-
-  REQUIRE(predictions.n_elem == testData.n_cols);
+  // Allow multiple trials, if needed.
+  bool success = false;
+  for (size_t trial = 0; trial < 5; ++trial)
+  {
+    // Loading data.
+    data::DatasetInfo info;
+    arma::mat trainData, testData;
+    arma::rowvec trainResponses, testResponses;
+    LoadBostonHousingDataset(trainData, testData, trainResponses, testResponses,
+        info);
+
+    // Add some noise.
+    arma::mat noise(trainData.n_rows, 100, arma::fill::randu);
+    arma::rowvec noiseResponses(100);
+    for (size_t i = 0; i < noiseResponses.n_elem; ++i)
+      noiseResponses[i] = 15 + Random(0, 10); // Random response.
+
+    // Concatenate data matrices.
+    arma::mat data = join_rows(trainData, noise);
+    arma::rowvec fullResponses = join_rows(trainResponses, noiseResponses);
+
+    // Now set weights.
+    arma::rowvec weights(trainData.n_cols + 100);
+    for (size_t i = 0; i < trainData.n_cols; ++i)
+      weights[i] = Random(0.9, 1.0);
+    for (size_t i = trainData.n_cols; i < trainData.n_cols + 100; ++i)
+      weights[i] = Random(0.0, 0.01); // Low weights for false points.
+
+    // Now build the decision tree using MADGain.
+    DecisionTreeRegressor<MADGain> d(data, fullResponses, weights);
+
+    // Now we can check that we get good performance on the test set.
+    arma::rowvec predictions;
+    d.Predict(testData, predictions);
+
+    REQUIRE(predictions.n_elem == testData.n_cols);
+
+    // Figure out the rmse.
+    double rmse = RMSE(predictions, testResponses);
+    if (rmse < 6.6)
+    {
+      success = true;
+      break;
+    }
+  }
 
-  // Figure out the rmse.
-  double rmse = RMSE(predictions, testResponses);
-  REQUIRE(rmse < 6.5);
+  REQUIRE(success == true);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/decision_tree_test.cpp 4.4.0-1/src/mlpack/tests/decision_tree_test.cpp
--- 4.3.0-2/src/mlpack/tests/decision_tree_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/decision_tree_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -946,8 +946,8 @@ TEST_CASE("WeightedDecisionTreeTest", "[
     noiseLabels[i] = RandInt(3); // Random label.
 
   // Concatenate data matrices.
-  arma::mat data = arma::join_rows(dataset, noise);
-  arma::Row<size_t> fullLabels = arma::join_rows(labels, noiseLabels);
+  arma::mat data = join_rows(dataset, noise);
+  arma::Row<size_t> fullLabels = join_rows(labels, noiseLabels);
 
   // Now set weights.
   arma::rowvec weights(dataset.n_cols + 1000);
@@ -1017,8 +1017,8 @@ TEST_CASE("CategoricalWeightedBuildTest"
   for (size_t i = 2000; i < 4000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(trainingData, randomNoise);
-  arma::Row<size_t> fullLabels = arma::join_rows(trainingLabels, randomLabels);
+  arma::mat fullData = join_rows(trainingData, randomNoise);
+  arma::Row<size_t> fullLabels = join_rows(trainingLabels, randomLabels);
 
   // Build the tree.
   DecisionTree<> tree(fullData, di, fullLabels, 5, weights, 10);
@@ -1059,8 +1059,8 @@ TEST_CASE("WeightedDecisionTreeInformati
     noiseLabels[i] = RandInt(3); // Random label.
 
   // Concatenate data matrices.
-  arma::mat data = arma::join_rows(dataset, noise);
-  arma::Row<size_t> fullLabels = arma::join_rows(labels, noiseLabels);
+  arma::mat data = join_rows(dataset, noise);
+  arma::Row<size_t> fullLabels = join_rows(labels, noiseLabels);
 
   // Now set weights.
   arma::rowvec weights(dataset.n_cols + 1000);
@@ -1130,8 +1130,8 @@ TEST_CASE("CategoricalInformationGainWei
   for (size_t i = 2000; i < 4000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(trainingData, randomNoise);
-  arma::Row<size_t> fullLabels = arma::join_rows(trainingLabels, randomLabels);
+  arma::mat fullData = join_rows(trainingData, randomNoise);
+  arma::Row<size_t> fullLabels = join_rows(trainingLabels, randomLabels);
 
   // Build the tree.
   DecisionTree<InformationGain> tree(fullData, di, fullLabels, 5, weights, 10);
diff -pruN 4.3.0-2/src/mlpack/tests/det_test.cpp 4.4.0-1/src/mlpack/tests/det_test.cpp
--- 4.3.0-2/src/mlpack/tests/det_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/det_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,10 +30,46 @@
 using namespace mlpack;
 using namespace std;
 
+template<typename MatType>
+class TestDTree : public DTree<MatType>
+{
+ public:
+  TestDTree() : DTree<MatType>() { }
+
+  TestDTree(const typename DTree<MatType>::StatType& maxVals,
+            const typename DTree<MatType>::StatType& minVals,
+            const size_t totalPoints) :
+      DTree<MatType>(maxVals, minVals, totalPoints) { }
+
+  TestDTree(MatType& data) : DTree<MatType>(data) { }
+
+  size_t& Start() { return DTree<MatType>::start; }
+  size_t& End() { return DTree<MatType>::end; }
+  double& LogVolume() { return DTree<MatType>::logVolume; }
+
+  bool FindSplit(const MatType& data,
+                 size_t& splitDim,
+                 typename MatType::elem_type& splitValue,
+                 double& leftError,
+                 double& rightError,
+                 const size_t minLeafSize = 5) const
+  {
+    return DTree<MatType>::FindSplit(data, splitDim, splitValue, leftError,
+        rightError, minLeafSize);
+  }
+
+  size_t SplitData(MatType& data,
+                   const size_t splitDim,
+                   const typename MatType::elem_type splitValue,
+                   arma::Col<size_t>& oldFromNew) const
+  {
+    return DTree<MatType>::SplitData(data, splitDim, splitValue, oldFromNew);
+  }
+};
+
 // Tests for the private functions.  We cannot perform these if we are on
 // Windows because we cannot make private functions accessible using the macro
 // trick above.
-#ifndef _WIN32
 TEST_CASE("TestGetMaxMinVals", "[DETTest]")
 {
   arma::mat testData(3, 5);
@@ -57,14 +93,14 @@ TEST_CASE("TestComputeNodeError", "[DETT
   arma::vec maxVals("7 7 8");
   arma::vec minVals("3 0 1");
 
-  DTree<arma::mat> testDTree(maxVals, minVals, 5);
+  TestDTree<arma::mat> testDTree(maxVals, minVals, 5);
   double trueNodeError = -log(4.0) - log(7.0) - log(7.0);
 
-  REQUIRE((double) testDTree.logNegError ==
+  REQUIRE((double) testDTree.LogNegError() ==
       Approx(trueNodeError).epsilon(1e-12));
 
-  testDTree.start = 3;
-  testDTree.end = 5;
+  testDTree.Start() = 3;
+  testDTree.End() = 5;
 
   double nodeError = testDTree.LogNegativeError(5);
   trueNodeError = 2 * log(2.0 / 5.0) - log(4.0) - log(7.0) - log(7.0);
@@ -96,7 +132,7 @@ TEST_CASE("TestFindSplit", "[DETTest]")
                { 5, 0, 1, 7, 1 },
                { 5, 6, 7, 1, 8 } };
 
-  DTree<arma::mat> testDTree(testData);
+  TestDTree<arma::mat> testDTree(testData);
 
   size_t obDim;
   double obLeftError, obRightError, obSplit;
@@ -107,7 +143,7 @@ TEST_CASE("TestFindSplit", "[DETTest]")
   double trueRightError = 2 * log(3.0 / 5.0) - (log(7.0) + log(4.0) +
       log(2.5));
 
-  testDTree.logVolume = log(7.0) + log(4.0) + log(7.0);
+  testDTree.LogVolume() = log(7.0) + log(4.0) + log(7.0);
   REQUIRE(testDTree.FindSplit(
       testData, obDim, obSplit, obLeftError, obRightError, 1));
 
@@ -126,7 +162,7 @@ TEST_CASE("TestSplitData", "[DETTest]")
                { 5, 0, 1, 7, 1 },
                { 5, 6, 7, 1, 8 } };
 
-  DTree<arma::mat> testDTree(testData);
+  TestDTree<arma::mat> testDTree(testData);
 
   arma::Col<size_t> oTest(5);
   oTest = { 1, 2, 3, 4, 5 };
@@ -157,7 +193,7 @@ TEST_CASE("TestSparseFindSplit", "[DETTe
 
   arma::sp_mat testData(realData);
 
-  DTree<arma::sp_mat> testDTree(testData);
+  TestDTree<arma::sp_mat> testDTree(testData);
 
   size_t obDim;
   double obLeftError, obRightError, obSplit;
@@ -169,7 +205,7 @@ TEST_CASE("TestSparseFindSplit", "[DETTe
   double trueRightError = 2 * log(4.0 / 7.0) -
       (log(7.0) + log(6.5) + log(8.0) + log(6.0));
 
-  testDTree.logVolume = log(7.0) + log(7.0) + log(8.0) + log(6.0);
+  testDTree.LogVolume() = log(7.0) + log(7.0) + log(8.0) + log(6.0);
   REQUIRE(testDTree.FindSplit(
       testData, obDim, obSplit, obLeftError, obRightError, 1));
 
@@ -191,7 +227,7 @@ TEST_CASE("TestSparseSplitData", "[DETTe
 
   arma::sp_mat testData(realData);
 
-  DTree<arma::sp_mat> testDTree(testData);
+  TestDTree<arma::sp_mat> testDTree(testData);
 
   arma::Col<size_t> oTest(7);
   oTest = { 1, 2, 3, 4, 5, 6, 7 };
@@ -213,8 +249,6 @@ TEST_CASE("TestSparseSplitData", "[DETTe
   REQUIRE(oTest[6] == 7);
 }
 
-#endif
-
 // Tests for the public functions.
 
 TEST_CASE("TestGrow", "[DETTest]")
@@ -264,12 +298,12 @@ TEST_CASE("TestGrow", "[DETTest]")
 
   // Test node errors for every node (these are private functions).
 #ifndef _WIN32
-  REQUIRE(testDTree.logNegError == Approx(rootError).epsilon(1e-12));
-  REQUIRE(testDTree.Left()->logNegError == Approx(lError).epsilon(1e-12));
-  REQUIRE(testDTree.Right()->logNegError == Approx(rError).epsilon(1e-12));
-  REQUIRE(testDTree.Right()->Left()->logNegError ==
+  REQUIRE(testDTree.LogNegError() == Approx(rootError).epsilon(1e-12));
+  REQUIRE(testDTree.Left()->LogNegError() == Approx(lError).epsilon(1e-12));
+  REQUIRE(testDTree.Right()->LogNegError() == Approx(rError).epsilon(1e-12));
+  REQUIRE(testDTree.Right()->Left()->LogNegError() ==
       Approx(rlError).epsilon(1e-12));
-  REQUIRE(testDTree.Right()->Right()->logNegError ==
+  REQUIRE(testDTree.Right()->Right()->LogNegError() ==
       Approx(rrError).epsilon(1e-12));
 #endif
 
diff -pruN 4.3.0-2/src/mlpack/tests/distribution_test.cpp 4.4.0-1/src/mlpack/tests/distribution_test.cpp
--- 4.3.0-2/src/mlpack/tests/distribution_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/distribution_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -917,9 +917,9 @@ TEST_CASE("GammaDistributionTrainStatist
 
   // Train object d2 with the data's statistics.
   GammaDistribution d2;
-  const arma::vec meanLogx = arma::mean(arma::log(data), 1);
+  const arma::vec meanLogx = arma::mean(log(data), 1);
   const arma::vec meanx = arma::mean(data, 1);
-  const arma::vec logMeanx = arma::log(meanx);
+  const arma::vec logMeanx = log(meanx);
   d2.Train(logMeanx, meanLogx, meanx);
 
   REQUIRE(d1.Alpha(0) == Approx(d2.Alpha(0)).epsilon(1e-7));
@@ -1457,7 +1457,7 @@ TEST_CASE("DiagonalGaussianDistributionT
   arma::mat observations(4, 10000);
 
   for (size_t i = 0; i < 10000; ++i)
-    observations.col(i) = (arma::sqrt(cov) % arma::randn<arma::vec>(4)) + mean;
+    observations.col(i) = (sqrt(cov) % arma::randn<arma::vec>(4)) + mean;
 
   DiagonalGaussianDistribution d;
 
@@ -1522,7 +1522,7 @@ TEST_CASE("DiagonalGaussianWeightedParam
   arma::vec probs("0.2 0.2 0.2 0.2 0.2");
 
   for (size_t i = 0; i < 5; ++i)
-    obs.col(i) = (arma::sqrt(cov) % arma::randn<arma::vec>(4)) + mean;
+    obs.col(i) = (sqrt(cov) % arma::randn<arma::vec>(4)) + mean;
 
   DiagonalGaussianDistribution d1;
   DiagonalGaussianDistribution d2;
diff -pruN 4.3.0-2/src/mlpack/tests/fastmks_test.cpp 4.4.0-1/src/mlpack/tests/fastmks_test.cpp
--- 4.3.0-2/src/mlpack/tests/fastmks_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/fastmks_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -592,9 +592,9 @@ TEST_CASE("FastMKSModelCosineTest", "[Fa
 
   FastMKS<CosineDistance> f(referenceData, ck);
 
-  FastMKSModel m(FastMKSModel::COSINE_DISTANCE);
-  FastMKSModel mNaive(FastMKSModel::COSINE_DISTANCE);
-  FastMKSModel mSingle(FastMKSModel::COSINE_DISTANCE);
+  FastMKSModel m(FastMKSModel::COSINE_SIMILARITY);
+  FastMKSModel mNaive(FastMKSModel::COSINE_SIMILARITY);
+  FastMKSModel mSingle(FastMKSModel::COSINE_SIMILARITY);
   util::Timers timers;
 
   m.BuildModel(timers, std::move(referenceCopy1), ck, false, false, 2.0);
diff -pruN 4.3.0-2/src/mlpack/tests/gmm_test.cpp 4.4.0-1/src/mlpack/tests/gmm_test.cpp
--- 4.3.0-2/src/mlpack/tests/gmm_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/gmm_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -167,7 +167,7 @@ TEST_CASE("GMMTrainEMMultipleGaussians",
     counts += 1;
 
     // Account for rounding errors (possibly necessary).
-    counts[gaussians - 1] += (data.n_cols - arma::accu(counts));
+    counts[gaussians - 1] += (data.n_cols - accu(counts));
 
     // Build each Gaussian individually.
     size_t point = 0;
@@ -497,7 +497,7 @@ TEST_CASE("GMMClassifyTest", "[GMMTest]"
   gmm.Component(2) = GaussianDistribution("-2 -2", "2.2 1.4; 1.4 5.1");
   gmm.Weights() = "0.6 0.25 0.15";
 
-  arma::mat observations = arma::trans(arma::mat(
+  arma::mat observations = trans(arma::mat(
     " 0  0;"
     " 0  1;"
     " 0  2;"
@@ -702,7 +702,7 @@ TEST_CASE("UseExistingModelTest", "[GMMT
   counts += 1;
 
   // Account for rounding errors (possibly necessary).
-  counts[gaussians - 1] += (data.n_cols - arma::accu(counts));
+  counts[gaussians - 1] += (data.n_cols - accu(counts));
 
   // Build each Gaussian individually.
   size_t point = 0;
diff -pruN 4.3.0-2/src/mlpack/tests/hoeffding_tree_test.cpp 4.4.0-1/src/mlpack/tests/hoeffding_tree_test.cpp
--- 4.3.0-2/src/mlpack/tests/hoeffding_tree_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/hoeffding_tree_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -1485,7 +1485,7 @@ TEST_CASE("HoeffdingTreeEmptyConstructor
   HoeffdingTree<> ht;
 
   // Just ensure that we can train without throwing an exception.
-  REQUIRE_NOTHROW(ht.Train(data, labels));
+  REQUIRE_NOTHROW(ht.Train(data, labels, 2));
 
   // Now, create a categorical dataset and retrain.
   arma::mat data2 = arma::mat(4, 3000);
@@ -1514,9 +1514,324 @@ TEST_CASE("HoeffdingTreeEmptyConstructor
   }
 
   // Ensure we can train without throwing an exception.
-  REQUIRE_NOTHROW(ht.Train(data2, info, labels2));
+  REQUIRE_NOTHROW(ht.Train(data2, info, labels2, 3));
 
   // Train while specifying the number of classes.
-  REQUIRE_NOTHROW(ht.Train(data, labels, false, true, 2));
-  REQUIRE_NOTHROW(ht.Train(data2, info, labels2, false, 3));
+  REQUIRE_NOTHROW(ht.Train(data, labels, 2, false, true));
+  REQUIRE_NOTHROW(ht.Train(data2, info, labels2, 3, false));
+}
+
+// Test all numeric Train() variants.
+TEST_CASE("HoeffdingTreeNumericTrainVariantTest", "[HoeffdingTreeTest]")
+{
+  // Generate data.
+  arma::mat data(5, 1000, arma::fill::randu);
+  // Generate labels.
+  arma::Row<size_t> labels(1000);
+  for (size_t i = 0; i < 500; ++i)
+    labels[i] = 0;
+  for (size_t i = 500; i < 1000; ++i)
+    labels[i] = 1;
+
+  // Create trees.
+  HoeffdingTree<> htEmpty, ht1(5, 2), ht2(5, 2), ht3(5, 2), ht4(5, 2),
+      ht5(5, 2), ht6(5, 2), ht7(5, 2);
+
+  htEmpty.Train(data, labels, 2);
+  ht1.Train(data, labels);
+  ht2.Train(data, labels, 2);
+  ht3.Train(data, labels, 2, true);
+  ht4.Train(data, labels, 2, false, 0.96);
+  ht5.Train(data, labels, 2, false, 0.97, 150);
+  ht6.Train(data, labels, 2, false, 0.98, 160, 110);
+  ht7.Train(data, labels, 2, false, 0.99, 170, 120, 110);
+
+  REQUIRE(htEmpty.NumClasses() == 2);
+  REQUIRE(htEmpty.NumSamples() == data.n_cols);
+
+  REQUIRE(ht1.NumClasses() == 2);
+  REQUIRE(ht1.NumSamples() == data.n_cols);
+
+  REQUIRE(ht2.NumClasses() == 2);
+  REQUIRE(ht2.NumSamples() == data.n_cols);
+
+  REQUIRE(ht3.NumClasses() == 2);
+  REQUIRE(ht3.NumSamples() == data.n_cols);
+
+  REQUIRE(ht4.NumClasses() == 2);
+  REQUIRE(ht4.NumSamples() > 0);
+  REQUIRE(ht4.SuccessProbability() == 0.96);
+
+  REQUIRE(ht5.NumClasses() == 2);
+  REQUIRE(ht5.NumSamples() > 0);
+  REQUIRE(ht5.SuccessProbability() == 0.97);
+  REQUIRE(ht5.MaxSamples() == 150);
+
+  REQUIRE(ht6.NumClasses() == 2);
+  REQUIRE(ht6.NumSamples() > 0);
+  REQUIRE(ht6.SuccessProbability() == 0.98);
+  REQUIRE(ht6.MaxSamples() == 160);
+  REQUIRE(ht6.CheckInterval() == 110);
+
+  REQUIRE(ht7.NumClasses() == 2);
+  REQUIRE(ht7.NumSamples() > 0);
+  REQUIRE(ht7.SuccessProbability() == 0.99);
+  REQUIRE(ht7.MaxSamples() == 170);
+  REQUIRE(ht7.CheckInterval() == 120);
+  REQUIRE(ht7.MinSamples() == 110);
+}
+
+// Test all categorical Train() variants.
+TEST_CASE("HoeffdingTreeCategoricalTrainVariantTest", "[HoeffdingTreeTest]")
+{
+  // Generate data.
+  arma::mat data(4, 9000);
+  arma::Row<size_t> labels(9000);
+  data::DatasetInfo info(4); // All features are numeric, except the fourth.
+  info.MapString<double>("0", 3);
+  for (size_t i = 0; i < 9000; i += 3)
+  {
+    data(0, i) = Random();
+    data(1, i) = Random();
+    data(2, i) = Random();
+    data(3, i) = 0.0;
+    labels[i] = 0;
+
+    data(0, i + 1) = Random();
+    data(1, i + 1) = Random() - 1.0;
+    data(2, i + 1) = Random() + 0.5;
+    data(3, i + 1) = 0.0;
+    labels[i + 1] = 2;
+
+    data(0, i + 2) = Random();
+    data(1, i + 2) = Random() + 1.0;
+    data(2, i + 2) = Random() + 0.8;
+    data(3, i + 2) = 0.0;
+    labels[i + 2] = 1;
+  }
+
+  // Create trees.
+  HoeffdingTree<> htEmpty, ht1(info, 3), ht2(info, 3), ht3(info, 3),
+      ht4(info, 3), ht5(info, 3), ht6(info, 3), ht7(info, 3);
+
+  htEmpty.Train(data, info, labels, 3);
+  ht1.Train(data, info, labels);
+  ht2.Train(data, info, labels, 3);
+  ht3.Train(data, info, labels, 3, true);
+  ht4.Train(data, info, labels, 3, false, 0.96);
+  ht5.Train(data, info, labels, 3, false, 0.97, 150);
+  ht6.Train(data, info, labels, 3, false, 0.98, 160, 110);
+  ht7.Train(data, info, labels, 3, false, 0.99, 170, 120, 110);
+
+  REQUIRE(htEmpty.NumClasses() == 3);
+  REQUIRE(htEmpty.NumSamples() == data.n_cols);
+
+  REQUIRE(ht1.NumClasses() == 3);
+  REQUIRE(ht1.NumSamples() == data.n_cols);
+
+  REQUIRE(ht2.NumClasses() == 3);
+  REQUIRE(ht2.NumSamples() == data.n_cols);
+
+  REQUIRE(ht3.NumClasses() == 3);
+  REQUIRE(ht3.NumSamples() == data.n_cols);
+
+  REQUIRE(ht4.NumClasses() == 3);
+  REQUIRE(ht4.NumSamples() > 0);
+  REQUIRE(ht4.SuccessProbability() == 0.96);
+
+  REQUIRE(ht5.NumClasses() == 3);
+  REQUIRE(ht5.NumSamples() > 0);
+  REQUIRE(ht5.SuccessProbability() == 0.97);
+  REQUIRE(ht5.MaxSamples() == 150);
+
+  REQUIRE(ht6.NumClasses() == 3);
+  REQUIRE(ht6.NumSamples() > 0);
+  REQUIRE(ht6.SuccessProbability() == 0.98);
+  REQUIRE(ht6.MaxSamples() == 160);
+  REQUIRE(ht6.CheckInterval() == 110);
+
+  REQUIRE(ht7.NumClasses() == 3);
+  REQUIRE(ht7.NumSamples() > 0);
+  REQUIRE(ht7.SuccessProbability() == 0.99);
+  REQUIRE(ht7.MaxSamples() == 170);
+  REQUIRE(ht7.CheckInterval() == 120);
+  REQUIRE(ht7.MinSamples() == 110);
+}
+
+// Test overloads of Reset().  (Also test NumSamples().)
+TEST_CASE("HoeffdingTreeResetTests", "[HoeffdingTreeTest]")
+{
+  // Generate data.
+  arma::mat data(5, 1000, arma::fill::randu);
+  // Generate labels.
+  arma::Row<size_t> labels(1000);
+  for (size_t i = 0; i < 500; ++i)
+    labels[i] = 0;
+  for (size_t i = 500; i < 1000; ++i)
+    labels[i] = 1;
+
+  // Train the tree.
+  HoeffdingTree<> ht(data, labels, 2);
+
+  REQUIRE(ht.NumSamples() > 0);
+  REQUIRE(ht.NumClasses() == 2);
+
+  // Reset the tree, changing nothing.
+  ht.Reset();
+  REQUIRE(ht.NumSamples() == 0);
+  REQUIRE(ht.NumChildren() == 0);
+
+  ht.Train(data, labels);
+  REQUIRE(ht.NumSamples() > 0);
+  REQUIRE(ht.NumClasses() == 2);
+
+  // Reset the tree, changing the dimensionality and number of classes.
+  ht.Reset(10, 3);
+  REQUIRE(ht.NumSamples() == 0);
+  REQUIRE(ht.NumChildren() == 0);
+
+  data = arma::randu<arma::mat>(10, 1000);
+  for(size_t i = 750; i < 1000; ++i)
+    labels[i] = 2;
+
+  ht.Train(data, labels);
+  REQUIRE(ht.NumSamples() > 0);
+  REQUIRE(ht.NumClasses() == 3);
+
+  // Reset the tree to work on categorical data with a different number of
+  // classes.
+  data::DatasetInfo info(10);
+  info.MapString<double>("0", 3);
+  data.row(9).fill(0.0);
+  for (size_t i = 0; i < 250; ++i)
+    labels[i] = 3;
+
+  ht.Reset(info, 4);
+  REQUIRE(ht.NumSamples() == 0);
+  REQUIRE(ht.NumChildren() == 0);
+
+  ht.Train(data, info, labels);
+
+  REQUIRE(ht.NumSamples() > 0);
+  REQUIRE(ht.NumClasses() == 4);
+}
+
+// Test that we can learn on floating-point data.
+TEST_CASE("HoeffdingTreeNumericFloatDataTest", "[HoeffdingTreeTest]")
+{
+  // We need to create a dataset with some amount of complexity, that must be
+  // split in a handful of ways to accurately classify the data.  An expanding
+  // spiral should do the trick here.  We'll make the spiral in two dimensions.
+  // The label will change as the index increases.
+  arma::fmat spiralDataset(2, 10000);
+  for (size_t i = 0; i < 10000; ++i)
+  {
+    // One circle every 20000 samples.  Plus some noise.
+    const float magnitude = 2.0 + (float(i) / 20000.0) + 0.5 * Random();
+    const float angle = (i % 20000) * (2 * M_PI) + Random();
+
+    const float x = magnitude * cos(angle);
+    const float y = magnitude * sin(angle);
+
+    spiralDataset(0, i) = x;
+    spiralDataset(1, i) = y;
+  }
+
+  arma::Row<size_t> labels(10000);
+  for (size_t i = 0; i < 2000; ++i)
+    labels[i] = 1;
+  for (size_t i = 2000; i < 4000; ++i)
+    labels[i] = 3;
+  for (size_t i = 4000; i < 6000; ++i)
+    labels[i] = 2;
+  for (size_t i = 6000; i < 8000; ++i)
+    labels[i] = 0;
+  for (size_t i = 8000; i < 10000; ++i)
+    labels[i] = 4;
+
+  // Now shuffle the dataset.
+  arma::uvec indices = arma::shuffle(arma::linspace<arma::uvec>(0, 9999,
+      10000));
+  arma::fmat d(2, 10000);
+  arma::Row<size_t> l(10000);
+  for (size_t i = 0; i < 10000; ++i)
+  {
+    d.col(i) = spiralDataset.col(indices[i]);
+    l[i] = labels[indices[i]];
+  }
+
+  // Split into a training set and a test set.
+  arma::fmat trainingData = d.cols(0, 4999);
+  arma::fmat testData = d.cols(5000, 9999);
+  arma::Row<size_t> trainingLabels = l.subvec(0, 4999);
+  arma::Row<size_t> testLabels = l.subvec(5000, 9999);
+
+  data::DatasetInfo info(2);
+
+  // Now build two decision trees; one in batch mode, and one in streaming mode.
+  // We need to set the confidence pretty high so that the streaming tree isn't
+  // able to have enough samples to build to the same leaves.
+  HoeffdingTree<GiniImpurity, HoeffdingFloatNumericSplit>
+      batchTree(trainingData, info, trainingLabels, 5, true, 0.99999999);
+  HoeffdingTree<GiniImpurity, HoeffdingFloatNumericSplit>
+      streamTree(trainingData, info, trainingLabels, 5, false, 0.99999999);
+
+  // Ensure that the performance of the batch tree is better.
+  size_t batchCorrect = 0;
+  size_t streamCorrect = 0;
+  for (size_t i = 0; i < 5000; ++i)
+  {
+    size_t streamLabel = streamTree.Classify(testData.col(i));
+    size_t batchLabel = batchTree.Classify(testData.col(i));
+
+    if (streamLabel == testLabels[i])
+      ++streamCorrect;
+    if (batchLabel == testLabels[i])
+      ++batchCorrect;
+  }
+
+  // The batch tree must be a bit better than the stream tree.  But not too
+  // much, since the accuracy is already going to be very high.
+  REQUIRE(batchCorrect >= streamCorrect);
+}
+
+// Test that we can learn on categorical floating-point data.
+TEST_CASE("HoeffdingTreeCategoricalFloatDataTest", "[HoeffdingTreeTest]")
+{
+  // Generate data.
+  arma::fmat dataset(4, 3000, arma::fill::randu);
+  arma::Row<size_t> labels(3000);
+  for (size_t i = 0; i < 3000; i += 3)
+  {
+    labels[i] = 0;
+
+    labels[i + 1] = 2;
+    dataset.col(i + 1) += 2.0;
+
+    labels[i + 2] = 1;
+    dataset.col(i + 2) -= 2.0;
+  }
+
+  data::DatasetInfo info(4); // All features are numeric, except the fourth.
+  info.MapString<double>("0", 3);
+  dataset.row(3).fill(0.0);
+
+  // Now build two decision trees; one in batch mode, and one in streaming mode.
+  HoeffdingTree<GiniImpurity, HoeffdingFloatNumericSplit>
+      batchTree(dataset, info, labels, 3, true);
+  HoeffdingTree<GiniImpurity, HoeffdingFloatNumericSplit>
+      streamTree(dataset, info, labels, 3, false);
+
+  // Make sure that we trained successfully.
+  arma::Row<size_t> batchPredictions, streamPredictions;
+  batchTree.Classify(dataset, batchPredictions);
+  streamTree.Classify(dataset, streamPredictions);
+
+  REQUIRE(batchPredictions.n_elem == labels.n_elem);
+  REQUIRE(streamPredictions.n_elem == labels.n_elem);
+
+  // Make sure that the accuracy is at least reasonable.  (This is a very loose
+  // test.)
+  REQUIRE(accu(batchPredictions == labels) > (labels.n_elem / 2));
+  REQUIRE(accu(streamPredictions == labels) > (labels.n_elem / 2));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/hpt_test.cpp 4.4.0-1/src/mlpack/tests/hpt_test.cpp
--- 4.3.0-2/src/mlpack/tests/hpt_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/hpt_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -29,7 +29,7 @@ TEST_CASE("CVFunctionTest", "[HPTTest]")
   arma::vec beta = arma::randn(5, 1);
   arma::rowvec ys = beta.t() * xs + 0.1 * arma::randn(1, 100);
 
-  SimpleCV<LARS, MSE> cv(0.2, xs, ys);
+  SimpleCV<LARS<>, MSE> cv(0.2, xs, ys);
 
   bool transposeData = true;
   bool useCholesky = false;
@@ -41,7 +41,7 @@ TEST_CASE("CVFunctionTest", "[HPTTest]")
 
   FixedArg<bool, 1> fixedUseCholesky{useCholesky};
   FixedArg<double, 3> fixedLambda1{lambda2};
-  CVFunction<decltype(cv), LARS, 4, FixedArg<bool, 1>, FixedArg<double, 3>>
+  CVFunction<decltype(cv), LARS<>, 4, FixedArg<bool, 1>, FixedArg<double, 3>>
       cvFun(cv, datasetInfo, 0.0, 0.0, fixedUseCholesky, fixedLambda1);
 
   double expected = cv.Evaluate(transposeData, useCholesky, lambda1, lambda2);
@@ -64,7 +64,7 @@ TEST_CASE("CVFunctionCategoricalTest", "
   arma::vec beta = arma::randn(5, 1);
   arma::rowvec ys = beta.t() * xs + 0.1 * arma::randn(1, 100);
 
-  SimpleCV<LARS, MSE> cv(0.2, xs, ys);
+  SimpleCV<LARS<>, MSE> cv(0.2, xs, ys);
 
   bool transposeData = true;
   bool useCholesky = false;
@@ -78,7 +78,7 @@ TEST_CASE("CVFunctionCategoricalTest", "
 
   FixedArg<bool, 1> fixedUseCholesky{useCholesky};
   FixedArg<double, 3> fixedLambda1{lambda2};
-  CVFunction<decltype(cv), LARS, 4, FixedArg<bool, 1>, FixedArg<double, 3>>
+  CVFunction<decltype(cv), LARS<>, 4, FixedArg<bool, 1>, FixedArg<double, 3>>
       cvFun(cv, datasetInfo, 0.0, 0.0, fixedUseCholesky, fixedLambda1);
 
   double expected = cv.Evaluate(transposeData, useCholesky, lambda1, lambda2);
@@ -99,10 +99,10 @@ template<typename MLAlgorithm,
          typename MatType = void,
          typename PredictionsType = void,
          typename WeightsType = void>
-class QuadraticFunction
+class QuadraticTestFunction
 {
  public:
-  QuadraticFunction(double a,
+  QuadraticTestFunction(double a,
                     double b,
                     double c,
                     double d,
@@ -137,7 +137,7 @@ TEST_CASE("CVFunctionGradientTest", "[HP
   double b = -1.5;
   double c = 2.5;
   double d = 3.0;
-  QuadraticFunction<LARS> lf(a, b, c, d);
+  QuadraticTestFunction<LARS<>> lf(a, b, c, d);
 
   // All values are numeric.
   IncrementPolicy policy(true);
@@ -145,7 +145,7 @@ TEST_CASE("CVFunctionGradientTest", "[HP
 
   double relativeDelta = 0.01;
   double minDelta = 0.001;
-  CVFunction<decltype(lf), LARS, 3> cvFun(lf, datasetInfo, relativeDelta,
+  CVFunction<decltype(lf), LARS<>, 3> cvFun(lf, datasetInfo, relativeDelta,
       minDelta);
 
   double x = 0.0;
@@ -181,7 +181,7 @@ void InitProneToOverfittingData(arma::ma
   arma::rowvec data = arma::linspace<arma::rowvec>(0.0, 10.0, N);
   xs = data;
   for (size_t i = 2; i <= M; ++i)
-    xs = arma::join_cols(xs, arma::pow(data, i));
+    xs = join_cols(xs, pow(data, i));
 
   // Responses that approximately follow the function y = 2 * x. Adding noise to
   // avoid having a polynomial of degree 1 that exactly fits the points.
@@ -202,7 +202,7 @@ void FindLARSBestLambdas(arma::mat& xs,
                          double& bestLambda2,
                          double& bestObjective)
 {
-  SimpleCV<LARS, MSE> cv(validationSize, xs, ys);
+  SimpleCV<LARS<>, MSE> cv(validationSize, xs, ys);
 
   bestObjective = std::numeric_limits<double>::max();
 
@@ -250,8 +250,8 @@ TEST_CASE("GridSearchTest", "[HPTTest]")
   for (double lambda2 : lambda2Set)
     datasetInfo.MapString<size_t>(lambda2, 1);
 
-  SimpleCV<LARS, MSE> cv(validationSize, xs, ys);
-  CVFunction<decltype(cv), LARS, 4, FixedArg<bool, 0>, FixedArg<bool, 1>>
+  SimpleCV<LARS<>, MSE> cv(validationSize, xs, ys);
+  CVFunction<decltype(cv), LARS<>, 4, FixedArg<bool, 0>, FixedArg<bool, 1>>
       cvFun(cv, datasetInfo, 0.0, 0.0, {transposeData}, {useCholesky});
 
   ens::GridSearch optimizer;
@@ -297,7 +297,7 @@ TEST_CASE("HPTTest", "[HPTTest]")
       expectedObjective);
 
   double actualLambda1, actualLambda2;
-  HyperParameterTuner<LARS, MSE, SimpleCV, GridSearch>
+  HyperParameterTuner<LARS<>, MSE, SimpleCV, GridSearch>
       hpt(validationSize, xs, ys);
   std::tie(actualLambda1, actualLambda2) = hpt.Optimize(Fixed(transposeData),
       Fixed(useCholesky), lambda1Set, lambda2Set);
@@ -322,13 +322,13 @@ TEST_CASE("HPTMaximizationTest", "[HPTTe
 {
   // Initializing a linearly separable dataset.
   arma::mat xs = arma::linspace<arma::rowvec>(0.0, 10.0, 50);
-  arma::Row<size_t> ys = arma::join_rows(arma::zeros<arma::Row<size_t>>(25),
+  arma::Row<size_t> ys = join_rows(arma::zeros<arma::Row<size_t>>(25),
       arma::ones<arma::Row<size_t>>(25));
 
   // We will train and validate on the same dataset.
   double validationSize = 0.5;
-  arma::mat doubledXs = arma::join_rows(xs, xs);
-  arma::Row<size_t> doubledYs = arma::join_rows(ys, ys);
+  arma::mat doubledXs = join_rows(xs, xs);
+  arma::Row<size_t> doubledYs = join_rows(ys, ys);
 
   // Defining lambdas to choose from. Zero should be preferred since big lambdas
   // are likely to restrict capabilities of logistic regression.
@@ -354,7 +354,7 @@ TEST_CASE("HPTMaximizationTest", "[HPTTe
  */
 TEST_CASE("HPTGradientDescentTest", "[HPTTest]")
 {
-  // Constructor arguments for the fake CV function (QuadraticFunction).
+  // Constructor arguments for the fake CV function (QuadraticTestFunction).
   double a = 1.0;
   double b = -1.5;
   double c = 2.5;
@@ -366,9 +366,9 @@ TEST_CASE("HPTGradientDescentTest", "[HP
   double zMin = -2.0;
 
   // We pass LARS just because some ML algorithm should be passed. We pass MSE
-  // to tell HyperParameterTuner that the objective function (QuadraticFunction)
-  // should be minimized.
-  HyperParameterTuner<LARS, MSE, QuadraticFunction,
+  // to tell HyperParameterTuner that the objective function
+  // (QuadraticTestFunction) should be minimized.
+  HyperParameterTuner<LARS<>, MSE, QuadraticTestFunction,
       GradientDescent> hpt(a, b, c, d, xMin, yMin, zMin);
 
   // Setting GradientDescent to find more close solution to the optimal one.
diff -pruN 4.3.0-2/src/mlpack/tests/image_load_test.cpp 4.4.0-1/src/mlpack/tests/image_load_test.cpp
--- 4.3.0-2/src/mlpack/tests/image_load_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/image_load_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -115,7 +115,7 @@ TEST_CASE("SaveImageMatAPITest", "[Image
   arma::Mat<unsigned char> im1;
   size_t dimension = info.Width() * info.Height() * info.Channels();
   im1 = arma::randi<arma::Mat<unsigned char>>(dimension, 1);
-  arma::mat input = arma::conv_to<arma::mat>::from(im1);
+  arma::mat input = ConvTo<arma::mat>::From(im1);
   REQUIRE(Save("APITest.bmp", input, info, false) == true);
 
   arma::mat output;
diff -pruN 4.3.0-2/src/mlpack/tests/kernel_test.cpp 4.4.0-1/src/mlpack/tests/kernel_test.cpp
--- 4.3.0-2/src/mlpack/tests/kernel_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/kernel_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -203,38 +203,38 @@ TEST_CASE("MDFullCovarianceTest", "[Kern
 /**
  * Simple test case for the cosine distance.
  */
-TEST_CASE("CosineDistanceSameAngleTest", "[KernelTest]")
+TEST_CASE("CosineSimilaritySameAngleTest", "[KernelTest]")
 {
   arma::vec a = "1.0 2.0 3.0";
   arma::vec b = "2.0 4.0 6.0";
 
-  REQUIRE(CosineDistance::Evaluate(a, b) == Approx(1.0).epsilon(1e-7));
-  REQUIRE(CosineDistance::Evaluate(b, a) == Approx(1.0).epsilon(1e-7));
+  REQUIRE(CosineSimilarity::Evaluate(a, b) == Approx(1.0).epsilon(1e-7));
+  REQUIRE(CosineSimilarity::Evaluate(b, a) == Approx(1.0).epsilon(1e-7));
 }
 
 /**
  * Now let's have them be orthogonal.
  */
-TEST_CASE("CosineDistanceOrthogonalTest", "[KernelTest]")
+TEST_CASE("CosineSimilarityOrthogonalTest", "[KernelTest]")
 {
   arma::vec a = "0.0 1.0";
   arma::vec b = "1.0 0.0";
 
-  REQUIRE(CosineDistance::Evaluate(a, b) == Approx(0.0).margin(1e-5));
-  REQUIRE(CosineDistance::Evaluate(b, a) == Approx(0.0).margin(1e-5));
+  REQUIRE(CosineSimilarity::Evaluate(a, b) == Approx(0.0).margin(1e-5));
+  REQUIRE(CosineSimilarity::Evaluate(b, a) == Approx(0.0).margin(1e-5));
 }
 
 /**
  * Some random angle test.
  */
-TEST_CASE("CosineDistanceRandomTest", "[KernelTest]")
+TEST_CASE("CosineSimilarityRandomTest", "[KernelTest]")
 {
   arma::vec a = "0.1 0.2 0.3 0.4 0.5";
   arma::vec b = "1.2 1.0 0.8 -0.3 -0.5";
 
-  REQUIRE(CosineDistance::Evaluate(a, b) ==
+  REQUIRE(CosineSimilarity::Evaluate(a, b) ==
       Approx(0.1385349024).epsilon(1e-7));
-  REQUIRE(CosineDistance::Evaluate(b, a) ==
+  REQUIRE(CosineSimilarity::Evaluate(b, a) ==
       Approx(0.1385349024).epsilon(1e-7));
 }
 
@@ -286,13 +286,6 @@ TEST_CASE("GaussianKernelTest", "[Kernel
   REQUIRE(gk.Normalizer(2) == Approx(1.5707963267948963).epsilon(1e-7));
   REQUIRE(gk.Normalizer(3) == Approx(1.9687012432153019).epsilon(1e-7));
   REQUIRE(gk.Normalizer(4) == Approx(2.4674011002723386).epsilon(1e-7));
-  /* check the convolution integral */
-  REQUIRE(gk.ConvolutionIntegral(a, b) ==
-      Approx(0.024304474038457577).epsilon(1e-7));
-  REQUIRE(gk.ConvolutionIntegral(a, c) ==
-      Approx(0.024304474038457577).epsilon(1e-7));
-  REQUIRE(gk.ConvolutionIntegral(b, c) ==
-      Approx(0.024304474038457577).epsilon(1e-7));
 }
 
 TEST_CASE("GaussianKernelSerializationTest", "[KernelTest]")
@@ -329,11 +322,6 @@ TEST_CASE("SphericalKernelTest", "[Kerne
   REQUIRE(sk.Normalizer(2) == Approx(0.78539816339744828).epsilon(1e-7));
   REQUIRE(sk.Normalizer(3) == Approx(0.52359877559829893).epsilon(1e-7));
   REQUIRE(sk.Normalizer(4) == Approx(0.30842513753404244).epsilon(1e-7));
-  /* check the convolution integral */
-  REQUIRE(sk.ConvolutionIntegral(a, b) == Approx(0.0).epsilon(1e-7));
-  REQUIRE(sk.ConvolutionIntegral(a, c) == Approx(0.0).epsilon(1e-7));
-  REQUIRE(sk.ConvolutionIntegral(b, c) ==
-      Approx(1.0021155029652784).epsilon(1e-7));
 }
 
 TEST_CASE("EpanechnikovKernelTest", "[KernelTest]")
@@ -356,11 +344,6 @@ TEST_CASE("EpanechnikovKernelTest", "[Ke
   REQUIRE(ek.Normalizer(2) == Approx(0.39269908169872414).epsilon(1e-7));
   REQUIRE(ek.Normalizer(3) == Approx(0.20943951023931956).epsilon(1e-7));
   REQUIRE(ek.Normalizer(4) == Approx(0.10280837917801415).epsilon(1e-7));
-  /* check the convolution integral */
-  REQUIRE(ek.ConvolutionIntegral(a, b) == Approx(0.0).epsilon(1e-7));
-  REQUIRE(ek.ConvolutionIntegral(a, c) == Approx(0.0).epsilon(1e-7));
-  REQUIRE(ek.ConvolutionIntegral(b, c) ==
-      Approx(1.5263455690698258).epsilon(1e-7));
 }
 
 TEST_CASE("PolynomialKernelTest", "[KernelTest]")
diff -pruN 4.3.0-2/src/mlpack/tests/knn_test.cpp 4.4.0-1/src/mlpack/tests/knn_test.cpp
--- 4.3.0-2/src/mlpack/tests/knn_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/knn_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,6 +17,18 @@
 using namespace mlpack;
 
 /**
+ * A couple of handful declarations for float32 testing.
+ * These will be removed when we refactor the Bounds to accept MatType.
+ * For now, we will keep the following declarations.
+ */
+template<typename MetricType>
+using FloatHRectBound = HRectBound<MetricType, float>;
+
+template<typename MetricType, typename StatisticType, typename MatType>
+using FloatKDTree = BinarySpaceTree<MetricType, StatisticType, MatType,
+                                    FloatHRectBound, MidpointSplit>;
+
+/**
  * Test that Unmap() works in the dual-tree case (see unmap.hpp).
  */
 TEST_CASE("KNNDualTreeUnmapTest", "[KNNTest]")
@@ -747,6 +759,48 @@ TEST_CASE("KNNSingleTreeVsNaive", "[KNNT
 }
 
 /**
+ * Test the single-tree nearest-neighbors method with the naive method.
+ *
+ * The main difference with the above test is that this one loads the reference
+ * dataset as a float32, and the distances as a float32 as well.
+ *
+ * Errors are produced if the results are not identical.
+ */
+TEST_CASE("KNNSingleTreeVsNaiveF32", "[KNNTest]")
+{
+  arma::fmat dataset;
+
+  // Hard-coded filename: bad?
+  // Code duplication: also bad!
+  if (!data::Load("test_data_3_1000.csv", dataset))
+    FAIL("Cannot load test dataset test_data_3_1000.csv!");
+
+  NeighborSearch<NearestNeighborSort,
+                 EuclideanDistance,
+                 arma::fmat,
+                 FloatKDTree> knn(dataset, SINGLE_TREE_MODE);
+
+  // Set up computation for naive mode.
+  NeighborSearch<NearestNeighborSort,
+                 EuclideanDistance,
+                 arma::fmat,
+                 FloatKDTree> naive(dataset, NAIVE_MODE);
+
+  arma::Mat<size_t> neighborsTree;
+  arma::fmat distancesTree;
+  knn.Search(15, neighborsTree, distancesTree);
+
+  arma::Mat<size_t> neighborsNaive;
+  arma::fmat distancesNaive;
+  naive.Search(15, neighborsNaive, distancesNaive);
+
+  for (size_t i = 0; i < neighborsTree.n_elem; ++i)
+  {
+    REQUIRE(neighborsTree[i] ==neighborsNaive[i]);
+    REQUIRE(distancesTree[i] == Approx(distancesNaive[i]).epsilon(1e-7));
+  }
+}
+/**
  * Test the cover tree single-tree nearest-neighbors method against the naive
  * method.  This uses only a random reference dataset.
  *
@@ -1775,11 +1829,11 @@ TEST_CASE("KNNGreedyTreeSearch", "[KNNTe
 
   // Check that all neighbour values are between 0 and 100, as only 100 points
   // are present in dataset.
-  REQUIRE(arma::accu(neighbors < 0 || neighbors >= 100) ==0);
+  REQUIRE(accu(neighbors < 0 || neighbors >= 100) ==0);
 
   // Check that all distances values are between 0.0 and sqrt(3) as arma::randu
   // generates a uniform distribution in [0, 1].
-  REQUIRE(arma::accu(distances < 0.0 || distances > std::sqrt(3.0)) == 0);
+  REQUIRE(accu(distances < 0.0 || distances > std::sqrt(3.0)) == 0);
 }
 
 /**
@@ -1812,16 +1866,16 @@ TEST_CASE("KNNSpillTreeSearchEnoughResul
 
   // Check that all neighbor values are between 0 and 100, as only 100 points
   // are present in dataset.
-  REQUIRE(arma::accu(neighborsDual < 0 || neighborsDual >= 100) == 0);
-  REQUIRE(arma::accu(neighborsSingle < 0 || neighborsSingle >= 100) == 0);
-  REQUIRE(arma::accu(neighborsGreedy < 0 || neighborsGreedy >= 100) == 0);
+  REQUIRE(accu(neighborsDual < 0 || neighborsDual >= 100) == 0);
+  REQUIRE(accu(neighborsSingle < 0 || neighborsSingle >= 100) == 0);
+  REQUIRE(accu(neighborsGreedy < 0 || neighborsGreedy >= 100) == 0);
 
   // Check that all distances values are between 0.0 and sqrt(3) as arma::randu
   // generates a uniform distribution in [0, 1].
-  REQUIRE(arma::accu(distancesDual < 0.0 || distancesDual > std::sqrt(3.0))
+  REQUIRE(accu(distancesDual < 0.0 || distancesDual > std::sqrt(3.0))
       == 0);
-  REQUIRE(arma::accu(distancesSingle < 0.0 || distancesSingle > std::sqrt(3.0))
+  REQUIRE(accu(distancesSingle < 0.0 || distancesSingle > std::sqrt(3.0))
       == 0);
-  REQUIRE(arma::accu(distancesGreedy < 0.0 || distancesGreedy > std::sqrt(3.0))
+  REQUIRE(accu(distancesGreedy < 0.0 || distancesGreedy > std::sqrt(3.0))
       == 0);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/lars_test.cpp 4.4.0-1/src/mlpack/tests/lars_test.cpp
--- 4.3.0-2/src/mlpack/tests/lars_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/lars_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -17,24 +17,32 @@
 
 using namespace mlpack;
 
+template<typename MatType, typename ResponsesType>
 void GenerateProblem(
-    arma::mat& X, arma::rowvec& y, size_t nPoints, size_t nDims)
+    MatType& X, ResponsesType& y, size_t nPoints, size_t nDims)
 {
-  X = arma::randn(nDims, nPoints);
-  arma::vec beta = arma::randn(nDims, 1);
+  X = arma::randn<MatType>(nDims, nPoints);
+  arma::Col<typename ResponsesType::elem_type> beta =
+      arma::randn<arma::Col<typename ResponsesType::elem_type>>(nDims);
   y = beta.t() * X;
 }
 
-void LARSVerifyCorrectness(arma::vec beta, arma::vec errCorr, double lambda)
+template<typename VecType, typename ElemType>
+void LARSVerifyCorrectness(const VecType& beta,
+                           const VecType& errCorr,
+                           ElemType lambda)
 {
   size_t nDims = beta.n_elem;
-  const double tol = 1e-10;
+
+  // floats require a much larger tolerance.
+  const ElemType tol = (std::is_same<ElemType, double>::value) ? 1e-8 : 5e-3;
+
   for (size_t j = 0; j < nDims; ++j)
   {
     if (beta(j) == 0)
     {
       // Make sure that |errCorr(j)| <= lambda.
-      REQUIRE(std::max(fabs(errCorr(j)) - lambda, 0.0) ==
+      REQUIRE(std::max(fabs(errCorr(j)) - lambda, (ElemType) 0.0) ==
           Approx(0.0).margin(tol));
     }
     else if (beta(j) < 0)
@@ -50,30 +58,33 @@ void LARSVerifyCorrectness(arma::vec bet
   }
 }
 
+template<typename MatType>
 void LassoTest(size_t nPoints, size_t nDims, bool elasticNet, bool useCholesky,
                bool fitIntercept, bool normalizeData)
 {
-  arma::mat X;
-  arma::rowvec y;
+  typedef typename MatType::elem_type ElemType;
+
+  MatType X;
+  arma::Row<ElemType> y;
 
   for (size_t i = 0; i < 100; ++i)
   {
     GenerateProblem(X, y, nPoints, nDims);
 
     // Armadillo's median is broken, so...
-    arma::vec sortedAbsCorr = sort(abs(X * y.t()));
-    double lambda1 = sortedAbsCorr(nDims / 2);
-    double lambda2;
+    arma::Col<ElemType> sortedAbsCorr = sort(abs(X * y.t()));
+    ElemType lambda1 = sortedAbsCorr(nDims / 2);
+    ElemType lambda2;
     if (elasticNet)
       lambda2 = lambda1 / 2;
     else
       lambda2 = 0;
 
-    LARS lars(useCholesky, lambda1, lambda2);
+    LARS<MatType> lars(useCholesky, lambda1, lambda2);
     lars.FitIntercept(fitIntercept);
     lars.NormalizeData(normalizeData);
-    arma::vec betaOpt;
-    lars.Train(X, y, betaOpt);
+    lars.Train(X, y);
+    arma::Col<ElemType> betaOpt = lars.Beta();
 
     if (fitIntercept)
     {
@@ -83,51 +94,53 @@ void LassoTest(size_t nPoints, size_t nD
 
     if (normalizeData)
     {
-      arma::vec stds = arma::stddev(X, 0, 1);
+      arma::Col<ElemType> stds = arma::stddev(X, 0, 1);
       stds.replace(0.0, 1.0);
 
       X.each_col() /= stds;
       betaOpt %= stds; // recover solution in normalized space
     }
 
-    arma::vec errCorr = (X * trans(X) + lambda2 *
-        arma::eye(nDims, nDims)) * betaOpt - X * y.t();
+    arma::Col<ElemType> errCorr = (X * trans(X) + lambda2 *
+        arma::eye<MatType>(nDims, nDims)) * betaOpt - X * y.t();
 
     LARSVerifyCorrectness(betaOpt, errCorr, lambda1);
   }
 }
 
-TEST_CASE("LARSTestLassoCholesky", "[LARSTest]")
+TEMPLATE_TEST_CASE("LARSTestLassoCholesky", "[LARSTest]", arma::fmat, arma::mat)
 {
-  LassoTest(100, 10, false, true, false, false);
-  LassoTest(100, 10, false, true, true, false);
-  LassoTest(100, 10, false, true, false, true);
-  LassoTest(100, 10, false, true, true, true);
+  LassoTest<TestType>(100, 10, false, true, false, false);
+  LassoTest<TestType>(100, 10, false, true, true, false);
+  LassoTest<TestType>(100, 10, false, true, false, true);
+  LassoTest<TestType>(100, 10, false, true, true, true);
 }
 
 
-TEST_CASE("LARSTestLassoGram", "[LARSTest]")
+TEMPLATE_TEST_CASE("LARSTestLassoGram", "[LARSTest]", arma::fmat, arma::mat)
 {
-  LassoTest(100, 10, false, false, false, false);
-  LassoTest(100, 10, false, false, true, false);
-  LassoTest(100, 10, false, false, false, true);
-  LassoTest(100, 10, false, false, true, true);
+  LassoTest<TestType>(100, 10, false, false, false, false);
+  LassoTest<TestType>(100, 10, false, false, true, false);
+  LassoTest<TestType>(100, 10, false, false, false, true);
+  LassoTest<TestType>(100, 10, false, false, true, true);
 }
 
-TEST_CASE("LARSTestElasticNetCholesky", "[LARSTest]")
+TEMPLATE_TEST_CASE("LARSTestElasticNetCholesky", "[LARSTest]", arma::fmat,
+    arma::mat)
 {
-  LassoTest(100, 10, true, true, false, false);
-  LassoTest(100, 10, true, true, true, false);
-  LassoTest(100, 10, true, true, false, true);
-  LassoTest(100, 10, true, true, true, true);
+  LassoTest<TestType>(100, 10, true, true, false, false);
+  LassoTest<TestType>(100, 10, true, true, true, false);
+  LassoTest<TestType>(100, 10, true, true, false, true);
+  LassoTest<TestType>(100, 10, true, true, true, true);
 }
 
-TEST_CASE("LARSTestElasticNetGram", "[LARSTest]")
+TEMPLATE_TEST_CASE("LARSTestElasticNetGram", "[LARSTest]", arma::fmat,
+    arma::mat)
 {
-  LassoTest(100, 10, true, false, false, false);
-  LassoTest(100, 10, true, false, true, false);
-  LassoTest(100, 10, true, false, false, true);
-  LassoTest(100, 10, true, false, true, true);
+  LassoTest<TestType>(100, 10, true, false, false, false);
+  LassoTest<TestType>(100, 10, true, false, true, false);
+  LassoTest<TestType>(100, 10, true, false, false, true);
+  LassoTest<TestType>(100, 10, true, false, true, true);
 }
 
 // Ensure that LARS doesn't crash when the data has linearly dependent features
@@ -148,15 +161,14 @@ TEST_CASE("CholeskySingularityTest", "[L
   // Test for a couple values of lambda1.
   for (double lambda1 = 0.0; lambda1 < 1.0; lambda1 += 0.1)
   {
-    LARS lars(true, lambda1, 0.0);
+    LARS<> lars(true, lambda1, 0.0);
     lars.FitIntercept(false);
     lars.NormalizeData(false);
-    arma::vec betaOpt;
-    lars.Train(X, y, betaOpt);
+    lars.Train(X, y);
 
-    arma::vec errCorr = (X * X.t()) * betaOpt - X * y.t();
+    arma::vec errCorr = (X * X.t()) * lars.Beta() - X * y.t();
 
-    LARSVerifyCorrectness(betaOpt, errCorr, lambda1);
+    LARSVerifyCorrectness(lars.Beta(), errCorr, lambda1);
   }
 }
 
@@ -176,15 +188,14 @@ TEST_CASE("NoCholeskySingularityTest", "
   // Test for a couple values of lambda1.
   for (double lambda1 = 0.0; lambda1 < 1.0; lambda1 += 0.1)
   {
-    LARS lars(false, lambda1, 0.0);
+    LARS<> lars(false, lambda1, 0.0);
     lars.FitIntercept(false);
     lars.NormalizeData(false);
-    arma::vec betaOpt;
-    lars.Train(X, y, betaOpt);
+    lars.Train(X, y);
 
-    arma::vec errCorr = (X * X.t()) * betaOpt - X * y.t();
+    arma::vec errCorr = (X * X.t()) * lars.Beta() - X * y.t();
 
-    LARSVerifyCorrectness(betaOpt, errCorr, lambda1);
+    LARSVerifyCorrectness(lars.Beta(), errCorr, lambda1);
   }
 }
 
@@ -205,32 +216,194 @@ TEST_CASE("PredictTest", "[LARSTest]")
     {
       for (double lambda2 = 0.0; lambda2 < 1.0; lambda2 += 0.2)
       {
-        LARS lars(useCholesky, lambda1, lambda2);
+        LARS<> lars(useCholesky, lambda1, lambda2);
         lars.FitIntercept(false);
         lars.NormalizeData(false);
-        arma::vec betaOpt;
-        lars.Train(X, y, betaOpt);
+        lars.Train(X, y);
 
         // Calculate what the actual error should be with these regression
         // parameters.
-        arma::vec betaOptPred = (X * X.t()) * betaOpt;
+        arma::vec betaOptPred = (X * X.t()) * lars.Beta();
         arma::rowvec predictions;
         lars.Predict(X, predictions);
         arma::vec adjPred = X * predictions.t();
 
+        const double tol = 1e-7;
+
         REQUIRE(predictions.n_elem == 1000);
         for (size_t i = 0; i < betaOptPred.n_elem; ++i)
         {
           if (std::abs(betaOptPred[i]) < 1e-5)
             REQUIRE(adjPred[i] == Approx(0.0).margin(1e-5));
           else
-            REQUIRE(adjPred[i] == Approx(betaOptPred[i]).epsilon(1e-7));
+            REQUIRE(adjPred[i] == Approx(betaOptPred[i]).epsilon(tol));
+        }
+
+        // Now check with single-point Predict(), in two ways: we will pass
+        // different types into Predict() to test templating support.  We allow
+        // a looser tolerance for predictions.
+        for (size_t i = 0; i < X.n_cols; ++i)
+          predictions[i] = lars.Predict(X.col(i));
+
+        adjPred = X * predictions.t();
+        for (size_t i = 0; i < betaOptPred.n_elem; ++i)
+        {
+          if (std::abs(betaOptPred[i]) < 1e-5)
+            REQUIRE(adjPred[i] == Approx(0.0).margin(1e-5));
+          else
+            REQUIRE(adjPred[i] == Approx(betaOptPred[i]).epsilon(tol));
+        }
+
+        for (size_t i = 0; i < X.n_cols; ++i)
+          predictions[i] = lars.Predict(X.unsafe_col(i));
+
+        adjPred = X * predictions.t();
+        for (size_t i = 0; i < betaOptPred.n_elem; ++i)
+        {
+          if (std::abs(betaOptPred[i]) < 1e-5)
+            REQUIRE(adjPred[i] == Approx(0.0).margin(1e-5));
+          else
+            REQUIRE(adjPred[i] == Approx(betaOptPred[i]).epsilon(tol));
         }
       }
     }
   }
 }
 
+// This is the same as PredictTest, but for arma::fmat, and it allows multiple
+// trials for run to deal with the lower precision of floats.
+TEST_CASE("PredictFloatTest", "[LARSTest]")
+{
+  for (size_t i = 0; i < 2; ++i)
+  {
+    // Run with both true and false.
+    bool useCholesky = bool(i);
+
+    arma::fmat X;
+    arma::frowvec y;
+
+    for (float lambda1 = 0.0; lambda1 < 1.0; lambda1 += 0.2)
+    {
+      for (float lambda2 = 0.0; lambda2 < 1.0; lambda2 += 0.2)
+      {
+        // For float data, sometimes the solutions are further away from the
+        // true solution due to precision issues, so we allow multiple trials.
+        bool success = false;
+        for (size_t trial = 0; trial < 3; ++trial)
+        {
+          // Generate a new problem so that we hopefully end up with a better
+          // fit.
+          GenerateProblem(X, y, 1000, 100);
+
+          LARS<arma::fmat> lars(useCholesky, lambda1, lambda2);
+          lars.FitIntercept(false);
+          lars.NormalizeData(false);
+          lars.Train(X, y);
+
+          // Calculate what the actual error should be with these regression
+          // parameters.
+          arma::fvec betaOptPred = (X * X.t()) * lars.Beta();
+          arma::frowvec predictions;
+          lars.Predict(X, predictions);
+          arma::fvec adjPred = X * predictions.t();
+
+          const float tol = 3e-5;
+
+          REQUIRE(predictions.n_elem == 1000);
+          bool trialSuccess = true;
+          for (size_t i = 0; i < betaOptPred.n_elem; ++i)
+          {
+            if (std::abs(betaOptPred[i]) < 1e-5)
+            {
+              if (adjPred[i] != Approx(0.0).margin(1e-5))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+            else
+            {
+              if (adjPred[i] != Approx(betaOptPred[i]).epsilon(tol))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+          }
+
+          // If this trial didn't succeed, skip to the next trial.
+          if (!trialSuccess)
+            continue;
+
+          // Now check with single-point Predict(), in two ways: we will pass
+          // different types into Predict() to test templating support.  We allow
+          // a looser tolerance for predictions.
+          for (size_t i = 0; i < X.n_cols; ++i)
+            predictions[i] = lars.Predict(X.col(i));
+
+          adjPred = X * predictions.t();
+          for (size_t i = 0; i < betaOptPred.n_elem; ++i)
+          {
+            if (std::abs(betaOptPred[i]) < 1e-5)
+            {
+              if (adjPred[i] != Approx(0.0).margin(1e-5))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+            else
+            {
+              if (adjPred[i] != Approx(betaOptPred[i]).epsilon(10 * tol))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+          }
+
+          // If this trial didn't succeed, skip to the next trial.
+          if (!trialSuccess)
+            continue;
+
+          for (size_t i = 0; i < X.n_cols; ++i)
+            predictions[i] = lars.Predict(X.unsafe_col(i));
+
+          adjPred = X * predictions.t();
+          for (size_t i = 0; i < betaOptPred.n_elem; ++i)
+          {
+            if (std::abs(betaOptPred[i]) < 1e-5)
+            {
+              if (adjPred[i] != Approx(0.0).margin(1e-5))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+            else
+            {
+              if (adjPred[i] != Approx(betaOptPred[i]).epsilon(10 * tol))
+              {
+                trialSuccess = false;
+                break;
+              }
+            }
+          }
+
+          // If this trial succeeded, we're done.
+          if (trialSuccess)
+          {
+            success = true;
+            break;
+          }
+        }
+
+        REQUIRE(success == true);
+      }
+    }
+  }
+}
+
 TEST_CASE("PredictRowMajorTest", "[LARSTest]")
 {
   arma::mat X;
@@ -239,18 +412,17 @@ TEST_CASE("PredictRowMajorTest", "[LARST
 
   // Set lambdas to 0.
 
-  LARS lars(false, 0, 0);
+  LARS<> lars(false, 0, 0);
   lars.FitIntercept(false);
   lars.NormalizeData(false);
-  arma::vec betaOpt;
-  lars.Train(X, y, betaOpt);
+  lars.Train(X, y);
 
   // Get both row-major and column-major predictions.  Make sure they are the
   // same.
   arma::rowvec rowMajorPred, colMajorPred;
 
   lars.Predict(X, colMajorPred);
-  lars.Predict(X.t(), rowMajorPred, true);
+  lars.Predict(X.t(), rowMajorPred, false);
 
   REQUIRE(colMajorPred.n_elem == rowMajorPred.n_elem);
   for (size_t i = 0; i < colMajorPred.n_elem; ++i)
@@ -275,19 +447,18 @@ TEST_CASE("LARSRetrainTest", "[LARSTest]
   arma::rowvec newY;
   GenerateProblem(newX, newY, 750, 75);
 
-  LARS lars(false, 0.1, 0.1);
+  LARS<> lars(false, 0.1, 0.1);
   lars.FitIntercept(false);
   lars.NormalizeData(false);
-  arma::vec betaOpt;
-  lars.Train(origX, origY, betaOpt);
+  lars.Train(origX, origY);
 
   // Now train on new data.
-  lars.Train(newX, newY, betaOpt);
+  lars.Train(newX, newY);
 
   arma::vec errCorr = (newX * trans(newX) + 0.1 *
-        arma::eye(75, 75)) * betaOpt - newX * newY.t();
+        arma::eye(75, 75)) * lars.Beta() - newX * newY.t();
 
-  LARSVerifyCorrectness(betaOpt, errCorr, 0.1);
+  LARSVerifyCorrectness(lars.Beta(), errCorr, 0.1);
 }
 
 /**
@@ -304,19 +475,18 @@ TEST_CASE("RetrainCholeskyTest", "[LARST
   arma::rowvec newY;
   GenerateProblem(newX, newY, 750, 75);
 
-  LARS lars(true, 0.1, 0.1);
+  LARS<> lars(true, 0.1, 0.1);
   lars.FitIntercept(false);
   lars.NormalizeData(false);
-  arma::vec betaOpt;
-  lars.Train(origX, origY, betaOpt);
+  lars.Train(origX, origY);
 
   // Now train on new data.
-  lars.Train(newX, newY, betaOpt);
+  lars.Train(newX, newY);
 
   arma::vec errCorr = (newX * trans(newX) + 0.1 *
-        arma::eye(75, 75)) * betaOpt - newX * newY.t();
+        arma::eye(75, 75)) * lars.Beta() - newX * newY.t();
 
-  LARSVerifyCorrectness(betaOpt, errCorr, 0.1);
+  LARSVerifyCorrectness(lars.Beta(), errCorr, 0.1);
 }
 
 /**
@@ -330,11 +500,11 @@ TEST_CASE("TrainingAndAccessingBetaTest"
 
   GenerateProblem(X, y, 1000, 100);
 
-  LARS lars1;
-  arma::vec beta;
-  lars1.Train(X, y, beta);
+  LARS<> lars1;
+  lars1.Train(X, y);
+  arma::vec beta = lars1.Beta();
 
-  LARS lars2;
+  LARS<> lars2;
   lars2.Train(X, y);
 
   REQUIRE(beta.n_elem == lars2.Beta().n_elem);
@@ -353,11 +523,11 @@ TEST_CASE("TrainingConstructorWithDefaul
 
   GenerateProblem(X, y, 1000, 100);
 
-  LARS lars1;
-  arma::vec beta;
-  lars1.Train(X, y, beta);
+  LARS<> lars1;
+  lars1.Train(X, y);
+  arma::vec beta = lars1.Beta();
 
-  LARS lars2(X, y);
+  LARS<> lars2(X, y);
 
   REQUIRE(beta.n_elem == lars2.Beta().n_elem);
   for (size_t i = 0; i < beta.n_elem; ++i)
@@ -380,11 +550,11 @@ TEST_CASE("TrainingConstructorWithNonDef
   double lambda1 = 0.2;
   double lambda2 = 0.4;
 
-  LARS lars1(useCholesky, lambda1, lambda2);
-  arma::vec beta;
-  lars1.Train(X, y, beta);
+  LARS<> lars1(useCholesky, lambda1, lambda2);
+  lars1.Train(X, y);
+  arma::vec beta = lars1.Beta();
 
-  LARS lars2(X, y, transposeData, useCholesky, lambda1, lambda2);
+  LARS<> lars2(X, y, transposeData, useCholesky, lambda1, lambda2);
 
   REQUIRE(beta.n_elem == lars2.Beta().n_elem);
   for (size_t i = 0; i < beta.n_elem; ++i)
@@ -410,30 +580,26 @@ TEST_CASE("LARSTrainReturnCorrelation",
   double lambda2 = 0.1;
 
   // Test with Cholesky decomposition and with lasso.
-  LARS lars1(true, lambda1, 0.0);
-  arma::vec betaOpt1;
-  double error = lars1.Train(X, y, betaOpt1);
+  LARS<> lars1(true, lambda1, 0.0);
+  double error = lars1.Train(X, y);
 
   REQUIRE(std::isfinite(error) == true);
 
   // Test without Cholesky decomposition and with lasso.
-  LARS lars2(false, lambda1, 0.0);
-  arma::vec betaOpt2;
-  error = lars2.Train(X, y, betaOpt2);
+  LARS<> lars2(false, lambda1, 0.0);
+  error = lars2.Train(X, y);
 
   REQUIRE(std::isfinite(error) == true);
 
   // Test with Cholesky decomposition and with elasticnet.
-  LARS lars3(true, lambda1, lambda2);
-  arma::vec betaOpt3;
-  error = lars3.Train(X, y, betaOpt3);
+  LARS<> lars3(true, lambda1, lambda2);
+  error = lars3.Train(X, y);
 
   REQUIRE(std::isfinite(error) == true);
 
   // Test without Cholesky decomposition and with elasticnet.
-  LARS lars4(false, lambda1, lambda2);
-  arma::vec betaOpt4;
-  error = lars4.Train(X, y, betaOpt4);
+  LARS<> lars4(false, lambda1, lambda2);
+  error = lars4.Train(X, y);
 
   REQUIRE(std::isfinite(error) == true);
 }
@@ -442,24 +608,26 @@ TEST_CASE("LARSTrainReturnCorrelation",
  * Test that LARS::ComputeError() returns error value less than 1
  * and greater than 0.
  */
-TEST_CASE("LARSTestComputeError", "[LARSTest]")
+TEMPLATE_TEST_CASE("LARSTestComputeError", "[LARSTest]", arma::fmat, arma::mat)
 {
-  arma::mat X;
-  arma::mat Y;
+  typedef TestType MatType;
+  typedef typename MatType::elem_type ElemType;
+
+  MatType X;
+  MatType Y;
 
   if (!data::Load("lars_dependent_x.csv", X))
     FAIL("Cannot load dataset lars_dependent_x.csv");
   if (!data::Load("lars_dependent_y.csv", Y))
     FAIL("Cannot load dataset lars_dependent_y.csv");
 
-  arma::rowvec y = Y.row(0);
+  arma::Row<ElemType> y = Y.row(0);
 
-  LARS lars1(true, 0.1, 0.0);
+  LARS<MatType> lars1(true, 0.1, 0.0);
   lars1.FitIntercept(false);
   lars1.NormalizeData(false);
-  arma::vec betaOpt1;
-  double train1 = lars1.Train(X, y, betaOpt1);
-  double cost = lars1.ComputeError(X, y);
+  ElemType train1 = lars1.Train(X, y);
+  ElemType cost = lars1.ComputeError(X, y);
 
   REQUIRE(cost <= 1);
   REQUIRE(cost >= 0);
@@ -483,9 +651,9 @@ TEST_CASE("LARSCopyConstructorTest", "[L
 
   // Check if the copy is accessible even after deleting the pointer to the
   // object.
-  LARS* glm1 = new LARS(false, .1, .1);
+  LARS<>* glm1 = new LARS<>(false, .1, .1);
   arma::rowvec predictions, predictionsFromCopiedModel;
-  std::vector<LARS> models;
+  std::vector<LARS<>> models;
   glm1->Train(features, targets);
   glm1->Predict(features, predictions);
   models.emplace_back(*glm1); // Call the copy constructor.
@@ -497,13 +665,13 @@ TEST_CASE("LARSCopyConstructorTest", "[L
   REQUIRE_NOTHROW(models[0].Train(features, targets));
 
   // Check if we can train the copied model.
-  LARS glm2(false, 0.1, 0.1);
+  LARS<> glm2(false, 0.1, 0.1);
   models.emplace_back(glm2); // Call the copy constructor.
   REQUIRE_NOTHROW(glm2.Train(features, targets));
   REQUIRE_NOTHROW(models[1].Train(features, targets));
 
   // Create a copy using assignment operator.
-  LARS glm3 = glm2;
+  LARS<> glm3 = glm2;
   models[1].Predict(features, predictions);
   glm3.Predict(features, predictionsFromCopiedModel);
   // The output of both models should be the same.
@@ -521,14 +689,14 @@ TEST_CASE("LARSFitInterceptTest", "[LARS
   arma::mat centeredFeatures = features.each_col() - arma::mean(features, 1);
   arma::rowvec centeredResponses = responses - arma::mean(responses);
 
-  LARS l1(features, responses, true, true, 0.001, 0.001, 1e-16, true, false);
-  LARS l2(centeredFeatures, centeredResponses, true, true, 0.001, 0.001, 1e-16, false, false);
+  LARS<> l1(features, responses, true, true, 0.001, 0.001, 1e-16, true, false);
+  LARS<> l2(centeredFeatures, centeredResponses, true, true, 0.001, 0.001, 1e-16, false, false);
 
   // The weights learned should be the same.
   REQUIRE(l1.Beta().n_elem == l2.Beta().n_elem);
   CheckMatrices(l1.Beta(), l2.Beta());
   REQUIRE(l1.Intercept() == Approx(arma::mean(responses) -
-      arma::dot(arma::mean(features, 1), l1.Beta())));
+      dot(arma::mean(features, 1), l1.Beta())));
 }
 
 // Make sure that Predict() provides reasonable enough solutions when we are
@@ -549,17 +717,16 @@ TEST_CASE("PredictFitInterceptTest", "[L
     {
       for (double lambda2 = 0.0; lambda2 < 1.0; lambda2 += 0.2)
       {
-        LARS lars(useCholesky, lambda1, lambda2);
+        LARS<> lars(useCholesky, lambda1, lambda2);
         lars.FitIntercept(true);
         lars.NormalizeData(false);
-        arma::vec betaOpt;
-        lars.Train(X, y, betaOpt);
+        lars.Train(X, y);
         const double intercept = arma::mean(y) -
-            arma::dot(arma::mean(X, 1), betaOpt);
+            dot(arma::mean(X, 1), lars.Beta());
 
         // Calculate what the actual error should be with these regression
         // parameters.
-        arma::vec betaOptPred = X.t() * betaOpt + intercept;
+        arma::vec betaOptPred = X.t() * lars.Beta() + intercept;
         arma::rowvec predictions;
         lars.Predict(X, predictions);
         arma::vec adjPred = predictions.t();
@@ -595,15 +762,14 @@ TEST_CASE("PredictNormalizeDataTest", "[
     {
       for (double lambda2 = 0.0; lambda2 < 1.0; lambda2 += 0.2)
       {
-        LARS lars(useCholesky, lambda1, lambda2);
+        LARS<> lars(useCholesky, lambda1, lambda2);
         lars.FitIntercept(false);
         lars.NormalizeData(true);
-        arma::vec betaOpt;
-        lars.Train(X, y, betaOpt);
+        lars.Train(X, y);
 
         // Calculate what the actual error should be with these regression
         // parameters.
-        arma::vec betaOptPred = (X * X.t()) * betaOpt;
+        arma::vec betaOptPred = (X * X.t()) * lars.Beta();
         arma::rowvec predictions;
         lars.Predict(X, predictions);
         arma::vec adjPred = X * predictions.t();
@@ -639,17 +805,16 @@ TEST_CASE("PredictFitInterceptNormalizeD
     {
       for (double lambda2 = 0.0; lambda2 < 1.0; lambda2 += 0.2)
       {
-        LARS lars(useCholesky, lambda1, lambda2);
+        LARS<> lars(useCholesky, lambda1, lambda2);
         lars.FitIntercept(true);
         lars.NormalizeData(true);
-        arma::vec betaOpt;
-        lars.Train(X, y, betaOpt);
+        lars.Train(X, y);
         const double intercept = arma::mean(y) -
-            arma::dot(arma::mean(X, 1), betaOpt);
+            dot(arma::mean(X, 1), lars.Beta());
 
         // Calculate what the actual error should be with these regression
         // parameters.
-        arma::vec betaOptPred = X.t() * betaOpt + intercept;
+        arma::vec betaOptPred = X.t() * lars.Beta() + intercept;
         arma::rowvec predictions;
         lars.Predict(X, predictions);
         arma::vec adjPred = predictions.t();
@@ -685,16 +850,16 @@ void CheckKKT(const arma::vec& beta,
               const arma::rowvec& y,
               const double lambda)
 {
-  const double epsilon = 1e-10; // For numerical precision.
+  const double epsilon = 1e-6; // For numerical precision.
 
-  arma::vec v = X.t() * X * beta - X.t() * y.t() + lambda * arma::sign(beta);
+  arma::vec v = X.t() * X * beta - X.t() * y.t() + lambda * sign(beta);
   // Active set indices with global numbering: could be empty.
   arma::uvec ia = arma::find(arma::abs(beta) > epsilon);
   // Zero indices with global numbering: could be empty.
   arma::uvec iz = arma::find(arma::abs(beta) <= epsilon);
 
   // Should be zero if beta is the solution.
-  const double crit = arma::dot(beta, v);
+  const double crit = dot(beta, v);
   REQUIRE(std::abs(crit) < epsilon);
 
   // v should be zero at the Active Set ia.
@@ -715,9 +880,8 @@ TEST_CASE("LARSTestKKT", "[LARSTest]")
   arma::field<arma::mat> F;
   F.load("lars_kkt.bin");
 
-  arma::vec beta;
   bool useCholesky = true;
-  LARS lars(useCholesky, 1.0, 0.0);
+  LARS<> lars(useCholesky, 1.0, 0.0);
 
   for (size_t i = 0; i < F.n_cols; i++)
   {
@@ -730,19 +894,19 @@ TEST_CASE("LARSTestKKT", "[LARSTest]")
 
     lars.FitIntercept(false);
     lars.NormalizeData(false);
-    lars.Train(X, y, beta, false);
-    CheckKKT(beta, X, y, 1.0);
+    lars.Train(X, y, false);
+    CheckKKT(lars.Beta(), X, y, 1.0);
 
     // Now try when we fit an intercept too.
     lars.FitIntercept(true);
     lars.NormalizeData(false);
-    lars.Train(X, y, beta, false);
+    lars.Train(X, y, false);
 
     // Now mean-center data before the check.
     X.each_row() -= xMean;
     y -= yMean;
 
-    CheckKKT(beta, X, y, 1.0);
+    CheckKKT(lars.Beta(), X, y, 1.0);
 
     X.each_row() += xMean;
     y += yMean;
@@ -750,9 +914,10 @@ TEST_CASE("LARSTestKKT", "[LARSTest]")
     // Now try when we normalize the data.
     lars.FitIntercept(false);
     lars.NormalizeData(true);
-    lars.Train(X, y, beta, false);
+    lars.Train(X, y, false);
 
     X.each_row() /= xStds;
+    arma::vec beta = lars.Beta();
     beta %= xStds.t();
 
     CheckKKT(beta, X, y, 1.0);
@@ -761,7 +926,8 @@ TEST_CASE("LARSTestKKT", "[LARSTest]")
 
     lars.FitIntercept(true);
     lars.NormalizeData(true);
-    lars.Train(X, y, beta, false);
+    lars.Train(X, y, false);
+    beta = lars.Beta();
 
     X.each_row() -= xMean;
     X.each_row() /= xStds;
@@ -771,3 +937,412 @@ TEST_CASE("LARSTestKKT", "[LARSTest]")
     CheckKKT(beta, X, y, 1.0);
   }
 }
+
+// Check that all variants of constructors appear to work.
+TEMPLATE_TEST_CASE("LARSConstructorVariantTest", "[LARSTest]", arma::fmat,
+    arma::mat)
+{
+  typedef TestType MatType;
+  typedef typename MatType::elem_type ElemType;
+
+  // The results of the training are not all that important here; the more
+  // important thing is just that all the overloads compile properly.  We do
+  // some basic sanity checks on the trained model nonetheless.
+  MatType X;
+  arma::Row<ElemType> y;
+
+  GenerateProblem(X, y, 1000, 100);
+  MatType Xt = X.t();
+
+  const arma::Col<ElemType> xMean = arma::mean(X, 1);
+  arma::Col<ElemType> xStds = arma::stddev(X, 0, 1);
+  xStds.replace(0.0, 1.0);
+  const ElemType yMean = arma::mean(y);
+
+  MatType centeredX = X.each_col() - xMean;
+  arma::Row<ElemType> centeredY = y - yMean;
+
+  MatType centeredUnitX = centeredX.each_col() / xStds;
+
+  MatType matGram = X * X.t();
+  MatType centeredUnitMatGram = centeredUnitX * centeredUnitX.t();
+
+  LARS<MatType> l1;
+  LARS<MatType> l2(false, 0.1, 0.2, 1e-15, false, false);
+  LARS<MatType> l3(X, y);
+  LARS<MatType> l4(Xt, y, false);
+  LARS<MatType> l5(Xt, y, false, false);
+  LARS<MatType> l6(Xt, y, false, false, 0.1);
+  LARS<MatType> l7(X, y, true, false, 0.11, 0.01);
+  LARS<MatType> l8(X, y, true, false, 0.12, 0.02, 1e-8);
+  LARS<MatType> l9(centeredX, centeredY, true, false, 0.13, 0.03, 1e-7, false);
+  LARS<MatType> l10(centeredUnitX, centeredY, true, false, 0.14, 0.04, 1e-6,
+      false, false);
+
+  REQUIRE(l1.BetaPath().size() == 0);
+
+  REQUIRE(l2.BetaPath().size() == 0);
+  REQUIRE(l2.UseCholesky() == false);
+  REQUIRE(l2.Lambda1() == Approx(0.1));
+  REQUIRE(l2.Lambda2() == Approx(0.2));
+  REQUIRE(l2.Tolerance() == Approx(1e-15));
+  REQUIRE(l2.FitIntercept() == false);
+  REQUIRE(l2.NormalizeData() == false);
+
+  REQUIRE(l3.Beta().n_elem == X.n_rows);
+
+  REQUIRE(l4.Beta().n_elem == X.n_rows);
+
+  REQUIRE(l5.Beta().n_elem == X.n_rows);
+  REQUIRE(l5.UseCholesky() == false);
+
+  REQUIRE(l6.Beta().n_elem == X.n_rows);
+  REQUIRE(l6.UseCholesky() == false);
+  REQUIRE(l6.Lambda1() == Approx(0.1));
+
+  REQUIRE(l7.Beta().n_elem == X.n_rows);
+  REQUIRE(l7.UseCholesky() == false);
+  REQUIRE(l7.Lambda1() == Approx(0.11));
+  REQUIRE(l7.Lambda2() == Approx(0.01));
+
+  REQUIRE(l8.Beta().n_elem == X.n_rows);
+  REQUIRE(l8.UseCholesky() == false);
+  REQUIRE(l8.Lambda1() == Approx(0.12));
+  REQUIRE(l8.Lambda2() == Approx(0.02));
+  REQUIRE(l8.Tolerance() == Approx(1e-8));
+
+  REQUIRE(l9.Beta().n_elem == X.n_rows);
+  REQUIRE(l9.UseCholesky() == false);
+  REQUIRE(l9.Lambda1() == Approx(0.13));
+  REQUIRE(l9.Lambda2() == Approx(0.03));
+  REQUIRE(l9.Tolerance() == Approx(1e-7));
+  REQUIRE(l9.FitIntercept() == false);
+
+  REQUIRE(l10.Beta().n_elem == X.n_rows);
+  REQUIRE(l10.UseCholesky() == false);
+  REQUIRE(l10.Lambda1() == Approx(0.14));
+  REQUIRE(l10.Lambda2() == Approx(0.04));
+  REQUIRE(l10.Tolerance() == Approx(1e-6));
+  REQUIRE(l10.FitIntercept() == false);
+  REQUIRE(l10.NormalizeData() == false);
+
+  // Now check constructors where we specify the Gram matrix.
+
+  const size_t dim = centeredUnitMatGram.n_rows;
+  LARS<MatType> l11(X, y, true, false, centeredUnitMatGram);
+  LARS<MatType> l12(Xt, y, false, false, centeredUnitMatGram, 0.1);
+
+  // If lambda2 > 0, then we have to adjust the Gram matrix to account for that.
+  MatType centeredUnitMatGramL13 = centeredUnitMatGram +
+      0.01 * arma::eye<MatType>(dim, dim);
+  LARS<MatType> l13(Xt, y, false, false, centeredUnitMatGramL13, 0.11, 0.01);
+
+  MatType centeredUnitMatGramL14 = centeredUnitMatGram +
+      0.02 * arma::eye<MatType>(dim, dim);
+  LARS<MatType> l14(Xt, y, false, false, centeredUnitMatGramL14, 0.12, 0.02,
+      1e-15);
+
+  MatType centeredUnitMatGramL15 = centeredUnitMatGram +
+      0.03 * arma::eye<MatType>(dim, dim);
+  LARS<MatType> l15(centeredX, centeredY, true, false, centeredUnitMatGramL15,
+      0.13, 0.03, 1e-14, false);
+
+  MatType centeredUnitMatGramL16 = centeredUnitMatGram +
+      0.04 * arma::eye<MatType>(dim, dim);
+  LARS<MatType> l16(centeredUnitX, centeredY, true, false,
+      centeredUnitMatGramL16, 0.14, 0.04, 1e-13, false, false);
+
+  REQUIRE(l11.Beta().n_elem == X.n_rows);
+  REQUIRE(l11.UseCholesky() == false);
+
+  REQUIRE(l12.Beta().n_elem == X.n_rows);
+  REQUIRE(l12.UseCholesky() == false);
+  REQUIRE(l12.Lambda1() == Approx(0.1));
+
+  REQUIRE(l13.Beta().n_elem == X.n_rows);
+  REQUIRE(l13.UseCholesky() == false);
+  REQUIRE(l13.Lambda1() == Approx(0.11));
+  REQUIRE(l13.Lambda2() == Approx(0.01));
+
+  REQUIRE(l14.Beta().n_elem == X.n_rows);
+  REQUIRE(l14.UseCholesky() == false);
+  REQUIRE(l14.Lambda1() == Approx(0.12));
+  REQUIRE(l14.Lambda2() == Approx(0.02));
+  REQUIRE(l14.Tolerance() == Approx(1e-15));
+
+  REQUIRE(l15.Beta().n_elem == X.n_rows);
+  REQUIRE(l15.UseCholesky() == false);
+  REQUIRE(l15.Lambda1() == Approx(0.13));
+  REQUIRE(l15.Lambda2() == Approx(0.03));
+  REQUIRE(l15.Tolerance() == Approx(1e-14));
+  REQUIRE(l15.FitIntercept() == false);
+
+  REQUIRE(l16.Beta().n_elem == X.n_rows);
+  REQUIRE(l16.UseCholesky() == false);
+  REQUIRE(l16.Lambda1() == Approx(0.14));
+  REQUIRE(l16.Lambda2() == Approx(0.04));
+  REQUIRE(l16.Tolerance() == Approx(1e-13));
+  REQUIRE(l16.FitIntercept() == false);
+  REQUIRE(l16.NormalizeData() == false);
+}
+
+// Check that all variants of Train() appear to work.
+TEMPLATE_TEST_CASE("LARSTrainVariantTest", "[LARSTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+  typedef typename MatType::elem_type ElemType;
+
+  // The results of the training are not all that important here; the more
+  // important thing is just that all the overloads compile properly.  We do
+  // some basic sanity checks on the trained model nonetheless.
+  MatType X;
+  arma::Row<ElemType> y;
+
+  GenerateProblem(X, y, 1000, 5);
+  MatType Xt = X.t();
+
+  const arma::Col<ElemType> xMean = arma::mean(X, 1);
+  arma::Col<ElemType> xStds = arma::stddev(X, 0, 1);
+  xStds.replace(0.0, 1.0);
+  const ElemType yMean = arma::mean(y);
+
+  MatType centeredX = X.each_col() - xMean;
+  arma::Row<ElemType> centeredY = y - yMean;
+
+  MatType centeredUnitX = centeredX.each_col() / xStds;
+
+  MatType matGram = X * X.t();
+  MatType centeredUnitMatGram = centeredUnitX * centeredUnitX.t();
+
+  LARS<MatType> l1, l2, l3, l4, l5, l6, l7, l8, l9, l10, l11, l12, l13, l14;
+
+  l1.Train(X, y);
+  l2.Train(Xt, y, false);
+  l3.Train(Xt, y, false, false);
+  l4.Train(Xt, y, false, false, 0.1);
+  l5.Train(Xt, y, false, false, 0.11, 0.01);
+  l6.Train(Xt, y, false, false, 0.12, 0.02, 1e-15);
+  l7.Train(centeredX, centeredY, true, false, 0.13, 0.03, 1e-14, false);
+  l8.Train(centeredX, centeredY, true, false, 0.14, 0.04, 1e-13, false, false);
+
+  l9.Train(Xt, y, false, false, centeredUnitMatGram);
+  l10.Train(X, y, true, false, centeredUnitMatGram, 0.15);
+
+  // If lambda2 > 0, then we have to adjust the Gram matrix.
+  const size_t dim = centeredUnitMatGram.n_rows;
+  MatType centeredUnitMatGramL11 = centeredUnitMatGram +
+      0.05 * arma::eye<MatType>(dim, dim);
+  l11.Train(X, y, true, false, centeredUnitMatGramL11, 0.16, 0.05);
+
+  MatType centeredUnitMatGramL12 = centeredUnitMatGram +
+      0.06 * arma::eye<MatType>(dim, dim);
+  l12.Train(X, y, true, false, centeredUnitMatGramL12, 0.17, 0.06, 1e-12);
+
+  MatType centeredUnitMatGramL13 = centeredUnitMatGram +
+      0.07 * arma::eye<MatType>(dim, dim);
+  l13.Train(centeredX, centeredY, true, false, centeredUnitMatGramL13, 0.18,
+      0.07, 1e-11, false);
+
+  MatType centeredUnitMatGramL14 = centeredUnitMatGram +
+      0.08 * arma::eye<MatType>(dim, dim);
+  l14.Train(centeredUnitX, centeredY, true, false, centeredUnitMatGramL14, 0.19,
+      0.08, 1e-10, false, false);
+
+  REQUIRE(l1.Beta().n_elem == X.n_rows);
+
+  REQUIRE(l2.Beta().n_elem == X.n_rows);
+
+  REQUIRE(l3.Beta().n_elem == X.n_rows);
+  REQUIRE(l3.UseCholesky() == false);
+
+  REQUIRE(l4.Beta().n_elem == X.n_rows);
+  REQUIRE(l4.UseCholesky() == false);
+  REQUIRE(l4.Lambda1() == Approx(0.1));
+
+  REQUIRE(l5.Beta().n_elem == X.n_rows);
+  REQUIRE(l5.UseCholesky() == false);
+  REQUIRE(l5.Lambda1() == Approx(0.11));
+  REQUIRE(l5.Lambda2() == Approx(0.01));
+
+  REQUIRE(l6.Beta().n_elem == X.n_rows);
+  REQUIRE(l6.UseCholesky() == false);
+  REQUIRE(l6.Lambda1() == Approx(0.12));
+  REQUIRE(l6.Lambda2() == Approx(0.02));
+  REQUIRE(l6.Tolerance() == Approx(1e-15));
+
+  REQUIRE(l7.Beta().n_elem == X.n_rows);
+  REQUIRE(l7.UseCholesky() == false);
+  REQUIRE(l7.Lambda1() == Approx(0.13));
+  REQUIRE(l7.Lambda2() == Approx(0.03));
+  REQUIRE(l7.Tolerance() == Approx(1e-14));
+  REQUIRE(l7.FitIntercept() == false);
+
+  REQUIRE(l8.Beta().n_elem == X.n_rows);
+  REQUIRE(l8.UseCholesky() == false);
+  REQUIRE(l8.Lambda1() == Approx(0.14));
+  REQUIRE(l8.Lambda2() == Approx(0.04));
+  REQUIRE(l8.Tolerance() == Approx(1e-13));
+  REQUIRE(l8.FitIntercept() == false);
+  REQUIRE(l8.NormalizeData() == false);
+
+  REQUIRE(l9.Beta().n_elem == X.n_rows);
+  REQUIRE(l9.UseCholesky() == false);
+
+  REQUIRE(l10.Beta().n_elem == X.n_rows);
+  REQUIRE(l10.UseCholesky() == false);
+  REQUIRE(l10.Lambda1() == Approx(0.15));
+
+  REQUIRE(l11.Beta().n_elem == X.n_rows);
+  REQUIRE(l11.UseCholesky() == false);
+  REQUIRE(l11.Lambda1() == Approx(0.16));
+  REQUIRE(l11.Lambda2() == Approx(0.05));
+
+  REQUIRE(l12.Beta().n_elem == X.n_rows);
+  REQUIRE(l12.UseCholesky() == false);
+  REQUIRE(l12.Lambda1() == Approx(0.17));
+  REQUIRE(l12.Lambda2() == Approx(0.06));
+  REQUIRE(l12.Tolerance() == Approx(1e-12));
+
+  REQUIRE(l13.Beta().n_elem == X.n_rows);
+  REQUIRE(l13.UseCholesky() == false);
+  REQUIRE(l13.Lambda1() == Approx(0.18));
+  REQUIRE(l13.Lambda2() == Approx(0.07));
+  REQUIRE(l13.Tolerance() == Approx(1e-11));
+  REQUIRE(l13.FitIntercept() == false);
+
+  REQUIRE(l14.Beta().n_elem == X.n_rows);
+  REQUIRE(l14.UseCholesky() == false);
+  REQUIRE(l14.Lambda1() == Approx(0.19));
+  REQUIRE(l14.Lambda2() == Approx(0.08));
+  REQUIRE(l14.Tolerance() == Approx(1e-10));
+  REQUIRE(l14.FitIntercept() == false);
+  REQUIRE(l14.NormalizeData() == false);
+}
+
+// Ensure that SelectBeta() works correctly.
+TEMPLATE_TEST_CASE("LARSSelectBetaTest", "[LARSTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+  typedef typename MatType::elem_type ElemType;
+
+  const ElemType tol = (std::is_same<ElemType, double>::value) ? 1e-5 : 5e-3;
+
+  // Train a model on a randomly generated problem.  Then, we will iterate
+  // through different selected lambda values, ensuring that the error on the
+  // training set is monotonically increasing.
+  MatType X;
+  arma::Row<ElemType> y;
+
+  GenerateProblem(X, y, 1000, 100);
+
+  LARS<MatType> lars(X, y);
+
+  // Ensure that the solution with no regularization is fully dense.
+  REQUIRE(lars.ActiveSet().size() == X.n_rows);
+  REQUIRE(lars.Beta().n_elem == X.n_rows);
+
+  // Now step through numerous different lambda values.
+  ElemType lastError = std::numeric_limits<ElemType>::max();
+  const ElemType errorTol = (std::is_same<ElemType, double>::value) ? 1e-8 :
+      0.05;
+  for (ElemType i = 5.0; i >= -5.0; i -= 0.1)
+  {
+    const ElemType selLambda1 = std::pow(10.0, (ElemType) i);
+    lars.SelectBeta(selLambda1);
+
+    REQUIRE(lars.Beta().n_elem == X.n_rows);
+    REQUIRE(lars.SelectedLambda1() == Approx(selLambda1).margin(tol));
+    REQUIRE(accu(lars.Beta() != 0.0) == lars.ActiveSet().size());
+    const ElemType newError = lars.ComputeError(X, y);
+    REQUIRE(newError <= lastError + errorTol);
+    lastError = newError;
+  }
+
+  // Lastly, step through values corresponding to the lambda path exactly.
+  // Here we can just check that we are looking at the right model via
+  // Intercept() and Beta().
+  for (size_t i = 0; i < lars.LambdaPath().size(); ++i)
+  {
+    lars.SelectBeta(lars.LambdaPath()[i]);
+
+    REQUIRE(lars.SelectedLambda1() == Approx(lars.LambdaPath()[i]).margin(tol));
+    REQUIRE(lars.Intercept() == Approx(lars.InterceptPath()[i]).margin(tol));
+    REQUIRE(arma::approx_equal(lars.Beta(), lars.BetaPath()[i], "absdiff",
+        tol));
+  }
+}
+
+// Test that SelectBeta() throws an error when the model is not trained.
+TEST_CASE("LARSSelectBetaUntrainedModelTest", "[LARSTest]")
+{
+  LARS<> lars;
+
+  REQUIRE_THROWS_AS(lars.SelectBeta(0.01), std::runtime_error);
+}
+
+// Test that SelectBeta() throws an exception when an invalid new lambda1 value
+// is specified.
+TEST_CASE("LARSSelectBetaInvalidLambda1Test", "[LARSTest]")
+{
+  arma::mat X;
+  arma::rowvec y;
+
+  GenerateProblem(X, y, 1000, 100);
+
+  LARS<> lars;
+  lars.Lambda1() = 1.0;
+  lars.Train(X, y);
+
+  REQUIRE_THROWS_AS(lars.SelectBeta(0.1), std::invalid_argument);
+}
+
+// Test that we can train a sparse model on dense data.
+TEMPLATE_TEST_CASE("LARSSparseModelDenseData", "[LARSTest]", float, double)
+{
+  typedef TestType eT;
+
+  // 1k-dimensional data.
+  arma::Mat<eT> data(1000, 500, arma::fill::randu);
+
+  arma::SpCol<eT> betaSp;
+  betaSp.sprandu(1000, 1, 0.1);
+  arma::Col<eT> beta = betaSp + arma::randu<arma::Col<eT>>(1000) * 0.01;
+
+  // Create slightly noisy responses.
+  arma::Row<eT> responses = beta.t() * data +
+      0.02 * arma::randu<arma::Row<eT>>(500);
+
+  LARS<arma::SpMat<eT>> lars1(data, responses, true, true, 0.5, 0.01);
+  LARS<arma::SpMat<eT>> lars2(true, 0.01, 1e-4);
+  lars2.Train(data, responses);
+
+  // Make sure we at least approximately recovered the solution vector.
+  REQUIRE(lars1.Beta().n_elem == 1000);
+  REQUIRE(lars2.Beta().n_elem == 1000);
+
+  REQUIRE((arma::norm(lars1.Beta() - beta, 2) / lars1.Beta().n_elem) < 0.01);
+  REQUIRE((arma::norm(lars2.Beta() - beta, 2) / lars2.Beta().n_elem) < 0.01);
+
+  // Make some predictions and ensure they are approximately correct.
+  arma::Row<eT> responses1, responses2;
+  lars1.Predict(data, responses1);
+  lars2.Predict(data, responses2);
+
+  REQUIRE(responses1.n_elem == responses.n_elem);
+  REQUIRE(responses2.n_elem == responses.n_elem);
+
+  REQUIRE((accu(arma::abs(responses - responses1)) / responses.n_elem)
+      < 0.1);
+  REQUIRE((accu(arma::abs(responses - responses2)) / responses.n_elem)
+      < 0.1);
+
+  // Make sure ComputeError returns something reasonable.
+  REQUIRE((lars1.ComputeError(data, responses) / responses.n_elem) < 1);
+  REQUIRE((lars2.ComputeError(data, responses) / responses.n_elem) < 1);
+
+  REQUIRE(lars1.ActiveSet().size() < 1000);
+  REQUIRE(lars1.ActiveSet().size() > 0);
+
+  REQUIRE(lars2.ActiveSet().size() < 1000);
+  REQUIRE(lars2.ActiveSet().size() > 0);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/lin_alg_test.cpp 4.4.0-1/src/mlpack/tests/lin_alg_test.cpp
--- 4.3.0-2/src/mlpack/tests/lin_alg_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/lin_alg_test.cpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,266 +0,0 @@
-/**
- * @file tests/lin_alg_test.cpp
- * @author Ryan Curtin
- *
- * Simple tests for things in the linalg__private namespace.
- * Partly so I can be sure that my changes are working.
- * Move to boost unit testing framework at some point.
- *
- * mlpack is free software; you may redistribute it and/or modify it under the
- * terms of the 3-clause BSD license.  You should have received a copy of the
- * 3-clause BSD license along with mlpack.  If not, see
- * http://www.opensource.org/licenses/BSD-3-Clause for more information.
- */
-#include <mlpack/core.hpp>
-
-#include "catch.hpp"
-#include "test_catch_tools.hpp"
-
-using namespace arma;
-using namespace mlpack;
-
-/**
- * Test for linalg__private::Center().  There are no edge cases here, so we'll
- * just try it once for now.
- */
-TEST_CASE("TestCenterA", "[LinAlgTest]")
-{
-  mat tmp(5, 5);
-  // [[0  0  0  0  0]
-  //  [1  2  3  4  5]
-  //  [2  4  6  8  10]
-  //  [3  6  9  12 15]
-  //  [4  8  12 16 20]]
-  for (int row = 0; row < 5; row++)
-    for (int col = 0; col < 5; col++)
-      tmp(row, col) = row * (col + 1);
-
-  mat tmp_out;
-  Center(tmp, tmp_out);
-
-  // average should be
-  // [[0 3 6 9 12]]'
-  // so result should be
-  // [[ 0  0  0  0  0]
-  //  [-2 -1  0  1  2 ]
-  //  [-4 -2  0  2  4 ]
-  //  [-6 -3  0  3  6 ]
-  //  [-8 -4  0  4  8]]
-  for (int row = 0; row < 5; row++)
-  {
-    for (int col = 0; col < 5; col++)
-    {
-      REQUIRE(tmp_out(row, col) ==
-          Approx((double) (col - 2) * row).epsilon(1e-7));
-    }
-  }
-}
-
-TEST_CASE("TestCenterB", "[LinAlgTest]")
-{
-  mat tmp(5, 6);
-  for (int row = 0; row < 5; row++)
-    for (int col = 0; col < 6; col++)
-      tmp(row, col) = row * (col + 1);
-
-  mat tmp_out;
-  Center(tmp, tmp_out);
-
-  // average should be
-  // [[0 3.5 7 10.5 14]]'
-  // so result should be
-  // [[ 0    0    0   0   0   0  ]
-  //  [-2.5 -1.5 -0.5 0.5 1.5 2.5]
-  //  [-5   -3   -1   1   3   5  ]
-  //  [-7.5 -4.5 -1.5 1.5 1.5 4.5]
-  //  [-10  -6   -2   2   6   10 ]]
-  for (int row = 0; row < 5; row++)
-  {
-    for (int col = 0; col < 6; col++)
-    {
-      REQUIRE(tmp_out(row, col) ==
-          Approx((double) (col - 2.5) * row).epsilon(1e-7));
-    }
-  }
-}
-
-TEST_CASE("TestOrthogonalize", "[LinAlgTest]")
-{
-  // Generate a random matrix; then, orthogonalize it and test if it's
-  // orthogonal.
-  mat tmp, orth;
-  if (!data::Load("fake.csv", tmp))
-    FAIL("Cannot load dataset fake.csv");
-  Orthogonalize(tmp, orth);
-
-  // test orthogonality
-  mat test = ColumnCovariance(orth);
-  double ival = test(0, 0);
-  for (size_t row = 0; row < test.n_rows; row++)
-  {
-    for (size_t col = 0; col < test.n_cols; col++)
-    {
-      if (row == col)
-      {
-        if (std::abs(test(row, col)) > 1e-10)
-          REQUIRE(test(row, col) == Approx(ival).epsilon(1e-11));
-      }
-      else
-      {
-        REQUIRE(test(row, col) == Approx(0.0).margin(1e-10));
-      }
-    }
-  }
-}
-
-// Test RemoveRows().
-TEST_CASE("TestRemoveRows", "[LinAlgTest]")
-{
-  // Run this test several times.
-  for (size_t run = 0; run < 10; ++run)
-  {
-    arma::mat input;
-    input.randu(200, 200);
-
-    // Now pick some random numbers.
-    std::vector<size_t> rowsToRemove;
-    size_t row = 0;
-    while (row < 200)
-    {
-      row += RandInt(1, (2 * (run + 1) + 1));
-      if (row < 200)
-      {
-        rowsToRemove.push_back(row);
-      }
-    }
-
-    // Ensure we're not about to remove every single row.
-    if (rowsToRemove.size() == 10)
-    {
-      rowsToRemove.erase(rowsToRemove.begin() + 4); // Random choice to remove.
-    }
-
-    arma::mat output;
-    RemoveRows(input, rowsToRemove, output);
-
-    // Now check that the output is right.
-    size_t outputRow = 0;
-    size_t skipIndex = 0;
-
-    for (row = 0; row < 200; ++row)
-    {
-      // Was this row supposed to be removed?  If so skip it.
-      if ((skipIndex < rowsToRemove.size()) && (rowsToRemove[skipIndex] == row))
-      {
-        ++skipIndex;
-      }
-      else
-      {
-        // Compare.
-        REQUIRE(accu(input.row(row) == output.row(outputRow)) == 200);
-
-        // Increment output row counter.
-        ++outputRow;
-      }
-    }
-  }
-}
-
-TEST_CASE("TestSvecSmat", "[LinAlgTest]")
-{
-  arma::mat X(3, 3);
-  X(0, 0) = 0; X(0, 1) = 1, X(0, 2) = 2;
-  X(1, 0) = 1; X(1, 1) = 3, X(1, 2) = 4;
-  X(2, 0) = 2; X(2, 1) = 4, X(2, 2) = 5;
-
-  arma::vec sx;
-  Svec(X, sx);
-  REQUIRE(sx(0) == Approx(0).epsilon(1e-9));
-  REQUIRE(sx(1) == Approx(M_SQRT2 * 1.).epsilon(1e-9));
-  REQUIRE(sx(2) == Approx(M_SQRT2 * 2.).epsilon(1e-9));
-  REQUIRE(sx(3) == Approx(3.).epsilon(1e-9));
-  REQUIRE(sx(4) == Approx(M_SQRT2 * 4.).epsilon(1e-9));
-  REQUIRE(sx(5) == Approx(5.).epsilon(1e-9));
-
-  arma::mat Xtest;
-  Smat(sx, Xtest);
-  REQUIRE(Xtest.n_rows == 3);
-  REQUIRE(Xtest.n_cols == 3);
-  for (size_t i = 0; i < 3; ++i)
-    for (size_t j = 0; j < 3; ++j)
-      REQUIRE(X(i, j) == Approx(Xtest(i, j)).epsilon(1e-9));
-}
-
-TEST_CASE("TestSparseSvec", "[LinAlgTest]")
-{
-  arma::sp_mat X;
-  X.zeros(3, 3);
-  X(1, 0) = X(0, 1) = 1;
-
-  arma::sp_vec sx;
-  Svec(X, sx);
-
-  const double v0 = sx(0);
-  const double v1 = sx(1);
-  const double v2 = sx(2);
-  const double v3 = sx(3);
-  const double v4 = sx(4);
-  const double v5 = sx(5);
-
-  REQUIRE(v0 == Approx(0).epsilon(1e-9));
-  REQUIRE(v1 == Approx(M_SQRT2 * 1.).epsilon(1e-9));
-  REQUIRE(v2 == Approx(0).epsilon(1e-9));
-  REQUIRE(v3 == Approx(0).epsilon(1e-9));
-  REQUIRE(v4 == Approx(0).epsilon(1e-9));
-  REQUIRE(v5 == Approx(0).epsilon(1e-9));
-}
-
-TEST_CASE("TestSymKronIdSimple", "[LinAlgTest]")
-{
-  arma::mat A(3, 3);
-  A(0, 0) = 1; A(0, 1) = 2, A(0, 2) = 3;
-  A(1, 0) = 2; A(1, 1) = 4, A(1, 2) = 5;
-  A(2, 0) = 3; A(2, 1) = 5, A(2, 2) = 6;
-  arma::mat Op;
-  SymKronId(A, Op);
-
-  const arma::mat X = A + arma::ones<arma::mat>(3, 3);
-  arma::vec sx;
-  Svec(X, sx);
-
-  const arma::vec lhs = Op * sx;
-  const arma::mat Rhs = 0.5 * (A * X + X * A);
-  arma::vec rhs;
-  Svec(Rhs, rhs);
-
-  REQUIRE(lhs.n_elem == rhs.n_elem);
-  for (size_t j = 0; j < lhs.n_elem; ++j)
-    REQUIRE(lhs(j) == Approx(rhs(j)).epsilon(1e-7));
-}
-
-TEST_CASE("TestSymKronId", "[LinAlgTest]")
-{
-  const size_t n = 10;
-  arma::mat A = arma::randu<arma::mat>(n, n);
-  A += A.t();
-
-  arma::mat Op;
-  SymKronId(A, Op);
-
-  for (size_t i = 0; i < 5; ++i)
-  {
-    arma::mat X = arma::randu<arma::mat>(n, n);
-    X += X.t();
-    arma::vec sx;
-    Svec(X, sx);
-
-    const arma::vec lhs = Op * sx;
-    const arma::mat Rhs = 0.5 * (A * X + X * A);
-    arma::vec rhs;
-    Svec(Rhs, rhs);
-
-    REQUIRE(lhs.n_elem == rhs.n_elem);
-    for (size_t j = 0; j < lhs.n_elem; ++j)
-      REQUIRE(lhs(j) == Approx(rhs(j)).epsilon(1e-7));
-  }
-}
diff -pruN 4.3.0-2/src/mlpack/tests/linear_regression_test.cpp 4.4.0-1/src/mlpack/tests/linear_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/linear_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/linear_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,25 +21,30 @@ using namespace mlpack;
  * Creates two 10x3 random matrices and one 10x1 "results" matrix.
  * Finds B in y=BX with one matrix, then predicts against the other.
  */
-TEST_CASE("LinearRegressionTestCase", "[LinearRegressionTest]")
+TEMPLATE_TEST_CASE("LinearRegressionTestCase", "[LinearRegressionTest]",
+    arma::fmat, arma::mat)
 {
+  typedef TestType MatType;
+  typedef arma::Row<typename MatType::elem_type> RowType;
+  typedef arma::Col<typename MatType::elem_type> ColType;
+
   // Predictors and points are 10x3 matrices.
-  arma::mat predictors(3, 10);
-  arma::mat points(3, 10);
+  MatType predictors(3, 10);
+  MatType points(3, 10);
 
   // Responses is the "correct" value for each point in predictors and points.
-  arma::rowvec responses(10);
+  RowType responses(10);
 
   // The values we get back when we predict for points.
-  arma::rowvec predictions(10);
+  RowType predictions(10);
 
   // We'll randomly select some coefficients for the linear response.
-  arma::vec coeffs;
+  ColType coeffs;
   coeffs.randu(4);
 
   // Now generate each point.
   for (size_t row = 0; row < 3; row++)
-    predictors.row(row) = arma::linspace<arma::rowvec>(0, 9, 10);
+    predictors.row(row) = arma::linspace<RowType>(0, 9, 10);
 
   points = predictors;
 
@@ -57,7 +62,7 @@ TEST_CASE("LinearRegressionTestCase", "[
         dot(coeffs.rows(1, 3), arma::ones<arma::rowvec>(3) * elem);
 
   // Initialize and predict.
-  LinearRegression lr(predictors, responses);
+  LinearRegression<MatType> lr(predictors, responses);
   lr.Predict(points, predictions);
 
   // Output result and verify we have less than 5% error from "correct" value
@@ -69,16 +74,20 @@ TEST_CASE("LinearRegressionTestCase", "[
 /**
  * Check the functionality of ComputeError().
  */
-TEST_CASE("ComputeErrorTest", "[LinearRegressionTest]")
+TEMPLATE_TEST_CASE("ComputeErrorTest", "[LinearRegressionTest]", arma::fmat,
+    arma::mat)
 {
-  arma::mat predictors;
+  typedef TestType MatType;
+  typedef arma::Row<typename MatType::elem_type> RowType;
+
+  MatType predictors;
   predictors = { {  0, 1, 2, 4, 8, 16 },
                  { 16, 8, 4, 2, 1,  0 } };
-  arma::rowvec responses = "0 2 4 3 8 8";
+  RowType responses = "0 2 4 3 8 8";
 
   // http://www.mlpack.org/trac/ticket/298
   // This dataset gives a cost of 1.189500337 (as calculated in Octave).
-  LinearRegression lr(predictors, responses);
+  LinearRegression<MatType> lr(predictors, responses);
 
   REQUIRE(lr.ComputeError(predictors, responses) ==
       Approx(1.189500337).epsilon(1e-5));
@@ -95,7 +104,7 @@ TEST_CASE("ComputeErrorPerfectFitTest",
                  { 0, 1, 2, 2, 2, 6 } };
   arma::rowvec responses = "0 2 4 3 8 8";
 
-  LinearRegression lr(predictors, responses);
+  LinearRegression<> lr(predictors, responses);
 
   REQUIRE(lr.ComputeError(predictors, responses) == Approx(0.0).margin(1e-25));
 }
@@ -116,7 +125,7 @@ TEST_CASE("RidgeRegressionTest", "[Linea
   // invertible.  If ridge regression is not working correctly, then the matrix
   // will not be invertible and the test should segfault (or something else
   // ugly).
-  LinearRegression lr(data, responses, 0.0001);
+  LinearRegression<> lr(data, responses, 0.0001);
 
   // Now just make sure that it predicts some more zeros.
   arma::rowvec predictedResponses;
@@ -167,7 +176,7 @@ TEST_CASE("RidgeRegressionTestCase", "[L
         dot(coeffs.rows(1, 3), arma::ones<arma::rowvec>(3) * elem);
 
   // Initialize and predict with very small lambda.
-  LinearRegression lr(predictors, responses, 0.001);
+  LinearRegression<> lr(predictors, responses, 0.001);
   lr.Predict(points, predictions);
 
   // Output result and verify we have less than 5% error from "correct" value
@@ -186,8 +195,8 @@ TEST_CASE("LinearRegressionTrainTest", "
   arma::mat dataset = arma::randu<arma::mat>(5, 1000);
   arma::rowvec responses = arma::randu<arma::rowvec>(1000);
 
-  LinearRegression lr(dataset, responses, 0.3);
-  LinearRegression lrTrain;
+  LinearRegression<> lr(dataset, responses, 0.3);
+  LinearRegression<> lrTrain;
   lrTrain.Lambda() = 0.3;
 
   lrTrain.Train(dataset, responses);
@@ -209,8 +218,8 @@ TEST_CASE("LinearRegressionTest", "[Line
   arma::rowvec responses;
   responses.randn(800);
 
-  LinearRegression lr(data, responses, 0.05); // Train the model.
-  LinearRegression xmlLr, jsonLr, binaryLr;
+  LinearRegression<> lr(data, responses, 0.05); // Train the model.
+  LinearRegression<> xmlLr, jsonLr, binaryLr;
 
   SerializeObjectAll(lr, xmlLr, jsonLr, binaryLr);
 
@@ -260,8 +269,157 @@ TEST_CASE("LinearRegressionTrainReturnOb
         dot(coeffs.rows(1, 3), arma::ones<arma::rowvec>(3) * elem);
 
   // Initialize and predict.
-  LinearRegression lr;
+  LinearRegression<> lr;
   double error = lr.Train(predictors, responses);
 
   REQUIRE(std::isfinite(error) == true);
 }
+
+/**
+ * Make sure all versions of Train() work correctly.
+ */
+TEMPLATE_TEST_CASE("LinearRegressionAllTrainVersionsTest",
+    "[LinearRegressionTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+  typedef arma::Row<typename MatType::elem_type> RowType;
+
+  // The data doesn't really matter for this test; mostly we want to make sure
+  // that all the Train() variants work properly.
+  MatType predictors;
+  predictors = { {  0, 1, 2, 4, 8, 16 },
+                 { 16, 8, 4, 2, 1,  0 } };
+  RowType responses = "0 2 4 3 8 8";
+  RowType weights = "1.0 1.1 1.2 0.8 0.9 1.0";
+
+  LinearRegression<MatType> lr1, lr2, lr3, lr4, lr5, lr6;
+
+  (void) lr1.Train(predictors, responses);
+  (void) lr2.Train(predictors, responses, 0.1);
+  (void) lr3.Train(predictors, responses, 0.2, false);
+  (void) lr4.Train(predictors, responses, weights);
+  (void) lr5.Train(predictors, responses, weights, 0.3);
+  (void) lr6.Train(predictors, responses, weights, 0.4, false);
+
+  // We don't care about the specifics of the trained model, but we want to just
+  // make sure everything appears to be correct from the sizes and
+  // hyperparameters.
+  REQUIRE(lr1.Lambda() == Approx(0.0).margin(1e-10));
+  REQUIRE(lr1.Intercept() == true);
+  REQUIRE(lr1.Parameters().n_elem == 3);
+
+  REQUIRE(lr2.Lambda() == Approx(0.1).margin(1e-10));
+  REQUIRE(lr2.Intercept() == true);
+  REQUIRE(lr2.Parameters().n_elem == 3);
+
+  REQUIRE(lr3.Lambda() == Approx(0.2).margin(1e-10));
+  REQUIRE(lr3.Intercept() == false);
+  REQUIRE(lr3.Parameters().n_elem == 2);
+
+  REQUIRE(lr4.Lambda() == Approx(0.0).margin(1e-10));
+  REQUIRE(lr4.Intercept() == true);
+  REQUIRE(lr4.Parameters().n_elem == 3);
+
+  REQUIRE(lr5.Lambda() == Approx(0.3).margin(1e-10));
+  REQUIRE(lr5.Intercept() == true);
+  REQUIRE(lr5.Parameters().n_elem == 3);
+
+  REQUIRE(lr6.Lambda() == Approx(0.4).margin(1e-10));
+  REQUIRE(lr6.Intercept() == false);
+  REQUIRE(lr6.Parameters().n_elem == 2);
+
+  // We can also check that the weighted model is different from the unweighted
+  // model.
+  REQUIRE(!arma::approx_equal(lr1.Parameters(), lr4.Parameters(), "absdiff",
+      1e-5));
+}
+
+/**
+ * Ensure that single-point Predict() returns the same results as multi-point
+ * Predict().
+ */
+TEMPLATE_TEST_CASE("LinearRegressionSinglePointPredictTest",
+    "[LinearRegressionTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+  typedef arma::Row<typename MatType::elem_type> RowType;
+
+  MatType predictors;
+  predictors = { {  0, 1, 2, 4, 8, 16 },
+                 { 16, 8, 4, 2, 1,  0 } };
+  RowType responses = "0 2 4 3 8 8";
+
+  LinearRegression<MatType> lr(predictors, responses, 0.1, true);
+
+  // Compute predictions for test points in batch.
+  RowType predictions;
+  lr.Predict(predictors, predictions);
+
+  // Now compute each prediction individually.
+  for (size_t i = 0; i < predictors.n_cols; ++i)
+  {
+    const double prediction = lr.Predict(predictors.col(i));
+    REQUIRE(prediction == Approx(predictions[i]));
+  }
+}
+
+// Make sure training on submatrices and subvectors works.
+TEST_CASE("LinearRegressionSubmatrixTrainingTest", "[LinearRegressionTest]")
+{
+  // The quality of the model doesn't matter---mostly this is a compilation
+  // test.
+  arma::mat predictors(100, 1000, arma::fill::randu);
+  arma::rowvec responses(1000, arma::fill::randu);
+  arma::rowvec weights(1000, arma::fill::randu);
+
+  LinearRegression<> lr1(predictors.cols(0, 499), responses.subvec(0, 499));
+  LinearRegression<> lr2;
+  lr2.Train(predictors.cols(0, 499), responses.subvec(0, 499));
+  LinearRegression<> lr3(predictors.cols(0, 499), responses.subvec(0, 499),
+      weights.subvec(0, 499));
+  LinearRegression<> lr4;
+  lr4.Train(predictors.cols(0, 499), responses.subvec(0, 499),
+      weights.subvec(0, 499));
+
+  REQUIRE(lr1.Parameters().n_elem == 101);
+  REQUIRE(lr2.Parameters().n_elem == 101);
+  REQUIRE(lr3.Parameters().n_elem == 101);
+  REQUIRE(lr4.Parameters().n_elem == 101);
+
+  arma::rowvec predictions;
+
+  lr1.Predict(predictors.cols(500, 999), predictions);
+  REQUIRE(predictions.n_cols == 500);
+
+  lr2.Predict(predictors.cols(500, 999), predictions);
+  REQUIRE(predictions.n_cols == 500);
+
+  lr3.Predict(predictors.cols(500, 999), predictions);
+  REQUIRE(predictions.n_cols == 500);
+
+  lr4.Predict(predictors.cols(500, 999), predictions);
+  REQUIRE(predictions.n_cols == 500);
+}
+
+// Make sure we can train on sparse data.
+TEST_CASE("LinearRegressionSparseTrainingTest", "[LinearRegressionTest]")
+{
+  // For this test the quality of the model doesn't matter---mostly this is a
+  // compilation test, but we check that the sizes of the returned predictions
+  // and the sizes of the learned models are correct.
+
+  // Generate sparse random data.
+  arma::sp_mat data;
+  data.sprandu(100, 5000, 0.3);
+
+  arma::rowvec responses(5000, arma::fill::randu);
+
+  LinearRegression<> lr(data, responses);
+
+  REQUIRE(lr.Parameters().n_elem == 101);
+
+  arma::rowvec predictions;
+  lr.Predict(data, predictions);
+
+  REQUIRE(predictions.n_elem == 5000);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/linear_svm_test.cpp 4.4.0-1/src/mlpack/tests/linear_svm_test.cpp
--- 4.3.0-2/src/mlpack/tests/linear_svm_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/linear_svm_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -221,7 +221,7 @@ TEST_CASE("LinearSVMFunctionRegularizati
     parameters.randu(inputSize, numClasses);
 
     double wL2SquaredNorm;
-    wL2SquaredNorm = arma::dot(parameters, parameters);
+    wL2SquaredNorm = dot(parameters, parameters);
 
     // Calculate regularization terms.
     const double smallRegTerm = 0.5 * wL2SquaredNorm;
@@ -312,7 +312,7 @@ TEST_CASE("LinearSVMFunctionRegularizati
     parameters.randu(inputSize, numClasses);
 
     double wL2SquaredNorm;
-    wL2SquaredNorm = 0.5 * arma::dot(parameters, parameters);
+    wL2SquaredNorm = 0.5 * dot(parameters, parameters);
 
     // Calculate regularization terms.
     const double smallRegTerm = 0.5 * wL2SquaredNorm;
@@ -486,7 +486,7 @@ TEST_CASE("LinearSVMLBFGSSimpleTest", "[
 
   // Compare training accuracy to 1.
   const double acc = lsvm.ComputeAccuracy(dataset, labels);
-  REQUIRE(acc == Approx(1.0).epsilon(0.005));
+  REQUIRE(acc == Approx(100.0).epsilon(0.005));
 }
 
 /**
@@ -514,12 +514,12 @@ TEST_CASE("LinearSVMGradientDescentSimpl
 
   // Create a linear svm object using custom gradient descent optimizer.
   ens::GradientDescent optimizer(stepSize, maxIterations, tolerance);
-  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, lambda,
-      delta, false, optimizer);
+  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, optimizer, lambda,
+      delta, false);
 
   // Compare training accuracy to 1.
   const double acc = lsvm.ComputeAccuracy(dataset, labels);
-  REQUIRE(acc == Approx(1.0).epsilon(0.005));
+  REQUIRE(acc == Approx(100.0).epsilon(0.005));
 }
 
 /**
@@ -632,8 +632,7 @@ TEST_CASE("LinearSVMFitIntercept", "[Lin
     }
 
     // Now train a svm object on it.
-    LinearSVM<arma::mat> svm(data, labels, numClasses, lambda,
-        delta, true, ens::L_BFGS());
+    LinearSVM<arma::mat> svm(data, labels, numClasses, lambda, delta, true);
 
     // Ensure that the error is close to zero.
     const double acc = svm.ComputeAccuracy(data, labels);
@@ -775,12 +774,12 @@ TEST_CASE("LinearSVMPSGDSimpleTest", "[L
   ens::ParallelSGD<ens::ConstantStep> optimizer(0,
       std::ceil((float) dataset.n_cols / omp_get_max_threads()),
       1e-5, true, decayPolicy);
-  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, lambda,
-      delta, false, optimizer);
+  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, optimizer, lambda,
+      delta, false);
 
   // Compare training accuracy to 1.
   const double acc = lsvm.ComputeAccuracy(dataset, labels);
-  REQUIRE(acc == Approx(1.0).epsilon(1e-2));
+  REQUIRE(acc == Approx(100.0).epsilon(1e-2));
 }
 
 /**
@@ -824,13 +823,13 @@ TEST_CASE("LinearSVMParallelSGDTwoClasse
 
     // Train linear svm object using Parallel SGD optimizer.
     // The threadShareSize is chosen such that each function gets optimized.
-    ens::ParallelSGD<ens::ConstantStep> optimizer(0,
+    ens::ParallelSGD<ens::ConstantStep> optimizer(100000,
         std::ceil((float) data.n_cols / omp_get_max_threads()),
         1e-5, true, decayPolicy);
-    LinearSVM<arma::mat> lsvm(data, labels, numClasses, lambda,
-        delta, false, optimizer);
+    LinearSVM<arma::mat> lsvm(data, labels, numClasses, optimizer, lambda,
+        delta, false);
 
-    // Compare training accuracy to 1.
+    // Compare training accuracy to 100.
     const double acc = lsvm.ComputeAccuracy(data, labels);
 
     // Create test dataset.
@@ -849,8 +848,8 @@ TEST_CASE("LinearSVMParallelSGDTwoClasse
     const double testAcc = lsvm.ComputeAccuracy(data, labels);
 
     // Larger tolerance is sometimes needed.
-    if (testAcc == Approx(1.0).epsilon(0.02) &&
-        acc == Approx(1.0).epsilon(0.02))
+    if (testAcc == Approx(100.0).epsilon(0.02) &&
+        acc == Approx(100.0).epsilon(0.02))
     {
       success = true;
       break;
@@ -863,23 +862,28 @@ TEST_CASE("LinearSVMParallelSGDTwoClasse
 #endif
 
 /**
- * Test sparse and dense linear svm and make sure they both work the
+ * Test sparse and dense linear svm training and make sure they both work the
  * same using the L-BFGS optimizer.
  */
-TEST_CASE("LinearSVMSparseLBFGSTest", "[LinearSVMTest]")
+TEMPLATE_TEST_CASE("LinearSVMSparseLBFGSTest", "[LinearSVMTest]", float, double)
 {
+  typedef TestType ElemType;
+  typedef typename arma::SpMat<ElemType> SparseMatType;
+  typedef typename arma::Mat<ElemType> MatType;
+
   // Create a random dataset.
-  arma::sp_mat dataset;
+  SparseMatType dataset;
   dataset.sprandu(10, 800, 0.3);
-  arma::mat denseDataset(dataset);
+  MatType denseDataset(dataset);
   arma::Row<size_t> labels(800);
   for (size_t i = 0; i < 800; ++i)
     labels[i] = RandInt(0, 2);
 
-  LinearSVM<arma::mat> lr(denseDataset, labels, 2, 0.3, 1,
-      false, ens::L_BFGS());
-  LinearSVM<arma::sp_mat> lrSparse(dataset, labels, 2, 0.3, 1,
-      false, ens::L_BFGS());
+  LinearSVM<> lr(denseDataset, labels, 2, 0.3, 1, false);
+  LinearSVM<> lrSparse(dataset, labels, 2, 0.3, 1, false);
+
+  // Make the initial points the same.
+  lrSparse.Parameters() = lr.Parameters();
 
   REQUIRE(lr.Parameters().n_elem == lrSparse.Parameters().n_elem);
   for (size_t i = 0; i < lr.Parameters().n_elem; ++i)
@@ -891,10 +895,15 @@ TEST_CASE("LinearSVMSparseLBFGSTest", "[
 
 /**
  * Test training of linear svm for multiple classes on a complex gaussian
- * dataset using L-BFGS optimizer.
+ * dataset using L-BFGS optimizer, with different types.
  */
-TEST_CASE("LinearSVMLBFGSMultipleClasses", "[LinearSVMTest]")
+TEMPLATE_TEST_CASE("LinearSVMLBFGSMultipleClasses", "[LinearSVMTest]", float,
+    double)
 {
+  typedef TestType ElemType;
+  typedef typename arma::Mat<ElemType> MatType;
+  typedef typename arma::Col<ElemType> VecType;
+
   const size_t points = 1000;
   const size_t inputSize = 5;
   const size_t numClasses = 5;
@@ -908,7 +917,7 @@ TEST_CASE("LinearSVMLBFGSMultipleClasses
   GaussianDistribution g4(arma::vec("4.0 1.0 1.0 2.0 7.0"), identity);
   GaussianDistribution g5(arma::vec("1.0 0.0 1.0 8.0 3.0"), identity);
 
-  arma::mat data(inputSize, points);
+  MatType data(inputSize, points);
   arma::Row<size_t> labels(points);
 
   // This loop can be removed when ensmallen PR #136 is merged into a version
@@ -921,32 +930,32 @@ TEST_CASE("LinearSVMLBFGSMultipleClasses
   {
     for (size_t i = 0; i < points / 5; ++i)
     {
-      data.col(i) = g1.Random();
+      data.col(i) = ConvTo<VecType>::From(g1.Random());
       labels(i) = 0;
     }
     for (size_t i = points / 5; i < (2 * points) / 5; ++i)
     {
-      data.col(i) = g2.Random();
+      data.col(i) = ConvTo<VecType>::From(g2.Random());
       labels(i) = 1;
     }
     for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
     {
-      data.col(i) = g3.Random();
+      data.col(i) = ConvTo<VecType>::From(g3.Random());
       labels(i) = 2;
     }
     for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
     {
-      data.col(i) = g4.Random();
+      data.col(i) = ConvTo<VecType>::From(g4.Random());
       labels(i) = 3;
     }
     for (size_t i = (4 * points) / 5; i < points; ++i)
     {
-      data.col(i) = g5.Random();
+      data.col(i) = ConvTo<VecType>::From(g5.Random());
       labels(i) = 4;
     }
 
     // Train linear svm object using L-BFGS optimizer.
-    LinearSVM<arma::mat> lsvm(data, labels, numClasses, lambda);
+    LinearSVM<MatType> lsvm(data, labels, numClasses, lambda);
 
     // Compare training accuracy to 1.
     const double acc = lsvm.ComputeAccuracy(data, labels);
@@ -956,27 +965,27 @@ TEST_CASE("LinearSVMLBFGSMultipleClasses
     // Create test dataset.
     for (size_t i = 0; i < points / 5; ++i)
     {
-      data.col(i) = g1.Random();
+      data.col(i) = ConvTo<VecType>::From(g1.Random());
       labels(i) = 0;
     }
     for (size_t i = points / 5; i < (2 * points) / 5; ++i)
     {
-      data.col(i) = g2.Random();
+      data.col(i) = ConvTo<VecType>::From(g2.Random());
       labels(i) = 1;
     }
     for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
     {
-      data.col(i) = g3.Random();
+      data.col(i) = ConvTo<VecType>::From(g3.Random());
       labels(i) = 2;
     }
     for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
     {
-      data.col(i) = g4.Random();
+      data.col(i) = ConvTo<VecType>::From(g4.Random());
       labels(i) = 3;
     }
     for (size_t i = (4 * points) / 5; i < points; ++i)
     {
-      data.col(i) = g5.Random();
+      data.col(i) = ConvTo<VecType>::From(g5.Random());
       labels(i) = 4;
     }
 
@@ -995,8 +1004,13 @@ TEST_CASE("LinearSVMLBFGSMultipleClasses
 /**
  * Testing single point classification (Classify()).
  */
-TEST_CASE("LinearSVMClassifySinglePointTest", "[LinearSVMTest]")
+TEMPLATE_TEST_CASE("LinearSVMClassifySinglePointTest", "[LinearSVMTest]", float,
+    double)
 {
+  typedef TestType ElemType;
+  typedef typename arma::Mat<ElemType> MatType;
+  typedef typename arma::Col<ElemType> VecType;
+
   const size_t points = 500;
   const size_t inputSize = 5;
   const size_t numClasses = 5;
@@ -1010,70 +1024,79 @@ TEST_CASE("LinearSVMClassifySinglePointT
   GaussianDistribution g4(arma::vec("4.0 1.0 1.0 2.0 7.0"), identity);
   GaussianDistribution g5(arma::vec("1.0 0.0 1.0 8.0 3.0"), identity);
 
-  arma::mat data(inputSize, points);
+  MatType data(inputSize, points);
   arma::Row<size_t> labels(points);
 
   for (size_t i = 0; i < points / 5; ++i)
   {
-    data.col(i) = g1.Random();
+    data.col(i) = ConvTo<VecType>::From(g1.Random());
     labels(i) = 0;
   }
   for (size_t i = points / 5; i < (2 * points) / 5; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<VecType>::From(g2.Random());
     labels(i) = 1;
   }
   for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
   {
-    data.col(i) = g3.Random();
+    data.col(i) = ConvTo<VecType>::From(g3.Random());
     labels(i) = 2;
   }
   for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
   {
-    data.col(i) = g4.Random();
+    data.col(i) = ConvTo<VecType>::From(g4.Random());
     labels(i) = 3;
   }
   for (size_t i = (4 * points) / 5; i < points; ++i)
   {
-    data.col(i) = g5.Random();
+    data.col(i) = ConvTo<VecType>::From(g5.Random());
     labels(i) = 4;
   }
 
   // Train linear svm object.
-  LinearSVM<arma::mat> lsvm(data, labels, numClasses, lambda);
+  LinearSVM<MatType> lsvm(data, labels, numClasses, lambda);
 
   // Create test dataset.
   for (size_t i = 0; i < points / 5; ++i)
   {
-    data.col(i) = g1.Random();
+    data.col(i) = ConvTo<VecType>::From(g1.Random());
     labels(i) = 0;
   }
   for (size_t i = points / 5; i < (2 * points) / 5; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<VecType>::From(g2.Random());
     labels(i) = 1;
   }
   for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
   {
-    data.col(i) = g3.Random();
+    data.col(i) = ConvTo<VecType>::From(g3.Random());
     labels(i) = 2;
   }
   for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
   {
-    data.col(i) = g4.Random();
+    data.col(i) = ConvTo<VecType>::From(g4.Random());
     labels(i) = 3;
   }
   for (size_t i = (4 * points) / 5; i < points; ++i)
   {
-    data.col(i) = g5.Random();
+    data.col(i) = ConvTo<VecType>::From(g5.Random());
     labels(i) = 4;
   }
 
-  lsvm.Classify(data, labels);
+  MatType scores;
+  lsvm.Classify(data, labels, scores);
 
   for (size_t i = 0; i < data.n_cols; ++i)
   {
     REQUIRE(lsvm.Classify(data.col(i)) == labels(i));
+
+    size_t prediction;
+    VecType scoresVec;
+    lsvm.Classify(data.col(i), prediction, scoresVec);
+
+    REQUIRE(prediction == labels(i));
+    REQUIRE(scoresVec.n_elem == scores.n_rows);
+    REQUIRE(arma::approx_equal(scoresVec, scores.col(i), "absdiff", 1e-5));
   }
 }
 
@@ -1188,8 +1211,259 @@ TEST_CASE("LinearSVMCallbackTest", "[Lin
   CallbackTestFunction cb;
 
   ens::L_BFGS opt;
-  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, lambda,
-      delta, false, opt, cb);
+  LinearSVM<arma::mat> lsvm(dataset, labels, numClasses, lambda, delta,
+      false, cb);
 
   REQUIRE(cb.calledEndOptimization == true);
 }
+
+// Test all variants of LinearSVM constructors.
+TEMPLATE_TEST_CASE("LinearSVMConstructorVariantTest", "[LinearSVMTest]",
+    arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+
+  // Create some random data.  The results here do not matter all that much;
+  // this is more of a test that all constructor variants successfully compile
+  // and produce models at all.
+  MatType dataset(10, 800, arma::fill::randu);
+  arma::Row<size_t> labels(800);
+  for (size_t i = 0; i < 800; ++i)
+    labels[i] = RandInt(0, 2);
+
+  LinearSVM<> lsvm1;
+  LinearSVM<> lsvm2(10, 2);
+  LinearSVM<> lsvm3(10, 2, 0.0002, 1.1, true);
+  LinearSVM<> lsvm4(dataset, labels, 2);
+  LinearSVM<> lsvm5(dataset, labels, 2, 0.0003, 1.2, true);
+  LinearSVM<> lsvm6(dataset, labels, 2, 0.0004, 1.3, true,
+      CallbackTestFunction());
+  LinearSVM<> lsvm7(dataset, labels, 2, 0.0005, 1.4, true,
+      CallbackTestFunction(), ens::TimerStop(1.0));
+
+  ens::StandardSGD sgd;
+  LinearSVM<> lsvm8(dataset, labels, 2, sgd);
+  LinearSVM<> lsvm9(dataset, labels, 2, sgd, 0.0006, 1.5, true);
+  LinearSVM<> lsvm10(dataset, labels, 2, sgd, 0.0007, 1.6, true,
+      CallbackTestFunction());
+  LinearSVM<> lsvm11(dataset, labels, 2, sgd, 0.0008, 1.7, true,
+      CallbackTestFunction(), ens::TimerStop(1.0));
+
+  // Check that the variants that did not train have reasonable values.
+  REQUIRE(lsvm1.Lambda() == Approx(0.0001));
+  REQUIRE(lsvm1.Delta() == Approx(1.0));
+  REQUIRE(lsvm1.FitIntercept() == false);
+
+  REQUIRE(lsvm2.Lambda() == Approx(0.0001));
+  REQUIRE(lsvm2.Delta() == Approx(1.0));
+  REQUIRE(lsvm2.FitIntercept() == false);
+  REQUIRE(lsvm2.NumClasses() == 2);
+
+  REQUIRE(lsvm3.Lambda() == Approx(0.0002));
+  REQUIRE(lsvm3.Delta() == Approx(1.1));
+  REQUIRE(lsvm3.FitIntercept() == true);
+  REQUIRE(lsvm3.NumClasses() == 2);
+
+  // Now check that the variants that trained have reasonable models of the
+  // right size.
+  REQUIRE(lsvm4.Lambda() == Approx(0.0001));
+  REQUIRE(lsvm4.Delta() == Approx(1.0));
+  REQUIRE(lsvm4.FitIntercept() == false);
+  REQUIRE(lsvm4.FeatureSize() == 10);
+  REQUIRE(lsvm4.NumClasses() == 2);
+
+  REQUIRE(lsvm5.Lambda() == Approx(0.0003));
+  REQUIRE(lsvm5.Delta() == Approx(1.2));
+  REQUIRE(lsvm5.FitIntercept() == true);
+  REQUIRE(lsvm5.FeatureSize() == 10);
+  REQUIRE(lsvm5.NumClasses() == 2);
+
+  REQUIRE(lsvm6.Lambda() == Approx(0.0004));
+  REQUIRE(lsvm6.Delta() == Approx(1.3));
+  REQUIRE(lsvm6.FitIntercept() == true);
+  REQUIRE(lsvm6.FeatureSize() == 10);
+  REQUIRE(lsvm6.NumClasses() == 2);
+
+  REQUIRE(lsvm7.Lambda() == Approx(0.0005));
+  REQUIRE(lsvm7.Delta() == Approx(1.4));
+  REQUIRE(lsvm7.FitIntercept() == true);
+  REQUIRE(lsvm7.FeatureSize() == 10);
+  REQUIRE(lsvm7.NumClasses() == 2);
+
+  REQUIRE(lsvm8.Lambda() == Approx(0.0001));
+  REQUIRE(lsvm8.Delta() == Approx(1.0));
+  REQUIRE(lsvm8.FitIntercept() == false);
+  REQUIRE(lsvm8.FeatureSize() == 10);
+  REQUIRE(lsvm8.NumClasses() == 2);
+
+  REQUIRE(lsvm9.Lambda() == Approx(0.0006));
+  REQUIRE(lsvm9.Delta() == Approx(1.5));
+  REQUIRE(lsvm9.FitIntercept() == true);
+  REQUIRE(lsvm9.FeatureSize() == 10);
+  REQUIRE(lsvm9.NumClasses() == 2);
+
+  REQUIRE(lsvm10.Lambda() == Approx(0.0007));
+  REQUIRE(lsvm10.Delta() == Approx(1.6));
+  REQUIRE(lsvm10.FitIntercept() == true);
+  REQUIRE(lsvm10.FeatureSize() == 10);
+  REQUIRE(lsvm10.NumClasses() == 2);
+
+  REQUIRE(lsvm11.Lambda() == Approx(0.0008));
+  REQUIRE(lsvm11.Delta() == Approx(1.7));
+  REQUIRE(lsvm11.FitIntercept() == true);
+  REQUIRE(lsvm11.FeatureSize() == 10);
+  REQUIRE(lsvm11.NumClasses() == 2);
+}
+
+// Test all variants of LinearSVM Train() functions.
+TEMPLATE_TEST_CASE("LinearSVMTrainVariantTest", "[LinearSVMTest]", arma::fmat,
+    arma::mat)
+{
+  typedef TestType MatType;
+
+  // Create some random data.  The results here do not matter all that much;
+  // this is more of a test that all constructor variants successfully compile
+  // and produce models at all.
+  MatType dataset(10, 800, arma::fill::randu);
+  arma::Row<size_t> labels(800);
+  for (size_t i = 0; i < 800; ++i)
+    labels[i] = RandInt(0, 2);
+
+  LinearSVM<> lsvm1(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm2(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm3(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm4(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm5(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm6(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm7(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm8(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm9(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm10(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm11(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm12(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm13(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm14(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm15(10, 2, 1.2, 0.5, true);
+  LinearSVM<> lsvm16(10, 2, 1.2, 0.5, true);
+
+  lsvm1.Train(dataset, labels, 2);
+  lsvm2.Train(dataset, labels, 2, CallbackTestFunction());
+  lsvm3.Train(dataset, labels, 2, CallbackTestFunction(), ens::TimerStop(1.0));
+  lsvm4.Train(dataset, labels, 2, 0.0002);
+  lsvm5.Train(dataset, labels, 2, 0.0003, 1.1);
+  lsvm6.Train(dataset, labels, 2, 0.0004, 1.2, false);
+  lsvm7.Train(dataset, labels, 2, 0.0005, 1.3, true, CallbackTestFunction());
+  lsvm8.Train(dataset, labels, 2, 0.0006, 1.4, false, CallbackTestFunction(),
+      ens::TimerStop(1.0));
+
+  ens::Adam adam;
+  lsvm9.Train(dataset, labels, 2, adam);
+  lsvm10.Train(dataset, labels, 2, adam, CallbackTestFunction());
+  lsvm11.Train(dataset, labels, 2, adam, CallbackTestFunction(),
+      ens::TimerStop(1.0));
+  lsvm12.Train(dataset, labels, 2, adam, 0.0007);
+  lsvm13.Train(dataset, labels, 2, adam, 0.0008, 1.5);
+  lsvm14.Train(dataset, labels, 2, adam, 0.0009, 1.6, false);
+  lsvm15.Train(dataset, labels, 2, adam, 0.001, 1.7, true,
+      CallbackTestFunction());
+  lsvm16.Train(dataset, labels, 2, adam, 0.0011, 1.8, false,
+      CallbackTestFunction(), ens::TimerStop(1.0));
+
+  // Check that all hyperparameters are set as expected, and that the model has
+  // the correct size.
+  REQUIRE(lsvm1.Lambda() == Approx(1.2));
+  REQUIRE(lsvm1.Delta() == Approx(0.5));
+  REQUIRE(lsvm1.FitIntercept() == true);
+  REQUIRE(lsvm1.FeatureSize() == 10);
+  REQUIRE(lsvm1.NumClasses() == 2);
+
+  REQUIRE(lsvm2.Lambda() == Approx(1.2));
+  REQUIRE(lsvm2.Delta() == Approx(0.5));
+  REQUIRE(lsvm2.FitIntercept() == true);
+  REQUIRE(lsvm2.FeatureSize() == 10);
+  REQUIRE(lsvm2.NumClasses() == 2);
+
+  REQUIRE(lsvm3.Lambda() == Approx(1.2));
+  REQUIRE(lsvm3.Delta() == Approx(0.5));
+  REQUIRE(lsvm3.FitIntercept() == true);
+  REQUIRE(lsvm3.FeatureSize() == 10);
+  REQUIRE(lsvm3.NumClasses() == 2);
+
+  REQUIRE(lsvm4.Lambda() == Approx(0.0002));
+  REQUIRE(lsvm4.Delta() == Approx(0.5));
+  REQUIRE(lsvm4.FitIntercept() == true);
+  REQUIRE(lsvm4.FeatureSize() == 10);
+  REQUIRE(lsvm4.NumClasses() == 2);
+
+  REQUIRE(lsvm5.Lambda() == Approx(0.0003));
+  REQUIRE(lsvm5.Delta() == Approx(1.1));
+  REQUIRE(lsvm5.FitIntercept() == true);
+  REQUIRE(lsvm5.FeatureSize() == 10);
+  REQUIRE(lsvm5.NumClasses() == 2);
+
+  REQUIRE(lsvm6.Lambda() == Approx(0.0004));
+  REQUIRE(lsvm6.Delta() == Approx(1.2));
+  REQUIRE(lsvm6.FitIntercept() == false);
+  REQUIRE(lsvm6.FeatureSize() == 10);
+  REQUIRE(lsvm6.NumClasses() == 2);
+
+  REQUIRE(lsvm7.Lambda() == Approx(0.0005));
+  REQUIRE(lsvm7.Delta() == Approx(1.3));
+  REQUIRE(lsvm7.FitIntercept() == true);
+  REQUIRE(lsvm7.FeatureSize() == 10);
+  REQUIRE(lsvm7.NumClasses() == 2);
+
+  REQUIRE(lsvm8.Lambda() == Approx(0.0006));
+  REQUIRE(lsvm8.Delta() == Approx(1.4));
+  REQUIRE(lsvm8.FitIntercept() == false);
+  REQUIRE(lsvm8.FeatureSize() == 10);
+  REQUIRE(lsvm8.NumClasses() == 2);
+
+  REQUIRE(lsvm9.Lambda() == Approx(1.2));
+  REQUIRE(lsvm9.Delta() == Approx(0.5));
+  REQUIRE(lsvm9.FitIntercept() == true);
+  REQUIRE(lsvm9.FeatureSize() == 10);
+  REQUIRE(lsvm9.NumClasses() == 2);
+
+  REQUIRE(lsvm10.Lambda() == Approx(1.2));
+  REQUIRE(lsvm10.Delta() == Approx(0.5));
+  REQUIRE(lsvm10.FitIntercept() == true);
+  REQUIRE(lsvm10.FeatureSize() == 10);
+  REQUIRE(lsvm10.NumClasses() == 2);
+
+  REQUIRE(lsvm11.Lambda() == Approx(1.2));
+  REQUIRE(lsvm11.Delta() == Approx(0.5));
+  REQUIRE(lsvm11.FitIntercept() == true);
+  REQUIRE(lsvm11.FeatureSize() == 10);
+  REQUIRE(lsvm11.NumClasses() == 2);
+
+  REQUIRE(lsvm12.Lambda() == Approx(0.0007));
+  REQUIRE(lsvm12.Delta() == Approx(0.5));
+  REQUIRE(lsvm12.FitIntercept() == true);
+  REQUIRE(lsvm12.FeatureSize() == 10);
+  REQUIRE(lsvm12.NumClasses() == 2);
+
+  REQUIRE(lsvm13.Lambda() == Approx(0.0008));
+  REQUIRE(lsvm13.Delta() == Approx(1.5));
+  REQUIRE(lsvm13.FitIntercept() == true);
+  REQUIRE(lsvm13.FeatureSize() == 10);
+  REQUIRE(lsvm13.NumClasses() == 2);
+
+  REQUIRE(lsvm14.Lambda() == Approx(0.0009));
+  REQUIRE(lsvm14.Delta() == Approx(1.6));
+  REQUIRE(lsvm14.FitIntercept() == false);
+  REQUIRE(lsvm14.FeatureSize() == 10);
+  REQUIRE(lsvm14.NumClasses() == 2);
+
+  REQUIRE(lsvm15.Lambda() == Approx(0.001));
+  REQUIRE(lsvm15.Delta() == Approx(1.7));
+  REQUIRE(lsvm15.FitIntercept() == true);
+  REQUIRE(lsvm15.FeatureSize() == 10);
+  REQUIRE(lsvm15.NumClasses() == 2);
+
+  REQUIRE(lsvm16.Lambda() == Approx(0.0011));
+  REQUIRE(lsvm16.Delta() == Approx(1.8));
+  REQUIRE(lsvm16.FitIntercept() == false);
+  REQUIRE(lsvm16.FeatureSize() == 10);
+  REQUIRE(lsvm16.NumClasses() == 2);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/lmnn_test.cpp 4.4.0-1/src/mlpack/tests/lmnn_test.cpp
--- 4.3.0-2/src/mlpack/tests/lmnn_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/lmnn_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -411,7 +411,7 @@ double KnnAccuracy(const arma::mat& data
       Map(labels(neighbors(j, i))) +=
           1 / std::pow(distances(j, i) + 1, 2);
 
-    size_t index = arma::conv_to<size_t>::from(arma::find(Map
+    size_t index = ConvTo<size_t>::From(arma::find(Map
         == arma::max(Map)));
 
     // Increase count if labels match.
diff -pruN 4.3.0-2/src/mlpack/tests/load_save_test.cpp 4.4.0-1/src/mlpack/tests/load_save_test.cpp
--- 4.3.0-2/src/mlpack/tests/load_save_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/load_save_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -193,15 +193,74 @@ TEST_CASE("LoadSparseTXTTest", "[LoadSav
   double temp = 0.1;
   for (int i = 0; it != it_end; ++it, temp += 0.1, ++i)
   {
-    REQUIRE((double)(*it) == Approx(temp).epsilon(1e-7));
-    REQUIRE((int)(it.row()) == i + 1);
-    REQUIRE((int)it.col() == i + 2);
+    REQUIRE((double) (*it) == Approx(temp).epsilon(1e-7));
+    REQUIRE((int) (it.row()) == i + 1);
+    REQUIRE((int) it.col() == i + 2);
   }
   // Remove the file.
   remove("test_sparse_file.txt");
 }
 
 /**
+ * Make sure sparse coordinate list autodetection works.
+ */
+TEST_CASE("LoadSparseAutodetectTest", "[LoadSaveTest]")
+{
+  fstream f;
+  f.open("test_file.csv", fstream::out);
+
+  f << "1, 3, 4.0" << endl;
+  f << "2, 4, 5.0" << endl;
+  f << "3, 6, -3.0" << endl;
+
+  f.close();
+
+  arma::sp_mat test;
+
+  REQUIRE(data::Load("test_file.csv", test, true) == true);
+
+  REQUIRE(test.n_rows == 7);
+  REQUIRE(test.n_cols == 4);
+
+  REQUIRE(test.at(3, 1) == Approx(4.0));
+  REQUIRE(test.at(4, 2) == Approx(5.0));
+  REQUIRE(test.at(6, 3) == Approx(-3.0));
+
+  remove("test_file.csv");
+}
+
+/**
+ * Make sure sparse coordinate list autodetection fails when the number of
+ * columns is wrong.
+ */
+TEST_CASE("LoadSparseAutodetectNotCoordinateListTest", "[LoadSaveTest]")
+{
+  fstream f;
+  f.open("test_file.csv", fstream::out);
+
+  f << "1, 0, 0, 4" << endl;
+  f << "0, 1, 0, 3" << endl;
+  f << "1, 0, 0, 0" << endl;
+
+  f.close();
+
+  arma::sp_mat test;
+
+  REQUIRE(data::Load("test_file.csv", test, true) == true);
+
+  REQUIRE(test.n_rows == 4);
+  REQUIRE(test.n_cols == 3);
+
+  REQUIRE(test.at(0, 0) == Approx(1.0));
+  REQUIRE(test.at(3, 0) == Approx(4.0));
+  REQUIRE(test.at(1, 1) == Approx(1.0));
+  REQUIRE(test.at(3, 1) == Approx(3.0));
+  REQUIRE(test.at(0, 2) == Approx(1.0));
+
+  remove("test_file.csv");
+}
+
+/**
  * Make sure a TSV is loaded correctly.
  */
 TEST_CASE("LoadTSVTest", "[LoadSaveTest]")
@@ -402,8 +461,8 @@ TEST_CASE("SaveSparseBinaryTest", "[Load
   {
     double val = (*it);
     REQUIRE(val == Approx(temp).epsilon(1e-7));
-    REQUIRE((int)(it.row()) == i);
-    REQUIRE((int)it.col() == i);
+    REQUIRE((int) (it.row()) == i);
+    REQUIRE((int) it.col() == i);
   }
 
   // Remove the file.
diff -pruN 4.3.0-2/src/mlpack/tests/local_coordinate_coding_test.cpp 4.4.0-1/src/mlpack/tests/local_coordinate_coding_test.cpp
--- 4.3.0-2/src/mlpack/tests/local_coordinate_coding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/local_coordinate_coding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -18,7 +18,10 @@
 using namespace arma;
 using namespace mlpack;
 
-void VerifyCorrectness(const vec& beta, const vec& errCorr, double lambda)
+template<typename MatType, typename VecType>
+void VerifyCorrectness(const MatType& beta,
+                       const VecType& errCorr,
+                       const double lambda)
 {
   const double tol = 0.1;
   size_t nDims = beta.n_elem;
@@ -43,15 +46,18 @@ void VerifyCorrectness(const vec& beta,
   }
 }
 
-
-TEST_CASE("LocalCoordinateCodingTestCodingStep",
-          "[LocalCoordinateCodingTest]")
+TEMPLATE_TEST_CASE("LocalCoordinateCodingTestCodingStep",
+    "[LocalCoordinateCodingTest]", arma::mat, arma::fmat)
 {
+  typedef TestType MatType;
+  typedef arma::Col<typename MatType::elem_type> VecType;
+
   double lambda1 = 0.1;
   uword nAtoms = 10;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  mat inX; // The .arm file is saved as an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // normalize each point since these are images
@@ -60,37 +66,40 @@ TEST_CASE("LocalCoordinateCodingTestCodi
     X.col(i) /= norm(X.col(i), 2);
   }
 
-  mat Z;
-  LocalCoordinateCoding lcc(X, nAtoms, lambda1, 10);
+  MatType Z;
+  LocalCoordinateCoding<MatType> lcc(X, nAtoms, lambda1, 10);
   lcc.Encode(X, Z);
 
-  mat D = lcc.Dictionary();
+  MatType D = lcc.Dictionary();
 
   for (uword i = 0; i < nPoints; ++i)
   {
-    vec sqDists = vec(nAtoms);
+    VecType sqDists(nAtoms);
     for (uword j = 0; j < nAtoms; ++j)
     {
       sqDists[j] = arma::norm(D.col(j) - X.col(i));
     }
-    mat Dprime = D * diagmat(1.0 / sqDists);
-    mat zPrime = Z.unsafe_col(i) % sqDists;
+    MatType Dprime = D * diagmat(1.0 / sqDists);
+    MatType zPrime = Z.unsafe_col(i) % sqDists;
 
-    vec errCorr = trans(Dprime) * (Dprime * zPrime - X.unsafe_col(i));
+    VecType errCorr = trans(Dprime) * (Dprime * zPrime - X.unsafe_col(i));
     VerifyCorrectness(zPrime, errCorr, 0.5 * lambda1);
   }
 }
 
-TEST_CASE("LocalCoordinateCodingTestDictionaryStep",
-          "[LocalCoordinateCodingTest]")
+TEMPLATE_TEST_CASE("LocalCoordinateCodingTestDictionaryStep",
+    "[LocalCoordinateCodingTest]", arma::mat, arma::fmat)
 {
+  typedef TestType MatType;
+
   const double tol = 0.1;
 
   double lambda = 0.1;
   uword nAtoms = 10;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  mat inX; // File is saved as an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // normalize each point since these are images
@@ -99,15 +108,15 @@ TEST_CASE("LocalCoordinateCodingTestDict
     X.col(i) /= norm(X.col(i), 2);
   }
 
-  mat Z;
-  LocalCoordinateCoding lcc(X, nAtoms, lambda, 10);
+  MatType Z;
+  LocalCoordinateCoding<MatType> lcc(X, nAtoms, lambda, 10);
   lcc.Encode(X, Z);
   uvec adjacencies = find(Z);
   lcc.OptimizeDictionary(X, Z, adjacencies);
 
-  mat D = lcc.Dictionary();
+  MatType D = lcc.Dictionary();
 
-  mat grad = zeros(D.n_rows, D.n_cols);
+  MatType grad = zeros<MatType>(D.n_rows, D.n_cols);
   for (uword i = 0; i < nPoints; ++i)
   {
     grad += (D - repmat(X.unsafe_col(i), 1, nAtoms)) *
@@ -118,26 +127,30 @@ TEST_CASE("LocalCoordinateCodingTestDict
   REQUIRE(norm(grad, "fro") == Approx(0.0).margin(tol));
 }
 
-TEST_CASE("LocalCoordinateCodingSerializationTest",
-          "[LocalCoordinateCodingTest]")
+TEMPLATE_TEST_CASE("LocalCoordinateCodingSerializationTest",
+    "[LocalCoordinateCodingTest]", arma::mat, arma::fmat)
 {
-  mat X = randu<mat>(100, 100);
+  typedef TestType MatType;
+
+  MatType X = randu<MatType>(100, 100);
   size_t nAtoms = 10;
 
-  LocalCoordinateCoding lcc(nAtoms, 0.05, 2 /* don't care about quality */);
+  LocalCoordinateCoding<MatType> lcc(nAtoms, 0.05,
+      2 /* don't care about quality */);
   lcc.Train(X);
 
-  mat Y = randu<mat>(100, 200);
-  mat codes;
+  MatType Y = randu<MatType>(100, 200);
+  MatType codes;
   lcc.Encode(Y, codes);
 
-  LocalCoordinateCoding lccXml(50, 0.1), lccJson(12, 0.0), lccBinary(0, 0.0);
+  LocalCoordinateCoding<MatType> lccXml(50, 0.1), lccJson(12, 0.0),
+      lccBinary(0, 0.0);
   SerializeObjectAll(lcc, lccXml, lccJson, lccBinary);
 
   CheckMatrices(lcc.Dictionary(), lccXml.Dictionary(), lccJson.Dictionary(),
       lccBinary.Dictionary());
 
-  mat xmlCodes, jsonCodes, binaryCodes;
+  MatType xmlCodes, jsonCodes, binaryCodes;
   lccXml.Encode(Y, xmlCodes);
   lccJson.Encode(Y, jsonCodes);
   lccBinary.Encode(Y, binaryCodes);
@@ -167,14 +180,17 @@ TEST_CASE("LocalCoordinateCodingSerializ
  * Test that LocalCoordinateCoding::Train() returns finite final objective
  * value.
  */
-TEST_CASE("LocalCoordinateCodingTrainReturnObjective",
-          "[LocalCoordinateCodingTest]")
+TEMPLATE_TEST_CASE("LocalCoordinateCodingTrainReturnObjective",
+    "[LocalCoordinateCodingTest]", arma::mat, arma::fmat)
 {
+  typedef TestType MatType;
+
   double lambda1 = 0.1;
   uword nAtoms = 10;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  mat inX; // File is saved as arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // Normalize each point since these are images.
@@ -183,7 +199,7 @@ TEST_CASE("LocalCoordinateCodingTrainRet
     X.col(i) /= norm(X.col(i), 2);
   }
 
-  LocalCoordinateCoding lcc(nAtoms, lambda1, 10);
+  LocalCoordinateCoding<MatType> lcc(nAtoms, lambda1, 10);
   double objVal = lcc.Train(X);
 
   REQUIRE(std::isfinite(objVal) == true);
diff -pruN 4.3.0-2/src/mlpack/tests/logistic_regression_test.cpp 4.4.0-1/src/mlpack/tests/logistic_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/logistic_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/logistic_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -79,7 +79,7 @@ TEST_CASE("LogisticRegressionFunctionRan
     for (size_t j = 0; j < points; ++j)
     {
       const double sigmoid = (1.0 / (1.0 + exp(-parameters[0] -
-          arma::dot(data.col(j), parameters.subvec(1, dimension)))));
+          dot(data.col(j), parameters.subvec(1, dimension)))));
       if (responses[j] == 1.0)
         loglikelihood += log(std::pow(sigmoid, responses[j]));
       else
@@ -510,7 +510,7 @@ TEST_CASE("LogisticRegressionLBFGSSimple
   LogisticRegression<> lr(data, responses);
 
   // Test sigmoid function.
-  arma::rowvec sigmoids = 1 / (1 + arma::exp(-lr.Parameters()[0]
+  arma::rowvec sigmoids = 1 / (1 + exp(-lr.Parameters()[0]
       - lr.Parameters().tail_cols(lr.Parameters().n_elem - 1) * data));
 
   // Large 0.1% error tolerance is because the optimizer may terminate before
@@ -534,7 +534,7 @@ TEST_CASE("LogisticRegressionSGDSimpleTe
   LogisticRegression<> lr(data, responses, sgd, 0.001);
 
   // Test sigmoid function.
-  arma::rowvec sigmoids = 1 / (1 + arma::exp(-lr.Parameters()[0]
+  arma::rowvec sigmoids = 1 / (1 + exp(-lr.Parameters()[0]
       - lr.Parameters().tail_cols(lr.Parameters().n_elem - 1) * data));
 
   // Large 0.1% error tolerance is because the optimizer may terminate before
@@ -558,7 +558,7 @@ TEST_CASE("LogisticRegressionLBFGSRegula
   LogisticRegression<> lr(data, responses, 0.001);
 
   // Test sigmoid function.
-  arma::rowvec sigmoids = 1 / (1 + arma::exp(-lr.Parameters()[0]
+  arma::rowvec sigmoids = 1 / (1 + exp(-lr.Parameters()[0]
       - lr.Parameters().tail_cols(lr.Parameters().n_elem - 1) * data));
 
   // Large error tolerance is because the optimizer may terminate before
@@ -584,7 +584,7 @@ TEST_CASE("LogisticRegressionSGDRegulari
   LogisticRegression<> lr(data, responses, sgd, 0.001);
 
   // Test sigmoid function.
-  arma::rowvec sigmoids = 1 / (1 + arma::exp(-lr.Parameters()[0]
+  arma::rowvec sigmoids = 1 / (1 + exp(-lr.Parameters()[0]
       - lr.Parameters().tail_cols(lr.Parameters().n_elem - 1) * data));
 
   // Large error tolerance is because the optimizer may terminate before
@@ -706,7 +706,7 @@ TEST_CASE("LogisticRegressionInstantiate
   LogisticRegression<> lr(data, responses, lbfgsOpt, 0.0005);
 
   // Test sigmoid function.
-  arma::rowvec sigmoids = 1 / (1 + arma::exp(-lr.Parameters()[0]
+  arma::rowvec sigmoids = 1 / (1 + exp(-lr.Parameters()[0]
       - lr.Parameters().tail_cols(lr.Parameters().n_elem - 1) * data));
 
   // Error tolerance is small because we tightened the optimizer tolerance.
@@ -721,7 +721,7 @@ TEST_CASE("LogisticRegressionInstantiate
   LogisticRegression<> lr2(data, responses, sgdOpt, 0.0005);
 
   // Test sigmoid function.
-  sigmoids = 1 / (1 + arma::exp(-lr2.Parameters()[0]
+  sigmoids = 1 / (1 + exp(-lr2.Parameters()[0]
       - lr2.Parameters().tail_cols(lr2.Parameters().n_elem - 1) * data));
 
   // Error tolerance is small because we tightened the optimizer tolerance.
@@ -873,7 +873,7 @@ TEST_CASE("ClassifyTest", "[LogisticRegr
   arma::Row<size_t> predictions;
   lr.Classify(data, predictions);
 
-  REQUIRE((double) arma::accu(predictions == responses) >= 900);
+  REQUIRE((double) accu(predictions == responses) >= 900);
 }
 
 /**
@@ -965,8 +965,9 @@ TEST_CASE("ClassifyProbabilitiesTest", "
     responses[i] = 1;
   }
 
+  arma::Row<size_t> predictions;
   arma::mat probabilities;
-  lr.Classify(data, probabilities);
+  lr.Classify(data, predictions, probabilities);
 
   REQUIRE(probabilities.n_cols == data.n_cols);
   REQUIRE(probabilities.n_rows == 2);
@@ -982,6 +983,16 @@ TEST_CASE("ClassifyProbabilitiesTest", "
     else
       REQUIRE(probabilities(1, i) == Approx(1.0).epsilon(0.10));
   }
+
+  // Classify a single point and make sure that we get reasonable probabilities.
+  arma::vec probabilities2;
+  size_t prediction;
+  lr.Classify(data.col(0), prediction, probabilities2);
+
+  REQUIRE(probabilities2.n_elem == 2);
+  REQUIRE(accu(probabilities2) == Approx(1.0));
+  REQUIRE(probabilities2[0] == Approx(probabilities(0, 0)));
+  REQUIRE(probabilities2[1] == Approx(probabilities(1, 0)));
 }
 
 /**
@@ -1078,3 +1089,173 @@ TEST_CASE("IncrementalTraining", "[Logis
 
   REQUIRE(acc == Approx(100.0).epsilon(0.03)); // 3% error tolerance.
 }
+
+// Test all constructor variants.  This test is more about checking that all
+// variants compile correctly than anything else.
+TEMPLATE_TEST_CASE("LogisticRegressionAllConstructorsTest",
+    "[LogisticRegressionTest]", arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels(1000, arma::fill::zeros);
+  labels.subvec(500, 999).fill(1);
+
+  // Empty constructor.
+  LogisticRegression<MatType> lr1;
+
+  // Specify data and labels and lambda, and optionally callbacks.
+  LogisticRegression<MatType> lr2(data, labels, 0.1);
+  LogisticRegression<MatType> lr3(data, labels, 0.1, ens::EarlyStopAtMinLoss());
+  LogisticRegression<MatType> lr4(data, labels, 0.1, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1000.0));
+
+  // Specify data and labels and initial point and lambda, and optionally
+  // callbacks.
+  arma::Row<typename MatType::elem_type> initialPoint(50, arma::fill::randu);
+  LogisticRegression<MatType> lr5(data, labels, initialPoint, 0.1);
+  LogisticRegression<MatType> lr6(data, labels, initialPoint, 0.1,
+      ens::EarlyStopAtMinLoss());
+  LogisticRegression<MatType> lr7(data, labels, initialPoint, 0.1,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1000.0));
+
+  // Specify data, labels, optimizer, and lambda, and optionally callbacks.
+  ens::AdaDelta optimizer(0.01);
+  LogisticRegression<MatType> lr8(data, labels, optimizer, 0.1);
+  LogisticRegression<MatType> lr9(data, labels, optimizer, 0.1,
+      ens::EarlyStopAtMinLoss());
+  LogisticRegression<MatType> lr10(data, labels, optimizer, 0.1,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1000.0));
+
+  // Specify data, labels, optimizer, initial point, and lambda, and optionally
+  // callbacks.
+  LogisticRegression<MatType> lr11(data, labels, optimizer, initialPoint, 0.1);
+  LogisticRegression<MatType> lr12(data, labels, optimizer, initialPoint, 0.1,
+      ens::EarlyStopAtMinLoss());
+  LogisticRegression<MatType> lr13(data, labels, optimizer, initialPoint, 0.1,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1000.0));
+
+  // Now we don't care what the training actually produced, but we do care that
+  // the model trained at all and has the right size (except for the first one).
+  REQUIRE(lr1.Parameters().n_elem == 1);
+  REQUIRE(lr2.Parameters().n_elem == 51);
+  REQUIRE(lr3.Parameters().n_elem == 51);
+  REQUIRE(lr4.Parameters().n_elem == 51);
+  REQUIRE(lr5.Parameters().n_elem == 51);
+  REQUIRE(lr6.Parameters().n_elem == 51);
+  REQUIRE(lr7.Parameters().n_elem == 51);
+  REQUIRE(lr8.Parameters().n_elem == 51);
+  REQUIRE(lr9.Parameters().n_elem == 51);
+  REQUIRE(lr10.Parameters().n_elem == 51);
+  REQUIRE(lr11.Parameters().n_elem == 51);
+  REQUIRE(lr12.Parameters().n_elem == 51);
+  REQUIRE(lr13.Parameters().n_elem == 51);
+}
+
+// Test all Train() variants.  This test is more about checking that all
+// variants compile correctly than anything else.
+TEMPLATE_TEST_CASE("LogisticRegressionAllTrainTest", "[LogisticRegressionTest]",
+    arma::mat, arma::fmat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels(1000, arma::fill::zeros);
+  labels.subvec(500, 999).fill(1);
+
+  // Construct all objects that we will use, but don't train.
+  LogisticRegression<MatType> lr1, lr2, lr3, lr4, lr5, lr6, lr7, lr8, lr9, lr10,
+      lr11, lr12, lr13, lr14, lr15, lr16, lr17, lr18;
+
+  // Specify data and labels only, and optionally callbacks.
+  lr1.Train(data, labels);
+  lr2.Train(data, labels, ens::EarlyStopAtMinLoss());
+  lr3.Train(data, labels, ens::EarlyStopAtMinLoss(), ens::TimerStop(1.0));
+
+  // Specify data and labels only, with a different optimizer, and optionally
+  // callbacks.
+  lr4.template Train<ens::GradientDescent>(data, labels);
+  lr5.template Train<ens::GradientDescent>(data, labels,
+      ens::EarlyStopAtMinLoss());
+  lr6.template Train<ens::GradientDescent>(data, labels,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1.0));
+
+  // Specify data, labels, and lambda, and optionally callbacks.
+  lr7.Train(data, labels, 0.1);
+  lr8.Train(data, labels, 0.1, ens::EarlyStopAtMinLoss());
+  lr9.Train(data, labels, 0.1, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1.0));
+
+  // Specify data, labels, and lambda, using a different optimizer, and
+  // optionally callbacks.
+  lr10.template Train<ens::StandardSGD>(data, labels, 0.1);
+  lr11.template Train<ens::StandardSGD>(data, labels, 0.1,
+      ens::EarlyStopAtMinLoss());
+  lr12.template Train<ens::StandardSGD>(data, labels, 0.1,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1.0));
+
+  // Specify data, labels, and an instantiated optimizer, optionally with
+  // callbacks.
+  ens::AdaDelta optimizer(0.01);
+  lr13.Train(data, labels, optimizer);
+  lr14.Train(data, labels, optimizer, ens::EarlyStopAtMinLoss());
+  lr15.Train(data, labels, optimizer, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1.0));
+
+  // Specify data, labels, an instantiated optimizer, and lambda, optionally
+  // with callbacks.
+  lr16.Train(data, labels, optimizer, 0.1);
+  lr17.Train(data, labels, optimizer, 0.1, ens::EarlyStopAtMinLoss());
+  lr18.Train(data, labels, optimizer, 0.1, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1.0));
+
+  REQUIRE(lr1.Parameters().n_elem == 51);
+  REQUIRE(lr2.Parameters().n_elem == 51);
+  REQUIRE(lr3.Parameters().n_elem == 51);
+  REQUIRE(lr4.Parameters().n_elem == 51);
+  REQUIRE(lr5.Parameters().n_elem == 51);
+  REQUIRE(lr6.Parameters().n_elem == 51);
+  REQUIRE(lr7.Parameters().n_elem == 51);
+  REQUIRE(lr8.Parameters().n_elem == 51);
+  REQUIRE(lr9.Parameters().n_elem == 51);
+  REQUIRE(lr10.Parameters().n_elem == 51);
+  REQUIRE(lr11.Parameters().n_elem == 51);
+  REQUIRE(lr12.Parameters().n_elem == 51);
+  REQUIRE(lr13.Parameters().n_elem == 51);
+  REQUIRE(lr14.Parameters().n_elem == 51);
+  REQUIRE(lr15.Parameters().n_elem == 51);
+  REQUIRE(lr16.Parameters().n_elem == 51);
+  REQUIRE(lr17.Parameters().n_elem == 51);
+  REQUIRE(lr18.Parameters().n_elem == 51);
+}
+
+// Make sure resetting the model does something.
+TEST_CASE("LogisticRegressionResetTest", "[LogisticRegressionTest]")
+{
+  // Create random data.
+  arma::mat data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels(1000, arma::fill::zeros);
+  labels.subvec(500, 999).fill(1);
+
+  // Create two logistic regression models.
+  LogisticRegression<> lr1, lr2;
+
+  ens::L_BFGS lbfgs1(1, 1); // Use only one iteration.
+  ens::L_BFGS lbfgs2(1, 1);
+
+  lr1.Train(data, labels, lbfgs1);
+  lr2.Train(data, labels, lbfgs2);
+
+  REQUIRE(
+      arma::approx_equal(lr1.Parameters(), lr2.Parameters(), "absdiff", 1e-5));
+
+  // Now reset one model and incrementally train the other.
+  lr1.Reset();
+  lr1.Train(data, labels, lbfgs1);
+  lr2.Train(data, labels, lbfgs2);
+
+  REQUIRE(
+      !arma::approx_equal(lr1.Parameters(), lr2.Parameters(), "absdiff", 1e-5));
+}
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/adaboost_probabilities_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/adaboost_probabilities_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/adaboost_probabilities_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/adaboost_probabilities_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,7 +62,7 @@ TEST_CASE_METHOD(AdaBoostPredictProbaTes
   REQUIRE(probabilities.n_cols == testSize);
 
   for (size_t i = 0; i < testSize; ++i)
-    REQUIRE(arma::accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
+    REQUIRE(accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
 }
 
 /**
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/adaboost_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/adaboost_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/adaboost_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/adaboost_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,8 +52,8 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   // Check that number of predicted labels is equal to the input test points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
 }
 
 /**
@@ -91,7 +91,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   REQUIRE(probabilities.n_cols == testSize);
 
   for (size_t i = 0; i < testSize; ++i)
-    REQUIRE(arma::accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
+    REQUIRE(accu(probabilities.col(i)) == Approx(1).epsilon(1e-7));
 }
 
 /**
@@ -120,7 +120,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   arma::Row<size_t> output;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   AdaBoostModel* model = params.Get<AdaBoostModel*>("output_model");
   ResetSettings();
@@ -131,7 +131,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   // Check that initial output and output using saved model are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
 }
 
 /**
@@ -182,7 +182,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   arma::Row<size_t> output;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   CleanMemory();
   ResetSettings();
@@ -197,7 +197,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   // Check that initial output and final output matrix are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
 }
 
 /**
@@ -221,30 +221,6 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
 }
 
 /**
- * This test can be removed in mlpack 4.0.0.  This tests that the output and
- * predictions outputs are the same.
- */
-TEST_CASE_METHOD(AdaBoostTestFixture, "AdaBoostOutputPredictionsTest",
-                 "[AdaBoostMainTest][BindingTests]")
-{
-  arma::mat trainData;
-  if (!data::Load("vc2.csv", trainData))
-    FAIL("Unable to load train dataset vc2.csv!");
-
-  arma::Row<size_t> labels;
-  if (!data::Load("vc2_labels.txt", labels))
-    FAIL("Unable to load label dataset vc2_labels.txt!");
-
-  SetInputParam("training", std::move(trainData));
-  SetInputParam("labels", std::move(labels));
-
-  RUN_BINDING();
-
-  CheckMatrices(params.Get<arma::Row<size_t>>("output"),
-                params.Get<arma::Row<size_t>>("predictions"));
-}
-
-/**
  * Weak learner should be either Decision Stump or Perceptron.
  */
 TEST_CASE_METHOD(AdaBoostTestFixture, "AdaBoostWeakLearnerTest",
@@ -285,7 +261,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   arma::Row<size_t> output;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   CleanMemory();
   ResetSettings();
@@ -298,9 +274,9 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   RUN_BINDING();
 
   arma::Row<size_t> outputPerceptron;
-  outputPerceptron = std::move(params.Get<arma::Row<size_t>>("output"));
+  outputPerceptron = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
-  REQUIRE(arma::accu(output != outputPerceptron) > 1);
+  REQUIRE(accu(output != outputPerceptron) > 1);
 }
 
 /**
@@ -339,7 +315,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  size_t correct = arma::accu(output == testLabels);
+  size_t correct = accu(output == testLabels);
   double accuracy1 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -357,7 +333,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy10 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -375,7 +351,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy100 = (double(correct) / double(testLabels.n_elem) * 100);
 
   REQUIRE(accuracy1 <= accuracy10);
@@ -417,7 +393,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  size_t correct = arma::accu(output == testLabels);
+  size_t correct = accu(output == testLabels);
   double accuracy1 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -434,7 +410,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy2 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -451,7 +427,7 @@ TEST_CASE_METHOD(AdaBoostTestFixture, "A
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy3 = (double(correct) / double(testLabels.n_elem) * 100);
 
   REQUIRE(accuracy1 <= accuracy2);
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/adaboost_train_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/adaboost_train_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/adaboost_train_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/adaboost_train_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -154,7 +154,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
       outputPerceptron);
 
-  REQUIRE(arma::accu(output != outputPerceptron) > 1);
+  REQUIRE(accu(output != outputPerceptron) > 1);
 }
 
 /**
@@ -193,7 +193,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  size_t correct = arma::accu(output == testLabels);
+  size_t correct = accu(output == testLabels);
   double accuracy1 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -211,7 +211,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy10 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -229,7 +229,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy100 = (double(correct) / double(testLabels.n_elem) * 100);
 
   REQUIRE(accuracy1 <= accuracy10);
@@ -271,7 +271,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  size_t correct = arma::accu(output == testLabels);
+  size_t correct = accu(output == testLabels);
   double accuracy1 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -288,7 +288,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy2 = (double(correct) / double(testLabels.n_elem) * 100);
 
   CleanMemory();
@@ -305,7 +305,7 @@ TEST_CASE_METHOD(AdaBoostFitTestFixture,
   params.Get<AdaBoostModel*>("output_model")->Classify(testData,
        output);
 
-  correct = arma::accu(output == testLabels);
+  correct = accu(output == testLabels);
   double accuracy3 = (double(correct) / double(testLabels.n_elem) * 100);
 
   REQUIRE(accuracy1 <= accuracy2);
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/bayesian_linear_regression_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/bayesian_linear_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/bayesian_linear_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/bayesian_linear_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -41,8 +41,8 @@ TEST_CASE_METHOD(BRTestFixture,
 
   RUN_BINDING();
 
-  BayesianLinearRegression* estimator =
-      params.Get<BayesianLinearRegression*>("output_model");
+  BayesianLinearRegression<>* estimator =
+      params.Get<BayesianLinearRegression<>*>("output_model");
 
   REQUIRE(estimator->DataOffset().n_elem == 0);
   REQUIRE(estimator->DataScale().n_elem == 0);
@@ -61,7 +61,7 @@ TEST_CASE_METHOD(BRTestFixture,
   const arma::rowvec omega = arma::randu<arma::rowvec>(m);
   arma::rowvec y =  omega * matX;
 
-  BayesianLinearRegression model;
+  BayesianLinearRegression<> model;
   model.Train(matX, y);
 
   arma::rowvec responses;
@@ -72,8 +72,8 @@ TEST_CASE_METHOD(BRTestFixture,
 
   RUN_BINDING();
 
-  BayesianLinearRegression* mOut =
-      params.Get<BayesianLinearRegression*>("output_model");
+  BayesianLinearRegression<>* mOut =
+      params.Get<BayesianLinearRegression<>*>("output_model");
 
   ResetSettings();
 
@@ -101,7 +101,7 @@ TEST_CASE_METHOD(BRTestFixture,
   const arma::rowvec omega = arma::randu<arma::rowvec>(m);
   arma::rowvec y =  omega * matX;
 
-  BayesianLinearRegression model;
+  BayesianLinearRegression<> model;
   model.Train(matX, y);
 
   arma::rowvec responses;
@@ -121,7 +121,7 @@ TEST_CASE_METHOD(BRTestFixture,
   // An error should occur.
   SetInputParam("input", std::move(matX));
   SetInputParam("input_model",
-                params.Get<BayesianLinearRegression*>("output_model"));
+                params.Get<BayesianLinearRegression<>*>("output_model"));
   SetInputParam("test", std::move(matXtest));
 
   REQUIRE_THROWS_AS(RUN_BINDING(), std::runtime_error);
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/cf_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/cf_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/cf_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/cf_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -421,7 +421,7 @@ TEST_CASE_METHOD(CFTestFixture, "CFNeigh
   const arma::Mat<size_t> output2 = params.Get<arma::Mat<size_t>>("output");
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
 }
 
 /**
@@ -505,8 +505,8 @@ TEST_CASE_METHOD(CFTestFixture, "CFInter
   REQUIRE(output3.n_cols == 7);
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
-  REQUIRE(arma::any(arma::vectorise(output1 != output3)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output3)));
 }
 
 /**
@@ -590,8 +590,8 @@ TEST_CASE_METHOD(CFTestFixture, "CFNeigh
   REQUIRE(output3.n_cols == 7);
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
-  REQUIRE(arma::any(arma::vectorise(output1 != output3)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output3)));
 }
 
 /**
@@ -688,6 +688,6 @@ TEST_CASE_METHOD(CFTestFixture, "CFNorma
   REQUIRE(output3.n_cols == 7);
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
-  REQUIRE(arma::any(arma::vectorise(output1 != output3)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output3)));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/dbscan_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/dbscan_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/dbscan_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/dbscan_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -133,7 +133,7 @@ TEST_CASE_METHOD(DBSCANTestFixture, "DBS
   arma::Row<size_t> output2;
   output2 = std::move(params.Get<arma::Row<size_t>>("assignments"));
 
-  REQUIRE(arma::accu(output1 != output2) > 1);
+  REQUIRE(accu(output1 != output2) > 1);
 }
 
 /**
@@ -168,7 +168,7 @@ TEST_CASE_METHOD(DBSCANTestFixture, "DBS
   arma::Row<size_t> output2;
   output2 = std::move(params.Get<arma::Row<size_t>>("assignments"));
 
-  REQUIRE(arma::accu(output1 != output2) > 1);
+  REQUIRE(accu(output1 != output2) > 1);
 }
 
 /**
@@ -422,5 +422,5 @@ TEST_CASE_METHOD(DBSCANTestFixture, "DBS
   arma::Row<size_t> randomOutput;
   randomOutput = std::move(params.Get<arma::Row<size_t>>("assignments"));
 
-  REQUIRE(arma::accu(orderedOutput != randomOutput) > 0);
+  REQUIRE(accu(orderedOutput != randomOutput) > 0);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/det_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/det_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/det_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/det_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -216,11 +216,11 @@ TEST_CASE_METHOD(DETTestFixture, "DETDif
 
   // Check that initial estimates and final estimates using two models are
   // different.
-  REQUIRE(arma::accu(trainingSetEstimates ==
+  REQUIRE(accu(trainingSetEstimates ==
       params.Get<arma::mat>("training_set_estimates")) <
       trainingSetEstimates.n_elem);
 
-  REQUIRE(arma::accu(testSetEstimates ==
+  REQUIRE(accu(testSetEstimates ==
       params.Get<arma::mat>("test_set_estimates")) <
       testSetEstimates.n_elem);
 }
@@ -261,11 +261,11 @@ TEST_CASE_METHOD(DETTestFixture, "DETDif
 
   // Check that initial estimates and final estimates using two models are
   // different.
-  REQUIRE(arma::accu(trainingSetEstimates ==
+  REQUIRE(accu(trainingSetEstimates ==
       params.Get<arma::mat>("training_set_estimates")) <
       trainingSetEstimates.n_elem);
 
-  REQUIRE(arma::accu(testSetEstimates ==
+  REQUIRE(accu(testSetEstimates ==
       params.Get<arma::mat>("test_set_estimates")) <
       testSetEstimates.n_elem);
 }
@@ -306,11 +306,11 @@ TEST_CASE_METHOD(DETTestFixture, "DETDif
 
   // Check that initial estimates and final estimates using two models are
   // different.
-  REQUIRE(arma::accu(trainingSetEstimates ==
+  REQUIRE(accu(trainingSetEstimates ==
       params.Get<arma::mat>("training_set_estimates")) <
       trainingSetEstimates.n_elem);
 
-  REQUIRE(arma::accu(testSetEstimates ==
+  REQUIRE(accu(testSetEstimates ==
       params.Get<arma::mat>("test_set_estimates")) <
       testSetEstimates.n_elem);
 }
@@ -351,11 +351,11 @@ TEST_CASE_METHOD(DETTestFixture, "DETSki
 
   // Check that initial estimates and final estimates using two models are
   // different.
-  REQUIRE(arma::accu(trainingSetEstimates ==
+  REQUIRE(accu(trainingSetEstimates ==
       params.Get<arma::mat>("training_set_estimates")) <
       trainingSetEstimates.n_elem);
 
-  REQUIRE(arma::accu(testSetEstimates ==
+  REQUIRE(accu(testSetEstimates ==
       params.Get<arma::mat>("test_set_estimates")) <
       testSetEstimates.n_elem);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/hmm_test_utils.hpp 4.4.0-1/src/mlpack/tests/main_tests/hmm_test_utils.hpp
--- 4.3.0-2/src/mlpack/tests/main_tests/hmm_test_utils.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/hmm_test_utils.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -49,7 +49,7 @@ struct InitHMMModel
          ++it)
     {
       arma::Col<size_t> maxSeqs =
-          arma::conv_to<arma::Col<size_t>>::from(arma::max(*it, 1)) + 1;
+          ConvTo<arma::Col<size_t>>::From(arma::max(*it, 1)) + 1;
       maxEmissions = arma::max(maxEmissions, maxSeqs);
     }
 
@@ -140,7 +140,7 @@ struct InitHMMModel
     for (size_t i = 0; i < e.size(); ++i)
     {
       e[i].Probabilities().randu();
-      e[i].Probabilities() /= arma::accu(e[i].Probabilities());
+      e[i].Probabilities() /= accu(e[i].Probabilities());
     }
   }
 
@@ -162,7 +162,7 @@ struct InitHMMModel
     {
       // Random weights.
       e[i].Weights().randu();
-      e[i].Weights() /= arma::accu(e[i].Weights());
+      e[i].Weights() /= accu(e[i].Weights());
 
       // Random means and covariances.
       for (int g = 0; g < 2; ++g)
@@ -185,7 +185,7 @@ struct InitHMMModel
     {
       // Random weights.
       e[i].Weights().randu();
-      e[i].Weights() /= arma::accu(e[i].Weights());
+      e[i].Weights() /= accu(e[i].Weights());
 
       // Random means and covariances.
       for (int g = 0; g < 2; ++g)
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/image_converter_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/image_converter_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/image_converter_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/image_converter_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -44,7 +44,7 @@ TEST_CASE_METHOD(ImageConverterTestFixtu
 TEST_CASE_METHOD(ImageConverterTestFixture, "SaveImageTest",
                  "[ImageConverterMainTest][BindingTests]")
 {
-  arma::mat testimage = arma::conv_to<arma::mat>::from(
+  arma::mat testimage = ConvTo<arma::mat>::From(
       arma::randi<arma::Mat<unsigned char>>((5 * 5 * 3), 2));
   SetInputParam<vector<string>>("input", {"test_image777.png",
       "test_image999.png"});
@@ -81,7 +81,7 @@ TEST_CASE_METHOD(ImageConverterTestFixtu
 TEST_CASE_METHOD(ImageConverterTestFixture, "IncompleteTest",
                  "[ImageConverterMainTest][BindingTests]")
 {
-  arma::mat testimage = arma::conv_to<arma::mat>::from(
+  arma::mat testimage = ConvTo<arma::mat>::From(
       arma::randi<arma::Mat<unsigned char>>((5 * 5 * 3), 2));
   SetInputParam<vector<string>>("input", {"test_image777.png",
       "test_image999.png"});
@@ -99,7 +99,7 @@ TEST_CASE_METHOD(ImageConverterTestFixtu
 TEST_CASE_METHOD(ImageConverterTestFixture, "InvalidInputTest",
                  "[ImageConverterMainTest][BindingTests]")
 {
-  arma::mat testimage = arma::conv_to<arma::mat>::from(
+  arma::mat testimage = ConvTo<arma::mat>::From(
       arma::randi<arma::Mat<unsigned char>>((5 * 5 * 3), 2));
   SetInputParam<vector<string>>("input", {"test_image777.png",
       "test_image999.png"});
@@ -119,7 +119,7 @@ TEST_CASE_METHOD(ImageConverterTestFixtu
 TEST_CASE_METHOD(ImageConverterTestFixture, "InvalidWidthTest",
                  "[ImageConverterMainTest][BindingTests]")
 {
-  arma::mat testimage = arma::conv_to<arma::mat>::from(
+  arma::mat testimage = ConvTo<arma::mat>::From(
       arma::randi<arma::Mat<unsigned char>>((5 * 5 * 3), 2));
   SetInputParam<vector<string>>("input", {"test_image777.png",
       "test_image999.png"});
@@ -138,7 +138,7 @@ TEST_CASE_METHOD(ImageConverterTestFixtu
 TEST_CASE_METHOD(ImageConverterTestFixture, "InvalidChannelTest",
                  "[ImageConverterMainTest][BindingTests]")
 {
-  arma::mat testimage = arma::conv_to<arma::mat>::from(
+  arma::mat testimage = ConvTo<arma::mat>::From(
       arma::randi<arma::Mat<unsigned char>>((5 * 5 * 3), 2));
   SetInputParam<vector<string>>("input", {"test_image777.png",
       "test_image999.png"});
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/kde_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/kde_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/kde_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/kde_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -518,6 +518,6 @@ TEST_CASE_METHOD(KDETestFixture, "KDEMai
 
   // Check whether results are equal.
   differences = arma::abs(estimations1 - estimations2);
-  const double sumDifferences = arma::accu(differences);
+  const double sumDifferences = accu(differences);
   REQUIRE(sumDifferences > 0);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/kernel_pca_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/kernel_pca_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/kernel_pca_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/kernel_pca_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -121,7 +121,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
   arma::mat output2 = params.Get<arma::mat>("output");
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
 }
 
 /**
@@ -204,7 +204,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
     arma::mat output2 = params.Get<arma::mat>("output");
 
     // The resulting matrices should be different.
-    REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+    REQUIRE(arma::any(vectorise(output1 != output2)));
   }
 }
 
@@ -239,7 +239,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
     arma::mat output2 = params.Get<arma::mat>("output");
 
     // The resulting matrices should be different.
-    REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+    REQUIRE(arma::any(vectorise(output1 != output2)));
   }
 }
 
@@ -266,7 +266,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
   arma::mat output2 = params.Get<arma::mat>("output");
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
 }
 
 /**
@@ -292,7 +292,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
   arma::mat output2 = params.Get<arma::mat>("output");
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
 }
 
 /**
@@ -329,7 +329,7 @@ TEST_CASE_METHOD(KernelPCATestFixture, "
   arma::mat output3 = params.Get<arma::mat>("output");
 
   // The resulting matrices should be different.
-  REQUIRE(arma::any(arma::vectorise(output1 != output2)));
-  REQUIRE(arma::any(arma::vectorise(output2 != output3)));
-  REQUIRE(arma::any(arma::vectorise(output1 != output3)));
+  REQUIRE(arma::any(vectorise(output1 != output2)));
+  REQUIRE(arma::any(vectorise(output2 != output3)));
+  REQUIRE(arma::any(vectorise(output1 != output3)));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/kmeans_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/kmeans_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/kmeans_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/kmeans_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -212,10 +212,10 @@ TEST_CASE_METHOD(KmTestFixture, "KmClust
 
   if (killEmptyOutput.n_elem == allowEmptyOutput.n_elem)
   {
-    REQUIRE(arma::accu(killEmptyOutput != allowEmptyOutput) > 1);
-    REQUIRE(arma::accu(killEmptyOutput != normalOutput) > 1);
+    REQUIRE(accu(killEmptyOutput != allowEmptyOutput) > 1);
+    REQUIRE(accu(killEmptyOutput != normalOutput) > 1);
   }
-  REQUIRE(arma::accu(normalOutput != allowEmptyOutput) > 1);
+  REQUIRE(accu(normalOutput != allowEmptyOutput) > 1);
 }
 
 /**
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_predict_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_predict_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_predict_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_predict_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -37,7 +37,7 @@ TEST_CASE_METHOD(LRPredictTestFixture, "
   arma::rowvec trainY = arma::randu<arma::rowvec>(N);
   arma::mat testX = arma::randu<arma::mat>(D - 1, M); // Wrong dimensionality.
 
-  LinearRegression* model = new LinearRegression();
+  LinearRegression<>* model = new LinearRegression<>();
   model->Train(trainX, trainY);
 
   SetInputParam("input_model", std::move(model));
@@ -60,7 +60,7 @@ TEST_CASE_METHOD(LRPredictTestFixture, "
   arma::rowvec trainY = arma::randu<arma::rowvec>(N);
   arma::mat testX = arma::randu<arma::mat>(D, M);
 
-  LinearRegression* model = new LinearRegression();
+  LinearRegression<>* model = new LinearRegression<>();
   model->Train(trainX, trainY);
 
   SetInputParam("input_model", std::move(model));
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -118,7 +118,7 @@ TEST_CASE_METHOD(LRTestFixture, "LRModel
 
   RUN_BINDING();
 
-  LinearRegression* model = params.Get<LinearRegression*>("output_model");
+  LinearRegression<>* model = params.Get<LinearRegression<>*>("output_model");
   const arma::rowvec testY1 = params.Get<arma::rowvec>("output_predictions");
 
   ResetSettings();
@@ -191,7 +191,7 @@ TEST_CASE_METHOD(LRTestFixture, "LRWrong
 
   RUN_BINDING();
 
-  LinearRegression* model = params.Get<LinearRegression*>("output_model");
+  LinearRegression<>* model = params.Get<LinearRegression<>*>("output_model");
 
   ResetSettings();
 
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_train_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_train_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/linear_regression_train_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/linear_regression_train_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -44,8 +44,7 @@ TEST_CASE_METHOD(LRFitTestFixture, "LRFi
   // The first solution.
   RUN_BINDING();
   arma::rowvec preds1;
-  params.Get<LinearRegression*>("output_model")->Predict(testX,
-      preds1);
+  params.Get<LinearRegression<>*>("output_model")->Predict(testX, preds1);
   const double testY1 = preds1(0);
 
   ResetSettings();
@@ -57,8 +56,7 @@ TEST_CASE_METHOD(LRFitTestFixture, "LRFi
   // The second solution.
   RUN_BINDING();
   arma::rowvec preds2;
-  params.Get<LinearRegression*>("output_model")->Predict(testX,
-      preds2);
+  params.Get<LinearRegression<>*>("output_model")->Predict(testX, preds2);
   const double testY2 = preds2(0);
 
   // Second solution has stronger regularization,
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/lmnn_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/lmnn_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/lmnn_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/lmnn_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -325,8 +325,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -365,11 +365,11 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   SetInputParam("linear_scan",  (bool) true);
 
   RUN_BINDING();
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -408,8 +408,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -450,8 +450,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -493,8 +493,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -536,8 +536,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -578,8 +578,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -625,8 +625,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
@@ -668,8 +668,8 @@ TEST_CASE_METHOD(LMNNTestFixture, "LMNND
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
-  REQUIRE(arma::accu(params.Get<arma::mat>("transformed_data") !=
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("transformed_data") !=
       transformedData) > 0);
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/local_coordinate_coding_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/local_coordinate_coding_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/local_coordinate_coding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/local_coordinate_coding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,7 +32,7 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCDim
                  "[LCCMainTest][BindingTests]")
 {
   arma::mat x;
-  x.load("mnist_first250_training_4s_and_9s.arm");
+  x.load("mnist_first250_training_4s_and_9s.csv");
   int rows = x.n_rows, cols = x.n_cols;
   arma::mat t = x;
   int atoms = 10;
@@ -58,7 +58,7 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCOut
                  "[LCCMainTest][BindingTests]")
 {
   arma::mat x;
-  x.load("mnist_first250_training_4s_and_9s.arm");
+  x.load("mnist_first250_training_4s_and_9s.csv");
   arma::mat t = x;
 
   SetInputParam("training", std::move(x));
@@ -71,8 +71,8 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCOut
   // Get the encoded output and dictionary after training.
   arma::mat initCodes = std::move(params.Get<arma::mat>("codes"));
   arma::mat initDict = std::move(params.Get<arma::mat>("dictionary"));
-  LocalCoordinateCoding* outputModel =
-      std::move(params.Get<LocalCoordinateCoding*>("output_model"));
+  LocalCoordinateCoding<>* outputModel =
+      std::move(params.Get<LocalCoordinateCoding<>*>("output_model"));
 
   ResetSettings();
 
@@ -129,7 +129,7 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCTra
                  "[LCCMainTest][BindingTests]")
 {
   arma::mat x;
-  x.load("mnist_first250_training_4s_and_9s.arm");
+  x.load("mnist_first250_training_4s_and_9s.csv");
   arma::mat t = x;
 
   t.shed_rows(1, 2);
@@ -150,7 +150,7 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCTra
                  "[LCCMainTest][BindingTests]")
 {
   arma::mat x;
-  x.load("mnist_first250_training_4s_and_9s.arm");
+  x.load("mnist_first250_training_4s_and_9s.csv");
 
   SetInputParam("training", x);
   SetInputParam("atoms", (int) 10);
@@ -158,8 +158,8 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCTra
 
   RUN_BINDING();
 
-  LocalCoordinateCoding* outputModel =
-      std::move(params.Get<LocalCoordinateCoding*>("output_model"));
+  LocalCoordinateCoding<>* outputModel =
+      std::move(params.Get<LocalCoordinateCoding<>*>("output_model"));
 
   // No need to input training data again.
   SetInputParam("input_model", std::move(outputModel));
@@ -175,7 +175,7 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCTra
                  "[LCCMainTest][BindingTests]")
 {
   arma::mat x;
-  x.load("mnist_first250_training_4s_and_9s.arm");
+  x.load("mnist_first250_training_4s_and_9s.csv");
   arma:: mat t = x;
   t.shed_rows(1, 2);
 
@@ -185,8 +185,8 @@ TEST_CASE_METHOD(LCCTestFixture, "LCCTra
 
   RUN_BINDING();
 
-  LocalCoordinateCoding* outputModel =
-      std::move(params.Get<LocalCoordinateCoding*>("output_model"));
+  LocalCoordinateCoding<>* outputModel =
+      std::move(params.Get<LocalCoordinateCoding<>*>("output_model"));
 
   SetInputParam("input_model", std::move(outputModel));
   SetInputParam("test", std::move(t));
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/logistic_regression_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/logistic_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/logistic_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/logistic_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -625,5 +625,27 @@ TEST_CASE_METHOD(LogisticRegressionTestF
       params.Get<arma::Row<size_t>>("predictions");
 
   // Check that the output changed when the decision boundary moved.
-  REQUIRE(arma::accu(output1 != output2) > 0);
+  REQUIRE(accu(output1 != output2) > 0);
+}
+
+/**
+ * Check that running the binding with print_training_accuracy set to true
+ * does not crash.
+ **/
+TEST_CASE_METHOD(LogisticRegressionTestFixture, "LRPrintTrainingAccuracyTest",
+                "[LogisticRegressionMainTest][BindingTests]")
+{
+  constexpr int N = 100;
+  constexpr int D = 5;
+
+  arma::mat trainX = arma::randu<arma::mat>(D, N);
+  arma::Row<size_t> trainY = arma::randi<arma::Row<size_t>>(N,
+      arma::distr_param(0, 1));
+
+  SetInputParam("training", trainX);
+  SetInputParam("labels", trainY);
+  SetInputParam("print_training_accuracy", true);
+
+  // Run the binding with print_training_accuracy set to true.
+  REQUIRE_NOTHROW(RUN_BINDING());
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/lsh_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/lsh_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/lsh_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/lsh_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -134,9 +134,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
@@ -171,9 +171,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
@@ -208,9 +208,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
@@ -248,9 +248,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
@@ -285,9 +285,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
@@ -322,9 +322,9 @@ TEST_CASE_METHOD(LSHTestFixture, "LSHDif
 
   // Check that initial outputs and final outputs using the two models are
   // different.
-  REQUIRE(arma::accu(neighbors ==
+  REQUIRE(accu(neighbors ==
       params.Get<arma::Mat<size_t>>("neighbors")) < neighbors.n_elem);
-  REQUIRE(arma::accu(distances ==
+  REQUIRE(accu(distances ==
       params.Get<arma::mat>("distances")) < distances.n_elem);
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/nbc_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/nbc_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/nbc_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/nbc_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,12 +62,12 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCOut
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 }
 
 /**
@@ -106,18 +106,18 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCLab
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 
   // Store outputs.
   arma::Row<size_t> output;
   arma::mat output_probs;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
-  output_probs = std::move(params.Get<arma::mat>("output_probs"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
+  output_probs = std::move(params.Get<arma::mat>("probabilities"));
 
   // Reset data passed.
   CleanMemory();
@@ -136,17 +136,17 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCLab
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 
   // Check that initial output and final output matrix
   // from two models are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
-  CheckMatrices(output_probs, params.Get<arma::mat>("output_probs"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
+  CheckMatrices(output_probs, params.Get<arma::mat>("probabilities"));
 }
 
 /**
@@ -178,8 +178,8 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCMod
 
   arma::Row<size_t> output;
   arma::mat output_probs;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
-  output_probs = std::move(params.Get<arma::mat>("output_probs"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
+  output_probs = std::move(params.Get<arma::mat>("probabilities"));
 
   // Reset passed parameters.
   NBCModel* m = params.Get<NBCModel*>("output_model");
@@ -194,17 +194,17 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCMod
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 
   // Check that initial output and final output
   // matrix using saved model are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
-  CheckMatrices(output_probs, params.Get<arma::mat>("output_probs"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
+  CheckMatrices(output_probs, params.Get<arma::mat>("probabilities"));
 }
 
 /**
@@ -260,18 +260,18 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCInc
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 
   // Store outputs.
   arma::Row<size_t> output;
   arma::mat output_probs;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
-  output_probs = std::move(params.Get<arma::mat>("output_probs"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
+  output_probs = std::move(params.Get<arma::mat>("probabilities"));
 
   CleanMemory();
   ResetSettings();
@@ -286,112 +286,15 @@ TEST_CASE_METHOD(NBCTestFixture, "NBCInc
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
-  REQUIRE(params.Get<arma::mat>("output_probs").n_rows == 2);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
+  REQUIRE(params.Get<arma::mat>("probabilities").n_rows == 2);
 
   // Check that initial output and final output matrix
   // from two models are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
-  CheckMatrices(output_probs, params.Get<arma::mat>("output_probs"));
-}
-
-/**
- * Ensure that the parameter 'output' and the parameter 'predictions' give the
- * same output.  This test case should be removed in mlpack 4 when the
- * deprecated parameter 'output' is removed.
- */
-TEST_CASE_METHOD(NBCTestFixture, "NBCOptionConsistencyTest",
-                "[NBCMainTest][BindingTests]")
-{
-  arma::mat inputData;
-  if (!data::Load("trainSet.csv", inputData))
-    FAIL("Cannot load train dataset trainSet.csv!");
-
-  // Get the labels out.
-  arma::Row<size_t> labels(inputData.n_cols);
-  for (size_t i = 0; i < inputData.n_cols; ++i)
-    labels[i] = inputData(inputData.n_rows - 1, i);
-
-  // Delete the last row containing labels from input dataset.
-  inputData.shed_row(inputData.n_rows - 1);
-
-  arma::mat testData;
-  if (!data::Load("testSet.csv", testData))
-    FAIL("Cannot load test dataset testSet.csv!");
-
-  // Delete the last row containing labels from test dataset.
-  testData.shed_row(testData.n_rows - 1);
-
-  // Input training data.
-  SetInputParam("training", std::move(inputData));
-  SetInputParam("labels", std::move(labels));
-
-  // Input test data.
-  SetInputParam("test", std::move(testData));
-
-  RUN_BINDING();
-
-  // Get the output from the 'output' parameter.
-  const arma::Row<size_t> testY1 =
-      std::move(params.Get<arma::Row<size_t>>("output"));
-
-  // Get output from 'predictions' parameter.
-  const arma::Row<size_t> testY2 =
-      params.Get<arma::Row<size_t>>("predictions");
-
-  // Both solutions must be equal.
-  CheckMatrices(testY1, testY2);
-}
-
-
-/**
- * This test ensures that the parameter 'output_probabilities' and the parameter
- * 'probabilities' give the same output.  This test case should be removed in
- * mlpack 4 when the deprecated parameter: 'output_probabilities' is removed.
- */
-TEST_CASE_METHOD(NBCTestFixture, "NBCOptionConsistencyTest2",
-                "[NBCMainTest][BindingTests]")
-{
-  arma::mat inputData;
-  if (!data::Load("trainSet.csv", inputData))
-    FAIL("Cannot load train dataset trainSet.csv!");
-
-  // Get the labels out.
-  arma::Row<size_t> labels(inputData.n_cols);
-  for (size_t i = 0; i < inputData.n_cols; ++i)
-    labels[i] = inputData(inputData.n_rows - 1, i);
-
-  // Delete the last row containing labels from input dataset.
-  inputData.shed_row(inputData.n_rows - 1);
-
-  arma::mat testData;
-  if (!data::Load("testSet.csv", testData))
-    FAIL("Cannot load test dataset testSet.csv!");
-
-  // Delete the last row containing labels from test dataset.
-  testData.shed_row(testData.n_rows - 1);
-
-  // Input training data.
-  SetInputParam("training", std::move(inputData));
-  SetInputParam("labels", std::move(labels));
-
-  // Input test data.
-  SetInputParam("test", std::move(testData));
-
-  RUN_BINDING();
-
-  // Get the output probabilites which is a deprecated parameter.
-  const arma::mat testY1 =
-      std::move(params.Get<arma::mat>("output_probs"));
-
-  // Get probabilities from 'predictions' parameter.
-  const arma::mat testY2 =
-      params.Get<arma::mat>("probabilities");
-
-  // Both solutions must be equal.
-  CheckMatrices(testY1, testY2);
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
+  CheckMatrices(output_probs, params.Get<arma::mat>("probabilities"));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/nca_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/nca_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/nca_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/nca_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -156,7 +156,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCANor
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
 }
 
 /**
@@ -198,7 +198,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCADif
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
 }
 
 /**
@@ -247,7 +247,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCADif
     RUN_BINDING();
 
     // Check that the output matrices are different.
-    success = (arma::accu(params.Get<arma::mat>("output") != output) > 0);
+    success = (accu(params.Get<arma::mat>("output") != output) > 0);
     if (success)
       break;
 
@@ -298,7 +298,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCADif
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
 }
 
 /**
@@ -339,7 +339,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCALin
   RUN_BINDING();
 
   // Check that the output matrices are different.
-  REQUIRE(arma::accu(params.Get<arma::mat>("output") != output) > 0);
+  REQUIRE(accu(params.Get<arma::mat>("output") != output) > 0);
 }
 
 /**
@@ -425,7 +425,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCADif
     RUN_BINDING();
 
     // Check that the output matrices are different.
-    success = (arma::accu(params.Get<arma::mat>("output") != output) > 0);
+    success = (accu(params.Get<arma::mat>("output") != output) > 0);
     if (success)
       break;
 
@@ -480,7 +480,7 @@ TEST_CASE_METHOD(NCATestFixture, "NCADif
     RUN_BINDING();
 
     // Check that the output matrices are different.
-    success = (arma::accu(params.Get<arma::mat>("output") != output) > 0);
+    success = (accu(params.Get<arma::mat>("output") != output) > 0);
     if (success)
       break;
 
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/perceptron_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/perceptron_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/perceptron_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/perceptron_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -62,10 +62,10 @@ TEST_CASE_METHOD(PerceptronTestFixture,
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
 }
 
 /**
@@ -104,16 +104,16 @@ TEST_CASE_METHOD(PerceptronTestFixture,
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
 
   inputData.shed_row(inputData.n_rows - 1);
 
   // Store outputs.
   arma::Row<size_t> output;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   // Reset data passed.
   CleanMemory();
@@ -130,48 +130,14 @@ TEST_CASE_METHOD(PerceptronTestFixture,
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
 
   // Check that initial output and final output matrix
   // from two models are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
-}
-
-/**
- * This test can be removed in mlpack 4.0.0. This tests that the output and
- * predictions outputs are the same.
- */
-TEST_CASE_METHOD(PerceptronTestFixture, "PerceptronOutputPredictionsCheck",
-                 "[PerceptronMainTest][BindingTests]")
-{
-  arma::mat trainX1;
-  arma::Row<size_t> labelsX1;
-
-  // Loading a train data set with 3 classes.
-  if (!data::Load("vc2.csv", trainX1))
-  {
-    FAIL("Could not load the train data (vc2.csv)");
-  }
-
-  // Loading the corresponding labels to the dataset.
-  if (!data::Load("vc2_labels.txt", labelsX1))
-  {
-    FAIL("Could not load the train data (vc2_labels.csv)");
-  }
-
-  SetInputParam("training", std::move(trainX1)); // Training data.
-  // Labels for the training data.
-  SetInputParam("labels", std::move(labelsX1));
-
-  // Training model using first training dataset.
-  RUN_BINDING();
-
-  // Check that the outputs are the same.
-  CheckMatrices(params.Get<arma::Row<size_t>>("output"),
-                params.Get<arma::Row<size_t>>("predictions"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
 }
 
 /**
@@ -202,7 +168,7 @@ TEST_CASE_METHOD(PerceptronTestFixture,
   RUN_BINDING();
 
   arma::Row<size_t> output;
-  output = std::move(params.Get<arma::Row<size_t>>("output"));
+  output = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   // Reset passed parameters.
   PerceptronModel* m = params.Get<PerceptronModel*>("output_model");
@@ -217,14 +183,14 @@ TEST_CASE_METHOD(PerceptronTestFixture,
   RUN_BINDING();
 
   // Check that number of output points are equal to number of input points.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_cols == testSize);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_cols == testSize);
 
   // Check output have only single row.
-  REQUIRE(params.Get<arma::Row<size_t>>("output").n_rows == 1);
+  REQUIRE(params.Get<arma::Row<size_t>>("predictions").n_rows == 1);
 
   // Check that initial output and final output matrix
   // using saved model are same.
-  CheckMatrices(output, params.Get<arma::Row<size_t>>("output"));
+  CheckMatrices(output, params.Get<arma::Row<size_t>>("predictions"));
 }
 
 /**
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/preprocess_split_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/preprocess_split_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/preprocess_split_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/preprocess_split_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -222,7 +222,7 @@ TEST_CASE_METHOD(
   REQUIRE(params.Get<arma::mat>("test").n_cols ==
       std::floor(0.1 * inputSize));
 
-  arma::mat concat = arma::join_rows(params.Get<arma::mat>("training"),
+  arma::mat concat = join_rows(params.Get<arma::mat>("training"),
       params.Get<arma::mat>("test"));
   CheckMatrices(inputData, concat);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/radical_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/radical_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/radical_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/radical_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -121,7 +121,7 @@ TEST_CASE_METHOD(RadicalTestFixture, "Ra
   RUN_BINDING();
 
   // Check that initial output and final output using two models are different.
-  REQUIRE(arma::accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
+  REQUIRE(accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
 }
 
 /**
@@ -153,7 +153,7 @@ TEST_CASE_METHOD(RadicalTestFixture, "Ra
   RUN_BINDING();
 
   // Check that initial output and final output using two models are different.
-  REQUIRE(arma::accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
+  REQUIRE(accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
 }
 
 /**
@@ -185,7 +185,7 @@ TEST_CASE_METHOD(RadicalTestFixture, "Ra
   RUN_BINDING();
 
   // Check that initial output and final output using two models are different.
-  REQUIRE(arma::accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
+  REQUIRE(accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
 }
 
 /**
@@ -218,5 +218,5 @@ TEST_CASE_METHOD(RadicalTestFixture, "Ra
   RUN_BINDING();
 
   // Check that initial output and final output using two models are different.
-  REQUIRE(arma::accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
+  REQUIRE(accu(Y == params.Get<arma::mat>("output_ic")) < Y.n_elem);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/softmax_regression_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/softmax_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/softmax_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/softmax_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -130,8 +130,8 @@ TEST_CASE_METHOD(
   predictions = std::move(params.Get<arma::Row<size_t>>("predictions"));
 
   // Reset passed parameters.
-  SoftmaxRegression* m = params.Get<SoftmaxRegression*>("output_model");
-  params.Get<SoftmaxRegression*>("output_model") = NULL;
+  SoftmaxRegression<>* m = params.Get<SoftmaxRegression<>*>("output_model");
+  params.Get<SoftmaxRegression<>*>("output_model") = NULL;
   CleanMemory();
   ResetSettings();
 
@@ -262,7 +262,7 @@ TEST_CASE_METHOD(
 
   // Input pre-trained model.
   SetInputParam("input_model",
-                params.Get<SoftmaxRegression*>("output_model"));
+                params.Get<SoftmaxRegression<>*>("output_model"));
 
   REQUIRE_THROWS_AS(RUN_BINDING(), std::runtime_error);
 }
@@ -308,7 +308,7 @@ TEST_CASE_METHOD(
 
   // Store output parameters.
   arma::mat modelParam;
-  modelParam = params.Get<SoftmaxRegression*>("output_model")->Parameters();
+  modelParam = params.Get<SoftmaxRegression<>*>("output_model")->Parameters();
 
   // Reset passed parameters.
   CleanMemory();
@@ -329,7 +329,7 @@ TEST_CASE_METHOD(
   for (size_t i = 0; i < modelParam.n_elem; ++i)
   {
     REQUIRE(modelParam[i] !=
-        params.Get<SoftmaxRegression*>("output_model")->Parameters()[i]);
+        params.Get<SoftmaxRegression<>*>("output_model")->Parameters()[i]);
   }
 }
 
@@ -374,7 +374,7 @@ TEST_CASE_METHOD(
 
   // Store output parameters.
   arma::mat modelParam;
-  modelParam = params.Get<SoftmaxRegression*>("output_model")->Parameters();
+  modelParam = params.Get<SoftmaxRegression<>*>("output_model")->Parameters();
 
   // Reset passed parameters.
   CleanMemory();
@@ -395,7 +395,7 @@ TEST_CASE_METHOD(
   for (size_t i = 0; i < modelParam.n_elem; ++i)
   {
     REQUIRE(modelParam[i] !=
-        params.Get<SoftmaxRegression*>("output_model")->Parameters()[i]);
+        params.Get<SoftmaxRegression<>*>("output_model")->Parameters()[i]);
   }
 }
 
@@ -440,7 +440,7 @@ TEST_CASE_METHOD(
 
   // Store output parameters.
   arma::mat modelParam;
-  modelParam = params.Get<SoftmaxRegression*>("output_model")->Parameters();
+  modelParam = params.Get<SoftmaxRegression<>*>("output_model")->Parameters();
 
   // Reset passed parameters.
   CleanMemory();
@@ -458,7 +458,7 @@ TEST_CASE_METHOD(
   // Check that initial parameters has 1 more parameter than
   // final parameters matrix.
   REQUIRE(
-      params.Get<SoftmaxRegression*>("output_model")->Parameters().n_cols ==
+      params.Get<SoftmaxRegression<>*>("output_model")->Parameters().n_cols ==
       modelParam.n_cols + 1);
 }
 
@@ -502,7 +502,7 @@ TEST_CASE_METHOD(
   // the probabilities sum to 1.
   for (size_t i = 0; i < probabilities.n_cols; ++i)
   {
-    const double sum = arma::accu(probabilities.col(i));
+    const double sum = accu(probabilities.col(i));
     REQUIRE(sum == Approx(1.0));
 
     size_t classPrediction = (size_t) arma::index_max(probabilities.col(i));
diff -pruN 4.3.0-2/src/mlpack/tests/main_tests/sparse_coding_test.cpp 4.4.0-1/src/mlpack/tests/main_tests/sparse_coding_test.cpp
--- 4.3.0-2/src/mlpack/tests/main_tests/sparse_coding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/main_tests/sparse_coding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -234,7 +234,7 @@ TEST_CASE_METHOD(SparseCodingTestFixture
   LoadData(inputData, testData);
 
   mat initialDictionary = inputData.cols(0, 1);
-  SparseCoding* c = new SparseCoding();
+  SparseCoding<>* c = new SparseCoding<>();
 
   // Input trained model and initial_dictionary.
   SetInputParam("input_model", c);
@@ -340,8 +340,8 @@ TEST_CASE_METHOD(SparseCodingTestFixture
   arma::mat codes = std::move(params.Get<arma::mat>("codes"));
 
   // Reset passed parameters.
-  SparseCoding* m = params.Get<SparseCoding*>("output_model");
-  params.Get<SparseCoding*>("output_model") = NULL;
+  SparseCoding<>* m = params.Get<SparseCoding<>*>("output_model");
+  params.Get<SparseCoding<>*>("output_model") = NULL;
   CleanMemory();
   ResetSettings();
 
@@ -420,11 +420,10 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
 
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
 
 /**
@@ -469,10 +468,9 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
 
 /**
@@ -518,11 +516,10 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
 
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
 
 /**
@@ -567,11 +564,10 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
 
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
 
 /**
@@ -616,11 +612,10 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
 
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
 
 /**
@@ -667,9 +662,8 @@ TEST_CASE_METHOD(SparseCodingTestFixture
 
   // Check that initial outputs and final outputs
   // using two models model are different.
-  REQUIRE(arma::accu(dictionary ==
+  REQUIRE(accu(dictionary ==
       params.Get<arma::mat>("dictionary")) < dictionary.n_elem);
 
-  REQUIRE(arma::accu(codes ==
-      params.Get<arma::mat>("codes")) < codes.n_elem);
+  REQUIRE(accu(codes == params.Get<arma::mat>("codes")) < codes.n_elem);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/metric_test.cpp 4.4.0-1/src/mlpack/tests/metric_test.cpp
--- 4.3.0-2/src/mlpack/tests/metric_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/metric_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,10 +34,10 @@ TEST_CASE("L1MetricTest", "[MetricTest]"
 
   ManhattanDistance lMetric;
 
-  REQUIRE((double) arma::accu(arma::abs(a1 - b1)) ==
+  REQUIRE((double) accu(arma::abs(a1 - b1)) ==
       Approx(lMetric.Evaluate(a1, b1)).epsilon(1e-7));
 
-  REQUIRE((double) arma::accu(arma::abs(a2 - b2)) ==
+  REQUIRE((double) accu(arma::abs(a2 - b2)) ==
       Approx(lMetric.Evaluate(a2, b2)).epsilon(1e-7));
 }
 
@@ -60,10 +60,10 @@ TEST_CASE("L2MetricTest", "[MetricTest]"
 
   EuclideanDistance lMetric;
 
-  REQUIRE((double) sqrt(arma::accu(arma::square(a1 - b1))) ==
+  REQUIRE((double) sqrt(accu(square(a1 - b1))) ==
       Approx(lMetric.Evaluate(a1, b1)).epsilon(1e-7));
 
-  REQUIRE((double) sqrt(arma::accu(arma::square(a2 - b2))) ==
+  REQUIRE((double) sqrt(accu(square(a2 - b2))) ==
       Approx(lMetric.Evaluate(a2, b2)).epsilon(1e-7));
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/nbc_test.cpp 4.4.0-1/src/mlpack/tests/nbc_test.cpp
--- 4.3.0-2/src/mlpack/tests/nbc_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/nbc_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -397,3 +397,104 @@ TEST_CASE("NaiveBayesClassifierHighDimen
   for (size_t i = 0; i < calcVec.n_cols; ++i)
     REQUIRE(calcVec(i) == testLabels(i));
 }
+
+/**
+ * Test that we can reset the model.
+ */
+TEST_CASE("NBCResetTest", "[NBCTest]")
+{
+  const char* trainFilename = "trainSet.csv";
+
+  arma::mat trainData;
+  if (!data::Load(trainFilename, trainData))
+    FAIL("Cannot load dataset");
+
+  // Get the labels out.
+  arma::Row<size_t> labels(trainData.n_cols);
+  for (size_t i = 0; i < trainData.n_cols; ++i)
+    labels[i] = trainData(trainData.n_rows - 1, i);
+  trainData.shed_row(trainData.n_rows - 1);
+
+  NaiveBayesClassifier<> nbc1, nbc2;
+
+  nbc1.Train(trainData, labels, 2);
+  nbc2.Train(trainData, labels, 2);
+
+  REQUIRE(approx_equal(nbc1.Probabilities(), nbc2.Probabilities(), "absdiff",
+      1e-5));
+  REQUIRE(approx_equal(nbc1.Means(), nbc2.Means(), "absdiff", 1e-5));
+  REQUIRE(approx_equal(nbc1.Variances(), nbc2.Variances(), "absdiff", 1e-5));
+
+  // Now reset one model but train the other.  Modify the training set slightly.
+  trainData += 0.1 * arma::randu<arma::mat>(trainData.n_rows, trainData.n_cols);
+
+  nbc1.Train(trainData, labels, 2);
+  nbc2.Reset();
+  nbc2.Train(trainData, labels, 2);
+
+  REQUIRE(!approx_equal(nbc1.Probabilities(), nbc2.Probabilities(), "absdiff",
+      1e-5));
+  REQUIRE(!approx_equal(nbc1.Means(), nbc2.Means(), "absdiff", 1e-5));
+  REQUIRE(!approx_equal(nbc1.Variances(), nbc2.Variances(), "absdiff", 1e-5));
+}
+
+/**
+ * Test that we can use a model for incremental point-by-point training.
+ */
+TEMPLATE_TEST_CASE("NBCIncrementalTest", "[NBCTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+
+  const char* trainFilename = "trainSet.csv";
+
+  MatType trainData;
+  if (!data::Load(trainFilename, trainData))
+    FAIL("Cannot load dataset");
+
+  // Get the labels out.
+  arma::Row<size_t> labels(trainData.n_cols);
+  for (size_t i = 0; i < trainData.n_cols; ++i)
+    labels[i] = trainData(trainData.n_rows - 1, i);
+  trainData.shed_row(trainData.n_rows - 1);
+
+  NaiveBayesClassifier<MatType> nbc(trainData.n_rows, 2);
+
+  for (size_t i = 0; i < trainData.n_cols; ++i)
+  {
+    nbc.Train(trainData.col(i), labels[i]);
+  }
+
+  // We don't care what the result is, we are more concerned with the fact that
+  // training worked at all.
+  REQUIRE(nbc.Probabilities().n_elem == 2);
+  REQUIRE(nbc.Means().n_rows == trainData.n_rows);
+  REQUIRE(nbc.Means().n_cols == 2);
+  REQUIRE(nbc.Variances().n_rows == trainData.n_rows);
+  REQUIRE(nbc.Variances().n_cols == 2);
+}
+
+/**
+ * Test that we can train on sparse data with different internal model types.
+ */
+TEMPLATE_TEST_CASE("NBCModelMatTypeTest", "[NBCTest]", float, double)
+{
+  typedef TestType ElemType;
+
+  NaiveBayesClassifier<arma::Mat<ElemType>> nbc;
+
+  // Create random data; 5000 points in 4 classes.
+  arma::SpMat<ElemType> data;
+  data.sprandu(100, 5000, 0.2);
+  arma::Row<size_t> labels =
+      arma::randi<arma::Row<size_t>>(5000, arma::distr_param(0, 3));
+
+  nbc.Train(data, labels, 4);
+
+  // We don't care what the result is, we are more concerned with the fact that
+  // training worked at all.
+  REQUIRE(nbc.Probabilities().n_elem == 4);
+  REQUIRE(nbc.Means().n_rows == data.n_rows);
+  REQUIRE(nbc.Means().n_cols == 4);
+  REQUIRE(nbc.Variances().n_rows == data.n_rows);
+  REQUIRE(nbc.Variances().n_cols == 4);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/nmf_test.cpp 4.4.0-1/src/mlpack/tests/nmf_test.cpp
--- 4.3.0-2/src/mlpack/tests/nmf_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/nmf_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -153,11 +153,11 @@ TEST_CASE("SparseNMFAcolDistTest", "[NMF
     // Get an initialization.
     arma::mat iw, ih;
     RandomAcolInitialization<>::Initialize(v, r, iw, ih);
-    GivenInitialization g(std::move(iw), std::move(ih));
+    GivenInitialization<> g(std::move(iw), std::move(ih));
 
     // The GivenInitialization will force the same initialization for both
     // Apply() calls.
-    AMF<SimpleResidueTermination, GivenInitialization> nmf(srt, g);
+    AMF<SimpleResidueTermination, GivenInitialization<>> nmf(srt, g);
     nmf.Apply(v, r, w, h);
     nmf.Apply(dv, r, dw, dh);
 
@@ -212,11 +212,11 @@ TEST_CASE("SparseNMFALSTest", "[NMFTest]
       // Get an initialization.
       arma::mat iw, ih;
       RandomAcolInitialization<>::Initialize(v, r, iw, ih);
-      GivenInitialization g(std::move(iw), std::move(ih));
+      GivenInitialization<> g(std::move(iw), std::move(ih));
 
       SimpleResidueTermination srt(1e-10, 10000);
-      AMF<SimpleResidueTermination, GivenInitialization, NMFALSUpdate> nmf(srt,
-          g);
+      AMF<SimpleResidueTermination, GivenInitialization<>, NMFALSUpdate>
+          nmf(srt, g);
       nmf.Apply(v, r, w, h);
       nmf.Apply(dv, r, dw, dh);
 
@@ -255,8 +255,8 @@ TEST_CASE("NonNegNMFDefaultTest", "[NMFT
   AMF<> nmf;
   nmf.Apply(v, r, w, h);
 
-  REQUIRE((arma::all(arma::vectorise(w) >= 0)
-      && arma::all(arma::vectorise(h) >= 0)));
+  REQUIRE((arma::all(vectorise(w) >= 0)
+      && arma::all(vectorise(h) >= 0)));
 }
 
 /**
@@ -277,8 +277,8 @@ TEST_CASE("NonNegNMFRandomDivTest", "[NM
       NMFMultiplicativeDivergenceUpdate> nmf(srt);
   nmf.Apply(v, r, w, h);
 
-  REQUIRE((arma::all(arma::vectorise(w) >= 0)
-      && arma::all(arma::vectorise(h) >= 0)));
+  REQUIRE((arma::all(vectorise(w) >= 0)
+      && arma::all(vectorise(h) >= 0)));
 }
 
 /**
@@ -298,6 +298,35 @@ TEST_CASE("NonNegNMFALSTest", "[NMFTest]
       NMFALSUpdate> nmf(srt);
   nmf.Apply(v, r, w, h);
 
-  REQUIRE((arma::all(arma::vectorise(w) >= 0)
-      && arma::all(arma::vectorise(h) >= 0)));
+  REQUIRE((arma::all(vectorise(w) >= 0)
+      && arma::all(vectorise(h) >= 0)));
+}
+
+/**
+ * Check that NoInitialization doesn't do anything to the elements of a matrix.
+ */
+TEMPLATE_TEST_CASE("NoInitializationTest", "[NMFTest]", float, double)
+{
+  typedef TestType eT;
+
+  arma::Mat<eT> W, H;
+  W.randu(100, 5);
+  H.randu(5, 50);
+  arma::Mat<eT> oldW = W;
+  arma::Mat<eT> oldH = H;
+
+  arma::Mat<eT> V(100, 50, fill::zeros);
+
+  REQUIRE_NOTHROW(NoInitialization::Initialize(V, 5, W, H));
+  REQUIRE_NOTHROW(NoInitialization::InitializeOne(V, 5, W, true));
+  REQUIRE_NOTHROW(NoInitialization::InitializeOne(V, 5, H, false));
+
+  REQUIRE(arma::approx_equal(oldW, W, "both", 1e-5, 1e-5));
+  REQUIRE(arma::approx_equal(oldH, H, "both", 1e-5, 1e-5));
+
+  // And if something is the wrong size, it should throw an exception.
+  W.set_size(10, 10);
+
+  REQUIRE_THROWS(NoInitialization::Initialize(V, 5, W, H));
+  REQUIRE_THROWS(NoInitialization::InitializeOne(V, 5, W, true));
 }
diff -pruN 4.3.0-2/src/mlpack/tests/pca_test.cpp 4.4.0-1/src/mlpack/tests/pca_test.cpp
--- 4.3.0-2/src/mlpack/tests/pca_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/pca_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -330,3 +330,180 @@ TEST_CASE("PCAScalingTest", "[PCATest]")
   // The eigenvalues should sum to three.
   REQUIRE(accu(eigval) == Approx(3.0).epsilon(0.001));
 }
+
+/**
+ * Test PCA on a subview of a matrix with different decomposition strategies.
+ */
+TEMPLATE_TEST_CASE("PCASubviewTest", "[PCATest]", ExactSVDPolicy,
+    RandomizedSVDPCAPolicy, RandomizedBlockKrylovSVDPolicy, QUICSVDPolicy)
+{
+  typedef TestType DecompositionPolicy;
+
+  // Generate an artifical dataset in 10 dimensions.
+  arma::mat data(3, 5000);
+
+  arma::vec mean("1.0 3.0 -12.0");
+  arma::mat cov("1.0 0.9 0.0;"
+                "0.9 1.0 0.0;"
+                "0.0 0.0 12.0");
+  GaussianDistribution g(mean, cov);
+
+  for (size_t i = 0; i < 5000; ++i)
+    data.col(i) = g.Random();
+
+  // Compute PCA on the first 2000 points.
+  arma::mat transData1, transData2, transData3, eigvec;
+  arma::vec eigval1, eigval2;
+
+  PCA<DecompositionPolicy> p;
+  p.Apply(data.cols(0, 1999), transData1);
+  p.Apply(data.cols(0, 1999), transData2, eigval1);
+  p.Apply(data.cols(0, 1999), transData3, eigval2, eigvec);
+
+  // Only check for deterministic policies.
+  if (std::is_same<DecompositionPolicy, ExactSVDPolicy>::value)
+  {
+    arma::mat trueTransData, trueEigvec;
+    arma::vec trueEigval;
+
+    arma::mat dataSub = data.cols(0, 1999);
+    p.Apply(dataSub, trueTransData, trueEigval, trueEigvec);
+
+    REQUIRE(arma::approx_equal(transData1, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(transData2, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(transData3, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigval1, trueEigval, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigval2, trueEigval, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigvec, trueEigvec, "both", 1e-5, 1e-5));
+  }
+}
+
+/**
+ * Test PCA on an input expression.
+ */
+TEMPLATE_TEST_CASE("PCAExpressionTest", "[PCATest]", ExactSVDPolicy,
+    RandomizedSVDPCAPolicy, RandomizedBlockKrylovSVDPolicy, QUICSVDPolicy)
+{
+  typedef TestType DecompositionPolicy;
+
+  // Generate an artifical dataset in 10 dimensions.
+  arma::mat data(3, 5000);
+
+  arma::vec mean("1.0 3.0 -12.0");
+  arma::mat cov("1.0 0.9 0.0;"
+                "0.9 1.0 0.0;"
+                "0.0 0.0 12.0");
+  GaussianDistribution g(mean, cov);
+
+  for (size_t i = 0; i < 5000; ++i)
+    data.col(i) = g.Random();
+
+  // Compute PCA on an expression involving the input matrix.
+  arma::mat transData1, transData2, transData3, eigvec;
+  arma::vec eigval1, eigval2;
+
+  PCA<DecompositionPolicy> p;
+  p.Apply(2 * data + 1, transData1);
+  p.Apply(2 * data + 1, transData2, eigval1);
+  p.Apply(2 * data + 1, transData3, eigval2, eigvec);
+
+  // Only check for deterministic policies.
+  if (std::is_same<DecompositionPolicy, ExactSVDPolicy>::value)
+  {
+    arma::mat trueTransData, trueEigvec;
+    arma::vec trueEigval;
+
+    arma::mat dataSub = 2 * data + 1;
+    p.Apply(dataSub, trueTransData, trueEigval, trueEigvec);
+
+    REQUIRE(arma::approx_equal(transData1, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(transData2, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(transData3, trueTransData, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigval1, trueEigval, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigval2, trueEigval, "both", 1e-5, 1e-5));
+    REQUIRE(arma::approx_equal(eigvec, trueEigvec, "both", 1e-5, 1e-5));
+  }
+}
+
+/**
+ * Test PCA on 32-bit data.
+ */
+TEMPLATE_TEST_CASE("PCAFloatTest", "[PCATest]", ExactSVDPolicy,
+    RandomizedSVDPCAPolicy, RandomizedBlockKrylovSVDPolicy, QUICSVDPolicy)
+{
+  typedef TestType DecompositionPolicy;
+
+  // Generate an artifical dataset in 10 dimensions.
+  arma::fmat data(3, 5000);
+
+  arma::vec mean("1.0 3.0 -12.0");
+  arma::mat cov("1.0 0.9 0.0;"
+                "0.9 1.0 0.0;"
+                "0.0 0.0 12.0");
+  GaussianDistribution g(mean, cov);
+
+  for (size_t i = 0; i < 5000; ++i)
+    data.col(i) = arma::conv_to<arma::fvec>::from(g.Random());
+
+  // Compute PCA on the floating-point data.
+  arma::fmat coeff, coeff1, score, score1;
+  arma::fvec eigVal, eigVal1;
+
+  PCA<DecompositionPolicy> p;
+  p.Apply(data, score1, eigVal1, coeff1);
+
+  princomp(coeff, score, eigVal, trans(data));
+
+  // Verify the PCA results based on the eigenvalues.  We don't check for
+  // QUIC-SVD, since that method has a lot of noise.
+  if (!std::is_same<DecompositionPolicy, QUICSVDPolicy>::value)
+  {
+    for (size_t i = 0; i < eigVal.n_elem; ++i)
+    {
+      if (eigVal[i] == 0.0)
+        REQUIRE(eigVal1[i] == Approx(0.0).margin(1e-5));
+      else
+        REQUIRE(eigVal[i] == Approx(eigVal1[i]).epsilon(1e-3));
+    }
+  }
+}
+
+/**
+ * Test that we can convert sparse input matrices to dense output matrices.
+ * We check the general shape of the output, but not the exact details---those
+ * are handled in other tests.
+ */
+TEMPLATE_TEST_CASE("PCASparseToDenseTest", "[PCATest]", float, double)
+{
+  typedef arma::Mat<TestType> MatType;
+  typedef arma::SpMat<TestType> SpMatType;
+
+  SpMatType dataset;
+  dataset.sprandu(1000, 50000, 0.01);
+  MatType transformedDataset1, transformedDataset2;
+
+  PCA<> p;
+  const double varRetained1 = p.Apply(dataset, transformedDataset1, 5);
+  const double varRetained2 = p.Apply(dataset, transformedDataset2, 0.6);
+
+  REQUIRE(transformedDataset1.n_cols == dataset.n_cols);
+  REQUIRE(transformedDataset2.n_cols == dataset.n_cols);
+  REQUIRE(varRetained1 >= 0.0);
+  REQUIRE(varRetained1 <= 1.0);
+  REQUIRE(varRetained2 >= 0.0);
+  REQUIRE(varRetained2 <= 1.0);
+  REQUIRE(transformedDataset1.n_rows == 5);
+  REQUIRE(transformedDataset2.n_rows <= dataset.n_rows);
+
+  // Ensure we get basically the same as if we had done it to a dense matrix.
+  MatType denseData1(dataset);
+  MatType denseData2(dataset);
+
+  const double varRetained3 = p.Apply(denseData1, 5);
+  const double varRetained4 = p.Apply(denseData2, 0.6);
+
+  REQUIRE(varRetained1 == Approx(varRetained3));
+  REQUIRE(varRetained2 == Approx(varRetained4));
+
+  REQUIRE(denseData2.n_rows == transformedDataset2.n_rows);
+}
diff -pruN 4.3.0-2/src/mlpack/tests/perceptron_test.cpp 4.4.0-1/src/mlpack/tests/perceptron_test.cpp
--- 4.3.0-2/src/mlpack/tests/perceptron_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/perceptron_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -42,20 +42,20 @@ TEST_CASE("SimpleWeightUpdateWeights", "
   wip.UpdateWeights(trainingPoint, weights, biases, incorrectClass,
                     correctClass);
 
-  CHECK(weights(0, 0) == -1);
-  CHECK(weights(1, 0) == 0);
-  CHECK(weights(2, 0) == 1);
-  CHECK(weights(3, 0) == 2);
-  CHECK(weights(4, 0) == 3);
-
-  CHECK(weights(0, 2) == 7);
-  CHECK(weights(1, 2) == 8);
-  CHECK(weights(2, 2) == 9);
-  CHECK(weights(3, 2) == 10);
-  CHECK(weights(4, 2) == 11);
+  REQUIRE(weights(0, 0) == -1);
+  REQUIRE(weights(1, 0) == 0);
+  REQUIRE(weights(2, 0) == 1);
+  REQUIRE(weights(3, 0) == 2);
+  REQUIRE(weights(4, 0) == 3);
+
+  REQUIRE(weights(0, 2) == 7);
+  REQUIRE(weights(1, 2) == 8);
+  REQUIRE(weights(2, 2) == 9);
+  REQUIRE(weights(3, 2) == 10);
+  REQUIRE(weights(4, 2) == 11);
 
-  CHECK(biases(0) == 1);
-  CHECK(biases(2) == 8);
+  REQUIRE(biases(0) == 1);
+  REQUIRE(biases(2) == 8);
 }
 
 /**
@@ -85,20 +85,20 @@ TEST_CASE("SimpleWeightUpdateInstanceWei
   wip.UpdateWeights(trainingPoint, weights, biases, incorrectClass,
                     correctClass, instanceWeight);
 
-  CHECK(weights(0, 0) == -3);
-  CHECK(weights(1, 0) == -4);
-  CHECK(weights(2, 0) == -5);
-  CHECK(weights(3, 0) == -6);
-  CHECK(weights(4, 0) == -7);
-
-  CHECK(weights(0, 2) == 9);
-  CHECK(weights(1, 2) == 12);
-  CHECK(weights(2, 2) == 15);
-  CHECK(weights(3, 2) == 18);
-  CHECK(weights(4, 2) == 21);
+  REQUIRE(weights(0, 0) == -3);
+  REQUIRE(weights(1, 0) == -4);
+  REQUIRE(weights(2, 0) == -5);
+  REQUIRE(weights(3, 0) == -6);
+  REQUIRE(weights(4, 0) == -7);
+
+  REQUIRE(weights(0, 2) == 9);
+  REQUIRE(weights(1, 2) == 12);
+  REQUIRE(weights(2, 2) == 15);
+  REQUIRE(weights(3, 2) == 18);
+  REQUIRE(weights(4, 2) == 21);
 
-  CHECK(biases(0) == -1);
-  CHECK(biases(2) == 10);
+  REQUIRE(biases(0) == -1);
+  REQUIRE(biases(2) == 10);
 }
 
 /**
@@ -120,10 +120,21 @@ TEST_CASE("And", "[PerceptronTest]")
   Row<size_t> predictedLabels;
   p.Classify(testData, predictedLabels);
 
-  CHECK(predictedLabels(0, 0) == 0);
-  CHECK(predictedLabels(0, 1) == 0);
-  CHECK(predictedLabels(0, 2) == 1);
-  CHECK(predictedLabels(0, 3) == 0);
+  REQUIRE(predictedLabels(0) == 0);
+  REQUIRE(predictedLabels(1) == 0);
+  REQUIRE(predictedLabels(2) == 1);
+  REQUIRE(predictedLabels(3) == 0);
+
+  // Test single-point classify too.
+  predictedLabels(0) = p.Classify(testData.col(0));
+  predictedLabels(1) = p.Classify(testData.col(1));
+  predictedLabels(2) = p.Classify(testData.col(2));
+  predictedLabels(3) = p.Classify(testData.col(3));
+
+  REQUIRE(predictedLabels(0) == 0);
+  REQUIRE(predictedLabels(1) == 0);
+  REQUIRE(predictedLabels(2) == 1);
+  REQUIRE(predictedLabels(3) == 0);
 }
 
 /**
@@ -146,10 +157,10 @@ TEST_CASE("Or", "[PerceptronTest]")
   Row<size_t> predictedLabels;
   p.Classify(testData, predictedLabels);
 
-  CHECK(predictedLabels(0, 0) == 1);
-  CHECK(predictedLabels(0, 1) == 1);
-  CHECK(predictedLabels(0, 2) == 1);
-  CHECK(predictedLabels(0, 3) == 0);
+  REQUIRE(predictedLabels(0, 0) == 1);
+  REQUIRE(predictedLabels(0, 1) == 1);
+  REQUIRE(predictedLabels(0, 2) == 1);
+  REQUIRE(predictedLabels(0, 3) == 0);
 }
 
 /**
@@ -174,7 +185,7 @@ TEST_CASE("Random3", "[PerceptronTest]")
   p.Classify(testData, predictedLabels);
 
   for (size_t i = 0; i < predictedLabels.n_cols; ++i)
-    CHECK(predictedLabels(0, i) == 0);
+    REQUIRE(predictedLabels(0, i) == 0);
 }
 
 /**
@@ -198,35 +209,40 @@ TEST_CASE("TwoPoints", "[PerceptronTest]
   Row<size_t> predictedLabels;
   p.Classify(testData, predictedLabels);
 
-  CHECK(predictedLabels(0, 0) == 0);
-  CHECK(predictedLabels(0, 1) == 1);
+  REQUIRE(predictedLabels(0, 0) == 0);
+  REQUIRE(predictedLabels(0, 1) == 1);
 }
 
 /**
  * This tests the convergence of the perceptron on a dataset which has a
- * non-linearly separable dataset.
+ * non-linearly separable dataset.  We test on multiple element types to ensure
+ * that MatType can be set correctly.
  */
-TEST_CASE("NonLinearlySeparableDataset", "[PerceptronTest]")
+TEMPLATE_TEST_CASE("NonLinearlySeparableDataset", "[PerceptronTest]", float,
+    double)
 {
-  mat trainData;
+  typedef TestType eT;
+
+  Mat<eT> trainData;
   trainData = { { 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7, 8 },
                 { 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2 } };
 
   Mat<size_t> labels;
   labels = { 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1 };
 
-  Perceptron<> p(trainData, labels.row(0), 2, 1000);
+  Perceptron<SimpleWeightUpdate, ZeroInitialization, Mat<eT>> p(trainData,
+      labels.row(0), 2, 1000);
 
-  mat testData;
+  Mat<eT> testData;
   testData = { { 3,   4,   5,   6 },
                { 3, 2.3, 1.7, 1.5 } };
   Row<size_t> predictedLabels;
   p.Classify(testData, predictedLabels);
 
-  CHECK(predictedLabels(0, 0) == 0);
-  CHECK(predictedLabels(0, 1) == 0);
-  CHECK(predictedLabels(0, 2) == 1);
-  CHECK(predictedLabels(0, 3) == 1);
+  REQUIRE(predictedLabels(0, 0) == 0);
+  REQUIRE(predictedLabels(0, 1) == 0);
+  REQUIRE(predictedLabels(0, 2) == 1);
+  REQUIRE(predictedLabels(0, 3) == 1);
 }
 
 TEST_CASE("SecondaryConstructor", "[PerceptronTest]")
@@ -266,3 +282,84 @@ TEST_CASE("InstanceWeightsConstructor",
 
   REQUIRE(p.Weights().n_elem > 0);
 }
+
+/**
+ * This tests that incremental training can be stopped with `Reset()`.
+ */
+TEST_CASE("IncrementalTrainingTest", "[PerceptronTest]")
+{
+  mat trainData = randu<mat>(10, 1000);
+  for (size_t i = 0; i < 500; ++i)
+    trainData.col(i) += 0.6;
+  for (size_t i = 500; i < 1000; ++i)
+    trainData.col(i) += 0.8;
+  Row<size_t> labels(1000);
+  labels.subvec(0, 499).zeros();
+  labels.subvec(500, 999).ones();
+
+  Perceptron<> p1, p2;
+
+  // This should result in the same model, because the default initialization is
+  // zeros.
+  p1.Train(trainData, labels, 2, 1);
+  p2.Train(trainData, labels, 2, 1);
+
+  REQUIRE(approx_equal(p1.Weights(), p2.Weights(), "absdiff", 1e-5));
+  REQUIRE(approx_equal(p1.Biases(), p2.Biases(), "absdiff", 1e-5));
+
+  // Resetting and retraining p2 should result in the same model.
+  p2.Reset();
+  p2.Train(trainData, labels, 2);
+
+  REQUIRE(approx_equal(p1.Weights(), p2.Weights(), "absdiff", 1e-5));
+  REQUIRE(approx_equal(p1.Biases(), p2.Biases(), "absdiff", 1e-5));
+
+  // Training p1 again should result in a different model.
+  p1.Train(trainData, labels, 2);
+
+  // The biases often converge to the same -1/0/1 values, but it's sufficient
+  // to just check the weights.
+  REQUIRE(!approx_equal(p1.Weights(), p2.Weights(), "absdiff", 1e-5));
+}
+
+// Test all forms of Train().
+TEST_CASE("TrainFormTest", "[PerceptronTest]")
+{
+  mat trainData;
+  trainData = { { 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7, 8 },
+                { 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2 } };
+
+  Row<size_t> labels;
+  labels = { 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1 };
+
+  rowvec weights(labels.n_elem);
+  weights.ones();
+
+  Perceptron<> p1(trainData, labels, 2);
+  Perceptron<> p2(trainData, labels, 2, weights);
+  Perceptron<> p3, p4, p5, p6;
+  p3.Train(trainData, labels, 2);
+  p4.Train(trainData, labels, 2, 50);
+  p5.Train(trainData, labels, 2, weights);
+  p6.Train(trainData, labels, 2, weights, 50);
+
+  mat testData;
+  testData = { { 3,   4,   5,   6 },
+               { 3, 2.3, 1.7, 1.5 } };
+  Row<size_t> predictions1, predictions2, predictions3, predictions4,
+      predictions5, predictions6;
+  Row<size_t> trueLabels = { 0, 0, 1, 1 };
+  p1.Classify(testData, predictions1);
+  p2.Classify(testData, predictions2);
+  p3.Classify(testData, predictions3);
+  p4.Classify(testData, predictions4);
+  p5.Classify(testData, predictions5);
+  p6.Classify(testData, predictions6);
+
+  REQUIRE(all(predictions1 == trueLabels));
+  REQUIRE(all(predictions2 == trueLabels));
+  REQUIRE(all(predictions3 == trueLabels));
+  REQUIRE(all(predictions4 == trueLabels));
+  REQUIRE(all(predictions5 == trueLabels));
+  REQUIRE(all(predictions6 == trueLabels));
+}
diff -pruN 4.3.0-2/src/mlpack/tests/policy_gradient_test.cpp 4.4.0-1/src/mlpack/tests/policy_gradient_test.cpp
--- 4.3.0-2/src/mlpack/tests/policy_gradient_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/policy_gradient_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -56,7 +56,7 @@ TEST_CASE("PendulumWithDDPG", "[PolicyGr
     int size = 1; 
     double mu = 0.0;
     double theta = 1.0; 
-    double sigma = 0.1;  
+    double sigma = 0.01;
 
     // Create an instance of the OUNoise class.
     OUNoise ouNoise(size, mu, theta, sigma);
@@ -108,7 +108,7 @@ TEST_CASE("PendulumWithGaussianDDPG", "[
     // Set up the GaussianNoise parameters.
     int size = 1; 
     double mu = 0.0; 
-    double sigma = 0.1; 
+    double sigma = 0.01;
 
     // Create an instance of the GaussianNoise class.
     GaussianNoise gaussianNoise(size, mu, sigma);
@@ -153,7 +153,7 @@ TEST_CASE("DDPGForMultipleActions", "[Po
   int size = 4;
   double mu = 0.0;
   double theta = 1.0;  
-  double sigma = 0.1; 
+  double sigma = 0.01;
 
   // Create an instance of the OUNoise class.
   OUNoise ouNoise(size, mu, theta, sigma);
@@ -184,7 +184,7 @@ TEST_CASE("OUNoiseTest", "[PolicyGradien
   int size = 3;
   double mu = 0.0;
   double theta = 0.15;
-  double sigma = 0.2;
+  double sigma = 0.02;
 
   // Create an instance of the OUNoise class.
   OUNoise ouNoise(size, mu, theta, sigma);
@@ -192,7 +192,7 @@ TEST_CASE("OUNoiseTest", "[PolicyGradien
   // Test the reset function.
   ouNoise.reset();
   arma::colvec state = ouNoise.sample();
-  REQUIRE(state.n_elem == size);
+  REQUIRE((int) state.n_elem == size);
 
   // Verify that the sample is not equal to the reset state.
   arma::colvec sample = ouNoise.sample();
@@ -206,14 +206,14 @@ TEST_CASE("GaussianNoiseTest", "[PolicyG
   // Set up the GaussianNoise parameters.
   int size = 5; 
   double mu = 0.0; 
-  double sigma = 0.1; 
+  double sigma = 0.00001;
 
   // Create an instance of the GaussianNoise class.
   GaussianNoise gaussianNoise(size, mu, sigma);
 
   // Test the sample function.
   arma::colvec noise = gaussianNoise.sample();
-  REQUIRE(noise.n_elem == size);
+  REQUIRE((int) noise.n_elem == size);
 
   // Verify that the noise vector has values drawn from a 
   // Gaussian distribution with the specified mean and standard deviation.
@@ -222,8 +222,8 @@ TEST_CASE("GaussianNoiseTest", "[PolicyG
 
   double meanErr = mean - mu;
   double stdDevErr = stdDev - sigma;
-  REQUIRE(meanErr <= 1e-5);
-  REQUIRE(stdDevErr <= 1e-5);
+  REQUIRE(meanErr <= 1e-4);
+  REQUIRE(stdDevErr <= 1e-4);
 }
 
 //! Test TD3 on Pendulum task.
diff -pruN 4.3.0-2/src/mlpack/tests/quic_svd_test.cpp 4.4.0-1/src/mlpack/tests/quic_svd_test.cpp
--- 4.3.0-2/src/mlpack/tests/quic_svd_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/quic_svd_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -34,7 +34,7 @@ TEST_CASE("QUICSVDReconstructionError",
   {
     // Obtain the SVD using default parameters.
     arma::mat u, v, sigma;
-    QUIC_SVD quicsvd(dataset, u, v, sigma);
+    QUIC_SVD<> quicsvd(dataset, u, v, sigma);
 
     // Reconstruct the matrix using the SVD.
     arma::mat reconstruct;
@@ -64,14 +64,14 @@ TEST_CASE("QUICSVDSingularValueError", "
 
   arma::mat s = arma::diagmat(arma::vec("1 0.1 0.01"));
 
-  arma::mat data = arma::trans(U * arma::diagmat(s) * V.t());
+  arma::mat data = trans(U * arma::diagmat(s) * V.t());
 
   arma::vec s1, s3;
   arma::mat U1, U2, V1, V2, s2;
 
   // Obtain the SVD using default parameters.
   arma::svd_econ(U1, s1, V1, data);
-  QUIC_SVD quicsvd(data, U1, V1, s2);
+  QUIC_SVD<> quicsvd(data, U1, V1, s2);
 
   s3 = arma::diagvec(s2);
   s1 = s1.subvec(0, s3.n_elem - 1);
@@ -87,5 +87,5 @@ TEST_CASE("QUICSVDSameDimensionTest", "[
 
   // Obtain the SVD using default parameters.
   arma::mat u, v, sigma;
-  QUIC_SVD quicsvd(dataset, u, v, sigma);
+  QUIC_SVD<> quicsvd(dataset, u, v, sigma);
 }
diff -pruN 4.3.0-2/src/mlpack/tests/random_forest_test.cpp 4.4.0-1/src/mlpack/tests/random_forest_test.cpp
--- 4.3.0-2/src/mlpack/tests/random_forest_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/random_forest_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -145,8 +145,8 @@ TEST_CASE("UnweightedNumericLearningTest
   dt.Classify(testDataset, dtPredictions);
 
   // Calculate the number of correct points.
-  size_t rfCorrect = arma::accu(rfPredictions == testLabels);
-  size_t dtCorrect = arma::accu(dtPredictions == testLabels);
+  size_t rfCorrect = accu(rfPredictions == testLabels);
+  size_t dtCorrect = accu(dtPredictions == testLabels);
 
   REQUIRE(rfCorrect >= dtCorrect * 0.9);
   REQUIRE(rfCorrect >= size_t(0.7 * testDataset.n_cols));
@@ -172,8 +172,8 @@ TEST_CASE("WeightedNumericLearningTest",
     noiseLabels[i] = RandInt(3); // Random label.
 
   // Concatenate data matrices.
-  arma::mat fullData = arma::join_rows(dataset, noise);
-  arma::Row<size_t> fullLabels = arma::join_rows(labels, noiseLabels);
+  arma::mat fullData = join_rows(dataset, noise);
+  arma::Row<size_t> fullLabels = join_rows(labels, noiseLabels);
 
   // Now set weights.
   arma::rowvec weights(dataset.n_cols + 1000);
@@ -201,8 +201,8 @@ TEST_CASE("WeightedNumericLearningTest",
   dt.Classify(testDataset, dtPredictions);
 
   // Calculate the number of correct points.
-  size_t rfCorrect = arma::accu(rfPredictions == testLabels);
-  size_t dtCorrect = arma::accu(dtPredictions == testLabels);
+  size_t rfCorrect = accu(rfPredictions == testLabels);
+  size_t dtCorrect = accu(dtPredictions == testLabels);
 
   REQUIRE(rfCorrect >= dtCorrect * 0.8);
   REQUIRE(rfCorrect >= size_t(0.7 * testDataset.n_cols));
@@ -238,8 +238,8 @@ TEST_CASE("UnweightedCategoricalLearning
   dt.Classify(testData, dtPredictions);
 
   // Calculate the number of correct points.
-  size_t rfCorrect = arma::accu(rfPredictions == testLabels);
-  size_t dtCorrect = arma::accu(dtPredictions == testLabels);
+  size_t rfCorrect = accu(rfPredictions == testLabels);
+  size_t dtCorrect = accu(dtPredictions == testLabels);
 
   REQUIRE(rfCorrect >= dtCorrect - 35);
   REQUIRE(rfCorrect >= size_t(0.7 * testData.n_cols));
@@ -280,8 +280,8 @@ TEST_CASE("WeightedCategoricalLearningTe
   for (size_t i = 2000; i < 4000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(trainingData, randomNoise);
-  arma::Row<size_t> fullLabels = arma::join_rows(trainingLabels, randomLabels);
+  arma::mat fullData = join_rows(trainingData, randomNoise);
+  arma::Row<size_t> fullLabels = join_rows(trainingLabels, randomLabels);
 
   // Build a random forest and a decision tree.
   RandomForest<> rf(fullData, di, fullLabels, 5, weights, 25 /* 25 trees */, 1,
@@ -296,8 +296,8 @@ TEST_CASE("WeightedCategoricalLearningTe
   dt.Classify(testData, dtPredictions);
 
   // Calculate the number of correct points.
-  size_t rfCorrect = arma::accu(rfPredictions == testLabels);
-  size_t dtCorrect = arma::accu(dtPredictions == testLabels);
+  size_t rfCorrect = accu(rfPredictions == testLabels);
+  size_t dtCorrect = accu(dtPredictions == testLabels);
 
   REQUIRE(rfCorrect >= dtCorrect - 25);
   REQUIRE(rfCorrect >= size_t(0.7 * testData.n_cols));
@@ -396,8 +396,8 @@ TEST_CASE("RandomForestNumericTrainRetur
     noiseLabels[i] = RandInt(3); // Random label.
 
   // Concatenate data matrices.
-  arma::mat fullData = arma::join_rows(dataset, noise);
-  arma::Row<size_t> fullLabels = arma::join_rows(labels, noiseLabels);
+  arma::mat fullData = join_rows(dataset, noise);
+  arma::Row<size_t> fullLabels = join_rows(labels, noiseLabels);
 
   // Now set weights.
   arma::rowvec weights(dataset.n_cols + 1000);
@@ -449,8 +449,8 @@ TEST_CASE("RandomForestCategoricalTrainR
   for (size_t i = 4000; i < 6000; ++i)
     weights[i] = Random(0.0, 0.001);
 
-  arma::mat fullData = arma::join_rows(d, randomNoise);
-  arma::Row<size_t> fullLabels = arma::join_rows(l, randomLabels);
+  arma::mat fullData = join_rows(d, randomNoise);
+  arma::Row<size_t> fullLabels = join_rows(l, randomLabels);
 
   // Test random forest on unweighted categorical dataset.
   RandomForest<> rf;
@@ -542,7 +542,7 @@ TEST_CASE("WarmStartTreesPredictionsQual
   rf.Classify(trainingData, oldPredictions);
 
   // Calculate the number of correct points.
-  size_t oldCorrect = arma::accu(oldPredictions == trainingLabels);
+  size_t oldCorrect = accu(oldPredictions == trainingLabels);
 
   rf.Train(trainingData, di, trainingLabels, 5, 20 /* 20 trees */, 1, 1e-7, 0,
       true /* warmStart */, MultipleRandomDimensionSelect(4));
@@ -552,7 +552,7 @@ TEST_CASE("WarmStartTreesPredictionsQual
   rf.Classify(trainingData, newPredictions);
 
   // Calculate the number of correct points.
-  size_t newCorrect = arma::accu(newPredictions == trainingLabels);
+  size_t newCorrect = accu(newPredictions == trainingLabels);
 
   REQUIRE(newCorrect >= oldCorrect);
 }
@@ -577,8 +577,8 @@ TEST_CASE("ExtraTreesAccuracyTest", "[Ra
     noiseLabels[i] = RandInt(3); // Random label.
 
   // Concatenate data matrices.
-  arma::mat fullData = arma::join_rows(dataset, noise);
-  arma::Row<size_t> fullLabels = arma::join_rows(labels, noiseLabels);
+  arma::mat fullData = join_rows(dataset, noise);
+  arma::Row<size_t> fullLabels = join_rows(labels, noiseLabels);
 
   // Now set weights.
   arma::rowvec weights(dataset.n_cols + 1000);
@@ -602,7 +602,7 @@ TEST_CASE("ExtraTreesAccuracyTest", "[Ra
   et.Classify(testDataset, predictions);
 
   // Calculate the prediction accuracy.
-  double accuracy = arma::accu(predictions == testLabels);
+  double accuracy = accu(predictions == testLabels);
   accuracy /= predictions.n_elem;
 
   REQUIRE(accuracy >= 0.85);
diff -pruN 4.3.0-2/src/mlpack/tests/randomized_svd_test.cpp 4.4.0-1/src/mlpack/tests/randomized_svd_test.cpp
--- 4.3.0-2/src/mlpack/tests/randomized_svd_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/randomized_svd_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -32,11 +32,10 @@ TEST_CASE("RandomizedSVDReconstructionEr
 
   arma::mat s = arma::diagmat(arma::vec("1 0.1 0.01"));
 
-  arma::mat data = arma::trans(U * arma::diagmat(s) * V.t());
+  arma::mat data = trans(U * arma::diagmat(s) * V.t());
 
   // Center the data into a temporary matrix.
-  arma::mat centeredData;
-  Center(data, centeredData);
+  arma::mat centeredData = data.each_col() - arma::mean(data, 1);
 
   arma::mat U1, U2, V1, V2;
   arma::vec s1, s2, s3;
diff -pruN 4.3.0-2/src/mlpack/tests/rectangle_tree_test.cpp 4.4.0-1/src/mlpack/tests/rectangle_tree_test.cpp
--- 4.3.0-2/src/mlpack/tests/rectangle_tree_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/rectangle_tree_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -1158,7 +1158,7 @@ TEST_CASE("RPlusPlusTreeTraverserTest",
 // to allow us to test by hand without adding hundreds of points.
 TEST_CASE("RTreeSplitTest", "[RectangleTreeTraitsTest]")
 {
-  arma::mat data = arma::trans(arma::mat("0.0 0.0;"
+  arma::mat data = trans(arma::mat("0.0 0.0;"
                                          "0.0 1.0;"
                                          "1.0 0.1;"
                                          "1.0 0.5;"
@@ -1253,7 +1253,7 @@ TEST_CASE("RTreeSplitTest", "[RectangleT
 // to allow us to test by hand without adding hundreds of points.
 TEST_CASE("RStarTreeSplitTest", "[RectangleTreeTraitsTest]")
 {
-  arma::mat data = arma::trans(arma::mat("0.0 0.0;"
+  arma::mat data = trans(arma::mat("0.0 0.0;"
                                          "0.0 1.0;"
                                          "1.0 0.1;"
                                          "1.0 0.5;"
diff -pruN 4.3.0-2/src/mlpack/tests/regularized_svd_test.cpp 4.4.0-1/src/mlpack/tests/regularized_svd_test.cpp
--- 4.3.0-2/src/mlpack/tests/regularized_svd_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/regularized_svd_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -52,7 +52,7 @@ TEST_CASE("RegularizedSVDFunctionRandomE
       const size_t item = data(1, j) + numUsers;
 
       const double rating = data(2, j);
-      const double ratingError = rating - arma::dot(parameters.col(user),
+      const double ratingError = rating - dot(parameters.col(user),
                                                     parameters.col(item));
       const double ratingErrorSquared = ratingError * ratingError;
 
@@ -218,7 +218,7 @@ TEST_CASE("RegularizedSVDFunctionOptimiz
   // Make rating entries based on the parameters.
   for (size_t i = 0; i < numRatings; ++i)
   {
-    data(2, i) = arma::dot(parameters.col(data(0, i)),
+    data(2, i) = dot(parameters.col(data(0, i)),
                            parameters.col(numUsers + data(1, i)));
   }
 
@@ -234,7 +234,7 @@ TEST_CASE("RegularizedSVDFunctionOptimiz
   arma::mat predictedData(1, numRatings);
   for (size_t i = 0; i < numRatings; ++i)
   {
-    predictedData(0, i) = arma::dot(optParameters.col(data(0, i)),
+    predictedData(0, i) = dot(optParameters.col(data(0, i)),
                                     optParameters.col(numUsers + data(1, i)));
   }
 
@@ -276,7 +276,7 @@ TEST_CASE("RegularizedSVDFunctionOptimiz
   // Make rating entries based on the parameters.
   for (size_t i = 0; i < numRatings; ++i)
   {
-    data(2, i) = arma::dot(parameters.col(data(0, i)),
+    data(2, i) = dot(parameters.col(data(0, i)),
                            parameters.col(numUsers + data(1, i)));
   }
 
@@ -299,7 +299,7 @@ TEST_CASE("RegularizedSVDFunctionOptimiz
   arma::mat predictedData(1, numRatings);
   for (size_t i = 0; i < numRatings; ++i)
   {
-    predictedData(0, i) = arma::dot(optParameters.col(data(0, i)),
+    predictedData(0, i) = dot(optParameters.col(data(0, i)),
                                     optParameters.col(numUsers + data(1, i)));
   }
 
diff -pruN 4.3.0-2/src/mlpack/tests/serialization.cpp 4.4.0-1/src/mlpack/tests/serialization.cpp
--- 4.3.0-2/src/mlpack/tests/serialization.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/serialization.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -14,7 +14,7 @@
 
 namespace mlpack {
 
-// Utility function to check the equality of two Armadillo matrices.
+// Utility function to check the equality of four Armadillo matrices.
 void CheckMatrices(const arma::mat& x,
                    const arma::mat& xmlX,
                    const arma::mat& jsonX,
@@ -39,15 +39,52 @@ void CheckMatrices(const arma::mat& x,
     const double val = x[i];
     if (val == 0.0)
     {
-      REQUIRE(xmlX[i] == Approx(0.0).margin(1e-6 / 100));
-      REQUIRE(jsonX[i] == Approx(0.0).margin(1e-6 / 100));
-      REQUIRE(binaryX[i] == Approx(0.0).margin(1e-6 / 100));
+      REQUIRE(xmlX[i] == Approx(0.0).margin(1e-8));
+      REQUIRE(jsonX[i] == Approx(0.0).margin(1e-8));
+      REQUIRE(binaryX[i] == Approx(0.0).margin(1e-8));
     }
     else
     {
-      REQUIRE(val == Approx(xmlX[i]).epsilon(1e-6 / 100));
-      REQUIRE(val == Approx(jsonX[i]).epsilon(1e-6 / 100));
-      REQUIRE(val == Approx(binaryX[i]).epsilon(1e-6 / 100));
+      REQUIRE(val == Approx(xmlX[i]).epsilon(1e-8));
+      REQUIRE(val == Approx(jsonX[i]).epsilon(1e-8));
+      REQUIRE(val == Approx(binaryX[i]).epsilon(1e-8));
+    }
+  }
+}
+
+void CheckMatrices(const arma::fmat& x,
+                   const arma::fmat& xmlX,
+                   const arma::fmat& jsonX,
+                   const arma::fmat& binaryX)
+{
+  // First check dimensions.
+  REQUIRE(x.n_rows == xmlX.n_rows);
+  REQUIRE(x.n_rows == jsonX.n_rows);
+  REQUIRE(x.n_rows == binaryX.n_rows);
+
+  REQUIRE(x.n_cols == xmlX.n_cols);
+  REQUIRE(x.n_cols == jsonX.n_cols);
+  REQUIRE(x.n_cols == binaryX.n_cols);
+
+  REQUIRE(x.n_elem == xmlX.n_elem);
+  REQUIRE(x.n_elem == jsonX.n_elem);
+  REQUIRE(x.n_elem == binaryX.n_elem);
+
+  // Now check elements.
+  for (size_t i = 0; i < x.n_elem; ++i)
+  {
+    const float val = x[i];
+    if (val == 0.0)
+    {
+      REQUIRE(xmlX[i] == Approx(0.0).margin(1e-6));
+      REQUIRE(jsonX[i] == Approx(0.0).margin(1e-6));
+      REQUIRE(binaryX[i] == Approx(0.0).margin(1e-6));
+    }
+    else
+    {
+      REQUIRE(val == Approx(xmlX[i]).epsilon(1e-6));
+      REQUIRE(val == Approx(jsonX[i]).epsilon(1e-6));
+      REQUIRE(val == Approx(binaryX[i]).epsilon(1e-6));
     }
   }
 }
diff -pruN 4.3.0-2/src/mlpack/tests/serialization.hpp 4.4.0-1/src/mlpack/tests/serialization.hpp
--- 4.3.0-2/src/mlpack/tests/serialization.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/serialization.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -219,6 +219,11 @@ void CheckMatrices(const arma::mat& x,
                    const arma::mat& jsonX,
                    const arma::mat& binaryX);
 
+void CheckMatrices(const arma::fmat& x,
+                   const arma::fmat& xmlX,
+                   const arma::fmat& jsonX,
+                   const arma::fmat& binaryX);
+
 void CheckMatrices(const arma::Mat<size_t>& x,
                    const arma::Mat<size_t>& xmlX,
                    const arma::Mat<size_t>& jsonX,
diff -pruN 4.3.0-2/src/mlpack/tests/serialization_test.cpp 4.4.0-1/src/mlpack/tests/serialization_test.cpp
--- 4.3.0-2/src/mlpack/tests/serialization_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/serialization_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -635,10 +635,10 @@ TEST_CASE("SoftmaxRegressionTest", "[Ser
     labels[i] = 0;
   for (size_t i = 500; i < 1000; ++i)
     labels[i] = 1;
-  SoftmaxRegression sr(dataset, labels, 2);
-  SoftmaxRegression srXml(dataset.n_rows, 2);
-  SoftmaxRegression srText(dataset.n_rows, 2);
-  SoftmaxRegression srBinary(dataset.n_rows, 2);
+  SoftmaxRegression<> sr(dataset, labels, 2);
+  SoftmaxRegression<> srXml(dataset.n_rows, 2);
+  SoftmaxRegression<> srText(dataset.n_rows, 2);
+  SoftmaxRegression<> srBinary(dataset.n_rows, 2);
 
   SerializeObjectAll(sr, srXml, srText, srBinary);
 
@@ -1061,20 +1061,18 @@ TEST_CASE("LARSTest", "[SerializationTes
   arma::vec beta = arma::randn(75, 1);
   arma::rowvec y = beta.t() * X;
 
-  LARS lars(true, 0.1, 0.1);
-  arma::vec betaOpt;
-  lars.Train(X, y, betaOpt);
+  LARS<> lars(true, 0.1, 0.1);
+  lars.Train(X, y);
 
   // Now, serialize.
-  LARS xmlLars(false, 0.5, 0.0), binaryLars(true, 1.0, 0.0),
+  LARS<> xmlLars(false, 0.5, 0.0), binaryLars(true, 1.0, 0.0),
       jsonLars(false, 0.1, 0.1);
 
   // Train jsonLars.
   arma::mat jsonX = arma::randn(25, 150);
   arma::vec jsonBeta = arma::randn(25, 1);
   arma::rowvec jsonY = jsonBeta.t() * jsonX;
-  arma::vec jsonBetaOpt;
-  jsonLars.Train(jsonX, jsonY, jsonBetaOpt);
+  jsonLars.Train(jsonX, jsonY);
 
   SerializeObjectAll(lars, xmlLars, binaryLars, jsonLars);
 
@@ -1552,12 +1550,12 @@ TEST_CASE("BayesianLinearRegressionTest"
   arma::vec omega = arma::randn(75, 1);
   arma::rowvec y = omega.t() * matX;
 
-  BayesianLinearRegression blr(false, false);
+  BayesianLinearRegression<> blr(false, false);
   blr.Train(matX, y);
   arma::vec omegaOpt = blr.Omega();
 
   // Now, serialize.
-  BayesianLinearRegression xmlBlr(false, false), binaryBlr(false, false),
+  BayesianLinearRegression<> xmlBlr(false, false), binaryBlr(false, false),
     textBlr(false, false);
 
   SerializeObjectAll(blr, xmlBlr, binaryBlr, textBlr);
diff -pruN 4.3.0-2/src/mlpack/tests/softmax_regression_test.cpp 4.4.0-1/src/mlpack/tests/softmax_regression_test.cpp
--- 4.3.0-2/src/mlpack/tests/softmax_regression_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/softmax_regression_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -33,7 +33,7 @@ TEST_CASE("SoftmaxRegressionFunctionEval
     labels(i) = RandInt(0, numClasses);
 
   // Create a SoftmaxRegressionFunction. Regularization term ignored.
-  SoftmaxRegressionFunction srf(data, labels, numClasses, 0);
+  SoftmaxRegressionFunction<> srf(data, labels, numClasses, 0);
 
   // Run a number of trials.
   for (size_t i = 0; i < trials; ++i)
@@ -49,8 +49,8 @@ TEST_CASE("SoftmaxRegressionFunctionEval
     {
       arma::mat hypothesis, probabilities;
 
-      hypothesis = arma::exp(parameters * data.col(j));
-      probabilities = hypothesis / arma::accu(hypothesis);
+      hypothesis = exp(parameters * data.col(j));
+      probabilities = hypothesis / accu(hypothesis);
 
       logLikelihood += log(probabilities(labels(j), 0));
     }
@@ -80,9 +80,9 @@ TEST_CASE("SoftmaxRegressionFunctionRegu
     labels(i) = RandInt(0, numClasses);
 
   // 3 objects for comparing regularization costs.
-  SoftmaxRegressionFunction srfNoReg(data, labels, numClasses, 0);
-  SoftmaxRegressionFunction srfSmallReg(data, labels, numClasses, 1);
-  SoftmaxRegressionFunction srfBigReg(data, labels, numClasses, 20);
+  SoftmaxRegressionFunction<> srfNoReg(data, labels, numClasses, 0);
+  SoftmaxRegressionFunction<> srfSmallReg(data, labels, numClasses, 1);
+  SoftmaxRegressionFunction<> srfBigReg(data, labels, numClasses, 20);
 
   // Run a number of trials.
   for (size_t i = 0; i < trials; ++i)
@@ -92,7 +92,7 @@ TEST_CASE("SoftmaxRegressionFunctionRegu
     parameters.randu(numClasses, inputSize);
 
     double wL2SquaredNorm;
-    wL2SquaredNorm = arma::accu(parameters % parameters);
+    wL2SquaredNorm = accu(parameters % parameters);
 
     // Calculate regularization terms.
     const double smallRegTerm = 0.5 * wL2SquaredNorm;
@@ -123,8 +123,8 @@ TEST_CASE("SoftmaxRegressionFunctionGrad
 
   // 2 objects for 2 terms in the cost function. Each term contributes towards
   // the gradient and thus need to be checked independently.
-  SoftmaxRegressionFunction srf1(data, labels, numClasses, 0);
-  SoftmaxRegressionFunction srf2(data, labels, numClasses, 20);
+  SoftmaxRegressionFunction<> srf1(data, labels, numClasses, 0);
+  SoftmaxRegressionFunction<> srf2(data, labels, numClasses, 20);
 
   // Create a random set of parameters.
   arma::mat parameters;
@@ -195,7 +195,7 @@ TEST_CASE("SoftmaxRegressionTwoClasses",
   }
 
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
+  SoftmaxRegression<> sr(data, labels, numClasses, lambda);
 
   // Compare training accuracy to 100.
   const double acc = sr.ComputeAccuracy(data, labels);
@@ -218,28 +218,31 @@ TEST_CASE("SoftmaxRegressionTwoClasses",
   REQUIRE(testAcc == Approx(100.0).epsilon(0.02));
 }
 
-TEST_CASE("SoftmaxRegressionFitIntercept", "[SoftmaxRegressionTest]")
+TEMPLATE_TEST_CASE("SoftmaxRegressionFitIntercept", "[SoftmaxRegressionTest]",
+    arma::fmat, arma::mat)
 {
+  typedef TestType MatType;
+
   // Generate a two-Gaussian dataset,
   // which can't be separated without adding the intercept term.
   GaussianDistribution g1(arma::vec("1.0 1.0 1.0"), arma::eye<arma::mat>(3, 3));
   GaussianDistribution g2(arma::vec("9.0 9.0 9.0"), arma::eye<arma::mat>(3, 3));
 
-  arma::mat data(3, 1000);
+  MatType data(3, 1000);
   arma::Row<size_t> responses(1000);
   for (size_t i = 0; i < 500; ++i)
   {
-    data.col(i) = g1.Random();
+    data.col(i) = ConvTo<MatType>::From(g1.Random());
     responses[i] = 0;
   }
   for (size_t i = 500; i < 1000; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<MatType>::From(g2.Random());
     responses[i] = 1;
   }
 
   // Now train a logistic regression object on it.
-  SoftmaxRegression lr(data, responses, 2, 0.01, true);
+  SoftmaxRegression<MatType> lr(data, responses, 2, 0.01, true);
 
   // Ensure that the error is close to zero.
   const double acc = lr.ComputeAccuracy(data, responses);
@@ -248,12 +251,12 @@ TEST_CASE("SoftmaxRegressionFitIntercept
   // Create a test set.
   for (size_t i = 0; i < 500; ++i)
   {
-    data.col(i) = g1.Random();
+    data.col(i) = ConvTo<MatType>::From(g1.Random());
     responses[i] = 0;
   }
   for (size_t i = 500; i < 1000; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<MatType>::From(g2.Random());
     responses[i] = 1;
   }
 
@@ -262,8 +265,11 @@ TEST_CASE("SoftmaxRegressionFitIntercept
   REQUIRE(testAcc == Approx(100.0).epsilon(0.02));
 }
 
-TEST_CASE("SoftmaxRegressionMultipleClasses", "[SoftmaxRegressionTest]")
+TEMPLATE_TEST_CASE("SoftmaxRegressionMultipleClasses",
+    "[SoftmaxRegressionTest]", arma::fmat, arma::mat)
 {
+  typedef TestType MatType;
+
   const size_t points = 5000;
   const size_t inputSize = 5;
   const size_t numClasses = 5;
@@ -277,37 +283,38 @@ TEST_CASE("SoftmaxRegressionMultipleClas
   GaussianDistribution g4(arma::vec("4.0 1.0 1.0 2.0 7.0"), identity);
   GaussianDistribution g5(arma::vec("1.0 0.0 1.0 8.0 3.0"), identity);
 
-  arma::mat data(inputSize, points);
+  MatType data(inputSize, points);
   arma::Row<size_t> labels(points);
 
   for (size_t i = 0; i < points / 5; ++i)
   {
-    data.col(i) = g1.Random();
+    // TODO: when GaussianDistribution is templatized, remove the conv_to.
+    data.col(i) = ConvTo<MatType>::From(g1.Random());
     labels(i) = 0;
   }
   for (size_t i = points / 5; i < (2 * points) / 5; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<MatType>::From(g2.Random());
     labels(i) = 1;
   }
   for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
   {
-    data.col(i) = g3.Random();
+    data.col(i) = ConvTo<MatType>::From(g3.Random());
     labels(i) = 2;
   }
   for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
   {
-    data.col(i) = g4.Random();
+    data.col(i) = ConvTo<MatType>::From(g4.Random());
     labels(i) = 3;
   }
   for (size_t i = (4 * points) / 5; i < points; ++i)
   {
-    data.col(i) = g5.Random();
+    data.col(i) = ConvTo<MatType>::From(g5.Random());
     labels(i) = 4;
   }
 
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
+  SoftmaxRegression<MatType> sr(data, labels, numClasses, lambda);
 
   // Compare training accuracy to 100.
   const double acc = sr.ComputeAccuracy(data, labels);
@@ -316,27 +323,27 @@ TEST_CASE("SoftmaxRegressionMultipleClas
   // Create test dataset.
   for (size_t i = 0; i < points / 5; ++i)
   {
-    data.col(i) = g1.Random();
+    data.col(i) = ConvTo<MatType>::From(g1.Random());
     labels(i) = 0;
   }
   for (size_t i = points / 5; i < (2 * points) / 5; ++i)
   {
-    data.col(i) = g2.Random();
+    data.col(i) = ConvTo<MatType>::From(g2.Random());
     labels(i) = 1;
   }
   for (size_t i = (2 * points) / 5; i < (3 * points) / 5; ++i)
   {
-    data.col(i) = g3.Random();
+    data.col(i) = ConvTo<MatType>::From(g3.Random());
     labels(i) = 2;
   }
   for (size_t i = (3 * points) / 5; i < (4 * points) / 5; ++i)
   {
-    data.col(i) = g4.Random();
+    data.col(i) = ConvTo<MatType>::From(g4.Random());
     labels(i) = 3;
   }
   for (size_t i = (4 * points) / 5; i < points; ++i)
   {
-    data.col(i) = g5.Random();
+    data.col(i) = ConvTo<MatType>::From(g5.Random());
     labels(i) = 4;
   }
 
@@ -355,12 +362,12 @@ TEST_CASE("SoftmaxRegressionTrainTest",
   for (size_t i = 500; i < 1000; ++i)
     labels[i] = size_t(1.0);
 
-  SoftmaxRegression sr(dataset.n_rows, 2);
-  SoftmaxRegression sr2(dataset.n_rows, 2);
+  SoftmaxRegression<> sr(dataset.n_rows, 2);
+  SoftmaxRegression<> sr2(dataset.n_rows, 2);
   sr.Parameters() = sr2.Parameters();
   ens::L_BFGS lbfgs;
-  sr.Train(dataset, labels, 2, std::move(lbfgs));
-  sr2.Train(dataset, labels, 2, std::move(lbfgs));
+  sr.Train(dataset, labels, 2, lbfgs);
+  sr2.Train(dataset, labels, 2, lbfgs);
 
   // Ensure that the parameters are the same.
   REQUIRE(sr.Parameters().n_rows == sr2.Parameters().n_rows);
@@ -386,10 +393,10 @@ TEST_CASE("SoftmaxRegressionOptimizerTra
     labels[i] = size_t(1.0);
 
   ens::L_BFGS lbfgs;
-  SoftmaxRegression sr(dataset.n_rows, 2, true);
+  SoftmaxRegression<> sr(dataset.n_rows, 2, true);
 
   ens::L_BFGS lbfgs2;
-  SoftmaxRegression sr2(dataset.n_rows, 2, true);
+  SoftmaxRegression<> sr2(dataset.n_rows, 2, true);
 
   sr.Lambda() = sr2.Lambda() = 0.01;
   sr.Parameters() = sr2.Parameters();
@@ -456,7 +463,7 @@ TEST_CASE("SoftmaxRegressionClassifySing
   }
 
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
+  SoftmaxRegression<> sr(data, labels, numClasses, lambda);
 
   // Create test dataset.
   for (size_t i = 0; i < points / 5; ++i)
@@ -539,7 +546,7 @@ TEST_CASE("SoftmaxRegressionComputeProba
   }
 
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
+  SoftmaxRegression<> sr(data, labels, numClasses, lambda);
 
   // Create test dataset.
   for (size_t i = 0; i < points / 5; ++i)
@@ -568,17 +575,29 @@ TEST_CASE("SoftmaxRegressionComputeProba
     labels(i) = 4;
   }
 
+  arma::Row<size_t> predictions;
   arma::mat probabilities;
-  sr.Classify(data, probabilities);
+  sr.Classify(data, predictions, probabilities);
 
+  REQUIRE(predictions.n_elem == data.n_cols);
   REQUIRE(probabilities.n_cols == data.n_cols);
   REQUIRE(probabilities.n_rows == sr.NumClasses());
 
   for (size_t i = 0; i < data.n_cols; ++i)
   {
-    REQUIRE(arma::sum(probabilities.col(i)) ==
-        Approx(1.0).epsilon(1e-7));
+    REQUIRE(predictions[i] < numClasses);
+    REQUIRE(sum(probabilities.col(i)) == Approx(1.0).epsilon(1e-7));
   }
+
+  // Test Classify() on a single point.
+  size_t prediction = sr.Classify(data.col(0));
+  REQUIRE(prediction == predictions[0]);
+
+  arma::vec probabilitiesVec;
+  sr.Classify(data.col(0), prediction, probabilitiesVec);
+  REQUIRE(prediction == predictions[0]);
+  REQUIRE(arma::approx_equal(probabilities.col(0), probabilitiesVec, "absdiff",
+      1e-5));
 }
 
 TEST_CASE("SoftmaxRegressionComputeProbabilitiesAndLabelsTest",
@@ -627,7 +646,7 @@ TEST_CASE("SoftmaxRegressionComputeProba
   }
 
   // Train softmax regression object.
-  SoftmaxRegression sr(data, labels, numClasses, lambda);
+  SoftmaxRegression<> sr(data, labels, numClasses, lambda);
 
   // Create test dataset.
   for (size_t i = 0; i < points / 5; ++i)
@@ -667,8 +686,7 @@ TEST_CASE("SoftmaxRegressionComputeProba
 
   for (size_t i = 0; i < data.n_cols; ++i)
   {
-    REQUIRE(arma::sum(probabilities.col(i)) ==
-        Approx(1.0).epsilon(1e-7));
+    REQUIRE(sum(probabilities.col(i)) == Approx(1.0).epsilon(1e-7));
     REQUIRE(testLabels(i) == labels(i));
   }
 }
@@ -688,7 +706,7 @@ TEST_CASE("SoftmaxImmediateTrainTest", "
     labels(i) = RandInt(0, numClasses);
 
   // Train without setting any parameters to the constructor.
-  SoftmaxRegression sr;
+  SoftmaxRegression<> sr;
   sr.Train(data, labels, numClasses);
 
   // Now classify some points.
@@ -701,3 +719,191 @@ TEST_CASE("SoftmaxImmediateTrainTest", "
   REQUIRE(arma::all(predictions >= 0));
   REQUIRE(arma::all(predictions <= 2));
 }
+
+// Test variants of constructor.  This test is more about checking that all
+// variants compile correctly than anything else.
+TEMPLATE_TEST_CASE("SoftmaxRegressionConstructorVariantTest",
+    "[SoftmaxRegressionTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels =
+      arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+
+  // Empty constructor.
+  SoftmaxRegression<MatType> sr1;
+
+  // No hyperparameters.
+  SoftmaxRegression<MatType> sr2(data, labels, 4);
+
+  // Specify hyperparameters only.
+  SoftmaxRegression<MatType> sr3(data, labels, 4, 0.001, true);
+
+  // Specify hyperparameters and a callback.
+  SoftmaxRegression<MatType> sr4(data, labels, 4, 0.002, true,
+      ens::EarlyStopAtMinLoss());
+
+  // Specify hyperparameters and two callbacks.
+  SoftmaxRegression<MatType> sr5(data, labels, 4, 0.003, true,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1000.0));
+
+  // Specify hyperparameters and optimizer.
+  ens::StandardSGD sgd(0.01);
+  SoftmaxRegression<MatType> sr6(data, labels, 4, sgd, 0.004, true);
+
+  // Specify hyperparameters, optimizer, and a callback.
+  SoftmaxRegression<MatType> sr7(data, labels, 4, sgd, 0.005, true,
+      ens::EarlyStopAtMinLoss());
+
+  // Specify hyperparameters, optimizer, and two callbacks.
+  SoftmaxRegression<MatType> sr8(data, labels, 4, sgd, 0.006, true,
+      ens::EarlyStopAtMinLoss(), ens::TimerStop(1000.0));
+
+  // Now we don't care what the training call actually produced, but we do care
+  // that the model trained ate all and has the right size (except for the first
+  // one).
+  REQUIRE(sr1.Parameters().n_elem == 0);
+
+  REQUIRE(sr2.Parameters().n_rows == 4);
+  REQUIRE(sr2.Parameters().n_cols == 51);
+
+  REQUIRE(sr3.Parameters().n_rows == 4);
+  REQUIRE(sr3.Parameters().n_cols == 51);
+  REQUIRE(sr3.Lambda() == Approx(0.001));
+  REQUIRE(sr3.FitIntercept() == true);
+
+  REQUIRE(sr4.Parameters().n_rows == 4);
+  REQUIRE(sr4.Parameters().n_cols == 51);
+  REQUIRE(sr4.Lambda() == Approx(0.002));
+  REQUIRE(sr4.FitIntercept() == true);
+
+  REQUIRE(sr5.Parameters().n_rows == 4);
+  REQUIRE(sr5.Parameters().n_cols == 51);
+  REQUIRE(sr5.Lambda() == Approx(0.003));
+  REQUIRE(sr5.FitIntercept() == true);
+
+  REQUIRE(sr6.Parameters().n_rows == 4);
+  REQUIRE(sr6.Parameters().n_cols == 51);
+  REQUIRE(sr6.Lambda() == Approx(0.004));
+  REQUIRE(sr6.FitIntercept() == true);
+
+  REQUIRE(sr7.Parameters().n_rows == 4);
+  REQUIRE(sr7.Parameters().n_cols == 51);
+  REQUIRE(sr7.Lambda() == Approx(0.005));
+  REQUIRE(sr7.FitIntercept() == true);
+
+  REQUIRE(sr8.Parameters().n_rows == 4);
+  REQUIRE(sr8.Parameters().n_cols == 51);
+  REQUIRE(sr8.Lambda() == Approx(0.006));
+  REQUIRE(sr8.FitIntercept() == true);
+}
+
+// Test variants of Train().  This test is more about checking that all variants
+// compile correctly than anything else.
+TEMPLATE_TEST_CASE("SoftmaxRegressionTrainVariantTest",
+    "[SoftmaxRegressionTest]", arma::fmat, arma::mat)
+{
+  typedef TestType MatType;
+
+  // Create random data.
+  MatType data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels =
+      arma::randi<arma::Row<size_t>>(1000, arma::distr_param(0, 3));
+
+  // Create objects that we will use.
+  SoftmaxRegression<MatType> sr1, sr2, sr3, sr4, sr5, sr6, sr7, sr8;
+
+  // No hyperparameters.
+  sr1.Train(data, labels, 4);
+
+  // Specify hyperparameters only.
+  sr2.Train(data, labels, 4, 0.001, false);
+
+  // Specify hyperparameters and a callback.
+  sr3.Train(data, labels, 4, 0.002, false, ens::EarlyStopAtMinLoss());
+
+  // Specify hyperparameters and two callbacks.
+  sr4.Train(data, labels, 4, 0.003, false, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1000.0));
+
+  // Specify hyperparameters and an optimizer.
+  ens::AdaDelta adaDelta(0.01);
+  sr5.Train(data, labels, 4, adaDelta, 0.004, true);
+
+  // Specify hyperparameters, an optimizer, and a callback.
+  sr6.Train(data, labels, 4, adaDelta, 0.005, true, ens::EarlyStopAtMinLoss());
+
+  // Specify hyperparameters, an optimizer, and two callbacks.
+  sr7.Train(data, labels, 4, adaDelta, 0.006, true, ens::EarlyStopAtMinLoss(),
+      ens::TimerStop(1000.0));
+
+  // Now we don't care what the training actually produced, but we do want to
+  // make sure that the model trained at all and has the right size.
+  REQUIRE(sr1.Parameters().n_rows == 4);
+  REQUIRE(sr1.Parameters().n_cols == 51);
+
+  REQUIRE(sr2.Parameters().n_rows == 4);
+  REQUIRE(sr2.Parameters().n_cols == 50);
+  REQUIRE(sr2.Lambda() == Approx(0.001));
+  REQUIRE(sr2.FitIntercept() == false);
+
+  REQUIRE(sr3.Parameters().n_rows == 4);
+  REQUIRE(sr3.Parameters().n_cols == 50);
+  REQUIRE(sr3.Lambda() == Approx(0.002));
+  REQUIRE(sr3.FitIntercept() == false);
+
+  REQUIRE(sr4.Parameters().n_rows == 4);
+  REQUIRE(sr4.Parameters().n_cols == 50);
+  REQUIRE(sr4.Lambda() == Approx(0.003));
+  REQUIRE(sr4.FitIntercept() == false);
+
+  REQUIRE(sr5.Parameters().n_rows == 4);
+  REQUIRE(sr5.Parameters().n_cols == 51);
+  REQUIRE(sr5.Lambda() == Approx(0.004));
+  REQUIRE(sr5.FitIntercept() == true);
+
+  REQUIRE(sr6.Parameters().n_rows == 4);
+  REQUIRE(sr6.Parameters().n_cols == 51);
+  REQUIRE(sr6.Lambda() == Approx(0.005));
+  REQUIRE(sr6.FitIntercept() == true);
+
+  REQUIRE(sr7.Parameters().n_rows == 4);
+  REQUIRE(sr7.Parameters().n_cols == 51);
+  REQUIRE(sr7.Lambda() == Approx(0.006));
+  REQUIRE(sr7.FitIntercept() == true);
+}
+
+// Make sure resetting a model does something.
+TEST_CASE("SoftmaxRegressionResetTest", "[SoftmaxRegressionTest]")
+{
+  // Create random data.
+  arma::mat data(50, 1000, arma::fill::randu);
+  arma::Row<size_t> labels(1000, arma::fill::zeros);
+  labels.subvec(500, 999).fill(1);
+
+  // Create two logistic regression models.
+  SoftmaxRegression<> sr1, sr2;
+
+  // Initialize models to zeros.
+  sr1.Parameters() = arma::zeros<arma::mat>(2, 51);
+  sr2.Parameters() = arma::zeros<arma::mat>(2, 51);
+
+  ens::L_BFGS lbfgs1(1, 1); // Use only one iteration.
+  ens::L_BFGS lbfgs2(1, 1);
+
+  sr1.Train(data, labels, 2, lbfgs1);
+  sr2.Train(data, labels, 2, lbfgs2);
+
+  REQUIRE(
+      arma::approx_equal(sr1.Parameters(), sr2.Parameters(), "absdiff", 1e-5));
+
+  // Now reset one model and incrementally train the other.
+  sr1.Reset();
+  sr1.Train(data, labels, 2, lbfgs1);
+  sr2.Train(data, labels, 2, lbfgs2);
+
+  REQUIRE(
+      !arma::approx_equal(sr1.Parameters(), sr2.Parameters(), "absdiff", 1e-5));
+}
diff -pruN 4.3.0-2/src/mlpack/tests/sparse_autoencoder_test.cpp 4.4.0-1/src/mlpack/tests/sparse_autoencoder_test.cpp
--- 4.3.0-2/src/mlpack/tests/sparse_autoencoder_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/sparse_autoencoder_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -91,14 +91,14 @@ TEST_CASE("SparseAutoencoderFunctionRand
       arma::mat hiddenLayer, outputLayer, diff;
 
       hiddenLayer = 1.0 /
-          (1 + arma::exp(-(parameters.submat(0, 0, l1 - 1, l2 - 1) *
+          (1 + exp(-(parameters.submat(0, 0, l1 - 1, l2 - 1) *
           data.col(j) + parameters.submat(0, l2, l1 - 1, l2))));
       outputLayer = 1.0 /
-          (1 + arma::exp(-(parameters.submat(l1, 0, l3 - 1, l2 - 1).t()
+          (1 + exp(-(parameters.submat(l1, 0, l3 - 1, l2 - 1).t()
           * hiddenLayer + parameters.submat(l3, 0, l3, l2 - 1).t())));
       diff = outputLayer - data.col(j);
 
-      reconstructionError += 0.5 * arma::sum(arma::sum(diff % diff));
+      reconstructionError += 0.5 * sum(sum(diff % diff));
     }
     reconstructionError /= points;
 
@@ -136,7 +136,7 @@ TEST_CASE("SparseAutoencoderFunctionRegu
 
     double wL2SquaredNorm;
 
-    wL2SquaredNorm = arma::accu(parameters.submat(0, 0, l3 - 1, l2 - 1) %
+    wL2SquaredNorm = accu(parameters.submat(0, 0, l3 - 1, l2 - 1) %
         parameters.submat(0, 0, l3 - 1, l2 - 1));
 
     // Calculate regularization terms.
@@ -188,17 +188,17 @@ TEST_CASE("SparseAutoencoderFunctionKLDi
       arma::mat hiddenLayer;
 
       hiddenLayer = 1.0 / (1 +
-          arma::exp(-(parameters.submat(0, 0, l1 - 1, l2 - 1) *
+          exp(-(parameters.submat(0, 0, l1 - 1, l2 - 1) *
           data.col(j) + parameters.submat(0, l2, l1 - 1, l2))));
       rhoCap += hiddenLayer;
     }
     rhoCap /= points;
 
     // Calculate divergence terms.
-    const double smallDivTerm = 5 * arma::accu(rho * arma::log(rho / rhoCap) +
-        (1 - rho) * arma::log((1 - rho) / (1 - rhoCap)));
-    const double bigDivTerm = 20 * arma::accu(rho * arma::log(rho / rhoCap) +
-        (1 - rho) * arma::log((1 - rho) / (1 - rhoCap)));
+    const double smallDivTerm = 5 * accu(rho * log(rho / rhoCap) +
+        (1 - rho) * log((1 - rho) / (1 - rhoCap)));
+    const double bigDivTerm = 20 * accu(rho * log(rho / rhoCap) +
+        (1 - rho) * log((1 - rho) / (1 - rhoCap)));
 
     REQUIRE(safNoDiv.Evaluate(parameters) + smallDivTerm ==
         Approx(safSmallDiv.Evaluate(parameters)).epsilon(1e-7));
diff -pruN 4.3.0-2/src/mlpack/tests/sparse_coding_test.cpp 4.4.0-1/src/mlpack/tests/sparse_coding_test.cpp
--- 4.3.0-2/src/mlpack/tests/sparse_coding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/sparse_coding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -8,10 +8,6 @@
  * 3-clause BSD license along with mlpack.  If not, see
  * http://www.opensource.org/licenses/BSD-3-Clause for more information.
  */
-
-// Note: We don't use BOOST_REQUIRE_CLOSE in the code below because we need
-// to use FPC_WEAK, and it's not at all intuitive how to do that.
-
 #include <mlpack/core.hpp>
 #include <mlpack/methods/sparse_coding.hpp>
 
@@ -22,9 +18,13 @@
 using namespace arma;
 using namespace mlpack;
 
-void SCVerifyCorrectness(vec beta, vec errCorr, double lambda)
+template<typename VecType>
+void SCVerifyCorrectness(const VecType& beta,
+                         const VecType& errCorr,
+                         double lambda)
 {
-  const double tol = 1e-12;
+  const double tol = std::is_same<typename VecType::elem_type, float>::value ?
+      1e-6 : 1e-12;
   size_t nDims = beta.n_elem;
   for (size_t j = 0; j < nDims; ++j)
   {
@@ -47,13 +47,18 @@ void SCVerifyCorrectness(vec beta, vec e
   }
 }
 
-TEST_CASE("SparseCodingTestCodingStepLasso", "[SparseCodingTest]")
+TEMPLATE_TEST_CASE("SparseCodingTestCodingStepLasso", "[SparseCodingTest]",
+    arma::mat, arma::fmat)
 {
+  typedef TestType MatType;
+  typedef arma::Col<typename MatType::elem_type> VecType;
+
   double lambda1 = 0.1;
   uword nAtoms = 25;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  arma::mat inX; // The .arm file contains an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // Normalize each point since these are images.
@@ -62,72 +67,82 @@ TEST_CASE("SparseCodingTestCodingStepLas
     X.col(i) /= norm(X.col(i), 2);
   }
 
-  SparseCoding sc(nAtoms, lambda1);
-  mat Z;
+  SparseCoding<MatType> sc(nAtoms, lambda1);
+  MatType Z;
   DataDependentRandomInitializer::Initialize(X, 25, sc.Dictionary());
   sc.Encode(X, Z);
 
-  mat D = sc.Dictionary();
+  MatType D = sc.Dictionary();
 
   for (uword i = 0; i < nPoints; ++i)
   {
-    vec errCorr = trans(D) * (D * Z.unsafe_col(i) - X.unsafe_col(i));
+    VecType errCorr = trans(D) * (D * Z.unsafe_col(i) - X.unsafe_col(i));
     SCVerifyCorrectness(Z.unsafe_col(i), errCorr, lambda1);
   }
 }
 
-TEST_CASE("SparseCodingTestCodingStepElasticNet", "[SparseCodingTest]")
+TEMPLATE_TEST_CASE("SparseCodingTestCodingStepElasticNet", "[SparseCodingTest]",
+    arma::mat, arma::fmat)
 {
+  typedef TestType MatType;
+  typedef arma::Col<typename MatType::elem_type> VecType;
+
   double lambda1 = 0.1;
   double lambda2 = 0.2;
   uword nAtoms = 25;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  arma::mat inX; // The .arm file contains an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // Normalize each point since these are images.
   for (uword i = 0; i < nPoints; ++i)
     X.col(i) /= norm(X.col(i), 2);
 
-  SparseCoding sc(nAtoms, lambda1, lambda2);
-  mat Z;
+  SparseCoding<MatType> sc(nAtoms, lambda1, lambda2);
+  MatType Z;
   DataDependentRandomInitializer::Initialize(X, 25, sc.Dictionary());
   sc.Encode(X, Z);
 
-  mat D = sc.Dictionary();
+  MatType D = sc.Dictionary();
 
   for (uword i = 0; i < nPoints; ++i)
   {
-    vec errCorr =
-      (trans(D) * D + lambda2 * eye(nAtoms, nAtoms)) * Z.unsafe_col(i)
-      - trans(D) * X.unsafe_col(i);
+    VecType errCorr =
+        (trans(D) * D + lambda2 * eye<MatType>(nAtoms, nAtoms)) *
+        Z.unsafe_col(i) - trans(D) * X.unsafe_col(i);
 
     SCVerifyCorrectness(Z.unsafe_col(i), errCorr, lambda1);
   }
 }
 
-TEST_CASE("SparseCodingTestDictionaryStep", "[SparseCodingTest]")
+TEMPLATE_TEST_CASE("SparseCodingTestDictionaryStep", "[SparseCodingTest]",
+    arma::mat, arma::fmat)
 {
-  const double tol = 1e-6;
+  typedef TestType MatType;
+
+  const double tol = std::is_same<typename MatType::elem_type, float>::value ?
+      0.01 : 1e-6;
 
   double lambda1 = 0.1;
   uword nAtoms = 25;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  arma::mat inX; // The .arm file contains an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // Normalize each point since these are images.
   for (uword i = 0; i < nPoints; ++i)
     X.col(i) /= norm(X.col(i), 2);
 
-  SparseCoding sc(nAtoms, lambda1, 0.0, 0, 0.01, tol);
-  mat Z;
+  SparseCoding<MatType> sc(nAtoms, lambda1, 0.0, 0, 0.01, tol);
+  MatType Z;
   DataDependentRandomInitializer::Initialize(X, 25, sc.Dictionary());
   sc.Encode(X, Z);
 
-  mat D = sc.Dictionary();
+  MatType D = sc.Dictionary();
 
   uvec adjacencies = find(Z);
   double normGradient = sc.OptimizeDictionary(X, Z, adjacencies);
@@ -135,25 +150,28 @@ TEST_CASE("SparseCodingTestDictionarySte
   REQUIRE(normGradient == Approx(0.0).margin(tol));
 }
 
-TEST_CASE("SerializationTest", "[SparseCodingTest]")
+TEMPLATE_TEST_CASE("SerializationTest", "[SparseCodingTest]", arma::mat,
+    arma::fmat)
 {
-  mat X = randu<mat>(100, 100);
+  typedef TestType MatType;
+
+  MatType X = randu<MatType>(100, 100);
   size_t nAtoms = 25;
 
-  SparseCoding sc(nAtoms, 0.05, 0.1);
+  SparseCoding<MatType> sc(nAtoms, 0.05, 0.1, 10);
   sc.Train(X);
 
-  mat Y = randu<mat>(100, 200);
-  mat codes;
+  MatType Y = randu<MatType>(100, 200);
+  MatType codes;
   sc.Encode(Y, codes);
 
-  SparseCoding scXml(50, 0.01), scJson(nAtoms, 0.05), scBinary(0, 0.0);
+  SparseCoding<MatType> scXml(50, 0.01), scJson(nAtoms, 0.05), scBinary(0, 0.0);
   SerializeObjectAll(sc, scXml, scJson, scBinary);
 
   CheckMatrices(sc.Dictionary(), scXml.Dictionary(), scJson.Dictionary(),
       scBinary.Dictionary());
 
-  mat xmlCodes, jsonCodes, binaryCodes;
+  MatType xmlCodes, jsonCodes, binaryCodes;
   scXml.Encode(Y, xmlCodes);
   scJson.Encode(Y, jsonCodes);
   scBinary.Encode(Y, binaryCodes);
@@ -192,22 +210,27 @@ TEST_CASE("SerializationTest", "[SparseC
 /**
  * Test that SparseCoding::Train() returns finite final objective value.
  */
-TEST_CASE("SparseCodingTrainReturnObjective", "[SparseCodingTest]")
+TEMPLATE_TEST_CASE("SparseCodingTrainReturnObjective", "[SparseCodingTest]",
+    arma::mat, arma::fmat)
 {
-  const double tol = 1e-6;
+  typedef TestType MatType;
+
+  const double tol = std::is_same<typename MatType::elem_type, float>::value ?
+      0.01 : 1e-6;
 
   double lambda1 = 0.1;
   uword nAtoms = 25;
 
-  mat X;
-  X.load("mnist_first250_training_4s_and_9s.arm");
+  arma::mat inX; // The .arm file contains an arma::mat.
+  inX.load("mnist_first250_training_4s_and_9s.csv");
+  MatType X = arma::conv_to<MatType>::from(inX);
   uword nPoints = X.n_cols;
 
   // Normalize each point since these are images.
   for (uword i = 0; i < nPoints; ++i)
     X.col(i) /= norm(X.col(i), 2);
 
-  SparseCoding sc(nAtoms, lambda1, 0.0, 0, 0.01, tol);
+  SparseCoding<MatType> sc(nAtoms, lambda1, 0.0, 0, 0.01, tol);
   double objVal = sc.Train(X);
 
   REQUIRE(std::isfinite(objVal) == true);
diff -pruN 4.3.0-2/src/mlpack/tests/split_data_test.cpp 4.4.0-1/src/mlpack/tests/split_data_test.cpp
--- 4.3.0-2/src/mlpack/tests/split_data_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/split_data_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -101,7 +101,7 @@ TEST_CASE("SplitShuffleDataResultMat", "
   REQUIRE(std::get<0>(value).n_cols == 8); // Train data.
   REQUIRE(std::get<1>(value).n_cols == 2); // Test data.
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   CheckMatEqual(input, concat);
 }
 
@@ -115,7 +115,7 @@ TEST_CASE("SplitDataResultMat", "[SplitD
   REQUIRE(std::get<0>(value).n_cols == 8); // Train data.
   REQUIRE(std::get<1>(value).n_cols == 2); // Test data.
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   // Order matters here.
   CheckMatrices(input, concat);
 }
@@ -130,7 +130,7 @@ TEST_CASE("ZeroRatioSplitData", "[SplitD
   REQUIRE(std::get<0>(value).n_cols == 10); // Train data.
   REQUIRE(std::get<1>(value).n_cols == 0); // Test data.
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   // Order matters here.
   CheckMatrices(input, concat);
 }
@@ -145,7 +145,7 @@ TEST_CASE("TotalRatioSplitData", "[Split
   REQUIRE(std::get<0>(value).n_cols == 0); // Train data.
   REQUIRE(std::get<1>(value).n_cols == 10); // Test data.
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   // Order matters here.
   CheckMatrices(input, concat);
 }
@@ -200,7 +200,7 @@ TEST_CASE("SplitDataLargerTest", "[Split
   REQUIRE(std::get<0>(value).n_cols == 497 - size_t(0.3 * 497));
   REQUIRE(std::get<1>(value).n_cols == size_t(0.3 * 497));
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   CheckMatEqual(input, concat);
 }
 
@@ -286,7 +286,7 @@ TEST_CASE("StratifiedSplitDataResultTest
   REQUIRE(static_cast<uvec>(find(std::get<3>(value) == 1)).n_rows == 2);
   REQUIRE(static_cast<uvec>(find(std::get<3>(value) == 2)).n_rows == 3);
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   CheckMatEqual(input, concat);
 }
 
@@ -315,9 +315,9 @@ TEST_CASE("StratifiedSplitLargerDataResu
   twoLabel.fill(2);
   threeLabel.fill(3);
 
-  Row<size_t> labels = arma::join_rows(zeroLabel, oneLabel);
-  labels = arma::join_rows(labels, twoLabel);
-  labels = arma::join_rows(labels, threeLabel);
+  Row<size_t> labels = join_rows(zeroLabel, oneLabel);
+  labels = join_rows(labels, twoLabel);
+  labels = join_rows(labels, threeLabel);
   const double testRatio = 0.3;
 
   const auto value = Split(input, labels, testRatio, false, true);
@@ -331,7 +331,7 @@ TEST_CASE("StratifiedSplitLargerDataResu
   REQUIRE(static_cast<uvec>(find(std::get<3>(value) == 2)).n_rows == 19);
   REQUIRE(static_cast<uvec>(find(std::get<3>(value) == 3)).n_rows == 9);
 
-  mat concat = arma::join_rows(std::get<0>(value), std::get<1>(value));
+  mat concat = join_rows(std::get<0>(value), std::get<1>(value));
   CheckMatEqual(input, concat);
 }
 
@@ -392,8 +392,8 @@ TEST_CASE("SplitMatrixLabeledData", "[Sp
   REQUIRE(std::get<2>(value).n_cols == 8);
   REQUIRE(std::get<3>(value).n_cols == 2);
 
-  mat inputConcat = arma::join_rows(std::get<0>(value), std::get<1>(value));
-  mat labelsConcat = arma::join_rows(std::get<2>(value), std::get<3>(value));
+  mat inputConcat = join_rows(std::get<0>(value), std::get<1>(value));
+  mat labelsConcat = join_rows(std::get<2>(value), std::get<3>(value));
 
   // Order matters here.
   CheckMatrices(input, inputConcat);
diff -pruN 4.3.0-2/src/mlpack/tests/string_encoding_test.cpp 4.4.0-1/src/mlpack/tests/string_encoding_test.cpp
--- 4.3.0-2/src/mlpack/tests/string_encoding_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/string_encoding_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -71,7 +71,7 @@ void CheckVectors(const vector<vector<Va
  */
 TEST_CASE("DictionaryEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   arma::mat output;
   DictionaryEncoding<SplitByAnyOf::TokenType> encoder;
@@ -108,7 +108,7 @@ TEST_CASE("DictionaryEncodingTest", "[St
  */
 TEST_CASE("UnicodeDictionaryEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   arma::mat output;
   DictionaryEncoding<SplitByAnyOf::TokenType> encoder;
@@ -142,7 +142,7 @@ TEST_CASE("UnicodeDictionaryEncodingTest
  */
 TEST_CASE("OnePassDictionaryEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   vector<vector<size_t>> output;
   DictionaryEncoding<SplitByAnyOf::TokenType> encoder(
@@ -179,10 +179,10 @@ TEST_CASE("OnePassDictionaryEncodingTest
  */
 TEST_CASE("SplitByAnyOfTokenizerTest", "[StringEncodingTest]")
 {
-  std::vector<MLPACK_STRING_VIEW> tokens;
-  MLPACK_STRING_VIEW line(stringEncodingInput[0]);
+  std::vector<std::string_view> tokens;
+  std::string_view line(stringEncodingInput[0]);
   SplitByAnyOf tokenizer(" ,.");
-  MLPACK_STRING_VIEW token = tokenizer(line);
+  std::string_view token = tokenizer(line);
 
   while (!token.empty())
   {
@@ -217,10 +217,10 @@ TEST_CASE("SplitByAnyOfTokenizerUnicodeT
     "\xE2\x93\x82\xE2\x93\x81\xE2\x93\x85\xE2\x92\xB6\xE2\x92\xB8\xE2\x93\x80"
   };
 
-  std::vector<MLPACK_STRING_VIEW> tokens;
-  MLPACK_STRING_VIEW line(stringEncodingUtf8Input[2]);
+  std::vector<std::string_view> tokens;
+  std::string_view line(stringEncodingUtf8Input[2]);
   SplitByAnyOf tokenizer(" ,.");
-  MLPACK_STRING_VIEW token = tokenizer(line);
+  std::string_view token = tokenizer(line);
   while (!token.empty())
   {
     tokens.push_back(token);
@@ -289,7 +289,7 @@ TEST_CASE("OnePassDictionaryEncodingIndi
  */
 TEST_CASE("StringEncodingCopyTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
   arma::sp_mat output;
   DictionaryEncoding<SplitByAnyOf::TokenType> encoderCopy;
   SplitByAnyOf tokenizer(" ,.");
@@ -325,7 +325,7 @@ TEST_CASE("StringEncodingCopyTest", "[St
  */
 TEST_CASE("StringEncodingMoveTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
   arma::sp_mat output;
   DictionaryEncoding<SplitByAnyOf::TokenType> encoderCopy;
   SplitByAnyOf tokenizer(" ,.");
@@ -384,21 +384,21 @@ void CheckDictionaries(const StringEncod
 
 /**
  * This is a specialization of the CheckDictionaries() function for
- * the MLPACK_STRING_VIEW token type.
+ * the std::string_view token type.
  */
 template<>
 void CheckDictionaries(
-    const StringEncodingDictionary<MLPACK_STRING_VIEW>& expected,
-    const StringEncodingDictionary<MLPACK_STRING_VIEW>& obtained)
+    const StringEncodingDictionary<std::string_view>& expected,
+    const StringEncodingDictionary<std::string_view>& obtained)
 {
   /* MapType is equal to
    *
-   * std::unordered_map<MLPACK_STRING_VIEW,
+   * std::unordered_map<std::string_view,
    *                    size_t,
-   *                    boost::hash<MLPACK_STRING_VIEW>>.
+   *                    boost::hash<std::string_view>>.
    */
   using MapType =
-      typename StringEncodingDictionary<MLPACK_STRING_VIEW>::MapType;
+      typename StringEncodingDictionary<std::string_view>::MapType;
 
   const std::deque<std::string>& expectedTokens = expected.Tokens();
   const std::deque<std::string>& tokens = obtained.Tokens();
@@ -451,9 +451,9 @@ TEST_CASE("StringEncodingDictionarySeria
 
   for (const string& line : stringEncodingInput)
   {
-    MLPACK_STRING_VIEW lineView(line);
+    std::string_view lineView(line);
 
-    MLPACK_STRING_VIEW token = tokenizer(lineView);
+    std::string_view token = tokenizer(lineView);
 
     while (!tokenizer.IsTokenEmpty(token))
     {
@@ -538,7 +538,7 @@ TEST_CASE("CharExtractDictionaryEncoding
  */
 TEST_CASE("BagOfWordsEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   arma::mat output;
   BagOfWordsEncoding<SplitByAnyOf::TokenType> encoder;
@@ -613,7 +613,7 @@ TEST_CASE("BagOfWordsEncodingTest", "[St
  */
 TEST_CASE("VectorBagOfWordsEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   vector<vector<size_t>> output;
   BagOfWordsEncoding<SplitByAnyOf::TokenType> encoder(
@@ -707,7 +707,7 @@ TEST_CASE("VectorBagOfWordsEncodingIndiv
  */
 TEST_CASE("RawCountSmoothIdfEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   arma::mat output;
   TfIdfEncoding<SplitByAnyOf::TokenType> encoder;
@@ -809,7 +809,7 @@ TEST_CASE("RawCountSmoothIdfEncodingTest
  */
 TEST_CASE("VectorRawCountSmoothIdfEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   vector<vector<double>> output;
   TfIdfEncoding<SplitByAnyOf::TokenType> encoder(
@@ -968,7 +968,7 @@ TEST_CASE("VectorRawCountSmoothIdfEncodi
  */
 TEST_CASE("TfIdfRawCountEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   arma::mat output;
   TfIdfEncoding<SplitByAnyOf::TokenType> encoder(
@@ -1018,7 +1018,7 @@ TEST_CASE("TfIdfRawCountEncodingTest", "
  */
 TEST_CASE("VectorTfIdfRawCountEncodingTest", "[StringEncodingTest]")
 {
-  using DictionaryType = StringEncodingDictionary<MLPACK_STRING_VIEW>;
+  using DictionaryType = StringEncodingDictionary<std::string_view>;
 
   vector<vector<double>> output;
   TfIdfEncoding<SplitByAnyOf::TokenType>
diff -pruN 4.3.0-2/src/mlpack/tests/svdplusplus_test.cpp 4.4.0-1/src/mlpack/tests/svdplusplus_test.cpp
--- 4.3.0-2/src/mlpack/tests/svdplusplus_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/svdplusplus_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -76,7 +76,7 @@ TEST_CASE("SVDPlusPlusEvaluate", "[SVDPl
       userVec += parameters.col(user).subvec(0, rank - 1);
 
       double ratingError = rating - userBias - itemBias -
-          arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+          dot(userVec, parameters.col(item).subvec(0, rank - 1));
       double ratingErrorSquared = ratingError * ratingError;
 
       cost += ratingErrorSquared;
@@ -145,7 +145,7 @@ TEST_CASE("SVDPlusPlusFunctionRegulariza
       {
         if (implicitVecsNormSquare(it.row()) < 0)
         {
-          implicitVecsNormSquare(it.row()) = arma::dot(
+          implicitVecsNormSquare(it.row()) = dot(
             parameters.col(implicitStart + it.row()).subvec(0, rank - 1),
             parameters.col(implicitStart + it.row()).subvec(0, rank - 1));
         }
@@ -382,7 +382,7 @@ TEST_CASE("SVDPlusPlusFunctionOptimize",
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     data(2, i) = userBias + itemBias +
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
   }
 
   // Make the SVD++ function and the optimizer.
@@ -421,7 +421,7 @@ TEST_CASE("SVDPlusPlusFunctionOptimize",
     userVec += optParameters.col(user).subvec(0, rank - 1);
 
     predictedData(0, i) = userBias + itemBias +
-        arma::dot(userVec, optParameters.col(item).subvec(0, rank - 1));
+        dot(userVec, optParameters.col(item).subvec(0, rank - 1));
   }
 
   // Calculate relative error.
@@ -489,7 +489,7 @@ TEST_CASE("SVDPlusPlusFunctionParallelOp
     userVec += parameters.col(user).subvec(0, rank - 1);
 
     data(2, i) = userBias + itemBias +
-        arma::dot(userVec, parameters.col(item).subvec(0, rank - 1));
+        dot(userVec, parameters.col(item).subvec(0, rank - 1));
   }
 
   // Make the SVD++ function and the optimizer.
@@ -535,7 +535,7 @@ TEST_CASE("SVDPlusPlusFunctionParallelOp
     userVec += optParameters.col(user).subvec(0, rank - 1);
 
     predictedData(0, i) = userBias + itemBias +
-        arma::dot(userVec, optParameters.col(item).subvec(0, rank - 1));
+        dot(userVec, optParameters.col(item).subvec(0, rank - 1));
   }
 
   // Calculate relative error.
diff -pruN 4.3.0-2/src/mlpack/tests/test_catch_tools.hpp 4.4.0-1/src/mlpack/tests/test_catch_tools.hpp
--- 4.3.0-2/src/mlpack/tests/test_catch_tools.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/test_catch_tools.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -21,15 +21,52 @@
 #define REQUIRE_RELATIVE_ERR(L, R, E) \
     REQUIRE(std::abs((R) - (L)) <= (E) * std::abs(R))
 
+// Simple wrapper class to prevent copies of Armadillo matrices.
+template <typename MatType, typename ElemType = typename MatType::elem_type,
+          typename = std::enable_if_t<std::is_same<ElemType,
+              typename MatType::elem_type>::value>>
+class MatProxy
+{
+public:
+  explicit MatProxy(const MatType& mat) : mat(mat) { }
+  
+  const arma::Mat<ElemType> mat;
+  arma::uword n_elem() const { return mat.n_elem; }
+  arma::uword n_rows() const { return mat.n_rows; }
+  arma::uword n_cols() const { return mat.n_cols; }
+  ElemType operator[](const size_t i) const { return mat[i]; }
+};
+
+template <typename ElemType>
+class MatProxy<arma::Mat<ElemType>, ElemType>
+{
+ public:
+  explicit MatProxy(const arma::Mat<ElemType>& mat) : mat(mat) { }
+
+  const arma::Mat<ElemType>& mat;
+  arma::uword n_elem() const { return mat.n_elem; }
+  arma::uword n_rows() const { return mat.n_rows; }
+  arma::uword n_cols() const { return mat.n_cols; }
+  ElemType operator[](const size_t i) const { return mat[i]; }
+};
+
 // Check the values of two matrices.
-inline void CheckMatrices(const arma::mat& a,
-                          const arma::mat& b,
+template <typename MatTypeA, typename MatTypeB,
+    typename = std::enable_if_t<arma::is_arma_type<MatTypeA>::value
+        && arma::is_arma_type<MatTypeB>::value
+        && std::is_same<typename MatTypeA::elem_type,
+                        typename MatTypeB::elem_type>::value
+        && !std::is_integral<typename MatTypeA::elem_type>::value>>
+inline void CheckMatrices(const MatTypeA& _a,
+                          const MatTypeB& _b,
                           double tolerance = 1e-5)
 {
-  REQUIRE(a.n_rows == b.n_rows);
-  REQUIRE(a.n_cols == b.n_cols);
+  MatProxy<MatTypeA> a(_a);
+  MatProxy<MatTypeB> b(_b);
+  REQUIRE(a.n_rows() == b.n_rows());
+  REQUIRE(a.n_cols() == b.n_cols());
 
-  for (size_t i = 0; i < a.n_elem; ++i)
+  for (size_t i = 0; i < a.n_elem(); ++i)
   {
     if (std::abs(a[i]) < tolerance / 2)
       REQUIRE(b[i] == Approx(0.0).margin(tolerance / 2));
@@ -63,16 +100,57 @@ inline void CheckFields(const FieldType&
     CheckMatrices(a(i), b(i));
 }
 
+
+// Simple wrapper class to prevent copies of Armadillo cubes.
+template <typename CubeType, typename ElemType = typename CubeType::elem_type,
+    typename = std::enable_if_t<std::is_same<ElemType,
+        typename CubeType::elem_type>::value>>
+class CubeProxy
+{
+ public:
+  explicit CubeProxy(const CubeType& cube) : cube(cube) { }
+
+  const arma::Cube<ElemType> cube;
+  arma::uword n_elem() const { return cube.n_elem; }
+  arma::uword n_rows() const { return cube.n_rows; }
+  arma::uword n_cols() const { return cube.n_cols; }
+  arma::uword n_slices() const { return cube.n_slices; }
+  ElemType operator[](const size_t i) const { return cube[i]; }
+};
+
+template <typename ElemType>
+class CubeProxy<arma::Cube<ElemType>, ElemType>
+{
+ public:
+  explicit CubeProxy(const arma::Cube<ElemType>& cube) : cube(cube) { }
+
+  const arma::Cube<ElemType>& cube;
+  arma::uword n_elem() const { return cube.n_elem; }
+  arma::uword n_rows() const { return cube.n_rows; }
+  arma::uword n_cols() const { return cube.n_cols; }
+  arma::uword n_slices() const { return cube.n_slices; }
+  ElemType operator[](const size_t i) const { return cube[i]; }
+};
+
 // Check the values of two cubes.
-inline void CheckMatrices(const arma::cube& a,
-                          const arma::cube& b,
+template <typename CubeTypeA, typename CubeTypeB,
+    typename = std::enable_if_t<arma::is_arma_cube_type<CubeTypeA>::value
+        && arma::is_arma_cube_type<CubeTypeB>::value
+        && std::is_same<typename CubeTypeA::elem_type,
+                        typename CubeTypeB::elem_type>::value
+        && !std::is_integral<typename CubeTypeA::elem_type>::value>,
+    typename = void>
+inline void CheckMatrices(const CubeTypeA& _a,
+                          const CubeTypeB& _b,
                           double tolerance = 1e-5)
 {
-  REQUIRE(a.n_rows == b.n_rows);
-  REQUIRE(a.n_cols == b.n_cols);
-  REQUIRE(a.n_slices == b.n_slices);
+  CubeProxy<CubeTypeA> a(_a);
+  CubeProxy<CubeTypeB> b(_b);
+  REQUIRE(a.n_rows() == b.n_rows());
+  REQUIRE(a.n_cols() == b.n_cols());
+  REQUIRE(a.n_slices() == b.n_slices());
 
-  for (size_t i = 0; i < a.n_elem; ++i)
+  for (size_t i = 0; i < a.n_elem(); ++i)
   {
     if (std::abs(a[i]) < tolerance / 2)
       REQUIRE(b[i] == Approx(0.0).margin(tolerance / 2));
diff -pruN 4.3.0-2/src/mlpack/tests/test_function_tools.hpp 4.4.0-1/src/mlpack/tests/test_function_tools.hpp
--- 4.3.0-2/src/mlpack/tests/test_function_tools.hpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/test_function_tools.hpp	2024-05-28 20:08:34.000000000 +0000
@@ -30,11 +30,11 @@ using namespace mlpack;
  * @param shuffledResponses Matrix object to store the shuffled responses into.
  */
 inline void LogisticRegressionTestData(arma::mat& data,
-                                arma::mat& testData,
-                                arma::mat& shuffledData,
-                                arma::Row<size_t>& responses,
-                                arma::Row<size_t>& testResponses,
-                                arma::Row<size_t>& shuffledResponses)
+                                       arma::mat& testData,
+                                       arma::mat& shuffledData,
+                                       arma::Row<size_t>& responses,
+                                       arma::Row<size_t>& testResponses,
+                                       arma::Row<size_t>& shuffledResponses)
 {
   // Generate a two-Gaussian dataset.
   GaussianDistribution g1(arma::vec("1.0 1.0 1.0"), arma::eye<arma::mat>(3, 3));
@@ -79,15 +79,15 @@ inline void LogisticRegressionTestData(a
   }
 }
 
-template<typename MatType>
+template<typename MatType, typename ResponsesType>
 void LoadBostonHousingDataset(MatType& trainData,
                               MatType& testData,
-                              arma::rowvec& trainResponses,
-                              arma::rowvec& testResponses,
+                              ResponsesType& trainResponses,
+                              ResponsesType& testResponses,
                               data::DatasetInfo& info)
 {
   MatType dataset;
-  arma::rowvec responses;
+  ResponsesType responses;
 
   // Defining categorical deimensions.
   info.SetDimensionality(13);
@@ -103,11 +103,11 @@ void LoadBostonHousingDataset(MatType& t
       trainResponses, testResponses, 0.3);
 }
 
-inline double RMSE(const arma::Row<double>& predictions,
-                   const arma::Row<double>& trueResponses)
+template<typename ElemType>
+inline ElemType RMSE(const arma::Row<ElemType>& predictions,
+                     const arma::Row<ElemType>& trueResponses)
 {
-  double mse = arma::accu(arma::square(predictions - trueResponses)) /
-      predictions.n_elem;
+  ElemType mse = accu(square(predictions - trueResponses)) / predictions.n_elem;
   return sqrt(mse);
 }
 
diff -pruN 4.3.0-2/src/mlpack/tests/tree_test.cpp 4.4.0-1/src/mlpack/tests/tree_test.cpp
--- 4.3.0-2/src/mlpack/tests/tree_test.cpp	2023-11-27 18:27:20.000000000 +0000
+++ 4.4.0-1/src/mlpack/tests/tree_test.cpp	2024-05-28 20:08:34.000000000 +0000
@@ -1451,7 +1451,7 @@ bool CheckHyperplaneSplit(const TreeType
     success = true;
     for (size_t k = 0; k < tree.Count(); ++k)
     {
-      ElemType result = arma::dot(mat.col(k), x);
+      ElemType result = dot(mat.col(k), x);
       if (result > 0)
       {
         x -= mat.col(k) * delta;
@@ -1861,7 +1861,7 @@ void CheckCovering(const TreeType& node)
 TEST_CASE("SimpleCoverTreeConstructionTest", "[TreeTest]")
 {
   // 20-point dataset.
-  arma::mat data = arma::trans(arma::mat("0.0 0.0;"
+  arma::mat data = trans(arma::mat("0.0 0.0;"
                                          "1.0 0.0;"
                                          "0.5 0.5;"
                                          "2.0 2.0;"
